// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

// from Common
#include "MAdMessage.h"

#include "AdaptInterface.h"
#include "MeshDataBaseInterface.h"
#include "MAdResourceManager.h"
#include "MAdTimeManager.h"
#include "NodalDataManager.h"
#include "MAdStatistics.h"
#include "ModelConstraintManager.h"
#include "History.h"
#include "SizeFieldBase.h"
#include "SizeFieldManager.h"
#include "DistanceFunction.h"
#include "MeshParametersManager.h"
#include "CallbackManager.h"
#include "regularizer.h"

// Linear operators
#include "MAdOperatorBase.h"
#include "LinearEdgeSplitOp.h"
#include "LinearEdgeCollapseOp.h"
#include "LinearFaceCollapseOp.h"
#include "LinearDESCOp.h"
#include "LinearEdgeSwapOp.h"
#include "LinearFaceSwapOp.h"
#include "LinearVertexMoveOp.h"
#include "LinearRegionRemoveOp.h"
#include "LinearSliverFaceHandler.h"
#include "LinearSliverRegionHandler.h"

// Curved operators
#include "CurvedEdgeSplitOp.h"
#include "CurvedEdgeCollapseOp.h"
#include "CurvedFaceCollapseOp.h"
#include "CurvedDESCOp.h"
#include "CurvedEdgeSwapOp.h"
#include "CurvedFaceSwapOp.h"
#include "CurvedSliverFaceHandler.h"
#include "CurvedSliverRegionHandler.h"

// repositioning
#include "GeoMatcher.h"
#include "NodesRepositioningOp.h"
#include "LaplaceSmoothingOp.h"
#include "MobileObject.h"

// standard C/C++
#include <iostream>
#include <stdlib.h>
#include <fstream>
#include <sstream>
#include <limits>

#ifdef PARALLEL
#include "mpi.h"
#include "MeshDataBaseParallelInterface.h"
#endif
using std::cerr;
using std::endl;
using std::set;
using std::vector;
using std::multiset;
using std::string;
using std::stringstream;
using std::cout; 

namespace MAd {


    IMeshAdapter * IMeshAdapter::BuildMeshAdapter(MDB_Mesh * mesh, ISizeField * sf)
    {
        pSField psf = dynamic_cast<SizeFieldBase *>(sf);
        return new CMeshAdapter(mesh, psf);
    }

  // -------------------------------------------------------------------
  void MeshAdapterAbortFct(void * data) {
    ((CMeshAdapter *) data) -> abort();
  }

  // -------------------------------------------------------------------
  CMeshAdapter::CMeshAdapter(pMesh m, ISizeField * sf): 
    mesh(m),
    eSplitOp(0), eCollapseOp(0), fCollapseOp(0), descOp(0), 
    eSwapOp(0), fSwapOp(0), vMoveOp(0), rRegionOp(0), sliverFOp(0), 
    sliverROp(0), geoTracker(0), 
    mpm(MeshParametersManagerSgl::instance()), 
    updateSFFrequency(0), verbosity(2), outputDirectoryPath(""), debugLevel(0)
  {
      pSField psf = dynamic_cast<SizeFieldBase *>(sf);
      SFManager = new SizeFieldManager(mesh, psf);

    MAdMsgSgl            ::instance().registerAbortFct(MeshAdapterAbortFct,this);
    MAdTimeManagerSgl    ::instance().initialize();
    MAdStatisticsSgl     ::instance().initialize();
    CallBackManagerSgl   ::instance().initialize();
    NodalDataManagerSgl  ::instance().initialize(mesh);
    MeshQualityManagerSgl::instance().initialize(mesh,SFManager->getSizeField(),MEANRATIO);
    HistorySgl           ::instance().initialize();
    HistorySgl           ::instance().closeJournal();
    ModelConstraintManagerSgl::instance().initialize(M_model(mesh));
    MeshParametersManagerSgl::instance().initialize();

    if (mesh) buildOperators();

    setDefaultValues();

    if ( debugLevel >= 1 && !checkTheMesh() ) abort(__LINE__,__FILE__);
  }

  // -------------------------------------------------------------------
  CMeshAdapter::~CMeshAdapter()
  {
    removeOperators();
    if (geoTracker) delete geoTracker;
    if (SFManager) delete SFManager;

    MAdTimeManagerSgl    ::instance().finalize();
    MAdStatisticsSgl     ::instance().finalize();
    CallBackManagerSgl   ::instance().finalize();
    NodalDataManagerSgl  ::instance().finalize();
    MeshQualityManagerSgl::instance().finalize();
    HistorySgl           ::instance().finalize();
    ModelConstraintManagerSgl::instance().finalize();
    MeshParametersManagerSgl::instance().finalize();
  }

  // -------------------------------------------------------------------
  // -------------------------------------------------------------------
  // -------------------------------------------------------------------

  // -------------------------------------------------------------------
  void CMeshAdapter::setDefaultValues()
  {
    maxIterationsNumber = std::numeric_limits<int>::max();

    setEdgeLenSqBounds             ( 1./3. , 3. );

    if ( M_dim(mesh) == 2 ) {
      setNoSwapQuality             ( 0.2 );
    }
    else { //  to be checked
      setNoSwapQuality             ( 0.1 );
    }

    mpm.setSliverTriBound          ( 0.01 );
    mpm.setSliverTetBound          ( 0.02 );

    // infinite loops control
    setSliverPermissionInESplit    ( true, 10. );
    setSliverPermissionInECollapse ( true, 0.1 );

    setCollapseOnBoundary          ( true, 1.e-6 );
    setSwapOnBoundary              ( true, 1.e-6 );
    
    activatedSplitOp = true;
    activatedSwapOp = true;
    activatedCollapseOp = true;

#ifdef PARALLEL
    load_balance_algorithm = DEFAULT_ALGORITHM;
    dataExchanger = new PWLSFieldDE((PWLSField*) SFManager->getSizeField());
#endif

    setQualityMeasure(MEANRATIO);

    updateSFFrequency = 0;
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::setQualityMeasure(evaluationType type)
  {
    MeshQualityManagerSgl::instance().setQualityMeasure(type);
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::printParameters() const
  {
    mpm.diagnostics();
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::setEdgeLenSqBounds(double lower, double upper)
  {
    if ( sqrt(lower) > 0.5*sqrt(upper) ) {
      stringstream ss;
      ss << "Edge length interval is too small ("
         << sqrt(lower) << ", " << sqrt(upper) <<")\n";
      MAdMsgSgl::instance().warning(-1, __FILE__, ss.str().c_str());
    }

    mpm.setLowerLengthSqBound(lower);
    mpm.setUpperLengthSqBound(upper);
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::setCollapseOnBoundary(bool accept, double tolerance) 
  {
    eCollapseOp->collapseOnBoundary(accept,tolerance);
    fCollapseOp->collapseOnBoundary(accept,tolerance);
    sliverROp  ->collapseOnBoundary(accept,tolerance);
    sliverFOp  ->collapseOnBoundary(accept,tolerance);
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::setSwapOnBoundary(bool accept, double tolerance) 
  {
    eSwapOp  ->swapOnBoundary(accept,tolerance);
    sliverROp->swapOnBoundary(accept,tolerance);
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::setMaxIterationsNumber(int max)
  {
    maxIterationsNumber = max;
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::setNoSwapQuality(double noSwapQuality)
  {
    mpm.setNoSwapQuality(noSwapQuality);
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::setSliverQuality(double sliverQuality)
  {
    mpm.setSliverTriBound(sliverQuality);
    mpm.setSliverTetBound(sliverQuality);
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::setSliverPermissionInESplit(bool perm, double bound)
  {
    mpm.setSliverPermissionInESplit(perm, bound);
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::setSliverPermissionInECollapse(bool perm, double bound)
  {
    mpm.setSliverPermissionInECollapse(perm, bound);
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::setGeoTracking(bool track, bool cavityEqualMesh, 
                                   int cavityThickness, double chi,
                                   bool strictChecking, bool forceRelocation)
  {
    if ( track && !geoTracker) {
      geoTracker = new geoMatcher(mesh,this);
      geoTracker->setCavityEqualMesh(cavityEqualMesh,cavityThickness);
      geoTracker->setStiffnessAlterationCoef(chi);
      geoTracker->setStrictChecking(strictChecking);
      geoTracker->setForceRelocation(forceRelocation);
      if ( !sliverFOp && !sliverROp ) {
        MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                    "Build sliver handler before setting up geo tracker");
      }
      geoTracker->setSliverFaceHandler(sliverFOp);
      geoTracker->setSliverRegionHandler(sliverROp);
    }
    else if ( !track && geoTracker) { 
      delete geoTracker;
      geoTracker = NULL;
    }
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::setInfiniteLength(double length)
  {
    mpm.setBigLength(length);
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::setSFUpdateFrequency(int freq)
  {
    updateSFFrequency = freq;
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::setVerbosity(int _verbosity)
  {
    verbosity = _verbosity;
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::clearConstraints() const
  {
    DeleteConstraint(mesh);
    ModelConstraintManagerSgl::instance().clear();
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::setConstraint( pEntity e ) const
  {
    EN_constrain( e );

    switch( EN_type( e ) )
    {
      case 1:
        EN_constrain( (pEntity)E_vertex( (pEdge)e, 0 ) );
        EN_constrain( (pEntity)E_vertex( (pEdge)e, 1 ) );
        break;
      case 2:
      {
        void* tmp = 0;
        pPList fEdges = F_edges( (pFace)e );
        while( pEdge edge = (pEdge)PList_next( fEdges, &tmp ) )
        {
          setConstraint( (pEntity)edge );
        }
        PList_delete( fEdges );
      }
      break;
      case 3:
      {
        void* tmp = 0;
        pPList rFaces = R_faces( (pRegion)e );
        while( pFace face = (pFace)PList_next( rFaces, &tmp ) )
        {
          setConstraint( (pEntity)face );
        }
        PList_delete( rFaces );
      }
      break;    
    }
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::setConstraint(int type, int id) const
  {
    ModelConstraintManagerSgl::instance().constrain(type,id);
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::removeConstraint(int type, int id) const
  {
    ModelConstraintManagerSgl::instance().unconstrain(type,id);
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::setConstraint(pGEntity e) const
  {
    ModelConstraintManagerSgl::instance().constrain(e);
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::removeConstraint(pGEntity e) const
  {
    ModelConstraintManagerSgl::instance().unconstrain(e);
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::removeOperators()
  {
    if (eSplitOp)    { delete eSplitOp;     eSplitOp=0;    }
    if (eCollapseOp) { delete eCollapseOp;  eCollapseOp=0; }
    if (fCollapseOp) { delete fCollapseOp;  fCollapseOp=0; }
    if (descOp)      { delete descOp;       descOp=0;      }
    if (eSwapOp)     { delete eSwapOp;      eSwapOp=0;     }
    if (fSwapOp)     { delete fSwapOp;      fSwapOp=0;     }
    if (vMoveOp)     { delete vMoveOp;      vMoveOp=0;     }
    if (rRegionOp)   { delete rRegionOp;    rRegionOp=0;   }
    if (sliverFOp)   { delete sliverFOp;    sliverFOp=0;   }
    if (sliverROp)   { delete sliverROp;    sliverROp=0;   }
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::buildOperators()
  {
    removeOperators();
    
    int order = M_maxOrder(mesh);
    
    if ( order < 2 )
    {
      eSplitOp    = new LinearedgeSplitOp         (mesh,SFManager->getSizeField());
      eCollapseOp = new LinearedgeCollapseOp      (mesh,SFManager->getSizeField());
      fCollapseOp = new LinearfaceCollapseOp      (mesh,SFManager->getSizeField());
      descOp      = new LinearDESCOp              (mesh,SFManager->getSizeField());
      eSwapOp     = new LinearedgeSwapOp          (mesh,SFManager->getSizeField());
      fSwapOp     = new LinearfaceSwapOp          (mesh,SFManager->getSizeField());
      vMoveOp     = new LinearvertexMoveOp        (mesh,SFManager->getSizeField(),false);
      rRegionOp   = new regionRemoveOp            (mesh,SFManager->getSizeField());
      sliverROp   = new LinearsliverRegionHandler (mesh,SFManager->getSizeField());
      sliverFOp   = new LinearsliverFaceHandler   (mesh,SFManager->getSizeField());
    }
    else
    {
//       MAdMsgSgl::instance().error(__LINE__,__FILE__, "AdaptOperators Not Yet Implemented for Order > 1");
      eSplitOp    = new CurvededgeSplitOp         (mesh,SFManager->getSizeField());
      eCollapseOp = new CurvededgeCollapseOp      (mesh,SFManager->getSizeField());
      fCollapseOp = new CurvedfaceCollapseOp      (mesh,SFManager->getSizeField());
      descOp      = new CurvedDESCOp              (mesh,SFManager->getSizeField());
      eSwapOp     = new CurvededgeSwapOp          (mesh,SFManager->getSizeField());
      fSwapOp     = new CurvedfaceSwapOp          (mesh,SFManager->getSizeField());
      vMoveOp     = new LinearvertexMoveOp        (mesh,SFManager->getSizeField(),false);
      rRegionOp   = new regionRemoveOp            (mesh,SFManager->getSizeField());
      sliverROp   = new CurvedsliverRegionHandler (mesh,SFManager->getSizeField());
      sliverFOp   = new CurvedsliverFaceHandler   (mesh,SFManager->getSizeField());
    }
    
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::addCallback(CBFunction CB, void* userData, 
                                CBFunc_move CB_move, void* userData_move)
  { 
    CallBackManagerSgl::instance().registerCallBack(CB,userData);
    if (CB_move) {
      CallBackManagerSgl::instance().registerCallBackMove(CB_move,userData_move);
    }
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::addSizeField(ISizeField * sf)
  {
      pSField psf = dynamic_cast<SizeFieldBase *>(sf);
      SFManager->addSizeField(psf);
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::setSizeFieldSmoothing(bool enable, double maxGrad)
  {
    SFManager->setSmoothing(enable,maxGrad);
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::incrementTime(double dt) 
  {
    MAdTimeManagerSgl::instance().incrementTime(dt);
    updateSizeField();
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::setTime(double t) 
  {
    MAdTimeManagerSgl::instance().setTime(t);
    updateSizeField();
  }

  // -------------------------------------------------------------------
  double CMeshAdapter::getTime() const {
    return MAdTimeManagerSgl::instance().getTime();
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::updateSizeField()
  {
    SFManager->update();
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::updateSizeField(pVertex pv)
  {
      SFManager->update(pv);
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::storeInitialCoordinates()
  {
    if ( getTime() != 0. ) {
      stringstream ss;
      ss << "Storing coordinates at time " << getTime() << endl;
      MAdMsgSgl::instance().warning(-1, __FILE__, ss.str().c_str());
    }
    NodalDataManagerSgl::instance().storeCoordinates();
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::removeStoredCoordinates()
  {
    NodalDataManagerSgl::instance().removeCoordinates();
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::registerObjects(mobileObjectSet* objs)
  {
    set<mobileObject*> allObj = objs->getObjects();
  
    set<mobileObject*>::const_iterator it = allObj.begin();
    for (; it != allObj.end(); it++) {
      set<LocalSizeField* > sizes = (*it)->getSizes();
      set<LocalSizeField* >::iterator it1 = sizes.begin();
      for (; it1 != sizes.end(); it1++) {
        SFManager->addSizeField(*it1);
      }
    }

    objects = objs;
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::registerData  (std::string name, 
                                   const std::vector<double> data) const
  {
    NodalDataManagerSgl::instance().registerData(name,data);
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::registerVData (std::string name, 
                                   const std::vector<std::vector<double> > data) const
  {
    NodalDataManagerSgl::instance().registerVData(name,data);
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::getMeshData   (std::string name, 
                                   std::vector<double> * data) const
  {
    NodalDataManagerSgl::instance().getMeshData(name,data);
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::getMeshVData  (std::string name, 
                                   std::vector<std::vector<double> > * data) const
  {
    NodalDataManagerSgl::instance().getMeshVData(name,data);
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::removeData    (std::string name) const
  {
    NodalDataManagerSgl::instance().removeData(name);
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::removeVData   (std::string name) const
  {
    NodalDataManagerSgl::instance().removeVData(name);
  }

  // -------------------------------------------------------------------
  // -------------------------------------------------------------------
  // -------------------------------------------------------------------

  // -------------------------------------------------------------------
  // LEVEL 1 OPERATIONS
  // -------------------------------------------------------------------

  // -------------------------------------------------------------------
  bool CMeshAdapter::splitEdge(pEdge edge, bool checkSize)
  {
      // set edge to split
      double reducSq = eSplitOp->setSplitEdge(edge);

      // check that the edge is not splitted in short edges
      if (checkSize)
      {
          double lenSq = SFManager->getSizeField()->SF_E_lengthSq(edge);
          if ((reducSq * lenSq) <= mpm.getLowerLengthSqBound()) return false;
      }

      // do the checks
      double worstShape;
      if (!eSplitOp->evaluate(&worstShape)) return false;

      // check that we do not create a worse sliver ( if asked for )
      // if we can create new slivers, check that the edge is long enough to allow it
      if (worstShape < mpm.getSliverTetBound())
      {
          double oriWorst;
          MeshQualityManagerSgl::instance().E_worstShape(edge, &oriWorst);
          if (worstShape < oriWorst) {
              if (!mpm.getSliverPermissionInESplit()) return false;
              if (mpm.getSliverUpperLengthSqBound() > 0. &&
                  mpm.getSliverUpperLengthSqBound() > SFManager->getSizeField()->SF_E_lengthSq(edge))
              {
                  return false;
              }
          }
      }

      // do the job
      eSplitOp->apply();
      return true;
  }

  // -------------------------------------------------------------------
  pVertex CMeshAdapter::splitEdge(pEdge edge, double t, bool checkShapes)
  {
    // set edge to split
    eSplitOp->setSplitEdge(edge, t);

    // do the checks
    if (checkShapes) {

        double worstShape;
        if (!eSplitOp->evaluate(&worstShape)) return 0;

        // check that we do not create a worse sliver ( if asked for )
        // if we can create new slivers, check that the edge is long enough to allow it
        if (worstShape < mpm.getSliverTetBound())
        {
            double oriWorst;
            MeshQualityManagerSgl::instance().E_worstShape(edge, &oriWorst);
            if (worstShape < oriWorst) {
                if (!mpm.getSliverPermissionInESplit()) return 0;
                if (mpm.getSliverUpperLengthSqBound() > 0. &&
                    mpm.getSliverUpperLengthSqBound() > SFManager->getSizeField()->SF_E_lengthSq(edge))
                {
                    return 0;
                }
            }
        }
    }

    // do the job
    eSplitOp->apply();
    return eSplitOp->getNewVertex();
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::collapseEdgeBrute(pEdge edge)
  {
      pVertex vDel = E_vertex(edge,0);
      pVertex vTgt = E_vertex(edge,1);
      if ( V_whatInType(vTgt) != 0 ) {
        vDel = E_vertex(edge,1);
        vTgt = E_vertex(edge,0);
      }
      assert(V_whatInType(vTgt)==0);
      eCollapseOp->setCollapseEdge(edge,vDel,vTgt);
      double shape;
      eCollapseOp->evaluate(&shape);
      eCollapseOp->apply();
  }

  // -------------------------------------------------------------------
  bool CMeshAdapter::collapseEdge(pEdge edge)
  {
    double shapes[2] = {-1.,-1.};
    int ok[2] = {0,0};

    for (int iDir=0; iDir<2; iDir++) {

      pVertex vDel = E_vertex(edge,iDir);
      pVertex vTgt = E_vertex(edge,1-iDir);
      eCollapseOp->setCollapseEdge(edge,vDel,vTgt);

      // do the checks
      ok[iDir] = eCollapseOp->evaluate(&(shapes[iDir]));

      // check that we do not create a worse sliver ( if asked for )
      // if we can create new slivers, check that the edge is short enough to allow it
      if ( shapes[iDir] < mpm.getSliverTetBound() ) 
        {
          double oriWorst;
          MeshQualityManagerSgl::instance().E_worstShape(edge, &oriWorst);
          if ( shapes[iDir] < oriWorst ) {
            if ( !mpm.getSliverPermissionInECollapse() ) ok[iDir] = 0;
            if ( mpm.getSliverLowerLengthSqBound() > 0. &&
                 mpm.getSliverLowerLengthSqBound() < SFManager->getSizeField()->SF_E_lengthSq(edge) )
              {
                ok[iDir] = 0;
              }
          }
        }
    }
  
    // find the best direction
    int best = -1; double worst = -1.;
    if ( ok[0] ) { best = 0; worst = shapes[0]; }
    if ( ok[1] && shapes[1] > worst ) {
      best = 1;
      worst = shapes[1];
    }
  
    // if possible apply an edge collapse
    if ( best >= 0 ) {
      pVertex vDel = E_vertex(edge,best);
      pVertex vTgt = E_vertex(edge,1-best);
      eCollapseOp->setCollapseEdge(edge,vDel,vTgt);
      eCollapseOp->apply();
      return true;
    }

    return false;
  }

  // -------------------------------------------------------------------
  bool CMeshAdapter::collapseFace(pFace face, pEdge edge)
  {
    for( int ClpsOnvt = 1; ClpsOnvt >=0; ClpsOnvt-- )
      {
        fCollapseOp->reset(face, edge, ClpsOnvt!=0);
        
        // do the checks
        double worstShape;
        if ( !fCollapseOp->evaluate(&worstShape) ) continue;
        
        // check that we do not create a sliver
        if ( worstShape < mpm.getSliverTetBound() ) continue;
        
        // do the job
        fCollapseOp->apply();
        return true;
      }
    return false;
  }

  // -------------------------------------------------------------------
  bool CMeshAdapter::DSplitCollapseEdge(pRegion pr, pEdge edge1, pEdge edge2)
  {
    descOp->setDESC(pr,edge1,edge2);

    // do the checks
    double worstShape;
    if ( !descOp->evaluate(&worstShape) ) return false;

    // check that we do not create a sliver
    if ( worstShape < mpm.getSliverTetBound() ) return false;

    // do the job
    descOp->apply();
    return true;
  }

  // -------------------------------------------------------------------
  bool CMeshAdapter::swapEdge(pEdge edge)
  {
    eSwapOp->setSwapEdge(edge);

    // do the checks
    double worstShape;
    if ( !eSwapOp->evaluate(&worstShape) ) return false;

    // check that we do not create a sliver
    if ( worstShape < mpm.getSliverTetBound() ) return false;

    // do the job
    eSwapOp->apply();
    return true;
  }

  // -------------------------------------------------------------------
  bool CMeshAdapter::swapFace(pFace face)
  {
    fSwapOp->setSwapFace(face);
  
    // do the checks
    double worstShape;
    if ( !fSwapOp->evaluate(&worstShape) ) return false;

    // check that we do not create a sliver
    if ( worstShape < mpm.getSliverTetBound() )  return false;

    // do the job
    fSwapOp->apply();
    return true;
  }

  // -------------------------------------------------------------------
  bool CMeshAdapter::moveVertex (pVertex v, double dxyz[3])
  {
    return vMoveOp->move(v,dxyz);
  }

  // -------------------------------------------------------------------
  bool CMeshAdapter::removeRegion(pRegion region)
  {
    rRegionOp->setRegion(region);
  
    // do the checks
    double worstShape;
    if ( !rRegionOp->evaluate(&worstShape) ) return false;

    // check that we do not create a sliver
    if ( worstShape < mpm.getSliverTetBound() )  return false;

    // do the job
    rRegionOp->apply();
    return true;
  }

  // -------------------------------------------------------------------
  bool CMeshAdapter::putVertex (pVertex v, double xyz[3])
  {
    double oriPos[3], dxyz[3];
    V_coord(v,oriPos);
    for (int i=0; i<3; i++) dxyz[i] = xyz[i] - oriPos[i];
    return moveVertex(v,dxyz);
  }

  // -------------------------------------------------------------------
//   bool CMeshAdapter::moveVertices (multiset<vDisplacement,vDisplacementLess>& vDisps)
//   {
//     return vMoveOp->move(vDisps);
//   }

  // -------------------------------------------------------------------
  // LEVEL 2 OPERATIONS
  // -------------------------------------------------------------------

  // -------------------------------------------------------------------
  double CMeshAdapter::LaplaceSmoothing()
  {
    MAdResourceManager& tm = MAdResourceManagerSgl::instance();

    double t0 = tm.getTime();

    double convergenceCriterion = 1e-2;
    double L2Disp = 0.;
    LaplaceSmoothingOp* laplOp = new LaplaceSmoothingOp(mesh,SFManager->getSizeField());
    laplOp->run(&L2Disp);
    double L2Disp0 = L2Disp; int i = 0;
    double L2DispTot = 0.;
    while ( ( L2Disp > convergenceCriterion * L2Disp0 ) && ( L2Disp > MAdTOL ) ) {
      laplOp->run(&L2Disp);
      i++;
      L2DispTot += L2Disp;
    }
    if (laplOp) delete laplOp;

    double dt = tm.getTime() - t0;
    if ( verbosity >= 2 ) {
      stringstream ss;
      ss << "Performed a Laplace smoothing: total L2 norm of the displacements: "
         << L2DispTot<<" in "<<dt<<" seconds\n";
      MAdMsgSgl::instance().info(-1, __FILE__, ss.str().c_str());
    }

    if ( debugLevel >= 3 && !checkTheMesh() ) abort(__LINE__,__FILE__);

    return L2DispTot;
  }

  // -------------------------------------------------------------------
  int CMeshAdapter::eSplitLoop() 
  {
    if (!activatedSplitOp) return 0;
    
    MAdResourceManager& tm = MAdResourceManagerSgl::instance();
    MAdStatistics&  stat = MAdStatisticsSgl ::instance();

    double t0 = tm.getTime();

    int nSplit = 0;
    int countE = 0; int oriNbEdges = M_numEdges(mesh);
    EIter ei = M_edgeIter(mesh);
    pEdge edge;
    while ( (countE < oriNbEdges) && ( edge = EIter_next(ei) ) )
    {
      if( !EN_constrained((pEntity)edge) )
      {
        double lengthSq = SFManager->getSizeField()->SF_E_lengthSq(edge);
        if ( lengthSq > mpm.getUpperLengthSqBound() )
          if (splitEdge(edge,true)) nSplit++;
      }
      countE++;
    }
    EIter_delete(ei);
        
    double dt = tm.getTime() - t0;
    stat.addCPUESplits(dt);
    stat.addNumESplits(nSplit);
    if ( verbosity >= 2 ) {
      stringstream ss;
      ss << "Performed "<< nSplit<<" edge splits in "<<dt<<" seconds";
      MAdMsgSgl::instance().info(-1, __FILE__, ss.str().c_str());
    }

    if ( debugLevel >= 3 && !checkTheMesh() ) abort(__LINE__,__FILE__);

    // --- Geometry tracking ---
    if (nSplit) snapVertices();
    
    return nSplit;
  }

  // -------------------------------------------------------------------
  int CMeshAdapter::eCollapseLoop() 
  {
    if (!activatedCollapseOp) return 0;
    
    MAdResourceManager& tm = MAdResourceManagerSgl::instance();
    MAdStatistics&  stat = MAdStatisticsSgl ::instance();

    double t0 = tm.getTime();

    int nCollapse = 0;
    int countE = 0; int oriNbEdges = M_numEdges(mesh);
    EIter ei = M_edgeIter(mesh);
    pEdge edge;
    while ( (countE < oriNbEdges) && ( edge = EIter_next(ei) ) )
    {
      if( !EN_constrained((pEntity)edge) )
      {
        double lengthSq = SFManager->getSizeField()->SF_E_lengthSq(edge);
        if ( lengthSq < mpm.getLowerLengthSqBound() )
          if (collapseEdge(edge)) nCollapse++;
      }
      countE++;
    }
    EIter_delete(ei);

    double dt = tm.getTime() - t0;
    stat.addCPUECollapses(dt);
    stat.addNumECollapses(nCollapse);
    if ( verbosity >= 2 ) {
      stringstream ss;
      ss << "Performed "<< nCollapse<<" edge collapses in "<<dt<<" seconds";
      MAdMsgSgl::instance().info(-1, __FILE__, ss.str().c_str());
    }

    if ( debugLevel >= 3 && !checkTheMesh() ) abort(__LINE__,__FILE__);

    return nCollapse;
  }

  // -------------------------------------------------------------------
  int CMeshAdapter::eSplitCollapseLoop() 
  {
    if (!activatedSplitOp || !activatedCollapseOp ) return 0;
    
    int nSplit = 0;
    int nCollapse = 0;
    int countE = 0; int oriNbEdges = M_numEdges(mesh);
    EIter ei = M_edgeIter(mesh);
    pEdge edge;
    while ( (countE < oriNbEdges) && ( edge = EIter_next(ei) ) ) {
      double lengthSq = SFManager->getSizeField()->SF_E_lengthSq(edge);
      if ( lengthSq > mpm.getUpperLengthSqBound() ) if (splitEdge(edge,true)) nSplit++;
      if ( lengthSq < mpm.getLowerLengthSqBound() ) if (collapseEdge(edge)) nCollapse++;
      countE++;
    }
    EIter_delete(ei);

    if ( debugLevel >= 3 && !checkTheMesh() ) abort(__LINE__,__FILE__);

    return nSplit+nCollapse;
  }

  // -------------------------------------------------------------------
  int CMeshAdapter::edgeSwapLoop()
  {
    if (!activatedSwapOp) return 0;
    
    MAdResourceManager& tm = MAdResourceManagerSgl::instance();
    MAdStatistics&  stat = MAdStatisticsSgl ::instance();

    double t0 = tm.getTime();

    int nb_eswap = 0;
    pEdge edge;
    EIter ei = M_edgeIter(mesh);
    int ne = M_numEdges(mesh);
    int i = 0;
    while( ( edge = EIter_next(ei) ) )
    {
      if( i > ne ) break;
      i++;
  
      if( EN_constrained((pEntity)edge) )  continue;
      if( E_numRegions(edge) > eSwapOp->getMaxNumRgns() )  continue;
  
      double worstShape;
      MeshQualityManagerSgl::instance().E_worstShape(edge, &worstShape);
  
      if( worstShape > mpm.getNoSwapQuality() ) continue;
  
      eSwapOp->setSwapEdge(edge);

      double newWorstShape;
      if( eSwapOp->evaluate(&newWorstShape) )
      {
        if( newWorstShape > worstShape )
        {
          if( M_dim(mesh) == 2 || 
              (  ( eSwapOp->getMaxLenSq() < mpm.getUpperLengthSqBound() ) &&
                 ( eSwapOp->getMinLenSq() > mpm.getLowerLengthSqBound() ) ) )
          {
            eSwapOp->apply();
            nb_eswap++;
          }
        }
      }
    }
    EIter_delete(ei);

    double dt = tm.getTime() - t0;
    stat.addCPUESwaps(dt);
    stat.addNumESwaps(nb_eswap);
    if ( verbosity >= 2 ) {
      stringstream ss;
      ss << "Performed "<< nb_eswap<<" edge swaps in "<<dt<<" seconds";
      MAdMsgSgl::instance().info(-1, __FILE__, ss.str().c_str());
    }

    if ( debugLevel >= 3 && !checkTheMesh() ) abort(__LINE__,__FILE__);

    return nb_eswap;
  }

  // -------------------------------------------------------------------
  int CMeshAdapter::faceSwapLoop() 
  {
    if (!activatedSwapOp) return 0;
    
    int nfswap = 0;
    pFace face;
    FIter fi = M_faceIter(mesh);
    int nf = M_numFaces(mesh);
    int i = 0;
    while ( ( face = FIter_next(fi) ) ) {
      if( i > nf ) break;
      i++;

      if( EN_constrained((pEntity)face) )  continue;

      double oriWorst;
      MeshQualityManagerSgl::instance().F_worstShape(face, &oriWorst);

      if ( oriWorst > mpm.getNoSwapQuality() ) continue;

      fSwapOp->setSwapFace(face);
      double newWorst;
      if( fSwapOp->evaluate(&newWorst) ) {
        if ( newWorst > oriWorst ) {
          fSwapOp->apply();
          nfswap++;
        }
      }
    }
    FIter_delete(fi);

    if ( debugLevel >= 3 && !checkTheMesh() ) abort(__LINE__,__FILE__);

    return nfswap;
  }

  // -------------------------------------------------------------------
  int CMeshAdapter::splitEveryEdgeOnce()
  {
    if (!activatedSplitOp) return 0;
    
    MAdResourceManager& tm = MAdResourceManagerSgl::instance();
    double t0 = tm.getTime();

    // list initial edges
    std::set<pEdge> initEdges;
    EIter ei = M_edgeIter(mesh);
    pEdge edge;
    while ( ( edge = EIter_next(ei) ) ) initEdges.insert(edge);
    EIter_delete(ei);
    
    // split every edge
    int numOp = 0;
    std::set<pEdge>::const_iterator it = initEdges.begin();
    for (; it != initEdges.end(); it++ ) {
      if ( splitEdge( *it ) ) numOp++;
    }

    double dt = tm.getTime() - t0;
    if ( verbosity >= 2 ) {
      stringstream ss;
      ss << "Performed "<< numOp<<" edge splits in "<<dt<<" seconds";
      MAdMsgSgl::instance().info(-1, __FILE__, ss.str().c_str());
    }

    if ( debugLevel >= 3 && !checkTheMesh() ) abort(__LINE__,__FILE__);

    // --- Geometry tracking ---
    if (numOp) snapVertices();
    
    return numOp;
  }

  // -------------------------------------------------------------------
  int CMeshAdapter::mOptimiseLoop()
  {
    if (M_dim(mesh) != 3) return 0;
    
    CavityRegularizer reg;
    
    MAdResourceManager& tm = MAdResourceManagerSgl::instance();
    
    double t0 = tm.getTime();
    
    /*
    // [1] Optimise Faces
    pPList Faces = PList_new();
    FIter Fi = M_faceIter(mesh);
    // [1.1] Get Faces Cavity
    while( pFace f = FIter_next(Fi) )
    {
      if ( F_whatInType(f) <= 2 ) 
        PList_append(Faces,f);
    }
    
    // [1.2] Optimise Operator
    reg.setCavity(Faces);
    reg.apply();
    
    // [1.3] Delete Temporary Iterators
    FIter_delete(Fi);
    PList_delete(Faces);
    */
    
    // [2] Optimise Regions
    pPList Regions = PList_new();
    // [2.1] Get Region Cavity
    RIter Ri = M_regionIter(mesh);
    while( pRegion r = RIter_next(Ri) )
    {
      if( EN_constrained((pEntity) r) )  continue;
      PList_append(Regions,r);
    }
    // [2.2] Optimise Operator
    reg.setCavity(Regions);
    reg.apply();
    
    // [2.3] Delete Temporary Iterators
    RIter_delete(Ri);
    PList_delete(Regions);
    
    
    double dt = tm.getTime() - t0;
    if ( verbosity >= 2 ) {
      stringstream ss;
      ss << "Performed Mesh Optimisation in "<<dt<<" seconds";
      MAdMsgSgl::instance().info(-1, __FILE__, ss.str().c_str());
    }
    
    return 0;
  }

  // -------------------------------------------------------------------
  int CMeshAdapter::removeSlivers()
  {
    MAdResourceManager& tm = MAdResourceManagerSgl::instance();
    MAdStatistics&  stat = MAdStatisticsSgl ::instance();

    double t0 = tm.getTime();
    
//     if (!activatedSplitOp || !activatedCollapseOp || !activatedSwapOp) return 0;

    int nbBefore, nbAfter;
    if (M_dim(mesh) == 3) sliverROp->removeSliverRegions(&nbBefore,&nbAfter);
    else                  sliverFOp->removeSliverFaces  (&nbBefore,&nbAfter);

    double dt = tm.getTime() - t0;
    stat.addCPURSlivers(dt);
    if ( verbosity >= 2 ) {
      stringstream ss;
      ss << "Removed slivers " << nbBefore << " -> " << nbAfter 
                      << " in "<<dt<<" seconds";
      MAdMsgSgl::instance().info(-1, __FILE__, ss.str().c_str());
    }

    if ( debugLevel >= 3 && !checkTheMesh() ) abort(__LINE__,__FILE__);

    return ( nbBefore - nbAfter );
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::snapVertices()
  {
    if (geoTracker) {
      MAdResourceManager& tm = MAdResourceManagerSgl::instance();
      double t0 = tm.getTime();

      if ( !(geoTracker->snap()) ) {
        MAdMsgSgl::instance().error(__LINE__,__FILE__,"Snapping failed");
      }

      double dt = tm.getTime() - t0;
      MAdMsgSgl::instance().info(__LINE__,__FILE__,
                                 "Performed vertex snapping in %f seconds",dt);
    }

    if ( debugLevel >= 3 && !checkTheMesh() ) abort(__LINE__,__FILE__);
  }

  // -------------------------------------------------------------------
  // LEVEL 3 OPERATIONS
  // -------------------------------------------------------------------

  // -------------------------------------------------------------------
  int CMeshAdapter::optimiseEdgeLength()
  {
    int nbMaxItOptiLen = 10;
    int nSplColl = 0;
    for (int iter=0; iter < nbMaxItOptiLen; iter++ ) {
      int nSC = eSplitCollapseLoop();
      stringstream ss;
      ss << "Applied "<<nSC<<" split or collapses\n";
      MAdMsgSgl::instance().info(-1, __FILE__, ss.str().c_str());
      if (!nSC) break;
      nSplColl += nSC;
    }
    return nSplColl;
  }

  // -------------------------------------------------------------------
  int CMeshAdapter::optimiseElementShape()
  {
    int nbMaxItOptiShp = 10;
    int nESwapTot = 0, nFSwapTot = 0;
    int nESwap = 0, nFSwap = 0;
    for (int iter=0; iter < nbMaxItOptiShp; iter++ ) {
      // find ill-shaped elements around each edge
      nESwap = edgeSwapLoop();
      stringstream ss;
      ss << "Applied " << nESwap<< " edge swaps"<<endl;
      MAdMsgSgl::instance().info(-1, __FILE__, ss.str().c_str());
      if ( M_dim(mesh) == 3 ) {
        // find ill-shaped elements around each face
        nFSwap = faceSwapLoop();
        stringstream ss1;                                 //redifinition by FB to remove shadow warnings
        ss1 << "Applied " << nFSwap<< " face swaps"<<endl; //  propagation
        MAdMsgSgl::instance().info(-1, __FILE__, ss1.str().c_str()); // propagation
      }
    
      if (nESwap+nFSwap == 0) break;
      nESwapTot += nESwap;
      nFSwapTot += nFSwap;
    }
    return ( nESwapTot + nFSwapTot );
  }

  // -------------------------------------------------------------------
  int CMeshAdapter::splitLongestEdges() 
  {
    MAdResourceManager& tm = MAdResourceManagerSgl::instance();
    MAdStatistics&  stat = MAdStatisticsSgl ::instance();

    double t0 = tm.getTime();

    int nSplitTot = 0;
    double boundSq = 4.*mpm.getUpperLengthSqBound();

    while (1)
      {
        int nSplit = 0;
        
        EIter ei = M_edgeIter(mesh);
        pEdge edge;
        double lengthSq;
        while ( ( edge = EIter_next(ei) ) ) {
          lengthSq = SFManager->getSizeField()->SF_E_lengthSq(edge);
          if ( lengthSq > boundSq ) if (splitEdge(edge,true)) nSplit++;
        }
        EIter_delete(ei);

        nSplitTot += nSplit;
        if ( debugLevel >= 3 && !checkTheMesh() ) abort(__LINE__,__FILE__);

        if ( nSplit == 0 ) break;
      }
    
    double dt = tm.getTime() - t0;
    stat.addCPUESplits(dt);
    stat.addNumESplits(nSplitTot);
    if ( verbosity >= 2 ) {
      stringstream ss;
      ss << "Performed "<< nSplitTot
                      <<" (very long) edge splits in "<<dt<<" seconds";
      MAdMsgSgl::instance().info(-1, __FILE__, ss.str().c_str());
    }

    // --- Geometry tracking ---
    if (nSplitTot) snapVertices();

    return nSplitTot;
  }

  // -------------------------------------------------------------------
  int CMeshAdapter::runOneIter()
  {
    // --- constrain parallel interfaces ---
#ifdef PARALLEL
    MAdResourceManager& tm = MAdResourceManagerSgl::instance();
    { 
      double t0= tm.getTime();
      UpdateParallelConstraint(mesh);
      double dt = tm.getTime() - t0;
      MAdMsgSgl::instance().info(__LINE__,__FILE__,
                                 "Performed constraint of parallel interfaces in %f seconds",dt);
      MPI_Barrier(M_getMPIComm(mesh));
    }
#endif

    // --- perform local mesh modifications ---
    int numTopoChg = 0;

    // first, split very long edges (avoid infinite loops split/collapse)
// #warning "First split very long edges (test)"
//     while (1) {
//       int nOp = splitLongestEdges();
//       if ( !nOp ) break;
//       numTopoChg += nOp;
//     }
    
    // NOTE : ORIGINAL
//     numTopoChg += eCollapseLoop(); // Collapse short edges
//     numTopoChg += edgeSwapLoop();  // Swap edges to improve elements quality
//     numTopoChg += removeSlivers(); // Remove slivers
//     numTopoChg += eSplitLoop();    // Split long edges
    
//     (L.Cadet)
    numTopoChg += eCollapseLoop(); // Collapse short edges
    numTopoChg += edgeSwapLoop();  // Swap edges to improve elements quality
    numTopoChg += removeSlivers(); // Remove slivers
    numTopoChg += mOptimiseLoop(); // Optimise Cavity 
    numTopoChg += eSplitLoop();    // Split long edges
    
    /* Loop Olivier */
    /*numTopoChg += removeSlivers();
    numTopoChg += eCollapseLoop();
    numTopoChg += edgeSwapLoop();
    numTopoChg += eSplitLoop();
    numTopoChg += edgeSwapLoop();*/



    // --- Reposition the vertices ---
//     LaplaceSmoothing(OPTIMAL);
//     LaplaceSmoothing(FAST);

#ifdef PARALLEL
    // --- move parallel interfaces ---
    { 
      MPI_Barrier(M_getMPIComm(mesh));
      double t0 = tm.getTime();
      DeleteParallelConstraint(mesh);
      double dt = tm.getTime() - t0;
      if( verbosity >= 2 ) {
        MAdMsgSgl::instance().info(__LINE__,__FILE__,
                                   "Performed unconstraint of parallel interface in %f seconds",dt);
      }

      t0 = tm.getTime();
      switch( load_balance_algorithm  ) {      
      case DEFAULT_ALGORITHM:
        Balance2( mesh, *dataExchanger );
        break;
      case METIS_ALGORITHM:
#ifdef _HAVE_PARMETIS_
// #warning "Temporary removed calls to load balancing with Metis (bug in it)"
//         BalanceMetis2( mesh, *dataExchanger );
#else
        MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                    "Could not balance load: Parmetis not enabled");
#endif
        break;
      default:
        MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                    "Unknown load balancing algorithm: %d",
                                    load_balance_algorithm);        
      }
      dt = tm.getTime() - t0;
      if( verbosity >= 2 ) {
        MAdMsgSgl::instance().info(__LINE__,__FILE__,
                                   "Performed load balancing in %f seconds",dt);
      }
    }
#endif

    // --- clean up the mesh by deleting dead elements ---
    M_clean(mesh);

// #warning "debug"
//     M_writeMsh(mesh,"test.msh",2);

    // --- check the mesh (debug) ---
    if ( debugLevel >= 2 && !checkTheMesh() )  abort(__LINE__,__FILE__);

    return numTopoChg;
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::uglyTheMesh(double avgQualThresh, int maxIt)
  {
    MeshQualityManagerSgl::instance().evaluateStatistics();
    double meanQuality  = MeshQualityManagerSgl::instance().getMeanShape();
  
    int iter = 0;
    while ( (meanQuality > avgQualThresh) && (iter < maxIt) ) {

      int numOp = 0;
      int oriNumFaces = M_numFaces(mesh);
      int count = 0;
      FIter fi = M_faceIter(mesh);
      pFace pf;
      while ( ( pf = FIter_next(fi) ) && ( count < oriNumFaces ) ) {
        int flag = 0;
        for (int j=0; j<3; j++) {
          pEdge pe = F_edge(pf,j);
          for( int ClpsOnvt = 0; ClpsOnvt < 2; ClpsOnvt++ ) {
            fCollapseOp->reset(pf, pe, ClpsOnvt!=0);
            double worstShape;
            if ( fCollapseOp->evaluate(&worstShape) ) {
              fCollapseOp->apply();
              flag = 1;
              break;
            }
          }
          if ( flag ) {
            numOp++;
            break;
          }
        }
        count++;
      }
      FIter_delete(fi);
      stringstream ss;
      ss << "Num face collapses: "<<numOp<<endl;
      MAdMsgSgl::instance().info(-1, __FILE__, ss.str().c_str());
    
      meanQuality  = MeshQualityManagerSgl::instance().getMeanShape();
      iter++;
    }
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::uglyTheMeshOli(double avgQualThresh, double minQual, int maxIt)
  {
    /* Based on edge collapse instead of face collapse */
    MeshQualityManagerSgl::instance().evaluateStatistics();
    double meanQuality  = MeshQualityManagerSgl::instance().getMeanShape();
  
    int iter = 0;
    while ( (meanQuality > avgQualThresh) && (iter < maxIt) ) {
      int numOp = 0;
      int oriNumEdges = M_numEdges(mesh);
      int count = 0;
      EIter ei = M_edgeIter(mesh);
      pEdge edge;
      while ( ( edge = EIter_next(ei) ) && ( count < oriNumEdges ) ) {
        double shapes[2] = {-1.,-1.};
        int ok[2] = {0,0};
        
        double oriWorst;
        MeshQualityManagerSgl::instance().E_worstShape(edge, &oriWorst);

        for (int iDir=0; iDir<2; iDir++) {
          pVertex vDel = E_vertex(edge,iDir);
          pVertex vTgt = E_vertex(edge,1-iDir);
          eCollapseOp->setCollapseEdge(edge,vDel,vTgt);

          // do the checks
          ok[iDir] = eCollapseOp->evaluate(&(shapes[iDir]));

          // check that the shape is worst, but not worse than the prescribed tolerance (minQual)
          // if we can create new slivers, check that the edge is short enough to allow it
          if ( shapes[iDir] > oriWorst || shapes[iDir] < minQual ) 
            ok[iDir] = 0;
        }

        // find the best direction
        int best = -1; double worst = -1.;
        if ( ok[0] ) { best = 0; worst = shapes[0]; }
        if ( ok[1] && shapes[1] < worst ) {
          best = 1;
          worst = shapes[1];
        }

        // if possible apply an edge collapse
        if ( best >= 0 ) {
          pVertex vDel = E_vertex(edge,best);
          pVertex vTgt = E_vertex(edge,1-best);
          eCollapseOp->setCollapseEdge(edge,vDel,vTgt);
          eCollapseOp->apply();
          numOp++;
        }
        count++;
      }
      EIter_delete(ei);
      stringstream ss;
      ss << "Num edge collapses: "<<numOp<<endl;
      MAdMsgSgl::instance().info(-1, __FILE__, ss.str().c_str());
    
      MeshQualityManagerSgl::instance().evaluateStatistics();
      meanQuality  = MeshQualityManagerSgl::instance().getMeanShape();
      printf("meanQuality: %e avgQuallThresh %e\n",meanQuality,avgQualThresh);
      iter++;
    }
  }


    


  // -------------------------------------------------------------------
  //! Removes elements with negative volume and returns \ingroup adaptation
  //! number of remaining negative volumes
  int CMeshAdapter::removeNegativeElements()
  {
    int nNeg = 0;
    int nRem = 0;
    RIter rIt= M_regionIter(mesh);
    while ( pRegion region = RIter_next(rIt) ) {
      if (R_volume(region) < 0.) {
        nNeg++;
        if ( removeRegion(region) ) nRem++;
      }
    }
    RIter_delete(rIt);

    stringstream ss;
    ss << "Negative elements removal: "
       << nNeg << " detected, "
       << nRem << " removed";
    MAdMsgSgl::instance().info(-1, __FILE__, ss.str().c_str());

    return (nNeg - nRem);
  }

  // -------------------------------------------------------------------
  // LEVEL 4 OPERATIONS
  // -------------------------------------------------------------------

  // -------------------------------------------------------------------
  int CMeshAdapter::run()
  {
    // --- Adaptation procedure ---

    MAdResourceManager& tm = MAdResourceManagerSgl::instance();
    double t0= tm.getTime();

    if( verbosity >= 1 )
      MAdMsgSgl::instance().info( -1, __FILE__, 
                                  "+++ Starting a MAdLib mesh adaptation procedure +++" );
    if( verbosity >= 2 )
      MAdMsgSgl::instance().info( -1, __FILE__, "");
    
    int iter;
    int numTopoChg = std::numeric_limits<int>::max();
    int numTopoChgPrevious = -1;
    int numTopoChgTwoPrevious = -2;
    
    for( iter = 1; numTopoChg != 0; iter++ )
    {
      if ( verbosity >= 1 ) {
        stringstream ss;
        ss << "   --- Mesh adaptation iteration "<<iter<<" ---";
        MAdMsgSgl::instance().info( -1, __FILE__, ss.str().c_str() );
      }

      numTopoChgTwoPrevious = numTopoChgPrevious;
      numTopoChgPrevious = numTopoChg;
      numTopoChg = runOneIter();

//       // for debugging
//       stringstream ss;
//       string iterStr;  ss << iter;  ss >> iterStr;
//       string debugName = outputDirectoryPath + "iter" + iterStr + ".msh";
//       M_writeMsh(mesh,debugName.c_str(),2);

#ifdef PARALLEL
      int numTopoChgGlob = 0;
      MPI_Allreduce(&numTopoChg,&numTopoChgGlob,1,MPI_INT,MPI_SUM,M_getMPIComm(mesh));
      numTopoChg = numTopoChgGlob;
#endif

      if ( updateSFFrequency && (iter%updateSFFrequency) == 0 ) updateSizeField();

      if ( verbosity >= 2 ) MAdMsgSgl::instance().info(-1, __FILE__, "");

      // --- Infinite loop detected ---
      if( ( numTopoChg == numTopoChgPrevious ) && ( numTopoChg == numTopoChgTwoPrevious ) ) {
        MAdStatisticsSgl::instance().addInfiniteLoops(1);
          MAdMsgSgl::instance().warning(-1,__FILE__, "Infinite loop detected");
        break;
      }
      
      // --- Maximum number of iterations reached ---
      if( iter == maxIterationsNumber && numTopoChg != 0 )
      {
        MAdMsgSgl::instance().warning(-1,__FILE__,
          "Reach maximum number of iterations");
        break;
      }
    }
    
    if( numTopoChg == 0 ) iter--; // To print correctly the number of iterations

    // --- Load balancing ---
#ifdef PARALLEL
    MPI_Barrier(M_getMPIComm(mesh));
    if( load_balance_algorithm != METIS_ALGORITHM ) {   
#ifdef _HAVE_PARMETIS_ 
// #warning "Temporary removed calls to load balancing with Metis (bug in it)"
      BalanceMetis2( mesh, *dataExchanger );
#endif
    }
#endif
    
    if ( debugLevel >= 1 && !checkTheMesh() )  abort(__LINE__,__FILE__);

    // --- Print conclusion ---
    stringstream ss;
    if( verbosity >= 1 ) ss << "+++ ";
    ss << "Ending a MAdLib mesh adaptation procedure (" 
       << tm.getTime() - t0 << " seconds - " << iter << " iterations)";
    if( verbosity >= 1 ) ss << "+++";
    MAdMsgSgl::instance().info( -1, __FILE__, ss.str().c_str() );
    if( verbosity >= 1 ) MAdMsgSgl::instance().info( -1, __FILE__, "" );
    
    return iter;
  }

#ifdef PARALLEL
  // -------------------------------------------------------------------
  void CMeshAdapter::setLoadBalancingAlgorithm( loadBalanceAlgorithm lbAlgo )
  {
    load_balance_algorithm = lbAlgo;
  }
  // -------------------------------------------------------------------
//   void CMeshAdapter::setDataExchanger( MDB_DataExchanger* dataExch )
//   {
//     dataExchanger = dataExch;
//   }
#endif

  // -------------------------------------------------------------------
  int CMeshAdapter::partlyMoveObjects(double t, double dt, double* part)
  {
    return objects->partlyMove(*vMoveOp,t,dt,part);
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::moveObjectsAndReposition(double t, double dt,
                                             bool subAdaptation,
                                             double qualityThreshold,
                                             double chi, bool meshIsCavity,
                                             int cavityThickness)
  {
    objects->moveAndReposition(mesh, t, dt, subAdaptation, qualityThreshold, chi, 
                               meshIsCavity, cavityThickness, this);

    /*
    objects->setupElasticRepositioning(mesh, t, dt, qualityThreshold, chi, 
                                       meshIsCavity, cavityThickness);
    int subIter = 0;
    double ratio = 0.;
    int achieved = -1;
    while ( achieved != 2 ) {
      achieved = objects->reposition(&ratio);
      MAdMsgSgl::instance().info(-1,__FILE__,
                                 "Advanced repositioning, achieved: %d, ratio: %f",achieved,ratio);
// #warning "output paper"
//       string name = "subIter";
//       stringstream ss;
//       string iterStr;  ss << subIter;  ss >> iterStr;
//       name = name + iterStr + "_1.pos";
//       writePos(name.c_str(),OD_MEANRATIO);

      if ( debugLevel >= 3 && !checkTheMesh() )  abort(__LINE__,__FILE__);
      if ( achieved <= 1 ) {
        if ( !subAdaptation ) {
          MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                      "Could not advance objects, ratio reached: %f",ratio);
        }
        else if ( !removeSlivers() && !optimiseElementShape() ) {
          MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                      "Could not advance objects, ratio reached: %f",ratio);
        }
      }
// #warning "output paper"
//       name = "subIter";
//       stringstream ss2;
//       iterStr ="";  ss2 << subIter;  ss2 >> iterStr;
//       name = name + iterStr + "_2.pos";
//       writePos(name.c_str(),OD_MEANRATIO);

      subIter++;
    }
    */

    if ( debugLevel >= 1 && !checkTheMesh() )  abort(__LINE__,__FILE__);
  }

  // -------------------------------------------------------------------
  // -------------------------------------------------------------------
  // -------------------------------------------------------------------
  void CMeshAdapter::nodalDataDiagnostics(std::ostream& out) const
  {
    NodalDataManagerSgl::instance().diagnostics(out);
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::openJournal() const
  {
    HistorySgl::instance().openJournal();
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::setReferenceJournal(std::string& name) const
  {
    HistorySgl::instance().loadJournal(name);
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::flushJournal(std::ostream& out) const
  {
    HistorySgl::instance().flushJournal(out);
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::enableSliverReports()
  {
    if ( !sliverROp ) {
      std::cerr<<"Error: no sliver region handler\n";
      exit(0);
    }

    sliverROp->enableReport(outputDirectoryPath);

    if ( !sliverFOp ) {
      MAdMsgSgl::instance().error(-1, __FILE__, 
                                  "Error: no sliver face handler\n");
      exit(0);
    }

    sliverFOp->enableReport(outputDirectoryPath);
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::testSliverOperators(bool test)
  {
    sliverROp->setTestAllOperators(test);
  }

  // -------------------------------------------------------------------
  bool CMeshAdapter::checkTheMesh(int verbosity_arg, MeshStatus * status) const
  {
    MAdMsgSgl::instance().info(-1,__FILE__,
                               "Checking mesh validity (debug level = %d)",
                               debugLevel);
    return checkMesh(mesh, CHECK_ALL, verbosity_arg, status);
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::infoMobileObjects(std::ostream& out) const
  {
    objects->describe(out);
  }

  // -------------------------------------------------------------------
  // -------------------------------------------------------------------
  // -------------------------------------------------------------------

  void CMeshAdapter::setOutputDirectoryPath(const char * _outputDirectoryPath)
  {
      outputDirectoryPath = _outputDirectoryPath;
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::writePos(const char * fn, MAdOutputData type) const
  {
    string fullName = outputDirectoryPath + fn;
    MAdGmshOutput(mesh, SFManager->getSizeField(), 
                  fullName.c_str(), type);
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::writeMsh(const char * fn) const
  {
    string fullName = outputDirectoryPath + fn;
    M_writeMsh(mesh,fullName.c_str(),2,NULL);
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::writeDistanceToWalls(std::string fnb) const
  {
    std::set<LocalSizeField*> localSF = SFManager->getLocalSizeFields();
    std::set<LocalSizeField*>::const_iterator it = localSF.begin();
    for (; it != localSF.end(); it++)
      {
        string fullName = outputDirectoryPath + fnb + "_" + (*it)->getName() + ".pos";
        MAdGmshOutput(mesh, *it, fullName.c_str(),
                      OD_DISTANCE);
      }
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::writeVolumicCurvature(std::string fnb) const
  {
    std::set<LocalSizeField*> localSF = SFManager->getLocalSizeFields();
    std::set<LocalSizeField*>::const_iterator it = localSF.begin();
    for (; it != localSF.end(); it++)
      {
        string fullName = outputDirectoryPath + fnb + "_" + (*it)->getName() + ".pos";
        MAdGmshOutput(mesh, *it, fullName.c_str(),
                      OD_CURVATURE_DIV);
      }
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::getModificationsInfo(int * nSplit, int * nColl, 
                                         int * nSwap, double * cpuSplit,
                                         double * cpuColl, double * cpuSwap,
                                         double * cpuSliv) const
  {
    *nSplit   = MAdStatisticsSgl::instance().getNumESplits();
    *nColl    = MAdStatisticsSgl::instance().getNumECollapses();
    *nSwap    = MAdStatisticsSgl::instance().getNumESwaps();
    *cpuSplit = MAdStatisticsSgl::instance().getCPUESplits();
    *cpuColl  = MAdStatisticsSgl::instance().getCPUECollapses();
    *cpuSwap  = MAdStatisticsSgl::instance().getCPUESwaps();
    *cpuSliv  = MAdStatisticsSgl::instance().getCPUSlivers();
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::getStatistics(double * meanQuality, 
                                  double * worstQuality) const
  {
    MeshQualityManagerSgl::instance().evaluateStatistics();
    *meanQuality  = MeshQualityManagerSgl::instance().getMeanShape();
    *worstQuality = MeshQualityManagerSgl::instance().getWorstShape();
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::printStatistics(std::ostream& out) const
  {
    MeshQualityManagerSgl::instance().evaluateStatistics();
    MeshQualityManagerSgl::instance().printStatistics(out);
    out << "\n\n";
    MAdStatisticsSgl::instance().print(out);
    out << "\n\n";
    if (geoTracker) geoTracker->printFailures(out);
  }

  // -------------------------------------------------------------------
  void CMeshAdapter::printSliverRegionStatistics(std::ostream& out) const 
  {
    if ( sliverROp ) sliverROp->printStats(out);
    //   if ( sliverROp ) sliverROp->printOperatorsTest(out);
  }
  // -------------------------------------------------------------------
  void CMeshAdapter::abort(int line, const char* file) const
  {
    cerr << "\n"
         << " ****************************\n"
         << "  CMeshAdapter is aborting... \n"
         << " ****************************\n\n";    

    string stName = outputDirectoryPath + "aborted_statistics";
    std::ofstream stOut(stName.c_str());
    printStatistics(stOut);

    string slName = outputDirectoryPath + "aborted_slivers";
    std::ofstream slOut(slName.c_str());
    printSliverRegionStatistics(slOut);

    string jName = outputDirectoryPath + "aborted_journal";
    std::ofstream jOut(jName.c_str());
    flushJournal(jOut);

    writePos("aborted.pos",OD_MEANRATIO);
    writeMsh("aborted.msh");

    if ( file && line >=0 ) {
      MAdMsgSgl::instance().info(-1,__FILE__,
                                 "Abort function called from file %s (line %d)",
                                 file,line);
    }
    else {
      MAdMsgSgl::instance().info(-1,__FILE__,"Abort!");
    }

    flush(std::cout);
    flush(std::cerr);

    exit(EXIT_FAILURE);
  }

  // -------------------------------------------------------------------

} // End of namespace MAd

