// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#ifndef _H_ADAPTINTERFACE
#define _H_ADAPTINTERFACE

#include "MAdExport.h"

// from Geo
#include "ModelInterface.h"

// from Mesh
#include "MeshDataBaseInterface.h"
#include "CheckMesh.h"
#ifdef PARALLEL
#include "MeshDataBaseComm.h"
#endif

// from Adapt
#include "IMeshAdapter.h"
#include "MAdOutput.h"

#include <set>
#include <string>
#include <vector>

namespace MAd {

  class SizeFieldManager;
  class geoMatcher;
  class mobileObjectSet;
  class MeshParametersManager;
  
  // Local mesh modification operators
  class edgeSplitOp;
  class edgeCollapseOp;
  class faceCollapseOp;
  class DESCOp;
  class edgeSwapOp;
  class faceSwapOp;
  class vertexMoveOp;
  class regionRemoveOp;
  class sliverFaceHandler;
  class sliverRegionHandler;
  
  // -------------------------------------------------------------------

  class CMeshAdapter : public IMeshAdapter
  {

  public:
  
    CMeshAdapter(pMesh m, ISizeField * sf=NULL);
    virtual ~CMeshAdapter();

    // --------------------------------------------

    // gives an overview of the current parameters
    virtual void printParameters() const;

    virtual void addCallback(CBFunction CB, void* userData, 
                     CBFunc_move CB_move=0, void* userData_move=0);
    virtual void addSizeField(ISizeField * sf);

    // whether or not the global size field is smoothed and maximum gradient
    virtual void setSizeFieldSmoothing(bool, double maxGrad=1.);

    // set the maximum number of iterations available to reach the 'convergence' 
    // of the global mesh adaptation procedure
    virtual void setMaxIterationsNumber(int max);

#ifdef PARALLEL
    // set load balancing algorithm
    virtual void setLoadBalancingAlgorithm( loadBalanceAlgorithm lbAlgo );
    // set the data Exchanger
//     virtual void setDataExchanger( MDB_DataExchanger* dataExch);
#endif

    // choose the quality criterion
    virtual void setQualityMeasure(evaluationType type);

    // impose the interval governing edges length in the transformed space
    virtual void setEdgeLenSqBounds(double lower, double upper);

    // set permission for operators to modify slightly boundaries 
    // (default: false) and tolerance for the relative volume/area 
    // modifications.
    // Edge collapse and face collapse:
    virtual void setCollapseOnBoundary(bool accept=true, double tolerance=1.e-6);
    // Edge swap:
    virtual void setSwapOnBoundary(bool accept=true, double tolerance=1.e-6);

    // impose the element quality from which a swap is not required
    virtual void setNoSwapQuality(double noSwapQuality);

    // impose the maximum quality of a sliver
    virtual void setSliverQuality(double sliverQuality);
    
    virtual void activateSplitOp(bool t = true) {activatedSplitOp = t;};
    virtual void activateSwapOp(bool t = true)  {activatedSwapOp = t;};
    virtual void activateCollapseOp(bool t = true)  {activatedCollapseOp = t;};
    
    // Allow or forbid operations creating a sliver. 
    // If allowed and a bound is prescribed, it is allowed only if the edge is 
    // longer/shorter than the bound (bound is expressed as the square of 
    // adimensional edge length)
    virtual void setSliverPermissionInESplit   (bool perm, double lenSqBound=-1.);
    virtual void setSliverPermissionInECollapse(bool perm, double lenSqBound=-1.);

    // tell if you want to project new vertices on geometric entities
    virtual void setGeoTracking(bool enable, bool cavityEqualMesh=false,
                        int cavityThickness=2, double chi=1.0,
                        bool strictChecking=false, bool forceRelocation=false);

    // set a value for a very huge edge length (default=1.e14)
    virtual void setInfiniteLength(double length);

    // frequency at which the size fields are updated inside the global loop
    // (useful for local and analytical SF)
    virtual void setSFUpdateFrequency(int freq);

    // Set verbosity:
    //  < 0 : no detail
    //    1 : global procedure details
    //  > 2 : iterations details
    virtual void setVerbosity(int _verbosity);

    // constraint a mesh/geometric entity and all its downward entities
    virtual void clearConstraints() const;
    virtual void setConstraint(pEntity e)        const;
    virtual void setConstraint(pGEntity ge)      const;
    virtual void setConstraint(int type, int id) const;
    // unconstrain a geometric entity
    virtual void removeConstraint(int type, int id) const;
    virtual void removeConstraint(pGEntity ge)      const;

    // manage the physical time
    virtual void   incrementTime(double dt);
    virtual void   setTime(double t);
    double getTime() const;
    virtual void   updateSizeField();
    virtual void   updateSizeField(pVertex pv);

    // functions to keep track of the initial coordinates
    virtual void storeInitialCoordinates();
    virtual void removeStoredCoordinates();

    // add predefined mobile objects
    virtual void registerObjects(mobileObjectSet* objs);

    // will attach/get the datas to the nodes of the mesh
    // order in vector is the node iterator order
    virtual void registerData  (std::string name, const std::vector<double>)               const;
    virtual void registerVData (std::string name, const std::vector<std::vector<double> >) const;
    virtual void getMeshData   (std::string name, std::vector<double> *)                   const;
    virtual void getMeshVData  (std::string name, std::vector<std::vector<double> > *)     const;
    virtual void removeData    (std::string name) const;
    virtual void removeVData   (std::string name) const;

  public:

    // ---------------- OPERATIONS ----------------

    // ----- Level 1 -----
    // ---> Elementary operations

    // Split the edge and if 'checkSize' is true, check that the two resulting 
    // edges are not short edges.
    bool splitEdge (pEdge e, bool checkSize=false);
    // split edge at the parametric location t and return the new vertex
    pVertex splitEdge(pEdge e, double t, bool checkShapes=false);
    bool collapseEdge (pEdge e);
    virtual void collapseEdgeBrute (pEdge e); // use it only if you know what your are doing
    bool collapseFace(pFace f, pEdge e);
    bool DSplitCollapseEdge(pRegion pr, pEdge edge1, pEdge edge2);
    bool swapEdge (pEdge e);
    bool swapFace (pFace f);
    bool removeRegion(pRegion region);
    bool moveVertex (pVertex v, double dxyz[3]);
    bool putVertex (pVertex v, double xyz[3]);
    //    bool moveVertices (std::multiset<vDisplacement,vDisplacementLess>& vDisps);

    // ----- Level 2 -----
    // ---> Loops on one elementary operation

    // node repositioning
    double LaplaceSmoothing();

    // topology operations
    int eSplitLoop();
    int eCollapseLoop();
    int eSplitCollapseLoop();
    int edgeSwapLoop();
    int faceSwapLoop();
    int splitEveryEdgeOnce();
    int mOptimiseLoop();
    
    // slivers handling
    int removeSlivers();

    // geometry matching
    virtual void snapVertices();

    // ----- Level 3 -----
    // ---> Procedures with a global objective
    int optimiseEdgeLength();
    int optimiseElementShape();
    int splitLongestEdges();
    int runOneIter();
    virtual void uglyTheMesh(double avgQualThresh, int maxIt);
    virtual void uglyTheMeshOli(double avgQualThresh, double minQual, int maxIt);
    int removeNegativeElements();

    // ----- Level 4 -----
    // ---> Global procedure
    int run();

    // ----- objects motion -----
    // move boundaries without repositioning nodes in the volume
    int partlyMoveObjects        (double t, double dt, double* part);
    // move boundaries and reposition all nodes with an elasticity analogy
    //   subAdaptation:    allow sliver elimination and mesh optimization if necessary
    //   qualityThreshold: at which quality do we optimize the mesh (if allowed)
    //   chi:              stiffness alteration coefficient (-1 = none)
    //   meshIsCavity:     true = elastic computation on whole mesh
    //   cavityThickness:  nb layers of elements if mesh is not the cavity
    virtual void moveObjectsAndReposition (double t, double dt, 
                                   bool subAdaptation=true,
                                   double qualityThreshold=0.,
                                   double chi=-1., 
                                   bool meshIsCavity=true, 
                                   int cavityThickness=3);
    
    // --------------------------------------------

  public:

    // ------ Diagnostics ------

    // get informations on mesh quality
    virtual void getStatistics(double * meanQuality, double * worstQuality) const;

    // get information on applied mesh modifications
    virtual void getModificationsInfo(int * nSplit, int * nColl, 
                              int * nSwap, double * cpuSplit,
                              double * cpuColl, double * cpuSwap,
                              double * cpuSliv) const;
    
    // about all datas attached to the nodes
    virtual void nodalDataDiagnostics(std::ostream& out) const;
  
    // journals listing all operations tested or applied
    virtual void setDebugLevel(int debug) { debugLevel = debug; }
    virtual void openJournal() const;
    virtual void setReferenceJournal(std::string& name) const;
    virtual void flushJournal(std::ostream& out) const;

    // sliver outputs
    virtual void enableSliverReports();
    virtual void testSliverOperators(bool test);

    // performs several checks to check the validity of the mesh
    bool checkTheMesh(int verbose=1, MeshStatus * status=NULL) const;

    // get infos about mobile objects
    virtual void infoMobileObjects(std::ostream& out=std::cout) const;

  public:

    // ------ Outputs ------

    // set the path to output directory
    void setOutputDirectoryPath(const char * _outputDirectoryPath);

    // write mesh with required postpro data in 'pos' format (Gmsh)
    // the output data can be any data listed in 'MAdOutput.h'
    void writePos(const char * fn, MAdOutputData type=OD_CONSTANT) const;

    // write mesh in 'msh' format (Gmsh)
    void writeMsh(const char * fn) const;

    // write a .pos file with the distance to walls for every local size field
    void writeDistanceToWalls(std::string fnBase) const;
    // write a .pos file with the 'volumic' curvature for every local size field
    virtual void writeVolumicCurvature(std::string fnBase) const;

    // get global data over the mesh
    virtual void printStatistics(std::ostream& out) const;
    virtual void printSliverRegionStatistics(std::ostream& out) const;

  public:

    // save all available informations to output directory and abort
    virtual void abort(int line=-1, const char* file=NULL) const;

    // --------------------------------------------

  private:

    virtual void setDefaultValues();
    virtual void buildOperators();
    virtual void removeOperators();

  private:

    pMesh mesh;
    SizeFieldManager * SFManager;

    mobileObjectSet * objects;

    // ----- Local mesh modification operators -----
    edgeSplitOp          *  eSplitOp;
    edgeCollapseOp       *  eCollapseOp;
    faceCollapseOp       *  fCollapseOp;
    DESCOp               *  descOp;
    edgeSwapOp           *  eSwapOp;
    faceSwapOp           *  fSwapOp;
    vertexMoveOp         *  vMoveOp;
    regionRemoveOp       *  rRegionOp;
    sliverFaceHandler    *  sliverFOp;
    sliverRegionHandler  *  sliverROp;
    bool activatedSplitOp;
    bool activatedSwapOp;
    bool activatedCollapseOp;

    // ----- Geometry related -----
    geoMatcher * geoTracker;

    // ----- Adaptation parameters -----
    int maxIterationsNumber;
    MeshParametersManager& mpm;
#ifdef PARALLEL
    loadBalanceAlgorithm load_balance_algorithm;
    MDB_DataExchanger*   dataExchanger;
#endif
    int updateSFFrequency;

    // ----- Output parameters -----
    int verbosity;
    std::string outputDirectoryPath;
    int debugLevel;
  };

  // -------------------------------------------------------------------

} // End of namespace MAd

#endif

