// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#ifndef _H_IMESHADAPTER
#define _H_IMESHADAPTER

#include "MAdExport.h"
#include "MeshDataBaseInterface.h"
#include "MAdOutput.h"

namespace MAd
{
    class MDB_Mesh;
    class MDB_MeshEntity;
    class MDB_Point;
    class PList;

    class ISizeField;
    class mobileObjectSet;

    // -------------------------------------------------------------------
    enum evaluationType {
        UNKNOWNEVALTYPE,
        MEANRATIO,
        GAMMA,
        MAX_DIHEDRAL_ANGLE,
        ORIENTEDMEANRATIO
    };
#ifdef PARALLEL
    enum loadBalanceAlgorithm {
        DEFAULT_ALGORITHM,
        METIS_ALGORITHM
    };
#endif

    enum operationType {
        MAd_UNKNOWNOPERATION,
        MAd_VERTEXMOVE,
        MAd_ESPLIT,
        MAd_ECOLLAPSE,
        MAd_FCOLLAPSE,
        MAd_DESPLTCLPS,
        MAd_ESWAP,
        MAd_FSWAP,
        MAd_RREMOVE
    };

    // -------------------------------------------------------------------
    // callback functions
    typedef void(*CBFunction)(PList *, PList *, void *, operationType, MDB_MeshEntity *);
    typedef void(*CBFunc_move)(MDB_Point *, double *, void *);

    // -------------------------------------------------------------------
    // -------------------------------------------------------------------
    class MAD_EXPORT IMeshAdapter
    {

    public:

        // -------------------------------------------------------------------

        static IMeshAdapter * BuildMeshAdapter(MDB_Mesh * mesh, ISizeField * sf = NULL);

        // -------------------------------------------------------------------

        virtual ~IMeshAdapter() {};

        // --------------------------------------------

        // gives an overview of the current parameters
        virtual void printParameters() const = 0;

        virtual void addCallback(CBFunction CB, void* userData,
            CBFunc_move CB_move = 0, void* userData_move = 0) = 0;
        virtual void addSizeField(ISizeField * sf) = 0;

        // whether or not the global size field is smoothed and maximum gradient
        virtual void setSizeFieldSmoothing(bool, double maxGrad = 1.) = 0;

        // set the maximum number of iterations available to reach the 'convergence' 
        // of the global mesh adaptation procedure
        virtual void setMaxIterationsNumber(int max) = 0;

#ifdef PARALLEL
        // set load balancing algorithm
        virtual void setLoadBalancingAlgorithm(loadBalanceAlgorithm lbAlgo) = 0;
#endif

        // choose the quality criterion
        virtual void setQualityMeasure(evaluationType type) = 0;

        // impose the interval governing edges length in the transformed space
        virtual void setEdgeLenSqBounds(double lower, double upper) = 0;

        // set permission for operators to modify slightly boundaries 
        // (default: false) and tolerance for the relative volume/area 
        // modifications.
        // Edge collapse and face collapse:
        virtual void setCollapseOnBoundary(bool accept = true, double tolerance = 1.e-6) = 0;
        // Edge swap:
        virtual void setSwapOnBoundary(bool accept = true, double tolerance = 1.e-6) = 0;

        // impose the element quality from which a swap is not required
        virtual void setNoSwapQuality(double noSwapQuality) = 0;

        // impose the maximum quality of a sliver
        virtual void setSliverQuality(double sliverQuality) = 0;

        virtual void activateSplitOp(bool t = true) = 0;
        virtual void activateSwapOp(bool t = true) = 0;
        virtual void activateCollapseOp(bool t = true) = 0;

        // Allow or forbid operations creating a sliver. 
        // If allowed and a bound is prescribed, it is allowed only if the edge is 
        // longer/shorter than the bound (bound is expressed as the square of 
        // adimensional edge length)
        virtual void setSliverPermissionInESplit(bool perm, double lenSqBound = -1.) = 0;
        virtual void setSliverPermissionInECollapse(bool perm, double lenSqBound = -1.) = 0;

        // tell if you want to project new vertices on geometric entities
        virtual void setGeoTracking(bool enable, bool cavityEqualMesh = false,
            int cavityThickness = 2, double chi = 1.0,
            bool strictChecking = false, bool forceRelocation = false) = 0;

        // set a value for a very huge edge length (default=1.e14)
        virtual void setInfiniteLength(double length) = 0;

        // frequency at which the size fields are updated inside the global loop
        // (useful for local and analytical SF)
        virtual void setSFUpdateFrequency(int freq) = 0;

        // constraint a mesh/geometric entity and all its downward entities
        virtual void clearConstraints() const = 0;
        virtual void setConstraint(pEntity e)        const = 0;
        virtual void setConstraint(pGEntity ge)      const = 0;
        virtual void setConstraint(int type, int id) const = 0;
        // unconstrain a geometric entity
        virtual void removeConstraint(int type, int id) const = 0;
        virtual void removeConstraint(pGEntity ge)      const = 0;

        // size field updates
        virtual void   updateSizeField() = 0;
        virtual void   updateSizeField(pVertex pv) = 0;

        // manage the physical time
        virtual void   incrementTime(double dt) = 0;
        virtual void   setTime(double t) = 0;
        virtual double getTime() const = 0;

        // add predefined mobile objects
        virtual void registerObjects(mobileObjectSet* objs) = 0;

        // ---------------- OPERATIONS ----------------

        // ----- Level 1 -----
        // ---> Elementary operations

        // Split the edge and if 'checkSize' is true, check that the two resulting 
        // edges are not short edges.
        virtual bool splitEdge(MDB_Edge * e, bool checkSize = false) = 0;
        // split edge at the parametric location t and return the new vertex
        virtual MDB_Point * splitEdge(MDB_Edge * e, double t, bool checkShapes = false) = 0;
        virtual bool collapseEdge(MDB_Edge * e) = 0;
        virtual bool collapseFace(MDB_Face * f, MDB_Edge * e) = 0;
        virtual bool DSplitCollapseEdge(MDB_Region * pr, MDB_Edge * edge1, MDB_Edge * edge2) = 0;
        virtual bool swapEdge(MDB_Edge * e) = 0;
        virtual bool swapFace(MDB_Face * f) = 0;
        virtual bool removeRegion(MDB_Region * region) = 0;
        virtual bool moveVertex(pVertex v, double dxyz[3]) = 0;
        virtual bool putVertex(pVertex v, double xyz[3]) = 0;

        // ----- Level 2 -----
        // ---> Loops on one elementary operation

        // node repositioning
        virtual double LaplaceSmoothing() = 0;

        // topology operations
        virtual int eSplitLoop() = 0;
        virtual int eCollapseLoop() = 0;
        virtual int eSplitCollapseLoop() = 0;
        virtual int edgeSwapLoop() = 0;
        virtual int faceSwapLoop() = 0;
        virtual int splitEveryEdgeOnce() = 0;
        virtual int mOptimiseLoop() = 0;

        // slivers handling
        virtual int removeSlivers() = 0;

        // geometry matching
        virtual void snapVertices() = 0;

        // ----- Level 3 -----
        // ---> Procedures with a global objective
        virtual int optimiseEdgeLength() = 0;
        virtual int optimiseElementShape() = 0;
        virtual int splitLongestEdges() = 0;
        virtual int runOneIter() = 0;
        virtual int removeNegativeElements() = 0;

        // ----- Level 4 -----
        // ---> Global procedure
        virtual int run() = 0;

        // ----- objects motion -----
        // move boundaries without repositioning nodes in the volume
        virtual int partlyMoveObjects(double t, double dt, double* part) = 0;
        // move boundaries and reposition all nodes with an elasticity analogy
        //   subAdaptation:    allow sliver elimination and mesh optimization if necessary
        //   qualityThreshold: at which quality do we optimize the mesh (if allowed)
        //   chi:              stiffness alteration coefficient (-1 = none)
        //   meshIsCavity:     true = elastic computation on whole mesh
        //   cavityThickness:  nb layers of elements if mesh is not the cavity
        virtual void moveObjectsAndReposition(double t, double dt,
            bool subAdaptation = true,
            double qualityThreshold = 0.,
            double chi = -1.,
            bool meshIsCavity = true,
            int cavityThickness = 3) = 0;

        // --------------------------------------------

        // ------ Outputs ------

        // set the path to output directory
        virtual void setOutputDirectoryPath(const char * prefix) = 0;

        // Set verbosity:
        //  < 0 : no detail
        //    1 : global procedure details
        //  > 2 : iterations details
        virtual void setVerbosity(int _verbosity) = 0;

        // write mesh with required postpro data in 'pos' format (Gmsh)
        // the output data can be any data listed in 'MAdOutput.h'
        virtual void writePos(const char * fn, MAdOutputData type = OD_CONSTANT) const = 0;

        // write mesh in 'msh' format (Gmsh)
        virtual void writeMsh(const char * fn) const = 0;

        virtual void getStatistics(double * meanQuality, double * worstQuality) const = 0;

    };

    // -------------------------------------------------------------------

} // End of namespace MAd

#endif

