// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#include "CurvedEdgeSplit.h"
#include "LinearOperatorTools.h"
#include "MeshDataBase.h"
#include "MAdMapping.h"
#include "MappingInterface.h"
#include "CallbackManager.h"

using namespace std;

namespace MAd {

 bool curvedEdgeSplit::checkConstraints() const
  {
   if(EN_constrained((pEntity) edge)) return false;
   return true;
  }

 bool curvedEdgeSplit::checkGeometry()
  {
   return true;
  }

 //----------------------------------------------------
 // method to get a cavity: elements surounding an edge
 void curvedEdgeSplit::getCavity(pPList * cavity) const
  {
   if (M_dim(mesh) ==3) *cavity = E_regions(edge);
   else                  *cavity = E_faces(edge);
  }

 //------------------------------------------------------
 // method devoted to the splitting of an edge
 // it results two edges of same order than original edge 
 void curvedEdgeSplit::splitCreateEdges(pEdge splittedEdge, double lscale, int patch, pVertex vMiddle, pEdge e1, pEdge e2)
  {
   // define a mapping from the 1d ref 
   pMapping mapping=Map_create(splittedEdge,lscale);
   // construct the middle vertex of the edge
   double uvw[3];
   for(int i=0; i<3; i++) uvw[i]=0.;
   double coord[3];
   Map_eval(mapping, uvw, xyz);
   pGEntity pGE=E_whatIn(splittedEdge);
   vMiddle=M_createV2(mesh, xyz, patch, pGE);
    
   // get the order of the original edge
   // int o=EN_order(edge);
   int o=EN_order(splittedEdge);
   // create left and right edge resulting from splitting
   pVertex *HOvertices = new pVertex[(o+1)-2];
   if(o>1)
    {
     // create the HO vertices of the left edge
     for(int i=1; i<o; i++)
      {
       uvw[0]=-1.+i*1./o;
       uvw[1]=0.;
       uvw[2]=0.;
       Map_eval(mapping, uvw, coord);
       HOvertices[i-1]=M_createV2(mesh,coord,patch,pGE);       
      }
     // e1 = M_createE(mesh,E_vertex(edge,0),vMiddle,o,HOvertices,pGE); 
     e1 = M_createE(mesh,E_vertex(splittedEdge,0),vMiddle,o,HOvertices,pGE); 
     // create the HO vertices of the right edge
     for(int i=1; i<o; i++)
      {
       uvw[0]=0.+i*1./o;
       uvw[1]=0.;
       uvw[2]=0.;
       Map_eval(mapping, uvw, coord);
       HOvertices[i-1]=M_createV2(mesh,coord,patch,/*splittedEdge*/pGE);       
      }
     // e2 = M_createE(mesh, vMiddle, E_vertex(edge,1), o, HOvertices, /*splittedEdge*/pGE);     
     e2 = M_createE(mesh, vMiddle, E_vertex(splittedEdge,1), o, HOvertices, /*splittedEdge*/pGE);     
    }
   delete mapping;
   delete [] HOvertices;
  }

 //-----------------------------------------------------------
 // method devoted to the creation of the median edge that link
 // the middle of the splitted edge and the opposite vertex 
 // the edge created is of same order of the splitted edge
 void curvedEdgeSplit::createMedianEdge(pFace face, pEdge splittedEdge, pVertex vMiddle, double lscale, int i, int patch, pEdge medianEdge)
  {
   // define a mapping from reference triangle
   pMapping mapping = Map_create(face, lscale);
   int o = EN_order(splittedEdge);
   pVertex *HOvertices = new pVertex[o-1];
   double uvw[3],uvwB[3],uvwE[3];
   double coord[3];
   pGEntity pGEFace=F_whatIn(face);
   // pGEntity pGEEdge=E_whatIn(edge);
   // pGEntity pGEEdge=E_whatIn(splittedEdge);
   // switch on splitted edge id 
   // to get the bounding vertices of the median edge
   switch(i){
    case 1:
    { 
     uvwB[0]=0.5; uvwB[1]=0.0; uvwB[2]=0.0;
     uvwE[0]=0.0; uvwE[1]=1.0; uvwE[2]=0.0;
     break;
    }
   case 2:
    {
     uvwB[0]=0.5; uvwB[1]=0.5; uvwB[2]=0.0;
     uvwE[0]=0.0; uvwE[1]=0.0; uvwE[2]=0.0;
     break;
    }
   case 3:
    {
     uvwB[0]=0.0; uvwB[1]=0.5; uvwB[2]=0.0;
     uvwE[0]=1.0; uvwE[1]=0.0; uvwE[2]=0.0;
     break;
    }
   default:
    {
     std::cout << "splitting methods are only implemented for triangles" << std::endl;
     std::cout << "edge id must be 1, 2 or 3" << std::endl;
     return;
    }
   }
   // build the high order points belonging to edges 
   // firstly in the parametric space, then in physical space using mapping 
   for(int k=1; k<o; k++)
    {
     for(int index=0; index<3; index++) uvw[index]=uvwB[index]+((k*1./o)*(uvwE[index]-uvwB[index]));
     Map_eval(mapping,uvw,coord);
     // HOvertices[k-1]=M_createV2(mesh,coord,patch,pGEEdge);
     HOvertices[k-1]=M_createV2(mesh,coord,patch,pGEFace);
    } 
   // build the "high order" median edge in physical space
   medianEdge = M_createE(mesh, vMiddle, F_edOpVt(face,splittedEdge), o, HOvertices, pGEFace);
   delete mapping;
   delete [] HOvertices;
  } 
 
 //---------------------------------------------------
 // method that creates the high order points located 
 // inside the face (order greater or equal to 3)
 void curvedEdgeSplit::createFaceHOVertices(pFace face, pEdge splittedEdge, double lscale, int patch, pVertex internalVertices1[], pVertex internalVertices2[])
  {
   double uvw[3];
   double coord[3];
   int o = EN_order(face);
   pMapping mapping=Map_create(face,lscale);
   for(int i=0; i<3; i++)
    {
     uvw[i] = 0.;
     coord[i] = 0.;
    }
   pGEntity pGEFace=F_whatIn(face);
   double *u1;
   double *v1;
   double *u2;
   double *v2;
//   int nb=((o-3)+1)*((o-3)+2)/2;
   int nb=(o-2)*(o-1)/2;
   u1=new double[nb];
   v1=new double[nb];
   u2=new double[nb];
   v2=new double[nb];
   double uv11[2];
   double uv12[2];
   double uv13[2];
   double uv21[2];
   double uv22[2];
   double uv23[2];
   
   if(splittedEdge==F_edge(face,1))
    {
// first face generated
     uv11[0]=0.0; uv11[1]=0.0;
     uv12[0]=0.5; uv12[1]=0.0;
     uv13[0]=0.0; uv13[1]=1.0;
// 2nd face generated
     uv21[0]=0.5; uv21[1]=0.0;
     uv22[0]=1.0; uv22[1]=0.0;
     uv23[0]=0.0; uv23[1]=1.0;
    }
   else if(splittedEdge==F_edge(face,2))
    {
// first face generated
     uv11[0]=0.0; uv11[1]=0.0;
     uv12[0]=1.0; uv12[1]=0.0;
     uv13[0]=0.5; uv13[1]=0.5;
// 2nd face generated
     uv21[0]=0.0; uv21[1]=0.0;
     uv22[0]=0.5; uv22[1]=0.5;
     uv23[0]=0.0; uv23[1]=1.0;      
    }   
   else if(splittedEdge==F_edge(face,3))
    {
// first face generated
     uv11[0]=0.0; uv11[1]=0.0;
     uv12[0]=1.0; uv12[1]=0.0;
     uv13[0]=0.0; uv13[1]=0.5;
// 2nd face generated
     uv21[0]=0.0; uv21[1]=0.5;
     uv22[0]=1.0; uv22[1]=0.0;
     uv23[0]=0.0; uv23[1]=1.0;      
    } 
   else
    {
     std::cout<< "edge id must be between 1, 2 or 3 !" << std::endl;
     return;
    }
// build the high-order points located inside the elements
// which define an element of order o-3 translated and reduced  
   controlPointsTriangle(uv11,uv12,uv13,o-3,u1,v1);
   controlPointsTriangle(uv21,uv22,uv23,o-3,u2,v2);
   for(int k=0; k<nb; k++)
    {
     // 1st new element
     uvw[0]=1/o*u1[k]+(uv12[0]-2*uv11[0]+uv13[0])/o;
     uvw[1]=1/o*v1[k]+(uv12[1]-2*uv11[1]+uv13[1])/o;
     uvw[2]=0.0;
     Map_eval(mapping,uvw,coord);
     internalVertices1[k]=M_createV2(mesh,coord,patch,pGEFace);
     // 2nd new element
     uvw[0]=1/o*u2[k]+(uv22[0]-2*uv21[0]+uv23[0])/o;
     uvw[1]=1/o*v2[k]+(uv22[1]-2*uv21[1]+uv23[1])/o;
     uvw[2]=0.0;
     Map_eval(mapping,uvw,coord);
     internalVertices2[k]=M_createV2(mesh,coord,patch,pGEFace);
    }
   Map_delete(mapping);
   delete [] u1;
   delete [] u2;
   delete [] v1;
   delete [] v2;
  }

 //-------------------------------------------------------
 // method that manages a 2D cavity in which an edge is splitted
 //
 void curvedEdgeSplit::splitCavity2d(double lscale,int patch)
  {
   pPList newEdges = PList_new();
   pPList newFace = PList_new();
   pVertex vMiddle=NULL;
   pEdge e1=NULL;
   pEdge e2=NULL;
   splitCreateEdges(edge, lscale, patch, vMiddle, e1, e2);
   PList_append(newEdges,e1);
   PList_append(newEdges,e2);
   // vertices creation
   int o = EN_order(edge);
   const int nb = (o+1)*(o+2)/2;
   pVertex *face1Vertices = new pVertex[nb];
   int *face1VertexId = new int[nb];
   pVertex *face2Vertices = new pVertex[nb];
   int *face2VertexId = new int[nb];
   // store the 2 first principal vertices of the two faces resulting from splitting
   face1Vertices[0]=E_vertex(edge,0); face2Vertices[0]=vMiddle;
   face1Vertices[1]=vMiddle; face2Vertices[1]=E_vertex(edge,1);
   // edge creation
   PList_append(newEdges, (pEntity) e1);
   PList_append(newEdges, (pEntity) e2);
   // get cavity
   pPList eFaces = E_faces(edge);
   void * temp = NULL;
   pFace f=NULL;
   pEdge medianEdge=NULL;
   //
   while((f=(pFace) PList_next(eFaces,&temp)))
    {
     // store the 3rd principal vertice 
     face1Vertices[2]=F_edOpVt(f,edge);
     face2Vertices[2]=F_edOpVt(f,edge);
     int i=1;
     bool found=false;
     while((!found)&&(i<4))
      {
       if(edge==F_edge(f,i)) {found=true;}
       else{i++;}
      }
     if(found){
      createMedianEdge(f, edge, vMiddle, lscale, i, patch, medianEdge);
      PList_append(newEdges, medianEdge);
      int nbIn=E_numPoints(edge);
      // store the high order control points inside edges
      for(int k=0; k<nbIn; k++) 
       {
        face1Vertices[3+k]=(pVertex) E_point(e1,k); face2Vertices[3+k]=(pVertex) E_point(e2,k);
        face1Vertices[3+k+(nbIn+1)]=(pVertex) E_point(medianEdge,k); face2Vertices[3+k+(nbIn+1)]=(pVertex) E_point(F_vtOpEd(f,E_vertex(edge,1)),k);
        face1Vertices[3+k+(2*nbIn+1)]=(pVertex) E_point(F_vtOpEd(f,E_vertex(edge,0)),k); face2Vertices[3+k+(2*nbIn+1)]=(pVertex) E_point(medianEdge,k);
       }
      // store the high-order points located in the face which 
      // exists when order is gep than 3
      // warning: MAdLib does not support order>4
      if(o>2)
       {
        int nbInt = ((o-3)+1)*((o-3)+2)/2; 
        pVertex *internalVertices1 = new pVertex[nbInt];
        pVertex *internalVertices2 = new pVertex[nbInt];
        createFaceHOVertices(f, edge, lscale, patch, internalVertices1, internalVertices2);
        for(int k=0; k<nbInt; k++)
         {
          face1Vertices[3+2*nbIn-1+k]=internalVertices1[k];
          face2Vertices[3+2*nbIn-1+k]=internalVertices2[k];
         } 
         delete [] internalVertices1;
         delete [] internalVertices2;
       }
      for(int k=0; k<nb; k++)
       {
        face1VertexId[k]=V_id(face1Vertices[k]);
        face2VertexId[k]=V_id(face2Vertices[k]);
       }
      pGEntity gent = (pGEntity) F_whatIn(f);
      // M_createTri is only implemented for order<5
      // see MeshDataBaseInterface.cc file
      PList_append(newFace,M_createTri(mesh, EN_order(f),face1VertexId,gent));
      PList_append(newFace,M_createTri(mesh, EN_order(f),face2VertexId,gent));   
     }
     // delete original face to keep only the faces resulting from splitting operation
     M_removeFace(mesh,f);
    }    
   // delete splitted edge
   M_removeEdge(mesh,edge);
   // free memory
   PList_delete(newEdges);
   PList_delete(newFace);
   delete [] face1Vertices;
   delete [] face2Vertices;
   delete [] face1VertexId;
   delete [] face2VertexId;
  }

//-------------------------------------------------------------------
// method that split a tetraedron considering a long edge
// it results 2 tetraedra built on the splitted edge 
//
 void curvedEdgeSplit::splitElement(pRegion originElt, pEdge splittedEdge, double lscale, int patch)
  {
   pEdge medianEdge[2] = {NULL,NULL};
   pEdge e1=NULL, e2=NULL;
   pVertex vMiddle=NULL;
   pVertex verts[2];
   pGEntity gent=(pGEntity) R_whatIn(originElt);
   int k=0;
   for(int indexF=0; indexF<R_numFaces(originElt); indexF++)
    {
     pFace pf=R_face(originElt,indexF);
     if(F_inClosure(pf,(pEntity)splittedEdge))
      {
       splitCreateEdges(splittedEdge, lscale, patch, vMiddle, e1, e2);
       verts[k]=F_edOpVt(pf,splittedEdge);
       for(int indexE=0; indexE<F_numEdges(pf); indexE++)
        {
         if((F_edge(pf,indexE))==splittedEdge)
          {  
           createMedianEdge(pf, splittedEdge, vMiddle, lscale, indexE, patch, medianEdge[k]);
           k++;
          }
        }
      }
    }
   const int numV=R_numVertices(originElt);
   pVertex *newVerts = new pVertex[numV];
   newVerts[0]=E_vertex(e1,0);
   newVerts[1]=vMiddle;
   newVerts[2]=verts[0];
   newVerts[3]=verts[1];
   pRegion newRegion;
   newRegion=M_createR(mesh,numV,newVerts,gent);
   pPList newRegions = PList_new();
   pPList oldRegions = PList_new();
   PList_append(oldRegions,originElt);
   PList_append(newRegions,newRegion);
   newVerts[0]=vMiddle;
   newVerts[1]=E_vertex(e2,1);
   newVerts[2]=verts[0];
   newVerts[3]=verts[1];
   newRegion=M_createR(mesh,numV,newVerts,gent);
   PList_append(newRegions,newRegion);
   CallBackManagerSgl::instance().callCallBacks(oldRegions, newRegions, MAd_ESPLIT, (pEntity)edge );
   PList_delete(newRegions); 
   PList_delete(oldRegions);
   delete [] newVerts;
  }  

//---------------------------------------------------------------------------
// split all the tetraedra belonging to a cavity that surrounds a long edge
//
 void curvedEdgeSplit::splitCavity3d(pEdge splittedEdge, double lscale, int patch)
  {
    // get cavity
    pPList eRegions = E_regions(edge);
    void * temp = NULL;
    pRegion r;
    //
    while((r=(pRegion) PList_next(eRegions,&temp)))
     {
      splitElement(r,splittedEdge,lscale,patch);
     }
    M_removeEdge(mesh, splittedEdge);
    PList_delete(eRegions);
  }

}
/* void curvedEdgeSplit::getCavity(pPList * cavity) const
  {
   if (dim==3) *cavity = E_regions(edge);
   else        *cavity = E_faces(edge);
  }

 void curvedEdgeSplit::setCavitySize(pPList cavity)
  {
   cavitySize=PList_size(cavity);
  }

 void curvedEdgeSplit::getMidCoord3d(pRegion elt, double lscale, double coord[3])
  {
   double uvw[3];
   int nEdge;
   int i;
   pMapping map=Map_create(elt,lscale);
   nEdge = R_numEdges(elt);
   i=1;
   while((i<=nEdge)&&(R_edge(elt,i)!=edge))
    {
     i++;
    }
   if(R_edge(elt,i)==edge)
    {
     switch(nEdge){
      case 6:
      // element is a tet
       if(i==1){
        uvw[0] = 0.5; uvw[1] = 0.0; uvw[2] = 0.0;
       }
       else if(i==2){
        uvw[0] = 0.5; uvw[1] = 0.5; uvw[2] = 0.0;
       }
       else if(i==3){
        uvw[0] = 0.0; uvw[1] = 0.5; uvw[2] = 0.0;
       }
       else if(i==4){
        uvw[0] = 0.0; uvw[1] = 0.0; uvw[2] = 0.5;
       }
       else if(i==5){
        uvw[0]=0.5; uvw[1] = 0.5; uvw[2] = 0.0;
       }
       else if(i==6){
        uvw[0] = 0.0; uvw[1] = 0.5; uvw[2] = 0.5;
       }
       else{
        printf("index is out of range edge numbering"); return ;
       }
       break;
      case 8:
      // element is a pyramid 
       if(i==1){
        uvw[0] = 0.0; uvw[1] = -1.0; uvw[2] = 0.0;
       }
       else if(i==2){
        uvw[0] = 1.0; uvw[1] = 0.0; uvw[2] = 0.0;
       }
       else if(i==3){
        uvw[0] = 0.0; uvw[1] = 1.0; uvw[2] = 0.0;
       }
       else if(i==4){
        uvw[0] = -1.0; uvw[1] = 0.0; uvw[2] = 0.0;
       }
       else if(i==5){
        uvw[0] = -0.5; uvw[1] = -0.5; uvw[2] = 0.5;
       }
       else if(i==6){
        uvw[0] = 0.5; uvw[1] = -0.5; uvw[2] = 0.5;
       }
       else if(i==7){
        uvw[0] = 0.5; uvw[1] = 0.5; uvw[2] = 0.5;
       }
       else if(i==8){
        uvw[0] = -0.5; uvw[1] = 0.5; uvw[2] = 0.5;
       }
       else{
        printf("index is out of range edge numbering"); return ;
       }         
       break;
      case 9:
      // element is a prism
       if(i==1){
        uvw[0] = 0.5; uvw[1] = 0.0; uvw[2] = -1.0;
       }
       else if(i==2){
        uvw[0] = 0.5; uvw[1] = 0.5; uvw[2] = -1.0;
       }
       else if(i==3){
        uvw[0] = 0.0; uvw[1] = 0.5; uvw[2] = -1.0;
       }
       else if(i==4){
        uvw[0] = 0.5; uvw[1] = 0.0; uvw[2] = 1.0;
       }
       else if(i==5){
        uvw[0] = 0.5; uvw[1] = 0.5; uvw[2] = 1.0;
       }
       else if(i==6){
        uvw[0] = 0.0; uvw[1] = 0.5; uvw[2] = 1.0;
       }
       else if(i==7){
        uvw[0] = 0.5; uvw[1] = 0.0; uvw[2] = 0.0;
       }
       else if(i==8){
        uvw[0] = 0.5; uvw[1] = 0.5; uvw[2] = 0.0;
       }
       else if(i==9){
        uvw[0] = 0.0; uvw[1] = 0.5; uvw[2] = 0.0;
       }
       else{
        printf("index is out of range edge numbering"); return ;
       }
       break;
      case 12:
      // element is a hexaedron
       if(i==1){
        uvw[0] = 0.0; uvw[1] = -1.0; uvw[2] = -1.0;
       }
       else if(i==2){
        uvw[0] = 1.0; uvw[1] = 0.0; uvw[2] = -1.0;
       }
       else if(i==3){
        uvw[0] = 0.0; uvw[1] = 1.0; uvw[2] = -1.0;
       }
       else if(i==4){
        uvw[0] = -1.0; uvw[1] = 0.0; uvw[2] = -1.0;
       }
       else if(i==5){
        uvw[0] = 0.0; uvw[1] = -1.0; uvw[2] = 1.0;          
       }
       else if(i==6){
        uvw[0] = 1.0; uvw[1] = 0.0; uvw[2] = 1.0;
       }
       else if(i==7){
        uvw[0] = 0.0; uvw[1] = 1.0; uvw[2] = 1.0;
       }
       else if(i==8){
        uvw[0] = -1.0; uvw[1] = 0.0; uvw[2] = 1.0;
       }
       else if(i==9){
        uvw[0] = 0.0; uvw[1] = -1.0; uvw[2] = 0.0;
       }
       else if(i==10){
        uvw[0] = 1.0; uvw[1] = 0.0; uvw[2] = 0.0;
       }
       else if(i==11){
        uvw[0] = 0.0; uvw[1] = 1.0; uvw[2] = 0.0;
       }
       else if(i==12){
        uvw[0] = -1.0; uvw[1] = 0.0; uvw[2] = 0.0;
       }
       else{
        printf("index is out of range edge numbering"); return ;
       }
       break;
      default:
       printf("polyedra are not supported by MAdLib"); return ;
       break;
     }
    }
   else
    {
     printf("edge doesn't belong to element"); return ;
    }
  // compute the image of the mid-point in the physical space
  Map_eval(map,uvw,coord);
  Map_delete(map);
 }

//-------------------------------------------------------------------------
 void curvedEdgeSplit::getMidCoord2d(pFace face, double lscale, double coord[3])
  {
   double uvw[3];
   int nEdge;
   int i;
   pMapping map=Map_create(face,lscale);
   nEdge = F_numEdges(face);
   i=1;
   while((i<=nEdge)&&(F_edge(face,i)!=edge))
    {
     i++;
    }
   if(F_edge(face,i)==edge)
    {
     switch(nEdge){
      case 3:
      // element is a triangle
       if(i==1){
        uvw[0] = 0.5; uvw[1] = 0.0; uvw[2] = 0.0;
       }
       else if(i==2){
        uvw[0] = 0.5; uvw[1] = 0.5; uvw[2] = 0.0;
       }
       else if(i==3){
        uvw[0] = 0.0; uvw[1] = 0.5; uvw[2] = 0.0;
       }
       else{
        printf("index is out of range edge numbering"); return ;
       }
       break;
      case 4:
      // element is a quadrangle
       if(i==1){
        uvw[0] = 0.0; uvw[1] = -1.0; uvw[2] = 0.0;
       }
       else if(i==2){
        uvw[0] = 1.0; uvw[1] = 0.0; uvw[2] = 0.0;
       }
       else if(i==3){
        uvw[0] = 0.0; uvw[1] = 1.0; uvw[2] = 0.0;
       }
       else if (i==4){
        uvw[0] = -1.0; uvw[1] = 0.0; uvw[2] = 0.0;
       }
       else{
        printf("index is out of range edge numbering"); return ;
       }
       break;
      default:
       printf("polyedra are not supported by MAdLib"); return ;
       break;
     }
    }
    else{
     printf("edge doesn't belong to element"); return ;
    }
   // compute the image of the mid-point in the physical space
   Map_eval(map,uvw,coord);
   Map_delete(map);
  }

//--------------------------------------------------------------------- 
 void curvedEdgeSplit::getMidCoord(double lscale)
  {
   pPList cavity=PList_new();
   getCavity(&cavity);
   setCavitySize(cavity);
   *midCoordList = new double[cavitySize];
   for(int i=0; i<cavitySize; i++){midCoordList[i] = new double[3];}
   int index=0;
   void * tmp =NULL;
   switch(dim){
    case 3:
     pRegion r;
     while ((r=(pRegion)PList_next(cavity,&tmp)))
     {
      getMidCoord3d(r, lscale, midCoordList[index]);
      index++;
     }
     break;
    case 2:
     pFace f;
     while ((f=(pFace)PList_next(cavity,&tmp)))
     {
      getMidCoord2d(f, lscale, midCoordList[index]);
      index++;
     }
     break;
    default:
     printf("dimension must be 2 or 3"); return;
     break;
    }
   PList_delete(cavity);
  }

//---------------------------------------------------------------------
 void curvedEdgeSplit::createSplitCavity2d(double lscale)
  {
    // get cavity
    pPList eFaces = E_faces(edge);
    void * temp = NULL;
    pFace f;
    double coord[3];
    pPList newEdges = PList_new();
    pPList newFaces = PList_new();
    pVertex vMiddle = NULL;
//
    while((f=(pFace) PList_next(eFaces,&temp)))
     {
      // get coordinates of middle edge
      getMidCoord2d(f,lscale,coord);
      // create a new vertex
      pGEntity eGEntity = E_whatIn(edge);
      vMiddle = M_createV(mesh, coord[0], coord[1], coord[2], -1 ,eGEntity);
      // create the 2 edges resulting from splitting
        //1) find the 2 vertices linking the original edge
      pVertex v0 = E_vertex(edge,0);
      pVertex v1 = E_vertex(edge,1);
        //2) create the 2 edges
      pEdge fEdges1[3];
      fEdges1[0] = M_createE(mesh,v0, vMiddle, eGEntity);
      PList_append(newEdges, (pEntity) fEdges1[0]);
      pEdge fEdges2[3];
      fEdges2[0] = M_createE(mesh, vMiddle, v1, eGEntity);
      PList_append(newEdges, (pEntity) fEdges2[0]);
      // create the edge linking the new vertex to the vertex opposed to the splitted edge
        //1) get the vertex opposite to edge belonging to face
      pVertex vOpposite = F_edOpVt(f,edge);
        //2) create the edges completing a face
      fEdges1[1] = M_createE(mesh, vMiddle, vOpposite, eGEntity);
      fEdges1[2] = M_createE(mesh, vOpposite, v0, eGEntity);
      fEdges2[1] = M_createE(mesh, v1, vOpposite, eGEntity);
      fEdges2[2] = M_createE(mesh, vOpposite, vMiddle, eGEntity); 
      
      // create the two new faces
      pGEntity fGEntity = F_whatIn(f);
      pFace newFace[2];
      newFace[0] = M_createF(mesh, 3, fEdges1, fGEntity);
      newFace[1] = M_createF(mesh, 3, fEdges2, fGEntity);
      for(int i=0; i<2; i++) {
       PList_append(newFaces, (pEntity) newFace[i]);
      }
     }
    pPList delEdge = PList_new();
    PList_append(delEdge,(pEntity) edge);
    CallBackManagerSgl::instance().callCallBacks(delEdge, newEdges, 
                                                 MAd_ESPLIT, (pEntity) vMiddle);
    PList_delete(delEdge);
    PList_delete(newEdges);
//
    temp = NULL;
    while( pFace face = (pFace)PList_next(eFaces,&temp) ) {

      EN_removeData((pEntity)face,"_EdgeSplitMarker_");
      
      for(int iE=0; iE<F_numEdges(face); iE++) {
        pEdge iEdge = F_edge(face,iE);
        if( iEdge != edge ) EN_removeData((pEntity)iEdge,"_EdgeSplitMarker_");
      }

      M_removeFace(mesh, face);
    }
    PList_delete(eFaces);
    
    temp = NULL;
    while( pFace face = (pFace)PList_next(newFaces,&temp) ) {
      EN_removeData((pEntity)face,"_EdgeSplitMarker_");
    }
    PList_delete(newFaces);
    
    // --- remove edge ---
    M_removeEdge(mesh,edge);
  }

//--------------------------------------------------------------------
 void curvedEdgeSplit::createSplitCavity3d(double lscale)
  {
    // get cavity
    pPList eRegions = E_regions(edge);
    void * temp = NULL;
    pRegion r;
    pVertex vMiddle = NULL;
    double coord[3];
    // vertices of the new regions introduced by splitting edge
    pVertex vertsR1[4];
    pVertex vertsR2[4];
    pPList newRegions = PList_new();
    //    
    //
    while((r=(pRegion) PList_next(eRegions,&temp)))
     {
      getMidCoord3d(r, lscale, coord);
      // create the new regions based on the middle of the edge
      pGEntity eGEntity = E_whatIn(edge);
      vMiddle = M_createV(mesh, coord[0], coord[1], coord[2], -1 ,eGEntity);
      // create the 2 edges resulting from splitting
        //1) find the 1st vertex of the original edge and start to build the first new region
      pVertex v0 = E_vertex(edge,0);
      vertsR1[0]=v0;
      vertsR1[1]=vMiddle;
        //2) find the 2nd vertex of the original edge and start to build the second new region
      pVertex v1 = E_vertex(edge,1);
      vertsR2[0]=vMiddle;
      vertsR2[1]=v1;

      pPList rFaces = R_faces(r);
      pFace f;
      void * tmp = NULL;
      bool found=false;
      while((f=(pFace) PList_next(rFaces,&tmp))&&(!found))
       {
        //3) find the 3rd vertex of the new regions: opposite vertex of a long edge
        pVertex vFOppo = F_edOpVt(f,edge);
        vertsR2[2]=vertsR1[2]=vFOppo;
        if(F_inClosure(f,(pEntity) edge))
         {
          //4) find the 4th vertex of the new regions: opposite vertex of the previously defined face
          pVertex vROppo = R_fcOpVt(r, f);
          vertsR2[3]=vertsR1[3]=vROppo;
          found=true;
         }
       }
      // create the new regions based on the vertices
      pRegion newRegion = M_createR(mesh, 4, vertsR1, eGEntity);
      PList_append(newRegions,newRegion);
      newRegion = M_createR(mesh, 4, vertsR2, eGEntity);
      PList_append(newRegions,newRegion);
     }
    CallBackManagerSgl::instance().callCallBacks(eRegions, newRegions, MAd_ESPLIT, (pEntity)edge );
    PList_delete(newRegions);      
    //  Remove old regions connected to edge
    pRegion drgn;
    while( ( drgn = (pRegion)PList_next(eRegions, &temp) ) )
      M_removeRegion(mesh,drgn);
    PList_delete(eRegions);
    // remove old faces connected to edge
    pFace face;
    pPList faces = E_faces(edge);
    while( ( face = (pFace)PList_next(faces, &temp) ) )
      M_removeFace(mesh, face);
    PList_delete(faces);

    // remove edge
    M_removeEdge(mesh, edge);
  }

}*/

//---------------------------------------------------------------------


