// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#include "CurvedOperatorTools.h"
#include "CallbackManager.h"
#include "MAdMessage.h"
#include "Constraint.h"
#include "MappingInterface.h"
#include "MAdDefines.h"
#include "regularizer.h"

#include <map>
#include <assert.h>

using namespace std;

namespace MAd {
  
  namespace CurvedToolBox {
    // ---------------------------------------------------------------------------------------
    
    // ---------------------------------------------------------------------------------------
    //! Check if a vertex is on a boundary(Line in 2D and Face in 3D)
    bool isVertexOnBoundary( pVertex V)
    {
      void *temp = NULL;
      
      // 3D config ??
      if (V_numRegions(V))
      {
        pPList fList = V_faces(V);
        while ( pFace face = (pFace) PList_next(fList,&temp) )
        {
          if ( (F_whatInType(face) <= 2) && (F_numRegions(face) == 1) ) 
          {
            PList_delete(fList);
            return true;
          }
        }
      }
      else
      {
        pPList eList = V_edges(V);
        while ( pEdge edge = (pEdge) PList_next(eList,&temp) )
        {
          if ( (E_whatInType(edge) <= 1) ) 
          {
            PList_delete(eList);
            return true;
          }
        }
      }
      return false;
    }
    
    // ---------------------------------------------------------------------------------------
    //! Curved Edge Length evaluated with Gauss quadratures order 7
    double edgeLength(pEdge edge, double uend)
    {
      const int nb = 4;
      double quad[4][2] = { {-0.8611363115,0.3478548451},
                            {-0.3399810435,0.6521451548},
                            { 0.3399810435,0.6521451548},
                            { 0.8611363115,0.3478548451} };
      double length = 0.0;
      double uvw[3],deriv[9];
      memset(uvw, 0 ,3*sizeof(double));
      
      double du = (uend + 1.0) / 2.0;
      
      pMapping mapping = Map_create(edge,1.0);
      for (int pt = 0 ; pt < nb ; pt++)
      {
        uvw[0] = -1.0 + ( quad[pt][0] + 1.0 )*du;
        Map_deval(mapping,uvw,deriv);
        double dtdu = sqrt(deriv[0]*deriv[0] + deriv[1]*deriv[1] + deriv[2]*deriv[2]);
        length += dtdu*quad[pt][1];
      }
      Map_delete(mapping);
      length *=du;
      return length;
    }
    //! Evaluate Hessian of an edge at parameter "u"
    double edgeHessian(pEdge edge, double u)
    {
      double deriv[9];
      pMapping mapping = Map_create(edge,1.0);
      
      double tmp[3] = {0.,0.,0.};
      tmp[0] = std::min(u + 0.05,1.0);
      Map_deval(mapping,tmp,deriv);
      double dtdup = sqrt(deriv[0]*deriv[0] + deriv[1]*deriv[1] + deriv[2]*deriv[2]);
      tmp[0] = std::max(u - 0.05,-1.0);
      Map_deval(mapping,tmp,deriv);
      double dtdum = sqrt(deriv[0]*deriv[0] + deriv[1]*deriv[1] + deriv[2]*deriv[2]);
      
      double H = (dtdup - dtdum) / 0.10;
        
      Map_delete(mapping);
      return H;
    }
    
    // ---------------------------------------------------------------------------------------
    //! Regularize HO nodes on edge based on curved length to ensure equidistance in physical space
    //! \note : not in use
    double regularizeEdge(pEdge &edge)
    {
      pMapping mapping = NULL;
      const double alpha = 0.9;
      int o = EN_order(edge);
      
      int nb = edge->getNbHighOrderPoints();
      
      double *uvw = new double[nb*3];
      double dt, error = 1.e+10;
      double xyz[3];
      double deriv[9];
      double t_ideal,t_curr;
      memset(uvw,0,3*nb*sizeof(double));
      
      double fullLength = edgeLength(edge);
      int it = 20;
      while(error > 1.e-6 && it--)
      {
        error = 0.0;
        // initialise with perfect u
        for (int pt = 0; pt < nb; pt++) uvw[3*pt] = -1.0 + (pt+1) * 2.0 / o;
        
        
        mapping = Map_create(edge,1.0);
        for (int pt = 0; pt < nb; pt++)
        {
          t_ideal = (pt+1) * 1.0 / o;
          t_curr = edgeLength(edge,uvw[3*pt]) / fullLength;
          dt = (t_curr-t_ideal);
          
          // TODO : To test ...
          mapping->invertedJacobian(&uvw[3*pt],deriv);
          
          double dudt = sqrt(deriv[0]*deriv[0] + deriv[3]*deriv[3] + deriv[6]*deriv[6] );
          
          uvw[3*pt] -= alpha*dt*dudt;
          
          error = max(error,abs(dt));
        }
        
        // move the HO nodes
        for (int pt = 0; pt < nb; pt++)
        {
          pVertex v = edge->getHighOrderPoint(pt);
          Map_eval(mapping,&uvw[3*pt],xyz);
          V_setCoord(v,xyz);
        }
      }
      Map_delete(mapping);
      
      
      delete [] uvw;
      return error;
    }
    // ---------------------------------------------------------------------------------------
    
    // ---------------------------------------------------------------------------------------
    //! Create a High Order (Linear) Edge from v1 to v2 of order "order" 
    pEdge createHOEdge( pMesh mesh, pVertex v1, pVertex v2, int order, pGEntity eGEntity)
    {
      return createLinearHOEdge(mesh, v1, v2, order, eGEntity);
    }
    
    //! Create an High Order Curved Tetra based on its Bounding vertices. 
    //! - High order Edges and Faces are created from Linear/Serendip elements if they does not exists
    //! - High order internal Nodes are created based on Serendip Tetra mapping
    pRegion createHOTetraFromVertices(pMesh mesh, pVertex verts[4], int o, pGEntity pGE)
    {
      
      pFace t1 , t2 , t3 , t4;
      
      // [1] Create the Edges
      pEdge e1 = mesh->find_edge(verts[0],verts[1]);
      if (!e1) e1 = createHOEdge(mesh, verts[0],verts[1], o, pGE);
      pEdge e2 = mesh->find_edge(verts[1],verts[2]);
      if (!e2) e2 = createHOEdge(mesh, verts[1],verts[2], o, pGE);
      pEdge e3 = mesh->find_edge(verts[2],verts[0]);
      if (!e3) e3 = createHOEdge(mesh, verts[2],verts[0], o, pGE);
      pEdge e4 = mesh->find_edge(verts[0],verts[3]);
      if (!e4) e4 = createHOEdge(mesh, verts[0],verts[3], o, pGE);
      pEdge e5 = mesh->find_edge(verts[1],verts[3]);
      if (!e5) e5 = createHOEdge(mesh, verts[1],verts[3], o, pGE);
      pEdge e6 = mesh->find_edge(verts[2],verts[3]);
      if (!e6) e6 = createHOEdge(mesh, verts[2],verts[3], o, pGE);
      
      // [2] Create the Faces
      t1 = mesh->find_triangle(e1, e2, e3);
      if(!t1)t1 = createHOFaceFromEdges(mesh, e1, e2, e3, pGE);
      t2 = mesh->find_triangle(e1, e4, e5);
      if(!t2)t2 = createHOFaceFromEdges(mesh, e1, e4, e5, pGE);
      t3 = mesh->find_triangle(e2, e6, e5);
      if(!t3)t3 = createHOFaceFromEdges(mesh, e2, e6, e5, pGE);
      t4 = mesh->find_triangle(e3, e4, e6);
      if(!t4)t4 = createHOFaceFromEdges(mesh, e3, e4, e6, pGE);
      
      // [3] Create the Region
      return createHOTetraFromFaces(mesh, t1, t2, t3, t4, pGE);
      
    }

    // Create a complete High Order Tetra based on its bounding faces 
    // - Interior Nodes are created from a serendipity Tetra 
    pRegion createHOTetraFromFaces(pMesh mesh, pFace f1, pFace f2, pFace f3, pFace f4, pGEntity pGE)
    {
      // Get order from Faces
      int o = max( max( EN_order(f1), max(EN_order(f2),EN_order(f3)) ) , EN_order(f4) );
      
      // [1] Create a Temporary Tetrahedron (complete if order < 4, serendipity for order > 3 )
      // \note : This constructor sorts the faces if a negative volume is detected ... Helpful for Steps [2] and [4]
      MDB_Tet* Tetra = mesh->add_tet (  (MDB_Triangle *) f1, (MDB_Triangle *) f2, 
                                        (MDB_Triangle *) f3, (MDB_Triangle *) f4, pGE);
      
      if (R_volume(Tetra) < 0.)
        MAdMsgSgl::instance().error(__LINE__,__FILE__,"Negative Volume for Tetra found in createHOTetraFromFaces ! (V = %f).",R_volume(Tetra));
      
      // Need interior Nodes ?
      if ( o > 3 ) 
      {
        // Consider here that "Tetra" is a serendipity Tetrahedron
        double coord[3],uvw[3];
        pMapping mapping = Map_create(Tetra, 1.0);
        
        // [2] Create HO internal nodes 
        int nb = (o-3)*(o-2)*(o-1)/6;
        pPoint* internalVertices = new pPoint[nb];
        int l = 0;
        for(int k = 1 ; k < o - 1 ; k++ )
          for(int j = 1 ; j < o - k - 1 ; j++ )
            for(int i = 1 ; i + j + k < o ; i++ )
            {
              uvw[0] = 0. + i * 1./o ;
              uvw[1] = 0. + j * 1./o ;
              uvw[2] = 0. + k * 1./o ;
              
              Map_eval(mapping,uvw,coord);
              internalVertices[l++]=M_createV2(mesh,coord,-1,pGE);
            }
        
        if (l != nb) MAdMsgSgl::instance().error(__LINE__,__FILE__,"Wrong number of HO nodes in createHOTetraFromFaces !.");
          
        // [3] Delete Temporary serendipity Tetrahedron's mapping
        Map_delete(mapping);
        
        // [4] Create High Order Complete Tetrahedron
        MDB_Tet* HOTetra = mesh->add_tet ( (MDB_Triangle *) Tetra->f1, (MDB_Triangle *) Tetra->f2,
                                          (MDB_Triangle *) Tetra->f3, (MDB_Triangle *) Tetra->f4, pGE, o, false, internalVertices);
        delete[] internalVertices;
        
        if (R_volume(HOTetra) < 0.)
          MAdMsgSgl::instance().error(__LINE__,__FILE__,"Negative Volume for High order Tetra found in createHOTetraFromFaces !.");
        
        // [5] Delete Temporary serendipity Tetrahedron
        M_removeRegion(mesh,Tetra);
        
        // [6] Return High Order Complete Tetrahedron
        return (pRegion) HOTetra;
      }
      // [6] Return Complete Tetrahedron of order < 4
      return (pRegion) Tetra;
    }

    //! Create a High Order Linear Edge from Vertex v1 to v2 of order "order"
    pEdge createLinearHOEdge(pMesh mesh, pVertex v1, pVertex v2, int order, pGEntity eGEntity)
    {
      double coord[3];
      /* -- Suppose here that we will create a linear Edge of order o between v1 and v2 */
      pVertex* HOvertices = new pVertex[order-1];
      for(int k=1; k<order; k++) 
      {
        for (int d = 0; d<3; d++)
        {
          coord[d] = V_coordComponent(v1,d) + k * 1./order * (V_coordComponent(v2,d) - V_coordComponent(v1,d));
        }
        HOvertices[k-1]=M_createV2(mesh,coord,-1,eGEntity);
      }
      pEdge e = (pEdge) mesh->add_edge( v1, v2, eGEntity, order, HOvertices);
      delete[] HOvertices;
      return e;
    }

    //! Create a complete High Order Triangle based on its bounding Edges
    //! - Interior Nodes are obtained from a serendipity Triangle mapping
    pFace createHOFaceFromEdges(pMesh mesh, pEdge e1, pEdge e2, pEdge e3, pGEntity pGE)
    {
      // Get order from edges
      int o = max( EN_order(e1), max(EN_order(e2),EN_order(e3)) );
      
      if ( o > 2 ) 
      {
        double uvw[3],coord[3];
        int nb= (o-2)*(o-1)/2;
        pPoint* internalVertices = new pPoint[nb];
        double *u1=new double[nb];
        double *v1=new double[nb];
        
        // [1] First, create a temporary serendipity Triangle to get a mapping
        pFace tmpFace = mesh->add_triangle( e1, e2, e3, pGE);
        pMapping mapping = Map_create(tmpFace, 1.0);
        // [2] Create interior Nodes for a Reference Simplex Triangle
        double uvA[] = {0.,0.} , uvB[] = {1.,0.} , uvC[] = {0.,1.};
        controlPointsTriangle(uvA,uvB,uvC,o-3,u1,v1);
        // Foreach interior node
        uvw[2]=0.0;
        for(int k=0; k<nb; k++)
        {
          uvw[0] = 1./o*u1[k]+(uvB[0]-2.*uvA[0]+uvC[0])/o;
          uvw[1] = 1./o*v1[k]+(uvB[1]-2.*uvA[1]+uvC[1])/o;
          Map_eval(mapping,uvw,coord);
          internalVertices[k]=M_createV2(mesh,coord,-1,pGE);
        }
        // [3] Delete Temporary Element and mapping
        Map_delete(mapping);
        M_removeFace(mesh,tmpFace);
        delete [] u1;
        delete [] v1;
        
        // [4] Create High Order Triangle
        pFace f = (pFace) mesh->add_triangle( e1, e2, e3, pGE, o, false, internalVertices);
        delete[] internalVertices;
        return f;
      }
      else
      {
        return (pFace) mesh->add_triangle( e1, e2, e3, pGE);
      }
      return NULL;
    }
    
    //! Create a complete High Order Triangle face in a specific Tetrahedron (specific to the Splitting Operator) 
    pFace createHOFaceFromRegion(pMesh mesh, pRegion region, pEdge splittedEdge, pVertex &vMiddle, double t, pGEntity pGE)
    {
      double uvw[3], coord[3];
      int o = EN_order(splittedEdge);
      
      // Get the Splitted Edge index
      int SplitCase = -1;
      for (int i = 0; i < R_numEdges(region) ; i++) 
        if (R_edge(region,i) == splittedEdge ) { SplitCase = i; break; }
      
      if (SplitCase == -1 || SplitCase > 5)
      {
        MAdMsgSgl::instance().error(__LINE__,__FILE__,"Edge split for curved mesh failed. createHOFaceFromRegion : Failed to acquire the splitted edge index.");
        return NULL;
      }
      
      // Define Connectivity Tables and Nodes parametric coordinates depending on the splitted edge index
      // Parametric coordinates of the 4 Tetrahedron's Vertices
      double TetraNodesUVW[4][3] = { {0.,0.,0.}, {1.,0.,0.}, {0.,1.,0.}, {0.,0.,1.} };
      // Parametric coordinates of the Splitting Vertex on the coresponding edge (will be modified hereafter to consider the splitting parameter "t")
      double SplitCaseToSplitUVW[6][3] = { {0.5,0.,0.}, {0.5,0.5,0.}, {0.,0.5,0.}, {0.,0.,0.5}, {0.5,0.,0.5}, {0.,0.5,0.5} };
      // Connectivity of the newly created face (the first node is implicitly the splitting Vertex)
      int SplitCaseToOtherNodesUVW[6][2] = { {2,3} , {3,0} , {3,1} , {1,2} , {2,0} , {0,1}  };
      
      pEdge edges[3] = {NULL, NULL, NULL};
      // Get Triangle edges and Sort them ...
      switch(SplitCase)
      {
        case 0:
          edges[0] = mesh->find_edge(vMiddle,R_vertex(region,2));
          edges[1] = mesh->find_edge(R_vertex(region,2),R_vertex(region,3));
          edges[2] = mesh->find_edge(R_vertex(region,3),vMiddle);
          
          if ( E_vertex(splittedEdge,0) == R_vertex(region,0) ) 
          {
            SplitCaseToSplitUVW[0][0] =  t   ;
            SplitCaseToSplitUVW[0][1] = 0.0  ;
            SplitCaseToSplitUVW[0][2] = 0.0  ;
          }
          else {
            SplitCaseToSplitUVW[0][0] =(1.-t);
            SplitCaseToSplitUVW[0][1] = 0.0  ;
            SplitCaseToSplitUVW[0][2] = 0.0  ;
          }
          break;
        case 1:
          edges[0] = mesh->find_edge(vMiddle,R_vertex(region,3));
          edges[1] = mesh->find_edge(R_vertex(region,3),R_vertex(region,0));
          edges[2] = mesh->find_edge(R_vertex(region,0),vMiddle);
          
          if ( E_vertex(splittedEdge,0) == R_vertex(region,1) ) 
          {
            SplitCaseToSplitUVW[1][0] =(1.-t);
            SplitCaseToSplitUVW[1][1] =  t   ;
            SplitCaseToSplitUVW[1][2] = 0.0  ;
          }
          else {
            SplitCaseToSplitUVW[1][0] =  t   ;
            SplitCaseToSplitUVW[1][1] =(1.-t);
            SplitCaseToSplitUVW[1][2] = 0.0  ;
          }
          break;
        case 2:
          edges[0] = mesh->find_edge(vMiddle,R_vertex(region,3));
          edges[1] = mesh->find_edge(R_vertex(region,3),R_vertex(region,1));
          edges[2] = mesh->find_edge(R_vertex(region,1),vMiddle);
          
          if ( E_vertex(splittedEdge,0) == R_vertex(region,0) ) 
          {
            SplitCaseToSplitUVW[2][0] = 0.0  ;
            SplitCaseToSplitUVW[2][1] =  t   ;
            SplitCaseToSplitUVW[2][2] = 0.0  ;
          }
          else {
            SplitCaseToSplitUVW[2][0] = 0.0  ;
            SplitCaseToSplitUVW[2][1] =(1.-t);
            SplitCaseToSplitUVW[2][2] = 0.0  ;
          }
          break;
        case 3:
          edges[0] = mesh->find_edge(vMiddle,R_vertex(region,1));
          edges[1] = mesh->find_edge(R_vertex(region,1),R_vertex(region,2));
          edges[2] = mesh->find_edge(R_vertex(region,2),vMiddle);
          
          if ( E_vertex(splittedEdge,0) == R_vertex(region,0) ) 
          {
            SplitCaseToSplitUVW[3][0] = 0.0  ;
            SplitCaseToSplitUVW[3][1] = 0.0  ;
            SplitCaseToSplitUVW[3][2] =  t   ;
          }
          else {
            SplitCaseToSplitUVW[3][0] = 0.0  ;
            SplitCaseToSplitUVW[3][1] = 0.0  ;
            SplitCaseToSplitUVW[3][2] =(1.-t);
          }
          break;
        case 4:
          edges[0] = mesh->find_edge(vMiddle,R_vertex(region,2));
          edges[1] = mesh->find_edge(R_vertex(region,2),R_vertex(region,0));
          edges[2] = mesh->find_edge(R_vertex(region,0),vMiddle);
          
          if ( E_vertex(splittedEdge,0) == R_vertex(region,3) ) 
          {
            SplitCaseToSplitUVW[4][0] =  t   ;
            SplitCaseToSplitUVW[4][1] = 0.0  ;
            SplitCaseToSplitUVW[4][2] =(1.-t);
          }
          else {
            SplitCaseToSplitUVW[4][0] =(1.-t);
            SplitCaseToSplitUVW[4][1] = 0.0  ;
            SplitCaseToSplitUVW[4][2] =  t   ;
          }
          break;
        case 5:
          edges[0] = mesh->find_edge(vMiddle,R_vertex(region,0));
          edges[1] = mesh->find_edge(R_vertex(region,0),R_vertex(region,1));
          edges[2] = mesh->find_edge(R_vertex(region,1),vMiddle);
          
          if ( E_vertex(splittedEdge,0) == R_vertex(region,2) ) 
          {
            SplitCaseToSplitUVW[5][0] = 0.0  ;
            SplitCaseToSplitUVW[5][1] =(1.-t);
            SplitCaseToSplitUVW[5][2] =  t   ;
          }
          else {
            SplitCaseToSplitUVW[5][0] = 0.0  ;
            SplitCaseToSplitUVW[5][1] =  t   ;
            SplitCaseToSplitUVW[5][2] =(1.-t);
          }
          break;
        default:
          MAdMsgSgl::instance().error(__LINE__,__FILE__,"Edge split for curved mesh failed. createHOFaceFromRegion : Unknown Tet SplitCase (ie. Splitted Edge local id = %d).",SplitCase);
          return NULL;
      }
      
      // Need to create HO Nodes for the new Triangle ?
      if ( o > 2 ) 
      {
        
        // New Face interior Nodes
        int nb= (o-2)*(o-1)/2;
        pPoint* internalVertices = new pPoint[nb];
        double *u1=new double[nb];
        double *v1=new double[nb];
        
        // Map the original Tetrahedron
        pMapping mapping = Map_create(region, 1.0);
        
        // [1] Create interior Nodes for a Reference Simplex Triangle
        double uvA[] = {0.,0.} , uvB[] = {1.,0.} , uvC[] = {0.,1.};
        controlPointsTriangle(uvA,uvB,uvC,o-3,u1,v1);
        // Foreach interior node
        double u,v;
        for(int k=0; k<nb; k++)
        {
          u = 1./o*u1[k]+(uvB[0]-2.*uvA[0]+uvC[0])/o;
          v = 1./o*v1[k]+(uvB[1]-2.*uvA[1]+uvC[1])/o;
          // [2] Map from the reference Simplex triangle to the Parametric space Tetrahedron depending on the Splitting Case
          for (int d = 0 ; d < 3 ; d++)
            uvw[d] = ( SplitCaseToSplitUVW[SplitCase][d]*(1.-u -v)  + u*TetraNodesUVW[ SplitCaseToOtherNodesUVW[SplitCase][0] ][d] 
                                                                    + v*TetraNodesUVW[ SplitCaseToOtherNodesUVW[SplitCase][1] ][d] );
          Map_eval(mapping,uvw,coord);
          internalVertices[k]=M_createV2(mesh,coord,-1,pGE);
        }
        // [3] Delete Temporary mapping
        Map_delete(mapping);
        delete [] u1;
        delete [] v1;
        
        // [4] Create High Order Triangle
        pFace f = (pFace) mesh->add_triangle( edges[0],edges[1],edges[2], pGE, o, false, internalVertices);
        delete [] internalVertices;
        return f;
      }
      
      // [5] Consider The complete 1st or 2nd Order Triangle based on edges 
      return (pFace) mesh->add_triangle( edges[0],edges[1],edges[2], pGE);
    }

    // -------------------------------------------------------------------
    /*  Copy and minor modifications of the similar routine developped by F.BENYOUCEF 
        in curvoperator/CurvedEdgeSplit.h */
    void controlPointsTriangle(double uv1[2], double uv2[2], double uv3[2], int o, double *uf, double *vf){
      
      if (o <=0) {uf[0]=uv1[0];vf[0]=uv1[1]; return;}     

      uf[0]= uv1[0]; vf[0]= uv1[1];
      uf[1]= uv2[0]; vf[1]= uv2[1];
      uf[2]= uv3[0]; vf[2]= uv3[1];
      // edges
      for (int k=0; k< (o-1); k++) {
      uf[3+k] = uv1[0]+k*(uv2[0]-uv1[0])/o; vf[3+k]=uv1[1]+k*(uv2[1]-uv1[1])/o;
      uf[3+o-1+k] = uv2[0]+k*(uv3[0]-uv2[0])/o ; vf[3+o-1+k] = uv2[1]+k*(uv3[1]-uv2[1])/o;
      uf[3+2*(o-1)+k] = uv3[0]+ k*(uv1[0]-uv3[0])/o; vf[3+2*(o-1)+k] = uv3[1]+ k*(uv1[1]-uv3[1])/o;
      }
      // if order greater than 2: control points in the face
      // these points are defined by an homothetic triangle of order o-3
      // translated from original element by a vector 1/o*vec(u)+1/o*vec(v)
      // where vec(u)=(1;0;0) and vec(v)=(0;1;0)
      if (o >2){
      double *uin;
      double *vin;
      int nbin = ((o-3)+1)*((o-3)+2)/2;
      uin = new double[nbin];
      vin = new double[nbin];
      double uv1temp[2];
      double uv2temp[2];
      double uv3temp[2];
      for(int k=0; k<2; k++) {
        uv1temp[k]=uv1[k]/o; 
        uv2temp[k]=uv2[k]/o; 
        uv3temp[k]=uv3[k]/o;
      }
      controlPointsTriangle(uv1temp,uv2temp,uv3temp,o-3,uin,vin);
      for (int k=0; k< nbin; k++){
        uf[3+k+3*(o-1)] = uin[k]+uv3temp[0]-2*uv1temp[0]+uv2temp[0];
        vf[3+k+3*(o-1)] = vin[k]+uv3temp[0]-2*uv1temp[0]+uv2temp[0];
      }
      delete []uin;
      delete []vin;
      }
      return;     
    }
    
    // -------------------------------------------------------------------
    //! Routine involved in CurvedVertexMoveOp operator
    //! move Vertex vertex to position target and call the callback fonction
    bool V_setPosition(pVertex vertex, double target[3])
    {
      if( EN_constrained((pEntity)vertex) ) return false;
      
      //! \todo 1: to Rebuild to consider correctly the callback ...
      //! \todo 2: Allow HO nodes Projection on Surface if considered ...
      
      // On boundary ?
//       bool onBnd = isVertexOnBoundary(vertex);
      
      CallBackManagerSgl::instance().callCallBackMoves(vertex,target);
      
      // Move Vertex
      pPoint point = V_point(vertex);
      P_setPos(point,target[0],target[1],target[2]);
      
      // Get Cavity
      pPList vR = V_regions(vertex);
      
      // Regularize Volumic HO Nodes
      CavityRegularizer regul;
      regul.setCavity(vR);
//       regul.apply();
      
      PList_delete(vR);
      
      return true;
    }

    // -------------------------------------------------------------------
    
    
    //! Find faces adjacent to vertex "v" lying on the boundary in 3D and append in "mylist"
    //! - bandwidth > 0 will extend to adjacent Vertices 
    void cavity2D( pVertex v, pPList mylist, pGEntity entity, int bandwidth)
    {
      pPList vFList = V_faces(v);
      void *temp = NULL;
      // Foreach face adjacent to vertex "v"
      while ( pFace face = (pFace) PList_next(vFList,&temp) )
      {
        
        // 2D ??
        if (V_numRegions(v) == 0)
        {
          
          if (entity != NULL && F_whatIn(face) != entity )
            continue;
          
          // Consider all adjacent faces without distinction
          PList_appUnique(mylist,face);
          continue;
        }
        
        // 3D case
        if ( F_whatInType(face) == 2 && F_numRegions(face) == 1 )
        {
          
//           if (entity == NULL)
//             entity = F_whatIn(face);
          
          
          if (entity != NULL && F_whatIn(face) != entity )
            continue;
          
          // Add all faces on boundaries
          PList_appUnique(mylist,face);
          
          // Need to extend to adjacent Nodes ?
          if (bandwidth>0)
          {
            pPList FvList = F_vertices(face);
            void *tmp = NULL;
            while ( pVertex vertix = (pVertex) PList_next(FvList,&tmp) )
            {
              if (vertix != v)
                cavity2D(vertix, mylist, entity, bandwidth-1);
            }
            PList_delete(FvList);
          }
        }
      }
      PList_delete(vFList);
    }
    
    
    
    // -------------------------------------------------------------------
    //! fonction involved in the CurvedEdgeCollapseOp operator
    //! collapse the edge edgeDel from vertex vDel to vTgt and call the callback fonction
    void E_collapse(pMesh mesh, 
                    pEdge edgeDel, 
                    pVertex vDel, 
                    pVertex vTgt)
    {
      pPList eRegions = E_regions(edgeDel);
      
      // --- 2D collapse case ---
      if( PList_size(eRegions) == 0 ) {
        PList_delete(eRegions);
        E_collapseOnGFace(mesh,edgeDel,vDel,vTgt);
        return;
      }
      
      void * temp = NULL;
      int o = EN_order(edgeDel);
      
      // --- build the new mesh ---      
      // --------------------------------------------------------------------------------------------------
      Regularizer regul;
      
      std::map<pGEntity,pPList> oldSurfaceCavitySet;
      pPList newEdgeCavity = PList_new();
      pPList newSurfaceCavity = PList_new();
      pPList newVolumeCavity = PList_new();
      
      // --------------------------------------------------------------------------------------------------
      // [1] Are we going to collapse an edge on a Boundary ?
      if ( E_whatInType(edgeDel) < 3 )
      {
        int nE = 0, nF = 0;
        
        // ------------------------------------------------------------------
        // [1.1] Defining the initial cavity foreach Geometrical entity
        temp = NULL;
        pPList cav2D = V_faces(vDel);
        while ( pFace face = (pFace)PList_next(cav2D,&temp) )
        {
          if (F_whatInType(face) != 2) continue;
          
          pGEntity fGen = F_whatIn(face);
          
          if ( oldSurfaceCavitySet.find(fGen) == oldSurfaceCavitySet.end() )
            oldSurfaceCavitySet[fGen] = PList_new();
        }
        PList_delete(cav2D);
        
        std::map<pGEntity,pPList>::iterator it;
        for (it=oldSurfaceCavitySet.begin(); it!=oldSurfaceCavitySet.end(); it++)
        {
          cavity2D(vDel,it->second,it->first);
        }
        
        // ------------------------------------------------------------------
        // [1.2] Create Boundary EDGES and FACES
        temp = NULL;
        cav2D = V_faces(vDel);
        while ( pFace face = (pFace)PList_next(cav2D,&temp) )
        {
          pGEntity fGen = F_whatIn(face);
          
          // On one of both faces to be deleted ?
          if ( F_inClosure(face,vTgt) ) continue;
          
          // is this face not on boundary ?
          if (F_whatInType(face) != 2) continue;
          
          pEdge oppE = F_vtOpEd(face,vDel);
          
          pFace oFace = F_exist(oppE,vTgt);
          if ( oFace && (F_whatIn(oFace) == fGen) ) continue;
          
          pVertex vA = E_vertex(oppE,0);
          pVertex vB = E_vertex(oppE,1);
          
          pEdge eA = E_exist(vTgt,vA);
          pEdge eB = E_exist(vTgt,vB);
          
          // [1.2.1] Create Boundary Edges
          if (!eA || (eA && E_whatInType(eA) == 3) )
          {
            pEdge oEdge = E_exist(vA,vDel);
            // [1.2.1.1] Create Edge
            eA = createHOEdge(mesh, vA, vTgt, o, E_whatIn( oEdge ) );
            
            if ( E_whatInType( oEdge ) == 1 )
            {
              // [1.2.1.2] Update HO Nodes ...
              pPList cavity = PList_new();
              
              PList_append(cavity, oEdge );
              PList_append(cavity, edgeDel );
              
              regul.setEntity(eA);
              regul.setConstraint(cavity);
              regul.apply();
              PList_delete(cavity);
            }
            if ( E_whatInType( oEdge ) == 2 )
            {
              regul.setEntity(eA);
              regul.setConstraint(oldSurfaceCavitySet[E_whatIn( oEdge )]);
              regul.apply();
            }
            
            PList_append(newEdgeCavity,eA);
            nE++;
          }
          
          if ( !eB || (eB && E_whatInType(eB) == 3) )
          {
            pEdge oEdge = E_exist(vB,vDel);
            // [1.2.1.1 bis] Create Edge
            eB = createHOEdge(mesh, vB, vTgt, o,  E_whatIn( oEdge ) );

            if ( E_whatInType( oEdge ) == 1 )
            {
              // [1.2.1.2 bis] Update HO Nodes ...
              pPList cavity = PList_new();
              
              PList_append(cavity, oEdge );
              PList_append(cavity, edgeDel );
              
              regul.setEntity(eB);
              regul.setConstraint(cavity);
              regul.apply();
              PList_delete(cavity);
            }
            if ( E_whatInType( oEdge ) == 2 )
            {
              regul.setEntity(eB);
              regul.setConstraint(oldSurfaceCavitySet[E_whatIn( oEdge )]);
              regul.apply();
            }
            
            PList_append(newEdgeCavity,eB);
            nE++;
          }
          
          // [1.2.2] Create Boundary Faces
          pFace newface = createHOFaceFromEdges(mesh, eA, eB, oppE, fGen);
          
          regul.setEntity(newface);
          regul.setConstraint(oldSurfaceCavitySet[fGen]);
          regul.apply();
          
          PList_append(newSurfaceCavity,newface);
          nF++;
          
        }
        PList_delete(cav2D);
        
        // ------------------------------------------------------------------
        // [1.3] Free memory
        for (it=oldSurfaceCavitySet.begin(); it!=oldSurfaceCavitySet.end(); it++)
        {
          PList_delete(oldSurfaceCavitySet[it->first]);
        }
        
      }
      
      // --------------------------------------------------------------------------------------------------
      // [2] Create HO Regions and internal Edges/Faces
      pVertex verts[4];
      temp = NULL;
      pPList newRegions = PList_new();
      pPList vDelRegions = V_regions(vDel);
      while ( pRegion region = (pRegion)PList_next(vDelRegions,&temp) )
        {
          if( PList_inList(eRegions,(pEntity)region) ) continue;
          
          pFace fExt = R_vtOpFc(region,vDel);
          
          // create the region
          pGEntity rGEntity = (pGEntity)R_whatIn(region);
          verts[0] = vTgt;
          for (int iV = 0; iV<3; iV++) verts[iV+1] = F_vertex(fExt,iV);
          
          // --- create HO Region and internal Edges/Faces
          pRegion newR = createHOTetraFromVertices(mesh, verts, o, rGEntity);
          
          regul.setEntity(newR);
          regul.setConstraint(NULL);
          regul.apply();
          
          PList_append(newRegions, (pEntity)newR);
          PList_append(newVolumeCavity, (pEntity)newR);
        }
      PList_delete(eRegions);
      
      
      
      // --------------------------------------------------------------------------------------------------
      // [3] Mesh Cleaning ...
      temp = NULL;
      while ( pFace face = (pFace)PList_next(newSurfaceCavity,&temp) )
      {
        if ( F_numRegions(face) != 1 )
          M_removeFace(mesh,face);
      }
      
      
      PList_delete(newEdgeCavity);
      PList_delete(newSurfaceCavity);
      PList_delete(newVolumeCavity);

      // [4] --- re-classify new boundary entities ---
      int gDim = E_whatInType(edgeDel);

      // Classify edges on lines
      if ( gDim == 1 )
        {
          pPList vDelEdges = V_edges(vDel);
          temp = NULL;
          pEdge pe;
          while ( ( pe = (pEdge)PList_next(vDelEdges,&temp) ) ) {
            if ( pe == edgeDel ) continue;
            if ( E_whatInType(pe) == 1 ) {
              pGEntity gEdge2 = E_whatIn(pe);
              pVertex oppV = E_otherVertex(pe,vDel);
              pEdge newLineEdge = E_exist(vTgt,oppV);
              E_setWhatIn(newLineEdge,gEdge2);
            }
          }
          PList_delete(vDelEdges);
        }
      // classify faces on surfaces
      if ( gDim <= 2 )
        {
          pGEntity gFace;
          pEdge oppE;
          pFace newSurfaceFace;

          pPList vDelFaces = V_faces(vDel);
          temp = NULL;
          pFace pf;
          while ( ( pf = (pFace)PList_next(vDelFaces,&temp) ) ) {
            if ( F_inClosure(pf,(pEntity)edgeDel) ) continue;
            if ( F_whatInType(pf) == 2 ) {
              gFace          = F_whatIn(pf);
              oppE           = F_vtOpEd(pf,vDel);
              newSurfaceFace = F_exist(oppE,vTgt);
              F_setWhatIn(newSurfaceFace,gFace);
            }
          }
          PList_delete(vDelFaces);
        }


      // --- call callback functions ---
      CallBackManagerSgl::instance().callCallBacks(vDelRegions,
                                                  newRegions,
                                                  MAd_ECOLLAPSE,
                                                  (pEntity)vDel);
      PList_delete(newRegions);

      // --- delete old cavity ---
      temp = NULL;
      while ( pRegion region = (pRegion)PList_next(vDelRegions,&temp) ) {
        M_removeRegion(mesh,region);
      }
      PList_delete(vDelRegions);

      pPList vFaces = V_faces(vDel);
      temp = NULL;
      while ( pFace face = (pFace)PList_next(vFaces,&temp) ) {
        M_removeFace(mesh,face);
      }
      PList_delete(vFaces);

      pPList vEdges = V_edges(vDel);
      temp = NULL;
      while ( pEdge edge = (pEdge)PList_next(vEdges,&temp) ) {
        M_removeEdge(mesh,edge);
      }
      PList_delete(vEdges);

      M_removeVertex(mesh,vDel);

    }
      

    // -------------------------------------------------------------------
    //! function involved in the CurvedEdgeCollapseOp operator
    //! collapse the edge edgeDel from vertex vDel to vTgt and call the callback fonction
    //! This function is specific for 2D cases and called by E_collapse()
    void E_collapseOnGFace(pMesh mesh, 
                          pEdge edgeDel, 
                          pVertex vDel, 
                          pVertex vTgt)
    {
      // --- list edges to be merged ---

      std::map<pEdge,pEdge> eOldToNew; // deleted edge -> corresponding new (or target) edge
      std::map<pEdge,bool>  eDirs;     // if direction of (deleted) edge has to be changed
      
      int o = EN_order(edgeDel);
      
      pPList eFaces = E_faces(edgeDel);
      void * temp1 = 0;
      while ( pFace face = (pFace)PList_next(eFaces,&temp1) ) {
        
        pEdge eDel = NULL, eTgt = NULL;

        pPList fEdges = F_edges(face);
        void * temp2 = 0;
        while ( pEdge edge = (pEdge)PList_next(fEdges, &temp2) ) {
          if ( edge != edgeDel ) {
            if (E_inClosure(edge,(pEntity)vDel)) eDel = edge;  // deleted edge
            else                                 eTgt = edge;  // target edge
          }
        }
        PList_delete(fEdges);

        eOldToNew[eDel] = eTgt;
        
        // Choose the classification of target edge
        if( E_whatInType(eDel) < E_whatInType(eTgt) ) {
          E_setWhatIn( eTgt, E_whatIn(eDel) );
        }

        // Choose the direction of target edge
        if( E_vertex(eDel,0) == E_vertex(eTgt,0) ||
            E_vertex(eDel,1) == E_vertex(eTgt,1) ) {
          eDirs[eDel] = true;
        }
        else {
          eDirs[eDel] = false;
        }
      }

      // --- build the new mesh ---

      pPList newFaces = PList_new();
      pPList vDelFaces = V_faces(vDel);
      void * temp3 = 0;
      
      while ( pFace face = (pFace)PList_next(vDelFaces,&temp3) ) {
        
        if( PList_inList(eFaces,(pEntity)face) ) continue;

        // need to create a new face
        pEdge fEdges[3];

        pEdge eExt = F_vtOpEd(face,vDel);

        for(int iE=0; iE<3; iE++ ) {

          pEdge edge = F_edge(face,iE);
          
          if( edge == eExt ) {
            fEdges[iE] = edge;
            continue;
          }
          
          if ( eOldToNew.find(edge) != eOldToNew.end() ) {
            fEdges[iE] = eOldToNew.find(edge)->second;
          }
          else {
            // need to create a new edge
            
            pVertex eVertices[2];
            for(int iV=0; iV<2; iV++) {
              pVertex pV = E_vertex(edge,iV);
              if( pV == vDel ) eVertices[iV] = vTgt;
              else             eVertices[iV] = pV;
            }
            
            // create the edge
            pGEntity eGEntity = E_whatIn(edge);
            /* -- Suppose here that we will create a linear Edge of order o 
                  between eVertices[0] and eVertices[1] */
            fEdges[iE] = createLinearHOEdge( mesh, eVertices[0], eVertices[1], o, eGEntity );
            eOldToNew[edge] = fEdges[iE];
            
            // On boundary ?? Project
            if ( E_whatInType(edge) <= 1 )
            {
              pPList cavity = PList_new();
              
              PList_append(cavity, E_exist(eVertices[0],vDel) );
              PList_append(cavity, E_exist(eVertices[1],vDel) );
              
              Regularizer regul;
              regul.setEntity(fEdges[iE]);
              regul.setConstraint(cavity);
              regul.apply();
              PList_delete(cavity);
            }
            
          }
        }
        
        // create the face
        pGEntity fGEntity = F_whatIn(face);
        pFace newFace = NULL;
        
        newFace = createHOFaceFromEdges(mesh, fEdges[0], fEdges[1], fEdges[2], fGEntity);
        Regularizer Faceregul;
        Faceregul.setEntity(newFace);
        Faceregul.apply();
        
        PList_append(newFaces,(pEntity)newFace);
      }

      PList_delete(eFaces);

      // --- call callback functions ---
      CallBackManagerSgl::instance().callCallBacks(vDelFaces,
                                                  newFaces,
                                                  MAd_ECOLLAPSE,
                                                  (pEntity)vDel);
      PList_delete(newFaces);

      // --- delete old cavity ---
      void * temp = NULL;
      while ( pFace face = (pFace)PList_next(vDelFaces,&temp) ) {
        M_removeFace(mesh,face);
      }
      PList_delete(vDelFaces);

      pPList vEdges = V_edges(vDel);
      temp = NULL;
      while ( pEdge edge = (pEdge)PList_next(vEdges,&temp) ) {
        M_removeEdge(mesh,edge);
      }
      PList_delete(vEdges);

      M_removeVertex(mesh,vDel);
    }
    
    // -------------------------------------------------------------------
    
    // -------------------------------------------------------------------
    
    /*  Copy and minor modifications of the similar routine developped by F.BENYOUCEF 
        in curvoperator/CurvedEdgeSplit.cc */
    //! function related to the E_split() function
    //! creates the HO median edge connecting vertices vMiddle and the opposite vertex of 
    //! splittedEdge wrt triangle face
    void createMedianEdge(  pMesh mesh, pFace &face, pEdge &splittedEdge, pVertex &vMiddle, 
                              double t, double lscale, int i, int patch, pEdge &medianEdge)
    {
      // define a mapping from reference triangle
      pMapping mapping = Map_create(face, lscale);
      int o = EN_order(splittedEdge);
      double uvw[3],uvwB[3],uvwE[3];
      double coord[3];
      pGEntity pGEFace=F_whatIn(face);
      
      // switch on splitted edge id to get the bounding vertices of the median edge
      switch ( i ) {
        case 0:
        {
          if( F_edgeDir(face,0) ) {
            uvwB[0]= t ; uvwB[1]=0.0; uvwB[2]=0.0;
            uvwE[0]=0.0; uvwE[1]=1.0; uvwE[2]=0.0;
          }
          else {
            uvwB[0]= (1.-t); uvwB[1]=0.0; uvwB[2]=0.0;
            uvwE[0]=0.0    ; uvwE[1]=1.0; uvwE[2]=0.0;
          }        
          break;
        }
      case 1:
        {
          if( F_edgeDir(face,1) ) {
            uvwB[0]=(1.-t); uvwB[1]= t ; uvwB[2]=0.0;
            uvwE[0]= 0.0  ; uvwE[1]=0.0; uvwE[2]=0.0;
          }
          else {
            uvwB[0]= t ; uvwB[1]=(1.-t); uvwB[2]=0.0;
            uvwE[0]=0.0; uvwE[1]= 0.0  ; uvwE[2]=0.0; 
          }
          break;
        }
      case 2:
        {
          if( F_edgeDir(face,2) ) {
            uvwB[0]=0.0; uvwB[1]=(1.-t); uvwB[2]=0.0;
            uvwE[0]=1.0; uvwE[1]=  0.0 ; uvwE[2]=0.0;
          }
          else {
            uvwB[0]=0.0; uvwB[1]= t ; uvwB[2]=0.0;
            uvwE[0]=1.0; uvwE[1]=0.0; uvwE[2]=0.0;
          }
          break;
        }
      default:
        {
          std::cerr << "splitting methods are only implemented for triangles" << std::endl;
          std::cerr << "edge id must be 0, 1 or 2" << std::endl;
          Map_delete(mapping);
          throw;
        }
      }
      pVertex* HOvertices = new pVertex[o-1];
      if ( o > 1 ) {
        for(int k=1; k<o; k++) {
          for(int index=0; index<3; index++) uvw[index]=uvwB[index]+((k*1./o)*(uvwE[index]-uvwB[index]));
          Map_eval(mapping,uvw,coord);
          HOvertices[k-1]=M_createV2(mesh,coord,patch,pGEFace);
        }
      }
      // build the median edge in physical space (arbitrary Order o)
      medianEdge = M_createE(mesh, vMiddle, F_edOpVt(face,splittedEdge), o, HOvertices, pGEFace);
      
      delete[] HOvertices;
      Map_delete(mapping);
    }
    
    //! function related to the E_split() function
    //! creates a linear HO edge connecting vertices vMiddle to the opposite vertex of splittedEdge
    //! wrt triangle face
    void createLinearMedianEdge(  pMesh mesh, pFace face, pEdge &splittedEdge, pVertex &vMiddle, 
                                    double lscale, int i, int patch, pEdge &medianEdge)
    {
      int o = EN_order(splittedEdge);
      pGEntity pGEFace=F_whatIn(face);

      medianEdge = createLinearHOEdge(mesh, vMiddle, F_edOpVt(face,splittedEdge), o, pGEFace);

    }
    
    /*  Copy and minor modifications of the similar routine developped by F.BENYOUCEF 
     in curvoperator/C*urvedEdgeSplit.cc */
    //! function related to the E_split() function
    //! Split the edge splittedEdge at curvilinear abscisse t resulting into 2 smaller edges e1, e2 and the mid-vertex vMiddle
    void splitCreateEdges(pMesh mesh, pEdge &splittedEdge, double lscale, int patch, double t, pVertex &vMiddle, pEdge &e1, pEdge &e2)
    {
      // define a mapping from the 1d ref 
      pMapping mapping=Map_create(splittedEdge,lscale);
      
      // construct the middle vertex of the edge
      double uvw[3];                // Mapping  Space Coordinates of the new Vertex
      double coord[3], xyz[3];      // Physical Space Coordinates of the new Vertex
      double u = -1.0 + t * 2.0;
      uvw[0] = u; uvw[1] = 0.; uvw[2] = 0.; 
      Map_eval(mapping, uvw, xyz);
      pGEntity pGE=E_whatIn(splittedEdge);
      vMiddle=M_createV2(mesh, xyz, patch, pGE);
      
      // get the order of the original edge
      int o=EN_order(splittedEdge);
      
      // create left and right edge resulting from splitting
      pVertex** HOvertices = new pVertex*[2];
      HOvertices[0] = new pVertex[o-1];
      HOvertices[1] = new pVertex[o-1];
  
      // create the HO vertices of the left edge
      if( o > 1 )
        for(int i=1; i<o; i++)
        {
          uvw[0]=-1.0 + i*(u+1.0)/o ;
          uvw[1]=0.;
          uvw[2]=0.;
          Map_eval(mapping, uvw, coord);
          HOvertices[0][i-1]=M_createV2(mesh,coord,patch,pGE);
        }
      e1 = M_createE(mesh,E_vertex(splittedEdge,0),vMiddle,o,HOvertices[0],pGE);
      // create the HO vertices of the right edge
      if( o > 1 )
        for(int i=1; i<o; i++)
        {
          uvw[0]= u + i*(1.0-u)/o ;
          uvw[1]=0.;
          uvw[2]=0.;
          Map_eval(mapping, uvw, coord);
          HOvertices[1][i-1]=M_createV2(mesh,coord,patch,pGE);
        }
      e2 = M_createE(mesh, vMiddle, E_vertex(splittedEdge,1), o, HOvertices[1], pGE);
      
      delete[] HOvertices[0];
      delete[] HOvertices[1];
      delete[] HOvertices;
      Map_delete(mapping);
    }
    
    
    // -------------------------------------------------------------------
    /*  Copy and minor modifications of the similar routine developped by F.BENYOUCEF 
        in curvoperator/CurvedEdgeSplit.cc */                                                                                
     //! function related to the E_split() function
     //! Creates the 2 smaller faces newF1 and newF2 resulting from the split of origface by edge medianEdge
     //! \note This procedure is applied in the parametric space of origface
    bool createHOFacesFromFace( pMesh mesh, pFace origface, pEdge splittedEdge, double t,            // inputs
                                double lscale, int patch, pEdge &e1, pEdge &e2, pEdge &medianEdge,    // inputs
                        pFace &newF1, pFace &newF2)                                                   // outputs
    {
      // --- Create a mapping from reference triangle --- //
      pMapping mapping = Map_create(origface, lscale);
      int o = EN_order(splittedEdge);
      
      pGEntity fGE = F_whatIn(origface);
      double coord[3];
      
      pEdge fEdgesF1[3];
      pEdge fEdgesF2[3];
      pEdge attachedEdgeF1;
      pEdge attachedEdgeF2;
      
      // --- Create Internal High Order points --- //
      // - Now, work in the Parametric Space UVW
      
      double uvw [3];
      double uv11[2];
      double uv12[2];
      double uv13[2];
      double uv21[2];
      double uv22[2];
      double uv23[2];
      
      // Spliting Case 0
      if(splittedEdge==F_edge(origface,0))
      {
  // first face generated
        uv11[0]=0.0; uv11[1]=0.0;
        uv13[0]=0.0; uv13[1]=1.0;
        
        if( F_edgeDir(origface,0) ) {
          uv12[0]= t      ; uv12[1]=0.0;
        }
        else {
          uv12[0]= (1.-t) ; uv12[1]=0.0;
        }
        
  // 2nd face generated
        if( F_edgeDir(origface,0) ) {
          uv21[0]= t      ; uv21[1]=0.0;
        }
        else {
          uv21[0]= (1.-t) ; uv21[1]=0.0;
        }
        uv22[0]=1.0; uv22[1]=0.0;
        uv23[0]=0.0; uv23[1]=1.0;
        
        if( F_edgeDir(origface,0) ) {
          fEdgesF1[0] = e1;
          fEdgesF1[1] = medianEdge;
          fEdgesF1[2] = F_edge(origface,2);
          fEdgesF2[0] = e2;
          fEdgesF2[1] = F_edge(origface,1);
          fEdgesF2[2] = medianEdge;
        }else{
          fEdgesF1[0] = e2;
          fEdgesF1[1] = medianEdge;
          fEdgesF1[2] = F_edge(origface,2);
          fEdgesF2[0] = e1;
          fEdgesF2[1] = F_edge(origface,1);
          fEdgesF2[2] = medianEdge;
        }
        attachedEdgeF1 = F_edge(origface,2);
        attachedEdgeF2 = F_edge(origface,1);
      }
      // Spliting Case 1
      else if(splittedEdge==F_edge(origface,1))
      {
  // first face generated
        uv11[0]=0.0; uv11[1]=0.0;
        uv12[0]=1.0; uv12[1]=0.0;
        if( F_edgeDir(origface,1) ) {
          uv13[0]= (1.-t) ; uv13[1]=  t   ;
        }
        else {
          uv13[0]=   t    ; uv13[1]=(1.-t);
        }
        
  // 2nd face generated
        uv21[0]=0.0; uv21[1]=0.0;
        uv23[0]=0.0; uv23[1]=1.0; 
        if( F_edgeDir(origface,1) ) {
          uv22[0]= (1.-t) ; uv22[1]=  t   ;
        }
        else {
          uv22[0]=   t    ; uv22[1]=(1.-t);
        }
        
        if( F_edgeDir(origface,1) ) {
          fEdgesF1[0] = F_edge(origface,0);
          fEdgesF1[1] = e1;
          fEdgesF1[2] = medianEdge;
          fEdgesF2[0] = medianEdge;
          fEdgesF2[1] = e2;
          fEdgesF2[2] = F_edge(origface,2);
        }else{
          fEdgesF1[0] = F_edge(origface,0);
          fEdgesF1[1] = e2;
          fEdgesF1[2] = medianEdge;
          fEdgesF2[0] = medianEdge;
          fEdgesF2[1] = e1;
          fEdgesF2[2] = F_edge(origface,2);
        }
        attachedEdgeF1 = F_edge(origface,0);
        attachedEdgeF2 = F_edge(origface,2);
      }
      // Spliting Case 2
      else if(splittedEdge==F_edge(origface,2))
      {
  // first face generated
        uv11[0]=0.0; uv11[1]=0.0;
        uv12[0]=1.0; uv12[1]=0.0;
        if( F_edgeDir(origface,2) ) {
          uv13[0]= 0.0 ; uv13[1]=(1.-t);
        }
        else {
          uv13[0]= 0.0 ; uv13[1]=  t   ;
        }
          
  // 2nd face generated
        uv22[0]=1.0; uv22[1]=0.0;
        uv23[0]=0.0; uv23[1]=1.0;
        if( F_edgeDir(origface,2) ) {
          uv21[0]= 0.0 ; uv21[1]=(1.-t);
        }
        else {
          uv21[0]= 0.0 ; uv21[1]=  t   ;
        }
        
        if( F_edgeDir(origface,2) ) {
          fEdgesF1[0] = F_edge(origface,0);
          fEdgesF1[1] = medianEdge;
          fEdgesF1[2] = e2;
          fEdgesF2[0] = medianEdge;
          fEdgesF2[1] = F_edge(origface,1);
          fEdgesF2[2] = e1;
        }else{
          fEdgesF1[0] = F_edge(origface,0);
          fEdgesF1[1] = medianEdge;
          fEdgesF1[2] = e1;
          fEdgesF2[0] = medianEdge;
          fEdgesF2[1] = F_edge(origface,1);
          fEdgesF2[2] = e2;
        }
        attachedEdgeF1 = F_edge(origface,0);
        attachedEdgeF2 = F_edge(origface,1);
      }
      else
      {
        std::cout<< "edge id must be between 0, 1 or 2 !" << std::endl;
        Map_delete(mapping);
        return false;
      }
      
      int nb= (o-2)*(o-1)/2;
      pPoint** internalVertices = new pPoint*[2];
      internalVertices[0] = new pPoint[nb];
      internalVertices[1] = new pPoint[nb];
      double *u1=new double[nb];
      double *v1=new double[nb];
      double u,v;
      if ( o > 2 ) {
        // [1] Create interior Nodes for a Reference Simplex Triangle
        double uvA[] = {0.,0.} , uvB[] = {1.,0.} , uvC[] = {0.,1.};
        controlPointsTriangle(uvA,uvB,uvC,o-3,u1,v1);
        // Foreach interior node
        for(int k=0; k<nb; k++)
        {
          u = 1./o*u1[k]+(uvB[0]-2.*uvA[0]+uvC[0])/o;
          v = 1./o*v1[k]+(uvB[1]-2.*uvA[1]+uvC[1])/o;
          // [2] Map from the reference Simplex triangle to the Parametric space triangle
          // 1st new element
          uvw[0] = ( uv11[0]*(1.-u -v) + u*uv12[0] + v*uv13[0] );
          uvw[1] = ( uv11[1]*(1.-u -v) + u*uv12[1] + v*uv13[1] );
          uvw[2]=0.0;
          Map_eval(mapping,uvw,coord);
          internalVertices[0][k]=M_createV2(mesh,coord,patch,fGE);
        }
        for(int k=0; k<nb; k++)
        {
          u = 1./o*u1[k]+(uvB[0]-2.*uvA[0]+uvC[0])/o;
          v = 1./o*v1[k]+(uvB[1]-2.*uvA[1]+uvC[1])/o;
          // [2] Map from the reference Simplex triangle to the Parametric space triangle
          // 2nd new element
          uvw[0] = ( uv21[0]*(1.-u -v) + u*uv22[0] + v*uv23[0] );
          uvw[1] = ( uv21[1]*(1.-u -v) + u*uv22[1] + v*uv23[1] );
          uvw[2]=0.0;
          Map_eval(mapping,uvw,coord);
          internalVertices[1][k]=M_createV2(mesh,coord,patch,fGE);
        }
      }
      
      // --- Create the new Faces --- //
      newF1 = mesh->add_triangle( fEdgesF1[0], fEdgesF1[1], fEdgesF1[2], fGE, o, false, internalVertices[0]);
      EN_attachDataP((pEntity)attachedEdgeF1,"_EdgeSplitMarker_",(void *)newF1);
      
      newF2 = mesh->add_triangle( fEdgesF2[0], fEdgesF2[1], fEdgesF2[2], fGE, o, false, internalVertices[1]);
      EN_attachDataP((pEntity)attachedEdgeF2,"_EdgeSplitMarker_",(void *)newF2);
      
      delete[] internalVertices[0];
      delete[] internalVertices[1];
      delete[] internalVertices;
      Map_delete(mapping);
      delete [] u1;
      delete [] v1;
      return true;
    }
    
    //! function related to the E_split() function
    //! Creates the 2 smaller faces newF1 and newF2 resulting from the split of origface by edge medianEdge
    //! \note This procedure is applied in the physical space.thus, it's not conform wrt original face origface.
    bool createHOFacesFromSplittedEdges( pMesh mesh, pFace origface, pEdge splittedEdge, double lscale, int patch,   // inputs
                        pEdge &e1, pEdge &e2, pEdge &medianEdge,                                    // inputs
                        pFace &newF1, pFace &newF2)                                                 // outputs
    {
      pGEntity fGE = F_whatIn(origface);
      
      pEdge fEdgesF1[3];
      pEdge fEdgesF2[3];
      pEdge attachedEdgeF1;
      pEdge attachedEdgeF2;
      
      // Spliting Case 0
      if(splittedEdge==F_edge(origface,0))
      {      
        if( F_edgeDir(origface,0) ) {
          fEdgesF1[0] = e1;
          fEdgesF1[1] = medianEdge;
          fEdgesF1[2] = F_edge(origface,2);
          fEdgesF2[0] = e2;
          fEdgesF2[1] = F_edge(origface,1);
          fEdgesF2[2] = medianEdge;
        }else{
          fEdgesF1[0] = e2;
          fEdgesF1[1] = medianEdge;
          fEdgesF1[2] = F_edge(origface,2);
          fEdgesF2[0] = e1;
          fEdgesF2[1] = F_edge(origface,1);
          fEdgesF2[2] = medianEdge;
        }
        attachedEdgeF1 = F_edge(origface,2);
        attachedEdgeF2 = F_edge(origface,1);
      }
      // Spliting Case 1
      else if(splittedEdge==F_edge(origface,1))
      {
        if( F_edgeDir(origface,1) ) {
          fEdgesF1[0] = F_edge(origface,0);
          fEdgesF1[1] = e1;
          fEdgesF1[2] = medianEdge;
          fEdgesF2[0] = medianEdge;
          fEdgesF2[1] = e2;
          fEdgesF2[2] = F_edge(origface,2);
        }else{
          fEdgesF1[0] = F_edge(origface,0);
          fEdgesF1[1] = e2;
          fEdgesF1[2] = medianEdge;
          fEdgesF2[0] = medianEdge;
          fEdgesF2[1] = e1;
          fEdgesF2[2] = F_edge(origface,2);
        }
        attachedEdgeF1 = F_edge(origface,0);
        attachedEdgeF2 = F_edge(origface,2);
      }
      // Spliting Case 2
      else if(splittedEdge==F_edge(origface,2))
      {      
        if( F_edgeDir(origface,2) ) {
          fEdgesF1[0] = F_edge(origface,0);
          fEdgesF1[1] = medianEdge;
          fEdgesF1[2] = e2;
          fEdgesF2[0] = medianEdge;
          fEdgesF2[1] = F_edge(origface,1);
          fEdgesF2[2] = e1;
        }else{
          fEdgesF1[0] = F_edge(origface,0);
          fEdgesF1[1] = medianEdge;
          fEdgesF1[2] = e1;
          fEdgesF2[0] = medianEdge;
          fEdgesF2[1] = F_edge(origface,1);
          fEdgesF2[2] = e2;
        }
        attachedEdgeF1 = F_edge(origface,0);
        attachedEdgeF2 = F_edge(origface,1);
      }
      else
      {
        std::cout<< "edge id must be between 0, 1 or 2 !" << std::endl;
        return false;
      }
      
      
      // --- Create the new Faces --- //
      newF1 = createHOFaceFromEdges(mesh, fEdgesF1[0], fEdgesF1[1], fEdgesF1[2], fGE);
      EN_attachDataP((pEntity)attachedEdgeF1,"_EdgeSplitMarker_",(void *)newF1);
      
      newF2 = createHOFaceFromEdges(mesh, fEdgesF2[0], fEdgesF2[1], fEdgesF2[2], fGE);
      EN_attachDataP((pEntity)attachedEdgeF2,"_EdgeSplitMarker_",(void *)newF2);
      return true;
    }
    
    // -------------------------------------------------------------------
    /*  Modification of the original routine for 1st order element to handle High Order triangles */
    //! function related to CurvedEdgeSplitOp
    //! this function split edge at curvilinear location t and the adjacent faces and regions
    pVertex E_split(pMesh mesh, pEdge edge, double xyz[3], double t)
    {
      const bool LinearStrategy = false;
      const double lscale = 1.0;
      pPList newEdges = PList_new();    // List of newly created Edges
      pPList newFaces = PList_new();    // List of newly created Faces
      pVertex newV = NULL;              // The new Vertex
      pEdge e1 = NULL;                  // The new First  Edge
      pEdge e2 = NULL;                  // The new Second Edge
      Regularizer regul;
      
      // [1] --- build the Mid vertex and the edges ---
      /////////////////////////////////////////////////////////////////////////
      splitCreateEdges(mesh, edge, lscale, -1, t , newV, e1, e2);               // SPLIT At PARAMETER t
      V_coord(newV,xyz);
      /////////////////////////////////////////////////////////////////////////
      
      int EwiT = E_whatInType(edge);
      
      if ( !newV || !e1 || !e2)
        MAdMsgSgl::instance().error(__LINE__,__FILE__,"Edge split for curved mesh failed. splitCreateEdges Failed.");   
      
      PList_append(newEdges,(pEntity)e1);
      PList_append(newEdges,(pEntity)e2);
      
      // [2] --- build the faces bordered by new edges ---
      pPList eFaces = E_faces(edge);
      void * temp = NULL;
      while( pFace face = (pFace)PList_next(eFaces,&temp) ) {
        
        // Edge id in the Face 
        int iEdge = 0;
        for( iEdge=0; iEdge< F_numEdges(face); iEdge++) {
          if( F_edge(face,iEdge) == edge ) break;
        }
        
        // Apply Curved edges only on surfaces in 3D configuration or in 2D configuration
        bool allowCurved = (E_numRegions(edge) != 0 && F_whatInType(face) <= 2 && F_numRegions(face) == 1) || (E_numRegions(edge)==0 && EwiT <= 2) ;
        
        // [2.1] Create median Edge and the new Faces
        pEdge newE = NULL;
        pFace newF1= NULL;
        pFace newF2= NULL;
        /////////////////////////////////////////////////////////////////////////
        if ( !LinearStrategy && allowCurved )
        {
          createMedianEdge(mesh, face, edge, newV, t, lscale, iEdge, -1, newE);           // SPLIT At PARAMETER t
        }
        else
          createLinearMedianEdge(mesh, face, edge, newV, lscale, iEdge, -1, newE);
        
        /////////////////////////////////////////////////////////////////////////
        
        
        PList_append(newEdges,(pEntity)newE);
        
        if ( !newE )
          MAdMsgSgl::instance().error(__LINE__,__FILE__,"Edge split for curved mesh failed. createMedianEdge Failed.");   
        
        // [2.2] Create the high order internal points
        /////////////////////////////////////////////////////////////////////////
        if ( !LinearStrategy && allowCurved )
        {
          createHOFacesFromFace( mesh, face, edge, t, lscale, -1,  e1, e2, newE, newF1, newF2);
        }
        else
          createHOFacesFromSplittedEdges( mesh, face, edge, lscale, -1,  e1, e2, newE, newF1, newF2);
        
        /////////////////////////////////////////////////////////////////////////
        
        EN_attachDataP((pEntity)face,"_EdgeSplitMarker_",(void *)newE);
        EN_attachDataP((pEntity)newF1,"_EdgeSplitMarker_",face);
        EN_attachDataP((pEntity)newF2,"_EdgeSplitMarker_",face);
        
        if ( !newF1 || !newF2 )
          MAdMsgSgl::instance().error(__LINE__,__FILE__,"Edge split for curved mesh failed. createHOFace Failed."); 
        
        PList_append(newFaces,(pEntity)newF1);
        PList_append(newFaces,(pEntity)newF2);
      }
      
      // [3] --- build the regions bordered by new edges ---
      pPList delRegs = PList_new();
      pPList eRegs = E_regions(edge);
      if (eRegs) {
        
        pPList newVolumeCavity = PList_new();
        
        temp = NULL;
        while( pRegion region = (pRegion)PList_next(eRegs, &temp) ) {

          pGEntity rGE = (pGEntity)R_whatIn(region);

          pFace extFaces[2];

          // [3.1] --- build the new face between the two new regions ---
          
          pEdge fEdges[3];
          for(int iF=0, eCount=0, fCount=0; iF<R_numFaces(region); iF++) {
            pFace face = R_face(region,iF);
            if( F_inClosure(face,(pEntity)edge) ) {
              fEdges[eCount] = (pEdge)EN_dataP((pEntity)face,"_EdgeSplitMarker_");
              eCount++;
            }
            else {
              extFaces[fCount++] = face;
            }
          }
          
          pEdge oppE = R_gtOppEdg(region,edge);
          fEdges[2] = oppE;
          pFace newF = NULL;
          /////////////////////////////////////////////////////////////////////////
          if (false)
          {
            newF = createHOFaceFromRegion(mesh, region, edge, newV, t, rGE);
          }
          else
          {
            newF = createHOFaceFromEdges(mesh, fEdges[0], fEdges[1], fEdges[2], rGE);
          }
          /////////////////////////////////////////////////////////////////////////
          
          // [3.3] -- build the regions ---
          pFace rFaces[4];
          
          rFaces[0] = extFaces[0] ;
          rFaces[1] = newF;
          int fIndex = 2;
          for(int i=0; i<F_numEdges(extFaces[0]); i++) {
            pEdge tmpE = F_edge(extFaces[0],i);
            if( tmpE != oppE ) {
              rFaces[fIndex] = (pFace)EN_dataP((pEntity)tmpE,"_EdgeSplitMarker_");
              fIndex++;
            }
          }
          /////////////////////////////////////////////////////////////////////////
          pRegion r1 = createHOTetraFromFaces(mesh, rFaces[0], rFaces[1], rFaces[2], rFaces[3], rGE);
          regul.setEntity(r1);
          regul.apply();
          /////////////////////////////////////////////////////////////////////////
          rFaces[0] = extFaces[1] ;
          rFaces[1] = newF;
          fIndex = 2;
          for( int i=0; i < F_numEdges(extFaces[1]); i++ ) {
            pEdge tmpE = F_edge(extFaces[1],i);
            if( tmpE != oppE ) {
              rFaces[fIndex] = (pFace)EN_dataP((pEntity)tmpE,"_EdgeSplitMarker_");
              fIndex++;
            }
          }
          /////////////////////////////////////////////////////////////////////////
          pRegion r2 = createHOTetraFromFaces(mesh, rFaces[0], rFaces[1], rFaces[2], rFaces[3], rGE);
          regul.setEntity(r2);
          regul.apply();
          /////////////////////////////////////////////////////////////////////////
          PList_append(delRegs,(pEntity)region);
          
          PList_append(newVolumeCavity,r1);
          PList_append(newVolumeCavity,r2);
        }
        
        
        PList_delete(newVolumeCavity);
      }
      PList_delete(eRegs);
      
      pPList delEdge = PList_new();
      PList_append(delEdge,(pEntity)edge);
      CallBackManagerSgl::instance().callCallBacks(delEdge, newEdges, 
                                                  MAd_ESPLIT, (pEntity)newV);
      PList_delete(delEdge);
      PList_delete(newEdges);
      
      // --- remove regions ---
      temp = NULL;
      while( pRegion region = (pRegion)PList_next(delRegs,&temp) ) {
        M_removeRegion(mesh, region);
      }
      PList_delete(delRegs);
      
      // --- remove and clean faces ---
      temp = NULL;
      while( pFace face = (pFace)PList_next(eFaces,&temp) ) {

        EN_removeData((pEntity)face,"_EdgeSplitMarker_");
        
        for(int iE=0; iE<F_numEdges(face); iE++) {
          pEdge iEdge = F_edge(face,iE);
          if( iEdge != edge ) EN_removeData((pEntity)iEdge,"_EdgeSplitMarker_");
        }

        M_removeFace(mesh, face);
      }
      PList_delete(eFaces);
      
      temp = NULL;
      while( pFace face = (pFace)PList_next(newFaces,&temp) ) {
        EN_removeData((pEntity)face,"_EdgeSplitMarker_");
      }
      PList_delete(newFaces);
      
      // --- remove edge ---
      M_removeEdge(mesh,edge);
      
      return newV ;
    }

    // -------------------------------------------------------------------
  }
}
