// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------


#include "pointprojector.h"
#include "MAdMessage.h"
#include "Constraint.h"
#include "MeshDataBase.h"
#include "MAdMapping.h"
#include "MappingInterface.h"
#include "MAdDefines.h"

#include <cmath>

namespace MAd 
{

  // Return the Projection of vector "vec" on the face represented by its mapping at parametric location "uvw"
  double projectedOnTangent(pMapping map, double uvw[3], double vec[3])
  {
    // Compute Local derivatives
    double deriv[9];
    Map_deval(map,uvw,deriv);
    
    // Project on the Surface
    double val = 0.0;
    val  = (vec[0]*deriv[0] + vec[1]*deriv[1] + vec[2]*deriv[2])*(vec[0]*deriv[0] + vec[1]*deriv[1] + vec[2]*deriv[2]);
    val += (vec[0]*deriv[3] + vec[1]*deriv[4] + vec[2]*deriv[5])*(vec[0]*deriv[3] + vec[1]*deriv[4] + vec[2]*deriv[5]);
    
    // Return the Norm of the 2D projection
    return std::sqrt(val);
  }
  
  
  // Project the 3D Point of coordinates "coord" on the set of Triangles "stencil"
  // - modify in place (coord : original coordinates as input and projected location at exit)
  bool projectOnStencil( pPList stencil, double coord[3] )
  {
    double uvw[3],xyz[3];
    double Selecteduvw[3] = {0.0,0.0,0.0};
    double Error = 1.e+12;
    bool found = false;
    
    void *temp = NULL;
    bool isFaces = true;
    while ( pEntity ent = (pEntity) PList_next(stencil,&temp) )
    {
      if (EN_mshType(ent) < MSH_TYPE_TRI || EN_mshType(ent) > MSH_TYPE_QUA)
      {
        isFaces = false; break;
      }
    }
    
    temp = NULL;
    pMapping mapping = NULL;
    pFace selectedFace = NULL;
    pEdge selectedEdge = NULL;
    // Faces ?
    if (isFaces)
    {
      // Foreach Triangle of the Stencil
      while ( pFace face = (pFace) PList_next(stencil,&temp) )
      {
        mapping = Map_create(face,1.0);
        found = found || mapping->invert(coord[0],coord[1],coord[2],uvw[0],uvw[1],uvw[2],1.e-10,100,20);
        Map_eval(mapping,uvw,xyz);
        bool inRef = Map_inRefElt(mapping,uvw,1.e-02);
        Map_delete(mapping);
        
        double trial = DISTANCE(xyz,coord);
        
        // Criterion : Closest point and inside the element
        if ( (trial <= Error && inRef) )
        {
          Error = trial;
          selectedFace = face;
          Selecteduvw[0] = uvw[0];
          Selecteduvw[1] = uvw[1];
          Selecteduvw[2] = uvw[2];
        }
        
      }
      if (!selectedFace)
      {
        return false;
      }
      
      mapping = Map_create(selectedFace,1.0);
      Map_eval(mapping,Selecteduvw,coord);
      Map_delete(mapping);
    }
    else
    {
      // Lines 
      // Foreach Edges of the Stencil
      while ( pEdge edge = (pEdge) PList_next(stencil,&temp) )
      {
        mapping = Map_create(edge,1.0);
        found = found || mapping->invert(coord[0],coord[1],coord[2],uvw[0],uvw[1],uvw[2],1.e-10,100,10);
        Map_eval(mapping,uvw,xyz);
        bool inRef = Map_inRefElt(mapping,uvw,1.e-02);
        Map_delete(mapping);
        
        double trial = DISTANCE(xyz,coord);
        
        // Criterion : Closest point and inside the element
        if ( (trial <= Error && inRef) )
        {
          Error = trial;
          selectedEdge = edge;
          Selecteduvw[0] = uvw[0];
          Selecteduvw[1] = uvw[1];
          Selecteduvw[2] = uvw[2];
        }
        
      }
      if (!selectedEdge)
      {
//         MAdMsgSgl::instance().error(__LINE__,__FILE__,"projectOnStencil() : Impossible to project on %d Edges.", PList_size(stencil));
        return false;
      }
      
      mapping = Map_create(selectedEdge,1.0);
      Map_eval(mapping,Selecteduvw,coord);
      Map_delete(mapping);
    }
    
    return true;
  }


  bool projectOnStencilV0( pPList stencil, double coord[3] )
  {
    //! Strategy : 
    // Foreach Triangle of the set, probe the surface at a finite number of location and project the point on it
    // Find the probed location MINIMIZING the projection distance from the original coordinates AND the scalar product on the surface 
    
    bool OK = true;
    // Generate Test nodes
    int o = 8, l = 0;
    int nb = (o-2)*(o-1)/2;
    double** probes = new double*[nb];
    for(int j = 1 ; j < o -  1 ; j++ )
      for(int i = 1 ; i + j < o ; i++ )
      {
        probes[l] = new double[3];
        probes[l][0] = 0. + i * 1./o ;
        probes[l][1] = 0. + j * 1./o ;
        probes[l][2] = 0.0;
        l++;
      }
      
      double uvw[3], xyz[3], diff[3];
    double tmpDist = 0.0;
    double tmpProj = 0.0;
    
    double selecteduvw[3];
    double selectedDist = 1.e+14;
    double selectedProj = 1.e+14;
    pFace selectedFace = NULL;
    pMapping mapping;
    
    void *temp = NULL;
    
    // Foreach Triangle of the Stencil
    while ( pFace face = (pFace) PList_next(stencil,&temp) )
    {
      mapping = Map_create(face,1.0);
      
      // Foreach Test node on the current Triangle
      for (int p = 0 ; p < nb ; p++)
      {
        memcpy(uvw,probes[p],3*sizeof(double));
        Map_eval(mapping,uvw,xyz);
        tmpDist = 0.0;
        
        // Evaluate the metrics ( distance from the original node : tmpdist, vector from original node to current test node : diff)
        for (int d = 0 ; d < 3 ; d++)
        {
          diff[d] = (xyz[d] - coord[d]);
          tmpDist += diff[d]*diff[d];
        }
        tmpDist = sqrt(tmpDist);
        
        // Compute the projection of "diff" on the plane formed by the two tangent vector of the triangle at the current test node
        tmpProj = projectedOnTangent(mapping, uvw, diff);
        
        // If the current test node is the best regarding to distance and projection, select him
        if ( (tmpProj < selectedProj) && (tmpDist < selectedDist) )
        {
          selectedFace = face;
          selectedDist = tmpDist;
          selectedProj = tmpProj;
          memcpy(selecteduvw,uvw,3*sizeof(double));
        }
      }
      
      Map_delete(mapping);
    }
    
    for (int p = 0 ; p < nb ; p++) delete[] probes[p];
    delete[] probes;
    
    if (!selectedFace)
      MAdMsgSgl::instance().error(__LINE__,__FILE__,"Project on Stencil : Impossible to find the best element to project on.");
    
    
    // Compute the new projected node location supposed to be on the stencil
    mapping = Map_create(selectedFace,1.0);
    Map_eval(mapping,selecteduvw,coord);
    Map_delete(mapping);
    
    if ( !OK )
    {
      MAdMsgSgl::instance().error(__LINE__,__FILE__,"Project on Stencil : The projected point is outside of the best element.");
      return false;
    }
    
    return true;
  }

  // -------------------------------------------------------------------
  pEdge projectEdgeOnSurface(pMesh mesh, pEdge &edge, pPList &surfSet, pGEntity EGen)
  {
    int order = EN_order(edge);
    pMapping mapping = NULL;
    double uvw[3],xyz[3];
    
    if ( order <= 1 ) return edge;
    
    pVertex vA = E_vertex(edge,0);
    pVertex vB = E_vertex(edge,1);
    
    pEdge curEdge = edge;
    
    uvw[1] = 0.0;uvw[2] = 0.0;
    for (int o = 2; o <= order ; o++)
    {
      pVertex* HOPoints = new pVertex[o-1];
      
      mapping = Map_create(curEdge,1.0);
      for (int pt = 1 ; pt < o ; pt++)
      {
        // [1] 
        uvw[0] = -1.0 + pt * 2.0 / o;
        // [1.1] 
        Map_eval(mapping,uvw,xyz);
        projectOnStencil(surfSet, xyz );
        // [1.2] 
        HOPoints[pt-1] = M_createV2(mesh,xyz,-1, EGen);
      }
      Map_delete(mapping);
      
      if ( o != 2 ) M_removeEdge(mesh,curEdge);
      curEdge = (pEdge) mesh->add_edge( vA, vB, EGen, o, HOPoints);
      delete[] HOPoints;
    }
    
    return curEdge;
  }

}
