// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------


#include "regularizer.h"
#include "MAdMessage.h"
#include "Constraint.h"
#include "MappingInterface.h"
#include "MAdDefines.h"

#include <map>
#include <assert.h>

using namespace std;

namespace RegulTools {
  
  //! Try to determine if a Face is inside a cavity composed of Regions
  //! Excludes :
  //! - faces on boundaries
  //! - faces on the border of the cavity
  bool faceInCavity(pPList cavity, pFace face)
  {
    
    // is on boundary ??
    if ( F_whatInType(face) <= 2 ) return false;
    
    // Check if face is inside the cavity
    int count = 0;
    for (int r = 0 ; r < F_numRegions(face) ; r++)
      if ( PList_inList(cavity, F_region(face,r) ) && F_numRegions(face) > 1 ) count++;
    
    if(count == 2)
      return true;
    
    return false;
  }
  
  //! Return the Vertex local id in the entity "ent" wrt MAdMapping::getVertices
  int getNodeIdInEntity(pEntity ent, pVertex vertex)
  {
    int i;
    vector<pVertex> tmp;
    MAdMapping::getVertices( (pEntity) ent,tmp);
    for( i = 0 ; i < (int) tmp.size() ; i++)
    {
      if ( vertex == tmp[i] )
        break;
    }
    tmp.clear();
    return (i != (int) tmp.size() ? i : -1);
  }
  
  // Cf. regularizer.h
  template<> void getFaceHONodesStencil< 3 >(pFace face, vector< stencil_structure > &stencil)
  {
    vector<pVertex> tmp;
    MAdMapping::getVertices(face,tmp);
    
    /*
    3rd Order Triangle
    
    2
    |\
    7 6
    |  \
    8 9 5
    |    \
    0-3-4-1
    Cross Stencil for node 9 --> [3,5,6,8]
    */
    
    stencil_structure st;
    st.v = face->getHighOrderPoint(0);
    st.faces.push_back(face);
    st.neighbors.push_back( tmp[3] );
    st.neighbors.push_back( tmp[5] );
    st.neighbors.push_back( tmp[6] );
    st.neighbors.push_back( tmp[8] );
    tmp.clear();
    stencil.push_back(st);
  }
  
  // Cf. regularizer.h
  template<> void getFaceHONodesStencil< 4 >(pFace face, vector< stencil_structure > &stencil)
  {
    vector<pVertex> tmp;
    MAdMapping::getVertices(face,tmp);
    
    /*
     4th Order Triangle
     
     2
     |\
     | \
     9  8
     |   \
    10 14 7
     |     \
     |      \
    11 12 13 6
     |        \
     0--3-4--5-1
     Cross Stencil for node 12 --> [3,13,14,11]
     Cross Stencil for node 13 --> [4, 6, 7,12]
     Cross Stencil for node 14 --> [12,7, 8,10]
    */
    
    int map[3][4] = { { 3,13,14,11},
                      { 4, 6, 7,12},
                      {12, 7, 8,10} };
    
    for (int pt = 0 ; pt < 3; pt++)
    {
      stencil_structure st;
      st.faces.push_back(face);
      st.v = face->getHighOrderPoint(pt);
      for (int n = 0 ; n < 4 ; n++)
        st.neighbors.push_back( tmp[map[pt][n]] );
      stencil.push_back(st);
    }
    
    tmp.clear();
  }
  
  // Cf. regularizer.h
  template<> void getFaceHONodesStencil< 5 >(pFace face, vector< stencil_structure > &stencil)
  {
    vector<pVertex> tmp;
    MAdMapping::getVertices(face,tmp);
    
    int map[6][4] = { { 3,16,18,14},
                      { 4,17,19,15},
                      { 5, 7, 8,16},
                      {15,19,20,13},
                      {16, 8, 9,18},
                      {18, 9,10,12} };
    
    for (int pt = 0 ; pt < 6; pt++)
    {
      stencil_structure st;
      st.faces.push_back(face);
      st.v = face->getHighOrderPoint(pt);
      for (int n = 0 ; n < 4 ; n++)
        st.neighbors.push_back( tmp[map[pt][n]] );
      stencil.push_back(st);
    }
    
    tmp.clear();
  }
  
  // Cf. regularizer.h
  template<> void getRegionHONodesStencil< 4 >(pRegion region, vector< stencil_structure > &stencil)
  {
    vector<pVertex> tmp;
    MAdMapping::getVertices(region,tmp);
    
    int map[1][6] = { {22,28,30,31,25,29} };
    
    for (int pt = 0 ; pt < 1; pt++)
    {
      stencil_structure st;
      st.regions.push_back(region);
      st.v = region->getHighOrderPoint(pt);
      for (int n = 0 ; n < 6 ; n++)
        st.neighbors.push_back( tmp[map[pt][n]] );
      stencil.push_back(st);
    }
    tmp.clear();
  }
  
  
  template<> void getFaceCavityHONodesStencil< 2 >(pPList cavity, vector< stencil_structure > &stencil)
  {   
    // Edges HO nodes
    int nte[3]    = { 0,1,2 };        // Node to Edge
    
    int ns = 2;
    double weights[3][6] = {
                            {1.,1.,0.,0.,0.,0.},
                            {1.,1.,0.,0.,0.,0.},
                            {1.,1.,0.,0.,0.,0.} };
      
    int map[3][6] = { { 0, 1,-1,-1,-1,-1},
                      { 1, 2,-1,-1,-1,-1},
                      { 0, 2,-1,-1,-1,-1} };
    
    void *temp = NULL;
    while( pFace face = (pFace)PList_next(cavity,&temp) ) 
    {
      // TODO : TO generalized based on "getRegionCavityHONodesStencil" ...
      bool goodEdge[3] = {true,true,true};
      // [0.1] Check by Edges Geometrical entity
      for (int e = 0 ; e < 3 ; e++)
      {
        pEdge edge = F_edge(face,e);
        if ( E_whatInType(edge) < 2 )
          goodEdge[e] = false;
      }
      
      
      vector<pVertex> tmp;
      MAdMapping::getVertices(face,tmp);
      
      for (int pt = 0 ; pt < 3; pt++)
      {
        if ( !goodEdge[nte[pt]] ) continue;
        
        bool OK = true;
        stencil_structure st;
        st.faces.push_back(face);
        st.v = tmp[pt+3];
        
        // Check if node already treated
        for( int i = 0 ; i < (int) stencil.size() ; i++)
          if ( stencil[i] == st ) 
          {
            OK = false;
            break;
          }
        if (!OK) continue;
        
        for (int n = 0 ; n < ns ; n++)
        {
          if(map[pt][n] != -1) 
          {
            st.neighbors.push_back( tmp[map[pt][n]] );
            st.weights.push_back(weights[pt][n]);
          }
          else
          {
            pEdge edge = F_edge(face, nte[pt] );
            pFace oface = NULL;
            for (int e = 0 ; e < E_numFaces(edge) ; e++)
            {
              oface = E_face(edge,e);
              if ( (oface != face) && PList_inList(cavity,oface) ) break;
              oface = NULL;
            }
            // is On boundary ?
            if (!oface)
            {
              OK = false; break;
            }
            st.faces.push_back(oface);
            vector<pVertex> tmpA;
            MAdMapping::getVertices( (pEntity) oface,tmpA);
            int id = getNodeIdInEntity(oface,st.v);
            for (int nn = 0 ; nn < ns ; nn++)
            {
              if(map[id-3][nn] != -1) 
              {
                if(find(st.neighbors.begin(), st.neighbors.end(), tmpA[map[id-3][nn]]) == st.neighbors.end())
                {
                  st.neighbors.push_back( tmpA[map[id-3][nn]] );
                  st.weights.push_back(weights[pt][n]);
                }
              }
            }
            tmpA.clear();
          }
        }
        if(OK) stencil.push_back(st);
      }
      
      // HO Faces Internal Nodes
      
      tmp.clear();
    }
  }
  
  template<> void getFaceCavityHONodesStencil< 3 >(pPList cavity, vector< stencil_structure > &stencil)
  {    
    
    // Edges HO nodes & Face HO nodes
    int ns = 6;
    int nte[6]    = { 0,0,1,1,2,2 };
    
    double weights[7][6] = {
                            {1.,1.,0.,0.,0.,0.},
                            {1.,1.,0.,0.,0.,0.},
                            {1.,1.,0.,0.,0.,0.},
                            {1.,1.,0.,0.,0.,0.},
                            {1.,1.,0.,0.,0.,0.},
                            {1.,1.,0.,0.,0.,0.},
                            {1.,1.,1.,1.,1.,1.} };
      
    int map[7][6] = { { 0, 4, 9,-1,-1,-1},
                      { 3, 1, 5,-1,-1,-1},
                      { 1, 6, 3,-1,-1,-1},
                      { 5, 2, 8,-1,-1,-1},
                      { 8, 2, 6,-1,-1,-1},
                      { 0, 7, 9,-1,-1,-1},
                      { 3, 5, 6, 8, 4, 7} };
      
    void *temp = NULL;
    while( pFace face = (pFace)PList_next(cavity,&temp) ) 
    {
      bool goodEdge[3] = {true,true,true};
      // [0.1] Check by Edges Geometrical entity
      for (int e = 0 ; e < 3 ; e++)
      {
        pEdge edge = F_edge(face,e);
        if ( E_whatInType(edge) < 2 )
          goodEdge[e] = false;
      }
      
      vector<pVertex> tmp;
      MAdMapping::getVertices(face,tmp);
      
      for (int pt = 0 ; pt < 7; pt++)
      {
        if ( pt < 6 && !goodEdge[nte[pt]] ) continue;
        
        bool OK = true;
        stencil_structure st;
        st.faces.push_back(face);
        st.v = tmp[pt+3];
        
        // Check if node already treated
        for( int i = 0 ; i < (int) stencil.size() ; i++)
        {
          if ( stencil[i] == st ) 
          {
            OK = false;
            break;
          }
        }
        if (!OK) continue;
        
        for (int n = 0 ; n < ns ; n++)
        {
          if(map[pt][n] != -1) 
          {
            st.neighbors.push_back( tmp[map[pt][n]] );
            st.weights.push_back(weights[pt][n]);
          }
          else
          {            
            pEdge edge = F_edge(face, nte[pt] );
            pFace oface = NULL;
            for (int e = 0 ; e < E_numFaces(edge) ; e++)
            {
              
              oface = E_face(edge,e);
              if ( (oface != face) && PList_inList(cavity,oface) ) break;
              oface = NULL;
            }
            // is On boundary ?
            if (!oface)
            {
              OK = false; break;
            }
            st.faces.push_back(oface);
            vector<pVertex> tmpA;
            MAdMapping::getVertices( (pEntity) oface,tmpA);
            int id = getNodeIdInEntity(oface,st.v);
            
            for (int nn = 0 ; nn < ns ; nn++)
            {
              if(map[id-3][nn] != -1) 
              {
                if(find(st.neighbors.begin(), st.neighbors.end(), tmpA[map[id-3][nn]]) == st.neighbors.end())
                {
                  st.neighbors.push_back( tmpA[map[id-3][nn]] );
                  st.weights.push_back(weights[pt][n]);
                }
              }
            }
            tmpA.clear();
          }
        }
        
        if(OK)
          stencil.push_back(st);
        
      }
      
      tmp.clear();
    }
  }
  
  template<> void getFaceCavityHONodesStencil< 4 >(pPList cavity, vector< stencil_structure > &stencil)
  {   
    // Edges HO nodes
    int nte[9]    = { 0,0,0,1,1,1,2,2,2 };
    
    double weights[12][4] = {
                              {1.,1.,0.,0.},
                              {1.,1.,0.,0.},
                              {1.,1.,0.,0.},
                              {1.,1.,0.,0.},
                              {1.,1.,0.,0.},
                              {1.,1.,0.,0.},
                              {1.,1.,0.,0.},
                              {1.,1.,0.,0.},
                              {1.,1.,0.,0.},
                              {1.,1.,1.,1.},
                              {1.,1.,1.,1.},
                              {1.,1.,1.,1.} };
      
    int map[12][4]= { { 0, 4,12,-1},
                      { 3, 5,13,-1},
                      { 4, 1, 6,-1},
                      { 1, 7, 4,-1},
                      { 6, 8,12,-1},
                      { 7, 2,10,-1},
                      {10, 2, 8,-1},
                      { 9,11,14,-1},
                      { 0,10,12,-1},
                      { 3,13,14,11},
                      { 4, 6, 7,12},
                      {12, 7, 8,10} };
      
    void *temp = NULL;
    while( pFace face = (pFace)PList_next(cavity,&temp) ) 
    {
      bool goodEdge[3] = {true,true,true};
      // [0.1] Check by Edges Geometrical entity
      for (int e = 0 ; e < 3 ; e++)
      {
        pEdge edge = F_edge(face,e);
        if ( E_whatInType(edge) < 2 )
          goodEdge[e] = false;
      }
      
      vector<pVertex> tmp;
      MAdMapping::getVertices(face,tmp);
      
      for (int pt = 0 ; pt < 12; pt++)
      {
        if ( pt < 9 && !goodEdge[nte[pt]] ) continue;
        
        bool OK = true;
        stencil_structure st;
        st.faces.push_back(face);
        st.v = tmp[pt+3];
        
        // Check if node already treated
        for( int i = 0 ; i < (int) stencil.size() ; i++)
        {
          if ( stencil[i] == st ) 
          {
            OK = false;
            break;
          }
        }
        if (!OK) continue;
        
        for (int n = 0 ; n < 4 ; n++)
        {
          if(map[pt][n] != -1) 
          {
            st.neighbors.push_back( tmp[map[pt][n]] );
            st.weights.push_back(weights[pt][n]);
          }
          else
          {
            pEdge edge = F_edge(face, nte[pt] );
            pFace oface = NULL;
            for (int e = 0 ; e < E_numFaces(edge) ; e++)
            {
              
              oface = E_face(edge,e);
              if ( (oface != face) && PList_inList(cavity,oface) ) break;
              oface = NULL;
            }
            // is On boundary ?
            if (!oface)
            {
              OK = false; break;
            }
            st.faces.push_back(oface);
            vector<pVertex> tmpA;
            MAdMapping::getVertices( (pEntity) oface,tmpA);
            int id = getNodeIdInEntity(oface,st.v);
            for (int nn = 0 ; nn < 4 ; nn++)
            {
              if(map[id-3][nn] != -1) 
              {
                if(find(st.neighbors.begin(), st.neighbors.end(), tmpA[map[id-3][nn]]) == st.neighbors.end())
                {
                  st.neighbors.push_back( tmpA[map[id-3][nn]] );
                  st.weights.push_back(weights[pt][n]);
                }
              }
            }
            tmpA.clear();
          }
        }
        if(OK) stencil.push_back(st);
      }
      
      tmp.clear();
    }
  }
  
  // ---------------------------------------------------------------------------------------------------------- //
  template<> void getRegionCavityHONodesStencil< 2 >(pPList cavity, vector< stencil_structure > &stencil)
  {    
    void *temp = NULL;
    while( pRegion region = (pRegion)PList_next(cavity,&temp) ) 
    {
      vector<pVertex> tmp;
      MAdMapping::getVertices(region,tmp);
      
      // [0] Check inner edges
      bool goodEdge[6] = {true,true,true,true,true,true};
      // [0.1] Check by Faces
      int faceToEdge[4][3] = { {0,1,2} , {0,3,4} , {1,4,5} , {2,3,5} };
      for (int f = 0 ; f < 4; f++)
      {
        pFace face = R_face(region,f);
        if ( !faceInCavity(cavity,face) )
        {
          goodEdge[ faceToEdge[f][0] ] = false;
          goodEdge[ faceToEdge[f][1] ] = false;
          goodEdge[ faceToEdge[f][2] ] = false;
        }
      }
      // [0.2] Check by Edges Geometrical entity
      for (int e = 0 ; e < 6 ; e++)
      {
        pEdge edge = R_edge(region,e);
        if ( E_whatInType(edge) <= 2 )
          goodEdge[e] = false;
      }
      
      // [1] Edges HO nodes (6 HO nodes on 6 edges [1 per edge])
      int edgeStencil[1][2] = { {0,2} };
      for (int e = 0 ; e < 6 ; e++)
      {
        if ( !goodEdge[e] ) continue;
        
        pEdge edge = R_edge(region,e);
        
        vector<pVertex> tmpE;
        MAdMapping::getVertices(edge,tmpE);
        for (int n = 0 ; n < 1 ; n++)
        {
          pVertex vert = tmpE[n+1];
          stencil_structure st;
          st.v = vert;
          st.lscale = E_length(edge); // TODO : To modify
          
          void * tempR = NULL;
          pPList ErL= E_regions(edge);
          while ( pRegion reg = (pRegion)PList_next(ErL,&tempR) ) {
              st.regions.push_back(reg);
          }
          PList_delete(ErL);
          
          st.neighbors.push_back( tmpE[ edgeStencil[n][0] ]);
          st.neighbors.push_back( tmpE[ edgeStencil[n][1] ]);
          st.weights.push_back(1.0);
          st.weights.push_back(1.0);
          
          // Check if node already treated
          bool OK = true;
          for( int i = 0 ; i < (int) stencil.size() ; i++)
          {
            if ( stencil[i] == st ) 
            {
              OK = false;
              break;
            }
          }
          if (OK) stencil.push_back(st);
        }
        tmpE.clear();
      }
      
      
      // [2] Faces HO nodes (0 HO nodes on 4 faces [nothing to do])
      
      
      // [3] Region HO nodes (0 HO nodes in 1 Region [nothing to do])
      
      
      tmp.clear();
    }
  }
  
  template<> void getRegionCavityHONodesStencil< 3 >(pPList cavity, vector< stencil_structure > &stencil)
  {    
    void *temp = NULL;
    while( pRegion region = (pRegion)PList_next(cavity,&temp) ) 
    {
      vector<pVertex> tmp;
      MAdMapping::getVertices(region,tmp);      
      
      // [0] Check inner edges
      bool goodEdge[6] = {true,true,true,true,true,true};
      // [0.1] Check by Faces
      int faceToEdge[4][3] = { {0,1,2} , {0,3,4} , {1,4,5} , {2,3,5} };
      for (int f = 0 ; f < 4; f++)
      {
        pFace face = R_face(region,f);
        if ( !faceInCavity(cavity,face) )
        {
          goodEdge[ faceToEdge[f][0] ] = false;
          goodEdge[ faceToEdge[f][1] ] = false;
          goodEdge[ faceToEdge[f][2] ] = false;
        }
      }
      
      // [0.1] Check by Edges Geometrical entity
      for (int e = 0 ; e < 6 ; e++)
      {
        pEdge edge = R_edge(region,e);
        if ( E_whatInType(edge) <= 2 )
          goodEdge[e] = false;
      }
      
      
      // [1] Edges HO nodes (12 HO nodes on 6 edges [2 per edge])
      int edgeStencil[2][2] = { {0,2} , {1,3} };
      for (int e = 0 ; e < 6 ; e++)
      {
        if ( !goodEdge[e] ) continue;
        
        pEdge edge = R_edge(region,e);
        
        vector<pVertex> tmpE;
        MAdMapping::getVertices(edge,tmpE);
        for (int n = 0 ; n < 2 ; n++)
        {
          pVertex vert = tmpE[n+1];
          stencil_structure st;
          st.v = vert;
          st.lscale = E_length(edge); // TODO : To modify
          
          void * tempR = NULL;
          pPList ErL= E_regions(edge);
          while ( pRegion reg = (pRegion)PList_next(ErL,&tempR) ) {
            if ( PList_inList(cavity,reg) )
              st.regions.push_back(reg);
          }
          PList_delete(ErL);
          
          st.neighbors.push_back( tmpE[ edgeStencil[n][0] ]);
          st.neighbors.push_back( tmpE[ edgeStencil[n][1] ]);
          st.weights.push_back(1.0);
          st.weights.push_back(1.0);
          
          // Check if node already treated
          bool OK = true;
          for( int i = 0 ; i < (int) stencil.size() ; i++)
          {
            if ( stencil[i] == st ) 
            {
              OK = false;
              break;
            }
          }
          if (OK) stencil.push_back(st);
        }
        tmpE.clear();
      }
      
      
      // [2] Faces HO nodes (4 HO nodes on 4 faces [1 per face])
      int faceStencil[1][4] = { {3,5,6,8} };
      for (int f = 0 ; f < 4 ; f++)
      {
        pFace face = R_face(region,f);
        
        if ( ! faceInCavity(cavity,face) ) continue;
        
        vector<pVertex> tmpF;
        MAdMapping::getVertices(face,tmpF);
        for (int n = 0 ; n < 1 ; n++)
        {
          pVertex vert = face->getHighOrderPoint(n);
          stencil_structure st;
          st.v = vert;
          st.lscale = sqrt(F_inscRadius(face));
          st.faces.push_back(face);
          
          void * tempR = NULL;
          pPList ErL= F_regions(face);
          while ( pRegion reg = (pRegion)PList_next(ErL,&tempR) ) {
            if ( PList_inList(cavity,reg) )
              st.regions.push_back(reg);
          }
          PList_delete(ErL);
          
          for (int s = 0 ; s < 4 ; s++)
          {
            st.neighbors.push_back( tmpF[ faceStencil[n][s] ]);
            st.weights.push_back( 1.0 );
          }
          
          
          // Check if node already treated
          bool OK = true;
          for( int i = 0 ; i < (int) stencil.size() ; i++)
          {
            if ( stencil[i] == st ) 
            {
              OK = false;
              break;
            }
          }
          if (OK) stencil.push_back(st);
        }
        tmpF.clear();
        
      }
      
      // [3] Region HO nodes (0 HO nodes in 1 Region [nothing to do])
      
      
      tmp.clear();
    }
  }
  
  template<> void getRegionCavityHONodesStencil< 4 >(pPList cavity, vector< stencil_structure > &stencil)
  {    
    void *temp = NULL;
    while( pRegion region = (pRegion)PList_next(cavity,&temp) ) 
    {
      vector<pVertex> tmp;
      MAdMapping::getVertices(region,tmp);      
      
      // [0] Check inner edges
      bool goodEdge[6] = {true,true,true,true,true,true};
      // [0.1] Check by Faces
      int faceToEdge[4][3] = { {0,1,2} , {0,3,4} , {1,4,5} , {2,3,5} };
      for (int f = 0 ; f < 4; f++)
      {
        pFace face = R_face(region,f);
        if ( !faceInCavity(cavity,face) )
        {
          goodEdge[ faceToEdge[f][0] ] = false;
          goodEdge[ faceToEdge[f][1] ] = false;
          goodEdge[ faceToEdge[f][2] ] = false;
        }
      }
      
      // [0.1] Check by Edges Geometrical entity
      for (int e = 0 ; e < 6 ; e++)
      {
        pEdge edge = R_edge(region,e);
        if ( E_whatInType(edge) <= 2 )
          goodEdge[e] = false;
      }
      
      // [1] Edges HO nodes (18 HO nodes on 6 edges [3 per edge])
      int edgeStencil[3][2] = { {0,2} , {1,3} , {2,4} };
      for (int e = 0 ; e < 6 ; e++)
      {
        if ( !goodEdge[e] ) continue;
        
        pEdge edge = R_edge(region,e);
        
        vector<pVertex> tmpE;
        MAdMapping::getVertices(edge,tmpE);
        for (int n = 0 ; n < 3 ; n++)
        {
          pVertex vert = tmpE[n+1];
          stencil_structure st;
          st.v = vert;
          st.lscale = E_length(edge); // TODO : To modify
          
          void * tempR = NULL;
          pPList ErL= E_regions(edge);
          while ( pRegion reg = (pRegion)PList_next(ErL,&tempR) ) {
            if ( PList_inList(cavity,reg) )
              st.regions.push_back(reg);
          }
          PList_delete(ErL);
          
          st.neighbors.push_back( tmpE[ edgeStencil[n][0] ]);
          st.neighbors.push_back( tmpE[ edgeStencil[n][1] ]);
          st.weights.push_back(1.0);
          st.weights.push_back(1.0);
          
          // Check if node already treated
          bool OK = true;
          for( int i = 0 ; i < (int) stencil.size() ; i++)
          {
            if ( stencil[i] == st ) 
            {
              OK = false;
              break;
            }
          }
          if (OK) stencil.push_back(st);
        }
        tmpE.clear();
      }
      
      // [2] Faces HO nodes (12 HO nodes on 4 faces [3 per face])
      int faceStencil[3][4] = { {3,11,13,14} , {4,6,7,12} , {7,8,10,12} };
      for (int f = 0 ; f < 4 ; f++)
      {
        pFace face = R_face(region,f);
        
        if ( ! faceInCavity(cavity,face) ) continue;
        
        vector<pVertex> tmpF;
        MAdMapping::getVertices(face,tmpF);
        for (int n = 0 ; n < 3 ; n++)
        {
          pVertex vert = face->getHighOrderPoint(n);
          stencil_structure st;
          st.v = vert;
          st.lscale = sqrt(F_inscRadius(face));
          st.faces.push_back(face);
          
          void * tempR = NULL;
          pPList ErL= F_regions(face);
          while ( pRegion reg = (pRegion)PList_next(ErL,&tempR) ) {
            if ( PList_inList(cavity,reg) )
              st.regions.push_back(reg);
          }
          PList_delete(ErL);
          
          for (int s = 0 ; s < 4 ; s++)
          {
            st.neighbors.push_back( tmpF[ faceStencil[n][s] ]);
            st.weights.push_back( 1.0 );
          }
          
          
          // Check if node already treated
          bool OK = true;
          for( int i = 0 ; i < (int) stencil.size() ; i++)
          {
            if ( stencil[i] == st ) 
            {
              OK = false;
              break;
            }
          }
          if (OK) stencil.push_back(st);
        }
        tmpF.clear();
        
      }
      
      // [3] Region HO nodes (1 HO nodes in 1 Region [1 per region])
      int regionStencil[1][6] = { {22,25,28,29,30,31} };
      for(int rn = 0 ; rn < 1 ; rn++)
      {
        pVertex vert = region->getHighOrderPoint(rn);
        stencil_structure st;
        st.v = vert;
        st.lscale = R_inscrRad(region);
        st.regions.push_back(region);
        for (int s = 0 ; s < 6 ; s++)
        {
          st.neighbors.push_back( tmp[ regionStencil[rn][s] ]);
          st.weights.push_back( 1.0 );
        }
        
        // Check if node already treated
        bool OK = true;
        for( int i = 0 ; i < (int) stencil.size() ; i++)
        {
          if ( stencil[i] == st ) 
          {
            OK = false;
            break;
          }
        }
        if (OK) stencil.push_back(st);
      }
      
      tmp.clear();
    }
  }
  // ---------------------------------------------------------------------------------------------------------- //
  //   
  // WARNING : Not working very well ... 
  //           Further work on the distorsionBasedQuality() method is required to evaluate correctly the distorsion criteria
  //
    Quality::Quality(pEntity ent) : forcePositive(false), entity(ent) {};
    
    double Quality::eval()
    {
      double val = 0.0;
      int type = EN_mshType(entity);
      
      switch (type)
      {
        case MSH_TYPE_TRI:
        {          
          // Compute Quality
          for (int q = 0 ; q < 16 ; q++)
          {
            double uvw[3] = { Triquad8[q][0] , Triquad8[q][1] , 0.0};
            double qual = distorsionBasedQuality(entity,uvw);
//             double qual = ConditionBasedQuality(entity,uvw);
            val += Triquad8[q][2]*qual;
          }
          break;
        }
        case MSH_TYPE_TET:
        {
          // Compute Quality
          for (int q = 0 ; q < 5 ; q++)
          {
            double uvw[3] = { Tetquad3[q][0] , Tetquad3[q][1] , Tetquad3[q][2] };
            double qual = distorsionBasedQuality(entity,uvw);
//             double qual = ConditionBasedQuality(entity,uvw);
            val += Tetquad3[q][3]*qual*6.0;
          }
          break;
        }
        default:
          MAdMsgSgl::instance().error(__LINE__,__FILE__,"Element type %d not supported in Quality evaluation.",type);
      }
      
      if (isnan(val)) return 0.0;
      
      return ( val == 0.0 ? 0.0 : 1.0/val);
    };
    
    // Quality based on mesh distorsion 
    double Quality::distorsionBasedQuality(pEntity ent, double uvw[3])
    {
      double qual = 0.0, dim = 2.0;
      double jac[9];
      
      // Evaluate the dimension of the problem
      switch( EN_mshType(ent) )
      {
        case MSH_TYPE_LIN:
          return 1.0;
        case MSH_TYPE_QUA:
          dim = 2.0;break;
        case MSH_TYPE_TRI:
          dim = 2.0;break;
        default:
          dim = 3.0;
      }
      
      double detW = 0.0;
      doubleMatrix W(3,3);
      W.set_all(0);
      // -------------------------------------------------------- //
      // IDEAL ELEMENTS (TRIANGLE and TETRAHEDRON)
      if (EN_mshType(ent) == MSH_TYPE_TRI)
      {
        doubleMatrix W2D(2,2);
        // Linear mapping between the reference element and the ideal element (W)
        W2D(0,0) = 1.0;W2D(0,1) = 0.5          ;
        W2D(1,0) = 0.0;W2D(1,1) = 0.5*sqrt(3.0);
        
        // Get W^{-1}
        W2D.invertInPlace();
        
        W(0,0) = W2D(0,0);W(0,1) = W2D(0,1);
        W(1,0) = W2D(1,0);W(1,1) = W2D(1,1);
        
        // Get det(W^{-1})
        detW = W2D.determinant();
      }
      if (EN_mshType(ent) == MSH_TYPE_TET)
      {
        // Linear mapping between the reference element and the ideal element (W)
        W(0,0) = 1.0;W(0,1) = 0.5         ;W(0,2) = 0.5              ;
        W(1,0) = 0.0;W(1,1) = 0.5*sqrt(3.);W(1,2) = sqrt(3.)/6.0     ;
        W(2,0) = 0.0;W(2,1) = 0.0         ;W(2,2) = sqrt(2.)/sqrt(3.);
        
        // Get W^{-1}
        W.invertInPlace();
        // Get det(W^{-1})
        detW = W.determinant();
      }
      // -------------------------------------------------------- //
      
      double detA = 0.0, sigma = 0.0, normS = 0.0, det = 0.0;
      pMapping mapping = Map_create(ent,1.0);
      
      // Get the Jacobian of the mapping between the reference and the physical space (A)
      Map_deval(mapping,uvw,jac);
      
      // Get the determinant of A : det(A)
      // NOTE : Map_determinant() returns always a non-negative determinant (See MAdMapping::determinant() ) for 2D Elements !!
      //        Thus for Triangles and Quads, it's not possible to evaluate if the element is tangled. 
      //        Need to evaluate localy the determinant for 2D Elements ...
      if (EN_dim(ent) == 2)
        detA = (jac[0]*jac[4] + jac[1]*jac[5] + jac[2]*jac[3]) - (jac[0]*jac[5] + jac[1]*jac[3] + jac[2]*jac[4]);
      else 
        detA = Map_determinant(mapping,uvw);
      
      Map_delete(mapping);
      doubleMatrix A(jac,3,3);
      doubleMatrix S(3,3);
      
      // Compute the transformation S = W^{-1}.A from the Ideal to the physical element
      S.set_all(0.0);
      W.transpose().mult(A.transpose(),S);
      
      // Get det(S) = det(A).det(W^{-1})
      det = detW*detA;
      
//       double epsilon = 1.e-15;
      
//       double delta = epsilon;
//       if (det <= 0.0) delta = sqrt(epsilon*(epsilon-det));
//       sigma = 0.5*( det + sqrt(det*det + delta*delta) );
      
      if (forcePositive)
      {
        // Asymptotic behaviour for det <= 0
        double delta = 0.0;
        if (det <= 0.0) delta = 0.05;
        sigma = 0.5*( det + sqrt(det*det + 4.0*delta*delta) );
      }
      else
      {
//         sigma = 0.5*( det + sqrt(det*det + epsilon*epsilon) );
        // Force in [0,1]
        sigma = 0.5*(det + abs(det));
//         sigma = det;
      }
//       sigma = 0.5*( det + sqrt(det*det + epsilon*epsilon) );
      
//       sigma = det;
      
      // Gets the Frobenius Norm of the transformation S
      normS = pow( S.frobenius() , 2.0 );
      
      // Compute the distorsion criterion
      qual = normS / ( dim * pow(sigma , 2.0/dim) );
      return qual;
    }
    
    
    // Condition number based quality 
    double Quality::ConditionBasedQuality(pEntity ent, double uvw[3])
    {
      double dim = 3.0;
      double qual = 0.0;
      double jac[9];
      
      doubleMatrix W(3,3);
      W.set_all(0);
      // -------------------------------------------------------- //
      // IDEAL ELEMENTS (TRIANGLE and TETRAHEDRON)
      if (EN_mshType(ent) == MSH_TYPE_TRI)
      {
        dim = 2.0;
        doubleMatrix W2D(2,2);
        
        W2D(0,0) = 1.0;W2D(0,1) = 0.5          ;
        W2D(1,0) = 0.0;W2D(1,1) = 0.5*sqrt(3.0);
        
        W2D.invertInPlace();
        
        W(0,0) = W2D(0,0);W(0,1) = W2D(0,1);
        W(1,0) = W2D(1,0);W(1,1) = W2D(1,1);
        
      }
      if (EN_mshType(ent) == MSH_TYPE_TET)
      {
        dim = 3.0;
        W(0,0) = 1.0;W(0,1) = 0.5         ;W(0,2) = 0.5              ;
        W(1,0) = 0.0;W(1,1) = 0.5*sqrt(3.);W(1,2) = sqrt(3.)/6.0     ;
        W(2,0) = 0.0;W(2,1) = 0.0         ;W(2,2) = sqrt(2.)/sqrt(3.);
        
        W.invertInPlace();
      }
      // -------------------------------------------------------- //
      
      double normS = 0.0, normSminusOne = 0.0;
      pMapping mapping = Map_create(ent,1.0);
      Map_deval(mapping,uvw,jac);
      Map_delete(mapping);
      doubleMatrix A(jac,3,3);
      doubleMatrix S(3,3);
      
      // Compute S
      S.set_all(0.0);
      W.mult(A,S);
      
      // Compute Frobenius Norm of S
      normS = S.frobenius();
      
      // Compute S^{-1}
      S.invertInPlace();
      
      // Compute Frobenius Norm of S^{-1}
      normSminusOne = S.frobenius();
      
      qual = 1.0/dim*normS*normSminusOne;
      return 1.0/qual;
    }
  
  double Quality::Tetquad2[4][4] = {
                                      {0.138196601125 , 0.138196601125 , 0.138196601125, 0.041666666667},
                                      {0.585410196625 , 0.138196601125 , 0.138196601125, 0.041666666667},
                                      {0.138196601125 , 0.585410196625 , 0.138196601125, 0.041666666667},
                                      {0.138196601125 , 0.138196601125 , 0.585410196625, 0.041666666667} };
  double Quality::Tetquad3[5][4] = {
                                      {0.250000000000 , 0.250000000000 , 0.250000000000,-0.133333333333},
                                      {0.166666666667 , 0.166666666667 , 0.166666666667, 0.075000000000},
                                      {0.166666666667 , 0.166666666667 , 0.500000000000, 0.075000000000},
                                      {0.166666666667 , 0.500000000000 , 0.166666666667, 0.075000000000},
                                      {0.500000000000 , 0.166666666667 , 0.166666666667, 0.075000000000} };
  
                                      
  double Quality::Triquad8[16][3] = { {0.333333333333 , 0.333333333333 , 0.144315607678},
                                      {0.459292588293 , 0.459292588293 , 0.0950916342673},
                                      {0.459292588293 , 0.0814148234146, 0.0950916342673},
                                      {0.0814148234146 , 0.459292588293, 0.0950916342673},
                                      {0.170569307752 , 0.170569307752 , 0.103217370535},
                                      {0.170569307752 , 0.658861384496 , 0.103217370535},
                                      {0.658861384496 , 0.170569307752 , 0.103217370535},
                                      {0.050547228317 , 0.050547228317 , 0.0324584976232},
                                      {0.050547228317 , 0.898905543366 , 0.0324584976232},
                                      {0.898905543366 , 0.050547228317 , 0.0324584976232},
                                      {0.263112829635 , 0.728492392955 , 0.0272303141744},
                                      {0.728492392955 , 0.008394777409 , 0.0272303141744},
                                      {0.008394777409 , 0.263112829635 , 0.0272303141744},
                                      {0.263112829635 , 0.008394777409 , 0.0272303141744},
                                      {0.728492392955 , 0.263112829635 , 0.0272303141744},
                                      {0.008394777409 , 0.728492392955 , 0.0272303141744} };
  double Quality::Triquad5[7][3] = {  {0.333333333333 , 0.333333333333 , 0.225000000000},
                                      {0.470142064105 , 0.470142064105 , 0.132394152789},
                                      {0.470142064105 , 0.059715871789 , 0.132394152789},
                                      {0.059715871789 , 0.470142064105 , 0.132394152789},
                                      {0.101286507323 , 0.101286507323 , 0.125939180545},
                                      {0.101286507323 , 0.797426985353 , 0.125939180545},
                                      {0.797426985353 , 0.101286507323 , 0.125939180545}};
  double Quality::Triquad4[6][3] = {  {0.44594849091597, 0.44594849091597, 0.22338158967801},
                                      {0.44594849091597, 0.10810301816807, 0.22338158967801},
                                      {0.10810301816807, 0.44594849091597, 0.22338158967801},
                                      {0.09157621350977, 0.09157621350977, 0.10995174365532},
                                      {0.09157621350977, 0.81684757298046, 0.10995174365532},
                                      {0.81684757298046, 0.09157621350977, 0.10995174365532} };
  double Quality::Triquad3[4][3] = {  {0.33333333333333, 0.33333333333333, -0.56250000000000},
                                      {0.20000000000000, 0.20000000000000,  0.52083333333333},
                                      {0.20000000000000, 0.60000000000000,  0.52083333333333},
                                      {0.60000000000000, 0.20000000000000,  0.52083333333333} };
  
  
}

namespace MAd {
  
  /* ***********************************************************************************
   * 
   *                                Regularizer::CONSTANTS
   * 
   * *********************************************************************************** */
  // CONSTANTS
  int Regularizer::Niter    =  10 ;
  double Regularizer::relax = 0.95;

  /* ***********************************************************************************
   * 
   *                        Regularizer::CONSTRUCTORS and DESTRUCTOR
   * 
   * *********************************************************************************** */
  Regularizer::Regularizer() : e(NULL), Geom(NULL)
  {

  }

  Regularizer::Regularizer(const Regularizer& other)
  {

  }

  Regularizer::~Regularizer()
  {

  }
  
  /* ***********************************************************************************
   * 
   *                                Regularizer::PUBLIC METHODS
   * 
   * *********************************************************************************** */
  
  void Regularizer::apply()
  {
    if (!e) return;
    
    int mshType = EN_mshType(e);
    switch (mshType)
    {
      // Edge
      case MSH_TYPE_LIN:
        applyEdge();
        break;
        
      // Face
      case MSH_TYPE_TRI:
//         applyFace();
        applyFaceMappingBased();
        break;
        
      // Region
      case MSH_TYPE_TET:
        applyRegion();
        break;
      
        // Not implemented
      default:
        throw;
    };
  }
  
  /* ***********************************************************************************
   * 
   *                                Regularizer::PRIVATE METHODS
   * 
   * *********************************************************************************** */
  double Regularizer::applyEdge()
  {
    pEdge edge = (pEdge) e;
    int order = EN_order(edge);
    
    // Has internal nodes ??
    if (order < 2) return 0.0;
    
    int nbHOnodes = edge->getNbHighOrderPoints();
    double du = 2.0/order;
    
    double old[3],coord[3],tmp[3];
    double uvw[3];
    double error = MAdBIG;
    int iter = Niter;
    while(error > 1.e-06 && iter--)
    {
      error = 0.0;
      // Foreach HO nodes
      for (int pt = 0 ; pt < nbHOnodes ; pt++)
      {
        memset(coord,0,3*sizeof(double));
        memset(  uvw,0,3*sizeof(double));
        
        double u = -1.0 + (pt+1)*2.0/order;
        
        // Get previous location
        pVertex vert = edge->getHighOrderPoint(pt);
        V_coord(vert,old);
        
        // Apply laplacian smoothing on "coord"
        pMapping mapping = Map_create(edge,1.0);
        
        double w = 1.0, weightSum = 0.0;
        uvw[0] = u + du;
        Map_eval(mapping,uvw,tmp);
        
        coord[0] += relax*tmp[0]*w;
        coord[1] += relax*tmp[1]*w;
        coord[2] += relax*tmp[2]*w;
        weightSum+= w;
        
        uvw[0] = u - du;
        Map_eval(mapping,uvw,tmp);
        
        coord[0] += relax*tmp[0]*w;
        coord[1] += relax*tmp[1]*w;
        coord[2] += relax*tmp[2]*w;
        weightSum+= w;
        
        coord[0] /= weightSum;
        coord[1] /= weightSum;
        coord[2] /= weightSum;
        
        uvw[0] = u;
        Map_eval(mapping,uvw,tmp);
        coord[0] += (1.0-relax)*tmp[0];
        coord[1] += (1.0-relax)*tmp[1];
        coord[2] += (1.0-relax)*tmp[2];
        Map_delete(mapping);
        
        // Need to be projected on a surface ?
        if (Geom)
          projectOnStencil(Geom,coord);
        
        // Relocate node
        V_setCoord(vert,coord);
        
        // How much did the node moved ?
        double dist = DISTANCE(old,coord);
        error = std::max(error,dist);
      }
    }
    
    return error;
  }

  
  // ----------------------------------------------------------------------------------- //
  
  double Regularizer::applyFace()
  {
    pFace face = (pFace) e;
    int order = EN_order(face);
    
    // Has internal nodes ??
    if (order < 3) return 0.0;
    
    int nbHOnodes = face->getNbHighOrderPoints();
    
    double du = 1.0/order;
    double *IndexToParam = new double[order+1];
    for (int i = 0 ; i < order+1 ; i++)
      IndexToParam[i] = 0.0 + i * du;
    
    // Stencil for laplacian (extra-diagonal terms)
    double Stencil[4][2] = { {-du,0.0} , {0.0,-du} , {du,0.0} , {0.0,du} };
    
    double old[3],coord[3],tmp[3];
    double uvw[3];
    double error = MAdBIG;
    int iter = Niter;
    while(error > 1.e-06 && iter--)
    {
      error = 0.0;
      // Foreach HO nodes
      for (int pt = 0 ; pt < nbHOnodes ; pt++)
      {
        memset(coord,0,3*sizeof(double));
        memset(  uvw,0,3*sizeof(double));
        
        // Get i,j indexes
        int j = pt/(order-2)+1;
        int i = pt%(order-2)+1;
        
        // Get previous location
        pVertex vert = face->getHighOrderPoint(pt);
        V_coord(vert,old);
        
        // Apply laplacian smoothing on "coord"
        double w = 1.0, weightSum = 0.0;
        pMapping mapping = Map_create(face,1.0);
        for(int p = 0 ; p < 4 ; p++)
        {
          uvw[0] = IndexToParam[i] + Stencil[p][0];
          uvw[1] = IndexToParam[j] + Stencil[p][1];
          Map_eval(mapping,uvw,tmp);
          
          coord[0] += relax*tmp[0]*w;
          coord[1] += relax*tmp[1]*w;
          coord[2] += relax*tmp[2]*w;
          weightSum+=w;
        }
        coord[0]/=weightSum;
        coord[1]/=weightSum;
        coord[2]/=weightSum;
        
        uvw[0] = IndexToParam[i];
        uvw[1] = IndexToParam[j];
        Map_eval(mapping,uvw,tmp);
        coord[0] += (1.0-relax)*tmp[0];
        coord[1] += (1.0-relax)*tmp[1];
        coord[2] += (1.0-relax)*tmp[2];
        Map_delete(mapping);
        
        // Need to be projected on a surface ?
        if (Geom)
          projectOnStencil(Geom,coord);
        
        // Relocate node
        V_setCoord(vert,coord);
        
        // How much did the node moved ?
        double dist = DISTANCE(old,coord);
        error = std::max(error,dist);
        
      }
      
    }
    
    delete [] IndexToParam;
    
    return error;
  }
  
  
  // ----------------------------------------------------------------------------------- //
  
  double Regularizer::applyRegion()
  {
    pRegion region = (pRegion) e;
    int order = EN_order(region);
    
    // Has internal nodes ??
    if (order < 4) return 0.0;
    
    // Stencil for laplacian (extra-diagonal terms)
    vector< RegulTools::stencil_structure > stencil;
    switch(order)
    {
      case 4:
        RegulTools::getRegionHONodesStencil< 4 >(region,stencil);
        break;
      default:
        MAdMsgSgl::instance().error(__LINE__,__FILE__,"CavityRegularizer::applyRegion() not implemented for order %d.",order); 
    }
    
    double old[3],coord[3],tmp[3];
    double error = MAdBIG;
    int iter = Niter;
    while(error > 1.e-06 && iter--)
    {
      error = 0.0;
      // Foreach HO nodes
      for (int pt = 0 ; pt < (int) stencil.size() ; pt++)
      {        
        // Get previous location
        pVertex vert = stencil[pt].v;
        V_coord(vert,old);
        V_coord(vert,coord);
        
        // Apply laplacian smoothing on "coord"
        coord[0] *= (1.0 -relax);
        coord[1] *= (1.0 -relax);
        coord[2] *= (1.0 -relax);
        double OneOverFour = 1.0 / 6.0;
        for (int n = 0 ; n < 6 ; n++)
        {
          V_coord( stencil[pt].neighbors[n] , tmp);
          coord[0] += relax*OneOverFour*tmp[0];
          coord[1] += relax*OneOverFour*tmp[1];
          coord[2] += relax*OneOverFour*tmp[2];
        }
        
        // Relocate node
        V_setCoord(vert,coord);
        
        // How much did the node moved ?
        double dist = DISTANCE(old,coord);
        error = std::max(error,dist);
        
      }
      
    }
    stencil.clear();
    return error;
  }
  
  
  // ----------------------------------------------------------------------------------- //
  
  double Regularizer::applyFaceMappingBased()
  {
    pFace face = (pFace) e;
    int order = EN_order(face);
    
    // Has internal nodes ??
    if (order < 3) return 0.0;
    
    vector< RegulTools::stencil_structure > stencil;
    switch(order)
    {
      case 3:
        RegulTools::getFaceHONodesStencil< 3 >(face,stencil);
        break;
      case 4:
        RegulTools::getFaceHONodesStencil< 4 >(face,stencil);
        break;
      case 5:
        RegulTools::getFaceHONodesStencil< 5 >(face,stencil);
        break;
      default:
        MAdMsgSgl::instance().error(__LINE__,__FILE__,"CavityRegularizer::applyFaceMappingBased() not implemented for order %d.",order); 
    }
    
    double old[3],coord[3],tmp[3];
    double error = MAdBIG;
    int iter = Niter;
    while(error > 1.e-06 && iter--)
    {
      error = 0.0;
      // Foreach HO nodes
      for (int pt = 0 ; pt < (int) stencil.size() ; pt++)
      {
        // Get previous location
        pVertex vert = stencil[pt].v;
        V_coord(vert,old);
        V_coord(vert,coord);
        
        // Apply laplacian smoothing
        coord[0] *= (1.0 -relax);
        coord[1] *= (1.0 -relax);
        coord[2] *= (1.0 -relax);
        double OneOverFour = 1.0 / 4.0;
        for (int n = 0 ; n < 4 ; n++)
        {
          V_coord( stencil[pt].neighbors[n] , tmp);
          coord[0] += relax*OneOverFour*tmp[0];
          coord[1] += relax*OneOverFour*tmp[1];
          coord[2] += relax*OneOverFour*tmp[2];
        }
        
        // Need to be projected on a surface ?
        if (Geom)
          projectOnStencil(Geom,coord);
        
        // Relocate node
        V_setCoord(vert,coord);
        
        // How much did the node moved ?
        double dist = DISTANCE(old,coord);
        error = std::max(error,dist);
        
      }
      
    }
    stencil.clear();
    return error;
  }
  
  /* ***********************************************************************************
   * 
   *                                CavityRegularizer::CONSTANTS
   * 
   * *********************************************************************************** */
  int CavityRegularizer::Niter    = 20;
  double CavityRegularizer::relax = 0.95;
  
  /* ***********************************************************************************
   * 
   *                        CavityRegularizer::CONSTRUCTORS and DESTRUCTOR
   * 
   * *********************************************************************************** */
  CavityRegularizer::CavityRegularizer() : order(1), cavity(NULL), constraints(NULL)
  {
    
  }

  CavityRegularizer::~CavityRegularizer()
  {
    
  }
  /* ***********************************************************************************
   * 
   *                           CavityRegularizer::PUBLIC METHODS
   * 
   * *********************************************************************************** */
  void CavityRegularizer::apply()
  {
    int type = MSH_TYPE_UNKNOWN;
    if (!cavity) return;
    
    void *temp = NULL; order = 1;
    while ( pEntity ent = PList_next(cavity,&temp) ) 
    {
      order = max(order, EN_order(ent));
      type = EN_mshType(ent);
    }
    
    if (order < 2) return;  // Only HO Nodes are modified
    
    //! \note : The boundaries of the cavity are forced as Dirichlet
    switch (type)
    {
      case MSH_TYPE_LIN:
        applyLine();break;
      case MSH_TYPE_TRI:
        applySurf();break;
      case MSH_TYPE_TET:
        applyVol();break;
      default:
        MAdMsgSgl::instance().error(__LINE__,__FILE__,"CavityRegularizer::apply() does not support Mesh Type %d.",type);
    }
    
  }

  /* ***********************************************************************************
   * 
   *                          CavityRegularizer::PRIVATE METHODS
   * 
   * *********************************************************************************** */
  void CavityRegularizer::applyLine()
  {
    vector<RegulTools::stencil_structure> stencil;
    
    void *temp = NULL;
    while( pEdge edge = (pEdge)PList_next(cavity,&temp) ) 
    {
      vector<pVertex> tmp;
      MAdMapping::getVertices(edge,tmp);
      
      for (int pt = 0 ; pt < order-1 ; pt++)
      {
        RegulTools::stencil_structure st;
        st.v = tmp[pt+1];
        st.neighbors.push_back( tmp[pt] );
        st.neighbors.push_back( tmp[pt+2] );
        st.weights.push_back( 1.0 );
        st.weights.push_back( 1.0 );
        stencil.push_back(st);
      }
      tmp.clear();
    }
    
    int nbPoints = (int) stencil.size();
    
    // APPLY LAPLACIAN SMOOTHING
    double old[3],coord[3],tmp[3];
    double error = MAdBIG;
    int iter = Niter;
    while(error > 1.e-06 && iter--)
    {
      error = 0.0;
      // Foreach HO nodes
      for (int pt = 0 ; pt < nbPoints ; pt++)
      {
        // Get previous location
        pVertex vert = stencil[pt].v;
        V_coord(vert,old);
        
        // Apply laplacian smoothing
        V_coord(vert,coord);
        coord[0] *= (1.0 - relax);
        coord[1] *= (1.0 - relax);
        coord[2] *= (1.0 - relax);
        
        static int nbNeighbors = 2;
        
        double weightSum = 0.0;
        for (int n = 0 ; n < nbNeighbors ; n++) 
          weightSum += stencil[pt].weights[n];
        
        for (int n = 0 ; n < nbNeighbors ; n++)
        {
          V_coord( stencil[pt].neighbors[n] , tmp);
          
          double weight = stencil[pt].weights[n]/weightSum;
          
          coord[0] += relax*weight*tmp[0];
          coord[1] += relax*weight*tmp[1];
          coord[2] += relax*weight*tmp[2];
        }
        
        // Need to be projected on a surface ?
        if (constraints)
          projectOnStencil(constraints,coord);
        
        // Relocate node
        V_setCoord(vert,coord);
        
        // How much did the node moved ?
        double dist = DISTANCE(old,coord);
        error = std::max(error,dist);
        
      }
      
    }
    stencil.clear();
  }
  
  // ----------------------------------------------------------------------------------- //
  
  void CavityRegularizer::applySurf()
  {
    vector<RegulTools::stencil_structure> stencil;
    
    switch(order)
    {
      case 2:
        RegulTools::getFaceCavityHONodesStencil< 2 >(cavity,stencil);break;
      case 3:
        RegulTools::getFaceCavityHONodesStencil< 3 >(cavity,stencil);break;
      case 4:
        RegulTools::getFaceCavityHONodesStencil< 4 >(cavity,stencil);break;
      default:
        MAdMsgSgl::instance().error(__LINE__,__FILE__,"CavityRegularizer::applySurf() does not support order %d.",order); 
    }
    int nbPoints = (int) stencil.size();
    
    
    // APPLY LAPLACIAN SMOOTHING
    double old[3],coord[3],tmp[3];
    double error = MAdBIG;
    int iter = Niter;
    while(error > 1.e-06 && iter--)
    {
      error = 0.0;
      // Foreach HO nodes
      for (int pt = 0 ; pt < nbPoints ; pt++)
      {
        // Get previous location
        pVertex vert = stencil[pt].v;
        V_coord(vert,old);
        
        // Get previous quality
        int nbTri = (int)(stencil[pt].faces.size());
        double old_quality = 0.0, new_quality = 0.0;
        for(int t = 0 ; t < nbTri ; t++)
        {
          pFace fa = stencil[pt].faces[t];
          RegulTools::Quality qual(fa);
          old_quality+= qual.eval();
        }
        
        // Apply laplacian smoothing
        V_coord(vert,coord);
        coord[0] *= (1.0 - relax);
        coord[1] *= (1.0 - relax);
        coord[2] *= (1.0 - relax);
        
        int nbNeighbors = (int) stencil[pt].neighbors.size();
        
        double weightSum = 0.0;
        for (int n = 0 ; n < nbNeighbors ; n++) 
          weightSum += stencil[pt].weights[n];
        
        for (int n = 0 ; n < nbNeighbors ; n++)
        {
          V_coord( stencil[pt].neighbors[n] , tmp);
          
          double weight = (relax*stencil[pt].weights[n])/weightSum;
          
          coord[0] += weight*tmp[0];
          coord[1] += weight*tmp[1];
          coord[2] += weight*tmp[2];
        }
        
        // Need to be projected on a surface ?
        if (constraints)
          projectOnStencil(constraints,coord);
        
        // Relocate node
        V_setCoord(vert,coord);
        
        // Get new quality
        new_quality = 0.0;
        for(int t = 0 ; t < nbTri ; t++)
        {
          pFace fa = stencil[pt].faces[t];
          RegulTools::Quality qual(fa);
          new_quality+= qual.eval();
        }
        
        // Give it 3 chances to improve the mesh quality ...
        int test = 3;
        while(test-- && (new_quality < old_quality) )
        {
          // Under-relaxation
          coord[0] = 0.5*(old[0] + coord[0]);
          coord[1] = 0.5*(old[1] + coord[1]);
          coord[2] = 0.5*(old[2] + coord[2]);
          
          // Need to be projected on a surface ?
          if (constraints)
            projectOnStencil(constraints,coord);
          
          // Relocate node
          V_setCoord(vert,coord);  
          
          // Get new quality
          new_quality = 0.0;
          for(int t = 0 ; t < nbTri ; t++)
          {
            pFace fa = stencil[pt].faces[t];
            RegulTools::Quality qual(fa);
            new_quality+= qual.eval();
          }
        }
        
        // If it still does not improve mesh quality, move back
        if (new_quality < old_quality)
          V_setCoord(vert,old);
        
        // How much did the node moved ?
        double dist = DISTANCE(old,coord);
        error = std::max(error,dist);
        
      }
      
    }
    stencil.clear();
  }
  
  // ----------------------------------------------------------------------------------- //
  void CavityRegularizer::applyVol()
  {
    vector<RegulTools::stencil_structure> stencil;
    
    switch(order)
    {
      case 2:
        RegulTools::getRegionCavityHONodesStencil< 2 >(cavity, stencil);break;
      case 3:
        RegulTools::getRegionCavityHONodesStencil< 3 >(cavity, stencil);break;
      case 4:
        RegulTools::getRegionCavityHONodesStencil< 4 >(cavity, stencil);break;
      default:
        MAdMsgSgl::instance().error(__LINE__,__FILE__,"CavityRegularizer::applyVol() does not support order %d.",order); 
    }
    int nbPoints = (int) stencil.size();
    
    if (!nbPoints) return;
    
    // APPLY LAPLACIAN SMOOTHING
    double old[3],coord[3],tmp[3];
    double error = MAdBIG;
    int iter = Niter;
    while(error > 1.e-06 && iter--)
    {
      error = 0.0;
      // Foreach HO nodes
      for (int pt = 0 ; pt < nbPoints ; pt++)
      {
        // Get previous location
        pVertex vert = stencil[pt].v;
        V_coord(vert,old);
        
        // --------------------------------------------------------------------------- //
        // [1] Apply laplacian smoothing
        V_coord(vert,coord);
        coord[0] *= (1.0 - relax);
        coord[1] *= (1.0 - relax);
        coord[2] *= (1.0 - relax);
        
        int nbNeighbors = (int) stencil[pt].neighbors.size();
        
        double weightSum = 0.0;
        for (int n = 0 ; n < nbNeighbors ; n++) 
          weightSum += stencil[pt].weights[n];
        
        for (int n = 0 ; n < nbNeighbors ; n++)
        {
          V_coord( stencil[pt].neighbors[n] , tmp);
          
          double weight = stencil[pt].weights[n]/weightSum;
          
          coord[0] += relax*weight*tmp[0];
          coord[1] += relax*weight*tmp[1];
          coord[2] += relax*weight*tmp[2];
        }
        
        // [2.1] Relocate node (laplacian smoothing)
        V_setCoord(vert,coord);
        
        
        // How much did the node moved ?
        double dist = DISTANCE(old,coord);
        error = std::max(error,dist);
        
      }
      
    }
    stencil.clear();
    
    
  }
  // ----------------------------------------------------------------------------------- //
  
  

  
  
}
