// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#ifndef REGULARIZER_H
#define REGULARIZER_H

#include "MeshDataBase.h"
#include "MAdMapping.h"
#include "MeshDataBaseInterface.h"
#include "pointprojector.h"
#include "MAdMessage.h"
#include "NullModel.h"

#include <vector>
#include <utility>

using namespace std;

namespace RegulTools {
  
  //! Structure of the stencil used by the laplacian smoothing operator per node
  typedef struct stencil_structure
  {
    //! Central Vertex
    pVertex v;
    //! Characteristic Length scale
    double lscale;
    //! Stencil neighbors
    vector<pVertex> neighbors;
    //! Corresponding weights associated to the Corresponding neighbors
    vector<double> weights;
    //! List of faces implied 
    vector<pFace> faces;
    //! List of regions implied
    vector<pRegion> regions;
    
    //! equality comparison
    bool operator==(const stencil_structure& a) const
    {
      return (v == a.v);
    }
    
    ~stencil_structure()
    {
      neighbors.clear();
      weights.clear();
      faces.clear();
      regions.clear();
    }
  } stencil_structure;
  
  //! Templatized fonctions (order dependant) returning in argument the list of stencil_structures for each nodes of the cavity
  //! Only the HO nodes are considered. 1st Order Vertices are not considered !!
  //! -> cavity is a list of Regions
  template<int order> void getRegionCavityHONodesStencil(pPList cavity, vector< stencil_structure > &stencil);
  //! -> cavity is a list of Faces
  template<int order> void getFaceCavityHONodesStencil(pPList cavity, vector< stencil_structure > &stencil);
  
  //! Templatized fonctions (order dependant) returning in argument the list of stencil_structures for each internal HO nodes of the Element
  //! -> HO internal nodes of the face
  template<int order> void getFaceHONodesStencil(pFace face, vector< stencil_structure > &stencil);
  //! -> HO internal nodes of the Region
  template<int order> void getRegionHONodesStencil(pRegion region, vector< stencil_structure > &stencil);
  
  
  //! Class Returning a quality criterion for the element "ent" (Triangle or Tetrahedron)
  class Quality
  {
  public:
    //! Constructor
    Quality(pEntity ent);
    
    //! Return the quality criterion for "entity"
    double eval();
    
    //! Force the criterion to live in [0,1]
    void setPositive(bool on = true) {forcePositive = on;};
        
  private:
    //! Quality based on mesh distorsion 
    double distorsionBasedQuality(pEntity ent, double uvw[3]);
    //! Quality based on Element Condition Number
    double ConditionBasedQuality(pEntity ent, double uvw[3]);
    
  public:
    bool forcePositive;
    pEntity entity;
    
    //! Quadrature sets
    static double Triquad8[16][3];
    static double Triquad5[7][3];
    static double Triquad4[6][3];
    static double Triquad3[4][3];
    
    static double Tetquad2[4][4];
    static double Tetquad3[5][4];
  };
  
}

namespace MAd {
  
  /*! ***********************************************************************************
   * 
   *                                ENTITY REGULARIZER
   * 
   *  Regularization Procedure Based on Laplacian smoothing for HO internal nodes of
   *  a SINGLE Element.
   * 
   * *********************************************************************************** */
  class Regularizer
  {
  public:
    // ------------------------------------------------------------------- //
    Regularizer();
    
    Regularizer(const Regularizer& other);
    
    ~Regularizer();
    
    // ------------------------------------------------------------------- //
    //! Set the Element to regularize
    void setEntity(pEntity ent) {e = ent;};
    
    //! If the Element is part of the boundary, define the element set on which
    //! we need to project the HO nodes after moving them
    void setConstraint(pPList stencil) { Geom = stencil;};    
    // ------------------------------------------------------------------- //
    //! Apply the laplacian smoothing loop procedure
    void apply();
    // ------------------------------------------------------------------- //

  private:
    // ------------------------------------------------------------------- //
    //! Specificators for each type of elements
    double applyEdge();
    double applyFace();
    double applyRegion();
    
    double applyFaceMappingBased();
    // ------------------------------------------------------------------- //
    
  public:
    static int Niter;
    static double relax;
    pEntity e;
      
    pPList Geom;
  };
  
  /*! ***********************************************************************************
   * 
   *                                CAVITY REGULARIZER
   * 
   *  Regularization Procedure Based on Laplacian smoothing for HO internal nodes of
   *  A SET OF ELEMENTS forming a cavity.
   * 
   * *********************************************************************************** */
  class CavityRegularizer
  {
  public:
    // ------------------------------------------------------------------- //
    CavityRegularizer();
    
    ~CavityRegularizer();
    
    // ------------------------------------------------------------------- //
    //! Define the set of elements forming the cavity you wants to regularize
    void setCavity(pPList cav) {cavity = cav;};
    
    //! If the cavity is part of a boundary, define the element set on which
    //! we need to project the HO nodes after moving them
    void addConstraint(pPList stencil) { constraints = stencil;};
    // ------------------------------------------------------------------- //
    //! Apply the laplacian smoothing loop procedure
    void apply();
    // ------------------------------------------------------------------- //
    
  private:
    // ------------------------------------------------------------------- //
    //! Specificators for each type of elements
    void applyLine();
    void applySurf();
    void applyVol();
    // ------------------------------------------------------------------- //
    
  public:
    static int Niter;
    static double relax;
    int order;
    pPList cavity;
    pPList constraints;
  };
  
  
}

#endif // REGULARIZER_H

