// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#include "DihedralAngleEvaluator.h"
#include "MeshSizeBase.h"
#include "MAdDefines.h"

#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
using std::string;
#include <math.h>

namespace MAd {

  // -------------------------------------------------------------------
  dihedralAngleEvaluator::dihedralAngleEvaluator(const DiscreteSF * f): 
    elementEvaluatorBase(f)
  {}

  // -------------------------------------------------------------------
  int dihedralAngleEvaluator::F_shape(const pFace face, const double normal[3],
                                      double * shape) const
  {
    double fxyz[3][3];
    pMSize pS[3];

    void *iter=0;
    int i = 0;
    pPList fvts=F_vertices(face,1);
    while( pVertex vt=(pVertex)PList_next(fvts,&iter) ) {
      pS[i]=sizeField->findSize(vt);
      V_coord(vt,fxyz[i++]);  
    }
    PList_delete(fvts);

    return XYZ_F_shape(fxyz,pS,normal,shape);  
  }
  
  // -------------------------------------------------------------------
  int dihedralAngleEvaluator::F_shapeWithDisp(const pFace face, const double normal[3],
                                              double disp[3][3], double * shape) const
  {
    double fxyz[3][3];
    pMSize pS[3];

    void *iter=0;
    int i = 0;
    pPList fvts=F_vertices(face,1);
    while( pVertex vt=(pVertex)PList_next(fvts,&iter) ) {
      pS[i]=sizeField->findSize(vt);
      V_coord(vt,fxyz[i++]);  
    }
    PList_delete(fvts);

    for (int j=0; j<3; j++) for (int k=0; k<3; k++) fxyz[j][k] += disp[j][k];

    return XYZ_F_shape(fxyz,pS,normal,shape); 
  }
  
  // -------------------------------------------------------------------
  int dihedralAngleEvaluator::XYZ_F_shape(const double xyz[3][3], const pMSize pS[3], 
                                          const double nor[3], double * shape) const
  {
    if( sizeField->getType() == NULLSFIELD ) {
      if( ! XYZ_F_shape(xyz,pS[0],nor,shape) )
        { *shape=0; return 0; }
      return 1;
    }

//     // use central size to compute shape
//     pMSize pSZ = MS_interpolate(pS[0], pS[1], pS[2], 0.33, 0.33);
//     int res = XYZ_F_shape(xyz, pSZ, nor, shape);
//     if ( pSZ ) delete pSZ;
//     return res;

    // compute shapes using sizes at every vertex and retain the worst one
    double shp[3];
    for(int i=0; i<3; i++) {
      if( ! XYZ_F_shape(xyz, pS[i], nor, &shp[i]) )
        { *shape = 0.; return 0; }
    }
    *shape = std::min(std::min(shp[0],shp[1]),shp[2]);
    return 1;

//     // compute shapes using sizes at every vertex and retain the mean
//     double shp[3];
//     for(int i=0; i<3; i++) {
//       if( ! XYZ_F_shape(xyz, pS[i], nor, &shp[i]) )
//         { *shape = 0.; return 0; }
//     }
//     *shape = MAdTHIRD * ( shp[0] + shp[1] + shp[2] );
//     return 1;
  }
  
  // -------------------------------------------------------------------
  int dihedralAngleEvaluator::XYZ_F_shape(const double xyz[3][3], const pMSize pS, 
                                          const double nor[3], double * shape) const
  {
    double area = XYZ_F_area(xyz,nor);
    if ( area < 0. ) { *shape = 0.; return 0; }

    double e[3][3];
    diffVec(xyz[1],xyz[0],e[0]);
    diffVec(xyz[2],xyz[0],e[1]);
    diffVec(xyz[2],xyz[1],e[2]);
    
    for (int i=0; i<3; i++) {
      if ( fabs(e[i][0]) <= MAdTOL && 
           fabs(e[i][1]) <= MAdTOL && 
           fabs(e[i][2]) <= MAdTOL ) { *shape = 0.; return 0; }
      normalizeVec(e[i],e[i]);
    }

    double alpha = 1000.;
    alpha = std::min(alpha, dotProd(e[0],e[1]));
    for (int i=0; i<3; i++) e[0][i] *= -1.;
    alpha = std::min(alpha, dotProd(e[0],e[2]));
    alpha = std::min(alpha, dotProd(e[1],e[2]));

    if ( alpha - MAdTOL < -1. || alpha + MAdTOL > 1. ) { *shape = 0.; return 0; }

    double bestAngle = 60.; // angle in an equilateral triangle
    double maxAngle  = acos(alpha)*MAdPITH*180.;
//     printf("Best angle: %f, max angle: %f, alpha: %f\n", bestAngle, maxAngle, alpha);
    assert ( maxAngle+MAdTOL > bestAngle);

    *shape = ( 180. - maxAngle ) / ( 180. - bestAngle );

    if(*shape < MAdTOL) return 0;
    else if (*shape > (1.+MAdTOL) ) *shape = 1.;

    return 1;
  }

  // -------------------------------------------------------------------
  int dihedralAngleEvaluator::R_shape(const pRegion rgn, double * shape) const
  {
    double rxyz[4][3];
    pMSize pS[4];
  
    void *iter=0;
    int i=0;
    pPList verts=R_vertices(rgn);
    while( pVertex vt=(pVertex)PList_next(verts,&iter) ) {
      pS[i]=sizeField->findSize(vt);
      V_coord(vt,rxyz[i++]);
    }  
    PList_delete(verts);

    return XYZ_R_shape(rxyz,pS,shape);
  }

  // -------------------------------------------------------------------
  int dihedralAngleEvaluator::R_shapeWithDisp(const pRegion rgn, double disp[4][3], 
                                              double * shape) const
  {
    double rxyz[4][3];
    pMSize pS[4];
  
    void *iter=0;
    int i=0;
    pPList verts=R_vertices(rgn);
    while( pVertex vt=(pVertex)PList_next(verts,&iter) ) {
      pS[i]=sizeField->findSize(vt);
      V_coord(vt,rxyz[i++]);
    }  
    PList_delete(verts);

    for (int j=0; j<4; j++) for (int k=0; k<3; k++) rxyz[j][k] += disp[j][k];

    return XYZ_R_shape(rxyz,pS,shape);
  }

  // -------------------------------------------------------------------
  int dihedralAngleEvaluator::XYZ_R_shape(const double xyz[4][3], const pMSize pS[4], 
                                          double * shape) const
  {
    if( sizeField->getType() == NULLSFIELD ) {
      if( ! XYZ_R_shape(xyz,pS[0],shape) )
        { *shape=0.; return 0; }
      return 1;
    }

    // use central size
    pMSize pSZ = MS_interpolate(pS[0], pS[1], pS[2], pS[3], 0.25, 0.25, 0.25);

    // use minimal size
    //     pMSize pSZ;
    //     double hmin = MAdBIG;
    //     for(int i=0; i<4; i++) {
    //       if( pS[i] ) {
    //         double s = pS[i]->minSize();
    //         if( s <  hmin ) { hmin = s; pSZ = pS[i]; }
    //       }
    //     }
    //     assert( hmin != MAdBIG );

    int res = XYZ_R_shape(xyz,pSZ,shape);
    if ( pSZ ) delete pSZ;
    return res;
  }

  // -------------------------------------------------------------------
  int dihedralAngleEvaluator::XYZ_R_shape(const double xyz[4][3], const pMSize pS, 
                                          double * shape) const
  {
    double vol;
    vol = R_XYZ_volume(xyz);
    if ( vol < 0. ) { *shape = 0.; return 0; }

    double nor[4][3];
    double face[3][3];
    for (int i=0; i<3; i++) { face[0][i] = xyz[0][i]; face[1][i] = xyz[2][i]; face[2][i] = xyz[1][i]; }
    XYZ_F_normal(face,nor[0]);
    for (int i=0; i<3; i++) { face[0][i] = xyz[0][i]; face[1][i] = xyz[1][i]; face[2][i] = xyz[3][i]; }
    XYZ_F_normal(face,nor[1]);
    for (int i=0; i<3; i++) { face[0][i] = xyz[1][i]; face[1][i] = xyz[2][i]; face[2][i] = xyz[3][i]; }
    XYZ_F_normal(face,nor[2]);
    for (int i=0; i<3; i++) { face[0][i] = xyz[0][i]; face[1][i] = xyz[3][i]; face[2][i] = xyz[2][i]; }
    XYZ_F_normal(face,nor[3]);
    
    for (int i=0; i<4; i++) {
      if ( fabs(nor[i][0]) <= MAdTOL && 
           fabs(nor[i][1]) <= MAdTOL && 
           fabs(nor[i][2]) <= MAdTOL ) { *shape = 0.; return 0; }
      normalizeVec(nor[i],nor[i]);
    }
    
//     double alphaMax = -1000.;

    double alpha = 1000.;
    for (int i=0; i<3; i++) nor[0][i] *= -1.;
    alpha = std::min(alpha, dotProd(nor[0],nor[1]));
    alpha = std::min(alpha, dotProd(nor[0],nor[2]));
    alpha = std::min(alpha, dotProd(nor[0],nor[3]));
//     alphaMax = std::max(alphaMax, dotProd(nor[0],nor[1]));
//     alphaMax = std::max(alphaMax, dotProd(nor[0],nor[2]));
//     alphaMax = std::max(alphaMax, dotProd(nor[0],nor[3]));
    for (int i=0; i<3; i++) nor[1][i] *= -1.;
    alpha = std::min(alpha, dotProd(nor[1],nor[2]));
    alpha = std::min(alpha, dotProd(nor[1],nor[3]));
//     alphaMax = std::max(alphaMax, dotProd(nor[1],nor[2]));
//     alphaMax = std::max(alphaMax, dotProd(nor[1],nor[3]));
    for (int i=0; i<3; i++) nor[2][i] *= -1.;
    alpha = std::min(alpha, dotProd(nor[2],nor[3]));
//     alphaMax = std::max(alphaMax, dotProd(nor[2],nor[3]));

    if ( alpha - MAdTOL < -1. || alpha + MAdTOL > 1. ) { *shape = 0.; return 0; }

    double bestAngle = 2*acos(sqrt(2./3.))*180.*MAdPITH; // dihedral angle in an equilateral tetrahedron
    double maxAngle  = acos(alpha)*MAdPITH*180.;
//     printf("Best angle: %f, max angle: %f, alpha: %f\n", bestAngle, maxAngle, alpha);
    assert ( maxAngle+MAdTOL > bestAngle);

    *shape = ( 180. - maxAngle ) / ( 180. - bestAngle );

    if(*shape < MAdTOL) return 0;
    else if (*shape > (1.+MAdTOL) ) *shape = 1.;

    return 1;
  }

  // -------------------------------------------------------------------

}
