// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#include "BoxSizeField.h"

#include "IsoMeshSize.h"
#include "MAdDefines.h"

#include <cfloat>

namespace MAd
{
    IBoxSizeField * CreateBoxSizeField()
    {
        return new BoxSizeField();
    }

    // -------------------------------------------------------------------
    // Constructors / destructors
    // -------------------------------------------------------------------
    BoxSizeField::BoxSizeField() :
        SizeFieldBase(), IBoxSizeField(),
        sizeIn(-1.), sizeOut(DBL_MAX), useGradient(false), gradient(DBL_MAX)
    {
    }
    BoxSizeField::~BoxSizeField()
    {}

    // -------------------------------------------------------------------
    // Evaluate a local size
    // -------------------------------------------------------------------
    pMSize BoxSizeField::eval(const double xyz[3]) const
    {
        double size = 0.;
        if (!useGradient)
        {
            bool isIn = true;
            for (int i = 0; i < 3; i++)
            {
                if (xyz[i] < box[i] || xyz[i] > box[3 + i])
                {
                    isIn = false;
                    break;
                }
            }
            size = isIn ? sizeIn : sizeOut;
        }
        else
        {
            double distSq = 0.;
            for (int i = 0; i < 3; i++)
            {
                if (xyz[i] < box[i])
                    distSq += (box[i] - xyz[i]) * (box[i] - xyz[i]);
                else if (xyz[i] > box[3 + i])
                    distSq += (box[3 + i] - xyz[i]) * (box[3 + i] - xyz[i]);
            }

            if (distSq < 1.e-12)
            {
                size = sizeIn;
            }
            else
            {
                if (sizeIn < sizeOut)
                    size = std::min(sizeOut, sizeIn + std::sqrt(distSq) * gradient);
                else
                    size = std::max(sizeOut, sizeIn - std::sqrt(distSq) * gradient);
            }
        }

        return new IsoMeshSize(size);
    }

    // -------------------------------------------------------------------
    MeshSizeBase * BoxSizeField::getSize(const pVertex pv) const
    {
        double xyz[3];
        V_coord(pv, xyz);
        return eval(xyz);
    }

    // -------------------------------------------------------------------
    MeshSizeBase * BoxSizeField::getSizeOnEntity(const pEntity, const double xyz[3]) const
    {
        return eval(xyz);
    }

    // -------------------------------------------------------------------
    // Length squared computation
    // -------------------------------------------------------------------

    // -------------------------------------------------------------------
    double BoxSizeField::SF_VV_lengthSq(const pVertex pV0,
        const pVertex pV1) const
    {
        double xyz[2][3];
        V_coord(pV0, xyz[0]);
        V_coord(pV1, xyz[1]);

        pMSize pS[2];
        pS[0] = getSize(pV0);
        pS[1] = getSize(pV1);

        double lSq = SF_XYZ_lengthSq(xyz[0], xyz[1], pS[0], pS[1]);

        if (pS[0]) delete pS[0];
        if (pS[1]) delete pS[1];

        return lSq;
    }

    // -------------------------------------------------------------------
    double BoxSizeField::SF_XYZ_lengthSq(const double xyz0[3],
        const double xyz1[3],
        const pMSize pS0,
        const pMSize pS1) const
    {
        if (pS0)
        {
            double e[3];
            diffVec(xyz0, xyz1, e);
            double lenSq0 = pS0->normSq(e);
            if (pS1)
            {
                double lenSq1 = pS1->normSq(e);
                return sqrt(lenSq0*lenSq1);
            }
            else return lenSq0;
        }
        else {
            MAdMsgSgl::instance().error(__LINE__, __FILE__, "No size defined");
        }
        return 0.;
    }

    // -------------------------------------------------------------------
    // Area squared computation
    // -------------------------------------------------------------------

    double BoxSizeField::SF_F_areaSq(const pFace face) const
    {
        double area = 0.;

        double xyz[3][3];
        F_coordP1(face, xyz);

        void * temp = 0;
        pPList fVerts = F_vertices(face, 1);
        while (pVertex pV = (pVertex)PList_next(fVerts, &temp))
        {
            pMSize pS = getSize(pV);
            area += SF_XYZ_areaSq(xyz, pS, 0);
            if (pS) delete pS;
        }
        PList_delete(fVerts);

        area /= F_numVertices(face);

        return area;
    }

    // -------------------------------------------------------------------
    double BoxSizeField::SF_XYZ_areaSq(const double fxyz[3][3],
        const pMSize pS,
        const double norDir[3]) const
    {
        if (!pS) {
            MAdMsgSgl::instance().error(__LINE__, __FILE__, "No size defined");
        }

        // get the two first edges
        double e01[3], e02[3];
        diffVec(fxyz[1], fxyz[0], e01);
        diffVec(fxyz[2], fxyz[0], e02);

        double nor[3];
        crossProd(e01, e02, nor);

        double l1SqInv = 1. / pS->lengthSqInDir(e01);
        double l2SqInv = 1. / pS->lengthSqInDir(e02);

        if (norDir && dotProd(norDir, nor) < MAdTOL) return 0.;

        double areaSq = 0.25 * dotProd(nor, nor) * l1SqInv * l2SqInv;
        if (areaSq < MAdTOL) return 0.;

        return areaSq;
    }

    // -------------------------------------------------------------------
    // Volume computation
    // -------------------------------------------------------------------

    double BoxSizeField::SF_R_volume(const pRegion region) const
    {
        double vol = 0.;

        double xyz[4][3];
        R_coordP1(region, xyz);

        pPList rVerts = R_vertices(region);
        void * temp = 0;
        while (pVertex pV = (pVertex)PList_next(rVerts, &temp))
        {
            pMSize pS = getSize(pV);
            vol += SF_XYZ_volume(xyz, pS);
            if (pS) delete pS;
        }
        PList_delete(rVerts);

        vol /= R_numVertices(region);

        return vol;
    }

    // -------------------------------------------------------------------
    double BoxSizeField::SF_XYZ_volume(const double xyz[4][3],
        const pMSize pS) const
    {
        if (!pS) {
            MAdMsgSgl::instance().error(__LINE__, __FILE__, "No size defined");
        }

        double physVol = R_XYZ_volume(xyz);

        return (physVol / (pS->size(0)*pS->size(1)*pS->size(2)));
    }

    // -------------------------------------------------------------------
    // Center of edge computation
    // -------------------------------------------------------------------

    // -------------------------------------------------------------------
    double BoxSizeField::SF_E_center(const pEdge edge, double center[3],
        double * reducSq, pMSize * cSize) const
    {
        return SF_VV_center(E_vertex(edge, 0), E_vertex(edge, 1), center, reducSq, cSize);
    }

    // -------------------------------------------------------------------
    double BoxSizeField::SF_VV_center(const pVertex v0, const pVertex v1,
        double center[3], double * reducSq,
        pMSize * cSize) const
    {
        double xyz[2][3];
        V_coord(v0, xyz[0]);
        V_coord(v1, xyz[1]);

        pMSize pS[2];
        pS[0] = getSize(v0);
        pS[1] = getSize(v1);

        double cParam = SF_XYZ_center(xyz, pS, center, reducSq, cSize);

        if (pS[0]) delete pS[0];
        if (pS[1]) delete pS[1];

        return cParam;
    }

    // -------------------------------------------------------------------
    void BoxSizeField::scale(double fact)
    {
        sizeIn *= fact;
        sizeOut *= fact;
    }

    // -------------------------------------------------------------------

}
