// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#ifndef _H_ISIZEFIELD
#define _H_ISIZEFIELD

#include "MAdExport.h"
#include "MeshDataBaseInterface.h"

namespace MAd {

    class MeshSizeBase;
    typedef class MeshSizeBase * pMSize;

    // -------------------------------------------------------------------
    enum sizeFieldType {
        UNKNOWNSFIELDTYPE,
        NULLSFIELD,
        DISCRETESFIELD,
        ANALYTICALSFIELD,
        LOCALSFIELD,
        BACKGROUNDSFIELD,
        BOXSFIELD
    };

    typedef pMSize(*sizeFunction)(const double[3], double);

    // -------------------------------------------------------------------
    // Size objects services
    MAD_EXPORT pMSize CreateIsoSize(double size);

    // -------------------------------------------------------------------
    class MAD_EXPORT ISizeField
    {
    public:

        virtual ~ISizeField() {};

        virtual sizeFieldType getType() const = 0;
        virtual const char * getName() const = 0;

        // get the size at a location
        // --- Memory is allocated ! ---
        virtual pMSize getSize(const pVertex) const = 0;
        virtual pMSize getSizeOnEntity(const pEntity, const double[3]) const = 0;

        // visualisation of the isotropic size in a .pos file
        virtual void printPosIsotropic(const pMesh mesh, const char * fileName) = 0;
        // visualisation of the anisotropic size in a .pos file
        virtual void printPosAnisotropic(const pMesh mesh, const char * fileName) = 0;

    };

    // -------------------------------------------------------------------
    class MAD_EXPORT IAnalyticalSizeField : public virtual ISizeField
    {
    public:
        virtual ~IAnalyticalSizeField() {};

        // set the size
        virtual void setSize(sizeFunction) = 0;
        virtual void setSize(const char * size) = 0;
    };

    MAD_EXPORT IAnalyticalSizeField * CreateAnalyticalSizeField();

    // -------------------------------------------------------------------
    class MAD_EXPORT ILocalSizeField : public virtual ISizeField
    {
    public:
        virtual ~ILocalSizeField() {};
        
        virtual void addGeometricEntity(int type, int tag) = 0;
        virtual void updateTree() = 0;

        virtual void setIsoSize(double _radius, const char * _sizeN) = 0;
        virtual void setAnisoSize(double _radius, const char * _sizeN, const char * _sizeT) = 0;
        virtual void setCurvatureLimiter(double onTgSize, double _maxCurv = 1.e12, double maxCurvGrad = 1.e12) = 0;
    };

    MAD_EXPORT ILocalSizeField * CreateLocalSizeField(pMesh m);

    // -------------------------------------------------------------------

    class MAD_EXPORT IPiecewiseLinearSizeField : public virtual ISizeField
    {
    public:
        virtual ~IPiecewiseLinearSizeField() {};
        
        virtual pMesh getMesh() const = 0;

        // set the size at a vertex 
        virtual void setVSize(pVertex, double[3][3], double[3]) = 0;
        virtual void setVSize(pVertex, pMSize) = 0;
        virtual void setVSize(pVertex, double) = 0;
        virtual void setVSize(int, double) = 0;

        // set a size at all vertices 
        virtual void setCurrentSize() = 0;
        virtual void setCurrentSizeScaled(double fact) = 0;
        virtual void setCurvatureSize(bool aniso,
            double alpha = 2., // prescribe 2*PI*alpha edges around a circle
            double hMin = 1.e-4) = 0; // minimal size
        virtual void setAllVSizes(pMSize) = 0;
        virtual void setAllVSizes(double[3][3], double[3]) = 0;
        virtual void setAllVSizes(double) = 0;
        
        // delete sizes
        virtual void cleanUp() = 0;
        virtual void deleteSize(pEntity) = 0;

        // Intersect with another size field
        virtual void intersect(const ISizeField *) = 0;
        virtual void intersect(pVertex pv, const ISizeField *) = 0;

        // smooth the size field
        virtual void smooth(double maxGradient) = 0;
    };

    MAD_EXPORT IPiecewiseLinearSizeField * CreatePiecewiseLinearSizeField(pMesh m);

    // -------------------------------------------------------------------
    class MAD_EXPORT IBackgroundSizeField : public virtual ISizeField
    {
    public:
        ~IBackgroundSizeField() {};

        virtual void loadData(const char * fileName) = 0;
    };

    MAD_EXPORT IBackgroundSizeField * CreateBackgroundSizeField();

    // -------------------------------------------------------------------
    class MAD_EXPORT IBoxSizeField : public virtual ISizeField
    {
    public:
        virtual ~IBoxSizeField() {};

        virtual void setBox(double box[6]) = 0;
        virtual void setSizeIn(double size) = 0;
        virtual void setSizeOut(double size) = 0;
        virtual void setGradient(bool useGradient, double grad) = 0;
    };

    MAD_EXPORT IBoxSizeField * CreateBoxSizeField();

    // -------------------------------------------------------------------
}

// -------------------------------------------------------------------

#endif
