// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#ifndef _H_PWLINEARSFIELD
#define _H_PWLINEARSFIELD

#include "MAdExport.h"
#include "DiscreteSF.h"
#ifdef PARALLEL
#include "MeshDataBaseComm.h"
#endif

#include <set>

namespace MAd {

  // -------------------------------------------------------------------
  class PWLSField : public DiscreteSF, public IPiecewiseLinearSizeField
  {
  public:

    PWLSField(pMesh, std::string name="");
    ~PWLSField();

    virtual sizeFieldType getType() const { return DiscreteSF::getType(); } // avoiding VS buggy warnings
    virtual const char * getName() const { return SizeFieldBase::getName(); } // avoiding VS buggy warnings
    virtual pMesh getMesh() const { return DiscreteSF::getMesh(); } // avoiding VS buggy warnings

    DiscreteSFType discretization() const { return VERTEX_P1_DSFTYPE; }

    // delete sizes
    void cleanUp();
    virtual void deleteSize(pEntity ent) { DiscreteSF::deleteSize(ent); } // avoiding VS buggy warnings

    // Intersect with another size field
    void intersect(const ISizeField *);
    void intersect(pVertex pv, const ISizeField *);

    // smooth the size field
    void smooth(double);

    // set the size at a vertex 
    virtual void setVSize(pVertex v, double dirs[3][3], double s[3]) { DiscreteSF::setSize((pEntity)v, dirs, s); }
    virtual void setVSize(pVertex v, pMSize s) { DiscreteSF::setSize((pEntity)v, s); }
    virtual void setVSize(pVertex v, double s) { DiscreteSF::setSize((pEntity)v, s); }
    virtual void setVSize(int v, double s) { DiscreteSF::setSize(v, s); }

    // set a size at all vertices 
    void setCurrentSize();
    void setCurrentSizeScaled(double fact);
    void setCurvatureSize(bool aniso, 
                          double alpha=2., // prescribe 2*PI*alpha edges around a circle
                          double hMin=1.e-4); // minimal size
    void setAllVSizes(pMSize);
    void setAllVSizes(double[3][3], double[3]);
    void setAllVSizes(double);
    void scale(double);

    // get the size at a location (allocates memory!)
    MeshSizeBase * getSize(const pVertex) const;
    MeshSizeBase * getSizeOnEntity(const pEntity, const double[3]) const;

    // get the size at a vertex (do not allocate space)
    pMSize findSize(const pVertex) const;
    pMSize findSize(const pVertex);

    // edge length (squared)
    double SF_VV_lengthSq(const pVertex, const pVertex) const;
    double SF_XYZ_lengthSq(const double[3], const double[3],
                           const pMSize, const pMSize=NULL) const;

    // face area (squared)
    double SF_F_areaSq(const pFace) const;
    double SF_XYZ_areaSq(const double[3][3], const pMSize,
                         const double[3]) const;

    // region volume
    double SF_R_volume(const pRegion) const;
    double SF_XYZ_volume(const double[4][3], const pMSize) const;

    // center and its associated size
    double SF_E_center(const pEdge, double[3], double * reducSq, pMSize *) const;
    double SF_VV_center(const pVertex, const pVertex,
                        double[3], double * reducSq, pMSize *) const;

    // interpolate the size at parametric location u
    pMSize SF_E_interpolate(const pEdge edge, double u) const;

    void registerInCallbackManager();

    // redeclaration to avoid VS buggy warnings
    virtual void printPosIsotropic(const pMesh m, const char * fileName) { SizeFieldBase::printPosIsotropic(m, fileName); }
    virtual void printPosAnisotropic(const pMesh m, const char * fileName) { SizeFieldBase::printPosAnisotropic(m, fileName); }

  protected:

    pMSize getSizeOnEdge(const pEdge, const double[3]) const;
    pMSize getSizeOnEdgeParam(const pEdge, const double) const;
    pMSize getSizeOnFace(const pFace, const double[3]) const;
    pMSize getSizeOnFaceParam(const pFace, const double[2]) const;
    pMSize getSizeOnRegion(const pRegion, const double[3]) const;
    pMSize getSizeOnRegionParam(const pRegion, const double[3]) const;

    void smoothOnEdge(const pEdge, double, std::set<pEdge>*);
  };

  // -------------------------------------------------------------------
#ifdef PARALLEL
  class PWLSFieldDE : public MDB_DataExchanger
  {
  public :
    
    PWLSField *field;
    
    PWLSFieldDE(PWLSField *f);
    virtual ~PWLSFieldDE();
    
    virtual void * sendData (pEntity pe,    // in
                             int iProcDest, // in
                             int &_size );
    
    virtual void receiveData (pEntity pe,      //in
                              int iProcSender, //in
                              void *buf );
    
    virtual void deleteExternalData(pEntity pe) const;
  };
#endif

}

// -------------------------------------------------------------------

#endif
