// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#include "SizeFieldManager.h"
#include "NullSField.h"
#include "MAdTimeManager.h"
#include "MeshQualityManager.h"
#include "MAdMessage.h"
#include "MeshSizeBase.h"

#include <iostream>
using std::cerr;
using std::cout;
using std::endl;
using std::set;

namespace MAd {

  // -------------------------------------------------------------------
  SizeFieldManager::SizeFieldManager(pMesh m, pSField psf): 
    localTime(-1.),smooth(false),maxGradient(1.)
  {
    if (!m) return;
    mesh = m;

    mainSF = new PWLSField(mesh,"Main size field");

    localTime = MAdTimeManagerSgl::instance().getTime();

    if (!psf) MAdMsgSgl::instance().warning(__LINE__,__FILE__,"No size field registered");

    else {
    
      mainSF->intersect(psf);
  
      sizeFieldType type = psf->getType();
      switch (type) {
      case NULLSFIELD: {
        break;
      }
      case DISCRETESFIELD: {
        PWLSField* linsf = dynamic_cast<PWLSField*>(psf);
        addPWLinearSF( linsf );
        break;
      }
      case ANALYTICALSFIELD: {
        AnalyticalSField* asf = dynamic_cast<AnalyticalSField*>(psf);
        addAnalyticalSF( asf );
        break;
      }
      case LOCALSFIELD: {
        LocalSizeField* locsf = dynamic_cast<LocalSizeField*>(psf);
        addLocalSF( locsf );
        break;
      }
      case BACKGROUNDSFIELD: {
        BackgroundSF* bg = dynamic_cast<BackgroundSF*>(psf);
        addBGSF( bg );
        break;
      }
      case BOXSFIELD: {
          BoxSizeField * box = dynamic_cast<BoxSizeField*>(psf);
          addBoxSizeField(box);
          break;
      }
      default: {
        MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                    "unknown size field type %d", type);
      }
      }
    }  
  }

  // -------------------------------------------------------------------
  SizeFieldManager::~SizeFieldManager() {
    if (mainSF) delete mainSF;
  }

  // -------------------------------------------------------------------
  void SizeFieldManager::setMesh(pMesh m)
  {
    mesh = m;
    mainSF = new PWLSField(mesh);
    MAdMsgSgl::instance().warning(__LINE__,__FILE__,"No size field registered");
  }

  // -------------------------------------------------------------------
  void SizeFieldManager::intersect(const pSField sf) {
    mainSF->intersect(sf);
  }

  // -------------------------------------------------------------------
  void SizeFieldManager::addSizeField(pSField sf) {

    mainSF->intersect(sf);
  
    sizeFieldType type = sf->getType();
    switch (type) {
    case NULLSFIELD: {
      break;
    }
    case DISCRETESFIELD: {
      PWLSField* linsf = dynamic_cast<PWLSField*>(sf);
      addPWLinearSF( linsf );
      break;
    }
    case ANALYTICALSFIELD: {
      AnalyticalSField* asf = dynamic_cast<AnalyticalSField*>(sf);
      addAnalyticalSF( asf );
      break;
    }
    case LOCALSFIELD: {
      LocalSizeField* locsf = dynamic_cast<LocalSizeField*>(sf);
      addLocalSF( locsf );
      break;
    }
    case BACKGROUNDSFIELD: {
      BackgroundSF* bg = dynamic_cast<BackgroundSF*>(sf);
      addBGSF( bg );
      break;
    }
    case BOXSFIELD: {
        BoxSizeField * box = dynamic_cast<BoxSizeField*>(sf);
        addBoxSizeField(box);
        break;
    }
    default: {
      MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                  "Unknown size field type %d",type);
    }
    }
  }

  // -------------------------------------------------------------------
  void SizeFieldManager::addPWLinearSF(PWLSField* sf) {
    sf->registerInCallbackManager();
    linears.insert(sf);
  }
  void SizeFieldManager::addAnalyticalSF(AnalyticalSField* sf) {
    analyticals.insert(sf);
  }
  void SizeFieldManager::addLocalSF(LocalSizeField* sf) {
    locals.insert(sf);
  }
  void SizeFieldManager::addBGSF(BackgroundSF* sf) {
    bgs.insert(sf);
  }
  void SizeFieldManager::addBoxSizeField(BoxSizeField * box) {
      boxs.insert(box);
  }

  // -------------------------------------------------------------------
  void SizeFieldManager::update()
  {
      MeshQualityManagerSgl::instance().clearAllShapes(); // not useful in isotropic case (be careful about the hypothesis)

      mainSF->cleanUp();

      set<PWLSField*>::const_iterator linIter = linears.begin();
      set<PWLSField*>::const_iterator linLast = linears.end();
      for (; linIter != linLast; linIter++) {
          mainSF->intersect(*linIter);
      }

      set<AnalyticalSField*>::const_iterator aIter = analyticals.begin();
      set<AnalyticalSField*>::const_iterator aLast = analyticals.end();
      for (; aIter != aLast; aIter++) {
          mainSF->intersect(*aIter);
      }

      set<LocalSizeField*>::iterator locIter = locals.begin();
      set<LocalSizeField*>::iterator locLast = locals.end();
      for (; locIter != locLast; locIter++) {
          (*locIter)->updateTree();
          mainSF->intersect(*locIter);
      }

      set<BackgroundSF*>::iterator bgIter = bgs.begin();
      set<BackgroundSF*>::iterator bgLast = bgs.end();
      for (; bgIter != bgLast; bgIter++) {
          mainSF->intersect(*bgIter);
      }

      set<BoxSizeField*>::iterator boxIter = boxs.begin();
      set<BoxSizeField*>::iterator boxLast = boxs.end();
      for (; boxIter != boxLast; boxIter++) {
          mainSF->intersect(*boxIter);
      }

      if (smooth)
          mainSF->smooth(maxGradient);

      localTime = MAdTimeManagerSgl::instance().getTime();
  }

  // -------------------------------------------------------------------
  void SizeFieldManager::update(pVertex pv)
  {
      set<PWLSField*>::const_iterator linIter = linears.begin();
      set<PWLSField*>::const_iterator linLast = linears.end();
      for (; linIter != linLast; linIter++) {
          mainSF->intersect(pv, *linIter);
      }

      set<AnalyticalSField*>::const_iterator aIter = analyticals.begin();
      set<AnalyticalSField*>::const_iterator aLast = analyticals.end();
      for (; aIter != aLast; aIter++) {
          mainSF->intersect(pv, *aIter);
      }

      set<LocalSizeField*>::iterator locIter = locals.begin();
      set<LocalSizeField*>::iterator locLast = locals.end();
      for (; locIter != locLast; locIter++) {
          (*locIter)->updateTree();
          mainSF->intersect(pv, *locIter);
      }

      set<BackgroundSF*>::iterator bgIter = bgs.begin();
      set<BackgroundSF*>::iterator bgLast = bgs.end();
      for (; bgIter != bgLast; bgIter++) {
          mainSF->intersect(pv, *bgIter);
      }

      set<BoxSizeField*>::iterator boxIter = boxs.begin();
      set<BoxSizeField*>::iterator boxLast = boxs.end();
      for (; boxIter != boxLast; boxIter++) {
          mainSF->intersect(pv, *boxIter);
      }

      localTime = MAdTimeManagerSgl::instance().getTime();
  }

  // -------------------------------------------------------------------

}
