// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#include "DistanceToGEntity.h"
#include "MAdMessage.h"
#include "MAdOutput.h"
using std::set;
using namespace MAd;

// -------------------------------------------------------------------
double DistanceSpaceFunction (const void * data, double xyz[3])
{
  const DistanceToGEntity * dist = (const DistanceToGEntity *) data;
  return dist->getDistance(xyz);
}

// -------------------------------------------------------------------
namespace MAd {

  // -------------------------------------------------------------------
  void M_distanceToWalls(std::string outName, std::string mshName, bool toFaces)
  {
    pGModel model = 0;
    GM_create(&model,"theModel");
    GM_readFromMSH(model, mshName.c_str());
    pMesh mesh = M_new(model);
    M_load(mesh,mshName.c_str());
    M_distanceToWalls(outName, mesh, toFaces);
  }

  // -------------------------------------------------------------------
  void M_distanceToWalls(std::string outName, pMesh mesh, bool toFaces)
  {
    DistanceToGEntity D = DistanceToGEntity();
    if ( M_dim(mesh) == 3 ) {
      pGEntity pge;
      GFIter gfit = GM_faceIter( M_model(mesh) );
      while ( ( pge = (pGEntity)GFIter_next(gfit) ) ) { D.addGEntity(pge); }
      GFIter_delete(gfit);
    }
    if ( M_dim(mesh) == 2 ) {
      pGEntity pge;
      GEIter geit = GM_edgeIter( M_model(mesh) );
      while ( ( pge = (pGEntity)GEIter_next(geit) ) ) { D.addGEntity(pge); }
      GEIter_delete(geit);
    }
  
    D.compute(mesh, toFaces);

    D.printDistanceToAllVertices(outName, mesh);
    //D.posDistanceToAllVertices(outName, mesh);
  }

  // -------------------------------------------------------------------
  void DistanceToGEntity::compute(pMesh m, bool toFaces)
  {
    if ( distFct ) delete distFct;
    distFct = new distanceFunction(m, toFaces);

    set<pVertex> wallVertices;
    set<pEntity> wallEntities;
    set<pGEntity>::const_iterator it    = entities.begin();
    set<pGEntity>::const_iterator itEnd = entities.end();
    for (; it != itEnd; it++) 
      {
        switch ( GEN_type(*it) ) {
        case 0:
          {
            VIter vit = M_vertexIter(m);
            while (pVertex pv = VIter_next(vit))
              {
                if ( EN_whatIn((pEntity)pv) == *it ) {
                  wallEntities.insert((pEntity)pv);
                  wallVertices.insert(pv);
                }
              }
            VIter_delete(vit);
            break;
          }
        case 1:
          {
            EIter eit = M_edgeIter(m);
            while (pEdge pe = EIter_next(eit))
              {
                if ( EN_whatIn((pEntity)pe) == *it )
                  {
                    wallEntities.insert((pEntity)pe);
                    wallVertices.insert(E_vertex (pe,0));
                    wallVertices.insert(E_vertex (pe,1));
                  }
              }
            EIter_delete(eit);
            break;
          }
        case 2:
          {
            FIter fit = M_faceIter(m);
            while (pFace pf = FIter_next(fit))
              {
                if ( EN_whatIn((pEntity)pf) == *it )
                  {
                    wallEntities.insert((pEntity)pf);
                    wallVertices.insert(F_vertex (pf,0));
                    wallVertices.insert(F_vertex (pf,1));
                    wallVertices.insert(F_vertex (pf,2));
                  }
              }
            FIter_delete(fit);
            break;
          }
        case 3:
          {
            MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                        "Distance to entities of dimension 3 not implemented");
            break;
          }
        }
      }
  
    distFct->computeTree(wallVertices, wallEntities);
    treeIsComputed = true;
  }


  // -------------------------------------------------------------------
  double DistanceToGEntity::getDistance(double xyz[3]) const
  {
    if ( !treeIsComputed ) {
      MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                  "The tree has to be computed first");
    }
    return distFct->computeDistance(xyz);
  }

  // -------------------------------------------------------------------
  void DistanceToGEntity::printDistanceToAllVertices(std::string filename, pMesh mesh) const
  {
    if ( !treeIsComputed ) {
      MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                  "The tree has to be computed first");
    }

    FILE * f = fopen(filename.c_str(),"w");
    if ( !f ) {
      MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                  "Could not open file %s",filename.c_str());
    }

    pVertex pv;
    double xyz[3], d;
    VIter vit = M_vertexIter(mesh);
    while ( ( pv = VIter_next(vit) ) ) {
      V_coord(pv,xyz);
      d = distFct->computeDistance(xyz);
      fprintf(f, "%f %f %f %f\n", xyz[0], xyz[1], xyz[2], d);
    }
    VIter_delete(vit);

    fclose(f);
  }

  // -------------------------------------------------------------------
  void DistanceToGEntity::posDistanceToAllVertices(std::string filename, pMesh mesh) const
  {
    if ( !treeIsComputed ) {
      MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                  "The tree has to be computed first");
    }

    MAdGmshOutput (mesh, DistanceSpaceFunction, this, filename.c_str());
  }

  // -------------------------------------------------------------------

}
