// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#ifndef CGNSINTERFACE__H
#define CGNSINTERFACE__H

#include "MeshDataBaseInterface.h"
#include "cgnslib.h"

#include <map>
#include <vector>
#include <string>

namespace MAd {

  class IJKZone;

  // ---------------------------------------------------------------------------
  //! \brief data structure for an ijk-face connection
  // ---------------------------------------------------------------------------
  
  class IJKConnection {
    
  public: 
    
    IJKConnection();
    ~IJKConnection();
    
  public:
    
    bool periodic() const;
    bool computeTransformation(const std::map<std::string,IJKZone*>& zones);
    
    bool getCorner(int ic,int local[3],int donor[3]) const;
    void setLocalName(const char* zname) {
      std::memcpy(localName,zname,35*sizeof(char));
    }
    
  public:
    
    pTransform transformation;
    
    int      transformIndex[3];
    cgsize_t localRange[6];
    cgsize_t donorRange[6];
    char connectionName[35];
    char localName[35];
    char donorName[35];
    
  };

  // ---------------------------------------------------------------------------
  //!\brief data structure for an ijk-structured block
  // ---------------------------------------------------------------------------

  class IJKZone {
    
  public:

    IJKZone();
    ~IJKZone();

  protected:
    
    void clean();
    void initialize();

  public:
    
    //! write size in readable format 
    std::string printBlockSize() const;

    //! get number of nodes
    int getNbNodes() const;
    
    //! compute index in the coordinate block 
    int getNodeIndex(int i,int j,int k) const;
    
    //! get node coordinates 
    bool getNodeXYZ(int i,int j,int k,double xyz[3]) const;

    //! get node coordinates 
    bool getNodeXYZ(int ijk[3],double xyz[3]) const;
    
    //! compute ijk corresponding to node id
    void getNodeIJK(int id,int& i,int& j,int& k) const;
    
    //! get number of nodes
    int getNbCells() const;

    //! compute index in the cell block 
    int getCellIndex(int i,int j,int k) const;    
    
    //! compute ijk corresponding to node id
    void getCellIJK(int id,int& i,int& j,int& k) const;
    
    //! get minimum and maximum indices for data, return total number
    int getDataRanges(cgsize_t minIdx[3], //! minimum index per direction
                      cgsize_t maxIdx[3], //! maximum index per direction
                      GridLocation_t loc  //! data location, only cell- or vertex-centered data
                      ) const;
    
    //! clear the coordinates
    void clearCoordinates();
    
    
  public:
    
    int groupIndex;

    int fileIndex;
    int baseIndex;
    int zoneIndex;
    
    char name[35];     //!< zone name
    cgsize_t nbIJK[9]; //!< block sizes (0-2: nodes, 3-5: elements, 6-8: vertices
    double* coord[3];  //!< ijk-stored block

    double boundingBox[3][2];
    
    int startNodeId;
    int startCellId;

    pMeshDataId ijkTag;

    //! patch connections and periodic transformations
    std::vector<IJKConnection*> connections; 
 
  };


  // ---------------------------------------------------------------------------
  //! \brief load the coordinates of the block and update the bounding box
  // ---------------------------------------------------------------------------
  
  bool readCGNSIJKZoneCoordinates(IJKZone& zone,double scale,const pTransform tfo);
    
  // ---------------------------------------------------------------------------
  //!\brief load all ijk zones from a structured CGNS as well as the periodicities
  //! the affine transformation tfo is applied to coordinates and periodicities
  // ---------------------------------------------------------------------------
    
  std::vector<IJKZone*> 
  loadCGNSIJKZones(const std::string& fName,//!< cgns file name    
                   double& scale,           //!< coordinate scale
                   const pTransform tfo,    //!< transformation from CGNS to current
                   std::map<int,TransformationSet>& per //!< minimal set of periodicities
                   );
  
  // ---------------------------------------------------------------------------  
  //!\brief create a mesh for a given ijk block
  // ---------------------------------------------------------------------------
  
  bool createMeshForIJKZone(pMesh pm,            //!< CGNS mesh container
                            IJKZone& zone,       //!< ijk-zone to be converted
                            const pTransform tfo //!< aligns cgns with current mesh
                            );

  // ---------------------------------------------------------------------------

}
  
#endif
