// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#include "MAdTransformation.h"
#include "MeshDataBase.h"
#include "MeshDataBaseInterface.h"
#include "MeshDataBaseIO.h"
#include "CGNSInterface.h"
#include "MAdVector3D.h"
#include <cgnslib.h>

#include <sstream>
#include <fstream>

namespace MAd {

  // ---------------------------------------------------------------------------

  IJKConnection::IJKConnection():transformation(NULL) {
    for (size_t i=0;i<6;i++) localRange[i] = donorRange[i] = 0;
  };
  
  // ---------------------------------------------------------------------------

  IJKConnection::~IJKConnection() {
    if (transformation) T_delete(transformation);
  }
  
  // ---------------------------------------------------------------------------
  
  bool IJKConnection::periodic() const {return transformation != NULL;}

  // ---------------------------------------------------------------------------
  
  bool IJKConnection::getCorner(int ic,int ijkLocal[3],int ijkDonor[3]) const {
    
    if (ic < 0 || ic > 3) return false;

    int kLocal = -1;
    for (int i=0;i<3;i++) if (localRange[i]==localRange[i+3]) {kLocal = i;break;}
    
    // int kDonor = -1;
    // for (int i=0;i<3;i++) if (donorRange[i]==donorRange[i+3]) {kDonor = i;break;}
        
    int iLocal = (kLocal+1)%3;
    int jLocal = (kLocal+2)%3;

    int ip = ic / 2;
    int jp = ic % 2;
    
    
    integerVector local(3);
    
    ijkLocal[iLocal] = localRange[iLocal+ip*3]-1;
    ijkLocal[jLocal] = localRange[jLocal+jp*3]-1;
    ijkLocal[kLocal] = localRange[kLocal] - 1;

    for (int i=0;i<3;i++) local(i) = ijkLocal[i] + 1 - localRange[i];
    
    integerMatrix T(3,3);
    
    for (int i=0;i<3;i++) {
      int sgn = transformIndex[i] > 0 ? 1:-1;
      for (int j=0;j<3;j++) {
        T(j,i) = (abs(transformIndex[i]) == j+1) ? sgn:0;
      }
    }

    integerVector donor(3);
    donor = 0;

    T.mult(local,donor);

    int begin[3];
    for (int k=0;k<3;k++) {
      int iDonor = abs(transformIndex[k])-1;
      if (transformIndex[k] > 0) begin[iDonor] = donorRange[iDonor];
      else                       begin[iDonor] = donorRange[iDonor+3];  
    }
    
    for (int i=0;i<3;i++) ijkDonor[i] = donor(i) + begin[i] - 1;
    
    
    // int iLocal = abs(transformIndex[iDonor]) - 1;
    // int jLocal = abs(transformIndex[jDonor]) - 1;

    // int ipDonor = (ic/2);
    // int jpDonor = (ic%2);
    
    // int ipLocal = (transformIndex[iDonor] > 0) ? ipDonor : (ipDonor+1)%2;
    // int jpLocal = (transformIndex[jDonor] > 0) ? jpDonor : (jpDonor+1)%2;
    
    

    // ijkDonor[iDonor] = donorRange[iDonor+ipDonor*3]-1;
    // ijkDonor[jDonor] = donorRange[jDonor+jpDonor*3]-1;

    // ijkDonor[kDonor] = donorRange[kDonor]-1;
    return true;
  }

  // ---------------------------------------------------------------------------

  bool IJKConnection::computeTransformation(const std::map<std::string,IJKZone*>& ijkzones) {
    
    std::map<std::string,IJKZone*>::const_iterator lIter = ijkzones.find(localName);
    std::map<std::string,IJKZone*>::const_iterator dIter = ijkzones.find(donorName);
    
    if (dIter == ijkzones.end()) return false;
    if (lIter == ijkzones.end()) return false;

    IJKZone* donor = dIter->second;
    IJKZone* local = lIter->second;
    
    double donor0[3];
    double donor1[3];
    double donor2[3];
    double donor3[3];

    double local0[3];
    double local1[3];
    double local2[3];
    double local3[3];

    
    int ijkLocal[3];
    int ijkDonor[3];
    
    getCorner(0,ijkLocal,ijkDonor);
    local->getNodeXYZ(ijkLocal,local0);
    donor->getNodeXYZ(ijkDonor,donor0);
    
    getCorner(1,ijkLocal,ijkDonor);
    local->getNodeXYZ(ijkLocal,local1);
    donor->getNodeXYZ(ijkDonor,donor1);
        
    getCorner(2,ijkLocal,ijkDonor);
    local->getNodeXYZ(ijkLocal,local2);
    donor->getNodeXYZ(ijkDonor,donor2);
    
    getCorner(3,ijkLocal,ijkDonor);
    local->getNodeXYZ(ijkLocal,local3);
    donor->getNodeXYZ(ijkDonor,donor3);
    
    transformation = T_create(donor0,donor1,donor2,
                              local0,local1,local2,1e-6);
    
    if (transformation == NULL) return false;
    
    // if (T_isIdentity(transformation)) T_delete(transformation);
    
    return true;
  }


  // ---------------------------------------------------------------------------
  
  IJKZone::IJKZone() {initialize();}
  
  // ---------------------------------------------------------------------------

  IJKZone::~IJKZone() {clean();}

  // ---------------------------------------------------------------------------

  void IJKZone::clean() {
    for (int i=0;i<3;i++) if (coord[i]) delete [] coord[i];
    
    std::vector<IJKConnection*>::iterator cIter = connections.begin();
    for (;cIter!=connections.end();++cIter) delete *cIter;
    connections.end();

    initialize();
  }

  // ---------------------------------------------------------------------------

  void IJKZone::initialize() {
    fileIndex = -1;
    baseIndex = -1;
    zoneIndex = -1;
    coord[0] = coord[1] = coord[2] = NULL;
    for (size_t i=0;i<9;i++) nbIJK[i] = 0;

    startNodeId = 0;
    startCellId = 0;
  } 

  // ---------------------------------------------------------------------------

  std::string IJKZone::printBlockSize() const { 
    std::ostringstream sizeString;
    sizeString << nbIJK[0] << "x" << nbIJK[1] << "x" << nbIJK[2];
    return sizeString.str();
  }

  // ---------------------------------------------------------------------------
  
  int IJKZone::getNbNodes() const {
    return nbIJK[0]*nbIJK[1]*nbIJK[2];
  }
  
  // ---------------------------------------------------------------------------
  
  int IJKZone::getNodeIndex(int i,int j,int k) const {
    return k*nbIJK[0]*nbIJK[1] + j*nbIJK[0] + i;
  }
  // ---------------------------------------------------------------------------
    
  void IJKZone::getNodeIJK(int idx,int&i,int&j,int&k) const {
    
    int tmpIdx = idx;

    i = tmpIdx % (nbIJK[0]);
    tmpIdx -= i;

    j = (tmpIdx % (nbIJK[0]*nbIJK[1])) / nbIJK[0];
    tmpIdx -= j * nbIJK[0];

    k = tmpIdx / (nbIJK[0]*nbIJK[1]);

    if (idx != getNodeIndex(i,j,k)) throw;
  }
  
  // ---------------------------------------------------------------------------
  
  bool IJKZone::getNodeXYZ(int i,int j,int k,double xyz[3]) const {
    int id = getNodeIndex(i,j,k);
    if (id > getNbNodes()) return false;
    for (int d=0;d<3;d++) xyz[d] = coord[d][id];
    return true;
  }
  
  // ---------------------------------------------------------------------------
  
  bool IJKZone::getNodeXYZ(int ijk[3],double xyz[3]) const {
    return getNodeXYZ(ijk[0],ijk[1],ijk[2],xyz);
  }
  
  // ---------------------------------------------------------------------------

  int IJKZone::getNbCells() const {
    return (nbIJK[0]-1)*(nbIJK[1]-1)*(nbIJK[2]-1);
  }
  
  // ---------------------------------------------------------------------------
  
  int IJKZone::getCellIndex(int i,int j,int k) const {
    return k*(nbIJK[0]-1)*(nbIJK[1]-1) + j*(nbIJK[0]-1) + i;
  }
  
  // ---------------------------------------------------------------------------
  
  void IJKZone::getCellIJK(int idx,int&i,int&j,int&k) const {
    
    int tmpIdx = idx;

    i = tmpIdx % (nbIJK[0]-1);
    tmpIdx -= i;

    j = (tmpIdx % ((nbIJK[0]-1)*(nbIJK[1]-1))) / (nbIJK[0]-1);
    tmpIdx -= j * (nbIJK[0]-1);

    k = tmpIdx / ((nbIJK[0]-1)*(nbIJK[1]-1));  

    if (idx != getCellIndex(i,j,k)) throw;

  }
  // ---------------------------------------------------------------------------
  
  int IJKZone::getDataRanges(cgsize_t minIdx[3],
                             cgsize_t maxIdx[3],
                             GridLocation_t location) const {
    
    minIdx[0] = 1;
    minIdx[1] = 1;
    minIdx[2] = 1;
    
    switch(location) {
    case CellCenter:
      for (int d=0;d<3;d++) maxIdx[d] = nbIJK[d] - 1;
      break;
    case Vertex:
      for (int d=0;d<3;d++) maxIdx[d] = nbIJK[d];
      break;
    default:
      throw;
    }
    
    int nbTotal = 1;
    for (size_t d=0;d<3;d++) nbTotal *= (maxIdx[d] > 0) ? maxIdx[d]:1;
    return nbTotal;
  }
  
  // ---------------------------------------------------------------------------
  
  void IJKZone::clearCoordinates() {
    for (int d=0;d<3;d++) {
      if (coord[d]) {
        delete [] coord[d];
        coord[d] = NULL;
      }
    }
  }
  
  // ---------------------------------------------------------------------------
  
  std::vector<IJKZone*> clearZones(std::vector<IJKZone*>& zones) {
    std::vector<IJKZone*>::iterator zIter;
    for (zIter=zones.begin();zIter!=zones.end();++zIter) delete *zIter;
    zones.clear();
    return zones;
  }

  // ---------------------------------------------------------------------------
  
  bool readCGNSIJKZoneCoordinates(IJKZone& zone,double scale,const pTransform tfo) {
    
    int fileIndex = zone.fileIndex;
    int baseIndex = zone.baseIndex;
    int zoneIndex = zone.zoneIndex;
    
    int nbCoords;
    cg_ncoords(fileIndex,baseIndex,zoneIndex,&nbCoords);
    
    cgsize_t minIndex[3];
    cgsize_t maxIndex[3];
    int nbNodes = zone.getDataRanges(minIndex,maxIndex,Vertex);
    
    for (int iCoord=0;iCoord<nbCoords;iCoord++) {
      
      if (zone.coord[iCoord]) delete [] zone.coord[iCoord];
      double* dCoord = new double[nbNodes];
      zone.coord[iCoord] = dCoord;
      
      DataType_t dataType;
      char coordName[35];
      if (cg_coord_info(fileIndex,baseIndex,zoneIndex,
                        iCoord+1,&dataType,coordName)) return false;
      
      bool check = true;

      switch(dataType) {
      case RealSingle:
        {
          if (check) {
            MAdMsgSgl::instance().warning(__LINE__,__FILE__,
                                          "Single precision coordinates in zone %i",
                                          zoneIndex);
          }
          check = false;
          float* fCoord = new float[nbNodes];
          if (cg_coord_read(fileIndex,
                            baseIndex,
                            zoneIndex,
                            coordName,
                            dataType,
                            minIndex,
                            maxIndex,
                            fCoord)) return false;
          for (int i=0;i<nbNodes;i++) dCoord[i] = fCoord[i]*scale;
          delete [] fCoord;
          break;
        }
      case RealDouble:
        {
          if (cg_coord_read(fileIndex,
                            baseIndex,
                            zoneIndex,
                            coordName,
                            dataType,
                            minIndex,
                            maxIndex,
                            dCoord)) return false;
          if (scale!=1) for (int i=0;i<nbNodes;i++) dCoord[i] *= scale;
          break;
        }
      default:
        {
          return false;
          break;
        }
      }
    }
    
    // transform if necessary

    if (!T_isIdentity(tfo)) {
        
      double* px = zone.coord[0];
      double* py = zone.coord[1];
      double* pz = zone.coord[2];
        
      for (int i=0;i<nbNodes;i++) {
          
        double xyz[3] = {*px,*py,*pz};
          
        T_transformCoordinates(tfo,xyz);
          
        *(px++) = xyz[0];
        *(py++) = xyz[1];
        *(pz++) = xyz[2];
      }
    }
    
    // --- update the bounding box

    for (size_t d=0;d<3;d++) {
      zone.boundingBox[d][0] =   std::numeric_limits<double>::max();
      zone.boundingBox[d][1] = - std::numeric_limits<double>::max();
      const double* coord = zone.coord[d];
      for (int i=0;i<nbNodes;i++) {
        zone.boundingBox[d][0] = std::min(zone.boundingBox[d][0],coord[i]);
        zone.boundingBox[d][1] = std::max(zone.boundingBox[d][1],coord[i]);
      }
    }
    
    return true;
  }

  // ---------------------------------------------------------------------------
  
  std::vector<IJKZone*> loadCGNSIJKZones(const std::string& fname,
                                         double& scale,
                                         const pTransform fwdTfo,
                                         std::map<int,TransformationSet>& periodicities) {
    
    bool master = true;
#ifdef PARALLEL
    int rank;
    MPI_Comm_rank(MPI_COMM_WORLD,&rank);
    master = rank == 0;
#endif

    std::vector<IJKZone*> zones;
    const MAdMsg& msg = MAdMsgSgl::instance();
    pTransform bckTfo = T_inverse(fwdTfo);
    
    char* name = new char[fname.size()+1];
    strcpy(name,fname.c_str());
    
    // --- open file -----------------------------------------------------------

    int fileIndex;


    if (cg_open(name,CG_MODE_READ,&fileIndex)) {
      
      if (master) msg.warning(__LINE__,__FILE__,"Error opening cgns file %s : %s",
                  name,cg_get_error());
      
      int fileType;
      if (cg_is_cgns(name,&fileType)) {
        if (master) msg.warning(__LINE__,__FILE__,"File %s is not a cgns file",name);
      }
      else {
        if (master) msg.warning(__LINE__,__FILE__,"File %s is in %s format",
                    name,(fileType == CG_FILE_ADF) ? "ADF" : "HDF5");
      }
      
      msg.error(__LINE__,__FILE__,"Could not open CGNS file %s",name); 
      return clearZones(zones);
    }

    int nbBases;
    int baseIndex(1);
    cg_nbases(fileIndex,&nbBases);
    
    // --- read length unit ---------------------------------------------------- 
    scale = 1;

    MassUnits_t massUnits;
    LengthUnits_t lengthUnits;
    TimeUnits_t timeUnits;
    TemperatureUnits_t temperatureUnits;
    AngleUnits_t angleUnits;
    
    cg_units_read(&massUnits,&lengthUnits,&timeUnits,&temperatureUnits,&angleUnits);

    switch (lengthUnits) {

    case Centimeter:
      if (master) msg.info(__LINE__,__FILE__,"CGNS database in cm");
      scale = 0.01;
      break;
    case Millimeter:
      if (master) msg.info(__LINE__,__FILE__,"CGNS database in mm");
      scale = 0.001;
      break;
    case Foot:
      if (master) msg.info(__LINE__,__FILE__,"CGNS database in feet");
      scale = 0.3048;
      break;
    case Inch:
      if (master) msg.info(__LINE__,__FILE__,"CGNS database in inch");
      scale = 0.0254;
      break;
    case Meter:
      if (master) msg.info(__LINE__,__FILE__,"CGNS database in meter");
      scale = 1;
      break;
    case CG_Null:
    case CG_UserDefined:
      if (master) msg.info(__LINE__,__FILE__,"CGNS database length scale undefined");
      scale = 1;
      break;
    }
    
    // --- read the different zones in the mesh --------------------------------

    int nbZones;
    cg_nzones(fileIndex,baseIndex,&nbZones);

    int startNodeId = 1;
    int startCellId = 1;
    
    for (int zoneIndex=1;zoneIndex<=nbZones;zoneIndex++) {
      
      IJKZone* zone = new IJKZone();
      zones.push_back(zone);
      
      zone->fileIndex = fileIndex;
      zone->baseIndex = baseIndex;
      zone->zoneIndex = zoneIndex;
      zone->groupIndex = zoneIndex;
      
      zone->startNodeId = startNodeId;
      zone->startCellId = startCellId;
      
      cg_zone_read(fileIndex,baseIndex,zoneIndex,zone->name,zone->nbIJK);
      
      // read coordinates and update bounding box
      if (!readCGNSIJKZoneCoordinates(*zone,scale,fwdTfo)) {
        msg.error(__LINE__,__FILE__,"Could not read coordinates",
                  zoneIndex,name); 
        cg_close(fileIndex);
        clearZones(zones);
      } 
      
      // --- read connectivity 

      int nbConnections;
      cg_n1to1(fileIndex,baseIndex,zoneIndex,&nbConnections);


      for (int connIndex=1;connIndex<=nbConnections;connIndex++) {

        IJKConnection* connection = new IJKConnection();
        zone->connections.push_back(connection);
        
        connection->setLocalName(zone->name);
        
        // --- read connected blocks

        cg_1to1_read(fileIndex,
                     baseIndex,
                     zoneIndex,
                     connIndex,
                     connection->connectionName,
                     connection->donorName,
                     connection->localRange,
                     connection->donorRange,
                     connection->transformIndex);
        
        // --- read periodic transformation 

        float center[3];
        float angle[3];
        float translation[3];

        if (!cg_1to1_periodic_read(fileIndex, 
                                   baseIndex, 
                                   zoneIndex, 
                                   connIndex,
                                   center, 
                                   angle, 
                                   translation)) {          
          
          if (pTransform pt = T_create(center,angle,translation)) {
            
            if (fwdTfo) {
              
              pTransform t1 = T_combine(bckTfo,pt);              
              pTransform t2 = T_combine(t1,fwdTfo);
              
              T_delete(pt);
              T_delete(t1);
              
              pt = t2;
            }
            connection->transformation = pt;
          }
        } 
      }
      
      startNodeId += zone->getNbNodes();
      // startCellId += zone->getNbCells(); 
    }
    
    
    // --- construct groups of connected blocks --------------------------------

    std::map<std::string,IJKZone*> nameToZone;

    std::vector<IJKZone*>::iterator zIter = zones.begin();
    for (zIter=zones.begin();zIter!=zones.end();++zIter) {
      nameToZone[(*zIter)->name] = *zIter;
    }

    // --- attempt to reconstruct the missing transformations ------------------

    
    // for (zIter=zones.begin();zIter!=zones.end();++zIter) {
    //   IJKZone* zone = *zIter;
    //   std::vector<IJKConnection*>::iterator cIter = zone->connections.begin();
    //   for (;cIter!=zone->connections.end();++cIter) {
    //     IJKConnection* conn = *cIter;
    //     if (conn->transformation == NULL)  conn->computeTransformation(nameToZone);
    //   }
    // }
    
    
    // --- construct groups of connected blocks --------------------------------
    
    bool groupCheck = true;
    int  groupIter = 0;
    while (groupCheck && groupIter < 10) {
      groupCheck = false;
      groupIter++;

      for (zIter=zones.begin();zIter!=zones.end();++zIter) {
        
        IJKZone* zone1 = *zIter;
        
        std::vector<IJKConnection*>::iterator cIter = zone1->connections.begin();
        for (;cIter!=zone1->connections.end();++cIter) {
          
          IJKConnection* conn = *cIter;
          std::string donorName = conn->donorName;
          
          std::map<std::string,IJKZone*>::iterator  nIter = nameToZone.find(donorName);
          if (nIter == nameToZone.end()) {
            if (master) msg.warning(__LINE__,__FILE__,"Could not find zone %s",donorName.c_str());
          }
          
          IJKZone* zone2 = nIter->second;
          
          
          int gIdx1 = zone1->groupIndex;
          int gIdx2 = zone2->groupIndex;
          
          if (gIdx1 != gIdx2) {
            groupCheck = true;
            int gIdx = std::min(gIdx1,gIdx2);
            zone1->groupIndex = gIdx;
            zone2->groupIndex = gIdx;
          }
        }
      }
    }

    if (groupCheck) {
      msg.warning(__LINE__,__FILE__,
                  "Error grouping all contiguous zones");
    }

    // --- construct the list of periodicities per group
    
    for (zIter=zones.begin();zIter!=zones.end();++zIter) {

      IJKZone* zone = *zIter;
      
      std::vector<IJKConnection*>::iterator cIter = zone->connections.begin();
      for (;cIter!=zone->connections.end();++cIter) {
        if ((*cIter)->transformation) {
          periodicities[zone->groupIndex].insert(T_duplicate((*cIter)->transformation));
        }
      }
    }
    
    // --- clear the coordinates
    
    for (zIter=zones.begin();zIter!=zones.end();++zIter) {
      IJKZone* zone = *zIter;
      zone->clearCoordinates();
    }
  
    delete [] name;
  
    return zones;
  }

  // ---------------------------------------------------------------------------
  
  bool createMeshForIJKZone(pMesh pm,IJKZone& zone,const pTransform tfo) {
    
    double scale = 1;
    
    if (!pm->model) {
      pGModel model = NULL;
      GM_create(&model,"");
      pm->model = model; 
    }

    pGEntity geom = (pGEntity) GM_regionByTag(pm->model,zone.zoneIndex);

    zone.ijkTag = MD_lookupMeshDataId("CGNSIJK");
    
    int startCellId = zone.startCellId;
    int startNodeId = zone.startNodeId;
      
    int nI = zone.nbIJK[0];
    int nJ = zone.nbIJK[1];
    int nK = zone.nbIJK[2];
    
    readCGNSIJKZoneCoordinates(zone,scale,tfo);
    
    double *px = zone.coord[0];
    double *py = zone.coord[1];
    double *pz = zone.coord[2];
    
    for (int i=0;i<nI;i++) {
      for (int j=0;j<nJ;j++) {
        for (int k=0;k<nK;k++) {
          int nodeIdx = zone.getNodeIndex(i,j,k);
          pm->add_point(nodeIdx+startNodeId,px[nodeIdx],py[nodeIdx],pz[nodeIdx]);
        }
      }
    }

    zone.clearCoordinates();
    
    for (int k=0;k<nK-1;k++) {
      for (int j=0;j<nJ-1;j++) {
        for (int i=0;i<nI-1;i++) {
          
          int elemIdx = zone.getCellIndex(i,j,k);
          
          pEntity pe = pm->add_hex(startNodeId + zone.getNodeIndex(i  ,j  ,k  ),
                                   startNodeId + zone.getNodeIndex(i+1,j  ,k  ),
                                   startNodeId + zone.getNodeIndex(i+1,j+1,k  ),
                                   startNodeId + zone.getNodeIndex(i  ,j+1,k  ),
                                   startNodeId + zone.getNodeIndex(i  ,j  ,k+1),
                                   startNodeId + zone.getNodeIndex(i+1,j  ,k+1),
                                   startNodeId + zone.getNodeIndex(i+1,j+1,k+1),
                                   startNodeId + zone.getNodeIndex(i  ,j+1,k+1),geom);
          
          if (pe) {
            EN_setID(pe,elemIdx + startCellId);
            EN_attachDataInt(pe,zone.ijkTag,elemIdx);
          }
        }
      }
    }
    
    pm->classify_unclassified_entities();
    pm->destroyStandAloneEntities();
    
    return true;
  }
  
  // ---------------------------------------------------------------------------
}
