// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#include "MAdMatrix.h"

#include <cstdio>


// --- blas functions used in MAdLib ---------------------------------

extern "C" {
  
  // -- matrix vector product 
  void dgemv_(char* tran,const int& m,const int& n,const double& alpha,const double* A,
              const int& tda,const double* x,const int& incx,const double& b,
              double* y,const int& incy);
  
  void sgemv_(char* tran,const int& m,const int& n,const float& alpha,const float* A,
              const int& tda,const float* x,const int& incx,const float& b,
              float* y,const int& incy);
  
  // -- matrix-matrix product
  void dgemm_(char* ta, char* tb,const int& M, const int& N, const int& K, 
              const double& ALPHA, 
              const double* A, const int& LDA, 
              const double* B, const int& LDB,
              const double& BETA, 
              double* C, const int& LDC );
  
  void sgemm_(char* ta, char* tb,const int& M, const int& N, const int& K, 
              const float& ALPHA, 
              const float* A, const int& LDA, 
              const float* B, const int& LDB,
              const float& BETA, 
              float* C, const int& LDC );
  
}

// --- LAPACK functions used in MAdLib -----------------------------------------

extern "C" {
  
  // LU decomposition and back/forward substitution

  void dgesv_(const int*,const int*,double*,const int*,int*,double*,const int*,int*);

  void sgesv_(const int*,const int*,float*, const int*,int*,float*,const int*,int*);
}


// --- templated specialisations using BLAS/LAPACK -----------------------------

namespace MAd {
  
  // --- Matrix inversion ------------------------------------------------------
  
  template<class SCALAR,void (*GESV) (const int*,
                                      const int*,
                                      SCALAR*,
                                      const int*,
                                      int*,
                                      SCALAR*,
                                      const int*,
                                      int*)> 
  bool invertLapack(const MAd_Matrix<SCALAR>& A,MAd_Matrix<SCALAR>& AInv) {
    
    if (A.size1() != A.size2()) return false;

    int nI = A.size1();
    
    AInv.resize(nI,nI);
    MAd_Matrix<SCALAR> inv(nI,nI);
    

    SCALAR* work = new SCALAR[nI*nI];
    SCALAR* colW = work;
    
    const SCALAR* colA = A.seePtr();
    for (int i=0;i<A.size1();i++) {
      std::memcpy(colW,colA,nI*sizeof(SCALAR));
      colW += nI;
      colA += A.ld();
    }
    
    SCALAR* pInv = AInv.touchPtr();
    int* pivot  = new int[nI];
    int info;
  
    for (int i=0;i<nI*nI;i++) pInv[i] = 0;
    for (int i=0;i<nI;i++)    pInv[i*nI + i] = 1;
  
    GESV(&nI,&nI,work,&nI,pivot,pInv,&nI,&info);
    
    delete [] pivot;
    delete [] work;
    return (info == 0);
    
  }

  // specialisation for double precision 

  template<> MAd_Matrix<double> MAd_Matrix<double>::inverse() const {
    
    MAd_Matrix<double> inv(size1(),size1());
    if (!invertLapack<double,dgesv_>(*this,inv)) {
      MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                  "Could not invert matrix of size %i x %i",
                                  size1(),size2());
    }
    return inv;
  }

  // specialisation for single precision 
  
  template<> MAd_Matrix<float> MAd_Matrix<float>::inverse() const {

    MAd_Matrix<float> inv(size1(),size1());
    if (!invertLapack<float,sgesv_>(*this,inv)) {
      MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                  "Could not invert matrix of size %i x %i",
                                  size1(),size2());
    }
    return inv;
  }
  
  // dummy specialisation for integers

  template<> MAd_Matrix<int> MAd_Matrix<int>::inverse() const {
    
    MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                "Can not invert integer matrix");
    return MAd_Matrix<int>();
  }

  
  template<> bool MAd_Matrix<double>::invertInPlace() {
    
    MAd_Matrix<double> inv(size1(),size1());
    if (!invertLapack<double,dgesv_>(*this,inv)) return false;
    *this = inv;
    return true;
  }

  // specialisation for single precision 
  
  template<> bool MAd_Matrix<float>::invertInPlace() {
    
    MAd_Matrix<float> inv(size1(),size2());
    if (!invertLapack<float,sgesv_>(*this,inv)) return false;
    *this = inv;
    return true;
  }
  
  // dummy specialisation for integers

  template<> bool MAd_Matrix<int>::invertInPlace() {
    MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                "Can not invert integer matrix");
    return false;
  }
  

  
};
