// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#ifndef _H_MATRIX_MAD
#define _H_MATRIX_MAD

#include "MAdMessage.h"
#include "MAdVector.h"

#include <cassert>
#include <cmath>
#include <limits>
#include <iostream>
#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <algorithm>
#include <iomanip>

// -------------------------------------------------------------------

namespace MAd {
  
  /* ===========================================================================*/
  /*! \brief View on dense matrix                                               */
  /* ===========================================================================*/

  template <class SCALAR> class mad_matrix_view {
  
  public:
  
    ~mad_matrix_view() {}
  
    mad_matrix_view(SCALAR* b,const SCALAR* e,size_t n1,size_t n2,size_t l):
      data(b),end(e),size1(n1),size2(n2),ld(l) {}
  
    SCALAR* const data;
    const SCALAR* end;
  
    size_t size1;
    size_t size2;
    size_t ld;
  };

  /* ===========================================================================*/
  /*! \brief View on constant dense matrix                                      */
  /* ===========================================================================*/

  template <class SCALAR> class mad_const_matrix_view {

  public:

    mad_const_matrix_view(const SCALAR* b,const SCALAR* e,size_t n1,size_t n2,size_t l):
      data(b),end(e),size1(n1),size2(n2),ld(l) {}

    ~mad_const_matrix_view() {}

    const SCALAR* const data;
    const SCALAR* const end;
    size_t size1;
    size_t size2;
    size_t ld;
  };

  // -------------------------------------------------------------------
  /*! \brief dense matrix class stored in column-major format */
  /*! \todo remove ld -> will only treat contiguous matrices */
  template <class SCALAR>
  class MAd_Matrix
  {
  private:
    int _r;               //!< number of rows
    int _c;               //!< number of columns
    int _ld;              //!< leading dimension = number of rows if original
    SCALAR *_data;        //!< modifiable handle contents 
    const SCALAR* _cData; //!< constant handle to contents 
    const SCALAR* _cEnd;  //!< constant pointer to 1 beyond the matrix
    bool _owner;          //!< false if this is a proxy
    
  public:

    MAd_Matrix();
    MAd_Matrix(int R,int C);
    MAd_Matrix(const MAd_Matrix& o);
    MAd_Matrix(SCALAR* d,int R,int C);
    
    ~MAd_Matrix();

  public:

    void resize(int R,int C);
    void allocate(int R,int C);

  public:
    
    inline int size1() const { return _r; }
    inline int size2() const { return _c; }
    inline int ld() const {return _ld;}

    inline const SCALAR* seePtr() const {return _cData;}
    inline       SCALAR* touchPtr() const {return _data;}

    inline SCALAR operator () (int i, int j) const
    {
      return _data[i + _ld * j];
    }
    inline SCALAR & operator () (int i, int j)
    {
      return _data[i + _ld * j];
    }
    
    inline SCALAR * operator [] (const size_t i)      
    {
      return _data + i*_c;
    }
    
    inline const SCALAR * operator [] (const size_t i) const    
    {
      return _data + i*_c;
    }
    
    MAd_Matrix & operator = (const MAd_Matrix &other)
    {
      if(this != &other){
        resize(other._r,other._c);
        memcpy(other);
      }
      return *this;
    }

    MAd_Matrix & operator += (const MAd_Matrix &other)
    {
      this->add(other);
      return *this;
    }

    MAd_Matrix & operator + (const MAd_Matrix &other) const
    {
      MAd_Matrix tmp(_r,_c);
      tmp = *this;
      tmp.add(other);
      return tmp;
    }

    MAd_Matrix & operator = (SCALAR a) 
    {
      set_all(a);
      return *this;
    }

    MAd_Matrix & operator *= (SCALAR a) 
    {
      scale(a);
      return *this;
    }

    MAd_Matrix & operator /= (SCALAR a) 
    {
      scale(1./a);
      return *this;
    }

    bool operator== (const MAd_Matrix& other) const {
      
      if (_r != other._r) return false;
      if (_c != other._c) return false;
      for (int j=0;j<_c;j++) {
        for (int i=0;i<_r;i++) {
          if ((*this)(i,j) != other(i,j)) return false;
        }
      }
      return true;
    }  

    bool isEqual(const MAd_Matrix& other,double tol=0) const {
      if (_r != other._r) return false;
      if (_c != other._c) return false;
      for (int j=0;j<_c;j++) {
        for (int i=0;i<_r;i++) {
          if (fabs((*this)(i,j)-other(i,j)) > tol) return false;
        }
      }
      return true;
    }
    
    void memcpy(const MAd_Matrix &other)
    {
      for (int j=0;j<_c;j++) for (int i=0;i<_r;i++) (*this)(i,j) = other(i,j);
    }

    void copy(const MAd_Matrix &a, int i0, int ni, int j0, int nj, 
              int desti0, int destj0)
    {
      for(int i = i0, desti = desti0; i < i0 + ni; i++, desti++)
        for(int j = j0, destj = destj0; j < j0 + nj; j++, destj++)
          (*this)(desti, destj) = a(i, j);
    }

    //!\brief \f$ c = c + data * b \f$
    inline void mult(const MAd_Matrix &b, MAd_Matrix &c) const
    {
      //      c.scale(0.);
      for(int i = 0; i < _r; i++)
        for(int j = 0; j < b.size2(); j++)
          for(int k = 0; k < _c; k++)
            c(i,j) += (*this)(i, k) * b(k, j);
    }

    //!\brief \f$ y = c_y y + c_x A * x \f$
    inline void mult(const MAd_Vector<SCALAR> &x, 
                     MAd_Vector<SCALAR> &y, 
                     bool t=false, 
                     const SCALAR c_x=1.0, 
                     const SCALAR c_y=1.0) const
    {
      
      if(!t){
        for(int i=0;i<_r;i++){
          y(i)=c_y*y(i);
          for(int j=0;j<_c;j++) y(i)+=c_x*(*this)(i,j) * x(j);
        }
      }
      else{
        for(int i=0;i<_r;i++){
          y(i)=c_y*y(i);
          for(int j=0;j<_c;j++) y(i)+=c_x*(*this)(j,i) * x(j);
        }
      }
    }



    // data = alpha * ( a * b ) + beta * data
    inline void blas_dgemm(const MAd_Matrix &a, const MAd_Matrix &b, 
                           SCALAR alpha=1., SCALAR beta=1.)
    {
      MAd_Matrix temp(a.size1(), b.size2()); // temp = 0;
      a.mult(b, temp); // temp = a * b
      temp.scale(alpha); // temp = alpha * ( a * b )
      scale(beta);
      add(temp);
    }

    inline MAd_Matrix blas_dgemm_transpose(MAd_Matrix &a, MAd_Matrix &b)
    {
      MAd_Matrix temp(a.size1(), b.size2());
      a.mult(b,temp);
      MAd_Matrix T(temp.size2(), temp.size1());
      for(int i = 0; i < temp.size1(); i++)
        for(int j = 0; j < temp.size2(); j++)
          T(j, i) = temp(i, j);
      return T;
    }

    inline void set_all(const SCALAR &m) 
    {
      for (int j=0;j<_c;j++) for (int i=0;i<_r;i++) (*this)(i,j) = m;
    }
    
    inline void set_identity() {
      set_all(0);
      for (int i=0;i<std::min(_r,_c);i++) (*this)(i,i) = 1;
    }
    
    inline void setValues(const SCALAR *M[]) 
    {
      for (int i=0;i<_r;i++) for (int j=0;j<_c;j++) (*this)(i,j) = M[i][j];
    }
    inline void scale(const SCALAR s)
    {
      if(s == 0.)
        for (int j=0;j<_c;j++) for (int i=0;i<_r;i++) (*this)(i,j) = 0;
      else
        for (int j=0;j<_c;j++) for (int i=0;i<_r;i++) (*this)(i,j)*= s;
    }
    inline void add(const SCALAR &a) 
    {
      for (int j=0;j<_c;j++) for (int i=0;i<_r;i++) (*this)(i,j) +=a;
    }
    inline void add(const MAd_Matrix &m) 
    {
      for (int j=0;j<_c;j++) for (int i=0;i<_r;i++) (*this)(i,j) += m(i, j);
    }
    
    inline MAd_Matrix transpose()
    {
      MAd_Matrix T(size2(),size1());
      for (int j=0;j<_c;j++) for (int i=0;i<_r;i++) T(j, i) = (*this)(i, j);
      return T;
    }
    
    inline SCALAR trace()
    {
      SCALAR t(0);
      for (int j=0;j<std::min(_c,_r);j++) t+= (*this)(j, j);
      return t;
    }
    
    SCALAR frobenius()
    {
      MAd_Matrix temp(size2(),size1());
      MAd_Matrix Strans = transpose();
      
      Strans.mult(*this,temp);
      
      return sqrt(temp.trace());
    }
    
    MAd_Matrix inverse() const;

    bool invertInPlace();
    
    inline SCALAR determinant() {
      
      if (size1() != size2()) {
        MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                    "Determinant computation requires square matrix");
      }
      
      SCALAR det(0);
      switch(size1()) {
      case 1:
        det = _data[0];
        break;
      case 2:
        det = _data[0]*_data[3] - _data[1]*_data[2];
        break;
      default:
        SCALAR fact = 1;
        for (int i=0;i<size1();i++,fact*=-1) {
          det += _data[i] * cofactor(i,0).determinant()*fact;
        }
        break;
      }
      return det;
    }
    
    MAd_Matrix cofactor(int i, int j) const 
    {
      int ni = size1();
      int nj = size2();
    
      MAd_Matrix cof(ni - 1, nj - 1);
      for(int I = 0; I < ni; I++){
        for(int J = 0; J < nj; J++){
          if(J != j && I != i)
            cof(I < i ? I : I - 1, J < j ? J : J - 1) = (*this)(I, J);
        }
      }
      return cof;
    }
    
    bool eig(MAd_Matrix &VL, // left eigenvectors 
             MAd_Vector<SCALAR> &DR, // Real part of eigenvalues
             MAd_Vector<SCALAR> &DI, // Im part of eigenvalues
             MAd_Matrix &VR,
             bool sortRealPart=false) {
      MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                  "Eigen value decomposition not yet implemented");
      return false;
    } 
    
    void print(std::string name="") const
    {
      printf("Matrix %s (%d, %d):\n",name.c_str(),_r,_c);
      for(int i = 0; i < _r ; ++i) {
        for(int j = 0; j < _c ; ++j) printf("%12.5E ",(*this)(i, j));
        printf("\n");
      }
    }

    bool print(std::ostream& out) const {
      
      if (_r*_c==0) return false;
      if (out.bad()) return false;

      out << "Matrix " << _r << " x " << _c << "\n";
      
      for(int i=0;i<_r;++i) {
        for(int j=0;j<_c;++j) out << std::setw(6) << (*this)(i,j) << " ";
        out << "\n";
      }
      return true;
    }


  public :
    // proxies integration
    /*!\brief proxy constructor */
    MAd_Matrix(mad_matrix_view<SCALAR> view) : 
      _r(view.size1),_c(view.size2),_ld(view.ld),
      _data(view.data),_cData(view.data),_cEnd(view.end) {}

    /*!\brief const proxy constructor */ 
    MAd_Matrix(mad_const_matrix_view<SCALAR> view) : 
      _r(view.size1),_c(view.size2),_ld(view.ld),
      _data(NULL),_cData(view.data),_cEnd(view.end){}
    //
    /*!\brief pointer to modifiable data */
    inline       SCALAR* data()       {return _data;}
    /*!\brief pointer to constant data */
    inline const SCALAR* data() const {return _cData;}
  
  public:

    /*!\brief size in bytes of the basic data type */
    static inline size_t dataTypeSize() {return sizeof(SCALAR);}

  public:

    /*! \brief column vector proxy */
    inline mad_vector_view<SCALAR> touchColumn(int i)  {
      return mad_vector_view<SCALAR>(_data + i*_ld,_cEnd,size1(),1);
    }
  
    /*! \brief column vector proxy */
    inline mad_const_vector_view<SCALAR> seeColumn(int i) const {
      return mad_const_vector_view<SCALAR>(_cData + i*_ld,_cEnd,size1(),1);
    }

    /*! \brief row vector proxy */
    inline mad_vector_view<SCALAR> touchRow(int i) {
      return mad_vector_view<SCALAR>(_data + i,_cEnd,size2(),_ld);
    }

    /*! \brief row vector proxy */
    inline mad_const_vector_view<SCALAR> seeRow(int i) const {
      return mad_const_vector_view<SCALAR>(_cData + i,_cEnd,size2(),_ld);
    }

  };
  
  // -------------------------------------------------------------------
  
  typedef MAd_Matrix<double> doubleMatrix;
  typedef MAd_Matrix<int> integerMatrix;

  // -------------------------------------------------------------------
  
  typedef MAd_Matrix<double> smallMatrix;
  typedef MAd_Vector<double> smallVector;

  // -------------------------------------------------------------------

}

#include "MAdMatrix.ipp"

#endif
