// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#include "MAdMessage.h"
#include "MAdConfig.h"

#include <cstdarg>
#include <iostream>
#include <sstream>
#include <stdlib.h>
#include <stdio.h>

using std::stringstream;
using std::string;
using std::ostream;

namespace MAd {

    void StdOutWriterFunction(const char * message, void * data)
    {
        std::cout << message;
    }
    void StdErrWriterFunction(const char * message, void * data)
    {
        std::cerr << message;
    }

  // -------------------------------------------------------------------
  MAdMsg::MAdMsg()
  {
      resetWriters();
  }

  // -------------------------------------------------------------------
  void MAdMsg::resetWriters()
  {
      outWriter = &StdOutWriterFunction;
      outData = NULL;
      errWriter = &StdErrWriterFunction;
      errData = NULL;
      prefix = "";
  }

  // -------------------------------------------------------------------
  void MAdMsg::registerAbortFct(AbortFunction fct, void * data)
  {
    abortFcts.push_back(std::make_pair(fct,data));
  }

  // -------------------------------------------------------------------
  void MAdMsg::write(const std::string& message) const
  {
      std::string prefixedMsg = message;
      addPrefix(&prefixedMsg);
      outWriter(prefixedMsg.c_str(), outData);
  }

  // -------------------------------------------------------------------
  void MAdMsg::info(int line,
                    const char* file,
                    const char* fmt,...) const
  {
    char buff[1024];
    va_list args;
    va_start (args, fmt);
    vsprintf(buff, fmt, args);
    va_end (args);

    std::string str = "Info: ";
    str += buff;
    appendLineAndFile(&str, line, file);
    addPrefix(&str);
    str += "\n";

    outWriter(str.c_str(), errData);
  }

  // -------------------------------------------------------------------
  void MAdMsg::warning(int line,
                       const char* file,
                       const char* fmt,...) const
  {
    char buff[1024];
    va_list args;
    va_start (args, fmt);
    vsprintf(buff, fmt, args);
    va_end (args);

    std::string str = "WARNING: ";
    str += buff;
    appendLineAndFile(&str, line, file);
    addPrefix(&str);
    str += "\n";

    errWriter(str.c_str(), errData);
  }

  // -------------------------------------------------------------------
  void MAdMsg::error(int line,
                     const char* file,
                     const char* fmt,...) const
  {
    char buff[1024];
    va_list args;
    va_start (args, fmt);
    vsprintf(buff, fmt, args);
    va_end (args);

    std::string str = "ERROR: ";
    str += buff;
    appendLineAndFile(&str, line, file);
    addPrefix(&str);
    str += "\n";

    errWriter(str.c_str(), errData);

    // call abort functions
    std::list<std::pair<AbortFunction,void*> >::const_iterator iter = abortFcts.begin();
    for (; iter != abortFcts.end(); iter++) {
      AbortFunction fct = (*iter).first;
      void * data       = (*iter).second;
      fct(data);
    }

    abort();
  }
  
  // -------------------------------------------------------------------
  void MAdMsg::setOutWriter(WriterFunction writer, void * data)
  {
      outWriter = writer;
      outData = data;
  }

  void MAdMsg::setErrWriter(WriterFunction writer, void * data)
  {
      errWriter = writer;
      errData = data;
  }

  void MAdMsg::setMessagePrefix(const char * _prefix)
  {
      prefix = _prefix;
  }
  void MAdMsg::addPrefix(std::string * message) const
  {
      // TODO: implement this: add the prefix to any new line contained in the message
      // For now, naive implementation that only works for one line:
      *message = prefix + *message;
  }

  void MAdMsg::appendLineAndFile(std::string * str, int line, const char * file) const
  {
      if (line >= 0)
      {
          *str += " (line ";
          stringstream ss;
          ss << line;
          std::string tmp;
          ss >> tmp;
          *str += tmp;
          *str += " in file \'";
          *str += file;
          *str += "\')";
      }
  }
}
