// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#ifndef _H_MADRESOURCEMANAGER
#define _H_MADRESOURCEMANAGER

#include "MAdSingleton.h"

#include <string>
#include <iostream>
#include <set>

#ifdef PARALLEL
#include <mpi.h>
#endif

namespace MAd {

  // -------------------------------------------------------------------
  //!\brief singleton managing the interface to the system 
  //! \ingroup tools
  class MAdResourceManager {

  public: // constructor/destructor
  
    MAdResourceManager();
    ~MAdResourceManager();

  public: // alter state
    
    //! \brief set timers and default communicator
    void initialize();
    
    //! \brief close all data
    void finalize();

    //! \brief reset the clock
    void reset();

  public:
    
    //! \brief retain program arguments
    void setArguments(int* ac,char*** av);

    //! \brief get program arguments
    void getArguments(int*& ac,char***& av);

  public: // resource interface
    
    //! \brief get elapsed time since last reset
    double elapsedTime() const;
    
    //! \brief get wall clock time
    double getTime() const;

    //! \brief Gives the amount of memory currently used on all processors
    //! provides resident set size RSS and virtual memory VM
    bool getMemoryUsage(double& RSS, double& VM) const;
    
    //! \brief Gives the amount of memory currently used locally on the processor
    //! provides resident set size RSS and virtual memory VM
    bool getLocalMemoryUsage(double& RSS,double& VM) const;

    //! \brief Write the memory usage to a stream s, preceded on all lines by prefix p
    void printMemoryUsage(std::string p="", std::ostream& s=std::cout) const;
    
  public: // communication interface
    
    //! \brief register a communication tag t to check for double use
    bool reserveCommunicationTag(int t);
    
    //! \brief release a communication tag t such that it can be used again
    bool releaseCommunicationTag(int t);

#ifdef PARALLEL
    
    //! \brief set the communicator - should be shared by the mesh and autopack
    //! \warning currently the mesh and autopack store their own communicator
    MPI_Comm setCommunicator(MPI_Comm comm);

    //! \brief get the communicator
    MPI_Comm getCommunicator();
#endif
    
  private:
    
    int* argc;    //!< program argument count
    char*** argv; //!< program arguments
    
    double initialTime; //!< wall clock time at last reset
    std::set<int> communicationTags; //!< all registered communication tags
    
#ifdef PARALLEL
    MPI_Comm communicator; //!< mpi communicator 
#endif
    
  };

  // -------------------------------------------------------------------

  typedef MAdSingleton<MAdResourceManager> MAdResourceManagerSgl;
}

#endif
