// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#ifndef _H_VECTOR_MAD
#define _H_VECTOR_MAD

#include "MAdMessage.h"
#include <cassert>
#include <cmath>
#include <limits>
#include <iostream>
#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <algorithm>
#include <iomanip>

// -------------------------------------------------------------------

namespace MAd {

  
  /* ===========================================================================*/
  /*! \brief proxy view on a vector                                             */
  /* ===========================================================================*/
  
  template <class SCALAR> 
  struct mad_vector_view {
    
  public:
    
    mad_vector_view(SCALAR* b,const SCALAR* e,size_t n,size_t l):
      size(n),stride(l),data(b),end(e) {}
    
    ~mad_vector_view() {}
    
    size_t size;
    size_t stride;
    SCALAR* const data;
    const SCALAR* const end;
  };
  
  /* ===========================================================================*/
  /*! \brief proxy view on a constant vector                                    */
  /* ===========================================================================*/
  
  template <class SCALAR> 
  struct mad_const_vector_view {
    
  public:
    
    mad_const_vector_view(const SCALAR* b,const SCALAR* e,size_t n,size_t l):
      size(n),stride(l),data(b),end(e) {}
    
    ~mad_const_vector_view() {}
    
    size_t size;
    size_t stride;
    const SCALAR* const data;
    const SCALAR* const end;
  };
  
  // -------------------------------------------------------------------
  // Basic vector / matrix
  // -------------------------------------------------------------------

  /*! \brief dense vector */
  
  template<class SCALAR>
  class MAd_Vector
  {
  private:
    int r;       //!<  number of rows of a vector
    int stride;  //!<  offset between two consecutive entries in case of a proxy
    
    bool owner;  //!<  allocated the data in the vector or proxy
    
    SCALAR *data;  //!<       pointer to data 
    const SCALAR* cData; //!< const pointer to data 

  private:
    
    //! \brief resize or allocate internal space 
    void resize(int R) {
      if (data && owner) delete [] data;
      if (R) data = R==0? NULL : new SCALAR[R];
      cData = data;
      for (int i=0;i<R;i++) data[i] = 0;
      owner = true;
      r = R;
      stride = 1;
    }
    
  public:
    
    //! \brief destructor
    ~MAd_Vector() { if(data && owner) delete [] data; }
  
    //! \brief empty constructor
    MAd_Vector():data(NULL) {resize(0);}

    //! \brief constructor based on the size of the vector
    MAd_Vector(int R):data(NULL) {resize(R);}
    
    //! \brief constructor based on a copy of another vector 
    MAd_Vector(const MAd_Vector &other):data(NULL) {*this = other;}
    
    //! \brief reinterpretation of an external table as a vector
    MAd_Vector(SCALAR* const ptr,int n,int s=1):
      r(n),stride(s),owner(false),data(ptr),cData(ptr) {}

    //! \brief constructor based on a modifiable proxy view
    MAd_Vector(mad_vector_view<SCALAR> view):
      r((int)(view.size)),
      stride((int)(view.stride)),
      owner(false),
      data(view.data),
      cData(view.data) {}
    
    //! \brief constructor based on a const proxy view 
    MAd_Vector(mad_const_vector_view<SCALAR> view):
      r((int)(view.size)),
      stride((int)(view.stride)),
      owner(false),
      data(NULL),
      cData(view.data) {}
  
  public:

    inline int size() const {return r;} 
  
    const double* seePtr() const {return cData;}
    
    double* touchPtr() {return data;}

    inline MAd_Vector& operator= (const MAd_Vector &other) {
      if ( this != &other ) {
        resize(other.r);
        for (int i = 0; i < r; ++i) data[i*stride] = other(i);
      }
      return *this;
    }

    // object = object + previously defined oject
    inline MAd_Vector& operator+= (const MAd_Vector& other) 
    {
      add(other);
      return *this;
    }
    
    inline MAd_Vector& operator-= (const MAd_Vector& other) 
    {
      subtract(other);
      return *this;
    }
    
    inline MAd_Vector& operator+ (const MAd_Vector& other) const
    {
      MAd_Vector tmp(size());
      tmp = *this;
      tmp.add(other);
      return tmp;
    }


    inline MAd_Vector& operator= (SCALAR a) 
    {
      set_all(a);
      return *this;
    }

    inline MAd_Vector& operator*= (SCALAR a) 
    {
      scale(a);
      return *this;
    }

    inline SCALAR operator () (int i) const
    {
      return data[i*stride];
    }
    inline SCALAR & operator () (int i)
    {
      return data[i*stride];
    }
    
    bool operator== (const MAd_Vector& other) const {
      if (size() != other.size()) return false;
      for (size_t i=0;i<size();i++) if (data[i*stride] != other(i)) return false;
      return true;
    }  
    
    inline SCALAR norm()
    {
      SCALAR n = 0.;
      for(int i = 0; i < r; ++i) n += data[i*stride] * data[i*stride];
      return sqrt(n);
    }
    inline void scale(const SCALAR s)
    {
      for (int i = 0; i < r; ++i) data[i*stride] *= s;
    }
    inline void set_all(const SCALAR &m) 
    {
      for (int i = 0; i < r; ++i) data[i*stride] = m;
    }
    inline void add(const MAd_Vector &v) 
    {
      for (int i = 0; i < r; ++i) data[i*stride] += v(i);    
    }
    inline void subtract(const MAd_Vector &v) 
    {
      for (int i = 0; i < r; ++i) data[i*stride] -= v(i);    
    }
    void print(std::string name="") const
    {
      printf("Vector %s:\n",name.c_str());
      for (int i = 0; i < r; ++i) printf("%12.5E ",data[i*stride]);
      printf("\n");
    }
  };
  
  // -------------------------------------------------------------------
  
  typedef MAd_Vector<double> doubleVector;
  typedef MAd_Vector<int> integerVector;
  
  // -------------------------------------------------------------------

}

#endif
