// -*- C++ -*- 
// -----------------------------------------------------------------------------
// MAdLiB - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------


#ifndef BASICFUNCTIONS_H
#define BASICFUNCTIONS_H

#include <cmath>
#include <iostream>
#include <complex>
//using std::complex;

// -----------------------------------------------------------------------------
/*! \brief factorial */
/*! \f$ n! = \prod_{i=1}^n i \f$ */
// -----------------------------------------------------------------------------

inline int factorial(int n) {
  int value = 1;
  for (int i=1;i<=n;i++) value *= i;
  return value;
}

// -----------------------------------------------------------------------------
/*! \brief  Pochhammer symbol \f$ \mathcal P(x,n) \f$  */
/*!  \f$ \mathcal P(x,n) = \prod_{i=0}^{n-1} (x+i) \f$ */
/*! \f$ \mathcal P(x,0) = 1 \f$  by convention (empty product) */
// -----------------------------------------------------------------------------

inline double Pochhammer(double x,int n) {
  double result(1.);
  for (int i=0;i<n;i++) result *= (x+i);
  return result;
}

// -----------------------------------------------------------------------------
/*! \brief nChooseK  */
/*! \f$ \begin{pmatrix} n \\ k \end{pmatrix} = \frac{n!}{(n-k)!k!} \f$ */
// -----------------------------------------------------------------------------

inline double nChooseK(int n, int k){

  return (double)(factorial(n))/(factorial(k)*factorial(n-k));

}

// -----------------------------------------------------------------------------
/*! \brief Kummer M function \f$ \mathcal M(a,b,x)\f$        */ 
/*! 
  \f{eqnarray*}{
  \mathcal M(a,b,x) = \sum_{n=0}^\infty \frac {(a)_n x^n}{(b)_n n!}  \\
  (a)_n = \prod_{i=0}^{n-1} (a+i) = \mathcal P(a,n)
  \f}
  with \f$ \mathcal P(x,n)\f$ the Pochhammer symbol
  \warning prone to floating point overflow leading to non-convergence of the series
*/
// -----------------------------------------------------------------------------
template<class dtype> dtype KummerM(dtype a,dtype b,dtype x) {
  
  dtype n   = 0.; 
  dtype nn  = 1.;
  dtype update = 1;
  dtype value = 0;

  do {
    value += update;
    update *= (a+n) * x / ((b+n) * nn);
    n += 1;
    nn += 1;
  } while (norm(update) > 1.e-12*norm(value));
  
  return value;
}

#endif
