// -*- C++ -*- 
// -----------------------------------------------------------------------------
// MAdLiB - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------


#include "Bergot.h"

#include <algorithm>
#include <vector>
#include <cmath>
#include <map>
#include <sstream>

// -----------------------------------------------------------------------------
Bergot::Bergot(int p):order(p)  {
  
  // allocate function information and fill 

  iOrder = new int[size()];
  jOrder = new int[size()];
  kOrder = new int[size()];
  
  legendre = new Legendre(order);

  int index = 0;
  for (int i=0;i<=order;i++) {
    for (int j=0;j<=order;j++) {
      int mIJ = std::max(i,j);
      for (int k=0;k<=(order - mIJ);k++,index++) {
        
        iOrder[index] = i;
        jOrder[index] = j;
        kOrder[index] = k;
        
        if (jacobi.find(mIJ) == jacobi.end()) {
          jacobi[mIJ] = new Jacobi(2*mIJ+2,0,order);
        }
      }
    }
  }
}

// -----------------------------------------------------------------------------

Bergot::~Bergot() {

  if (iOrder) delete [] iOrder;
  if (jOrder) delete [] jOrder;
  if (kOrder) delete [] kOrder;
  
  if (legendre) delete legendre;
  std::map<size_t,Jacobi*>::iterator jIter = jacobi.begin();
  for (;jIter!=jacobi.end();++jIter) delete jIter->second;
  jacobi.clear();
}

// -----------------------------------------------------------------------------

std::string Bergot::name() const {
  std::ostringstream nameStream;
  nameStream << "Family of orthogonal functions of order " 
             << order << " on the pyramid";
  return nameStream.str();
}

// -----------------------------------------------------------------------------

size_t Bergot::size() const {
  int n = order + 1;
  int n2 = n*n;
  int n3 = n2*n;
  return (2*n3+3*n2+n)/6;
}

// -----------------------------------------------------------------------------

size_t Bergot::dim() const { return 3;}

// -----------------------------------------------------------------------------

size_t Bergot::maxOrder() const {return order;}

// -----------------------------------------------------------------------------

int Bergot::elementType() const {return MSH_PYRAMID;}

// -----------------------------------------------------------------------------

bool Bergot::orthogonal() const {return true;}

// -----------------------------------------------------------------------------

void Bergot::polynomialOrders(int i,int& pu,int& pv,int& pw) const {
  pu = iOrder[i];
  pv = jOrder[i];
  pw = kOrder[i];
}

// -----------------------------------------------------------------------------

std::vector<std::vector<int> > Bergot::polynomialOrders() const {

  std::vector<std::vector<int> > orders;
  for (size_t i=0;i<size();i++) {
    orders[0].push_back(iOrder[i]);
    orders[1].push_back(jOrder[i]);
    orders[2].push_back(kOrder[i]);
  }
  return orders;
}


// -----------------------------------------------------------------------------

void Bergot::evaluateValues(double uvw[3],std::vector<double>& f) const {

  f.resize(size());
  
  const double u(uvw[0]);
  const double v(uvw[1]);
  const double w(uvw[2]);

  // first evaluate all function values 
  
  // when w = 1 all functions have finite values since multiplied by (1-w)^mIJ
  /*
    Actually, if we use the mapping Pyramid <-> Cube, we notice that all inputs of 
    the Legendre polynomials have finite values (as the Jacobi one).
    Therefore the evaluation is zero, except for mIJ = 0 since pow(.,0) = 1
  */

  double uhat = (w == 1.) ? 1 : u/(1.-w);
  std::vector<double> uFcts;
  legendre->evaluateValues1D(uhat,uFcts);

  double vhat = (w == 1.) ? 1 : v/(1.-w);
  std::vector<double> vFcts;
  legendre->evaluateValues1D(vhat,vFcts);

  double what = 2.*w - 1.;
  std::map<size_t,std::vector<double> > wFcts;
  std::map<size_t,Jacobi*>::const_iterator jIter=jacobi.begin();

  for (;jIter!=jacobi.end();jIter++) {
    size_t a = jIter->first;
    std::vector<double>& wf = wFcts[a];
    jIter->second->evaluateValues1D(what,wf);
  }
  
  // recombine to find shape function values 

  size_t index = 0;
  for (int i=0;i<=order;i++) {
    for (int j=0;j<=order;j++) {
      size_t mIJ = std::max(i,j);
      double fact = std::pow(1-w,(int) mIJ); // pow(.,0) = 1, it is ok even if basis is zero by continuity
      std::vector<double>& wf = (wFcts.find(mIJ))->second;
      for (size_t k=0;k<=(size_t) (order - mIJ);k++,index++) {
        f[index] = uFcts[i] * vFcts[j] * wf[k] * fact;
      }
    }
  }
}

// -----------------------------------------------------------------------------

void Bergot::evaluateDerivatives(double uvw[3], // use w = {.99;1} and compare ; values should be close to each other as it is supposed to be continuous
                                 std::vector<double>& du,
                                 std::vector<double>& dv,
                                 std::vector<double>& dw) const {
  
  // copy coordinates for convenience

  const double u(uvw[0]);
  const double v(uvw[1]);
  const double w(uvw[2]);

  // first evaluate all functions and derivatives in the point (u,v,w)


	double uhat = (w == 1.) ? 1 : u/(1.-w);
  std::vector<double> uFcts;
  legendre->evaluateValues1D(uhat,uFcts);
  
  std::vector<double> uGrads;
  legendre->evaluateDerivatives1D(uhat,uGrads);


  double vhat = (w == 1.) ? 1 : v/(1.-w);
  std::vector<double> vFcts;
  legendre->evaluateValues1D(vhat,vFcts);

  std::vector<double> vGrads;
  legendre->evaluateDerivatives1D(vhat,vGrads);

  double what = 2.*w - 1.;
  std::map<size_t,std::vector<double> > wFcts;
  std::map<size_t,std::vector<double> > wGrads;
  std::map<size_t,Jacobi*>::const_iterator jIter=jacobi.begin();
  
  for (;jIter!=jacobi.end();jIter++) {
    size_t a = jIter->first;
    std::vector<double>& wf = wFcts[a];
    jIter->second->evaluateValues1D(what,wf);
    std::vector<double>& wg = wGrads[a];
    jIter->second->evaluateDerivatives1D(what,wg);
  }
  
  // now recombine to find the shape function gradients

  int index = 0;
  for (int i=0;i<=order;i++) {
    for (int j=0;j<=order;j++) {

      int mIJ = std::max(i,j);

      std::vector<double>& wf = (wFcts .find(mIJ))->second;
      std::vector<double>& wg = (wGrads.find(mIJ))->second;

      double wPowM2 = std::pow(1.-w, (int) mIJ - 2);
      double facU = u*wPowM2;
      double facV = v*wPowM2;
      if ( w == 1. ){
        if (mIJ == 0.){
          facU = 0.;
          facV = 0.;
        }
        else if (mIJ == 1.){
          facU = 1.;
          facV = 1.;
        }
      }
      double wPowM1 = (mIJ == 0) ? 0 : std::pow(1.-w, (int) mIJ - 1);
      double wPowM0 = std::pow(1.-w, (int) mIJ);
			
      for (size_t k=0;k<= (size_t) (order - mIJ);k++,index++) {

        du[index] = uGrads[i] * vFcts[j]  * wf[k] * wPowM1;

        dv[index] = uFcts[i]  * vGrads[j] * wf[k] * wPowM1;

        dw[index]  = 0;
        dw[index] += facU   * uGrads[i] * vFcts[j]  * wf[k];
        dw[index] += facV   * uFcts[i]  * vGrads[j] * wf[k];
        dw[index] -= mIJ * uFcts[i]  * vFcts[j]  * wf[k] * wPowM1;
        dw[index] +=   2 * uFcts[i]  * vFcts[j]  * wg[k] * wPowM0;
            
      }
    }
  }
}

// -----------------------------------------------------------------------------

double Bergot::evaluateWeightFunction(double uvw[3]) const {return 1;}
