// -*- C++ -*- 
// -----------------------------------------------------------------------------
// MAdLiB - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------


#ifndef BERGOT_H
#define BERGOT_H

#include "FunctionFamily.h"
#include "Legendre.h"
#include "Jacobi.h"

#include <map>
#include <vector>
#include <string>


/*!\brief Family of orthogonal functions conforming with standard function
  spaces of boundaries, as defined by Bergot et al. (SIAM JSciComp 2010) */
/*! 
  The canonical pyramid is defined as
  \f{eqnarray*}{
  -1 + \zeta &\leq \xi,\eta \leq& 1 - \zeta \\
  0 &\leq \zeta \leq& 1
  \f}
  and the functions \f$\psi_{ijk}\f$ of up to order p are given by
  \f{eqnarray*}{
  \psi_{ijk} &=& P_i\left(\frac{\xi}{1-\zeta}\right) P_j\left(\frac{\eta}{1-\zeta}\right) \left(1-\zeta\right)^{\mu_{ij}} P^{(2\mu_{ij}+2,0)}_{k}\left(2\zeta -1\right) \\
v  \mu_{ij} &=& \max(i,j) \\
  0 &\leq i,j \leq& p \\
  0 &\leq k\leq& p-\mu_{ij} 
  \f}
  with \f$P_n\f$ the Legendre polynomial of order n, and \f$P^{(a,b)}_n\f$ the Jacobi polynomial of order n and exponent pair \f$(a,b)\f$
  (see Abramowitz and Stegun, chapter 22)
*/

class Bergot : public HierarchicFunctionFamily {
  
public:
  
  Bergot(int);
  ~Bergot();

public:
  
  virtual std::string name()   const;
  virtual size_t size()        const;
  virtual size_t dim()         const;
  virtual size_t maxOrder()    const;
  virtual int    elementType() const;
  virtual bool   orthogonal()  const;

public:
  
  void polynomialOrders(int,int&,int&,int& ) const ;
  std::vector<std::vector<int> > polynomialOrders() const ;
 
public:
  
  virtual void evaluateValues(double [3],std::vector<double>&) const;
  
  virtual void evaluateDerivatives(double [3],
                                   std::vector<double>&,
                                   std::vector<double>&,
                                   std::vector<double>&) const;
  
  virtual double evaluateWeightFunction(double[3]) const;

private:

  int order;   //!< interpolation order

  int *iOrder; //!< order of \f$\hat \xi \f$ polynomial 
  int *jOrder; //!< order of \f$\hat \eta \f$ polynomial 
  int *kOrder; //!< order of \f$\hat \zeta \f$ polynomial 

  
private:

  //! list of Legendre polynomials up to order p 
  Legendre* legendre;
  
  //! list of Jacobi polynomials up to order p in function of index i
  //! (\f$ \alpha = 2*i + 2\f$)
  std::map<size_t,Jacobi*> jacobi; 
  
};

#endif
