// -*- C++ -*- 
// -----------------------------------------------------------------------------
// MAdLiB - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------


#include <vector>
using std::vector;
#include <iostream>
using std::ostream;
using std::flush;
using std::ostream;
#include <limits>
using std::numeric_limits;


#include "FunctionFamily.h"

// -----------------------------------------------------------------------------

const double FunctionFamily::eps = 1e3*numeric_limits<double>::epsilon();

// -----------------------------------------------------------------------------

const double FunctionFamily::over2Eps = 1./(2*FunctionFamily::eps);

// -----------------------------------------------------------------------------

void FunctionFamily::evaluateDerivatives(double uvw[3],
                                         vector<double>& du,
                                         vector<double>& dv,
                                         vector<double>& dw) const {
  

  du.resize(size());
  dv.resize(size());
  dw.resize(size());
  
  vector<double>* der[3] = {&du,&dv,&dw};
  
  double uvwEps[3] = {uvw[0],uvw[1],uvw[2]};

  vector<double> valPEps(size());
  vector<double> valMEps(size());

  for (size_t iDim=0;iDim<dim();iDim++) {

    uvwEps[iDim] = uvw[iDim] + eps;
    evaluateValues(uvwEps,valPEps);
    uvwEps[iDim] = uvw[iDim] - eps;
    evaluateValues(uvwEps,valMEps);
    uvwEps[iDim] = uvw[iDim];

    for (size_t i=0;i<size();i++) (*der)[i][iDim] = (valPEps[i] - valMEps[i])*over2Eps;

  }
}

// -----------------------------------------------------------------------------

void Hierarchic1DFunctionFamily::polynomialOrders(int i,int& pu,int& pv,int& pw) const {
  pu = i - 1;
  pv = pw = 0;
}

// -----------------------------------------------------------------------------

vector<vector<int> > Hierarchic1DFunctionFamily::polynomialOrders() const {

  vector<vector<int> > orders;
  for (int i=0;i<=int(maxOrder());i++) orders[0].push_back(i);
  return orders;
}

// -----------------------------------------------------------------------------

void Hierarchic1DFunctionFamily::evaluateValues(double uvw[3],vector<double>& val) const {
  val.resize(size());
  evaluateValues1D(uvw[0],val);
}

// -----------------------------------------------------------------------------

void Hierarchic1DFunctionFamily::evaluateDerivatives(double uvw[3],
                                                     vector<double>& du,
                                                     vector<double>& dv,
                                                     vector<double>& dw) const {
  
  du.resize(size());
  dv.resize(size());
  dw.resize(size());
  
  for (size_t i=0;i<size();i++) {dv[i] = dw[i] = 0;}
  evaluateDerivatives1D(uvw[0],du);
}

// -----------------------------------------------------------------------------

double Hierarchic1DFunctionFamily::evaluateWeightFunction(double uvw[3]) const {
  return evaluateWeightFunction1D(uvw[0]);
}

// -----------------------------------------------------------------------------

void Hierarchic1DFunctionFamily::plotValues(ostream& out,int nbPoints)  const {

  vector<double> tmp(size());
  size_t nb = (nbPoints <= 0) ? 200 : nbPoints;
  double dx = (range().second - range().first)/nb;
  
  for (size_t i=0;i<=nb;i++) {
    double x = (range()).first + i*dx;    
    evaluateValues1D(x,tmp);
    out << x << " ";
    for (size_t j=0;j<size();j++) out << " " << tmp[j];
    out << "\n";
  }
  out << flush;
}

// -----------------------------------------------------------------------------

void Hierarchic1DFunctionFamily::plotDerivatives(ostream& out,int nbPoints)  const {
  
  size_t nb = (nbPoints <= 0) ? 200 : nbPoints;
  double dx = (range().second - range().first)/nb;
  
  vector<double> der(size());

  for (size_t i=0;i<=nb;i++) {
    double x = range().first + i*dx;    
    evaluateDerivatives1D(x,der);
    out << x << " ";
    for (size_t j=0;j<size();j++) out << " " << der[j];
    out << "\n";
  }
  out << flush;
}

// -----------------------------------------------------------------------------

void Hierarchic1DFunctionFamily::plotDerivativesFD(ostream& out,int nbPoints)  const {
  
  size_t nb = (nbPoints <= 0) ? 200 : nbPoints;
  double dx = (range().second - range().first)/nb;
  
  vector<double> pEps(size());
  vector<double> mEps(size());

  for (size_t i=0;i<=nb;i++) {
    double x = range().first + i*dx;    
    evaluateValues1D(x+eps,pEps);
    evaluateValues1D(x-eps,mEps);
    out << x << " ";
    for (size_t j=0;j<size();j++) out << " " << (pEps[j] - mEps[j])*over2Eps;
    out << "\n";
  }
  out << flush;
}
