// -*- C++ -*- 
// -----------------------------------------------------------------------------
// MAdLiB - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------


#ifndef FUNCTIONFAMILY_H
#define FUNCTIONFAMILY_H

#include <string>
//using std::string;
#include <vector>
using std::vector;
#include <sstream>
#include <utility>
//using std::pair;
//using std::make_pair;
#include <iostream>
//using std::ostream;

#include "BasicFunctions.h"

// -----------------------------------------------------------------------------
/*! \brief Base class for a family of functions defined on geometric domains  */
// -----------------------------------------------------------------------------

class FunctionFamily {
  
public:

  virtual ~FunctionFamily() {}

public:
  
  enum ElementType {MSH_LINE=1,
                    MSH_TRIANGLE=2,
                    MSH_QUADRRILATERAL=3,
                    MSH_TETRAHEDRON=4,
                    MSH_PYRAMID=7,
                    MSH_WEDGE=6,
                    MSH_HEXAHEDRON=5,
                    UNDEFINED=0};

public:
  
  virtual std::string name() const = 0;
  virtual size_t      size() const = 0;
  virtual size_t      dim()  const = 0;
  virtual int         elementType() const = 0;
  virtual bool        orthogonal()  const = 0;
 
public:
  
  virtual void evaluateValues(double [3],std::vector<double>&) const = 0;
  virtual void evaluateDerivatives(double [3],
                                   std::vector<double>&,
                                   std::vector<double>&,
                                   std::vector<double>&) const = 0;
  
  virtual double evaluateWeightFunction(double[3]) const = 0;

protected:
  
  static const double eps; 
  static const double over2Eps;
  
};

// -----------------------------------------------------------------------------
/*! \brief Base class for a hierarchic family of polynomial functions  */
// -----------------------------------------------------------------------------

class HierarchicFunctionFamily : public FunctionFamily {
  
public:

  virtual ~HierarchicFunctionFamily() {};

public:
  
  virtual std::string name()        const = 0;
  virtual size_t      size()        const = 0;
  virtual size_t      dim()         const = 0;
  virtual size_t      maxOrder()    const = 0;
  virtual int         elementType() const = 0;
  virtual bool        orthogonal()  const = 0;

public: 
  
  virtual void polynomialOrders(int i,int& ,int& ,int&) const = 0;
  virtual std::vector<std::vector<int> > polynomialOrders() const = 0;
  
public:
  
  virtual void evaluateValues(double [3],std::vector<double>&) const = 0;
  virtual void evaluateDerivatives(double [3],
                                   std::vector<double>&,
                                   std::vector<double>&,
                                   std::vector<double>&) const = 0;
  
  virtual double evaluateWeightFunction(double[3]) const = 0;
  
};

// -----------------------------------------------------------------------------
/*! \brief Base class for a hierarchic family of 1D functions  */
// -----------------------------------------------------------------------------

class Hierarchic1DFunctionFamily : public HierarchicFunctionFamily {
  
public:
  
  virtual ~Hierarchic1DFunctionFamily() {}
  
public:

  size_t  dim() const {return 1;}
  virtual std::string name()     const = 0;
  virtual size_t size()          const = 0;
  size_t maxOrder()      const {return (size_t) ((int) size() -1);}
  int    elementType()   const {return MSH_LINE;}
  virtual std::pair<double,double> range() const = 0;  

public:

  void polynomialOrders(int i,int& ,int& ,int&) const;
  std::vector<std::vector<int> > polynomialOrders() const;
  
public:

  void evaluateValues(double[3],std::vector<double>&) const;
  void evaluateDerivatives(double[3],
                           std::vector<double>&,
                           std::vector<double>&,
                           std::vector<double>&) const;
  double evaluateWeightFunction(double[3]) const;

public:

  virtual void   evaluateValues1D(double,std::vector<double>&)      const = 0;
  virtual void   evaluateDerivatives1D(double,std::vector<double>&) const = 0;
  virtual double evaluateWeightFunction1D(double)                   const = 0;

public:

  virtual void plotValues(std::ostream&,int=-1) const;
  virtual void plotDerivatives(std::ostream&,int=-1) const;
  virtual void plotDerivativesFD(std::ostream&,int=-1) const;
  
};

#endif
