// -*- C++ -*- 
// -----------------------------------------------------------------------------
// MAdLiB - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------


#ifndef JACOBI_H
#define JACOBI_H

#include "Jacobi.h"

#include <sstream>
using std::ostringstream;
using std::pair;
using std::make_pair;
using std::string;
using std::vector;
// -----------------------------------------------------------------------------

Jacobi::Jacobi(double a,double b,int o):
  alpha(a),beta(b),n(o),alphaPlusBeta(a+b),a2MinusB2(a*a - b*b) {}

// -----------------------------------------------------------------------------

Jacobi::~Jacobi() {;}

// -----------------------------------------------------------------------------

string Jacobi::name() const {
  
  ostringstream nn;
  nn << "Jacobi polynomial of order " << n << " with exponent pair (" << alpha << "," << beta << ")";
  return nn.str();
}

// -----------------------------------------------------------------------------

size_t Jacobi::size() const {return n+1;}

// -----------------------------------------------------------------------------
void Jacobi::evaluateValues1D(double x,vector<double>& fcn) const {

	fcn.resize(n+1);

	fcn[0] = 1;
	if (n>0) fcn[1] = .5 * ( 2 * (alpha+1) + (alphaPlusBeta+2) * (x-1) );

	for (int i=1; i<n; i++){

	  double ii = (double) i;
	  
	  double np1 =  ii+1;
	  double a0i = 2*ii + alphaPlusBeta;
	  double a1i = ii + alphaPlusBeta + 1;
	  
	  double an = (a0i + 1) * (a0i + 2) / ( 2 * np1 * a1i );
	  double bn = -a2MinusB2 * (a0i + 1) / ( 2 * np1 * a1i * a0i );
	  double cn = (ii + alpha) * (ii + beta) * (a0i + 2) / ( np1 * a1i * a0i);

	  fcn[i+1] = ( an * x - bn ) * fcn[i] - cn * fcn[i-1];
	  
	}

}
// -----------------------------------------------------------------------------
void Jacobi::evaluateDerivatives1D(double x,vector<double>& der) const {

	der.resize(n+1);

  vector<double> tmp(size());
  evaluateValues1D(x,tmp);
  
  der[0] = 0;
  if (n>0) der[1] = 0.5*(alphaPlusBeta + 2.);

  for (int i=1;i<n;i++) {

		double ii = (double) i;

		double np1 =  ii+1;
		double a0i = 2*ii + alphaPlusBeta;
		double a1i = ii + alphaPlusBeta + 1;

		double an = (a0i + 1) * (a0i + 2) / ( 2 * np1 * a1i );
		double bn = -a2MinusB2 * (a0i + 1) / ( 2 * np1 * a1i * a0i );
		double cn = (ii + alpha) * (ii + beta) * (a0i + 2) / ( np1 * a1i * a0i);
  
    der[i+1] = an * tmp[i] + ( an * x - bn ) * der[i] - cn * der[i-1];
  }
}

// -----------------------------------------------------------------------------

double Jacobi::evaluateWeightFunction1D(double x) const {
  return pow(1-x,alpha)*pow(1+x,beta);
}

// -----------------------------------------------------------------------------

pair<double,double> Jacobi::range() const {
  return make_pair(-1,1);
}

// -----------------------------------------------------------------------------
#endif
