// -*- C++ -*- 
// -----------------------------------------------------------------------------
// MAdLiB - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------


#include "MAd_HexInterpolant.h"

#include <utility>
using std::swap;
#include <cmath>
using std::max;

//---------------------------------------------------------------------
// Hex
/*  
 *         |0|                     order 0 (#1)
 *
 *
 *          v
 *   3----------2
 *   |\     ^   |\
 *   | \    |   | \
 *   |  \   |   |  \
 *   |   7------+---6
 *   |   |  +-- |-- | -> u         order 1 (#8)
 *   0---+---\--1   | 
 *    \  |    \  \  |
 *     \ |     \  \ |
 *      \|      w  \|
 *       4----------5
 *
 *
 *          
 *   3----10----2
 *   |\         |\
 *   | 19   24  | 18
 *   11 \ 20    9   \
 *   |   7----14----6
 *   |25 |  26  | 23|              order 2 (#27)
 *   0---|-8----1   | 
 *    \  15   21 \  13
 *     16|  22   17 |
 *      \|         \|
 *       4----12----5
 *
 *
 *        ...
*/

//-------------------------------
//  Public Members
//-------------------------------

MAd_HexInterpolant::MAd_HexInterpolant(const int &o)
  : MAd_Interpolant(MSH_HEX_8, o), ndof(size_t(size(o))), 
    lefs(o), ndofe(lefs.size()), 
    OrientedFaceDof(48, int(getClosureSpaceSize())) {
  
  quad = new MAd_QuadInterpolant(o); // not ideal, but can not be gotten from factor

  NodeNumberu = new double[ndof];
  NodeNumberv = new double[ndof];
  NodeNumberw = new double[ndof];
  idFromIJK = new int[ndof];
  nodePositionHex(order(), NodeNumberu, NodeNumberv, NodeNumberw, idFromIJK);
  
  coef.resize(size(),size());
  for (int i=0; i< int(ndof); i++){
    for (int j=0; j< int(ndof); j++) {
      coef(i,j)=getCoef(i,j);
    }
  }

  int facenumber, dir, rotate;
  for (int i=0; i< 48 ; i++){
    getOrientationFromCode(i, facenumber, dir, rotate);
    getClosureSpace(facenumber, dir, rotate , &OrientedFaceDof(i,0));
  }
  
  
}

// -----------------------------------------------------------------------------

MAd_HexInterpolant::~MAd_HexInterpolant() {
  delete quad;
  delete[] NodeNumberu;
  delete[] NodeNumberv;
  delete[] NodeNumberw;
  delete[] idFromIJK;
  return;
}

// -----------------------------------------------------------------------------

void MAd_HexInterpolant::fcts (const double &u, const double &v, const double &w, 
				      doubleVector &f) const {
  const int n=int(ndof); 
  doubleVector m(n);
  f = 0; 
  for (int i=0; i< int(ndof); i++) m(i) = monomeof(u,v,w,i);
  coef.mult(m,f,false);
  return;
}

// -----------------------------------------------------------------------------

void MAd_HexInterpolant::grads (const double &u, const double &v, const double &w, 
                                      doubleVector gr[3]) const {
  const int n=int(ndof);
  doubleVector m(n);
  for (int k=0; k< 3; k++){
    gr[k] = 0;
    for (int i=0; i< int(ndof); i++) m(i) = dmonomeof(u,v,w,i,k);
    coef.mult(m,gr[k],false);
  }
  
  return;
}

// -----------------------------------------------------------------------------

void MAd_HexInterpolant::getIJKfromIdx(int idx, int &i, int &j, int &k) const {
  i = (int)((NodeNumberu[idx]*order()+order())/2.);
  j = (int)((NodeNumberv[idx]*order()+order())/2.);
  k = (int)((NodeNumberw[idx]*order()+order())/2.);
  return;
}

// -----------------------------------------------------------------------------

double MAd_HexInterpolant::getCoef(int id1,int id2) const {
  int i1,i2,j1,j2,k1,k2;
  getIJKfromIdx(id1, i1, j1, k1);
  getIJKfromIdx(id2, i2, j2, k2);
  double coefu = lefs.getCoef(order(), i1, i2);
  double coefv = lefs.getCoef(order(), j1, j2);
  double coefw = lefs.getCoef(order(), k1, k2);
  return coefu*coefv*coefw;
}

// -----------------------------------------------------------------------------

void MAd_HexInterpolant::getMonomialExponents(int i, int &powerU, int &powerV, int &powerW) const {
  int iu,iv,iw;
  getIJKfromIdx(i, iu, iv, iw);
  powerU = iu;
  powerV = iv;
  powerW = iw;
}

// -----------------------------------------------------------------------------

double MAd_HexInterpolant::monomeof(const double &u, const double &v, const double &w, int i) const{
  int powerU, powerV, powerW;
  getMonomialExponents(i, powerU, powerV, powerW);
  return pow(u,powerU)*pow(v,powerV)*pow(w,powerW);
}

// -----------------------------------------------------------------------------

double MAd_HexInterpolant::dmonomeof(const double &u, const double &v, const double &w, int i, int k) const{
  double Zero=0.;
  int powerU, powerV, powerW;
  getMonomialExponents(i, powerU, powerV, powerW);
  switch (k){
  case 0:
    if (powerU > 0)
      return powerU * pow(u,powerU-1)*pow(v,powerV)*pow(w,powerW);
    else return Zero;
    break;
  case 1:    
    if (powerV > 0)
      return powerV * pow(u,powerU)*pow(v,powerV-1)*pow(w,powerW);
    else return Zero;
    break;
  case 2:    
    if (powerW > 0)
      return powerW * pow(u,powerU)*pow(v,powerV)*pow(w,powerW-1);
    else return Zero;
    break;
  default:
    throw;
    break;
  }
}

// -----------------------------------------------------------------------------

int MAd_HexInterpolant::getClosureSpaceSize (int iFace) const {
  return MAd_QuadInterpolant::size(order());
}

// -----------------------------------------------------------------------------

  const int* MAd_HexInterpolant::getClosure (int iFace, int iSign, int iRotate) const {
  return OrientedFaceDof[getCodeFromOrientation(iFace, iSign, iRotate)];
}


// -----------------------------------------------------------------------------
void MAd_HexInterpolant::getClosureSpace(const int &iFace,
                                         const int &sign, 
                                         const int &rotate,
                                         int* facedofs) const {
  if(order()==0) facedofs[0]=0;
  else {
    int n = 0;
    // vertices
    for (size_t i=0; i<4; i++) facedofs[i]=fToN[iFace][i];
    n += 4;
    //edges
    for (int ie=0; ie<4; ie++){
      for (int k=0; k<order()-1; k++){ 
	int km = k;
	if (fToEdir[iFace][ie]==-1) km = order()-2-k;
	facedofs[n+k] = 8+fToE[iFace][ie]*(order()-1)+km;
      }
      n +=order()-1;
    }
    //inside
    size_t ndofqi = size_t(MAd_QuadInterpolant::size(order()-2));
    for (int k=0; k<int(ndofqi); k++){
      // offsetNode + offsetEdges + offsetFaceTri + offsetFaceQuad
      facedofs[n+k] = 8+12*(order()-1)+iFace*int(ndofqi)+k;
    }
    // update rotation and direction
    if (rotate>0) changeOrientation(facedofs,rotate,order());
    if (sign<=0) changeDir(facedofs,order());
  }
  return;
}

// -----------------------------------------------------------------------------

void MAd_HexInterpolant::nodeNumberCoordinate(const int &i, double &u, double &v, double &w) const{

  u=NodeNumberu[i];
  v=NodeNumberv[i];
  w=NodeNumberw[i];
  return;
}

// -----------------------------------------------------------------------------

int MAd_HexInterpolant::size(int q){
  return (q+1)*(q+1)*(q+1);
}

//-------------------------------
//  Private Members
//-------------------------------

void MAd_HexInterpolant::nodePositionHex(int o, double *u, double *v, double *w, int *idIJK) const {
  if (o==0) {
    u[0]=0;
    v[0]=0;
    w[0]=0;
    idIJK[0] = 0;
    return;
  }
  // hex vertices
  u[0] = -o;
  v[0] = -o;
  w[0] = -o;
  idIJK[0] = 0;
  u[1] = o;
  v[1] = -o;
  w[1] = -o;
  idIJK[o] = 1;
  u[2] = o;
  v[2] = o;
  w[2] = -o;
  idIJK[o*(o+1)+o] = 2;
  u[3] = -o;
  v[3] = o;
  w[3] = -o;
  idIJK[o*(o+1)] = 3;
  u[4] = -o;
  v[4] = -o;
  w[4] = o;
  idIJK[o*(o+1)*(o+1)] = 4;
  u[5] = o;
  v[5] = -o;
  w[5] = o;
  idIJK[o*(o+1)*(o+1)+o] = 5;
  u[6] = o;
  v[6] = o;
  w[6] = o;
  idIJK[o*(o+1)*(o+1)+o*(o+1)+o] = 6;
  u[7] = -o;
  v[7] = o;
  w[7] = o;
  idIJK[o*(o+1)*(o+1)+o*(o+1)] = 7;

  int n = 8;

  // edges
  for (int m=0; m<o-1; m++){
    u[n+m] = 2*(m+1)-o;          v[n+m] = -o;                 w[n+m] = -o;      
    u[n+(o-1)+m] = o;            v[n+(o-1)+m] = 2*(m+1)-o;    w[n+(o-1)+m] = -o;
    u[n+2*(o-1)+m] = o-2*(m+1);  v[n+2*(o-1)+m] = o;          w[n+2*(o-1)+m] = -o;
    u[n+3*(o-1)+m] = -o;         v[n+3*(o-1)+m] = o-2*(m+1);  w[n+3*(o-1)+m] = -o;

    u[n+4*(o-1)+m] = -o;         v[n+4*(o-1)+m] = -o;         w[n+4*(o-1)+m] = 2*(m+1)-o;      
    u[n+5*(o-1)+m] = o;          v[n+5*(o-1)+m] = -o;         w[n+5*(o-1)+m] = 2*(m+1)-o;
    u[n+6*(o-1)+m] = o;          v[n+6*(o-1)+m] = o;          w[n+6*(o-1)+m] = 2*(m+1)-o;
    u[n+7*(o-1)+m] = -o;         v[n+7*(o-1)+m] = o;          w[n+7*(o-1)+m] = 2*(m+1)-o;

    u[n+8*(o-1)+m] = 2*(m+1)-o;  v[n+8*(o-1)+m] = -o;         w[n+8*(o-1)+m] = o;      
    u[n+9*(o-1)+m] = o;          v[n+9*(o-1)+m] = 2*(m+1)-o;  w[n+9*(o-1)+m] = o;
    u[n+10*(o-1)+m] = o-2*(m+1); v[n+10*(o-1)+m] = o;         w[n+10*(o-1)+m] = o;
    u[n+11*(o-1)+m] = -o;        v[n+11*(o-1)+m] = o-2*(m+1); w[n+11*(o-1)+m] = o;
  }
  for (int e=0; e<12; e++){
    for (int m=0; m<o-1; m++){
      int i = (int) round((u[n+e*(o-1)+m] + o)/2.);
      int j = (int) round((v[n+e*(o-1)+m] + o)/2.);
      int k = (int) round((w[n+e*(o-1)+m] + o)/2.);
      idIJK[k*(o+1)*(o+1)+j*(o+1)+i] = n+e*(o-1)+m;
    }
  }
  n += 12*(o-1);

  // faces
  if (o > 1){
    for (int iquad=0; iquad<6; iquad++){
      int nodes[4][3];
      for (int i=0; i<4; i++){
	for (int j=0; j<3; j++){
	  nodes[i][j] = NToC[fToN[iquad][i]][j];
	}
      }
      int ndoftmp = MAd_QuadInterpolant::size(o-2);
      double* utmp = new double[ndoftmp];
      double* vtmp = new double[ndoftmp];
      double* wtmp = new double[ndoftmp];
      createQuad(o-2, o, nodes, utmp, vtmp, wtmp);
      for (int m=0; m< ndoftmp; m++){
	u[n+m] = utmp[m];
	v[n+m] = vtmp[m];
	w[n+m] = wtmp[m];
	int i = (int) round((u[n+m] + o)/2.);
	int j = (int) round((v[n+m] + o)/2.);
	int k = (int) round((w[n+m] + o)/2.);
	idIJK[k*(o+1)*(o+1)+j*(o+1)+i] = n+m;
      }
      n += ndoftmp;
      delete []utmp;
      delete []vtmp;
      delete []wtmp;
    }
  }

  // inside the hex, recursively
  if (o > 1){
    int nbdoftmp = size(o-2);
    double* utmp = new double[nbdoftmp];
    double* vtmp = new double[nbdoftmp];
    double* wtmp = new double[nbdoftmp];
    int* ijktmp = new int[nbdoftmp];
    nodePositionHex(o-2, utmp, vtmp, wtmp, ijktmp);
    for (int m=0; m< nbdoftmp; m++){
      u[n+m] = utmp[m]*(o-2);
      v[n+m] = vtmp[m]*(o-2);
      w[n+m] = wtmp[m]*(o-2);
      int i = (int) round((u[n+m] + o)/2.);
      int j = (int) round((v[n+m] + o)/2.);
      int k = (int) round((w[n+m] + o)/2.);
      idIJK[k*(o+1)*(o+1)+j*(o+1)+i] = n+m;
    }
    delete[] utmp;
    delete[] vtmp;
    delete[] wtmp;
    delete[] ijktmp;
    n += nbdoftmp;
  }
  for (int m=0; m<size(o); m++){
    u[m] = u[m]/o;
    v[m] = v[m]/o;
    w[m] = w[m]/o;
  }
  return;
}

// -----------------------------------------------------------------------------

int MAd_HexInterpolant::getIdxFromIJK(int i, int j, int k) const {
  int index = k*(order()+1)*(order()+1)+j*(order()+1)+i;
  if (index>= (int) ndof || index<0) throw;
  return idFromIJK[index];
}
 
// -----------------------------------------------------------------------------

void MAd_HexInterpolant::changeDir(int* idx, int o) const {
  
  int ndofq = quad->size();
  int* tmp = new int[ndofq];
  for (int k=0; k<ndofq; k++) tmp[k] = idx[k];
  for (int j=0; j<=o; j++){
    for (int i=0; i<=o; i++){
      if (i!=j) idx[quad->getIdxFromIJ(i,j)] = tmp[quad->getIdxFromIJ(j,i)];
    }
  }
  
  delete[] tmp;
  return;
}

// -----------------------------------------------------------------------------

void MAd_HexInterpolant::changeOrientation(int* idx, int rotate, int o) const {
  
  int ndofq = quad->size();
  int* tmp = new int[ndofq];
  for (int rot=0; rot<rotate; rot++) {
    for (int j=0; j<=o; j++){
      for (int i=0; i<=o; i++){
        tmp[quad->getIdxFromIJ(o-i,j)] = idx[quad->getIdxFromIJ(i,j)];
      }
    }
    for (int k=0; k<ndofq; k++) idx[k] = tmp[k];
    changeDir(idx,o);
  }
  delete[] tmp;
  return;
}

// -----------------------------------------------------------------------------

// create a quad with coordinates given by pt.
void MAd_HexInterpolant::createQuad(const int o, int ordhex, const int pt[4][3], 
					  double *u, double *v, double *w) {
  int n = 0;
  double mid[3];
  mid[0] = (pt[0][0]+pt[1][0]+pt[2][0]+pt[3][0])/4.;
  mid[1] = (pt[0][1]+pt[1][1]+pt[2][1]+pt[3][1])/4.;
  mid[2] = (pt[0][2]+pt[1][2]+pt[2][2]+pt[3][2])/4.;
  int scale[3];
  for (int k=0; k<3; k++) {
    if (mid[k]==0) scale[k] = o;
    else scale[k] = ordhex;
  }
  if(o==0) {
    u[n] = mid[0]*scale[0];
    v[n] = mid[1]*scale[1];
    w[n] = mid[2]*scale[2];
  }
  else if(o>0){
    //vertices
    for (int i=0; i<4; i++){
      u[i] = pt[i][0]*scale[0];  v[i] = pt[i][1]*scale[1];  w[i] = pt[i][2]*scale[2];
    }
    n += 4;
    //edges
    for (int i=0; i<4; i++){
      int dir[3];
      dir[0] = int(u[(i+1)%4]-u[i]);   dir[1] = int(v[(i+1)%4]-v[i]);   dir[2] = int(w[(i+1)%4]-w[i]);
      for (int j=0; j<(o-1); j++){
	u[n+j] = u[i]+(j+1)*dir[0]/scale[0];   v[n+j] = v[i]+(j+1)*dir[1]/scale[1];   w[n+j] = w[i]+(j+1)*dir[2]/scale[2];
      }
      n += o-1;
    }
    // inside, recursively
    if (o>1){
      int nbdoftmp = MAd_QuadInterpolant::size(o-2);
      double* utmp = new double[nbdoftmp];
      double* vtmp = new double[nbdoftmp];
      double* wtmp = new double[nbdoftmp];
      createQuad(o-2, ordhex, pt, utmp, vtmp, wtmp);
      for (int m=0; m< nbdoftmp; m++){
	u[n+m] = utmp[m];
	v[n+m] = vtmp[m];
	w[n+m] = wtmp[m];
      }
      delete[] utmp;
      delete[] vtmp;
      delete[] wtmp;
      n += nbdoftmp;
    }
  }
  return;
}

// -----------------------------------------------------------------------------

const int MAd_HexInterpolant::fToN[6][4] = {{0,3,2,1},{0,1,5,4},{1,2,6,5},
                                            {2,3,7,6},{3,0,4,7},{4,5,6,7}};

// -----------------------------------------------------------------------------

const int MAd_HexInterpolant::fToE[6][4] = {{3,2,1,0},{0,5,8,4},{1,6,9,5},
                                            {2,7,10,6},{3,4,11,7},{8,9,10,11}};

// -----------------------------------------------------------------------------

const int MAd_HexInterpolant::fToEdir[6][4] = {{-1,-1,-1,-1},{1,1,-1,-1},{1,1,-1,-1},
                                               {1,1,-1,-1},{1,1,-1,-1},{1,1,1,1}};

// -----------------------------------------------------------------------------

const int MAd_HexInterpolant::NToC[8][3] = {{-1,-1,-1},{1,-1,-1},{1,1,-1},{-1,1,-1},
                                            {-1,-1,1}, {1,-1,1}, {1,1,1}, {-1,1,1}};

// -----------------------------------------------------------------------------
