// -*- C++ -*- 
// -----------------------------------------------------------------------------
// MAdLiB - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#include "MAd_Interpolant.h"
#include "MAd_VertexInterpolant.h"
#include "MAd_EdgeInterpolant.h"
#include "MAd_TriInterpolant.h"
#include "MAd_QuadInterpolant.h"
#include "MAd_TetInterpolant.h"
#include "MAd_PrismInterpolant.h"
#include "MAd_PyramidInterpolant.h"
#include "MAd_HexInterpolant.h"

#include "MshTags.h"

#include <map>
using std::map;
using std::pair;
#include <cstdlib>
#include <limits>
using std::numeric_limits;
using std::min;
using std::max;
using std::string;
using std::ofstream;

// -----------------------------------------------------------------------------

bool MAd_Interpolant::computePointIndices(int iRot,int iSwp,
                                        const doubleMatrix& uvwTfo,
                                        int*& renum,
                                        double tolerance) const {

  map<int,int>  correspondingNode;
  double maxError = 0;

  doubleMatrix uvwOri(uvwTfo.size1(),uvwTfo.size2());

  if (!findOriginalCoordinates(iRot,iSwp,uvwTfo,uvwOri)) return false;

  int nb = int(uvwTfo.size1());

  for (int i=0;i<nb;i++) {
        
    double error = numeric_limits<double>::max();
    int index = -1;
    
    for (int j=0;j<nb;j++) {
      double err = 0;
      for (int k=0;k<int(dim());k++) { 
        double du = uvwTfo(j,k) - uvwOri(i,k);
        err += du * du;
      }
      if (err < error) {
        error = err;
        index = j;
      }
    }
    
    if (index == -1) return false;

    if (correspondingNode.find(index) != correspondingNode.end()) return false;
    correspondingNode[index] = i;
    maxError = max(maxError,error);
  }
  
  if (sqrt(maxError) > tolerance) return false;

  map<int,int>::iterator rIter = correspondingNode.begin();
  for (;rIter!=correspondingNode.end();rIter++) renum[rIter->second] = rIter->first;
  return true;
}

//------------------------------------------------------------------------------

double MAd_Interpolant::computeLebesgueConstant() const{
  
  double Lebesgue_cst; // returned value
  double eps = numeric_limits<double>::epsilon(); // machine epsilon

  size_t myDim = dim(); //  1 or 2 or 3D

  int M = 100;// number of sample points

  int ten = M/10; // number of selection

  double **X = new double* [M];
  for(int i=0; i<M; i++) X[i] = new double [3];
  
  // random sampling
  for (int i=0; i<M; i++)
    {
      switch (myDim){
        
      case 3:
        X[i][2] = 2 * ((double)(rand())) / RAND_MAX - 1;
      case 2:
        X[i][1] = 2 * ((double)(rand())) / RAND_MAX - 1;
      case 1:
        X[i][0] = 2 * ((double)(rand())) / RAND_MAX - 1;
        break; 
      }
    }


  int N = size(); // number of shape functions
  doubleVector f(N);
  double lambda = 0.0; // local value of Lebesgue fct

  map<double,double*> localLebesgue;
  
  for(size_t count=0; count<20; count++){
    
    for (int i=0; i<M; i++){
      if (inReferenceElement(X[i][0],X[i][1],X[i][2])){
        fcts(X[i][0],X[i][1],X[i][2],f);
        // computation of the local value of Lebesgue fct
        for (int j=0; j<N; j++) lambda += fabs(f(j));
        // ensuring there are different values for key values (cf. map specifications)
        double pert = 100*eps*(2* ((double) rand() )/RAND_MAX - 1); 
        // sorting with corresponding coordinates
        localLebesgue.insert(pair<double,double*>(lambda+pert,X[i]) ); 
      }
    }

    // selection of the 10% of points giving the largest local values of Lebesgue fct
    int myCount = 0;
    for(map<double,double*>::reverse_iterator it = localLebesgue.rbegin() ; 
        myCount < ten ; it++){

      double* myX = it->second; // coordinates
      // we are going to re-use the memory of too small values
      map<double,double*>::reverse_iterator tooSmall = localLebesgue.rbegin(); // useless points
      for(int t = 0 ; t<(ten+10*myCount) ; t++)	tooSmall++;
      
      // we compute 10 new points close to points giving the largest local values of Lebesgue fct
      for(size_t ii=0; ii<10; ii++){
        
        double* myPoint = tooSmall->second;
        
        myPoint[0] = myX[0] + .1/M * (2 * ((double)(rand())) / RAND_MAX - 1);
        myPoint[1] = myX[1] + .1/M * (2 * ((double)(rand())) / RAND_MAX - 1);
        myPoint[2] = myX[2] + .1/M * (2 * ((double)(rand())) / RAND_MAX - 1);
        
        tooSmall++;
      }
      
      myCount++;
    }
    Lebesgue_cst = localLebesgue.rbegin()->first; // we register the current Lebesgue cst
    localLebesgue.clear(); // we clear the structure for next iteration
  }

  // free memory :-)
  for(int i=0; i<M; i++) delete [] X[i];
  
  delete []X;

  return Lebesgue_cst;
  
}// end function


//------------------------
bool MAd_Interpolant::inReferenceElement(double u, double v, double w) const {

  double eps = numeric_limits<double>::epsilon();

  switch(getElementType())
    {
    case MSH_LIN_2 :
      if(u < -1.-eps || u > 1+eps)return false;
      break;
    case MSH_TRI_3 : 
      if(u < -eps || v < -eps || 1.-u-v < -eps)return false;
      break;
    case MSH_TET_4 :
      if(u < -eps || v < -eps || w < -eps || 1.-u-v-w < -eps)return false;
      break;
    case MSH_QUA_4 :
      if((u < -1.-eps) ||(u > 1.+eps) ||(v > 1.+eps) ||(v < -1.-eps))return false;
      break; 
    case MSH_PYR_5 :
      if(( 1.-u-v-w < -eps) || ( 1.-u+v-w < -eps) || ( 1.+u+v-w < -eps) || ( 1.+u-v-w < -eps) || w < -eps) return false;
    case MSH_PRI_6 :
      if((w < -1.-eps || w > 1.+eps) || (u < -eps || v < -eps || 1.-u-v < -eps)) return false;
      break; 
    case MSH_HEX_8 :
      if( ((u < -1.-eps) ||(u > 1.+eps) ||(v > 1.+eps) ||(v < -1.-eps)) || (w < -1.-eps || w > 1+eps)) return false;
      break;
    default: break; // need to throw an exception here
    }
  return true;
}

