// -*- C++ -*- 
// -----------------------------------------------------------------------------
// MAdLiB - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------


#ifndef MADINTERPOLANT_H
#define MADINTERPOLANT_H

#include "MAdMatrix.h"
using namespace MAd; 
#include "MshTags.h"

#include <map>

// -----------------------------------------------------------------------------
/*! \brief Base class for all shape function classes  */ 
// -----------------------------------------------------------------------------

class MAd_Interpolant {
  
public:
  
  MAd_Interpolant(const int &t,const int& o):typeE(t),p(o) {};
  virtual ~MAd_Interpolant(){};

public:

  /*! \brief get interpolation order */
  int order() const {return p;};

  /*! \brief get basic element type (corresponding linear type) */
  int getElementType() const {return typeE;}

  /*! \brief get number of shape functions / control points */
  virtual int size () const = 0;

  /*! \brief get dimension of parametric coordinate system */
  virtual size_t dim() const = 0;

  /*! \brief evaluate shape functions in (u,v,w) */
  virtual void fcts(const double& u,const double& v,const double& w, 
                    doubleVector& f) const = 0;

  /*! \brief evaluate shape function gradients in (u,v,w) */
  virtual void grads(const double& u,const double& v,const double& w,
                     doubleVector gr[3]) const = 0;
  
  /*! \brief get coefficients of i-th interpolant to j-th monomial */
  virtual double getCoef ( int i, int j) const {return coef (i,j);}

  /*! \brief clone the interpolant at another order */
  virtual MAd_Interpolant *clone(int p) const = 0;
  
  virtual void getMonomialExponents(int iMonomial , 
                                    int & powerOfU,  
                                    int & powerOfV,  
                                    int & powerOfW) const=0;
public:
  
  /*! \brief recompute the indices of a rotationally symmetric set of points
    in a rotated element in the original element */
  
  bool computePointIndices(int iRot, //!< index of the first node 
                           int iOri, //!< relative orientation 
                           const doubleMatrix& uvw, //!< parametric coordinates
                           int*& renum, //!< original node for this orientation 
                           double tolerance=0. //!< tolerance applied
                           ) const;

  /*! \brief in place computation of the coordinates of a rotated element in
      the original element */
  virtual 
  bool findOriginalCoordinates(int iRot, //!< index of the first node
                               int iOri, //!< relative orientation
                               const doubleMatrix& tfo,//!< transformed coordinates
                               doubleMatrix& ori //!< original coordinates 
                               ) const {throw;}

  /*! \brief return the Lebesgue constant of the nodes related to the
     function space class */
  double computeLebesgueConstant() const;

  /*! \brief check whether (u,v,w) is inside the reference element */
  bool inReferenceElement(double u, double v, double w) const;
  
  virtual bool reorientData(doubleMatrix&,int) const {return false;}
  virtual int nbRotations()    const {abort(); return -1;}
  virtual int nbOrientations() const {abort(); return -1;}

  /*! \brief get parametric coordinates (u,v,w) of control point i */
  virtual void nodeNumberCoordinate(const int& i, 
                                    double& u, 
                                    double& v, 
                                    double& w) const=0;

protected:
  const int typeE; //!< basic element type
  const int p;     //!< interpolation order 
  doubleMatrix coef; 

};

#endif            
