// -*- C++ -*- 
// -----------------------------------------------------------------------------
// MAdLiB - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------


#include <cmath>
#include <algorithm>
using std::swap;


#include "MAd_PrismInterpolant.h"

//---------------------------------------------------------------------
// Prism
/*
 *
 *               w
 *               ^
 *               |
 *               3
 *             ,/|`\
 *           ,/  |  `\
 *         ,/    |    `\
 *        4------+------5
 *        |      |      |
 *        |    ,/|`\    |
 *        |  ,/  |  `\  |
 *        |,/    |    `\|              order 1
 *       ,|      |      `\
 *     ,/ |      0      | `\
 *    u   |    ,/ `\    |    v
 *        |  ,/     `\  |
 *        |,/         `\|
 *        1-------------2
 *
 *
 *               w
 *               ^
 *               |
 *               3
 *             ,/|`\
 *            9  16 11
 *         ,/    |    `\
 *        4------10-----5
 *        |      12     |
 *        |    ,/| `\   |
 *        |  17  20  19 |
 *        |,/    |    `\|             order 2
 *       ,13     18     14
 *     ,/ |      0      | `\
 *    u   |    ,/ `\    |    v
 *        |   6  15 8   |
 *        |,/         `\|
 *        1------7------2
 *
 */

//-------------------------------
//  Public Members
//-------------------------------

MAd_PrismInterpolant::MAd_PrismInterpolant(const int &o)
  : MAd_Interpolant(MSH_PRI_6, o), ndof(size(o)), lefs(o), ndofe(size_t(lefs.size())), ltfs(o), ndoft(size_t(ltfs.size())) {

  NodeNumberu = new double[ndof];
  NodeNumberv = new double[ndof];
  NodeNumberw = new double[ndof];
  idTri = new int[ndof];
  nodePositionPrism(order(), order(), NodeNumberu, NodeNumberv, NodeNumberw);
  getIdTri(idTri);

  coef.resize(int(ndof),int(ndof));
  for (int i=0; i< int(ndof); i++) for (int j=0; j< int(ndof); j++) coef(i,j) = getCoef(i,j);
  
}

MAd_PrismInterpolant::~MAd_PrismInterpolant() {
  delete[] NodeNumberu;
  delete[] NodeNumberv;
  delete[] NodeNumberw;
  delete[] idTri;
  return;
}

void MAd_PrismInterpolant::fcts(const double &u, const double &v, const double &w, 
                                      doubleVector &f) const {
  const int n=int(ndof);
  doubleVector m(n);
  f = 0; 
  for (int i=0; i< int(ndof); i++) m(i) = monomeof(u,v,w,i);
  coef.mult(m,f,false);
  return;
}

void MAd_PrismInterpolant::grads(const double &u, const double &v, const double &w, 
                                       doubleVector gr[3]) const {
  const int n=int(ndof);
  doubleVector m(n);
  for (int k=0; k< 3; k++){
    gr[k] = 0;
    for (int i=0; i< int(ndof); i++) m(i) = dmonomeof(u,v,w,i,k);
    coef.mult(m,gr[k],false);
  }
  return;
}

void MAd_PrismInterpolant::getIdTri(int* id) const {
  size_t found=0;
  for (int i=0; i< int(ndof); i++){
    for (int j=0; j< int(ndoft); j++){
      double ut,vt,wt;
      ltfs.nodeNumberCoordinate(j, ut, vt, wt);
      if (NodeNumberu[i]==ut && NodeNumberv[i]==vt) {
        id[i]=j;
        found++;
      }
    }
  }
  if (found!=ndof) throw;
}

double MAd_PrismInterpolant::getCoef(int id1,int id2) const {
  int i1,i2;
  i1 = (int)(.5*(NodeNumberw[id1]*order()+order()));
  i2 = (int)(.5*(NodeNumberw[id2]*order()+order()));
  double coefe = MAd_EdgeInterpolant::getCoef(order(), i1, i2);
  double coeft = MAd_TriInterpolant::getCoef(order(), idTri[id1], idTri[id2]);
  return coefe*coeft;
}

void MAd_PrismInterpolant::getMonomialExponents(int i, int &powerU, int &powerV, int &powerW) const {
  ltfs.getMonomialExponents(idTri[i], powerU, powerV, powerW);
  powerW = (int)(.5*(NodeNumberw[i]*order()+order()));
  return;
}

double MAd_PrismInterpolant::monomeof(const double &u, const double &v, const double &w, int i) const{
  int powerU, powerV, powerW;
  getMonomialExponents(i, powerU, powerV, powerW);
  return pow(u,powerU)*pow(v,powerV)*pow(w,powerW);
};

double MAd_PrismInterpolant::dmonomeof(const double &u, const double &v, const double &w, int i, int k) const{
  double Zero=0.;
  int powerU, powerV, powerW;
  getMonomialExponents(i, powerU, powerV, powerW);
  switch (k){
  case 0:
    if (powerU > 0)
      return powerU * pow(u,powerU-1)*pow(v,powerV)*pow(w,powerW);
    else return Zero;
    break;
  case 1:    
    if (powerV > 0)
      return powerV * pow(u,powerU)*pow(v,powerV-1)*pow(w,powerW);
    else return Zero;
    break;
  case 2:    
    if (powerW > 0)
      return powerW * pow(u,powerU)*pow(v,powerV)*pow(w,powerW-1);
    else return Zero;
    break;
  default:
    throw;
    break;
  }
}

// size_t MAd_PrismInterpolant::getClosureSpaceSize (int iFace) const {
//   if (iFace<2 && iFace>=0) return MAd_TriInterpolant::size(order());
//   if (iFace<5 && iFace>=2) return MAd_QuadInterpolant::size(order());
//   return MAd_QuadInterpolant::size(order());
// }

// const int* MAd_PrismInterpolant::getClosure (int iFace, int iSign, int iRotate) const {
//   if (iFace<2) return OrientedFaceDofTri[getCodeFromOrientation(iFace, iSign, iRotate)/*,0*/];
//   else return OrientedFaceDofQuad[getCodeFromOrientation(iFace, iSign, iRotate)/*,0*/];
// }

// // -----------------------------------------------------------------------------
// void MAd_PrismInterpolant::getClosureSpace(const int &iFace,
//                                                  const int &sign, 
//                                                  const int &rotate,
//                                                  int* facedofs) const {
//   if(order()==0) facedofs[0]=0;
//   else {
//     int n = 0;
//     int stri = MAd_TriInterpolant::size(order()-3);
//     int squad = MAd_QuadInterpolant::size(order()-2);
//     if (iFace<2) {
//       // vertices
//       for (size_t i=0; i<3; i++) facedofs[i]=fToNTri[iFace][i];
//       n += 3;
      
//       //edges
//       for (int ie=3*iFace; ie<3*(iFace+1); ie++){
//         for (int k=0; k<order()-1; k++){
//           facedofs[n+k] = 6+ie*(order()-1)+k;
//         }
//         n +=order()-1;
//       }
      
//       //inside
//       for (int k=0; k<stri; k++){
//         facedofs[n+k] = 6+9*(order()-1)+iFace*stri+k;
//       }
//       n +=stri;
      
//       // update rotation and direction
//       if (rotate>0) changeOrientationTri(facedofs,rotate,order());
//       if (sign<=0) changeDirTri(facedofs,order());
//     }
//     else {
//       // vertices
//       for (size_t i=0; i<4; i++) facedofs[i]=fToNQuad[iFace-2][i];
//       n += 4;
      
//       //edges
//       for (int ie=0; ie<4; ie++){
//         for (int k=0; k<order()-1; k++){
//           int km = k;
//           if (fToEdirQuad[iFace-2][ie]==-1) km = order()-2-k;
//           facedofs[n+k] = 6+fToEQuad[iFace-2][ie]*(order()-1)+km;
//         }
//         n +=order()-1;
//       }

//       //inside
//       for (int k=0; k<squad; k++){
//         // offsetNode + offsetEdges + offsetFaceTri + offsetFaceQuad
//         facedofs[n+k] = 6+9*(order()-1)+2*stri+(iFace-2)*squad+k;
//       }
//       n +=squad;

//       // update rotation and direction
//       if (rotate>0) changeOrientationQuad(facedofs,rotate,order());
//       if (sign<=0) changeDirQuad(facedofs,order());
      
//     }
//   }
//   return;
// }

void MAd_PrismInterpolant::nodeNumberCoordinate(const int &i, double &u, double &v, double &w) const{


  u=NodeNumberu[i];
  v=NodeNumberv[i];
  w=NodeNumberw[i];
  return;
}

int MAd_PrismInterpolant::size(int q){
  return (q+1)*(q+1)*(q+2)/2;
}




//-------------------------------
//  Private Members
//-------------------------------
void MAd_PrismInterpolant::nodePositionPrism(int OTri, int OQuad, double *u, double *v, double *w) const {
  
  int nbDof = int((OQuad+1)*(OTri+1)*(OTri+2)/2.);
  int n = 0;
  if (OQuad==0 && OTri==0) {
    u[0] = 0.;  v[0] = 0.;  w[0] = 0.; 
    return;
  }
  else if (OTri>0 && OQuad>0){
    // vertices
    u[0] = 0.;       v[0] = 0.;       w[0] = -OQuad;
    u[1] = OTri;     v[1] = 0.;       w[1] = -OQuad;
    u[2] = 0.;       v[2] = OTri;     w[2] = -OQuad;
    
    u[3] = 0.;       v[3] = 0.;       w[3] = OQuad;
    u[4] = OTri;     v[4] = 0.;       w[4] = OQuad;
    u[5] = 0.;       v[5] = OTri;     w[5] = OQuad;
    n += 6;
    
    // edges tri
    for (int k=0; k<(OTri-1); k++){
      u[6+k] = k+1;                        v[6+k] = 0.;                            w[6+k] = -OQuad;
      u[6+OTri-1+k] = OTri-(k+1);          v[6+OTri-1+k] = k+1;                    w[6+OTri-1+k] = -OQuad;
      u[6+2*(OTri-1)+k] = 0. ;             v[6+2*(OTri-1)+k] = OTri-(k+1);         w[6+2*(OTri-1)+k] = -OQuad;
      
      u[6+3*(OTri-1)+k] = k+1;            v[6+3*(OTri-1)+k] = 0.;             w[6+3*(OTri-1)+k] = OQuad;
      u[6+4*(OTri-1)+k] = OTri-(k+1);     v[6+4*(OTri-1)+k] = k+1;            w[6+4*(OTri-1)+k] = OQuad;
      u[6+5*(OTri-1)+k] = 0. ;            v[6+5*(OTri-1)+k] = OTri-(k+1);     w[6+5*(OTri-1)+k] = OQuad;
    }
    n += 6*(OTri-1);
    
    // edges quad
    for (int k=0; k<(OQuad-1); k++){
      u[n+k] = 0.;                v[n+k] = 0.;                w[n+k] = 2*(k+1)-OQuad;
      u[n+(OQuad-1)+k] = OTri;    v[n+(OQuad-1)+k] = 0.;      w[n+(OQuad-1)+k] = 2*(k+1)-OQuad;
      u[n+2*(OQuad-1)+k] = 0. ;   v[n+2*(OQuad-1)+k] = OTri;  w[n+2*(OQuad-1)+k] = 2*(k+1)-OQuad;
    }
    n += 3*(OQuad-1);
  }
  else if (OTri==0){
    // vertices
    u[0] = 0.;       v[0] = 0.;       w[0] = -OQuad;
    u[1] = 0.;       v[1] = 0.;       w[1] = OQuad;
    n += 2;
    // edge
    for (int k=0; k<(OQuad-1); k++){
      u[6+6*(OQuad-1)+k] = 0.;            v[6+6*(OQuad-1)+k] = 0.;            w[6+6*(OQuad-1)+k] = 2*(k+1)-OQuad;
    }
    n += OQuad-1;
  }
  else if (OQuad==0){
    // vertices
    u[0] = 0.;       v[0] = 0.;       w[0] = 0.;
    u[1] = OTri;     v[1] = 0.;       w[1] = 0.;
    u[2] = 0.;       v[2] = OTri;     w[2] = 0.;
    n += 3;
    // edges tri
    for (int k=0; k<(OTri-1); k++){
      u[6+k] = k+1;                        v[6+k] = 0.;                            w[6+k] = 0.;
      u[6+OTri-1+k] = OTri-(k+1);          v[6+OTri-1+k] = k+1;                    w[6+OTri-1+k] = 0.;
      u[6+2*(OTri-1)+k] = 0. ;             v[6+2*(OTri-1)+k] = OTri-(k+1);         w[6+2*(OTri-1)+k] = 0.;
    }
    n += OTri-1;
  }
  
  int oTri = OTri-3;
  int oQuad = OQuad-2;
  int oQuadTri = OTri-2; //edges between tri and quad

  // face tri
  if (oTri >= 0){
    for (int itri=0; itri<2; itri++){
      MAd_TriInterpolant tri(oTri);
      int nbdoftmp = tri.size();
      for (int k=0; k< nbdoftmp; k++){
        double utmp, vtmp, wtmp;
        tri.nodeNumberCoordinate(k, utmp, vtmp, wtmp);
        u[n+k] = utmp*oTri+1.;
        v[n+k] = vtmp*oTri+1.;
        w[n+k] = OQuad*pow(-1,itri+1);
      }
      n += nbdoftmp;
    }
  }
  
  // face quad
  if (oQuad >= 0 && oQuadTri >= 0){
    for (int iquad=0; iquad<3; iquad++){
      int nodes[4][3];
      for (int i=0; i<4; i++){
        for (int j=0; j<3; j++){
          nodes[i][j] = NToCQuad[fToNQuad[iquad][i]][j];
        }
      }
      int offsetu = int((nodes[0][0]+nodes[1][0]+nodes[2][0]+nodes[3][0])/2.);
      int offsetv = int((nodes[0][1]+nodes[1][1]+nodes[2][1]+nodes[3][1])/2.);
      int nbdoftmp = (oQuad+1)*(oQuadTri+1);
      double* utmp = new double[nbdoftmp];
      double* vtmp = new double[nbdoftmp];
      double* wtmp = new double[nbdoftmp];
      createQuadPrism(oQuad, oQuadTri, OQuad, OTri, nodes, utmp, vtmp, wtmp);
      for (int k=0; k< nbdoftmp; k++){
        u[n+k] = utmp[k]*oQuadTri+offsetu;
        v[n+k] = vtmp[k]*oQuadTri+offsetv;
        w[n+k] = wtmp[k]*oQuad;
      }
      n += nbdoftmp;
      delete []utmp;
      delete []vtmp;
      delete []wtmp;
    }
  }
  
  // inside the prism, recursively
  if (oQuad >= 0 && oTri >= 0){
    int nbdoftmp = int((oQuad+1)*(oTri+1)*(oTri+2)/2.);
    double* utmp = new double[nbdoftmp];
    double* vtmp = new double[nbdoftmp];
    double* wtmp = new double[nbdoftmp];
    nodePositionPrism(oTri, oQuad, utmp, vtmp, wtmp);
    for (int k=0; k< nbdoftmp; k++){
      u[n+k] = utmp[k]*oTri+1.;
      v[n+k] = vtmp[k]*oTri+1.;
      w[n+k] = wtmp[k]*oQuad;
    }
    delete []utmp;
    delete []vtmp;
    delete []wtmp;
  }
  
  for (int k=0; k< nbDof; k++){
    if (OTri!=0){
      u[k] = u[k]/OTri;
      v[k] = v[k]/OTri;
    }
    if (OQuad!=0) w[k] = w[k]/OQuad;
  }
  //if (OQuad==OTri) throw;
  return;
}

// create a quad with coordinates given by pt. The order of the edges can be different (for prism faces)
void MAd_PrismInterpolant::createQuadPrism(int oQuad, int oTri, int OQuad, int OTri, const int pt[4][3], 
                                                 double *u, double *v, double *w) {
  int n = 0;
  double mid[3];
  mid[0] = (pt[0][0]+pt[1][0]+pt[2][0]+pt[3][0])/4.;
  mid[1] = (pt[0][1]+pt[1][1]+pt[2][1]+pt[3][1])/4.;
  mid[2] = (pt[0][2]+pt[1][2]+pt[2][2]+pt[3][2])/4.;
  int scale[3];
  for (int k=0; k<2; k++) {
    if (mid[k]==0) scale[k] = OTri;
    else scale[k] = oTri;
  }
  if (mid[2]==0) scale[2] = oQuad;
  else scale[2] = OQuad;
  int offsetu = int((pt[0][0]+pt[1][0]+pt[2][0]+pt[3][0])/2.);
  int offsetv = int((pt[0][1]+pt[1][1]+pt[2][1]+pt[3][1])/2.);
  if(oQuad==0 && oTri==0) {
    u[n] = mid[0]*scale[0];
    v[n] = mid[1]*scale[1];
    w[n] = mid[2]*scale[2];
  }
  else if(oQuad==0 && oTri>0){
    for (int i=0; i<4; i++){
      u[i] = pt[i][0]*scale[0];   v[i] = pt[i][1]*scale[1];   w[i] = mid[2]*scale[2];
    }
  }
  else if(oTri==0 && oQuad>0){
    for (int i=0; i<4; i++){
      u[i] = mid[0]*scale[0];   v[i] = mid[1]*scale[1];   w[i] = pt[i][2]*scale[2];
    }
  }
  else if(oQuad>0 && oTri>0){
    //vertices
    for (int i=0; i<4; i++){
      u[i] = pt[i][0]*scale[0];  v[i] = pt[i][1]*scale[1];  w[i] = pt[i][2]*scale[2];
    }
    n += 4;
    //edges
    for (int i=0; i<4; i++){
      int dir[3];
      dir[0] = int((u[(i+1)%4]-u[i])/scale[0]);   dir[1] = int((v[(i+1)%4]-v[i])/scale[1]);   dir[2] = int((w[(i+1)%4]-w[i])/scale[2]);
      int o = oTri;
      if (dir[2]==0) o = oQuad;
      for (int j=0; j<(o-1); j++){
        u[n+j] = u[i]+(j+1)*dir[0];   v[n+j] = v[i]+(j+1)*dir[1];   w[n+j] = w[i]+(j+1)*dir[2];
      }
      n += o-1;
    }
    // inside, recursively
    if (oQuad>1 && oTri>1){
      int nbdoftmp = MAd_EdgeInterpolant::size(oQuad-2)*MAd_EdgeInterpolant::size(oTri-2);
      double* utmp = new double[nbdoftmp];
      double* vtmp = new double[nbdoftmp];
      double* wtmp = new double[nbdoftmp];
      createQuadPrism(oQuad-2, oTri-2, OQuad, OTri, pt, utmp, vtmp, wtmp);
      for (int m=0; m< nbdoftmp; m++){
        u[n+m] = utmp[m]*(oTri-2)+offsetu;
        v[n+m] = vtmp[m]*(oTri-2)+offsetv;
        w[n+m] = wtmp[m]*(oQuad-2);
      }
      delete[] utmp;
      delete[] vtmp;
      delete[] wtmp;
      n += nbdoftmp;
    }
  }
  for (int k=0; k< n; k++){
    if (oTri!=0){
      u[k] = u[k]/oTri;
      v[k] = v[k]/oTri;
    }
    if (oQuad!=0) w[k] = w[k]/oQuad;
  }
  return;
}
void MAd_PrismInterpolant::changeDirTri(int* idx, int o) const {
  int stri = MAd_TriInterpolant::size(o);
  int* tmp = new int[stri];
  int n=0;
  int locOrder = o;
  while (locOrder>=0){
    if (locOrder==0) tmp[n] = idx[n];
    else {
      // vertices
      tmp[n] = idx[n]; tmp[n+1] = idx[n+2]; tmp[n+2] = idx[n+1];
      // edges
      for (int k=0; k<(locOrder-1); k++){
        tmp[n+3+k] = idx[n+3+3*(locOrder-1)-(k+1)]; 
        tmp[n+3+(locOrder-1)+k] = idx[n+3+2*(locOrder-1)-(k+1)]; 
        tmp[n+3+2*(locOrder-1)+k] = idx[n+3+(locOrder-1)-(k+1)];
      }
    }
    // inside, recursively
    n = n+3*locOrder;
    locOrder = locOrder-3;
  }
  for (int k=0; k<stri; k++) idx[k] = tmp[k];

  delete[] tmp;
  return;
}

void MAd_PrismInterpolant::changeOrientationTri(int* idx, int rotate, int o) const {
  int stri = MAd_TriInterpolant::size(o);
  int* tmp = new int[stri];
  int n=0;
  int locOrder = o;
  while (locOrder>=0){
    if (locOrder==0) tmp[n] = idx[n];
    else {
      for (int edge=0; edge<3; edge++){
        // vertices
        tmp[n+edge] = idx[n+(edge+rotate)%3];
        // edges
        for (int k=0; k<(locOrder-1); k++) tmp[n+3+edge*(locOrder-1)+k] = idx[n+3+((edge+rotate)%3)*(locOrder-1)+k];
      }
    }
    // inside, recursively
    n = n+3*locOrder;
    locOrder = locOrder-3;
  }
  for (int k=0; k<stri; k++) idx[k] = tmp[k];
  delete[] tmp;
  return;
}

void MAd_PrismInterpolant::changeDirQuad(int* idx, int o) const {
  int squad = MAd_QuadInterpolant::size(o);
  int* tmp = new int[squad];
  int n=0;
  int locOrder = o;
  while (locOrder>=0){
    if (locOrder==0) tmp[n] = idx[n];
    else {
      // vertices
      tmp[n] = idx[n]; tmp[n+1] = idx[n+3]; tmp[n+2] = idx[n+2]; tmp[n+3] = idx[n+1];
      // edges
      for (int k=0; k<(locOrder-1); k++){
        tmp[n+4+k] = idx[n+4+4*(locOrder-1)-(k+1)]; 
        tmp[n+4+(locOrder-1)+k] = idx[n+4+3*(locOrder-1)-(k+1)]; 
        tmp[n+4+2*(locOrder-1)+k] = idx[n+4+2*(locOrder-1)-(k+1)];
        tmp[n+4+3*(locOrder-1)+k] = idx[n+4+(locOrder-1)-(k+1)];
      }
    }
    // inside, recursively
    n = n+4*locOrder;
    locOrder = locOrder-2;
  }
  for (int k=0; k<squad; k++) idx[k] = tmp[k];

  delete[] tmp;
  return;
}

void MAd_PrismInterpolant::changeOrientationQuad(int* idx, int rotate, int o) const {
  int squad = MAd_QuadInterpolant::size(o);
  int* tmp = new int[squad];
  int n=0;
  int locOrder = o;
  while (locOrder>=0){
    if (locOrder==0) tmp[n] = idx[n];
    else {
      for (int i=0; i<4; i++){
        // vertices
        tmp[n+i] = idx[n+(i+rotate)%4];
        // edges
        for (int k=0; k<(locOrder-1); k++) tmp[n+4+i*(locOrder-1)+k] = idx[n+4+((i+rotate)%4)*(locOrder-1)+k];
      }
    }
    // inside, recursively
    n = n+4*locOrder;
    locOrder = locOrder-2;
  }
  for (int k=0; k<squad; k++) idx[k] = tmp[k];
  delete[] tmp;
  return;
}

int MAd_PrismInterpolant::fToNTri[2][3] = {{0,1,2},{3,4,5}};
int MAd_PrismInterpolant::fToNQuad[3][4] = {{0,3,4,1},{1,4,5,2},{3,0,2,5}};
int MAd_PrismInterpolant::fToEQuad[3][4] = {{6,3,7,0},{7,4,8,1},{6,2,8,5}};
int MAd_PrismInterpolant::fToEdirQuad[3][4] = {{1,1,-1,-1},{1,1,-1,-1},{-1,-1,1,1}};
int MAd_PrismInterpolant::NToCQuad[6][3] = {{0,0,-1},{1,0,-1},{0,1,-1},
                                                  {0,0,1},{1,0,1},{0,1,1}};
