// -*- C++ -*- 
// -----------------------------------------------------------------------------
// MAdLiB - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------


#include "MAd_PyramidInterpolant.h"
#include "MAd_TriInterpolant.h"
#include "MAd_QuadInterpolant.h"

#include <cmath>
#include <limits>

#include "MAdVector3D.h"

// -----------------------------------------------------------------------------

void computeFaceTransformation(size_t nb,
                               const std::vector<MAd_Vector3D>&  corners,
                               const MAd_Interpolant* fs,
                               const MAd_Interpolant* bfs,
                               int rot,int swap,
                               int* dof) {



  MAd_Interpolant* linear = bfs->clone(1);


  // --- compute positions of oriented face

  std::vector<MAd_Vector3D> oriented(nb);

  for (size_t i=0;i<nb;i++) {
    size_t j = (i+rot)%nb;
    oriented[i] = corners[j];
  }

  // --- only triangles and quadrangles -> swap only requires switching 1 w/ 2 or 1 w/ 3 resp.

  if (swap) {
    MAd_Vector3D tmp(oriented[1]);
    oriented[1] = oriented[nb-1];
    oriented[nb-1] = tmp;
  }

  // --- coordinates of face points

  doubleVector linearFcts(linear->size());
  doubleVector elmtFcts(fs->size());
  double eps(100.*std::numeric_limits<double>::epsilon());

  for (int i=0;i<bfs->size();i++) { 
    double u,v,w;
    bfs->nodeNumberCoordinate(i,u,v,w);

    linear->fcts(u,v,w,linearFcts);

    MAd_Vector3D pos(0,0,0);

    for (int j=0;j<linearFcts.size();j++) {
      for (size_t d=0;d<3;d++) pos[d] += oriented[j][d]*linearFcts(j);
    }
    
    fs->fcts(pos[0],pos[1],pos[2],elmtFcts);

    int found = 0;
    for (int j=0;j<elmtFcts.size();j++) {
      if (fabs(elmtFcts(j)) > eps) {
        dof[i] = j;
        found++;
      }
    }

    if (found != 1) {
      /*      if (found == 0) {
              rm.fatal(__LINE__,__FILE__,false,
              "ComputeFaceTransformation: no corresponding points found for dof %d on bounding face",i);
              }
              else {
              rm.fatal(__LINE__,__FILE__,false,
              "ComputeFaceTransformation: too many (%d) corresponding points found for dof %d on face",
              found,i);
              }*/
    }
  }
  delete linear;
}


// -----------------------------------------------------------------------------
//  definition of edges and faces in the pyramid template
// -----------------------------------------------------------------------------

const int MAd_PyramidInterpolant::edgeDef[8][2] = {{0,1},{0,3},{0,4},
                                                   {1,2},{1,4},
                                                   {2,3},{2,4},
                                                   {3,4}};

const int MAd_PyramidInterpolant::triaDef[4][3] = {{0,1,4},{3,0,4},//{0,4,3},
                                                   {1,2,4},
                                                   {3,2,4}};//{2,3,4}

const int MAd_PyramidInterpolant::quadDef[4]    = {0,1,2,3};

const double MAd_PyramidInterpolant::vCoord[5][3] = {{-1,-1,0},{1,-1,0},{1,1,0},{-1,1,0},{0,0,1}};

// -----------------------------------------------------------------------------
MAd_PyramidInterpolant::MAd_PyramidInterpolant(const int &o): 
  MAd_Interpolant(MSH_PYR_5, o),
  nDof(size(o)),
  uvw(NULL),
  bergot(new Bergot(o)),
  iOrder(NULL),
  jOrder(NULL),
  kOrder(NULL),
  dOrder(NULL),
  nbDofTriangle  ((o+1)*(o+2)/2),
  nbDofQuadrangle((o+1)*(o+1)),
  quadDof(NULL),
  triaDof(NULL) {
  
  // --- compute point positions

  uvw = new double[nDof][3];
  computePoints(o,uvw);

  // --- determine transformation matrix from Monomial basis toward Lagrangian one

  coef.resize(nDof,nDof);

  iOrder = new int[nDof];
  jOrder = new int[nDof];
  kOrder = new int[nDof];
  dOrder = new int[nDof];

  for (int j=0;j<nDof;j++){
    int pu, pv, pw;
    bergot->polynomialOrders(j,pu,pv,pw);
    iOrder[j] = pu;
    jOrder[j] = pv;
    kOrder[j] = pw;
    dOrder[j] = std::max(pu,pv) - pu - pv;
  }
  

  for (int i=0;i<nDof;i++) {
    
    double u = uvw[i][0];
    double v = uvw[i][1];
    double w = uvw[i][2];
    double oneMinW = w == 1 ? 1e-12 : 1-w;
    
    for (int j=0;j<nDof;j++){
      coef(j,i) = (std::pow(u,iOrder[j])*
                   std::pow(v,jOrder[j])*
                   std::pow(w,kOrder[j])*
                   std::pow(oneMinW,dOrder[j]));
    }
  }

  coef.invertInPlace();
  
  // --- now compute the correspondence of quad and element nodes

  MAd_QuadInterpolant quad(order());

  quadDof = new int[8*quad.size()];

  std::vector<MAd_Vector3D> quadCorners;
  quadCorners.push_back(MAd_Vector3D(vCoord[quadDef[0]]));
  quadCorners.push_back(MAd_Vector3D(vCoord[quadDef[1]]));
  quadCorners.push_back(MAd_Vector3D(vCoord[quadDef[2]]));
  quadCorners.push_back(MAd_Vector3D(vCoord[quadDef[3]]));


  for (int rot=0;rot<4;rot++) {
    for (int ori=0;ori<2;ori++) {
      int* dof = quadDof + (rot*2 + ori) * quad.size();
      computeFaceTransformation(4,quadCorners,this,&quad,rot,ori,dof);
    }
  }


  // --- compute the correspondence between triangle nodes and element nodes
  MAd_TriInterpolant tri(order());

  triaDof = new int[24*tri.size()];

  for (size_t iFace=0;iFace<4;iFace++) {

    std::vector<MAd_Vector3D> triCorners;
    triCorners.push_back(MAd_Vector3D(vCoord[triaDef[iFace][0]]));
    triCorners.push_back(MAd_Vector3D(vCoord[triaDef[iFace][1]]));
    triCorners.push_back(MAd_Vector3D(vCoord[triaDef[iFace][2]]));

    for (int rot=0;rot<3;rot++) {
      for (int ori=0;ori<2;ori++) {
        int* dof = triaDof + (iFace*6 + rot*2 + ori)* tri.size();
        computeFaceTransformation(3,triCorners,this,&tri,rot,ori,dof);
      }
    }
  }
}

// -----------------------------------------------------------------------------

MAd_PyramidInterpolant::~MAd_PyramidInterpolant() {
  
  if (uvw)     delete [] uvw;
  if (bergot)  delete bergot;
  if (triaDof) delete [] triaDof;
  if (quadDof) delete [] quadDof;

  if (iOrder) delete [] iOrder;
  if (jOrder) delete [] jOrder;
  if (kOrder) delete [] kOrder;
  if (dOrder) delete [] dOrder;
  
}

// -----------------------------------------------------------------------------

std::string MAd_PyramidInterpolant::name () const {
  std::ostringstream n;
  n << "Equidistant Lagrangian function space of order " 
    << order() << " on pyramid";
  return n.str();
}

// -----------------------------------------------------------------------------

int MAd_PyramidInterpolant::computePoints(int o,double (*parametric)[3]) const {

  int i=0;

  if (o==0) {
    parametric[i][0] = 0;
    parametric[i][1] = 0;
    parametric[i][2] = 0;
    i++;
  }

  else {

    for (;i<5;i++) for (int d=0;d<3;d++) parametric[i][d] = vCoord[i][d];

    // edge nodes

    if (o > 1) {
      // edge 1
      for (int e=0;e<8;e++) {
        for (int io=1;io<o;io++,i++) {
          double f1 = (double) io / o;
          double f2 = 1. - f1;
          for (int d=0;d<3;d++) {
            parametric[i][d] = (parametric[edgeDef[e][0]][d] * f2 + 
                                parametric[edgeDef[e][1]][d] * f1);
          }
        }
      }

      // quad points

      MAd_QuadInterpolant* quad = new MAd_QuadInterpolant(o-2);

      double dksi = 1.-2./(o);

      for (int q=0;q<quad->size();q++,i++) { 
        double u,v,w;
        quad->nodeNumberCoordinate(q,u,v,w);
        parametric[i][0] = u*dksi;
        parametric[i][1] = v*dksi;
        parametric[i][2] = 0;
      }

      delete quad;

      // triangle points

      if (o > 2) {

        MAd_TriInterpolant* tri = new MAd_TriInterpolant(o-3);

        MAd_Vector3D *triCoord = new MAd_Vector3D[tri->size()];

        double overP = 1./o;
        double du    = 1.*(o-3)/o;
        for (int q=0;q<tri->size();q++) { 
          double u,v,w;
          tri->nodeNumberCoordinate(q,u,v,w);
          triCoord[q] = MAd_Vector3D(overP+u*du,overP+v*du,0);
        }

        for (int t=0;t<4;t++) { // loop over triangles t

          MAd_Vector3D uvw0(parametric[triaDef[t][0]]);
          MAd_Vector3D uvw1(parametric[triaDef[t][1]]);
          MAd_Vector3D uvw2(parametric[triaDef[t][2]]);

          MAd_Vector3D du01(uvw0,uvw1);
          MAd_Vector3D du02(uvw0,uvw2);

          for (int q=0;q<tri->size();q++,i++) { 
            MAd_Vector3D uvwP = uvw0;
            uvwP += triCoord[q][0] * du01;
            uvwP += triCoord[q][1] * du02;
            parametric[i][0] = uvwP[0];
            parametric[i][1] = uvwP[1];
            parametric[i][2] = uvwP[2];
          }
        }
        delete [] triCoord;
        delete tri;

        double (* uvwInside)[3] = parametric + i;
        int nb = int(computePoints(o-3,uvwInside));

        double factor = 1.*(o-3)/o;

        for (int k=0;k<nb;k++) {
          for (int d=0;d<3;d++) {
            uvwInside[k][d] *= factor;
          }
          uvwInside[k][2] += 1./(o);
        }
        i+=nb;

      }
    }
  }
  return i;

}

// -----------------------------------------------------------------------------
void MAd_PyramidInterpolant::fcts(const double &u, 
                                  const double &v, 
                                  const double &w,
                                  doubleVector &f) const {
  doubleVector fs(nDof);
  monomialFcts(u,v,w,fs);
  coef.mult(fs,f,false,1,0);
}

// -----------------------------------------------------------------------------
void MAd_PyramidInterpolant::grads(const double &u, 
                                   const double &v, 
                                   const double &w,
                                   doubleVector gr[3]) const {
    
  doubleVector grs[3] = {doubleVector(nDof),
                         doubleVector(nDof),
                         doubleVector(nDof)};
  monomialGrads(u,v,w,grs);
  for (size_t k=0;k<3;k++) coef.mult(grs[k],gr[k],false,1,0);
}

// -----------------------------------------------------------------------------
MAd_Interpolant * MAd_PyramidInterpolant::clone(int o) const {
  return new MAd_PyramidInterpolant(o);
}

// -----------------------------------------------------------------------------
void MAd_PyramidInterpolant::nodeNumberCoordinate(const int & i, 
                                                  double &u, 
                                                  double &v, 
                                                  double &w) const {
  
  u = uvw[i][0];
  v = uvw[i][1];
  w = uvw[i][2];
  return;
}

// ------------------------------------------------------------------------------

void MAd_PyramidInterpolant::getMonomialExponents(int i,
                                                  int& pu,
                                                  int& pv,
                                                  int& pw) const {
  bergot->polynomialOrders(i,pu,pv,pw);
}

// ------------------------------------------------------------------------------
/*! the number of dof is 
  \f$ \sum_{k=1}^{p+1} k^2 = \frac{2*(p+1)^3+3*(p+1)^2+(p+1)}{6} \f$ */

int MAd_PyramidInterpolant::size(int order) {
  int n = order+1;
  int n2 = n*n;
  int n3 = n2*n;
  return (2*n3+3*n2+n)/6;
}

// -----------------------------------------------------------------------------
/*! evaluate the monomials */ 

void MAd_PyramidInterpolant::monomialFcts(const double& u,
                                          const double& v,
                                          const double& w,
                                          doubleVector& f) const {


  double oneMinW = (w == 1) ? 1e-12 : 1.-w;

  for (int iDof=0;iDof<nDof;iDof++) {
    
    f(iDof) = (std::pow(u,iOrder[iDof]) * 
               std::pow(v,jOrder[iDof]) * 
               std::pow(w,kOrder[iDof]) * 
               std::pow(oneMinW,dOrder[iDof]));
    
  }
}

// -----------------------------------------------------------------------------
/*! evaluate the gradients of the simple monomials instead of Bergot */

void MAd_PyramidInterpolant::monomialGrads(const double& u,
                                           const double& v,
                                           const double& w,
                                           doubleVector gr[3]) const {

  
  double oneMinW = (w == 1) ? 1e-12 : 1.-w;
  
  for (int iDof=0;iDof<nDof;iDof++) {

    double upm1 = (iOrder[iDof] ==0) ? 0 : std::pow(u,iOrder[iDof] - 1);
    double vpm1 = (jOrder[iDof] ==0) ? 0 : std::pow(v,jOrder[iDof] - 1);
    double wpm1 = (kOrder[iDof] ==0) ? 0 : std::pow(w,kOrder[iDof] - 1);

    double up = (iOrder[iDof] == 0) ? 1 : upm1 * u;
    double vp = (jOrder[iDof] == 0) ? 1 : vpm1 * v;
    double wp = (kOrder[iDof] == 0) ? 1 : wpm1 * w;
    
    double wM1P = std::pow(oneMinW,dOrder[iDof]);
    
    gr[0](iDof) = iOrder[iDof] * upm1 * vp * wp * wM1P;
    gr[1](iDof) = jOrder[iDof] * up * vpm1 * wp * wM1P;
    gr[2](iDof) = kOrder[iDof] * up * vp * wpm1 * wM1P;
    
    if (dOrder[iDof] != 0) {
      gr[2](iDof) -= dOrder[iDof]*up*vp*wp*std::pow(oneMinW,dOrder[iDof]-1);
    }
  }    
}

// -----------------------------------------------------------------------------
