// -*- C++ -*- 
// -----------------------------------------------------------------------------
// MAdLiB - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// ------------------------------------------------------------------


#include "MAd_QuadInterpolant.h"

#include <cmath>
using std::max;

//---------------------------------------------------------------------
// Quad
/*  
    j
    |_i
    -
   |0|             order 0 (#1)
    -  
   3-2
   | |             order 1 (#4)
   0-1 

   3-6-2
   |   |
   7 8 5           order 2 (#9)
   |   |
   0-4-1 

   3- 9- 8- 2
   |        |
   10 14 15 7
   |        |       order 3 (#16)
   11 12 13 6
   |        |
   0- 4- 5- 1  
     
   3-12-11-10- 2
   |           |
   13 22 23 24 9
   |           |
   14 19 20 21 8   order 4 (#25)
   |           |
   15 16 17 18 7
   |           |
   0- 4- 5- 6- 1  
   
     ...
*/

//-------------------------------
//  Public Members
//-------------------------------
MAd_QuadInterpolant::MAd_QuadInterpolant(const int &o)
  :MAd_Interpolant(MSH_QUA_4, o), 
   ndof(size(o)), 
   lefs(o), 
   ndofe(lefs.size(o)), 
   codeClosure (8,int(getClosureSpaceSize())) {
  coef.resize(size(),size());
  
  NodeNumberu = new double[ndof];
  NodeNumberv = new double[ndof];
  idFromIJ = new int[ndof];
  nodePositionQuad(order(), NodeNumberu, NodeNumberv, idFromIJ);
  
  for (int i=0; i< int(ndof); i++){
    for (int j=0; j< int(ndof); j++)
      coef(i,j)=getCoef(i,j);
  }

  for (int i=0; i<4; i++){
    int  k = getOrientationCode(i, 1);
    getClosureSpace (i, 1, codeClosure[k]);
    k = getOrientationCode(i,-1);
    getClosureSpace (i,-1, codeClosure[k]);
  }
}

MAd_QuadInterpolant::~MAd_QuadInterpolant() {
  delete[] NodeNumberu;
  delete[] NodeNumberv;
  delete[] idFromIJ;
  return;
}

void MAd_QuadInterpolant::fcts (const double &u, const double &v, const double &w, 
                                      doubleVector &f) const {
  const int n=int(ndof);
  doubleVector m(n);
  f= 0.;
  for (int i=0; i<int(ndof) ; i++) m(i) = monomeof(u,v,i);
  coef.mult(m,f,false);
  return;
}

void MAd_QuadInterpolant::grads (const double &u, const double &v, const double &w, 
                                 doubleVector gr[3]) const {
  const int n=int(ndof);
  doubleVector m(n);
  for (int k=0; k< 2; k++){ 
    gr[k]= 0.;
    for (int i=0; i< int(ndof) ; i++) m(i) = dmonomeof(u,v,i,k);
    coef.mult(m,gr[k],false);
  }
  return;
}

void MAd_QuadInterpolant::getIJfromIdx(int idx, int &i, int &j) const {
  int o = (int)order();
  i = (int)((NodeNumberu[idx]*o+o)/2.);
  j = (int)((NodeNumberv[idx]*o+o)/2.);
  return;
}

double MAd_QuadInterpolant::getCoef(int id1,int id2) const {
  int i1,i2,j1,j2;
  getIJfromIdx(id1, i1, j1);
  getIJfromIdx(id2, i2, j2);
  double coefu = MAd_EdgeInterpolant::getCoef(order(), i1, i2);
  double coefv = MAd_EdgeInterpolant::getCoef(order(), j1, j2);
  return coefu*coefv;
}

void MAd_QuadInterpolant::getMonomialExponents(int i, int &powerU, int &powerV,  int &powerW) const{
  int iu,iv;
  getIJfromIdx(i, iu, iv);
  powerU = iu;
  powerV = iv;
  powerW=0;
}

double MAd_QuadInterpolant::monomeof(const double &u, const double &v, int i) const{
  int powerU, powerV, powerW;
  getMonomialExponents(i, powerU, powerV, powerW);
  return pow(u,powerU)*pow(v,powerV);
};

double MAd_QuadInterpolant::dmonomeof(const double &u, const double &v, int i, int k) const{
  double Zero=0.;
  int powerU, powerV, powerW;
  getMonomialExponents(i, powerU, powerV, powerW);
  switch (k){
  case 0:
    if (powerU > 0)
      return powerU * pow(u,powerU-1)*pow(v,powerV);
    else return Zero;
    break;
  case 1:    
    if (powerV > 0)
      return powerV * pow(u,powerU)*pow(v,powerV-1);
    else return Zero;
    break;
  default:
    throw;
    break;
  }
}

size_t MAd_QuadInterpolant::getClosureSpaceSize(int iFace) const {
  return order()+1;
}

const int*  MAd_QuadInterpolant::getClosure (int iEd, int iSign, int iRotate) const { 
  return  codeClosure[getOrientationCode(iEd, iSign)];
} // replace double*

// -----------------------------------------------------------------------------
void MAd_QuadInterpolant::getClosureSpace(const int &edgenumber,
                                                const int &sign, 
                                                int* edgedofs) const { //replace int*

  if (order() == 0) edgedofs[0] = 0;
  else {

    int k0=sign==1?0:order();
    int k1=sign==1?order():0;
    if (edgenumber==0) {
      edgedofs[k0] = 0;
      edgedofs[k1] = 1;
    }
    if (edgenumber==1) {
      edgedofs[k0] = 1;
      edgedofs[k1] = 2;
    }
    if (edgenumber==2) {
      edgedofs[k0] = 2;
      edgedofs[k1] = 3;
    }
    if (edgenumber==3) {
      edgedofs[k0] = 3;
      edgedofs[k1] = 0;
    }
    int n = 3 + edgenumber*(order()-1);
    
    for (int i=1; i<order(); i++){
      int im = i;
      if (sign==-1) im = order()-i;
      edgedofs[im] = n+i;
    }
  }
  return;
}

void MAd_QuadInterpolant::nodeNumberCoordinate(const int & i, double &u, double &v, double &w) const{

  u=NodeNumberu[i];
  v=NodeNumberv[i];
  w=0.;
  return;
}

int MAd_QuadInterpolant::size(int q){
  return (q+1)*(q+1);
}

bool MAd_QuadInterpolant::reorientData(doubleMatrix& sol,int orientation) const {

  if (sol.size1() != size()) return false;

  // if (orientation != 1) cout << "Reorienting quad data " << endl;

  doubleMatrix tmp(sol);

  if (orientation < 0) {
    int n=0;
    int locOrder = order();
    while (locOrder>=0){
      if (locOrder==0) doubleVector (tmp.touchRow(n)) = (const doubleVector)(sol.seeRow(n));
      else {
        // vertices
        for (int j=0;j<sol.size2();j++) {
          tmp(n,j) = sol(n,j); tmp(n+1,j) = sol(n+3,j); tmp(n+2,j) = sol(n+2,j); tmp(n+3,j) = sol(n+1,j);
          // edges
          for (int k=0; k<(locOrder-1); k++){
            tmp(n+4+k,j) = sol(n+4+4*(locOrder-1)-(k+1),j);
            tmp(n+4+(locOrder-1)+k,j) = sol(n+4+3*(locOrder-1)-(k+1),j);
            tmp(n+4+2*(locOrder-1)+k,j) = sol(n+4+2*(locOrder-1)-(k+1),j);
            tmp(n+4+3*(locOrder-1)+k,j) = sol(n+4+(locOrder-1)-(k+1),j);
          }
        }
      }
      // inside, recursively
      n = n+4*locOrder;
      locOrder = locOrder-2;
    }
    sol = tmp;
  }
  
  // if (abs(orientation) > 4) cout << "Too many rotations " << endl;

  for (int ori=1;ori<abs(orientation);ori++) {
    int n=0;
    int locOrder = order();
    while (locOrder>=0){
      for (int j=0; j<sol.size2(); j++) {
        if (locOrder==0) tmp(n,j) = sol(n,j);
        else {
          for (int i=0; i<4; i++){
            // vertices
            tmp(n+i,j) = sol(n+(i+1)%4,j);
            // edges
            for (int k=0; k<(locOrder-1); k++) tmp(n+4+i*(locOrder-1)+k,j) = sol(n+4+((i+1)%4)*(locOrder-1)+k,j);
          }
        }
      }
      // inside, recursively
      n = n+4*locOrder;
      locOrder = locOrder-2;
    }
    sol = tmp;
  }
  return true;
}

// -----------------------------------------------------------------------------

bool MAd_QuadInterpolant::findOriginalCoordinates(int iRot,int iSwp,
                                                        const doubleMatrix& uvwTfo,
                                                        doubleMatrix& uvwOri) const {
  
  
  if (abs(iSwp) != 1) return false;
  if (iRot > 3 || iRot < 0) return false;
  
  const doubleVector uTfo(uvwTfo.seeColumn(0));
  const doubleVector vTfo(uvwTfo.seeColumn(1));
  
  doubleMatrix uvwTmp(uvwTfo.size1(),2);
  
  doubleVector uTmp(uvwTmp.touchColumn(0));
  doubleVector vTmp(uvwTmp.touchColumn(1));
  
  switch(iSwp) {
  case 1:
    uTmp = uTfo;
    vTmp = vTfo;
    break;
 
  case -1:
    uTmp = vTfo;
    vTmp = uTfo;
    break;
  default:
    return false;
  }

  doubleVector uOri(uvwOri.touchColumn(0));
  doubleVector vOri(uvwOri.touchColumn(1));
  
  switch (iRot) {
  case 0:
    uOri = uTmp;
    vOri = vTmp;
    break;
  case 1:
    uOri = vTmp;
    uOri *= -1;
    vOri = uTmp;
    break;
  case 2:
    uOri = uTmp;
    vOri = vTmp;
    uOri *= -1;
    vOri *= -1;
    break;
  case 3:
    uOri = vTmp;
    vOri = uTmp;
    vOri *= -1;
    break;
  default:
    return false;
  }

  return true;
}


//-------------------------------
//  Private Members
//-------------------------------

int MAd_QuadInterpolant::getOrientationCode(int edgenumber, int dir) const 
{
  if (dir > 0)
    return edgenumber*2;
  else
    return edgenumber*2+1;
}

void MAd_QuadInterpolant::nodePositionQuad(int o, double *u, double *v, int *idIJ) {
  //int o = (int) order();
  if (o==0) {
    u[0]=0.;
    v[0]=0.;
    idIJ[0] = 0;
    return;
  }
  // quad vertices
  u[0] = -o;
  v[0] = -o;
  idIJ[0] = 0;
  u[1] = o;
  v[1] = -o;
  idIJ[o] = 1;
  u[2] = o;
  v[2] = o;
  idIJ[o+o*(o+1)] = 2;
  u[3] = -o;
  v[3] = o;
  idIJ[o*(o+1)] = 3;
  int n = 4;

  // edges
  for (int k=0; k<o-1; k++){
    u[n+k] = 2*(k+1)-o;          v[n+k] = -o;
    u[n+(o-1)+k] = o;            v[n+(o-1)+k] = 2*(k+1)-o;
    u[n+2*(o-1)+k] = o-2*(k+1);  v[n+2*(o-1)+k] = o;
    u[n+3*(o-1)+k] = -o;         v[n+3*(o-1)+k] = o-2*(k+1);
  }
  for (int e=0; e<4; e++){
    for (int k=0; k<o-1; k++){
      int i = (int) round((u[n+e*(o-1)+k] + o)/2.);
      int j = (int) round((v[n+e*(o-1)+k] + o)/2.);
      idIJ[i+j*(o+1)] = n+e*(o-1)+k;
    }
  }
  n += 4*(o-1);

  // inside the quad, recursively
  if (o > 1){
    int nbdoftmp = size(o-2);
    double* utmp = new double[nbdoftmp];
    double* vtmp = new double[nbdoftmp];
    int* ijtmp = new int[nbdoftmp];
    nodePositionQuad(o-2, utmp, vtmp, ijtmp);
    for (int k=0; k< nbdoftmp; k++){
      u[n+k] = utmp[k]*(o-2);
      v[n+k] = vtmp[k]*(o-2);
      int i = (int) round((u[n+k] + o)/2.);
      int j = (int) round((v[n+k] + o)/2.);
      idIJ[i+j*(o+1)] = n+k;
    }
    delete []utmp;
    delete []vtmp;
    delete []ijtmp;
    n += nbdoftmp;
  }
  
  for (int m=0; m<n; m++){
    u[m] = u[m]/o;
    v[m] = v[m]/o;
  }
  return;     
}

size_t MAd_QuadInterpolant::getIdxFromIJ(int i, int j) const {
  size_t n = j*(order()+1)+i;
  if (n>=ndof) throw;
  return idFromIJ[n];
}

int MAd_QuadInterpolant::directionEdge[4][2] = {{1,0},// edge 0
                                                {0,1},// edge 1
                                                {-1,0},// edge 2
                                                {0,-1}};// edge 3


