// -*- C++ -*- 
// -----------------------------------------------------------------------------
// MAdLiB - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -----------------------------------------------------------------------------


#include <iostream>
#include <algorithm>
#include <cmath>
#include <cstdio>
#include <cassert>
#include <vector>

#include "MeshDataBaseInterface.h"
#include "MshTags.h"

#include "MAdMapping.h"
#include "MAd_TetInterpolant.h"
#include "MAd_TriInterpolant.h"
#include "MAd_QuadInterpolant.h"
#include "MAd_HexInterpolant.h"
#include "MAd_PrismInterpolant.h"
#include "MAd_PyramidInterpolant.h"
#include "MAd_Interpolant.h"
#include "MAd_InterpolantFactory.hpp"

#include "MAdVector3D.h"

// -----------------------------------------------------------------------------
// compute the modulus of a 3D vector
double norm3(double a[3])
{
  return std::sqrt(a[0] * a[0] + a[1] * a[1] + a[2] * a[2]);
}

// -----------------------------------------------------------------------------
// normalize a 3D vector
double norme(double a[3])
{
  double mod = norm3(a);
  if(mod != 0.0){
    a[0] /= mod;
    a[1] /= mod;
    a[2] /= mod;
  }
  return mod;
}

// -----------------------------------------------------------------------------
// compute the cross product of vector a and b and store the result in c
void prodve(double a[3], double b[3], double c[3])
{
  c[2] = a[0] * b[1] - a[1] * b[0];
  c[1] = -a[0] * b[2] + a[2] * b[0];
  c[0] = a[1] * b[2] - a[2] * b[1];
}

// -----------------------------------------------------------------------------
// solve a 3 equations linear system using Cramer's formulae 
int sys3x3(double mat[3][3], double b[3], double res[3])
{
  double ud;
  int i;
  double det = (mat[0][0] * (mat[1][1] * mat[2][2] - mat[1][2] * mat[2][1]) -
                mat[0][1] * (mat[1][0] * mat[2][2] - mat[1][2] * mat[2][0]) +
                mat[0][2] * (mat[1][0] * mat[2][1] - mat[1][1] * mat[2][0]));

  if(det == 0.0) {
    res[0] = res[1] = res[2] = 0.0;
    return (0);
  }

  ud = 1. / (det);

  res[0] = b[0] * (mat[1][1] * mat[2][2] - mat[1][2] * mat[2][1]) -
    mat[0][1] * (b[1] * mat[2][2] - mat[1][2] * b[2]) +
    mat[0][2] * (b[1] * mat[2][1] - mat[1][1] * b[2]);

  res[1] = mat[0][0] * (b[1] * mat[2][2] - mat[1][2] * b[2]) -
    b[0] * (mat[1][0] * mat[2][2] - mat[1][2] * mat[2][0]) +
    mat[0][2] * (mat[1][0] * b[2] - b[1] * mat[2][0]);

  res[2] = mat[0][0] * (mat[1][1] * b[2] - b[1] * mat[2][1]) -
    mat[0][1] * (mat[1][0] * b[2] - b[1] * mat[2][0]) +
    b[0] * (mat[1][0] * mat[2][1] - mat[1][1] * mat[2][0]);

  for(i = 0; i < 3; i++)
    res[i] *= ud;
  return (1);
}

// -----------------------------------------------------------------------------
/*! \brief swaps a higher order quad */

void quadChangeDir(vector<pVertex>& id,int order) {
  const MAd_QuadInterpolant* quad =  
    dynamic_cast<const MAd_QuadInterpolant*> 
    (MAd_InterpolantFactorySgl::instance().find(MSH_TYPE_QUA,order));
  
  int ndofq = quad->size();
  vector<pVertex> tmp(quad->size());
  for (int j=0; j<=order; j++){
    for (int i=0; i<=order; i++){
      tmp[quad->getIdxFromIJ(i,j)] = id[quad->getIdxFromIJ(j,i)]; 
    }
  }
  for (int k=0; k<ndofq; k++){
    id[k] = tmp[k];
  }
  return;
}

// -----------------------------------------------------------------------------
/*! \brief rotates a higher order quad */

void quadChangeOrientation(vector<pVertex>& id,int order) {
  
  const MAd_QuadInterpolant* quad =  
    dynamic_cast<const MAd_QuadInterpolant*> 
    (MAd_InterpolantFactorySgl::instance().find(MSH_TYPE_QUA,order));
  int ndofq = quad->size();
  vector<pVertex> tmp(quad->size());
  for (int j=0; j<=order; j++){
    for (int i=0; i<=order; i++){
      tmp[quad->getIdxFromIJ(i,j)] = id[quad->getIdxFromIJ(j,order-i)]; 
    }
  }
  for (int k=0; k<ndofq; k++) id[k] = tmp[k];
  return;
}

// -----------------------------------------------------------------------------

void triChangeDir(vector<pVertex>& id, int order) {
  int ndoft = MAd_TriInterpolant::size(order);
  vector<pVertex> tmp;
  int n=0;
  int locOrder = order;
  while (locOrder>=0){
    if (locOrder==0) tmp.push_back(id[n]);
    else if (locOrder>2){
      // vertices
      tmp.push_back(id[n]); tmp.push_back(id[n+2]); tmp.push_back(id[n+1]);
      // edges
      for (int j=3; j>0; j--){
        for (int k=0; k<(locOrder-1); k++){
          tmp.push_back(id[n+3+j*(locOrder-1)-(k+1)]);
        }
      }
    }
    // inside, recursively
    n = n+3*locOrder;
    locOrder = locOrder-3;
  }
  for (int k=0; k<ndoft; k++) id[k] = tmp[k];

  return;
}

// -----------------------------------------------------------------------------

void triChangeOrientation(vector<pVertex>& id, int order) {
  int ndoft = MAd_TriInterpolant::size(order);
  vector<pVertex> tmp;
  int n=0;
  int locOrder = order;
  while (locOrder>=0){
    if (locOrder==0) tmp.push_back(id[n]);
    else if (locOrder>2){
      // vertices
      for (int edge=0; edge<3; edge++){
        tmp.push_back(id[n+(edge+1)%3]);
      }
      // edges
      for (int edge=0; edge<3; edge++){
        for (int k=0; k<(locOrder-1); k++) tmp.push_back(id[n+3+((edge+1)%3)*(locOrder-1)+k]);
      }
    }
    // inside, recursively
    n = n+3*locOrder;
    locOrder = locOrder-3;
  }
  for (int k=0; k<ndoft; k++) id[k] = tmp[k];

  return;

}
                         
// -----------------------------------------------------------------------------

void reorientTriangleHigherOrderPoints(vector<pVertex>& ho,
                                       int nbPoints,int orientation,int dir) {

  if (nbPoints <= 1) return;
  if (nbPoints >  3) {
    MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                "Reorientation of interior points"
                                " of triangle of order > 4"
                                " not yet supported");
  }
  

  if (dir == 1 && orientation == 0) return;

  vector<pVertex> tmp(ho.size());
  for (int i=0;i<3;i++) tmp[i] = ho[(i+orientation)%3];
  if (dir == 1) ho = tmp;
  else {
    ho[0] = tmp[0];
    ho[1] = tmp[2];
    ho[2] = tmp[1];
  }
}


// -----------------------------------------------------------------------------
// triangle template
// -----------------------------------------------------------------------------
//
// vertices:
// ---------
// 
//   V0 V1 V2 
// 
// edges:
// ------
//
//   edge orientation such that a closed loop is formed 
// 
//      | v0 v1 |
//   ------------   
//   E0 | V0 V1 |
//   E1 | V1 V2 |
//   E2 | V2 V0 |
//   
// face:
// ------
// 
//   loop (eta(ksi)), ksi e0 (v0-v1) , eta -e2 (v0-v2)
//
// -----------------------------------------------------------------------------

const double MAdMapping::triNodeUVW[3][2] = {{0.,0.},{1.,0.},{0.,1.}};
const int    MAdMapping::triEdgeToNodes[3][2] = {{0,1},{1,2},{2,0}};

// -----------------------------------------------------------------------------
// quadrangle template
// -----------------------------------------------------------------------------
//
// vertices:
// ---------
//
// V3 V2
// v0 V1
//
// edges:
// ------
//
//   edge orientation such that a closed loop is formed
//
//      | v0 v1 |
//   ------------
//   E0 | V0 V1 |
//   E1 | V1 V2 |
//   E2 | V2 V3 |
//   E3 | V3 V0 |
//
// face:
// ------
//
//   loop (eta(ksi)), ksi e0 (v0-v1) , eta -e2 (v0-v2) ???
//
// -----------------------------------------------------------------------------

const double MAdMapping::quadNodeUVW[4][2] = {{-1.,-1.},{1.,-1.},{1.,1.},{-1.,1.}};
const int    MAdMapping::quadEdgeToNodes[4][2] = {{0,1},{1,2},{2,3},{3,0}};

// -----------------------------------------------------------------------------
// tetrahedron template
// -----------------------------------------------------------------------------
//
// vertices:
// ---------
// 
//   V0 V1 V2 V3
// 
// edges:
// ------
//  
//      | v0 v1 |
//   ------------
//   E0 | V0 V1 |
//   E1 | V1 V2 |
//   E2 | V2 V0 |
//   E3 | V0 V3 |
//   E4 | V1 V3 |
//   E5 | V2 V3 |
//   
// faces:
// ------
// 
//   loop (eta(ksi)), ksi e0 (v0-v1) , eta -e2 (v0-v2)
// 
//      |  e0  e1  e2 | v0 v1 v2 |
//   -----------------------------     
//   F0 | +E0 +E1 +E2 | V0 V1 V2 |
//   F1 | +E0 +E4 -E3 | V0 V1 V3 |
//   F2 | +E1 +E5 -E4 | V1 V2 V3 |
//   F3 | -E2 +E5 -E3 | V0 V2 V3 |
//
//   
// volume:
// -------
// 
//   loop (zeta(eta(ksi))), ksi (V0-V1), eta (V0-V2), zeta (V0-V3)
// 
// -----------------------------------------------------------------------------

const double MAdMapping::tetNodeUVW[4][3] = {{0.,0.,0.},{1.,0.,0.},{0.,1.,0.},{0.,0.,1.}};
const int    MAdMapping::tetEdgeToNodes[6][2] = {{0,1},{1,2},{2,0},{0,3},{1,3},{2,3}};
const int    MAdMapping::tetFaceToNodes[4][3] = {{0,1,2},{0,1,3},{1,2,3},{0,2,3}};

// -----------------------------------------------------------------------------
// hexahedron template
// -----------------------------------------------------------------------------
//
// vertices:
// ---------
// 
//   V0 V1 V2 V3 V4 V5 V6 V7 V8
// 
// edges:
// ------
//  
//      | v0 v1 |
//   ------------
//   E0 | V0 V1 |
//   E1 | V1 V2 |
//   E2 | V2 V3 |
//   E3 | V3 V0 |
//   E4 | V4 V5 |
//   E5 | V5 V6 |
//   E6 | V6 V7 |
//   E7 | V7 V4 |
//   E8 | V0 V4 |
//   E9 | V1 V5 |
//   E10| V2 V6 |
//   E11| V3 V7 |
//   
// faces:
// ------
// 
//   loop (eta(ksi)), ksi e0 (v0-v1) , eta -e2 (v0-v2) ??
// 
//      |  e0  e1  e2 e3  | v0 v1 v2 v3 |
//   ------------------------------------     
//   F0 | -E3 -E2 -E1 -E0 | V0 V3 V2 V1 |
//   F1 | +E0 +E9 -E4 -E8 | V0 V1 V5 V4 |
//   F2 | +E1 +E10-E5 -E9 | V1 V2 V6 V5 |
//   F3 | +E2 +E11-E6 -E10| V2 V3 V7 V6 |
//   F4 | -E3 +E8 -E7 -E11| V3 V0 V4 V7 |
//   F5 | +E4 +E5 +E6 +E7 | V4 V5 V6 V7 |
//
//   
// volume:
// -------
// 
//   loop (zeta(eta(ksi))), ksi (V0-V1), eta (V0-V2), zeta (V0-V3) ??
// 
// -----------------------------------------------------------------------------

const double MAdMapping::hexNodeUVW[8][3] = {{-1.,-1.,-1.},
                                             {1.,-1.,-1.},
                                             {1.,1.,-1.},
                                             {-1.,1.,-1.},
                                             {-1.,-1., 1.},
                                             {1.,-1., 1.},
                                             {1.,1., 1.},
                                             {-1.,1., 1.}};

const int MAdMapping::hexEdgeToNodes[12][2] = {{0,1},{1,2},{2,3},{3,0},
                                               {0,4},{1,5},{2,6},{3,7},
                                               {4,5},{5,6},{6,7},{7,4}};

const int MAdMapping::hexFaceToNodes[6][4] = {{0,3,2,1},{0,1,5,4},{1,2,6,5},
                                              {2,3,7,6},{3,0,4,7},{4,5,6,7}};

// -----------------------------------------------------------------------------

const double MAdMapping::prismNodeUVW[6][3] = {{0,0,-1},{1,0,-1},{0,1,-1},
                                               {0,0, 1},{1,0, 1},{0,1, 1}};
const int MAdMapping::prismEdgeToNodes[9][2] = {{0,1},{1,2},{2,0},
                                                {3,4},{4,5},{5,0},
                                                {0,3},{1,4},{2,5}};
const int MAdMapping::prismFaceToNodesTri[2][3] = {{0,1,2},{3,4,5}};
const int MAdMapping::prismFaceToNodesQuad[3][4] = {{0,3,4,1},{1,4,5,2},{3,0,2,5}};

// -----------------------------------------------------------------------------

const double MAdMapping::pyramidNodeUVW[5][3] = {{-1.,-1.,0.},
                                                 {1.,-1.,0.},
                                                 {1.,1.,0.},
                                                 {-1.,1.,0.},
                                                 {0.,0.,1.}};
const int MAdMapping::pyramidEdgeToNodes[8][2] = {{0,1},
                                                  {0,3},
                                                  {0,4},
                                                  {1,2},
                                                  {1,4},
                                                  {2,3},
                                                  {2,4},
                                                  {3,4}};
const int MAdMapping::pyramidFaceToNodesTri[4][3] = {{0,1,4},{3,0,4},{1,2,4},{3,2,4}};
const int MAdMapping::pyramidFaceToNodesQuad[1][4] = {{0,1,2,3}};

// ----------------------------------------------------------------------------------
//! \brief collect the vertices of the mapping in the correct order 
//! \todo correct in case of a hybrid element mappings ? 

void MAdMapping::getVertices(pEntity e, vector<pVertex  > &vl) {
  
  
  int mshType = EN_mshType(e);
  int order   = EN_order(e);

  switch (mshType){
  
  case MSH_TYPE_LIN: 
    {
      vl.push_back(E_vertex((pEdge) e , 0));
      for (int i=0;i<(order-1);i++){
        vl.push_back(E_point((pEdge) e , i));
      }
      vl.push_back(E_vertex((pEdge) e , 1));
      break;
    }
  case MSH_TYPE_TRI:
    { 
      pFace pf = (pFace) e;
      int nv = EN_numVertices(e);
      const int (*edgeToNodes)[2] = triEdgeToNodes;
      for (int i=0;i<nv;i++) vl.push_back(F_vertex(pf,i));

      if (order > 1) {
        for (int iedge=0;iedge<nv;iedge++){
          
          pEdge pe = F_edge(pf,iedge);
          pVertex firstv = vl[edgeToNodes[iedge][0]];
                  
          if (E_vertex(pe,0) == firstv){
            for (int i=0;i<(order-1);i++) vl.push_back(E_point(pe,i));
          }
          else {
            for (int i=order-2;i>=0;i--)  vl.push_back(E_point(pe,i));
          }
        }
                
        if (order > 2) {
          int nbInPoints = std::min(F_numPoints(pf),(order -2)*(order-1)/2);
          if (nbInPoints>0) {// Lagrange mapping
            for (int i=0;i<nbInPoints;i++) vl.push_back(F_point(pf, i));
          }
        }
      }
      break;
    }
  case MSH_TYPE_QUA:
    {
      
      pFace pf = (pFace) e;
      int nv = EN_numVertices(e);
      for (int i=0;i<nv;i++) vl.push_back(F_vertex(pf,i));
      if (order > 1) {
        // points on edges
        for (int i=0;i<4;i++){
          pEdge pe = F_edge(pf,i);
          if (E_vertex(pe,0)==F_vertex(pf,i)) for (int k=0;k<=order-2;k++) vl.push_back(E_point(pe,k));
          else                                for (int k=order-2;k>=0;k--) vl.push_back(E_point(pe,k));
        }
        // inside quad
        for (int k=0;k<F_numPoints(pf);k++) vl.push_back(F_point(pf,k));
      }
      break;
    }
  case MSH_TYPE_TET: 
    {  
      pRegion r = (pRegion)e;
      for (int i=0;i<R_numVertices(r);i++) vl.push_back(R_vertex(r,i));
      if (order > 1) {

        int nbEdgePoints = 0;
        pPList edges = R_edges(r);
                
        void* ptr = NULL;
                
        int iedge = 0;
        while (pEdge edge = (pEdge) PList_next(edges,&ptr)) {
                  
          nbEdgePoints = E_numPoints(edge);
          if (iedge ==0 &&  nbEdgePoints == 0) break;
          const int* nodes = tetEdgeToNodes[iedge];
                  
          if (EN_id((pEntity) E_vertex(edge,0)) == EN_id ((pEntity) vl[nodes[0]])) {
            for (int inode=0;inode<nbEdgePoints;inode++)    vl.push_back(E_point(edge,inode));
          }
          else {
            for (int inode=nbEdgePoints-1;inode>=0;inode--) vl.push_back(E_point(edge,inode));
          }
          iedge++;
        }
                
        PList_delete(edges);

        // KH: check orientation for p>3

        if (order > 2) {

          int nbFacePoints = 0;
          // pPList faces = R_faces(r,0);
          pPList faces = R_faces(r);
          void *ptrf = NULL;
                  
          int iface = 0;
                  
          while (pFace f = (pFace) PList_next(faces,&ptrf)) {
                    
            int od  = R_faceOri(r,iface);
            int dir = od > 0 ? 1 : -1;
            int ori = dir * od - 1;
                    
            int nbF = F_numPoints(f);
            if (iface ==0)  nbFacePoints = nbF;
            else if (nbFacePoints != nbF) {
              exit(1);
              break;
            }
                    
            if (nbFacePoints == 0) break;
                    
                    
            vector<pVertex> tmp;
            for (int inode =0;inode<nbFacePoints;inode++) tmp.push_back(F_point(f,inode));
                    
            reorientTriangleHigherOrderPoints(tmp,nbFacePoints,ori,dir);
            vl.insert(vl.end(),tmp.begin(),tmp.end());
                    
            iface++;
          }
                  
          PList_delete(faces);
                  
          if (order > 3) {
            // KH: check orientation for p>4
                    
            int nbRegionPoints = R_numPoints(r);
            for (int inode=0;inode<nbRegionPoints;inode++) vl.push_back(R_point(r,inode));
          }
        }
      }
      break;
    }
  case MSH_TYPE_HEX: 
    {
      
      pRegion r = (pRegion)e;
      for (int i=0;i<R_numVertices(r);i++) vl.push_back(R_vertex(r,i));
    
      if (order>1) {
      
        
        int faceOrigin[6][2] = {{0,3},{0,1},{1,2},
                                {2,3},{3,0},{4,5}};
        
        int edgeOrigin[12] = {0,1,2,3,
                              0,1,2,3,
                              4,5,6,7};
        
        // hex edges
        for (int ie=0;ie<12;ie++) {
          pEdge pe = R_edge(r,ie);
          if (E_vertex(pe,0)==R_vertex(r,edgeOrigin[ie])) for (int i=0;i<=order-2;i++) vl.push_back(E_point(pe,i));
          else                                            for (int i=order-2;i>=0;i--) vl.push_back(E_point(pe,i));
        }
                
        // hex faces
        for (int iFace=0;iFace<6;iFace++) {
          pFace pf = (pFace) R_face(r,iFace);
          vector<pVertex> vert;
          vector<pVertex> tmpface;
          if (order==2) vl.push_back(F_point(pf,0));
          else {
            // quad vertices
            for (int i=0;i<4;i++) vert.push_back(F_vertex(pf,i));
            // inside
            for (int i=0;i<MAd_QuadInterpolant::size(order-2);i++) {
              tmpface.push_back(F_point(pf,i));
            }
            while(vert[0]!=R_vertex(r,faceOrigin[iFace][0])) {
              quadChangeOrientation(vert,1);
              quadChangeOrientation(tmpface,order-2);
            }
            if (vert[1]!=R_vertex(r,faceOrigin[iFace][1])) {
              quadChangeDir(vert,1);
              quadChangeDir(tmpface,order-2);
            }
            for (int i=0;i<MAd_QuadInterpolant::size(order-2);i++) {
              vl.push_back(tmpface[i]);
            }
          }
        }
        // inside hex
        for (int i=0;i<MAd_HexInterpolant::size(order-2);i++) {
          vl.push_back(R_point(r,i));
        }
      }
      break;
    }
  case MSH_TYPE_PRI: 
    {
      
      pRegion r = (pRegion) e;

      for (int i=0;i<R_numVertices(r);i++) vl.push_back(R_vertex(r,i));
      
      if (order>1) {
        int eToN[9][2] = {{0,1},{1,2},{2,0},{3,4},{4,5},{5,3},{0,3},{1,4},{2,5}};
        int fToNTri[2][3] = {{0,1,2},{3,4,5}}; // tri
        int fToNQuad[3][4] = {{0,3,4,1},{1,4,5,2},{3,0,2,5}}; //quad
        int edgeMDB[9] = {0,2,1,3,5,4,6,7,8};
                
        // points on edges
        for (int edge=0;edge<9;edge++){
          pEdge re = R_edge(r,edgeMDB[edge]);
          if (E_vertex(re,0)==R_vertex(r,eToN[edge][0])) for (int i=0;i<=order-2;i++) vl.push_back(E_point(re,i));
          else                                           for (int i=order-2;i>=0;i--) vl.push_back(E_point(re,i));
        }
        // points on faces
        for (int iFace=0;iFace<5;iFace++){
          pFace pf = (pFace) R_face(r,iFace);
          vector<pVertex> vert;
          vector<pVertex> tmpface;
          if (iFace<2 && order>2){ // tri
            if (order==3) vl.push_back(F_point(pf,0));
            else {
              // tri vertices
              for (int i=0;i<3;i++) vert.push_back(F_vertex(pf,i));
              for (int i=0;i<MAd_TriInterpolant::size(order-3);i++) {
                tmpface.push_back(F_point(pf,i));
              }
              while(vert[0]!=R_vertex(r,fToNTri[iFace][0])) {
                triChangeOrientation(vert,1);
                triChangeOrientation(tmpface,order-3);
              }
              if (vert[1]!=R_vertex(r,fToNTri[iFace][1])) {
                triChangeDir(vert,1);
                triChangeDir(tmpface,order-3);
              }
              // inside tri
              for (int i=0;i<MAd_TriInterpolant::size(order-3);i++) {
                vl.push_back(tmpface[i]);
              }
            }
          }
          else if (iFace>1) { // quad
            if (order==2) vl.push_back(F_point(pf,0));
            else {
              // quad vertices
              for (int i=0;i<4;i++) vert.push_back(F_vertex(pf,i));
              for (int i=0;i<MAd_QuadInterpolant::size(order-2);i++) {
                tmpface.push_back(F_point(pf,i));
              }
              while(vert[0]!=R_vertex(r,fToNQuad[iFace-2][0])) {
                quadChangeOrientation(vert,1);
                quadChangeOrientation(tmpface,order-2);
              }
              if (vert[1]!=R_vertex(r,fToNQuad[iFace-2][1])) {
                quadChangeDir(vert,1);
                quadChangeDir(tmpface,order-2);
              }
              // inside quad
              for (int i=0;i<MAd_QuadInterpolant::size(order-2);i++) {
                vl.push_back(tmpface[i]);
              }
            }
          }
        }
        // inside prism
        if (order>2) {
          int nbptsinside = MAd_PrismInterpolant::size(order)-6-9*MAd_EdgeInterpolant::size(order-2)-2*MAd_TriInterpolant::size(order-3)-3*MAd_QuadInterpolant::size(order-2);
          for (int i=0;i<nbptsinside;i++) {
            vl.push_back(R_point(r,i));
          }
        }
      }
      break;
    }
  case MSH_TYPE_PYR:
    {
      
      pRegion r = (pRegion) e;
      // pyramid vertex
      for (int i=0;i<R_numVertices(r);i++) vl.push_back(R_vertex(r,i));
      
      if (order==2) { // order > 1
        int eToN[8][2] = {{0,1},{0,3},{0,4},{1,2},{1,4},{2,3},{2,4},{3,4}};
        
        // points on edges
        for (int edge=0;edge<8;edge++){
          pEdge re = R_edge(r,edge);
          if (E_vertex(re,0)==R_vertex(r,eToN[edge][0])) for (int i=0;i<=order-2;i++) vl.push_back(E_point(re,i));
          else                                           for (int i=order-2;i>=0;i--) vl.push_back(E_point(re,i));
        }
        // points on faces ~ only on the quad
        pFace pf = (pFace) R_face(r,0);        
        vl.push_back(F_point(pf,0));
        // inside pyramid ~ none for second order
      }
      if (order>2)
        throw; // not implemented
      
      break;    	
    }
  default:
    throw;
  }
}

// -----------------------------------------------------------------------------
/* creation and initialization of the MAdMapping class */
MAdMapping::MAdMapping (pEntity e,double s) : 
  ent(e), 
  entityTag (EN_mshTag(e)),
  entityDim (EN_type(e)),
  entityType(EN_mshType(e)),
  entityOrder(EN_order(e)),
  entityIncomplete(isSerendipityMSH(entityTag)),
  numVertices(EN_numVertices(e)),
  lengthScale(s) {
  
  double bbox[3][2];
  EN_box(e,bbox);
  
  double dd = 0;
  for (int d=0;d<3;d++) dd += (bbox[d][1]-bbox[d][0])*(bbox[d][1]-bbox[d][0]);
  overD = 1./std::sqrt(dd)/lengthScale;
  
  interpolant = MAd_InterpolantFactorySgl::instance().find(entityType,
                                                           entityOrder);
  
  nbKnots = getNumVerticesForElementTypeMSH(entityTag);

  knots.resize(nbKnots,3);
  
  vector<pVertex > verts;
  
  getVertices (ent,verts);
  double xyz[3];
  
  for(int i=0;i <int(verts.size());i++) {
    V_coord(verts[i], xyz);
    knots(i,0) = xyz[0];
    knots(i,1) = xyz[1];
    knots(i,2) = xyz[2];  
  }
  
  knots /= lengthScale;
  
  nbKnots = (int)(verts.size());
  getSerendipCoefficients();
  
  knotsCoef.resize(3,nbKnots);
  knotsCoef.blas_dgemm(knots.transpose(), coef, 1., 0.);
  
  }
// -----------------------------------------------------------------------------
int MAdMapping::order() const {
  return entityOrder;
}

// -----------------------------------------------------------------------------
int MAdMapping::geomOrder() const {
  return entityOrder;
}

// -----------------------------------------------------------------------------
void MAdMapping::evalLinear(double u, double v, double w, 
                            double &x, double &y, double &z) const{
  double xyz[3] = {0.,0.,0.};

  int basicTag = getBasicElementTypeMSH(entityTag);
  int np = getNumVerticesForElementTypeMSH(basicTag);

  double* shapeFun = new double[np];
  
  switch(entityType) {
  case MSH_TYPE_HEX: {
    shapeFun[0] = (1-u)*(1-v)*(1-w)/8.; 
    shapeFun[1] = (1+u)*(1-v)*(1-w)/8.;
    shapeFun[2] = (1+u)*(1+v)*(1-w)/8.;
    shapeFun[3] = (1-u)*(1+v)*(1-w)/8.; 
    shapeFun[4] = (1-u)*(1-v)*(1+w)/8.;
    shapeFun[5] = (1+u)*(1-v)*(1+w)/8.;
    shapeFun[6] = (1+u)*(1+v)*(1+w)/8.; 
    shapeFun[7] = (1-u)*(1+v)*(1+w)/8.;
    break;
  }
  case MSH_TYPE_PRI: {
    shapeFun[0] = (1-u-v)*(1-w)/2.;
    shapeFun[1] = (u)*(1-w)/2.;
    shapeFun[2] = (v)*(1-w)/2.;
    shapeFun[3] = (1-u-v)*(1+w)/2.;
    shapeFun[4] = (u)*(1+w)/2.;
    shapeFun[5] = (v)*(1+w)/2.;
    break;
  }
  default:
    MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                "evalLinear not implemented yet for type %d",
                                entityType);
    break;  
  } 
  
  for(int d=0; d<3; d++){
    for(int i=0; i<np; i++){
      xyz[d] += shapeFun[i] * knots(i,d);
    }
  } 

  x=xyz[0]; y=xyz[1]; z=xyz[2];   
  
  delete [] shapeFun;
}

// -----------------------------------------------------------------------------
void MAdMapping::eval(double u, double v, double w, 
                      double &x, double &y, double &z) const
{
  
  switch(entityType) {
  case MSH_TYPE_PYR: {
    double* f = new double[nbKnots];   
    GeomShapeFunction(u,v,w,f); 
    x = 0.0, y = 0.0, z = 0.0;
    for(int i=0;i<int(nbKnots);i++){
      double fct = f[i];
      x +=  knots(i,0) * fct;
      y +=  knots(i,1) * fct;
      z +=  knots(i,2) * fct;
    }
    delete[] f;
    break;
  }
  default:{
    doubleVector xyz(3);
    const int nbK=int(nbKnots);
    doubleVector monomes(nbK);
    monomeof(u,v,w,monomes);
    knotsCoef.mult(monomes,xyz,false,1.,0); // 1.); 
    
    x = xyz(0);
    y = xyz(1); 
    z = xyz(2);
    break;
  }
  }
}

// -----------------------------------------------------------------------------
void MAdMapping::deval(double u, double v, double w, 
                       double &dxdu, double &dydu, double &dzdu,
                       double &dxdv, double &dydv, double &dzdv,
                       double &dxdw, double &dydw, double &dzdw) const
{
  dxdu = 0.0, dydu = 0.0, dzdu = 0.0;
  dxdv = 0.0, dydv = 0.0, dzdv = 0.0;
  dxdw = 0.0, dydw = 0.0, dzdw = 0.0;
  const int nbK=int(nbKnots);
  doubleMatrix dus(nbK,3);
  
  GradGeomShapeFunction  (u ,  v ,  w, dus);
  
  for(int i=0;i<int(nbKnots);i++) {
    
    const double xx =  knots(i,0);
    const double yy =  knots(i,1);
    const double zz =  knots(i,2);
    
    dxdu += xx * dus(i,0);
    dydu += yy * dus(i,0);
    dzdu += zz * dus(i,0);

    dxdv += xx * dus(i,1);
    dydv += yy * dus(i,1);
    dzdv += zz * dus(i,1);

    dxdw += xx * dus(i,2);
    dydw += yy * dus(i,2);
    dzdw += zz * dus(i,2);
  }
}

// ----------------------------------------------------------------------------

bool MAdMapping::inReferenceElement(double u,double v,double w,
                                    double eps) const {
  switch(entityType) {
  case MSH_TYPE_LIN :
    if(u < -1.-eps || 
       u >  1.+eps) return false;
    break;
  case MSH_TYPE_TRI : 
    if(u < -eps || 
       v < -eps || 
       1.-u-v < -eps) return false;
    break;
  case MSH_TYPE_TET :
    if(u < -eps || 
       v < -eps || 
       w < -eps || 
       1.-u-v-w < -eps) return false;
    break;
  case MSH_TYPE_QUA :
    if((u < -1.-eps) ||
       (u >  1.+eps) ||
       (v >  1.+eps) ||
       (v < -1.-eps)) return false;
    break; 
  case MSH_TYPE_PYR :
    if((u < (-1.+w-eps)) || 
       (u > ( 1.-w+eps)) || 
       (v < (-1.+w-eps)) || 
       (v > ( 1.-w+eps)) || 
       (w < -eps) || 
       (w > (1.+eps))) return false;
    break;
  case MSH_TYPE_PRI :
    if((w < -1.-eps || 
        w > 1.+eps) || 
       (u < -eps || 
        v < -eps || 
        1.-u-v < -eps)) return false;
    break; 
  case MSH_TYPE_HEX :
    if( (u < -1.-eps) ||
        (u >  1.+eps) ||
        (v >  1.+eps) ||
        (v < -1.-eps) || 
        (w < -1.-eps) || 
        (w >  1.+eps)) return false;
    break;
  default: 
    break; // need to throw an exception here
  }
  return true;
}

// ----------------------------------------------------------------------------
double MAdMapping::parametricProjection(double& u,double& v, double& w, 
                                        const double eps) const {
  
  double oldU(u);
  double oldV(v);
  double oldW(w);
  
  double oneeps  = 1.-eps;
  double zeroeps = 0.+eps;
  
  switch (entityType) {
    
  case MSH_TYPE_LIN:
    {
      u = std::min(oneeps,max(-oneeps,u));
      break;
    }
  case MSH_TYPE_PRI: // same limitations on u,v as triangle
    {
      w = std::min(oneeps,max(-oneeps,w));
    }
  case MSH_TYPE_TRI:
    {
      u = max(zeroeps,u);
      v = max(zeroeps,v);
      double uPV = u + v;
      if (uPV > oneeps) {
        double uMV = u - v;
        if (uMV >  oneeps) {u = oneeps;v = zeroeps;}
        if (uMV < -oneeps) {u = zeroeps;v = oneeps;}
        u = (oneeps + u - v)*0.5;
        v = (oneeps - u); // normally (1-u+v)/2 but we modified u in place
      }
      break;
    }
  case MSH_TYPE_QUA:
    {
      u = std::min(oneeps,max(-oneeps,u));
      v = std::min(oneeps,max(-oneeps,v));
      break;
    }
  case MSH_TYPE_TET:
    {
      u = max(zeroeps,u);
      v = max(zeroeps,v);
      w = max(zeroeps,w);
    
      double un(u),vn(v),wn(w);

      // proj onto uw-edge
      if(w > oneeps-u+2*v){
        un = std::min(oneeps, max( 0.5*(oneeps+u-w) ,zeroeps));
        vn = zeroeps;
        wn = std::min(oneeps, max( 0.5*(oneeps-u+w) ,zeroeps));
      }
      // proj onto vw-edge
      else if(w > oneeps+2*u-v){
        un = zeroeps;
        vn = std::min(oneeps, max( 0.5*(oneeps+v-w) ,zeroeps));
        wn = std::min(oneeps, max( 0.5*(oneeps-v+w) ,zeroeps));
      }
      // proj onto uv-edge
      else if(w < (u+v-oneeps)/2.){
        un = std::min(oneeps, max( 0.5*(oneeps+u-v) ,zeroeps));
        vn = std::min(oneeps, max( 0.5*(oneeps-u+v) ,zeroeps));
        wn = zeroeps;
      }
      // proj onto u+v+w=1 plane
      else if(u+v+w>oneeps){
        un = (2*u-v-w+oneeps)/3;
        vn = (-u+2*v-w+oneeps)/3;
        wn = (-u-v+2*w+oneeps)/3;
      }

      u = un;
      v = vn;
      w = wn;

      break;
    }
  case MSH_TYPE_HEX:
    {
      u = std::min(oneeps,max(-oneeps,u));
      v = std::min(oneeps,max(-oneeps,v));
      w = std::min(oneeps,max(-oneeps,w));
      break;
    }
  case MSH_TYPE_PYR:
    {
      w = std::min(oneeps,max(zeroeps,w));
      u = std::min(oneeps-w,max(-oneeps+w,u));
      v = std::min(oneeps-w,max(-oneeps+w,v));
      break;
    }
  default:
    {
      throw;
      break;
    }
  }

  return std::sqrt((u-oldU)*(u-oldU) + 
                   (v-oldV)*(v-oldV) + 
                   (w-oldW)*(w-oldW));
} 

// ------------------------------------------------------------------
bool MAdMapping::projectAlongDirection(const double* ori,
                                       const double* dir,
                                       double* uvw,
                                       double* xyz,
                                       double& distance,
                                       double tol) const {

  MAd_Vector3D oriV(ori);
  MAd_Vector3D dirV(dir); 
  MAd_Vector3D xyzV;
  MAd_Vector3D uvwV;
  

  MAd_Vector3D e1,e2,e3;
  dirV.computeOrthonormalBasis(e1,e2,e3);
  
  COG (uvwV[0],uvwV[1],uvwV[2]);
  eval(uvwV[0],uvwV[1],uvwV[2],
       xyzV[0],xyzV[1],xyzV[2]);

  //! first project the point on the plane through the COG, normal to dir

  MAd_Vector3D dr(oriV,xyzV);
  oriV -= (dr*e1)*e1;
  dr = xyzV - oriV;
  
  //! now perform Newton-Raphson iterations to find the projection

  switch (entityDim) {

  case 1: 
    {
      throw;
      break;
    }
  case 2: 
    {

      MAd_Vector3D drdu;
      MAd_Vector3D drdv;
      MAd_Vector3D drdw;

      deval(uvwV[0],uvwV[1],uvwV[2],
            drdu[0],drdu[1],drdu[2],
            drdv[0],drdv[1],drdv[2],
            drdw[0],drdw[1],drdw[2]);

      double overResRef = 1./std::sqrt(drdu*drdu + drdv*drdv);

      e2 /= overResRef;
      e3 /= overResRef;

      int maxIterations = 20;
      int nbIterations = 0;

      doubleMatrix A(2,2);
      
      doubleVector res(2);
      res(0) = e2 * dr;
      res(1) = e3 * dr;
      
      doubleVector duv(2);
      doubleVector uv(2);
      uv(0)=uvwV[0];
      uv(1)=uvwV[1];
      
      while (res.norm() > 1e-10 && nbIterations++ < maxIterations) { 
        
        A(0,0) = e2*drdu;
        A(0,1) = e2*drdv;
        
        A(1,0) = e3*drdu;
        A(1,1) = e3*drdv;
        A=A.inverse();
        
        A.mult(res,uv,false,-1,1);
        uvwV[0]=uv(0);
        uvwV[1]=uv(1);

        eval(uvwV[0],uvwV[1],uvwV[2],
             xyzV[0],xyzV[1],xyzV[2]);
       
        deval(uvwV[0],uvwV[1],uvwV[2],
              drdu[0],drdu[1],drdu[2],
              drdv[0],drdv[1],drdv[2],
              drdw[0],drdw[1],drdw[2]);

        
        dr = xyzV - oriV;
        
        res(0) = e2*dr;
        res(1) = e3*dr;
      }
      
      if (nbIterations > maxIterations) return false;

      break;
    }
  case 3: // not applicable ? 
    {
      throw;
      break;
    }
  }

  std::memcpy(uvw,uvwV.data(),3*sizeof(double));
  std::memcpy(xyz,xyzV.data(),3*sizeof(double));
  
  distance = dr.norm();
  return inReferenceElement(uvwV[0],uvwV[1],uvwV[2],tol);
}

// -----------------------------------------------------------
//! \todo replace by simple arithmetic average of vertex coordinates
void MAdMapping::COG(double &u, double &v, double &w) const{
  switch(entityType){
  case MSH_TYPE_LIN:
    u = v = w = 0;
    break;
  case MSH_TYPE_TRI:
    u = v = 1./3.;
    w = 0;
    break;
  case MSH_TYPE_QUA:
    u = v = w = 0;
    break;
  case MSH_TYPE_TET:
    u = v = w = 1./4.;
    break;
  case MSH_TYPE_HEX:
    u = v = w = 0;
    break;
  case MSH_TYPE_PRI:
    u = v = 1./3.;
    w = 0;
    break;
  case MSH_TYPE_PYR:
    u = v = 0;
    w = 1./5.;
    break;
  default: 
    throw; 
    break; // need to throw an exception here
  }
}

// -----------------------------------------------------------------------------
bool MAdMapping::project(double  xo,double  yo,double  zo,
                         double& up,double& vp,double& wp,
                         double& distance) const {
  

  bool inClosure = true;
  
  switch (entityDim) {

  case 0: // trivial 
    { 
      double xp,yp,zp;
      COG(up,vp,wp);
      eval(up,vp,wp,xp,yp,zp);
      
      double dx = xo - xp;
      double dy = yo - yp;
      double dz = zo - zp;
      
      distance = std::sqrt(dx*dx+dy*dy+dz*dz);
      
      break;
    }
  case 1: // orthogonalise projection vector wrt parametric coordinate gradient
    {
      
      double tolerance = 1e-6;
      double res       = 1;      
      double xp,yp,zp;

      up = std::min(max(up,-1.),1.);
      
      eval(up,vp,wp,xp,yp,zp);

      int nbIter = 100;
      int iter = 0;
      
      while (res > tolerance && iter++ <nbIter) {

        eval(up,vp,wp,xp,yp,zp);
        
        double dxdu,dydu,dzdu;
        double dxdv,dydv,dzdv;
        double dxdw,dydw,dzdw;
        
        deval(up,vp,wp,
              dxdu,dydu,dzdu,
              dxdv,dydv,dzdv,
              dxdw,dydw,dzdw);

        double g11 = dxdu*dxdu + dydu*dydu + dzdu*dzdu ;
        
        double dx = xo - xp;
        double dy = yo - yp;
        double dz = zo - zp;

        double ds   = std::sqrt(dx*dx+dy*dy+dz*dz);
        double prod = (dx*dxdu + dy*dydu + dz*dzdu);
        
        res = fabs(prod/(ds*std::sqrt(g11)));

        up += prod/g11;
        
        if (up > 1.) {
          up = 1.;
          inClosure = false;
          break;
        }

        if (up < -1.) {
          up = -1.;
          inClosure = false;
          break;
        }
      }
      break;
    }    
  case 2:
    {     
      project2D(xo,yo,zo,up,vp,wp,distance);

      break;
    }
  default:
    {
      exit(1);
      break;
    }
  }
  
  double xp,yp,zp;
  eval(up,vp,wp,xp,yp,zp);
  distance = std::sqrt((xp-xo)*(xp-xo) + (yp-yo)*(yp-yo) + (zp-zo)*(zp-zo));

  return inClosure;  
  
}

// -----------------------------------------------------------------------------

bool MAdMapping::project2D(double  xo,double  yo,double  zo,
                           double& up,double& vp,double& wp,
                           double& distance) const {

  
  int nbConst(0);
  double C[4][2];
  double b[4]; 
  vector<int> aConst;

  const double tol = 1e-12;
  const double eps = 1e-6;

  // Parametric constraints: A.[u,v]-b >= 0
  // --------------------------------------

  switch(entityType){
  case MSH_TYPE_TRI:
    { 
      C[0][0]= 1.; C[0][1]= 0.;
      C[1][0]= 0.; C[1][1]= 1.;
      C[2][0]=-1.; C[2][1]=-1.;
      C[3][0]= 0.; C[3][1]= 0.;
          
      b[0]= 0.;
      b[1]= 0.;
      b[2]=-1.;
      b[3]= 0.;
          
      nbConst =3;	  
      break;
    }
        
  case MSH_TYPE_QUA:
    {
      C[0][0]= 1.; C[0][1]= 0.; // u>=-1
      C[1][0]=-1.; C[1][1]= 0.; // u<=1
      C[2][0]= 0.; C[2][1]= 1.; // v>=-1
      C[3][0]= 0.; C[3][1]=-1.; // v<=1
          
      b[0]=-1.;
      b[1]=-1.;
      b[2]=-1.;
      b[3]=-1.;
          
      nbConst =4;	  
      break;
    }
  }



  // Initial guess
  // ------------- 

  doubleVector sol(2+nbConst);
  
  sol    = 1.0;
  sol(0) = up;
  sol(1) = vp;

  for(int j=0; j<nbConst; j++){
    double constraint = C[j][0]*up+C[j][1]*vp-b[j];
    if(constraint < -tol){
      MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                  "Projection: initial guess outside"
                                  " of element %e %e\n",up,vp);
    }    
  }

  int ite = 0;
  bool found = false;
  double norm = 1.0;

  for(; ite<30 && !found; ite++){  

    up = sol(0);
    vp = sol(1);
  
    doubleVector F(2+nbConst);
      
    double f[2];
    distanceJacobianProduct(xo,yo,zo,up,vp,wp,f);
    F(0) = f[0];
    F(1) = f[1];
    
    for(int j=0; j<nbConst; j++){
      F(0)  -=  C[j][0]*sol(2+j);
      F(1)  -=  C[j][1]*sol(2+j);
      F(2+j) = (C[j][0]*up + C[j][1]*vp - b[j])*sol(2+j);
    }

    // Stopping criterion
    // ------------------

    norm = F.norm();
    if(norm < tol) found = true;

    // Jacobian
    // --------

    doubleMatrix J(2+nbConst,2+nbConst);
    
    double fEps[2];

    distanceJacobianProduct(xo,yo,zo,up+eps,vp,wp,fEps);
    J(0,0) = (fEps[0]-f[0])/eps;
    J(1,0) = (fEps[1]-f[1])/eps;

    distanceJacobianProduct(xo,yo,zo,up,vp+eps,wp,fEps);
    J(0,1) = (fEps[0]-f[0])/eps;
    J(1,1) = (fEps[1]-f[1])/eps;

    for(int j=0; j<nbConst; j++){
      J(0,2+j)   = -C[j][0];
      J(1,2+j)   = -C[j][1];
      J(2+j,0)   = C[j][0]*sol(2+j);
      J(2+j,1)   = C[j][1]*sol(2+j);
      J(2+j,2+j) = C[j][0]*up + C[j][1]*vp - b[j];
    }
    
    J = J.inverse();

    // Linear solve
    // ------------

    doubleVector dSol(2+nbConst);
    J.mult(F,dSol,false,1.,0.);
    
    // Keep constraints non-negative
    // -----------------------------
    double alpha =1;
    for(int j=0; j<nbConst; j++){
      double  c = C[j][0]*up + C[j][1]*vp - b[j];
      double  d = C[j][0]*dSol(0) + C[j][1]*dSol(1);
      if(d > tol) alpha = std::min(alpha,c/d);
    }

    // Update: sol' = sol - a*dSol
    // ---------------------------
    dSol *= alpha;     
    sol  -= dSol; 
  }
    
  if(!found){
    MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                "Calculation of closest point failed,"
                                " %d %e",ite,norm);
  }

  double xp,yp,zp;
  eval(up,vp,wp,xp,yp,zp);
  distance = std::sqrt((xp-xo)*(xp-xo)+(yp-yo)*(yp-yo)+(zp-zo)*(zp-zo));

  return true;
}

// -----------------------------------------------------------------------------
  
void MAdMapping::distanceJacobianProduct(double xo,double yo,double zo,
                                         double up,double vp,double wp,
                                         double* P) const {
  
  double xpIter,ypIter,zpIter;
  eval(up,vp,wp,xpIter,ypIter,zpIter);

  double dxdu,dydu,dzdu;
  double dxdv,dydv,dzdv;
  double dxdw,dydw,dzdw;
        
  deval(up,vp,wp,
        dxdu,dydu,dzdu,
        dxdv,dydv,dzdv,
        dxdw,dydw,dzdw);
        
  double dx = xpIter - xo;
  double dy = ypIter - yo;
  double dz = zpIter - zo;

  P[0] = dx*dxdu + dy*dydu + dz*dzdu;
  P[1] = dx*dxdv + dy*dydv + dz*dzdv;
}


// -----------------------------------------------------------------------------
void MAdMapping::invertLinear(double x, double y, double z,
                              double &u, double &v, double &w) const{
  
  switch(entityType){
  case MSH_TYPE_HEX:
    {
      doubleVector b(3);
      b(0) = 8*x;
      b(1) = 8*y; 
      b(2) = 8*z;

      doubleMatrix M(3,3);
      for(int i=0; i<3; i++){
        M(i,0) =-knots(0,i)+knots(1,i)+knots(2,i)-knots(3,i)
          -knots(4,i)+knots(5,i)+knots(6,i)-knots(7,i);
        M(i,1) =-knots(0,i)-knots(1,i)+knots(2,i)+knots(3,i)
          -knots(4,i)-knots(5,i)+knots(6,i)+knots(7,i); 
        M(i,2) =-knots(0,i)-knots(1,i)-knots(2,i)-knots(3,i)
          +knots(4,i)+knots(5,i)+knots(6,i)+knots(7,i);
        
        for(int j=0; j<8; j++) b(i) -= knots(j,i);
      }
      
      doubleVector uvw(3);
      M=M.inverse();   
      M.mult(b,uvw,false,1,0);
      u = uvw(0); 
      v = uvw(1);  
      w = uvw(2);
      break;
    }
  case MSH_TYPE_TET:
    {
      doubleVector b(3);
      b(0) = x-knots(0,0);
      b(1) = y-knots(0,1); 
      b(2) = z-knots(0,2);

      doubleMatrix M(3,3);
      for(int i=0; i<3; i++){
        M(i,0) = knots(1,i)-knots(0,i);
        M(i,1) = knots(2,i)-knots(0,i); 
        M(i,2) = knots(3,i)-knots(0,i);
      }
      
      doubleVector uvw(3);
      M=M.inverse();   
      M.mult(b,uvw,false,1,0);
      u = uvw(0); 
      v = uvw(1);  
      w = uvw(2);
      break;
    }
  }
}

// -----------------------------------------------------------------------------
bool MAdMapping::invert(double xp, double yp, double zp,
                        double &Upos, double &Vpos, double &Wpos,
                        double eps, int maxiter,int nbSamples) const
{
  
  
  double   x_est, y_est, z_est;
  doubleMatrix InvJacMatrix(3,3);
  double Error = std::numeric_limits<double>::max();
  
  if (nbSamples == 1) {
    COG(Upos,Vpos,Wpos);
    invertLinear(xp,yp,zp,Upos,Vpos,Wpos);
    
    eval(Upos,Vpos,Wpos,x_est,y_est,z_est);
    
    Error = std::sqrt(((xp-x_est)*(xp-x_est) + 
                       (yp-y_est)*(yp-y_est) + 
                       (zp-z_est)*(zp-z_est))*overD);
    
    if (Error < eps) return true;
  }
    
  if (nbSamples > 1) {
    
    int iPt = dim() > 0 ? nbSamples : 1;
    int jPt = dim() > 1 ? nbSamples : 1;
    int kPt = dim() > 2 ? nbSamples : 1;
    
    for(int i=0; i<iPt; i++){
      for(int j=0; j<jPt; j++){
        for(int k=0; k<kPt; k++){
          
          double U0 = 2*(i+1)/(iPt+1.)-1;
          double V0 = 2*(j+1)/(jPt+1.)-1;
          double W0 = 2*(k+1)/(kPt+1.)-1;

          if(inReferenceElement(U0,V0,W0)){
            
            eval(U0,V0,W0,x_est,y_est,z_est);
          
            double trial = std::sqrt(((xp-x_est)*(xp-x_est) + 
                                      (yp-y_est)*(yp-y_est) + 
                                      (zp-z_est)*(zp-z_est))*overD);
          
            if(trial < Error){
              Error = trial;
              Upos  = U0;
              Vpos  = V0;
              Wpos  = W0;
            }
          }
        }
      }
    }
  }

  int iter = 0;
  
  while (Error > eps && iter++ < maxiter){

    eval(Upos,Vpos,Wpos,x_est,y_est,z_est);
    
    Error = std::sqrt(((xp-x_est)*(xp-x_est) + 
                       (yp-y_est)*(yp-y_est) + 
                       (zp-z_est)*(zp-z_est))*overD);

    if (Error < eps) break;

    invertedJacobian(Upos,Vpos,Wpos,InvJacMatrix);

    Upos += (InvJacMatrix(0,0) * (xp-x_est) + 
             InvJacMatrix(1,0) * (yp-y_est) +
             InvJacMatrix(2,0) * (zp-z_est)) ;
    Vpos += (InvJacMatrix(0,1) * (xp-x_est) + 
             InvJacMatrix(1,1) * (yp-y_est) +
             InvJacMatrix(2,1) * (zp-z_est)) ;
    Wpos += (InvJacMatrix(0,2) * (xp-x_est) + 
             InvJacMatrix(1,2) * (yp-y_est) +
             InvJacMatrix(2,2) * (zp-z_est)) ;
    
  }

  if (Error < eps) return true;
  return false;
}

// -----------------------------------------------------------------------------

double MAdMapping::invertedJacobian(const double* uvw,double* invJac) const {

  double jac[3][3];
  
  double detJac = 0;

  // jac_ij = dxj/dui
  deval (uvw[0],uvw[1],uvw[2],
         jac[0][0],jac[0][1],jac[0][2],
         jac[1][0],jac[1][1],jac[1][2],
         jac[2][0],jac[2][1],jac[2][2]);
  
  switch(entityDim) {
    
  case 1:
    {
      detJac = sqrt ( jac[0][0]*jac[0][0] + 
                      jac[0][1]*jac[0][1] +
                      jac[0][2]*jac[0][2] );
      
      int vrai=0;
      for(int i=0;i<3;i++){
        if(jac[0][i]==0){   // if this component of the normal vector is zero
          vrai =1;
          for(int j=0;j<3;j++){
            if (j==i)
              jac[1][j] = 1; // then the component of the second normal must be one,
            else
              jac[1][j] = 0; // and the other components of the second normal are zero.
          }
          continue;
        }
      }
      // surface equation with normal vector n : n1*X + n2*Y + n3*Z = 0
      if(!vrai){ // looking for the second normal vector in the plane z = 0
        double temp = sqrt ( jac[0][0]*jac[0][0] + jac[0][1]*jac[0][1] );
        jac[1][0] = -jac[0][1]/temp;
        jac[1][1] =  jac[0][0]/temp;
        jac[1][2] =  0;
      }
      // The third normal vector
      jac[2][0] = ( jac[0][1]*jac[1][2] - jac[0][2]*jac[1][1] )/ detJac;
      jac[2][1] = ( jac[0][2]*jac[1][0] - jac[0][0]*jac[1][2] )/ detJac;
      jac[2][2] = ( jac[0][0]*jac[1][1] - jac[0][1]*jac[1][0] )/ detJac; 
    
      break;
    }
  case 2: 
    {
      double d3 = jac[0][0]*jac[1][1] - jac[0][1] * jac[1][0];   
      double d2 = jac[0][2]*jac[1][0] - jac[0][0] * jac[1][2];  
      double d1 = jac[0][1]*jac[1][2] - jac[0][2] * jac[1][1];  
      
      detJac = sqrt ( d1*d1 + d2*d2 + d3*d3 );
      
      jac[2][0] = d1/detJac;  
      jac[2][1] = d2/detJac; 
      jac[2][2] = d3/detJac;
      
      break;
    }
  case 3:
    {
      detJac = (jac[0][0]*jac[1][1]*jac[2][2] + jac[0][2] *jac[1][0]*jac[2][1] +
                jac[0][1]*jac[1][2]*jac[2][0] - jac[0][2] *jac[1][1]*jac[2][0] -
                jac[0][0]*jac[1][2]*jac[2][1] - jac[0][1] *jac[1][0]*jac[2][2]);
    }
    break;
  }

  // out InvJac_ij = duj/dxi
  
  double overDJ = 1/detJac;
  
  invJac[0*3+0] = (jac[1][1] * jac[2][2] - jac[1][2] * jac[2][1]) *overDJ;
  invJac[0*3+1] = -(jac[1][0] * jac[2][2] - jac[1][2] * jac[2][0]) *overDJ;
  invJac[0*3+2] = (jac[1][0] * jac[2][1] - jac[1][1] * jac[2][0]) *overDJ;

  invJac[1*3+0] = -(jac[0][1] * jac[2][2] - jac[0][2] * jac[2][1]) *overDJ;
  invJac[1*3+1] = (jac[0][0] * jac[2][2] - jac[0][2] * jac[2][0]) *overDJ;
  invJac[1*3+2] = -(jac[0][0] * jac[2][1] - jac[0][1] * jac[2][0]) *overDJ;

  invJac[2*3+0] = (jac[0][1] * jac[1][2] - jac[0][2] * jac[1][1]) *overDJ;
  invJac[2*3+1] = -(jac[0][0] * jac[1][2] - jac[0][2] * jac[1][0]) *overDJ;
  invJac[2*3+2] = (jac[0][0] * jac[1][1] - jac[0][1] * jac[1][0]) *overDJ;
  
  return detJac;
}
// -----------------------------------------------------------------------------

void MAdMapping::metricTensor(const double* uvw,double* metric) const {

  double jac[3][3];
  // jac_ij = dxj/dui
  deval (uvw[0],uvw[1],uvw[2],
         jac[0][0],jac[0][1],jac[0][2],
         jac[1][0],jac[1][1],jac[1][2],
         jac[2][0],jac[2][1],jac[2][2]);

  int idx=0;
  for (int i=0;i<3;i++) {
    for (int j=0;j<3;j++,idx++) {
      metric[idx] = 0;
      for (int k=0;k<3;k++) metric[idx] += jac[i][k] * jac[j][k];
    }
  }
}

// -------------------------------------------------------------------------
double MAdMapping::invertedJacobian(double u,double v,double w,
                                    doubleMatrix &invJac) const {

  double ij[9];
  double uvw[3] = {u,v,w};
  double dj = invertedJacobian(uvw,ij);
  
  for (int i=0;i<3;i++) for (int j=0;j<3;j++) invJac(i,j) = ij[j*3+i];
  return dj;
}

// -----------------------------------------------------------------------------
void MAdMapping::normalVector(pEntity border,
                              double u,double v,double w,
                              double n[3]) const {
  
  // In case of a mesh mapping, the normal to an entity which is on
  // a border = sum of grad geom shape functions which node are not
  // on the border entity
  
  n[0] = n[1] = n[2] = 0.0;
  doubleMatrix jInv(3,3);
  invertedJacobian(u,v,w,jInv);

  vector<pVertex> verts;
  getVertices (ent,verts);
  vector<pVertex > vb;
  getVertices (border,vb);

  const int nbK=int(nbKnots);
  doubleMatrix dus(nbK,3);
  GradGeomShapeFunction  (u ,  v ,  w, dus);
  for(int i=0;i<int(nbKnots);i++){
    vector<pVertex>::iterator it = find (vb.begin(),vb.end(),verts[i]);
    if(it == vb.end()){
      doubleVector gr(3);
      gr(0) = dus(i,0);
      gr(1) = dus(i,1);
      gr(2) = dus(i,2);
      doubleVector grT(3);
      jInv.mult(gr,grT,false,1,0);
      n[0]-= grT(0);
      n[1]-= grT(1);
      n[2]-= grT(2);
    }
  }
  norme(n);
}

// ------------------------------------------------------------------------------
void MAdMapping::normalVector (pEntity border ,
                               int direction,
                               int rotation,
                               const double* uvw,
                               double* normal) const {
  
  switch (entityDim) {
  case 2: {
    int edgePosition = getEdgeNumber(border);
    double u,v,w(0);
    getEdgePointsOnFace(edgePosition, direction,uvw[0],u,v);
    normalVector(border, u, v, w, normal);
    break;
  }
  case 3: {
    int facePosition = getFaceNumber(border);
    double u,v,w;
    getFacePointsOnElement(facePosition,direction,rotation,uvw[0],uvw[1],u,v,w);
    normalVector(border,u,v,w,normal);
    break;
  }
  default:
    {
      MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                  "Cannot compute boundary normals "
                                  "for elements of dimension %d",
                                  entityDim);
      throw;
    }
  }
}


// -----------------------------------------------------------------------------
double MAdMapping::determinant(const double* uvw) const  {

  double jac[3][3];
  double detJac = 0;
  
  deval (uvw[0],uvw[1],uvw[2],
         jac[0][0],jac[0][1],jac[0][2],
         jac[1][0],jac[1][1],jac[1][2],
         jac[2][0],jac[2][1],jac[2][2]);

  switch(entityDim) {
    
  case 1:
    {
      detJac = sqrt ( jac[0][0]*jac[0][0] + 
                      jac[0][1]*jac[0][1] +
                      jac[0][2]*jac[0][2] );
      break;
    }
  case 2: 
    {
      double d3 = jac[0][0]*jac[1][1] - jac[0][1] * jac[1][0];   
      double d2 = jac[0][2]*jac[1][0] - jac[0][0] * jac[1][2];  
      double d1 = jac[0][1]*jac[1][2] - jac[0][2] * jac[1][1];  
      
      detJac = sqrt ( d1*d1 + d2*d2 + d3*d3 );
      break;
    }
  case 3:
    {
      detJac = (jac[0][0]*jac[1][1]*jac[2][2] + jac[0][2] *jac[1][0]*jac[2][1] +
                jac[0][1]*jac[1][2]*jac[2][0] - jac[0][2] *jac[1][1]*jac[2][0] -
                jac[0][0]*jac[1][2]*jac[2][1] - jac[0][1] *jac[1][0]*jac[2][2]);
      break;
    }
  }
  return detJac;
}

// -----------------------------------------------------------------------------
void MAdMapping::boundingBox (double bbox[3][2]) const {
  
  for (int k=0;k<3;k++) {
    bbox[k][0] = bbox[k][1] = knots(0,k);
    for(int i=1;i<int(nbKnots);i++) { 
      bbox[k][0] = std::min(bbox[k][0],knots(i,k));
      bbox[k][1] = std::max(bbox[k][1],knots(i,k));
    }
  }
}

// -----------------------------------------------------------------------------
bool MAdMapping::getVertexUVW(int i,double& u,double& v,double& w) const {

  if(EN_mshType((pEntity) ent)==MSH_TYPE_PNT){
    u = v = w = 0;
  }
  else{
    if(i < interpolant->size()){
      interpolant->nodeNumberCoordinate(i,u,v,w);    
    }
    else{
      u = v = w = 0;
      MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                  "Parametric node %d >= %d",
                                  i,interpolant->size());
      return false;
    }
  }
  return true;
}

// -----------------------------------------------------------------------------
void MAdMapping::getEdgePointsOnFace(int k, int dir, double quadPt, double &u, double &v) const {
  // applies a linear mapping from the 1D reference element to the 2D reference face's edge number k
  double x1,y1,x2,y2;
  switch (numVertices) {
  case 3:
    x1=triNodeUVW[triEdgeToNodes[k][0]][0];
    y1=triNodeUVW[triEdgeToNodes[k][0]][1];
    x2=triNodeUVW[triEdgeToNodes[k][1]][0];
    y2=triNodeUVW[triEdgeToNodes[k][1]][1];
    break;
  case 4:
    x1=quadNodeUVW[quadEdgeToNodes[k][0]][0];
    y1=quadNodeUVW[quadEdgeToNodes[k][0]][1];
    x2=quadNodeUVW[quadEdgeToNodes[k][1]][0];
    y2=quadNodeUVW[quadEdgeToNodes[k][1]][1];
    break;
  default:
    throw;
  }
  
  if (dir==-1){
    double tempx=x1;
    double tempy = y1;
    x1=x2;y1=y2;x2=tempx;y2=tempy;
  }
  double phi1 = 0.5*(1-quadPt);
  double phi2 = 0.5*(1+quadPt);
  u = x1*phi1 + x2*phi2;
  v = y1*phi1 + y2*phi2;
}

// -----------------------------------------------------------------------------
void MAdMapping::getFacePointsOnElement(int k, int dir, int rot,
                                        double ksi,double eta,
                                        double &u,double &v,double &w) const {
  
  // applies a linear mapping from the 2D reference face to the 3D reference element
  
  int mshTag = getElementTypeMSH(EN_mshTag(ent));
  switch(mshTag) {
  case MSH_TYPE_TET: 
    {
      MAd_Vector3D tetFaceNodeUVW[3];  
    
      tetFaceNodeUVW[0] = tetNodeUVW[tetFaceToNodes[k][0]];
      tetFaceNodeUVW[1] = tetNodeUVW[tetFaceToNodes[k][1]];
      tetFaceNodeUVW[2] = tetNodeUVW[tetFaceToNodes[k][2]];
    
      // rotate around positive direction
    
      MAd_Vector3D faceNodeUVW[3];
    
      for (int i=0;i<3;i++) faceNodeUVW[i] = tetFaceNodeUVW[(dir*i+rot+3)%3];
      
      MAd_Vector3D uvw = (1.-ksi-eta) * faceNodeUVW[0] + ksi*faceNodeUVW[1] + eta*faceNodeUVW[2];

      u = uvw[0];
      v = uvw[1];
      w = uvw[2];  
      break;
    }
  case MSH_TYPE_HEX: 
    {
      MAd_Vector3D hexFaceNodeUVW[4];  
    
      hexFaceNodeUVW[0] = hexNodeUVW[hexFaceToNodes[k][0]];
      hexFaceNodeUVW[1] = hexNodeUVW[hexFaceToNodes[k][1]];
      hexFaceNodeUVW[2] = hexNodeUVW[hexFaceToNodes[k][2]];
      hexFaceNodeUVW[3] = hexNodeUVW[hexFaceToNodes[k][3]];

      MAd_Vector3D faceNodeUVW[4];
      for (int i=0;i<4;i++) faceNodeUVW[i] = hexFaceNodeUVW[(dir*i+rot+4)%4];

      MAd_Vector3D uvw = 0.25*((1.-ksi)*(1.-eta)*faceNodeUVW[0] + 
                               (1.+ksi)*(1.-eta)*faceNodeUVW[1] + 
                               (1.+ksi)*(1.+eta)*faceNodeUVW[2] + 
                               (1.-ksi)*(1.+eta)*faceNodeUVW[3]);
      
      u = uvw[0];
      v = uvw[1];
      w = uvw[2];  
      
      break;
    }
  case MSH_TYPE_PRI: 
    {
      if (k<2){
        MAd_Vector3D priFaceNodeUVW[3];
        priFaceNodeUVW[0] = prismNodeUVW[prismFaceToNodesTri[k][0]];
        priFaceNodeUVW[1] = prismNodeUVW[prismFaceToNodesTri[k][1]];
        priFaceNodeUVW[2] = prismNodeUVW[prismFaceToNodesTri[k][2]];
        
        MAd_Vector3D faceNodeUVW[3];
        for (int i=0;i<3;i++) faceNodeUVW[i] = priFaceNodeUVW[(dir*i+rot+3)%3];
        
        MAd_Vector3D uvw = (1.-ksi-eta) * faceNodeUVW[0] + ksi*faceNodeUVW[1] + eta*faceNodeUVW[2];
        
        u = uvw[0];
        v = uvw[1];
        w = uvw[2];  
      }
      else {
        MAd_Vector3D priFaceNodeUVW[4];
        priFaceNodeUVW[0] = prismNodeUVW[prismFaceToNodesQuad[k-2][0]];
        priFaceNodeUVW[1] = prismNodeUVW[prismFaceToNodesQuad[k-2][1]];
        priFaceNodeUVW[2] = prismNodeUVW[prismFaceToNodesQuad[k-2][2]];
        priFaceNodeUVW[3] = prismNodeUVW[prismFaceToNodesQuad[k-2][3]];
      
        MAd_Vector3D faceNodeUVW[4];
        for (int i=0;i<4;i++) faceNodeUVW[i] = priFaceNodeUVW[(dir*i+rot+4)%4];
      
        MAd_Vector3D uvw = 0.25*((1.-ksi)*(1.-eta)*faceNodeUVW[0] + 
                                 (1.+ksi)*(1.-eta)*faceNodeUVW[1] + 
                                 (1.+ksi)*(1.+eta)*faceNodeUVW[2] + 
                                 (1.-ksi)*(1.+eta)*faceNodeUVW[3]);

        u = uvw[0];
        v = uvw[1];
        w = uvw[2];  
      }
      break;
    }

  case MSH_TYPE_PYR: {
    if (k==0){
      MAd_Vector3D pyrFaceNodeUVW[4];
      pyrFaceNodeUVW[0] = pyramidNodeUVW[pyramidFaceToNodesQuad[k][0]];
      pyrFaceNodeUVW[1] = pyramidNodeUVW[pyramidFaceToNodesQuad[k][1]];
      pyrFaceNodeUVW[2] = pyramidNodeUVW[pyramidFaceToNodesQuad[k][2]];
      pyrFaceNodeUVW[3] = pyramidNodeUVW[pyramidFaceToNodesQuad[k][3]];
      
      MAd_Vector3D faceNodeUVW[4];
      for (int i=0;i<4;i++) faceNodeUVW[i] = pyrFaceNodeUVW[(dir*i+rot+4)%4];
      
      MAd_Vector3D uvw = 0.25*((1.-ksi)*(1.-eta)*faceNodeUVW[0] + 
                               (1.+ksi)*(1.-eta)*faceNodeUVW[1] + 
                               (1.+ksi)*(1.+eta)*faceNodeUVW[2] + 
                               (1.-ksi)*(1.+eta)*faceNodeUVW[3]);

      u = uvw[0];
      v = uvw[1];
      w = uvw[2];  
    }
    else {
      MAd_Vector3D pyrFaceNodeUVW[3];
      pyrFaceNodeUVW[0] = pyramidNodeUVW[pyramidFaceToNodesTri[k-1][0]];
      pyrFaceNodeUVW[1] = pyramidNodeUVW[pyramidFaceToNodesTri[k-1][1]];
      pyrFaceNodeUVW[2] = pyramidNodeUVW[pyramidFaceToNodesTri[k-1][2]];
      
      MAd_Vector3D faceNodeUVW[3];
      for (int i=0;i<3;i++) faceNodeUVW[i] = pyrFaceNodeUVW[(dir*i+rot+3)%3];
      
      MAd_Vector3D uvw = (1.-ksi-eta) * faceNodeUVW[0] + ksi*faceNodeUVW[1] + eta*faceNodeUVW[2];
      
      u = uvw[0];
      v = uvw[1];
      w = uvw[2];  
    }
    break;
  }
  default: 
    throw;
  }
  return;
}

// -----------------------------------------------------------------------------
int MAdMapping::getEdgeNumber(pEntity e) const{
  //return the position (0,1,2 or 3) of the edge e (according to face ent) in the data base
  assert(EN_type(e)==1);
  for (int iedge=0;iedge<numVertices;iedge++){
    if (F_edge((pFace)ent,iedge)==(pEdge)e){
      return iedge;
    }
  }
  throw;
}

// -----------------------------------------------------------------------------
int MAdMapping::getFaceNumber(pEntity e) const {

  assert(EN_type(e) == 2);
  
  int numFaces;
  
  switch(entityType){
  case MSH_TYPE_TET: {numFaces = 4; break;}
  case MSH_TYPE_HEX: {numFaces = 6; break;} 
  case MSH_TYPE_PRI: {numFaces = 5; break;}
  case MSH_TYPE_PYR: {numFaces = 5; break;}
  default: throw;
  }
  for (int iface=0;iface<numFaces;iface++) 
    if (R_face((pRegion) ent,iface) == (pFace) e) return iface;
  throw;
}

// -----------------------------------------------------------------------------
double MAdMapping::getMaxRadius() const {
  double maxRadius;
  
  switch(entityType) {
  case MSH_TYPE_TET: {

    double xyz0[3], xyz1[3], xyz2[3], xyz3[3], mat[3][3], b[3], res[3];

    pPList vertices = R_vertices((pRegion) ent);

    void *iter=0;
    pVertex v0 = (pVertex) PList_next(vertices, &iter);
    pVertex v1 = (pVertex) PList_next(vertices, &iter);
    pVertex v2 = (pVertex) PList_next(vertices, &iter);
    pVertex v3 = (pVertex) PList_next(vertices, &iter);
    PList_delete(vertices);
    V_coord(v0, xyz0);
    V_coord(v1, xyz1);
    V_coord(v2, xyz2);
    V_coord(v3, xyz3);

    b[0] = (xyz1[0]*xyz1[0] - xyz0[0]*xyz0[0] +
            xyz1[1]*xyz1[1] - xyz0[1]*xyz0[1] +
            xyz1[2]*xyz1[2] - xyz0[2]*xyz0[2]);
    b[1] = (xyz2[0]*xyz2[0] - xyz1[0]*xyz1[0] +
            xyz2[1]*xyz2[1] - xyz1[1]*xyz1[1] +
            xyz2[2]*xyz2[2] - xyz1[2]*xyz1[2]);
    b[2] = (xyz3[0]*xyz3[0] - xyz2[0]*xyz2[0] +
            xyz3[1]*xyz3[1] - xyz2[1]*xyz2[1] +
            xyz3[2]*xyz3[2] - xyz2[2]*xyz2[2]);

    for (int i=0; i<3; i++){
      mat[0][i] = xyz1[i] - xyz0[i];
      mat[1][i] = xyz2[i] - xyz1[i];
      mat[2][i] = xyz3[i] - xyz2[i];
    }
    if(!sys3x3(mat, b, res)) {
      fprintf(stderr, "Coplanar points in circum sphere computation - region (%d,%d,%d,%d) is probably a sliver !!\n",
              EN_id((pEntity) v0),EN_id((pEntity) v1),EN_id((pEntity) v2),EN_id((pEntity) v3));
      res[0] = res[1] = res[2] = 10.0e10;
    }

    maxRadius = res[0];
    break;
  }
  case MSH_TYPE_HEX: 
  case MSH_TYPE_PRI: 
  case MSH_TYPE_PYR: {
    maxRadius = R_inscrRad((pRegion) ent);
    break;
  }
  default: throw;
  }
  return maxRadius;
}

// -----------------------------------------------------------------------------
void MAdMapping::getSerendipCoefficients(){

  coef.resize(nbKnots,nbKnots);
  powerof.resize(nbKnots,entityDim);

  switch (entityType) { 
  case MSH_TYPE_LIN: 
    {
      for (int i=0; i< nbKnots; i++){
        powerof(i,0) = i;
        for (int j=0; j< nbKnots; j++) {
          switch(entityOrder) {
          case 1: coef(i,j) = coef1_edge[i][j]; break;
          case 2: coef(i,j) = coef2_edge[i][j]; break;
          case 3: coef(i,j) = coef3_edge[i][j]; break;
          case 4: coef(i,j) = coef4_edge[i][j]; break;
          case 5: coef(i,j) = coef5_edge[i][j]; break;
          default : throw;
          }
        }
      }
      break;
    }
  case MSH_TYPE_TRI: 
    {
      if (entityIncomplete){
        for (int i=0; i< nbKnots; i++){
          for (int j=0; j< nbKnots; j++){
            switch(entityOrder) {
            case 1: coef(i,j) = coef1_tri[i][j]; break;
            case 2: coef(i,j) = coef2_tri[i][j]; break;
            case 3: coef(i,j) = coef3_tri[i][j]; break;
            case 4: coef(i,j) = coef4_tri[i][j]; break;
            case 5: coef(i,j) = coef5_tri[i][j]; break;
            default : throw;
            }
          }
          for (int j=0; j< 2; j++){
            switch(entityOrder) {
            case 1: powerof(i,j) = P1_tri[i][j]; break;
            case 2: powerof(i,j) = P2_tri[i][j]; break;
            case 3: powerof(i,j) = P3_tri[i][j]; break;
            case 4: powerof(i,j) = P4_tri[i][j]; break;
            case 5: powerof(i,j) = P5_tri[i][j]; break;
            default : throw;
            }
          }
        }
      }
      else {
        for (int i=0; i< nbKnots; i++){
          for (int j=0; j< nbKnots; j++){
            switch(entityOrder) {
            case 1: coef(i,j) = coef1_tri[i][j]; break;
            case 2: coef(i,j) = coef2_tri[i][j]; break;
            case 3: coef(i,j) = coef3_tri_lagrange[i][j]; break;
            case 4: coef(i,j) = coef4_tri_lagrange[i][j]; break;
            case 5: coef(i,j) = coef5_tri_lagrange[i][j]; break;
            default : throw;
            }
          }
          for (int j=0; j< 2; j++)
            powerof(i,j) = P_tri_lagrange[i][j];
        }
      }
      break;
    }
  case MSH_TYPE_QUA: 
    {
      if (entityIncomplete){
        for (int i=0; i< nbKnots; i++){
          for (int j=0; j< nbKnots; j++){
            switch(entityOrder) {
            case 1: coef(i,j) = coef1_quad[i][j]; break;
            case 2: coef(i,j) = coef2_quad[i][j]; break;
            case 3: coef(i,j) = coef3_quad[i][j]; break;
              //               case 4: coef(i,j) = coef4_quad[i][j]; break;
            default : throw;
            }
          }
          for (int j=0; j< 2; j++){
            switch(entityOrder) {
            case 1: powerof(i,j) = P1_quad[i][j]; break;
            case 2: powerof(i,j) = P2_quad[i][j]; break;
            case 3: powerof(i,j) = P3_quad[i][j]; break;
              //               case 4: powerof(i,j) = P4_quad[i][j]; break;
            default : throw;
            }
          }
        }
      }
      else {
        if (!interpolant) throw;
        for (int id2=0; id2< nbKnots; id2++){
          for (int id1=0; id1< nbKnots; id1++){
            coef(id1,id2)=interpolant->getCoef(id1,id2);
          }
          int powerU,powerV,powerW;
          interpolant->getMonomialExponents(id2,powerU,powerV,powerW);
          powerof(id2,0)=powerU;
          if (entityDim > 1) powerof(id2,1)=powerV;
          if (entityDim > 2) powerof(id2,2)=powerW;
        }
      }
      break;
    }
  case MSH_TYPE_TET: 
    {
      for (int i=0; i< nbKnots; i++){
        for (int j=0; j< nbKnots; j++)
          switch(entityOrder) {
          case 1: coef(i,j) = coef1_tet[i][j]; break;
          case 2: coef(i,j) = coef2_tet[i][j]; break;
          case 3: coef(i,j) = coef3_tet[i][j]; break;
          case 4: coef(i,j) = entityIncomplete ? coef4_tet[i][j] : coef4_tet_lagrange[i][j]; break;
          default : throw;
          }
        for (int j=0; j< 3; j++)
          switch(entityOrder) {
          case 1: powerof(i,j) = P1_tet[i][j]; break;
          case 2: powerof(i,j) = P2_tet[i][j]; break;
          case 3: powerof(i,j) = P3_tet[i][j]; break;
          case 4: powerof(i,j) = entityIncomplete ? P4_tet[i][j] : P4_tet_lagrange[i][j]; break;
          default : throw;
          }
      }
      break;
    }
  case MSH_TYPE_PYR: 
    {
      const MAd_PyramidInterpolant* pyr = dynamic_cast<const MAd_PyramidInterpolant*> (interpolant);
      for (int id2=0; id2< nbKnots; id2++){
        for (int id1=0; id1< nbKnots; id1++){
          coef(id1,id2)=pyr->getCoef(id1,id2);
        }
        int powerU,powerV,powerW;
        interpolant->getMonomialExponents(id2,powerU,powerV,powerW);
        powerof(id2,0)=powerU;
        powerof(id2,1)=powerV;
        powerof(id2,2)=powerW;
      }
      break;
    }
  default:
    {
      if (!interpolant) throw;
      for (int id2=0; id2< nbKnots; id2++){
        for (int id1=0; id1< nbKnots; id1++){
          coef(id1,id2)=interpolant->getCoef(id1,id2);
        }
        int powerU,powerV,powerW;
        interpolant->getMonomialExponents(id2,powerU,powerV,powerW);
        powerof(id2,0)=powerU;
        if (entityDim > 1) powerof(id2,1)=powerV;
        if (entityDim > 2) powerof(id2,2)=powerW;
      }
      break;
    }
  }
}

// -----------------------------------------------------------------------------

void MAdMapping::GeomShapeFunction ( double u, double v, double w, double* result) const
{
  
  switch(entityType) {

  case MSH_TYPE_PYR: {
    const int nbK=int(nbKnots);
    doubleVector f(result,nbK);
    interpolant->fcts(u,v,w,f);
    break;
  }

  default:{
    const int nbK=int(nbKnots);
    doubleVector res(result,nbK);
    
    doubleVector monomes(nbK);
    monomeof(u,v,w,monomes);
    coef.mult(monomes,res,false,1,0); 
  }
  }
}

// -----------------------------------------------------------------------------

void MAdMapping::GradGeomShapeFunction (double u, double v, double w , doubleMatrix &grads) const
{
  
  switch(entityType) {
  case MSH_TYPE_PYR: 
    {
      doubleVector gr[3] = {grads.touchColumn(0),grads.touchColumn(1),grads.touchColumn(2)};
      interpolant->grads(u,v,w,gr);
      break;
    }
    
  default:
    {
      doubleMatrix dMonomes(nbKnots,3);
      dmonomeof(u,v,w,dMonomes);
      grads.blas_dgemm(coef,dMonomes,1,0);
    } 
  }
}

// -----------------------------------------------------------------------------
bool MAdMapping::computeCut(const MAd_Vector3D& point,
                            const MAd_Vector3D& normal,int nb,
                            doubleMatrix& param,
                            doubleMatrix& coord) const {
  
  if (EN_type(ent) != 2) return false;

  MAd_Vector3D nn = normal;
  nn.normalise();

  pFace pf = (pFace) ent;
  int nbV = F_numVertices(pf);
  
  // --- check if all points on the same side 
  
  int nbAbove = 0;
  double position[4];
  
  for (int iV=0;iV<nbV;iV++) {
    MAd_Vector3D xyz;
    V_coord(F_vertex(pf,iV),xyz.data());
    position[iV] = (xyz - point) * nn;
    if (position[iV]>=0) nbAbove++;
  }
  
  if (nbAbove == 0 || nbAbove == nbV) return false;
  
  // --- compute edge intersections

  int iC = 0;
  MAd_Vector3D endPoint[4];  

  for (int iVm=0;iVm<nbV;iVm++) {
    
    int iVp = (iVm+1)%nbV;
    
    if (position[iVm] * position[iVp] < 0) {
      
      double t = fabs(position[iVp] / (position[iVm] - position[iVp]));
      
      MAd_Vector3D pm;
      getVertexUVW(iVm,pm[0],pm[1],pm[2]);
      MAd_Vector3D pp;
      getVertexUVW(iVp,pp[0],pp[1],pp[2]);
      MAd_Vector3D dp(pm,pp);
      
      double dd = dp.norm();

      int iter = 0;
      double distance = 0;
      double eps = 1.e-8;
      double over2eps = 1./(2.*eps);

      MAd_Vector3D ppc = pm + t*dp;
      MAd_Vector3D rc;
      eval(ppc[0],ppc[1],ppc[2],rc[0],rc[1],rc[2]);
      
      do {
        
        MAd_Vector3D ppp = pm + (t+eps) * dp;
        MAd_Vector3D ppm = pm + (t-eps) * dp;
        
        MAd_Vector3D rp;
        MAd_Vector3D rm;

        eval(ppp[0],ppp[1],ppp[2],rp[0],rp[1],rp[2]);
        eval(ppm[0],ppm[1],ppm[2],rm[0],rm[1],rm[2]);
        
        MAd_Vector3D drdt = (rp-rm) * over2eps;
        
        t -= (rc - point)*nn / (drdt * nn) * 0.2;
        t = std::min(1.,max(0.,t));

        ppc = pm + t*dp;
        eval(ppc[0],ppc[1],ppc[2],rc[0],rc[1],rc[2]);

        distance = fabs((rc - point) * nn);
        
      } while (iter++ < 100 && distance > 0.0001*dd);
      
      endPoint[iC++] = ppc;
      
    } 
  }
  
  // --- interpolate parametric coordinates 
  //     this is not the ultimate solution, ideally we should assure that the points are on the surface !!
  
  // param.resize(nb,3);
  // coord.resize(nb,3);

  for (int i=0;i<nb;i++) {
    
    double frac = (double) i / (double) (nb-1);
    
    param(i,0) = (1.-frac) * endPoint[0][0] + frac * endPoint[1][0];
    param(i,1) = (1.-frac) * endPoint[0][1] + frac * endPoint[1][1];
    param(i,2) = 0;
    
    double xyz[3];
    
    eval(param(i,0),param(i,1),param(i,2),xyz[0],xyz[1],xyz[2]);
    coord(i,0) = xyz[0];
    coord(i,1) = xyz[1];
    coord(i,2) = xyz[2];    
    
  }
  
  return true;
}


// -----------------------------------------------------------------------------

void MAdMapping::monomeof(const double &u, 
                          const double &v, 
                          const double& w, 
                          doubleVector& monomes) const{

  int eType = int(EN_type(ent));
 
  double uvw[3] = {u,v,w};
  doubleMatrix uvwPow(eType,entityOrder+1);
  uvwPow=1;
 
  for(int i=0;i<eType  ;i++){
    for(int k=1;k<=entityOrder;k++){
      uvwPow(i,k) = uvwPow(i,k-1)*uvw[i];
    }
  }

  switch (eType){
  case 3: 
    {
      for(int i=0;i<int(nbKnots);i++) {
        monomes(i) = (uvwPow(0,(int)powerof(i,0))*
                      uvwPow(1,(int)powerof(i,1))*
                      uvwPow(2,(int)powerof(i,2))); 
      }
      break;
    }
  case 2: 
    {
      for(int i=0;i<int(nbKnots);i++) {
        monomes(i) = (uvwPow(0,(int)powerof(i,0))*
                      uvwPow(1,(int)powerof(i,1))); 
      }
      break;
    }
  case 1:
    {
      for(int i=0;i<int(nbKnots);i++) {
        monomes(i) = uvwPow(0,(int)powerof(i,0)); 
      }
      break;
    }
  default:
    MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                "Mapping for dimension %i not implemented",
                                EN_type(ent));
    throw;
  }
}

// -----------------------------------------------------------------------------

void MAdMapping::dmonomeof(const double &u, 
                           const double &v, 
                           const double &w, 
                           doubleMatrix& dMonomes) const{

  int eType = int(EN_dim(ent));   
  double uvw[3] = {u,v,w};
  doubleMatrix uvwPow(eType,entityOrder+1);
  uvwPow=1;
 
  for(int i=0;i<eType  ;i++){
    for(int k=1;k<=entityOrder;k++){
      uvwPow(i,k) = uvwPow(i,k-1)*uvw[i];
    }
  }
  dMonomes=0;
  
  switch (eType){
  case 3:
    { 
      for(int i=0; i<int(nbKnots); i++){
        for(int k=0; k<3; k++){  
          if (powerof(i,k) > 0){
            double dd = powerof(i,k);
            
            switch (k) {
            case 0: 
              {
                dMonomes(i,k) = dd * (uvwPow(0,(int)powerof(i,0)-1) * 
                                      uvwPow(1,(int)powerof(i,1)) * 
                                      uvwPow(2,(int)powerof(i,2))); 
                                      
                break;
              }
            case 1: 
              {
                dMonomes(i,k) = dd * (uvwPow(0,(int)powerof(i,0)) * 
                                      uvwPow(1,(int)powerof(i,1)-1) * 
                                      uvwPow(2,(int)powerof(i,2))); 
                break;
              }
            case 2: 
              {
                dMonomes(i,k) = dd * (uvwPow(0,(int)powerof(i,0)) * 
                                      uvwPow(1,(int)powerof(i,1)) * 
                                      uvwPow(2,(int)powerof(i,2)-1)); 
                break;	  
              }
            default: throw;
            }
          }
        }
      }
      break;
    }
  case 2:
    {
      for(int i=0; i<int(nbKnots); i++){
        for(int k=0; k<2; k++){ 
          if (powerof(i,k) > 0.){
            double dd = powerof(i,k);
            switch (k){
            case 0: 
              {
                dMonomes(i,k) = dd * (uvwPow(0,(int)powerof(i,0)-1)*
                                      uvwPow(1,(int)powerof(i,1))); 
                break;	   
              }
            case 1: 
              {
                dMonomes(i,k) = dd * (uvwPow(0,(int)powerof(i,0))*
                                      uvwPow(1,(int)powerof(i,1)-1));
                break;
              }
            default: throw;
            }
          }
        }
      }
      break;
    }
  case 1:
    {
      for(int i=0; i<int(nbKnots); i++){
        for(int k=0; k<1; k++){ 
          if (i>0) dMonomes(i,k) = i*uvwPow(0,i-1); 
        }}
      break;
    }
  }
}


//! \todo simplex mappings are explicitly coded, should be computed based on control point location.

// --- First order mapping

const double MAdMapping::pos1_tet[4][3] = {{0,0,0},
                                           {1,0,0},
                                           {0,1,0},
                                           {0,0,1}};

const int MAdMapping::P1_tet[4][3] = {{0,0,0}, // 1
                                      {1,0,0},
                                      {0,1,0},
                                      {0,0,1}}; // u v w

const double MAdMapping::coef1_tet[4][4] = {{1.,-1.,-1.,-1.}, // 1-u-v-w
                                            {0.,1.,0.,0.},    // u
                                            {0.,0.,1.,0.},    // v
                                            {0.,0.,0.,1.}};   // w

// --- Second order mapping

const double MAdMapping::pos2_tet[10][3] = {{0.,0.,0.},
                                            {1.,0.,0.},
                                            {0.,1.,0.},
                                            {0.,0.,1.},
                                            {0.5,0.,0.},
                                            {0.5,0.5,0.},
                                            {0.,0.5,0.},
                                            {0.,0.,0.5},
                                            {0.5,0.,0.5},
                                            {0.,0.5,0.5}};

const int MAdMapping::P2_tet[10][3] = {
  {0,0,0},
  //order 1:
  {1,0,0},{0,1,0},{0,0,1},
  // order 2:
  {2,0,0},{0,2,0},{0,0,2},{1,1,0},{0,1,1},{1,0,1}};

const double MAdMapping::coef2_tet[10][10] = {{1,-3,-3,-3,2,2,2,4,4,4},
                                              {0,-1,0,0,2,0,0,0,0,0},
                                              {0,0,-1,0,0,2,0,0,0,0},
                                              {0,0,0,-1,0,0,2,0,0,0},
                                              {0,4,0,0,-4,0,0,-4,0,-4},
                                              {0,0,0,0,0,0,0,4,0,0},
                                              {0,0,4,0,0,-4,0,-4,-4,0},
                                              {0,0,0,4,0,0,-4,0,-4,-4},
                                              {0,0,0,0,0,0,0,0,0,4},
                                              {0,0,0,0,0,0,0,0,4,0}};

// --- third order mapping


// regenerated automatically

const int MAdMapping::P3_tet[20][3] =   {{0,0,0},
                                         {1,0,0},{0,1,0},{0,0,1},
                                         {2,0,0},{1,1,0},{0,2,0},{1,0,1},{0,1,1},{0,0,2},
                                         {3,0,0},{2,1,0},{1,2,0},{0,3,0},{2,0,1},{1,1,1},{0,2,1},{1,0,2},{0,1,2},{0,0,3}};


const double MAdMapping::coef3_tet[20][20] = {
  {1.,-5.5,-5.5,-5.5,9.0,18.0,9.0,18.0,18.0,9.0,-4.5,-13.5,-13.5,-4.5,-13.5,-27.0,-13.5,-13.5,-13.5,-4.5},
  {0.0,1.,0.,0.,-4.5,0.0,0.0,0.0,0.0,0.0,4.5,0.0,0.0,0.,0.0,0.0,0.0,0.0,0.0,0.0},
  {0.,0.0,1.,0.0,0.,0.0,-4.5,0.,0.,0.0,0.0,0.0,0.0,4.5,0.,0.,0.0,0.,0.,0.},
  {0.,0.,0.,1.0,0.0,0.,0.0,0.,0.,-4.5,0.,0.0,0.0,0.,0.0,0.0,0.0,0.0,0.0,4.5},
  {0.,9.0,0.,0.0,-22.5,-22.5,0.0,-22.5,0.0,0.,13.5,27.0,13.5,0.,27.0,27.0,0.,13.5,0.0,0.0},
  {0.0,-4.5,0.0,0.,18.0,4.5,0.,4.5,0.,0.0,-13.5,-13.5,0.,0.0,-13.5,0.0,0.0,0.0,0.0,0.0},
  {0.,0.,0.,0.0,0.0,-4.5,0.0,0.,0.,0.,0.0,13.5,0.0,0.,0.0,0.0,0.0,0.0,0.0,0.0},
  {0.,0.,0.,0.0,0.0,-4.5,0.0,0.0,0.0,0.0,0.0,0.0,13.5,0.0,0.,0.,0.0,0.,0.,0.},
  {0.0,0.0,-4.5,0.0,0.,4.5,18.0,0.0,4.5,0.,0.,0.,-13.5,-13.5,0.,0.,-13.5,0.,0.,0.},
  {0.,0.,9.0,0.0,0.0,-22.5,-22.5,0.,-22.5,0.,0.0,13.5,27.0,13.5,0.0,27.0,27.0,0.0,13.5,0.0},
  {0.,0.,0.,9.0,0.0,0.,0.0,-22.5,-22.5,-22.5,0.0,0.0,0.0,0.,13.5,27.0,13.5,27.0,27.0,13.5},
  {0.0,0.0,0.0,-4.5,0.,0.0,0.,4.5,4.5,18.0,0.,0.,0.,0.0,0.,0.,0.,-13.5,-13.5,-13.5},
  {0.,0.,0.,0.,0.0,0.,0.0,-4.5,0.0,0.0,0.0,0.0,0.0,0.,13.5,0.0,0.0,0.0,0.0,0.0},
  {0.,0.,0.,0.,0.0,0.,0.0,-4.5,0.,0.0,0.0,0.0,0.0,0.,0.0,0.0,0.0,13.5,0.0,0.0},
  {0.,0.,0.,0.0,0.0,0.,0.0,0.,-4.5,0.,0.0,0.0,0.0,0.,0.0,0.0,13.5,0.0,0.0,0.0},
  {0.,0.,0.,0.,0.0,0.,0.0,0.,-4.5,0.0,0.0,0.0,0.0,0.,0.0,0.0,0.0,0.0,13.5,0.0},
  {0.0,0.0,0.0,0.0,0.,27.0,0.,0.,0.,0.,0.,-27.0,-27.0,0.0,0.,-27.0,0.,0.,0.,0.},
  {0.0,0.0,0.0,0.,0.,0.0,0.,27.0,0.0,0.0,0.,0.,0.,0.0,-27.0,-27.0,0.0,-27.0,0.,0.},
  {0.,0.,0.,0.,0.0,0.,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.,0.0,27.0,0.0,0.0,0.0,0.0},
  {0.0,0.0,0.0,0.,0.,0.0,0.,0.0,27.0,0.0,0.,0.,0.,0.0,0.,-27.0,-27.0,0.,-27.0,0.}
};

// ---fourth order mapping

const int MAdMapping::P4_tet[34][3] = {{0,0,0},
                                       {1,0,0},{0,1,0},{0,0,1},
                                       {2,0,0},{1,1,0},{0,2,0},{0,1,1},{0,0,2},{1,0,1},
                                       {3,0,0},{2,1,0},{1,2,0},{0,3,0},{0,2,1},{0,1,2},{0,0,3},{1,0,2},{2,0,1},
                                       {4,0,0},{3,1,0},{2,2,0},{1,3,0},{0,4,0},{3,0,1},{2,1,1},{1,2,1},{0,3,1},
                                       {2,0,2},{1,1,2},{0,2,2},{1,0,3},{0,1,3},{0,0,4}};

const double MAdMapping::coef4_tet[34][34] = {
  {1,-8.33333333,-8.33333333,-8.33333333,23.33333333,46.66666667,23.33333333,46.66666667,23.33333333,46.66666667,-26.66666667,-80,-80,-26.66666667,-80,-80,-26.66666667,-80,-80,10.66666667,42.66666667,64,42.66666667,10.66666667,42.66666667,-32,-32,42.66666667,64,-32,64,42.66666667,42.66666667,10.66666667},
  {0,-1,0,0,7.33333333,0,0,0,0,0,-16,0,0,0,0,0,0,0,0,10.66666667,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
  {0,0,-1,0,0,3.33067e-15,7.33333333,0,0,0,0,0,0,-16,0,0,0,0,0,0,0,0,0,10.66666667,0,0,0,0,0,0,0,0,0,0},
  {0,0,0,-1,0,0,0,0,7.33333333,0,0,0,0,0,0,0,-16,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,10.66666667},
  {0,16,0,0,-69.33333333,-69.33333333,0,0,0,-69.33333333,96,192,96,0,0,0,0,96,192,-42.66666667,-128,-128,-42.66666667,0,-128,-64,64,0,-128,64,0,-42.66666667,0,0},
  {0,-12,0,0,76,28,0,0,0,28,-128,-144,-16,0,0,0,0,-16,-144,64,128,64,0,0,128,96,-32,0,64,-32,0,0,0,0},
  {0,5.33333333,0,0,-37.33333333,-5.33333333,0,0,0,-5.33333333,74.66666667,32,0,0,0,0,0,0,32,-42.66666667,-42.66666667,0,0,0,-42.66666667,0,0,0,0,0,0,0,0,0},
  {0,0,0,0,0,5.33333333,0,0,0,0,0,-32,0,0,0,0,0,0,0,0,42.66666667,0,0,0,0,0,0,0,0,0,0,0,0,0},
  {0,0,0,0,0,4,0,0,0,0,0,-16,-16,0,0,0,0,0,0,0,0,64,0,0,0,0,0,0,0,0,0,0,0,0},
  {0,0,0,0,0,5.33333333,0,0,0,0,0,0,-32,0,0,0,0,0,0,0,0,0,42.66666667,0,0,0,0,0,0,0,0,0,0,0},
  {0,0,5.33333333,0,0,-5.33333333,-37.33333333,-5.33333333,0,0,0,-8.88178e-16,32,74.66666667,32,0,0,0,0,0,0,0,-42.66666667,-42.66666667,0,0,0,-42.66666667,0,0,0,0,0,0},
  {0,0,-12,0,0,28,76,28,0,0,0,-16,-144,-128,-144,-16,0,0,0,0,0,64,128,64,0,-32,96,128,0,-32,64,0,0,0},
  {0,0,16,0,0,-69.33333333,-69.33333333,-69.33333333,0,0,0,96,192,96,192,96,0,0,0,0,-42.66666667,-128,-128,-42.66666667,0,64,-64,-128,0,64,-128,0,-42.66666667,0},
  {0,0,0,16,0,0,0,-69.33333333,-69.33333333,-69.33333333,0,0,0,0,96,192,96,192,96,0,0,0,0,0,-42.66666667,64,64,-42.66666667,-128,-64,-128,-128,-128,-42.66666667},
  {0,0,0,-12,0,0,0,28,76,28,0,0,0,0,-16,-144,-128,-144,-16,0,0,0,0,0,2.4869e-14,-32,-32,0,64,96,64,128,128,64},
  {0,0,0,5.33333333,0,0,0,-5.33333333,-37.33333333,-5.33333333,0,0,0,0,0,32,74.66666667,32,0,0,0,0,0,0,0,0,0,0,0,0,0,-42.66666667,-42.66666667,-42.66666667},
  {0,0,0,0,0,0,0,0,0,5.33333333,0,0,0,0,0,0,0,0,-32,0,0,0,0,0,42.66666667,0,0,0,0,0,0,0,0,0},
  {0,0,0,0,0,0,0,0,0,4,0,0,0,0,0,0,0,-16,-16,0,0,0,0,0,0,0,0,0,64,0,0,0,0,0},
  {0,0,0,0,0,0,0,0,0,5.33333333,0,0,0,0,0,0,0,-32,0,0,0,0,0,0,0,0,0,0,0,0,0,42.66666667,0,0},
  {0,0,0,0,0,0,0,5.33333333,0,0,0,0,0,0,-32,0,0,0,0,0,0,0,0,0,0,0,0,42.66666667,0,0,0,0,0,0},
  {0,0,0,0,0,0,0,4,0,0,0,0,0,0,-16,-16,0,0,0,0,0,0,0,0,0,0,0,0,0,0,64,0,0,0},
  {0,0,0,0,0,0,0,5.33333333,0,0,0,0,0,0,0,-32,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,42.66666667,0},
  {0,0,0,0,0,96,0,0,0,0,0,-224,-224,0,0,0,0,0,0,0,128,256,128,0,0,32,32,0,0,-96,0,0,0,0},
  {0,0,0,0,0,-32,0,0,0,0,0,160,32,0,0,0,0,0,0,0,-128,-128,0,0,0,-96,32,0,0,32,0,0,0,0},
  {0,0,0,0,0,-32,0,0,0,0,0,32,160,0,0,0,0,0,0,0,0,-128,-128,0,0,32,-96,0,0,32,0,0,0,0},
  {0,0,0,0,0,0,0,0,0,96,0,0,0,0,0,0,0,-224,-224,0,0,0,0,0,128,32,-96,0,256,32,0,128,0,0},
  {0,0,0,0,0,0,0,0,0,-32,0,0,0,0,0,0,0,32,160,0,0,0,0,0,-128,-96,32,0,-128,32,0,0,0,0},
  {0,0,0,0,0,0,0,0,0,-32,0,0,0,0,0,0,0,160,32,0,0,0,0,0,0,32,32,0,-128,-96,0,-128,0,0},
  {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,96,-32,0,0,-32,0,0,0,0},
  {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,-32,96,0,0,-32,0,0,0,0},
  {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,-32,-32,0,0,96,0,0,0,0},
  {0,0,0,0,0,0,0,96,0,0,0,0,0,0,-224,-224,0,0,0,0,0,0,0,0,0,-96,32,128,0,32,256,0,128,0},
  {0,0,0,0,0,0,0,-32,0,0,0,0,0,0,160,32,0,0,0,0,0,0,0,0,0,32,-96,-128,0,32,-128,0,0,0},
  {0,0,0,0,0,0,0,-32,0,0,0,0,0,0,32,160,0,0,0,0,0,0,0,0,0,32,32,0,0,-96,-128,0,-128,0}
};


const double MAdMapping::pos4_tet_lagrange[35][3] = {
  {0,0,0},{1,0,0},{0,1,0},{0,0,1},                 // principal vertices
  {0.25,0,0},{0.5,0,0},{0.75,0,0},                 // edge 0-1
  {0.75,0.25,0},{0.5,0.5,0},{0.25,0.75,0},         // edge 1-2
  {0,0.75,0},{0,0.5,0},{0,0.25,0},                 // edge 2-0
  {0,0,0.25},{0,0,0.5},{0,0,0.75},                 // edge 0-3
  {0.75,0,0.25},{0.5,0,0.5},{0.25,0,0.75},         // edge 1-3
  {0,0.75,0.25},{0,0.5,0.5},{0,0.25,0.75},         // edge 2-3
  {0.25,0.25,0},{0.5,0.25,0},{0.25,0.5,0},         // face 0-1-2
  {0.25,0,0.25},{0.5,0,0.25},{0.25,0,0.5},         // face 0-1-3
  {0.5,0.25,0.25},{0.25,0.5,0.25},{0.25,0.25,0.5}, // face 1-2-3
  {0,0.25,0.25},{0,0.5,0.25},{0,0.25,0.5},         // face 0-2-3
  {0.25,0.25,0.25}                                 // center
};

const int MAdMapping::P4_tet_lagrange[35][3] = {
  {0,0,0},
  {1,0,0},{0,1,0},{0,0,1},
  {2,0,0},{1,1,0},{0,2,0},{1,0,1},{0,1,1},{0,0,2},
  {3,0,0},{2,1,0},{1,2,0},{0,3,0},{2,0,1},{1,1,1},{0,2,1},{1,0,2},{0,1,2},{0,0,3},
  {4,0,0},{3,1,0},{2,2,0},{1,3,0},{0,4,0},{3,0,1},{2,1,1},{1,2,1},{0,3,1},{2,0,2},{1,1,2},{0,2,2},{1,0,3},{0,1,3},{0,0,4}
};

const double MAdMapping::coef4_tet_lagrange[35][35] = {
  {1,-8.33333333,-8.33333333,-8.33333333,23.33333333,46.66666667,23.33333333,46.66666667,46.66666667,23.33333333,-26.66666667,-80,-80,-26.66666667,-80,-160,-80,-80,-80,-26.66666667,10.66666667,42.66666667,64,42.66666667,10.66666667,42.66666667,128,128,42.66666667,64,128,64,42.66666667,42.66666667,10.66666667},
  {0,-1,0,0,7.33333333,0,0,0,0,0,-16,0,0,0,0,0,0,0,0,0,10.66666667,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
  {0,0,-1,0,0,0,7.33333333,0,0,0,0,0,0,-16,0,0,0,0,0,0,0,0,0,0,10.66666667,0,0,0,0,0,0,0,0,0,0},
  {0,0,0,-1,0,0,0,0,0,7.33333333,0,0,0,0,0,0,0,0,0,-16,0,0,0,0,0,0,0,0,0,0,0,0,0,0,10.66666667},
  {0,16,0,0,-69.33333333,-69.33333333,0,-69.33333333,0,0,96,192,96,0,192,192,0,96,0,0,-42.66666667,-128,-128,-42.66666667,0,-128,-256,-128,0,-128,-128,0,-42.66666667,0,0},
  {0,-12,0,0,76,28,0,28,0,0,-128,-144,-16,0,-144,-32,0,-16,0,0,64,128,64,0,0,128,128,0,0,64,0,0,0,0,0},
  {0,5.33333333,0,0,-37.33333333,-5.33333333,0,-5.33333333,0,0,74.66666667,32,0,0,32,0,0,0,0,0,-42.66666667,-42.66666667,0,0,0,-42.66666667,0,0,0,0,0,0,0,0,0},
  {0,0,0,0,0,5.33333333,0,0,0,0,0,-32,0,0,0,0,0,0,0,0,0,42.66666667,0,0,0,0,0,0,0,0,0,0,0,0,0},
  {0,0,0,0,0,4,0,0,0,0,0,-16,-16,0,0,0,0,0,0,0,0,0,64,0,0,0,0,0,0,0,0,0,0,0,0},
  {0,0,0,0,0,5.33333333,0,0,0,0,0,0,-32,0,0,0,0,0,0,0,0,0,0,42.66666667,0,0,0,0,0,0,0,0,0,0,0},
  {0,0,5.33333333,0,0,-5.33333333,-37.33333333,0,-5.33333333,0,0,0,32,74.66666667,0,0,32,0,0,0,0,0,0,-42.66666667,-42.66666667,0,0,0,-42.66666667,0,0,0,0,0,0},
  {0,0,-12,0,0,28,76,0,28,0,0,-16,-144,-128,0,-32,-144,0,-16,0,0,0,64,128,64,0,0,128,128,0,0,64,0,0,0},
  {0,0,16,0,0,-69.33333333,-69.33333333,0,-69.33333333,0,0,96,192,96,0,192,192,0,96,0,0,-42.66666667,-128,-128,-42.66666667,0,-128,-256,-128,0,-128,-128,0,-42.66666667,0},
  {0,0,0,16,0,0,0,-69.33333333,-69.33333333,-69.33333333,0,0,0,0,96,192,96,192,192,96,0,0,0,0,0,-42.66666667,-128,-128,-42.66666667,-128,-256,-128,-128,-128,-42.66666667},
  {0,0,0,-12,0,0,0,28,28,76,0,0,0,0,-16,-32,-16,-144,-144,-128,0,0,0,0,0,0,0,0,0,64,128,64,128,128,64.},
  {0,0,0,5.33333333,0,0,0,-5.33333333,-5.33333333,-37.33333333,0,0,0,0,0,0,0,32,32,74.66666667,0,0,0,0,0,0,0,0,0,0,0,0,-42.66666667,-42.66666667,-42.66666667},
  {0,0,0,0,0,0,0,5.33333333,0,0,0,0,0,0,-32,0,0,0,0,0,0,0,0,0,0,42.66666667,0,0,0,0,0,0,0,0,0},
  {0,0,0,0,0,0,0,4,0,0,0,0,0,0,-16,0,0,-16,0,0,0,0,0,0,0,0,0,0,0,64,0,0,0,0,0},
  {0,0,0,0,0,0,0,5.33333333,0,0,0,0,0,0,0,0,0,-32,0,0,0,0,0,0,0,0,0,0,0,0,0,0,42.66666667,0,0},
  {0,0,0,0,0,0,0,0,5.33333333,0,0,0,0,0,0,0,-32,0,0,0,0,0,0,0,0,0,0,0,42.66666667,0,0,0,0,0,0},
  {0,0,0,0,0,0,0,0,4,0,0,0,0,0,0,0,-16,0,-16,0,0,0,0,0,0,0,0,0,0,0,0,64,0,0,0},
  {0,0,0,0,0,0,0,0,5.33333333,0,0,0,0,0,0,0,0,0,-32,0,0,0,0,0,0,0,0,0,0,0,0,0,0,42.66666667,0},
  {0,0,0,0,0,96,0,0,0,0,0,-224,-224,0,0,-224,0,0,0,0,0,128,256,128,0,0,256,256,0,0,128,0,0,0,0},
  {0,0,0,0,0,-32,0,0,0,0,0,160,32,0,0,32,0,0,0,0,0,-128,-128,0,0,0,-128,0,0,0,0,0,0,0,0},
  {0,0,0,0,0,-32,0,0,0,0,0,32,160,0,0,32,0,0,0,0,0,0,-128,-128,0,0,0,-128,0,0,0,0,0,0,0},
  {0,0,0,0,0,0,0,96,0,0,0,0,0,0,-224,-224,0,-224,0,0,0,0,0,0,0,128,256,128,0,256,256,0,128,0,0},
  {0,0,0,0,0,0,0,-32,0,0,0,0,0,0,160,32,0,32,0,0,0,0,0,0,0,-128,-128,0,0,-128,0,0,0,0,0},
  {0,0,0,0,0,0,0,-32,0,0,0,0,0,0,32,32,0,160,0,0,0,0,0,0,0,0,0,0,0,-128,-128,0,-128,0,0},
  {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,-32,0,0,0,0,0,0,0,0,0,0,128,0,0,0,0,0,0,0,0},
  {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,-32,0,0,0,0,0,0,0,0,0,0,0,128,0,0,0,0,0,0,0},
  {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,-32,0,0,0,0,0,0,0,0,0,0,0,0,0,0,128,0,0,0,0},
  {0,0,0,0,0,0,0,0,96,0,0,0,0,0,0,-224,-224,0,-224,0,0,0,0,0,0,0,128,256,128,0,256,256,0,128,0},
  {0,0,0,0,0,0,0,0,-32,0,0,0,0,0,0,32,160,0,32,0,0,0,0,0,0,0,0,-128,-128,0,0,-128,0,0,0},
  {0,0,0,0,0,0,0,0,-32,0,0,0,0,0,0,32,32,0,160,0,0,0,0,0,0,0,0,0,0,0,-128,-128,0,-128,0},
  {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,256,0,0,0,0,0,0,0,0,0,0,-256,-256,0,0,-256,0,0,0,0}
};



// -----------------------------------------------------------------------------
//   1             u             v
const double MAdMapping::coef1_tri[3][3]={
  {  1.00000000,  -1.00000000,  -1.00000000},
  {  0.00000000,   1.00000000,   0.00000000},
  {  0.00000000,   0.00000000,   1.00000000}
};

// -----------------------------------------------------------------------------
//   1             u             v             u^2           v^2           uv
const double MAdMapping::coef2_tri[6][6]={
  {  1.00000000,  -3.00000000,  -3.00000000,   2.00000000,   2.00000000,   4.00000000},
  {  0.00000000,  -1.00000000,   0.00000000,   2.00000000,  -0.00000000,  -0.00000000},
  {  0.00000000,   0.00000000,  -1.00000000,  -0.00000000,   2.00000000,  -0.00000000},
  {  0.00000000,   4.00000000,   0.00000000,  -4.00000000,  -0.00000000,  -4.00000000},
  {  0.00000000,   0.00000000,   0.00000000,  -0.00000000,  -0.00000000,   4.00000000},
  {  0.00000000,   0.00000000,   4.00000000,  -0.00000000,  -4.00000000,  -4.00000000}
};

// -----------------------------------------------------------------------------
//   1             u             v             u^2           v^2           u^3           v^3           u^2 v         u v^2
const double MAdMapping::coef3_tri[9][9]={
  {  1.00000000,  -5.50000000,  -5.50000000,   9.00000000,   9.00000000,  -4.50000000,  -4.50000000,   4.50000000,   4.50000000},
  {  0.00000000,   1.00000000,   0.00000000,  -4.50000000,  -0.00000000,   4.50000000,  -0.00000000,  -0.00000000,  -0.00000000},
  {  0.00000000,   0.00000000,   1.00000000,  -0.00000000,  -4.50000000,  -0.00000000,   4.50000000,  -0.00000000,  -0.00000000},
  {  0.00000000,   9.00000000,   0.00000000, -22.50000000,  -0.00000000,  13.50000000,   0.00000000,   4.50000000,  -9.00000000},
  {  0.00000000,  -4.50000000,  -0.00000000,  18.00000000,   0.00000000, -13.50000000,  -0.00000000,  -9.00000000,   4.50000000},
  {  0.00000000,   0.00000000,   0.00000000,  -0.00000000,  -0.00000000,  -0.00000000,   0.00000000,   9.00000000,  -4.50000000},
  {  0.00000000,   0.00000000,  -0.00000000,  -0.00000000,   0.00000000,  -0.00000000,  -0.00000000,  -4.50000000,   9.00000000},
  {  0.00000000,   0.00000000,  -4.50000000,  -0.00000000,  18.00000000,  -0.00000000, -13.50000000,   4.50000000,  -9.00000000},
  {  0.00000000,   0.00000000,   9.00000000,  -0.00000000, -22.50000000,  -0.00000000,  13.50000000,  -9.00000000,   4.50000000}
};

// -----------------------------------------------------------------------------
//   1             u             v            u^2           v^2          u^3            v^3           u^4           v^4            u^3 v         u v^3         u^2 v^2
const double MAdMapping::coef4_tri[12][12]={
  {  1.00000000,  -8.33333333,  -8.33333333,  23.33333333,  23.33333333, -26.66666667, -26.66666667,  10.66666667,  10.66666667,   9.33333333,   9.33333333,  -2.66666667},
  {  0.00000000,  -1.00000000,   0.00000000,   7.33333333,  -0.00000000, -16.00000000,   0.00000000,  10.66666667,  -0.00000000,  -0.00000000,  -0.00000000,  -0.00000000},
  {  0.00000000,   0.00000000,  -1.00000000,  -0.00000000,   7.33333333,  -0.00000000, -16.00000000,  -0.00000000,  10.66666667,  -0.00000000,  -0.00000000,  -0.00000000},
  {  0.00000000,  16.00000000,  -0.00000000, -69.33333333,   0.00000000,  96.00000000,  -0.00000000, -42.66666667,   0.00000000,  -5.33333333, -16.00000000,  21.33333333},
  {  0.00000000, -12.00000000,   0.00000000,  76.00000000,  -0.00000000, -128.00000000,   0.00000000,  64.00000000,  -0.00000000,  12.00000000,  12.00000000, -40.00000000},
  {  0.00000000,   5.33333333,  -0.00000000, -37.33333333,   0.00000000,  74.66666667,  -0.00000000, -42.66666667,   0.00000000, -16.00000000,  -5.33333333,  21.33333333},
  {  0.00000000,   0.00000000,   0.00000000,  -0.00000000,  -0.00000000,  -0.00000000,   0.00000000,  -0.00000000,  -0.00000000,  16.00000000,   5.33333333, -21.33333333},
  {  0.00000000,   0.00000000,  -0.00000000,  -0.00000000,   0.00000000,  -0.00000000,  -0.00000000,  -0.00000000,   0.00000000, -12.00000000, -12.00000000,  40.00000000},
  {  0.00000000,   0.00000000,   0.00000000,  -0.00000000,  -0.00000000,  -0.00000000,   0.00000000,  -0.00000000,  -0.00000000,   5.33333333,  16.00000000, -21.33333333},
  {  0.00000000,   0.00000000,   5.33333333,  -0.00000000, -37.33333333,  -0.00000000,  74.66666667,  -0.00000000, -42.66666667,  -5.33333333, -16.00000000,  21.33333333},
  {  0.00000000,   0.00000000, -12.00000000,  -0.00000000,  76.00000000,  -0.00000000, -128.00000000,  -0.00000000,  64.00000000,  12.00000000,  12.00000000, -40.00000000},
  {  0.00000000,   0.00000000,  16.00000000,  -0.00000000, -69.33333333,  -0.00000000,  96.00000000,  -0.00000000, -42.66666667, -16.00000000,  -5.33333333,  21.33333333}
};

// -----------------------------------------------------------------------------
//   1             u             v            u^2           v^2          u^3            v^3           u^4           v^4            u^5          v^5           u^4 v^1       u^3 v^2        u^2 v^3      u^1 v^4

const double MAdMapping::coef5_tri[15][15]={
  {  1.00000000, -11.41666667, -11.41666667,  46.87500000,  46.87500000, -88.54166667, -88.54166667,  78.12500000,  78.12500000, -26.04166667, -26.04166667,  10.41666667,   5.20833333,   5.20833333,  10.41666667},
  {  0.00000000,   1.00000000,  -0.00000000, -10.41666667,   0.00000000,  36.45833333,  -0.00000000, -52.08333333,   0.00000000,  26.04166667,   0.00000000,  -0.00000000,   0.00000000,  -0.00000000,   0.00000000},
  {  0.00000000,   0.00000000,   1.00000000,  -0.00000000, -10.41666667,  -0.00000000,  36.45833333,  -0.00000000, -52.08333333,   0.00000000,  26.04166667,  -0.00000000,   0.00000000,  -0.00000000,   0.00000000},
  {  0.00000000,  25.00000000,  -0.00000000, -160.41666667,   0.00000000, 369.79166667,  -0.00000000, -364.58333333,   0.00000000, 130.20833333,  -0.00000000,   6.25000000, -38.54166667,  60.41666667, -25.00000000},
  {  0.00000000, -25.00000000,   0.00000000, 222.91666667,  -0.00000000, -614.58333333,   0.00000000, 677.08333333,  -0.00000000, -260.41666667,   0.00000000, -16.66666667,  95.83333333, -122.91666667,  25.00000000},
  {  0.00000000,  16.66666667,  -0.00000000, -162.50000000,   0.00000000, 510.41666667,  -0.00000000, -625.00000000,   0.00000000, 260.41666667,  -0.00000000,  25.00000000, -122.91666667,  95.83333333, -16.66666667},
  {  0.00000000,  -6.25000000,   0.00000000,  63.54166667,  -0.00000000, -213.54166667,   0.00000000, 286.45833333,  -0.00000000, -130.20833333,   0.00000000, -25.00000000,  60.41666667, -38.54166667,   6.25000000},
  {  0.00000000,   0.00000000,  -0.00000000,  -0.00000000,   0.00000000,  -0.00000000,  -0.00000000,  -0.00000000,   0.00000000,   0.00000000,  -0.00000000,  25.00000000, -60.41666667,  38.54166667,  -6.25000000},
  {  0.00000000,   0.00000000,   0.00000000,  -0.00000000,  -0.00000000,  -0.00000000,   0.00000000,  -0.00000000,  -0.00000000,   0.00000000,   0.00000000, -25.00000000, 122.91666667, -95.83333333,  16.66666667},
  {  0.00000000,   0.00000000,  -0.00000000,  -0.00000000,   0.00000000,  -0.00000000,  -0.00000000,  -0.00000000,   0.00000000,   0.00000000,  -0.00000000,  16.66666667, -95.83333333, 122.91666667, -25.00000000},
  {  0.00000000,   0.00000000,   0.00000000,  -0.00000000,  -0.00000000,  -0.00000000,   0.00000000,  -0.00000000,  -0.00000000,   0.00000000,   0.00000000,  -6.25000000,  38.54166667, -60.41666667,  25.00000000},
  {  0.00000000,   0.00000000,  -6.25000000,  -0.00000000,  63.54166667,  -0.00000000, -213.54166667,  -0.00000000, 286.45833333,   0.00000000, -130.20833333,   6.25000000, -38.54166667,  60.41666667, -25.00000000},
  {  0.00000000,   0.00000000,  16.66666667,  -0.00000000, -162.50000000,  -0.00000000, 510.41666667,  -0.00000000, -625.00000000,   0.00000000, 260.41666667, -16.66666667,  95.83333333, -122.91666667,  25.00000000},
  {  0.00000000,   0.00000000, -25.00000000,  -0.00000000, 222.91666667,  -0.00000000, -614.58333333,  -0.00000000, 677.08333333,   0.00000000, -260.41666667,  25.00000000, -122.91666667,  95.83333333, -16.66666667},
  {  0.00000000,   0.00000000,  25.00000000,  -0.00000000, -160.41666667,  -0.00000000, 369.79166667,  -0.00000000, -364.58333333,   0.00000000, 130.20833333, -25.00000000,  60.41666667, -38.54166667,   6.25000000}
};


// -----------------------------------------------------------------------------

const int MAdMapping::P0_tri[1][2] = {
  {0,0} // 1
};

// -----------------------------------------------------------------------------

const int MAdMapping::P1_tri[3][2] = {
  {0,0}, // 1
  {1,0}, // u
  {0,1}  // v
};

// -----------------------------------------------------------------------------

const int MAdMapping::P2_tri[6][2] = {
  {0,0}, // 1
  {1,0}, // u
  {0,1}, // v
  {2,0}, // u^2
  {0,2}, // v^2
  {1,1}  // u * v
};

// -----------------------------------------------------------------------------

const int MAdMapping::P3_tri[9][2] = {
  {0,0}, // 1
  {1,0}, // u
  {0,1}, // v
  {2,0}, // u^2
  {0,2}, // v^2
  {3,0}, // u^3
  {0,3}, // v^3
  {2,1}, // u^2 v
  {1,2}  // u v^2
};

// ----------------------------------------------------------------------------

const int MAdMapping::P4_tri[12][2] = {
  {0,0}, // 1
  {1,0}, // u
  {0,1}, // v
  {2,0}, // u^2
  {0,2}, // v^2
  {3,0}, // u^3
  {0,3}, // v^3
  {4,0}, // u^4
  {0,4}, // v^4
  {3,1}, // u^3 v
  {1,3}, // u v^3
  {2,2}  // u^2 v^2
};

// -----------------------------------------------------------------------------

const int MAdMapping::P5_tri[15][2] = {
  {0,0}, // 1
  {1,0}, // u
  {0,1}, // v
  {2,0}, // u^2
  {0,2}, // v^2
  {3,0}, // u^3
  {0,3}, // v^4
  {4,0}, // u^4
  {0,4}, // v^4
  {5,0}, // u^5
  {0,5}, // v^5
  {4,1}, // u^4 v
  {3,2}, // u^3 v^2
  {2,3}, // u^2 v^3
  {1,4}  // u v^4
};

// -----------------------------------------------------------------------------
//   1             u             v             uv
const double MAdMapping::coef1_quad[4][4]={
  {  0.25000000,  -0.25000000,  -0.25000000,   0.25000000},
  {  0.25000000,   0.25000000,  -0.25000000,  -0.25000000},
  {  0.25000000,   0.25000000,   0.25000000,   0.25000000},
  {  0.25000000,  -0.25000000,   0.25000000,  -0.25000000},
};

// -----------------------------------------------------------------------------
//   1             u             v             uv            u^2           v^2            u^2v            uv^2         
const double MAdMapping::coef2_quad[8][8]={
  { -0.25000000,   0.00000000,   0.00000000,   0.25000000,   0.25000000,   0.25000000,  -0.25000000,  -0.25000000},
  { -0.25000000,   0.00000000,   0.00000000,  -0.25000000,   0.25000000,   0.25000000,  -0.25000000,   0.25000000},
  { -0.25000000,   0.00000000,   0.00000000,   0.25000000,   0.25000000,   0.25000000,   0.25000000,   0.25000000},
  { -0.25000000,   0.00000000,   0.00000000,  -0.25000000,   0.25000000,   0.25000000,   0.25000000,  -0.25000000},
  {  0.50000000,   0.00000000,  -0.50000000,   0.00000000,   0.50000000,   0.00000000,   0.50000000,   0.00000000},
  {  0.50000000,   0.50000000,   0.00000000,   0.00000000,   0.00000000,   0.50000000,   0.00000000,  -0.50000000},
  {  0.50000000,   0.00000000,   0.50000000,   0.00000000,  -0.50000000,   0.00000000,  -0.50000000,   0.00000000},
  {  0.50000000,  -0.50000000,   0.00000000,   0.00000000,   0.00000000,  -0.50000000,   0.00000000,   0.50000000},
};

// -----------------------------------------------------------------------------
//! \todo : TO RECOMPUTE ......... OK
//   1             u             v             uv            u^2           v^2            u^2v            uv^2       u^3           v^3           u^3v          uv^3         
const double MAdMapping::coef3_quad[12][12]={
  { -0.31250000,   0.31250000,   0.31250000,  -0.31250000,   0.28125000,   0.28125000,  -0.28125000,  -0.28125000,  -0.28125000,  -0.28125000,   0.28125000,   0.28125000},
  { -0.31250000,  -0.31250000,   0.31250000,   0.31250000,   0.28125000,   0.28125000,  -0.28125000,   0.28125000,   0.28125000,  -0.28125000,  -0.28125000,  -0.28125000},
  { -0.31250000,  -0.31250000,  -0.31250000,  -0.31250000,   0.28125000,   0.28125000,   0.28125000,   0.28125000,   0.28125000,   0.28125000,   0.28125000,   0.28125000},
  { -0.31250000,   0.31250000,  -0.31250000,   0.31250000,   0.28125000,   0.28125000,   0.28125000,  -0.28125000,  -0.28125000,   0.28125000,  -0.28125000,  -0.28125000},
  {  0.28125000,  -0.84375000,  -0.28125000,   0.84375000,  -0.28125000,   0.00000000,   0.28125000,   0.00000000,   0.84375000,   0.00000000,  -0.84375000,   0.00000000},
  {  0.28125000,   0.84375000,  -0.28125000,  -0.84375000,  -0.28125000,   0.00000000,   0.28125000,   0.00000000,  -0.84375000,   0.00000000,   0.84375000,   0.00000000},
  {  0.28125000,   0.28125000,  -0.84375000,  -0.84375000,   0.00000000,  -0.28125000,   0.00000000,  -0.28125000,   0.00000000,   0.84375000,   0.00000000,   0.84375000},
  {  0.28125000,   0.28125000,   0.84375000,   0.84375000,   0.00000000,  -0.28125000,   0.00000000,  -0.28125000,   0.00000000,  -0.84375000,   0.00000000,  -0.84375000},
  {  0.28125000,   0.84375000,   0.28125000,   0.84375000,  -0.28125000,   0.00000000,  -0.28125000,   0.00000000,  -0.84375000,   0.00000000,  -0.84375000,   0.00000000},
  {  0.28125000,  -0.84375000,   0.28125000,  -0.84375000,  -0.28125000,   0.00000000,  -0.28125000,   0.00000000,   0.84375000,   0.00000000,   0.84375000,   0.00000000},  
  {  0.28125000,  -0.28125000,   0.84375000,  -0.84375000,   0.00000000,  -0.28125000,   0.00000000,   0.28125000,   0.00000000,  -0.84375000,   0.00000000,   0.84375000},
  {  0.28125000,  -0.28125000,  -0.84375000,   0.84375000,   0.00000000,  -0.28125000,   0.00000000,   0.28125000,   0.00000000,   0.84375000,   0.00000000,  -0.84375000},
};

// -----------------------------------------------------------------------------

const int MAdMapping::P1_quad[4][2] = {
  {0,0}, // 1
  {1,0}, // u
  {0,1}, // v
  {1,1}, // uv
};

// -----------------------------------------------------------------------------

const int MAdMapping::P2_quad[8][2] = {
  {0,0}, // 1
  {1,0}, // u
  {0,1}, // v
  {1,1}, // u v
  {2,0}, // u^2
  {0,2}, // v^2
  {2,1}, // u^2 v
  {1,2}, // u v^2
};

// -----------------------------------------------------------------------------

const int MAdMapping::P3_quad[12][2] = {
  {0,0}, // 1
  {1,0}, // u
  {0,1}, // v
  {1,1}, // u v
  {2,0}, // u^2
  {0,2}, // v^2
  {2,1}, // u^2 v
  {1,2}, // u v^2
  {3,0}, // u^3
  {0,3}, // v^3
  {3,1}, // u^3 v
  {1,3}, // u v^3
};

// -----------------------------------------------------------------------------

const double MAdMapping::coef1_edge[2][2] = {{0.5, -0.5},
                                             {0.5, 0.5}};

// -----------------------------------------------------------------------------

const double MAdMapping::coef2_edge[3][3] = {{0, -0.5, 0.5},
                                             {1, 0, -1},
                                             {0, 0.5, 0.5}};

// -----------------------------------------------------------------------------

const double MAdMapping::coef3_edge[4][4] = {{-0.0625, 0.0625, 0.5625, -0.5625},
                                             {0.5625, -1.6875, -0.5625, 1.6875},
                                             {0.5625, 1.6875, -0.5625, -1.6875},
                                             {-0.0625, -0.0625, 0.5625, 0.5625}};

// -----------------------------------------------------------------------------

const double MAdMapping::coef4_edge[5][5] = {{0, 0.166666666666667, -0.166666666666667, -0.666666666666667, 0.666666666666667},
                                             {0, -1.33333333333333, 2.66666666666667, 1.33333333333333, -2.66666666666667},
                                             {1, -5.80822592402949e-20, -5, 0, 4},
                                             {0, 1.33333333333333, 2.66666666666667, -1.33333333333333, -2.66666666666667},
                                             {0, -0.166666666666667, -0.166666666666667, 0.666666666666667, 0.666666666666667}};

// -----------------------------------------------------------------------------

const double MAdMapping::coef5_edge[6][6] = {{0.01171875, -0.01171875, -0.325520833333333, 0.325520833333333, 0.813802083333333, -0.813802083333333},
                                             {-0.09765625, 0.162760416666667, 2.5390625, -4.23177083333333, -2.44140625, 4.06901041666667},
                                             {0.5859375, -2.9296875, -2.21354166666667, 11.0677083333333, 1.62760416666667, -8.13802083333333},
                                             {0.5859375, 2.9296875, -2.21354166666667, -11.0677083333333, 1.62760416666667, 8.13802083333333},
                                             {-0.09765625, -0.162760416666667, 2.5390625, 4.23177083333333, -2.44140625, -4.06901041666667},
                                             {0.01171875, 0.01171875, -0.325520833333333, -0.325520833333333, 0.813802083333333, 0.813802083333333}};

const double MAdMapping::coef3_tri_lagrange[10][10] = {
  {1, -5.5, -5.5, 9, 9, 18, -4.5, -4.5, -13.5, -13.5},
  {0, 1, 0, -4.5, 0, 0, 4.5, 0, 0, 0},
  {0, 0, 1, 0, -4.5, 0, 0, 4.5, 0, 0},
  {0, 9, 0, -22.5, 0, -22.5, 13.5, 0, 27, 13.5},
  {0, -4.5, 0, 18, 0, 4.5, -13.5, 0, -13.5, 0},
  {0, 0, 0, 0, 0, -4.5, 0, 0, 13.5, 0},
  {0, 0, 0, 0, 0, -4.5, 0, 0, 0, 13.5},
  {0, 0, -4.5, 0, 18, 4.5, 0, -13.5, 0, -13.5},
  {0, 0, 9, 0, -22.5, -22.5, 0, 13.5, 13.5, 27},
  {0, 0, 0, 0, 0, 27, 0, 0, -27, -27}};




const double MAdMapping::coef4_tri_lagrange[15][15] = {
  {1, -8.33333333333333, -8.33333333333333, 23.3333333333333, 23.3333333333333, 46.6666666666667, -26.6666666666667, -26.6666666666667, -80, -80, 10.6666666666667, 10.6666666666667, 42.6666666666667, 42.6666666666667, 64},
  {0, -1, 0, 7.33333333333333, 0, 0, -16, 0, 0, 0, 10.6666666666667, 0, 0, 0, 0},
  {0, 0, -1, 0, 7.33333333333333, 0, 0, -16, 0, 0, 0, 10.6666666666667, 0, 0, 0},
  {0, 16, 0, -69.3333333333333, 0, -69.3333333333333, 96, 0, 192, 96, -42.6666666666667, 0, -128, -42.6666666666667, -128},
  {0, -12, 0, 76, 0, 28, -128, 0, -144, -16, 64, 0, 128, 0, 64},
  {0, 5.33333333333333, 0, -37.3333333333333, 0, -5.33333333333333, 74.6666666666667, 0, 32, 0, -42.6666666666667, 0, -42.6666666666667, 0, 0},
  {0, 0, 0, 0, 0, 5.33333333333333, 0, 0, -32, 0, 0, 0, 42.6666666666667, 0, 0},
  {0, 0, 0, 0, 0, 4, 0, 0, -16, -16, 0, 0, 0, 0, 64},
  {0, 0, 0, 0, 0, 5.33333333333333, 0, 0, 0, -32, 0, 0, 0, 42.6666666666667, 0},
  {0, 0, 5.33333333333333, 0, -37.3333333333333, -5.33333333333333, 0, 74.6666666666667, 0, 32, 0, -42.6666666666667, 0, -42.6666666666667, 0},
  {0, 0, -12, 0, 76, 28, 0, -128, -16, -144, 0, 64, 0, 128, 64},
  {0, 0, 16, 0, -69.3333333333333, -69.3333333333333, 0, 96, 96, 192, 0, -42.6666666666667, -42.6666666666667, -128, -128},
  {0, 0, 0, 0, 0, 96, 0, 0, -224, -224, 0, 0, 128, 128, 256},
  {0, 0, 0, 0, 0, -32, 0, 0, 160, 32, 0, 0, -128, 0, -128},
  {0, 0, 0, 0, 0, -32, 0, 0, 32, 160, 0, 0, 0, -128, -128}};




const double MAdMapping::coef5_tri_lagrange[21][21] = {
  {1, -11.4166666666667, -11.4166666666667, 46.875, 46.875, 93.75, -88.5416666666667, -88.5416666666667, -265.625, -265.625, 78.125, 78.125, 312.5, 312.5, 468.75, -26.0416666666667, -26.0416666666667, -130.208333333333, -130.208333333333, -260.416666666667, -260.416666666667},
  {0, 1, 0, -10.4166666666667, 0, 0, 36.4583333333333, 0, 0, 0, -52.0833333333333, 0, 0, 0, 0, 26.0416666666667, 0, 0, 0, 0, 0},
  {0, 0, 1, 0, -10.4166666666667, 0, 0, 36.4583333333333, 0, 0, 0, -52.0833333333333, 0, 0, 0, 0, 26.0416666666667, 0, 0, 0, 0},
  {0, 25, 0, -160.416666666667, 0, -160.416666666667, 369.791666666667, 0, 739.583333333333, 369.791666666667, -364.583333333333, 0, -1093.75, -364.583333333333, -1093.75, 130.208333333333, 0, 520.833333333333, 130.208333333333, 781.25, 520.833333333333},
  {0, -25, 0, 222.916666666667, 0, 97.9166666666667, -614.583333333333, 0, -739.583333333333, -125, 677.083333333333, 0, 1406.25, 52.0833333333333, 781.25, -260.416666666667, 0, -781.25, 0, -781.25, -260.416666666667},
  {0, 16.6666666666667, 0, -162.5, 0, -37.5, 510.416666666667, 0, 322.916666666667, 20.8333333333333, -625, 0, -781.25, 0, -156.25, 260.416666666667, 0, 520.833333333333, 0, 260.416666666667, 0},
  {0, -6.25, 0, 63.5416666666667, 0, 6.25, -213.541666666667, 0, -57.2916666666667, 0, 286.458333333333, 0, 156.25, 0, 0, -130.208333333333, 0, -130.208333333333, 0, 0, 0},
  {0, 0, 0, 0, 0, -6.25, 0, 0, 57.2916666666667, 0, 0, 0, -156.25, 0, 0, 0, 0, 130.208333333333, 0, 0, 0},
  {0, 0, 0, 0, 0, -4.16666666666667, 0, 0, 31.25, 20.8333333333333, 0, 0, -52.0833333333333, 0, -156.25, 0, 0, 0, 0, 260.416666666667, 0},
  {0, 0, 0, 0, 0, -4.16666666666667, 0, 0, 20.8333333333333, 31.25, 0, 0, 0, -52.0833333333333, -156.25, 0, 0, 0, 0, 0, 260.416666666667},
  {0, 0, 0, 0, 0, -6.25, 0, 0, 0, 57.2916666666667, 0, 0, 0, -156.25, 0, 0, 0, 0, 130.208333333333, 0, 0},
  {0, 0, -6.25, 0, 63.5416666666667, 6.25, 0, -213.541666666667, 0, -57.2916666666667, 0, 286.458333333333, 0, 156.25, 0, 0, -130.208333333333, 0, -130.208333333333, 0, 0},
  {0, 0, 16.6666666666667, 0, -162.5, -37.5, 0, 510.416666666667, 20.8333333333333, 322.916666666667, 0, -625, 0, -781.25, -156.25, 0, 260.416666666667, 0, 520.833333333333, 0, 260.416666666667},
  {0, 0, -25, 0, 222.916666666667, 97.9166666666667, 0, -614.583333333333, -125, -739.583333333333, 0, 677.083333333333, 52.0833333333333, 1406.25, 781.25, 0, -260.416666666667, 0, -781.25, -260.416666666667, -781.25},
  {0, 0, 25, 0, -160.416666666667, -160.416666666667, 0, 369.791666666667, 369.791666666667, 739.583333333333, 0, -364.583333333333, -364.583333333333, -1093.75, -1093.75, 0, 130.208333333333, 130.208333333333, 520.833333333333, 520.833333333333, 781.25},
  {0, 0, 0, 0, 0, 250, 0, 0, -979.166666666667, -979.166666666667, 0, 0, 1250, 1250, 2500, 0, 0, -520.833333333333, -520.833333333333, -1562.5, -1562.5},
  {0, 0, 0, 0, 0, 41.6666666666667, 0, 0, -354.166666666667, -41.6666666666667, 0, 0, 833.333333333333, 0, 312.5, 0, 0, -520.833333333333, 0, -520.833333333333, 0},
  {0, 0, 0, 0, 0, 41.6666666666667, 0, 0, -41.6666666666667, -354.166666666667, 0, 0, 0, 833.333333333333, 312.5, 0, 0, 0, -520.833333333333, 0, -520.833333333333},
  {0, 0, 0, 0, 0, -125, 0, 0, 906.25, 281.25, 0, 0, -1562.5, -156.25, -1718.75, 0, 0, 781.25, 0, 1562.5, 781.25},
  {0, 0, 0, 0, 0, 31.25, 0, 0, -187.5, -187.5, 0, 0, 156.25, 156.25, 1093.75, 0, 0, 0, 0, -781.25, -781.25},
  {0, 0, 0, 0, 0, -125, 0, 0, 281.25, 906.25, 0, 0, -156.25, -1562.5, -1718.75, 0, 0, 0, 781.25, 781.25, 1562.5}
};

const int MAdMapping::P_tri_lagrange[78][2] = {
  //order 0 :
  {0,0},
  //order 1:
  {1,0},
  {0,1},
  // order 2:
  {2,0},
  {0,2},
  {1,1},
  // order 3:
  {3,0},
  {0,3},
  {2,1},
  {1,2},
  //order 4:
  {4,0},
  {0,4},
  {3,1},
  {1,3},
  {2,2},
  // order 5:
  {5,0},
  {0,5},
  {4,1},
  {1,4},
  {3,2},
  {2,3},
  //order 6
  {6,0},
  {0,6},
  {5,1},
  {1,5},
  {4,2},
  {2,4},
  {3,3},
  //order 7
  {7,0},
  {0,7},
  {6,1},
  {1,6},
  {5,2},
  {2,5},
  {4,3},
  {3,4},
  //order 8
  {8,0},
  {0,8},
  {7,1},
  {1,7},
  {6,2},
  {2,6},
  {5,3},
  {3,5},
  {4,4},
  //order 9
  {9,0},
  {0,9},
  {8,1},
  {1,8},
  {7,2},
  {2,7},
  {6,3},
  {3,6},
  {5,4},
  {4,5},
  // order 10
  {10,0},
  {0,10},
  {9,1},
  {1,9},
  {8,2},
  {2,8},
  {7,3},
  {3,7},
  {6,4},
  {4,6},
  {5,5},
  // order 11
  {11,0},
  {0,11},
  {10,1},
  {1,10},
  {9,2},
  {2,9},
  {8,3},
  {3,8},
  {7,4},
  {4,7},
  {6,5},
  {5,6}
};


