// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#ifndef H_MADMAPPING
#define H_MADMAPPING

#include <vector>
#include <list>
#include <cmath>
#include <algorithm>
using std::max;

#include "MeshDataBase.h"
#include "MeshDataBaseInterface.h"
#include "MAd_Interpolant.h"

using namespace MAd;
#include <map>
#include "MAdMatrix.h"
#include "MAdVector3D.h"


/*!\brief Generic Lagrange mapping */
/*!The geometry is defined by Lagrangian interpolants, and hence exactly maps
  the knots \f$(\vec U_i,\vec X_i)\f$ \f{eqnarray*} \vec X(\vec U) &= \sum_i
  \vec X_i \cdot \psi_i(u,v,w) \\ \psi_i(\vec U_j) &= \delta_{ij} \f}
*/

double norme(double a[3]);
void prodve(double a[3], double b[3], double c[3]);

using namespace MAd;

class MAdMapping 
{
protected:
  pEntity ent;              //!< mesh entity
  int entityTag;            //!< Gmsh tag for element type 
  int entityDim;            //!< dimension 
  int entityType;           //!< Gmsh base line type 
  int entityOrder;          //!< interpolation order 
  bool entityIncomplete;    //!< Serendipity element ? 
  int numVertices;          //!< number of principal vertices
  int nbKnots;              //!< total number of control points
  double lengthScale;       //!< rescaling length
  double overD;             //!< 1 over element diameter 
  doubleMatrix knots ;      //!< control point coordinates
  doubleMatrix coef;        //!< polynomial leading coefficients
  doubleMatrix powerof;     //!< monomial powers for the expansion in the polynomial
  doubleMatrix knotsCoef;   //!< generalized coordinates of the control points
  
  const MAd_Interpolant* interpolant;
  
  void getSerendipCoefficients();
  
  /*\brief get values of all monomials in (u,v,w) */
  void monomeof(const double& u, 
                const double& v,
                const double& w, 
                doubleVector& m) const;
  
  /* \brief get derivatives of all monomials in (u,v,w) */
  void dmonomeof(const double& u, 
                 const double& v, 
                 const double& w,
                 doubleMatrix& g) const;
  
  int geomOrder() const;
  
  virtual int order() const;

public:

  static void getVertices(pEntity e, std::vector<pVertex  > &vl );
  
public:  

  MAdMapping(pEntity me,double s);

  virtual ~MAdMapping(){};

  inline pEntity getEntity () const {return ent;}

  virtual void normalVector (pEntity border , 
                             double u, 
                             double v, 
                             double w, 
                             double n[3]) const;

  void normalVector (pEntity border,
                     int direction,
                     int orientation,
                     const double* uvw,
                     double* normal) const;

  bool computeCut(const MAd_Vector3D& point,
                  const MAd_Vector3D& normal,
                  int nbPoints,
                  doubleMatrix& coord,
                  doubleMatrix& param) const;

  double determinant(const double* uvw) const;


  void eval(double u, double v, double w,
            double &x, double &y, double &z) const;

  void evalLinear(double u, double v, double w,
                  double &x, double &y, double &z) const;

  void deval(double u, double v, double w,
             double &dxdu, double &dydu, double &dzdu,
             double &dxdv, double &dydv, double &dzdv,
             double &dxdw, double &dydw, double &dzdw) const;

  void COG (double &u, double &v, double &w) const;

  double parametricProjection(double &u,
                              double &v,
                              double &w,
                              double eps=0) const;

  bool project(double xo,double yo,double zo,
               double& up,double& vp,double& wp,double& distance) const;
  
  bool projectAlongDirection(const double* ori, //!< point to be projected
                             const double* dir, //!< direction
                             double* uvw,       //!< parametric coordinates of projection
                             double* xyz,       //!< Cartesian coordinates of projection
                             double& distance,  //!< parametric distance
                             double tol=1e-10   //!< convergence criterion
                             ) const;

  double invertedJacobian(double x, double y, double z,doubleMatrix& Invjac) const;
  
  double invertedJacobian(const double* uvw,double* invJac) const;

  /*! \brief compute the metric tensor \f$ \mathcal M :
    ds^2 = d \xi^T \cdot \mathcal M d \xi \f$ */
  void metricTensor(const double* uvw, //!< point in the element
                    double* metric     //!< metric tensor 
                    ) const;
  
  void invertLinear(double x, double y, double z,
                    double &u, double &v, double &w) const;

  bool invert(double x, double y, double z,
              double &u, double &v, double &w,
              double eps=1e-10,int maxiter=50,
              int np=1) const;

  bool inReferenceElement(double u, double v, double w,double eps=1e-5) const;

  void boundingBox (double bbox[3][2]) const;

  int getNumberOfKnots() const {return nbKnots;}

  int getEdgeNumber(pEntity e) const;

  int getFaceNumber(pEntity f) const;

  bool getVertexUVW(int i,double& u,double& v,double& w) const;

  void getEdgePointsOnFace(int k, int dir, double quadPt,double &u, double &v) const;

  void getFacePointsOnElement(int k,int dir,int rot,double ksi,double eta,
                              double& u,double& v,double& w) const;
  
  int dim() const {return EN_dim(ent);}

  double getMaxRadius() const;

  inline double getCoef(int i, int j) const{return coef(i,j);};

  inline void getMonomialExponents(int i, int &pu, int &pv, int &pw) const {
    pu = (int) powerof(i,0);
    pv = (powerof.size2() > 1) ? (int) powerof(i,1) : 0;
    pw = (powerof.size2() > 2) ? (int) powerof(i,2) : 0;
    return;};


protected:

  virtual void GeomShapeFunction(double u,double v,double w,
                                 double* ) const;
  
  virtual void GradGeomShapeFunction(double u,double v,double w,
                                     doubleMatrix&) const;
  
  bool project2D(double xo,double yo,double zo,
                 double& up,double& vp,double& wp,
                 double& distance) const;
  
  void distanceJacobianProduct(double,double,double,
                               double,double,double,
                               double*) const;

protected: 

  /*!\brief triangle template - nodes */
  static const double triNodeUVW[3][2];
  
  /*!\brief triangle template - edges */
  static const int triEdgeToNodes[3][2];

  /*!\brief quadrangle template - nodes */
  static const double quadNodeUVW[4][2];
  
  /*!\brief quadrangle template - edges */
  static const int quadEdgeToNodes[4][2];

  /*!\brief tetrahedron template  - nodes */
  static const double tetNodeUVW[4][3];
  
  /*!\brief tetrahedron template  - edges */
  static const int tetEdgeToNodes[6][2];
  
  /*!\brief tetrahedron template  - faces */
  static const int tetFaceToNodes[4][3];

  /*!\brief hexahedron template  - nodes */
  static const double hexNodeUVW[8][3];
  
  /*!\brief hexahedron template  - edges */
  static const int hexEdgeToNodes[12][2];
  
  /*!\brief hexahedron template  - faces */
  static const int hexFaceToNodes[6][4];

  /*!\brief prism template  - nodes */
  static const double prismNodeUVW[6][3];
  
  /*!\brief prism template  - edges */
  static const int prismEdgeToNodes[9][2];
  
  /*!\brief prism template  - faces */
  static const int prismFaceToNodesTri[2][3];
  static const int prismFaceToNodesQuad[3][4];

  /*!\brief pyramid template  - nodes */
  static const double pyramidNodeUVW[5][3];
  
  /*!\brief prism template  - edges */
  static const int pyramidEdgeToNodes[8][2];
  
  /*!\brief prism template  - faces */
  static const int pyramidFaceToNodesTri[4][3];
  static const int pyramidFaceToNodesQuad[1][4];


protected: // edge polynomials

  static const double coef1_edge[2][2];
  static const double coef2_edge[3][3];
  static const double coef3_edge[4][4];
  static const double coef4_edge[5][5];
  static const double coef5_edge[6][6];

protected: // Serendipity triangle interpolants
  
  static const double coef1_tri[3][3];
  static const double coef2_tri[6][6];
  static const double coef3_tri[9][9];
  static const double coef4_tri[12][12];
  static const double coef5_tri[15][15];
  
  static const int P0_tri[1][2];
  static const int P1_tri[3][2];
  static const int P2_tri[6][2];
  static const int P3_tri[9][2];
  static const int P4_tri[12][2];
  static const int P5_tri[15][2];
  
protected: // Serendipity quadrangle interpolants
  
  static const int P1_quad[4][2];
  static const int P2_quad[8][2];
  static const int P3_quad[12][2];
  static const int P4_quad[16][2];
  static const double coef1_quad[4][4];
  static const double coef2_quad[8][8];
  static const double coef3_quad[12][12];
  static const double coef4_quad[16][16];

protected: // Lagrange triangle interpolants
  
  static const double coef3_tri_lagrange[10][10];
  static const double coef4_tri_lagrange[15][15];
  static const double coef5_tri_lagrange[21][21];
  
  static const int P_tri_lagrange[78][2];  

protected : // Lagrange quadrangle interpolants
  
  static const double coef1_quad_lagrange[4][4];

protected: // tetrahedron interpolants

  static const double pos1_tet[4][3];
  static const double pos2_tet[10][3];
  //   static const double pos3_tet[20][3];
  //   static const double pos4_tet[34][3];
  static const double pos4_tet_lagrange[35][3];
  
  static const int P1_tet[4][3];
  static const int P2_tet[10][3];
  static const int P3_tet[20][3];
  static const int P4_tet[34][3];
  static const int P4_tet_lagrange[35][3];
  
  static const double coef1_tet[4][4];
  static const double coef2_tet[10][10];
  static const double coef3_tet[20][20];
  static const double coef4_tet[34][34];
  static const double coef4_tet_lagrange[35][35];
  
};
#endif
