// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#ifndef _H_MAPPINGINTERFACE
#define _H_MAPPINGINTERFACE

#include "MAdExport.h"
#include "MeshDataBaseInterface.h"

class MAdMapping;

/*! \defgroup api_mapping Element mapping 
  \ingroup api */

namespace MAd {

  /*!\typedef type definition of pointers to a MAdMapping class object */
  typedef class MAdMapping * pMapping;

  /*!\brief constructor of mapping object */
  /*!\ingroup api_mapping */
  MAD_EXPORT 
  pMapping Map_create(pEntity, //!< pointer to an entity object
                      double); //!< reference length scale
  
  /*!\brief destructor of mapping object  */
  /*!\ingroup api_mapping */
  MAD_EXPORT 
  void Map_delete(pMapping); //!< pointer to mapping object

  /*!\brief gives the dimension of the entity required to define a mapping */
  /*!\ingroup api_mapping */
  MAD_EXPORT 
  int Map_dim(pMapping); //!< pointer to mapping object 

  /*!\brief returns the number of knots of the element */
  /*!\ingroup api_mapping */
  MAD_EXPORT 
  int Map_getNumberOfKnots(pMapping map);

  /*!\brief gives the parametric coordinates of control point i */
  /*! returns false if not existent */
  /*!\ingroup api_mapping */
  MAD_EXPORT 
  bool Map_getVertexUVW(pMapping mapping, //!< pointer to a mapping object
                        int idx, //!< index 
                        double* uvw); //!< parametric coordinates of the vertex
  
  /*!\brief computes the parametric coordinates of the center of gravity */
  /*!\ingroup api_mapping */
  MAD_EXPORT 
  void Map_COG(pMapping, //!< pointer to a mapping object
               double*); //!< parametric coordinates of the center of gravity
  
  /*!\brief computes the edge index related to face definition */
  /*!\ingroup api_mapping */
  MAD_EXPORT 
  int Map_getEdgeNumber(pMapping, //!< pointer to a mapping object
                        pEntity); //!< pointer to an entity object
  
  /*!\brief computes the face index related to element definition */
  /*!\ingroup api_mapping */
  MAD_EXPORT 
  int Map_getFaceNumber(pMapping, //!< pointer to a mapping object
                        pEntity); //!< pointer to an entity object
  
  /*!\brief computes two points coordinates to define a bounding box
     around an element */
  /*!\ingroup api_mapping */
  MAD_EXPORT 
  void Map_boundBox(pMapping, //!< pointer to a mapping object
                    double[3][2]); //!< coordinate min/max
  
  /*!\brief computes the inscribed circle radius */
  /*!\ingroup api_mapping */
  MAD_EXPORT 
  double Map_getMaxRadius(pMapping //!< pointer to a mapping object
                          );          
  
  /*!\brief map a 1D reference element to an edge of a face */
  /*!\ingroup api_mapping */
  MAD_EXPORT 
  void Map_getEdgePtsOnFace(pMapping, //!< pointer to a mapping object
                            int, //!< index of the edge in the face
                            int, //!< direction
                            double, //!< parametric coordinates in edge
                            double*); //!< face coordinates 
  
  /*!\brief map a 2D reference element to a face of an element */
  /*!\ingroup api_mapping */
  MAD_EXPORT 
  void Map_getFacePtsOnElt(pMapping, //!< pointer to a mapping object
                           int,      //!< index of the face 
                           int,      //!< direction
                           int,      //!< number of rotations
                           double,   //!< 1st parametric face coordinate
                           double,   //!< 2nd parametric face coordinate
                           double*); //!< coordinates in the element

  /*!\brief evaluates the coordinates of points in the physical space
     given their parametric coordinates */
  /*!\ingroup api_mapping */
  MAD_EXPORT 
  void Map_eval(pMapping,       //!< pointer to a mapping object
                const double*,  //!< pointer to parametric coordinates
                double*);       //!< pointer to physical coordinates
  
  /*!\brief evaluates the derivatives of physical coordinates towards
     parametric coordinates */
  /*!\ingroup api_mapping */
  MAD_EXPORT 
  void Map_deval(pMapping,      //!< pointer to a mapping object
                 const double*, //!< pointer to parametric coordinates
                 double*);      //!< derivative values at physical points
  
  /*!\brief compute the Jacobian of the mapping at the given point */
  /*!\ingroup api_mapping */
  MAD_EXPORT 
  double Map_determinant(pMapping mapping,    //!< pointer to mapping
                         const double* uvw);  //!< parametric coordinates
  
  /*! \brief computes the inverse of the Jacobian of the mapping and
      returns the determinant */
  /*!\ingroup api_mapping */
  MAD_EXPORT 
  double Map_invertedJacobian(pMapping mapping,  //!< pointer to mapping
                              const double* uvw, //!< parametric coordinates
                              double* invJac);   //!< inverse of the mapping Jacobian (row major)

  /*! \brief computes local metric \f$ \mathcal M \f$ stored in row-major */
  /*! \f$ \mathcal M_{ij} : ds^2 = d \xi^T \cdot \mathcal M \cdot d \xi \f$ */
  /*! \f$ \mathcal M_{ij} = \dd{x_k}{\xi_i} \dd{x_k}{\xi_j} \f$ */
  /*!\ingroup api_mapping */
  void Map_metricTensor(pMapping mapping,
                        const double* uvw,
                        double* metric);
  
  /*!\brief computes the parametric coordinates of the projected point
     (on the parametric space) */
  /*!\ingroup api_mapping */
  MAD_EXPORT 
  double Map_paraProject(pMapping mapping, //!< mapping object
                         double* uvw,      //!< parametric coordinates
                         double eps = 0    //!< distance inside the element
                         );   
  
  /*!\brief project a point along a given direction on a face */
  /*!\ingroup api_mapping */
  MAD_EXPORT 
  bool Map_projAlongDir(pMapping mapping,   //!< pointer to a mapping object
                        const double* pnt,  //!< physical coordinates 
                        const double* dir,  //!< defining projection direction
                        double* uvw,        //!< parametric coordinates 
                        double* xyz,        //!< projected point
                        double& distance,   //!< projection distance 
                        double tol=1e-10);  //!< checking tolerance
  
  /*!\brief bool computes the parametric coordinates of the projected points */
  /*!\ingroup api_mapping */
  MAD_EXPORT 
  bool Map_project(pMapping, //!< pointer to a mapping object
                   const double*, //!< physical coordinates of the points
                   double*, //!< parametric coordinates of the projected points
                   double&); //!< distance between origin and projection

  /*!\brief computes the parametric coordinates of a point */
  /*!\ingroup api_mapping */
  MAD_EXPORT 
  bool Map_invert(pMapping,        //!< pointer to a mapping object
                  const double*,   //!< physical coordinates of a point
                  double*,         //!< parametric coordinates of a point
                  double=1e-10,    //!< tolerance
                  int=50,          //!< maximum number of iterations
                  int np=1         //!< number of samples in each direction 
                  );                  
  
  /*!\brief computes the normal to a border for given orientation/rotation */
  /*!\ingroup api_mapping */
  MAD_EXPORT 
  void Map_normal(pMapping mapping,  //!< element mapping
                  pEntity  border,   //!< boundary          
                  int direction,     //!< pointing in/out     
                  int rotation,      //!< number of rotations 
                  const double* uvw, //!< parametric coordinates in face
                  double* normal     //!< resulting normal
                  );
  
  /*!\brief compute a cut of a face with a plane */
  /*!\ingroup api_mapping */
  MAD_EXPORT 
  bool Map_planeCut(pMapping mapping, //!< pointer to a mapping object
                    const double* origin, //!< origin of the plane
                    const double* normal, //!< normal vector to the plane
                    int nbPoints,         //!< number of points along the cut
                    double* xyz,          //!< coordinates of the points
                    double* uvw);         //!< parametric coordinates
  
  /*!\brief check if a point belongs to the reference element */
  /*!\ingroup api_mapping */
  MAD_EXPORT 
  bool Map_inRefElt(pMapping mapping, //!< pointer to a mapping object
                    const double* uvw, //!< parametric coordinates 
                    double tol=1e-5);  //!< tolerance
}
#endif

