// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#include "CheckMesh.h"

#include "MAdMessage.h"
#include "MeshDataBase.h"

#include <stdio.h>
#include <sstream>

namespace MAd {


    // -------------------------------------------------------------------
    //! Dump informations on the mesh
    //! \ingroup api_mesh_mesh
    void M_info(const pMesh mesh)
    {
        std::stringstream out;
        
        out << "\n";
        out << "Mesh statistics:\n";
        out << "  Number of tetrahedra        : " << M_numTets(mesh) << "\n";
        out << "  Number of hexahedra         : " << M_numHexes(mesh) << "\n";
        out << "  Number of prisms            : " << M_numPrisms(mesh) << "\n";
        out << "  Number of pyramids          : " << M_numPyramids(mesh) << "\n";
        out << "  Number of triangles         : " << M_numTriangles(mesh) << "\n";
        out << "  Number of quadrangles       : " << M_numQuads(mesh) << "\n";
        out << "  Number of edges             : " << M_numEdges(mesh) << "\n";
        out << "  Number of vertices          : " << M_numVertices(mesh) << "\n";
        out << "  Number of high order points : " << M_numPoints(mesh) << "\n";
        out << "\n";

        std::string iStr = out.str();
        MAdMsgSgl::instance().write(iStr);
    }

    // -------------------------------------------------------------------
    //! Dumps quality informations about the region. \ingroup api_mesh_region
    void R_info_quality(const pRegion region, std::ostream& out)
    {
        out << "  Quality informations:\n";
        double volume = R_volume(region);
        out << "    Volume      : " << volume << "\n";
        if (volume < 0.) out << "  *** Negative volume ***\n";
        else {
            out << "    r/R ratio : " << R_inscrRad(region) / R_circumRad(region) << "\n";
            double meanRatio3;
            R_meanRatioCube(region, &meanRatio3);
            out << "    Cubic mean ratio : " << meanRatio3 << "\n";
        }
    }

    // -------------------------------------------------------------------
    //! Dumps topology informations about the region. \ingroup api_mesh_region
    void R_info_topology(const pRegion region, std::ostream& out)
    {
        out << "  Topology informations:\n";
        out << "\n    Faces       (ids):   type  geom ent.      tag   orient\n";
        for (int iF = 0; iF<R_numFaces(region); iF++) {
            pFace face = R_face(region, iF);
            pGEntity pGE = F_whatIn(face);
            int gTag = pGE ? GEN_tag(pGE) : -1;
            int gDim = pGE ? GEN_type(pGE) : -1;
            out << " " << face << " (" << EN_id((pEntity)face) << ") "
                << gDim << "   " << pGE << " " << gTag << " "
                << R_faceDir(region, iF) << "\n";
        }
        out << "\n    Edges       (ids):   type  model ent.      tag   orient\n";
        pPList rEdges = R_edges(region);
        void * temp = NULL;
        while (pEdge edge = (pEdge)PList_next(rEdges, &temp)) {
            pGEntity pGE = E_whatIn(edge);
            int gTag = pGE ? GEN_tag(pGE) : -1;
            int gDim = pGE ? GEN_type(pGE) : -1;
            out << " " << edge << " (" << EN_id((pEntity)edge) << ") "
                << gDim << "   " << pGE << " " << gTag << "\n";
        }
        PList_delete(rEdges);

        out << "\nVertices    (ids):   type  model ent.      tag               coordinates\n";
        pPList rVerts = R_vertices(region);
        temp = NULL;
        while (pVertex pV = (pVertex)PList_next(rVerts, &temp)) {
            double xyz[3];
            V_coord(pV, xyz);
            pGEntity pGE = V_whatIn(pV);
            int gTag = pGE ? GEN_tag(pGE) : -1;
            int gDim = pGE ? GEN_type(pGE) : -1;
            out << " " << pV << " (" << EN_id((pEntity)pV) << ") "
                << gDim << "   " << pGE << " " << gTag << " "
                << xyz[0] << " " << xyz[1] << " " << xyz[2] << "\n";
        }
        PList_delete(rVerts);
    }

    // -------------------------------------------------------------------
    //! Dumps all informations about the region. \ingroup api_mesh_region
    void R_info(const pRegion region, const char * name)
    {
        std::stringstream out;

        out << "\nRegion \'" << name << "\' (" << region
            << ", Id: " << (int)EN_id((pEntity)region)
            << ") informations:\n";

        out << "  Classification:  ";
        pGEntity pGE = (pGEntity)R_whatIn(region);
        if (!pGE) {
            out << "NULL";
        }
        else {
            out << "GEntity: " << pGE
                << ", dim: " << GEN_type(pGE)
                << ", tag: " << GEN_tag(pGE);
        }
        out << "\n";

        R_info_quality(region, out);
        R_info_topology(region, out);

        out << "\n";

        std::string iStr;
        out >> iStr;
        MAdMsgSgl::instance().write(iStr);
    }
    // -------------------------------------------------------------------
    //! Dumps informations about the face. \ingroup api_mesh_face
    void F_info(const pFace face, const char * name)
    {
        std::stringstream out;

        out << "\n";
        out << "Face \'" << name << "\' " << face
            << ", id: " << EN_id((pEntity)face) << "\n";
        out << "  Classified on " << face->g
            << " with dim " << GEN_type(face->g)
            << " and tag " << GEN_tag(face->g) << "\n";

        out << "\n--- Regions:\n";
        for (int iR = 0; iR<face->getNbRegions(); iR++) {
            pRegion region = face->getRegion(iR);
            if (!region) {
                out << "Region " << iR << ": " << region << "\n";
            }
            else {
                out << "Region " << iR << " (" << region
                    << ", Id: " << (int)EN_id((pEntity)region)
                    << ") informations:\n";
                out << "  Classification:  ";
                pGEntity pGE = (pGEntity)R_whatIn(region);
                if (!pGE) {
                    out << "NULL";
                }
                else {
                    out << "GEntity: " << pGE
                        << ", dim: " << GEN_type(pGE)
                        << ", tag: " << GEN_tag(pGE);
                }
                out << "  Vertex opposite to face: \n";
                V_info(R_fcOpVt(region, face), "");
                out << "\n";
            }
        }

        out << "\n--- Edges:\n";
        for (int iE = 0; iE<face->getNbEdges(); iE++) {
            pEdge edge = face->getEdge(iE);
            out << "Edge " << iE << ":\n";
            E_info(edge, "");
        }

        out << "\n";

        std::string iStr;
        out >> iStr;
        MAdMsgSgl::instance().write(iStr);
    }

    // -------------------------------------------------------------------
    //! Dumps informations about the edge. \ingroup api_mesh_edge
    void E_info(const pEdge edge, const char * name)
    {
        std::stringstream out;

        out << "\n";
        out << "Edge \'" << name << "\' " << edge
            << ", id: " << EN_id((pEntity)edge) << "\n";
        pGEntity pGE = EN_whatIn((pEntity)edge);
        out << "  - Classification: dim: " << GEN_type(pGE) << ", tag: " << GEN_tag(pGE) << "\n";
        pVertex pv0 = E_vertex(edge, 0);
        pVertex pv1 = E_vertex(edge, 1);
        out << "  - Vertices:\n"
            << "     * Vertex 0:\n";
        V_info(pv0, "");
        out << "     * Vertex 1:\n";
        V_info(pv1, "");
        out << "\n";

        std::string iStr;
        out >> iStr;
        MAdMsgSgl::instance().write(iStr);
    }

    void V_info(const pVertex vertex, const char * name)
    {
        std::stringstream out;

        out << "\n";
        out << "Vertex \'" << name << "\' " << vertex
            << ", id: " << EN_id((pEntity)vertex) << "\n";
        pGEntity pGE = EN_whatIn((pEntity)vertex);
        out << "  - Classification: dim: " << GEN_type(pGE) << ", tag: " << GEN_tag(pGE) << "\n";
        out << "  - Num edges: " << vertex->edges.size() << "\n";
        out << "  - Parameters:\n";
        double u, v;
        if (vertex->getParams(&u, &v)) {
            out << "     Point is parametric: (u,v) = (" << u << ", " << v << ")\n";
        }
        else {
            out << "      Point is not parametric\n";
        }
        out << "  - Coordinates:\n";
        double xyz[3];
        V_coord(vertex, xyz);
        out << "     ( " << xyz[0] << ", " << xyz[1] << ", " << xyz[2] << " )\n";
        out << "\n";

        std::string iStr;
        out >> iStr;
        MAdMsgSgl::instance().write(iStr);
    }



  // -------------------------------------------------------------------
  // checkGeometricalCompatibility:

  // *** For 2D meshes ***

  // --- Faces ---
  //   - check every face is classified on a surface

  // --- Edges ---
  // For every edge:
  //   - if is classified on a surface: the edge is used by exactly 
  //     2 faces with the same classification as the edge,
  //   - if classified on a line: the edge is used by exactly 1 face
  //     or the edge is used by exactely 2 faces which are classified
  //     on distinct faces.
  //   - is not classified on a point
  
  // --- Vertices ---
  // For every vertex:
  //   - if classified on a surface: all edges using it have to be 
  //     classified on the same surface,
  //   - if classified on a line: exactly two edges classified on the 
  //     same line use it. No other edge classified on a line use it.
  //   - if classified on a point, at least two edges using it are
  //     classified on lines (it can be the same line: revoluted disk)

  // *** For 3D meshes ***

  // --- Regions ---
  //  - check every mesh region is classified on a model region

  // --- Faces ---
  // For every face:
  //   - if classified on a model region: the face is used by exactly 
  //     2 mesh regions with the same classification as the face,
  //   - if classified on a surface: the face is used by exactly 1 
  //     mesh region or the face is used by exactly 2 mesh regions
  //     which are classified on distinct model regions
  //   - is not classified on a line
  //   - is not classified on a point

  // --- Edges ---
  // For every edge:
  //   - if classified on a model region: check that all faces using 
  //     it are classified on the same model region,
  //   - if classified on a surface: check that exactly 2 faces using it 
  //     are classified on the same surface, and that no face is classified 
  //     on a different surface,
  //   - if classified on a line: check that at least two faces using it 
  //     are classified on surfaces (it can be the same surface: revoluted 
  //     cylinder).
  //   - is not classified on a point

  // --- Vertices ---
  // For every vertex:
  //   - if classified on a model region: all edges using it have to be 
  //     classified on the same model region,
  //   - if classified on a surface: at least 3 edges using it are 
  //     classified on the same surface, and no edge can be classified 
  //     on a line or on a different surface.
  //   - if classified on a line: exactly 2 edges using it are 
  //     classified on the same line and no other edge is classified 
  //     on a line,
  //   - if classified on a point: at least 1 edge using it is 
  //     classified on a line (only 1: for revoluted cones).

  int checkGeomCompatibility2D(MDB_Mesh * mesh, int verbose)
  {
    int gDim;
    int flag = 1;

    std::stringstream out;

    // --- Faces ---
    pFace pf;
    FIter fIt = M_faceIter(mesh);
    while ( flag && ( pf = FIter_next(fIt) ) ) {
      gDim = F_whatInType(pf);
      if ( gDim != 2 ) {
        flag = 0;
        if (verbose) {
          out <<  "Mesh face not classified on a model face\n";
          F_info(pf,"");
        }
        break;
      }
    }
    FIter_delete(fIt);
    if (!flag) {
        std::string iStr;
        out >> iStr;
        MAdMsgSgl::instance().write(iStr);
        return 0;
    }

    // --- Edges ---
    pEdge pe;
    EIter eIt = M_edgeIter(mesh);
    while ( flag && ( pe = EIter_next(eIt) ) ) {
      gDim = E_whatInType(pe);
      switch (gDim) {
        case 2: {
          int nFac = E_numFaces(pe);
          if ( nFac != 2 ) {
            flag = 0;
            if (verbose) {
              out << "Edge classif on dim 2 with not exactly 2 faces but "
                  << nFac << "\n";
              E_info(pe,"");
            }
          }
          else if ( ( E_whatIn(pe) != (pGEntity)F_whatIn( E_face(pe,0) ) ) ||
                    ( E_whatIn(pe) != (pGEntity)F_whatIn( E_face(pe,1) ) ) ) {
            flag = 0;
            if (verbose) {
              out << "Edge classif on dim 2 used by faces classif on different entities\n";
              E_info(pe,"");
            }
          }
          break;
        }
        case 1: {
          int nFac = E_numFaces(pe);
          if ( nFac == 2 ) {
            if ( (pGEntity)F_whatIn( E_face(pe,0) ) ==
                 (pGEntity)F_whatIn( E_face(pe,1) ) ) { 
              flag = 0;
              if (verbose) {
                out <<"Edge classif on dim 1 used by exactely 2 faces classif on the same entity\n";
                E_info(pe,"");
              }
            }
          }
          else if ( nFac != 1 ) {
            flag = 0;
            if (verbose) {
              out << "Edge classif on dim 1 with not exactly 1 face but "
                  << nFac << "\n";
              E_info(pe,"");
            }
          }
          break;
        }
        default: {
          flag = 0;
          if (verbose) {
            out << "Edge classif on wrong dimension\n";
            E_info(pe,"");
          }
        }
      }
    }
    EIter_delete(eIt);
    if (!flag) {
        std::string iStr;
        out >> iStr;
        MAdMsgSgl::instance().write(iStr);
        return 0;
    }

    // --- Vertices ---
    pVertex pv;
    pGEntity pgv;
    pPList vEdges;
    void * temp;
    VIter vIt = M_vertexIter(mesh);
    while ( flag && ( pv = VIter_next(vIt) ) ) {
      pgv  = V_whatIn(pv);
      gDim = GEN_type(pgv);
      switch (gDim) {
        case 2: {
          vEdges = V_edges(pv);
          temp = NULL;
          while ( ( pe = (pEdge) PList_next(vEdges,&temp) ) ) {
            if ( E_whatIn(pe) != pgv ) {
              flag = 0;
              if (verbose) {
                out << "Vertex classif on a surface but used by edges "
                    << "not classif on the same surface\n";
                V_info(pv,"");
              }
            }
          }
          PList_delete(vEdges);
          break;
        }
        case 1: {
          int nbE = 0;
          vEdges = V_edges(pv);
          temp = NULL;
          while ( ( pe = (pEdge) PList_next(vEdges,&temp) ) ) {
            if ( E_whatInType(pe) == 1 ) {
              if ( E_whatIn(pe) != pgv ) {
                flag = 0;
                if (verbose) {
                  out << "Vertex classif on a line used by an edge classif "
                      << "on another line with tag " << GEN_tag(E_whatIn(pe)) <<"\n";
                  V_info(pv,"");
                }
              }
              nbE++;
            }
          }
          PList_delete(vEdges);
          if ( nbE != 2 ) {
            flag = 0;
            if (verbose) {
              out << "Vertex classif on a line and not used by exactly "
                  << "2 edges classif on lines but " << nbE << "\n";
              V_info(pv,"");
            }
          }
          break;
        }
        case 0: {
          int nbE = 0;
          vEdges = V_edges(pv);
          temp = NULL;
          while ( ( pe = (pEdge) PList_next(vEdges,&temp) ) ) {
            if ( E_whatInType(pe) == 1 ) nbE++;
          }
          PList_delete(vEdges);
          if ( nbE < 2 ) {
            flag = 0;
            if (verbose) {
              out << "Vertex classif on a point and not used by at least"
                  << " 2 edges classif on lines but " <<  nbE << "\n";
              V_info(pv,"");
            }
          }
          break;
        }
        default: {
          flag = 0;
          if( verbose ) {
            out << "Vertex classif on non-existing dimension\n";
            V_info(pv,"");
          }
        }
      }
    }
    VIter_delete(vIt);

    std::string iStr;
    out >> iStr;
    MAdMsgSgl::instance().write(iStr);

    if ( !flag ) return 0;
    return flag;
  }

  int checkGeomCompatibility(MDB_Mesh * mesh, int verbose)
  {
#ifdef PARALLEL
    // In a mesh partitioned with MDB, the parallel boundary 
    // faces are not classified on a model face and the following 
    // check fails
    return 1;
#endif

    if ( M_dim(mesh) < 2 ) return 1;

    if ( M_dim(mesh) < 3 )
      return checkGeomCompatibility2D( mesh, verbose );

    int gDim;
    int flag = 1;
    std::stringstream out;

    // --- Regions ---
    pRegion pr;
    RIter rIt= M_regionIter(mesh);
    while ( ( pr = RIter_next(rIt) ) )
      {
        gDim = R_whatInType(pr);
        if ( gDim != 3 ) {
          flag = 0;
          if (verbose) {
            out <<  "Mesh region not classified on a model region\n";
            R_info_topology(pr,out);
          }
          break;
        }
      }
    RIter_delete(rIt);
    if (!flag) {
        std::string iStr;
        out >> iStr;
        MAdMsgSgl::instance().write(iStr);
        return 0;
    }

    // -- Faces ---
    pFace pf;
    FIter fIt= M_faceIter(mesh);
    while ( flag && ( pf = FIter_next(fIt) ) )
      {
        gDim = F_whatInType(pf);
        switch (gDim) {
        case 3: {
          int nRgn = F_numRegions(pf); 
          if ( nRgn != 2 ) {
            flag = 0;
            if (verbose) {
              out << "Face classif on dim 3 with not exactly 2 regions but "
                  << nRgn << "\n";
              F_info(pf,"");
            }
          }
          else if ( ( F_whatIn(pf) != (pGEntity)R_whatIn( F_region(pf,0) ) ) || 
               ( F_whatIn(pf) != (pGEntity)R_whatIn( F_region(pf,1) ) ) ) {
            flag = 0;
            if (verbose) {
              out << "Face classif on dim 3 used by regions classif on different entities\n";
              F_info(pf,"");
            }
          }
          break;
        }
        case 2: {
          int nRgn = F_numRegions(pf);
          if ( nRgn == 2 ) {
            if ( (pGEntity)R_whatIn( F_region(pf,0) ) ==
                 (pGEntity)R_whatIn( F_region(pf,1) ) ) { 
              flag = 0;
              if (verbose) {
                out <<"Face classif on dim 2 used by exactely 2 regions classif on the same entity\n";
                F_info(pf,"");
              }
            }
          }
          else if ( nRgn != 1 ) {
            flag = 0;
            if (verbose) {
              out << "Face classif on dim 2 with not exactly 1 region but "
                  << nRgn << "\n";
              F_info(pf,"");
            }
          }
          break;
        }
        default: {
          flag = 0;
          if (verbose) {
            out << "Face classif on wrong dimension\n";
            F_info(pf,"");
          }
        }
        }
      }
    FIter_delete(fIt);
    if (!flag) {
        std::string iStr;
        out >> iStr;
        MAdMsgSgl::instance().write(iStr);
        return 0;
    }

    // -- Edges ---
    pEdge pe;
    pGEntity pge;
    pPList eFaces;
    void * temp;
    EIter eIt = M_edgeIter(mesh);
    while ( flag && ( pe = EIter_next(eIt) ) )
      {
        pge  = E_whatIn(pe);
        gDim = GEN_type(pge);
        switch (gDim) {
        case 3: {
          eFaces = E_faces(pe);
          temp = NULL;
          while ( ( pf = (pFace) PList_next(eFaces,&temp) ) ) {
            if ( F_whatIn(pf) != pge ) {
              flag = 0;
              if (verbose) {
                out << "Edge classif on a region but used by faces "
                    << "not classif on the same region\n";
                E_info(pe,"");
              }
            }
          }
          PList_delete(eFaces);
          break;
        }
        case 2: {
          int nbBF = 0;
          eFaces = E_faces(pe);
          temp = NULL;
          while ( ( pf = (pFace) PList_next(eFaces,&temp) ) ) {
            if ( F_whatInType(pf) == 2 ) {
              if ( F_whatIn(pf) != pge ) {
                flag = 0;
                if (verbose) {
                  out << "Edge classif on a surface used by a face classif "
                      << "on another surface with tag " << GEN_tag(F_whatIn(pf)) <<"\n";
                  E_info(pe,"");
                }
              }
              nbBF++;
            }
          }
          PList_delete(eFaces);
          if ( nbBF != 2 ) {
            flag = 0;
            if (verbose) {
              out << "Edge classif on a surface and not used by exactly "
                  << "2 faces classif on surfaces but " << nbBF << "\n";
              E_info(pe,"");
            }
          }
          break;
        }
        case 1: {
          int nbBF = 0;
          eFaces = E_faces(pe);
          temp = NULL;
          while ( ( pf = (pFace) PList_next(eFaces,&temp) ) ) {
            if ( F_whatInType(pf) == 2 ) nbBF++;
          }
          PList_delete(eFaces);
          if ( nbBF < 2 ) {
            flag = 0;
            if (verbose) {
              out << "Edge classif on a line and not used by at least"
                  << " 2 faces classif on surfaces but " <<  nbBF << "\n";
              E_info(pe,"");
            }
          }
          break;
        }
        default: {
          flag = 0;
          if (verbose) {
            out << "Edge classif on wrong dimension\n";
            E_info(pe,"");
          }
        }
        }
      }
    EIter_delete(eIt);
    if (!flag) {
        std::string iStr;
        out >> iStr;
        MAdMsgSgl::instance().write(iStr);
        return 0;
    }

    // -- Vertices ---
    pPList vEdges;
    pVertex pv;
    VIter vIt = M_vertexIter(mesh);
    while ( flag && ( pv = VIter_next(vIt) ) )
      {
        pge  = V_whatIn(pv);
        gDim = GEN_type(pge);
        switch (gDim) {
        case 3: {
          vEdges = V_edges(pv);
          temp = NULL;
          while ( ( pe = (pEdge) PList_next(vEdges,&temp) ) ) {
            if ( E_whatIn(pe) != pge ) {
              flag = 0;
              if (verbose) {
                out << "Vertex classif on a region but used by edges "
                    << "not classif on the same region\n";
                V_info(pv,"");
              }
            }
          }
          PList_delete(vEdges);
          break;
        }
        case 2: {
          int nbE = 0;
          vEdges = V_edges(pv);
          temp = NULL;
          while ( ( pe = (pEdge) PList_next(vEdges,&temp) ) ) {
            if ( E_whatInType(pe) == 3 ) continue;
            if ( E_whatIn(pe) != pge ) {
              flag = 0;
              if (verbose) {
                out << "Vertex classif on a surface but used by edges "
                    << "not classif on the same surface or on a region\n";
                V_info(pv,"");
              }
            }
            nbE++;
          }
          PList_delete(vEdges);
          if ( nbE < 3 ) {
            flag = 0;
            if (verbose) {
              out << "Vertex classif on a surface and not used by at least"
                  << " 3 edges classif on surfaces  but " << nbE << "\n";
              V_info(pv,"");
            }
          }
          break;
        }
        case 1: {
          int nbE = 0;
          vEdges = V_edges(pv);
          temp = NULL;
          while ( ( pe = (pEdge) PList_next(vEdges,&temp) ) ) {
            if ( E_whatInType(pe) > 1 ) continue;
            if ( E_whatIn(pe) != pge ) {
              flag = 0;
              if (verbose) {
                out << "Vertex classif on a line and used by an edge "
                    << "classif on another line\n";
                V_info(pv,"");
              }
            }
            nbE++;
          }
          PList_delete(vEdges);
          if ( nbE != 2 ) {
            flag = 0;
            if (verbose) {
              out << "Vertex classif on a line and not used by exactly"
                  << " 2 edges classif on the same line\n";
              V_info(pv,"");
            }
          }
          break;
        }
        case 0: {
          int nbE = 0;
          std::set<pGEntity> glines;
          vEdges = V_edges(pv);
          temp = NULL;
          while ( ( pe = (pEdge) PList_next(vEdges,&temp) ) ) {
            if ( E_whatInType(pe) > 1 ) continue;
            nbE++;
            glines.insert(E_whatIn(pe));
          }
          PList_delete(vEdges);
          if ( nbE == 0 ) {
            flag = 0;
            if (verbose) {
              out << "Vertex classif on a point and not used by an"
                  << " edge classif on a line\n";
              V_info(pv,"");
            }
          }
//           if ( nbE < 2 ) {
//             flag = 0;
//             if (verbose) {
//               out << "Vertex classif on a point and used by less than"
//                   << " 2 edges classif on lines\n";
//               V_info(pv,"");
//             }
//           }
//           if ( glines.size() < 2 ) {
//             flag = 0;
//             if (verbose) {
//               out << "Vertex classif on a point and not used by edges"
//                   << " classified on different lines\n";
//               V_info(pv,"");
//             }
//           }
          break;
        }
        default: {
          flag = 0;
          if (verbose) {
            out << "Vertex classif on non-existing dimension\n";
            V_info(pv,"");
          }
        }
        }
      }
    VIter_delete(vIt);

    std::string iStr;
    out >> iStr;
    MAdMsgSgl::instance().write(iStr);

    if ( !flag ) return 0;
    return flag;
  }

  // -------------------------------------------------------------------
  int checkRegionsVolume(MDB_Mesh * mesh, int verbose)
  {
      std::stringstream out;

    int flag = 1;
    pRegion r;
    RIter rIt= M_regionIter(mesh);
    while ( ( r = RIter_next(rIt) ) ) {
      if (R_volume(r) < 0.) {
        flag = 0;
        if (verbose) {
          out << "Negative volume found !\n";
          R_info_topology(r,out);
        }
        break;
      }  
    }
    RIter_delete(rIt);

    std::string iStr;
    out >> iStr;
    MAdMsgSgl::instance().write(iStr);

    return flag;
  }

  // -------------------------------------------------------------------
  int checkEdgeToRegionConnectivity(MDB_Mesh * mesh, int verbose)
  {
#ifdef PARALLEL
    // In a mesh partitioned with MDB, the parallel boundary 
    // faces are not classified on a model face and the following 
    // check fails
    return 1;
#endif

    if ( M_dim(mesh) <= 2 )
        return 1;

    std::stringstream out;
    int flag = 1;
    pEdge edge;
    EIter eIt = M_edgeIter(mesh);
    while ( ( edge = EIter_next(eIt) ) ) {

      if ( E_whatInType(edge) != 3 ) continue;

      pFace face = E_face(edge,0);
      pRegion start_region = F_region(face,0);
      
      pFace current_face = face;
      pRegion current_region = start_region;
      pFace next_face;
      pRegion next_region = NULL;
      while ( next_region != start_region )
      {
        next_face = E_otherFace(edge, current_face, current_region);
        next_region = F_region(next_face, 0);
        if( next_region == current_region ) {
          next_region = F_region(next_face, 1);
        }
        current_face   = next_face;
        current_region = next_region;
        
        if ( !next_region ) {
          flag = 0;
          if (verbose) {
            out << "Found a wrong edge to region connectivity !\n";
            E_info(edge,"");
          }
          break;
        }
      }
    }
    EIter_delete(eIt);

    std::string iStr;
    out >> iStr;
    MAdMsgSgl::instance().write(iStr);

    return flag;
  }

  // -------------------------------------------------------------------
  // Obsolete: included in checkGeomCompatibility()
  int checkFaceToRegionConnectivity(MDB_Mesh * mesh, int verbose) {

#ifdef PARALLEL
    // In a mesh partitioned with MDB, the parallel boundary 
    // faces are not classified on a model face and the following 
    // check fails
    return 1;
#endif

    if ( M_dim(mesh) <= 2 )
        return 1;

    std::stringstream out;
    int flag = 1;
    pFace f;
    FIter fIt= M_faceIter(mesh);
    while ( ( f = FIter_next(fIt) ) ) {
      int ok = 1;
      int nRgn = F_numRegions(f); 
      if ( verbose && nRgn == 0 ) out << "nRgn: " << nRgn << "\n";
      int type = F_whatInType(f);
      if      (              nRgn == 0 ) ok = 0;
      else if ( type == 2 && nRgn >  2 ) ok = 0;
      else if ( type == 3 && nRgn != 2 ) ok = 0;
      if (!ok) {
        flag = 0;
        if (verbose) {
          out << "Found a wrong face to region connectivity !\n";
          F_info(f,"");
        }
        break;
      }  
    }
    FIter_delete(fIt);

    std::string iStr;
    out >> iStr;
    MAdMsgSgl::instance().write(iStr);

    return flag;
  }

  // -------------------------------------------------------------------
  int checkFaceToVertexConnectivity(MDB_Mesh * mesh, int verbose)
  {
      std::stringstream out;
    int flag = 1;
    pFace f;
    pVertex v0,v1,v2;
    FIter fIt = M_faceIter(mesh);
    while ( ( f = FIter_next(fIt) ) ) {
      v0 = F_vertex(f,0);
      v1 = F_vertex(f,1);
      v2 = F_vertex(f,2);
      if ( v0 == v1 || v0 == v2 || v1 == v2 ) {
        flag = 0;
        if (verbose) {
          out << "Found a wrong face to vertex connectivity !\n";
          F_info(f,"");
        }
        break;
      }
    }
    FIter_delete(fIt);

    std::string iStr;
    out >> iStr;
    MAdMsgSgl::instance().write(iStr);

    return flag;
  }

  // -------------------------------------------------------------------
  // Check that there is at least one face attached to each edge in 2D, 
  // one region and two faces in 3D.
  int checkEdgeConnectivity(MDB_Mesh * mesh, int verbose)
  {

#ifdef PARALLEL
    // In a mesh partitioned with MDB, the parallel boundary 
    // edges are not classified on a model line and the following 
    // check fails
    return 1;
#endif

    int dim = M_dim(mesh);
    if ( dim <= 1 )
        return 1;

    std::stringstream out;
    int flag = 1;
    pEdge e;
    EIter eIt= M_edgeIter(mesh);
    while ( ( e = EIter_next(eIt) ) ) {
      int nFace = E_numFaces(e);
      int nRgn  = E_numRegions(e);
      if ( dim == 2 ) {
        if ( (nRgn != 0) || (nFace < 1) ) {
          flag = 0; break;
        }
      }
      else if ( dim == 3 ) {
        if ( (nRgn < 1) || (nFace < 2) ) {
          flag = 0; break;
        }
      }
    }
    EIter_delete(eIt);
    if ( (!flag) && verbose ) {
      out << "Found an ill-connected edge !\n";
      E_info(e,"");
    }

    std::string iStr;
    out >> iStr;
    MAdMsgSgl::instance().write(iStr);

    return flag;
  }

  // -------------------------------------------------------------------
  int checkEntityPointers(MDB_Mesh * mesh, int verbose)
  {
      std::stringstream out;
    int flag = 1;

    VIter vIt = M_vertexIter(mesh);
    pVertex pv;
    while ( ( pv = VIter_next(vIt) ) ) {
      if ( !pv ) {
        flag = 0;
        if (verbose) out << "Found a null pointer when iterating on vertices\n";
      } 
    }
    VIter_delete(vIt);

    EIter eIt = M_edgeIter(mesh);
    pEdge pe;
    while ( ( pe = EIter_next(eIt) ) ) {
      if ( !pe ) {
        flag = 0;
        if (verbose) out << "Found a null pointer when iterating on edges\n";
      } 
    }
    EIter_delete(eIt);

    FIter fIt = M_faceIter(mesh);
    pFace pf;
    while ( ( pf = FIter_next(fIt) ) ) {
      if ( !pf ) {
        flag = 0;
        if (verbose) out << "Found a null pointer when iterating on faces\n";
      } 
    }
    FIter_delete(fIt);

    RIter rIt = M_regionIter(mesh);
    pRegion pr;
    while ( ( pr = RIter_next(rIt) ) ) {
      if ( !pr ) {
        flag = 0;
        if (verbose) out << "Found a null pointer when iterating on regions\n";
      } 
    }
    RIter_delete(rIt);

    std::string iStr;
    out >> iStr;
    MAdMsgSgl::instance().write(iStr);

    return flag;
  }

  // -------------------------------------------------------------------
  int checkIterators(MDB_Mesh * mesh, int verbose)
  {
      std::stringstream out;
    int flag = 1;

    int nbMeshV = M_numVertices(mesh);
    int nbIterV = 0;
    VIter vIt= M_vertexIter(mesh);
    while ( VIter_next(vIt) )  nbIterV++;
    VIter_delete(vIt);
    if ( nbMeshV != nbIterV ) {
      flag = 0;
      if (verbose) {
        out << "Incoherent number of vertices: " 
            << nbMeshV << " in mesh, " 
            << nbIterV << " in iterator\n";
      }
    }

    int nbMeshE = M_numEdges(mesh);
    int nbIterE = 0;
    EIter eIt= M_edgeIter(mesh);
    while ( EIter_next(eIt) )  nbIterE++;
    EIter_delete(eIt);
    if ( nbMeshE != nbIterE ) {
      flag = 0;
      if (verbose) {
        out << "Incoherent number of edges: "
            << nbMeshE << " in mesh, "
            << nbIterE << " in iterator\n";
      }
    }

    int nbMeshF = M_numFaces(mesh);
    int nbIterF = 0;
    FIter fIt= M_faceIter(mesh);
    while ( FIter_next(fIt) )  nbIterF++;
    FIter_delete(fIt);
    if ( nbMeshF != nbIterF ) {
      flag = 0;
      if (verbose) {
        out << "Incoherent number of faces: "
            << nbMeshF << " in mesh, "
            << nbIterF << " in iterator\n";
      }
    }

    int nbMeshR = M_numRegions(mesh);
    int nbIterR = 0;
    RIter rIt= M_regionIter(mesh);
    while ( RIter_next(rIt) )  nbIterR++;
    RIter_delete(rIt);
    if ( nbMeshR != nbIterR ) {
      flag = 0;
      if (verbose) {
        out << "Incoherent number of regions: "
            << nbMeshR << " in mesh, "
            << nbIterR << " in iterator\n";
      }
    }

    std::string iStr;
    out >> iStr;
    MAdMsgSgl::instance().write(iStr);

    return flag;
  }

  // -------------------------------------------------------------------
  int checkParameters(MDB_Mesh * mesh, int verbose)
  {
    if ( !M_isParametric(mesh) )
        return 1;

    std::stringstream out;
    int flag = 1;

    pVertex pv;
    double tmp0,tmp1;
    VIter vIt= M_vertexIter(mesh);
    while ( ( pv = VIter_next(vIt) ) ) {
      int gDim = V_whatInType(pv);
      bool param = V_params(pv,&tmp0,&tmp1);
      if ( ( gDim==0 || gDim==1 || gDim==2 ) && !param ) {
        flag = 0;
        if (verbose) {
          out << "Non parametric point classified on geo entity with dim " <<gDim<<"\n";
          V_info(pv,"");
        }
      }
      if ( ( gDim==3 ) && param ) {
        flag = 0;
        if (verbose) {
          out << "Parametric point classified on geo entity with dim " <<gDim<<"\n";
          V_info(pv,"");
        }
      }
    }
    VIter_delete(vIt);

    std::string iStr;
    out >> iStr;
    MAdMsgSgl::instance().write(iStr);

    return flag;
  }

  // -------------------------------------------------------------------
  bool checkMesh(MDB_Mesh * mesh, checkType type, int verbose, MeshStatus * status)
  {
    switch (type) {
    case CHECK_ALL: {
      if ( !checkRegionsVolume(mesh, verbose) ) {
        if(status) *status = NEGATIVE_VOLUME; 
        return false;
      }
      if ( !checkGeomCompatibility(mesh, verbose) ) {
        if(status) *status = GEOM_INCOMPATIBILITY;
        return false;
      }
      if ( !checkEdgeToRegionConnectivity(mesh, verbose) ) {
        if(status) *status = WRONG_EDGE_TO_RGN_CONN;
        return false;
      }
      if ( !checkFaceToRegionConnectivity(mesh, verbose) ) {
        if(status) *status = WRONG_FACE_TO_RGN_CONN;
        return false;
      }
      if ( !checkFaceToVertexConnectivity(mesh, verbose) ) {
        if(status) *status = WRONG_FACE_TO_VTX_CONN;
        return false;
      }
      if ( !checkEdgeConnectivity(mesh, verbose) ) {
        if(status) *status = WRONG_EDGE_CONN;
        return false;
      }
      if ( !checkEntityPointers(mesh, verbose) ) {
        if(status) *status = WRONG_ENTITY_POINTERS;
        return false;
      }
      if ( !checkIterators(mesh, verbose) ) {
        if(status) *status = WRONG_ITERATORS;
        return false;
      }
      if ( !checkParameters(mesh, verbose) ) {
        if(status) *status = WRONG_PARAMETERS;
        return false;
      }
      break;
    }
    case CHECK_VOLUME: {
      if ( !checkRegionsVolume(mesh, verbose) ) {
        if(status) *status = NEGATIVE_VOLUME;
        return false;
      }
      break;
    }
    case CHECK_GEOM_COMPATIBILITY: {
      if ( !checkGeomCompatibility(mesh, verbose) ) {
        if(status) *status = GEOM_INCOMPATIBILITY;
        return false;
      }
      break;
    }
    case CHECK_EDGE_TO_RGN_CONN: {
      if ( !checkEdgeToRegionConnectivity(mesh, verbose) ) {
        if(status) *status = WRONG_EDGE_TO_RGN_CONN;
        return false;
      }
      break;
    }
    case CHECK_FACE_TO_RGN_CONN: {
      if ( !checkFaceToRegionConnectivity(mesh, verbose) ) {
        if(status) *status = WRONG_FACE_TO_RGN_CONN;
        return false;
      }
      break;
    }
    case CHECK_FACE_TO_VTX_CONN: {
      if ( !checkFaceToVertexConnectivity(mesh, verbose) ) {
        if(status) *status = WRONG_FACE_TO_VTX_CONN;
        return false;
      }
      break;
    }
    case CHECK_EDGE_CONN: {
      if ( !checkEdgeConnectivity(mesh, verbose) ) {
        if(status) *status = WRONG_EDGE_CONN;
        return false;
      }
      break;
    }
    case CHECK_ENTITY_POINTERS: {
      if ( !checkEntityPointers(mesh, verbose) ) {
        if(status) *status = WRONG_ENTITY_POINTERS;
        return false;
      }
      break;
    }
    case CHECK_ITERATORS: {
      if ( !checkIterators(mesh, verbose) ) {
        if(status) *status = WRONG_ITERATORS;
        return false;
      }
      break;
    }
    case CHECK_PARAMETERS: {
      if ( !checkParameters(mesh, verbose) ) {
        if(status) *status = WRONG_PARAMETERS;
        return false;
      }
      break;
    }
    default: {
      MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                  "Not a valid checkType: %d",type);
      return false;
    }
    }

    if(status) *status = VALID;
    return true;
  }

  // -------------------------------------------------------------------

}
