// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#ifndef H_MESHDATABASE
#define H_MESHDATABASE
#include <set>
#include <vector>
#include <algorithm>
#include <list>
#include <map>
#include <cmath>
#include <assert.h>
#include <cstdarg>
#include <cstring>
#include <iostream>
#include <limits>
#include <functional>
#include "ModelInterface.h"
#include "MeshDataBaseIterators.h"
#include "MeshDataBaseAttachable.h"
#include "MeshDataBaseMiniMesh.h"
#include "MAdMessage.h"
#include "MathUtils.h"
#include "MshTags.h"
#ifdef PARALLEL
#include "mpi.h"
#endif

/*! \defgroup internal Internal functionalities */
/*! \defgroup int_mesh Mesh related functions and classes 
  \ingroup internal */

/*! \defgroup int_mesh_mesh      Mesh operations  
  \ingroup int_mesh */
/*! \defgroup int_mesh_entity    Entity operations 
  \ingroup int_mesh */
/*! \defgroup int_mesh_region    Region operations 
  \ingroup int_mesh */
/*! \defgroup int_mesh_face      Face operations 
  \ingroup int_mesh */
/*! \defgroup int_mesh_edge      Edge operations
  \ingroup int_mesh */
/*! \defgroup int_mesh_vertex    Vertex operations 
  \ingroup int_mesh */
/*! \defgroup int_mesh_point     Point operations 
  \ingroup int_mesh */
/*! \defgroup int_mesh_ios       Mesh in- and output 
  \ingroup int_mesh */
/*! \defgroup int_mesh_parallel  Communication 
  \ingroup int_mesh */

namespace MAd {

  class MDB_Region;
  class MDB_Tet;
  class MDB_Hex;
  class MDB_Prism;
  class MDB_Pyramid;
  class MDB_Quad;
  class MDB_Face;
  class MDB_Edge;
  class MDB_Triangle;
  class MDB_Mesh;
  class MDB_Point;

  typedef std::list<MDB_Point *>      MDB_ListP ;
  typedef std::list<MDB_Edge *>       MDB_ListE ;
  typedef std::list<MDB_Triangle *>   MDB_ListF ;
  typedef std::list<MDB_Quad*>        MDB_ListQ ;
  typedef std::list<MDB_Face*>        MDB_ListFace;
  typedef std::list<MDB_Tet *>        MDB_ListT ;
  typedef std::list<MDB_Hex *>        MDB_ListH ;
  typedef std::list<MDB_Prism *>      MDB_ListPr;
  typedef std::list<MDB_Pyramid *>    MDB_ListPy;
  typedef std::list<MDB_Region *>     MDB_ListR ;
  typedef std::vector<MDB_Triangle *> MDB_VectorF ;
  typedef std::vector<MDB_Quad *>     MDB_VectorQ ;
  typedef std::vector<MDB_Face *>     MDB_VectorFace ;
  typedef std::vector<MDB_Edge *>     MDB_VectorE ;
  typedef std::vector<MDB_Tet *>      MDB_VectorT ;
  typedef std::vector<MDB_Hex *>      MDB_VectorH ;
  typedef std::vector<MDB_Prism *>    MDB_VectorPr;
  typedef std::vector<MDB_Pyramid *>  MDB_VectorPy;

  class MDB_MeshEntity : public mAttachableDataContainer  
  {
  public :
    int iD;
    bool deleted;
    pGEntity g;
    virtual int getDim()   const = 0;
    virtual int getOrder() const = 0;
    virtual int getMshTag() const = 0;
    MDB_MeshEntity ( int _ID = 0, pGEntity _g = 0) : iD(_ID), deleted (false), g(_g) {}    
    virtual ~MDB_MeshEntity() {}
    virtual bool setClassification(pGEntity pg) {
      if (GEN_type(pg) < getDim())    return false;
      if (g) if (GEN_type(g)  < GEN_type(pg)) return false;
      g = pg;
      return true;
    }
  };

  /*! \brief Vertex */ 

  class MDB_Point : public MDB_MeshEntity
  {
  public:
    // SHOULD BE PRIVATE, 2B DONE
    double X,Y,Z;
    MDB_VectorE edges;

    virtual int getOrder() const {return 0;}
    virtual int getDim()   const {return 0;}
    virtual int getMshTag() const {return MSH_PNT;}
    virtual bool isParametric() const { return false; }
    virtual bool getParams(double * u, double * v) const { return false; }

    inline bool operator < (const MDB_Point & other) const
    {
      return iD < other.iD;
    }
    inline void del(MDB_Edge *e)
    {
      MDB_VectorE::iterator it  = edges.begin();
      MDB_VectorE::iterator ite = edges.end();
      while(it!=ite){
        if(*it == e){
          edges.erase(it);
          break;
        }
        ++it;
      }
    }
    void getFaces(MDB_ListFace &) const;  
    void getTets(MDB_ListT &) const; 
    void getHexes(MDB_ListH &) const;
  
    //void getHexs     (MDB_ListH &) const;

    MDB_Point(int id, double x=0, double y=0, double z=0) 
      : MDB_MeshEntity(id),X(x),Y(y),Z(z)
    {     
      //    edges.reserve(12);
    }
    virtual ~MDB_Point() {}
  };

  /*! \brief Parametric vertex */ 

  class MDB_PointParam : public MDB_Point
  {
  public:

    MDB_PointParam(int id, double x=0, double y=0, double z=0, double u=0, double v=0)
      :  MDB_Point(id,x,y,z),U(u),V(v)
    {}
    ~MDB_PointParam() {}

    virtual int getOrder() const {return 0;}
    bool isParametric() const { return true; }
    bool getParams(double * u, double * v) const { *u=U; *v=V; return true; }

  private:

    double U,V;
  };

/*! \brief Class for arbitrary order equidistant Lagrange edges */ 

class MDB_Edge: public MDB_MeshEntity
  {
    MDB_VectorFace _faces;
  public:
    // Variables
    int order;
    MDB_Point *p1,*p2;
    std::vector<MDB_Point *> HighOrderPoints;
    
    // Constructors
    
    // Ex : P1 Edge
    // x--------------------------------------x
    // A                                      B
    MDB_Edge(MDB_Point *A, MDB_Point *B)
      : p1(A),p2(B)
    {
      order=1;
      p1->edges.push_back(this);
      p2->edges.push_back(this);
    }
    // Ex : P3 Edge
    // x------------o------------o------------x
    // A        HOpoints[0]  HOpoints[1]      B
    MDB_Edge(MDB_Point *A, MDB_Point *B, std::vector<MDB_Point *> &HOpoints)
      : p1(A),p2(B)
    {
      p1->edges.push_back(this);
      p2->edges.push_back(this);
      
      order = ( (int) HOpoints.size() )+1;
      
      if ( order > 1 ) {
        for (int i = 0 ; i < order-1 ; i++) {
          HighOrderPoints.push_back( HOpoints[i] );
          HighOrderPoints[i]->edges.push_back(this);
        }
      }
    }
    // Ex : P3 Edge
    // x------------o------------o------------x
    // A        HOpoints[0]  HOpoints[1]      B
    MDB_Edge(MDB_Point *A, MDB_Point *B, int _order, MDB_Point ** HOpoints)
      : p1(A),p2(B)
    {
      order = _order;
      p1->edges.push_back(this);
      p2->edges.push_back(this);
      
      if ( order > 1 ) {
        for (int i = 0 ; i < order-1 ; i++) {
          HighOrderPoints.push_back( HOpoints[i] );
          HighOrderPoints[i]->edges.push_back(this);
        }
      }
    }
    
    // Destructors
    virtual ~MDB_Edge();
    
    // Class members
    virtual int getDim()   const {return 1;}
    
    virtual int getOrder() const {return order;}
    
    virtual int getMshTag() const {
      switch (getOrder()) {
          case 1:  return MSH_LIN_2;
          case 2:  return MSH_LIN_3;
          case 3:  return MSH_LIN_4;
          case 4:  return MSH_LIN_5;
          default: return MSH_UNDEFINED_ELEM;
      }
    }

    inline MDB_Point* getVertex(int i) const {
      if (i>1||i<0) return NULL;
      return (i==0) ? p1 : p2;
    }

    inline MDB_Face* faces(int i) const
    {
      if ( _faces.size() == 0 ) return NULL;
      if ( _faces.size() == 1 && i == 1 ) return NULL;
      return _faces [i];
    }

    inline double length() const
    {
      return sqrt((p1->X-p2->X)*(p1->X-p2->X)+(p1->Y-p2->Y)*(p1->Y-p2->Y)+(p1->Z-p2->Z)*(p1->Z-p2->Z));
    }
    
    double piecewiselength() const
    {
      if ( getOrder() == 1 ) return length();
      
      double len = 0;
      len += linearlength (p1, HighOrderPoints.front() );
      len += linearlength (HighOrderPoints.back(), p2 );
      if (getNbHighOrderPoints()>1)
        for (int i = 0 ; i < getNbHighOrderPoints()-1 ; i++) 
          len += linearlength (HighOrderPoints[i], HighOrderPoints[i+1] );
      return len;
    }

    inline int numfaces() const
    {
      return (int)(_faces.size());
    }

    inline MDB_Point * commonvertex(const MDB_Edge *other) const
    {
      if(p1 == other->p1 || p1 == other->p2) return p1;
      if(p2 == other->p1 || p2 == other->p2) return p2;
      printf("two edges %d %d -- %d %d without common vertex\n",p1->iD,p2->iD, other->p1->iD, other->p2->iD);
      throw;
    }

    inline MDB_Point * othervertex(const MDB_Point *p) const
    {
      if(p1 == p) return p2;
      if(p2 == p) return p1;
      return 0;
    }

    inline void addface(MDB_Face *f)
    {
      _faces.push_back(f);
    }

    inline MDB_Face * otherFace(const MDB_Face *f) const
    {
      if(numfaces()!=2) return NULL;
      if(f == _faces[0]) return _faces[1];
      if(f == _faces[1]) return _faces[0];
      MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                    "Requested other Face does not exists");
      return 0;
    }

    inline void del( MDB_Face *t )
    {
      _faces.erase(std::remove_if(_faces.begin(),_faces.end() , std::bind2nd(std::equal_to<MDB_Face*>(), t)) , 
                   _faces.end());
    }
  
    inline void oppositeof(MDB_Point * oface[2]) const; 
  
    virtual int getNbHighOrderPoints () const {return (int) HighOrderPoints.size();}
    
    virtual MDB_Point *getHighOrderPoint (int i) {
      if ( i >= getNbHighOrderPoints() )
        MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                    "Requested Edges High Order Point %i does not exists (Nb High Order Point : %i)",
                                    i,getNbHighOrderPoints() );
      
      return HighOrderPoints[i];
    }

    virtual void swapHONodes() {
      int nb = getNbHighOrderPoints();
      if ( nb > 1 ) {
        for (int i = 0 ; i < nb/2 ; i++) std::swap(HighOrderPoints[i] , HighOrderPoints[nb-i-1] );
      }
    }
    
    virtual int align(MDB_Point* po1,MDB_Point* po2);

    virtual bool setClassification(pGEntity pg) {
      if (!MDB_MeshEntity::setClassification(pg)) return false;
      p1->setClassification(pg);
      p2->setClassification(pg);
      
      if (getNbHighOrderPoints()) {
        for (int i = 0 ; i < getNbHighOrderPoints() ; i++)
          HighOrderPoints[i]->setClassification(pg);
      }
      
      return true;
    }
  private:
    
    inline double linearlength(MDB_Point* A,MDB_Point* B) const
    {
      return sqrt((A->X-B->X)*(A->X-B->X)+(A->Y-B->Y)*(A->Y-B->Y)+(A->Z-B->Z)*(A->Z-B->Z));
    }
  };

  /*! \brief Base class for edged Face and base class for higher order versions */ 
  
  class MDB_Face: public MDB_MeshEntity
  {
  protected:
    std::vector<MDB_Region *> _regions;
  public :
    virtual int        getNbEdges ()  const = 0;
    virtual int        getNbNodes ()  const = 0;
    virtual int        getNbVertices() const { return getNbNodes();}
    virtual MDB_Edge * getEdge  (int) const = 0;
    virtual int getDim() const {return 2;}
    virtual int getMshTag() const = 0;
    virtual bool complete() const = 0;

    inline MDB_Region* getRegion (int i) const
    {
      if (getNbRegions()==0)return 0;
      if (getNbRegions()==1 && i == 1)return 0;
      return _regions[i];
    }

    inline MDB_Region * opposite_region(MDB_Region *t)
    {
      if(t == getRegion(0))return getRegion(1);
      if(t == getRegion(1))return getRegion(0);
      throw;
    }

    inline int getNbRegions() const 
    {
      return (int)(_regions.size());
    }
  
    inline void addregion(MDB_Region *t)
    {
      _regions.push_back(t);
    }

    inline void del( MDB_Region *t )
    {
      _regions.erase(std::remove_if(_regions.begin(),_regions.end() , std::bind2nd(std::equal_to<MDB_Region*>(), t)) , 
                     _regions.end());
    }

    inline MDB_Edge * commonedge(const MDB_Face *other) const
    {
      for (int i=0;i<getNbEdges ();i++)
        for (int j=0;j<other->getNbEdges ();j++)
          {
            MDB_Edge *e = getEdge(i);
            if (e == other->getEdge(j)) return  e;
          }
      throw;
    }

    inline MDB_Point *getNode(int i) const
    {
      if (!i)
        {
          int nbe = getNbEdges ();
          return getEdge(0)->commonvertex(getEdge(nbe-1));
        }
      return getEdge(i)->commonvertex(getEdge(i-1));
    }

    inline MDB_Point *getVertex(int i) const { return getNode(i);}

    inline void getNodes(MDB_Point **n) const
    {
      for (int i=0;i<getNbEdges ();i++)n[i] = getNode(i);
    }
    virtual MDB_Edge * find_edge(const MDB_Point*, const MDB_Point*) const = 0;

    virtual MDB_Edge * getOpposingEdge(const MDB_Edge*) const {
      return NULL;
    }

    MDB_Face() /*: r1(0), r2(0) */{};
    virtual MDB_Point *getHighOrderPoint (int i) {throw;}
    virtual int getNbHighOrderPoints() const {return 0;}
  
    virtual int align(MDB_Point* v0,MDB_Point* v1,MDB_Point* v2,MDB_Point* v3) = 0;

    virtual bool setClassification(pGEntity pg) {
      if (!MDB_MeshEntity::setClassification(pg)) return false;
      for (int i=0;i<getNbEdges();i++) getEdge(i)->setClassification(pg);
      for (int i=0;i<getNbHighOrderPoints();i++) getHighOrderPoint(i)->setClassification(pg);
      return true;
    }


  protected:
  
    virtual int orientation(MDB_Point* v0,MDB_Point* v1,MDB_Point* v2,MDB_Point* v3) const;
  
  };

  /*! \brief Generic order equidistant Lagrange triangle with complete interpolation */ 

  class MDB_Triangle: public MDB_Face
  {
  public:

    MDB_Edge *e1,*e2,*e3;
    
    
    // Constructors
    
    /* Ex : 1st order Triangle
    C
    X
    |\
    | \
    |  \
    |   \
    |    \
    X-----X
    A     B
    */
    MDB_Triangle(MDB_Edge *A, MDB_Edge *B, MDB_Edge *C)
      : e1(A),e2(B),e3(C),nbPoints(0),point(NULL), isComplete(false)
    { 
      e1->addface(this);
      e2->addface(this);
      e3->addface(this);
      
      order = std::numeric_limits<int>::max();
      order = std::min(e1->getOrder(),order);
      order = std::min(e2->getOrder(),order);
      order = std::min(e3->getOrder(),order);
      
      isComplete = order < 3 ? true : false;
      
      // Allow Serendip ?
      /* 
      if ( order > 2 ) {
       MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                  "Incoherent interpolation orders for triangle and edges (%i). Internal Nodes required.",
                                  order);
       }
      */
      
    }
    
    /* Ex : 3rd order Triangle
    C
    X
    |\
    o o
    |  \
    o o o
    |    \
    X-o-o-X
    A     B
    */
    MDB_Triangle(MDB_Edge *A, MDB_Edge *B, MDB_Edge *C, int o, MDB_Point** P)
      : e1(A),e2(B),e3(C),nbPoints(std::max(0,(o-2)*(o-1)/2)),point(NULL), isComplete(false)
    {
      e1->addface(this);
      e2->addface(this);
      e3->addface(this);
      
      order = std::numeric_limits<int>::max();
      order = std::min(e1->getOrder(),order);
      order = std::min(e2->getOrder(),order);
      order = std::min(e3->getOrder(),order);
      
      
      if (order != o) {
        MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                    "Incoherent interpolation orders for triangle and edges (%i vs %i)",
                                    o,order);
      }
      
      if (order > 5) {
        MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                    "Complete triangle of order %d not yet implemented", order);
      }
      
      isComplete = (order < 3 || nbPoints > 0) ? true : false;
      point = nbPoints > 0 ? new MDB_Point*[nbPoints] : NULL;
      
      for (int i=0;i<nbPoints;i++) point[i] = P[i];
      
    }
    
    virtual ~MDB_Triangle() {if (point) delete [] point;}
    
    
    virtual int        getOrder()     const {return order;}
    int getMshTag() const {
      // is a Serendipity Element ?
      if (isComplete)
      {
        switch (order) {
            case 1:  return MSH_TRI_3;
            case 2:  return MSH_TRI_6;
            case 3:  return MSH_TRI_10;
            case 4:  return MSH_TRI_15;
            default: return MSH_UNDEFINED_ELEM;
        }
      }
      else
      {
        switch (order) {
            case 1:  return MSH_TRI_3;
            case 2:  return MSH_TRI_6;
            case 3:  return MSH_TRI_9;
            case 4:  return MSH_TRI_12;
            default: return MSH_UNDEFINED_ELEM;
        }
      }
      return MSH_UNDEFINED_ELEM;
    }
    
    virtual int        getNbEdges ()  const {return 3;}
    virtual int        getNbNodes ()  const {return 3;}
    virtual MDB_Edge * getEdge  (int i) const 
    {
      switch (i){
      case 0 : return e1;
      case 1 : return e2;
      case 2 : return e3;
      case 3 : return (MDB_Edge *)  NULL;
      }
      throw;
    }

    virtual bool complete() const {return isComplete;}
    virtual MDB_Edge * find_edge(const MDB_Point*, const MDB_Point*) const;

    int getNbHighOrderPoints() const {return nbPoints;}
    MDB_Point* getHighOrderPoint(int i) {return (i<nbPoints) ?  point[i] : NULL;}

    using MDB_Face::getOpposingEdge;
    
    virtual int alignP1(MDB_Point* v0,MDB_Point* v1,MDB_Point* v2,MDB_Point* v3);
    
    virtual int align(MDB_Point* v0,MDB_Point* v1,MDB_Point* v2,MDB_Point* v3) {


      if (v3 != NULL) {
        MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                    "Aligning a triangle with a quadrilateral is not allowed");
      }
    
      int rot = alignP1(v0,v1,v2,NULL);

      /* we only store the internal points (the higher order nodes on the edges are stored elsewhere)
         this means that this corresponds to a triangle of order = order - 3 */ 
      rotateHOPoints(rot,point,order-3);
    
      return rot;
    }
    
    /*! rotate a triangular set of points such that the node on position abs(rot) - 1
      is migrated to node 0; swap if rot < 0 */

    void rotateHOPoints(int rot,MDB_Point** op,int recursiveOrder) {
    
      if (recursiveOrder <= 0) return; // only one node needs no further reorientation 
      if (rot   == 1) return; // already aligned 
  
      int nbRecursivePoints = (recursiveOrder+1)*(recursiveOrder+2)/2;
      int nbEdgePoints  = (recursiveOrder - 1);
  
      MDB_Point** np = new MDB_Point*[nbRecursivePoints];
  
      // first rearrange principal vertices
  
      memcpy(np,op,nbRecursivePoints*sizeof(MDB_Point*));
  
      for (int i=0;i<3;i++) np[i] = op[(i + abs(rot) - 1)%3];
      if (rot < 0) std::swap(np[1],np[2]);
  
  
      if (recursiveOrder > 1)  {
    
        // then edge nodes
      
        if (rot > 0) {
          for (int i=0;i<3;i++) {
            MDB_Point** nb = np + 3 + i * nbEdgePoints;
            MDB_Point** ob = op + 3 + (i + rot - 1) % 3 * nbEdgePoints;
            for (int j=0;j<nbEdgePoints;j++) memcpy(nb,ob,nbEdgePoints * sizeof(MDB_Point*));
          }
        }
      
        if (rot < 0) {
          for (int i=0;i<3;i++) {
            MDB_Point** nb = np + 2 + ((abs(rot) - i + 1) % 3 + 1)  * nbEdgePoints;      
            MDB_Point** ob = op + 3 + i * nbEdgePoints;
            for (int j=0;j<nbEdgePoints;j++) *nb-- = *ob++;
          }
        }
      
        rotateHOPoints(rot,np + 3 * (nbEdgePoints+1) , recursiveOrder - 3);
      }
    
      memcpy(op,np,nbRecursivePoints * sizeof(MDB_Point*));
      delete [] np;
    }

  protected:
    
    int order;
    int nbPoints;
    MDB_Point** point;
    bool isComplete;

  };

  /*! \brief Generic order equidistant Lagrange quad with complete interpolation */ 
  
  class MDB_Quad : public MDB_Face {

  public:
    MDB_Edge *e1,*e2,*e3,*e4;
  
    MDB_Quad(MDB_Edge *A, MDB_Edge *B, MDB_Edge *C, MDB_Edge *D)
      : e1(A),e2(B),e3(C),e4(D), nbPoints(0), point(NULL) 
    { 
      e1->addface(this);
      e2->addface(this);
      e3->addface(this);
      e4->addface(this);

      order = std::numeric_limits<int>::max();
      order = std::min(e1->getOrder(),order);
      order = std::min(e2->getOrder(),order);
      order = std::min(e3->getOrder(),order);
      order = std::min(e4->getOrder(),order);
      
      // Serendipity Quad ?
      isComplete = order < 2 ? true : false;
    }
  
    MDB_Quad(MDB_Edge *A, MDB_Edge *B, MDB_Edge *C,MDB_Edge *D, int o,MDB_Point** P):
      e1(A),e2(B),e3(C),e4(D),nbPoints(std::max(0,(o-1)*(o-1))),point(NULL) {
      
      e1->addface(this);
      e2->addface(this);
      e3->addface(this);
      e4->addface(this);

      order = std::numeric_limits<int>::max();
      order = std::min(e1->getOrder(),order);
      order = std::min(e2->getOrder(),order);
      order = std::min(e3->getOrder(),order);
      order = std::min(e4->getOrder(),order);
      
      if (order != o) {
        MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                    "Incoherent order of quad wrt edges (%i vs %i)", o,order);
      }
    
      order = o;
      
      if (order > 4) {
        MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                    "Complete quad of order %d not yet implemented", order);
      }
      
      // Simple serendipity check ...
      isComplete = (order < 2 || nbPoints > 0) ? true : false;
      point = nbPoints > 0 ? new MDB_Point*[nbPoints] : NULL;
      for (int i=0;i<nbPoints;i++) point[i] = P[i];
    }
    
    ~MDB_Quad() {if (point) delete [] point;}

    virtual bool       complete()     const {return isComplete;}
    virtual int        getOrder()     const {return order;}
    virtual int        getNbEdges ()  const {return 4;}
    virtual int        getNbNodes ()  const {return 4;}
    virtual MDB_Edge * getEdge  (int i) const 
    {
      switch (i){
      case 0 : return e1;
      case 1 : return e2;
      case 2 : return e3;
      case 3 : return e4;
      }
      throw;
    }
    virtual MDB_Edge * getOpposingEdge(const MDB_Edge* orig)  const { 
      if (orig == e1) return e3;
      if (orig == e3) return e1;
      if (orig == e2) return e4;
      if (orig == e4) return e2;
      return NULL;
    }
    
    int alignP1(MDB_Point*,MDB_Point*,MDB_Point*,MDB_Point*);
    int align(MDB_Point*,MDB_Point*,MDB_Point*,MDB_Point*);
    MDB_Edge * find_edge(const MDB_Point*, const MDB_Point*) const;
    int getNbHighOrderPoints() const {return nbPoints;}
    MDB_Point* getHighOrderPoint(int i) {return (i<nbPoints) ?  point[i] : NULL;}
    virtual int        getMshTag()    const {
      if (isComplete) {
        switch (order) {
            case 1 : return MSH_QUA_4;
            case 2 : return MSH_QUA_9;
            case 3 : return MSH_QUA_16;
            case 4 : return MSH_QUA_25;
            default: return MSH_UNDEFINED_ELEM;
        }
      }
      else {
        switch (order) {
          case 1 : return MSH_QUA_4;
          case 2 : return MSH_QUA_8;
          case 3 : return MSH_QUA_12;
          default: return MSH_UNDEFINED_ELEM;
        }
      }
    }

  protected:
    int order;
    int nbPoints;
    MDB_Point** point;
    bool isComplete;

  };

  /*! \brief base class for three-dimensional element */
 
  class MDB_Region: public MDB_MeshEntity
  {
  public:
    virtual MDB_Face *getFace (int i) const = 0;
    virtual MDB_Edge *getEdge (int i) const = 0;
    virtual MDB_Point *getVertex (int i) const = 0;
    virtual MDB_Point *getHighOrderPoint (int i) {throw;}
    virtual int getNbHighOrderPoints() const {return 0;}
    virtual void      getNodes(MDB_Point **n) const = 0;
    virtual int       getNbFace () const = 0;
    virtual int       getNbEdge () const = 0;
    virtual int       getNbVertex () const = 0;
    virtual int       getFaceDir (int) const = 0;
    virtual int       getFaceOrientation(int) const {throw;}
    virtual MDB_Face* getOpposingFace(MDB_Face*) const {return NULL;}
            MDB_Face* getOpposingFace(MDB_Face* orig,int& orientation) const {
      MDB_Face* dest = getOpposingFace(orig);
      if (dest)
  orientation = getOpposingFaceOrientation(orig,dest);
      return dest;
    }
    virtual bool getParametricCoordinates(const double[3],double[3]) const {return false;}
    virtual void getParametricControlPoints(double[][3]) const {throw;}
    virtual bool contains(const double[3],double) const { return false;}
    
    virtual bool setClassification(pGEntity pg) {
      if (!MDB_MeshEntity::setClassification(pg)) return false;
      for (int i=0;i<getNbFace();i++) getFace(i)->setClassification(pg);
      for (int i=0;i<getNbHighOrderPoints();i++) getHighOrderPoint(i)->setClassification(pg);
      return true;
    }

  private:
    
    int getOpposingFaceOrientation(MDB_Face* f1,MDB_Face* f2) const;
  };

  /*! \brief Generic order equidistant Lagrange hexahedron with complete interpolation (internal nodes) */
  /*! \warning only implemented up to order 2 for Serendipity Element and up to order 4 for Complete Hex*/
  class MDB_Hex: public MDB_Region
  {
  public:
    MDB_Quad  *f1,*f2,*f3,*f4,*f5,*f6;
  
    virtual int getOrder() const  {return order;}
    virtual int getDim()   const  {return 3;}
    virtual int getMshTag()const {
      if( isComplete ) {
        switch (getOrder()) {
          case 1:  return MSH_HEX_8;
          case 2:  return MSH_HEX_27;
          case 3:  return MSH_HEX_64;
          case 4:  return MSH_HEX_125;
          default: return MSH_UNDEFINED_ELEM;
        }
      }
      else {
        switch (getOrder()) {
          case 1:  return MSH_HEX_8;
          case 2:  return MSH_HEX_20;
          default: return MSH_UNDEFINED_ELEM;
        }
      }
    }
    virtual int       getNbFace () const {return 6;}
    virtual int       getNbEdge () const {return 12;}
    virtual int       getNbVertex () const{return 8;}
    virtual MDB_Face *getFace (int i) const
    {
      switch(i){
      case 0:return f1;
      case 1:return f2;
      case 2:return f3;
      case 3:return f4;
      case 4:return f5;
      case 5:return f6;
      }
      throw;
    }
    virtual MDB_Edge *getEdge (int n) const
    {
      switch (n){
      case 0  : return f1->commonedge ( f2 );
      case 1  : return f1->commonedge ( f3 );
      case 2  : return f1->commonedge ( f4 );
      case 3  : return f1->commonedge ( f5 );
      case 4  : return f2->commonedge ( f5 );
      case 5  : return f2->commonedge ( f3 );    
      case 6  : return f3->commonedge ( f4 );    
      case 7  : return f4->commonedge ( f5 );    
      case 8  : return f2->commonedge ( f6 );    
      case 9  : return f3->commonedge ( f6 );    
      case 10 : return f4->commonedge ( f6 );    
      case 11 : return f5->commonedge ( f6 );    
      }
      throw;    
    }

    virtual MDB_Point *getVertex (int n) const /*SL OK*/
    {
      MDB_Edge *e1=0,*e2=0; 
      switch (n){
      case 0 : e1 = getEdge(0);e2 = getEdge(3);break;
      case 1 : e1 = getEdge(0);e2 = getEdge(1);break;
      case 2 : e1 = getEdge(1);e2 = getEdge(2);break;
      case 3 : e1 = getEdge(2);e2 = getEdge(3);break;
      case 4 : e1 = getEdge(4);e2 = getEdge(8);break;
      case 5 : e1 = getEdge(5);e2 = getEdge(9);break;
      case 6 : e1 = getEdge(6);e2 = getEdge(10);break;
      case 7 : e1 = getEdge(7);e2 = getEdge(11);break;
      default: throw;
      }

      return e1->commonvertex ( e2 );
    }

    virtual int getFaceDir (int n) const;


    virtual void getNodes(MDB_Point **n) const
    {
      MDB_Point *o[4];
      f1->getNodes(n);
      f2->getNodes(o);
      for(int i=0;i<4;i++) n[i+4]=o[i];
    }

    virtual MDB_Face* getOpposingFace(MDB_Face* orig) const{
      if (orig == f1) return f6;
      if (orig == f6) return f1;
      if (orig == f2) return f4;
      if (orig == f4) return f2;
      if (orig == f3) return f5;
      if (orig == f5) return f3;
      return NULL;
    }
    
    MDB_Hex(MDB_Quad *A, MDB_Quad *B, MDB_Quad *C, MDB_Quad *D, MDB_Quad *E,  MDB_Quad *F);
    MDB_Hex(MDB_Quad *A,MDB_Quad *B,MDB_Quad *C,MDB_Quad *D,MDB_Quad *E,MDB_Quad *F,
                    int o,MDB_Point** pp);
    
    virtual ~MDB_Hex() {if (point) delete [] point;}
    
    
    MDB_Point *getHighOrderPoint (int i) { return ((i < nbPoint) ? point[i] : NULL );}
    int getNbHighOrderPoints() const {return nbPoint;}
    
  protected:
  
    int order;
    bool isComplete;
    static int gmshPoint4[27];
    int nbPoint;
    MDB_Point** point;

  };
  
  /*! \brief Generic order equidistant Lagrange tetrahedron */ 

  class MDB_Tet: public MDB_Region
  {
  public:
    MDB_Triangle  *f1,*f2,*f3,*f4;
    
    MDB_Tet(MDB_Triangle *A, MDB_Triangle *B, MDB_Triangle *C, MDB_Triangle *D)
      : f1(A),f2(B),f3(C),f4(D), nbPoints(0)
    {
      f1->addregion(this);
      f2->addregion(this);
      f3->addregion(this);
      f4->addregion(this);
      
      order = std::numeric_limits<int>::max();
      order = std::min(order,f1->getOrder());
      order = std::min(order,f2->getOrder());
      order = std::min(order,f3->getOrder());
      order = std::min(order,f4->getOrder());
      
      isComplete = (order < 3 ) ? true : false;
      point = NULL;
      
    }
    
    MDB_Tet(MDB_Triangle *A, MDB_Triangle *B, MDB_Triangle *C, MDB_Triangle *D,int o,MDB_Point** pp):
      f1(A),f2(B),f3(C),f4(D), nbPoints(std::max(0,(o-3)*(o-2)*(o-1)/6))
    {
      f1->addregion(this);
      f2->addregion(this);
      f3->addregion(this);
      f4->addregion(this);
      
      order = std::numeric_limits<int>::max();
      order = std::min(order,f1->getOrder());
      order = std::min(order,f2->getOrder());
      order = std::min(order,f3->getOrder());
      order = std::min(order,f4->getOrder());

      if (order != o) {
        MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                    "Incoherent order %d with faces %d not yet implemented",
                                    o,order);
      }
      if (order > 5) {
        MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                    "Complete tetrahedron of order %d not yet implemented",
                                    order);
      }
      
      isComplete = (order < 3 || nbPoints > 0) ? true : false;
      point = nbPoints ? new MDB_Point*[nbPoints] : NULL;
      for (int i=0;i<nbPoints;i++) point[i] = pp[i];
    }
    
    
    virtual ~MDB_Tet() {if (point) delete [] point;}
    
    
    virtual int getOrder() const {return order;}
    virtual int getDim()   const {return 3;}
    virtual int getMshTag() const {
      if (isComplete) {
        switch (order) {
        case 1: return MSH_TET_4;
        case 2: return MSH_TET_10;
        case 3: return MSH_TET_20;
        case 4: return MSH_TET_35;
        default: return MSH_UNDEFINED_ELEM;
        }
        return MSH_UNDEFINED_ELEM;
      }
      else
      {
        switch (order) {
        case 1: return MSH_TET_4;
        case 2: return MSH_TET_10;
        case 3: return MSH_TET_20;
        case 4: return MSH_TET_34;
        default: return MSH_UNDEFINED_ELEM;
        }
        return MSH_UNDEFINED_ELEM;
      }
      return MSH_UNDEFINED_ELEM;
    }

    virtual int       getNbFace () const {return 4;}
    virtual int       getNbEdge () const {return 6;}
    virtual int       getNbVertex () const{return 4;}
    virtual MDB_Face *getFace (int i) const
    {
      switch(i){
      case 0:return f1;
      case 1:return f2;
      case 2:return f3;
      case 3:return f4;
        // NP default NULL added to avoid call to throw 
        //    in classify_unclassified_entities
      default: return (MDB_Face *)  NULL;
      }
      throw;
    }
    virtual MDB_Edge *getEdge (int n) const
    {
      switch (n){
      case 0 : return f1->commonedge ( f2 );
      case 1 : return f1->commonedge ( f3 );
      case 2 : return f1->commonedge ( f4 );
      case 3 : return f2->commonedge ( f4 );
      case 4 : return f2->commonedge ( f3 );
      case 5 : return f3->commonedge ( f4 );    
      }
      throw;    
    }

    virtual MDB_Point *getVertex (int n) const
    {
      MDB_Edge *e1=0,*e2=0; 
      switch (n){
      case 0 : e1 = getEdge(0);e2 = getEdge(2);break;
      case 1 : e1 = getEdge(0);e2 = getEdge(1);break;
      case 2 : e1 = getEdge(1);e2 = getEdge(2);break;
      case 3 : e1 = getEdge(3);e2 = getEdge(5);break;
      default: throw;
      }
      return e1->commonvertex ( e2 );
    }

    // return 1 if the face direction points to outside of the tet
    // return 0                                 inside
    virtual int getFaceDir (int n) const;
    // return number of times we need to rotate the element to make
    // the principal node coincide with that of the tetrahedron (before normal inversion)
    virtual int getFaceOrientation (int n) const;
    
    using MDB_Region::getOpposingFace;

    virtual void getNodes(MDB_Point **n) const
    {
      MDB_Point *o[3];
      f1->getNodes(n);
      f2->getNodes(o);	  
      n[3] = 0; //for stupid gcc warning
      if(o[0] != n[0] && o[0] != n[1] &&o[0] != n[2])n[3] = o[0];
      if(o[1] != n[0] && o[1] != n[1] &&o[1] != n[2])n[3] = o[1];
      if(o[2] != n[0] && o[2] != n[1] &&o[2] != n[2])n[3] = o[2];
    }
    
    bool getParametricCoordinates(const double xyz[3],double res[3])  const {
      
      double rxyz[4][3];
      for (int i=0;i<4;i++) {
        MDB_Point* pp = getVertex(i);
        rxyz[i][0] = pp->X;
        rxyz[i][1] = pp->Y;
        rxyz[i][2] = pp->Z;
      }
      
      double mat[3][3];
      mat[0][0] = rxyz[1][0] - rxyz[0][0];
      mat[0][1] = rxyz[1][1] - rxyz[0][1];
      mat[0][2] = rxyz[1][2] - rxyz[0][2];
      mat[1][0] = rxyz[2][0] - rxyz[0][0];
      mat[1][1] = rxyz[2][1] - rxyz[0][1];
      mat[1][2] = rxyz[2][2] - rxyz[0][2];
      mat[2][0] = rxyz[3][0] - rxyz[0][0];
      mat[2][1] = rxyz[3][1] - rxyz[0][1];
      mat[2][2] = rxyz[3][2] - rxyz[0][2];
      
      double invMat[3][3];
      inverseMat(mat,invMat);
      
      double vec[3];
      diffVec(xyz,rxyz[0],vec);
      
      vecMat(vec,invMat,res);

      double residu = 1.;
      for (int i=0;i<3;i++) {
        if (res[i] > 1 || res[i] < 0) return false;
        residu -= res[i];
      }
      return (residu >= 0 && residu <= 1);
    }
      
    bool contains(const double xyz[3],double tol) const {
      
      double tmp[4];
      if (getParametricCoordinates(xyz,tmp)) return true;
      tmp[3] = 1. - tmp[0] - tmp[1] - tmp[2];
      
      for (int i=0; i<4; i++) {
        if ( tmp[i] < -tol || tmp[i] > 1.+tol ) return false;
      }
      return true;
    }
  
    
    int getNbHighOrderPoints() const {return nbPoints;}
    MDB_Point* getHighOrderPoint(int i) {return (i<nbPoints) ? point[i] : NULL;}
    
  
    
  protected:

    int order;
    int nbPoints;
    bool isComplete;
    MDB_Point** point;
    static int face_vtx[4][3];
  
  };
  
  /*! \brief Generic order equidistant Lagrange prism */
  
  class MDB_Prism: public MDB_Region
  {
  public:
    MDB_Triangle  *f1,*f2;
    MDB_Quad      *f3,*f4,*f5;
  
    virtual int getOrder() const {return order;}
    virtual int getDim()   const {return 3;} /*SL OK*/
    virtual int getMshTag()const {
      if (isComplete) {
        switch (getOrder()) {
            case 1: return MSH_PRI_6;
            case 2: return MSH_PRI_18;
            default: return MSH_UNDEFINED_ELEM;
        }
      }
      else {
        switch(getOrder()) {
          case 1: return MSH_PRI_6;
          case 2: return MSH_PRI_15;
          default: return MSH_UNDEFINED_ELEM;
        }
      }
    }
    
    virtual int       getNbFace () const {return 5;} /*SL OK*/ 
    virtual int       getNbEdge () const {return 9;} /*SL OK*/
    virtual int       getNbVertex () const{return 6;} /*SL OK*/
    virtual MDB_Face *getFace (int i) const /*SL OK*/
    {
      switch(i){
      case 0:return f1;
      case 1:return f2;
      case 2:return f3;
      case 3:return f4;
      case 4:return f5;
      case 5:return (MDB_Face *) NULL;
      }
      throw;
    }
    virtual MDB_Edge *getEdge (int n) const /*SL OK*/
    {
      switch (n){
      case 0  : return f1->commonedge ( f3 );
      case 1  : return f1->commonedge ( f5 );
      case 2  : return f1->commonedge ( f4 );
      case 3  : return f2->commonedge ( f3 );
      case 4  : return f2->commonedge ( f5 );
      case 5  : return f2->commonedge ( f4 );    
      case 6  : return f3->commonedge ( f5 );    
      case 7  : return f3->commonedge ( f4 );    
      case 8  : return f4->commonedge ( f5 );       
      }
      throw;    
    }
  
    virtual MDB_Point *getVertex (int n) const /*SL OK*/
    {
      MDB_Edge *e1=0,*e2=0; 
      switch (n){
      case 0 : e1 = getEdge(0);e2 = getEdge(1);break;
      case 1 : e1 = getEdge(0);e2 = getEdge(2);break;
      case 2 : e1 = getEdge(1);e2 = getEdge(2);break;
      case 3 : e1 = getEdge(4);e2 = getEdge(3);break;
      case 4 : e1 = getEdge(3);e2 = getEdge(5);break;
      case 5 : e1 = getEdge(4);e2 = getEdge(5);break;
      default: throw;
      }
      return e1->commonvertex ( e2 );
    }

    virtual int getFaceDir (int n) const
    { std::cerr << "not yet implemented\n"; abort(); return 0;}


    virtual void getNodes(MDB_Point **n) const
    {
      std::cerr << "not yet implemented\n"; abort();
      /*    MDB_Point *o[3];
            f1->getNodes(n);
            f2->getNodes(o);    
            n[3] = 0; //for stupid gcc warning
            if(o[0] != n[0] && o[0] != n[1] &&o[0] != n[2])n[3] = o[0];
            if(o[1] != n[0] && o[1] != n[1] &&o[1] != n[2])n[3] = o[1];
            if(o[2] != n[0] && o[2] != n[1] &&o[2] != n[2])n[3] = o[2];*/
    }


    virtual MDB_Face* getOpposingFace(MDB_Face* orig) const {
      if (orig == f1) return f2;
      if (orig == f2) return f1;
      return NULL;
    }
  
    MDB_Prism(MDB_Triangle *A, MDB_Triangle *B, MDB_Quad *C, 
              MDB_Quad *D, MDB_Quad *E) /*SL OK*/
    : f1(A),f2(B),f3(C),f4(D),f5(E), nbPoints(0), point(NULL)
    { 
      f1->addregion(this);
      f2->addregion(this);
      f3->addregion(this);
      f4->addregion(this);
      f5->addregion(this);
      
      order = std::numeric_limits<int>::max();
      order = std::min(order,f1->getOrder());
      order = std::min(order,f2->getOrder());
      order = std::min(order,f3->getOrder());
      order = std::min(order,f4->getOrder());
      order = std::min(order,f5->getOrder());
      
      isComplete = order < 2 ? true : false;
    }
    
    MDB_Prism(MDB_Triangle *A, MDB_Triangle *B, MDB_Quad *C, 
              MDB_Quad *D, MDB_Quad *E,int o,MDB_Point** ho)
    : f1(A),f2(B),f3(C),f4(D),f5(E)
    { 
      f1->addregion(this);
      f2->addregion(this);
      f3->addregion(this);
      f4->addregion(this);
      f5->addregion(this);
      
      bool faceComplete = f1->complete() && 
                          f2->complete() && 
                          f3->complete() && 
                          f4->complete() && 
                          f5->complete();

      order = std::numeric_limits<int>::max();
      order = std::min(order,f1->getOrder());
      order = std::min(order,f2->getOrder());
      order = std::min(order,f3->getOrder());
      order = std::min(order,f4->getOrder());
      order = std::min(order,f5->getOrder());
      
      if (order != o) { 
        MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                    "Incoherent order of prism wrt faces (min) (%i vs %i)", o,order);
      }
      
      nbPoints = std::max(0,(order+1)*(order+1)*(order+2)/2 
                  - 6                           // principal vertices 
                  - 9 * (order-1)               // edge vertices
                  - 3 * (order-1)*(order-1)     // quad vertices
                  - 2 * (order-1)*(order-2)/2); // triangle vertices
      
      point = nbPoints >  0 ? new MDB_Point*[nbPoints]:NULL;
      isComplete = faceComplete && (order < 3 || nbPoints > 0);
      for (int i=0;i<nbPoints;i++) point[i] = ho[i];
    }
    
    virtual ~MDB_Prism() {
      if (nbPoints > 0) {
        for (int i=0;i<nbPoints;i++) delete point[i];
        delete [] point;
      }
    }
    
    
    virtual int getNbHighOrderPoints() const {return nbPoints;}
    virtual MDB_Point* getHighOrderPoint(int i) {return (i<nbPoints) ? point[i] : NULL;}
    
  protected:
    
    int order;
    int nbPoints;
    bool isComplete;
    MDB_Point** point;

  };
  
  /*! \brief Generic order equidistant Lagrange pyramid */
  
  class MDB_Pyramid: public MDB_Region
  {
  public:
    
    MDB_Quad      *quad;
    MDB_Triangle  *tri1,*tri2,*tri3,*tri4;
  
    virtual int getOrder() const {return order;}
    virtual int getDim()   const {return 3;} 
    virtual int getMshTag() const {
      if (isComplete) 
      {
        switch(order) {
          case 1: return MSH_PYR_5;
          case 2: return MSH_PYR_14;
          default: return MSH_UNDEFINED_ELEM;
        }
      }
      else 
      {
        switch( order ) {
          case 1 : return MSH_PYR_5;
          case 2 : return MSH_PYR_13;
          default: return MSH_UNDEFINED_ELEM;
        }
      }
    }

    virtual int       getNbFace () const {return 5;}
    virtual int       getNbEdge () const {return 8;}
    virtual int       getNbVertex () const{return 5;}
    
    virtual MDB_Face *getFace (int i) const
    {
      switch(i){
      case 0:return quad;
      case 1:return tri1;
      case 2:return tri2;
      case 3:return tri3;
      case 4:return tri4;
      case 5:return (MDB_Face *) NULL;
      }
      throw;
    }
    virtual MDB_Edge *getEdge (int n) const
    {
      switch (n){
      case 0  : return quad->commonedge ( tri1 );
      case 1  : return quad->commonedge ( tri2 );
      case 2  : return tri1->commonedge ( tri2 );
      case 3  : return quad->commonedge ( tri3 );
      case 4  : return tri3->commonedge ( tri1 );
      case 5  : return quad->commonedge ( tri4 );    
      case 6  : return tri4->commonedge ( tri3 );   
      case 7  : return tri4->commonedge ( tri2 );
      }
      throw;    
    }
  
    virtual MDB_Point *getVertex (int n) const /*SL OK*/
    {
      MDB_Edge *e1=0,*e2=0; 
      switch (n){
      case 0 : e1 = getEdge(0);e2 = getEdge(1);break;
      case 1 : e1 = getEdge(0);e2 = getEdge(3);break;
      case 2 : e1 = getEdge(3);e2 = getEdge(5);break;
      case 3 : e1 = getEdge(1);e2 = getEdge(5);break;
      case 4 : e1 = getEdge(2);e2 = getEdge(4);break;
      default: throw;
      }
      return e1->commonvertex ( e2 );
    }

    virtual int getFaceDir (int n) const
    { std::cerr << "not yet implemented\n"; abort(); return 0;}


    virtual void getNodes(MDB_Point **n) const
    {
      std::cerr << "not yet implemented\n"; abort();
    }
    /// -- KH OK
    MDB_Pyramid(MDB_Quad* Q,MDB_Triangle *A,MDB_Triangle *B,MDB_Triangle *C,MDB_Triangle *D)
    : quad(Q),tri1(A),tri2(B),tri3(C),tri4(D), nbHOPoints(0), hoPoint(NULL)
    { 
      quad->addregion(this);
      tri1->addregion(this);
      tri2->addregion(this);
      tri3->addregion(this);
      tri4->addregion(this);
      
      order = std::numeric_limits<int>::max();
      order = std::min(order,tri1->getOrder());
      order = std::min(order,tri2->getOrder());
      order = std::min(order,tri3->getOrder());
      order = std::min(order,tri4->getOrder());
      order = std::min(order,quad->getOrder());
      
      isComplete = (order < 2) ? true : false;
    }
    
    MDB_Pyramid(MDB_Quad* Q,MDB_Triangle *A,MDB_Triangle *B,MDB_Triangle *C,
                        MDB_Triangle *D,int o,MDB_Point** pp)
    : quad(Q),tri1(A),tri2(B),tri3(C),tri4(D)
    {
      quad->addregion(this);
      tri1->addregion(this);
      tri2->addregion(this);
      tri3->addregion(this);
      tri4->addregion(this);
      
      bool faceComplete = tri1->complete() && 
                          tri2->complete() &&
                          tri3->complete() && 
                          tri4->complete() && 
                          quad->complete();

      order = std::numeric_limits<int>::max();
      order = std::min(order,tri1->getOrder());
      order = std::min(order,tri2->getOrder());
      order = std::min(order,tri3->getOrder());
      order = std::min(order,tri4->getOrder());
      order = std::min(order,quad->getOrder());
      
      if (order != o) {
        MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                    "Incoherent order of pyramid wrt faces (min) (%i vs %i)", o,order);
      }
      
      nbHOPoints = getNbPoints(order-3);
      
      if (nbHOPoints > 0) {
        hoPoint = new MDB_Point*[nbHOPoints];
        std::memcpy(pp,hoPoint,nbHOPoints*sizeof(MDB_Point*));
      }

      isComplete = faceComplete && (order < 3 || nbHOPoints > 0); 
    }
    
    using MDB_Region::getOpposingFace;

    virtual ~MDB_Pyramid() {
      if (nbHOPoints > 0) {
        for (int i=0;i<nbHOPoints;i++) delete hoPoint[i];
        delete [] hoPoint;
      }
    } 
    
    
    virtual int getNbHighOrderPoints() const {return nbHOPoints;}
    virtual MDB_Point* getHighOrderPoint(int i) {return (i<nbHOPoints) ? hoPoint[i] : NULL;}
    
  protected:
    int getNbPoints(int o) const {
      int nbPoints = 0;
      for (int i=0;i<=o;i++) nbPoints += (i+1)*(i+1);
      return nbPoints;
    }
    
  protected:  
    int order;
    int nbHOPoints;
    bool isComplete;
    MDB_Point** hoPoint;    
  };


  class PointLessThan
  {
  public:
    bool operator()(const MDB_Point* ent1, const MDB_Point* ent2) const
    {
      return *ent1 < *ent2;
    }
  };

  class MDB_Mesh : public MDB_MiniMesh
  {        
    // no copies allowed
    MDB_Mesh(const MDB_Mesh &other);

    int getRank() const {return rank;}
    
  private:
    bool parametric;
#ifdef PARALLEL
    MPI_Comm  mpi_communicator;
#endif
    int rank;

    // bounding boxes
    double Min[3],Max[3];
    bool boundingBox;
    double scale;

  private: 
    
    bool meshRead;
    std::string name;
    std::string meshFileName;

  public:
    
    bool setMeshFileName(const char* fname) {
      if (meshFileName.size() != 0) return false;
      meshFileName = fname;
      return true;
    }

    bool setBoundingBox(const double bbox[3][2]) {
      if (boundingBox) return false;
      boundingBox = true;
      for (int i=0;i<3;i++) {
        Min[i] = bbox[i][0];
        Max[i] = bbox[i][1];
      }
      return true;
    }
    
  public:

    // this is an add on to the classical interface
    // one can build lists of geometrical "features" that
    // encompass more than one geometrical entities
    std::multimap<int, pGEntity > geomFeatures_Tags;
    // sometimes, features have names
    std::map<std::string, int >         geomFeatures_Names;

    bool shrinked;

    // constructor of an empty mesh
#ifdef PARALLEL
    // In PARALLEL, the default value for the MPI communicator is set to 
    // MPI_COMM_WORLD. When MPI_COMM_WORLD is used, the mesh is assumed 
    // distributed and must exist on all processes. The MPI_COMM_SELF 
    // communicator can be used to create a local mesh on a specific 
    // process in PARALLEL.
    MDB_Mesh(MPI_Comm mpi_comm = MPI_COMM_WORLD);
#else
    MDB_Mesh();
#endif
    // the destructor
    ~MDB_Mesh();
    // initialize first vertex id and increments
    void initializeIdData();
    // load the mini mesh into the binary file f, mesh is still empty
    bool isParametric() const { return parametric; }
    void load   ( FILE *f );
    // expand the mini mesh in the bi directional data structure
    // then, delete the mini mesh. Mesh adaptation can be performed
    // only when the mesh is expanded 
    void expand ( );
    // shrink the mini mesh from the bi-directional data structure
    // then, delete the bi-directional data structure. The mesh
    // is "idle", i.e. is shrinked in memory, allowing the solver 
    // to take the whole memory space
    void shrink ( );
    bool isShrinked() const { return shrinked; };  
    // gives the bounding box of the mesh
    void getBoundingBox(double[3][2]);
    // gives the bounding box of the transformed mesh
    void getBoundingBox(double[3][2], const pTransform& pt);
    // gives the scale of the mesh
    double getScale() const {return scale;}
    // set the scale of the mesh
    void   setScale(double s);
    // flush the mini mesh
    void flush ( FILE *f ) const;

    bool setName(const std::string& n) {
      if (name.size() != 0) return false;
      name = n;
      return true;
    }
    
    const std::string& getName() const {return name;}
    
    // data about vertex ids (globally unique ids in parallel)
    int maxId;       // max vertex id
    int idIncrement; // id increment size (=nbProcs)
    // the geometric model on which the mesh entities are classified
    pGModel model; 
    
    // name 
    // the datas
    int nbPoints, nbEdgePoints, nbEdges, nbTriangles, nbQuads, nbTets, nbHexes, nbPrisms, nbPyramids, nbPartitions;
    std::set<MDB_Point*,PointLessThan>     points; 
    std::set<MDB_Point*,PointLessThan>     ho_points; 
    MDB_ListE edges; 
    MDB_ListF triangles; 
    MDB_ListQ quads; 
    MDB_ListT tets; 
    MDB_ListH hexes; 
    MDB_ListPr prisms; 
    MDB_ListPy pyramids;
    // Points operations
    MDB_Point * add_point(int num , double x, double y,double z, pGEntity=0);
    MDB_PointParam * add_pointParam(int num , double x, double y, double z, 
                                    double u, double v, pGEntity=0);
    void del_point(MDB_Point *p);  
    MDB_Point *find_point(int num);
    void deleteTheHighOrderPoint(MDB_Point *p);
    // Edges operations
    /*! straight edge from point id's */
    MDB_Edge  * add_edge(int p1, int p2, pGEntity=0);
    //  MDB_Edge  * add_edge(int p1, int p2, int p3, pGEntity=0);
    /*! straight edge from points */ 
    MDB_Edge  * add_edge(MDB_Point *p1, MDB_Point *p2, pGEntity=0);
    //  MDB_Edge  * add_edge(MDB_Point *p1, MDB_Point *p2, MDB_Point *p3, pGEntity=0);
    /*! straight/curved edge from point id's */
    MDB_Edge  * add_edge(int numberOfPoints, pGEntity pg, int p1, ...);
    /*! straight/curved edge from points */
    MDB_Edge  * add_edge(int numberOfPoints, pGEntity pg, MDB_Point *firstPoint, ...);
    /*! straight/curved edge from points (gmsh order) */
    MDB_Edge  * add_edge(MDB_Point* p1,MDB_Point*p2,pGEntity,int o,MDB_Point** iP);
    void del_edge(MDB_Edge *e);
    MDB_Edge  *find_edge(int p1, int p2);
    MDB_Edge  *find_edge(MDB_Point *p1, MDB_Point *p2)const;
    MDB_Edge  *find_clone_edge(MDB_Edge *edge)const;
    // Faces operations
    void del_face(MDB_Face *f);
    // Triangles operations
    /*! straight triangle from vertex id's */ 
    MDB_Triangle *add_triangle(int p1, int p2, int p3, pGEntity=0); 
    /*! straight triangle from vertices */ 
    MDB_Triangle *add_triangle(MDB_Point* v0, MDB_Point* v1, MDB_Point* v2, pGEntity=0); 
    /* curvilinear triangle */
    //  MDB_Triangle *add_triangle(int p1, int p2, int p3, int p4, int p5, int p6, pGEntity=0);
    /*! straight/serendipity triangle from edges */
    MDB_Triangle *add_triangle(MDB_Edge *e1,MDB_Edge *e2,MDB_Edge *e3, pGEntity=0);
    /*! straight/complete triangle triangle from edges and internal points */
    MDB_Triangle *add_triangle(MDB_Edge *e1,MDB_Edge *e2,MDB_Edge *e3, pGEntity,int order,
                               bool serendip=true,MDB_Point** iP=NULL);
    /*! straight/complete triangle triangle from point id's*/
    //MDB_Triangle *add_triangle(pGEntity,int order,bool serendip=true,int* pp=NULL);
    /*! straight/serendipity triangle from point id's */ 
    MDB_Triangle *add_triangle(int order, bool complete, pGEntity pg, int p1, int p2, ...);
    void del_triangle(MDB_Triangle *t);
    MDB_Triangle  *find_triangle(MDB_Edge *e1,MDB_Edge *e2,MDB_Edge *e3);
    // Quad operations
    /*! straight quadrangle from vertex id's */ 
    MDB_Quad *add_quad(int p1, int p2, int p3, int p4, pGEntity=0); 
    /*! straight quadrangle from vertices */ 
    MDB_Quad *add_quad(MDB_Point* v0, MDB_Point* v1, MDB_Point* v2, MDB_Point* v3, pGEntity=0); 
    MDB_Quad *add_quad(MDB_Edge *e1,MDB_Edge *e2,MDB_Edge *e3,MDB_Edge *e4, pGEntity=0);
    /*! straight/serendipity quad from point id's */ 
    MDB_Quad *add_quad(int order, bool complete, pGEntity pg, int p1, int p2, int p3, ...);
    /*! complete high order quad from points id's */
    MDB_Quad *add_quad(pGEntity pg,int order,bool serendip,int* p);
    /*! straight/complete quad from edges and internal points */
    MDB_Quad *add_quad(MDB_Edge *e1,MDB_Edge *e2,MDB_Edge *e3,MDB_Edge *e4, pGEntity,int order,
                               bool serendip=true,MDB_Point** iP=NULL);
    void del_quad(MDB_Quad *q);
    MDB_Quad  *find_quad(MDB_Edge *e1,MDB_Edge *e2,MDB_Edge *e3,MDB_Edge *e4);
    // Tets operations
    /*! straight tet from point id's */
    MDB_Tet *add_tet(int p1, int p2, int p3,int p4, pGEntity=0);
    /*! straight/serendipity/complete triangle from point id's (in gmsh order) */
    MDB_Tet *add_tet(pGEntity,int order,bool serendip,int* points);
    /*! straight tet from points */ 
    MDB_Tet *add_tet(MDB_Point *p1, MDB_Point *p2,MDB_Point *p3, MDB_Point *p4, pGEntity=0);
    /*! straight/serendipity tetrahedron from faces */
    MDB_Tet *add_tet(MDB_Triangle  *,MDB_Triangle  *,MDB_Triangle  *,MDB_Triangle  *, pGEntity=0);
    /*! straight/complete tetrahedron from faces and points */
    MDB_Tet *add_tet(MDB_Triangle  *,MDB_Triangle  *,MDB_Triangle  *,MDB_Triangle  *, pGEntity,
                     int order,bool serendip=true,MDB_Point** iP=NULL);
  
    /*! straight/complete tet from faces and internal point id's */
    MDB_Tet *find_tet(MDB_Triangle * t1, MDB_Triangle * t2, MDB_Triangle * t3,  MDB_Triangle * t4);
    void del_tet(MDB_Tet *);
    // Hexes operations
    MDB_Hex *add_hex(int p1, int p2, int p3,int p4,int p5, int p6, int p7, int p8, pGEntity=0);
    MDB_Hex *add_hex(MDB_Point *p1, MDB_Point *p2,MDB_Point *p3, MDB_Point *p4,
                     MDB_Point *p5, MDB_Point *p6,MDB_Point *p7, MDB_Point *p8, pGEntity=0);
    MDB_Hex *add_hex(MDB_Quad  *,MDB_Quad  *,MDB_Quad  *,MDB_Quad  *,MDB_Quad  *,MDB_Quad  *, pGEntity=0);
    MDB_Hex *add_hex(int order,MDB_Quad  *,MDB_Quad  *,MDB_Quad  *,MDB_Quad  *,MDB_Quad  *,MDB_Quad  *, int*, pGEntity=0);
    MDB_Hex *add_hex(pGEntity,int order,int* pp, bool serendip = false);
    MDB_Hex *find_hex(MDB_Quad  *,MDB_Quad  *,MDB_Quad  *,MDB_Quad  *,MDB_Quad  *,MDB_Quad  *);
    void del_hex(MDB_Hex *);
    // Prisms operations
    MDB_Prism *add_prism(int p1, int p2, int p3,int p4,int p5, int p6, pGEntity=0);
    MDB_Prism *add_prism(MDB_Point *p1, MDB_Point *p2,MDB_Point *p3, 
                         MDB_Point *p4, MDB_Point *p5, MDB_Point *p6, pGEntity=0);
    MDB_Prism *add_prism(pGEntity,int order,bool serendip,int* points);
    MDB_Prism *add_prism(MDB_Triangle  *,MDB_Triangle  *,MDB_Quad  *,MDB_Quad  *,MDB_Quad  *, pGEntity=0);
    MDB_Prism *find_prism(MDB_Triangle  *,MDB_Triangle  *,MDB_Quad  *,MDB_Quad  *,MDB_Quad  *);
    void del_prism(MDB_Prism *);
    // Pyramids
    MDB_Pyramid* add_pyramid(pGEntity pg,int order, bool serendip,int* points);
    void del_pyramid(MDB_Pyramid * pyra);
    
    /*! \brief classify entities that are not classified by inheritance */
    void classify_unclassified_entities();
#ifndef _HAVE_GMSH_
    /*! \brief complete the model such that a coherent BREP is found */
    /*! \caveat we assume that the boundary elements of dim-1 are correctly tagged */
    void completeNullModel();
    
    /*! \brief complete the model closures using mesh entity classifications */
    /*! \warning For parallel computations, the necessary exchange required
        for reconstructing the full topology is yet to be implemented */
    void updateModelClosures();
#endif
    void destroyStandAloneEntities();
    void destroyConnectionData();

#ifdef PARALLEL
    // return the mpi commicator used in mesh constructor. This function
    // should be used to set the MPI communicator in MPI instructions.
    MPI_Comm& getMPIComm( );
#endif
    
    // check if there is no identical geometrical tags between regions and faces
    /*   void checkGeomTagsCoherence () const; */
  };

  typedef std::set<MDB_Point*,PointLessThan> MDB_SetV;

  class MDB_RIter : public MDB_Iter < MDB_ListR , MDB_Region , pGEntity>
  {
  public:
    MDB_RIter (MDB_ListR*_l): MDB_Iter< MDB_ListR , MDB_Region , pGEntity > (_l){}
    MDB_RIter (MDB_ListR*_l,pGEntity _g): MDB_Iter< MDB_ListR , MDB_Region , pGEntity > (_l,_g){}
  };
  class MDB_TIter : public MDB_Iter < MDB_ListT , MDB_Tet , pGEntity>
  {
  public:
    MDB_TIter (MDB_ListT*_l): MDB_Iter< MDB_ListT , MDB_Tet , pGEntity > (_l){}
    MDB_TIter (MDB_ListT*_l,pGEntity _g): MDB_Iter< MDB_ListT , MDB_Tet , pGEntity > (_l,_g){}
  };
  class MDB_HIter : public MDB_Iter < MDB_ListH , MDB_Hex , pGEntity>
  {
  public:
    MDB_HIter (MDB_ListH*_l): MDB_Iter< MDB_ListH , MDB_Hex , pGEntity > (_l){}
    MDB_HIter (MDB_ListH*_l,pGEntity _g): MDB_Iter< MDB_ListH , MDB_Hex , pGEntity > (_l,_g){}
  };
  class MDB_PIter : public MDB_Iter < MDB_ListPr , MDB_Prism , pGEntity>
  {
  public:
    MDB_PIter (MDB_ListPr*_l): MDB_Iter< MDB_ListPr , MDB_Prism , pGEntity > (_l){}
    MDB_PIter (MDB_ListPr*_l,pGEntity _g): MDB_Iter< MDB_ListPr , MDB_Prism , pGEntity > (_l,_g){}
  };
  class MDB_PyIter : public MDB_Iter < MDB_ListPy , MDB_Pyramid , pGEntity>
  {
  public:
    MDB_PyIter (MDB_ListPy*_l): MDB_Iter< MDB_ListPy ,MDB_Pyramid , pGEntity > (_l){}
    MDB_PyIter (MDB_ListPy*_l,pGEntity _g): MDB_Iter< MDB_ListPy , MDB_Pyramid , pGEntity > (_l,_g){}
  };
  class MDB_EIter : public MDB_Iter < MDB_ListE , MDB_Edge , pGEntity >
  {
  public:
    MDB_EIter (MDB_ListE*_l): MDB_Iter< MDB_ListE , MDB_Edge, pGEntity > (_l){}
    MDB_EIter (MDB_ListE*_l,pGEntity _g,int _c): MDB_Iter< MDB_ListE , MDB_Edge, pGEntity > (_l,_g,_c){}
  };
  class MDB_FIter : public MDB_Iter < MDB_ListF , MDB_Triangle , pGEntity >
  {
  public:
    MDB_FIter (MDB_ListF*_l): MDB_Iter <MDB_ListF , MDB_Triangle ,pGEntity> (_l){}
    MDB_FIter (MDB_ListF*_l,pGEntity _g,int _c): MDB_Iter <MDB_ListF , MDB_Triangle ,pGEntity> (_l,_g,_c){}
  };
  class MDB_QIter : public MDB_Iter < MDB_ListQ , MDB_Quad , pGEntity>
  {
  public:
    MDB_QIter (MDB_ListQ*_l): MDB_Iter <MDB_ListQ , MDB_Quad , pGEntity> (_l){}
    MDB_QIter (MDB_ListQ*_l,pGEntity _g,int _c): MDB_Iter <MDB_ListQ , MDB_Quad ,pGEntity> (_l,_g,_c){}
  };
  class MDB_VIter : public MDB_Iter < MDB_SetV , MDB_Point ,pGEntity >
  {
  public:
    MDB_VIter (MDB_SetV *_l): MDB_Iter < MDB_SetV , MDB_Point , pGEntity > (_l){}
    MDB_VIter (MDB_SetV *_l,pGEntity _g,int _c): MDB_Iter < MDB_SetV , MDB_Point , pGEntity > (_l,_g,_c){}
  };

  class MDB_FaceIter
  {
  public:
    MDB_FIter itf;
    MDB_QIter itq;

    MDB_Triangle *t;
    MDB_Quad *q;
  public:
    // MDB_FaceIter ():t(NULL),q(NULL) {}
    MDB_FaceIter ( MDB_ListF* _t , MDB_ListQ* _q):itf(_t),itq(_q),t(0),q(0){}
    MDB_FaceIter ( MDB_ListF* _t , MDB_ListQ* _q, pGEntity _g,int _c):itf(_t,_g,_c),itq(_q,_g,_c),t(0),q(0){}
    inline MDB_Face * next ()
    {
      t = itf.next();
      if (t) return (MDB_Face*)t;
      q = itq.next();
      return (MDB_Face*)q;
    }
    inline void reset()
    {
      itf.reset();
      itq.reset();
    }
    inline void cleanup()
    {
      itf.cleanup();
      itq.cleanup();
    }

  };

  class MDB_RegionIter
  {
  public:
    MDB_TIter itt;
    MDB_HIter ith;
    MDB_PIter itp;
    MDB_PyIter itpy;

    MDB_Tet *t;
    MDB_Hex *h;
    MDB_Prism *p;
    MDB_Pyramid *py;
  public:
    MDB_RegionIter ( MDB_ListT* _t , MDB_ListH* _h, MDB_ListPr* _p, MDB_ListPy* _py ):itt(_t),ith(_h),itp(_p),itpy(_py),t(0),h(0),p(0),py(0){}
    MDB_RegionIter ( MDB_ListT* _t , MDB_ListH* _h, MDB_ListPr* _p, MDB_ListPy* _py, pGEntity _g):itt(_t,_g),ith(_h,_g),itp(_p,_g),itpy(_py,_g),t(0),h(0),p(0),py(0){}
    inline MDB_Region * next ()
    {
      t = itt.next();
      if (t) return (MDB_Region*)t;
      h = ith.next();
      if (h) return (MDB_Region*)h;
      p = itp.next();
      if (p) return (MDB_Region*)p;
      py = itpy.next();
      return (MDB_Region*)py;
      
    }
    inline void reset()
    {
      itt.reset();
      ith.reset();
      itp.reset();
      itpy.reset();
    }
    inline void cleanup()
    {
      itt.cleanup();
      ith.cleanup();
      itp.cleanup();
      itpy.cleanup();
    }

  };

} // End of namespace MAd



#endif






