// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------


#include "MeshDataBase.h"
#include "MeshDataBaseInterface.h"
#include "MeshDataBaseParallelInterface.h"
#include "MeshDataBaseComm.h"
#include "MeshDataBaseExtrusion.h"
#include "MeshDataBaseExtrusionInterface.h"

#include <limits>
#include <map>
#include <vector>
#include <cstdlib>
#include <string>

#ifdef PARALLEL
#include "mpi.h"
#include "autopack.h"
#endif

namespace MAd {
    
  // ---------------------------------------------------------------------------
  
  typedef std::map< ExtrusionData* , pFace > ExtrudedRegionSeedMap;

  // ---------------------------------------------------------------------------
  
  typedef std::map< pFace , ExtrusionData* > ExtrudedRegionSendMap;
  
  // ---------------------------------------------------------------------------

  class MDB_extrudedRegionDataExchanger : public MDB_DataExchanger {

  public:
    
    MDB_extrudedRegionDataExchanger(int t,
                                    const ExtrudedRegionSendMap& s1,
                                    ExtrudedRegionSeedMap& s2):
      MDB_DataExchanger(t),send(s1),seed(s2) {}
    ~MDB_extrudedRegionDataExchanger() {;}

  public:
    
    // -------------------------------------------------------------------------
    using MDB_DataExchanger::sendData; 
    void* sendData(pEntity pe,int iProcDest,int& sizebuf) {
      void* buf = NULL;
      sizebuf = 0;
      ExtrudedRegionSendMap::const_iterator sIter = send.find((pFace) pe);
      if (sIter!=send.end()) {
        sizebuf = sizeof(ExtrusionData);
        buf = malloc(sizebuf);
        ExtrusionData* data = (ExtrusionData*) buf;
        *(data) = *(sIter->second);
      }
      return buf;
    }
    
    // -------------------------------------------------------------------------
    
    void receiveData(pEntity pe,int iProcSend,void* buf) {
      pFace pf = (pFace) pe;
      const ExtrusionData* other = reinterpret_cast<const ExtrusionData*> (buf);
      ExtrusionData* data = new ExtrusionData(*other);
      seed[data] = pf;
    }
    
  protected:
    
    const ExtrudedRegionSendMap& send;
          ExtrudedRegionSeedMap& seed;
  };
  
  // ---------------------------------------------------------------------------

  class FinaliseExtrusionDataExchanger : public MDB_ExtrusionDataExchanger {
    
  public:
    
    FinaliseExtrusionDataExchanger(pMeshDataId te,int tc):
      MDB_ExtrusionDataExchanger(tc),tagExt(te) {}
    ~FinaliseExtrusionDataExchanger() {}
    
  public:
    
    void* sendData(pEntity pf,pEntity pr,int iProc,int index,int& dataSize) {
      dataSize = 5*sizeof(int);
      void* buf = malloc(dataSize);
      int* ibuf = reinterpret_cast<int*>(buf);
      *(ibuf++) = EN_mshBasicTag(pf);
      for (int i=0;i<F_numVertices((pFace) pf);i++) {
        *(ibuf++) = EN_id((pEntity) F_vertex((pFace) pf,i));
      }
      if (F_numVertices((pFace) pf) == 3) (*ibuf) = -1;
      return buf;
    }
    
    void receiveData(pEntity pf,int iProc,int index,int count,void* buf) {
      int* ibuf = reinterpret_cast<int*>(buf);
      int type = *ibuf;
      if (EN_mshBasicTag((pEntity) pf) != type) {
        MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                    "Incompatible face type %d (%d,%d,%d,%d) with seed of type %d (%d,%d,%d,%d)",
                                    type,
                                    ibuf[1],ibuf[2],ibuf[3],ibuf[4],
                                    EN_mshTag(pf),
                                    EN_id(F_vertex((pFace) pf,0)),
                                    EN_id(F_vertex((pFace) pf,1)),
                                    EN_id(F_vertex((pFace) pf,2)),
                                    F_numVertices ((pFace) pf) == 4 ? 
                                    EN_id(F_vertex((pFace) pf,3)):-1);
      }
      void* tmpptr(NULL);
      if (EN_getDataPtr((pEntity) pf,tagExt,&tmpptr)) {
        MAd_extrusionData* ed = reinterpret_cast<MAd_extrusionData*>(tmpptr);
        ed->layerIdx++;
        eltCount[pf] = ed->layerIdx;
      }
    }
  public:

    pMeshDataId tagExt;
    std::map<pEntity,int> eltCount;
    
  };

  // ---------------------------------------------------------------------------
  
  pMeshDataId M_flagExtrudedRegions(pMesh pm,
                                    const std::set<pGFace>& beginfaces,
                                    const std::set<pGFace>& finalfaces,
                                    const std::string& name,
                                    int depth=std::numeric_limits<int>::max()) {

    std::string extrName = "extrusion " + name;
    pMeshDataId tagExt = MD_getMeshDataId(extrName.c_str());
    if ((int) tagExt!=-1) return tagExt;

    tagExt = MD_newMeshDataId(extrName.c_str());
    
    int myRank = 0;

#ifdef PARALLEL
    MPI_Comm_rank(pm->getMPIComm(),&myRank);
#endif

    ExtrudedRegionSeedMap seed;

    // --- detect all faces on my partition 
    
    for (std::set<pGFace>::const_iterator gIter=beginfaces.begin();gIter!=beginfaces.end();++gIter) {
      pGFace pgf = *gIter;
      FIter fIter = M_classifiedFaceIter(pm,(pGEntity) pgf,0);
      while (pFace pf = FIter_next(fIter)) {
        ExtrusionData* data = new ExtrusionData;
        data->seedElmt    = pf;
        data->seedProc    = myRank;
        data->layerIdx    = 0;
        seed[data]        = pf;
      }
      FIter_delete(fIter);
    }
    
    // --- flag all faces on my partition
    for (ExtrudedRegionSeedMap::iterator sIter = seed.begin();sIter!=seed.end();++sIter) {
      void* data = (void*) new MAd_extrusionData(*(sIter->first),NULL);
      EN_attachDataPtr((pEntity) sIter->second,tagExt,data);
    }

    // --- propagate 
    
#ifdef PARALLEL
    int totalNbSend = 0;
#endif
    
    do {

#ifdef PARALLEL
      ExtrudedRegionSendMap send;
#endif
      
      for (ExtrudedRegionSeedMap::iterator sIter = seed.begin();sIter!=seed.end();++sIter) {
        
        ExtrusionData* ed = sIter->first;
        
        pFace   pf = sIter->second;
        pRegion pr = F_region(pf,0); // correct : starting out at boundary of extrusion region       

        int count = 0;
        
        do {
          
          int newOrientation = 0;
          if ((pf = R_gtOppFace(pr,pf,newOrientation))) {
            ed->layerIdx++;
            count++;
            EN_attachDataPtr((pEntity) pf,tagExt,(void*) new MAd_extrusionData(*ed,pr));
            pr = F_otherRegion(pf,pr);
            if (finalfaces.find((pGFace)EN_whatIn((pEntity) pf)) == finalfaces.end()) sIter->second = pf;
          }
          else pr = NULL;

        } while (pf && pr && count<depth);
        

#ifdef PARALLEL
        if (pf) {
          if (F_isInterface(pf) && finalfaces.find((pGFace)EN_whatIn((pEntity) pf)) == finalfaces.end()) send[pf] = new ExtrusionData(*ed);
        }
#endif 
        delete sIter->first;
      }
      seed.clear();
      
#ifdef PARALLEL
      int nbSend = send.size();
      totalNbSend = 0;
      
      MPI_Allreduce(&nbSend,&totalNbSend,1,MPI_INTEGER,MPI_SUM,M_getMPIComm(pm));
      
      if (totalNbSend) {
        int tagCom = 1003;
        MDB_extrudedRegionDataExchanger de(tagCom,send,seed);
        exchangeDataOnFaces(pm,de);
        send.clear();
      }
    } 
    while (totalNbSend != 0);
#else
    } while (!seed.empty());
#endif
  
    int tagCom = 1006;
    FinaliseExtrusionDataExchanger de(tagExt,tagCom);
    reduceDataOnExtrudedFace(pm,tagExt,de);

    return tagExt;
  }

  // ---------------------------------------------------------------------------
  
  pMeshDataId M_flagExtrudedRegions(pMesh pm,
                                    const std::set<pGFace>& beginfaces,
                                    const std::string& name,
                                    const MDB_criterionExtrudedRegion& criterion) {

    std::string extrName = "extrusion " + name;
    pMeshDataId tagExt = MD_getMeshDataId(extrName.c_str());
    if ((int) tagExt!=-1) return tagExt;

    tagExt = MD_newMeshDataId(extrName.c_str());
    
    int myRank = 0;

#ifdef PARALLEL
    MPI_Comm_rank(pm->getMPIComm(),&myRank);
#endif

    ExtrudedRegionSeedMap seed;

    // --- detect all faces on my partition 
    
    for (std::set<pGFace>::const_iterator gIter=beginfaces.begin();gIter!=beginfaces.end();++gIter) {
      pGFace pgf = *gIter;
      FIter fIter = M_classifiedFaceIter(pm,(pGEntity) pgf,0);
      while (pFace pf = FIter_next(fIter)) { 
        ExtrusionData* data = new ExtrusionData;
        data->seedElmt    = pf;
        data->seedProc    = myRank;
        data->layerIdx    = 0;
        seed[data]        = pf;
      }
      FIter_delete(fIter);
    }
    
    // --- propagate 
    
#ifdef PARALLEL
    int totalNbSend = 0;
#endif
    
    do {

#ifdef PARALLEL
      ExtrudedRegionSendMap send;
#endif
      
      for (ExtrudedRegionSeedMap::iterator sIter = seed.begin();sIter!=seed.end();++sIter) {
        
        ExtrusionData* ed = sIter->first;
        
        pFace   pf  = sIter->second;
        pRegion pr  = F_region(pf,0); // correct : starting out at boundary of extrusion region       
        if(isExtruded(pr,pf)){         

          // flag all seed faces 
          // -------------------
          void* data = (void*) new MAd_extrusionData(*(sIter->first),NULL);
          EN_attachDataPtr((pEntity) sIter->second,tagExt,data);
          
          int count = 0;
          do {
            
            int newOrientation = 0;
            if ((pf = R_gtOppFace(pr,pf,newOrientation))){

              pr = F_otherRegion(pf,pr);            

              // flag if anisotropy criterion for next element is fullfilled,
              // or if we're at the last face of the current partition    
              // -----------------------------------------------------------            
              if(criterion(pf,pr,tagExt) || !pr){
                ed->layerIdx++;
                count++;  
                EN_attachDataPtr((pEntity) pf,tagExt,(void*) new MAd_extrusionData(*ed,pr));
                sIter->second = pf;
              }
              else{pr=NULL; pf=NULL;}
            }
            else{pr=NULL;}

          } while (pf && pr);
        
          
#ifdef PARALLEL
          if (pf) {
            if (F_isInterface(pf)) send[pf] = new ExtrusionData(*ed);
          }
#endif 
        }        
        delete sIter->first;
      }
      seed.clear();
      
#ifdef PARALLEL
      int nbSend = send.size();
      totalNbSend = 0;
      
      MPI_Allreduce(&nbSend,&totalNbSend,1,MPI_INTEGER,MPI_SUM,M_getMPIComm(pm));
      
      if (totalNbSend) {
        int tagCom = 1003;
        MDB_extrudedRegionDataExchanger de(tagCom,send,seed);
        exchangeDataOnFaces(pm,de);
        send.clear();
      }
    } 
    while (totalNbSend != 0);
#else
    } while (!seed.empty());
#endif
  
    int tagCom = 1006;
    FinaliseExtrusionDataExchanger de(tagExt,tagCom);
    reduceDataOnExtrudedFace(pm,tagExt,de);

    return tagExt;
  }

  /*! get extrusion data */

  bool EN_getExtrusionData(pEntity pe,
                           pMeshDataId tag,
                           pEntity& seedElmt,
                           int&     seedProc,
                           pEntity& neighbour,
                           int& index) {
    
    void* buf = NULL;
    if (EN_getDataPtr((pEntity) pe,tag,&buf)) {
      MAd_extrusionData* data = reinterpret_cast<MAd_extrusionData*> (buf);
      seedElmt  = data->seedElmt;
      seedProc  = data->seedProc;
      neighbour = data->neighbour;
      index     = data->layerIdx;
      return true;
    }
    return false;
  }

  /*! reduce data on the extruded face */
  
  void reduceDataOnExtrudedFace (pMesh m, pMeshDataId tagData, MDB_ExtrusionDataExchanger &de )
  { 
    
    int mySize = 1;
    int myRank = 0;

#ifdef PARALLEL
    MPI_Comm_size(m->getMPIComm(), &mySize);
    MPI_Comm_rank(m->getMPIComm(), &myRank);
#endif    
  
    int *sendcounts = new int[mySize];
    for(int i=0;i<mySize;i++)sendcounts[i]=0;
    
    FIter fit = M_faceIter(m);
    pFace pf;
    while ( (pf = FIter_next(fit)) ) {

      pEntity remote(NULL);
      int iProc(-1);
      pEntity elmt(NULL);
      int index(-1);
      
      if (EN_getExtrusionData(pf,tagData,remote,iProc,elmt,index)) {
        
        if (elmt != NULL) { // non-trivial extrusion face

          int sizebuf = 0;
          void *buf = de.sendData (pf,elmt,iProc,index,sizebuf);
          
#ifdef PARALLEL
          size_t sendSize = sizebuf + sizeof(pEntity) + sizeof(int);
#endif
          
          if (buf != NULL) {

            if (iProc == myRank) {
              
              pEntity seed(NULL);
              int proc(-1);
              pEntity neighbour(NULL);
              int nbLayers(-1);
              
              if (EN_getExtrusionData(remote,tagData,seed,proc,neighbour,nbLayers)) {
                if (seed != remote) std::cout << "Not a seed face " << std::endl;
                if (proc != myRank) std::cout << "Seed partition wrong" << std::endl;
                de.receiveData ((pEntity) remote,myRank,index,nbLayers,buf);
              }
            }
            else {
              
#ifdef PARALLEL
              char *msg = (char*)AP_alloc(iProc,de.tag(),sendSize);
              memcpy(&msg[0],&remote,sizeof(pEntity));
              memcpy(&msg[sizeof(pEntity)],&index,sizeof(int));
              memcpy(&msg[sizeof(pEntity)+sizeof(int)],buf,sizebuf);
              AP_send(msg);
              sendcounts[iProc] ++;
#else
              throw;
#endif
            }
            free(buf);
          }
        }
      }
    }
    FIter_delete(fit);

#ifdef PARALLEL 
    AP_check_sends(AP_NOFLAGS);
    AP_reduce_nsends(sendcounts);
  
    int message=0;
    int count=0;
  
    while (!AP_recv_count(&count) || message<count) 
      {
        void *msg;
        int from;
        int tag;
        int size;
        int rc;
        rc=AP_recv(MPI_ANY_SOURCE, de.tag(), AP_BLOCKING|AP_DROPOUT,
                   &msg, &size, &from, &tag);
        if (rc) 
          {
            message++;
            char * tmp = (char *) msg;
            //pFace pf;
            int index;
            memcpy(&pf,&tmp[0],sizeof(pEntity));
            memcpy(&index,&tmp[sizeof(pEntity)],sizeof(int));
            
            pEntity remote;
            int iProc;
            pEntity neighbour;
            int nbLayers;
            
            if (EN_getExtrusionData(pf,tagData,remote,iProc,neighbour,nbLayers)) {
              de.receiveData ((pEntity) pf,from,index,nbLayers,&tmp[sizeof(pEntity)+sizeof(int)]);
            }
            AP_free(msg);
          }
      }    
    AP_check_sends(AP_WAITALL); 
    MPI_Barrier(m->getMPIComm());
#endif
    delete [] sendcounts;
  }
  
  // ---------------------------------------------------------------------------
  
  pMeshDataId M_flagExtrudedRegions(pMesh pm,
                                    const std::set<int>& startTags,
                                    const std::set<int>& finalTags,
                                    const std::string& name,
                                    int depth=std::numeric_limits<int>::max()) {

    pGModel pgm = M_model(pm);

    std::set<pGFace> startFaces;
    std::set<pGFace> finalFaces;
    
    GFIter gfIter = GM_faceIter(pgm);
    while (pGFace pgf = GFIter_next(gfIter)) {
      int tag = GEN_tag((pGEntity) pgf);
      if (startTags.find(tag) != startTags.end()) startFaces.insert(pgf);
      if (finalTags.find(tag) != finalTags.end()) finalFaces.insert(pgf);
    }
    GFIter_delete(gfIter);
    
    return M_flagExtrudedRegions(pm,startFaces,finalFaces,name,depth);
  }

  // ---------------------------------------------------------------------------
  
  pMeshDataId M_flagExtrudedRegions(pMesh pm,
                                    const std::set<int>& startTags,
                                    const std::string& name,
                                    const MDB_criterionExtrudedRegion& criterion) {
    pGModel pgm = M_model(pm);

    std::set<pGFace> startFaces;
    
    GFIter gfIter = GM_faceIter(pgm);
    while (pGFace pgf = GFIter_next(gfIter)) {
      int tag = GEN_tag((pGEntity) pgf);
      if (startTags.find(tag) != startTags.end()) startFaces.insert(pgf);
    }
    GFIter_delete(gfIter);
    
    return M_flagExtrudedRegions(pm,startFaces,name,criterion);
  }

  // ---------------------------------------------------------------------------

  bool isExtruded(pRegion pr, pFace pf){

    int basicTag = EN_mshBasicTag((pEntity) pr);

    switch(basicTag){
    case MSH_QUA_4: return true; 
    case MSH_HEX_8: return true; 
    case MSH_PRI_6:              
      { 
        if(pf == R_face(pr,0)) return true;
        if(pf == R_face(pr,1)) return true; 
      }
    }    
    return false;
  }; 

  // ---------------------------------------------------------------------------

  bool MDB_criterionExtrudedRegion::operator()(pFace pf, pRegion pr, pMeshDataId faceTag) const{ 
   
    if(!pf || !pr) return false;
 
    if(!isExtruded(pr,pf)) return false;
 
    // Check that no other face than pf was tagged
    // -> prevent intersecting extrusion lines
    // -------------------------------------------
    
    void *temp_ptr = NULL;
    void *fIter    = NULL;

    pPList faces = R_faces(pr);
    while (pFace f = (pFace) PList_next(faces,&fIter)) {
          
      if(f!=pf && EN_getDataPtr(f,faceTag,&temp_ptr)){
        PList_delete(faces); 
        return false;
      }  
  
    }
    PList_delete(faces);     

    // Otherwise stop in isotropic region 
    // ----------------------------------
 
    double V = R_volume(pr);
    double A = F_area(pf);
    double height = V/A; 

    double minSide = std::numeric_limits<double>::max();

    void  *eIter = NULL;
    pPList edges = F_edges(pf);
    while (pEdge e = (pEdge) PList_next(edges,&eIter)) {
      minSide = std::min(minSide,E_length(e));
    }
    PList_delete(edges);     
        
    return alpha*height < minSide;    
  };


};
