// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------


#include "MAdExport.h"
#include "MeshDataBaseInterface.h"
#include "MeshDataBaseParallelInterface.h"

#include <set>
#include <string>

/*! \defgroup extrusion Extrusion data \ingroup parallel */

namespace MAd {    
  
  // ---------------------------------------------------------------------------
  /*! \brief tag extruded regions from faces belonging to geometrical classifications 
    \ingroup extrusion */
  /*! argument list
    
     - pMesh: mesh
     - set of start faces
     - set of final faces
     - string : name of the extrusion tag (pMeshDataId)
     - depth specifies how far to go 
     
  */
  // ---------------------------------------------------------------------------
  
  MAD_EXPORT pMeshDataId M_flagExtrudedRegions(pMesh,
                                    const std::set<pGFace>&,
                                    const std::set<pGFace>&,
                                    const std::string&,
                                    int);
  
  
  // ---------------------------------------------------------------------------
  /*! \brief tag extruded regions from faces belonging to geometrical classifications 
    \ingroup extrusion */
  /*! argument list
    
     - pMesh: mesh
     - set of start face tags 
     - set of final face tags
     - string : name of the extrusion tag (pMeshDataId)
     - depth specifies how far to go 
     
  */
  // ---------------------------------------------------------------------------
  

  MAD_EXPORT pMeshDataId M_flagExtrudedRegions(pMesh,
                                    const std::set<int>&,
                                    const std::set<int>&,
                                    const std::string&,
                                    int);

  // ---------------------------------------------------------------------------
  /*! \brief get extrusion data from d-1 entities \ingroup extrusion */
  /*! input: 
    <ul>
    <li> pEntity : copy from extrusion base
    <li> pMeshDataId: mesh tag corresponding to this extrusion
    </ul>
    output
    <ul>
    <li> pEntity : extrusion base / seed
    <li> int     : seed processor
    <li> pEntity : d-dimensional neighbour below current entity
    <li> int     : layer index
    </ul>
  */
  // ---------------------------------------------------------------------------
  
  MAD_EXPORT bool EN_getExtrusionData(pEntity,pMeshDataId,pEntity&,int&,pEntity&,int&);
  
  // ---------------------------------------------------------------------------
  /*! \brief data exchanger for operations on extruded regions */
  // ---------------------------------------------------------------------------
  
  class MAD_EXPORT MDB_ExtrusionDataExchanger {
    
  public:
    
    MDB_ExtrusionDataExchanger(int t):tagCom(t) {}
    virtual ~MDB_ExtrusionDataExchanger() {}
    
  public:
    
    int tag() const {return tagCom;}
    //! my face and region, proc, index, orientation, size of message
    virtual void* sendData   (pEntity,pEntity,int,int,int&) = 0;
    //! extrusion base, sending proc, index of extruded face, total count, data
    virtual void  receiveData(pEntity,int,int,int,void*) = 0;
    
    virtual void  deleteExternalData(pEntity pe) {};

  protected:
    
    int tagCom;
    
  };
  

  // ---------------------------------------------------------------------------
  /*! \brief Reduces data from all faces to the starting face of the extrusion zone  
    \ingroup extrusion */
  // ---------------------------------------------------------------------------

  MAD_EXPORT void reduceDataOnExtrudedFace (pMesh,pMeshDataId,MDB_ExtrusionDataExchanger&);
  
  // ---------------------------------------------------------------------------
  /*! \brief Broadcast data from the starting face to all faces of the extrusion zone  
   \ingroup extrusion */
  // ---------------------------------------------------------------------------

  MAD_EXPORT void broadcastDataFromExtrudedFace(pMesh,pMeshDataId,MDB_ExtrusionDataExchanger&);

  // ---------------------------------------------------------------------------
  /*! \brief criterion to determine end of boundary-layer extruded regions    */
  // ---------------------------------------------------------------------------

  class MAD_EXPORT MDB_criterionExtrudedRegion {

  public:
    
    MDB_criterionExtrudedRegion(const double a=2):alpha(a){};
    virtual ~MDB_criterionExtrudedRegion(){;}
    virtual bool operator()(pFace, pRegion, pMeshDataId) const; 

  private:
    
    double alpha;    
  };

  // ---------------------------------------------------------------------------
  /*! \brief tag extruded regions from faces belonging to geometrical classifications 
    \ingroup extrusion */
  /*! argument list
    
     - pMesh: mesh
     - set of start face tags 
     - string : name of the extrusion tag (pMeshDataId)
     - criterion : functor determining end of extruded region
  */
  // ---------------------------------------------------------------------------

  MAD_EXPORT pMeshDataId M_flagExtrudedRegions(pMesh,
                                    const std::set<pGFace>&,
                                    const std::string&,
                                    const MDB_criterionExtrudedRegion&);


  MAD_EXPORT pMeshDataId M_flagExtrudedRegions(pMesh,
                                    const std::set<int>&,
                                    const std::string&,
                                    const MDB_criterionExtrudedRegion&);

};
