// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#include "MeshDataBase.h"
#include "MeshDataBaseInterface.h"
#include "MeshDataBaseIO.h"
#include "MeshDataBaseGEntity2Physical.h"
#include "MshTags.h"
#include "MAdMessage.h"
#include "MAdResourceManager.h"
#include "CheckMesh.h"

#include <cstdlib>
#include <cstring>
#include <set>
#include <map>
#include <utility>
#include <cassert>
#include <string>
#include <sstream>

#include <float.h>

#ifdef PARALLEL
#include "mpi.h"
#include "autopack.h"
#endif

#include "MeshDataBaseParallelInterface.h"
#include "MAdMatrix.h"
#include "MAdTransformation.h"
#include "NullModel.h"


namespace MAd {

  // ---------------------------------------------------------------------------
  // seek a specific section in the file
  // ---------------------------------------------------------------------------

  bool findSection(FILE* fp,
                    const std::string& name,
                    bool rewindFirst=true) {
    
    char line[256];
    if (rewindFirst) rewind(fp);
    while(!feof(fp)) {
      do {
        if(!fgets(line, sizeof(line), fp))
          break;
        if(feof(fp))
          break;
      } while(line[0] != '$'); 
      std::string lineString(line);
      if (!lineString.compare(1,name.size(),name)) return true;
    } 

    MAdMsgSgl::instance().info(__LINE__,__FILE__,"Could not find section %s",name.c_str());
    
    return false;
  }
  
  // ---------------------------------------------------------------------------
  // read and write baseline periodicities 
  // ---------------------------------------------------------------------------
  
  void SaveGmshPeriodicities(pMesh pm,FILE* fp,bool distributed) {

    pGModel pgm = M_model(pm);
  
    int myRank(0);
    
#ifdef PARALLEL
    MPI_Comm_rank(pm->getMPIComm(),&myRank);
#endif

    if (myRank == 0 || distributed) {
      fprintf(fp,"$Periodicities\n");
      fprintf(fp,"%d\n",GM_numPeriodicities(pgm));
      
      double mat[16];
      GPIter gpIter = GM_periodicityIter(pgm,false);
      
      while (pTransform tfo = GPIter_next(gpIter)) {
        T_copyAffineTransformation(tfo,mat);
        fprintf(fp,"Affine");
        for (int i=0;i<16;i++) fprintf(fp," %.*g",DBL_DIG,mat[i]);
        fprintf(fp,"\n");
      }
      fprintf(fp,"$EndPeriodicities\n");
      GPIter_delete(gpIter);
    }
    
    if (!distributed) {
      unsigned long position = ftell(fp);
#ifdef PARALLEL
      MPI_Bcast(&position,1,MPI_LONG_INT,0,pm->getMPIComm());
#endif 
      fseek(fp,position,0);
    }
  }
  
  bool ReadGmshPeriodicities(FILE* fp,TransformationSet& tfoSet) {

    if (findSection(fp,"Periodicities")) {
      
      char line[256];
      fgets(line,sizeof(line),fp);
      
      int nbPeriodicities;
      std::istringstream lineStream(line);
      lineStream >> nbPeriodicities;
      
      for (int iPeriod=0;iPeriod<nbPeriodicities;iPeriod++) {
        
        // --- read transformation

        fgets(line,sizeof(line),fp);
        std::istringstream tfoStream(line);
        std::string tfoType;   
        tfoStream >> tfoType;
        pTransform tfo(NULL);

        if (!tfoType.compare(0,6,"Affine")) {
          doubleMatrix mat(4,4);
          for (int i=0;i<4;i++) for (int j=0;j<4;j++) tfoStream >> mat(i,j);
          tfo = new MAdTransformation(mat);
        }

        if (tfo) {
          if (tfoSet.find(tfo) == tfoSet.end()) tfoSet.insert(tfo);
          else delete tfo;
        }
      }
      
      fgets(line,sizeof(line),fp);
      
      std::string checkLine(line);
      if (checkLine.compare(0,17,"$EndPeriodicities")) throw;
      return true;
    }
    
    return false;
  }
  
  // ---------------------------------------------------------------------------
  // read and write bounding boxes 
  // ---------------------------------------------------------------------------

  void SaveGmshBoundingBox(pMesh pm,FILE* fp,bool distributed) {
    
    double bb[3][2];
    M_box(pm,bb);
    double myBB[6] = {bb[0][0],bb[0][1],
                      bb[1][0],bb[1][1],
                      bb[2][0],bb[2][1]};
    
    int nbProc(1);
    int myRank(0);
    
#ifdef PARALLEL
    MPI_Comm_size(pm->getMPIComm(),&nbProc);
    MPI_Comm_rank(pm->getMPIComm(),&myRank);
    double* allBB = (myRank == 0)? new double[6*nbProc] : NULL;
    MPI_Gather(myBB,6,MPI_DOUBLE,allBB,6,MPI_DOUBLE,0,pm->getMPIComm());
#else
    double* allBB = myBB;
#endif
    
    if (myRank == 0 || distributed) {
      fprintf(fp,"$BoundingBoxes\n");
      fprintf(fp,"%d\n",distributed ? 1:nbProc);

      int minProc = distributed ? myRank:0;
      int maxProc = distributed ? myRank:nbProc;

      double* bbp = distributed ? myBB : allBB;

      for (int iProc=minProc;iProc<maxProc;iProc++,bbp+=6) {
        fprintf(fp,"%d",iProc);
        for (int i=0;i<6;i++) fprintf(fp," %.*g",DBL_DIG,bbp[i]);
        fprintf(fp,"\n");   
      }
      fprintf(fp,"$EndBoundingBoxes\n");
    }
    
    if (!distributed) {
      unsigned long position = ftell(fp);
#ifdef PARALLEL
      MPI_Bcast(&position,1,MPI_LONG_INT,0,pm->getMPIComm());
#endif 
      fseek(fp,position,0);
    }
    if (allBB != myBB) delete [] allBB;
  }
      

  bool ReadGmshBoundingBox(FILE* fp,double bb[3][2],int myRank=1) {
    
    if (findSection(fp,"BoundingBoxes")) {
      
      char line[256];
      fgets(line,sizeof(line),fp);
      std::istringstream nbLine(line);

      int nbPartitions;
      nbLine >> nbPartitions;
      
      for (int iPart=0;iPart<nbPartitions;iPart++) {
        fgets(line,sizeof(line),fp);
        std::istringstream partString(line);
        int rank = -1;
        partString >> rank;
        if (rank == myRank) {
          for (int i=0;i<3;i++) for (int j=0;j<2;j++) partString >> bb[i][j];
          return true;
        }
      }
    }
    return false;
  }

  
  // -------------------------------------------------------------------
  // ------------- Load a Gmsh mesh ------------------------------------
  // -------------------------------------------------------------------
  
  void LoadGmshMesh (pMesh m,const char *filename,int part,bool switchPhysical)
  {  

    const MAdMsg& msg = MAdMsgSgl::instance();

#ifdef PROFILE
    const MAdResourceManager& mrm = MAdResourceManagerSgl::instance();
    double dt = - mrm.getTime();
#endif
    
    double scale = m->getScale();

    FILE *fp = fopen(filename, "r");
    if(!fp) msg.error(__LINE__,__FILE__,"Unknown File %s",filename);

    int myRank = 0;
#ifdef PARALLEL 
    int nbProc = 1;
    MPI_Comm_size(m->getMPIComm(), &nbProc);
    MPI_Comm_rank(m->getMPIComm(), &myRank);
    int rank = (part==-1)? myRank:part;
#endif

    char String[256];

    if (!m->model) {
      pGModel model = NULL;
      GM_create(&model,"");
      m->model = model; 
    }

    // ----------------------------------------------
    // ------ Reading format
    // ----------------------------------------------

    int version = -1;
    int subversion = 0;
    do {
      if(!fgets(String, sizeof(String),fp))
        break;
      if(feof(fp))
        break;
    } while(String[0] != '$' ||
            strncmp(&String[1], "MeshFormat",10));
  
    if(feof(fp)) {
      version = 1;
#ifdef PROFILE
      if (myRank==0) msg.info(__LINE__,__FILE__,"Reading Gmsh format %d",version);
#endif
    }
    else {

      int vv1 = 0;
      int vv2 = 0;
      
      fgets(String,sizeof(String),fp);
      std::string format(String);
      
      size_t dotPos = format.find_first_of(".");

      if (dotPos == std::string::npos) {
        std::istringstream formatBuf(format);
        formatBuf >> version;
        formatBuf >> vv1;
        formatBuf >> vv2;
        
#ifdef PROFILE
        if (myRank==0) {
          msg.info(__LINE__,__FILE__,"Reading Gmsh format %d %d %d",version,vv1,vv2);
        }
#endif
      }
      else {
        std::istringstream versBuf(format.substr(0,dotPos));
        versBuf >> version;
        std::istringstream subVersBuf(format.substr(dotPos+1,format.size()));
        subVersBuf >> subversion;
        subVersBuf >> vv1;
        subVersBuf >> vv2;
#ifdef PROFILE
        if (myRank==0) {
          msg.info(__LINE__,__FILE__,
                   "Reading Gmsh format %d.%d %d %d",
                   version,subversion,vv1,vv2);
        }
#endif
      }
    }

    rewind(fp);

#ifdef PARALLEL
    if(!switchPhysical) {  // tmp - this test should be removed if OK for Argo
      if ( version == 1 && nbProc > 1 ) {
        if (myRank==0) msg.error(__LINE__,__FILE__,
                                "Cannot load Gmsh mesh '%s': reading not implemented for format msh1 in parallel if all partitions are in a single file",
                                filename);
      }
    }
#endif

    // ----------------------------------------------------------------
    // ------ Reading elements to find local nodes and check dimensions 
    // ----------------------------------------------------------------

    std::set<int> localNodeIds;

    int dim = 0;

    do {
      if(!fgets(String, sizeof(String), fp))
        break;
      if(feof(fp))
        break;
    } while(String[0] != '$' ||
            ( strncmp(&String[1], "ELM",      3) &&
              strncmp(&String[1], "Elements", 8)    ) );

    if(!feof(fp)) {

#ifdef PROFILE 
      double dtNodes = - mrm.getTime();
#endif

      int nbElems_world, NbTags, verts[256], Tag;
      int Num, Type, Physical, Elementary, Nbr_Nodes, Partition;
      fscanf(fp, "%d", &nbElems_world);

      std::set<int> nbParts;
      
      for(int i_Element = 0; i_Element < nbElems_world; i_Element++) {
        
        std::set<int> partitions;
        std::set<int> ghosts;
        if(version == 1){
          fscanf(fp, "%d %d %d %d %d",
                 &Num, &Type, &Physical, &Elementary, &Nbr_Nodes);
          Partition = -1;
        }
        else{
          fscanf(fp, "%d %d %d", &Num, &Type, &NbTags);
          Elementary = Physical = Partition = -1;
          int nbPartitions = 0;
          
          for(int j = 0; j < NbTags; j++){
            fscanf(fp, "%d", &Tag);
            
            if     (j == 0) Physical = Tag;
            else if(j == 1) Elementary = Tag;
            else if(subversion <  2 && j == 2) partitions.insert(Tag-1);
            else if(subversion >= 2 && j == 2 && NbTags > 3) nbPartitions = Tag;
            else if(subversion >= 2 && j == 3) partitions.insert(Tag-1);
            else if(j >= 4 && j < 4 + nbPartitions - 1) ghosts.insert(-Tag);
            
          }
          Nbr_Nodes = getNumVerticesForElementTypeMSH(Type);
        }
        
        nbParts.insert(partitions.begin(),partitions.end());
        
        dim = std::max(dim,getDimForElementTypeMSH(Type));

        for(int j = 0; j < Nbr_Nodes; j++) {
          fscanf(fp, "%d", &verts[j]);
        }

#ifdef PARALLEL
        // only take nodes on my partition 
        // but internal faces do not (necessarily) have parallel classification
        if (partitions.find(rank) != partitions.end() || partitions.size() == 0) 
#endif
          {
            // avoid partition boundary
            if (Elementary >= 0) {
              for(int j = 0; j < Nbr_Nodes; j++) {
                localNodeIds.insert(verts[j]);
              }
            }
          }
      }
#ifdef PROFILE 
      dtNodes += mrm.getTime();
      if (myRank==0) {
        msg.info(__LINE__,__FILE__,"Read elements to find out nodes in %e seconds",dtNodes);
      }
#endif

      m->nbPartitions = (int)(nbParts.size());
    }

    rewind(fp);

    // ----------------------------------------------
    // ------ Reading nodes
    // ----------------------------------------------

    do {
      if(!fgets(String, sizeof(String), fp))
        break;
      if(feof(fp))
        break;
    } while(String[0] != '$' ||
            (strncmp(&String[1], "NOD", 3) &&
             strncmp(&String[1], "NOE", 3) &&
             strncmp(&String[1], "Nodes", 5) &&
             strncmp(&String[1], "ParametricNodes",15) ));
    
    if(!feof(fp)) {
#ifdef PROFILE  
      double dtNodes = -mrm.getTime();
#endif
      
      int nbNodes_world;
      fscanf(fp, "%d", &nbNodes_world);
      for(int i_Node = 0; i_Node < nbNodes_world; i_Node++) {
        int id;
        double x,y,z;
        fscanf(fp, "%d %lf %lf %lf", &id, &x, &y, &z);
        if ( localNodeIds.find(id) != localNodeIds.end() ) {
          m->add_point(id,x*scale,y*scale,z*scale);
        }
      }
#ifdef PROFILE   
      dtNodes += mrm.getTime();
      if (myRank==0) {
        msg.info(__LINE__,__FILE__,"Read nodes in %e seconds",dtNodes);
      }
#endif
      
    }

    rewind(fp);

    // ----------------------------------------------
    // ------ Reading parametric nodes
    // ----------------------------------------------

    if ( version == 2 ) {
      do {
        if(!fgets(String, sizeof(String), fp)) break;
        if(feof(fp)) break;
      } while(String[0] != '$' ||
              strncmp(&String[1], "ParametricNodes", 15)  );
    
      if(!feof(fp)) {
        int nbNodes_world;
        fscanf(fp, "%d", &nbNodes_world);
        for(int i_Node = 0; i_Node < nbNodes_world; i_Node++) {
          int id, gDim, gTag;
          double x,y,z;
          fscanf(fp, "%d %lf %lf %lf %d %d", &id, &x, &y, &z, &gDim, &gTag);
          if ( gDim == 3 ) {
            if ( localNodeIds.find(id) != localNodeIds.end() ) {
              m->add_point(id,x*scale,y*scale,z*scale);
            }
          }
          else {
            double u,v;
            if ( gDim == 0 ) {
              u = v = 0.;
            }
            if ( gDim == 1 ) {
              fscanf(fp, "%lf", &u);
              v = 0.;
            } 
            if ( gDim == 2 ) {
              fscanf(fp, "%lf %lf", &u, &v);
            }
            m->add_pointParam(id,x*scale,y*scale,z*scale,u,v);
          }
        }
      }

      rewind(fp);
    }

    // ----------------------------------------------
    // ------ Reading elements
    // ----------------------------------------------

    std::map<int, std::map<int,int> > dim_elem2Phys;  // Used if switchPhysical=1; used when reading periodicities
    
    do {
      if(!fgets(String, sizeof(String), fp))
        break;
      if(feof(fp))
        break;
    } while(String[0] != '$' ||
            ( strncmp(&String[1], "ELM",      3) &&
              strncmp(&String[1], "Elements", 8)    ) );

    if(!feof(fp)) {

#ifdef PROFILE 
      double dtElements = -mrm.getTime();
#endif

      int nbElems_world, NbTags, verts[256], Tag;
      int Num, Type, Physical, Elementary, Nbr_Nodes, Partition;
      fscanf(fp, "%d", &nbElems_world);


      for(int i_Element = 0; i_Element < nbElems_world; i_Element++) {
        
        std::set<int> partitions;
        std::set<int> ghosts;
      
        if(version == 1){
          fscanf(fp, "%d %d %d %d %d",&Num,&Type,&Physical,&Elementary,&Nbr_Nodes);
          Partition = -1;
        }
        else{
          fscanf(fp, "%d %d %d", &Num, &Type, &NbTags);
          Elementary = Physical = Partition = -1;
          int nbPartitions = -1;
          for(int j = 0; j < NbTags; j++) {
            fscanf(fp, "%d", &Tag);
            
            //KH: code from Gmsh ... svn version 7874 - version 2.2 may not be stable ??
            
            if     (j == 0) Physical = Tag;
            else if(j == 1) Elementary = Tag;
            else if(subversion <  2 && j == 2)               partitions.insert(Tag-1);
            else if(subversion >= 2 && j == 2 && NbTags > 3) nbPartitions = Tag;
            else if(subversion >= 2 && j == 3)               partitions.insert(Tag-1);
            else if(j >= 4 && j < 4 + nbPartitions - 1)      ghosts.insert(-Tag+1);
            
          }
          Nbr_Nodes = getNumVerticesForElementTypeMSH(Type);
        }

        if(switchPhysical) {
          dim_elem2Phys[getDimForElementTypeMSH(Type)][Elementary] = Physical;
        }
              
        for(int j = 0; j < Nbr_Nodes; j++) {
          fscanf(fp, "%d", &verts[j]);
        }

#ifdef PARALLEL 
        // only take elements on my partition 
        // but internal faces do not (necessarily) have parallel classification
        if (partitions.find(rank) != partitions.end() || partitions.size() == 0) {
          // avoid ghost elements from a lower dimension since not full rank 
          if (getDimForElementTypeMSH(Type) == dim || ghosts.size() == 0) {
#endif
            if (Elementary >= 0) { // a negative elementary tag corresponds to a partition boundary
        
              int order = getMappingOrderMSH(Type);
        
              pGEntity geom = 0;
            
              int elemTag = switchPhysical ? Physical : Elementary;
//               int physTag = switchPhysical ? Elementary : Physical;
              

              switch (getDimForElementTypeMSH(Type)) {
              case 0:
                geom = (pGEntity) GM_vertexByTag(m->model,elemTag);
                break;
              case 1:
                geom = (pGEntity) GM_edgeByTag(m->model,elemTag);
                break;
              case 2:
                geom = (pGEntity) GM_faceByTag(m->model,elemTag);
                break;
              case 3:
                geom = (pGEntity) GM_regionByTag(m->model,elemTag);
                break;
    
              }
        
              if (!geom) {
                msg.error(__LINE__,__FILE__,
                          "Cannot find entity %i of dimension %i",
                          elemTag,getDimForElementTypeMSH(Type));
              }
              
              // Even in case of switchPhysical, assign Physical - done like this in Morfeo GmshMesh reader !!
              if ( Physical > 0) GEN_setPhysical(geom,GEN_type(geom),Physical); 
            
              pEntity pe = NULL;

              switch (Type) {
    
                // --- single point --------------------------------------------
    
              case MSH_PNT:
                {
                  MDB_Point *p = m->find_point(verts[0]);
                  p->g = geom;
                  break;
                }
                
                // --- edges ---------------------------------------------------

                // straight edge
              case MSH_LIN_2:
                {
                  pe = m->add_edge(verts[0],verts[1],geom);
                  break;
                }
                // quadratic edge
              case MSH_LIN_3:
                {
                  pe = m->add_edge(3,geom,verts[0],verts[2],verts[1]);
                  break;
                }
                // cubic edge
              case MSH_LIN_4:
                {
                  pe = m->add_edge(4,geom, verts[0],verts[2],verts[3],verts[1]);
                  break;
                }
                // quartic edge
              case MSH_LIN_5:
                {
                  pe = m->add_edge(5,geom, 
                                   verts[0],verts[2],verts[3],
                                   verts[4],verts[1]);
                  break;
                }
                
                // --- triangles -----------------------------------------------

                // straight edged triangle
              case MSH_TRI_3:
                {
                  pe = m->add_triangle(verts[0],verts[1],verts[2],geom); 
                  break;
                }
                // quadratic triangle 
              case MSH_TRI_6:
                {
                  pe = m->add_triangle(2,0,geom,
                                       verts[0],verts[3],verts[1],
                                       verts[4],verts[2],verts[5]); 
                  break;
                }
                // 3rd order serendipity triangle
              case MSH_TRI_9:
                {
                  pe = m->add_triangle(3,0,geom,
                                       verts[0],verts[3],verts[4],
                                       verts[1],verts[5],verts[6],
                                       verts[2],verts[7],verts[8]); 
                  break;
                }
                // 3rd order complete triangle
              case MSH_TRI_10:
                {
                  pe = m->add_triangle(3,1,geom,
                                       verts[0],verts[3],verts[4],verts[1],
                                       verts[5],verts[6],verts[2],verts[7],
                                       verts[8],verts[9]); 
                  break;
                }
                // 4th order serendipity triangle 
              case MSH_TRI_12:
                {
                  pe = m->add_triangle(4,0,geom,
                                       verts[0],verts[3],verts[4],verts[5],
                                       verts[1],verts[6],verts[7],verts[8],
                                       verts[2],verts[9],verts[10],verts[11]); 
                  break;
                }
                
                // 4th order complete triangle (LAGRANGE !)
              case MSH_TRI_15:
                {
                  pe = m->add_triangle(4,1,geom,
                                       verts[0],verts[3],verts[4],
                                       verts[5],verts[1],verts[6],
                                       verts[7],verts[8],verts[2],
                                       verts[9],verts[10],verts[11],
                                       verts[12],verts[13],verts[14]); 
                  break;
                }
                
                // --- quadrangles ---------------------------------------------

                // straight sided quadrangle
              case MSH_QUA_4:
                {
                  pe = m->add_quad(verts[0],verts[1],verts[2],verts[3],geom); 
                  break;
                }
                // high order quadrangles
              case MSH_QUA_9:
              case MSH_QUA_16:
              case MSH_QUA_25:
                {
                  pe = m->add_quad(geom,order,false,verts);
                  break;
                }
                // 8-node serendip quad
              case MSH_QUA_8:
                {
                  pe = m->add_quad(geom, 2, true, verts);
                  break;
                }
                
                // --- tetrahedra ----------------------------------------------
                
                // straight faced tetrahedron
              case MSH_TET_4:
                {
                  pe = m->add_tet(verts[0],verts[1],verts[2],verts[3],geom); 
                  break;
                }
                break;
                // Comlete high-order tetrahedra, 2nd order
              case MSH_TET_10:
              case MSH_TET_20:
              case MSH_TET_35:
                {
                  pe = m->add_tet(geom,order,false,verts);
                  break;
                }
                // 34 node tetrahedron, 4th order incomplete
              case MSH_TET_34:
                {
                  pe = m->add_tet(geom,4,true,verts);
                  break;
                }
                // --- prisms --------------------------------------------------
                
                // straight edged prism 
              case MSH_PRI_6:
                {
                  pe = m->add_prism(verts[0],verts[1],verts[2],
                                    verts[3],verts[4],verts[5],geom); 
                  break;
                }
                // 2nd order incomplete prism
              case MSH_PRI_15:
                {
                  pe = m->add_prism(geom,2,true,verts);
                  break;
                }
                // 2nd order complete prism 
              case MSH_PRI_18:
                {
                  pe = m->add_prism(geom,2,false,verts);
                  break;
                }
                
                // --- hexahedra -----------------------------------------------
                
                // straight-edged hexahedron 
              case MSH_HEX_8:
                {
                  pe = m->add_hex(verts[0],verts[1],verts[2],verts[3],
                                  verts[4],verts[5],verts[6],verts[7],geom); 
                  break;
                }
              case MSH_HEX_20:
                {
                  pe = m->add_hex(geom, 2, verts, true);
                  break;
                }
              case MSH_HEX_27:
              case MSH_HEX_64:
              case MSH_HEX_125:
                {
                  pe = m->add_hex(geom,order,verts);
                  break;
                }
                
                // --- pyramids ------------------------------------------------
              case MSH_PYR_5:
              case MSH_PYR_14:
                {
                  pe = m->add_pyramid(geom,order,false,verts);
                  break;
                }
              case MSH_PYR_13:          
                {
                  pe = m->add_pyramid(geom,order, true,verts);
                  break;
                }
                // --- unknown element type ------------------------------------
                
              default:
                if (myRank==0) {
                    char elemTypeName[256];
                    getElementNameMSH(Type, elemTypeName);
                    msg.error(__LINE__, __FILE__,
                        "Element of type %s (tag %i) is (currently) not supported",
                        elemTypeName, Type);
                }
                throw;
                break; 
              }
              
              
              if (pe) EN_setID(pe,Num);

              bool find = false;
              for (std::multimap<int, pGEntity>::iterator it = 
                     m->geomFeatures_Tags.lower_bound(Elementary);
                   it != m->geomFeatures_Tags.upper_bound(Elementary);++it)
                if (it->second == geom)find = true;
              if (!find)
                m->geomFeatures_Tags.insert(std::pair<int,pGEntity>(Elementary, geom));
            }
#ifdef PARALLEL
          }
        }
#endif
      }
#ifdef PROFILE  
      dtElements += mrm.getTime();
      if (myRank==0) {
        msg.info(__LINE__,__FILE__,"Read elements in %e",dtElements);
      }
#endif
    }
    
    
    if(!switchPhysical) {   // tmp - this test should be removed if OK for Argo
      if ((M_numRegions(m) == 0) && (M_numFaces(m) == 0) && (M_numEdges(m) == 0)) {
        msg.error(__LINE__,__FILE__,"Empty mesh on processor %d",myRank);
      }
    }
    
#ifdef PROFILE 
    double rss;
    double vm;
    mrm.getLocalMemoryUsage(rss,vm);
    dt += mrm.getTime();
    if (myRank==0) {
      msg.info(__LINE__,__FILE__,
               "Reading took %.2f seconds, rss = %.2f MB, vm = %.2f MB",
               dt,rss,vm);
    }
    dt = - mrm.getTime();
#endif

    // transfer classification from parents to children if not yet classified
#ifdef PROFILE  
    double dtClassifications = - mrm.getTime();      
 #endif   
    m->classify_unclassified_entities();
#ifdef PROFILE
    if (myRank==0) {
      dtClassifications += mrm.getTime();
      msg.info(__LINE__,__FILE__,"Classified elements in %2.f seconds",dtClassifications);
    }
#endif
    
#ifdef PROFILE
    dt += mrm.getTime();
    mrm.getLocalMemoryUsage(rss,vm);
    if (myRank==0) {
      msg.info(__LINE__,__FILE__,
               "Updating entity classication took %.2f s, rss = %.2f MB and vm = %.2f MB",
               dt,rss,vm);    
    }
    dt = - mrm.getTime();
#endif

    // remove unused elements - orphan entities are allowed if switchPhysical is activated
    if(!switchPhysical) {
      m->destroyStandAloneEntities();
    }

#ifdef PROFILE   
    dt += mrm.getTime();    
    mrm.getLocalMemoryUsage(rss,vm);
    if (myRank==0) {
      msg.info(__LINE__,__FILE__,
               "Deleting stand alone entities took %.2f s, rss = %.2f MB and vm = %.2f MB",
               dt,rss,vm);
    }
#endif
    
    rewind(fp);


    // ----------------------------------------------
    // ------ Reading periodicity properties
    // ----------------------------------------------

    int periodicInfo = 0;
    
    // KH: added Gmsh periodicity information (periodicInfo == 3)  
    // completion of multiple connections should be expensive -> prefer cenaero encoding if available

    while(periodicInfo == 0 || periodicInfo == 3) {
      do {
        if(!fgets(String, sizeof(String), fp))
          break;
        if(feof(fp))
          break;
      } while(String[0] != '$');
      if(feof(fp))
        break;
      else if(!strncmp(&String[1], "PERIODICNOD", 11)) {
        msg.info(__LINE__,__FILE__,"Format \"PERIODICNOD\" is no longer supported");
        periodicInfo = 1;
      } 
      else if (!strncmp(&String[1],"PeriodicNodes",13)) {
        periodicInfo = 2;
      }
      else if (!strncmp(&String[1],"Periodic",8) && strlen(&String[1]) == 9) {
        periodicInfo = 3;
      }

      do {
        if(!fgets(String, sizeof(String), fp))
          throw;
        if(feof(fp))
          throw;
      } while(String[0] != '$');
    }
    rewind(fp);

    // ----------------------------------------------
    // ------ Reading periodic data
    // ----------------------------------------------

    if (periodicInfo) {
      
      // pMeshDataId tagPeriodicPoint    = MD_lookupMeshDataId("PeriodicPoint");
      pMeshDataId tagPeriodicVertexID = MD_lookupMeshDataId("PeriodicVertexID");

      std::map<int,std::set<int> > connections;

      switch (periodicInfo) {
        
      case 2:
        {
          fseek(fp,0,SEEK_SET);
          while(1) {
            do {
              if(!fgets(String, sizeof(String), fp))
                break;
              if(feof(fp))
                break;
            } while(String[0] != '$');
    
            if(feof(fp))
              break;

            if((!strncmp(&String[1],"PER", 3)) || 
               (!strncmp(&String[1],"PeriodicNodes",13))) {
            
              int nPeriod;
              fscanf(fp, "%d ", &nPeriod);
            
              for(int ip = 0; ip < nPeriod; ip++) {
                
                int Num;
                fscanf(fp, "%d ", &Num);
              
                std::set<int> vtcs;

                for (int i=0;i<Num;i++) {
                  int vtxId;
                  fscanf(fp, "%d ", &vtxId);
                  vtcs.insert(vtxId);
                }
              
                for (std::set<int>::iterator vIter = vtcs.begin();vIter!=vtcs.end();++vIter) {
                  std::map<int,std::set<int> >::iterator cIter = connections.find(*vIter);
                  if (cIter!=connections.end()) {
                    vtcs.insert(cIter->second.begin(),cIter->second.end());
                  }
                }
                  
                for (std::set<int>::iterator vIter = vtcs.begin();vIter!=vtcs.end();++vIter) {
                  connections[*vIter].insert(vtcs.begin(),vtcs.end());
                }
                int trafo;
                fscanf(fp, "%d ",&trafo);
              }
            }
          }          
          break;
        }
      case 3: 
        // periodicity information stored by Gmsh 
        //
        // only pairs are written -> collection routines are needed to detect
        // multiple periodicities
        //
        {
          fseek(fp,0,SEEK_SET);
          while(1) {
            do {
              if(!fgets(String, sizeof(String), fp))
                break;
              if(feof(fp))
                break;
            } while(String[0] != '$');
    
            if(feof(fp))
              break;

            if((!strncmp(&String[1],"Periodic",8))&&(strlen(&String[1]) == 9)) {
              
              int nPeriod;
              fscanf(fp, "%d ", &nPeriod);
              
              for (int iPeriod = 0;iPeriod<nPeriod;iPeriod++) {
                
                int genDim(-1);
                int genID1(-1);
                int genID2(-1);

                fscanf(fp,"%d %d %d\n",&genDim,&genID1,&genID2);
                if(switchPhysical) {
                  genID1 = dim_elem2Phys[genDim][genID1];
                  genID2 = dim_elem2Phys[genDim][genID2];
                }

                long int posBeforeTfo = ftell(fp);

                fgets(String, sizeof(String), fp);
                std::istringstream tfoLine(String);
                std::string transformationType;
                tfoLine >> transformationType;
                
                pTransform tfo = NULL; 
                if (transformationType.compare("Affine") == 0) {
                  doubleMatrix mat(4,4);
                  for (int i=0;i<4;i++) for (int j=0;j<4;j++) tfoLine >> mat(i,j);
                  mat(3,3) = 1.;
                  tfo = new MAdTransformation(mat);
                }
                
                if (tfo) {
                  // transformation from second (master) to first entity
                  pGEntity ge1 = GM_entityByTag(m->model,genDim,genID1);
                  pGEntity ge2 = GM_entityByTag(m->model,genDim,genID2);
                  GM_addTransformation(m->model,tfo,ge2,ge1); 

                  T_invert(tfo);
                  GM_addTransformation(m->model,tfo,ge1,ge2);
                  
                  delete tfo;
                }
                
                else fseek(fp,posBeforeTfo,SEEK_SET);
                
                int nbVtcs(0);
                fscanf(fp,"%d",&nbVtcs);
                
                for (int iVtx=0;iVtx<nbVtcs;iVtx++) {
                  int vtxID[2];
                  fscanf(fp,"%d %d",vtxID,vtxID+1);
                  std::set<int> vtcs;
                  
                  for (int i=0;i<2;i++) {
                    vtcs.insert(vtxID[i]);
                    std::map<int,std::set<int> >::iterator cIter = connections.find(vtxID[i]);
                    if (cIter != connections.end()) {
                      vtcs.insert(cIter->second.begin(),cIter->second.end());
                    }
                  }

                  for (std::set<int>::iterator vIter = vtcs.begin();vIter!=vtcs.end();++vIter) {
                    connections[*vIter].insert(vtcs.begin(),vtcs.end());
                  }
                }
              }
            }
          }
          break;
        }
      default:
        throw;
      }
      
      std::map<int,std::set<int> >::iterator cIter = connections.begin();
      for (;cIter!=connections.end();++cIter) {
        
        int vtxId = cIter->first;
        pVertex vt1 = m->find_point(vtxId);
                
        if (vt1) {
          
          void* tmp_periodic = NULL;
          int havePeriodic = EN_getDataPtr((pEntity) vt1,tagPeriodicVertexID,&tmp_periodic);
          std::map<int,std::vector<int> > *  myConnections = NULL;
          
          if (havePeriodic) {
            myConnections = (std::map<int,std::vector<int> > * )  (tmp_periodic);
          }
          else {
            myConnections = new std::map<int,std::vector<int> >;
            EN_attachDataPtr((pEntity) vt1,tagPeriodicVertexID,myConnections);
          }
          
          std::set<int>::const_iterator vIter2 = cIter->second.begin();
          for (;vIter2!=cIter->second.end();++vIter2) {
            if (vtxId != *vIter2) {
              myConnections->insert(std::make_pair(*vIter2,std::vector<int>()));
            }
          }
        }
      }

      // computeTransformations(m);
    }
    
    fclose(fp);
    // ----------------------------------------------
    // ------ Tagging inter-partition nodes
    // ----------------------------------------------

    pMeshDataId tag = MD_lookupMeshDataId("RemotePoint");
    


#ifdef PROFILE
    if (myRank==0) {
      dt = - mrm.getTime();
    }
#endif
    
    // --- establishing vertex connectivity ----------------------------

    V_createInfoInterface(m,tag);

#ifdef PROFILE  
    dt += mrm.getTime();
    mrm.getLocalMemoryUsage(rss,vm);
    if (myRank==0) {
      msg.info(__LINE__,__FILE__,
               "Establishing node connectivity took %.2f s, "
               "rss = %.2f MB and vm = %.2f MB",
               dt,rss,vm);
    }
    dt = - mrm.getTime();            
#endif

#ifdef PARALLEL
    int dim_max(dim);
    MPI_Allreduce( &dim, &dim_max, 1, MPI_INTEGER, MPI_MAX, m->getMPIComm() );
    dim = dim_max;
#endif

    switch (dim) {

    case 2: 
      {
    
        // --- establishing edge connectivity ------------------------------
        
        E_createInfoInterface(m,tag);
        // 
        // #ifdef PROFILE
        //     dt += mrm.getTime();
        //     mrm.getLocalMemoryUsage(rss,vm);
        //     if (myRank==0) {
        //       msg.info(__LINE__,__FILE__,
        //                "Establishing edge connectivity took %.2f secs, "
        //                "rss = %.2f MB and vm = %.2f MB",
        //                dt,rss,vm);
        //     }
        //     dt = - mrm.getTime();
        // #endif
        break;
      }
      
    case 3:
      {
        // --- establishing face connectivity ------------------------------
        
        F_createInfoInterface(m,tag);
        
#ifdef PROFILE   
        dt += mrm.getTime();
        mrm.getLocalMemoryUsage(rss,vm);
        if (myRank==0) {
          msg.info(__LINE__,__FILE__,
                   "Establishing face connectivity took %.2f secs, "
                   "rss = %.2f MB and vm = %.2f MB",
                   dt,rss,vm);  
        }
        dt = - mrm.getTime();
#endif
        break;
      }
    }
    
    // --- initializing identity ---------------------------------------
    
    m->initializeIdData();
      
#ifdef PROFILE 
    dt += mrm.getTime();
    mrm.getLocalMemoryUsage(rss,vm);
    if (myRank == 0) {
      msg.info(__LINE__,__FILE__,
               "Initialization of the id data took %.2f secs,"
               "memory usage: rss = %.2f MB and vm = %.2f MB",
               dt,rss,vm);  
      M_info(m);
    }
    dt = -mrm.getTime();
#endif
  }

  void LoadGmshMeshExperimental (pMesh m,const char *filename,int part,bool switchPhysical)
  {

    const MAdMsg& msg = MAdMsgSgl::instance();

#ifdef PROFILE
    const MAdResourceManager& mrm = MAdResourceManagerSgl::instance();
    double dt = - mrm.getTime();
#endif
    
    double scale = m->getScale();

    FILE *fp = fopen(filename, "rb");
    if(!fp) msg.error(__LINE__,__FILE__,"Unknown File %s",filename);

    int myRank = 0;
#ifdef PARALLEL 
    int nbProc = 1;
    MPI_Comm_size(m->getMPIComm(), &nbProc);
    MPI_Comm_rank(m->getMPIComm(), &myRank);
    int rank = (part==-1)? myRank:part;
#endif

    char String[256];

    if (!m->model) {
      pGModel model = NULL;
      GM_create(&model,"");
      m->model = model; 
    }

    int version = 1;
    int subversion = 0;
    int format = 0, size;
    bool binary = false, swap = false;
    std::map<int,double*>    mapPoints;
#ifdef PARALLEL    
    std::vector<pEntity> mdbPointsGhosts;
    std::vector<pEntity> mdbEdgesGhosts;
    std::vector<pEntity> mdbFacesGhosts;
#endif    
    std::map<int,std::set<int> > connections;
    int dim = 0;
    bool periodic = false;

    do 
    {
      if(!fgets(String,sizeof(String),fp))
      {
        break;
      }
      if(String[0]=='$')
      {
        if(!strncmp(&String[1],"MeshFormat",10))
        {
          fgets(String,sizeof(String),fp);
          std::string fversion(String);          
          size_t dotPos = fversion.find_first_of(".");
          std::istringstream versBuf(fversion.substr(0,dotPos));
          versBuf >> version;
          if(dotPos!=std::string::npos)
          {
            std::istringstream subVersBuf(fversion.substr(dotPos+1,fversion.size()));
            subVersBuf>>subversion;
            subVersBuf>>format;
            subVersBuf>>size;
          }
#ifdef PARALLEL
    if ( version == 1 && nbProc > 1 ) {
      if (myRank==0) msg.error(__LINE__,__FILE__,
                               "Cannot Gmsh mesh %s: parallel not implemented for format msh1",
                               filename);
    }
#endif

#ifdef PROFILE
          if(myRank==0)
          {
            msg.info(__LINE__,__FILE__,"Reading Gmsh version %d.%d",version,subversion);
          }
#endif
          if(format){
            binary = true;
#ifdef PROFILE
          if(myRank==0) {  
            msg.info(__LINE__,__FILE__,"Mesh file is in binary format");
          }
#endif 
            int one;            
            fread(&one,sizeof(int), 1, fp);        
            if(one != 1){
              swap = true;
#ifdef PROFILE
              msg.info(__LINE__,__FILE__,"Swapping bytes from binary file");
#endif
            }
          }
        }
        // --- Read List of nodes --- //      
        if( (!strncmp(&String[1],"NOD",3)||!strncmp(&String[1],"NOE",3)||
              !strncmp(&String[1],"Nodes",5)) )
        {
          if(!feof(fp))
          {
            int nbNodes;
            fscanf(fp,"%d \n",&nbNodes);
            for(int i_Node=0; i_Node<nbNodes; i_Node++)
            {
              int id;
              double* xyz = new double[3];
              if(!binary)
              {
                fscanf(fp,"%d %lf %lf %lf",&id,xyz,xyz+1,xyz+2);
              }
              else
              { 
                fread(&id,sizeof(int),1,fp);
                if(swap) SwapBytes((char*)&id, sizeof(int), 1);
                fread(xyz, sizeof(double), 3, fp);
                if(swap) SwapBytes((char*)xyz, sizeof(double), 3);
              }
              std::pair<int,double*> nodeInfo = std::pair<int,double*>(id,xyz);
              mapPoints.insert(nodeInfo);
            }
          }
        }
        else if( !strncmp(&String[1],"ParametricNodes",15) )
        {
          if ( version == 2 )
          {
            if(!feof(fp))
            {
              int nbNodes;
              fscanf(fp,"%d \n",&nbNodes);
              for(int i_Node=0; i_Node<nbNodes; i_Node++)
              {
                int id;
                double* xyz =new double[3];
                if(!binary)
                {
                  fscanf(fp,"%d %lf %lf %lf",&id,xyz,xyz+1,xyz+2);
                }
                else
                {
                  fread(&id, sizeof(int), 1, fp);
                  if(swap) SwapBytes((char*)&id, sizeof(int), 1);
                  fread(xyz, sizeof(double), 3, fp);
                  if(swap) SwapBytes((char*)xyz, sizeof(double), 3);
                }
                std::pair<int,double*> nodeInfo = std::pair<int,double*>(id,xyz);
                mapPoints.insert(nodeInfo);
              }
            }
          }
          else
          {
            if(!feof(fp))
            {
              int nbNodes;
              fscanf(fp,"%d \n",&nbNodes);
              for(int i_Node = 0; i_Node < nbNodes; i_Node++)
              {
                int id,gDim,gTag;
                double* xyz = new double[3];
                if(!binary)
                {                
                  fscanf(fp,"%d %lf %lf %lf %d %d",&id,xyz,xyz+1,xyz+2,&gDim,&gTag);
                }
                else
                {
                  fread(&id, sizeof(int), 1, fp);
                  if(swap) SwapBytes((char*)&id, sizeof(int), 1);
                  fread(xyz, sizeof(double), 3, fp);
                  if(swap) SwapBytes((char*)xyz, sizeof(double), 3);
                  fread(&gDim, sizeof(int), 1, fp);
                  if(swap) SwapBytes((char*)&gDim, sizeof(int), 1);
                  fread(&gTag, sizeof(int), 1, fp);
                  if(swap) SwapBytes((char*)&gTag, sizeof(int), 1);
                }
                if ( gDim == 3 )
                {
                  std::pair<int,double*> nodeInfo = std::pair<int,double*>(id,xyz);
                  mapPoints.insert(nodeInfo);
                }
                else {
                  double u,v;
                  if ( gDim == 0 )
                  {
                    u = v = 0.;
                  }
                  else if ( gDim == 1 )
                  {
                    if(!binary)
                    {
                      fscanf(fp,"%lf",&u);
                    }
                    else
                    {
                      fread(&u, sizeof(double), 1, fp);
                      if(swap) SwapBytes((char*)&u, sizeof(double), 1);
                    }
                    v = 0.;
                  } 
                  else if ( gDim == 2 )
                  {
                    if(!binary)
                    {
                      fscanf(fp,"%lf %lf",&u,&v);
                    }
                    else
                    {
                      fread(&u, sizeof(double), 1, fp);
                      if(swap) SwapBytes((char*)&u, sizeof(double), 1);
                      fread(&v, sizeof(double), 1, fp);
                      if(swap) SwapBytes((char*)&v, sizeof(double), 1);
                    }
                  }
                  m->add_pointParam(id,xyz[0]*scale,xyz[1]*scale,xyz[2]*scale,u,v);
                }
              }
            }          
          }  
        }
        else if( (!strncmp(&String[1],"ELM",3)||!strncmp(&String[1],"Elements",8)) )
        {
          if(!feof(fp))
          {
            int nbElements, NbTags, verts[256], Tag;
            int Num, Type, Physical, Elementary, Nbr_Nodes, Partition;
            int nbPartitions = 0;
            fscanf(fp,"%d \n",&nbElements);
            std::set<int> nbParts;
            if(!binary)
            {
              for(int i_Element=0; i_Element<nbElements; i_Element++)
              {
                std::set<int> partitions;
                std::set<int> ghosts;
                if(version == 1)
                {
                  fscanf(fp,"%d %d %d %d %d",&Num,&Type,&Physical,&Elementary,&Nbr_Nodes);
                  Partition = -1;
                }
                else
                {
                  Elementary = Physical = Partition = -1;
                  fscanf(fp, "%d %d %d", &Num, &Type, &NbTags);
                  for(int j = 0; j < NbTags; j++)
                  {
                    fscanf(fp,"%d",&Tag);
                    if     (j == 0) Physical = Tag;
                    else if(j == 1) Elementary = Tag;
                    else if(subversion <  2 && j == 2) partitions.insert(Tag-1);
                    else if(subversion >= 2 && j == 2 && NbTags > 3) nbPartitions = Tag;
                    else if(subversion >= 2 && j == 3) partitions.insert(Tag-1);
                    else if(j >= 4 && j < 4 + nbPartitions - 1) ghosts.insert(-Tag);
                  }
                  Nbr_Nodes = getNumVerticesForElementTypeMSH(Type);
                }                
                nbParts.insert(partitions.begin(),partitions.end());
                dim = std::max(dim,getDimForElementTypeMSH(Type));

                for(int j = 0; j < Nbr_Nodes; j++)
                {
                    fscanf(fp, "%d", &verts[j]);
                }

#ifdef PARALLEL
            // only take nodes on my partition 
            // but internal faces do not (necessarily) have parallel classification
                if ((partitions.find(rank) != partitions.end() || partitions.size() == 0)) //&&(ghosts.size()==0))
#endif
                {  
                  // avoid partition boundary
                  if (Elementary >= 0)
                  {
                    int order = getMappingOrderMSH(Type);
              
                    pGEntity geom = 0;
            
                    int elemTag = switchPhysical ? Physical : Elementary;
//                     int physTag = switchPhysical ? Elementary : Physical;

                    switch (getDimForElementTypeMSH(Type))
                    {
                      case 0:
                        geom = (pGEntity) GM_vertexByTag(m->model,elemTag);
                        break;
                      case 1:
                        geom = (pGEntity) GM_edgeByTag(m->model,elemTag);
                        break;
                      case 2:
                        geom = (pGEntity) GM_faceByTag(m->model,elemTag);
                        break;
                      case 3:
                        geom = (pGEntity) GM_regionByTag(m->model,elemTag);
                        break;
                    }
              
                    if (!geom) {
                      msg.error(__LINE__,__FILE__,
                                "Cannot find entity %i of dimension %i",
                                elemTag,getDimForElementTypeMSH(Type));
                    }
                    
                    if ( Physical > 0) GEN_setPhysical(geom,GEN_type(geom),Physical);
                    for(int j = 0; j < Nbr_Nodes; j++) {
                      MDB_Point* p = m->find_point(verts[j]);
                      if ( !p ) {
                        std::map<int,double*>::iterator it = mapPoints.find(verts[j]);
                        m->add_point(it->first,scale*it->second[0],scale*it->second[1],scale*it->second[2]);
                      }
                    }
                    pEntity pe = NULL;

                    switch (Type) 
                    {
                      // --- single point --------------------------------------------
                      case MSH_PNT:
                      {
                        pe = m->find_point(verts[0]);
                        pe->g = geom;
                        break;
                      }
                      // --- edges ---------------------------------------------------
                      // straight edge
                      case MSH_LIN_2:
                      {
                        pe = m->add_edge(verts[0],verts[1],geom);
                        break;
                      }
                      // quadratic edge
                      case MSH_LIN_3:
                      {
                        pe = m->add_edge(3,geom,verts[0],verts[2],verts[1]);              
                        break;
                      }
                      // cubic edge
                      case MSH_LIN_4:
                      {
                        pe = m->add_edge(4,geom, verts[0],verts[2],verts[3],verts[1]);
                        break;
                      }
                      // quartic edge
                      case MSH_LIN_5:
                      {
                        pe = m->add_edge(5,geom, 
                                        verts[0],verts[2],verts[3],
                                        verts[4],verts[1]);
                        break;
                      }
                      // --- triangles -----------------------------------------------
                      // straight edged triangle
                      case MSH_TRI_3:
                      {
                        pe = m->add_triangle(verts[0],verts[1],verts[2],geom);
                        break;
                      }
                      // quadratic triangle 
                      case MSH_TRI_6:
                      {
                        pe = m->add_triangle(2,0,geom,
                                          verts[0],verts[3],verts[1],
                                          verts[4],verts[2],verts[5]);
                        break;
                      }
                      // 3rd order serendipity triangle
                      case MSH_TRI_9:
                      {
                        pe = m->add_triangle(3,0,geom,
                                            verts[0],verts[3],verts[4],
                                            verts[1],verts[5],verts[6],
                                            verts[2],verts[7],verts[8]);
                        break;
                      }
                      // 3rd order complete triangle
                      case MSH_TRI_10:
                      {
                        pe = m->add_triangle(3,1,geom,
                                            verts[0],verts[3],verts[4],verts[1],
                                            verts[5],verts[6],verts[2],verts[7],
                                            verts[8],verts[9]);
                        break;
                      }
                      // 4th order serendipity triangle 
                      case MSH_TRI_12:
                      {
                        pe = m->add_triangle(4,0,geom,
                                            verts[0],verts[3],verts[4],verts[5],
                                            verts[1],verts[6],verts[7],verts[8],
                                            verts[2],verts[9],verts[10],verts[11]);
                        break;
                      }
                      // 4th order complete triangle (lagrange !)
                      case MSH_TRI_15:
                      {
                        pe = m->add_triangle(4,1,geom,
                                            verts[0],verts[3],verts[4],
                                            verts[5],verts[1],verts[6],
                                            verts[7],verts[8],verts[2],
                                            verts[9],verts[10],verts[11],
                                            verts[12],verts[13],verts[14]);
                        break;
                      }
                      // --- quadrangles ---------------------------------------------
                      // straight sided quadrangle
                      case MSH_QUA_4:
                      {
                        pe = m->add_quad(verts[0],verts[1],verts[2],verts[3],geom);
                        break;
                      }
                      // high order quadrangles
                      case MSH_QUA_9:
                      case MSH_QUA_16:
                      case MSH_QUA_25:
                      {
                        pe = m->add_quad(geom,order,false,verts);
                        break;
                      }
                      // 8-node serendip quad
                      case MSH_QUA_8:
                      {
                        pe = m->add_quad(geom, 2, true, verts);
                        break;
                      }
                      // --- tetrahedra ----------------------------------------------
                      // straight faced tetrahedron
                      case MSH_TET_4:
                      {
                        pe = m->add_tet(verts[0],verts[1],verts[2],verts[3],geom);
                        break;
                      }
                      break;
                      // Comlete high-order tetrahedra, 2nd order
                      case MSH_TET_10:
                      case MSH_TET_20:
                      case MSH_TET_35:
                      {
                        pe = m->add_tet(geom,order,false,verts);
                        break;
                      }
                      // 34 node tetrahedron, 4th order incomplete
                      case MSH_TET_34:
                      {
                        pe = m->add_tet(geom,4,true,verts);
                        break;
                      }
                      // --- prisms --------------------------------------------------
                      // straight edged prism 
                      case MSH_PRI_6:
                      {
                        pe = m->add_prism(verts[0],verts[1],verts[2],
                                          verts[3],verts[4],verts[5],geom);
                        break;
                      }
                      // 2nd order incomplete prism
                      case MSH_PRI_15:
                      {
                        pe = m->add_prism(geom,2,true,verts);
                        break;
                      }
                      // 2nd order complete prism 
                      case MSH_PRI_18:
                      {
                        pe = m->add_prism(geom,2,false,verts);
                        break;
                      }
                      
                      // --- hexahedra -----------------------------------------------
                      // straight-edged hexahedron 
                      case MSH_HEX_8:
                      {
                        pe = m->add_hex(verts[0],verts[1],verts[2],verts[3],
                                        verts[4],verts[5],verts[6],verts[7],geom);
                        break;
                      }
                      case MSH_HEX_20:
                      {
                        pe = m->add_hex(geom, 2, verts, true);
                        break;
                      }
                      case MSH_HEX_27:
                      case MSH_HEX_64:
                      case MSH_HEX_125:
                      {
                        pe = m->add_hex(geom,order,verts);
                        break;
                      }
                      // --- pyramids ------------------------------------------------
                      case MSH_PYR_5:
                      case MSH_PYR_14:
                      {
                        pe = m->add_pyramid(geom,order,false,verts);
                        break;
                      }
                      case MSH_PYR_13:          
                      {
                        pe = m->add_pyramid(geom,order, true,verts);
                        break;
                      }
                      // --- unknown element type ------------------------------------
                      default:
                        if (myRank==0) {
                            char elemTypeName[256];
                            getElementNameMSH(Type, elemTypeName);
                            msg.error(__LINE__, __FILE__,
                                "Element of type %s (tag %i) is (currently) not supported",
                                elemTypeName, Type);
                        }
                      throw;
                      break; 
                    }
                  
#ifdef PARALLEL
                    if(ghosts.size()!=0)
                    {
                      if(getDimForElementTypeMSH(Type)==0)
                      {
                        mdbPointsGhosts.push_back(pe);
                      }
                      else if(getDimForElementTypeMSH(Type)==1)
                      {    
                        mdbEdgesGhosts.push_back(pe);
                      }
                      else if(getDimForElementTypeMSH(Type)==2)
                      {    
                        mdbFacesGhosts.push_back(pe);
                      }
                    } 
#endif 
                    if (pe&&Type!=MSH_PNT) EN_setID(pe,Num);

                    bool find = false;
                    for (std::multimap<int, pGEntity>::iterator it = m->geomFeatures_Tags.lower_bound(Elementary);
                        it != m->geomFeatures_Tags.upper_bound(Elementary);++it)
                    {
                      if (it->second == geom)find = true;
                    }
                    if (!find) m->geomFeatures_Tags.insert(std::pair<int,pGEntity>(Elementary, geom));
                  }
                }
              }
            }
            else
            {
              int numElementsPartial = 0;
              while(numElementsPartial < nbElements)
              {
                std::set<int> partitions;
                std::set<int> ghosts;
                int header[3];
                fread(header, sizeof(int), 3, fp);
                if(swap) SwapBytes((char*)header, sizeof(int), 3);
                Type = header[0];
                int nbElms = header[1];
                NbTags = header[2];
                Nbr_Nodes = getNumVerticesForElementTypeMSH(Type);
                unsigned int n = 1 + NbTags + Nbr_Nodes;
                int *data = new int[n];
                for(int i = 0; i < nbElms; i++)
                {
                  fread(data, sizeof(int), n, fp);
                  if(swap) SwapBytes((char*)data, sizeof(int), n);
                  Num = data[0];
                  Physical = (NbTags > 0) ? data[1]: 0;
                  Elementary = (NbTags > 1) ? data[2] : 0;
                  nbPartitions = (subversion >= 2 && NbTags > 3) ? data[3] : 0;
                  if(subversion<2 && NbTags== 2) partitions.insert(data[3]-1);
                  if(version >= 2 && NbTags > 3) partitions.insert(data[4]-1);
                  if(nbPartitions>0)
                  {
                    for(int k=0; k<nbPartitions-1; k++)
                    {
                      ghosts.insert(-data[5 + k]);
                    }
                  }
                }
                nbParts.insert(partitions.begin(),partitions.end());
                dim = std::max(dim,getDimForElementTypeMSH(Type));
                
                for(int j=0; j<Nbr_Nodes; j++)
                {
                  verts[j] = data[ n-Nbr_Nodes+j ];
                }
#ifdef PARALLEL
            // only take nodes on my partition 
            // but internal faces do not (necessarily) have parallel classification
                if ((partitions.find(rank) != partitions.end() || partitions.size() == 0)) //&&(ghosts.size()==0))
#endif
                {  
                  // avoid partition boundary
                  if (Elementary >= 0)
                  {
                    int order = getMappingOrderMSH(Type);
              
                    pGEntity geom = 0;
            
                    int elemTag = switchPhysical ? Physical : Elementary;
//                     int physTag = switchPhysical ? Elementary : Physical;

                    switch (getDimForElementTypeMSH(Type))
                    {
                      case 0:
                        geom = (pGEntity) GM_vertexByTag(m->model,elemTag);
                        break;
                      case 1:
                        geom = (pGEntity) GM_edgeByTag(m->model,elemTag);
                        break;
                      case 2:
                        geom = (pGEntity) GM_faceByTag(m->model,elemTag);
                        break;
                      case 3:
                        geom = (pGEntity) GM_regionByTag(m->model,elemTag);
                        break;
                    }
              
                    if (!geom) {
                      msg.error(__LINE__,__FILE__,
                                "Cannot find entity %i of dimension %i",
                                elemTag,getDimForElementTypeMSH(Type));
                    }
              
                    if ( Physical > 0) GEN_setPhysical(geom,GEN_type(geom),Physical);
                    for(int j = 0; j < Nbr_Nodes; j++)
                    {
                      MDB_Point* p = m->find_point(verts[j]);
                      if ( !p )
                      {
                        std::map<int,double*>::iterator it = mapPoints.find(verts[j]);
                        m->add_point(it->first,scale*it->second[0],scale*it->second[1],scale*it->second[2]);
                      }
                    }

                    pEntity pe = NULL;

                    switch (Type) 
                    {
                      // --- single point --------------------------------------------
                      case MSH_PNT:
                      {
                        pe = m->find_point(verts[0]);
                        pe->g = geom;
                        break;
                      }
                      // --- edges ---------------------------------------------------
                      // straight edge
                      case MSH_LIN_2:
                      {
                        pe = m->add_edge(verts[0],verts[1],geom);
                        break;
                      }
                      // quadratic edge
                      case MSH_LIN_3:
                      {
                        pe = m->add_edge(3,geom,verts[0],verts[2],verts[1]);              
                        break;
                      }
                      // cubic edge
                      case MSH_LIN_4:
                      {
                        pe = m->add_edge(4,geom, verts[0],verts[2],verts[3],verts[1]);
                        break;
                      }
                      // quartic edge
                      case MSH_LIN_5:
                      {
                        pe = m->add_edge(5,geom, 
                                        verts[0],verts[2],verts[3],
                                        verts[4],verts[1]);
                        break;
                      }
                      // --- triangles -----------------------------------------------
                      // straight edged triangle
                      case MSH_TRI_3:
                      {
                        pe = m->add_triangle(verts[0],verts[1],verts[2],geom);
                        break;
                      }
                      // quadratic triangle 
                      case MSH_TRI_6:
                      {
                        pe = m->add_triangle(2,0,geom,
                                          verts[0],verts[3],verts[1],
                                          verts[4],verts[2],verts[5]);
                        break;
                      }
                      // 3rd order serendipity triangle
                      case MSH_TRI_9:
                      {
                        pe = m->add_triangle(3,0,geom,
                                            verts[0],verts[3],verts[4],
                                            verts[1],verts[5],verts[6],
                                            verts[2],verts[7],verts[8]);
                        break;
                      }
                      // 3rd order complete triangle
                      case MSH_TRI_10:
                      {
                        pe = m->add_triangle(3,1,geom,
                                            verts[0],verts[3],verts[4],verts[1],
                                            verts[5],verts[6],verts[2],verts[7],
                                            verts[8],verts[9]);
                        break;
                      }
                      // 4th order serendipity triangle 
                      case MSH_TRI_12:
                      {
                        pe = m->add_triangle(4,0,geom,
                                            verts[0],verts[3],verts[4],verts[5],
                                            verts[1],verts[6],verts[7],verts[8],
                                            verts[2],verts[9],verts[10],verts[11]);
                        break;
                      }
                      // 4th order complete triangle (lagrange !)
                      case MSH_TRI_15:
                      {
                        pe = m->add_triangle(4,1,geom,
                                            verts[0],verts[3],verts[4],
                                            verts[5],verts[1],verts[6],
                                            verts[7],verts[8],verts[2],
                                            verts[9],verts[10],verts[11],
                                            verts[12],verts[13],verts[14]);
                        break;
                      }
                      // --- quadrangles ---------------------------------------------
                      // straight sided quadrangle
                      case MSH_QUA_4:
                      {
                        pe = m->add_quad(verts[0],verts[1],verts[2],verts[3],geom);
                        break;
                      }
                      // high order quadrangles
                      case MSH_QUA_9:
                      case MSH_QUA_16:
                      case MSH_QUA_25:
                      {
                        pe = m->add_quad(geom,order,false,verts);
                        break;
                      }
                      // 8-node serendip quad
                      case MSH_QUA_8:
                      {
                        pe = m->add_quad(geom, 2, true, verts);
                        break;
                      }
                      // --- tetrahedra ----------------------------------------------
                      // straight faced tetrahedron
                      case MSH_TET_4:
                      {
                        pe = m->add_tet(verts[0],verts[1],verts[2],verts[3],geom);
                        break;
                      }
                      break;
                      // Comlete high-order tetrahedra, 2nd order
                      case MSH_TET_10:
                      case MSH_TET_20:
                      case MSH_TET_35:
                      {
                        pe = m->add_tet(geom,order,false,verts);
                        break;
                      }
                      // 34 node tetrahedron, 4th order incomplete
                      case MSH_TET_34:
                      {
                        pe = m->add_tet(geom,4,true,verts);
                        break;
                      }
                      // --- prisms --------------------------------------------------
                      // straight edged prism 
                      case MSH_PRI_6:
                      {
                        pe = m->add_prism(verts[0],verts[1],verts[2],
                                          verts[3],verts[4],verts[5],geom);
                        break;
                      }
                      // 2nd order incomplete prism
                      case MSH_PRI_15:
                      {
                        pe = m->add_prism(geom,2,true,verts);
                        break;
                      }
                      // 2nd order complete prism 
                      case MSH_PRI_18:
                      {
                        pe = m->add_prism(geom,2,false,verts);
                        break;
                      }
                      
                      // --- hexahedra -----------------------------------------------
                      // straight-edged hexahedron 
                      case MSH_HEX_8:
                      {
                        pe = m->add_hex(verts[0],verts[1],verts[2],verts[3],
                                        verts[4],verts[5],verts[6],verts[7],geom);
                        break;
                      }
                      case MSH_HEX_20:
                      {
                        pe = m->add_hex(geom, 2, verts, true);
                        break;
                      }
                      case MSH_HEX_27:
                      case MSH_HEX_64:
                      case MSH_HEX_125:
                      {
                        pe = m->add_hex(geom,order,verts);
                        break;
                      }
                      // --- pyramids ------------------------------------------------
                      case MSH_PYR_5:
                      case MSH_PYR_14:
                      {
                        pe = m->add_pyramid(geom,order,false,verts);
                        break;
                      }
                      case MSH_PYR_13:          
                      {
                        pe = m->add_pyramid(geom,order, true,verts);
                        break;
                      }
                      // --- unknown element type ------------------------------------
                      default:
                        if (myRank==0) {
                            char elemTypeName[256];
                            getElementNameMSH(Type, elemTypeName);
                            msg.error(__LINE__, __FILE__,
                                "Element of type %s (tag %i) is (currently) not supported",
                                elemTypeName, Type);
                        }
                      throw;
                      break; 
                    }
                  
#ifdef PARALLEL
                    if(ghosts.size()!=0)
                    {
                      if(getDimForElementTypeMSH(Type)==0)
                      {
                        mdbPointsGhosts.push_back(pe);
                      }
                      else if(getDimForElementTypeMSH(Type)==1)
                      {    
                        mdbEdgesGhosts.push_back(pe);
                      }
                      else if(getDimForElementTypeMSH(Type)==2)
                      {    
                        mdbFacesGhosts.push_back(pe);
                      }
                    } 
#endif 
                    if (pe&&Type!=MSH_PNT) EN_setID(pe,Num);

                    bool find = false;
                    for (std::multimap<int, pGEntity>::iterator it = m->geomFeatures_Tags.lower_bound(Elementary);
                        it != m->geomFeatures_Tags.upper_bound(Elementary);++it)
                    {
                      if (it->second == geom)find = true;
                    }
                    if (!find) m->geomFeatures_Tags.insert(std::pair<int,pGEntity>(Elementary, geom));
                  }
                }
                delete [] data;
                numElementsPartial += nbElms;
              }
            }
            m->nbPartitions = (int)(nbParts.size());
          }
        }      
        else if( !strncmp(&String[1], "PERIODICNOD", 11) )
        {
          msg.info(__LINE__,__FILE__,"Format \"PERIODICNOD\" is no longer supported");
        } 
        else if( !strncmp(&String[1],"PeriodicNodes",13) ) 
        {
          periodic = true;
          int nPeriod;
          fscanf(fp, "%d ", &nPeriod);
        
          for(int ip = 0; ip < nPeriod; ip++)
          {          
            int Num;
            fscanf(fp, "%d ", &Num);
          
            std::set<int> vtcs;

            for (int i=0;i<Num;i++)
            {
              int vtxId;
              fscanf(fp, "%d ", &vtxId);
              vtcs.insert(vtxId);
            }
          
            for (std::set<int>::iterator vIter = vtcs.begin();vIter!=vtcs.end();++vIter)
            {
              std::map<int,std::set<int> >::iterator cIter = connections.find(*vIter);
              if (cIter!=connections.end())
              {
                vtcs.insert(cIter->second.begin(),cIter->second.end());
              }
            }
              
            for (std::set<int>::iterator vIter = vtcs.begin();vIter!=vtcs.end();++vIter)
            {
              connections[*vIter].insert(vtcs.begin(),vtcs.end());
            }
            int trafo;
            fscanf(fp, "%d ",&trafo);
          }
        }
        else if( (!strncmp(&String[1],"Periodic",8)&&(strlen(&String[1])==9)) )
        {
        periodic = true;
        int nPeriod;
        fscanf(fp, "%d ", &nPeriod);
        
        for (int iPeriod = 0;iPeriod<nPeriod;iPeriod++)
        {  
          int genDim(-1);
          int genID1(-1);
          int genID2(-1);
          
          fscanf(fp,"%d %d %d\n",&genDim,&genID1,&genID2);

          long int posBeforeTfo = ftell(fp);

          fgets(String, sizeof(String), fp);
          std::istringstream tfoLine(String);
          std::string transformationType;
          tfoLine >> transformationType;
          
          pTransform tfo = NULL; 
          if (transformationType.compare("Affine") == 0)
          {
            doubleMatrix mat(4,4);
            for (int i=0;i<4;i++) for (int j=0;j<4;j++) tfoLine >> mat(i,j);
            mat(3,3) = 1.;
            tfo = new MAdTransformation(mat);
          }
          
          if (tfo)
          {
            // transformation from second (master) to first entity
            pGEntity ge1 = GM_entityByTag(m->model,genDim,genID1);
            pGEntity ge2 = GM_entityByTag(m->model,genDim,genID2);
            GM_addTransformation(m->model,tfo,ge2,ge1); 

            T_invert(tfo);
            GM_addTransformation(m->model,tfo,ge1,ge2);
            
            delete tfo;
          }          
          else
          {
            fseek(fp,posBeforeTfo,SEEK_SET);
          }
          
          int nbVtcs(0);
          fscanf(fp,"%d",&nbVtcs);
          
          for (int iVtx=0;iVtx<nbVtcs;iVtx++)
          {
            int vtxID[2];
            fscanf(fp,"%d %d",vtxID,vtxID+1);
            
            std::set<int> vtcs;
            
            for (int i=0;i<2;i++)
            {
              vtcs.insert(vtxID[i]);
              std::map<int,std::set<int> >::iterator cIter = connections.find(vtxID[i]);
              if (cIter != connections.end())
              {
                vtcs.insert(cIter->second.begin(),cIter->second.end());
              }
            }

            for (std::set<int>::iterator vIter = vtcs.begin();vIter!=vtcs.end();++vIter)
            {
              connections[*vIter].insert(vtcs.begin(),vtcs.end());
            }
          }
        }
      }
      }
    } while(!feof(fp));

//-------------clean ----------------------------------------------------------
    for (std::map<int,double*>::iterator it = mapPoints.begin(); it!=mapPoints.end(); ++it)
    {
      if(it->second) delete [] it->second;
    }
    mapPoints.clear();
    
#ifdef PARALLEL    
   if(dim==3)
   {
     for (std::vector<pEntity>::iterator it = mdbFacesGhosts.begin() ; it != mdbFacesGhosts.end(); ++it)
     {
       m->del_face(dynamic_cast<MDB_Face*>(*it));
       delete *it;
     }
      
     for (std::vector<pEntity>::iterator it = mdbEdgesGhosts.begin() ; it != mdbEdgesGhosts.end(); ++it)
     {
       m->del_edge(dynamic_cast<MDB_Edge*>(*it));
       delete *it;
     }
      
     for (std::vector<pEntity>::iterator it = mdbPointsGhosts.begin() ; it != mdbPointsGhosts.end(); ++it)
     {
       m->del_point(dynamic_cast<MDB_Point*>(*it));
       delete *it;
     }
   }
   else if(dim==2)
   {
     for (std::vector<pEntity>::iterator it = mdbEdgesGhosts.begin() ; it != mdbEdgesGhosts.end(); ++it)
     {
       m->del_edge(dynamic_cast<MDB_Edge*>(*it));
       delete *it;
     }
      
     for (std::vector<pEntity>::iterator it = mdbPointsGhosts.begin() ; it != mdbPointsGhosts.end(); ++it)
     {
       m->del_point(dynamic_cast<MDB_Point*>(*it));
       delete *it;
     }
   }
   else if(dim==1)
   {
     for (std::vector<pEntity>::iterator it = mdbPointsGhosts.begin() ; it != mdbPointsGhosts.end(); ++it)
     {
       m->del_point(dynamic_cast<MDB_Point*>(*it));
       delete *it;
     }      
   }
   mdbFacesGhosts.clear();
   mdbEdgesGhosts.clear();
   mdbPointsGhosts.clear();
    
#endif
    
    if ((M_numRegions(m) == 0) && (M_numFaces(m) == 0) && (M_numEdges(m) == 0)) {
      msg.error(__LINE__,__FILE__,"Empty mesh on processor %d",myRank);
    }    
    
#ifdef PROFILE
    double rss;
    double vm;
    mrm.getLocalMemoryUsage(rss,vm);
    dt += mrm.getTime();
    if (myRank==0) {
      msg.info(__LINE__,__FILE__,
               "Reading took %.2f seconds, rss = %.2f MB, vm = %.2f MB",
               dt,rss,vm);
    }
    dt = - mrm.getTime();
#endif

    // transfer classification from parents to children if not yet classified
    m->classify_unclassified_entities();

#ifdef PROFILE
    dt += mrm.getTime();
    mrm.getLocalMemoryUsage(rss,vm);
    if (myRank==0) {
      msg.info(__LINE__,__FILE__,
               "Updating entity classication took %.2f s, rss = %.2f MB and vm = %.2f MB",
               dt,rss,vm);    
    }
    dt = - mrm.getTime();
#endif

    // remove unused elements - orphan entities are allowed if switchPhysical is activated
    if(!switchPhysical) {
      m->destroyStandAloneEntities();
    }

#ifdef PROFILE
    dt += mrm.getTime();    
    mrm.getLocalMemoryUsage(rss,vm);
    if (myRank==0) {
      msg.info(__LINE__,__FILE__,
               "Deleting stand alone entities took %.2f s, rss = %.2f MB and vm = %.2f MB",
               dt,rss,vm);
    }
#endif

    if(periodic)
    {
      pMeshDataId tagPeriodicVertexID = MD_lookupMeshDataId("PeriodicVertexID");
      std::map<int,std::set<int> >::iterator cIter = connections.begin();
      for (;cIter!=connections.end();++cIter)
      {
        int vtxId = cIter->first;
        pVertex vt1 = m->find_point(vtxId);
                
        if (vt1)
        {
          void* tmp_periodic = NULL;
          int havePeriodic = EN_getDataPtr((pEntity) vt1,tagPeriodicVertexID,&tmp_periodic);
          std::map<int,std::vector<int> > *  myConnections = NULL;
          
          if (havePeriodic)
          {
            myConnections = (std::map<int,std::vector<int> > * )(tmp_periodic);
          }
          else 
          {
            myConnections = new std::map<int,std::vector<int> >;
            EN_attachDataPtr((pEntity) vt1,tagPeriodicVertexID,myConnections);
          }
          
          std::set<int>::const_iterator vIter2 = cIter->second.begin();
          for (;vIter2!=cIter->second.end();++vIter2)
          {
            if (vtxId != *vIter2) myConnections->insert(std::make_pair(*vIter2,std::vector<int>()));
          }
        }
      }
    }
    
    fclose(fp);

    // ----------------------------------------------
    // ------ Tagging inter-partition nodes
    // ----------------------------------------------

    pMeshDataId tag = MD_lookupMeshDataId("RemotePoint");
    


#ifdef PROFILE
    dt = - mrm.getTime();
#endif
    
    // --- establishing vertex connectivity ----------------------------

    V_createInfoInterface(m,tag);

#ifdef PROFILE
    dt += mrm.getTime();
    mrm.getLocalMemoryUsage(rss,vm);
    if (myRank==0) {
      msg.info(__LINE__,__FILE__,
               "Establishing node connectivity took %.2f s, rss = %.2f MB and vm = %.2f MB",
               dt,rss,vm);
    }
    dt = - mrm.getTime();            
#endif

    
    // --- establishing edge connectivity ------------------------------
    
    E_createInfoInterface(m,tag);

#ifdef PROFILE
    dt += mrm.getTime();
    mrm.getLocalMemoryUsage(rss,vm);
    if (myRank==0) {
      msg.info(__LINE__,__FILE__,
               "Establishing edge connectivity took %.2f secs, rss = %.2f MB and vm = %.2f MB",
               dt,rss,vm);
    }
    dt = - mrm.getTime();
#endif
    
    // --- establishing face connectivity ------------------------------
    
    F_createInfoInterface(m,tag);

#ifdef PROFILE
    dt += mrm.getTime();
    mrm.getLocalMemoryUsage(rss,vm);
    if (myRank==0) {
      msg.info(__LINE__,__FILE__,
               "Establishing face connectivity took %.2f secs, rss = %.2f MB and vm = %.2f MB",
               dt,rss,vm);  
    }
#endif
    
    // --- initializing identity ---------------------------------------
    
#ifdef PROFILE
    dt = - mrm.getTime();
#endif
    m->initializeIdData();
      
#ifdef PROFILE 
    dt += mrm.getTime();
    mrm.getLocalMemoryUsage(rss,vm);
    if (myRank == 0) {
      msg.info(__LINE__,__FILE__,
               "Initialization of the id data took %.2f secs,"
               "rss = %.2f MB and vm = %.2f MB",
               dt,rss,vm);  
      M_info(m);
    }
#endif
  }
  // -------------------------------------------------------------------
  // ------------- Save a Gmsh mesh ------------------------------------
  // -------------------------------------------------------------------

  void SaveGmshMesh (const pMesh m,
                     const char *filename, 
                     int version, 
                     bool saveAll, 
                     const int * partitionTable)
  {

    const MAdMsg& msg = MAdMsgSgl::instance();

    if ( version == 1 && partitionTable ) {
      msg.error(__LINE__,__FILE__,
                "Error while saving mesh on file %s: SaveGmshMesh not implemented for msh1 format with partitioning",filename);
    }

    double scale = m->getScale();
    double overScale = 1./scale;

    int meshDim = M_dim(m);
    int nbModelVertex = 0;
    pMeshDataId tagPeriodicPoint = MD_lookupMeshDataId("PeriodicPoint");
    // Create a reverse map of physical tags
    GEntity2Physical gentity2phys(m->geomFeatures_Tags);
    FILE *f = fopen(filename, "w");
    if ( !f ) {
      throw;
    }
    if ( version != 1 ) {
      fprintf(f, "$MeshFormat\n2.1 0 8\n$EndMeshFormat\n");
    }
    // ------ Writting the nodes ------
    if ( version == 1 || !(m->isParametric()) )
      {
        if ( version == 1 ) { fprintf(f, "$NOD\n");   } 
        else                { fprintf(f, "$Nodes\n"); }
        fprintf(f, "%d\n", m->nbPoints);
        VIter vit = M_vertexIter(m); 
        while (VIter_next(vit)){}
        VIter_reset(vit);
        pVertex pv;  
        int NN = 0;
        while ((pv = VIter_next(vit)))
          {
            NN++;
            if(pv->g)
              {
                int dim = GEN_type(pv->g); 
                if(dim == 0)  nbModelVertex++;
                fprintf(f, "%d %.16g %.16g %.16g\n", pv->iD, 
                        pv->X*overScale, pv->Y*overScale, pv->Z*overScale);
              }
            else
              {
                fprintf(f, "%d %.16g %.16g %.16g\n", pv->iD, 
                        pv->X*overScale, pv->Y*overScale, pv->Z*overScale);
              }
          }
        if (NN != m->nbPoints) {
          msg.error(__LINE__,__FILE__,
                    "Vertex iterator saw %d vertices while there are %d points in the mesh",
                    NN,m->nbPoints);
        }
        VIter_delete(vit);
        if ( version == 1 ) { fprintf(f, "$ENDNOD\n");   } 
        else                { fprintf(f, "$EndNodes\n"); }
      }
    // ------ Writting the parametric nodes ------
    else
      {
        fprintf(f, "$ParametricNodes\n");
        fprintf(f, "%d\n", m->nbPoints);
        pVertex pv;
        VIter vit = M_vertexIter(m);
        while ((pv = VIter_next(vit)))
          {
            if(pv->g)
              {
                int gdim = GEN_type(pv->g);
                int gtag = GEN_tag(pv->g);
                fprintf(f, "%d %.16g %.16g %.16g %d %d", pv->iD, 
                        pv->X*overScale, pv->Y*overScale, pv->Z*overScale, 
                        gdim, gtag);
                double u[2]; V_params(pv,&u[0],&u[1]);
                switch (gdim) {
                case 0: { nbModelVertex++; break; }
                case 1: { fprintf(f, " %.16g", u[0]); break; }
                case 2: { fprintf(f, " %.16g %.16g", u[0], u[1]); break; }
                case 3: { break; }
                }
                fprintf(f, "\n");
              }
            else {
              msg.error(__LINE__,__FILE__,
                        "Parametric point without classification");
            }
          }
        VIter_delete(vit);
        fprintf(f, "$EndParametricNodes\n");
      }
    // ------ Writting the elements ------
    {
      if ( version == 1 ) { fprintf(f, "$ELM\n");   } 
      else                { fprintf(f, "$Elements\n"); }
    
      bool physic = m->model ? GM_physical(m->model) : false;
      long int nbClasVerts = 0;
      long int nbClasEdges = 0;
      long int nbClasFaces = 0;
      long int nbClasRegs = 0;
      {
        VIter vit = M_vertexIter(m);
        pVertex pv;
        int numV=0;
        while ((pv = VIter_next(vit)))
          {
            numV++;
            // cout << "numV: " << numV << endl;
            if( GEN_type(pv->g) == 0 ){
              if ( saveAll ) { nbClasVerts++; continue; }
              if ( physic ) {
                if ( GEN_physTag(pv->g) != 0 ) nbClasVerts++;
              }
              else nbClasVerts++;
            }
          }
        VIter_delete(vit);
      } 
      {
        EIter eit = M_edgeIter(m);
        pEdge pe; 
        while ((pe = EIter_next(eit)))
          {
            if( GEN_type(pe->g) == 1 ){
              if ( saveAll ) { nbClasVerts++; continue; }
              if ( physic ) {
                if ( GEN_physTag(pe->g) != 0 ) nbClasEdges++;
              }
              else nbClasEdges++;
            }
          }
        EIter_delete(eit);
      }  
      {
        FIter fit = M_faceIter(m);
        pFace pf;
        int nbFace=0;
        //const int nbFaceTot=M_numFaces(m)-1;
        //std::cout<<"nbFaceTot: " << nbFaceTot<<std::endl; 
        while ((pf = FIter_next(fit))/*&&(nbFace<nbFaceTot)*/) //pb detected by Farid here
          {
            nbFace++;
            // cout << "nbFace: " << nbFace << endl;
            if( GEN_type(pf->g) == 2 ){
              // cout << "saveAll: " << saveAll << endl;
              if ( saveAll ) { nbClasVerts++; continue; }
              // cout << "physic: " << physic << endl;
              if ( physic ) {
                if ( GEN_physTag(pf->g) != 0 ) nbClasFaces++;
              }
              else nbClasFaces++;
            }
          }
        FIter_delete(fit);
      }
      {
        RIter rit = M_regionIter(m);
        pRegion pr;
        while ((pr = RIter_next(rit)))
          {
            if( GEN_type(pr->g) == 3 ){
              if ( saveAll ) { nbClasVerts++; continue; }
              if ( physic ) {
                if ( GEN_physTag(pr->g) != 0 ) nbClasRegs++;
              }
              else nbClasRegs++;
            }
          }
        RIter_delete(rit);
      } 

      fprintf(f, "%ld\n", (nbClasVerts + nbClasEdges + nbClasFaces + nbClasRegs) );

      int k = 1;
      // --- Nodes ---
      {
        VIter vit = M_vertexIter(m);
        pVertex pv;  
        while ((pv = VIter_next(vit)))
          {
            if(pv->g)
              {
                int dim = GEN_type(pv->g);
                int nbTags = 3;
                int tag = GEN_tag (pv->g); 
                int phys = 0;
                if ( GEN_type(pv->g) == 0 ) phys = GEN_physTag(pv->g); 
                if ( !saveAll && physic && phys == 0 ) continue;
                //                 int phys = gentity2phys.get_first_tag(pv->g);
                int partition = -1;
                if ( partitionTable ) {
                  switch(meshDim) {
                  case 0:
                    partition = partitionTable[pv->iD];
                    break;
                  case 1:
                    for (int i=0; i < V_numEdges(pv); i++) {
                      int ePart = partitionTable[V_edge(pv,i)->iD];
                      partition = std::min(ePart,partition);
                      if (partition < 0) partition = ePart;
                    }
                    break;
                  case 2: {
                    pPList vFaces = V_faces(pv);
                    void *tmp_ptr = 0;
                    pFace pf;
                    while( (pf = (pFace) PList_next(vFaces,&tmp_ptr)) ) {
                      int fPart = partitionTable[pf->iD];
                      partition = std::min(fPart,partition);
                      if (partition < 0) partition = fPart;
                    }
                    PList_delete(vFaces);
                    break;
                  }
                  case 3: {
                    pPList vRegs = V_regions(pv);
                    void *tmp_ptr = 0;
                    pRegion pr;
                    while( (pr = (pRegion) PList_next(vRegs,&tmp_ptr)) ) {
                      int rPart = partitionTable[pr->iD];
                      partition = std::min(rPart,partition);
                      if (partition < 0) partition = rPart;
                    }
                    PList_delete(vRegs);
                    break;
                  }
                  }
                }

                if(dim == 0) {
                  if ( version == 1 ) { fprintf(f, "%d %d %d %d %d %d\n",
                                                k++, 15, phys, tag, 1, pv->iD); }
                  else                { fprintf(f, "%d %d %d %d %d %d %d\n",
                                                k++, 15, nbTags, phys, tag, partition + 1, pv->iD); }
                }
              }
          }
        VIter_delete(vit);    
      }
      // --- Edges ---
      {      
        EIter eit = M_edgeIter(m);
        pEdge pe;  
        while ((pe = EIter_next(eit)))
          {
            if(pe->g)
              {
                int dim = GEN_type(pe->g);
                int nbTags = 3;
                int tag = GEN_tag (pe->g);
                int phys = 0;
                if ( GEN_type(pe->g) == 1 ) phys = GEN_physTag(pe->g);
                if ( !saveAll && physic && phys == 0 ) continue; 
                //                 int phys = gentity2phys.get_first_tag(pe->g);
                int partition = -1;
                if ( partitionTable ) {
                  switch(meshDim) {
                  case 1:
                    partition = partitionTable[pe->iD];
                    break;
                  case 2:
                    for (int i=0; i < E_numFaces(pe); i++) {
                      int fPart = partitionTable[E_face(pe,i)->iD];
                      partition = std::min(fPart,partition);
                      if (partition < 0) partition = fPart;
                    }
                    break;
                  case 3: {
                    pPList eRegs = E_regions(pe);
                    void *tmp_ptr = 0;
                    pRegion pr;
                    while( (pr = (pRegion) PList_next(eRegs,&tmp_ptr)) ) {
                      int rPart = partitionTable[pr->iD];
                      partition = std::min(rPart,partition);
                      if (partition < 0) partition = rPart;
                    }
                    PList_delete(eRegs);
                    break;
                  }
                  }
                }
        
                if(dim ==1) {
                  if ( version == 1 ) { fprintf(f, "%d %d %d %d %d %d %d\n", 
                                                k++, 1, phys, tag, 2, 
                                                pe->p1->iD, pe->p2->iD); }
                  else                { fprintf(f, "%d %d %d %d %d %d %d %d\n",
                                                k++, 1, nbTags, phys, tag, partition + 1, 
                                                pe->p1->iD, pe->p2->iD); }
                }
              }
          }
        EIter_delete(eit);
      }
      // --- Faces ---  
      {
        FIter fit = M_faceIter(m);
        pFace pf;  
        while ((pf = FIter_next(fit)))
          {
            if(pf->g)
              {
                MDB_Point *nod[4];
                int dim = GEN_type(pf->g);
                int nbTags = 3;
                int tag = GEN_tag (pf->g); 
                int phys = 0;
                if ( GEN_type(pf->g) == 2 ) phys = GEN_physTag(pf->g);
                if ( !saveAll && physic && phys == 0 ) continue;
                //                 int phys = gentity2phys.get_first_tag(pf->g);
                int partition = -1;
                if ( partitionTable ) {
                  switch(meshDim) {
                  case 2:
                    partition = partitionTable[pf->iD];
                    break;
                  case 3:
                    for (int i=0; i < F_numRegions(pf); i++) {
                      int rPart = partitionTable[F_region(pf,i)->iD];
                      partition = std::min(rPart,partition);
                      if (partition < 0) partition = rPart;
                    }
                    break;
                  }
                }
        
                pf->getNodes(nod);
                  if(dim == 2) {
                    if (pf->getNbEdges() == 4) { //quad
                      if ( version == 1 ) { fprintf(f, "%d %d %d %d %d %d %d %d %d\n",
                                                    k++, 3, phys, tag, 4,
                                                    nod[0]->iD, nod[1]->iD, nod[2]->iD, nod[3]->iD); }
                      else                { fprintf(f, "%d %d %d %d %d %d %d %d %d %d\n",
                                                    k++, 3, nbTags, phys, tag, partition + 1, 
                                                    nod[0]->iD, nod[1]->iD, nod[2]->iD, nod[3]->iD); }
                    }
                    else { //triangle
                      if ( version == 1 ) { fprintf(f, "%d %d %d %d %d %d %d %d\n",
                                                    k++, 2, phys, tag, 3,
                                                    nod[0]->iD, nod[1]->iD, nod[2]->iD); }
                      else                { fprintf(f, "%d %d %d %d %d %d %d %d %d\n",
                                                    k++, 2, nbTags, phys, tag, partition + 1, 
                                                    nod[0]->iD, nod[1]->iD, nod[2]->iD); }
                    }
                  }
                }
              }
            FIter_delete(fit);
          } 
        {
          RIter rit = M_regionIter(m);
          pRegion pr;  
          while ((pr = RIter_next(rit)))
            {
              if(pr->g)
                {
                  MDB_Point *nod[8];
                  //int dim = GEN_type(pr->g);
                  int nbTags = 3;
                  int tag = GEN_tag (pr->g);
                  int phys = 0;
                  if ( GEN_type(pr->g) == 3 ) phys = GEN_physTag(pr->g);
                  if ( !saveAll && physic && phys == 0 ) continue;
                  //                 int phys = gentity2phys.get_first_tag(pr->g);
                  int partition = 0;
                  if ( partitionTable ) {
                    partition = partitionTable[pr->iD];
                  }
        
                  int numVer =  pr->getNbVertex();
                  pPList ll = R_vertices(pr);
                  for(int i=0;i<numVer;i++) nod[i] = (pVertex)PList_item(ll, i);
                  PList_delete(ll);
              
                  if(numVer==4) {
                    if ( version == 1 ) { fprintf(f, "%d %d %d %d %d %d %d %d %d\n", 
                                                  k++, 4, phys, tag, 4,
                                                  nod[0]->iD, nod[1]->iD, nod[2]->iD, nod[3]->iD); }
                    else                { fprintf(f, "%d %d %d %d %d %d %d %d %d %d\n",
                                                  k++, 4, nbTags, phys, tag, partition + 1, 
                                                  nod[0]->iD, nod[1]->iD, nod[2]->iD, nod[3]->iD); }
                  }
                  else if(numVer==8) {
                    if ( version == 1 ) { fprintf(f, "%d %d %d %d %d %d %d %d %d %d %d %d %d\n", 
                                                  k++, 5, phys, tag, 8,
                                                  nod[0]->iD, nod[1]->iD, nod[2]->iD, nod[3]->iD,
                                                  nod[4]->iD, nod[5]->iD, nod[6]->iD, nod[7]->iD); }
                    else                { fprintf(f, "%d %d %d %d %d %d %d %d %d %d %d %d %d %d\n",
                                                  k++, 5, nbTags, phys, tag, partition + 1, 
                                                  nod[0]->iD, nod[1]->iD, nod[2]->iD, nod[3]->iD,
                                                  nod[4]->iD, nod[5]->iD, nod[6]->iD, nod[7]->iD); }
                  }
                  else if(numVer==6) { 
                    if ( version == 1 ) { fprintf(f, "%d %d %d %d %d %d %d %d %d %d %d\n", 
                                                  k++, 6, phys, tag, 6,
                                                  nod[0]->iD, nod[1]->iD, nod[2]->iD, nod[3]->iD,
                                                  nod[4]->iD, nod[5]->iD); }
                    else                { fprintf(f, "%d %d %d %d %d %d %d %d %d %d %d %d\n",
                                                  k++, 6, nbTags, phys, tag, partition + 1, 
                                                  nod[0]->iD, nod[1]->iD, nod[2]->iD, nod[3]->iD,
                                                  nod[4]->iD, nod[5]->iD); }
                  }
                  else if(numVer==5) {
                    if( version == 1  ) { fprintf(f, "%d %d %d %d %d %d %d %d %d %d\n",
                                                  k++, 7, phys, tag, 5,
                                                  nod[0]->iD, nod[1]->iD, nod[2]->iD, nod[3]->iD,
                                                  nod[4]->iD); }
                    else                { fprintf(f, "%d %d %d %d %d %d %d %d %d %d %d\n",
                                                  k++, 7, nbTags, phys, tag, partition + 1,
                                                  nod[0]->iD, nod[1]->iD, nod[2]->iD, nod[3]->iD,
                                                  nod[4]->iD); }
                  }
                }
            }
          RIter_delete(rit);
        }
        if ( version == 1 ) { fprintf(f, "$ENDELM\n");   } 
        else                { fprintf(f, "$EndElements\n"); }
        {
          int nperiodic = 0,nref = 0;;
          VIter vit = M_vertexIter(m);
          pVertex pv;  
          while ((pv = VIter_next(vit)))
            {
              void *temp_ptr; 
              int isPeriodic = EN_getDataPtr((pEntity) pv , tagPeriodicPoint, &temp_ptr);
              if(isPeriodic) {
                std::vector<std::pair<std::vector<int> , pVertex> > *recup = (std::vector<std::pair<std::vector<int> , pVertex> > *) temp_ptr;
                if(!nperiodic) {
                  nref = (int)((*recup)[0].first.size());
                }
                nperiodic+=(int)((*recup).size());
              }
            }
          VIter_delete(vit);  
          if(nperiodic) {
            fprintf(f, "$PERIODICNOD\n");
            fprintf(f, "%d %d\n", nperiodic,nref);
            int num = 1;
            vit = M_vertexIter(m);
            while ((pv = VIter_next(vit)))
              {
                void *temp_ptr; 
                int isPeriodic = EN_getDataPtr((pEntity) pv , tagPeriodicPoint, &temp_ptr);
                if(isPeriodic) {
                  std::vector<std::pair<std::vector<int> , pVertex> > *recup = (std::vector<std::pair<std::vector<int> , pVertex> > *) temp_ptr;
                  for(unsigned int rk=0 ; rk<(*recup).size() ; rk++) {
                    fprintf(f, "%d", num++);
                    for(unsigned int j=0 ; j<(*recup)[rk].first.size() ; j++) 
                      fprintf(f, " %d", (*recup)[rk].first[j]);
                    fprintf(f, " %d %d \n", EN_id((pEntity) pv),EN_id((pEntity) (*recup)[rk].second));
                  }
                  //delete recup;
                  //EN_deleteData((pEntity) pv , tagPeriodicPoint);
                }
              }
            VIter_delete(vit);
            fprintf(f, "$ENDPERIODICNOD\n");  
          }
        }  
      
        {
          pMeshDataId newTagPeriodic= MD_lookupMeshDataId("PeriodicVertexID");
    pGModel model = M_model(m);
    const int nt = model ? GM_numTransformations(model) : 0;
          double xyz0[3],xyz1[3];
    const double tol = 1e-10;
          if (nt) {
      fprintf(f,"$Periodic\n");
      fprintf(f,"%d\n",nt);
      GTIter titer = GM_transformationIter(model);
      pGEntity pGE1,pGE2;
      while (pTransform transfo = GTIter_next(titer,pGE2,pGE1)) {
        // if (GEN_tag(pGE1) < GEN_tag(pGE2)) {
        fprintf(f,"%d %d %d\n",GEN_type(pGE1),GEN_tag(pGE1),GEN_tag(pGE2));
        double M[16];
        T_copyAffineTransformation(transfo,&(M[0]));
        fprintf(f,"Affine");
        for(int i = 0; i < 16; i++)
    fprintf(f," %.16g",M[i]);
        fprintf(f,"\n");
        // Get all the vertices belonging to the GE + closure
        int entDim = GEN_type(pGE1);
        std::vector<int> tmpMask(M_numVertices(m),0);
        std::vector<std::pair<int,int> > tmpMap;
        if (entDim == 0) {
    VIter viter = M_classifiedVertexIter(m,pGE1,false);
    while (pVertex pv = VIter_next(viter)) {
      void *temp_ptr; 
      EN_getDataPtr((pEntity) pv , newTagPeriodic, &temp_ptr);
      std::map<int,std::vector<int> >* conn = (std::map<int,std::vector<int> >*) temp_ptr;
      for (std::map<int,std::vector<int> >::iterator cIter = conn->begin();cIter!=conn->end();++cIter){
        int vId2 = cIter->first;
        pVertex v2 = M_vertex(m,vId2);
        pGEntity pGEtmp = V_whatIn(v2);
        const int id1 = EN_id(pv), id2 = EN_id(v2);
        if ((pGEtmp == pGE2) && (tmpMask[id1] == 0)){
          tmpMap.push_back(std::make_pair(id1,id2));
          tmpMask[id1] = 1;
        }
      }
    }
    VIter_delete(viter);
        } else if (entDim == 1) {
    EIter eiter = M_classifiedEdgeIter(m,pGE1,false);
    while (pEdge pe = EIter_next(eiter)) {
      for(int i = 0; i < E_numVertices(pe); i++){
        pVertex pv = E_vertex(pe,i);
        V_coord(pv,xyz0);
        void *temp_ptr; 
        EN_getDataPtr((pEntity) pv , newTagPeriodic, &temp_ptr);
        std::map<int,std::vector<int> >* conn = (std::map<int,std::vector<int> >*) temp_ptr;
        for (std::map<int,std::vector<int> >::iterator cIter = conn->begin();cIter!=conn->end();++cIter){
          int vId2 = cIter->first;
          pVertex v2 = M_vertex(m,vId2);
          V_coord(v2,xyz1);
          T_transformCoordinates(transfo,xyz1);
          double err = 0.;
          for(int l = 0; l < 3; l++)
      err += (xyz0[l] - xyz1[l])*(xyz0[l] - xyz1[l]);
          
          const int id1 = EN_id(pv), id2 = EN_id(v2);
          if ((err < tol*tol) && (tmpMask[id1] == 0)){
      tmpMap.push_back(std::make_pair(id1,id2));
      tmpMask[id1] = 1;
          }
        }
      }
    }
    EIter_delete(eiter);
        } else if (entDim == 2) {
    FIter fiter = M_classifiedFaceIter(m,pGE1,false);
    while (pFace pf = FIter_next(fiter)) {
      for(int i = 0; i < F_numVertices(pf); i++){
        pVertex pv = F_vertex(pf,i);
        V_coord(pv,xyz0);
        void *temp_ptr; 
        EN_getDataPtr((pEntity) pv , newTagPeriodic, &temp_ptr);
        std::map<int,std::vector<int> >* conn = (std::map<int,std::vector<int> >*) temp_ptr;
        for (std::map<int,std::vector<int> >::iterator cIter = conn->begin();cIter!=conn->end();++cIter){
          int vId2 = cIter->first;
          pVertex v2 = M_vertex(m,vId2);
          V_coord(v2,xyz1);
          T_transformCoordinates(transfo,xyz1);
          double err = 0.;
          for(int l = 0; l < 3; l++)
      err += (xyz0[l] - xyz1[l])*(xyz0[l] - xyz1[l]);
          
          const int id1 = EN_id(pv), id2 = EN_id(v2);
          if ((err < tol*tol) && (tmpMask[id1] == 0)){
      tmpMap.push_back(std::make_pair(id1,id2));
      tmpMask[id1] = 1;
          }
        }
      }
    }
    FIter_delete(fiter);
        }
        fprintf(f,"%d\n",(int) tmpMap.size());
        for(size_t i = 0; i < tmpMap.size(); i++) {
    fprintf(f,"%d %d\n",tmpMap[i].first,tmpMap[i].second);
        }
        // const int n = M_numClassifiedVertices(m,pGE1);
        // fprintf(f,"%d\n",n);
        // VIter viter = M_classifiedVertexIter(m,pGE1,true);
        // int nn = 0;
        // while (pVertex pv = VIter_next(viter)) {
        //  void *temp_ptr; 
        //  int isPeriodic = EN_getDataPtr((pEntity) pv , newTagPeriodic, &temp_ptr);
        //  std::map<int,std::vector<int> >* conn = (std::map<int,std::vector<int> >*) temp_ptr;
        //  for (std::map<int,std::vector<int> >::iterator cIter = conn->begin();cIter!=conn->end();++cIter){
        //    int vId2 = cIter->first;
        //    pVertex v2 = M_vertex(m,vId2);
        //    pGEntity pGEtmp = V_whatIn(v2);
        //    if (pGEtmp == pGE2){
        //      nn++;
        //      fprintf(f,"%d %d\n",EN_id(pv),EN_id(v2));
        //    }
        //  }
        // }
      }
      GTIter_delete(titer);
      fprintf(f,"$EndPeriodic\n");
    }
        }
      }
      fclose(f);
      //MD_deleteMeshDataId(tagPeriodicPoint);

    }


    void SaveGmshMeshPer (const pMesh m,const char *filename,
                          MDB_DataExchangerPeriodic &deperiodic, int version)
    {
      int nbModelVertex = 0;
      pMeshDataId tagPeriodicPoint = MD_lookupMeshDataId("PeriodicPoint");
      // Create a reverse map of physical tags
      GEntity2Physical gentity2phys(m->geomFeatures_Tags);
      FILE *f = fopen(filename, "w");
      {
        fprintf(f, "$NOD\n");
        fprintf(f, "%d\n", m->nbPoints);
        VIter vit = M_vertexIter(m); 
        while (VIter_next(vit)){}
        VIter_reset(vit);
        pVertex pv;  
        int NN = 0;
        while ((pv = VIter_next(vit)))
          {
            if(pv->g)
              {
                NN++;
                int dim = GEN_type(pv->g); 
                if (pv->deleted)printf("ouuch\n");
                if(dim == 0)
                  nbModelVertex++;
                void *temp_ptr; 
                int isPeriodic = EN_getDataPtr((pEntity) pv , tagPeriodicPoint, &temp_ptr);
                if(pv->X==2.25 && pv->Y==0.25) printf("%e %e %e : %d\n",pv->X,pv->Y,pv->Z,pv->iD);
                if(pv->X==2.75 && pv->Y==0.25) printf("%e %e %e : %d\n",pv->X,pv->Y,pv->Z,pv->iD);
                if(isPeriodic) {
                  std::vector<std::pair<std::vector<int> , pVertex> > *recup = (
                                                                                std::vector<std::pair<std::vector<int> , pVertex> > *) temp_ptr;
                          
                  // the use of allocation is done due to ms visual c++ compiler 
                  // that due not support c99 standard (it uses the c95 et c++98 standards)
                  int* ntrans = new int[deperiodic.nbRefPeriodic()];
                  for(int kk = 0 ; kk<deperiodic.nbRefPeriodic() ; kk++) {
                    ntrans[kk] = 0;
                  }
                  int nump = EN_id((pEntity) pv);
                  pVertex pnump = pv;
                  for(unsigned int k=0 ; k<(*recup).size() ; k++) {
                    if ( nump > EN_id((pEntity) (*recup)[k].second) ) {
                      nump  = EN_id((pEntity) (*recup)[k].second);
                      pnump =  (*recup)[k].second;
                      for(unsigned int j=0 ; j<(*recup)[k].first.size() ; j++) ntrans[j] = (-1) * (*recup)[k].first[j];
                    }             
                  }
                  if (nump == EN_id((pEntity) pv)) {  
                    fprintf(f, "%d %.16g %.16g %.16g\n", pv->iD, pv->X, pv->Y, pv->Z);          
                  } else {
                    double X,Y,Z;
                    X = pnump->X;
                    Y = pnump->Y;
                    Z = pnump->Z; 
                    for(int kk = 0 ; kk<deperiodic.nbRefPeriodic() ; kk++) {
                      if(ntrans[kk]){
                        int inv1 = (ntrans[kk] < 0) ? 1 : 0;
                        for(int nb = 0 ; nb<abs(ntrans[kk]) ; nb++) { //nb de fois qu'il faut appliquer la transfo
                          deperiodic.fperiodic(inv1,X,Y,Z,kk+1,&X,&Y,&Z);
                        }  
                      }
                    }
                    fprintf(f, "%d %.16g %.16g %.16g\n", pv->iD, X, Y, Z);    
                    if((fabs(X-pv->X) > 1e-12) || (fabs(Y-pv->Y) > 1e-12) || (fabs(Z-pv->Z) > 1e-12)  ) {
                      printf("point %d : %e %e %e \n",EN_id((pEntity) pv),pv->X, pv->Y, pv->Z);
                      printf("-- corresp %d : %e %e %e\n",nump,pnump->X,pnump->Y,pnump->Z);
                      printf("*end* %e %e %e\n",X,Y,Z);       
                    }
                  }
                  delete [] ntrans;
                } else {
                  fprintf(f, "%d %.16g %.16g %.16g\n", pv->iD, pv->X, pv->Y, pv->Z);
                }
              }
            else
              {
                NN++;
                fprintf(f, "%d %.16g %.16g %.16g\n", pv->iD, pv->X, pv->Y, pv->Z);
              }
            //     throw;
          }
        if (NN != m->nbPoints) 
          {
            printf("%d != %d\n",NN,m->nbPoints);
            throw;
          }
        VIter_delete(vit);    
        fprintf(f, "$ENDNOD\n");
      }
      {
        fprintf(f, "$ELM\n");
    
        int nbClasEdges = 0;
        int nbClasFaces = 0;
        {
          EIter eit = M_edgeIter(m);
          pEdge pe;  
          while ((pe = EIter_next(eit)))
            {
              int dim = GEN_type(pe->g); 
              if(dim == 1)nbClasEdges++;
            }
          EIter_delete(eit);
        }    
        {
          FIter fit = M_faceIter(m);
          pFace pf;  
          while ((pf = FIter_next(fit)))
            {
              int dim = GEN_type(pf->g); 
              if(dim == 2)nbClasFaces++;
              else if(F_numRegions(pf)==1) nbClasFaces++;
            }
          FIter_delete(fit);
        } 

        fprintf(f, "%ld\n", (long int)(nbClasEdges + nbModelVertex + nbClasFaces + m->nbTets) );

        int k = 1;
        {
          VIter vit = M_vertexIter(m);
          pVertex pv;  
          while ((pv = VIter_next(vit)))
            {
              if(pv->g)
                {
                  int dim = GEN_type(pv->g); 
                  int tag = GEN_tag (pv->g); 
                  int phys = gentity2phys.get_first_tag(pv->g);
        
                  if(dim == 0)
                    fprintf(f, "%d %d %d %d %d %d\n",
                            k++, 15, phys,tag, 1,pv->iD);
                }
            }
          VIter_delete(vit);    
        }
        {      
          EIter eit = M_edgeIter(m);
          pEdge pe;  
          while ((pe = EIter_next(eit)))
            {
              if(pe->g)
                {
                  int dim = GEN_type(pe->g); 
                  int tag = GEN_tag (pe->g); 
                  int phys = gentity2phys.get_first_tag(pe->g);
        
                  if(dim ==1)
                    fprintf(f, "%d %d %d %d %d %d %d\n", k++, 1, phys,tag, 2,pe->p1->iD, pe->p2->iD);
                }
            }
          EIter_delete(eit);
        }
        {
          FIter fit = M_faceIter(m);
          pFace pf;  
          while ((pf = FIter_next(fit)))
            {
              if(pf->g)
                {
                  MDB_Point *nod[4];
                  int dim = GEN_type(pf->g); 
                  int tag = GEN_tag (pf->g); 
                  int phys = gentity2phys.get_first_tag(pf->g);
        
                  pf->getNodes(nod);
                  if(dim == 2 || F_numRegions(pf)==1){
                    if (0 && nod[3]) //quad
                      fprintf(f, "%d %d %d %d %d %d %d %d %d\n",
                              k++, 3, phys,tag, 4,
                              nod[0]->iD, nod[1]->iD, nod[2]->iD, nod[3]->iD);
                    else //triangle
                      fprintf(f, "%d %d %d %d %d %d %d %d\n",
                              k++, 2, phys,tag, 3,
                              nod[0]->iD, nod[1]->iD, nod[2]->iD);

                  }
                }
            }
          FIter_delete(fit);
        } 
        {
          RIter rit = M_regionIter(m);
          pRegion pr;  
          while ((pr = RIter_next(rit)))
            {
              if(pr->g)
                {
                  MDB_Point *nod[8];
                  //int dim = GEN_type(pr->g); 
                  int tag = GEN_tag (pr->g);  
                  int phys = gentity2phys.get_first_tag(pr->g);
        
                  int numVer =  pr->getNbVertex();
                  pPList ll = R_vertices(pr);
                  for(int i=0;i<numVer;i++) nod[i] = (pVertex)PList_item(ll, i);
                  PList_delete(ll);
              
                  if(numVer==4) fprintf(f, "%d %d %d %d %d %d %d %d %d\n", k++, 4, phys,tag, 4,
                                        nod[0]->iD, nod[1]->iD, nod[2]->iD, nod[3]->iD);
                  else if(numVer==8) fprintf(f, "%d %d %d %d %d %d %d %d %d %d %d %d %d\n", k++, 5, phys,tag, 5,
                                             nod[0]->iD, nod[1]->iD, nod[2]->iD, nod[3]->iD, nod[4]->iD, nod[5]->iD, nod[6]->iD, nod[7]->iD);
                }
            }
          RIter_delete(rit);
        }
        fprintf(f, "$ENDELM\n");
        {
          int nperiodic = 0,nref = 0;;
          VIter vit = M_vertexIter(m);
          pVertex pv;  
          while ((pv = VIter_next(vit)))
            {
              void *temp_ptr; 
              int isPeriodic = EN_getDataPtr((pEntity) pv , tagPeriodicPoint, &temp_ptr);
              if(isPeriodic) {
                std::vector<std::pair<std::vector<int> , pVertex> > *recup = (std::vector<std::pair<std::vector<int> , pVertex> > *) temp_ptr;
                if(!nperiodic) {
                  nref = (int)((*recup)[0].first.size());
                }
                nperiodic+=(int)((*recup).size());
              }
            }
          VIter_delete(vit);  
          if(nperiodic) {
            fprintf(f, "$PERIODICNOD\n");
            fprintf(f, "%d %d\n", nperiodic,nref);
            int num = 1;
            vit = M_vertexIter(m);
            while ((pv = VIter_next(vit)))
              {
                void *temp_ptr; 
                int isPeriodic = EN_getDataPtr((pEntity) pv , tagPeriodicPoint, &temp_ptr);
                if(isPeriodic) {
                  std::vector<std::pair<std::vector<int> , pVertex> > *recup = (std::vector<std::pair<std::vector<int> , pVertex> > *) temp_ptr;
                  for(unsigned int rk=0 ; rk<(*recup).size() ; rk++) {
                    fprintf(f, "%d", num++);
                    for(unsigned int j=0 ; j<(*recup)[rk].first.size() ; j++) fprintf(f, " %d", (*recup)[rk].first[j]);
                    fprintf(f, " %d %d \n", EN_id((pEntity) pv),EN_id((pEntity) (*recup)[rk].second));
                  }
                }
              }
            VIter_delete(vit);
            fprintf(f, "$ENDPERIODICNOD\n");  
          }
        }  
      }
      fclose(f);
    }

  }


  
  //}
