// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#include "MeshDataBase.h"
#include "MeshDataBaseInterface.h"
#include "MeshDataBaseParallelInterface.h"
#include "MeshDataBaseIO.h"
#include "MeshDataBaseGEntity2Physical.h"
#include "MshTags.h"
#include "MAdMessage.h"
#include "MAdResourceManager.h"

#include <cstdlib>
#include <cstring>
#include <set>
#include <map>
#include <utility>
#include <cassert>
#include <string>
#include <sstream>
#include <fstream>

#ifdef PARALLEL
#include "mpi.h"
#include "autopack.h"
#endif

#include "MeshDataBaseParallelInterface.h"
#include "MAdMatrix.h"
#include "ModelInterface.h"
#include "TransformationInterface.h"
#include "MAdTransformation.h"

using namespace std;

namespace MAd {

  bool insertHOPoints(pEdge pe,pVertex v0,pVertex v1,
                      std::vector<pVertex>& points) {
    
    if (E_vertex(pe,0) == v0 && E_vertex(pe,1) == v1) {
      for (int i=0;i<E_numPoints(pe);i++) points.push_back(E_point(pe,i));
      return true;
    }

    if (E_vertex(pe,1) == v0 && E_vertex(pe,0) == v1) {
      for (int i=E_numPoints(pe)-1;i>=0;i--) points.push_back(E_point(pe,i));
      return true;
    }


    return false;
  }

  bool insertHOPoints(pFace pf,
                     pVertex v0,pVertex v1,pVertex v2,pVertex v3,
                     std::vector<pVertex>& points) {
    

    int order = EN_order((pEntity) pf);
    int nbVtx = EN_numVertices((pEntity) pf);
    
    if (order == 1) return true;
    if (order > 4) throw;

    if  (nbVtx == 3 && order == 2) return true;
    if ((nbVtx == 3 && order == 3) || (nbVtx == 4 && order == 2)) {
      points.push_back(F_point(pf,0));
      return true;
    }
    
    int rot = 0;
    for (;rot<nbVtx;rot++) if (v0 == F_vertex(pf,rot)) break;
    if (rot == nbVtx) return false;
    bool swap = true;
    if (v1 == F_vertex(pf,(rot+1)%nbVtx)) swap = false;
    
    if (swap) 
      for (int i=0;i<nbVtx;i++) points.push_back(F_point(pf,(nbVtx+rot-i)%nbVtx));
    else
      for (int i=0;i<nbVtx;i++) points.push_back(F_point(pf,(i+rot)%nbVtx));
    
    if ((nbVtx == 3 && order == 4) || (nbVtx == 4 && order == 3)) return true;

    if (swap)
      for (int i=0;i<nbVtx;i++) points.push_back(F_point(pf,4+(4-i+rot)%nbVtx));
    else 
      for (int i=0;i<nbVtx;i++) points.push_back(F_point(pf,4+(i+rot)%nbVtx));
    
    if ((nbVtx == 3 && order == 5)) return true;
    
    points.push_back(F_point(pf,8));
    return true;
  }

  void SaveGmshMeshHighOrder(const pMesh mesh,const char *filename, bool newNum) {
    
    std::ofstream gmsh(filename);
    gmsh.precision(16);
    
    const MAdMsg& msg = MAdMsgSgl::instance();
      
    double scale = mesh->getScale();
    double overScale = 1./scale;
    
    // --- count classified elements
      
    long int nbElmts = 0;
    
    VIter vIter = M_vertexIter(mesh);
    while (pVertex pv=VIter_next(vIter)) if (GEN_type(pv->g)==0) nbElmts++;

    EIter eIter = M_edgeIter(mesh);
    while (pEdge pe=EIter_next(eIter))   if (GEN_type(pe->g)==1) nbElmts++;

    FIter fIter = M_faceIter(mesh);
    while (pFace pf=FIter_next(fIter))   if (GEN_type(pf->g)==2) nbElmts++;
      
    RIter rIter = M_regionIter(mesh);
    while (pRegion pr=RIter_next(rIter)) if (GEN_type(pr->g)==3) nbElmts++;
    
    gmsh << "$MeshFormat\n2.1 0 8\n$EndMeshFormat\n";
    
    // --- Writing the points -------------------------------------------------
      
    // --- non parametric version 

    if (!(mesh->isParametric())) {
      gmsh <<  "$Nodes\n";
      gmsh << M_numVertices(mesh) + M_numPoints(mesh) << "\n";
      VIter_reset(vIter);
      while (pVertex pv = VIter_next(vIter)) {
        gmsh << EN_id((pEntity) pv) << " " 
             << pv->X*overScale << " " 
             << pv->Y*overScale << " " 
             << pv->Z*overScale << "\n";
      }
      VIter pIter = M_pointIter(mesh);
      while (pVertex pv = VIter_next(pIter)) {
        gmsh << EN_id((pEntity) pv) << " " 
             << pv->X*overScale << " " 
             << pv->Y*overScale << " " 
             << pv->Z*overScale << "\n";
      } 
      gmsh << "$EndNodes\n";             
      VIter_delete(pIter);
    }
      
    // --- parametric nodes                       
      
    else {
      gmsh << "$ParametricNodes\n";
      gmsh << M_numVertices(mesh) + M_numPoints(mesh) << "\n";
      VIter_reset(vIter);
      while (pVertex pv = VIter_next(vIter)) {
        if(pv->g) {
          int gdim = GEN_type(pv->g);
          int gtag = GEN_tag(pv->g);
          gmsh << pv->iD << " " 
               << pv->X*overScale << " " 
               << pv->Y*overScale << " " 
               << pv->Z*overScale << " " 
               << gdim << " " << gtag;
          double u[2]; V_params(pv,&u[0],&u[1]);
          switch (gdim) {
          case 0: { break;}
          case 1: { gmsh << " " << u[0]; break;}
          case 2: { gmsh << " " << u[0] << " " << u[1]; break;}
          case 3: { break;}
          }
          gmsh << "\n";
        }
        else {
          msg.error(__LINE__,__FILE__,
                    "Parametric point %i has no classification",pv->iD);
        }
      } 
      
      VIter pIter = M_pointIter(mesh);
      while (pVertex pv = VIter_next(pIter)) {
        if(pv->g) {
          int gdim = GEN_type(pv->g);
          int gtag = GEN_tag(pv->g);
          gmsh << pv->iD << " " 
               << pv->X*overScale << " " 
               << pv->Y*overScale << " " 
               << pv->Z*overScale << " " 
               << gdim << " " << gtag;
          double u[2]; V_params(pv,&u[0],&u[1]);
          switch (gdim) {
          case 0: { break;}
          case 1: { gmsh << " " << u[0]; break;}
          case 2: { gmsh << " " << u[0] << " " << u[1]; break;}
          case 3: { break;}
          }
          gmsh << "\n";
        }
        else {
          msg.error(__LINE__,__FILE__,
                    "Parametric point %i has no classification",pv->iD);
        }
      }
      gmsh << "$EndParametricNodes\n";
      VIter_delete(pIter);
    }
      
    // --- writing the elements ----------------------------------------------

    gmsh << "$Elements\n" << nbElmts << "\n";

    int elmtId = 1;

    // --- write nodes

    VIter_reset(vIter);
    while (pVertex pv = VIter_next(vIter)) {
      if (GEN_type(pv->g) == 0) {
        
        int type = EN_mshTag((pEntity) pv);
        int elmtTag = GEN_tag(pv->g);
        int physTag = GEN_physTag(pv->g);

        int id = newNum ? elmtId++ : EN_id((pEntity) pv);
        
        gmsh << id << " " << type << " " << 2 << " "
             << (physTag ? physTag:elmtTag) << " " << elmtTag << " " 
             << EN_id((pEntity) pv) << "\n";
      }
    }
    
    // --- write edges 
      
    EIter_reset(eIter);
    while (pEdge pe = EIter_next(eIter)) {
      if (GEN_type(pe->g) == 1) {
        
        int type = EN_mshTag((pEntity) pe);
        int elmtTag = GEN_tag(pe->g);
        int physTag = GEN_physTag(pe->g);
        
        int id = newNum ? elmtId++ : EN_id((pEntity) pe);

        gmsh << id << " " << type << " " << 2 << " "
             << (physTag ? physTag:elmtTag) << " " << elmtTag << " "
             << EN_id((pEntity) pe->p1) << " " 
             << EN_id((pEntity) pe->p2);
        for (int i=0;i<E_numPoints(pe);i++) {
          gmsh << " " << EN_id((pEntity) E_point(pe,i));
        }
        gmsh << "\n";
      }
    }
      
    // --- write faces 
    
    FIter_reset(fIter);
    while (pFace pf = FIter_next(fIter)) {
      
      
      if (GEN_type(pf->g) == 2) {
      
        int type = EN_mshTag((pEntity) pf);
        int elmtTag = GEN_tag(pf->g);
        int physTag = GEN_physTag(pf->g); 

        int id = newNum ? elmtId++ : EN_id((pEntity) pf);

        gmsh << id << " " << type << " " << 2 << " "
             << (physTag ? physTag:elmtTag) << " " << elmtTag;
          
        std::vector<pVertex> pts;

        // --- main vertices

        for (int i=0;i<F_numVertices(pf);i++) pts.push_back(F_vertex(pf,i));
        
        // --- edge nodes 
        
        for (int i=0;i<F_numEdges(pf);i++) {
          int iEdge;
          const int* idx = getMshEdgePoints(type,i,iEdge);
          pVertex v0 = pf->getVertex(idx[0]);
          pVertex v1 = pf->getVertex(idx[1]);
          pEdge pe = pf->getEdge(iEdge);
          insertHOPoints(pe,v0,v1,pts);
        }
        
        // --- face nodes 
        
        for (int i=0;i<pf->getNbHighOrderPoints();i++) {
          pts.push_back(pf->getHighOrderPoint(i));
        }
        
        // --- 
        
        std::vector<pVertex>::iterator pIter = pts.begin();
        for (;pIter != pts.end();pIter++) gmsh << " " << (*pIter)->iD;

        // ---
        
        gmsh << "\n";
      }
    }
    
    // --- write regions        

    RIter_reset(rIter);
    while (pRegion pr = RIter_next(rIter)) {
      
      int type = EN_mshTag((pEntity) pr);
      int elmtTag = GEN_tag(pr->g);
      int physTag = GEN_physTag(pr->g); 

      int id = newNum ? elmtId++ : EN_id((pEntity) pr);

      gmsh << id << " " << type << " " << 2 << " "
           << (physTag ? physTag:elmtTag) << " " << elmtTag;
      
      std::vector<pVertex> pts;

      // --- principal vertices
      
      for (int i=0;i<R_numVertices(pr);i++) pts.push_back(R_vertex(pr,i));
      
      // --- edge nodes
      
      for (int iEdge=0;iEdge<R_numEdges(pr);iEdge++) {
        int idx(-1);
        const int* node = getMshEdgePoints(type,iEdge,idx);
        pVertex v0 = pr->getVertex(node[0]);
        pVertex v1 = pr->getVertex(node[1]);
        pEdge pe = pr->getEdge(idx);
        insertHOPoints(pe,v0,v1,pts);
      }

      // --- face nodes

      for (int iFace=0;iFace<R_numFaces(pr);iFace++) {
        int idx(-1);
        const int* node = getMshFacePoints(type,iFace,idx);
        pVertex v0 = (node[0]>=0) ? pr->getVertex(node[0]) : NULL;
        pVertex v1 = (node[1]>=0) ? pr->getVertex(node[1]) : NULL;
        pVertex v2 = (node[2]>=0) ? pr->getVertex(node[2]) : NULL;
        pVertex v3 = (node[3]>=0) ? pr->getVertex(node[3]) : NULL;
        pFace pf = pr->getFace(idx);
        insertHOPoints(pf,v0,v1,v2,v3,pts);
      }
      
      // --- internal nodes

      for (int i=0;i<pr->getNbHighOrderPoints();i++) {
        pts.push_back(pr->getHighOrderPoint(i));
      }
      
      // --- now write
      
      std::vector<pVertex>::iterator pIter = pts.begin();
      for (;pIter!=pts.end();++pIter) gmsh << " " << (*pIter)->iD;
      
      gmsh << "\n";
    }
      
    // --- close element region 
      
    gmsh <<  "$EndElements\n"; 
      
    // --- write periodic node section ----------------------------------------
    

    pGModel model = M_model(mesh);
    pMeshDataId periodicTag = MD_lookupMeshDataId("PeriodicVertexID");
    int nbTransformations = GM_numTransformations(model);    
    nbTransformations = 0;
    if (nbTransformations) {  

      gmsh << "$Periodic\n";
      gmsh << nbTransformations << "\n";
      
      GTIter gtIter = GM_transformationIter(model);

      pGEntity ge1;
      pGEntity ge2;
      
      
      while (pTransform gt = GTIter_next(gtIter,ge1,ge2)) {
        
        double mat[16];
        T_copyAffineTransformation(gt,mat);

        gmsh << GEN_type(ge1) << " " 
             << GEN_tag(ge1)  << " " 
             << GEN_tag(ge2)  << "\n";
        
        gmsh << "Affine";
        for (int i=0;i<16;i++) gmsh << " " << mat[i];
        gmsh << "\n";

        VIter pvIter = M_classifiedVertexIter(mesh,ge1,false);
        
        std::map<int,int> vtxPairs;
        
        while (pVertex pv1 = VIter_next(pvIter)) {
          void* tmp_periodic;
          if (EN_getDataPtr((pEntity) pv1,periodicTag,&tmp_periodic)) {
            std::map<int,std::vector<int> > *  conn = 
              (std::map<int,std::vector<int> >*) (tmp_periodic);
            std::map<int,std::vector<int> >::iterator cIter = conn->begin();
            for (;cIter!=conn->end();cIter++) {
              int id2 = cIter->first;
              pVertex pv2 = M_vertex(mesh,id2);
              pGEntity pg2 = EN_whatIn((pEntity) pv2);
              if (pg2 == ge2) {
                vtxPairs[EN_id(pv1)] = EN_id(pv2);
                break;
              }
            }
          }
        }
        
        gmsh << vtxPairs.size() << std::endl;
        
        for (std::map<int,int>::iterator vtpIter=vtxPairs.begin();
             vtpIter != vtxPairs.end();++vtpIter) {
          gmsh << vtpIter->first << " " << vtpIter->second << "\n";
        }
      }
      GTIter_delete(gtIter);
      gmsh << "$EndPeriodic\n";
    }
    // else {

    //   VIter_reset(vIter);
    //   std::map<int,std::set<int> > corr;
    //   std::set<int> treated;

    //   while (pVertex pv = VIter_next(vIter)) {
    //     void* tmp(NULL);
    //     if (EN_getDataPtr((pEntity) pv,periodicTag,&tmp)) {

    //       int id = EN_id((pEntity) pv);
    //       if (treated.find(id) == treated.end()) {
            
    //         std::map<int,std::vector<int> > *  conn = 
    //           (std::map<int,std::vector<int> >*) (tmp);
            
    //         std::map<int,std::vector<int> >::iterator connIter = conn->begin();
            
    //         for (;connIter!=conn->end();connIter++) {
    //           int idc = connIter->first;
    //           corr[id].insert(idc);
    //           treated.insert(idc);
    //         }
    //       }
    //     }
    //   }

    //   gmsh << "$PeriodicNodes\n";
    //   gmsh << corr.size() << "\n";
    //   std::map<int,std::set<int> >::iterator corrIter = corr.begin();
    //   for (;corrIter!=corr.end();++corrIter) {

    //     std::set<int>& copies = corrIter->second;
        
    //     gmsh << copies.size() + 1 << " " << corrIter->first;
    //     for (std::set<int>::iterator copyIter = copies.begin();
    //          copyIter!=copies.end();++copyIter) gmsh << " " << *copyIter;
    //     gmsh << " 0\n";
    //   }
    //   gmsh << "$EndPeriodicNodes\n";
    // }
    
    

    // --- delete the iterators 
    
    VIter_delete(vIter);
    EIter_delete(eIter);
    FIter_delete(fIter);
    RIter_delete(rIter);

    // --- close the file 
    
    gmsh.close();
  }
}
