// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#ifndef H_MESHDATABASEINTERFACE
#define H_MESHDATABASEINTERFACE

#include "MAdExport.h"
#include "ModelInterface.h"

#include "MshTags.h"

#ifdef PARALLEL
#include "mpi.h"
#endif

/*! \defgroup api       Application Programming Interface (API) */

/*! \defgroup api_mesh  Mesh and element operations and objects 
  \ingroup api */

/*! \defgroup api_mesh_mesh      Mesh operations 
  \ingroup api_mesh */
/*! \defgroup api_mesh_entity    Entity operations 
  \ingroup api_mesh */
/*! \defgroup api_mesh_region    Region operations 
  \ingroup api_mesh */
/*! \defgroup api_mesh_face      Face operations 
  \ingroup api_mesh */
/*! \defgroup api_mesh_edge      Edge operations 
  \ingroup api_mesh */
/*! \defgroup api_mesh_vertex    Vertex operations 
  \ingroup api_mesh */
/*! \defgroup api_mesh_point     Point operations 
  \ingroup api_mesh */
/*! \defgroup api_mesh_ios       Mesh in- and output 
  \ingroup api_mesh */
/*! \defgroup api_mesh_parallel  Communication 
  \ingroup api_mesh */
/*! \defgroup api_mesh_tags      Element tags and types
  \ingroup api_mesh */

namespace MAd {
  
  /**************************/
  /*  Forward declarations  */
  /**************************/

  class MDB_DataExchangerPeriodic;

  class MDB_Mesh;  

  class MDB_MeshEntity;
  class MDB_Region;
  class MDB_Face;
  class MDB_Edge;
  class MDB_Point;
  
  class MDB_RegionIter;
  class MDB_FaceIter;
  class MDB_EIter;
  class MDB_VIter;

  class PList;

  /**********************/
  /*  Type definitions  */
  /**********************/
  
  typedef unsigned int pMeshDataId;

  typedef class MDB_Mesh  * pMesh;  

  typedef class MDB_MeshEntity * pEntity;
  typedef class MDB_Region     * pRegion;
  typedef class MDB_Face       * pFace;
  typedef class MDB_Edge       * pEdge;
  typedef class MDB_Point      * pVertex;
  typedef class MDB_Point      * pPoint;
  
  typedef class MDB_RegionIter * RIter;
  typedef class MDB_FaceIter   * FIter;
  typedef class MDB_EIter      * EIter;
  typedef class MDB_VIter      * VIter;

  typedef class PList * pPList;
  
  /********************/
  /*  Mesh Operators  */
  /********************/

#ifdef PARALLEL
  MAD_EXPORT pMesh   M_new(pGModel, MPI_Comm mpi_comm=MPI_COMM_WORLD);
#else
  MAD_EXPORT pMesh   M_new(pGModel);
#endif
  MAD_EXPORT void    M_delete(pMesh);
  MAD_EXPORT void    M_load(pMesh, const char * filename, int rank=-1);
  MAD_EXPORT void    M_clean(pMesh);
  MAD_EXPORT bool    M_setModel(pMesh,pGModel);
  MAD_EXPORT const char* M_getName(pMesh);
  MAD_EXPORT bool    M_setName(pMesh,const char*);
  
  MAD_EXPORT void    M_scale(pMesh,double);

  MAD_EXPORT void    M_shrink(pMesh);
  MAD_EXPORT void    M_expand(pMesh);

  MAD_EXPORT pGModel M_model(pMesh);
  MAD_EXPORT int     M_dim(pMesh);

  MAD_EXPORT int     M_edgesMaxOrder(pMesh);
  MAD_EXPORT int     M_facesMaxOrder(pMesh);
  MAD_EXPORT int     M_regionsMaxOrder(pMesh);
  MAD_EXPORT int     M_maxOrder(pMesh);

  MAD_EXPORT bool    M_isParametric(pMesh);

  MAD_EXPORT void    M_writeMsh(const pMesh, const char *name, int version=2, const int * partitionTable=NULL);
  MAD_EXPORT void    M_writeMshPer(const pMesh, const char *name, MDB_DataExchangerPeriodic &, int version);

  /*! \brief write the bounding box information per partition and the periodic connections */
  /*! Returns false if file could not be opened */
  MAD_EXPORT bool    M_writeMshBoundingBox(const pMesh,const char* name,bool writeAll);

  MAD_EXPORT bool    M_readMshBoundingBox(const pMesh,const char* name,int part);

#ifdef _HAVE_METIS_
  MAD_EXPORT void    M_Partition(pMesh, int nbParts, const char * filename);
#endif
  
  MAD_EXPORT void M_box(pMesh, double[3][2]);
  MAD_EXPORT void M_box(pMesh, pTransform, double[3][2]);
  MAD_EXPORT void M_transformableBox(pMesh,double[8][3]);
  MAD_EXPORT bool M_boxContains(pMesh, double[3], double);
  MAD_EXPORT bool M_boxContains(pMesh, double[3][2], double);

  MAD_EXPORT double M_getScale(pMesh);

  MAD_EXPORT int M_numRegions(pMesh);
  MAD_EXPORT int M_numTets(pMesh);
  MAD_EXPORT int M_numHexes(pMesh);
  MAD_EXPORT int M_numPrisms(pMesh);
  MAD_EXPORT int M_numPyramids(pMesh);
  MAD_EXPORT int M_numFaces(pMesh);
  MAD_EXPORT int M_numQuads(pMesh);
  MAD_EXPORT int M_numTriangles(pMesh);
  MAD_EXPORT int M_numEdges(pMesh);
  MAD_EXPORT int M_numVertices(pMesh);
  MAD_EXPORT int M_numPoints(pMesh);
  MAD_EXPORT int M_numPartitions(pMesh);

  MAD_EXPORT RIter M_regionIter(pMesh);
  MAD_EXPORT FIter M_faceIter(pMesh);
  MAD_EXPORT EIter M_edgeIter(pMesh);
  MAD_EXPORT VIter M_vertexIter(pMesh);
  MAD_EXPORT VIter M_pointIter(pMesh);

  MAD_EXPORT int M_numClassifiedRegions (pMesh, pGEntity);
  MAD_EXPORT int M_numClassifiedFaces   (pMesh, pGEntity);
  MAD_EXPORT int M_numClassifiedEdges   (pMesh, pGEntity);
  MAD_EXPORT int M_numClassifiedVertices(pMesh, pGEntity);

  MAD_EXPORT RIter   M_classifiedRegionIter(pMesh, pGEntity);
  MAD_EXPORT FIter   M_classifiedFaceIter  (pMesh, pGEntity, int closure);
  MAD_EXPORT EIter   M_classifiedEdgeIter  (pMesh, pGEntity, int closure);
  MAD_EXPORT VIter   M_classifiedVertexIter(pMesh, pGEntity, int closure);
  MAD_EXPORT pVertex M_classifiedVertex    (pMesh, pGVertex);

  MAD_EXPORT void M_classifyEntities(pMesh);

  /* mesh entity creation routines */
  MAD_EXPORT pRegion M_createR(pMesh, int nFace, pFace   faces[], pGEntity pg=NULL);
  MAD_EXPORT pRegion M_createR(pMesh, int nVert, pVertex verts[], pGEntity pg=NULL);
  MAD_EXPORT pRegion M_createR(pMesh, int nVert, int   vertIds[], pGEntity pg=NULL);
  MAD_EXPORT pRegion M_createTet(pMesh, int order, bool serendip, int vertIds[], pGEntity pg=NULL);
  MAD_EXPORT pFace   M_createF(pMesh, int nEdge, pEdge   edges[], pGEntity pg=NULL);
  MAD_EXPORT pFace   M_createF(pMesh, int nVert, pVertex verts[], pGEntity pg=NULL);
  MAD_EXPORT pFace   M_createF(pMesh, int nVert, int   vertIds[], pGEntity pg=NULL);
  MAD_EXPORT pFace   M_createTri(pMesh, int order, int vertIds[], pGEntity pg=NULL);
  MAD_EXPORT pFace   M_createQuad(pMesh, int order, int vertIds[], pGEntity pg=NULL);
  MAD_EXPORT pEdge   M_createE(pMesh, pVertex v1, pVertex v2, pGEntity pg=NULL);
  MAD_EXPORT pEdge   M_createE(pMesh, int   vId1, int   vId2, pGEntity pg=NULL);
  MAD_EXPORT pEdge   M_createE(pMesh, pVertex v1, pVertex v2, int order=1,
                    pVertex pts[]=NULL, pGEntity pg=NULL);
  MAD_EXPORT pVertex M_createV(pMesh, double x, double y, double z, int patch, pGEntity pg=NULL);
  MAD_EXPORT pVertex M_createV2(pMesh, double xyz[3], int patch, pGEntity pg=NULL);
  MAD_EXPORT pVertex M_createVP(pMesh, double x, double y, double z,
                     double u, double v, int patch, pGEntity pg=NULL);

  /* mesh entity deletion routines */
  MAD_EXPORT void M_removeRegion(pMesh, pRegion);
  MAD_EXPORT void M_removeFace  (pMesh, pFace);
  MAD_EXPORT void M_removeEdge  (pMesh, pEdge);
  MAD_EXPORT void M_removeVertex(pMesh, pVertex);
  MAD_EXPORT void P_delete(pPoint);

  /* extra access to entities */
  MAD_EXPORT pRegion M_region(pMesh, int []);
  MAD_EXPORT pFace   M_face  (pMesh, int []);
  MAD_EXPORT pEdge   M_edge  (pMesh, int, int);
  MAD_EXPORT pVertex M_vertex(pMesh, int);

  /* detection of extruded regions */
  
  MAD_EXPORT RIter M_extrudedFace_faceIter(pMesh,pFace,pMeshDataId);
  MAD_EXPORT FIter M_extrudedFace_regionIter(pMesh,pFace,pMeshDataId);
  MAD_EXPORT FIter M_extrudedFaceIter(pMesh,pMeshDataId);


  /***********************/
  /* Geometric model ops */
  /***********************/
  /* deprecated (rather use ModelInterface) */

  void    M_setGeomFeature    (pMesh, int, pGEntity);
  int     M_numGeomFeatures   (pMesh);
  int     M_geomFeatureId     (pMesh, int ith);
  pPGList M_geomFeature       (pMesh, int id);
  void    M_printGeomFeatures (pMesh);

  /********************/
  /* Entity Iter Ops  */
  /********************/

  MAD_EXPORT pRegion RIter_next  (RIter);
  MAD_EXPORT void    RIter_delete(RIter);
  MAD_EXPORT void    RIter_reset (RIter);
  MAD_EXPORT pFace   FIter_next  (FIter);
  MAD_EXPORT void    FIter_delete(FIter);
  MAD_EXPORT void    FIter_reset (FIter);
  MAD_EXPORT pEdge   EIter_next  (EIter);
  MAD_EXPORT void    EIter_delete(EIter);
  MAD_EXPORT void    EIter_reset (EIter);
  MAD_EXPORT pVertex VIter_next  (VIter);
  MAD_EXPORT void    VIter_delete(VIter);
  MAD_EXPORT void    VIter_reset (VIter);

  /********************/
  /* Entity Operators */
  /********************/

  //! \brief set the identity tag of the element 
  //! \warning this function does not invalidate topology, adjacencies, ... even in parallel
  //! \warning however, the identity is not updated by default on other partitions
  //! \ingroup api_mesh_entity 
  MAD_EXPORT void     EN_setID(pEntity, int id);
  
  //! \brief get the identity tag of the element 
  //! \ingroup api_mesh_entity 
  MAD_EXPORT int      EN_id   (pEntity);

  //! \brief describe the element to standard output
  //! \ingroup api_mesh_entity 
  MAD_EXPORT void     EN_print(pEntity);
  
  //! \brief returns dimension of element pe 
  //! \ingroup api_mesh_entity 
  //! \warning obsolete : use EN_dim
  MAD_EXPORT int      EN_type(pEntity); // obsolete: returns dim, use EN_dim() instead
  
  //! \brief returns dimension of element pe 
  //! \ingroup api_mesh_entity 
  MAD_EXPORT int      EN_dim(pEntity);

  //! \brief returns interpolation order of the element pe 
  //! \ingroup api_mesh_entity
  MAD_EXPORT int      EN_order(pEntity);
  
  //! \brief returns element type 
  //! \ingroup api_mesh_entity
  MAD_EXPORT int      EN_mshType(pEntity);

  //! \brief returns tag of complete element type 
  //! \ingroup api_mesh_entity
  MAD_EXPORT int      EN_mshTag(pEntity);
  
  //! \brief returns tag of corresponding linear element type  
  //! \ingroup api_mesh_entity
  MAD_EXPORT int      EN_mshBasicTag(pEntity);
  
  //! \brief returns dimension of the geometric entity on which element e is classified
  //! \ingroup api_mesh_entity
  MAD_EXPORT int      EN_whatInType(pEntity);

  //! \brief returns element classification  
  //! \ingroup api_mesh_entity
  MAD_EXPORT pGEntity EN_whatIn(pEntity);
  
  //! \brief set element classification  
  //! \ingroup api_mesh_entity
  MAD_EXPORT void     EN_setWhatIn(pEntity,pGEntity);
  
  //! \brief get number of principal vertices
  //! \ingroup api_mesh_entity
  MAD_EXPORT int      EN_numVertices(const pEntity);
  
  //! \brief get a list of principal vertices
  //! \ingroup api_mesh_entity
  MAD_EXPORT pPList   EN_vertices(const pEntity);
  
  //! \brief return element bounding box (min/max per direction)
  //! \ingroup api_mesh_entity
  MAD_EXPORT void     EN_box(const pEntity, double [3][2]);
  
  //! \brief convert safely to vertex, returns NULL if not successful 
  //! \ingroup api_mesh_entity
  MAD_EXPORT pVertex  EN_convertToVertex(pEntity);
  
  //! \brief convert safely to edge, returns NULL if not successful 
  //! \ingroup api_mesh_entity
  MAD_EXPORT pEdge    EN_convertToEdge(pEntity);
  
  //! \brief convert safely to face, returns NULL if not successful 
  //! \ingroup api_mesh_entity
  MAD_EXPORT pFace    EN_convertToFace(pEntity);
  
  //! \brief convert safely to region, returns NULL if not successful
  //! \ingroup api_mesh_entity
  MAD_EXPORT pRegion  EN_convertToRegion(pEntity);
  
  //! \brief attach an integer data d to the entity e on tag t
  //! \ingroup api_mesh_entity
  MAD_EXPORT void EN_attachDataInt(pEntity e, pMeshDataId t, int d);

  //! \brief attach a double d to the entity e on tag t
  //! \ingroup api_mesh_entity
  MAD_EXPORT void EN_attachDataDbl(pEntity e, pMeshDataId t, double d);

  //! \brief attach a general pointer d to the entity e on tag t
  //! \ingroup api_mesh_entity
  //! \warning data will not be destroyed upon destruction of the element
  MAD_EXPORT void EN_attachDataPtr(pEntity e, pMeshDataId t, void *d);
  
  //! \brief attach a general pointer d to the entity e on tag t
  //! \ingroup api_mesh_entity
  //! \warning data will not be destroyed upon destruction of the element
  MAD_EXPORT void EN_attachDataP  (pEntity e, const char * t, void *d);

  //! \brief attach a general pointer d to the entity e on tag t
  //! \ingroup api_mesh_entity
  MAD_EXPORT void EN_attachDataI  (pEntity e, const char * t, int d);

  //! \brief replace data stored on tag t on entity e by d
  //! \ingroup api_mesh_entity  
  MAD_EXPORT void EN_modifyDataInt(pEntity e, pMeshDataId t, int d);
  
  //! \brief replace double data stored on tag t on entity e by d
  //! \ingroup api_mesh_entity  
  MAD_EXPORT void EN_modifyDataDbl(pEntity e, pMeshDataId t, double d);
  
  //! \brief replace pointer data stored on tag t on entity e by d
  //! \ingroup api_mesh_entity  
  MAD_EXPORT void EN_modifyDataPtr(pEntity e, pMeshDataId t, void *d);

  // int  EN_modifyDataP  (pEntity, const char *, void *);
  // int  EN_modifyDataI  (pEntity, const char *, int);
  
  //! \brief remove data stored on tag t on entity e
  //! \ingroup api_mesh_entity  
  //! \warning does not deallocate pointer data
  MAD_EXPORT void EN_deleteData(pEntity e, pMeshDataId t);
  
  //! \brief remove data stored on tag t on entity e
  //! \ingroup api_mesh_entity  
  //! \warning does not deallocate pointer data
  MAD_EXPORT void EN_removeData(pEntity e, const char * t);
  
  //! \brief copy integer data stored on tag t on entity e to d, returns 0 if not available
  //! \ingroup api_mesh_entity  
  MAD_EXPORT int  EN_getDataInt(pEntity e, pMeshDataId t, int * d);
  
  //! \brief copy double stored on tag t on entity e to d, returns 0 if not available
  //! \ingroup api_mesh_entity  
  MAD_EXPORT int  EN_getDataDbl(pEntity e, pMeshDataId t, double *d);
  
  //! \brief copy pointer stored on tag t on entity e to d, returns 0 if not available
  //! \ingroup api_mesh_entity  
  MAD_EXPORT int  EN_getDataPtr(pEntity e, pMeshDataId t, void **d);
  
  //! \brief recuperate pointer stored on tag t on entity e, returns NULL if not available
  //! \ingroup api_mesh_entity  
  MAD_EXPORT void * EN_dataP(pEntity e, const char * t);
  
  //! \brief recuperate pointer stored on tag t on entity e, returns NULL if not available
  //! \warning undefined if not available
  //! \ingroup api_mesh_entity  
  MAD_EXPORT int    EN_dataI(pEntity e, const char * t);
  
  /********************/
  /* Region Operators */
  /********************/

  MAD_EXPORT pGRegion R_whatIn(pRegion);
  MAD_EXPORT int      R_whatInType(pRegion);
  MAD_EXPORT void     R_setWhatIn(pRegion, pGEntity);
  MAD_EXPORT void     R_forceWhatIn(pRegion region, pGEntity what);
  MAD_EXPORT int      R_numPoints(pRegion);
  MAD_EXPORT int      R_numPointsTot(pRegion);
  MAD_EXPORT pPoint   R_point(pRegion,int);
  MAD_EXPORT pPList   R_pointsTot(pRegion);
  MAD_EXPORT int      R_numVertices(pRegion);
  MAD_EXPORT pPList   R_vertices(pRegion);
  MAD_EXPORT pVertex  R_vertex(pRegion, int);
  MAD_EXPORT pVertex  R_fcOpVt(const pRegion, const pFace);
  MAD_EXPORT int      R_numEdges(pRegion);
  MAD_EXPORT pPList   R_edges(pRegion);
  MAD_EXPORT pEdge    R_edge(pRegion, int);
  MAD_EXPORT pEdge    R_gtOppEdg(const pRegion, const pEdge);
  MAD_EXPORT pFace    R_gtOppFace(const pRegion, const pFace,int&);
  MAD_EXPORT int      R_numFaces(pRegion);
  MAD_EXPORT pFace    R_face(pRegion, int);
  MAD_EXPORT pPList   R_faces(pRegion);
  MAD_EXPORT pFace    R_vtOpFc(const pRegion, const pVertex);
  MAD_EXPORT int      R_inClosure(pRegion, pEntity);
  MAD_EXPORT int      R_faceDir(pRegion, int);
  MAD_EXPORT int      R_faceOri(pRegion, int);
  MAD_EXPORT int      R_dirUsingFace(pRegion,pFace);
  MAD_EXPORT int      R_oriUsingFace(pRegion,pFace);

  MAD_EXPORT void     R_coordP1(const pRegion, double [][3]);
  MAD_EXPORT void     R_coord(const pRegion, double [][3]);
  MAD_EXPORT double   R_volume(const pRegion);
  MAD_EXPORT double   R_XYZ_volume (const double [][3]);
  MAD_EXPORT double   R_circumRad(const pRegion);
  MAD_EXPORT double   R_inscrRad(const pRegion);
  MAD_EXPORT bool     R_meanRatioCube(const pRegion, double *);
  MAD_EXPORT bool     R_XYZ_meanRatioCube(const double [][3], double *);
  MAD_EXPORT bool     R_XYZ_isFlat(const double [][3]);
  MAD_EXPORT void     R_linearParams(const pRegion, const double[3], double[3]);
  MAD_EXPORT void     R_center(const pRegion, double [3]);
  MAD_EXPORT void     R_jacobian(const pRegion, double [3][3]);
  MAD_EXPORT double   R_invJacobian(const pRegion, double [3][3]);
  MAD_EXPORT void     R_box(const pRegion, double [3][2]);
  MAD_EXPORT bool     R_inBox(const pRegion, const double [3], double);
  MAD_EXPORT bool     R_contains(const pRegion, const double [3], double);

  /********************/
  /*  Face Operators  */
  /********************/

  MAD_EXPORT pFace    F_exist(pEntity, pEntity, pEntity, pEntity);
  MAD_EXPORT pFace    F_exist(pEdge,pVertex);

  MAD_EXPORT pGEntity F_whatIn(pFace);
  MAD_EXPORT int      F_whatInType(pFace);
  MAD_EXPORT void     F_setWhatIn(pFace, pGEntity);
  MAD_EXPORT void     F_forceWhatIn(pFace face, pGEntity what);

  MAD_EXPORT pPoint   F_point(pFace, int);
  MAD_EXPORT int      F_numPoints(pFace);
  MAD_EXPORT int      F_numPointsTot(pFace);
  MAD_EXPORT int      F_numVertices(pFace);
  MAD_EXPORT pPList   F_vertices(pFace, int dir=1);
  MAD_EXPORT pVertex  F_vertex(pFace, int);
  MAD_EXPORT pEdge    F_vtOpEd(const pFace, const pVertex);
  MAD_EXPORT int      F_numEdges(pFace);
  MAD_EXPORT pPList   F_edges(pFace);
  MAD_EXPORT pEdge    F_edge(pFace, int);
  MAD_EXPORT int      F_edgeDir(pFace, int);
  MAD_EXPORT pEdge    F_findEdge(const pFace, const pVertex, const pVertex);
  MAD_EXPORT pVertex  F_edOpVt(const pFace, const pEdge);
  MAD_EXPORT int      F_dirUsingEdge(pFace, pEdge);
  MAD_EXPORT int      F_numRegions(pFace);
  MAD_EXPORT pPList   F_regions(pFace);
  MAD_EXPORT pRegion  F_region(pFace, int);
  MAD_EXPORT pRegion  F_otherRegion(const pFace, const pRegion);
  MAD_EXPORT int      F_inClosure(pFace, pEntity);
  MAD_EXPORT void     F_chDir(pFace);
  MAD_EXPORT int      F_align(pFace,pVertex,pVertex,pVertex,pVertex);
  MAD_EXPORT int      F_orientation(pFace);

  MAD_EXPORT void     F_coordP1(const pFace, double [][3]);
  MAD_EXPORT void     F_coord(const pFace, double [][3]);
  MAD_EXPORT bool     F_params(const pFace, double [][2]);
  MAD_EXPORT double   F_area(const pFace, const double *dir=NULL);
  MAD_EXPORT double   XYZ_F_area(const double [][3], const double *dir=NULL);
  MAD_EXPORT double   F_areaSq(const pFace, const double *dir=NULL);
  MAD_EXPORT double   XYZ_F_areaSq(const double [][3], const double *dir=NULL);
  MAD_EXPORT void     F_jacobian(const pFace, double [3][2]);
  MAD_EXPORT void     F_linearParams(const pFace, const double[3], double[2]);
  MAD_EXPORT void     F_center(const pFace, double [3]);
  MAD_EXPORT void     F_normal(const pFace, double [3]);
  MAD_EXPORT void     XYZ_F_normal(const double [3][3], double [3]);
  MAD_EXPORT bool     F_volumeRatio(const pFace, double *);
  MAD_EXPORT double   F_worstVolumeRatio(const pPList faces);
  MAD_EXPORT double   F_circRadius(const pFace);
  MAD_EXPORT double   F_inscRadius(const pFace);
  MAD_EXPORT void     F_box(const pFace, double [3][2]);

  MAD_EXPORT int      F_checkTransform(const pTransform,const pFace,const pFace);
  

  /********************/
  /*  Edge Operators  */
  /********************/

  MAD_EXPORT pEdge    E_exist(pVertex, pVertex);

  MAD_EXPORT pGEntity E_whatIn(pEdge);
  MAD_EXPORT int      E_whatInType(pEdge);
  MAD_EXPORT void     E_setWhatIn(pEdge, pGEntity);
  MAD_EXPORT void     E_forceWhatIn(pEdge edge, pGEntity what);

  MAD_EXPORT int      E_numPoints(pEdge);
  MAD_EXPORT pPoint   E_point(pEdge, int);
  MAD_EXPORT pVertex  E_vertex(pEdge, int);
  MAD_EXPORT int      E_numVertices(pEdge);
  MAD_EXPORT pPList   E_vertices(const pEdge);
  MAD_EXPORT pVertex  E_otherVertex(pEdge, pVertex);
  MAD_EXPORT int      E_numFaces(pEdge);
  MAD_EXPORT pPList   E_faces(const pEdge);
  MAD_EXPORT pFace    E_face(pEdge, int);
  MAD_EXPORT pFace    E_otherFace(pEdge, pFace);
  MAD_EXPORT pFace    E_otherFace(pEdge, pFace, pRegion);
  MAD_EXPORT int      E_numRegions(pEdge);
  MAD_EXPORT pPList   E_regions(pEdge);
  MAD_EXPORT int      E_inClosure(pEdge, pEntity);
  MAD_EXPORT int      E_dir(pEdge, pVertex, pVertex);
  MAD_EXPORT int      E_align(pEdge, pVertex,pVertex);

  MAD_EXPORT void     E_coordP1(const pEdge, double [][3]);
  MAD_EXPORT void     E_coord(const pEdge, double [][3]);
  MAD_EXPORT void     E_coordP1(const pEdge edge, double t, double xyz[3]);
  MAD_EXPORT bool     E_params(const pEdge, double [2][2]);
  MAD_EXPORT double   E_length(const pEdge);
  MAD_EXPORT double   E_lengthSq(const pEdge);
  MAD_EXPORT double   E_linearParams(const pEdge, const pVertex);
  MAD_EXPORT double   E_linearParams(const pEdge, const double[3]);
  MAD_EXPORT void     E_center(const pEdge, double[3]);
  MAD_EXPORT void     E_cavityCenter(const pEdge, double[3]);

  MAD_EXPORT int      E_checkTransform(const pTransform,const pEdge,const pEdge);

  /********************/
  /* Point Operators  */
  /********************/

  MAD_EXPORT double P_x(pPoint);
  MAD_EXPORT double P_y(pPoint);
  MAD_EXPORT double P_z(pPoint);
  MAD_EXPORT void   P_setPos(pPoint, double x, double y, double z);

  MAD_EXPORT void   P_setID(pPoint, int);
  MAD_EXPORT int    P_id(pPoint);

  MAD_EXPORT double P_param1(pPoint);
  MAD_EXPORT void   P_setParam1(pPoint, double);

  MAD_EXPORT bool   P_checkTransform(const pTransform,const pPoint,const pPoint);
  
  MAD_EXPORT pPoint P_transform(const pTransform,const pPoint);

  /********************/
  /* Vertex Operators */
  /********************/
  
  //! \brief return the entity on which vertex pv is classified 
  //! \ingroup api_mesh_vertex
  MAD_EXPORT pGEntity V_whatIn(pVertex pv);
  
  //! \brief return the dimension of the entity on which vertex pv is classified 
  //! \ingroup api_mesh_vertex
  MAD_EXPORT int      V_whatInType(pVertex);
  
  //! \brief set the classification of vertex pv to pg 
  //! will not modify if node already classified
  //! \ingroup api_mesh_vertex
  MAD_EXPORT void     V_setWhatIn(pVertex pv, pGEntity pg);

  //! \brief change the classification of vertex pv to pg
  //! \ingroup api_mesh_vertex
  MAD_EXPORT void     V_forceWhatIn(pVertex vert, pGEntity pg);

  MAD_EXPORT pPoint   V_point(pVertex);
  
  //! \brief return the number of edges attached to vertex pv 
  //! \ingroup api_mesh_vertex
  MAD_EXPORT int      V_numEdges(pVertex);
  
  //! \brief return the list of edges attached to vertex pv 
  //! \ingroup api_mesh_vertex
  MAD_EXPORT pPList   V_edges(pVertex);

  //! \brief return n-th edge attached to vertex pv 
  //! \ingroup api_mesh_vertex
  MAD_EXPORT pEdge    V_edge(pVertex, int);

  //! \brief return the number of faces attached to vertex pv 
  //! \ingroup api_mesh_vertex
  MAD_EXPORT int      V_numFaces(pVertex);
  
  //! \brief return the list of faces attached to vertex pv 
  //! \ingroup api_mesh_vertex
  MAD_EXPORT pPList   V_faces(pVertex);
  
  //! \brief return the number of regions attached to vertex pv 
  //! \ingroup api_mesh_vertex
  MAD_EXPORT int      V_numRegions(pVertex);
  
  //! \brief return the list of regions attached to vertex pv 
  //! \ingroup api_mesh_vertex
  MAD_EXPORT pPList   V_regions(pVertex);

  //! \brief return the identity tag of the vertex pv 
  //! \ingroup api_mesh_vertex
  MAD_EXPORT int      V_id(const pVertex pv);

  //! \brief list the coordinates of point pv in xyz
  //! \ingroup api_mesh_vertex
  MAD_EXPORT void     V_coord(const pVertex pv, double xyz[3]);
  
  //! \brief return coordinate i of vertex pv
  //! \ingroup api_mesh_vertex
  MAD_EXPORT double   V_coordComponent(const pVertex pv,int d);
  
  //! \brief Set the position of vertex p to xyz 
  //! \ingroup api_mesh_vertex
  MAD_EXPORT void     V_setCoord(pVertex pv, double xyz[3]);

  //! \brief Lists the parametric coordinates of the vertex pv in u and v
  //! \ingroup api_mesh_vertex
  //! returns false is the vertex is not parametric (this includes in the volume)
  //! v is only filled if pv is classified on an entity of dimension 2
  MAD_EXPORT bool     V_params(pVertex pv, double * u, double *v);
  
  //! \brief compute the mean length square of the edges adjacent to the vertex 
  //! \ingroup api_mesh_vertex
  MAD_EXPORT double   V_meanEdgeLenSq(const pVertex);
  
  //! \brief returns the center of the cavity surrounding a vertex. 
  //! \ingroup api_mesh_vertex
  MAD_EXPORT void     V_cavityCenter(const pVertex, double[3]);
  
  //! \brief give the square of the distance of point p to vertex pv
  //! \ingroup api_mesh_vertex
  MAD_EXPORT double   V_distanceSquared(const pVertex pv,const double* p);
  
  //! \brief give the distance of point p to vertex pv
  //! \ingroup api_mesh_vertex
  MAD_EXPORT double   V_distance(const pVertex pv,const double* p);

  MAD_EXPORT int      V_checkTransform(const pTransform,const pVertex,const pVertex);

  /***************************/
  /* Entities list operators */
  /***************************/

  MAD_EXPORT pPList  PList_new();
  MAD_EXPORT void    PList_delete         (pPList);
  MAD_EXPORT void    PList_deallocate     (pPList);
  MAD_EXPORT void    PList_clear          (pPList);
  MAD_EXPORT pPList  PList_appPListUnique (pPList, pPList source);
  MAD_EXPORT pPList  PList_appUnique      (pPList, pEntity);
  MAD_EXPORT pPList  PList_append         (pPList, pEntity);
  MAD_EXPORT int     PList_size           (pPList);
  MAD_EXPORT pEntity PList_item           (pPList, int n);
  MAD_EXPORT pEntity PList_next           (pPList, void ** restart);
  MAD_EXPORT int     PList_inList         (pPList, pEntity);
  MAD_EXPORT void    PList_remItem        (pPList, pEntity);
  
  /***********************/
  /* Attached data tools */
  /***********************/
  
  /*! \brief create an exclusive mesh data id */
  MAD_EXPORT pMeshDataId MD_newMeshDataId(const char * tag="");
  
  /*! \brief find preexisting mesh data id and create new if not available */
  MAD_EXPORT pMeshDataId MD_lookupMeshDataId(const char * tag);
  
  /*! \brief find preexisting mesh data id and return -1 if not available */
  MAD_EXPORT pMeshDataId MD_getMeshDataId(const char * tag);

  /*! \brief remove mesh data id */
  MAD_EXPORT void        MD_deleteMeshDataId(pMeshDataId);
  
  MAD_EXPORT void M_fix(pMesh pm);


}

#endif 


