// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#include "MAdResourceManager.h"
#include "MeshDataBase.h"
#include "MeshDataBaseIO.h"
#include "MeshDataBaseInterface.h"
#include "MeshDataBaseParallelInterface.h"
#include "MeshDataBaseParallelIO.h"
#include "MeshDataBaseCommCheck.h"
#include "MeshDataBaseComm.h"
#ifdef PARALLEL
#include "mpi.h"
#include "autopack.h"
#endif

#include <set>
#include <iterator>
#include <algorithm>
#include <cassert>

namespace MAd {
  
  struct edge_comm {
    int distProc;  // remote processor 
    pEdge pe;      // edge 
    pVertex p1,p2; // vertices on remote processor
    int id1,id2;   // local identity tags
    int classTag;  // tag of local classification
    int classDim;  // dimension of local classification
  };  
  
  struct face_comm
  {
    int     distProc;        // remote processor 
    pFace   pf;              // face
    pVertex p1,p2,p3,p4;     // vertices on remote processor
    int     id1,id2,id3,id4; // local identity tags 
    int classTag;            // tag of local classification
    int classDim;            // dimension of local classification
  };

  // -------------------------------------------------------------------
  
  MDB_DataExchanger::MDB_DataExchanger(int t):tagComm(t) {
    MAdResourceManagerSgl::instance().reserveCommunicationTag(tagComm);
  }

  // -------------------------------------------------------------------

  MDB_DataExchanger::~MDB_DataExchanger() {
    MAdResourceManagerSgl::instance().releaseCommunicationTag(tagComm);
  }
  
  // -------------------------------------------------------------------
  /*! \brief write mesh in parallel format \ingroup internal */
#ifdef PARALLEL
  void M_writeParallel(pMesh m, const char * filename, int version)
  {
    SaveGmshMeshParallel(m, filename, version);
  }
#endif

  // -------------------------------------------------------------------
  /*! \brief flags true if entity pe is on a partition boundary \ingroup internal
    \warning only valid after creation of interface exchange information (based
    on data with tag "RemotePoint") */
  
  bool EN_isInterface(pEntity pe)
  {
    pMeshDataId tagData = MD_lookupMeshDataId("RemotePoint");
    void *tmp;
    int isInterface = EN_getDataPtr(pe, tagData, &tmp);
    if (isInterface) return true;
    return false;
  }
  
  // -------------------------------------------------------------------
  /*! \brief flags true if vertex pv is on a partition boundary \ingroup parallel
    \warning only valid after V_createInfoInterface (based on EN_isInterface) */
  
  bool V_isInterface(pVertex pv) { return EN_isInterface((pEntity) pv);}
  
  // -------------------------------------------------------------------
  /*! \brief flags true if edge pe is on a partition boundary \ingroup parallel
    \warning only valid after E_createInfoInterface (based on EN_isInterface)  */
  
  bool E_isInterface(pEdge pe)   { return EN_isInterface((pEntity) pe);}
  
  // -------------------------------------------------------------------
  /*! \brief flags true if face pf is on a partition boundary \ingroup parallel
    \warning only valid after F_createInfoInterface  (based on EN_isInterface) */

  bool F_isInterface(pFace pf)   { return EN_isInterface((pEntity) pf);}

  // -------------------------------------------------------------------
  /*!  \brief Lists all processors on which the vertex pv has a copy \ingroup internal 
    The data with tag "RemotePoint" is created during
    V_createInfoInterface and stores a list of pairs consisting of processor
    id and vertex pointer
  */
  
  int V_listInterface(pVertex pv, std::vector<int>* distProcs)
  {
    pMeshDataId tagVertex = MD_lookupMeshDataId("RemotePoint");
    
    void *tmp;
    int isInterface = EN_getDataPtr((pEntity) pv, tagVertex, &tmp);
    if(isInterface) {
      
      ConnectionMap* connections = reinterpret_cast<ConnectionMap*> (tmp);
      
      std::set<int> parts;
      ConnectionMap::const_iterator cIter = connections->begin();
      for (;cIter!=connections->end();++cIter) parts.insert(cIter->first);
      
      distProcs->insert(distProcs->begin(),parts.begin(),parts.end());
    }
    return (int)(distProcs->size());
  }
  
  // -------------------------------------------------------------------
  /*!  \brief Returns true if the node is a (periodic) copy of the node with global id rId \ingroup internal */
  /*! This information is based on data read with the mesh and hence will always be correct */
  
  bool V_corresponds(pVertex pv,int rId) 
  {
    if (EN_id((pEntity) pv) == rId) return true;
    
    pMeshDataId tagVertex = MD_lookupMeshDataId("PeriodicVertexID");
    void* tmpptr = NULL;
    if (!EN_getDataPtr((pEntity) pv,tagVertex,&tmpptr)) return false;
    std::map<int,std::vector<int> >* conn = (std::map<int,std::vector<int> >*) tmpptr;
    return (conn->find(rId) != conn->end());
  }

  // -------------------------------------------------------------------
  
  /*! \brief return true if the edge corresponds (possibly) \ingroup internal */
  /*! \warning error for edges transformed onto oneself */
  
  bool E_corresponds(pEdge pe,int rv1,int rv2) {

    pVertex pv1 = E_vertex(pe,0);
    pVertex pv2 = E_vertex(pe,1);

    int lv1 = EN_id((pEntity) pv1);
    int lv2 = EN_id((pEntity) pv2);
    
    int lmin = std::min(lv1,lv2);
    int lmax = std::max(lv1,lv2);

    int rmin = std::min(rv1,rv2);
    int rmax = std::max(rv1,rv2);

    // identical edge - most of the cases
    
    if (lmin == rmin && lmax == rmax) return true;

    // only one shared node -> problems
    
    if (lmin == rmax && lmax != rmax) return false;
    if (lmin != rmin && lmax == rmax) return false;
    
    // all other cases should be ok ? (Koen)
    
    if (V_corresponds(pv1,rv1) && V_corresponds(pv2,rv2)) return true;
    if (V_corresponds(pv1,rv2) && V_corresponds(pv2,rv1)) return true;

    return false;

  }

  // -------------------------------------------------------------------
  /*! \brief return true if the face corresponds \ingroup internal */
  /*! \warning not yet fully functional outside of F_createInfoInterface */
  /*! <ul>
      <li> we do not really verify periodic nodes 
      <li> error for edges transformed onto oneself
      </ul>
  */  

  bool F_corresponds(pFace pf,int rv1,int rv2,int rv3,int rv4) {
    
    std::set<int> lverts;

    for (int i=0;i<F_numVertices(pf);i++) {
      lverts.insert(EN_id((pEntity) F_vertex(pf,i)));
    }
    
    std::set<int> rverts;
    rverts.insert(rv1);
    rverts.insert(rv2);
    rverts.insert(rv3);
    if (rv4 != -1) rverts.insert(rv4);
    
    // obvious checks ... 
    if (rverts.size() != lverts.size()) return false;

    // same face ... 
    if (rverts == lverts) return true;

    // only some shared nodes is not ok ? 
    std::set<int> nonShared;
    std::set_difference(lverts.begin(),lverts.end(),
                        rverts.begin(),rverts.end(),
                        std::insert_iterator< std::set<int> > (nonShared,
                                                               nonShared.begin()));
    
    if (nonShared.size() != lverts.size()) return false;
    return true;
  }

  // -------------------------------------------------------------------
  /*!  \brief Returns true if the node is a (periodic) copy of the node with
       global id rId \ingroup parallel
    
    This information is based on data read with the mesh and hence will always
    be correct; only valid for check on local partition */
  
  bool V_corresponds(pVertex p1,pVertex p2) 
  {
    return (V_corresponds(p1,EN_id((pEntity) p2)) && 
            V_corresponds(p2,EN_id((pEntity) p1)));
  }

  // --------------------------------------------------------------------

  /*! \brief returns true if the node is flagged as periodic \ingroup parallel

    This information is based on data read with the mesh and hence will always
    be correct */

  bool V_isPeriodic(pVertex pv) {
    
    pMeshDataId tagVertex = MD_lookupMeshDataId("PeriodicVertexID");
    void* tmpptr = NULL;
    if (!EN_getDataPtr((pEntity) pv,tagVertex,&tmpptr)) return false;
    return true;
  }
    
  // -------------------------------------------------------------------
  /*!\brief  An edge flagged as potentially on a parallel interface  \ingroup internal

    \warning the check is not correct for 2D meshes with internal edges */    
  
  bool E_isPotentialInterface(pMesh mesh, pEdge pe)
  {

    switch (M_dim(mesh)) {
    case 2:
      {     
      
        // --- check that the edge has less than two neighbouring faces ---
        
        int numf = pe->numfaces();
        if( numf == 2 ) return false;
        
        // --- check that the edge is [not alone or] <= Not OK for Morfeo/Crack
        //     with multiple neighbouring faces (debug) ---
        
        assert(numf <= 1);
        return true;
        
        // KH: for the moment an edge is always classified on an edge ... (nullmodel)
        //     hence the following check is not usable ... 
        /// 
        // if edge is classified on an edge, still two possibilities
        // 
        //   1. periodic edge: check whether or not nodes are periodic - ok
        //   2. internal edge: not yet treated correctly
        
        
        if ( E_whatInType(pe) == 1 ) {
          for (int i=0;i<2;i++) if (!V_isPeriodic(E_vertex(pe,i))) return false;
          return true;
        }
        
        break;
      }
    case 3:
      {
        // --- check that the edge has at least one neighbouring face that is
        // --- potentially shared ---
        
        int numf = pe->numfaces();
        for (int iF=0; iF < numf; iF++) {
          if ( F_isPotentialInterface(mesh,E_face(pe,iF))) return true;
        }
        break;
      }
    }

    return false;
  }

  // ---------------------------------------------------------------------------
  /*!\brief Verifies whether edge pe is located at a parallel interface boundary 
    and provides information about all possible copies \ingroup internal */
  
  bool E_isPotentialInterface(pMesh mesh, pEdge pe, std::vector<edge_comm>& distVt) {
    
    distVt.clear();
    
    if ( !E_isPotentialInterface(mesh,pe) ) return false;

    pMeshDataId tagData = MD_lookupMeshDataId("RemotePoint");
    pVertex pv1 = pe->p1;
    pVertex pv2 = pe->p2;
    void * tmp1, *tmp2;
    int isInterface1 = EN_getDataPtr((pEntity) pv1 , tagData, &tmp1);
    int isInterface2 = EN_getDataPtr((pEntity) pv2 , tagData, &tmp2);
  
    if( !(isInterface1 && isInterface2) ) return false;

    // --- check that the two nodes are on the same parallel boundary ---
    
    const ConnectionMap* conn1 = reinterpret_cast<const ConnectionMap* > (tmp1);
    const ConnectionMap* conn2 = reinterpret_cast<const ConnectionMap* > (tmp2);

    ConnectionMap::const_iterator rIter1;
    ConnectionMap::const_iterator rIter2;

    pGEntity pg = EN_whatIn((pEntity) pe);
    int classDim = GEN_type(pg);
    int classTag = GEN_tag(pg);
    
    for (rIter1=conn1->begin();rIter1!=conn1->end();++rIter1) {
      
      int remoteProc = rIter1->first;
      pEntity remote1 = rIter1->second.remoteEntity;

      for (rIter2=conn2->begin();rIter2!=conn2->end();++rIter2) {
        
        if (rIter2->first == remoteProc) {
          
          edge_comm ee;
          ee.distProc = remoteProc;
          ee.p1 = (pVertex) remote1;
          ee.p2 = (pVertex) rIter2->second.remoteEntity;
          
          ee.id1 = EN_id((pEntity) E_vertex(pe,0));
          ee.id2 = EN_id((pEntity) E_vertex(pe,1));
          ee.pe = pe;
          ee.classDim = classDim;
          ee.classTag = classTag;

          distVt.push_back(ee);
        }
      }
    }
    return (!distVt.empty());
  }

  // -------------------------------------------------------------------
  /*! \brief Verify whether pf may potentially be an interface \ingroup internal 
    \warning will flag false for internal boundaries !!! */ 
  
  // -------------------------------------------------------------------

  bool F_isPotentialInterface(pMesh mesh, pFace pf)
  {
   
    // --- check that the mesh is 3D ---
    
    if ( M_dim(mesh) != 3 ) return false;
  
    // --- check that the face has less than two neighbouring regions ---

    int numreg = pf->getNbRegions();
    if (numreg == 2) return false;
    return true;

    // OK : for the moment a face is always classified on a geometric face (nullmodel ...)
    
    // --- if the face is classified on a topological face, we can still have 
    //     1. a periodic face: all nodes are periodic - ok
    //     2. an internal boundary - check is not ok 
    
    int modelDim = GEN_type(pf->g);

    if (modelDim == 2) {
      for (int i=0;i<F_numVertices(pf);i++) {
        if (!V_isPeriodic(F_vertex(pf,i))) return false;
      }
      return true;
    }
    
    // --- check that the face is not alone (debug) ---
    assert(numreg);
    return true;
  }


  /*! \brief flag face pface as potential interface and list possible communications 
    \ingroup internal */

  bool F_isPotentialInterface(pMesh mesh, pFace pface, std::vector<face_comm>& distFace)
  {

    if ( !F_isPotentialInterface(mesh,pface) ) return false;

    // --- check that the three nodes are on a parallel boundary ---

    pMeshDataId tagData = MD_lookupMeshDataId("RemotePoint");
    pVertex nod[4];

    distFace.clear();


    switch (pface->getNbNodes()) {

    case 3:
      {
        pface->getNodes(nod);
        void * tmp1, *tmp2, *tmp3;
        int isInterface1 = EN_getDataPtr((pEntity) nod[0], tagData, &tmp1);
        int isInterface2 = EN_getDataPtr((pEntity) nod[1], tagData, &tmp2);
        int isInterface3 = EN_getDataPtr((pEntity) nod[2], tagData, &tmp3);
      
        int id1 = EN_id((pEntity) nod[0]);
        int id2 = EN_id((pEntity) nod[1]);
        int id3 = EN_id((pEntity) nod[2]);

        pGEntity classification = EN_whatIn((pEntity) pface);
        
        int classTag = GEN_tag(classification);
        int classDim = GEN_type(classification);        

        if( !(isInterface1 && isInterface2 && isInterface3) ) return false;

        const ConnectionMap* conn1 = reinterpret_cast<const ConnectionMap* > (tmp1);
        const ConnectionMap* conn2 = reinterpret_cast<const ConnectionMap* > (tmp2);
        const ConnectionMap* conn3 = reinterpret_cast<const ConnectionMap* > (tmp3);
        
        ConnectionMap::const_iterator rIter1 = conn1->begin();
        for (;rIter1!=conn1->end();++rIter1) {
          int remoteProc1 = rIter1->first;
          ConnectionMap::const_iterator rIter2 = conn2->begin();
          for (;rIter2!=conn2->end();++rIter2) {
            if (remoteProc1 == rIter2->first) {
              ConnectionMap::const_iterator rIter3 = conn3->begin();
              for (;rIter3!=conn3->end();++rIter3) {
                if (remoteProc1 == rIter3->first) {

                  face_comm fc;
                
                  fc.distProc = remoteProc1;

                  fc.p1 = (pVertex) rIter1->second.remoteEntity;
                  fc.p2 = (pVertex) rIter2->second.remoteEntity;
                  fc.p3 = (pVertex) rIter3->second.remoteEntity;
                  fc.p4 = NULL;
                  
                  fc.id1 = id1;
                  fc.id2 = id2;
                  fc.id3 = id3;
                  fc.id4 = -1;

                  fc.pf = pface;

                  fc.classTag = classTag;
                  fc.classDim = classDim;
                
                  distFace.push_back(fc);

                }
              }
            }
          }
        }

        break;
      }
    
    case 4:
      {
        pface->getNodes(nod);
        void * tmp1, *tmp2, *tmp3, *tmp4;
        int isInterface1 = EN_getDataPtr((pEntity) nod[0], tagData, &tmp1);
        int isInterface2 = EN_getDataPtr((pEntity) nod[1], tagData, &tmp2);
        int isInterface3 = EN_getDataPtr((pEntity) nod[2], tagData, &tmp3);
        int isInterface4 = EN_getDataPtr((pEntity) nod[3], tagData, &tmp4);
        
        int id1 = EN_id((pEntity) nod[0]);
        int id2 = EN_id((pEntity) nod[1]);
        int id3 = EN_id((pEntity) nod[2]);
        int id4 = EN_id((pEntity) nod[3]);
    
        pGEntity classification = EN_whatIn((pEntity) pface);
        
        int classTag = GEN_tag(classification);
        int classDim = GEN_type(classification);
        if(!(isInterface1 && isInterface2 && isInterface3 && isInterface4)) return false;
    
        
        const ConnectionMap* conn1 = reinterpret_cast<const ConnectionMap* > (tmp1);
        const ConnectionMap* conn2 = reinterpret_cast<const ConnectionMap* > (tmp2);
        const ConnectionMap* conn3 = reinterpret_cast<const ConnectionMap* > (tmp3);
        const ConnectionMap* conn4 = reinterpret_cast<const ConnectionMap* > (tmp4);
        
        ConnectionMap::const_iterator rIter1 = conn1->begin();
        for (;rIter1!=conn1->end();++rIter1) {
          int remoteProc1 = rIter1->first;
          ConnectionMap::const_iterator rIter2 = conn2->begin();
          for (;rIter2!=conn2->end();++rIter2) {
            if (remoteProc1 == rIter2->first) {
              ConnectionMap::const_iterator rIter3 = conn3->begin();
              for (;rIter3!=conn3->end();++rIter3) {
                if (remoteProc1 == rIter3->first) {
                  ConnectionMap::const_iterator rIter4 = conn4->begin();
                  for (;rIter4!=conn4->end();++rIter4) {
                    if (remoteProc1 == rIter4->first) {

                      face_comm fc;
                      
                      fc.distProc = remoteProc1;
                      
                      fc.p1 = (pVertex) rIter1->second.remoteEntity;
                      fc.p2 = (pVertex) rIter2->second.remoteEntity;
                      fc.p3 = (pVertex) rIter3->second.remoteEntity;
                      fc.p4 = (pVertex) rIter4->second.remoteEntity;
                  
                      fc.id1 = id1;
                      fc.id2 = id2;
                      fc.id3 = id3;
                      fc.id4 = id4;

                      fc.classTag = classTag;
                      fc.classDim = classDim;

                      fc.pf = pface;
                      
                      distFace.push_back(fc);
                    }
                  }
                }
              }
            }
          }
        }
        break;
      }
    }
    return (distFace.size() != 0);
  }

  // -------------------------------------------------------------------
  /*! \brief Flag region if located at an interface. \ingroup parallel
    \warning As we partition along edges in 2D and faces in 3D, the answer is
    always no  */

  bool R_isInterface(pRegion pr) {return false;}

  // -------------------------------------------------------------------
  // do a tentative connection - since we send all potential interface nodes
  // to all processors, we send the minimum information : the node id 
  // 
  // result is that we store the connected partitions for a shared node
  // more detailed communication follows 
  
  void createRemotePointLists (pMesh mesh, pMeshDataId tagVertex)
  {
    std::set<pVertex> bdryNodes;
    std::set<int>     bdryNodeId;

    // locate all points on a potential parallel boundary

#ifdef PARALLEL
    int myRank = 0;
    int nbproc = 1;
    
    MPI_Comm_size(mesh->getMPIComm(), &nbproc);
    MPI_Comm_rank(mesh->getMPIComm(), &myRank);

    // --- list local nodes on parallel boundaries ---
    switch (M_dim(mesh)) {
    case 3:
      {
        FIter fit = M_faceIter(mesh);
        while ( pFace pf = FIter_next(fit) ) {
          if ( F_isPotentialInterface(mesh,pf) ) {
            for (int iv = 0; iv < F_numVertices(pf); iv++) {
              pVertex pv = F_vertex(pf,iv);
              bdryNodes .insert(pv);
              bdryNodeId.insert(pv->iD);
            }
          }
        }
        FIter_delete(fit);
        break;
      }
    case 2:
      {
        EIter eit = M_edgeIter(mesh);
        while ( pEdge pe = EIter_next(eit) ) {
          if ( E_isPotentialInterface(mesh,pe) ) {
            for (int iv = 0; iv < 2; iv++) {
              pVertex pv = E_vertex(pe,iv);            
              bdryNodes.insert(pv);
              bdryNodeId.insert(pv->iD);
            }
          }
        }
        EIter_delete(eit);
        break;
      }
    }
#endif

    // add all periodic nodes --> introduce transformations ?
    
    pMeshDataId tagPeriodic = MD_lookupMeshDataId("PeriodicVertexID");  
    
    VIter vit = M_vertexIter(mesh);
    
    while (pVertex pv = VIter_next(vit)) {
      
      void* tmp_periodic = NULL;
      int isPeriodic = EN_getDataPtr((pEntity) pv,tagPeriodic,&tmp_periodic);

      if (isPeriodic) {
      
        void* tmp_remote = NULL;
        int hasRemote  = EN_getDataPtr((pEntity) pv,tagVertex,&tmp_remote);
        
        if (!hasRemote) {
          ConnectionMap* remote = new ConnectionMap;
          EN_attachDataPtr((pEntity) pv,tagVertex,remote);
        }
        else {
          ConnectionMap* remote = (ConnectionMap*) tmp_remote;
          remote->clear();
        }
        
        std::map<int,std::vector<int> > * conn = (std::map<int,std::vector<int> > *) tmp_periodic;
        std::map<int,std::vector<int> >::iterator citer = conn->begin();
        bdryNodes.insert(pv);
        for (;citer != conn->end();++citer) bdryNodeId.insert(citer->first);
      }
    }
    VIter_delete(vit);
    
#ifdef PARALLEL
    MPI_Barrier(mesh->getMPIComm());

    int commTag = 669;
    MAdResourceManagerSgl::instance().reserveCommunicationTag(commTag);
    
    // --- send list ---
    
    int* sendcounts = new int[nbproc];
    int sendSize = bdryNodeId.size()*sizeof(int);
    
    void* send = malloc(sendSize);
    int * bufCast = reinterpret_cast<int *>(send);
    std::set<int>::const_iterator cIter = bdryNodeId.begin();
    std::set<int>::const_iterator cLast = bdryNodeId.end();
    for (;cIter!=cLast;++cIter) *(bufCast++) = *cIter;
    
    for (int iProcDest=0; iProcDest < nbproc; iProcDest++) {
      if ( iProcDest == myRank ) {
        sendcounts[iProcDest] = 0;
        continue;    
      }
      else {
        sendcounts[iProcDest] = 1;
        void *buf = AP_alloc(iProcDest,commTag,sendSize);
        memcpy(buf,send,sendSize);
        AP_send(buf);
      }
    }

    free(send);

    AP_check_sends(AP_NOFLAGS);
    AP_reduce_nsends(sendcounts);
    AP_flush();

    // --- receive lists ---

    int message = 0;
    int count;

    while (!AP_recv_count(&count) || message < count) {

      void *recv;
      int from;
      int tag;
      int recvSize;
      
    
      if (AP_recv (MPI_ANY_SOURCE, commTag, AP_BLOCKING|AP_DROPOUT,
                   &recv, &recvSize, &from, &tag)) {
    
        int* irecv = reinterpret_cast<int*>(recv);

        std::set<int> distNodes;
        int nbDistNodes = recvSize / sizeof(int);
        for (int iDistN = 0; iDistN < nbDistNodes; iDistN++) {
          distNodes.insert(irecv[iDistN]);
        }
        
        std::set<pVertex>::const_iterator vIter = bdryNodes.begin();
        std::set<pVertex>::const_iterator vLast = bdryNodes.end();
        
        for (;vIter!=vLast;++vIter) {

          pVertex pv = *vIter;
          int     id = EN_id((pEntity) pv);

          if (distNodes.find(id) != distNodes.end()) {
            void *tmp;
            int exist = EN_getDataPtr((pEntity) pv, tagVertex, &tmp);
            ConnectionMap * parts = NULL;
            if (exist) {
              parts = reinterpret_cast<ConnectionMap*>(tmp);
            } else if (!exist) {
              parts = new ConnectionMap;
              EN_attachDataPtr((pEntity) pv, tagVertex, parts);
            }
            parts->insert(std::make_pair(from,EntityConnection(NULL,NULL)));
          }
        } 
        message++;
        AP_free(recv);
      }
    }
    
    // --- wait until all have finished sending ---
  
    MAdResourceManagerSgl::instance().releaseCommunicationTag(commTag);

    AP_check_sends(AP_WAITALL);
    MPI_Barrier(mesh->getMPIComm());
    delete [] sendcounts;
#endif
  }
  
  // -------------------------------------------------------------------
  
  void V_createInfoInterface(pMesh mesh, pMeshDataId tagVertex)
  {
    int myRank = 0;
    size_t nbRemoteVertices = 0;

    pGModel model = M_model(mesh);
    MAdMsg& msg = MAdMsgSgl::instance();
    
#ifdef PARALLEL        
    if(mesh->getMPIComm()==MPI_COMM_SELF) return;
    int mysize = 1;
    MPI_Comm_size(mesh->getMPIComm(), &mysize);
    MPI_Comm_rank(mesh->getMPIComm(), &myRank);
#endif

    // --- Delete previous information ---
    VIter vit = M_vertexIter(mesh);
    while ( pVertex pv = VIter_next(vit) ) {
      void * tmp_ptr;
      if( EN_getDataPtr((pEntity) pv,tagVertex,&tmp_ptr)) {
        EN_deleteData((pEntity) pv, tagVertex);
      }
    }
    VIter_delete(vit);

    // --- Create lists of distant procs ---

    createRemotePointLists(mesh, tagVertex);

#ifdef PARALLEL
    // --- Send node pointers ---
    int *sendcounts = new int[mysize];
    for(int i=0; i<mysize; i++) sendcounts[i] = 0;
#endif

    pMeshDataId tagPeriodic = MD_lookupMeshDataId("PeriodicVertexID");

#ifdef PARALLEL
    int commTag = 445;
    MAdResourceManagerSgl::instance().reserveCommunicationTag(commTag);
#endif
    
    vit = M_vertexIter(mesh);
    while ( pVertex pv = VIter_next(vit) ) {

      std::vector<int> distProcTab;
      V_listInterface(pv, &distProcTab);

      std::set<int> distProcs;
      distProcs.insert(distProcTab.begin(),distProcTab.end());
      
      void* tmp_remote = NULL;
      int haveRemote = EN_getDataPtr((pEntity) pv,tagVertex,&tmp_remote);
      ConnectionMap* remote = NULL;
      if (haveRemote) {
        remote  = (ConnectionMap*) (tmp_remote);
        remote->clear();
      }
      
      void* tmp_periodic = NULL;
      int havePeriodic = EN_getDataPtr((pEntity) pv,tagPeriodic,&tmp_periodic);
      std::map<int,std::vector<int> >* periodic = NULL;

      if (havePeriodic) {
        
        if (!haveRemote) {
          msg.error(__LINE__,__FILE__,
                    "Periodic node %d without connection table",
                    EN_id((pEntity) pv));
        }
        
        pGEntity localClass = EN_whatIn((pEntity) pv);
        
        // insert data on local processor
        
        periodic = (std::map<int,std::vector<int> >*) tmp_periodic;
        std::map<int,std::vector<int> >::const_iterator citer = periodic->begin();
        for (;citer!=periodic->end();++citer) {
          pVertex cvtx = mesh->find_point(citer->first);
          if (cvtx && cvtx != pv) {
            
            pGEntity remoteClass = EN_whatIn((pEntity) cvtx);
            pTransform tfo = GM_transformation(model,localClass,remoteClass);

            remote->insert(std::make_pair(myRank,EntityConnection(cvtx,tfo)));
            nbRemoteVertices++;
          }
        }
      }
      
#ifdef PARALLEL
      
      int nbConnections = 1;
      if (periodic) nbConnections += periodic->size();
      int sendSize = sizeof(pVertex) + (3 + nbConnections ) * sizeof(int);
      void* send = malloc(sendSize);
      
      pVertex *vSend = (pVertex *) send;

      int lId = EN_id((pEntity) pv);
      
      *(vSend++) = pv;
      int* iSend = (int*) (vSend);
      
      *(iSend++) = GEN_tag (V_whatIn(pv));
      *(iSend++) = GEN_type(V_whatIn(pv));
      *(iSend++) = nbConnections;
      *(iSend++) = lId;

      
      if (periodic) {
        std::map<int,std::vector<int> >::const_iterator cIter = periodic->begin();
        for (;cIter!=periodic->end();++cIter) {
          int rId = cIter->first;
          *(iSend++) = (rId == lId) ? -rId : rId; // avoid sending once again to oneself
        }
      }
      
      // make sure that we send only once to different processors

      std::set<int>::iterator dIter = distProcs.begin();
      for (;dIter!=distProcs.end();++dIter) {

        int distProc = *dIter;
        if (distProc != myRank) {    
          void *buf = AP_alloc(distProc,commTag,sendSize);
          memcpy(buf,send,sendSize);
          AP_send(buf);
          sendcounts[distProc]++;
        }
      }
      
      free(send);
#endif
    }
    VIter_delete(vit);
    
#ifdef PARALLEL
    AP_check_sends(AP_NOFLAGS);
    AP_reduce_nsends(sendcounts);
    AP_flush();

    // --- Receive node pointers ---
    
    int message=0, count;
    
    while (!AP_recv_count(&count) || message<count) {
      
      void *recv;
      int  from;
      int  tag;
      int  recvSize;
      
      if (AP_recv(MPI_ANY_SOURCE, commTag, AP_BLOCKING|AP_DROPOUT,
                  &recv, &recvSize, &from, &tag))  {
        
        message++;
        pVertex* vRecv = reinterpret_cast<pVertex*>(recv);
        pVertex pRecv = *(vRecv++);
        int* iRecv = reinterpret_cast<int*>(vRecv);
        
        int gTag, gDim;
        gTag   = *(iRecv++);
        gDim   = *(iRecv++);
        int nbConn = *(iRecv++);       
        
        pGEntity remoteClass = GM_entityByTag(model,gTag,gDim);
        
        for (int iConn=0;iConn<nbConn;iConn++) {
          
          int pId = *(iRecv++);
          pVertex vtx = mesh->find_point(pId);
          
          if (vtx) {     
            
            pGEntity localClass = EN_whatIn((pEntity) vtx);
            pTransform tfo = GM_transformation(model,localClass,remoteClass);

            void* tmp_remote =  NULL;
            int isInterface = EN_getDataPtr((pEntity) vtx, tagVertex, &tmp_remote);
            if(! isInterface) {
              msg.error(__LINE__,__FILE__,"Problem in V_createInfoInterface");
            }
            
            ConnectionMap *conn = (ConnectionMap*) tmp_remote;  
            conn->insert(std::make_pair(from,EntityConnection(pRecv,tfo)));

            nbRemoteVertices++;
            
          }
        }
        AP_free(recv);
      }
    }
    
    MAdResourceManagerSgl::instance().releaseCommunicationTag(commTag);
    
    AP_check_sends(AP_WAITALL);
    MPI_Barrier(mesh->getMPIComm());
    delete [] sendcounts;

    MDB_CommCheck cc(mesh->getMPIComm());
#else
    MDB_CommCheck cc;
#endif
  
    exchangeDataOnVertices(mesh,cc);

  }
  
  // -------------------------------------------------------------------
  /*! \brief Create edge correspondance and assure coherent orientation \ingroup parallel */ 
  void E_createInfoInterface(pMesh mesh, pMeshDataId tagEdge)
  { 
    int myRank = 0;
    
#ifdef PARALLEL    
    if(mesh->getMPIComm()==MPI_COMM_SELF) return;
    int nproc = 1;
    MPI_Comm_size(mesh->getMPIComm(), &nproc);
    MPI_Comm_rank(mesh->getMPIComm(), &myRank);
#endif
    
    MAdMsg& msg = MAdMsgSgl::instance();
    pGModel model = M_model(mesh);
    
    // --- Delete previous information ---
    
    EIter eit = M_edgeIter(mesh);
    while ( pEdge pe = EIter_next(eit) ) {
      void * tmp_ptr;
      if( EN_getDataPtr((pEntity) pe, tagEdge, &tmp_ptr) ) {
        EN_deleteData((pEntity) pe, tagEdge);
      }
    }
    EIter_delete(eit);


    // --- establish local connections and assure coherent orientation --- 
    // 
    //     * coherent reorientation on other partitions assumes coherent orientation here 
    //     * all periodic edges have same connections across partititions
    //       hence subsequent reorientation should remain coherent

    
    size_t nbRemoteEdges = 0;
    
    eit = M_edgeIter(mesh);
    while ( pEdge pe = EIter_next(eit) ) {

      std::vector<edge_comm> remoteEdges;
      
      if (E_isPotentialInterface(mesh,pe,remoteEdges)) {
      
        int id0 = EN_id((pEntity) E_vertex(pe,0));
        int id1 = EN_id((pEntity) E_vertex(pe,1));
        
        std::pair<int,int> lNodeID(std::min(id0,id1),
                                   std::max(id0,id1));
        
        std::vector<edge_comm >::const_iterator riter = remoteEdges.begin();

        pVertex p1Local = E_vertex(pe,0);
        pVertex p2Local = E_vertex(pe,1);
        
        pGEntity localClass = EN_whatIn((pEntity) pe);

        for (;riter!=remoteEdges.end();++riter) {
        
          int distProc = riter->distProc;

          if (distProc == myRank) {
            
            edge_comm comm = *riter;
            
            pVertex p1 = comm.p1;
            pVertex p2 = comm.p2;
            
            pEdge remote = E_exist(p1,p2);
            
            if (remote) {
              
              // --- get points again, since alignment may be different from comm 

              pVertex p1Remote = E_vertex(remote,0);
              pVertex p2Remote = E_vertex(remote,1);

              // --- verify orientation and realign if necessary 

              int orientation = 0;
              if (V_corresponds(p1Remote,p1Local) && 
                  V_corresponds(p2Remote,p2Local)) orientation =  1;
              if (V_corresponds(p1Remote,p2Local) && 
                  V_corresponds(p2Remote,p1Local)) orientation = -1;
                
              if (orientation == 0) {
                msg.error(__LINE__,__FILE__,
                          "Found a non-aligned periodic edge connection %d-%d to %d-%d",
                          EN_id((pEntity) p1),
                          EN_id((pEntity) p2),
                          EN_id((pEntity) p1Local),
                          EN_id((pEntity) p2Local));
              }
              
              std::pair<int,int> rNodeID(std::min(EN_id((pEntity) p1Remote),
                                                  EN_id((pEntity) p2Remote)),
                                         std::max(EN_id((pEntity) p1Remote),
                                                  EN_id((pEntity) p2Remote))); 
  
            
              if (rNodeID < lNodeID) { 
                lNodeID = rNodeID; // remember to what we aligned the edge 
                if (orientation == -1) E_align(pe,p2Local,p1Local);
              }

              // --- get transformation

              pGEntity remoteClass = EN_whatIn(remote);
              pTransform tfo = GM_transformation(model,localClass,remoteClass);
              
              // --- add the current connection 

              void* tmp;
              ConnectionMap* connections;
              
              if(EN_getDataPtr((pEntity) pe, tagEdge, (void**)&tmp)) {
                connections = (ConnectionMap*) tmp;
              }
              else {
                connections = new ConnectionMap ;
                EN_attachDataPtr((pEntity) pe,tagEdge,connections);
              }
              
              connections->insert(std::make_pair(myRank,EntityConnection(remote,tfo)));
              nbRemoteEdges++;
            }
          }
        }
      }
    }
    EIter_delete(eit);
    
#ifdef PARALLEL

    int commTag = 446;
    MAdResourceManagerSgl::instance().reserveCommunicationTag(commTag);
    
    
    int *sendcounts = new int[nproc];
    for(int i=0;i<nproc;i++)sendcounts[i]=0;
    
    // --- send edge info ---

    eit = M_edgeIter(mesh);
    while ( pEdge pe = EIter_next(eit) ) {

      std::vector<edge_comm> remote;

      if (E_isPotentialInterface(mesh,pe,remote)) {

        std::vector<edge_comm >::const_iterator rIter = remote.begin();

        for (;rIter!=remote.end();++rIter) {
        
          int distProc = rIter->distProc;

          if (distProc != myRank) {
            void *buf = AP_alloc(distProc,commTag,sizeof(edge_comm));
            edge_comm* castbuf = (edge_comm*) buf;
            std::memcpy(buf,&(*rIter),sizeof(edge_comm));
            castbuf->distProc = myRank;
            AP_send(buf);
            sendcounts[distProc]++;
          }
        }
      }
    }
    EIter_delete(eit);
    
    // --- synchronise sends --- 
    
    AP_check_sends(AP_NOFLAGS);
    AP_reduce_nsends(sendcounts);
    //AP_flush();

    // --- Receive edge info ---
    
    int message=0,count;
    while (!AP_recv_count(&count) ||message<count) {

      void *recv;
      int from;
      int tag;
      int recvSize;
      
      if (AP_recv(MPI_ANY_SOURCE,commTag, AP_BLOCKING|AP_DROPOUT,
                  &recv, &recvSize, &from, &tag)) {
        message++;
        edge_comm * comm = (edge_comm*) recv;
        pEdge pe = E_exist(comm->p1,comm->p2);
        if(pe) {
        

          pGEntity localClass = EN_whatIn((pEntity) pe);

          int lowestRank = myRank;
          
          // avoid connections between two edges that span full periodicity
          // ie. periodic nodes n1 and n3 while both edges n1-n2 and n2-n3 exist

          int id1 = comm->id1;
          int id2 = comm->id2;

          if (E_corresponds(pe,id1,id2)) {
            
            ConnectionMap* connections;
            
            if(EN_getDataPtr((pEntity) pe, tagEdge, (void**)&connections)) {
              lowestRank = std::min(myRank,connections->begin()->first);
            }
            else {
              connections = new ConnectionMap;
              EN_attachDataPtr((pEntity) pe, tagEdge, (void*)connections);
            }

            pGEntity remoteClass = GM_entityByTag(model,comm->classTag,comm->classDim);
            pTransform tfo = GM_transformation(model,localClass,remoteClass);
            
            connections->insert(std::make_pair(from,EntityConnection(comm->pe,tfo)));
            if (from < lowestRank) {
              int dir = E_align(pe,comm->p1,comm->p2);
              if (dir == 0) {
                msg.error(__LINE__,__FILE__,
                          "Could not align edge %i-%i with a remote from proc %i",
                          EN_id(E_vertex(pe,0)),
                          EN_id(E_vertex(pe,1)),
                          from);
              }
            }
            
            nbRemoteEdges++;

          }
        }
        AP_free(recv);
      }
    }

    // --- synchronise receives --- 

    MAdResourceManagerSgl::instance().releaseCommunicationTag(commTag);

    AP_check_sends(AP_WAITALL);
    MPI_Barrier(mesh->getMPIComm());
    
    // --- clean ship ---
    
    delete [] sendcounts;

#endif 
    
    eit = M_edgeIter(mesh);
    
    while ( pEdge localEdge = EIter_next(eit) ) {
      
      if (E_isInterface(localEdge)) {

        pVertex p1Local = E_vertex(localEdge,0);
        pVertex p2Local = E_vertex(localEdge,1);
  
        void* tmp1;
        EN_getDataPtr((pEntity) localEdge, tagEdge, (void**)&tmp1);
        ConnectionMap* connections = reinterpret_cast<ConnectionMap*> (tmp1);
        
        int alignProc = connections->begin()->first;

        ConnectionMap::const_iterator eIter = connections->lower_bound(myRank);
        for (;eIter!=connections->upper_bound(myRank);++eIter) {
          
          pEdge remoteEdge = (pEdge) eIter->second.remoteEntity;
          
          void* tmp2;
          EN_getDataPtr((pEntity) remoteEdge, tagEdge, (void**)&tmp2);
          ConnectionMap* remoteConnections = (ConnectionMap*) tmp2;
          
          int alignProcRemote = remoteConnections->begin()->first;
          
          pVertex p1Remote = E_vertex(remoteEdge,0);
          pVertex p2Remote = E_vertex(remoteEdge,1);
          
          int orientation = 0;
          if (V_corresponds(p1Remote,p1Local) && 
              V_corresponds(p2Remote,p2Local)) orientation =  1;
          if (V_corresponds(p1Remote,p2Local) && 
              V_corresponds(p2Remote,p1Local)) orientation = -1;
          
    
          if (alignProcRemote == alignProc) {
            if (orientation == -1) 
              msg.error(__LINE__,__FILE__,
                        "Got a local connection of non-aligned edges %d-%d to %d-%d on processor %d",
                        EN_id((pEntity) p1Remote),EN_id((pEntity) p2Remote),
                        EN_id((pEntity) p1Local) ,EN_id((pEntity) p2Local),myRank);
          }
          
          if (alignProcRemote < alignProc) {
            alignProc = alignProcRemote;
            if (orientation == -1) E_align(localEdge,p2Local,p1Local);
            
          }
        }
      }
    }
    EIter_delete(eit);

#ifdef PARALLEL
    MDB_CommCheck cc(mesh->getMPIComm());
#else
    MDB_CommCheck cc;
#endif
    
    exchangeDataOnEdges(mesh,cc);
  }
  
  // -------------------------------------------------------------------
  /*! \brief Establish face to face correspondance and assure coherent orientation 
    \ingroup parallel */ 
  //
  
  void F_createInfoInterface(pMesh mesh, pMeshDataId tagFace)
  {
    
    int myRank = 0;
    size_t nbRemoteFaces = 0;
    
    MAdMsg& msg = MAdMsgSgl::instance();
    pGModel model = M_model(mesh);
    
#ifdef PARALLEL
    if(mesh->getMPIComm()==MPI_COMM_SELF) return;
    int nproc;
    MPI_Comm_size(mesh->getMPIComm(), &nproc);
    MPI_Comm_rank(mesh->getMPIComm(), &myRank);
#endif

    // --- delete previous information ---

    FIter fit = M_faceIter(mesh);
    while ( pFace pf = FIter_next(fit) ) {
      void * tmp;
      if( EN_getDataPtr((pEntity) pf, tagFace, &tmp) ) {
        EN_deleteData((pEntity) pf, tagFace);
      }
    }
    FIter_delete(fit);

    // --- first establish local communication and coherent orientation ---
    //     * coherent reorientation on other partitions assumes coherent orientation here 
    //     * all periodic faces have same connections across partititions
    //       hence subsequent reorientation should remain coherent
    
    
    fit = M_faceIter(mesh);
    while ( pFace pf = FIter_next(fit) ) {
      int distProc = -1;

      std::vector<pVertex> distVt;
      std::vector<face_comm> rd;
      
      if ( F_isPotentialInterface(mesh, pf, rd) ) {
        
        std::vector<face_comm>::iterator rIter = rd.begin();

        pGEntity localClass = EN_whatIn((pEntity) pf);
        
        std::vector<int> lNodeID;
        for (int i=0;i<F_numVertices(pf);i++) lNodeID.push_back(EN_id((pEntity) F_vertex(pf,i)));
                
        for (;rIter!=rd.end();++rIter) {
          
          distProc = rIter->distProc;
          
          if (distProc == myRank) {
            
            face_comm& comm = *rIter;
            
            pVertex p1recv,p2recv,p3recv,p4recv;
            int nbNodes = (comm.p4 == NULL) ? 3:4;

            if (nbNodes != F_numVertices(pf)) {
              msg.error(__LINE__,__FILE__,
                        "Non-coherent number of vertices %i vs %i",
                        nbNodes,F_numVertices(pf));
            }
            
            p1recv = comm.p1;
            p2recv = comm.p2;
            p3recv = comm.p3;
            p4recv = comm.p4;
            
            assert(p1recv);assert(p2recv);assert(p3recv);
            
            pFace remoteFace = F_exist(p1recv,p2recv,p3recv,p4recv);
            
            if (remoteFace) {
              
	      // --- get the transformation if available

              pGEntity remoteClass = EN_whatIn((pEntity) remoteFace);
              pTransform tfo = GM_transformation(model,localClass,remoteClass);
	      if ( !tfo ) continue;
	      
              // --- check presence of multiple connections

              void* tmp = NULL;
              EN_getDataPtr((pEntity) pf,tagFace,&tmp);
              if (tmp != NULL) {
		ConnectionMap* thisConnection = (ConnectionMap*) tmp;
		ConnectionMap::iterator iteCon = thisConnection->find(myRank);
		if ( iteCon->second.remoteEntity == remoteFace ) continue;
		else msg.error(__LINE__,__FILE__,"Multiple connections for a face");
              }

              // --- compare ordered lists of nodes and reorder
              
              std::vector<int> rNodeID;
              int nv  = F_numVertices(remoteFace);
              for (int i=0;i<nv;i++) rNodeID.push_back(EN_id((pEntity) F_vertex(remoteFace,i)));
              
              if (rNodeID > lNodeID) {
                lNodeID = rNodeID; 
                F_align(remoteFace,p1recv,p2recv,p3recv,p4recv);
              }
              
              // --- attach the connection information

              ConnectionMap* connections = new ConnectionMap;
              connections->insert(std::make_pair(myRank,EntityConnection(remoteFace,tfo)));
              EN_attachDataPtr((pEntity) pf, tagFace, connections);

              // --- update the counter
              
              nbRemoteFaces++;
            }
          }
        }
      }
    }
    FIter_delete(fit);

#ifdef PARALLEL

    int *sendcounts = new int[nproc];
    for(int i=0;i<nproc;i++)sendcounts[i]=0;

    int commTag = 444;
    MAdResourceManagerSgl::instance().reserveCommunicationTag(commTag);

    // --- send faces ---     

    fit = M_faceIter(mesh);
    while ( pFace pf = FIter_next(fit) ) {
      int distProc = -1;
      std::vector<pVertex> distVt;
      
      std::vector<face_comm> rd;
      
      if ( F_isPotentialInterface(mesh, pf, rd) ) {
        
        std::vector<face_comm>::iterator rIter = rd.begin();
        
        for (;rIter!=rd.end();++rIter) {
          
          distProc = rIter->distProc;
          
          if (distProc != myRank) {
            
            void *buf = AP_alloc(distProc,commTag,sizeof(face_comm));
            face_comm *castbuf = (face_comm *) buf;
            std::memcpy(buf,&(*rIter),sizeof(face_comm));
            castbuf->distProc = myRank;
            
            AP_send(buf);
            sendcounts[distProc]++;
          }
        }
      }
    }
    FIter_delete(fit);

    // --- synchronise sends --- 

    AP_check_sends(AP_NOFLAGS);
    AP_reduce_nsends(sendcounts);
    //AP_flush();
  
    // --- Receive face info ---
    int message=0, count;
    while (!AP_recv_count(&count) || message<count) {
      void *recv;
      int from;
      int tag;
      int recvSize;
      
      if (AP_recv(MPI_ANY_SOURCE,commTag, AP_BLOCKING|AP_DROPOUT,
                  &recv, &recvSize, &from, &tag)) {
        
        message++;
        face_comm * castRecv = (face_comm*) recv;
        
        pVertex p1recv = castRecv->p1;
        pVertex p2recv = castRecv->p2;
        pVertex p3recv = castRecv->p3;
        pVertex p4recv = castRecv->p4;

        int id1recv = castRecv->id1;
        int id2recv = castRecv->id2;
        int id3recv = castRecv->id3;
        int id4recv = castRecv->id4;
      
        assert(p1recv);assert(p2recv);assert(p3recv);
        // p4recv = nbNodes == 4 ? castbuf -> p4 : NULL;
      
        int recvProc = castRecv->distProc;
        assert(recvProc==from);
      
        pFace pf = F_exist(p1recv,p2recv,p3recv,p4recv);
        
        if (pf) {

          pGEntity localClass = EN_whatIn((pEntity) pf);

          if (F_corresponds(pf,id1recv,id2recv,id3recv,id4recv)) {

            // --- check for multiple connections
            
            void* tmp = NULL;
            if (EN_getDataPtr((pEntity) pf,tagFace,&tmp)) { 
              msg.error(__LINE__,__FILE__,"Multiple connections for a face");
            }

            // --- reorient the face, lowest ranking processor is master
            
            int orientation = 1;
            if (from < myRank) orientation = F_align(pf,p1recv,p2recv,p3recv,p4recv);  
            if (orientation == 0) {
              msg.error(__LINE__,__FILE__,
                        "Could not align face (%i,%i,%i,%i) with remote connection on %i",
                        EN_id((pEntity) F_vertex(pf,0)),
                        EN_id((pEntity) F_vertex(pf,1)),
                        EN_id((pEntity) F_vertex(pf,2)),
                        F_numVertices(pf) == 4 ? EN_id((pEntity) F_vertex(pf,3)):-1,
                        from);
            }
            
            // --- get the transformation

            pGEntity remoteClass = GM_entityByTag(model,
                                                  castRecv->classTag,
                                                  castRecv->classDim);
            
            pTransform tfo = GM_transformation(model,localClass,remoteClass);

            // --- add the connection data to the face
            
            pFace remote = castRecv->pf;
            ConnectionMap* connections = new ConnectionMap;
            connections->insert(std::make_pair(recvProc,EntityConnection(remote,tfo)));
            EN_attachDataPtr((pEntity) pf, tagFace, connections);
            
            nbRemoteFaces++;

          }
          else {
            msg.info(__LINE__,__FILE__,
                     "non-corresponding face %d-%d-d% vs %d-%d-d%",
                     id1recv,id2recv,id3recv,
                     EN_id((pEntity) p1recv),
                     EN_id((pEntity) p2recv),
                     EN_id((pEntity) p3recv));
          }
        }
        AP_free(recv);
      }
    }

    // --- synchronise receives --- 

    MAdResourceManagerSgl::instance().releaseCommunicationTag(commTag);

    AP_check_sends(AP_WAITALL);
    MPI_Barrier(mesh->getMPIComm());

    // --- clean ship --- 
    
    delete [] sendcounts;

    MDB_CommCheck cc(mesh->getMPIComm());
#else
    MDB_CommCheck cc;
#endif
    
    exchangeDataOnFaces(mesh,cc);
  }
  
  // -------------------------------------------------------------------
#ifdef PARALLEL
  MPI_Comm& M_getMPIComm( pMesh pm){ return pm->getMPIComm(); }
#endif

  // -------------------------------------------------------------------

  bool P_reserveCommTag(int t) {
    return MAdResourceManagerSgl::instance().reserveCommunicationTag(t);
  }

  // -------------------------------------------------------------------

  bool P_releaseCommTag(int t) {
    return MAdResourceManagerSgl::instance().releaseCommunicationTag(t);
  }

    void exchangeDataOnVertices (pMesh m, MDB_DataExchanger &de )
  { 
  
    int mysize = 1;
    int myrank = 0;

#ifdef PARALLEL
    MPI_Comm_size(m->getMPIComm(), &mysize);
    MPI_Comm_rank(m->getMPIComm(), &myrank); 
#endif    
    
    int *sendcounts = new int[mysize];
    for(int i=0;i<mysize;i++)sendcounts[i]=0;
  
    pMeshDataId tagData = MD_lookupMeshDataId("RemotePoint");
    VIter vit = M_vertexIter(m);
    pVertex pv; 
    while ((pv = VIter_next(vit))) {
      void *tmp;
      if (EN_getDataPtr((pEntity) pv , tagData, &tmp)) {

        ConnectionMap* connections = reinterpret_cast<ConnectionMap* >(tmp);

        ConnectionMap::iterator cIter = connections->begin();
        
        for (;cIter!=connections->end();++cIter) {
          
          int sizebuf;      
          int remoteProc = cIter->first;  // (*recup)[j].first;
          pEntity remote = cIter->second.remoteEntity; // (*recup)[j].second;
          pTransform tfo = cIter->second.transform; 
    
          void *buf = de.sendData ((pEntity) pv, tfo, remoteProc, sizebuf );
          if (buf)
          {
            if (remoteProc == myrank)
            {
              de.receiveData (remote, myrank,buf);
              free(buf);
            }
            else
            {
#ifdef PARALLEL
              char *msg = (char*)AP_alloc(remoteProc,de.tag(),sizeof(pEntity*)+sizebuf);
              memcpy(&msg[0],&remote,sizeof(pEntity));
              memcpy(&msg[sizeof(pEntity)],buf,sizebuf);
              free(buf);
              AP_send(msg);
              sendcounts[remoteProc] ++;
#else
              throw;
#endif
            }
          } 
        }
      }
    }      
    VIter_delete(vit);  

#ifdef PARALLEL 
    if(m->getMPIComm() != MPI_COMM_SELF){

      AP_check_sends(AP_NOFLAGS);
      AP_reduce_nsends(sendcounts);
  
      int message=0;
      int count;
  
      while (!AP_recv_count(&count) || message<count)   { 
        void *msg;
        int from;
        int tag;
        int size;
        int rc;
        rc=AP_recv(MPI_ANY_SOURCE, de.tag(), AP_BLOCKING|AP_DROPOUT,
                   &msg, &size, &from, &tag);
        if (rc)  {
          message++;
          char * tmp = (char *) msg;
          pEntity * pvRcv = (pEntity *) &tmp[0];
#ifdef DEBUG    
          assert(*pvRcv);
#endif
          de.receiveData (*pvRcv,from, &tmp[sizeof(pEntity*)]);
          AP_free(msg);
        }
      }    
      
      AP_check_sends(AP_WAITALL); 
      MPI_Barrier(m->getMPIComm());
    }
#endif
    
    delete [] sendcounts;
  }

  // ----------------------------------------------------------------------------
  
  void exchangeDataOnEdges (pMesh m, MDB_DataExchanger &de )
  { 
    int mysize = 1;
    int myrank = 0;

#ifdef PARALLEL
    MPI_Comm_size(m->getMPIComm(), &mysize);
    MPI_Comm_rank(m->getMPIComm(), &myrank);
#endif    
  
    int *sendcounts = new int[mysize];
    for(int i=0;i<mysize;i++)sendcounts[i]=0;
    
    pMeshDataId tagData = MD_lookupMeshDataId("RemotePoint");
    EIter eit = M_edgeIter(m);
    pEdge pe;  
    while( (pe = EIter_next(eit)) ) {

      void* tmpptr = NULL;
      if (EN_getDataPtr(pe,tagData,&tmpptr)) {
        
        ConnectionMap* connections = reinterpret_cast<ConnectionMap*> (tmpptr);
        ConnectionMap::iterator cIter = connections->begin();
        
        for (;cIter!=connections->end();++cIter) {
          
          int remoteProc = cIter->first;
          pEntity remote = cIter->second.remoteEntity;
          pTransform tfo = cIter->second.transform;
          
          int sizebuf = 0;
          void *buf = de.sendData ((pEntity) pe, tfo, remoteProc, sizebuf );
#ifdef PARALLEL
          size_t sendSize = sizebuf + sizeof(pEntity);
#endif
  
          if (remoteProc == myrank) de.receiveData (remote,myrank,buf);
          else {
#ifdef PARALLEL
            char *msg = (char*)AP_alloc(remoteProc,de.tag(),sendSize);
            memcpy(&msg[0],&remote,sizeof(pEntity));
            memcpy(&msg[sizeof(pEntity)],buf,sizebuf);
            AP_send(msg);
            sendcounts[remoteProc] ++;
#else
            throw;
#endif
          }
          free(buf); 
        }
      }
    }

    EIter_delete(eit);
              

#ifdef PARALLEL 
    if(m->getMPIComm() != MPI_COMM_SELF){
      AP_check_sends(AP_NOFLAGS);
      AP_reduce_nsends(sendcounts);
  
      int message=0;
      int count;
  
      while (!AP_recv_count(&count) || message<count) 
  {
    void *msg;
    int from;
    int tag;
    int size;
    int rc;
    rc=AP_recv(MPI_ANY_SOURCE, de.tag(), AP_BLOCKING|AP_DROPOUT,
         &msg, &size, &from, &tag);
    if (rc) 
      {
        message++;
        char * tmp = (char *) msg;
        pEdge peRcv;
        memcpy(&peRcv,&tmp[0],sizeof(pEntity));
        de.receiveData ((pEntity) peRcv,from, &tmp[sizeof(pEntity)]);
        AP_free(msg);
      }
  }    
      AP_check_sends(AP_WAITALL); 
      MPI_Barrier(m->getMPIComm());
    }
#endif
    delete [] sendcounts;
  }

  // ----------------------------------------------------------------------------

  void exchangeDataOnFaces (pMesh m, MDB_DataExchanger &de )
  { 
    
    int mysize = 1;
    int myrank = 0;

#ifdef PARALLEL
    MPI_Comm_size(m->getMPIComm(), &mysize);
    MPI_Comm_rank(m->getMPIComm(), &myrank);
#endif    
  
    int *sendcounts = new int[mysize];
    for(int i=0;i<mysize;i++)sendcounts[i]=0;
    
    pMeshDataId tagData = MD_lookupMeshDataId("RemotePoint");
    FIter fit = M_faceIter(m);
    pFace pf;  
    while( (pf = FIter_next(fit)) ) {

      void* tmp = NULL;
      if (EN_getDataPtr((pEntity) pf,tagData,&tmp)) {
        
        ConnectionMap* connection = reinterpret_cast<ConnectionMap* > (tmp);
        
        if (connection->size() != 1) throw;
        
        int remoteProc = connection->begin()->first;
        pEntity remote = connection->begin()->second.remoteEntity;
        pTransform tfo = connection->begin()->second.transform;
        
        int sizebuf = 0;
        void *buf = de.sendData (pf, tfo, remoteProc, sizebuf );
#ifdef PARALLEL
        size_t sendSize = sizebuf + sizeof(pEntity);
#endif
        if (buf != NULL) {
          if (remoteProc == myrank) de.receiveData ((pEntity) remote,myrank,buf);
          else {
            
#ifdef PARALLEL
            char *msg = (char*)AP_alloc(remoteProc,de.tag(),sendSize);
            memcpy(&msg[0],&remote,sizeof(pEntity));
            memcpy(&msg[sizeof(pEntity)],buf,sizebuf);
            AP_send(msg);
            sendcounts[remoteProc] ++;
#else
            throw;
#endif
          }
          free(buf);
        }
      }
    }
    FIter_delete(fit);

#ifdef PARALLEL 
    if(m->getMPIComm() != MPI_COMM_SELF){

      AP_check_sends(AP_NOFLAGS);
      AP_reduce_nsends(sendcounts);
  
      int message=0;
      int count;
  
      while (!AP_recv_count(&count) || message<count) 
  {
    void *msg;
    int from;
    int tag;
    int size;
    int rc;
    rc=AP_recv(MPI_ANY_SOURCE, de.tag(), AP_BLOCKING|AP_DROPOUT,
         &msg, &size, &from, &tag);
    if (rc) 
      {
        message++;
        char * tmp = (char *) msg;
        pFace pfRcv;
        memcpy(&pfRcv,&tmp[0],sizeof(pEntity));
        de.receiveData ((pEntity) pfRcv,from, &tmp[sizeof(pEntity)]);
        AP_free(msg);
      }
  }    
      AP_check_sends(AP_WAITALL); 
      MPI_Barrier(m->getMPIComm());
    }
#endif
    delete [] sendcounts;
  }

  
  
}
