// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#ifndef H_MESHDATABASEPARALLELINTERFACE
#define H_MESHDATABASEPARALLELINTERFACE

#include "MeshDataBaseInterface.h"

#include <vector>

namespace MAd {
    
  /*! \brief functor base class providing call backs during parallel communication 
    \ingroup parallel */
    class MDB_DataExchanger
    {
    private:
        int tagComm;
    public:
        // get a tag for the communication
        int tag() const { return tagComm; }

        /*! \brief allocation call back for message from pe to corresponding entities */
        virtual void * sendData(pEntity pe,     //!< sending entity
            int iProcDest, //!< destination processor
            int &_size      //!< message size in byte
        ) = 0;

        /*! \brief allocation call back for message from pe to corresponding entities */
        virtual void * sendData(pEntity pe,          //!< sending entity
            const pTransform pt, //!< corresponding transformation 
            int iProcDest,      //!< destination processor
            int &_size           //!< message size in byte
        )
        {
            return sendData(pe, iProcDest, _size);
        }

        /*! \brief reception call back of message from element corresponding to pe, from iProc */
        /*! \warning should not delete the message, this is done externally */
        virtual void receiveData(pEntity pe,       //!< receiving entity
            int iProcSender, //!< sending processor
            void *buf         //!< message
        ) = 0;

        // In case the user has to delete a data when 'pe' is deleted
        virtual void deleteExternalData(pEntity pe) const {}

        MDB_DataExchanger(int);
        virtual ~MDB_DataExchanger();
    };

    class MDB_DataExchangerPeriodic
    {
    public:
        virtual int nbRefPeriodic() const = 0;
        virtual void   fperiodic(const int inv,
            const double X,
            const double Y,
            const double Z,
            int numref,      // number of the transformation
            double *Xnew,
            double *Ynew,
            double *Znew) = 0;
        virtual ~MDB_DataExchangerPeriodic() {};
    };

  // -------------------------------------------------------------------

#ifdef PARALLEL
  void M_writeParallel(pMesh, const char*, int version=2);
#endif 

  // -------------------------------------------------------------------

  /*! \brief Return true if on a parallel or periodic interface \ingroup internal */ 
  bool EN_isInterface(pEntity pv);

  /*! \brief Return true if on a parallel or periodic interface \ingroup parallel */ 
  bool V_isInterface(pVertex pv);

  /*! \brief Return true if pv is a (periodic) copy of vertex with tag id \ingroup parallel */
  bool V_corresponds(pVertex pv,int id);

  /*! \brief Return size of the list and list is an array containing 
    proc numbers where pv exist except calling proc \ingroup parallel */ 
  int  V_listInterface(pVertex pv, std::vector<int>* list);

  /*! \brief  Return true if edge is //possibly// on a parallel or periodic interface \ingroup internal */ 
  bool E_isPotentialInterface(pMesh, pEdge);

  /*! \brief  Return true if on a parallel or periodic interface \ingroup parallel */ 
  bool E_isInterface(pEdge);

  /*! \brief Return true if the edge corresponds either directly or periodically \ingroup parallel */
  bool E_corresponds(pEdge,int,int);
  
  /*! \brief  Return true if face is potentially on a parallel interface \ingroup internal */ 
  bool F_isPotentialInterface(pMesh, pFace);

  /*! \brief  Return true if face is on a parallel interface \ingroup parallel */ 
  bool F_isInterface(pFace);

  /*! \brief  Return true if on a parallel interface (always false) \ingroup parallel */ 
  bool R_isInterface(pRegion pr);

  // -------------------------------------------------------------------

  /*! \brief Fill the attached pointer containing the distant proc
    numbers and the distant node pointers \ingroup parallel */ 
  void V_createInfoInterface(pMesh mesh, pMeshDataId tagVertex);

  /*! \brief Fill the attached pointer containing the distant proc
    numbers and the distant edge pointers \ingroup parallel */ 
  void E_createInfoInterface(pMesh mesh, pMeshDataId tagEdge);

  /*! \brief Fill the attached pointer containing the distant proc
    numbers and the distant face pointers \ingroup parallel */ 
  void F_createInfoInterface(pMesh mesh, pMeshDataId tagFace);
  
  // -------------------------------------------------------------------
#ifdef PARALLEL

  // interface elt migration
  void Balance(pMesh mesh,MDB_DataExchanger &de);
  void Balance2(pMesh mesh,MDB_DataExchanger &de);
  void BalanceRandom(pMesh mesh,MDB_DataExchanger &de);
  int BalanceManifold(pMesh mesh,MDB_DataExchanger &de);

#ifdef _HAVE_PARMETIS_
  // load balancing with metis
  void BalanceMetis(pMesh mesh,MDB_DataExchanger &de);
  void BalanceMetis2(pMesh mesh,MDB_DataExchanger &de);
#endif

  // migration of elt tagged tagElt
  void loadBalancing(pMesh mesh,pMeshDataId tagElt, MDB_DataExchanger &de);  
  void loadBalancing2(pMesh mesh,pMeshDataId tagElt, MDB_DataExchanger &de);
#endif

  //Move periodic interfaces
  void BalancePeriodic(pMesh mesh,int dim,MDB_DataExchanger &de,
                       MDB_DataExchangerPeriodic &deperiodic,std::vector<std::vector<int> >& transfo);

// -------------------------------------------------------------------

//migration of periodic elt
void PeriodicInterfaceMigration(MAd::pMesh mesh,MAd::pMeshDataId tagElt,MAd::pMeshDataId tagMove,
                                MAd::pMeshDataId tagTransfo, MDB_DataExchanger &de,
                                MDB_DataExchangerPeriodic &deperiodic); 


void GroupPeriodicTetra(MAd::pMesh mesh, MDB_DataExchanger &de,
                        MDB_DataExchangerPeriodic &deperiodic);

// -------------------------------------------------------------------
#ifdef PARALLEL  
  /*! \brief Return the mpi communicator associated to the mesh.
       This function should be used to set the MPI communicator in MPI
       instructions related to the mesh \ingroup parallel */
  MPI_Comm& M_getMPIComm( pMesh );
#endif
  
  // -------------------------------------------------------------------
  /*! \brief register communication tag to avoid collisions with
    internal exchanges \ingroup parallel */
  bool P_reserveCommTag(int);
  
  /*! \brief release communication tag to allow use by others 
    \ingroup parallel */
  bool P_releaseCommTag(int);

  /*! \brief apply call back to exchanges on nodes 
    \ingroup parallel */

  void exchangeDataOnVertices (pMesh m, MDB_DataExchanger &de );

  /*! \brief apply call back to exchanges on edges
    \ingroup parallel */

  void exchangeDataOnEdges    (pMesh m, MDB_DataExchanger &de );

  /*! \brief apply call back to exchanges on faces
    \ingroup parallel */
  
  void exchangeDataOnFaces    (pMesh m, MDB_DataExchanger &de );

}

#endif
