// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#include "MeshDataBase.h"
#include "MAdMapping.h"
#include "MappingInterface.h"
#include "MeshDataBaseInterface.h"
#include "MeshDataBaseExtrusion.h"
#include "MeshDataBaseUntangling.h"

#include <limits>
#include <cstdlib>
#include <string>

// ---------------------------------------------------------------------------

class Criterion {

public:

  Criterion(int m):maxLayers(m){;}

public:
    
  bool operator()(pFace pf, pRegion pr, int l, pMeshDataId tag) const {

    void *tmp = NULL;
    if(!pf || !pr) return false;
    if(l>=maxLayers) return false;
    if(!isExtruded(pr,pf)) return false;
    if(EN_getDataPtr(pr,tag,&tmp)) return false;

    return true;
  } 

private:

  int maxLayers;

};

// ---------------------------------------------------------------------------

class Layer{
  
public:
  
  Layer(int i,double xyz[3], double h){
    index = i; height = h;
    x = xyz[0]; y = xyz[1]; z = xyz[2];
  };  
  
public:
  
  int getIndex() const {return index;}
  void setHeight(double h) {height = h;}
  double getHeight() const {return height;}
  void getCoordinates(double* xyz) const {
    xyz[0] = x; xyz[1] = y; xyz[2] = z;
  };

private:

  int index;
  double x,y,z;
  double height;
};

// ---------------------------------------------------------------------------
  
double distance(double a[3], double b[3]){

  return sqrt((a[0]-b[0])*(a[0]-b[0]) 
              +(a[1]-b[1])*(a[1]-b[1])
              +(a[2]-b[2])*(a[2]-b[2]));
}

// ---------------------------------------------------------------------------

bool getParametricCoordinates(pMapping mapping, double* coord, double* uvw){

  int nbKnots = mapping->getNumberOfKnots();

  double xyz[3];

  for(int i=0; i<nbKnots; i++){
    mapping->getVertexUVW(i,uvw[0],uvw[1],uvw[2]);
    mapping->eval(uvw[0],uvw[1],uvw[2],xyz[0],xyz[1],xyz[2]);
    if(distance(coord,xyz)<1e-12) return true;
  }

  return false;
}
  
// ---------------------------------------------------------------------------

double R_projectToFace(pRegion pr, pFace pf, double* uvw){

  const MAdMsg& msg = MAdMsgSgl::instance();

  int basicTag = EN_mshBasicTag((pEntity) pr);

  double uvw0[3] = {uvw[0],uvw[1],uvw[2]};

  switch(basicTag){
  case MSH_HEX_8:
    { 
      if(pf == R_face(pr,0)){ uvw[2] = -1.; break;}
      if(pf == R_face(pr,5)){ uvw[2] = +1.; break;}
      if(pf == R_face(pr,1)){ uvw[1] = -1.; break;}
      if(pf == R_face(pr,3)){ uvw[1] = +1.; break;}
      if(pf == R_face(pr,4)){ uvw[0] = -1.; break;}
      if(pf == R_face(pr,2)){ uvw[0] = +1.; break;}
    }
  case MSH_PRI_6:
    { 
      if(pf == R_face(pr,0)){ uvw[2] = -1.; break;}
      else if(pf == R_face(pr,1)){ uvw[2] = +1.; break;}
      else msg.error(__LINE__,__FILE__,
                     "Encountered non-extrusion face in prism",
                     basicTag);

    }
  default:
    msg.error(__LINE__,__FILE__,
              "Unsupported element type %d",
              basicTag);
      
  }    

  return distance(uvw,uvw0);
}

// ---------------------------------------------------------------------------

std::vector<Layer> layerPositions(pMesh theMesh,
                                  pFace pf,
                                  pRegion pr,
                                  pMeshDataId untangledId,
                                  const Criterion& criterion) {

  int index = 0;
  double totalHeight = 0;
  std::vector<Layer> stack;
  
  if(!criterion(pf,pr,index,untangledId)) return stack;

  while( criterion(pf,pr,index,untangledId) ){
     
    // --- get centroid
            
    double xyz[3];
    F_center(pf,xyz);

    // --- insert in stack           
            
    if(stack.size()==0){
      totalHeight = 0;
      stack.push_back(Layer(index,xyz,totalHeight));
    }
    else{
      double xyz0[3];
      stack.back().getCoordinates(xyz0);
      totalHeight += distance(xyz,xyz0);
      stack.push_back(Layer(index,xyz,totalHeight));
    }

    index++;

    // --- move to next element in stack

    int newOrientation = 0;            
    pf = R_gtOppFace(pr,pf,newOrientation);
    pr = F_otherRegion(pf,pr);     
  }
   
  // --- add last face

  if(pf){ 
    double xyz[3];
    double xyz0[3];
    F_center(pf,xyz);
    stack.back().getCoordinates(xyz0);
    totalHeight += distance(xyz,xyz0);
    stack.push_back(Layer(index,xyz,totalHeight));  
    index++;
  }
          
  // --- calculate position

  std::vector<Layer>::iterator it = stack.begin();
  for(; it!=stack.end(); it++){
    double height = it->getHeight();
    if(totalHeight > 1e-12) it->setHeight(height/totalHeight);
  }
        
  // --- return / exit

  return stack;
}

// ---------------------------------------------------------------------------
  
void curveElementStack(pMesh theMesh,
                       pFace pf,
                       pRegion pr,
                       pMeshDataId untangledId,
                       const Criterion& criterion){

  const MAdMsg& msg = MAdMsgSgl::instance();

  int index  = 0;

  // --- position of each layer in element stack

  std::vector<Layer> stack = layerPositions(theMesh,pf,pr,untangledId,criterion);
  
  while(criterion(pf,pr,index,untangledId)) {
          
    // --- curve element (located above the face)

    pPList rPoints   = R_pointsTot(pr);
    pMapping mapping = Map_create((pEntity) pr,1.0);   
       
    void * temp = NULL;
    while (pPoint pP = (pPoint) PList_next(rPoints, &temp)) {
         
      void* buff = NULL;
      if(!EN_getDataPtr((pEntity) pP,untangledId,&buff)){
              
        double uvw[3];
        double xyz[3];
           
        xyz[0] = P_x(pP);
        xyz[1] = P_y(pP);
        xyz[2] = P_z(pP);
              
        // --- get parametric nodes                             

        if(!getParametricCoordinates(mapping,xyz,uvw)){
          msg.error(__LINE__,__FILE__,
                    "Failed to find parametric coordinates");
        }
                
        // --- eval linear mapping
           
        mapping->evalLinear(uvw[0],uvw[1],uvw[2],
                            xyz[0],xyz[1],xyz[2]);

        // --- project parametric coordinate
                
        double pDist = R_projectToFace(pr,pf,uvw);
                
        // --- evaluate mappings
           
        double xyzC[3]; // curved
        double xyzL[3]; // linear
          
        mapping->eval(uvw[0],uvw[1],uvw[2],
                      xyzC[0],xyzC[1],xyzC[2]);
          
        mapping->evalLinear(uvw[0],uvw[1],uvw[2],
                            xyzL[0],xyzL[1],xyzL[2]);
          
        // --- fetch layer height
                
        double h0 = stack[index].getHeight();
        double h1 = stack[index+1].getHeight();
        double fact1 = (1-h1)/(1-h0);              
        double fact  = (1-0.5*pDist) * 1.0 + (0.5*pDist) * fact1;
          
        //double fact = 1;
        //printf("%d %e %e %e %e %e\n",index,fact,fact1,h0,h1,pDist);
          
        // --- propagate deformation
          
        xyz[0] += fact*(xyzC[0] - xyzL[0]);
        xyz[1] += fact*(xyzC[1] - xyzL[1]);
        xyz[2] += fact*(xyzC[2] - xyzL[2]);
          
        P_setPos(pP,xyz[0],xyz[1],xyz[2]);
           
        // --- flag node
              
        void* untangled = (void*) new bool(true);
        EN_attachDataPtr((pEntity) pP,untangledId,untangled);
               
      }       
    }
      
    Map_delete(mapping);    
    PList_delete(rPoints);
            
    // --- flag element

    void* untangled = (void*) new bool(true);
    EN_attachDataPtr((pEntity) pr,untangledId,untangled);

    // --- move to next element in stack

    int newOrientation = 0;            
    pf = R_gtOppFace(pr,pf,newOrientation);
    pr = F_otherRegion(pf,pr);     
       
    index++;
  } 
}

// ---------------------------------------------------------------------------
  
void propagateOutwards(pMesh theMesh,
                       const std::set<pGFace>& beginFaces, 
                       pMeshDataId untangledId,
                       const Criterion& criterion){

  FIter fIter = M_faceIter(theMesh);
  while (pFace pf = FIter_next(fIter)) {

    void* buff = NULL;
      
    // --- curved element left ?

    bool curvedL = false;      
    if(pRegion pr = F_region(pf,0)){
      curvedL = (EN_getDataPtr((pEntity) pr,untangledId,&buff) != 0);
    }

    // --- curved element right ?

    bool curvedR = false;      
    if(pRegion pr = F_region(pf,1)){
      curvedR = (EN_getDataPtr((pEntity) pr,untangledId,&buff) != 0);
    }

    pRegion pr = NULL;
    if(curvedR && !curvedL) pr = F_region(pf,0); // Left wasnt curved
    if(curvedL && !curvedR) pr = F_region(pf,1); // Right wasnt curved

    // --- curve element stack

    curveElementStack(theMesh,pf,pr,untangledId,criterion);

  }
  FIter_delete(fIter);

}

namespace MAd {

  // ---------------------------------------------------------------------------
  
  void M_untangleBoundaryLayers(pMesh theMesh,
                                const std::set<int>& beginTags,
                                int nbLayers) {

    pGModel pgm = M_model(theMesh);

    std::set<pGFace> beginFaces;
    
    GFIter gfIter = GM_faceIter(pgm);
    while (pGFace pgf = GFIter_next(gfIter)) {
      int tag = GEN_tag((pGEntity) pgf);
      if (beginTags.find(tag) != beginTags.end()) beginFaces.insert(pgf);
    }
    GFIter_delete(gfIter);
    
    M_untangleBoundaryLayers(theMesh,beginFaces,nbLayers);
  }

  // ---------------------------------------------------------------------------

  void M_untangleBoundaryLayers(pMesh theMesh,
                                const std::set<pGFace>& beginFaces,
                                int nbLayers) {

    // --- flag extrusion regions

    std::string name = "boundaryLayerExtrusions_untangled";

    // --- prepare untangling flag

    pMeshDataId untangledId = MD_newMeshDataId(name.c_str());

    // --- criterion

    Criterion criterion(nbLayers);

    // --- loop faces and curve element stack

    std::set<pGFace>::const_iterator gIter=beginFaces.begin();
    for (; gIter!=beginFaces.end(); ++gIter) {

      pGFace pgf = *gIter;
      FIter fIter = M_classifiedFaceIter(theMesh,(pGEntity) pgf,0);
      while (pFace pf = FIter_next(fIter)) { 

        pRegion pr = F_region(pf,0);
        curveElementStack(theMesh,pf,pr,untangledId,criterion);

      }
      FIter_delete(fIter);
    }   

    // --- propagate deformation outwards

    propagateOutwards(theMesh,beginFaces,untangledId,criterion);
 
    // --- clean elements

    RIter rIter = M_regionIter(theMesh);
    while (pRegion pr = RIter_next(rIter)) {

      pr->deleteDataPtr<bool>(untangledId);

    }
    RIter_delete(rIter);

    // --- clean principal vertices
    
    VIter vIter = M_vertexIter(theMesh);
    while (pVertex pv = VIter_next(vIter)) {

      pv->deleteDataPtr<bool>(untangledId);

    }
    VIter_delete(vIter);

    // --- clean high-order points
    
    vIter = M_pointIter(theMesh);
    while (pVertex pv = VIter_next(vIter)) {

      pv->deleteDataPtr<bool>(untangledId);

    }
    VIter_delete(vIter);

  }

};
