// -*- C++ -*- 
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#include "MshTags.h"
#include "MAdMessage.h"

#include <sstream>
#include <cstdlib>
#include <cstring>
#include <string>


static const int MSH_LIN_EDGE_NODES[2]     = {0,1};

static const int MSH_TRI_EDGE[3] = {0,1,2};
static const int MSH_TRI_EDGE_NODES[3][2]  = {{0,1},{1,2},{2,0}};

static const int MSH_QUA_EDGE[4] = {0,1,2,3};
static const int MSH_QUA_EDGE_NODES[4][2]  = {{0,1},{1,2},{2,3},{3,0}};

static const int MSH_TET_EDGE[6] = {0,1,2,3,5,4};
static const int MSH_TET_EDGE_NODES[6][2]  = {{0,1},{1,2},{2,0},
                                              {3,0},{3,2},{3,1}};

static const int MSH_HEX_EDGE[12] = {0,3,4,1,5,2,6,7,8,11,9,10};
static const int MSH_HEX_EDGE_NODES[12][2] = {{0,1},{0,3},{0,4},
                                              {1,2},{1,5},
                                              {2,3},{2,6},
                                              {3,7},
                                              {4,5},{4,7},
                                              {5,6},
                                              {6,7}};

static const int MSH_PRI_EDGE[9] = {0,1,6,2,7,8,3,4,5};
static const int MSH_PRI_EDGE_NODES[9][2]  = {{0,1},{0,2},{0,3},
                                              {1,2},{1,4},
                                              {2,5},
                                              {3,4},{3,5},
                                              {4,5}};

static const int MSH_PYR_EDGE[8] = {0,1,2,3,4,5,6,7};
static const int MSH_PYR_EDGE_NODES[8][2]  = {{0,1},{0,3},{0,4},
                                              {1,2},{1,4},
                                              {2,3},{2,4},
                                              {3,4}};

// Gmsh face (node) numbers in MAdLib definition

static const int MSH_TRI_FACE_NODES[3]    = {0,1,2};
static const int MSH_QUA_FACE_NODES[4]    = {0,1,2,3};


static const int MSH_HEX_FACE[6] = {0,1,4,2,3,5};
static const int MSH_HEX_FACE_NODES[6][4] = {{0,3,2,1},
                                             {0,1,5,4},
                                             {0,4,7,3},
                                             {1,2,6,5},
                                             {2,3,7,6},
                                             {4,5,6,7}};
static const int MSH_TET_FACE[4] = {0,1,3,2};
static const int MSH_TET_FACE_NODES[4][3] = {{0,2,1},
                                             {0,1,3},
                                             {0,3,2},
                                             {3,1,2}};

static const int MSH_PRI_FACE[6] = {0,1,2,4,3};
static const int MSH_PRI_FACE_NODES[6][4] = {{0,2,1,-1},
                                             {3,4,5,-1},
                                             {0,1,4, 3},
                                             {0,3,5, 2},
                                             {1,2,5, 4}};

static const int MSH_PYR_FACE[5] = {0,1,2,3,4};
static const int MSH_PYR_FACE_NODES[5][4] = {{0,1,4,-1},
                                             {3,0,4,-1},
                                             {1,2,4,-1},
                                             {2,3,4,-1},
                                             {0,3,2,1}};


namespace  MAd {

  
  std::string getElementName(int type) {

    std::string name;
  
    switch (type) {
    case MSH_PNT     : name = "Point"; break;
    case MSH_LIN_2   : name = "Linear edge"; break;
    case MSH_LIN_3   : name = "Quadratic edge"; break;
    case MSH_LIN_4   : name = "Cubic edge"; break;
    case MSH_LIN_5   : name = "Quartic edge"; break;
    case MSH_TRI_3   : name = "Linear triangle"; break;
    case MSH_TRI_6   : name = "Quadratic triangle"; break;
    case MSH_TRI_9   : name = "Cubic serendipity triangle"; break;
    case MSH_TRI_10  : name = "Cubic triangle"; break;
    case MSH_TRI_12  : name = "Quartic serendipity triangle"; break;
    case MSH_TRI_15  : name = "Quartic triangle"; break;
    case MSH_QUA_4   : name = "Bilinear Quadrangle"; break;
    case MSH_QUA_8   : name = "Quadratic serendipity quadrangle"; break;
    case MSH_QUA_9   : name = "Quadratic quadrangle"; break;
    case MSH_QUA_16  : name = "Cubic quadrangle"; break;
    case MSH_QUA_25  : name = "Quartic quadrangle"; break;
    case MSH_TET_4   : name = "Linear tetrahedron"; break;
    case MSH_TET_10  : name = "Quadratic tetrahedron"; break;
    case MSH_HEX_8   : name = "Trilinear hexahedron"; break;
    case MSH_HEX_20  : name = "Quadratic edge serendipity hexahedron"; break;
    case MSH_HEX_27  : name = "Quadratic hexahedron"; break;
    case MSH_HEX_64  : name = "Cubic hexahedron"; break;
    case MSH_HEX_125 : name = "Quartic hexahedron"; break;
    case MSH_PRI_6   : name = "Linear prism"; break;
    case MSH_PRI_15  : name = "Quadratic edge serendipity prism"; break;
    case MSH_PRI_18  : name = "Quadratic serendipity prism"; break;
    case MSH_PYR_5   : name = "Linear pyramid"; break;
    case MSH_PYR_13  : name = "Quadratic edge serendipity pyramid"; break;
    case MSH_PYR_14  : name = "Quadratic pyramid"; break;
    case MSH_TET_20  : name = "Cubic tetrahedron";break;
    case MSH_TET_35  : name = "Quartic tetrahedron";break;
    case MSH_TET_34  : name = "Quartic serendipity tetrahedron";break;
    default:
        MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                    "Getting element type name for unsupported type %d\n", 
                                    type);     
        name = "unknown element";
        break;
    }
    return name;
  }



  //! get total number of nodes for a given element type
  
  int getNumVerticesForElementTypeMSH(int type) {
    
    switch (type) {

    case MSH_PNT     : return 1;

    case MSH_LIN_2   : return 2;
    case MSH_LIN_3   : return 3;
    case MSH_LIN_4   : return 4;
    case MSH_LIN_5   : return 5;

    case MSH_TRI_3   : return 3;
    case MSH_TRI_6   : return 6;
    case MSH_TRI_9   : return 9;
    case MSH_TRI_10  : return 10;
    case MSH_TRI_12  : return 12;
    case MSH_TRI_15  : return 15;

    case MSH_QUA_4   : return 4;
    case MSH_QUA_8   : return 8;
    case MSH_QUA_9   : return 9; 
    case MSH_QUA_12  : return 12;
    case MSH_QUA_16  : return 16;
    case MSH_QUA_25  : return 25;

    case MSH_TET_4   : return 4;
    case MSH_TET_10  : return 10;
    case MSH_TET_20  : return 20;
    case MSH_TET_35  : return 35;
    case MSH_TET_34  : return 34;

    case MSH_HEX_8   : return 8;
    case MSH_HEX_20  : return 20;
    case MSH_HEX_27  : return 27;
    case MSH_HEX_64  : return 64;
    case MSH_HEX_125 : return 125;

    case MSH_PRI_6   : return 6;
    case MSH_PRI_15  : return 15;
    case MSH_PRI_18  : return 18;

    case MSH_PYR_5   : return 5;
    case MSH_PYR_13  : return 13;
    case MSH_PYR_14  : return 14;

    default: 
      {
        MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                    "Getting number of points for unsupported type %d\n", 
                                    type);
        break;
      }
    }
    return 0;
  }

  //! return dimension for a given element type
  
  int getDimForElementTypeMSH(int type) {

    int dim(-1);
    
    switch (type) {
    case MSH_PNT     : dim =  0; break;

    case MSH_LIN_2   : 
    case MSH_LIN_3   : 
    case MSH_LIN_4   : 
    case MSH_LIN_5   : dim =  1; break;

    case MSH_TRI_3   : 
    case MSH_TRI_6   : 
    case MSH_TRI_9   : 
    case MSH_TRI_10  : 
    case MSH_TRI_12  : 
    case MSH_TRI_15  : 

    case MSH_QUA_4   : 
    case MSH_QUA_8   : 
    case MSH_QUA_9   :     
    case MSH_QUA_12  : 
    case MSH_QUA_16  : 
    case MSH_QUA_25  : dim =  2; break;

    case MSH_TET_4   : 
    case MSH_TET_10  : 
    case MSH_TET_20  : 
    case MSH_TET_35  : 
    case MSH_TET_34  : 

    case MSH_HEX_8   : 
    case MSH_HEX_20  : 
    case MSH_HEX_27  : 
    case MSH_HEX_64  :
    case MSH_HEX_125 :

    case MSH_PRI_6   : 
    case MSH_PRI_15  : 
    case MSH_PRI_18  : 

    case MSH_PYR_5   : 
    case MSH_PYR_13  : 
    case MSH_PYR_14  : dim =  3; break;

    default: 
      {
        MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                    "Getting element dimension for unsupported type %d\n", 
                                    type);
        dim = -1;
        break;
      }
    }
    return dim;
  }

  //! return human readable name for given element type
  
  void getElementNameMSH(int type, char * name)
  {
    std::string nameMSH = getElementName(type);
    strcpy(name,nameMSH.c_str());
  }
  
  //! get element type for generic order type
  
  int getElementTypeMSH(int type) {

    int bType;
  
    switch (type) {
    case MSH_PNT     : bType = MSH_TYPE_PNT; break;

    case MSH_LIN_2   :
    case MSH_LIN_3   :
    case MSH_LIN_4   :
    case MSH_LIN_5   : bType = MSH_TYPE_LIN; break;

    case MSH_TRI_3   : 
    case MSH_TRI_6   : 
    case MSH_TRI_9   : 
    case MSH_TRI_10  : 
    case MSH_TRI_12  : 
    case MSH_TRI_15  : bType = MSH_TYPE_TRI; break;

    case MSH_QUA_4   : 
    case MSH_QUA_8   : 
    case MSH_QUA_9   : 
    case MSH_QUA_12  : 
    case MSH_QUA_16  : 
    case MSH_QUA_25  : bType = MSH_TYPE_QUA; break;

    case MSH_TET_4   : 
    case MSH_TET_10  : 
    case MSH_TET_20  : 
    case MSH_TET_35  : 
    case MSH_TET_34  : bType = MSH_TYPE_TET;break;

    case MSH_HEX_8   : 
    case MSH_HEX_20  : 
    case MSH_HEX_27  : 
    case MSH_HEX_64  : 
    case MSH_HEX_125 : bType = MSH_TYPE_HEX ; break;

    case MSH_PRI_6   : 
    case MSH_PRI_15  : 
    case MSH_PRI_18  : bType = MSH_TYPE_PRI; break;

    case MSH_PYR_5   : 
    case MSH_PYR_13  : 
    case MSH_PYR_14  : bType = MSH_TYPE_PYR; break;

    default:
      {
        MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                    "Getting element type for unsupported element type %d\n", 
                                    type);
        bType = MSH_TYPE_UNKNOWN;
        break;
      }
    }
    return bType;
  }
  
  
  // ---------------------------------------------------------------------------
  //! get linear counterpart / basic element type 
  
  int getBasicElementTypeMSH(int type) {
    
    int basicType = MSH_UNDEFINED_ELEM;
  
    switch (type) {
    case MSH_PNT     : basicType = MSH_PNT; break;

    case MSH_LIN_2   :
    case MSH_LIN_3   :
    case MSH_LIN_4   :
    case MSH_LIN_5   : basicType = MSH_LIN_2; break;

    case MSH_TRI_3   : 
    case MSH_TRI_6   : 
    case MSH_TRI_9   : 
    case MSH_TRI_10  : 
    case MSH_TRI_12  : 
    case MSH_TRI_15  : basicType = MSH_TRI_3; break;

    case MSH_QUA_4   : 
    case MSH_QUA_8   : 
    case MSH_QUA_9   : 
    case MSH_QUA_12  : 
    case MSH_QUA_16  : 
    case MSH_QUA_25  : basicType = MSH_QUA_4; break;

    case MSH_TET_4   : 
    case MSH_TET_10  : 
    case MSH_TET_20  : 
    case MSH_TET_35  : 
    case MSH_TET_34  : basicType = MSH_TET_4;break;

    case MSH_HEX_8   : 
    case MSH_HEX_20  : 
    case MSH_HEX_27  : 
    case MSH_HEX_64  : 
    case MSH_HEX_125 : basicType = MSH_HEX_8 ; break;

    case MSH_PRI_6   : 
    case MSH_PRI_15  : 
    case MSH_PRI_18  : basicType = MSH_PRI_6; break;

    case MSH_PYR_5   : 
    case MSH_PYR_13  : 
    case MSH_PYR_14  : basicType = MSH_PYR_5; break;

    
    default:
      {
        MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                    "Getting linear equivalent for unsupported element type %d\n", 
                                    type);
        break;
      }
    }
    return basicType;
  }
  
  // ---------------------------------------------------------------------------
  //! return human readable name for basic element 
    
  void getBasicElementNameMSH(int type, char * name) {
    getElementNameMSH(getBasicElementTypeMSH(type), name);
  }
  
  // ---------------------------------------------------------------------------
  //! return interpolation order for element type
    
  int getMappingOrderMSH(int type) {
    
    switch (type) {
    case MSH_PNT     : return 0; break;

    case MSH_LIN_2   : 
    case MSH_TRI_3   : 
    case MSH_QUA_4   : 
    case MSH_TET_4   : 
    case MSH_HEX_8   : 
    case MSH_PRI_6   : 
    case MSH_PYR_5   : return 1; break;
      
    case MSH_LIN_3   : 
    case MSH_TRI_6   : 
    case MSH_QUA_8   : 
    case MSH_QUA_9   : 
    case MSH_TET_10  : 
    case MSH_HEX_20  : 
    case MSH_HEX_27  : 
    case MSH_PRI_15  : 
    case MSH_PRI_18  : 
    case MSH_PYR_13  : 
    case MSH_PYR_14  : return 2; break;

    case MSH_LIN_4   : 
    case MSH_TRI_9   : 
    case MSH_TRI_10  : 
    case MSH_QUA_12  :
    case MSH_QUA_16  :
    case MSH_TET_20  : 
    case MSH_HEX_64  : return 3; break;
      
    case MSH_LIN_5   : 
    case MSH_TRI_12  : 
    case MSH_TRI_15  : 
    case MSH_QUA_25  :
    case MSH_TET_35  : 
    case MSH_TET_34  : 
    case MSH_HEX_125 : return 4; break;

    default: 
      MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                  "Getting order of unsupported type of element %d\n", 
                                  type);
    }
    return -1;
  }

  
  // ---------------------------------------------------------------------------
  //! return interpolation order for element type
    
  bool isSerendipityMSH(int type) {
    
    bool serendipity = false;

    switch (type) {
      
    case MSH_PNT     :

    case MSH_LIN_2   :
    case MSH_LIN_3   :
    case MSH_LIN_4   :
    case MSH_LIN_5   : 

    case MSH_TRI_3   : 
    case MSH_TRI_6   : 
    case MSH_TRI_10  : 
    case MSH_TRI_15  : 

    case MSH_QUA_4   : 
    case MSH_QUA_9   : 
    case MSH_QUA_16  : 
    case MSH_QUA_25  : 

    case MSH_TET_4   : 
    case MSH_TET_10  : 
    case MSH_TET_20  : 
    case MSH_TET_35  : 

    case MSH_HEX_8   : 
    case MSH_HEX_27  : 
    case MSH_HEX_64  : 
    case MSH_HEX_125 : 

    case MSH_PRI_6   : 
    case MSH_PRI_18  : 

    case MSH_PYR_5   : 
    case MSH_PYR_14  : 
      {
        serendipity = false;
        break;
      }

    case MSH_TRI_9   : 
    case MSH_TRI_12  : 
    case MSH_QUA_8   : 
    case MSH_QUA_12  : 
    case MSH_HEX_20  : 
    case MSH_TET_34  : 
    case MSH_PRI_15  : 
    case MSH_PYR_13  : 
      {
        serendipity = true;
        break;
      }


    default: 
      MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                  "Getting order of unsupported type of element %d\n", 
                                  type);
    }
    return serendipity;
  }


  //! return ordered edge nodes in reference element
    
  const int* getMshEdgePoints(int type,int edge,int& idx) {
    
    idx = -1;

    if (edge>=0) {
      switch (type) {
      case MSH_LIN_2   :
      case MSH_LIN_3   :
      case MSH_LIN_4   :
      case MSH_LIN_5   : 
        {
          if (edge>=1) return NULL;
          idx = 0;
          return MSH_LIN_EDGE_NODES;
          break;
        }
      case MSH_TRI_3   : 
      case MSH_TRI_6   : 
      case MSH_TRI_9   : 
      case MSH_TRI_10  : 
      case MSH_TRI_12  : 
      case MSH_TRI_15  : 
        {
          if (edge>=3) return NULL;
          idx = MSH_TRI_EDGE[edge];
          return MSH_TRI_EDGE_NODES[edge];
          break;
        }
      case MSH_QUA_4   : 
      case MSH_QUA_8   : 
      case MSH_QUA_9   : 
      case MSH_QUA_12  : 
      case MSH_QUA_16  : 
      case MSH_QUA_25  : 
        {
          if (edge>=4) return NULL;
          idx = MSH_QUA_EDGE[edge];
          return MSH_QUA_EDGE_NODES[edge];
          break;
        }
      case MSH_TET_4   : 
      case MSH_TET_10  : 
      case MSH_TET_20  : 
      case MSH_TET_35  : 
      case MSH_TET_34  : 
        {
          if (edge>=6) return NULL;
          idx = MSH_TET_EDGE[edge];
          return MSH_TET_EDGE_NODES[edge];
          break;
        }
      case MSH_HEX_8   : 
      case MSH_HEX_20  : 
      case MSH_HEX_27  : 
      case MSH_HEX_64  : 
      case MSH_HEX_125 :  
        {
          if (edge>=12) return NULL;
          idx = MSH_HEX_EDGE[edge];
          return MSH_HEX_EDGE_NODES[edge];
          break;
        }
      case MSH_PRI_6   : 
      case MSH_PRI_15  : 
      case MSH_PRI_18  :   
        {
          if (edge>=9) return NULL;
          idx = MSH_PRI_EDGE[edge];
          return MSH_PRI_EDGE_NODES[edge];
          break;
        }
      case MSH_PYR_5   : 
      case MSH_PYR_13  : 
      case MSH_PYR_14  : 
        {
          if (edge>=8) return NULL;
          idx = MSH_PYR_EDGE[edge];
          return MSH_PYR_EDGE_NODES[edge];
          break;
        }
      default:
        {
          MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                      "Getting edge points of unsupported type of element %d\n", 
                                      type);
          break;
        }
      }
    }
    return NULL;
  }
  
  //! return ordered face nodes in reference element 
    
  const int* getMshFacePoints(int type,int face,int& idx) {

    idx = -1;

    if (face>=0) {
      
      switch (type) {
      case MSH_TRI_3   : 
      case MSH_TRI_6   : 
      case MSH_TRI_9   : 
      case MSH_TRI_10  : 
      case MSH_TRI_12  : 
      case MSH_TRI_15  : 
        {
          if (face>=1) return NULL;
          idx = 0;
          return MSH_TRI_FACE_NODES;
          break;
        }
      case MSH_QUA_4   : 
      case MSH_QUA_8   : 
      case MSH_QUA_9   : 
      case MSH_QUA_12  : 
      case MSH_QUA_16  : 
      case MSH_QUA_25  : 
        {
          if (face>=1) return NULL;
          idx = 0;
          return MSH_QUA_FACE_NODES;
          break;
        }
      case MSH_TET_4   : 
      case MSH_TET_10  : 
      case MSH_TET_20  : 
      case MSH_TET_35  : 
      case MSH_TET_34  : 
        {
          if (face>=6) return NULL;
          idx = MSH_TET_FACE[face];
          return MSH_TET_FACE_NODES[face];
          break;
        }
      case MSH_HEX_8   : 
      case MSH_HEX_20  : 
      case MSH_HEX_27  : 
      case MSH_HEX_64  : 
      case MSH_HEX_125 :  
        {
          if (face>=6) return NULL;
          idx = MSH_HEX_FACE[face];
          return MSH_HEX_FACE_NODES[face];
          break;
        }
      case MSH_PRI_6   : 
      case MSH_PRI_15  : 
      case MSH_PRI_18  : 
       
        {
          if (face>=5) return NULL;
          idx = MSH_PRI_FACE[face];
          return MSH_PRI_FACE_NODES[face];
          break;
        }
      case MSH_PYR_5   : 
      case MSH_PYR_13  : 
      case MSH_PYR_14  : 
        {
          if (face>=8) return NULL;
          idx = MSH_PYR_FACE[face];
          return MSH_PYR_FACE_NODES[face];
          break;
        }
      default:
        {
          MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                      "Getting face points of unsupported type of element %d\n", 
                                      type);
          break;
        }
      }
    }
    return NULL;
  }
}

