// -*- C++ -*- 
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

// Gmsh mesh definitions 

#ifndef MSHTAGS__H
#define MSHTAGS__H

#include "MAdExport.h"

namespace MAd {
    
  // ---------------------------------------------------------------------------
  //! \brief Basic element type
  //! \ingroup api_msh_tags
  // ---------------------------------------------------------------------------
  
  enum MshElementType {
    
    MSH_TYPE_UNKNOWN  =  0, //!< unknown
    MSH_TYPE_PNT      =  1, //!< point
    MSH_TYPE_LIN      =  2, //!< edge
    MSH_TYPE_TRI      =  3, //!< triangle
    MSH_TYPE_QUA      =  4, //!< quadrangle
    MSH_TYPE_TET      =  5, //!< tetrahedron
    MSH_TYPE_PYR      =  6, //!< pyramid
    MSH_TYPE_PRI      =  7, //!< wedge
    MSH_TYPE_HEX      =  8, //!< hexahedron
    MSH_TYPE_POLYG    =  9, //!< polygon
    MSH_TYPE_POLYH    = 10  //!< polyhedron
  };

  // ---------------------------------------------------------------------------
  //! \brief element type 
  //! \ingroup api_mesh_tags
  // ---------------------------------------------------------------------------

  enum MshElementTag {
  
    MSH_PNT   = 15, //!< point

    // --- edges ----------------------------------------------------------------
 
    MSH_LIN_2 = 1,  //!< linear edge
    MSH_LIN_3 = 8,  //!< quadratic edge
    MSH_LIN_4 = 26, //!< cubic edge
    MSH_LIN_5 = 27, //!< quartic edge

    // --- complete triangles ---------------------------------------------------
 
    MSH_TRI_3  = 2,  //!< linear quadrangle
    MSH_TRI_6  = 9,  //!< quadratic quadrangle
    MSH_TRI_10 = 21, //!< cubic quadrangle
    MSH_TRI_15 = 23, //!< quartic quadrangle

    // --- serendipity triangles ------------------------------------------------

    MSH_TRI_9  = 20, //!< cubic serendipity quadrangle
    MSH_TRI_12 = 22, //!< quartic serendipity quadrangle

    // --- complete quadrilaterals ----------------------------------------------

    MSH_QUA_4  = 3, //!< quadratic quadrangle
    MSH_QUA_9  = 10, //!< quadratic quadrangle
    MSH_QUA_16 = 36, //!< cubic quadrangle
    MSH_QUA_25 = 37 , //!< quartic quadrangle

    // --- serendipity quadrilaterals -------------------------------------------

    MSH_QUA_8  = 16, //!< quadratic serendipity quadrangle
    MSH_QUA_12 = 39, //!< cubic serendipity quadrangle

    // --- complete tetrahedra --------------------------------------------------

    MSH_TET_4  = 4, //!< linear tetrahedron
    MSH_TET_10 = 11, //!< quadratic tetrahedron
    MSH_TET_20 = 29, //!< cubic tetrahedron
    MSH_TET_35 = 30, //!< quadratic tetrahedron

    // --- incomplete tetrahedra ------------------------------------------------

    MSH_TET_34 = 32, //!< face serendipity tetrahedron

    // --- hexahedra ------------------------------------------------------------
 
    MSH_HEX_8   = 5 , //!< linear hexahedron
    MSH_HEX_27  = 12, //!< quadratic hexahedron
    MSH_HEX_64  = 92, //!< cubic hexahedron
    MSH_HEX_125 = 93, //!< quartic hexahedron

    // --- incomplete hexahedra -------------------------------------------------
 
    MSH_HEX_20 = 17, //!< edge serendipity quadratic hexahedron

    // --- wedges ---------------------------------------------------------------

    MSH_PRI_6  = 6,  //!< linear wedge
    MSH_PRI_18 = 13, //!< quadratic wedge
    MSH_PRI_40 = 90, //!< cubic wedge
    MSH_PRI_75 = 91, //!< quartic wedge

    // --- incomplete wedges ----------------------------------------------------

    MSH_PRI_15 = 18, //!< edge serendipity quadratic wedge

    // --- pyramids -------------------------------------------------------------

    MSH_PYR_5  = 7,  //!< linear pyramid
    MSH_PYR_14 = 14, //!< quadratic pyramid

    // --- incomplete pyramids -------------------------------------------------- 

    MSH_PYR_13 = 19, //! edge serendipity quadratic pyramid

    // --- general purpose tags -------------------------------------------------

    MSH_UNDEFINED_ELEM = 33 //!< undefined element
 
  };

  // ----------------------------------------------------------------------------
  //  general purpose tags
  // ----------------------------------------------------------------------------
  
  //! maximum number of nodes in any element
  //! \ingroup api_mesh_tags
  #define MSH_MAX_ELEMENT_NODES 125
  
  // ----------------------------------------------------------------------------
  //  getting information about tags
  // ----------------------------------------------------------------------------

  //! get total number of nodes for a given element type
  //! \ingroup api_mesh_tags
  MAD_EXPORT int getNumVerticesForElementTypeMSH(int tag);
  
  //! get dimension for a given element type 
  //! \ingroup api_mesh_tags
  MAD_EXPORT int getDimForElementTypeMSH(int type);
  
  //! get a human readable name for an element type
  //! name: memory must be allocated by the caller
  //! \ingroup api_mesh_tags
  MAD_EXPORT void getElementNameMSH(int type, char * name);
  
  //! get element type 
  //! \ingroup api_mesh_tags
  MAD_EXPORT int getElementTypeMSH(int type);
  
  //! get linear counterpart / basic element type 
  //! \ingroup api_mesh_tags
  MAD_EXPORT int getBasicElementTypeMSH(int type);
  
  //! return human readable name for basic element 
  //! name: memory must be allocated by the caller
  //! \ingroup api_mesh_tags
  MAD_EXPORT void getBasicElementNameMSH(int type, char * name);

  //! return interpolation order for element type
  //! \ingroup api_mesh_tags
  MAD_EXPORT int getMappingOrderMSH(int type);
  
  //! return true if this is an incomplete element 
  //! \ingroup api_mesh_tags
  MAD_EXPORT bool isSerendipityMSH(int type);

  //! return ordered edge nodes in reference element
  //! \ingroup api_mesh_tags
  MAD_EXPORT const int* getMshEdgePoints(int type,int edge,int& idx);
 
  //! return ordered face nodes in reference element 
  //! \ingroup api_mesh_tags
  MAD_EXPORT const int* getMshFacePoints(int type,int face,int& idx);

}
#endif
