// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#include "MAdModel.h"
#include "TransformationInterface.h"

namespace MAd {

  // -----------------------------------------------------------------

  MAdModel::~MAdModel() {

    TransformationMap::iterator tmIter;
    for (tmIter=transformations.begin();tmIter!=transformations.end();++tmIter) {
      delete tmIter->second;
    }
    transformations.clear();
   
    // MAdTransformation stored in periodicites already deleted from transformations above 
    periodicities.clear();
    
    TransformationSet::iterator tsIter;
    for (tsIter=combinedPeriodicities.begin();tsIter!=combinedPeriodicities.end();++tsIter) {
      delete *tsIter;
    }
    combinedPeriodicities.clear();
    
    std::set<Physical*>::iterator it;
    for(it=physicals.begin(); it!=physicals.end(); ++it) delete *it;
    physicals.clear();
  }

  // -----------------------------------------------------------------
  
  Physical * MAdModel::getPhysical(int d, int t)
  {
    std::set<Physical *>::const_iterator it = physicals.begin();
    for (; it != physicals.end(); ++it)
      {
        if ( (*it)->dim() == d && (*it)->tag() == t ) return *it;
      }

    Physical * newp = new Physical(d,t);
    physicals.insert(newp);
    return newp;
  }

  // -----------------------------------------------------------------

  bool MAdModel::addTransformation(MAdGEntity* ge1,
                                   MAdGEntity* ge2,
                                   const MAdTransformation* tfo) {

    if (ge1->dim() != ge2->dim()) return false;

    TransformationMap::iterator fIter = transformations.find(std::make_pair(ge1,ge2));
    TransformationMap::iterator iIter = transformations.find(std::make_pair(ge2,ge1));
    
    if (fIter != transformations.end()) {      
      if (iIter == transformations.end()) throw;
      return fIter->second->compatible(*tfo);
    }
    
    MAdTransformation* forward = tfo->duplicate();
    MAdTransformation* inverse = tfo->inverse();
    
    transformations[std::make_pair(ge1,ge2)] = forward;
    transformations[std::make_pair(ge2,ge1)] = inverse;
    
    periodicities.insert(forward);
    
    return true;
  }
  
  // -----------------------------------------------------------------

  bool MAdModel::addPeriodicity(const MAdTransformation* tfo) {

    TransformationSet::const_iterator pIter = periodicities.find((pTransform) tfo);
    
    if (pIter == periodicities.end()) {
      periodicities.insert(tfo->duplicate());
      periodicities.insert(tfo->inverse());
    }
    return true;
  }

  // -----------------------------------------------------------------
  
  MAdTransformation* MAdModel::getTransformation(MAdGEntity* ge1,
                                                 MAdGEntity* ge2) const {
    
    
    TransformationMap::const_iterator 
      tfoIter = transformations.find(std::make_pair(ge1,ge2));
    if (tfoIter == transformations.end()) return NULL;
    return tfoIter->second;
  }

  // -----------------------------------------------------------------

  int nbCombinations(int n,int i) {
    if (i>n)  return 0;
    int result = 1;
    for (int k=i;k<=n;k++) result *= k;
    for (int k=0;k<=n-i;k++) result /= k;
    return result;
  }

  // -----------------------------------------------------------------
  
  std::set<std::set<int> > allCombinations(int n) {
    
    std::set<std::set<int> > result;
    
    int identity = 0;
    result.insert(std::set<int>(&identity,&identity+1));
    
    for (int i=0;i<n;i++) {
      for (int j=0;j<n;j++) {
        std::set<std::set<int> > tmp = result;
        std::set<std::set<int> >::iterator tIter=tmp.begin();
        for (;tIter!=tmp.end();++tIter) {
          std::set<int> ftmp = *tIter;
          if (ftmp.find(-j-1) == ftmp.end()) {
            ftmp.insert(j+1);
            result.insert(ftmp);            
          }
          std::set<int> itmp = *tIter;
          if (itmp.find(j+1) == itmp.end()) {
            itmp.insert(-j-1);
            result.insert(itmp);
          }
        }
      }
    }

    std::set<int> empty;
    empty.insert(0);
    result.erase(empty);

    return result;
  }
  
  // -----------------------------------------------------------------
  //! \brief foresee all combinations of the periods
  //! \warning does not work for rotations since order is important
  
  bool MAdModel::combinePeriodicities() {

    
    // --- compute all possible combinations in index table
    
    size_t nbPeriods = periodicities.size();  
    std::set<std::set<int> > combinations = allCombinations((int)(nbPeriods));
    if (combinedPeriodicities.size() == combinations.size()) return false;
    
    // --- clear existing entries

    TransformationSet::iterator pIter = combinedPeriodicities.begin();
    for (;pIter!=combinedPeriodicities.end();++pIter) delete *pIter;
    combinedPeriodicities.clear();

    // --- copy periodic transformations
    
    std::vector<MAdTransformation*> forward;
    std::vector<MAdTransformation*> inverse;
    
    
    for (pIter=periodicities.begin();pIter!=periodicities.end();++pIter) {
      forward.push_back((*pIter)->duplicate());
      inverse.push_back((*pIter)->inverse());
    }
    
    combinedPeriodicities.insert(forward.begin(),forward.end());
    combinedPeriodicities.insert(inverse.begin(),inverse.end());
    
    // --- generate combined transformations

    std::set<std::set<int> >::iterator cIter;
    for (cIter=combinations.begin();cIter!=combinations.end();++cIter) {
      MAdTransformation* newTfo = T_createIdentity();
      const std::set<int>& cc = *cIter;
      for (std::set<int>::const_iterator ccIter=cc.begin();
           ccIter!=cc.end();++ccIter) {
        int i=*ccIter;
        if (i != 0) {
          MAdTransformation* tfo = i>0 ? forward[i-1] : inverse[-i-1];
          T_apply(tfo,newTfo);
        }
      }
      
      std::pair<TransformationSet::iterator,bool> check = 
        combinedPeriodicities.insert(newTfo);
      if (!check.second) delete newTfo;
    }
    

    return true;
  }
  
  // -----------------------------------------------------------------

}
