// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

// from Common/
#include "MathUtils.h"
#include "MAdMessage.h"
#include "MAdDefines.h"

#include "TransformationInterface.h"
#include "ModelInterface.h"
#ifdef _HAVE_GMSH_
#include "GmshModel.h"
#else
#include "NullModel.h"
#endif
#include "GM_Iterators.h"
#include "PGList.h"

#include <string.h>

namespace MAd {

  // -------------------------------------------------------------------
  void GM_create(pGModel* model, const char * name)
  {
    if (*model) delete (*model);
    std::string s(name);
#ifdef _HAVE_GMSH_
    *model = new GmshModel(s);
#else
    *model = new NullModel(s);
#endif
  }

  // -------------------------------------------------------------------
  void GM_delete(pGModel model)
  { 
    if (model) { delete model; model=NULL; }
  }
  
  // -------------------------------------------------------------------
  enum GeoFileFormat {
    FORMAT_MSH,
    FORMAT_GEO,
    FORMAT_STEP,
    FORMAT_BREP,
    FORMAT_IGES,
    FORMAT_UNKNOWN
  };

  // -------------------------------------------------------------------
  std::vector<std::string> SplitFileName(std::string fileName)
  {
    // returns [path, baseName, extension]
    size_t idot = fileName.find_last_of('.');
    size_t islash = fileName.find_last_of("/\\");
    std::vector<std::string> s(3);
    if(idot < std::string::npos)
      s[2] = fileName.substr(idot);
    if(islash < std::string::npos)
      s[0] = fileName.substr(0, islash + 1);
    s[1] = fileName.substr(s[0].size(), fileName.size() - s[0].size() - s[2].size());
    return s;
  }

  // -------------------------------------------------------------------
  GeoFileFormat guessFormatFromExtension(const std::string fileName)
  {
    std::string ext = SplitFileName(fileName)[2];
    if     ( !strcmp(ext.c_str(),".msh" ) )  return FORMAT_MSH;
    if     ( !strcmp(ext.c_str(),".geo" ) )  return FORMAT_GEO;
    if     ( !strcmp(ext.c_str(),".stp" ) )  return FORMAT_STEP;
    if     ( !strcmp(ext.c_str(),".step") )  return FORMAT_STEP;
    if     ( !strcmp(ext.c_str(),".brep") )  return FORMAT_BREP;
    if     ( !strcmp(ext.c_str(),".iges") )  return FORMAT_IGES;
    return FORMAT_UNKNOWN;
  }

  // -------------------------------------------------------------------
  int GM_read(pGModel model, const char * fileName)
  {
    GeoFileFormat format = guessFormatFromExtension(fileName);
    if ( format == FORMAT_MSH  ) return GM_readFromMSH (model,fileName);
    if ( format == FORMAT_GEO  ) return GM_readFromGEO (model,fileName);
    if ( format == FORMAT_STEP ) return GM_readFromSTEP(model,fileName);
    if ( format == FORMAT_BREP ) return GM_readFromBREP(model,fileName);
    if ( format == FORMAT_IGES ) return GM_readFromIGES(model,fileName);
    MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                "Unknown geo file format %d",format);
    return 0;
  }

  // -------------------------------------------------------------------
  int GM_readFromMSH(pGModel model, const char * name)
  {
      std::string s(name);
      return (model->readMSH(s));
  }

  // -------------------------------------------------------------------
  int GM_readFromGEO(pGModel model, const char * name)
  {
      std::string s(name);
      return (model->readGEO(s));
  }

  // -------------------------------------------------------------------
  int GM_readFromSTEP(pGModel model, const char * name)
  {
      std::string s(name);
      return (model->readSTEP(s));
  }

  // -------------------------------------------------------------------
  int GM_readFromBREP(pGModel model, const char * name)
  {
      std::string s(name);
      return (model->readBREP(s));
  }

  // -------------------------------------------------------------------
  int GM_readFromIGES(pGModel model, const char * name)
  {
      std::string s(name);
      return (model->readIGES(s));
  }

  // -------------------------------------------------------------------
  bool GM_physical(const pGModel model)
  {
    return model->physical();
  }

  // -------------------------------------------------------------------
  pGEntity GM_entityByTag(pGModel model, int dim, int tag)
  {
    return model->getEntityByTag(dim,tag);
  }
  pGRegion GM_regionByTag(const pGModel model, int tag)
  {
    return model->getRegionByTag(tag);
  }
  pGFace GM_faceByTag(const pGModel model, int tag)
  {
    return model->getFaceByTag(tag);
  }
  pGEdge GM_edgeByTag(const pGModel model, int tag)
  {
    return model->getEdgeByTag(tag);
  }
  pGVertex GM_vertexByTag(const pGModel model, int tag)
  {
    return model->getVertexByTag(tag);
  }

  // -------------------------------------------------------------------
  int GM_numVertices(const pGModel model)
  {
    return model->getNumVertices();
  }
  int GM_numEdges(const pGModel model)
  {
    return model->getNumEdges();
  }
  int GM_numFaces(const pGModel model)
  {
    return model->getNumFaces();
  }
  int GM_numRegions(const pGModel model)
  {
    return model->getNumRegions();
  }

  // -------------------------------------------------------------------
  GRIter GM_regionIter(pGModel model)
  {
    return new GM_RegionIterator(model);
  }
  GFIter GM_faceIter(pGModel model)
  {
    return new GM_FaceIterator(model);
  }
  GEIter GM_edgeIter(pGModel model)
  {
    return new GM_EdgeIterator(model);
  }
  GVIter GM_vertexIter(pGModel model)
  {
    return new GM_VertexIterator(model);
  }
  GTIter GM_transformationIter(pGModel model) 
  {
    return new GM_TransformationIterator(model);
  }
  GPIter GM_periodicityIter(pGModel model,bool combine) 
  {
    return new GM_PeriodicityIterator(model,combine);
  }
  

  pGRegion   GRIter_next(GRIter iter) { return iter->next(); }
  pGFace     GFIter_next(GFIter iter) { return iter->next(); }
  pGEdge     GEIter_next(GEIter iter) { return iter->next(); }
  pGVertex   GVIter_next(GVIter iter) { return iter->next(); }
  pTransform GTIter_next(GTIter iter,pGEntity& ge1,pGEntity& ge2) {
    return iter->next(ge1,ge2);
  }
  pTransform GPIter_next(GPIter iter) { return iter->next();}

  void GRIter_delete(GRIter iter) { delete (iter); }
  void GFIter_delete(GFIter iter) { delete (iter); }
  void GEIter_delete(GEIter iter) { delete (iter); }
  void GVIter_delete(GVIter iter) { delete (iter); }
  void GTIter_delete(GTIter iter) { delete (iter); }
  void GPIter_delete(GPIter iter) { delete (iter); }

  void GRIter_reset(GRIter iter) { iter->reset(); }
  void GFIter_reset(GFIter iter) { iter->reset(); }
  void GEIter_reset(GEIter iter) { iter->reset(); }
  void GVIter_reset(GVIter iter) { iter->reset(); }
  void GTIter_reset(GTIter iter) { iter->reset(); }
  void GPIter_reset(GPIter iter) { iter->reset(); }

  // -------------------------------------------------------------------
  int GEN_tag(const pGEntity ent)
  {
    return ent->tag();
  }

  // -------------------------------------------------------------------
  int GEN_type(const pGEntity ent)
  {
    return ent->dim();
  }

  // -------------------------------------------------------------------
  void GEN_setPhysical(pGEntity ent, int dim, int tag)
  {
    ent->setPhysical(dim,tag);
  }

  // -------------------------------------------------------------------
  int GEN_physTag(const pGEntity ent)
  {
    return ent->pTag();
  }

  // -------------------------------------------------------------------
  int GEN_physDim(const pGEntity ent)
  {
    return ent->pDim();
  }

#ifndef _HAVE_GMSH_
  pPGList GEN_closure(const pGEntity pGE)
  {
      pPGList closure = PGList_new();
      std::set<NullGEntity *> asSet = pGE->closure;
      for (std::set<NullGEntity *>::const_iterator it = asSet.begin(); it != asSet.end(); it++)
          PGList_append(closure, *it);
      return closure;
  }
#endif
  
  // -------------------------------------------------------------------

  pGEdge GEN_convertToGEdge(pGEntity pe) {
#ifdef DEBUG
    return dynamic_cast<pGEdge>(pe);
#else 
    return static_cast<pGEdge>(pe);
#endif
  }
  
  // -------------------------------------------------------------------

  pGFace GEN_convertToGFace(pGEntity pe) {
#ifdef DEBUG
    return dynamic_cast<pGFace>(pe);
#else 
    return static_cast<pGFace>(pe);
#endif
  }  
  
  // -------------------------------------------------------------------
  
  pGRegion GEN_convertToGRegion(pGEntity pe) {
#ifdef DEBUG
    return dynamic_cast<pGRegion>(pe);
#else 
    return static_cast<pGRegion>(pe);
#endif
  }

#ifdef _HAVE_GMSH_
  // -------------------------------------------------------------------
  pPGList GEN_closure(const pGEntity pGE)
  {
      pPGList closure = PGList_new();

      int type = GEN_type(pGE);
      switch (type) {
      case 0: {
          break;
      }
      case 1: {
          std::list<pGVertex> vList = GE_vertices((pGEdge)pGE);
          std::list<pGVertex>::const_iterator vIter = vList.begin();
          for (; vIter != vList.end(); vIter++) {
              PGList_append(closure, *vIter);
          }
          break;
      }
      case 2: {
          std::list<pGEdge> eList = GF_edges((pGFace)pGE);
          std::list<pGEdge>::const_iterator eIter = eList.begin();
          for (; eIter != eList.end(); eIter++) {
              PGList_append(closure, *eIter);
          }
          break;
      }
      case 3: {
          std::list<pGFace> fList = GR_faces((pGRegion)pGE);
          std::list<pGFace>::const_iterator fIter = fList.begin();
          for (; fIter != fList.end(); fIter++) {
              PGList_append(closure, *fIter);
          }
          break;
      }
      }

      return closure;
  }

  // -------------------------------------------------------------------
  pGVertex GEN_convertToGVertex(pGEntity pe) {
#ifdef DEBUG
    return dynamic_cast<pGVertex>(pe);
#else 
    return static_cast<pGVertex>(pe);
#endif
  }

  // -------------------------------------------------------------------
  pPGList GR_faces(const pGRegion pGR)
  {
      pPGList faces = PGList_new();
      std::list<pGFace> list = pGR->faces();
      for (std::list<pGFace>::const_iterator it = list.begin(); it != list.end(); it++)
          PGList_append(faces, *it);
      return faces;
  }

  // -------------------------------------------------------------------
  int GF_numRegions(const pGFace f)
  {
    return f->numRegions();
  }

  // -------------------------------------------------------------------
  pPGList GF_edges(const pGFace pGF)
  {
      pPGList edges = PGList_new();
      std::list<pGEdge> list = pGF->edges();
      for (std::list<pGEdge>::const_iterator it = list.begin(); it != list.end(); it++)
          PGList_append(edges,*it);
      return edges;
  }

  // -------------------------------------------------------------------
  bool GF_getParams(const pGFace pGF, const double xyz[3], 
                    double params[2])
  {
    //   SPoint2 sP = pGF->parFromPoint(SPoint3(xyz));
    //   params[0] = sP.x();
    //   params[1] = sP.y();

    pGF->XYZtoUV( xyz[0], xyz[1], xyz[2],
                  params[0], params[1], 1.);

    //    throw;

    // a test should be done to check that xyz was on the surface

    return true;
  }

  // -------------------------------------------------------------------
  void GF_closestPoint(const pGFace pGF, const double xyz[3],
                       const double initGuess[2], double xyzOnF[3])
  {
    //   GPoint gP = pGF->closestPoint( SPoint3(xyz), initGuess );
    //   xyzOnF[0] = gP.x();
    //   xyzOnF[1] = gP.y();
    //   xyzOnF[2] = gP.z();
    throw;
  }

  // -------------------------------------------------------------------
  void GF_xyz(const pGFace pGF, double u, double v, double xyz[3])
  {
    GPoint gP = pGF->point(u,v);
    xyz[0] = gP.x();
    xyz[1] = gP.y();
    xyz[2] = gP.z();
  }

  // -------------------------------------------------------------------
  double GF_curvatureDiv(const pGFace surface, const double u[2], 
                         double cMaxBound)
  {
    SPoint2 param(u[0],u[1]);
    double curv = surface->curvatureDiv(param);
    if ( isnan(curv) ) {
      MAdMsgSgl::instance().warning(__LINE__,__FILE__,"NaN curvature");
      curv = cMaxBound;
    }
    return std::min(curv,cMaxBound);
  }

  // -------------------------------------------------------------------
 
  double GF_curvatures(const pGFace surface, const double u[2],
                       double dirMax[3], double dirMin[3],
                       double *curvMax, double *curvMin, 
                       double cMaxBound)
  {
    SPoint2 param(u[0],u[1]);
    SVector3 dirMaxTmp = SVector3();
    SVector3 dirMinTmp = SVector3();

    surface->curvatures(param, &dirMaxTmp, &dirMinTmp, curvMax, curvMin);

    dirMax[0] = dirMaxTmp.x();
    dirMax[1] = dirMaxTmp.y();
    dirMax[2] = dirMaxTmp.z();
    if ( ( dotProd(dirMax,dirMax) <= MAdTOL ) ||
         ( isnan(dirMax[0]) || isnan(dirMax[1]) || isnan(dirMax[2]) ) )
      {
        MAdMsgSgl::instance().warning(__LINE__,__FILE__,
                                      "NaN direction for maximum curvature");
        dirMax[0] = 0.36436431;
        dirMax[1] = 0.76356436;
        dirMax[2] = 0.96983673;
      }
    
    dirMin[0] = dirMinTmp.x();
    dirMin[1] = dirMinTmp.y();
    dirMin[2] = dirMinTmp.z();
    if ( ( dotProd(dirMin,dirMin) <= MAdTOL ) ||
         ( isnan(dirMin[0]) || isnan(dirMin[1]) || isnan(dirMin[2]) ) )
      {
        MAdMsgSgl::instance().warning(__LINE__,__FILE__,
                                      "Inconsistent direction for minimum curvature (u,v)=(%f,%f), direction: %f, %f, %f, curvature: %f",
                                      u[0],u[1],dirMin[0],dirMin[1],dirMin[2],*curvMin);
        double tmp[3] = { 0.86684859, 0.69576964, 0.39876864 };
        crossProd(tmp,dirMax,dirMin);
      }
    
    if ( isnan(*curvMax) ) {
      MAdMsgSgl::instance().warning(__LINE__,__FILE__,"NaN maximum curvature");
      *curvMax = cMaxBound;
    }
    *curvMax = std::min(*curvMax,cMaxBound);
    if ( isnan(*curvMin) ) {
      MAdMsgSgl::instance().warning(__LINE__,__FILE__,"NaN minimum curvature");
      *curvMin = cMaxBound;
    }
    *curvMin = std::min(*curvMin,cMaxBound);

    return *curvMax;
  }

  // -------------------------------------------------------------------
  void GF_centerOnGeodesic(const pGFace face, double t, 
                           const double e[2][2], double c[2])
  {
    SPoint2 pt1(e[0][0],e[0][1]);
    SPoint2 pt2(e[1][0],e[1][1]);
    SPoint2 res = face->geodesic(pt1,pt2,t);
    c[0] = res.x();
    c[1] = res.y();
  }

  // -------------------------------------------------------------------
  pPGList GE_vertices(const pGEdge pGE)
  {
      pPGList vertices = PGList_new();
      std::list<pGVertex> list = pGE->vertices();
      for (std::list<pGVertex>::const_iterator it = list.begin(); it != list.end(); it++)
          PGlist_append(vertices,*it);
      return vertices;
  }

  // -------------------------------------------------------------------
  void GE_closestPoint(const pGEdge pGE, const double xyz[3], 
                       double xyzOnE[3])
  {
    //   GPoint gP = pGE->closestPoint( SPoint3(xyz) );
    //   xyzOnE[0] = gP.x();
    //   xyzOnE[1] = gP.y();
    //   xyzOnE[2] = gP.z();
    throw;
  }

  // -------------------------------------------------------------------
  void GE_xyz(const pGEdge pGE, double u, double xyz[3])
  {
    GPoint gP = pGE->point(u);
    xyz[0] = gP.x();
    xyz[1] = gP.y();
    xyz[2] = gP.z();
  }

  // -------------------------------------------------------------------
  void GE_reparamOnFace(const pGEdge edge, const pGFace face, 
                        double uOnE, double uOnF[2], double uClose[2])
  {
    //   assert( edge->isSeam(face) );

    SPoint2 pt = edge->reparamOnFace(face, uOnE, 0);
    uOnF[0] = pt.x();
    uOnF[1] = pt.y();

    if ( uClose )
      {
        double dist = 
          ( uClose[0] - pt.x() ) * ( uClose[0] - pt.x() ) +
          ( uClose[1] - pt.y() ) * ( uClose[1] - pt.y() );

        SPoint2 pt2 = edge->reparamOnFace(face, uOnE, 1);
        double dist2 = 
          ( uClose[0] - pt2.x() ) * ( uClose[0] - pt2.x() ) +
          ( uClose[1] - pt2.y() ) * ( uClose[1] - pt2.y() );

        if ( dist2 < dist ) {
          uOnF[0] = pt2.x();
          uOnF[1] = pt2.y();
        }
      }
  }

  // -------------------------------------------------------------------
  
  bool GE_isSeam(const pGEdge edge, const pGFace face)
  {
    return edge->isSeam(face);
  }

  // -------------------------------------------------------------------
  double GE_curvature(const pGEdge line, double u, double cMaxBound)
  {
    double curv = line->curvature(u);
    if ( isnan(curv) ) {
      MAdMsgSgl::instance().warning(__LINE__,__FILE__,"NaN curvature");
      curv = cMaxBound;
    }
    return std::min(curv,cMaxBound);
  }

  // -------------------------------------------------------------------
 
  pPGList GV_edges(const pGVertex pGV)
  {
      pPGList edges = PGList_new();
      std::list<pGEdge> list = pGV->edges();
      for (std::list<pGEdge>::const_iterator it = list.begin(); it != list.end(); it++)
          PGList_append(edges, *it);
      return edges;
  }

  // -------------------------------------------------------------------
  void GV_reparamOnFace(const pGVertex vertex, const pGFace face, 
                        double uOnF[2], double uClose[2])
  {
    SPoint2 pt = vertex->reparamOnFace(face, 0);
    uOnF[0] = pt.x();
    uOnF[1] = pt.y();

    if ( uClose )
      {
        double dist = 
          ( uClose[0] - pt.x() ) * ( uClose[0] - pt.x() ) +
          ( uClose[1] - pt.y() ) * ( uClose[1] - pt.y() );

        SPoint2 pt2 = vertex->reparamOnFace(face, 1);
        double dist2 = 
          ( uClose[0] - pt2.x() ) * ( uClose[0] - pt2.x() ) +
          ( uClose[1] - pt2.y() ) * ( uClose[1] - pt2.y() );

        if ( dist2 < dist ) {
          uOnF[0] = pt2.x();
          uOnF[1] = pt2.y();
        }
      }
  }

  // -------------------------------------------------------------------
  void GV_reparamOnEdge(const pGVertex vertex, const pGEdge edge, 
                        double * uOnE, double uClose)
  {
    Range<double> range = edge->parBounds(0);

    if ( uClose >= 0. && 
         vertex == edge->getBeginVertex() && 
         vertex == edge->getEndVertex() ) {
      *uOnE = range.low();
      double dist  = ( uClose - *uOnE ) * ( uClose - *uOnE );
      double dist2 = ( uClose - range.high() ) * ( uClose - range.high() );
      if ( dist2 < dist ) *uOnE = range.high();
      return;
    }

    if ( vertex == edge->getBeginVertex() ) { *uOnE = range.low();  return; }
    if ( vertex == edge->getEndVertex() )   { *uOnE = range.high(); return; }

    MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                "GVertex not in the closure of the GEdge");
  }

  // -------------------------------------------------------------------
  
  bool GV_isOnSeam(const pGVertex vert, const pGFace face)
  {
    return vert->isOnSeam(face);
  }

#else

  // -------------------------------------------------------------------

  int GR_numFaces(const pGRegion pr)  {return (int)pr->closure.size();}

  // -------------------------------------------------------------------

  int GF_numRegions(const pGFace pf)  {return (int)pf->inClosure.size();}
  
  // -------------------------------------------------------------------

  int GF_numEdges  (const pGFace pf)  {return (int)pf->closure.size();}

  // -------------------------------------------------------------------

  int GE_numFaces   (const pGEdge pe) {return (int)pe->inClosure.size();}

  // -------------------------------------------------------------------

  int GE_numVertices(const pGEdge pe) {return (int)pe->closure.size();}
  
  // -------------------------------------------------------------------

  int GV_numEdges(const pGVertex pv)  {return (int)pv->inClosure.size();}
  
  // -------------------------------------------------------------------
  
  void GF_centerOnGeodesic(const pGFace face, double t,
                           const double e[2][2], double c[2])
  {
    MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                "Using geodesics requires Gmsh");
  }

#endif

  // -------------------------------------------------------------------

  // -------------------------------------------------------------------
  // PGList functions
  // -------------------------------------------------------------------

  // -------------------------------------------------------------------
  PGList * PGList_new()
  {
    return new PGList();
  }

  // -------------------------------------------------------------------
  void PGList_delete(PGList * lst)
  {
    if (lst) delete lst;
    lst = NULL;
  }

  // -------------------------------------------------------------------
  void PGList_clear(PGList * lst)
  {
    lst->clear();
  }

  // -------------------------------------------------------------------
  PGList * PGList_appUnique(PGList * lst, pGEntity ent)
  {
    for ( unsigned int i=0; i<lst->entities.size(); i++ ) {
      if ( lst->entities[i] == ent ) return lst;
    }
    lst->entities.push_back(ent);
    return lst;
  }

  // -------------------------------------------------------------------
  PGList * PGList_appPGListUnique(PGList * lst, PGList * source)
  {
    for ( unsigned int iSrc=0; iSrc<source->entities.size(); iSrc++ ) {
      PGList_appUnique(lst,source->entities[iSrc]);
    }
    return lst;
  }

  // -------------------------------------------------------------------
  PGList * PGList_append(PGList * lst, pGEntity ent)
  {
    lst->entities.push_back(ent);
    return lst;
  }

  // -------------------------------------------------------------------
  int PGList_size(PGList * lst)
  {
    return (int)(lst->entities.size());
  }

  // -------------------------------------------------------------------
  pGEntity PGList_item(PGList * lst, int i)
  {
    return lst->entities[i];
  }

  // -------------------------------------------------------------------
  pGEntity PGList_next(PGList * lst, void **restart)
  {
    if( *(int*)(restart) >= (int)lst->entities.size() ) return NULL;
    return lst->entities[(*(int*)(restart))++];
  }

  // -------------------------------------------------------------------
  int PGList_inList(PGList * lst, pGEntity ent)
  {
    for ( unsigned int i=0; i<lst->entities.size(); i++ ) {
      if ( lst->entities[i] == ent ) return 1;
    }
    return 0;
  }

  // -------------------------------------------------------------------
  void PGList_remItem(PGList * lst, pGEntity ent)
  {
    std::vector<pGEntity>::iterator eIter = lst->entities.begin();
    for (; eIter != lst->entities.end() ; eIter++) {
      if ( *eIter == ent ) lst->entities.erase(eIter);
    }
  }

  // -------------------------------------------------------------------
  
  /*! \warning add creation / retrieval of transformations */
  
  pTransform GM_transformation(const pGModel model,
                               const pGEntity ge1,
                               const pGEntity ge2) {
    return model->getTransformation(ge1,ge2);
  }

  // -------------------------------------------------------------------

  void GM_addTransformation(pGModel model,
                            const pTransform fTfo,
                            const pGEntity ge1,
                            const pGEntity ge2) {
    model->addTransformation(ge1,ge2,fTfo);
  }
  
  // -------------------------------------------------------------------

  void GM_addPeriodicity(pGModel model,
                            const pTransform fTfo) {
    model->addPeriodicity(fTfo);
  }
  
  // -------------------------------------------------------------------

  int GM_numTransformations(const pGModel model) {
    return model->nbTransformations();
  }
  
  // -------------------------------------------------------------------

  int GM_numPeriodicities(const pGModel model) {
    return model->nbPeriodicities();
  }
  
  // -------------------------------------------------------------------
  
}

