// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#ifndef _H_MODELINTERFACE
#define _H_MODELINTERFACE

#include "MAdExport.h"
#include "TransformationInterface.h"

/*! \defgroup api_model  Model and entity operations and objects 
  \ingroup api */

/*! \defgroup api_model_model     Model operations 
  \ingroup api_model */
/*! \defgroup api_model_entity    Model entity operations 
  \ingroup api_model */
/*! \defgroup api_model_region    Model region operations 
  \ingroup api_model */
/*! \defgroup api_model_face      Model face operations 
  \ingroup api_model */
/*! \defgroup api_model_edge      Model edge operations 
  \ingroup api_model */
/*! \defgroup api_model_vertex    Model vertex operations 
  \ingroup api_model */
/*! \defgroup api_model_transformation  Model periodicities and transformations 
  \ingroup api_model */

namespace MAd {

  // -------------------------------------------------------------------
#ifdef _HAVE_GMSH_
  typedef class GmshGEntity MAdGEntity;
  typedef class GmshGEntityLessThan MAdGEntityLessThan;
  typedef class GmshGRegion MAdGRegion;
  typedef class GmshGFace   MAdGFace;
  typedef class GmshGEdge   MAdGEdge;
  typedef class GmshGVertex MAdGVertex;
#else
  typedef class NullGEntity MAdGEntity;
  typedef class NullGEntityLessThan MAdGEntityLessThan;
  typedef class NullGRegion MAdGRegion;
  typedef class NullGFace   MAdGFace;
  typedef class NullGEdge   MAdGEdge;
  typedef class NullGVertex MAdGVertex;
#endif
  
  typedef MAdGEntity * pGEntity;
  typedef MAdGRegion * pGRegion;
  typedef MAdGFace   * pGFace;
  typedef MAdGEdge   * pGEdge;
  typedef MAdGVertex * pGVertex;

  typedef class MAdModel * pGModel;

  typedef class GM_RegionIterator         * GRIter;
  typedef class GM_FaceIterator           * GFIter;
  typedef class GM_EdgeIterator           * GEIter;
  typedef class GM_VertexIterator         * GVIter;
  typedef class GM_TransformationIterator * GTIter;
  typedef class GM_PeriodicityIterator    * GPIter;

  typedef class PGList * pPGList;

  // -------------------------------------------------------------------

  // --- Model operators ---

  //! \brief create an empty model
  //! \ingroup api_model_model
  MAD_EXPORT void GM_create(pGModel * model, const char * name="");

  //! \brief delete the model
  //! \ingroup api_model_model
  MAD_EXPORT void GM_delete(pGModel model);

  //! \brief read a model, guess file format from extension
  //! \ingroup api_model_model
  MAD_EXPORT int  GM_read(pGModel model, const char * name);

  // read particular file formats
  MAD_EXPORT int  GM_readFromMSH(pGModel model, const char * name);
  MAD_EXPORT int  GM_readFromGEO(pGModel model, const char * name);
  MAD_EXPORT int  GM_readFromSTEP(pGModel model, const char * name);
  MAD_EXPORT int  GM_readFromBREP(pGModel model, const char * name);
  MAD_EXPORT int  GM_readFromIGES(pGModel model, const char * name);

  // void GM_diagnostics(const pGModel model);
  
  //! \brief returns true if physical tags are used in the model 
  //! \ingroup api_model_model
  MAD_EXPORT bool GM_physical(const pGModel model);
  
  //! \brief find or create the model entity by tag 
  //! \ingroup api_model_model
  MAD_EXPORT pGEntity GM_entityByTag(const pGModel model, int type, int tag);
  
  //! \brief find or create the model region by tag 
  //! \ingroup api_model_model
  MAD_EXPORT pGRegion GM_regionByTag(const pGModel model, int tag);
  
  //! \brief find or create the model face by tag 
  //! \ingroup api_model_model
  MAD_EXPORT pGFace   GM_faceByTag  (const pGModel model, int tag);
  
  //! \brief find or create the model edge by tag 
  //! \ingroup api_model_model
  MAD_EXPORT pGEdge   GM_edgeByTag  (const pGModel model, int tag);
  
  //! \brief find or create the model vertex by tag 
  //! \ingroup api_model_model
  MAD_EXPORT pGVertex GM_vertexByTag(const pGModel model, int tag);

  //! \brief give the number of model vertices
  //! \ingroup api_model_model
  MAD_EXPORT int GM_numVertices(const pGModel);

  //! \brief give the number of model edges
  //! \ingroup api_model_model
  MAD_EXPORT int GM_numEdges(const pGModel);
  
  //! \brief give the number of model faces
  //! \ingroup api_model_model
  MAD_EXPORT int GM_numFaces(const pGModel);
  
  //! \brief give the number of model regions
  //! \ingroup api_model_model
  MAD_EXPORT int GM_numRegions(const pGModel);
  
  // --- periodicities 

  //! \brief give the number of pairs of periodic entities
  //! \ingroup api_model_periodic
  MAD_EXPORT int GM_numTransformations(const pGModel);
  
  //! \brief give the number of fundamental periodicities
  //! \ingroup api_model_periodic
  MAD_EXPORT int GM_numPeriodicities(const pGModel);
  
  /*! \brief find the transformation that transforms e1 into e2 */
  //! \ingroup api_model_periodic
  MAD_EXPORT pTransform GM_transformation(const pGModel model,
                                          const pGEntity ge1,
                                          const pGEntity ge2);
  
  /*! \brief add a transformation that transforms e1 into e2 */
  //! \ingroup api_model_periodic
  MAD_EXPORT void GM_addTransformation(pGModel model,
                                       const pTransform tfo,
                                       const pGEntity ge1,
                                       const pGEntity ge2);
  
  /*! \brief add a fundamental periodicity */
  //! \ingroup api_model_periodic
  MAD_EXPORT void GM_addPeriodicity(pGModel,const pTransform tfo);
  
  // --- Iterators ---

  /*! \brief iterator over the regions in the model */
  //! \ingroup api_model_region
  MAD_EXPORT GRIter GM_regionIter(pGModel);
  
  /*! \brief iterator over the faces in the model */
  //! \ingroup api_model_face
  MAD_EXPORT GFIter GM_faceIter(pGModel);
  
  /*! \brief iterator over the edges in the model */
  //! \ingroup api_model_edge
  MAD_EXPORT GEIter GM_edgeIter(pGModel);
  
  /*! \brief iterator over the vertices in the model */
  //! \ingroup api_model_vertex
  MAD_EXPORT GVIter GM_vertexIter(pGModel);
  
  /*! \brief iterator over the transformations between pairs of geometric entities */
  //! \ingroup api_model_periodic
  MAD_EXPORT GTIter GM_transformationIter(pGModel);
  
  /*! \brief iterator over the list of periodic transformations */
  /*! if combine is true all combinations are provided, otherwise a minimal set */
  //! \ingroup api_model_periodic
  MAD_EXPORT GPIter GM_periodicityIter(pGModel,bool combine);
  
  /*! \brief increment iterator over regions in the model */
  //! \ingroup api_model_region
  MAD_EXPORT pGRegion   GRIter_next(GRIter);
  
  /*! \brief increment iterator over faces in the model */
  //! \ingroup api_model_face
  MAD_EXPORT pGFace     GFIter_next(GFIter);
  
  /*! \brief increment iterator over edges in the model */
  //! \ingroup api_model_edge
  MAD_EXPORT pGEdge     GEIter_next(GEIter);
  
  /*! \brief increment iterator over vertices in the model */
  //! \ingroup api_model_vertex
  MAD_EXPORT pGVertex   GVIter_next(GVIter);

  /*! \brief increment iterator over periodic connections in the model */
  /*! provides connected entities, transformation from left to right */
  //! \ingroup api_model_periodic
  MAD_EXPORT pTransform GTIter_next(GTIter it,    //!< iterator 
                                    pGEntity& e1, //!< first connected entity
                                    pGEntity& e2  //!< second connected entity   
                                    );
  
  /*! \brief increment iterator over model periodicities */
  //! \ingroup api_model_periodic
  MAD_EXPORT pTransform GPIter_next(GPIter);
  
  /*! \brief delete iterator over regions in the model */
  //! \ingroup api_model_region
  MAD_EXPORT void GRIter_delete(GRIter);
  
  /*! \brief delete iterator over faces in the model */
  //! \ingroup api_model_face
  MAD_EXPORT void GFIter_delete(GFIter);

  /*! \brief delete iterator over edges in the model */
  //! \ingroup api_model_edge
  MAD_EXPORT void GEIter_delete(GEIter);
  
  /*! \brief delete iterator over vertices in the model */
  //! \ingroup api_model_vertex
  MAD_EXPORT void GVIter_delete(GVIter);

  /*! \brief delete iterator over periodic connections in the model */
  //! \ingroup api_model_periodic
  MAD_EXPORT void GTIter_delete(GTIter);

  /*! \brief delete iterator over periodicities in the model */
  //! \ingroup api_model_periodic
  MAD_EXPORT void GPIter_delete(GPIter);

  /*! \brief reset iterator over regions in the model */
  //! \ingroup api_model_region
  MAD_EXPORT void GRIter_reset(GRIter);
  
  /*! \brief reset iterator over faces in the model */
  //! \ingroup api_model_face
  MAD_EXPORT void GFIter_reset(GFIter);
  
  /*! \brief reset iterator over edges in the model */
  //! \ingroup api_model_edge
  MAD_EXPORT void GEIter_reset(GEIter);
  
  /*! \brief reset iterator over vertices in the model */
  //! \ingroup api_model_vertex
  MAD_EXPORT void GVIter_reset(GVIter);
  
  /*! \brief reset iterator over periodic connections in the model */
  //! \ingroup api_model_periodic
  MAD_EXPORT void GTIter_reset(GTIter);
  
  /*! \brief reset iterator over periodicities in the model */
  //! \ingroup api_model_periodic
  MAD_EXPORT void GPIter_reset(GPIter);

  // --- Entity operators ---
  
  //! \brief get model entity identifier
  //! \ingroup api_model_entity
  MAD_EXPORT int  GEN_tag(const pGEntity);  //!< model entity tag

  //! \brief get model entity type / dimension
  //! \ingroup api_model_entity
  MAD_EXPORT int  GEN_type(const pGEntity); //!< model entity dimension

  //! \brief set model entity physical classification
  //! \ingroup api_model_entity
  MAD_EXPORT void GEN_setPhysical(pGEntity, int dim, int tag);

  //! \brief get model entity physical classification
  //! \ingroup api_model_entity
  MAD_EXPORT int  GEN_physTag(const pGEntity);

  //! \brief get dimension of model entity physical classification
  //! \ingroup api_model_entity
  MAD_EXPORT int  GEN_physDim(const pGEntity);
  
  //! \brief return a list of all entities in the closure of the specified entity
  //! \ingroup api_model_entity
  // get the closure of an entity - Allocates memory! 
  // returns: the closure entities. The object has to be freed by the caller using PGList_delete
  MAD_EXPORT pPGList GEN_closure(const pGEntity);

  //! \brief convert safely to vertex, returns NULL if not successful 
  //! \ingroup api_model_entity
  MAD_EXPORT pGVertex  GEN_convertToGVertex(pGEntity);
  
  //! \brief convert safely to edge, returns NULL if not successful 
  //! \ingroup api_model_entity
  MAD_EXPORT pGEdge    GEN_convertToGEdge(pGEntity);
  
  //! \brief convert safely to face, returns NULL if not successful 
  //! \ingroup api_model_entity
  MAD_EXPORT pGFace    GEN_convertToGFace(pGEntity);
  
  //! \brief convert safely to region, returns NULL if not successful
  //! \ingroup api_model_entity
  MAD_EXPORT pGRegion  GEN_convertToGRegion(pGEntity);


#ifdef _HAVE_GMSH_

  // --- Region operators ---

  //! \brief get closure faces of a model region
  //! \ingroup api_model_region
  MAD_EXPORT pPGList GR_faces(const pGRegion); // Allocates memory!

  // --- Face operators ---
  
  //! \brief get number of model regions adjacent to face
  //! \ingroup api_model_face
  MAD_EXPORT int GF_numRegions(const pGFace);

  //! \brief get closure edges of a model face
  //! \ingroup api_model_face
  MAD_EXPORT pPGList GF_edges(const pGFace); // Allocates memory!
  
  //! \brief computes the parameter location of xyz in the surface.
  //! Returns false if xyz is not in the surface.  
  //! \ingroup api_model_face
  MAD_EXPORT bool GF_getParams(const pGFace gf,
                               const double xyz[3],
                               double uv[2]);
  
  //! \brief computes the coordinates of the point on the surface closest to xyz
  //! \ingroup api_model_face
  //! \warning currently not implemented
  MAD_EXPORT void GF_closestPoint(const pGFace gf, 
                                  const double xyz[3],
                                  const double uv[2], 
                                  double xyzC[3]);
  
  //! \brief evaluate Cartesian coordinates xyz for parametric location u,v
  //! \ingroup api_model_face
  MAD_EXPORT void GF_xyz(const pGFace gf,double u,double v,double xyz[3]);
  
  //! \brief gets the curvature of the surface as the divergence of its normal
  //! the result is bounded by cMaxBound. NaN curvatures are turned into cMaxBound.
  //! \ingroup api_model_face
  MAD_EXPORT double GF_curvatureDiv(const pGFace gf,    //!< face
                                    const double uv[2], //!< position
                                    double cMaxBound    //!< clipping value
                                    );
  
  //! \brief get surface curvatures at position  
  //! Compute the min and max curvatures and the corresponding directions.
  //! Returns the maximum curvature.
  //! Min and max curvatures are bounded by cMaxBound. 
  //! NaN curvatures are turned into cMaxBound.
  //! \ingroup api_model_face
  MAD_EXPORT double GF_curvatures(const pGFace gf,  //!< face
                                  const double uv[2], //!< parametric position
                                  double dirMax[3],//!< direction of maximum curvature
                                  double dirMin[3],//!< direction of minimum curvature
                                  double *curvMax, //!< maximum curvature
                                  double *curvMin, //!< minimum curvature
                                  double cMaxBound //!< saturation
                                  );
  
  //! \brief computes the parametric coordinates of the point on a geodesic  
  //! \ingroup api_model_face
  MAD_EXPORT void GF_centerOnGeodesic(const pGFace face, //!< surface
                                      double t, //!< line position (0<=t<=1)
                                      const double e[2][2], //!< end points
                                      double r[2] //!< parametric coordinates of result
                                      );

  // --- Edge operators ---

  //! \brief returns the number of faces adjacent to the edge   
  //! \ingroup api_model_edge
  MAD_EXPORT int GE_numFaces(const pGEdge);

  //! \brief returns the number of vertices in the closuer of the edge
  //! \ingroup api_model_edge
  MAD_EXPORT pPGList GE_vertices(const pGEdge); // Allocates memory!

  //! \brief returns the Cartesian coordinate of the closest point on the edge
  //! \ingroup api_model_edge
  //! \warning currently not implemented
  MAD_EXPORT void GE_closestPoint(const pGEdge, 
                                  const double[3] xyz, 
                                  double[3]);

  //! \brief computes the Cartesian coordinates from parametric coordinate
  //! \ingroup api_model_edge
  MAD_EXPORT void GE_xyz(const pGEdge, double, double[3]);

  //! \brief reparametrize a point specified on an edge on one of the adjacent faces
  //! Given an edge which is a seam of the face, and a point of 
  //! the edge with parametric coordinate uOnE on the edge, 
  //! find the parametric coordinates of the point on the face.
  //! uClose are the parametric coordinates of a close point of 
  //! face used to determine the direction for the parametrization.
  //! \ingroup api_model_edge
  MAD_EXPORT void GE_reparamOnFace(const pGEdge, const pGFace,
                                   double, double[2], double uClose[2]=NULL);

  //! \brief returns true if the edge is a seam for the given face.
  //! \ingroup api_model_edge
  MAD_EXPORT bool GE_isSeam(const pGEdge, const pGFace);

  //! \brief gets the curvature of the line at that point bounded by cMaxBound
  //! \ingroup api_model_edge
  MAD_EXPORT double GE_curvature(const pGEdge, double, double);

  // --- Vertex operators ---
  
  //! \brief returns the number of edges including the vertex
  //! \ingroup api_model_vertex
  MAD_EXPORT int GV_numEdges(const pGVertex);

  //! \brief returns a list of the lines including the vertex
  //! \ingroup api_model_vertex
  MAD_EXPORT pPGList GV_edges(const pGVertex);

  //! \brief find parametric coordinates of vertex on adjacent face
  //! Given a geometric vertex which is on the face, find the parametric 
  //! coordinates of the vertex on the face.
  //! uClose are the parametric coordinates of a close point of 
  //! face used to determine the direction for the parametrization.
  //! \ingroup api_model_vertex
  MAD_EXPORT void GV_reparamOnFace(const pGVertex, const pGFace, double [2],
                                   double uClose[2]=NULL);
  
  //! \brief find parametric coordinates of vertex on adjacent edge
  //! Given a geometric vertex which is on the edge, find the parametric 
  //! coordinates of the vertex on the edge.
  //! uClose is the parametric coordinate of a close point of 
  //! edge used to determine the direction for the parametrization.
  //! \ingroup api_model_vertex
  MAD_EXPORT void GV_reparamOnEdge(const pGVertex, const pGEdge, double *,
                                   double uClose=-1.);

  //! \brief Return true if the vertex is on a seam of the given face
  //! \ingroup model_api_vertex
  MAD_EXPORT bool GV_isOnSeam(const pGVertex, const pGFace);
#else
  
  MAD_EXPORT int GR_numFaces(const pGRegion);

  MAD_EXPORT int GF_numRegions(const pGFace);
  MAD_EXPORT int GF_numEdges(const pGFace);
  
  MAD_EXPORT int GE_numFaces(const pGEdge);
  MAD_EXPORT int GE_numVertices(const pGEdge);
  
  MAD_EXPORT int GV_numEdges(const pGVertex);
  
  MAD_EXPORT void GF_centerOnGeodesic(const pGFace face, double t,
                                      const double e[2][2], double c[2]);
#endif

  // --- pGList operators ---

  MAD_EXPORT pPGList  PGList_new();
  MAD_EXPORT void     PGList_delete          (pPGList);
  MAD_EXPORT void     PGList_deallocate      (pPGList);
  MAD_EXPORT void     PGList_clear           (pPGList);
  MAD_EXPORT pPGList  PGList_appPGListUnique (pPGList, pPGList source);
  MAD_EXPORT pPGList  PGList_appUnique       (pPGList, pGEntity);
  MAD_EXPORT pPGList  PGList_append          (pPGList, pGEntity);
  MAD_EXPORT int      PGList_size            (pPGList);
  MAD_EXPORT pGEntity PGList_item            (pPGList, int n);
  MAD_EXPORT pGEntity PGList_next            (pPGList, void ** restart);
  MAD_EXPORT int      PGList_inList          (pPGList, pGEntity);
  MAD_EXPORT void     PGList_remItem         (pPGList, pGEntity);

  // -------------------------------------------------------------------

}

#endif
