#ifndef _FEM_H
#define _FEM_H

#include <vector>
#include <stdexcept>

static int polyExp2d[6][2] = {{0,0},
			      {1,0},
			      {0,1},
			      {2,0},
			      {0,2},
			      {1,1}};


static double coeff_tri_1[3][3] = {{1, -1, -1}, 
				  {0, 1, 0}, 
				  {0, 0, 1}};

static double coeff_tri_2[6][6] = {{1, -3, -3, 2, 2, 4}, 
				   {0, -1, 0, 2, 0, 0}, 
				   {0, 0, -1, 0, 2, 0}, 
				   {0, 4, 0, -4, 0, -4}, 
				   {0, 0, 0, 0, 0, 4}, 
				   {0, 0, 4, 0, -4, -4}};

static int polyExp3d[10][3] = {{0,0,0},
			       {1,0,0},
			       {0,1,0},
			       {0,0,1},
			       {2,0,0},
			       {0,2,0},
			       {0,0,2},
			       {1,1,0},
			       {0,1,1},
			       {1,0,1}};

static double coeff_tet_1[4][4]   = {{1,-1,-1,-1},
				     {0,1,0,0},
				     {0,0,1,0},
				     {0,0,0,1}};

static double coeff_tet_2[10][10] = {{1,-3,-3,-3,2,2,2,4,4,4},
				     {0,-1,0,0,2,0,0,0,0,0},
				     {0,0,-1,0,0,2,0,0,0,0},
				     {0,0,0,-1,0,0,2,0,0,0},
				     {0,4,0,0,-4,0,0,-4,0,-4},
				     {0,0,0,0,0,0,0,4,0,0},
				     {0,0,4,0,0,-4,0,-4,-4,0},
				     {0,0,0,4,0,0,-4,0,-4,-4},
				     {0,0,0,0,0,0,0,0,0,4},
				     {0,0,0,0,0,0,0,0,4,0}};

class TriFEM {
 public:
  TriFEM(int order){
    _n = 0;
    if (order == 1) {
      _n = 3;
      for(int i = 0; i < _n; i++){
	std::vector<double> tmp(_n);
	for(int j = 0; j < _n; j++)
	  tmp[j] = coeff_tri_1[i][j];
	_coeffs.push_back(tmp);
      }
    } else if (order == 2) {
      _n = 6;
       for(int i = 0; i < _n; i++){
	 std::vector<double> tmp(_n);
	for(int j = 0; j < _n; j++)
	  tmp[j] = coeff_tri_2[i][j];
	_coeffs.push_back(tmp);
       }
    } else {
      throw std::runtime_error("invalid order");
    }
  }
  inline double eval(int i, double u, double v) const  {
    if (i >= _n)
      throw std::runtime_error("eval: i > n");
    return eval_poly(_coeffs[i],u,v);
  }

  inline double deval(int i, double u, double v, int k) const {
    if (i >= _n)
      throw std::runtime_error("eval: i > n");
    return deval_poly(_coeffs[i],u,v,k);
  }

  inline int n(){return _n;}

  ~TriFEM(){}
  
private:
  inline double eval_poly(const std::vector<double> &c, double u, double v) const {

    double res = 0;
    int eu,ev;
    for(size_t i = 0; i < c.size(); i++){
      eu = polyExp2d[i][0];
      ev = polyExp2d[i][1];
      res += c[i] * pow(u,eu) * pow(v,ev);
    }
    
    return res;
  }
  inline double deval_poly(const std::vector<double> &c, double u, double v, int k) const {

    double res = 0;
    int eu,ev;
    if (k == 0) {
      for(size_t i = 0; i < c.size(); i++){
	eu = polyExp2d[i][0];
	ev = polyExp2d[i][1];
	res += c[i] * eu * pow(u,max(eu-1,0)) * pow(v,ev);
      }
    } else if (k == 1) {
      for(size_t i = 0; i < c.size(); i++){
	eu = polyExp2d[i][0];
	ev = polyExp2d[i][1];
	res += c[i] * ev * pow(u,eu)  * pow(v,max(ev-1,0));
      }
    } else {
      throw std::runtime_error("deval: k != 0,1");
    }
    
    return res;
  }
  int _n;
  std::vector<std::vector<double> > _coeffs;
};

class TetFEM {
 public:
  TetFEM(int order){
    _n = 0;
    if (order == 1) {
      _n = 4;
      for(int i = 0; i < _n; i++){
	std::vector<double> tmp(_n);
	for(int j = 0; j < _n; j++)
	  tmp[j] = coeff_tet_1[i][j];
	_coeffs.push_back(tmp);
      }
    } else if (order == 2) {
      _n = 10;
       for(int i = 0; i < _n; i++){
	 std::vector<double> tmp(_n);
	for(int j = 0; j < _n; j++)
	  tmp[j] = coeff_tet_2[i][j];
	_coeffs.push_back(tmp);
       }
    } else {
      throw std::runtime_error("invalid order");
    }
  }
  inline double eval(int i, double u, double v, double w) const  {
    if (i >= _n)
      throw std::runtime_error("eval: i > n");
    return eval_poly(_coeffs[i],u,v,w);
  }

  inline double deval(int i, double u, double v, double w, int k) const {
    if (i >= _n)
      throw std::runtime_error("eval: i > n");
    return deval_poly(_coeffs[i],u,v,w,k);
  }

  inline int n(){return _n;}

  void test() {

    double pts[10][3] = {{0.,0.,0.},
			 {1.,0.,0.},
			 {0.,1.,0.},
			 {0.,0.,1.},
			 {.5,0.,0.},
			 {.5,.5,0.},
			 {0.,.5,0.},
			 {0.,0.,.5},
			 {.5,0.,.5},
			 {0.,.5,.5}};

    for(int i = 0; i < _n; i++) {
      for(int j = 0; j < _n; j++) {
	const double val = eval(i,pts[j][0],pts[j][1],pts[j][2]);
	cout << val << " " ;
      }
      cout << endl;
    }
			 
			
  }

  ~TetFEM(){}
  
private:
  inline double eval_poly(const std::vector<double> &c, double u, double v, double w) const {

    double res = 0;
    int eu,ev,ew;
    for(size_t i = 0; i < c.size(); i++){
      eu = polyExp3d[i][0];
      ev = polyExp3d[i][1];
      ew = polyExp3d[i][2];
      res += c[i] * pow(u,eu) * pow(v,ev) * pow(w,ew);
    }
    
    return res;
  }

  inline double deval_poly(const std::vector<double> &c, double u, double v, double w, int k) const {

    double res = 0;
    int eu,ev,ew;
    if (k == 0) {
      for(size_t i = 0; i < c.size(); i++){
	eu = polyExp3d[i][0];
	ev = polyExp3d[i][1];
	ew = polyExp3d[i][2];
	res += c[i] * eu * pow(u,max(eu-1,0)) * pow(v,ev) * pow(w,ew);
      }
    } else if (k == 1) {
      for(size_t i = 0; i < c.size(); i++){
	eu = polyExp3d[i][0];
	ev = polyExp3d[i][1];
	ew = polyExp3d[i][2];
	res += c[i] * ev * pow(u,eu) * pow(v,max(ev-1,0)) * pow(w,ew);
      }
    } else if (k == 2) {
      for(size_t i = 0; i < c.size(); i++){
	eu = polyExp3d[i][0];
	ev = polyExp3d[i][1];
	ew = polyExp3d[i][2];
	res += c[i] * ew * pow(u,eu) * pow(v,ev) * pow(w,max(ew-1,0));
      }
    } else {
      throw std::runtime_error("deval: k != 0,1,2");
    }
    return res;
  }

  int _n;
  std::vector<std::vector<double> > _coeffs;
};

#endif
