#include <iostream>
#include <fstream>
#include <vector>
#include <algorithm>

#include "MAdLib/MAdLib.h"
// #include "MAdLib/MappingInterface.h"
// #include "MAdLib/MeshDataBaseIO.h"
#include "gmshIO.h"

#include "entityIds.h"
#include "normals.h"
#include "deform.h"
#include "layers.h"

#include <boost/program_options.hpp>

namespace po = boost::program_options;
using namespace std;
using namespace MAd;

// Test if a sparse matrix is symmetric.
double is_symmetric(SparseMatrix &A){

  SparseMatrix B = A.transpose();
  B = B - A;

  double err = 0.;
  int n = 0;
  for(int irow = 0; irow < B.outerSize(); irow++)
    for (SparseMatrix::InnerIterator it(B,irow); it; ++it){
      n++;
      err += abs(it.value());
    }
  return err / n;

}

int main(int argc, char* argv[]) {
  cout << "starting..." << endl;

  po::options_description desc("Options");
  desc.add_options()
    ("help", "produce help message")
    ("fi", po::value<string>(), "nom du maillage a traiter")
    ("np", po::value<int>(), "nombre de couches de prismes a inserer")
    ("nf", po::value<vector<int> >(), "numéro des frontieres")
    ("fo", po::value<string>(), "nom du maillage avec prismes")
    ;

  po::variables_map vm;        
  po::store(po::parse_command_line(argc, argv, desc), vm);
  po::notify(vm);    

  if (vm.count("help")) {
    cout << desc << "\n";
    return 0;
  }

  string ifname,fo;
  int np;
  vector<int> tags;

if (vm.count("fi"))
  ifname=vm["fi"].as<string>();
 else{
    cout << desc << "\n";
    return 0;
  } 
if (vm.count("np"))
  np=vm["np"].as<int>();
 else{
    cout << desc << "\n";
    return 0;
  }   
if (vm.count("fo"))
  fo=vm["fo"].as<string>();
 else{
    cout << desc << "\n";
    return 0;
  } 
if (vm.count("nf"))
  tags=vm["nf"].as<vector<int> >();
 else{
    cout << desc << "\n";
    return 0;
  } 
 
    
  MAdLibInitialize(&argc,&argv);
  pGModel model = 0;
  const string modelName = "theModel";
  GM_create(&model,modelName);
  pMesh   mesh = M_new(model);
  cout << "loading " << ifname << " ... " << endl;
  M_load(mesh,ifname.c_str());

  cout << "# or vertices : " << M_numVertices(mesh) << endl;
  cout << "# or regions  : " << M_numRegions(mesh) << endl;
  cout << "# or faces    : " << M_numFaces(mesh) << endl;
  cout << "# or edges    : " << M_numEdges(mesh) << endl;
  cout << "dim           : " << M_dim(mesh) << endl;
  vector<int> orders;
  bool isSimplex = getMeshOrders(mesh,M_dim(mesh),orders);
  cout << "simplex       : " << isSimplex << endl;
  cout << "orders        : (" ;
  for(size_t i = 0; i < orders.size(); i++)
    cout << orders[i] << " ";
  cout << ")" << endl;

  map<int,vector<pEntity> > bMap;
  getBoundaries(mesh,bMap);
  cout << "boundaries" << endl;
  for( map<int,vector<pEntity> >::iterator it = bMap.begin();
       it != bMap.end();
       it++) {
    cout << "  id = " << it->first << ", # of elements: " << it->second.size() << endl;
  }
    
  if (orders.size() != 1) {
    cout << "the mesh contains different orders, aborting" << endl;
    return -1;
  }
  const int dim   = M_dim(mesh);
  const int order = orders[0];
  const int nDofs = M_numVertices(mesh) + M_numPoints(mesh);

  cout << "FEM: # of DOFs = " << nDofs << endl;

  const bool periodic = isPeriodic(mesh);
  cout << "periodic? " << periodic << endl;
  map<pTransform,vector<pair<int,int> > > pairs;

  setIDs(mesh);

  if (periodic){
    double err = periodicityError(mesh);
    getPeriodicIDPairs(mesh, pairs);
    cout << "periodicity error: " << err << endl;;
  }

 
    vector<int> isBc;
    
    computeNodeNormals(mesh,tags);

    SparseMatrix P;
    if (periodic)
      getPeriodicityMat(mesh, pairs, P);

    Eigen::VectorXd rhs(dim * nDofs), sol;
    getRHS(mesh, tags, isBc, rhs);
    rhs *= .5;
    cout << "Max boundary node displacement: " << rhs.maxCoeff()  << endl;
    int nsteps = 1;
    for(int istep = 0; istep < nsteps; istep++) {
      Eigen::VectorXd rhs_step = rhs / nsteps;

      SparseMatrix mat(dim * nDofs,dim * nDofs);
      getMat(mesh,order, isBc,mat, rhs_step);

      if (periodic) {
	rhs_step = P.transpose() * rhs_step;
	mat = P.transpose() * mat * P;
      }
      bool is_mat_symm = is_symmetric(mat) < 1e-10;
      cout << "  symm matrix ? " << is_mat_symm << endl;
      const double tol = 1e-12;
      const int maxIt = 500;
      // Eigen::BiCGSTAB<SparseMatrix> solver;
      Eigen::ConjugateGradient<SparseMatrix> solver;
      solver.setTolerance(tol);
      solver.setMaxIterations(maxIt);
      solver.compute(mat);
      sol = solver.solve(rhs_step);
      cout << "  # of iterations: " << solver.iterations() << endl;
      cout << "  estimated error: " << solver.error()      << endl;

      if (periodic)
	sol = P * sol;

      cout << "Max node displacement: " << sol.maxCoeff()  << endl;
      deformMesh(mesh,sol);

    }
    sol *= nsteps;
    // sol = .1*rhs;

    addLayerElements(mesh, tags, isBc, sol, np);

    SaveGmshMeshHighOrderPerso(mesh,fo.c_str());

    MeshStatus status = VALID;
    if ( !checkMesh(mesh, CHECK_ALL, 1, std::cout, &status) ) {
      cout << "the mesh is invalid, status = " << status << endl;
      return -1;
    }else {
      cout << "the mesh is valid" << endl;
    }
    // M_writeMsh(mesh,"res.msh");

    if (periodic) {
      double err = periodicityError(mesh);
      cout << "periodicity error: " << err  << endl;;
    }

 

  M_delete(mesh);
  GM_delete(model);
  
  MAdLibFinalize();

  cout << "done" << endl;



}
