#include "tools.h"

using namespace std;
using namespace MAd;

bool isPeriodic(const pMesh mesh){

  pGModel model = M_model(mesh);

  pMeshDataId tagPeriodic = MD_lookupMeshDataId("PeriodicVertexID");

  pVertex v;
  VIter vIt= M_vertexIter(mesh);
  while ( ( v = VIter_next(vIt) ) ) {
    int vId = V_id(v);
    void* tmpptr = NULL;
    int isPeriodic = EN_getDataPtr((pEntity) v,tagPeriodic,&tmpptr);
    if (isPeriodic)
      return true;
  }
  return false;

}


pVertex getPeriodicVertex(const pMesh mesh, const pTransform transfo, const pVertex v){

  const pGModel model = M_model(mesh);
  const pMeshDataId tagPeriodic = MD_lookupMeshDataId("PeriodicVertexID");
  const bool periodic = isPeriodic(mesh);

  const pGEntity pGE = V_whatIn(v);
  double xyzFrom[3], xyzTo[3];
  void* tmpptr = NULL;
  int isPeriodic = EN_getDataPtr((pEntity) v,tagPeriodic,&tmpptr);
  if (isPeriodic){
    map<int,vector<int> > * conn = (map<int,vector<int> > *) tmpptr;
    bool isFirst = true;
    for (map<int,vector<int> >::iterator citer = conn->begin(); citer != conn->end(); ++citer) {
	pVertex v1 = M_vertex(mesh,citer->first);
	V_coord(v,xyzFrom);
	V_coord(v1,xyzTo);
	T_transformCoordinates(transfo,xyzFrom);	
	double locErr = 0.;
	for(int i = 0; i < 3; i++)
	  locErr += (xyzFrom[i] - xyzTo[i])*(xyzFrom[i] - xyzTo[i]);
	if (locErr < 1e-16) return v1;
	// pGEntity pGE1 = V_whatIn(v1);
	// // pTransform transfo = GM_transformation(model,pGE1,pGE);
	// if (transfo == GM_transformation(model,pGE,pGE1))
	//   return v1;
    }
  }
  return NULL;
}

double periodicityError(const pMesh mesh) {

  const int dim    = M_dim(mesh);
  const pGModel model = M_model(mesh);

  double xyzFrom[3], xyzTo[3];

  double err = 0.;
  int n = 0;
  
  pTransform t;
  pGEntity pGEfrom, pGEto;
  GTIter tIt=GM_transformationIter(model);
  while ( ( t = GTIter_next(tIt,pGEfrom, pGEto) ) ){
    const int d = GEN_type(pGEfrom);
    const int tagFrom = GEN_tag(pGEfrom);
    const int tagTo   = GEN_tag(pGEto);
    if (tagFrom < tagTo) {
      pVertex v;
      VIter   vIt= M_classifiedVertexIter(mesh, pGEfrom, 0);
      while ( ( v = VIter_next(vIt) ) ){
	pVertex v2 = getPeriodicVertex(mesh, t, v);
	if (!v2) throw runtime_error("unable to find the matching vertex");
	V_coord(v,xyzFrom);
	V_coord(v2,xyzTo);
	T_transformCoordinates(t,xyzFrom);	
	double locErr = 0.;
	for(int i = 0; i < dim; i++)
	  locErr += (xyzFrom[i] - xyzTo[i])*(xyzFrom[i] - xyzTo[i]);
	err += locErr;
	n++;
      }
      

      VIter_delete(vIt);
      pEdge e;
      EIter eIt= M_classifiedEdgeIter(mesh, pGEfrom, 0);
      while ( ( e = EIter_next(eIt) ) ){
      	pVertex v00 = E_vertex(e,0);
      	pVertex v10 = E_vertex(e,1);
      	pVertex v01 = getPeriodicVertex(mesh, t, v00);
      	pVertex v11 = getPeriodicVertex(mesh, t, v10);
      	pEdge   e1  = M_edge(mesh,V_id(v01),V_id(v11));
	if (!e1) throw runtime_error("unable to find the matching edge");
	
      	for(int i = 0 ; i < E_numPoints(e); i++) {
      	  if (i > 0) throw runtime_error("more than 1 point");
      	  pPoint p0   = E_point(e,i);
      	  pPoint p1   = E_point(e1,i);
	  V_coord((pVertex) p0,xyzFrom);
	  V_coord((pVertex) p1,xyzTo);
	  T_transformCoordinates(t,xyzFrom);
	  double locErr = 0.;
	  for(int i = 0; i < dim; i++)
	    locErr += (xyzFrom[i] - xyzTo[i])*(xyzFrom[i] - xyzTo[i]);
	  err += locErr;
	  n++;
      	}
      }
      EIter_delete(eIt);
    }
  }
  GTIter_delete(tIt);

  return sqrt(err) / n;

}


bool getMeshOrders(pMesh mesh, int dim, vector<int> &orders){

  orders.clear();
  if (dim == 2) {
    pFace f;
    FIter fIt= M_faceIter(mesh);
    while ( ( f = FIter_next(fIt) ) ) {
      const int order = EN_order((pEntity) f);
      if (!isIn(orders,order))
	orders.push_back(order);
      if (F_numVertices(f) != 3)
	return false;
    }
    FIter_delete(fIt);
    return true;
  } else if (dim == 3) {
    pRegion r;
    RIter rIt= M_regionIter(mesh);
    while ( ( r = RIter_next(rIt) ) ) {
      const int order = EN_order((pEntity) r);
      if (!isIn(orders,order))
	orders.push_back(order);
      if (R_numVertices(r) != 4)
	return false;
    }
    RIter_delete(rIt);
    return true;
  }
  return false;

}

bool getBoundaries(pMesh mesh,map<int,vector<pEntity> > &bMap) {

  const int dim = M_dim(mesh);
  const pGModel model = M_model(mesh);
  
  if (dim == 2) {
    pGEdge pGE;
    GEIter geIt = GM_edgeIter(model);
    while ( ( pGE = GEIter_next(geIt) ) ) {
      const int tag = GEN_tag((pGEntity) pGE);
      const int n   = M_numClassifiedEdges(mesh,(pGEntity)  pGE);
      bMap[tag].resize(n);
      pEdge e;
      EIter eIt = M_classifiedEdgeIter  (mesh, (pGEntity) pGE,0);
      int iEdge = 0;
      while ( ( e = EIter_next(eIt) ) ) {
	bMap[tag][iEdge++] = (pEntity) e;
      }
      EIter_delete(eIt);
    }
    GEIter_delete(geIt);
  } else if (dim == 3) {
    pGFace pGE;
    GFIter gfIt = GM_faceIter(model);
    while ( ( pGE = GFIter_next(gfIt) ) ) {
      const int tag = GEN_tag((pGEntity) pGE);
      const int n   = M_numClassifiedFaces(mesh,(pGEntity)  pGE);
      bMap[tag].resize(n);
      pFace f;
      FIter fIt = M_classifiedFaceIter  (mesh, (pGEntity) pGE,0);
      int iFace = 0;
      while ( ( f = FIter_next(fIt) ) ) {
	bMap[tag][iFace++] = (pEntity) f;
      }
      FIter_delete(fIt);
    }
    GFIter_delete(gfIt);
  }

  return false;
}
