// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------
// Testcase to check mapping for 1st order triangles element
// Load a mesh (square discretised through first order triangle)
// Build an entity through the given node label
// Compute the jacobian determinant of the mapping at each node
// Compare the computed values to the reference values in the txt file.
// for more details on the mapping features 
// please refer to the ELCI report L.2.4.2 (farid.benyoucef@cenaero.fr)

#include <string>
#include <fstream>
#include <iostream>
#include <sstream>
#include <cmath>
#include "ModelInterface.h"
#include "IMeshAdapter.h"
#include "ISizeField.h"
#include "PWLinearSField.h"
#include "MappingInterface.h"
#include "CheckMesh.h"
#ifdef _HAVE_MKL_
#include "mkl_service.h"
#endif

using namespace MAd;
using namespace std;

static void setSolutionBasedSize(IPiecewiseLinearSizeField * sizeField)
{
    double edge_length_mean = 0.;
    double jump_mean = 0.;
    double jump_std = 0.;
    pMeshDataId solutionID = MD_getMeshDataId("SolutionDataID");
    pMeshDataId solutionJumpID = MD_lookupMeshDataId("SolutionJumpDataID");
    pMesh mesh = sizeField->getMesh();

    // getStat
    {
        VIter vi = M_vertexIter(mesh);
        while (pVertex vert = VIter_next(vi)) {
            double v1(0.), tmp(0.);
            EN_getDataDbl((pEntity)vert, solutionID, &v1);

            for (int e = 0; e < V_numEdges(vert); e++)
            {
                double v2(0.);
                pEdge edge = V_edge(vert, e);
                EN_getDataDbl((pEntity)E_otherVertex(edge, vert), solutionID, &v2);
                double jump = abs(v2 - v1);
                jump_mean += jump;
                jump_std += jump*jump;
                tmp += jump;

                edge_length_mean += E_length(edge);
            }
            tmp /= V_numEdges(vert);
            EN_attachDataDbl((pEntity)vert, solutionJumpID, tmp);
        }
        edge_length_mean /= M_numEdges(mesh);
        jump_mean /= M_numEdges(mesh);
        jump_std /= M_numEdges(mesh);
        jump_std = std::sqrt(abs(jump_std - jump_mean*jump_mean));

        VIter_delete(vi);
    }

    // set sizes
    {
        cout << jump_mean << "( " << jump_std << " )" << endl;
        VIter vi = M_vertexIter(mesh);
        //! Foreach Vertices
        while (pVertex vert = VIter_next(vi)) {
            //! Get the precomputed jump of the solution
            double jp(0.0), coeff(1.0);
            EN_getDataDbl((pEntity)vert, solutionJumpID, &jp);
            double lenSq = V_meanEdgeLenSq(vert);
            coeff = std::sqrt(lenSq);

            if (jp > jump_mean + 0.4*jump_std) coeff = 0.5*std::sqrt(lenSq);
            if (jp < jump_mean - 0.1*jump_std) coeff = 2.0*std::sqrt(lenSq);

            pMSize pSV = MAd::CreateIsoSize(coeff);
            sizeField->setVSize(vert, pSV);
        }
        VIter_delete(vi);
    }
};

void attachSolution( pMesh mesh )
{
  pMeshDataId solID = MD_lookupMeshDataId("SolutionDataID");
  VIter vi = M_vertexIter(mesh);
  //! Foreach Vertices
  while ( pVertex vert = VIter_next(vi) ) {
    //! Get the precomputed jump of the solution
    double value(0.0);
    
    double x = V_coordComponent(vert,0);
    double y = V_coordComponent(vert,1);
    double z = V_coordComponent(vert,2);
    double r = std::sqrt( x*x+y*y+z*z);
    
//     value = sin( 2.*3.14*std::sqrt(x*x+y*y+z*z) );
    value = tanh( 16.0*(r-0.5) ) / tanh(16.0/2.0);
    
    EN_attachDataDbl((pEntity) vert,solID, value);
  }
  VIter_delete(vi);
  
}

int main(int argc, char* argv[])
{
  
#ifdef _HAVE_MKL_
  atexit(mkl_free_buffers);
#endif 

  pGModel null_model = 0; 
  GM_create(&null_model);
  pMesh p_mesh = M_new(null_model);

  std::string mshName( argv[1] );
//   mshName += "/naca0012.msh";
//   mshName += "/sphere.msh_old";
//   mshName += "/sphere_notregularized.msh";
//   mshName += "/sphere_regularized.msh";
//   mshName += "/dome.msh";
//   mshName += "/sphere.msh";
  mshName += "/damier.msh";

  M_load(p_mesh, mshName.c_str());
  M_writeMsh(p_mesh, "./initialMesh_SolutionBasedSF.msh", 2);
  int initOrder = M_maxOrder(p_mesh);

/* Start of code lines to check mapping functionalities   */
  attachSolution( p_mesh );
  IPiecewiseLinearSizeField * sizeField = CreatePiecewiseLinearSizeField(p_mesh);
  setSolutionBasedSize(sizeField);
  sizeField->printPosIsotropic(p_mesh,"./initialMesh_SolutionBasedSF.pos");
  IMeshAdapter* adapter = IMeshAdapter::BuildMeshAdapter(p_mesh, sizeField);

  if (!checkMesh(p_mesh)) {
    cerr << "Problem with initial Mesh !!" << endl;
    delete adapter;
    delete sizeField;
    M_delete( p_mesh );
    GM_delete( null_model ); 
    return 1;
  }

//     adapter->setEdgeLenSqBounds(1.0,2.5);
    adapter->setEdgeLenSqBounds(1.0/12.0,2.5);
// 	adapter->setEdgeLenSqBounds(1.0/9.0,1.0);
  adapter->setSwapOnBoundary(false);
  adapter->setCollapseOnBoundary(false);   //! \warning PROBLEM HERE !!!!!! Force to False
// 	adapter->setMaxIterationsNumber(2);
	adapter->run();

  std::string pref = "./";
	adapter->setOutputDirectoryPath(pref.c_str());
  std::string fn="adapted_mesh_SolutionBasedSF.msh";
  adapter->writeMsh(fn.c_str());
  
  /* The resulting mesh is conform ? */
  if (!checkMesh(p_mesh)) {
    cerr << "Problem with Mesh !!" << endl;
//     M_delete( p_mesh );
//     GM_delete( null_model ); 
//     return 1;
  }
  
// memory cleaning
  delete adapter;
  delete sizeField;
	M_delete( p_mesh );
	
/* Start of code lines to check final order  */	
	p_mesh = M_new(null_model);
	M_load(p_mesh, (pref+fn).c_str());
	int finOrder = M_maxOrder(p_mesh);
	
	cout << "Initial Mesh Order : " << initOrder << endl;
	cout << "Final   Mesh Order : " << finOrder << endl;
	
	M_delete( p_mesh );
	GM_delete( null_model ); 
	
	if (initOrder != finOrder)
		return 1;
	else
		return 0;
	
}
