// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------
// Testcase to check mapping for 1st order triangles element
// Load a mesh (square discretised through first order triangle)
// Build an entity through the given node label
// Compute the jacobian determinant of the mapping at each node
// Compare the computed values to the reference values in the txt file.
// for more details on the mapping features 
// please refer to the ELCI report L.2.4.2 (farid.benyoucef@cenaero.fr)

#include <string>
#include <fstream>
#include <iostream>
#include <sstream>
#include "MAdConfig.h"
#include "ModelInterface.h"
#include "IMeshAdapter.h"
#include "ISizeField.h"
#include "MappingInterface.h"
#ifdef _HAVE_MKL_
#include "mkl_service.h"
#endif
#ifdef _WIN32
#ifdef _MADLIB_STATIC_LINKING_
#define _HAVE_REGUL_TOOLS_ 1
#else
#undef _HAVE_REGUL_TOOLS_
#endif 
#else
#define _HAVE_REGUL_TOOLS_ 1
#endif

#ifdef _HAVE_REGUL_TOOLS_
#include "regularizer.h"
#endif
using namespace MAd;
using namespace std;

int main(int argc, char* argv[])
{
  
#ifdef _HAVE_MKL_
  atexit(mkl_free_buffers);
#endif   
  
  pGModel null_model = 0; 
  GM_create(&null_model);
  pMesh p_mesh = M_new(null_model);

  std::string mshName( argv[1] );
//     mshName += "/sphere.msh";
  mshName += "/sphere_coarse.msh";
//   mshName += "/phill_tetra.msh";
//   mshName += "/extruded_nozzle.msh";
//   mshName += "/testNotNURBS.msh";
//   mshName += "/SphereInSphere.msh";
//   mshName += "/Naca_extrude.msh";

  M_load(p_mesh, mshName.c_str());
  M_writeMsh(p_mesh, "./initialMesh_tetra2O.msh", 2);
  int initOrder = M_maxOrder(p_mesh);

#ifdef _HAVE_REGUL_TOOLS_
  double MinQuality = 1.e+16;
  double MeanQuality = 0.0;
  double MaxQuality = -1.e+16;

  // -------------------------------------------------------------------- //
  // FACES
  FIter Fi = M_faceIter(p_mesh);
  while( pFace face = FIter_next(Fi) )
  {
    RegulTools::Quality qual(face);
    double q = qual.eval();
    
    MinQuality = min(MinQuality,q);
    MeanQuality+= q;
    MaxQuality = max(MaxQuality,q);
  }
  FIter_delete(Fi);
  MeanQuality /= M_numFaces(p_mesh);
  printf("Distortion based Mesh Quality (Faces):\n min  = %f \n mean = %f \n max  = %f \n",MinQuality,MeanQuality,MaxQuality);
  // -------------------------------------------------------------------- //
  // -------------------------------------------------------------------- //
  // REGION
  MinQuality = 1.e+16;
  MeanQuality= 0.0;
  MaxQuality =-1.e+16;
  RIter Ri = M_regionIter(p_mesh);
  while( pRegion region = RIter_next(Ri) )
  {
    RegulTools::Quality qual(region);
    double q = qual.eval();
    
    MinQuality = min(MinQuality,q);
    MeanQuality+= q;
    MaxQuality = max(MaxQuality,q);
  }
  RIter_delete(Ri);
  MeanQuality /= M_numRegions(p_mesh);
  printf("Distortion based Mesh Quality (Regions):\n min  = %f \n mean = %f \n max  = %f \n",MinQuality,MeanQuality,MaxQuality);
  // -------------------------------------------------------------------- //
#endif
  
/* Start of code lines to check Adaptation functionalities   */

// 	PWLSField* sizeField=new PWLSField(p_mesh);
// 	sizeField->setCurrentSize();

  IAnalyticalSizeField* sizeField = CreateAnalyticalSizeField();
//   sizeField->setSize(" max( sqrt( (x)*(x) + (z)*(z) ) , 0.10 )");
//   sizeField->setSize("max( tanh( 16.0*abs(y+0.5) ) / tanh(16.0/2.0) , 0.01 )");
  sizeField->setSize("min( max( abs(z-1.0) , 0.2 ) , 0.8)");
//   sizeField->setSize("min( max( abs(y) , 0.1 ) , 0.9)");
//   sizeField->setSize("min( max( abs(y+0.5) , 0.1 ) , 0.9)");
//   sizeField->setSize("max( abs(x-1) , 0.10 )");
//   sizeField->setSize("max( 2.0*(x+1.5) , 0.10 )");
//   sizeField->setSize("max( 2.0*(z+0.5) , 0.10 )");
  IMeshAdapter* adapter = IMeshAdapter::BuildMeshAdapter(p_mesh, sizeField);
//   adapter->setEdgeLenSqBounds(1.0,2.5);
  adapter->setEdgeLenSqBounds(1.0/9.0,1.0);
//   adapter->activateCollapseOp(true);
//   adapter->activateSwapOp(true);
//   adapter->activateSplitOp(true);
  
//   adapter->setEdgeLenSqBounds( 1.0/3.0, 3.0 );
//   adapter->setNoSwapQuality( 0.1 );
  adapter->setNoSwapQuality( 0.6 );
  adapter->setSliverQuality( 0.02 );
  //   adapter->setSliverQuality( 0.2 );
  adapter->setCollapseOnBoundary(true,100.);
  adapter->setSwapOnBoundary(true,100.);
  adapter->setSliverPermissionInESplit( true, 1.5 );
  adapter->setSliverPermissionInECollapse( false, 0.1 ); 
  
  
  adapter->setMaxIterationsNumber(15);
//   adapter->setMaxIterationsNumber(4);
//   adapter->setMaxIterationsNumber(3);
//   adapter->setMaxIterationsNumber(2);
//   adapter->setMaxIterationsNumber(1);
  adapter->run();
  
//   adapter->eCollapseLoop(); 
//   adapter->edgeSwapLoop();
//   adapter->eSplitCollapseLoop();
//   adapter->removeSlivers();
//   adapter->eSplitLoop();
//   
//   adapter->eCollapseLoop(); 

#ifdef _HAVE_REGUL_TOOLS_
  // -------------------------------------------------------------------- //
  // REGION
  MinQuality = 1.e+16;
  MeanQuality= 0.0;
  MaxQuality =-1.e+16;
  Ri = M_regionIter(p_mesh);
  while( pRegion region = RIter_next(Ri) )
  {
    RegulTools::Quality qual(region);
    double q = qual.eval();
    
    MinQuality = min(MinQuality,q);
    MeanQuality+= q;
    MaxQuality = max(MaxQuality,q);
  }
  RIter_delete(Ri);
  MeanQuality /= M_numRegions(p_mesh);
  printf("Distortion based Mesh Quality (Regions):\n min  = %f \n mean = %f \n max  = %f \n",MinQuality,MeanQuality,MaxQuality);
  // -------------------------------------------------------------------- //
  // -------------------------------------------------------------------- //
  // FACES
  MinQuality = 1.e+16;
  MeanQuality= 0.0;
  MaxQuality =-1.e+16;
  Fi = M_faceIter(p_mesh);
  while( pFace face = FIter_next(Fi) )
  {
    RegulTools::Quality qual(face);
    double q = qual.eval();
    
    MinQuality = min(MinQuality,q);
    MeanQuality+= q;
    MaxQuality = max(MaxQuality,q);
  }
  FIter_delete(Fi);
  MeanQuality /= M_numFaces(p_mesh);
  printf("Distortion based Mesh Quality (Faces):\n min  = %f \n mean = %f \n max  = %f \n",MinQuality,MeanQuality,MaxQuality);
  // -------------------------------------------------------------------- //
#endif

	std::string pref = "./";
	adapter->setOutputDirectoryPath(pref.c_str());
	std::string fn="adapted_mesh_tetra2O.msh";
  adapter->writeMsh(fn.c_str());
  
  // -------------------------------------------------------------------- //
  
  /* The resulting mesh is conform ? */
//  if (!checkMesh(p_mesh)) {
//     delete adapter;
//     delete sizeField;
//     M_delete( p_mesh );
//     GM_delete( null_model ); 
//     return 1;
//   }
  
	
// memory cleaning
  delete adapter;
  delete sizeField;
	M_delete( p_mesh );
	
/* Start of code lines to check final order  */	
	p_mesh = M_new(null_model);
	M_load(p_mesh, (pref+fn).c_str());
	int finOrder = M_maxOrder(p_mesh);
	
	cout << "Initial Mesh Order : " << initOrder << endl;
	cout << "Final   Mesh Order : " << finOrder << endl;
	
	M_delete( p_mesh );
	GM_delete( null_model ); 
	
	if (initOrder != finOrder)
		return 1;
	else
		return 0;
	
}
