// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------
// Testcase to check mapping for 2nd order tetraedra element
// Load a mesh (circle discretised through 2nd order tetraedra)
// Build an entity through the given node label
// Compute the jacobian determinant of the mapping at each node
// Compare the computed values to the reference values in the txt file.
// for more details on the mapping features 
// please refer to the ELCI report L.2.4.2 (farid.benyoucef@cenaero.fr)

#include <string>
#include <fstream>
#include <iostream>
#include <sstream>
#include <algorithm>
#include <cmath>
#include <vector>
#include "ModelInterface.h"
#include "MappingInterface.h"
#include "PWLinearSField.h"
#ifdef _HAVE_MKL_
#include "mkl_service.h"
#endif
const double lengthscale=1.0;

using namespace MAd;

int main(int argc, char* argv[])
{
  
#ifdef _HAVE_MKL_
  atexit(mkl_free_buffers);
#endif  

  pGModel null_model = 0; 
  GM_create(&null_model);
  pMesh p_mesh = M_new(null_model);

  std::string mshName( argv[1] );
  mshName += "/cylinder.msh";
  
  M_load(p_mesh, mshName.c_str());
   
/* Start of code lines to check mapping functionalities   */
  pVertex verts[10];
  verts[0]=M_vertex(p_mesh,99);
  verts[1]=M_vertex(p_mesh,768);
  verts[2]=M_vertex(p_mesh,161);
  verts[3]=M_vertex(p_mesh,84);
  verts[4]=M_vertex(p_mesh,1403);
  verts[5]=M_vertex(p_mesh,1402);
  verts[6]=M_vertex(p_mesh,1399);
  verts[7]=M_vertex(p_mesh,1398);
  verts[8]=M_vertex(p_mesh,1401);
  verts[9]=M_vertex(p_mesh,1400);


  //
  int verts_id[10];
  for(int i=0; i<10; i++){verts_id[i]=V_id(verts[i]);}
  pRegion tet2o=M_createTet(p_mesh, 2, false, verts_id);
  //  
  double xyz[10][3];
  R_coord(tet2o,xyz);
  pMapping p_maps = Map_create((pEntity) tet2o, lengthscale);

  std::vector<double> detJac;
  double uvw[3];
  
  for (int i=0;i<3;i++) {
    uvw[0] = i*0.5;
    for (int j=2-i;j<3;j++) {
      uvw[1] = j*0.5;
      for (int k=2-i-j;k<3;k++) {
        uvw[2] = k*0.5;
        detJac.push_back(Map_determinant(p_maps,uvw));
      }
    }
  }
  
  Map_delete(p_maps);
  M_removeRegion(p_mesh, tet2o);
  M_delete(p_mesh);
  GM_delete( null_model );

  // compare computed data to reference value computed analytically 
  // written in the qa_values_check.txt file
  // non-regression tescase database   
  
  std::string refName( argv[1] );
  refName += "/qa_values_check.txt";
  std::ifstream ref_file( refName.c_str() );
 
  int nbr_ref_values;
  ref_file>>nbr_ref_values; 
  bool works_fine=true;
  double eps=1e-06;
  for(int i=0; i<std::min(10,nbr_ref_values); i++){
    double ref;
    ref_file>> ref;
    if (std::fabs(ref-detJac[i]) > eps) {
      std::cout << "Error in quadratic tetrahedron mapping on point " << i 
                << " : " << detJac[i] << " vs " << ref << std::endl;
      works_fine = false;
    }
  }
  
  return works_fine ? 0:1;
}
