// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------
// Testcase to check mapping for 1st order triangles element
// Load a mesh (square discretised through first order triangle)
// Build an entity through the given node label
// Compute the jacobian determinant of the mapping at each node
// Compare the computed values to the reference values in the txt file.
// for more details on the mapping features 
// please refer to the ELCI report L.2.4.2 (farid.benyoucef@cenaero.fr)

#include <string>
#include <fstream>
#include <iostream>
#include <sstream>
#include <algorithm>
#include <cmath>
#include "ModelInterface.h"
#include "MappingInterface.h"
#include "PWLinearSField.h"
#ifdef _HAVE_MKL_
#include "mkl_service.h"
#endif
const double lengthscale=1.0;

using namespace MAd;
using namespace std;

int main(int argc, char* argv[])
{
  
#ifdef _HAVE_MKL_
  atexit(mkl_free_buffers);
#endif  

  pGModel null_model = 0; 
  GM_create(&null_model);
  pMesh p_mesh = M_new(null_model);

  std::string mshName( argv[1] );
  mshName += "/carre.msh";
  
  M_load(p_mesh, mshName.c_str());
  int numF=M_numFaces(p_mesh);
  cout << "numF: " << numF << endl;
  int numVM=M_numVertices(p_mesh);
  cout << "numV: " << numVM << endl; 
   
/* Start of code lines to check mapping functionalities   */
  pVertex verts[3];
  verts[0]=M_vertex(p_mesh,1);
  verts[1]=M_vertex(p_mesh,5);
  verts[2]=M_vertex(p_mesh,30); 
//
  pFace tri=M_createF(p_mesh, 3, verts); //create a triangle entity connected by previously defined nodes label
//  
  pMapping p_maps = Map_create((pEntity) tri, lengthscale); //create a mapping object


  double detJac[3];
  double uvw[3];

  int idx = 0;
  for (int i=0;i<2;i++) {
    uvw[0] = i*1.;
    for (int j=0;j<2-i;j++) {
      uvw[1] = j*1.;
      uvw[2] = 0;
      
      detJac[idx++] = Map_determinant(p_maps,uvw);
    }
  }
  
  Map_delete(p_maps);
  M_removeFace(p_mesh, tri);
  // End of code lines to check mapping functionalities 
  M_delete( p_mesh );
  GM_delete( null_model );

  // compare computed data to reference value computed analytically 
  // written in the qa_values_check.txt file
  
  // non-regression tescase database  
  std::string refName( argv[1] );
  refName += "/qa_values_check.txt";
  std::ifstream ref_file( refName.c_str() );
  
  int nbr_ref_values;
  ref_file >> nbr_ref_values;
  
  bool works_fine=true;
  double eps=1e-06;
  for(int i=0; i<std::min(3,nbr_ref_values); i++){
    double ref;
    ref_file>>ref;
    if (std::fabs(ref-detJac[i]) > eps) {
      std::cout << "Error in linear triangle mapping on point " << i 
                << " : " << detJac[i] << " vs " << ref << std::endl;
      works_fine = false;
    }
  }

  return works_fine ? 0:1;
}
