// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------
// Testcase to check mapping for 1st order triangles element
// Load a mesh (square discretised through first order triangle)
// Build an entity through the given node label
// Compute the jacobian determinant of the mapping at each node
// Compare the computed values to the reference values in the txt file.
// for more details on the mapping features 
// please refer to the ELCI report L.2.4.2 (farid.benyoucef@cenaero.fr)

#include <string>
#include <fstream>
#include <iostream>
#include <sstream>
#include "MAdConfig.h"
#include "ModelInterface.h"
#include "IMeshAdapter.h"
#include "ISizeField.h"
#include "MappingInterface.h"

#ifdef _HAVE_MKL_
#include "mkl_service.h"
#endif
#ifdef _WIN32
  #ifdef _MADLIB_STATIC_LINKING_
    #define _HAVE_REGUL_TOOLS_ 1
  #else
    #undef _HAVE_REGUL_TOOLS_
  #endif 
#else
  #define _HAVE_REGUL_TOOLS_ 1
#endif

#ifdef _HAVE_REGUL_TOOLS_
#include "regularizer.h"
#endif

using namespace MAd;
using namespace std;

int main(int argc, char* argv[])
{
  
#ifdef _HAVE_MKL_
  atexit(mkl_free_buffers);
#endif 
  
  double coord[3],uvw[3];
  std::string mshName( argv[1] );
  pGModel null_model = 0; 
  GM_create(&null_model);
  pMesh p_mesh = M_new(null_model);

  // -------------------------------------------------------------------- //
  
  pVertex TetVerts[4], TriVerts[3];
  TetVerts[0] = M_createV(p_mesh,0.,0.,0.,-1);
  TetVerts[1] = M_createV(p_mesh,1.,0.,0.,-1);
  TetVerts[2] = M_createV(p_mesh,0.5,sqrt(3.)/2.,0.,-1);
  TetVerts[3] = M_createV(p_mesh,0.5,sqrt(3.)/6.,sqrt(2.)/sqrt(3.),-1);
  
  TriVerts[0] = TetVerts[0];
  TriVerts[1] = TetVerts[1];
  TriVerts[2] = TetVerts[2];
  
  pFace   Tri = M_createF(p_mesh,3,TriVerts);
  pRegion Tet = M_createR(p_mesh,4,TetVerts);
  
  // -------------------------------------------------------------------- //
#ifdef _HAVE_REGUL_TOOLS_
  RegulTools::Quality TriQual(Tri);
  double TriQuality = TriQual.eval();
  
  RegulTools::Quality TetQual(Tet);
  double TetQuality = TetQual.eval();
  
  printf("  --------------------------------------------------------------------\n");
  printf("Distortion based Mesh Quality of 1st Order Triangle   (TRI):%f\n",TriQuality);
  printf("Distortion based Mesh Quality of 1st Order Tetrahedra (TET):%f\n",TetQuality);
  printf("  --------------------------------------------------------------------\n");
#endif
  M_removeRegion(p_mesh,Tet);
  M_delete( p_mesh );
  // -------------------------------------------------------------------- //
  pGModel model = NULL;
  GM_create(&model,"");
  p_mesh = M_new(model);
  mshName += "/triangle.msh";
  M_load(p_mesh, mshName.c_str());
  
  pFace Tri3o = NULL;
  FIter Fi = M_faceIter(p_mesh);
  Tri3o = FIter_next(Fi);
  FIter_delete(Fi);
  
#ifdef _HAVE_REGUL_TOOLS_
  RegulTools::Quality qu(Tri3o);
  
//   qu.setPositive();
  
  double Tri3oQuality = qu.eval();
  printf("Distortion based Mesh Quality of 3st Order Triangle   (TRI):%f\n",Tri3oQuality);
  printf("  --------------------------------------------------------------------\n");
#endif

  pVertex vert = F_vertex(Tri3o,2);
  
  V_coord(vert,coord);
  printf("Distortion of 3rd Triangle Node for y in range [%f,%f] : \n",coord[1],coord[1]-29*0.05);
  for (int c = 0 ; c < 30 ; c++)
  {
    V_setCoord(vert,coord);

#ifdef _HAVE_REGUL_TOOLS_
    printf(" (%f,%f) --> %f\n",coord[0],coord[1],qu.eval());
#endif
    coord[1]-=0.05;
  }  
  
  for (int i = 0 ; i < 11 ; i++)
  {
    pMapping mapping = Map_create((MAd::pEntity)(Tri3o),1.0);
    uvw[0] = 0.5 - i/10.*0.5;
    uvw[1] = 0.0 + i/10.;
    uvw[2] = 0.0;
    double det = Map_determinant(mapping,uvw);
    double jac[9];
    Map_deval(mapping,uvw,jac);
    Map_delete(mapping);
    
    double d = (jac[0]*jac[4] + jac[1]*jac[5] + jac[2]*jac[3]) - (jac[0]*jac[5] + jac[1]*jac[3] + jac[2]*jac[4]);
    printf(" u,v = (%f,%f) Determinant = (%f , %f)\n",uvw[0],uvw[1],det,d);
  }
//   M_writeMsh(p_mesh, "./Modified_TriangleO3.msh", 2);
  printf("  --------------------------------------------------------------------\n");
  // -------------------------------------------------------------------- //
  
  M_delete( p_mesh );
  GM_delete( null_model ); 
  GM_delete( model ); 
  
  // -------------------------------------------------------------------- //
#ifdef _HAVE_REGUL_TOOLS_
  if ( abs(TriQuality - 1.0) < 1.e-10 && abs(TetQuality - 1.0) < 1.e-10 && abs(Tri3oQuality - 1.0) < 1.e-10 )
  {
    printf("OK\n");
    return 0;
  }
  else
  {
    printf("NOK\n");
    return 1;
  }
#else
  return 0;
#endif
  // -------------------------------------------------------------------- //
  

}
