// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#include "MAdTransformation.h"

#include <cstdlib>
#include <cstring>
#include <limits>

namespace MAd {
  
  /*! \brief empty constructor */
  MAdTransformation::MAdTransformation():
    affine(4,4) { 
    affine.set_identity();
    updateDecomposition();
  }

  
  /*! \brief constructor based upon specified affine transformation */
  MAdTransformation::MAdTransformation(const doubleMatrix& aff):
    affine(aff) {
    updateDecomposition();
  }
  
  /*! \brief copy constructor */
  MAdTransformation::MAdTransformation(const MAdTransformation& other):
    affine(other.affine) {
    updateDecomposition();
  }
  
  /*! \brief destructor */
  MAdTransformation::~MAdTransformation() {}
  
  /*! \brief chain with another rotation */
  void MAdTransformation::applyAffineTransformation(const double* mat) {
    
    doubleMatrix newAffine(4,4);
    const double* pMat = mat;
    for (int i=0;i<4;i++) for (int j=0;j<4;j++) newAffine(i,j) = *(pMat++);
    affine.blas_dgemm(affine,newAffine,1.,0.);
    updateDecomposition();
  }
  
  /*! \brief invert the affine transformation */
  bool MAdTransformation::invert() {
    affine = affine.inverse();
    updateDecomposition();
    return true;
  }
 
  /*! \brief scale translation */
  void MAdTransformation::scaleTranslation(double s) {
    for(int i=0; i<3; i++) affine(i,3) *= s;
    updateDecomposition(); 
  }
  
  /*! \brief raise to nth power */
  bool MAdTransformation::raiseToPower(int n) {
    if(n==0) 
      affine.set_identity();
    else{
      if(n<0) affine = affine.inverse();
      doubleMatrix tmp1 = affine;
      for(int i=1; i<abs(n); i++){
        doubleMatrix tmp2 = affine;
        affine.blas_dgemm(tmp1,tmp2,1,0);
      }
    }
    updateDecomposition(); 
    return true;
  }
  
  /*! \brief copy the transformation to an external array */
  void MAdTransformation::copyAffineTransformation(double* mat) const {
    double* pMat = mat;
    for (int i=0;i<4;i++) for (int j=0;j<4;j++) *(pMat++) = affine(i,j);
  }
  
  /*! \brief transform coordinates - implies rotation and translation */

  void MAdTransformation::transformCoordinates(double* xyz, const int n)  const {
    if (isIdentity()) return;
    doubleVector vec1(4);
    doubleVector vec2(4);
    std::memcpy(vec1.touchPtr(),xyz,3*sizeof(double));
    vec1(3) = 1;
    for(int i=0; i<n; i++) affine.mult(vec1,vec2,false);
    std::memcpy(xyz,vec2.seePtr(),3*sizeof(double));
  }

  /*! \brief transform a tensor of order 1 - implies only rotation */
  /*! if inverse -> invert transformation */
  
  void MAdTransformation::transformVector(double* vec,bool inv) const {
    if (isIdentity()) return;
    doubleVector vec1(4);
    doubleVector vec2(4);
    std::memcpy(vec1.touchPtr(),vec,3*sizeof(double));
    vec1(3) = 0;
    affine.mult(vec1,vec2,inv);
    std::memcpy(vec,vec2.seePtr(),3*sizeof(double));
  }
  
  /*! \brief transform a second order tensor - implies rotation */

  void MAdTransformation::transformTensor2(double*,bool) const {
    if (isIdentity()) return;
    throw;
  }
  
  /*! \brief transform a second order tensor - implies rotation */

  void MAdTransformation::transformSymTensor2(double*,bool) const {
    if (isIdentity()) return;
    throw;
  }
  
  /*! \brief print to screen */

  void MAdTransformation::print() const {
    printf("Affine transformation");
    affine.print("A");
  }

  bool MAdTransformation::operator< (const MAdTransformation& other) const {
    
    for (int i=0;i<3;i++) {
      for (int j=0;j<4;j++) {
        double da = affine(i,j) - other.affine(i,j);
        if (fabs(da) > 1.e-4) {
          if (da > 0.0) return false;
          if (da < 0.0) return true;
        }
      }
    }
    return false;
  }

  bool MAdTransformation::describe(std::ostream& out,bool printAffine) const {
    
    if (out.bad()) return false;

    bool check(true);

    if (isIdentity()) {
      out << "Identity transformation\n";
      check = false;
    }
    if (rotation) {
      out << "Rotation of " << angle * 180./acos(-1.) << " around axis "
          << "(" << axis[0] << "," << axis[1] << "," << axis[2] << ")"
          << ", passing through "
          << "(" << center[0] << "," << center[1] << "," << center[2] << ")\n";
      check = false;
    }
    if (translation) {
      out << "Translation by (" 
          << displacement[0] << "," 
          << displacement[1] << "," 
          << displacement[2] << ")\n";
      check = false;
    }

    if (printAffine || check) {
      out << "\nAffine matrix = \n";
      affine.print(out);
    }
    return true;
  }

  bool MAdTransformation::isCyclic(int& nb) const {

    if (isIdentity()) {nb = 1; return true;}
    static const double twoPi = 2*acos(-1.);
    static const double tol = std::numeric_limits<double>::epsilon()*1e2;
    if (translation) return false;
    
    if (rotation) {
      int i = int(twoPi / fabs(angle));
      if (fabs(twoPi - i*fabs(angle)) > tol) return false;
      nb = i;
      return true;
    }
    return false;
  }
  
  bool MAdTransformation::isIdentity() const {
    if (translation) return false;
    if (rotation)    return false;
    if (scaling)     return false;
    return true;
  }
  
  bool MAdTransformation::isRotation(double* a,
                                     double* c,
                                     double& aa) const {
    if (!rotation) return false;
    std::memcpy(a,axis,3*sizeof(double));
    std::memcpy(c,center,3*sizeof(double));
    aa = angle;
    return true;
  }

  bool MAdTransformation::isTranslation(double* t) const {
    
    if (!translation) return false;
    std::memcpy(t,displacement,3*sizeof(double));
    return true;
  }

  bool MAdTransformation::isScaling() const {
    return scaling;
  }

  void MAdTransformation::updateDecomposition()  {

    static const double tol = std::numeric_limits<double>::epsilon()*1e2;

    // initialize
    
    for (int i=0;i<3;i++) displacement[i] = affine(i,3);
    for (int i=0;i<3;i++) axis[i] = center[i] = 0;
    angle = 0;
    
    rotation = translation = scaling = false;

    // --- check for rotation

    for (int i=0;i<3;i++) {
      for (int j=0;j<3;j++) {
        rotation = rotation || (i!=j && fabs(affine(i,j)) > tol);
      }
    }
    
    // --- decompose the rotation

    if (rotation) {
      
      // get the axis 
      
      double RmRT[3][3];  
      for (int i=0;i<3;i++) {
        for (int j=0;j<3;j++) {
          RmRT[i][j] = (affine(i,j) - affine(j,i));
          rotation = rotation || (RmRT[i][j] != 0);
        }
      }
      
      axis[0] = RmRT[1][2];
      axis[1] = RmRT[2][0];
      axis[2] = RmRT[0][1];

      // normalize the axis

      double da = 0;
      for (int i=0;i<3;i++) da += axis[i]*axis[i];
      da = 1./sqrt(da);
      for (int i=0;i<3;i++) axis[i] *= da;
      
      // compute two orthogonal axes axis2 and axis3

      int mi = 0;
      for (int i=1;i<3;i++) if (fabs(axis[i]) > fabs(axis[mi])) mi = i;
      
      double axis2[3] = {0,0,0};
      axis2[(mi+1)%3] = 1;
      da = 0;
      for (int i=0;i<3;i++) da += axis[i]*axis2[i];
      for (int i=0;i<3;i++) axis2[i] -= axis[i]*da;
      da = 0;
      for (int i=0;i<3;i++) da += axis2[i]*axis2[i];
      da = 1./sqrt(da);
      for (int i=0;i<3;i++) axis2[i] *= da;
      
      
      double axis3[3];
      for (int i=0;i<3;i++) {
        int ip1 = (i+1)%3;
        int ip2 = (i+2)%3;
        axis3[i] = axis[ip1]*axis2[ip2] - axis[ip2]*axis2[ip1];
      }
      
      // now transform axis2 and use projections on axis2 and axis3 to find
      // the angle

      double prod[3];
      for (int i=0;i<3;i++) {
        prod[i] = 0;
        for (int j=0;j<3;j++) { 
          prod[i] += affine(i,j)*axis2[j];
        }
      }
      
      double cosA = 0;
      double sinA = 0;
      
      for (int i=0;i<3;i++) {
        cosA += axis2[i]*prod[i];
        sinA += axis3[i]*prod[i];
      }
      
      angle = std::atan2(sinA,cosA);
      
      // decompose the translation in a rotation center and a displacement
      // along the axis

      da = 0;
      for (int i=0;i<3;i++) da += displacement[i]*axis[i];
      for (int i=0;i<3;i++) displacement[i] = da * axis[i];
      
      for (int i=0;i<3;i++) {
        center[i] = 0;
        for (int j=0;j<3;j++) {
          center[i] -= affine(j,i)*(affine(j,3) - da*axis[i]);
        }
      }
    }
    
    // now check for a non-trivial displacement

    for (int i=0;i<3;i++) {
      translation = translation || (fabs(displacement[i]) > tol);
    }

    // now check for coordinate scaling

    if(!rotation){
      for (int i=0;i<3;i++) {
        scaling = scaling || (fabs(affine(i,i)-1) > tol);
      }
    }
    
  }
  
  int MAdTransformation::computeAlignedMultiple(const double* x,
                                                const double* y) const {
    
    if (rotation && translation) throw;
    
    if (rotation) {
      int maxNbRotations = int(2*acos(-1)/angle + 1);
      double tx[3] = {x[0],x[1],x[2]};
      double dist = std::numeric_limits<double>::max();
      int nbRotations = 0;
      for (int iRot=0;iRot<maxNbRotations;iRot++) {
        double newdist = 0;
        for (int i=0;i<3;i++) newdist += (y[i]-tx[i])*(y[i]-tx[i]);
        if (newdist < dist) { 
          nbRotations = iRot;
          dist = newdist;
        }
        double tmp[3] = {0,0,0};
        for (int i=0;i<3;i++) for (int j=0;j<3;j++) tmp[i] += affine(i,j)*tx[j];
        std::memcpy(tx,tmp,3*sizeof(double));
      }
      return nbRotations;
    }
    
    if (translation) {
      double da = 0;
      double dp = 0;
      for (int i=0;i<3;i++) {
        double dx = y[i] - x[i];
        da += dx*displacement[i];
        dp += dx*dx;
      }
      return int(da/dp);
    }
    return false;
  }
}
