// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#ifndef H__MADTRANSFORMATION
#define H__MADTRANSFORMATION

#include <cstring>
#include <set>

#include "MAdMatrix.h"

namespace MAd {
  
  class MAdTransformation {
    
  public:
    
    MAdTransformation();
    MAdTransformation(const MAdTransformation&);
    MAdTransformation(const doubleMatrix&);
    ~MAdTransformation();

  private:
    
    void updateDecomposition();

  public:

    MAdTransformation* duplicate() const {return new MAdTransformation(*this);}
    MAdTransformation* inverse() const {
      MAdTransformation* newTfo = new MAdTransformation(*this);
      newTfo->invert();
      return newTfo;
    }

    bool compatible(const MAdTransformation& other) const {
      return (affine == other.affine);
    }

    void copyMatrix(double mat[16]) const {
      int idx = 0;
      for (int i=0;i<4;i++) for (int j=0;j<4;j++) mat[idx++] = affine(i,j);
    }

    bool operator < (const MAdTransformation&) const;

    bool compose(MAdTransformation* other) {
      if (!other) return false;
      doubleMatrix tmp(4,4);
      tmp.blas_dgemm(other->affine,affine,1,0);
      affine = tmp;
      return true;
    }

          doubleMatrix& matrix()       {return affine;}
    const doubleMatrix& matrix() const {return affine;}
    
  public:

    bool invert();
    bool raiseToPower(int);
    void scaleTranslation(double);
    void applyAffineTransformation(const double*);
    void copyAffineTransformation(double*) const;
   
  public:
    
    bool describe(std::ostream&,bool=false) const;
    
  public:
    
    /*! \brief return true if is an identity transformation */
    bool isIdentity() const;

    /*! \brief return true if there is a rotation */
    bool isRotation(double* axis,double* origin,double& angle) const;
    
    /*! \brief return true if there is a non-trivial translation 
      (ie. one that does not result from a non-trivial rotation center) */
    bool isTranslation(double* displacement) const;

    /*! \brief return true if there is a coordinate scaling */
    bool isScaling() const;

    /*! \brief check whether there is an integer multiple that recovers
        the original position */
    bool isCyclic(int& nbPeriods) const;

    /*! \brief compute the number of applications that aligns best 2 points */
    int computeAlignedMultiple(const double*,const double*) const;
    
  protected:
    
    enum Transformed{FORWARD=1,BACKWARD=-1,NOT=0};

  public:
    
    void transformCoordinates(double* xyz,    //!< coordinates, in place
                              const int nb=1  //!< number of applications
                              ) const;
    void transformVector     (double* vec,    //!< components, in place
                              bool inv =false //!< apply inverse
                              )  const;
    void transformTensor2    (double* tens,   //!< components, in place
                              bool inv =false //!< apply inverse
                              )  const;
    void transformSymTensor2 (double* vec,    //!< components, in place
                              bool inv =false //!< apply inverse
                              )  const;

  public:

    void print() const;
    
  protected:
    
    /*! \brief compact storage of an affine matrix */
    /*! The affine transformation 
      \f$\mathbf x^\prime = \mathbf R \cdot \mathbf \x + \mathbf d\f$ 
      is stored as a partitioned matrix \f$ \mathbf A \in \mathbb R^{4\times 4} \f$
      \f$ \mathbf A = 
      \begin{matrix}
      \mathbf R & \mathbf d \\ 
      \mathbf o^T & 1 \end{matrix} 
      \f$
    */
       
    doubleMatrix affine; 
    
    bool rotation;
    bool translation;
    bool scaling;
    
    double axis[3];
    double center[3];
    double angle;
    double displacement[3];
    
  };
  
  // // ---------------------------------------------------------------------------

  // struct MAdTransformationLess {
  //   bool operator() (const MAdTransformation* a,
  //                    const MAdTransformation* b) const {
  //     if (!a) return true;
  //     if (!b) return false;
  //     return (*a) < (*b);}
  // };
  
  // typedef std::set<MAdTransformation*,MAdTransformationLess> TransformationSet;
  
}

// -----------------------------------------------------------------------------

#endif
