// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------


#include "TransformationInterface.h"
#include "MAdTransformation.h"
#include "MAdVector3D.h"

#include <cstdio>
#include <limits>


namespace MAd {

  // -------------------------------------------------------------------
  // --- describe ------------------------------------------------------
  // -------------------------------------------------------------------

  void T_describe(const pTransform pt,std::ostream& out,bool affine) {
    if (pt) pt->describe(out,affine);
    else out << "Empty transformation";
  }

  // -------------------------------------------------------------------
  
  char* T_describe(const pTransform pt,bool affine) {
    std::ostringstream out;
    T_describe(pt,out,affine);
    return strdup(out.str().c_str());
  }

  // -------------------------------------------------------------------
  // --- get meta information ------------------------------------------
  // -------------------------------------------------------------------
  
  bool T_isIdentity(const pTransform pt) {
    if (!pt) return true;
    return pt->isIdentity();
  }

  // -------------------------------------------------------------------
  
  bool T_isRotation(const pTransform pt,
                    double* axis,
                    double* origin,
                    double& angle) {
    if (!pt) return false;
    return pt->isRotation(axis,origin,angle);
  }

  // -------------------------------------------------------------------

  bool T_isTranslation(const pTransform pt,
                       double* displacement) {
    if (!pt) return false;
    return pt->isTranslation(displacement);
  }

  // -------------------------------------------------------------------
  
  bool T_isScaling(const pTransform pt) {
    if (!pt) return false;
    return pt->isScaling();
  }

  // -------------------------------------------------------------------

  int T_alignedMultiple(const pTransform pt,
                        const double* p1,
                        const double* p2) {
    if (!pt) return std::numeric_limits<int>::max();
    return pt->computeAlignedMultiple(p1,p2);
  }
  
  // -------------------------------------------------------------------
  // --- creation ------------------------------------------------------
  // -------------------------------------------------------------------

  pTransform T_createIdentity() {
    return new MAdTransformation();
  }
  
  // -------------------------------------------------------------------
  
  template<class FLOAT> 
  pTransform template_T_create(const FLOAT* mat,const FLOAT* vec) {
    
    if (mat == NULL && vec == NULL) return NULL;
    pTransform tfo = T_createIdentity();

    double affine[16];
    for (size_t i=0;i<16;i++) affine[i] = 0;
    for (size_t i=0;i<4;i++) affine[i*5] = 1;
    
    if (mat) {
      for (size_t i=0;i<3;i++) {
        for (size_t j=0;j<3;j++) {
          affine[i*4+j] = mat[i*3+j];
        }
      }
    }
    
    if (vec) {
      for (size_t i=0;i<3;i++) {
        affine[i*4+3] = vec[i];
      }
    }

    T_applyAffineTransformation(tfo,affine);
    return tfo;
  }

  // -------------------------------------------------------------------

  pTransform T_create(const float* mat,const float* trans) {
    return template_T_create<float>(mat,trans);
  }
  
  // -------------------------------------------------------------------

  pTransform T_create(const double* mat,const double* trans) {
    return template_T_create<double>(mat,trans);
  }
  
  // -------------------------------------------------------------------

  template<class FLOAT>
  pTransform template_T_create(const FLOAT* center,
                               const FLOAT* angle,
                               const FLOAT* translation) {
    
    if (center == NULL && angle == NULL && translation == NULL) return NULL;
  
    doubleMatrix compoundRotation(3,3);
    compoundRotation.set_identity();
    MAd_Vector<double> dispv(3);

    if (center != NULL && angle != NULL) {
    
      for (int i=0;i<3;i++) {
        
        if (angle[i] != 0) {
          
          doubleMatrix tmp(compoundRotation);
          doubleMatrix rotation(3,3);
          rotation.set_identity();

          compoundRotation = 0;
          
          int ii = (i+1)%3;
          int jj = (i+2)%3;
          
          double ca = cos(angle[i]);
          double sa = sin(angle[i]);
          
          // rotation with alpha
          
          rotation(ii,ii) = ca;
          rotation(ii,jj) = sa;
          rotation(jj,ii) = -sa;
          rotation(jj,jj) = ca;
          
          rotation.mult(tmp,compoundRotation);
          
        }
      }
      
      // compute displacement from rotation center
    
      MAd_Vector<double> centerv(3);
      for (int i=0;i<3;i++) centerv(i) = center[i];
      
      compoundRotation.mult(centerv,dispv,false);
    }
    
    // add specified translation
    if (translation != NULL) {
      for (int i=0;i<3;i++) dispv(i) = - translation[i];
    }
    
    return T_create(compoundRotation.seePtr(),dispv.seePtr());
  }
  
  // -------------------------------------------------------------------

  pTransform T_create(const float* center,
                      const float* angle,
                      const float* translation) {
    return template_T_create<float>(center,angle,translation);
  }
  
  // -------------------------------------------------------------------

  pTransform T_create(const double* center,
                      const double* angle,
                      const double* translation) {
    return template_T_create<double>(center,angle,translation);
  }

  // -------------------------------------------------------------------

  pTransform T_create(const double* src0,
                      const double* src1,
                      const double* src2,
                      const double* tgt0,
                      const double* tgt1,
                      const double* tgt2,
                      double tol) {

    const MAdMsg& msg = MAdMsgSgl::instance();

    MAd_Vector3D dS2T[3] = {MAd_Vector3D(src0,tgt0),
                            MAd_Vector3D(src1,tgt1),
                            MAd_Vector3D(src2,tgt2)};
    
    // --- check for identity transformation ---------------------------
    
    bool check = true;
    
    for (int i=0;i<3;i++) check = check && (dS2T[i].norm() < tol);
    if (check)  return T_createIdentity();

    // --- check for translation ---------------------------------------
    
    double distTol = dS2T[0].norm() * tol;
    check = true;
    for (int i=0;i<3;i++) {
      check &= (MAd_Vector3D(dS2T[i],dS2T[(i+1)%3]).norm() < distTol);
    }
    if (check) return T_create(NULL,dS2T[0].data());
    
    // --- check for generic transformation ----------------------------

    // --- isolate the rotation matrix first

    MAd_Vector3D dSrc[3] = {MAd_Vector3D(src0,src1),
                            MAd_Vector3D(src1,src2),
                            MAd_Vector3D(src2,src0)};
    MAd_Vector3D dTgt[3] = {MAd_Vector3D(tgt0,tgt1),
                            MAd_Vector3D(tgt1,tgt2),
                            MAd_Vector3D(tgt2,tgt0)};

    dSrc[2] = dSrc[0] ^ dSrc[1];
    dTgt[2] = dTgt[0] ^ dTgt[1];
    
    doubleMatrix mDSrc(3,3);
    doubleMatrix mDTgt(3,3);

    for (int i=0;i<3;i++) for (int j=0;j<3;j++) mDSrc(i,j) = dSrc[j][i];
    for (int i=0;i<3;i++) for (int j=0;j<3;j++) mDTgt(i,j) = dTgt[j][i];

    if (!mDSrc.invertInPlace()) {
      msg.warning(__LINE__,__FILE__,
                  "Provided collinear points for computing a transformation,"
                  "returning NULL");
      return NULL;
    }
    
    doubleMatrix T(3,3);
    T.blas_dgemm(mDTgt,mDSrc,1,0);
    
    // --- compute the displacement 
    
    doubleVector xSrc(3);
    std::memcpy(xSrc.touchPtr(),src1,3*sizeof(double));
    
    doubleVector xTgt(3);
    std::memcpy(xTgt.touchPtr(),tgt1,3*sizeof(double));
    
    T.mult(xSrc,xTgt,false,-1,1);

    // --- copy both to the affine 4x4 matrix and create the transformation 

    doubleMatrix affine(4,4);
    affine.set_identity();
    
    for (int i=0;i<3;i++) {
      for (int j=0;j<3;j++) affine(i,j) = T(i,j);
      affine(i,3) = xTgt(i);
    }
    
    pTransform tfo = new MAdTransformation(affine);
    
    // --- now check whether we have the correct transformation
    
    double tfoSrc[3];
    const double* cSrc[3] = {src0,src1,src2};
    const double* cTgt[3] = {tgt0,tgt1,tgt2};
    
    for (int i=0;i<3;i++) {
      std::memcpy(tfoSrc,cSrc[i],3*sizeof(double));
      T_transformCoordinates(tfo,tfoSrc);
      MAd_Vector3D diff(tfoSrc,cTgt[i]);

      if (diff.norm() > tol) {
        msg.warning(__LINE__,__FILE__,
                    "Checking point %i in computed transformation, "
                    "tolerance = %g",i,diff.norm());
        T_delete(tfo);
        return NULL;
      }
    }
    
    return tfo;
  }

  // -------------------------------------------------------------------

  void T_delete(pTransform& pt) {
    if (pt) delete pt;
    pt = NULL;
  }

  // -------------------------------------------------------------------

  pTransform T_duplicate(const pTransform pt) {
    if (!pt) return NULL;
    return new MAdTransformation(*pt);
  }

  // -------------------------------------------------------------------

  pTransform T_inverse(const pTransform pt) {
    if (!pt) return NULL;
    pTransform ppt = T_duplicate(pt);
    ppt->invert();
    return ppt;
  }

  // -------------------------------------------------------------------

  pTransform T_power(const pTransform pt, int n) {
    if (!pt) return NULL;
    pTransform ppt = T_duplicate(pt);
    ppt->raiseToPower(n);
    return ppt;
  }
  
  // -------------------------------------------------------------------
  // --- modification --------------------------------------------------
  // -------------------------------------------------------------------

  bool T_invert(pTransform pt) {
    if (!pt) return false;
    return pt->invert();
  }

  // -------------------------------------------------------------------

  bool T_applyAffineTransformation(pTransform pt,const double* matrix) {
    if (!pt) return false;
    pt->applyAffineTransformation(matrix);
    return true;
  }

  // -------------------------------------------------------------------
  
  bool T_apply(const pTransform t1,pTransform t2) {
    if (t1 && t2) return t2->compose(t1);
    return false;
  }

  // -------------------------------------------------------------------

  bool T_raiseToPower(const pTransform pt, int n) {
    if (!pt) return false;
    return pt->raiseToPower(n);
  }

  // -------------------------------------------------------------------
  
  pTransform T_combine(const pTransform t1,const pTransform t2) {
    if (t1 && t2) {
      doubleMatrix affine(4,4);
      t1->matrix().mult(t2->matrix(),affine);
      return new MAdTransformation(affine);
    }
    if (t1) return T_duplicate(t1);
    if (t2) return T_duplicate(t2);
    return NULL;
  } 
 
  // -------------------------------------------------------------------
  //  --- recuperation in simple format --------------------------------
  // -------------------------------------------------------------------
  
  bool T_copyAffineTransformation(pTransform pt,double* matrix) {
    if (!pt) {
      for (int i=0;i<16;i++) matrix[i] = 0;
      for (int i=0;i<4 ;i++) matrix[i*5] = 1;
      return false;
    }
    pt->copyAffineTransformation(matrix);
    return true;
  }
  
  // -------------------------------------------------------------------
  // --- application ---------------------------------------------------
  // -------------------------------------------------------------------
  
  bool T_transformCoordinates(pTransform pt,double* xyz,const int n) {
    if (!pt) return false;
    pt->transformCoordinates(xyz,n);
    return true;
  }

  // -------------------------------------------------------------------

  bool T_transformBoundingBox(pTransform pt,
                              const double oldBB[3][2],
                              double newBB[3][2]) {
    if (!pt) return false;
    double minXYZ[3] = {oldBB[0][0],oldBB[1][0],oldBB[2][0]};
    double maxXYZ[3] = {oldBB[0][1],oldBB[1][1],oldBB[2][1]};

    T_transformCoordinates(pt,minXYZ);
    T_transformCoordinates(pt,maxXYZ);
    
    for (int d=0;d<3;d++) {
      newBB[d][0] = std::min(minXYZ[d],maxXYZ[d]);
      newBB[d][0] = std::max(minXYZ[d],maxXYZ[d]);
    } 
    return true;
  }
  
  // -------------------------------------------------------------------

  bool T_transformVector(pTransform pt,double* vec,bool inv) {
    if (!pt) return false;
    pt->transformVector(vec,inv);
    return true;
  }
  
  // -------------------------------------------------------------------
  
  bool T_transformTensor2(pTransform pt,double* ten,bool inv) {
    if (!pt) return false;
    pt->transformTensor2(ten,inv);
    return true;
  }
  
  // -------------------------------------------------------------------

  bool T_transformSymTensor2(pTransform pt,double* ten,bool inv) {
    if (!pt) return false;
    pt->transformSymTensor2(ten,inv);
    return true;
  }
  
  // -----------------------------------------------------------------

  bool pTransformLess::operator() (const pTransform a,
                                   const pTransform b) const {
    if (!a) return true;
    if (!b) return false;
    return (*a) < (*b);
  }

  // -------------------------------------------------------------------
  
};


