// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// -------------------------------------------------------------------

#ifndef H__TRANSFORMATIONINTERFACE
#define H__TRANSFORMATIONINTERFACE

#include "MAdExport.h"

#include <iostream>
#include <set>

namespace MAd {

  class MAdTransformation;

  // --- typedef -----------------------------------------------------

  /*! \brief transformation combining rotations and displacements */
  /*! transformation is constructed recursively starting from identity */
  
  typedef class MAdTransformation* pTransform;

  // --- describe ----------------------------------------------------
  
  /*! \brief add description to an output stream */
  /*! if affine checked, also write the matrix */
  MAD_EXPORT void T_describe(const pTransform,std::ostream&,bool affine=false);
  
  /*! \brief add description to a char* string */
  /*! if affine checked, also write the matrix */
  MAD_EXPORT char* T_describe(const pTransform,bool affine=false);

  // --- find basic properties ---------------------------------------

  /*! \brief check whether it is an identity */
  MAD_EXPORT bool T_isIdentity(const pTransform);

  /*! \brief check whether it is a rotation and give axis, center and angle */
  MAD_EXPORT bool T_isRotation(const pTransform,double*,double*,double&);
  
  /*! \brief return whether it is a non-trivial translation (ie. not just a
      non-trivial rotation center) and give the displacement */
  MAD_EXPORT bool T_isTranslation(const pTransform,double*);
 
  /*! \brief return whether it is a coordinate scaling */
  MAD_EXPORT bool T_isScaling(const pTransform);

  /*! \brief find multiple that aligns two points best */
  MAD_EXPORT int T_alignedMultiple(const pTransform,double*,double*);
  
  // --- creation ----------------------------------------------------
  
  /*! \brief create a void transformation */
  MAD_EXPORT pTransform T_createIdentity();

  /*! \brief create based upon rotation matrix and translation */
  MAD_EXPORT pTransform T_create(const float* rotationMatrix,
                      const float* translation);
  
  /*! \brief create based upon rotation matrix and translation */
  MAD_EXPORT pTransform T_create(const double* rotationMatrix,
                      const double* translation);

  /*! \brief create based upon rotation center, rotation angles and displacement */
  MAD_EXPORT pTransform T_create(const float* center,
                      const float* angle,
                      const float* translation);

  /*! \brief compute the transformation between point triplets \f$ \mathbf s_i \f$ and \f$ \mathbf t_i \f$ */
  /*! These triplets are specified in order of correspondence, ie. s0 corresponds to t0 etc. */
  /*! The following transformations will be tested in order
    - identity transformation if \f$ ||\mathbf s_i - \mathbf t_i|| < tol \forall i \f$
    - translation if \f$ || \Delta \mathbf s_{ij} - \Delta \mathbf t_{ij} || < tol \forall i,j \f$
    - a general affine transformation. 
    with \f$\Delta \mathbf x_{ij} = \mathbf x_i - \mathbf x_j \f$
  */
  MAD_EXPORT pTransform T_create(const double* s0, //!< source point 0
                                 const double* s1, //!< source point 1
                                 const double* s2, //!< source point 2
                                 const double* t0, //!< target point 0
                                 const double* t1, //!< target point 1
                                 const double* t2, //!< target point 2
                                 double tol=1e-12  //!< tolerance for the transformation
                                 );
  
  /*! \brief create based upon rotation center, rotation angles and displacement */
  MAD_EXPORT pTransform T_create(const double* center,
                                 const double* angle,
                                 const double* translation);

  /*! \brief delete a transformation */
  MAD_EXPORT void T_delete(pTransform&);
  
  /*! \brief provide an identical copy of a transformation */
  MAD_EXPORT pTransform T_duplicate(const pTransform);

  /*! \brief create inverse transformation */
  MAD_EXPORT pTransform T_inverse(const pTransform);

  /*! \brief chain */
  MAD_EXPORT pTransform T_combine(const pTransform,const pTransform);

  /*! \brief repeat n times */
  MAD_EXPORT pTransform T_power(const pTransform, int);

  // --- modification ------------------------------------------------

  /*! \brief invert the transformation */
  MAD_EXPORT bool T_invert(pTransform);

  /*! \brief apply a affine transformation */
  MAD_EXPORT bool T_applyAffineTransformation(pTransform,   //!< transformation
                                   const double* //!< 4x4 row-major
                                   );

  /*! \brief compose transformations \f$ T_2 \rightarrow T_1 o T_2 \f$ */
  MAD_EXPORT bool T_apply(const pTransform t1,  //!< first transformation
                     pTransform t2   //!< second transformation
               );

  // --- application -------------------------------------------------

  /*! \brief transformation of coordinates (in place) */
  MAD_EXPORT bool T_transformCoordinates(const pTransform pt, //!< transformation
                              double* coord,       //!< transformed coordinates
                              const int n=1);      //!< number of applications
  
  /*! \brief transformation of bounding box (out of place) */
  /*! \warning does not exactly correspond to bounding box computed on the transformed entity */
  MAD_EXPORT bool T_transformBoundingBox(const pTransform pt,      //!< transformation
                              const double oldBB[3][2], //! original bounding box
                              double newBB[3][2]);      //!< transformed bounding box

  /*! \brief transformation of a tensor of first order (in place) */
  MAD_EXPORT bool T_transformVector(const pTransform pt, //!< transformation
                         double* vec,         //!< transformed vector
                         bool inverse=false   //!< apply inverse
                         );
  
  /*! \brief transform a tensor of second order */
  /*! Stored in row major ordering, ie. \f$ T_{ij} = *(pT + 3*i + j) \f$ */
  MAD_EXPORT bool T_transformTensor2(const pTransform pt,  //!< transformation
                          double* tensor,       //!< transformed tensor
                          bool inverse=false    //!< apply inverse
                          );

  /*! \brief transform a tensor of second order */
  /*! Stored in row major ordering, ie. \f$ T_{ij} = *(pT + 3*i + j) \f$ */
  MAD_EXPORT bool T_transformSymTensor2(const pTransform pt,  //!< transformation
                             double* tensor,       //!< transformed tensor
                             bool inverse=false    //!< apply inverse
                             );

  /*! \brief compose transformations \f$ T_2 \rightarrow T_1 o T_2 \f$ */

  MAD_EXPORT bool T_apply(const pTransform T2, //!< transformation to be applied
               pTransform T1        //!< 
               );

  /*! \brief raise to nth power */

  MAD_EXPORT bool T_raiseToPower(const pTransform pt,
                      int n);

  // --- elementary transformation transcription ---------------------
  
  /*! \brief get the affine transformation matrix */
  /*! this matrix has size 4x4
    - the first 3 columns provide the rotation matrix 
    - the last    column provides the translation vector 
  */
  MAD_EXPORT bool T_copyAffineTransformation(const pTransform,double*);

  // -----------------------------------------------------------------

  struct MAD_EXPORT pTransformLess {
    bool operator() (const pTransform a,const pTransform b) const;
  };
  
  typedef std::set<pTransform,pTransformLess> TransformationSet;
  
  
};

#endif
