// -*- C++ -*- 
// -----------------------------------------------------------------------------
// MAdLiB - Copyright (C) 2008-2009 Université catholique de Louvain 
// -----------------------------------------------------------------------------
//
// See the Copyright.txt and License.txt files for license information. 
// ------------------------------------------------------------------


//-------------------------------
//  Lagrange Tetrahedrons
//-------------------------------

// from DGBase
#include "MAd_TetInterpolant.h"
#include "MAd_TriInterpolant.h"

#include <iostream>
using std::cout;
using std::endl;

// -----------------------------------------------------------------------------
//  Publics Members
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
/*! \brief Equidistant Lagrange function space for the tetrahedron */

MAd_TetInterpolant::MAd_TetInterpolant(const int &o)
  : MAd_Interpolant(MSH_TET_4, o), ndof(size(o))
{
  coef.resize(int(ndof),int(ndof));
  for (int i=0; i< int(ndof); i++){
    for (int j=0; j< int(ndof); j++)
      switch(order()) {
      case 0: coef(i,j) = coef0[i][j]; break;
      case 1: coef(i,j) = coef1[i][j]; break;
      case 2: coef(i,j) = coef2[i][j]; break;
      case 3: coef(i,j) = coef3[i][j]; break;
      case 4: coef(i,j) = coef4[i][j]; break;
      case 5: coef(i,j) = coef5[i][j]; break;
      case 6: coef(i,j) = coef6[i][j]; break;
      case 7: coef(i,j) = coef7[i][j]; break;
      case 8: coef(i,j) = coef8[i][j]; break;
      case 9: coef(i,j) = coef9[i][j]; break;
      case 10: coef(i,j) = coef10[i][j]; break;
      default : 
        cout << "Lagrange Shape Function on Tet only defined up to order 10" << endl; 
        throw; break;
      } 
  }

  pp = P;
  
  NodeNumberu = new double[ndof];
  NodeNumberv = new double[ndof];
  NodeNumberw = new double[ndof];
  nodeposition(order(), NodeNumberu, NodeNumberv, NodeNumberw);

  return;
}

// -----------------------------------------------------------------------------
MAd_TetInterpolant::~MAd_TetInterpolant(){
  
  delete[] NodeNumberu;
  delete[] NodeNumberv;
  delete[] NodeNumberw;
  return;
}

// -----------------------------------------------------------------------------
void MAd_TetInterpolant::fcts(const double &u, const double &v, const double &w, 
				    doubleVector &f) const {
  doubleVector m(ndof);
  f = 0.;
  for (int i=0; i< ndof ; i++) m(i) = monomeof(u,v,w,i);
  coef.mult(m,f,false);
  return;
}

// -----------------------------------------------------------------------------
void MAd_TetInterpolant::grads(const double &u, const double &v, const double &w, 
				     doubleVector gr[3]) const {
  doubleVector m(ndof);
  for (int k=0; k< 3; k++){ 
    gr[k] = 0.;
    for (int i=0; i< ndof ; i++) m(i) = dmonomeof(u,v,w,i,k);
    coef.mult(m,gr[k],false);
  }
  return;
}

// -----------------------------------------------------------------------------
void MAd_TetInterpolant::getMonomialExponents (int iMonomial, int &powerOfU, int &powerOfV, int &powerOfW) const {
  powerOfU = P[iMonomial][0];
  powerOfV = P[iMonomial][1];
  powerOfW = P[iMonomial][2];
}

// -----------------------------------------------------------------------------
MAd_Interpolant * MAd_TetInterpolant::clone(int o) const {
  return new MAd_TetInterpolant(o);
}

// -----------------------------------------------------------------------------
void MAd_TetInterpolant::nodeNumberCoordinate(const int & i, double &u, double &v, double &w) const {

  u= NodeNumberu[i];
  v= NodeNumberv[i];  
  w= NodeNumberw[i];
  return;
}
   
// -----------------------------------------------------------------------------
int MAd_TetInterpolant::size(int q) {
  return (q+1)*(q+2)*(q+3)/6;
}

// -----------------------------------------------------------------------------
//  Private Members
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
int MAd_TetInterpolant::nbdoftriangle(int o) const{
  return MAd_TriInterpolant::size(o);
}

// -----------------------------------------------------------------------------
void MAd_TetInterpolant::nodepositionface0(int o,  double *u,  double *v,  double *w){
   int ndofT = nbdoftriangle(o);
   if (o ==0) {u[0]=0.;v[0]=0.; return;}     
   
   u[0]= 0.;    v[0]= 0.;  w[0] = 0.;
   u[1]= o; v[1]= 0;   w[1] = 0.;
   u[2]= 0.;    v[2]= o; w[2] = 0.;
   // edges
   for (int k=0; k< (o-1); k++){
     u[3+k] = k+1; v[3+k]=0.; w[3+k] = 0.;
     u[3+o-1+k] = o -1 -k ; v [3+o-1+k] = k+1; w[3+o-1+k] = 0.;
     u[3+2*(o-1)+k] = 0. ; v [3+2*(o-1)+k] = o -1 -k; w [3+2*(o-1)+k] = 0.;
   }
   if (o >2){
     int nbdoftemp = nbdoftriangle(o-3);
     nodepositionface0(o-3,&u[3+3*(o-1)], &v[3+3*(o-1)], &w[3+3*(o-1)]);
      for (int k=0; k< nbdoftemp; k++){
     	u[3+k+3*(o-1)] = u[3+k+3*(o-1)]*(o-3)+1.;
        v[3+k+3*(o-1)] = v[3+k+3*(o-1)]*(o-3)+1.;
	w[3+k+3*(o-1)] = w[3+k+3*(o-1)]*(o-3);
      }
   }
   for (int k=0; k< ndofT; k++){
     u[k] = u[k]/o;
     v[k] = v[k]/o;
     w[k] = w[k]/o;
   }
   return;     
}


// -----------------------------------------------------------------------------
void MAd_TetInterpolant::nodepositionface1(int o,  double *u,  double *v,  double *w){
   int ndofT = nbdoftriangle(o);
   if (o ==0) {u[0]=0.;v[0]=0.; return;}     
   
   u[0]= 0.;    v[0]= 0.;  w[0] = 0.;
   u[1]= o; v[1]= 0;   w[1] = 0.;
   u[2]= 0.;    v[2]= 0.;  w[2] = o;
   // edges
   for (int k=0; k< (o-1); k++){
     u[3+k] = k+1; v[3+k]=0.; w[3+k] = 0.;
     u[3+o-1+k] = o -1 -k; v[3+o-1+k] = 0.;     w[3+o-1+k] = k+1; 
     u[3+2*(o-1)+k] = 0. ;     v[3+2*(o-1)+k] = 0.; w[3+2*(o-1)+k] = o -1 -k; 
   }
   if (o >2){
     int nbdoftemp = nbdoftriangle(o-3);
     nodepositionface1(o-3,&u[3+3*(o-1)], &v[3+3*(o-1)], &w[3+3*(o-1)]);
      for (int k=0; k< nbdoftemp; k++){
	u[3+k+3*(o-1)] = u[3+k+3*(o-1)]*(o-3)+1.;
        v[3+k+3*(o-1)] = v[3+k+3*(o-1)]*(o-3);
	w[3+k+3*(o-1)] = w[3+k+3*(o-1)]*(o-3)+1.;
      }
   }
   for (int k=0; k< ndofT; k++){
     u[k] = u[k]/o;
     v[k] = v[k]/o;
     w[k] = w[k]/o;
   }
   return;
}

// -----------------------------------------------------------------------------
void MAd_TetInterpolant::nodepositionface2(int o,  double *u,  double *v,  double *w){
   int ndofT = nbdoftriangle(o);
   if (o ==0) {u[0]=0.;v[0]=0.; return;}     
   
   u[0]= o;    v[0]= 0.;  w[0] = 0.;
   u[1]= 0.; v[1]= o;   w[1] = 0.;
   u[2]= 0.;    v[2]= 0.;   w[2] = o;
   // edges
   for (int k=0; k< (o-1); k++){
     u[3+k] = o -1 -k;     v[3+k]= k+1;                  w[3+k] = 0.;
     u[3+o-1+k] = 0.;      v[3+o-1+k] = o -1 -k; w[3+o-1+k] = k+1; 
     u[3+2*(o-1)+k] = k+1; v[3+2*(o-1)+k] = 0.;      w[3+2*(o-1)+k] = o -1 -k; 
   }
   if (o >2){
     int nbdoftemp = nbdoftriangle(o-3);
     nodepositionface2(o-3,&u[3+3*(o-1)], &v[3+3*(o-1)], &w[3+3*(o-1)]);
      for (int k=0; k< nbdoftemp; k++){
        u[3+k+3*(o-1)] = u[3+k+3*(o-1)]*(o-3)+1.;
        v[3+k+3*(o-1)] = v[3+k+3*(o-1)]*(o-3)+1.;
	w[3+k+3*(o-1)] = w[3+k+3*(o-1)]*(o-3)+1.;
      }
   }
   for (int k=0; k< ndofT; k++){
     u[k] = u[k]/o;
     v[k] = v[k]/o;
     w[k] = w[k]/o;
   }
   return;
}

// -----------------------------------------------------------------------------
void MAd_TetInterpolant::nodepositionface3(int o,  double *u,  double *v,  double *w){
   int ndofT = nbdoftriangle(o);
   if (o ==0) {u[0]=0.;v[0]=0.; return;}     
   
   u[0]= 0.;    v[0]= 0.;  w[0] = 0.;
   u[1]= 0.;    v[1]= o;   w[1] = 0.;
   u[2]= 0.;    v[2]= 0.;   w[2] = o;
   // edges
   for (int k=0; k< (o-1); k++){
     u[3+k] = 0.;             v[3+k]= k+1;                  w[3+k] = 0.;
     u[3+o-1+k] = 0.;     v[3+o-1+k] = o -1 -k; w[3+o-1+k] = k+1; 
     u[3+2*(o-1)+k] = 0.; v[3+2*(o-1)+k] = 0.;      w[3+2*(o-1)+k] = o -1 -k; 
   }
   if (o >2){
     int nbdoftemp = nbdoftriangle(o-3);
     nodepositionface3(o-3,&u[3+3*(o-1)], &v[3+3*(o-1)], &w[3+3*(o-1)]);
      for (int k=0; k< nbdoftemp; k++){
	u[3+k+3*(o-1)] = u[3+k+3*(o-1)]*(o-3);
        v[3+k+3*(o-1)] = v[3+k+3*(o-1)]*(o-3)+1.;
	w[3+k+3*(o-1)] = w[3+k+3*(o-1)]*(o-3)+1.;
      }
   }
   for (int k=0; k< ndofT; k++){
     u[k] = u[k]/o;
     v[k] = v[k]/o;
     w[k] = w[k]/o;
   }
   return;
}

// -----------------------------------------------------------------------------
void MAd_TetInterpolant::nodeposition (int o,  double *u,  double *v,  double *w){
      //nodes
      if (o ==0) {u[0]=0.;v[0]=0.; w[0]=0.; return;}
      u[0]= 0.;
      u[1]= o;
      u[2]= 0.;
      u[3]= 0.;
      
      v[0]= 0.;
      v[1]= 0;
      v[2]= o;
      v[3]= 0.;
      
      w[0]= 0.;
      w[1]= 0;
      w[2]= 0.;
      w[3]= o;
      //edges
      for (int k=0; k< (o-1); k++) {
      	u[4+k] = k+1;
	u[4+o-1+k]  = o-1-k;
	u[4+2*(o-1)+k]  = 0.; 
	u[4+3*(o-1)+k]  = 0.;
	u[4+4*(o-1)+k]  = o-1-k;
	u[4+5*(o-1)+k]  = 0.;
	
	v[4+k] = 0.;
	v[4+o-1+k]      = k+1;
	v[4+2*(o-1)+k]  = o-1-k; 
	v[4+3*(o-1)+k]  = 0.;
	v[4+4*(o-1)+k]  = 0.;
	v[4+5*(o-1)+k]  = o-1-k;
	
	w[4+k] = 0.;
	w[4+o-1+k]      = 0.;
	w[4+2*(o-1)+k]  = 0.; 
	w[4+3*(o-1)+k]  = k+1;
	w[4+4*(o-1)+k]  = k+1;
	w[4+5*(o-1)+k]  = k+1; 
      }
      int ns = 4+6*(o-1);
   if (o >2) {  
     int nbdofface = nbdoftriangle(o-3);
     nodepositionface0(o-3, &u[ns], &v[ns], &w[ns]);
     for (int i=0; i < nbdofface; i++){
       u[ns+i] = u[ns+i]*(o-3) + 1.;
       v[ns+i] = v[ns+i]*(o-3) + 1.;
       w[ns+i] = w[ns+i]*(o-3);
     }
     ns = ns + nbdofface;
     nodepositionface1(o-3, &u[ns], &v[ns], &w[ns]);

     for (int i=0; i < nbdofface; i++){
       u[ns+i] = u[ns+i]*(o-3) + 1.;
       v[ns+i] = v[ns+i]*(o-3) ;
       w[ns+i] = w[ns+i]*(o-3) + 1.;
     }
     ns = ns + nbdofface;
     nodepositionface2(o-3, &u[ns], &v[ns], &w[ns]);

     for (int i=0; i < nbdofface; i++){
       u[ns+i] = u[ns+i]*(o-3) + 1.;
       v[ns+i] = v[ns+i]*(o-3) + 1.;
       w[ns+i] = w[ns+i]*(o-3) + 1.;
     }
     ns =ns +nbdofface;
      nodepositionface3(o-3, &u[ns], &v[ns], &w[ns]);
     for (int i=0; i < nbdofface; i++){
       u[ns+i] = u[ns+i]*(o-3);
       v[ns+i] = v[ns+i]*(o-3) + 1.;
       w[ns+i] = w[ns+i]*(o-3) + 1.;
     }
     ns =ns +nbdofface;
   }
   if (o  > 4 ){
      int ndofTT = size(o -4);
      nodeposition (o-4, &u[ns], &v[ns], &w[ns]);
      for (int k=0; k< ndofTT; k++) {
        u[ns+k] = 1.+u[ns+k]*(o-4);
	v[ns+k] = 1.+v[ns+k]*(o-4);
	w[ns+k] = 1.+w[ns+k]*(o-4);
      }
      }
      if (o == 4) {u[ns] = 1.; v[ns]=1.; w[ns]=1.; }
      
   int ndofTT = size(o);
   for (int i=0; i< ndofTT; i++){
     u[i] = u[i]/o;
     v[i] = v[i]/o;
     w[i] = w[i]/o;
   }   	
  return;  
 }

   
// -----------------------------------------------------------------------------
 int MAd_TetInterpolant::P[][3] = {    
 //order 0 :    
    {0,0,0},      
 //order 1:   
    {1,0,0},	
    {0,1,0},
    {0,0,1},
 // order 2:
    {2,0,0},	
    {0,2,0},
    {0,0,2},
    {1,1,0},
    {0,1,1},
    {1,0,1},
 // order 3:
    {3,0,0},
    {0,3,0},
    {0,0,3},
    {2,1,0},
    {2,0,1},
    {1,2,0},
    {0,2,1},
    {1,0,2},
    {0,1,2},
    {1,1,1},
 //order 4:
    {4,0,0},
    {0,4,0},
    {0,0,4},
    {3,1,0},
    {3,0,1},
    {1,3,0},
    {0,3,1},
    {1,0,3},
    {0,1,3},
    {2,2,0},
    {2,0,2},
    {0,2,2},
    {2,1,1},
    {1,2,1},
    {1,1,2},
 // order 5:
    {5,0,0},
    {0,5,0},
    {0,0,5},
    {4,1,0},
    {4,0,1},
    {1,4,0},
    {0,4,1},
    {1,0,4},
    {0,1,4},
    {3,2,0},
    {3,0,2},
    {2,3,0},
    {0,3,2},
    {2,0,3},
    {0,2,3},
    {3,1,1},
    {1,3,1},
    {1,1,3},
    {2,2,1},
    {2,1,2},
    {1,2,2},
 //order 6
    {6,0,0},
    {0,6,0},
    {0,0,6},
    {5,1,0},
    {5,0,1},
    {1,5,0},
    {0,5,1},
    {1,0,5},
    {0,1,5},
    {4,2,0},
    {4,0,2},
    {2,4,0},
    {0,4,2},
    {2,0,4},
    {0,2,4},
    {4,1,1},
    {1,4,1},
    {1,1,4},
    {3,3,0},
    {3,0,3},
    {0,3,3},
    {3,2,1},
    {3,1,2},
    {2,3,1},
    {1,3,2},
    {2,1,3},
    {1,2,3},
    {2,2,2},
  //order 7
    {7,0,0},
    {0,7,0},
    {0,0,7},
    {6,1,0},
    {6,0,1},
    {1,6,0},
    {0,6,1},
    {1,0,6},
    {0,1,6},
    {5,2,0},
    {5,0,2},
    {2,5,0},
    {0,5,2},
    {2,0,5},
    {0,2,5},
    {5,1,1},
    {1,5,1},
    {1,1,5},
    {4,3,0},
    {4,0,3},
    {3,4,0},
    {0,4,3},    
    {3,0,4},
    {0,3,4},
    {4,2,1},
    {4,1,2},
    {2,4,1},
    {1,4,2},
    {2,1,4},
    {1,2,4},
    {3,3,1},
    {3,1,3},
    {1,3,3},
    {3,2,2},
    {2,3,2},
    {2,2,3},
  //order 8
    {8,0,0},
    {0,8,0},
    {0,0,8},
    {7,1,0},
    {7,0,1},
    {1,7,0},
    {0,7,1},
    {1,0,7},
    {0,1,7},
    {6,2,0},
    {6,0,2},
    {2,6,0},
    {0,6,2},
    {2,0,6},
    {0,2,6},
    {6,1,1},
    {1,6,1},
    {1,1,6},
    {5,3,0},
    {5,0,3},
    {3,5,0},
    {0,5,3},    
    {3,0,5},
    {0,3,5},
    {5,2,1},
    {5,1,2},
    {2,5,1},
    {1,5,2},
    {2,1,5},
    {1,2,5},
    {4,4,0},
    {4,0,4},
    {0,4,4},
    {4,3,1},
    {4,1,3},
    {3,4,1},
    {1,4,3},
    {3,1,4},
    {1,3,4},
    {4,2,2},
    {2,4,2},
    {2,2,4},
    {3,3,2},
    {3,2,3},
    {2,3,3},
 //order 9
    {9,0,0},
    {0,9,0},
    {0,0,9},
    {8,1,0},
    {8,0,1},
    {1,8,0},
    {0,8,1},
    {1,0,8},
    {0,1,8},
    {7,2,0},
    {7,0,2},
    {2,7,0},
    {0,7,2},
    {2,0,7},
    {0,2,7},
    {7,1,1},
    {1,7,1},
    {1,1,7},
    {6,3,0},
    {6,0,3},
    {3,6,0},
    {0,6,3},    
    {3,0,6},
    {0,3,6},
    {6,2,1},
    {6,1,2},
    {2,6,1},
    {1,6,2},
    {2,1,6},
    {1,2,6},
    {5,4,0},   
    {5,0,4},
    {4,5,0},   
    {0,5,4},
    {4,0,5},
    {0,4,5},
    {5,3,1},
    {5,1,3},
    {3,5,1},
    {1,5,3},
    {3,1,5},
    {1,3,5},
    {5,2,2},
    {2,5,2},
    {2,2,5},
    {4,4,1},
    {4,1,4},
    {1,4,4},
    {4,3,2},
    {4,2,3},
    {3,4,2},
    {2,4,3},
    {3,2,4},
    {2,3,4},
    {3,3,3},
 //order 10
    {10,0,0},
    {0,10,0},
    {0,0,10},
    {9,1,0},
    {9,0,1},
    {1,9,0},
    {0,9,1},
    {1,0,9},
    {0,1,9},
    {8,2,0},
    {8,0,2},
    {2,8,0},
    {0,8,2},
    {2,0,8},
    {0,2,8},
    {8,1,1},
    {1,8,1},
    {1,1,8},
    {7,3,0},
    {7,0,3},
    {3,7,0},
    {0,7,3},    
    {3,0,7},
    {0,3,7},
    {7,2,1},
    {7,1,2},
    {2,7,1},
    {1,7,2},
    {2,1,7},
    {1,2,7},
    {6,4,0},   
    {6,0,4},
    {4,6,0},   
    {0,6,4},
    {4,0,6},
    {0,4,6},
    {6,3,1},
    {6,1,3},
    {3,6,1},
    {1,6,3},
    {3,1,6},
    {1,3,6},
    {6,2,2},
    {2,6,2},
    {2,2,6},
    {5,5,0},
    {5,0,5},
    {0,5,5},
    {5,4,1},
    {5,1,4},
    {4,5,1},
    {1,5,4},
    {4,1,5},
    {1,4,5},
    {5,3,2},
    {5,2,3},
    {3,5,2},
    {2,5,3},
    {3,2,5},
    {2,3,5},
    {4,4,2},
    {2,4,4},
    {2,4,4},
    {4,3,3},
    {3,4,3},
    {3,3,4}
    };
    
// -----------------------------------------------------------------------------
double MAd_TetInterpolant::monomeof(const double &u, const double &v, const double &w, int i) const{
  return pow(u,pp[i][0])*pow(v,pp[i][1])*pow(w,pp[i][2]);
}

// -----------------------------------------------------------------------------
double MAd_TetInterpolant::dmonomeof(const double &u, const double &v, const double &w, int i, int k) const{

  switch (k){
  case 0:
    if (pp[i][0] > 0)
      return pp[i][0] * pow(u,pp[i][0]-1)*pow(v,pp[i][1])*pow(w,pp[i][2]);
    else return 0.;
    break;
  case 1:    
    if (pp[i][1] > 0)
      return pp[i][1] * pow(u,pp[i][0])*pow(v,pp[i][1]-1)*pow(w,pp[i][2]);
    else return 0.;  
    break;
  case 2:
    if (pp[i][2] > 0)
     return pp[i][2] * pow(u,pp[i][0])*pow(v,pp[i][1])*pow(w,pp[i][2]-1);
    else return 0.;  
    break; 
  default:
    cout <<  "must have 0< k < 2 in dmonomeof " << endl;
    throw;
    break;
  }
}


// -----------------------------------------------------------------------------
//Tetrahedron
// order 1
//     2---3 Face 1 : 0 1 2 {{0,1,2},{0,1,3},{1,2,3},{0,2,3}}
//    /|\  | Face 2 : 0 1 3
//   / | \ | Face 3 : 1 2 3
//  /  |  \| Face 4 : 0 2 3
// 3---0---1 
//     |  /
//     | /
//     |/
//     3

// order 2
//     2-9-3
//    /|\  |
//   9 6 5 8
//  /  |  \|
// 3-7-0-4-1
//     |  /
//     7 8 
//     |/  
//     3         

// order 5
//         Surface DOF                 Volume DOF
//                  
//                   2-24-25-26-27--3              52  (1/5,1/5,1/5)
//                 / |\             |              53  (2/5,1/5,1/5)
//                /  | \            |              54  (1/5,2/5,1/5)
//              24  12 11 41 44 42  23             55  (1/5,1/5,2/5)
//              /    |   \          |
//             /     |    \         |
//           25  48 13 30 10 43 45  22
//           /       |      \       | 
//          /        |       \      |
//        26  50 49 14 33 32  9 40  21
//        /          |         \    |
//       /           |          \   |
//     27  47 51 46 15 28 31 29  8  20
//     /             |            \ |
//    /              |             \|
//   3--19-18-17-16--0--4--5--6--7--1
//                   |             /
//                   |            /
//                  16 34 37 35 20
//                   |          /
//                   |         /
//                  17 39 38 21 
//                   |       /
//                   |      / 
//                  18 36 22   
//                   |    /
//                   |   /
//                  19 23
//                   | /
//                   |/
//                   3


//  order 6
//                  
//                      2--29-30-31-32-33-03                
//                    / |\                |              Internal DOF :
//                   /  | \               |                            
//                 29  14  13 55 59 60 56 28
//                 /    |   \             |              74  (1/6,1/6,1/6)
//                /     |    \            |              75  (1/6,3/6,1/6)
//              30  65 15 36  12 58 63 61 27             76  (3/6,1/6,1/6)
//              /       |      \          |              77  (1/6,1/6,3/6) 
//             /        |       \         |              78  (2/6,1/6,1/6)
//           31  69 68 16 41 40  11 57 62 26             79  (2/6,2/6,1/6)
//           /          |         \       |              80  (1/6,2/6,1/6)
//          /           |          \      |              81  (1/6,1/6,2/6)
//        32  70 73 67 17 42 43 39  10 54 25             82  (2/6,1/6,2/6)
//        /             |            \    |              83  (1/6,2/6,2/6)
//       /              |             \   |              
//     33  66 71 72 64 18 34 37 38 35   9 24             
//     /                |               \ |              
//    /                 |                \|
//   3--23-22-21-20-19--0--4--5--6--7--8--1
//                      |                /
//                      |               /
//                     19 44 47 48 45  24
//                      |             /
//                      |            /
//                     20 52 53 49  25 
//                      |          /
//                      |         / 
//                     21 51 50  26   
//                      |       /
//                      |      /
//                     22 46  27
//                      |    /
//                      |   / 
//                     23  28
//                      | /
//                      |/
//                      3
//
// -----------------------------------------------------------------------------

double MAd_TetInterpolant::coef0[1][1]   = {{1}};

// -----------------------------------------------------------------------------
double MAd_TetInterpolant::coef1[4][4]   = {{1,-1,-1,-1},
                                                  {0,1,0,0},
                                                  {0,0,1,0},
                                                  {0,0,0,1}};
                        
// -----------------------------------------------------------------------------
double MAd_TetInterpolant::coef2[10][10] = {{1,-3,-3,-3,2,2,2,4,4,4},
                                                  {0,-1,0,0,2,0,0,0,0,0},
                                                  {0,0,-1,0,0,2,0,0,0,0},
                                                  {0,0,0,-1,0,0,2,0,0,0},
                                                  {0,4,0,0,-4,0,0,-4,0,-4},
                                                  {0,0,0,0,0,0,0,4,0,0},
                                                  {0,0,4,0,0,-4,0,-4,-4,0},
                                                  {0,0,0,4,0,0,-4,0,-4,-4},
                                                  {0,0,0,0,0,0,0,0,0,4},
                                                  {0,0,0,0,0,0,0,0,4,0}};
                        
// -----------------------------------------------------------------------------
double MAd_TetInterpolant::coef3[20][20] = {{1, -5.5, -5.5, -5.5, 9., 9, 9., 18, 18, 18, -4.5, -4.5, -4.5, -13.5, -13.5, -13.5, -13.5, -13.5, -13.5, -27},
                                                  {0, 1., 0, 0, -4.5, 0, 0, 0, 0, 0, 4.5, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 1., 0, 0, -4.5, 0, 0, 0, 0, 0, 4.5, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 1., 0, 0, -4.5, 0, 0, 0, 0, 0, 4.5, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 9, 0, 0, -22.5, 0, 0, -22.5, 0, -22.5, 13.5, 0, 0, 27, 27, 13.5, 0, 13.5, 0, 27},
                                                  {0, -4.5, 0, 0, 18, 0, 0, 4.5, 0, 4.5, -13.5, 0, 0, -13.5, -13.5, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, -4.5, 0, 0, 0, 0, 0, 13.5, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, -4.5, 0, 0, 0, 0, 0, 0, 0, 13.5, 0, 0, 0, 0},
                                                  {0, 0, -4.5, 0, 0, 18, 0, 4.5, 4.5, 0, 0, -13.5, 0, 0, 0, -13.5, -13.5, 0, 0, 0},
                                                  {0, 0, 9, 0, 0, -22.5, 0, -22.5, -22.5, 0, 0, 13.5, 0, 13.5, 0, 27, 27, 0, 13.5, 27},
                                                  {0, 0, 0, 9, 0, 0, -22.5, 0, -22.5, -22.5, 0, 0, 13.5, 0, 13.5, 0, 13.5, 27, 27, 27},
                                                  {0, 0, 0, -4.5, 0, 0, 18, 0, 4.5, 4.5, 0, 0, -13.5, 0, 0, 0, 0, -13.5, -13.5, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, -4.5, 0, 0, 0, 0, 13.5, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, -4.5, 0, 0, 0, 0, 0, 0, 0, 13.5, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, -4.5, 0, 0, 0, 0, 0, 0, 0, 13.5, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, -4.5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 13.5, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 27, 0, 0, 0, 0, 0, -27, 0, -27, 0, 0, 0, -27},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 27, 0, 0, 0, 0, -27, 0, 0, -27, 0, -27},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 27},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 27, 0, 0, 0, 0, 0, 0, 0, -27, 0, -27, -27}};

// -----------------------------------------------------------------------------
double MAd_TetInterpolant::coef4[35][35] = {{1, -8.33333333333333, -8.33333333333327, -8.33333333333333, 23.3333333333333, 23.3333333333329, 23.3333333333333, 46.6666666666666, 46.6666666666667, 46.6666666666667, -26.6666666666667, -26.666666666666, -26.6666666666667, -80, -80, -79.9999999999996, -80, -80, -80.0000000000001, -160, 10.6666666666667, 10.6666666666663, 10.6666666666667, 42.6666666666667, 42.6666666666667, 42.6666666666663, 42.6666666666666, 42.6666666666667, 42.6666666666667, 63.9999999999999, 64, 64, 128, 128, 128}, 
                                                  {0, -1, 0, 0, 7.33333333333333, 0, 0, 0, 0, 0, -16, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10.6666666666667, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                  {0, 0, -1, 0, 0, 7.33333333333335, 0, 0, 0, 0, 0, -16, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10.6666666666667, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                  {0, 0, 0, -1, 0, 0, 7.33333333333333, 0, 0, 0, 0, 0, -16, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10.6666666666667, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                  {0, 16, 0, 0, -69.3333333333334, 0, 0, -69.3333333333332, 0, -69.3333333333334, 96.0000000000001, 0, 0, 192, 192, 95.9999999999992, 0, 96, 0, 192, -42.6666666666667, 0, 0, -128, -128, -42.666666666666, 0, -42.6666666666667, 0, -128, -128, 0, -256, -128, -128}, 
                                                  {0, -12, 0, 0, 76, 0, 0, 27.9999999999999, 0, 28, -128, 0, 0, -144, -144, -15.9999999999997, 0, -16, 0, -31.9999999999998, 64, 0, 0, 128, 128, 0, 0, 0, 0, 63.9999999999997, 64, 0, 128, 0, 0}, 
                                                  {0, 5.33333333333333, 0, 0, -37.3333333333333, 0, 0, -5.33333333333332, 0, -5.33333333333331, 74.6666666666667, 0, 0, 31.9999999999999, 31.9999999999999, 0, 0, 0, 0, 0, -42.6666666666667, 0, 0, -42.6666666666666, -42.6666666666666, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, 0, 0, 5.33333333333335, 0, 0, 0, 0, 0, -32, 0, 0, 0, 0, 0, 0, 0, 0, 0, 42.6666666666667, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, 0, 0, 3.99999999999995, 0, 0, 0, 0, 0, -16, 0, -15.9999999999999, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 63.9999999999999, 0, 0, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, 0, 0, 5.33333333333331, 0, 0, 0, 0, 0, 0, 0, -31.9999999999998, 0, 0, 0, 0, 0, 0, 0, 0, 0, 42.6666666666666, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                  {0, 0, 5.33333333333333, 0, 0, -37.3333333333333, 0, -5.33333333333342, -5.33333333333334, 0, 0, 74.6666666666667, 0, 0, 0, 32.0000000000002, 32, 0, 0, 0, 0, -42.6666666666667, 0, 0, 0, -42.6666666666668, -42.6666666666667, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                  {0, 0, -12, 0, 0, 75.9999999999997, 0, 28.0000000000001, 28, 0, 0, -128, 0, -16.0000000000003, 0, -144, -144, 0, -16, -32, 0, 63.9999999999998, 0, 0, 0, 128, 128, 0, 0, 64.0000000000002, 0, 64, 0, 128, 0}, 
                                                  {0, 0, 15.9999999999999, 0, 0, -69.3333333333327, 0, -69.3333333333333, -69.3333333333333, 0, 0, 95.9999999999989, 0, 96.0000000000002, 0, 191.999999999999, 192, 0, 96, 192, 0, -42.6666666666661, 0, -42.6666666666668, 0, -127.999999999999, -128, 0, -42.6666666666667, -128, 0, -128, -128, -256, -128}, 
                                                  {0, 0, 0, 16, 0, 0, -69.3333333333334, 0, -69.3333333333333, -69.3333333333334, 0, 0, 96.0000000000001, 0, 96, 0, 96, 192, 192, 192, 0, 0, -42.6666666666667, 0, -42.6666666666667, 0, -42.6666666666666, -128, -128, 0, -128, -128, -128, -128, -256}, 
                                                  {0, 0, 0, -12, 0, 0, 76, 0, 28, 28, 0, 0, -128, 0, -16, 0, -15.9999999999999, -144, -144, -31.9999999999999, 0, 0, 64, 0, 0, 0, 0, 128, 128, 0, 63.9999999999999, 63.9999999999999, 0, 0, 128}, 
                                                  {0, 0, 0, 5.33333333333333, 0, 0, -37.3333333333333, 0, -5.3333333333333, -5.33333333333331, 0, 0, 74.6666666666667, 0, 0, 0, 0, 31.9999999999999, 31.9999999999999, 0, 0, 0, -42.6666666666667, 0, 0, 0, 0, -42.6666666666666, -42.6666666666666, 0, 0, 0, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 5.33333333333334, 0, 0, 0, 0, -32, 0, 0, 0, 0, 0, 0, 0, 0, 0, 42.6666666666667, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0, 0, 0, 0, -16, 0, 0, -16, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 0, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 5.33333333333334, 0, 0, 0, 0, 0, 0, 0, -32, 0, 0, 0, 0, 0, 0, 0, 0, 0, 42.6666666666667, 0, 0, 0, 0, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 5.33333333333333, 0, 0, 0, 0, 0, 0, 0, -32, 0, 0, 0, 0, 0, 0, 0, 0, 0, 42.6666666666667, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 4.00000000000001, 0, 0, 0, 0, 0, 0, 0, -16, 0, -16, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 5.33333333333333, 0, 0, 0, 0, 0, 0, 0, 0, 0, -32, 0, 0, 0, 0, 0, 0, 0, 0, 0, 42.6666666666667, 0, 0, 0, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, 0, 0, 95.9999999999997, 0, 0, 0, 0, 0, -224, 0, -223.999999999998, 0, 0, 0, -224, 0, 0, 0, 128, 0, 127.999999999999, 0, 0, 0, 255.999999999999, 0, 0, 256, 256, 128}, 
                                                  {0, 0, 0, 0, 0, 0, 0, -31.9999999999999, 0, 0, 0, 0, 0, 160, 0, 31.9999999999996, 0, 0, 0, 32, 0, 0, 0, -128, 0, 0, 0, 0, 0, -128, 0, 0, -128, 0, 0}, 
                                                  {0, 0, 0, 0, 0, 0, 0, -31.9999999999998, 0, 0, 0, 0, 0, 31.9999999999999, 0, 159.999999999999, 0, 0, 0, 32.0000000000001, 0, 0, 0, 0, 0, -127.999999999999, 0, 0, 0, -128, 0, 0, 0, -128, 0}, 
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 96.0000000000001, 0, 0, 0, 0, -224, 0, 0, -224, 0, -224, 0, 0, 0, 0, 128, 0, 0, 128, 0, 0, 256, 0, 256, 128, 256}, 
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, -32, 0, 0, 0, 0, 160, 0, 0, 32.0000000000001, 0, 32, 0, 0, 0, 0, -128, 0, 0, 0, 0, 0, -128, 0, -128, 0, 0}, 
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, -32, 0, 0, 0, 0, 32.0000000000001, 0, 0, 160, 0, 32, 0, 0, 0, 0, 0, 0, 0, -128, 0, 0, -128, 0, 0, 0, -128}, 
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -32, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 128, 0, 0}, 
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -32.0000000000001, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 128, 0}, 
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -31.9999999999999, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 128}, 
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 96, 0, 0, 0, 0, 0, 0, 0, -224, 0, -224, -224, 0, 0, 0, 0, 0, 0, 128, 0, 128, 0, 0, 256, 128, 256, 256}, 
                                                  {0, 0, 0, 0, 0, 0, 0, 0, -32, 0, 0, 0, 0, 0, 0, 0, 160, 0, 32, 32, 0, 0, 0, 0, 0, 0, -128, 0, 0, 0, 0, -128, 0, -128, 0}, 
                                                  {0, 0, 0, 0, 0, 0, 0, 0, -32, 0, 0, 0, 0, 0, 0, 0, 32, 0, 160, 32, 0, 0, 0, 0, 0, 0, 0, 0, -128, 0, 0, -128, 0, 0, -128}, 
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 256, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -256, -256, -256}};

// -----------------------------------------------------------------------------
double MAd_TetInterpolant::coef5[56][56] = {{1, -11.4166666666667, -11.4166666666669, -11.4166666666667, 46.8750000000007, 46.8750000000015, 46.8750000000007, 93.7500000000035, 93.7500000000002, 93.7500000000003, -88.5416666666689, -88.5416666666698, -88.5416666666689, -265.625000000009, -265.625000000001, -265.625000000019, -265.625, -265.625000000001, -265.625000000003, -531.250000000004, 78.1250000000027, 78.1250000000025, 78.1250000000027, 312.500000000009, 312.500000000002, 312.500000000031, 312.499999999999, 312.500000000001, 312.500000000004, 468.750000000034, 468.750000000003, 468.750000000005, 937.500000000006, 937.500000000004, 937.500000000008, -26.0416666666679, -26.0416666666674, -26.0416666666679, -130.208333333337, -130.208333333335, -130.208333333348, -130.208333333333, -130.208333333334, -130.208333333336, -260.416666666683, -260.416666666668, -260.416666666695, -260.416666666668, -260.416666666668, -260.416666666672, -520.833333333338, -520.833333333333, -520.833333333338, -781.250000000003, -781.250000000006, -781.250000000006},
                                                  {0, 1.00000000000005, 0, 0, -10.4166666666672, 0, 0, 0, 0, 0, 36.458333333335, 0, 0, 0, 0, 0, 0, 0, 0, 0, -52.0833333333354, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 26.0416666666675, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 1.00000000000005, 0, 0, -10.4166666666671, 0, 0, 0, 0, 0, 36.4583333333349, 0, 0, 0, 0, 0, 0, 0, 0, 0, -52.0833333333353, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 26.0416666666675, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 1.00000000000005, 0, 0, -10.4166666666672, 0, 0, 0, 0, 0, 36.458333333335, 0, 0, 0, 0, 0, 0, 0, 0, 0, -52.0833333333354, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 26.0416666666675, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 25.0000000000004, 0, 0, -160.41666666667, 0, 0, -160.416666666675, 0, -160.416666666667, 369.791666666677, 0, 0, 739.583333333355, 739.583333333338, 369.791666666712, 0, 369.791666666667, 0, 739.583333333338, -364.583333333347, 0, 0, -1093.75000000002, -1093.75000000001, -364.583333333405, 0, -364.583333333333, 0, -1093.75000000008, -1093.75000000001, 0, -2187.50000000001, -1093.75000000001, -1093.75, 130.208333333339, 0, 0, 520.833333333341, 520.833333333338, 130.208333333367, 0, 130.208333333333, 0, 781.250000000041, 781.250000000005, 520.833333333403, 0, 520.833333333336, 0, 1562.50000000001, 520.833333333337, 520.833333333337, 1562.50000000001, 1562.50000000001, 781.250000000004},
                                                  {0, -25.0000000000007, 0, 0, 222.916666666673, 0, 0, 97.9166666666744, 0, 97.9166666666679, -614.583333333353, 0, 0, -739.583333333354, -739.583333333341, -125.000000000043, 0, -125.000000000001, 0, -250.000000000005, 677.083333333358, 0, 0, 1406.25000000002, 1406.25000000001, 52.0833333334018, 0, 52.0833333333314, 0, 781.250000000084, 781.250000000007, 0, 1562.50000000002, 156.250000000013, 156.250000000007, -260.416666666677, 0, 0, -781.250000000007, -781.250000000008, 0, 0, 0, 0, -781.250000000042, -781.250000000007, -260.416666666736, 0, -260.416666666669, 0, -1562.50000000001, 0, 0, -781.250000000015, -781.250000000011, 0},
                                                  {0, 16.6666666666673, 0, 0, -162.500000000006, 0, 0, -37.5000000000037, 0, -37.5000000000011, 510.416666666685, 0, 0, 322.916666666677, 322.916666666673, 20.8333333333543, 0, 20.8333333333342, 0, 41.6666666666708, -625.000000000023, 0, 0, -781.25000000001, -781.250000000011, 0, 0, 0, 0, -156.250000000041, -156.250000000004, 0, -312.500000000017, 0, 0, 260.416666666677, 0, 0, 520.833333333337, 520.833333333339, 0, 0, 0, 0, 260.416666666688, 260.416666666671, 0, 0, 0, 0, 520.833333333346, 0, 0, 0, 0, 0},
                                                  {0, -6.25000000000028, 0, 0, 63.5416666666694, 0, 0, 6.25000000000092, 0, 6.25000000000046, -213.541666666675, 0, 0, -57.2916666666696, -57.291666666669, 0, 0, 0, 0, 0, 286.458333333344, 0, 0, 156.250000000003, 156.250000000004, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -130.208333333338, 0, 0, -130.208333333335, -130.208333333335, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, -6.25000000000047, 0, 0, 0, 0, 0, 57.2916666666675, 0, 0, 0, 0, 0, 0, 0, 0, 0, -156.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 130.208333333333, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, -4.16666666666739, 0, 0, 0, 0, 0, 31.2500000000013, 0, 20.8333333333372, 0, 0, 0, 0, 0, 0, 0, -52.0833333333349, 0, 0, 0, 0, 0, -156.250000000007, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 260.41666666667, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, -4.16666666666737, 0, 0, 0, 0, 0, 20.833333333336, 0, 31.2500000000034, 0, 0, 0, 0, 0, 0, 0, 0, 0, -52.08333333334, 0, 0, 0, -156.250000000005, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 260.41666666667, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, -6.25000000000043, 0, 0, 0, 0, 0, 0, 0, 57.2916666666697, 0, 0, 0, 0, 0, 0, 0, 0, 0, -156.250000000005, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 130.208333333335, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, -6.25000000000026, 0, 0, 63.5416666666694, 0, 6.25000000000062, 6.25000000000008, 0, 0, -213.541666666675, 0, 0, 0, -57.2916666666707, -57.291666666667, 0, 0, 0, 0, 286.458333333344, 0, 0, 0, 156.250000000007, 156.249999999999, 0, 0, 0, 0, 0, 0, 0, 0, 0, -130.208333333338, 0, 0, 0, -130.208333333337, -130.208333333334, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 16.6666666666673, 0, 0, -162.500000000006, 0, -37.5000000000031, -37.5000000000006, 0, 0, 510.416666666686, 0, 20.8333333333399, 0, 322.916666666686, 322.916666666669, 0, 20.8333333333359, 41.6666666666683, 0, -625.000000000024, 0, 0, 0, -781.250000000034, -781.250000000003, 0, 0, -156.250000000028, 0, -156.250000000005, 0, -312.500000000002, 0, 0, 260.416666666677, 0, 0, 0, 520.833333333351, 520.833333333336, 0, 0, 0, 0, 260.41666666669, 260.416666666667, 0, 0, 0, 520.833333333333, 0, 0, 0, 0},
                                                  {0, 0, -25.0000000000008, 0, 0, 222.916666666674, 0, 97.9166666666736, 97.9166666666676, 0, 0, -614.583333333355, 0, -125.000000000016, 0, -739.583333333374, -739.583333333336, 0, -125.000000000005, -250.000000000005, 0, 677.083333333359, 0, 52.0833333333486, 0, 1406.25000000007, 1406.25, 0, 52.0833333333394, 781.250000000061, 0, 781.250000000011, 156.250000000012, 1562.50000000001, 156.250000000009, 0, -260.416666666677, 0, 0, 0, -781.250000000035, -781.250000000004, 0, 0, -260.416666666692, 0, -781.250000000051, -781.250000000004, 0, -260.416666666676, 0, -1562.5, 0, -781.250000000012, 0, -781.250000000008},
                                                  {0, 0, 25.0000000000006, 0, 0, -160.416666666672, 0, -160.416666666674, -160.416666666667, 0, 0, 369.791666666679, 0, 369.791666666685, 0, 739.583333333378, 739.583333333334, 0, 369.791666666671, 739.583333333338, 0, -364.583333333346, 0, -364.583333333351, 0, -1093.75000000007, -1093.75, 0, -364.583333333341, -1093.75000000007, 0, -1093.75000000001, -1093.75000000001, -2187.5, -1093.75000000001, 0, 130.208333333338, 0, 130.20833333334, 0, 520.833333333369, 520.833333333334, 0, 130.208333333337, 520.833333333366, 0, 781.25000000006, 781.250000000004, 0, 520.833333333344, 520.83333333334, 1562.5, 520.833333333338, 1562.50000000001, 781.250000000004, 1562.5},
                                                  {0, 0, 0, 25.0000000000004, 0, 0, -160.41666666667, 0, -160.416666666667, -160.416666666667, 0, 0, 369.791666666677, 0, 369.791666666669, 0, 369.791666666666, 739.583333333338, 739.583333333339, 739.583333333341, 0, 0, -364.583333333347, 0, -364.583333333335, 0, -364.583333333329, -1093.75000000001, -1093.75000000001, 0, -1093.75000000001, -1093.75000000001, -1093.75000000001, -1093.75, -2187.50000000002, 0, 0, 130.208333333339, 0, 130.208333333335, 0, 130.208333333331, 520.833333333336, 520.833333333339, 0, 520.833333333338, 0, 520.833333333337, 781.250000000006, 781.250000000011, 520.833333333341, 520.83333333333, 1562.50000000002, 781.250000000005, 1562.50000000002, 1562.50000000001},
                                                  {0, 0, 0, -25.0000000000007, 0, 0, 222.916666666673, 0, 97.9166666666678, 97.916666666668, 0, 0, -614.583333333353, 0, -125.000000000003, 0, -125.000000000001, -739.583333333341, -739.583333333341, -250.000000000008, 0, 0, 677.083333333358, 0, 52.0833333333366, 0, 52.083333333332, 1406.25000000001, 1406.25000000001, 0, 781.250000000013, 781.250000000011, 156.250000000013, 156.250000000006, 1562.50000000003, 0, 0, -260.416666666677, 0, 0, 0, 0, -781.250000000006, -781.250000000007, 0, -260.416666666674, 0, -260.416666666672, -781.250000000008, -781.250000000011, 0, 0, -1562.50000000002, 0, -781.250000000018, -781.250000000015},
                                                  {0, 0, 0, 16.6666666666673, 0, 0, -162.500000000006, 0, -37.5000000000011, -37.5000000000011, 0, 0, 510.416666666685, 0, 20.8333333333357, 0, 20.8333333333359, 322.916666666673, 322.916666666673, 41.6666666666725, 0, 0, -625.000000000023, 0, 0, 0, 0, -781.25000000001, -781.25000000001, 0, -156.250000000008, -156.250000000009, 0, 0, -312.500000000016, 0, 0, 260.416666666677, 0, 0, 0, 0, 520.833333333338, 520.833333333338, 0, 0, 0, 0, 260.416666666672, 260.416666666673, 0, 0, 520.833333333346, 0, 0, 0},
                                                  {0, 0, 0, -6.25000000000028, 0, 0, 63.5416666666694, 0, 6.25000000000045, 6.25000000000043, 0, 0, -213.541666666675, 0, 0, 0, 0, -57.2916666666688, -57.2916666666688, 0, 0, 0, 286.458333333344, 0, 0, 0, 0, 156.250000000004, 156.250000000003, 0, 0, 0, 0, 0, 0, 0, 0, -130.208333333338, 0, 0, 0, 0, -130.208333333335, -130.208333333335, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, -6.25000000000001, 0, 0, 0, 0, 57.2916666666669, 0, 0, 0, 0, 0, 0, 0, 0, 0, -156.250000000001, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 130.208333333334, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, -4.16666666666659, 0, 0, 0, 0, 31.2499999999999, 0, 0, 20.8333333333331, 0, 0, 0, 0, 0, 0, -52.0833333333332, 0, 0, 0, 0, 0, -156.250000000001, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 260.416666666667, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, -4.16666666666675, 0, 0, 0, 0, 20.8333333333337, 0, 0, 31.2500000000003, 0, 0, 0, 0, 0, 0, 0, 0, 0, -52.0833333333336, 0, 0, -156.250000000001, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 260.416666666668, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, -6.24999999999998, 0, 0, 0, 0, 0, 0, 0, 57.2916666666667, 0, 0, 0, 0, 0, 0, 0, 0, 0, -156.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 130.208333333333, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, -6.25000000000009, 0, 0, 0, 0, 0, 0, 0, 57.291666666667, 0, 0, 0, 0, 0, 0, 0, 0, 0, -156.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 130.208333333333, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, -4.16666666666659, 0, 0, 0, 0, 0, 0, 0, 31.2500000000001, 0, 20.8333333333327, 0, 0, 0, 0, 0, 0, 0, -52.0833333333335, 0, 0, 0, 0, -156.250000000001, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 260.416666666668, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, -4.16666666666677, 0, 0, 0, 0, 0, 0, 0, 20.8333333333333, 0, 31.2500000000008, 0, 0, 0, 0, 0, 0, 0, 0, 0, -52.0833333333343, 0, 0, -156.250000000001, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 260.416666666668, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, -6.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 57.2916666666667, 0, 0, 0, 0, 0, 0, 0, 0, 0, -156.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 130.208333333333, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 250.000000000015, 0, 0, 0, 0, 0, -979.166666666702, 0, -979.166666666751, 0, 0, 0, -979.166666666667, 0, 0, 0, 1250.00000000003, 0, 1250.00000000013, 0, 0, 0, 2500.00000000014, 0, 0, 2500, 2500, 1249.99999999999, 0, 0, 0, -520.833333333345, 0, -520.833333333398, 0, 0, 0, -1562.50000000007, 0, -1562.50000000012, 0, 0, 0, -1562.5, -1562.5, -520.83333333333, -3125, -1562.49999999999, -1562.49999999999},
                                                  {0, 0, 0, 0, 0, 0, 0, 41.66666666667, 0, 0, 0, 0, 0, -354.166666666673, 0, -41.6666666666888, 0, 0, 0, -41.6666666666664, 0, 0, 0, 833.333333333337, 0, 0, 0, 0, 0, 312.500000000041, 0, 0, 312.499999999999, 0, 0, 0, 0, 0, -520.833333333333, 0, 0, 0, 0, 0, -520.833333333352, 0, 0, 0, 0, 0, -520.833333333333, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 41.6666666666699, 0, 0, 0, 0, 0, -41.666666666674, 0, -354.166666666687, 0, 0, 0, -41.6666666666673, 0, 0, 0, 0, 0, 833.333333333367, 0, 0, 0, 312.500000000029, 0, 0, 0, 312.500000000003, 0, 0, 0, 0, 0, 0, -520.833333333351, 0, 0, 0, 0, 0, -520.833333333358, 0, 0, 0, 0, -520.833333333336, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, -125.00000000001, 0, 0, 0, 0, 0, 906.250000000025, 0, 281.250000000061, 0, 0, 0, 281.249999999999, 0, 0, 0, -1562.50000000002, 0, -156.250000000096, 0, 0, 0, -1718.75000000011, 0, 0, -1718.75, -312.499999999997, -156.249999999996, 0, 0, 0, 781.250000000007, 0, 0, 0, 0, 0, 1562.50000000005, 0, 781.250000000093, 0, 0, 0, 1562.5, 0, 0, 1562.5, 781.249999999991, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 31.2500000000047, 0, 0, 0, 0, 0, -187.500000000013, 0, -187.500000000024, 0, 0, 0, -31.249999999999, 0, 0, 0, 156.250000000015, 0, 156.250000000038, 0, 0, 0, 1093.75000000004, 0, 0, 156.249999999997, 156.249999999999, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -781.250000000019, 0, -781.250000000032, 0, 0, 0, 0, 0, 0, -781.249999999998, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, -125.00000000001, 0, 0, 0, 0, 0, 281.250000000025, 0, 906.250000000058, 0, 0, 0, 281.250000000001, 0, 0, 0, -156.250000000025, 0, -1562.50000000009, 0, 0, 0, -1718.75000000009, 0, 0, -312.500000000003, -1718.75000000001, -156.249999999994, 0, 0, 0, 0, 0, 781.250000000047, 0, 0, 0, 781.25000000004, 0, 1562.50000000007, 0, 0, 0, 0, 1562.50000000001, 0, 1562.50000000001, 0, 781.249999999997},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 250, 0, 0, 0, 0, -979.166666666667, 0, 0, -979.166666666666, 0, -979.166666666668, 0, 0, 0, 0, 1250, 0, 0, 1250, 0, 0, 2500.00000000001, 0, 2500, 1250, 2500.00000000001, 0, 0, 0, 0, -520.833333333335, 0, 0, -520.833333333331, 0, 0, -1562.5, 0, 0, -1562.5, 0, -1562.5, -520.833333333327, -1562.5, -1562.5, -3125.00000000001, -1562.5},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 41.6666666666665, 0, 0, 0, 0, -354.166666666668, 0, 0, -41.6666666666646, 0, -41.6666666666668, 0, 0, 0, 0, 833.333333333337, 0, 0, 0, 0, 0, 312.499999999997, 0, 312.500000000001, 0, 0, 0, 0, 0, 0, -520.833333333336, 0, 0, 0, 0, 0, -520.833333333333, 0, 0, 0, 0, -520.833333333335, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 41.6666666666667, 0, 0, 0, 0, -41.6666666666668, 0, 0, -354.166666666668, 0, -41.6666666666664, 0, 0, 0, 0, 0, 0, 0, 833.333333333336, 0, 0, 312.500000000002, 0, 0, 0, 312.5, 0, 0, 0, 0, 0, 0, 0, -520.833333333335, 0, 0, 0, 0, 0, -520.833333333335, 0, 0, 0, -520.833333333333, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, -125, 0, 0, 0, 0, 906.250000000002, 0, 0, 281.249999999997, 0, 281.250000000001, 0, 0, 0, 0, -1562.5, 0, 0, -156.249999999994, 0, 0, -1718.75, 0, -1718.75, -156.25, -312.500000000001, 0, 0, 0, 0, 781.250000000004, 0, 0, 0, 0, 0, 1562.5, 0, 0, 781.249999999999, 0, 1562.5, 0, 0, 781.250000000002, 1562.5, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 31.25, 0, 0, 0, 0, -187.500000000001, 0, 0, -187.5, 0, -31.2500000000005, 0, 0, 0, 0, 156.250000000001, 0, 0, 156.249999999999, 0, 0, 1093.75000000001, 0, 156.250000000001, 0, 156.250000000001, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -781.250000000004, 0, 0, -781.250000000003, 0, 0, 0, 0, 0, -781.250000000001, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, -125, 0, 0, 0, 0, 281.250000000001, 0, 0, 906.250000000001, 0, 281.25, 0, 0, 0, 0, -156.250000000001, 0, 0, -1562.5, 0, 0, -1718.75000000001, 0, -312.500000000002, -156.249999999994, -1718.75, 0, 0, 0, 0, 0, 0, 0, 781.250000000001, 0, 0, 781.250000000006, 0, 0, 1562.50000000001, 0, 0, 0, 1562.5, 0, 1562.5, 781.249999999998},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 41.6666666666666, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -312.5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 520.833333333334, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 41.6666666666674, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -312.500000000003, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 520.833333333337, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 41.6666666666664, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -312.500000000001, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 520.833333333333, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 31.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -156.25, -156.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 781.25, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 31.2499999999999, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -156.249999999998, -156.249999999999, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 781.249999999998},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 31.2500000000001, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -156.25, 0, -156.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 781.25, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 250, 0, 0, 0, 0, 0, 0, 0, -979.166666666665, 0, -979.166666666673, -979.166666666674, 0, 0, 0, 0, 0, 0, 1249.99999999999, 0, 1250.00000000001, 0, 0, 2500.00000000001, 1250.00000000002, 2500, 2500.00000000002, 0, 0, 0, 0, 0, 0, -520.83333333333, 0, -520.833333333338, 0, 0, 0, -1562.50000000001, 0, -1562.50000000002, -520.833333333344, -1562.5, -1562.50000000002, -1562.50000000001, -1562.50000000002, -3125.00000000001},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 41.6666666666671, 0, 0, 0, 0, 0, 0, 0, -354.166666666669, 0, -41.6666666666686, -41.6666666666697, 0, 0, 0, 0, 0, 0, 833.333333333336, 0, 0, 0, 0, 312.500000000005, 0, 312.500000000005, 0, 0, 0, 0, 0, 0, 0, -520.833333333335, 0, 0, 0, 0, 0, -520.833333333336, 0, 0, 0, -520.833333333335, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 41.6666666666667, 0, 0, 0, 0, 0, 0, 0, -41.6666666666666, 0, -354.166666666668, -41.6666666666671, 0, 0, 0, 0, 0, 0, 0, 0, 833.333333333337, 0, 0, 312.500000000002, 0, 0, 312.500000000002, 0, 0, 0, 0, 0, 0, 0, 0, -520.833333333335, 0, 0, 0, 0, 0, -520.833333333335, 0, 0, -520.833333333335, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, -125.000000000001, 0, 0, 0, 0, 0, 0, 0, 906.250000000003, 0, 281.250000000005, 281.250000000008, 0, 0, 0, 0, 0, 0, -1562.5, 0, -156.250000000006, 0, 0, -1718.75000000001, -156.250000000014, -1718.75000000001, -312.500000000021, 0, 0, 0, 0, 0, 0, 781.250000000002, 0, 0, 0, 0, 0, 1562.50000000001, 0, 781.250000000013, 0, 1562.5, 0, 781.25000000001, 0, 1562.50000000002},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 31.25, 0, 0, 0, 0, 0, 0, 0, -187.5, 0, -187.500000000001, -31.2500000000023, 0, 0, 0, 0, 0, 0, 156.249999999999, 0, 156.250000000001, 0, 0, 1093.75, 0, 156.250000000002, 156.250000000006, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -781.250000000003, 0, -781.250000000004, 0, 0, 0, 0, 0, -781.250000000003},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, -125, 0, 0, 0, 0, 0, 0, 0, 281.249999999998, 0, 906.250000000003, 281.250000000002, 0, 0, 0, 0, 0, 0, -156.249999999994, 0, -1562.50000000001, 0, 0, -1718.75000000001, -156.250000000006, -312.499999999997, -1718.75000000001, 0, 0, 0, 0, 0, 0, 0, 0, 781.250000000003, 0, 0, 0, 781.250000000002, 0, 1562.50000000001, 0, 0, 1562.50000000001, 0, 781.250000000008, 1562.5},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1250, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2812.5, -2812.5, -2812.50000000001, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1562.5, 1562.5, 1562.50000000001, 3125, 3125.00000000001, 3125},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -312.5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1875, 312.499999999999, 312.500000000001, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1562.5, 0, 0, -1562.5, -1562.5, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -312.500000000002, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 312.500000000003, 1875.00000000001, 312.500000000004, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1562.50000000001, 0, -1562.5, 0, -1562.50000000001},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -312.499999999999, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 312.5, 312.499999999992, 1875, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1562.5, 0, -1562.5, -1562.49999999999}};
                 
// -----------------------------------------------------------------------------
double MAd_TetInterpolant::coef6[84][84] = {{1, -14.6999999999992, -14.6999999999973, -14.6999999999992, 81.1999999999899, 81.1999999999682, 81.1999999999899, 162.399999999973, 162.399999999999, 162.399999999996, -220.499999999955, -220.499999999874, -220.499999999955, -661.499999999994, -661.49999999999, -661.499999999654, -661.500000000009, -661.499999999954, -661.499999999981, -1322.99999999995, 314.999999999908, 314.999999999773, 314.999999999908, 1260.00000000015, 1260, 1259.99999999885, 1260, 1259.99999999986, 1259.9999999999, 1889.99999999939, 1889.99999999989, 1890.00000000007, 3779.99999999978, 3779.99999999979, 3779.99999999984, -226.799999999913, -226.799999999806, -226.799999999913, -1134.00000000023, -1134.00000000001, -1133.99999999859, -1133.99999999998, -1133.99999999983, -1133.99999999984, -2267.99999999976, -2267.99999999992, -2267.99999999831, -2268.00000000015, -2267.99999999977, -2268.00000000007, -4535.99999999967, -4535.99999999966, -4535.99999999983, -6803.99999999945, -6803.99999999947, -6803.99999999973, 64.7999999999695, 64.7999999999367, 64.7999999999695, 388.800000000094, 388.800000000008, 388.799999999413, 388.799999999986, 388.799999999929, 388.799999999918, 972.000000000061, 971.999999999979, 971.99999999888, 972.000000000057, 971.999999999872, 972.000000000005, 1943.99999999985, 1943.99999999983, 1943.99999999989, 1295.99999999934, 1295.99999999989, 1296.00000000013, 3887.99999999963, 3887.9999999996, 3887.99999999963, 3887.99999999981, 3887.99999999969, 3887.99999999995, 5831.99999999944},
                                                  {0, -0.999999999999524, 0, 0, 13.6999999999942, 0, 0, 0, 0, 0, -67.4999999999748, 0, 0, 0, 0, 0, 0, 0, 0, 0, 152.99999999995, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -161.999999999954, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64.7999999999839, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, -0.999999999999513, 0, 0, 13.699999999994, 0, 0, 0, 0, 0, -67.4999999999739, 0, 0, 0, 0, 0, 0, 0, 0, 0, 152.999999999947, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -161.999999999952, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64.7999999999833, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, -0.999999999999524, 0, 0, 13.6999999999942, 0, 0, 0, 0, 0, -67.4999999999748, 0, 0, 0, 0, 0, 0, 0, 0, 0, 152.99999999995, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -161.999999999954, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64.7999999999839, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 35.9999999999957, 0, 0, -313.199999999946, 0, 0, -313.199999999921, 0, -313.199999999986, 1043.99999999976, 0, 0, 2087.99999999999, 2087.99999999996, 1043.99999999898, 0, 1043.99999999986, 0, 2087.99999999991, -1673.99999999951, 0, 0, -5022.00000000044, -5021.99999999998, -1673.99999999666, 0, -1673.9999999996, 0, -5021.99999999821, -5021.99999999967, 0, -10043.9999999996, -5021.99999999974, -5021.99999999969, 1295.99999999954, 0, 0, 5184.00000000064, 5184.00000000002, 1295.99999999597, 0, 1295.99999999953, 0, 7775.99999999935, 7775.99999999978, 5183.99999999495, 0, 5183.99999999933, 0, 15551.9999999993, 5183.99999999966, 5183.99999999958, 15551.9999999992, 15551.999999999, 7775.99999999963, -388.79999999984, 0, 0, -1944.00000000025, -1944.00000000001, -388.799999998365, 0, -388.799999999809, 0, -3888.00000000025, -3887.99999999996, -1943.99999999666, 0, -1943.99999999962, 0, -7775.99999999964, -1943.99999999984, -1943.99999999972, -3887.99999999804, -3887.99999999971, 0, -11663.9999999994, -11663.9999999993, -7775.99999999957, -3887.99999999978, -7775.99999999943, -3887.99999999988, -11663.9999999993},
                                                  {0, -44.9999999999902, 0, 0, 526.499999999878, 0, 0, 256.499999999908, 0, 256.499999999982, -2074.49999999946, 0, 0, -2610.00000000002, -2609.99999999995, -535.499999998777, 0, -535.499999999844, 0, -1070.99999999995, 3698.99999999891, 0, 0, 7884.00000000058, 7883.99999999994, 485.999999995989, 0, 485.999999999557, 0, 4670.99999999797, 4670.99999999968, 0, 9341.99999999969, 1457.9999999998, 1457.99999999977, -3077.99999999899, 0, 0, -9396.00000000078, -9395.99999999997, -161.999999995222, 0, -161.999999999496, 0, -9719.99999999951, -9719.99999999983, -3563.99999999397, 0, -3563.99999999935, 0, -19439.9999999995, -647.999999999804, -647.999999999753, -10691.9999999995, -10691.9999999994, -971.999999999588, 971.999999999647, 0, 0, 3888.00000000029, 3887.99999999999, 0, 0, 0, 0, 5832.00000000046, 5832, 971.999999995979, 0, 971.999999999639, 0, 11663.9999999997, 0, 0, 3887.99999999776, 3887.99999999977, 0, 11663.9999999995, 11663.9999999996, 3887.99999999978, 0, 3887.99999999969, 0, 5831.9999999996},
                                                  {0, 39.999999999988, 0, 0, -507.999999999852, 0, 0, -147.999999999947, 0, -147.999999999989, 2231.99999999935, 0, 0, 1692.00000000006, 1691.99999999997, 179.999999999246, 0, 179.999999999921, 0, 359.999999999963, -4355.9999999987, 0, 0, -6120.00000000046, -6119.99999999995, -71.9999999974996, 0, -71.9999999997815, 0, -1835.99999999893, -1835.9999999999, 0, -3671.99999999979, -215.99999999983, -215.999999999799, 3887.99999999879, 0, 0, 8424.00000000055, 8423.99999999995, 0, 0, 0, 0, 5184.00000000007, 5184.00000000002, 647.999999996407, 0, 647.999999999809, 0, 10367.9999999997, 0, 0, 1943.99999999967, 1943.99999999959, 0, -1295.99999999958, 0, 0, -3888.00000000019, -3887.99999999997, 0, 0, 0, 0, -3888.00000000047, -3888.00000000004, 0, 0, 0, 0, -7775.99999999984, 0, 0, -1295.99999999881, -1295.99999999998, 0, -3887.99999999966, -3887.99999999978, 0, 0, 0, 0, 0},
                                                  {0, -22.4999999999917, 0, 0, 296.999999999898, 0, 0, 49.4999999999846, 0, 49.4999999999972, -1381.49999999956, 0, 0, -598.500000000041, -598.499999999995, -26.9999999997568, 0, -26.999999999984, 0, -53.9999999999804, 2888.99999999911, 0, 0, 2376.0000000002, 2375.99999999999, 0, 0, 0, 0, 296.999999999733, 297.000000000016, 0, 593.999999999885, 0, 0, -2753.99999999918, 0, 0, -3726.00000000021, -3725.99999999998, 0, 0, 0, 0, -972.0000000002, -972.000000000063, 0, 0, 0, 0, -1943.99999999988, 0, 0, 0, 0, 0, 971.999999999714, 0, 0, 1944.00000000007, 1943.99999999998, 0, 0, 0, 0, 972.000000000237, 972.000000000035, 0, 0, 0, 0, 1943.99999999994, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 7.19999999999693, 0, 0, -97.1999999999624, 0, 0, -7.19999999999824, 0, -7.19999999999962, 467.999999999836, 0, 0, 90.0000000000102, 90.0000000000002, 0, 0, 0, 0, 0, -1025.99999999967, 0, 0, -378.000000000039, -377.999999999999, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1036.7999999997, 0, 0, 648.000000000034, 647.999999999997, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -388.799999999895, 0, 0, -388.800000000013, -388.799999999998, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 7.19999999999854, 0, 0, 0, 0, 0, -90.0000000000113, 0, 0, 0, 0, 0, 0, 0, 0, 0, 378.00000000004, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -648.000000000034, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 388.800000000012, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 4.49999999999274, 0, 0, 0, 0, 0, -49.499999999977, 0, -26.999999999926, 0, 0, 0, 0, 0, 0, 0, 161.999999999978, 0, 0, 0, 0, 0, 296.999999999809, 0, 0, 0, 0, 0, 0, 0, 0, -162.000000000007, 0, 0, 0, 0, 0, -971.999999999814, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 971.999999999947, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 3.99999999999425, 0, 0, 0, 0, 0, -35.9999999999946, 0, -35.9999999999387, 0, 0, 0, 0, 0, 0, 0, 72.000000000009, 0, 71.9999999998264, 0, 0, 0, 323.99999999987, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -647.999999999904, 0, -647.999999999705, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1295.99999999985, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 4.4999999999986, 0, 0, 0, 0, 0, -26.9999999999992, 0, -49.4999999999736, 0, 0, 0, 0, 0, 0, 0, 0, 0, 161.999999999905, 0, 0, 0, 296.999999999933, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -161.999999999884, 0, 0, 0, 0, 0, -971.999999999821, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 971.999999999888, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 7.20000000000054, 0, 0, 0, 0, 0, 0, 0, -89.9999999999935, 0, 0, 0, 0, 0, 0, 0, 0, 0, 377.999999999964, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -647.999999999947, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 388.799999999974, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 7.19999999999707, 0, 0, -97.1999999999627, 0, -7.19999999999966, -7.20000000000034, 0, 0, 467.999999999837, 0, 0, 0, 89.9999999999816, 90.000000000001, 0, 0, 0, 0, -1025.99999999967, 0, 0, 0, -377.999999999927, -377.999999999999, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1036.7999999997, 0, 0, 0, 647.999999999886, 648, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -388.799999999895, 0, 0, 0, -388.799999999945, -388.799999999996, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, -22.4999999999925, 0, 0, 296.999999999905, 0, 49.4999999999976, 49.4999999999998, 0, 0, -1381.49999999958, 0, -27.0000000000529, 0, -598.499999999893, -598.5, 0, -26.9999999999921, -53.9999999999875, 0, 2888.99999999915, 0, 0, 0, 2375.99999999958, 2376, 0, 0, 296.999999999936, 0, 297.000000000011, 0, 593.999999999959, 0, 0, -2753.9999999992, 0, 0, 0, -3725.99999999938, -3725.99999999999, 0, 0, 0, 0, -971.9999999996, -972.000000000068, 0, 0, 0, -1943.99999999989, 0, 0, 0, 0, 0, 971.999999999722, 0, 0, 0, 1943.99999999971, 1943.99999999998, 0, 0, 0, 0, 971.99999999966, 972.000000000023, 0, 0, 0, 1943.99999999994, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 39.999999999989, 0, 0, -507.99999999986, 0, -147.999999999983, -147.999999999997, 0, 0, 2231.99999999938, 0, 180.000000000102, 0, 1691.99999999962, 1692, 0, 179.999999999957, 359.999999999979, 0, -4355.99999999875, 0, -72.0000000004533, 0, -6119.99999999862, -6119.99999999999, 0, -71.9999999998677, -1835.99999999953, 0, -1835.99999999998, -215.99999999994, -3671.99999999984, -215.999999999977, 0, 3887.99999999883, 0, 0, 0, 8423.99999999813, 8423.99999999997, 0, 0, 648.000000000139, 0, 5183.99999999832, 5184.00000000009, 0, 647.999999999947, 0, 10367.9999999996, 0, 1943.99999999967, 0, 1943.99999999978, 0, -1295.99999999959, 0, 0, 0, -3887.99999999915, -3887.99999999997, 0, 0, 0, 0, -3887.99999999876, -3888.00000000005, 0, 0, 0, -7775.99999999981, 0, -1295.9999999995, 0, -1296.00000000002, 0, 0, -3887.99999999963, -3887.99999999961, 0, 0, 0},
                                                  {0, 0, -44.9999999999888, 0, 0, 526.499999999861, 0, 256.499999999945, 256.499999999992, 0, 0, -2074.4999999994, 0, -535.50000000004, 0, -2609.9999999992, -2609.99999999999, 0, -535.499999999898, -1070.99999999995, 0, 3698.99999999881, 0, 486.000000000507, 0, 7883.9999999973, 7883.99999999998, 0, 485.99999999966, 4670.99999999858, 0, 4670.9999999999, 1457.99999999973, 9341.99999999967, 1457.99999999994, 0, -3077.99999999891, 0, -162.000000000755, 0, -9395.99999999654, -9395.99999999995, 0, -161.999999999556, -3563.99999999941, 0, -9719.99999999611, -9720.00000000002, 0, -3563.99999999975, -647.999999999557, -19439.9999999993, -648.000000000074, -10691.9999999992, -971.999999999485, -10691.9999999997, 0, 971.999999999624, 0, 0, 0, 3887.9999999985, 3887.99999999997, 0, 0, 972.000000000132, 0, 5831.9999999974, 5832.00000000002, 0, 971.999999999834, 0, 11663.9999999996, 0, 3887.9999999985, 0, 3887.99999999998, 3887.9999999995, 0, 11663.9999999993, 11663.9999999995, 0, 3888.00000000006, 5831.99999999942},
                                                  {0, 0, 35.9999999999921, 0, 0, -313.199999999905, 0, -313.199999999934, -313.199999999994, 0, 0, 1043.99999999961, 0, 1043.99999999998, 0, 2087.99999999914, 2088, 0, 1043.99999999991, 2087.99999999991, 0, -1673.99999999926, 0, -1674.00000000036, 0, -5021.99999999717, -5022, 0, -1673.99999999968, -5021.99999999838, 0, -5022, -5021.99999999959, -10043.9999999996, -5021.9999999998, 0, 1295.99999999934, 0, 1296.00000000058, 0, 5183.99999999649, 5183.99999999996, 0, 1295.99999999954, 5183.99999999916, 0, 7775.99999999575, 7776.00000000013, 0, 5183.99999999989, 5183.99999999937, 15551.9999999993, 5183.99999999982, 15551.9999999989, 7775.9999999991, 15551.9999999996, 0, -388.799999999779, 0, -388.800000000257, 0, -1943.99999999852, -1943.99999999997, 0, -388.799999999782, -1944.00000000003, 0, -3887.99999999723, -3888.00000000005, 0, -1943.99999999987, -1943.99999999969, -7775.99999999962, -1943.99999999988, -3887.99999999828, 0, -3888.0000000001, -7775.9999999993, -3887.99999999926, -11663.9999999992, -11663.9999999995, -3887.99999999956, -7776.00000000004, -11663.9999999991},
                                                  {0, 0, 0, 35.9999999999957, 0, 0, -313.199999999946, 0, -313.200000000001, -313.19999999999, 0, 0, 1043.99999999976, 0, 1043.99999999997, 0, 1044.00000000002, 2087.99999999989, 2087.99999999998, 2087.99999999989, 0, 0, -1673.99999999951, 0, -1673.99999999998, 0, -1674, -5021.99999999968, -5021.99999999985, 0, -5021.99999999969, -5022.00000000023, -5021.99999999953, -5021.99999999949, -10043.9999999996, 0, 0, 1295.99999999954, 0, 1296.00000000002, 0, 1295.99999999994, 5183.99999999962, 5183.99999999972, 0, 5183.99999999968, 0, 5184.00000000042, 7775.99999999943, 7776.00000000031, 5183.99999999944, 5183.99999999918, 15551.9999999996, 7775.9999999986, 15551.9999999988, 15551.9999999994, 0, 0, -388.79999999984, 0, -388.800000000016, 0, -388.799999999958, -1943.99999999983, -1943.99999999984, 0, -1943.99999999989, 0, -1944.00000000016, -3887.99999999969, -3888.00000000009, -1943.99999999979, -1943.99999999957, -7775.99999999979, 0, -3887.99999999966, -3888.00000000038, -3887.9999999992, -7775.99999999924, -3887.999999999, -7775.99999999965, -11663.9999999993, -11663.9999999999, -11663.9999999985},
                                                  {0, 0, 0, -44.9999999999902, 0, 0, 526.499999999878, 0, 256.500000000004, 256.499999999991, 0, 0, -2074.49999999946, 0, -535.49999999997, 0, -535.500000000023, -2609.9999999999, -2610.00000000003, -1070.99999999989, 0, 0, 3698.99999999891, 0, 485.999999999965, 0, 485.999999999983, 7883.99999999971, 7884, 0, 4670.99999999973, 4671.00000000043, 1457.99999999954, 1457.99999999949, 9341.99999999969, 0, 0, -3077.99999999899, 0, -161.999999999997, 0, -161.999999999891, -9395.99999999965, -9395.99999999988, 0, -3563.99999999964, 0, -3564.00000000064, -9719.99999999954, -9720.0000000007, -647.999999999467, -647.999999999148, -19439.9999999997, -971.999999998574, -10691.9999999989, -10691.9999999995, 0, 0, 971.999999999647, 0, 0, 0, 0, 3887.99999999984, 3887.9999999999, 0, 971.999999999851, 0, 972.000000000249, 5831.99999999978, 5832.00000000032, 0, 0, 11663.9999999999, 0, 3887.99999999966, 3888.0000000006, 0, 3887.99999999932, 0, 3887.99999999976, 11663.9999999994, 11664, 5831.99999999845},
                                                  {0, 0, 0, 39.999999999988, 0, 0, -507.999999999852, 0, -148.000000000004, -147.999999999995, 0, 0, 2231.99999999935, 0, 179.999999999987, 0, 180.00000000002, 1691.99999999996, 1692.00000000005, 359.99999999994, 0, 0, -4355.9999999987, 0, -71.9999999999757, 0, -71.9999999999911, -6119.99999999986, -6120.00000000008, 0, -1835.99999999991, -1836.00000000036, -215.999999999741, -215.999999999694, -3671.99999999986, 0, 0, 3887.99999999879, 0, 0, 0, 0, 8423.99999999982, 8424.00000000002, 0, 647.999999999849, 0, 648.00000000052, 5183.99999999987, 5184.00000000067, 0, 0, 10367.9999999999, 0, 1943.99999999942, 1943.99999999976, 0, 0, -1295.99999999958, 0, 0, 0, 0, -3887.99999999991, -3887.99999999997, 0, 0, 0, 0, -3887.99999999996, -3888.00000000035, 0, 0, -7776.00000000002, 0, -1295.99999999985, -1296.00000000049, 0, 0, 0, 0, -3887.99999999973, -3888.00000000008, 0},
                                                  {0, 0, 0, -22.4999999999917, 0, 0, 296.999999999898, 0, 49.4999999999999, 49.4999999999982, 0, 0, -1381.49999999956, 0, -26.9999999999962, 0, -27.0000000000011, -598.499999999986, -598.500000000009, -53.9999999999819, 0, 0, 2888.99999999911, 0, 0, 0, 0, 2375.99999999996, 2376.00000000002, 0, 296.999999999984, 297.000000000097, 0, 0, 593.99999999997, 0, 0, -2753.99999999918, 0, 0, 0, 0, -3725.99999999994, -3726.00000000001, 0, 0, 0, 0, -971.999999999991, -972.000000000206, 0, 0, -1944.00000000002, 0, 0, 0, 0, 0, 971.999999999714, 0, 0, 0, 0, 1943.99999999997, 1943.99999999999, 0, 0, 0, 0, 972.000000000008, 972.000000000116, 0, 0, 1944.00000000002, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 7.19999999999693, 0, 0, -97.1999999999624, 0, -7.19999999999918, -7.19999999999954, 0, 0, 467.999999999836, 0, 0, 0, 0, 89.9999999999971, 89.999999999996, 0, 0, 0, -1025.99999999967, 0, 0, 0, 0, -377.999999999993, -377.999999999994, 0, 0, 0, 0, 0, 0, 0, 0, 1036.7999999997, 0, 0, 0, 0, 647.999999999991, 647.999999999993, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -388.799999999895, 0, 0, 0, 0, -388.799999999995, -388.799999999997, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 7.19999999999993, 0, 0, 0, 0, -90.0000000000014, 0, 0, 0, 0, 0, 0, 0, 0, 0, 378, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -647.999999999997, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 388.799999999997, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 4.49999999999879, 0, 0, 0, 0, -49.4999999999924, 0, 0, -26.99999999999, 0, 0, 0, 0, 0, 0, 161.999999999989, 0, 0, 0, 0, 0, 296.999999999959, 0, 0, 0, 0, 0, 0, 0, 0, -161.999999999994, 0, 0, 0, 0, 0, -971.999999999959, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 971.999999999991, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 3.99999999999851, 0, 0, 0, 0, -35.9999999999949, 0, 0, -35.9999999999864, 0, 0, 0, 0, 0, 0, 71.9999999999905, 0, 0, 71.9999999999676, 0, 0, 323.999999999953, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -647.999999999943, 0, 0, -647.999999999918, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1295.99999999994, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 4.50000000000011, 0, 0, 0, 0, -26.9999999999992, 0, 0, -49.5000000000002, 0, 0, 0, 0, 0, 0, 0, 0, 0, 162, 0, 0, 296.999999999993, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -161.999999999996, 0, 0, 0, 0, 0, -971.999999999989, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 971.999999999997, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 7.19999999999985, 0, 0, 0, 0, 0, 0, 0, -89.9999999999983, 0, 0, 0, 0, 0, 0, 0, 0, 0, 377.999999999994, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -647.999999999991, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 388.799999999994, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 7.20000000000028, 0, 0, 0, 0, 0, 0, 0, -90.0000000000009, 0, 0, 0, 0, 0, 0, 0, 0, 0, 378.000000000002, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -648.000000000001, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 388.799999999998, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 4.49999999999836, 0, 0, 0, 0, 0, 0, 0, -49.4999999999949, 0, -26.9999999999867, 0, 0, 0, 0, 0, 0, 0, 161.999999999991, 0, 0, 0, 0, 296.99999999995, 0, 0, 0, 0, 0, 0, 0, 0, 0, -161.999999999999, 0, 0, 0, 0, 0, -971.999999999938, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 971.999999999982, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 3.99999999999884, 0, 0, 0, 0, 0, 0, 0, -35.9999999999908, 0, -35.9999999999895, 0, 0, 0, 0, 0, 0, 0, 71.9999999999799, 0, 71.9999999999678, 0, 0, 323.999999999942, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -647.99999999993, 0, -647.999999999905, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1295.99999999993, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 4.50000000000086, 0, 0, 0, 0, 0, 0, 0, -27.0000000000026, 0, -49.5000000000063, 0, 0, 0, 0, 0, 0, 0, 0, 0, 162.000000000015, 0, 0, 297.00000000002, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -162.000000000014, 0, 0, 0, 0, 0, -972.00000000004, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 972.000000000025, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 7.19999999999949, 0, 0, 0, 0, 0, 0, 0, 0, 0, -89.9999999999971, 0, 0, 0, 0, 0, 0, 0, 0, 0, 377.999999999996, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -647.999999999992, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 388.799999999996, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 539.999999999817, 0, 0, 0, 0, 0, -3077.99999999985, 0, -3077.99999999776, 0, 0, 0, -3077.99999999986, 0, 0, 0, 6426.00000000069, 0, 6425.9999999928, 0, 0, 0, 12851.9999999955, 0, 0, 12851.9999999993, 12851.9999999996, 6425.99999999966, 0, 0, 0, -5832.00000000124, 0, -5831.99999999132, 0, 0, 0, -17495.9999999974, 0, -17495.9999999885, 0, 0, 0, -17495.9999999987, -17495.9999999993, -5831.99999999954, -34991.9999999988, -17495.9999999985, -17495.9999999997, 0, 0, 0, 1944.00000000055, 0, 1943.99999999645, 0, 0, 0, 7775.99999999993, 0, 7775.99999999268, 0, 0, 0, 7775.99999999929, 7775.99999999962, 1943.9999999997, 11663.9999999952, 0, 0, 23327.9999999989, 11663.9999999986, 23327.9999999992, 11663.9999999997, 7775.99999999927, 7776.00000000019, 23327.9999999992},
                                                  {0, 0, 0, 0, 0, 0, 0, -53.9999999999784, 0, 0, 0, 0, 0, 648.000000000023, 0, 53.999999999687, 0, 0, 0, 53.9999999999951, 0, 0, 0, -2538.00000000019, 0, 0, 0, 0, 0, -593.999999999558, 0, 0, -593.999999999986, 0, 0, 0, 0, 0, 3888.00000000022, 0, 0, 0, 0, 0, 1944.00000000003, 0, 0, 0, 0, 0, 1943.99999999998, 0, 0, 0, 0, 0, 0, 0, 0, -1944.00000000007, 0, 0, 0, 0, 0, -1944.00000000019, 0, 0, 0, 0, 0, -1943.99999999999, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, -53.9999999999991, 0, 0, 0, 0, 0, 54.0000000000635, 0, 647.999999999902, 0, 0, 0, 53.9999999999956, 0, 0, 0, 0, 0, -2537.9999999996, 0, 0, 0, -593.999999999958, 0, 0, 0, -593.999999999998, 0, 0, 0, 0, 0, 0, 3887.99999999944, 0, 0, 0, 0, 0, 1943.99999999957, 0, 0, 0, 0, 1943.99999999998, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1943.99999999974, 0, 0, 0, 0, 0, -1943.99999999963, 0, 0, 0, 0, -1943.99999999999, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, -359.999999999804, 0, 0, 0, 0, 0, 3491.9999999998, 0, 1331.99999999765, 0, 0, 0, 1331.99999999996, 0, 0, 0, -9612.00000000057, 0, -1619.99999999255, 0, 0, 0, -11231.9999999953, 0, 0, -11231.9999999997, -3239.99999999987, -1619.99999999994, 0, 0, 0, 10368.0000000011, 0, 647.999999991183, 0, 0, 0, 21383.9999999973, 0, 11663.9999999879, 0, 0, 0, 21383.9999999993, 1943.99999999977, 647.999999999965, 23327.9999999996, 11663.9999999997, 1943.99999999981, 0, 0, 0, -3888.00000000046, 0, 0, 0, 0, 0, -11663.9999999999, 0, -3887.99999999235, 0, 0, 0, -11663.9999999995, 0, 0, -11663.9999999949, 0, 0, -23327.9999999995, -11663.9999999995, -11663.9999999998, 0, -3888.00000000005, 0, -11664},
                                                  {0, 0, 0, 0, 0, 0, 0, 179.999999999903, 0, 0, 0, 0, 0, -2015.99999999995, 0, -395.999999998783, 0, 0, 0, -395.999999999993, 0, 0, 0, 7020.0000000004, 0, 215.999999996083, 0, 0, 0, 3995.9999999978, 0, 0, 3995.99999999995, 431.999999999975, 215.999999999986, 0, 0, 0, -9072.0000000006, 0, 0, 0, 0, 0, -11015.9999999991, 0, -1943.99999999393, 0, 0, 0, -11015.9999999998, 0, 0, -3887.99999999993, -1944.00000000008, 0, 0, 0, 0, 3888.00000000023, 0, 0, 0, 0, 0, 7776.00000000022, 0, 0, 0, 0, 0, 7775.99999999989, 0, 0, 3887.99999999762, 0, 0, 7775.99999999987, 3888, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, -35.9999999999489, 0, 0, 0, 0, 0, 359.99999999988, 0, 251.99999999947, 0, 0, 0, 36.000000000003, 0, 0, 0, -971.999999999939, 0, -215.999999998397, 0, 0, 0, -2483.99999999872, 0, 0, -323.999999999987, -216.000000000024, 0, 0, 0, 0, 648.000000000059, 0, 0, 0, 0, 0, 6479.99999999891, 0, 1943.9999999972, 0, 0, 0, 647.999999999908, 0, 0, 1944.0000000001, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3887.99999999972, 0, 0, 0, 0, 0, 0, 0, 0, -3887.99999999864, 0, 0, -3888.00000000002, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, -35.9999999999748, 0, 0, 0, 0, 0, 251.999999999979, 0, 359.999999999682, 0, 0, 0, 36.0000000000156, 0, 0, 0, -216.000000000123, 0, -971.999999998995, 0, 0, 0, -2483.99999999927, 0, 0, -216.000000000002, -324.000000000024, 0, 0, 0, 0, 0, 0, 647.999999998806, 0, 0, 0, 1943.9999999995, 0, 6479.99999999823, 0, 0, 0, 0, 648.000000000006, 0, 1944.00000000012, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3887.9999999989, 0, 0, 0, 0, 0, 0, -3887.9999999992, 0, 0, 0, 0, -3888.00000000007, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 179.999999999968, 0, 0, 0, 0, 0, -396.000000000105, 0, -2015.99999999943, 0, 0, 0, -395.999999999992, 0, 0, 0, 216.000000000597, 0, 7019.99999999799, 0, 0, 0, 3995.99999999908, 0, 0, 431.999999999829, 3995.99999999997, 216.000000000145, 0, 0, 0, 0, 0, -9071.99999999742, 0, 0, 0, -1943.99999999988, 0, -11015.9999999971, 0, 0, 0, 0, -11015.9999999999, 0, -3887.99999999984, 0, -1944.0000000003, 0, 0, 0, 0, 0, 3887.99999999888, 0, 0, 0, 0, 0, 7775.99999999799, 0, 0, 0, 0, 7775.99999999989, 0, 3887.999999999, 0, 0, 0, 0, 7775.99999999982, 3888.00000000005, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, -359.999999999866, 0, 0, 0, 0, 0, 1331.99999999991, 0, 3491.99999999829, 0, 0, 0, 1331.99999999993, 0, 0, 0, -1620.00000000066, 0, -9611.99999999449, 0, 0, 0, -11231.9999999964, 0, 0, -3239.99999999948, -11231.9999999998, -1619.99999999999, 0, 0, 0, 648.000000001167, 0, 10367.9999999933, 0, 0, 0, 11663.9999999978, 0, 21383.9999999911, 0, 0, 0, 1943.99999999892, 21383.9999999995, 648.000000000132, 23327.9999999993, 1943.99999999907, 11664.0000000002, 0, 0, 0, 0, 0, -3887.9999999972, 0, 0, 0, -3887.99999999983, 0, -11663.9999999943, 0, 0, 0, 0, -11663.9999999997, 0, -11663.9999999962, 0, 0, -11663.9999999992, 0, -23327.9999999994, -11663.9999999999, 0, -3888.00000000053, -11663.9999999998},
                                                  {0, 0, 0, 0, 0, 0, 0, 134.999999999871, 0, 0, 0, 0, 0, -1106.99999999976, 0, -1106.99999999858, 0, 0, 0, -296.999999999992, 0, 0, 0, 1944.00000000008, 0, 1943.99999999564, 0, 0, 0, 8747.99999999662, 0, 0, 2105.99999999984, 2105.99999999996, 162.000000000103, 0, 0, 0, -972.000000000465, 0, -971.999999994875, 0, 0, 0, -13607.9999999973, 0, -13607.9999999924, 0, 0, 0, -1943.99999999948, -1943.99999999982, 0, -14580, -971.999999999928, -972.000000000118, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5831.99999999941, 0, 5831.99999999537, 0, 0, 0, 0, 0, 0, 11663.9999999964, 0, 0, 11663.9999999998, 0, 11663.9999999999, 0, 0, 0, 5832.00000000035},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 539.999999999971, 0, 0, 0, 0, -3077.9999999999, 0, 0, -3077.9999999997, 0, -3077.99999999984, 0, 0, 0, 0, 6425.99999999993, 0, 0, 6425.99999999914, 0, 0, 12851.9999999991, 0, 12851.9999999994, 6425.9999999995, 12851.9999999994, 0, 0, 0, 0, -5832.00000000003, 0, 0, -5831.99999999902, 0, 0, -17495.9999999991, 0, 0, -17495.9999999983, 0, -17495.9999999992, -5831.99999999928, -17495.999999999, -17495.9999999985, -34991.9999999982, -17495.9999999993, 0, 0, 0, 0, 1944.00000000004, 0, 0, 1943.9999999996, 0, 0, 7775.99999999972, 0, 0, 7775.99999999905, 0, 7775.99999999965, 1943.99999999962, 7775.99999999946, 0, 11663.9999999991, 0, 11663.9999999991, 23327.9999999989, 7775.99999999904, 7775.99999999978, 23327.9999999988, 11663.9999999997, 23327.9999999982},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, -53.999999999997, 0, 0, 0, 0, 647.999999999992, 0, 0, 53.9999999999783, 0, 53.9999999999945, 0, 0, 0, 0, -2537.99999999998, 0, 0, 0, 0, 0, -593.999999999991, 0, -593.999999999967, 0, 0, 0, 0, 0, 0, 3887.99999999997, 0, 0, 0, 0, 0, 1944.00000000004, 0, 0, 0, 0, 1943.99999999995, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1943.99999999998, 0, 0, 0, 0, 0, -1944.00000000003, 0, 0, 0, 0, -1943.99999999997, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, -53.9999999999993, 0, 0, 0, 0, 53.9999999999997, 0, 0, 647.999999999991, 0, 53.9999999999945, 0, 0, 0, 0, 0, 0, 0, -2537.99999999996, 0, 0, -593.999999999992, 0, 0, 0, -593.999999999951, 0, 0, 0, 0, 0, 0, 0, 3887.99999999994, 0, 0, 0, 0, 0, 1944, 0, 0, 0, 1943.99999999996, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1943.99999999997, 0, 0, 0, 0, 0, -1944.00000000001, 0, 0, 0, -1944, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, -359.999999999966, 0, 0, 0, 0, 3491.99999999988, 0, 0, 1331.99999999968, 0, 1331.99999999996, 0, 0, 0, 0, -9611.99999999987, 0, 0, -1619.99999999911, 0, 0, -11231.9999999991, 0, -11231.9999999998, -1619.99999999989, -3239.99999999985, 0, 0, 0, 0, 10368, 0, 0, 647.999999999017, 0, 0, 21383.9999999992, 0, 0, 11663.9999999982, 0, 21383.9999999996, 647.999999999935, 1943.9999999998, 11663.9999999996, 23327.9999999994, 1943.99999999988, 0, 0, 0, 0, -3888.00000000001, 0, 0, 0, 0, 0, -11663.9999999997, 0, 0, -3887.99999999901, 0, -11663.9999999998, 0, 0, 0, -11663.9999999991, 0, -11663.9999999996, -23327.9999999995, -3887.99999999982, 0, -11663.9999999996, 0, -11663.9999999995},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 179.999999999983, 0, 0, 0, 0, -2015.99999999994, 0, 0, -395.999999999852, 0, -395.999999999992, 0, 0, 0, 0, 7019.99999999991, 0, 0, 215.999999999589, 0, 0, 3995.99999999964, 0, 3995.99999999993, 216.000000000007, 431.999999999963, 0, 0, 0, 0, -9071.99999999992, 0, 0, 0, 0, 0, -11015.9999999997, 0, 0, -1943.99999999932, 0, -11015.9999999999, 0, 0, -1943.99999999998, -3887.99999999978, 0, 0, 0, 0, 0, 3887.99999999997, 0, 0, 0, 0, 0, 7775.99999999994, 0, 0, 0, 0, 7775.99999999991, 0, 0, 0, 3887.99999999973, 0, 3887.99999999994, 7775.99999999981, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, -35.9999999999901, 0, 0, 0, 0, 359.999999999954, 0, 0, 251.999999999911, 0, 36.0000000000038, 0, 0, 0, 0, -971.999999999934, 0, 0, -215.999999999763, 0, 0, -2483.99999999967, 0, -324.000000000011, 0, -216.000000000016, 0, 0, 0, 0, 647.999999999966, 0, 0, 0, 0, 0, 6479.99999999963, 0, 0, 1943.9999999994, 0, 647.999999999997, 0, 0, 0, 1944.00000000003, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3887.99999999986, 0, 0, 0, 0, 0, 0, 0, 0, -3887.99999999967, 0, 0, -3888.00000000001, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, -35.999999999996, 0, 0, 0, 0, 251.999999999982, 0, 0, 359.999999999959, 0, 35.9999999999961, 0, 0, 0, 0, -215.999999999984, 0, 0, -971.999999999897, 0, 0, -2483.99999999983, 0, -215.999999999989, 0, -323.99999999999, 0, 0, 0, 0, 0, 0, 0, 647.999999999891, 0, 0, 1943.99999999978, 0, 0, 6479.9999999997, 0, 0, 0, 647.999999999977, 0, 1943.99999999997, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3887.99999999984, 0, 0, 0, 0, 0, -3887.99999999981, 0, 0, 0, 0, 0, -3887.99999999998, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 179.999999999994, 0, 0, 0, 0, -395.999999999981, 0, 0, -2015.99999999994, 0, -395.999999999951, 0, 0, 0, 0, 215.999999999981, 0, 0, 7019.9999999998, 0, 0, 3995.99999999981, 0, 431.999999999805, 215.999999999782, 3995.99999999977, 0, 0, 0, 0, 0, 0, 0, -9071.99999999975, 0, 0, -1943.99999999972, 0, 0, -11015.9999999997, 0, 0, 0, -11015.9999999997, 0, -3887.99999999952, -1943.99999999971, 0, 0, 0, 0, 0, 0, 0, 3887.99999999988, 0, 0, 0, 0, 0, 7775.99999999986, 0, 0, 0, 7775.99999999989, 0, 3887.99999999974, 0, 0, 0, 0, 0, 7775.99999999967, 3887.99999999985, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, -359.999999999979, 0, 0, 0, 0, 1331.99999999992, 0, 0, 3491.99999999978, 0, 1331.99999999989, 0, 0, 0, 0, -1619.99999999993, 0, 0, -9611.99999999937, 0, 0, -11231.9999999992, 0, -3239.99999999958, -1619.99999999959, -11231.9999999995, 0, 0, 0, 0, 648.000000000028, 0, 0, 10367.9999999993, 0, 0, 11663.9999999991, 0, 0, 21383.9999999986, 0, 1943.99999999953, 647.999999999346, 21383.9999999993, 1943.99999999883, 23327.9999999988, 11663.9999999994, 0, 0, 0, 0, 0, 0, 0, -3887.9999999997, 0, 0, -3887.99999999965, 0, 0, -11663.9999999993, 0, 0, 0, -11663.9999999996, 0, -11663.9999999991, 0, 0, -11663.9999999993, 0, -3887.99999999994, -23327.9999999991, -11663.9999999997, -11663.9999999986},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 134.999999999978, 0, 0, 0, 0, -1106.99999999991, 0, 0, -1106.99999999979, 0, -296.999999999995, 0, 0, 0, 0, 1943.9999999999, 0, 0, 1943.99999999944, 0, 0, 8747.9999999992, 0, 2105.99999999997, 161.999999999985, 2106, 0, 0, 0, 0, -971.999999999995, 0, 0, -971.999999999408, 0, 0, -13607.9999999991, 0, 0, -13607.9999999985, 0, -1943.99999999992, 0, -1943.99999999999, -971.99999999992, -14579.9999999999, -972.000000000025, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5831.99999999966, 0, 0, 5831.9999999992, 0, 0, 0, 0, 0, 11663.9999999992, 0, 0, 11663.9999999999, 0, 0, 11664, 0, 5831.99999999995},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -54.0000000000051, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 594.000000000013, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1944.00000000001, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1944, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -53.9999999999992, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 593.999999999995, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1943.99999999998, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1943.99999999999, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -53.9999999999951, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 593.999999999955, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1943.99999999996, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1944, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -35.9999999999859, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 323.999999999959, 215.999999999961, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -647.999999999962, 0, 0, -1943.99999999989, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3887.99999999995, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -35.9999999999969, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 215.999999999979, 323.999999999962, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -647.999999999975, 0, -1943.99999999988, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3887.99999999991, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -35.9999999999942, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 323.999999999999, 215.999999999942, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -647.999999999985, 0, 0, 0, -1943.99999999999, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3888.00000000003, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -35.9999999999902, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 215.999999999976, 323.999999999983, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -647.999999999918, 0, 0, -1943.99999999993, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3887.99999999993, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -35.9999999999962, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 215.99999999999, 0, 323.999999999992, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -647.999999999985, 0, -1943.99999999997, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3887.99999999998, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -36.000000000003, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 324.000000000009, 0, 216.000000000013, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -648, 0, 0, 0, -1944.00000000003, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3888, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -26.9999999999943, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 161.999999999968, 161.999999999984, 161.999999999969, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -971.999999999956, -971.999999999913, -971.999999999971, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5831.99999999995},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 539.999999999987, 0, 0, 0, 0, 0, 0, 0, -3077.99999999997, 0, -3077.99999999983, -3077.99999999983, 0, 0, 0, 0, 0, 0, 6425.99999999993, 0, 6425.99999999938, 0, 0, 12851.9999999998, 6425.9999999993, 12851.9999999992, 12851.9999999995, 0, 0, 0, 0, 0, 0, -5831.99999999989, 0, -5831.99999999912, 0, 0, 0, -17495.9999999999, 0, -17495.9999999995, -5831.99999999922, -17495.9999999986, -17495.9999999995, -17495.9999999977, -17495.9999999982, -34991.999999999, 0, 0, 0, 0, 0, 0, 1943.99999999995, 0, 1943.99999999959, 0, 0, 0, 7775.99999999992, 0, 7775.99999999961, 1943.99999999973, 7775.9999999993, 7775.9999999997, 0, 0, 11664, 7775.99999999872, 7775.99999999885, 11663.9999999983, 23327.9999999993, 11663.9999999989, 23327.9999999999, 23327.9999999975},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, -53.9999999999983, 0, 0, 0, 0, 0, 0, 0, 647.999999999999, 0, 53.9999999999824, 53.9999999999979, 0, 0, 0, 0, 0, 0, -2538.00000000001, 0, 0, 0, 0, -593.999999999975, 0, -593.999999999947, 0, 0, 0, 0, 0, 0, 0, 3888.00000000001, 0, 0, 0, 0, 0, 1944, 0, 0, 0, 1943.9999999999, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1944.00000000001, 0, 0, 0, 0, 0, -1944.00000000002, 0, 0, 0, -1943.99999999996, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, -54.0000000000018, 0, 0, 0, 0, 0, 0, 0, 54.0000000000079, 0, 648.00000000002, 53.999999999991, 0, 0, 0, 0, 0, 0, 0, 0, -2538.00000000004, 0, 0, -594.000000000133, 0, 0, -594.00000000002, 0, 0, 0, 0, 0, 0, 0, 0, 3888.00000000002, 0, 0, 0, 0, 0, 1944.00000000026, 0, 0, 1944.00000000008, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1944, 0, 0, 0, 0, 0, -1944.00000000015, 0, 0, -1944.00000000005, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, -359.999999999978, 0, 0, 0, 0, 0, 0, 0, 3491.99999999992, 0, 1331.99999999975, 1331.99999999993, 0, 0, 0, 0, 0, 0, -9611.9999999999, 0, -1619.99999999925, 0, 0, -11231.9999999993, -1619.99999999972, -11231.9999999995, -3239.99999999985, 0, 0, 0, 0, 0, 0, 10367.9999999999, 0, 647.999999999086, 0, 0, 0, 21383.9999999992, 0, 11663.9999999987, 647.99999999977, 21383.9999999992, 1943.99999999995, 11663.9999999988, 1943.99999999934, 23327.9999999994, 0, 0, 0, 0, 0, 0, -3888, 0, 0, 0, 0, 0, -11663.9999999997, 0, -3887.99999999926, 0, -11663.9999999996, 0, 0, 0, -11663.9999999993, -3887.99999999935, 0, -11663.9999999989, -23327.9999999993, 0, -11664, -11663.9999999988},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 179.99999999999, 0, 0, 0, 0, 0, 0, 0, -2015.99999999997, 0, -395.999999999887, -395.999999999991, 0, 0, 0, 0, 0, 0, 7019.99999999997, 0, 215.999999999679, 0, 0, 3995.99999999971, 216.000000000004, 3995.99999999981, 432.000000000022, 0, 0, 0, 0, 0, 0, -9072.00000000001, 0, 0, 0, 0, 0, -11015.9999999997, 0, -1943.99999999945, 0, -11015.9999999996, 0, -1943.99999999966, 0, -3887.99999999973, 0, 0, 0, 0, 0, 0, 3888.00000000001, 0, 0, 0, 0, 0, 7775.99999999992, 0, 0, 0, 7775.99999999985, 0, 0, 0, 3887.99999999972, 0, 0, 3887.99999999955, 7775.99999999962, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, -35.999999999991, 0, 0, 0, 0, 0, 0, 0, 359.999999999952, 0, 251.999999999917, 36.000000000006, 0, 0, 0, 0, 0, 0, -971.999999999922, 0, -215.999999999781, 0, 0, -2483.99999999966, 0, -323.999999999989, -216.000000000054, 0, 0, 0, 0, 0, 0, 647.999999999953, 0, 0, 0, 0, 0, 6479.99999999956, 0, 1943.9999999994, 0, 647.999999999982, 0, 0, 0, 1944.00000000001, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3887.99999999982, 0, 0, 0, 0, 0, 0, 0, -3887.99999999963, 0, 0, 0, -3887.99999999994, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, -35.9999999999979, 0, 0, 0, 0, 0, 0, 0, 251.999999999979, 0, 359.99999999998, 36.0000000000077, 0, 0, 0, 0, 0, 0, -215.999999999943, 0, -971.999999999943, 0, 0, -2483.99999999988, 0, -216.000000000003, -324.00000000007, 0, 0, 0, 0, 0, 0, 0, 0, 647.999999999927, 0, 0, 0, 1943.99999999979, 0, 6479.99999999984, 0, 0, 648.000000000145, 0, 0, 1944.00000000004, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3887.99999999991, 0, 0, 0, 0, 0, -3887.99999999988, 0, 0, 0, 0, 0, -3888.00000000011, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 180.000000000004, 0, 0, 0, 0, 0, 0, 0, -396.000000000015, 0, -2016.00000000004, -395.999999999957, 0, 0, 0, 0, 0, 0, 215.999999999966, 0, 7020.00000000006, 0, 0, 3996.00000000033, 215.999999999756, 431.999999999731, 3995.99999999998, 0, 0, 0, 0, 0, 0, 0, 0, -9071.99999999998, 0, 0, 0, -1944.00000000041, 0, -11016.0000000006, 0, 0, -11016.0000000001, 0, -1943.99999999954, -3887.99999999991, 0, 0, 0, 0, 0, 0, 0, 0, 3887.99999999995, 0, 0, 0, 0, 0, 7776.00000000031, 0, 0, 7776.00000000009, 0, 0, 3888.00000000045, 0, 0, 0, 0, 3887.99999999983, 7776.00000000027, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, -359.999999999994, 0, 0, 0, 0, 0, 0, 0, 1331.99999999997, 0, 3491.99999999993, 1331.9999999999, 0, 0, 0, 0, 0, 0, -1619.99999999986, 0, -9611.99999999971, 0, 0, -11232, -1619.99999999946, -3239.99999999945, -11231.9999999998, 0, 0, 0, 0, 0, 0, 647.999999999784, 0, 10367.9999999995, 0, 0, 0, 11664, 0, 21384, 647.999999999375, 1943.99999999909, 21383.9999999999, 1943.99999999831, 11663.9999999986, 23327.9999999995, 0, 0, 0, 0, 0, 0, 0, 0, -3887.99999999976, 0, 0, 0, -3887.99999999993, 0, -11663.9999999999, 0, 0, -11663.9999999999, 0, 0, -11664.0000000001, 0, -3887.9999999991, 0, -11663.9999999997, -11663.9999999993, -23328.0000000002, -11663.999999998},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 134.999999999982, 0, 0, 0, 0, 0, 0, 0, -1106.9999999999, 0, -1106.99999999983, -296.999999999995, 0, 0, 0, 0, 0, 0, 1943.9999999998, 0, 1943.99999999952, 0, 0, 8747.99999999927, 161.999999999886, 2105.99999999992, 2106.00000000005, 0, 0, 0, 0, 0, 0, -971.999999999839, 0, -971.999999999444, 0, 0, 0, -13607.999999999, 0, -13607.9999999987, 0, -1943.99999999988, -1944.00000000018, -971.999999999615, -971.999999999695, -14579.9999999999, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5831.9999999995, 0, 5831.99999999931, 0, 0, 0, 0, 0, 11663.9999999992, 0, 0, 0, 11663.9999999998, 0, 11664.0000000001, 5831.99999999944},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4319.99999999972, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -15983.9999999989, -15983.9999999991, -15983.9999999988, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 19439.9999999987, 19439.9999999986, 19439.9999999981, 38879.9999999973, 38879.9999999968, 38879.9999999987, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7775.99999999951, -7775.99999999925, -7775.99999999896, 0, 0, 0, -23327.9999999984, -23327.9999999981, -23327.9999999981, -23327.9999999995, -23327.9999999977, -23327.9999999995, -46655.9999999971},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 432.000000000002, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4319.99999999999, -432.000000000017, -432.000000000006, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 11664, 0, 0, 3888.00000000001, 3888, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7775.99999999998, 0, 0, 0, 0, 0, -7775.99999999998, -7775.99999999998, 0, 0, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 431.999999999977, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -431.999999999881, -4319.9999999999, -431.999999999987, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 11663.9999999998, 0, 3887.99999999958, 0, 3888.00000000011, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7775.99999999989, 0, 0, 0, 0, 0, 0, -7775.99999999967, -7776.00000000006, 0, 0, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 431.999999999943, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -431.999999999803, -431.999999999761, -4319.99999999976, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 11663.9999999996, 0, 3887.99999999937, 3887.99999999965, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7775.99999999982, 0, 0, 0, 0, 0, 0, 0, -7775.99999999955, -7775.99999999979, 0},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1619.99999999991, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 13283.9999999996, 3563.9999999997, 3563.99999999969, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -23327.9999999995, -1943.99999999959, -1943.99999999953, -25271.999999999, -25271.9999999991, -3887.99999999948, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 11663.9999999998, 0, 0, 0, 0, 0, 23327.9999999993, 23327.9999999994, 11663.9999999994, 0, 11663.9999999995, 0, 23327.999999999},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 323.999999999926, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2267.99999999972, -2267.99999999971, -323.99999999972, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1943.99999999966, 1943.99999999957, 0, 15551.9999999992, 1943.99999999931, 1943.99999999949, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -11663.9999999995, 0, -11663.9999999994, 0, 0, 0, -11663.9999999992},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1619.99999999983, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3563.99999999936, 13283.9999999994, 3563.99999999936, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1943.99999999925, -23327.9999999991, -1943.99999999898, -25271.9999999983, -3887.99999999822, -25271.9999999994, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 11663.9999999995, 0, 0, 0, 0, 11663.999999999, 0, 23327.9999999988, 23327.9999999997, 0, 11663.9999999999, 23327.9999999984},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1619.99999999985, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3563.99999999945, 3563.9999999995, 13283.9999999993, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1943.99999999941, -1943.99999999924, -23327.9999999988, -3887.99999999872, -25271.999999998, -25271.9999999992, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 11663.9999999993, 0, 0, 0, 0, 11663.9999999989, 0, 11663.9999999999, 23327.9999999985, 23327.9999999995, 23327.9999999983},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 323.999999999989, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2267.99999999994, -323.999999999967, -2267.99999999997, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1943.9999999999, 0, 1943.99999999994, 1943.99999999987, 15551.9999999998, 1943.99999999999, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -11663.9999999999, 0, 0, -11663.9999999999, 0, -11663.9999999999},
                                                  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 323.999999999948, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -323.999999999833, -2267.99999999991, -2267.99999999969, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1943.99999999988, 1943.99999999943, 1943.99999999982, 1943.99999999922, 15551.9999999998, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -11664, 0, -11663.9999999998, -11663.9999999996}};

// -----------------------------------------------------------------------------
double MAd_TetInterpolant::coef7[120][120] = {{1, -18.1500000000015, -18.1500000000256, -18.1500000000015, 127.672222222242, 127.672222222634, 127.672222222242, 255.344444444599, 255.344444444557, 255.344444444465, -460.668055555666, -460.668055557955, -460.668055555666, -1382.00416666608, -1382.00416666683, -1382.00416666977, -1382.0041666669, -1382.00416666673, -1382.00416666813, -2764.00833333446, 933.722222222539, 933.722222228936, 933.722222222539, 3734.88888888467, 3734.88888888973, 3734.8888889052, 3734.88888888851, 3734.88888888935, 3734.88888889572, 5602.33333333608, 5602.33333333281, 5602.33333333691, 11204.6666666706, 11204.6666666722, 11204.6666666768, -1073.78055555606, -1073.78055556528, -1073.78055555606, -5368.90277776884, -5368.90277777967, -5368.90277781387, -5368.90277777579, -5368.90277777904, -5368.90277779165, -10737.8055555495, -10737.8055555538, -10737.8055555753, -10737.8055555573, -10737.8055555541, -10737.8055555683, -21475.6111111184, -21475.6111111153, -21475.611111138, -32213.4166666861, -32213.416666688, -32213.4166667176, 653.605555555937, 653.605555562575, 653.605555555937, 3921.63333332496, 3921.6333333354, 3921.63333336895, 3921.63333333065, 3921.63333333461, 3921.63333334639, 9804.08333332204, 9804.08333333191, 9804.08333336874, 9804.08333333079, 9804.08333333293, 9804.08333335202, 19608.1666666731, 19608.1666666603, 19608.1666666961, 13072.1111111133, 13072.1111111063, 13072.1111111176, 39216.3333333579, 39216.3333333506, 39216.3333333487, 39216.3333333998, 39216.3333333621, 39216.3333334167, 58824.5000001004, -163.401388889, -163.401388890886, -163.401388889, -1143.80972221926, -1143.80972222299, -1143.80972223516, -1143.80972222112, -1143.8097222227, -1143.80972222682, -3431.42916666123, -3431.42916666655, -3431.42916668614, -3431.42916666474, -3431.42916666692, -3431.42916667601, -6862.8583333358, -6862.85833332688, -6862.85833334584, -5719.04861110728, -5719.04861110813, -5719.04861111793, -5719.04861111094, -5719.04861110899, -5719.04861111685, -17157.1458333434, -17157.1458333379, -17157.145833332, -17157.1458333566, -17157.1458333458, -17157.1458333736, -22876.1944444563, -22876.1944444536, -22876.1944445049, -34314.2916667222, -34314.2916667386, -34314.2916667414}, 
                                                    {0, 1.00000000000176, 0, 0, -17.1500000000265, 0, 0, 0, 0, 0, 110.522222222377, 0, 0, 0, 0, 0, 0, 0, 0, 0, -350.145833333779, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 583.576388889557, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -490.204166667166, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 163.401388889035, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 1.00000000000172, 0, 0, -17.1500000000264, 0, 0, 0, 0, 0, 110.522222222377, 0, 0, 0, 0, 0, 0, 0, 0, 0, -350.145833333784, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 583.576388889572, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -490.204166667179, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 163.40138888904, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 1.00000000000176, 0, 0, -17.1500000000265, 0, 0, 0, 0, 0, 110.522222222377, 0, 0, 0, 0, 0, 0, 0, 0, 0, -350.145833333779, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 583.576388889557, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -490.204166667166, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 163.401388889035, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 49.0000000000113, 0, 0, -546.35000000016, 0, 0, -546.350000000297, 0, -546.350000000038, 2431.4888888898, 0, 0, 4862.97777777367, 4862.97777777822, 2431.48888889717, 0, 2431.48888888871, 0, 4862.97777778043, -5552.31250000263, -2.75129324382076e-08, 0, -16656.9374999794, -16656.9375000025, -5552.31250004783, 0, -5552.31249999966, -1.33998112538848e-08, -16656.9374999926, -16656.9374999973, -1.92439166193407e-08, -33313.8750000113, -16656.9375000171, -16656.9375000171, 6886.20138889293, 4.02111875404451e-08, 0, 27544.8055555157, 27544.8055555613, 6886.20138900194, 0, 6886.20138888885, 2.29961364036202e-08, 41317.208333281, 41317.2083333274, 27544.8055555719, 3.72496128525934e-08, 27544.8055555467, 5.01893019238824e-08, 82634.4166666937, 27544.8055555782, 27544.8055555986, 82634.4166667356, 82634.4166667025, 41317.2083334513, -4411.83750000305, -2.93144145028321e-08, 0, -22059.1874999639, -22059.1875000059, -4411.83750011776, 0, -4411.83750000034, -1.8626174927593e-08, -44118.3749999369, -44118.3749999964, -22059.1875000614, -2.83561486666262e-08, -22059.1874999915, -5.77627717516132e-08, -88236.750000027, -22059.1875000006, -22059.1875000419, -44118.3749999408, -44118.3749999818, -6.84392897182928e-08, -132355.125000094, -132355.125000036, -88236.7500000811, -44118.375000173, -88236.7500000295, -44118.375000183, -132355.12500023, 1143.80972222312, 0, 0, 6862.8583333208, 6862.85833333548, 1143.80972226695, 0, 1143.80972222238, 0, 17157.1458333078, 17157.1458333333, 6862.85833337702, 0, 6862.85833333053, 2.40480208777032e-08, 34314.2916666774, 6862.85833332281, 6862.8583333491, 22876.1944444061, 22876.1944444349, 17157.1458333145, 2.91726266383205e-08, 17157.1458333243, 3.74719445580192e-08, 68628.5833333772, 68628.5833333477, 34314.2916666898, 17157.145833401, 34314.2916666715, 17157.1458334204, 68628.5833334005, 68628.5833333379, 22876.1944445961, 102942.875000138, 68628.5833335145, 68628.5833334791}, 
                                                    {0, -73.500000000036, 0, 0, 1076.77500000052, 0, 0, 562.275000000061, 0, 562.275000000038, -5615.19583333634, -1.89287297796808e-08, 0, -7294.46666665737, -7294.46666666723, -1679.27083334182, 0, -1679.27083333271, 0, -3358.54166666931, 14205.9166666754, 5.38712672378669e-08, 0, 30862.8541666281, 30862.8541666699, 2451.02083339254, 0, 2451.02083333087, 1.27543723671819e-08, 19107.9583332936, 19107.9583333283, 2.11562776684769e-08, 38215.9166666798, 7353.06250001834, 7353.0625000106, -18907.8750000132, -7.96022381711106e-08, 0, -58474.3541665977, -58474.3541666741, -1750.72916682231, 0, -1750.72916666195, -1.92803217151713e-08, -61975.8124998707, -61975.8124999924, -24160.0624999351, -3.67887025539984e-08, -24160.0624999806, -5.3134662360343e-08, -123951.625000038, -7002.91666668966, -7002.91666669719, -72480.1875000877, -72480.1875000271, -10504.3750001162, 12745.3083333432, 5.8585013146884e-08, 0, 51471.4374999399, 51471.437500007, 490.204166840978, 0, 490.204166662814, 1.30934691013487e-08, 78432.6666665382, 78432.6666666647, 14706.1249999885, 1.76683720736991e-08, 14706.1249999757, 6.09081100940491e-08, 156865.333333374, 2451.02083333528, 2451.02083335584, 53922.458333128, 53922.4583333022, 7.90140041319588e-08, 161767.375000115, 161767.375000037, 58824.5000001138, 4902.04166683884, 58824.4999999985, 4902.04166684273, 88236.7500002446, -3431.42916666958, -1.6977898523164e-08, 0, -17157.1458333128, -17157.1458333356, -6.98877853403497e-08, 0, 0, 0, -34314.291666621, -34314.2916666681, -3431.4291666934, 0, -3431.42916665663, -2.45001467563157e-08, -68628.5833333503, 1.3187391649435e-08, 0, -34314.2916665669, -34314.2916666532, -17157.1458332301, -3.09456288229184e-08, -17157.1458333155, -4.4349824608968e-08, -102942.875000057, -102942.875000021, -17157.145833377, -6.30973077183137e-08, -17157.1458333144, -8.41794785918298e-08, -68628.5833334315, -68628.5833333217, -1.64035960681515e-07, -102942.875000157, -34314.2916668684, -34314.2916667996}, 
                                                    {0, 81.6666666667289, 0, 0, -1291.69444444536, 0, 0, -434.194444444182, 0, -434.194444444461, 7412.6111111164, 2.01544446226219e-08, 0, 6264.51388887872, 6264.51388888926, 852.736111115066, 0, 852.736111110179, 0, 1705.47222222413, -20325.1319444597, -5.80520538887482e-08, 0, -30746.1388888517, -30746.1388888913, -733.638888929482, 0, -733.638888884637, 0, -11154.6458332766, -11154.6458333275, -1.71143505732691e-08, -22309.2916666763, -2200.91666668136, -2200.91666666936, 28828.6736111342, 8.66501596177961e-08, 0, 65827.4166666047, 65827.4166666721, 233.430555678896, 0, 233.430555546403, 1.16880867624858e-08, 45402.2430554084, 45402.2430555499, 8636.93055542666, 2.78716004453403e-08, 8636.93055553051, 3.82250280903449e-08, 90804.4861111443, 933.722222242245, 933.722222238191, 25910.7916667359, 25910.7916666812, 1400.58333342021, -20425.1736111284, -6.42514050540478e-08, 0, -63726.5416666145, -63726.5416666709, -1.49781328740523e-07, 0, 0, 0, -68628.5833332037, -68628.5833333352, -2451.02083323773, 0, -2451.02083329763, -4.41241442869163e-08, -137257.166666701, -1.29989255674896e-08, 0, -27778.236110843, -27778.2361110776, -6.39677683270525e-08, -83334.7083334114, -83334.7083333604, -9804.08333342353, -1.25720597575219e-07, -9804.08333331538, -1.23563668644071e-07, -14706.1250001883, 5719.04861111619, 1.87400222914013e-08, 0, 22876.1944444269, 22876.1944444455, 6.33984795545526e-08, 0, 0, 0, 34314.2916666255, 34314.2916666695, 0, 0, -1.57400236276905e-08, 1.71035374638076e-08, 68628.5833333482, 0, 0, 22876.1944443286, 22876.1944444329, 5719.04861096085, 2.38455781223963e-08, 5719.04861108917, 3.5748832983951e-08, 68628.5833333753, 68628.5833333519, 3.86697151441712e-08, 4.56485302559303e-08, -2.96233775046048e-08, 5.9595300243348e-08, 22876.1944445254, 22876.1944444261, 1.31095448738944e-07, 34314.2916667931, 1.56869287731024e-07, 9.37381530898415e-08}, 
                                                    {0, -61.2500000000633, 0, 0, 1004.50000000094, 0, 0, 218.458333333058, 0, 218.458333333335, -6062.04861111656, -1.21160596680637e-08, 0, -3318.04861110506, -3318.04861111129, -257.250000000493, 0, -257.249999999216, 0, -514.500000000971, 17607.3333333491, 3.53419094535961e-08, 0, 17607.333333313, 17607.3333333347, 100.041666682954, 0, 100.041666662892, 0, 3601.49999996155, 3601.49999999598, 0, 7203.00000000529, 300.125000007413, 300.12499999855, -26377.6527778015, -5.33032915287615e-08, 0, -41550.6388888569, -41550.638888892, -5.90327138650261e-08, 0, 0, 0, -16456.854166578, -16456.8541666648, -1283.86805545833, -1.30594983593588e-08, -1283.8680555371, -1.46264743416149e-08, -32913.7083333535, -1.04146614504053e-08, 0, -3851.60416670099, -3851.60416667301, -3.88175060355073e-08, 19608.1666666844, 3.98075300102848e-08, 0, 44118.3749999739, 44118.3750000018, 7.75502110914727e-08, 0, 0, 0, 29412.2499999281, 29412.2500000039, -8.61092822665432e-08, 0, -2.85420970640445e-08, 1.78840489588502e-08, 58824.5000000199, 1.42275368347655e-08, 0, 4902.04166649226, 4902.04166664536, 2.91220615741538e-08, 14706.1250000305, 14706.1250000181, 4.17929418946398e-08, 5.06518133345458e-08, -1.12856270267985e-08, 4.34479416652917e-08, 8.69002312891798e-08, -5719.04861111631, -1.16876023392961e-08, 0, -17157.1458333247, -17157.1458333335, -3.44956391288022e-08, 0, 0, 0, -17157.1458333126, -17157.1458333359, 1.83355980895427e-08, 0, 1.28927437693802e-08, 0, -34314.2916666753, 0, 0, -5719.04861104129, -5719.04861110542, 1.03038656606282e-07, -1.07059642866666e-08, 1.54591829135384e-08, -1.53451816249537e-08, -17157.1458333533, -17157.1458333464, -1.86182212147525e-08, -1.99093034556906e-08, 1.91398004407829e-08, -2.0776920589372e-08, -4.193328354637e-08, 0, -5.82716568046972e-08, -6.50949152088522e-08, -7.18081162373064e-08, -3.77653788487073e-08}, 
                                                    {0, 29.4000000000381, 0, 0, -492.450000000568, 0, 0, -63.6999999998881, 0, -63.7000000000082, 3064.13333333664, 0, 0, 997.558333331429, 997.558333333462, 34.299999999747, 0, 34.2999999997178, 0, 68.6000000003747, -9253.8541666762, -1.16370153631889e-08, 0, -5552.31249999395, -5552.31250000076, 0, 0, 0, 0, -500.2083333201, -500.208333332144, 0, -1000.41666666898, 0, 0, 14496.0375000143, 1.77231816727683e-08, 0, 14005.8333333243, 14005.833333335, 1.59576125335998e-08, 0, 0, 0, 2451.02083330504, 2451.02083333418, -3.53802055926095e-08, 0, 0, 0, 4902.04166667471, 0, 0, 1.09770157280339e-08, 0, 0, -11274.6958333441, -1.33249054587867e-08, 0, -16176.7374999927, -16176.7375000009, -2.25222382601818e-08, 0, 0, 0, -4902.04166664485, -4902.04166666977, 3.396731737028e-08, 0, 1.08069490681045e-08, 0, -9804.08333334071, 0, 0, 5.85353921871067e-08, 0, 0, 0, 0, -1.22134935884623e-08, 0, 0, 0, -2.36166286210327e-08, 3431.4291666698, 0, 0, 6862.85833333096, 6862.85833333347, 1.0520634682176e-08, 0, 0, 0, 3431.42916666078, 3431.42916666804, 0, 0, 0, 0, 6862.85833333666, 0, 0, -2.2014735717811e-08, 0, -3.57083541321576e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.39239626550362e-08, 0, 1.08554872863074e-08, 1.88174138909562e-08, 1.85491508818217e-08, 0}, 
                                                    {0, -8.16666666667925, 0, 0, 138.697222222411, 0, 0, 8.16666666665151, 0, 8.16666666667343, -880.843055556653, 0, 0, -130.530555555329, -130.530555555616, 0, 0, 0, 0, 0, 2734.47222222539, 0, 0, 750.312499999289, 750.312500000279, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4435.18055556032, 0, 0, -1984.15972222124, -1984.15972222282, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3594.83055555911, 0, 0, 2451.02083333252, 2451.0208333337, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1143.80972222326, 0, 0, -1143.80972222197, -1143.80972222231, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, -8.1666666666407, 0, 0, 0, 0, 0, 130.530555555233, 0, 0, 0, 0, 0, 0, 0, 0, 0, -750.312499998951, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1984.15972222066, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2451.02083333205, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1143.80972222182, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, -4.90000000002005, 0, 0, 0, 0, 0, 71.4583333332206, 0, 34.3000000003975, 0, 0, 0, 0, 0, 0, 0, -350.145833332884, 0, 0, 0, 0, 0, -500.208333333032, 0, 0, 0, 0, 0, 0, 0, 0, 700.291666665794, 0, 0, 0, 0, 0, 2451.02083333148, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -490.204166665965, 0, 0, 0, 0, 0, -4902.04166666558, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3431.42916666635, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, -4.08333333334147, 0, 0, 0, 0, 0, 52.4027777774992, 0, 42.8750000003119, 0, 0, 0, 0, 0, 0, 0, -200.083333331571, 0, -100.041666668758, 0, 0, 0, -550.229166666513, 0, 0, 0, 0, 0, 0, 0, 0, 233.430555551916, 0, 0, 0, 0, 0, 2100.87499999668, 0, 1283.86805555743, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2451.0208333269, 0, 0, 0, 0, 0, 0, 0, 0, -4902.04166666439, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5719.04861110945, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, -4.08333333336335, 0, 0, 0, 0, 0, 42.8750000000295, 0, 52.4027777783095, 0, 0, 0, 0, 0, 0, 0, -100.041666666456, 0, -200.0833333354, 0, 0, 0, -550.229166668199, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 233.430555560009, 0, 0, 0, 1283.868055557, 0, 2100.87500000669, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2451.0208333433, 0, 0, 0, 0, 0, 0, -4902.04166667388, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5719.04861111652, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, -4.90000000001068, 0, 0, 0, 0, 0, 34.3000000000035, 0, 71.4583333335431, 0, 0, 0, 0, 0, 0, 0, 0, 0, -350.145833334533, 0, 0, 0, -500.20833333437, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 700.291666668585, 0, 0, 0, 0, 0, 2451.02083333765, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -490.204166667634, 0, 0, 0, 0, 0, -4902.04166667211, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3431.42916666893, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, -8.1666666666594, 0, 0, 0, 0, 0, 0, 0, 130.5305555555, 0, 0, 0, 0, 0, 0, 0, 0, 0, -750.312499999802, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1984.15972222207, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2451.02083333343, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1143.80972222223, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, -8.16666666667811, 0, 0, 138.697222222401, 0, 8.16666666666587, 8.16666666666372, 0, 0, -880.843055556618, 0, 0, 0, -130.530555555562, -130.530555555539, 0, 0, 0, 0, 2734.47222222531, 0, 0, 0, 750.312500000093, 750.312499999875, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4435.1805555602, 0, 0, 0, -1984.15972222247, -1984.15972222187, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3594.83055555902, 0, 0, 0, 2451.02083333349, 2451.02083333313, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1143.80972222324, 0, 0, 0, -1143.80972222221, -1143.80972222211, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 29.4000000000313, 0, 0, -492.450000000506, 0, -63.6999999999965, -63.6999999999982, 0, 0, 3064.13333333636, 0, 34.2999999999755, 0, 997.558333333517, 997.558333333409, 0, 34.2999999998301, 68.6000000000349, 0, -9253.8541666755, 0, 0, 0, -5552.31250000158, -5552.31250000044, 0, 0, -500.208333333085, 0, -500.208333332758, 0, -1000.4166666672, 0, 0, 14496.0375000132, 0, 0, 0, 14005.8333333372, 14005.8333333339, 0, 0, 0, 0, 2451.02083333603, 2451.02083333287, 0, 0, 0, 4902.04166666797, 0, 0, 0, 0, 0, -11274.6958333431, 0, 0, 0, -16176.7375000031, -16176.7375000004, 0, 0, 0, 0, -4902.04166667391, -4902.04166666781, 0, 0, 0, -9804.08333333603, 0, 0, 0, 0, 0, 0, 0, -1.02512202683778e-08, 0, -1.27172348460603e-08, 0, 0, 3431.4291666695, 0, 0, 0, 6862.8583333341, 6862.8583333333, 0, 0, 0, 0, 3431.4291666714, 3431.42916666773, 0, 0, 0, 6862.85833333452, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, -61.2500000000461, 0, 0, 1004.50000000078, 0, 218.458333333426, 218.458333333409, 0, 0, -6062.04861111587, 0, -257.2500000001, 0, -3318.04861111331, -3318.04861111208, 0, -257.250000000064, -514.500000000551, 0, 17607.3333333472, 0, 100.041666666118, 0, 17607.3333333456, 17607.3333333377, 0, 100.041666665776, 3601.50000000519, 0, 3601.50000000167, 300.124999999174, 7203.00000000497, 300.12500000558, 0, -26377.6527777983, 0, 0, 0, -41550.6388889149, -41550.6388888967, 0, 0, -1283.86805555625, 0, -16456.8541666969, -16456.8541666753, 0, -1283.86805554752, 0, -32913.7083333468, -1.71257657141693e-08, -3851.60416666119, 0, -3851.60416669728, 0, 19608.1666666816, 0, 0, 0, 44118.3750000222, 44118.3750000062, 0, 0, 0, 0, 29412.2500000504, 29412.2500000147, 0, -1.5159980685496e-08, 0, 58824.5000000166, 1.85208901182321e-08, 4902.04166668579, 0, 4902.04166666423, -2.04828421740324e-08, 0, 14706.1249999999, 14706.1250000578, 1.38855185451407e-08, 6.40934220839309e-08, 3.6827020296217e-08, 0, -5719.04861111536, 0, 0, 0, -17157.1458333399, -17157.1458333349, 0, 0, 0, 0, -17157.1458333592, -17157.145833341, 0, 0, 0, -34314.2916666736, 0, 0, 0, -5719.04861113147, -5719.04861111327, 0, 0, 0, 0, -17157.1458333303, -17157.1458333631, 0, -3.55530412061378e-08, 1.48099652990726e-08, 0, -4.50328849839335e-08, 0, -1.73768725108226e-08, -3.2506014911204e-08}, 
                                                    {0, 0, 81.6666666667117, 0, 0, -1291.69444444525, 0, -434.194444444864, -434.194444444723, 0, 0, 7412.61111111603, 0, 852.736111111661, 0, 6264.51388889673, 6264.5138888916, 0, 852.73611111243, 1705.47222222431, 0, -20325.1319444585, 0, -733.638888885672, 0, -30746.1388889282, -30746.1388889005, 0, -733.638888892922, -11154.6458333554, 0, -11154.6458333392, -2200.91666666835, -22309.291666683, -2200.91666668535, 0, 28828.6736111314, 0, 233.43055554636, 0, 65827.4166667465, 65827.4166666886, 0, 233.430555563532, 8636.93055557106, 0, 45402.2430556574, 45402.2430555754, 0, 8636.93055555098, 933.722222220344, 90804.4861111516, 933.72222227611, 25910.7916666767, 1400.58333335316, 25910.7916667655, 0, -20425.1736111254, 0, 0, 0, -63726.5416667361, -63726.5416666853, 0, 0, -2451.02083332711, 0, -68628.5833334858, -68628.5833333646, 0, -2451.02083332222, 0, -137257.166666711, -6.13802662629068e-08, -27778.2361111925, 0, -27778.2361111057, -9804.08333332145, 0, -83334.7083333512, -83334.7083334903, -5.44619991023107e-08, -9804.08333350591, -14706.1250001289, 0, 5719.04861111505, 0, 0, 0, 22876.1944444662, 22876.1944444501, 0, 0, 0, 0, 34314.2916667397, 34314.2916666832, 0, 0, 0, 68628.5833333506, 2.71144610110345e-08, 5719.04861112896, 0, 22876.1944445139, 22876.1944444456, 0, 5719.04861110094, 0, 0, 68628.5833333376, 68628.5833334109, 3.26241428157799e-08, 9.03420944160323e-08, 22876.1944444325, 1.12238342990613e-08, 22876.1944445643, 3.04157258805095e-08, 34314.2916667508, 1.1750708025134e-07}, 
                                                    {0, 0, -73.5000000000447, 0, 0, 1076.77500000078, 0, 562.275000000747, 562.275000000411, 0, 0, -5615.19583333791, 0, -1679.27083333393, 0, -7294.46666667974, -7294.46666666962, 0, -1679.27083333626, -3358.54166667053, 0, 14205.9166666792, 0, 2451.02083332579, 0, 30862.8541667303, 30862.8541666781, 0, 2451.02083334577, 19107.958333368, 0, 19107.9583333398, 7353.0625000078, 38215.9166666924, 7353.06250003396, 0, -18907.8750000175, 0, -1750.72916664527, 0, -58474.3541667953, -58474.3541666883, 0, -1750.72916669388, -24160.0625000232, 0, -61975.8125001534, -61975.8125000121, 0, -24160.0625000043, -7002.91666667393, -123951.625000053, -7002.91666676039, -72480.1875000443, -10504.3750000539, -72480.1875001695, 0, 12745.3083333452, 0, 490.204166644797, 0, 51471.4375001145, 51471.4375000186, 0, 490.204166694359, 14706.1249999877, 0, 78432.666666892, 78432.6666666866, 0, 14706.1250000084, 2451.02083333815, 156865.333333377, 2451.02083344134, 53922.4583334542, 0, 53922.4583333166, 58824.5000000315, 4902.04166668929, 161767.375000051, 161767.375000245, 4902.04166676951, 58824.5000002811, 88236.7500002717, 0, -3431.42916666983, 0, 0, 0, -17157.1458333706, -17157.1458333391, 0, -1.04276564033917e-08, -3431.42916665247, 0, -34314.2916667735, -34314.2916666786, 0, -3431.4291666742, 0, -68628.5833333467, -4.70828909733371e-08, -17157.1458333601, 0, -34314.2916667672, -34314.2916666572, 0, -17157.1458333208, -17157.1458333429, 0, -102942.875000011, -102942.87500011, -5.668892822929e-08, -17157.1458334757, -68628.583333345, -2.60732362518707e-08, -68628.5833335244, -34314.2916667779, -102942.875000192, -34314.2916668948}, 
                                                    {0, 0, 49.0000000000498, 0, 0, -546.350000000813, 0, -546.350000000573, -546.350000000316, 0, 0, 2431.48888889358, 0, 2431.48888888845, 0, 4862.97777778812, 4862.97777777912, 0, 2431.48888889212, 4862.9777777812, 0, -5552.31250001286, 0, -5552.31249999083, 0, -16656.9375000513, -16656.9375000032, 0, -5552.31250001494, -16656.9375000205, 0, -16656.937500006, -16656.93750001, -33313.8750000196, -16656.9375000305, 0, 6886.20138890702, 0, 6886.20138886648, 0, 27544.8055556627, 27544.80555556, 0, 6886.20138892094, 27544.8055555595, 0, 41317.208333432, 41317.2083333354, 0, 27544.8055555731, 27544.8055555702, 82634.4166666951, 27544.8055556369, 82634.4166667184, 41317.208333391, 82634.4166668157, 0, -4411.83750001269, 0, -4411.83749997787, 0, -22059.1875000998, -22059.1875000028, 0, -4411.83750003156, -22059.1874999797, 0, -44118.3750001518, -44118.3749999992, 0, -22059.1875000293, -22059.1875000119, -88236.7500000076, -22059.1875000919, -44118.3750000611, 0, -44118.3749999951, -88236.7500000561, -44118.3750000383, -132355.125000049, -132355.125000203, -44118.3750000913, -88236.7500002447, -132355.125000273, 0, 1143.80972222572, 0, 1143.80972221434, 0, 6862.85833336758, 6862.85833333387, 0, 1143.80972223369, 6862.85833331947, 0, 17157.1458334084, 17157.1458333333, 0, 6862.85833335018, 6862.85833333786, 34314.2916666619, 6862.85833337283, 17157.1458333415, 0, 22876.1944445011, 22876.1944444361, 0, 17157.1458333332, 34314.2916666879, 17157.1458333398, 68628.5833333377, 68628.5833334141, 17157.1458333788, 34314.291666787, 68628.5833333595, 22876.1944444718, 68628.5833335023, 68628.5833334685, 102942.875000196, 68628.5833335485}, 
                                                    {0, 0, 0, 49.0000000000113, 0, 0, -546.35000000016, 0, -546.350000000166, -546.350000000094, 0, 0, 2431.4888888898, 0, 2431.48888888957, 0, 2431.48888888782, 4862.97777777839, 4862.97777778103, 4862.97777777953, 0, 0, -5552.31250000263, 0, -5552.31250000304, 0, -5552.31249999235, -16656.9375000032, -16656.9375000172, 0, -16656.9374999996, -16656.9375000009, -16656.9375000052, -16656.9375000035, -33313.8750000208, 0, 0, 6886.20138889293, 0, 6886.20138889579, 0, 6886.20138887224, 27544.8055555633, 27544.8055555923, 0, 27544.8055555506, 0, 27544.8055555344, 41317.2083333338, 41317.2083333545, 27544.8055555612, 27544.8055555419, 82634.4166667243, 41317.2083333499, 82634.4166667153, 82634.4166667448, 0, 0, -4411.83750000305, 0, -4411.83750000736, 0, -4411.83749998419, -22059.1875000076, -22059.1875000359, 0, -22059.1874999955, 0, -22059.1874999634, -44118.3750000066, -44118.3750000408, -22059.1875000018, -22059.1874999624, -88236.7500000666, 0, -44118.3749999887, -44118.3749999891, -44118.3750000164, -88236.750000036, -44118.3749999877, -88236.750000077, -132355.125000075, -132355.125000135, -132355.125000176, 0, 0, 1143.80972222312, 0, 1143.80972222501, 0, 1143.80972221674, 6862.85833333606, 6862.8583333462, 0, 6862.858333333, 0, 6862.85833331566, 17157.145833338, 17157.1458333558, 6862.85833333324, 6862.85833331148, 34314.291666696, 0, 17157.1458333249, 0, 17157.145833316, 22876.1944444411, 22876.1944444491, 17157.1458333358, 34314.2916666725, 17157.1458333075, 34314.29166668, 68628.5833333692, 68628.5833333986, 22876.1944444371, 68628.5833333605, 68628.5833334176, 68628.5833334307, 68628.5833334412, 102942.875000142}, 
                                                    {0, 0, 0, -73.500000000036, 0, 0, 1076.77500000052, 0, 562.275000000082, 562.275000000208, 0, 0, -5615.19583333634, 0, -1679.2708333347, 0, -1679.27083333062, -7294.46666666832, -7294.46666666972, -3358.54166666654, 0, 0, 14205.9166666754, 0, 2451.02083333894, 0, 2451.0208333194, 30862.8541666743, 30862.8541666849, 0, 19107.9583333351, 19107.9583333246, 7353.06249999874, 7353.06249998593, 38215.9166666779, 0, 0, -18907.8750000132, 0, -1750.72916667872, 0, -1750.72916663984, -58474.3541666841, -58474.354166708, 0, -24160.0624999963, 0, -24160.0624999506, -61975.8125000069, -61975.8125000044, -7002.9166666618, -7002.91666660692, -123951.625000036, -10504.3749999678, -72480.187500031, -72480.1875000079, 0, 0, 12745.3083333432, 0, 490.20416667924, 0, 490.20416664305, 51471.4375000167, 51471.4375000422, 0, 14706.1249999972, 0, 14706.1249999321, 78432.6666666851, 78432.6666666943, 2451.02083332723, 2451.02083324574, 156865.333333381, 0, 53922.4583333219, 53922.4583332847, 4902.04166663221, 58824.5000000217, 4902.04166656865, 58824.4999999574, 161767.375000051, 161767.375000025, 88236.7500000533, 0, 0, -3431.42916666958, 0, 0, 0, 0, -17157.1458333393, -17157.1458333489, 0, -3431.42916666799, 0, -3431.42916663639, -34314.2916666775, -34314.2916666853, 0, 4.0821188879382e-08, -68628.5833333568, 0, -17157.1458333229, 0, -17157.1458332952, -34314.2916666646, -34314.2916666543, 1.97408976857773e-08, -17157.1458333341, 6.98335678722625e-08, -17157.1458332848, -102942.875000024, -102942.875000009, 5.60838797510273e-08, -68628.5833333532, -68628.5833333354, -34314.2916667099, -34314.2916666701, -102942.875000058}, 
                                                    {0, 0, 0, 81.6666666667289, 0, 0, -1291.69444444536, 0, -434.194444444477, -434.194444444702, 0, 0, 7412.6111111164, 0, 852.736111112639, 0, 852.736111108766, 6264.51388889117, 6264.51388889079, 1705.47222222065, 0, 0, -20325.1319444597, 0, -733.638888894624, 0, -733.638888877877, -30746.1388888986, -30746.1388889006, 0, -11154.6458333376, -11154.6458333221, -2200.91666666012, -2200.91666664466, -22309.2916666653, 0, 0, 28828.6736111342, 0, 233.430555567099, 0, 233.430555535579, 65827.416666688, 65827.4166666944, 0, 8636.93055555653, 0, 8636.93055551041, 45402.2430555674, 45402.2430555452, 933.722222209747, 933.722222151957, 90804.4861111136, 1400.58333327308, 25910.7916666682, 25910.7916666146, 0, 0, -20425.1736111284, 0, -1.17249559865057e-08, 0, 1.69255253711709e-08, -63726.5416666868, -63726.5416666964, 0, -2451.0208333357, 0, -2451.02083327745, -68628.5833333574, -68628.583333337, 1.05063855500977e-08, 8.86995505043564e-08, -137257.16666668, 0, -27778.2361111066, -27778.2361110557, 5.98274236719539e-08, -9804.08333332697, 1.26747023510956e-07, -9804.08333321453, -83334.7083333374, -83334.7083332564, -14706.1249999088, 0, 0, 5719.04861111619, 0, 0, 0, 0, 22876.1944444517, 22876.1944444558, 0, 0, 0, -2.37714164911072e-08, 34314.2916666795, 34314.2916666731, 0, -3.80943553795454e-08, 68628.5833333426, 0, 5719.04861110407, 0, 5719.04861107553, 22876.1944444447, 22876.1944444231, -2.9886152066855e-08, 0, -7.54526295646438e-08, -7.57616080907373e-08, 68628.583333337, 68628.5833332927, -7.09338214503675e-08, 22876.1944444429, 22876.1944443834, -3.00566913195786e-08, -9.47920916020126e-08, 34314.2916666155}, 
                                                    {0, 0, 0, -61.2500000000633, 0, 0, 1004.50000000094, 0, 218.45833333335, 218.458333333531, 0, 0, -6062.04861111656, 0, -257.250000001062, 0, -257.249999998762, -3318.04861111301, -3318.04861111207, -514.499999998899, 0, 0, 17607.3333333491, 0, 100.041666670216, 0, 100.041666660817, 17607.3333333409, 17607.3333333388, 0, 3601.50000000476, 3601.49999999291, 300.124999996, 300.124999985618, 7202.99999999703, 0, 0, -26377.6527778015, 0, 0, 0, 1.03927319841986e-08, -41550.638888905, -41550.6388889021, 0, -1283.86805555965, 0, -1283.86805553004, -16456.854166679, -16456.8541666582, 0, 4.38239747763932e-08, -32913.7083333278, 3.97122088841209e-08, -3851.60416666537, -3851.60416662217, 0, 0, 19608.1666666844, 0, 0, 0, 0, 44118.3750000149, 44118.3750000147, 0, 0, 0, -2.97149957142432e-08, 29412.2500000211, 29412.2499999978, 0, -5.24942746514631e-08, 58824.5000000008, 0, 4902.04166667082, 4902.0416666327, -3.31292821027498e-08, 0, -7.86433725169494e-08, -9.07025643464139e-08, 14706.1249999941, 14706.124999931, -9.6912108768881e-08, 0, 0, -5719.04861111631, 0, 0, 0, 0, -17157.1458333388, -17157.1458333389, 0, 0, 0, 1.2280991815805e-08, -17157.1458333435, -17157.1458333346, 0, 2.17434550683006e-08, -34314.2916666687, 0, 0, 0, 2.01656997617722e-08, -5719.0486111145, -5719.04861109712, 1.68561839566723e-08, 0, 4.46817207446478e-08, 5.3536516917035e-08, -17157.1458333327, -17157.1458332996, 4.19489718121939e-08, 0, 5.02742162369945e-08, 3.52402941205736e-08, 8.56033901815037e-08, 6.12865725769702e-08}, 
                                                    {0, 0, 0, 29.4000000000381, 0, 0, -492.450000000568, 0, -63.7000000000195, -63.7000000000889, 0, 0, 3064.13333333664, 0, 34.3000000004544, 0, 34.2999999996858, 997.558333334228, 997.558333333775, 68.5999999998068, 0, 0, -9253.8541666762, 0, 0, 0, 0, -5552.31250000335, -5552.31250000206, 0, -500.208333336043, -500.208333331219, 0, 0, -1000.41666666653, 0, 0, 14496.0375000143, 0, 0, 0, 0, 14005.8333333404, 14005.8333333382, 0, 0, 0, 0, 2451.02083334007, 2451.0208333314, 0, -1.34243567189005e-08, 4902.04166666635, -1.0721846936014e-08, 0, -1.36722250967042e-08, 0, 0, -11274.6958333441, 0, 0, 0, 0, -16176.7375000063, -16176.7375000054, 0, 0, 0, 0, -4902.04166667741, -4902.04166666731, 0, 1.55930334645989e-08, -9804.08333333554, 0, 0, 1.1123924413817e-08, 0, 0, 2.27592117837267e-08, 2.98928560084203e-08, 0, 2.34725029731898e-08, 3.85942081448479e-08, 0, 0, 3431.4291666698, 0, 0, 0, 0, 6862.85833333569, 6862.85833333532, 0, 0, 0, 0, 3431.42916667142, 3431.42916666747, 0, 0, 6862.85833333447, 0, 0, 0, 0, 0, 0, 0, 0, -1.37408920510735e-08, -1.81330543506068e-08, 0, -1.05471139331499e-08, -1.14704978195258e-08, 0, -1.57539890094003e-08, -1.01225601550108e-08, -3.15863361311153e-08, -2.30502602903811e-08}, 
                                                    {0, 0, 0, -8.16666666667925, 0, 0, 138.697222222411, 0, 8.16666666667562, 8.16666666668671, 0, 0, -880.843055556653, 0, 0, 0, 0, -130.530555555758, -130.530555555679, 0, 0, 0, 2734.47222222539, 0, 0, 0, 0, 750.312500000692, 750.312500000481, 0, 0, 0, 0, 0, 0, 0, 0, -4435.18055556032, 0, 0, 0, 0, -1984.15972222368, -1984.15972222328, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3594.83055555911, 0, 0, 0, 0, 2451.0208333346, 2451.02083333448, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1143.80972222326, 0, 0, 0, 0, -1143.80972222272, -1143.80972222263, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -8.16666666666261, 0, 0, 0, 0, 130.530555555519, 0, 0, 0, 0, 0, 0, 0, 0, 0, -750.312499999941, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1984.15972222225, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2451.02083333323, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1143.80972222216, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -4.89999999999724, 0, 0, 0, 0, 71.4583333333445, 0, 0, 34.2999999999926, 0, 0, 0, 0, 0, 0, -350.145833333271, 0, 0, 0, 0, 0, -500.208333333363, 0, 0, 0, 0, 0, 0, 0, 0, 700.291666666188, 0, 0, 0, 0, 0, 2451.02083333215, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -490.204166666488, 0, 0, 0, 0, 0, -4902.04166666532, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3431.42916666659, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -4.08333333334412, 0, 0, 0, 0, 52.4027777777951, 0, 0, 42.8750000000916, 0, 0, 0, 0, 0, 0, -200.083333333469, 0, 0, -100.041666667549, 0, 0, -550.229166666699, 0, 0, 0, 0, 0, 0, 0, 0, 233.430555556378, 0, 0, 0, 0, 0, 2100.87500000141, 0, 0, 1283.86805555788, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2451.02083333465, 0, 0, 0, 0, 0, 0, 0, 0, -4902.04166666798, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5719.04861111224, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -4.08333333333933, 0, 0, 0, 0, 42.8750000001241, 0, 0, 52.4027777778077, 0, 0, 0, 0, 0, 0, -100.041666667165, 0, 0, -200.083333333025, 0, 0, -550.229166667113, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 233.430555555856, 0, 0, 1283.86805555473, 0, 0, 2100.8749999995, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2451.02083333542, 0, 0, 0, 0, 0, -4902.04166666676, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5719.04861111059, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -4.89999999999475, 0, 0, 0, 0, 34.2999999999609, 0, 0, 71.4583333332402, 0, 0, 0, 0, 0, 0, 0, 0, 0, -350.145833333264, 0, 0, -500.208333332649, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 700.291666666101, 0, 0, 0, 0, 0, 2451.02083333235, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -490.204166666288, 0, 0, 0, 0, 0, -4902.04166666403, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3431.4291666659, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -8.16666666667589, 0, 0, 0, 0, 0, 0, 0, 130.530555555661, 0, 0, 0, 0, 0, 0, 0, 0, 0, -750.312500000354, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1984.1597222231, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2451.02083333412, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1143.80972222256, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -8.16666666666401, 0, 0, 0, 0, 0, 0, 0, 130.530555555538, 0, 0, 0, 0, 0, 0, 0, 0, 0, -750.312499999873, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1984.15972222191, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2451.02083333316, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1143.80972222214, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -4.89999999999372, 0, 0, 0, 0, 0, 0, 0, 71.458333333253, 0, 34.2999999999866, 0, 0, 0, 0, 0, 0, 0, -350.145833333024, 0, 0, 0, 0, -500.20833333271, 0, 0, 0, 0, 0, 0, 0, 0, 0, 700.291666666305, 0, 0, 0, 0, 0, 2451.02083333278, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -490.204166666192, 0, 0, 0, 0, 0, -4902.04166666595, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3431.42916666627, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -4.08333333333334, 0, 0, 0, 0, 0, 0, 0, 52.402777777618, 0, 42.8750000000427, 0, 0, 0, 0, 0, 0, 0, -200.083333332705, 0, -100.041666667471, 0, 0, -550.229166666381, 0, 0, 0, 0, 0, 0, 0, 0, 0, 233.430555554245, 0, 0, 0, 0, 0, 2100.87499999614, 0, 1283.86805555417, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2451.02083332802, 0, 0, 0, 0, 0, 0, 0, -4902.04166666357, 0, 0, 0, 0, 0, 0, -1.12639205989176e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5719.04861110864, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -4.0833333333297, 0, 0, 0, 0, 0, 0, 0, 42.8750000000388, 0, 52.4027777777361, 0, 0, 0, 0, 0, 0, 0, -100.041666666712, 0, -200.083333332898, 0, 0, -550.229166665365, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 233.430555555163, 0, 0, 0, 1283.86805555354, 0, 2100.87499999909, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2451.02083333364, 0, 0, 0, 0, 0, -4902.04166666184, 0, 0, 0, 0, 0, 0, 1.32184738064319e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5719.04861110893, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -4.8999999999895, 0, 0, 0, 0, 0, 0, 0, 34.2999999998464, 0, 71.4583333332169, 0, 0, 0, 0, 0, 0, 0, 0, 0, -350.145833333121, 0, 0, -500.208333332983, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 700.291666666101, 0, 0, 0, 0, 0, 2451.02083333098, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -490.204166665773, 0, 0, 0, 0, 0, -4902.04166666299, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.02967369040867e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3431.42916666526, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -8.16666666666481, 0, 0, 0, 0, 0, 0, 0, 0, 0, 130.530555555583, 0, 0, 0, 0, 0, 0, 0, 0, 0, -750.312500000143, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1984.15972222271, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2451.02083333401, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1143.80972222248, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 1029.00000000129, 0, 0, 0, -1.44026668946688e-08, 0, -7871.84999999423, 0, -7871.85000002673, 0, 0, 0, -7871.85000000608, 0, 3.91721108397445e-08, 0, 23509.7916666239, 0, 23509.7916668043, 1.67103254070873e-08, 0, 2.50290443085173e-08, 47019.5833333635, 0, 3.33834768772048e-08, 47019.5833333574, 47019.5833333727, 23509.7916667106, 0, -5.52424481535758e-08, 0, -34314.2916665727, 0, -34314.2916669634, -2.42626245192652e-08, 0, -4.66957240824618e-08, -102942.874999949, 0, -102942.875000201, -7.4971226565613e-08, 0, -7.91995700258044e-08, -102942.875000051, -102942.875000061, -34314.2916667715, -205885.750000141, -102942.875000078, -102942.875000262, 0, 3.89176658339725e-08, 0, 24510.2083332435, 0, 24510.2083336192, 1.6014101106026e-08, 0, 4.20470905654108e-08, 98040.833333218, 0, 98040.8333336741, 7.84124452504809e-08, 0, 8.93616812190808e-08, 98040.8333333845, 98040.8333333467, 24510.2083334466, 147061.250000061, 0, 1.00603674688863e-07, 294122.500000205, 147061.250000068, 294122.500000164, 147061.250000395, 98040.8333334167, 98040.8333337336, 294122.500000484, 0, -1.08396557122946e-08, 0, -6862.8583333016, 0, -6862.85833343469, 0, 0, -1.41612385745159e-08, -34314.2916666065, 0, -34314.2916668453, -2.97059425266059e-08, 0, -3.96578563461005e-08, -34314.2916666864, -34314.291666652, -6862.85833338043, -68628.5833333095, 0, -68628.5833334257, -4.90848291789393e-08, 0, -4.89067802336821e-08, -137257.16666676, -68628.5833333535, -137257.166666708, -68628.5833335024, -34314.2916666973, -34314.2916668563, -205885.75000013, -68628.5833333491, -68628.5833336489, -205885.750000275, -205885.750000385, -137257.166666987}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 68.5999999998724, 0, 0, 0, 0, 0, -1069.01666666433, 0, -68.5999999999875, 0, 0, 0, -68.6000000000673, 0, 1.26752731225397e-08, 0, 5902.45833332572, 0, 0, 0, 0, 0, 1000.41666665207, 0, 0, 1000.41666666684, 0, 0, 0, -1.91399097799864e-08, 0, -14706.1249999882, 0, -2.02514514384661e-08, 0, 0, 0, -4902.04166663388, 0, 3.68776971809005e-08, 0, 0, 0, -4902.04166666973, 0, 0, 0, 0, 0, 0, 1.43332214933867e-08, 0, 16666.9416666571, 0, 2.75054369064644e-08, 0, 0, 0, 9804.08333330758, 0, -3.28688188229701e-08, 0, 0, 0, 9804.08333333682, 0, 0, -6.40471038802944e-08, 0, 0, 0, 0, 0, 0, 0, 0, 1.69406389281587e-08, 0, 0, 0, -6862.85833333016, 0, -1.23060835963172e-08, 0, 0, 0, -6862.85833332599, 0, 0, 0, 0, 0, -6862.85833333477, 0, 0, 2.55083183505549e-08, 0, 3.79961258723081e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.22559810940731e-08, -1.76861336052517e-08, -1.51510849123416e-08, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 68.5999999999859, 0, 0, 0, 0, 0, -68.5999999998315, 0, -1069.01666666685, 0, 0, 0, -68.6000000000773, 0, 0, 0, 0, 0, 5902.4583333351, 0, 0, 0, 1000.41666666638, 0, 0, 0, 1000.41666666701, 0, 0, 0, 0, 0, 0, -14706.1250000043, 0, 0, 0, 0, 0, -4902.04166666992, 0, 0, 0, 0, -4902.04166666882, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16666.9416666705, 0, 0, 0, 0, 0, 9804.08333334084, 0, 0, 0, 0, 9804.08333333692, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.02301208639422e-08, 0, 0, 0, 0, 0, -6862.85833333442, 0, 0, 0, 0, 0, -6862.85833333763, 0, 0, 0, 0, -6862.8583333345, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, -857.500000000951, 0, 0, 0, 2.65287130178707e-08, 0, 10561.5416666528, 0, 4559.0416666936, 0, 0, 0, 4559.04166667069, 0, -7.33679333524453e-08, 0, -40867.0208332609, 0, -8953.72916681818, 0, 0, -1.55675240566182e-08, -49820.7499999819, 0, -1.8585076720989e-08, -49820.7500000224, -17907.4583333597, -8953.7291666893, 0, 1.05638528979206e-07, 0, 70379.3124998553, 0, 7703.20833368283, 0, 0, 2.8563073009162e-08, 148461.833333162, 0, 85785.7291667473, 3.36084325349343e-08, 0, 5.14139812256868e-08, 148461.833333391, 23109.6250000303, 7703.20833338018, 171571.458333475, 85785.7291667016, 23109.6250001714, 0, -7.61153259099994e-08, 0, -56373.4791665336, 0, -2451.02083369002, 0, 0, -2.48402223786755e-08, -171571.458333108, 0, -63726.5416668967, -2.73538663773768e-08, 0, -6.02852278806778e-08, -171571.458333399, -9804.083333312, -2451.02083338188, -176473.499999842, 0, -5.94647577406024e-08, -352947.000000227, -176473.500000044, -191179.62500019, -14706.1250002673, -63726.5416666599, -9804.08333358894, -191179.625000366, 0, 2.16759138561048e-08, 0, 17157.1458332869, 0, 1.32925992126154e-07, 0, 0, 0, 68628.5833332365, 0, 17157.1458334831, 0, 0, 2.69391371314352e-08, 68628.5833333602, -3.08492436546759e-08, 2.03943833484885e-08, 102942.874999879, 0, 68628.583333302, 2.47233567158543e-08, 0, 3.27941997976898e-08, 205885.750000107, 102942.875000023, 68628.5833333995, 1.03733769954228e-07, 17157.1458333141, 1.1799313244185e-07, 205885.750000161, 68628.5833333152, 2.20798779109569e-07, 205885.750000234, 102942.875000326, 68628.5833335335}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 571.666666666769, 0, 0, 0, -2.67527573432704e-08, 0, -8041.44444443002, 0, -2038.94444445774, 0, 0, 0, -2038.94444444664, 0, 7.53417473147281e-08, 0, 37815.7499999391, 0, 2401.00000008989, 0, 0, 0, 26210.9166666081, 0, 1.27263979230245e-08, 26210.9166666804, 4802.00000001764, 2401.00000000858, 0, -1.10234731762187e-07, 0, -76098.3611110004, 0, -933.722222454241, 0, 0, -1.21318250260264e-08, -103176.305555356, 0, -28011.6666665779, -2.11604678871327e-08, 0, -3.90341325332612e-08, -103176.305555594, -2801.16666668855, -933.722222233192, -56023.3333334295, -28011.6666666776, -2801.16666677821, 0, 8.05466723301702e-08, 0, 68628.583333236, 0, 2.56598692168999e-07, 0, 0, 0, 147061.249999795, 0, 9804.08333332832, 1.04223020212603e-08, 0, 4.47902004424199e-08, 147061.250000046, 0, 0, 88236.7499996955, 0, 5.20965977518199e-08, 176473.500000151, 88236.7500000196, 29412.2500001359, 1.79890926158033e-07, 9804.08333329208, 1.63595265005091e-07, 29412.2500002354, 0, -2.32109156473271e-08, 0, -22876.1944444111, 0, -1.01515817017053e-07, 0, 0, 0, -68628.5833332584, 0, -4.37154545189875e-08, 0, 0, -1.66311908002857e-08, -68628.5833333526, 1.81770453365407e-08, 0, -68628.5833331793, 0, -22876.1944442966, -2.01178406031291e-08, 0, -3.11483478602428e-08, -137257.166666737, -68628.5833333495, -5.32866598727806e-08, -6.65885509589631e-08, 3.86513565481082e-08, -7.20054286190609e-08, -68628.5833334457, -22876.1944444156, -1.60197562857913e-07, -68628.5833334922, -2.2837137886136e-07, -1.30899731517198e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, -257.249999999781, 0, 0, 0, 1.46355414923047e-08, 0, 3858.74999999217, 0, 557.375000003008, 0, 0, 0, 557.375000000617, 0, -4.20454693457219e-08, 0, -20058.3541666382, 0, -300.125000030301, 0, 0, 0, -7753.22916662409, 0, 0, -7753.2291666713, -600.250000006907, -300.125000000268, 0, 6.25181781770645e-08, 0, 45869.1041666191, 0, 9.17704384066085e-08, 0, 0, 0, 35014.5833332225, 0, 3851.60416657085, 1.24071215714958e-08, 0, 1.82956756190262e-08, 35014.5833333489, 0, 0, 7703.20833336783, 3851.6041666682, 4.20545446725218e-08, 0, -4.62730574319928e-08, 0, -46569.3958332934, 0, -1.11357112369762e-07, 0, 0, 0, -61275.5208332365, 0, 6.97205815408562e-08, 0, 0, -2.21809835308111e-08, -61275.5208333523, 0, 1.11936592513015e-08, -14706.1249997994, 0, -2.74719788448781e-08, -29412.2500000535, -14706.1250000059, -4.73111449431647e-08, -7.17851396906995e-08, 2.92189662369996e-08, -5.24134680244992e-08, -7.59422876772386e-08, 0, 1.3470539293426e-08, 0, 17157.1458333199, 0, 4.67840770106386e-08, 0, 0, 0, 34314.2916666361, 0, 0, 0, 0, 0, 34314.2916666747, 0, 0, 17157.145833246, 0, -1.11932096392341e-07, 1.16476444373072e-08, 0, 1.61058466029439e-08, 34314.2916666924, 17157.1458333413, 1.79668624825839e-08, 2.80691623234981e-08, -2.45720857535763e-08, 1.96067913020853e-08, 4.18829948114401e-08, -1.50366962156632e-08, 6.70091927701307e-08, 6.42797872888811e-08, 8.58113925636663e-08, 4.62079706966563e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 42.8750000001313, 0, 0, 0, 0, 0, -593.1041666652, 0, -343.000000003119, 0, 0, 0, -42.8750000002041, 0, 0, 0, 2651.10416665879, 0, 300.125000016776, 0, 0, 0, 4701.95833333103, 0, 0, 550.229166670135, 300.125000001796, 0, 0, 0, 0, -4551.8958333176, 0, -3.74758134396859e-08, 0, 0, 0, -20658.6041666461, 0, -3851.60416667321, 0, 0, 0, -2100.87500000622, 0, 0, -3851.60416667941, 0, 0, 0, 0, 0, 2451.02083331889, 0, 3.72742159396447e-08, 0, 0, 0, 34314.2916666405, 0, 2.21615924795158e-08, 0, 0, 0, 2451.02083334204, 0, 0, 14706.1249999767, 0, 0, 14706.125000034, 0, 1.60987564758293e-08, 2.33528481765899e-08, -1.67540199342755e-08, 1.19791006092465e-08, 0, 0, 0, 0, 0, 0, -1.3561299761576e-08, 0, 0, 0, -17157.1458333227, 0, -1.45138229298284e-08, 0, 0, 0, 0, 0, 0, -17157.1458333168, 0, 0, 0, 0, 0, -17157.1458333483, 0, 0, -1.13134311709728e-08, 0, 0, -1.23933659306538e-08, 0, 0, 0, -2.80324956475894e-08, -2.04828278166382e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 38.1111111112898, 0, 0, 0, 0, 0, -438.277777776787, 0, -438.277777781528, 0, 0, 0, -38.1111111111203, 0, 0, 0, 1333.88888888068, 0, 1333.8888889072, 0, 0, 0, 5002.0833333396, 0, 0, 400.166666665412, 400.16666666618, 0, 0, 0, 0, -933.722222202876, 0, -933.72222225954, 0, 0, 0, -14939.5555555484, 0, -14939.5555555938, 0, 0, 0, -933.7222222253, -933.722222223122, 0, -4201.75000000599, 0, 0, 0, 0, 0, -1.90931119420812e-08, 0, 3.39223366246577e-08, 0, 0, 0, 9804.08333330874, 0, 9804.08333339161, 0, 0, -1.15980492645198e-08, 0, 0, 0, 42484.3611111343, 0, 0, 9804.08333333327, 0, 9804.0833333439, 0, 0, 0, 4.51414946022745e-08, 0, 0, 0, 0, 0, -1.14239181042423e-08, 0, 0, 0, 1.48138158929332e-08, 0, -2.83598171342766e-08, 0, 0, 0, 0, 0, 0, -22876.1944444436, 0, -22876.1944444691, 0, 0, 0, 0, 0, 0, 0, 0, 0, -22876.1944444547, 1.27766796968815e-08, 0, -2.49665813159919e-08, -1.17454704046667e-08, 1.21241585714041e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 42.8750000001661, 0, 0, 0, 0, 0, -343.00000000021, 0, -593.10416666967, 0, 0, 0, -42.8749999998048, 0, 0, 0, 300.124999997784, 0, 2651.10416668041, 0, 0, 0, 4701.95833334493, 0, 0, 300.125000000833, 550.229166664865, 0, 0, 0, 0, 0, 0, -4551.89583335882, 0, 0, 0, -3851.60416667884, 0, -20658.6041667137, 0, 0, 0, 0, -2100.87499999126, 0, -3851.60416665862, 0, 0, 0, 0, 0, 0, 0, 2451.02083335418, 0, 0, 0, 0, 0, 34314.2916667308, 0, 0, 0, 0, 2451.02083331805, 0, 14706.1250000476, 0, 0, 0, 0, 14706.1249999805, 0, 0, 0, -3.97777883696312e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -17157.1458333612, 0, 0, 0, 0, 0, 0, -1.46601807627415e-08, 0, -17157.1458333684, 0, 0, 0, 0, 0, -17157.1458333191, 0, 0, 0, 0, 0, 0, 1.8331147705113e-08, 1.85655790392769e-08, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, -257.250000000189, 0, 0, 0, 0, 0, 557.375000000024, 0, 3858.75000000426, 0, 0, 0, 557.375000000576, 0, 0, 0, -300.124999996915, 0, -20058.3541666887, 0, 0, 0, -7753.22916667984, 0, -1.05482909139011e-08, -600.250000000566, -7753.22916667111, -300.125000004804, 0, 0, 0, 0, 0, 45869.1041667108, 1.65653039032677e-08, 0, 0, 3851.60416667607, 0, 35014.5833333967, 3.31925363918428e-08, 0, 1.65841976396539e-08, 0, 35014.5833333453, 1.14258290504973e-08, 7703.20833332937, 0, 3851.6041667015, 0, 0, 0, 0, 0, -46569.395833371, -1.34812455499415e-08, 0, 0, 0, 0, -61275.5208334277, -3.92238777780246e-08, 0, -1.02676150457105e-08, 0, -61275.5208333446, 0, -14706.1250000545, 0, -3.53266454377294e-08, 0, 0, -29412.2499999848, -14706.1250000599, -1.13174507990528e-08, -5.07284155607205e-08, -1.81706878336361e-08, 0, 0, 0, 0, 0, 17157.1458333447, 0, 0, 0, 0, 0, 34314.2916667106, 1.63370895581182e-08, 0, 0, 0, 34314.2916666704, 0, 1.35627418858644e-08, 0, 17157.1458333793, 2.13990879712355e-08, 0, 1.27147999255485e-08, 0, 0, 34314.2916666561, 17157.1458333666, 0, 2.27596094337808e-08, 0, 0, 4.64663309991103e-08, 0, 1.11097166087575e-08, 2.33328626999038e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 571.666666667578, 0, 0, 0, 0, 0, -2038.94444444472, 0, -8041.44444446152, 0, 0, 0, -2038.94444444673, 0, 0, 0, 2400.99999998598, 0, 37815.7500000824, 2.55446787855672e-08, 0, 0, 26210.9166667183, 0, 2.40631078614698e-08, 4802.00000000502, 26210.9166666851, 2401.00000001706, 0, 0, 0, -933.722222183166, 0, -76098.3611112738, -4.79063927591296e-08, 0, -1.16591004037589e-08, -28011.6666667053, 0, -103176.305555785, -7.68554384270508e-08, 0, -3.36905690044727e-08, -2801.1666666778, -103176.3055556, -933.7222222641, -56023.333333356, -2801.1666666886, -28011.6666667799, 0, 0, 0, -4.00010534787792e-08, 0, 68628.5833334745, 4.03986968751075e-08, 0, 1.00511990476705e-08, 9804.08333331389, 0, 147061.250000328, 9.79801818564551e-08, 0, 2.45404813393619e-08, 1.33804488024903e-08, 147061.250000042, 4.40148241784486e-08, 88236.7500002014, 0, 6.64431328740909e-08, 29412.2500000335, 1.59432291666846e-08, 176473.500000013, 88236.7500001901, 3.59554877745141e-08, 9804.08333349203, 29412.2500001259, 0, 0, 0, 1.42674855571326e-08, 0, -22876.194444489, -1.2629766391338e-08, 0, 0, 2.44453280639399e-08, 0, -68628.5833334837, -4.3382072672145e-08, 0, 0, 0, -68628.583333347, -1.85130991685427e-08, -22876.1944444954, 0, -68628.5833334948, -4.40023887857742e-08, 0, -1.97763967228142e-08, -1.71137728991101e-08, 0, -137257.166666662, -68628.5833334362, -1.53600308634834e-08, -7.27429299128939e-08, -68628.583333345, 0, -22876.1944445789, -6.21980871801743e-08, -68628.5833334264, -1.03020290815798e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, -857.500000001732, 0, 0, 0, 0, 0, 4559.0416666653, 0, 10561.5416666983, 0, 0, 0, 4559.04166667178, 0, -1.03089863265687e-08, 0, -8953.72916663358, 0, -40867.0208334857, -3.05585120809365e-08, 0, -1.8033573254001e-08, -49820.7500000768, 0, -3.00409728678167e-08, -17907.4583333495, -49820.7500000363, -8953.72916670454, 0, 1.1752882863725e-08, 0, 7703.20833324832, 0, 70379.3125003069, 5.54781914008548e-08, 0, 3.46194726103665e-08, 85785.7291666974, 0, 148461.833333689, 8.71596870999225e-08, 0, 4.78506255150193e-08, 23109.6250000329, 148461.833333405, 7703.20833342639, 171571.45833343, 23109.6250000548, 85785.7291668921, 0, 0, 0, -2451.02083324764, 0, -56373.4791669421, -4.62389000597307e-08, 0, -3.35018099572038e-08, -63726.5416665987, 0, -171571.458333855, -1.12505067777802e-07, 0, -4.53560715550539e-08, -9804.0833333672, -171571.45833338, -2451.02083343946, -176473.50000027, 0, -7.03895152846572e-08, -191179.62500014, -14706.1250000373, -352947.000000105, -176473.500000357, -9804.08333340647, -63726.5416669971, -191179.625000367, 0, 0, 0, -3.07145122491528e-08, 0, 17157.1458334242, 1.42259728872257e-08, 0, 1.20218318609464e-08, 17157.1458332793, 0, 68628.5833335793, 5.04376136196833e-08, 0, 2.19528685930826e-08, 1.36714408501485e-08, 68628.5833333403, 4.64503476437234e-08, 68628.5833333855, 0, 102942.87500023, 4.55806485361264e-08, 0, 2.07365994023766e-08, 68628.5833333989, 0, 205885.750000023, 102942.875000173, 3.29812967476943e-08, 17157.1458334894, 205885.750000081, 0, 68628.583333592, 102942.875000194, 205885.750000285, 68628.5833335831}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 428.750000001493, 0, 0, 0, 0, 0, -4530.4583333283, 0, -4530.45833336238, 0, 0, 0, -1529.20833333489, 0, 2.00993294291489e-08, 0, 12505.2083332861, 0, 12505.2083334753, 0, 0, 0, 46019.1666667177, 0, 0, 14305.9583333434, 14305.9583333421, 1800.75000000908, 0, -2.59975349999463e-08, 0, -13305.5416665556, 0, -13305.5416669574, -1.28431541892491e-08, 0, -1.37811461518985e-08, -114847.833333306, 0, -114847.833333611, 0, 0, 1.54684599155161e-08, -27311.3750000294, -27311.3750000077, -700.291666686943, -129553.958333399, -14706.1250000009, -14706.1250000558, 0, 1.68428901582987e-08, 0, 4902.04166655745, 0, 4902.04166693332, 1.16205367769214e-08, 0, 1.72864926182319e-08, 107844.916666542, 0, 107844.916667094, 1.31304766905307e-08, 0, -1.59454852930791e-08, 14706.125000036, 14706.1249999794, 3.40999493966363e-08, 205885.750000159, 0, -3.5565923285905e-08, 220591.875000125, 14706.1250000045, 220591.875000085, 14706.1250000949, 4902.04166664866, 4902.04166674113, 117649.00000016, 0, 0, 0, 3.89332266857174e-08, 0, -9.03378828863551e-08, 0, 0, 0, -34314.2916665918, 0, -34314.2916668749, -1.00264582392817e-08, 0, 0, -1.560072422639e-08, 1.89575101871339e-08, -1.95030792180166e-08, -102942.875000004, 0, -102942.875000163, 1.26447152785128e-08, 0, 2.50468823961976e-08, -102942.875000064, 0, -102942.875000027, -4.21520810566985e-08, 1.06774632889777e-08, -3.48138889087061e-08, -205885.750000081, 2.02881398576088e-08, -5.3695637913028e-08, -102942.875000101, -102942.875000137, -34314.2916667237}, 
                                                    {0, 0, 0, 0, 0, 0, 0, -171.500000000589, 0, 0, 0, 0, 0, 2172.33333332873, 0, 1572.08333334623, 0, 0, 0, 371.583333334122, 0, -1.97440320100188e-08, 0, -8303.45833330393, 0, -2801.16666673298, 0, 0, 0, -19508.1250000052, 0, 0, -4301.79166667296, -3001.25000000474, -200.083333336222, 0, 2.68134842453414e-08, 0, 11204.6666666035, 0, 1400.58333347583, 0, 0, 0, 71429.7499999446, 0, 32213.4166667421, 0, 0, 0, 13305.5416666842, 2801.16666667267, 0, 34314.2916667108, 2100.87499999503, 1400.58333336107, 0, -1.83540376927403e-08, 0, -4902.04166660667, 0, -1.36707852922563e-07, 0, 0, 0, -88236.749999906, 0, -14706.1250001419, 0, 0, 1.069254526683e-08, -9804.08333335549, 1.20171096855903e-08, 0, -98040.8333333188, 0, 1.55933574860717e-08, -102942.87500008, -4902.04166666863, -29412.2500000648, -4.91317636722145e-08, 2.67292810025692e-08, -4.36432350501489e-08, -14706.1250001089, 0, 0, 0, -2.11423119572657e-08, 0, 4.82886577871986e-08, 0, 0, 0, 34314.2916666211, 0, 7.74759138463717e-08, 0, 0, 0, 0, -1.04917402566208e-08, 0, 68628.5833332961, 0, 34314.2916666906, 0, 0, 0, 68628.5833333724, 0, 2.7738916578213e-08, 1.6809332271815e-08, -1.90030811828177e-08, 1.94548644650142e-08, 68628.5833333864, -2.01227488038105e-08, 3.16692181247787e-08, 34314.2916667282, 9.55060206861305e-08, 3.45115989413411e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, -171.500000000706, 0, 0, 0, 0, 0, 1572.08333333262, 0, 2172.33333334649, 0, 0, 0, 371.583333333236, 0, 0, 0, -2801.16666664997, 0, -8303.45833339499, 0, 0, 0, -19508.1250000384, 0, 0, -3001.25000000073, -4301.79166666537, -200.083333333153, 0, 0, 0, 1400.58333328921, 0, 11204.6666667864, 1.20314760283401e-08, 0, 0, 32213.4166666882, 0, 71429.7500001721, 0, 0, -1.11134358039024e-08, 2801.16666667356, 13305.5416666591, 0, 34314.2916666678, 1400.58333332908, 2100.87499999983, 0, 0, 0, 4.50872474665229e-08, 0, -4902.0416667704, -1.10761000448265e-08, 0, 0, -14706.124999969, 0, -88236.7500002424, -1.30341089258647e-08, 0, 1.08478939886713e-08, -1.03977672520024e-08, -9804.0833333207, 0, -98040.8333334845, 0, 1.88837874706429e-08, -29412.2500000162, 0, -102942.874999992, -4902.04166666703, 0, 0, -14706.1249999978, 0, 0, 0, -1.62200876069144e-08, 0, 3.30902185445965e-08, 0, 0, 0, -2.80536093953045e-08, 0, 34314.2916667767, 0, 0, 0, 0, 0, 0, 34314.2916667021, 0, 68628.5833334547, 0, 0, -1.42830396353439e-08, 1.21281280086294e-08, 0, 68628.5833333232, 0, 0, 0, 68628.5833333386, 0, 0, 0, 34314.2916666601, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 1029.00000000013, 0, 0, 0, 0, -7871.85000000134, 0, 0, -7871.85000000023, 0, -7871.85000000436, 0, 0, 0, 0, 23509.7916666737, 0, 0, 23509.7916666697, 2.74131240458606e-08, 0, 47019.5833333288, 3.70541863547476e-08, 47019.5833333477, 23509.7916666936, 47019.5833333678, 0, 0, 0, 0, -34314.2916666835, 0, 0, -34314.2916666755, -4.80480662838863e-08, 0, -102942.874999984, 0, -7.01801897438687e-08, -102942.874999992, -1.06639008715703e-07, -102942.875000023, -34314.2916667065, -102942.875000088, -102942.875000089, -205885.750000074, -102942.875000191, 0, 0, 0, 0, 24510.2083333513, 0, 0, 24510.2083333426, 3.96706778663412e-08, 0, 98040.8333333182, 0, 5.27671018220695e-08, 98040.8333333374, 1.24344386189414e-07, 98040.8333333486, 24510.2083333429, 98040.8333334246, 0, 147061.249999966, 1.41078057554447e-07, 147061.250000106, 294122.500000059, 98040.8333334309, 98040.8333335973, 294122.500000098, 147061.250000308, 294122.500000364, 0, 0, 0, 0, -6862.85833334023, 0, 0, -6862.85833333664, -1.24710828226229e-08, 0, -34314.2916666639, 0, -1.40661258566033e-08, -34314.2916666723, -5.20223653389557e-08, -34314.2916666707, -6862.85833332438, -34314.2916667021, 0, -68628.5833333098, 0, -5.90533300439756e-08, -68628.5833333227, -8.11968570678524e-08, -68628.5833333753, -137257.166666679, -34314.2916666882, -34314.2916667702, -137257.166666707, -68628.5833334844, -68628.583333402, -205885.750000029, -68628.5833335626, -205885.7500002, -137257.166666929, -205885.750000255}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 68.5999999999698, 0, 0, 0, 0, -1069.01666666649, 0, 0, -68.5999999995534, 0, -68.6000000000152, 0, 0, 0, 0, 5902.45833333292, 0, 0, 0, 0, 0, 1000.41666666444, 0, 1000.41666666597, 0, 0, 0, 0, 0, 0, -14706.1249999993, 0, 0, 0, 0, 0, -4902.04166666369, 0, 0, 0, 0, -4902.0416666666, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16666.9416666659, 0, 0, 0, 0, 0, 9804.08333333223, 0, 0, -1.18876388650018e-08, 0, 9804.08333333281, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6862.85833333293, 0, 0, 0, 0, 0, -6862.8583333335, 0, 0, 0, 0, -6862.85833333303, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 68.600000000048, 0, 0, 0, 0, -68.6000000002583, 0, 0, -1069.01666666715, 0, -68.5999999998703, 0, 0, 0, 0, 0, 0, 0, 5902.45833333551, 0, 0, 1000.41666666763, 0, 0, 0, 1000.41666666618, 0, 0, 0, 0, 0, 0, 0, -14706.1250000046, 0, 0, 0, 0, 0, -4902.04166666979, 0, 0, 0, -4902.0416666653, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16666.9416666711, 0, 0, 0, 0, 0, 9804.08333333858, 0, 0, 0, 9804.08333333292, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6862.85833333488, 0, 0, 0, 0, 0, -6862.85833333619, 0, 0, 0, -6862.8583333337, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -857.5000000001, 0, 0, 0, 0, 10561.5416666679, 0, 0, 4559.04166666646, 0, 4559.04166666973, 0, 0, 0, 0, -40867.0208333399, 0, 0, -8953.72916666766, -1.91196072173038e-08, 0, -49820.7499999942, -3.27294273317511e-08, -49820.75000001, -8953.72916668622, -17907.4583333541, 0, 0, 0, 0, 70379.3125000157, 0, 0, 7703.2083333379, 2.95882192584558e-08, 0, 148461.833333315, 0, 6.08382647971159e-08, 85785.7291666541, 8.98452734542217e-08, 148461.833333353, 7703.20833336348, 23109.6250000569, 85785.7291667382, 171571.458333382, 23109.6250001302, 0, 0, 0, 0, -56373.4791666834, 0, 1.05486627208686e-08, -2451.02083333831, -2.07854609947777e-08, 0, -171571.458333317, 0, -3.73780920998452e-08, -63726.5416666637, -9.9583430486513e-08, -171571.458333348, -2451.02083334161, -9804.08333338845, 0, -176473.499999961, -1.30909838384909e-07, -176473.500000082, -352947.000000041, -63726.5416667634, -9804.08333351308, -191179.625000061, -14706.1250002244, -191179.625000251, 0, 0, 0, 0, 17157.1458333396, 0, 0, 0, 0, 0, 68628.5833333305, 0, 0, 17157.1458333349, 4.10224756267071e-08, 68628.5833333378, 0, 1.81312671275789e-08, 0, 102942.874999974, 0, 5.25625419594714e-08, 68628.5833333206, 7.47706298056645e-08, 102942.875000033, 205885.750000012, 17157.1458333693, 6.5839139655031e-08, 68628.5833333548, 1.17064158971584e-07, 68628.583333404, 205885.750000013, 1.71872482175365e-07, 205885.750000134, 68628.5833335085, 102942.875000156}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 571.666666666684, 0, 0, 0, 0, -8041.44444444471, 0, 0, -2038.9444444439, 0, -2038.9444444456, 0, 0, 0, 0, 37815.7500000019, 0, 0, 2400.99999999848, 1.04113112100683e-08, 0, 26210.9166666605, 1.97259694860537e-08, 26210.9166666674, 2401.00000000877, 4802.00000000729, 0, 0, 0, 0, -76098.3611111162, 0, 0, -933.722222220463, -1.47526997419873e-08, 0, -103176.305555541, 0, -3.45689054162524e-08, -28011.6666666474, -5.07260168771522e-08, -103176.305555557, -933.722222237242, -2801.16666669595, -28011.6666666947, -56023.3333333476, -2801.16666673712, 0, 0, 0, 0, 68628.5833333387, 0, 0, 0, 0, 0, 147061.249999989, 0, 1.25413528648165e-08, 9804.08333331471, 5.3961077125016e-08, 147061.249999999, 0, 2.83121381121486e-08, 0, 88236.7499999633, 8.07584772151933e-08, 88236.750000018, 176473.500000002, 9804.08333337653, 9.52092439132045e-08, 29412.2500000193, 1.39847766053853e-07, 29412.2500001374, 0, 0, 0, 0, -22876.1944444463, 0, 0, 0, 0, 0, -68628.583333332, 0, 0, 0, -2.38077738887213e-08, -68628.5833333324, 0, 0, 0, -68628.5833333133, 0, -2.93901866875273e-08, -22876.194444427, -4.66566787211046e-08, -68628.583333338, -137257.166666663, -2.07021993055646e-08, -2.99218525014843e-08, 0, -8.02265125427828e-08, -22876.1944444751, -68628.5833333203, -1.09884438515685e-07, -68628.5833333914, -8.63112708690065e-08, -7.6540748484248e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -257.249999999945, 0, 0, 0, 0, 3858.74999999964, 0, 0, 557.374999999073, 0, 557.375000000192, 0, 0, 0, 0, -20058.3541666659, 0, 0, -300.124999996053, 0, 0, -7753.22916666095, 0, -7753.22916666521, -300.125000002883, -600.249999999969, 0, 0, 0, 0, 45869.1041666655, 0, 0, 0, 0, 0, 35014.5833333236, 0, 1.53598302008438e-08, 3851.60416664562, 1.71938612043847e-08, 35014.5833333325, 0, 0, 3851.60416667415, 7703.20833333668, 2.58250935743128e-08, 0, 0, 0, 0, -46569.3958333321, 0, 0, 0, 0, 0, -61275.5208333275, 0, 0, 2.6592906332102e-08, -1.80279702553342e-08, -61275.520833331, 0, 0, 0, -14706.1249999722, -3.24646157484438e-08, -14706.124999996, -29412.2499999953, -1.21128202251323e-08, -3.40538695182848e-08, 0, -4.98975959168705e-08, -5.13341982809928e-08, 0, 0, 0, 0, 17157.1458333327, 0, 0, 0, 0, 0, 34314.2916666661, 0, 0, -1.19383832619302e-08, 0, 34314.2916666655, 0, 0, 0, 17157.1458333219, 0, 1.15913365182189e-08, -1.73358135506364e-08, 1.79341704434762e-08, 17157.1458333314, 34314.2916666631, 0, 0, 0, 3.00967866746359e-08, 1.03133750730156e-08, -1.32098975158156e-08, 4.68659859187866e-08, 2.54993084369212e-08, 2.96730248205525e-08, 2.13157563200338e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 42.8750000000198, 0, 0, 0, 0, -593.104166666734, 0, 0, -343.000000000241, 0, -42.8749999999289, 0, 0, 0, 0, 2651.10416666672, 0, 0, 300.125000001381, 0, 0, 4701.95833333364, 0, 550.229166666599, 0, 300.125000000046, 0, 0, 0, 0, -4551.89583333345, 0, 0, 0, 0, 0, -20658.604166665, 0, 0, -3851.60416666965, 0, -2100.874999999, 0, 0, 0, -3851.60416666288, 0, 0, 0, 0, 0, 2451.02083333364, 0, 0, 0, 0, 0, 34314.2916666636, 0, 0, 0, 0, 2451.02083333258, 0, 0, 0, 14706.125, 0, 0, 14706.1249999989, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -17157.1458333321, 0, 0, 0, 0, 0, 0, 0, 0, -17157.1458333314, 0, 0, 0, 0, 0, -17157.1458333324, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 38.1111111111753, 0, 0, 0, 0, -438.277777778395, 0, 0, -438.277777778241, 0, -38.1111111110743, 0, 0, 0, 0, 1333.88888889149, 0, 0, 1333.88888889114, 0, 0, 5002.08333333515, 0, 400.166666665785, 0, 400.166666665895, 0, 0, 0, 0, -933.722222227766, 0, 0, -933.722222227191, 0, 0, -14939.5555555575, 0, 0, -14939.555555563, 0, -933.722222220492, 0, -933.722222220315, 0, -4201.7499999975, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9804.08333333476, 0, 0, 9804.08333334548, 0, 0, 0, 0, 0, 42484.3611111162, 0, 0, 9804.08333332784, 0, 0, 9804.08333332859, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -22876.1944444445, 0, 0, -22876.1944444501, 0, 0, 0, 0, 0, 0, 0, 0, -22876.1944444417, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 42.8749999999891, 0, 0, 0, 0, -343.000000000121, 0, 0, -593.104166666256, 0, -42.874999999904, 0, 0, 0, 0, 300.125000001134, 0, 0, 2651.10416666522, 0, 0, 4701.95833333102, 0, 300.124999999949, 0, 550.229166666124, 0, 0, 0, 0, 0, 0, 0, -4551.895833331, 0, 0, -3851.60416666244, 0, 0, -20658.6041666594, 0, 0, 0, -2100.87499999983, 0, -3851.6041666668, 0, 0, 0, 0, 0, 0, 0, 0, 2451.02083333156, 0, 0, 0, 0, 0, 34314.2916666592, 0, 0, 0, 2451.02083333189, 0, 14706.1249999892, 0, 0, 0, 0, 0, 14706.1249999978, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -17157.1458333307, 0, 0, 0, 0, 0, 0, 0, 0, -17157.1458333273, 0, 0, 0, 0, 0, -17157.1458333316, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -257.250000000114, 0, 0, 0, 0, 557.375000000767, 0, 0, 3858.75000000086, 0, 557.374999999393, 0, 0, 0, 0, -300.125000003514, 0, 0, -20058.354166671, 0, 0, -7753.22916666669, 0, -600.249999995898, -300.124999996747, -7753.22916666471, 0, 0, 0, 0, 0, 0, 0, 45869.1041666767, 0, 0, 3851.60416666411, 0, 0, 35014.583333334, 1.05170992109261e-08, 0, 0, 35014.5833333307, -1.70153130082471e-08, 7703.20833331844, 3851.60416666143, 0, 0, 0, 0, 0, 0, 0, -46569.3958333434, 0, 0, 0, 0, 0, -61275.5208333395, 0, 0, 0, -61275.5208333341, 0, -14706.1249999881, -1.48150925482363e-08, 2.03118266569629e-08, 1.74042972118213e-08, 2.4695984401171e-08, 0, -29412.249999986, -14706.1249999957, 1.89615143606779e-08, 0, 0, 0, 0, 0, 0, 0, 17157.145833337, 0, 0, 0, 0, 0, 34314.2916666716, 0, 0, 0, 34314.291666669, 0, 0, 0, 0, 17157.1458333286, 0, -1.1670360581351e-08, -1.01154461411452e-08, 0, 0, 34314.2916666601, 17157.1458333323, -1.64409536827886e-08, -1.07574581801088e-08, 0, -2.09323768857807e-08, -1.6016202296816e-08, -1.35106452228249e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 571.666666666859, 0, 0, 0, 0, -2038.94444444615, 0, 0, -8041.44444444542, 0, -2038.94444444411, 0, 0, 0, 0, 2401.00000000855, 0, 0, 37815.7500000059, 1.07356830435013e-08, 0, 26210.9166666643, 1.42389381750771e-08, 4801.99999999527, 2401.00000000048, 26210.9166666683, 0, 0, 0, 0, -933.722222242133, 0, 0, -76098.3611111263, -1.87909227859531e-08, 0, -28011.6666666564, 0, -3.1589868735874e-08, -103176.305555552, -3.94926097733672e-08, -2801.16666665346, -933.722222224043, -103176.305555565, -2801.16666665615, -56023.3333333229, -28011.6666666908, 0, 0, 0, 0, 2.05638397921132e-08, 0, 0, 68628.583333349, 1.62390230467599e-08, 0, 9804.08333332723, 0, 2.84582494819374e-08, 147061.250000008, 4.46559185580507e-08, -1.49462767483776e-08, 0, 147061.250000014, 0, 88236.7499999694, 5.49246486502752e-08, -1.76718392192369e-08, 29412.2499999781, -2.13567064212239e-08, 9804.08333337082, 176473.49999999, 88236.7500000324, 29412.2500000262, 0, 0, 0, 0, 0, 0, 0, -22876.1944444501, 0, 0, 0, 0, 0, -68628.5833333417, -1.76045576263841e-08, 0, 0, -68628.5833333419, 0, -22876.1944444253, 0, -2.54461252773178e-08, -68628.5833333219, -3.00863880296068e-08, 1.17306004145652e-08, 1.63082946906365e-08, 1.33467387939174e-08, -1.89830652822578e-08, -137257.16666666, -68628.583333347, 1.51531484605346e-08, -68628.5833333205, -22876.1944444752, -1.281747933207e-08, -2.42909780901756e-08, -68628.5833333568}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -857.500000000226, 0, 0, 0, 0, 4559.04166666889, 0, 0, 10561.5416666677, 0, 4559.04166666861, 0, 0, 0, 0, -8953.72916667767, 0, 0, -40867.0208333403, -2.23333487395216e-08, 0, -49820.749999997, -2.89112408725393e-08, -17907.4583333388, -8953.72916667925, -49820.7500000186, 0, 0, 0, 0, 7703.20833335908, 0, 0, 70379.312500018, 3.96787158969802e-08, 0, 85785.7291666508, 0, 5.69582627534294e-08, 148461.833333333, 8.66788832191867e-08, 23109.6250000064, 7703.20833334923, 148461.83333338, 23109.6250000379, 171571.458333368, 85785.7291667746, 0, 0, 0, 0, -2451.02083336052, 0, 0, -56373.479166685, -3.35044438146485e-08, 0, -63726.5416666521, 0, -4.64058972265973e-08, -171571.458333352, -1.01428670414448e-07, -9804.08333333562, -2451.02083333047, -171571.458333383, 0, -176473.499999966, -1.14502466067311e-07, -14706.1250000474, -191179.625000027, -9804.08333336027, -63726.5416668147, -352947.000000045, -176473.500000168, -191179.625000208, 0, 0, 0, 0, 1.03962700854276e-08, 0, 0, 17157.1458333398, 1.09274839379478e-08, 0, 17157.1458333313, 0, 1.32061610675776e-08, 68628.5833333476, 4.14808152251678e-08, 0, 0, 68628.5833333544, 0, 68628.5833333078, 0, 4.97735863335729e-08, 102942.874999993, 6.64608027254616e-08, 1.72209325250665e-08, 68628.5833333353, 0, 17157.1458333918, 205885.750000014, 102942.875000078, 2.06876288211579e-08, 205885.750000014, 68628.5833334601, 102942.875000123, 68628.583333482, 205885.750000148}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 428.750000000183, 0, 0, 0, 0, -4530.45833333542, 0, 0, -4530.45833333404, 0, -1529.20833333456, 0, 0, 0, 0, 12505.2083333431, 0, 0, 12505.2083333387, 0, 0, 46019.1666666665, 1.66829814995022e-08, 14305.9583333398, 1800.75000000817, 14305.9583333403, 0, 0, 0, 0, -13305.541666689, 0, 0, -13305.5416666802, -1.29047951955781e-08, 0, -114847.833333324, 0, -3.56115626787505e-08, -114847.833333344, -4.69092106842181e-08, -27311.3750000143, -700.291666680199, -27311.3750000116, -14706.1250000351, -129553.958333355, -14706.1250000439, 0, 0, 0, 0, 4902.04166669039, 0, 0, 4902.0416666798, 0, 0, 107844.916666654, 0, 3.14128340834151e-08, 107844.916666696, 5.0873982201531e-08, 14706.125000014, 0, 14706.1250000061, 0, 205885.749999988, 6.92465205509669e-08, 14706.1250000514, 220591.875000031, 4902.04166671065, 4902.04166673198, 220591.875000017, 14706.1250000622, 117649.000000098, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -34314.2916666637, 0, 0, -34314.2916666837, -1.93317880935786e-08, 0, 0, 0, 0, -102942.874999983, 0, -3.16454146790174e-08, -102942.875000007, -3.83188179125213e-08, -2.32680539558982e-08, -102942.875000014, -1.44406122074337e-08, -2.69691736127837e-08, -102942.875000001, -2.67289120446555e-08, -3.39602554387964e-08, -205885.750000013, -5.05364346838079e-08, -102942.87500007, -34314.2916667384, -102942.875000055}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -171.500000000134, 0, 0, 0, 0, 2172.33333333439, 0, 0, 1572.08333333448, 0, 371.583333333477, 0, 0, 0, 0, -8303.45833333732, 0, 0, -2801.16666667272, 0, 0, -19508.1250000033, 0, -4301.79166666705, -200.083333333314, -3001.2500000013, 0, 0, 0, 0, 11204.6666666752, 0, 0, 1400.58333334635, 0, 0, 71429.7499999995, 0, 0, 32213.4166666841, 1.29012486031781e-08, 13305.5416666663, 0, 2801.16666666666, 2100.87500000308, 34314.2916666628, 1400.58333333571, 0, 0, 0, 0, -4902.04166667583, 0, 0, -1.18877270294651e-08, 0, 0, -88236.7499999953, 0, 0, -14706.1250000301, -1.12547260630087e-08, -9804.08333333399, 0, 0, 0, -98040.833333341, -1.80108523341429e-08, -4902.04166667665, -102942.875, 0, 0, -29412.2499999987, -1.70393457031689e-08, -14706.125000025, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 34314.2916666653, 0, 0, 1.49303644515725e-08, 0, 0, 0, 0, 0, 68628.5833333293, 0, 0, 34314.2916666796, 1.10881728604972e-08, 0, 68628.5833333334, 0, 0, 0, 1.14291726747551e-08, 0, 68628.5833333304, 0, 34314.2916666657, 1.13774002556394e-08, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -171.500000000064, 0, 0, 0, 0, 1572.08333333444, 0, 0, 2172.33333333307, 0, 371.583333333452, 0, 0, 0, 0, -2801.16666667271, 0, 0, -8303.45833333328, 0, 0, -19508.1249999974, 0, -3001.25000000064, -200.083333336976, -4301.79166666616, 0, 0, 0, 0, 1400.58333334738, 0, 0, 11204.6666666685, 0, 0, 32213.4166666597, 0, 1.47163942575793e-08, 71429.7499999948, 1.17805240066982e-08, 2801.16666666897, 0, 13305.5416666658, 1400.58333334468, 34314.2916666709, 2100.87500001306, 0, 0, 0, 0, -1.46618715939981e-08, 0, 0, -4902.04166666914, 0, 0, -14706.1249999946, 0, -1.44363891803399e-08, -88236.750000001, -1.43905191194901e-08, 0, 0, -9804.08333332717, 0, -98040.8333333169, -1.9772845315889e-08, -1.03301907645621e-08, -29412.2500000011, -1.65957332581045e-08, -2.08856944620778e-08, -102942.874999995, -4902.04166666906, -14706.1250000117, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 34314.2916666696, 0, 0, 0, 0, 0, 34314.2916666558, 0, 1.13992389858709e-08, 68628.5833333284, 0, 0, 0, 0, 0, 68628.5833333272, 0, 1.27013165376949e-08, 68628.5833333332, 1.59467957461482e-08, 2.83198386301894e-08, 2.10802885963008e-08, 34314.2916666794}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 68.6000000000397, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1000.4166666661, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4902.04166666691, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -9804.08333333313, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6862.85833333315, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.04077520552725e-08, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 68.6000000001046, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1000.41666666749, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4902.04166667045, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -9804.08333333916, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.31496662523315e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6862.85833333555, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 68.5999999998978, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1000.41666666634, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4902.0416666657, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -9804.08333333328, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6862.85833333384, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 42.8750000000464, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -550.229166667994, -300.125000000279, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2100.87499999895, 0, 0, 3851.60416666613, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2451.02083333292, 0, 0, 0, 0, 0, -14706.1250000083, 0, 0, 0, 0, 0, 1.84495695790547e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 17157.1458333356, 0, 0, 0, 0, 0, 0, 0, 0, -1.58504392716429e-08, 0, 1.06466232019746e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 38.1111111112397, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -400.166666664962, -400.166666666939, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 933.722222222014, 933.722222227649, 0, 4201.75000000355, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.15980492645196e-08, 0, -1.30112647584164e-08, 0, 0, 0, 0, -9804.08333332766, 0, -9804.08333334288, 0, 0, 0, -4.1052993886769e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.02571142250867e-08, 0, 0, 0, 22876.194444454, -1.01041271224881e-08, 0, 1.86692124440991e-08, 1.16568899393703e-08, -1.02391282733358e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 42.8750000000812, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -300.125000001971, -550.229166668283, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2100.87500000276, 0, 3851.60416666969, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2451.02083333264, 0, 0, 0, 0, -1.14576584383998e-08, 0, -14706.1250000047, -1.05706319170879e-08, 0, 0, 2.34053996447857e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 17157.1458333329, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 42.8750000001856, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -550.229166668158, -300.125000001459, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2100.87500000181, 0, 0, 0, 3851.60416667389, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2451.02083333532, 0, 0, 0, 0, 0, 0, 0, -14706.1250000129, 0, -1.56180470299049e-08, -1.83074704686569e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 17157.1458333387, 0, 0, 0, 0, 1.18009142308131e-08, 0, 1.34875338511919e-08, 1.46195115342652e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 38.111111111298, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -400.166666667723, -400.166666668676, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 933.722222224715, 933.722222228615, 0, 0, 4201.75000001131, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -9804.08333334134, 0, -9804.08333334564, -1.21452823589734e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 22876.1944444502, 0, 0, 1.18439732353922e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 42.8750000001241, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -300.125000000987, -550.229166667476, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2100.87500000363, 0, 0, 3851.60416666939, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2451.02083333735, 0, 0, 0, 0, 0, 0, 0, 0, -14706.1250000056, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 17157.1458333367, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 42.8750000000077, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -300.12500000053, 0, -550.229166666755, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2100.875000001, 0, 3851.60416666942, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2451.0208333332, 0, 0, 0, 0, 0, 0, 0, -14706.1250000019, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 17157.1458333334, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 38.1111111112252, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -400.16666666674, 0, -400.166666666769, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 933.722222222485, 0, 933.722222222436, 0, 4201.75000000169, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -9804.08333333366, 0, 0, -9804.08333333394, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 22876.1944444456, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 42.8750000000625, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -550.229166667337, 0, -300.125000000862, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2100.87500000063, 0, 0, 0, 3851.60416666607, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2451.02083333412, 0, 0, 0, 0, 0, 0, -14706.1250000037, 0, 0, 0, 0, -1.23438505247264e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 17157.1458333346, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 28.5833333334106, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -300.125000000499, -200.083333334508, -200.083333333937, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 700.291666669111, 0, 0, 2100.875000003, 2100.87500000691, 1400.58333333932, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4902.0416666654, -4902.04166666955, 0, 0, 0, 0, -14706.1249999892, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 34314.2916666774, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 28.583333333322, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -200.083333333273, -300.124999999715, -200.083333333344, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 700.29166666574, 0, 2100.87499999937, 1400.58333333272, 2100.87500000078, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4902.04166666417, -4902.04166666895, 0, 0, -14706.1250000023, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 34314.29166667, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 28.5833333335017, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -200.083333334023, -200.083333333371, -300.125000001516, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 700.291666666467, 1400.58333333498, 2100.87499999853, 2100.87500000383, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.11315844801095e-08, -4902.04166666798, -4902.04166668185, -14706.1250000246, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 34314.291666671}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 1029.00000000033, 0, 0, 0, 0, 0, 0, 0, -7871.84999999748, 0, -7871.85000000611, -7871.85000000639, 0, 0, 0, 0, 0, 0, 23509.7916666504, 0, 23509.7916667016, 0, 0, 47019.583333325, 23509.7916666854, 47019.58333336, 47019.5833333974, 0, 0, 0, 0, 0, 0, -34314.2916666338, 0, -34314.2916667466, 0, 0, 0, -102942.874999926, 0, -102942.875000018, -34314.291666685, -102942.875000017, -102942.875000177, -102942.875000082, -102942.875000144, -205885.750000254, 0, 0, 0, 0, 0, 0, 24510.2083333044, 0, 24510.2083334137, 0, 0, 0, 98040.8333332277, 0, 98040.8333333984, 24510.2083333386, 98040.8333333036, 98040.8333335359, 0, 0, 147061.249999918, 98040.8333333984, 98040.833333433, 147061.25000005, 294122.500000285, 147061.250000241, 294122.500000436, 294122.500000547, 0, 0, 0, 0, 0, 0, -6862.85833332382, 0, -6862.85833336275, 0, 0, 0, -34314.2916666208, 0, -34314.2916667094, -6862.85833333337, -34314.2916666412, -34314.2916667541, 0, 0, 0, -68628.5833332641, 0, -68628.5833333149, -34314.29166668, -34314.2916666814, -68628.583333317, -137257.166666751, -68628.5833334565, -137257.166666882, -68628.5833333457, -68628.5833334198, -205885.750000261, -137257.16666694, -205885.750000361, -205885.750000456}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 68.5999999999929, 0, 0, 0, 0, 0, 0, 0, -1069.01666666666, 0, -68.5999999997929, -68.6000000000902, 0, 0, 0, 0, 0, 0, 5902.4583333335, 0, 0, 0, 0, 1000.41666666553, 0, 1000.41666666757, 0, 0, 0, 0, 0, 0, 0, -14706.1250000003, 0, 0, 0, 0, 0, -4902.04166666553, 0, 0, 0, -4902.04166666912, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16666.9416666668, 0, 0, 0, 0, 0, 9804.08333333373, 0, 0, 0, 9804.08333333656, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.11896394223368e-08, 0, 0, 0, 0, 0, 0, 0, -6862.85833333329, 0, 0, 0, 0, 0, -6862.85833333395, 0, 0, 0, -6862.85833333481, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 68.5999999999733, 0, 0, 0, 0, 0, 0, 0, -68.5999999993752, 0, -1069.01666666667, -68.5999999994131, 0, 0, 0, 0, 0, 0, 0, 0, 5902.45833333407, 0, 0, 1000.41666666314, 0, 0, 1000.4166666636, 0, 0, 0, 0, 0, 0, 0, 0, -14706.1250000024, 0, 0, 0, 1.09799611303034e-08, 0, -4902.04166665974, 0, 1.67431309909133e-08, -4902.04166665994, 1.41365769111409e-08, 0, 2.15697640587877e-08, 0, 0, 0, 0, 0, 0, 0, 0, 16666.9416666696, 0, 0, 0, -1.22770938154509e-08, 0, 9804.08333332744, 0, -1.95386661993005e-08, 9804.0833333271, 0, 0, -1.55127602541953e-08, -1.36967197919997e-08, 0, -2.60406003456395e-08, -3.71235977237987e-08, 0, -3.13863076195056e-08, -3.18541251985714e-08, 0, 0, 0, 0, 0, 0, 0, 0, -6862.85833333455, 0, 0, 0, 0, 0, -6862.85833333159, 0, 0, -6862.85833333142, 0, 0, 0, 0, 0, 0, 0, 0, 1.48763934285408e-08, 2.05718250756132e-08, 0, 1.63836918853996e-08, 1.30447467771126e-08, 0, 2.36427219399364e-08, 1.44270904925003e-08, 3.07183601483532e-08, 2.4753198722563e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -857.500000000361, 0, 0, 0, 0, 0, 0, 0, 10561.5416666657, 0, 4559.04166667143, 4559.04166667405, 0, 0, 0, 0, 0, 0, -40867.0208333259, 0, -8953.72916669495, 0, 0, -49820.7499999865, -8953.72916668433, -49820.7500000418, -17907.4583334009, 0, 0, 0, 0, 0, 0, 70379.3124999875, 0, 7703.20833340157, 0, 0, 0, 148461.833333255, 0, 85785.7291666563, 7703.20833334281, 148461.833333407, 23109.6250001912, 85785.729166764, 23109.6250001385, 171571.458333615, 0, 0, 0, 0, 0, 0, -56373.4791666583, 0, -2451.0208334045, 0, 0, 0, -171571.458333236, 0, -63726.5416666966, -2451.02083332935, -171571.458333387, -9804.08333355822, 0, 0, -176473.499999878, -63726.5416667253, -9804.08333340489, -176473.500000112, -352947.000000341, -14706.1250002523, -191179.62500048, -191179.625000552, 0, 0, 0, 0, 0, 0, 17157.1458333312, 0, 2.66217035006713e-08, 0, 0, 0, 68628.5833332963, 0, 17157.1458333619, 0, 68628.5833333485, 9.76997074693854e-08, 0, 0, 0, 102942.874999914, 0, 68628.583333289, 17157.1458333386, 0, 102942.875000036, 205885.750000127, 1.40724453194516e-07, 68628.5833335802, 68628.5833333605, 7.69404480493461e-08, 205885.75000028, 68628.5833335807, 205885.750000384, 102942.875000482}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 571.666666666897, 0, 0, 0, 0, 0, 0, 0, -8041.44444444514, 0, -2038.94444444638, -2038.94444444808, 0, 0, 0, 0, 0, 0, 37815.7500000021, 0, 2401.00000001066, 0, 0, 26210.9166666604, 2401.00000000571, 26210.9166666925, 4802.0000000315, 0, 0, 0, 0, 0, 0, -76098.3611111166, 0, -933.722222248831, 0, 0, 0, -103176.305555526, 0, -28011.6666666508, -933.722222218893, -103176.305555618, -2801.16666676026, -28011.6666667094, -2801.16666671909, -56023.3333334733, 0, 0, 0, 0, 0, 0, 68628.5833333395, 0, 2.87841360612746e-08, 0, 0, 0, 147061.249999969, 0, 9804.0833333295, 0, 147061.250000068, 1.13929132944683e-07, 0, 0, 88236.7499999331, 9804.08333334267, 0, 88236.7500000715, 176473.500000183, 1.10903691123548e-07, 29412.2500002412, 29412.2500002299, 0, 0, 0, 0, 0, 0, -22876.1944444467, 0, -1.11182513809824e-08, 0, 0, 0, -68628.5833333244, 0, 0, 0, -68628.5833333615, -5.0635059153881e-08, 0, 0, 0, -68628.5833332912, 0, -22876.1944444142, 0, 0, -68628.5833333732, -137257.166666747, -6.92524669448641e-08, -1.28424946603037e-07, -22876.1944444545, -2.44920633964926e-08, -68628.5833334746, -8.37096349844126e-08, -68628.5833335018, -2.17790047311322e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -257.250000000044, 0, 0, 0, 0, 0, 0, 0, 3858.75000000032, 0, 557.375000000064, 557.375000000882, 0, 0, 0, 0, 0, 0, -20058.354166668, 0, -300.124999999957, 0, 0, -7753.22916666464, -300.125000000073, -7753.22916667451, -600.250000007769, 0, 0, 0, 0, 0, 0, 45869.1041666692, 0, 0, 0, 0, 0, 35014.5833333295, 0, 3851.60416665521, 0, 35014.5833333547, 2.42416334389161e-08, 3851.60416667352, 0, 7703.20833337312, 0, 0, 0, 0, 0, 0, -46569.3958333353, 0, 0, 0, 0, 0, -61275.5208333322, 0, 1.48200566692934e-08, 0, -61275.5208333588, -2.90346623554639e-08, 0, 0, -14706.1249999811, 0, 1.00261317536525e-08, -14706.1250000191, -29412.250000057, -2.02076781561026e-08, -7.14369818397241e-08, -4.21954990576085e-08, 0, 0, 0, 0, 0, 0, 17157.1458333338, 0, 0, 0, 0, 0, 34314.2916666676, 0, 0, 0, 34314.2916666776, 1.30018531223508e-08, 0, 0, 0, 17157.1458333245, 0, -1.19882737687381e-08, 0, 0, 17157.1458333471, 34314.2916666941, 1.54588708301802e-08, 3.81008659895254e-08, 0, 0, 4.65745687123804e-08, 0, 3.52717461082733e-08, 4.77969846235186e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 42.8749999999656, 0, 0, 0, 0, 0, 0, 0, -593.104166665834, 0, -343.000000000074, -42.8750000000678, 0, 0, 0, 0, 0, 0, 2651.10416666289, 0, 300.125000001733, 0, 0, 4701.95833332929, 0, 550.229166666601, 300.125000002096, 0, 0, 0, 0, 0, 0, -4551.89583332638, 0, 0, 0, 0, 0, -20658.6041666505, 0, -3851.60416666159, 0, -2100.87499999942, 0, 0, 0, -3851.60416666876, 0, 0, 0, 0, 0, 0, 2451.02083332761, 0, 0, 0, 0, 0, 34314.2916666457, 0, 0, 0, 2451.02083333258, 1.24963403542088e-08, 0, 0, 14706.1249999807, 0, 0, 0, 14706.1249999944, 1.21295862472802e-08, 1.04897190407906e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -17157.1458333246, 0, 0, 0, 0, 0, 0, 0, 0, -17157.1458333198, 0, 0, 0, 0, 0, -17157.1458333284, 0, 0, 0, 0, 0, 0, 0, -1.47921576175481e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 38.1111111111173, 0, 0, 0, 0, 0, 0, 0, -438.277777777182, 0, -438.277777777997, -38.1111111113146, 0, 0, 0, 0, 0, 0, 1333.88888888659, 0, 1333.88888889163, 0, 0, 5002.08333332714, 0, 400.166666664916, 400.166666669292, 0, 0, 0, 0, 0, 0, -933.722222219447, 0, -933.722222230417, 0, 0, 0, -14939.5555555327, 0, -14939.5555555461, 0, -933.722222215988, -933.722222233286, 0, -1.0961803790178e-08, -4201.7499999926, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9804.08333330781, 0, 9804.08333333194, 0, 0, 1.5818782469242e-08, 0, 0, 42484.3611110768, -1.2544172511514e-08, 1.15285896882732e-08, -1.24069904130268e-08, 9804.08333330812, 2.50794532805611e-08, 9804.08333332666, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -22876.1944444228, 0, -22876.1944444317, 0, 0, 0, 1.52483149263004e-08, -1.51825122015242e-08, 0, 1.19312593413453e-08, -1.35578860014287e-08, -22876.1944444244, 0, 0, -1.5122974014977e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 42.8749999999328, 0, 0, 0, 0, 0, 0, 0, -342.99999999921, 0, -593.104166666093, -42.874999999973, 0, 0, 0, 0, 0, 0, 300.124999997085, 0, 2651.10416666516, 0, 0, 4701.95833332786, 0, 300.124999998029, 550.229166668097, 0, 0, 0, 0, 0, 0, 0, 0, -4551.89583333157, 0, 0, 0, -3851.60416665181, 0, -20658.6041666537, 0, 0, -2100.87500000264, 0, 0, -3851.60416666751, 0, 0, 0, 0, 0, 0, 0, 0, 2451.02083333255, 0, 0, 0, -1.70932691819627e-08, 0, 34314.2916666539, 0, -1.37267713722444e-08, 2451.02083333449, 0, 0, 14706.1249999777, 0, 0, 0, 0, 0, 14706.1250000011, 1.67897861180638e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -17157.1458333289, 0, 0, 0, 0, 0, 0, 1.18738683728713e-08, 0, -17157.1458333218, 0, 0, 0, 0, 0, -17157.1458333302, 0, 0, 0, -1.05654076088545e-08, -1.01981479496657e-08, -1.5281482827204e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -257.249999999887, 0, 0, 0, 0, 0, 0, 0, 557.374999997641, 0, 3858.74999999983, 557.374999998265, 0, 0, 0, 0, 0, 0, -300.124999990519, 0, -20058.354166669, 0, 0, -7753.22916665343, -300.124999994457, -600.249999981534, -7753.22916666005, 0, 0, 0, 0, 0, 0, -1.66850581532179e-08, 0, 45869.104166675, 0, 0, 0, 3851.60416662505, 0, 35014.5833333079, 0, -5.55982849580772e-08, 35014.5833333201, -4.483421220382e-08, 3851.60416665412, 7703.2083332768, 0, 0, 0, 0, 0, 0, 1.37748513037857e-08, 0, -46569.3958333438, 0, 0, 0, 4.8471769027052e-08, 0, -61275.5208333127, 0, 6.85107846538836e-08, -61275.5208333218, 0, 0, -14706.1249999434, 4.85521209993958e-08, 0, 8.74819372019413e-08, 1.03855273137425e-07, -14706.1249999839, -29412.2499999208, 7.21456758168944e-08, 0, 0, 0, 0, 0, 0, 0, 0, 17157.1458333378, 0, 0, 0, -1.99510302222659e-08, 0, 34314.2916666612, 0, -2.92631401988839e-08, 34314.2916666642, 0, 0, 0, -3.18229697169702e-08, 0, 17157.1458333073, -2.03727590943156e-08, 0, -5.17505522688001e-08, -6.20063607343527e-08, 17157.1458333241, 34314.2916666224, -4.74876616582409e-08, 0, -6.13445759255309e-08, -3.18003802047559e-08, -7.43435665502346e-08, -5.05891951502929e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 571.666666666522, 0, 0, 0, 0, 0, 0, 0, -2038.94444443948, 0, -8041.44444444535, -2038.94444444307, 0, 0, 0, 0, 0, 0, 2400.99999997893, 0, 37815.750000012, 0, 0, 26210.9166666395, 2400.99999999535, 4801.99999997367, 26210.9166666717, 0, 0, 0, 0, 0, 0, -933.722222184763, 0, -76098.3611111448, 0, 0, 0, -28011.6666665703, 0, -103176.305555513, -933.722222211807, -2801.16666657316, -103176.305555579, -2801.16666660643, -28011.666666688, -56023.3333332896, 0, 0, 0, 0, 0, 0, -3.07598031685851e-08, 0, 68628.5833333713, 0, 0, 0, 9804.08333321693, 0, 147061.24999998, -1.21235105985106e-08, -1.25308699414235e-07, 147061.250000033, 0, 0, 88236.7499998748, -7.62105243017863e-08, 9804.08333335201, -1.44040084560233e-07, 29412.2499998767, 88236.7500000462, 176473.499999953, 29412.2500000099, 0, 0, 0, 0, 0, 0, 0, 0, -22876.1944444595, 0, 0, 0, 4.82395185135519e-08, 0, -68628.5833333352, 0, 5.58323186191747e-08, -68628.5833333521, 0, 0, 0, -22876.1944443682, 0, -68628.5833332812, 3.67490919115191e-08, 0, 9.53486691313325e-08, 9.14323363260917e-08, -68628.5833333555, -137257.166666634, 8.67744938130928e-08, -22876.1944444721, -68628.5833332735, -1.41110847014695e-08, 4.0766378827525e-08, -68628.5833333617}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -857.499999999953, 0, 0, 0, 0, 0, 0, 0, 4559.04166665965, 0, 10561.5416666706, 4559.04166666912, 0, 0, 0, 0, 0, 0, -8953.72916663342, 0, -40867.0208333636, 0, 0, -49820.7499999688, -8953.729166675, -17907.4583333235, -49820.7500000414, 0, 0, 0, 0, 0, 0, 7703.20833327152, 0, 70379.3125000751, 0, 0, 0, 85785.7291665261, 0, 148461.83333331, 7703.20833333906, 23109.6249999196, 148461.833333455, 23109.6249999954, 85785.7291667775, 171571.458333432, 0, 0, 0, 0, 0, 0, -2451.0208332816, 0, -56373.4791667452, 0, 0, 0, -63726.5416664831, 0, -171571.458333363, -2451.02083333045, -9804.08333319747, -171571.458333477, 0, 0, -176473.499999841, -9804.08333330743, -63726.5416667571, -14706.1249999002, -191179.625000028, -176473.500000193, -352947.000000205, -191179.625000328, 0, 0, 0, 0, 0, 0, -1.62567626165588e-08, 0, 17157.1458333628, 0, 0, 0, 17157.1458332549, 0, 68628.5833333648, 0, -6.67619999120612e-08, 68628.5833333983, 0, 0, 0, 68628.5833332205, 0, 102942.874999949, -2.21521628863234e-08, 17157.14583335, -9.40788166240154e-08, 68628.5833332828, 102942.875000095, 205885.750000097, -6.96777893738824e-08, 68628.58333342, 205885.750000081, 68628.5833335103, 102942.875000165, 205885.750000297}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 428.749999999911, 0, 0, 0, 0, 0, 0, 0, -4530.45833332733, 0, -4530.45833333549, -1529.20833333712, 0, 0, 0, 0, 0, 0, 12505.208333306, 0, 12505.208333355, 0, 0, 46019.1666666322, 1800.75000001164, 14305.95833334, 14305.958333378, 0, 0, 0, 0, 0, 0, -13305.5416666188, 0, -13305.5416667249, 0, 0, 0, -114847.83333319, 0, -114847.833333294, -700.291666674614, -27311.374999978, -27311.375000133, -14706.1250000305, -14706.125000117, -129553.958333458, 0, 0, 0, 0, 0, 0, 4902.04166662965, 0, 4902.0416667292, 0, 0, 0, 107844.916666487, 0, 107844.916666678, 0, 14706.1249999467, 14706.1250001581, 0, 0, 205885.74999982, 4902.04166666645, 4902.04166675356, 14706.1249999913, 220591.875000072, 14706.1250002173, 220591.875000247, 117649.000000348, 0, 0, 0, 0, 0, 0, 1.06826184811286e-08, 0, -2.35941255631282e-08, 0, 0, 0, -34314.2916665933, 0, -34314.2916666908, 0, 2.7784851022793e-08, -6.92256681305591e-08, 0, 0, 0, -102942.874999876, 0, -102942.874999938, 1.39407156847253e-08, -1.59167183323776e-08, 2.76987835206253e-08, -102942.874999984, -1.16453282345474e-07, -102942.875000131, 2.83018826901806e-08, -9.13884979014853e-08, -205885.750000092, -34314.2916668344, -102942.875000197, -102942.87500033}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -171.499999999959, 0, 0, 0, 0, 0, 0, 0, 2172.33333333065, 0, 1572.08333333423, 371.583333334502, 0, 0, 0, 0, 0, 0, -8303.45833332131, 0, -2801.16666667653, 0, 0, -19508.124999983, -200.083333335954, -4301.79166666839, -3001.25000001498, 0, 0, 0, 0, 0, 0, 11204.6666666465, 0, 1400.58333336051, 0, 0, 0, 71429.7499999299, 0, 32213.4166666466, 0, 13305.5416666627, 2801.1666667181, 2100.8750000043, 1400.58333337114, 34314.2916666949, 0, 0, 0, 0, 0, 0, -4902.04166665209, 0, -2.94393759971036e-08, 0, 0, 0, -88236.7499999133, 0, -14706.1250000055, 0, -9804.0833333263, -6.81256411656819e-08, 0, 0, -98040.8333332424, 1.45412063160269e-08, -2.32453736124548e-08, -4902.0416666649, -102942.874999993, -8.52228404459559e-08, -29412.2500000731, -14706.1250000947, 0, 0, 0, 0, 0, 0, 0, 0, 1.11363109881483e-08, 0, 0, 0, 34314.2916666325, 0, 1.19373100260849e-08, 0, 0, 3.12842136954675e-08, 0, 0, 0, 68628.5833332705, 0, 34314.291666636, -1.24132365954056e-08, 0, 0, 68628.5833333142, 5.16058805492031e-08, 4.75326252746704e-08, -1.57648614010992e-08, 3.33542564924402e-08, 68628.5833333385, 3.79061581523764e-08, 34314.2916667113, 1.10248366747675e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -171.499999999904, 0, 0, 0, 0, 0, 0, 0, 1572.08333333062, 0, 2172.33333333319, 371.583333334258, 0, 0, 0, 0, 0, 0, -2801.1666666557, 0, -8303.45833333788, 0, 0, -19508.1249999804, -200.083333336929, -3001.24999999643, -4301.7916666801, 0, 0, 0, 0, 0, 0, 1400.58333331534, 0, 11204.6666666826, 0, 0, 0, 32213.4166666001, 0, 71429.7499999653, 0, 2801.16666664237, 13305.5416667064, 1400.58333333133, 2100.87500004417, 34314.2916666845, 0, 0, 0, 0, 0, 0, 1.3272536923649e-08, 0, -4902.04166668584, 0, 0, 0, -14706.1249999223, 0, -88236.7499999811, 0, 3.82708616975819e-08, -9804.08333337935, 0, 0, -98040.8333332382, 1.58434909193746e-08, -3.95850603233729e-08, 2.83114020759925e-08, -29412.2499999854, -4902.04166674655, -102942.875000037, -14706.1250000983, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.06757644611359e-08, 0, 34314.2916666668, 0, -1.78888255994255e-08, 2.08524636941224e-08, 0, 0, 0, 34314.2916666091, 0, 68628.583333293, -1.22136460632629e-08, 0, -2.49347071154579e-08, -2.54691346727874e-08, 4.11816518086941e-08, 68628.5833333496, -2.52071871058515e-08, 4.20147045487651e-08, 68628.5833333304, 5.24346169417517e-08, 4.77309605364126e-08, 34314.2916667625}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.28365617609987e-08, 12005.0000000118, 0, 0, 0, 0, 0, 0, -1.53090211244647e-08, 0, -5.17209780587004e-08, 0, 0, -6.14732901223868e-08, -63826.583333373, -63826.5833334053, -63826.5833334248, 0, 0, 0, 0, 0, 0, 1.37527909499984e-08, 0, 9.61506655550359e-08, 0, 0, 0, 1.16501961942655e-07, 0, 1.75431652385719e-07, 125352.208333392, 125352.208333454, 125352.208333561, 250704.416666887, 250704.416666861, 250704.416667128, 0, 0, 0, 0, 0, 0, 0, 0, -8.41897919342263e-08, 0, 0, 0, -9.91789007718164e-08, 0, -2.11581873030746e-07, -107844.916666705, -107844.916666728, -107844.916666911, 0, 0, -2.11761321488107e-07, -323534.750000262, -323534.75000015, -323534.750000253, -323534.750000647, -323534.750000278, -323534.750000728, -647069.500000855, 0, 0, 0, 0, 0, 0, 0, 0, 2.77719038478089e-08, 0, 0, 0, 3.02891876655709e-08, 0, 9.13930499441894e-08, 34314.291666676, 34314.2916666676, 34314.2916667644, 0, 0, 0, 8.98157734637562e-08, 0, 1.21107583149273e-07, 137257.166666771, 137257.166666699, 137257.16666673, 137257.166666936, 137257.166666798, 137257.166667022, 205885.750000162, 205885.75000009, 205885.750000526, 411771.500000451, 411771.50000063, 411771.500000624}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -600.25000000044, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 8303.45833333432, 600.250000003725, 600.250000003753, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.32729119761592e-08, 0, -2.08166991836375e-08, -37115.4583333339, 0, 0, -7703.20833334277, -7703.20833333796, -3.0614998557615e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.27219092822102e-08, 63726.5416666664, 0, 0, 0, 0, 3.12613956826559e-08, 29412.250000008, 29412.2500000009, 1.29947118619284e-08, 4.391785126639e-08, 0, 5.81915403241332e-08, 3.99110356783659e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -34314.2916666664, 0, 0, 0, 0, 0, -1.27853374852058e-08, 0, -1.85179871380858e-08, -34314.2916666687, -34314.2916666659, 0, -1.56074079301542e-08, 0, -3.02945241456473e-08, 0, 0, -4.71742521686526e-08, -1.50655177501525e-08, -3.1804752584506e-08, -3.67040574374472e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -600.250000001127, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.0585443429329e-08, 600.250000002432, 8303.45833334416, 600.250000006044, 0, 0, 0, 0, 0, 0, -1.66371644480829e-08, 0, 0, 0, 0, 0, -3.23499834797873e-08, 0, -1.82503958690918e-08, 0, -37115.4583333636, -1.19503616357438e-08, -7703.20833335358, 0, -7703.20833337899, 0, 0, 0, 0, 0, 0, 1.38997616780133e-08, 0, 0, 0, 0, 0, 3.87965520526544e-08, 0, 1.4047313889376e-08, 0, 63726.5416667012, 1.0666687220649e-08, 0, 0, 3.78049156830451e-08, 1.81219816434463e-08, 0, 29412.2500000345, 29412.2500000839, 0, 6.00887837911628e-08, 6.19108187640234e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.60910502742408e-08, 0, 0, 0, -34314.2916666808, 0, 0, 0, 0, -2.26638739024346e-08, 0, -1.48530123263323e-08, 0, 0, -34314.2916666842, -34314.291666713, 0, -2.57726395521257e-08, -1.48183084387294e-08, 0, -5.48169213727297e-08, -2.7296699449906e-08, -5.55715080989086e-08, -4.15675066579867e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -600.249999999702, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 600.249999997266, 600.249999998279, 8303.45833333329, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.20311277522444e-08, 0, 0, -37115.4583333366, 1.08902198463825e-08, -7703.20833332739, -7703.20833333323, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.52243233996314e-08, 0, 0, 63726.5416666738, 0, 0, 1.55873310263246e-08, -1.77777077544491e-08, -1.20192043540992e-08, -1.73158012038135e-08, 0, 29412.2499999964, 29412.2500000048, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -34314.2916666711, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -34314.2916666669, -34314.2916666706, 1.40636069128518e-08, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6002.50000000632, 0, 0, 0, 0, 0, 0, 0, 0, 2.61295755712163e-08, 0, 0, 3.5397502692167e-08, 63426.4166666913, 21408.9166667022, 21408.916666712, 0, 0, 0, 0, 0, 0, 0, 0, -4.48040104541554e-08, 0, 0, 0, -6.32644105997577e-08, 0, -1.04165443024124e-07, -175072.916666708, -25210.500000056, -25210.5000001129, -200283.416666796, -200283.41666677, -50421.0000002089, 0, 0, 0, 0, 0, 0, -1.55251002258656e-08, 0, 3.60082692531587e-08, 0, 0, 0, 4.54932950785264e-08, 0, 1.22009460032992e-07, 186277.583333365, 9804.08333335154, 9804.08333345433, 0, 0, 1.30576473126528e-07, 382359.250000165, 382359.250000095, 205885.750000172, 29412.2500002904, 205885.750000152, 29412.2500003331, 411771.50000038, 0, 0, 0, 0, 0, 0, 0, 0, -1.08493825094085e-08, 0, 0, 0, 0, 0, -5.02848822147531e-08, -68628.5833333424, 0, -4.71082165283669e-08, 0, 0, 0, -5.37462670848735e-08, 0, -7.60046342609524e-08, -205885.750000067, -205885.75000003, -68628.5833333932, -1.18702245777178e-07, -68628.5833334101, -1.67768091058676e-07, -205885.750000123, -205885.750000062, -2.38526969388409e-07, -411771.500000207, -205885.750000287, -205885.750000264}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2401.00000000238, 0, 0, 0, 0, 0, 0, 0, 0, -1.06882290450712e-08, 0, 0, -1.68357616692181e-08, -30412.6666666742, -5202.16666668077, -5202.16666668556, 0, 0, 0, 0, 0, 0, 0, 0, 1.65043869616587e-08, 0, 0, 0, 2.73735814539108e-08, 0, 5.33668746948026e-08, 116248.416666676, 2801.16666668665, 2801.16666671586, 60225.0833333799, 60225.0833333696, 5602.33333342863, 0, 0, 0, 0, 0, 0, 1.23386797910681e-08, 0, -1.14444245752684e-08, 0, 0, 0, -1.301901519262e-08, 0, -6.08523469882979e-08, -156865.333333338, 0, -5.07391127460953e-08, 0, 0, -7.00559985016714e-08, -186277.583333384, -186277.583333359, -29412.2500000675, -1.2502338299876e-07, -29412.2500000606, -1.73073688155843e-07, -58824.500000149, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.33559079595192e-08, 68628.583333334, 0, 1.79597213433207e-08, 0, 0, 0, 2.65214697348772e-08, 0, 4.26899560393655e-08, 137257.166666684, 137257.166666672, 2.58064868338533e-08, 4.3713447074266e-08, 3.15237473994783e-08, 9.17980835334342e-08, 68628.5833333773, 68628.5833333538, 1.23166073445926e-07, 137257.166666734, 1.12056250698685e-07, 1.24106150599722e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -400.166666667386, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4601.91666666818, 3201.33333333708, 400.166666672437, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -14005.8333333327, -2801.16666667425, -1.80277225569916e-08, -36415.1666666806, -4201.75000000953, -2801.1666666814, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9804.08333332979, 0, 2.36945848204911e-08, 0, 0, -1.4818499148951e-08, 107844.916666681, 9804.08333333391, 29412.2500000258, 1.82436395414899e-08, 2.37597151912387e-08, 2.13252224615588e-08, 29412.2500000294, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.08370113921218e-08, 0, 0, 0, 0, 0, 0, -68628.5833333378, 0, -1.25769762830068e-08, 0, -1.71286498582028e-08, -1.20454929250705e-08, -68628.5833333507, 0, 0, -68628.5833333446, -2.61435705447182e-08, -2.13320460751271e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -400.16666666736, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3201.33333333582, 4601.91666667384, 400.166666668542, 0, 0, 0, 0, 0, 0, -1.09055577355686e-08, 0, 0, 0, 0, 0, 0, 0, 0, -2801.16666667077, -14005.8333333573, 0, -36415.166666688, -2801.16666666966, -4201.75000000981, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.25890799657437e-08, 0, 0, 0, 9804.08333336391, 1.22403665423789e-08, 0, 0, -1.68308577248627e-08, 29412.2500000266, 0, 107844.916666707, 9804.08333336698, 0, 0, 29412.2500000302, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.29792963361652e-08, 0, 0, 0, 0, 0, 0, 1.29090820567459e-08, -1.1023436880924e-08, 0, -68628.5833333556, -2.8350908568838e-08, 0, 0, -68628.5833333572, 0, 0, -2.02014426909083e-08, -68628.5833333718, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2401.00000000457, 0, 0, 0, 0, 0, 0, -2.59148694512392e-08, 0, -1.13330978655808e-08, 0, 0, -2.7180100844688e-08, -5202.16666667818, -30412.6666667064, -5202.1666666953, 0, 0, 0, 0, 0, 0, 4.75776967040402e-08, 0, 1.99405093814307e-08, 0, 0, 0, 7.87582319603521e-08, 0, 4.72146763514073e-08, 2801.16666667976, 116248.416666769, 2801.16666673315, 60225.0833334221, 5602.33333338165, 60225.0833335155, 0, 0, 0, 0, 0, 0, -4.04502330663725e-08, 0, -1.79447435384717e-08, 0, 0, 0, -9.73561733462582e-08, 0, -4.42534321751906e-08, 0, -156865.333333443, -7.14941666608126e-08, 0, 0, -7.96851622957872e-08, -29412.250000089, -2.13809914469171e-08, -186277.583333469, -186277.58333364, -6.48143130713173e-08, -29412.2500002516, -58824.5000002996, 0, 0, 0, 0, 0, 0, 1.2921119149917e-08, 0, 0, 0, 0, 0, 4.22100387239303e-08, 0, 1.78749007137451e-08, 0, 68628.5833333765, 2.93570361156219e-08, 0, 0, 0, 4.76743925751461e-08, 0, 3.0065152323872e-08, 3.19414987371139e-08, 0, 137257.166666727, 137257.166666827, 3.1685095653965e-08, 1.1567247381264e-07, 68628.5833333995, 0, 68628.5833335363, 1.44940605491563e-07, 137257.166666917, 2.08189587212911e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6002.50000001006, 0, 0, 0, 0, 0, 0, 2.77239135350651e-08, 0, 3.42409800096987e-08, 0, 0, 4.33196165116616e-08, 21408.9166666953, 63426.4166667387, 21408.9166667408, 0, 0, 0, 0, 0, 0, -4.63051805544384e-08, 0, -6.62837758200359e-08, 0, 0, 0, -9.87083601520725e-08, 0, -9.9460114659434e-08, -25210.5000000345, -175072.916666819, -25210.5000001853, -200283.416666856, -50421.0000001381, -200283.416667067, 0, 0, 0, 0, 0, 0, 3.64290798874139e-08, 0, 6.12536121774758e-08, 0, 0, 0, 1.08810010107135e-07, 0, 1.16950978664044e-07, 9804.08333335001, 186277.583333464, 9804.08333353855, 0, 0, 1.19650418584174e-07, 205885.750000208, 29412.2500000766, 382359.250000247, 382359.250000602, 29412.2500002036, 205885.750000602, 411771.500000709, 0, 0, 0, 0, 0, 0, -1.07787906130063e-08, 0, -2.12011881490713e-08, 0, 0, 0, -4.42882862248186e-08, 0, -5.32871635997661e-08, 0, -68628.5833333737, -8.48532094684432e-08, 0, 0, 0, -5.87721633246789e-08, 0, -5.73607402945364e-08, -68628.5833334121, 0, -205885.750000084, -205885.750000279, -1.03335655622751e-07, -68628.583333624, -205885.75000014, -4.13846083110638e-08, -205885.750000444, -205885.750000351, -411771.500000552, -205885.750000513}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6002.50000000672, 0, 0, 0, 0, 0, 0, 0, 0, 4.0059568079685e-08, 0, 0, 4.20181654108473e-08, 21408.9166666906, 21408.916666702, 63426.4166667249, 0, 0, 0, 0, 0, 0, 0, 0, -7.50801076555659e-08, 0, 0, 0, -6.81940516017902e-08, 0, -1.39229849275618e-07, -25210.500000037, -25210.5000000392, -175072.916666812, -50421.0000001152, -200283.416666799, -200283.416666948, 0, 0, 0, 0, 0, 0, -1.29104955546311e-08, 0, 6.53151099618777e-08, 0, 0, 0, 4.38933619659549e-08, 0, 1.73894759850329e-07, 9804.08333335904, 9804.08333332594, 186277.583333486, 0, 0, 1.64168219517602e-07, 29412.2500001394, 205885.750000114, 29412.2500000941, 205885.750000362, 382359.250000186, 382359.250000466, 411771.500000551, 0, 0, 0, 0, 0, 0, 0, 0, -2.14412072724601e-08, 0, 0, 0, 0, 0, -7.45619775731685e-08, 0, 1.83356133825062e-08, -68628.5833333935, 0, 0, 0, -6.36688496464773e-08, 0, -1.03167187787966e-07, -5.6230445755655e-08, -68628.5833333623, 0, -68628.5833334649, -205885.75000008, -205885.750000228, -6.24426041725988e-08, -205885.750000071, -205885.750000325, -205885.750000301, -205885.750000374, -411771.500000415}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2401.00000000146, 0, 0, 0, 0, 0, 0, 0, 0, -1.65552325412017e-08, 0, 0, -1.71009971743708e-08, -5202.16666666904, -5202.16666667504, -30412.6666666822, 0, 0, 0, 0, 0, 0, 0, 0, 3.2096046769853e-08, 0, 0, 0, 3.19754947618214e-08, 0, 5.3791677633342e-08, 2801.16666666545, 2801.16666667792, 116248.416666711, 5602.33333335112, 60225.0833333599, 60225.0833334071, 0, 0, 0, 0, 0, 0, 0, 0, -2.91087433737498e-08, 0, 0, 0, -2.67094649305902e-08, 0, -6.68062575860515e-08, 0, 0, -156865.333333387, 0, 0, -6.68107181049969e-08, -1.30428438560637e-08, -29412.2500000105, 0, -29412.2500001034, -186277.583333379, -186277.583333453, -58824.5000001372, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.91244467705088e-08, 0, 0, 68628.5833333577, 0, 0, 0, 2.90874559495e-08, 0, 3.97161024303659e-08, 0, 0, 0, 4.31282776939938e-08, 137257.16666669, 137257.166666725, 0, 68628.5833333444, 68628.5833334203, 6.93372393954388e-08, 9.84411240465011e-08, 137257.166666773}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -400.166666667404, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4601.91666667061, 400.166666669383, 3201.33333333831, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.47931284605916e-08, -14005.833333342, 0, -2801.16666667868, -4201.75000001285, -36415.1666666832, -2801.16666668188, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.72404919286106e-08, 9804.08333334188, 0, 1.19279599620179e-08, 0, 0, 1.88768017203635e-08, 9804.08333335161, 107844.916666689, 1.52821782363688e-08, 1.37938209573982e-08, 29412.2500000237, 2.68552400163149e-08, 29412.2500000337, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.18369074358699e-08, 0, -68628.5833333437, 0, 0, -1.09116037138012e-08, -1.4744445867631e-08, -1.16707552097933e-08, -68628.5833333489, -1.38309413727367e-08, -68628.5833333566, -1.7894725013517e-08, -2.36443959147336e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -400.166666667425, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3201.33333333747, 400.166666671761, 4601.91666667048, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.61679648407608e-08, 0, -1.51951381213133e-08, -2801.16666667554, -1.06538101139521e-08, -14005.8333333384, -2801.16666668582, -36415.1666666818, -4201.75000002309, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.71139989679402e-08, 0, 1.51879336372477e-08, 0, 0, 9804.08333333482, 0, 0, 2.58679458938528e-08, 2.71164994743943e-08, 29412.2500000217, 2.4052355518483e-08, 3.79520757776095e-08, 107844.91666668, 9804.08333335972, 29412.250000054, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.36548404410258e-08, 0, -1.27664241583523e-08, -1.22401912924723e-08, 0, 0, -1.89593740070447e-08, -68628.5833333358, 0, -1.6754656844748e-08, -68628.5833333442, -2.43978020269957e-08, -3.75158644623609e-08, -3.88994348153166e-08, -68628.5833333628}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -400.16666666803, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 400.166666670355, 4601.91666667561, 3201.3333333458, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.480633468809e-08, 0, -14005.8333333468, -2801.16666669848, -4201.75000002398, -2801.16666669179, -36415.1666667347, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.97259810808878e-08, 0, 9804.08333333967, 3.26114309091312e-08, 0, 0, 1.25782977618169e-08, 2.40855797510948e-08, 1.54251271019401e-08, 9804.08333335626, 107844.916666753, 3.83580133075168e-08, 29412.2500001183, 29412.2500001264, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.20684488339747e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, -68628.583333364, -1.70371271407104e-08, -5.97355446468061e-08, -1.19266479703019e-08, -1.00987250833976e-08, -68628.5833334132, -6.18323430603e-08, -68628.58333342, -1.05291246633636e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -400.166666668041, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 400.16666667066, 3201.33333334047, 4601.91666667902, 0, 0, 0, 0, 0, 0, 0, 0, -1.48185359979831e-08, 0, 0, 0, 0, 0, -1.55750507013598e-08, 0, -2801.16666667594, -14005.8333333669, -2801.16666668802, -4201.75000002446, -36415.1666667196, 0, 0, 0, 0, 0, 0, 0, 0, 1.37548412464913e-08, 0, 0, 0, 0, 0, 2.21633692730589e-08, 0, 0, 9804.08333337179, 0, 0, 1.11984951593524e-08, 2.25609868387846e-08, 1.42265216531499e-08, 1.99154831021191e-08, 29412.2500000662, 9804.0833333742, 107844.916666755, 29412.2500001005, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.05974764388295e-08, 0, 0, -1.6095151673268e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.44608867322515e-08, -2.17362973250815e-08, -68628.5833333781, -1.11302896715741e-08, -1.12901707205349e-08, -68628.5833333899, -4.85506145453133e-08, -6.94553759410864e-08, -68628.5833334126}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1800.75000000282, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -16506.875000009, -16506.8750000184, -3901.62500001837, 0, 0, 0, 0, 0, 0, 0, 0, 2.01421567277972e-08, 0, 0, 0, 0, 0, 0, 29412.2500000114, 29412.2500000424, 2100.87500005163, 147061.250000062, 31513.1250000329, 31513.1250000721, 0, 0, 0, 0, 0, 0, 0, 0, -2.1080535233945e-08, 0, 0, 0, -1.17361645550858e-08, 0, 0, -14706.1250000049, -14706.1250000379, -6.92148920085671e-08, 0, 0, 1.44358337633745e-08, -235298.000000079, -29412.2500000142, -235298.000000098, -29412.2500001215, -14706.125000058, -14706.1250000854, -250004.125000143, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.07356140483808e-08, 3.34344536783334e-08, 0, 0, 0, 0, 0, -1.08324275877909e-08, 102942.875000032, 0, 102942.875000042, 6.7467561508972e-08, 4.05588572597523e-08, 4.24256805452888e-08, 205885.750000067, 1.17861884521691e-08, 5.818982044687e-08, 205885.750000076, 205885.750000132, 102942.875000078}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1800.75000000298, 0, 0, 0, 0, 0, 0, 0, 0, -1.25628881918943e-08, 0, 0, -2.22265736400119e-08, -16506.875000016, -3901.62500001413, -16506.8750000194, 0, 0, 0, 0, 0, 0, 0, 0, 1.93898437411873e-08, 0, 0, 0, 4.3871694713257e-08, 0, 6.63988735855632e-08, 29412.2500000352, 2100.87500001688, 29412.2500000394, 31513.1250000607, 147061.250000064, 31513.1250000883, 0, 0, 0, 0, 0, 0, 0, 0, -1.34917493541775e-08, 0, 0, 0, -3.40350747221716e-08, 0, -7.46906152385612e-08, -14706.1250000337, 0, -14706.1250000319, 0, 0, -9.45104894499525e-08, -29412.2500000899, -235298.000000086, -14706.1250000641, -14706.1250001145, -235298.000000074, -29412.2500001357, -250004.125000185, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.86542853273399e-08, 1.18978520157053e-08, 0, 0, 0, 0, 0, 4.11684794353947e-08, 0, 5.49176920090061e-08, 4.19024918759674e-08, 102942.875000039, 1.60441752944194e-08, 4.27698762587562e-08, 102942.875000025, 6.27498205824375e-08, 5.17086000690192e-08, 205885.750000049, 9.88861284886717e-08, 205885.750000123, 102942.875000119, 205885.750000117}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -300.1250000007, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2401.00000000324, 2401.00000000251, 2401.0000000062, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.11868070132536e-08, -2100.87500000693, -2100.87499999779, -2100.87500001435, -18907.8750000112, -18907.8750000139, -18907.8750000314, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.56493691500691e-08, 0, -1.14235803189508e-08, 1.27499446706231e-08, 0, 0, 2.73490146098277e-08, 14706.1250000183, 14706.1250000149, 14706.1250000045, 14706.1250000382, 14706.1250000163, 14706.1250000582, 147061.25000005, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.02147954746729e-08, 0, 0, 0, 0, 0, 0, 0, 0, -1.79728316750627e-08, 0, 0, 0, -1.08171219563334e-08, 0, -2.97803900515243e-08, 0, 0, -4.32472565267892e-08, -102942.875000027, -102942.875000032, -102942.875000037}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1800.7500000049, 0, 0, 0, 0, 0, 0, 0, 0, -2.27269198500583e-08, 0, 0, -1.73337446475107e-08, -3901.6250000154, -16506.875000027, -16506.8750000444, 0, 0, 0, 0, 0, 0, 0, 0, 4.42109024636298e-08, 0, 0, 0, 1.67033532402621e-08, 0, 6.5688824048142e-08, 2100.87500001958, 29412.2500000303, 29412.2500001139, 31513.1250000824, 31513.1250000916, 147061.250000221, 0, 0, 0, 0, 0, 0, 0, 0, -3.93649571072812e-08, 0, 0, 0, 0, 0, -8.8412371673088e-08, -1.04730822579844e-08, -14706.1249999965, -14706.1250001196, 0, 0, -5.9757081365774e-08, -14706.1250000929, -14706.1250000626, -29412.2500000693, -235298.000000274, -29412.2500001371, -235298.000000382, -250004.125000411, 0, 0, 0, 0, 0, 0, 0, 0, 1.31047423427002e-08, 0, 0, 0, 0, 0, 3.97095971540726e-08, 0, -1.16241377066809e-08, 4.57707681108061e-08, 0, 0, 0, 1.51951687802446e-08, 0, 4.37605924650564e-08, 3.60097013300399e-08, 0, 0, 102942.875000094, 6.28220783764172e-08, 102942.875000192, 4.35267108039405e-08, 3.95004485927345e-08, 205885.750000257, 102942.875000205, 205885.75000028, 205885.75000033}};

// -----------------------------------------------------------------------------
double MAd_TetInterpolant::coef8[165][165] = {{1, -21.7428571428631, -21.7428571428436, -21.7428571428631, 187.498412698525, 187.498412698451, 187.498412698525, 374.996825395815, 374.996825396696, 374.99682539668, -854.400000000831, -854.400000002595, -854.400000000831, -2563.19999999746, -2563.19999999946, -2563.19999998049, -2563.19999999604, -2563.19999999655, -2563.20000000183, -5126.40000001363, 2280.53333333654, 2280.53333334995, 2280.53333333654, 9122.13333334139, 9122.13333333537, 9122.13333322138, 9122.13333328991, 9122.13333331315, 9122.1333333538, 13683.1999999152, 13683.199999975, 13683.2000000176, 27366.4000001028, 27366.4000000432, 27366.4000002004, -3686.40000000704, -3686.40000004583, -3686.40000000704, -18432.0000000562, -18432.0000000135, -18431.9999996929, -18431.9999998612, -18431.9999999526, -18432.0000000388, -36863.9999998111, -36863.9999999486, -36863.9999996119, -36863.9999998271, -36863.9999998562, -36864.0000003179, -73728.0000001971, -73727.9999998958, -73728.0000009647, -110592.000000556, -110592.000001056, -110592.000000677, 3549.86666667518, 3549.8666667304, 3549.86666667518, 21299.2000001154, 21299.2000000228, 21299.1999995526, 21299.1999998163, 21299.1999999438, 21299.2000000054, 53247.9999997953, 53247.9999999921, 53247.9999992533, 53247.9999993144, 53247.9999997117, 53248.0000007778, 106496.000000104, 106495.999999602, 106496.000001765, 70997.3333326341, 70997.3333330193, 70997.3333337039, 212992.00000112, 212992.00000168, 212992.000001137, 212991.999999468, 212992.000003755, 212992.000003597, 319488.000003197, -1872.45714286264, -1872.45714290118, -1872.45714286264, -13107.2000001018, -13107.2000000156, -13107.1999996706, -13107.1999998898, -13107.1999999689, -13107.199999969, -39321.5999999251, -39321.6000000435, -39321.5999993125, -39321.5999992381, -39321.5999997597, -39321.6000006478, -78643.1999999574, -78643.199999632, -78643.2000013622, -65535.9999993583, -65535.999999774, -65535.9999991878, -65535.999999629, -65535.9999995015, -65536.0000011788, -196608.000000727, -196608.000000933, -196608.000001107, -196607.999997587, -196608.000004699, -196608.00000535, -262144.000002228, -262144.000004363, -262144.000002091, -393216.000003585, -393216.000001835, -393216.000008599, 416.101587303038, 416.101587313619, 416.101587303038, 3328.81269844594, 3328.81269841665, 3328.81269831625, 3328.81269838837, 3328.81269840596, 3328.81269839639, 11650.8444444516, 11650.8444444682, 11650.8444441951, 11650.8444441786, 11650.8444443741, 11650.8444446115, 23301.6888888491, 23301.6888887899, 23301.6888892628, 23301.6888886693, 23301.6888888546, 23301.6888885602, 23301.6888884681, 23301.6888886593, 23301.6888895368, 69905.0666667819, 69905.0666667815, 69905.0666671027, 69905.0666653007, 69905.0666685626, 69905.0666690166, 29127.1111107364, 29127.1111108875, 29127.1111113931, 116508.44444544, 116508.444445959, 116508.444445774, 116508.444443643, 116508.444447307, 116508.444447118, 174762.666667655, 174762.666666383, 174762.666672677, 233016.888890646, 233016.888893728, 233016.888892259}, 
                                                    {0, -1.00000000000189, 0, 0, 20.7428571428917, 0, 0, 0, 0, 0, -166.755555555809, 0, 0, 0, 0, 0, 0, 0, 0, 0, 687.644444445422, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1592.888888891, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2093.51111111369, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1456.35555555721, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.19616743177176e-08, 0, 0, 416.10158730202, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, -1.00000000000022, 0, 0, 20.7428571428622, 0, 0, 0, 0, 0, -166.75555555559, 0, 0, 0, 0, 0, 0, 0, 0, 0, 687.644444444604, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1592.88888888921, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2093.51111111149, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1456.35555555583, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 416.10158730165, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, -1.00000000000189, 0, 0, 20.7428571428917, 0, 0, 0, 0, 0, -166.755555555809, 0, 0, 0, 0, 0, 0, 0, 0, 0, 687.644444445422, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1592.888888891, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2093.51111111369, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1456.35555555721, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.17579475045204e-08, 0, 0, 416.10158730202, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 64.000000000039, 0, 0, -879.542857143586, 0, 0, -879.542857139218, 0, -879.542857142639, 4963.55555556099, -1.11805524307428e-08, 0, 9927.11111110319, 9927.11111111353, 4963.55555548533, 2.20956643999743e-08, 4963.55555554703, 5.68509292850263e-08, 9927.11111118267, -14973.1555555764, 1.47055766128975e-08, 0, -44919.4666666936, -44919.4666666905, -14973.1555551542, 0, -14973.1555554978, -2.87664569247937e-07, -44919.4666663926, -44919.4666666305, -5.17735792056619e-07, -89838.9333336865, -44919.4666671122, -44919.4666675697, 26168.8888889344, 1.5275020671959e-08, 0, 104675.555555747, 104675.555555633, 26168.8888877833, -1.85391491355767e-07, 26168.8888887404, 6.16585605261984e-07, 157013.33333266, 157013.333333331, 104675.5555544, 1.23929158775551e-06, 104675.55555524, 2.46786777912371e-06, 314026.666667238, 104675.555556437, 104675.555559489, 314026.66666882, 314026.666670172, 157013.333337811, -26578.4888889442, -6.03098124219159e-08, 0, -132892.444444853, -132892.444444551, -26578.4888872598, 3.98071661585259e-07, -26578.4888886906, -6.31496186032241e-07, -265784.888888014, -265784.888889129, -132892.444442349, -9.75152605922312e-07, -132892.444443728, -4.46526808321335e-06, -531569.777777991, -132892.444445139, -132892.44445156, -265784.888886735, -265784.888888325, -5.5679982271175e-06, -797354.66667, -797354.6666717, -531569.777783018, -265784.888894961, -531569.777789782, -265784.888905048, -797354.66667909, 14563.5555555911, 5.78414610893082e-08, 0, 87381.3333337075, 87381.3333333997, 14563.5555543425, -3.33689048606292e-07, 14563.5555554355, 3.0981266609932e-07, 218453.333332878, 218453.333333659, 87381.3333314633, 2.76567516474803e-08, 87381.3333326967, 3.4002954691495e-06, 436906.666666469, 87381.3333334249, 87381.3333390469, 291271.111108892, 291271.111110863, 218453.33333119, 4.4134605464914e-06, 218453.333332178, 7.61966403216119e-06, 873813.333335021, 873813.333336156, 436906.666672293, 218453.333335449, 436906.666681824, 218453.33335423, 873813.333340535, 873813.333346007, 291271.111128927, 1310720.00001147, 873813.333345936, 873813.333363205, -3328.81269842202, -1.88350191897896e-08, 0, -23301.6888890147, -23301.6888889046, -3328.81269805217, 9.89153359210391e-08, -3328.81269838196, -5.91808154658225e-08, -69905.0666666239, -69905.0666667926, -23301.6888882059, 1.65473282267993e-07, -23301.6888886915, -9.04054167745321e-07, -139810.133333195, -23301.6888888231, -23301.6888905994, -116508.444443587, -116508.444444517, -69905.0666659157, -1.02288403260635e-06, -69905.0666660767, -3.20295384006555e-06, -349525.333333418, -349525.33333376, -139810.133335561, -69905.0666662451, -139810.133339582, -69905.0666753547, -116508.444443336, -116508.444443969, -3.45197773261974e-06, -466033.777780504, -466033.777782053, -349525.333337925, -116508.4444496, -349525.333341631, -116508.444457384, -699050.666669645, -349525.333337158, -349525.333353178, -699050.666673209, -699050.666681494, -466033.777793052}, 
                                                    {0, -112.000000000114, 0, 0, 1987.20000000212, -1.39221084190203e-08, 0, 1091.19999999353, 0, 1091.20000000016, -13051.022222238, 8.6830322559756e-08, 0, -17372.4444444346, -17372.4444444547, -4321.42222209699, -7.73025491440899e-08, -4321.42222221586, -1.17262289023112e-07, -8642.84444457092, 43488.7111111716, -2.7531237351884e-07, 0, 95894.7555556049, 95894.7555556109, 8917.33333260055, 2.41934635819598e-07, 8917.33333326791, 5.49558924535519e-07, 61323.3777773934, 61323.3777778168, 1.10865092942698e-06, 122646.755556002, 26752.000001011, 26752.0000014568, -81464.8888890202, 4.85510814247428e-07, 0, -254520.888889201, -254520.888889021, -10126.2222201593, -3.07277932381044e-07, -10126.22222202, -1.19217594839299e-06, -274773.33333238, -274773.333333639, -111843.555554001, -3.34650807669293e-06, -111843.555555438, -4.48963494923462e-06, -549546.666667069, -40504.8888919187, -40504.8888950256, -335530.666669888, -335530.666671212, -60757.3333417714, 87017.2444446047, -4.85969020206659e-07, 0, 354076.444445108, 354076.444444591, 6007.46666355174, 1.52993595404821e-07, 6007.46666633227, 1.30626716856558e-06, 546133.333332032, 546133.333334114, 111047.111108389, 4.51314404464969e-06, 111047.111110582, 7.58349482273975e-06, 1092266.6666662, 30037.3333378435, 30037.3333444294, 384113.777774948, 384113.777777981, 1.10976077087387e-05, 1152341.33333683, 1152341.33333868, 444188.444453705, 60074.6666829029, 444188.444460276, 60074.6666933198, 666282.666685004, -49516.0888889915, 2.58117894976395e-07, 0, -249036.80000061, -249036.800000076, -1456.35555318886, 4.5284072425956e-08, -1456.35555532445, -7.2156284872745e-07, -502442.666665979, -502442.666667369, -56797.8666642949, -2.755598837732e-06, -56797.8666660438, -5.63032783593019e-06, -1004885.33333245, -8738.13333646399, -8738.1333424342, -509724.444441338, -509724.444445079, -262143.999997575, -1.06126990689948e-05, -262143.999999463, -1.3265842869159e-05, -1529173.33333389, -1529173.33333545, -283989.333344523, -21845.3333467622, -283989.333353839, -21845.3333659257, -1048576.00000941, -1048576.00001482, -29127.1111462175, -1572864.00001325, -567978.666689984, -567978.666708996, 11650.8444444713, -5.61284658236514e-08, 0, 69905.0666668726, 69905.0666666807, -7.20413321796027e-07, -3.98756749611472e-08, -6.81235784069055e-08, 1.62108868702797e-07, 174762.666666605, 174762.666666889, 11650.8444435628, 6.10358083210566e-07, 11650.8444442085, 1.50285197311238e-06, 349525.33333295, 8.58899394233896e-07, 2.83422175672724e-06, 233016.888887611, 233016.888889327, 69905.066665969, 3.39167367426847e-06, 69905.066666211, 5.21028414264066e-06, 699050.666666037, 699050.666666597, 69905.0666714988, 3.92402078393056e-06, 69905.0666752695, 1.32803150746143e-05, 174762.666665262, 174762.66666658, 6.59271009728675e-06, 699050.66666937, 699050.666671067, 349525.333339997, 1.42712271362841e-05, 349525.333343423, 2.20885839442978e-05, 1048576.00000229, 174762.666676829, 174762.666694329, 699050.666675271, 699050.666685132, 233016.888913561}, 
                                                    {0, 149.333333333526, 0, 0, -2848.71111111469, 2.39682810557661e-08, 0, -1056.71111110332, 0, -1056.71111111169, 20403.2000000266, -1.59235548882188e-07, 0, 18568.5333333241, 18568.5333333481, 2943.99999984914, 8.78529520568004e-08, 2944.00000000069, 1.14458947877848e-07, 5888.00000010934, -73181.8666667688, 5.41518663178837e-07, 0, -116588.088888948, -116588.088888947, -4039.1111101926, -3.26995045482786e-07, -4039.1111110731, -5.22438585115534e-07, -47445.3333329568, -47445.3333334652, -1.07875507310581e-06, -94890.6666669507, -12117.3333343177, -12117.3333345364, 145408.000000221, -1.0270240463608e-06, 0, 344746.666667, 344746.666666761, 2730.6666640282, 5.76508340298704e-07, 2730.66666649922, 1.16987098840532e-06, 255999.999999219, 256000.000000575, 59391.9999983194, 3.43357541673784e-06, 59392.0000002933, 4.12180800028069e-06, 511999.999999952, 10922.6666700609, 10922.6666717333, 178176.000002323, 178176.000003111, 16384.0000075956, -162747.733333603, 1.10732693633733e-06, 0, -522467.555556224, -522467.555555607, -728.177773733955, -5.89315709724045e-07, -728.177777416144, -1.34807105486768e-06, -591644.444443595, -591644.444445492, -36408.8888856854, -5.19066494620632e-06, -36408.888889045, -6.87555960524814e-06, -1183288.88888781, -3640.88889466656, -3640.88889808871, -267605.33333065, -267605.333334701, -1.03743022754198e-05, -802816.000001313, -802816.000002697, -145635.555563233, -7281.77779430219, -145635.555567013, -7281.77780017153, -218453.333347254, 96119.4666668397, -6.32909799102357e-07, 0, 393216.000000591, 393215.999999993, -3.09063040035406e-06, 2.5654370346782e-07, -2.8035975404131e-07, 7.85357821292952e-07, 611669.333333112, 611669.333334056, 8738.13333043943, 3.66511855446974e-06, 8738.13333314604, 5.09666499792699e-06, 1223338.66666541, 4.49234171403287e-06, 7.64768535252599e-06, 436906.666664116, 436906.666668299, 131071.999997363, 1.0446801262891e-05, 131072.000000962, 1.17569835756634e-05, 1310719.99999858, 1310719.99999999, 43690.6666767987, 1.63207047908716e-05, 43690.6666821781, 2.6558054298269e-05, 524288.000005772, 524288.000009543, 3.21631080898904e-05, 786432.000006953, 87381.3333535619, 87381.3333653587, -23301.6888889342, 1.48320870494776e-07, 0, -116508.444444638, -116508.444444433, 9.48239145714443e-07, -3.88950205475644e-08, 9.06160710273939e-08, -1.85529880776846e-07, -233016.888888983, -233016.888889071, 1.08323451012045e-06, -9.50011154065056e-07, 1.52219829126403e-07, -1.37328326059708e-06, -466033.777777294, -1.38134225275269e-06, -2.45950075189662e-06, -233016.888887929, -233016.888889591, -23301.6888879667, -3.5778653177766e-06, -23301.6888890195, -4.52735228286903e-06, -699050.666665448, -699050.666665942, -4.71417287068408e-06, -5.67282270804149e-06, -6.57892676319117e-06, -1.07067265244875e-05, -116508.444443108, -116508.444445109, -5.9262137344764e-06, -466033.777778587, -466033.777780031, -116508.444449105, -1.45913366519735e-05, -116508.444451442, -1.8703507766803e-05, -699050.666666414, -1.01871015275906e-05, -2.08838959000254e-05, -233016.888894058, -233016.888901119, -1.99282013850269e-05}, 
                                                    {0, -140.000000000208, 0, 0, 2764.00000000385, -1.90883704179454e-08, 0, 710.666666660672, 0, 710.666666667349, -20721.7777778063, 1.28034465993512e-07, 0, -13100.4444444379, -13100.4444444575, -1338.66666654938, -4.62316725062748e-08, -1338.66666667143, -5.60652698265337e-08, -2677.33333338417, 78172.444444554, -4.41916198059788e-07, 0, 88078.2222222738, 88078.2222222637, 1109.333332596, 1.79487395998048e-07, 1109.33333332457, 2.43517398407176e-07, 22961.7777774938, 22961.7777779215, 5.39434619353478e-07, 45923.5555556772, 3328.000000465, 3328.00000054109, -162929.777778015, 8.48035208790009e-07, 0, -282168.888889149, -282168.888888927, -341.333331208228, -3.26756202360127e-07, -341.333333243058, -5.59760141993631e-07, -136533.333332886, -136533.333333869, -17635.5555540813, -1.74978853947324e-06, -17635.5555560044, -1.9883561144105e-06, -273066.666666587, -1365.33333502704, -1365.33333565895, -52906.6666675493, -52906.6666679672, -2048.00000359995, 190236.444444734, -9.251956578946e-07, 0, 461482.666667144, 461482.666666637, -3.26234796209769e-06, 3.98135843758815e-07, -2.57206910134922e-07, 6.64409704369154e-07, 357262.222222042, 357262.222223052, 5006.22221912565, 2.84163664103272e-06, 5006.22222276109, 3.28731542104922e-06, 714524.444443765, 3.05134272222817e-06, 4.18357509391414e-06, 91022.2222200722, 91022.2222238305, 5.14434455734674e-06, 273066.666666759, 273066.66666767, 20024.8888920469, 7.96562101771723e-06, 20024.8888938786, 1.00951791132643e-05, 30037.3333393689, -116508.44444463, 5.33168613929932e-07, 0, -371370.666667059, -371370.666666608, 2.4727020980203e-06, -1.94318920472974e-07, 2.16538628460081e-07, -3.90450895339631e-07, -415061.333333656, -415061.33333379, 2.89323653969883e-06, -2.16646135306782e-06, -1.60666314791696e-07, -2.33438270051289e-06, -830122.66666596, -2.42929180657836e-06, -3.42947833294932e-06, -182044.444442934, -182044.444446097, -21845.333330565, -5.32555535127685e-06, -21845.3333349151, -5.54349937734271e-06, -546133.333332007, -546133.333333112, -4.41668275144756e-06, -8.55466234440385e-06, -7.06631008356247e-06, -1.15930413317657e-05, -87381.3333351177, -87381.333337499, -1.52117823542878e-05, -131072.000001825, -9.15461516809831e-06, -1.37794428251909e-05, 29127.1111111596, -1.25652682674496e-07, 0, 116508.444444565, 116508.444444419, -7.55402700854968e-07, 3.26929781434268e-08, -7.38934459132909e-08, 9.28907912205681e-08, 174762.666666897, 174762.66666675, -1.06897464621087e-06, 5.71558168229011e-07, -4.12235725168921e-08, 5.92286011858256e-07, 349525.33333306, 7.87613312409618e-07, 1.1374975940195e-06, 116508.444444039, 116508.444445051, -1.24179570718023e-06, 1.87640151774476e-06, 4.71331758052695e-07, 2.09072364306769e-06, 349525.333332513, 349525.333332901, 2.20857425806759e-06, 3.01637363154422e-06, 2.92399241288412e-06, 4.55813231157687e-06, 29127.111109982, 29127.1111120148, 2.79882413895548e-06, 116508.444444227, 116508.444445331, 1.6067751239121e-06, 7.07318493963116e-06, 3.15821850571947e-06, 8.19472185576855e-06, 174762.666665886, 4.69627274734514e-06, 8.91065089387852e-06, 1.54218204577653e-06, 4.85302644165519e-06, 8.33879425224174e-06}, 
                                                    {0, 89.6000000001458, 0, 0, -1804.80000000269, 0, 0, -311.46666666416, 0, -311.466666667165, 13912.1777777977, -5.00279886887894e-08, 0, 5907.91111110911, 5907.9111111191, 358.399999949474, 1.24431214943735e-08, 358.400000003864, 1.33757994459424e-08, 716.800000014063, -54294.755555632, 1.71959736480904e-07, 0, -41426.4888889208, -41426.4888889139, -136.533333007962, -4.75400130188452e-08, -136.533333335575, -4.90476285440258e-08, -6382.93333320039, -6382.93333341808, -1.38603314767044e-07, -12765.8666667212, -409.600000107997, -409.600000128915, 117464.177777943, -3.26122816852073e-07, 0, 140174.222222364, 140174.222222245, -9.19176907057191e-07, 7.61664674853614e-08, -3.24182550684485e-08, 1.20837120452107e-07, 40277.3333331579, 40277.3333336276, 2275.55555478916, 4.47797877634144e-07, 2275.55555584125, 4.8700267617434e-07, 80554.6666667301, 4.06043322733431e-07, 5.79582997874407e-07, 6826.66666689065, 6826.66666702369, 9.1540771545631e-07, -142358.755555757, 3.53239907445402e-07, 0, -244303.644444684, -244303.644444428, 1.40239493244535e-06, -1.29559556744766e-07, 1.16705791238871e-07, -1.50637053859675e-07, -114688.000000078, -114688.000000426, 1.67243562031995e-06, -7.98289795375197e-07, -3.84297299103101e-07, -8.10383084825343e-07, -229375.999999933, -7.64149829662949e-07, -9.99751501948597e-07, -12743.1111099605, -12743.1111120429, -1.33281137352875e-06, -38229.3333334836, -38229.3333337781, -6.52783373752312e-07, -1.78858368285726e-06, -1.26167860028837e-06, -2.34726625730759e-06, -1.58325793758158e-06, 90294.0444445738, -2.00818500094703e-07, 0, 209715.200000183, 209715.199999969, -1.04187070920793e-06, 6.60114321573499e-08, -1.04215077186505e-07, 8.29156611653514e-08, 148548.266667022, 148548.266666875, -1.54972288777717e-06, 6.82449794838415e-07, 1.71008753745024e-07, 5.01420491383604e-07, 297096.533333266, 6.07900810746139e-07, 7.46762389180372e-07, 29127.1111104658, 29127.1111120211, -1.70774837145182e-06, 1.39329053088285e-06, 1.00533942557495e-06, 1.29915070611789e-06, 87381.3333330873, 87381.3333335587, 9.35788389456193e-07, 2.13711954588402e-06, 1.86198908885644e-06, 2.52000379845018e-06, 4.28389389097636e-07, 1.28867893017569e-06, 3.62215242942352e-06, 2.67467978346148e-07, 2.25600289293818e-06, 3.21838798651699e-06, -23301.6888889227, 4.65556681593976e-08, 0, -69905.0666667189, -69905.0666666532, 3.16309853019364e-07, 0, 3.6504569533984e-08, -1.84475249455151e-08, -69905.066666848, -69905.0666666973, 5.58032750655229e-07, -1.7217325642587e-07, 0, -9.83883077534824e-08, -139810.133333294, -2.16589790409013e-07, -2.6555274537688e-07, -23301.688888792, -23301.6888892048, 8.42394996152859e-07, -5.11375980168102e-07, -3.39641221539028e-07, -4.79163115254489e-07, -69905.0666664985, -69905.0666666804, -5.30289678106636e-07, -6.74232595715367e-07, -6.81768451251115e-07, -9.22313977430438e-07, 6.42510548300623e-07, -5.48726815870934e-07, -6.42427412304237e-07, 9.43783079282947e-08, -3.45550854347238e-07, -2.88504552410033e-07, -1.55181307130133e-06, -8.8758520496247e-07, -1.63806916447527e-06, 1.57846170881605e-07, -9.36639128790363e-07, -1.96130264331749e-06, -3.47500256840495e-07, -1.17267540841699e-06, -1.55762998424307e-06}, 
                                                    {0, -37.333333333398, 0, 0, 761.955555556748, 0, 0, 79.9999999995555, 0, 80.000000000231, -5984.0000000088, 0, 0, -1546.66666666712, -1546.66666666993, -42.6666666566364, 0, -42.6666666683024, 0, -85.3333333367166, 23918.9333333671, -2.79914509234261e-08, 0, 11158.7555555687, 11158.755555567, -6.78736232069157e-08, 0, 0, 0, 779.377777748122, 779.377777806425, 1.58429749872401e-08, 1558.75555558107, 1.24179974771298e-08, 1.8471693852008e-08, -53248.0000000731, 4.97662983305776e-08, 0, -39253.3333333825, -39253.333333349, 1.78744714112081e-07, 0, 0, 0, -5119.99999996642, -5120.00000009564, 1.93977597668808e-07, -4.67491859739796e-08, -9.50471234598482e-08, -4.86846242988147e-08, -10240.0000000724, -4.89499084035424e-08, -8.85023741292238e-08, -7.86517885912758e-08, -9.14118397730557e-08, -1.45686371818027e-07, 66628.2666667558, -5.17935396901928e-08, 0, 71907.5555556327, 71907.5555555622, -2.69145302982707e-07, 2.1782817122193e-08, -2.96198133310418e-08, 1.34097993357162e-08, 15473.7777778431, 15473.7777779165, -4.30681204019914e-07, 9.40347386422915e-08, 1.27690459983299e-07, 8.7903066994234e-08, 30947.555555643, 9.06703432387833e-08, 1.26812889803315e-07, -3.21527539241295e-07, 2.8490374914834e-07, 1.43953926805164e-07, 1.95275100824254e-07, 2.05961152539452e-07, 9.68589206419119e-08, 1.74821334798525e-07, 2.19401233368959e-07, 2.82266063383319e-07, 3.61672177849052e-07, -43690.6666667239, 2.7524392480445e-08, 0, -65536.0000000551, -65535.9999999988, 1.89300291942169e-07, -1.04017332705871e-08, 2.84650932106001e-08, 0, -21845.3333334872, -21845.3333334046, 3.81203458676407e-07, -1.14158334330172e-07, -6.36276034472757e-08, -3.74413418091982e-08, -43690.6666667512, -6.3934506515821e-08, -5.82504190342345e-08, 1.52756669307248e-07, -2.79039312574973e-07, 5.17352265787729e-07, -1.50584733000809e-07, -3.1941013669777e-07, -1.07580918214793e-07, -1.34561317010122e-07, -2.30778717093496e-07, -1.1201116039494e-07, -2.67561047400099e-07, -3.05081784646384e-07, -2.37841228976283e-07, -1.97531598561146e-07, -3.33882199972105e-07, -4.19872037837495e-07, -1.57401134051415e-07, -4.62408266749652e-07, -4.78330301797742e-07, 11650.8444444594, 0, 0, 23301.6888889036, 23301.688888887, -5.7374175884139e-08, 0, -1.00025864962469e-08, 0, 11650.844444515, 11650.8444444574, -1.31743558524877e-07, 2.83795220117978e-08, 0, 0, 23301.6888889115, 2.68808584138655e-08, 2.91947251866074e-08, 0, 9.80164211408939e-08, -2.73167194683878e-07, 6.19096083727637e-08, 1.14499284776426e-07, 3.97130767264462e-08, 4.03451689775952e-08, 9.02396638217295e-08, 7.46750393049531e-08, 4.86848358899101e-08, 6.48573220755054e-08, 5.74238953851685e-08, -1.91910657696413e-07, 1.73778394287484e-07, 4.4015833760216e-08, 6.10574465521228e-08, 1.30766055721154e-07, 7.24690030285235e-08, 9.95433714160281e-08, 1.62929747533629e-07, 9.02974616113032e-08, 1.05705353037274e-07, 6.0629726029493e-08, 2.21292628606566e-07, 1.95251991427238e-07, 2.41711230913324e-07, 1.56622537679252e-07}, 
                                                    {0, 9.14285714287374, 0, 0, -188.342857143163, 0, 0, -9.14285714286102, 0, -9.14285714292011, 1499.02222222448, 0, 0, 179.20000000042, 179.200000000788, 0, 0, 0, 0, 0, -6098.48888889754, 0, 0, -1319.82222222529, -1319.82222222535, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 13880.8888889076, 0, 0, 4778.66666667574, 4778.66666667226, 0, 0, 0, 0, 0, 1.72536639617548e-08, -1.4729851735665e-08, 0, 1.67039713493311e-08, 0, 2.26996957545509e-08, 0, 1.18593838407471e-08, 2.86534269491473e-08, 2.37275123620725e-08, 2.07933470787793e-08, -17840.3555555784, 0, 0, -9102.22222223526, -9102.22222222746, 1.22375778772928e-08, 0, 0, 0, -1.70421066649277e-08, -2.68006100460605e-08, 3.74135737273041e-08, 0, -2.11358213290298e-08, 0, -3.46828564931116e-08, 0, -1.48814387460381e-08, 2.94841551334197e-08, -4.33557811666552e-08, 0, -7.56614707446825e-08, -5.46568270311786e-08, -2.70483422511529e-08, -1.75653455184077e-08, -4.75450064283593e-08, -3.29410918346579e-08, -6.16524216814978e-08, 12066.9460317607, 0, 0, 8738.13333334212, 8738.13333333611, 0, 0, 0, 0, 2.95619545405351e-08, 1.6198196368648e-08, -3.01340342475248e-08, 0, 1.15512124113923e-08, 0, 3.21778702538752e-08, 0, 0, 0, 4.4482120072864e-08, -5.56593071354722e-08, 0, 4.93836896298764e-08, 0, 6.72402571170139e-08, 6.5592124719749e-08, 2.92048713407559e-08, 2.41278853493943e-08, 4.55576635570727e-08, 1.99200487348744e-08, 6.55241028102226e-08, 6.93185835912247e-08, 2.62703747644332e-08, 6.11640980744662e-08, 8.19259695056763e-08, 7.09451095734414e-08, -3328.81269841652, 0, 0, -3328.81269841495, -3328.81269841317, 0, 0, 0, 0, -1.27080541540018e-08, 0, 1.00503757977974e-08, 0, 0, 0, 0, 0, 0, 0, -1.67860629305335e-08, 3.69697594276458e-08, 0, -1.863148921556e-08, 0, -2.10261806940033e-08, -2.85735472799217e-08, -1.24198342060294e-08, 0, 0, 0, 2.03107584166376e-08, -2.76594162108251e-08, 0, -3.60654279822196e-08, -3.4564523927089e-08, -3.45426481451309e-08, 0, -2.42625621305317e-08, 0, -5.56916208909781e-08, 0, -1.14240238473966e-08, -9.08800049287309e-08, -3.67612237590229e-08, -6.18404316785145e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 9.1428571427958, 0, 0, 0, 0, 0, -179.19999999973, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1319.82222222241, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4778.66666666963, 0, 1.01836617614967e-08, 0, 0, 0, 1.04221996325434e-08, 0, 2.30164841844877e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9102.22222222829, 0, -1.44135600268046e-08, 0, 0, 0, 0, 0, -4.81340342444446e-08, 0, 0, 0, 0, 0, 0, -4.99518506204947e-08, 0, 0, 1.57052481419773e-08, 0, -1.40715318136687e-08, 0, 0, 1.26863434335986e-08, 0, 0, 0, 0, -8738.13333333803, 0, 0, 0, 0, 0, -1.55839646667898e-08, 0, 3.72452175229817e-08, 0, 0, 0, 0, 0, 0, 2.95259578529248e-08, 0, 7.29977662468239e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.68160004790782e-08, 0, 0, 0, 0, 0, 3328.81269841398, 0, 0, 0, 0, 0, 0, 0, -1.19543158601673e-08, 0, 0, 0, 0, 0, 0, 0, 0, -4.26053343922381e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.93397664869521e-08, 0, 0, 0, 0, 0, -1.54692191341481e-08, 0, 0, 0, -3.66886872985797e-08, 0, 3.46664474374893e-08, -1.60927183856765e-08, 1.84586092947788e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 5.33333333307764, 0, 0, 0, 0, 0, -97.4222222217554, 0, -42.6666666611615, 0, 0, 0, 0, 0, -2.41642331439915e-08, 0, 639.999999999275, 0, -3.13604624569625e-08, 0, 0, 0, 779.377777764406, 0, 0, 0, 0, 0, 0, 5.03317167823222e-08, 0, -1934.22222222233, 0, 1.12493557077746e-07, -1.11267461390528e-08, 0, 0, -5119.99999999013, 0, 6.2902559925128e-08, 0, 0, 0, 1.45163213159277e-08, 0, 1.050220744406e-08, 2.73959821407193e-08, 1.22583625118719e-08, 3.29484985135839e-08, 0, -5.66638729896392e-08, 0, 2730.66666666895, 0, -1.68980312957165e-07, 0, 0, 0, 15473.7777777622, 0, -1.32106439430131e-07, 1.49510423400027e-08, 0, 0, -2.85139692486461e-08, 1.73765682157267e-08, -3.15662336680911e-08, -5.82559340731932e-08, 0, 1.14196701350277e-08, -1.19435334408826e-07, -1.90617603293829e-08, 0, 0, -2.99469049101566e-08, -5.18823630999565e-08, -3.75377547789086e-08, 0, 3.44091994311467e-08, 0, -1456.35555555867, 0, 1.41531516387056e-07, 0, 0, 0, -21845.3333333041, 0, 1.63318486559698e-07, 1.62250235320057e-08, 0, 1.24614604417936e-08, 3.50780634634461e-08, 0, -1.30953539630453e-08, 5.5996582822769e-08, 0, 7.21480506292846e-08, -1.3399575440795e-08, 0, 0, 1.07452332381806e-07, 8.80537720873316e-08, 0, 5.81368935117984e-08, 2.65023069144474e-08, 1.40963962721806e-08, 8.45798805465716e-08, 2.21695657119524e-08, 2.78634048330611e-08, 0, 8.87568476864233e-08, 8.65930969025622e-08, 0, 0, 0, 0, 0, -4.20198835602604e-08, 0, 0, 0, 11650.8444444352, 0, -6.21684517345159e-08, 0, 0, 0, -1.37668273913824e-08, 0, 0, -1.33815099596809e-08, 0, 0, -1.09436124768568e-08, 0, 0, -2.81069411772559e-08, -2.81872190340642e-08, 0, 0, 1.89003751888941e-08, 2.66780569856506e-08, -2.58057089359685e-08, 0, 0, -5.47891191124343e-08, -3.32097238609523e-08, -5.89019958939727e-08, 5.61907884542461e-08, -1.55935290136621e-08, 0, -8.6452316099829e-08, 7.90191986033371e-08, 0, -2.121133840722e-07, -3.2155211370791e-08, -9.3979113818297e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 4.26666666655675, 0, 0, 0, 0, 0, -71.1111111107009, 0, -51.1999999993873, 0, 0, 0, 0, 0, 0, 0, 398.22222222166, 0, 136.533333323381, 0, 0, 0, 853.333333337484, 0, 0, 0, -1.2797209212813e-08, 0, 0, 1.34853132420351e-08, 0, -910.222222222994, 0, 0, 2.40519246873389e-08, 0, 0, -4778.66666663757, 0, -2275.55555557732, -1.05007239128045e-08, 0, 0, 1.50609952775097e-08, -1.15718316440242e-08, 0, 0, 3.48412898577627e-08, -5.16675236430729e-08, 0, -1.55525332451929e-08, 0, 728.177777783229, 0, 0, 2.16657014757053e-08, 0, 0, 10922.6666666146, 0, 3.67761568336741e-08, -2.33607303269217e-08, 0, 0, -1.85348451920704e-08, -1.82721285854951e-08, 7.38657728286701e-08, 12743.1111111114, 0, -1.87182701678144e-08, 7.87316979587094e-08, -1.30656796475821e-07, 0, 2.38444443323461e-08, -5.62051749849305e-08, 8.63235864619323e-08, -1.05277964022114e-07, 0, 0, 0, 0, 0, -1.37005119463e-08, 0, 0, 0, -8738.13333332359, 0, -1.03435029995306e-07, -3.01105800008499e-08, 0, -2.78621132670254e-08, 0, -1.06029781241447e-08, 2.78622381743606e-08, -29127.1111111046, 0, -3.42934835343852e-08, 1.30773374117771e-08, 0, 0, -1.37999392634035e-08, 2.40773889193477e-08, 0, -1.90290979377296e-07, 2.09854151914507e-08, -2.22751932974847e-08, -1.98473536043818e-08, 1.64703878000355e-07, -8.22779668901787e-08, 3.19754559833658e-07, -3.71201612986055e-08, 4.08339106492312e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.35790035144171e-08, 0, 0, 0, 0, 0, 1.39687277495088e-08, 23301.6888888538, 0, -1.62863021479155e-08, 1.95589279470408e-08, 0, 0, -2.59525174266615e-08, -2.70107784805127e-08, 3.3930293506882e-08, 3.0086088790104e-08, -5.42136814759205e-08, -6.03893212202542e-08, 0, 0, -1.48260772430872e-08, 0, -7.26121962263543e-08, 7.47307091285504e-08, -8.6816431426556e-08, -3.34874299978182e-08, 0, -5.95065822922008e-08, -1.53318054265502e-07, -8.72067054137845e-08, 2.14102387956068e-07, -1.53338302187579e-07, 9.50352691356582e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 4.00000000020935, 0, 0, 0, 0, 0, -58.6666666672845, 0, -58.6666666689624, 0, 0, 0, 0, 0, 2.46772679056538e-08, 0, 256.000000001877, 0, 256.000000027584, 0, 0, 0, 860.444444432132, 0, 0, 0, 1.59321366053331e-08, 0, 0, -4.97267104909036e-08, 0, -341.333333342458, 0, -341.333333384528, -3.0074335074948e-08, 0, 0, -3754.66666664585, 0, -3754.6666666622, 1.42755855115357e-08, 0, 0, -1.62439249333999e-08, 0, 1.35460717353318e-08, -2.22769673353369e-08, -1.4943830483919e-08, 5.20501667118266e-08, 0, 6.06473621175441e-08, 0, 2.0334235959841e-08, 0, 1.12076276055444e-07, -3.51353097205911e-08, 0, 0, 5006.2222221617, 0, 5006.22222231892, 1.83779621629565e-08, 0, -1.42213688508932e-08, 2.23617845780212e-08, 3.62201811297523e-08, -1.22751192029706e-07, 16383.9999999239, 0, 1.69048350532024e-08, -8.33386871058834e-08, 6.16170498167794e-08, -2.12645715698652e-08, 1.93379933300263e-08, 0, -1.01987251195787e-07, 5.24032474266084e-08, 0, -3.80070445020409e-08, 0, -1.93924905804669e-08, 0, -6.04584135596077e-08, 0, 0, 0, 9.16029948774156e-08, 0, -3.98088317104264e-08, 3.29244654235197e-08, 0, 3.79883554519207e-08, 0, 1.23068677803241e-08, -3.06413141956496e-08, -21845.3333331427, 0, -21845.3333333516, 0, 0, 2.38732301507713e-08, 0, 7.53269440943835e-08, -4.12416948326846e-08, 2.08207259038085e-07, 5.62677283619603e-08, 1.465732171483e-08, -3.29778172071189e-08, -1.13571990378494e-08, 4.7147797501778e-08, -1.27104931922759e-07, 9.91337958439256e-08, 1.21301964768651e-07, 0, 0, 0, 0, 0, 2.10329008868361e-08, 0, 0, 0, -2.20642375205436e-08, 0, 3.05952110506384e-08, 0, 0, 0, 0, 0, -1.62147840138466e-08, -5.10868788933786e-08, 0, 9.57459963134859e-08, -1.91344233994986e-08, 0, 0, 3.15139384914742e-08, 0, -2.83393955667461e-08, -1.43595831673818e-08, 4.04460081260213e-08, 8.6568443219683e-08, 29127.1111110773, 0, 2.15531414427904e-08, 3.85209924430418e-08, -1.05882339288588e-08, -5.57485569493266e-08, 1.4729372272607e-07, -7.56686085961996e-08, 6.43597912418974e-08, -2.00957169360468e-08, 1.71808902455883e-07, 2.13809560938648e-08, -3.41746470237259e-07, -1.70876684639313e-08, -2.01355100214414e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 4.26666666637722, 0, 0, 0, 0, 0, -51.1999999983083, 0, -71.1111111065323, 0, 0, 0, 0, 0, 1.45646980428064e-08, 0, 136.533333330512, 0, 398.222222190263, 0, 0, 0, 853.333333311515, 0, 0, 0, -1.28083604510007e-08, 0, 0, -3.28596986806139e-08, 0, 0, 0, -910.222222162675, 2.27758671363121e-08, 0, 0, -2275.55555546123, 0, -4778.66666654946, -1.38772943339557e-08, 0, 0, 0, 0, 0, 3.08077459875376e-08, 3.23755271819846e-08, -2.88144693104059e-08, 0, 3.77800220786997e-08, 0, 0, 0, 728.177777667822, 3.47312707976822e-08, 0, 0, -1.25679622975008e-07, 0, 10922.6666663826, 0, 0, 1.57345760956937e-08, 0, -5.24346461352923e-08, 1.17692829616672e-07, 12743.1111108535, 0, 0, 9.10038452809776e-08, -7.80180576072803e-08, 6.22504655602315e-08, -4.98223056313186e-08, -6.43076057224102e-08, 1.06584947966968e-07, -9.01398131743469e-08, 0, -2.26248433149946e-08, 0, -1.13123063647267e-08, 0, 5.65284560584562e-08, 0, 0, 0, 5.21999351966599e-08, 0, -8738.1333331435, -2.43379823671766e-08, 0, -3.31079383477825e-08, 0, 1.55490166242195e-08, 2.22569323398108e-08, 1.98388710156586e-07, 0, -29127.1111107674, -1.7263234585634e-08, 0, -3.88460098728733e-08, -3.89334287204677e-08, -4.81249344407432e-08, 1.71394473462419e-08, -1.39961867304134e-07, 0, -4.19046315849208e-08, -3.86207683423526e-08, 8.19120125358712e-08, -7.0726657144152e-08, 5.97379377906268e-08, 0, -5.66755863553647e-08, 0, 0, 0, 0, 0, -1.7973541641281e-08, 0, 0, 0, -1.63041334682848e-08, 0, -8.07398789727129e-08, 0, 0, 0, 0, -1.65611668968443e-08, 1.18325018999568e-08, -1.08473395640403e-07, 0, 23301.6888886913, 1.0702228183063e-08, 0, 0, 0, 0, 2.80588684224652e-08, 0, -4.13838542193819e-08, -6.08989018546839e-08, -1.5541389384381e-07, 0, -1.86038665492243e-08, 0, -2.90332043362845e-08, 9.85916445850851e-08, -9.64463640007109e-08, 4.19980128234486e-08, -4.16759861290061e-08, 1.21721170456733e-08, -1.81527616358732e-07, -1.40401686383582e-08, 2.56044930609588e-07, -2.17159851779137e-08, 1.56174795710287e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 5.333333333062, 0, 0, 0, 0, 0, -42.6666666659239, 0, -97.4222222165852, 0, 0, 0, 0, 0, 0, 0, 0, 0, 639.999999966251, 0, 0, 0, 779.377777749995, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1934.22222211173, -1.42055029007773e-08, 0, 0, 1.86634027179617e-08, 0, -5119.99999984614, 0, 0, 0, 0, 0, 0, -2.58985834972184e-08, -1.7015672211161e-08, 1.1700092320258e-08, 0, 0, 0, 1.51651271515574e-08, 0, 2730.66666652041, -1.43938597364432e-08, 0, 0, 2.01093864961899e-08, 0, 15473.7777774539, 0, 0, 0, 0, 0, -6.64998578771053e-08, -2.02083304912112e-07, 0, 0, -3.05071316624345e-08, 2.96022493306699e-08, -4.8311752806262e-08, 1.74206151612331e-08, 4.02566426549628e-08, -6.1746235848987e-08, 4.28435897239077e-08, 0, 0, 0, -1.04734179748679e-08, 0, -1456.35555543937, 0, 0, 0, -2.72636360060875e-08, 0, -21845.3333329667, 0, 0, 1.80846798436892e-08, 0, 1.98177535947998e-08, -1.3360384084182e-08, 1.25244369673807e-07, 0, 3.39335332716861e-07, 1.24079368013363e-08, 0, 2.98736397303273e-08, 0, 4.12232488831827e-08, 1.51349619475165e-08, 1.07734169931531e-07, 0, 2.29026987109634e-08, 0, -2.30258964736321e-08, 4.99399098511209e-08, 2.62912004803749e-08, 2.62759012080827e-08, 4.49234557836202e-08, 0, 0, 0, 0, 0, -3.12508034046704e-08, 0, 0, 0, 1.66085885180996e-08, 0, 11650.8444443207, 0, 0, 0, 0, 0, 0, 0, 0, -1.53497230983175e-07, 0, 0, 0, 0, 0, -3.23491182475014e-08, -1.87794286625432e-08, 2.00727967828753e-08, 3.24787381913548e-08, -1.07008950891384e-07, 0, 0, 1.72089206774905e-08, 0, -4.94922474095476e-08, 2.5576872676067e-08, -3.89900727733519e-08, 2.92797163215972e-08, -1.93996780452897e-08, 6.34942195452099e-08, 0, -1.35068393105623e-07, 0, -1.11920750809817e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 9.14285714282524, 0, 0, 0, 0, 0, 0, 0, -179.199999999896, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1319.82222222377, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.37377958103073e-08, 0, 0, 0, -4778.66666668101, 0, 0, 0, 2.78878480357077e-08, 0, 0, 0, 0, 0, 0, 0, 0, 1.02718309850533e-08, 0, 0, 0, -1.56598827632478e-08, 0, 0, 0, 9102.22222224506, 0, 0, 0, -3.50930768264081e-08, 0, 0, 0, 0, 0, 0, 0, 2.16001828239019e-08, -7.87017885198468e-08, 0, 0, 1.01049130091687e-08, -1.23949383522962e-08, 1.3326703665887e-08, -2.06144559168007e-08, -1.85280848934136e-08, 1.7702541529652e-08, -1.68004330430441e-08, 0, 0, 0, 0, 0, -8738.13333335612, 0, 0, 0, 0, 0, -4.40851597745942e-08, 0, 0, 0, 0, 0, 0, 8.34079450223034e-08, 0, 7.53954853849292e-08, 0, 0, -1.13490237801335e-08, 0, 0, 0, -1.24508199671527e-08, 0, 0, 0, 1.52340122459915e-08, 0, -1.85149601611619e-08, 0, -1.34878896467697e-08, 0, 0, 0, 0, 0, 3328.81269841904, 0, 0, 0, 0, 0, 2.01810027984843e-08, 0, 0, 0, 0, 0, 0, -3.4584238391404e-08, 0, -2.38920222681936e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.74265741082824e-08, 0, 0, 0, 0, 1.62351791929739e-08, -1.04279231278842e-08, 0, -1.07987910390381e-08, 0, -2.65463916685165e-08, 0, 3.39799367346336e-08, 0, 3.51324166421387e-08}, 
                                                    {0, 0, 9.14285714284279, 0, 0, -188.34285714261, 0, -9.14285714282128, -9.14285714284471, 0, 0, 1499.02222222055, 0, 0, 0, 179.199999999672, 179.200000000026, 0, 0, 0, 0, -6098.48888888316, 0, 0, 0, -1319.82222222195, -1319.8222222222, 0, 0, 0, 0, 0, 0, 0, 0, 0, 13880.8888888778, 0, 0, 0, 4778.66666666745, 4778.66666666571, 0, 0, -1.24902382654726e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.32654176497592e-08, 0, -17840.3555555432, 0, 0, 0, -9102.22222222055, -9102.22222222095, 0, 0, 1.36961717923189e-08, 0, -3.46125046428778e-08, 0, 0, 0, 0, 0, 0, 3.54616911615902e-08, 0, -1.21710981004055e-08, 0, 0, 0, 1.49194838120808e-08, 0, 2.3985617618994e-08, 2.30056039165324e-08, 0, 12066.9460317386, 0, 0, 0, 8738.13333332697, 8738.13333332993, 0, 0, 0, 0, 5.01961683324022e-08, 0, 0, 0, 0, -1.01209690663085e-08, 0, -5.56818918031358e-08, 0, -1.02761288968174e-08, 1.92918181924734e-08, 0, 3.03598797557716e-08, 0, 0, -1.92302638051207e-08, -3.60321208774065e-08, 0, -1.73956017642087e-08, 0, 0, -5.62745575820026e-08, 1.73170580774083e-08, -6.45639608145197e-08, -3.27484086113835e-08, 0, -3328.81269841087, 0, 0, 0, -3328.8126984088, -3328.81269841164, 0, 0, 0, 0, -2.01000551282763e-08, 0, 0, 0, 0, 0, 0, 2.05409912332781e-08, 0, -1.14461756156677e-08, 0, 0, 0, 0, 0, 0, 1.1941515393473e-08, 0, 0, 2.9720820719716e-08, 0, 0, 0, 0, 0, 4.3271883758191e-08, 0, 0, 0, 4.05327699389582e-08, 0, 0, 0, 0}, 
                                                    {0, 0, -37.3333333332148, 0, 0, 761.955555553476, 0, 79.9999999995193, 79.9999999999365, 0, 0, -5983.99999998594, 0, -42.6666666636907, 0, -1546.66666665997, -1546.66666666714, 0, -42.6666666656501, -85.333333332973, 0, 23918.9333332849, 0, 0, 0, 11158.7555555276, 11158.7555555558, 0, 0, 779.377777732225, 0, 779.377777782917, 0, 1558.75555554656, 0, 0, -53247.9999999055, 0, -1.11989029293614e-08, 0, -39253.333333277, -39253.333333327, 0, 2.58396555324814e-08, 1.49308020222745e-07, 0, -5119.99999984191, -5120.00000002602, 0, -2.36004314008109e-08, 0, -10239.9999999747, 4.1677723687439e-08, 0, 3.31046149509828e-08, 8.03826987989591e-08, 0, 66628.2666665614, 0, 4.18158571917875e-08, 0, 71907.5555554826, 71907.5555555443, 0, -4.47326393023425e-08, -1.35105121505796e-07, 0, 15473.7777776243, 15473.7777778093, 0, 2.29454456448829e-08, 1.85144761997993e-08, 30947.5555555243, -7.17629584532373e-08, -5.72035032008016e-07, 0, 1.05861160941103e-07, 2.8038341068344e-08, 0, -1.20725995094094e-08, -2.26374333924601e-07, -6.39167796884473e-08, -2.14527404609926e-07, -1.37720292354217e-07, 0, -43690.666666604, 0, -3.8055244930804e-08, 0, -65535.9999999353, -65535.9999999872, 0, 3.65771212762068e-08, -1.16081085093158e-08, 0, -21845.3333333272, -21845.3333333146, 0, 0, -1.37249990507282e-08, -43690.6666666609, 5.02751176735252e-08, 5.92564555152057e-07, 0, 5.97465661007227e-07, -1.15624361751114e-07, 0, -1.45619575736675e-07, -1.16624466743294e-07, 5.3522681169447e-08, 2.96737738832664e-08, 3.08180686444954e-07, 9.54414703763284e-08, 2.50584469595499e-07, -4.05522857888185e-08, 3.13789235301578e-08, 4.4231304682213e-07, -3.10406649956323e-08, 4.79158585631497e-07, 3.35431268410928e-07, 0, 11650.8444444291, 0, 1.11508176411574e-08, 0, 23301.6888888635, 23301.6888888848, 0, -1.00845886611174e-08, 3.99517827852961e-08, 0, 11650.8444444752, 11650.8444444361, 0, 0, 0, 23301.6888888949, -2.05507501526307e-08, -1.73068362606503e-07, 0, -1.75981300425382e-07, 3.41751632045007e-08, 0, 3.11518002752827e-08, 3.66395238850472e-08, -1.6115995139477e-08, 0, -1.0201840616616e-07, -3.65134450392165e-08, -9.64377171612057e-08, -3.97388635295216e-07, 0, 5.92096947852297e-08, 1.76953968045435e-08, -4.58198392722654e-08, 0, -2.93983016708161e-07, -2.66009629813383e-08, -2.19847175150074e-07, 4.63011867155971e-08, -2.45733417862846e-07, -1.37928061081079e-07, -2.33099879334773e-08, 0, -1.98911416078729e-07}, 
                                                    {0, 0, 89.5999999996198, 0, 0, -1804.79999999345, 0, -311.466666664004, -311.466666665941, 0, 0, 13912.1777777344, 0, 358.399999985988, 0, 5907.91111106741, 5907.911111105, 0, 358.399999991327, 716.799999993588, 0, -54294.7555554086, 0, -136.533333314885, 0, -41426.4888886588, -41426.4888888522, 0, -136.533333286204, -6382.93333308006, 0, -6382.93333327596, -409.599999993241, -12765.8666665882, -409.599999915679, 0, 117464.177777496, 0, 5.79385400218629e-08, 0, 140174.222221647, 140174.222222123, 0, -1.30840647784255e-07, 2275.55555491712, 0, 40277.3333322023, 40277.3333330809, 0, 2275.55555540176, 2.99215448261462e-08, 80554.6666664009, -3.18884665469533e-07, 6826.6666665751, -2.18788423851818e-07, 6826.66666597123, 0, -142358.755555246, 0, -1.7033005720596e-07, 0, -244303.644443658, -244303.644444327, 0, 1.93618629846571e-07, 4.9403831461062e-07, 0, -114687.999998083, -114687.999999492, 0, 2.69906343022664e-07, -1.24606824265077e-07, -229375.999999667, 5.72537972740558e-07, -12743.1111083576, 0, -12743.1111107109, -7.23660223602045e-08, 1.75136347654677e-07, -38229.3333329571, -38229.333331118, 4.95322283748662e-07, 1.83121402289668e-06, 1.22655456762153e-06, 0, 90294.0444442626, 0, 1.46442133507868e-07, 0, 209715.199999416, 209715.199999928, 0, -1.44808032447602e-07, 8.08118577938914e-08, 0, 148548.266665246, 148548.266666143, 0, -3.11672811934263e-07, 1.1782153404688e-07, 297096.533333188, -4.61872124292737e-07, -2.42299233952287e-06, 0, 29127.1111076525, 29127.1111104666, 0, -3.13413378440561e-07, 4.23060047385317e-07, -1.83826995140431e-07, 87381.3333330081, 87381.3333306799, -7.09572163076112e-07, -2.03700383978256e-06, 0, -3.97425550423648e-07, -3.64877003755552e-06, -4.58788586744261e-07, -3.19410310239382e-06, -2.36935072754924e-06, 0, -23301.6888888448, 0, -4.16618846413709e-08, 0, -69905.0666664862, -69905.0666666508, 0, 3.9646248156482e-08, -1.5802442925112e-07, 0, -69905.0666662679, -69905.0666664859, 0, 1.2041915067224e-07, -4.02477752491505e-08, -139810.133333344, 1.61413207847286e-07, 6.33821287833589e-07, 0, -23301.6888875343, -23301.6888885523, 0, 1.33716858584975e-07, -1.88868509873102e-07, 5.4785683091301e-08, -69905.066666597, -69905.0666656989, 2.91938699940046e-07, 7.95354898708737e-07, 1.78501213294364e-06, 0, 2.63644971817849e-07, -6.85476415252664e-08, 2.33726426836593e-07, 1.18373275459744e-07, 2.28676746231593e-06, 2.59549454531731e-07, 1.87272153968868e-06, -4.24634832027542e-08, 1.72921448244542e-06, 1.18358949914973e-06, 7.81364600746203e-07, 4.44267473115938e-07, 2.23804951321368e-06}, 
                                                    {0, 0, -139.999999999431, 0, 0, 2763.99999999086, 0, 710.6666666594, 710.666666663313, 0, 0, -20721.7777777214, 0, -1338.66666663219, 0, -13100.4444443167, -13100.4444444007, 0, -1338.666666631, -2677.33333330447, 0, 78172.4444442671, 0, 1109.33333330518, 0, 88078.2222215036, 88078.222221992, 0, 1109.33333319116, 22961.7777770815, 0, 22961.7777773673, 3327.99999995279, 45923.5555551959, 3327.99999971348, 0, -162929.777777461, 0, -341.333333508993, 0, -282168.888886991, -282168.888888348, 0, -341.333333018233, -17635.5555539934, 0, -136533.333329973, -136533.333331463, 0, -17635.5555544523, -1365.33333337829, -273066.66666532, -1365.33333244623, -52906.6666660039, -2047.99999942278, -52906.6666637936, 0, 190236.444444117, 0, 4.39432715451074e-07, 0, 461482.666664004, 461482.666666074, 0, -4.05325016072358e-07, 5006.22222104068, 0, 357262.222215816, 357262.222218589, 0, 5006.22222074067, 2.53720234687059e-07, 714524.44444248, -1.42030621435145e-06, 91022.2222152494, 0, 91022.222218418, 20024.8888885438, -4.29280198049413e-07, 273066.666664538, 273066.666657034, -9.5496786476879e-07, 20024.8888823376, 30037.3333285123, 0, -116508.444444263, 0, -3.68624978273661e-07, 0, -371370.666664732, -371370.666666359, 0, 2.77951769905094e-07, -1.02585541442919e-07, 0, -415061.333327848, -415061.333330149, 0, 1.15388324723155e-06, -2.55743700319703e-07, -830122.666665482, 1.12718086601174e-06, -21845.3333276387, 0, -182044.444435088, -182044.444439042, 0, -21845.333330171, -4.60880736739764e-07, 3.08855016407732e-07, -546133.333331291, -546133.333321545, 1.10932972323698e-06, 6.28043542965291e-06, -87381.3333324368, 4.3546903174823e-07, -87381.3333186658, 2.54623561565921e-06, -131071.999987941, 7.2344713714456e-06, 0, 29127.111111069, 0, 1.06110870606186e-07, 0, 116508.444443874, 116508.44444439, 0, -7.3367514233181e-08, 3.10411475288337e-07, 0, 174762.666664881, 174762.666665658, 0, -3.69404961144598e-07, 8.86923107830402e-08, 349525.333333147, -3.91350221711037e-07, -1.45584002363307e-06, 0, 116508.444440438, 116508.444441918, 0, -9.73635000401047e-07, 2.7278281187946e-07, -1.05488904749897e-07, 349525.333332725, 349525.333328741, -4.22714831649464e-07, -2.23047302403491e-06, 29127.1111068022, 0, 29127.1111088091, -1.20813439927385e-07, -2.72814377318514e-07, 116508.444443761, 116508.444434847, -1.19433703791588e-07, -6.80749583623992e-06, -4.62553274677353e-07, 174762.666659575, -3.0815446374901e-06, -3.78356687435668e-06, -1.59099214571529e-06, -8.68789384826059e-06}, 
                                                    {0, 0, 149.333333333014, 0, 0, -2848.7111111076, 0, -1056.71111110029, -1056.71111110503, 0, 0, 20403.1999999902, 0, 2943.99999995169, 0, 18568.5333331362, 18568.5333332357, 0, 2943.99999994385, 5887.99999995143, 0, -73181.8666666773, 0, -4039.11111109452, 0, -116588.088887761, -116588.088888364, 0, -4039.11111097309, -47445.3333322636, 0, -47445.3333324351, -12117.3333332756, -94890.6666659484, -12117.3333329786, 0, 145408.000000094, 0, 2730.66666698609, 0, 344746.666663639, 344746.666665462, 0, 2730.66666652542, 59391.9999976795, 0, 255999.999994741, 255999.999995516, 0, 59391.9999979564, 10922.6666667501, 511999.999997111, 10922.666666121, 178175.999998733, 16383.9999997254, 178175.999994684, 0, -162747.733333505, 0, -728.177778508501, 0, -522467.555551279, -522467.555554284, 0, -728.17777770859, -36408.8888870097, 0, -591644.444434048, -591644.444435485, 0, -36408.8888870767, -3640.88888916463, -1183288.88888443, -3640.88888870234, -267605.333323129, 0, -267605.333324436, -145635.55555476, -7281.77777808086, -802815.999995957, -802815.999980065, -7281.77777920379, -145635.555545874, -218453.333325379, 0, 96119.4666667998, 0, 6.15565712118e-07, 0, 393215.999996917, 393215.999999398, 0, -2.61222328360826e-08, 8738.13333321156, 0, 611669.333324018, 611669.333325607, 0, 8738.1333326222, 2.27198345055987e-07, 1223338.66666385, 1.05677378995732e-07, 131071.9999918, 0, 436906.666652794, 436906.666653277, 0, 131071.999993615, 43690.6666667845, 1.42719005206497e-07, 1310719.99999608, 1310719.99997441, 2.72360423227981e-06, 43690.6666602642, 524287.999998475, 2.7388974547904e-06, 524287.999972086, 87381.3333290914, 786431.999977616, 87381.3333252941, 0, -23301.6888889276, 0, -1.83455915077017e-07, 0, -116508.444443553, -116508.444444356, 0, 0, -3.26847599267193e-07, 0, -233016.888885719, -233016.888886499, 0, 1.04150099981363e-07, -6.84777519789686e-08, -466033.777777261, 0, -23301.6888867138, 0, -233016.888882768, -233016.888882567, 0, -23301.688887459, -1.03145368512969e-07, 7.17943232156155e-08, -699050.666665532, -699050.666656296, -1.3637354271008e-06, 1.34433876558168e-06, -116508.444438359, 0, -116508.444439044, -116508.444444272, -8.80326495423668e-07, -466033.777776937, -466033.777757725, -2.43548686645528e-06, -116508.444433419, 1.15245536894312e-06, -699050.666652525, 1.15122903897221e-06, -233016.888881894, 7.96123295761696e-07, -233016.888873666}, 
                                                    {0, 0, -112.000000000053, 0, 0, 1987.20000000338, 0, 1091.19999999075, 1091.19999999593, 0, 0, -13051.0222222622, 0, -4321.422222184, 0, -17372.4444442718, -17372.4444443553, 0, -4321.42222219808, -8642.84444442374, 0, 43488.7111113045, 0, 8917.33333334311, 0, 95894.7555545679, 95894.7555550354, 0, 8917.33333342077, 61323.3777768458, 0, 61323.3777769872, 26752.0000000754, 122646.755554976, 26752.0000000417, 0, -81464.8888893581, 0, -10126.2222225692, 0, -254520.888886229, -254520.888887693, 0, -10126.2222227475, -111843.555553518, 0, -274773.333328797, -274773.333328623, 0, -111843.555554485, -40504.8888891724, -549546.666664014, -40504.8888903615, -335530.666666164, -60757.3333345485, -335530.666662687, 0, 87017.2444450508, 0, 6007.46666742304, 0, 354076.444440677, 354076.444443242, 0, 6007.46666753627, 111047.111109277, 0, 546133.333324321, 546133.333323348, 0, 111047.111111976, 30037.3333336262, 1092266.66666246, 30037.3333372925, 384113.777769201, 0, 384113.777769428, 444188.444444926, 60074.6666689388, 1152341.33333129, 1152341.33331386, 60074.6666737309, 444188.444441397, 666282.66666279, 0, -49516.0888892883, 0, -1456.3555562051, 0, -249036.799997291, -249036.799999507, 0, -1456.35555616502, -56797.8666662787, 0, -502442.666658456, -502442.666657951, 0, -56797.8666687853, -8738.13333341806, -1004885.3333308, -8738.1333370913, -262143.999992944, 0, -509724.444433019, -509724.444429968, 0, -262143.99999603, -283989.333334041, -21845.3333344173, -1529173.33333169, -1529173.33330613, -21845.3333439011, -283989.333337535, -1048576.00000145, -29127.1111196356, -1048575.99997661, -567978.666665781, -1572863.99998137, -567978.666670237, 0, 11650.8444445499, 0, 2.01426604956831e-07, 0, 69905.0666658902, 69905.0666666238, 0, 1.6130638407734e-07, 11650.8444446212, 0, 174762.666663812, 174762.666663994, 0, 11650.844445352, 0, 349525.333332953, 1.18425288707405e-06, 69905.0666646564, 0, 233016.888883856, 233016.888881716, 0, 69905.066666834, 69905.0666667846, 0, 699050.666666518, 699050.666655217, 4.82511245997781e-06, 69905.0666704508, 174762.666661699, 0, 174762.666661515, 349525.333334451, 2.79765168090597e-06, 699050.666667842, 699050.666646332, 6.64355626687883e-06, 349525.33332744, 174762.666666015, 1048575.99998671, 174762.666674287, 699050.666661911, 233016.888893225, 699050.666656245}, 
                                                    {0, 0, 64.0000000000667, 0, 0, -879.542857145358, 0, -879.542857138351, -879.54285714205, 0, 0, 4963.55555558258, 0, 4963.55555553971, 0, 9927.11111102557, 9927.11111107838, 0, 4963.55555556144, 9927.11111112914, 0, -14973.1555556819, 0, -14973.1555555744, 0, -44919.4666661785, -44919.4666664343, 0, -14973.1555556987, -44919.4666662322, 0, -44919.4666664429, -44919.4666668578, -89838.9333332233, -44919.4666670482, 0, 26168.8888891912, 0, 26168.8888890985, 0, 104675.555554234, 104675.555554978, 0, 26168.8888893851, 104675.555554588, 0, 157013.333331275, 157013.333331336, 0, 104675.555555914, 104675.555555957, 314026.66666569, 104675.555557917, 314026.666667409, 157013.333335469, 314026.66666642, 0, -26578.4888892774, 0, -26578.4888893329, 0, -132892.444442555, -132892.444443845, 0, -26578.4888895816, -132892.444443474, 0, -265784.88888485, -265784.888884079, 0, -132892.444446704, -132892.44444469, -531569.777775965, -132892.444449418, -265784.88888506, 0, -265784.88888649, -531569.777779754, -265784.88889221, -797354.666668012, -797354.666658935, -265784.888897713, -531569.777783058, -797354.666670165, 0, 14563.5555558108, 0, 14563.5555559449, 0, 87381.3333319647, 87381.3333330867, 0, 14563.5555559945, 87381.3333330208, 0, 218453.333329629, 218453.333328909, 0, 87381.3333359567, 87381.3333332828, 436906.666665536, 87381.3333376134, 218453.333330023, 0, 291271.111106264, 291271.111104988, 0, 218453.333334188, 436906.666668093, 218453.333334979, 873813.333334962, 873813.333319878, 218453.333345325, 436906.666678107, 873813.333337635, 291271.111121099, 873813.333328507, 873813.333338116, 1310719.99999625, 873813.333348281, 0, -3328.81269848003, 0, -3328.81269853749, 0, -23301.6888884949, -23301.6888888685, 0, -3328.81269851972, -23301.6888889327, 0, -69905.0666653558, -69905.0666652813, 0, -23301.6888898096, -23301.6888888246, -139810.133333182, -23301.6888901889, -69905.0666656367, 0, -116508.444442367, -116508.444441015, 0, -69905.0666680964, -139810.133333571, -69905.0666667584, -349525.333334215, -349525.333327167, -69905.0666718413, -139810.133339277, -116508.444442297, 0, -116508.444442984, -349525.33333543, -116508.444447753, -466033.777780539, -466033.777769117, -116508.444451587, -349525.333335971, -349525.333334319, -699050.666661648, -349525.333346031, -699050.666667598, -466033.777786544, -699050.666668135}, 
                                                    {0, 0, 0, 64.000000000039, 0, 0, -879.542857143586, 0, -879.542857139598, -879.542857142032, 0, 0, 4963.55555556099, 0, 4963.55555554903, 0, 4963.55555553259, 9927.1111110947, 9927.11111106594, 9927.1111111228, 0, 0, -14973.1555555764, 0, -14973.1555555474, 0, -14973.1555554268, -44919.4666665806, -44919.466666418, 0, -44919.466666515, -44919.4666665209, -44919.4666668937, -44919.4666666213, -89838.933333576, 0, 0, 26168.8888889344, 0, 26168.8888889162, 0, 26168.888888508, 104675.555555374, 104675.555554817, 0, 104675.555555168, 0, 104675.555555003, 157013.333332572, 157013.333333036, 104675.555556067, 104675.555554672, 314026.666667724, 157013.333334515, 314026.666669095, 314026.666667579, 0, 0, -26578.4888889442, 0, -26578.4888889591, 0, -26578.4888883453, -132892.444444264, -132892.444443282, 0, -132892.444444147, 0, -132892.444442986, -265784.88888746, -265784.888888306, -132892.444444754, -132892.444442101, -531569.777779147, 0, -265784.888887072, -265784.888888767, -265784.888891763, -531569.777782353, -265784.88889046, -531569.777775465, -797354.666674448, -797354.66667253, -797354.666675129, 0, 0, 14563.5555555911, 0, 14563.555555611, 0, 14563.555555183, 87381.3333332587, 87381.3333324419, 0, 87381.3333333391, 0, 87381.3333317766, 218453.33333221, 218453.333332447, 87381.333333241, 87381.3333310934, 436906.666667101, 0, 218453.333331859, 0, 218453.333332444, 291271.11110851, 291271.111111951, 218453.333335424, 436906.666669523, 218453.333333766, 436906.66665965, 873813.333341965, 873813.333341544, 291271.111115794, 873813.333344066, 873813.333336553, 873813.333344519, 873813.333338819, 1310720.00002056, 0, 0, -3328.81269842202, 0, -3328.812698427, 0, -3328.81269831334, -23301.6888888805, -23301.6888886272, 0, -23301.6888889453, 0, -23301.6888883386, -69905.066666356, -69905.0666661901, -23301.6888887865, -23301.688888161, -139810.133333209, 0, -69905.0666663311, 0, -69905.0666659127, -116508.444443307, -116508.444444871, -69905.0666670973, -139810.133333798, -69905.0666664984, -139810.133329421, -349525.333336354, -349525.333336495, 0, -116508.444443218, -116508.444444781, -116508.444446789, -349525.333337447, -116508.444446704, -349525.33333103, -466033.777784177, -466033.77778257, -349525.333336953, -349525.333332579, -699050.666679707, -466033.777783777, -699050.666679972, -699050.666675996}, 
                                                    {0, 0, 0, -112.000000000114, 0, 0, 1987.20000000212, 0, 1091.19999998855, 1091.19999999819, 0, 0, -13051.022222238, 0, -4321.4222222043, 0, -4321.42222213311, -17372.444444412, -17372.4444442775, -8642.8444444198, 0, 0, 43488.7111111716, 0, 8917.33333328874, 0, 8917.33333296619, 95894.7555553927, 95894.7555546323, 0, 61323.3777774483, 61323.3777768717, 26752.0000001896, 26751.9999996031, 122646.75555546, 0, 0, -81464.8888890202, 0, -10126.2222222008, 0, -10126.2222213262, -254520.888888558, -254520.88888637, 0, -111843.555554639, 0, -111843.55555316, -274773.333331783, -274773.333329635, -40504.8888893692, -40504.888886418, -549546.666665987, -60757.3333339784, -335530.666669213, -335530.666666051, 0, 0, 87017.2444446047, 0, 6007.4666667046, 0, 6007.46666544885, 354076.444444132, 354076.444440859, 0, 111047.111110198, 0, 111047.111107337, 546133.333330497, 546133.333325891, 30037.3333335447, 30037.333327564, 1092266.66666401, 0, 384113.777773996, 384113.777771624, 60074.6666690723, 444188.444449916, 60074.6666653447, 444188.444438487, 1152341.3333406, 1152341.333334, 666282.666676654, 0, 0, -49516.0888889915, 0, -1456.35555559886, 0, -1456.35555470266, -249036.799999889, -249036.799997472, 0, -56797.8666663634, 0, -56797.8666635233, -502442.666664459, -502442.666659452, -8738.13333312685, -8738.13332774471, -1004885.33332972, 0, -262143.999996749, 0, -262143.999994404, -509724.444439289, -509724.444437107, -21845.3333350315, -283989.333336683, -21845.3333287627, -283989.333321553, -1529173.33333989, -1529173.33333315, -29127.1111134383, -1048576.0000122, -1048575.99999735, -567978.666681438, -567978.666672766, -1572864.0000225, 0, 0, 11650.8444444713, 0, 1.09820473471025e-08, 0, -2.40206144579284e-07, 69905.0666666614, 69905.0666659693, 0, 11650.8444444497, 0, 11650.8444434285, 174762.666666057, 174762.666664019, -1.43547679083209e-07, -1.93607768526996e-06, 349525.333331744, 0, 69905.0666658158, 0, 69905.0666645263, 233016.88888668, 233016.888885496, 2.94721773128583e-07, 69905.0666671531, -2.81722239148202e-06, 69905.0666604255, 699050.666668226, 699050.666665375, 0, 174762.666664194, 174762.666663853, 1.58324101308314e-06, 349525.333338164, 1.6918576109732e-08, 349525.333328315, 699050.666673312, 699050.666667066, 174762.666671509, 174762.666664697, 1048576.00001236, 233016.888896903, 699050.666683728, 699050.666678087}, 
                                                    {0, 0, 0, 149.333333333526, 0, 0, -2848.71111111469, 0, -1056.71111109565, -1056.71111110903, 0, 0, 20403.2000000266, 0, 2943.99999997631, 0, 2943.99999986492, 18568.5333332991, 18568.5333331091, 5887.99999996631, 0, 0, -73181.8666667688, 0, -4039.11111103334, 0, -4039.11111057383, -116588.088888718, -116588.088887673, 0, -47445.333332965, -47445.3333318475, -12117.3333334484, -12117.3333328755, -94890.6666664384, 0, 0, 145408.000000221, 0, 2730.66666655363, 0, 2730.66666545666, 344746.666666316, 344746.666663462, 0, 59391.9999989176, 0, 59391.9999958593, 255999.999998335, 255999.999993803, 10922.6666669193, 10922.6666641438, 511999.999998609, 16384.0000001593, 178176.000001989, 178175.999998669, 0, 0, -162747.733333603, 0, -728.177777685071, 0, -728.177776233957, -522467.555555208, -522467.555551149, 0, -36408.8888876571, 0, -36408.8888827912, -591644.444441415, -591644.444432373, -3640.8888888523, -3640.8888831317, -1183288.88888482, 0, -267605.33332927, -267605.33332135, -7281.77777882841, -145635.555559844, -7281.77777531335, -145635.555549396, -802816.00000526, -802815.999996556, -218453.33334196, 0, 0, 96119.4666668397, 0, -4.75634379572878e-08, 0, -1.03953336173145e-06, 393215.999999868, 393215.999996986, 0, 8738.13333271411, 0, 8738.13332867605, 611669.333330932, 611669.333322235, -2.86592318295664e-07, -5.65899758185628e-06, 1223338.66666197, 0, 131071.999996354, 0, 131071.999989054, 436906.666661287, 436906.666651792, 4.08926216139454e-07, 43690.6666689995, -5.8855750204339e-06, 43690.6666561026, 1310720.00000408, 1310719.99999331, -3.09251500211869e-07, 524288.000009404, 524287.999993004, 87381.333345772, 87381.3333395584, 786432.000017463, 0, 0, -23301.6888889342, 0, 1.46480398483666e-08, 0, 2.85649407823666e-07, -116508.444444434, -116508.444443633, 0, 1.04956322246958e-07, 0, 1.41552206860527e-06, -233016.888888205, -233016.888885042, 1.4124624424857e-07, 2.00413601028824e-06, -466033.777775892, 0, -23301.6888878523, 0, -23301.6888850287, -233016.888886589, -233016.888882252, 7.51698732298444e-08, -2.5425410734879e-07, 3.45862078480581e-06, 5.43220205416822e-06, -699050.666667261, -699050.666662368, 0, -116508.44444185, -116508.444438107, -1.89245746381376e-07, -116508.444448025, 1.60638012884076e-06, -116508.444438666, -466033.777782654, -466033.777773708, -3.80714951775533e-06, 8.28765568551737e-07, -699050.666675062, -7.15180146308086e-06, -233016.88890315, -233016.888898839}, 
                                                    {0, 0, 0, -140.000000000208, 0, 0, 2764.00000000385, 0, 710.6666666565, 710.666666665388, 0, 0, -20721.7777778063, 0, -1338.66666664967, 0, -1338.66666657278, -13100.4444444243, -13100.4444442944, -2677.3333333205, 0, 0, 78172.444444554, 0, 1109.33333326771, 0, 1109.33333296893, 88078.2222221192, 88078.2222214143, 0, 22961.7777775392, 22961.7777766764, 3328.00000006687, 3327.99999982773, 45923.5555554306, 0, 0, -162929.777778015, 0, -341.333333209548, 0, -341.333332557034, -282168.888888672, -282168.88888679, 0, -17635.5555548182, 0, -17635.5555523759, -136533.333332273, -136533.333328691, -1365.33333342759, -1365.33333235039, -273066.666665798, -2048.00000025635, -52906.6666676421, -52906.6666660847, 0, 0, 190236.444444734, 0, -1.39008484886964e-07, 0, -9.31147147002588e-07, 461482.666666432, 461482.666663826, 0, 5006.22222132199, 0, 5006.22221764432, 357262.222220276, 357262.222213334, -9.99929500425704e-08, -2.3365838331091e-06, 714524.444441908, 0, 91022.2222196375, 91022.2222126018, 5.49244265635909e-07, 20024.8888907835, -5.71108242906548e-07, 20024.8888866138, 273066.666669025, 273066.666664028, 30037.3333386035, 0, 0, -116508.44444463, 0, 8.77674866607695e-08, 0, 6.02350031029884e-07, -371370.66666657, -371370.666664753, 0, 5.47072128736726e-07, 0, 3.33655371895118e-06, -415061.33333174, -415061.33332534, 1.89862976829807e-07, 2.43845664100446e-06, -830122.666663875, 0, -21845.3333309168, 0, -21845.3333243675, -182044.444441071, -182044.444432418, 0, -7.31056447828275e-07, 2.16098401517208e-06, 3.94473521141413e-06, -546133.333334933, -546133.333328008, -4.43313688954316e-08, -87381.3333373394, -87381.333329036, -6.54276848817666e-06, -5.72122049324136e-06, -131072.000008855, 0, 0, 29127.1111111596, 0, -2.61452042694678e-08, 0, -1.59963327941292e-07, 116508.444444433, 116508.444443935, 0, -1.25246051259267e-07, 0, -9.68453347669254e-07, 174762.666666194, 174762.66666396, -8.26813679220304e-08, -8.9398996298969e-07, 349525.333332236, 0, -7.24057815665723e-07, 0, -3.12248865505631e-06, 116508.444442985, 116508.444439183, -1.16130421215279e-07, 0, -1.47281192329642e-06, -2.02827704241967e-06, 349525.333333411, 349525.333330151, 0, 29127.1111094788, 29127.1111057414, -1.78116312505211e-08, 1.33562428406375e-06, -7.13903420646981e-07, -2.87369095772758e-06, 116508.444446385, 116508.444441035, 1.69435670467103e-06, 1.5844809391136e-06, 174762.666670466, 4.53128396814898e-06, 6.95335975335738e-06, 6.17253548105622e-06}, 
                                                    {0, 0, 0, 89.6000000001458, 0, 0, -1804.80000000269, 0, -311.466666662996, -311.466666666313, 0, 0, 13912.1777777977, 0, 358.399999993619, 0, 358.399999964599, 5907.91111110504, 5907.91111105538, 716.799999999509, 0, 0, -54294.755555632, 0, -136.533333305305, 0, -136.533333201228, -41426.4888888548, -41426.488888588, 0, -6382.93333324353, -6382.93333288914, -409.600000033645, -409.599999987888, -12765.8666666332, 0, 0, 117464.177777943, 0, -5.90336628252689e-08, 0, -2.62782403135537e-07, 140174.222222149, 140174.222221443, 0, 2275.55555525998, 0, 2275.55555424411, 40277.3333329259, 40277.3333314234, 5.14887600041817e-08, -1.14855136115657e-07, 80554.66666638, 2.07322347191274e-07, 6826.66666695447, 6826.66666656714, 0, 0, -142358.755555757, 0, 7.25703767469111e-08, 0, 2.85123330116303e-07, -244303.644444353, -244303.644443403, 0, 3.63110401608499e-07, 0, 1.82350175723011e-06, -114687.999999236, -114687.999996387, 1.64055386358553e-08, 3.19112592894721e-07, -229375.999999101, 0, -12743.1111101113, -12743.1111069685, -3.56408711173403e-07, -5.10892053771405e-07, -3.09769820528673e-07, 2.75182310879402e-07, -38229.3333339451, -38229.3333323011, -1.92303935842683e-06, 0, 0, 90294.0444445738, 0, -4.73544581769531e-08, 0, -1.75411625874599e-07, 209715.199999961, 209715.199999309, 0, -2.45815582460509e-07, 0, -1.23643940756625e-06, 148548.266666014, 148548.266663454, -4.25525020439397e-08, -4.38635847520638e-07, 297096.533332393, 0, -9.82818431937493e-07, 0, -3.88580655615116e-06, 29127.1111098009, 29127.1111058518, 9.20919160045623e-08, 9.45629044647861e-08, -3.66898445732013e-08, -5.3406116623567e-07, 87381.3333336386, 87381.3333311285, 4.42919806307218e-07, 9.34953720473582e-07, -1.46324699433487e-06, 2.15306285801764e-06, 2.66457735840424e-06, 2.52107785495476e-06, 0, 0, -23301.6888889227, 0, 1.43583425404986e-08, 0, 4.42643728768179e-08, -69905.0666666609, -69905.066666484, 0, 6.08911797416672e-08, 0, 3.3561377748658e-07, -69905.0666664652, -69905.0666655937, 2.21059509341642e-08, 1.74860855309428e-07, -139810.133332966, 0, 3.00913356041831e-07, 0, 1.32946007002162e-06, -23301.6888883062, -23301.688886624, 0, 2.53227890097863e-08, 2.02348745202076e-07, 2.39809406915856e-07, -69905.0666665935, -69905.0666654042, 0, 6.42368232285174e-07, 2.40255732575361e-06, -1.23167629523876e-07, -2.40895142771113e-07, -1.75542195747036e-08, 7.96250329819779e-07, -3.52348205670392e-07, 1.45884515956215e-06, -4.61483619645621e-07, -1.33127400076607e-06, -9.22887981620388e-07, -1.70356180936051e-06, -1.88633275327487e-06, -2.10463294381116e-06}, 
                                                    {0, 0, 0, -37.333333333398, 0, 0, 761.955555556748, 0, 79.9999999993369, 80.0000000000271, 0, 0, -5984.0000000088, 0, -42.6666666658417, 0, -42.6666666599069, -1546.66666666646, -1546.66666665616, -85.3333333341922, 0, 0, 23918.9333333671, 0, 0, 0, -2.33410361011775e-08, 11158.7555555522, 11158.7555554976, 0, 779.377777761996, 779.377777686581, 1.04535367873309e-08, 0, 1558.75555555301, 0, 0, -53248.0000000731, 0, 1.1619683327626e-08, 0, 4.24124687967224e-08, -39253.333333326, -39253.3333331794, 0, 6.04194603688236e-08, 0, 2.7291939822333e-07, -5119.99999991994, -5119.99999958856, -2.26070748066514e-08, -1.87277654828418e-08, -10239.9999999601, -5.16046987969842e-08, -5.31357117137667e-08, 0, 0, 0, 66628.2666667558, 0, -1.53033769383271e-08, 0, -3.58042525270042e-08, 71907.5555555411, 71907.5555553528, 0, -6.61767439308836e-08, 0, -3.56351154138197e-07, 15473.7777776204, 15473.7777770036, 0, 4.71087428722355e-08, 30947.5555553787, 0, -2.04058412730524e-07, -9.18329451867325e-07, 1.08417084003069e-07, 1.02868237396867e-07, 1.50367132453454e-07, 1.14317773187498e-08, 1.19009651238306e-07, -2.35992063969611e-07, 3.21718670327635e-07, 0, 0, -43690.6666667239, 0, 0, 0, 2.12836131436882e-08, -65535.9999999952, -65535.9999998677, 0, 5.06778041057586e-08, 0, 2.02435727338945e-07, -21845.3333331897, -21845.33333264, 0, 0, -43690.6666664999, 0, 2.21420122603423e-07, 0, 8.44240434966341e-07, 2.77371255453452e-07, 1.21328588733507e-06, -3.86925746351417e-08, -1.98240369218707e-08, -1.33707957499047e-07, -1.2129098299242e-08, -3.94247338567187e-08, 5.27234196432282e-07, -1.82972226587555e-07, -1.48411678650619e-07, 2.92255959670271e-07, -3.94386421140517e-07, -4.93529031575731e-07, -2.73402959894203e-07, 0, 0, 11650.8444444594, 0, 0, 0, 0, 23301.6888888881, 23301.6888888541, 0, -1.21709092240301e-08, 0, -4.62600928986255e-08, 11650.8444443983, 11650.8444442124, 0, 0, 23301.6888888227, 0, -6.63322418044388e-08, 0, -2.80751867515034e-07, -1.31513930715455e-07, -5.13782341191937e-07, 1.58705505204355e-08, 0, 2.5569260366515e-08, 4.1047678115409e-08, -3.0965159446069e-08, -2.86875752925737e-07, 0, -1.42694643971946e-07, -5.48532409557707e-07, 5.91906880889095e-08, 3.6416232134646e-08, 5.66546354378512e-08, -1.29908148738965e-07, 1.85095274774339e-08, -3.54632911920798e-07, 4.64668628084036e-08, 3.65461749324039e-07, 4.97878275618122e-08, 2.48031531795721e-07, 2.12445849176897e-07, 2.59962086332033e-07}, 
                                                    {0, 0, 0, 9.14285714287374, 0, 0, -188.342857143163, 0, -9.14285714283861, -9.14285714289909, 0, 0, 1499.02222222448, 0, 0, 0, 0, 179.200000000411, 179.199999999541, 0, 0, 0, -6098.48888889754, 0, 0, 0, 0, -1319.82222222358, -1319.82222221925, 0, 0, 0, 0, 0, 0, 0, 0, 13880.8888889076, 0, 0, 0, 0, 4778.66666666966, 4778.66666665664, 0, 0, 0, -1.70982228298187e-08, 0, -2.96595876804556e-08, 0, 0, 0, 0, 1.26996168508439e-08, 0, 0, 0, -17840.3555555784, 0, 0, 0, 0, -9102.22222222425, -9102.22222220919, 0, 0, 0, 1.86085119450074e-08, 0, 5.81102703573908e-08, 0, -2.14910800869331e-08, 0, 0, 0, 7.31290238479507e-08, -1.28998922666762e-08, -2.71681926180517e-08, -1.94845686443938e-08, 0, -3.71503247794768e-08, 0, -2.65754566824421e-08, 0, 0, 12066.9460317607, 0, 0, 0, 0, 8738.13333333498, 8738.13333332532, 0, 0, 0, 0, 0, -5.53215288758158e-08, 0, 0, 0, 0, -1.546069938889e-08, 0, -5.6276149352653e-08, -1.49394591993319e-08, -1.03454018345709e-07, 0, 2.14977427550673e-08, 1.82182242006734e-08, 1.12944894642769e-08, 2.86886942357411e-08, -4.26220843934195e-08, 2.90433544188595e-08, 4.14893144922139e-08, 0, 6.24761695101834e-08, 4.00663626453281e-08, 2.48731782657051e-08, 0, 0, -3328.81269841652, 0, 0, 0, 0, -3328.81269841307, -3328.81269841057, 0, 0, 0, 0, 0, 2.09384876214173e-08, 0, 0, 0, 0, 0, 0, 2.20048369399362e-08, 0, 4.38768436923299e-08, 0, 0, 0, -1.16302058330001e-08, 0, 2.47892392156171e-08, 0, 0, 4.48836577058139e-08, -1.03377025335474e-08, -1.64423419311605e-08, 0, 0, 0, 3.16988700521602e-08, 0, -4.42783784594229e-08, 0, 1.72684554672762e-08, -1.41351798316111e-08, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 9.1428571428549, 0, 0, 0, 0, -179.200000000098, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1319.82222222247, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4778.66666666615, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9102.22222222049, 0, 0, 0, 0, 0, 0, 0, 0, 1.04153608118052e-08, 0, 0, 0, 0, 0, 2.28880856794814e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -8738.13333333203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.28684258790334e-08, 0, 0, -3.20452305184769e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.26917475631989e-08, 0, 0, 0, 0, 0, 3328.8126984122, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.29959142509666e-08, 0, 0, 0, 0, 0, 0, 0, 0, 1.86304081404752e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 5.33333333333668, 0, 0, 0, 0, -97.4222222220889, 0, 0, -42.6666666667405, 0, 0, 0, 0, 0, 0, 639.99999999857, 0, 0, 0, 0, 0, 779.377777778661, 0, 0, 0, 0, 0, 0, 0, 0, -1934.222222218, 0, 0, 0, 0, 0, -5120.00000000736, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2730.66666665933, 0, 0, 0, 0, 0, 15473.777777782, 0, 0, 1.26026127357249e-08, 0, 0, 0, 0, 0, 3.12270447470184e-08, 0, 0, 0, 0, 1.0548473328175e-08, 0, 0, -1.47328896005269e-08, 0, 0, 0, 0, -1456.35555554828, 0, 0, 0, 0, 0, -21845.333333334, 0, 0, 0, 0, 0, 0, 0, 0, -2.57547146596065e-08, 0, 0, -1.56603607506201e-08, 0, 0, -1.58759382001803e-08, 0, 0, 0, 0, 0, 2.898682842128e-08, 0, 0, 7.13402770988975e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 11650.8444444416, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.20128333776159e-08, 0, 0, 0, 0, 0, 0, 0, 0, 1.30165666040628e-08, 0, 0, 0, 0, 0, 0, 1.43498121532076e-08, 0, 0, -3.23997003404453e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 4.26666666667364, 0, 0, 0, 0, -71.1111111115294, 0, 0, -51.200000000175, 0, 0, 0, 0, 0, 0, 398.222222225107, 0, 0, 136.533333333365, 0, 0, 853.33333333479, 0, 0, 0, 0, 0, 0, 0, 0, -910.22222222953, 0, 0, 0, 0, 0, -4778.66666666505, 0, 0, -2275.55555555493, 0, 0, 0, 0, 0, 0, -1.12850741433643e-08, 0, 0, 0, 0, 728.177777788718, 0, 0, 0, 0, 0, 10922.6666666891, 0, 0, -2.20837017671562e-08, 0, 0, 0, 0, 0, 12743.1111110792, 0, 0, 2.52358061112051e-08, 0, -1.59267912641499e-08, 0, 1.25499107694275e-08, 3.21072593778543e-08, 0, 0, 0, 0, -1.05412432249561e-08, 0, 0, 0, 0, 0, -8738.13333335517, 0, 0, 2.43416853428893e-08, 0, -1.09297957790569e-08, 0, 0, 0, -29127.1111110925, 0, 0, 2.98110359456574e-08, 0, 0, 2.6040145813228e-08, 0, 0, -2.09139672867392e-08, -1.2482193111535e-08, 1.20496027154274e-08, -6.17360669073734e-08, 0, 0, -1.15870994471543e-07, 1.35698752117772e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.22348811978612e-08, 0, 0, 0, 0, 0, 0, 0, 0, 23301.6888889022, 0, 0, 0, 0, 0, 2.56756414594912e-08, 0, 0, -1.66889009604344e-08, 1.01028969499976e-08, 0, 0, 0, 1.86522154493022e-08, -1.74657200418951e-08, 0, 1.25540650374959e-08, -1.70566204940451e-08, 0, 0, 1.77640541969221e-08, -2.31651541993717e-08, 0, 1.1869287392795e-08, 6.65814670394097e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 3.99999999997026, 0, 0, 0, 0, -58.6666666663195, 0, 0, -58.6666666653121, 0, 0, 0, 0, 0, 0, 256.000000000787, 0, 0, 255.999999993696, 0, 0, 860.444444432196, 0, 0, 0, 0, 0, 0, 0, 0, -341.333333342247, 0, 0, -341.333333321679, 0, 0, -3754.66666664706, 0, 0, -3754.666666606, 0, 0, 0, 0, 0, 0, 1.21158517875292e-08, 0, 0, 0, 0, 1.72508654961878e-08, 0, 0, 0, 0, 0, 5006.22222219338, 0, 0, 5006.2222221273, 0, 0, 0, 0, 0, 16383.9999998701, 0, 1.45799491487088e-08, -1.7366686948532e-08, 1.46176143770023e-08, 1.9687788094894e-08, 1.67245878308137e-08, 0, -2.33336107174099e-08, 0, 0, 0, 0, -1.08653757930804e-08, 0, 0, 0, 0, 0, 0, 0, 0, 5.98954261032589e-08, 0, 1.00742891406355e-08, 0, 0, 0, -21845.3333332383, 0, 0, -21845.3333331243, 0, -1.64852816158611e-08, -4.53887126323373e-08, 0, 0, 1.29251202870229e-08, 0, -2.42050849713555e-08, 4.79284028499629e-08, 0, -4.00176470500688e-08, 8.90127983914077e-08, -3.89090397893043e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.72106977426594e-08, 0, 0, 0, 0, 0, -3.55055239125367e-08, 0, 0, -9.10685091207922e-08, 0, 0, -2.47510736002027e-08, 0, 0, 2.5532174031127e-08, 0, 0, 29127.1111109962, 0, -1.9462713142638e-08, 2.38142931749488e-08, 1.22737032161713e-08, -1.49583824085353e-08, 3.56607669927155e-08, 0, 2.51498397814485e-08, -1.58203613935561e-08, 3.840893996635e-08, 1.28375416237555e-08, 1.52763219711476e-08, -6.9687287805425e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 4.26666666665063, 0, 0, 0, 0, -51.2000000002553, 0, 0, -71.1111111109389, 0, 0, 0, 0, 0, 0, 136.533333334807, 0, 0, 398.222222220681, 0, 0, 853.333333331905, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -910.222222220093, 0, 0, -2275.55555554392, 0, 0, -4778.66666665472, 0, 0, 0, 0, 0, 0, -1.2306064973458e-08, 0, 0, 0, 0, 0, 0, 0, 728.177777773491, 0, 0, 1.51910746945424e-08, 0, 0, 10922.6666666373, 0, 0, 0, 0, 0, 12743.1111110713, 0, 0, 2.44953846529523e-08, 0, 0, 0, 1.06328638879318e-08, 2.75189652131994e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -8738.13333330536, 0, -1.06375449684613e-08, 0, 0, 0, 4.95583897179908e-08, 0, 0, -29127.111111065, 0, 0, 2.43426219726303e-08, 0, 1.2094107306327e-08, -2.20298886116321e-08, -1.12279277901178e-08, 1.51312623503474e-08, -4.62104157620795e-08, 0, 2.48012578344788e-08, -4.75280687036342e-08, 1.30471424792176e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 23301.6888888672, 0, 0, 2.14126341306607e-08, 0, 0, -1.65948443701081e-08, 0, 0, -1.99224363440544e-08, 0, 1.71759190583734e-08, 0, 0, 1.48002717952584e-08, -2.35501919682027e-08, 0, -1.51887410716417e-08, 1.56839268079565e-08, -2.43204228848559e-08, -1.04140959309839e-08, -1.20675507833208e-08, 5.84358280762889e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 5.33333333328813, 0, 0, 0, 0, -42.666666666016, 0, 0, -97.4222222215106, 0, 0, 0, 0, 0, 0, 0, 0, 0, 639.999999997026, 0, 0, 779.377777771234, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1934.22222221519, 0, 0, 1.41639519591918e-08, 0, 0, -5119.9999999742, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2730.66666666276, 0, 0, -3.17515202929483e-08, 0, 0, 15473.7777777345, 0, 0, 0, 0, 0, -6.03614352290392e-08, 0, 0, 0, 0, 0, 0, 0, -1.01647137154606e-08, 0, 0, 0, 0, 0, 0, 0, -1456.35555555343, 0, 0, 1.44877277968357e-08, 0, 0, -21845.3333333044, 0, 0, 0, 0, 0, 3.86330765717129e-08, 0, 0, 8.00200190343466e-08, 0, 0, -1.40448970285971e-08, 0, -1.04275780419833e-08, 0, 0, 0, 1.49925668083586e-08, 0, -1.96590912917965e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 11650.8444444357, 0, 0, 0, 0, 0, -1.67997574248616e-08, 0, 0, -2.93939720406017e-08, 0, 0, 0, 0, 0, 0, 0, 0, -3.55252807551302e-08, 0, 0, 0, 0, 0, 1.47471189809434e-08, 0, 1.03848816658307e-08, 0, 1.58022314059033e-08, 1.07156089024059e-08, 1.4177298006214e-08, -2.37280157864385e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 9.14285714283388, 0, 0, 0, 0, 0, 0, 0, -179.199999999721, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1319.8222222207, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4778.66666666355, 0, 0, 1.02015680347736e-08, 0, 0, 1.28477015692855e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9102.22222221728, 0, 0, 0, 0, 0, -2.41474511497593e-08, 0, 0, 0, 0, 0, -2.90166695116065e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -8738.13333333089, 0, 0, 0, 0, 0, 2.21060332261664e-08, 0, 0, 0, 0, 0, 3.27991585002972e-08, 0, 0, 3.65531533930163e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3328.8126984121, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.76842676058447e-08, 0, 0, 0, 0, 0, 0, 0, 0, -1.88617963470256e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 9.14285714284895, 0, 0, 0, 0, 0, 0, 0, -179.200000000058, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1319.82222222245, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4778.66666666735, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.16811455739084e-08, 0, 0, 0, 0, 0, 0, 9102.22222222294, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.2635625530258e-08, 0, 0, 0, 0, 0, -2.59951137087559e-08, -1.73204473410649e-08, 0, 0, 0, 0, 0, 0, -8738.13333333193, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.94729908663771e-08, 0, -2.78198703303552e-08, 0, 0, 0, 1.76551757474483e-08, 0, 1.77238598700775e-08, 0, -1.09068728640653e-08, 5.41261437180223e-08, 0, 5.98616868813624e-08, 5.06453117373085e-08, 0, 0, 0, 0, 0, 0, 3328.81269841224, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.01215083952424e-08, 0, 0, 0, -3.68953660980882e-08, 0, 0, 0, -3.23196025455376e-08, 0, 0, 0, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 5.3333333333008, 0, 0, 0, 0, 0, 0, 0, -97.4222222220044, 0, -42.6666666665583, 0, 0, 0, 0, 0, 0, 0, 639.999999998883, 0, 0, 0, 0, 779.377777776221, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1934.22222222019, 0, 0, 0, 0, 0, -5119.99999998563, 0, 0, 0, -1.67174220792042e-08, 1.01222732757789e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2730.66666666515, 0, 0, 0, 0, 0, 15473.7777777542, 0, -2.2857166248261e-08, 0, 4.91098191195714e-08, -2.74223902837066e-08, 0, 0, -3.26295715818555e-08, 4.0635546524686e-08, -2.14272837964609e-08, 4.16447244288156e-08, -5.20485285774845e-08, 0, -4.13410151459306e-08, 1.48936936430133e-08, 0, 0, 0, 0, 0, 0, -1456.35555555699, 0, 0, 0, 0, 0, -21845.3333333372, 0, 3.50375341193311e-08, 0, -4.57095044904407e-08, 3.42553462879698e-08, 0, 0, 0, 5.4565699974478e-08, 0, 6.35306515246529e-08, -2.70392275770854e-08, -1.63203795907278e-08, -9.14231939037457e-08, 3.32221275675729e-08, 2.89762318898864e-08, 7.45696641465209e-08, -8.10771829667576e-08, 3.04041675244849e-08, 5.14600832265541e-08, -5.11448849074472e-08, -7.94212771470613e-08, -6.12599471425404e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 11650.8444444401, 0, -1.22024895549038e-08, 0, 1.72350773663122e-08, 0, 0, 0, 0, -2.49111171508906e-08, 0, -1.26551389672961e-08, 1.4011998781837e-08, 0, 4.77544676028763e-08, 0, -1.81117232622503e-08, -2.42453139340977e-08, 0, 0, -2.07200499992964e-08, 1.9965560424622e-08, -1.51612422782084e-08, 4.02221420602522e-08, -1.83230818584637e-08, -1.7723869145122e-08, -6.96019850545123e-08, 3.49839680733687e-08, 5.23235134511814e-08, -4.64085828741842e-08, 1.11560460708995e-08, 1.44460935425428e-08, -5.37084722083557e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 4.26666666670214, 0, 0, 0, 0, 0, 0, 0, -71.1111111115941, 0, -51.199999999688, 0, 0, 0, 0, 0, 0, 0, 398.222222225226, 0, 136.53333333262, 0, 0, 853.333333332463, 0, 0, 0, 0, 0, 0, 0, 0, 0, -910.222222230643, 0, 0, 0, 0, 0, -4778.66666666551, 0, -2275.55555555146, 0, 0, 1.72149235078058e-08, 0, 1.37722515539929e-08, 3.37009978255132e-08, 0, 0, 0, 0, 0, 0, 728.177777794293, 0, 0, 0, 0, 0, 10922.666666676, 0, 2.01756419457937e-08, 0, -1.41429691315607e-08, -3.94607098090061e-08, 0, 0, 12743.1111111193, 1.86728377508582e-08, 3.30300454384505e-08, 1.75859812383195e-08, -4.9202431151802e-08, -5.4033546676279e-08, -2.33281318134118e-08, 0, 0, 0, 0, 0, 0, 0, -1.24960484164997e-08, 0, 0, 0, 0, 0, -8738.13333331701, 0, -2.33500879585836e-08, 0, 0, 2.68179751904823e-08, 0, 0, 0, -29127.1111111005, 0, -2.84200350190299e-08, -3.8906896891162e-08, 0, 1.16513776710266e-08, 1.12339720999394e-07, 6.66478684883628e-08, 0, -1.94706550949282e-08, 1.93503083731901e-08, 1.21668429802314e-07, -5.99333883992171e-08, 1.48135565645934e-07, 1.50823787129932e-07, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.36267330800156e-08, 0, 0, 0, 23301.6888888826, 0, 0, 1.07822854710072e-08, 0, 0, -6.07413591923565e-08, -3.18440515129646e-08, -2.25845865209822e-08, 0, 0, -1.77789105006918e-08, 3.15471834399275e-08, 0, 0, -1.42724565614718e-07, -3.95490991969604e-08, 0, 1.3707848529349e-08, -8.95097971808596e-08, 1.42048118991918e-08, 4.77366118515422e-08, 0, -6.4051533481567e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 3.99999999988622, 0, 0, 0, 0, 0, 0, 0, -58.6666666655072, 0, -58.6666666651831, 0, 0, 0, 0, 0, 0, 0, 255.999999995358, 0, 255.999999992828, 0, 0, 860.444444427947, 0, 0, 0, 0, 0, 0, 0, 0, 0, -341.333333328359, 0, -341.333333321024, 0, 0, 0, -3754.66666661425, 0, -3754.66666661617, 1.16565530104801e-08, 2.19848783245724e-08, 1.50306275259602e-08, -1.62677491227633e-08, 0, 2.71568582810026e-08, 0, 0, 0, 0, 0, 0, 0, 0, -1.62887390787498e-08, 0, 0, 0, 5006.22222212114, 0, 5006.2222221209, -1.93940525554822e-08, -6.57570015306565e-08, 0, 0, 0, 16383.999999823, 0, -4.13234699680689e-08, 0, -1.7642781396714e-07, 0, -5.73010608428455e-08, -4.79008834177465e-08, 0, 0, 0, 0, 0, 0, 0, 0, 1.10769684015576e-08, 0, 0, 0, 6.43568083496631e-08, 0, 6.25999587943032e-08, 1.44273609024186e-08, 2.39460631672784e-08, 2.45609519769667e-08, 0, 0, 0, -21845.3333331396, 0, -21845.3333331731, 2.68740242863363e-08, 2.96272888875079e-08, -1.29967682420839e-08, 2.04614758739302e-07, -2.08394171473117e-08, -1.45503252603427e-08, -3.18249814448446e-08, -1.53014630600128e-08, 1.82805604854103e-07, 1.25167024606799e-07, -2.96132775577163e-08, -8.26803610295854e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.81687173750172e-08, 0, 0, 0, 0, 0, 0, 0, 0, -6.46966175868001e-08, 0, -5.60590614571623e-08, -2.23924630795579e-08, -1.85850288883396e-08, 1.23681326977128e-08, -9.55909117894636e-08, 0, 5.11469363803972e-08, 0, 0, 29127.1111110267, 0, -1.41229875726823e-08, 4.97462980214435e-08, -1.04089229859767e-07, 5.31600301565053e-08, -7.25319881382018e-08, 0, -6.39613742450913e-08, -2.77383440985388e-08, 5.60223696670154e-08, 1.98176292617909e-08, 1.42500230552756e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 4.26666666658452, 0, 0, 0, 0, 0, 0, 0, -51.1999999995329, 0, -71.1111111099135, 0, 0, 0, 0, 0, 0, 0, 136.53333333241, 0, 398.222222216792, 0, 0, 853.33333332819, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -910.222222203597, 0, 0, 0, -2275.55555552835, 0, -4778.66666663678, 0, 2.09866523405226e-08, 0, 4.0895513052966e-08, 0, 1.24311314378279e-08, 0, 0, 0, 0, 0, 0, 0, 0, 728.177777758745, 0, 0, 0, -3.62055577034313e-08, 0, 10922.6666666426, -2.93657988727556e-08, -1.12529787009679e-08, -2.62995147305551e-08, 0, 0, 12743.1111110649, -4.66988384003324e-08, 1.00919273504428e-08, -2.50737709073649e-08, -3.93508118968083e-08, 6.39115791791826e-08, 0, 1.32953133165492e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.57646192750772e-08, 0, -8738.13333329714, 3.16951187784662e-08, 6.08948102645478e-08, 0, 0, 0, 0, 6.07857654368425e-08, 0, -29127.1111110371, 7.18392123709844e-08, 1.29285714895482e-08, 3.14485236389401e-08, 1.22277188721526e-07, -4.97668991525002e-08, 7.47847843750069e-08, 3.19062545290204e-08, -3.86863260228028e-08, -1.96648107985131e-08, -1.17376994974337e-07, 2.72630821960029e-08, -2.08898770190715e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.16125595266055e-08, 0, -1.97194519617015e-08, 0, 0, 0, 0, -4.18473025124016e-08, 0, 23301.6888888788, -1.95926943833607e-08, -1.25216723358989e-08, -1.9515929839017e-08, -5.67014047569218e-08, 1.97042384832989e-08, -6.1031465151472e-08, 0, 0, -3.39317694967296e-08, -3.69303146100977e-08, 1.41808486561473e-08, -4.17896747594433e-08, -6.96867247464639e-08, 0, 1.83859853488382e-08, -7.86994459967229e-08, 5.41687180331809e-08, 6.08571774202057e-08, -7.46011312503926e-08, 0, -5.67498303746463e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 5.33333333331183, 0, 0, 0, 0, 0, 0, 0, -42.6666666666455, 0, -97.4222222212483, 0, 0, 0, 0, 0, 0, 0, 0, 0, 639.999999994088, 0, 0, 779.377777770674, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1934.22222221256, 0, 0, 0, 0, 0, -5119.99999997079, 0, -2.96887357882918e-08, -1.30750987626162e-08, -4.39335087053422e-08, -3.24750623479362e-08, -4.75931998935355e-08, 0, 0, 0, 0, 0, 0, 0, 0, 2730.66666665191, 0, 0, 0, 0, 0, 15473.7777777033, 0, 2.4651076708829e-08, 4.94486811390959e-08, 0, 0, -4.65153732131502e-08, 2.85724205244788e-08, 3.95229152841156e-08, 6.39950765103423e-08, 1.38540470140836e-07, 5.92431663464368e-08, 1.08945281353011e-07, 1.69523650490947e-07, 0, 0, 0, 0, 0, 0, 0, 0, -1456.35555554543, 0, 0, 0, 0, 0, -21845.33333328, 0, -5.20147181366605e-08, -2.95866081397587e-08, 0, 0, 0, 2.46067624149472e-08, 0, 5.03596358153878e-08, -2.53065319024096e-08, 0, -6.59622222718083e-08, -1.84760297829738e-07, -1.02224757071601e-07, -1.60506471338008e-07, -3.8281838065399e-08, -9.94571532335271e-08, -1.33849656594041e-07, -2.60175767522018e-08, -2.66356604181124e-07, -2.8660478470641e-07, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 11650.8444444332, 0, 1.50359350215695e-08, 0, 0, 0, 0, 0, 0, -3.99309640398626e-08, 0, 0, 2.42211196831414e-08, 6.70398120425532e-08, 4.32020206378137e-08, 7.18136362219679e-08, 0, 0, -2.24849809883612e-08, 1.83469814310957e-08, 2.99435863344129e-08, 4.19894550204432e-08, 1.10809286342876e-07, 8.14690479729855e-08, 8.10092378647926e-08, 4.99336525158255e-08, 9.68595928671789e-08, 1.48707219566983e-07, 1.67749242324477e-07, 1.45964194466231e-07, 2.65988171499626e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 9.1428571427734, 0, 0, 0, 0, 0, 0, 0, 0, 0, -179.199999998851, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1319.82222221636, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4778.66666665053, 0, 0, 0, 2.53848552787108e-08, 0, 4.10519480727778e-08, 0, 0, 1.28860822178583e-08, 1.11184826101057e-08, 1.0976343991668e-08, 1.87041720210375e-08, 0, 0, 0, 0, 0, 0, 0, 0, 9102.22222220222, 0, 0, 0, -2.93289724622393e-08, 0, -7.63055013909084e-08, 0, 1.01390398923638e-08, -3.86106773232567e-08, 0, 0, -9.35967193352286e-08, 0, -1.82568497684528e-08, 0, -3.34124414695026e-08, -2.28119693085319e-08, -7.14970555140356e-08, -4.01086353320995e-08, 0, 0, 0, 0, 0, 0, 0, 0, -8738.13333332123, 0, 0, 0, 0, 0, 6.92995187496151e-08, 0, 0, 3.1841714039055e-08, 0, 0, 0, 7.36146084641355e-08, 0, 1.25067712539701e-07, 1.08230579628564e-08, 0, 0, 2.02113115589277e-08, 3.88518290175321e-08, 1.08140890186614e-07, 0, 3.0526334058701e-08, 6.6905886225735e-08, 0, 2.71620929355522e-08, 7.69418536884946e-08, 0, 0, 0, 0, 0, 0, 0, 0, 3328.8126984096, 0, 0, 0, 0, 0, -2.50886366402092e-08, 0, 0, -1.05522660012755e-08, 0, 0, 0, -2.76404119045214e-08, 0, -5.2138790977098e-08, 0, 0, 0, 0, -1.86953119873944e-08, -4.86307540089998e-08, 0, 0, -5.39126657763024e-08, 0, -1.24673812520726e-08, 0, -3.41500863019196e-08, -2.34118017386799e-08, -6.61213441515813e-08, -2.68249778683951e-08, 1.41466488292994e-08, -4.29202566588531e-08, -5.11557615651007e-08, -3.6206543205798e-08, -6.58658007783702e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 1791.99999998755, 0, 0, 0, -8.70251591755591e-08, 0, -17459.1999999569, 0, -17459.1999997632, 4.40610530118609e-08, 0, -9.24463807092452e-08, -17459.2000001241, 0, 4.06234716651212e-07, 0, 69142.7555556018, 0, 69142.7555542378, -4.54037716140937e-07, 0, 6.73110874306737e-07, 138285.511109906, 0, 1.71672208521473e-07, 138285.511111806, 138285.511111482, 69142.755557424, 0, -9.75912487956971e-07, 0, -142677.333333921, 0, -142677.333329805, 1.36788600983569e-06, 0, -1.86495752828614e-06, -428031.999996969, 0, -428031.999994751, 1.93079303407581e-06, 0, -3.02332227306504e-06, -428032.000001228, -428031.999999448, -142677.333342488, -856064.00000334, -428032.000007567, -428032.000005844, 0, 1.26062769245373e-06, 0, 162019.555556887, 0, 162019.555550536, -1.79986809863268e-06, 0, 2.45975265477802e-06, 648078.222218375, 0, 648078.222212455, -6.56952938165285e-06, 0, 7.94279824417218e-06, 648078.222222893, 648078.222219556, 162019.555573475, 972117.333323085, 0, 1.53708207333089e-06, 1944234.66667251, 972117.33334394, 1944234.66667385, 972117.33333233, 648078.222250367, 648078.222250837, 1944234.66668733, 0, -8.32898919302129e-07, 0, -96119.4666679053, 0, -96119.4666630386, 1.09078342558648e-06, 0, -1.58612714849084e-06, -480597.333331256, 0, -480597.333324567, 6.93969693910014e-06, 0, -7.59065430544097e-06, -480597.333333211, -480597.33333048, -96119.4666820027, -961194.666656668, 0, -961194.666655153, 4.88652007299322e-06, 0, -8.12025494837848e-06, -1922389.33333636, -961194.666672301, -1922389.33334053, -961194.666653436, -480597.33337076, -480597.333376117, -2883584.00001277, -961194.666695717, -961194.666684518, -2883584.00001964, -2883584.00001281, -1922389.33339162, 0, 2.20807908487672e-07, 0, 23301.6888893065, 0, 23301.6888878476, -2.46370282965577e-07, 0, 4.06779369290329e-07, 139810.133333051, 0, 139810.133330229, -2.39309328646145e-06, 0, 2.45914506992444e-06, 139810.13333318, 139810.133332385, 23301.6888937175, 349525.333329732, 0, 349525.333328812, -4.03448091519702e-06, 0, 4.92184887006128e-06, 699050.666666784, 349525.333334109, 699050.666669429, 349525.333325214, 139810.133349524, 139810.13335285, 466033.777772234, 0, 1.01663362024264e-06, 1398101.33333859, 466033.777787201, 1398101.33334156, 466033.77777459, 349525.333353554, 349525.333353146, 1398101.33333765, 1398101.33333274, 699050.666709044, 2097152.00000907, 1398101.33336256, 1398101.33335603}, 
                                                    {0, 0, 0, 0, 0, 0, 0, -85.3333333323817, 0, 0, 0, -1.521143907071e-08, 0, 1644.08888888622, 0, 85.333333316293, 0, 0, 0, 85.3333333332591, 0, 5.21556840674273e-08, 0, -11798.7555555568, 0, 1.03886022051443e-07, 0, 0, 0, -1558.75555549937, 0, -1.5054390097857e-08, -1558.75555554946, 0, 0, 0, -1.00098015112932e-07, 0, 41187.5555555812, 0, -2.99453359642134e-07, 1.15199913859429e-08, 0, 0, 10239.9999999126, 0, -2.88791124062375e-07, 4.10670256302162e-08, 0, 5.23136687316452e-08, 10239.9999999733, 1.74385097636566e-08, 4.67492462626944e-08, -4.14490201217587e-08, 0, 6.44234670747048e-08, 0, 1.08457412679881e-07, 0, -74638.2222222746, 0, 4.46506462628489e-07, -1.35937591412046e-08, 0, -1.29034402571358e-08, -30947.5555555351, 0, 6.04067193690659e-07, -1.08985780982035e-07, 0, -8.63370997522565e-08, -30947.55555548, -6.5080912029905e-08, -6.45550464022506e-08, 4.58609003526825e-07, 0, -1.55373596938792e-07, 1.50768350320916e-07, 4.66661795319933e-08, 6.69166169595562e-08, -8.73217553501116e-08, -4.79539805476181e-08, -1.56703499554408e-07, -7.6548279256763e-08, 0, -6.19335919116269e-08, 0, 66992.3555555978, 0, -3.35433164863683e-07, 0, 0, 0, 43690.6666667373, 0, -5.71905300539084e-07, 9.79333107982104e-08, 0, 2.49798813663313e-08, 43690.6666666064, 4.19460322105259e-08, 4.38890626611351e-08, -2.92022281775202e-07, 0, -6.56254243431367e-07, 1.6398430844058e-07, 0, 1.13086601218969e-07, -2.09693549034439e-07, -7.37829380431875e-08, -1.64942009551316e-08, 1.55155182588354e-07, 1.41689837656908e-07, 1.58883423992427e-07, -1.39508432587447e-07, 3.67983265410846e-08, 3.20179423471329e-07, -2.12827493637203e-07, 8.87058213534766e-08, 1.61901996527621e-07, 0, 1.44827530343553e-08, 0, -23301.6888889011, 0, 1.00446866273933e-07, 0, 0, 0, -23301.6888889341, 0, 2.01244271307857e-07, -2.37495795080911e-08, 0, 0, -23301.6888888672, -2.31475205926673e-08, -1.62018097902006e-08, 5.03371665946532e-08, 0, 3.03843104600941e-07, -5.09659958957456e-08, 0, -3.99942315518302e-08, 8.11016781946097e-08, 4.35244628683769e-08, -2.44640651659689e-08, -2.47596528534544e-08, -5.62356749738719e-08, -6.47367279028256e-08, 2.52480468980098e-07, 0, -4.97556365722778e-08, 1.38297616450409e-07, 2.23271556007068e-08, 8.71831880242989e-08, -1.16436495088911e-07, -6.98849225638934e-08, -6.07065643493371e-08, 1.73333141828879e-07, -3.25685922798675e-08, -1.56404770130107e-07, 2.26905096098237e-07, -2.27965180423096e-08, 9.07494117034124e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, -85.333333332591, 0, 0, 0, -1.49099812377999e-08, 0, 85.3333333280022, 0, 1644.08888887748, 0, 0, 0, 85.3333333326811, 0, 5.12164325067074e-08, 0, 1.15883446465129e-08, 0, -11798.7555555009, 0, 0, 0, -1558.75555547118, 0, 0, 0, -1558.75555554749, 0, 0, -9.92615446924717e-08, 0, 0, 0, 41187.5555554402, 0, 0, 0, -2.27275791200417e-07, 0, 10239.9999996323, 1.12899814814118e-08, 0, 1.14589993389722e-08, 0, 10239.9999999656, 0, 0, 0, 0, 0, 1.09834465416279e-07, 0, -3.10777860871214e-08, 0, -74638.2222220982, 0, 0, -1.60276198343871e-08, 1.9813927982839e-07, 0, -30947.555554954, -4.10411069879498e-08, 0, -4.80061899713298e-08, -2.14928521098008e-08, -30947.5555554743, -1.04755059159416e-08, 9.43802114821456e-07, 0, -5.26216172731864e-08, -6.00474332152591e-08, 5.29269485243702e-08, 2.75698738105697e-08, 1.20813507745164e-07, 2.482080622615e-08, -2.38844951663659e-08, 3.08304888946061e-08, 0, -6.46838563283061e-08, 0, 2.95329491638732e-08, 0, 66992.3555554854, 0, 0, 1.10062683208963e-08, 0, 0, 43690.6666662585, 4.1658576753962e-08, 0, 1.96623693591347e-08, 2.27496876306923e-08, 43690.6666666194, 0, -8.30035563271574e-07, 0, -1.188715550025e-06, 3.22273625573523e-08, 0, 0, 8.15580378311438e-08, -3.41443372435113e-08, -4.36375167931088e-08, -1.40233943097192e-07, -4.37655339086787e-08, -4.35237918322299e-08, 0, -1.05968537938621e-07, 1.79151634572305e-08, -8.58970727910163e-08, -1.0894546799484e-07, -5.82073135390663e-08, 0, 1.57165046486343e-08, 0, 0, 0, -23301.6888888716, 0, 0, 0, -3.82522292190251e-08, 0, -23301.6888887947, 0, 0, 0, 0, -23301.6888888811, 0, 2.16993471573148e-07, 0, 4.64146020793812e-07, -4.9716571370281e-08, 0, -1.116455896353e-08, -5.07197083029418e-08, 0, 0, 6.16556559393764e-08, 0, 1.00150714417854e-08, 6.10455656627662e-07, 0, 0, 0, 5.9123739250282e-08, 1.29027627914131e-08, 0, 4.14401942649803e-08, 9.16010665845756e-08, 0, 5.65864987710811e-08, 1.54495615772874e-07, -3.55011840137709e-08, 6.40825791092958e-08, 1.36567091733386e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, -1791.99999998468, 0, 0, 0, -3.01521148699277e-08, 0, 27016.5333332913, 0, 12680.533333043, 5.88915340176509e-08, 0, 1.67185696107917e-07, 12680.5333335476, 0, 0, 0, -136772.266666727, 0, -35327.9999984011, -6.22241997740217e-08, 0, -9.16716449903318e-07, -172100.266665409, 0, -1.10518427539352e-06, -172100.266667476, -70656.000001392, -35328.0000027045, 0, 2.13640125093169e-07, 0, 331093.333334043, 0, 48469.3333290019, -1.81543555591832e-07, 0, 2.25912708114971e-06, 710655.999996379, 0, 428031.999995252, 2.1373543589065e-06, 0, 5.51633332645046e-06, 710656.000000795, 145408.000003465, 48469.3333454376, 856064.000005368, 428032.000008875, 145408.000012688, 0, -4.36399475709894e-07, 0, -420522.666668365, 0, -32767.9999936806, 4.50193867652767e-07, 0, -2.80335353489663e-06, -1294335.99999446, 0, -518826.666658864, -1.05674955673076e-06, 0, -1.06942819544307e-05, -1294335.99999938, -131072.000004304, -32768.0000229272, -1359871.99998996, 0, -1.03164638897654e-05, -2719744.00000631, -1359872.00000993, -1556480.00001465, -196608.000018819, -518826.66669917, -131072.000044708, -1556480.00003117, 0, 3.59656715795723e-07, 0, 270882.133334994, 0, 8738.13332861392, -3.41455292181235e-07, 0, 1.77092421175011e-06, 1092266.66666299, 0, 305834.66666008, -6.2990015938363e-07, 0, 8.94209360647739e-06, 1092266.66666533, 43690.6666692802, 8738.1333528371, 1660245.33332166, 0, 1135957.33332462, 6.85928786712472e-06, 0, 1.53865939098811e-05, 3320490.66666796, 1660245.33333652, 1223338.66668372, 87381.333343231, 305834.666709858, 43690.666725477, 3407872.0000171, 1135957.33336295, 87381.3333816156, 3407872.00002309, 1835008.00003412, 1223338.66674377, 0, -1.0866338425758e-07, 0, -69905.0666672491, 0, 1.40498076563983e-06, 8.51788196067738e-08, 0, -4.52530916810066e-07, -349525.333332586, 0, -69905.0666643271, 4.80038164883914e-07, 0, -2.67909825753009e-06, -349525.333332777, -6.34813980063236e-07, -6.20126266706218e-06, -699050.666661817, 0, -349525.333330805, -1.0596812937645e-06, 0, -6.85278656160531e-06, -1398101.33333229, -699050.666666194, -349525.333340439, -1.11099005860191e-06, -69905.06668547, -2.52880523578932e-05, -699050.66666148, 0, -6.15363140322779e-06, -2097152.00000552, -699050.666674822, -1398101.33334533, -1.47055430841208e-05, -349525.333354487, -3.4310402043919e-05, -2097152.00000326, -699050.666679024, -349525.333386354, -2097152.0000142, -1398101.33336749, -699050.66670688}, 
                                                    {0, 0, 0, 0, 0, -3.88048753962893e-08, 0, 1493.33333331858, 1.2260609657948e-08, 0, 0, 2.54310202315094e-07, 0, -25500.444444414, 0, -7580.44444417021, -1.21208633489911e-07, 0, -1.66481740041727e-07, -7580.44444461124, 0, -8.42804358054817e-07, 0, 153031.111111147, 0, 14279.1111095309, 4.71971537834072e-07, 0, 8.05776328812905e-07, 119523.555554725, 0, 1.39769222868375e-06, 119523.555555916, 28558.2222236829, 14279.1111130308, 0, 1.56272560500867e-06, 0, -425756.444444925, 0, -11832.8888844027, -9.30726571868806e-07, 0, -1.88452156006441e-06, -620771.555553184, 0, -206847.999997071, -4.27843022705756e-06, 0, -5.46107672575692e-06, -620771.555555142, -35498.6666717313, -11832.8888972118, -413696.000003531, -206848.000004744, -35498.6666776925, 0, -1.64674072919169e-06, 0, 602567.111112315, 0, 3640.88888209044, 9.96121040609575e-07, 0, 2.28329157667282e-06, 1354410.6666628, 0, 156558.222217659, 6.23465781875e-06, 0, 9.3811308231075e-06, 1354410.66666448, 14563.5555641726, 3640.88890453062, 904760.88888305, 0, 1.26670910577952e-05, 1809521.77777932, 904760.888891416, 469674.666678856, 21845.3333569175, 156558.222240968, 14563.5555894008, 469674.666687067, 0, 9.22729540527415e-07, 0, -422343.111112321, 0, 5.24610734243197e-06, -5.5548035660156e-07, 0, -1.43994775559012e-06, -1310719.99999734, 0, -43690.6666628407, -4.3948352932084e-06, 0, -7.48914744669679e-06, -1310719.99999764, -6.92874448239938e-06, -1.34224858790518e-05, -1398101.33332589, 0, -553415.111107228, -1.24073226273148e-05, 0, -1.51363864737891e-05, -2796202.6666638, -1398101.33333075, -174762.666682895, -2.24558787510692e-05, -43690.6666926292, -4.18705700498289e-05, -1660245.33334286, -553415.111124873, -4.5592845019949e-05, -1660245.33334204, -262144.000029216, -174762.666715666, 0, -2.13704064781726e-07, 0, 116508.444444876, 0, -1.61122015735045e-06, 1.30300337938808e-07, 0, 3.64608773342288e-07, 466033.777777231, 0, -1.40924878457102e-06, 1.20207806861741e-06, 0, 2.19813198118337e-06, 466033.77777694, 2.14799773850721e-06, 4.3436604024892e-06, 699050.66666337, 0, 116508.444443861, 4.15373895450924e-06, 0, 6.00367794611096e-06, 1398101.33333108, 699050.666664421, 7.48410113829312e-06, 8.02426173514834e-06, 1.1589318288712e-05, 1.7552813665469e-05, 466033.777774933, 0, 7.36139527314123e-06, 1398101.33333493, 466033.777779815, 466033.777785862, 2.00033049124911e-05, 116508.444455637, 2.77276019956018e-05, 1398101.33333108, 1.4897340766671e-05, 3.35461549790429e-05, 699050.666673672, 466033.7777956, 3.03512502919624e-05}, 
                                                    {0, 0, 0, 0, 0, 3.60393573446367e-08, 0, -895.999999988667, 0, 0, 0, -2.47610552450251e-07, 0, 16255.9999999769, 0, 3114.6666664584, 6.07438326459033e-08, 0, 7.47849065274284e-08, 3114.66666672741, 0, 8.62202640419451e-07, 0, -106609.777777788, 0, -3583.99999873104, -2.57892540269226e-07, 0, -3.46302537560737e-07, -52849.7777772607, 0, -6.75652854907252e-07, -52849.7777778031, -7168.00000061414, -3584.00000071274, 0, -1.67732565114859e-06, 0, 329728.000000235, 0, 1365.33332960514, 5.35821217808659e-07, 0, 7.91186390753463e-07, 308565.333332244, 0, 56661.3333311054, 2.2296807629142e-06, 0, 2.4916141067393e-06, 308565.333332697, 4096.00000235869, 1365.33333645808, 113322.6666675, 56661.3333346681, 4096.00000448429, 0, 1.84820168795289e-06, 0, -515185.777778356, 0, 5.72434916590068e-06, -5.87157026944269e-07, 0, -9.33176170713683e-07, -784611.55555414, 0, -20024.8888845769, -3.53445314467727e-06, 0, -4.03035624105963e-06, -784611.55555377, -4.29954125271488e-06, -5.81337579141268e-06, -289450.666663314, 0, -6.14931477066257e-06, -578901.333332016, -289450.666666138, -60074.6666707731, -1.09605050816338e-05, -20024.8888951014, -1.30914493227436e-05, -60074.6666733476, 0, -1.07852266812918e-06, 0, 393216.000000564, 0, -4.42660471065905e-06, 3.20681842079542e-07, 0, 5.95613840302374e-07, 873813.333332584, 0, -4.12813574716782e-06, 2.68464645786109e-06, 0, 3.1483385102672e-06, 873813.333331615, 3.5559558809695e-06, 4.99490477846275e-06, 567978.666663345, 0, 87381.3333300717, 6.5919456017808e-06, 0, 7.00670464942634e-06, 1135957.33332983, 567978.66666393, 6.13694316901036e-06, 1.16175326376201e-05, 9.19239135193272e-06, 1.57396465970412e-05, 262144.000001345, 87381.3333370272, 1.9833679298799e-05, 262143.999999927, 1.08736217002386e-05, 1.66074346946255e-05, 0, 2.58989496442759e-07, 0, -116508.444444641, 0, 1.35716929624994e-06, -7.33839107924351e-08, 0, -1.50545063067002e-07, -349525.333333265, 0, 1.55774852337233e-06, -7.61841038555111e-07, 0, -8.90298140249907e-07, -349525.333332742, -1.13056613331912e-06, -1.63003353388012e-06, -349525.333332063, 0, 1.12954352464044e-06, -2.30621357298504e-06, 0, -2.62803701985241e-06, -699050.666664796, -349525.333331687, -3.09410243161998e-06, -4.47201709595176e-06, -4.17949410460496e-06, -6.44024975184135e-06, -116508.44444282, 0, -3.62053478978871e-06, -349525.333332391, -116508.444444225, -2.13156189713362e-06, -9.68537069919449e-06, -3.68402861991513e-06, -1.12634080586676e-05, -349525.333330317, -6.2924920572826e-06, -1.18210740427445e-05, -2.26152336407658e-07, -4.72559855288443e-06, -1.13584395467146e-05}, 
                                                    {0, 0, 0, 0, 0, -1.43567333698478e-08, 0, 358.3999999951, 0, 0, 0, 9.90163540573677e-08, 0, -6741.33333332183, 0, -767.999999910683, -1.37799483964446e-08, 0, -1.57402919763836e-08, -768.000000008844, 0, -3.47121665191269e-07, 0, 46660.2666666682, 0, 409.599999445905, 6.0957263627096e-08, 0, 6.38605997348511e-08, 13619.1999997629, 0, 1.52476032920128e-07, 13619.1999999732, 819.200000106493, 409.600000122351, 0, 6.79594232096616e-07, 0, -154965.333333424, 0, 1.63622198663367e-06, -1.24438032905538e-07, 0, -1.42612126072353e-07, -85333.3333329389, 0, -6826.66666549879, -5.13419312752077e-07, 0, -5.46924396823957e-07, -85333.3333330883, -4.50791732337986e-07, -5.7695230952445e-07, -13653.3333332829, -6826.66666681351, -8.11855776604081e-07, 0, -7.52264364357268e-07, 0, 263236.266666873, 0, -2.50078428211078e-06, 1.33489040643371e-07, 0, 1.6661166636506e-07, 240298.666666368, 0, -2.46678052850706e-06, 8.85531646307247e-07, 0, 8.43788661688744e-07, 240298.666666062, 8.79125481470021e-07, 1.06356745663129e-06, 38229.3333316716, 0, 1.42746768694829e-06, 76458.6666658417, 38229.3333329084, 4.14920774178672e-07, 2.06650993053292e-06, 1.01292665630089e-06, 2.35992651855526e-06, 8.88079254534555e-07, 0, 4.40094054568932e-07, 0, -218453.333333519, 0, 1.91794710247655e-06, -7.50624767875745e-08, 0, -1.06233488885282e-07, -305834.666666679, 0, 2.42311833889933e-06, -7.02390284496555e-07, 0, -6.06731512039247e-07, -305834.666666116, -7.56471108541707e-07, -8.83871261235362e-07, -87381.3333321144, 0, 2.16382085379789e-06, -1.5925543918309e-06, 0, -1.5411195256732e-06, -174762.666665268, -87381.3333324421, -8.70932546376767e-07, -2.35972585490152e-06, -1.7367912625612e-06, -2.80167705074817e-06, 4.81256477224186e-07, -5.13150367542283e-07, -3.9278869516475e-06, 1.09108109275359e-06, -1.37384547241231e-06, -2.3927167465484e-06, 0, -1.05744692876074e-07, 0, 69905.0666667267, 0, -5.8294145717696e-07, 1.67964343452747e-08, 0, 2.47585854744788e-08, 139810.133333415, 0, -8.97754716487327e-07, 1.82598905328874e-07, 0, 1.53316311747662e-07, 139810.133333145, 2.58800716273961e-07, 2.893465578999e-07, 69905.0666663387, 0, -9.70053207639736e-07, 5.62610051840957e-07, 0, 5.5075662454625e-07, 139810.13333268, 69905.0666661754, 5.56032991590253e-07, 8.92317327451421e-07, 8.04043831845218e-07, 1.12930584246113e-06, -8.56218665741044e-07, 0, 8.2300389886959e-07, -6.77433860375875e-07, -2.18004978529274e-07, 4.42861347582055e-08, 1.95815885174725e-06, 7.14661870493842e-07, 1.99073362251542e-06, -1.17605186813637e-06, 8.66283687448294e-07, 1.94382913274481e-06, -8.67458458072011e-07, 3.73018114118109e-07, 1.49257178312672e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, -51.1999999980552, 0, 0, 0, -5.11587787412562e-08, 0, 904.533333329343, 0, 460.799999963932, 0, 0, 0, 51.2000000010759, 0, 1.83655474311193e-07, 0, -5631.99999999381, 0, -409.599999771633, -1.22882217689278e-08, 0, -1.35576072152622e-08, -8089.59999992919, 0, -1.44577409680504e-08, -853.333333310632, -409.600000014757, -1.93268078918353e-08, 0, -3.66956728486508e-07, 0, 15701.3333333354, 0, -6.93423862521946e-07, 2.42196407326545e-08, 0, 2.63150283135474e-08, 49834.6666665159, 0, 6826.66666635764, 7.6122159031374e-08, 0, 5.69954003312465e-08, 4778.66666652944, 1.20333994064184e-07, 6.54126481672801e-08, 6826.66666658927, -6.70203622437658e-08, 5.05375228037209e-08, 0, 4.14576990156952e-07, 0, -19660.8000000324, 0, 1.08855537359451e-06, -2.55951853740387e-08, 0, -1.54876386501463e-08, -136533.333333061, 0, 6.66116622258669e-07, -6.38811206057337e-08, 0, -4.03326293387708e-08, -10922.6666664052, -1.90730537630714e-07, -1.88908810929577e-07, -38229.3333329972, 0, -7.59380432544539e-08, -38229.3333329343, 4.92286098532427e-07, -1.11174950074665e-07, -4.90974753399536e-07, 0, -2.50623050395855e-07, 2.67614387967617e-07, 0, -2.47790638280084e-07, 0, 8738.13333337973, 0, -8.52176817719984e-07, 1.14223820187226e-08, 0, 2.07629898502225e-08, 166024.5333331, 0, -7.09288174849658e-07, 5.00510696588637e-08, 0, 1.33173133924968e-07, 8738.13333309416, 2.22232119809712e-07, 1.79225907455012e-07, 87381.3333329377, 0, -2.73923832335531e-07, 1.86186523538499e-07, 0, 2.44639540502297e-07, 87381.3333327183, -6.68574164803378e-07, 2.18546242727056e-07, 5.09378990645559e-07, 1.49932180475666e-07, 4.38274888709842e-07, -3.38021878927367e-07, -3.57182084655506e-07, 5.25953328650128e-07, -9.14427211246519e-07, -3.16205759436837e-07, -3.83845595211952e-07, 0, 6.08819637235564e-08, 0, -1.98290144951798e-08, 0, 2.62549979018815e-07, 0, 0, 0, -69905.0666665998, 0, 2.89894917984987e-07, -1.33980312515627e-08, 0, -4.97155525929438e-08, 7.56449526467223e-08, -5.85356658493356e-08, -5.48456554451478e-08, -69905.0666664712, 0, 9.58881672247767e-08, -7.07929996202551e-08, 0, -6.92112510034126e-08, -69905.0666663616, 3.01515690898088e-07, -1.50717702372464e-07, -2.94066043702681e-07, -1.24194303106064e-07, -1.75654315276291e-07, 1.37270190064491e-07, 0, -1.65750409324211e-07, 2.64777895538408e-07, 3.65722408803078e-07, -5.38354955427702e-08, -4.0519184513282e-07, 5.66598399683701e-08, -4.11161422241383e-07, 6.61738498627274e-07, 0, -2.59882384083983e-08, 5.52374101741511e-07, 5.86447977388528e-07, -3.38260749988592e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, -42.6666666670207, 0, 0, 0, 1.54976438316415e-08, 0, 668.444444445137, 0, 554.666666673143, 0, 0, 0, 42.6666666678611, 0, -6.62079379330946e-08, 0, -3356.44444445061, 0, -1877.33333340394, 0, 0, 0, -8647.11111109021, 0, -1.47299874098033e-08, -625.777777771473, -512.000000024383, 0, 0, 1.50947753799907e-07, 0, 6371.55555560186, 0, 1365.33333359994, 1.14172258374716e-08, 0, 3.13288201471247e-08, 43007.9999997731, 0, 28899.5555556764, 0, 0, 4.93056246318198e-08, 2730.66666666191, 1365.33333337526, 1.45069431128576e-07, 7509.33333336733, -1.07643853274275e-07, 1.79349552469811e-07, 0, -1.8832024919413e-07, 0, -3640.88888900828, 0, -4.47338224549705e-07, -2.40385688222404e-08, 0, -4.86677970744417e-08, -80099.5555550639, 0, -20024.888889516, -5.12563285345515e-08, 0, -1.18467198565534e-07, -3640.88888887238, -1.97483716510509e-07, -2.0858156591578e-07, -140174.222221947, 0, -1.13943266186671e-07, -32767.9999999577, 2.70073690927107e-07, -20024.8888889904, -1.37135632464681e-07, 2.72223391849674e-08, -4.62154140004242e-07, 9.23118586352291e-08, 0, 1.21316915536995e-07, 0, 1.23622544183684e-07, 0, 3.52606347483085e-07, 0, 0, 2.15735366532864e-08, 43690.6666662731, 0, 8.27102713724922e-07, 5.67459849527392e-08, 0, -1.39951140859429e-08, 1.86399423560055e-08, 2.89326367336979e-08, 2.34729026083438e-07, 247580.444443691, 0, 87381.3333336118, 2.02541191403386e-08, 0, -1.39038686561808e-08, 43690.6666666664, -3.20282115299206e-07, 3.43429607343706e-07, 2.49111434926583e-07, 1.86107013800021e-07, 5.9789479273647e-07, 87381.3333336645, -4.87136615235473e-07, 5.59024799147296e-07, -5.50548690238951e-07, 4.42923270686319e-07, 6.10012746433889e-07, 0, -3.15480788156965e-08, 0, -4.52254408277214e-08, 0, -1.06558240622099e-07, 0, 0, 0, 1.02558713047057e-07, 0, -3.39147244841532e-07, -4.98844172460287e-08, 0, 1.6147047493343e-08, 0, -1.24214653424041e-08, -8.88712018134967e-08, -116508.444444049, 0, -3.55286839392946e-07, 6.06298607860106e-08, 0, -1.52981793348436e-08, -1.92096414875231e-08, 1.23077142632315e-07, -1.38381145225302e-07, -4.29593441945083e-08, -8.49500153095741e-08, -2.67858909391362e-07, -116508.444444303, 0, 9.35528696619699e-08, -116508.444444594, 2.97660406912479e-07, -1.53404108149959e-07, -1.28193351970591e-07, 1.62170527024841e-07, -3.94153049504386e-07, 4.71635833241767e-07, -3.36487404548575e-07, -6.32663368868918e-07, 4.1989856982236e-07, 2.3559197761483e-07, 9.70760396325514e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, -42.6666666659825, 0, 0, 0, 4.7628863402514e-08, 0, 554.666666660354, 0, 668.444444430432, 0, 0, 0, 42.6666666652955, 0, -1.8419049111529e-07, 0, -1877.33333332846, 0, -3356.44444438728, 2.25756992168157e-08, 0, 1.76997808250515e-08, -8647.11111093759, 0, 5.14606093208413e-08, -512.000000007608, -625.777777730264, -1.84628959958603e-08, 0, 3.90956053750399e-07, 0, 1365.33333338292, 0, 6371.55555544434, -7.34695330184198e-08, 0, -3.37135274836596e-08, 28899.5555549748, 0, 43007.9999993407, -1.61903756682679e-07, 0, -1.50109098902993e-07, 1365.33333333726, 2730.66666647606, 4.37707012361209e-08, 7509.33333331558, 7.7965006092835e-08, -1.39118546440062e-07, 0, -4.63449813072039e-07, 0, -1.44391602215433e-07, 0, -3640.88888877471, 9.95940674493091e-08, 0, 1.38970872655399e-08, -20024.8888879727, 0, -80099.5555544481, 3.09283383417129e-07, 0, 1.73312474440815e-07, -3.27523512627631e-08, -3640.88888833025, -2.60276269347808e-07, -140174.222220523, 0, 4.21763350318821e-07, -20024.8888890579, 4.22833775488344e-08, -32767.9999999349, 6.69852227360936e-07, -2.48016145052327e-07, -1.27140325628335e-07, 1.58155443124735e-07, 0, 2.87775912315066e-07, 0, 1.49555782907189e-07, 0, -5.75686890247639e-08, -3.73411064588477e-08, 0, 0, -6.65335131462883e-07, 0, 43690.666665731, -2.83444272584206e-07, 0, -6.46462065298445e-08, 2.37276820131366e-08, -3.45079845580138e-07, 1.8156775946261e-07, 87381.3333314524, 0, 247580.444442716, -4.98841624957238e-07, 0, -3.55445997514557e-07, 1.26287088350736e-07, -1.11412845729752e-07, 43690.6666663527, -1.08709729310284e-06, 3.65110681370058e-07, 2.56831432597127e-07, 87381.3333334694, 1.14501954688e-07, -4.1433539395776e-07, -3.36042661395247e-07, -1.21315200562444e-06, -1.08041434462988e-07, 0, -7.28627825407573e-08, 0, -5.37764451313685e-08, 0, 0, 0, 0, 0, 1.71407512505878e-07, 0, 3.15978887763738e-07, 1.43044594308425e-07, 0, 0, -1.23355932391884e-08, 1.13828011770054e-07, -6.17815101630117e-08, 7.61314467905438e-07, 0, -116508.444443794, 7.75558788543318e-08, 0, 1.21105789153674e-07, -2.62706562863602e-08, 1.08524164787859e-08, 1.3967201095732e-07, 5.13431850037627e-07, -2.58885233806302e-07, -1.94724047004882e-07, -116508.444443497, 0, 1.62210424344605e-07, -1.05622056857889e-07, 6.01153815366942e-08, -116508.444444619, 3.94106950425464e-07, -1.96590182844251e-07, 2.46714283772993e-07, -3.51156327844945e-08, 9.72755156730786e-07, 1.17680430107847e-07, -1.46794547391801e-07, 3.25716471217322e-08, 3.21500116525393e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, -51.1999999972741, 0, 0, 0, 1.42727446019173e-08, 0, 460.79999998954, 0, 904.533333280887, -1.02575870231503e-08, 0, 0, 51.1999999961725, 0, -6.05305894053024e-08, 0, -409.600000000731, 0, -5631.99999967621, 4.98152187506476e-08, 0, 1.5817594748628e-08, -8089.59999975279, 0, 7.6190750557381e-08, -409.599999994473, -853.333333273425, 2.97889829082546e-08, 0, 1.37759728242121e-07, 0, 6.38729096272548e-08, 0, 15701.3333324169, -1.03406413259643e-07, 0, -3.42731276004593e-08, 6826.66666624509, 0, 49834.6666653032, -3.5360949604413e-07, 0, -2.2224486668209e-07, 3.73653610068769e-08, 4778.66666633717, -2.99214555050035e-08, 6826.66666652312, -1.58790237524385e-08, -3.29797333069363e-07, 0, -1.72444350233064e-07, 0, -1.25782790723229e-07, 0, -19660.7999986763, 1.1537420006429e-07, 0, 6.15087896649967e-08, 2.59950025554703e-07, 0, -136533.33333032, 6.84324610653954e-07, 0, 3.73740288118606e-07, -7.15570541703204e-08, -10922.6666660735, 1.21386476662579e-07, -38229.3333312295, 0, 7.58695545676566e-07, 1.40698169842995e-07, -1.17346474695768e-07, -38229.3333327489, 1.14654375399108e-06, -6.23289411752397e-08, 6.88396401237803e-07, 3.85470928666924e-07, 0, 1.11880272875276e-07, 0, 9.81154574562084e-08, 0, 8738.13333238697, -9.2141421385072e-08, 0, -3.63172391334619e-08, 2.53780347653013e-08, 0, 166024.533330389, -5.54263802016864e-07, 0, -2.74151197372726e-07, 7.4981396265498e-08, 8738.13333263455, -6.00922835080598e-08, -1.39963497600599e-06, 0, 87381.3333300524, -9.20870404186325e-07, 0, -6.42058412985715e-07, -1.18567448289663e-08, 1.83117157346531e-07, 87381.3333325295, -1.65246292522262e-06, 1.33317550204267e-07, -5.09697220599054e-07, -2.5946967981928e-07, 2.01910295863807e-07, -1.5411479214128e-06, 1.7878422745766e-07, -1.01721104495557e-06, -9.38046685335114e-08, 0, -2.93514496756463e-08, 0, -2.80758760974863e-08, 0, 2.65513606530834e-07, 2.2065157375959e-08, 0, 0, -5.63003516691061e-08, 0, -69905.0666656014, 1.31335868440615e-07, 0, 9.0710929184214e-08, -2.61326733989102e-08, 2.61665381865863e-07, 2.45129953413513e-08, 3.10777691424627e-07, 0, -69905.066665047, 5.23859154690861e-07, 0, 2.18067128123403e-07, -1.92707270783871e-08, -5.37684965154031e-08, -69905.0666662388, 7.67114116639028e-07, 0, 2.48970195275291e-07, 1.18116841475501e-06, 0, 4.19292316013998e-07, 0, -1.06842910326202e-07, 2.66900472895435e-07, 1.08843016256242e-06, -6.87535608147343e-08, 3.70702493109914e-07, 3.06956275681997e-08, 8.2576788446001e-07, -2.63586853225091e-07, 6.77575292186793e-07, -1.52528627244329e-07, 5.49058901939115e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 358.399999994902, 0, 0, 0, 3.53391658081351e-08, 0, -767.99999997338, 0, -6741.33333324386, 1.67617842532938e-08, 0, 0, -767.99999998939, 0, -1.11159735206889e-07, 0, 409.599999967228, 0, 46660.2666661601, -8.75003622352561e-08, 0, -2.72263065131561e-08, 13619.1999994938, 0, -1.29904025572961e-07, 819.199999985207, 13619.1999998576, 409.599999928947, 0, 1.97538642185212e-07, 0, -7.75753553223401e-08, 0, -154965.333332013, 2.13936614862971e-07, 0, 4.5992702689835e-08, -6826.6666655674, 0, -85333.3333308212, 6.09223954236633e-07, 0, 3.72737118666941e-07, -5.88277534860832e-08, -85333.3333326854, 1.58629016668983e-07, -13653.3333329901, 1.33126264086791e-07, -6826.66666592872, 0, -2.01290261881579e-07, 0, 2.2469519402266e-07, 0, 263236.266664879, -2.57588542928941e-07, 0, -1.60228916512837e-08, -8.17134291790219e-07, 0, 240298.666661738, -1.1092481247653e-06, 0, -4.70906332173666e-07, 1.54347567722793e-07, 240298.666665489, -2.68579581418418e-07, 38229.3333283458, 0, -1.22728741008447e-06, -1.57838901822549e-07, -1.34662470688524e-07, 76458.666665506, 38229.3333306381, -2.09837966526859e-07, -1.5015548965593e-06, -1.29615902184321e-06, 0, 1.09710741193999e-07, 0, -1.88765278929357e-07, 0, -218453.333332102, 1.5427817084103e-07, 0, 0, -4.88394775740008e-08, 0, -305834.666662383, 8.92812800738422e-07, 0, 3.81257167068075e-07, -1.70817575437078e-07, -305834.666665593, 2.1954160194349e-07, 3.89104639768936e-06, 0, -87381.3333263726, 1.64850779554858e-06, 0, 1.23396164887113e-06, -1.36684521622592e-07, -4.28695083038444e-08, -174762.6666652, -87381.3333299034, 1.86609552999238e-07, 1.5586295974167e-06, 7.2954154693366e-07, 2.29590500818301e-07, 3.27236366894005e-06, 7.96537011640438e-07, 3.00159231561003e-06, 1.42300137288439e-06, 0, -2.48182121995759e-08, 0, 5.2609028865095e-08, 0, 69905.0666663257, -3.681118033349e-08, 0, 0, 1.89138125289968e-07, 0, 139810.133331942, -3.04178346476373e-07, 0, -1.23151021454848e-07, 6.54054329883759e-08, 139810.133332992, -6.37814094663071e-08, -9.44056771761283e-07, 0, 69905.0666635921, -7.28997982958022e-07, 0, -3.73981419388112e-07, 1.67053838422146e-07, 2.49491655884285e-08, 139810.133332704, 69905.0666651814, -8.30506863026893e-08, -6.29882750494378e-07, -3.04657375866217e-06, 0, -7.69157068218162e-07, -1.46271564681522e-07, -9.5905644189157e-08, -5.73997059997166e-07, -2.15682597431218e-06, -8.85595655154171e-08, -1.56183493799705e-06, -2.29411654458321e-07, -1.93914557667635e-06, -7.88569138316938e-07, -1.30244312100924e-06, -4.89595010742684e-07, -2.20487575011004e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, -895.999999984966, 0, 0, 0, 1.53539096925604e-08, 0, 3114.6666665945, 0, 16255.999999725, -1.11641976551449e-07, 0, -5.83147479005564e-08, 3114.66666661654, 0, -1.12766789601078e-07, 0, -3583.99999994544, 0, -106609.777776183, 5.78074556010136e-07, 0, 1.65404477112674e-07, -52849.7777762543, 0, 8.90172418558095e-07, -7167.99999991905, -52849.7777770072, -3583.9999997476, 0, 3.21370290350893e-07, 0, 1365.33333365378, 0, 329727.99999571, -1.39083763829003e-06, 0, -2.41533015541506e-07, 56661.3333301035, 0, 308565.333325677, -4.01525129975663e-06, 0, -2.41319412275892e-06, 4096.00000008511, 308565.333329847, 1365.33333302323, 113322.666664852, 4095.99999994918, 56661.3333293494, 0, -4.52786624283038e-07, 0, -7.89055984962628e-07, 0, -515185.777771795, 1.67486947260768e-06, 0, 1.50481426378947e-07, -20024.8888861952, 0, -784611.555539988, 7.67391523468444e-06, 0, 3.05812258376388e-06, -3.09320120595792e-07, -784611.555549019, 1.5625003330651e-07, -289450.666652437, 0, 8.38597042116018e-06, -60074.6666648683, -4.8550869927651e-07, -578901.333327112, -289450.66665199, -1.29760481311276e-06, -20024.888881831, -60074.6666608561, 0, 3.1528875836959e-07, 0, 6.70872814789917e-07, 0, 393215.9999958, -1.0132098912433e-06, 0, -2.58584676065161e-08, -4.17553339052961e-07, 0, 873813.333319104, -6.47497868127294e-06, 0, -2.00003052939585e-06, 3.5366877220083e-07, 873813.333327578, 2.29242006824146e-08, 87381.3333222747, 0, 567978.666646707, -1.10380049750342e-05, 0, -7.27053121779095e-06, -9.19461527767688e-07, 7.32225033658356e-07, 1135957.33332526, 567978.66664713, 2.30345600208133e-06, -5.72071765589074e-06, 262143.999995407, 2.24189268302924e-06, 87381.3333148559, -3.29701025752284e-06, 262143.999983817, -4.23214825336264e-06, 0, -8.6570393757353e-08, 0, -1.99399173829995e-07, 0, -116508.444443272, 2.35932294517714e-07, 0, 0, -2.99853313569733e-07, 0, -349525.333328452, 2.05591638536358e-06, 0, 5.61121516583659e-07, -1.37557264567137e-07, -349525.33333148, -2.15800214250032e-08, 2.88718904144376e-06, 0, -349525.33332426, 4.97764640438967e-06, 0, 2.23338079459484e-06, 1.16984195994505e-07, -2.54167877815265e-07, -699050.66666314, -349525.333324766, -1.04615835500975e-06, 1.93022769138827e-06, -116508.444436082, 0, 4.8051653074181e-06, 1.35950029525174e-06, -8.85562447035633e-07, -349525.333330282, -116508.444431567, -1.78162594541457e-06, 7.62994993869652e-06, 1.1450923176723e-06, -349525.333322126, 7.08903709981969e-07, 6.40883373430721e-06, 3.21404624449021e-07, 9.57306410975902e-06}, 
                                                    {0, 0, 0, 0, 0, 2.12878093699702e-08, 0, 1493.3333333103, -1.40033522983023e-08, 0, 0, -1.90706598641395e-07, 0, -7580.44444433558, 0, -25500.4444440145, 2.51060799488671e-07, 0, 1.02186594806046e-07, -7580.44444437057, 0, 8.08769215986007e-07, 0, 14279.1111110739, 0, 153031.111108656, -1.35268080345435e-06, 0, -1.6371470002567e-07, 119523.555553048, 0, -1.9447773227604e-06, 28558.2222221948, 119523.55555401, 14279.1111110861, 0, -1.82936197359588e-06, 0, -11832.8888895835, 0, -425756.444437889, 3.27992298705514e-06, 0, -9.27041619446431e-08, -206847.999994395, 0, -620771.555543325, 9.43974466131532e-06, 0, 4.55668395822522e-06, -35498.6666669456, -620771.555548184, -11832.8888907634, -413695.999996881, -35498.6666685589, -206847.999992791, 0, 2.26967885186536e-06, 0, 3640.88889051619, 0, 602567.111101996, -3.93491222824937e-06, 0, 5.29965376956486e-07, 156558.22221673, 0, 1354410.66664207, -1.86976254782942e-05, 0, -4.41988153096121e-06, 14563.5555560283, 1354410.66665256, 3640.88889391876, 904760.888865943, 0, -1.86187925195712e-05, 469674.666663478, 21845.3333372625, 1809521.77776623, 904760.888856919, 14563.5555659336, 156558.222214144, 469674.666659043, 0, -1.45817109090121e-06, 0, -1.43107655294821e-06, 0, -422343.111104736, 2.31636143460022e-06, 0, -5.12206457499949e-07, -43690.6666647905, 0, -1310719.9999774, 1.61353541662717e-05, 0, 1.88093991616733e-06, -3.95909347832816e-07, -1310719.99998781, -4.86556154536127e-06, -553415.111092288, 0, -1398101.33330242, 2.63848801782319e-05, 0, 1.41778758399555e-05, -174762.666664319, -3.05901657311768e-06, -2796202.66665145, -1398101.33328835, -1.54398204677589e-05, -43690.6666659579, -1660245.33332579, -1.28657414424508e-05, -553415.11107428, -262143.999996762, -1660245.33330165, -174762.666671197, 0, 3.78951927726463e-07, 0, 4.50189472557248e-07, 0, 116508.444442679, -5.24766529641957e-07, 0, 1.65328513024857e-07, 8.32585940874154e-08, 0, 466033.777769918, -5.10818365954119e-06, 0, -2.98878678380182e-07, 1.40018744547877e-07, 466033.777773919, 1.62404943091212e-06, 116508.44443896, 0, 699050.666652812, -1.2193678823672e-05, 0, -3.63007390140266e-06, -8.08780058090449e-07, 8.28387253058347e-07, 1398101.33332666, 699050.666646517, 7.00574590073846e-06, 1.65274436487205e-06, 466033.777764638, 0, -1.07578685064805e-05, 466033.777775541, 4.70855503842611e-06, 1398101.3333286, 466033.777748363, 9.55826648380205e-06, 116508.44443231, -1.6360990619315e-06, 1398101.33330936, 9.12442071354031e-06, 699050.666655633, 6.08535837474964e-06, 466033.777762436}, 
                                                    {0, 0, 0, 0, 0, -2.88689973598831e-08, 0, -1791.99999997893, 0, 0, 0, 2.51177125893362e-07, 0, 12680.533333241, 0, 27016.5333329345, -2.2264938591315e-07, 0, -1.73501375810934e-08, 12680.5333333542, 0, -1.04996765627108e-06, 0, -35328.0000000186, 0, -136772.266664442, 1.32163715559125e-06, 0, -3.59853327402574e-07, -172100.266664331, 0, 1.53478333143191e-06, -70656.0000004015, -172100.266665691, -35328.0000011375, 0, 2.35661425722697e-06, 0, 48469.3333341879, 0, 331093.333327459, -3.28483157851056e-06, 0, 1.53550239210578e-06, 428031.999994389, 0, 710655.999989194, -9.21311380186059e-06, 0, -1.96985725642202e-06, 145408.000000965, 710655.99999409, 48469.3333406198, 856063.999999607, 145408.000005961, 428031.999997455, 0, -2.91009442838561e-06, 0, -32768.0000019697, 0, -420522.666658507, 3.90597434094766e-06, 0, -2.47517143801246e-06, -518826.666660276, 0, -1294335.999979, 1.97187742577388e-05, 0, -1.59996625537047e-06, -131072.000000862, -1294335.999988, -32768.0000157785, -1359871.99997911, 0, 1.55435925515517e-05, -1556480.00000108, -196608.000009157, -2719743.99999561, -1359871.99997313, -131072.000025196, -518826.666676684, -1556480.00000538, 0, 1.86326462258042e-06, 0, 8738.1333351271, 0, 270882.133327608, -2.21650617140815e-06, 0, 1.8039595868094e-06, 305834.66666366, 0, 1092266.66664758, -1.75612010918203e-05, 0, 3.91419424143776e-06, 43690.6666670184, 1092266.66665628, 8738.13334763412, 1135957.33331469, 0, 1660245.33330719, -2.63232093219976e-05, 0, -7.62658970737952e-06, 1223338.66666682, 87381.3333387876, 3320490.66665977, 1660245.3332893, 43690.6667021181, 305834.666692143, 3407872.00000207, 87381.3333604921, 1135957.33331181, 1835008.00000786, 3407871.99998199, 1223338.66670279, 0, -4.82919518396306e-07, 0, -5.89705795355869e-07, 0, -69905.066665076, 4.73600314294363e-07, 0, -5.05333251677023e-07, -69905.0666663693, 0, -349525.333326661, 5.57819879406258e-06, 0, -1.68765869748335e-06, -6.78450066746677e-08, -349525.333330113, -4.6734859900773e-06, -349525.333327247, 0, -699050.666655492, 1.30526815116154e-05, 0, 6.43818587864462e-08, -349525.333332861, -1.0494529747571e-06, -1398101.33333014, -699050.66664598, -1.58671541801885e-05, -69905.066680469, -699050.666654978, 0, 9.16629904395761e-06, -1398101.33333493, -9.08175750573099e-06, -2097152.00000148, -699050.66663966, -1.97773601383406e-05, -349525.333334263, -699050.666667462, -2097151.99998099, -349525.333365175, -2097151.99999677, -699050.666687234, -1398101.33333309}, 
                                                    {0, 0, 0, 0, 0, 3.24818218342693e-08, 0, 1119.99999998449, 0, 0, 0, -2.79687992948382e-07, 0, -14645.3333332675, 0, -14645.3333330332, 1.25315904185057e-07, 0, 0, -5685.33333341918, 0, 1.16562522302205e-06, 0, 56192.0000000419, 0, 56191.9999984446, -7.6308131238791e-07, 0, 2.80832551220512e-07, 184063.999998101, 0, -8.16397615510718e-07, 66901.3333338442, 66901.3333332095, 10709.3333349278, 0, -2.61543274821054e-06, 0, -94549.3333341863, 0, -94549.3333294161, 1.92006083328811e-06, 0, -1.06337955774546e-06, -647509.333327794, 0, -647509.333325743, 5.12900265015833e-06, 0, 8.94858442578698e-07, -197973.33333415, -197973.333330837, -8874.66667492395, -759808.000001679, -112298.666673094, -112298.666669578, 0, 3.23110092156492e-06, 0, 73728.0000020826, 0, 73727.9999946696, -2.31719522462579e-06, 0, 1.65418226474231e-06, 980309.333325219, 0, 980309.333320216, -1.11648845336238e-05, 0, 1.12948190611134e-06, 223914.666667046, 223914.666660286, 2730.66668341351, 1813162.66664997, 0, -8.35523761363942e-06, 2042538.6666696, 229376.000008131, 2042538.66666892, 229375.999991346, 81920.000025272, 81920.0000202493, 1144149.33334774, 0, -2.06957699676677e-06, 0, -21845.3333353612, 0, -21845.3333296164, 1.3290268114407e-06, 0, -1.21127356364627e-06, -677205.333328009, 0, -677205.333322142, 1.0105010637407e-05, 0, -2.22777308304257e-06, -87381.3333332662, -87381.3333270835, -1.48130005895321e-05, -1922389.33331533, 0, -1922389.33331623, 1.4547014132156e-05, 0, 3.79729432875283e-06, -2053461.33333402, -131072.000003575, -2053461.33333417, -131071.999978428, -21845.3333680679, -21845.3333661624, -3932160.00000732, -87381.3333574763, -87381.3333370402, -2097152.00001269, -2097152.0000026, -742741.333380644, 0, 5.36463483309717e-07, 0, 7.07468326184739e-07, 0, -1.03019892888447e-06, -2.886495044978e-07, 0, 3.48502115592737e-07, 174762.666665515, 0, 174762.666662805, -3.22838688535284e-06, 0, 9.20155912870014e-07, -5.11046525302937e-08, -2.1492465652895e-06, 4.82255377805254e-06, 699050.666659615, 0, 699050.666660494, -7.31991920828706e-06, 0, 3.60821924883534e-08, 699050.666666077, 3.15492458440138e-07, 699050.666666406, -1.15684770610333e-05, 1.54570705477303e-05, 1.55276448828962e-05, 1048575.99999097, 0, -4.88856543713299e-06, 2097152.00000311, 6.99899313641023e-06, 2097152.00000493, -1.00716188192843e-05, 1.76333032145218e-05, 1.10737063379338e-05, 1048576.00000159, 1048575.99999347, 174762.666702669, 2097152.00000407, 699050.66668974, 699050.666681373}, 
                                                    {0, 0, 0, 0, 0, -1.60730089490829e-08, 0, 223.999999995585, 0, 0, 0, 1.10060882486111e-07, 0, -3765.33333332516, 0, -2271.99999991786, -1.56175886083976e-08, 0, -1.9652108097134e-08, -480.000000019366, 0, -3.78755772332194e-07, 0, 21631.9999999974, 0, 4095.99999952577, 7.57999395480476e-08, 0, 1.10605859975699e-07, 37674.6666664656, 0, 1.50524300883959e-07, 7551.99999998245, 4352.00000019875, 256.000000219462, 0, 7.28069399049483e-07, 0, -53589.3333334199, 0, -2047.99999863771, -1.90407906210819e-07, 0, -2.67037738368539e-07, -211285.333332622, 0, -64170.6666661368, -4.88313672214728e-07, 0, -5.55409106638028e-07, -38229.3333330199, -4096.00000080643, -1.03545971770921e-06, -67925.333333569, -3754.66666682688, -2048.0000012582, 0, -7.9533314298156e-07, 0, 57344.000000285, 0, -2.1008393652126e-06, 2.48195061727933e-07, 0, 3.14286102462764e-07, 502442.666665218, 0, 30037.3333327757, 7.25694870017076e-07, 0, 8.75729387430359e-07, 73727.9999992707, 1.54088954190209e-06, 2.02207727070552e-06, 322218.666665508, 0, 1.10200864445283e-06, 338602.666666, 16383.99999886, 60074.6666682229, 3.40255750297336e-06, 1.61752376818073e-06, 3.76928661189534e-06, 30037.333334586, 0, 4.62044183164187e-07, 0, -21845.33333366, 0, 1.64753141791824e-06, -1.39846689109489e-07, 0, -2.23704113176625e-07, -502442.666665461, 0, 3.63047232457588e-07, -6.07855555169333e-07, 0, -8.37949051123422e-07, -43690.6666660081, -1.2562600655255e-06, -1.88416533270261e-06, -611669.333331216, 0, -131071.999999973, -1.28866201826216e-06, 0, -1.47317463358466e-06, -633514.666665232, -21845.3333312514, -2.424532938398e-06, -3.63213510572221e-06, -2.80383343025332e-06, -4.94770002384716e-06, -262144.000000634, 1.53073361862293e-07, -5.40524141587431e-06, -131071.999998482, -2.89904545052959e-06, -4.24590932504879e-06, 0, -1.10895585361856e-07, 0, 1.26673669957312e-07, 0, -5.12986383711135e-07, 3.76903624131968e-08, 0, 6.45570991661805e-08, 174762.666666338, 0, -1.57572171555102e-07, 2.3808769442367e-07, 0, 2.72863787565437e-07, -2.09954316589122e-07, 3.75714176191507e-07, 6.28965422913869e-07, 349525.333332223, 0, 4.38733570902285e-07, 3.88316617854329e-07, 0, 5.47881683588163e-07, 349525.333332563, -1.01365629944708e-06, 1.18111455309737e-06, 1.5787774406028e-06, 1.38122868240919e-06, 2.09226314838117e-06, 174762.666666171, 0, 7.06604639731256e-07, 349525.333333162, -9.88594933628498e-07, 1.05042018108583e-06, 2.7146850201103e-06, 6.45625790186536e-07, 3.48587618967674e-06, 174762.666664567, 2.06760244927066e-06, 3.69719430614943e-06, -7.74357102798855e-07, 1.37501850913798e-07, 3.53149671957721e-06}, 
                                                    {0, 0, 0, 0, 0, 1.43613155750101e-08, 0, 223.999999991382, 0, 0, 0, -1.16352327053505e-07, 0, -2271.99999996128, 0, -3765.33333316694, 7.23696855306436e-08, 0, 4.43387421852999e-08, -479.999999976095, 0, 4.63637753075122e-07, 0, 4095.99999999416, 0, 21631.9999990181, -3.6572870011582e-07, 0, -1.3911015509966e-07, 37674.6666657224, 0, -5.73605743741405e-07, 4351.99999998362, 7551.99999955632, 255.999999930835, 0, -1.00821144345741e-06, 0, -2048.00000025148, 0, -53589.3333306277, 8.76722570750569e-07, 0, 2.6118153388516e-07, -64170.6666646329, 0, -211285.333328589, 2.43795634331874e-06, 0, 1.61501949856341e-06, -4096.00000009087, -38229.3333311557, -8.03033956960692e-08, -67925.3333324576, -2048.0000003701, -3754.66666469525, 0, 1.21916850097584e-06, 0, 5.67622710934135e-07, 0, 57343.9999961562, -1.0754181738994e-06, 0, -3.00787976493043e-07, 30037.3333312417, 0, 502442.666656713, -4.69523003046644e-06, 0, -2.31282240006848e-06, 1.99723475950044e-07, 73727.9999955444, 3.38764943157201e-07, 322218.666658352, 0, -5.25477426670822e-06, 60074.666665849, 9.84394288074463e-07, 338602.666663182, 16383.9999922296, 1.81551747465316e-06, -2.87962868272182e-06, 30037.3333315775, 0, -7.69738519916664e-07, 0, -4.9972045317966e-07, 0, -21845.3333306163, 6.84655647553015e-07, 0, 1.53511569080617e-07, 9.27088885820241e-07, 0, -502442.666657144, 4.00632546447655e-06, 0, 1.54591546625246e-06, -1.86920506510543e-07, -43690.6666624913, -3.59417866697517e-07, -131071.999993354, 0, -611669.33332179, 6.53332157855697e-06, 0, 4.43586927879571e-06, 4.98501312589763e-07, -9.30157034379933e-07, -633514.666661488, -21845.3333219651, -2.52950165100384e-06, 1.7173327215849e-06, -262143.999997835, -2.61335871644223e-06, 9.04623174346434e-06, -2.09184553690277e-07, -131071.999992223, -5.99227833576582e-07, 0, 1.97619288426005e-07, 0, 1.59837907225333e-07, 0, -7.53411555334485e-07, -1.64553491846607e-07, 0, -2.8745337884706e-08, -9.95358630020311e-08, 0, 174762.666663246, -1.23613861395884e-06, 0, -4.51454807379964e-07, 6.76995123177462e-08, -1.45582366552687e-06, 1.0221055159173e-07, -1.99227592580456e-06, 0, 349525.33332794, -2.98371288761623e-06, 0, -1.48186578925249e-06, -1.8290233901159e-07, 3.19434975068924e-07, 349525.333330806, -5.27857429821416e-06, 1.05949497737104e-06, -5.41606689606816e-07, 174762.666662034, 0, -2.78896653160031e-06, -5.09451037158256e-07, 1.00969527793846e-06, 349525.333331769, -6.82903965316572e-06, 1.6911561416874e-06, -3.00659079536857e-06, 8.68847403379569e-08, 174762.666660264, 1.61765851749468e-06, -2.51976149833958e-06, 1.62823582327281e-06, -3.02584739989073e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, -597.333333326812, 0, 0, 0, 0, 0, 9244.44444442407, 0, 6855.1111109865, 2.45324375578354e-08, 0, 5.22097137711845e-08, 2076.44444451731, 0, -8.19701308572439e-08, 0, -46421.3333333505, 0, -19000.8888883017, -8.37574354206124e-08, 0, -2.97635064471255e-07, -103651.555554885, 0, -2.46656066681309e-07, -29696.0000001923, -21390.2222227232, -2389.33333424263, 0, 3.01080726124093e-07, 0, 96028.4444448286, 0, 20024.8888874734, 1.98114442909919e-07, 0, 8.04229794018989e-07, 497891.555552903, 0, 268970.666664933, 3.76440668519593e-07, 0, 1.21918121231468e-06, 126520.888888824, 40960.0000015291, 910.222226508194, 300373.33333493, 31402.6666688976, 21845.3333373043, 0, -4.70791392341154e-07, 0, -87381.3333343942, 0, -7281.7777758064, -2.98998649908573e-07, 0, -1.10781815693823e-06, -953912.888883971, 0, -262143.999998608, -2.04218183951391e-07, 0, -2.59320312557568e-06, -185685.333332658, -21845.3333358074, -8.56065590061556e-06, -1121393.77777219, 0, -1.54379321516084e-06, -1230620.44444572, -109226.66666731, -535210.666671969, -21845.333339413, -10922.6666765076, -7281.77779186396, -283989.33334107, 0, 3.45750663866704e-07, 0, 29127.111112233, 0, -1.47379420949612e-06, 2.41246254535398e-07, 0, 7.64684263062326e-07, 786431.999996182, 0, 87381.3333330435, 3.67388314481866e-08, 0, 2.39813328099161e-06, 87381.3333326182, 1.98943504722192e-06, 7.65669150435499e-06, 1572863.99999155, 0, 902940.444441787, 5.51151614351901e-07, 0, 2.62667491147239e-06, 1660245.33333281, 87381.3333312016, 262144.000006983, 4.15973202745504e-06, 1.4390074237853e-05, 1.92100838825089e-05, 1835008.00000573, 29127.1111173059, 1.40687178460058e-05, 961194.666668847, 262144.000009019, 87381.3333553994, 0, -9.77390025741186e-08, 0, -4.14256123257883e-07, 0, 4.48410467750027e-07, -7.60780664393903e-08, 0, -2.10495800828622e-07, -233016.888887904, 0, 0, -3.08270984881457e-08, 0, -8.14823239677797e-07, 2.25479054015086e-07, -6.21034083865017e-07, -2.54334877392004e-06, -699050.666662678, 0, -233016.888889158, 6.58284361458805e-08, 0, -1.31226266338421e-06, -699050.666665945, 1.3727955676364e-06, -3.1202636209458e-06, -1.18415132490763e-06, -6.68911000063429e-06, -8.61772997526303e-06, -699050.666663793, 0, -6.95236737798452e-07, -1398101.33333511, -2.48454556717722e-07, -699050.666671276, -4.283697346189e-06, -5.79712413064563e-06, -1.00647596683899e-05, -699050.666664564, -3.87138371019106e-06, -1.64942584930855e-05, -699050.666668864, -233016.888896217, -1.10294642590078e-05}, 
                                                    {0, 0, 0, 0, 0, 1.74159885289331e-08, 0, 199.111111110466, 0, 0, 0, -1.39691857312319e-07, 0, -2815.99999998879, 0, -2815.99999998517, 1.40053191682575e-08, 0, 0, -426.666666673288, 0, 5.52881618805684e-07, 0, 11719.1111111191, 0, 11719.1111111476, -8.34701275002443e-08, 0, 0, 39367.1111107066, 0, -1.22782628830583e-07, 5575.11111115254, 5575.11111106136, 227.555555725536, 0, -1.19487863984166e-06, 0, -16384.0000002189, 0, -16384.0000003578, 2.06797454542073e-07, 0, -5.39398011651309e-08, -159743.999998329, 0, -159743.999998884, 6.4419519281772e-07, 0, 2.78012947973379e-07, -19114.666666727, -19114.6666662484, -9.86174366252105e-07, -72362.6666668112, -2730.66666712216, -2730.6666668989, 0, 1.43587497618872e-06, 0, 7281.77777835966, 0, 7281.77777853356, -2.42132019223612e-07, 0, 1.63743755223755e-07, 211171.555552666, 0, 211171.555554586, -1.14515424062073e-06, 0, -1.01415142823614e-07, 14563.5555556267, 14563.5555545608, 2.19235706490871e-06, 611669.333329239, 0, -1.16381018725752e-06, 243939.555556002, 7281.77777790125, 243939.555555756, 7281.77777587557, 2.41447041946683e-06, 2.43900475146329e-06, 32768.000000442, 0, -9.00721474291739e-07, 0, -5.94652428789375e-07, 0, -6.67874241897802e-07, 9.99711422368136e-08, 0, -1.14327984759243e-07, -87381.333331186, 0, -87381.3333331538, 1.00076016224171e-06, 0, 9.21033126586906e-08, -8.80378969848087e-08, 8.5916188808221e-07, -1.95412170467081e-06, -699050.6666612, 0, -699050.666663686, 1.92547460961815e-06, 0, 1.14701999404743e-06, -174762.666666967, 4.01275213991837e-07, -174762.666666877, 3.30073084462556e-06, -3.821169909619e-06, -3.91188833855857e-06, -786432.000001339, -1.34424129308289e-06, 2.23435558941137e-08, -87381.3333324527, -87381.3333315856, -4.19824508933394e-06, 0, 2.29795247083406e-07, 0, 2.13895418344858e-07, 0, 2.20245323030137e-07, -2.14214626038985e-08, 0, 2.99084880029984e-08, -5.444663063097e-07, 0, 6.82041145747949e-08, -3.60226619381364e-07, 0, 0, 4.51544120788276e-08, -2.63641284521758e-07, 6.69958130590492e-07, 233016.888886496, 0, 233016.888888306, -7.29960179916612e-07, 0, -2.79666774516346e-07, 6.81063537383503e-08, -1.76937150601822e-07, 6.07805430889521e-08, -1.73779357899079e-06, 1.88275136009613e-06, 1.99903657490807e-06, 466033.777775525, 0, -8.83551358664599e-07, 466033.777778459, 0, 466033.777778741, -1.54689680046189e-06, 1.52401967883876e-06, 7.43356250897138e-07, -6.55728954825854e-07, -1.92475950180612e-06, 3.53206461520683e-06, 233016.888888546, 1.24281167849637e-06, -2.61323917476332e-07}, 
                                                    {0, 0, 0, 0, 0, -4.04782952493376e-08, 0, -597.333333320111, 1.01994554124336e-08, 0, 0, 3.2525989071792e-07, 0, 6855.1111110443, 0, 9244.44444418758, -1.69250285869383e-07, 0, -8.17292052728563e-08, 2076.44444441752, 0, -1.289194532342e-06, 0, -19000.8888888975, 0, -46421.3333319449, 8.98429325144538e-07, 0, 2.09008818932859e-07, -103651.555553784, 0, 1.30934346177285e-06, -21390.2222223061, -29695.9999990894, -2389.33333360919, 0, 2.79165292959338e-06, 0, 20024.8888895055, 0, 96028.4444408831, -2.19034999990524e-06, 0, -2.30533914106151e-07, 268970.666662096, 0, 497891.555547574, -6.01035980009342e-06, 0, -3.32054868228139e-06, 40960.0000003185, 126520.888884043, 910.222224553591, 300373.333331991, 21845.333335645, 31402.6666635741, 0, -3.36315401617992e-06, 0, -7281.77777923783, 0, -87381.3333285279, 2.68844559513952e-06, 0, 7.82695480169765e-08, -262143.999994206, 0, -953912.888873524, 1.1716777378835e-05, 0, 3.93010219425454e-06, -21845.3333337365, -185685.333323219, -5.23797212316215e-06, -1121393.77776176, 0, 1.20685025681024e-05, -535210.666665642, -21845.3333372635, -1230620.44443806, -109226.666648609, -7281.77778781449, -10922.6666664359, -283989.333333045, 0, 2.11584695565665e-06, 0, 1.36241665585796e-06, 0, 29127.1111078528, -1.62379503195564e-06, 0, 4.62991962235672e-08, 87381.3333299846, 0, 786431.999986035, -1.01764707801987e-05, 0, -2.37384163622982e-06, 2.96754631822022e-07, 87381.3333240526, 4.73589745004417e-06, 902940.444429511, 0, 1572863.99998048, -1.64184620414606e-05, 0, -9.74172987502017e-06, 262143.999999007, 2.71561373622193e-06, 1660245.33332367, 87381.3333053971, 1.40604166584507e-05, 5.21729913291355e-06, 1835007.99999728, 1.1164724108779e-05, 29127.1110939448, 262144.000002125, 961194.666651301, 87381.3333459442, 0, -5.41433662556148e-07, 0, -4.58437971439961e-07, 0, 8.6642804266643e-07, 3.84138471061814e-07, 0, -3.51570761474565e-08, 6.30264197174506e-07, 0, -233016.888883995, 3.23734677615778e-06, 0, 6.11495419445416e-07, -1.05785968908129e-07, 3.13298412813222e-06, -1.5446998164863e-06, -233016.888883685, 0, -699050.666658364, 7.44870351902393e-06, 0, 2.86637518981558e-06, 5.4401807886308e-07, -7.65593836708402e-07, -699050.666661924, 1.33300688821398e-05, -6.25255864351457e-06, -3.22974626889898e-06, -699050.66665781, 0, 6.80930475673894e-06, -699050.666666199, -3.82266549303628e-06, -1398101.33333139, 1.63070187274004e-05, -7.94459298459044e-06, 3.83221992363177e-06, -3.18635109053982e-07, -699050.666652231, -1.18107831624368e-05, -699050.666662764, -8.48133253968564e-06, -233016.888885184}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 1791.99999999829, 0, 0, 0, 0, -17459.1999999937, 0, -3.52055223697682e-08, -17459.1999999584, -5.41287600521876e-08, -17459.2000000947, 0, 0, 0, 0, 69142.7555555962, 0, -1.10547584249063e-08, 69142.7555553261, 1.63001845831619e-07, 0, 138285.511110776, 1.08171547431548e-06, 138285.51111186, 69142.7555560462, 138285.511112332, 0, 0, 0, 0, -142677.333333574, 0, 5.28914931513415e-07, -142677.333332839, 8.27274248860479e-08, 0, -428031.999999296, 0, -3.5575781889212e-06, -428031.999998104, -4.55207050097482e-06, -428032.000001773, -142677.333333314, -428032.000004761, -428032.000004232, -856064.000006788, -428032.000006914, 0, 0, 0, 0, 162019.55555597, 0, -1.23794332867243e-06, 162019.555555058, -7.72635724154323e-07, 0, 648078.222222135, 0, 4.09279021760437e-06, 648078.222218546, 6.83168111248852e-06, 648078.222223917, 162019.555553056, 648078.222229435, 0, 972117.333328949, 1.3696208134614e-05, 972117.33334213, 1944234.66667934, 648078.222230229, 648078.222230708, 1944234.666687, 972117.333358541, 1944234.66669207, 0, 0, 0, 0, -96119.4666669665, 0, 1.1126516032291e-06, -96119.4666664571, 8.8314762104655e-07, 0, -480597.333334024, 0, -1.60034870661323e-06, -480597.333330446, -3.96078245515659e-06, -480597.333333914, -96119.4666630946, -480597.333337871, 0, -961194.666663471, 0, -1.35198038974558e-05, -961194.666659873, -1.65042331713183e-05, -961194.666673306, -1922389.33334289, -480597.333339528, -480597.333333038, -1922389.3333559, -961194.666697283, -961194.666681494, -2883584.00002632, -961194.666696885, -2883584.00003076, -1922389.33335756, -2883584.00005589, 0, 0, 0, 0, 23301.6888889676, 0, -3.54009793068066e-07, 23301.6888888604, -3.08943742755775e-07, 0, 139810.133333715, 0, 4.35390868203578e-08, 139810.133332551, 6.18355515567603e-07, 139810.133333337, 23301.6888874055, 139810.133334263, 0, 349525.333332837, 0, 4.20165328390456e-06, 349525.333330279, 6.12551105965296e-06, 349525.333334903, 699050.666669165, 139810.133334949, 139810.13333035, 699050.666674875, 349525.333344864, 0, 466033.777774639, 9.06230561312723e-06, 466033.777784566, 1398101.33334405, 349525.333341262, 349525.333342145, 1398101.33334835, 466033.777799675, 1398101.33334456, 699050.666671988, 1398101.33336759, 1398101.33335044, 2097152.00003295, 1398101.33336329}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -85.3333333333162, 0, 0, 0, 0, 1644.08888888935, 0, 0, 85.333333333538, 0, 85.3333333335191, 0, 0, 0, 0, -11798.7555555544, 0, 0, 0, 0, 0, -1558.75555557193, -1.51050318624379e-08, -1558.75555555472, 0, 0, 0, 0, 0, 0, 41187.5555555434, 0, 1.22137894291481e-08, -1.20964793162735e-08, 0, 0, 10240.0000000591, 0, 4.13715602264874e-08, 6.60628174172692e-08, 5.2348587606313e-08, 10239.9999999975, 2.22948271728488e-08, 4.59661247403553e-08, 0, 0, 7.07342019136853e-08, 0, 0, 0, 0, -74638.2222221944, 0, -1.47455041439513e-08, 3.69504934763203e-08, -1.29034402573145e-08, 0, -30947.5555556283, 0, -1.09796617076328e-07, -9.90135224780829e-08, -8.63463850477075e-08, -30947.5555555302, -6.54750815165046e-08, -5.94917485750539e-08, 0, -2.37305263682639e-07, -1.55554792634224e-07, 1.62088570960396e-08, -2.14290845708814e-08, 0, -1.12575109021655e-07, -5.73393622572675e-08, -1.65969152779831e-07, -1.71186975743756e-07, 0, 0, 0, 0, 66992.3555555312, 0, 0, -3.09420619930836e-08, 0, 0, 43690.6666666845, 0, 9.88993997256822e-08, 4.50125292737223e-08, 2.49798813669004e-08, 43690.66666665, 3.85414318245379e-08, 5.50036253907909e-08, 0, 2.21524997588882e-07, 0, 1.64350234731783e-07, 2.6831657043385e-07, 1.13127118870246e-07, -2.61712635304099e-08, 1.80427685209157e-08, 3.14567097699581e-08, 1.5479851584813e-07, 1.4322073881762e-07, 1.68183496834831e-07, 4.24989113175543e-08, 8.37345896154712e-08, 3.23680601771385e-07, 0, 1.6748179028992e-07, 2.31087869751726e-07, 0, 0, 0, 0, -23301.6888888812, 0, 0, 1.0870588700532e-08, 0, 0, -23301.6888888829, 0, -2.41693399364502e-08, 0, 0, -23301.6888888813, -2.00422134864399e-08, -2.09421401127175e-08, 0, -6.86180137241281e-08, 0, -5.12047133948722e-08, -9.15961051890636e-08, -3.99942315522597e-08, 0, 0, -3.07222916873846e-08, -2.99974865837662e-08, -4.96918900335467e-08, -6.21446754672781e-08, 0, -1.40644681921447e-07, -4.98096601061687e-08, 0, -2.47502241687688e-08, 0, -1.17912574512425e-07, -5.62105102970375e-08, -5.83017604326702e-08, 3.18153311849998e-08, -6.87613410430848e-08, -1.72781117881788e-07, 7.66355083980066e-08, -1.02184638441966e-07, 5.45591935947847e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -85.3333333330637, 0, 0, 0, 0, 85.3333333303529, 0, 0, 1644.08888888531, 0, 85.3333333331997, 0, 0, 0, 0, 1.18552487779277e-08, 0, 1.46069930731252e-08, -11798.755555538, 1.93390580695605e-08, 0, -1558.75555552007, 3.39285408920472e-08, 0, 0, -1558.75555555328, 0, 0, 0, 0, -2.40437222108708e-08, 0, -3.44642114060152e-08, 41187.5555555176, -4.75503611981893e-08, 0, -1.06494378796675e-07, 0, -1.08053390299232e-07, 10239.9999998502, -1.20947324843715e-07, 0, -2.36213545425748e-08, 10239.9999999546, 4.03205596065541e-08, 1.03129917054994e-08, -1.30329652995686e-08, 0, 0, 0, 0, 2.87782563214831e-08, 0, 4.36662031202637e-08, -74638.2222221767, 6.51713905292284e-08, 0, 1.39207814464948e-07, 0, 1.87001732948921e-07, -30947.5555552848, 2.02409896427843e-07, 0, 9.76971310129681e-08, -30947.5555554793, 0, 3.4324537817203e-07, 2.5887928398699e-07, -5.15261799343419e-08, -2.42837767393423e-08, -1.17927884441062e-08, 5.00707884879917e-08, 0, 5.24885918497906e-08, -1.7182008836222e-07, 0, 0, 0, 0, -1.90166040692672e-08, 0, -2.23978949739065e-08, 66992.3555555327, -4.7875566838753e-08, 0, -9.25488872060924e-08, 0, -1.76332236373233e-07, 43690.6666664401, -1.77063060400724e-07, 0, -5.28292187901317e-08, 43690.6666665945, 0, -3.26807126189704e-07, 0, -3.3999225966095e-07, -4.40660304132931e-07, -3.15396552297835e-07, 2.58711597698219e-08, -3.76766480874524e-08, -1.25243951179373e-08, -1.55107187791152e-07, -4.05365629029665e-08, -1.41941664989808e-07, 0, 0, -1.89310933021142e-07, 1.12172072692337e-07, 1.36433145000032e-07, 1.63321714321741e-07, 0, 0, 0, 0, 0, 0, 0, -23301.6888888843, 1.32562796452987e-08, 0, 2.44060329539739e-08, 0, 5.75643775916683e-08, -23301.688888818, 5.02575898974674e-08, 0, 1.62517923355272e-08, -23301.6888888604, 0, 1.04828452697979e-07, 0, 1.01993121454764e-07, 1.92759024124384e-07, 1.17846402208987e-07, 0, 0, 4.13987089140153e-08, 7.57050345557437e-08, 2.17597724640704e-08, 9.02213706328605e-08, 0, 2.12984027074779e-07, 1.5886656247064e-07, 0, 0, 1.10115670490422e-08, 1.10414411997743e-07, 1.55131724141406e-08, 9.22378454110885e-08, -9.76239260666281e-08, -3.98743597434399e-08, -1.85261578262194e-08, -3.16189850178213e-07, -1.40590223217031e-07, -2.40038126825765e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -1.03007881072424e-08, -1791.9999999991, 0, 0, 0, 0, 27016.5333333428, 0, 1.2053547327477e-07, 12680.5333332967, 1.64787677361592e-07, 12680.5333335152, 0, 0, 0, 0, -136772.266666787, 0, -3.42461382011267e-07, -35327.9999997687, -6.90767703634439e-07, 0, -172100.266666446, -2.03390965702196e-06, -172100.266667564, -35328.0000014072, -70656.0000020458, 0, 0, 0, 0, 331093.333333772, 0, 1.9818176940605e-07, 48469.3333327998, 1.14553250862226e-06, 0, 710655.999999814, 0, 6.78024855908754e-06, 428031.999998311, 7.94439774557964e-06, 710656.000001705, 48469.3333368191, 145408.000007944, 428032.000006034, 856064.000008011, 145408.000012934, 0, 0, 0, 0, -420522.666667332, 0, 4.61905500317022e-07, -32767.9999993901, -7.62492057315368e-07, 0, -1294336.000001, 0, -9.25666476136332e-06, -518826.66666309, -1.23421825276692e-05, -1294336.00000125, -32768.0000036854, -131072.000013031, 0, -1359871.99999645, -2.22861108089621e-05, -1359872.00000966, -2719744.00001313, -518826.666681371, -131072.000024804, -1556480.00002422, -196608.000040701, -1556480.00003318, 0, 0, 0, 0, 270882.13333379, 0, -7.15852465724165e-07, 8738.1333330385, 1.13023249983857e-07, 0, 1092266.66666832, 0, 5.48553017820783e-06, 305834.666663719, 8.24965806715806e-06, 1092266.66666679, 8738.13333458023, 43690.66667645, 0, 1660245.33333136, 0, 2.28678518651474e-05, 1135957.33332679, 2.55536081555222e-05, 1660245.33333942, 3320490.66667625, 305834.666682158, 43690.6666845397, 1223338.6666946, 87381.3333814594, 1135957.33335226, 3407872.00002738, 87381.3333899905, 3407872.00003352, 1223338.66670677, 1835008.00007021, 0, 0, 0, 0, -69905.0666667845, 0, 2.70160938108469e-07, 5.85948928409666e-08, 4.62976209972306e-08, 0, -349525.333334028, 0, -1.15612717119086e-06, -69905.0666658389, -1.91436109339561e-06, -349525.333333165, 6.4439980114684e-08, -2.71362702221754e-06, 0, -699050.666666805, 0, -7.69936342995231e-06, -349525.333330167, -9.51373360397864e-06, -699050.666667783, -1398101.33333586, -69905.0666724938, -4.13237787669351e-06, -349525.333344078, -1.86069113164739e-05, 0, -699050.666663788, -1.39550961136527e-05, -699050.666673966, -2097152.00001069, -349525.333344863, -2.32595607377661e-05, -1398101.33334908, -3.60980359899359e-05, -2097152.00001119, -349525.333348864, -699050.666709829, -1398101.33335353, -2097152.00003579, -699050.666709257}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 1.08686715391794e-08, 1493.33333333322, 0, 0, 0, 0, -25500.444444458, 0, -1.14805429455773e-07, -7580.44444443019, -1.60965202577011e-07, -7580.4444446012, 0, 0, 0, 0, 153031.111111203, 0, 3.7736214505939e-07, 14279.1111109901, 7.48160181024695e-07, 0, 119523.555555566, 1.62923270857112e-06, 119523.555556077, 14279.1111125141, 28558.2222239242, 0, 0, 0, 0, -425756.444444704, 0, -5.15087973869944e-07, -11832.8888885569, -1.55849785776647e-06, 0, -620771.555555966, 0, -5.16799246601749e-06, -206847.999999482, -6.4083037444759e-06, -620771.5555561, -11832.8888933118, -35498.6666734502, -206848.000004227, -413696.000005101, -35498.6666778455, 0, 0, 0, 0, 602567.111111442, 0, 2.20023101757606e-07, 3640.8888883821, 1.62616015185471e-06, 0, 1354410.66666806, 0, 7.18750726433352e-06, 156558.222220733, 1.05953483803979e-05, 1354410.66666639, 3640.8888956587, 14563.5555672954, 0, 904760.888888413, 1.6448368900826e-05, 904760.888893561, 1809521.77778442, 156558.222234229, 14563.5555799141, 469674.666683207, 21845.3333665642, 469674.666690104, 0, 0, 0, 0, -422343.111111307, 0, 1.18868752168979e-07, 3.09043574787674e-07, -8.44970284874807e-07, 0, -1310720.00000145, 0, -4.52951813683282e-06, -43690.6666651704, -7.81044924992293e-06, -1310719.9999992, -4.66431246585004e-06, -9.64929674228774e-06, 0, -1398101.33333385, 0, -1.63716615272771e-05, -553415.111108811, -1.91098352934737e-05, -1398101.33333491, -2796202.66667035, -43690.6666811951, -2.17945540276539e-05, -174762.666686954, -3.93526480600901e-05, -553415.111122572, -1660245.33334906, -4.83276335625549e-05, -1660245.33335198, -174762.666699487, -262144.000050121, 0, 0, 0, 0, 116508.444444487, 0, -8.20571946752245e-08, -8.52964839854141e-08, 1.78221228000343e-07, 0, 466033.777778283, 0, 1.09470851569787e-06, -5.02655052734234e-07, 2.12485312329992e-06, 466033.777777396, 1.25383075210352e-06, 2.95901083245454e-06, 0, 699050.666667364, 0, 5.46127833446328e-06, 116508.444443016, 7.42667480951163e-06, 699050.666666441, 1398101.33333396, 6.12654196608497e-06, 7.1194438861477e-06, 8.25332666647516e-06, 1.56546021835622e-05, 0, 466033.77777697, 9.86705589964409e-06, 466033.777781178, 1398101.33333867, 116508.444452335, 2.1725809704054e-05, 466033.777787434, 2.93672808977342e-05, 1398101.33333796, 1.56128853980076e-05, 3.13429051427811e-05, 466033.777790152, 699050.6666887, 3.29593283213009e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -896.000000000023, 0, 0, 0, 0, 16256.000000007, 0, 5.5610504394777e-08, 3114.66666666668, 7.58685030171595e-08, 3114.66666673272, 0, 0, 0, 0, -106609.777777799, 0, -1.99216046931924e-07, -3583.99999997093, -3.5912543397033e-07, 0, -52849.777777884, -7.1721624897539e-07, -52849.7777779368, -3584.00000063799, -7168.00000073023, 0, 0, 0, 0, 329728.000000008, 0, 3.44056019356335e-07, 1365.33333320986, 7.96786706022294e-07, 0, 308565.333333823, 0, 2.26468819156567e-06, 56661.3333335773, 2.7594936338366e-06, 308565.333333307, 1365.33333549286, 4096.00000299561, 56661.3333348589, 113322.666668567, 4096.00000488635, 0, 0, 0, 0, -515185.777777717, 0, -2.87911424619718e-07, 2.97216302808591e-07, -9.19909838790298e-07, 0, -784611.555556469, 0, -3.3378793950399e-06, -20024.8888890156, -4.65051262375783e-06, -784611.555554965, -3.66412202240114e-06, -5.27186992810511e-06, 0, -289450.666667906, -6.97294987000066e-06, -289450.666667662, -578901.333335416, -20024.8888936267, -1.11064791676352e-05, -60074.6666734381, -1.40800955811115e-05, -60074.6666757435, 0, 0, 0, 0, 393215.999999918, 0, 9.50199354513551e-08, -2.19726976908435e-07, 5.24977980621399e-07, 0, 873813.333333957, 0, 2.25954507483288e-06, -1.97980294726844e-07, 3.43061763740635e-06, 873813.333332607, 2.73579444004971e-06, 4.54714857424709e-06, 0, 567978.666668157, 0, 6.92397203613145e-06, 87381.3333341198, 7.87726735047754e-06, 567978.666666267, 1135957.33333395, 6.2250463014827e-06, 1.07318600944724e-05, 8.75777040353768e-06, 1.65396173227717e-05, 87381.3333369347, 262144.000005907, 2.09027826121793e-05, 262144.000005806, 1.35304907777694e-05, 2.02728260647791e-05, 0, 0, 0, 0, -116508.444444412, 0, 0, 7.27994506440953e-08, -1.24562163693189e-07, 0, -349525.333333483, 0, -6.07672246923979e-07, 1.53014757508346e-07, -9.61796734634875e-07, -349525.333333044, -8.25399092946941e-07, -1.45505676118141e-06, 0, -349525.333333983, 0, -2.35531519988524e-06, -3.57164404773401e-08, -3.09401189709787e-06, -349525.333332845, -699050.666666546, -2.76122034775336e-06, -3.81176018924315e-06, -3.65816068424071e-06, -6.61142383542395e-06, 0, -116508.444445004, -3.98592313991447e-06, -116508.444445124, -349525.333335141, -2.71869821442989e-06, -9.81151794776403e-06, -3.90384680114163e-06, -1.22800443717172e-05, -349525.333334132, -7.04545188465079e-06, -1.2771092333019e-05, -4.07332434613192e-06, -8.12389770490076e-06, -1.32334348902216e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 358.399999999991, 0, 0, 0, 0, -6741.33333333559, 0, -1.35746175894712e-08, -768.000000001612, -1.63803445654235e-08, -768.000000011869, 0, 0, 0, 0, 46660.2666666644, 0, 5.08975557441506e-08, 409.599999997041, 7.2342940754588e-08, 0, 13619.200000066, 1.60661042493202e-07, 13619.2000000207, 409.600000128257, 819.200000139424, 0, 0, 0, 0, -154965.333333292, 0, -9.99012548447253e-08, 3.45264793394441e-08, -1.622446206904e-07, 0, -85333.3333335807, 0, -5.17329995611819e-07, -6826.66666691043, -5.83676359371793e-07, -85333.3333332913, -4.37875740307976e-07, -5.99000544057695e-07, -6826.6666669102, -13653.3333336612, -9.89948238243221e-07, 0, 0, 0, 0, 263236.266666566, 0, 1.02253102344479e-07, -1.28768183745069e-07, 2.03248763065993e-07, 0, 240298.666667021, 0, 8.50944888009597e-07, 3.40454708534187e-07, 1.00051580929493e-06, 240298.666666473, 8.19265716786386e-07, 1.04405954783582e-06, 0, 38229.3333342408, 1.58883680668128e-06, 38229.3333333889, 76458.666667043, 7.83573667775356e-07, 2.17288032013529e-06, 1.34914974432037e-06, 2.77925893460952e-06, 1.57822523216242e-06, 0, 0, 0, 0, -218453.333333242, 0, -5.50557811990689e-08, 1.03893239129525e-07, -1.0982360233796e-07, 0, -305834.666666811, 0, -6.11689737399305e-07, -1.27730513524857e-07, -6.437673642921e-07, -305834.66666646, -6.31375913250436e-07, -9.50168163398575e-07, 0, -87381.3333342146, 0, -1.55235939046188e-06, -9.3129119307637e-07, -1.59718005362119e-06, -87381.3333331328, -174762.666666699, -1.16808592382888e-06, -2.20146308207466e-06, -1.87419798677225e-06, -3.1748755452498e-06, -5.27509770895279e-07, -1.16358697979834e-06, -4.1499672653094e-06, -6.95554870222989e-07, -2.32849444713445e-06, -3.92571549689698e-06, 0, 0, 0, 0, 69905.066666637, 0, 0, -3.73086580215364e-08, 2.54971791000138e-08, 0, 139810.133333337, 0, 1.70291487047487e-07, -1.44462185008272e-08, 1.68341367603313e-07, 139810.133333253, 2.15407758184549e-07, 3.14337587708941e-07, 0, 69905.0666669677, 0, 5.63286778167265e-07, 2.77913196980632e-07, 6.31188102541466e-07, 69905.0666665346, 139810.133333281, 5.39985879851978e-07, 7.88963691009359e-07, 7.83824956798052e-07, 1.25578155938962e-06, 0, 5.05692986061013e-07, 8.12705854679629e-07, 2.49467036419218e-08, 3.77344868545799e-07, 3.83769609629005e-07, 1.97504761250035e-06, 8.21640327737883e-07, 2.30422616442945e-06, 0, 1.3171764599754e-06, 2.41230433336878e-06, 6.02166256376337e-07, 1.4524959911916e-06, 2.0746492839197e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -51.2000000000573, 0, 0, 0, 0, 904.533333333936, 0, 0, 460.800000001258, 0, 51.2000000031803, 0, 0, 0, 0, -5632.00000000044, 0, 0, -409.60000000514, -2.20320056067768e-08, 0, -8089.60000001192, -2.24566153850467e-08, -853.333333340303, -2.78933194541482e-08, -409.600000032153, 0, 0, 0, 0, 15701.3333333296, 0, 0, 1.37781567282826e-08, 4.59452384517931e-08, 0, 49834.6666667156, 0, 7.89121107555094e-08, 6826.66666669481, 9.36206969418614e-08, 4778.66666665058, 8.96730391436849e-08, 8.92958956958263e-08, 9.08071175221997e-08, 6826.6666667349, 2.06872672204342e-07, 0, 0, 0, 0, -19660.799999987, 0, 1.00286169623126e-08, 0, -5.21247353504843e-08, 0, -136533.33333344, 0, -2.63045664784288e-08, -2.5525836759423e-08, -1.97026271245062e-07, -10922.666666648, -1.26221873850756e-07, -1.84044580301589e-07, 0, -38229.3333334518, -2.36649615277318e-07, -2.30943195487745e-08, -38229.3333334111, -2.27615064403837e-07, -5.14197350324781e-07, -2.94633383835561e-07, -6.37933316350356e-07, -9.84417849249883e-08, 0, 0, 0, 0, 8738.13333331954, 0, -1.218600250408e-08, 0, 2.43531033025931e-08, 0, 166024.533333411, 0, -4.42219284393551e-08, 0, 1.70208986175935e-07, 8738.13333329295, 1.05414579578696e-07, 2.08727062788145e-07, 0, 87381.3333334707, 0, 1.44662611549879e-07, 4.39958980323385e-08, 3.00553861765173e-07, -3.60947801313217e-08, 87381.3333332765, 3.465264744846e-07, 3.51203070645621e-07, 2.76827310527008e-07, 7.8065078181449e-07, 4.83582138647405e-08, 1.32991600881894e-07, 7.37305327880058e-07, 1.34495342323283e-07, 3.6063498662043e-07, 9.04012302560295e-07, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -69905.0666666878, 0, 0, 0, -6.47406084479726e-08, 1.40770361672762e-08, -2.41353737387175e-08, -6.40834695094775e-08, 0, -69905.0666667358, 0, -7.06015396936914e-08, 2.30229961881791e-08, -1.49642728997195e-07, 5.06810493462813e-08, -69905.0666666456, -1.10525360413573e-07, -1.73515527090356e-07, -1.24214919136612e-07, -3.09742138032303e-07, 0, -3.51855801958731e-08, -1.55257422891243e-07, 0, -7.05810917263399e-08, -1.21229590995147e-07, -4.19067209245378e-07, -9.04356676589724e-08, -7.32715290212503e-07, -3.18398039177774e-08, -3.27024837080677e-07, -4.36297417252584e-07, -4.01767243903381e-07, -2.1658231074429e-07, -7.94791748263506e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -42.666666666584, 0, 0, 0, 0, 668.444444445244, 0, 0, 554.666666661917, 0, 42.6666666667004, 0, 0, 0, 0, -3356.44444446301, 0, 0, -1877.33333330614, 0, 0, -8647.11111106896, -1.38548574287577e-08, -625.777777777787, 0, -511.999999998435, 0, 0, 0, 0, 6371.55555563196, 0, 0, 1365.33333327087, -1.59357680944971e-08, 0, 43007.9999999147, 0, 1.11087963805237e-08, 28899.5555552956, 6.84854838916409e-08, 2730.66666669733, -1.74771317504453e-08, 1365.3333334008, -4.52762337543641e-08, 7509.33333336035, -1.08566206284971e-08, 0, 0, 0, 0, -3640.88888901979, 0, 2.50517668204003e-08, 3.97965226791654e-08, 5.03612441132441e-08, 0, -80099.5555555647, 0, -5.49930382554892e-08, -20024.8888883697, 0, -3640.88888888785, 2.11449605557715e-08, -3.69828683266215e-08, 0, -140174.222221523, -6.38270465391826e-08, 2.55407875911679e-08, -32768.0000000405, 2.33815504984805e-08, 0, -20024.8888888587, 1.33384076235933e-08, -4.60773870147173e-07, 0, 0, 0, 0, 9.98328082555549e-08, 0, -3.41981523327688e-08, -1.71621656202522e-08, -2.11207753098654e-08, 0, 43690.6666668309, 0, 8.58275968634603e-08, -3.85886994999318e-07, 3.22686295967486e-08, 1.78881899065665e-08, -1.20569785499883e-07, -5.52519260110852e-08, 0, 247580.444443961, 0, 2.56955064290921e-07, 87381.3333322971, 2.51150542340549e-07, -2.3047591997152e-08, 43690.6666668634, -1.81228452329236e-07, 4.21250374321009e-08, 5.44108588713695e-08, 8.15494621384399e-08, 1.64859227461588e-07, 87381.3333335281, 3.19390959474744e-07, 3.65304150750103e-07, 3.85485788306446e-07, 9.08109187618543e-08, 0, 0, 0, 0, -2.87840326021965e-08, 0, 0, 0, 0, 0, -8.94992223265785e-08, 0, -1.07928263108195e-08, 9.48022973005168e-08, 2.4672910891969e-08, 0, 5.99403640655162e-08, 2.51998893681753e-08, 0, -116508.444444367, 0, 3.29492821947151e-08, 4.35105310057723e-07, -1.74743691708152e-08, 0, -4.20549374504602e-08, 2.41453986903777e-08, 7.24318534495695e-08, 1.813424075231e-08, -2.40570691546607e-08, 0, -116508.444443966, -9.3432351831479e-08, -6.25426121635186e-08, -116508.444444469, -2.14833150155059e-08, 0, -5.56712876399751e-08, 0, 0, 7.06228724854311e-08, -4.28732200903643e-07, 5.77440460774037e-07, -1.92998976604572e-07, -1.76350608522797e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -42.6666666664656, 0, 0, 0, 0, 554.666666666717, 0, 0, 668.444444437987, 0, 42.6666666666581, 0, 0, 0, 0, -1877.33333334535, 0, 0, -3356.44444441116, 0, 0, -8647.11111105259, 0, -512.000000015539, 1.03608845553361e-08, -625.777777785063, 0, 0, 0, 0, 1365.33333339231, 0, -2.53805606388446e-08, 6371.55555549796, -3.55315954829977e-08, 0, 28899.5555554148, 0, 6.70779935168028e-08, 43007.9999996882, 0, 1365.33333336624, -7.66940250295072e-08, 2730.66666660093, 7.02042915440915e-08, 7509.33333341008, 3.05382339225173e-08, 0, 0, 0, 0, -1.03382746061657e-07, 0, 3.49642596757374e-08, -3640.8888888304, 4.03722243786217e-08, 0, -20024.8888888515, 0, -5.27174073222859e-08, -80099.5555550123, -2.74473765312558e-08, -9.3756466832642e-08, 2.3732101830931e-07, -3640.88888877936, 0, -140174.222221475, -2.51574703670827e-07, -1.9948147083106e-07, -20024.8888891552, 4.76296413619058e-08, 7.43148047307523e-08, -32768.0000002444, -1.87100219656545e-07, 1.18102849131436e-07, 0, 0, 0, 0, 8.06200373132868e-08, 0, -1.85331749636464e-08, 0, -5.01004051706527e-08, 0, 9.30346193821577e-08, 0, -2.63170184386537e-08, 43690.6666662741, -7.47597791085276e-08, 4.57254405494812e-08, -1.56992446342051e-07, -5.3790528209976e-08, 0, 87381.3333326842, 0, 1.41015118822471e-07, 247580.444443334, 3.57875353500815e-08, 1.78637660546605e-07, 1.68595992065897e-07, -2.5177681710176e-08, -3.84548001807505e-07, 43690.6666667593, 6.93058724153587e-08, -4.50972386880321e-08, 87381.3333336257, 0, 3.0912803170874e-07, -4.48377771601276e-07, 3.59811899750478e-07, 0, 0, 0, 0, -2.33211294807374e-08, 0, 0, 0, 1.46557313644414e-08, 0, -6.14012300413389e-08, 0, 1.07908791546017e-08, 9.38218769355957e-08, 0, -1.35027150778467e-08, 4.19693727211702e-08, 2.48717417581228e-08, 0, 1.32308706864112e-07, 0, -1.70193287623266e-07, -116508.444443955, -7.10314642267278e-08, -5.39735884868266e-08, -9.33330896490684e-08, 9.84178806052726e-08, 1.41428862870857e-07, -2.38277044094709e-08, 2.44796572532024e-08, 0, -116508.444443873, -1.37856190778552e-07, -4.01048843604461e-08, -2.39711180430228e-07, 0, -2.54540243649684e-08, -116508.444444568, -2.20669422459665e-07, -3.26514137116973e-07, 2.02335781736068e-07, 1.05693559616865e-07, -6.59082196521155e-07, -3.20463808645253e-07, -1.77874554603267e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -51.1999999996511, 0, 0, 0, 0, 460.799999996409, 0, 0, 904.533333328193, 0, 51.1999999987643, 0, 0, 0, 0, -409.599999988906, 0, 0, -5631.99999997585, 1.22971647160813e-08, 0, -8089.59999995282, 0, -409.600000001833, 2.12436522576329e-08, -853.333333320528, 0, 0, 0, 0, -1.76137214947534e-08, 0, 0, 15701.3333332819, -3.35561833107886e-08, 0, 6826.66666653521, 0, 2.25610843521824e-08, 49834.6666664608, 0, 3.04887665784975e-08, -9.93075316233916e-08, 4778.66666664607, -5.10926900942649e-08, 6826.66666673522, -9.42374909683301e-08, 0, 0, 0, 0, 1.70771252044105e-08, 0, 1.47181627833117e-08, -19660.7999999632, 6.62637290516664e-08, 0, 1.61422826004259e-07, 0, -8.26997852342031e-08, -136533.333332961, 1.0278290881874e-07, -2.05204000449653e-08, 1.62923774370248e-07, -10922.6666665663, 0, -38229.3333328145, -1.76429689012473e-08, 3.73601073254795e-08, -2.1031034750878e-07, 1.89674937317166e-07, 2.61823290148198e-07, -38229.3333334159, 3.19317083654312e-07, -2.16060789016229e-07, 0, 0, 0, 0, -1.18986466850321e-08, 0, -2.13474341008492e-08, 8738.13333331791, -3.95444876721645e-08, 0, -7.18883119260767e-08, 0, 7.97014321275334e-08, 166024.533333059, -1.01920163664372e-07, 2.15286264775907e-08, -2.03777394203337e-07, 8738.13333317615, 0, -4.12737376469732e-07, 0, 1.83928332097871e-07, 87381.3333326734, 1.01078228782155e-07, 3.03441223717922e-08, 2.69013020212703e-07, -3.76996279958564e-07, -1.91121508702051e-07, 87381.333333415, -3.5323088845535e-07, -4.50582540269279e-08, 3.9243690274992e-07, -3.33415049137433e-07, 4.49313205626791e-07, 1.16187456771554e-07, 2.17933718743664e-07, 0, 0, 0, 0, 0, 0, 0, 0, 1.0705671897111e-08, 0, 1.19042631246867e-08, 0, -2.22719768206103e-08, -69905.0666665966, 5.88832383521393e-08, 0, 8.24064726846009e-08, 5.98247923632405e-08, 0, 1.13644763280894e-07, 0, 0, -69905.0666664069, 4.67240489012108e-08, 0, -7.05541888558732e-08, 1.49797610614795e-07, 1.29129304305405e-07, -69905.0666666563, 1.71921011579198e-07, 0, 2.99675273551166e-07, -4.59287193024786e-08, 0, -1.80259714066104e-07, 1.2164489649994e-07, 1.80989860099599e-07, -1.84722106048365e-07, 2.85165752004844e-07, -1.74926941723706e-07, 1.58467433921465e-07, -1.75417971396144e-07, 1.85881091376095e-07, -4.01360514360792e-07, 4.77543182524983e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 358.399999998756, 0, 0, 0, 0, -767.999999986438, 0, 2.17045227588076e-08, -6741.33333331638, 2.87227217109016e-08, -767.999999995388, 0, 0, 0, 0, 409.599999949095, 0, -1.01942285829202e-07, 46660.2666665852, -1.55381035455392e-07, 0, 13619.1999998353, -2.14885494592999e-07, 819.200000016433, 409.599999923391, 13619.1999999575, 0, 0, 0, 0, 9.87889079974411e-08, 0, 2.4376714205317e-07, -154965.333333158, 4.02918398784908e-07, 0, -6826.66666618057, 0, 6.79155004104279e-07, -85333.3333326346, 8.1228310703572e-07, -3.82807141684684e-08, 3.92683238339159e-07, -85333.3333330317, -6.98066723836061e-08, -13653.3333335336, -6826.66666640309, 0, 0, 0, 0, -1.14258211111734e-07, 0, -3.2864182748758e-07, 263236.266666482, -5.81327745683543e-07, 0, -6.3117643140318e-07, 0, -1.13130908055134e-06, 240298.666665403, -1.55719385901076e-06, 7.08258799752047e-08, -9.64792118310895e-07, 240298.666665952, 0, 38229.3333316799, -1.74022144877624e-06, 1.68179904363888e-07, 5.11578582465657e-07, -4.17756740972136e-07, -9.70333558548027e-07, 76458.6666669264, 38229.3333324323, 1.34162480006716e-06, 0, 0, 0, 0, 7.59242667875145e-08, 0, 2.10598458864395e-07, -218453.333333244, 4.10715001746307e-07, 0, 3.87410049413036e-07, 0, 1.00298919327962e-06, -305834.666665648, 1.40952089838071e-06, -4.97977419417432e-08, 8.57738237707888e-07, -305834.666665866, 0, 1.49385258516587e-06, 0, 1.86684513111543e-06, -87381.3333312248, 1.96639305822769e-06, -1.46728393079668e-07, -2.81702106290268e-07, 9.89277859585166e-07, 1.60482160732528e-06, -174762.666666623, -87381.3333317618, 2.13384919024196e-07, -7.23175911001092e-07, 1.68519129375288e-06, -1.75322922932088e-06, -1.09237546482044e-06, -1.83705807097959e-06, 0, 0, 0, 0, -2.18249786243821e-08, 0, -5.70066752691313e-08, 69905.0666666524, -1.1365110725141e-07, 0, -9.60843528697021e-08, 0, -3.2559996290217e-07, 139810.133333031, -4.82016651386532e-07, 1.45353552112122e-08, -3.0458732644148e-07, 139810.133333022, 0, -4.63442365635845e-07, 0, -7.03459315988839e-07, 69905.0666657752, -8.82102005610415e-07, 1.33643925608853e-08, 7.22480704296557e-08, -6.00964935062666e-07, -8.54618614830729e-07, 139810.133333182, 69905.0666657575, 0, -9.99120425435363e-07, -9.65632680200206e-07, -3.98567602780069e-08, 2.74754494220643e-07, -3.84096209663695e-07, -1.10308319429445e-06, 2.6183283310446e-07, -1.09420384874272e-06, 8.25982224687882e-07, -8.93080252999206e-08, 6.1408340682324e-07, 1.44688801089632e-06, 1.791093671203e-06, 1.45111508077362e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -895.999999996846, 0, 0, 0, 0, 3114.66666663466, 0, -6.86383654705983e-08, 16255.9999999528, -9.96345211672847e-08, 3114.66666663889, 0, 0, 0, 0, -3583.99999989525, 0, 3.14470286118141e-07, -106609.777777551, 5.62988560254469e-07, 0, -52849.7777773245, 5.88845013613287e-07, -7168.00000007045, -3583.99999959668, -52849.777777572, 0, 0, 0, 0, 1365.33333317194, 0, -7.99126845039853e-07, 329727.999999531, -1.55405093438664e-06, 0, 56661.3333320358, 0, -1.58635362518425e-06, 308565.333331343, -2.40198866414686e-06, 4096.00000029672, 1365.33333128652, 308565.333332089, 4095.99999987368, 113322.666667687, 56661.3333319603, 0, 0, 0, 0, 1.45977086847306e-07, 0, 1.13433230585106e-06, -515185.777777324, 2.27031767098207e-06, 0, -20024.888887363, 0, 2.56925732789384e-06, -784611.555551988, 5.07374411426354e-06, -4.27813580530394e-07, 4.592553314024e-06, -784611.555552383, 0, -289450.666661882, 4.03969944613339e-06, -6.18963144781589e-07, -60074.6666696606, 2.81019712659949e-06, -20024.8888838429, -578901.333335273, -289450.666662819, -60074.6666707037, 0, 0, 0, 0, -8.57916997256729e-08, 0, -7.93017266528486e-07, 393215.999999811, -1.62438570871239e-06, 0, -7.76404802698273e-07, 0, -2.33717699018651e-06, 873813.333330556, -4.99176083807004e-06, 2.51908411838e-07, -4.28848677253055e-06, 873813.333329805, 0, 87381.333329121, 0, -3.60231793188719e-06, 567978.666660421, -5.13689968875716e-06, 7.42377440942457e-07, 2.40721504238967e-06, -5.39809176299896e-06, -7.7865169624947e-06, 1135957.33333391, 567978.666660238, -1.29088411855728e-06, 262144.000005031, 87381.333326757, 7.8234028027532e-06, 9.62945032160958e-07, 262144.00000716, 0, 0, 0, 0, 2.51597923691648e-08, 0, 2.25147188057371e-07, -116508.444444424, 4.52185957108223e-07, 0, 1.44717974869258e-07, 0, 7.95715161594647e-07, -349525.333332553, 1.81222720820485e-06, -5.87325359548203e-08, 1.50800820507765e-06, -349525.333331962, 0, 1.21320120285489e-06, 0, 1.29786477449098e-06, -349525.333330708, 2.5863202602851e-06, -1.73803748480348e-07, -6.8798870782176e-07, 2.93407609591899e-06, 3.93168528598146e-06, -699050.666666199, -349525.333329608, 0, -116508.444441456, 1.96121814584276e-06, 9.55986461469985e-08, -2.35633940249647e-06, 1.85654129725061e-06, 4.3082366581504e-06, -349525.333335492, -116508.444440397, -3.63904240176814e-06, 2.5883680412307e-06, -349525.333335876, -4.04669140692718e-06, -8.02267211458411e-06, -3.76326405941229e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -1.35728742524396e-08, 1493.33333332864, 0, 0, 0, 0, -7580.4444444027, 0, 1.0591440155609e-07, -25500.4444443632, 1.75613829824206e-07, -7580.44444438177, 0, 0, 0, 0, 14279.1111110183, 0, -4.8946769828908e-07, 153031.111110713, -1.05958439091216e-06, 0, 119523.555554785, -6.82597770861443e-07, 28558.2222224454, 14279.1111102067, 119523.555555122, 0, 0, 0, 0, -11832.8888888824, 0, 1.37567529352794e-06, -425756.444443644, 3.11173343363611e-06, 0, -206847.999997902, 0, 1.10281914076833e-06, -620771.55555197, 3.17155652783845e-06, -35498.6666676767, -11832.8888841287, -620771.555553009, -35498.6666665118, -413696.000002661, -206847.99999726, 0, 0, 0, 0, 3640.88888905109, 0, -2.12856183492035e-06, 602567.111110389, -4.66429830834928e-06, 0, 156558.222220157, 0, -1.6244059214287e-06, 1354410.66666021, -7.95716681728318e-06, 14563.5555569462, 3640.88887825553, 1354410.66665982, 0, 904760.888880155, -2.30039282253856e-06, 21845.3333360217, 469674.666674785, 14563.5555495403, 156558.222211178, 1809521.77778339, 904760.888882576, 469674.666675174, 0, 0, 0, 0, -1.55276227220351e-07, 0, 1.61366629904249e-06, -422343.111110858, 3.3955484832215e-06, 0, -43690.6666660653, 0, 1.92051743620965e-06, -1310719.99999509, 8.79270115252355e-06, -7.94977023420046e-07, 1.01514070997377e-05, -1310719.99999219, 0, -553415.11110369, 0, 1.65578577990544e-08, -1398101.33332151, 4.62297729776777e-06, -3.13072317798594e-06, -174762.666673827, 1.18892559225913e-05, -43690.6666491104, -2796202.66666919, -1398101.33332213, 6.85963996191573e-07, -1660245.33334747, -553415.111100574, -262144.000019218, -174762.666665867, -1660245.33335096, 0, 0, 0, 0, 4.26731039020314e-08, 0, -4.8331961885966e-07, 116508.444444434, -9.55174388886244e-07, 0, 5.32193368449327e-08, 0, -8.31371085723627e-07, 466033.777776466, -3.44080865247889e-06, 1.71419055615907e-07, -3.58632153545519e-06, 466033.777774696, 0, 116508.444442517, 0, 2.60124267566474e-07, 699050.666661679, -3.25894077758599e-06, 9.45184875556448e-07, 2.11265330842002e-06, -6.4363713500968e-06, -8.93837566578712e-06, 1398101.33333259, 699050.66665968, 0, 466033.777772092, -1.75837764402491e-07, 1.18754962781679e-06, 466033.777784738, -2.92004076439032e-06, 116508.444436827, 1398101.33333966, 466033.777771983, 9.16292214302624e-06, -7.99387359571843e-06, 1398101.33334055, 7.97061272290355e-06, 699050.666686229, 466033.777784692}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -1791.9999999961, 0, 0, 0, 0, 12680.5333333054, 0, -4.27754623349017e-08, 27016.5333332533, -1.03745321075234e-07, 12680.5333333141, 0, 0, 0, 0, -35327.9999999942, 0, 2.76869111673544e-07, -136772.266666261, 7.57410780012401e-07, 0, -172100.266665925, -2.8973211260764e-07, -70656.000000523, -35327.9999994561, -172100.266666695, 0, 0, 0, 0, 48469.3333335866, 0, -1.13015316775797e-06, 331093.333332519, -2.61262283595686e-06, 0, 428031.999998129, 0, 2.23733100480062e-06, 710655.999996278, 6.26281948507458e-07, 145408.000001804, 48469.3333293105, 710655.999999223, 145408.00000187, 856064.000004759, 428032.000000783, 0, 0, 0, 0, -32768.000000543, 0, 2.11111572535432e-06, -420522.66666596, 4.31575022529135e-06, 0, -518826.666665333, 0, -3.13018930584607e-06, -1294335.99999315, 1.86856406683774e-06, -131072.000002259, -32767.9999898063, -1294335.99999576, 0, -1359871.99999093, -8.38944908880318e-06, -196608.000007009, -1556480.00001219, -131071.999998928, -518826.666662211, -2719744.00001154, -1359872.00000581, -1556480.00001747, 0, 0, 0, 0, 8738.13333375733, 0, -1.7874121600001e-06, 270882.133333113, -3.32751117001794e-06, 0, 305834.666666929, 0, 1.17754991231183e-06, 1092266.66666149, -4.4672468380853e-06, 43690.6666678623, 8738.13332288422, 1092266.66666046, 0, 1135957.33332602, 0, 1.1176358403765e-05, 1660245.33332045, 7.75937643667604e-06, 87381.3333399733, 1223338.6666773, 43690.6666597222, 305834.666653162, 3320490.66667546, 1660245.333337, 87381.3333408284, 3407872.00002191, 1135957.33334069, 1835008.00002984, 1223338.66667635, 3407872.00003667, 0, 0, 0, 0, -1.16219668411778e-07, 0, 5.65302640219243e-07, -69905.066666669, 9.71491459475366e-07, 0, -69905.0666670546, 0, 9.08882740443606e-08, -349525.333331982, 2.3477225671733e-06, -2.22141514281087e-07, 3.84347091434469e-06, -349525.333330558, 0, -349525.333331698, 0, -4.13939361013901e-06, -699050.666661135, -1.48424329525747e-06, -1.91208487808107e-06, -349525.333336465, 4.64841416946723e-06, -69905.0666583503, -1398101.33333473, -699050.666665478, 0, -699050.666660519, -6.58679473815312e-06, -4.91714987608906e-06, -1398101.33334371, -1.84480365000068e-06, -349525.333332982, -2097152.00001052, -699050.66667316, -699050.666679721, -349525.333328939, -2097152.00001908, -699050.666680967, -2097152.00002957, -1398101.33335171}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 1119.99999999792, 0, 0, 0, 0, -14645.3333333317, 0, -3.82171803773577e-08, -14645.333333267, -1.75771080777903e-08, -5685.33333336566, 0, 0, 0, 0, 56192.0000001363, 0, 7.16441330785687e-08, 56191.999999642, -1.11938831887952e-07, 0, 184063.999999411, 1.12234880295814e-06, 66901.3333338463, 10709.3333334159, 66901.3333337523, 0, 0, 0, 0, -94549.333333976, 0, 3.31053872413497e-07, -94549.3333326225, 9.4151056132015e-07, 0, -647509.33333208, 0, -4.78686011558655e-06, -647509.333329949, -3.78762344757183e-06, -197973.333334951, -8874.66666545594, -197973.333334237, -112298.666669161, -759808.00000399, -112298.666670464, 0, 0, 0, 0, 73728.000001099, 0, -1.10486628243201e-06, 73727.9999994284, -2.01378686037366e-06, 0, 980309.333333315, 0, 7.17996981281055e-06, 980309.333326919, 3.92831599819358e-06, 223914.666668707, 2730.66666203961, 223914.666666137, 0, 1813162.66665837, 1.49814625616704e-05, 229376.000006577, 2042538.66667645, 81920.0000033453, 81920.0000057189, 2042538.66667592, 229376.000013051, 1144149.33334883, 0, 0, 0, 0, -21845.333334152, 0, 1.1627793626866e-06, -21845.3333332008, 1.71246456786801e-06, 0, -677205.333334898, 0, -4.42887123412115e-06, -677205.333328536, -7.86689145535894e-07, -87381.333334449, 5.75623315124331e-06, -87381.3333310334, 0, -1922389.33332718, 0, -1.7982558769306e-05, -1922389.33332068, -1.52311516680844e-05, -131072.000005757, -2053461.33334278, -21845.3333337845, -21845.3333331102, -2053461.33334065, -131072.000013045, -87381.333342756, -3932160.00001611, -87381.3333548501, -2097152.00002339, -742741.333346519, -2097152.00003087, 0, 0, 0, 0, 2.26696877154973e-07, 0, -3.9917592608668e-07, 2.27387556658558e-08, -5.24837458182699e-07, 0, 174762.666667525, 0, 9.06709680188438e-07, 174762.666665474, -5.42595918802358e-07, 2.06132704418604e-07, -2.37275924537398e-06, -1.29905959597537e-06, 0, 699050.666665688, 0, 6.63369603137362e-06, 699050.666661119, 4.62258171526816e-06, 1.59205558531387e-06, 699050.666669819, -9.73974774791137e-07, -2.46957667427188e-06, 699050.666668121, 3.3377245266341e-06, 0, 1048575.99999401, 1.01505012122565e-05, 4.85147844481243e-06, 2097152.00000811, 4.41618572791193e-06, 8.66114421877274e-06, 2097152.00000721, 1.42102147621056e-05, 1048576.00001039, 174762.666667703, 1048576.00001684, 699050.666677801, 2097152.00002123, 699050.666684375}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 224.000000000067, 0, 0, 0, 0, -3765.3333333367, 0, 0, -2271.99999999651, -1.94372307134598e-08, -480.000000018951, 0, 0, 0, 0, 21632.000000032, 0, 1.39709163825951e-08, 4095.99999997317, 1.07880992092396e-07, 0, 37674.6666666399, 1.90699950387912e-07, 7552.00000004959, 256.000000174668, 4352.00000019782, 0, 0, 0, 0, -53589.3333334444, 0, 3.02557469397202e-08, -2047.99999994137, -2.25367778500207e-07, 0, -211285.333333358, 0, -5.37169795539791e-07, -64170.666666361, -8.42154204676078e-07, -38229.3333333583, -5.32576564454066e-07, -4096.00000077755, -3754.66666712705, -67925.3333338387, -2048.00000138963, 0, 0, 0, 0, 57344.0000001683, 0, -1.15882904387311e-07, -4.49835790975445e-08, 2.01990729350502e-07, 0, 502442.666667005, 0, 5.24431756535948e-07, 30037.3333326243, 1.37967879595246e-06, 73727.9999999089, 7.44061497168432e-07, 1.27362822449372e-06, 0, 322218.666665972, 1.87389641251568e-06, 16384.0000002896, 338602.666667227, 1.4622726789517e-06, 3.34022448546841e-06, 60074.6666682264, 4.17522505047525e-06, 30037.3333361246, 0, 0, 0, 0, -21845.3333334493, 0, 1.42327540977373e-07, 0, -1.10373941531978e-07, 0, -502442.666667156, 0, -2.01548302283196e-07, 6.00080085416755e-07, -1.16649192193927e-06, -43690.6666665262, -3.64894433154752e-07, -1.04805554340313e-06, 0, -611669.333333039, 0, -1.8540944298279e-06, -131071.999998584, -2.60842778436844e-06, -21845.3333331884, -633514.666666989, -1.72030605699442e-06, -2.75406147413666e-06, -1.77160909667145e-06, -5.1677840779771e-06, -1.12534201577416e-06, -262144.000001315, -6.39342929278175e-06, -131072.000002192, -4.27949894012628e-06, -5.81774841308295e-06, 0, 0, 0, 0, 3.00195028118084e-08, 0, -4.58053712170904e-08, 0, 2.48467422340117e-08, 0, 174762.666666867, 0, 4.02009319847503e-08, -1.66733112733461e-07, 3.35836518550098e-07, -6.02378828933029e-08, 3.22862781266873e-08, 3.0410252721288e-07, 0, 349525.333333433, 0, 4.6294813066742e-07, -7.12711573229226e-07, 1.01603275066371e-06, -1.78983944959374e-07, 349525.333333376, 7.08466538580867e-07, 8.87660548180329e-07, 6.70119353092397e-07, 2.03674484014505e-06, 0, 174762.666666067, 1.25849292967884e-06, 2.41977411175949e-07, 349525.333333741, 9.25635934510995e-07, 2.89866318129856e-06, 6.72704237706672e-07, 4.24018106628515e-06, 174762.666666967, 2.06668937259598e-06, 3.88196350817306e-06, 1.0623004111609e-06, 2.23742186200539e-06, 4.80223391229397e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 223.999999998732, 0, 0, 0, 0, -2271.99999999003, 0, 0, -3765.33333330944, 1.45458931013917e-08, -479.999999989963, 0, 0, 0, 0, 4095.99999998647, 0, -1.62017857457932e-08, 21631.9999998838, -1.06874125170095e-07, 0, 37674.6666664463, 2.59998073793573e-08, 4352.00000004549, 255.999999840825, 7551.99999993232, 0, 0, 0, 0, -2048.00000003783, 0, 7.54296352781689e-08, -53589.333333109, 3.55006812937945e-07, 0, -64170.6666660839, 0, -3.55169212741941e-07, -211285.333332288, 7.21326208934513e-08, -4096.00000023505, 7.9736089569508e-07, -38229.3333329393, -2047.99999988019, -67925.3333338441, -3754.66666613335, 0, 0, 0, 0, 1.05063168702591e-07, 0, -1.64796213972526e-07, 57343.9999998161, -5.60208414402921e-07, 0, 30037.3333328265, 0, 7.05544393169078e-07, 502442.666664815, -6.38978472818457e-07, 3.36517280351641e-07, -1.68024167005765e-06, 73727.9999989269, 0, 322218.666664093, 8.91082628843097e-07, 4.10366330386024e-07, 60074.6666684331, -1.22788184084e-06, -1.83365867458147e-06, 338602.666667729, 16383.9999986157, 30037.3333340616, 0, 0, 0, 0, -8.59537092533301e-08, 0, 1.55833635546012e-07, -21845.3333332844, 4.15288369681991e-07, 0, 6.15171227993111e-08, 0, -4.98798323839593e-07, -502442.666665318, 9.56859994497169e-07, -1.98879006100337e-07, 1.6245629695622e-06, -43690.6666654807, 0, -131071.999997856, 0, -1.57965641302377e-06, -611669.333329817, -4.57439602059601e-07, -5.97529952226862e-07, -1.82451495545531e-06, 2.14333031416261e-06, 2.52782284101525e-06, -633514.666667219, -21845.3333311852, 3.32625168723535e-07, -262144.000002936, 1.9806819506984e-06, -3.20074652800338e-06, 1.27192825693984e-06, -131072.000001646, 0, 0, 0, 0, 2.36650554818479e-08, 0, -5.24221499005247e-08, 0, -1.16590449135547e-07, 0, 5.90966637609268e-08, 0, 1.25014657596263e-07, 174762.666666332, -4.13208081001645e-07, 4.06924368726586e-08, -5.77850254570372e-07, -4.58547834343185e-07, 0, -5.19006024857335e-07, 0, 6.83922577116585e-07, 349525.333331868, -1.20740026712129e-07, 1.96956326086427e-07, 6.04099759521544e-07, -1.09759823054458e-06, -1.20688005855017e-06, 349525.333333261, -1.23165379187909e-06, 0, 174762.666664961, 7.00153335407025e-07, 1.5523089096267e-07, 1.62925677388452e-06, -5.86028162330086e-07, -1.05043456072938e-06, 349525.33333465, -1.13985525809799e-06, 1.75810182250263e-06, -1.76181265301503e-06, 174762.666667133, 7.04212146588913e-07, 2.90685629664631e-06, -1.19149659394992e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -597.333333332966, 0, 0, 0, 0, 9244.44444445313, 0, 2.41990472280379e-08, 6855.11111108349, 4.12200010106944e-08, 2076.44444448853, 0, 0, 0, 0, -46421.3333334537, 0, -2.53218865626952e-08, -19000.8888887251, -1.73217222153536e-07, 0, -103651.555555321, -6.26402571759418e-07, -29696.0000002264, -2389.33333369405, -21390.2222226889, 0, 0, 0, 0, 96028.4444449102, 0, -1.99234241251364e-07, 20024.8888885539, 1.96276270328807e-07, 0, 497891.555555189, 0, 2.08490783912348e-06, 268970.666665055, 2.55433701939827e-06, 126520.888889385, 910.222222969424, 40960.0000016238, 31402.6666682133, 300373.333335107, 21845.3333369584, 0, 0, 0, 0, -87381.3333340881, 0, 5.72672982869156e-07, -7281.77777750831, 8.63508691512853e-08, 0, -953912.888889463, 0, -2.52299104905807e-06, -262143.999996764, -3.81806978939918e-06, -185685.333333835, -2.8876802991167e-07, -21845.3333355659, 0, -1121393.77777387, -7.25149612170569e-06, -109226.666669075, -1230620.44444764, -10922.6666703964, -7281.7777854893, -535210.666671177, -21845.3333447121, -283989.333342096, 0, 0, 0, 0, 29127.1111116643, 0, -5.90268049699011e-07, -5.87159056242929e-08, -2.12247839318263e-07, 0, 786432.000001368, 0, 1.20796673352695e-06, 87381.3333308371, 2.59080372865704e-06, 87381.3333335342, -6.46473844440857e-07, 1.35794295004461e-06, 0, 1572863.99999741, 0, 7.60164684071715e-06, 902940.444437951, 8.684169607965e-06, 87381.3333348213, 1660245.33333621, 3.60199307202834e-06, 5.1803934849146e-06, 262144.000004342, 1.32645950711244e-05, 29127.1111157959, 1835008.00000536, 1.78656308141244e-05, 961194.666675849, 87381.3333444267, 262144.000017552, 0, 0, 0, 0, -1.5293006184523e-07, 0, 1.93135787184542e-07, -1.26182037499204e-08, 8.83743111836523e-08, 0, -233016.888889529, 0, -1.6333511341753e-07, 6.34149811164705e-07, -5.72299688706644e-07, 0, 4.73151290125783e-07, -2.43193491422505e-07, 0, -699050.666666473, 0, -2.40464500060933e-06, -233016.888885924, -3.17704300639747e-06, -2.41887522412602e-07, -699050.666667622, -1.20022600827171e-06, -9.62773889095177e-07, -1.34726248366843e-06, -4.86541038084398e-06, 0, -699050.666663664, -4.81713345346964e-06, -1.75905444985406e-06, -1398101.33333572, -2.93291833253493e-06, -7.75174454016965e-06, -699050.666669034, -1.16602621166224e-05, -699050.666669855, -4.13622394866413e-06, -1.06431824720844e-05, -233016.88889383, -699050.666675291, -1.26947476788531e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 199.111111110555, 0, 0, 0, 0, -2816.00000000231, 0, 0, -2815.99999997687, 0, -426.66666666449, 0, 0, 0, 0, 11719.1111111778, 0, -2.58345137718314e-08, 11719.1111109857, -5.76084934043406e-08, 0, 39367.1111109045, 8.06457912002059e-08, 5575.11111116173, 227.555555496408, 5575.11111110387, 0, 0, 0, 0, -16384.000000289, 0, 1.26135685355996e-07, -16383.9999997513, 2.29800344262312e-07, 0, -159743.999999548, 0, -4.08055565786383e-07, -159743.999998811, -2.52331802974691e-07, -19114.6666668813, 4.34079592815266e-07, -19114.6666665397, -2730.66666678095, -72362.6666670082, -2730.66666670818, 0, 0, 0, 0, 7281.77777827317, 0, -2.48303258312926e-07, 7281.77777757602, -3.99768983159259e-07, 0, 211171.555555538, 0, 5.94651760976523e-07, 211171.555553337, 1.10107392813258e-07, 14563.5555558851, -1.12233494277812e-06, 14563.5555550643, 0, 611669.33333034, 1.4381106324389e-06, 7281.77777843499, 243939.555556675, -2.96022028651166e-07, -4.55980850027743e-07, 243939.555556138, 7281.77777835411, 32768.000000982, 0, 0, 0, 0, -3.79044313564826e-07, 0, 2.17017387310613e-07, 5.06441087146837e-08, 3.19296758915284e-07, 0, -87381.3333339076, 0, -3.07563097990421e-07, -87381.3333317056, 2.30848240788554e-07, -2.22625386296194e-07, 1.17495574488943e-06, 6.59943368035971e-07, 0, -699050.666664355, 0, -1.8048803350648e-06, -699050.666662186, -1.32875502926103e-06, -6.99614308855925e-07, -174762.666667925, 8.68997377933168e-07, 1.37216297583119e-06, -174762.666666861, -3.70103711518933e-07, -5.5596215705509e-07, -786432.000001559, -1.06532376647633e-06, -87381.3333355797, 3.01036367718592e-07, -87381.3333353842, 0, 0, 0, 0, 1.08864316348044e-07, 0, -6.94386938749635e-08, 0, -9.57941032045884e-08, 0, 3.26258522419532e-07, 0, 1.85934297817129e-08, -3.93355396738141e-07, -1.7946584968539e-07, 5.66943749291709e-08, -4.40162437587859e-07, -2.83337739068474e-07, 0, 233016.888888478, 0, 6.83535724807524e-07, 233016.888886956, 3.54253958486778e-07, 2.37340076848982e-07, 4.75862836430759e-07, -5.34301336638775e-07, -8.93214846127826e-07, -1.21957275861248e-07, -9.77072545415621e-08, 0, 466033.777775599, 1.03341216378498e-06, 4.33801809524286e-07, 466033.777778709, 8.49682993083958e-08, 2.13759420356091e-07, 466033.777778327, 8.91460315560308e-07, 1.29148063269385e-06, -1.04930172940028e-06, 1.03148861567446e-06, 6.68942775136024e-07, 233016.888890855, 6.57548710066109e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -597.333333330955, 0, 0, 0, 0, 6855.11111109915, 0, 0, 9244.44444438658, -4.28391207786101e-08, 2076.44444441683, 0, 0, 0, 0, -19000.8888889377, 0, 4.2605977931505e-08, -46421.3333330389, 3.47137447129503e-07, 0, -103651.555555027, -2.19546299376533e-07, -21390.2222223969, -2389.3333329008, -29695.9999998087, 0, 0, 0, 0, 20024.8888892381, 0, -2.78977359764814e-07, 96028.4444438721, -1.21637531176673e-06, 0, 268970.666665371, 0, 1.67734739397648e-06, 497891.555552827, 2.51020475491389e-07, 40960.0000008317, 910.222219846011, 126520.888887683, 21845.3333333699, 300373.333334901, 31402.6666655643, 0, 0, 0, 0, -7281.77777843136, 0, 6.47261078139864e-07, -87381.3333328732, 1.96004269407401e-06, 0, -262143.999999276, 0, -3.0563072980368e-06, -953912.888883934, 1.37633533490467e-06, -21845.3333345262, 5.3657652634016e-06, -185685.333329902, 0, -1121393.777771, -4.66618690387023e-06, -21845.3333355799, -535210.666672097, -7281.77777503634, -10922.6666620542, -1230620.44444749, -109226.666664496, -283989.333336591, 0, 0, 0, 0, 5.05360671430439e-07, 0, -6.17139902468946e-07, 29127.1111109994, -1.47808752558695e-06, 0, 87381.3333338588, 0, 2.10713562283053e-06, 786431.999996382, -2.70566559636181e-06, 7.17626585308702e-07, -5.32616447742209e-06, 87381.3333293616, 0, 902940.444439007, 0, 7.10670042793035e-06, 1572863.99999032, 3.26964193729885e-06, 2.65053816530021e-06, 262144.00000567, -5.60116405417777e-06, -7.50187303485362e-06, 1660245.33333455, 87381.3333288536, 1.00094952650036e-06, 1835008.0000087, 29127.1111087415, 262144.00001052, 87381.3333303259, 961194.666673721, 0, 0, 0, 0, -1.42803147336346e-07, 0, 2.09289034887162e-07, -1.55992616054119e-08, 4.22352488759376e-07, 0, -4.27488102395804e-07, 0, -4.67092306560892e-07, -233016.888888004, 1.30309105143503e-06, -1.53810697817588e-07, 1.93282917630786e-06, 1.5739721047357e-06, 0, -233016.888887739, 0, -2.99695514139354e-06, -699050.66666256, -2.48959973918862e-07, -8.66596382224502e-07, -1.92394254731055e-06, 3.05960822345689e-06, 3.95347096661556e-06, -699050.666666166, 3.09107758870129e-06, 0, -699050.666661972, -3.45188644327523e-06, -1.42217337721122e-06, -699050.666671557, 6.41384606916302e-07, 1.71347386672819e-06, -1398101.33333702, 1.07550250191266e-06, -5.81015445911256e-06, 5.38695146038662e-06, -699050.666669373, -2.90821763755209e-06, -699050.6666762, -233016.888889712}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -85.333333333762, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.51050318624536e-08, 1558.75555555578, 0, 0, 0, 0, 0, 0, 0, 0, -1.22137894291385e-08, 0, 0, 0, 0, 0, -4.13715602265102e-08, 0, -5.23485876063232e-08, -10240.0000000008, -2.34613253269345e-08, -5.27340395841536e-08, 0, -1.80530058459889e-08, -7.62919205611785e-08, 0, 0, 0, 0, 0, 0, 1.47455041439276e-08, 0, 1.29034402573017e-08, 0, 0, 0, 1.09796617076312e-07, 0, 8.63463850477645e-08, 30947.5555555348, 6.6733870577688e-08, 7.33703377457426e-08, 0, 0, 1.55554792634504e-07, 0, 3.39865234925828e-08, 1.12500681402796e-08, 1.06495617580107e-07, 7.70769032676296e-08, 1.80543756081649e-07, 1.86011022981609e-07, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -9.88993997256996e-08, 0, -2.49798813669314e-08, -43690.6666666532, -3.93265509046614e-08, -5.18855303127228e-08, 0, 0, 0, -1.64350234731863e-07, 0, -1.13127118870208e-07, 1.98024181639437e-08, -3.54660362097496e-08, -3.64770748554556e-08, -1.72630001365375e-07, -1.52226563105153e-07, -1.77782924272271e-07, -5.1059447567936e-08, -1.0482281838126e-07, -3.44364160168021e-07, -3.80022135648846e-08, -2.27710022717418e-07, -2.65944413976633e-07, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.41693399364501e-08, 0, 0, 23301.6888888823, 2.02477773806701e-08, 1.92725648396552e-08, 0, 0, 0, 5.12047133948845e-08, 0, 3.99942315522929e-08, 0, 0, 3.18459297813314e-08, 3.07098359788831e-08, 5.71114469175672e-08, 7.15386219449749e-08, 0, 0, 4.98096601062656e-08, 0, 3.09065794972147e-08, 0, 1.20487232264969e-07, 7.92473183449994e-08, 6.49982189480215e-08, -2.07305183531121e-08, 8.04604876393846e-08, 2.00575163864598e-07, -6.56391359493823e-08, 1.41190841324457e-07, -2.89071078413752e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -85.3333333330162, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1558.75555555292, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.04532982056463e-08, 0, -1.13370248941116e-08, 0, -10239.9999999963, 0, 0, 0, -1.20088630608998e-08, 0, 0, 0, 0, 0, 0, 0, 0, 1.60276198344419e-08, 0, 0, 0, 3.88928269862084e-08, 0, 4.79875194825135e-08, 1.43821428840359e-08, 30947.5555555428, 0, 0, 0, 5.19674147736652e-08, 4.54931675146023e-08, -2.27243274700165e-08, 0, -4.2515341850567e-08, 0, 3.77661810317885e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.100626832093e-08, 0, 0, 0, -3.92888636920627e-08, 0, -1.96623693593547e-08, -1.38347483395288e-08, -43690.6666666876, 0, 0, 0, 0, -3.08290321116159e-08, 0, 0, -4.65185924743613e-08, 0, 0, 1.56562893400206e-08, 2.84394824193381e-08, 4.19766123099278e-08, 0, 4.95086127983078e-08, -6.86630913568985e-08, 6.42697979897591e-08, 3.71390690226081e-08, 3.99099805545229e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 23301.6888889061, 0, 0, 0, 0, 4.87065353134304e-08, 0, 1.11645589636588e-08, 2.90155488376816e-08, 0, 1.50641869228544e-08, 0, 0, 0, 0, 0, 0, 0, -2.06124064468601e-08, 0, 4.68765776835709e-08, -1.47540257042474e-08, -9.07373621768879e-08, 0, 0, -1.44318682856794e-07, 7.76764026845825e-08, -5.12766374529306e-08, -1.21602023796428e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -85.3333333332818, 0, 0, 0, 0, 0, 0, -1.4606993073127e-08, 0, -1.93390580695553e-08, 0, 0, -3.3928540892042e-08, 0, 0, 1558.75555555349, 0, 0, 0, 0, 0, 0, 3.446421140602e-08, 0, 4.75503611981886e-08, 0, 0, 0, 1.08053390299239e-07, 0, 1.20947324843729e-07, 0, 2.35337817381757e-08, -10239.9999999574, -4.18886484086872e-08, -1.29877402544372e-08, 1.2001388919695e-08, 0, 0, 0, 0, 0, 0, -4.36662031202773e-08, 0, -6.51713905292364e-08, 0, 0, 0, -1.87001732948941e-07, 0, -2.02409896427826e-07, 0, -9.77341793865586e-08, 30947.5555554859, 0, 0, -2.58879283986875e-07, 5.23396313286698e-08, 2.40374011048384e-08, 1.21888335800659e-08, -5.67194850494464e-08, 0, -4.86125950160016e-08, 1.72721491765387e-07, 0, 0, 0, 0, 0, 0, 2.23978949739033e-08, 0, 4.78755668387463e-08, 0, 0, 0, 1.76332236373232e-07, 0, 1.77063060400722e-07, 0, 5.2840550779365e-08, -43690.6666665927, 0, 0, 0, 3.39992259660941e-07, 0, 3.15396552297926e-07, -2.69658409670278e-08, 3.4843502830772e-08, 1.17052299302574e-08, 1.46803601334492e-07, 4.28773597865166e-08, 1.40148372558648e-07, 0, 0, 1.80959795588277e-07, -1.22522266347204e-07, -1.64353555774e-07, -1.71669676583327e-07, 0, 0, 0, 0, 0, 0, 0, 0, -1.32562796452996e-08, 0, 0, 0, -5.75643775916695e-08, 0, -5.0257589897469e-08, 0, -1.62493721813623e-08, 23301.6888888596, 0, 0, 0, -1.01993121454771e-07, 0, -1.17846402208992e-07, 0, 0, -4.13525439458476e-08, -7.60521894941727e-08, -1.92835729237694e-08, -8.54141878882697e-08, 0, 0, -1.58866562470638e-07, 0, 0, -1.0805906163967e-08, -1.10147770422126e-07, 0, -9.07566348857917e-08, 9.97752585816487e-08, 4.2796840519799e-08, 2.74576475046094e-08, 3.19246261859708e-07, 1.58179716613455e-07, 2.48997101898766e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -51.20000000324, 0, 0, 0, 0, 0, 0, 1.26024306261743e-08, 0, 1.35595532950013e-08, 0, 0, 1.45214917508966e-08, 853.33333334377, 409.600000020742, 3.28287506985455e-08, 0, 0, 0, 0, 0, 0, -2.52578907135105e-08, 0, -2.63153436624934e-08, 0, 0, 0, -7.6516576801619e-08, 0, -5.70270466207004e-08, -4778.6666666707, -1.07993109497011e-07, -9.44095063936866e-08, -6826.66666673346, -1.05684647524559e-07, -8.3440371607248e-08, 0, 0, 0, 0, 0, 0, 2.7351424841349e-08, 0, 1.54876386505495e-08, 0, 0, 0, 6.49744894010688e-08, 0, 4.03372544538693e-08, 10922.666666667, 1.3150628268899e-07, 2.16970643492386e-07, 0, 0, 7.6108114569122e-08, 38229.333333438, 5.69509668379689e-08, 2.7393752249021e-07, 4.57248029433227e-07, 2.96460932495651e-07, 2.98483858233826e-07, 3.15303868853551e-07, 0, 0, 0, 0, 0, 0, -1.28771092748591e-08, 0, -2.07629898504232e-08, 0, 0, 0, -5.14219702075155e-08, 0, -1.33173133927513e-07, -8738.13333331674, -1.56781221405029e-07, -1.93257732987216e-07, 0, 0, 0, -1.86551231410605e-07, 0, -2.44659722814619e-07, -87381.3333333637, 0, -2.37289161336431e-07, -3.88235904395147e-07, -3.58502133722205e-07, -4.5626011468663e-07, -1.58573943868424e-07, -3.13712201293407e-07, -4.62199187978896e-07, -3.64647867818977e-07, -2.25550875163376e-07, -3.34811646534206e-07, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.40232355894015e-08, 0, 4.97155525936026e-08, 0, 3.4068800916858e-08, 5.81731468960412e-08, 0, 0, 0, 7.1057856174703e-08, 0, 6.9211251006015e-08, 69905.066666652, -2.6986613701624e-08, 1.11861186027262e-07, 2.23059580787207e-07, 1.72319014797199e-07, 1.76753874623127e-07, 0, 0, 1.65777319068323e-07, 8.23774031426721e-08, 5.65562874646347e-08, 2.01344261241762e-07, 3.13931174529536e-07, 1.87081178630881e-07, 3.6288100241333e-07, 9.59648954566915e-08, 1.06186748567588e-07, 3.02875455611818e-07, 1.56902118997264e-07, 2.14458018988146e-07, 6.86525836203524e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -42.6666666683391, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.48713854311921e-08, 625.777777780073, 512.00000002031, 1.8313845471382e-08, 0, 0, 0, 0, 0, 0, -1.67882557808218e-08, 0, -3.13302659443059e-08, 0, 0, 0, 0, 0, -4.93579200279183e-08, -2730.66666667803, -1365.3333333431, -1.74850442983571e-07, -7509.33333336105, -5.449526674579e-08, -1.71133599979915e-07, 0, 0, 0, 0, 0, 0, 3.37780318721332e-08, 0, 4.8667797073872e-08, 0, 0, 0, 5.37644428101106e-08, 0, 1.18488403595777e-07, 3640.88888886377, 1.43306289211595e-07, 2.37959089255694e-07, 0, 0, 1.14203348994574e-07, 32767.9999999415, 1.66495307952953e-07, 20024.8888888617, 0, 3.91568736044394e-07, 4.1619144439456e-07, 2.85638332708079e-07, 0, 0, 0, 0, 0, 0, -1.77555381831723e-08, 0, -2.15735366528132e-08, 0, 0, 0, -6.00401417083448e-08, 0, 1.39951140858999e-08, 1.66647095335441e-08, 0, -2.43060112704081e-07, 0, 0, 0, -2.07384381293685e-08, 0, 1.38113376155401e-08, -43690.6666666271, -9.76823645754598e-08, -1.24185750805069e-07, 0, -5.71897146787203e-07, -5.32265724340621e-07, -87381.3333334511, -3.54980468518737e-07, -2.79228401736947e-07, -2.20753836316081e-07, -6.4851464532472e-07, -1.29168867906405e-06, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.15356686599378e-08, 0, -1.61470474935707e-08, 0, 1.2629013445864e-08, 8.73524900287052e-08, 0, 0, 0, -6.03657803123688e-08, 0, 1.52981793360885e-08, 0, 0, 4.45928951045846e-08, -6.88923155049364e-08, 2.03276322607179e-07, 2.39459638315801e-07, 0, 0, -9.34294949436486e-08, 116508.444444461, 1.53761652203559e-07, -5.27812688657785e-08, -7.21335348475806e-08, 2.58825898098065e-07, 2.401557991111e-07, -8.4891127988992e-08, 3.062826043354e-07, 9.34502102813145e-07, -1.09759764065405e-07, 5.36418043766793e-07, 1.33357517772168e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -42.6666666656105, 0, 0, 0, 0, 0, 0, -2.5843550059509e-08, 0, -1.77466572206524e-08, 0, 0, -5.25300023842001e-08, 512.000000005764, 625.777777734935, 2.1591431726427e-08, 0, 0, 0, 0, 0, 0, 8.35983182431237e-08, 0, 3.37242451001481e-08, 0, 0, 0, 1.68333692242396e-07, 0, 1.5086357812137e-07, -1365.33333331592, -2730.66666652182, -4.60508844868535e-08, -7509.33333329562, -1.37463328328375e-07, 1.36440283696455e-07, 0, 0, 0, 0, 0, 0, -1.15287095976687e-07, 0, -1.38970872657464e-08, 0, 0, 0, -3.26356499558146e-07, 0, -1.73469666147105e-07, -2.0254601251483e-08, 3640.88888846414, 2.54325989817381e-07, 0, 0, -4.25745501304683e-07, 20024.8888889073, 9.74190320743568e-08, 32767.9999998809, -6.0594481965443e-07, 4.09912028177569e-07, 5.22804589712161e-08, 2.58176438312833e-08, 0, 0, 0, 0, 0, 0, 4.8930079285414e-08, 0, 0, 0, 0, 0, 3.03564353575573e-07, 0, 6.46462065298999e-08, 2.98337290872269e-08, 1.85933285481646e-07, -1.72801629966598e-07, 0, 0, 0, 5.07127104061691e-07, 0, 3.56131924959616e-07, 7.55986244494752e-08, 0, -43690.666666362, 8.97883267763747e-07, -5.12870407444121e-07, -1.24144861533698e-07, -87381.3333331196, -4.47015789868096e-07, 5.35965773714114e-07, 1.06181014821361e-07, 8.69701058435816e-07, -1.343331878932e-07, 0, 0, 0, 0, 0, 0, -1.12897525514908e-08, 0, 0, 0, 0, 0, -1.51489778828093e-07, 0, 0, 0, -5.03446334571899e-08, 5.86033623569911e-08, 0, 0, 0, -8.32696922100054e-08, 0, -1.21105789153353e-07, -5.38831188748333e-08, 0, -1.08259831507451e-07, -3.94427225334344e-07, 3.02913502359331e-07, 1.37274298041397e-07, 0, 0, -1.63124994272311e-07, -1.15941715396986e-07, 1.17463626276154e-07, 116508.444444444, -4.3028633967766e-07, 3.71560125714199e-07, -3.84729438173488e-07, 1.04105023791643e-07, -6.90809735417507e-07, -2.86196145783669e-08, 2.62486897656551e-07, 2.19404528882011e-07, -1.51658341179795e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.10109866824112e-08, 0, 0, -51.1999999965752, 0, 0, 0, 0, 0, 0, -5.48754136485907e-08, 0, -1.58753810573565e-08, 0, 0, -7.75598709927524e-08, 409.599999991782, 853.333333283881, -2.76202606096939e-08, 0, 0, 0, 0, 0, 0, 1.19554394369758e-07, 0, 3.42871327415783e-08, 0, 0, 0, 3.61867881637932e-07, 0, 2.23172958974801e-07, -1.78188732456905e-08, -4778.66666639855, 3.03595596861271e-08, -6826.66666650964, 0, 3.00280905685553e-07, 0, 0, 0, 0, 0, 0, -1.41513360753661e-07, 0, -6.15087896649194e-08, 0, 0, 0, -7.06378775950529e-07, 0, -3.73945696855571e-07, 2.30765283526044e-08, 10922.6666662298, -1.30461604669768e-07, 0, 0, -7.6357541264948e-07, -2.29894386677199e-07, 1.785076391142e-07, 38229.3333327561, -1.00597285297838e-06, 1.13391960803263e-07, -6.80839457171379e-07, -3.27125382471079e-07, 0, 0, 0, 0, 0, 0, 1.12722480410793e-07, 0, 3.63172391334187e-08, 0, 0, 0, 5.80533712893312e-07, 0, 2.74151197372416e-07, -2.25384741668347e-08, -8738.13333280334, 7.18878326218882e-08, 0, 0, 0, 9.30957327463226e-07, 0, 6.42954742021154e-07, 1.283790579042e-07, -2.44664948199951e-07, -87381.3333326061, 1.41702339277041e-06, -1.65676034328484e-07, 5.20495249233423e-07, 4.20626316871904e-07, -3.4396737157862e-07, 1.48426591164884e-06, -2.35129835352253e-07, 8.68031350459725e-07, 1.31321928626787e-08, 0, 0, 0, 0, 0, 0, -2.82968685517774e-08, 0, 0, 0, 0, 0, -1.42596714278503e-07, 0, -9.0710929184126e-08, 0, -1.94362087940266e-07, -2.94452136269023e-08, 0, 0, 0, -5.30733587931461e-07, 0, -2.18067128123185e-07, -3.37099753791141e-08, 6.68195362717964e-08, 69905.0666663111, -6.38397850583637e-07, 0, -2.53584234361125e-07, 0, 0, -4.20487421394845e-07, -1.12048105540837e-07, 2.01214589936931e-07, -3.55138538029161e-07, -1.01448610380251e-06, 1.31532551836703e-07, -3.7221323342158e-07, 0, -6.75501869431115e-07, 2.7929670486014e-07, -6.73949322521067e-07, 2.40775894542873e-07, -4.43276056805337e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -51.1999999998682, 0, 0, 0, 0, 0, 0, 1.08469432909683e-08, 0, 0, 0, 0, 0, -1.06460063078566e-08, 853.333333336118, 409.599999986795, 0, 0, 0, 0, 0, 0, -3.47563544390303e-08, 0, -1.50656686337664e-08, 0, 0, 0, -3.03854564049434e-08, 0, 0, 5.65848460435668e-08, -4778.66666674544, 7.91157373196301e-08, -1.18318828825059e-08, 1.00804473784468e-07, -6826.6666665406, 0, 0, 0, 0, 0, 0, 4.64466613819401e-08, 0, 0, 0, 0, 0, 2.19061930163484e-08, 0, -3.94044516495119e-08, -6.59403471070365e-08, 10922.666666831, -8.01184126875429e-08, 0, 0, -4.99137895482743e-08, -1.44830494061714e-07, -1.90088955304909e-07, 9.42724634328227e-08, 38229.3333333351, -2.79378027534206e-07, -3.4256232499558e-07, -4.43892280277321e-07, 0, 0, 0, 0, 0, 0, -3.19190358371136e-08, 0, 0, 0, 0, 0, 2.5835962037992e-08, 0, -3.21466330467821e-08, 5.24115870532007e-08, -8738.13333358317, 4.69920657346656e-08, 0, 0, 0, 0, 0, -7.70950324938271e-08, 1.52825437519278e-07, 2.75399618070131e-07, -2.09855357653983e-07, -87381.3333334934, 2.44923708873237e-07, 2.52326697528803e-07, 8.61035686025905e-08, 2.61681741774797e-07, 5.31622392165595e-07, 7.01154775801364e-07, 6.28363903572463e-07, 9.68877692658585e-07, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.82731379864566e-08, 0, 2.62905504056106e-08, -1.7985978391786e-08, 8.78068807294845e-08, -1.68116020367906e-08, 0, 0, 0, 0, 0, -1.4196692282788e-08, -8.87263044678487e-08, -8.71697381892534e-08, 1.2531715420175e-07, 69905.0666668024, -6.79137249179355e-08, -6.59978616033129e-08, 0, 0, 2.04409794943914e-08, -4.43165143135743e-08, -1.38409515435928e-07, 7.37723822160787e-08, -1.82610653557106e-07, -9.98328927736464e-08, -2.10888378730213e-07, -2.44774045838505e-07, -2.34989277058621e-07, -3.34479655376869e-07, -3.20492211499106e-07, -5.00028286196174e-07, -3.20838965042084e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -42.6666666684203, 0, 0, 0, 0, 0, 0, 0, 0, 3.19294851981914e-08, 0, 0, 3.6523493533483e-08, 1.21865452791501e-08, 625.777777806644, 512.000000015088, 0, 0, 0, 0, 0, 0, -1.9258431198637e-08, 0, -6.73618157535849e-08, 0, 0, 0, -1.67208166251488e-08, 0, -1.46821367114556e-07, -7.87762800953425e-08, -2730.66666673986, -1365.33333347534, 0, -6.66128668597305e-08, -7509.33333343112, 0, 0, 0, 0, 0, 0, 4.27203432222983e-08, 0, 1.09571680124873e-07, 0, 0, 0, -2.81836066424332e-08, 0, 3.57317061566944e-07, 7.98665041221279e-08, 3640.88888914411, 2.17333844108895e-07, 0, 0, 1.30942062925424e-07, 2.21595266499733e-07, 8.49437877669568e-08, 2.30425579322451e-07, 32768.0000000482, 2.92289765872692e-07, 20024.8888891674, 1.70619656347802e-07, 0, 0, 0, 0, 0, 0, -1.82045515827529e-08, 0, -7.73966475349438e-08, 0, 0, 0, -7.29748768587954e-08, 0, -3.24577779968044e-07, -6.22253003807132e-08, -1.0551198421359e-07, -2.23125203027134e-07, 0, 0, 0, -1.39235234792304e-08, 0, -1.81815999394419e-07, -2.19560929328969e-07, -2.38960952717613e-07, -3.28169202216069e-07, -43690.6666666557, -3.46901402620237e-07, -4.46037587629684e-07, -4.45893643503286e-07, -1.98715536813407e-07, -87381.3333336377, -7.37425927157916e-07, -5.44371656915e-07, -7.11332668317109e-07, 0, 0, 0, 0, 0, 0, 0, 0, 1.86816568188006e-08, 0, 0, 0, 1.46925230732562e-08, 0, 8.99599935989902e-08, 2.20788624478463e-08, 3.42134626892577e-08, 7.92736457347545e-08, 0, 0, 0, -9.74443788782275e-08, 0, 1.92840035111361e-07, 1.27342694517045e-07, 5.87248652975986e-08, 1.77991592028543e-07, 4.13910633032983e-08, 1.26556938589805e-07, 2.2536582864959e-07, 0, 0, 2.57422879577027e-08, 1.7765961058707e-07, 9.55665242668804e-08, 2.11021784520212e-07, 116508.44444447, 8.72504534473136e-08, 2.69069668714771e-07, 8.87578212345797e-08, 4.25504746520146e-07, 3.70036379255531e-07, 2.75780565889065e-08, 2.62098237239999e-07, -3.25346038708761e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -42.6666666668461, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.32539966806701e-08, 0, 511.999999983035, 625.777777777517, 0, 0, 0, 0, 0, 0, 1.89011059711562e-08, 0, -4.14699917497398e-08, 0, 0, 0, 3.77779811920643e-08, 0, -3.54420414806693e-08, 3.18922515205264e-08, -1365.33333330762, -2730.66666662708, -6.69189512438381e-08, 0, -7509.33333327582, 0, 0, 0, 0, 0, 0, -3.98608638285828e-08, 0, 3.5663580297285e-08, 0, 0, 0, -8.80608065143796e-08, 0, 1.08918635494503e-07, 0, -1.4071757754624e-07, 3640.88888892409, 0, 0, 2.07100108312941e-08, -3.47272702977214e-08, 6.21569181661087e-08, -7.00824520887947e-08, 20024.8888888461, -7.87011084239756e-08, 32767.9999999717, 7.93532638002827e-08, 0, 0, 0, 0, 0, 0, 1.25758955456572e-08, 0, -2.05975488231915e-08, 0, 0, 0, 1.95687668135636e-07, 0, -1.60761897653634e-07, 0, -1.76956372420165e-08, -3.74034459796092e-08, 0, 0, 0, 1.3426171910384e-07, 0, -2.49539960254928e-07, -1.11028059343219e-08, 1.09826576001543e-07, 1.13832001993961e-07, 3.38993833387998e-08, 2.02962480668129e-08, -43690.6666668457, 1.74449383389917e-07, -8.49371552404153e-08, -87381.3333330497, 4.09223028798364e-07, 3.08184105156e-07, 3.03469179251102e-07, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.81237107795666e-08, 0, 8.78645229132323e-08, 0, 1.36935929365099e-08, 1.26971504592977e-08, 0, 0, 0, -1.99551662696329e-08, 0, 7.6140795285307e-08, -3.42331478470426e-08, -2.88740191699983e-08, -8.57709346034655e-08, -3.04071865816225e-08, 2.57993493539295e-08, 1.24749288734455e-07, 0, 0, 2.14968971887263e-08, -2.54048109798839e-08, 2.97315375297572e-08, -8.55482016782195e-08, -1.18103365250368e-07, 9.43252254547927e-08, 116508.444444438, 7.12436917744795e-08, -2.78029192355639e-07, 0, 1.75045882764253e-07, 1.7152602829887e-08, 2.29494628455529e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -51.1999999987818, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 409.599999997096, 853.33333333204, 0, 0, 0, 0, 0, 0, 1.75433263897064e-08, 0, 1.63807005518239e-08, 0, 0, 0, 1.3576776305093e-08, 0, 0, 0, 4.37549773307329e-08, -4778.6666667018, 9.60273862248465e-08, 0, -6826.66666667541, 0, 0, 0, 0, 0, 0, -1.74541983536825e-08, 0, -1.11790799256934e-08, 0, 0, 0, -5.63725497871417e-08, 0, 1.5803252019347e-08, -2.06050825551449e-08, -2.53499657016781e-08, 10922.6666666642, 0, 0, -1.54160453129723e-08, -8.37146828778021e-08, -8.18837347930592e-08, -1.35639708921209e-07, -1.12070992815229e-07, 8.48055033720097e-08, 38229.3333334228, -1.90168257450411e-07, 0, 0, 0, 0, 0, 0, 2.25116241651819e-08, 0, 0, 0, 0, 0, 0, 0, 0, 1.41206823785523e-08, 1.12065866560382e-07, -8738.13333333306, 0, 0, 0, -2.79675926861378e-08, 0, 3.7234216899454e-08, 8.40226616087402e-08, 0, 1.61739494079539e-07, 2.11821463185775e-07, -7.31027488175731e-08, -87381.3333334166, 1.0760074353716e-07, 2.93856172060028e-08, -2.31362937780124e-07, -1.0088505594692e-07, 1.09949188867766e-07, -2.33856212040318e-07, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.85735459089162e-08, 0, -3.90136683511334e-08, 0, 0, 0, 0, -3.46724809715658e-08, 0, 2.240336935288e-08, 0, -1.14514769494385e-08, -4.0501422735331e-08, -9.71158695804613e-08, 0, 69905.0666666921, 0, 0, 1.37693339670559e-08, -5.28900115247208e-08, -2.76746356263874e-08, -6.86393663238311e-08, 2.80981262672519e-08, -1.16229262344538e-08, 1.229481737076e-07, -1.52825093292118e-07, -7.33575310152253e-08, 1.07772766790852e-07, -3.025730254394e-07, 0, -7.52822942800525e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -51.1999999984265, 0, 0, 0, 0, 0, 0, 0, 0, -1.22971647160989e-08, 0, 0, 0, 409.600000000736, -2.16008164136217e-08, 853.333333319181, 0, 0, 0, 0, 0, 0, 0, 0, 3.35561833107941e-08, 0, 0, 0, -2.25610843521755e-08, 0, 0, -2.77415395010285e-08, 1.00501352578531e-07, -4778.66666663589, 5.81742820578047e-08, -6826.66666672326, 9.99814787778643e-08, 0, 0, 0, 0, 0, 0, -1.47181627832686e-08, 0, -6.62637290516361e-08, 0, 0, 0, 8.2699785234241e-08, 0, -1.02782908818822e-07, 1.58133776942757e-08, -1.63977217887328e-07, 10922.6666665444, 0, 0, 1.76429689008305e-08, -4.37967623307236e-08, 2.03694510608584e-07, -1.93927645149931e-07, -2.46315796066326e-07, 38229.3333333921, -3.37977114088851e-07, 2.03107460014292e-07, 0, 0, 0, 0, 0, 0, 2.13474341008552e-08, 0, 3.9544487672184e-08, 0, 0, 0, -7.97014321275215e-08, 0, 1.01920163664408e-07, -1.80306816979685e-08, 2.04442777770726e-07, -8738.13333318196, 0, 0, 0, -1.83928332097766e-07, 0, -1.01078228782336e-07, -2.45371491032046e-08, -2.53210118242209e-07, 3.82416441452965e-07, 2.1973762708594e-07, -87381.3333334152, 3.62517383968425e-07, 5.42942750075827e-08, -3.60331262692034e-07, 3.65618805352439e-07, -4.06109472307292e-07, -2.01849359438134e-08, -1.77493893486553e-07, 0, 0, 0, 0, 0, 0, 0, 0, -1.07056718971076e-08, 0, 0, 0, 2.22719768206125e-08, 0, -5.88832383521358e-08, 0, -8.26003183431028e-08, -5.71502880551088e-08, 0, 0, 0, 0, 0, -4.67240489012229e-08, 0, 6.90365186215779e-08, -1.50861181996813e-07, -1.29159480913717e-07, 69905.0666666467, -1.87187192935627e-07, 0, 0, 4.59287193024193e-08, 0, 1.75348843979766e-07, -1.24030787737524e-07, -1.84269413378229e-07, 1.62269272830282e-07, -2.93219451034303e-07, 1.61507281589222e-07, -1.7136189339186e-07, 1.39317336033427e-07, -2.01164894753473e-07, 3.38151691005644e-07, -8.45850499654005e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -42.6666666664154, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 512.000000011222, -1.28692255758894e-08, 625.777777780913, 0, 0, 0, 0, 0, 0, 2.53805606388413e-08, 0, 3.55315954829779e-08, 0, 0, 0, -6.70779935166749e-08, 0, 0, -1365.33333335166, 8.13063793538734e-08, -2730.66666659721, -6.30436479586782e-08, -7509.33333339209, -2.04109191926342e-08, 0, 0, 0, 0, 0, 0, -3.49642596757474e-08, 0, -4.03722243786317e-08, 0, 0, 0, 5.27174073220708e-08, 0, 2.74473765311116e-08, 7.1683952474132e-08, -2.4302989067077e-07, 3640.88888878334, 0, 0, 2.51574703670938e-07, 1.68522450700129e-07, 20024.8888890829, -6.95588119473716e-08, -1.29612115841398e-07, 32768.0000001979, 1.73563888786813e-07, -1.83178387453147e-07, 0, 0, 0, 0, 0, 0, 1.85331749635863e-08, 0, 5.01004051705769e-08, 0, 0, 0, 2.63170184386836e-08, 0, 7.47597791086283e-08, -2.77090387801012e-08, 1.60349146857995e-07, 9.04857381351712e-08, 0, 0, 0, -1.41015118822097e-07, 0, -3.57875353490951e-08, -1.47602581762095e-07, -1.12428343327402e-07, 3.78619666430788e-08, 3.60393633160057e-07, -43690.6666666751, -3.9776243878421e-08, 7.18298525502063e-08, -87381.3333335511, 0, -2.54819177427042e-07, 3.77418619181977e-07, -3.0218944771194e-07, 0, 0, 0, 0, 0, 0, 0, 0, -1.46557313644322e-08, 0, 0, 0, -1.07908791546013e-08, 0, 0, 0, -4.27291376470229e-08, -3.80176680755418e-08, 0, 0, 0, 1.70193287623093e-07, 0, 7.10314642264949e-08, 4.23439578949664e-08, 6.61983755672412e-08, -1.03005996888982e-07, -1.47910085430942e-07, 0, -1.5514024417546e-08, 0, 0, 1.37856190778278e-07, 2.46559454749726e-08, 1.88228413743244e-07, -1.47152006723662e-08, 1.34856140737283e-08, 116508.444444545, 2.12643431294634e-07, 2.90571973448654e-07, -2.1160029789982e-07, -1.09362757059211e-07, 6.22337422337017e-07, 3.34857890332777e-07, 1.75771865047265e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -42.6666666658555, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.38548574287008e-08, 625.777777773369, 0, 511.999999993991, 0, 0, 0, 0, 0, 0, 0, 0, 1.59357680945146e-08, 0, 0, 0, -1.11087963804316e-08, 0, -6.84854838915588e-08, -2730.66666668451, 2.16875100432287e-08, -1365.33333337793, 6.49552362980703e-08, -7509.33333332557, 2.70126320581731e-08, 0, 0, 0, 0, 0, 0, -2.50517668203146e-08, 0, -5.03612441131873e-08, 0, 0, 0, 5.4993038255453e-08, 0, 0, 3640.88888886757, -2.55038968011964e-08, 0, 0, 0, 6.38270465382307e-08, -5.63059109706081e-08, 32767.9999999905, -4.21015059558327e-08, 1.15270891567283e-08, 20024.8888887871, -5.94139626288522e-08, 4.05361394029903e-07, 0, 0, 0, 0, 0, 0, 3.4198152332759e-08, 0, 2.11207753098869e-08, 0, 0, 0, -8.5827596863398e-08, 0, -3.2268629596604e-08, 0, 1.23157576857273e-07, 4.96629668282126e-08, 0, 0, 0, -2.56955064290463e-07, 0, -2.51150542340509e-07, 5.33222462052257e-08, -43690.6666667978, 1.98631534560399e-07, 1.08631475132825e-08, -2.15471926981166e-08, -5.32699390689502e-08, -1.30741363185414e-07, -87381.333333428, -2.53050070943571e-07, -2.40985969482263e-07, -1.96549389249027e-07, 3.30236926960323e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.07928263108195e-08, 0, -2.46729108919871e-08, 0, -6.06230902987353e-08, -2.20993877073555e-08, 0, 0, 0, -3.29492821948012e-08, 0, 1.74743691706631e-08, 0, 2.36485542162499e-08, -2.83010424667578e-08, -7.45555430592937e-08, -4.43701199639318e-08, 0, 0, 0, 9.34323518311906e-08, 4.82710538194843e-08, 116508.444444433, 1.09386865419143e-08, 0, 0, -1.60949169077764e-08, -5.94447724107518e-08, -1.05613777891397e-07, 3.46882457972767e-07, -6.27737179208995e-07, 4.84074866707466e-08, -6.6748057143223e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -51.2000000032017, 0, 0, 0, 0, 0, 0, 0, 0, 2.20320056067927e-08, 0, 0, 2.24566153850503e-08, 853.333333339101, 2.70784629751821e-08, 409.600000030997, 0, 0, 0, 0, 0, 0, 0, 0, -4.59452384518044e-08, 0, 0, 0, -7.89121107554649e-08, 0, -9.36206969417837e-08, -4778.66666664644, -8.82507562788891e-08, -9.12673295982925e-08, -9.05140783752333e-08, -6826.66666673174, -2.04876162348559e-07, 0, 0, 0, 0, 0, 0, -1.00286169623283e-08, 0, 5.21247353504696e-08, 0, 0, 0, 2.63045664783412e-08, 0, 1.97026271245037e-07, 10922.6666666428, 1.24415100141344e-07, 1.93423484779321e-07, 0, 0, 2.36649615277502e-07, 1.36243502363265e-08, 38229.3333333856, 2.20743043648248e-07, 4.87661844518124e-07, 2.83265086978538e-07, 6.39424876399684e-07, 7.6967739351095e-08, 0, 0, 0, 0, 0, 0, 1.21860025040551e-08, 0, -2.43531033026292e-08, 0, 0, 0, 4.42219284393563e-08, 0, -1.70208986175902e-07, -8738.13333328954, -1.04347356681422e-07, -1.91370271924696e-07, 0, 0, 0, -1.44662611549784e-07, 0, -3.00553861764735e-07, 4.59730981307272e-08, -87381.3333332601, -3.43587921931594e-07, -3.69574160389802e-07, -2.45672460388923e-07, -7.72309545949606e-07, -4.15178871210758e-08, -1.09778512557302e-07, -7.50598149639399e-07, -1.25922009805406e-07, -4.20167761435452e-07, -8.99685000574114e-07, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6.47406084479641e-08, -1.45836745712699e-08, 2.39144261486102e-08, 5.78917619564565e-08, 0, 0, 0, 7.06015396936257e-08, 0, 1.49642728997113e-07, -5.44671122189061e-08, 69905.0666666358, 1.09153023810635e-07, 1.70988021941376e-07, 1.21980126087545e-07, 3.18672834952368e-07, 0, 0, 1.55257422891163e-07, 0, 5.24225929270535e-08, 1.17683356084312e-07, 4.14517703713632e-07, 8.98878678642401e-08, 7.32853126745305e-07, 2.37104065949412e-08, 3.29152924165799e-07, 4.46779264386438e-07, 3.92536864996582e-07, 2.43899035655819e-07, 8.06633984680968e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -32.0000000029492, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.63504256572233e-08, 469.333333354227, 256.000000028883, 256.000000035454, 0, 0, 0, 0, 0, 0, 0, 0, -2.5307526046863e-08, 0, 0, 0, -8.43678546823419e-08, 0, -9.52941362520509e-08, -2048.00000004152, -9.6734407475202e-08, -1.00495539360657e-07, -3754.66666682973, -3754.66666679301, -2048.00000034233, 0, 0, 0, 0, 0, 0, 0, 0, 7.66155195601008e-08, 0, 0, 0, 2.17998290836216e-07, 0, 2.82349828326796e-07, 2730.66666672402, 1.21489633205075e-07, 2.13794920146422e-07, 0, 0, 3.61122800907636e-07, 16384.000000309, 16384.0000001902, 3.54075792997854e-07, 5.63064781122316e-07, 2.37258795831843e-07, 8.94280420260972e-07, 30037.3333341386, 0, 0, 0, 0, 0, 0, -3.00040165977624e-08, 0, -4.21399193402956e-08, 0, 0, 0, -1.59126238031292e-07, 0, -1.47865099038665e-07, -4.30267730670053e-08, -1.64373342168886e-07, -1.2444116587165e-07, 0, 0, 0, -2.1606780620529e-07, 0, -1.72374814822579e-07, -21845.3333335689, -21845.3333334642, -4.35035022964432e-07, -7.55615494451422e-07, -2.41046788967717e-07, -7.14439449862147e-07, -4.07713546109732e-07, -1.60736174991816e-07, -1.36827219332421e-06, -131072.000000458, -1.42827148645288e-06, -1.28646341921887e-06, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.88468178032778e-08, 1.29251679319543e-08, 7.84989299683306e-08, 4.79546822319981e-08, 0, 0, 0, 1.90596370706502e-07, 0, 1.24482450893206e-07, 9.24854721790018e-08, 7.19778044465203e-08, 1.72744465307249e-07, 2.01244098625845e-07, 1.17220296848952e-07, 3.09391411501543e-07, 0, 0, 9.97120810146811e-08, 9.6887943894763e-08, 7.66197587168495e-08, 1.86043715165951e-07, 6.54626301139111e-07, 1.20628848751747e-07, 3.55393664418499e-07, 174762.666666988, 7.76863796297512e-07, 3.08502493936567e-07, 1.09742391051498e-06, 6.20593400618156e-07, 6.37328147556555e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -31.9999999991159, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 255.999999994981, 469.33333332085, 256.000000019931, 0, 0, 0, 0, 0, 0, 2.72787775897452e-08, 0, -2.52298541811005e-08, 0, 0, 0, 0, 0, -7.01983521136433e-08, 5.32180047546504e-08, -2047.99999994087, -9.6017175190308e-08, -3754.66666665403, -2048.00000004936, -3754.66666683891, 0, 0, 0, 0, 0, 0, -4.22445702511786e-08, 0, 7.1832934402151e-08, 0, 0, 0, 5.85469050402164e-08, 0, 2.53331711128682e-07, -1.06494162600254e-07, 2730.66666650829, 2.16591034436647e-07, 0, 0, 2.4667058385494e-07, -1.37712844682359e-07, 1.54259079130442e-08, 16383.9999998711, 16384.000000005, 2.45179644665975e-07, 6.03482922544656e-07, 30037.333333688, 0, 0, 0, 0, 0, 0, 0, 0, -3.86497830285165e-08, 0, 0, 0, 0, 0, -1.30942985698336e-07, 9.86299267941731e-08, 7.86885272449507e-08, -1.54713515035043e-07, 0, 0, 0, -4.55220366189989e-08, 0, -1.42352888018187e-07, 2.52088830617571e-07, 9.22311291011027e-08, -21845.3333332296, -21845.3333333282, -3.35078006172902e-07, -5.41935236709195e-07, 2.55913096751112e-07, -1.3434948539186e-07, -6.99653031419469e-07, 3.55788675368908e-07, -131072.000000551, -7.86242087625354e-07, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.15330811498605e-08, 0, 4.81202443045819e-08, -3.46594664792721e-08, 0, 5.9030351820549e-08, 0, 0, 0, 1.05426366390932e-07, 0, 1.03990826401671e-07, -1.11331634641009e-07, -3.82142885947956e-08, -4.30150851010838e-08, -1.04174135426813e-07, 1.84411641851513e-07, 2.77847286237451e-07, 0, 0, 2.76655894937612e-08, -2.25413394574017e-07, 3.08549301988324e-08, -1.59844773642737e-07, 2.19947442715688e-07, 1.89994267690086e-07, 1.0370615788769e-07, -1.21170354542242e-07, 174762.666666916, 2.66730830294655e-07, 4.3451271397178e-07, 2.33331446355449e-07, 2.9327869515414e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -31.9999999987646, 0, 0, 0, 0, 0, 0, 0, 0, -1.46929926680183e-08, 0, 0, 0, 255.999999997912, 255.999999994419, 469.333333321152, 0, 0, 0, 0, 0, 0, -2.22702391548e-08, 0, 4.41172774675565e-08, 0, 0, 0, -5.27173479495102e-08, 0, 3.59199201809948e-08, 0, 1.07680126182751e-08, -2047.99999994918, -2047.99999999494, -3754.66666664943, -3754.66666660079, 0, 0, 0, 0, 0, 0, 3.10801435094023e-08, 0, -5.90690378369105e-08, 0, 0, 0, 1.36894715247422e-07, 0, -1.11629261440431e-07, -1.09328316089576e-08, -2.50655725022202e-08, 2730.66666656877, 0, 0, 6.92569475731975e-08, 2.70228144642437e-08, -1.17032310686776e-08, 0, -1.31741541926016e-07, 16383.999999979, 16383.9999998667, 30037.3333332502, 0, 0, 0, 0, 0, 0, -2.22568285428279e-08, 0, 3.81086795861422e-08, 0, 0, 0, -1.4354561010037e-07, 0, 1.21763156888341e-07, 1.20326529872389e-08, 0, 1.09713799727955e-07, 0, 0, 0, -1.64628358391159e-07, 0, 6.81691194849756e-08, -1.03382499920066e-08, 3.93515273219229e-08, -2.89268495742743e-08, 1.40048756508821e-07, -21845.3333332945, -21845.3333330914, -1.09406623942891e-07, -6.7897493475576e-08, 1.68241817376578e-07, 2.40083403895026e-07, 3.33639309078887e-07, -131071.999999827, 0, 0, 0, 0, 0, 0, 0, 0, -1.00057029898457e-08, 0, 0, 0, 3.73349885719602e-08, 0, -5.39503348461455e-08, 0, 0, -4.19930386326231e-08, 0, 0, 0, 1.10963373998251e-07, 0, -5.73495834953613e-08, 0, -2.22078450029604e-08, 1.61437687583345e-08, -4.01434304752336e-08, -1.91062470840667e-08, -1.19318756117383e-07, 0, 0, 4.78924671492151e-08, 5.2789075568694e-08, 2.07162915782426e-08, 6.51514028999863e-08, -8.67111061732402e-08, 3.96131437895678e-08, -1.21423569706651e-07, -6.78393373755652e-08, -1.48028504210412e-07, 174762.666666629, -1.15225780266311e-07, -6.78857457302757e-08, -3.17926952376122e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -28.4444444481692, 0, 0, 0, 0, 0, 0, 1.12148467408323e-08, 0, 2.34519361682559e-08, 0, 0, 1.79511371294262e-08, 341.333333347782, 341.333333363576, 227.555555572257, 0, 0, 0, 0, 0, 0, -3.58923827436648e-08, 0, -3.66800009398321e-08, 0, 0, 0, -3.42612378762876e-08, 0, -8.17737579736516e-08, -910.222222283656, -910.222222326685, -8.03226839266101e-08, -4096.00000007128, -2730.66666673526, -2730.66666662629, 0, 0, 0, 0, 0, 0, 5.09641539468434e-08, 0, -1.68677761034322e-08, 0, 0, 0, -1.21067506459914e-07, 0, 2.12423542964488e-08, 1.07009590740051e-07, 2.09265704117285e-07, 9.54937506763094e-08, 0, 0, -3.23088568822879e-08, 10922.6666668266, 7281.77777794658, 10922.6666669651, 7281.77777803315, 1.99232923762211e-07, -5.63263875623205e-08, 32768.0000000676, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.01161023851494e-07, 0, -1.51728749815086e-07, -8.48669351150811e-08, -8.81932447797706e-08, -1.35928614442199e-07, 0, 0, 0, -4.31266773828168e-08, 0, -3.23940733877682e-07, -2.11093779559049e-07, -2.27861826901917e-07, -2.89867126396057e-07, 2.34899713037842e-08, -2.50525787170873e-07, -3.95703416798161e-07, -29127.1111115309, -3.49162307083649e-07, 1.59494607202864e-07, -87381.3333339784, -87381.3333332445, 2.47998842715724e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.87209575797583e-08, 0, 4.40477494298954e-08, 2.63579828374112e-08, 0, 2.53253177947028e-08, 0, 0, 0, -1.64091898136578e-07, 0, 9.95361475677639e-08, 6.87103217205552e-08, 6.58921594543586e-08, 1.04813758589134e-07, 7.78751724729688e-08, 6.80489399011232e-08, 1.34196788439678e-07, 0, 0, 1.29839960664844e-07, 2.83707840204844e-07, 1.43207126401399e-07, 2.85005459287979e-07, -1.4971033291404e-07, 9.70131209604474e-08, 5.26115788269748e-07, 1.40416302828974e-07, -3.94746256768973e-08, 5.0254073130518e-07, 233016.888888547, 5.19092742856332e-08, 4.28432804825909e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -28.4444444466553, 0, 0, 0, 0, 0, 0, 0, 0, 2.14039223283057e-08, 0, 0, 3.36574240393806e-08, 227.555555566239, 341.333333350134, 341.333333355994, 0, 0, 0, 0, 0, 0, -1.51952098410608e-08, 0, -5.08849212461034e-08, 0, 0, 0, -9.13269851079366e-08, 0, -1.37744737985158e-07, -2.78479371982079e-08, -910.222222263967, -910.222222314349, -2730.66666674599, -2730.66666674065, -4096.0000001393, 0, 0, 0, 0, 0, 0, 1.68073811049343e-08, 0, 7.02711370984023e-08, 0, 0, 0, 1.2252103552981e-07, 0, 2.65588667707338e-07, 1.85167223128432e-08, 6.25633381082437e-08, 1.82839907850119e-07, 0, 0, 2.95781285047806e-07, 1.72499530385743e-07, 9.20840100038584e-08, 7281.7777779831, 10922.6666669063, 7281.77777798898, 10922.6666671394, 32768.0000002614, 0, 0, 0, 0, 0, 0, 0, 0, -4.50395821405258e-08, 0, 0, 0, -8.07444414041932e-08, 0, -2.274919744755e-07, 0, -4.12276279609196e-08, -1.53555263757854e-07, 0, 0, 0, -2.84264107973761e-07, 0, -3.93904490775161e-07, -1.42878468984945e-07, -6.88530274284722e-08, -2.0276633622646e-07, -2.3351995885835e-07, -2.86320161729457e-07, -5.79459959848703e-07, -2.8847416827479e-07, -2.07697767764842e-07, -29127.111111736, -4.58894637832637e-07, -87381.3333338577, -87381.3333340162, 0, 0, 0, 0, 0, 0, 0, 0, 1.18028840204396e-08, 0, 0, 0, 2.53027774706699e-08, 0, 7.17291346969888e-08, 0, 0, 5.20877631747704e-08, 0, 0, 0, 8.42921607309718e-08, 0, 1.83193473967878e-07, 4.29857483054666e-08, 1.80664899884392e-08, 7.91842503694235e-08, 5.39563407248275e-08, 1.1398688493569e-07, 2.54027029230815e-07, 0, 0, 1.85510753058254e-07, 1.39156338837936e-07, 8.03556393843019e-08, 1.62289790910396e-07, 2.56361054873406e-07, 1.52777311572541e-07, 4.16095026427444e-07, 1.4692550359999e-07, 2.2936174184561e-07, 3.54797731326837e-07, 3.3278995439467e-07, 3.36094675862903e-07, 233016.888889292}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -28.4444444449889, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.29039176353549e-08, 341.333333338419, 227.555555560462, 341.333333333044, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.92277568713259e-08, 0, -4.75008961759451e-08, -910.222222240566, -1.08058998260286e-08, -910.222222217187, -2730.66666671051, -4096.000000003, -2730.66666670637, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.04293893529699e-07, 0, 5.15430279060029e-08, 3.07829323734302e-08, 1.28023410688823e-08, -3.56232096697413e-08, 0, 0, 1.49965180278027e-07, 7281.77777788179, 10922.6666666926, 9.84800202341361e-08, 8.78852765432598e-08, 10922.666666666, 7281.77777783651, 32768.0000000666, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.67382666347594e-08, 0, -1.09936625701144e-08, -2.49113451196443e-08, 1.20974612591174e-08, 4.08220830264787e-08, 0, 0, 0, -1.98672682940557e-07, 0, -1.8357163339162e-07, -1.28331218466012e-07, -2.62547721631896e-08, -7.44870196177379e-08, -7.88431152805294e-08, 3.84882844598249e-08, -8.26944061887294e-08, -2.043572769697e-07, -29127.1111110976, -1.60518062111206e-07, -87381.3333334631, -1.47659084411321e-07, -87381.3333334478, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.41438264272213e-08, 0, 0, 0, 0, -1.7400373644545e-08, 0, 0, 0, 7.61627251567597e-08, 0, 4.26307615252466e-08, 3.91685396474146e-08, 0, 2.41413639650037e-08, 2.9391949028309e-08, -2.43802228001031e-08, 2.34081658397374e-08, 0, 0, 8.99014174838537e-08, 1.10864358341594e-07, 0, 1.10855739723833e-07, 1.20906894880192e-07, 0, 9.17513422247649e-08, 4.70059412702238e-08, 2.80178951109333e-08, 3.99017377726809e-08, 8.32115232544084e-08, 233016.888888905, 1.578044287475e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 1791.99999999943, 0, 0, 0, 0, 0, 0, 0, -17459.1999999867, 0, -17459.1999999939, -17459.2000000297, 0, 0, 0, 0, 0, 0, 69142.755555397, 0, 69142.7555556764, 0, 0, 138285.511110964, 69142.7555559365, 138285.511111062, 138285.511111703, 0, 0, 0, 0, 0, 0, -142677.333332937, 0, -142677.333333797, 0, 0, 0, -428031.999997737, 0, -428032.000001393, -142677.333334053, -428031.99999884, -428032.000003652, -428032.000002187, -428032.000004289, -856064.00000046, 0, 0, 0, 0, 0, 0, 162019.555555294, 0, 162019.555556118, 0, 0, 0, 648078.222216019, 0, 648078.222227454, 162019.555555644, 648078.222220014, 648078.22222955, 0, 0, 972117.33333223, 648078.222227452, 648078.222229347, 972117.333337916, 1944234.6666553, 972117.333350051, 1944234.66667921, 1944234.66667755, 0, 0, 0, 0, 0, 0, -96119.4666667487, 0, -96119.4666669257, 0, 0, 0, -480597.333327441, 0, -480597.333338866, -96119.4666660299, -480597.333332375, -480597.333339093, 0, 0, 0, -961194.666659381, 0, -961194.666671971, -480597.333337046, -480597.333336667, -961194.666671711, -1922389.3333112, -961194.666688205, -1922389.33335775, -961194.666677526, -961194.666687519, -2883583.9999976, -1922389.33334761, -2883584.00000104, -2883584.00003497, 0, 0, 0, 0, 0, 0, 23301.6888889843, 0, 23301.6888889217, 0, 0, 0, 139810.133331516, 0, 139810.133335174, 23301.6888885393, 139810.133333495, 139810.133334866, 0, 0, 0, 349525.333328645, 0, 349525.33333748, 139810.133334006, 139810.133333257, 349525.33333576, 699050.66665621, 349525.333342079, 699050.666678784, 0, 0, 466033.777777303, 349525.33333856, 349525.333340651, 466033.777784306, 1398101.33332025, 466033.777792057, 1398101.33334292, 699050.666670088, 1398101.33332781, 1398101.3333603, 1398101.33334046, 1398101.33335436, 2097152.0000097}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -85.3333333332541, 0, 0, 0, 0, 0, 0, 0, 1644.0888888893, 0, 85.3333333321896, 85.333333333114, 0, 0, 0, 0, 0, 0, -11798.755555556, 0, 0, 0, 0, -1558.75555555893, 0, -1558.75555555619, 1.02115712758071e-08, 0, 0, 0, 0, 0, 0, 41187.5555555534, 0, -3.16303617372186e-08, 0, 0, 0, 10240.000000013, 0, 1.44788188023353e-08, 0, 10240.0000000201, -5.25408352216424e-08, 2.4289502675025e-08, -2.04990025114366e-08, -6.1355469934696e-08, 0, 0, 0, 0, 0, 0, -74638.2222222186, 0, 5.11966920545019e-08, 0, 0, 0, -30947.5555555664, 0, 0, -2.48637594352948e-08, -30947.5555556335, 1.02486668633142e-07, 0, 0, -7.41245230990819e-08, -6.0303398505665e-08, 0, -5.58846456108275e-08, 2.16516932185701e-07, 4.8117273301944e-08, 2.53410343301017e-07, 9.32544544564687e-08, 0, 0, 0, 0, 0, 0, 66992.3555555532, 0, -3.99056990250684e-08, 0, 0, 0, 43690.6666666536, 0, -4.02848214704401e-08, 1.96646356216282e-08, 43690.6666667708, -8.72346895953838e-08, 0, 0, 0, 6.08165651371118e-08, 0, 7.58062608717332e-08, 9.86864385231438e-08, 0, 1.18643269337446e-07, -2.19314165327145e-07, -1.1859304028085e-07, -3.25219524270152e-07, 1.31971939749504e-07, -2.72871906219545e-08, -4.58168458487297e-07, 7.2475389311401e-08, -3.24374096509998e-07, -3.01029012352746e-07, 0, 0, 0, 0, 0, 0, -23301.6888888888, 0, 1.13774779227544e-08, 0, 0, 0, -23301.6888888776, 0, 2.04445337696292e-08, 0, -23301.6888889362, 3.11489495596069e-08, 0, 0, 0, -1.07606005726042e-08, 0, -1.97817813461902e-08, -3.49403899399863e-08, 0, -5.92263813842592e-08, 4.62363248514291e-08, 4.98171015261933e-08, 1.24302249543547e-07, 0, 0, -3.93216367368536e-08, -4.83139587874754e-08, 2.76543619499793e-08, -5.29140723698528e-08, 2.64584285025902e-07, 3.23916816358263e-08, 2.92920672883491e-07, -7.30434924192593e-08, 1.42805300056977e-07, 1.77912599705016e-07, -2.52052477562983e-08, -2.35819281961307e-08, 2.70540989117091e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -85.3333333323972, 0, 0, 0, 0, 0, 0, 0, 85.3333333250476, 0, 1644.08888887475, 85.3333333322183, 0, 0, 0, 0, 0, 0, 2.73457758496372e-08, 0, -11798.7555554806, 0, 0, -1558.7555554441, 0, 0, -1558.75555553467, 0, 0, 0, 0, 0, 0, -4.52228674063448e-08, 0, 41187.5555553683, 0, 0, 0, -3.09356564809197e-07, 0, 10239.9999995154, -1.12082183895269e-08, 1.71771352697772e-08, 10239.9999998881, 4.88513696383508e-08, 0, -5.70096807870104e-08, 0, 0, 0, 0, 0, 0, 4.13375565836411e-08, 0, -74638.2222219776, 0, 0, 0, 3.92166758119141e-07, 0, -30947.5555546367, 2.92451657963887e-08, -2.79091196968517e-08, -30947.5555552995, 0, 0, 1.04367035529354e-06, -5.83102131069747e-08, 3.28443247968039e-08, -1.35758781624941e-07, 1.947588401629e-08, 4.91953385468556e-08, 3.62391715368386e-07, -2.3430170905675e-07, 0, 0, 0, 0, 0, 0, -2.08221955233287e-08, 0, 66992.3555553972, 0, 0, 0, -2.37194467444958e-07, 0, 43690.666665866, -2.92772063804134e-08, 1.22641652389712e-08, 43690.6666664153, 0, 0, 0, -9.35601124396148e-07, 0, -1.34691455279682e-06, 0, -9.2360008193669e-08, 1.32301666852422e-07, 7.87037841122448e-08, -1.2006427825844e-07, -6.18485655620269e-07, 1.40035330319789e-07, -2.80465242852227e-08, -4.10022580535892e-07, 1.76579729801336e-07, 5.10228512431845e-07, 1.75976224984133e-07, 0, 0, 0, 0, 0, 0, 0, 0, -23301.688888848, 0, 0, 0, 5.57402098877135e-08, 0, -23301.6888886296, 1.03308713207477e-08, 0, -23301.6888887989, 0, 0, 0, 2.94918073393406e-07, 0, 5.85564426600705e-07, 0, 4.89784103622766e-08, -2.85839223832495e-08, -6.40732427825595e-08, 7.15020604996616e-08, 3.04815483638633e-07, 0, 0, 6.05781492894444e-07, -3.37230799867408e-08, 4.22860319227384e-08, -6.56038906461273e-08, 1.34287633544293e-07, 1.87785710213548e-08, 4.06829263734835e-07, 0, -3.48835366864848e-07, -1.46441541209769e-08, -2.8290757372947e-07, -1.73212300354278e-07, -2.94216682427921e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -1792.00000000239, 0, 0, 0, 0, 0, 0, 0, 27016.5333333137, 0, 12680.5333333747, 12680.5333333327, 0, 0, 0, 0, 0, 0, -136772.266666444, 0, -35328.0000005513, 0, 0, -172100.266666237, -35328.0000001428, -172100.26666638, -70656.0000001992, 0, 0, 0, 0, 0, 0, 331093.333332988, 0, 48469.333335267, 0, 0, 0, 710655.999995122, 0, 428032.000001621, 48469.3333334709, 710655.999998859, 145408.000002959, 428032.000000888, 145408.000001959, 856063.999995937, 0, 0, 0, 0, 0, 0, -420522.666666931, 0, -32768.0000028294, 0, 0, 0, -1294335.99998828, 0, -518826.666674957, -32767.9999994333, -1294335.99999993, -131072.000007743, 0, 0, -1359871.99999351, -518826.666669338, -131072.000002996, -1359872.00000352, -2719743.99997573, -196608.000011659, -1556480.00000227, -1556479.9999975, 0, 0, 0, 0, 0, 0, 270882.133334114, 0, 8738.13333522752, 0, 0, 0, 1092266.66665634, 0, 305834.666676608, 8738.13333228237, 1092266.66666942, 43690.6666738574, 0, 0, 0, 1660245.33331589, 0, 1135957.33333651, 305834.666668359, 43690.6666668119, 1660245.33334062, 3320490.6666317, 87381.3333510828, 1223338.66668471, 1135957.33334187, 87381.3333474383, 3407871.99997327, 1223338.66666707, 3407871.99998107, 1835008.00001121, 0, 0, 0, 0, 0, 0, -69905.0666670418, 0, -4.82649129423349e-07, 0, 0, 0, -349525.333330338, 0, -69905.0666703763, 4.64765761729773e-07, -349525.333335302, -2.21923855820305e-06, 0, 0, 0, -699050.666657094, 0, -349525.333338264, -69905.0666666806, 9.98027368828457e-07, -699050.666671374, -1398101.33331909, -8.04529330426684e-06, -349525.33334513, 0, 0, -699050.666662133, -349525.333337209, -4.54054849631135e-06, -699050.666673147, -2097151.99997187, -1.14166597550817e-05, -1398101.33333246, -349525.333331888, -2097151.99998646, -699050.666683078, -1398101.33333119, -699050.666674632, -2097151.99998933}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 1493.33333333109, 0, 0, 0, 0, 0, 0, 0, -25500.4444443889, 0, -7580.44444442137, -7580.44444440004, 0, 0, 0, 0, 0, 0, 153031.11111079, 0, 14279.1111112134, 0, 0, 119523.55555477, 14279.11111102, 119523.555555064, 28558.2222217489, 0, 0, 0, 0, 0, 0, -425756.444443891, 0, -11832.8888895046, 0, 0, 0, -620771.55555061, 0, -206847.99999905, -11832.8888886628, -620771.555554233, -35498.6666660101, -206847.999999489, -35498.6666659235, -413695.999993544, 0, 0, 0, 0, 0, 0, 602567.111111009, 0, 3640.88888987567, 0, 0, 0, 1354410.6666562, 0, 156558.222223873, 3640.88888835286, 1354410.66666661, 14563.555555818, 0, 0, 904760.888879507, 156558.222222192, 14563.5555548828, 904760.888888998, 1809521.77775342, 21845.3333343769, 469674.666655694, 469674.666656475, 0, 0, 0, 0, 0, 0, -422343.111111547, 0, -6.66096103036446e-07, 0, 0, 0, -1310719.99999111, 0, -43690.6666698286, 6.81920555839247e-07, -1310720.00000261, -6.07006977935243e-07, 0, 0, 0, -1398101.33331635, 0, -553415.111106925, -43690.6666666321, 1.29149996917042e-06, -1398101.33333698, -2796202.66663636, -3.10242905294234e-06, -174762.666663012, -553415.111113788, -2.68165518381946e-06, -1660245.33329697, -174762.66666036, -1660245.33330782, -262143.999988546, 0, 0, 0, 0, 0, 0, 116508.444444701, 0, 1.65518277961158e-07, 0, 0, 0, 466033.777775202, 0, 1.33438930817466e-06, -2.81348054253012e-07, 466033.777779572, 1.18759009098531e-07, 0, 0, 0, 699050.666658249, 0, 116508.444444994, -3.1388912449409e-07, -8.51304339715966e-07, 699050.666669815, 1398101.33332192, 1.6610752450221e-06, 1.24997087533015e-06, 0, 0, 466033.777771554, 116508.444445555, 7.8254178201242e-07, 466033.777780601, 1398101.3333059, 3.09117819066336e-06, 466033.77776538, -2.33695681114455e-06, 1398101.33331869, -8.55085633679245e-07, 466033.77777107, -2.43746416490676e-06, 699050.666646114}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -895.999999997537, 0, 0, 0, 0, 0, 0, 0, 16255.9999999696, 0, 3114.66666663326, 3114.66666663738, 0, 0, 0, 0, 0, 0, -106609.777777627, 0, -3583.99999987723, 0, 0, -52849.7777773511, -3583.99999992353, -52849.7777775443, -7167.99999958877, 0, 0, 0, 0, 0, 0, 329727.999999733, 0, 1365.33333309875, 0, 0, 0, 308565.33333125, 0, 56661.3333322376, 1365.33333321774, 308565.333332899, 4095.99999859364, 56661.3333329712, 4095.99999896078, 113322.666663128, 0, 0, 0, 0, 0, 0, -515185.777777674, 0, 2.98161996352739e-07, 0, 0, 0, -784611.555551464, 0, -20024.8888876818, 6.78903477187324e-08, -784611.555556045, 2.25849939239846e-06, 0, 0, -289450.666662134, -20024.8888887381, 1.2192087561894e-06, -289450.666666405, -578901.333322132, 2.26262046048648e-06, -60074.6666580242, -60074.6666602111, 0, 0, 0, 0, 0, 0, 393216.000000127, 0, -2.06368202359899e-07, 0, 0, 0, 873813.333329886, 0, -7.48163770557475e-07, -7.14363920378722e-08, 873813.333335047, -1.89402062380366e-06, 0, 0, 0, 567978.666659722, 0, 87381.3333298302, 2.43966064047425e-07, -1.02392657845739e-06, 567978.666668216, 1135957.33332068, -2.63814573199184e-06, -7.86233089276332e-06, 87381.3333340929, -1.66521857305438e-06, 262143.999980704, -4.03919186205587e-06, 262143.999986113, -1.09350390791488e-05, 0, 0, 0, 0, 0, 0, -116508.444444534, 0, 5.73569582353006e-08, 0, 0, 0, -349525.333332318, 0, 1.96963587925904e-07, 3.40349614499455e-08, -349525.333334338, 6.68546836113503e-07, 0, 0, 0, -349525.333330059, 0, 8.82269827653441e-07, -4.80649273094343e-08, 3.97759463287988e-07, -349525.333334751, -699050.666662302, 1.06616632345849e-06, 2.50392652910886e-06, 0, 0, -116508.44444152, -3.13959829238956e-07, 6.95097503557214e-07, -116508.444445305, -349525.333320804, 7.40416290427687e-07, 8.91625371342558e-06, 9.93468965224814e-07, -349525.333326045, 4.74464311323711e-06, 4.05788783725139e-06, 3.27926334308841e-06, 1.21065514009869e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 358.39999999929, 0, 0, 0, 0, 0, 0, 0, -6741.33333332879, 0, -767.99999998976, -767.999999993698, 0, 0, 0, 0, 0, 0, 46660.2666666425, 0, 409.599999942347, 0, 0, 13619.1999999276, 409.599999990445, 13619.1999999655, 819.199999881144, 0, 0, 0, 0, 0, 0, -154965.333333284, 0, 1.58967806698427e-07, 0, 0, 0, -85333.3333330202, 0, -6826.66666644228, 0, -85333.3333333544, 5.04573583175973e-07, -6826.66666667044, 3.12326394958969e-07, -13653.3333325153, 0, 0, 0, 0, 0, 0, 263236.266666639, 0, -2.39439728147588e-07, 0, 0, 0, 240298.666666052, 0, -4.207214059381e-07, 8.7052437958297e-08, 240298.666667055, -9.36928850866922e-07, 0, 0, 38229.3333326966, 1.62056997826806e-07, -3.23905431950959e-07, 38229.3333334502, 76458.6666642754, -8.58239012854425e-07, -2.4899074552699e-06, -1.51515212557885e-06, 0, 0, 0, 0, 0, 0, -218453.333333347, 0, 1.72125610093649e-07, 0, 0, 0, -305834.666666102, 0, 4.56702922834961e-07, -7.90744686256818e-08, -305834.666667332, 8.15514116285721e-07, 0, 0, 0, -87381.3333323295, 0, 5.97897269847113e-07, -3.57162098793596e-07, 2.46537919065644e-07, -87381.3333339711, -174762.666664178, 1.26263557726912e-06, 2.85418330770328e-06, -4.76785710460546e-07, 7.16674098161802e-07, 4.64166954053768e-06, 6.50697571757839e-07, 3.37245243209987e-06, 3.29978529873396e-06, 0, 0, 0, 0, 0, 0, 69905.0666666807, 0, -4.72516870796556e-08, 0, 0, 0, 139810.133333149, 0, -1.6737075549274e-07, 2.31880551846644e-08, 139810.133333664, -2.81396364258021e-07, 0, 0, 0, 69905.0666661699, 0, -2.2270797487419e-07, 1.39023737920898e-07, -7.79014555891567e-08, 69905.0666670957, 139810.133332609, -5.39605429451416e-07, -1.09529789569772e-06, 0, 0, -4.11753401057758e-07, 1.95181862569004e-07, -3.10474795363383e-07, 2.67262857663184e-07, -2.75220964075683e-06, -4.75189481526197e-07, -2.56887743563857e-06, 3.17880371662462e-08, -1.62156371442264e-06, -1.76847611929644e-06, -7.5372124181131e-07, -8.00929536345208e-07, -3.04782168926845e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -51.2000000000297, 0, 0, 0, 0, 0, 0, 0, 904.533333335207, 0, 460.799999998184, 51.1999999991897, 0, 0, 0, 0, 0, 0, -5632.00000001372, 0, -409.599999988593, 0, 0, -8089.59999998509, 0, -853.333333353477, -409.599999956161, 0, 0, 0, 0, 0, 0, 15701.3333333823, 0, -2.8402898325943e-08, 0, 0, 0, 49834.6666666064, 0, 6826.66666659377, -1.22125454219162e-08, 4778.66666682698, -2.02939953616492e-07, 9.52947240603726e-08, -1.48549560831001e-07, 6826.66666645816, 0, 0, 0, 0, 0, 0, -19660.8000000909, 0, 4.49070957772579e-08, 0, 0, 0, -136533.33333324, 0, 1.2527209356913e-07, 0, -10922.666667112, 3.95718382610897e-07, 0, 0, -38229.3333331058, -1.79005064073695e-07, 2.18813937185195e-07, -4.30495769366284e-07, -38229.3333328404, 4.90386495432209e-07, 7.10130620774127e-07, 3.72070326866374e-07, 0, 0, 0, 0, 0, 0, 8738.13333341427, 0, -2.78925852639617e-08, 0, 0, 0, 166024.533333283, 0, -1.23304177635544e-07, 0, 8738.13333386515, -3.64308418096948e-07, 0, 0, 0, 87381.333332967, 0, -2.48532362903925e-07, 1.45286537493248e-07, -1.93929745272462e-07, 7.71448826984639e-07, 87381.3333329133, -6.67891001510405e-07, -8.15167894912163e-07, 5.17941694560729e-07, -5.31474279559085e-07, -1.2652919212914e-06, -1.85253008191624e-07, -6.15416639549759e-07, -1.10962684459356e-06, 0, 0, 0, 0, 0, 0, -2.70630830596373e-08, 0, 0, 0, 0, 0, -69905.0666666667, 0, 4.09271519917067e-08, 0, -2.2376195959642e-07, 1.27284012106467e-07, 0, 0, 0, -69905.0666665021, 0, 8.50951192043346e-08, -4.0423006337428e-08, 6.0904881030935e-08, -4.14508735286691e-07, -69905.0666666131, 2.92974716642396e-07, 3.11039506588627e-07, 0, 0, 1.65009585327573e-07, -1.87244106347697e-07, 2.12209712839253e-07, -3.63509174407064e-07, 7.88798303214128e-07, 3.38505184805057e-07, 6.75240283587434e-07, 0, 1.90632241117413e-07, 6.05087926299908e-07, 6.09519722158163e-08, 4.20436135167793e-07, 9.0336167769078e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -42.6666666667, 0, 0, 0, 0, 0, 0, 0, 668.444444444989, 0, 554.666666664402, 42.6666666656074, 0, 0, 0, 0, 0, 0, -3356.44444445318, 0, -1877.33333333282, 0, 0, -8647.11111107702, 0, -625.777777768788, -511.999999984128, 0, 0, 0, 0, 0, 0, 6371.55555560574, 0, 1365.33333335374, 0, 0, 0, 43007.9999998058, 0, 28899.5555554853, 3.82710986806388e-08, 2730.66666659586, 1365.33333331171, 9.59015279977336e-08, 1.72070618304501e-08, 7509.33333314802, 0, 0, 0, 0, 0, 0, -3640.88888900164, 0, -5.28229371054279e-08, 0, 0, 0, -80099.5555551727, 0, -20024.888888945, -1.39491610907122e-08, -3640.88888878805, 0, 0, 0, -140174.222221678, -3.28658398467187e-07, -7.59862869782917e-08, -2.95680695003439e-07, -32767.9999990042, -1.80843858959078e-08, -20024.8888887927, -1.38397748454394e-08, 0, 0, 0, 0, 0, 0, 1.07619059988635e-07, 0, 3.28419404669566e-08, 0, 0, 0, 43690.6666664048, 0, 1.98083717469571e-07, -1.07363931188064e-08, -1.91432155445703e-08, -2.77677090087315e-08, 0, 0, 0, 247580.444443612, 0, 87381.3333330381, 3.38445168563404e-07, 9.96818073619272e-08, 2.93647094828903e-07, 43690.6666651738, 5.04099515384316e-08, 5.10856699715241e-07, 6.72417966684494e-07, 1.42791247755423e-07, 87381.3333321787, 5.03606932840768e-07, -2.10459356151757e-07, 2.79367165011863e-07, 0, 0, 0, 0, 0, 0, -3.67936450450359e-08, 0, 0, 0, 0, 0, 4.71486802883014e-08, 0, -8.43112686374924e-08, 0, -2.39402753595415e-08, 3.29690498393612e-08, 0, 0, 0, -116508.444444056, 0, 0, -9.7548546052328e-08, 0, -1.72866871652857e-07, 6.89076470647094e-07, 2.36492043387788e-08, -3.34731859728795e-07, 0, 0, -116508.444444039, -3.63220108630281e-07, -8.1237373013788e-08, -4.30752390363162e-07, -116508.444443281, -7.11243664068762e-08, 0, -2.30555213025421e-07, 3.90110576735502e-07, -3.04429551408928e-07, -4.26325669954424e-07, -2.52229179752929e-07, 3.52563365188408e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -42.666666665831, 0, 0, 0, 0, 0, 0, 0, 554.666666659017, 0, 668.44444443253, 42.6666666609438, 0, 0, 0, 0, 0, 0, -1877.33333330144, 0, -3356.44444439323, 0, 0, -8647.1111110061, 2.29543396692237e-08, -511.999999941866, -625.777777749033, 0, 0, 0, 0, 0, 0, 1365.33333327119, 0, 6371.55555545175, 0, 0, 0, 28899.5555551576, 0, 43007.9999996784, -1.20038738475468e-07, 1365.33333309613, 2730.66666660574, -9.64365535673218e-08, 1.01511467589495e-08, 7509.33333307716, 0, 0, 0, 0, 0, 0, 5.48579946146013e-08, 0, -3640.88888876176, 0, 0, 0, -20024.8888882075, 0, -80099.5555551477, 2.3816455582196e-07, 4.82147648418247e-07, -3640.88888878335, 0, 0, -140174.222221306, 3.05238552967676e-07, 5.04827800091842e-08, 2.6185820388462e-07, -20024.8888879057, -2.72810272484596e-07, -32767.9999996985, 4.78445468436295e-08, 0, 0, 0, 0, 0, 0, -1.88755112513312e-08, 0, -7.168455336456e-08, 0, 0, 0, -5.35992367145611e-07, 0, 43690.666666398, -2.31149767763214e-07, -4.26834881007014e-07, -9.26652065163584e-08, 0, 0, 0, 87381.3333321154, 0, 247580.444443673, -5.04057484835472e-07, -3.05341007154284e-07, -2.6857054224029e-07, -1.58284228802042e-06, 3.41125756796782e-07, 43690.6666665159, -2.59980860752935e-07, 3.54386054737112e-07, 87381.3333324112, -2.70922773957863e-07, -3.58069738487782e-07, 1.2495701811757e-07, 0, 0, 0, 0, 0, 0, 0, 0, 1.72790542277079e-08, 0, 0, 0, 1.55064700194717e-07, 0, 6.23471127655864e-08, 8.22456080956293e-08, 1.35882116554091e-07, 3.18798198919593e-08, 0, 0, 0, 5.14828700490708e-07, 0, -116508.444444249, 2.32965752719334e-07, 1.7329795701848e-07, 1.5944443462145e-07, 7.48392545349616e-07, -1.63200752253955e-07, 0, 0, 0, -116508.444443956, 2.2855421871288e-07, -4.55317955624927e-08, 1.30386049441065e-07, 9.44238706691664e-07, -3.32037211314969e-07, -116508.444444282, 3.41068915431808e-07, 2.82848589627586e-07, -2.38547756724575e-07, -1.34637352239005e-08, -1.22204963657891e-07, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -51.1999999994788, 0, 0, 0, 0, 0, 0, 0, 460.799999997299, 0, 904.533333323034, 51.2000000026651, 0, 0, 0, 0, 0, 0, -409.599999997463, 0, -5631.99999994615, 0, 0, -8089.59999994062, 0, -409.600000028655, -853.333333371702, 0, 0, 0, 0, 0, 0, 0, 0, 15701.3333332029, 0, 0, 0, 6826.66666654432, 0, 49834.6666663887, -1.80130848998716e-08, 7.76398166954271e-08, 4778.66666681313, 4.97017774041937e-08, 2.00591731399665e-07, 6826.66666694627, 0, 0, 0, 0, 0, 0, 0, 0, -19660.7999998462, 0, 0, 0, 1.47275768588212e-07, 0, -136533.333332799, 9.51341590084721e-08, -8.97160916173064e-08, -10922.6666669227, 0, 0, -38229.3333328583, 6.26336872513906e-08, -1.91831258988032e-07, -1.7909786776004e-07, -6.1755821305364e-07, -6.85745023184201e-07, -38229.3333341178, -8.65424958176468e-07, 0, 0, 0, 0, 0, 0, 0, 0, 8738.13333323796, 0, 0, 0, -9.98982842017689e-08, 0, 166024.533332881, -9.26097111593363e-08, 3.14783663163365e-08, 8738.13333353273, 0, 0, 0, -3.37305639076908e-07, 0, 87381.3333327431, -1.56146513583105e-07, 1.11221645176655e-08, 1.26667311888625e-07, 5.65314355972314e-07, 8.98948186477947e-07, 87381.3333342463, 0, 7.36624955267809e-07, 1.15350569464186e-06, 5.43541373894091e-07, 1.40126772011314e-06, 2.10461823022473e-06, 0, 0, 0, 0, 0, 0, 0, 0, 2.24161567419585e-08, 0, 0, 0, 2.95231278917178e-08, 0, -69905.066666515, 3.04642481273964e-08, 0, -6.04093139934281e-08, 0, 0, 0, 1.11169768323028e-07, 0, -69905.0666664058, 8.23517542256927e-08, 5.66972183337048e-08, -4.37339081098165e-08, -1.43610751121027e-07, -3.6210610295076e-07, -69905.0666669938, 0, 0, 2.36637440300006e-07, 5.72927452401717e-08, -2.10213751035599e-07, -5.22025630268282e-08, -5.11047462629578e-07, -5.2119640040398e-07, -6.79247488740144e-07, 8.73560035091484e-08, -6.49261771546611e-07, -1.23382928167217e-06, -4.56263828306527e-07, -9.49528728384622e-07, -1.41735954339458e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 358.399999995332, 0, 0, 0, 0, 0, 0, 0, -767.999999959029, 0, -6741.33333326258, -767.999999996269, 0, 0, 0, 0, 0, 0, 409.599999855972, 0, 46660.2666662926, 0, 0, 13619.1999994724, 409.600000013229, 819.199999989948, 13619.199999951, 0, 0, 0, 0, 0, 0, 2.73539035194243e-07, 0, -154965.33333239, 0, 0, 0, -6826.66666518937, 0, -85333.3333310779, 3.48469458707411e-08, 8.5476368415619e-08, -85333.3333329963, -1.92859363697537e-07, -6826.66666695359, -13653.3333333044, 0, 0, 0, 0, 0, 0, -3.00586041230367e-07, 0, 263236.26666543, 0, 0, 0, -2.02602409721044e-06, 0, 240298.666662388, -1.69501814648372e-07, -3.15925682809755e-07, 240298.666665763, 0, 0, 38229.3333285873, 1.63798672356151e-07, 3.01388282649499e-07, 3.37131425604294e-07, 0, 38229.3333340605, 76458.6666659027, 2.20152167913469e-06, 0, 0, 0, 0, 0, 0, 1.81213138135537e-07, 0, -218453.333332521, 0, 0, 0, 1.38124531877049e-06, 0, -305834.666662918, 1.74629123741131e-07, 4.12515434063883e-07, -305834.666665678, 0, 0, 0, 4.31018159626855e-06, 0, -87381.3333273733, 1.36324116060834e-07, 1.51401076438285e-07, 2.86759191551128e-08, 1.131775271488e-07, -87381.3333339053, -174762.666664889, -2.9551316428992e-07, -1.01210072863271e-06, 8.90057212153176e-07, -2.02609621679838e-06, -3.52116493357496e-06, -3.75712015772962e-06, 0, 0, 0, 0, 0, 0, -4.63678419804473e-08, 0, 69905.0666664551, 0, 0, 0, -3.76416355291198e-07, 0, 139810.133332094, -5.92021144206674e-08, -1.75312075305198e-07, 139810.133332955, 0, 0, 0, -1.44683888224227e-06, 0, 69905.0666640803, -1.07643889538509e-07, -1.89188254365633e-07, -1.86111532342926e-07, -1.37767004397429e-07, 69905.0666667488, 139810.133332258, 0, 0, -2.68226235250144e-06, 0, 2.00539731669501e-07, 3.7421294526747e-08, -4.70156091627659e-07, 6.04802151289472e-07, -1.09759599633273e-06, 2.31338982192433e-07, 1.67585444081973e-06, 1.68487872903333e-06, 1.92763840115922e-06, 2.41533926995424e-06, 3.06631898504152e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -895.999999986839, 0, 0, 0, 0, 0, 0, 0, 3114.66666655291, 0, 16255.9999998042, 3114.66666664353, 0, 0, 0, 0, 0, 0, -3583.9999995762, 0, -106609.777776752, 0, 0, -52849.7777763956, -3584.00000000657, -7167.99999977391, -52849.7777776418, 0, 0, 0, 0, 0, 0, 1365.3333324516, 0, 329727.999997403, 0, 0, 0, 56661.3333294121, 0, 308565.333327628, 1365.33333313603, 4095.99999875489, 308565.333332544, 4095.99999999281, 56661.3333345281, 113322.666666208, 0, 0, 0, 0, 0, 0, 1.05461509443437e-06, 0, -515185.777774347, 0, 0, 0, -20024.888883187, 0, -784611.555544841, 7.18283217358738e-07, 2.97673350012358e-06, -784611.555553276, 0, 0, -289450.66665496, 3.89004951037314e-07, -20024.8888905407, 1.15151660235912e-06, -60074.6666641618, -289450.666670451, -578901.333331958, -60074.6666728608, 0, 0, 0, 0, 0, 0, -6.75803708245832e-07, 0, 393215.999997725, 0, 0, 0, -4.19021629837572e-06, 0, 873813.333323902, -7.39317741485184e-07, -3.04579578059208e-06, 873813.333330666, 0, 0, 0, 87381.3333224435, 0, 567978.666652514, -1.18949782312821e-06, -4.46817920603218e-08, -2.97751525620312e-06, -4.8274344980927e-06, 567978.666670328, 1135957.33333002, -9.81878159866841e-07, 87381.333338815, 262143.999997249, 6.19944256185635e-06, 7.08635338621358e-06, 262144.000013606, 0, 0, 0, 0, 0, 0, 1.79381727312157e-07, 0, -116508.444443848, 0, 0, 0, 1.22960416260839e-06, 0, -349525.333330178, 2.53507159704134e-07, 1.11542174899867e-06, -349525.333332269, 0, 0, 0, 3.84658802018795e-06, 0, -349525.333327272, 6.31316258852161e-07, 5.51091403684076e-07, 1.82774946352731e-06, 2.72864501340144e-06, -349525.333334321, -699050.666664344, 0, 0, -116508.444438118, 6.46063004746892e-07, -1.59070700045049e-06, 1.0956413712388e-06, 2.71437345015209e-06, -116508.444447849, -349525.333331468, -5.60465190692963e-07, -1.76675997631265e-06, -349525.333340346, -4.53883893526316e-06, -8.99408889909367e-06, -9.02288669572628e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 1493.3333333142, 0, 0, 0, 0, 0, 0, 0, -7580.44444428563, 0, -25500.4444441612, -7580.44444439276, 0, 0, 0, 0, 0, 0, 14279.1111104963, 0, 153031.111109659, 0, 0, 119523.555553678, 14279.1111111574, 28558.2222215964, 119523.555555353, 0, 0, 0, 0, 0, 0, -11832.888887555, 0, -425756.444440786, 0, 0, 0, -206847.999994448, 0, -620771.555548303, -11832.8888887391, -35498.6666632321, -620771.555554758, -35498.666666459, -206848.000002721, -413695.999998582, 0, 0, 0, 0, 0, 0, 3640.88888726899, 0, 602567.111106243, 0, 0, 0, 156558.222213547, 0, 1354410.66665361, 3640.88888783099, 14563.5555478764, 1354410.666664, 0, 0, 904760.888873933, 14563.5555556475, 156558.22222687, 21845.3333297441, 469674.666657402, 904760.888898304, 1809521.77777786, 469674.666677228, 0, 0, 0, 0, 0, 0, 1.03435265292732e-06, 0, -422343.111107843, 0, 0, 0, -43690.6666599407, 0, -1310719.99998864, 1.28031749862934e-06, 7.30754035611996e-06, -1310719.9999964, 0, 0, 0, -553415.111096154, 0, -1398101.33331689, 1.0015324114353e-06, -43690.6666678638, 7.58442631806473e-06, -174762.666649833, -1398101.33334318, -2796202.66666693, 1.01220541760983e-06, -553415.11112537, -1660245.33332781, -174762.666679662, -262144.000006495, -1660245.33336139, 0, 0, 0, 0, 0, 0, -2.72287268996901e-07, 0, 116508.444443577, 0, 0, 0, -2.02569287888286e-06, 0, 466033.777773943, -4.73481386444877e-07, -2.50262606789333e-06, 466033.777776203, 0, 0, 0, 116508.444438726, 0, 699050.666660042, -7.42127238923285e-07, -7.59687075833835e-07, -4.20514202807946e-06, -8.77052668820216e-06, 699050.666669687, 1398101.33333244, 0, 0, 466033.777770096, -1.45381659250503e-07, 116508.444449263, -1.84115726844546e-06, 466033.777769718, 466033.777786686, 1398101.33333345, 2.11419222815622e-06, -2.4847441591824e-06, 1398101.33334976, 7.36181975250843e-06, 466033.777797214, 699050.666681291}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -1791.99999999048, 0, 0, 0, 0, 0, 0, 0, 12680.5333332749, 0, 27016.5333331763, 12680.5333333108, 0, 0, 0, 0, 0, 0, -35327.9999997681, 0, -136772.266665913, 0, 0, -172100.266665825, -35328.0000002225, -70655.9999996243, -172100.266666759, 0, 0, 0, 0, 0, 0, 48469.3333328422, 0, 331093.333331467, 0, 0, 0, 428031.999996921, 0, 710655.999997676, 48469.333333629, 145407.999997393, 710656.000001013, 145408.000001312, 428032.0000037, 856063.999999703, 0, 0, 0, 0, 0, 0, -32767.9999995275, 0, -420522.666664087, 0, 0, 0, -518826.666660666, 0, -1294335.99999719, -32767.9999993259, -131071.999994238, -1294336.00000127, 0, 0, -1359871.99999426, -131072.00000373, -518826.666673032, -196608.000000951, -1556479.99998944, -1359872.00001369, -2719744.00000832, -1556480.00001293, 0, 0, 0, 0, 0, 0, 8738.13333315005, 0, 270882.133331494, 0, 0, 0, 305834.666661547, 0, 1092266.66666463, 8738.13333207488, 43690.6666617673, 1092266.66666646, 0, 0, 0, 1135957.33332467, 0, 1660245.33333076, 43690.6666689074, 305834.666668742, 87381.33333192, 1223338.66664546, 1660245.33334924, 3320490.66668212, 87381.3333401894, 1135957.3333527, 3407871.99999938, 1223338.66668334, 1835008.00000635, 3407872.00003694, 0, 0, 0, 0, 0, 0, 1.77891330038836e-08, 0, -69905.0666661473, 0, 0, 0, -69905.0666651328, 0, -349525.333332569, 5.36245502879686e-07, 1.40876601976982e-06, -349525.333332764, 0, 0, 0, -349525.333329034, 0, -699050.666666767, -1.9793215654212e-07, -69905.0666658797, 1.11046229511507e-06, -349525.333322602, -699050.666672272, -1398101.33334019, 0, 0, -699050.666664168, -3.70130305148163e-06, -349525.333340047, -3.17703334978918e-06, -1398101.33332332, -699050.666679324, -2097152.00000844, -349525.33333673, -699050.666662868, -2097152.00002489, -699050.666676538, -1398101.33335707, -2097152.00001604}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 1120.00000000544, 0, 0, 0, 0, 0, 0, 0, -14645.3333333969, 0, -14645.3333333756, -5685.33333334071, 0, 0, 0, 0, 0, 0, 56192.0000002782, 0, 56192.0000004379, 0, 0, 184064.000000126, 10709.333333407, 66901.3333334311, 66901.3333335298, 0, 0, 0, 0, 0, 0, -94549.3333342156, 0, -94549.3333348579, 0, 0, 0, -647509.333331801, 0, -647509.333335872, -8874.66666644932, -197973.333333666, -197973.333335616, -112298.666668172, -112298.666668, -759807.999999427, 0, 0, 0, 0, 0, 0, 73728.0000016779, 0, 73728.0000022584, 0, 0, 0, 980309.333328724, 0, 980309.333342042, 2730.66666539585, 223914.666668323, 223914.666672367, 0, 0, 1813162.66666618, 81920.0000029469, 81920.0000010606, 229376.00000557, 2042538.66665653, 229376.000008134, 2042538.66667547, 1144149.33333607, 0, 0, 0, 0, 0, 0, -21845.3333348605, 0, -21845.3333348533, 0, 0, 0, -677205.333329396, 0, -677205.333343187, 1.61346835527817e-06, -87381.333336458, -87381.3333384503, 0, 0, 0, -1922389.33332569, 0, -1922389.33334066, -21845.3333348201, -21845.3333312899, -131072.000008353, -2053461.33331565, -131072.00001235, -2053461.33335543, -87381.3333426532, -87381.333342679, -3932159.99999569, -742741.333335808, -2097151.99999835, -2097152.00001588, 0, 0, 0, 0, 0, 0, 5.12501466861976e-07, 0, 3.86214812815397e-07, 0, 0, 0, 174762.666665714, 0, 174762.666670292, -6.29359431813641e-07, 1.70147064511379e-06, 1.53148805748256e-06, 0, 0, 0, 699050.666661875, 0, 699050.66667259, -2.22500838134681e-08, -1.87093174855581e-06, 4.49381982249766e-06, 699050.666658935, 5.43956992763934e-06, 699050.66667904, 0, 0, 1048575.99999931, 4.07252415085496e-06, 2.42997516623193e-06, 6.5842182679233e-06, 2097151.99998735, 8.0235299897745e-06, 2097152.00000923, 174762.666664879, 1048575.99999699, 1048576.0000159, 699050.66667026, 699050.666674297, 2097152.00000218}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 224.00000000105, 0, 0, 0, 0, 0, 0, 0, -3765.33333334874, 0, -2272.00000000165, -479.999999997049, 0, 0, 0, 0, 0, 0, 21632.0000000949, 0, 4096.00000002605, 0, 0, 37674.6666666356, 255.999999983151, 7552.00000007649, 4351.99999984972, 0, 0, 0, 0, 0, 0, -53589.3333336638, 0, -2048.00000011328, 0, 0, 0, -211285.333332991, 0, -64170.6666666573, 6.40011750813028e-08, -38229.333333925, -4095.99999947298, -3754.66666706783, -2047.99999944918, -67925.3333323887, 0, 0, 0, 0, 0, 0, 57344.0000006047, 0, 1.75952654606894e-07, 0, 0, 0, 502442.666665958, 0, 30037.3333337697, -1.69161415921948e-07, 73728.0000017764, -8.08287185485797e-07, 0, 0, 322218.666665579, 7.93916309476799e-07, -9.48064895009167e-07, 16384.0000020655, 338602.666663656, -1.51375526324905e-06, 60074.6666644946, 30037.333331528, 0, 0, 0, 0, 0, 0, -21845.3333338616, 0, -1.15218138001129e-07, 0, 0, 0, -502442.666666245, 0, -6.64574160517694e-07, 1.89659954294675e-07, -43690.666668895, 7.26998538067537e-07, 0, 0, 0, -611669.333331169, 0, -131071.99999953, -6.11240380430222e-07, 8.88511222917707e-07, -21845.3333368763, -633514.666663427, 1.66638361527786e-06, 1.01724059916257e-06, -2.62537554502064e-06, 1.68120975686697e-06, -262143.999994117, 1.08968137669123e-06, -131071.999997279, 3.73672884841592e-06, 0, 0, 0, 0, 0, 0, 1.73531168469934e-07, 0, 2.71609697976394e-08, 0, 0, 0, 174762.666666646, 0, 2.67431336335338e-07, -7.0284637896996e-08, 9.66592145363627e-07, -2.83809554218414e-07, 0, 0, 0, 349525.333332275, 0, 1.5305093673909e-07, 1.18822535519938e-07, -3.58680886186577e-07, 1.9891655924169e-06, 349525.3333324, -7.1228435594466e-07, 4.77776454202271e-08, 0, 0, 174762.666665727, 1.03655824591606e-06, -7.12534931231048e-07, 2.02643742655815e-06, 349525.333328979, -8.69570516376438e-07, -1.9619532159398e-06, -3.54514974344746e-07, 174762.666665614, -1.43588150659678e-06, -1.60975575217396e-07, -1.69677126322174e-06, -3.97819657507856e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 223.999999997062, 0, 0, 0, 0, 0, 0, 0, -2271.99999997829, 0, -3765.33333328541, -479.999999994038, 0, 0, 0, 0, 0, 0, 4095.99999992772, 0, 21631.9999997724, 0, 0, 37674.666666338, 255.999999962935, 4351.99999993631, 7552.00000004441, 0, 0, 0, 0, 0, 0, -2047.99999986278, 0, -53589.3333328182, 0, 0, 0, -64170.6666656779, 0, -211285.333332135, 2.89648777738542e-07, -4095.99999962117, -38229.3333336413, -2047.99999977794, -3754.66666718573, -67925.3333335406, 0, 0, 0, 0, 0, 0, -1.46256064060154e-07, 0, 57343.9999993851, 0, 0, 0, 30037.3333317564, 0, 502442.666664751, -7.03393407080855e-07, -9.17450566656801e-07, 73728.0000005649, 0, 0, 322218.666664085, -9.81503708604474e-07, 3.21399882463396e-07, -5.81346317190483e-07, 60074.6666661701, 16384.0000024286, 338602.666668429, 30037.3333350878, 0, 0, 0, 0, 0, 0, 8.17305340883974e-08, 0, -21845.3333329621, 0, 0, 0, 1.24074004522308e-06, 0, -502442.666665191, 6.59307082952778e-07, 9.17107326478481e-07, -43690.6666670833, 0, 0, 0, -131071.999997258, 0, -611669.333330853, 1.45517326731147e-06, 6.46166848807267e-07, 8.81472140448428e-07, 1.8759754392194e-06, -21845.3333365563, -633514.666669308, 1.04359278601995e-06, -2.72599731363524e-06, -262144.000001605, -3.37194796697785e-07, -1.82398366312646e-06, -131072.000006012, 0, 0, 0, 0, 0, 0, -1.89875324989083e-08, 0, -8.83533868286921e-08, 0, 0, 0, -3.7539712594319e-07, 0, 174762.666666215, -2.21243094098641e-07, -3.24019025680833e-07, 1.14957395100461e-07, 0, 0, 0, -1.10729090275242e-06, 0, 349525.333332493, -6.52965536508554e-07, -5.25124125657271e-07, -4.59136519690274e-07, -1.21334944445443e-06, 1.33740942740922e-06, 349525.333334434, 0, 0, 174762.666665414, -7.17988021636098e-07, 6.57222032315246e-07, -4.52899300940391e-07, -3.22416021158512e-07, 2.13887811884502e-06, 349525.333335217, -1.15058616404364e-06, 3.2069467312494e-07, 174762.666670742, 4.26064523721507e-07, 2.75812489971706e-06, 3.55826587409209e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -597.333333337434, 0, 0, 0, 0, 0, 0, 0, 9244.44444448833, 0, 6855.11111114276, 2076.44444444127, 0, 0, 0, 0, 0, 0, -46421.3333335491, 0, -19000.8888891754, 0, 0, -103651.555555579, -2389.333333328, -29696.0000001095, -21390.2222220689, 0, 0, 0, 0, 0, 0, 96028.4444451885, 0, 20024.8888898599, 0, 0, 0, 497891.555554361, 0, 268970.666667818, 910.222222055514, 126520.888889744, 40960.0000000809, 31402.6666677078, 21845.3333328935, 300373.333331523, 0, 0, 0, 0, 0, 0, -87381.3333347651, 0, -7281.77777919995, 0, 0, 0, -953912.888885803, 0, -262144.000004503, 6.604516998387e-07, -185685.33333628, -21845.3333344012, 0, 0, -1121393.77777586, -10922.6666687345, -7281.77777675549, -109226.666671456, -1230620.44443602, -21845.3333335032, -535210.666665871, -283989.333331139, 0, 0, 0, 0, 0, 0, 29127.1111124103, 0, 9.44475542212504e-07, 0, 0, 0, 786431.999997629, 0, 87381.3333386755, -8.26347000486263e-07, 87381.3333374569, 1.09599611136808e-06, 0, 0, 0, 1572863.99999347, 0, 902940.444446596, 1.25500496080771e-06, -1.80247645792954e-06, 87381.3333412434, 1660245.33332199, 1.40944661708828e-06, 262144.000005701, 29127.1111178718, 3.14681143415466e-07, 1835007.99998851, 87381.3333324029, 961194.666662145, 262143.999998024, 0, 0, 0, 0, 0, 0, -4.358001855315e-07, 0, -2.34790844540066e-07, 0, 0, 0, -233016.888888418, 0, -2.01359574152167e-06, 3.25581876436202e-07, -1.92039512126961e-06, -2.78421586416515e-07, 0, 0, 0, -699050.666663081, 0, -233016.888891494, -1.0362823076327e-07, 1.03649082211002e-06, -4.36385734146229e-06, -699050.666662463, -7.63932822905551e-07, -4.3903303942999e-06, 0, 0, -699050.666664812, -2.71740963823446e-06, 2.46933074625192e-07, -5.04374320391855e-06, -1398101.33332189, -1.15067099657201e-06, -699050.66666624, 9.48208247075232e-07, -699050.66666444, -2.07707762940492e-06, -233016.888890185, 7.90642019084677e-07, -699050.666660633}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 199.111111110556, 0, 0, 0, 0, 0, 0, 0, -2815.99999999341, 0, -2815.99999998759, -426.666666657175, 0, 0, 0, 0, 0, 0, 11719.1111110848, 0, 11719.1111111113, 0, 0, 39367.1111109179, 227.555555555962, 5575.11111099913, 5575.11111106924, 0, 0, 0, 0, 0, 0, -16384.0000000164, 0, -16384.0000001287, 0, 0, 0, -159743.999998871, 0, -159743.999999763, 5.82042533714659e-08, -19114.6666661118, -19114.6666667988, -2730.66666683375, -2730.66666690406, -72362.6666659055, 0, 0, 0, 0, 0, 0, 7281.77777796052, 0, 7281.77777803595, 0, 0, 0, 211171.555553271, 0, 211171.555556209, -2.88859591058434e-07, 14563.5555545478, 14563.5555560579, 0, 0, 611669.333331018, 5.41320252978677e-07, 4.4052689583669e-07, 7281.77777818099, 243939.555551399, 7281.77777903169, 243939.55555573, 32768.0000004647, 0, 0, 0, 0, 0, 0, -2.37579527863299e-07, 0, -1.90248279836757e-07, 0, 0, 0, -87381.3333315343, 0, -87381.3333345705, 3.76423965879909e-07, 6.69955536482763e-07, -4.4367701959449e-07, 0, 0, 0, -699050.666662536, 0, -699050.666665917, -2.18458391527251e-07, 1.20489141517086e-07, -4.26488864752015e-07, -174762.666659996, -1.72669887826096e-06, -174762.666669338, -1.49629639081452e-06, -1.87721300658664e-06, -786431.99999618, -1.38158014397041e-06, -87381.3333325868, -87381.3333355295, 0, 0, 0, 0, 0, 0, 9.18765700595825e-08, 0, 4.913820290039e-08, 0, 0, 0, -4.65828551215241e-07, 0, 5.33826526416848e-07, -1.49411733839225e-07, -1.17474060861249e-07, 1.06298278245597e-07, 0, 0, 0, 233016.888886902, 0, 233016.888889329, -5.36883386754632e-08, -2.65878785544191e-07, 2.05971831968544e-07, -3.08064023851261e-06, 7.08935072954075e-07, 1.77274546158834e-06, 0, 0, 466033.777776288, 6.73512731322364e-07, 6.086306574092e-07, 9.01581660343461e-07, 466033.777773195, 1.39584030323322e-06, 466033.777778315, 2.06397303237728e-07, -1.27453948593124e-06, 2.21824749222074e-06, 1.48672168533856e-06, 1.66955721145537e-06, 233016.888888667}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -597.333333329802, 0, 0, 0, 0, 0, 0, 0, 6855.11111108489, 0, 9244.44444438238, 2076.44444442724, 0, 0, 0, 0, 0, 0, -19000.8888887921, 0, -46421.3333331275, 0, 0, -103651.55555504, -2389.3333333033, -21390.2222219901, -29696.0000000749, 0, 0, 0, 0, 0, 0, 20024.8888887625, 0, 96028.4444441661, 0, 0, 0, 268970.666664407, 0, 497891.55555445, 910.222221840277, 40959.9999986895, 126520.888889885, 21845.333333201, 31402.6666679078, 300373.333332934, 0, 0, 0, 0, 0, 0, -7281.77777781145, 0, -87381.333333156, 0, 0, 0, -262143.999995644, 0, -953912.888888716, 1.1732252921197e-06, -21845.3333306218, -185685.333335591, 0, 0, -1121393.77777326, -7281.77777732718, -10922.6666680157, -21845.333332595, -535210.666660866, -109226.666672742, -1230620.44444892, -283989.333336555, 0, 0, 0, 0, 0, 0, 1.60405482813487e-07, 0, 29127.1111110604, 0, 0, 0, 87381.333329787, 0, 786432.000000954, -1.24706128524132e-06, -2.21894256765052e-06, 87381.3333352164, 0, 0, 0, 902940.44443731, 0, 1572863.99999824, -1.26232453743424e-06, -9.1774967990044e-07, -1.36276663118699e-06, 262143.999988811, 87381.3333415421, 1660245.33334319, 4.30003549604288e-07, 29127.1111186173, 1835007.99999944, 87381.3333356988, 262144.00000044, 961194.666680243, 0, 0, 0, 0, 0, 0, -7.51360537529969e-08, 0, 0, 0, 0, 0, 1.00658078142319e-06, 0, -233016.888889401, 4.46488120941777e-07, 6.12155666203153e-07, -5.37532735159303e-07, 0, 0, 0, -233016.888885486, 0, -699050.666667207, 7.28623664115087e-07, 1.09768557383714e-06, 6.55529927619994e-07, 5.65888743781776e-06, -3.40020113120095e-06, -699050.666671757, 0, 0, -699050.666664316, -1.52158596240319e-07, -2.13395242097755e-06, -3.53990148796427e-07, -699050.666661158, -5.73320915005862e-06, -1398101.33333847, 1.34415051888109e-06, 2.28181723588741e-06, -699050.66667722, -1.7983819083677e-06, -233016.888896114, -699050.666672113}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -1.01421611118311e-08, 0, 0, 0, 0, 0, 0, 0, 5.40521999344647e-08, 0, 2.07769885969959e-07, 28672.0000002121, 0, 0, 0, 0, 0, 0, 1.7670251573222e-07, 0, -1.18896807593472e-06, 0, 0, -1.69126666906006e-06, -202888.533334766, -202888.533334311, -202888.533336173, 0, 0, 0, 0, 0, 0, -1.15868874526039e-06, 0, 2.80434706750208e-06, 0, 0, 0, 2.72253198186372e-06, 0, 9.61859048033445e-06, 565248.00000321, 565248.000000486, 565248.000012974, 1130496.00000792, 1130496.00001333, 1130496.0000119, 0, 0, 0, 0, 0, 0, 1.97970557253538e-06, 0, -3.17348026574302e-06, 0, 0, 0, 8.80381668074328e-07, 0, -1.9544764305745e-05, -775509.333336275, -775509.333331072, -775509.333357287, 0, 0, -1.85360541962764e-05, -2326528.00001622, -2326528.00002249, -2326528.00001676, -2326528.00000738, -2326528.00004473, -2326528.00005094, -4653056.00004309, 0, 0, 0, 0, 0, 0, -1.41554366766226e-06, 0, 1.73377958436337e-06, 0, 0, 0, -3.96809956583786e-06, 0, 1.64176175967139e-05, 524288.000000958, 524287.999996999, 524288.000019216, 0, 0, 0, 1.0743675908171e-05, 0, 2.98181296868238e-05, 2097152.00001217, 2097152.0000154, 2097152.00001597, 2097151.99998899, 2097152.00005552, 2097152.00007171, 3145728.00002961, 3145728.00005255, 3145728.00004479, 6291456.00004897, 6291456.0000375, 6291456.00010407, 0, 0, 0, 0, 0, 0, 3.73132590777698e-07, 0, -3.72745491720466e-07, 0, 0, 0, 1.86161225764076e-06, 0, -4.87873077350931e-06, -139810.133333332, -139810.133332303, -139810.133338963, 0, 0, 0, -8.28556913944027e-07, 0, -1.39106104261884e-05, -699050.666669584, -699050.666670204, -699050.666672454, -699050.66665761, -699050.66668934, -699050.666698055, 0, 0, -1.17213239650801e-05, -1398101.33334676, -1398101.33335317, -1398101.33335028, -1398101.33333816, -1398101.3333668, -1398101.33337366, -2796202.66668278, -2796202.66667405, -2796202.66673707, -4194304.00002798, -4194304.0000571, -4194304.00004856}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.56037510434174e-08, 0, 1.88264387339198e-08, 819.200000015915, 0, 0, 0, 0, 0, 0, -6.3499986370313e-08, 0, -8.59024940496402e-08, 0, 0, -1.75182534244154e-07, -14472.5333333668, -819.200000149796, -819.200000175639, 0, 0, 0, 0, 0, 0, 1.25159145558203e-07, 0, 1.88559964352928e-07, 0, 0, 0, 5.93846572413473e-07, 0, 6.4070340599245e-07, 90111.9999999686, 5.48263914411281e-07, 7.17406936097329e-07, 13653.3333336614, 13653.333333799, 1.08948812832292e-06, 0, 0, 0, 0, 0, 0, -1.29604527185736e-07, 0, -2.18736401716488e-07, 0, 0, 0, -9.15919377410526e-07, 0, -1.04085306374896e-06, -251221.33333315, -9.5307198610494e-07, -1.31498446735536e-06, 0, 0, -1.66494492125149e-06, -76458.6666668388, -76458.6666671155, -1.06551689144274e-06, -2.58420932946313e-06, -1.69886571365216e-06, -3.12589239933959e-06, -1.92003283030258e-06, 0, 0, 0, 0, 0, 0, 6.79328904739593e-08, 0, 1.30586592188473e-07, 0, 0, 0, 6.63111707606877e-07, 0, 7.76940498219689e-07, 314572.799999784, 7.89644984960811e-07, 1.11647909447967e-06, 0, 0, 0, 1.73891062187265e-06, 0, 1.84183977643525e-06, 174762.666666507, 174762.666666733, 1.41846858635659e-06, 2.66572067101915e-06, 2.2297050470822e-06, 3.65393612812726e-06, 7.06837817429217e-07, 1.52419246961458e-06, 4.69264577122454e-06, 1.16884424949761e-06, 2.80734005458498e-06, 4.36669287362992e-06, 0, 0, 0, 0, 0, 0, -1.41587404511689e-08, 0, -3.19313283354657e-08, 0, 0, 0, -1.84314722636889e-07, 0, -2.18056920196923e-07, -139810.133333251, -2.49922849877861e-07, -3.60995087976408e-07, 0, 0, 0, -6.34344634341945e-07, 0, -7.00399353547506e-07, -139810.133333189, -139810.133333251, -6.54150160822533e-07, -1.0118798869235e-06, -9.78575939600537e-07, -1.47152560479078e-06, 0, 0, -9.78483173748182e-07, -1.12320438514828e-07, -4.35828740647549e-07, -5.89903432381724e-07, -2.29300143103883e-06, -1.08164789363971e-06, -2.68807602270008e-06, -1.27189950754901e-07, -1.46308332037393e-06, -2.81554869252965e-06, -7.87379870073041e-07, -1.82015586750061e-06, -2.85627602828262e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.55575312359841e-08, 0, 0, 819.199999990737, 0, 0, 0, 0, 0, 0, 8.02412670887361e-08, 0, 2.34905048776131e-08, 0, 0, 1.3389675133041e-07, -819.199999971834, -14472.533333222, -819.199999922597, 0, 0, 0, 0, 0, 0, -1.8927768921146e-07, 0, -3.09455412279121e-08, 0, 0, 0, -5.89686979405179e-07, 0, -3.77378765636841e-07, -2.54060203312261e-08, 90111.9999995936, -2.35820863369008e-07, 13653.3333329931, -1.82474177057922e-07, 13653.3333325554, 0, 0, 0, 0, 0, 0, 2.2429882088595e-07, 0, 1.71180596398472e-08, 0, 0, 0, 1.11596962036491e-06, 0, 5.10582222345225e-07, -1.42197370742844e-08, -251221.333332705, 3.66635432872659e-07, 0, 0, 1.2842302159281e-06, 4.48462331793351e-07, 1.63742321250452e-07, -76458.6666656803, -76458.6666643706, 3.69266271882252e-07, 1.79368208989027e-06, 1.60865592749353e-06, 0, 0, 0, 0, 0, 0, -1.29192856246842e-07, 0, 0, 0, 0, 0, -9.52026156310767e-07, 0, -3.49110534020283e-07, 3.46156944763002e-08, 314572.799999574, -2.91824770674597e-07, 0, 0, 0, -1.66518428496453e-06, 0, -1.15805107538617e-06, -2.56023165062483e-07, -5.59829226178448e-08, 174762.666665619, 174762.666664046, -3.47146827613708e-07, -1.82047974803279e-06, -9.69196353681657e-07, -1.60118951819355e-07, -3.58551301245212e-06, -1.37395273610968e-06, -3.26407830488672e-06, -2.25255597408587e-06, 0, 0, 0, 0, 0, 0, 2.85493893280306e-08, 0, 0, 0, 0, 0, 2.99648195645682e-07, 0, 9.68604710489131e-08, -1.48314015978646e-08, -139810.133333232, 9.16370841886262e-08, 0, 0, 0, 7.33093888448618e-07, 0, 3.88178111670262e-07, 4.87308873567843e-08, 0, -139810.133332968, -139810.133332319, 1.37123419784973e-07, 7.00113575902109e-07, 0, 0, 7.50295367403472e-07, 3.21412435348177e-07, 1.18075396505467e-08, 5.48734654387869e-07, 2.10493267238527e-06, 3.60391075202997e-08, 1.77471018495371e-06, 4.36070113985092e-07, 1.85876308423101e-06, 1.073213476875e-06, 1.49205900464415e-06, 8.55243979581129e-07, 2.37806938460425e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.28771820570839e-08, 0, -2.93566718516957e-08, 819.199999994074, 0, 0, 0, 0, 0, 0, 1.06135550593985e-07, 0, 1.56280798861172e-07, 0, 0, 2.11940300860836e-07, -819.200000010058, -819.199999920077, -14472.5333332888, 0, 0, 0, 0, 0, 0, -2.61310468442875e-07, 0, -4.19299099336734e-07, 0, 0, 0, -6.92731780409399e-07, 0, -8.11720283109916e-07, 4.03010758248051e-08, -4.374471292664e-07, 90111.9999997315, -2.86268453166169e-08, 13653.3333335202, 13653.3333330757, 0, 0, 0, 0, 0, 0, 3.4609602584126e-07, 0, 5.92506825609233e-07, 0, 0, 0, 1.18768163033851e-06, 0, 1.54139060699145e-06, -4.73850992523758e-08, 9.91345914715431e-07, -251221.333332614, 0, 0, 1.75563749408929e-06, -8.06004553261601e-08, -4.22139809052941e-07, 5.5670799978932e-07, 1.08719146191036e-06, -76458.6666670025, -76458.6666658481, -1.14108762925753e-06, 0, 0, 0, 0, 0, 0, -2.33110083029573e-07, 0, -4.19560376376302e-07, 0, 0, 0, -1.00176656931507e-06, 0, -1.41905049369598e-06, 3.28490120537096e-08, -9.70536846038388e-07, 314572.799999199, 0, 0, 0, -1.83887753842937e-06, 0, -2.00362727512724e-06, 6.02712203403214e-08, 2.77160203431703e-07, -1.15405363657795e-06, -1.82069997042115e-06, 174762.66666669, 174762.666665173, -3.25373638628758e-07, 6.78524114330376e-07, -1.46148931072301e-06, 1.83970073676514e-06, 9.68731059476696e-07, 2.05687226373116e-06, 0, 0, 0, 0, 0, 0, 6.27856112024207e-08, 0, 1.17082562676134e-07, 0, 0, 0, 3.25325587625954e-07, 0, 5.0059019729545e-07, -1.19736139270777e-08, 3.43792502712904e-07, -139810.133333022, 0, 0, 0, 7.38131796960422e-07, 0, 8.59698636257563e-07, -1.07150214423374e-08, -5.84899040174592e-08, 6.42408407392823e-07, 9.52881356539313e-07, -139810.13333319, -139810.133332448, 0, 0, 9.51863346233199e-07, 9.43446333592321e-08, -2.42276654605128e-07, 4.54535734880158e-07, 1.07782256664352e-06, -2.45046061667581e-07, 9.74617981767831e-07, -6.66781315247732e-07, 1.66668037441427e-07, -7.12449894199028e-07, -1.13678960505616e-06, -1.78676417164212e-06, -1.36557801442873e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 1.71055045279516e-08, 0, 0, 0, 0, 0, 0, 0, -1.57056068711979e-07, 0, -2.76886219581691e-07, -17920.0000003339, 0, 0, 0, 0, 0, 0, 3.6257402783909e-07, 0, 1.338586622198e-06, 0, 0, 2.62864111324549e-06, 234325.333335044, 90965.333335633, 90965.3333372015, 0, 0, 0, 0, 0, 0, -5.90915158403308e-08, 0, -2.86348001956844e-06, 0, 0, 0, -7.40125176230991e-06, 0, -1.13857535980874e-05, -899072.000003374, -171349.333338791, -171349.333349283, -1070421.33334431, -1070421.33334857, -342698.666687452, 0, 0, 0, 0, 0, 0, -6.75009068015068e-07, 0, 3.06899115100564e-06, 0, 0, 0, 8.35537003321878e-06, 0, 1.98583897563232e-05, 1512789.33333624, 141994.666672408, 141994.666694989, 0, 0, 2.74199407912769e-05, 3167573.33335196, 3167573.33335727, 1796778.66669308, 425984.000034114, 1796778.66671471, 425984.000069044, 3593557.33339149, 0, 0, 0, 0, 0, 0, 7.66651485618186e-07, 0, -1.64638818628889e-06, 0, 0, 0, -3.75827960701038e-06, 0, -1.50277127385538e-05, -1179648.0000009, -43690.6666692795, -43690.6666893433, 0, 0, 0, -2.4453719255954e-05, 0, -3.48000529750227e-05, -3582634.66667936, -3582634.66668334, -1310720.00002763, -174762.66668684, -1310720.00005815, -174762.666752993, -3670016.00003564, -3670016.00005224, -262144.000083232, -7340032.00005912, -3932160.00006812, -3932160.00012471, 0, 0, 0, 0, 0, 0, -2.55005830486095e-07, 0, 3.61486185402037e-07, 0, 0, 0, 4.45806884027425e-07, 0, 4.11033311307392e-06, 349525.333333316, 3.57472100292348e-07, 6.79428234107039e-06, 0, 0, 0, 7.19233840181349e-06, 0, 1.41638881755271e-05, 1398101.33333616, 1398101.33333759, 349525.333343775, 3.05139845877482e-06, 349525.333357061, 3.5577154940305e-05, 0, 0, 1.6906635165035e-05, 2097152.00001431, 2097152.00001934, 1398101.3333546, 2.88067411945993e-05, 1398101.33336539, 5.63001009972528e-05, 4194304.00001955, 1398101.33335896, 1398101.33341245, 4194304.00003782, 4194304.0000634, 2097152.0000708}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -1.64211074958614e-08, 0, 0, 0, 0, 0, 0, 0, 1.67420617643324e-07, 0, 2.32608591112286e-07, 9557.3333335839, 0, 0, 0, 0, 0, 0, -5.8872398356711e-07, 0, -1.10447601534978e-06, 0, 0, -2.12473627884411e-06, -147911.111112047, -33223.1111132221, -33223.111113807, 0, 0, 0, 0, 0, 0, 9.67695297673705e-07, 0, 2.45778592068854e-06, 0, 0, 0, 6.55704271859119e-06, 0, 8.43985418480553e-06, 742741.333334627, 38229.3333400707, 38229.3333441823, 475136.000007121, 475136.000008587, 76458.6666829366, 0, 0, 0, 0, 0, 0, -7.96926413905033e-07, 0, -2.83125817461941e-06, 0, 0, 0, -9.07591873862509e-06, 0, -1.4489822580454e-05, -1536455.11111156, -14563.5555659788, -14563.5555750668, 0, 0, -2.05065281856656e-05, -2024334.22223155, -2024334.22223343, -502442.666686325, -43690.6667007117, -502442.666694256, -43690.6667153512, -1004885.33337058, 0, 0, 0, 0, 0, 0, 3.06095833741671e-07, 0, 1.65283145765938e-06, 0, 0, 0, 5.85100999840546e-06, 0, 1.12391785210606e-05, 1398101.33333289, 7.8336450787059e-06, 1.6160109836577e-05, 0, 0, 0, 1.97883478003131e-05, 0, 2.43757317627469e-05, 2970965.33333822, 2970965.33333963, 174762.666689881, 3.05574311719713e-05, 174762.666700745, 5.84534709500624e-05, 1747626.66668694, 1747626.66669304, 6.66378471272693e-05, 3495253.33336506, 524288.00005172, 524288.000077607, 0, 0, 0, 0, 0, 0, -3.91148010914512e-08, 0, -3.90827706554604e-07, 0, 0, 0, -1.43873700044862e-06, 0, -3.24231847269456e-06, -466033.777777478, -2.28385560354164e-06, -5.06447596316903e-06, 0, 0, 0, -6.65143166201737e-06, 0, -9.7506128220435e-06, -1398101.33333401, -1398101.33333447, -9.7410024314262e-06, -1.01365821438923e-05, -1.42987048106617e-05, -2.39725827873748e-05, 0, 0, -1.21514584707235e-05, -1398101.33334015, -1398101.33334201, -466033.777791095, -2.88563843859543e-05, -466033.777794494, -4.04242075461078e-05, -2796202.66667543, -2.43966732334773e-05, -4.84709346437321e-05, -1398101.33335587, -1398101.33336901, -4.97180849692515e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7.44006244516021e-08, 0, -9.83713205232122e-08, -3584.00000009402, 0, 0, 0, 0, 0, 0, 2.8658274602362e-07, 0, 4.65982641046003e-07, 0, 0, 8.79596601060708e-07, 60245.3333335957, 7680.000000855, 7680.00000101904, 0, 0, 0, 0, 0, 0, -5.38014579123337e-07, 0, -1.05780480913542e-06, 0, 0, 0, -2.8446827873121e-06, 0, -3.36100356520111e-06, -346112.000000098, -4096.0000029735, -4096.00000410101, -120832.000002305, -120832.000002865, -8192.00000640233, 0, 0, 0, 0, 0, 0, 5.30524221918439e-07, 0, 1.26214366373954e-06, 0, 0, 0, 4.23437404131392e-06, 0, 5.69014897066747e-06, 857429.3333328, 4.94566581821791e-06, 7.45056965090076e-06, 0, 0, 8.32272940209593e-06, 611669.3333354, 611669.333336399, 60074.6666736066, 1.46594893160863e-05, 60074.6666763587, 1.84233421747388e-05, 120149.33334632, 0, 0, 0, 0, 0, 0, -2.63752777082768e-07, 0, -7.69248476485581e-07, 0, 0, 0, -2.97491927352313e-06, 0, -4.43042690166143e-06, -917503.999999236, -3.93928417467439e-06, -6.29251973268936e-06, 0, 0, 0, -8.40967301715986e-06, 0, -9.5368268335587e-06, -1179648.00000018, -1179648.000001, -8.71485330129397e-06, -1.45311951886201e-05, -1.22778414645497e-05, -2.16103258464796e-05, -262144.000005618, -262144.000008605, -2.68897511405627e-05, -524288.000009306, -1.93843664578427e-05, -2.75539409283491e-05, 0, 0, 0, 0, 0, 0, 5.18901941511664e-08, 0, 1.90134779571091e-07, 0, 0, 0, 8.03377768480883e-07, 0, 1.29965438749864e-06, 349525.333333026, 1.20338705516228e-06, 2.02331300395557e-06, 0, 0, 0, 2.9956967151201e-06, 0, 3.75107785225195e-06, 699050.666666275, 699050.66666651, 3.82998804441575e-06, 5.29359274397825e-06, 5.27029574167616e-06, 8.78369963582079e-06, 0, 0, 4.88593428499818e-06, 349525.333334701, 349525.333335845, 4.11985156865782e-06, 1.26854718854996e-05, 5.76140833605016e-06, 1.55226832262304e-05, 699050.666668458, 9.87953536228753e-06, 1.71681875100961e-05, 6.94111072885257e-06, 1.18787587595738e-05, 1.86788947445005e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.93285325916784e-08, 0, 2.33553359406222e-08, 512.000000029155, 0, 0, 0, 0, 0, 0, -8.75963343876293e-08, 0, -1.16395459042875e-07, 0, 0, -1.77251737516472e-07, -8021.33333344246, -4608.00000023959, -512.000000308702, 0, 0, 0, 0, 0, 0, 2.10746815547852e-07, 0, 2.92348369057375e-07, 0, 0, 0, 5.7502502376969e-07, 0, 6.50867851392748e-07, 40277.333333481, 4096.00000082697, 1.26092121747875e-06, 71680.0000007996, 7509.33333434286, 4096.00000168127, 0, 0, 0, 0, 0, 0, -2.76390829170948e-07, 0, -3.90901622023166e-07, 0, 0, 0, -9.50259089084584e-07, 0, -1.15812475050549e-06, -76458.6666667157, -1.42933361597273e-06, -2.35829288538704e-06, 0, 0, -1.46398288108865e-06, -354986.666667713, -32768.0000012292, -60074.6666688629, -3.65965886205448e-06, -3.32056245106941e-06, -4.72694501111092e-06, -60074.6666709206, 0, 0, 0, 0, 0, 0, 1.8648837981447e-07, 0, 2.65844032516793e-07, 0, 0, 0, 7.75414340398564e-07, 0, 9.8581415016924e-07, 43690.6666666224, 1.20322502829101e-06, 2.06090405922789e-06, 0, 0, 0, 1.50643941915821e-06, 0, 1.6457481454675e-06, 655360.000000578, 43690.6666671885, 2.61721586599008e-06, 3.69833715443726e-06, 4.20147615436945e-06, 5.61159750581518e-06, 262144.000002151, 3.12162686779722e-06, 6.17393941575479e-06, 262144.000003684, 6.1642012395936e-06, 8.54989831870121e-06, 0, 0, 0, 0, 0, 0, -5.07126178301799e-08, 0, -7.23554007389638e-08, 0, 0, 0, -2.47241190216838e-07, 0, -3.21710605370225e-07, 2.70491628786785e-08, -3.91054919185208e-07, -6.84222680957377e-07, 0, 0, 0, -5.80015734922749e-07, 0, -6.72364134490462e-07, -349525.333333441, -1.25685869069704e-08, -1.11681349181708e-06, -1.42031617902731e-06, -1.81029065402501e-06, -2.36154350030733e-06, 0, 0, -8.06581650140152e-07, -349525.333334055, -9.30842365423657e-07, -1.35910480850117e-06, -2.81192358551984e-06, -2.06836606043226e-06, -3.47256187991118e-06, -349525.333334268, -3.10983908410553e-06, -5.22983653273239e-06, -2.77972842123155e-06, -4.13895727225602e-06, -5.48588814696286e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.29055671255863e-08, 0, 0, 455.111111123152, 0, 0, 0, 0, 0, 0, 7.57841187858074e-08, 0, -2.12975611760056e-08, 0, 0, 1.06747781983412e-07, -5916.44444451537, -5916.44444442692, -455.111111326674, 0, 0, 0, 0, 0, 0, -1.80612561186254e-07, 0, 9.06008080133619e-08, 0, 0, 0, -6.21386627616739e-07, 0, -1.97687898631659e-07, 20024.8888890026, 20024.8888886047, 1.13255018276853e-06, 76458.6666668367, 5461.33333432459, 5461.33333351022, 0, 0, 0, 0, 0, 0, 2.03237083113119e-07, 0, -1.46875979118697e-07, 0, 0, 0, 1.29630734442157e-06, 0, 8.04513685223596e-08, -14563.5555556076, -14563.5555549392, -2.33520085545529e-06, 0, 0, 1.20380127975835e-06, -254862.222222509, -14563.5555570638, -254862.222222361, -14563.5555537244, -3.83342690702988e-06, -2.1151063665221e-06, -65536.0000017078, 0, 0, 0, 0, 0, 0, -1.05661826892331e-07, 0, 1.10494214863506e-07, 0, 0, 0, -1.1366805195192e-06, 0, 5.96254371563502e-08, -1.24186869734827e-08, -5.00977047334132e-07, 2.08814751844549e-06, 0, 0, 0, -1.89846015912828e-06, 0, -8.24294881966668e-07, 174762.666666703, 9.35570412550992e-07, 174762.666666741, -3.45980662878462e-06, 5.18999815021446e-06, 3.87924905346592e-06, 815559.111111741, 4.16029574631079e-06, -1.06065832386923e-06, 174762.666668576, 174762.666665997, 6.1232203085964e-06, 0, 0, 0, 0, 0, 0, 1.97185497267586e-08, 0, -3.25797820835197e-08, 0, 0, 0, 3.55536391984067e-07, 0, -3.8298168293895e-08, 0, 1.37451601141996e-07, -6.85236408498021e-07, 0, 0, 0, 9.05321030495784e-07, 0, 1.80130626945312e-07, 6.93050892409074e-08, -2.00382700664685e-07, -4.14032074408581e-08, 1.65559547139616e-06, -2.29071089647219e-06, -1.94937775282955e-06, 0, 0, 7.55332227069569e-07, -466033.777778027, -1.46974021212353e-06, -466033.777778248, 2.21492273427218e-06, -2.87015127147041e-06, -6.39748058308091e-07, -4.72888860778859e-07, 1.44021458345918e-06, -4.85772245847278e-06, -466033.777778021, -3.45920971816149e-06, -1.01471669552609e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7.4933573026379e-08, 0, -4.23634860662047e-08, 511.999999976698, 0, 0, 0, 0, 0, 0, 3.90694069316681e-07, 0, 1.37716961400312e-07, 0, 0, 5.71535787512361e-07, -4607.9999999671, -8021.33333291191, -511.999999960966, 0, 0, 0, 0, 0, 0, -9.60736684546507e-07, 0, -2.36005453556785e-07, 0, 0, 0, -2.46765345923631e-06, 0, -1.54843180342298e-06, 4095.99999994586, 40277.3333313317, 1.77709961734435e-07, 71679.9999990393, 4096.00000056319, 7509.33333161494, 0, 0, 0, 0, 0, 0, 1.20814499109728e-06, 0, 2.28955042091047e-07, 0, 0, 0, 4.72038551239185e-06, 0, 2.06027937211088e-06, 1.32109001323152e-07, -76458.6666626797, -5.20576893537492e-07, 0, 0, 5.02710448791137e-06, -60074.666665223, -1.3679737157081e-06, -354986.666662971, -32767.9999926993, -2.4427505416538e-06, 2.37125059006499e-06, -60074.6666656922, 0, 0, 0, 0, 0, 0, -7.5215275273511e-07, 0, -1.14861786052264e-07, 0, 0, 0, -4.1014995257976e-06, 0, -1.4149724805535e-06, -1.48773042990539e-07, 43690.6666631141, 4.6782892221229e-07, 0, 0, 0, -6.52713289806229e-06, 0, -4.29695791734462e-06, -1.39283042968483e-06, 1.14928113000242e-06, 655359.999994919, 43690.6666562023, 3.17531142938183e-06, -1.40764537956926e-06, 262143.999996574, 3.63160745566912e-06, -8.38687701477722e-06, 3.48670214552996e-07, 262143.999993661, 1.97410696227576e-06, 0, 0, 0, 0, 0, 0, 1.83972285690373e-07, 0, 2.17949414745138e-08, 0, 0, 0, 1.3198325095098e-06, 0, 4.03334563075286e-07, 5.61213921644794e-08, 1.16936236032097e-06, -1.42735419037707e-07, 0, 0, 0, 2.90516450403279e-06, 0, 1.37787496284982e-06, 5.67572199865174e-07, -3.37855620528545e-07, -349525.333330998, 4.86017519249451e-06, -1.32153475921974e-06, 3.64252439222703e-07, 0, 0, 2.76588964419807e-06, 1.40853504617445e-06, -1.55967603822925e-06, -349525.333331086, 6.4796254959414e-06, -2.3154088795858e-06, 3.12086518845309e-06, -1.66971892233631e-07, -349525.333327976, -2.06675851781779e-06, 1.91790853693194e-06, -2.48422281026268e-06, 2.2088945121249e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9.93580072189468e-08, 0, 4.3213016491728e-08, -3583.99999995469, 0, 0, 0, 0, 0, 0, -5.28873176802417e-07, 0, -7.62442293097767e-08, 0, 0, -8.15028576022352e-07, 7679.99999985727, 60245.3333326852, 7679.99999975909, 0, 0, 0, 0, 0, 0, 1.25843777738697e-06, 0, -2.95490716213325e-08, 0, 0, 0, 3.87175821447241e-06, 0, 1.97354131922026e-06, -4095.99999970799, -346111.999997428, -4095.99999974548, -120831.999998121, -8191.99999980128, -120831.999996071, 0, 0, 0, 0, 0, 0, -1.48294614758922e-06, 0, 2.06108644856224e-07, 0, 0, 0, -7.57500476344019e-06, 0, -2.02011312863662e-06, -4.19055717820234e-07, 857429.333329067, 2.23752908784709e-07, 0, 0, -7.86817076464215e-06, 60074.666663607, 1.06035779874011e-07, 611669.333327846, 611669.33331971, 1.30115693971667e-06, 60074.6666594787, 120149.333326611, 0, 0, 0, 0, 0, 0, 8.41405086527482e-07, 0, -2.05085430836406e-07, 0, 0, 0, 6.53133057253371e-06, 0, 9.01730065891768e-07, 3.42718195342282e-07, -917503.99999688, -4.59560349795453e-07, 0, 0, 0, 1.09896255548212e-05, 0, 6.1758027933914e-06, 2.7194053491718e-06, -4.38420823444064e-08, -1179647.99999406, -1179647.99998302, -2.75139440729026e-06, 5.246267148333e-06, -262143.999994278, -2.58267148557951e-06, -262143.99998082, 5.90336855385741e-06, -524287.999983706, 5.98561061647019e-06, 0, 0, 0, 0, 0, 0, -1.81680463544352e-07, 0, 6.68301009174286e-08, 0, 0, 0, -2.05430297829239e-06, 0, -1.35934745966039e-07, -1.12632925806959e-07, 349525.333332512, 1.80108609885133e-07, 0, 0, 0, -4.99967648658141e-06, 0, -1.67795068983029e-06, -9.90077189972102e-07, -6.94714993012224e-08, 699050.666664513, 699050.666659672, 1.36966735933164e-06, -1.37592729893727e-06, 0, 0, -4.60231437656321e-06, -2.38772208183264e-06, 1.11305451909567e-06, 349525.333330312, 349525.333320429, 2.31397879022847e-06, -8.09376306186051e-06, -2.41397671014303e-06, 699050.666656609, -1.06736301143163e-06, -7.13524590758486e-06, -1.49493650451457e-06, -1.03511730409487e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.95493981234161e-07, 0, -1.10875206239646e-08, 9557.33333329382, 0, 0, 0, 0, 0, 0, 1.16676767810707e-06, 0, -3.78097910438038e-07, 0, 0, 1.40166727129037e-06, -33223.1111110863, -147911.111109972, -33223.1111114207, 0, 0, 0, 0, 0, 0, -2.86361203426881e-06, 0, 1.59344032938667e-06, 0, 0, 0, -8.56820129078913e-06, 0, -1.5740274020544e-06, 38229.3333331545, 742741.333328185, 38229.3333369676, 475135.999997473, 76458.666669389, 475135.999994183, 0, 0, 0, 0, 0, 0, 3.3346925937542e-06, 0, -2.52864046795377e-06, 0, 0, 0, 1.83726535093244e-05, 0, -2.36300810767346e-06, -14563.5555547875, -1536455.11110232, -14563.5555647236, 0, 0, 1.44116602412262e-05, -502442.666662471, -43690.6666715618, -2024334.22221417, -2024334.22219348, -43690.6666811146, -502442.666663874, -1004885.33332767, 0, 0, 0, 0, 0, 0, -1.81152459631895e-06, 0, 1.81255631308067e-06, 0, 0, 0, -1.63925123794877e-05, 0, 4.76331941487095e-06, -9.70212658856608e-07, 1398101.33332705, 8.84630609306608e-06, 0, 0, 0, -2.4907223235104e-05, 0, -6.53601092360819e-06, 174762.666661907, 2.92025135883064e-06, 2970965.33332479, 2970965.33329317, 2.18373443275821e-05, 174762.666675785, 1747626.66665963, 1.78694766581119e-05, 1747626.66663315, 524287.999996036, 3495253.3333065, 524288.000009691, 0, 0, 0, 0, 0, 0, 3.61288847228411e-07, 0, -4.94629920611507e-07, 0, 0, 0, 5.19417039180345e-06, 0, -2.04348925067072e-06, 3.8750426394298e-07, -466033.777776245, -2.97311711721207e-06, 0, 0, 0, 1.23446055780082e-05, 0, -4.02562224862969e-07, 2.1654950726468e-06, -3.90803385847428e-07, -1398101.33333047, -1398101.33331602, -1.00967290762649e-05, -7.14664402981705e-06, 0, 0, 8.65804519561974e-06, -466033.777774542, -6.78879190786867e-06, -1398101.33333019, -1398101.33330435, -1.35666964225069e-05, -466033.777769475, 2.92271582426682e-06, -2796202.66664748, -1.44940378547908e-05, -1398101.33332377, -7.90602203192524e-06, -1398101.33332075}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9.32675982799746e-08, 0, -1.76616869027973e-07, -17920.0000001181, 0, 0, 0, 0, 0, 0, -8.84879266368757e-07, 0, 1.45691694266249e-06, 0, 0, -1.14737312322045e-07, 90965.3333341375, 234325.333333236, 90965.3333354561, 0, 0, 0, 0, 0, 0, 2.42092731235379e-06, 0, -4.41917506374419e-06, 0, 0, 0, 6.13162956098428e-06, 0, -5.07731810599559e-06, -171349.333334948, -899071.999997559, -171349.333345296, -1070421.33333613, -342698.666676134, -1070421.33333574, 0, 0, 0, 0, 0, 0, -2.8118870864963e-06, 0, 6.23550213352042e-06, 0, 0, 0, -1.69256830789131e-05, 0, 1.65376346877403e-05, 141994.666667713, 1512789.33332846, 141994.666692116, 0, 0, -3.03984104562077e-06, 1796778.66667268, 425984.000014917, 3167573.33333868, 3167573.33331502, 425984.000037601, 1796778.66669518, 3593557.33335092, 0, 0, 0, 0, 0, 0, 1.40702725090718e-06, 0, -4.21399223046956e-06, 0, 0, 0, 1.64209450994076e-05, 0, -1.77531117189412e-05, -43690.6666665494, -1179647.99999698, -43690.6666895837, 0, 0, 0, 1.83897389866408e-05, 0, -1.17984954047877e-05, -1310720.00000341, -174762.666675765, -3582634.66667291, -3582634.66663025, -174762.666718831, -1310720.00005389, -3670016.00001271, -262144.000042053, -3670015.99999678, -3932160.00001946, -7340031.99999986, -3932160.00006229, 0, 0, 0, 0, 0, 0, -2.288149822206e-07, 0, 1.11073552389136e-06, 0, 0, 0, -5.33620978448062e-06, 0, 6.33809817131685e-06, -2.38601455020541e-07, 349525.333332962, 7.44641387531182e-06, 0, 0, 0, -1.14998258851484e-05, 0, 9.46626961059619e-06, 349525.333333508, 1.6566098519584e-06, 1398101.33333645, 1398101.33331596, 2.32025616853465e-05, 349525.333361088, 0, 0, -2.15313661285991e-06, 1398101.33333891, 1.51314556862914e-05, 2097152.00000842, 2097151.99997786, 2.99805750752785e-05, 1398101.33334997, 1398101.33333759, 4194303.99999123, 1398101.3333844, 4194304.00000718, 2097152.00003317, 4194304.00001441}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.99491164394421e-08, 0, -3.83142151532015e-08, -17920.0000000385, 0, 0, 0, 0, 0, 0, 7.45426436767206e-08, 0, 2.40805549835277e-08, 0, 0, 1.49911809261301e-06, 90965.3333341623, 90965.3333332938, 234325.33333401, 0, 0, 0, 0, 0, 0, 4.86745941197662e-07, 0, 6.25597551531258e-07, 0, 0, 0, -5.21721956010495e-06, 0, -6.39344480980141e-06, -171349.333335863, -171349.333330844, -899072.000002606, -342698.666670984, -1070421.33334014, -1070421.33333785, 0, 0, 0, 0, 0, 0, -1.38150183116764e-06, 0, -1.7501125061919e-06, 0, 0, 0, 6.44837956525534e-06, 0, 9.97034006801413e-06, 141994.666669621, 141994.666659183, 1512789.33333594, 0, 0, 1.99535405961983e-05, 425984.000012439, 1796778.6666814, 425984.000005864, 1796778.66666733, 3167573.3333533, 3167573.33335522, 3593557.33335867, 0, 0, 0, 0, 0, 0, 1.32048844917406e-06, 0, 1.70771416486032e-06, 0, 0, 0, -3.07171798042259e-06, 0, -6.11321267404227e-06, -43690.6666681306, -43690.6666587506, -1179647.9999997, 0, 0, 0, -1.99592087598039e-05, 0, -2.4476253583073e-05, -174762.666678178, -1310720.00001139, -174762.666668005, -1310719.99998894, -3582634.66668728, -3582634.66669463, -262144.000018806, -3670016.00003052, -3670016.00002512, -3932160.00003674, -3932160.00001948, -7340032.00005781, 0, 0, 0, 0, 0, 0, -4.40908315295666e-07, 0, -5.69960600346103e-07, 0, 0, 0, 3.58659897712571e-07, 0, 1.10501195844166e-06, 2.46760913088595e-07, -2.86510908565552e-06, 349525.333332396, 0, 0, 0, 6.4571007330642e-06, 0, 9.42487146759819e-06, 3.5042985264042e-06, 349525.333336149, -1.0365171280033e-06, 349525.333325524, 1398101.33333987, 1398101.33334362, 0, 0, 1.34062279012602e-05, 1.04574300834804e-05, 1398101.33334637, 8.58774954105738e-06, 1398101.33333796, 2097152.00001722, 2097152.00002184, 1398101.33334765, 1398101.33333318, 4194304.00003521, 2097152.00002041, 4194304.00003832, 4194304.00002934}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 1.26921219848336e-08, 0, 0, 0, 0, 0, 0, 0, -9.43752878997218e-08, 0, -1.57482663913346e-07, 9557.3333332629, 0, 0, 0, 0, 0, 0, 4.65420609643684e-07, 0, 9.45414100700101e-07, 0, 0, 4.44906932663426e-07, -33223.111111332, -33223.1111101285, -147911.111110686, 0, 0, 0, 0, 0, 0, -1.39165353662583e-06, 0, -2.84409950546643e-06, 0, 0, 0, -5.33488293259856e-07, 0, -1.90995049750545e-06, 38229.3333344056, 38229.3333280761, 742741.333331215, 76458.6666663737, 475136.000002855, 475135.999997258, 0, 0, 0, 0, 0, 0, 2.20407311391568e-06, 0, 4.32586490385379e-06, 0, 0, 0, 1.0704487556892e-06, 0, 5.29670794559249e-06, -14563.5555569439, -14563.5555440415, -1536455.11110511, 0, 0, -5.2086686509815e-07, -43690.666669942, -502442.666674873, -43690.6666601135, -502442.666653991, -2024334.22222974, -2024334.22221858, -1004885.33334164, 0, 0, 0, 0, 0, 0, -1.71645823320457e-06, 0, -3.1939530191725e-06, 0, 0, 0, -1.60699655255278e-06, 0, -6.44749435221251e-06, 7.55668884303905e-07, -1.09740997284923e-05, 1398101.33332618, 0, 0, 0, 2.37965953650864e-06, 0, -4.59858923183853e-07, 4.00483262558809e-06, 174762.666673398, -1.27333393828474e-05, 174762.666645966, 2970965.33333856, 2970965.33332708, 4.63131998559296e-07, 1747626.66668314, 1747626.66665849, 524288.000019327, 524287.999996424, 3495253.33335559, 0, 0, 0, 0, 0, 0, 5.19765747226391e-07, 0, 9.10692444494837e-07, 0, 0, 0, 7.97772305734748e-07, 0, 2.70974962107956e-06, -1.60939698184281e-07, 3.82100782880367e-06, -466033.777774862, 0, 0, 0, -8.9376832045419e-07, 0, 1.23450710283058e-06, -1.36304087551529e-06, -1.66718254241592e-06, 6.7601906382486e-06, 1.04119707689099e-05, -1398101.33333362, -1398101.3333287, 0, 0, -1.73476958355701e-06, -2.25096078820953e-06, -466033.777785378, 2.41058356104654e-06, -466033.777770041, -1398101.33334162, -1398101.333331, -8.4949769464556e-06, 1.05180692174766e-05, -2796202.66667805, -6.66664322629624e-06, -1398101.33335515, -1398101.33334099}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7.45234844197571e-08, 0, 1.05916203381208e-07, -3583.9999999629, 0, 0, 0, 0, 0, 0, -3.43025727358717e-07, 0, -5.90355136475202e-07, 0, 0, -5.97629942893915e-07, 7680.00000002628, 7679.99999952381, 60245.3333330824, 0, 0, 0, 0, 0, 0, 8.89954195445934e-07, 0, 1.65080963817588e-06, 0, 0, 0, 1.6532930451378e-06, 0, 2.3952945388089e-06, -4096.00000021056, -4095.99999760466, -346111.999998718, -8191.99999946098, -120832.000000895, -120831.999998393, 0, 0, 0, 0, 0, 0, -1.25884575465496e-06, 0, -2.38936022177651e-06, 0, 0, 0, -2.75469337351007e-06, 0, -5.00330711393605e-06, 2.51668484666645e-07, -5.14756251160416e-06, 857429.333330003, 0, 0, -3.91143657519646e-06, 8.33788961010832e-08, 60074.6666692569, -3.77334574003716e-06, 60074.6666605853, 611669.333335421, 611669.33332956, 120149.333336386, 0, 0, 0, 0, 0, 0, 9.02563632027039e-07, 0, 1.71027122797648e-06, 0, 0, 0, 2.44474960364845e-06, 0, 4.97573116828503e-06, -1.25850000060858e-07, 4.88798730221328e-06, -917503.99999629, 0, 0, 0, 3.40868175062039e-06, 0, 4.92317155100184e-06, -2.95816130375638e-07, -2.02345497572925e-06, 6.52429744144868e-06, 9.30952096213045e-06, -1179648.00000083, -1179647.99999387, 2.11474130393434e-06, -262144.000005058, -262143.999993531, -6.91456833542757e-06, 9.68617863263221e-07, -524288.000007579, 0, 0, 0, 0, 0, 0, -2.56758666791616e-07, 0, -4.78592004749144e-07, 0, 0, 0, -8.5678450919618e-07, 0, -1.84550848333572e-06, 2.94264257714778e-08, -1.7021801821216e-06, 349525.333331884, 0, 0, 0, -1.30918376229939e-06, 0, -2.41246924836159e-06, 8.78790209813991e-08, 4.79985647235625e-07, -3.35040278993221e-06, -4.6266774642807e-06, 699050.666666246, 699050.666663057, 0, 0, -1.75536883708489e-06, -3.76283473596722e-07, 2.26174734409688e-06, -2.33993433084721e-06, -4.55439894052169e-06, 349525.333335663, 349525.333329755, 2.82785511657308e-06, -3.85925959759568e-06, 699050.66666973, 2.53307711834064e-06, 7.92670572062381e-06, 3.12378089293903e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.18106571906532e-08, 0, 2.31190601852133e-08, 512.000000020569, 0, 0, 0, 0, 0, 0, -2.26925715480749e-08, 0, -1.13660248953104e-07, 0, 0, -2.17050376045053e-07, -8021.33333339321, -512.000000198285, -4608.00000022273, 0, 0, 0, 0, 0, 0, -2.04667132124061e-08, 0, 2.50675304547073e-07, 0, 0, 0, 6.21537650221859e-07, 0, 9.37448340927779e-07, 40277.3333333672, 6.18492573143918e-07, 4096.00000084553, 7509.33333392347, 71680.0000005656, 4096.0000016997, 0, 0, 0, 0, 0, 0, 1.0647516323991e-07, 0, -2.78606248910659e-07, 0, 0, 0, -7.42430047371863e-07, 0, -1.66202862427872e-06, -76458.666666585, -8.53331916349535e-07, -1.4376893169513e-06, 0, 0, -2.23501921342226e-06, -32768.0000005217, -354986.666667263, -1.76628818279507e-06, -3.85445665524111e-06, -60074.6666683155, -4.99723111337624e-06, -60074.6666701133, 0, 0, 0, 0, 0, 0, -1.12323524379524e-07, 0, 1.52513860872339e-07, 0, 0, 0, 3.60674540314382e-07, 0, 1.31435702097761e-06, 43690.6666665365, 5.22051479902424e-07, 1.13435976026191e-06, 0, 0, 0, 2.07016223603217e-06, 0, 2.78080259918974e-06, 43690.6666666802, 655360.000000299, 2.11845400064875e-06, 3.4735980443201e-06, 1.86994610732362e-06, 5.81226838329584e-06, 1.45764664847474e-06, 262144.000001252, 7.68967135905607e-06, 262144.000002424, 5.5604112367982e-06, 6.88321329896598e-06, 0, 0, 0, 0, 0, 0, 3.85418116560847e-08, 0, -3.26450438069132e-08, 0, 0, 0, -4.52981678156221e-08, 0, -3.84683336353321e-07, 5.5271515062581e-08, -1.09042340029625e-07, -3.39971121474128e-07, 0, 0, 0, -6.53544501373599e-07, 0, -1.14051520155644e-06, 1.14837213132316e-07, -349525.333333391, -8.70538001929228e-07, -1.07844751476893e-06, -7.37081623381383e-07, -2.32156054236302e-06, 0, 0, -1.35820501069282e-06, -3.01829861435424e-07, -349525.333333707, -1.08445401762302e-06, -3.52713958088819e-06, -6.85138761196676e-07, -4.56088902108646e-06, -349525.333333844, -2.79022845995749e-06, -4.08452961355276e-06, -2.05098259560079e-06, -2.69505051682987e-06, -5.33299980732604e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 455.11111110706, 0, 0, 0, 0, 0, 0, 2.05111321692187e-08, 0, 5.58562976327108e-08, 0, 0, -9.35497088354124e-08, -5916.44444448086, -455.111111047349, -5916.44444441787, 0, 0, 0, 0, 0, 0, -1.17427341627554e-07, 0, -2.27635532165328e-07, 0, 0, 0, 4.67283322657224e-07, 0, 2.99832699150007e-07, 20024.8888890613, -4.4291098452821e-07, 20024.8888886977, 5461.33333343203, 76458.6666668936, 5461.33333335603, 0, 0, 0, 0, 0, 0, 2.45906051343999e-07, 0, 4.02498930335055e-07, 0, 0, 0, -6.98945654505671e-07, 0, -1.61650420718309e-07, -14563.5555558227, 1.13179298162694e-06, -14563.5555549377, 0, 0, -1.58807581271504e-06, -14563.5555561808, -254862.222223097, 2.87412351727711e-07, 4.56968352547634e-07, -254862.22222254, -14563.5555560586, -65536.000000781, 0, 0, 0, 0, 0, 0, -2.18965194924518e-07, 0, -3.22831623428415e-07, 0, 0, 0, 3.74301364624977e-07, 0, -2.19854578218997e-07, 1.77057506938656e-07, -1.20023513929592e-06, -7.7067916774496e-07, 0, 0, 0, 2.00355301800352e-06, 0, 1.51232666265038e-06, 6.9254890426863e-07, 174762.666667681, -8.61045638771911e-07, -1.35918858543851e-06, 174762.666666565, 3.23780371671027e-07, 6.26045018016419e-07, 815559.111112261, 1.09427748885772e-06, 174762.666668644, -4.18075668332963e-07, 174762.666668437, 0, 0, 0, 0, 0, 0, 7.08134018838172e-08, 0, 9.69229112754464e-08, 0, 0, 0, -4.27372562089337e-08, 0, 1.7729826731851e-07, -4.27929112047159e-08, 4.52464879971864e-07, 3.22738493342094e-07, 0, 0, 0, -7.5969844996371e-07, 0, -3.96884720011175e-07, -2.26950468517667e-07, -3.82696045682996e-07, 5.29481606165432e-07, 8.82964608876718e-07, 2.3781568816346e-07, 1.20252976523316e-07, 0, 0, -1.12331358126759e-06, -4.79112707921034e-07, -466033.777778511, -1.4707956464311e-07, -2.86799713791873e-07, -466033.777778134, -9.19521022953935e-07, -1.14421582899198e-06, 1.11750815239942e-06, -8.80750160890603e-07, -5.58724473881062e-07, -466033.777779473, -6.2360592246249e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.62685865524966e-08, 511.999999987441, 0, 0, 0, 0, 0, 0, 0, 0, 1.2156711783812e-07, 0, 0, -2.79795267603536e-08, -4608.00000003351, -511.999999830192, -8021.33333324314, 0, 0, 0, 0, 0, 0, -5.31593961233241e-08, 0, -3.99124090405485e-07, 0, 0, 0, 4.07886560691229e-07, 0, -1.08052541074742e-07, 4096.00000019797, -8.18673474339159e-07, 40277.3333328553, 4095.99999984473, 71680.0000004332, 7509.33333270263, 0, 0, 0, 0, 0, 0, 1.33716070463027e-07, 0, 6.19277452239763e-07, 0, 0, 0, -8.42439108416255e-07, 0, 7.50607734259329e-07, -2.77548594513294e-07, 1.71737405107533e-06, -76458.6666654423, 0, 0, -9.60339576415389e-07, -3.69958382380023e-07, -60074.6666682871, 1.27071556828803e-06, 2.01357921018321e-06, -354986.666667569, -32767.9999984102, -60074.666667173, 0, 0, 0, 0, 0, 0, -1.33576807003102e-07, 0, -4.53397049268061e-07, 0, 0, 0, 6.42343933939887e-07, 0, -1.07862315138575e-06, 1.4870504964114e-07, -1.6286575126335e-06, 43690.6666653276, 0, 0, 0, 1.74428477141406e-06, 0, 3.89270482573533e-07, 5.42496476246106e-07, 1.65371375943216e-06, -2.14944525773431e-06, -2.703833209617e-06, 655360.000000376, 43690.6666642048, -2.9111273922769e-07, 262144.000002804, -2.22092387067128e-06, 2.76025290349776e-06, -1.74458885462308e-06, 262144.000001266, 0, 0, 0, 0, 0, 0, 4.66383087546766e-08, 0, 1.26596152125368e-07, 0, 0, 0, -1.62349646168233e-07, 0, 4.67158415847822e-07, -2.43977145600068e-08, 5.73884422030538e-07, 5.15130497593529e-07, 0, 0, 0, -7.94885951114569e-07, 0, 1.78089610207896e-07, -1.81378832684515e-07, -5.34152697156555e-07, 1.09173295039958e-06, 1.25753780391243e-06, -349525.333333192, 1.37713892814586e-06, 0, 0, -7.48045802555643e-07, -1.75677988283381e-07, -1.55356738335408e-06, 5.45901016405663e-07, 1.16344364303457e-06, -349525.333334593, 1.29651170560936e-06, -1.59514456537804e-06, 1.96092540226674e-06, -349525.33333366, -4.90127129934419e-07, -2.69248925741415e-06, 5.38657722020073e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.72270529107834e-08, 0, 2.06452589125985e-08, 512.000000003004, 0, 0, 0, 0, 0, 0, -7.65203753551494e-08, 0, -1.21387945240839e-07, 0, 0, -1.183190494567e-07, -511.999999964877, -8021.33333341101, -4608.00000000776, 0, 0, 0, 0, 0, 0, 2.04920083318278e-07, 0, 3.3851363553889e-07, 0, 0, 0, 2.00120401752677e-07, 0, 5.91270043092964e-07, -1.79569013388572e-07, 40277.3333338682, 4096.00000019435, 7509.33333334668, 4095.99999973012, 71679.9999999466, 0, 0, 0, 0, 0, 0, -3.13759732230457e-07, 0, -4.66161751359775e-07, 0, 0, 0, -1.76330584861918e-07, 0, -1.39634926182473e-06, 3.4867919582215e-07, -76458.6666680674, -6.55083904158849e-07, 0, 0, -6.74017998995622e-07, 4.1503343508468e-07, 8.30380063739977e-07, -32768.0000009165, -354986.666666747, 1.81409901915183e-07, -60074.6666667973, -60074.6666654204, 0, 0, 0, 0, 0, 0, 2.44526223851234e-07, 0, 3.08340545977762e-07, 0, 0, 0, 1.40458717181315e-07, 0, 1.42287824347003e-06, -3.09306095344285e-07, 43690.6666682034, 7.38179976846849e-07, 0, 0, 0, 1.14764335227979e-07, 0, 1.28100731164377e-06, -6.81756782588115e-07, -9.69662672992764e-07, 43690.6666686571, 655360.000000757, 4.28521670654639e-07, 1.10506919763018e-06, 3.42005335970116e-07, -6.62234266117173e-07, 262143.999999255, -2.41382477054178e-06, 262143.999999244, -1.68598328779478e-06, 0, 0, 0, 0, 0, 0, -7.46130150533871e-08, 0, -7.80607178284258e-08, 0, 0, 0, -6.79442152188279e-08, 0, -5.15146764216008e-07, 1.03997422588965e-07, -5.97320300851022e-07, -2.70663283836483e-07, 0, 0, 0, 8.34558649781808e-08, 0, -7.48270139874151e-07, 3.19056203231643e-07, 4.01107181817039e-07, -1.15498766482498e-06, -349525.333333986, -3.71614502703002e-07, -8.58876488654149e-07, 0, 0, -2.34543735949129e-07, 1.27464038179335e-07, 4.30677563826196e-07, -7.19911792270686e-07, -349525.333332832, -4.63542412774326e-08, 3.12467326087498e-08, 1.37909132399209e-06, -349525.333333855, 2.06123964585918e-07, 1.01247452875653e-06, 1.58128585623793e-06, 1.39840887906432e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.22364102708616e-08, 0, 3.11067691625734e-08, 455.111111122258, 0, 0, 0, 0, 0, 0, -2.33068822559292e-08, 0, -1.99078247529875e-07, 0, 0, -1.63461452516955e-07, -455.111111184253, -5916.44444452217, -5916.44444455718, 0, 0, 0, 0, 0, 0, 2.99960316715219e-08, 0, 5.61844042127005e-07, 0, 0, 0, 1.18659538623761e-07, 0, 1.03325056639107e-06, 2.41544786637724e-07, 20024.8888891518, 20024.8888895116, 5461.33333376458, 5461.33333373002, 76458.6666670884, 0, 0, 0, 0, 0, 0, -4.91021836924758e-08, 0, -7.70399981517657e-07, 0, 0, 0, 2.70713718334644e-07, 0, -2.52137601361894e-06, -3.68642999770092e-07, -14563.555556107, -14563.5555569954, 0, 0, -1.18496762093092e-06, -1.10405593317453e-06, -7.96068584398661e-07, -14563.5555565879, -254862.222222584, -14563.5555569129, -254862.222224269, -65536.0000013802, 0, 0, 0, 0, 0, 0, 4.99184939940224e-08, 0, 5.06746221657912e-07, 0, 0, 0, -3.80842665991525e-07, 0, 2.54311427607286e-06, 2.66160975819028e-07, 5.9114526229584e-07, 1.41473738571039e-06, 0, 0, 0, 5.28829600509158e-08, 0, 2.56735278422106e-06, 1.18784422350881e-06, 6.31849434840414e-07, 1.18452008077445e-06, 174762.666666772, 1.91557928882802e-06, 174762.66667038, 1.64356525534532e-06, 1.68408052355923e-06, 815559.111112446, 2.07859821064002e-06, 174762.666668464, 174762.666669564, 0, 0, 0, 0, 0, 0, -1.7974797202471e-08, 0, -1.28223076910696e-07, 0, 0, 0, 1.25122449757084e-07, 0, -9.0295627193779e-07, -7.50697018719112e-08, -2.34753552061353e-07, -4.95217246641939e-07, 0, 0, 0, 2.8421243937595e-07, 0, -1.48842417618172e-06, -4.55981627175184e-07, -1.57274015020366e-07, -5.28855286096127e-07, -8.50045525308894e-08, -8.92755700849812e-07, -2.03483646610019e-06, 0, 0, -5.72745560906629e-07, -8.7677475111974e-07, -6.82731934267636e-07, -8.08097707577102e-07, -466033.777777734, -1.1321447936547e-06, -466033.777779468, -9.12090965947853e-07, -9.12708276921625e-07, -2.18028067453446e-06, -1.54913137268703e-06, -1.83244560590546e-06, -466033.777779233}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 511.999999991936, 0, 0, 0, 0, 0, 0, 3.74786053002911e-08, 0, 2.22907614472983e-08, 0, 0, 2.20389259612391e-08, -511.999999964115, -4607.99999991447, -8021.33333329645, 0, 0, 0, 0, 0, 0, -1.09728456377272e-07, 0, -5.52887120395078e-08, 0, 0, 0, -1.04717401145423e-07, 0, 4.213616681889e-08, -9.34687466952778e-08, 4095.99999963497, 40277.3333332822, 4095.99999967977, 7509.33333325122, 71679.9999997359, 0, 0, 0, 0, 0, 0, 1.64374312632564e-07, 0, 8.33789704972113e-08, 0, 0, 0, 2.73496852130399e-07, 0, -1.79355635822369e-07, 1.35868338518114e-07, 6.85083314593692e-07, -76458.6666665859, 0, 0, -1.48972881768973e-07, 5.16150304820035e-07, 2.33452528462339e-07, 9.94117100020448e-07, -60074.6666656314, -32768.0000001822, -354986.666666776, -60074.6666658776, 0, 0, 0, 0, 0, 0, -1.22122261044275e-07, 0, -6.5287970440957e-08, 0, 0, 0, -3.03260281597516e-07, 0, 1.76791567438853e-07, -9.11052741257364e-08, -5.75406033352937e-07, 43690.6666665542, 0, 0, 0, 5.93744621636924e-08, 0, 4.63268098011173e-07, -3.85221543496039e-07, -3.90421265713767e-07, -1.2099576981681e-06, -1.52463792007134e-06, 43690.6666670006, 655360.000000542, -9.42982534234544e-07, 2.76467707917919e-07, 262143.999999887, -1.15124294037057e-06, -2.11782650173764e-06, 262144.000000287, 0, 0, 0, 0, 0, 0, 3.57430127634661e-08, 0, 2.01306730365465e-08, 0, 0, 0, 1.19193058381106e-07, 0, -5.45832477823816e-08, 2.03216835855531e-08, 1.78901240573533e-07, 5.44762174420717e-08, 0, 0, 0, 3.12741540778271e-08, 0, -2.49991807209153e-07, 1.02788707086232e-07, 1.99452224368225e-07, 4.93178245340361e-07, 7.15864383897876e-07, -1.10742337189384e-07, -349525.333333594, 0, 0, -2.27346205947087e-07, 2.80515889078343e-07, -1.16584940980401e-08, 5.48126442016769e-07, 3.41181242508028e-07, -3.43677313304033e-07, -349525.333333826, 6.61983296139866e-07, 1.50536360716799e-06, -349525.33333394, 1.25554164727005e-06, -4.29408727661668e-08, 3.20406121037085e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.07144479587945e-08, 0, 8.89729305805482e-08, 7168.00000016748, 0, 0, 0, 0, 0, 0, 5.42005137734415e-07, 0, -6.95478592432814e-07, 0, 0, 5.62953103624297e-08, -82261.3333343492, -82261.3333338506, -24917.3333357228, 0, 0, 0, 0, 0, 0, -1.54784176468396e-06, 0, 2.03474657626317e-06, 0, 0, 0, -2.99095693963104e-06, 0, 2.34016712986982e-06, 228010.666668755, 228010.666666142, 28672.0000112041, 914773.33333808, 256682.666677197, 256682.666674038, 0, 0, 0, 0, 0, 0, 1.95388542929808e-06, 0, -2.83088549059399e-06, 0, 0, 0, 8.57559760988861e-06, 0, -7.12064609970082e-06, -240298.666668543, -240298.666663475, -10922.666688423, 0, 0, 1.01125379634241e-06, -2315605.33334218, -491520.000016786, -2315605.33334251, -491519.999999784, -262144.000036092, -262144.000033296, -2577749.33336357, 0, 0, 0, 0, 0, 0, -1.11633895364893e-06, 0, 1.93379357314262e-06, 0, 0, 0, -8.64327452419509e-06, 0, 7.18083212806097e-06, 87381.3333340425, 87381.3333298615, 1.86978014301686e-05, 0, 0, 0, -8.26260186220543e-06, 0, 5.76791609351536e-06, 2184533.33333911, 262144.000011553, 2184533.33334126, 262143.999985216, 87381.333379739, 87381.3333822133, 4194304.00001585, 262144.000039533, 262144.000019754, 4456448.00003308, 4456448.00002433, 2359296.00007399, 0, 0, 0, 0, 0, 0, 2.30772864253074e-07, 0, -5.28301098177722e-07, 0, 0, 0, 2.91258018442157e-06, 0, -2.42162744323174e-06, -7.38209169963413e-08, 1.15003295321296e-06, -5.94769888013473e-06, 0, 0, 0, 5.47710443942415e-06, 0, -4.17260962710737e-06, -699050.666667753, -2.94861655360581e-06, -699050.666669179, 9.44110175952782e-06, -1.98645890183754e-05, -2.23159257134806e-05, 0, 0, 5.15928167430145e-07, -2097152.00000659, -1.44147367012011e-05, -2097152.000009, 4.32424430418547e-06, -2.58175432169258e-05, -2.14943564332281e-05, -2097152.00001058, -2097152.00000375, -699050.666717666, -4194304.0000183, -2097152.00003945, -2097152.00003167}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.46800050151074e-08, 0, -7.30180523398769e-08, -2389.33333344232, 0, 0, 0, 0, 0, 0, 1.6142126978144e-07, 0, 3.95360052721685e-07, 0, 0, 4.41285790673895e-07, 33792.0000005037, 24234.6666674125, 5120.00000121008, 0, 0, 0, 0, 0, 0, -3.55593080860577e-07, 0, -1.0236131160355e-06, 0, 0, 0, -9.35997317199985e-07, 0, -1.98472611982035e-06, -140629.33333429, -43690.6666688658, -2730.66667186111, -337237.333336491, -66901.3333377915, -46421.3333387936, 0, 0, 0, 0, 0, 0, 4.8895332501091e-07, 0, 1.36587108914909e-06, 0, 0, 0, 9.18460629429208e-07, 0, 3.98749249768192e-06, 196608.000000905, 21845.3333368235, 9.88943277131063e-06, 0, 0, 3.28010350638663e-06, 1354410.66667178, 229376.000006348, 567978.666674696, 43690.6666753426, 32768.0000147629, 21845.3333509579, 600746.666682509, 0, 0, 0, 0, 0, 0, -3.68232838303447e-07, 0, -9.16785680808754e-07, 0, 0, 0, -4.88375695629191e-07, 0, -3.55300091482457e-06, -87381.3333337805, -2.87081838596593e-06, -8.57514729548378e-06, 0, 0, 0, -2.02535321797188e-06, 0, -4.79773351227052e-06, -1747626.66667027, -174762.66667056, -262144.00000914, -6.15614746618967e-06, -1.88655805327524e-05, -2.30013633106069e-05, -1922389.33334315, -87381.333348186, -1.78257523650703e-05, -2009770.66668215, -524288.000019029, -262144.000033885, 0, 0, 0, 0, 0, 0, 1.1274319859134e-07, 0, 2.46621967464836e-07, 0, 0, 0, 1.39981871594771e-07, 0, 1.16058011409714e-06, 1.00499962880021e-07, 9.46766875927776e-07, 2.78568222059548e-06, 0, 0, 0, 3.6810198926919e-07, 0, 2.26491317474113e-06, 699050.66666762, 8.55697500590548e-07, 3.78836524001334e-06, 1.78378369212213e-06, 8.12424817118533e-06, 1.01269132396963e-05, 0, 0, 1.69219533828363e-06, 1398101.3333372, 4.97802729958106e-06, 699050.666672628, 5.40569505581786e-06, 9.71766129886857e-06, 1.21414011625432e-05, 1398101.33333809, 8.10714217873764e-06, 2.21115546700678e-05, 1398101.3333437, 699050.666683698, 1.80309080755036e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.61157629739974e-07, 0, 5.81034842404595e-08, -2389.33333332413, 0, 0, 0, 0, 0, 0, -8.97936439230335e-07, 0, -7.41819064736901e-08, 0, 0, -1.14500268310584e-06, 24234.6666668016, 33791.9999992313, 5120.00000051419, 0, 0, 0, 0, 0, 0, 2.24993380129484e-06, 0, -1.15080744501837e-07, 0, 0, 0, 5.71774572045284e-06, 0, 2.42644756708762e-06, -43690.6666668825, -140629.33332906, -2730.66666996164, -337237.333332214, -46421.3333369587, -66901.3333313174, 0, 0, 0, 0, 0, 0, -2.80741349194335e-06, 0, 3.56940144477795e-07, 0, 0, 0, -1.16883096181868e-05, 0, -2.08189480421051e-06, 21845.333333201, 196607.999991153, 6.95817179872322e-06, 0, 0, -1.04086128539222e-05, 567978.666665027, 43690.6666732145, 1354410.66666081, 229375.999983916, 21845.3333473762, 32768.0000030038, 600746.66667068, 0, 0, 0, 0, 0, 0, 1.70497020607883e-06, 0, -3.21884089919127e-07, 0, 0, 0, 1.04369897659023e-05, 0, 7.38043826516931e-07, 3.93543605110176e-07, -87381.3333254437, -6.14747992583401e-06, 0, 0, 0, 1.54446448236152e-05, 0, 7.13253190131456e-06, -262143.999997615, -4.77375605355282e-06, -1747626.66665796, -174762.66664047, -1.85703179979589e-05, -9.51601460116894e-06, -1922389.33332917, -1.7096025691357e-05, -87381.3333193987, -524288.000006798, -2009770.66665719, -262144.000020648, 0, 0, 0, 0, 0, 0, -4.0160338901393e-07, 0, 1.04581439888997e-07, 0, 0, 0, -3.37949569248942e-06, 0, -8.69682103379362e-08, -1.79179206830078e-07, -2.5539358456082e-06, 1.97204549297748e-06, 0, 0, 0, -7.344510327388e-06, 0, -1.63735277916965e-06, -1.27093214427284e-06, 1.2536652640413e-06, 699050.666662542, -1.26266223251177e-05, 8.01631617735708e-06, 5.16994357252686e-06, 0, 0, -5.81286944161831e-06, 699050.666664804, 6.66607887939084e-06, 1398101.33333059, -1.54413325056691e-05, 1.13168345189418e-05, -1.58787275933222e-06, 1.96266310344037e-06, 1398101.33332212, 1.64961821411205e-05, 1398101.33333236, 1.35772555959181e-05, 699050.666668026}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.0624023598952e-07, 0, 8.54432965620911e-08, 7168.00000008235, 0, 0, 0, 0, 0, 0, -3.41521913372534e-07, 0, -1.78572004659588e-07, 0, 0, -1.84220532710799e-06, -82261.3333341282, -24917.3333339015, -82261.3333342085, 0, 0, 0, 0, 0, 0, 1.92427702713906e-07, 0, -3.00965239715092e-07, 0, 0, 0, 7.18277115428823e-06, 0, 6.53090149741007e-06, 228010.666668977, 28672.0000002929, 228010.666669327, 256682.666671587, 914773.333338841, 256682.666674004, 0, 0, 0, 0, 0, 0, 5.51009861327202e-07, 0, 1.24769116564749e-06, 0, 0, 0, -1.08175387936404e-05, 0, -8.6556150847968e-06, -240298.666669578, -10922.6666641687, -240298.666669143, 0, 0, -2.22306989107064e-05, -491520.000011523, -2315605.33334541, -262144.000009808, -262144.000013014, -2315605.33334696, -491520.00002387, -2577749.3333572, 0, 0, 0, 0, 0, 0, -8.2110223002625e-07, 0, -1.26320735397787e-06, 0, 0, 0, 6.9849545918622e-06, 0, 4.32679650521319e-06, 87381.3333349872, -4.18868404521702e-06, 87381.3333332463, 0, 0, 0, 2.508461453926e-05, 0, 2.35311434478663e-05, 262144.00001009, 2184533.33334437, 87381.3333408394, 87381.3333394582, 2184533.33334591, 262144.000026105, 262144.000019037, 4194304.00002035, 262144.000037226, 4456448.0000313, 2359296.0000265, 4456448.00004574, 0, 0, 0, 0, 0, 0, 3.18518503586182e-07, 0, 4.15317378458505e-07, 0, 0, 0, -1.58341918623769e-06, 0, -4.76217871825563e-07, -3.36295284664298e-07, 1.89310728073497e-06, 6.90580733189363e-07, 0, 0, 0, -9.22161675105234e-06, 0, -7.96038182773091e-06, -2.98791985739474e-06, -699050.666670281, -1.71187808533255e-06, 5.2662911841993e-07, -699050.666670246, -8.72161583708175e-06, 0, 0, -1.43878181773962e-05, -9.23985662234781e-06, -2097152.00000958, -9.87501673640299e-06, -1.53096352707836e-05, -2097152.00000998, -2.40056500297066e-05, -2097152.00001333, -699050.666674295, -2097152.00002584, -2097152.00001919, -4194304.00002743, -2097152.00002807}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.37232623640593e-08, 0, -6.16590436727062e-08, -2389.33333339366, 0, 0, 0, 0, 0, 0, 9.66571485419202e-08, 0, 2.49320690249623e-07, 0, 0, 8.17454461204266e-07, 33792.000000318, 5120.0000005378, 24234.6666672727, 0, 0, 0, 0, 0, 0, 7.29031078733927e-08, 0, -3.8115953356489e-07, 0, 0, 0, -2.67809760709443e-06, 0, -3.28687491065126e-06, -140629.333334073, -2730.66666801465, -43690.6666687534, -66901.3333357204, -337237.33333551, -46421.3333383095, 0, 0, 0, 0, 0, 0, -4.43704129310943e-07, 0, 1.85839891988367e-07, 0, 0, 0, 3.48483529213676e-06, 0, 5.24288174268201e-06, 196608.000000878, 1.19751872901291e-06, 21845.3333363654, 0, 0, 9.17682623709005e-06, 229376.000004087, 1354410.66667047, 32768.0000059312, 21845.3333439244, 567978.66667203, 43690.6666816389, 600746.666678536, 0, 0, 0, 0, 0, 0, 4.81369441135592e-07, 0, 5.98475161890166e-08, 0, 0, 0, -1.85209911871443e-06, 0, -3.60493627523336e-06, -87381.3333338388, -1.18057319211092e-07, -1.89137219152541e-06, 0, 0, 0, -9.24566062497235e-06, 0, -1.07158274322442e-05, -174762.666669684, -1747626.66667008, -6.09625209573686e-06, -8.27549354898714e-06, -262144.000005233, -1.70222495257999e-05, -87381.3333409614, -1922389.33333921, -2.34859208125415e-05, -2009770.66667863, -262144.000016854, -524288.00002235, 0, 0, 0, 0, 0, 0, -1.67156008612958e-07, 0, -5.60482466986633e-08, 0, 0, 0, 2.8698200842216e-07, 0, 8.76176396061258e-07, 1.06252633803698e-07, -2.11167654579553e-07, 3.99440608600944e-07, 0, 0, 0, 3.08895189070977e-06, 0, 4.0727881316936e-06, 7.98395914575904e-07, 699050.666667874, 2.18183238971108e-06, 1.98391745568585e-06, 1.7213587028788e-06, 6.42700380175789e-06, 0, 0, 5.77544460943597e-06, 2.99912238807713e-06, 1398101.3333359, 4.53314257082299e-06, 1.04096325242124e-05, 699050.666669348, 1.43793065199226e-05, 1398101.33333784, 7.27758243412499e-06, 1.28093370882851e-05, 699050.666675187, 1398101.33334393, 1.62403100578293e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.90445490186067e-08, -2389.33333330388, 0, 0, 0, 0, 0, 0, 2.78352502733207e-08, 0, -3.5272951555654e-07, 0, 0, 3.39209658567939e-07, 24234.6666668413, 5119.99999958751, 33791.9999997771, 0, 0, 0, 0, 0, 0, 1.14877964147658e-07, 0, 1.27092087125687e-06, 0, 0, 0, -2.28423023225441e-06, 0, -5.4768619699477e-07, -43690.6666674888, -2730.66666436373, -140629.333332008, -46421.3333335842, -337237.333334799, -66901.3333322889, 0, 0, 0, 0, 0, 0, -4.61813628016322e-07, 0, -2.06634250494423e-06, 0, 0, 0, 4.22644217307741e-06, 0, -1.1568277376163e-06, 21845.3333344729, -5.28651348192604e-06, 196607.99999626, 0, 0, 6.05613058979662e-06, 43690.6666695304, 567978.666671863, 21845.3333312603, 32767.9999955917, 1354410.66666963, 229375.999998187, 600746.666669943, 0, 0, 0, 0, 0, 0, 5.13604838141338e-07, 0, 1.55252389011309e-06, 0, 0, 0, -3.0995346887592e-06, 0, 2.73552561456743e-06, -6.37508245155908e-07, 5.28415037389732e-06, -87381.3333289974, 0, 0, 0, -9.02167446378583e-06, 0, -4.40365033065612e-06, -3.24815173481396e-06, -262144.00000527, 4.86009849794929e-06, 7.41317404597716e-06, -1747626.6666678, -174762.666662607, -2.54336563206039e-06, -1922389.33334203, -87381.3333318743, -524288.000010343, -262143.999996666, -2009770.66667404, 0, 0, 0, 0, 0, 0, -1.87005304765335e-07, 0, -4.40163127279053e-07, 0, 0, 0, 7.73763876625684e-07, 0, -1.36329216121045e-06, 1.19616588426738e-07, -1.92278661051154e-06, -1.72773725089963e-06, 0, 0, 0, 3.85458212895324e-06, 0, 5.17914652524941e-07, 1.04983103633658e-06, 1.7057481299162e-06, -2.78601644160877e-06, -3.91381589558544e-06, 699050.666666084, -3.02568780779289e-06, 0, 0, 4.33396309098069e-06, 2.26362906412621e-06, 699050.666671466, 2.82855493395124e-07, -1.24981683400734e-06, 1398101.33333714, -3.79502645290203e-07, 5.65744120157607e-06, -5.69669130991363e-06, 1398101.33333653, 2.96229201031017e-06, 1398101.33334295, 699050.666667515}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.11975418003123e-08, 0, 2.17315240825498e-08, 341.333333357943, 0, 0, 0, 0, 0, 0, -8.25501087200531e-08, 0, -9.95554042643899e-08, 0, 0, -2.09003026574561e-07, -4437.33333347645, -3072.00000022566, -3072.00000019287, 0, 0, 0, 0, 0, 0, 1.62223516121491e-07, 0, 2.21563264175901e-07, 0, 0, 0, 6.27451005848187e-07, 0, 8.14311649227849e-07, 15018.6666670867, 2730.66666740968, 2730.66666732369, 39594.6666676933, 39594.6666673669, 27306.66666809, 0, 0, 0, 0, 0, 0, -1.79933007504722e-07, 0, -2.55322985036023e-07, 0, 0, 0, -8.40776736619825e-07, 0, -1.44605430758112e-06, -10922.6666673243, -1.16076789044655e-06, -1.06982628340955e-06, 0, 0, -1.89302125850571e-06, -131072.000002103, -131072.000001309, -32768.0000026733, -21845.333336639, -32768.0000015645, -21845.333337124, -349525.333337026, 0, 0, 0, 0, 0, 0, 1.08725776821365e-07, 0, 1.48566553234613e-07, 0, 0, 0, 5.42971361336371e-07, 0, 1.16586129639249e-06, 5.15855542113319e-07, 8.7801640765961e-07, 8.03671646103348e-07, 0, 0, 0, 1.68714046164156e-06, 0, 2.35559855816131e-06, 87381.3333353366, 87381.333334628, 2.91235059063925e-06, 3.19802687619347e-06, 1.63895356425444e-06, 4.40042741785772e-06, 87381.3333369563, 87381.3333349729, 5.71327614540445e-06, 1135957.33333753, 262144.000006183, 262144.000005538, 0, 0, 0, 0, 0, 0, -2.77803406276103e-08, 0, -3.49973585658313e-08, 0, 0, 0, -1.42367852584386e-07, 0, -3.47924456784688e-07, -1.57891615904237e-07, -2.59249436734356e-07, -2.23915054542865e-07, 0, 0, 0, -5.20214991964973e-07, 0, -9.95281272865538e-07, -7.21721101449688e-07, -5.13937040168343e-07, -1.12360683268076e-06, -1.13571939275723e-06, -6.18636596283493e-07, -1.78347600520302e-06, 0, 0, -1.08815111663359e-06, -1.65090092748704e-06, -7.09412596183044e-07, -1.97939460316829e-06, -2.59995233255329e-06, -7.81833527023127e-07, -3.36714202023299e-06, -699050.666668503, -3.28738427750147e-06, -3.03700840862098e-06, -699050.666670279, -699050.666669252, -4.34415281248069e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.50373376215914e-08, 0, 2.80836068969622e-08, 341.333333349426, 0, 0, 0, 0, 0, 0, -3.81330699013662e-08, 0, -1.56637438835798e-07, 0, 0, -2.07217781954907e-07, -3072.00000006321, -4437.33333343965, -3072.00000021724, 0, 0, 0, 0, 0, 0, 4.28791047866025e-08, 0, 3.96594625625226e-07, 0, 0, 0, 4.39338958118893e-07, 0, 1.03838497288588e-06, 2730.66666673573, 15018.6666669118, 2730.66666765431, 39594.6666671129, 27306.6666673653, 39594.66666779, 0, 0, 0, 0, 0, 0, -3.30592019289709e-08, 0, -5.05480829595145e-07, 0, 0, 0, -2.97957753773143e-07, 0, -2.11519006237559e-06, 5.02021455757394e-08, -10922.66666692, -1.90696108362083e-06, 0, 0, -1.9901356988033e-06, -32768.0000005623, -21845.333334257, -131072.000001218, -131072.000001779, -21845.3333357419, -32768.0000041362, -349525.33333604, 0, 0, 0, 0, 0, 0, 2.61490756734144e-08, 0, 3.20624475837741e-07, 0, 0, 0, 0, 0, 1.86328978418777e-06, -1.38621804537505e-07, 1.41435995983735e-07, 1.63563055638439e-06, 0, 0, 0, 1.32961619952427e-06, 0, 3.00918539375238e-06, 1.35495188105723e-07, 5.39204486922768e-07, 87381.3333346436, 87381.3333342097, 3.1969464140948e-06, 5.66360949058143e-06, 87381.3333347565, 2.32017341434516e-06, 87381.3333382158, 262144.000002149, 1135957.33333733, 262144.000005846, 0, 0, 0, 0, 0, 0, -1.10267155325852e-08, 0, -8.12272477624929e-08, 0, 0, 0, 4.41978605396471e-08, 0, -5.96256343805068e-07, 6.82837733194042e-08, -4.3368875043478e-08, -5.16979903498931e-07, 0, 0, 0, -2.37602678956025e-07, 0, -1.41221588460913e-06, 8.63793906269825e-08, -1.09597867654929e-07, -4.90175479773129e-07, -3.33389324364855e-08, -1.41708217487161e-06, -2.5299729313825e-06, 0, 0, -1.19005664055121e-06, -4.242345035922e-07, -8.13205545429154e-07, -9.00322494654756e-07, -1.52485354514093e-06, -1.61338984277591e-06, -3.64520837123879e-06, -4.41180015390283e-07, -699050.666668514, -3.91656317115436e-06, -699050.66666871, -2.27352681325576e-06, -699050.666670533}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.59068115166589e-08, 0, 0, 341.333333338229, 0, 0, 0, 0, 0, 0, -7.99225236937991e-08, 0, -1.58118539013041e-08, 0, 0, -1.53320783231047e-07, -3072.00000004612, -3072.00000005463, -4437.33333336037, 0, 0, 0, 0, 0, 0, 1.7929460545807e-07, 0, 0, 0, 0, 0, 6.98209823386718e-07, 0, 4.34410459489652e-07, 2730.66666679571, 2730.66666681856, 15018.6666667529, 27306.6666670695, 39594.6666668556, 39594.666666974, 0, 0, 0, 0, 0, 0, -2.02254831228147e-07, 0, 3.60286737720282e-08, 0, 0, 0, -1.29265591057138e-06, 0, -4.85096264997391e-07, -1.36370252104772e-07, -1.83789602250742e-07, -10922.6666667188, 0, 0, -1.68572497097757e-06, -21845.3333343697, -32768.0000003488, -21845.3333343734, -32768.0000006114, -131072.00000062, -131072.00000108, -349525.333334108, 0, 0, 0, 0, 0, 0, 1.10947227586495e-07, 0, -2.93967823858581e-08, 0, 0, 0, 1.0731435073332e-06, 0, 1.97631956270833e-07, 5.41239960124389e-08, 1.08180393294344e-07, -6.98536562683432e-08, 0, 0, 0, 2.19923814383252e-06, 0, 1.5279128841365e-06, 9.83354662896914e-07, 1.54439147843922e-07, 1.06769779985101e-06, 4.46824523711739e-07, 87381.3333340192, 87381.333334579, 2.07648505635352e-06, 87381.3333342605, 87381.333335118, 262144.00000101, 262144.00000069, 1135957.33333507, 0, 0, 0, 0, 0, 0, -2.30919418287815e-08, 0, 0, 0, 0, 0, -3.31974347535447e-07, 0, -1.15280249217081e-08, 0, -2.53828988725381e-08, 5.59211716726353e-08, 0, 0, 0, -9.41919148291678e-07, 0, -4.52148152575472e-07, -3.14249608307485e-07, 2.1841254415389e-08, -3.8890377456936e-07, -1.29919778917938e-07, -2.05450629843959e-07, -4.08131999656894e-07, 0, 0, -1.06758740076594e-06, -1.09924940575065e-06, -3.41192072374908e-07, -1.17626215634226e-06, -8.10819678206546e-07, -6.23257571481932e-07, -1.23324915990563e-06, -3.4499690236715e-07, -9.94362208493818e-08, -699050.666667872, -7.4247565611622e-07, -699050.666667613, -699050.666667453}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -1.73300636111986e-08, 0, 0, 0, 0, 0, 0, 0, 1.60536687603693e-07, 0, 2.43950340625273e-07, 7168.00000009994, 0, 0, 0, 0, 0, 0, -5.35568260075388e-07, 0, -1.40025895147922e-06, 0, 0, -1.82031082266481e-06, -24917.3333338019, -82261.3333341811, -82261.3333344116, 0, 0, 0, 0, 0, 0, 1.0346727805361e-06, 0, 3.68481033573005e-06, 0, 0, 0, 3.90770130683953e-06, 0, 9.16144159787971e-06, 28672.0000010124, 228010.666669623, 228010.666672304, 256682.666670376, 256682.666669828, 914773.333338741, 0, 0, 0, 0, 0, 0, -1.29583019309018e-06, 0, -4.8317922919691e-06, 0, 0, 0, -3.49006723306871e-06, 0, -1.96002985105204e-05, -10922.6666675655, -240298.666672418, -240298.666678608, 0, 0, -1.62557795547782e-05, -262144.000007349, -262144.000004003, -491520.000011164, -2315605.33334093, -491520.000012999, -2315605.33335685, -2577749.33334441, 0, 0, 0, 0, 0, 0, 9.22898814259802e-07, 0, 3.09742027764444e-06, 0, 0, 0, 1.50112668532744e-06, 0, 1.82163396111608e-05, 2.38284643937131e-07, 87381.3333391117, 87381.3333441063, 0, 0, 0, 1.04226082090263e-05, 0, 2.54805018325115e-05, 87381.3333394983, 87381.3333346287, 262144.000014252, 2184533.33333787, 262144.000018862, 2184533.33337057, 262144.000015994, 262144.000013249, 4194304.00002345, 2359296.00000915, 4456448.0000151, 4456448.00002898, 0, 0, 0, 0, 0, 0, -2.7000993909619e-07, 0, -7.76124854309913e-07, 0, 0, 0, -3.36694957087605e-07, 0, -6.12133798987998e-06, 3.50719372391457e-08, -2.22879553326989e-06, -3.49325885340466e-06, 0, 0, 0, -2.05494245396254e-06, 0, -1.26230808473436e-05, -1.84160314114323e-06, 3.74613307774319e-07, -6.53166145925552e-06, -699050.666668257, -8.54943922638141e-06, -699050.666684992, 0, 0, -9.26688134783697e-06, -7.33504471314368e-06, -4.22205765124906e-06, -1.00321491500529e-05, -2097152.00000563, -1.0428065679336e-05, -2097152.00002157, -699050.66666813, -2097152.00000785, -2097152.00002319, -2097152.00000916, -2097152.00001176, -4194304.00001508}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7.3258904250553e-08, 0, -1.22516089912258e-07, -2389.33333337299, 0, 0, 0, 0, 0, 0, 2.49829936789316e-07, 0, 7.36337178525008e-07, 0, 0, 7.31425749265099e-07, 5120.00000009393, 33792.0000003959, 24234.6666670724, 0, 0, 0, 0, 0, 0, -5.61429282855217e-07, 0, -2.02123301820961e-06, 0, 0, 0, -1.09754769688186e-06, 0, -3.98886033518423e-06, -2730.66666669304, -140629.333335166, -43690.666669053, -66901.3333344594, -46421.333334008, -337237.333334947, 0, 0, 0, 0, 0, 0, 8.33529369752423e-07, 0, 2.73341149221851e-06, 0, 0, 0, 3.26301365503812e-07, 0, 9.19714808914072e-06, -2.89628405703428e-07, 196608.000004226, 21845.3333389443, 0, 0, 5.41561983592274e-06, 32768.0000013405, 21845.3333335344, 229376.000004841, 1354410.66666902, 43690.666670721, 567978.666674256, 600746.666668451, 0, 0, 0, 0, 0, 0, -6.68582291448245e-07, 0, -1.78649852841627e-06, 0, 0, 0, 2.26746703564414e-07, 0, -9.02661170945525e-06, 4.46902115335306e-07, -87381.3333378341, -5.48151243859249e-06, 0, 0, 0, -1.75654242491518e-06, 0, -9.96753696525935e-06, -4.62493581790768e-07, 7.69586765275103e-07, -174762.666674261, -1747626.66666919, -7.10830984470198e-06, -262144.000013988, -87381.333338137, -3.32544247070793e-06, -1922389.3333385, -262143.999999621, -2009770.66667122, -524288.000006194, 0, 0, 0, 0, 0, 0, 2.11326799721774e-07, 0, 4.51249109893233e-07, 0, 0, 0, -6.31821374369356e-08, 0, 3.15745967807523e-06, -1.94419600345752e-07, 1.7523010643885e-06, 1.88701438224865e-06, 0, 0, 0, -3.52270351941369e-07, 0, 5.4449195071579e-06, -9.02439442922003e-08, -5.94761098767964e-07, 3.90396055828837e-06, 699050.66666822, 3.63461265559526e-06, 7.6690533318981e-06, 0, 0, 2.66155094118187e-06, 1.56357623388924e-06, 8.28066279712121e-07, 3.82286418690604e-06, 1398101.33333356, 3.43609038172833e-06, 699050.666672912, -1.37339030932494e-06, 1398101.33333747, 6.73082342924462e-06, 699050.666668574, 1.26493006290983e-06, 1398101.33333592}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.15531486547313e-08, 0, -4.40799324916812e-08, -2389.33333333856, 0, 0, 0, 0, 0, 0, 3.76755615826144e-08, 0, 2.81319052543764e-07, 0, 0, 3.6462879718169e-07, 5120.00000008647, 24234.6666666516, 33792.0000001343, 0, 0, 0, 0, 0, 0, 0, 0, -7.62116154543068e-07, 0, 0, 0, -6.71269569507748e-07, 0, -2.14803522961038e-06, -2730.66666697761, -43690.6666663891, -140629.333334333, -46421.3333337129, -66901.3333340272, -337237.333333744, 0, 0, 0, 0, 0, 0, -6.91294385254119e-08, 0, 9.99261705888093e-07, 0, 0, 0, 3.11427054048018e-07, 0, 4.82451782374399e-06, 4.19777219208798e-07, 21845.333332783, 196608.000002216, 0, 0, 3.87528524564481e-06, 21845.3333349936, 32768.0000011807, 43690.6666670626, 567978.666665485, 229376.000003348, 1354410.66667131, 600746.66666807, 0, 0, 0, 0, 0, 0, 7.10779917510411e-08, 0, -6.30945038171782e-07, 0, 0, 0, 1.40296659193209e-07, 0, -4.56374329641658e-06, -2.68876788149559e-07, 3.16691814555003e-07, -87381.3333352984, 0, 0, 0, -2.43517683088192e-06, 0, -6.54865515941371e-06, -2.02707135995992e-06, -3.34124581633332e-07, -1.11446325945942e-07, -262143.999997074, -174762.666671318, -1747626.66667508, -2.53849867349766e-06, -87381.3333377241, -1922389.33333692, -262144.000002022, -524287.999998873, -2009770.66667407, 0, 0, 0, 0, 0, 0, -2.54410826158511e-08, 0, 1.54023364483236e-07, 0, 0, 0, -1.06216192841478e-07, 0, 1.54405530973823e-06, 7.59205918279389e-08, -2.91800956708084e-08, 6.19111580787516e-07, 0, 0, 0, 4.46875992390761e-07, 0, 3.32001781582694e-06, 8.12491356764422e-07, -2.11243190136618e-07, 5.6135377127992e-08, -1.40012951278555e-06, 1.93385248659149e-06, 699050.666670902, 0, 0, 2.45761062091015e-06, 1.85389100994229e-06, 1.49915218743635e-06, 1.17728574095602e-06, 699050.666666551, 3.28126391880671e-06, 1398101.33333831, 4.28020304830322e-07, -1.89284909880199e-06, 1398101.33333959, 4.98467159106563e-07, 699050.666670907, 1398101.33333585}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -8.29317026918646e-08, 0, -9.36726180374627e-08, -1792.0000000746, 0, 0, 0, 0, 0, 0, 3.15014222507168e-07, 0, 4.45530376187719e-07, 0, 0, 9.19784108593559e-07, 18176.00000041, 18176.0000006343, 18176.0000007137, 0, 0, 0, 0, 0, 0, -5.93639457503508e-07, 0, -1.01191009304914e-06, 0, 0, 0, -2.83351315362001e-06, 0, -3.7114646706124e-06, -32768.0000010022, -32768.0000018696, -32768.0000028369, -180224.000003071, -180224.000002535, -180224.000004675, 0, 0, 0, 0, 0, 0, 6.29160193284405e-07, 0, 1.19974358991882e-06, 0, 0, 0, 3.87697982956446e-06, 0, 6.58915743785302e-06, 16384.0000012113, 16384.0000026109, 16384.0000049888, 0, 0, 8.99270146399016e-06, 294912.000006067, 294912.000004035, 294912.000008091, 294912.000009349, 294912.000007367, 294912.000014769, 1753088.00001156, 0, 0, 0, 0, 0, 0, -3.65543467930334e-07, 0, -7.31231906698976e-07, 0, 0, 0, -2.54532637745209e-06, 0, -5.2724541581684e-06, -7.10837804765033e-07, -1.82925153692179e-06, -3.92255969930813e-06, 0, 0, 0, -8.3861499328649e-06, 0, -1.11668242855231e-05, -131072.000005138, -131072.000003058, -131072.000008677, -131072.000007437, -131072.000008928, -131072.000018591, -262144.000011711, -262144.000007662, -262144.000020308, -2621440.00001166, -2621440.00001759, -2621440.00002117, 0, 0, 0, 0, 0, 0, 9.0543107044468e-08, 0, 1.83796931076298e-07, 0, 0, 0, 6.74044726659453e-07, 0, 1.56694989012867e-06, 1.58958598319043e-07, 5.26648361934848e-07, 1.13971264297314e-06, 0, 0, 0, 2.72009703224556e-06, 0, 4.64602517067281e-06, 1.57302919011048e-06, 9.17371795538128e-07, 3.27968806919364e-06, 2.14090199791458e-06, 3.64823029463942e-06, 7.78250978334571e-06, 0, 0, 5.39970801620094e-06, 5.24094714118525e-06, 2.86172285545837e-06, 6.67098121054799e-06, 8.10454366735207e-06, 4.78918348954404e-06, 1.35382288136681e-05, 1048576.00000411, 1048576.00000849, 1048576.00001325, 2097152.00001026, 2097152.00000914, 2097152.00001454}};


// -----------------------------------------------------------------------------
double MAd_TetInterpolant::coef9[220][220] = {{1, -25.4607142859045, -25.4607142867922, -25.4607142859045, 261.763392861411, 261.763392882119, 261.763392861411, 523.526785702391, 523.526785726581, 523.526785719153, -1453.82142860868, -1453.82142880233, -1453.82142860868, -4361.4642855827, -4361.46428579421, -4361.46428556995, -4361.46428586122, -4361.46428575912, -4361.46428588221, -8722.92857222817, 4869.49218766938, 4869.49218861218, 4869.49218766938, 19477.9687491613, 19477.9687506128, 19477.9687491835, 19477.9687512531, 19477.9687501975, 19477.9687508392, 29216.9531242154, 29216.9531254582, 29216.9531260588, 58433.9062550756, 58433.9062589582, 58433.9062594417, -10295.9859379464, -10295.9859406034, -10295.9859379464, -51479.9296843394, -51479.9296898005, -51479.9296846754, -51479.9296924535, -51479.9296881675, -51479.9296902185, -102959.859372045, -102959.859378347, -102959.859373449, -102959.859384325, -102959.859375365, -102959.859373867, -205919.718765398, -205919.718791006, -205919.718794453, -308879.578175702, -308879.578170143, -308879.578206649, 13839.6093757094, 13839.6093802021, 13839.6093757094, 83037.6562433235, 83037.6562546245, 83037.6562443094, 83037.6562601009, 83037.6562515817, 83037.656256375, 207594.140614811, 207594.140636428, 207594.14062144, 207594.140658857, 207594.140623711, 207594.140609443, 415188.281276156, 415188.281345717, 415188.281363505, 276792.187503953, 276792.187503529, 276792.187508632, 830376.562622449, 830376.562602845, 830376.562681664, 830376.562792769, 830376.562649471, 830376.562763933, 1245564.84411397, -11467.1049113774, -11467.1049158742, -11467.1049113774, -80269.7343672505, -80269.7343801146, -80269.7343687106, -80269.7343863053, -80269.7343771602, -80269.7343840646, -240809.20310543, -240809.203143664, -240809.203116979, -240809.203178511, -240809.203122754, -240809.203095985, -481618.406275609, -481618.406370554, -481618.406414276, -401348.671881689, -401348.671888761, -401348.671892891, -401348.671928706, -401348.671870962, -401348.671838731, -1204046.01577342, -1204046.01575224, -1204046.01593089, -1204046.01613358, -1204046.01588509, -1204046.01604227, -1605394.68781747, -1605394.68772603, -1605394.6881689, -2408092.03192159, -2408092.03223913, -2408092.0320844, 5338.13504498354, 5338.13504743151, 5338.13504498354, 42705.0803525194, 42705.0803600839, 42705.0803536622, 42705.080363832, 42705.080358644, 42705.0803638235, 149467.781232523, 149467.781264533, 149467.781241474, 149467.781288606, 149467.78124908, 149467.781230018, 298935.562513531, 298935.562578183, 298935.562624043, 298935.562498467, 298935.562518243, 298935.56251543, 298935.562573823, 298935.562491888, 298935.562435216, 896806.687590301, 896806.687580639, 896806.687747495, 896806.687918214, 896806.687740948, 896806.687845817, 373669.453155182, 373669.453126115, 373669.453124848, 1494677.81274275, 1494677.81267308, 1494677.81284691, 1494677.81328065, 1494677.8127267, 1494677.81311697, 2242016.71930867, 2242016.71985211, 2242016.71956465, 2989355.62621182, 2989355.62598736, 2989355.62650457, -1067.62700900182, -1067.62700956039, -1067.62700900182, -9608.64307924641, -9608.64308104497, -9608.6430796211, -9608.6430820192, -9608.64308076795, -9608.64308228802, -38434.5723156455, -38434.5723257873, -38434.5723182799, -38434.572331893, -38434.5723214691, -38434.5723168993, -76869.1446458459, -76869.1446635591, -76869.1446801927, -89680.6687461714, -89680.6687578727, -89680.6687535395, -89680.6687802778, -89680.6687468462, -89680.6687220344, -269042.006272367, -269042.006270345, -269042.006328077, -269042.006380254, -269042.006340772, -269042.006369729, -134521.003136446, -134521.003128623, -134521.003141053, -134521.003143923, -134521.003121441, -134521.003100791, -538084.012568283, -538084.012553246, -538084.012637378, -538084.012830277, -538084.012595492, -538084.012711735, -807126.018923253, -807126.019181469, -807126.019061138, -672605.015753403, -672605.015701367, -672605.015970321, -1345210.03174448, -1345210.03165619, -1345210.03190793, -1345210.03210253, -1345210.03169936, -1345210.03193773, -1793613.37591859}, 
                                                    {0, 1.00000000011252, 0, 0, -24.4607142882813, 0, 0, 0, 0, 0, 237.302678594408, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1216.51875010777, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3652.97343779371, 0, 0, 0, 0, 0, 0, 0, 0, 1.81653696529338e-08, 1.81265633614203e-08, 1.29960255890182e-08, 0, 1.30857635838311e-08, 0, 3.64115848583424e-08, 1.34899842407776e-08, 1.3416414645917e-08, 3.91085350770481e-08, 3.87043144201016e-08, 2.00466097055596e-08, -6643.01250048095, 0, 0, -1.51407313353502e-08, -1.55449519922968e-08, 0, 0, 0, 0, -4.00247799348108e-08, -4.03918107849677e-08, -2.36374333439926e-08, 0, -2.39478730179599e-08, 0, -8.10229181524846e-08, -2.38759234605368e-08, -2.39042172722748e-08, -4.54267734539826e-08, -4.52489175017945e-08, 0, -1.36705551767591e-07, -1.36576204567973e-07, -9.42522291325076e-08, -4.84471058825875e-08, -9.52449852320569e-08, -4.6877111969934e-08, -1.40599013320752e-07, 7196.59687546633, 0, 0, 1.40801788006684e-08, 1.42725866680848e-08, 0, 0, 0, 0, 4.93529910272628e-08, 4.88161902012507e-08, 2.47079263715477e-08, 0, 2.4445185431432e-08, 0, 9.79363505848596e-08, 2.49415670339204e-08, 2.4239842133511e-08, 7.60054206239147e-08, 7.61929754844459e-08, 6.03413070621173e-08, 0, 6.01731500182723e-08, 0, 2.27059047119837e-07, 2.26360555188876e-07, 1.23075579949727e-07, 6.206489899796e-08, 1.23321370892882e-07, 6.14763564499299e-08, 2.42826899921056e-07, 2.43719398440589e-07, 8.33880875461546e-08, 3.63412482329295e-07, 2.43784086251253e-07, 2.44275611294142e-07, -4270.50803596069, 0, 0, 0, 0, 0, 0, 0, 0, -3.0324194710829e-08, -3.06297849306247e-08, -1.37107392106373e-08, 0, -1.38805127392061e-08, 0, -6.05222680860606e-08, -1.37018467682992e-08, -1.3579771973582e-08, -5.93399107629011e-08, -6.12963333423977e-08, -4.15032985756625e-08, 0, -4.11863894100861e-08, 0, -1.83203439974022e-07, -1.8333281559535e-07, -8.30357009817817e-08, -4.1090991942383e-08, -8.19362213633212e-08, -4.14127541148446e-08, -6.78511753449129e-08, -6.66579182961868e-08, 0, -2.64743107436516e-07, -2.67459483893617e-07, -2.04468676656688e-07, -6.80128522390078e-08, -2.08724316053122e-07, -7.07356804241499e-08, -3.94864059671818e-07, -2.04442812901107e-07, -2.03770184725727e-07, -4.1351631807629e-07, -4.10774077863607e-07, -2.76203593330138e-07, 1067.62700898321, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.52027936906052e-08, 0, 0, 1.8981442195809e-08, 1.86095583387669e-08, 1.13258789014026e-08, 0, 1.11714673067809e-08, 0, 5.71569387375348e-08, 5.70534695043534e-08, 2.26849046214284e-08, 1.14358265079773e-08, 2.3009896210624e-08, 1.11399387492384e-08, 2.64562984853001e-08, 2.65630077933565e-08, 2.23340421712237e-08, 0, 2.23542535593424e-08, 0, 1.07054006548424e-07, 1.0598688504615e-07, 6.6547784172144e-08, 2.16703135436092e-08, 6.83813254909182e-08, 2.20907026005079e-08, 1.57560691604886e-07, 6.58331202885165e-08, 6.74111930720755e-08, 1.11393717361352e-07, 1.09414656890294e-07, 2.80230523230784e-08, 2.19204423501651e-07, 2.2305908942144e-07, 1.7022409792844e-07, 1.12874772639771e-07, 1.67966931030605e-07, 1.13579744720482e-07, 2.21817288093007e-07}, 
                                                    {0, 0, 1.00000000007104, 0, 0, -24.460714287326, 0, 0, 0, 0, 0, 237.30267858576, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1216.51875006672, 0, 1.44337845842113e-08, 0, 2.66615771004335e-08, 0, 0, 0, 3.58831616646728e-08, 0, 0, 0, 0, 0, 0, 3652.97343768006, 0, -3.98276718271973e-08, 0, -9.20174676777053e-08, 0, 0, 0, -1.5944909322326e-07, 0, -1.74896903319941e-07, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6643.01250029135, 0, 6.24415026351244e-08, 0, 1.8485829851406e-07, 0, 0, 0, 3.46831147150274e-07, 0, 4.27656114136005e-07, 0, 0, 0, 0, 0, 0, 6.24436182113347e-07, 0, 0, -1.58744333695876e-08, 0, -1.20026656931311e-08, 0, 0, 0, 0, 0, 7196.59687527798, 0, -5.56614465629934e-08, 0, -2.10205765923038e-07, 0, 0, 0, -4.01984745067233e-07, 0, -5.47679224234932e-07, 0, 0, 0, 0, 0, 0, -1.01709328794689e-06, 0, -1.12758771846035e-06, 0, 0, 0, 1.53683770640303e-08, 0, 0, 1.74612905158317e-08, 0, -2.92880603828032e-08, 1.00644531578739e-08, 0, 4.72782836006595e-08, 4.113128886595e-08, -7.43437745403588e-08, -3.30192224557401e-08, 0, -4270.50803585911, 0, 2.69873023928113e-08, 0, 1.31988899811083e-07, 0, 0, 0, 2.3900329236609e-07, 0, 3.61788673021327e-07, 0, 0, 0, 0, 0, 0, 7.88844374908009e-07, 0, 9.46687026983561e-07, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.20770585665197e-06, 0, 0, -4.66109925274136e-08, 0, 1.65031768454633e-08, 4.87242577510307e-08, 0, -7.23722076928307e-08, 3.66185530213321e-08, 5.70281172496452e-08, 0, -2.10344904038894e-08, 8.98556769394008e-08, -3.90937507190326e-08, 0, 1067.62700896006, 0, 0, 0, -3.35483583885232e-08, 0, 0, 0, -5.55380593971321e-08, 0, -9.46602813708807e-08, 0, 0, 0, 0, 0, 0, -2.35879346431406e-07, 0, -3.14303725990598e-07, 0, 0, 0, 0, 0, 0, 1.00251557358303e-08, 0, 0, -4.59810110335629e-07, 0, -5.12896417149151e-07, 0, 0, 0, 0, 0, 0, -1.81150747803221e-08, 0, -1.76379548182625e-08, -1.24838732451679e-08, -2.07669088680476e-08, 1.35218479672982e-08, 2.70986695697506e-08, 0, -1.34152420681417e-08, 0, 0, -4.25771237712758e-08, -4.98856506972103e-08, 0, 0, -1.63464589884098e-08}, 
                                                    {0, 0, 0, 1.00000000011252, 0, 0, -24.4607142882813, 0, 0, 0, 0, 0, 237.302678594408, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1216.51875010777, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3652.97343779371, 0, 0, 0, 0, 0, 0, 0, 1.30857635838311e-08, 0, 1.3139121790573e-08, 1.81354558037583e-08, 1.80804811122925e-08, 1.33590170037223e-08, 1.34544109187118e-08, 3.62628256311837e-08, 2.0061161620788e-08, 3.90664993688006e-08, 3.95871353475741e-08, 0, 0, -6643.01250048095, 0, 0, 0, 0, -1.53396086943758e-08, -1.53412251790996e-08, 0, -2.33698393969917e-08, 0, -2.35113155611089e-08, -4.08558520348379e-08, -4.0682849089535e-08, -2.40230590975443e-08, -2.40182060906591e-08, -8.08127254003921e-08, 0, -4.55270168231436e-08, -4.5407368531869e-08, -4.74527226401733e-08, -9.55877652586423e-08, -4.7564284955115e-08, -9.56362669057853e-08, -1.35664294020899e-07, -1.37035385705531e-07, -1.42591005669601e-07, 0, 0, 7196.59687546633, 0, 0, 0, 0, 1.41214080429108e-08, 1.43752600934022e-08, 0, 2.43861677517998e-08, 0, 2.44176980856992e-08, 4.86642051100716e-08, 4.90942966280272e-08, 2.45397728804164e-08, 2.46262743530679e-08, 9.79234187070688e-08, 0, 6.13534751892075e-08, 0, 5.95910734091376e-08, 7.45728669926393e-08, 7.53813083065324e-08, 6.16186355273385e-08, 1.24071590335006e-07, 6.09654193794995e-08, 1.21064203995047e-07, 2.28468991281261e-07, 2.25998405767314e-07, 8.07719828799236e-08, 2.38920506490103e-07, 2.40796055095416e-07, 2.50368032084225e-07, 2.51144115281932e-07, 3.63347794518631e-07, 0, 0, -4270.50803596069, 0, 0, 0, 0, 0, 0, 0, -1.36444491261045e-08, 0, -1.381987857485e-08, -3.06330179000724e-08, -3.07187093540051e-08, -1.38328140053545e-08, -1.36355566837665e-08, -6.1081706803634e-08, 0, -4.12187262099906e-08, 0, -4.05477180720482e-08, -6.09438615128965e-08, -6.0229197629269e-08, -4.10149993967934e-08, -8.21076042711866e-08, -4.19091321646192e-08, -8.27769923716914e-08, -1.83125848707277e-07, -1.8334577589485e-07, 0, -6.56101803997444e-08, -6.68810429260702e-08, -6.9193177409943e-08, -2.05387038931804e-07, -6.8656376583931e-08, -2.06771119337645e-07, -2.61108368704299e-07, -2.62712319454295e-07, -2.07961136311496e-07, -2.03239835627755e-07, -3.95329720959126e-07, -2.7941143798671e-07, -4.09170183957031e-07, -4.10049779020483e-07, 0, 0, 1067.62700898321, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.5363674776836e-08, 0, 1.11302362881815e-08, 0, 1.12308882194156e-08, 1.91277678140978e-08, 1.91027069718075e-08, 1.14560378960959e-08, 2.26372058875768e-08, 1.13549827318593e-08, 2.21359748309169e-08, 5.62773578849374e-08, 5.60833370855107e-08, 0, 2.23720384440185e-08, 0, 2.20939355699556e-08, 2.64854023157568e-08, 2.69429776267316e-08, 2.22944294137051e-08, 6.73368134584962e-08, 2.26169944994581e-08, 6.70134454594518e-08, 1.06342582739671e-07, 1.08108224594616e-07, 6.71072228897174e-08, 6.87337973204194e-08, 1.59177574232672e-07, 2.76673404187022e-08, 1.11122091084326e-07, 1.10895712168713e-07, 1.11658891910338e-07, 1.67921655247483e-07, 1.11587752371634e-07, 1.70340513250267e-07, 2.21416314616363e-07, 2.20808374251646e-07, 2.23278959765594e-07}, 
                                                    {0, 81.0000000015749, 0, 0, -1333.31785717827, -2.05863971420212e-07, 0, -1333.31785713918, -2.48360969451413e-08, -1333.31785716346, 9202.97410745311, 1.94889631928927e-06, 0, 18405.9482140085, 18405.9482146063, 9202.97410744655, 1.90339974581959e-07, 9202.97410735478, 3.92070698453923e-07, 18405.9482162674, -34932.7687514189, -9.6114901054451e-06, 0, -104798.306247283, -104798.306252409, -34932.7687534811, -1.41832802438714e-06, -34932.7687510394, -2.86521807509178e-06, -104798.306251055, -104798.306252066, -9.5529059490931e-07, -209596.612512943, -104798.306272133, -104798.306271533, 80033.948441263, 2.74040629990315e-05, 0, 320135.793738017, 320135.793758869, 80033.9484525224, 5.94942622007868e-06, 80033.9484411397, 1.14971874428523e-05, 480203.690622452, 480203.690639504, 320135.793769669, 4.43822397533353e-06, 320135.793752706, 1.3431333844544e-06, 960407.381289228, 320135.793856219, 320135.793852212, 960407.381377165, 960407.381357195, 480203.690798063, -113669.325006006, -4.67850772919836e-05, 0, -568346.624973212, -568346.62501756, -113669.325035124, -1.35453751928834e-05, -113669.325008545, -2.67076645647277e-05, -1136693.24996737, -1136693.25004759, -568346.625061474, -1.95666407415862e-05, -568346.624997376, -6.13892375812083e-07, -2273386.50006646, -568346.625260703, -568346.625269456, -1136693.25006502, -1136693.25002006, 7.62157315449968e-06, -3410079.75029845, -3410079.75024869, -2273386.50048714, -1136693.25062068, -2273386.50035704, -1136693.25053778, -3410079.75078603, 97984.4343806672, 4.71178209685932e-05, 0, 587906.606218271, 587906.606269211, 97984.434422432, 1.73877158736333e-05, 97984.4343867201, 3.42377472594541e-05, 1469766.51554587, 1469766.51570089, 587906.606334001, 3.22536406736641e-05, 587906.60624308, 9.04117726490447e-06, 2939533.03131611, 587906.606592374, 587906.606649475, 1959688.68756885, 1959688.68756474, 1469766.51580616, 1.34051039605742e-05, 1469766.51562002, -5.61420334882285e-05, 5879066.06284653, 5879066.06281738, 2939533.03213404, 1469766.51674033, 2939533.03189335, 1469766.51650611, 5879066.0633404, 5879066.06305764, 1959688.68877727, 8818599.09517425, 5879066.06467624, 5879066.06421138, -46975.5883957798, -2.57846929086213e-05, 0, -328829.118730965, -328829.118760914, -46975.5884272871, -1.1731048206156e-05, -46975.5884011652, -2.26805458735706e-05, -986487.356174476, -986487.356308512, -328829.118806224, -2.27702997396253e-05, -328829.118746999, -1.88061033377953e-05, -1972974.71253565, -328829.118979354, -328829.119055268, -1644145.59375658, -1644145.5938285, -986487.356422141, -3.48102605994359e-05, -986487.356225394, 7.06909654889931e-05, -4932436.78145127, -4932436.78146104, -1972974.71325762, -986487.357201467, -1972974.71310826, -986487.357031519, -1644145.59393993, -1644145.5937693, 4.42004039490896e-05, -6576582.37561827, -6576582.37544659, -4932436.78226829, -1644145.59524627, -4932436.78183359, -1644145.5949031, -9864873.56366354, -4932436.7838251, -4932436.78295926, -9864873.56512637, -9864873.56448497, -6576582.37798488, 9608.64308098962, 5.90194341755922e-06, 0, 76869.1446382971, 76869.1446453686, 9608.64309065976, 3.33567696971664e-06, 9608.64308271047, 5.96289688201307e-06, 269042.006224391, 269042.006267538, 76869.1446585024, 6.01264350770718e-06, 76869.1446431456, 9.71700039078239e-06, 538084.012508069, 76869.1447050364, 76869.1447350275, 538084.012483846, 538084.012532422, 269042.006305741, 1.50320667023972e-05, 269042.006238892, -2.45132850482867e-05, 1614252.03754849, 1614252.03755631, 538084.012747292, 269042.006556674, 538084.012729907, 269042.006536009, 672605.015684996, 672605.01564467, 538084.012600394, -1.51151597400758e-06, 538084.012492492, -4.18522815248068e-05, 2690420.06266347, 2690420.0626456, 1614252.03793791, 538084.013146622, 1614252.03775429, 538084.012911599, 4035630.09410999, 1614252.03857331, 1614252.03818957, 2690420.06287539, 2690420.06270774, 672605.016237567, 5380840.12603423, 5380840.12580845, 4035630.09528284, 2690420.06423901, 4035630.09466438, 2690420.0638183, 5380840.12679322}, 
                                                    {0, -162.000000005839, -2.44918328902124e-08, 0, 3395.63571441735, 5.93053490028537e-07, 0, 1937.63571437657, 3.89273441011015e-08, 1937.63571432061, -27125.3089297289, -5.65877005251705e-06, 0, -36811.8964284602, -36811.896429073, -9686.58750265706, -3.18616600260567e-07, -9686.58750040272, -6.56343684029646e-07, -19373.1750020645, 113455.18125532, 2.81267283130695e-05, 0, 253186.256244743, 253186.256253624, 26275.8937728161, 1.73405536152235e-06, 26275.8937521298, 5.20052588013853e-06, 166006.968761156, 166006.968753782, 2.57881303961794e-06, 332013.937513746, 78827.6812733426, 78827.6812704925, -278309.418764187, -8.0775070570967e-05, 0, -876754.631222738, -876754.631262798, -41826.375091729, -6.17800205851731e-06, -41826.375007554, -2.05799130382074e-05, -960407.381257235, -960407.381275364, -403788.543849943, -4.05060483159533e-06, -403788.543756864, -1.84958528826799e-05, -1920814.76254177, -167305.500121706, -167305.500096961, -1211365.63138301, -1211365.63135796, -250958.250153242, 415557.337522752, 0.000138765418344779, 0, 1701349.31243693, 1701349.31252445, 39119.9627046124, 1.42864786848041e-05, 39119.9625175526, 4.5359437594831e-05, 2649823.8749248, 2649823.87507874, 572037.187809039, 4.87231208952123e-06, 572037.187500586, 6.12468044404019e-05, 5299647.75006966, 195599.812817915, 195599.812760878, 1896949.1252193, 1896949.12504493, 2.17522280871522e-05, 5690847.37529718, 5690847.37526095, 2288148.75055065, 391199.625586545, 2288148.75035749, 391199.625433449, 3432223.12571723, -372008.700021601, -0.000140473066430998, 0, -1879972.53742491, -1879972.53752593, -19929.0377610748, -1.99317442132367e-05, -19929.0375241443, -5.43412509915023e-05, -3819732.18729845, -3819732.18762105, -471653.88796277, 9.13095154225596e-07, -471653.887491736, -0.000108857720292702, -7639464.37507089, -119574.225439815, -119574.225397633, -3919377.37516781, -3919377.37512194, -2059333.87558854, -2.10752759478569e-05, -2059333.87501826, -5.10084629409346e-05, -11758132.1253157, -11758132.125342, -2358269.4385756, -298935.563630112, -2358269.43815035, -298935.563180576, -8237335.50090496, -8237335.50060754, -398580.751049137, -12356003.2513005, -4716538.8771013, -4716538.87641021, 183631.845546906, 7.72036708391327e-05, 0, 1106061.58120508, 1106061.58126419, 4270.50821357829, 1.41336380549173e-05, 4270.50805311353, 3.42588984622169e-05, 2780100.73105731, 2780100.73134211, 209254.894096699, -6.70776330893315e-06, 209254.893744677, 9.63463645324954e-05, 5560201.46253944, 29893.5565580994, 29893.556563295, 3736694.53122735, 3736694.53138247, 1165848.6943418, 5.74600823459259e-06, 1165848.69372985, 5.62555369687892e-05, 11210083.5939152, 11210083.5939814, 1255529.3634742, 89680.6697582196, 1255529.36311937, 89680.6693693635, 2839887.8442224, 2839887.84381352, 2.71278343207724e-05, 11359551.3756077, 11359551.3755293, 5829243.46995767, 149467.78260354, 5829243.46938885, 149467.782106087, 17039327.0635291, 3138823.40894805, 3138823.40757501, 11658486.9400163, 11658486.9391968, 4185097.87753429, -38434.5723238639, -1.77310707050981e-05, 0, -269042.006239314, -269042.006253093, -4.99888908736631e-05, -4.31160160838672e-06, -5.05733960979303e-06, -8.51149240708186e-06, -807126.018685177, -807126.018777697, -38434.5724267563, 2.84838014731218e-06, -38434.5723213049, -3.28997623723024e-05, -1614252.03750936, -8.67352389485413e-05, -9.71011846145903e-05, -1345210.03119368, -1345210.0313015, -269042.006460531, 5.33966884164174e-06, -269042.00623614, -2.58828950971831e-05, -4035630.09378792, -4035630.09381179, -269042.006580986, -0.000335410669130505, -269042.006487629, -0.000237917867714371, -1345210.03136808, -1345210.0312913, -807126.019011403, -7.14604220204548e-06, -807126.018753996, -7.99108153117857e-06, -5380840.12513652, -5380840.12518019, -1614252.03805633, -0.000617233506344269, -1614252.03777845, -0.000275609752818771, -8071260.18780533, -807126.019945989, -807126.019268405, -4035630.09417164, -4035630.09400565, -0.000496997293876076, -8071260.18843011, -8071260.18821548, -4035630.09534289, -1345210.03286293, -4035630.09449751, -1345210.03225371, -5380840.12654976}, 
                                                    {0, 252.000000012727, 3.70309353125885e-08, 0, -5660.10000028754, -9.0465794856866e-07, 0, -2258.10000023045, -1.67795492273852e-08, -2258.10000003229, 48984.0750025376, 8.70080718481303e-06, 0, 47071.8000001631, 47071.800000403, 8293.72500597675, 3.40202796212071e-08, 8293.72500043881, 3.60321202843821e-07, 16587.4500012468, -219410.775011713, -4.3585278613223e-05, 0, -363360.937491171, -363360.937502574, -15992.4375497605, 2.13688347815118e-07, -15992.4375024704, -3.51551538834544e-06, -159942.60002457, -159942.600004194, 4.61399443787329e-07, -319885.200009124, -47977.3125138512, -47977.3125108395, 568879.706281385, 0.000126082668057613, 0, 1387829.19370232, 1387829.19375776, 17085.9376980297, -7.6641501982545e-07, 17085.9375086005, 1.49682265633968e-05, 1086105.2062595, 1086105.20627574, 284241.656458619, -1.34928775164399e-05, 284241.656261291, 9.76879054304874e-06, 2172210.41252959, 68343.7500778896, 68343.7500504766, 852724.968835612, 852724.968816005, 102515.625073164, -887949.337550561, -0.000218019480191228, 0, -2913699.0936407, -2913699.09376251, -9595.46293677156, -1.557802834208e-06, -9595.46251932909, -3.38236911069764e-05, -3422996.71858946, -3422996.71882063, -278637.469411365, 4.29069145717779e-05, -278637.46876102, -4.89573924822469e-05, -6845993.43754988, -47977.3127102889, -47977.3126373531, -1666288.96911667, -1666288.96881532, 2.28743879229792e-05, -4998866.90644573, -4998866.90641823, -1114549.87536215, -95954.6253068984, -1114549.87520826, -95954.6251764341, -1671824.81291661, 823733.550048235, 0.000221925848452631, 0, 3428901.61862166, 3428901.61876101, 2214.3380481881, 6.83724344346109e-06, 2214.33752648031, 3.94901618110713e-05, 5480485.31210911, 5480485.31259863, 142824.769778431, -6.56924239408868e-05, 142824.768749158, 0.000101047421787191, 10960970.6250543, 13286.0253028147, 13286.0252166957, 4107596.06265477, 4107596.06264092, 1372889.25104254, -7.98313911269991e-05, 1372889.25007228, 1.3297020831764e-05, 12322788.1877049, 12322788.1877232, 714123.844479968, 33215.0631322771, 714123.844120157, 33215.062723358, 5491557.00058761, 5491557.00038171, 44286.7504402716, 8237335.50081452, 1428247.68875235, 1428247.68820973, -418509.787525101, -0.000122567752356011, 0, -2122442.49367419, -2122442.49375452, -0.000366171918912955, -6.04731641322208e-06, -1.94213721783759e-05, -2.49886657982164e-05, -4334565.65589491, -4334565.65632157, -29893.5570458667, 5.0724612634708e-05, -29893.5562438156, -8.94582427926262e-05, -8669131.31253139, -0.000221595033588113, -0.000178970160390921, -4484033.43731793, -4484033.43762527, -567977.569863214, 9.77681259608293e-05, -567977.5687761, -5.5033522619294e-05, -13452100.3126101, -13452100.31265, -179361.338178382, -0.000579357602180005, -179361.33784721, -0.000204079860029647, -2391484.50065647, -2391484.50012168, 4.47649530093514e-05, -9565938.00038307, -9565938.00035922, -2839887.84456827, -0.000686244312794595, -2839887.84412132, -0.000290486812009326, -14348907.0006439, -448403.345437747, -448403.344334339, -5679775.68911165, -5679775.68848496, -597871.126338453, 89680.6687554865, 2.82594572657629e-05, 0, 538084.012482155, 538084.012500501, 0.000100796315894979, 2.41448465598179e-06, 5.82527545060355e-06, 5.9775483799427e-06, 1345210.03113427, 1345210.03127142, 0.000246485074487645, -1.45837279867963e-05, -1.91443213396604e-06, 2.97055785559189e-05, 2690420.06250772, 6.44656957979524e-05, 5.81046018144109e-05, 1793613.37485788, 1793613.37504207, 89680.6691707759, -4.25994149921544e-05, 89680.6687480063, 3.03972528673652e-05, 5380840.1250291, 5380840.1250377, 0.000235524576972394, 0.000194357904931449, 0.000135664273734887, 8.28804643876502e-05, 1345210.03135382, 1345210.03130541, 538084.012903028, -3.86224275482686e-05, 538084.012536759, -4.17113322460568e-06, 5380840.12507876, 5380840.12511975, 538084.012887187, 0.000328663703901314, 538084.01265712, 4.3904465231337e-05, 8071260.18768581, 0.000768505146564351, 0.00018670879169864, 2690420.06277728, 2690420.06265957, 0.000203978511936865, 5380840.12557694, 5380840.12546318, 1345210.03228957, 0.000930564869109592, 1345210.03165505, 0.000480946556591932, 1793613.37592311}, 
                                                    {0, -283.500000017956, -3.53548442332555e-08, 0, 6580.23750040656, 8.70841439213854e-07, 0, 1902.48750028303, 0, 1902.48750002042, -59404.3875035996, -8.44746911296894e-06, 0, -41490.7312501557, -41490.7312502085, -5049.84375726765, 1.68740130334115e-07, -5049.84375033338, -6.33622611759482e-08, -10099.6875005298, 278498.503141678, 4.26589919469005e-05, 0, 340634.362488634, 340634.362500849, 6629.34381058803, -1.38731769936033e-06, 6629.34375189213, 1.16044500343065e-06, 103210.453153444, 103210.453128686, -2.27607921639013e-06, 206420.90625491, 19888.0312551455, 19888.0312534976, -753879.403169872, -0.000124313387280278, 0, -1394602.05931608, -1394602.05937542, -4305.65649134302, 5.33549524596031e-06, -4305.65625601352, -5.84922139196727e-06, -763023.796869244, -763023.796894962, -126606.797125247, 1.76300602646928e-05, -126606.796888982, 1.55825189166635e-06, -1526047.59376875, -17222.6250306796, -17222.6250147498, -379820.390666987, -379820.390656336, -25833.9375188174, 1222683.35632258, 0.000216378824929295, 0, 3127751.71861861, 3127751.71874635, 1107.16928024426, -8.52335531065903e-06, 1107.16876234933, 1.45957109719352e-05, 2661818.2028864, 2661818.20316899, 76579.1726944183, -4.63872727668878e-05, 76579.1718960988, 1.14217246024129e-05, 5323636.40628288, 5535.84382820869, 5535.84378905287, 832221.8441163, 832221.843823747, -4.34330356697981e-05, 2496665.53136225, 2496665.53133855, 306316.687658147, 11071.6875958569, 306316.687581557, 11071.6875255872, 459475.031454725, -1172491.70631951, -0.000221457064671614, 0, -3902769.84359891, -3902769.84374173, -0.000660058162685199, 6.22286200732657e-06, -1.65955223417467e-05, -1.67109641242131e-05, -4691627.5776068, -4691627.57817315, -18268.2856828923, 6.50075516030409e-05, -18268.2843836883, -4.14885074881715e-05, -9383255.15628906, -0.00011270541586141, -6.66157932968287e-05, -2383180.73437604, -2383180.73450561, -440099.579280742, 9.25768550666089e-05, -440099.578242549, 3.76879862369418e-05, -7149542.20325855, -7149542.20324684, -91341.4221773391, -0.000208326468386858, -91341.4220032165, 3.48481705362021e-05, -1760398.3127687, -1760398.31266545, -5.44388142569514e-05, -2640597.4692386, -182682.844317988, -182682.844020957, 612817.903161311, 0.00012291372789507, 0, 2540952.2811622, 2540952.28124259, 0.000436314813307565, -3.67494942347932e-06, 1.25272719179173e-05, 1.16692708360707e-05, 4035630.09330365, 4035630.09377885, 0.00103022016170231, -4.87956827999723e-05, -5.59906893824469e-06, 3.86710478932313e-05, 8071260.1875224, 8.515878440626e-05, 6.18051071095656e-05, 2989355.62462155, 2989355.62509015, 89680.6700590476, -9.68258490822084e-05, 89680.6688240046, 5.49395025028343e-06, 8968066.87509012, 8968066.87508296, 0.000271626400616934, 0.000185285596144958, 0.000109754089265098, -3.10270469941092e-05, 971540.578694402, 971540.578283824, -7.3466336372415e-05, 3886162.31269799, 3886162.3126779, 448403.344098815, 0.000198745328465449, 448403.343862951, -1.0318422185839e-05, 5829243.46917031, 0.000742912246865379, 0.000161660759234434, 896806.688408666, 896806.688019936, 0.000523062100166579, -134521.003132966, -2.84467871584143e-05, 0, -672605.015604586, -672605.015622482, -0.00011874755501477, 2.9919787540449e-07, -3.96007960686992e-06, -2.73068820259176e-06, -1345210.03110921, -1345210.031258, -0.000321873054513733, 1.4320149095096e-05, 3.76522372043285e-06, -1.30602750105437e-05, -2690420.06250533, -2.47935468763329e-05, -2.24549762983227e-05, -1345210.03103366, -1345210.03127245, -0.000517212518056961, 3.87263539382325e-05, -1.30133632829347e-05, -9.6092981113767e-06, -4035630.09377933, -4035630.09376864, -9.21974103464354e-05, -5.62600858629464e-05, -4.36996359481819e-05, 1.14100718066361e-05, -672605.015641585, -672605.015684595, -134521.003533485, 4.68003290198735e-05, -134521.003189191, 2.04631649728545e-05, -2690420.06254765, -2690420.06255334, -0.000158171037613935, -0.000101085339243981, -4.31276191776783e-05, 6.68126443775952e-05, -4035630.09388206, -0.000324163094264153, -5.46871451863307e-06, -672605.015741335, -672605.015678252, -1.37588870158017e-05, -1345210.03159387, -1345210.03155154, -0.000526883410373272, -0.000388296360387059, -0.000185987487421855, -0.000192601848850403, -0.000491829992415429}, 
                                                    {0, 226.800000016991, 2.22308612551326e-08, 0, -5366.25000038546, -5.52040149899766e-07, 0, -1113.75000020892, 0, -1113.75000001076, 49708.6875034227, 5.40558479881747e-06, 0, 24953.0625000038, 24953.0625001053, 2035.12500543611, -1.04972224324278e-07, 2035.12500018318, 0, 4070.25000020933, -240251.06251591, -2.75250119184849e-05, 0, -212799.656240484, -212799.656250145, -1640.25004578509, 7.51261326542298e-07, -1640.25000092173, -1.66851559603554e-07, -43079.3437700504, -43079.3437526267, 1.52640393356909e-06, -86158.6875024515, -4920.75000157671, -4920.75000077359, 671948.915667955, 8.07948417574246e-05, 0, 913332.206203021, 913332.206248078, 492.075183249835, -3.14208208106601e-06, 492.075002264524, 1.24546169975376e-06, 335615.653107345, 335615.653138278, 32722.9876870981, -9.17838628903957e-06, 32722.9875114316, -3.25427158266412e-06, 671231.306260944, 1968.30001015435, 1968.30000207661, 98168.9625185943, 98168.9625138969, 2952.45000224493, -1125621.56256971, -0.000141540274753899, 0, -2158979.06239791, -2158979.06249213, -0.00040225842526909, 5.28220332712032e-06, -3.52283512119619e-06, -4.07542425889734e-06, -1254791.24978421, -1254791.25002456, -9226.40688209858, 1.87125738676861e-05, -9226.40626867663, 3.94862231616048e-06, -2509582.50001927, -2.10263201428187e-05, -3.51029017109305e-06, -230660.156481719, -230660.15630909, 2.6096114877248e-05, -691980.468809772, -691980.468792605, -36905.6250557033, -2.52275507700132e-05, -36905.6250257277, 7.74924098669159e-06, -55358.4375968051, 1112704.593817, 0.000145598283962035, 0, 2839887.84363485, 2839887.84373747, 0.000498430751244735, -4.18661521298597e-06, 4.49602861075202e-06, 4.24303427001912e-06, 2399788.26519091, 2399788.26564448, 0.0010292279389972, -2.37893131964746e-05, 8.18017556614415e-06, 7.68795661908548e-06, 4799576.53127431, 2.88220390980388e-05, 8.62977422062106e-06, 730731.374888517, 730731.37509725, 58126.3602024898, -3.99792004833221e-05, 58126.3594784865, -3.60276163611538e-05, 2192194.12507922, 2192194.1250629, 9.28453456277175e-05, 5.7652048379623e-05, 3.0641999952224e-05, -7.01336360789412e-05, 232505.437603737, 232505.437560256, -4.27686087618105e-05, 348758.156527068, 0.000233729248618867, 9.14277562153327e-05, -597871.125035109, -8.11983360673847e-05, 0, -1943081.15618414, -1943081.15624043, -0.000327420285216648, 3.43984567422719e-06, -3.81887852098051e-06, -3.96361098175918e-06, -2242016.71839053, -2242016.71875671, -0.000820545390132191, 1.79149169098517e-05, 5.7647449027162e-06, -8.2020675280789e-06, -4484033.43751345, -2.20790216630814e-05, -1.27388871378371e-05, -1046274.46836263, -1046274.4688093, -0.000986957957539913, 3.61309786947727e-05, -7.10269265744747e-05, 1.16817084650788e-05, -3138823.40631162, -3138823.40629582, -7.44314532841885e-05, -4.81391413475358e-05, -1.7241178600804e-05, 5.9091929057887e-05, -149467.781567633, -149467.781382482, 4.09421898011112e-05, -597871.125101029, -597871.125077395, -0.000111371136072015, -3.34591405490488e-05, -8.98133811778463e-06, 5.42386461389947e-05, -896806.687749772, -0.000287363730610186, -2.73677533983586e-05, -0.000474068655042095, -0.000253416676854423, -0.000175783299845547, 134521.003132726, 1.8857011366987e-05, 0, 538084.012484874, 538084.012497084, 8.84890696312707e-05, -5.2251161308575e-07, 1.40990593269492e-06, 9.07618741161588e-07, 807126.018639449, 807126.018750999, 0.000257801215973604, -5.62796696286874e-06, -4.1762382870354e-06, 2.89151862967154e-06, 1614252.03750308, 5.9935890254787e-06, 6.17691166493922e-06, 538084.012305527, 538084.012511541, 0.000402875131746749, -1.47209019981812e-05, 1.45637879406427e-05, -1.29210492411097e-06, 1614252.03752159, 1614252.03751002, 2.29289876164682e-05, 1.00314242600497e-05, 5.95290432691435e-06, -2.65623303802622e-05, 134521.003079097, 134521.003172592, 0.000275819643343227, -2.41207078220228e-05, 5.66104178954356e-05, -1.65213676618178e-05, 538084.012530747, 538084.012519598, 4.64012455547575e-05, 2.4747873105107e-05, 3.06505939794134e-06, -6.226677008194e-05, 807126.018834159, 0.000113994541800625, -3.02733369314592e-05, 3.92067752707739e-05, 9.17419291173338e-06, -3.0518100282198e-05, 0.000185756580040931, 0.000174188167226151, 0.000239404926963871, 0.000140153200230584, 7.30352994173102e-05, 8.88331837007911e-05, 0.000235425659603284}, 
                                                    {0, -126.000000010772, 0, 0, 3019.05000024478, 2.22590665454516e-07, 0, 429.750000095528, 0, 429.750000004925, -28453.612502179, -2.20425288477405e-06, 0, -9803.36249995332, -9803.36250005957, -486.000002521106, 2.68726259941912e-08, -486.000000067331, 0, -972.00000008466, 140501.081260158, 1.13258655717468e-05, 0, 85789.6312451802, 85789.6312501065, 182.250021486042, -1.77637388730656e-07, 182.250000234427, 3.41788096808842e-08, 10534.0500087294, 10534.0500012959, -4.56893829697821e-07, 21068.1000010753, 546.750000478614, 546.750000228621, -402804.393777506, -3.34988844251869e-05, 0, -380770.368727098, -380770.368749072, -8.65233109757106e-05, 8.70663778194625e-07, -1.30952209862616e-07, -2.44541699944762e-07, -85046.9624862908, -85046.9625066124, -3745.23758661855, 2.5975797502934e-06, -3745.23750554944, 9.25423516195859e-07, -170093.925005189, -3.20773855496818e-06, -5.30972440436198e-07, -11235.7125070021, -11235.712505804, -2.67640619528452e-07, 693087.637544771, 5.90899989926215e-05, 0, 937402.874951241, 937402.874995989, 0.000190015487892612, -1.4170536552945e-06, -6.62304030652293e-07, 9.91134945244638e-07, 333995.906134908, 333995.906261998, 0.000300913959566935, -3.22891529874905e-06, 8.44422351542291e-06, -1.96505466197881e-06, 667991.812509162, 4.60900959262065e-06, -1.12261917187672e-07, 27679.2188408988, 27679.2187796748, -6.97075892872708e-06, 83037.6562753311, 83037.6562680098, 1.63986027476504e-05, 8.15770554605196e-06, 9.88644969648406e-06, -8.88940856036823e-07, 3.81787135719835e-05, -704159.325043148, -6.10887131482785e-05, 0, -1290958.76244589, -1290958.7624935, -0.000234823570383575, 8.32706991225164e-07, 9.9492614247251e-07, -6.59994626150381e-07, -680908.781029623, -680908.781258949, -0.000498072680339266, 3.26706968169341e-06, -2.1987938886889e-06, -1.75428343426332e-06, -1361817.56251202, -5.48874480158615e-06, -2.90542062287335e-07, -94109.3436546703, -94109.3437993009, -0.000371421004504256, 6.07424113316256e-06, -5.11694682619729e-05, 1.27488417321012e-05, -282328.0312843, -282328.031279325, -2.40103870931005e-05, -1.51098787035896e-05, -1.06905609963044e-05, 2.63100145149772e-05, -3.56009887034919e-05, -2.39072632631638e-05, 2.3681177826092e-05, -0.000118494576363844, -8.14744294962184e-05, -3.01047295749839e-05, 388616.231272671, 3.42442775250906e-05, 0, 926700.243719491, 926700.243744998, 0.000153675446365018, -1.17744190422916e-06, -3.25218366000621e-07, 1.10207996651172e-06, 687551.793572837, 687551.79375223, 0.000401046529897775, -2.52728227065198e-06, -4.79486058444729e-06, 1.34332859838004e-06, 1375103.58750668, 4.09203459572684e-06, 1.94771779779002e-06, 149467.781029709, 149467.781280511, 0.000463320806086741, -4.14847220971119e-06, 3.46273198684093e-05, -3.29247216716499e-06, 448403.343778352, 448403.343773574, 1.54181467579738e-05, 1.19136640030678e-05, 3.64516745597105e-06, -1.89409495186275e-05, 0.00010822105033468, 6.61341801451668e-05, -7.92946182496645e-06, 4.72143776695336e-05, 3.34139813362397e-05, 3.18203320368129e-05, 1.29594743443953e-06, -8.39702010441407e-07, -1.50766224036567e-05, 0.00010450833407457, 9.13064614019318e-05, 1.11986427826071e-05, 0.000186736976927109, 9.81704952821031e-05, 5.11018603582816e-05, -89680.6687550018, -7.97968184319469e-06, 0, -269042.006243071, -269042.006248473, -4.13529531833619e-05, 1.25141282870291e-07, -7.64011753915098e-08, -2.41738917635027e-07, -269042.006196551, -269042.006249967, -0.000126637110115882, 1.03556065749949e-06, 2.80940766600436e-06, -3.20240524421964e-07, -538084.012501598, -8.26005480470166e-07, -1.45704490087756e-06, -89680.6686470644, -89680.6687560088, -0.000194118011180982, 2.50829186417273e-06, -6.87820730617623e-06, 8.01264166578733e-07, -269042.006260133, -269042.006255859, -3.76575002316165e-06, -3.46419744500813e-07, 2.06010440529172e-07, 1.1290970531073e-05, 4.28107281263377e-05, -2.43053990997573e-05, -0.000118187768645599, 5.93278089396157e-06, -2.76957246920131e-05, 5.05991877141661e-06, -1.57291119170202e-05, -8.35851400642817e-06, -1.35578429844464e-05, -6.65937568161388e-06, -1.33671436550484e-06, 2.18967104977558e-05, -3.59430497289216e-05, -3.16989153466535e-05, 1.18676426745754e-05, -1.39074591281763e-05, -3.55634165228126e-06, 1.46189806386228e-05, -7.11176877640297e-05, -7.14725950582564e-05, -8.7507997021894e-05, -4.23973055040441e-05, -2.43219760535785e-05, -4.77699047437587e-05, -8.27423578140785e-05}, 
                                                    {0, 46.2857142901213, 0, 0, -1118.95714295744, -5.21994818708365e-08, 0, -98.3571428826204, 0, -98.3571428589207, 10677.2464294662, 5.23823033119455e-07, 0, 2272.91785712456, 2272.91785716863, 52.0714292457579, 0, 52.071428587472, 0, 104.142857170737, -53563.2750041813, -2.71514621701826e-06, 0, -20266.1999986621, -20266.2000000888, -5.80019502875348e-06, 2.67089403710065e-08, -1.88915590095609e-08, 0, -1148.17500220137, -1148.17500038793, 4.63865200129921e-08, -2296.35000035746, -1.26194688488379e-07, -1.02298967946405e-07, 156520.85626135, 8.0883572055568e-06, 0, 92223.0562437926, 92223.0562499168, 2.34950398610239e-05, -1.05145360416297e-07, -1.84263956246076e-07, 5.67309386767288e-08, 9513.4499952337, 9513.45000206196, 2.29580547807554e-05, -4.37458842278773e-07, 1.5048439355993e-06, -5.06282178930685e-08, 19026.9000017348, 7.87625557340595e-07, 4.04637371767773e-07, 1.86272583987971e-06, 1.84534980907396e-06, 2.48950161016593e-07, -275315.96251852, -1.43699697678853e-05, 0, -234350.718736957, -234350.718749179, -5.16072366784072e-05, 1.72568775698872e-07, 7.19879321366576e-07, -1.72798814399157e-07, -38750.9062166047, -38750.9062539972, -8.15646326966803e-05, 1.76781948310627e-07, -1.93496763754877e-06, 1.11662279826796e-07, -77501.8125032287, -7.61115901630341e-07, -5.33452670781451e-07, -2.0448077016495e-05, -8.4948868467707e-06, 6.66239543884184e-07, -7.27608871240699e-06, -5.90338817777179e-06, -3.37940407697703e-06, -2.36564029853517e-06, -4.05620516589851e-06, -1.60393006765408e-06, -9.99678531943972e-06, 286598.539303605, 1.49182692570035e-05, 0, 335472.131235702, 335472.13124847, 6.36640955609046e-05, -2.96942510309741e-08, -9.7991411096464e-07, 2.20547465138656e-08, 83037.6561878661, 83037.6562534468, 0.000136840122038713, 1.72803105958508e-07, -1.92918896976663e-07, 6.01433981477366e-07, 166075.312504251, 6.87729564746957e-07, 6.18641899744785e-07, -3.46800552436995e-05, 1.47087074096823e-05, 9.58088924079153e-05, 2.28331606611543e-07, 1.36990207070843e-05, -2.53196430225235e-06, 9.99597835705457e-06, 1.07409442772613e-05, 5.18300159708941e-06, 2.21222975765149e-06, 4.81416293223717e-06, -3.22165973291673e-06, 9.57650080477346e-06, 9.29355255194864e-06, -4.49315469990087e-06, 3.25524026508821e-05, 2.06576228382611e-05, 7.29852998015335e-06, -162279.305366564, -8.41067600158969e-06, 0, -251959.974099182, -251959.974105883, -4.15748741617435e-05, 2.27563119075905e-07, 5.59178368701663e-07, -2.00835475555552e-07, -89680.6687012413, -89680.6687512263, -0.000111188093889186, -2.40274063277047e-07, 2.03434012925254e-06, -2.64482918279321e-07, -179361.33750248, -4.23262495648882e-07, -4.02795254046658e-07, 6.69028910979272e-05, -9.71518076071477e-06, -0.000124071056187457, -6.14461969990507e-07, -8.76606563440628e-06, 4.18007970854656e-08, -8.83115185352499e-06, -9.16606092387661e-06, -2.31211027419513e-06, -2.16535912775589e-06, -2.83131193123512e-06, 1.45770553680412e-06, -2.0269538598297e-05, -1.83398045908694e-05, -4.03599468332075e-07, -1.63634019927841e-05, -1.20299995205676e-05, -7.56449527066016e-06, 2.76717224234795e-06, -3.07544757881214e-06, 1.55411662366451e-06, -2.8287886272454e-05, -1.78634962854567e-05, -5.34332929989834e-06, -4.42121867230898e-05, -2.67761766503837e-05, -9.63215898095058e-06, 38434.5723235114, 1.9661910881468e-06, 0, 76869.1446410653, 76869.1446424558, 1.11616788043568e-05, 0, -1.09253249137722e-07, 4.27010604813851e-08, 38434.5723069238, 38434.5723215507, 3.53038443148266e-05, -1.28586116488251e-08, -1.00652936262208e-06, 0, 76869.1446434995, -2.02248350272911e-08, 2.88942477022529e-07, -2.98082012689297e-05, 2.14650062007958e-06, 5.31359010905851e-05, -5.4484524384732e-08, 1.54053144667185e-06, -6.74264062401175e-08, 3.18921997707331e-06, 2.60863265771558e-06, 4.09507072339184e-07, -5.05550078429716e-07, 2.10777013898719e-07, -2.04351175884776e-06, -1.59609266592442e-05, 7.04901384114297e-06, 2.91438130515826e-05, -4.40740098789489e-07, 7.26928608518668e-06, -6.51758791344166e-07, 5.38639569821461e-06, 3.74010760381964e-06, 4.2028347504374e-06, 1.76618761570503e-06, 2.22793830078503e-06, -3.45147254593987e-06, 9.91722779191753e-06, 4.79911543431644e-06, -1.0327820916991e-06, 4.28313737555037e-06, 3.07907138236587e-06, -2.51957647989048e-06, 1.68338275782328e-05, 1.7879208044748e-05, 2.11847454178324e-05, 8.30301977886799e-06, 6.19686396963298e-06, 1.70480286787534e-05, 1.70191421301564e-05}, 
                                                    {0, -10.125000001055, 0, 0, 246.3991071669, 0, 0, 10.1250000031467, 0, 10.1250000004307, -2373.15535735774, -5.60634563916785e-08, 0, -236.274107142217, -236.27410714919, -8.08253781289746e-08, 0, 0, 0, 0, 12050.1421885062, 2.92040594212819e-07, 0, 2136.88124984958, 2136.88125002993, 6.96593105341708e-07, 0, 0, 0, 2.53405516296467e-07, 6.15260250603214e-08, 1.52895250297207e-08, 7.71569820895415e-08, 2.90250177973356e-08, 3.50151246423773e-08, -35747.1984402371, -8.73782661500663e-07, 0, -9913.26093680092, -9913.2609375446, -2.8329525853305e-06, -1.84046720685355e-08, 3.50644436086618e-08, 0, 6.21775148065576e-07, -3.33557045019017e-07, -2.68335054881519e-06, 2.77254907424575e-08, -2.0832319710635e-07, 0, -3.45830526842682e-07, -1.31097284743007e-07, -1.40893897327737e-07, -3.10919903991962e-07, -3.56070332046211e-07, -1.54233477698474e-07, 64031.2593794752, 1.56443698639427e-06, 0, 25833.9374985267, 25833.9374999927, 6.20032127151618e-06, 1.79430671827865e-08, -1.29249034680146e-07, 0, -4.00702664984121e-06, 7.0034591679775e-07, 9.65593842783133e-06, -4.63079445637422e-08, 2.29548491493577e-07, 9.88483725092109e-08, 7.39164251890789e-07, 1.55259851982709e-07, 2.70789341404271e-07, 2.0662216046666e-06, 1.17779814463275e-06, 1.01728949895645e-07, 1.26854742259157e-06, 1.22970819424976e-06, 4.57176407573889e-07, 5.97588379904379e-07, 1.10249338668459e-06, 5.50378153175085e-07, 1.46159073782839e-06, -68090.8781293315, -1.62406367181848e-06, 0, -38197.3218733913, -38197.3218749056, -7.62512299344381e-06, 0, 1.79172447888304e-07, 1.84810211928788e-08, 7.454369912574e-06, -7.14935790194651e-07, -1.63763850428661e-05, -3.11563645323722e-08, 7.21477177350314e-08, -2.61293205970541e-08, -9.07212960677129e-07, -1.73938462481713e-07, -3.39409485334626e-07, 4.89747609932721e-06, -2.13430120686969e-06, -1.09254105078099e-05, 1.19500228229671e-07, -1.70506022836136e-06, 3.59109955761428e-07, -1.92302045869714e-06, -2.54185269144865e-06, -1.05090371821608e-06, -2.01892568962003e-08, -1.09696853912554e-06, -5.81551240469246e-07, -1.82606576061456e-06, -2.35131585889074e-06, -5.51731863583786e-07, -5.32843187684025e-06, -4.09162499429413e-06, -1.0883227888468e-06, 39502.1993326423, 9.2287140929786e-07, 0, 29893.5562491149, 29893.5562498939, 4.97727683661855e-06, -1.11002201869556e-08, -1.07899082679299e-07, 0, -5.70371320240494e-06, 3.44224691248425e-07, 1.34305179145127e-05, 1.44292250016607e-07, -3.01839044743998e-07, 2.99696416133324e-08, 5.64839567187738e-07, 8.12466515573269e-08, 6.44001678477817e-08, -8.49292067450724e-06, 1.51954546438601e-06, 1.4528988555447e-05, -3.84747531242386e-08, 1.0282021778714e-06, -2.41639989489499e-08, 1.83748570421546e-06, 2.1314042442417e-06, 3.45559973416644e-07, 4.42031586875713e-07, 1.00835289210007e-06, 3.54232775846183e-07, 1.57581172535614e-06, 2.35899916739239e-06, 1.96069047539626e-07, 3.51604508007946e-06, 2.45759317920841e-06, 1.1870433863544e-06, -1.499270315063e-06, 1.77160415648529e-06, -8.42044860428742e-07, 5.17766388406681e-06, 1.71338443881974e-06, 1.16983919393251e-06, 4.86072411969336e-06, 5.57132696762648e-06, 1.71983983542842e-06, -9608.64308086322, -2.16287310470808e-07, 0, -9608.64308015858, -9608.64308031769, -1.33694362399843e-06, 0, 2.48452316590575e-08, 0, 1.68442111691156e-06, -6.36496280455268e-08, -4.29859258412112e-06, -4.69751433289038e-08, 1.40516959152719e-07, 0, -1.56958251869219e-07, 0, -2.24345959531582e-08, 3.63093552832959e-06, -3.76256082074148e-07, -6.34132416868366e-06, -2.16045516503641e-08, -1.55415760539428e-07, 1.41992662684502e-08, -6.63361259942153e-07, -6.60465086855513e-07, -9.68680283001259e-08, 5.33002658655561e-08, -1.93093506238341e-07, 0, 2.29489009965122e-06, -9.48983737032786e-07, -3.13797073149723e-06, -1.06210817382339e-07, -8.55975511442972e-07, 3.8313065714441e-08, -1.13136511598246e-06, -1.07142034634092e-06, -9.5187295523814e-07, -4.44837221194317e-07, -7.90437766669967e-07, 4.44107732487369e-08, -1.83846699038694e-06, -1.06900842457104e-07, -4.39677644756788e-07, -8.34667447023646e-07, -1.12881308925462e-06, -1.53544102912031e-07, -2.61778443437913e-06, -2.4728053132868e-06, -3.00028039977673e-06, -1.00807123476153e-06, -1.25783128882223e-06, -2.88328356956086e-06, -2.02789204157489e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, -10.1250000026763, 0, 0, 0, 5.60634563911381e-08, 0, 236.274107136119, 0, 7.73689079150201e-08, 0, 0, 0, 0, 0, -2.92040594210179e-07, 0, -2136.88124981757, 0, -6.83150065106895e-07, 0, 0, 0, -2.15224647732617e-07, 0, -1.52895250288464e-08, 0, 1.24310364230055e-08, 0, 0, 8.73782661493208e-07, 0, 9913.26093671233, 0, 2.80187445570726e-06, 1.84046720658414e-08, 0, 0, -7.89290107622626e-07, 0, 2.56186385813636e-06, -2.77254907437322e-08, 0, 0, 0, 0, 1.0400292001929e-08, -7.94648017133146e-08, 0, -3.99579059131307e-08, 0, -1.56443698638165e-06, 0, -25833.9374983864, 0, -6.15605048355184e-06, -1.79430671784849e-08, 0, 0, 4.38153925728144e-06, 0, -9.43650381947239e-06, 4.63079445726778e-08, 0, -9.88483725062814e-08, 1.74455906135334e-08, 6.93020238344011e-08, 0, -1.64455229401216e-06, 0, -1.01728949902117e-07, 4.0532556873421e-08, 4.22341160317646e-08, 4.60881798493701e-07, -1.24374674655671e-07, -8.49420471351294e-08, -7.1935129245083e-08, -6.8895201018613e-08, 0, 1.62406367180586e-06, 0, 38197.3218732638, 0, 7.58693072166438e-06, 0, 0, -1.84810212072295e-08, -7.9062299345532e-06, 0, 1.6147869796237e-05, 3.11563645176496e-08, 0, 2.61293205958563e-08, -2.62862557780066e-08, -5.83145619001552e-08, 8.66646174051924e-08, -5.60043579367232e-06, 0, 1.03640386460712e-05, -1.19500228231506e-07, 0, -3.59109955749385e-07, -2.63539059220546e-07, 2.4120648690196e-07, -1.30125889394395e-07, -4.95116196020469e-07, -1.9312720675539e-07, -1.10571381168748e-07, -4.29143218672893e-07, -2.66235830550684e-08, -3.07937188053751e-07, 1.95066201831549e-06, 1.63626485026008e-06, -1.21116201179548e-06, 0, -9.22871409291001e-07, 0, -29893.556249053, 0, -4.95896334447213e-06, 1.11002201913288e-08, 0, 0, 5.98385209694414e-06, 0, -1.33024420920757e-05, -1.44292250007749e-07, 0, -2.99696416106377e-08, 0, 5.082514829426e-08, 6.71191789264775e-08, 9.05740119345164e-06, 0, -1.41470854975755e-05, 3.84747531469164e-08, 0, 2.41639989175257e-08, -1.36195840661907e-07, -2.86385803217528e-07, 5.02243772080439e-07, -2.63290351740315e-08, -1.38650547093809e-07, 9.2711671395809e-08, -9.59741757626846e-07, 0, -1.96069047565189e-07, -1.18741301179379e-06, -7.15590330537747e-08, 6.38234659052692e-07, 2.06470957632474e-06, 3.10097980410483e-07, 1.50437522316859e-06, -1.28031578761493e-06, 1.50876250023208e-07, 7.46933718359473e-07, -5.75273920408909e-07, -1.54000772097306e-06, 1.00694260995576e-06, 0, 2.16287310469246e-07, 0, 9608.64308014618, 0, 1.33319702523917e-06, 0, 0, 0, -1.75440311956882e-06, 0, 4.26864871012429e-06, 4.69751433261131e-08, 0, 0, 1.31410270087378e-08, -3.34373369961618e-08, 0, -3.80620686084196e-06, 0, 6.23657285867488e-06, 2.16045516425114e-08, 0, -1.41992662612791e-08, 1.18536558541461e-07, 1.17957803736987e-07, -1.30716311187707e-07, -1.5866031604859e-07, -1.99109783824048e-08, -1.12299317425501e-07, -2.54001556794341e-06, 0, 2.93174572971285e-06, 1.06210817381311e-07, 0, -3.8313065695375e-08, 1.22594972503118e-07, 6.13180239229971e-08, 3.31294277253054e-07, 1.670117546673e-07, 8.1849087174593e-08, -3.04562918057091e-07, 2.48318772746885e-07, -5.17670849012887e-07, -1.68841734425907e-07, -2.30591344832113e-07, 5.04415661900465e-08, -9.8823081403799e-08, 5.47249479247897e-07, 2.62415204080181e-07, 1.36040348334769e-06, -9.28811873503546e-08, -2.040556403932e-07, 1.63269352444168e-06, -1.76008386309589e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, -5.78571428640516, 0, 0, 0, 4.24083679480704e-08, 0, 127.575000014394, 0, 52.0714285729716, 0, 0, 0, 0, 0, -2.04803672536327e-07, 0, -1057.05000008795, 0, 6.4483790719821e-08, -1.74819384810216e-08, 0, 0, -1148.17500022134, 0, 5.1578738715517e-08, -1.12836822384655e-08, -6.50144706927383e-08, -1.58066287001237e-08, 0, 5.79324612739209e-07, 0, 4305.65625015373, 0, -4.6225525570754e-07, -6.39530578913738e-08, 0, -1.7563432908733e-08, 9513.45000133485, 0, 6.57002917648099e-07, 1.05830522102094e-07, 0, -3.01223631075715e-08, -1.38163654395259e-08, 1.89726797202285e-07, -1.18581708217507e-07, 5.57061303801372e-07, 5.09151221679663e-08, 6.38728193981185e-08, 0, -9.50079231120013e-07, 0, -9226.40625013421, 0, 1.02117560307827e-06, 7.1914807507928e-08, 0, -1.99380689201394e-08, -38750.906252576, 0, -3.20505704028948e-07, -1.51388573916907e-07, 0, 3.6118014502726e-07, -1.90192421400173e-08, -7.91669761704637e-07, 2.46964627102698e-07, -5.02134584437335e-06, 0, 3.73523612528062e-07, -8.82997592112514e-07, -1.62283981338928e-07, -2.73675649214043e-06, 4.86103431075331e-07, 3.02671629302564e-07, 1.10843132234356e-06, -4.83005565992727e-07, 0, 9.66140297234924e-07, 0, 9964.51874996196, 0, -1.18056235787264e-06, -2.29497370594817e-08, 0, 7.26460669412774e-08, 83037.6562532685, 0, -1.44904661105095e-06, -1.10623505457778e-07, 0, -1.09251869760649e-07, -6.79838974254049e-08, 1.13993202909535e-06, -5.11019336145179e-07, 9.98317527033196e-06, 0, 5.7949287697119e-06, 3.45858226106597e-07, 0, 1.33561947052301e-06, 2.0368579817724e-06, -1.20602591607785e-06, 2.63349850737562e-06, 2.08265493866222e-06, 3.91166161480297e-07, -1.57748131939551e-06, 3.63464181614301e-06, 1.95712792420448e-07, -9.16080156157307e-07, -4.6223245580638e-06, -2.61880395467174e-06, 4.33481118915847e-06, 0, -4.90907244165143e-07, 0, -4270.50803557599, 0, 7.65016870575962e-07, -6.4339853078202e-08, 0, 0, -89680.6687511616, 0, 2.45935582293653e-06, 5.13770612851405e-07, 0, 1.21403385750162e-07, 1.41178517127605e-07, -1.03332935723642e-06, -1.19540151024555e-07, -7.82726299683735e-06, 0, -2.51583680158155e-06, -1.10451102344676e-07, 0, -1.20174402136175e-07, -2.54710706653077e-07, 1.71652571459247e-06, -3.26839264459765e-06, -4.63694540193531e-07, 8.39180274559262e-07, 1.78021822153064e-06, -9.69019807820552e-06, 0, 6.75643126510954e-07, 2.29345469180829e-06, 5.35722668882721e-07, -5.2031252524563e-06, -6.04614827260592e-06, -1.04622091724042e-06, -1.90998673639944e-06, 1.12955520828221e-06, -4.12087869910946e-06, -1.89170065071943e-06, -3.82744132063825e-06, 1.85869228535018e-06, -4.7734994749612e-06, 0, 1.11915398647128e-07, 0, -4.40969179711905e-08, 0, -2.16739444284481e-07, 0, 0, 0, 38434.5723215803, 0, -1.08616917381564e-06, -1.58844843029139e-07, 0, 0, -8.539359659408e-08, 3.80088687323619e-07, 0, 1.98208448481049e-06, 0, -1.83329440821556e-07, -1.00042309718139e-07, 0, 6.86837620811725e-08, -2.91110976938001e-07, -7.81260911711985e-07, 9.31546425676485e-07, 7.75358653382493e-07, -2.19736530073438e-08, -2.18326105008357e-07, 4.51435961655597e-06, 0, 3.44038342548862e-06, -4.104035957983e-07, 0, 1.20549976474219e-07, 6.03217572795382e-07, -4.51287038697033e-07, -2.49613331377828e-07, -1.14123615560185e-06, -3.8969969333811e-07, -4.23853107689097e-07, -2.42547520373844e-08, 3.49386634234712e-06, -2.83647884901797e-07, 2.34308281588576e-06, -2.22589306796334e-07, -1.36018902135796e-06, 1.07122444940975e-06, 8.80645386713422e-07, -1.75866387186039e-06, 1.47425222239955e-06, 1.37176549750954e-06, -7.24626257908437e-06, 3.09464316494283e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, -4.50000000022935, 0, 0, 0, 5.8299325641717e-08, 0, 92.4750000086148, 0, 60.7500000112094, 0, 0, 0, 0, 0, -2.70999917957459e-07, 0, -683.437500061457, 0, -182.250000085976, 2.98247431565496e-08, 0, 1.42014632607328e-08, -1248.4124999979, 0, -1.00144168763658e-07, 0, 3.57924377792344e-08, 3.63939945434191e-08, 0, 7.10154825471626e-07, 0, 2323.68750026077, 0, 4.00390904323269e-07, 1.39933851644295e-07, 0, 3.50260973830604e-08, 9226.40625054232, 0, 3745.23749959746, -2.20346669081172e-07, 0, 8.20076140009014e-08, -1.71037961237215e-08, 2.10835026543031e-07, 7.29061528307497e-08, -4.36538345778611e-07, -3.77773184283099e-08, 6.93197302185313e-08, 0, -1.17446043724898e-06, 0, -3690.56250027011, 0, -6.04158751373808e-07, -2.10788319270491e-07, 0, 4.71725734539449e-08, -31369.78125293, 0, 2.55700709162922e-06, 2.77343941379612e-07, 0, -7.63477041019685e-07, 5.08098874165716e-07, 0, 8.33754439841437e-08, -27679.2187502545, 0, -7.44420619506443e-07, -7.78757205967134e-08, 6.54075297736469e-07, 2.12427383420527e-06, -2.58315850857097e-06, -3.4638126730976e-07, -1.33733840545658e-06, -1.87766651844687e-06, 0, 1.07187297485531e-06, 0, 2214.33750013347, 0, 6.07335884782051e-07, 1.36470067824193e-07, 0, -1.58264241752265e-07, 49822.5937525812, 0, -5.09010197342709e-06, 2.40451042011217e-07, 0, 2.55962574651313e-07, -8.14078002140082e-07, 2.26172007143421e-07, 1.35267011631476e-07, 94109.3437535391, 0, -8.04093864541168e-07, -5.61565065620229e-07, 0, -2.75853164922864e-06, -2.2558869252762e-06, 7.55817766026911e-07, 9.92082521331615e-07, -7.09151110588911e-07, -1.83057160195511e-06, 2.48701104755241e-07, -9.33287439653205e-07, -8.6689642439771e-07, 2.11396525256131e-06, 1.66717979590547e-05, 1.50036118193634e-05, -6.56401368621896e-06, 0, -6.25762553510926e-07, 0, -6.72110325021487e-08, 0, -5.75018998264039e-07, 9.8902532739426e-08, 0, -2.12674301117302e-08, -29893.5562522969, 0, 3.78577070164349e-06, -1.03516793964476e-06, 0, -2.60893546409716e-07, 5.52631328299526e-07, 1.61027689803265e-07, 7.82963407820686e-07, -149467.781254867, 0, 3.57034706917834e-06, 1.98408846507019e-07, 0, 2.64356569731826e-07, -1.83511476231841e-08, -6.76038062727945e-07, 4.95431322718639e-07, -2.58348928131394e-06, -6.24140659168983e-07, -5.35036142155343e-07, -3.39966149362258e-06, 0, -1.13161593663912e-06, -9.56542507283119e-06, -2.87961197696934e-07, 2.57893285109268e-06, 1.0911681214735e-05, 3.16234170048157e-06, 6.75196237510231e-06, -6.65008314245273e-06, -4.41874043440779e-06, 4.62814969613578e-06, -8.069449566071e-06, -8.79757155600689e-06, 2.29046789955266e-06, 0, 1.34164953948268e-07, 0, 0, 0, 2.65814360504893e-07, 3.1668076079497e-08, 0, 1.8850147504752e-08, 5.23591796838328e-07, 0, -9.54894607244641e-07, 3.02943042312621e-07, 0, 0, -9.07225027347413e-08, -1.42143886262995e-07, -1.26438758685376e-07, 89680.6687521661, 0, -1.81702150433209e-06, 2.40473916092805e-07, 0, -1.53765167078218e-07, 6.57120609641931e-07, 1.61274997774325e-07, 1.72017580977896e-07, -2.23477488014655e-07, -3.57640282350883e-07, -7.84259872473714e-07, 2.21347932794291e-06, 0, -5.37931524434562e-07, 8.90690071425247e-07, 0, -1.96895876175352e-07, -2.0021528954129e-07, -5.05728464621604e-08, 4.21545331054611e-06, 3.59003474584486e-06, 9.79214295595947e-08, -1.61594824067502e-06, 8.48370372646204e-07, -1.25412707656724e-06, -8.34276838921111e-07, -1.34266405913475e-06, 3.57403281605931e-08, 2.11556291307565e-06, 4.5889250343533e-06, 1.62907945710295e-06, 1.35012387028718e-05, 1.88969605700928e-06, -1.39005070901444e-06, 1.5749189413384e-05, 2.34360761470559e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, -4.04999999822942, 0, 0, 0, -2.19485134744422e-08, 0, 75.9374999948148, 0, 66.824999946271, 0, 0, 0, 0, 0, 9.50695791551411e-08, 0, -478.406250048363, 0, -328.049999545724, -3.96755157757296e-08, 0, -1.41340392791754e-08, -1252.96874995366, 0, 1.17892953364764e-07, 4.13928569379564e-08, -5.63983188634109e-08, -4.21768532477468e-08, 0, -2.09505050626991e-07, 0, 1230.18750012453, 0, 492.074998099186, -1.46554606906777e-07, 0, -4.13704062856416e-08, 7893.70312547321, 0, 6150.93749930926, 2.80633561271041e-07, 0, -1.34423516791895e-07, -2.996891148212e-07, -2.27319040847097e-07, -1.30026818875644e-07, 3.60104850087581e-07, -1.11338380415359e-07, 2.28075567940417e-07, 0, 3.25310226485842e-07, 0, -1107.16875042056, 0, 3.66742396229372e-06, 2.55711524650194e-07, 0, -6.24479740170665e-08, -20298.0937503952, 0, -9226.40624763361, -3.43878842079558e-07, 0, 1.03443120740138e-06, 1.75093868290694e-07, -3.04843825382321e-07, 6.56744428083642e-08, -38750.9062523153, 0, 8.31200719510182e-07, 1.84926009474632e-06, 1.13876661749382e-07, -2.47143368776329e-06, 1.61005984421318e-06, 8.25461567117924e-07, 7.13777382120923e-07, 2.11115424003246e-06, 0, -2.03001588745837e-07, 0, 5.03363603977224e-07, 0, -4.00239421513681e-06, -1.51409684132907e-07, 0, 2.05285118512346e-07, 18268.2843773854, 0, -4.63412285977893e-06, -3.27197068109424e-07, 0, -3.59150554556724e-07, -1.21139815056783e-07, -2.61025971421946e-07, -3.95719149064786e-07, 99645.187506165, 0, 58126.3593759066, 5.20655444322221e-07, 0, 3.43577170326697e-06, -1.36177922153959e-06, -2.9222950428183e-06, -3.2999620616182e-07, 3.87455669325398e-06, 1.4315875761788e-06, -3.05364962854677e-07, -5.27263153589291e-07, -3.62512141772423e-07, 1.2253050752175e-06, -2.50132419017462e-05, -1.36481557326365e-05, 9.90345294527193e-06, 0, 1.79499392508857e-07, 0, -2.10075240239458e-07, 0, 2.50134839688494e-06, -1.72065504515113e-07, 0, 2.78287163749585e-08, -7.72107034715284e-07, 0, 4.95345520985907e-06, 1.22619974712432e-06, 0, 3.22307515855407e-07, 0, 0, -7.92589196117494e-07, -89680.6687546227, 0, 9.22653188455039e-07, -3.18984671909153e-07, 0, -3.19616483925892e-07, 3.76981758028136e-06, 2.69557206495691e-06, -1.09832467462701e-06, 1.085550788771e-06, 1.65656029293572e-06, 2.20765673461118e-06, -149467.781255031, 0, 1.04183636963546e-06, 1.64115254563224e-05, 2.15313622469107e-06, -4.30889085378899e-06, -2.02561460480373e-05, -2.67223369696544e-06, -1.03426446491838e-05, 1.37827304628096e-05, -2.97731477092267e-06, -5.27042689056772e-06, 1.06719976024959e-05, 1.24573855407657e-05, -1.04623142293907e-05, 0, -2.93559875890089e-08, 0, 5.7385802927838e-08, 0, -7.3204582515589e-07, -2.49027775861308e-08, 0, -2.44511803597394e-08, 2.24232592187458e-07, 0, -1.96262191588182e-06, -3.22872620112827e-07, 0, -1.57472847544146e-08, -3.37488005740817e-08, 7.01271405852635e-08, 1.13679678177607e-07, 9.68863061385851e-07, 0, -1.12552663522669e-06, -3.28944813296952e-07, 0, 1.92114621736868e-07, -2.23322972362888e-06, -8.15163836322725e-07, -1.81063997747978e-07, 5.63121048283362e-07, 1.76970585777059e-07, 2.6612771348953e-07, 134521.003127914, 0, 2.87744513501842e-06, -1.17567462399749e-06, 0, 1.8935165696748e-07, -2.26040349448584e-06, -5.67953382334138e-07, -4.50073085799724e-06, -1.86201556400059e-06, -7.14985266167193e-07, 2.503649350363e-06, -3.86244494698494e-06, 4.72059143008418e-06, 8.76254929452131e-08, 4.69026921304102e-07, -4.02031934949462e-07, 2.17467228236773e-07, -8.69182348978523e-06, -5.0297470376273e-06, -1.26604941939281e-05, 5.63385766760493e-06, -3.81650772600308e-07, -1.72486533820829e-05, -2.67633726784095e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, -4.04999999397595, 0, 0, 0, 3.89353636962131e-08, 0, 66.8249999467817, 0, 75.9374999137773, 0, 0, 0, -1.28504042017994e-08, 0, -1.85701278968073e-07, 0, -328.049999715183, 0, -478.406249486582, 2.94869973818532e-08, 0, 0, -1252.96874928877, 0, -8.48438476931047e-08, 4.68749620108344e-08, 1.26549998527979e-07, 3.40850904670207e-08, 0, 4.85791404219466e-07, 0, 492.074999227876, 0, 1230.18749834527, 1.13399271547698e-07, 0, 2.88040691915247e-08, 6150.9374973724, 0, 7893.70312191204, -2.19725383291236e-07, 0, 1.44862534056563e-07, -1.84919572227313e-07, -7.27844337306846e-08, 1.57660061642607e-07, -9.11759955403937e-07, -7.88118089195606e-08, -1.96857454073094e-07, 0, -7.98325183153924e-07, 0, 1.47383125358462e-06, 0, -1107.16874632402, -2.65302677570335e-07, 0, 4.65095737895272e-08, -9226.40624494872, 0, -20298.093741851, 2.81991377554467e-07, 0, -9.24274736498091e-07, 1.02381883693315e-06, 1.43575154856705e-06, -1.31076177541072e-07, -38750.9062417334, 0, -4.81804454318187e-07, 1.70525472279146e-06, 1.03787271117396e-06, 2.99207491008237e-06, -1.72564980445162e-06, -7.47199501000272e-07, -6.40816137497257e-07, -8.93918717158123e-07, 0, 6.93930979776909e-07, 0, -1.52473354744422e-06, 0, -4.51742806763582e-06, 2.0252778030149e-07, 0, -1.61768787744602e-07, -8.24874529899798e-06, 0, 18268.2843629725, 3.25084587855179e-07, 0, 3.10639227021347e-07, -1.21668250263954e-06, -1.57667077885986e-06, 3.92470276292873e-07, 58126.3593623208, 0, 99645.1874891317, -2.48084366430622e-07, 0, -2.63044824509481e-06, -5.94767719237899e-06, -5.87564312976059e-08, -1.42413935876036e-06, -1.81375146653164e-06, -4.16727844116135e-07, 3.23309045209418e-07, -4.99455523371837e-06, -7.1066633881497e-07, 1.38618901454953e-06, 1.32350691958974e-05, 1.0941908878288e-05, -8.10076284545414e-06, 0, -4.43513362647953e-07, 0, 7.48059132707971e-07, 0, 2.78218794517922e-06, 1.04401000998738e-07, 0, -1.8188009106701e-08, 5.21460917725614e-06, 0, 8.93630738063478e-06, -8.98626939078346e-07, 0, -2.26881608064917e-07, 7.58241918689265e-07, 1.43459501040233e-06, 4.67028862629768e-07, 1.02093602768402e-05, 0, -89680.6687414795, 4.30173050727936e-07, 0, 1.99076727276247e-07, 4.16957968657057e-06, -4.55321761103892e-07, 2.61675033402773e-06, -1.21974637547511e-06, -2.185554166477e-06, -2.13885869354274e-06, -149467.781241993, 0, -4.66320810038721e-07, -4.53977815897086e-06, -1.16850379641107e-07, 6.50671730982012e-06, 1.35582432010104e-05, 2.5143725189891e-06, 5.2623099468582e-06, 1.31821062196503e-06, -6.65128647055242e-07, 1.59618154244784e-06, -2.67069702988011e-06, -2.04348904294252e-06, 2.1676424439332e-06, 0, 8.98252891170227e-08, 0, -1.67669078623462e-07, 0, -5.90302689792707e-07, 2.55053809430738e-08, 0, 1.9758976444461e-08, -1.45966896071393e-06, 0, -2.63338162296839e-06, 1.89936586534541e-07, 0, 2.82061731960383e-08, -1.47918889302781e-07, -5.20354948956603e-07, -1.63911620041204e-08, -3.41651906577747e-06, 0, -3.25322491207527e-06, 2.7270271936226e-07, 0, -1.37678721999845e-07, -5.49779974154591e-07, 2.00233665017243e-07, -7.47869452405903e-07, -3.9643654158042e-07, 2.99172753824049e-07, -2.19669703077936e-07, -3.36006013645746e-06, 0, 134521.003121696, 9.84673106505447e-07, 0, -1.03190254356463e-07, -2.79580482189172e-06, -2.09102250706516e-07, 3.0744994342735e-06, 2.62634610398959e-06, 6.77417934614406e-07, -2.08775586196951e-06, -2.36350458164721e-06, -1.40526690649747e-06, 1.58949483322488e-06, -1.77946652642517e-06, -3.04116293982512e-07, 5.07946329256271e-07, -1.04243908195391e-06, -1.84481103269824e-06, 8.67876486338307e-06, -9.31464682282846e-07, 9.27954426786087e-08, 1.73675041608978e-05, -3.53443458865212e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, -4.49999999864252, 0, 0, 0, 7.15305429895164e-08, 0, 60.7499999744937, 0, 92.4749999872517, 0, 0, 0, 0, 0, -3.17988228730393e-07, 0, -182.249999823951, 0, -683.437499911693, 0, 0, 0, -1248.41249997933, 0, 3.5061541320927e-08, 0, -1.24180088571055e-08, 0, 0, 8.32224654416939e-07, 0, -7.53650573643979e-07, 0, 2323.68749954793, -7.44192045271064e-08, 0, -1.03013580068161e-08, 3745.23749974312, 0, 9226.4062504331, 1.00246619850764e-07, 0, -1.06248936322095e-07, -5.34773126141494e-08, -3.46291125922236e-07, -1.242038591208e-07, 1.68722974136455e-07, 7.98743894688922e-08, -1.69928739838102e-07, 0, -1.30101837019879e-06, 0, 1.54227191782699e-06, 0, -3690.56249910933, 2.30677874094702e-07, 0, -1.68964157489387e-08, 2.37575592716836e-06, 0, -31369.7812515031, -1.50375924671424e-07, 0, 5.46014459214431e-07, -2.64981058446975e-07, -2.62302125527454e-08, 2.22209428891494e-08, -27679.2187532175, 0, 7.21888826422648e-08, 4.79206278593032e-07, -1.33301347237775e-07, -5.55693976379833e-07, 2.21045014000597e-06, -1.90686292781561e-07, 9.96476977786363e-07, -3.19161384052558e-08, 0, 1.25498680951227e-06, 0, -1.74130731694854e-06, 0, 2214.33749892183, -2.37220134000199e-07, 0, 7.50839166028063e-08, -4.24154860047531e-06, 0, 49822.5937512857, -2.39993490331056e-07, 0, -1.60575532670384e-07, 2.7510951148547e-07, -6.46010334780434e-07, -4.19581691781426e-08, 3.93401058404352e-06, 0, 94109.3437584047, 1.19159833936702e-08, 0, 1.19492903843449e-06, 2.14597804589346e-07, -6.28610710329071e-07, -2.40547595441233e-07, -2.05722731137402e-06, 1.66880906399544e-06, -1.30744025194314e-06, 8.01553296230572e-07, 2.26157706870771e-07, -5.02322503855005e-06, -8.88761034854827e-06, -1.6082914554069e-06, 9.34780409083826e-06, 0, -5.92490184974039e-07, 0, 1.05009791650639e-06, 0, 6.54568744363268e-07, 2.73214511956245e-08, 0, 0, 4.1495774574978e-06, 0, -29893.5562500372, 4.02669700886721e-07, 0, 7.90670013426915e-08, -2.25311372985477e-07, 6.65782012403154e-07, -2.67332754108785e-07, 0, 0, -149467.78125709, -3.89168742153431e-07, 0, -3.33626203310676e-08, -4.91307469820335e-08, 7.26532022621767e-07, 1.34654767150012e-07, 2.91959374489372e-06, -5.66199472471636e-07, 2.50706760522067e-06, -9.4392566839512e-06, 0, -1.37442378156705e-08, 5.27623445215046e-06, 1.01713409022813e-06, -3.1274684881825e-06, -1.56388272085896e-06, -2.44345713871848e-06, 4.25891528201736e-06, 3.5173511171875e-06, -5.68965519674203e-06, -6.77099402965045e-06, 2.93259833722058e-06, -2.75068558681312e-06, -9.1168096473214e-06, 0, 1.32635105475591e-07, 0, -2.44047325720823e-07, 0, -2.34758845154697e-07, -3.11019939303937e-08, 0, 0, -1.36250362804887e-06, 0, -2.3071961128056e-07, -4.37907731963972e-08, 0, -2.60593981123068e-08, 6.04182959363503e-08, -2.20291431007234e-07, 0, -1.01772119072027e-06, 0, 89680.6687523165, -1.41682149916054e-07, 0, 5.06091249558955e-08, -4.83077646187035e-07, -1.5917946962508e-07, -5.30561314210336e-07, -7.97369826299779e-07, 4.26658308884893e-07, -2.95562585933485e-07, 3.12444162988229e-06, 0, 5.78643571691806e-06, -5.25669345570744e-07, 0, 2.44586670392923e-08, -1.2693177572603e-07, -3.42019493957438e-07, -1.65248295030177e-06, -3.44059417190892e-06, 4.54279997096216e-07, -9.90772328024904e-07, -6.47353042625083e-07, 2.42569792208701e-06, 1.17470928184906e-06, -3.54125847234062e-07, 2.92599412382782e-07, -4.05510395015594e-06, -2.49168382587313e-06, -2.64295446522266e-07, -2.31893886902632e-07, 8.17442020262716e-06, 2.45258246678007e-06, -3.46240764092213e-06, 3.31445644109018e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, -5.78571428192276, 0, 0, 0, -8.05873083890372e-08, 0, 52.071428537142, 0, 127.574999932161, 0, 0, 0, 0, 0, 3.86285105658916e-07, 0, 1.86949144316374e-07, 0, -1057.04999952868, 0, 0, 0, -1148.17499951466, 0, 0, -4.50846088461726e-08, 5.25443691922927e-08, -3.92204803610482e-08, 0, -1.07086722097845e-06, 0, -5.41258616690271e-07, 0, 4305.65624830308, 0, 0, 0, -1.9734574986793e-06, 0, 9513.44999751563, -2.64492228378731e-08, 0, 5.20344034091918e-08, 2.24361873592139e-07, -1.89883819133521e-07, 3.2034943665159e-07, 1.04460674622155e-07, 4.83262381826438e-07, -3.58905805936315e-07, 0, 1.77546558189812e-06, 0, 9.57210714012003e-07, 0, -9226.40624639356, -4.97581386725534e-08, 0, 0, 4.18208322700085e-06, 0, -38750.9062435747, 6.74490347199363e-08, 0, -2.08457046965142e-07, -2.93687509182188e-07, 9.95162683732682e-07, -8.16466194115577e-07, 7.87993158595953e-06, 0, 7.44421865792219e-08, -9.81549691397858e-07, -1.55702941929687e-06, 0, 2.19370420845518e-06, -2.15115300757747e-06, 6.17600174075742e-07, -8.12869271911352e-07, 0, -1.75407213857305e-06, 0, -9.95229687887595e-07, 0, 9964.51874567621, 4.1688482529885e-08, 0, -1.85261112393693e-08, -5.22190510165227e-06, 0, 83037.6562413535, 8.30561402369824e-08, 0, 4.39927117373527e-08, 3.75871183761285e-07, -9.26287619433446e-07, 1.19210269124132e-06, -1.27220850674841e-05, 0, -1.42390348146198e-05, 4.99595193438715e-08, 0, -2.7826864272851e-07, 1.45972832974126e-06, 2.22066955315083e-06, -2.35935729143019e-07, -5.30488507514101e-06, 3.80545197998709e-06, -1.196230039547e-07, 9.19689401363389e-07, 4.73043350837684e-06, -4.8527139633086e-06, 6.69238303561887e-06, -2.67940338997678e-06, 5.44565684892719e-07, 0, 9.15366607222221e-07, 0, 5.50170504079541e-07, 0, -4270.50803290398, 2.37649288037234e-08, 0, 0, 3.25228970984722e-06, 0, -89680.6687439321, -6.30136248696158e-08, 0, 0, -1.89716799657533e-07, 4.1624388719358e-07, -8.04509390949459e-07, 9.66436893307092e-06, 0, 1.1585555297957e-05, 2.16028717309865e-07, 0, -3.17304432997756e-08, -1.05796433225052e-06, -1.60119099187071e-06, 7.89838095820104e-08, 4.497624974966e-06, -3.59478877576827e-06, -1.49117156715323e-06, 1.479656289716e-05, 0, 1.26742547097181e-07, -3.19290975460262e-06, -4.2718806467325e-06, -1.32281845978359e-07, 1.12109397413071e-05, -4.98527184153896e-06, 3.37402364814024e-06, -5.22209125764528e-06, 9.06012558938803e-06, -3.57778071177836e-06, -3.4409153154994e-06, -6.30308312460319e-06, 3.27544006161784e-06, 0, -2.06638412846623e-07, 0, -1.3036336689073e-07, 0, -7.14456562638955e-07, 0, 0, 0, -8.3142927515663e-07, 0, 38434.5723196651, -2.83491079969883e-08, 0, 1.42488361238737e-08, 3.19032870541872e-08, -7.96415003767669e-08, 2.51942911538032e-07, -2.94241095044894e-06, 0, -3.84330568235623e-06, 4.37136656318832e-08, 0, 0, 5.43102165880032e-07, 4.13483382463951e-07, 2.7528525623807e-07, -1.28733935484116e-06, 1.20055571323804e-06, 5.35946459542158e-07, -5.77706411329045e-06, 0, -6.42804036827504e-06, 1.69737505770771e-07, 0, 0, 7.45777527245628e-07, 1.26849303548167e-06, 1.12535538973666e-06, -4.15882924685325e-06, 2.35017489557542e-06, -1.27833277234924e-06, 1.82009102071542e-06, -4.52509220722836e-06, 2.44327777963257e-06, 1.40549092541751e-06, 2.20208502390734e-06, -4.11683999385675e-06, 2.48313327045115e-06, 3.7307447877825e-06, -1.66439616083867e-06, -7.52030100484499e-06, 4.69516132461625e-06, 1.81560194825941e-06, 1.92518069950693e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, -10.1249999994131, 0, 0, 0, -1.82446918593398e-08, 0, 0, 0, 236.274107132681, 0, 0, 0, 0, 0, 9.60801352991624e-08, 0, 2.2377159144306e-08, 0, -2136.88124991377, 0, 0, 0, 3.70849677865128e-08, 0, 0, 0, -1.36092936952956e-08, 2.11220682458199e-08, 0, -2.88011743338788e-07, 0, -8.24591353937791e-08, 0, 9913.26093710912, 0, 0, 0, -1.4223254578479e-08, 0, -2.19895384195929e-07, 0, 0, -1.57529422051466e-08, -2.37760846420698e-08, 1.30190996646338e-07, -1.2277774138373e-07, -8.52012851230955e-08, -9.81814414077796e-08, -1.69853516191438e-07, 0, 5.09757009142394e-07, 0, 1.87238887975192e-07, 0, -25833.9374990529, 2.94032234088703e-08, 0, 0, -1.0596008391236e-08, 0, 8.62109842403534e-07, -1.66458885358422e-08, 0, 4.67795465033513e-08, -1.63910169240068e-08, -7.4881673532256e-07, 2.36373875409637e-07, -2.28378714029727e-07, 0, -4.26068985691229e-08, 2.93561834285836e-07, 1.39779560813107e-07, 7.19585353184285e-08, 2.80888416339374e-07, 3.55560764110507e-07, 9.07413980018485e-07, 1.14776365851964e-06, 0, -5.22761978673403e-07, 0, -2.31860376631776e-07, 0, 38197.3218737075, -3.78916850118237e-08, 0, 0, -1.10722676782157e-07, 0, -1.79222974654888e-06, -1.80188818273203e-08, 0, 0, 0, 1.26019401132555e-06, -2.70484372791483e-07, 1.22389232813581e-06, 0, 7.38771637666048e-08, -2.38505669569571e-08, 0, 1.61206301718109e-08, -3.0529437035345e-07, 7.08038721975829e-08, 6.75134053007614e-07, 5.27399720302312e-07, -4.19118653030846e-07, -1.43350083530525e-06, -3.86063671344689e-07, -4.54609370723451e-07, -2.27974288277843e-06, -2.31401395803483e-06, -2.18292361062799e-06, -2.75502019687031e-06, 0, 2.95128240620506e-07, 0, 1.46803359304371e-07, 0, -29893.5562491093, 1.8660041425309e-08, 0, 0, 2.40385804663773e-07, 0, 1.66056834122446e-06, 0, 0, 0, 0, -1.0280871058555e-06, 1.84036793816175e-07, -1.04906418572096e-06, 0, 7.94665048722877e-07, -7.02652406189338e-08, 0, 2.18165698327596e-08, 8.35880760839505e-08, -1.39450622556865e-07, -1.17355795559363e-06, -1.50490405505828e-06, 1.7413003161137e-07, 1.27340888863103e-06, -1.57885718904831e-06, 0, -6.03026045922673e-08, 6.24281971787991e-07, 6.20077127047474e-08, -2.78453777222942e-07, 1.07859434238067e-06, 2.00488457689e-07, 3.11000225177894e-06, 8.45682923560082e-07, 5.67179901640081e-07, 3.04867354206373e-06, 3.83920209742912e-06, 3.13185809308311e-06, 5.92161416621843e-06, 0, -6.77476078898692e-08, 0, -3.6202460566784e-08, 0, 9608.64308009736, 0, 0, 0, -1.17820504122524e-07, 0, -5.70508284476123e-07, 1.56818000285572e-08, 0, 0, 0, 3.2341108749997e-07, -6.01027373023122e-08, 2.83028292634136e-07, 0, -5.24316147703317e-07, 0, 0, 0, -9.01542937349459e-08, 5.03405452223918e-08, 5.31098621626442e-07, 7.983054599411e-07, -1.94490686112261e-08, -3.46881331409581e-07, 9.44388609775699e-07, 0, 5.3617532339326e-07, -2.69288074234229e-08, 0, 0, -1.33223810776369e-07, 9.64503794206049e-08, 2.4291289967498e-07, -1.50153775056981e-08, -2.31038158328836e-08, -9.16648390520255e-07, 5.02831895196133e-08, 4.70046851577524e-07, -1.09678585859124e-06, -3.58441267351419e-07, 5.2707092342116e-08, -1.23969762991426e-06, -1.08287622976362e-06, -8.99394186714629e-07, -9.91997819918076e-07, -1.79318915666408e-06, -2.00879554551594e-06, -3.49734886238491e-06, -3.54977389111115e-06}, 
                                                    {0, 0, -10.1250000005857, 0, 0, 246.399107155866, 0, 10.1249999970752, 10.1249999996418, 0, 0, -2373.15535725599, 0, 3.07927068342165e-08, 0, -236.274107093294, -236.274107140266, 0, 0, 0, 0, 12050.1421880134, 0, -1.62174750578573e-07, 0, 2136.88124965439, 2136.88124998515, 0, -4.22520166955003e-08, -3.85992089455429e-07, 0, 0, 0, -1.65501747548815e-08, -2.12086442832424e-08, 0, -35747.1984388506, 0, 4.66690068599205e-07, 0, -9913.26093621542, -9913.26093744343, 0, 1.99002605826361e-07, 1.55725907250807e-06, 0, 1.92743048917565e-06, -3.09901188341618e-08, 0, 8.22525186233514e-08, 3.34883647616891e-08, -2.86526833621209e-08, 7.59332235697714e-08, 1.24109385902715e-07, 7.61482544374868e-08, 2.83212483581853e-07, 0, 64031.2593771316, 0, -7.90396267607598e-07, 0, 25833.9374972588, 25833.9374997807, 0, -4.21982755007424e-07, -3.33152689534477e-06, 0, -5.05624371329752e-06, 5.64472439711598e-08, 0, -5.63527906004907e-07, -9.82940774045956e-08, 3.21691226686871e-07, -1.56100073443855e-07, -5.83098658790546e-06, 0, 2.61065000847559e-07, -3.66301754004638e-07, 2.42190859973103e-07, -2.29128039130055e-07, -1.03220885027629e-06, 1.03841941179065e-07, -1.46677682008804e-06, -8.5757263352641e-07, 0, -68090.8781269877, 0, 7.70887748024624e-07, 0, -38197.32187166, -38197.3218747342, 0, 4.64776114994094e-07, 4.00839093603662e-06, 0, 7.20996898448754e-06, -3.79475479485676e-07, 0, 1.57355610987104e-06, 1.84176266163411e-07, -7.33253643772332e-07, 2.24968481310787e-07, 8.63534007965795e-06, 0, 1.08625536746968e-05, -1.7706292410791e-06, 0, 4.24384828008908e-08, 7.37188552424897e-07, -4.493383648604e-07, -1.2142246406501e-07, 9.63657454846882e-07, -7.12078290162558e-07, 2.06679572223316e-06, 2.32322337492155e-07, -1.08325073320457e-06, 2.50515700443692e-06, 2.52610417009184e-06, 1.4838966280099e-06, 3.05083083779512e-06, 0, 39502.1993313656, 0, -4.0716080630294e-07, 0, 29893.5562478322, 29893.5562497882, 0, -2.42637911984078e-07, -2.54679307684378e-06, 0, -5.24034968662081e-06, -5.14626503961705e-08, 0, -1.41940910426219e-06, -1.10649203089053e-07, 8.06393159260101e-07, -8.20320617588695e-08, -6.57253614596741e-06, 0, -1.00419600321227e-05, 1.03789728588413e-06, 0, -7.63917247852661e-07, -4.33011182443001e-07, 5.34476540250299e-08, 5.35518242029687e-07, 5.80665408694086e-08, 8.27253706735505e-07, -3.2039057331095e-06, -1.009764011968e-05, 0, 1.97832596929357e-06, -7.35624076283317e-08, 1.64366907900455e-06, 0, -2.69102958747062e-06, 2.20789245165593e-06, -4.3373178052497e-06, -1.45779366178499e-06, -1.55145898219937e-06, -2.53104935543745e-06, -4.81063584689945e-06, -4.60190072415966e-06, -9.33747221670349e-06, 0, -9608.64308057168, 0, 8.92893966215948e-08, 0, -9608.64307977325, -9608.64308029216, 0, 6.85425446350599e-08, 6.55380450322099e-07, 0, 1.50375302735828e-06, -1.43006415605057e-08, 0, 4.1729642616231e-07, 3.58622042801649e-08, -2.81465270568229e-07, 6.17108893878093e-08, 1.98705715711206e-06, 0, 3.59851055312834e-06, -2.53840282651353e-07, 0, 3.75252740010525e-07, 1.91510766760632e-07, 1.28061306367902e-08, -2.85903585316872e-07, -2.60254747132574e-07, -3.68679702437877e-07, 5.10397776355425e-07, 3.50020521369365e-06, 0, 4.44664977179395e-06, -8.10696004903508e-07, 0, -5.53147732716695e-07, 2.01377125957513e-07, -3.67664638868279e-07, -3.18616675975892e-07, 1.10597412911025e-06, -8.17767362489646e-07, 2.0765660366494e-06, 4.60718174034421e-07, 6.34902476275896e-07, 4.58310507829815e-07, 2.5789868373888e-07, -1.20488211322745e-06, 1.19619712446054e-06, 1.12703335232729e-06, 8.81193573819238e-07, 2.16437743834706e-06, 4.46289818237622e-06, 4.43912990214901e-07, 3.66210899362117e-06, 3.78509859503239e-06}, 
                                                    {0, 0, 46.285714287426, 0, 0, -1118.95714289314, 0, -98.3571428379719, -98.3571428521175, 0, 0, 10677.2464288658, 0, 52.0714283814512, 0, 2272.91785681227, 2272.91785710367, 0, 52.0714284853197, 104.142857134636, 0, -53563.2750012463, 0, 9.87674795530433e-07, 0, -20266.1999976306, -20266.1999998263, 0, 6.67050430686951e-07, -1148.17499767138, 0, -1148.17499964295, -4.54089197261505e-08, -2296.34999991368, 2.23123551960141e-07, 0, 156520.856253036, 0, -2.92588668783962e-06, 0, 92223.0562409514, 92223.0562493544, 0, -2.75709457968046e-06, -8.77749634588267e-06, 0, 9513.44998812168, 9513.44999978925, 0, -2.77447782684408e-06, 1.86853342974528e-07, 19026.900000538, -1.27535729260076e-06, -4.35817687620006e-07, 3.31949601950287e-07, -3.13484672470452e-06, 0, -275315.962504406, 0, 5.19111411620099e-06, 0, -234350.718730255, -234350.718748222, 0, 5.84853549119854e-06, 1.85466589278112e-05, 0, -38750.9062176481, -38750.9062505845, 0, 1.105851927373e-05, -1.39430587075757e-08, -77501.8125040588, 3.70137062221052e-06, 3.22281774029424e-05, 0, 7.60655378417191e-07, 4.47590480159742e-07, -3.8938588264889e-06, -7.22240223558593e-07, 9.21347196810918e-06, -3.012913809541e-06, 1.60561288573318e-05, 6.89581962235539e-06, 0, 286598.539289457, 0, -5.34646253326503e-06, 0, 335472.131225494, 335472.13124765, 0, -6.61366410947086e-06, -2.2954874459851e-05, 0, 83037.656201876, 83037.6562515075, 0, -2.17077139526619e-05, -4.50220835111191e-07, 166075.31250899, -6.03361608334434e-06, -4.45793594362417e-05, 0, -6.13399279811199e-05, 8.51617725236812e-06, 0, -8.89641875534951e-06, -1.75095218638766e-06, 6.19777749016884e-06, 6.6267550129032e-06, -7.33039275553963e-06, 7.85863278541871e-06, -3.01272298567256e-05, 2.27408673149621e-06, 1.4501350366329e-05, -3.51948794192083e-05, -6.08679633645363e-06, -1.70695533324056e-05, -2.07637409812929e-05, 0, -162279.305358853, 0, 2.9622967342986e-06, 0, -251959.974091006, -251959.974105397, 0, 3.7269088134378e-06, 1.53141280556284e-05, 0, -89680.6687136349, -89680.6687494918, 0, 1.81431077296811e-05, 4.69826946763297e-07, -179361.337508335, 4.18156865957922e-06, 3.32857938556554e-05, 0, 5.89317830565424e-05, -8.08399816588541e-06, 0, 1.65611059385847e-05, 1.36286664767391e-06, -3.09467417386959e-06, -1.08320458354921e-05, -9.53236678947084e-07, -7.48623726088648e-06, 3.41287212439682e-05, 5.15774528498528e-05, 0, -7.48676626956847e-06, -1.46908797456062e-06, -1.59096242013001e-05, -6.91422910702491e-06, 2.98603529721845e-05, -1.97141582769966e-05, 5.38685781674984e-05, -9.52513070281271e-07, 1.23220891847252e-05, 2.32331875620906e-05, 2.06569443857297e-05, 1.35701891853833e-05, 5.73889961443129e-05, 0, 38434.5723217513, 0, -6.76787994029966e-07, 0, 76869.1446384797, 76869.1446423288, 0, -9.1769820961448e-07, -4.1766800371772e-06, 0, 38434.5723105429, 38434.5723212248, 0, -5.63671926199674e-06, -1.79864949657703e-07, 76869.1446458335, -1.40135481917861e-06, -1.02698240543845e-05, 0, -2.19545714716245e-05, 2.02390425827688e-06, 0, -8.05413741843946e-06, -8.44529358811092e-07, 5.20945649244551e-07, 5.35997348863951e-06, 3.1458172721665e-06, 2.7923155202034e-06, -1.03954878724764e-05, -1.67194874979055e-05, 0, -2.34611234695641e-05, 5.49797278031151e-06, 0, 1.83326362729699e-07, -4.64807035656855e-07, 4.97167458930293e-06, 6.00963146869311e-06, -8.40479742963165e-06, 8.54936838245288e-06, -2.50900164929564e-05, 9.43408217725159e-07, -3.53248520947633e-06, -6.34840236790562e-06, 2.19679032926804e-06, 1.12668198652515e-05, -1.78500283191998e-05, -3.71656421647191e-06, 8.14484226466876e-07, -9.91215480350175e-06, -2.72688318652978e-05, -4.13195263896115e-07, -3.00341122765904e-05, -1.98568725489256e-05}, 
                                                    {0, 0, -126.000000001683, 0, 0, 3019.05000002543, 0, 429.749999923836, 429.749999978304, 0, 0, -28453.6125001133, 0, -485.999999263498, 0, -9803.36249868836, -9803.36249986552, 0, -485.999999587175, -972.000000064004, 0, 140501.081249998, 0, 182.249996120377, 0, 85789.6312405418, 85789.6312497385, 0, 182.249996632534, 10534.0499912679, 0, 10534.049998173, 546.750000931691, 21068.1000009025, 546.749999924588, 0, -402804.393748611, 0, 1.18686897611579e-05, 0, -380770.368713622, -380770.368749275, 0, 1.36745088126553e-05, -3745.23746769042, 0, -85046.9624556987, -85046.9625012037, 0, -3745.23748386869, -3.31010644546988e-06, -170093.925009059, 2.71933576429474e-06, -11235.7125057298, -7.76418654503857e-06, -11235.7124955619, 0, 693087.637495611, 0, -2.17439546765556e-05, 0, 937402.874920197, 937402.874997484, 0, -2.90258576147154e-05, -6.96750896351675e-05, 0, 333995.906128413, 333995.906263342, 0, -6.27147933049416e-05, 4.70107219441382e-06, 667991.812534425, -1.13571008024273e-05, 27679.2186360527, 0, 27679.218743401, 1.72459676259396e-05, 2.82981310600043e-05, 83037.6562809751, 83037.6562550436, 3.10840497990031e-05, -4.40326290813555e-05, 2.13018693827965e-05, 0, -704159.324993829, 0, 2.30918019862237e-05, 0, -1290958.76240051, -1290958.7624958, 0, 3.29431051980556e-05, 8.93083520340655e-05, 0, -680908.781066849, -680908.781271671, 0, 0.000115039208930407, -3.31025120627646e-06, -1361817.56256189, 2.10591916630963e-05, 0.000156409103788538, 0, -94109.3435353771, -94109.3437923859, 0, 6.00890267121806e-05, -1.76720478615717e-05, -3.89768825410886e-05, -282328.031327372, -282328.031305251, -5.83331992574406e-05, 0.000115319650501817, -5.25825016096685e-05, -8.46469714824872e-05, 7.23513723542234e-05, -7.61937567706196e-05, -6.33066358261849e-05, -3.05778810248863e-05, 0, 388616.231245794, 0, -1.31093143146473e-05, 0, 926700.243684296, 926700.243746321, 0, -1.87145487590988e-05, -6.17182594603202e-05, 0, 687551.793609517, 687551.793761559, 0, -9.48253082882624e-05, 8.69836224465313e-07, 1375103.58755184, -1.58353240499227e-05, -0.000118727677002027, 0, 149467.781043267, 149467.781302474, 0, -0.000100365647208648, 1.10337733121306e-05, 2.23703005576933e-05, 448403.343837906, 448403.343830045, 5.1474930095943e-05, -0.000136267768746562, -0.000174017033560462, 0, 2.26767072557246e-05, 3.78316698321148e-05, 8.05089029575065e-05, 8.19533749766458e-05, 0, 0.000101577160007212, -0.000171202274756999, 8.05054852327419e-05, 0.000107423875455615, 1.23211165651581e-06, 0.000104553498594205, 0.000103231803952088, 2.10716202804839e-05, 0, -89680.6687488723, 0, 3.05805589067559e-06, 0, -269042.006232176, -269042.006248777, 0, 4.44529357245458e-06, 1.73994799608032e-05, 0, -269042.006207132, -269042.006252684, 0, 2.93827596707851e-05, 8.03933175919388e-08, -538084.012517172, 5.15414803932241e-06, 3.79615633154693e-05, 0, -89680.6686722902, -89680.6687667025, 0, 4.90234677072041e-05, -1.49105918369294e-06, -4.40417745855462e-06, -269042.00628767, -269042.00628821, -1.81941546817371e-05, 4.92918937339033e-05, 5.61685519566461e-05, 0, 7.84977503594244e-05, -2.74781583819996e-05, 0, 1.0223086353018e-05, -9.09850266217623e-06, -2.48903162387073e-05, -4.43951764417384e-05, -2.59210616505945e-05, -4.43122497529933e-05, 0.000102561333004387, -2.75566416459708e-05, -5.09953244956847e-05, 7.55853379590386e-07, -2.85283002747312e-05, -4.97681775994677e-05, 3.69741035881039e-05, -5.35745162910749e-05, -6.29639176119558e-05, -5.87304929156347e-05, -3.98402491281927e-05, -5.97483827000957e-05, 1.69787189874351e-05, -6.5504539745196e-05}, 
                                                    {0, 0, 226.799999998109, 0, 0, -5366.24999992454, 0, -1113.74999980338, -1113.74999997136, 0, 0, 49708.6874990471, 0, 2035.12499810075, 0, 24953.0624966701, 24953.0625000602, 0, 2035.12499926193, 4070.25000067451, 0, -240251.062494258, 0, -1640.2499897609, 0, -212799.656226153, -212799.656252349, 0, -1640.24999289476, -43079.3437285053, 0, -43079.343748522, -4920.75000531549, -86158.6875085836, -4920.75000545985, 0, 671948.91560612, 0, 492.074967732338, 0, 913332.206158436, 913332.206259397, 0, 492.074969590735, 32722.9874200238, 0, 335615.653018937, 335615.653149295, 0, 32722.9874666091, 1968.30001713218, 671231.306304276, 1968.30001787225, 98168.9625461601, 2952.45004477174, 98168.9625438917, 0, -1125621.56246422, 0, 6.04005615894447e-05, 0, -2158979.06229934, -2158979.06251687, 0, 6.57561887846029e-05, -9226.40607084754, 0, -1254791.24971119, -1254791.25009842, 0, -9226.40609606506, -2.77045345813225e-05, -2509582.50016435, -3.08951935903238e-05, -230660.155986336, 0, -230660.156279436, -36905.6251159531, -0.000118871540215366, -691980.468944382, -691980.468986463, -0.000158065539328323, -36905.6250602815, -55358.4377822573, 0, 1112704.59371099, 0, -6.51685835844414e-05, 0, 2839887.84350003, 2839887.84376567, 0, -7.44804834530046e-05, -0.000238610685894716, 0, 2399788.26518846, 2399788.26577327, 0, -0.000292985907062751, 2.50683977109308e-05, 4799576.53150904, 3.06353179727859e-05, 58126.3590052392, 0, 730731.374522416, 730731.375224532, 0, 58126.3592562128, 0.000136374768486878, 0.000149314565805927, 2192194.12539448, 2192194.12555092, 0.000276604594695824, -5.45495171294488e-05, 232505.437817468, 0.000305618341213993, 232505.437820107, 0.000580537890712049, 348758.157070698, 0.000612195964270543, 0, -597871.124977299, 0, 3.73847330183477e-05, 0, -1943081.15608521, -1943081.15625683, 0, 4.18382679533534e-05, 0.000169543958245365, 0, -2242016.71841211, -2242016.71884904, 0, 0.000244824142245234, -1.16209571846781e-05, -4484033.43770185, -2.10269771462884e-05, 0.000293112003948631, 0, -1046274.4682965, -1046274.46900891, 0, 0.00025865375998148, -8.83948276582696e-05, -8.79045857229758e-05, -3138823.40663563, -3138823.40680536, -0.000248454413921114, 0.00016147204720451, -149467.780869582, 0, -149467.781386782, -0.00023700723133584, -0.000253639956773962, -597871.125408969, -597871.125643963, -0.000343146159203031, -1.52035942877366e-05, -0.00049618211918664, -896806.688521359, -0.000506950327872618, -0.0010245341981646, -0.000854655988279348, -0.00115893363383082, 0, 134521.00311954, 0, -8.79357414484613e-06, 0, 538084.012455433, 538084.012501041, 0, -9.62664309589611e-06, -4.87504569378912e-05, 0, 807126.018645596, 807126.018776227, 0, -7.57166584508569e-05, 1.98070511732997e-06, 1614252.03756285, 5.57516877250962e-06, -9.8253980257758e-05, 0, 538084.012329407, 538084.012589739, 0, -0.000135329304223548, 2.08597831592727e-05, 1.90790972664094e-05, 1614252.03764883, 1614252.03771029, 9.06839858079578e-05, -7.13415705310966e-05, -0.000126200707329772, 0, 134521.002960608, 134521.003245506, 0, -1.48250301544196e-05, 6.77658348715762e-05, 7.4675748800853e-05, 538084.012686852, 538084.012847824, 0.000148813411008035, -0.000126404920022787, 0.000156202804230573, 807126.019177859, 0.00017091567998753, 0.000143875875824726, 0.000138596565069151, 0.000181353731330696, 0.000423217944318847, 0.000371880556826016, 0.000569260642863802, 0.000754508110664461, 0.000410840672234134, 0.00046086667973507, 0.000781570240547596}, 
                                                    {0, 0, -283.49999999296, 0, 0, 6580.23749978913, 0, 1902.48749965779, 1902.48750003741, 0, 0, -59404.3874976914, 0, -5049.84374670452, 0, -41490.7312443161, -41490.7312512494, 0, -5049.8437499153, -10099.6875024348, 0, 278498.503112253, 0, 6629.34373199166, 0, 340634.362459801, 340634.362511692, 0, 6629.34374472689, 103210.453089246, 0, 103210.453133242, 19888.03126612, 206420.906280011, 19888.0312740968, 0, -753879.403085429, 0, -4305.65619215617, 0, -1394602.05922162, -1394602.05941855, 0, -4305.65622102426, -126606.796741668, 0, -763023.796704616, -763023.796973303, 0, -126606.796857298, -17222.6250498101, -1526047.59391611, -17222.6250995058, -379820.390779269, -25833.937635163, -379820.390838724, 0, 1222683.356178, 0, 1107.16864040697, 0, 3127751.71841545, 3127751.71883353, 0, 1107.16868350493, 76579.1715661796, 0, 2661818.20266676, 2661818.20345097, 0, 76579.171691328, 5535.84383434893, 5323636.40670728, 5535.84397993693, 832221.843344093, 0, 832221.843949253, 306316.687871538, 11071.687817095, 2496665.53185738, 2496665.53218164, 11071.6879572509, 306316.688048835, 459475.032269063, 0, -1172491.70617388, 0, 0.00011906905626024, 0, -3902769.84333501, -3902769.84383822, 0, 7.3726354607186e-05, -18268.2839504387, 0, -4691627.57743062, -4691627.57860467, 0, -18268.2839741053, -8.15935805902397e-05, -9383255.15692071, -0.000312224371458028, -440099.577544228, 0, -2383180.73367972, -2383180.73507107, 0, -440099.578101978, -91341.4223276052, -0.000382483763900152, -7149542.2042618, -7149542.20500883, -0.000794215684833511, -91341.4224809913, -1760398.31350011, -0.000742568806743982, -1760398.31427437, -182682.845646768, -2640597.4717141, -182682.846023401, 0, 612817.903081759, 0, -6.85198130697016e-05, 0, 2540952.28097777, 2540952.28129893, 0, -3.87287403863293e-05, -0.000307924213054976, 0, 4035630.09320751, 4035630.09408373, 0, -0.000346751623825712, 4.22221332956408e-05, 8071260.18799818, 0.000233315004303247, 89680.6682671385, 0, 2989355.62435244, 2989355.62575194, 0, 89680.6684124893, 0.000293215301847272, 0.000227945043116223, 8968066.87603014, 8968066.87674361, 0.000732216201463323, 0.000304876723357269, 971540.577590737, 0, 971540.578603533, 448403.344515123, 0.00055961622724717, 3886162.31368233, 3886162.31506673, 0.000799920603914122, 448403.344927803, 0.00155316907402774, 5829243.47226262, 0.00204585185599013, 896806.691029461, 0.00277559730267282, 896806.691936837, 0, -134521.003114803, 0, 1.61461229874644e-05, 0, -672605.015551804, -672605.015636129, 0, 7.97569546238899e-06, 8.9619865124917e-05, 0, -1345210.03108048, -1345210.03134119, 0, 0.000106972356345341, -8.76971487284376e-06, -2690420.06264898, -6.9928996763275e-05, 0.000169702006235842, 0, -1345210.03100541, -1345210.0315174, 0, 0.000205394900525143, -7.43776229987175e-05, -5.28846395682048e-05, -4035630.09412078, -4035630.0943623, -0.000275294531514851, -7.64396161392083e-05, -134521.002942676, 0, -672605.015409974, -672605.015959296, 0, -134521.003163343, -0.000227272093240491, -0.000159703400072674, -2690420.06300046, -2690420.06372013, -0.000347871280525235, -0.000185856054181203, -0.000478102708645454, -4035630.09518193, -0.000728575184168009, -672605.01605184, -0.000278485976617737, -672605.016601721, -0.00141973996559064, -0.00111492064604368, -1345210.03322185, -1345210.03398037, -0.00127622351100517, -0.00190873598642399, -0.002716740327006}, 
                                                    {0, 0, 251.999999991857, 0, 0, -5660.09999976522, 0, -2258.09999960529, -2258.10000015336, 0, 0, 48984.0749974997, 0, 8293.72499627357, 0, 47071.7999935259, 47071.8000028374, 0, 8293.72500145635, 16587.4500046447, 0, -219410.77498645, 0, -15992.4374797486, 0, -363360.937454744, -363360.93752271, 0, -15992.4375035976, -159942.599960644, 0, -159942.600022551, -47977.3125290357, -319885.200056575, -47977.3125488784, 0, 568879.70620848, 0, 17085.9374344367, 0, 1387829.19357864, 1387829.1938333, 0, 17085.9375005668, 284241.65610644, 0, 1086105.2060673, 1086105.2064286, 0, 284241.656280635, 68343.7500884085, 2172210.41279473, 68343.750214517, 852724.969041343, 102515.625241464, 852724.969193982, 0, -887949.337425188, 0, -9595.46237492413, 0, -2913699.09337871, -2913699.09391255, 0, -9595.46249380771, -278637.468412162, 0, -3422996.71826136, -3422996.7193073, 0, -278637.468652143, -47977.3126517299, -6845993.43827156, -47977.3130251218, -1666288.96834642, 0, -1666288.9691523, -1114549.87569566, -95954.6255408291, -4998866.90736654, -4998866.90805432, -95954.6257951381, -1114549.87627742, -1671824.8144625, 0, 823733.549921535, 0, 2214.33736356638, 0, 3428901.6182926, 3428901.61892701, 0, 2214.3374988292, 142824.768274545, 0, 5480485.31175365, 5480485.31331763, 0, 142824.768445978, 13286.0251502035, 10960970.6260842, 13286.0257460443, 1372889.24942875, 0, 4107596.0618468, 4107596.0636559, 0, 1372889.25019165, 714123.84460523, 33215.0631439795, 12322788.189511, 12322788.1909495, 33215.0638759455, 714123.8455005, 5491557.00187217, 44286.7511948542, 5491557.00375469, 1428247.69105431, 8237335.50568413, 1428247.69188666, 0, -418509.787455722, 0, 7.86207128113903e-05, 0, -2122442.49345228, -2122442.49385221, 0, -7.51058251544167e-06, -29893.5558991377, 0, -4334565.65565975, -4334565.65682796, 0, -29893.5559729055, -8.05329322150005e-05, -8669131.31327948, -0.000566497999547902, -567977.568256812, 0, -4484033.43689671, -4484033.43872229, 0, -567977.568507678, -179361.338047062, -0.000387288238010648, -13452100.3142489, -13452100.3155686, -0.00128117545862161, -179361.338753963, -2391484.49954789, 0, -2391484.50081923, -2839887.84520088, -0.00086228427409155, -9565938.00213631, -9565938.00493652, -0.0012392125591554, -2839887.84677062, -448403.346636068, -14348907.006619, -448403.347812841, -5679775.69423127, -597871.130126646, -5679775.69606417, 0, 89680.6687396181, 0, -1.85315539126715e-05, 0, 538084.012420634, 538084.012524349, 0, 4.21865018319521e-06, -0.00010320908707569, 0, 1345210.03106347, 1345210.03141012, 0, -8.3961399977295e-05, 1.78154383530072e-05, 2690420.06272688, 0.000172173685234538, 89680.6685684438, 0, 1793613.37476742, 1793613.37544162, 0, 89680.6685604857, 0.000140343918363105, 9.27122409780836e-05, 5380840.12560165, 5380840.12603814, 0.000487975662591775, 0.000405951458930991, 538084.012347088, 0, 1345210.03108503, 1345210.03177416, 0, 538084.012614802, 538084.012932626, 0.000243434638366395, 5380840.12587596, 5380840.12723511, 0.000536871652303292, 538084.013330395, 0.000883233850426154, 8071260.19016782, 0.0014851262315253, 2690420.06328933, 0.000399707028316075, 2690420.06454726, 1345210.03394953, 0.00202164735163302, 5380840.12876926, 5380840.13014818, 0.00230384744670198, 1345210.03503203, 1793613.3801186}, 
                                                    {0, 0, -161.999999996156, 0, 0, 3395.63571416972, 0, 1937.63571400342, 1937.63571447043, 0, 0, -27125.3089273126, 0, -9686.58749745861, 0, -36811.8964239493, -36811.8964314847, 0, -9686.5875021816, -19373.1750051296, 0, 113455.181243069, 0, 26275.8937365802, 0, 253186.256217832, 253186.25627198, 0, 26275.8937600578, 166006.968723548, 0, 166006.968774933, 78827.681281451, 332013.937561994, 78827.6813061645, 0, -278309.418728497, 0, -41826.3749564252, 0, -876754.631128717, -876754.631330183, 0, -41826.3750273542, -403788.543660286, 0, -960407.381127844, -960407.381419524, 0, -403788.543809321, -167305.500094658, -1920814.76280952, -167305.500253742, -1211365.63157477, -250958.250263045, -1211365.63176828, 0, 415557.337460897, 0, 39119.9624159838, 0, 1701349.31223884, 1701349.31265791, 0, 39119.9625522401, 572037.187290261, 0, 2649823.87466754, 2649823.87551373, 0, 572037.187518859, 195599.812661472, 5299647.7507798, 195599.813133352, 1896949.12477062, 0, 1896949.12540474, 2288148.75077171, 391199.625578345, 5690847.37622464, 5690847.37701403, 391199.625851293, 2288148.75159019, 3432223.12723216, 0, -372008.699958716, 0, -19929.0374074978, 0, -1879972.53718096, -1879972.53767447, 0, -19929.0375668911, -471653.887196867, 0, -3819732.18697979, -3819732.1882532, 0, -471653.887411773, -119574.225159566, -7639464.37605887, -119574.225909499, -2059333.8747062, 0, -3919377.37464379, -3919377.3760607, 0, -2059333.87529173, -2358269.43844221, -298935.563185549, -11758132.1271551, -11758132.128699, -298935.563957897, -2358269.4398775, -8237335.50207976, -398580.751234324, -8237335.50444205, -4716538.87899844, -12356003.2564089, -4716538.88002672, 0, 183631.845512309, 0, 4270.50798207652, 0, 1106061.58104449, 1106061.58135239, 0, 4270.50808344689, 209254.893519942, 0, 2780100.73083013, 2780100.73178558, 0, 209254.89365473, 29893.5563357207, 5560201.46323994, 29893.5569407505, 1165848.69348895, 0, 3736694.53090751, 3736694.53236226, 0, 1165848.69368986, 1255529.36309002, 89680.6691639168, 11210083.5955753, 11210083.5969381, 89680.6701001841, 1255529.36436322, 2839887.84358033, 0, 2839887.84451166, 5829243.47035828, 149467.782130674, 11359551.377339, 11359551.3805106, 149467.782480279, 5829243.47268899, 3138823.40948469, 17039327.0698717, 3138823.41100426, 11658486.9450995, 4185097.88075382, 11658486.947234, 0, -38434.5723159226, 0, 1.26931289707848e-05, 0, -269042.006195859, -269042.006274851, 0, -1.38639889180791e-05, -38434.5722523132, 0, -807126.018615503, -807126.018898975, 0, -38434.5722945244, -1.92594340484728e-05, -1614252.03771033, -0.000209842550154589, -269042.006146391, 0, -1345210.03110887, -1345210.03165611, 0, -269042.006154795, -269042.006399675, -9.98811690233675e-05, -4035630.09435824, -4035630.09479675, -0.000513742383126189, -269042.006877361, -807126.018704429, 0, -1345210.03119872, -1345210.03171602, 0, -807126.018878111, -1614252.03797219, -0.000249905970643674, -5380840.12594661, -5380840.12740769, -0.000522098361372081, -1614252.03877392, -807126.019734699, -8071260.19044452, -807126.020515965, -4035630.09462052, -0.000383917584223179, -4035630.09617409, -4035630.09679093, -1345210.0335111, -8071260.191751, -8071260.19324592, -1345210.03381598, -4035630.09815174, -5380840.13075515}, 
                                                    {0, 0, 81.0000000007136, 0, 0, -1333.31785715204, 0, -1333.31785703806, -1333.31785723547, 0, 0, 9202.97410717727, 0, 9202.97410622467, 0, 18405.9482126123, 18405.9482155993, 0, 9202.97410835651, 18405.9482173935, 0, -34932.7687499245, 0, -34932.7687451803, 0, -104798.306238511, -104798.306259762, 0, -34932.768756385, -104798.306241102, 0, -104798.306261679, -104798.306269091, -209596.612536737, -104798.306285417, 0, 80033.9484366751, 0, 80033.9484213726, 0, 320135.793706916, 320135.793785847, 0, 80033.9484573144, 320135.793722584, 0, 480203.690585482, 480203.690700687, 0, 320135.793779778, 320135.793807054, 960407.381426063, 320135.793913503, 960407.381447948, 480203.690788854, 960407.381573895, 0, -113669.324997734, 0, -113669.324967901, 0, -568346.624907924, -568346.625071149, 0, -113669.325040456, -568346.624934084, 0, -1136693.24988929, -1136693.25023062, 0, -568346.625019154, -568346.625096332, -2273386.5004274, -568346.625412355, -1136693.24994893, 0, -1136693.25017527, -2273386.50047028, -1136693.25036111, -3410079.75073206, -3410079.75121025, -1136693.25053115, -2273386.50103489, -3410079.75139583, 0, 97984.4343720249, 0, 97984.4343388175, 0, 587906.606138922, 587906.606329186, 0, 97984.4344260825, 587906.606146797, 0, 1469766.51544356, 1469766.51596598, 0, 587906.606233924, 587906.60634412, 2939533.03181035, 587906.606843378, 1469766.5155786, 0, 1959688.6874341, 1959688.6879656, 0, 1469766.51573325, 2939533.03181845, 1469766.51605681, 5879066.06376237, 5879066.06465175, 1469766.51653299, 2939533.03286613, 5879066.06375602, 1959688.68826901, 5879066.0652693, 5879066.0650068, 8818599.09768324, 5879066.06567958, 0, -46975.5883909273, 0, -46975.5883715503, 0, -328829.118679689, -328829.118796834, 0, -46975.5884276363, -328829.118666206, 0, -986487.356098252, -986487.356492916, 0, -328829.118731793, -328829.11879993, -1972974.71287919, -328829.119198424, -986487.356199537, 0, -1644145.59367523, -1644145.5942522, 0, -986487.3562039, -1972974.71284856, -986487.356513931, -4932436.78229547, -4932436.78305473, -986487.357086386, -1972974.71381502, -1644145.59377737, 0, -1644145.59405112, -4932436.78221262, -1644145.5943087, -6576582.37639665, -6576582.37830354, -1644145.59450408, -4932436.78382464, -4932436.78328719, -9864873.56696188, -4932436.78431253, -9864873.56720044, -6576582.37864065, -9864873.56851699, 0, 9608.64307986014, 0, 9608.64307526625, 0, 76869.1446247207, 76869.1446543499, 0, 9608.64308998856, 76869.1446164695, 0, 269042.006200375, 269042.006317186, 0, 76869.1446399584, 76869.1446539802, 538084.012604744, 76869.1447782927, 269042.006223203, 0, 538084.012461941, 538084.012687349, 0, 269042.00620491, 538084.012586906, 269042.006314343, 1614252.03783908, 1614252.03807608, 269042.006566275, 538084.012949191, 538084.012520138, 0, 672605.01564511, 672605.01582234, 0, 538084.012532382, 1614252.03777599, 538084.012663104, 2690420.06305992, 2690420.06390808, 538084.012814706, 1614252.03838744, 1614252.03812078, 4035630.09551547, 1614252.03865323, 2690420.06301789, 672605.015862821, 2690420.06399663, 4035630.09563597, 2690420.06394636, 5380840.12760699, 5380840.12848716, 2690420.06413308, 4035630.09651251, 5380840.12856834}, 
                                                    {0, 0, 0, 81.0000000015749, 0, 0, -1333.31785717827, 0, -1333.31785713707, -1333.31785716376, 0, 0, 9202.97410745311, 0, 9202.97410751102, 0, 9202.97410692068, 18405.9482144581, 18405.948214245, 18405.9482161292, 0, 0, -34932.7687514189, 0, -34932.7687528204, 0, -34932.7687495322, -104798.306250635, -104798.306249086, 0, -104798.306252073, -104798.306246732, -104798.306261937, -104798.306268515, -209596.612525057, 0, 0, 80033.948441263, 0, 80033.9484481547, 0, 80033.9484381836, 320135.79375184, 320135.793746667, 0, 320135.793764816, 0, 320135.793747262, 480203.690627047, 480203.690589001, 320135.793786742, 320135.793821437, 960407.381372077, 480203.690737964, 960407.381371632, 960407.381457141, 0, 0, -113669.325006006, 0, -113669.325021605, 0, -113669.325002734, -568346.625004053, -568346.624997334, 0, -568346.625050379, 0, -568346.625029282, -1136693.24999607, -1136693.24986805, -568346.625061771, -568346.625138203, -2273386.50031134, 0, -1136693.25001581, -1136693.24990566, -1136693.25028796, -2273386.50028562, -1136693.25035017, -2273386.50067866, -3410079.75041914, -3410079.75071526, -3410079.75096054, 0, 0, 97984.4343806672, 0, 97984.4343991352, 0, 97984.4343774244, 587906.606255188, 587906.606253591, 0, 587906.606332647, 0, 587906.606319433, 1469766.51561804, 1469766.51542194, 587906.606307647, 587906.606390247, 2939533.03169534, 0, 1469766.51568442, 0, 1469766.51558462, 1959688.68748433, 1959688.68715696, 1469766.51598726, 2939533.03160715, 1469766.51612427, 2939533.03233045, 5879066.06323096, 5879066.06365145, 1959688.6881447, 5879066.06316028, 5879066.06419615, 5879066.06436667, 5879066.06490238, 8818599.09608549, 0, 0, -46975.5883957798, 0, -46975.5884069097, 0, -46975.5883936142, -328829.118753304, -328829.118756268, 0, -328829.118814489, 0, -328829.118804928, -986487.356247832, -986487.356110074, -328829.11877805, -328829.118819786, -1972974.71283115, 0, -986487.356329866, 0, -986487.356318649, -1644145.59371886, -1644145.59334647, -986487.356473682, -1972974.71272336, -986487.356590726, -1972974.71331463, -4932436.78191874, -4932436.78220232, 0, -1644145.59375302, -1644145.59348531, -1644145.59426253, -4932436.78176944, -1644145.59433355, -4932436.78311431, -6576582.37567273, -6576582.37660448, -4932436.78286993, -4932436.78368953, -9864873.5648254, -6576582.37810757, -9864873.56542291, -9864873.56644945, 0, 0, 9608.64308098962, 0, 9608.64308369412, 0, 9608.64308043681, 76869.1446436588, 76869.1446453299, 0, 76869.1446621387, 0, 76869.1446569362, 269042.006250457, 269042.006214233, 76869.1446483785, 76869.1446562099, 538084.012598346, 0, 269042.006284874, 0, 269042.006296955, 538084.012488678, 538084.01234679, 269042.006304866, 538084.012554353, 269042.006340848, 538084.01272891, 1614252.03774774, 1614252.0378257, 0, 538084.012514872, 0, 538084.012456602, 672605.015609093, 672605.015432825, 538084.01264819, 1614252.03766, 538084.012691359, 1614252.03825741, 2690420.06277961, 2690420.06305261, 1614252.03801293, 1614252.03837161, 4035630.0946363, 672605.015840777, 2690420.06273812, 2690420.06334684, 2690420.06383743, 4035630.09501199, 2690420.06402495, 4035630.09584401, 5380840.12636505, 5380840.12685862, 5380840.12754556}, 
                                                    {0, 0, 0, -162.000000005839, 0, 0, 3395.63571441735, 0, 1937.63571424941, 1937.63571432286, 0, 0, -27125.3089297289, 0, -9686.58750069136, 0, -9686.58749878325, -36811.8964288458, -36811.8964285182, -19373.1750015484, 0, 0, 113455.18125532, 0, 26275.8937552272, 0, 26275.8937423957, 253186.256250714, 253186.256249816, 0, 166006.968754005, 166006.968739339, 78827.6812585743, 78827.6812564314, 332013.937532779, 0, 0, -278309.418764187, 0, -41826.375019608, 0, -41826.3749762141, -876754.631251106, -876754.631251418, 0, -403788.543777743, 0, -403788.543703477, -960407.381254951, -960407.381176745, -167305.500024146, -167305.499982674, -1920814.76268862, -250958.250053642, -1211365.63139254, -1211365.63144852, 0, 0, 415557.337522752, 0, 39119.9625395945, 0, 39119.9624545971, 1701349.31250129, 1701349.31251103, 0, 572037.18759268, 0, 572037.187428949, 2649823.87499607, 2649823.87477332, 195599.81253569, 195599.812373597, 5299647.75052265, 0, 1896949.12503238, 1896949.12471649, 391199.625162182, 2288148.75031146, 391199.625008208, 2288148.75041972, 5690847.37559685, 5690847.37587826, 3432223.12595617, 0, 0, -372008.700021601, 0, -19929.0375440424, 0, -19929.0374464466, -1879972.53750039, -1879972.53752317, 0, -471653.887651063, 0, -471653.887451487, -3819732.18749104, -3819732.18718359, -119574.22502626, -119574.22474876, -7639464.37577383, 0, -2059333.87511095, 0, -2059333.87464122, -3919377.37498084, -3919377.37425574, -298935.562713392, -2358269.43785961, -298935.562233572, -2358269.43788602, -11758132.1261705, -11758132.12664, -398580.75009146, -8237335.50086426, -8237335.50148457, -4716538.87700834, -4716538.87676058, -12356003.2528641, 0, 0, 183631.845546906, 0, 4270.50806126321, 0, 4270.50800129571, 1106061.58124913, 1106061.58126965, 0, 209254.89386733, 0, 209254.893729806, 2780100.73124722, 2780100.73104757, 29893.5562585159, 29893.5560315395, 5560201.46307889, 0, 1165848.69389681, 0, 1165848.69360195, 3736694.53120485, 3736694.53048152, 89680.6688746, 1255529.36270098, 89680.6683486812, 1255529.36260434, 11210083.5948866, 11210083.59526, 0, 2839887.84375976, 2839887.84300144, 149467.781339356, 5829243.46936481, 149467.780989447, 5829243.46996979, 11359551.376008, 11359551.3766321, 3138823.4080799, 3138823.40737677, 17039327.0657549, 4185097.87770836, 11658486.9412386, 11658486.9413597, 0, 0, -38434.5723238639, 0, -6.05189320185425e-06, 0, 9.035404462551e-06, -269042.006249379, -269042.006256048, 0, -38434.5723563109, 0, -38434.5723136818, -807126.018750379, -807126.018701683, -7.09138360076601e-07, 7.10047488455015e-05, -1614252.03767095, 0, -269042.006313881, 0, -269042.006244389, -1345210.03123446, -1345210.03097955, -2.61909550366719e-05, -269042.006290789, 0.000172562746058533, -269042.006217164, -4035630.09417923, -4035630.09429809, 0, -807126.01877762, 0, -807126.018531726, -1345210.03122395, -1345210.03082028, -1.77241125678847e-05, -1614252.03766623, 0.000220321528301254, -1614252.03791306, -5380840.12545858, -5380840.12564385, -807126.019336551, -807126.01891153, -8071260.18886807, 0.000111990413726785, -4035630.09406901, -4035630.09435408, -1345210.03255901, -4035630.09541697, -1345210.03217592, -4035630.09534941, -8071260.18944957, -8071260.18956861, -5380840.12779103}, 
                                                    {0, 0, 0, 252.000000012727, 0, 0, -5660.10000028754, 0, -2258.10000000307, -2258.10000003373, 0, 0, 48984.0750025376, 0, 8293.72500067251, 0, 8293.72499866157, 47071.8000001956, 47071.8000010063, 16587.4500006195, 0, 0, -219410.775011713, 0, -15992.4375050275, 0, -15992.4374898434, -363360.937499963, -363360.93750843, 0, -159942.600003872, -159942.599992049, -47977.3124987231, -47977.3124881895, -319885.200031305, 0, 0, 568879.706281385, 0, 17085.9375187553, 0, 17085.9374657239, 1387829.19374758, 1387829.19378521, 0, 284241.656276553, 0, 284241.656194412, 1086105.20625449, 1086105.20620378, 68343.7499875804, 68343.7498814784, 2172210.4127211, 102515.624934693, 852724.968845493, 852724.968844115, 0, 0, -887949.337550561, 0, -9595.46253770473, 0, -9595.46243262967, -2913699.0937422, -2913699.093832, 0, -278637.468838207, 0, -278637.46863979, -3422996.71874525, -3422996.71862913, -47977.3124658906, -47977.3121225508, -6845993.43817832, 0, -1666288.96877957, -1666288.9684543, -95954.6248573649, -1114549.87513983, -95954.6245183473, -1114549.87485702, -4998866.90682575, -4998866.90698505, -1671824.81287602, 0, 0, 823733.550048235, 0, 2214.33754165265, 0, 2214.33741989043, 3428901.61873726, 3428901.6188552, 0, 142824.76889384, 0, 142824.768649485, 5480485.31248909, 5480485.31237623, 13286.0249525794, 13286.0244295439, 10960970.6260557, 0, 1372889.25009984, 0, 1372889.24953576, 4107596.06248177, 4107596.06179775, 33215.0623167125, 714123.843833026, 33215.0613422225, 714123.843039253, 12322788.1888522, 12322788.1892569, 44286.7491969631, 5491557.00063483, 5491557.00045767, 1428247.68840305, 1428247.68724109, 8237335.50204382, 0, 0, -418509.787525101, 0, -2.39807514242037e-05, 0, 5.18251595813081e-05, -2122442.4937393, -2122442.49381859, 0, -29893.5563618973, 0, -29893.5561992192, -4334565.65624383, -4334565.65621001, 3.12917834759426e-05, 0.000424819434797429, -8669131.31330724, 0, -567977.568884652, 0, -567977.56849641, -4484033.43746118, -4484033.43683758, 0.000137251732200514, -179361.337490763, 0.00114795423874039, -179361.336684343, -13452100.3139398, -13452100.314374, 0, -2391484.50000152, -2391484.49915352, 0.00064409877074948, -2839887.84404166, 0.00132942876724929, -2839887.84344913, -9565938.00097194, -9565938.00097526, -448403.344632931, -448403.342296085, -14348907.0030151, -597871.125348917, -5679775.69010887, -5679775.688907, 0, 0, 89680.6687554865, 0, 5.63655229325534e-06, 0, -1.38141283818791e-05, 538084.012496438, 538084.01251792, 0, 3.32385793430783e-05, 0, -1.52967876914478e-05, 1345210.03124831, 1345210.03125331, -7.92815365449575e-06, -0.000125346237083859, 2690420.06274072, 0, 89680.668809495, 0, 89680.6687136353, 1793613.37498833, 1793613.37478892, -4.40695985618339e-05, -2.07568373673925e-05, -0.000408988556843138, -0.000316249820211734, 5380840.12556675, 5380840.1257374, 0, 538084.012521424, 0, 538084.012217314, 1345210.03122324, 1345210.03080245, -0.000212987504609868, 538084.012527176, -0.000667743730898454, 538084.012244945, 5380840.12552072, 5380840.12556791, 0.000274954957060036, -0.000931220962786563, 8071260.18900236, -0.000541597084305069, 2690420.06271665, 2690420.06247046, 0.000391277970534016, 1345210.03240432, -0.000560639465005429, 1345210.03113734, 5380840.12668551, 5380840.12621156, 1793613.37638402}, 
                                                    {0, 0, 0, -283.500000017956, 0, 0, 6580.23750040656, 0, 1902.48750005209, 1902.4875000154, 0, 0, -59404.3875035996, 0, -5049.84375035661, 0, -5049.84374944772, -41490.7312500081, -41490.7312516113, -10099.687500138, 0, 0, 278498.503141678, 0, 6629.34375268004, 0, 6629.34374453273, 340634.362499015, 340634.362512771, 0, 103210.453126686, 103210.453124716, 19888.0312442051, 19888.0312360476, 206420.906271338, 0, 0, -753879.403169872, 0, -4305.65626011996, 0, -4305.65623093757, -1394602.05936937, -1394602.05942605, 0, -126606.79688754, 0, -126606.796850025, -763023.796874336, -763023.796875913, -17222.6249708497, -17222.6248918087, -1526047.59392144, -25833.9374014169, -379820.390661277, -379820.390644495, 0, 0, 1222683.35632258, 0, 1107.16877050405, 0, 1107.16871198914, 3127751.71873517, 3127751.71886258, 0, 76579.1719191003, 0, 76579.1718234597, 2661818.20311401, 2661818.20314979, 5535.84368483873, 5535.84344208179, 5323636.40680936, 0, 832221.843754207, 832221.843603861, 11071.687243011, 306316.687483871, 11071.6869799497, 306316.687178007, 2496665.53161627, 2496665.53168978, 459475.031183311, 0, 0, -1172491.70631951, 0, -2.26150276440051e-05, 0, 4.64023880919516e-05, -3902769.84372824, -3902769.84388761, 0, -18268.2844498424, 0, -18268.2843410104, -4691627.57810724, -4691627.57821905, 7.82486944324761e-05, 0.00042790560303758, -9383255.15714885, 0, -440099.57815619, 0, -440099.577854461, -2383180.73434472, -2383180.73404513, 0.000353994773416031, -91341.4217483035, 0.00108397055042956, -91341.4210051282, -7149542.2041311, -7149542.20442151, 0.000952200241689953, -1760398.31275751, -1760398.31235931, -182682.843681076, -182682.84257522, -2640597.46961636, 0, 0, 612817.903161311, 0, 1.29551017651953e-05, 0, -3.1076145465107e-05, 2540952.28123327, 2540952.28133645, 0, 5.99233544049969e-05, 0, -5.00208232193832e-06, 4035630.09373707, 4035630.09386591, -4.639619264375e-05, -0.000298487371041926, 8071260.18819945, 0, 89680.6688055345, 0, 89680.6686018034, 2989355.62496591, 2989355.62472835, -0.000255960390787017, -0.000145648956582479, -0.000984158578384437, -0.000824065500455775, 8968066.87614836, 8968066.87652029, 0, 971540.578094902, 971540.577625762, -0.000807097154408083, 448403.343715313, -0.0013238862319528, 448403.342901442, 3886162.31309714, 3886162.31290331, -2.91094550065067e-05, -0.00215037728307071, 5829243.47058725, -0.00105938928844252, 896806.68846522, 896806.687100392, 0, 0, -134521.003132966, 0, -3.02499947476528e-06, 0, 8.65343531121028e-06, -672605.015619811, -672605.015646959, 0, -1.81001794159535e-05, 0, 5.2302972087777e-07, -1345210.03124546, -1345210.03129635, 1.08982111337319e-05, 8.39253799043518e-05, -2690420.0627109, 0, -2.7792235018937e-05, 0, 1.50526936690936e-05, -1345210.03124151, -1345210.0311833, 7.6121985510347e-05, 5.42647834941438e-05, 0.000328225178028024, 0.000276409723496282, -4035630.09421697, -4035630.09437785, 0, -134521.003122505, 0, -134521.002945168, -672605.01559863, -672605.015374463, 0.0002721371605116, 7.78979009615897e-05, 0.000604877503828063, 0.000435187869612109, -2690420.06287688, -2690420.06289598, 3.07956721891636e-05, 0.00107601761866494, -4035630.09482607, 0.000556060525202307, -672605.015691005, -672605.015339211, 0.000277583914055835, -0.00039113864668994, 0.00111760171404566, 0.000952120257330765, -1345210.03215494, -1345210.03157235, -8.11308767180478e-05}, 
                                                    {0, 0, 0, 226.800000016991, 0, 0, -5366.25000038546, 0, -1113.75000004466, -1113.75000000257, 0, 0, 49708.6875034227, 0, 2035.12500009824, 0, 2035.12499997047, 24953.0624999207, 24953.0625010475, 4070.25000004503, 0, 0, -240251.06251591, 0, -1640.25000077011, 0, -1640.24999873252, -212799.656248883, -212799.656258118, 0, -43079.343750033, -43079.3437521089, -4920.74999618625, -4920.74999356436, -86158.6875091733, 0, 0, 671948.915667955, 0, 492.07500308671, 0, 492.074995387103, 913332.206244686, 913332.206281877, 0, 32722.9875017675, 0, 32722.9874956329, 335615.653121206, 335615.653139047, 1968.29998107238, 1968.29995460897, 671231.306328779, 2952.44994531574, 98168.9625053715, 98168.9624977292, 0, 0, -1125621.56256971, 0, -6.49048760198946e-06, 0, 9.0262322515166e-06, -2158979.0624869, -2158979.06256896, 0, -9226.40625981747, 0, -9226.40624022031, -1254791.24998672, -1254791.25005476, 4.18572302113567e-05, 0.000118276913312114, -2509582.50026523, 0, -230660.15623779, -230660.156213804, 0.000156685104247799, -36905.6249545309, 0.000254517989158128, -36905.6248155902, -691980.46888816, -691980.468918409, -55358.4373739496, 0, 0, 1112704.593817, 0, 7.19680038201605e-06, 0, -1.1189724700828e-05, 2839887.84373158, 2839887.84383262, 0, 2.04057626481925e-05, 0, 3.997556751914e-06, 2399788.26560586, 2399788.26573734, -4.90166438805118e-05, -0.000149018855948682, 4799576.53168459, 0, 58126.3593605067, 0, 58126.359283307, 730731.374965336, 730731.374932114, -0.000227644035965022, -0.000118026186014524, -0.000486392835351164, -0.000433827230727502, 2192194.12543396, 2192194.12557835, -0.00050551036736879, 232505.437536975, 232505.437322727, -0.000252499790781941, -0.000767679277492672, 348758.156435631, 0, 0, -597871.125035109, 0, -4.09397888947136e-06, 0, 8.04723657102614e-06, -1943081.15623636, -1943081.15630116, 0, -1.8329651268653e-05, 0, -1.35579974213819e-05, -2242016.7187358, -2242016.71885622, 2.81576930178762e-05, 9.43998067025494e-05, -4484033.43784249, 0, 1.94632824323803e-07, 0, 5.29438269464187e-05, -1046274.46872044, -1046274.46872072, 0.000168015197095082, 0.00010954339842571, 0.000408290782014719, 0.000374531599349848, -3138823.40677045, -3138823.40696484, 0, -149467.781206931, -149467.781080489, 0.000451977458946732, 0.000103565482218275, 0.000639965257955327, 0.000526506171967325, -597871.12520461, -597871.125110619, 0.000228367513539518, 0.00119557382658206, -896806.688176713, 0.000840361797878249, -0.000118788544872237, 0.000588585392008234, 0, 0, 134521.003132726, 0, 9.44123457561023e-07, 0, -2.4523561524597e-06, 538084.012495889, 538084.01251277, 0, 5.90095349962721e-06, 0, 5.38605916056043e-06, 807126.018745159, 807126.018786723, -6.45593135633551e-06, -2.43404535796126e-05, 1614252.03760421, 0, 4.44028713475644e-06, 0, -3.58850154296204e-06, 538084.012492699, 538084.012508301, -4.99421654402236e-05, -3.61684459445335e-05, -0.000125885040757151, -0.000111855361172212, 1614252.0377168, 1614252.03780307, 0, -1.52622355479547e-05, 0, -7.02278787622999e-05, 134521.003102941, 134521.00304939, -0.000158980269370155, -6.84496797396278e-05, -0.000271114833862023, -0.000249883411991376, 538084.012653928, 538084.012681351, -9.10193283154933e-05, -0.00054883567539037, 807126.019203964, -0.000282447087542833, -7.86691789394231e-06, -0.000187767905598801, -0.000295746043337875, 2.28711861149942e-05, -0.000679197160107536, -0.000661010022690777, 0.000284537972727323, -2.48268612636877e-05, -0.000239967750928932}, 
                                                    {0, 0, 0, -126.000000010772, 0, 0, 3019.05000024478, 0, 429.750000016442, 429.7500000001, 0, 0, -28453.612502179, 0, -486.000000015156, 0, -486.000000040177, -9803.36249995476, -9803.36250033399, -972.000000044801, 0, 0, 140501.081260158, 0, 182.250000119856, 0, 182.249999914468, 85789.6312494432, 85789.631252564, 0, 10534.0499997897, 10534.0500008753, 546.749999003848, 546.7499987624, 21068.1000024081, 0, 0, -402804.393777506, 0, -5.62020378835991e-07, 0, 4.07842884011906e-07, -380770.368747417, -380770.368759881, 0, -3745.23749932404, 0, -3745.23749888245, -85046.962497573, -85046.9625063529, 5.44310169220223e-06, 8.20488289643229e-06, -170093.925020623, 1.36330262082619e-05, -11235.7125000741, -11235.712498632, 0, 0, 693087.637544771, 0, 1.29466671742429e-06, 0, -3.54145942680405e-07, 937402.874993745, 937402.875020967, 0, 7.00953030656237e-08, 0, -4.06876690566498e-06, 333995.906242635, 333995.906272955, -1.24853614061669e-05, -1.81441381866582e-05, 667991.812569274, 0, 27679.2187408609, 27679.2187427089, -4.40223692361555e-05, -1.65227436164443e-05, -5.82368057569546e-05, -4.85379457079859e-05, 83037.6562803829, 83037.6562882406, -4.21237770610487e-05, 0, 0, -704159.325043148, 0, -1.44171128138322e-06, 0, 0, -1290958.76249132, -1290958.76252446, 0, -2.71167826209784e-06, 0, -3.3591516826151e-07, -680908.781239681, -680908.781293785, 1.4420898693514e-05, 1.72764020176633e-05, -1361817.56261479, 0, 1.50171019853312e-05, 0, 2.73587852000803e-05, -94109.3437302963, -94109.3437472937, 6.81829414710139e-05, 4.01234964798191e-05, 0.000100685623781666, 0.000106846664963467, -282328.031355397, -282328.031395198, 0.000130140581001493, 4.78483381868764e-06, 5.88582219093455e-05, 9.5795014153234e-05, 0.000211689808152993, -1.82960973421657e-05, 0, 0, 388616.231272671, 0, 7.98236299668494e-07, 0, -1.79780672383384e-07, 926700.24374365, 926700.24376499, 0, 3.56587981746494e-06, 0, 3.59222044773484e-06, 687551.793742423, 687551.793788594, -8.04907795390814e-06, -6.46094358459572e-06, 1375103.58759103, 0, -8.730474548846e-06, 0, -2.2444211312383e-05, 149467.781234121, 149467.781263236, -5.18666686059911e-05, -3.4785219011805e-05, -7.40779264026599e-05, -8.68526970261272e-05, 448403.343879324, 448403.343932328, 0, -2.54886696627374e-05, -3.96769748467594e-05, -0.000122624731302032, -4.18361715929375e-05, -0.000150418286570302, -0.00015825832192678, 3.61333685955658e-05, 2.29846136835072e-05, -9.07129166446853e-05, -0.000313587224599486, 0.00014987882987139, -0.000257139072079013, -2.22726857422839e-05, -0.000211285959893616, 0, 0, -89680.6687550018, 0, -1.77333102122704e-07, 0, 1.31542441803928e-07, -269042.006248102, -269042.006253581, 0, -1.34315994398319e-06, 0, -2.05333247452273e-06, -269042.006247499, -269042.006262502, 1.81766029246695e-06, 5.07174690280965e-07, -538084.012527894, 0, 1.15824792023519e-06, 0, 4.97410655160588e-06, -89680.6687459455, -89680.6687617942, 1.55015185802745e-05, 1.08110615234239e-05, 1.87632530354416e-05, 2.19978817919436e-05, -269042.006304525, -269042.006328509, 0, 1.03251089315315e-05, 0, 2.21580929042219e-05, 1.14974475960789e-05, 1.19723373058536e-05, 4.60957451491943e-05, 2.40595246984942e-05, 5.74349193709784e-05, 7.60199915061747e-05, -3.38014522189579e-05, -4.94972787146261e-05, 3.61744459731627e-05, 0.000136797635858587, -0.00010865384689357, 7.22718515265815e-05, 9.67670486907232e-06, 5.66890137837966e-05, 0.000101177251160248, 1.96615639151563e-05, 0.000190073288806066, 0.000201665478970946, -5.43623268777706e-05, 2.574413111541e-05, 9.72897226211704e-05}, 
                                                    {0, 0, 0, 46.2857142901213, 0, 0, -1118.95714295744, 0, -98.3571428600822, -98.3571428576524, 0, 0, 10677.2464294662, 0, 52.0714285769899, 0, 52.0714285833112, 2272.91785713796, 2272.91785719452, 104.142857171219, 0, 0, -53563.2750041813, 0, -2.75281645131812e-08, 0, 0, -20266.1999998886, -20266.2000003838, 0, -1148.17499996882, -1148.17500012907, 0, -3.6788271463287e-08, -2296.35000048582, 0, 0, 156520.85626135, 0, 1.15209335024275e-07, 0, -9.07801056870706e-08, 92223.0562494156, 92223.0562513967, 0, -1.69894976756302e-07, 0, -6.83759586816049e-07, 9513.44999948172, 9513.45000129932, -4.06474664198125e-07, 8.99295684793332e-08, 19026.900003341, -5.73138083607531e-07, 8.52257238135694e-07, 7.8502848179098e-07, 0, 0, -275315.96251852, 0, -2.62039985165408e-07, 0, -1.01407284801697e-07, -234350.718748529, -234350.718752759, 0, 7.06864436536216e-08, 0, 2.62587445212136e-06, -38750.906248324, -38750.9062538805, 1.28644897928886e-06, -8.87480340295183e-07, -77501.8125099062, 0, 2.34384107409876e-06, 2.34513430113037e-06, 3.69543722430016e-06, -1.98769140363171e-07, 2.79699358638273e-06, 3.14329306254083e-06, -6.44333272033617e-06, -7.4856879733856e-06, -7.75452542600516e-07, 0, 0, 286598.539303605, 0, 2.76676507975666e-07, 0, 4.02951510903728e-07, 335472.1312479, 335472.131252934, 0, 5.13842036103944e-07, 0, -1.76318125607304e-06, 83037.6562476179, 83037.6562575118, -1.53477788347919e-06, 2.25155844398345e-06, 166075.312516173, 0, -4.25836003972247e-06, 0, -7.18980934663662e-06, -4.97820169064502e-06, 1.51358159822916e-07, -7.14739598895342e-06, -3.88800279612417e-06, -3.5879628060681e-06, -9.16851728440621e-06, 1.75734379140132e-05, 2.27818454070646e-05, -1.05759966841996e-05, 2.74066289823553e-06, 1.08367467037164e-06, -6.42228702597253e-06, -9.42029698637679e-06, 1.38016579816467e-05, 0, 0, -162279.305366564, 0, -1.43033152533167e-07, 0, -2.59809028684871e-07, -251959.974105583, -251959.974108967, 0, -7.78401189858575e-07, 0, 9.70081821842791e-07, -89680.6687482487, -89680.6687560991, 8.35474258834933e-07, -3.05866623647213e-06, -179361.337512224, 0, 2.67997425618225e-06, 0, 7.96447213161952e-06, 4.02030006087443e-06, -4.03716906242506e-06, 5.83679041229274e-06, 3.71348536392681e-06, 1.11652693606281e-06, 8.95578991302835e-06, -1.90870836916157e-05, -2.4735023083473e-05, 0, 6.79287753222319e-06, 7.28851609799371e-06, 9.72143031538416e-06, 2.10756228138233e-06, 9.88848891537901e-06, 1.65369146158174e-05, -8.08745026419334e-06, -1.2738059188925e-05, 6.47472875484527e-06, 2.55362114938498e-05, -3.45423045464056e-05, 1.91704492733562e-05, -4.21528207589148e-06, 1.83921102521312e-05, 0, 0, 38434.5723235114, 0, 3.01206281680863e-08, 0, 7.01755156857204e-08, 76869.1446423835, 76869.144643196, 0, 3.12644534234353e-07, 0, 7.89318352804278e-08, 38434.5723208525, 38434.572323227, -2.01863724710336e-07, 1.14606133117291e-06, 76869.144646574, 0, -4.35092079251662e-07, 0, -2.81048079031271e-06, -1.01185817641315e-06, 3.04910697154989e-06, -1.83873326049757e-06, -1.21096489136306e-06, 8.40500949359521e-07, -1.68008959206607e-06, 7.51545388149174e-06, 1.07200834779343e-05, 0, -2.90385162457869e-06, 0, -5.69567794947162e-06, -3.04292392591106e-06, -8.66158920979243e-07, -4.89614595469228e-06, -2.95113462066738e-06, -3.14866007428232e-06, -1.12708621793493e-05, 5.28620030404095e-06, 9.82028846449311e-06, -3.63195900109059e-06, -1.20951164464231e-05, 1.89003801566056e-05, -6.81571782589206e-06, 2.20691991510945e-07, -5.36712386440859e-06, -8.87077898458119e-06, 3.1109957973678e-07, -1.65016860251425e-05, -2.09264407426949e-05, 1.5836580288034e-05, 8.86310289056254e-06, -2.67830222009255e-06}, 
                                                    {0, 0, 0, -10.125000001055, 0, 0, 246.3991071669, 0, 10.1250000003732, 10.1250000003093, 0, 0, -2373.15535735774, 0, 0, 0, 0, -236.274107145666, -236.274107148231, 0, 0, 0, 12050.1421885062, 0, 1.07370947282174e-08, 0, 0, 2136.88125000458, 2136.88125003221, 0, 1.71156196080908e-08, 3.27130977424692e-08, 5.01633015712733e-08, 4.41647108142517e-08, 9.91709947561909e-08, 0, 0, -35747.1984402371, 0, -3.23399550813562e-08, 0, 2.25338149173246e-08, -9913.26093748033, -9913.26093759588, 0, -4.61197017529979e-08, 0, 7.51415276564235e-08, -3.13067463811492e-08, -2.72936454987486e-07, -9.65029915321258e-08, -2.02638881130912e-07, -4.74919101069986e-07, -3.22482643551556e-07, -4.15405067114242e-07, -4.42335771020138e-07, 0, 0, 64031.2593794752, 0, 6.00977700131127e-08, 0, 3.12387669174591e-08, 25833.937499913, 25833.9375001209, 0, 1.23511039859808e-07, 0, -5.22523069759566e-07, 0, 3.58237788448811e-07, 1.3230556989958e-07, 5.00821662359494e-07, 1.03496352184348e-06, 0, -5.80093324784178e-08, -2.37846198761712e-07, 5.2064214420732e-07, 1.21968458775626e-06, 9.57072061613449e-07, 1.08444548150476e-06, 1.72134215291082e-06, 1.69319969295257e-06, 2.71139626492522e-06, 0, 0, -68090.8781293315, 0, -5.77852483234222e-08, 0, -7.59901464416806e-08, -38197.3218748386, -38197.3218750602, 0, -2.10759774004557e-07, 0, 1.93877924896008e-07, 2.99397895367706e-08, -6.65443345483385e-07, -1.24221840167328e-07, -4.45026910704918e-07, -1.64392066316489e-06, 0, 2.33275884123186e-07, 0, 4.6495113177023e-07, 2.17925403038001e-07, -1.52529885563215e-07, -4.9421096441441e-07, -8.07453777528045e-07, -1.50603193639053e-06, -9.46029224261508e-07, -3.03174096491862e-06, -3.33127484083791e-06, -1.21975003358456e-06, -2.16488049568862e-06, -5.04098837770067e-06, -2.33316826256809e-06, -6.88117582001526e-06, -6.08848094689509e-06, 0, 0, 39502.1993326423, 0, 2.8608587727186e-08, 0, 4.5277645411999e-08, 29893.5562498629, 29893.5562500839, 0, 1.82181459263462e-07, 0, -2.82956532586509e-07, -3.78042969330667e-08, 3.45660715357155e-07, 7.79452007180433e-08, 5.80766056556302e-07, 9.3326301525662e-07, 0, -1.10276520425861e-07, 0, -9.52197782866834e-07, -1.9361325249433e-07, 5.49414778877146e-07, 3.39079183496821e-07, 4.41649393441133e-07, 7.2301225200803e-07, -1.73537920556909e-07, 2.52697694928784e-06, 2.166477846906e-06, 0, -4.7569207212973e-07, -2.72822323742359e-07, 1.91407017427145e-06, 2.16269918893486e-06, 1.72714376870194e-06, 2.47958665759158e-06, 3.12882470951444e-06, 5.10989260533937e-06, 2.71084388895193e-06, 3.27793441511048e-06, 9.04287222195177e-06, 4.57624011251325e-06, 4.21769865143893e-06, 2.98758006020835e-06, 0, 0, -9608.64308086322, 0, 0, 0, -1.50365629921747e-08, -9608.64308031146, -9608.64308034192, 0, -6.1237316529474e-08, 0, 1.22813555278957e-08, 2.11059955212171e-08, -6.4289859227806e-08, -1.22684840710634e-08, -1.76520960851029e-07, -2.71384559567538e-07, 0, -1.64903232157625e-08, 0, 4.38914280831941e-07, 2.95254648621873e-08, -4.00092033900602e-07, -5.53790678062055e-08, -6.81526658953614e-08, -2.74183161439853e-07, -2.58160584734046e-08, -8.19530588837448e-07, -1.10325243597228e-06, 0, 2.41667800035247e-07, 0, 6.44768891418629e-07, 2.4198539957165e-07, -2.68069322869463e-07, -2.57058477322175e-07, -2.93502062536528e-07, -3.35363935865434e-07, 3.01609843228592e-07, -1.02534739253495e-06, -1.78452770295815e-06, -7.01601615033289e-07, -6.19486711574725e-07, -3.41861862939183e-06, -2.07094222683472e-07, -1.12590275546814e-06, -9.91663994307336e-07, -1.99479222685333e-06, -1.86753344661205e-06, -2.26031806967736e-06, -1.34603745609897e-06, -5.28646941845919e-06, -5.48152816800068e-06, -4.14814958406183e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -10.1249999999603, 0, 0, 0, 0, 236.274107143092, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2136.88124999791, 0, 0, 1.17842442291583e-08, 0, 0, -2.33451564954357e-08, 0, 0, 0, 0, 0, 0, 0, 0, 9913.26093745601, 0, 0, -6.61425732335921e-08, 0, 0, 1.66042085456009e-07, 0, 0, 8.68365064279799e-08, 0, 0, 0, 0, 2.59246336237743e-08, 0, 0, 0, 0, 0, 0, -25833.9374998524, 0, 0, 1.73519822647621e-07, 0, 0, -3.25833309333667e-07, 0, 0, -1.01138831388084e-08, 0, 0, 0, -3.96830366058626e-08, 0, -7.56128833988274e-07, 0, -4.4072047435851e-08, 0, -4.03197724564841e-08, -3.05058253647498e-08, -1.39812905987056e-07, -3.57351440765143e-08, -7.60878863707625e-08, 0, 0, 0, 0, 38197.3218747781, 0, 0, -2.17364719670575e-07, 0, 0, 2.63075768169863e-07, 0, 0, -3.00662964362621e-07, 0, 2.97791724374511e-08, 0, 2.35912371067311e-08, 0, 1.4313415125297e-06, 0, 0, 1.14368836666829e-06, 0, 7.76804348316398e-08, 1.91767121471689e-07, 3.84533741212326e-08, -5.75786241844022e-08, 1.47519512480979e-07, 1.19238544621849e-07, 0, 1.32451994940437e-07, 9.58236155787109e-08, 0, 1.70207676895697e-07, 1.43323336773431e-08, 0, 0, 0, 0, -29893.556249832, 0, 0, 1.26212574826576e-07, 0, 0, -6.40857966693886e-08, 0, 0, 4.29914867184922e-07, 0, 0, 0, 0, 0, -9.55064945425906e-07, 0, 0, -6.46299120044418e-07, 0, 0, -1.51576884148521e-07, -7.9348810557184e-08, -3.11192163042427e-08, -1.01247410120389e-07, -6.2361111905147e-08, 0, -1.74292919971904e-06, 0, 1.3564780263007e-07, 7.82457246947165e-08, 8.4237244173283e-08, 7.55329643686598e-08, -1.72186847478658e-07, -2.13581372377236e-08, -2.0092828998958e-07, 5.68766487703519e-08, 0, -4.66419619899576e-07, -2.122886675083e-07, -1.65265995395776e-07, 0, 0, 0, 0, 9608.64308030529, 0, 0, -2.8591830418209e-08, 0, 0, 0, 0, 0, -1.7046083315207e-07, 0, 0, 0, 0, 0, 2.00984749549103e-07, 0, 0, 5.06644505435581e-08, 0, 1.90107038435888e-08, 1.66932855580996e-08, 1.79772213998718e-08, 0, 0, 0, 0, 7.03858268757496e-07, 0, 0, 6.4975050969171e-07, 0, 2.82682703401548e-08, 2.96004492858391e-08, 0, 6.80159358108938e-08, 8.00808193941251e-08, 5.03426141794333e-08, 1.19395407888151e-07, 0, -2.03019750677673e-08, 3.41337829331354e-08, 4.72465142056279e-08, 0, 0, 1.48140091340755e-07, 6.86072658372714e-08, 5.29168314744872e-08, -1.09382721144202e-07, 2.02554454640658e-07, 1.08874227187654e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -5.78571428585433, 0, 0, 0, 0, 127.574999999056, 0, 0, 52.071428577176, 0, 0, 0, 0, 0, 0, -1057.04999998201, 0, 0, -5.39461854831636e-08, 0, 0, -1148.17500001995, 0, 0, 0, 0, 0, 0, 0, 0, 4305.65624995416, 0, 0, 2.14661684740805e-07, 0, 0, 9513.44999992415, 0, 0, 2.23820214609242e-07, 0, 4.6664157063106e-08, 0, 1.60663701740451e-08, -5.84557842627663e-08, 8.16375556002682e-08, 0, 0, 0, 0, 0, -9226.4062499322, 0, 0, -4.97805674624643e-07, 0, 0, -38750.9062495918, 0, 0, -8.39322763442994e-07, 0, -1.23501531909868e-07, 0, 4.96763063602857e-08, 0, -1.47709206041863e-07, 0, -3.4777195418677e-08, -3.84626044607127e-07, 5.7253883128372e-08, 6.77753602255275e-08, 2.14530491208755e-07, 2.14049920411805e-08, 7.833898271148e-08, 0, 0, 0, 0, 9964.51874994464, 0, 0, 6.26643007204526e-07, 0, 0, 83037.6562492532, 0, 0, 1.30549343635087e-06, 0, 5.56405682204027e-08, 0, 1.47337123908612e-08, 0, -6.81759938340813e-07, 0, 0, 1.46068024752992e-06, 0, 3.10645083279752e-08, -9.08527721048478e-08, -3.22750130537546e-08, 2.96365623142429e-07, -1.26657099994631e-07, -1.69752730217851e-07, 2.42773709318679e-07, 2.47973471485793e-07, -2.45218574472678e-07, 8.94442794695901e-07, -1.42839255554622e-07, 4.91823346840114e-07, 0, 0, 0, 0, -4270.50803570936, 0, 0, -3.89910548242524e-07, 0, 0, -89680.6687494831, 0, 0, -1.01579671264758e-06, 0, -1.05616631539682e-07, 0, -4.46586034284091e-08, 0, 1.04931910606086e-06, 0, 0, -1.57997877013073e-06, 0, -2.42456811798894e-07, 5.57774469341687e-08, 2.51279811058288e-07, 1.26529340427801e-07, 1.08597134793974e-07, 1.6109404776445e-07, 0, -6.23440473819457e-07, 0, -8.55638348595091e-07, -1.10609129889764e-06, -5.21175312504931e-07, -4.21109564025929e-07, -4.69800209716138e-08, -1.09141922648523e-07, -1.66537841953134e-07, -6.60574090913272e-07, -2.19674709701375e-07, 9.90558994418134e-07, -3.48529425648959e-07, 2.04261011837947e-08, 0, 0, 0, 0, 1.08032028796959e-08, 0, 0, 1.05615463263397e-07, 0, 0, 38434.5723213147, 0, 0, 2.85475757114075e-07, 0, 1.57992611122031e-08, 0, 1.72299761683624e-08, 0, -4.05236150182285e-07, 0, 0, 6.49445783891352e-07, 0, 0, 1.0304521565123e-07, -5.17624553904872e-08, 0, 1.10916778932731e-07, 4.43373442503594e-08, 0, -2.00942501077396e-07, 0, 0, 7.34197397243958e-07, 0, 1.65527475502969e-08, 2.44053373252264e-07, 8.88666414779485e-08, -3.03952875226024e-07, -1.15011596949024e-07, -1.76508084818008e-07, -1.54800834057314e-07, 5.73015779331611e-08, 5.99274681646977e-08, -3.45447769629571e-08, 1.57835367819019e-07, 1.2414628056851e-07, 5.44567715301948e-07, 7.2921241310523e-08, 0, 5.86816905860943e-08, 1.22261203615374e-06, -1.3014761951698e-07, 3.26049442023788e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -4.50000000016295, 0, 0, 0, 0, 92.4750000012365, 0, 0, 60.7499999987773, 0, 0, 0, 0, 0, 0, -683.43750003244, 0, 0, -182.249999994826, 0, 0, -1248.41249994925, 0, 0, 0, 0, 0, 0, 0, 0, 2323.68750011664, 0, 0, 4.04578245286774e-08, 0, 0, 9226.40624993783, 0, 0, 3745.23749963735, 0, 0, 0, 0, 2.02899686806636e-07, -3.40922898745286e-08, 3.50424052373002e-08, 0, 0, 0, 0, -3690.56250030245, 0, 0, -9.81448117338541e-08, 0, 0, -31369.7812501057, 0, 0, 9.15698633774114e-07, 0, 2.12538159968181e-08, 0, -2.23095352188906e-07, 0, -27679.218749164, 0, -1.96478651516638e-07, 1.90071445030054e-07, -2.26032532016494e-07, -1.85850467368257e-07, -8.56958680404187e-07, -1.77246136749097e-07, -5.8958538725474e-07, 0, 0, 0, 0, 2214.33750046367, 0, 0, 1.55992000758219e-07, 0, 0, 49822.5937506476, 0, 0, -9.60440473596284e-07, 0, 2.20805126220836e-07, 0, 6.22374574928626e-08, 0, 94109.3437492291, 0, 0, -2.72150500336079e-06, 0, 4.55802445070462e-07, 1.23283078622815e-06, 2.0056449040503e-07, -6.23434747226207e-07, 7.32579512869791e-07, 5.26396194914915e-07, -1.53184438753108e-07, 7.69658242975581e-07, 6.94278542406053e-07, -6.29752143346665e-07, 9.19115372457776e-07, -4.74839533895072e-07, 0, 0, 0, 0, -3.16658244497537e-07, 0, 0, -1.66959963926612e-07, 0, 0, -29893.5562507121, 0, 0, 5.50725935406166e-07, 0, 0, -2.87257022201763e-08, 0, 0, -149467.781249517, 0, 0, 1.81077514534049e-06, 0, 8.41789265013806e-08, -1.01733191471563e-06, -6.2175584462491e-07, -3.10481314147029e-07, -5.68189455982125e-07, -3.3362451910196e-07, 0, 2.6490367864869e-06, 0, 1.47385784073397e-06, 9.39466302313784e-07, 8.90427260670333e-07, 8.46071794833559e-07, -1.27284273786277e-06, 1.68026859193387e-07, -1.14459153254503e-06, 1.06923251893613e-06, -1.98600526532415e-07, -3.44943077708574e-06, -6.37006228534567e-07, -5.00503179041152e-07, 0, 0, 0, 0, 7.47469804686007e-08, 0, 0, 5.05026769035617e-08, 0, 0, 2.47631628312974e-07, 0, 0, -7.05460002196561e-08, 0, 2.6272190483388e-08, 1.74643939207566e-08, -1.47217917930682e-08, 0, 89680.6687498734, 0, 0, -5.99602373248124e-07, 0, 1.45645687943347e-07, 5.57463502670331e-08, 1.26224119753383e-07, 1.00048949263222e-08, -1.48594168382314e-07, -5.96788270943112e-08, 0, -7.90099975587364e-07, 0, 0, -1.2943031655899e-06, 0, 1.66762062115633e-07, 1.58185192311495e-07, -1.21004558351749e-07, 6.66232589195091e-07, 6.32255017975619e-07, 3.8290224430579e-07, 9.64751149154583e-07, -2.7660978700179e-08, 2.09883045404448e-08, 2.57170965334281e-07, 1.88855570166773e-07, -2.594610500228e-07, -6.34712662942054e-07, 7.25736974419491e-07, 4.03136413432729e-07, 5.80681578287557e-08, -1.78607192143408e-06, 7.54996257220742e-07, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -4.05000000053787, 0, 0, 0, 0, 75.9375000147544, 0, 0, 66.8250000027391, 0, 0, 0, 0, 0, 0, -478.406250111769, 0, 0, -328.049999988082, 0, 0, -1252.96875006376, 0, -1.661454797448e-08, 0, 0, 0, 0, 0, 0, 1230.18750050014, 0, 0, 492.074999903669, 0, 0, 7893.70312548425, 0, 0, 6150.93749990372, 0, 8.31304734848405e-08, 0, 6.94840439307728e-08, -2.05214698205137e-07, 2.04188150818796e-07, 1.90593146715524e-08, 0, 0, 0, 0, -1107.16875104052, 0, 0, 5.25058058687904e-08, 0, 0, -20298.0937520788, 0, 0, -9226.40624919648, 0, -2.5532055526971e-07, -2.20127685160873e-08, 2.17697568985511e-07, 0, -38750.9062496323, 0, -2.54245240484121e-08, -7.07603173968963e-07, 1.25479039500688e-07, 9.55601436141455e-08, 7.20883380386372e-07, 1.11095131736983e-07, 3.39712641563535e-07, 0, 0, 0, 0, 1.09298333387586e-06, 0, 0, 6.16622050842274e-08, 0, 0, 18268.2843785411, 0, 0, -1.39209845947003e-06, 0, -3.17814731454871e-08, 2.77216538427379e-08, 7.30859473773743e-08, 0, 99645.187500862, 0, 0, 58126.3593730202, 0, -1.525938099844e-07, -7.5428054400918e-07, -1.23134697974777e-07, 1.03240425062675e-06, -1.84084436375774e-07, -4.97585652036511e-07, 6.1336325851659e-07, 7.37884578604542e-08, -7.46380381243361e-07, 1.83049040266482e-06, -6.45704457424485e-07, 1.96598002439557e-06, 0, 0, 0, 0, -6.23090282714822e-07, 0, 0, -4.49697308303061e-08, 0, 0, -2.88719762081574e-06, 0, 0, 7.79271800522923e-07, 0, -1.94672825033256e-07, 3.14868121051028e-08, -3.76538840159423e-08, 0, -89680.6687529127, 0, 0, 3.57083855283416e-06, 0, -5.03912071455799e-07, 6.41848545301556e-07, 7.44158377982343e-07, 3.4172584795679e-07, 2.45440892620669e-07, 1.56551552624908e-07, 0, -149467.781248907, 0, -2.57381759730676e-06, -2.52354077110093e-06, -1.56370893657506e-06, -1.37622621640997e-06, 8.42096100455264e-07, -5.23972702547836e-07, 3.28624247199922e-07, -1.89327153302708e-06, -7.64115309459215e-07, 3.94485209657223e-06, -1.10890084269617e-06, -5.82258402298052e-07, 0, 0, 0, 0, 1.59222017652999e-07, 0, 0, 1.57693211819355e-08, 0, 0, 9.13943162212395e-07, 0, 0, -1.94048618930025e-07, 0, 1.19676926601448e-08, -2.11930122655797e-08, 4.31534408154407e-08, 0, 1.66850760530471e-06, 0, 0, -1.21468037048128e-06, 0, -9.64086168197943e-08, 1.70055263479284e-07, -1.47174531173043e-07, 0, 3.68763526167429e-07, 1.11399909508811e-07, 0, 134521.003125123, 0, 0, -1.41254389797262e-06, 0, -6.14591382170807e-08, 2.35906637285929e-07, 2.39893000388423e-07, -8.11085779244709e-07, -4.79600393825473e-07, -3.72854931549478e-07, -7.71062254583189e-07, 1.59308685088479e-07, 4.71307964575799e-07, -2.27180329088342e-07, 6.04885980231438e-07, 4.56033164786036e-07, 1.60752520699625e-06, 1.19956093143957e-07, -2.11089416695853e-07, 2.98023683439147e-07, 3.02619940066831e-06, -8.34373557867924e-07, 8.06539298893736e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -4.05000000122332, 0, 0, 0, 0, 66.8250000199605, 0, 0, 75.9375000091193, 0, 0, 0, 0, 0, 0, -328.05000015495, 0, 0, -478.406250036382, 0, 0, -1252.96875014563, 0, 0, 0, 0, 0, 0, 0, 0, 492.075000512449, 0, 0, 1230.18750012361, 0, 0, 6150.93750104505, 0, 0, 7893.70312528409, 0, 0, 1.57758255407441e-08, -1.2304905890146e-08, 2.41452076620027e-07, -7.65751298765598e-08, 3.22032950904445e-08, 0, 0, 0, 0, -9.90117894172191e-07, 0, 0, -1107.16875010835, 0, 0, -9226.40625307732, 0, 0, -20298.0937502513, 0, 9.50591233396965e-08, -1.82400229379411e-08, -3.67862348114241e-07, 0, -38750.9062522556, 0, -1.54181222768457e-07, 1.50686695591198e-07, -2.09098139045408e-07, -1.63443884651079e-07, -1.03818195035197e-06, -3.28966219972502e-07, -6.24586173923418e-07, 0, 0, 0, 0, 1.14430210603157e-06, 0, 0, 6.3804436744009e-08, 0, 0, 4.83936339531568e-06, 0, 0, 18268.2843752042, 0, 1.97043248114577e-07, 0, 9.53262146200335e-08, 0, 58126.3593809995, 0, 0, 99645.1875008434, 0, 4.44361796213707e-07, 1.29972729354781e-06, 2.86515435866873e-07, -8.22187796202994e-07, 5.71415857387537e-07, 7.62794714652988e-07, -2.46251099342346e-07, 8.59020585617811e-07, 1.00072835628621e-06, -5.06540666762949e-07, 1.14525333820723e-06, -1.28071594376256e-06, 0, 0, 0, 0, -6.64139278417575e-07, 0, 0, -4.83251789958765e-08, 0, 0, -3.61475474611884e-06, 0, 0, -6.34327178679831e-08, 0, 7.23428849856137e-08, -4.65425590005631e-08, -1.25288895952105e-07, 0, -5.97313059462493e-06, 0, 0, -89680.6687503711, 0, 2.03965234956531e-07, -1.09073893954669e-06, -7.27093823808191e-07, -3.73023723934941e-07, -6.19213754698841e-07, -2.86257882953205e-07, 0, -149467.78125241, 0, 2.08254874589789e-06, 1.1586586149237e-06, 1.2441217477121e-06, 1.05065998429371e-06, -1.78373637412674e-06, 2.0786665497027e-07, -1.37875311219847e-06, 1.18182700835552e-06, -9.35168957394972e-08, -4.84507929338719e-06, -1.28550151561628e-07, -5.033025143746e-08, 0, 0, 0, 0, 1.43451995611353e-07, 0, 0, 1.47608128552494e-08, 0, 0, 1.01265539768473e-06, 0, 0, 6.24367429130501e-08, 0, 1.76281631987426e-08, 2.18251658542562e-08, 0, 0, 2.1414231517458e-06, 0, 0, -1.18267574874645e-07, 0, 1.80561618689724e-07, 0, 1.4982341715244e-07, 3.29377995148333e-08, -2.09153993650753e-07, 0, 0, 1.93384244810048e-06, 0, 0, 134521.003125224, 0, 2.02392920933007e-07, 2.29984202469979e-07, -1.32762174826627e-07, 7.13530580146227e-07, 7.68766681574309e-07, 3.7948221530728e-07, 1.02716864537655e-06, 0, -2.6108136979766e-07, 3.53541118761049e-07, -3.62233981807128e-07, -2.92960329408323e-07, -1.06402365264703e-06, 3.9282957679858e-07, 5.84244869325486e-07, 4.37426527639494e-08, -1.77861475465532e-06, 1.43214115313571e-06, -2.82058555102468e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -4.5000000010345, 0, 0, 0, 0, 60.7500000179242, 0, 0, 92.4750000142817, 0, 0, 0, 0, 0, 0, -182.250000112012, 0, 0, -683.437500067679, 0, 0, -1248.41250016787, 0, 0, 0, 0, 0, 0, 0, 0, 4.02331650304408e-07, 0, 0, 2323.687500167, 0, 0, 3745.23750082333, 0, 0, 9226.40625067097, 0, 3.6238524712021e-08, 0, 4.92769013793388e-08, -1.04611153565138e-07, 1.25475362993526e-07, 2.08665252741503e-08, 0, 0, 0, 0, -7.70905615369576e-07, 0, 0, -3690.56250036048, 0, 0, -2.30639290540157e-06, 0, 0, -31369.781251331, 0, -1.31462389130833e-07, 0, 1.77868916212291e-07, 0, -27679.2187521822, 0, -3.01239261666591e-08, -2.44880630947362e-07, 3.5726991180461e-08, 0, 4.21270913588364e-07, 1.16207213013497e-07, 1.05465290566873e-07, 0, 0, 0, 0, 7.79434616636123e-07, 0, 0, 2214.33750039299, 0, 0, 3.21185253758954e-06, 0, 0, 49822.5937515726, 0, -2.97915874230355e-08, 1.81712949676452e-08, 1.88577094232217e-08, 0, 3.99493079388607e-06, 0, 0, 94109.343753409, 0, -8.48597425477075e-08, -4.15833828063719e-07, -1.18194774644953e-07, 5.90807644212097e-07, 1.38927718474164e-07, -2.92639574699993e-07, 3.1364976149924e-07, -1.27179521963678e-08, -4.4930422059023e-07, 6.47749313121675e-07, -3.08638888910219e-07, 1.54353929788178e-06, 0, 0, 0, 0, -4.27897796777734e-07, 0, 0, -2.07598269357852e-07, 0, 0, -2.25703748387915e-06, 0, 0, -29893.556251132, 0, -9.65051358945107e-08, 2.08890781633453e-08, 6.68205049825474e-08, 0, -4.17944428213756e-06, 0, 0, -149467.781252015, 0, -2.52683136284594e-07, 4.06631392201944e-07, 3.53599896978948e-07, 1.49325885798096e-07, 6.02833739578336e-08, 0, 0, -4.10030597223969e-06, 0, -1.46048847946881e-06, -1.04119788308966e-06, -9.1486586863196e-07, -7.70403686759e-07, 6.01018336615571e-07, -2.97200425468645e-07, 4.18703408434574e-07, -8.85133373452697e-07, -5.1363207833124e-07, 2.63050237304097e-06, -8.18520659920128e-07, -6.88719542280079e-07, 0, 0, 0, 0, 1.05518853563156e-07, 0, 0, 3.87967842318281e-08, 0, 0, 6.5104708069621e-07, 0, 0, 3.17136143906696e-07, 0, 0, -1.08231985495736e-08, 0, 0, 1.62839260246813e-06, 0, 0, 89680.6687506778, 0, -7.68467895412556e-08, 8.09583768232983e-08, -7.2095942280156e-08, -1.08800384440449e-08, 2.11901616828957e-07, 0, 0, 1.73964628298033e-06, 0, 0, 1.16225827172536e-06, 0, -7.68043148993122e-08, 0, 1.0395086851798e-07, -3.12693645276402e-07, -2.4879651764232e-07, -1.19029968511375e-07, -3.84895763803209e-07, 5.42815157523398e-08, 4.60846746463358e-07, -1.63821918574694e-07, 6.70063993388646e-07, 2.25274141689871e-07, 9.49679392600468e-07, 2.90427680894444e-07, -1.90693693743274e-07, 1.55337619903175e-07, 1.22732958067185e-06, -7.50029743114928e-07, 3.81067755283771e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -5.78571428566509, 0, 0, 0, 0, 52.0714285707959, 0, 0, 127.574999998014, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1057.04999998964, 0, 0, -1148.17499999282, 0, 0, 0, 0, 0, 0, 0, 0, -1.42417231952792e-08, 0, 0, 4305.65624996773, 0, 0, 0, 0, 0, 9513.44999994929, 0, 0, 0, 0, 5.267681301715e-08, -2.33845697511033e-08, 0, 0, 0, 0, 0, 3.74786021919804e-08, 0, 0, -9226.40624987924, 0, 0, 6.82184981659044e-08, 0, 0, -38750.9062498825, 0, 2.50326295656863e-08, -1.14321951220172e-08, -1.17130314735704e-07, 0, -7.97318210585138e-08, 0, -2.82709657450337e-08, -2.04429146726016e-08, -4.19590229175868e-08, -2.18399131324381e-08, -2.70970570198572e-07, -1.11935960218396e-07, -1.16708119136733e-07, 0, 0, 0, 0, -2.7785517198223e-08, 0, 0, 9964.51874982675, 0, 0, -1.27619923295122e-07, 0, 0, 83037.6562497828, 0, 3.77918575153988e-08, 0, 3.94958929083613e-08, 0, 3.42748208935183e-07, 0, 0, -2.86380534739773e-08, 0, 9.38364499224304e-08, 2.76933771270605e-07, 1.03883993071211e-07, -1.87131009409276e-07, 3.97599921307246e-08, 1.98762265414114e-07, -4.42880914282654e-08, 2.16689012328587e-07, 2.6693204568934e-07, 4.14149802543445e-08, 2.57390216712305e-07, -4.65889029467693e-07, 0, 0, 0, 0, 1.08914646235229e-08, 0, 0, -4270.50803559259, 0, 0, 1.21483127551871e-07, 0, 0, -89680.6687497414, 0, 2.15149454035718e-08, -1.30226574408464e-08, -6.76931371632872e-08, 0, -1.21684876088833e-07, 0, 0, -9.03249570415219e-08, 0, 5.00210256813949e-08, -2.60346847901489e-07, -1.48040654675417e-07, -7.32375757305952e-08, -1.23995645144005e-07, -4.82931394356579e-08, 0, -8.31641521209692e-08, 0, 5.05915838076326e-07, 1.12556079528819e-07, 3.10413666893037e-07, 2.32370310968208e-07, -4.7283929035492e-07, 1.86042615155671e-08, -4.19241465550311e-07, 1.74575852566282e-07, -3.62149866290834e-08, -1.34676718791643e-06, 1.3850676533818e-08, 1.01710861381692e-07, 0, 0, 0, 0, 0, 0, 0, -3.03752308462094e-08, 0, 0, -5.41870901547179e-08, 0, 0, 38434.5723213227, 0, 0, 0, 0, 0, -4.26894940923479e-08, 0, 0, -1.74180251973131e-08, 0, 4.55940159849142e-08, 0, 3.16376818824619e-08, 1.24156200528866e-08, -4.66199781186072e-08, 2.93887575384042e-08, 0, 1.37581541839136e-07, 0, 0, 9.69984925106993e-08, 0, 6.47113457332406e-08, 9.18516695953235e-08, -1.48279085773966e-08, 1.04200174972081e-07, 1.71236955289754e-07, 6.31588590684453e-08, 2.02738313780876e-07, 2.92101947547796e-08, -1.28170289060011e-07, 9.65181091278127e-08, -2.34624287320047e-07, -4.35213354278252e-08, -2.85341224981692e-07, -1.11033426723613e-08, 1.758895291124e-07, 5.26655522326712e-08, -1.90362005597811e-07, 4.73730295482138e-07, 4.69250454454864e-08}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -10.124999999839, 0, 0, 0, 0, 0, 0, 0, 236.274107139568, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2136.88124997257, 0, 0, 2.1065248956852e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9913.26093739175, 0, 0, -7.53669889301413e-08, 0, 0, -1.36208213177547e-07, 0, 0, 0, 0, -1.08169312654366e-08, 0, 0, 0, 0, 0, 0, -1.58269820407553e-08, 0, 0, -25833.9374997727, 0, 0, 9.59235685150525e-08, 0, 0, 3.68421347715828e-07, 0, 0, 0, 3.05124552047479e-08, 0, 4.79678643123087e-07, 0, 0, 0, 0, 0, 6.50942653850217e-08, 2.63600346920533e-08, 1.72348144439044e-08, 0, 0, 0, 0, 1.95929765364663e-08, 0, 0, 38197.321874711, 0, 0, -1.77554726562612e-08, 0, 0, -4.81799811530995e-07, 0, 0, 0, -1.23151081445786e-08, 0, -7.94647745843682e-07, 0, 0, -9.20885097350872e-07, 0, -1.85056315280122e-08, -5.49979490631677e-08, -2.75940289091688e-08, 4.45496746405265e-08, 1.6873436539933e-08, -4.53640594430781e-08, 1.01613985983985e-08, -4.61125056233411e-08, -5.91512259216064e-08, -3.02461316783614e-08, -5.29666969583808e-08, 1.41824556739624e-07, 0, 0, 0, 0, -1.02950955776818e-08, 0, 0, -29893.5562498009, 0, 0, -5.41056367979114e-08, 0, 0, 3.17943191489534e-07, 0, 0, 0, 2.13414204592946e-08, 0, 4.92179578294116e-07, 0, 0, 7.58093771414958e-07, 0, -1.16514401929724e-08, 5.99476274635338e-08, 3.01309531757924e-08, 1.33084692076278e-08, 2.90986283813908e-08, 0, 0, 1.09176203980171e-06, 0, -1.17970189694538e-07, 0, -7.43616390266759e-08, -5.39806103192863e-08, 1.14976404973486e-07, 0, 9.59038725629404e-08, -2.48751311719435e-08, 0, 3.11661895457236e-07, 0, -7.31045234744722e-08, 0, 0, 0, 0, 0, 0, 0, 9608.64308029906, 0, 0, 3.12934425232832e-08, 0, 0, -9.10879981850755e-08, 0, 0, 0, 0, 0, -8.82609867976692e-08, 0, 0, -2.04796797370994e-07, 0, -1.06923329596835e-08, 0, 0, 0, 0, -1.05544612948982e-08, 0, -4.47892801796999e-07, 0, 0, -4.87110867835332e-07, 0, -1.67766694708601e-08, -2.75041126060479e-08, 0, -1.34579015399572e-08, -2.59986567880819e-08, 0, -4.79228205261739e-08, 0, 4.27453541579017e-08, -2.25244621046416e-08, 7.9424355065251e-08, 0, 6.97990289026015e-08, 2.01789311118038e-08, -4.12296221910076e-08, -1.35473896702061e-08, 1.62945468169836e-08, -1.13219183506781e-07, 0}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -10.1249999996435, 0, 0, 0, 0, 0, 0, 0, 236.274107140431, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2136.88124998706, 0, 4.2170525198135e-08, 0, 0, 0, 0, 2.80122369662974e-08, 0, 0, 0, 0, 0, 0, 0, 9913.26093745015, 0, -1.97479041464893e-07, 0, 0, 0, 3.24570665344304e-08, 0, -7.94989222550219e-08, -1.74668351970856e-08, -1.17280602991354e-07, 3.83941470929084e-08, -3.22493455621859e-08, 2.52263590178017e-08, -1.00414072832615e-07, 0, 0, 0, 0, 0, 0, -25833.9374998015, 0, 4.25203658985923e-07, 0, 0, 0, -5.19999188708375e-08, 0, 5.68335458354831e-07, 9.72350738808463e-08, 3.59606999575294e-07, -8.00080412625003e-08, 0, 0, -2.59695935044796e-07, 1.86539809025055e-07, -3.27345889431197e-07, 2.19053950978565e-07, 8.05334431491562e-07, -4.78643704275042e-07, 4.95264548103356e-07, -2.27711015067367e-07, 0, 0, 0, 0, 0, 0, 38197.3218747563, 0, -4.59918852221195e-07, 0, 0, 0, 3.85167927552388e-07, 0, -1.5668342344422e-06, -1.6777773996008e-07, -5.24178856716801e-07, 5.90923930091439e-08, 0, 0, 0, 1.77846004587369e-06, 0, -4.45229944512395e-08, -5.02894823935381e-07, 3.59987029557157e-07, -5.4742286057543e-07, -1.58907853215837e-06, 1.10108235040434e-06, -4.54004198059536e-07, 1.22371284571154e-07, 1.50816081901254e-06, -6.33366769769878e-07, -7.22713078221937e-07, 1.04634935433309e-06, 1.03956055167535e-07, 0, 0, 0, 0, 0, 0, -29893.5562498093, 0, 2.46003765426811e-07, 0, 0, 0, 4.82813619084865e-08, 0, 1.41857445962802e-06, 9.60239481922055e-08, 2.58541417135703e-07, -1.24634329142271e-07, 0, 0, 0, -1.02896716902983e-06, 0, 7.49495376257596e-07, 2.79195229537097e-07, 1.62921899487426e-07, 6.4586949678569e-07, 1.45014360890499e-06, -9.60113910122218e-07, 2.0105212678946e-06, 0, 0, -1.9857114507471e-06, -5.67606204352322e-08, -1.70841172505755e-06, 7.33267752951281e-08, 1.08036934939936e-06, -2.39699103489851e-06, 1.6667234620125e-06, 3.95426014759426e-07, 5.91674326809582e-07, -5.61194200307202e-07, 1.2336423475403e-06, 8.77949548688018e-07, 3.60734991139049e-06, 0, 0, 0, 0, 0, 0, 9608.64308029834, 0, -6.83682553943902e-08, 0, 0, 0, 1.60293718298732e-08, 0, -4.16849772831707e-07, -3.26938928798623e-08, -7.09592144054281e-08, 0, 0, 0, 0, 2.54185440457705e-07, 0, -3.74355681398646e-07, -1.40732798232683e-07, -8.05087621989333e-08, -3.05335621963958e-07, -6.13281640509745e-07, 3.62164735553791e-07, -1.28082531110273e-07, 0, 0, 0, 8.16204885450479e-07, 0, 5.52583756398871e-07, -5.07192426157451e-08, 2.42700639482331e-07, -2.52155353810888e-08, -1.01518472757367e-06, 8.58441409244507e-07, -1.00955030332901e-06, -4.12365384291097e-07, -9.01928594263144e-07, 5.20538711860626e-07, -5.24028786799807e-08, 1.18513429543679e-06, 1.00759230099177e-07, -1.04765576372309e-07, -2.06365711275557e-08, -9.91912287234148e-07, -2.2720304999906e-06, 1.49519271215163e-06, -4.26195588271366e-07, -1.65297130162148e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -5.78571428557156, 0, 0, 0, 0, 0, 0, 0, 127.574999995834, 0, 52.0714285715172, 0, 0, 0, 0, 0, 0, 0, -1057.04999996769, 0, 0, 0, 0, -1148.17499996221, -1.5351770078991e-08, -3.95235134397721e-08, 3.08862012173006e-08, 0, 0, 0, 0, 0, 0, 4305.65624982642, 0, 8.10032739962789e-08, 0, 0, 0, 9513.45000007658, 0, -2.86524464299068e-07, 1.26394721083332e-07, 1.69720652668065e-07, -2.94190006787213e-07, 1.22383191852713e-07, 2.41761817535132e-07, -1.08036552832686e-07, 0, 0, 0, 0, 0, 0, -9226.40624973904, 0, -1.37589182225333e-07, 0, 0, 0, -38750.9062501311, 0, 3.05595476944591e-07, -3.80896547920107e-07, -2.28084490938508e-07, 7.63853479551249e-07, 0, 0, 1.22665316354869e-07, -6.07730043995006e-07, -2.40351909086199e-07, -4.22694871373256e-07, -1.21629224650802e-06, -2.9749948787899e-07, -2.98984960519175e-07, -6.96179934986543e-07, 0, 0, 0, 0, 0, 0, 9964.51874954455, 0, 9.27480412616667e-08, 0, 0, 0, 83037.6562488211, 0, 6.66834586345608e-07, 4.4781276311321e-07, 1.92200485428193e-07, -1.1135910006052e-06, 0, 0, 0, -1.84736723783656e-06, 0, 3.87642972909896e-07, 4.80252501543446e-07, 1.24643620116456e-07, 9.51148700413106e-07, 2.58129793814116e-06, -8.60519520128184e-07, -2.4966882661491e-06, -6.44182745092913e-07, -4.00038488818964e-07, -2.04444408799822e-07, 4.96244579280415e-06, -1.67232897782745e-06, 3.88229405360437e-06, 0, 0, 0, 0, 0, 0, -4270.5080354418, 0, -5.08232123506291e-08, 0, 0, 0, -89680.6687497239, 0, -4.36252905188816e-07, -3.14859478369192e-07, 9.19916173074142e-08, 1.24931450137492e-06, 0, 0, 0, 9.15823671504617e-07, 0, -7.62649670748224e-07, -1.88740804603976e-07, -8.7533202088865e-07, -6.08083264677113e-07, -1.8630436598276e-06, 1.15521485519803e-06, -1.22332999867006e-06, 0, 0, 1.36855428529848e-06, 1.52739068671882e-06, 3.52905906330056e-07, -1.39356967168979e-06, -3.84676875281112e-06, 8.95191169028539e-07, -1.14219790776969e-06, -1.64492871535004e-06, 0, -1.94147542510695e-06, -6.56320331329212e-06, -1.00415103313374e-05, -1.3503559301224e-05, 0, 0, 0, 0, 0, 0, -5.69793817254975e-08, 0, 6.10519312690264e-08, 0, 0, 0, 38434.5723211345, 0, 1.79888862251928e-08, 1.04200356711915e-07, -9.3799742216596e-08, -2.79513433873413e-07, 0, 0, 0, -1.78978178427453e-07, 0, -3.77816144797556e-07, 1.128069505715e-07, 2.11361975189602e-07, 3.7593752688288e-07, 5.05755099583157e-07, -6.30130221991614e-07, -1.33819976079039e-06, 0, 0, 0, -1.23510386118888e-07, 0, -4.11425845072909e-07, -5.98237788942274e-07, 4.73451137667254e-07, -4.44230096836648e-07, 2.85420925158389e-06, -5.8936985827127e-07, -7.51174377381334e-07, 1.01958263261325e-06, 1.00711642835206e-06, -8.67430203979794e-07, -3.98132648588971e-07, -9.05645625977736e-07, -1.76743743655746e-06, 1.70701807154111e-06, 4.0509297232457e-06, 3.90823391210503e-06, 8.0898537774427e-06, 1.91155143317171e-06, 2.63154292629858e-06, 3.76641792014862e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -4.50000000020037, 0, 0, 0, 0, 0, 0, 0, 92.4750000252095, 0, 60.7499999865177, 0, 0, 0, 0, 0, 0, 0, -683.437500227289, 0, -182.249999830236, 0, 0, -1248.41250015119, -7.33510974071302e-08, -2.12394771498242e-07, 4.96356672114809e-08, 0, 0, 0, 0, 0, 0, 2323.68750078405, 0, -9.4186302267894e-07, 0, 0, 0, 9226.4062514801, 0, 3745.23750001254, 3.07476692616769e-07, 1.47688545300256e-06, -1.10641122505549e-06, 9.55003316519961e-07, -3.49956083525829e-08, 1.83620118919196e-06, 0, 0, 0, 0, 0, 0, -3690.56250132265, 0, 1.90268031408736e-06, 0, 0, 0, -31369.7812552482, 0, 3.38051628037302e-06, -4.45555830412554e-07, -4.22605983577435e-06, 5.16697868789879e-06, 0, 0, -27679.218753739, -1.74041253661778e-06, -1.23942825067329e-06, -4.33480393840387e-06, -1.13655461051114e-05, 2.71986928926153e-06, 1.15074161383704e-07, -7.39270060046749e-06, 0, 0, 0, 0, 0, 0, 2214.33750153356, 0, -2.0103046003701e-06, 0, 0, 0, 49822.59375911, 0, -8.51288844742713e-06, 9.25077851363601e-07, 5.33829334521601e-06, -9.01573772735016e-06, 0, 0, 0, 94109.3437638996, 0, -1.28095457326637e-06, 2.13217238672507e-06, 3.05373851416476e-06, 7.4513325050847e-06, 2.36184230181148e-05, -8.43783899918956e-06, -9.42449006396734e-06, 5.67222699290933e-06, 1.66527697367511e-06, 2.11538227443365e-05, 9.52877016008383e-06, 3.27578966202827e-05, 3.99572902414863e-06, 0, 0, 0, 0, 0, 0, -7.57985065383826e-07, 0, 1.04074603993686e-06, 0, 0, 0, -29893.5562550579, 0, 6.95203286433099e-06, -6.18942817352519e-07, -3.33617279853029e-06, 6.23090216103492e-06, 0, 0, 0, -149467.781263973, 0, 6.32692606102138e-06, -1.40952635557724e-06, -1.35042796770344e-06, -6.54071375885056e-06, -2.07411277094304e-05, 1.15923950472347e-05, 2.1119453725392e-05, 0, 0, -9.35311106496906e-06, -3.67396287139201e-06, -3.97102329270877e-06, -2.9130727438141e-06, -3.56163810270214e-05, 4.52876585313928e-06, -7.53969529539168e-06, -1.08561313912168e-05, -4.37046997060329e-05, 1.1118666447204e-05, -2.70769102722907e-05, -6.06985302385436e-06, -2.74192416270725e-05, 0, 0, 0, 0, 0, 0, 1.49333590203089e-07, 0, -2.96432523052e-07, 0, 0, 0, 1.58476226715178e-06, 0, -1.99737414268038e-06, 1.83879037142087e-07, 9.2076994545808e-07, -1.90156141881988e-06, 0, 0, 0, 89680.6687543727, 0, -2.57815619844701e-06, 4.65997510737437e-07, 4.43051624497775e-07, 1.88495213611459e-06, 6.56833580966415e-06, -4.7962866425368e-06, -7.46790800978705e-06, 0, 0, 0, 4.84187480268742e-06, 0, 9.47270488502229e-07, 7.97858223557097e-07, 7.71477802964948e-07, 3.01523350224686e-06, 1.61316645409459e-05, -3.17086725855394e-06, -3.3278802051967e-06, 1.8113420574459e-06, 1.60307716567349e-05, -9.99686270268997e-06, 1.05729875958952e-06, 1.49119683551312e-06, 1.77067402457867e-05, 7.4068479247215e-06, 4.3081927873759e-06, 1.42768485642336e-05, 2.31286495663413e-05, -1.72000167698028e-06, 1.23132529703715e-05, 2.34020400431917e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -4.05000000165006, 0, 0, 0, 0, 0, 0, 0, 75.9375000294661, 0, 66.8250000162441, 0, 0, 0, 0, 0, 0, 0, -478.406250184563, 0, -328.050000037625, 0, 0, -1252.96875036895, 7.01783948792489e-08, -1.2038017482473e-08, 2.01375236830532e-07, 0, 0, 0, 0, 0, 0, 1230.18750066644, 0, 492.075000071646, 0, 0, 0, 7893.70312724709, 0, 6150.93750081673, -2.9082005394936e-07, 3.40357469071118e-07, -1.22931459521817e-06, -2.68132890161955e-07, -8.88419414015219e-07, -7.28171128110342e-07, 0, 0, 0, 0, 0, 0, -1107.16875109745, 0, 1.39463963927183e-07, 0, 0, 0, -20298.0937551034, 0, -9226.40625140592, 1.70144933563678e-07, -1.05732384718673e-06, 3.07750049972203e-06, 0, 0, -38750.9062562203, 1.16673131203045e-06, 8.93834696028098e-07, 1.86740042056439e-07, 7.41416142084552e-07, 4.29069737827072e-06, 5.7719159788588e-06, 5.47682610577129e-06, 0, 0, 0, 0, 0, 0, 8.59955667546257e-07, 0, -2.42904699866826e-07, 0, 0, 0, 18268.2843803603, 0, 3.96828620975872e-07, -6.71510427165764e-07, 2.23096841566175e-06, -4.37339089663362e-06, 0, 0, 0, 99645.1875126853, 0, 58126.3593851782, -1.33598458709187e-06, -1.53374244976648e-06, 1.75673360492495e-06, 2.35059135810075e-06, -1.06650527830169e-05, -1.31333348236502e-05, -1.87329150413029e-06, -4.47366766499544e-06, -5.03828925998236e-07, -2.64639796326373e-06, -4.15041584799052e-06, -1.59062493335876e-05, 0, 0, 0, 0, 0, 0, -5.51044874151449e-07, 0, 2.2276278186567e-07, 0, 0, 0, -4.23238727304577e-06, 0, 6.77979492071423e-07, 2.61163214073061e-07, -2.02641690866574e-06, 3.4532217834929e-06, 0, 0, 0, -89680.6687609751, 0, -4.06464168446022e-06, 1.18230302549992e-06, 5.96306553063501e-07, -2.80441211423932e-06, -2.6224495255263e-06, 8.78106888670671e-06, 1.41681432849692e-05, 0, 0, -149467.781262207, 3.17762040394227e-06, 2.5867156997251e-06, -1.60347006634582e-07, -1.52234887985613e-06, 5.25326900211972e-06, 1.29899105191966e-05, 2.1364416952241e-06, 8.13552466617704e-06, 3.0022071341205e-05, 5.89977366132684e-06, 6.64833491455309e-06, 8.84690333619106e-06, 0, 0, 0, 0, 0, 0, 1.14339703103041e-07, 0, 0, 0, 0, 0, 8.79089306081675e-07, 0, -4.34127097724541e-07, -7.09905586702016e-08, 6.28398724738358e-07, -9.05631453557517e-07, 0, 0, 0, 4.19068881272662e-06, 0, -4.35027299277425e-08, -3.13787574754659e-07, -2.21464930367264e-07, 1.43974732461939e-06, 1.0950431220603e-06, -2.59939077492619e-06, -5.835200219568e-06, 0, 0, 0, 134521.003132495, 0, 4.4493215105346e-06, -1.39065723035413e-06, 8.79444140992304e-07, -5.45056319357096e-07, 2.71811968312638e-06, -3.8898792365923e-06, -6.77824491028063e-06, 4.57964546817631e-07, -5.44506830288909e-07, -1.58160479870868e-05, -1.83212573195855e-06, -1.46812304910588e-06, -8.66393518185754e-07, 2.25452124814982e-06, -3.5157371819095e-07, 1.69439564501752e-06, 1.00749896867601e-06, -7.97812720778177e-06, -2.08599447782253e-05, -3.96359291953486e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -4.05000000212759, 0, 0, 0, 0, 0, 0, 0, 66.8250000298126, 0, 75.9375000332732, 0, 0, 0, 0, 0, 0, 0, -328.050000195289, 0, -478.406250187453, 0, 0, -1252.96875032271, 6.76439994702239e-08, 6.98917188821423e-08, -6.67307381153167e-08, 0, 0, 0, 0, 0, 0, 492.075000590037, 0, 1230.18750059607, 0, 0, 0, 6150.93750191203, 0, 7893.70312653157, -1.22932224766042e-07, -3.66863817362048e-07, 5.9268462597159e-07, -8.54180721245454e-07, 1.12349567100347e-08, -6.00319360618745e-07, 0, 0, 0, 0, 0, 0, -1.09301220028012e-06, 0, -1107.16875111411, 0, 0, 0, -9226.40625496748, 0, -20298.0937531836, 3.07919064604701e-07, 6.37178850191523e-07, -2.4591777759426e-06, 0, 0, -38750.9062570715, 1.64004384615614e-06, 8.29669437581798e-07, 2.47815492103359e-06, 1.11542127226305e-06, -1.44373713079912e-06, -9.03497946695566e-07, 1.42971857464194e-06, 0, 0, 0, 0, 0, 0, 1.06247914130074e-06, 0, 1.32024702262253e-06, 0, 0, 0, 6.08558379359065e-06, 0, 18268.2843790554, -2.00952727702248e-07, -6.20823721086288e-07, 4.65313035643739e-06, 0, 0, 0, 58126.3593867375, 0, 99645.1875081536, -2.82722997313838e-06, 2.36994023978112e-07, -3.87791768614906e-06, -3.96507630717184e-06, 4.01480518092231e-06, -6.65415178821257e-07, -5.23732205300882e-06, 5.41313408998593e-07, -4.84286832880748e-06, -2.90425615790578e-06, -9.09016632333812e-06, -5.45548196677613e-06, 0, 0, 0, 0, 0, 0, -5.64716253763126e-07, 0, -7.38752309237416e-07, 0, 0, 0, -3.73307419317376e-06, 0, -2.25619714700746e-06, 2.76063983678248e-07, 2.44914711229829e-07, -3.96632237777893e-06, 0, 0, 0, -1.1014832992708e-05, 0, -89680.6687574164, 1.11365621772793e-06, 4.10763284997528e-07, 3.00095219078296e-06, 2.49792051466257e-06, -5.02270759985792e-06, -3.79446064980728e-06, 0, 0, -149467.781263334, 6.12043241156606e-06, -1.68723002302699e-06, 5.97233728731545e-06, -2.4961875239592e-06, -3.40104279671206e-06, -1.14822028131305e-06, 1.82412875811344e-06, 2.90030213412767e-06, 8.17135530310216e-06, -1.28714737993974e-06, 1.93646959988383e-06, 2.36701040216336e-06, 0, 0, 0, 0, 0, 0, 1.59524281180012e-07, 0, 2.15252985279311e-07, 0, 0, 0, 8.84259125981015e-07, 0, 6.56888754454944e-07, -7.06232171120135e-08, -2.84072415139594e-08, 1.40458352624976e-06, 0, 0, 0, 3.64088047980264e-06, 0, 1.55836839304112e-06, -5.19365372032663e-07, -2.97222247028194e-07, -6.92878145651964e-07, -1.36139744059194e-06, 2.28895646231815e-06, 1.56981307524628e-06, 0, 0, 0, 5.52398049925993e-06, 0, 134521.003128549, -1.86582745383647e-06, -3.34431142574394e-08, -2.94924996734877e-06, 9.76537268067019e-07, 2.33702964320815e-06, -6.15940306142352e-07, -5.01434300956905e-07, -1.79360323779998e-06, -4.37427454278066e-06, -3.44288448918122e-06, 1.14093339750066e-06, -4.66837079923873e-07, 9.86936516389517e-07, -2.94079831593319e-08, -4.83760900626833e-07, -1.18914293150936e-06, -6.49419867955898e-06, -4.42229740934367e-06, -1.62758737407669e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -4.50000000154034, 0, 0, 0, 0, 0, 0, 0, 60.7500000200393, 0, 92.4750000339102, 0, 0, 0, 0, 0, 0, 0, -182.250000076265, 0, -683.437500234248, 0, 0, -1248.41250037493, -4.56265568989007e-08, 1.11079235314445e-07, -1.98488029425781e-07, 0, 0, 0, 0, 0, 0, 1.00777614077085e-07, 0, 2323.68750067946, 0, 0, 0, 3745.23750111423, 0, 9226.40625231314, 4.39495233087621e-08, -7.33566239672702e-07, 1.3801812774064e-06, 1.55819707165794e-07, 8.15150925758007e-07, 1.11075548506901e-06, 0, 0, 0, 0, 0, 0, -1.78399856874443e-07, 0, -3690.56250119935, 0, 0, 0, -2.24802720803382e-06, 0, -31369.7812549155, 1.89482475320121e-07, 2.40381899437736e-06, -3.39653116017423e-06, 0, 0, -27679.2187553571, -3.00149616026168e-07, -2.79960372589046e-06, 1.6040992576713e-06, -9.6942956549558e-07, -3.28180913221837e-06, -5.64071920153116e-06, -5.92127777018781e-06, 0, 0, 0, 0, 0, 0, 3.3749526732755e-07, 0, 2214.3375010382, 0, 0, 0, 3.87244196477954e-06, 0, 49822.5937563163, -2.00533054082371e-07, -4.18910510004198e-06, 5.16360818928459e-06, 0, 0, 0, 1.1388328863528e-05, 0, 94109.3437618678, 6.90352841574584e-07, 1.3355329770029e-06, -5.98471999942985e-06, -1.30268645713821e-06, 7.16895028026648e-06, 1.52535402248644e-05, -1.83674175147389e-06, 5.97870427541048e-06, 1.49221931990237e-05, 1.18939474613523e-05, 1.47766320104794e-05, 1.63747495794392e-05, 0, 0, 0, 0, 0, 0, -1.14781152503168e-07, 0, -7.504591912533e-07, 0, 0, 0, -1.72449081076536e-06, 0, -29893.5562544991, 1.39738234141287e-07, 2.88018225645349e-06, -3.95908952269166e-06, 0, 0, 0, -8.68728106848735e-06, 0, -149467.781258963, 6.63832380722809e-07, -6.23805793844152e-07, 7.27957910031467e-06, 6.32925630158263e-06, -6.12546901120384e-06, -1.17868671397242e-05, 0, 0, -1.17685706318068e-05, -1.34455327103708e-06, -4.84720332118118e-06, 5.60938590437715e-06, -5.45842276297555e-06, -5.02941449890559e-06, -2.01299125338856e-05, -1.28789237168342e-05, -2.23729218524439e-06, -2.39015740321134e-05, -1.8421495974121e-05, -1.72661863303896e-05, -1.12739405551525e-05, 0, 0, 0, 0, 0, 0, 1.00674620540498e-08, 0, 7.60297923385755e-08, 0, 0, 0, 5.18719795265316e-07, 0, 1.15936944898968e-06, -4.89147686330587e-08, -7.65645733040418e-07, 1.12070619108575e-06, 0, 0, 0, 2.15756096985727e-06, 0, 89680.6687537773, -5.77468972887131e-08, -3.71674642368913e-08, -2.8243458402046e-06, -2.51075086653221e-06, 2.06247854175914e-06, 5.10085944185813e-06, 0, 0, 0, 4.07951530460931e-06, 0, 5.58530702485054e-06, -1.89718173913547e-07, 6.82977843442097e-07, -2.88903381819313e-06, -1.09877602173493e-06, 2.32996301683148e-06, 9.60662463445699e-06, 3.08306530899997e-06, -3.48748601060819e-06, 1.05132342533384e-05, -1.71714829403798e-06, 2.52066741571711e-06, 7.8694866651945e-06, 7.40972700510472e-06, 9.97733959134219e-06, 4.86845468437657e-06, 4.5915691210286e-06, 1.41669268210427e-05, 1.83033153616698e-05, 2.10780105036364e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -5.78571428613631, 0, 0, 0, 0, 0, 0, 0, 52.0714285618092, 0, 127.575000017178, 0, 0, 0, 0, 0, 0, 0, 7.97740930156178e-08, 0, -1057.05000014195, 0, 0, -1148.17500004047, 1.22989722473514e-08, 1.36207540630199e-08, -1.38485328911619e-07, 0, 0, 0, 0, 0, 0, -2.41520408577816e-07, 0, 4305.65625058315, 0, 0, 0, 1.37635143031426e-07, 0, 9513.45000029558, 4.07578819027387e-08, -3.7564325282891e-07, 9.44636569932836e-07, -5.74682318911821e-07, 5.7625757600955e-07, 3.18117266169123e-07, 0, 0, 0, 0, 0, 0, 6.6428182119814e-07, 0, -9226.40625124793, 0, 0, 0, -1.54420952988509e-07, 0, -38750.90625234, -2.15797143696235e-07, 1.10533026382266e-06, -3.33193383168714e-06, 0, 0, -1.50522650526418e-06, 1.26411940176988e-06, -1.05213767959888e-06, 2.16407731283496e-06, 2.77673912579758e-07, -2.56215442965732e-06, -2.7261499424682e-06, 1.23751742924915e-06, 0, 0, 0, 0, 0, 0, -9.14525203600996e-07, 0, 9964.51875152141, 0, 0, 0, -1.42403553202239e-06, 0, 83037.6562537943, 9.93919964362088e-08, -1.03362521756372e-06, 4.94111024800767e-06, 0, 0, 0, -3.88430832961019e-07, 0, 3.00990197464723e-06, -1.85531789580703e-06, 3.68889961598873e-06, -3.42335098905313e-06, -8.32438067239325e-07, 5.32746703388559e-06, 6.41729556666349e-06, -4.4309913869222e-06, 5.7073728851395e-06, -2.23132969834062e-06, 3.07511755825999e-06, -8.58056115516813e-06, 2.58420889635231e-06, 0, 0, 0, 0, 0, 0, 5.21544677350962e-07, 0, -4270.50803641126, 0, 0, 0, 6.66149422913797e-07, 0, -89680.6687532032, 0, 1.18026243406116e-06, -4.10621650416967e-06, 0, 0, 0, 2.26288261788591e-07, 0, -3.83571341068424e-06, 1.21436286745944e-06, -2.89636193903206e-06, 1.69464294699065e-06, -1.13458635776799e-06, -5.95539572657511e-06, -9.85508591331972e-06, 0, 0, -1.22169551518158e-06, 7.06815541035286e-06, -5.07783417347739e-06, 5.76097335751567e-06, -2.89313920798628e-07, -6.27969958012629e-06, -5.73610739643343e-07, -2.03464310710319e-06, 3.94436635342752e-06, 7.55429185368039e-07, 3.76060947874308e-06, -1.05390345782572e-05, -9.48681809613733e-06, 0, 0, 0, 0, 0, 0, -1.31917437745987e-07, 0, 2.36047042599451e-07, 0, 0, 0, -3.49880565483232e-07, 0, 38434.5723224802, 0, -3.69161813688213e-07, 1.24314685594269e-06, 0, 0, 0, 2.09068682356343e-07, 0, 1.20362588707652e-06, -4.4113542273732e-07, 9.80564570138092e-07, -4.52300757198727e-07, 1.32008297251288e-07, 2.55339199690092e-06, 3.44069291018235e-06, 0, 0, 0, 9.58923862036588e-07, 0, 4.20455561587564e-07, -2.11297500805883e-06, 3.01079324592538e-06, -1.91030355243835e-06, 2.17030188000088e-06, 3.04492099009265e-06, 2.30485876786587e-06, 1.53527859896866e-06, -4.83513797810918e-07, 1.16158342212077e-06, -2.79622832260194e-06, 2.97026452839775e-06, 1.15899317786097e-06, 3.93085212058529e-08, 4.7197544912482e-06, -2.70990212469422e-06, 2.38728587924021e-06, -3.75516260049135e-07, -4.69480923307329e-06, 5.53808433424202e-07}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -10.1249999999029, 0, 0, 0, 0, 0, 0, 0, 0, 0, 236.274107142133, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2136.88125000019, 0, 0, 0, -1.00497659024258e-08, 0, -2.38347316470704e-08, 0, 0, 0, 0, 0, 0, -5.36119445416912e-08, 0, 9913.2609375073, 0, 0, 0, -1.96628218334055e-07, 0, 1.05421495443197e-07, -2.74340193688606e-08, 7.99513856981349e-08, 1.37604879638924e-07, 1.46830797028454e-07, 4.44444744919615e-08, 7.85385105127521e-08, 0, 0, 0, 0, 0, 0, 1.3032021049698e-08, 0, -25833.9374999806, 0, 0, 0, 7.4195767812725e-07, 0, 1.62748189388366e-08, 9.48989240715425e-08, -2.8261907481347e-07, -3.16450762047919e-07, 0, 0, 6.59515509368018e-07, -8.31228633068342e-08, -3.51415625456371e-07, -5.22276816493738e-07, -2.07811733816507e-07, -5.2142848622612e-07, -4.54551795461439e-07, -1.41202342885502e-06, 0, 0, 0, 0, 0, 0, 3.7797874661082e-08, 0, 38197.3218749326, 0, 0, 0, -4.2239317155537e-07, 0, 2.13583323567452e-07, -9.66038281654489e-08, 2.15695802315216e-07, 7.52515727166076e-07, 0, 0, 0, -1.02632299345979e-06, 0, -5.50429808621479e-07, -6.01674808797855e-08, -2.80124506703831e-07, 9.60741888405598e-07, -2.4109668397472e-07, 9.05988445165139e-07, 1.26775981691111e-06, 4.45743199147994e-07, -3.44944457311358e-08, 2.85465215616549e-06, 7.82619272120667e-08, 4.64899004970227e-06, 2.57842521907902e-06, 0, 0, 0, 0, 0, 0, -2.69641532661362e-08, 0, -29893.556250022, 0, 0, 0, 4.11032355051174e-07, 0, -6.55218208584687e-08, 5.46928223702559e-08, -4.55480510960223e-07, -3.94326646490903e-07, 0, 0, 0, 1.33410084073044e-06, 0, 1.50657399071673e-08, 5.71555918214676e-08, 2.66857913837823e-07, -3.68559869994465e-07, 9.28684386339312e-07, -8.62634020642531e-07, -4.7848949947624e-07, 0, 0, 8.88892291335797e-07, -1.1551277593039e-06, -2.52718047454142e-07, -1.01180990436147e-06, -5.16090757331644e-07, -7.79521243561087e-07, -2.64323551825932e-06, -8.85610424253549e-07, -1.33192194630375e-06, -5.34655517656201e-06, -2.3263855578934e-06, -3.89234567223396e-07, 9.04555043200559e-07, 0, 0, 0, 0, 0, 0, 1.12899642334958e-08, 0, 9608.64308032952, 0, 0, 0, -4.22252295329074e-08, 0, 0, -1.68245177893065e-08, 1.47595949936019e-07, 1.33446347108183e-07, 0, 0, 0, -5.43665590850033e-07, 0, 2.24820701447778e-07, -3.88333644908663e-08, -1.43880479613966e-07, 1.75586988215598e-07, -1.81202754605499e-07, 2.85493103034204e-07, 5.87992899617415e-07, 0, 0, 0, -8.50993893162175e-07, 0, 2.29438546689272e-08, 6.40256162614739e-08, -3.07501684925063e-07, 1.23785623174375e-07, -9.16659801998109e-07, 7.084417614699e-08, 8.13678558902756e-07, 1.20703081376828e-07, 0, 1.90512046530104e-06, -3.26432283937253e-08, 1.53533915324021e-08, -4.92413395515724e-08, 8.76957607452529e-07, 2.76084865010247e-07, 1.25351210127565e-06, -2.11684805079367e-07, 3.2079248535747e-06, 3.53249733323274e-06, 2.0612586676291e-06}, 
                                                    {0, 0, -1.52077233990509e-08, 0, 0, 3.48561380816611e-07, 0, 2915.99999975815, 2.27166471058689e-07, 0, 0, -3.19585993862016e-06, 0, -34877.4428550317, 0, -34877.4428534799, -2.2598707338058e-06, 0, -3.71552909819389e-06, -34877.4428639511, 0, 1.52541140313459e-05, 0, 174358.574988879, 0, 174358.574975688, 1.355466866866e-05, 0, 2.59206345552016e-05, 348717.149981946, 0, 2.43099232881508e-05, 348717.15003999, 348717.150075619, 174358.575079336, 0, -4.21758566585606e-05, 0, -472966.087460525, 0, -472966.087410483, -4.82433356648205e-05, 0, -9.54319453035921e-05, -1418898.2624539, 0, -1418898.26242603, -9.76161563639084e-05, 0, -0.000119495204355082, -1418898.26261332, -1418898.26285709, -472966.087896877, -2837796.52540728, -1418898.26284267, -1418898.26314261, 0, 7.0086499154774e-05, 0, 752874.749917411, 0, 752874.749813574, 9.92111405014802e-05, 0, 0.000199414318052319, 3011498.99987465, 0, 3011498.99977671, 0.000248590539556949, 0, 0.000323045464789192, 3011499.00018485, 3011499.00087508, 752874.751076456, 4517248.50000092, 0, 0.000309641964061295, 9034497.00092684, 4517248.50069982, 9034497.00155934, 4517248.50223006, 3011499.00118957, 3011499.00218643, 9034497.0026891, 0, -6.90051446568997e-05, 0, -704159.324904002, 0, -704159.324784266, -0.000117612759632319, 0, -0.000235146360569768, -3520796.62476439, 0, -3520796.62459515, -0.00034956312118268, 0, -0.000494517680934787, -3520796.62517969, -3520796.62615999, -704159.32660806, -7041593.25010245, 0, -7041593.25008719, -0.000535904433855135, 0, -0.000493381882587109, -14083186.5010551, -7041593.25080545, -14083186.5028086, -7041593.25390259, -3520796.62722743, -3520796.6288713, -21124779.7526591, -7041593.251552, -7041593.25504098, -21124779.7545816, -21124779.757464, -14083186.5062734, 0, 3.7057852145628e-05, 0, 358722.674942546, 0, 358722.674871412, 7.44922528616053e-05, 0, 0.000144963513547747, 2152336.04978603, 0, 2152336.04963769, 0.000251258365900103, 0, 0.000397512434099582, 2152336.05009717, 2152336.05078947, 358722.676225583, 5380840.12502071, 0, 5380840.12501149, 0.000498797284184533, 0, 0.000422769578952464, 10761680.2506033, 5380840.12549694, 10761680.2523956, 5380840.12830622, 2152336.05216654, 2152336.0535619, 7174453.50039534, 0, 0.000478308881699969, 21523360.5019534, 7174453.50102513, 21523360.5032017, 7174453.50557913, 5380840.12667517, 5380840.13009106, 21523360.5035574, 21523360.5086853, 10761680.2566643, 32285040.7587255, 21523360.5066257, 21523360.5110221, 0, -8.35543659425601e-06, 0, -76869.1446290317, 0, -76869.144612201, -1.95123502223534e-05, 0, -3.64831883244737e-05, -538084.012428235, 0, -538084.012377345, -7.17477585662054e-05, 0, -0.000127313362587972, -538084.012522505, -538084.012718151, -76869.1450108547, -1614252.03746458, 0, -1614252.03745785, -0.000177071530866043, 0, -0.000157669668730818, -3228504.07514169, -1614252.03762653, -3228504.07578302, -1614252.0385741, -538084.013331149, -538084.013811567, -2690420.06267666, 0, -2690420.06269537, -0.000216512909294448, 0, -0.000157728035513796, -8071260.18802021, -2690420.06278891, -8071260.18886422, -2690420.06483052, -1614252.03826704, -1614252.03953663, -8071260.18856068, -8071260.19106239, -3228504.07781898, -10761680.2511876, -2690420.06295789, -2690420.06503827, -16142520.3783662, -10761680.2524538, -16142520.3800998, -10761680.2562729, -8071260.19068191, -8071260.19275156, -16142520.381274}, 
                                                    {0, 0, 0, 0, 0, 5.70996418338242e-08, 0, 104.142857167089, 0, 0, 0, -5.66231401065326e-07, 0, -2400.49285711392, 0, -104.142857804457, 0, 0, 0, -104.142857145541, 0, 2.91994988954385e-06, 0, 21323.2499986191, 0, 5.68015021786901e-06, 0, 0, 1.74395858597646e-08, 2296.35000226524, 0, -9.79652587318659e-08, 2296.35000003338, 1.40385796077465e-08, -6.7025740639886e-08, 0, -8.66768181826571e-06, 0, -96528.712493586, 0, -2.29042980866714e-05, 1.69098418318458e-07, 0, -3.91675057539523e-08, -19026.8999958789, 0, -2.31131615570955e-05, 3.31628320178679e-07, 0, 8.0750580994977e-08, -19026.9000002609, -4.36902027845399e-07, 2.95000438938919e-07, -7.59173984886122e-07, -2.9294672747992e-07, 5.20919252395045e-07, 0, 1.5320048998954e-05, 0, 243577.124986525, 0, 5.04030192533325e-05, -2.44483583225188e-07, 0, 1.92736883281103e-07, 77501.8124676429, 0, 8.09780282473175e-05, -2.53933744312909e-08, 0, -4.7284242485166e-07, 77501.8124999739, 5.99300576857238e-07, -9.054546096949e-07, 2.37287803346163e-05, 0, -1.0397631563748e-06, 2.55446630524359e-06, 3.25210817083054e-07, 2.24896517246139e-06, -1.16265712158015e-07, -7.3162688890324e-07, -1.5909799311669e-06, 4.43351599112982e-06, 0, -1.58844095541872e-05, 0, -345436.649985153, 0, -6.2325623712586e-05, 5.26439881186393e-08, 0, -9.47008133982647e-08, -166075.312439286, 0, -0.000134446244504098, -6.21796004454315e-08, 0, -4.92182111731743e-07, -166075.312500161, -8.46123161584501e-07, 1.04469340448342e-06, 2.75948329928372e-05, 0, -9.92838622240978e-05, -5.7418983271747e-07, 0, 1.19634483166525e-06, -2.68743462928846e-06, 5.94974131397149e-07, -2.82465056433448e-06, -2.0701481220396e-06, 5.49637633198759e-07, 7.92017542960406e-06, -3.61671205266225e-06, 1.32126419097508e-06, 9.11979355329212e-06, -1.28228634334306e-05, -7.25011029667601e-06, -1.47251755052469e-06, 0, 8.901583245727e-06, 0, 256230.482134514, 0, 4.07341392938288e-05, -1.63223266015326e-07, 0, 1.92690735584461e-07, 179361.337451261, 0, 0.000108199178444005, -2.73496549608389e-07, 0, 1.43079532533057e-07, 179361.337499913, 9.01484554101061e-07, -9.83307536572746e-08, -6.13983023154882e-05, 0, 0.000125007890426704, 7.24913072240468e-07, 0, 7.83736052334783e-08, 1.79867374239098e-06, -6.8446738413341e-07, 2.02682192336558e-06, 9.19886991218373e-07, -1.88575626633799e-06, -5.14023939650778e-06, 2.74153945898131e-05, 0, -2.72043658032145e-07, 4.12007389939231e-06, 7.56222752035849e-07, 4.96680291160446e-06, 7.54933361772523e-07, -5.4083181888951e-06, -2.55872973265886e-06, 1.00993349854883e-05, 1.3772309664205e-05, -1.16807543153704e-06, 2.95069423102808e-05, 7.09567541065633e-06, 2.47523949943385e-06, 0, -2.07810648678759e-06, 0, -76869.144640973, 0, -1.09294488876099e-05, 1.090577255486e-08, 0, -3.39895580440325e-08, -76869.1446282204, 0, -3.40938701735595e-05, 1.71703454688415e-07, 0, 0, -76869.1446427972, -2.24673777652121e-07, -1.40597943309013e-07, 2.85455394741071e-05, 0, -5.25194058125731e-05, 1.54526834134622e-07, 0, 0, -5.69680135096529e-07, 5.48158625304739e-07, -3.90527084139826e-07, 1.80710873132741e-07, 8.26128506649139e-07, 2.81991404019723e-06, 1.24578398007096e-05, 0, -3.17317730637046e-05, 8.5114369458361e-07, 0, 5.3120881477939e-07, -1.70132580221772e-06, 1.28731630175575e-06, -1.32609380695608e-06, 4.81324848976078e-07, 1.32275748151008e-06, 5.01204298807304e-06, -2.95298200704806e-06, -5.63416357117056e-06, 3.87182045510296e-06, -2.10443554058963e-06, 1.95221610663468e-06, 5.00320204802373e-06, -8.80812052800042e-06, -9.00863482540508e-06, -1.23756271579812e-05, -4.98624912049602e-06, -6.17190473405659e-07, -3.98210183159773e-06, -1.02736377858416e-05}, 
                                                    {0, 0, 0, 0, 0, -3.32415843997196e-08, 0, 104.142857130074, 0, 0, 0, 3.21103630497018e-07, 0, -104.142857030679, 0, -2400.49285691537, 1.33616360656362e-08, 0, 4.11393300442645e-08, -104.142857144187, 0, -1.6108953255099e-06, 0, -5.70458608331035e-07, 0, 21323.2499982853, -2.16803259324614e-08, 0, -3.07336266698525e-07, 2296.34999870147, 0, -2.6486927839677e-07, 6.9590634261022e-08, 2296.35000007464, -1.56004617419565e-07, 0, 4.6459085272318e-06, 0, 1.81100789231307e-06, 0, -96528.7124931259, 1.203986159804e-08, 0, 1.17485305976419e-06, 4.08184027037005e-06, 0, -19026.8999930817, 5.94063541423913e-07, 0, 1.7923022378382e-06, -3.98700447208614e-07, -19026.9000011372, 1.00319794672218e-06, 8.2703452826599e-08, -3.60817745988883e-07, 2.43786154094838e-06, 0, -7.97660740038859e-06, 0, -3.54242292510982e-06, 0, 243577.124984412, 2.59046624696906e-07, 0, -2.42417247758076e-06, -8.33618022886764e-06, 0, 77501.8124795805, -9.39975860638448e-08, 0, -5.90881901981765e-06, 8.54171637798477e-07, 77501.8125061917, -2.77751566776818e-06, -1.38479545387278e-05, 0, -2.91842886242111e-06, 1.19623944013207e-06, 2.2915596670672e-06, 2.26396336271432e-06, -6.02605276817091e-06, 8.34885644790223e-07, -1.24759202729764e-05, -8.94067988395027e-06, 0, 8.04270125248831e-06, 0, 4.00833628188443e-06, 0, -345436.649980052, -2.75544388418358e-07, 0, 2.82523771689194e-06, 1.13002532974112e-05, 0, -166075.31246714, 6.03905523938296e-07, 0, 8.98750591661905e-06, -9.87059011722115e-07, -166075.312512459, 4.14838360873893e-06, 1.46718590173195e-05, 0, 2.82357909173589e-05, 5.020527954779e-06, 0, 8.99268697957699e-06, -3.57173482799184e-06, -5.14544360690056e-06, -1.03213030048525e-05, 1.74892131552508e-06, -3.196250005197e-06, 2.32849697157407e-05, -3.10033740041117e-06, -6.83221329848473e-06, 3.59532292834527e-05, 1.12526826956494e-06, 2.64079247154941e-05, 2.30790510818813e-05, 0, -4.39162827706027e-06, 0, -2.38953055435637e-06, 0, 256230.482129418, 2.22665270860389e-07, 0, -1.72939474339411e-06, -8.62315389784156e-06, 0, 179361.337473362, 1.32123615471361e-07, 0, -6.50413273289951e-06, 2.80376730138189e-07, 179361.337510212, -3.024947597e-06, -1.00815332505373e-05, 0, -3.0162178340285e-05, -1.40674298665639e-07, 0, -1.08512748255553e-05, 2.18360539636314e-06, 4.85620459765201e-06, 1.59306027848258e-05, 6.87349745334346e-06, 3.92957747662749e-06, -1.70369069948223e-05, -1.59701139145696e-05, 0, -8.24498671012283e-06, 3.62538399952048e-06, 5.825487407484e-06, 7.50302193788786e-06, -3.30199932400317e-05, 5.42704858869587e-06, -4.48583362901569e-05, 8.32772204347261e-06, -1.47796954510376e-05, -2.60629685240315e-05, -1.46853322936836e-05, -8.7029235696713e-06, -4.34700749182833e-05, 0, 1.00133854108617e-06, 0, 5.78629935497555e-07, 0, -76869.1446391585, -4.80076300147283e-08, 0, 4.06526417385981e-07, 2.68236032368791e-06, 0, -76869.144634318, 3.05126133487743e-08, 0, 2.10170408500107e-06, -3.81173971635049e-08, -76869.1446464043, 8.33596962845201e-07, 3.45466805455845e-06, 0, 1.23596786141998e-05, 8.6157259765929e-08, 0, 4.39134804520892e-06, -6.11462864234602e-07, -1.31474156088357e-06, -7.7659955708079e-06, -6.42636639346637e-06, -1.46486693591949e-06, 7.19541973917879e-06, 3.32051734310987e-06, 0, 8.28838924163045e-06, 1.51138727541263e-06, 0, 4.17939723415631e-06, -1.4926879700394e-06, -3.54150723511058e-06, -7.41043463316801e-06, 6.76950922050391e-06, -4.19459223985455e-06, 1.63727372082818e-05, -5.48066622651662e-06, 2.41126186847935e-07, 7.91682521246719e-06, -4.30957191092691e-06, -4.35649943912468e-06, 2.08456963936968e-05, 2.50983061270452e-06, -3.82806672223492e-07, 7.06334815928813e-06, 2.22384853628355e-05, 1.07973408226264e-05, 2.92522198600479e-05, 2.01070350900557e-05}, 
                                                    {0, 0, 4.742910332624e-08, 0, 0, -1.12335476558629e-06, 0, -3401.99999985234, -2.72050354369441e-07, 0, 0, 1.05164036401742e-05, 0, 61102.3499985125, 0, 30484.3499984581, 2.37519625561875e-06, 0, 4.81520460966024e-06, 30484.3500057326, 0, -5.12507768052765e-05, 0, -386324.43749083, 0, -111965.287492132, -1.15177482404375e-05, 0, -3.55259535702998e-05, -498289.72499559, 0, -3.0255642289931e-05, -498289.725031072, -223930.575062492, -111965.287567087, 0, 0.000144424249702805, 0, 1223585.49370972, 0, 215897.906224323, 3.60381730075279e-05, 0, 0.000132619547104613, 2663068.89376025, 0, 1655381.30625418, 9.18062105355158e-05, 0, 0.000181846100520802, 2663068.89382858, 647693.719059334, 215897.906603424, 3310762.61282199, 1655381.30652031, 647693.719243726, 0, -0.000243898855623505, 0, -2173741.31240431, 0, -230660.156207796, -7.103737736856e-05, 0, -0.000275246899487991, -6751884.09371466, 0, -2865721.78120497, -0.000172649227787517, 0, -0.000543413273845311, -6751884.09386214, -922640.62579307, -230660.157252769, -7213204.40655728, 0, -0.000408068786685028, -14426408.8131485, -7213204.40675731, -8597165.3451038, -1383960.93920426, -2865721.78224458, -922640.62670043, -8597165.3457285, 0, 0.000243282070589572, 0, 2205480.14988091, 0, 129538.743726005, 8.5592025772908e-05, 0, 0.000319206359400892, 8951459.34355666, 0, 2723635.12479177, 0.000189875352800469, 0, 0.000851108223482505, 8951459.34385207, 647693.719837057, 129538.745295678, 13751035.8755918, 0, 9599153.06316161, 0.000521420963848295, 0, 0.000861742490726866, 27502071.7505852, 13751035.8755328, 10894540.5026449, 1295387.44057052, 2723635.12696872, 647693.721914521, 28797459.1896498, 9599153.06374592, 1295387.44116996, 28797459.1906283, 16341810.7557522, 10894540.5045457, 0, -0.000131994667054921, 0, -1195742.24992604, 0, -29893.5562592167, -5.60234544526014e-05, 0, -0.000193454499077056, -6008604.80601461, 0, -1345210.03099736, -0.000115363483668221, 0, -0.000672978433553891, -6008604.80630732, -179361.338255923, -29893.5574566445, -12106890.2815931, 0, -6277646.81295095, -0.00035822989526628, 0, -0.000861570114902288, -24213780.5627275, -12106890.2815483, -6726050.158631, -448403.34641786, -1345210.03323259, -179361.340577328, -12256358.0635915, 0, -0.00068305607844002, -36769074.1888261, -12256358.0632949, -25110587.252951, -597871.129072969, -6277646.81394073, -448403.34750321, -36769074.1896474, -13452100.3196984, -6726050.16120112, -37665880.8814405, -25110587.254513, -13452100.3204399, 0, 2.99983285414222e-05, 0, 269042.00623187, 0, 1.01854591929589e-05, 1.53145780118186e-05, 0, 4.7772239096214e-05, 1614252.03741041, 0, 269042.006154534, 3.02009035262073e-05, 0, 0.000209196762963148, 1614252.03751518, 0.000212465021707433, 0.000368661871915774, 4035630.09378262, 0, 1614252.03757405, 9.64091345596865e-05, 0, 0.000333744205053534, 8071260.18753133, 4035630.09381849, 1614252.03830229, 0.000881211039750208, 269042.007028889, 0.0011837658223136, 5380840.12544222, 0, 4035630.09423989, 0.000225766030996726, 0, 0.000316373622346131, 16142520.3752461, 5380840.12521718, 8071260.18886984, 0.00173185567428626, 1614252.03818984, 0.001573008874014, 16142520.3755487, 4035630.09687827, 1614252.03968831, 16142520.3760086, 4035630.09414836, 0.00179641418274529, 24213780.5647239, 16142520.3765422, 16142520.3791732, 5380840.12974043, 8071260.18971401, 4035630.09741871, 16142520.3793579}, 
                                                    {0, 0, -6.43637622932262e-08, 0, 0, 1.54489908698157e-06, 0, 3402.00000012615, 1.22923555828568e-07, 0, 0, -1.46035139293394e-05, 0, -68756.85000023, 0, -22829.8500037672, -6.39598188087975e-07, 0, -2.45838276083516e-06, -22829.8500025676, 0, 7.18731449011105e-05, 0, 506582.099995572, 0, 60598.1250310324, 1.20155957757619e-06, 0, 2.01172580822901e-05, 429399.225023632, 0, 9.94053150814112e-06, 429399.225013114, 121196.25002807, 60598.1250294082, 0, -0.000204462431359872, 0, -1822235.73746516, 0, -79552.1251211002, -1.91406841204324e-06, 0, -7.87247003478256e-05, -2799414.67508034, 0, -1056731.06266341, -2.20328064786933e-07, 0, -8.76854946249225e-05, -2799414.6750266, -238656.375150949, -79552.1251629325, -2113462.12514812, -1056731.06261541, -238656.375190328, 0, 0.000348261312203583, 0, 3579845.6249034, 0, 51667.875272824, 7.05035310103451e-06, 0, 0.000166186738515763, 8336980.68755312, 0, 1280625.18789665, -5.81179905731087e-05, 0, 0.000308128405085448, 8336980.68752214, 206671.500411151, 51667.8754824795, 5986092.37561974, 0, 0.000103780527334195, 11972184.7502589, 5986092.375185, 3841875.56319799, 310007.250667933, 1280625.18795307, 206671.500643633, 3841875.56332329, 0, -0.00034998678022555, 0, -3932663.39987252, 0, -13286.0253629251, -1.58231816860483e-05, 0, -0.000193147079621934, -12522078.5623557, 0, -763946.437952218, 0.000121549628691992, 0, -0.000521168714982902, -12522078.5625091, -66430.1255927484, -13286.0257695439, -13983541.3134016, 0, -6144786.56377331, 2.59418883235597e-05, 0, -0.000375232906440372, -27967082.6251325, -13983541.3126331, -3055785.75147659, -132860.251264404, -763946.438430802, -66430.1262164254, -18434359.6885462, -6144786.5630399, -132860.251286272, -18434359.6887145, -4583678.6275688, -3055785.75182121, 0, 0.000191103133720543, 0, 2271910.2749186, 0, 0.000262253059118404, 1.50346577550718e-05, 0, 0.000116973068138699, 9267002.43726522, 0, 179361.337756568, -9.88145294595783e-05, 0, 0.000424008188668359, 9267002.43750475, 0.000431806745673386, 0.000618415771365593, 14348907.0004931, 0, 2989355.62602563, -0.000133533811435701, 0, 0.000464454501101173, 28697813.9999634, 14348907.0000211, 896806.688894811, 0.00113006217463476, 179361.338459628, 0.00123757837500414, 10163809.1264789, 0, 0.000147499690816066, 30491427.3755069, 10163809.1253195, 11957422.5016493, 0.00149724425012556, 2989355.62565622, 0.00124900155502304, 30491427.3756492, 1793613.37848232, 896806.68946925, 17936133.7528547, 11957422.5018119, 1793613.37824626, 0, -4.36662643973107e-05, 0, -538084.012479766, 0, -7.83128801618789e-05, -5.32797132384621e-06, 0, -2.86653342460865e-05, -2690420.06240336, 0, -6.32820736078822e-05, 2.86662244984534e-05, 0, -0.000133113922199243, -2690420.06250274, -0.000126383393290783, -0.000194136182934754, -5380840.12505799, 0, -538084.012784686, 7.20450820335242e-05, 0, -0.000197428461180438, -10761680.2499689, -5380840.12498623, -0.000485891300105097, -0.000378086392404915, -0.000386829698174825, -0.000497200002858866, -5380840.12553283, 0, -2690420.063162, 1.03785153680012e-05, 0, -0.000121838145572268, -16142520.3750217, -5380840.12506944, -2690420.06331801, -0.000654294049806211, -538084.012816175, -0.000507831547588187, -16142520.3750816, -0.00159967619548754, -0.000862542019805502, -10761680.250526, -2690420.06267676, -0.000586175754083687, -16142520.3758429, -10761680.2505756, -5380840.12707068, -0.00208795835581687, -2690420.06340718, -0.00140096672797015, -5380840.12682487}, 
                                                    {0, 0, 5.39523030394938e-08, 0, 0, -1.30764198845223e-06, 0, -2551.50000027643, -1.74120740477336e-08, 0, 0, 1.2465023460857e-05, 0, 54629.4375008071, 0, 12529.6875067802, -2.5991911058731e-07, 0, 6.03083416419932e-07, 12529.6875007184, 0, -6.18890961692993e-05, 0, -436306.499994523, 0, -22895.1563039042, 3.10739853759922e-06, 0, -6.06102743230696e-06, -252530.156285102, 0, 1.01834452650248e-06, -252530.156254089, -45790.3125079348, -22895.1562574067, 0, 0.000177582153120156, 0, 1721134.82808257, 0, 18452.8127088345, -1.17689557578366e-05, 0, 2.57996419227323e-05, 1825803.28132065, 0, 433128.515872484, -2.80038354557522e-05, 0, 1.71245693700441e-05, 1825803.28125711, 55358.4375461643, 18452.8125412291, 866257.031302437, 433128.515658615, 55358.4375404502, 0, -0.000304887455394254, 0, -3686871.93738922, 0, -5535.84420773912, 2.10756866928022e-05, 0, -5.58330983125227e-05, -6166929.93747099, 0, -326614.781946734, 0.000103558410419669, 0, -8.91870316137724e-05, -6166929.93750012, -22143.3751274892, -5535.84387768532, -2801136.93810924, 0, 3.93242286788155e-05, -5602273.87509735, -2801136.93754926, -979844.343999266, -33215.0626435403, -326614.781375998, -22143.3751146763, -979844.344030019, 0, 0.000308558705698171, 0, 4367780.71860977, 0, 0.000578656165608066, -1.98459749252832e-05, 0, 6.56471141949727e-05, 10412922.0934834, 0, 91341.4228637277, -0.000161949145964639, 0, 0.000173168882647244, 10412922.0937405, 0.000189898650652027, 0.000214397810836859, 7813843.4537495, 0, 1860043.50138146, -0.000178143618279749, 0, 4.62780528901844e-05, 15627686.9062971, 7813843.45313105, 365365.688031879, 0.000279821383722678, 91341.4221175664, 0.000184413877143972, 5580130.50037683, 1860043.50014401, 0.000206378238023891, 5580130.50047851, 548048.532131063, 365365.688036754, 0, -0.000169507612931787, 0, -2690420.06241286, 0, -0.000392782988010803, 8.94394596202975e-06, 0, -4.01455502608549e-05, -8519663.53093998, 0, -0.000715733757082294, 0.000120803316868658, 0, -0.000145178266103279, -8519663.5312425, -0.000146934182997968, -0.000179500735464818, -9416470.21893631, 0, -448403.345030764, 0.000219394290168087, 0, -0.00012056052605811, -18832940.4374899, -9416470.21870305, -0.000502459333113377, -0.000243402249744853, -0.000251355464704099, -0.000190752307346552, -4035630.09493358, 0, 8.32220115293876e-05, -12106890.2814118, -4035630.09383031, -1793613.37560907, -0.000284398297873951, -448403.343904977, -0.000118961933421984, -12106890.281539, -0.00117037398609363, -0.000493455201403001, -2690420.06363732, -1793613.37570052, -0.000987686122059886, 0, 3.89325971313664e-05, 0, 672605.015603737, 0, 0.000110199568989646, -1.35357873123935e-06, 0, 9.81131770408437e-06, 2690420.06238615, 0, 0.000213134151537805, -3.45862554299886e-05, 0, 4.62492638980692e-05, 2690420.0624983, 4.45521717360581e-05, 5.91657433480478e-05, 4035630.09370739, 0, 0.000431555817407447, -8.76111431078934e-05, 0, 6.0219984704881e-05, 8071260.1874912, 4035630.09371944, 0.000176082610256854, 7.57648378465575e-05, 0.000107087466585535, 8.79209792672238e-05, 2690420.06283409, 0, 672605.016209719, -7.56757756496541e-05, 0, 4.18456243820127e-06, 8071260.18749514, 2690420.06250263, 0.000297337758931372, 0.000117957462573966, 6.1426260843435e-05, 1.24137101523973e-05, 8071260.18754077, 0.000536832535648642, 0.000169341886165425, 2690420.06268023, 672605.015654204, 5.71027370848431e-05, 4035630.09411723, 2690420.06277514, 0.00079986327844806, 0.000666750955777108, 0.000321731366188397, 0.000360981996086566, 0.000715104376629747}, 
                                                    {0, 0, -3.0580601769699e-08, 0, 0, 7.4828378078235e-07, 0, 1360.80000021891, 0, 0, 0, -7.19741547275156e-06, 0, -30156.3000003716, 0, -4641.30000547365, 1.57304372432305e-07, 0, -6.29934327966917e-08, -4641.30000016, 0, 3.60611221019898e-05, 0, 253017.674993195, 0, 5248.80004439908, -1.36814661063608e-06, 0, 9.55317506646217e-07, 97521.9750247962, 0, -1.37002342921126e-06, 97521.9750012047, 10497.6000021352, 5248.80000087584, 0, -0.000104361844352356, 0, -1061979.86245982, 0, -1968.30017381355, 5.03146490973388e-06, 0, -4.77701687207562e-06, -750988.462521015, 0, -104319.90019476, 1.19215680441096e-05, 0, 9.17760554880507e-07, -750988.462502535, -5904.90001407633, -1968.30000304167, -208639.800018573, -104319.900007526, -5904.90000546476, 0, 0.000180550707628174, 0, 2438723.69990521, 0, 0.000379525225764526, -9.15128447353747e-06, 0, 1.09564799720395e-05, 2760540.74988276, 0, 36905.6255998161, -3.7439973174347e-05, 0, 1.03236367010127e-05, 2760540.74999969, 3.74241670334144e-05, 1.1160173405099e-05, 730731.375358003, 0, -3.01664970188955e-05, 1461462.750045, 730731.375011832, 110716.875079248, 2.95216089519889e-05, 36905.6250153627, -9.42405164242946e-06, 110716.875099647, 0, -0.000183929876281296, 0, -3095643.82488697, 0, -0.00047238047582312, 9.43139779415229e-06, 0, -1.32976898009132e-05, -5154977.69968145, 0, -0.000918841702839202, 5.38890757395463e-05, 0, -3.13310337486001e-05, -5154977.6999946, -5.6702790013376e-05, -2.44864060277904e-05, -2291839.31268334, 0, -232505.438455113, 7.75284289108663e-05, 0, 2.17171344324225e-05, -4583678.62504242, -2291839.31249276, -0.000156285736660769, -7.33064569627703e-05, -1.54369816571004e-05, 5.7693728682513e-05, -697516.31262549, -232505.437514635, 1.23702548151422e-05, -697516.312726577, -0.000304420633633091, -0.000132795096965098, 0, 0.000101604008941287, 0, 2032761.82493253, 0, 0.000313595371286777, -5.01292992866127e-06, 0, 8.49890348353965e-06, 4663394.77470064, 0, 0.000703551429628974, -3.89148017847094e-05, 0, 2.80702291339915e-05, 4663394.7749945, 4.62072742254157e-05, 2.51097027352783e-05, 3228504.07487277, 0, 0.000997580872991024, -8.2013948830385e-05, 0, 5.86273016949388e-06, 6457008.15002479, 3228504.07496966, 0.000140765110672541, 6.82536798307556e-05, 1.59042004319103e-05, -6.10766020515079e-05, 597871.125603002, 0, -5.91687103509519e-05, 1793613.37506765, 597871.125004129, 0.000175625208023201, 4.50402124787126e-05, -8.16696907511298e-06, -8.83354094194027e-05, 1793613.37519059, 0.00037935935303542, 5.33517055996724e-05, 0.000492949378753089, 0.000278390500194736, 0.00028054946641201, 0, -2.3444700037956e-05, 0, -538084.012484005, 0, -8.58492564404352e-05, 1.03353566666917e-06, 0, -2.07354223866328e-06, -1614252.0374007, 0, -0.000216021264375507, 1.10949936129728e-05, 0, -9.32814005189853e-06, -1614252.03749812, -1.39859872916517e-05, -9.82737302988436e-06, -1614252.03738422, 0, -0.000373086317214618, 2.97833558477864e-05, 0, -7.53238383168743e-06, -3228504.07500666, -1614252.0374818, -4.78880663445969e-05, -2.10976647526386e-05, -1.1163961959245e-05, 1.71159250069585e-05, -538084.012596954, 0, -0.000357902659130189, 3.55825394898028e-05, 0, 1.44220583629365e-05, -1614252.03751143, -538084.012488881, -7.55597785815735e-05, -2.26028819265678e-05, 1.20134468606393e-05, 6.96505665291612e-05, -1614252.03755131, -0.000170421045519189, 1.74468104662332e-05, -4.98376555543873e-05, 1.50524191437421e-05, 3.27998061756604e-05, -0.000196899569436908, -0.000146356395820863, -0.000297414343083022, -0.000205357931285918, -9.86624644502528e-05, -6.12762604866938e-05, -0.000290908798209372}, 
                                                    {0, 0, 1.09480756271448e-08, 0, 0, -2.71561244670343e-07, 0, -486.000000096289, 0, 0, 0, 2.64714344906244e-06, 0, 11020.0500000158, 0, 1032.75000252372, -2.56571036638632e-08, 0, 0, 1032.75000003037, 0, -1.34336810450491e-05, 0, -95581.0124956406, 0, -546.750021091625, 2.04216435679524e-07, 0, -1.03568538086656e-07, -22316.5125099555, 0, 3.9494177991506e-07, -22316.5125002801, -1093.50000045228, -546.74999990888, 0, 3.93251840249998e-05, 0, 419042.86872795, 0, 8.38079639142431e-05, -9.73529713574188e-07, 0, 5.67507426729109e-07, 179320.331245537, 0, 11235.7125892438, -2.08007587090913e-06, 0, -7.26600317842829e-07, 179320.331250972, 3.70550872227648e-06, -1.13731502584721e-06, 22471.4250052409, 11235.7125015378, -1.15882184447341e-07, 0, -6.87046431750045e-05, 0, -1014904.68745149, 0, -0.000183530074409244, 1.65906859593737e-06, 0, -1.52007894892102e-06, -699361.593656548, 0, -0.000295218587587484, 4.8286355723078e-06, 0, 3.68993838576262e-07, -699361.593749999, -8.54416476996811e-06, 2.93937228899205e-06, -83037.6563739308, 0, 6.97466571203501e-06, -166075.312515614, -83037.6562519632, -1.94166310062595e-05, -6.45107521597275e-06, 1.61267858138337e-06, 8.36106843235614e-06, -2.91704686795122e-05, 0, 7.05649343206764e-05, 0, 1368460.57494474, 0, 0.000227163733684895, -1.49323495721135e-06, 0, 1.63830529028654e-06, 1411640.15605423, 0, 0.000474002377548969, -6.54239190655261e-06, 0, 4.28190869820204e-06, 1411640.15624886, 1.26898040798414e-05, -2.29881455284425e-06, 282328.031227191, 0, 0.000406585218308861, -1.06851079094609e-05, 0, -7.29282156313506e-06, 564656.062517126, 282328.031246378, 3.34187256151133e-05, 2.12938378748722e-05, -6.88615302268923e-06, -3.43790141016625e-05, 3.16898220175939e-05, -4.841218486907e-06, -1.94874225133352e-05, 7.77978586106919e-05, 7.69409275453628e-05, 2.66292736639681e-05, 0, -3.9243527817106e-05, 0, -956593.799968109, 0, -0.000149241969758592, 1.04208602279897e-06, 0, -1.32964937312575e-06, -1404997.14358378, 0, -0.000373730041433307, 4.94740795388354e-06, 0, -3.62991287785647e-06, -1404997.14374833, -1.0870371781268e-05, -5.0664395732403e-07, -448403.343591334, 0, -0.000470048216344245, 1.03460060723096e-05, 0, 7.05390282670735e-07, -896806.687512296, -448403.34374105, -2.76323417439766e-05, -1.94179647256378e-05, 8.86759141548388e-06, 3.06756721166282e-05, -0.000179848410765405, 0, 1.08812022289479e-05, -2.27736902997249e-05, 2.64545520501673e-06, -3.9879646238437e-05, -1.10949961571385e-05, 1.84695200821364e-05, 2.88850181557496e-05, -6.9999039598255e-05, -9.95407837652205e-05, 4.54411892127848e-07, -0.000166689732387729, -7.70269165643142e-05, -6.25577334940023e-05, 0, 9.1052819066916e-06, 0, 269042.006242638, 0, 4.03419710886021e-05, -1.94623153755103e-07, 0, 3.12321610272197e-07, 538084.012448127, 0, 0.000116456188015313, -1.57691903715937e-06, 0, 1.13431404682493e-06, 538084.012499334, 3.09151653828772e-06, 8.94285429580692e-07, 269042.006164658, 0, 0.0001880352049209, -3.98068111451019e-06, 0, 2.72115761235525e-07, 538084.012503601, 269042.006244551, 8.4760372910002e-06, 5.28908523813024e-06, -2.33145471518834e-06, -1.19844262507974e-05, -7.98476653496137e-06, 0, 0.000139963930066918, -7.01400003153138e-06, 0, -4.24681897679436e-06, 7.88372715029097e-06, -6.21009586638614e-06, 1.32350964755978e-05, 5.37764783747471e-06, -9.16817582169398e-06, -2.72525856176796e-05, 2.1136998020141e-05, 4.37298522913695e-05, -1.68611125915872e-05, 1.2681690482689e-05, -1.04126088440486e-05, -1.80296969199307e-05, 6.63105182179773e-05, 5.23196281580703e-05, 8.46972062714783e-05, 5.05316619268212e-05, 2.10478351138903e-05, 1.46025761110518e-05, 8.34780646648126e-05}, 
                                                    {0, 0, 0, 0, 0, 5.4878010138722e-08, 0, 60.7500000056372, 0, 0, 0, -5.0118988993531e-07, 0, -1309.16250013094, 0, -607.500000048165, 0, 0, 0, -60.7500000154482, 0, 2.37881539128514e-06, 0, 10474.8187508333, 0, 546.749999825411, -5.64037900999388e-08, 0, 5.5188265154219e-08, 13030.8750016024, 0, 1.62096218553332e-07, 1248.41250007189, 546.750000387427, 1.35489432819335e-07, 0, -6.53645442535605e-06, 0, -40596.1875019647, 0, 2.00531348618906e-06, -3.70679162902157e-08, 0, -3.57991824195383e-07, -103499.775011442, 0, -11235.7125034312, -2.97157210297509e-07, 0, -2.80830812327893e-07, -9226.40624956602, -2.06365404600933e-06, 4.10217772312208e-08, -11235.7125019999, 2.20286032416377e-08, -1.82409910884811e-06, 0, 1.07891046197529e-05, 0, 81192.3750018444, 0, -5.44010619750569e-06, -3.12266213259136e-08, 0, 4.81771430299535e-07, 396735.468778247, 0, -6.06653817337055e-06, -1.87706421484542e-06, 0, 2.35953786436836e-06, 31369.7812488866, 6.31981551972687e-06, 3.75865609262294e-07, 83037.656287474, 0, 7.40513836069233e-07, 83037.6562547023, -9.09414677225865e-07, 1.05650098169447e-05, 5.86639428513524e-06, 8.67522090212875e-07, -6.50360690895173e-07, 8.5889579204341e-06, 0, -1.0548094147374e-05, 0, -79716.1500001644, 0, 6.67192681228536e-06, 5.24057898089733e-07, 0, -8.20046422511905e-07, -780553.968781594, 0, 2.68831835123905e-05, 3.03487118273169e-06, 0, -2.78358783846886e-06, -49822.5937465017, -9.86366576180916e-06, -7.26884328500346e-07, -282328.031333021, 0, -3.99487297321928e-05, 5.17243184195178e-06, 0, -2.69748851950356e-06, -282328.031254501, 4.95719240965338e-06, -2.28812015784636e-05, -1.1140448400527e-05, 3.90521688432855e-06, -6.01843475895449e-07, -1.94867286827427e-05, -2.0849240913216e-07, -1.5789955454648e-05, -1.66978933710186e-05, -3.86539144950298e-05, -1.68698041860514e-05, 0, 5.62501284559187e-06, 0, 29893.5562492436, 0, -3.67597467723259e-06, 3.64533487324361e-08, 0, 2.48836836805695e-07, 747338.906265948, 0, -2.98259054005306e-05, -1.38495774351262e-06, 0, 2.54747782582178e-06, 29893.5562474402, 7.98521161127441e-06, -4.23125568546063e-07, 448403.343822058, 0, 6.96276882794931e-06, -6.39594270888371e-06, 0, 2.32272531413628e-06, 448403.343752705, -9.90265469086222e-06, 2.04749009388941e-05, 1.4243324952908e-05, -1.37986213752361e-06, -6.44668459422312e-06, 8.11515748305851e-05, 0, -1.820124467834e-06, 1.0563812933237e-05, -5.72913699328491e-06, 2.52504190766059e-05, 5.57461152712646e-06, 5.90575947740364e-06, -1.28900498065415e-05, 1.74808488733403e-05, 3.41941060396642e-05, 5.90148096434477e-06, 3.61696867605079e-05, 3.53903809970261e-05, 4.01700324403427e-05, 0, -1.2597650174601e-06, 0, 3.30840919337996e-07, 0, 7.0783532786905e-07, 3.78137947929116e-08, 0, -8.9432840164591e-08, -269042.006252767, 0, 1.08374369592307e-05, 2.38415337326029e-07, 0, -8.1476649523797e-07, 7.44096063025205e-07, -2.45327246113703e-06, 2.28101871463896e-07, -269042.006271479, 0, 6.85564412110896e-06, 1.2319153341717e-06, 0, -9.19614760582804e-07, -269042.006250098, 4.70059495811393e-06, -7.21929746749089e-06, -5.84370352298432e-06, -4.38503273143481e-07, -1.23793911733558e-07, -3.94706649731456e-05, 0, -2.32921456048237e-05, 1.90529066184525e-07, 0, -6.16203918181687e-07, -2.83730928553549e-06, 1.78415422597178e-06, -1.04797370121091e-05, -4.8510445055334e-06, 1.76414940982992e-06, 4.12480716371807e-06, -4.30302473561958e-06, -1.74915385704967e-05, -6.39687776491236e-07, -8.82342262784811e-06, 7.66111533403825e-07, -1.6593741302819e-06, -2.39121072222073e-05, -8.70884530453199e-06, -2.88934472300486e-05, -2.25004827551636e-05, -1.55370762798331e-05, 1.27397139209615e-06, -2.95735315746633e-05}, 
                                                    {0, 0, 0, 0, 0, 3.26234862001421e-08, 0, 48.5999999964352, 0, 0, 0, -3.00695244967651e-07, 0, -959.850000049202, 0, -704.699999879101, 1.33675085007293e-08, 0, -1.72496487643388e-08, -48.5999999947522, 0, 1.44560734928617e-06, 0, 6652.12500067504, 0, 2624.39999886762, -1.30622494428463e-07, 0, 2.06365993417826e-07, 13869.2250000148, 0, -1.77636696557811e-07, 911.249999865435, 656.099999934577, 0, 0, -4.02353372721886e-06, 0, -20503.1250023662, 0, -1968.29999542537, 6.75517195444156e-07, 0, -6.98020780820055e-07, -95544.5625062597, 0, -51175.7999958209, 1.88439144665064e-06, 0, -6.15228056816142e-07, -5740.87499903408, -1968.29999971699, -7.96168454768087e-07, -12301.8749999439, 5.64431543179262e-08, -4.00062633182476e-07, 0, 6.71923720646873e-06, 0, 28048.2750039353, 0, -9.45368986243039e-06, -1.21497470144076e-06, 0, 1.82476341192878e-06, 291554.437521285, 0, 36905.6249777839, -3.67902772799774e-06, 0, 1.2368241896868e-06, 14762.2499961364, -8.11995346823329e-07, 1.54347005237101e-06, 346912.875014421, 0, -2.56319712548698e-06, 77501.8124941639, -2.34502569003364e-06, 36905.6250020751, 1.61067443396301e-06, 8.51726629484855e-08, 3.64242450281624e-06, 3.98899989544926e-06, 0, -6.63657594408951e-06, 0, -13286.0250034289, 0, 9.88004272260026e-06, 6.69182102020352e-07, 0, -2.02603907690203e-06, -391937.737529364, 0, 4.5660472924172e-05, 4.6060439136044e-06, 0, -3.4097385489704e-06, -13286.0249961381, 1.89095076720223e-06, -2.64975154867723e-06, -1029666.9375548, 0, -232505.437502997, 3.91975222271416e-06, 0, 1.27690793687004e-06, -199290.374985661, 8.65231139137686e-06, -5.44542413757773e-06, -1.313553473416e-07, -4.72566525388772e-06, 1.28765503709204e-06, -232505.437499106, 2.40306453424691e-06, 3.88297152485279e-06, 9.33300494259209e-06, -1.95547077463608e-05, -4.38511476864523e-06, 0, 3.57026040693712e-06, 0, 1.50158681608961e-06, 0, -5.02157861258733e-06, -5.99946319634184e-07, 0, 1.49465907831193e-06, 179361.337518764, 0, -4.05373886562473e-05, -4.82655312636673e-06, 0, 2.51558502417917e-06, -2.65832147938817e-06, -2.07672887340353e-06, 3.93409120411921e-06, 1315316.47505478, 0, -2.11720549821925e-05, -4.77554464841937e-06, 0, 1.44468929534438e-06, 179361.33748306, -7.55926470816365e-06, 6.32081074640044e-06, -7.26677395972528e-06, 2.29743176258405e-06, -1.31944211607394e-06, 597871.125047612, 0, -2.77485096504258e-06, 597871.124991271, -1.11350419801189e-06, 9.80803593135367e-06, 8.70125249123143e-06, -9.24093580927976e-06, 7.07566668982458e-06, -2.66605592166561e-05, 4.07320427714481e-05, 6.10775507030599e-06, 1.11030964480728e-05, -1.96939008755346e-05, 1.31221544095561e-05, 0, -8.05556659390759e-07, 0, -2.53230652156707e-07, 0, 9.30230203313807e-07, 5.58432805838565e-08, 0, -3.01190627853946e-07, -4.61090690252006e-06, 0, 1.31582131311351e-05, 1.46536582823637e-06, 0, -8.59874249812873e-07, 6.3375606955187e-07, 3.90567036472933e-07, -1.61539229036045e-06, -538084.012518678, 0, 1.48593899420216e-05, 3.25621023164085e-06, 0, -1.83602200972859e-06, 5.42430632538108e-06, 2.85958259707672e-06, -1.90053081661194e-06, 3.45827267662827e-06, -4.47631005796828e-07, 1.73310429439084e-06, -538084.01252891, 0, -1.08203575075046e-05, 3.70814928932131e-06, 0, -2.07717547615744e-07, -538084.012489526, 3.73793625594332e-06, -9.36659602690921e-06, 8.05104871126556e-07, 4.42190608868429e-07, 6.68098857265941e-07, 1.42944627317043e-05, -1.35429353962341e-05, -1.65105887562808e-06, -5.61483517455937e-08, -1.02212781498143e-07, -3.06801457521468e-06, 1.29924705269711e-05, 1.02404216413938e-05, -2.11999153646218e-05, -1.84860530832931e-05, 7.57331724194685e-06, -1.82057646083751e-05, -8.59787505807474e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 45.562499958752, 0, 0, 0, -9.30880410303064e-08, 0, -797.343749697015, 0, -797.343749275756, -1.46906524672958e-08, 0, -5.02897219820889e-08, -45.5624999929931, 0, 4.85643722152152e-07, 0, 4442.34374878875, 0, 4442.34374502854, 6.54460609416017e-08, 0, 3.42842522529685e-07, 13907.9531210156, 0, 1.22413161400835e-07, 751.781249582262, 751.781249674268, 3.16778569536513e-07, 0, -1.42694391291602e-06, 0, -9226.40624681716, 0, -9226.4062318131, 1.44060682006242e-07, 0, -1.33642321346827e-06, -76989.2343636441, 0, -76989.2343550956, -1.00480751720563e-06, 0, -1.33764066859136e-06, -3690.56249792438, -3690.56249885647, -2.27227060400341e-06, -12404.3906214497, -3.91925829291573e-07, -4.32511978265181e-07, 0, 2.4743702573086e-06, 0, 5535.84374474439, 0, 5535.8437123372, -1.88614802060784e-07, 0, 2.67550880513324e-06, 157771.546851939, 0, 157771.546819248, 2.43489614383638e-06, 0, 5.2018523023075e-06, 5535.8437443908, 5535.84374442707, 7.47164968218007e-06, 420724.124968152, 0, 3.35702900365565e-06, 60894.2812340505, -8.48979037543753e-06, 60894.2812422261, 7.59999373543069e-06, 5.41736439620014e-06, 1.20515744984757e-05, -1.66216418356207e-07, 0, -2.51337368042575e-06, 0, 5.02369265398007e-06, 0, 4.43595963138712e-05, 1.02095745140761e-07, 0, -2.58932950410437e-06, -91341.4218443747, 0, -91341.4217867631, 5.3891381463848e-08, 0, -1.01989119499294e-05, 6.8903071834521e-06, 1.09674945524802e-05, -1.23396408575976e-05, -838680.328088491, 0, -838680.328073799, -5.56227312805075e-06, 0, -8.41966371458759e-06, -91341.4218330948, 1.46077102635839e-05, -91341.4218660016, -1.5082996087234e-05, -1.5612039723007e-05, -2.11264319769778e-05, -298935.5624649, 8.48136045909889e-06, -1.82038165770112e-05, 2.77618600871643e-05, 3.93812215528571e-06, -2.13188850616632e-05, 0, 1.38244726775899e-06, 0, -2.52670244053985e-06, 0, -2.72698702249013e-05, -1.49270194315713e-07, 0, 1.57826563387862e-06, -2.17646034974718e-05, 0, -7.26082445196277e-05, 3.89949199377067e-07, 0, 8.10484397468887e-06, -3.67204138619021e-06, -1.00484720868929e-05, 9.87923973698127e-06, 448403.343720429, 0, 448403.343699344, 4.76635160595584e-06, 0, 1.13706995732969e-05, -3.86690097012815e-05, -9.28460386351336e-06, -1.10520057892124e-05, 8.86521929317507e-06, 1.64986970158576e-05, 3.36303285889803e-05, 1569411.70310872, 0, 6.76515693090892e-06, 448403.343693142, -1.30795389641977e-05, 448403.343738066, 3.27179471253366e-05, -8.23298542336367e-07, 4.80478116559034e-05, -6.21175870844531e-05, 2.78183131761155e-05, 4.11843407326634e-05, -1.02342320523242e-05, -1.42269825974725e-05, 6.880775114015e-05, 0, -3.1768635268775e-07, 0, 5.13977308207938e-07, 0, 6.79208365943105e-06, 1.16976630659161e-08, 0, -4.69581739771639e-07, 6.12104035740086e-06, 0, 2.36268174616065e-05, 2.61384849149531e-07, 0, -2.30012993229369e-06, 6.37665707806394e-07, 3.85089868744624e-06, -3.09697046350806e-06, 1.19806902004985e-05, 0, 2.24470422595259e-05, -1.05003499883903e-06, 0, -2.61848658773337e-06, 1.31900625893056e-05, 2.45378542144536e-06, 5.75276482581441e-06, 2.22846168120325e-06, -6.05800521936957e-06, -7.5759717641344e-06, -672605.015621909, 0, -672605.0156222, -1.57693960895472e-06, 0, -5.37417824603517e-07, 2.52581480815312e-05, 5.29919904156042e-06, 4.24155971472809e-06, -1.54856922683563e-05, -3.03752559057515e-06, -1.22887847103266e-05, 2.88510992212753e-05, -1.60188461092698e-05, -1.6430172602452e-05, -672605.015614026, 6.8251487574002e-06, -6.7607039047764e-06, 4.3320618050581e-05, 1.81258003528894e-05, -4.2940579944179e-06, -4.46849010507989e-05, 4.23774736798958e-06, -4.53563402217818e-05, 1.3237864659428e-05}, 
                                                    {0, 0, 0, 0, 0, 6.80592547896931e-08, 0, 48.5999999662005, 0, 0, 0, -5.94463861530656e-07, 0, -704.699999613582, 0, -959.849999543923, -5.82202251628297e-08, 0, -1.09046686315121e-07, -48.6000000365108, 0, 2.70153706795461e-06, 0, 2624.3999976886, 0, 6652.12499717751, 2.49434478052295e-07, 0, 7.79010992367705e-07, 13869.2249968835, 0, 1.21289088287349e-06, 656.100000203375, 911.250000339521, 8.93222761545177e-07, 0, -7.11681368214436e-06, 0, -1968.29999192428, 0, -20503.1249897943, -1.1915426001936e-06, 0, -2.44397615910573e-06, -51175.7999874304, 0, -95544.562489731, -3.45460008934179e-06, 0, -5.2273442096011e-06, -1968.29999945805, -5740.87499976751, -5.02251856293899e-06, -12301.8750010173, -2.54138234052986e-06, -7.95407316932901e-06, 0, 1.1282653965728e-05, 0, -1.59086557892818e-05, 0, 28048.2749782858, 1.59794986442944e-06, 0, 4.88209399170298e-06, 36905.6249631536, 0, 291554.437475726, 6.61851153536295e-06, 0, 1.41734205613224e-05, -1.32267315011264e-06, 14762.2500003278, 1.3312400764167e-05, 346912.874985896, 0, 1.55368674765992e-05, 36905.624995768, 4.25349489018551e-06, 77501.812506118, 1.16861650936616e-05, 1.203442723398e-05, 1.84154603141484e-05, 1.97409849048086e-05, 0, -1.06186953725961e-05, 0, 1.7403378201069e-05, 0, -13286.0249741865, -1.19918421885799e-06, 0, -6.02426595290699e-06, 6.05467578901854e-05, 0, -391937.737458897, -8.92556150863338e-06, 0, -1.52247226577717e-05, 1.86907996103548e-06, -13286.0250003517, -1.7324563861293e-05, -232505.43747363, 0, -1029666.93750576, -1.47419146120574e-05, 0, -2.6721810762121e-05, 7.11970482760543e-06, -2.03455511671711e-06, -199290.375009337, -1.35138683182306e-05, -2.65620694182092e-05, -5.08233012002581e-05, -232505.437512079, -2.09908893726018e-05, -3.74237133195565e-05, -1.89911066872504e-05, -8.96200144684801e-05, -8.73324136143925e-05, 0, 5.46543748213998e-06, 0, -9.86433309242259e-06, 0, -1.60170688555736e-05, 9.54173303435832e-07, 0, 3.01142667359089e-06, -4.84476261560447e-05, 0, 179361.33746331, 6.14766077595774e-06, 0, 1.32483889306373e-05, -9.59632433820388e-07, -3.19055559505749e-06, 1.37197428542294e-05, -4.17520751361813e-05, 0, 1315316.4750096, 1.27792686436813e-05, 0, 1.69658001327531e-05, -1.2066477056905e-05, -2.15592109937996e-06, 179361.337506733, 1.0055066052068e-05, 2.82661142440991e-05, 3.02767895472513e-05, 597871.12502602, 0, 1.60728113183847e-05, -2.51249660175194e-06, -5.71854470282314e-06, 597871.125005308, -3.61440425676092e-06, 2.15841947394898e-05, -6.01819974415188e-06, 7.69087292404436e-07, 0.000101432471241672, 8.91245503800494e-05, 3.19462746107945e-05, 7.3906632637085e-05, 0.000159346278939934, 0, -1.18469376380635e-06, 0, 2.26469054194282e-06, 0, 4.02328503136193e-06, -7.19000282099582e-08, 0, -6.72847430374696e-07, 1.47563215941742e-05, 0, 1.26055516330666e-05, -2.21023681287791e-06, 0, -4.04196684896574e-06, 1.52923566459934e-07, 6.10702149830366e-07, -3.62937187889888e-06, 2.09780903855727e-05, 0, -538084.012500635, -4.25928576977934e-06, 0, -8.90755252369335e-06, 4.18160029662469e-06, -3.11786680032061e-07, 7.62050317794346e-07, -7.02040835395585e-06, -1.21380895197554e-05, -1.66612836522667e-05, -6.18253585384385e-06, 0, -538084.012522768, -8.13518771286878e-06, 0, -1.10230516035481e-05, 7.35301769787973e-06, 3.51305317636643e-07, -538084.01250994, 6.23866115508094e-07, -1.15598183555347e-05, -1.74382953965922e-05, 5.8332272270752e-06, -3.65241454728987e-05, -4.93480318400342e-05, -5.11740537285358e-06, -6.80455147770806e-06, -1.76631056530905e-05, -1.14604044294569e-06, 7.46130250617491e-06, -4.91998968731117e-05, -9.00177697730051e-05, -4.96607420995912e-05, -6.31623636601674e-05, -8.86910155516629e-05}, 
                                                    {0, 0, 0, 0, 0, -4.12552854365069e-08, 0, 60.7499999672537, 0, 0, 0, 3.99798958965255e-07, 0, -607.499999665377, 0, -1309.16249945895, -2.66097139322201e-08, 0, -4.5495959305476e-08, -60.7500000369492, 0, -1.99561391358481e-06, 0, 546.749998052435, 0, 10474.8187462508, 1.47255649696848e-07, 0, 2.46529129988997e-07, 13030.8749963476, 0, 2.21777083021266e-07, 546.750000375662, 1248.41249978669, 5.98826084043061e-07, 0, 5.70515119869541e-06, 0, 6.38382028932832e-06, 0, -40596.1874860896, -3.66851161957088e-07, 0, -9.6275295051198e-07, -11235.7124845544, 0, -103499.774983122, -1.56726984749056e-06, 0, -1.43262174437791e-06, -1.75300560920812e-06, -9226.40624861473, -4.13865497612076e-06, -11235.7125014569, -4.99809765066754e-06, 2.1218876946736e-06, 0, -9.69862661764474e-06, 0, -1.20349488743685e-05, 0, 81192.3749705364, 7.49475560605576e-07, 0, 1.54989948828679e-06, -3.81406109017544e-05, 0, 396735.468707579, 3.32742466788492e-06, 0, 4.29312434904184e-06, 3.70420306786554e-06, 31369.7812444918, 1.37670762303977e-05, 83037.6562018131, 0, 3.52543778584496e-06, 6.44216062329967e-06, 1.23504459405149e-05, 83037.6562497194, -1.50522954475368e-05, 2.34227740650411e-05, 6.51298577457267e-06, 1.79928510450249e-05, 0, 9.68371203122838e-06, 0, 1.29523820034525e-05, 0, -79716.1499641949, -9.86481865362026e-07, 0, -1.34152704171736e-06, 5.2310724115937e-05, 0, -780553.968690579, -4.99539781271908e-06, 0, -8.0496214634029e-06, -4.06534575598616e-06, -49822.5937422714, -2.25496098267419e-05, 8.2730384464058e-05, 0, -282328.031172531, -1.14478458639852e-05, 0, -8.07825075822027e-06, -8.14683660436062e-06, -1.66181728692413e-05, -282328.031247979, 3.47169776579045e-05, -4.5767701146723e-05, -1.54721955464997e-05, -5.07907682329722e-06, -3.2012223103658e-05, 2.88231244707892e-05, -3.53458674245659e-05, 3.00725497459345e-06, -5.48649717502212e-05, 0, -5.23972068040895e-06, 0, -7.40623040524434e-06, 0, 29893.5562269281, 4.59866182017296e-07, 0, 8.11427293181858e-07, -3.68352645213618e-05, 0, 747338.906205971, 3.64524653122438e-06, 0, 3.67638296568557e-06, 2.49492054106869e-06, 29893.5562473301, 1.5705239020393e-05, -7.18753702533839e-05, 0, 448403.343687385, 5.81331919859327e-06, 0, 1.14915958970586e-05, 1.05681806094604e-05, 1.2457827145235e-05, 448403.343746012, -2.81845701279798e-05, 4.46468623135703e-05, 3.02515974002353e-05, -7.98891838187093e-05, 0, 1.51622528270602e-05, 8.6986298591444e-06, 3.45029241581424e-05, 1.36395554907301e-05, -4.24172785571641e-05, 5.12004800762423e-05, 1.16777358264492e-07, 3.19870741676362e-05, -3.95825492231134e-05, 8.65401754594621e-05, 3.78916530000303e-05, 7.38593246526109e-05, 2.54517674940222e-05, 0, 1.18490114889272e-06, 0, 1.74445077166862e-06, 0, 6.10174442674527e-06, -2.05056974859905e-07, 0, -2.14705064348255e-07, 1.03847425959339e-05, 0, -269042.006236602, -8.67786281631602e-07, 0, -1.07533113103902e-06, -5.0434753601391e-07, 1.10438588854906e-06, -4.51889683581593e-06, 2.47416620563524e-05, 0, -269042.006229589, -2.64299190290124e-06, 0, -1.96040617103894e-06, -3.13926669470819e-06, -3.30828697316989e-06, -269042.006249357, 1.0682630430876e-05, -1.62506144277795e-05, -1.23560434235545e-05, 3.32312501841128e-05, 0, 2.93207390154695e-05, -4.58914813434399e-06, 0, -3.62807195223256e-06, -5.38400957912625e-06, -8.1533977665134e-06, -3.28598965396151e-08, 2.56265346096387e-05, -2.15887198271631e-05, 4.16479300220164e-06, -1.51354360381677e-05, 2.37219892154548e-05, -3.40810334738701e-05, -2.40771822796132e-06, -1.46526906423395e-05, 3.73132696917602e-05, -2.35182622525865e-05, -4.11496313072954e-05, -1.27100883311325e-06, 1.13573196649339e-05, -5.37331627682133e-05, -3.69575554239883e-05, -2.84114833025171e-05}, 
                                                    {0, 0, 0, 0, 0, 1.68605885933301e-07, 0, -485.9999999184, 0, 0, 0, -1.62825882482791e-06, 0, 1032.74999923939, 0, 11020.0499985945, -1.96415325924977e-08, 0, -1.90672919125235e-07, 1032.75000011832, 0, 8.1523913595302e-06, 0, -546.749995880192, 0, -95581.0124897419, -3.61551020927683e-07, 0, 1.61027796572673e-06, -22316.5124914934, 0, 1.30781738108355e-06, -1093.50000107105, -22316.5125010913, -546.75000077609, 0, -2.34497285455919e-05, 0, -1.3310041831505e-05, 0, 419042.868709979, 1.95720161591715e-06, 0, -6.24593442212024e-06, 11235.7124694004, 0, 179320.331207405, -7.07264018041089e-07, 0, -1.00588312644871e-05, 4.30578589183487e-06, 179320.331258906, 3.61839560697678e-06, 22471.4250064711, 9.96526040220696e-06, 11235.7124958459, 0, 4.01514511146634e-05, 0, 2.56278310130063e-05, 0, -1014904.68741127, -5.73288750723731e-06, 0, 1.34431351357615e-05, 6.9101104800632e-05, 0, -699361.59363157, -7.66372597808732e-06, 0, 3.45713177470895e-05, -8.19064170762425e-06, -699361.593784971, -1.10082217349308e-05, -83037.6561513253, 0, 1.51315339125656e-05, -2.32648261172276e-05, -2.65367284843717e-05, -166075.312530417, -83037.6562457235, -3.92693651476854e-05, 2.28235733818343e-05, -1.81905820612175e-05, 0, -4.03811060413107e-05, 0, -2.83723679617877e-05, 0, 1368460.57488829, 7.21558948780825e-06, 0, -1.60121262022571e-05, -9.45353129606811e-05, 0, 1411640.15606793, 1.30006725828013e-05, 0, -5.32123623678996e-05, 8.37172830440968e-06, 1411640.15631503, 1.75675643766916e-05, -0.000135107994535444, 0, 282328.031069648, -6.01552287344e-06, 0, -5.29688435051348e-05, 3.66376170428497e-05, 3.98703981583814e-05, 564656.062579205, 282328.031279832, 8.260733642956e-05, -5.52606003863286e-05, 5.33350328201748e-05, 6.91160412364734e-05, -0.000101528626646889, 8.59671530682673e-05, 5.56498427654068e-06, 4.27619221141146e-05, 0, 2.1999518088337e-05, 0, 1.66190467806918e-05, 0, -956593.799925748, -4.72826716523364e-06, 0, 9.70555175157396e-06, 6.92504198536799e-05, 0, -1404997.14360627, -1.29715201998765e-05, 0, 4.21754439458861e-05, -3.38973228188429e-06, -1404997.14380563, -1.21226809584861e-05, 0.000107558613059179, 0, -448403.343576818, -1.45814343233901e-05, 0, 6.37354845796164e-05, -2.83017286864156e-05, -3.15795810372185e-05, -896806.687594713, -448403.343815189, -8.49470597638906e-05, 2.77279382078426e-05, 0.00013587002495558, 0, 2.70407480920336e-05, -5.22649434936152e-05, -5.85026548008763e-05, -8.50853307473782e-05, 6.88841932579153e-05, -8.54688566828128e-05, 0.00013042597968396, -0.000103315782662051, -3.92056469137135e-05, -5.8820331395098e-05, -0.00010232401626979, -0.000117602684642445, -2.20584354085612e-05, 0, -5.0061437379721e-06, 0, -3.98002108780895e-06, 0, 269042.006229805, 1.30126287919553e-06, 0, -2.2879732983397e-06, -2.05148603110102e-05, 0, 538084.012454565, 3.59190383138907e-06, 0, -1.31327257584761e-05, 4.63569632494812e-07, 538084.012518334, 3.63156317524405e-06, -3.70980411946534e-05, 0, 269042.006184007, 6.88780973380921e-06, 0, -2.90546917827938e-05, 7.84386869628271e-06, 8.04614554353721e-06, 538084.012542671, 269042.006287214, 3.19854899709113e-05, -1.40446815245964e-05, -4.40642682145234e-05, 0, -5.87396630390064e-05, 2.11491593106129e-07, 0, -2.27240228953182e-05, 2.05868159138902e-05, 2.05478937537302e-05, 4.62032354539656e-05, -3.2149810627858e-06, 4.54133079149884e-05, -6.35865051563212e-05, 4.65922476534786e-05, 3.29186052217962e-05, 3.0850682438808e-05, 3.5641295450104e-05, 3.10026508087706e-05, -7.91559866424742e-05, 6.0062810983321e-05, 6.80826628224868e-05, 4.53486502659449e-05, 9.2850586947052e-06, 4.79235512534953e-05, -3.66083004960024e-06, 4.20931943763653e-05}, 
                                                    {0, 0, 1.56930611811915e-08, 0, 0, -4.00795308208767e-07, 0, 1360.7999997151, 2.60462314186371e-08, 0, 0, 3.92356996167014e-06, 0, -4641.29999722598, 0, -30156.2999952913, -4.55678596841708e-07, 0, -2.27880617969855e-07, -4641.30000104306, 0, -1.98578727828284e-05, 0, 5248.79998443418, 0, 253017.674966633, 4.00947357696567e-06, 0, 3.79816128745454e-07, 97521.9749707553, 0, 2.59150873948947e-06, 10497.6000067293, 97521.9750095886, 5248.80001232205, 0, 5.75957196748125e-05, 0, -1968.29994876459, 0, -1061979.86237217, -1.64089343501811e-05, 0, 1.03677073973122e-06, -104319.899888793, 0, -750988.462363034, -2.43038127938011e-05, 0, -7.9643059624292e-07, -5904.90002158125, -750988.462551397, -1968.30006565727, -208639.800052893, -5904.90006530628, -104319.900062548, 0, -9.92424874176294e-05, 0, -9.8672010963879e-05, 0, 2438723.69972027, 3.71552014379624e-05, 0, -6.12470560358754e-06, 36905.6247329128, 0, 2760540.74963397, 8.43336714752718e-05, 0, -2.41866135794047e-05, 3.9216698760955e-05, 2760540.75015484, 0.000186297788228961, 730731.374676649, 0, 4.33813788283364e-05, 110716.875141328, 0.000140472829516201, 1461462.75019788, 730731.375212318, 0.000261045269859045, 36905.6252203123, 110716.875347177, 0, 0.000100293919249063, 0, 0.000108428371163025, 0, -3095643.82465177, -4.48505883841455e-05, 0, 9.73171249600396e-06, 0.000376970474967574, 0, -5154977.69944546, -0.000128461707125196, 0, 5.44064660797118e-05, -3.90144193503727e-05, -5154977.700252, -0.000281540534830882, -232505.437016481, 0, -2291839.31196762, -0.000141691797564451, 0, 6.1216522261727e-06, -0.000196303532814173, -0.000175930722535621, -4583678.62539259, -2291839.31293743, -0.000533761232858313, -0.000384516872098857, -697516.312855238, -0.000347700979056721, -232505.437845382, -0.000683556695449571, -697516.313298457, -0.000920142569146535, 0, -5.48450843339881e-05, 0, -6.28944636225591e-05, 0, 2032761.824771, 2.86551217762321e-05, 0, -6.85511839987087e-06, -0.000277853080054013, 0, 4663394.77456337, 0.000101317945059274, 0, -5.22350535874484e-05, 1.84612460766752e-05, 4663394.77520508, 0.000208039915646533, -0.000419188289950542, 0, 3228504.07451849, 0.000168656656782185, 0, -6.59541846470778e-05, 0.000142574712576394, 0.000122507258241971, 6457008.15038629, 3228504.07546539, 0.000545644785859167, 0.000404310260996558, 597871.12458093, 0, 8.77088524203544e-05, 0.000313700781675367, 0.000235027970359214, 1793613.37544315, 597871.125434372, 0.000443201194724653, 0.000296811523082175, 0.000610194800053277, 1793613.37588939, 0.00109410698202361, 0.00111218730314194, 0.00107828196730299, 0.00138817870753463, 0, 1.25172326036614e-05, 0, 1.49247013077827e-05, 0, -538084.012438358, -7.60787759987881e-06, 0, 1.59064101521411e-06, 8.17847869024858e-05, 0, -1614252.03736178, -2.95118730480764e-05, 0, 1.69372790579074e-05, -3.10585370828242e-06, -1614252.03756481, -6.10737376604637e-05, 0.000151668192887811, 0, -1614252.0373205, -6.43029507190905e-05, 0, 4.48935906248421e-05, -3.84943090274633e-05, -3.23627892484972e-05, -3228504.0751523, -1614252.03769032, -0.000210900960749513, -0.00014698267714152, 0.000152500462044165, 0, -538084.012342524, -6.93499471757671e-05, 0, 3.28695359968757e-06, -0.000109352920187034, -6.69774094896738e-05, -1614252.03769686, -538084.012725117, -0.00022449466799425, -9.28904651166844e-05, -0.000223854615988119, -1614252.0378687, -0.000492519092163224, -0.000178471188313053, -0.000117632042406675, -9.56792220492013e-05, -0.000495469175136176, -0.000427335265049732, -0.000577316467546412, -0.000731855298038627, -0.000563242209713466, -0.000697115761851259, -0.000854756742550549}, 
                                                    {0, 0, -1.99057532449094e-08, 0, 0, 5.27729052007197e-07, 0, -2551.49999937224, -1.9298761052169e-07, 0, 0, -5.2822925745309e-06, 0, 12529.6874938838, 0, 54629.4374898716, 2.38425131824128e-06, 0, 2.77200898279307e-06, 12529.6875040422, 0, 2.7229477877673e-05, 0, -22895.156215716, 0, -436306.499929859, -1.60837529654814e-05, 0, -1.62939229579337e-05, -252530.156187418, 0, -2.36818071531317e-05, -45790.3125233951, -252530.156289156, -22895.1563009364, 0, -8.01370907120853e-05, 0, 18452.8123868041, 0, 1721134.82786042, 5.99245636492028e-05, 0, 5.239855208344e-05, 433128.515386759, 0, 1825803.28096964, 0.000121042149903317, 0, 9.17322640265781e-05, 55358.4375687822, 1825803.28143607, 18452.812770098, 866257.031453955, 55358.4377225175, 433128.515961971, 0, 0.000139662446241483, 0, -5535.84353241714, 0, -3686871.93692796, -0.00012481470660932, 0, -9.68123298445137e-05, -326614.780663125, 0, -6166929.93676675, -0.000337284233431614, 0, -0.000192433152457643, -22143.3751216378, -6166929.93799111, -5535.84450717092, -2801136.93686857, 0, -0.000334010876577377, -979844.344246502, -33215.0629411196, -5602273.87573762, -2801136.93860012, -22143.3758644991, -326614.782460849, -979844.345215209, 0, -0.000142394607601925, 0, -0.00023777143973443, 0, 4367780.71804613, 0.000145805490982366, 0, 0.000105341803589625, 91341.4210299072, 0, 10412922.0926386, 0.000490533100484076, 0, 0.000233323488400529, 0.000121855284053718, 10412922.0944697, 0.00113756306634176, 1860043.49903268, 0, 7813843.4521796, 0.000681045101029679, 0, 0.000422681304979057, 365365.688142067, 0.00051046146026946, 15627686.9075719, 7813843.45510002, 0.00174205257825259, 91341.4241023476, 5580130.50130562, 0.0010529779201697, 1860043.50244429, 548048.533806038, 5580130.50373269, 365365.691303031, 0, 7.84074613199561e-05, 0, 0.000136886862811959, 0, -2690420.06204274, -9.14209146425417e-05, 0, -6.02170499966616e-05, 0.000628387868789078, 0, -8519663.53036756, -0.000367462259845934, 0, -0.000154244880856328, -6.42615116331969e-05, -8519663.53179349, -0.000854033791257793, -448403.342864537, 0, -9416470.21792682, -0.00066676731431294, 0, -0.000250231476682932, -0.000433034783756587, -0.000331132689822548, -18832940.4386633, -9416470.22055824, -0.00177177240436244, -0.00213393209473084, -4035630.09305241, 0, -0.0005829646404082, -1793613.3760791, -0.000616787679251008, -12106890.2827467, -4035630.09639787, -0.00130763066997979, -448403.346188917, -0.00208494782056024, -12106890.2853383, -0.00437535011489872, -2690420.06701828, -0.00399580427724654, -1793613.38118194, 0, -1.79928051568798e-05, 0, -3.22255384796876e-05, 0, 672605.015503455, 2.34796067603405e-05, 0, 1.47992557082306e-05, -0.000185079179148395, 0, 2690420.06221849, 0.00010738970422055, 0, 4.47921228108698e-05, 1.32707725829358e-05, 2690420.06266388, 0.000252085799077345, -0.000334499238158226, 0, 4035630.09343991, 0.000242780841522463, 0, 5.19989549304259e-05, 0.000113390755819297, 8.94973791619864e-05, 8071260.18790926, 4035630.09439977, 0.000689546289361032, 0.000798152710075088, 672605.015357254, 0, 2690420.06228183, 0.000299806707377114, 0, 0.000160044456212003, 0.000345884775170584, 0.000158514729229741, 8071260.18811644, 2690420.06364743, 0.00065486913515388, 0.000976912334979355, 0.000689482332424134, 8071260.18911459, 0.00195232162982339, 2690420.06309785, 0.000301233357675933, 672605.016729675, 0.00186714882922668, 0.00143244646642798, 4035630.09621598, 2690420.06582985, 0.00208525341490894, 0.00311012370968282, 0.00355840491333923}, 
                                                    {0, 0, 1.18986309776541e-08, 0, 0, -3.46536194274537e-07, 0, 3401.99999913794, 4.26632219318324e-07, 0, 0, 3.63667247305918e-06, 0, -22829.8499919155, 0, -68756.8499861332, -4.892136716563e-06, 0, -6.49726305812099e-06, -22829.8500084266, 0, -1.94716674188216e-05, 0, 60598.1249561376, 0, 506582.099905045, 3.1015718403015e-05, 0, 4.14150915983984e-05, 429399.224917256, 0, 5.13255444232538e-05, 121196.25004741, 429399.225086091, 60598.1251049904, 0, 5.90212256441098e-05, 0, -79552.1248563132, 0, -1822235.73714566, -0.000111735634264434, 0, -0.000142470365761473, -1056731.06220193, 0, -2799414.67463255, -0.000236471359163706, 0, -0.00022315459302723, -238656.37513543, -2799414.67540064, -79552.1255497509, -2113462.12544466, -238656.375435054, -1056731.0632666, 0, -0.000105192168986597, 0, 51667.8747241628, 0, 3579845.62424231, 0.000227775061866293, 0, 0.000281650896880723, 1280625.18677091, 0, 8336980.68653216, 0.00063219633495458, 0, 0.000540339467071806, 206671.500232069, 8336980.68850638, 51667.8765295736, 5986092.37426764, 0, 0.000697336130363135, 3841875.56354151, 310007.250839787, 11972184.7516307, 5986092.37754497, 206671.501628397, 1280625.19023004, 3841875.56565967, 0, 0.000109082090779086, 0, -13286.0246989937, 0, -3932663.39907936, -0.000264267904304077, 0, -0.000319917803754201, -763946.436436032, 0, -12522078.5610079, -0.000910770742965254, 0, -0.000748089379555391, -66430.1252340947, -12522078.5638952, -13286.0273000528, -6144786.56143855, 0, -13983541.3114694, -0.00130545600442777, 0, -0.00101905391633494, -3055785.75127474, -132860.250941043, -27967082.6278692, -13983541.3169569, -66430.1282297671, -763946.442520586, -18434359.6903388, -132860.251937526, -6144786.56834425, -4583678.63034913, -18434359.6958849, -3055785.75793327, 0, -6.08437564529601e-05, 0, -0.000172799461982918, 0, 2271910.27441038, 0.000163863957007372, 0, 0.000190703500416336, 179361.336698669, 0, 9267002.43629713, 0.000673620548097236, 0, 0.000555969252773615, 0.000128890152577855, 9267002.43849968, 0.0017450801723255, 2989355.62399735, 0, 14348906.9990902, 0.00123788892405104, 0, 0.000770202386079498, 896806.688303296, 0.000589828087890281, 28697814.0024389, 14348907.0038944, 0.00325273424153179, 179361.342247139, 10163809.1244187, 0, 0.00116077639161715, 11957422.5022367, 0.00109561891943127, 30491427.3782588, 10163809.131287, 0.00230264197404948, 2989355.63099828, 1793613.37920637, 30491427.3843827, 896806.696423557, 17936133.7598631, 1793613.38311945, 11957422.5133607, 0, 1.41015216894625e-05, 0, 4.05258844835008e-05, 0, -538084.012345667, -4.16532254143088e-05, 0, -4.72571765740332e-05, 0.000237799689988795, 0, -2690420.06211287, -0.000196762566781696, 0, -0.000169669556978437, -2.92722766177229e-05, -2690420.06279004, -0.000519585774105196, -538084.012102204, 0, -5380840.12463657, -0.000447258378558432, 0, -0.000244763030404925, -0.000203241896526996, -0.000158066490542411, -10761680.2508129, -5380840.12632785, -0.00126477504897936, -0.0017733262331833, -2690420.06228777, 0, -5380840.1248704, -0.000546677631408873, 0, -0.000372857192708769, -2690420.06316778, -0.000269707639885615, -16142520.3763383, -5380840.12764477, -0.00114087234298436, -538084.014961967, -0.00131634016314224, -16142520.3787297, -0.00393969119820604, -10761680.2512754, -0.000514496568208939, -2690420.06532208, -5380840.12892007, -0.00286684272593424, -16142520.3805406, -10761680.2573285, -0.00413299058105989, -2690420.06915968, -5380840.13254704}, 
                                                    {0, 0, 0, 0, 0, -6.02635984984457e-08, 0, -3401.99999931257, -4.67556429966167e-07, 0, 0, 2.94413645323234e-07, 0, 30484.3499940839, 0, 61102.3499887786, 5.03963869906051e-06, 0, 7.42601019469259e-06, 30484.3500101734, 0, -1.80137771245608e-07, 0, -111965.287470057, 0, -386324.437422672, -3.06950171083036e-05, 0, -4.98602575073779e-05, -498289.724938773, 0, -5.46323202297541e-05, -223930.575057212, -498289.725109161, -111965.287623586, 0, -2.61622918560192e-06, 0, 215897.906152452, 0, 1223585.49346145, 0.00010854971968399, 0, 0.000178040297332997, 1655381.30606214, 0, 2663068.89346835, 0.000233021627714541, 0, 0.000258653112536273, 647693.718910278, 2663068.89425943, 215897.906886659, 3310762.61306892, 647693.71925667, 1655381.30721485, 0, 8.78618757652369e-06, 0, -230660.156059656, 0, -2173741.31188721, -0.000220336884344403, 0, -0.000362440309801579, -2865721.78080914, 0, -6751884.09296376, -0.000600495302360856, 0, -0.000674860551721596, -922640.625264893, -6751884.09500575, -230660.158004699, -7213204.40584672, 0, -0.000736140667981161, -8597165.34504882, -1383960.93847649, -14426408.8146386, -7213204.40951762, -922640.626827843, -2865721.78465306, -8597165.34767766, 0, -1.22329119704098e-05, 0, 129538.743538743, 0, 2205480.14926493, 0.000256635892050173, 0, 0.000418938097121601, 2723635.12433148, 0, 8951459.34247719, 0.000852422270890422, 0, 0.000990064690676102, 647693.719012567, 8951459.34544203, 129538.746384063, 9599153.06207598, 0, 13751035.8744412, 0.00128962087634239, 0, 0.00118373491109967, 10894540.5015146, 1295387.43857699, 27502071.7537991, 13751035.8806773, 647693.722288969, 2723635.13120704, 28797459.1911664, 1295387.43968675, 9599153.07008244, 16341810.7565728, 28797459.1982228, 10894540.5095683, 0, 8.10520217808212e-06, 0, -29893.5561271863, 0, -1195742.24953735, -0.000159372426943569, 0, -0.000253888475965112, -1345210.03072191, 0, -6008604.80519331, -0.000623121658402006, 0, -0.000763860876085406, -179361.337644787, -6008604.80742533, -29893.5582562516, -6277646.81207486, 0, -12106890.2806655, -0.00118897291624256, 0, -0.00100038329551397, -6726050.1571467, -448403.344404105, -24213780.5657225, -12106890.2860982, -179361.341013212, -1345210.03706363, -12256358.0625217, 0, -0.00120676534805498, -25110587.2527684, -597871.126263955, -36769074.1918178, -12256358.0706975, -448403.346221077, -6277646.82038824, -13452100.3175574, -36769074.1997511, -6726050.16678379, -37665880.8874668, -13452100.3222521, -25110587.2665635, 0, -2.0999031547701e-06, 0, -2.90454413948722e-05, 0, 269042.006131372, 4.07096303207728e-05, 0, 6.32694295136821e-05, 269042.006087351, 0, 1614252.03715535, 0.000180981752422945, 0, 0.000236571902864343, 3.39767323624643e-05, 1614252.03783191, 0.000600474513214422, 1614252.03729933, 0, 4035630.09347355, 0.000423851726086726, 0, 0.00035472042653375, 1614252.03771688, 0.000168674304817981, 8071260.18855688, 4035630.09535467, 0.00135966286496128, 269042.008406167, 4035630.09380978, 0, 5380840.12507623, 0.00053254934641471, 0, 0.000427652250921986, 8071260.18827303, 0.000314024863374299, 16142520.3768081, 5380840.12840394, 0.0011895325831487, 1614252.04075311, 4035630.0952704, 16142520.3799423, 1614252.04207775, 16142520.3766411, 0.000565487847525603, 4035630.09755029, 16142520.3798317, 5380840.12845781, 24213780.5696761, 16142520.3842419, 4035630.09858792, 8071260.19566742, 16142520.3842818}, 
                                                    {0, 0, -3.62129342754467e-08, 0, 0, 8.20930616389487e-07, 0, 2551.49999932383, 4.95941369090336e-07, 0, 0, -7.42710391678869e-06, 0, -40085.8874948808, 0, -40085.8874886613, -4.66260827042873e-06, 0, -8.64148138119048e-06, -17122.3875071095, 0, 3.49315756592644e-05, 0, 199550.081227924, 0, 199550.081171035, 2.34129095869852e-05, 0, 6.12221853588452e-05, 605771.662444144, 0, 6.09238799220813e-05, 244998.675034269, 244998.675071593, 45448.5938430617, 0, -9.50542979650446e-05, 0, -468701.437429517, 0, -468701.437205862, -7.2841765171588e-05, 0, -0.000222065392061264, -2793017.69988178, 0, -2793017.69971982, -0.000211136044002754, 0, -0.000339204236652661, -997066.968828956, -997066.969080638, -59664.0942643363, -3381047.32534366, -588029.625331611, -588029.625661931, 0, 0.00015538790528636, 0, 575727.749857661, 0, 575727.749384558, 0.000137717126407835, 0, 0.000452270487085397, 5984247.09350862, 0, 5984247.09288551, 0.000449978906245747, 0, 0.000968607391868851, 1765934.15635628, 1765934.1570675, 38750.907731234, 10817038.6873329, 0, 0.000845812787332603, 12660474.6568981, 1843435.96927882, 12660474.6576493, 1843435.97086265, 691980.47005329, 691980.47125027, 7368208.03370097, 0, -0.000150552336931777, 0, -358722.674836678, 0, -358722.674281672, -0.000157079999806611, 0, -0.000517153317660144, -6626404.96833479, 0, -6626404.96725208, -0.000558205655560152, 0, -0.00146446854029574, -1444855.21884961, -1444855.21984749, -9964.52102921828, -18085601.5314099, 0, -18085601.5309963, -0.00119026582899264, 0, -0.00166942738356202, -20277795.656776, -2192194.12547289, -20277795.6589031, -2192194.12860526, -398580.752705797, -398580.754923855, -37665880.8771792, -1494677.81385176, -1494677.81751978, -21074957.1598577, -21074957.1631115, -7822147.22502594, 0, 7.96366453881013e-05, 0, 89680.6686530013, 0, 89680.6683149816, 9.65785817071248e-05, 0, 0.000309538835009588, 3676907.41837646, 0, 3676907.41746215, 0.000376658104158328, 0, 0.00113046858607151, 448403.343811955, 448403.344496985, 0.00176954668064136, 13900503.6563398, 0, 13900503.6557696, 0.000914714358459532, 0, 0.00160683994373211, 14797310.3438985, 896806.687738903, 14797310.3460768, 896806.690555559, 89680.6715319462, 89680.6736271679, 20626553.8131833, 0, 0.00141250715727686, 42149914.3137765, 896806.688103514, 42149914.3154604, 896806.692598619, 448403.345423575, 448403.349362371, 22420167.1897791, 22420167.1957759, 4035630.1010957, 43046721.0074173, 15245713.6931453, 15245713.698101, 0, -1.7708174426225e-05, 0, 2.32003158645455e-05, 0, 0.000106243519970634, -2.46352244899636e-05, 0, -7.59938718056944e-05, -807126.018624794, 0, -807126.018324601, -0.000105172992587669, 0, -0.000343955442390442, -1.85832011816807e-05, -0.000207762526502088, -0.00053543107264205, -4035630.09370623, 0, -4035630.09344557, -0.0002851171364042, 0, -0.000606781244358235, -4035630.09375231, -5.67584522431807e-05, -4035630.09451135, -0.00100287998133574, -0.00109867270054329, -0.00187160017361109, -8071260.18787745, 0, -8071260.18781195, -0.00049838594253615, 0, -0.000630099279383255, -16142520.3752022, -0.000107036192472705, -16142520.3763611, -0.00207141169031846, -0.000869539716357962, -0.00252913156179254, -8071260.18804589, -8071260.19096249, -807126.022143709, -24213780.563515, -0.000316195160424701, -0.00244032905162008, -24213780.5649524, -8071260.1891793, -24213780.5672337, -8071260.19349272, -4035630.09673865, -4035630.09910437, -16142520.3805078}, 
                                                    {0, 0, 0, 0, 0, -2.30773843478752e-07, 0, -291.600000015339, 0, 0, 0, 2.11447443238404e-06, 0, 6087.150000514, 0, 3244.05000002342, -7.06553799047029e-08, 0, 9.38907472854607e-08, 619.650000068643, 0, -1.00767871119844e-05, 0, -46391.7375038129, 0, -5904.89999814287, 7.87183294290201e-07, 0, -9.80697901191814e-07, -67058.8875053905, 0, -9.66367611703702e-08, -12274.5375001723, -6232.95000119177, -328.050000947439, 0, 2.78000413728849e-05, 0, 167920.593760324, 0, 2952.44998836891, -2.41834541716645e-06, 0, 4.27094635945531e-06, 503023.668797, 0, 116621.775006041, -4.90820676302381e-06, 0, 3.08616260131084e-06, 85498.0312475834, 5904.90000610586, 4.71949161360725e-06, 122772.712506616, 6150.93750147422, 2952.45000708699, 0, -4.60549803074119e-05, 0, -306685.743762792, 0, 2.78369803544517e-05, 4.82834432313086e-06, 0, -8.64337115113949e-06, -1777005.84387507, 0, -55358.4374512489, 2.27503058766034e-05, 0, -1.65435144140391e-05, -265720.499991834, -1.91991961053838e-05, -1.53639496439779e-05, -808233.187644855, 0, 5.8023785479387e-06, -846984.093755708, -38750.9062464062, -110716.875039169, -1.58670622840486e-05, -1.25190958767078e-05, -1.23873158301825e-05, -55358.4375365005, 0, 4.51711698522787e-05, 0, 269042.006256778, 0, -3.13391647258147e-05, -5.76255499893406e-06, 0, 1.08754094894578e-05, 3128858.88764924, 0, -0.000147889285134351, -3.43786093884203e-05, 0, 2.7411966232634e-05, 368687.193734991, 3.01985531893621e-05, 2.49058749342476e-05, 2491129.68785419, 0, 348758.156388355, -4.19896360947435e-05, 0, 9.59780228795305e-06, 2590774.87499117, 99645.187474165, 8.97776430774532e-05, 2.15212757362913e-05, 1.67552641891768e-05, 2.65953589952425e-05, 697516.312562349, 8.4379995416423e-06, 4.14774873411807e-05, 348758.156289049, 0.000152858855192333, 8.44559028203323e-05, 0, -2.41554326734474e-05, 0, -89680.6687504832, 0, 1.60628460381907e-05, 2.3450960785184e-06, 0, -6.05778029658207e-06, -2600739.39383219, 0, 0.000150603139419804, 2.46002382540011e-05, 0, -2.27060541457063e-05, -179361.337489675, -2.42386898441375e-05, -1.90529846359578e-05, -3407865.41281886, 0, 3.7380405083137e-06, 5.09774994556405e-05, 0, -1.86695114445894e-05, -3497546.08123229, -89680.6687094916, -8.56770589704639e-05, -2.0624076301609e-05, -2.22152977017713e-05, -7.0787479674239e-06, -896806.687837413, 0, 1.99595351463322e-05, -1793613.37501761, 1.41017080150178e-05, -0.000102822243309179, -1.18094753588935e-05, -2.18492249795884e-05, 2.37819859281815e-05, -896806.687509704, -0.00016727264464031, -6.72098499344203e-05, -0.000124435081926019, -0.00010511467410763, -0.000161854645335415, 0, 5.42260131797124e-06, 0, -5.22177674524146e-07, 0, -2.78024553438311e-06, -5.41964556560365e-07, 0, 1.49156530575059e-06, 807126.018766637, 0, -5.25139330927289e-05, -6.6095198582022e-06, 0, 7.31224295676496e-06, -2.69502316700696e-06, 8.05505176321539e-06, 6.08406174771754e-06, 1614252.03759904, 0, -4.19067716419981e-05, -1.79897192678488e-05, 0, 1.04683961434222e-05, 1614252.03749214, -1.82271577307725e-05, 3.07967367409071e-05, 8.8832990055329e-06, 1.117461533003e-05, 1.04556398819845e-05, 807126.018922603, 0, 9.32029252217778e-05, -1.39943063332401e-05, 0, 2.11767518898184e-06, 1614252.03749084, -9.19663497407315e-06, 5.39018430210853e-05, 2.77562779112035e-06, 1.60805043050424e-06, -5.72072788204187e-06, 807126.018739359, 7.65533147810693e-05, 2.59742291808344e-05, 2.91722587540055e-05, 1.42659766088094e-06, 5.75625320957249e-06, 4.99393211253352e-05, 1.52197642438659e-05, 0.000123298116800283, 9.99021200470477e-05, 5.79157130083265e-05, 3.72994680357323e-05, 0.000114825006604503}, 
                                                    {0, 0, 0, 0, 0, 3.95854976397215e-08, 0, -291.599999841273, -3.81650291708223e-08, 0, 0, -4.7433494245132e-07, 0, 3244.04999832375, 0, 6087.14999753952, 3.84846969643917e-07, 0, 6.73806377658963e-07, 619.65000041009, 0, 2.74102918313082e-06, 0, -5904.89999013685, 0, -46391.7374834873, -1.89422635408477e-06, 0, -4.37685132845679e-06, -67058.887483506, 0, -5.60637124271555e-06, -6232.95000260775, -12274.5375019753, -328.050006687808, 0, -8.69327047958551e-06, 0, 2952.44996687082, 0, 167920.593689148, 6.23076186403699e-06, 0, 1.48532000291793e-05, 116621.774931164, 0, 503023.668680559, 2.15964638555952e-05, 0, 2.86268986159921e-05, 5904.90000795158, 85498.0312536627, 3.99570690881369e-05, 122772.712515264, 2952.45003040106, 6150.93753005202, 0, 1.59360280984941e-05, 0, 6.35596115560152e-05, 0, -306685.743620874, -1.09890764530913e-05, 0, -2.76456885132832e-05, -55358.4373209668, 0, -1777005.84357872, -4.46646692706127e-05, 0, -7.72285589494228e-05, -1.62297909869572e-05, -265720.500002386, -0.000118247926457004, -808233.187325206, 0, -8.03582293429803e-05, -110716.875036136, -6.48193608435182e-05, -846984.093785259, -38750.9062927277, -0.000134116894220522, -0.000148371393557774, -55358.4376643887, 0, -1.68317995134434e-05, 0, -6.89038289700206e-05, 0, 269042.006093769, 1.20971351443021e-05, 0, 2.97977784078603e-05, -0.000260948651219782, 0, 3128858.88724999, 5.96414159339066e-05, 0, 0.00010633345893699, 1.76836531198883e-05, 368687.193751941, 0.000179972130025893, 348758.155946069, 0, 2491129.68726486, 0.00012843278574757, 0, 0.000151989827575242, 4.76511596374469e-05, 7.90740311171436e-05, 2590774.87504517, 99645.1875216898, 0.000272757134199221, 0.000327138473725764, 697516.312582484, 0.000169034353066999, 0.000178690956593331, 0.000266297368863442, 348758.156568513, 0.000518912913244256, 0, 9.50708547979785e-06, 0, 3.93078121929699e-05, 0, -89680.6686501195, -7.41866259334071e-06, 0, -1.65140784920959e-05, 0.00019272070251754, 0, -2600739.39355546, -4.08788010713678e-05, 0, -7.28055326250793e-05, -1.10772265620352e-05, -179361.337505851, -0.000132738394289291, 0.000293127038032993, 0, -3407865.41231947, -9.07654064818495e-05, 0, -0.000137315648643864, -3.91579837883653e-05, -5.36986179421473e-05, -3497546.08127515, -89680.6687624453, -0.000277171403725376, -0.000332748281227543, -896806.687287245, 0, -0.000155352888666413, -6.84544821292936e-05, -0.000114913085085458, -1793613.37508841, -6.34707185980211e-05, -0.000234792846983424, -0.000240157160353377, -0.000211895963362181, -896806.687714655, -0.000669496484231456, -0.000381434279964433, -0.000533505884782632, -0.000660739885774449, 0, -2.22313817984108e-06, 0, -9.17915525730699e-06, 0, -2.61369650517018e-05, 1.94432892284051e-06, 0, 3.86590320025796e-06, -5.60459356289046e-05, 0, 807126.018688575, 1.17675732803715e-05, 0, 2.11445551767992e-05, 2.47079743876944e-06, 1.53870893577772e-06, 3.82264063503368e-05, -0.000111017418191849, 0, 1614252.03743794, 3.09179867515754e-05, 0, 4.55480582026731e-05, 1.06587408825942e-05, 1.66695871462564e-05, 1614252.03750526, 4.36513473781014e-06, 0.000107484293434974, 0.000136721416055703, -9.57940222179098e-05, 0, 807126.01869303, 5.77239200068346e-05, 0, 6.42835670587109e-05, 2.18465893072145e-05, 2.90038219040264e-05, 1614252.03752597, -6.35865965659118e-06, 0.00010798581623172, 0.000125996663192182, 7.69444871499133e-05, 807126.018794276, 0.000305583892305085, 3.11482318733947e-05, 5.81404610084155e-05, 2.5000250325479e-05, 0.000169159116858404, 0.000183403062528879, 0.000182380468146195, 0.000274191450824885, 0.000342236081920844, 0.000399681159556068, 0.000406074655822172}, 
                                                    {0, 0, 4.02793064913757e-08, 0, 0, -9.38836262364186e-07, 0, -1700.99999968136, -2.00740517124385e-07, 0, 0, 8.65100645300395e-06, 0, 31316.6249985845, 0, 23662.1249940378, 1.39095745401148e-06, 0, 3.83708565898031e-06, 8353.12500232531, 0, -4.1420677771241e-05, 0, -196921.124999201, 0, -90441.5624551584, -4.95544152857403e-06, 0, -2.9328571411247e-05, -425143.687480229, 0, -2.09110207021239e-05, -143294.062507993, -105705.000022487, -15263.4375332625, 0, 0.000114605369479736, 0, 556659.843748445, 0, 149672.812326934, 1.22458253806992e-05, 0, 0.000110228640474209, 2559610.12503661, 0, 1532608.59360494, 4.41865059476169e-05, 0, 0.000142713812509853, 787320.000003092, 311647.500110132, 12301.8752002515, 1775570.62509784, 242962.031348486, 174276.562707462, 0, -0.000190147586081606, 0, -787935.09373489, 0, -114407.437137176, -2.33493589701799e-05, 0, -0.000229260474177969, -6759265.21891055, 0, -2367495.8431751, -4.98458420734559e-05, 0, -0.000450221281160815, -1749326.62497075, -346912.875288829, -3690.56310258169, -8224418.53143928, 0, -0.00026530638184933, -9362957.06258883, -1138538.53132307, -4915829.25050319, -354294.000637643, -180837.562954876, -125479.12584317, -2729170.9694671, 0, 0.00018660292272265, 0, 548048.531219464, 0, 33215.0620860696, 3.12267510131989e-05, 0, 0.000263815995857605, 8785384.03140959, 0, 1677360.6551389, 1.83390622847033e-05, 0, 0.000719225847543274, 1693968.18745488, 132860.250401606, 0.000954496342339115, 17537553.000817, 0, 10944363.0938613, 0.000229399989690973, 0, 0.000654399253249112, 19331166.3748526, 1793613.3749375, 5081904.56359278, 199290.376095396, 49822.5947823408, 33215.064259722, 22586242.5006591, 697516.312859285, 132860.251475625, 12339395.7194929, 5181549.75221638, 1826828.43949498, 0, -9.97824481509187e-05, 0, -149467.781226046, 0, 0.000239952989763967, -2.15667989071827e-05, 0, -0.000158857835485201, -5530307.9062904, 0, -448403.342754048, 1.70121594020811e-06, 0, -0.000573246440059101, -597871.124976184, -0.000275209086533276, -0.000762007560965954, -16142520.3757394, 0, -6427114.59345872, -9.38947602955605e-05, 0, -0.00071022029582785, -17039327.0622248, -896806.687381706, -1793613.37602786, -0.000922464455922295, -0.00109384638592844, -0.00183920716592299, -16740391.5010431, 0, -0.000403424608246793, -34078654.125128, -597871.12503285, -19430811.5637232, -0.00143703293436706, -149467.781781854, -0.00175764304368737, -17936133.7500302, -2690420.06554653, -448403.346260272, -19729747.1271062, -6875517.93894781, -1793613.37841797, 0, 2.23907931626747e-05, 0, -6.71548377669375e-06, 0, -5.49142532828102e-05, 6.28095411120343e-06, 0, 3.90259523692202e-05, 1345210.03124189, 0, -0.000332916269286411, -1.73531000027709e-06, 0, 0.000176691912210548, -2.33205587562151e-06, 7.85549598663528e-05, 0.000235863010063419, 5380840.12520277, 0, 1345210.030993, 4.38074611078304e-06, 0, 0.00028737109275364, 5380840.12489326, -4.81478494790098e-05, 0.000346164771101669, 0.000298231026746191, 0.000439734945182325, 0.000718897135248778, 8071260.18806146, 0, 5380840.12535291, 8.9440722942277e-05, 0, 0.000226471068765989, 16142520.3748724, -4.48180218271663e-05, 5380840.12564829, 0.000579824212386051, 0.000309162808691208, 0.000833824053006983, 8071260.18733596, 0.00136711322336152, 0.00121732058565578, 16142520.3753674, 6.94453670871904e-05, 0.000678199721835102, 16142520.3753723, 5380840.12521732, 8071260.18919726, 0.00205052929399322, 1345210.03213075, 0.00175999243874308, 5380840.12656401}, 
                                                    {0, 0, -2.32317721929101e-08, 0, 0, 5.44797990636723e-07, 0, 850.499999946518, 3.11840347180292e-08, 0, 0, -5.03160854612965e-06, 0, -16934.0625005446, 0, -10555.3124984862, 8.31438831391051e-08, 0, -8.2775010410956e-07, -2900.81250039541, 0, 2.41562298004135e-05, 0, 120034.406256846, 0, 29387.8124854671, -2.10252836767407e-06, 0, 7.29736488740979e-06, 206830.968752888, 0, 1.83982752822123e-06, 54424.4062505878, 32668.3125035502, 3280.50000663085, 0, -6.70545455121864e-05, 0, -393660.000020781, 0, -30754.6874377743, 8.88794911200062e-06, 0, -2.90219220366051e-05, -1427222.53133306, 0, -547638.468716766, 1.39943579166869e-05, 0, -2.87970861006715e-05, -346912.874992738, -62739.5625178254, -1230.18754381703, -605457.281267459, -57818.8125160612, -33215.062532715, 0, 0.000111646786090759, 0, 638467.312527665, 0, 11071.6873658452, -1.71827443148239e-05, 0, 6.11041757723753e-05, 4487724.00024764, 0, 540667.406012966, -6.85323600560897e-05, 0, 0.000110823729809586, 944783.999971082, 33215.0625490517, 0.000137424210202339, 3448830.65650351, 0, 1.20751057689085e-05, 3775445.43748852, 326614.781236699, 1101632.9063676, 33215.0625776954, 20298.0938481063, 11071.6876555957, 581263.59388861, 0, -0.000109955823274338, 0, -498225.937516656, 0, 0.000150716307536571, 1.67877313101856e-05, 0, -7.0892271827753e-05, -6842302.87530311, 0, -182682.843220955, 0.000118015132166069, 0, -0.000189261101951441, -1096097.062459, -7.0666733782607e-05, -0.00022370929732442, -9200572.31324152, 0, -3039178.21894139, 8.5715189247229e-05, 0, -0.000119947533981409, -9898088.62488325, -697516.31241302, -548048.531535437, -9.53021023645308e-05, -0.000225490334962965, -0.000341417364048301, -6178001.62514015, -99645.1875515017, -0.000216335654634178, -3238468.59381816, -548048.531715998, -182682.844170244, 0, 5.8998273398e-05, 0, 149467.781253049, 0, -8.31422682422335e-05, -7.98863464020057e-06, 0, 4.27817628978066e-05, 4932436.78141726, 0, -0.000495257720013655, -9.20393347955325e-05, 0, 0.000152687968693715, 448403.343725252, 5.26741745608347e-05, 0.000180658581419705, 10313276.9069392, 0, 896806.687372343, -0.00013629360044499, 0, 0.000168719757481731, 10761680.2498538, 448403.34364438, 0.000278449289184791, 5.02741610748377e-05, 0.000241696971038007, 0.000362786048548551, 6427114.59448907, 0, 9.08196721295934e-06, 13003696.9687072, 149467.781220391, 2690420.062848, 0.000144219375545117, 0.000106752828281698, 0.000279260676146309, 6726050.15609825, 0.00064935084029988, 0.000552814570391593, 2690420.06289998, 896806.687782629, 0.000701938926299122, 0, -1.32813540195e-05, 0, 4.97522017780729e-07, 0, 1.73884936179489e-05, 1.33577991600856e-06, 0, -1.05144361433752e-05, -1345210.03128371, 0, 0.000167429384941509, 2.69254652096012e-05, 0, -4.73442804861935e-05, 5.03553368111875e-06, -1.6558738043605e-05, -5.73102408950831e-05, -4035630.09396391, 0, 0.00014564307705863, 6.03717357335796e-05, 0, -7.34573818276891e-05, -4035630.09369706, 4.26909388760205e-05, -9.66980309411897e-05, -6.52660553757467e-06, -9.84141695524176e-05, -0.000145906940725723, -4035630.09414751, 0, -1345210.03145767, 3.40566964776394e-05, 0, -3.80144265557602e-05, -8071260.18741748, 3.3316249294571e-05, -0.00019602234375956, -3.01746365813357e-05, -5.59568248012505e-05, -0.000137419239719123, -4035630.09363043, -0.000293518051821931, -0.000261257921522901, -4035630.09383931, -7.30405504221613e-06, -7.53328132698048e-05, -4035630.09375968, -1345210.03125295, -0.000410393185201222, -0.000422001703539278, -0.000206539676251552, -0.000370961065041459, -0.000330895964034852}, 
                                                    {0, 0, 0, 0, 0, -1.13252206539317e-07, 0, -242.999999910583, -1.28402431433424e-08, 0, 0, 1.06820687556786e-06, 0, 4525.87499955005, 0, 3796.87499828112, 2.56457241869073e-08, 0, 3.35113539105183e-07, 516.375000053882, 0, -5.22606802116049e-06, 0, -28476.5624995519, 0, -17085.9374874801, 2.87514471118e-07, 0, -2.7469806148013e-06, -70166.2499923587, 0, -6.19662152047308e-07, -9066.93749917605, -7517.81249967312, -273.375001938129, 0, 1.47269321808975e-05, 0, 75861.5625005995, 0, 24603.7499532579, -2.71194855371678e-06, 0, 1.03791206501171e-05, 435281.343745325, 0, 310212.281200794, -2.39604982505627e-06, 0, 1.13070429736626e-05, 50847.749994207, 28294.3124986904, 1.57170146416532e-05, 131425.031246041, 4510.68750404262, 3690.56250712065, 0, -2.48142006997734e-05, 0, -84882.937501317, 0, -11071.6874036614, 5.08433041239634e-06, 0, -2.25888068106789e-05, -1131157.40626431, 0, -428105.24983457, 8.6443348820193e-06, 0, -3.73360003645327e-05, -107026.312480332, -22143.3749911875, -4.79743260015637e-05, -1861888.78123035, 0, -1.08415233278188e-05, -730731.374959843, -22143.3749795431, -488999.531253126, -11071.687525878, -2.93106163511518e-05, -6.55508388735665e-05, -60894.2812776346, 0, 2.46736249482489e-05, 0, 33215.062500609, 0, -0.000109762140895366, -3.46924191780301e-06, 0, 2.4707220048708e-05, 1212349.78127234, 0, 182682.843460285, -2.14525137738855e-05, 0, 6.74689995161028e-05, 66430.1249787622, -2.03244982072684e-05, 7.95956447564247e-05, 4550463.56258324, 0, 2358269.43741645, 5.66193146080358e-06, 0, 4.70316068152244e-05, 1511285.34363442, 33215.0624483054, 365365.687524786, 2.91329980408061e-05, 8.85477027656495e-05, 0.000123570820153237, 2657204.99995369, -1.06422010751553e-05, 6.13966038020322e-05, 298935.562442273, 182682.843828009, 0.000120249779639634, 0, -1.33433222662952e-05, 0, 1.47285323436154e-07, 0, 6.43885131387603e-05, 2.76450729503636e-06, 0, -1.58655610976961e-05, -448403.343760265, 0, 0.000246475061643439, 2.05403784666231e-05, 0, -5.40587128508032e-05, 1.20793543914575e-05, 2.01310131925383e-05, -6.82618850957198e-05, -4334565.65634629, 0, -896806.687352432, 8.52863470262298e-06, 0, -6.52229579766654e-05, -896806.687380297, 3.90792590649471e-05, -2.34566879069359e-05, 7.843977280292e-06, -8.76274752287001e-05, -0.000161998320567383, -4782969.0001069, 0, -2.51364784924926e-05, -5231372.34363032, 2.62814414636253e-05, -1793613.37504384, -9.00383399672539e-05, 3.35314756574417e-06, -0.000185909147753471, -448403.343563512, -0.000203109342450368, -0.00020962827364369, -896806.687555159, 2.95007157268833e-05, -0.000239551748676364, 0, 3.02340511009621e-06, 0, -1.32040743486632e-07, 0, -1.5101694267049e-05, -3.18602104206811e-07, 0, 3.88362940517558e-06, 1.10915268765093e-06, 0, -8.01588036613834e-05, -7.11068031040732e-06, 0, 1.64272153578625e-05, -2.1591501594556e-06, -7.0982893367229e-06, 2.23584421328716e-05, 1345210.03127895, 0, -8.08433530887425e-05, -1.07096142845173e-05, 0, 2.56028605445397e-05, -3.98875521662783e-05, -1.23194074542137e-05, 3.6529554936727e-06, -1.69061523881594e-05, 3.27823270974768e-05, 5.03918938469351e-05, 2690420.06257643, 0, 1345210.03127867, -4.58898334513013e-06, 0, 1.4007307224506e-05, 2690420.0624258, -2.17384569898511e-05, 3.70989948189745e-05, 2.1926387196124e-05, 1.72424486574649e-05, 6.65194678309609e-05, -9.64862147491348e-05, 8.47895195648995e-05, 9.63882797807134e-05, 2690420.0624943, -1.02424242879246e-05, 3.17311656172179e-05, 1345210.0311395, -5.08648304182743e-05, 9.46404070958829e-05, 0.000161830792272782, 1.19104749585207e-05, 0.000193187241765991, 3.27037029134423e-05}, 
                                                    {0, 0, 0, 0, 0, -1.9151937357879e-07, 0, -242.9999998238, -4.83250887695674e-08, 0, 0, 1.70567119078391e-06, 0, 3796.8749983543, 0, 4525.87499726433, 4.5105180056937e-07, 0, 7.98660347301937e-07, 516.375000349625, 0, -7.89123113211481e-06, 0, -17085.9374914621, 0, -28476.5624820545, -2.25401491157636e-06, 0, -5.45137890032315e-06, -70166.2499834473, 0, -6.58088370207731e-06, -7517.81250093356, -9066.93750291185, -273.375006420843, 0, 2.11162244169182e-05, 0, 24603.7499724434, 0, 75861.5624348402, 7.42117074049164e-06, 0, 1.87345125105898e-05, 310212.281193511, 0, 435281.343681162, 2.36679958232143e-05, 0, 3.20580449440309e-05, 28294.3124979791, 50847.7500097638, 3.63588559291663e-05, 131425.031257345, 3690.5625189254, 4510.68754546036, 0, -3.39447902878282e-05, 0, -11071.6874478911, 0, -84882.9373638424, -1.26836938815753e-05, 0, -3.717637579727e-05, -428105.24985668, 0, -1131157.40606953, -5.30310341312529e-05, 0, -8.96268933456009e-05, -22143.374991412, -107026.312520768, -0.000103832914813723, -1861888.78114628, 0, -8.78370560252935e-05, -488999.531234325, -11071.6875135869, -730731.37504527, -22143.3751257397, -8.67095788064726e-05, -0.000162280308740148, -60894.281383245, 0, 3.23516243614102e-05, 0, -5.54826195904363e-05, 0, 33215.0623392836, 1.26274719826599e-05, 0, 4.23702478989896e-05, 182682.843530187, 0, 1212349.78095854, 6.57027026052322e-05, 0, 0.000124794133385099, -1.20822952712952e-05, 66430.1250211957, 0.000152883826488604, 2358269.43735645, 0, 4550463.56240763, 0.000112081986053371, 0, 0.000157433495369272, 365365.687430361, -4.84019416004563e-06, 1511285.34383802, 33215.0627108604, 0.000192287536185356, 0.000345793084544576, 2657205.00003198, 8.22123770149571e-05, 0.000308286888534047, 182682.843874759, 298935.562943296, 0.000463919527089682, 0, -1.6844615261333e-05, 0, 3.07898649690148e-05, 0, 9.90996950575394e-05, -7.88309603567516e-06, 0, -2.36467556573749e-05, 0.000170053820774649, 0, -448403.343503417, -4.81795149701728e-05, 0, -9.89285723213926e-05, 5.70551928452045e-06, -1.16520552043312e-06, -0.000119410556670149, -896806.687332173, 0, -4334565.6561697, -9.75096903867706e-05, 0, -0.000134849768141661, 8.61359040057235e-05, 1.29535708466826e-05, -896806.687565223, -0.000175531892547192, -0.000201635826217002, -0.000320300316210953, -4782969.00003502, 0, -0.000116504165449378, -1793613.37493501, 2.01479634194546e-05, -5231372.34382581, -0.00025278422038698, -0.000110913377484029, -0.000277399207556545, 2.24680271040003e-05, -448403.344330205, -0.000535226181740573, -896806.687819098, -0.000344620635767164, -0.000868943294777072, 0, 3.69025205588624e-06, 0, -6.93509479220137e-06, 0, -2.47601248374105e-05, 1.52817662658661e-06, 0, 5.75515136291683e-06, -5.07355219110179e-05, 0, -8.18274718448876e-05, 1.43203434286948e-05, 0, 2.97324900705194e-05, -5.61338458884479e-07, -1.08403467873332e-06, 3.48963670521716e-05, -7.83832433658444e-05, 0, 1345210.0312053, 3.28170819376556e-05, 0, 5.35114092496743e-05, -3.10023405107935e-05, -1.58842876287312e-06, 1.0460247816906e-05, 5.9199496376913e-05, 8.09588149400902e-05, 0.000125155727691467, 1345210.03126083, 0, 2690420.06255397, 4.70539649945599e-05, 0, 5.36246772479777e-05, -5.27711995273573e-05, -8.22215961189373e-06, 2690420.06255228, 0.000114164929418979, 6.59137309935136e-05, 0.000152944189716129, -4.11310113130011e-05, 0.000240068867567771, 0.000262837783636757, 2690420.06252214, 1.2483590395125e-05, 0.000141574922926094, 1.68012570391757e-05, 9.74930341289033e-06, 1345210.03153593, 0.000502082259872984, 0.000207830753132397, 0.000408629959709925, 0.00037116714635023}, 
                                                    {0, 0, 0, 0, 0, 9.11690756450072e-08, 0, 850.499999530819, 1.83103529735326e-07, 0, 0, -5.9828147102948e-07, 0, -10555.3124953157, 0, -16934.0624927109, -1.86179571974936e-06, 0, -3.10412780377985e-06, -2900.81250201397, 0, 1.79970718327916e-06, 0, 29387.812473686, 0, 120034.406201509, 9.53954243809094e-06, 0, 2.05044060185201e-05, 206830.968703586, 0, 2.57428167587467e-05, 32668.312510435, 54424.4062657941, 3280.50003029316, 0, -2.38954346519329e-06, 0, -30754.687412854, 0, -393659.999822436, -3.09298478343105e-05, 0, -7.01718377678146e-05, -547638.4685701, 0, -1427222.53105581, -9.88584577161818e-05, 0, -0.000129066072846128, -62739.5625268307, -346912.875057379, -1230.18767108482, -605457.281328892, -33215.0626154794, -57818.8126887498, 0, 3.20083989318451e-07, 0, 11071.6873335925, 0, 638467.312125551, 5.60101698351839e-05, 0, 0.000135007915077047, 540667.405787194, 0, 4487723.9995116, 0.000217482895383836, 0, 0.000347171003885119, 33215.0625482169, 944784.000122632, 0.000492921968628712, 3448830.65583514, 0, 0.000361925734168313, 1101632.90640783, 33215.0627092692, 3775445.43775956, 326614.781740286, 11071.6879808978, 20298.0945126239, 581263.594493486, 0, 2.65458776956576e-06, 0, 0.000179480105796781, 0, -498225.937050999, -6.06835263089994e-05, 0, -0.000149268141452111, -182682.843065349, 0, -6842302.87426052, -0.000284260849351843, 0, -0.000485719049442988, -5.087838063395e-05, -1096097.06265424, -0.000743142085021433, -3039178.21806956, 0, -9200572.31199338, -0.000548683866250431, 0, -0.000661578611346878, -548048.531413015, -0.000229436522702172, -9898088.62541609, -697516.313229874, -0.000990903123853754, -0.00157255981846726, -6178001.62544474, -0.000543822470544046, -99645.1888465402, -548048.532385976, -3238468.59558448, -182682.845916486, 0, -2.72261194790979e-06, 0, -0.00010161474740866, 0, 149467.780966482, 3.71592567501804e-05, 0, 8.46213455012826e-05, -0.000513258985013528, 0, 4932436.78065531, 0.000199348268103752, 0, 0.000353812095833606, 3.31633330155132e-05, 448403.343852552, 0.000560514630169611, 896806.686815744, 0, 10313276.905855, 0.000431027082908335, 0, 0.00058734682240203, 8.83512193521555e-05, 0.000146343431416343, 10761680.2503091, 448403.344347128, 0.00101769256645452, 0.00153283381468943, 6427114.59341562, 0, 0.000617509126977562, 2690420.06278471, 0.000266319591606538, 13003696.9692532, 149467.7823596, 0.000711897378910665, 0.0015098793520126, 0.000806292741378326, 6726050.15810207, 0.00261954921027571, 2690420.0644736, 0.00204488761739833, 896806.690826506, 0, 8.49829047484483e-07, 0, 2.34995358937063e-05, 0, 7.30730173485814e-05, -9.02073866653215e-06, 0, -2.02122944787154e-05, 0.00015059305575661, 0, -1345210.03105824, -5.75032549378735e-05, 0, -0.000104065325298973, -8.61761239604793e-06, -2.78273557883602e-05, -0.000163966263087969, 0.000273941669126835, 0, -4035630.09359787, -0.000144763714921222, 0, -0.00020764998048456, -1.78863367502078e-05, -4.64201483525298e-05, -4035630.09383419, -0.000197841473880064, -0.000399713626532168, -0.000602039708467824, -1345210.03109518, 0, -4035630.09370783, -0.000236362676429915, 0, -0.000259984987462991, -6.19049137638414e-05, -6.30754649237883e-05, -8071260.18769804, -0.000400527904556446, -0.00035299340834019, -0.000728044986940876, -0.00024602560967415, -4035630.09441426, -0.00121542936826774, -4035630.093935, -0.000143108226560418, -0.000557851645253874, -0.000729738079266298, -0.0006091913815969, -4035630.0948816, -1345210.03292852, -0.00118106669098977, -0.00177154277968759, -0.00180625997758526}, 
                                                    {0, 0, 1.84573464997331e-08, 0, 0, -3.88866723283709e-07, 0, -1700.99999922079, -4.09975576411129e-07, 0, 0, 3.32287509203396e-06, 0, 23662.1249929216, 0, 31316.6249875423, 4.10959761310507e-06, 0, 6.95849329675244e-06, 8353.12500499286, 0, -1.47389347255309e-05, 0, -90441.5624634905, 0, -196921.1249164, -2.14276558592548e-05, 0, -4.72690915555778e-05, -425143.687426445, 0, -5.35993580434013e-05, -105705.000024063, -143294.062545643, -15263.4375699013, 0, 3.78536488002388e-05, 0, 149672.812382317, 0, 556659.843444196, 6.80440775255285e-05, 0, 0.000166306486821527, 1532608.59350346, 0, 2559610.12467587, 0.000202236151229843, 0, 0.000277673750585845, 311647.500057901, 787320.000194267, 12301.8753902077, 1775570.62521537, 174276.562752199, 242962.031725541, 0, -5.85874190596509e-05, 0, -114407.437275856, 0, -787935.093109985, -0.000126687038528082, 0, -0.000331296141699794, -2367495.84314939, 0, -6759265.21788543, -0.00045841509036365, 0, -0.000760468034702332, -346912.875088501, -1749326.625456, -3690.56362127855, -8224418.53071258, 0, -0.000744663613533084, -4915829.25040865, -354294.000412924, -9362957.06331099, -1138538.53267378, -125479.126014822, -180837.564346538, -2729170.97051487, 0, 5.3990904624377e-05, 0, 33215.0622577431, 0, 548048.530492311, 0.000140950902794428, 0, 0.000373118625447996, 1677360.65535383, 0, 8785384.02988176, 0.000596948982695946, 0, 0.00110960925636453, 132860.250092211, 1693968.18809377, 0.00170943743280707, 10944363.0930361, 0, 17537552.9993359, 0.00112156910794031, 0, 0.00137663865895017, 5081904.56285281, 199290.375405262, 19331166.3764418, 1793613.37733692, 33215.0646199726, 49822.5974382313, 22586242.5012768, 132860.25106428, 697516.316030484, 5181549.75262082, 12339395.7234421, 1826828.44229948, 0, -2.73057686309876e-05, 0, 0.000137288807948569, 0, -149467.780779643, -8.49129905770791e-05, 0, -0.000219264880136909, -448403.343063874, 0, -5530307.90512278, -0.000416638557119943, 0, -0.000838516288783919, -6.01475143929618e-05, -597871.125392569, -0.00131493763709575, -6427114.59301816, 0, -16142520.3743931, -0.000908521495770429, 0, -0.00126301640501239, -1793613.37512423, -0.000241671092321007, -17039327.0636931, -896806.689452801, -0.0021803406736472, -0.00364022044689056, -16740391.4998746, 0, -0.00123402414684439, -19430811.5632537, -0.000462987875278023, -34078654.1266268, -597871.128384685, -0.00134648043175186, -149467.785226643, -2690420.06421818, -17936133.7553568, -448403.349463043, -19729747.1300612, -1793613.37933214, -6875517.94534638, 0, 5.83596298620527e-06, 0, -3.17205267669272e-05, 0, -0.000118781135088146, 2.0918309837875e-05, 0, 5.32913689965395e-05, -0.000204806073654326, 0, 1345210.03088328, 0.000118755525961294, 0, 0.000251640294216832, 1.72230968456125e-05, 0.00010706202768978, 0.000392820795454849, 1345210.03093178, 0, 5380840.12472492, 0.000299813339409821, 0, 0.000462985945839957, 1.22158943980523e-05, 7.22653195858951e-05, 5380840.12536626, 0.000640097000824358, 0.000858613692268358, 0.00140014881960907, 5380840.12496568, 0, 8071260.1875744, 0.000462824120793462, 0, 0.00052104426358345, 5380840.12512662, 8.80100297642701e-05, 16142520.3757129, 0.00133977529783925, 0.00070570694495415, 0.00182664804816271, 0.000459902522304131, 8071260.18962357, 0.00265254147513905, 16142520.3755726, 0.00024551508947912, 0.00163632759774586, 8071260.18920846, 0.00127415243643099, 16142520.3780843, 5380840.12924894, 0.0023727178703075, 1345210.03532385, 5380840.12908777}, 
                                                    {0, 0, -1.87930775907252e-08, 0, 0, 4.30616312564033e-07, 0, 755.999999636952, 1.36074265212997e-07, 0, 0, -3.93103758071516e-06, 0, -12784.4999971699, 0, -12784.4999939179, -1.17835806431923e-06, 0, -2.42764329687784e-06, -2578.50000118056, 0, 1.86404937762189e-05, 0, 68343.7499880775, 0, 68343.7499585298, 5.55396216589542e-06, 0, 1.73996049505047e-05, 213232.499967303, 0, 1.58420210280993e-05, 40641.750003082, 40641.750010099, 2916.00001914454, 0, -5.10572415665962e-05, 0, -144888.749964519, 0, -144888.749848121, -1.57205508845786e-05, 0, -6.27524016832948e-05, -1109082.37491984, 0, -1109082.37484071, -5.07869427202184e-05, 0, -9.15215602263157e-05, -186441.749997135, -186441.750044589, -1093.5001153347, -633956.625034735, -42646.5000557562, -42646.5001234015, 0, 8.38689087614303e-05, 0, 132860.249933184, 0, 132860.24968531, 2.87118748463456e-05, 0, 0.000129208604990545, 2236480.87483419, 0, 2236480.87453274, 0.000111619816595347, 0, 0.000265957726883374, 280482.749974453, 280482.750104838, 0.000339908975698961, 5447270.24984528, 0, 0.00021087946625891, 2841733.12498854, 162384.750028539, 2841733.12519161, 162384.750379532, 14762.2502567668, 14762.2505071062, 620014.500392021, 0, -8.1529906747939e-05, 0, -44286.7499277212, 0, -44286.7496349379, -3.34318804003184e-05, 0, -0.000146580143909394, -1926473.62474494, 0, -1926473.62421357, -0.000130733414642084, 0, -0.00040851338954331, -132860.249969299, -132860.250128376, -0.000528657783653814, -9787371.74994946, 0, -9787371.74978477, -0.000272862474312876, 0, -0.000424369981287787, -4052237.62482011, -132860.249957257, -4052237.62540539, -132860.250648285, -0.000600508292770696, -0.00104195434771368, -12068139.3751807, -44286.7501881424, -44286.7508773046, -2325054.37535168, -2325054.37619854, -199290.376207827, 0, 4.32187347504906e-05, 0, -4.07910220254716e-05, 0, -0.000219969656389373, 1.94810588463268e-05, 0, 8.75054319791165e-05, 597871.124791611, 0, 597871.124332102, 8.80178853779771e-05, 0, 0.000324297610573476, -1.4794085453357e-05, 6.68078236853595e-05, 0.00042321702126093, 7174453.49993218, 0, 7174453.49970935, 0.000214780036255721, 0, 0.000418716647977638, 1793613.37476348, -5.20692408369518e-05, 1793613.37535735, 0.000527823011034666, 0.000628058709004987, 0.00108275403682858, 13153164.7502326, 0, 0.000311605062792079, 14946778.1248925, -2.43449044361166e-05, 14946778.125434, 0.000870209399259946, 0.000269233415893461, 0.00106159149790907, 1793613.37490026, 1793613.3766958, 0.00152371207021243, 8370195.7510552, 597871.125783146, 597871.12716462, 0, -9.62217942866772e-06, 0, 9.32120715783249e-06, 0, 5.34429876240097e-05, -4.79264204776597e-06, 0, -2.14106225616394e-05, 6.51927844311771e-05, 0, 0.000219490320261028, -2.44135023540044e-05, 0, -9.89480323908194e-05, 1.07010483104594e-06, -1.55100806030768e-05, -0.000129657783130664, -1793613.37494243, 0, -1793613.37483494, -6.22425271701022e-05, 0, -0.000164994350292248, 8.46032803464542e-05, 1.43833191592378e-05, -0.000103242716182736, -0.0001641450898314, -0.000249542319401271, -0.000408906152314899, -5380840.12513761, 0, -5380840.1251287, -9.95899186985098e-05, 0, -0.000143087874281543, -5380840.1248656, 3.64176571136019e-05, -5380840.12524808, -0.000361927940469233, -0.000181682820796219, -0.000498568122964198, 0.000134357748499661, -0.000730153908915638, -0.000739473541041057, -10761680.2501174, -2.24968282487481e-05, -0.000409762439073011, -5380840.12506989, -6.02788792427545e-05, -5380840.12586869, -0.00128820616371372, -0.000485208117835685, -0.0011476813134811, -1793613.37588995}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -1.0165805631086e-07, 2916.00000008495, 0, 0, 0, 0, -34877.4428585828, 0, 1.34399971081566e-06, -34877.44285792, 1.81913665074563e-06, -34877.442861469, 0, 0, 0, 0, 174358.575010993, 0, -5.63309047587667e-06, 174358.575003402, -1.2685095312738e-05, 0, 348717.150008581, -2.16225423496764e-05, 348717.150031822, 174358.575047897, 348717.150045966, 0, 0, 0, 0, -472966.08754135, 0, 1.2582458049499e-05, -472966.087511177, 4.18471252649676e-05, 0, -1418898.26256032, 0, 7.38530216945446e-05, -1418898.26251099, 0.000140596741940783, -1418898.26260306, -472966.087716045, -1418898.26270056, -1418898.26280091, -2837796.52526755, -1418898.26288244, 0, 0, 0, 0, 752874.750083712, 0, -1.74839074032706e-05, 752874.750025598, -7.1657429243261e-05, 0, 3011499.00020152, 0, -0.000117089929275275, 3011498.99999136, -0.000416757317311753, 3011499.00018105, 752874.750495736, 3011499.00049045, 0, 4517248.50007582, -0.00041432569048764, 4517248.50075834, 9034497.00069496, 3011499.00107169, 3011499.00135331, 9034497.00085673, 4517248.50110395, 9034497.00190825, 0, 0, 0, 0, -704159.325093657, 0, 1.47921930910543e-05, -704159.32503422, 6.56961409799851e-05, 0, -3520796.62532708, 0, 0.000111123396045758, -3520796.62497966, 0.000598657019835902, -3520796.6251771, -704159.325601619, -3520796.62569082, 0, -7041593.25025758, 0, 0.000503545100926467, -7041593.24996404, 0.00101001279216505, -7041593.25094921, -14083186.5009557, -3520796.62682109, -3520796.62737239, -14083186.5014527, -7041593.25162105, -7041593.25195194, -21124779.7515564, -7041593.25266966, -21124779.753729, -14083186.5049994, -21124779.7540823, 0, 0, 0, 0, 358722.67505466, 0, -5.90674656222353e-06, 358722.675023413, -3.05943893542284e-05, 0, 2152336.05025411, 0, -6.70728925654202e-05, 2152336.04999584, -0.00040905485891951, 2152336.05008963, 358722.675368026, 2152336.05050909, 0, 5380840.12532887, 0, -0.000273859698126043, 5380840.12489417, -0.00104468428346347, 5380840.12558804, 10761680.2506643, 2152336.05147521, 2152336.05195878, 10761680.2512989, 5380840.12630286, 0, 7174453.5000496, -0.00088770215996707, 7174453.50154194, 21523360.5013832, 5380840.12716587, 5380840.12815434, 21523360.5015887, 7174453.50210675, 21523360.5032873, 10761680.2556372, 21523360.5039014, 21523360.5063174, 32285040.7552392, 21523360.5067931, 0, 0, 0, 0, -76869.1446558617, 0, 5.72774481857511e-07, -76869.1446492017, 5.66305135319542e-06, 0, -538084.012576034, 0, 1.97777393038142e-05, -538084.012503519, 0.000107111493487928, -538084.012518058, -76869.1447335447, -538084.012649588, 0, -1614252.03764009, 0, 6.15328824899362e-05, -1614252.03745766, 0.000380354309256045, -1614252.03764548, -3228504.07518218, -538084.012957099, -538084.013103675, -3228504.07546774, -1614252.03794967, 0, -2690420.06257279, 0, 0.000256052611320944, -2690420.06245414, 0.000492555954476831, -2690420.06294759, -8071260.18798231, -1614252.03837495, -1614252.03887222, -8071260.18815423, -2690420.06312339, -8071260.18858049, -3228504.07727005, -8071260.18899904, -2690420.06333078, -10761680.2506555, -2690420.06367043, -10761680.2526678, -16142520.3773234, -8071260.19093382, -8071260.19131793, -16142520.3773818, -10761680.2528836, -16142520.3793433}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 104.142857142838, 0, 0, 0, 0, -2400.49285714264, 0, 0, -104.142857150376, 0, -104.142857145808, 0, 0, 0, 0, 21323.2499999399, 0, -1.04130915734652e-08, 1.72767243254315e-08, 1.74434371780155e-08, 0, 2296.3500002504, -9.79185072627844e-08, 2296.35000003856, -1.58985131842111e-08, -2.07315601577891e-08, 0, 0, 0, 0, -96528.7124995106, 0, 1.7468987058637e-07, 9.80887901583254e-08, -3.91675057152153e-08, 0, -19026.9000012964, 0, 3.31288184639856e-07, -1.22676800890143e-06, 8.06714362345486e-08, -19026.9000003267, -9.24090572056576e-08, 0, -3.0888279112095e-07, -3.99883084739508e-07, 2.06130309121135e-07, 0, 0, 0, 0, 243577.124998545, 0, -2.59298444844656e-07, -4.05115468753599e-07, 1.92736883205774e-07, 0, 77501.812502051, 0, -2.42704504295669e-08, 1.86718024761377e-06, -4.7284242509444e-07, 77501.8125003461, -3.89354766414235e-07, -1.88569331878137e-07, 0, 6.90195352660047e-06, -1.0391782358813e-06, 1.18120002587934e-06, 1.1400581355248e-06, 1.11564214908602e-07, 4.63813150434864e-07, 0, 8.47366561683321e-07, 3.37337583493786e-06, 0, 0, 0, 0, -345436.649997905, 0, 7.48679119956632e-08, 5.11180594220668e-07, -9.47008133000188e-08, 0, -166075.312500851, 0, -6.3850537290812e-08, -1.67743615810191e-07, -4.92182111459928e-07, -166075.312500806, 9.57890654022469e-07, 2.59043368073318e-07, 0, -1.11289944518937e-05, 0, -5.76178562882847e-07, -1.28397420012686e-05, 1.19634483217952e-06, -8.5800203597835e-07, -1.19242891606381e-06, 1.01794226423317e-06, -2.25268591886672e-06, -9.61921575655271e-07, 2.74056451830417e-06, 7.44159602126329e-07, -2.38090077117717e-07, 3.67064969578702e-06, -8.78675492439717e-06, -2.56022894874461e-06, -2.80945440915726e-06, 0, 0, 0, 0, 256230.482141347, 0, -1.80761904373143e-07, -2.44985817803289e-07, 1.92690735530508e-07, 0, 179361.337499568, 0, -2.72702680848018e-07, -1.54810303887025e-06, 1.43079532335423e-07, 179361.337500431, -7.23114135951251e-07, 3.14927169438641e-07, 0, 6.34318744019269e-06, 0, 7.28071644163707e-07, 8.76704184238133e-06, 7.83736047297101e-08, 7.74541116649874e-07, 1.69751737364381e-06, -1.11046050329366e-06, 1.29096752013112e-07, -7.4638018573882e-07, -7.32840107526585e-07, 0, 1.64189029782291e-05, -2.7204365876828e-07, -8.39802393822104e-08, 2.51612842027332e-06, -4.69168846945062e-07, 5.67369134321866e-06, -3.58451240246886e-06, 5.95725172042495e-06, 5.72683839164473e-06, 8.0448407950053e-06, 1.55390346276684e-06, 1.90766566961439e-05, 3.2223208152746e-06, 2.92723757928677e-06, 0, 0, 0, 0, -76869.1446424183, 0, 1.65216305955221e-08, 1.91282583367574e-08, -3.39895580298189e-08, 0, -76869.1446425818, 0, 1.71904694739965e-07, 8.44858572971383e-07, 0, -76869.144642948, 2.94469741836778e-07, -2.3266202839513e-07, 0, -1.02184177985813e-06, 0, 1.52631690844782e-07, -1.75681139342398e-06, 0, -3.84298765623804e-07, -4.8716115480398e-07, 5.46491450256929e-07, 5.96086946864647e-07, 4.12088692361393e-07, 1.33568589960202e-06, 0, -5.83679858211438e-06, 0, 8.51143694880931e-07, -7.15106006919954e-06, 5.31208815043126e-07, 6.75804121969641e-08, 1.05763575080003e-07, 5.68631511156188e-07, -9.87898062624619e-07, 6.8823454903392e-07, 1.12698021871589e-06, -2.03561919599338e-06, -2.6514428265179e-06, 1.89815328710744e-06, 1.03869488631023e-06, 1.21055473210039e-06, 8.64274753729796e-07, -3.55448477468576e-06, -6.18702482725857e-06, -6.12015555706472e-06, -4.60517554342531e-06, -7.1125514802616e-07, -8.21729401701525e-06, -6.04942196782979e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 104.142857141381, 0, 0, 0, 0, -104.142857133513, 0, 0, -2400.49285711094, -1.37639405358287e-08, -104.142857130795, 0, 0, 0, 0, -2.95102671429627e-08, 0, 8.95110302232567e-08, 21323.2499997473, 1.00223339228683e-07, 0, 2296.34999980416, -3.57627049913038e-08, -2.29867539954838e-07, -1.54834473035093e-07, 2296.34999982803, 0, 0, 0, 0, 2.75189068349209e-08, 0, -4.45315801273758e-07, -96528.712499023, -3.4903854712713e-07, 0, 6.74534117528546e-07, 0, -2.58408409883157e-07, -19026.8999987413, 6.17272504562797e-07, 7.28917945129601e-07, 7.73163148283979e-07, -19026.899999812, 1.96841639356473e-06, 1.6747609825727e-06, 1.73209629839074e-06, 0, 0, 0, 0, 4.15195609416035e-08, 0, 6.23874734872249e-07, 243577.124997842, 1.03461239969598e-06, 0, -1.04601509528054e-06, 0, 2.37344343999903e-06, 77501.8124966686, -3.3761092192558e-08, -1.64600030451445e-06, -1.40399783312254e-06, 77501.8125008189, 0, -4.00475177925346e-06, -1.8817779013527e-07, -4.98541857373467e-06, -7.53358530213681e-06, -6.91485684161481e-06, -8.56944448817342e-06, -5.70351736182794e-06, -4.47567579895996e-06, -1.55293650694367e-05, 0, 0, 0, 0, -9.09815002975381e-08, 0, -7.67554441821954e-07, -345436.649997216, -1.40467245107967e-06, 0, 5.58608810892376e-07, 0, -2.03125514118288e-06, -166075.312495552, 3.01327516792096e-07, 1.71393741189421e-06, -3.76715960288033e-07, -166075.312500995, 0, 6.23557078424673e-06, 0, 8.66525892889858e-07, 7.90479276345537e-06, 1.47563519107212e-06, 6.99578358673639e-06, 9.7254419531192e-06, 1.06695955564619e-05, 1.17124400139726e-05, 4.39268095881304e-06, 3.11747084540979e-06, 1.30027663898728e-05, 1.18419177452967e-05, 3.04524123879898e-05, 1.99536646518203e-05, 4.98881872419408e-05, 2.57772151902927e-05, 0, 0, 0, 0, 5.64236905562811e-08, 0, 4.37681044684568e-07, 256230.482140931, 6.53198552102523e-07, 0, 1.27358439939822e-07, 0, 2.41357433221429e-06, 179361.337496849, 9.94393440264798e-07, -9.19267563134262e-07, -2.81080434309139e-07, 179361.337502441, 0, -4.13729194241994e-06, 0, 1.71743926562281e-06, -6.25264931813142e-06, -2.49218021399616e-06, -5.77953404218135e-06, -7.3255618100319e-06, -3.9982055775223e-06, -5.98525929844911e-06, -1.30509993351021e-06, 3.37390793034054e-06, 0, -9.25405546655715e-06, -4.02232341061266e-06, -1.54963687292902e-05, -1.76397110081408e-05, -1.58486023252179e-05, -3.07390757483623e-05, -1.48699125568383e-05, -2.14659148362471e-05, -2.37439680052921e-05, -4.13202598255024e-05, -2.62261611680222e-05, -4.51518975758214e-05, -2.66047975451174e-05, -3.73740651952485e-05, 0, 0, 0, 0, -1.00770887034365e-08, 0, -7.93573618745758e-08, -76869.1446423049, -2.75550317743317e-07, 0, -1.34652476588053e-07, 0, -6.09719936605336e-07, -76869.1446418916, -4.56086390140053e-07, 1.67115993153412e-07, -1.1294975717754e-07, -76869.1446435879, 0, 9.10947410555817e-07, 0, -1.00710168971866e-06, 1.7486988915826e-06, 8.7669808494553e-07, 1.44078622536406e-06, 1.8311077306532e-06, 6.12976297638942e-07, 1.3433268254655e-06, -2.48122969011439e-07, -7.40024008199324e-08, 0, 3.61869349601436e-06, 0, 4.92940311328349e-07, 3.95719819130393e-06, 2.66928450633356e-06, 4.07521463466991e-06, 5.94061270056841e-06, 3.0249220376758e-06, 8.70528891296897e-06, 2.30660267839736e-06, 3.45928448209263e-06, 8.70486881259721e-06, 1.44523139383368e-05, 4.67494894690784e-06, 4.62532442066828e-06, 7.49194155246053e-06, 1.04412619804371e-05, 2.14650024509047e-05, 1.39974754212069e-05, 2.90775918882065e-05, 2.87776671841286e-05, 2.02590797000244e-05, 2.46869207234169e-05, 2.99170619045701e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 5.20123132982892e-08, -3402.00000013104, 0, 0, 0, 0, 61102.3500021312, 0, -2.56908971870987e-07, 30484.3500012722, -1.38424288145971e-06, 30484.350004443, 0, 0, 0, 0, -386324.437516365, 0, -1.39504785377696e-06, -111965.287506103, 1.03269734048785e-05, 0, -498289.725012706, 1.14983255668728e-05, -498289.725033496, -111965.287553284, -223930.575039082, 0, 0, 0, 0, 1223585.49381181, 0, 1.09593302056858e-05, 215897.906271548, -3.60978638402235e-05, 0, 2663068.89384038, 0, -2.69455705691337e-05, 1655381.30626506, -8.57369975924444e-05, 2663068.89386046, 215897.906527956, 647693.718903146, 1655381.30656003, 3310762.61275747, 647693.71907213, 0, 0, 0, 0, -2173741.31262576, 0, -2.85445024381323e-05, -230660.156301069, 6.58060963739013e-05, 0, -6751884.09405307, 0, 1.96555321547248e-05, -2865721.78123287, 0.000270202778250826, -6751884.09394821, -230660.15697526, -922640.625358098, 0, -7213204.40636129, 0.00020275816116824, -7213204.40698829, -14426408.8132401, -2865721.78249455, -922640.626323341, -8597165.34454204, -1383960.93821898, -8597165.34533334, 0, 0, 0, 0, 2205480.15014159, 0, 3.94055436503198e-05, 129538.743820078, -6.55375347756695e-05, 0, 8951459.34424379, 0, -7.87087090884457e-06, 2723635.12496347, -0.000404715530733145, 8951459.3439487, 129538.744736917, 647693.719252903, 0, 13751035.875382, 0, -0.000176003026702388, 9599153.06244197, -0.000566437663639477, 13751035.8758784, 27502071.7510956, 2723635.12738161, 647693.721385341, 10894540.5013117, 1295387.43833509, 9599153.0646099, 28797459.1892029, 1295387.43959302, 28797459.1906183, 10894540.5045792, 16341810.7528269, 0, 0, 0, 0, -1195742.25008309, 0, -2.92813756740155e-05, -29893.5562994286, 3.45582586835779e-05, 0, -6008604.80663705, 0, 9.34791294220567e-06, -1345210.03124063, 0.000279033568937675, -6008604.80635331, -29893.55692435, -179361.33787878, 0, -12106890.2817348, 0, 5.98524572691226e-05, -6277646.81233365, 0.000644131094777389, -12106890.2817667, -24213780.5633084, -1345210.03338515, -179361.339887593, -6726050.15740679, -448403.344322755, 0, -12256358.0625758, 0.000397151916784858, -12256358.0640448, -36769074.1891929, -6277646.8151419, -448403.346708054, -25110587.2517581, -597871.12620154, -36769074.190274, -6726050.16208229, -13452100.314773, -25110587.2555091, -37665880.8789201, -13452100.3175505, 0, 0, 0, 0, 269042.00626983, 0, 8.82963127805886e-06, 1.38919034787456e-05, -7.6778940801043e-06, 0, 1614252.03761729, 0, -4.32160398343657e-06, 269042.006255401, -7.27191528259163e-05, 1614252.0375213, 0.000185805448259436, 0.000113603675730843, 0, 4035630.09395685, 0, -1.55894186738078e-05, 1614252.03743118, -0.000245642524776462, 4035630.09387323, 8071260.18773459, 269042.006971928, 0.000800558489974922, 1614252.03791589, 0.00020511388221091, 0, 5380840.12510615, 0, -7.90827087368855e-05, 4035630.09368462, -0.000257214284042831, 5380840.12541109, 16142520.3756419, 1614252.0386885, 0.00142706460659332, 8071260.1882183, 0.000203103367592155, 16142520.3759222, 1614252.04009998, 4035630.0945369, 4035630.09470372, 16142520.3758496, 0.000953966715161441, 16142520.377228, 24213780.5643511, 8071260.19085385, 4035630.09703964, 16142520.37661, 5380840.12667013, 16142520.3782242}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 3402.0000000988, 0, 0, 0, 0, -68756.8500014836, 0, -2.87897379732859e-07, -22829.850001058, 5.40523227589145e-07, -22829.8500025636, 0, 0, 0, 0, 506582.100011496, 0, 3.51269934612588e-06, 60598.1250057495, -3.83810989431443e-06, 0, 429399.225008891, -3.01285062003929e-06, 429399.225019486, 60598.1250319499, 121196.250020449, 0, 0, 0, 0, -1822235.73754364, 0, -1.71273445585907e-05, -79552.1250219624, 1.32273503839002e-05, 0, -2799414.6750645, 0, 7.77738781027995e-06, -1056731.06250884, 2.28981656073625e-05, -2799414.67506589, -79552.1251795766, -238656.375077607, -1056731.06267332, -2113462.12514088, -238656.375166231, 0, 0, 0, 0, 3579845.62508946, 0, 4.08535719888084e-05, 51667.8750535177, -2.4806525031996e-05, 0, 8336980.68771669, 0, -5.68269293868901e-06, 1280625.18748277, -6.2193166978647e-05, 8336980.68762514, 51667.8755016579, 206671.500179768, 0, 5986092.37507564, -6.14489481931775e-05, 5986092.37539817, 11972184.7504279, 1280625.18824479, 206671.500744609, 3841875.56293746, 310007.250323814, 3841875.56326405, 0, 0, 0, 0, -3932663.40010176, 0, -5.45345551070576e-05, -13286.0250747082, 2.63983611102328e-05, 0, -12522078.5628553, 0, -1.45739749932932e-06, -763946.437466813, 8.66125367303328e-05, -12522078.5626341, -13286.0257257632, -66430.1252498309, 0, -13983541.3127601, 0, 7.80605939784586e-05, -6144786.56245676, 0.000147466989796709, -13983541.3129689, -27967082.6256597, -763946.439017928, -66430.1265909099, -3055785.75072669, -132860.250330716, -6144786.56371612, -18434359.6885113, -132860.251119149, -18434359.6889886, -3055785.7523996, -4583678.62624357, 0, 0, 0, 0, 2271910.27506017, 0, 3.89254590288337e-05, 5.37234223013088e-05, -1.64615992156061e-05, 0, 9267002.43778381, 0, 4.57317096762861e-06, 179361.33748904, -4.96541012246647e-05, 9267002.43757626, 0.000524050613494434, 0.000189970472571905, 0, 14348907.0003238, 0, -5.50773677041614e-05, 2989355.62487848, -0.000161080548618728, 14348907.0002759, 28697814.0005061, 179361.33893511, 0.00152081469184503, 896806.688141935, 0.000175762154733108, 0, 10163809.125052, -0.000120388140977813, 10163809.1258561, 30491427.3760385, 2989355.62661553, 0.00172527634481601, 11957422.5010621, 0.00057574157170127, 30491427.3763126, 896806.690782563, 1793613.37586443, 11957422.5027049, 17936133.7517997, 1793613.37741308, 0, 0, 0, 0, -538084.012514371, 0, -1.10481719120118e-05, -1.54864089900536e-05, 4.37336986812061e-06, 0, -2690420.0625884, 0, -3.62687442478194e-06, -3.38481926011066e-06, 9.99864776820466e-06, -2690420.06251742, -0.000152270431962327, -5.72631849774694e-05, 0, -5380840.1251387, 0, 2.23622598297961e-05, -538084.012447451, 5.5087070933333e-05, -5380840.12506894, -10761680.2501548, -0.00050788545269887, -0.000537704377586812, -0.000233626604406082, -6.95319996007787e-05, 0, -5380840.12506626, 0, 3.00890786892424e-05, -2690420.06245937, 6.50770647721746e-05, -5380840.12521871, -16142520.3754096, -538084.013269408, -0.00086946515447577, -2690420.06293237, -7.93324388627873e-05, -16142520.3754312, -0.00154586414762122, -0.000259660583377716, -2690420.06307087, -10761680.2505368, -0.000594298752309125, -10761680.2510506, -16142520.375858, -2690420.06428856, -0.00174928981567912, -5380840.12571636, -0.00068556397923324, -5380840.12649507}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -2551.50000004184, 0, 0, 0, 0, 54629.4375004921, 0, -2.84718523656098e-08, 12529.6875005948, -2.18153553306825e-07, 12529.6875008133, 0, 0, 0, 0, -436306.500003609, 0, -1.75527745059391e-07, -22895.1562538871, 1.09330237955193e-06, 0, -252530.156254064, 1.99227796369206e-06, -252530.156257132, -22895.1562593059, -45790.3125062604, 0, 0, 0, 0, 1721134.82813798, 0, 2.38143072905971e-06, 18452.8125154415, -2.82628979077415e-06, 0, 1825803.28127654, 0, -1.15572818787057e-05, 433128.515633497, -7.55166099908122e-06, 1825803.28127688, 18452.8125525802, 55358.4375239305, 433128.515679198, 866257.031298169, 55358.4375447451, 0, 0, 0, 0, -3686871.93752589, 0, -7.41826337520989e-06, -5535.84378697719, 4.32526158032943e-06, 0, -6166929.9375795, 0, 2.88628419626449e-05, -326614.781256809, 1.02788326747388e-05, -6166929.93755765, -5535.84389911616, -22143.3750575306, 0, -2801136.93755018, 3.84950930576244e-05, -2801136.93763213, -5602273.87516255, -326614.781474483, -22143.3752073769, -979844.343898396, -33215.0625784451, -979844.343976295, 0, 0, 0, 0, 4367780.71877927, 0, 1.1573120403826e-05, 5.1102473062278e-05, -3.90751682992659e-06, 0, 10412922.0938728, 0, -3.49613179401212e-05, 91341.4218759599, -3.71072676458831e-06, 10412922.0938191, 0.000217503102371447, 8.28038317758003e-05, 0, 7813843.45324042, 0, -7.68817131532851e-05, 1860043.50004241, -5.18588702120586e-05, 7813843.45329989, 15627686.9065134, 91341.4223320826, 0.000455410333525516, 365365.687741434, 6.5955726854649e-05, 1860043.50037254, 5580130.50035914, 0.000320559604415059, 5580130.50047488, 365365.688217005, 548048.531593723, 0, 0, 0, 0, -2690420.0625171, 0, -9.14502929529565e-06, -3.69119393187441e-05, 2.55551906247607e-06, 0, -8519663.53134889, 0, 2.36044573389239e-05, -1.38189408653926e-06, -5.42957957679425e-06, -8519663.53129331, -0.000158770955816087, -6.567518155502e-05, 0, -9416470.21885862, 0, 7.30531330237372e-05, -448403.343755586, 3.44487542603346e-05, -9416470.21887051, -18832940.4377139, -0.000432268243292986, -0.000437844923506178, -0.000214290875666159, -1.10034277791959e-05, 0, -4035630.09382863, 6.83153534500256e-05, -4035630.09401768, -12106890.2816354, -448403.344231356, -0.000535986903151503, -1793613.37536363, -0.000147663870840695, -12106890.2817012, -0.000973535773020107, -0.000209878610252333, -1793613.37581041, -2690420.06306116, -0.000670848502674116, 0, 0, 0, 0, 672605.015628926, 0, 2.77848388306962e-06, 1.08485646763128e-05, -8.7551727840554e-07, 0, 2690420.0625324, 0, -5.47415785210663e-06, 2.05319536617339e-06, 3.91502369891333e-06, 2690420.0625107, 4.65806534087855e-05, 2.088968757465e-05, 0, 4035630.09379076, 0, -2.69869970053831e-05, -6.46099335947075e-06, -5.47869071159048e-06, 4035630.09378668, 8071260.18756898, 0.000153483961330781, 0.000155297416982007, 8.12734681550995e-05, 1.2261154863038e-05, 0, 2690420.06253424, 0, -2.95958421540518e-05, 672605.015646716, -2.05780616400675e-05, 2690420.06257007, 8071260.18765661, 0.000229022370664217, 0.000269659307237838, 0.000140814078478489, 1.06285647227951e-05, 8071260.18765714, 0.000456369912139302, 4.58322356926355e-05, 672605.015788659, 2690420.06268575, 0.000202259105574946, 2690420.06283347, 4035630.09403627, 0.000538636010389957, 0.000522536467975157, 0.000230887221914279, 0.000183043181129509, 0.000460316278782936}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 1360.80000001116, 0, 0, 0, 0, -30156.30000006, 0, 9.89840995989497e-08, -4641.3000002623, 1.08488266889032e-07, -4641.30000014604, 0, 0, 0, 0, 253017.675000085, 0, -5.5130911984836e-07, 5248.80000190145, -4.86678291730702e-07, 0, 97521.9750020542, -1.59636463720462e-06, 97521.9750017846, 5248.80000114301, 10497.6000008945, 0, 0, 0, 0, -1061979.862499, 0, 1.71452021947479e-06, -1968.30000713803, 1.13147568449442e-06, 0, -750988.462509947, 0, 8.35005787140715e-06, -104319.900009682, 5.92866354421333e-06, -750988.462508257, -1968.30000631797, -5904.90000301234, -104319.90001018, -208639.800011076, -5904.90000344935, 0, 0, 0, 0, 2438723.69999614, 0, -2.76024011736854e-06, 1.58888080582949e-05, -8.16955869691077e-07, 0, 2760540.7500184, 0, -1.74789894039952e-05, 36905.6250198618, -1.2065667096676e-05, 2760540.75001916, 1.71585283970608e-05, 7.36512448106272e-06, 0, 730731.375044081, -2.8135756809327e-05, 730731.375030325, 1461462.75004399, 36905.6250321798, 2.43241676447257e-05, 110716.875027657, -4.45085352122816e-06, 110716.875046935, 0, 0, 0, 0, -3095643.82499431, 0, 2.11390952829172e-06, -2.12509630565526e-05, 1.45627505030868e-07, 0, -5154977.70001665, 0, 2.05818566727926e-05, -1.75551776612084e-05, 9.4268821285189e-06, -5154977.70002597, -2.34506627219945e-05, -1.28956144564544e-05, 0, -2291839.31256487, 0, 4.57463275506603e-05, -232505.437585996, 4.42532998265187e-05, -2291839.31255054, -4583678.62507546, -5.73973785343756e-05, -6.24298102518708e-05, -4.14356450465855e-05, 3.41745981503237e-05, -232505.437552802, -697516.312575544, -6.18716305206387e-06, -697516.312635379, -0.000132408632018913, -4.4458457073716e-05, 0, 0, 0, 0, 2032761.82499582, 0, -1.37599803560037e-06, 1.53325762280697e-05, 4.95292107801448e-07, 0, 4663394.77501015, 0, -1.4769467491503e-05, 6.08513348971966e-06, -2.59937346826593e-06, 4663394.77501677, 1.53109482665394e-05, 1.46992844036929e-05, 0, 3228504.07503127, 0, -3.67078062590281e-05, 6.52755916397041e-05, -3.11762792304641e-05, 3228504.07504198, 6457008.15006673, 5.10711104196998e-05, 5.61324069349705e-05, 3.42391081384428e-05, -3.94892069024685e-05, 0, 597871.125099721, -5.09221891985709e-05, 597871.125039932, 1793613.37509072, 5.04957195530634e-05, 7.78282842489551e-05, 5.15933991752743e-05, -2.78182258629044e-05, 1793613.37514566, 0.000183394610415442, -4.43169590447624e-06, 0.000199879369594042, 0.000131742117520606, 9.39656857669089e-05, 0, 0, 0, 0, -538084.012498702, 0, 2.35644485522421e-07, -4.62507595338767e-06, -8.72028386266154e-08, 0, -1614252.03750377, 0, 4.06414751468385e-06, -4.68792180582945e-07, -4.65571209466987e-07, -1614252.03750419, -4.18732271253853e-06, -5.78000061182241e-06, 0, -1614252.03750386, 0, 1.29889418479116e-05, -1.56881106989136e-05, 7.05539005588031e-06, -1614252.03751577, -3228504.07502215, -1.70677804554819e-05, -1.7970491702401e-05, -1.47460923169217e-05, 1.15692562221663e-05, 0, -538084.012529641, 0, 2.16902796991132e-05, -4.83975973087601e-05, 1.92096921673285e-05, -538084.012511344, -1614252.03753691, -2.31033259999097e-05, -4.26585197964983e-05, -1.79272584752688e-05, 3.01283663410531e-05, -1614252.03755547, -7.72432906050065e-05, 1.98844417118274e-05, -1.02114230630734e-05, -3.00367448472151e-05, -1.78750735918722e-05, -9.04944500681195e-05, -9.1607026298432e-05, -0.000110569371431866, -9.15896429846357e-05, -4.21494351661767e-05, -2.00666881735237e-05, -0.00011647807538098}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -486.000000001679, 0, 0, 0, 0, 11020.0499999957, 0, -2.84527272485869e-08, 1032.75000007305, -2.20212613647408e-08, 1032.75000002159, 0, 0, 0, 0, -95581.0124997379, 0, 1.59891226209229e-07, -546.750000491043, 8.0523111135802e-08, 0, -22316.5125009548, 5.34075825505579e-07, -22316.5125003396, -546.750000039251, -1093.5000000392, 0, 0, 0, 0, 419042.868748135, 0, -7.44508924062969e-07, 1.49576461136808e-06, -2.0944422094389e-07, 0, 179320.331254307, 0, -2.54501648559761e-06, 11235.7125052369, -1.63419857616489e-06, 179320.331251942, 2.56279197741597e-07, 4.50171672785759e-08, 11235.7125017853, 22471.425002217, -4.96728612037054e-07, 0, 0, 0, 0, -1014904.68749481, 0, 1.27315620754438e-06, -2.78517565466188e-06, -1.76229031540734e-07, 0, -699361.593755654, 0, 3.89922031725572e-06, -1.04906685501949e-05, 4.15325859814334e-06, -699361.593754049, 2.9691448634755e-07, 4.45033676885449e-07, 0, -83037.6562749021, 8.39758333104495e-06, -83037.6562563828, -166075.312508772, -2.45141253470633e-06, -2.3963073091871e-06, -2.67132107881461e-06, 2.70766451770852e-06, -1.23817215637212e-05, 0, 0, 0, 0, 1368460.57499291, 0, -8.731373661627e-07, 3.59593308453041e-06, 2.16868368464938e-07, 0, 1411640.15625059, 0, -4.53019246111801e-06, 7.51168818253508e-06, -3.00487318110525e-06, 1411640.15625629, -1.36890341760302e-06, -9.58937073297676e-08, 0, 282328.031285997, 0, -1.02435615558574e-05, 5.06185025505584e-05, -1.36171236930281e-05, 282328.03126108, 564656.062514479, 1.4489313176349e-07, 7.85760932179251e-06, 3.78810281611659e-06, -1.59823886933721e-05, 1.7746597373206e-06, 8.84275024610564e-06, -1.47818304094561e-05, 4.2872175382324e-05, 2.56825023510473e-05, 7.84901133858929e-06, 0, 0, 0, 0, -956593.799995082, 0, 8.45847090021163e-07, -2.73976912270586e-06, -4.91656570233696e-07, 0, -1404997.14374731, 0, 3.65564629455596e-06, -2.8349578719569e-07, 1.30202355485171e-06, -1404997.1437539, 1.87517931951093e-06, -2.37284848048315e-06, 0, -448403.343766497, 0, 7.07896307607892e-06, -3.8578831373849e-05, 8.78313256004856e-06, -448403.343759204, -896806.687514236, 1.22803532410279e-06, -2.82776097917246e-06, 5.25296953866368e-08, 1.2902925731995e-05, 0, -5.92488718183131e-05, 1.41859169719562e-05, -1.90505811765299e-06, -1.38924754658516e-05, 4.32091956977301e-06, -1.02377141060374e-05, 6.51569241378025e-06, 6.76193871436399e-06, -4.12590826492249e-05, -4.25355234198178e-05, 6.09872304268649e-06, -6.67367596571153e-05, -3.19157009443889e-05, -1.96976426764037e-05, 0, 0, 0, 0, 269042.006248602, 0, -1.57302720738651e-07, 9.27062943108337e-07, 1.11811390898524e-07, 0, 538084.012498997, 0, -1.12445897827285e-06, -1.14923829034472e-06, -1.28682215842568e-07, 538084.012500954, -7.11005473420433e-07, 1.29708550146777e-06, 0, 269042.006251114, 0, -3.16097723783278e-06, 9.300687622057e-06, -2.0513387101301e-06, 269042.006253988, 538084.012504632, -9.82971481254399e-07, -6.00427188988743e-07, 2.47423468409487e-07, -5.96697259889648e-06, 0, 1.88131694462771e-05, 0, -7.00174513101178e-06, 2.82493758547127e-05, -6.17801499755895e-06, 4.11577326836133e-07, 4.22602989293359e-06, -2.34311005255491e-06, 4.49292020914419e-06, -2.38744609106499e-06, -1.07854967526e-05, 1.49040048754889e-05, 1.3655103074276e-05, -1.02280326786448e-05, -4.73793752541084e-06, -8.97779075903029e-07, -3.4103369755415e-06, 2.43653358749973e-05, 3.02728890021852e-05, 2.73044017594404e-05, 1.9487026674494e-05, 5.85455116378357e-06, 1.32779642635118e-05, 3.65322253945637e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 60.7500000015633, 0, 0, 0, 0, -1309.16249999723, 0, 0, -607.50000003884, 2.15315041564687e-08, -60.7500000056016, 0, 0, 0, 0, 10474.8187499753, 0, 0, 546.750000385295, -1.2891750552239e-07, 0, 13030.8749999867, 2.28138594334435e-08, 1248.41250011211, 8.76944046087845e-08, 546.750000088669, 0, 0, 0, 0, -40596.1875000318, 0, -2.88435823525646e-07, -1.71491289701174e-06, 4.18959823332032e-07, 0, -103499.774999286, 0, 1.68839097965441e-07, -11235.7125005334, 6.27057642516114e-07, -9226.40625028142, 8.45713863517057e-08, -5.71122626339616e-08, -2.29228847974789e-07, -11235.7125002363, -2.44915355224373e-07, 0, 0, 0, 0, 81192.3750004504, 0, 4.13694917958715e-07, 3.98677303158043e-06, -8.6207848679661e-07, 0, 396735.468747439, 0, -9.50986440468949e-07, 3.31658729868448e-06, -1.42472689428633e-06, 31369.7812515273, 1.23527747107868e-06, 1.47961972048232e-07, 0, 83037.6562485789, -6.84548484556684e-07, 5.60612893480157e-07, 83037.6562528152, 1.67163869808601e-06, -4.31694967765461e-07, 2.36261016277349e-06, -4.19884532721607e-06, -4.23897175948924e-06, 0, 0, 0, 0, -79716.1500010507, 0, -1.84236909479013e-07, -5.12742522941698e-06, 6.0139049894072e-07, 0, -780553.968746688, 0, 1.02712728433937e-06, -6.62967507159909e-06, 4.50319403980654e-06, -49822.5937515109, -2.62816195806504e-06, 6.16825264967371e-08, 0, -282328.031242887, 0, 4.73817747568967e-06, -2.31613907692306e-06, 3.62681360842991e-06, -5.15409700267604e-06, -282328.031256365, -4.27231506193415e-06, 3.10555260717356e-06, -1.02010702590617e-06, 5.66241547580147e-07, -5.89550485046033e-06, -8.01191799937674e-06, 4.97896692714834e-08, 3.03929009283623e-06, -5.03412364783891e-06, 7.08218540350611e-06, 0, 0, 0, 0, 29893.5562509584, 0, 3.02038529692711e-07, 3.41443038358105e-06, -5.89155965883791e-07, 0, 747338.906248492, 0, -9.43020290068744e-08, 5.80398420188038e-06, -2.38445860614289e-06, 29893.5562510831, 1.37066622271193e-06, -1.29115721291891e-07, 0, 448403.343741069, 0, -3.14048110613525e-06, 5.9464419993077e-06, -5.75501696134398e-06, 3.32501054580055e-06, 448403.343754267, 3.65054158207613e-06, 3.60369056626318e-07, 2.99848741527185e-06, -5.33578213098388e-06, 0, -3.40979220765241e-06, -5.12483920806522e-06, 5.16715145739867e-06, 4.00898515150119e-06, 1.66467104625106e-06, -1.39466654594839e-05, 1.09001512899092e-05, -2.28528965958785e-05, 4.96929991178446e-06, -1.0779048920907e-06, -1.06683317971934e-05, -2.45037242401883e-05, 4.68863630061896e-06, -4.32159856590467e-06, 0, 0, 0, 0, -3.11600473646591e-07, 0, -2.16345674093078e-08, -9.38496851007847e-07, 1.1107737915579e-07, 0, -269042.006249946, 0, -2.15556127130983e-07, -1.88800312405563e-06, 4.48229767200651e-07, -3.18625422558433e-07, -6.14772493934797e-07, 1.68445565589292e-07, 0, -269042.006246698, 0, 4.19160294101438e-07, -2.86706158582121e-06, 1.40383971024743e-06, -1.13724502250414e-06, -269042.006251456, -1.41816969915297e-06, -1.42622866791248e-06, -1.42267087265101e-06, -5.45795111797933e-07, 0, 3.85110557541109e-06, 0, 1.78274164542161e-07, -1.31326370472593e-06, 1.3149921015856e-06, -1.78665580788946e-06, -4.17718473337018e-06, -2.15372090946006e-06, 1.71460332580875e-06, -2.53512194076681e-06, 2.29326088843736e-06, -1.48108403286446e-06, 1.38361118217701e-06, 3.97774715047973e-07, -3.11865827953553e-06, -5.90566193934624e-06, -7.28917952856025e-07, -1.66241785171531e-06, 2.83024573304111e-06, 4.83781300057199e-06, 7.14969349837724e-06, 1.82843773659184e-07, 1.84187766047936e-05, 1.33978763773587e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 48.600000003321, 0, 0, 0, 0, -959.850000069455, 0, -2.17130121958812e-08, -704.700000005587, -2.28732759512215e-08, -48.60000001758, 0, 0, 0, 0, 6652.12500066117, 0, 1.06137568844318e-07, 2624.39999992512, 7.90820599178398e-08, 0, 13869.2249999601, 1.82184624232571e-07, 911.250000158361, 2.16886788471804e-07, 656.100000094523, 0, 0, 0, 0, -20503.1250028204, 0, -1.43173215233741e-07, -1968.2999998931, -2.74326001218299e-07, 0, -95544.5625016077, 0, -1.45784317892464e-06, -51175.7999973299, -1.06989709556329e-06, -5740.87500077326, -1.81973178269108e-06, -1968.30000031669, -1.35888014150412e-06, -12301.8750014013, -2.18032793788352e-06, 0, 0, 0, 0, 28048.2750062346, 0, 4.46602636384385e-07, 4.99813276599958e-07, 2.69334189373111e-07, 0, 291554.437509265, 0, 3.11215133221373e-06, 36905.624990936, 2.46104323852723e-06, 14762.2500011219, 2.81462124012909e-06, 7.40281154840429e-07, 0, 346912.874993135, 3.43733029770397e-06, 3.72778478291133e-06, 77501.8125050688, 4.45018173449474e-06, 7.83574693545675e-06, 36905.6250026295, 1.01383715119833e-05, 1.43513243626513e-05, 0, 0, 0, 0, -13286.0250074251, 0, -3.93292300235109e-07, -1.38441372483423e-06, -5.04752958062722e-08, 0, -391937.737518615, 0, -5.15404453466879e-06, 1.18254072158028e-05, -6.54265048972242e-06, -13286.0250013151, -2.70934350626027e-06, -6.99160485391771e-07, 0, -1029666.93749868, 0, -9.74039781803761e-06, -232505.437474729, -8.30934558070997e-06, -1.33939188945834e-08, -199290.375006484, -7.05407590841535e-06, -1.89943075826721e-05, -4.00474923115327e-06, -7.40472948041339e-06, -6.61145841445431e-06, -232505.437508559, -2.65876110430813e-05, -1.45188702344034e-05, -1.8617499176355e-05, -2.73511562498531e-05, 0, 0, 0, 0, 4.55444934276469e-06, 0, 2.15433683973129e-07, 1.32886773331525e-06, 1.71075364116703e-07, 0, 179361.337516434, 0, 3.03562517487805e-06, -6.72185388466023e-06, 3.01331703659335e-06, 8.90538022338305e-07, 3.23890872375868e-06, -1.02418969909878e-06, 0, 1315316.47500904, 0, 6.74361632608892e-06, -2.6383321469611e-05, 8.94195789699626e-06, 2.46792107971133e-06, 179361.337505821, 5.37999153588793e-06, 1.59715114579863e-05, 8.55740724396783e-07, 4.82428861597806e-06, 0, 597871.124978795, 1.19444911870577e-05, 6.08829744087351e-06, 597871.125018109, 1.42247984484573e-05, 4.62190018183209e-05, 1.07327464744036e-05, 3.80292832029166e-05, 1.02840581783914e-05, 1.54588530216171e-05, 2.83753554260497e-05, 4.54486153332598e-05, 1.75581106641916e-05, 2.54904493265854e-05, 0, 0, 0, 0, -1.12868233063646e-06, 0, -8.98893840869484e-08, -4.13038235805063e-07, 2.41977215805764e-08, 0, -5.34971692527869e-06, 0, -5.77433105962092e-07, 1.24564128064414e-06, -5.77397462267392e-07, -1.59985302594744e-07, -6.00117226440068e-07, 3.44391621886547e-07, 0, -538084.012506161, 0, -2.45691288031623e-06, 8.8397609790515e-06, -2.07655323982461e-06, -1.04788471017369e-07, -1.78179427767658e-06, -1.49966793721356e-06, -5.16571865705399e-06, -6.51726391757364e-08, -8.89947851142282e-07, 0, -538084.012495563, 0, -3.52696327585191e-06, 1.38600188943512e-05, -4.02990818150419e-06, -1.9343649675558e-06, -538084.012505176, -2.61084359259756e-06, -1.34314920473202e-05, -3.67849287348385e-06, -6.25998383760191e-06, -6.65901001297285e-06, -1.10516848940457e-05, -4.86659255161023e-06, -1.6407416997369e-06, -4.33468727894737e-06, -1.20976905161324e-05, -1.15478244598055e-05, -9.95723103219592e-06, -1.79083812244766e-05, -2.50725372881643e-05, -1.9366480339924e-05, -2.86753912919156e-05, -1.26038509441481e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 45.5625000094599, 0, 0, 0, 0, -797.343750184303, 0, 0, -797.343750064421, 5.08876656820741e-08, -45.5625000388087, 0, 0, 0, 0, 4442.34375142386, 0, 4.257697145028e-08, 4442.34375015117, -3.17308038221476e-07, 0, 13907.9531261084, -1.71806834116752e-07, 751.781250301641, 5.03006222831621e-07, 751.781250333963, 0, 0, 0, 0, -9226.40625535255, 0, -4.62026026512056e-07, -9226.40625021464, 1.33161517950259e-06, 0, -76989.2343831693, 0, 1.19138080833843e-06, -76989.2343760798, 1.54285961351945e-06, -3690.56250089985, -1.95386743205711e-06, -3690.56250076763, -2.05634449639622e-06, -12404.3906272944, -1.94016538977657e-06, 0, 0, 0, 0, 5535.84376073427, 0, 8.77721738544575e-07, 5535.84375043065, -2.63157007949859e-06, 0, 157771.546902424, 0, -1.69794885900219e-06, 157771.546872276, -4.78586818391083e-06, 5535.84375220156, 7.62668880458647e-06, 5535.84374958862, 0, 420724.125010489, -3.35138655723159e-06, 4.1485018219758e-06, 60894.2812572029, 9.30826957703834e-06, 9.39496544156265e-06, 60894.2812576569, -2.65029418766434e-06, 1.86601962254641e-06, 0, 0, 0, 0, -1.18756150624239e-05, 0, -1.37796111197832e-06, -7.88095308604322e-07, 2.98292034798082e-06, 0, -91341.4219196184, 0, 4.00575347044676e-06, -91341.421868992, 1.17775184630554e-05, -2.3397083544613e-06, -1.27817543757581e-05, 3.33345513927559e-06, 0, -838680.328161885, 0, 9.80586216778175e-06, -838680.328119479, 1.07391500060928e-05, -1.01019858860213e-05, -91341.4218887408, -1.98624002780544e-05, -1.56195446009921e-05, -91341.4218853023, 1.02779906845441e-06, -1.35263456091488e-05, -298935.562522645, -1.53243648366065e-06, -2.35910975912457e-05, -2.98834478658562e-05, -7.36626576226943e-06, 0, 0, 0, 0, 6.83607066128114e-06, 0, 1.03903043679475e-06, 5.82458563589217e-07, -2.08960240796386e-06, 0, 3.43786254223825e-05, 0, -2.08070798531828e-06, -3.56137601070935e-06, -7.36263479150181e-06, 1.22123159135693e-06, 7.99816113729871e-06, -1.61855124313611e-06, 0, 448403.343797402, 0, -6.8087600524587e-06, 448403.343733803, -1.280238669687e-05, 2.83932655050691e-06, 9.85439356439117e-06, 2.09543712871633e-05, 1.60966063887108e-05, 6.93761849020034e-06, -7.56542232465025e-06, 0, 1569411.70313278, -1.45445021597248e-05, 8.53245198644412e-06, 448403.343762966, 1.01326893211915e-05, -1.00943430992269e-05, 448403.343767316, -2.00883069117737e-05, 2.43935622622e-05, 4.85350940942668e-05, -7.63476275221801e-06, 6.89387902211939e-06, 3.55547274837892e-05, 4.02424110021667e-05, 0, 0, 0, 0, -1.60425969939979e-06, 0, -2.4914863098115e-07, -1.81242424833973e-07, 4.793305041718e-07, 0, -1.01602490967226e-05, 0, 1.06469189700299e-07, 6.27077408967706e-07, 1.9455869791916e-06, -3.78201206067169e-07, -2.8591473913996e-06, 7.34722206243995e-07, 0, -2.02010793654931e-05, 0, 2.87186983197945e-06, 6.90689057060013e-06, 3.40868843956032e-06, -1.23671459484766e-06, -3.14930629571539e-06, -8.14695822116255e-06, -6.12015421420665e-06, -1.47618112376281e-06, 3.84199081568159e-06, 0, -672605.015636227, 0, 4.30743133222994e-06, -672605.015618867, 5.01387794054024e-06, -1.66466887060701e-06, -7.88485976503214e-06, -1.05193178429868e-05, -4.18132463001416e-06, -6.64377716076673e-06, 3.59310976179472e-06, -4.97161087121811e-06, -1.65344343247663e-05, 5.95030182423679e-07, -7.98665177935439e-06, -672605.015637935, -4.96979817930681e-06, -6.57221062774418e-06, -1.7892721029421e-05, -1.13075081644126e-05, -1.16022181430748e-05, -1.5904629630103e-06, 6.71879254464303e-06, -2.49137676116396e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 48.6000000116984, 0, 0, 0, 0, -704.700000197164, 0, 4.74669881327551e-08, -959.850000129814, 5.30966706403662e-08, -48.6000000050132, 0, 0, 0, 0, 2624.40000135158, 0, -2.0208005555646e-07, 6652.12500058686, -4.05130862803502e-07, 0, 13869.225001672, -8.97985806832353e-07, 656.100000129726, -1.94273498946385e-07, 911.249999837773, 0, 0, 0, 0, -1968.30000470774, 0, 8.38434060918102e-07, -20503.1250016159, 1.50704372960646e-06, 0, -51175.800009505, 0, 2.55488108542307e-06, -95544.5625053994, 4.80352003947136e-06, -1968.30000109345, 3.96228569905822e-07, -5740.87499929681, -3.70249856180674e-07, -12301.8750017307, 2.71572144921002e-06, 0, 0, 0, 0, 9.06002922689956e-06, 0, -1.41831803051342e-06, 28048.2750028527, -2.46486322493364e-06, 0, 36905.6250272923, 0, -2.44699016887828e-06, 291554.437509451, -1.3981040750208e-05, 1.06742125669151e-06, -1.91621922869213e-06, 14762.2499979028, 0, 346912.875022746, -1.28780288241902e-05, 4.22454060881334e-06, 36905.62500469, -9.75152349522301e-07, -8.84896063288927e-06, 77501.8125024464, -1.13257559383156e-05, 2.25786988723127e-06, 0, 0, 0, 0, -9.73345510515604e-06, 0, 1.16254043270812e-06, -13286.0250029475, 2.7259629920101e-06, 0, -4.02415610131733e-05, 0, 2.3636496965806e-06, -391937.737510885, 1.79913959653133e-05, -1.80024545973946e-06, 2.88912120509197e-06, -13286.0249972839, 0, -232505.437549777, 0, 1.21748199878581e-05, -1029666.93752459, 2.85977299392508e-05, -3.42268753640428e-06, -8.56409122823855e-06, 3.71615878420781e-06, 7.50554027561505e-06, -199290.375008883, 3.4753003001645e-05, -9.94235696111794e-07, -232505.437513794, 5.20770088906293e-05, -2.02586376363777e-05, 2.96372093992559e-05, 9.33470747784434e-06, 0, 0, 0, 0, 5.50022910704668e-06, 0, -9.03085154799942e-07, 1.64376566861694e-06, -1.26950516274096e-06, 0, 2.93397366598583e-05, 0, -1.99836457491572e-06, 179361.337507527, -1.21096538705919e-05, 7.0277000462196e-07, -3.40201948210379e-06, -3.75361918864032e-07, 0, 5.09646548157615e-05, 0, -3.43750151262947e-06, 1315316.4750136, -3.03552577046134e-05, 2.66013697416765e-06, 8.50435047880608e-06, -1.00035132279883e-06, -8.02712351627718e-06, 179361.337506019, -2.42934959909031e-05, 0, 597871.125035167, -2.33971569218405e-05, 4.51935121945108e-07, 2.00968446782364e-05, 1.37702544433679e-06, -2.95932050777892e-05, 597871.125007979, -4.70938969186204e-05, 1.98670741159367e-05, -1.37403634770717e-05, -1.55167586591344e-05, 2.80456943702091e-05, 1.24927901365849e-05, -3.60802271671439e-05, 0, 0, 0, 0, -1.27502626837238e-06, 0, 1.9143254206244e-07, -3.43878621924563e-07, 2.60246429769467e-07, 0, -8.35446133889562e-06, 0, 1.04716042206339e-06, -2.31402492877478e-06, 3.5752322090352e-06, -7.10530065806697e-08, 1.17182160149652e-06, -2.08288298954672e-07, 0, -1.91543218468808e-05, 0, 1.41658756221915e-06, -538084.012503674, 1.19176735218467e-05, -9.62444530020142e-07, -2.48575100776947e-06, 4.50565509841759e-07, 3.38681309671055e-06, -1.54244126418438e-06, 1.24662188703228e-05, 0, -1.87178471235592e-05, 0, 7.66270515407982e-06, -538084.01250802, 1.50681569090481e-05, -1.98726704759389e-06, -6.37083871324339e-06, -1.80597360497203e-06, 1.02467709189068e-05, -538084.012508144, 2.73341703795135e-05, -9.31091877520637e-06, 2.14768926098963e-06, 9.02851592050309e-06, -1.89427820772517e-06, -7.01168693467956e-06, 2.93842349674463e-05, -3.30306193197512e-06, -1.64080594526485e-05, 5.71945008597358e-07, 1.47899441120938e-05, 2.79328752629118e-06, 1.40954549902578e-05, 5.94377945948595e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 60.7500000027003, 0, 0, 0, 0, -607.500000048239, 0, 4.33021014073157e-08, -1309.16250002749, 4.55434237979477e-08, -60.749999980372, 0, 0, 0, 0, 546.75000032429, 0, -2.90311248063049e-07, 10474.8187501259, -3.0807035013588e-07, 0, 13030.8750004325, -4.21053691897855e-07, 546.749999987456, -3.88991349762846e-07, 1248.41250009265, 0, 0, 0, 0, -1.09655116853617e-06, 0, 9.01126364772763e-07, -40596.1875002647, 6.93166626180669e-07, 0, -11235.7125023776, 0, 1.81239670725875e-06, -103499.775001571, 3.07792426029619e-06, 3.41514334900192e-07, 2.73399619675465e-06, -9226.40625071543, 1.07192136074147e-06, -11235.7124999843, 2.70272381165274e-06, 0, 0, 0, 0, 2.04205932973674e-06, 0, -1.50780732917879e-06, 81192.3750002127, -1.39932146674711e-06, 0, 6.28572044614645e-06, 0, -5.77503736420868e-06, 396735.46875299, -1.07784700554505e-05, 3.24433303499544e-07, -7.838746743249e-06, 31369.7812538147, 0, 83037.6562568986, -1.12917289814931e-05, -2.17431300334485e-06, 2.98250269761751e-06, -6.60019764949689e-06, -1.35764435283151e-05, 83037.6562565558, -7.04763474436864e-06, -2.78368494790995e-06, 0, 0, 0, 0, -2.16319878775432e-06, 0, 2.38630784312644e-06, -79716.1499999237, 8.60351703609142e-07, 0, -8.63067163636676e-06, 0, 5.04927197360105e-06, -780553.968752997, 1.47528886080152e-05, 3.07796249119268e-07, 1.35994289732465e-05, -49822.593759157, 0, -1.38828425697199e-05, 0, 1.4789018312826e-05, -282328.031259495, 2.76590228013517e-05, 3.43733526608786e-07, -1.50106398477002e-06, 1.54157417458117e-05, 3.25976752792357e-05, -282328.031260587, 7.40756286946452e-06, 7.31134967195604e-06, -3.38142427390292e-06, 1.26598045615536e-05, 1.47361084695099e-05, 1.17102396500191e-05, 1.62136302597957e-06, 0, 0, 0, 0, 1.22276260895896e-06, 0, -1.41218314014579e-06, 29893.5562497397, 6.68647449780346e-08, 0, 5.80308016525949e-06, 0, -3.39904453370582e-06, 747338.906251479, -1.19447131834833e-05, -1.0625675075856e-07, -8.17722094273766e-06, 29893.5562549464, 0, 1.29876968823475e-05, 0, -1.1721499952937e-05, 448403.343756474, -2.60048392085897e-05, 2.37768687408626e-06, -6.24612677443525e-08, -2.23784088292023e-05, -2.8676246954405e-05, 448403.343764657, -6.91133543169061e-06, 0, 1.22889949322481e-05, -2.07248931315487e-05, 3.0754083953008e-06, 5.9055845128301e-06, -9.24390734467447e-06, -3.40876536704038e-05, 2.11635679920832e-05, -3.51526192435678e-05, 8.41388742131238e-07, -5.87332340498279e-05, 1.90969788945238e-05, -3.77375616612045e-05, -2.11195239639493e-06, -6.16847421946922e-05, 0, 0, 0, 0, -2.87987867323349e-07, 0, 3.49678222546145e-07, 1.05237304456187e-07, -4.00300343823233e-08, 0, -1.5405760055759e-06, 0, 8.58393376001073e-07, -269042.006250179, 3.12583066159337e-06, 4.95824440400623e-08, 2.66522967876269e-06, -1.7301895191569e-06, 0, -4.45120472720758e-06, 0, 2.2076245142e-06, -269042.00625184, 1.00249832309832e-05, -1.2969958448442e-07, 6.79264709934303e-07, 8.53864302476803e-06, 9.14011913856413e-06, -269042.006256368, -3.05959126599604e-06, 0, -5.9804499969533e-06, 0, 5.60543920180104e-06, -3.91284611636851e-06, 1.03167163723836e-05, 0, -1.15262065722863e-06, 1.17599481225349e-05, 1.72900569013982e-05, -4.72196748242981e-06, 1.73074883255131e-05, -2.72679430774335e-06, 2.87126688237945e-05, -9.14883654808766e-06, 5.74070276232925e-06, -4.45895070180973e-06, 2.84490386393922e-05, -2.79918710027105e-06, -1.71098501540893e-06, 1.93190905563252e-05, 4.33090876224061e-05, -2.25454228012958e-05, 2.41518135777112e-05, 2.33159910642457e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -485.99999999712, 0, 0, 0, 0, 1032.75000001112, 0, 0, 11020.0499999081, 7.36472131329799e-08, 1032.74999989068, 0, 0, 0, 0, -546.750000198282, 0, -2.28416951292502e-07, -95581.0124991899, -5.57010045385425e-07, 0, -22316.5124996759, 5.29655893530419e-07, -1093.49999831495, -546.749998001922, -22316.5124996853, 0, 0, 0, 0, 9.46597226064001e-07, 0, 1.54760702235134e-06, 419042.868746663, 2.69457201615604e-06, 0, 11235.7124996697, 0, -6.88389881374971e-07, 179320.331246819, -6.59648947435503e-06, -6.49987386617206e-06, -1.15630471735313e-05, 179320.331254786, -1.70550288456911e-05, 22471.4249894107, 11235.7124857591, 0, 0, 0, 0, -2.12467189721054e-06, 0, -2.40489877945294e-06, -1014904.68749227, -7.635929855473e-06, 0, -1.86358194592396e-06, 0, -6.2910258729193e-06, -699361.593740617, 1.48569394911721e-05, 1.27772272875344e-05, 2.718221231734e-05, -699361.593776345, 0, -83037.6562413897, 2.00086172431189e-05, 4.42978830100914e-05, 4.47404205988233e-05, 6.84205182341299e-05, 7.93646836739617e-05, -166075.3124842, -83037.6562250829, 0.000101926885276134, 0, 0, 0, 0, 2.44490145079899e-06, 0, 2.01071208518305e-06, 1368460.57498967, 1.14387816294783e-05, 0, 5.85327610919582e-06, 0, 1.09839912497114e-05, 1411640.1562367, -1.959756871895e-05, -1.41445638722599e-05, -2.90721585981549e-05, 1411640.15629731, 0, -1.07178000014715e-05, 0, -2.73142449433531e-05, 282328.031229422, -5.7250800336519e-05, -5.83407961788429e-05, -6.95296485400668e-05, -0.00012091817750241, -0.000141036094938076, 564656.062516717, 282328.031259775, -0.000123678725428319, -7.21063674363089e-05, -0.000188215281160161, -0.000175536357441364, -0.000349223844357512, -0.000144106216434199, 0, 0, 0, 0, -1.41061818087956e-06, 0, -8.77182044402079e-07, -956593.799992624, -7.41354008717362e-06, 0, -5.83556873295752e-06, 0, -1.33354014866057e-05, -1404997.14374007, 8.28008090113892e-06, 7.45854354797569e-06, 1.95281087699547e-05, -1404997.14379226, 0, 3.72792758825687e-06, 0, 5.30682057045696e-06, -448403.343733013, 5.91187272542144e-05, 4.15746243676711e-05, 5.56370561000207e-05, 9.16106502304348e-05, 0.000101192215657473, -896806.68754697, -448403.343802896, 0, 2.34245336084081e-05, 6.31014745521703e-05, 0.000111348755014927, 0.000108067245915993, 0.000150971048462541, 0.000247923432966937, 4.56946727975896e-05, 0.00015503255978596, 0.000166056053467494, 0.000412120588019271, 6.61532146991845e-05, 0.000390038870346635, 0.000182344390676334, 0.000380125749562221, 0, 0, 0, 0, 3.22469709491354e-07, 0, 2.10459031670546e-08, 269042.00624785, 2.4253561791309e-06, 0, 1.9343091201795e-06, 0, 3.89528813900191e-06, 538084.012496694, -6.23298574406996e-07, -1.44426613131674e-06, -4.80684694344438e-06, 538084.012513212, 0, 6.2038056114078e-07, 0, 4.02512749918733e-06, 269042.006245389, -2.06507974524148e-05, -1.14784456522275e-05, -1.61217741819685e-05, -2.76593663352152e-05, -2.7649641448751e-05, 538084.012524579, 269042.006275717, 0, -8.1382209252811e-06, 0, -1.57241083799281e-05, -1.11780838049785e-05, -3.54786935498294e-05, -3.36249284908782e-05, -4.2632178429594e-05, -5.7085681480977e-05, -9.0089981153642e-05, 1.12447176758553e-06, -3.21613560577593e-05, -5.8211319321446e-05, -0.000167184872846746, 2.10482504580123e-05, -5.59349074003546e-05, -4.17904276907029e-05, -0.000112669358544904, -0.000149137861380084, -9.34196848678463e-05, -0.000248222437732798, -0.000284711664439952, -6.33465176162911e-05, -0.00017039010040471, -0.00023981718161986}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 1360.80000001125, 0, 0, 0, 0, -4641.30000031882, 0, 4.6735072072438e-07, -30156.2999999911, 1.27218234628009e-07, -4641.29999962237, 0, 0, 0, 0, 5248.80000244986, 0, -2.40175679768295e-06, 253017.674999149, -5.99949719766997e-07, 0, 97521.9750017645, -7.13936712117006e-06, 10497.5999949993, 5248.79999039232, 97521.9750014105, 0, 0, 0, 0, -1968.30000909321, 0, 5.86761690337882e-06, -1061979.86249526, -1.57271602058604e-06, 0, -104319.900013103, 0, 2.63574066110426e-05, -750988.462501346, 5.19006864663858e-05, -5904.8999803341, -1968.29993901743, -750988.462532895, -5904.89993838866, -208639.799979429, -104319.89994703, 0, 0, 0, 0, 1.81505559340668e-05, 0, -1.20310495332908e-05, 2438723.6999874, 1.23016252331001e-05, 0, 36905.6250430784, 0, -3.40336033910966e-05, 2760540.74999493, -0.000142875055255633, -3.83052653799532e-05, -0.000162946952913549, 2760540.75014009, 0, 730731.375015514, -0.000180365239840818, -0.000144973070871885, 110716.874895609, -0.000286449058889764, 36905.6246648302, 1461462.75003136, 730731.374929077, 110716.874712323, 0, 0, 0, 0, -1.99137925281997e-05, 0, 1.71349656426517e-05, -3095643.82498153, -2.60848675464339e-05, 0, -6.96615462296266e-05, 0, 1.90097222138752e-05, -5154977.69998887, 0.000190512784443234, 4.3646215066972e-05, 0.000219272489858416, -5154977.70024752, 0, -232505.437549848, 0, 0.000242457028293299, -2291839.312495, 0.000443242418885401, 0.000178089086618011, 0.000185336729486423, 0.000565266461173747, 0.000658198688828538, -4583678.62522986, -2291839.31257054, 0.000474319198389985, -697516.312378446, -232505.436786712, 0.000485428510945011, 0.00120619797209019, -697516.312211181, 0, 0, 0, 0, 1.13776042508451e-05, 0, -1.21882593780426e-05, 2032761.82498587, 2.10119158141456e-05, 0, 5.36251433520577e-05, 0, 3.96967701031563e-06, 4663394.77499021, -0.000117036098593632, -2.39224879510156e-05, -0.000154731722053192, 4663394.77520328, 0, 6.55761577739115e-05, 0, -0.000121277507973382, 3228504.07498731, -0.000440573066592225, -0.000116147823929823, -0.000157633126897838, -0.000502185690611973, -0.000537957738669228, 6457008.15033466, 3228504.07522615, 0, 597871.125002981, -0.0004403900371613, -0.00036777580721054, -0.00024513130167126, -0.000627469471916012, -0.00101035183757086, 1793613.37504339, 597871.124355297, -0.000400172953748566, -0.00165273568431053, 1793613.37512265, -0.00127851090300018, -0.000345002299457271, -0.00137787033999789, 0, 0, 0, 0, -2.64969245169791e-06, 0, 3.68663705617297e-06, -538084.012495661, -7.12059516816185e-06, 0, -1.57352928058697e-05, 0, -1.33705697085522e-06, -1614252.03749603, 2.58387739743957e-05, 5.00266455368861e-06, 4.4384096568996e-05, -1614252.03756328, 0, -2.85602954477004e-05, 0, 1.38176606725863e-05, -1614252.03749705, 0.000156344746060562, 3.27075612246603e-05, 4.9735502962759e-05, 0.00016934493133434, 0.000167243868752482, -3228504.07514873, -1614252.03761977, 0, -1.06368105405922e-05, 0, 0.000130211115190535, -538084.012489204, 0.000237469991089424, 0.000107737938723311, 0.000114496479839632, 0.000283037003432667, 0.000402630702041492, -1614252.03758689, -538084.012317261, 0.000131472019133758, 0.000722477506729736, -1614252.03772925, 0.000230874627428316, 7.5810021728585e-05, 0.000526722958006405, 0.000425792147211919, 0.000165332509409877, 0.000903307408843248, 0.00109471821177963, -3.1242963289964e-05, 0.000521906841186317, 0.000735019574833925}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 1.14291447061886e-07, -2551.50000006527, 0, 0, 0, 0, 12529.6875012935, 0, -2.313130998317e-06, 54629.4375004908, -1.59724838290085e-06, 12529.6874995297, 0, 0, 0, 0, -22895.1562594495, 0, 1.31300268100087e-05, -436306.500001074, 1.0242049761376e-05, 0, -252530.156258927, 2.98552539743096e-05, -45790.3124949763, -22895.1562297753, -252530.156258179, 0, 0, 0, 0, 18452.8125342915, 0, -3.9830100626009e-05, 1721134.82812478, -2.82513894467149e-05, 0, 433128.515681993, 0, -0.000120470514667913, 1825803.28126985, -0.000187895255802636, 55358.4374814254, 18452.8123564638, 1825803.28133835, 55358.4374045883, 866257.031263345, 433128.515537125, 0, 0, 0, 0, -5535.84381760939, 0, 7.67677190816447e-05, -3686871.93749421, 3.5581752454191e-05, 0, -326614.781426149, 0, 0.0002190635650125, -6166929.93752009, 0.000530709004527806, -22143.3749643689, -5535.84333286922, -6166929.93782451, 0, -2801136.93760226, 0.000644428073470596, -33215.0623219341, -979844.34371758, -22143.3744500025, -326614.780575081, -5602273.87525047, -2801136.93740419, -979844.34346928, 0, 0, 0, 0, 7.38766337149354e-05, 0, -9.27087354799944e-05, 4367780.71873717, -1.32039354373819e-05, 0, 91341.4221486686, 0, -0.00022258282760066, 10412922.0937639, -0.000730444206830586, -4.25208967029754e-05, -0.000613250727113207, 10412922.0943006, 0, 1860043.50026292, 0, -0.000897694658090768, 7813843.45318438, -0.00148819443763708, -0.000174037643779696, 365365.687389273, -0.00121180425124055, 91341.4204404114, 15627686.9070167, 7813843.45325611, -0.000807060499517166, 5580130.50017096, 1860043.49853713, 548048.530997749, 365365.685573001, 5580130.50010221, 0, 0, 0, 0, -4.22382903392812e-05, 0, 5.98918496666639e-05, -2690420.0624877, -6.62046050279822e-06, 0, -0.000205564770487926, 0, 0.000124548084932031, -8519663.53125823, 0.000483777665011033, 2.48596371431677e-05, 0.000457526789952322, -8519663.53168858, 0, -448403.344052176, 0, 0.000544169431802638, -9416470.218754, 0.00146757310724797, 9.36170507924716e-05, 0.000122749728836256, 0.00117443658904194, 0.00127887800948159, -18832940.438417, -9416470.2191444, 0, -4035630.09386587, 0.00145806544556309, 0.00053936426398354, -1793613.37494027, 0.00122789271874687, -448403.341612428, -12106890.2817707, -4035630.09225457, 3.11778198608689e-05, 0.00303800031156384, -12106890.2821904, 0.00166350176423353, -2690420.06293689, -1793613.37291593, 0, 0, 0, 0, 9.88000286733662e-06, 0, -1.59690554772408e-05, 672605.015620633, 5.5298283185771e-06, 0, 5.95875648094129e-05, 0, -3.56536633660025e-05, 2690420.0625016, -0.000121499976624374, -5.49804235073526e-06, -0.000137456677804364, 2690420.06263483, 0, 0.000121964241422756, 0, -0.000120664358891711, 4035630.09374955, -0.000523170708770369, -2.52389322571337e-05, -4.68316723316801e-05, -0.000422427362891351, -0.000434345929469941, 8071260.18787581, 4035630.09396218, 0, 672605.01570917, 0, -0.000464513536309747, 2690420.06249983, -0.00076134192487146, -0.000140524549682611, -7.84783995008353e-05, -0.000611108306537921, -0.000887162408600471, 8071260.18784061, 2690420.06196229, 2.25676738758445e-05, -0.00143607837788078, 8071260.18820271, -0.000425571919658361, 2690420.06255499, 672605.014396415, -0.000386903083231545, 0.000272333517583984, -0.00149119872313983, -0.00193954812367139, 4035630.09439077, 2690420.06178791, -0.000773387424876566}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -2.9011477609837e-07, 3402.00000013645, 0, 0, 0, 0, -22829.850002523, 0, 4.73559920186698e-06, -68756.8500011551, 4.28964368023542e-06, -22829.8500006682, 0, 0, 0, 0, 60598.1250185914, 0, -2.69457629962596e-05, 506582.100003922, -2.87384848438501e-05, 0, 429399.225017017, -5.89701256572735e-05, 121196.250007748, 60598.1249880207, 429399.225022002, 0, 0, 0, 0, -79552.1250680877, 0, 8.43796302734712e-05, -1822235.73750842, 9.30755464964917e-05, 0, -1056731.06261025, 0, 0.000239437865752776, -2799414.67503558, 0.000352840307646334, -238656.37503002, -79552.1248686555, -2799414.67513874, -238656.375003076, -2113462.12513671, -1056731.06251079, 0, 0, 0, 0, 51667.8751351647, 0, -0.000160434954776854, 3579845.62501258, -0.000159631083482406, 0, 1280625.18784726, 0, -0.000472441152638959, 8336980.68752998, -0.00100455857296919, 206671.500055545, 51667.8745484592, 8336980.68791226, 0, 5986092.37518718, -0.00116345477102405, 310007.250126464, 3841875.56284397, 206671.49969335, 1280625.18708112, 11972184.7506324, 5986092.37515011, 3841875.56300749, 0, 0, 0, 0, -13286.0251486395, 0, 0.00018443267147449, -3932663.40001057, 0.000147118783912735, 0, -763946.438045035, 0, 0.000538243648057181, -12522078.5625183, 0.00141718484760658, -66430.125051481, -13286.0242565257, -12522078.5631372, 0, -6144786.56300959, 0, 0.00165022821919585, -13983541.3125746, 0.002608844409121, -132860.250256993, -3055785.75045024, -66430.1240293485, -763946.436350864, -27967082.6263454, -13983541.3128206, -132860.249768057, -18434359.6885331, -6144786.56144869, -4583678.62656388, -3055785.74972048, -18434359.6893531, 0, 0, 0, 0, 8.54846070280293e-05, 0, -0.000114360099332236, 2271910.27500317, -6.84195515730098e-05, 0, 179361.337912333, 0, -0.000344172800824436, 9267002.43751758, -0.000973888183701916, 2.26849693486613e-05, -0.000600862000318016, 9267002.43798303, 0, 2989355.6255987, 0, -0.00108356669108232, 14348906.9999553, -0.00257122988978309, 0.000197511726886901, 896806.687786916, -0.00111447360515665, 179361.336302949, 28697814.0013533, 14348907.0004201, 0, 10163809.1252048, -0.00252910051742375, -2.15677961762096e-06, 11957422.5008028, -0.000783309591485796, 2989355.62324369, 30491427.3763473, 10163809.1235383, 1793613.37674933, 896806.685937523, 30491427.3775435, 1793613.37587329, 17936133.7531738, 11957422.5003242, 0, 0, 0, 0, -2.0107576984979e-05, 0, 2.9009280813664e-05, -538084.0124997, 1.21544955906431e-05, 0, -0.000120334751675027, 0, 9.93419079279824e-05, -2690420.06250751, 0.000258766209692909, -3.89744421352329e-06, 0.000190248191249579, -2690420.06264345, 0, -538084.012744137, 0, 0.000283498695667379, -5380840.12497699, 0.000921930131843798, -5.22270014810539e-05, -6.74801303454158e-05, 0.000446919693548432, 0.000466549385146132, -10761680.2505071, -5380840.12518635, 0, -2690420.06266373, 0, 0.000833631177290482, -5380840.12498205, 0.00129578783422876, -3.52149875264048e-05, -2690420.06273362, 0.000498738284658307, -538084.01173788, -16142520.3756473, -5380840.12429895, -0.000653236835696485, 0.000989631473835261, -16142520.3762416, 0.000220177974634694, -10761680.2504585, -2690420.06123364, -0.000803585423415932, -5380840.12664713, 0.000388963454001585, -2690420.0616794, -16142520.3769089, -10761680.2503015, -5380840.12613644}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 3.16970634163348e-07, -3402.00000014889, 0, 0, 0, 0, 30484.3500026549, 0, -4.423914965591e-06, 61102.350001288, -5.09082700608699e-06, 30484.3500032081, 0, 0, 0, 0, -111965.287519984, 0, 2.34019931593827e-05, -386324.43750496, 3.51025896833195e-05, 0, -498289.725016564, 5.83388061359834e-05, -223930.575028224, -111965.287526227, -498289.725039887, 0, 0, 0, 0, 215897.906324477, 0, -7.00211613696384e-05, 1223585.49376394, -0.000118941872034228, 0, 1655381.3063621, 0, -0.000222672497584116, 2663068.89377755, -0.000352791041287045, 647693.718847946, 215897.90632239, 2663068.89392683, 647693.718977718, 3310762.6127749, 1655381.30651619, 0, 0, 0, 0, -230660.156399877, 0, 0.000128413111582167, -2173741.31252867, 0.000216713284835295, 0, -2865721.78161572, 0, 0.000428003787840675, -6751884.09375549, 0.0010192258754382, -922640.625175106, -230660.156315612, -6751884.0941735, 0, -7213204.40641178, 0.00108632116037321, -1383960.93815817, -8597165.34450953, -922640.625617361, -2865721.78196871, -14426408.8134352, -7213204.40700037, -8597165.34548023, 0, 0, 0, 0, 129538.743916837, 0, -0.000142145867882035, 2205480.15003531, -0.000217799670688935, 0, 2723635.12558632, 0, -0.000486982456773906, 8951459.34373948, -0.00146565147563669, 647693.71891825, 129538.743693166, 8951459.3443297, 0, 9599153.06299544, 0, -0.00148547921455139, 13751035.8749935, -0.00246200410452793, 1295387.43839003, 10894540.5010705, 647693.719506454, 2723635.12600556, 27502071.7516179, 13751035.8759592, 1295387.43884664, 28797459.1893215, 9599153.06371145, 16341810.7538269, 10894540.5036324, 28797459.1914998, 0, 0, 0, 0, -29893.5563470274, 0, 8.46386017705303e-05, -1195742.25002212, 0.000114065353761291, 0, -1345210.03170049, 0, 0.000316418398583532, -6008604.80625566, 0.00102573918941883, -179361.337580828, -29893.5561133496, -6008604.80666268, 0, -6277646.81311319, 0, 0.000962406597671625, -12106890.2811193, 0.00246157480348975, -448403.344323389, -6726050.15699317, -179361.337891373, -1345210.03189186, -24213780.5639317, -12106890.2819459, 0, -12256358.0626382, 0.00229590801649847, -597871.126186204, -25110587.2516577, -448403.344906543, -6277646.81370146, -36769074.1894341, -12256358.0629148, -13452100.3161553, -6726050.15959293, -36769074.1915003, -13452100.3178926, -37665880.8808726, -25110587.2552724, 0, 0, 0, 0, 2.30595119488258e-05, 0, -2.05481998313906e-05, 269042.00625538, -2.43390714978965e-05, 0, 269042.006383215, 0, -9.14379986512424e-05, 1614252.03750763, -0.000278849752072862, 1.5089770900195e-05, -6.25673000908701e-05, 1614252.03761729, 0, 1614252.03775714, 0, -0.000256238751062559, 4035630.09369645, -0.000891167449806145, 0.000142253101050689, 1614252.03769914, 5.1302724834122e-05, 269042.006373388, 8071260.1880002, 4035630.09397206, 0, 4035630.09389857, 0, -0.000738611754640112, 5380840.12494101, -0.00118980863902824, 0.000368746709162055, 8071260.18807153, 0.000349581326715801, 1614252.03801225, 16142520.3758018, 5380840.12486421, 4035630.09501368, 1614252.03860815, 16142520.3765628, 0.000501486045050629, 16142520.3758215, 4035630.09380693, 5380840.12754929, 16142520.3778265, 4035630.09614981, 8071260.18991176, 24213780.5653482, 16142520.3772364, 16142520.3790059}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -2.3910115860358e-07, 2551.50000021879, 0, 0, 0, 0, -40085.8875038102, 0, 2.32124916056022e-06, -40085.8875019426, 4.6089978992809e-06, -17122.3875034827, 0, 0, 0, 0, 199550.081279005, 0, -9.06528867795196e-06, 199550.081258081, -3.42747342691153e-05, 0, 605771.662523021, -3.85995738332544e-05, 244998.675032553, 45448.5937899735, 244998.675026729, 0, 0, 0, 0, -468701.437609115, 0, 2.25875637202552e-05, -468701.43752533, 0.000124206550950488, 0, -2793017.70015944, 0, 0.000114772960874844, -2793017.7000336, 0.00025777587179491, -997066.968865991, -59664.0939488549, -997066.9688196, -588029.625277384, -3381047.32526289, -588029.625210762, 0, 0, 0, 0, 575727.750221436, 0, -3.71508532484242e-05, 575727.750056819, -0.000241483068324696, 0, 5984247.09427859, 0, -0.000178976507274837, 5984247.09374117, -0.000799597827965983, 1765934.15645981, 38750.906738308, 1765934.15633678, 0, 10817038.687714, -0.000640554155148454, 1843435.96947252, 12660474.6571081, 691980.469776321, 691980.469654397, 12660474.6570145, 1843435.96894099, 7368208.03269538, 0, 0, 0, 0, -358722.675248592, 0, 3.7053432715336e-05, -358722.675075359, 0.000260491114816799, 0, -6626404.969606, 0, 0.000198801021407565, -6626404.96871438, 0.00121630968040104, -1444855.2189539, -9964.51936294564, -1444855.21881237, 0, -18085601.5319449, 0, 0.00071100508833856, -18085601.531193, 0.00160733144278105, -2192194.12590757, -20277795.6575882, -398580.751843966, -398580.751838523, -20277795.6573937, -2192194.1246208, -1494677.81435554, -37665880.8770124, -1494677.81327635, -21074957.1595499, -7822147.22252458, -21074957.1585786, 0, 0, 0, 0, 89680.6688957566, 0, -1.8372608337283e-05, 89680.6688012908, -0.000146933935297543, 0, 3676907.4194137, 0, -0.000143743284835692, 3676907.41874897, -0.000887395548074095, 448403.34384724, 0.000378313203400362, 448403.343779, 0, 13900503.6571269, 0, -0.000393395321184798, 13900503.6560098, -0.00174341054594271, 896806.68804505, 14797310.3447526, 89680.6703150527, 89680.6703812568, 14797310.3446391, 896806.686824956, 0, 20626553.8126644, -0.0012474165814036, 896806.688937531, 42149914.3146624, 448403.345913807, 448403.345289918, 42149914.3145225, 896806.686870184, 22420167.1907338, 4035630.09834741, 22420167.1889532, 15245713.6926004, 43046721.0041911, 15245713.6911467, 0, 0, 0, 0, -3.49020260522444e-05, 0, 3.1910591348766e-06, -1.37900506170866e-05, 3.42401564241606e-05, 0, -807126.018948067, 0, 4.54303560303024e-05, -807126.018761248, 0.000250454207448673, -1.74280387168959e-05, -9.4161667003376e-05, -5.69865596185097e-06, 0, -4035630.094122, 0, 0.00011044324728319, -4035630.09365184, 0.000667869534460839, -0.000126666934596973, -4035630.0940373, -0.000501114159704638, -0.000521216924806303, -4035630.09403218, 0.000284682682122934, 0, -8071260.18770362, 0, 0.00033479226614714, -8071260.18739688, 0.000717892458094144, -0.000400661243938142, -16142520.375837, -0.000926496540412362, -0.000852626428767763, -16142520.375785, 0.000721736025743526, -8071260.18865065, -807126.020732776, -8071260.18782564, -0.000816504286242182, -24213780.5635981, -2.18870868913325e-05, -8071260.18974895, -24213780.5647688, -4035630.09657689, -4035630.09605641, -24213780.5641608, -8071260.1882703, -16142520.3778728}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -291.600000010325, 0, 0, 0, 0, 6087.15000010112, 0, 2.63968349794856e-08, 3244.05000013501, -7.1928697256695e-08, 619.6500000772, 0, 0, 0, 0, -46391.7375009634, 0, -2.99252315227432e-07, -5904.90000112375, 5.88614800456737e-07, 0, -67058.8874999074, -2.31001590040045e-07, -12274.5375009004, -328.050001069425, -6232.95000055459, 0, 0, 0, 0, 167920.593754531, 0, 1.86838360175191e-06, 2952.45000534194, -2.06124097619538e-06, 0, 503023.668750336, 0, 2.01327289187852e-06, 116621.774997544, -1.47074888028799e-06, 85498.0312532679, 5.59887550436131e-06, 5904.90000082551, 6150.93750537886, 122772.712504649, 2952.45000528194, 0, 0, 0, 0, -306685.743761185, 0, -3.6134177816298e-06, -1.36007770698954e-05, 4.68549391210636e-06, 0, -1777005.84375666, 0, -4.03561068991549e-06, -55358.4374963058, 4.62157033194233e-06, -265720.500008864, -1.73888459517742e-05, -1.5136757701959e-06, 0, -808233.187484958, -2.22388178080139e-06, -38750.9062643403, -846984.093773595, -2.45074038488668e-05, -2.03139568682467e-05, -110716.875017268, -1.39138874319875e-06, -55358.437512094, 0, 0, 0, 0, 269042.006264912, 0, 3.18316497128e-06, 1.86664451688959e-05, -4.54347159629153e-06, 0, 3128858.88751896, 0, 8.76732246016562e-06, 2.46499442682559e-06, -1.02130377002186e-05, 368687.193759284, 2.60094035293532e-05, 1.32841876871266e-07, 0, 2491129.68747616, 0, 3.43559046651384e-06, 348758.15621786, -3.35210958207378e-06, 99645.1875255333, 2590774.87504169, 5.1215109719617e-05, 4.60172979949498e-05, 2.02478054047116e-05, 2.87648332315672e-06, 5.30431326395618e-05, 697516.312557061, 5.24610534009654e-05, 348758.15626917, 6.36600232972923e-05, 1.42939373420007e-05, 0, 0, 0, 0, -89680.668760121, 0, -2.53938820377419e-06, -1.30798938169947e-05, 3.26941479260373e-06, 0, -2600739.39377108, 0, -7.50084124726648e-06, -7.88205083822122e-06, 7.4658164416082e-06, -179361.337506156, -1.91993523085463e-05, 2.7498655342763e-06, 0, -3407865.41248666, 0, -5.8207646332557e-06, 2.26753720178487e-05, 1.08722293477922e-05, -89680.6687703834, -3497546.08128176, -4.56803618028193e-05, -5.34096293083907e-05, -2.04240649713652e-05, 1.83263380868743e-05, 0, -896806.68745659, -3.00632816681642e-06, -4.68103040910226e-05, -1793613.37506257, -6.0953049207943e-05, -5.84587607469118e-05, -7.3684250880719e-05, 1.06566958347002e-05, -896806.687539181, -5.79819866832914e-05, 3.46008190204274e-06, -9.98670582787012e-06, -3.74837195114718e-05, -2.59916012591893e-05, 0, 0, 0, 0, 2.7387101504245e-06, 0, 5.35778759059755e-07, 3.67019096993129e-06, -8.2016244352593e-07, 0, 807126.018758211, 0, 2.50886061333087e-06, 4.05504752658451e-06, -1.83280267244522e-06, 1.54717326673008e-06, 5.77266185129403e-06, -1.35800158121687e-06, 0, 1614252.03749945, 0, 4.50159416805831e-06, -4.88485165354126e-06, -3.87884651229774e-06, 5.30430681737644e-06, 1614252.03751022, 1.59062121107981e-05, 2.26443697639712e-05, 7.99744663834907e-06, -2.15381442706676e-07, 0, 807126.018731241, 0, 7.13306602621521e-06, -1.7483298521216e-05, 1.15223202164198e-06, 1.54897406669113e-05, 1614252.03753013, 2.6948374000826e-05, 3.33359846640515e-05, 2.39849959344724e-05, -2.29703495611731e-07, 807126.018767889, 3.32338818219476e-05, 3.5895536049261e-07, 2.75055051073141e-05, 3.87336741148242e-05, 3.35949216187739e-05, 2.9888488712522e-05, 6.49496137352061e-06, 3.08968132204883e-05, 2.97837761431362e-05, 2.40814879170051e-05, -2.13923774378947e-05, 9.64243806468493e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 3.21240377759702e-08, -291.600000026301, 0, 0, 0, 0, 3244.05000045082, 0, -4.77868968491135e-07, 6087.15000028283, -4.62852774256336e-07, 619.64999987866, 0, 0, 0, 0, -5904.90000308336, 0, 2.75058089725614e-06, -46391.737501291, 3.17795602777461e-06, 0, -67058.8875038414, 5.88985353257836e-06, -6232.95000012554, -328.049996604049, -12274.5374998237, 0, 0, 0, 0, 2952.45001068079, 0, -9.02998499522235e-06, 167920.593753354, -9.59895717845353e-06, 0, 116621.775021663, 0, -2.15879994913036e-05, 503023.6687628, -3.62442139810914e-05, 5904.90000115359, -2.15375525116919e-05, 85498.0312518584, 2952.44999251382, 122772.712504623, 6150.93747427675, 0, 0, 0, 0, -2.04124647152076e-05, 0, 1.67531803609852e-05, -306685.743755224, 1.68083364126713e-05, 0, -55358.4375608571, 0, 4.45295917546434e-05, -1777005.84377305, 0.000111143035994195, -4.08706629016029e-06, 6.44760739594078e-05, -265720.500014941, 0, -808233.187554679, 0.000112602687731637, 2.51412431331272e-06, -110716.875025416, 5.49776972146749e-05, 0.000122101375587566, -846984.093788779, -38750.9061660112, -55358.4374722779, 0, 0, 0, 0, 2.18949043487256e-05, 0, -2.1006608656274e-05, 269042.006254565, -1.44614823429758e-05, 0, 8.81812822878955e-05, 0, -4.57820199666026e-05, 3128858.8875254, -0.000154886933559132, 2.83580459449454e-06, -0.000104436187552351, 368687.193788974, 0, 348758.156366761, 0, -0.000135922668705434, 2491129.68756414, -0.000264495066450466, 9.95462431258041e-06, 3.20234746591956e-05, -0.000133915843791711, -0.00024345699581096, 2590774.87508529, 99645.1873472522, -5.36813447013637e-05, 697516.312563471, -0.000294688567435487, -1.25387296119483e-05, -0.00021847328368728, 348758.156189153, 0, 0, 0, 0, -1.24020132974578e-05, 0, 1.32734050705592e-05, -89680.6687519074, 4.91964135520717e-06, 0, -6.29952285274435e-05, 0, 3.21366203018036e-05, -2600739.39376597, 0.000112811340618673, -8.13673178438004e-07, 7.55315567013829e-05, -179361.337528671, 0, -0.00011665076065371, 0, 8.47926410916988e-05, -3407865.41253866, 0.000266409694747923, -1.88056789472331e-05, -2.18963117937441e-05, 0.000156458952999152, 0.000219669913133724, -3497546.08135028, -89680.6686331477, 0, -896806.687588269, 0.00021459802515428, -1.56855990677145e-06, -7.98252595361175e-05, 8.82710690167254e-05, 0.000358606141654077, -1793613.37512848, 0.000398290601710085, -6.52848283342599e-05, 0.000414310058577339, -896806.687519513, 0.000163093847791374, -3.08339126311618e-05, 0.000469691707555823, 0, 0, 0, 0, 2.89489530151967e-06, 0, -3.31078682074835e-06, 2.52623535652233e-07, -7.42379389222681e-07, 0, 1.76377549682381e-05, 0, -1.00604930474944e-05, 807126.018754122, -3.0375034505075e-05, -1.807207545291e-07, -2.40856779580771e-05, 1.05220808103887e-05, 0, 4.27488132051076e-05, 0, -1.97582424238369e-05, 1614252.03751077, -0.000101428018685039, 4.02890199384969e-06, 3.5319590524309e-06, -6.06141531430972e-05, -7.62257263453673e-05, 1614252.03754069, -2.50024465775371e-05, 0, 4.586004782601e-05, 0, -6.18993163293328e-05, 807126.01877281, -0.000118246927776069, 4.13417699258149e-06, 2.30455183881433e-05, -6.98550346522014e-05, -0.000152167697508347, 1614252.03755241, -0.000192487425187156, 4.06877422966514e-05, -0.000194019316216592, 807126.018771553, -3.06726864551449e-05, 4.44361683132677e-05, -0.000263432062698142, 8.51565900785436e-06, 6.23525838547229e-05, -0.000146951396484566, -0.000321842039072464, 9.10561847388348e-05, -0.000211684681164044, -0.000153356729118422}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 5.26571175741382e-08, -1701.00000014561, 0, 0, 0, 0, 31316.6250025142, 0, 1.35953853197402e-07, 23662.1250012197, -1.65623476631489e-06, 8353.12500217173, 0, 0, 0, 0, -196921.125019939, 0, -4.11510102637137e-06, -90441.5625051862, 1.34988447770541e-05, 0, -425143.687512629, 7.36460720157268e-06, -143294.062517844, -15263.4375296291, -105705.000014328, 0, 0, 0, 0, 556659.843827386, 0, 2.07374240354042e-05, 149672.812518557, -5.22467298597345e-05, 0, 2559610.12509722, 0, -5.09484643214607e-06, 1532608.59375409, -6.86150886120418e-05, 787320.000061053, 12301.8751679746, 311647.500030812, 242962.031410581, 1775570.62513338, 174276.562643979, 0, 0, 0, 0, -787935.093911074, 0, -4.75689556532638e-05, -114407.437546589, 0.000107954856509004, 0, -6759265.21909511, 0, -2.9143651708625e-05, -2367495.84369925, 0.000227645415246243, -1749326.62511115, -3690.56296708169, -346912.875016799, 0, -8224418.53132584, 0.000117805487093203, -1138538.53162701, -9362957.06294256, -180837.563183797, -125479.125693842, -4915829.25037945, -354294.000160651, -2729170.96945913, 0, 0, 0, 0, 548048.53143487, 0, 6.08214431554365e-05, 33215.0625662436, -0.000123531198332344, 0, 8785384.0318331, 0, 5.88013459693206e-05, 1677360.65617561, -0.00037108939477253, 1693968.18760827, 0.000672765309095102, 132860.249967811, 0, 17537553.0003659, 0, -8.19449613380892e-05, 10944363.0935732, -0.000372885823761175, 1793613.37544493, 19331166.3757067, 49822.5951216968, 33215.0640122099, 5081904.56304484, 199290.37489686, 697516.313630591, 22586242.5010835, 132860.250859385, 12339395.7202488, 1826828.43968733, 5181549.75105749, 0, 0, 0, 0, -149467.781360512, 0, -4.25981155593929e-05, -4.74004916927499e-05, 7.62812579272694e-05, 0, -5530307.90671691, 0, -4.99350175888698e-05, -448403.343723982, 0.000276573075675873, -597871.125054195, -0.000479780998089023, 4.26515077174378e-05, 0, -16142520.3755248, 0, 3.94422894549538e-06, -6427114.59347105, 0.000453280457482803, -896806.687746243, -17039327.0630362, -0.00128855966341856, -0.00145295196903699, -1793613.37539928, 0.000331204079754532, 0, -16740391.499959, 0.000241175078476842, -597871.125816579, -34078654.1261705, -149467.782707947, -0.00146401544799381, -19430811.5636046, -0.000107800387670833, -17936133.7514013, -448403.346721936, -2690420.06298107, -6875517.94002145, -19729747.1268285, -1793613.3771241, 0, 0, 0, 0, 2.69077218503827e-05, 0, 1.16901056243156e-05, 1.33017437953632e-05, -1.90169226969998e-05, 0, 1345210.03139313, 0, 1.65499315123175e-05, 3.74182955185132e-06, -8.04312494988096e-05, 1.06550180075033e-05, 0.000139240761020717, -1.77821373073034e-05, 0, 5380840.12523786, 0, 3.20043252539341e-06, 1345210.03114736, -0.000178442444168721, 5.23264524372691e-05, 5380840.12515947, 0.000453572186628649, 0.000516328267334792, 0.000119615094988642, -0.000150130077404278, 0, 8071260.18757886, 0, -4.08699969904368e-05, 5380840.12486511, -0.000159064130397213, 0.000217146323062673, 16142520.3754719, 0.000691149266256607, 0.00077688230877251, 5380840.12542773, -0.000176138077287521, 8071260.18798196, 0.00139302272125139, 3.53256125779265e-05, 0.000541177356322067, 16142520.3756294, 0.000409432594782939, 5380840.12602037, 16142520.3759471, 1345210.03284276, 0.00152775631816264, 8071260.18816342, 0.000434084240221222, 5380840.12637123}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 850.500000049615, 0, 0, 0, 0, -16934.0625007866, 0, -2.41807641256959e-07, -10555.3125004197, 3.15035672987056e-07, -2900.81250063281, 0, 0, 0, 0, 120034.406256724, 0, 2.32774183399575e-06, 29387.8125022801, -2.84540800355327e-06, 0, 206830.968752286, 8.56480941437015e-07, 54424.4062553305, 3280.50000900358, 32668.3125044175, 0, 0, 0, 0, -393660.00002774, 0, -1.06613437078986e-05, -30754.6875102504, 1.13209745271823e-05, 0, -1427222.53127368, 0, -1.02825894073775e-05, -547638.46874109, 5.11462577505396e-06, -346912.875018055, -1230.18755106457, -62739.5625101716, -57818.8125448131, -605457.281286199, -33215.0625472035, 0, 0, 0, 0, 638467.312560797, 0, 2.27643908069692e-05, 11071.6875277667, -2.56829853116391e-05, 0, 4487724.00009942, 0, 2.80902442320501e-05, 540667.406212981, -2.23529404250824e-05, 944784.000037492, 0.000145288013117717, 33215.0625100652, 0, 3448830.65622967, 1.3644417730147e-05, 326614.78135382, 3775445.43763055, 20298.0939489869, 11071.6877152367, 1101632.90636181, 33215.0625738256, 581263.593928423, 0, 0, 0, 0, -498225.937573059, 0, -2.62339013455573e-05, -4.03669863650259e-05, 2.97151574961559e-05, 0, -6842302.87518497, 0, -4.60801992564746e-05, -182682.843708741, 4.74108059676117e-05, -1096097.06253726, -0.000213343245491525, 4.40841519763088e-06, 0, -9200572.31254068, 0, -3.5945760191253e-05, -3039178.21861308, 4.91033001038373e-06, -697516.312629974, -9898088.6252121, -0.000408530052353561, -0.00046110513617798, -548048.531403428, -5.49382059510509e-05, -99645.1878434981, -6178001.62532306, -0.000393261777920914, -3238468.59408005, -182682.844364011, -548048.531538746, 0, 0, 0, 0, 149467.781295478, 0, 1.84582666204402e-05, 2.93182357426269e-05, -1.9928505358808e-05, 0, 4932436.78140906, 0, 3.72523265560964e-05, -1.23086501172744e-05, -3.9376448563625e-05, 448403.343771505, 0.000158039815751287, -1.5941149684297e-05, 0, 10313276.9063606, 0, 3.67751131723486e-05, 896806.687351183, -3.21682098782169e-05, 448403.34382978, 10761680.2501605, 0.000382786365958778, 0.000455711146444356, 0.000118448326322193, -4.23950427578014e-05, 0, 6427114.59363655, 2.54591385214114e-05, 149467.781512682, 13003696.9691, 0.000447030149922487, 0.000541157882671176, 2690420.06286505, 0.000172941416784136, 6726050.15657111, 0.000782224978476717, 0.000154741928444893, 896806.68806208, 2690420.06293386, 0.000594422957694369, 0, 0, 0, 0, -1.14689340620046e-05, 0, -4.76099320771636e-06, -8.3756902931094e-06, 5.19910052651771e-06, 0, -1345210.03130168, 0, -1.22722827634901e-05, -2.46817160507102e-06, 1.19061811499131e-05, -4.93584632565568e-06, -4.68124897643105e-05, 7.49766543142069e-06, 0, -4035630.09381196, 0, -1.62610510661632e-05, 4.88093540180991e-05, 1.52285357746437e-05, -1.70838482502991e-05, -4035630.09379947, -0.000135998895099979, -0.000168713437433849, -3.72844023345191e-05, 1.43434339114433e-05, 0, -4035630.09373205, 0, -2.42936899562528e-05, -1345210.03116837, -5.55778967297044e-06, -7.3354570296336e-05, -8071260.1876479, -0.000207556651411454, -0.000268912222365588, -0.000132031371114132, -3.30498710378189e-05, -4035630.0938634, -0.000377875637993558, -2.82915130726396e-05, -0.000173320388454517, -4035630.09394971, -0.00022749108130098, -1345210.03150221, -4035630.09394034, -0.000405845075450253, -0.000439015999223536, -0.000192221595645013, -0.000134958598682543, -0.000318536876293085}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, -243.000000029134, 0, 0, 0, 0, 4525.87500057395, 0, 8.90211405704251e-08, 3796.87500015918, -8.77431905518215e-08, 516.375000239896, 0, 0, 0, 0, -28476.5625047542, 0, -7.60672399694507e-07, -17085.9375001871, 9.01565542249144e-07, 0, -70166.2500022722, -3.13103235977453e-07, -9066.93750186117, -273.375003360193, -7517.81250174249, 0, 0, 0, 0, 75861.5625187303, 0, 2.94351063556316e-06, 24603.7500004207, -4.00588259414926e-06, 0, 435281.343771631, 0, 3.21333114906638e-06, 310212.281245451, -8.07963039198339e-07, 50847.7500063517, 1.87353505098188e-05, 28294.3125039387, 4510.6875162197, 131425.031265002, 3690.56251997419, 0, 0, 0, 0, -84882.9375390518, 0, -6.53955144881621e-06, -11071.687502759, 9.34707326715797e-06, 0, -1131157.40633252, 0, -9.04550224308837e-06, -428105.249970819, 6.36252607312607e-06, -107026.312511773, -5.05657247187341e-05, -22143.3750001272, 0, -1861888.78125484, -4.88049559621455e-06, -22143.3750357601, -730731.375048705, -6.75771387547558e-05, -8.78090228895014e-05, -488999.531291577, -11071.6875405045, -60894.2813350095, 0, 0, 0, 0, 33215.062544567, 0, 8.31880116428838e-06, 5.97866432656087e-06, -1.19178281048617e-05, 0, 1212349.78139325, 0, 1.14704465795467e-05, 182682.843706632, -1.42997294108751e-05, 66430.1250119132, 7.32415865944863e-05, -1.07862887088469e-05, 0, 4550463.56257578, 0, 1.07173595517354e-05, 2358269.43742297, 1.15185219272427e-06, 33215.0625383618, 1511285.34382754, 0.000133927286135946, 0.000177644637192145, 365365.687554202, 2.93280941838099e-05, 0.000100871306981135, 2657205.00012021, 0.00015574625513832, 298935.562667536, 0.000247472515066044, 182682.843914911, 0, 0, 0, 0, -2.63844556261018e-05, 0, -5.7990739767614e-06, -5.17731381277464e-06, 7.81150587889637e-06, 0, -448403.343865672, 0, -8.81800927530996e-06, 2.49462309815929e-05, 1.37244966511099e-05, -6.49634654573865e-06, -5.44414594831612e-05, 1.31072138763525e-05, 0, -4334565.65637487, 0, -6.66275218215483e-06, -896806.687396974, 4.82881534589154e-06, -2.03116890039223e-05, -896806.687559585, -0.000127143077001389, -0.000164219606880975, -2.94943965647687e-05, 8.50075971258405e-06, 0, -4782968.99996239, -5.29733261519625e-06, -6.98785220966373e-05, -5231372.34387994, -0.000133408014566807, -0.00020984586321253, -1793613.37511795, -0.00010605404497705, -448403.34389653, -0.000314381402357485, -0.000110045016538421, -0.000274504559590796, -896806.687732756, -0.000313080269051412, 0, 0, 0, 0, 6.34923802032236e-06, 0, 1.60393095271139e-06, 1.59428113105126e-06, -2.09198452655016e-06, 0, 3.54325820134329e-05, 0, 3.0101712901542e-06, -4.5146474906512e-06, -4.73472299203503e-06, 1.46960712056878e-06, 1.60835514734557e-05, -5.10438484430784e-06, 0, 1345210.03130904, 0, 1.40048682729474e-06, -3.77867797034355e-05, -3.41155667192986e-06, 3.77218028005682e-06, 1.83186339264369e-05, 4.56893074525581e-05, 5.75960322461743e-05, 5.51751012938951e-06, -1.07749588200588e-05, 0, 2690420.06251285, 0, 1.79709864539437e-06, 1345210.03120122, 7.61998649004463e-07, 1.66073781354044e-05, 2690420.06255188, 6.32584653485432e-05, 9.48413670701565e-05, 4.15595452386843e-05, 1.77459475650808e-05, 5.02786564667899e-05, 0.000141978656371502, 1.92726152018187e-05, 4.58623045534656e-05, 2690420.06256829, 8.14988253051228e-05, 9.8232244168244e-05, 1345210.03136668, 0.000157616440430952, 0.000196033656224663, 0.000102242039128082, 0.000113876697563946, 0.000178580948154657}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 2.28303209332614e-08, -243.000000051318, 0, 0, 0, 0, 3796.87500091226, 0, -2.45592835817449e-07, 4525.87500048658, -4.6511457715538e-07, 516.375000171617, 0, 0, 0, 0, -17085.9375065912, 0, 8.65854427359161e-07, -28476.5625019336, 3.38490844417818e-06, 0, -70166.2500067563, 4.67717867130605e-06, -7517.81250188349, -273.375001373981, -9066.93750057475, 0, 0, 0, 0, 24603.7500238022, 0, -2.3370835839387e-06, 75861.5625049981, -1.27984077017118e-05, 0, 310212.28129228, 0, -1.56051304625344e-05, 435281.343766548, -2.88524583336548e-05, 28294.3125083883, 7.74402266171634e-06, 50847.749998849, 3690.56251216653, 131425.031266914, 4510.68749716351, 0, 0, 0, 0, -11071.6875467445, 0, 3.7213481460671e-06, -84882.937509114, 2.38031144680449e-05, 0, -428105.250129868, 0, 2.10454744590217e-05, -1131157.40626936, 8.67592962320466e-05, -22143.3750142861, -2.35473755308421e-05, -107026.312487971, 0, -1861888.78133176, 7.60511357737094e-05, -11071.6875389237, -488999.531306644, -4.18538513807297e-05, -3.04814043633992e-06, -730731.375043858, -22143.3749406089, -60894.2812909391, 0, 0, 0, 0, 5.09686583348417e-05, 0, -1.92064933807027e-06, 33215.0625103351, -2.62817813149522e-05, 0, 182682.843950381, 0, -2.72491945178177e-05, 1212349.78126657, -0.000132101602565426, 1.68119879690869e-05, 3.59098653749132e-05, 66430.1249757305, 0, 2358269.43770861, 0, -9.65493093620372e-05, 4550463.56255224, -0.000182192203702823, 5.58393884311115e-05, 365365.687599955, 7.90444177705864e-05, 3.88853072243664e-05, 1511285.34382614, 33215.0623307704, 7.48611998483686e-05, 2657205.00015554, -0.00018982493182841, 182682.843942946, 5.17159001873775e-05, 298935.562517433, 0, 0, 0, 0, -2.9083819671815e-05, 0, 1.0817700424497e-06, -6.17238594470037e-06, 1.50115426760881e-05, 0, -0.000150011316442986, 0, 1.87547448175711e-05, -448403.343763034, 9.12798496526443e-05, -7.96634530398429e-06, -1.8636122204391e-05, 1.19494609510918e-05, 0, -896806.687734224, 0, 5.18561357880907e-05, -4334565.65625085, 0.000195206437193222, -3.00820325095549e-05, -8.12862866767738e-05, -8.4428568842394e-05, -4.03256388472051e-05, -896806.687551888, 0.000161368081617378, 0, -4782969.00010801, 0.000163250646058652, -5.25599838409705e-05, -1793613.3751645, -7.57436188715822e-05, 0.000130446784277259, -5231372.3438787, 0.000292911659163806, -0.000206017745049719, -0.00015579410178114, -448403.343675981, -0.00020365417418974, -896806.687720429, 3.81278457556827e-06, 0, 0, 0, 0, 6.78838525599101e-06, 0, -1.24208540722129e-07, 1.45335810796001e-06, -3.41582427023229e-06, 0, 4.34968623359708e-05, 0, -5.66893472982328e-06, 5.06945606331915e-06, -2.59542626668269e-05, 1.6105466423689e-06, 5.71514538999583e-06, -3.17130966043994e-06, 0, 9.31229081596701e-05, 0, -1.85122732134656e-05, 1345210.03124492, -7.17093270465149e-05, 9.61451839026167e-06, 2.439735803806e-05, 3.04953199022282e-05, 1.34354609861528e-05, 1.34346471628079e-05, -7.1261468309141e-05, 0, 1345210.03132269, 0, -5.14614095575943e-05, 2690420.06250751, -8.84032743806241e-05, 1.85163211296441e-05, 6.65628585203679e-05, 5.01664009127782e-05, -2.19310242414511e-05, 2690420.06256342, -0.00015426506489691, 7.54239405651079e-05, 7.28228789665065e-05, -4.44074690204565e-05, 3.90076066962501e-05, 2690420.06258822, -0.000121600956587517, 7.83027230832196e-05, 0.000157755255463627, 8.13104266994673e-05, -1.49355938225327e-06, 1345210.03128701, -9.05024474199032e-05, 8.63545730565636e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -1.24475898283993e-07, 850.500000099563, 0, 0, 0, 0, -10555.312501734, 0, 1.75842238043669e-06, -16934.062501001, 1.94982737714035e-06, -2900.81249994871, 0, 0, 0, 0, 29387.8125122613, 0, -9.55768229040062e-06, 120034.406254308, -1.360094642671e-05, 0, 206830.968763551, -2.29431885478801e-05, 32668.3125035042, 3280.49999309663, 54424.4062506337, 0, 0, 0, 0, -30754.6875436141, 0, 3.09461283155443e-05, -393660.000011387, 4.5566922581231e-05, 0, -547638.468830967, 0, 8.37064866956699e-05, -1427222.53128849, 0.00013976772034571, -62739.5625175532, -1230.18745144633, -346912.875003207, -33215.0625018092, -605457.281290495, -57818.8124355285, 0, 0, 0, 0, 11071.6875848582, 0, -5.87518121797138e-05, 638467.312519618, -8.25161624611682e-05, 0, 540667.406490049, 0, -0.000159897296440074, 4487724.00005809, -0.000421301463089014, 33215.0625365445, -0.000151448124601135, 944784.000017283, 0, 3448830.65642559, -0.000412677114658171, 33215.0625635781, 1101632.90641175, 11071.6874211062, 20298.0934467421, 3775445.43766493, 326614.78096274, 581263.593772769, 0, 0, 0, 0, -9.2053294867092e-05, 0, 6.87846602369249e-05, -498225.937520216, 8.09976099927276e-05, 0, -182682.844111755, 0, 0.000183311903348626, -6842302.87505965, 0.000610380728587147, -3.78619459993548e-05, 0.000251823416399816, -1096097.06255177, 0, -3039178.21916321, 0, 0.000515136322485282, -9200572.31265643, 0.000965968138910033, -0.000129864644416732, -548048.531507526, 0.000246961028105949, 0.000578916936462669, -9898088.62532084, -697516.311883813, 9.12613689636956e-06, -6178001.62541748, -99645.1864885072, -548048.531602782, -182682.843344231, -3238468.59369308, 0, 0, 0, 0, 5.24444157809887e-05, 0, -4.3130725347719e-05, 149467.781260812, -3.86983232164837e-05, 0, 0.000266193113835692, 0, -0.000130414082863792, 4932436.78129108, -0.000437282227906366, 1.7486754837741e-05, -0.000200858127820381, 448403.343797194, 0, 896806.687942436, 0, -0.000319446225009299, 10313276.9063103, -0.000993753977538587, 0.000108494366778578, 0.000197212511176313, -0.000305317677497355, -0.000533170733460582, 10761680.2503143, 448403.343217542, 0, 6427114.59400483, -0.00081601525511044, 9.28302918984935e-05, 2690420.06297277, -0.000123279495713075, -0.00110690953618156, 13003696.9692674, 149467.779819735, 0.00056319929781147, -0.000751870859954392, 6726050.15621863, 1.89355426268186e-05, 2690420.06303068, 896806.686449676, 0, 0, 0, 0, -1.22577566714161e-05, 0, 1.06666647754482e-05, -2.24640472619427e-06, 7.8596854401204e-06, 0, -7.62128276522621e-05, 0, 4.0413615228114e-05, -1345210.03126263, 0.000120083184727368, -2.78659271186615e-06, 6.47667997455503e-05, -1.8677898876389e-05, 0, -0.000170425502135596, 0, 8.74533464566484e-05, -4035630.09376133, 0.000373842213858476, -2.9967471338478e-05, -5.2701577157711e-05, 0.000126603649337112, 0.000194706741018275, -4035630.09386566, 0.000170640073916928, 0, -1345210.03140152, 0, 0.000244573071910068, -4035630.09379228, 0.000444636274012933, -4.40411661468445e-05, -0.00017098851555969, 0.000110470442864161, 0.000427744381288444, -8071260.18772293, 0.000709210896229843, -0.000248931971088366, 0.000372661835002761, -4035630.0937466, 1.92224235384546e-05, -4035630.09400136, 0.000832526235875616, -0.000197809002860961, -0.000454492774843057, 0.00019841089881704, 0.00075564111076965, -4035630.09406646, -1345210.03061788, 0.00015351631999562}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 2.4387203702238e-07, -1701.0000001965, 0, 0, 0, 0, 23662.1250034855, 0, -3.0767134751248e-06, 31316.6250018046, -4.10097305248076e-06, 8353.12500112009, 0, 0, 0, 0, -90441.5625256325, 0, 1.58819358100698e-05, -196921.125007198, 2.94799978535756e-05, 0, -425143.687524069, 4.2004973182344e-05, -105705.00001584, -15263.4375054294, -143294.062509037, 0, 0, 0, 0, 149672.812593798, 0, -5.04837338938664e-05, 556659.84376952, -0.00010419056796421, 0, 1532608.59390534, 0, -0.000147700317013083, 2559610.12505215, -0.000260195503793007, 311647.500064467, 12301.8750074919, 787320.000017161, 174276.562600121, 1775570.62514574, 242962.031249355, 0, 0, 0, 0, -114407.437686136, 0, 9.5719906004161e-05, -787935.093787822, 0.000197423691417137, 0, -2367495.84423813, 0, 0.000275605002152524, -6759265.21879599, 0.000789293599475102, -346912.875122164, -3690.56247233485, -1749326.62500213, 0, -8224418.53152345, 0.000727328218026024, -354294.000333634, -4915829.25051713, -125479.125254815, -180837.562460329, -9362957.06295287, -1138538.53096319, -2729170.9693242, 0, 0, 0, 0, 33215.0627048601, 0, -0.000107826271927635, 548048.531294993, -0.000207990282293421, 0, 1677360.65701362, 0, -0.000334688256235463, 8785384.03127903, -0.00117302710572689, 132860.250124843, -0.000114904600425238, 1693968.18748811, 0, 10944363.0944916, 0, -0.000907954727892094, 17537553.0001185, -0.00171873914282703, 199290.375479246, 5081904.56333103, 33215.0628122374, 49822.5936863962, 19331166.3757208, 1793613.37405731, 132860.250606246, 22586242.5012574, 697516.311529099, 5181549.75173652, 1826828.43839865, 12339395.7195862, 0, 0, 0, 0, -0.000117976310130656, 0, 6.55973396784163e-05, -149467.781277941, 0.000111825643829299, 0, -448403.344326755, 0, 0.000243052626968266, -5530307.90627926, 0.000848434389518816, -5.96437809009797e-05, 0.000136300711070907, -597871.124991942, 0, -6427114.59461064, 0, 0.000573916698237335, -16142520.3749323, 0.00179640476664552, -0.000319902893685112, -1793613.37563637, -0.000152593499780728, 0.000115361178646758, -17039327.0630771, -896806.68651872, 0, -16740391.5003279, 0.00144515221084956, -0.000566011060013799, -19430811.5638932, -0.000539112871298804, -149467.780417007, -34078654.1262981, -597871.123047742, -2690420.06444778, -448403.344555615, -17936133.7504408, -1793613.37697927, -19729747.1273376, -6875517.93777407, 0, 0, 0, 0, 2.77946121684559e-05, 0, -1.56668773721923e-05, 6.85669922783089e-06, -2.51308984165919e-05, 0, 0.000168378647941164, 0, -7.52370083940898e-05, 1345210.03126438, -0.000237683521971279, 1.06161173427495e-05, -5.28215406316441e-05, -1.30326114705933e-07, 0, 1345210.03159804, 0, -0.000170116412102966, 5380840.12495871, -0.000678779610008936, 8.22683598000716e-05, 0.000180872111573893, 1.21090703677867e-05, -7.87538238980095e-05, 5380840.12518375, -0.000368241981931258, 0, 5380840.12524915, 0, -0.000437012447501443, 8071260.18748857, -0.000789395122475402, 0.000178287759298981, 5380840.12553151, 0.0001727666618918, -0.00024948450897134, 16142520.3755174, -0.00112723175617868, 0.000751127610885165, 0.000262116854089998, 8071260.18754897, 0.000234382637840663, 16142520.3757075, -0.000992742636289577, 0.00106184133872094, 8071260.18898022, 0.000785304118619746, 1345210.03125529, 16142520.376, 5380840.1246025, 5380840.1260127}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -5.01895057176023e-08, 756.000000114804, 0, 0, 0, 0, -12784.5000021036, 0, 3.26508627228043e-07, -12784.5000009344, 1.16856654351129e-06, -2578.50000091978, 0, 0, 0, 0, 68343.7500160488, 0, -3.60937528430875e-07, 68343.7500031761, -9.12453083632502e-06, 0, 213232.500012778, -8.631080929142e-06, 40641.7500082754, 2916.00001134646, 40641.7500054312, 0, 0, 0, 0, -144888.750060155, 0, -8.35075216875548e-07, -144888.750008408, 3.53528182436345e-05, 0, -1109082.3750903, 0, 2.37914618724427e-05, -1109082.37501706, 6.01021295506059e-05, -186441.750030968, -1093.50006275884, -186441.750006428, -42646.5000681497, -633956.625068972, -42646.5000474943, 0, 0, 0, 0, 132860.250121316, 0, 4.62363810784178e-06, 132860.250018542, -7.06459951996076e-05, 0, 2236480.87530007, 0, -2.6659398742801e-05, 2236480.87498637, -0.000189819125731493, 280482.750054601, 0.000171083226213013, 280482.74997626, 0, 5447270.25012219, -0.000142467426894186, 162384.75017151, 2841733.12522987, 14762.2502667426, 14762.2502394758, 2841733.12518501, 162384.749989637, 620014.500324584, 0, 0, 0, 0, -44286.7501350651, 0, -1.05161770032132e-05, -44286.7500253826, 8.11234098201231e-05, 0, -1926473.62548523, 0, 3.37651865716083e-05, -1926473.62496274, 0.000297700218394247, -132860.250056922, -0.000243942368852799, -132860.249930074, 0, -9787371.7504021, 0, 0.00016804842627889, -9787371.74996373, 0.000364521357870256, -132860.250208976, -4052237.62537881, -0.000515117936973112, -0.000537404699347918, -4052237.6252642, -132860.249742716, -44286.7504305536, -12068139.3755779, -44286.750063086, -2325054.37582576, -199290.375841486, -2325054.37547639, 0, 0, 0, 0, 7.84207875809681e-05, 0, 8.29002985049891e-06, 1.76363286415857e-05, -4.85763605778124e-05, 0, 597871.125374881, 0, -2.28514816543364e-05, 597871.124984008, -0.000216709994446602, 2.76824795989357e-05, 0.000164791880936988, -5.50804154835817e-05, 0, 7174453.50050778, 0, -9.46169555976157e-05, 7174453.49985378, -0.000401984203122928, 0.000113513260138821, 1793613.37529581, 0.000490831715262327, 0.000501159555519376, 1793613.37516407, -0.000341000433446831, 0, 13153164.750099, -0.000305391635423416, 0.000300446297346091, 14946778.1256284, 0.000522227052374759, 0.000318096129318658, 14946778.1255311, -0.000310542186894038, 1793613.37580164, 0.00119150532472309, 1793613.37514779, 597871.126192102, 8370195.75104355, 597871.125840757, 0, 0, 0, 0, -1.8578318723778e-05, 0, -2.71514224026787e-06, -4.7483151030762e-06, 1.18314301638024e-05, 0, -0.000111288560294274, 0, 6.47566364698089e-06, -1.19073340488974e-06, 6.31321130320106e-05, -5.41252259138906e-06, -4.76171760407807e-05, 1.83795755196088e-05, 0, -1793613.37521502, 0, 3.42455921516939e-05, -1793613.3749387, 0.000152699164770741, -2.7445962359278e-05, -8.83189116865725e-05, -0.000172783591513175, -0.000171390668132534, -3.85458110856408e-05, 0.000158401885736917, 0, -5380840.1251211, 0, 9.37277196057023e-05, -5380840.12494093, 0.000171468857659927, -8.10451663165055e-05, -5380840.12525061, -0.00026125091724538, -0.000209669649726139, -5380840.12521303, 0.000265271069789113, -0.000284908428586642, -0.000539122494673108, 3.98830750208511e-05, -0.000195683624032573, -10761680.2503271, 3.59304706932038e-05, -0.000467611282487427, -5380840.12560774, -0.000638673199507382, -0.000548516649337986, -5380840.12534819, -0.000112107602224322, -1793613.37567555}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104.142857143454, 0, 0, 0, 0, 0, 0, 1.0413091573378e-08, 0, -1.74434371785125e-08, 0, 0, 9.79185072626673e-08, -2296.35000000935, 2.86525014874964e-08, 4.32167007169366e-08, 0, 0, 0, 0, 0, 0, -1.74689870585965e-07, 0, 3.9167505718094e-08, 0, 0, 0, -3.31288184633029e-07, 0, -8.06714362284326e-08, 19026.9000001993, 6.43930339764128e-08, -9.24029828590357e-08, 2.12413840701965e-07, 2.20615749459444e-07, -2.67793199273317e-07, 0, 0, 0, 0, 0, 0, 2.59298444845794e-07, 0, -1.92736883213491e-07, 0, 0, 0, 2.42704504316113e-08, 0, 4.72842425067517e-07, -77501.8125000245, 4.37172301716503e-07, 4.15626568465405e-07, 0, 0, 1.03917823585974e-06, -7.63730410217282e-07, -2.36898327845033e-07, 6.99366843132583e-08, -3.66204152024378e-07, 6.41540642111873e-07, -6.12711098103181e-07, -2.84810084489796e-06, 0, 0, 0, 0, 0, 0, -7.48679119980445e-08, 0, 9.4700813303496e-08, 0, 0, 0, 6.38505372995432e-08, 0, 4.92182111499315e-07, 166075.312500427, -9.97991166849589e-07, -4.78636105250009e-07, 0, 0, 0, 5.76178562901051e-07, 0, -1.19634483215515e-06, 1.75394876729604e-07, -1.5275174583978e-07, -1.2533621720381e-06, 2.10022607987624e-06, 5.77665676369359e-08, -3.3232038073023e-06, -1.30152806490748e-06, -1.5405771153646e-06, -3.97780000762668e-06, 6.7048511927548e-06, 1.36266885659059e-06, 1.60511668201875e-06, 0, 0, 0, 0, 0, 0, 1.8076190437242e-07, 0, -1.92690735531079e-07, 0, 0, 0, 2.727026808425e-07, 0, -1.43079532366358e-07, -179361.337500183, 7.39082766701823e-07, -1.79628389091706e-07, 0, 0, 0, -7.2807164415904e-07, 0, -7.8373604831908e-08, -2.12918130412401e-07, -5.87314870525617e-07, 1.23550395285474e-06, -1.36537361197058e-07, 1.33831323937371e-06, 8.839837526977e-07, 0, 0, 2.72043658698605e-07, 7.12123642800472e-07, -8.49351368212524e-07, 8.96148958341083e-07, -5.37866901972063e-06, 5.26646629644139e-06, -5.58363986086639e-06, -3.76735699151471e-06, -7.11559572183978e-06, -8.74517707689131e-07, -1.73245555832821e-05, -8.4200806042598e-07, -1.6084371741336e-06, 0, 0, 0, 0, 0, 0, -1.65216305955773e-08, 0, 3.39895580291669e-08, 0, 0, 0, -1.71904694738162e-07, 0, 0, 76869.1446428826, -2.98052703232418e-07, 1.84618050668479e-07, 0, 0, 0, -1.52631690842686e-07, 0, 0, 2.08701089820676e-07, 1.66853637901278e-07, -5.7862063611654e-07, -6.14606476638451e-07, -6.7130256623159e-07, -1.50436619087727e-06, 0, 0, 0, -8.51143694864903e-07, 0, -5.31208815018472e-07, -3.27329596576713e-07, -8.80862783277134e-07, -6.85630741097757e-07, 1.05304108751019e-06, -1.1352240176852e-06, -1.21972300996308e-06, 1.3180655453409e-06, 2.53522736149731e-06, -1.91356691595153e-06, -1.26381769334951e-06, -1.91497121554242e-06, -1.02752450795279e-06, 2.43721969387404e-06, 4.88711911366503e-06, 5.55970737455035e-06, 4.08864483601351e-06, -9.65308323032686e-07, 6.86095123175053e-06, 4.54654176255491e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.19028136799037e-08, 0, -4.11369490779606e-08, 104.142857142852, 0, 0, 0, 0, 0, 0, 0, 0, 3.07350377070288e-07, 0, 0, 2.64802667704714e-07, -3.86232376713188e-08, -2296.35000006386, 5.46057579967906e-08, 0, 0, 0, 0, 0, 0, 4.9119047968703e-08, 0, -1.1748530597932e-06, 0, 0, 0, -5.93338570678529e-07, 0, -1.79259220583892e-06, 1.4413779303486e-07, 19026.9000003868, -5.91059420813195e-07, 3.33834939473603e-07, 2.3390608122846e-07, -1.69809350715531e-07, 0, 0, 0, 0, 0, 0, -4.18079397699254e-07, 0, 2.42417247761889e-06, 0, 0, 0, 9.00826883189239e-08, 0, 5.9088190199791e-06, -1.35104388137291e-07, -77501.8125018519, 2.29987693088233e-06, 0, 0, 2.92057187378486e-06, -1.6061360225285e-06, -2.27549147349352e-07, -2.46982375181832e-06, -1.27691030444102e-06, 4.34482743974093e-07, 3.149091579283e-06, 4.63333921985374e-07, 0, 0, 0, 0, 0, 0, 5.09568968559116e-07, 0, -2.82523771691854e-06, 0, 0, 0, -5.92822386175856e-07, 0, -8.98750591669887e-06, 1.78300861234407e-07, 166075.312503371, -4.04982896625082e-06, 0, 0, 0, -5.02781419277177e-06, 0, -8.99268698017511e-06, 2.75992597366396e-06, 3.39803850215909e-07, 5.11047765941596e-06, 3.98524440179618e-06, -1.3275885968286e-06, -9.79117370310962e-06, 4.13210971335583e-06, 7.5006487168305e-07, -6.27990134980224e-06, 6.78094895640766e-06, 9.82855717844002e-07, -5.6938322197286e-07, 0, 0, 0, 0, 0, 0, -4.04157016990816e-07, 0, 1.72939474340579e-06, 0, 0, 0, -1.47796527283409e-07, 0, 6.50413273289086e-06, 9.50219463121911e-08, -179361.337501956, 3.09083998017196e-06, 0, 0, 0, 1.52246558252137e-07, 0, 1.0851274825792e-05, -8.98970614332046e-07, -6.28947279125944e-07, -5.81625763013692e-06, -3.17406912393879e-06, 1.78697855406325e-06, 8.76943693010286e-06, 0, 0, 8.24498671110846e-06, -3.22325788690346e-06, 2.34624434762526e-06, -4.32246478797629e-06, 1.9101172904684e-06, 3.30128029889432e-06, 1.28490013629884e-05, -5.9425704841189e-06, -9.98418438177688e-06, 6.89403632480541e-06, -1.12213192539885e-05, -7.55057964085556e-06, -1.29461549163496e-05, 0, 0, 0, 0, 0, 0, 1.0541203647639e-07, 0, -4.06526417388552e-07, 0, 0, 0, -2.18118353044155e-08, 0, -2.10170408499214e-06, -1.87795856748849e-08, 76869.1446436534, -8.3698637674228e-07, 0, 0, 0, -9.3100615282535e-08, 0, -4.39134804520506e-06, 3.58654456906446e-07, 9.33055645076138e-08, 2.49644659582111e-06, 1.9551304634424e-06, -9.15255766974602e-07, -5.2042897803931e-06, 0, 0, 0, -1.51138727588221e-06, 0, -4.17939723445917e-06, 1.07798354567684e-06, 1.3313662171383e-07, 3.37096889858282e-06, 2.49901674617048e-06, -3.44514167317401e-07, -5.70593863623957e-06, 1.47721828903084e-06, 4.86500568090673e-06, -3.74949680497058e-06, 3.5021089876551e-06, -4.10100186379533e-07, -1.29187801803897e-06, 2.71638220229e-06, 9.79295435010813e-07, 6.91065212909315e-06, 9.64681507692824e-06, -3.70414133995578e-06, -1.03711969620705e-06, 7.08455744816228e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.37639405358491e-08, 104.142857130735, 0, 0, 0, 0, 0, 0, -8.95110302233255e-08, 0, -1.00223339228764e-07, 0, 0, 3.57627049908612e-08, 2.30409829102464e-07, 1.54639392871643e-07, -2296.34999982844, 0, 0, 0, 0, 0, 0, 4.45315801273764e-07, 0, 3.49038547127082e-07, 0, 0, 0, 2.58408409882854e-07, 0, -6.17272504562861e-07, -7.33261323942086e-07, -7.72791636812063e-07, 19026.8999998075, -1.97636985206289e-06, -1.68055645110528e-06, -1.74166174911559e-06, 0, 0, 0, 0, 0, 0, -6.23874734873057e-07, 0, -1.03461239969627e-06, 0, 0, 0, -2.37344344000221e-06, 0, 3.37610921880415e-08, 1.64297779224233e-06, 1.40253188127654e-06, -77501.8125008419, 0, 0, 1.8817779013363e-07, 4.98267462811857e-06, 7.51903167858056e-06, 6.9291381049666e-06, 8.59433422460124e-06, 5.6882844884765e-06, 4.46841983887115e-06, 1.55893516280208e-05, 0, 0, 0, 0, 0, 0, 7.6755444182198e-07, 0, 1.40467245108033e-06, 0, 0, 0, 2.03125514118161e-06, 0, -3.01327516797776e-07, -1.71063985709508e-06, 3.75588674757632e-07, 166075.312501015, 0, 0, 0, -8.66525892887536e-07, 0, -1.47563519106118e-06, -7.00490941424538e-06, -9.79662563850773e-06, -1.06787135401429e-05, -1.17408091119567e-05, -4.50217573587794e-06, -3.07975599590074e-06, -1.30209047327077e-05, -1.18943444313506e-05, -3.04911205755145e-05, -1.99040674468828e-05, -4.98639067844771e-05, -2.60423572063989e-05, 0, 0, 0, 0, 0, 0, -4.37681044684179e-07, 0, -6.53198552102689e-07, 0, 0, 0, -2.41357433221375e-06, 0, -9.94393440265078e-07, 9.16091591829661e-07, 2.82189534749495e-07, -179361.337502474, 0, 0, 0, -1.71743926562631e-06, 0, 2.49218021400442e-06, 5.78095218861841e-06, 7.32292684081742e-06, 4.0032361313283e-06, 5.99374134124561e-06, 1.28245099531587e-06, -3.31887839237644e-06, 0, 0, 4.02232341062117e-06, 1.55047739595275e-05, 1.75814811874755e-05, 1.58508521621288e-05, 3.07350115118633e-05, 1.47591207333896e-05, 2.14554110179096e-05, 2.37635849732302e-05, 4.12536982926777e-05, 2.64628168631733e-05, 4.52267478335894e-05, 2.66362453766815e-05, 3.78120157250536e-05, 0, 0, 0, 0, 0, 0, 7.9357361874546e-08, 0, 2.75550317743621e-07, 0, 0, 0, 6.09719936604512e-07, 0, 4.56086390140712e-07, -1.6636016391429e-07, 1.12711430077183e-07, 76869.1446435952, 0, 0, 0, 1.0071016897191e-06, 0, -8.76698084946357e-07, -1.44118772591424e-06, -1.84641938966397e-06, -6.14301740823722e-07, -1.35037067690701e-06, 2.84420315484044e-07, 9.78572213875746e-08, 0, 0, 0, -4.92940311329765e-07, 0, -2.66928450633116e-06, -4.0774794364291e-06, -5.90912438880855e-06, -3.02788294400951e-06, -8.74650753123294e-06, -2.41558606704609e-06, -3.50407795530771e-06, -8.69004069300136e-06, -1.44673146560957e-05, -4.80444056375451e-06, -4.62903938406478e-06, -7.66092167874666e-06, -1.04479127885827e-05, -2.14335428601725e-05, -1.40320983840873e-05, -2.90721419548023e-05, -2.88155923143368e-05, -2.00752919592372e-05, -2.46191236293684e-05, -3.00194279720579e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 60.7500000019587, 0, 0, 0, 0, 0, 0, 6.77960047476233e-08, 0, -5.51895914556412e-08, 0, 0, -1.61488085654575e-07, -1248.41250000252, -546.750000056457, -1.5445554374839e-07, 0, 0, 0, 0, 0, 0, -1.48370955981553e-08, 0, 3.57991824031989e-07, 0, 0, 0, 2.92102128323159e-07, 0, 2.80858068526765e-07, 9226.40624971838, -8.58002215134313e-08, 5.0984478416514e-07, 11235.7124995221, -1.78140224100526e-07, 1.00789043182063e-06, 0, 0, 0, 0, 0, 0, 1.63322385194613e-07, 0, -4.81771429971129e-07, 0, 0, 0, 1.89774537393949e-06, 0, -2.35953786363672e-06, -31369.7812503913, -3.29054889899692e-07, -2.25599319194439e-06, 0, 0, -7.40715274334816e-07, -83037.6562486429, -1.57342518107896e-07, 7.99063715055776e-07, -5.55601294667639e-08, -1.65548439855971e-06, -1.61740179297051e-06, 8.28632139192715e-07, 0, 0, 0, 0, 0, 0, -7.13482915601561e-07, 0, 8.20046422101432e-07, 0, 0, 0, -3.07944118957583e-06, 0, 2.783587837524e-06, 49822.5937496564, 7.44881933221735e-07, 3.43613270339564e-06, 0, 0, 0, -5.17174695033737e-06, 0, 2.6974885178516e-06, 282328.031248269, 1.09686898318688e-06, 3.09456059584537e-07, -2.13635009199073e-06, -2.02871771259369e-07, 1.01701100683478e-05, -1.39860526824005e-06, 1.68168773571904e-06, 1.4642846737657e-05, -5.78367843003325e-06, 2.90226662375091e-06, 8.64712725935902e-06, 0, 0, 0, 0, 0, 0, 1.06088655929708e-07, 0, -2.48836836556851e-07, 0, 0, 0, 1.43344143756223e-06, 0, -2.54747782511362e-06, -29893.5562501281, -9.08313699323224e-07, -1.31157571565796e-06, 0, 0, 0, 6.39485493846356e-06, 0, -2.32272531279237e-06, -448403.343750302, 6.83767073870021e-08, -1.30066655873857e-07, -1.22245396610419e-06, -2.79550367113277e-06, -4.90747064033314e-06, 0, 0, 1.82012446990887e-06, 4.26390962915401e-06, 2.03230993260463e-06, 2.42995985580801e-06, 2.040710126377e-06, -9.22954704278044e-06, 3.052357486577e-06, 5.3242293343019e-07, 1.22090586155344e-05, -8.54352893687121e-06, 1.77224120012004e-05, -1.29835185227362e-05, -1.22605518524004e-05, 0, 0, 0, 0, 0, 0, -8.13845961815673e-08, 0, 8.94328400992037e-08, 0, 0, 0, -2.59380538457941e-07, 0, 8.14766495017421e-07, 6.99466631060307e-08, 3.1816159200062e-07, 1.75637049676755e-07, 0, 0, 0, -1.23126267142728e-06, 0, 9.19614760153091e-07, 269042.00624987, 1.84768113167761e-07, 3.18284670007687e-07, 1.27463603413631e-06, 1.91494573621138e-06, 4.31622086352431e-06, 0, 0, 0, -1.90529067104838e-07, 0, 6.16203917433931e-07, -1.14810705462284e-06, 1.67703565182353e-06, -1.92314136185288e-06, 1.58840887737058e-07, 5.00240739897275e-07, 3.0553198137241e-06, 9.53141010407158e-07, -3.12423058083203e-06, 6.01252742049243e-06, -2.15420735279553e-06, 4.6588552735255e-07, 3.29543642505208e-06, 4.92785817264349e-07, -3.36122765834844e-06, -5.66919762230778e-06, 1.47920335609278e-06, 1.12966304709442e-05, -3.81081988719756e-06, 1.39057732205908e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.54561056656966e-08, 0, 1.72078258536445e-08, 48.6000000017526, 0, 0, 0, 0, 0, 0, 1.49282537229063e-07, 0, -2.0637014214498e-07, 0, 0, 1.78563040455467e-07, -911.250000013446, -656.099999877772, -1.81438252128718e-08, 0, 0, 0, 0, 0, 0, -7.60652971123601e-07, 0, 6.98020780786016e-07, 0, 0, 0, -1.89211774873374e-06, 0, 6.15313313224303e-07, 5740.87500017968, 1968.29999941662, 1.03390193510919e-06, 12301.8750002887, 4.22829509332814e-07, -6.10329648756356e-07, 0, 0, 0, 0, 0, 0, 1.43201175033815e-06, 0, -1.82476341183847e-06, 0, 0, 0, 3.71078190339694e-06, 0, -1.23682418970195e-06, -14762.2499994577, 1.63545379107617e-06, -2.80694312398521e-06, 0, 0, 2.56256703691314e-06, -77501.8124988617, -8.53190727854458e-07, -36905.6250000714, 4.44217876029852e-06, -1.52492235738382e-06, -1.82407622865546e-06, -7.7756252572749e-07, 0, 0, 0, 0, 0, 0, -9.8104601923993e-07, 0, 2.02603907683969e-06, 0, 0, 0, -4.67488068655233e-06, 0, 3.40973854883751e-06, 13286.0250007403, -2.08908536118368e-06, 4.64128888350092e-06, 0, 0, 0, -3.91760992131684e-06, 0, -1.27690793715161e-06, 199290.374999363, -9.16578149355038e-07, -5.69950608798433e-07, -1.09402725193549e-05, 8.07286270277645e-06, 2.23992937482096e-06, 232505.437499377, 2.32537387331901e-06, -1.7968423287398e-05, -1.66316781216296e-06, -1.63072018794185e-06, -1.47854337699971e-06, 0, 0, 0, 0, 0, 0, 8.35168115901286e-07, 0, -1.49465907824921e-06, 0, 0, 0, 4.90198563448207e-06, 0, -2.51558502409676e-06, -3.26278814142628e-07, 1.69370840261856e-06, -5.18115226777464e-06, 0, 0, 0, 4.77214216950053e-06, 0, -1.44468929517844e-06, -179361.337497335, 6.29879330925716e-08, -9.62348663694234e-07, 1.41962786141606e-05, -5.81818554316499e-06, -6.19576457381033e-06, 0, 0, 2.774850965409e-06, -597871.125001215, -3.99892689760789e-06, 1.79329139351576e-07, 1.26528441934294e-05, 3.79633805830571e-06, -2.72494238170421e-07, 6.81950447514116e-06, -6.1939548000958e-06, -1.01590494874116e-05, 1.49597504130393e-06, 1.08462857759509e-05, 1.99406772754934e-05, 0, 0, 0, 0, 0, 0, -1.27924054395617e-07, 0, 3.01190627839286e-07, 0, 0, 0, -1.49827732817139e-06, 0, 8.59874249801205e-07, 8.55549894574266e-08, -3.80395580348991e-07, 1.85373645489325e-06, 0, 0, 0, -3.25416874417929e-06, 0, 1.83602200968868e-06, -3.40561346342597e-07, -5.23621783202487e-08, 3.457457961332e-07, -4.86171239195859e-06, 1.51235907704328e-06, 2.13625693285227e-06, 0, 0, 0, -3.70814928952895e-06, 0, 2.07717547491826e-07, 538084.012498712, -1.61646296709769e-06, 2.99329762825226e-06, -1.0039733228618e-05, 2.35064008071211e-06, 1.2263162191944e-06, -4.59213846097144e-06, -2.96998158523709e-07, 4.96675038864281e-06, 7.97165130821175e-08, 1.84528876690588e-06, -5.55334774971813e-06, -6.79380317705917e-06, -4.11398618294876e-07, -2.19911621361293e-07, -8.62519158599118e-06, -1.58632199296107e-06, 6.98074841224144e-06, 1.33908545757895e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.36042379018503e-08, 0, 5.02917143260416e-08, 45.5625000249934, 0, 0, 0, 0, 0, 0, -5.55931356187566e-08, 0, -3.4286055195428e-07, 0, 0, -1.22439063123014e-07, -751.781250043252, -751.781250070615, -3.24965811835222e-07, 0, 0, 0, 0, 0, 0, -1.89884876856528e-07, 0, 1.33642321360116e-06, 0, 0, 0, 1.00480101364893e-06, 0, 1.33801117282042e-06, 3690.56250049794, 3690.56250044182, 2.03036855531281e-06, 12404.390625236, 1.40529263423236e-06, -4.35284552424735e-08, 0, 0, 0, 0, 0, 0, 3.08167317729825e-07, 0, -2.67550880540587e-06, 0, 0, 0, -2.43364060400432e-06, 0, -5.20185230289255e-06, -5535.84375081673, -5535.84374974496, -6.59154183924274e-06, 0, 0, -3.35976721791262e-06, -60894.2812539937, 1.2833869326364e-06, -60894.2812501671, -3.2912596721874e-06, -5.59208158077172e-06, -8.87007828845096e-06, -5.15044629423488e-06, 0, 0, 0, 0, 0, 0, -2.7850244040301e-07, 0, 2.58932950434129e-06, 0, 0, 0, -6.00241499978195e-08, 0, 1.01989119509453e-05, 7.70101174827873e-07, -1.45080497627893e-06, 1.09840932778011e-05, 0, 0, 0, 5.57158305523324e-06, 0, 8.41966371590172e-06, 91341.4218761618, -1.47047830306298e-06, 91341.4218750072, 6.73127534108394e-06, 1.31249093403569e-05, 1.62883185262334e-05, 298935.562498342, -4.18591677452367e-07, 2.34770218497363e-06, 1.57788238300031e-05, -1.2646392622125e-05, 1.73756893679111e-05, 0, 0, 0, 0, 0, 0, 2.86232482732948e-07, 0, -1.57826563403282e-06, 0, 0, 0, -3.78594176063134e-07, 0, -8.1048439754393e-06, -5.71143977607048e-07, 1.42992903059135e-06, -8.73527575606738e-06, 0, 0, 0, -4.78113796022544e-06, 0, -1.13706995744283e-05, -2.24718520571366e-06, -6.55442111747381e-07, 4.94206711327195e-07, -3.86027599563851e-06, -1.25494174150167e-05, -3.16185187868623e-05, 0, 0, -6.76515693194974e-06, -448403.343747206, 1.80782486887047e-06, -448403.343755544, -9.54315499289415e-06, -1.68889050226568e-06, -3.49037994040693e-05, -6.28708403217493e-06, -8.66362620191977e-08, -2.42671790135657e-05, -3.34232726858279e-05, -2.08555868270134e-05, -3.58099113982546e-05, 0, 0, 0, 0, 0, 0, -5.4928265553589e-08, 0, 4.69581739801973e-07, 0, 0, 0, -2.68786091218736e-07, 0, 2.30012993250099e-06, 2.10891490527279e-07, -6.86785788552525e-07, 2.70687975541533e-06, 0, 0, 0, 1.05890681124233e-06, 0, 2.61848658815008e-06, 7.89778590789789e-07, 1.1876970400316e-08, 2.42121528211602e-07, -1.90572592784611e-06, 4.23338476112648e-06, 7.23186287376219e-06, 0, 0, 0, 1.57693960929247e-06, 0, 5.3741782515953e-07, 1.33938884664189e-06, -1.02743463151522e-06, -1.56590068106455e-07, 5.2807800694467e-06, 2.48321777299639e-06, 8.56948534153509e-06, 2.68566379235569e-06, 1.10318005158356e-06, 1.0032616953328e-05, 672605.015627125, -4.29218454816706e-06, -2.32678592735044e-06, 2.65200750389779e-06, 1.4875114034628e-05, 6.51412824956934e-06, 1.22823545511888e-05, 1.98693803394684e-06, 2.41092970195365e-05, 1.80037728296041e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6.21045395094586e-08, 0, 1.09147489709257e-07, 48.6000000654308, 0, 0, 0, 0, 0, 0, -2.83995372855007e-07, 0, -7.79019080746034e-07, 0, 0, -1.21510550931402e-06, -656.100000374531, -911.250000766348, -1.00690649723923e-06, 0, 0, 0, 0, 0, 0, 1.34838033169166e-06, 0, 2.44397615914748e-06, 0, 0, 0, 3.47285065330305e-06, 0, 5.22751042544072e-06, 1968.30000048228, 5740.87500213265, 5.1475726652254e-06, 12301.8750028944, 2.30849623876611e-06, 1.01429166156112e-05, 0, 0, 0, 0, 0, 0, -1.99512981381725e-06, 0, -4.88209399180693e-06, 0, 0, 0, -6.69229974012715e-06, 0, -1.41734205613929e-05, -1.08949798345956e-06, -14762.2500064458, -1.32502475527225e-05, 0, 0, -1.5538095896587e-05, -36905.625004565, -6.37775809434781e-06, -77501.8125107999, -2.15330136550958e-05, -7.51317035122936e-06, -2.4677645566173e-05, -1.67513472229048e-05, 0, 0, 0, 0, 0, 0, 1.76543734737756e-06, 0, 6.02426595307254e-06, 0, 0, 0, 9.0821347905609e-06, 0, 1.52247226578915e-05, 8.10366594833225e-07, 13286.0250103369, 1.69235810397099e-05, 0, 0, 0, 1.47460912406711e-05, 0, 2.67218107613361e-05, 1.18558755070514e-05, 8.04305408550499e-06, 199290.375016369, 3.15396135013448e-05, 1.47747519768978e-05, 6.21752293858638e-05, 232505.437524056, 1.92081160699116e-05, 6.54158174337011e-05, 3.36087983825093e-05, 7.42764613694543e-05, 5.85409208045321e-05, 0, 0, 0, 0, 0, 0, -1.37751276113537e-06, 0, -3.01142667366814e-06, 0, 0, 0, -6.3146372100197e-06, 0, -1.32483889309223e-05, -4.53345175342986e-07, -5.25043165062613e-06, -1.34639171724353e-05, 0, 0, 0, -1.27859021135414e-05, 0, -1.69658001322263e-05, -5.04264883328998e-06, -1.73792435904347e-06, -179361.337515839, -2.62123061313157e-05, -1.61675044692083e-05, -3.92780555214445e-05, 0, 0, -1.60728113171266e-05, -1.73637962334181e-05, 8.09191314703312e-07, -597871.125007208, -3.49097725404251e-05, -1.29898539264059e-05, -3.09272283085476e-05, -3.41520334434587e-05, -7.19707048830806e-05, -5.22300588282854e-05, -3.63374219844128e-05, -6.17632922964e-05, -0.000102298669307657, 0, 0, 0, 0, 0, 0, 2.00427979342084e-07, 0, 6.7284743039422e-07, 0, 0, 0, 2.28066097542503e-06, 0, 4.04196684908414e-06, 1.27941328615435e-07, 2.04633013602198e-06, 3.63100543487332e-06, 0, 0, 0, 4.26326585180944e-06, 0, 8.90755252364597e-06, 1.50962223703021e-06, 9.67094475202741e-07, 4.28754157050542e-06, 1.23993314948552e-05, 7.45631891330829e-06, 1.96108380789914e-05, 0, 0, 0, 8.13518771236663e-06, 0, 1.10230516029972e-05, 4.599632151152e-06, 2.36581304028298e-06, 538084.012509574, 1.58543427672034e-05, 5.98297294430516e-06, 3.35793487411567e-05, 8.07934646167838e-06, 2.6028789686039e-05, 3.32958040561316e-05, 8.75564480768598e-06, 6.8134904897422e-06, 4.5708296830054e-05, 2.47125923232335e-05, 3.37380334197787e-06, 2.98254750240716e-05, 5.04000485092157e-05, 2.89197715761471e-05, 3.01168723380895e-05, 7.06153563531147e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.98066521331143e-08, 0, 4.5563772815463e-08, 60.7500000230091, 0, 0, 0, 0, 0, 0, -1.75792336201672e-07, 0, -2.4652766002228e-07, 0, 0, -2.23273840861404e-07, -546.750000020282, -1248.41250014117, -3.15677301115731e-07, 0, 0, 0, 0, 0, 0, 4.96903367002359e-07, 0, 9.62752950594772e-07, 0, 0, 0, 1.57968958359654e-06, 0, 1.43259153624655e-06, 2.25819467547625e-07, 9226.40625114125, 2.08507004133752e-06, 11235.7125005983, 1.06439138195004e-06, 8.85942382803651e-07, 0, 0, 0, 0, 0, 0, -1.08059271379253e-06, 0, -1.54989948849293e-06, 0, 0, 0, -3.37811388695148e-06, 0, -4.29312434922821e-06, -5.74395985778515e-07, -31369.7812514537, -7.38573773692598e-06, 0, 0, -3.52521453195531e-06, -1.1187557816373e-06, -5.50624277247173e-07, -83037.6562502821, -6.76820121659288e-06, -6.49212750081991e-06, -1.29456729286795e-05, -1.12928360934322e-05, 0, 0, 0, 0, 0, 0, 1.46163108947228e-06, 0, 1.34152704194121e-06, 0, 0, 0, 5.10429822983398e-06, 0, 8.04962146393327e-06, 4.85665479004034e-07, 49822.5937512036, 1.30634049013126e-05, 0, 0, 0, 1.14470868006858e-05, 0, 8.0782507583594e-06, -1.65792305931125e-06, 3.06859507513494e-07, 282328.031251017, 1.26863356341311e-05, 1.30698761180384e-05, 1.75745929483187e-05, -2.32943005522232e-06, -3.22819798106909e-06, 2.22570258135009e-05, 6.06327492849624e-06, 1.29364539978979e-05, 2.27690447222496e-05, 0, 0, 0, 0, 0, 0, -8.1781680727891e-07, 0, -8.11427293333657e-07, 0, 0, 0, -3.76324476440565e-06, 0, -3.6763829662313e-06, -3.66823426923027e-07, -29893.5562530699, -8.94459232571518e-06, 0, 0, 0, -5.81211362785371e-06, 0, -1.14915958974023e-05, -2.09700328313801e-06, -1.50654684623906e-06, -448403.343748423, -1.30208727997397e-05, -1.52636613247601e-05, -2.55298136102906e-05, 0, 0, -1.51622528267444e-05, 1.2621348207459e-06, -5.1282982619852e-06, -6.05512318999825e-06, -4.01416769338702e-05, -7.05874940042088e-06, -4.45266176358244e-05, -6.12047937056634e-08, -8.19543420652448e-06, -3.8022677714297e-05, -2.34103718151047e-05, -1.01397573421793e-05, -2.77998552475106e-05, 0, 0, 0, 0, 0, 0, 3.14672103628662e-07, 0, 2.1470506438466e-07, 0, 0, 0, 9.18563622287814e-07, 0, 1.07533113120694e-06, 0, 5.14198890153679e-07, 2.68339849039397e-06, 0, 0, 0, 2.64226856042886e-06, 0, 1.9604061712897e-06, 2.9263788049399e-07, 2.99563539201418e-07, 269042.006249628, 1.80707251960366e-06, 6.19135614914236e-06, 9.0068562094997e-06, 0, 0, 0, 4.58914813422261e-06, 0, 3.62807195218785e-06, 3.49101607271932e-07, -5.4414847901847e-07, -3.96997733217037e-06, 1.32489979440581e-05, 2.19026299225181e-06, 7.6794882494275e-06, 2.9072325522094e-06, 4.05437585470919e-06, 1.2845115837856e-05, -3.44565888272323e-06, -3.55621524470251e-06, 3.8875656770591e-06, 9.97007571518871e-06, 1.21721040235639e-05, 8.29837386750112e-06, 1.71642641159013e-05, 1.28127748378864e-05, 2.64540723459627e-05, 2.67860550806957e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.1480595140808e-08, 0, 3.37363948510679e-08, 60.750000052987, 0, 0, 0, 0, 0, 0, -1.57673968883336e-07, 0, -2.55229630789081e-07, 0, 0, -2.84598018852835e-07, -2.71741895437316e-07, -1248.41250064374, -546.750000672511, 0, 0, 0, 0, 0, 0, 9.85924703799287e-07, 0, 8.44494698660051e-07, 0, 0, 0, 7.14603985207666e-07, 0, 7.84820840073858e-07, 2.25702172586391e-07, 9226.40625329251, 3.52066566439942e-06, 3.56154509209588e-06, 1.67727667861148e-06, 11235.7125071893, 0, 0, 0, 0, 0, 0, -1.44618579724432e-06, 0, -2.09178343500853e-06, 0, 0, 0, -1.6783908582354e-06, 0, -2.10262405109036e-06, -3.79355646270503e-08, -31369.7812583792, -9.29616120940384e-06, 0, 0, -1.34145463918119e-07, -5.20594596235394e-06, -4.0227525729377e-06, -1.29950681701937e-05, -83037.6562723245, -7.80446117044262e-06, -1.77240751760469e-05, -2.48214636335763e-05, 0, 0, 0, 0, 0, 0, 1.88565147590273e-06, 0, 2.98825901000096e-06, 0, 0, 0, 7.44592349573442e-06, 0, -3.34805124115084e-07, -7.85184542011724e-07, 49822.5937612236, 1.32730054522055e-05, 0, 0, 0, 8.62215205694924e-06, 0, -3.19238731227942e-07, 6.39940434412205e-06, 1.80524739074611e-06, 2.24823299321562e-05, 282328.031292778, 1.6469481558104e-05, 3.91467452672208e-05, 2.45677969878506e-05, 1.15706231172824e-05, 6.70332908823181e-05, 2.39954643878509e-05, 9.2951942175679e-05, 5.64578881832239e-05, 0, 0, 0, 0, 0, 0, -1.14222077323581e-06, 0, -1.94938042224405e-06, 0, 0, 0, -4.66893149053673e-06, 0, -6.08854007582482e-07, 3.68049635854224e-07, -29893.5562586501, -1.07578422539831e-05, 0, 0, 0, -3.87979586673961e-06, 0, 3.64434155594497e-06, -3.82824969912044e-06, 3.54380725511162e-06, -1.87389790089292e-05, -448403.343791077, -1.67138777682163e-05, -1.65445887654654e-05, 0, 0, -4.51781327769565e-06, -1.96632877152779e-05, -5.14623491863179e-06, -2.20716162152871e-05, -6.48757668035933e-05, -1.8790771985795e-05, -4.46870608158892e-05, -1.76787279572402e-05, -9.07882976243922e-05, -6.19308065807905e-05, -5.52650122839959e-05, -4.00032425163794e-05, -8.79623825871506e-05, 0, 0, 0, 0, 0, 0, 2.15538429618599e-07, 0, 4.02666357394732e-07, 0, 0, 0, 2.06749537536338e-06, 0, 8.62859148038604e-07, -1.85074663035049e-07, 2.31713453232977e-06, 2.80773468709413e-06, 0, 0, 0, 2.24881330765292e-06, 0, 6.64526351315744e-07, 6.95030675656739e-07, -1.01291355255208e-06, 4.93665103499736e-06, 269042.006264008, 7.34656627677326e-06, 1.2333555319941e-05, 0, 0, 0, 2.180088046248e-06, 0, 1.89695703034734e-06, 5.21279048768693e-06, -1.11450385044622e-06, 1.01091373924032e-05, 2.47553001951247e-05, 6.82677975917287e-06, 1.81725071477144e-05, 2.28706871382651e-06, 3.58961264479488e-05, 2.94741588278475e-05, 7.62178302227139e-06, 5.12678462889571e-06, 3.8269232466849e-05, 2.58286609873364e-05, 9.4532127763757e-06, 3.77135138973639e-05, 5.35602785393328e-05, 3.7323508971414e-05, 4.85261726140101e-05, 6.16683174465305e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.35166662478154e-08, 0, 3.59199035244161e-08, 48.6000000372641, 0, 0, 0, 0, 0, 0, -2.19620610375409e-07, 0, -1.87873909565102e-07, 0, 0, -1.75598403054387e-08, -1.42418865828864e-07, -911.250000522321, -656.100000362252, 0, 0, 0, 0, 0, 0, 4.44434905299341e-07, 0, 1.19280661463734e-06, 0, 0, 0, 1.50840799728754e-06, 0, 4.91341602794533e-08, 8.72717719247152e-07, 5740.87500250923, 1968.30000176921, 1.68903114866839e-06, 2.79034853722079e-06, 12301.8750010402, 0, 0, 0, 0, 0, 0, -1.73665645756988e-06, 0, -2.0209839446921e-06, 0, 0, 0, -4.45886441210175e-06, 0, -1.85802878997967e-06, -1.64273735579633e-06, -14762.2500053993, -6.59545687541125e-06, 0, 0, -1.92094075517902e-06, -6.17829683729144e-06, -2.8335062688748e-06, -7.56166589358431e-06, -77501.8125118962, -1.10336284619903e-05, -36905.6250179072, -1.10339705873666e-05, 0, 0, 0, 0, 0, 0, 1.4468700178326e-06, 0, 1.8695543865236e-06, 0, 0, 0, 8.90684947126658e-07, 0, 1.09575956557351e-05, 1.4765771529545e-06, 13286.0250073271, 1.16937602154858e-05, 0, 0, 0, -6.57702673034201e-06, 0, 1.8042483521182e-06, 4.68574574855654e-06, 4.08070970964492e-06, 1.70393883376808e-05, 199290.375019031, 1.67734856826902e-05, 1.38296129691113e-05, 8.87296555835982e-06, 5.80564683507734e-06, 232505.437508946, 2.93193649967731e-05, -4.58812944110101e-06, 3.55449162587018e-05, 0, 0, 0, 0, 0, 0, -8.58980719720372e-07, 0, -8.98755362461128e-07, 0, 0, 0, -3.77231783608641e-06, 0, -6.81537597995028e-06, -2.15788476040949e-07, -5.14812871065642e-06, -6.19613174612712e-06, 0, 0, 0, 1.04386737869825e-07, 0, -1.01631598766988e-05, -4.28950097998309e-06, -6.51893040117649e-06, -1.40599464227155e-05, -179361.337516212, -1.69655225542725e-05, -4.40737636647041e-05, 0, 0, 3.56425154559979e-06, -5.0006583035819e-06, -1.03013542424291e-05, -2.69695575956367e-05, -597871.125036609, -1.32379428729062e-05, -4.37940350920094e-05, -1.29071145014296e-05, -1.38673243013013e-05, -4.68334056843445e-05, -5.33812156661365e-05, -5.83206439724249e-05, -8.18105385988451e-05, 0, 0, 0, 0, 0, 0, 3.72158718771154e-07, 0, 4.5386048048242e-07, 0, 0, 0, 6.48905065133422e-08, 0, 1.49830258635839e-06, 0, 1.1780065958857e-06, 2.47562772543256e-06, 0, 0, 0, 1.34364428214058e-06, 0, -2.00569632979051e-07, 7.11214318769373e-07, 9.63449032361176e-07, 6.32834839334463e-06, 4.49334684183144e-06, 4.94947893190747e-06, 7.20058290866244e-06, 0, 0, 0, -8.36373809156219e-07, 0, -3.75372203978439e-06, 2.62389254477921e-06, 4.49445933892921e-06, 4.82737811265281e-06, 538084.01251896, 5.91490477006751e-06, 1.24804950007853e-05, 7.94846286689258e-06, 1.18099782699894e-05, 1.14579754431462e-05, 7.11521393088584e-06, 1.81329457023422e-06, -6.40809566011133e-06, 1.92176016450125e-05, 2.64325330773724e-05, 2.82597230599554e-05, 4.29801117557939e-05, 1.57367715992505e-05, 3.02359369581373e-05, 1.09825535435035e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.97168444703782e-08, 0, -4.23262043014862e-08, 45.5625000094028, 0, 0, 0, 0, 0, 0, 0, 0, 1.81996547840875e-07, 0, 0, 5.64696031937853e-07, -3.05834986732481e-07, -751.781249909662, -751.781249997892, 0, 0, 0, 0, 0, 0, 6.09725171675807e-08, 0, -1.03499241846233e-06, 0, 0, 0, -3.68392600428253e-06, 0, -1.75351128861807e-06, 1.17651086676233e-06, 3690.56250053624, 3690.5624998186, 2.89960481019509e-06, 1.11906972383973e-06, 12404.3906260766, 0, 0, 0, 0, 0, 0, -7.49844759454792e-07, 0, 8.75601524177545e-07, 0, 0, 0, 4.59275833426902e-06, 0, 9.04959276559673e-06, -1.01548011777391e-06, -5535.8437529765, -5535.8437455262, 0, 0, 1.26283159378358e-05, -9.15661020948368e-06, -7.39536745288459e-06, -8.44106533769749e-06, -60894.2812553855, 2.02825021575496e-06, -60894.2812390504, -1.73158677152889e-05, 0, 0, 0, 0, 0, 0, 2.79819298270281e-06, 0, -8.47642522000022e-07, 0, 0, 0, 3.69143478833343e-06, 0, -1.64214544263723e-05, 2.83682396335668e-06, 3.01977171832544e-06, -6.6170206415176e-06, 0, 0, 0, -3.08273981027425e-06, 0, -2.46140678348338e-05, 1.53222457919226e-05, 8.37904888329296e-06, 7.56482252719468e-06, 91341.4218904511, 7.2448098441024e-07, 91341.4218593132, 2.07855650940179e-05, 1.20350790568286e-05, 298935.562501468, 3.52976603514713e-05, 6.69853126853275e-05, 6.25920704254049e-06, 0, 0, 0, 0, 0, 0, -1.52264331018276e-06, 0, -7.60590843457991e-07, 0, 0, 0, 5.15088620431365e-07, 0, 8.43462760840019e-06, -1.8016720082858e-06, -2.00223274783434e-06, 2.87311743392761e-07, 0, 0, 0, -1.02134881737914e-06, 0, 2.77881618383706e-05, -6.92782419293285e-06, -4.81833503157143e-06, -4.850460468903e-06, -7.05429637104563e-06, 6.98473051985067e-06, 3.68703161906183e-05, 0, 0, 2.00401648783919e-05, -2.95016070742307e-05, -1.27241677018174e-05, -8.61387298605314e-06, -448403.343749831, 7.56273625691849e-06, -448403.343734903, -4.71112307680241e-05, -7.72261220221376e-05, 4.23086742709357e-06, -7.59143460009862e-05, -2.43790783653731e-06, 1.23740953545891e-05, 0, 0, 0, 0, 0, 0, 3.40697410953541e-07, 0, -1.68939521121618e-07, 0, 0, 0, 1.16025942119872e-06, 0, -1.8888246876234e-06, 5.1173035994447e-07, 1.07948141698673e-06, 6.41562378661581e-08, 0, 0, 0, -1.8130749291315e-06, 0, -4.42609917061089e-06, 2.96224558851373e-06, 1.68841337234084e-06, 3.37298010973141e-07, 2.86408780412136e-06, -3.07015618844753e-06, -9.14625469080876e-06, 0, 0, 0, -1.0088082823955e-05, 0, -8.48731031304286e-06, 8.61333898519496e-06, 5.49547631657698e-07, 9.96502040315016e-06, -2.96807076614998e-06, 2.83594479167879e-07, -1.27417908715782e-05, 1.00742342202259e-05, 2.13705988512134e-05, -2.72746813077442e-06, 1.14125177405507e-05, 6.055940016505e-06, 672605.015633215, 1.90271031871759e-05, 1.62596365135133e-05, 2.97567229033153e-05, 7.05695443979884e-06, 3.43811348872225e-06, 2.69088125264365e-05, 5.4510861902559e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -8.19499331567205e-08, 0, -7.95476322192482e-08, 48.6000000240683, 0, 0, 0, 0, 0, 0, 4.71925309919419e-07, 0, 5.04012467071301e-07, 0, 0, 8.13744970184221e-07, -1.02333419748546e-07, -656.100000627628, -911.250000176718, 0, 0, 0, 0, 0, 0, -8.36578825378926e-07, 0, -1.38640258390637e-06, 0, 0, 0, -3.14320327576144e-06, 0, -5.20335852281412e-06, 3.55844420542303e-07, 1968.30000144271, 5740.8750002534, 2.10000248413444e-06, 1.35036602925456e-06, 12301.8750046598, 0, 0, 0, 0, 0, 0, 2.02819541291306e-06, 0, 3.11363037627156e-06, 0, 0, 0, 8.9343020561462e-06, 0, 5.35339277445171e-06, -1.56443352395613e-06, -1.37399862074994e-06, -14762.2500015901, 0, 0, 1.29921263459747e-05, -6.30004789281444e-06, 1.10698959093113e-06, -7.49060262142539e-06, -36905.6250101598, -4.56819259170909e-06, -77501.8125198244, -1.3151142815223e-05, 0, 0, 0, 0, 0, 0, -2.98885082465983e-06, 0, -3.21492791426517e-06, 0, 0, 0, -1.8771599924274e-05, 0, -5.1166279498169e-06, -3.03623701467796e-07, 3.91656224680089e-06, 13286.024997279, 0, 0, 0, -3.92285533769189e-05, 0, -2.00577429334705e-05, 9.22806183509958e-06, 1.35206092886634e-06, 1.34364861816767e-05, 1.28671916553751e-05, 9.89408819927116e-07, 199290.375032404, 2.01125000115148e-05, 6.38452881910639e-06, 232505.437528758, 4.35074822925071e-05, 2.56109502627623e-05, 6.94109973829477e-05, 0, 0, 0, 0, 0, 0, 1.13937051368462e-06, 0, 3.53229770447194e-06, 0, 0, 0, 7.89044660629965e-06, 0, 3.08054879805585e-06, 2.80955328895926e-07, -1.18602772295857e-06, 4.99201447214036e-06, 0, 0, 0, 3.34807222128784e-05, 0, 1.21708283246304e-05, -8.22391587065564e-06, 1.40309135122316e-06, -1.3759465955983e-05, -1.68567286276699e-05, -2.31491955782093e-06, -179361.337532852, 0, 0, 3.55713339279436e-05, -7.62133376294968e-06, 7.44758887740587e-06, -1.9427280739544e-05, -2.69051562001747e-05, -1.01861145289298e-05, -597871.125028056, -2.1364716410361e-05, -3.0479289014707e-05, -5.55310535898627e-05, -4.68399004704066e-05, -0.000108281883327016, -0.000173583964281493, 0, 0, 0, 0, 0, 0, -2.60689971532719e-07, 0, -5.27045246740874e-07, 0, 0, 0, -2.57197567256892e-06, 0, -6.37658184884288e-07, -1.25848107100415e-07, 1.27314645095369e-07, -2.34027873660155e-06, 0, 0, 0, -7.70049353146696e-06, 0, -5.60385894715084e-06, 1.57584438355172e-06, -4.67613940646495e-07, 4.28872426260934e-06, 4.20026182067393e-06, 6.52238444787648e-07, 7.10240869763421e-06, 0, 0, 0, -1.09417817790263e-05, 0, -1.37619394759691e-05, 5.75557492086337e-06, 1.1873961675389e-06, 5.53519012775453e-06, 1.69074383827087e-05, 1.27897432752888e-06, 538084.01251923, 9.2339351956355e-06, 1.60850756093823e-05, 2.34811142577671e-05, 9.07139254167499e-06, -1.87341721308708e-06, 1.92660412076709e-05, 3.41857328931862e-05, 2.99670945589507e-05, 2.80107936049196e-05, 7.3342303630361e-05, 3.81595734125234e-05, 3.8163369484843e-05, 6.01024116007301e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.62650019580624e-08, 0, 5.86303367672164e-08, 60.7500000205817, 0, 0, 0, 0, 0, 0, -1.75467985572521e-07, 0, -3.73055080910304e-07, 0, 0, -5.53558497180841e-07, -5.94676975399366e-08, -546.74999999078, -1248.41250072425, 0, 0, 0, 0, 0, 0, 1.52982205929045e-07, 0, 1.32049101037552e-06, 0, 0, 0, 1.23351067747831e-06, 0, 3.77965996113477e-06, -3.53295331745105e-07, 1.36379683736812e-07, 9226.40625430647, -4.54936478149355e-08, 1.91390908791222e-06, 11235.7125039672, 0, 0, 0, 0, 0, 0, -1.00589674845293e-06, 0, -2.35195767787917e-06, 0, 0, 0, -1.12030793773324e-06, 0, -6.04776453319982e-06, 1.14203448221781e-06, 0, -31369.7812612966, 0, 0, -3.71513563159797e-06, 2.52536646944921e-06, -5.36251608553813e-06, 2.45875047059353e-06, -1.32035733619534e-05, -1.28823629460028e-05, -83037.6562717907, -1.92414505762824e-05, 0, 0, 0, 0, 0, 0, 1.29106071350787e-06, 0, 2.23385194481669e-06, 0, 0, 0, 7.13359978399642e-06, 0, 1.11398527255842e-05, -8.68079508578861e-07, -3.25644628320039e-06, 49822.5937694604, 0, 0, 0, 1.01282348037983e-05, 0, 7.79568550240229e-06, -5.59170270737948e-06, 4.80240488258554e-08, -8.08947313866013e-06, 1.52665653557472e-05, 2.63005325640881e-05, 282328.031292832, -1.13697622347327e-05, 7.89317661835789e-06, 6.33003119013425e-05, 1.75580604743732e-06, 4.67543075433485e-05, 4.78645151101383e-05, 0, 0, 0, 0, 0, 0, -5.38003382184664e-07, 0, -2.14895300559449e-06, 0, 0, 0, -3.19184895112647e-06, 0, -7.08459337098359e-06, 5.54678811232069e-07, 4.19664022246386e-08, -29893.5562624797, 0, 0, 0, -6.5805061808332e-06, 0, -1.04474083463693e-05, 3.25928320773074e-06, 7.77504170755778e-07, 1.461056961212e-05, -5.5171922714124e-06, -2.5121566923126e-05, -448403.343787011, 0, 0, -8.85008855052819e-06, 1.74759661137556e-06, -1.16189163751049e-05, 1.12577452074655e-05, -5.59587467613561e-05, -2.13416140731214e-05, -7.39328629200255e-05, -7.10848806992559e-06, -1.68496285961098e-05, -8.55579474251797e-05, -2.33508102609682e-05, -1.54849465483558e-05, -2.17685884311477e-05, 0, 0, 0, 0, 0, 0, 1.61172431112685e-07, 0, 2.44143312897296e-07, 0, 0, 0, 1.2703229418644e-06, 0, 1.90290276702816e-06, -1.60584375309802e-07, -1.16985019834821e-07, 3.83149148769891e-06, 0, 0, 0, 1.69933581585615e-06, 0, 5.05729436985136e-06, -1.13145159179131e-06, -1.05744114371351e-06, -5.1210102978026e-06, 2.08379713669883e-06, 9.04312541631481e-06, 269042.006266885, 0, 0, 0, -2.38102897540594e-07, 0, 4.57186152785192e-06, -3.25720292622269e-06, -7.27356321753467e-07, -1.02836705842293e-05, 1.29175632550815e-05, 8.14092439933833e-06, 2.51285646469557e-05, 7.93538896134074e-07, 6.10091631426633e-07, 3.33258313522953e-05, -9.59563651510585e-06, 4.89028200606428e-06, 2.09905240114517e-05, 5.55802131457897e-06, 7.27808344809631e-06, 8.75310124107855e-06, 1.8747073545943e-05, 3.87043589870205e-05, 5.78212511199821e-05, 2.6018757381475e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.33021014072408e-08, 0, -4.55434237978462e-08, 60.7499999775672, 0, 0, 0, 0, 0, 0, 2.90311248062574e-07, 0, 3.08070350135363e-07, 0, 0, 4.21053691897217e-07, -546.749999958454, 4.06006616710096e-07, -1248.41250005989, 0, 0, 0, 0, 0, 0, -9.01126364771612e-07, 0, -6.93166626178098e-07, 0, 0, 0, -1.81239670725005e-06, 0, -3.07792426028597e-06, -4.31743620149434e-07, -2.7818152786041e-06, 9226.40625060776, -1.1704430602393e-06, 11235.7124997726, -2.80776122863083e-06, 0, 0, 0, 0, 0, 0, 1.50780732917734e-06, 0, 1.39932146674014e-06, 0, 0, 0, 5.77503736420842e-06, 0, 1.07784700554259e-05, -1.10332120226048e-07, 7.92085973392166e-06, -31369.7812534785, 0, 0, 1.12917289814631e-05, 2.52869013230552e-06, -2.1795608854995e-06, 6.8186625599335e-06, 1.37896406598265e-05, -83037.6562557713, 7.54709981656414e-06, 3.46011751256454e-06, 0, 0, 0, 0, 0, 0, -2.38630784312516e-06, 0, -8.60351703605611e-07, 0, 0, 0, -5.04927197358424e-06, 0, -1.47528886079512e-05, -5.24658153032477e-07, -1.36693083683652e-05, 49822.5937588193, 0, 0, 0, -1.47890183127858e-05, 0, -2.76590228012895e-05, -8.55704040908292e-07, 6.3968267020905e-07, -1.57097308357885e-05, -3.28763363912616e-05, 282328.031259504, -8.21887826586552e-06, -7.83508700491532e-06, 1.90041622116582e-06, -1.32552518985085e-05, -1.66011788991761e-05, -1.3066460830398e-05, -2.86262568070901e-06, 0, 0, 0, 0, 0, 0, 1.41218314014285e-06, 0, -6.68647449791726e-08, 0, 0, 0, 3.39904453369568e-06, 0, 1.19447131834497e-05, 2.28458709791633e-07, 8.21116619261901e-06, -29893.556254691, 0, 0, 0, 1.17214999529234e-05, 0, 2.60048392084593e-05, -2.00476671289658e-06, 8.24071768716931e-07, 2.25630999610251e-05, 2.88577618572929e-05, -448403.3437638, 7.25037557850915e-06, 0, 0, 2.07248931314324e-05, -2.53301086585136e-06, -4.23641915244466e-06, 9.6998959050807e-06, 3.4637045085803e-05, -1.90504495634482e-05, 3.59141855187696e-05, 1.00571289948552e-06, 5.98357843806357e-05, -1.78926393461181e-05, 3.97512304302053e-05, 4.63701592513993e-06, 6.27876099349758e-05, 0, 0, 0, 0, 0, 0, -3.49678222546025e-07, 0, 4.00300343815963e-08, 0, 0, 0, -8.58393375997047e-07, 0, -3.12583066158573e-06, -7.74718179804387e-08, -2.67259092026263e-06, 1.68022555310834e-06, 0, 0, 0, -2.20762451419573e-06, 0, -1.00249832309517e-05, 2.36901151310656e-08, -9.15562429218921e-07, -8.58686610883839e-06, -9.17881530046475e-06, 269042.00625602, 2.83015211882206e-06, 0, 0, 0, -5.60543920177325e-06, 0, -1.03167163723485e-05, -1.992529064534e-07, 5.5871785060102e-07, -1.18947548302802e-05, -1.7317547685895e-05, 4.17625255323334e-06, -1.74180057351015e-05, 2.32280289485855e-06, -2.89731628030595e-05, 8.82070280096352e-06, -5.94910508986182e-06, 4.04096176189917e-06, -2.87405465265044e-05, 1.9102291751904e-06, 5.14516085624057e-07, -2.00595567059198e-05, -4.40215720579675e-05, 2.05043265338922e-05, -2.56555609544877e-05, -2.4614656887035e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.7466988132436e-08, 0, -5.30966706399507e-08, 48.5999999989405, 0, 0, 0, 0, 0, 0, 2.02080055555324e-07, 0, 4.05130862800204e-07, 0, 0, 8.97985806826088e-07, -656.100000076254, 2.19937428269578e-07, -911.249999784609, 0, 0, 0, 0, 0, 0, -8.3843406091554e-07, 0, -1.50704372959759e-06, 0, 0, 0, -2.55488108541514e-06, 0, -4.80352003944369e-06, 1968.30000083637, -4.65877765859302e-07, 5740.87499902934, 9.98821902527427e-08, 12301.8750012258, -3.00144548428746e-06, 0, 0, 0, 0, 0, 0, 1.41831803050929e-06, 0, 2.46486322492385e-06, 0, 0, 0, 2.44699016883456e-06, 0, 1.39810407501129e-05, -6.00301928764423e-07, 2.01900134210761e-06, -14762.2499975526, 0, 0, 1.28780288241068e-05, -3.50435246436742e-06, -36905.6250034694, 1.55047935735877e-06, 9.49066398130002e-06, -77501.8125010309, 1.19132929209576e-05, -6.22046327645924e-07, 0, 0, 0, 0, 0, 0, -1.16254043270367e-06, 0, -2.72596299199565e-06, 0, 0, 0, -2.36364969657525e-06, 0, -1.79913959652387e-05, 1.22117411494627e-06, -2.99988157563445e-06, 13286.0249967626, 0, 0, 0, -1.21748199877642e-05, 0, -2.85977299390247e-05, 2.28380803960902e-06, 5.98887493836538e-06, -4.32327550430904e-06, -8.2219931646204e-06, 199290.375006316, -3.56129254508471e-05, -2.43498065862364e-07, 232505.437510033, -5.33863425673629e-05, 1.71970122506129e-05, -3.16980226662739e-05, -1.40522382917143e-05, 0, 0, 0, 0, 0, 0, 9.03085154802719e-07, 0, 1.26950516273302e-06, 0, 0, 0, 1.99836457491234e-06, 0, 1.21096538705214e-05, -3.51241683395866e-07, 3.46306028148942e-06, 5.607042670294e-07, 0, 0, 0, 3.43750151255797e-06, 0, 3.03552577045195e-05, -1.7902800277562e-06, -6.78326580468101e-06, 1.37124491170535e-06, 8.4453713955875e-06, -179361.337504017, 2.55772251957191e-05, 0, 0, 2.33971569217682e-05, 7.18280616444608e-07, -1.67951575675703e-05, -5.23217298261406e-07, 3.04088194229693e-05, -597871.125005277, 4.81839102336068e-05, -1.68469624437462e-05, 1.52300749904791e-05, 2.05212410638518e-05, -2.50542050396648e-05, -6.91423902690177e-06, 4.12314328992026e-05, 0, 0, 0, 0, 0, 0, -1.91432542062203e-07, 0, -2.60246429766081e-07, 0, 0, 0, -1.04716042206475e-06, 0, -3.57523220901418e-06, -1.42916657038751e-08, -1.18497418339837e-06, 1.43987102011388e-07, 0, 0, 0, -1.41658756220574e-06, 0, -1.19176735217821e-05, 6.98060932589762e-07, 1.98826393390853e-06, -5.41229082809467e-07, -3.50285662042971e-06, 1.1282001038668e-06, -1.25873877540239e-05, 0, 0, 0, -7.66270515404579e-06, 0, -1.5068156908951e-05, 1.54545141768332e-06, 5.12386661138385e-06, 1.53717547035071e-06, -1.07855684484673e-05, 538084.012506582, -2.81242224509396e-05, 7.64816344371837e-06, -2.77259735908916e-06, -1.11590174470123e-05, 1.45006233040609e-06, 4.46223569121717e-06, -2.97835594114235e-05, 1.67893892887735e-06, 1.37279058363179e-05, -1.75536828629065e-06, -1.61895813457306e-05, -4.02896911508326e-06, -1.47294725838469e-05, -8.70789566576447e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.08876656818098e-08, 45.5625000309785, 0, 0, 0, 0, 0, 0, -4.25769714512792e-08, 0, 3.17308038220196e-07, 0, 0, 1.71806834115407e-07, -751.781250211983, -4.5857703350035e-07, -751.781250247488, 0, 0, 0, 0, 0, 0, 4.62026026514089e-07, 0, -1.33161517949575e-06, 0, 0, 0, -1.19138080831588e-06, 0, -1.54285961349347e-06, 3690.56250057369, 1.83570082883584e-06, 3690.56250047525, 1.76600276837925e-06, 12404.3906266722, 1.6803850177545e-06, 0, 0, 0, 0, 0, 0, -8.77721738546863e-07, 0, 2.63157007948028e-06, 0, 0, 0, 1.6979488590078e-06, 0, 4.78586818385115e-06, -5535.84375138452, -7.43052190478906e-06, -5535.84374868695, 0, 0, 3.35138655715589e-06, -3.00880407434975e-06, -60894.281254662, -8.70608083508596e-06, -8.89650719902564e-06, -60894.2812554616, 3.87730215763947e-06, 7.82730596546056e-08, 0, 0, 0, 0, 0, 0, 1.37796111197815e-06, 0, -2.98292034797057e-06, 0, 0, 0, -4.00575347040509e-06, 0, -1.17775184628878e-05, 1.3957871495257e-06, 1.26153170036013e-05, -4.25329564513357e-06, 0, 0, 0, -9.80586216767648e-06, 0, -1.07391500059378e-05, 8.29229512831757e-06, 91341.421885503, 1.90752513592902e-05, 1.50188110784324e-05, 91341.4218823501, -3.0347581847106e-06, 1.19337374865921e-05, 298935.562518002, 1.17465551853805e-07, 1.80487595203288e-05, 2.6327815834145e-05, 4.13227229258681e-06, 0, 0, 0, 0, 0, 0, -1.03903043680028e-06, 0, 2.08960240796119e-06, 0, 0, 0, 2.08070798529339e-06, 0, 7.36263479141133e-06, -6.25437169458435e-07, -7.91931076911911e-06, 2.37892620864121e-06, 0, 0, 0, 6.80876005242939e-06, 0, 1.28023866965329e-05, -1.42053073033503e-06, -7.00420265592664e-06, -2.04782347414667e-05, -1.5693145446983e-05, -4.73403092687173e-06, 8.27237420376978e-06, 0, 0, 1.45445021594273e-05, -6.73864515265007e-06, -448403.343757477, -8.80709260528197e-06, 1.14150436186071e-05, -448403.343761239, 2.17685092981977e-05, -1.85103809810957e-05, -4.57304832026574e-05, 1.06649479778916e-05, -1.1073267675329e-06, -2.84705801130751e-05, -3.7329889880759e-05, 0, 0, 0, 0, 0, 0, 2.49148630981022e-07, 0, -4.79330504173791e-07, 0, 0, 0, -1.06469189689812e-07, 0, -1.94558697917146e-06, 2.22905465329153e-07, 2.84252021970481e-06, -9.06697675243384e-07, 0, 0, 0, -2.87186983196985e-06, 0, -3.40868843947364e-06, 8.05649513379917e-07, 2.24037231262052e-06, 8.02633171452996e-06, 6.03119103351934e-06, 5.3884031403448e-07, -4.4064449975167e-06, 0, 0, 0, -4.30743133216199e-06, 0, -5.01387794044486e-06, 9.7875166339682e-07, 5.78535395974518e-06, 1.01426265092025e-05, 4.16201760667417e-06, 5.07396785586257e-06, -3.71966258978417e-06, 3.37630609198491e-06, 1.59459675096661e-05, -1.43124600256932e-06, 7.33108963199161e-06, 672605.015636399, 4.32580390315377e-06, 3.77216020779691e-06, 1.40337224849667e-05, 9.21385040250221e-06, 9.89226414360429e-06, -3.87706572781502e-06, -1.06386059799841e-05, 2.13094549180113e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.17130121961822e-08, 0, 2.28732759516966e-08, 48.60000000973, 0, 0, 0, 0, 0, 0, -1.06137568845163e-07, 0, -7.90820599217241e-08, 0, 0, -1.8218462423969e-07, -911.25000007278, -1.86483506069328e-07, -656.100000028555, 0, 0, 0, 0, 0, 0, 1.43173215235138e-07, 0, 2.74326001228781e-07, 0, 0, 0, 1.45784317893273e-06, 0, 1.06989709559652e-06, 5740.87500033268, 1.74285889304592e-06, 1968.3000000126, 9.86907715616731e-07, 12301.875000694, 1.87508818414304e-06, 0, 0, 0, 0, 0, 0, -4.46602636384993e-07, 0, -2.69334189385848e-07, 0, 0, 0, -3.11215133225791e-06, 0, -2.46104323864153e-06, -14762.2500001725, -2.707947635882e-06, -3.13066628735129e-07, 0, 0, -3.43733029779823e-06, -2.53994402228417e-06, -77501.8125029408, -3.70698839468511e-06, -7.18591666865661e-06, -36905.6250007104, -9.53658560656879e-06, -1.22845799679834e-05, 0, 0, 0, 0, 0, 0, 3.93292300236567e-07, 0, 5.04752958222256e-08, 0, 0, 0, 5.15404453466984e-06, 0, 6.54265048981606e-06, 13286.0250000774, 2.60498484736792e-06, 3.16794133296159e-08, 0, 0, 0, 9.7403978181319e-06, 0, 8.30934558096953e-06, -2.05102415881527e-06, 199290.375002133, 6.31561578839597e-06, 1.82682653625655e-05, 7.78623377347221e-07, 6.58350553157787e-06, 4.75705161546319e-06, 232505.437503012, 2.53498351681603e-05, 9.60797545898586e-06, 1.59717092149752e-05, 2.17151636221942e-05, 0, 0, 0, 0, 0, 0, -2.15433683968093e-07, 0, -1.71075364125937e-07, 0, 0, 0, -3.03562517487837e-06, 0, -3.01331703667657e-06, -7.78772189589271e-08, -3.18454326911819e-06, 1.31851589409718e-06, 0, 0, 0, -6.74361632615917e-06, 0, -8.94195789712942e-06, -7.59305366795688e-07, -179361.337502474, -4.95231806675917e-06, -1.55949381058446e-05, 1.60996950653094e-06, -3.4863562393303e-06, 0, 0, -1.1944491187149e-05, -4.10851129818531e-06, -597871.125012341, -1.29988686699694e-05, -4.55252226315062e-05, -6.57378804390288e-06, -3.69700361188463e-05, -4.75620964123516e-06, -1.36438919395394e-05, -2.27455818317593e-05, -4.09828985110528e-05, -9.51218573634416e-06, -1.97547359764273e-05, 0, 0, 0, 0, 0, 0, 8.98893840864608e-08, 0, -2.41977215768996e-08, 0, 0, 0, 5.77433105959615e-07, 0, 5.77397462291863e-07, -5.70624239765166e-08, 5.88674063242833e-07, -4.54749308437468e-07, 0, 0, 0, 2.45691288032938e-06, 0, 2.07655323990064e-06, -4.44439397164722e-07, 7.55026415919982e-07, 1.39975219874696e-06, 5.06658297370224e-06, -4.47638625746423e-07, 7.67903342031647e-07, 0, 0, 0, 3.52696327589093e-06, 0, 4.02990818161306e-06, 1.12581699812901e-06, 538084.012502808, 2.25045936775742e-06, 1.29322897803606e-05, 1.64486935589135e-06, 5.44528407685449e-06, 3.87679118383563e-06, 1.03233446367075e-05, 2.60726741461181e-06, 9.37103587032727e-07, 7.20173170740647e-07, 1.17327579840377e-05, 8.90180864551268e-06, 5.77409790019515e-06, 1.6319543747874e-05, 2.34744091537958e-05, 1.71827938810417e-05, 2.79299815176414e-05, 8.90131954673883e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.15315041562944e-08, 60.7500000009195, 0, 0, 0, 0, 0, 0, 0, 0, 1.28917505521433e-07, 0, 0, -2.28138594342162e-08, -1248.41250004187, -6.62347818734719e-08, -546.750000045737, 0, 0, 0, 0, 0, 0, 2.88435823524711e-07, 0, -4.1895982332724e-07, 0, 0, 0, -1.68839097956598e-07, 0, -6.27057642502884e-07, 9226.4062499516, -1.3274038716861e-07, -7.27006311023985e-08, 1.10279609224716e-08, 11235.712499775, 1.35058281451207e-07, 0, 0, 0, 0, 0, 0, -4.13694917956715e-07, 0, 8.6207848678583e-07, 0, 0, 0, 9.50986440476039e-07, 0, 1.42472689425495e-06, -31369.7812506731, -1.16748632213686e-06, 2.13527397988439e-07, 0, 0, 6.84548484522752e-07, 4.09037052097523e-07, -83037.6562507357, -1.27698329406825e-06, 5.92631048980018e-07, -9.85796110371154e-07, 4.64324284773435e-06, 5.53354033680211e-06, 0, 0, 0, 0, 0, 0, 1.84236909474343e-07, 0, -6.01390498932242e-07, 0, 0, 0, -1.0271272843265e-06, 0, -4.5031940397194e-06, 49822.5937504122, 2.57908153027203e-06, -4.03590446110512e-07, 0, 0, 0, -4.73817747564751e-06, 0, -3.6268136083495e-06, 3.40001421170804e-06, 282328.031252951, 3.83271958869431e-06, -3.23933801082076e-06, -6.56029144432989e-07, -1.19021679708842e-06, 4.61960378340768e-06, 4.08414017608835e-06, -4.4434154705309e-07, -7.19872250928439e-06, 3.1409929418048e-06, -9.11206849600128e-06, 0, 0, 0, 0, 0, 0, -3.02038529691742e-07, 0, 5.891559658804e-07, 0, 0, 0, 9.43020289990072e-08, 0, 2.38445860609466e-06, -29893.5562503469, -1.35130636471887e-06, 4.39309484371068e-07, 0, 0, 0, 3.14048110612878e-06, 0, 5.75501696118746e-06, -1.83314929031235e-06, -448403.343751267, -3.41576359525861e-06, -2.60274052426309e-07, -1.90704279401805e-06, 5.36578134558553e-06, 0, 0, 5.1248392079402e-06, -3.5614649974822e-06, 9.36042771335939e-07, -7.40368613562076e-07, 1.43068551662282e-05, -6.73733807482303e-06, 2.32390725194141e-05, 5.69445294183262e-08, 2.39363612106262e-06, 1.23973436929296e-05, 2.84496151399367e-05, 4.68317333516305e-07, 5.91733721776164e-06, 0, 0, 0, 0, 0, 0, 2.1634567408369e-08, 0, -1.11077379156104e-07, 0, 0, 0, 2.15556127135047e-07, 0, -4.48229767189529e-07, 1.17109966774438e-07, 6.11535480641272e-07, -2.44545486531611e-07, 0, 0, 0, -4.19160294099016e-07, 0, -1.40383971020139e-06, 6.54396523200112e-07, 269042.006250467, 1.36586785778436e-06, 1.41059061631695e-06, 1.01003260987058e-06, 4.04680814103402e-07, 0, 0, 0, -1.7827416451875e-07, 0, -1.31499210153919e-06, 1.10978455444919e-06, 2.08180075758229e-06, 1.91070538584939e-06, -1.65889861810909e-06, 1.40075695067895e-06, -2.18064548430153e-06, -3.80926951572666e-07, -1.62270288314105e-06, -8.55480834028248e-07, 2.57509654508023e-06, 4.10546225469044e-06, 6.05544614355578e-07, -6.00723835666567e-07, -6.13677493722196e-06, -6.19015327383027e-06, -7.86411909789273e-06, -3.350792880015e-06, -2.00889936554447e-05, -3.71345727131235e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 36.4500000158067, 0, 0, 0, 0, 0, 0, -3.0962205946284e-08, 0, 0, 0, 0, 4.48970367219407e-07, -683.437500058619, -328.050000001366, -328.049999964077, 0, 0, 0, 0, 0, 0, -3.72062306539672e-07, 0, -1.51896293595298e-07, 0, 0, 0, -1.26661235731876e-06, 0, -6.92657115970026e-07, 4305.65625075932, 9.69103741314852e-07, 4.76846017624031e-07, 6150.9375011146, 6150.93750153379, 2952.44999810895, 0, 0, 0, 0, 0, 0, 2.89563916685344e-07, 0, -1.26115710302689e-07, 0, 0, 0, -7.63794602336342e-07, 0, 2.72742490093432e-06, -11071.6874997687, -1.62656145205282e-06, -1.02389224965738e-07, 0, 0, 3.39939610066097e-06, -38750.906254329, -38750.9062497232, -3.23580031522136e-06, -2.23893071696323e-06, -1.96543762090584e-06, -1.94025182431993e-07, -55358.4375030834, 0, 0, 0, 0, 0, 0, 8.34962020741592e-07, 0, -7.37395340081525e-07, 0, 0, 0, -2.0687545838456e-07, 0, -1.47908690944882e-06, 9964.51875076303, 1.89156769258593e-06, 6.85337395342073e-07, 0, 0, 0, 3.79793400932086e-06, 0, -4.81855547791167e-06, 99645.1875014129, 99645.1874986884, 5.82240735219484e-06, 1.08238870525074e-05, 1.15967656790573e-05, -1.12771634406278e-05, 2.07792186143978e-06, 6.21791089636184e-07, -2.16604661105151e-05, 348758.15626726, -1.17443275561028e-05, -5.86466232207035e-06, 0, 0, 0, 0, 0, 0, -1.40231384027554e-08, 0, -1.90825444808502e-08, 0, 0, 0, 9.43240659497892e-07, 0, 1.24275554742324e-06, 8.40822139457759e-08, -1.18299385344033e-06, -2.70827124413773e-06, 0, 0, 0, -5.25088684586405e-06, 0, 1.00458419169505e-06, -89680.6687486102, -89680.6687512043, -7.92891959025923e-06, -4.72248418674575e-06, -6.17616888953945e-06, -4.93728031731678e-06, 0, 0, 4.20646279800526e-07, -3.27908186113834e-06, -5.38186061630836e-06, -1.20655470393571e-05, -1.02071421928598e-05, 9.00832947820044e-06, -2.30090612104373e-05, -896806.687501473, -2.83201232050286e-05, 2.93406888228434e-06, -4.87867181089125e-05, 1.31524843057802e-05, 2.43648623383415e-07, 0, 0, 0, 0, 0, 0, 1.32683963366653e-07, 0, -1.07952266199852e-08, 0, 0, 0, -5.9627252751245e-07, 0, -6.30705819361732e-07, -4.99334665579721e-08, 1.55556085357359e-07, 1.20512552953408e-06, 0, 0, 0, -1.57725875841876e-07, 0, -1.28414968156203e-06, -3.73503223164452e-08, 1.05915244554426e-07, 2.9453217868726e-06, -1.63182381999685e-06, 7.86056184945627e-07, -5.20244620680307e-06, 0, 0, 0, -2.30161542556807e-06, 0, -2.9515891881197e-06, 1.27453919303449e-06, -2.18812591713158e-06, 6.54142100243674e-06, 4.84836360532379e-06, 2.81712286989919e-06, -2.47224633566156e-06, 807126.018748453, 1.14723321405377e-05, -5.83250595331664e-06, 5.17113344153399e-06, -9.01488022172314e-07, -6.62201774669621e-06, -3.70026726284689e-06, 1.32019504522251e-05, 1.45617271955402e-05, 7.87363519217283e-06, -1.83555460073014e-05, 1.73768466454989e-05, -3.48483402404027e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.58694866066826e-08, 0, 5.95023937452193e-08, 36.4500000589788, 0, 0, 0, 0, 0, 0, -2.06099323153049e-08, 0, -5.02166276405151e-07, 0, 0, -4.08155224727996e-07, -328.050000366724, -683.43750046056, -328.050000825925, 0, 0, 0, 0, 0, 0, -6.70910042339718e-07, 0, 1.61378750552331e-06, 0, 0, 0, 9.2119217817223e-07, 0, 3.66735036319627e-06, 1.45229164415544e-06, 4305.65625136684, 4.48889644308676e-06, 6150.9375023813, 2952.45000333609, 6150.93750552564, 0, 0, 0, 0, 0, 0, 8.76895904666941e-07, 0, -3.50524091357305e-06, 0, 0, 0, -3.07003390233955e-06, 0, -7.58130847380763e-06, -1.94184709724258e-06, -11071.687503332, -1.08144881349996e-05, 0, 0, -1.15195256111492e-05, -1.08685005828099e-05, -4.50262502891622e-06, -38750.906258773, -38750.9062674498, -8.09941609328361e-06, -2.22179229698431e-05, -55358.4375150783, 0, 0, 0, 0, 0, 0, -9.00345951839987e-07, 0, 3.69514272987414e-06, 0, 0, 0, -1.65877866065976e-06, 0, 1.13539111527933e-05, 3.2674853760672e-06, 9964.51875335698, 1.53639128549819e-05, 0, 0, 0, 1.00563823183853e-05, 0, 2.10410471062429e-05, 1.66500728046136e-05, 4.79571608727838e-06, 99645.1875086529, 99645.1875220957, 2.11368095022667e-05, 4.76621096232804e-05, 1.50665363616353e-05, 1.00403919422449e-05, 3.53675643760091e-05, 5.18927504231331e-05, 348758.156279783, 4.73813938880853e-05, 0, 0, 0, 0, 0, 0, 5.80123160175713e-07, 0, -2.5925920503347e-06, 0, 0, 0, 3.02672720839176e-06, 0, -1.03637077637387e-05, -1.61665127691556e-06, 1.37569665151716e-06, -1.38248300883951e-05, 0, 0, 0, -1.45736069704733e-05, 0, -1.57542517549425e-05, -7.4211213424021e-06, -6.06292202037311e-06, -89680.6687562202, -89680.6687526571, -1.84205115085489e-05, -5.38527715643783e-05, 0, 0, -1.23769032131643e-05, -1.74623508653459e-05, 8.08842327244583e-06, -6.690978800137e-06, -2.77015577296982e-05, 9.04338267968123e-06, -6.7910781460068e-05, -4.76861941938176e-05, -896806.687551408, -5.0910220247782e-05, -8.7149353774265e-05, -5.56114723291735e-05, -0.000100998479588503, 0, 0, 0, 0, 0, 0, -1.97004248857479e-07, 0, 6.18766936086538e-07, 0, 0, 0, -1.38917236218387e-06, 0, 2.44921769179984e-06, 5.67589769875545e-07, 2.24647126125603e-07, 4.61029876466662e-06, 0, 0, 0, 8.05153779745127e-07, 0, 6.32166386541203e-06, 2.92351030827857e-06, 1.65279659224488e-06, 2.55986720771702e-06, 3.16990364612566e-07, 6.14934697685677e-06, 1.31198122824553e-05, 0, 0, 0, 4.32602692333741e-06, 0, 7.15422645815403e-06, 6.62973247097508e-06, -1.01002397557131e-06, 7.54292275906572e-06, 1.16186606011605e-05, 4.44197006220796e-06, 3.66736811153068e-05, 1.25943852206524e-05, 807126.018763136, 2.7021093172294e-05, 1.31757929665222e-05, -3.10316447658348e-06, 3.4803263460114e-05, 2.04096522971509e-05, 1.80060995900088e-05, 3.43196603439275e-05, 4.81441109737371e-05, 3.77073234238317e-06, 6.1585413630667e-05, 5.89241019749842e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.35796511899467e-08, 0, 0, 36.4500000133207, 0, 0, 0, 0, 0, 0, 1.86489943646532e-07, 0, -6.40781053565536e-08, 0, 0, 2.84969422762441e-07, -328.049999984266, -328.04999985872, -683.437500001319, 0, 0, 0, 0, 0, 0, -4.33391043303361e-07, 0, -5.27154037208581e-07, 0, 0, 0, -1.42184598076797e-06, 0, -1.15281656185647e-06, 5.8576725818269e-07, -2.29730337448632e-06, 4305.65625047864, 2952.44999887065, 6150.93750078394, 6150.93750150591, 0, 0, 0, 0, 0, 0, 2.26370088565373e-06, 0, 2.08268411367471e-07, 0, 0, 0, -1.77000360313411e-06, 0, -9.617987187408e-07, 1.37635695255325e-07, 6.82066107467851e-06, -11071.6874989581, 0, 0, -1.97585749762465e-06, -1.15087062546656e-07, 3.64126431644442e-06, 8.45737201983067e-06, 3.32040673701354e-06, -38750.9062442622, -38750.9062497067, -55358.4374880873, 0, 0, 0, 0, 0, 0, -1.70497754336849e-06, 0, -4.14228546115861e-07, 0, 0, 0, -3.68544138241757e-06, 0, -5.94191051627942e-06, 5.50540616512817e-07, -6.62456500568908e-06, 9964.51874109086, 0, 0, 0, 7.00163890499366e-07, 0, 3.2332546223732e-06, 4.25259354895169e-06, 2.25500440865917e-06, -2.18462181025668e-05, -2.45391618190673e-06, 99645.1874990263, 99645.1875093253, -7.51098290455841e-06, 1.10595354838535e-05, -3.9401199160036e-05, 2.76320461382174e-05, -4.38348803220014e-05, 348758.156241487, 0, 0, 0, 0, 0, 0, 1.18371377255082e-06, 0, 1.29666213335706e-06, 0, 0, 0, 2.85398642675218e-06, 0, -4.82073162034119e-07, -9.28345836019853e-08, 1.17077201517423e-05, 4.11854488710066e-07, 0, 0, 0, -6.49524102071827e-06, 0, 6.41501165482521e-06, 2.97531009549271e-06, -9.09798462247649e-07, 8.81085012093265e-06, 5.37049200980244e-06, -89680.6687438779, -89680.6687569549, 0, 0, 6.06357434917794e-06, 1.75107059325045e-05, 1.0102158264346e-05, 3.40967304857708e-05, 3.82103960759669e-05, 2.15751765957604e-05, 1.75135043627709e-05, -9.90410712647159e-06, -1.15956707592294e-05, -896806.687457665, 2.16388973481207e-06, -2.12338554168336e-05, -4.84107201309182e-05, 0, 0, 0, 0, 0, 0, -4.36768635564188e-07, 0, -1.67590250586266e-07, 0, 0, 0, -9.38404332491391e-07, 0, 4.33363333018327e-07, 9.3873563167076e-08, -3.01422525798172e-06, -8.19834007730042e-07, 0, 0, 0, 1.7757490387665e-07, 0, -2.96287672143606e-07, 7.87006139290411e-07, 1.61615220522962e-06, -4.1480543247772e-06, -4.87469633999435e-06, -3.04129836459103e-06, -4.79311794080096e-06, 0, 0, 0, -1.22683245985708e-08, 0, -4.02685063630591e-06, -1.31730708068257e-06, -1.35839059602175e-07, 5.13094484859031e-07, -4.68677328828421e-06, 1.79060809509254e-06, 7.95978943258052e-06, -2.29070537460794e-06, -1.76959423620119e-07, 807126.018747519, 6.83403022459299e-07, -3.60905412933343e-07, 1.94319035153976e-05, -1.80506912097198e-05, 5.92370890172431e-06, -1.91318532744857e-05, 6.78460907726072e-06, -3.65139031684851e-05, -3.38384172279069e-06, 2.11916901617707e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.30731013359152e-08, 0, 1.9970690045771e-08, 30.3750000238805, 0, 0, 0, 0, 0, 0, -2.31989450395364e-07, 0, -1.36371018905624e-07, 0, 0, -9.75127606201139e-07, -501.187500180869, -410.062500763756, -273.375000523304, 0, 0, 0, 0, 0, 0, 1.89272602547364e-06, 0, 8.26477949889021e-07, 0, 0, 0, 3.29696585562426e-06, 0, 1.9751525836406e-06, 2460.37499918099, 1230.1875029459, 1.65362957342351e-06, 6766.03125144786, 4510.68749887403, 3690.56250711937, 0, 0, 0, 0, 0, 0, -3.26003923727183e-06, 0, -9.89195288572484e-07, 0, 0, 0, -1.24709363554278e-06, 0, -7.42777306316701e-06, -3690.56250093084, -1.03757310838248e-05, -7.20209882865594e-06, 0, 0, -9.17208388495761e-06, -33215.0624995344, -22143.3750030037, -20298.0937621961, -11071.6875112156, -1.45898678156937e-06, -9.32487754400433e-06, -60894.2812532889, 0, 0, 0, 0, 0, 0, 1.81961997402967e-06, 0, 2.84496667520471e-06, 0, 0, 0, 3.42783635307398e-06, 0, 6.02647776837732e-06, -1.860374274121e-06, 1.71693050718622e-05, 1.02433324786661e-05, 0, 0, 0, -2.08185372421739e-07, 0, 1.4800568542112e-05, 49822.5937528183, 33215.0625045846, 2.86483708139282e-05, 1.73495700096201e-05, -7.26558045928809e-06, 4.34655995558839e-05, 99645.1875252597, 7.64480619641494e-06, 7.21919852623604e-05, 298935.562485372, 182682.843814327, 3.14530583276622e-05, 0, 0, 0, 0, 0, 0, -2.25392421199973e-06, 0, -9.29629565184056e-07, 0, 0, 0, -4.59725805659994e-06, 0, -4.78093670379954e-06, 3.84000665298054e-07, -1.4147380673363e-05, -4.34732120784012e-06, 0, 0, 0, 4.2284165274408e-06, 0, -6.71076655642291e-06, -3.91242291161235e-06, 2.435395184946e-06, -2.46668867938076e-05, -2.91784267314766e-05, -4.04081789298066e-06, -1.12576217225265e-05, 0, 0, -7.96168192614665e-06, -149467.781265362, 6.68792259853075e-06, -2.17548896003623e-05, -2.08485651383322e-05, -2.82535085293285e-05, 1.05097463778977e-05, -448403.343752439, -1.79964723623593e-05, -2.2553081594771e-05, -896806.687465647, -3.2128707962032e-05, -4.56105002197868e-05, 0, 0, 0, 0, 0, 0, 3.46105690612013e-07, 0, 2.7622941222326e-07, 0, 0, 0, 2.14493879533323e-06, 0, 1.93091023520102e-06, -5.63205122840871e-08, 4.61030317700075e-06, 6.08290752279831e-07, 0, 0, 0, 2.15650113657605e-06, 0, 3.91347843947151e-06, -4.92949555969145e-07, -2.67626661088411e-07, 8.54604772280612e-06, 1.68797788412673e-05, 5.53867275937399e-06, 1.79803408072419e-05, 0, 0, 0, 8.17768809033302e-06, 0, 8.61480420104368e-06, 2.85907404073091e-06, 5.49218574628204e-06, 9.57602635111489e-06, 1.16160984764082e-05, 8.4039312685703e-07, 1.75687554173276e-05, 1.97168574241821e-06, 1.48090466672317e-05, 2.12779056952626e-05, 7.57340500119487e-06, 5.49791583794546e-06, 3.2619471621345e-05, 1345210.03126893, -2.30748398274132e-05, 1.28443629949138e-05, 2.02520734831124e-05, 3.91285447261946e-05, -1.79817190021455e-05, 2.31344553449783e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7.81825197647067e-08, 0, 3.11480357780338e-08, 30.3750000448221, 0, 0, 0, 0, 0, 0, -5.88643861299768e-07, 0, -4.78326052316279e-08, 0, 0, 1.16687533541844e-07, -410.062500213906, -501.187500555016, -273.375000156528, 0, 0, 0, 0, 0, 0, 1.65509677873064e-06, 0, 5.98910726500674e-07, 0, 0, 0, 1.5596585884291e-06, 0, -3.28160058470406e-07, 1230.1875011203, 2460.37500584263, 1.00409324287506e-06, 6766.03125418529, 3690.56250216481, 4510.68749881, 0, 0, 0, 0, 0, 0, -4.47251397620071e-06, 0, -1.22723640499131e-06, 0, 0, 0, -7.22412844734074e-06, 0, 4.35485576548108e-06, -1.56496003499199e-06, -3690.56251528666, -4.44910227456216e-06, 0, 0, 5.86183474686367e-06, -20298.0937572145, -11071.6875050787, -33215.062516848, -22143.3750199283, -1.0411355474579e-05, -5.28551368211368e-06, -60894.2812756982, 0, 0, 0, 0, 0, 0, 6.011497496372e-06, 0, 6.63578747312926e-07, 0, 0, 0, 1.84993830227074e-05, 0, 2.187319936741e-06, 3.04848307742212e-06, 1.93219593177006e-05, 1.43084014434427e-05, 0, 0, 0, 1.35874586146135e-05, 0, -1.32731211399834e-05, 2.47786568749611e-06, 2.17503845251242e-06, 49822.5937896471, 33215.0625510987, 2.08663889763961e-05, -2.34914815704887e-05, 99645.1875170266, -6.07355237402113e-06, 1.41432975387334e-05, 182682.843758996, 298935.562559613, -4.70218263401299e-07, 0, 0, 0, 0, 0, 0, -2.83223848755162e-06, 0, -1.67738854423113e-06, 0, 0, 0, -1.34063720938846e-05, 0, 2.38121821450386e-06, -1.67138936401374e-06, -1.93453927217611e-05, -9.43409603459435e-06, 0, 0, 0, -1.07705563481537e-05, 0, 7.88775998919042e-08, -6.40688895644973e-06, -5.1799216062676e-06, -3.26996920111791e-05, -5.44702059819712e-05, -2.18310920673227e-05, -3.97091707864445e-07, 0, 0, -5.85349070424967e-06, -1.90593104372226e-05, -2.60613649868997e-05, -149467.781304042, -8.34809131517869e-05, -1.51975301292179e-05, -3.89873007929245e-05, -2.4767113318187e-06, -448403.343835863, -2.60682947094273e-05, -896806.687582671, 3.03437451086809e-05, 2.62507912715985e-05, 0, 0, 0, 0, 0, 0, 9.91943715221643e-07, 0, 3.25387500125515e-07, 0, 0, 0, 3.72437788382113e-06, 0, -6.87259276070179e-07, 2.33542016798737e-07, 4.52485115547783e-06, 3.82887156417896e-06, 0, 0, 0, 5.8203746086348e-06, 0, -2.88835916340508e-06, 1.69749435844608e-06, 4.71623166850299e-07, 1.23087289477296e-05, 1.70826849673792e-05, 7.71630239189936e-06, 0, 0, 0, 0, -1.57830358108396e-06, 0, -3.71172504852426e-06, 3.70016631038988e-06, -5.34642327133022e-07, 1.16360563029799e-05, 3.18766414369808e-05, 6.43500531589561e-06, -1.67819852805647e-05, 6.1901965886147e-06, 4.68451398246188e-05, -7.85880155997867e-06, 3.5833313122307e-06, 1.81957330271495e-06, -2.46281091573976e-05, 2.80797157861524e-05, 1.88757923123005e-05, 1345210.03130564, 2.68407468562745e-05, 2.54315917933615e-06, 1.50453225590043e-05, -1.46835341434501e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.76540551281839e-08, 0, -6.22384453189659e-08, 30.3750000806313, 0, 0, 0, 0, 0, 0, 1.33845008151257e-07, 0, 4.73462851895553e-07, 0, 0, -1.75605161509428e-07, -273.375000654168, -501.187501311977, -410.06250076746, 0, 0, 0, 0, 0, 0, 1.09788610218521e-06, 0, -1.75930055264068e-06, 0, 0, 0, 8.67310423046471e-07, 0, -2.82978281359976e-06, 1.08562925669897e-06, 2460.3750041297, 1230.18750204989, 4510.68750686606, 3690.56250318528, 6766.03126263216, 0, 0, 0, 0, 0, 0, -2.39577982823374e-08, 0, 3.8395589650031e-06, 0, 0, 0, 2.70378477558977e-06, 0, 1.57576263337777e-06, -3.71849801947654e-06, -3690.56250777268, -5.19006615659543e-06, 0, 0, 1.40912940810234e-06, -1.24065079910881e-05, -1.16271903760114e-05, -22143.3750231056, -33215.0625233415, -11071.6875118436, -20298.0937744677, -60894.2812907606, 0, 0, 0, 0, 0, 0, -1.36797583243169e-06, 0, -2.65308441870563e-06, 0, 0, 0, -2.80024550002031e-06, 0, -1.15830715110001e-05, 3.50762961803281e-07, 1.12689622186888e-05, 4.17594112211687e-07, 0, 0, 0, -5.37798146131368e-06, 0, -3.4963231750209e-06, 1.91136837822005e-05, 1.75484929653235e-05, 33215.06253832, 49822.5937842797, 7.69935092493911e-06, 5.9436748987426e-05, 5.8124344827081e-05, 3.7968991769659e-05, 99645.1875920053, 6.63127361673324e-05, 298935.562642275, 182682.843869021, 0, 0, 0, 0, 0, 0, 0, 0, 3.19139151692916e-06, 0, 0, 0, -1.8481187045356e-07, 0, 8.37817770447818e-06, -8.37993347241416e-07, -7.38999014413537e-06, 2.80429283249706e-06, 0, 0, 0, 1.33536740509136e-05, 0, 1.12306324078199e-05, -1.53100087035711e-05, -1.48493371896541e-06, -3.10647732709858e-05, -4.11996136693063e-05, -8.75418978480368e-06, -1.22205235418232e-05, 0, 0, 1.7412995987039e-06, -3.69351309991553e-05, -2.00031213915461e-05, -4.56368923951361e-05, -149467.781299377, -5.55893446973842e-05, -1.34828587744919e-05, -5.40046464105251e-05, -448403.343852633, -9.4161450408874e-05, -7.84650597192708e-05, -0.000152755611975236, -896806.687699995, 0, 0, 0, 0, 0, 0, -2.03059504082112e-07, 0, -5.92667859427628e-07, 0, 0, 0, 7.70258485093065e-07, 0, -1.31431992497953e-06, 7.23628255073278e-08, 1.64776173763479e-06, -2.7845747520745e-06, 0, 0, 0, 4.88576567825734e-07, 0, -2.8856741128228e-06, 2.6186305051467e-06, 8.74138115063185e-07, 6.71757324706195e-06, 1.36328936576081e-05, 5.03196372504656e-06, 1.61800256259757e-05, 0, 0, 0, 5.78343207126147e-06, 0, 8.21192322737044e-07, 1.23181184769657e-05, 1.02324030417674e-05, 1.78584912237443e-05, 1.99877426264786e-05, 1.10717032418352e-05, 2.17560884328352e-05, 1.66705953297061e-05, 3.96818015715676e-05, 4.86926720793104e-05, 1.81460840795534e-05, 1.85527558938352e-05, 4.81571015281782e-05, 6.97730526980062e-05, 3.49345721297338e-05, 5.90639829214238e-05, 1345210.0313338, 9.39584029945303e-05, 4.17474677480507e-05, 0.000120702615138277}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.81179295180977e-08, 0, 0, 30.3749999973723, 0, 0, 0, 0, 0, 0, 2.5130108650642e-07, 0, 8.81407844123321e-08, 0, 0, 3.06957146478414e-07, -273.375000215368, -410.062499953458, -501.187500215954, 0, 0, 0, 0, 0, 0, -1.91357652514664e-06, 0, 3.43749086685758e-07, 0, 0, 0, -1.49350015769819e-06, 0, 5.06964945263972e-07, 9.0760353942999e-07, 1230.18750065955, 2460.37500037236, 3690.5625024814, 4510.6875022738, 6766.03124996805, 0, 0, 0, 0, 0, 0, 1.68893139085197e-06, 0, -2.28701116597573e-08, 0, 0, 0, 6.81269571764828e-06, 0, 5.16598465176254e-06, -2.00843982270209e-06, 0, -3690.56249997173, 0, 0, 6.26200821898741e-06, -6.88801517048831e-06, -1.06297163851319e-05, -11071.6875090047, -20298.0937636392, -22143.3750107909, -33215.0625163301, -60894.2812836045, 0, 0, 0, 0, 0, 0, -2.02279079757554e-06, 0, -6.84122468930143e-07, 0, 0, 0, -4.75795129772944e-06, 0, 2.56106703725415e-06, 3.45069202527017e-06, -7.60292547028351e-06, 6.1406227480624e-06, 0, 0, 0, -8.08695337455146e-06, 0, -8.76359846091774e-06, 4.37449530359187e-06, 7.86633017419849e-06, 1.43345855392702e-05, 7.81890782232694e-06, 33215.0625065058, 49822.5937549141, 7.69561640616001e-06, -4.30167811420918e-06, 99645.1875581158, 2.49959032637339e-05, 182682.843822885, 298935.56255093, 0, 0, 0, 0, 0, 0, 1.58983153669712e-06, 0, -1.25751498508533e-06, 0, 0, 0, 4.83135705360876e-06, 0, 3.0957082230038e-06, -1.81712228645901e-06, 2.26046437407559e-06, 1.61032130745749e-06, 0, 0, 0, 8.6130143782611e-06, 0, -3.85309674431584e-06, -6.14998278313058e-06, -1.14276198575046e-05, 3.7262756943064e-06, 1.01548132163617e-05, -1.1502351102376e-06, -1.83467551413046e-05, 0, 0, 1.68449003851935e-06, -2.00963021830911e-05, -2.33330878993638e-05, -3.2409169479952e-05, -7.06106754661387e-05, -7.86541499133055e-06, -149467.781338475, -3.47952747540308e-05, -3.60994353716252e-05, -448403.343841449, -9.48254158789397e-05, -3.77141314049891e-05, -896806.687539478, 0, 0, 0, 0, 0, 0, -1.68690031804512e-07, 0, 1.03214407554719e-07, 0, 0, 0, -1.79219178657116e-06, 0, -2.01554462241481e-06, 4.12056093654526e-07, -1.33356743398193e-06, 6.27160130135852e-07, 0, 0, 0, -3.84783324636916e-06, 0, -1.17393214392863e-06, 1.07494658629722e-06, 1.72999585196537e-06, -9.69927919981593e-07, -4.83841757418255e-06, -2.41725150303272e-06, 3.50679297614385e-06, 0, 0, 0, -6.64987639414012e-06, 0, 3.6258321643122e-07, 1.90736649259063e-06, 3.54469653337086e-06, -1.95995196755054e-06, 1.16003598267262e-05, -5.15801329229282e-06, 1.57344314794092e-05, 1.58964481762759e-05, 9.79873790110341e-06, 1.27175315386614e-05, 1.03961598179952e-06, -2.91386755605694e-07, -2.04983823179669e-06, 3.82338840911354e-05, 3.02601864378961e-05, 5.3757599095031e-05, 4.2393454912851e-05, 4.47497778941013e-05, 1345210.03132671, 3.56209481199749e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.56155657426893e-08, 0, 3.304022218887e-08, 30.3750000146515, 0, 0, 0, 0, 0, 0, -1.24321782900624e-08, 0, -1.20457343242025e-07, 0, 0, -8.63329068452889e-07, -410.062500253653, -273.375000479184, -501.18750040412, 0, 0, 0, 0, 0, 0, 1.84787712174346e-07, 0, 9.05095942534872e-07, 0, 0, 0, 3.10385628264919e-06, 0, 3.3107219453907e-06, 1230.18750002162, 2.14263969640217e-06, 2460.37500099408, 3690.56250280739, 6766.03125070504, 4510.68750504332, 0, 0, 0, 0, 0, 0, -1.07300909321894e-07, 0, -4.39989588424863e-07, 0, 0, 0, 1.61621119514811e-06, 0, -8.78384308000389e-06, -2.39805503476114e-06, -2.68127645391423e-06, -3690.56250439511, 0, 0, -7.31763300823919e-06, -11071.6875028507, -20298.0937593009, -1.09176776094882e-05, -1.4049849442644e-05, -33215.06250779, -22143.3750170647, -60894.2812769194, 0, 0, 0, 0, 0, 0, -1.64362017821695e-06, 0, 8.17650502212424e-07, 0, 0, 0, 4.53284270444165e-07, 0, 1.39965785289058e-05, 7.64473277156763e-07, -9.38073432787444e-07, 8.88992877786143e-06, 0, 0, 0, 9.14562275929985e-06, 0, 1.80092053507829e-05, 1.87589717454525e-06, 1.02044286152826e-05, 1.9656387877912e-05, 1.49789524435109e-05, 49822.5937483899, 33215.062531143, 1.85974111375152e-05, 99645.1875056861, 9.18162270349324e-05, 182682.843753715, 9.76289789563341e-05, 298935.562555388, 0, 0, 0, 0, 0, 0, 6.41710294129299e-07, 0, -4.42812187500402e-07, 0, 0, 0, 1.83432740730531e-08, 0, -4.65318582687028e-06, -1.21530990292805e-06, -4.16850779176363e-06, 3.78607092092139e-07, 0, 0, 0, 7.24764697927425e-06, 0, -2.26625289455217e-05, -9.31675394839331e-06, -4.82232180923537e-06, -4.28350975947932e-06, -1.73041039641761e-05, -2.86444494662146e-06, -9.09055906917615e-06, 0, 0, -2.43111174616655e-05, -2.06023928030305e-05, -2.09417401321547e-05, -3.1179485566862e-05, -7.21925891016651e-05, -149467.781286883, -4.89487259951608e-05, -1.5323388768869e-05, -3.14584968705264e-05, -448403.343815409, -3.52483154888388e-05, -896806.687534543, -4.57074440757532e-05, 0, 0, 0, 0, 0, 0, -1.61413250920894e-07, 0, -1.24326528125201e-08, 0, 0, 0, 1.64425640633827e-07, 0, 1.36031478599326e-06, 2.07546287140325e-07, 1.67206344532178e-07, -5.29234562578743e-07, 0, 0, 0, 5.20890344386929e-07, 0, 6.40033986212399e-06, 8.18115018110357e-07, 1.10696812182016e-06, -5.45079622426942e-07, 6.18431532684227e-06, 2.89906660234189e-06, 1.49219308477719e-05, 0, 0, 0, 8.04439615341412e-06, 0, 1.36172897349839e-05, 3.35698487567066e-06, 7.22722234232203e-06, -2.59130730996414e-06, 1.64408898090293e-05, -7.53461577699789e-07, 1.2529138546912e-05, 1.21299757567901e-05, 1.50845540524503e-05, 1.58300498709126e-05, -2.07603379920128e-06, 7.30112176035014e-06, 1.6316851702346e-05, 4.24552890333965e-05, 9.8502302934688e-06, 3.93579190489177e-05, 3.52674731981582e-05, 1345210.0313124, 1.85104991190217e-05, 2.75997600989871e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.11212403517768e-08, 0, 1.12700131196663e-08, 30.3750000249913, 0, 0, 0, 0, 0, 0, -7.51310634517286e-08, 0, -5.65486543448064e-08, 0, 0, -8.47183509566414e-08, -501.187500154674, -273.37500028193, -410.062500207562, 0, 0, 0, 0, 0, 0, 2.97541325415987e-07, 0, 1.8214623951794e-07, 0, 0, 0, 3.53734974606551e-07, 0, 4.26730053632799e-07, 2460.37500046364, 1.44376435038056e-06, 1230.1875006731, 4510.68750187538, 6766.03125119899, 3690.56250167875, 0, 0, 0, 0, 0, 0, -6.21886857186074e-07, 0, -3.40748029697476e-07, 0, 0, 0, -6.50910961921591e-07, 0, -9.30233403504327e-07, -3690.56250058272, -4.43773440329675e-06, -1.8137872330361e-06, 0, 0, -1.24722363863084e-06, -22143.3750039513, -33215.0625032639, -7.3644167882688e-06, -8.41846216618408e-06, -20298.0937549164, -11071.6875042965, -60894.281261306, 0, 0, 0, 0, 0, 0, 8.53654955192789e-07, 0, 3.52855311618567e-07, 0, 0, 0, 1.99077235537506e-06, 0, 1.42674730322025e-06, 5.35461241897165e-07, 6.05571514109066e-06, 3.29284018033672e-06, 0, 0, 0, 1.67737376769975e-06, 0, 2.33312782246987e-06, 33215.0625050294, 49822.5937535068, 1.47012268172538e-05, 1.43042176179323e-05, 6.86116480394755e-06, 2.65329634660832e-06, 1.53289551492018e-05, 99645.1875059095, 8.23446177332007e-06, 298935.562515978, 3.10876401080667e-05, 182682.843760543, 0, 0, 0, 0, 0, 0, -4.69999539352444e-07, 0, -1.36481400472446e-07, 0, 0, 0, -1.35020097424004e-06, 0, -1.21338735538338e-06, -6.84640069365093e-08, -3.84052349287724e-06, -2.19725358115759e-06, 0, 0, 0, -1.69309764050092e-06, 0, -2.05557899585843e-06, -2.1579435621964e-06, -2.25305781906386e-06, -1.22470011296062e-05, -1.27786198945962e-05, -7.02760263420549e-06, -5.18961405777757e-06, 0, 0, -3.18376829165682e-06, -1.01421162264689e-05, -149467.781257654, -1.82807507848202e-05, -1.435826336906e-05, -8.79181450488192e-06, -2.98687302504526e-06, -448403.343765307, -3.7808717349954e-05, -1.16044838631993e-05, -3.10448944982863e-05, -896806.687523108, -2.13152817383986e-05, 0, 0, 0, 0, 0, 0, 1.488422846149e-07, 0, 3.1222489370461e-08, 0, 0, 0, 3.24402557542874e-07, 0, 3.1081815604776e-07, -6.58078330803601e-08, 1.10777519866661e-06, 6.97650075964816e-07, 0, 0, 0, 4.67046801278125e-07, 0, 5.92761912191144e-07, 4.1084020323289e-07, 2.05974701999154e-07, 4.19421093537834e-06, 4.65217291482611e-06, 2.15761164513153e-06, 1.82648916839646e-06, 0, 0, 0, 7.53339103664367e-07, 0, 1.04772345653724e-06, 2.73286119436616e-06, 1.68330871503574e-06, 7.28702074674195e-06, 7.06790410616971e-06, 4.03848824061836e-06, 2.08100759617719e-06, 3.84437992663024e-06, 1.67924725368333e-05, 4.21327176914182e-06, 6.68462079359131e-06, 3.38043141380319e-06, 1.89768970172947e-06, 1.93053908080455e-05, 1345210.03125769, 2.40832252306473e-05, 1.71692327302149e-05, 6.26009840520724e-06, 5.8295878636046e-06, 2.05238785503744e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.77102454607079e-08, 0, -5.96133117415383e-08, 27.000000025625, 0, 0, 0, 0, 0, 0, 2.08575851755096e-07, 0, 2.15453892226218e-07, 0, 0, 1.10054451791523e-06, -364.500000010953, -364.500000058925, -364.499999880154, 0, 0, 0, 0, 0, 0, -1.16711930627284e-06, 0, -1.46685096022117e-06, 0, 0, 0, -4.20576213111847e-06, 0, -3.28381449530995e-06, 1093.50000103006, 1093.50000015408, 1093.50000001338, 4920.74999996202, 4920.75000155564, 4920.75000002355, 0, 0, 0, 0, 0, 0, 2.2808394797741e-06, 0, 1.22852830896017e-06, 0, 0, 0, -6.38003551549189e-07, 0, 7.91653751845861e-06, 1.56215680371132e-06, -1.3210481107405e-06, 4.6461888914128e-06, 0, 0, 5.00280223026999e-06, -14762.2500050289, -14762.2499940537, -14762.2499993834, -14762.2500047373, -14762.2499910843, -14762.2499979923, -66430.1249880296, 0, 0, 0, 0, 0, 0, -5.30044435293595e-07, 0, -2.08059273188066e-06, 0, 0, 0, -7.30918042571746e-06, 0, -1.37198624000174e-05, 6.55880652520546e-08, 5.31483890177003e-06, -1.28752274216564e-05, 0, 0, 0, -6.07542828855138e-06, 0, -1.15955055962783e-05, 5.67052356005728e-06, -2.66763493715436e-06, -5.00558827197981e-06, 2.00817886443954e-05, -4.94986743778338e-07, -4.97772785488159e-06, 44286.7499933833, 44286.7500017563, 44286.7499443221, 199290.375046148, 199290.374943919, 199290.374983076, 0, 0, 0, 0, 0, 0, 7.14851868711774e-07, 0, 1.26793919064702e-06, 0, 0, 0, 6.33197629599747e-06, 0, 3.05280520353722e-06, 1.27908335748631e-06, 4.11057237269956e-06, 5.42117282702823e-07, 0, 0, 0, -1.13150440641562e-05, 0, 1.82346429907836e-05, 8.25248771463462e-06, -1.73616832550394e-06, -1.11437024990022e-05, -4.75484451809585e-06, 1.17961005993854e-05, -1.31677618011877e-05, 0, 0, 1.99517733646038e-05, 1.53998455510399e-05, 2.05558800753501e-05, 2.16494014464552e-05, 2.31780096645774e-05, 5.50317532135398e-05, -9.31095139422441e-06, -1.98282588488017e-05, -4.84251196389824e-05, 5.00948873037356e-05, -597871.125050964, -597871.125004959, -597871.125065724, 0, 0, 0, 0, 0, 0, -2.05274682541635e-07, 0, -1.87513666641521e-07, 0, 0, 0, -2.35687520143799e-06, 0, -9.532325625084e-07, -2.49075030984849e-07, -2.96016931668565e-07, 1.16433326456302e-07, 0, 0, 0, -1.60499219499983e-06, 0, -4.28144959756863e-06, -4.88311852181889e-07, 1.18900134220207e-06, 5.18041238927113e-06, -1.94419277317064e-06, -7.05936535240344e-06, -1.21833603970507e-05, 0, 0, 0, -5.86067090675261e-06, 0, -1.13257238907105e-05, -2.32634482770119e-06, -5.01742894924655e-06, 1.23001419787207e-05, 5.82839925694718e-06, -4.04324303137887e-06, 1.02823512703823e-05, -2.43128573210448e-06, 1.51393506511716e-05, -1.08098418341085e-05, 8.54873681644567e-06, -9.6550275453874e-06, 2.17561997119983e-05, -2.85550701645334e-05, 1.25613333890165e-05, -3.111079909301e-06, 2.66531323572951e-05, -6.71593927763941e-05, 2.98292199064919e-05, 1793613.37502386}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 2916.00000013472, 0, 0, 0, 0, 0, 0, 0, -34877.4428591879, 0, -34877.4428588543, -34877.4428639803, 0, 0, 0, 0, 0, 0, 174358.575016807, 0, 174358.575007324, 0, 0, 348717.150017213, 174358.575043751, 348717.150081957, 348717.150081705, 0, 0, 0, 0, 0, 0, -472966.087564194, 0, -472966.087519464, 0, 0, 0, -1418898.26263557, 0, -1418898.2625147, -472966.087634085, -1418898.26286968, -1418898.262865, -1418898.2629631, -1418898.26289375, -2837796.5258055, 0, 0, 0, 0, 0, 0, 752874.750128666, 0, 752874.75004248, 0, 0, 0, 3011499.00044982, 0, 3011498.99988572, 752874.750225478, 3011499.00082947, 3011499.00087168, 0, 0, 4517248.50023846, 3011499.00113816, 3011499.00091749, 4517248.50161213, 9034497.00296962, 4517248.50122337, 9034497.00262386, 9034497.00356148, 0, 0, 0, 0, 0, 0, -704159.325142204, 0, -704159.325062373, 0, 0, 0, -3520796.62569115, 0, -3520796.62473238, -704159.325212442, -3520796.62600165, -3520796.62620483, 0, 0, 0, -7041593.25084089, 0, -7041593.2498768, -3520796.62641331, -3520796.62612514, -7041593.2525752, -14083186.5050541, -7041593.25192563, -14083186.5039776, -7041593.25283411, -7041593.25193448, -21124779.7572487, -14083186.5066393, -21124779.7598491, -21124779.7581647, 0, 0, 0, 0, 0, 0, 358722.675083121, 0, 358722.675048271, 0, 0, 0, 2152336.05049495, 0, 2152336.0497922, 358722.675105586, 2152336.05062783, 2152336.05089038, 0, 0, 0, 5380840.12602024, 0, 5380840.12453112, 2152336.05087766, 2152336.05068045, 5380840.12698044, 10761680.2540095, 5380840.12665104, 10761680.2530683, 0, 0, 7174453.50031496, 5380840.12722571, 5380840.12652337, 7174453.50286399, 21523360.5085053, 7174453.50175601, 21523360.5068634, 10761680.2555585, 21523360.5107359, 21523360.507568, 21523360.5120947, 21523360.5098514, 32285040.7656404, 0, 0, 0, 0, 0, 0, -76869.1446631316, 0, -76869.1446574368, 0, 0, 0, -538084.012634364, 0, -538084.012445902, -76869.1446643967, -538084.012662233, -538084.012766619, 0, 0, 0, -1614252.03790484, 0, -1614252.03727154, -538084.012719645, -538084.012658809, -1614252.03810173, -3228504.07619767, -1614252.03810105, -3228504.07599318, 0, 0, 0, -2690420.06283117, 0, -2690420.06237884, -1614252.03815459, -1614252.03796198, -2690420.06355098, -8071260.19102096, -2690420.0631586, -8071260.18983661, -3228504.07670854, -8071260.19156309, -8071260.19015626, -2690420.06355885, -2690420.06313497, -10761680.2540247, -8071260.1924917, -8071260.19161419, -10761680.2563926, -16142520.383915, -10761680.254308, -16142520.3821946, -16142520.3845606}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 104.142857137733, 0, 0, 0, 0, 0, 0, 0, -2400.49285710099, 0, -104.142857056816, -104.142857136972, 0, 0, 0, 0, 0, 0, 21323.2499998085, 0, -6.73045253310854e-07, 0, 0, 2296.34999960458, 7.90502015426418e-08, 2296.34999990026, -1.07115226247953e-07, 0, 0, 0, 0, 0, 0, -96528.7124992435, 0, 2.67150213589884e-06, 0, 0, 0, -19026.8999998725, 0, 3.0565354399227e-06, -2.49213727505821e-07, -19026.8999997043, 8.72831268709672e-07, -2.86388804669395e-07, -9.35159634580593e-07, 1.29574482525738e-06, 0, 0, 0, 0, 0, 0, 243577.12499813, 0, -5.71889482259931e-06, 0, 0, 0, 77501.8125007088, 0, -1.13802751262348e-05, 5.67514488419427e-08, 77501.8124991931, -3.18293559379686e-06, 0, 0, -8.88968524591775e-07, 1.2132059437886e-06, 3.50668418474272e-06, 1.33301758813998e-06, -3.25846500154e-06, 4.20730934097198e-06, -6.94609167258069e-06, 2.7754464384103e-06, 0, 0, 0, 0, 0, 0, -345436.649997417, 0, 6.50479468802333e-06, 0, 0, 0, -166075.312500364, 0, 2.10172875117263e-05, 3.6127134326445e-07, -166075.312499221, 5.85809992211485e-06, 0, 0, 0, -6.694236717279e-06, 0, 8.50858928284836e-06, 6.39868558448013e-07, -3.57645855299715e-06, -2.16691174198215e-06, 4.45191470369484e-06, -6.30151866846509e-06, 1.88694405650506e-05, -3.7282674867297e-06, -1.27701412539061e-05, 1.21179025108374e-05, -1.06588132802287e-05, -6.06821838660365e-06, -8.04052243528158e-06, 0, 0, 0, 0, 0, 0, 256230.482141018, 0, -3.68534609575978e-06, 0, 0, 0, 179361.337499236, 0, -1.77052879689639e-05, -2.93648125245178e-07, 179361.33749952, -4.58719921135446e-06, 0, 0, 0, 7.14893715974538e-06, 0, -1.5766793507069e-05, -1.16223038652866e-06, 9.8877991372207e-07, 1.39006616351218e-06, -5.60496644184324e-06, 3.98864666800915e-06, -2.36228377897715e-05, 0, 0, 6.13813648486325e-06, 4.76023380880087e-07, 1.15342569720667e-05, 6.30043905568679e-06, -3.62230922460088e-06, 1.52887474264375e-05, -2.49550866720893e-05, 5.83247528511043e-06, 4.47475976552659e-06, 1.81347890171181e-06, 1.39615938968921e-05, 2.04609586658392e-05, 9.31044672036633e-06, 0, 0, 0, 0, 0, 0, -76869.1446423278, 0, 8.56268134839379e-07, 0, 0, 0, -76869.144642371, 0, 5.6169751888459e-06, 9.22406671504767e-08, -76869.1446428429, 1.43261397424107e-06, 0, 0, 0, -1.8449477877368e-06, 0, 8.43046112205976e-06, 6.30293222745101e-07, 1.6791420503968e-07, -5.93630975612386e-07, 2.35781242311156e-06, -9.05532186677321e-07, 9.1000444524198e-06, 0, 0, 0, -5.39919867703066e-06, 0, 2.26579990989527e-07, 6.17152387445383e-07, -3.18890088199393e-06, -2.12618214526289e-06, 5.50534126619232e-07, -5.77795789269122e-06, 1.589147804278e-05, -7.76575659080928e-08, 1.07499521715265e-06, 1.00706473484275e-06, -2.1654681856286e-06, -7.89255854291571e-06, 4.20706803218456e-06, -5.31658766377154e-06, -8.99411212194157e-06, -6.13663936617071e-06, -6.56898043173741e-06, -1.30349453188645e-05, -4.88183414995685e-07, -1.31925881818671e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 104.142857144282, 0, 0, 0, 0, 0, 0, 0, -104.142857130848, 0, -2400.49285718665, -104.142857145987, 0, 0, 0, 0, 0, 0, -1.37635516864621e-07, 0, 21323.2500003948, 0, 0, 2296.35000001206, -1.699160580634e-07, -1.44359497571976e-07, 2296.35000031361, 0, 0, 0, 0, 0, 0, 4.60787032915594e-07, 0, -96528.7125016195, 0, 0, 0, 1.04802058508923e-06, 0, -19026.9000009052, 8.83424434316612e-07, 7.92652515375802e-07, -19026.9000028953, 1.87401616008034e-06, 1.06353648243464e-07, 4.05541033997267e-07, 0, 0, 0, 0, 0, 0, -7.45916358406938e-07, 0, 243577.125003441, 0, 0, 0, -3.37856365256269e-06, 0, 77501.8125046825, -2.0078710629892e-06, -1.1201608868119e-06, 77501.8125103025, 0, 0, -2.58055032192148e-06, -6.75895993519382e-06, -2.34253725580975e-06, -6.76381136026566e-06, -7.05245741308315e-06, 4.06976301434415e-06, 5.10910225649276e-06, -5.84800424795901e-06, 0, 0, 0, 0, 0, 0, 6.69483183151111e-07, 0, -345436.650003945, 0, 0, 0, 4.13204771179158e-06, 0, -166075.312509457, 2.34175341749094e-06, -2.70328666723777e-07, -166075.312517475, 0, 0, 0, 9.89819913293151e-06, 0, -2.63307115671851e-07, 1.12866456196223e-05, 4.71750771272392e-06, 9.28519014330637e-06, 1.49406100787556e-05, -1.41373111038231e-05, -2.07417590288646e-05, 1.82276043153983e-05, 6.67537704853677e-07, 1.019926698602e-05, 1.27044665212985e-05, 2.71674967071311e-05, -1.73508210855549e-06, 0, 0, 0, 0, 0, 0, -3.37453646004824e-07, 0, 256230.482145134, 0, 0, 0, -2.16579086712046e-06, 0, 179361.337508161, -1.37691076231063e-06, 1.36029137794174e-06, 179361.337514148, 0, 0, 0, -9.76976295571389e-06, 0, 5.55020825936894e-06, -8.69267039623436e-06, -3.73149934177197e-06, -4.25823772681534e-06, -1.09338057573456e-05, 1.82211013575438e-05, 2.76153499538632e-05, 0, 0, -8.61116266894832e-06, -1.99540367015904e-05, -4.97906717862851e-06, -1.8612255684409e-05, -2.43709195836035e-05, 4.77341261131859e-06, 3.12620113706373e-06, -1.19838554867868e-05, -3.75319865052286e-05, 1.8320382130291e-05, -3.22503342724658e-05, -1.08100748059143e-06, -2.52349791756266e-05, 0, 0, 0, 0, 0, 0, 7.72323945208195e-08, 0, -76869.1446433837, 0, 0, 0, 3.94753697689987e-07, 0, -76869.1446454236, 3.30913512766961e-07, -6.57700742560508e-07, -76869.144647261, 0, 0, 0, 3.03457794518701e-06, 0, -3.53331016888608e-06, 2.66840939675116e-06, 1.11409645391911e-06, 1.52207981777847e-08, 2.40934429883966e-06, -7.90025517643067e-06, -1.20557518957989e-05, 0, 0, 0, 5.58917750063575e-06, 0, 1.45697611703418e-06, 7.8141408767542e-06, 2.41599641660507e-06, 5.9442188997146e-06, 1.163657376686e-05, -4.34828382866068e-06, -8.09842158566098e-06, 4.4790182862877e-06, 1.51643250854729e-05, -1.37366651239528e-05, 1.06034760753069e-05, 1.47476496032641e-06, 8.52041989776903e-06, 1.34185259609982e-05, 1.51235979393822e-06, 2.33634225510877e-05, 2.50219020022887e-05, -6.93025562856913e-06, 3.81441308844449e-06, 1.08432515559765e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -3402.00000035215, 0, 0, 0, 0, 0, 0, 0, 61102.3500061865, 0, 30484.3500037466, 30484.350010487, 0, 0, 0, 0, 0, 0, -386324.437548243, 0, -111965.287514546, 0, 0, -498289.725049824, -111965.287567021, -498289.72513779, -223930.575110751, 0, 0, 0, 0, 0, 0, 1223585.49392702, 0, 215897.906283823, 0, 0, 0, 2663068.89411476, 0, 1655381.3063452, 215897.906457237, 2663068.8944285, 647693.719203811, 1655381.30699393, 647693.719292026, 3310762.61370176, 0, 0, 0, 0, 0, 0, -2173741.31284869, 0, -230660.156317065, 0, 0, 0, -6751884.09487682, 0, -2865721.78117271, -230660.156602897, -6751884.09538751, -922640.626020874, 0, 0, -7213204.40707877, -2865721.78305362, -922640.626270986, -7213204.40900629, -14426408.8173225, -1383960.9390576, -8597165.34734672, -8597165.34898381, 0, 0, 0, 0, 0, 0, 2205480.1503843, 0, 129538.743849012, 0, 0, 0, 8951459.3454141, 0, 2723635.12462737, 129538.744090219, 8951459.3458651, 647693.720135796, 0, 0, 0, 13751035.8773019, 0, 9599153.06295243, 2723635.12724019, 647693.720291975, 13751035.8796653, 27502071.7586737, 1295387.4397661, 10894540.5049509, 9599153.06723421, 1295387.44003658, 28797459.1985846, 10894540.5096333, 28797459.2028714, 16341810.7613151, 0, 0, 0, 0, 0, 0, -1195742.25022493, 0, -29893.5563301501, 0, 0, 0, -6008604.80742831, 0, -1345210.03091034, -29893.5564259698, -6008604.8076637, -179361.338542531, 0, 0, 0, -12106890.2837308, 0, -6277646.81216555, -1345210.03265917, -179361.338407989, -12106890.2850391, -24213780.5696769, -448403.345607624, -6726050.15970344, 0, 0, -12256358.06402, -6277646.81621263, -448403.345765904, -12256358.0675472, -36769074.2013574, -597871.127151879, -25110587.2598621, -6726050.16418318, -36769074.2049916, -13452100.3222175, -25110587.268457, -13452100.3261605, -37665880.8984158, 0, 0, 0, 0, 0, 0, 269042.006304691, 0, 2.52599414487349e-05, 0, 0, 0, 1614252.03782376, 0, 269042.006153406, 3.80103607909531e-05, 1614252.03788732, 0.000322393105177262, 0, 0, 0, 4035630.09467255, 0, 1614252.03722266, 269042.006610561, 0.000204516178022914, 4035630.09496327, 8071260.18973655, 0.000677787069806006, 1614252.03854691, 0, 0, 0, 5380840.12599207, 0, 4035630.09393245, 1614252.03860848, 0.000602111403666448, 5380840.12694408, 16142520.380933, 0.00075591551431182, 8071260.19057659, 1614252.03989549, 16142520.3818514, 4035630.09686954, 4035630.09561831, 0.000814365778151079, 16142520.381356, 8071260.19498028, 4035630.09939489, 16142520.3851502, 24213780.5764957, 5380840.1306508, 16142520.385316, 16142520.3891184}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 3402.0000003502, 0, 0, 0, 0, 0, 0, 0, -68756.8500067393, 0, -22829.8500031282, -22829.8500088177, 0, 0, 0, 0, 0, 0, 506582.100053426, 0, 60598.1250077045, 0, 0, 429399.225051079, 60598.1250556944, 429399.22512146, 121196.250085068, 0, 0, 0, 0, 0, 0, -1822235.73769407, 0, -79552.1250046451, 0, 0, 0, -2799414.67540281, 0, -1056731.06257039, -79552.1251718779, -2799414.67563226, -238656.375321567, -1056731.06312726, -238656.375423126, -2113462.12597185, 0, 0, 0, 0, 0, 0, 3579845.62537619, 0, 51667.8750047496, 0, 0, 0, 8336980.68875061, 0, 1280625.18729314, 51667.8752938392, 8336980.68909061, 206671.500679226, 0, 0, 5986092.37589346, 1280625.18899479, 206671.500991243, 5986092.37740915, 11972184.7541814, 310007.251173106, 3841875.56513874, 3841875.56668021, 0, 0, 0, 0, 0, 0, -3932663.40040781, 0, -13286.0250295403, 0, 0, 0, -12522078.5643309, 0, -763946.43686975, -13286.0252878259, -12522078.5646224, -66430.1259066019, 0, 0, 0, -13983541.3150596, 0, -6144786.56292612, -763946.439342789, -66430.1261822408, -13983541.3167149, -27967082.6328838, -132860.251643193, -3055785.75313399, -6144786.56651626, -132860.251965338, -18434359.6965005, -3055785.75762203, -18434359.7002861, -4583678.63363906, 0, 0, 0, 0, 0, 0, 2271910.27523476, 0, 4.18464110989017e-05, 0, 0, 0, 9267002.43878608, 0, 179361.336939356, 0.000154021990399108, 9267002.43895462, 0.000701846465780196, 0, 0, 0, 14348907.0027218, 0, 2989355.62449963, 179361.338665674, 0.000673674311231501, 14348907.0035262, 28697814.0067428, 0.00132757196900553, 896806.689286193, 0, 0, 10163809.1267462, 2989355.62810363, 0.00155763127713807, 10163809.1293994, 30491427.3870233, 0.00166131668371001, 11957422.5072371, 896806.693672432, 30491427.3899936, 1793613.38187752, 11957422.5150325, 1793613.38555362, 17936133.7688485, 0, 0, 0, 0, 0, 0, -538084.01255607, 0, -1.68331871761658e-05, 0, 0, 0, -2690420.0628525, 0, 0.000165377061440187, -3.47383727047258e-05, -2690420.06290629, -0.000225858454335603, 0, 0, 0, -5380840.12599286, 0, -538084.012117224, -0.000303075667155078, -0.000145846916971624, -5380840.12616054, -10761680.2521616, -0.000492370155208818, -0.000463768380842244, 0, 0, 0, -5380840.12612587, 0, -2690420.06272934, -538084.013425354, -0.000454238158777548, -5380840.12673224, -16142520.3803445, -0.000583011052502901, -2690420.06438984, -0.00182274184931697, -16142520.3810088, -0.00199747574299875, -2690420.06409666, -0.000646416316354223, -10761680.2552216, -2690420.06848864, -0.00433005604458585, -10761680.2584571, -16142520.3867058, -0.00422065547354127, -5380840.1329088, -5380840.13632681}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -2551.50000015445, 0, 0, 0, 0, 0, 0, 0, 54629.4375035248, 0, 12529.6875009675, 12529.6875044428, 0, 0, 0, 0, 0, 0, -436306.500029118, 0, -22895.1562474892, 0, 0, -252530.156276177, -22895.1562778292, -252530.156313618, -45790.3125405486, 0, 0, 0, 0, 0, 0, 1721134.82822971, 0, 18452.8124741881, 0, 0, 0, 1825803.28148382, 0, 433128.51563722, 18452.8125839586, 1825803.28159319, 55358.4376378789, 433128.515940115, 55358.4377082848, 866257.031745796, 0, 0, 0, 0, 0, 0, -3686871.93769803, 0, -5535.84369039566, 0, 0, 0, -6166929.93823341, 0, -326614.781020356, -5535.84389066945, -6166929.93838008, -22143.3752511004, 0, 0, -2801136.93801174, -326614.781974075, -22143.3754805566, -2801136.93873963, -5602273.87725637, -33215.0630683467, -979844.344999748, -979844.345854134, 0, 0, 0, 0, 0, 0, 4367780.71895837, 0, -5.62614624633513e-05, 0, 0, 0, 10412922.0948121, 0, 91341.4213035499, 0.000137154744452315, 10412922.0949383, 0.000298582378618705, 0, 0, 0, 7813843.45466827, 0, 1860043.50017536, 91341.4227352977, 0.000551952720791838, 7813843.45533439, 15627686.9106322, 0.000751688435622758, 365365.688728885, 1860043.50196844, 0.000972619485400741, 5580130.50471738, 365365.691237065, 5580130.50663948, 548048.535563615, 0, 0, 0, 0, 0, 0, -2690420.06261609, 0, 1.93740471662117e-05, 0, 0, 0, -8519663.53198497, 0, 0.000501551298478173, -7.49551482893078e-05, -8519663.53206903, -0.000229029224186465, 0, 0, 0, -9416470.22037614, 0, -448403.343315978, -0.000534929559878166, -0.000296932649102371, -9416470.22062456, -18832940.4413047, -0.000568364748463426, -0.000418107728013709, 0, 0, -4035630.09482993, -448403.345198582, -0.00075425042730724, -4035630.09594821, -12106890.287888, -0.000816324785832231, -1793613.37848537, -0.00290329469830539, -12106890.289189, -0.00325063727448033, -1793613.38252765, -0.00523950199428935, -2690420.07230576, 0, 0, 0, 0, 0, 0, 672605.015651748, 0, -8.18200685155275e-07, 0, 0, 0, 2690420.06269902, 0, -0.000152288569300416, 1.73004688896862e-05, 2690420.06272957, 7.58655053516618e-05, 0, 0, 0, 4035630.09433142, 0, -0.000302055815079104, 0.000139248734182279, 6.00836937552882e-05, 4035630.09437536, 8071260.18873223, 0.000204583254347533, 3.17234037844777e-05, 0, 0, 0, 2690420.06319977, 0, 672605.015765132, 0.000416455537354242, 0.000211647041182005, 2690420.06337327, 8071260.19051786, 0.000275920193398162, 0.000719843760730613, 0.000814242582894899, 8071260.19071513, 0.00084908337134351, 672605.01638004, 0.000319317706486641, 2690420.06530606, 0.00288883226103613, 0.00208681482259413, 2690420.06680359, 4035630.10001484, 0.00209738629781419, 0.00403278488652355, 0.00579156477814922}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 1360.80000000038, 0, 0, 0, 0, 0, 0, 0, -30156.300000657, 0, -4641.29999950282, -4641.30000122749, 0, 0, 0, 0, 0, 0, 253017.675007021, 0, 5248.79999319307, 0, 0, 97521.975003305, 5248.80000833976, 97521.9750186973, 10497.6000100793, 0, 0, 0, 0, 0, 0, -1061979.86252574, 0, -1968.29996882873, 0, 0, 0, -750988.462563109, 0, -104319.899971817, -1968.30002461959, -750988.462604982, -5904.90002489924, -104319.900094243, -5904.90006248772, -208639.800142131, 0, 0, 0, 0, 0, 0, 2438723.70004676, 0, -6.70593622422814e-05, 0, 0, 0, 2760540.75021514, 0, 36905.6248230531, 3.90114043437525e-05, 2760540.7502727, 1.66267119789054e-05, 0, 0, 730731.375115569, 36905.6252100348, 0.000146610248260165, 730731.375378564, 1461462.7506952, 0.000168326702834409, 110716.875314484, 110716.875645705, 0, 0, 0, 0, 0, 0, -3095643.82504596, 0, 7.29155261502317e-05, 0, 0, 0, -5154977.70031288, 0, 0.00035785925789672, -3.6399875809929e-05, -5154977.70037101, 1.29738517589308e-05, 0, 0, 0, -2291839.31296507, 0, -232505.437429558, -0.000230749911891816, -0.000158986943367183, -2291839.31318269, -4583678.62638789, -0.000195879752729251, -0.000171726638574057, -232505.438077555, -0.000315511266702968, -697516.313907, -0.00113510830972764, -697516.314579461, -0.00134205605488564, 0, 0, 0, 0, 0, 0, 2032761.82502338, 0, -3.8182726527776e-05, 0, 0, 0, 4663394.7752126, 0, -0.000301761117310772, 2.0108896138578e-05, 4663394.77525589, -1.39555664240296e-05, 0, 0, 0, 3228504.07550552, 0, -0.000299392056799751, 0.00013566436798832, 7.57409862464639e-05, 3228504.07558254, 6457008.15121107, 0.000117184547785678, -0.000134593675744667, 0, 0, 597871.125306018, 0.000390439822655321, 0.00024181554259111, 597871.125664894, 1793613.37711953, 0.000263819219201228, 0.000893308957422109, 0.000829809379170852, 1793613.37750818, 0.000952276553496746, 0.00229696598581763, 0.00166212746216954, 0.00312795447125773, 0, 0, 0, 0, 0, 0, -538084.012504781, 0, 7.90914605947877e-06, 0, 0, 0, -1614252.03755642, 0, 9.24615084397717e-05, -4.69969127290752e-06, -1614252.03757197, 3.21989911639452e-06, 0, 0, 0, -1614252.03767821, 0, 0.000169744918890144, -3.40726452813253e-05, -1.26110398146671e-05, -1614252.037696, -3228504.07539232, -3.48444822989667e-05, 9.87568971439909e-05, 0, 0, 0, -538084.012725518, 0, -1.34898771124139e-05, -0.000102183371568491, -6.39400972743456e-05, -538084.012765357, -1614252.03848358, -8.07677206086246e-05, -8.17205200901527e-05, -0.00021145812882425, -1614252.03851182, -0.000212060593046955, -0.000212806577103826, -0.000108150628120285, -0.000867941882353185, -0.000841437617656255, -0.000641351898238774, -0.00132327983790304, -0.00203685828732564, -0.00069219128247268, -0.00127476423681288, -0.00184677677141085}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -485.999999980251, 0, 0, 0, 0, 0, 0, 0, 11020.0499999168, 0, 1032.74999959535, 1032.75000013553, 0, 0, 0, 0, 0, 0, -95581.0125001986, 0, -546.7499965033, 0, 0, -22316.5124985201, -546.75000136599, -22316.5125022772, -1093.50000069625, 0, 0, 0, 0, 0, 0, 419042.868750918, 0, -1.41309137440154e-05, 0, 0, 0, 179320.331255624, 0, 11235.7124827085, 4.10943872040934e-06, 179320.33126407, -1.94366495671811e-06, 11235.7125140748, 1.06672127983698e-05, 22471.4250148879, 0, 0, 0, 0, 0, 0, -1014904.68750054, 0, 2.97938478341181e-05, 0, 0, 0, -699361.593776979, 0, 6.96119420272256e-05, -5.71555813578297e-06, -699361.59378726, 1.6391707605646e-05, 0, 0, -83037.6562506546, -3.24225083429879e-05, -2.88455777445841e-05, -83037.6563077738, -166075.312589184, -3.07111777247569e-05, -1.31090478215137e-05, -9.88959326399789e-05, 0, 0, 0, 0, 0, 0, 1368460.57499859, 0, -3.33061875051611e-05, 0, 0, 0, 1411640.15628957, 0, -0.000127582865512475, 4.50094838214207e-06, 1411640.15630184, -3.40866384879119e-05, 0, 0, 0, 282328.031316382, 0, -5.99109368525099e-05, 2.98014813948575e-05, 3.0085371382623e-05, 282328.031354479, 564656.062690197, 3.26564189379088e-05, -5.96593960555035e-05, 8.99763498251855e-05, 7.42441520891304e-05, 0.000153353630814951, 0.000188183759110402, 0.000318130943951915, 0.000207356250719458, 0, 0, 0, 0, 0, 0, -956593.799997862, 0, 1.85178502820564e-05, 0, 0, 0, -1404997.14377485, 0, 0.000105093693555701, -2.47070189473099e-06, -1404997.14378597, 2.73004080715053e-05, 0, 0, 0, -448403.34382858, 0, 0.000111015542694497, -1.49265636361365e-05, -1.1743827567309e-05, -448403.343838533, -896806.68766512, -1.27969171425599e-05, 0.000120437313914951, 0, 0, -3.68044995405429e-05, -5.25526758388815e-05, -6.08077897104924e-05, -0.000110175074771634, -0.000291501957747676, -7.03374156718367e-05, -2.58140755772541e-05, -0.000128189592149665, -0.000381796976886879, -0.000123631060354455, -0.000364415552863067, -0.000291953251036973, -0.000476995623120899, 0, 0, 0, 0, 0, 0, 269042.006249141, 0, -4.19659811238657e-06, 0, 0, 0, 538084.01250596, 0, -3.24207890892061e-05, 5.76659194083853e-07, 538084.012510394, -8.31354882607833e-06, 0, 0, 0, 269042.006276693, 0, -5.62420347515035e-05, 3.00829072362392e-06, 1.01313927991087e-06, 269042.006280145, 538084.012552752, 1.31154810494173e-06, -5.35819727032011e-05, 0, 0, 0, 3.89472224288485e-05, 0, -9.21892251132854e-06, 1.06585342297084e-05, 1.59661179654885e-05, 4.33470507211317e-05, 0.000142864602384222, 2.23064651645899e-05, -5.30115272679347e-05, 2.71850110846076e-05, 0.00014817947891041, 1.60798995808552e-05, 3.35857423041186e-05, 3.32113706993208e-05, 0.000103938770248681, 0.000129500922316319, 0.000113863663094496, 0.000207927300837836, 0.000318949557135509, 0.00013242801154545, 0.00018078304422844, 0.000305358031136698}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 60.7500000018787, 0, 0, 0, 0, 0, 0, 0, -1309.16250006965, 0, -607.499999994798, -60.7500000807462, 0, 0, 0, 0, 0, 0, 10474.8187506238, 0, 546.749999700914, 0, 0, 13030.8750004987, 6.24162031344576e-07, 1248.41250160714, 546.75000032456, 0, 0, 0, 0, 0, 0, -40596.1875021385, 0, 1.40648514174963e-06, 0, 0, 0, -103499.77500588, 0, -11235.712498847, -2.12370376764165e-06, -9226.40625913415, 1.90225005214213e-06, -7.4833287633981e-06, -3.48154197881093e-06, -11235.7125122849, 0, 0, 0, 0, 0, 0, 81192.375003625, 0, -2.65926859839848e-06, 0, 0, 0, 396735.468768855, 0, -1.02481030696383e-05, 4.29437115653845e-06, 31369.7812731264, -1.18896652075601e-05, 0, 0, 83037.6562610012, 1.52238129875596e-05, 9.94926462869524e-06, 2.98448183343827e-05, 83037.6563167153, 2.31373213052039e-06, 2.06658765202075e-05, 5.3393001286084e-05, 0, 0, 0, 0, 0, 0, -79716.1500032615, 0, 2.4014147196734e-06, 0, 0, 0, -780553.968776871, 0, 2.10970616381865e-05, -5.40348613317968e-06, -49822.5937791417, 2.21919758935451e-05, 0, 0, 0, -282328.031287851, 0, 1.11302846487963e-06, -1.68820280926294e-05, -1.27585868809413e-05, -5.28879586524982e-05, -282328.03138452, 1.49372421661841e-05, 6.4901514822952e-06, -3.79794332743156e-05, -1.47479540682878e-05, -0.00013211198317684, -9.09234680263724e-05, -0.000183170273733607, -9.96587193702125e-05, 0, 0, 0, 0, 0, 0, 29893.5562514762, 0, -8.30813070462134e-07, 0, 0, 0, 747338.906268226, 0, -1.72205425945746e-05, 3.880237682302e-06, 29893.5562682698, -1.73895205333667e-05, 0, 0, 0, 448403.343790123, 0, -1.70217085792403e-05, 9.75871042229985e-06, 7.38433286485575e-06, 4.36472067404909e-05, 448403.343868512, -2.65269876678659e-05, -3.78103429516927e-05, 0, 0, 2.34687564458939e-05, 2.02418429011835e-05, 1.53332329080109e-05, 4.19731723504927e-05, 0.000203460995893063, -4.52129607853117e-07, 7.80110958242072e-05, 6.45127443465458e-05, 0.000225708499679006, 4.49131762146237e-05, 0.000187395816893673, 0.00013514243509914, 0.000274671100046529, 0, 0, 0, 0, 0, 0, -2.56049252005138e-07, 0, 4.80770633087359e-08, 0, 0, 0, -269042.006254801, 0, 5.03911905365961e-06, -1.15573028187456e-06, -4.48826196069014e-06, 5.04662618541959e-06, 0, 0, 0, -269042.006264372, 0, 9.79874452070408e-06, -2.98459574101435e-06, -1.9646020013823e-06, -1.36217882171526e-05, -269042.006288075, 1.18120636869219e-05, 1.98416775923998e-05, 0, 0, 0, -1.62690079224461e-05, 0, -1.94475832976167e-06, -3.35538958569458e-06, -4.94431043559327e-06, -1.64495686491646e-05, -9.57889950037485e-05, 6.89978689153083e-06, -3.34867507444709e-06, -1.56365068057577e-05, -9.28898161137584e-05, 6.48708457066437e-06, -8.18186393429895e-06, -3.47337581936e-06, -8.07291438723524e-05, -6.19173482036501e-05, -5.34659849506723e-05, -0.000111109930437147, -0.000185094367557879, -4.7512351513429e-05, -0.000107074029213971, -0.000162657138481951}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 48.6000000109841, 0, 0, 0, 0, 0, 0, 0, -959.85000031987, 0, -704.700000019237, -48.6000000427948, 0, 0, 0, 0, 0, 0, 6652.12500246061, 0, 2624.3999992688, 0, 0, 13869.2250025276, 3.21954669539619e-07, 911.250001963531, 656.099999222519, 0, 0, 0, 0, 0, 0, -20503.1250082859, 0, -1968.29999548268, 0, 0, 0, -95544.5625204601, 0, -51175.8000034239, -1.64626400698225e-06, -5740.87501356, -1968.29998964668, -6.79044859471939e-06, 1.05202822544536e-06, -12301.8750101236, 0, 0, 0, 0, 0, 0, 28048.2750148681, 0, -1.06744646127369e-05, 0, 0, 0, 291554.437559861, 0, 36905.6249857429, 3.7503483061751e-06, 14762.2500372706, -3.92043878611913e-05, 0, 0, 346912.875051733, 1.25161095887282e-05, 7.12134116131417e-06, 3.44798094594156e-05, 77501.8125837382, -2.44006070600749e-05, 36905.6249936731, 3.5970527393252e-05, 0, 0, 0, 0, 0, 0, -13286.0250149768, 0, 1.19902776271315e-05, 0, 0, 0, -391937.73758241, 0, 4.25483516201459e-05, -4.58100160876035e-06, -13286.0250512304, 6.42057655535376e-05, 0, 0, 0, -1029666.93763618, 0, -232505.437536052, -1.27551045581313e-05, -1.67408308240546e-05, -7.19882663230177e-05, -199290.375185121, 8.13108026253023e-05, 0.000100486816976703, -3.69156990331361e-05, 2.37082809409103e-06, -232505.437647707, -7.99609714527453e-05, -0.00018946284066053, 3.80553508523129e-06, 0, 0, 0, 0, 0, 0, 8.05488135781616e-06, 0, -6.63145817921705e-06, 0, 0, 0, 179361.337554469, 0, -4.08280328466957e-05, 3.05305084371074e-06, 3.51526818178131e-05, -4.79466226178833e-05, 0, 0, 0, 1315316.47513626, 0, -1.43388029763025e-05, 8.28359730978668e-06, 1.24625352255042e-05, 6.76005743965061e-05, 179361.337664204, -9.37560330247033e-05, -0.000149932387736833, 0, 0, 597871.125106998, 1.5202859975728e-05, 2.49456697848396e-05, 4.59287051209047e-05, 597871.125279587, -3.80481248259119e-05, 4.26373322187118e-05, 7.33143906993817e-05, 0.00026386080298682, -0.000147972994592803, 0.000203713123181072, 6.94695067528142e-05, 0.000226078706394054, 0, 0, 0, 0, 0, 0, -1.81583402357417e-06, 0, 1.38929333649707e-06, 0, 0, 0, -1.40276137948607e-05, 0, 1.32221967331331e-05, -8.56836406554202e-07, -9.70608821731742e-06, 1.34119163259677e-05, 0, 0, 0, -538084.012547276, 0, 1.73778923307101e-05, -2.47075819074791e-06, -2.86365025262782e-06, -2.42938887475877e-05, -5.22604386524236e-05, 3.64581011493983e-05, 6.54270442183035e-05, 0, 0, 0, -538084.012561562, 0, -2.06464518781707e-05, -2.0188491558543e-06, -1.3074889714891e-05, -2.15425441160669e-05, -538084.012637043, 3.04300894239562e-05, 3.5801059041388e-05, -2.16091667690693e-05, -0.0001137584694013, 0.000118349287294272, -6.19020087618359e-06, -5.36226568583175e-06, -0.000106814884373194, -7.53339218365947e-05, -5.25548491210326e-05, -0.000126788710312223, -0.00020056865575289, 2.851518041312e-05, -2.48833351403958e-05, -0.000145654757304361}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 45.5625000171663, 0, 0, 0, 0, 0, 0, 0, -797.343750294552, 0, -797.343750218682, -45.5624999435139, 0, 0, 0, 0, 0, 0, 4442.34375202445, 0, 4442.34375099369, 0, 0, 13907.9531280523, -4.29277590612212e-07, 751.7812496127, 751.781249298069, 0, 0, 0, 0, 0, 0, -9226.40625679209, 0, -9226.40625245935, 0, 0, 0, -76989.2343933042, 0, -76989.2343854925, 1.05269637380378e-06, -3690.56250050115, -3690.56249651681, 3.08171208741005e-06, 3.48434892582691e-06, -12404.390619106, 0, 0, 0, 0, 0, 0, 5535.84376252217, 0, 5535.84375424855, 0, 0, 0, 157771.546924009, 0, 157771.54688946, -1.7015957905281e-06, 5535.84375545618, 5535.84374244555, 0, 0, 420724.125057189, -5.934659759741e-06, -2.23701039647413e-06, -5.47095537823353e-06, 60894.2812464826, -1.67267850930046e-05, 60894.2812142491, -1.87523026389635e-05, 0, 0, 0, 0, 0, 0, -1.31930371181685e-05, 0, -4.90756873284036e-06, 0, 0, 0, -91341.4219399627, 0, -91341.4218810955, 1.90852220179027e-06, -1.18474554442576e-05, 4.69220050634349e-06, 0, 0, 0, -838680.328250655, 0, -838680.328194769, 7.2111886867615e-06, -1.09792812797995e-06, 3.51193904830601e-06, -91341.4218837565, 3.36960270583159e-05, -91341.4217853825, 1.68076584264e-05, 9.01299018924902e-06, -298935.56247279, -4.9457606986988e-06, 2.27417112929723e-06, 0.000106783606194471, 0, 0, 0, 0, 0, 0, 7.54577710501461e-06, 0, 3.55741485681247e-06, 0, 0, 0, 4.20870584818408e-05, 0, -9.30051725178346e-07, -1.15272910898374e-06, 1.15222734862878e-05, 2.86090908749781e-06, 0, 0, 0, 448403.343867724, 0, 448403.343781135, -5.64640125768194e-06, -9.72557109919353e-07, 3.92805682040388e-06, 9.50459923507582e-06, -2.5719799565233e-05, -9.25227503818701e-05, 0, 0, 1569411.70323909, -2.03620861548762e-05, 7.61298191315263e-06, -2.21600896698226e-05, 448403.343774265, -1.73851839467236e-05, 448403.343673073, 2.35341473005566e-05, 2.2030377395883e-05, -0.000207064892112737, 5.18367571087497e-05, -4.56566297912531e-05, -6.99377394871226e-05, 0, 0, 0, 0, 0, 0, -1.82658690769273e-06, 0, -1.04542823098149e-06, 0, 0, 0, -1.06579187772406e-05, 0, 8.79359582411524e-07, 2.62754843771745e-07, -4.24278425957648e-06, -2.8846380461786e-06, 0, 0, 0, -3.97920987456682e-05, 0, -2.88570300861598e-06, 1.66803379287393e-06, 1.23337847427397e-06, -4.35976142054419e-06, -2.05139380742751e-06, 4.30909376932166e-06, 3.19318437043452e-05, 0, 0, 0, -672605.015683418, 0, -672605.015657661, 8.83226620306841e-06, -4.48330859420572e-06, 8.33288544831887e-06, -1.90319437929366e-05, 7.52037427482562e-06, 5.30494337997285e-05, -8.64945236766181e-06, -1.02018899489329e-05, 0.00010976807348361, 1.65215861212359e-05, -3.84717968232934e-06, -672605.015626628, -3.46379817611758e-05, -1.24925469863866e-05, -3.62091011155912e-05, -7.95494122479021e-06, 7.51945265133727e-05, 0.000104264206808954, -2.36289496298986e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 48.6000000144652, 0, 0, 0, 0, 0, 0, 0, -704.700000169842, 0, -959.850000274465, -48.6000000348286, 0, 0, 0, 0, 0, 0, 2624.40000082327, 0, 6652.12500176805, 0, 0, 13869.2250029091, 9.76820924684713e-08, 656.099999631238, 911.250001717984, 0, 0, 0, 0, 0, 0, -1968.30000223722, 0, -20503.1250054567, 0, 0, 0, -51175.8000118202, 0, -95544.5625157118, -6.19312347364695e-08, -1968.29999525704, -5740.8750114996, 6.19335846777195e-07, -6.42265094921868e-06, -12301.8750091806, 0, 0, 0, 0, 0, 0, 3.65235801965817e-06, 0, 28048.2750092466, 0, 0, 0, 36905.6250258565, 0, 291554.43753883, -9.28674491209061e-07, -1.65896735277052e-05, 14762.2500333912, 0, 0, 346912.875049917, 9.84665857389646e-07, 1.29302683629816e-05, -1.34182645714953e-05, 36905.625011925, 3.13202351397687e-05, 77501.8125590188, 4.30785291382583e-05, 0, 0, 0, 0, 0, 0, -3.55376595161383e-06, 0, -13286.0250089713, 0, 0, 0, -3.09635076553628e-05, 0, -391937.737551658, 2.46781958210815e-06, 2.54935597872552e-05, -13286.0250494593, 0, 0, 0, -232505.437581486, 0, -1029666.93758872, -1.24574765138105e-06, -1.00227017054353e-05, 4.10829746153593e-05, 1.13844345158932e-05, -6.32566617692828e-05, -199290.37512721, 1.49580869646367e-05, -4.55639631101125e-05, -232505.43760606, -0.000101778000580631, -7.807928921008e-05, -0.000148415029660611, 0, 0, 0, 0, 0, 0, 1.92572712795093e-06, 0, 4.40348755334605e-06, 0, 0, 0, 1.86613771753417e-05, 0, 179361.337534774, -2.29414605267541e-06, -1.81815615107036e-05, 3.65464543147691e-05, 0, 0, 0, 6.69568085856229e-05, 0, 1315316.47507488, -8.2885444950117e-07, 5.22392688691736e-07, -4.55071550538696e-05, -3.31196146612196e-05, 6.2021682192533e-05, 179361.337624225, 0, 0, 597871.125093693, -9.89945938375354e-06, 2.93127910987532e-05, -4.15487365475469e-05, 7.44147686913635e-05, 5.59143861733619e-05, 597871.125163035, 9.17141612262715e-05, 2.75213674933224e-05, 0.000173392876475189, 0.000159543751349121, 0.000204490867912328, 0.000230787189805017, 0, 0, 0, 0, 0, 0, -4.57445973483078e-07, 0, -8.79170397917491e-07, 0, 0, 0, -4.41387564421269e-06, 0, -9.40709219279187e-06, 7.16322740636007e-07, 4.88125444318554e-06, -1.05858556625565e-05, 0, 0, 0, -2.10952630613207e-05, 0, -538084.012524577, 9.14145752721776e-07, 2.03425351313528e-06, 1.6708592891985e-05, 1.71255218981951e-05, -2.32355769450699e-05, -4.60457974217943e-05, 0, 0, 0, -3.77511462446395e-05, 0, -538084.012536326, 3.44824010835755e-06, -5.21420827189146e-06, 2.55851460395948e-05, -1.45389799586636e-05, -2.55961647923636e-05, -538084.012578581, -2.55978054824264e-05, 1.48207784581289e-05, -7.17800543709808e-05, 1.65892693685986e-05, -1.88417590656972e-05, -6.84957682513827e-05, -8.17499839875935e-05, -9.08258374664277e-05, -5.31485771578675e-05, -9.5591952569813e-05, -9.99656791593169e-05, -0.000133530238179159, -0.00018117802696857}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 60.7500000104033, 0, 0, 0, 0, 0, 0, 0, -607.500000032066, 0, -1309.16250028911, -60.7500000628759, 0, 0, 0, 0, 0, 0, 546.749999794261, 0, 10474.8187521473, 0, 0, 13030.8750019692, 1.02844835920184e-07, 546.749999555629, 1248.41250237448, 0, 0, 0, 0, 0, 0, 1.38145219822414e-06, 0, -40596.1875077402, 0, 0, 0, -11235.7125055091, 0, -103499.775012966, -3.19831777089793e-08, 4.90427713016594e-06, -9226.40626565045, 3.82755059591794e-06, -8.67566454114916e-06, -11235.7125099342, 0, 0, 0, 0, 0, 0, -3.50298170352092e-06, 0, 81192.3750153479, 0, 0, 0, 8.79313632113948e-06, 0, 396735.468788269, -3.08440462533834e-07, -1.55346475638058e-05, 31369.7812967237, 0, 0, 83037.6562808562, -1.12546943192342e-05, 2.13689275627934e-05, -2.34666365485692e-05, 1.43010910196914e-05, 4.18801880574562e-05, 83037.6563101013, 2.85176541294111e-05, 0, 0, 0, 0, 0, 0, 4.59858946395872e-06, 0, -79716.1500168398, 0, 0, 0, -8.01481236571611e-06, 0, -780553.968808278, 9.65615185431065e-07, 2.36408160902237e-05, -49822.5938222214, 0, 0, 0, -4.11162923468437e-05, 0, -282328.031315568, 1.76228560191361e-05, -3.08825815610302e-05, 5.18871920309728e-05, -5.56067443243395e-06, -8.78392074748058e-05, -282328.031388659, 4.9820252137236e-05, -6.84212808757707e-05, -9.32364667594289e-05, -6.14168131731323e-05, -2.929405729328e-05, -0.000118831411560321, 0, 0, 0, 0, 0, 0, -3.05574916918582e-06, 0, 29893.5562596827, 0, 0, 0, 3.23148049232731e-06, 0, 747338.906294522, -1.09013241388949e-06, -1.76767730455366e-05, 29893.5563053406, 0, 0, 0, 3.10024005157379e-05, 0, 448403.343815487, -1.47859996046406e-05, 2.19585774124386e-05, -4.92032713852181e-05, -5.17973907071574e-06, 8.82646749842501e-05, 448403.343896958, 0, 0, 5.28127845887399e-05, -5.06630173823665e-05, 6.30232512537536e-05, -7.18150791824201e-05, 7.48123516410455e-05, 8.3012715297092e-05, 0.000140394189737686, 6.07890569811961e-05, -3.8132776603524e-06, 0.000154608231696325, 5.25032362728011e-05, 0.000147385792256312, 0.000130466398198067, 0, 0, 0, 0, 0, 0, 8.0800595820727e-07, 0, -2.24640717229673e-06, 0, 0, 0, -2.00241160712723e-07, 0, -269042.006263263, 4.22257119016696e-07, 5.21646504188341e-06, -1.65353317227417e-05, 0, 0, 0, -9.67950096701757e-06, 0, -269042.006275387, 4.75523412549962e-06, -5.94107838513161e-06, 1.70427727174844e-05, 4.53355266852085e-06, -3.43861561906729e-05, -269042.006308432, 0, 0, 0, -1.90050249485193e-05, 0, -2.42801353978115e-05, 1.97292071358537e-05, -2.31941286038939e-05, 3.31236531703496e-05, -2.59502849978372e-05, -3.78857906126642e-05, -7.25453548088103e-05, -2.10918634004368e-05, 1.33099148016558e-05, -7.53955913645778e-05, 3.56971869885612e-05, -3.41777470361862e-05, -5.49154232565178e-05, -2.80202043860952e-05, -7.20323570030795e-05, -3.4690888139623e-06, -5.18035740924899e-05, -7.99144225626595e-05, -8.15342659643226e-05, -9.63218600668589e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -486.000000020659, 0, 0, 0, 0, 0, 0, 0, 1032.75000001786, 0, 11020.0500005293, 1032.75000003237, 0, 0, 0, 0, 0, 0, -546.749999498612, 0, -95581.0125041655, 0, 0, -22316.5125020911, -546.74999865069, -1093.49999802127, -22316.5125038263, 0, 0, 0, 0, 0, 0, -2.19971536728468e-06, 0, 419042.868766091, 0, 0, 0, 11235.7125020688, 0, 179320.331265352, -6.74496304872289e-06, -1.36128652296672e-05, 179320.33128152, -1.93969721144461e-05, 11235.7125041634, 22471.4250037732, 0, 0, 0, 0, 0, 0, 4.47667603760529e-06, 0, -1014904.68753379, 0, 0, 0, -2.90501309633096e-07, 0, -699361.593802631, 1.49268024877441e-05, 3.34782040174244e-05, -699361.593855513, 0, 0, -83037.6562640205, 6.00074836066739e-05, 6.89420469878132e-06, 8.45256852023831e-05, 4.40469194352678e-05, -83037.6563068911, -166075.312580283, 3.27510400995711e-05, 0, 0, 0, 0, 0, 0, -5.32324409830114e-06, 0, 1368460.57503908, 0, 0, 0, 2.20106431753417e-06, 0, 1411640.15634134, -1.74241225693406e-05, -3.90336165705979e-05, 1411640.1564231, 0, 0, 0, -3.21943150843028e-06, 0, 282328.031294033, -9.21444789573004e-05, -2.17775959326494e-05, -0.000152211633288721, -0.000112064623055848, 282328.031414684, 564656.062748135, -0.000186110669094104, 3.51468160713773e-05, -2.57162735233755e-06, -6.95519337453411e-05, -0.00021958766783118, 8.47095775409341e-05, 0, 0, 0, 0, 0, 0, 3.53279392500739e-06, 0, -956593.800023364, 0, 0, 0, -3.82285636341784e-06, 0, -1404997.14382591, 1.03598134701379e-05, 2.25180326137783e-05, -1404997.1438872, 0, 0, 0, 3.93479750469583e-06, 0, -448403.343814194, 7.00729990998354e-05, 2.06814272915017e-05, 0.0001194981635603, 9.32487454840299e-05, -448403.343944891, -896806.687800594, 0, 0, 4.75731379471443e-06, 0.000182505513250085, 3.53636482398074e-06, 0.000223935173403884, 0.000108691029261093, -9.05664273948889e-05, -0.000118460688548474, 6.15644978799532e-05, 0.000339439661119483, -0.000242416904941137, 0.000246211771313057, -6.87334604441739e-05, 0.000132322448777522, 0, 0, 0, 0, 0, 0, -9.86948268448301e-07, 0, 269042.006255643, 0, 0, 0, 1.43647409129748e-06, 0, 538084.012523938, -2.48777011719322e-06, -5.24670331306574e-06, 538084.012541991, 0, 0, 0, 1.50364980214796e-06, 0, 269042.006281685, -2.11676217531845e-05, -7.05162624177965e-06, -3.39596421094832e-05, -2.61203634252563e-05, 269042.006331608, 538084.012626756, 0, 0, 0, -9.28649896784967e-06, 0, 4.29126701420479e-06, -6.76538982487238e-05, -7.68604642533358e-06, -8.98742801266564e-05, -5.50708485801279e-05, 5.95033313910155e-05, 0.00010265850175316, -2.40017196301075e-05, -0.000152994947043288, 0.000158041299170541, -0.000108711876847781, 1.04505200635961e-05, -2.0150990210461e-05, -9.1850767864714e-05, 2.63509173369657e-05, -0.000201692837353395, -0.000170418569926814, 9.89944914849182e-05, 1.79230355533014e-05, -4.3714898815153e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 1360.80000008157, 0, 0, 0, 0, 0, 0, 0, -4641.30000010437, 0, -30156.3000021156, -4641.30000026177, 0, 0, 0, 0, 0, 0, 5248.79999804983, 0, 253017.675016092, 0, 0, 97521.9750097714, 5248.79999610823, 10497.599992006, 97521.9750191678, 0, 0, 0, 0, 0, 0, -1968.29999065624, 0, -1061979.8625604, 0, 0, 0, -104319.900016612, 0, -750988.462565016, -1968.29997871358, -5904.89993425792, -750988.462645098, -5904.89993071429, -104319.900039199, -208639.800049694, 0, 0, 0, 0, 0, 0, -2.07561338287775e-05, 0, 2438723.70012466, 0, 0, 0, 36905.6250239633, 0, 2760540.75020537, -4.93553384505863e-05, -0.000186641276791607, 2760540.75046178, 0, 0, 730731.37508067, -0.00020034776787228, 36905.6250242562, -0.000352157781730159, 110716.874909046, 730731.375312909, 1461462.75047674, 110716.875036171, 0, 0, 0, 0, 0, 0, 2.65406698542617e-05, 0, -3095643.8251427, 0, 0, 0, -3.7274111180227e-05, 0, -5154977.70034297, 6.05139214594427e-05, 0.000255565873458615, -5154977.70073244, 0, 0, 0, -232505.437555799, 0, -2291839.312678, 0.00029666875416406, 3.1220144037514e-05, 0.000716996415149413, 0.000377659988289392, -2291839.31330231, -4583678.62625747, 0.000707749250697612, -232505.437784822, -697516.312854609, -5.26093692391601e-05, 0.000531253549879678, -697516.313338313, 0, 0, 0, 0, 0, 0, -1.83032094985729e-05, 0, 2032761.82508472, 0, 0, 0, 2.97171995028658e-05, 0, 4663394.77528179, -3.77971186821958e-05, -0.00017445210227061, 4663394.77556663, 0, 0, 0, 6.06226049516371e-05, 0, 3228504.07522247, -0.000225304938821903, -5.74332408434656e-05, -0.000635735492650919, -0.000385040429960559, 3228504.07589114, 6457008.15140707, 0, 0, 597871.12501584, -0.000652259200421727, 9.22793548558637e-05, -0.000952821780695846, -8.19301846893081e-05, 597871.125522105, 1793613.37588549, 4.05506912190387e-05, -0.00111699671732269, 1793613.37652061, -0.000483660774975459, 0.000868815714925087, 0.000282470513809062, 0, 0, 0, 0, 0, 0, 5.15480591175876e-06, 0, -538084.012520343, 0, 0, 0, -7.59104242465258e-06, 0, -1614252.03758882, 9.43399555881132e-06, 4.78697180352961e-05, -1614252.03766997, 0, 0, 0, -3.38201446417018e-05, 0, -1614252.03760694, 6.84404852593551e-05, 2.53863373063623e-05, 0.000206476478859946, 0.00013253414517471, -1614252.0378594, -3228504.07556834, 0, 0, 0, 6.275313596633e-06, 0, -538084.012508172, 0.00023292001173121, 4.85015706384666e-06, 0.00042321758271207, 0.000101451330967692, -538084.012802564, -1614252.03806062, 7.72079383452303e-06, 0.000586345868387481, -1614252.0383465, 0.000431039272598139, -0.000102636835330957, -0.00012863428679661, 0.000131564149395099, -0.000348741011998663, 0.000608238518708829, 0.000285055814098142, -0.000699367823993081, -0.000522977156390902, -0.000325184369576427}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -2551.50000013871, 0, 0, 0, 0, 0, 0, 0, 12529.687499842, 0, 54629.4375039866, 12529.6875009245, 0, 0, 0, 0, 0, 0, -22895.156243648, 0, -436306.50002989, 0, 0, -252530.156267629, -22895.1562463826, -45790.3124858326, -252530.156298071, 0, 0, 0, 0, 0, 0, 18452.8124729646, 0, 1721134.82823479, 0, 0, 0, 433128.515650163, 0, 1825803.28136577, 18452.8124731605, 55358.4373541344, 1825803.28158703, 55358.4373953755, 433128.515760209, 866257.03144138, 0, 0, 0, 0, 0, 0, -5535.8436919348, 0, -3686871.93772322, 0, 0, 0, -326614.781283566, 0, -6166929.93785056, -5535.8436806063, -22143.3745386027, -6166929.93852266, 0, 0, -2801136.9376191, -22143.3747238235, -326614.781439802, -33215.0618272441, -979844.343811802, -2801136.93831995, -5602273.87634056, -979844.344324426, 0, 0, 0, 0, 0, 0, -7.22782592479716e-05, 0, 4367780.71900269, 0, 0, 0, 91341.4219338681, 0, 10412922.0943285, -9.26020380802092e-05, -0.000686318183276321, 10412922.0953242, 0, 0, 0, 1860043.50007654, 0, 7813843.45332783, -0.000395079084464025, 91341.4219831734, -0.00155319877217829, 365365.686971229, 7813843.45502283, 15627686.9093829, -0.00124747763269841, 1860043.50090717, 5580130.50164917, 365365.688656046, 548048.531335855, 5580130.50304593, 0, 0, 0, 0, 0, 0, 4.83930268413542e-05, 0, -2690420.06264842, 0, 0, 0, -4.52398860880082e-05, 0, -8519663.53172795, 6.18716155912698e-05, 0.000499232113409334, -8519663.53244445, 0, 0, 0, -448403.343880485, 0, -9416470.21900214, 0.000305054080686731, 2.01436594499015e-05, 0.00150114921082613, 0.000756014701911731, -9416470.22074949, -18832940.4407623, 0, 0, -4035630.09364925, 0.00109256646274011, -448403.344200205, 0.00193511904279438, -1793613.37573645, -4035630.09510859, -12106890.2840432, -0.00102547502527906, 0.00131454518630082, -12106890.2856373, -0.000706275563067626, -1793613.37852227, -2690420.06546301, 0, 0, 0, 0, 0, 0, -1.32125916555601e-05, 0, 672605.015660225, 0, 0, 0, 8.80729766929459e-06, 0, 2690420.0626519, -1.62068829295265e-05, -0.000143372738452862, 2690420.06285318, 0, 0, 0, 8.00576232152759e-05, 0, 4035630.09388674, -9.50949696831858e-05, -3.22241610440162e-05, -0.000521914115666787, -0.000317793714556023, 4035630.09453487, 8071260.18876366, 0, 0, 0, 672605.015586767, 0, 2690420.06241776, -0.000379410402583315, 6.35551038523416e-05, -0.000939189132812731, 8.72118612652142e-05, 2690420.06321551, 8071260.18897823, 0.000286495050576617, -0.00097509135667263, 8071260.18965473, -0.000840261643497681, 672605.015951966, 2690420.06328381, 0.000509487512757701, 0.00147950114744537, -0.000507437246476015, 0.000630035742378718, 2690420.06472769, 4035630.09602441, 0.00224911056176093}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 3402.00000009905, 0, 0, 0, 0, 0, 0, 0, -22829.8499993189, 0, -68756.8500035147, -22829.8500025018, 0, 0, 0, 0, 0, 0, 60598.1249911582, 0, 506582.100025059, 0, 0, 429399.225013883, 60598.1250093313, 121196.250001872, 429399.225073486, 0, 0, 0, 0, 0, 0, -79552.1249664201, 0, -1822235.73758761, 0, 0, 0, -1056731.06252184, 0, -2799414.67507599, -79552.125018292, -238656.374894267, -2799414.67544723, -238656.375023038, -1056731.0627807, -2113462.12545756, 0, 0, 0, 0, 0, 0, 51667.8749305569, 0, 3579845.62517436, 0, 0, 0, 1280625.18755767, 0, 8336980.68768327, 51667.875011006, 206671.499545143, 8336980.68874761, 0, 0, 5986092.37503862, 206671.500092405, 1280625.18807078, 310007.249646826, 3841875.56342172, 5986092.37626156, 11972184.7523037, 3841875.56444591, 0, 0, 0, 0, 0, 0, -13286.0249148557, 0, -3932663.40019602, 0, 0, 0, -763946.437615789, 0, -12522078.5627888, -13286.0249840125, -66430.1242189314, -12522078.5643315, 0, 0, 0, -6144786.56257946, 0, -13983541.3123398, -66430.1251090403, -763946.438101738, -132860.248712532, -3055785.75064523, -13983541.3149887, -27967082.6295418, -132860.249448784, -6144786.56423335, -18434359.6917416, -3055785.75397022, -4583678.6285316, -18434359.6939348, 0, 0, 0, 0, 0, 0, -5.62322273306962e-05, 0, 2271910.27511466, 0, 0, 0, 179361.337584691, 0, 9267002.43775284, -2.44558214012746e-05, -0.000618978017044285, 9267002.4388557, 0, 0, 0, 2989355.62525298, 0, 14348906.9997833, 2.76997006641942e-05, 179361.337781213, -0.00149014346901737, 896806.687411607, 14348907.0024027, 28697814.0042262, 0, 0, 10163809.124643, -0.000472229254475197, 2989355.62617289, -0.00152533453534643, 11957422.5035054, 10163809.1270323, 30491427.3803709, 896806.691026939, 1793613.37690474, 30491427.3824344, 1793613.38039958, 11957422.5080388, 17936133.7592758, 0, 0, 0, 0, 0, 0, 1.50269568472124e-05, 0, -538084.01252705, 0, 0, 0, -1.58636431488042e-05, 0, -2690420.06258574, 8.77190018680555e-06, 0.000187278327872194, -2690420.06289555, 0, 0, 0, -538084.012649418, 0, -5380840.12494839, 8.22170989382072e-06, -3.62686085661488e-05, 0.000571751276857719, 0.000209432101020977, -5380840.12590655, -10761680.2515283, 0, 0, 0, -2690420.06243783, 0, -5380840.12470176, 0.00017931528255236, -538084.012791789, 0.000901825866144297, -2690420.06360506, -5380840.12591549, -16142520.3773278, -0.00108196345437577, -3.75677405678844e-06, -16142520.3781032, 0.000667137835414631, -2690420.06312374, -10761680.2522331, -0.00258084673333398, -2690420.06600023, -0.00166517847074762, -5380840.12897302, -10761680.2541987, -16142520.3802806, -5380840.13142746}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -3402.00000006456, 0, 0, 0, 0, 0, 0, 0, 30484.3500002758, 0, 61102.3500016565, 30484.3500055355, 0, 0, 0, 0, 0, 0, -111965.287502913, 0, -386324.437508429, 0, 0, -498289.725009481, -111965.287535073, -223930.575054459, -498289.725087502, 0, 0, 0, 0, 0, 0, 215897.906263954, 0, 1223585.4937734, 0, 0, 0, 1655381.30632022, 0, 2663068.89374562, 215897.906357102, 647693.718928457, 2663068.89417994, 647693.719088678, 1655381.30666573, 3310762.61328548, 0, 0, 0, 0, 0, 0, -230660.156278369, 0, -2173741.3125475, 0, 0, 0, -2865721.78151423, 0, -6751884.09360671, -230660.156423166, -922640.625252816, -6751884.09480755, 0, 0, -7213204.40629743, -922640.625879672, -2865721.7822307, -1383960.93842059, -8597165.34627775, -7213204.40767206, -14426408.8154271, -8597165.34734853, 0, 0, 0, 0, 0, 0, 129538.743777138, 0, 2205480.15006226, 0, 0, 0, 2723635.12543822, 0, 8951459.34346965, 129538.74389956, 647693.718884689, 8951459.34520625, 0, 0, 0, 9599153.06292835, 0, 13751035.8745403, 647693.719859007, 2723635.12619166, 1295387.43850938, 10894540.5037367, 13751035.8773207, 27502071.7547638, 1295387.43922427, 9599153.06479788, 28797459.194797, 10894540.5070634, 16341810.7589843, 28797459.1966623, 0, 0, 0, 0, 0, 0, -29893.5562624904, 0, -1195742.25004334, 0, 0, 0, -1345210.03156538, 0, -6008604.80604928, -29893.5563149051, -179361.337484802, -6008604.80730475, 0, 0, 0, -6277646.81319664, 0, -12106890.2805222, -179361.338172441, -1345210.0319397, -448403.34418893, -6726050.15878791, -12106890.2832406, -24213780.5663299, 0, 0, -12256358.0622704, -448403.345117129, -6277646.81432451, -597871.126149356, -25110587.2578159, -12256358.0646913, -36769074.1950039, -6726050.16236263, -13452100.3211307, -36769074.1964131, -13452100.3243011, -25110587.2615922, -37665880.8913021, 0, 0, 0, 0, 0, 0, 2.45758431879694e-06, 0, 269042.006262028, 0, 0, 0, 269042.00632993, 0, 1614252.0374501, 1.1052520840543e-05, -2.58323884260781e-05, 1614252.03780807, 0, 0, 0, 1614252.03781667, 0, 4035630.09345684, 0.00016022432391941, 269042.006396868, 4.69146169012907e-05, 1614252.03811877, 4035630.09446341, 8071260.18875686, 0, 0, 0, 4035630.09385817, 0, 5380840.12463285, 0.000387223584893584, 1614252.03804156, 0.000207429740577278, 8071260.1905128, 5380840.12582899, 16142520.3777183, 1614252.03939497, 4035630.0965588, 16142520.3781869, 0.000386704818296577, 4035630.0945671, 16142520.3790206, 4035630.09885938, 8071260.19253313, 5380840.13053001, 16142520.3835916, 16142520.38031, 24213780.5704413, 16142520.3855395}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 2551.50000025578, 0, 0, 0, 0, 0, 0, 0, -40085.8875048073, 0, -40085.887502878, -17122.3875073532, 0, 0, 0, 0, 0, 0, 199550.081287201, 0, 199550.081259743, 0, 0, 605771.662544164, 45448.5938007945, 244998.675098757, 244998.675084854, 0, 0, 0, 0, 0, 0, -468701.437635541, 0, -468701.437514474, 0, 0, 0, -2793017.70031174, 0, -2793017.70009787, -59664.0939147439, -997066.969207877, -997066.969080221, -588029.625571244, -588029.625456503, -3381047.32600429, 0, 0, 0, 0, 0, 0, 575727.750265518, 0, 575727.75001938, 0, 0, 0, 5984247.09469412, 0, 5984247.09371057, 38750.9065311988, 1765934.15726883, 1765934.15690666, 0, 0, 10817038.6882604, 691980.470189385, 691980.469927734, 1843435.97073621, 12660474.6602046, 1843435.96999478, 12660474.6593729, 7368208.03588865, 0, 0, 0, 0, 0, 0, -358722.675290973, 0, -358722.675031728, 0, 0, 0, -6626404.97012937, 0, -6626404.96844598, -9964.51901268923, -1444855.21996366, -1444855.21955082, 0, 0, 0, -18085601.5332871, 0, -18085601.5317072, -398580.751808368, -398580.751491363, -2192194.12808676, -20277795.662991, -2192194.1265139, -20277795.6603871, -1494677.81600518, -1494677.81491005, -37665880.884921, -7822147.22779844, -21074957.1695928, -21074957.166464, 0, 0, 0, 0, 0, 0, 89680.6689198232, 0, 89680.6687806838, 0, 0, 0, 3676907.41971125, 0, 3676907.41845316, 0.000129344331466314, 448403.344508962, 448403.34432483, 0, 0, 0, 13900503.658453, 0, 13900503.6560097, 89680.6698722386, 89680.6696233363, 896806.689805317, 14797310.3489618, 896806.688491381, 14797310.3463744, 0, 0, 20626553.8138246, 448403.346507298, 448403.345907528, 896806.690861076, 42149914.3245999, 896806.68935681, 42149914.3216131, 4035630.10146137, 22420167.2019835, 22420167.1957121, 15245713.7043207, 15245713.7008706, 43046721.0215662, 0, 0, 0, 0, 0, 0, -4.14735601680624e-05, 0, -1.09820713801171e-05, 0, 0, 0, -807126.019006775, 0, -807126.018659772, -2.67587420014339e-05, -0.000198872527812688, -0.00017783232795863, 0, 0, 0, -4035630.09457696, 0, -4035630.09353131, -0.000280164111061097, -0.000186391798941336, -0.000690456317324383, -4035630.09524758, -0.000320314932468429, -4035630.09445063, 0, 0, 0, -8071260.18837023, 0, -8071260.18763882, -0.000809813343350609, -0.000681728482986867, -0.00116606912203301, -16142520.3799944, -0.000521333759588769, -16142520.3778394, -807126.021101118, -8071260.19284131, -8071260.18973545, -0.00121291055508934, -0.000891699887659455, -24213780.5683651, -4035630.10077605, -4035630.09962427, -8071260.19631092, -24213780.5751372, -8071260.19284549, -24213780.5720077, -16142520.388673}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -291.600000037535, 0, 0, 0, 0, 0, 0, 0, 6087.15000086792, 0, 3244.05000024394, 619.650000622178, 0, 0, 0, 0, 0, 0, -46391.7375067738, 0, -5904.89999952937, 0, 0, -67058.8875069385, -328.050004023589, -12274.537511623, -6232.9500034937, 0, 0, 0, 0, 0, 0, 167920.593773158, 0, 2952.4499946405, 0, 0, 0, 503023.668806971, 0, 116621.77500781, 1.31204935854593e-05, 85498.0313165735, 5904.8999968235, 6150.93755142608, 2952.4500233485, 122772.712585971, 0, 0, 0, 0, 0, 0, -306685.743791602, 0, 1.1829239724946e-05, 0, 0, 0, -1777005.84392131, 0, -55358.4374613525, -2.45040979188372e-05, -265720.500171308, 4.68562891787207e-05, 0, 0, -808233.187631652, -0.000106858827765465, -6.63269009640724e-05, -38750.9064626242, -846984.094204626, -2.24331200468946e-05, -110716.875166229, -55358.4378422208, 0, 0, 0, 0, 0, 0, 269042.006291443, 0, -1.02096819983882e-05, 0, 0, 0, 3128858.88774123, 0, -0.000107851897396961, 2.76186355021567e-05, 368687.193974415, -9.30012965145941e-05, 0, 0, 0, 2491129.687864, 0, 348758.156324195, 0.000120040295761159, 8.61672554695298e-05, 99645.1878901164, 2590774.87591348, -6.94617476555924e-05, 1.991902687263e-05, 0.000289566844306535, 0.000104287254902835, 697516.313421726, 0.000597405803819906, 348758.15745389, 0.0006041105150276, 0, 0, 0, 0, 0, 0, -89680.6687717849, 0, 2.7434439045167e-06, 0, 0, 0, -2600739.39391335, 0, 9.70844295703749e-05, -1.7856972721408e-05, -179361.337647831, 7.10689798144734e-05, 0, 0, 0, -3407865.41287197, 0, 5.91792165273423e-05, -7.16409503150775e-05, -4.95029276213759e-05, -89680.6690840646, -3497546.08204998, 0.000127847028995332, 0.000176166390451483, 0, 0, -896806.687764038, -0.000179269059186389, -0.000116934629373573, -0.000328792341261872, -1793613.37638806, -1.67938953605938e-05, -0.000601872716627298, -0.000451773842942405, -896806.689000263, -0.000243279528577996, -0.00127397652668723, -0.000825872660929189, -0.0017496270222142, 0, 0, 0, 0, 0, 0, 4.73287019653462e-06, 0, 3.18575280815248e-07, 0, 0, 0, 807126.018793141, 0, -2.95378791329371e-05, 4.93791583473999e-06, 3.8927805561894e-05, -1.91885063687037e-05, 0, 0, 0, 1614252.03763158, 0, -5.17516150230261e-05, 1.98470296653673e-05, 1.09884207927189e-05, 0.000109942279748489, 1614252.03775707, -5.55107506603362e-05, -0.000101415236024689, 0, 0, 0, 807126.018909039, 0, 4.44984444815081e-05, 4.23739973414315e-05, 3.98299311500222e-05, 0.000134477858232809, 1614252.03814457, -3.10916645899888e-05, 6.98764092455081e-05, 0.000118855941828356, 807126.019372275, -6.26680244442229e-05, 8.61705353147083e-05, 3.74810195311454e-05, 0.000574121555276204, 0.000452649568201762, 0.00034785258021173, 0.00075431919987169, 0.00119516726519006, 0.00026006076066607, 0.000651204051060448, 0.00102344393868468}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -291.600000056394, 0, 0, 0, 0, 0, 0, 0, 3244.05000032698, 0, 6087.15000140108, 619.650000370141, 0, 0, 0, 0, 0, 0, -5904.90000015228, 0, -46391.7375100291, 0, 0, -67058.8875108325, -328.050000768709, -6232.9499987969, -12274.5375128652, 0, 0, 0, 0, 0, 0, 2952.44999739533, 0, 167920.59378467, 0, 0, 0, 116621.775032755, 0, 503023.668817694, 0, 5904.899977258, 85498.0313326123, 2952.44998856546, 6150.93754644795, 122772.712567653, 0, 0, 0, 0, 0, 0, 8.4880707075615e-06, 0, -306685.743815812, 0, 0, 0, -55358.4375580132, 0, -1777005.84394189, 4.18275149772243e-06, 8.07887513339125e-05, -265720.500238737, 0, 0, -808233.187666177, 3.36382667857424e-05, -9.85667072267472e-05, 0.000101326513340429, -110716.875120608, -38750.9064744733, -846984.094136304, -55358.4377284096, 0, 0, 0, 0, 0, 0, -1.22708991398016e-05, 0, 269042.006319471, 0, 0, 0, 6.16777793786485e-05, 0, 3128858.88778495, -9.93823492598813e-06, -0.000127747593159615, 368687.194106452, 0, 0, 0, 348758.156475882, 0, 2491129.68783718, -5.53952596981862e-05, 0.000115585251587203, -0.000258563274309759, 6.80703459257401e-05, 99645.1879624134, 2590774.87583949, -0.000198623194024794, 0.000330455771403179, 697516.313177896, 0.000448857053078367, 0.000360188196265583, 348758.157115433, 0, 0, 0, 0, 0, 0, 8.61266394595895e-06, 0, -89680.6687880285, 0, 0, 0, -3.30622595939277e-05, 0, -2600739.39396218, 9.33037225717929e-06, 9.59827552097634e-05, -179361.33776489, 0, 0, 0, -0.00017539685332518, 0, -3407865.41282777, 5.00362249049188e-05, -6.54520965990741e-05, 0.000264321789348732, 2.91604221136355e-05, -89680.6692045377, -3497546.08208949, 0, 0, -896806.687775172, 0.000190865952617701, -0.000257257729281003, 0.000336359149437885, -0.000547995608897555, -0.000409146523713807, -1793613.3759762, -0.000394216043310157, -0.000141314670059509, -896806.688587963, -0.000553851774701759, -0.0010213744370088, -0.00116980598273758, 0, 0, 0, 0, 0, 0, -2.34677603238733e-06, 0, 8.38882175389834e-06, 0, 0, 0, 6.19620950308964e-06, 0, 807126.018811847, -3.11440460531753e-06, -2.79253173254792e-05, 7.69807120671173e-05, 0, 0, 0, 5.64789349632672e-05, 0, 1614252.03762429, -1.72106121868799e-05, 1.28801455058938e-05, -9.50312706462858e-05, -3.31007212000819e-05, 0.000171969984604567, 1614252.03781779, 0, 0, 0, 9.93567279727289e-05, 0, 807126.018870309, -7.20515004439402e-05, 7.9279822786428e-05, -0.000171067937769218, 0.000171587415643921, 0.00018827690440388, 1614252.03797398, 0.00011928285451628, -4.01408661599212e-05, 807126.019243821, -0.000157578125918096, 0.000147400652684122, 0.000402081284309036, 0.0002637408726214, 0.000443382563826715, 0.000141490655986633, 0.000498722123884557, 0.000556328801862692, 0.000717618796462211, 0.000783974655889711}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -1701.00000028325, 0, 0, 0, 0, 0, 0, 0, 31316.6250059794, 0, 23662.1250022107, 8353.12500503661, 0, 0, 0, 0, 0, 0, -196921.125047096, 0, -90441.5625028505, 0, 0, -425143.687544935, -15263.437535141, -143294.062579788, -105705.00004143, 0, 0, 0, 0, 0, 0, 556659.843919377, 0, 149672.812489444, 0, 0, 0, 2559610.12535321, 0, 1532608.59381494, 12301.8751176238, 787320.000427557, 311647.500103455, 242962.031675108, 174276.562751485, 1775570.62562452, 0, 0, 0, 0, 0, 0, -787935.094075853, 0, -114407.437474851, 0, 0, 0, -6759265.21983249, 0, -2367495.84357127, -3690.56270793558, -1749326.62606929, -346912.875090901, 0, 0, -8224418.53205515, -180837.563541913, -125479.125708841, -1138538.53289378, -9362957.06540805, -354294.000523674, -4915829.25148272, -2729170.97162759, 0, 0, 0, 0, 0, 0, 548048.531601385, 0, 33215.0624923778, 0, 0, 0, 8785384.03281802, 0, 1677360.65569884, 0.000205343553670573, 1693968.188904, 132860.250045085, 0, 0, 0, 17537553.0022583, 0, 10944363.0941698, 49822.5950494598, 33215.0634259379, 1793613.37784707, 19331166.3805078, 199290.375327209, 5081904.56362446, 697516.315252403, 132860.251326952, 22586242.5062263, 1826828.44318029, 12339395.7279087, 5181549.75537924, 0, 0, 0, 0, 0, 0, -149467.781451928, 0, -1.56729516430445e-05, 0, 0, 0, -5530307.90733792, 0, -448403.34325504, -0.000110719168888365, -597871.125948817, -6.0019657251803e-05, 0, 0, 0, -16142520.377397, 0, -6427114.59336582, -0.000813810197992878, -0.000536752885057345, -896806.689855245, -17039327.0671225, -1.67356940055758e-07, -1793613.37500179, 0, 0, -16740391.5015367, -149467.783362436, -0.00133541866759665, -597871.12792735, -34078654.1336824, -0.000847316159783493, -19430811.5671036, -448403.348579481, -17936133.7607203, -2690420.06558177, -6875517.94877424, -1793613.38267953, -19729747.1382954, 0, 0, 0, 0, 0, 0, 4.84233992178615e-05, 0, 9.62051852655921e-06, 0, 0, 0, 1345210.03154264, 0, -0.000147884276851323, 2.58650100811139e-05, 0.000262188127330204, 3.86045063127899e-05, 0, 0, 0, 5380840.12587703, 0, 1345210.03094195, 0.000208862043228623, 0.000111012442108013, 0.000768562086588515, 5380840.12643664, -1.69730369205007e-05, -0.000182933546943339, 0, 0, 0, 8071260.18848454, 0, 5380840.12520218, 0.000616508820973634, 0.000438601513946795, 0.0011336262618104, 16142520.3788675, 0.00015144120285613, 5380840.125926, 0.00144889079599478, 8071260.19172793, 0.000253657913797479, 0.00102885559653702, 0.000550805960280887, 16142520.3787513, 1345210.03587108, 0.00350700832904812, 5380840.13126457, 16142520.3834951, 0.00261294500606041, 8071260.19257273, 5380840.13355936}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 850.500000165985, 0, 0, 0, 0, 0, 0, 0, -16934.0625034749, 0, -10555.3125013189, -2900.81250228304, 0, 0, 0, 0, 0, 0, 120034.406276594, 0, 29387.8125016797, 0, 0, 206830.968777492, 3280.50001482745, 54424.4062894857, 32668.3125151545, 0, 0, 0, 0, 0, 0, -393660.000092568, 0, -30754.6874934279, 0, 0, 0, -1427222.53145995, 0, -547638.468795384, -1230.18754913387, -346912.875224146, -62739.5625159969, -57818.8126852989, -33215.0625930559, -605457.281533975, 0, 0, 0, 0, 0, 0, 638467.312672032, 0, 11071.6874834198, 0, 0, 0, 4487724.00062727, 0, 540667.406167042, 8.92184224520791e-05, 944784.000579912, 33215.0624396834, 0, 0, 3448830.65674821, 20298.0941691772, 11071.6877710298, 326614.782008378, 3775445.43895787, 33215.0626395533, 1101632.90681503, 581263.594936543, 0, 0, 0, 0, 0, 0, -498225.937679011, 0, 7.56626148178502e-06, 0, 0, 0, -6842302.87588964, 0, -182682.843454748, -9.36188406714561e-05, -1096097.06327355, 0.000137847135014295, 0, 0, 0, -9200572.31383528, 0, -3039178.21907611, -0.000500652213773279, -0.000360669984289719, -697516.313880659, -9898088.62790843, 6.73974512011245e-05, -548048.531401636, -99645.1886422796, -0.000472704301687639, -6178001.62792483, -182682.845982231, -3238468.59783408, -548048.53326096, 0, 0, 0, 0, 0, 0, 149467.781349136, 0, 6.98713636750111e-06, 0, 0, 0, 4932436.78185769, 0, -0.000274698940840401, 5.51535630877682e-05, 448403.344273551, -8.9503294568844e-05, 0, 0, 0, 10313276.9076215, 0, 896806.687357986, 0.000309395382107726, 0.000210415946311812, 448403.344935003, 10761680.2525334, -0.00022112629639208, -0.000426496013249737, 0, 0, 6427114.59472318, 0.000801968356758202, 0.000521751116719171, 149467.782523586, 13003696.973132, 0.00022393004453731, 2690420.06440643, 0.00182178697745963, 6726050.16128595, 0.000768978611576757, 896806.692173312, 0.00291584164783294, 2690420.06820146, 0, 0, 0, 0, 0, 0, -2.28803760311842e-05, 0, -5.06996163687242e-06, 0, 0, 0, -1345210.03141173, 0, 8.34516496793629e-05, -1.40941170705414e-05, -0.000143046682297073, 1.54605048497332e-05, 0, 0, 0, -4035630.09423988, 0, 0.000153620987100274, -8.19827404715693e-05, -4.46414513548197e-05, -0.000395280839254589, -4035630.09456172, 9.18164558753153e-05, 0.000253757337785669, 0, 0, 0, -4035630.09433437, 0, -1345210.03141923, -0.000218524567539924, -0.000179152323628724, -0.000515662271321386, -8071260.18952895, 1.11446795410961e-05, -0.000234177807429269, -0.000521938501796358, -4035630.0958207, 0.000204466367955089, -0.000396289466531525, -0.000207120240810267, -4035630.09555205, -0.00180460206948887, -0.00132177913718777, -1345210.0339681, -4035630.09763424, -0.000869302691164077, -0.00200492395465255, -0.00352690754294265}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -243.000000065447, 0, 0, 0, 0, 0, 0, 0, 4525.87500149789, 0, 3796.87500045239, 516.37500019694, 0, 0, 0, 0, 0, 0, -28476.5625113005, 0, -17085.9374997239, 0, 0, -70166.2500122961, -273.375001598661, -9066.93750701847, -7517.81249826815, 0, 0, 0, 0, 0, 0, 75861.5625389953, 0, 24603.7499921322, 0, 0, 0, 435281.343840925, 0, 310212.281274426, 7.78625613282942e-06, 50847.75004945, 28294.3124716397, 4510.68752879974, 3690.5625023363, 131425.031283899, 0, 0, 0, 0, 0, 0, -84882.9375725139, 0, -11071.6874796607, 0, 0, 0, -1131157.40651425, 0, -428105.249974279, -1.63493893072833e-05, -107026.312639761, -22143.3748930776, 0, 0, -1861888.78148584, -6.76935088400494e-05, -4.80308632143376e-05, -22143.3751411063, -730731.375289067, -11071.6874356401, -488999.531209301, -60894.2814118984, 0, 0, 0, 0, 0, 0, 33215.0625761279, 0, -2.14286987189815e-05, 0, 0, 0, 1212349.78161697, 0, 182682.843627349, 1.78595861065154e-05, 66430.1251985573, -0.000163597477582975, 0, 0, 0, 4550463.56310202, 0, 2358269.43768938, 8.1489702102125e-05, 9.09874081648807e-05, 33215.0627828234, 1511285.34439364, -0.000236541116864772, 365365.687110087, 0.000186768604091733, 4.95667586126903e-05, 2657205.0004795, 0.000431501146158626, 298935.56328227, 182682.843741647, 0, 0, 0, 0, 0, 0, -4.28863052192076e-05, 0, 9.55208420700332e-06, 0, 0, 0, -448403.343995716, 0, 0.000123086670053441, -1.07047409276686e-05, -0.000142745154856009, 0.000109001553924679, 0, 0, 0, -4334565.6568539, 0, -896806.687478234, -5.1147550980614e-05, -5.87851980094238e-05, -0.000264428037279929, -896806.688069806, 0.000266348681778813, 0.000547220444256569, 0, 0, -4782969.00047252, -0.000109579431216277, -0.000150846032046193, -0.000207993023269244, -5231372.34472246, 6.65745886101586e-05, -1793613.37506044, -0.000423883131004868, -448403.344807723, 0.000624182312572914, -0.00101505680661795, -0.000388643395630031, -896806.688361834, 0, 0, 0, 0, 0, 0, 1.01461476958733e-05, 0, -1.25768469535172e-06, 0, 0, 0, 6.40785689612885e-05, 0, -3.96670666993505e-05, 2.82782265135916e-06, 4.16366029652149e-05, -2.5793212499068e-05, 0, 0, 0, 1345210.03146197, 0, -5.56319751703636e-05, 1.35483549434618e-05, 1.17301734738699e-05, 9.67518799901543e-05, 0.000179958747624491, -9.45638442144277e-05, -0.000226392143223826, 0, 0, 0, 2690420.06277183, 0, 1345210.03134657, 2.38260982498598e-05, 6.66199138609778e-05, 8.9806649378976e-05, 2690420.06298384, -7.16380135223353e-05, -0.000176935751971723, 0.000132152268259989, 0.000443484146388172, -0.00047581639403405, 4.20011771121703e-05, 5.6323879765352e-05, 2690420.06285275, 0.000429180583779421, 0.000295781136301433, 0.00060656015420475, 1345210.03202786, -9.68981019219439e-05, 2.68990037581289e-05, 0.000698452395775586}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -243.00000005365, 0, 0, 0, 0, 0, 0, 0, 3796.87500078769, 0, 4525.87500086273, 516.37500005073, 0, 0, 0, 0, 0, 0, -17085.9375049569, 0, -28476.5625048384, 0, 0, -70166.250010308, -273.375000399697, -7517.81249965485, -9066.93750365784, 0, 0, 0, 0, 0, 0, 24603.7500164002, 0, 75861.5625135622, 0, 0, 0, 310212.281302072, 0, 435281.343795388, 1.90742540636235e-06, 28294.3124942874, 50847.7500230539, 3690.56250545083, 4510.68751685384, 131425.031277087, 0, 0, 0, 0, 0, 0, -11071.6875308297, 0, -84882.9375223554, 0, 0, 0, -428105.250131861, 0, -1131157.40634045, -1.56754813517056e-06, -22143.3749777797, -107026.312564628, 0, 0, -1861888.78143345, -3.0174508989431e-05, -4.88416822390702e-05, -11071.6874977031, -488999.531330439, -22143.3750619412, -730731.375128052, -60894.2814246636, 0, 0, 0, 0, 0, 0, 3.34819517791339e-05, 0, 33215.0625221634, 0, 0, 0, 182682.843923119, 0, 1212349.7813463, -2.38703159191511e-06, -3.32238466083978e-05, 66430.1251013349, 0, 0, 0, 2358269.43786039, 0, 4550463.56276182, 4.50374655912057e-05, 5.31011538603057e-05, -5.22851229209339e-05, 365365.687572364, 33215.0625989864, 1511285.34395225, 2.50014902226677e-05, 0.000130464442153609, 2657205.00037343, 0.000477798647467593, 182682.844223676, 298935.562908133, 0, 0, 0, 0, 0, 0, -1.97911923021328e-05, 0, -1.21452321818074e-05, 0, 0, 0, -0.000112013395510919, 0, -448403.343805436, 3.79184059372714e-06, 1.94258035487791e-05, -8.37954080469656e-05, 0, 0, 0, -896806.687830418, 0, -4334565.65641775, -2.569675098189e-05, -1.40514484312997e-05, 7.00421667109653e-05, 1.06594999928702e-05, -9.47082918237701e-05, -896806.687653624, 0, 0, -4782969.0003469, -2.65135981993095e-05, -0.000115084229510155, 5.93961490404046e-05, -1793613.37540984, -0.000119700744473519, -5231372.34416088, -0.000463109283674991, -0.000394651098139, -448403.343990725, -0.000882778379111962, -0.000751440591144777, -896806.688438716, 0, 0, 0, 0, 0, 0, 4.96267455622405e-06, 0, 2.80457469506868e-06, 0, 0, 0, 2.82761184199079e-05, 0, 1.37302469640094e-05, -1.31126303500097e-06, -3.03214526172409e-06, 2.75694577260689e-05, 0, 0, 0, 0.000111850260821248, 0, 1345210.03129183, 5.09239226458131e-06, -4.41555135247652e-06, -2.42981103215825e-05, -2.65995523496954e-05, 4.25417785366798e-05, 5.28678038127998e-05, 0, 0, 0, 1345210.03141566, 0, 2690420.06260794, 5.96121537566705e-06, 2.84021271138041e-05, -5.18745492789324e-05, 0.000137903186167408, 4.62820901773358e-05, 2690420.06263841, 0.000140417623870085, 7.87041277910362e-05, 2.2812063746608e-05, -3.11944408813305e-05, 6.04586020326906e-05, 2690420.06276555, 0.000440292534498498, 0.000393765638164595, 0.000393217729155339, 0.000526223352045227, 0.000247116501564833, 1345210.03162263, 0.000822582599416602}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 850.500000132059, 0, 0, 0, 0, 0, 0, 0, -10555.3125009714, 0, -16934.062503129, -2900.81250101575, 0, 0, 0, 0, 0, 0, 29387.8125024451, 0, 120034.406271481, 0, 0, 206830.968776136, 3280.50000376088, 32668.3125005749, 54424.4062822727, 0, 0, 0, 0, 0, 0, -30754.6875027943, 0, -393660.000071201, 0, 0, 0, -547638.468838073, 0, -1427222.53140317, -1230.18750772441, -62739.562458674, -346912.875196923, -33215.0625057371, -57818.812625781, -605457.28146075, 0, 0, 0, 0, 0, 0, 11071.6875000527, 0, 638467.312630557, 0, 0, 0, 540667.406426539, 0, 4487724.00040889, 3.90280055331361e-06, 33215.0623238257, 944784.000544942, 0, 0, 3448830.65665436, 11071.6875249094, 20298.0940248905, 33215.0623588247, 1101632.90673266, 326614.781801632, 3775445.43858094, 581263.594604622, 0, 0, 0, 0, 0, 0, 4.12133863774665e-06, 0, -498225.93763377, 0, 0, 0, -182682.843959999, 0, -6842302.87558184, 1.04496729106808e-05, 0.000297205403886786, -1096097.06329219, 0, 0, 0, -3039178.2193534, 0, -9200572.31324592, -2.20092040448284e-05, -0.000309587430183182, 0.000505405030011743, -548048.531678388, -697516.313554495, -9898088.62715096, 0.000239079408541322, -99645.1883307662, -6178001.62722879, -182682.845493159, -548048.533025292, -3238468.59649707, 0, 0, 0, 0, 0, 0, -4.23003339138213e-06, 0, 149467.781321092, 0, 0, 0, 0.000126432501823364, 0, 4932436.78167184, -1.53055796941204e-05, -0.000227461383364955, 448403.344331428, 0, 0, 0, 896806.68800841, 0, 10313276.9069232, -9.945811993934e-06, 0.00015625944880028, -0.000581647874115512, 6.36330622026194e-05, 448403.344744025, 10761680.2520035, 0, 0, 6427114.59440323, -0.000225616312574358, 0.000632290227577239, -0.000633918516078059, 2690420.06450667, 149467.782191128, 13003696.9716581, 0.00153209629494588, 0.00123771110183085, 6726050.15929048, 0.00263861655688257, 896806.690925911, 2690420.06693807, 0, 0, 0, 0, 0, 0, 1.24932328812855e-06, 0, -1.51707799959696e-05, 0, 0, 0, -2.83025021962372e-05, 0, -1345210.03137072, 5.74300155991556e-06, 6.55956497624852e-05, -0.000168260370381106, 0, 0, 0, -0.000173022327105005, 0, -4035630.09399501, 1.02588655918506e-05, -2.3687238808063e-05, 0.000219098333440409, 6.52808628342168e-05, -0.000371836417159857, -4035630.09447359, 0, 0, 0, -1345210.0315096, 0, -4035630.09400188, 8.93723176099231e-05, -0.000179645257562772, 0.000374002728129276, -0.000654415276399751, -0.000408334764757195, -8071260.18878391, -0.000462331630459313, -0.000180858255788805, -4035630.09497425, 0.000295470238008541, -0.000339550593261792, -4035630.09510095, -0.00121803834378325, -0.00150207319819203, -0.000989143688989679, -0.00213128801519655, -1345210.0329291, -4035630.096195, -0.00300614574539128}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -1701.00000017498, 0, 0, 0, 0, 0, 0, 0, 23662.1250022431, 0, 31316.6250033133, 8353.12500270684, 0, 0, 0, 0, 0, 0, -90441.5625133097, 0, -196921.125018712, 0, 0, -425143.687536515, -15263.4375177519, -105705.000026043, -143294.062551626, 0, 0, 0, 0, 0, 0, 149672.812543903, 0, 556659.843801693, 0, 0, 0, 1532608.59392727, 0, 2559610.12516014, 12301.8750576238, 311647.500072081, 787320.000254605, 174276.562673062, 242962.031501784, 1775570.62549663, 0, 0, 0, 0, 0, 0, -114407.437582826, 0, -787935.093833401, 0, 0, 0, -2367495.84420852, 0, -6759265.21905964, -3690.56259279356, -346912.875053677, -1749326.62560024, 0, 0, -8224418.53185389, -125479.125477492, -180837.563140727, -354294.00041394, -4915829.25163204, -1138538.53208447, -9362957.06444325, -2729170.97106759, 0, 0, 0, 0, 0, 0, 33215.0625890473, 0, 548048.531329226, 0, 0, 0, 1677360.65687691, 0, 8785384.03158175, 7.43797557540472e-05, 132860.24994965, 1693968.18829686, 0, 0, 0, 10944363.094965, 0, 17537553.0007678, 33215.0631141276, 49822.5945454807, 199290.375284489, 5081904.56480443, 1793613.37622488, 19331166.3781004, 132860.250804247, 697516.314023775, 22586242.5052607, 1826828.44227523, 5181549.75600749, 12339395.7246385, 0, 0, 0, 0, 0, 0, -5.17465777444116e-05, 0, -149467.781290315, 0, 0, 0, -448403.344166516, 0, -5530307.90644852, -2.72734109784754e-05, 8.48800962265617e-05, -597871.125568433, 0, 0, 0, -6427114.59494864, 0, -16142520.3754411, -0.000369594055649611, -0.000442480999287979, 4.90000408275398e-05, -1793613.37640173, -896806.688453103, -17039327.0648816, 0, 0, -16740391.5009959, -0.000630134198280803, -149467.782578039, -0.000270853980727608, -19430811.5682255, -597871.126337228, -34078654.1306121, -448403.347952766, -2690420.068184, -17936133.7552234, -1793613.38324139, -6875517.94539162, -19729747.1364724, 0, 0, 0, 0, 0, 0, 1.28595696667062e-05, 0, 8.37828085294451e-06, 0, 0, 0, 0.000104965164638408, 0, 1345210.03130055, 3.3434382498952e-06, -2.90422577891347e-05, 0.000165914107471676, 0, 0, 0, 1345210.03168653, 0, 5380840.12510834, 8.6744179851028e-05, 8.3011296525858e-05, -7.23331424307388e-05, 0.000277075320866008, 0.000332062922466604, 5380840.12574969, 0, 0, 0, 5380840.12552248, 0, 8071260.18773991, 0.000171371433946883, 0.000408857925390571, -9.68962580026436e-05, 5380840.12714548, 0.00044345700079499, 16142520.377033, 0.0012900401234762, 0.00173063707288204, 8071260.18910529, 4.12962727711137e-05, 0.000599739239622519, 16142520.3781979, 0.00362580605294029, 1345210.03481553, 0.00383652210156635, 8071260.19375416, 5380840.12833621, 16142520.3804514, 5380840.13267366}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 756.000000128712, 0, 0, 0, 0, 0, 0, 0, -12784.5000027518, 0, -12784.5000011616, -2578.50000087449, 0, 0, 0, 0, 0, 0, 68343.7500209167, 0, 68343.7500020485, 0, 0, 213232.500024195, 2916.00000772, 40641.7500170215, 40641.7500072892, 0, 0, 0, 0, 0, 0, -144888.750074364, 0, -144888.749994152, 0, 0, 0, -1109082.37517048, 0, -1109082.37505718, -1093.50003061848, -186441.750094855, -186441.750004419, -42646.5001044294, -42646.5000631682, -633956.625153422, 0, 0, 0, 0, 0, 0, 132860.250143388, 0, 132860.249978217, 0, 0, 0, 2236480.8754979, 0, 2236480.8749868, 5.54779718309963e-05, 280482.750234659, 280482.749947812, 0, 0, 5447270.25045289, 14762.2502904163, 14762.2502335849, 162384.750388784, 2841733.12577154, 162384.750084529, 2841733.12532226, 620014.500858123, 0, 0, 0, 0, 0, 0, -44286.750156353, 0, -44286.7499759022, 0, 0, 0, -1926473.62570181, 0, -1926473.62482601, -5.15779352602132e-05, -132860.250309736, -132860.249900311, 0, 0, 0, -9787371.75114497, 0, -9787371.75037031, -0.000380130047911055, -0.000329108647085091, -132860.250622929, -4052237.62641382, -132860.249888563, -4052237.62498802, -44286.7506925821, -44286.7504347112, -12068139.3767989, -199290.377023401, -2325054.37780468, -2325054.37643741, 0, 0, 0, 0, 0, 0, 9.15447229469214e-05, 0, -1.01709803350712e-05, 0, 0, 0, 597871.125475961, 0, 597871.124813754, 2.53079021107185e-05, 0.000218481201187943, -5.45190157499431e-05, 0, 0, 0, 7174453.50116954, 0, 7174453.49997571, 0.000239511815534682, 0.000182060972514041, 0.000490855459081262, 1793613.37609055, -0.000221588822641875, 1793613.37458773, 0, 0, 13153164.7508747, 0.000540917445409816, 0.000534527053497295, 0.00061864190456842, 14946778.1276209, 0.000198182558912634, 14946778.1262078, 0.00187992486258792, 1793613.37813078, 1793613.37525378, 597871.128983353, 597871.127713923, 8370195.75420208, 0, 0, 0, 0, 0, 0, -2.25114739714425e-05, 0, 9.92909701012737e-07, 0, 0, 0, -0.000125116039200356, 0, 6.09831694872219e-05, -5.5509084718844e-06, -6.53550000202366e-05, 4.94577553703292e-06, 0, 0, 0, -1793613.37541868, 0, -1793613.37490624, -5.99049155567413e-05, -2.98593398338134e-05, -0.000163220569644718, -0.000301690815783554, 7.90824751665184e-05, 0.000216619776107894, 0, 0, 0, -5380840.12551484, 0, -5380840.12516486, -0.000155604779878158, -0.000184634012583648, -0.000204474068719384, -5380840.12614022, 1.40527703635306e-05, -5380840.12511544, -0.000581800102631074, -0.00113124001744557, 0.000387446033380961, -0.00019849808519245, -0.000222709735523572, -10761680.2511847, -0.00177532185652435, -0.00143530269643209, -0.0020860842735716, -5380840.12774357, -0.000714490856075753, -5380840.12636484, -1793613.37816702}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -7.89568453687543e-08, 0, 0, 0, 0, 0, 0, 0, 5.36377637125744e-07, 0, 1.06322871033753e-06, 61236.0000139474, 0, 0, 0, 0, 0, 0, -7.28681871705165e-06, 0, -8.3240536600364e-06, 0, 0, 4.67064600808209e-06, -548718.300091431, -548718.300164199, -548718.300153022, 0, 0, 0, 0, 0, 0, 3.82051594498426e-05, 0, 3.93846242831158e-05, 0, 0, 0, -2.77071013098116e-07, 0, -5.77598149466781e-05, 2015375.17527849, 2015375.17575652, 2015375.17569877, 4030750.3509484, 4030750.35074912, 4030750.35140026, 0, 0, 0, 0, 0, 0, -9.4200022657904e-05, 0, -0.000105050052935322, 0, 0, 0, -9.43620320831145e-05, 0, 0.000181334627359193, -3886162.31297151, -3886162.31426867, -3886162.31425817, 0, 0, 0.000206951079659247, -11658486.9398018, -11658486.939224, -11658486.9409742, -11658486.9424933, -11658486.9398705, -11658486.9420315, -23316973.8812225, 0, 0, 0, 0, 0, 0, 0.000122927206169068, 0, 0.000149174399897651, 0, 0, 0, 0.0002235325601239, 0, -0.00020918839745886, 4151882.81295191, 4151882.81472139, 4151882.81500349, 0, 0, 0, -9.80467425798415e-05, 0, -0.00069755115921538, 16607531.2528289, 16607531.2521649, 16607531.2559597, 16607531.2586032, 16607531.2540163, 16607531.257454, 24911296.8811742, 24911296.8785946, 24911296.8863226, 49822593.7612467, 49822593.7670914, 49822593.7639274, 0, 0, 0, 0, 0, 0, -8.32113962130685e-05, 0, -0.000105295811431449, 0, 0, 0, -0.000190969473028551, 0, 7.86374149208701e-05, -2331697.38772755, -2331697.38893439, -2331697.38934847, 0, 0, 0, -0.000163229490583388, 0, 0.000750315474070719, -11658486.9392484, -11658486.9389052, -11658486.9423738, -11658486.9445564, -11658486.9411258, -11658486.9438898, 0, 0, 0.000592792342757209, -23316973.8798614, -23316973.8777822, -23316973.8819168, -23316973.8877765, -23316973.8785106, -23316973.8860207, -46633947.7593215, -46633947.7693823, -46633947.7639282, -69950921.6455736, -69950921.6408624, -69950921.6499474, 0, 0, 0, 0, 0, 0, 2.30930571786935e-05, 0, 2.91177264850623e-05, 0, 0, 0, 5.71902036765314e-05, 0, 2.09777382776239e-06, 538084.012546166, 538084.012876342, 538084.01304365, 0, 0, 0, 0.000115604762848923, 0, -0.000253419334754407, 3228504.07543497, 3228504.07536528, 3228504.07653671, 3228504.07719857, 3228504.07632396, 3228504.07723216, 0, 0, 0, -0.000100110499676121, 0, -0.000407992715073466, 8071260.18892912, 8071260.18837848, 8071260.19029024, 8071260.19278749, 8071260.18893849, 8071260.19164186, 16142520.3779195, 16142520.3827763, 16142520.3805127, 10761680.2526487, 10761680.2511838, 10761680.2558308, 32285040.7583574, 32285040.7563926, 32285040.7614757, 32285040.7641343, 32285040.7572691, 32285040.7616088, 48427561.1397013}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.27589644519484e-08, 0, 1.49771463796003e-08, -1093.50000000991, 0, 0, 0, 0, 0, 0, -2.27687230956426e-07, 0, -2.53335196772595e-08, 0, 0, -3.72587739849353e-07, 23564.9250001611, 1093.50000002605, 1093.50000006503, 0, 0, 0, 0, 0, 0, 7.5934601966061e-07, 0, -1.48547603103859e-07, 0, 0, 0, 2.25291435724082e-06, 0, 1.35334050760089e-06, -188546.737500843, -1.49990131029788e-08, -1.10098119359927e-07, -22471.4250007201, -22471.4250008872, -1.55900995099978e-07, 0, 0, 0, 0, 0, 0, -1.43647859274472e-06, 0, 6.58000461551594e-07, 0, 0, 0, -5.79696569120795e-06, 0, -1.79372073437314e-06, 730731.375002361, -2.18095477622386e-07, 5.91956745709173e-07, 0, 0, -7.65686805659414e-06, 166075.312502474, 166075.312503431, 5.55834147767725e-07, 1.89840106557396e-06, 5.30044596780986e-07, -2.43035724136723e-06, 8.24361532922388e-06, 0, 0, 0, 0, 0, 0, 1.58662028177681e-06, 0, -1.0369147905883e-06, 0, 0, 0, 7.60963365064865e-06, 0, 2.21285343341959e-07, -1461462.7500034, 8.26225572152335e-07, -2.16303861316687e-06, 0, 0, 0, 1.54153085060812e-05, 0, 1.09196351749889e-05, -564656.062505003, -564656.062507044, 8.79356066363012e-07, -5.00700876759182e-06, 1.47428808987609e-06, 8.62338960978805e-06, 3.02977467482496e-06, 1.03113012964021e-07, 1.71001185926144e-06, -2.51081579832806e-05, -2.21991615367788e-05, -9.85435470862818e-06, 0, 0, 0, 0, 0, 0, -9.51935745948843e-07, 0, 7.40493406796667e-07, 0, 0, 0, -5.08908773209083e-06, 0, 1.24545427042616e-06, 1434890.70000234, -1.0480213741957e-06, 2.85847807930635e-06, 0, 0, 0, -1.34738180145475e-05, 0, -6.46040724671929e-06, 896806.687505887, 896806.68750694, -1.80900548924178e-06, 3.93720785914508e-06, -6.30914668033818e-07, -8.61269009643803e-06, 0, 0, -1.60060414414777e-05, -6.37925941738625e-06, 4.3154458214001e-07, -9.3071186445294e-06, 6.69040104684542e-06, -7.84550689894051e-06, -1.16519047176728e-05, 2.82522640382848e-05, 2.55457643090061e-05, -2.12298431883692e-06, 3.84274622026344e-05, 3.0699940303533e-05, 2.53854071763323e-05, 0, 0, 0, 0, 0, 0, 2.38687316921452e-07, 0, -2.01244230995237e-07, 0, 0, 0, 1.38383951671995e-06, 0, -6.86084279214056e-07, -538084.012500573, 4.09862539105056e-07, -1.21946373127781e-06, 0, 0, 0, 4.39223990924918e-06, 0, 1.13172394983944e-06, -538084.012502706, -538084.012502596, 8.39098910307153e-07, -5.8936755090777e-07, -7.68207184479567e-07, 2.40774698435513e-06, 0, 0, 0, 7.19227419803349e-06, 0, 5.56181107998515e-06, 2.40465204590814e-06, -8.55734482914632e-07, 4.95219984571135e-06, -4.92157554967833e-06, 4.76703281070026e-06, 7.99106072845065e-06, -1.12865243018019e-05, -9.81536583639234e-06, 4.85368288145762e-06, 8.30081244005305e-06, 4.94075775869381e-06, 8.38993988592696e-07, -1.83690726885884e-05, -1.86372934461524e-05, -1.81457556540533e-05, -1.82767335216385e-05, -7.65864733043398e-06, -2.85104696219301e-06, -2.99327201694563e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.08184466794373e-08, 0, 1.56850123623582e-07, -1093.50000015214, 0, 0, 0, 0, 0, 0, 6.00000504778068e-07, 0, -1.35510048453635e-06, 0, 0, -1.02126207826012e-06, 1093.50000086956, 23564.9250020541, 1093.50000154676, 0, 0, 0, 0, 0, 0, -3.31501143246527e-06, 0, 5.40143972347759e-06, 0, 0, 0, -2.41970405281785e-08, 0, 9.27508209848864e-06, -2.01882957630338e-06, -188546.737511878, -6.62251019456606e-06, -22471.4250108502, -7.18080878487788e-06, -22471.4250146923, 0, 0, 0, 0, 0, 0, 8.13800646375328e-06, 0, -1.1351351700673e-05, 0, 0, 0, 9.41433534109167e-06, 0, -3.2468693696371e-05, 2.28492229625715e-06, 730731.375033777, 1.56469957357426e-05, 0, 0, -1.50053086668892e-05, 2.48685523549924e-05, 1.07671726953018e-05, 166075.312545238, 166075.312567729, 2.50269492380352e-05, 3.68801136222127e-05, 6.87963476486857e-05, 0, 0, 0, 0, 0, 0, -1.04985238867153e-05, 0, 1.30238671921145e-05, 0, 0, 0, -2.06113068621064e-05, 0, 5.35471674916992e-05, -7.83271501363155e-07, -1461462.75005048, -2.16044173336634e-05, 0, 0, 0, -2.5797004437411e-06, 0, 5.32880822383723e-05, -3.01968389836884e-05, -6.75024958139128e-06, -564656.062585998, -564656.062641709, -4.84580131816467e-05, -3.73555171447188e-05, -7.52936373910937e-05, -2.18090037315047e-05, -0.000129738653874847, -0.000110961538790715, -0.000220125741716817, -0.000152581744224658, 0, 0, 0, 0, 0, 0, 6.94211401031279e-06, 0, -7.7561713292106e-06, 0, 0, 0, 1.78320254715456e-05, 0, -4.15665899376602e-05, -6.64350608806029e-07, 1434890.70003843, 1.57887848560967e-05, 0, 0, 0, 1.84184610183846e-05, 0, -6.73798261359945e-05, 1.87929641054496e-05, 6.47057031746934e-07, 896806.68757846, 896806.687632934, 4.97837394277656e-05, 1.55959639497123e-05, 0, 0, -2.25229348182914e-05, 6.083990366836e-05, 2.74801994501828e-06, 8.87831636087884e-05, 0.000196989623889016, 2.61472463513578e-05, 8.07758074064198e-05, 7.98878132393629e-05, 0.000274420825042011, 0.000144101496989596, 0.00024321119838072, 0.000155320167210244, 0.00032808355680642, 0, 0, 0, 0, 0, 0, -1.85167023474371e-06, 0, 1.88530694091551e-06, 0, 0, 0, -5.74883411351138e-06, 0, 1.22698666102174e-05, 4.76529765490481e-07, -538084.012511735, -4.57170148752417e-06, 0, 0, 0, -9.11096153926288e-06, 0, 2.83901654311661e-05, -4.72655868436396e-06, 7.70205676034557e-07, -538084.012528199, -538084.012546406, -1.99417252796443e-05, -2.55095792434042e-06, 0, 0, 0, -2.39157963748317e-06, 0, 2.08270658661118e-05, -1.92685091247647e-05, 2.42778812145328e-06, -3.77002284496759e-05, -9.55373761616456e-05, -1.51589178877982e-05, -7.13710540988214e-06, -2.13456150775886e-05, -0.000117151287266389, -5.43130737194516e-05, -3.46064486675929e-05, 4.59743921382596e-07, -7.48481895778975e-05, -9.24716218934665e-05, -5.05984121731013e-05, -0.000144897245612814, -0.000214048633606339, -7.06572097998259e-05, -0.000141718087096282, -0.0001817690428957}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.38411141012585e-08, 0, -1.32277549900369e-07, -1093.49999990825, 0, 0, 0, 0, 0, 0, 4.03884936865733e-07, 0, 9.30065126296563e-07, 0, 0, 2.39026036526781e-08, 1093.49999834385, 1093.49999797635, 23564.9250003781, 0, 0, 0, 0, 0, 0, -1.70058922828148e-06, 0, -4.015063026534e-06, 0, 0, 0, -5.45120796110829e-07, 0, 2.81682951321028e-06, 6.95848428536388e-06, 1.14731288561982e-05, -188546.73750897, 1.72263749234126e-05, -22471.4249910927, -22471.4249895854, 0, 0, 0, 0, 0, 0, 3.41079552791033e-06, 0, 9.9878875333601e-06, 0, 0, 0, 7.41133381066463e-06, 0, -8.80917495793006e-06, -1.41186409854872e-05, -2.72616836885566e-05, 730731.375037392, 0, 0, -1.62934816114853e-05, -4.71662877713765e-05, -4.013338702177e-05, -7.11483174143414e-05, -6.6465962590933e-05, 166075.312496349, 166075.312496402, -8.35759265628512e-05, 0, 0, 0, 0, 0, 0, -3.30177279869234e-06, 0, -1.36726335743008e-05, 0, 0, 0, -1.81175910337198e-05, 0, 8.45771599332089e-06, 1.52139946067179e-05, 3.24023725339087e-05, -1461462.75006651, 0, 0, 0, 1.71860101395075e-05, 0, 4.9455114834015e-05, 6.44690477429162e-05, 7.05738814124069e-05, 0.000129334761954744, 0.000126152595770767, -564656.062541573, -564656.062551918, 0.000135633672906296, 6.58634043636802e-05, 0.000125632960646129, 0.000175505445293299, 0.000303777469320996, 9.84509477036806e-05, 0, 0, 0, 0, 0, 0, 1.41518542658835e-06, 0, 9.56249309276992e-06, 0, 0, 0, 1.652725043774e-05, 0, -1.19548753012051e-06, -8.121880002272e-06, -1.96080272235375e-05, 1434890.70005461, 0, 0, 0, 1.27368561040206e-06, 0, -4.86713189077447e-05, -4.5206166319991e-05, -5.71556594674116e-05, -0.000106423914193198, -9.5893307266205e-05, 896806.687571301, 896806.687589366, 0, 0, -5.42513860015574e-05, -0.000113664569386549, -9.7924009983144e-05, -0.000162694207332519, -0.000192501781942502, -2.61192258634997e-05, -8.18540534257482e-05, -0.000160829778456141, -0.000396144936560042, 1.72927601039322e-05, -0.000368959595434993, -0.000169622390331922, -0.000360943199447188, 0, 0, 0, 0, 0, 0, -1.82218334279773e-07, 0, -2.66949949202856e-06, 0, 0, 0, -5.16561108086731e-06, 0, -1.27960419263145e-06, 1.62939934955889e-06, 4.93206646616062e-06, -538084.012517021, 0, 0, 0, -5.72446331504931e-06, 0, 1.55935030825353e-05, 1.27160620687546e-05, 1.74647069953471e-05, 3.2833350477064e-05, 2.56264421024625e-05, -538084.012533403, -538084.012542455, 0, 0, 0, 1.59622112774461e-05, 0, 3.09068320219343e-05, 3.70793572939119e-05, 4.38222409392008e-05, 6.75143345497985e-05, 7.73474185593844e-05, -8.34183251167981e-06, 7.31494970829899e-06, 5.77767646274972e-05, 0.000166889901347921, -5.35992343001046e-05, 6.57494863406937e-05, 3.79083467797351e-05, 9.19834027551757e-05, 0.000144359821476697, 8.74274241749266e-05, 0.000240177387804268, 0.000266824078980065, 2.60320993687802e-05, 0.000113847253254791, 0.000215501375188517}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 2.13305218035889e-07, 0, 0, 0, 0, 0, 0, 0, -2.50322591264763e-06, 0, -3.08554057999797e-06, -45927.00001105, 0, 0, 0, 0, 0, 0, 1.68713081685494e-05, 0, 2.21849762035377e-05, 0, 0, 2.07867285661667e-05, 721545.975069608, 308202.97513227, 308202.975111877, 0, 0, 0, 0, 0, 0, -6.20240754670652e-05, 0, -8.55322781828621e-05, 0, 0, 0, -9.45967453679045e-05, 0, -9.81751268485702e-05, -3591901.46270727, -818074.688169365, -818074.688018723, -4409976.15071053, -4409976.15054064, -1636149.37594504, 0, 0, 0, 0, 0, 0, 0.0001294738965212, 0, 0.000187428384286546, 0, 0, 0, 0.000245631263128907, 0, 0.000257790015805634, 8436625.8753467, 1073953.68921447, 1073953.68886959, 0, 0, 0.000298557817665066, 17947205.4391082, 17947205.4387856, 10584533.2528733, 3221861.06607464, 10584533.2517422, 3221861.06535379, 21169066.5039596, 0, 0, 0, 0, 0, 0, -0.000156697940451961, 0, -0.000230821975035197, 0, 0, 0, -0.000321935038799122, 0, -0.000405437459603352, -10363099.5003323, -697516.314823751, -697516.314541098, 0, 0, 0, -0.000597123883367387, 0, -0.000427671985343928, -31786814.8143265, -31786814.8141958, -12455648.4429686, -2790065.256682, -12455648.4406107, -2790065.25474349, -33181847.4422579, -33181847.4403146, -4185097.88218356, -66363694.8819617, -37366945.3240796, -37366945.3205265, 0, 0, 0, 0, 0, 0, 0.000102900863394726, 0, 0.000147526146928616, 0, 0, 0, 0.00020536741047259, 0, 0.000347507872978577, 6457008.15016858, 179361.339098006, 179361.339060293, 0, 0, 0, 0.000548611929389832, 0, 0.00031028411551315, 26007393.9385422, 26007393.9386651, 7174453.50487755, 896806.693382499, 7174453.50295617, 896806.691795868, 0, 0, 0.000552161331493435, 39459494.2534167, 39459494.252386, 26904200.6309922, 1793613.38370392, 26904200.627944, 1793613.38161236, 78918988.5056502, 28697814.0144393, 28697814.0077915, 80712601.8883861, 80712601.8839861, 43046721.014602, 0, 0, 0, 0, 0, 0, -2.82242338073923e-05, 0, -3.79104947547881e-05, 0, 0, 0, -5.27998395355911e-05, 0, -0.000119909936657058, -1614252.03753436, -0.000441419196457976, -0.000470742752639704, 0, 0, 0, -0.000174670313800409, 0, -0.000107936951450437, -8071260.18774156, -8071260.18782143, -1614252.03914784, -0.00194759036968654, -1614252.03862059, -0.00160169666683555, 0, 0, 0, -0.000281415337180006, 0, -0.000157273486443169, -16142520.3758956, -16142520.3758274, -8071260.19017017, -0.00384631522950382, -8071260.18877708, -0.00248634623077453, -32285040.751747, -8071260.19382198, -8071260.19070122, -16142520.3771234, -16142520.3761244, -0.0036801236370976, -48427561.130184, -48427561.1286528, -32285040.7582308, -16142520.384019, -32285040.7539189, -16142520.3811161, -48427561.133374}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -1.26191724107667e-07, 0, 0, 0, 0, 0, 0, 0, 1.41285813593855e-06, 0, 1.78708428640522e-06, 30618.0000048517, 0, 0, 0, 0, 0, 0, -9.08417592380797e-06, 0, -1.42771503819855e-05, 0, 0, -9.61364489035082e-06, -563699.250029467, -150356.250058667, -150356.250046485, 0, 0, 0, 0, 0, 0, 3.54663812285223e-05, 0, 5.68080015723484e-05, 0, 0, 0, 2.22293420637738e-05, 0, 6.5665772850043e-05, 3544580.25008575, 274741.875324899, 274741.875226628, 2579293.12529628, 2579293.12522187, 549483.750346582, 0, 0, 0, 0, 0, 0, -8.10573497986322e-05, 0, -0.000121972554944174, 0, 0, 0, -3.38681302639072e-05, 0, -0.000222214741007716, -10019877.1876433, -221433.750904662, -221433.750630568, 0, 0, -0.000109176557900968, -14171760.0006255, -14171760.0005292, -4373316.56382264, -664301.251394704, -4373316.56327234, -664301.250994089, -8746633.12637638, 0, 0, 0, 0, 0, 0, 0.000106437838821355, 0, 0.000144069303619641, 0, 0, 0, 3.18666979583494e-05, 0, 0.000378541592593018, 14182831.6876422, 66430.1263155163, 66430.1259731174, 0, 0, 0, 9.79773950211316e-05, 0, 0.000282826268580365, 31487879.2506456, 31487879.2507104, 3255076.12775026, 265720.502846022, 3255076.12647874, 265720.501695396, 20493693.5645819, 20493693.5637912, 398580.752486535, 40987387.1272863, 9765228.37933206, 9765228.37760884, 0, 0, 0, 0, 0, 0, -7.40485829366544e-05, 0, -8.76498486364369e-05, 0, 0, 0, -2.00006300112506e-05, 0, -0.000318321120421751, -9864873.5625774, -0.000961286000126418, -0.000761479347496079, 0, 0, 0, -2.7529611596004e-05, 0, -0.000313721056840353, -30491427.3753296, -30491427.3755068, -896806.690128406, -0.00269542774448494, -896806.688962867, -0.00162511416388523, 0, 0, -0.000191716022876519, -32285040.7513879, -32285040.7511325, -12555293.6279176, -0.00328731685606248, -12555293.6264808, -0.00214727170675027, -64570081.5017112, -3587226.75597238, -3587226.75253292, -37665880.8797309, -37665880.877857, -5380840.12979303, 0, 0, 0, 0, 0, 0, 2.10006330789238e-05, 0, 2.13622060411038e-05, 0, 0, 0, 7.91304172157364e-06, 0, 0.000104273711757057, 2690420.06251733, 0.000279899363522595, 0.000233589602342461, 0, 0, 0, -1.0102403114998e-05, 0, 0.000131345879094546, 10761680.2500702, 10761680.2501484, 0.000939424537910829, 0.000949960425686489, 0.000567877661546029, 0.00064367433326471, 0, 0, 0, 5.40319852287638e-05, 0, 9.9481676529012e-05, 16142520.3753241, 16142520.3754078, 2690420.06390366, 0.00156071584746612, 2690420.06317269, 0.000791867804017134, 32285040.75048, 0.00283784423717807, 0.00108655947098695, 10761680.2509971, 10761680.2505945, 0.00128897900719546, 32285040.7516755, 32285040.7511257, 10761680.2532583, 0.00329094345857329, 10761680.2512736, 0.00180423794659499, 16142520.3776427}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 1.70815768011882e-08, 0, 0, 0, 0, 0, 0, 0, 4.40256078009569e-08, 0, -3.80730717573871e-07, -15309.0000011927, 0, 0, 0, 0, 0, 0, -8.73283168351454e-07, 0, 4.23899828301654e-06, 0, 0, -9.27393735298534e-07, 304813.125007913, 52214.6250128945, 52214.6250117003, 0, 0, 0, 0, 0, 0, 2.15628515946795e-06, 0, -1.85429220920542e-05, 0, 0, 0, 2.1782815336706e-05, 0, -1.28801943426718e-05, -2160619.31252523, -59049.0000727287, -59049.0000617609, -979639.312568935, -979639.312559727, -118098.000067897, 0, 0, 0, 0, 0, 0, -8.41972890520346e-08, 0, 4.05443610201756e-05, 0, 0, 0, -7.69276107556182e-05, 0, 7.11541001325792e-05, 7085880.00004815, 22143.3752091579, 22143.3751789845, 0, 0, -3.01771230652119e-05, 6244431.75014608, 6244431.75014902, 1040738.62531286, 66430.1252644492, 1040738.6252232, 66430.1252086044, 2081477.25029331, 0, 0, 0, 0, 0, 0, -6.05911269008029e-06, 0, -4.75341136308764e-05, 0, 0, 0, 0.000120140589485675, 0, -0.000136847139770024, -11492411.6250561, -0.0003127231684906, -0.000281216464682684, 0, 0, 0, 0.000137327289623662, 0, -3.91367397283731e-05, -17006112.0001508, -17006112.000213, -365365.68817521, -0.000551227505033811, -365365.687941361, -0.000384646907455283, -5879066.0629844, -5879066.06288117, -0.00047389293924131, -11758132.1254937, -1096097.06340697, -1096097.06308543, 0, 0, 0, 0, 0, 0, 7.79979807219983e-06, 0, 2.84743837004006e-05, 0, 0, 0, -8.77949736108116e-05, 0, 0.000117961798188832, 8968066.87503581, 0.000234346807251469, 0.00022296916496284, 0, 0, 0, -0.000169747257420674, 0, 9.46183082758894e-05, 19729747.1250838, 19729747.1251658, 0.000664369645099812, 0.000526697363710993, 0.000440449452675476, 0.000385631551044606, 0, 0, -5.56421134480073e-05, 12555293.6252997, 12555293.6253401, 1793613.37571308, 0.000643044482821725, 1793613.37544404, 0.000431850449426574, 25110587.2503543, 0.00129833987593687, 0.000606576712103939, 5380840.12599417, 5380840.12568263, 0.00099754219364065, 0, 0, 0, 0, 0, 0, -3.00259371036038e-06, 0, -6.81760786910731e-06, 0, 0, 0, 2.40308210309575e-05, 0, -3.84007541062964e-05, -2690420.06250918, -7.0028661320943e-05, -6.92974273464252e-05, 0, 0, 0, 7.0402616641734e-05, 0, -4.7925428703199e-05, -8071260.18752314, -8071260.18755396, -0.000243229362839958, -0.000184978683730228, -0.000172413416171327, -0.00015781997728998, 0, 0, 0, 5.17511524565221e-05, 0, -7.95783622430248e-06, -8071260.18756133, -8071260.18762704, -0.00035728430895345, -0.000305724018714103, -0.000200216954886506, -0.00016242796841541, -16142520.3750935, -0.000633108870504667, -0.000264591650404965, -2690420.06273204, -2690420.06267623, -0.000270609820290039, -8071260.1878248, -8071260.18776954, -0.000718128556887388, -0.000703847561681325, -0.000338809586924462, -0.000385588374553025, -0.000573061119103501}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.65894010148727e-07, 0, -3.45478120745163e-08, 5248.80000014647, 0, 0, 0, 0, 0, 0, 1.22380025845102e-06, 0, -2.88460028995622e-07, 0, 0, 1.77354462583626e-06, -109568.700001317, -11153.7000010832, -11153.7000014282, 0, 0, 0, 0, 0, 0, -4.04309557085048e-06, 0, 2.28955360004687e-06, 0, 0, 0, -1.26611336467924e-05, 0, -4.15026209208996e-06, 835051.27500536, 5904.90000616801, 5904.90000776806, 220941.675007766, 220941.675009433, 11809.8000038696, 0, 0, 0, 0, 0, 0, 7.20709812599998e-06, 0, -6.12776757881482e-06, 0, 0, 0, 3.58709897814646e-05, 0, -5.13598224818382e-07, -3022570.6875129, -1.85345912032078e-05, -2.42329538415887e-05, 0, 0, 3.47734177318293e-05, -1538964.56251878, -1538964.56252794, -110716.87503033, -1.76058841428139e-05, -110716.875031815, -6.82267597014678e-06, -221433.750037521, 0, 0, 0, 0, 0, 0, -7.15248922628602e-06, 0, 7.96634756615151e-06, 0, 0, 0, -5.07435574009589e-05, 0, 1.30962586432672e-05, 5520343.38751755, 2.77145561124874e-05, 4.17045597900745e-05, 0, 0, 0, -8.00151609671609e-05, 0, -3.81971450843405e-05, 4782969.00002593, 4782969.00004786, 6.66656974404035e-05, 4.55706236814619e-05, 6.03679178023718e-05, 7.52856340715468e-06, 697516.312536023, 697516.312555305, 1.54758788051918e-05, 1395032.62508827, 0.000106683475947389, 6.32286088270554e-05, 0, 0, 0, 0, 0, 0, 3.6943569899896e-06, 0, -5.07749586973862e-06, 0, 0, 0, 3.56848177469289e-05, 0, -1.63483401039824e-05, -4842756.11251207, -2.02213232027831e-05, -3.61157278557843e-05, 0, 0, 0, 7.76539245664603e-05, 0, 1.49560412758554e-05, -6636369.48752363, -6636369.48754289, -6.74446837603961e-05, -4.43155060040274e-05, -6.32013727157617e-05, -1.16063446228114e-05, 0, 0, 6.85275196654757e-05, -1793613.37500855, -1793613.37505342, -4.99876205147159e-05, -5.41389668431771e-05, -4.99014243591724e-05, 1.15225098054278e-05, -3587226.75008579, -0.000156378903996654, -4.44659858601444e-05, -0.000140433240420283, -0.00012053313118506, -0.000111510097965858, 0, 0, 0, 0, 0, 0, -7.68632808654188e-07, 0, 1.26678228968135e-06, 0, 0, 0, -9.89046164272183e-06, 0, 6.28723409645576e-06, 1614252.03750311, 5.91714437166933e-06, 1.20850087439632e-05, 0, 0, 0, -2.78772069659555e-05, 0, 2.92834395262772e-07, 3228504.07501002, 3228504.07501537, 2.5213534356637e-05, 1.36094570007835e-05, 2.7208888126605e-05, 8.97061928497396e-06, 0, 0, 0, -3.42780521714452e-05, 0, -2.02513237161064e-05, 1614252.0374982, 1614252.03752257, 2.78247320942284e-05, 3.29102994789894e-05, 2.02753705170987e-05, -1.4322564970178e-05, 3228504.07503207, 7.58556816321241e-05, 1.27085498475238e-05, 2.54209232915734e-06, 1.67308678447832e-05, 1.41179145090537e-05, 5.17987649214892e-05, 6.36270260280012e-05, 9.0479529930444e-05, 9.09794007051307e-05, 4.79059276714313e-05, 2.5048070408798e-05, 9.55701940841801e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 8.23660162163167e-08, 0, -9.11482806663511e-08, -656.10000005055, 0, 0, 0, 0, 0, 0, -8.21773675833777e-07, 0, 9.81508462858247e-07, 0, 0, -3.55743029101527e-07, 12957.9750001816, 6561.00000004837, 656.100000997236, 0, 0, 0, 0, 0, 0, 3.08922832249993e-06, 0, -4.11905006527115e-06, 0, 0, 0, 6.20319352421697e-06, 0, -2.39371476519818e-06, -89803.6875002478, -5904.89999979677, -7.36874340182343e-06, -128923.650000037, -12301.8750030585, -5904.90000118141, 0, 0, 0, 0, 0, 0, -5.87886975895333e-06, 0, 8.76948686022182e-06, 0, 0, 0, -2.21027822808819e-05, 0, 1.38160895107011e-05, 276792.187500557, 5.51255328966439e-07, 2.35478330729062e-05, 0, 0, -9.20022795984844e-06, 885734.999996195, 77501.812503374, 110716.875002297, -8.81773683676593e-06, 1.89419827272649e-05, 1.76327592217718e-05, 110716.875003599, 0, 0, 0, 0, 0, 0, 6.01962571720116e-06, 0, -1.01380141479392e-05, 0, 0, 0, 3.48365814148418e-05, 0, -2.59328793198167e-05, -378651.712501526, -2.8380229081333e-06, -3.74889876417966e-05, 0, 0, 0, 3.81864433382602e-05, 0, -4.77924680412643e-06, -2690420.06249015, -199290.37500196, -1.33619553672726e-05, 2.50897588407473e-05, -4.52201678964428e-05, -5.22620707506454e-05, -697516.312494781, -1.94002219722186e-05, 2.99714326996173e-06, -697516.312491307, 6.57820560895516e-06, -4.25468408614249e-05, 0, 0, 0, 0, 0, 0, -3.15352707028777e-06, 0, 6.07686284015441e-06, 0, 0, 0, -2.58173358899805e-05, 0, 2.1463298595796e-05, 179361.337501736, 3.49795432252354e-06, 2.93425833052068e-05, 0, 0, 0, -4.57182604770276e-05, 0, 1.76649272481538e-05, 3587226.74999187, 179361.337501825, 1.9868158456141e-05, -2.50764038276526e-05, 4.72882137763341e-05, 6.03924748230891e-05, 0, 0, -2.03801814335195e-05, 1793613.37498415, 8.70272681050977e-06, 8.17836951260318e-06, -3.152517785171e-05, 3.07850204986418e-05, 2.27212098781032e-05, 1793613.37497731, -5.6711754154944e-06, 7.93854121626334e-05, -2.48020176240071e-05, 2.9074505145105e-05, 2.17493602377792e-05, 0, 0, 0, 0, 0, 0, 6.60912377522019e-07, 0, -1.48077007889522e-06, 0, 0, 0, 7.32459145623756e-06, 0, -6.68153713665405e-06, -6.50791358801886e-07, -1.40701230644492e-06, -8.98099445051152e-06, 0, 0, 0, 1.81424338622675e-05, 0, -9.1842464603576e-06, -1614252.03749804, -1.22257775444071e-06, -9.0963756860482e-06, 8.90249694337771e-06, -1.84086711238859e-05, -2.47254849892648e-05, 0, 0, 0, 1.62959217621364e-05, 0, 8.33914001808169e-07, -1614252.03749152, -2.02032639285571e-06, -1.01243563966539e-05, 1.98618692912475e-05, -1.57602132087576e-05, -1.84679192781797e-05, -1614252.03748894, -1.3227568592301e-06, -4.26389097469798e-05, 2.6003704319118e-06, -6.28878833909925e-06, 7.06431071939573e-06, 2.37665094132736e-05, -1.22392552584555e-06, 8.15643773444355e-06, -7.20402185854587e-08, -3.38443199345352e-05, -3.01819570070375e-05, -6.84458044817899e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 1.61067095791678e-08, 0, 0, 0, 0, 0, 0, 0, -6.07998079930599e-08, 0, -3.54972556733546e-07, -546.750000160459, 0, 0, 0, 0, 0, 0, -1.26565388203213e-07, 0, 2.88339905852715e-06, 0, 0, 1.59227995658681e-06, 9568.12500061994, 7927.87500111327, 546.750002512946, 0, 0, 0, 0, 0, 0, 1.1450619891042e-06, 0, -1.12055986001435e-05, 0, 0, 0, -8.7953889973544e-07, 0, -1.32831701367939e-05, -53308.1250012864, -29524.5000042451, -1.75318461532757e-05, -138191.062505494, -9021.37500590277, -7381.125010953, 0, 0, 0, 0, 0, 0, -2.66035071122154e-06, 0, 2.35780020994902e-05, 0, 0, 0, -7.48754910597442e-06, 0, 4.4763773428691e-05, 110716.875002289, 22143.3750107455, 5.67418275454131e-05, 0, 0, 2.00208098513156e-05, 763946.437509529, 44286.7500040791, 509297.625024278, 22143.3750143489, 3.41700484468393e-05, 6.47930028705348e-05, 121788.562531394, 0, 0, 0, 0, 0, 0, 2.86007824221431e-06, 0, -2.75521867241375e-05, 0, 0, 0, 1.82272490075306e-05, 0, -7.34954772858736e-05, -66430.1250036495, -1.63362848491022e-05, -9.23305874576784e-05, 0, 0, 0, -5.47823505791852e-06, 0, -6.18321753587572e-05, -1561107.93750652, -66430.1249974536, -365365.687552794, -4.74744513075453e-06, -8.41687906465486e-05, -0.00016806789607901, -2756850.18753999, -2.3165430969487e-05, -6.94372749462106e-05, -597871.125027171, -365365.687565264, -0.00012625916906969, 0, 0, 0, 0, 0, 0, -1.4312536770165e-06, 0, 1.67951906629725e-05, 0, 0, 0, -1.61743438822679e-05, 0, 5.88396495556679e-05, 3.40029473862688e-06, 1.29138907597202e-05, 7.35711431276443e-05, 0, 0, 0, -1.27181569847436e-05, 0, 7.19337245344897e-05, 896806.687502082, -3.31348120089382e-06, 5.29666829424177e-05, -3.00160483211882e-06, 9.23353655151458e-05, 0.000190063470073905, 0, 0, 3.3098160419282e-05, 5380840.12502367, 3.18218502688206e-07, 1793613.37506601, 1.50864497518716e-05, 4.21697707724013e-05, 0.000135723516865325, 896806.687495869, 7.17309434687831e-05, 0.000207140825823744, 1793613.37506099, 8.81701615159786e-05, 0.000131155898361127, 0, 0, 0, 0, 0, 0, 2.57094458954348e-07, 0, -4.15985881740819e-06, 0, 0, 0, 5.07150487385564e-06, 0, -1.8358125593381e-05, -1.22026230289935e-06, -4.05924883100675e-06, -2.2802909550806e-05, 0, 0, 0, 8.5297766006797e-06, 0, -2.95163389845957e-05, -6.80569315753388e-07, 6.21455197859163e-07, -1.96264535888782e-05, 1.67797325627848e-06, -3.72652381149915e-05, -7.8129887465845e-05, 0, 0, 0, -3.58870474516011e-06, 0, -2.26221114261306e-05, -2690420.06250415, 3.05844436094539e-06, -3.4395452453177e-05, 8.05980662263066e-06, -2.36527774274814e-05, -8.31721729193064e-05, 9.17099340602793e-06, -3.26094137051497e-05, -0.000113306911418117, -2690420.06252346, -3.64222638824006e-06, -2.6916824329333e-05, -2690420.06250828, -1.27942725927327e-05, -4.77034207943723e-05, -4.62107937309461e-05, -7.65143466350586e-05, -0.000103292143988943, -6.06630521704623e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 5.07992740953484e-08, 0, 0, 0, 0, 0, 0, 0, -5.37190051087678e-07, 0, -8.30076332142735e-07, -546.750000533405, 0, 0, 0, 0, 0, 0, 2.9130650026357e-06, 0, 5.49926989431091e-06, 0, 0, 6.47004605648247e-06, 7927.87500252763, 9568.12500537537, 546.750006873154, 0, 0, 0, 0, 0, 0, -9.39350643701017e-06, 0, -1.93334232374726e-05, 0, 0, 0, -2.52754088163005e-05, 0, -3.17310847727227e-05, -29524.5000070195, -53308.1250248188, -3.70775156651135e-05, -138191.06252421, -7381.12502239157, -9021.37504901091, 0, 0, 0, 0, 0, 0, 1.79524525927351e-05, 0, 3.84036122030879e-05, 0, 0, 0, 6.04456991361436e-05, 0, 8.5272037581533e-05, 22143.3750124149, 110716.875062085, 0.000105975412749589, 0, 0, 8.19840890549273e-05, 509297.625055534, 22143.3750385782, 763946.437590273, 44286.7501626311, 8.47993773420755e-05, 0.000177542328168107, 121788.562661957, 0, 0, 0, 0, 0, 0, -1.97620386516807e-05, 0, -4.30338266472163e-05, 0, 0, 0, -8.45992301895714e-05, 0, -0.000126981453324282, -1.35865598153604e-05, -66430.1250845085, -0.000162906418570432, 0, 0, 0, -0.000125699595107077, 0, -0.000144160374230025, -365365.687571227, -3.95986295939694e-05, -1561107.93766299, -66430.1252912334, -0.000175343987984213, -0.000342045771928861, -2756850.18764866, -8.62946774311735e-05, -0.000364220203236913, -365365.687759847, -597871.125446529, -0.000379526902942252, 0, 0, 0, 0, 0, 0, 1.15446560151523e-05, 0, 2.53241442021244e-05, 0, 0, 0, 6.19994796730127e-05, 0, 9.65473541092176e-05, 8.20757132198038e-06, 5.88420170243255e-05, 0.000125519882872538, 0, 0, 0, 0.000108328132727363, 0, 0.000134770890542821, 4.95467819850173e-05, 2.2739070847491e-05, 896806.687641354, 0.000260203827468572, 0.000182014270711482, 0.00033988087056957, 0, 0, 0.000122357656152503, 1793613.3751172, 4.09610291311338e-05, 5380840.12516886, 0.000390743168051753, 0.000106116565147746, 0.000387089928517112, 0.000205949996882849, 896806.688034697, 0.000428639730023531, 1793613.37550199, 0.000351521911959753, 0.000628719461168659, 0, 0, 0, 0, 0, 0, -2.76857600549794e-06, 0, -6.08053886315537e-06, 0, 0, 0, -1.82137662604784e-05, 0, -2.90452307951971e-05, -2.07155192561485e-06, -1.64194309600612e-05, -3.78492424133192e-05, 0, 0, 0, -3.86661881417329e-05, 0, -5.06230500870475e-05, -1.4521572402984e-05, -5.6510017613394e-06, -4.7203832469489e-05, -8.99600069487903e-05, -7.21252241755232e-05, -0.000132079254937855, 0, 0, 0, -4.54756614128453e-05, 0, -4.99129521992031e-05, -3.7185176163989e-05, -7.14207043602037e-06, -2690420.06257024, -0.000168322033306139, -5.53156012786951e-05, -0.000170849635826604, -6.56621557583839e-05, -0.000229431742129806, -0.000199036254019151, -2690420.06256141, -1.86334135959232e-05, -0.000172553685556739, -0.000199408595419329, -0.000123100003451456, -2690420.06279568, -0.000362973435869101, -0.000166908562496878, -0.000294466739233174, -0.000339404998245988}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 4.14999469711475e-08, 0, 0, 0, 0, 0, 0, 0, -4.16942420909127e-07, 0, -7.33680625274822e-07, -656.100000464447, 0, 0, 0, 0, 0, 0, 2.0595054803828e-06, 0, 4.87902537269883e-06, 0, 0, 6.02271807734462e-06, 6561.00000206142, 12957.9750042052, 656.100006731595, 0, 0, 0, 0, 0, 0, -6.21815616299589e-06, 0, -1.64669875350569e-05, 0, 0, 0, -2.25854341069776e-05, 0, -3.22944086066146e-05, -5904.90000586511, -89803.6875166853, -3.78198146303446e-05, -128923.650017807, -5904.90002039533, -12301.8750484929, 0, 0, 0, 0, 0, 0, 1.17850393629468e-05, 0, 3.11509294277307e-05, 0, 0, 0, 4.80102524420807e-05, 0, 8.48098674240962e-05, 1.12551528783427e-05, 276792.187537011, 0.000107933556488167, 0, 0, 9.18789346809937e-05, 110716.875043134, 3.31607715062628e-05, 885735.00005397, 77501.8126454151, 7.96372092074446e-05, 0.000200194253390899, 110716.875153285, 0, 0, 0, 0, 0, 0, -1.35917844536877e-05, 0, -3.34929211387705e-05, 0, 0, 0, -5.85716169070926e-05, 0, -0.000117687370091896, -1.27751964235804e-05, -378651.712548218, -0.000163570088175529, 0, 0, 0, -0.000138493179137733, 0, -0.000173030874681068, -6.09924168572763e-05, -3.79376867243275e-05, -2690420.06257638, -199290.375223733, -0.00016872084457332, -0.000397330117832009, -697516.312593144, -5.96589658320441e-05, -0.000420901185193096, -0.000243925680021428, -697516.312880382, -0.000417179464670557, 0, 0, 0, 0, 0, 0, 8.63795114478547e-06, 0, 1.91066705430703e-05, 0, 0, 0, 3.84861332986384e-05, 0, 8.3169240391166e-05, 7.47592267757566e-06, 179361.337534758, 0.000124033058889918, 0, 0, 0, 0.000105345383980453, 0, 0.000153069900399435, 4.46867804321221e-05, 2.81304159743833e-05, 3587226.75005404, 179361.337671305, 0.000181105583466976, 0.000383700956544673, 0, 0, 0.000167729791876974, 8.44830384992773e-05, 1.60634583539588e-05, 1793613.37507636, 0.000416137100689708, 6.73333339930804e-05, 0.0005058586771543, 0.000203537150097504, 1793613.37538345, 0.000504814267357618, 0.000430354862842646, 0.000360429367847791, 0.000685516940044869, 0, 0, 0, 0, 0, 0, -2.29670904549624e-06, 0, -4.4846701365e-06, 0, 0, 0, -1.06490485857735e-05, 0, -2.35937728693662e-05, -1.72109742203109e-06, -1.06396600863366e-05, -3.68335603640335e-05, 0, 0, 0, -3.17269628482277e-05, 0, -5.18697220689417e-05, -1.28387035594855e-05, -8.83959256446923e-06, -1614252.03751646, -5.2251474223816e-05, -7.37096370953009e-05, -0.000143121005851954, 0, 0, 0, -6.20499469291378e-05, 0, -7.14377935160309e-05, -3.04556272141329e-05, -2.62019576313009e-06, -1614252.03752091, -0.000155327344475746, -4.03246502471224e-05, -0.000224440163031889, -6.93952097200034e-05, -1614252.03763496, -0.000238435778014006, -3.2187440571205e-05, 5.2643984304444e-06, -0.000235053253323886, -0.000180615425337441, -0.000121141191100204, -0.000213926511982031, -0.000359010679650564, -0.00017979907066604, -0.000376341593116306, -0.0003539589428476}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -5.21322049944027e-08, 0, 0, 0, 0, 0, 0, 0, 7.28424384397044e-07, 0, 6.35102112922002e-07, 5248.80000153207, 0, 0, 0, 0, 0, 0, -5.83985812097383e-06, 0, -3.34856717831536e-06, 0, 0, -5.21833324972668e-06, -11153.7000078777, -109568.700017587, -11153.7000186678, 0, 0, 0, 0, 0, 0, 2.367182562056e-05, 0, 9.85662746987427e-06, 0, 0, 0, 3.44336445261224e-05, 0, 1.26519316861707e-05, 5904.9000199786, 835051.275087556, 5904.90009489717, 220941.675086389, 11809.800070089, 220941.67515515, 0, 0, 0, 0, 0, 0, -5.1712329561959e-05, 0, -1.79810918660557e-05, 0, 0, 0, -0.000112607199990733, 0, -6.40927691175193e-06, -2.90226623529442e-05, -3022570.68772587, -0.000256148340610313, 0, 0, -7.13739389877869e-05, -110716.87520053, -0.000113399518654215, -1538964.56282737, -1538964.56307952, -0.000260123560219781, -110716.875525669, -221433.750600622, 0, 0, 0, 0, 0, 0, 6.28809826025872e-05, 0, 2.01309235781863e-05, 0, 0, 0, 0.000177458237526808, 0, -1.34354284073594e-05, 2.2761524514021e-05, 5520343.38781677, 0.000379414791813639, 0, 0, 0, 0.000204751589126827, 0, 3.52001576592658e-05, 0.000251011556683174, 9.7844537993749e-05, 4782969.00057507, 4782969.00106453, 0.000520886418535581, 0.00088355010353364, 697516.313057182, 0.000244920024665583, 697516.313851593, 0.000961136271164359, 1395032.62673029, 0.00146712028159537, 0, 0, 0, 0, 0, 0, -4.01488235897359e-05, 0, -1.23814563169757e-05, 0, 0, 0, -0.000135455533108728, 0, 1.56793281080131e-05, -7.65558342598735e-06, -4842756.11273075, -0.0002863804153611, 0, 0, 0, -0.000225455759529289, 0, 3.93169408983382e-05, -0.000161229960818731, -4.37428271904631e-05, -6636369.48799072, -6636369.48842771, -0.000533937093950061, -0.000763086241507814, 0, 0, -0.000134075081221016, -0.000462084756748101, -8.02868719458416e-05, -1793613.37561032, -1793613.37663351, -0.000295897835691001, -0.00129961391620009, -0.000721206542575545, -3587226.75200374, -0.00159055057484938, -0.00192998879680823, -0.0014154524032764, -0.0027528584015365, 0, 0, 0, 0, 0, 0, 1.04739191614261e-05, 0, 3.14031330752832e-06, 0, 0, 0, 4.07877538767718e-05, 0, -4.06363287867348e-06, 2.99019768325914e-07, 1614252.03756833, 8.52557227584197e-05, 0, 0, 0, 8.44898518907451e-05, 0, -3.04735002156275e-05, 4.14610467544518e-05, 7.22872165607857e-06, 3228504.07516686, 3228504.07530949, 0.000211083883122664, 0.000269540478018487, 0, 0, 0, 9.07568173964291e-05, 0, 1.33155863655219e-05, 0.000149511133451732, 2.21639236164794e-06, 1614252.03774043, 1614252.03820643, 0.000160125058614398, 0.000458839372916354, 0.000207944966742277, 3228504.07581083, 0.000678816119535504, 0.000244072904888315, 2.68120453869853e-05, 0.00075892630923254, 0.000754519966558903, 0.000470241460517661, 0.00108886053170967, 0.00162029771321226, 0.000698253963988997, 0.00134822148057161, 0.00148757330315271}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 2.81480208614402e-07, 0, 0, 0, 0, 0, 0, 0, -3.41405550651187e-06, 0, -4.04977633277477e-06, -15309.0000064849, 0, 0, 0, 0, 0, 0, 2.3393535744181e-05, 0, 2.53278390410906e-05, 0, 0, 3.1362145185995e-05, 52214.6250356161, 304813.125072688, 52214.6250775966, 0, 0, 0, 0, 0, 0, -8.7482831522477e-05, 0, -8.70648631399021e-05, 0, 0, 0, -0.000158172440190735, 0, -0.000122774367442056, -59049.0001000536, -2160619.31284536, -59049.0003928039, -979639.31287126, -118098.000304319, -979639.313139969, 0, 0, 0, 0, 0, 0, 0.000182913941857932, 0, 0.000174893184964769, 0, 0, 0, 0.000453713115918552, 0, 0.000267862385092013, 22143.3751640391, 7085880.00085494, 22143.3760666544, 0, 0, 0.000414914137134339, 1040738.62589628, 66430.1255608549, 6244431.75135983, 6244431.75228124, 66430.126089285, 1040738.62719901, 2081477.25251451, 0, 0, 0, 0, 0, 0, -0.000215880133037234, 0, -0.000202606567245239, 0, 0, 0, -0.00068118313116029, 0, -0.000354542491075873, -0.000154390319248611, -11492411.6261572, -0.0015899652900788, 0, 0, 0, -0.000880237936309427, 0, -0.00050926677495398, -365365.688652804, -0.000577206837092525, -17006112.0023228, -17006112.0040419, -0.00212164442901708, -365365.691343277, -5879066.06490538, -0.00120305354938449, -5879066.06783588, -1096097.06672607, -11758132.1319915, -1096097.06850766, 0, 0, 0, 0, 0, 0, 0.000134741803723388, 0, 0.000124498858684603, 0, 0, 0, 0.000508185014473678, 0, 0.000265476603220699, 7.55285505729867e-05, 8968066.87581622, 0.00120576637115999, 0, 0, 0, 0.000885582967574304, 0, 0.000307014988546627, 0.000758721328617515, 0.000319708860627189, 19729747.1269222, 19729747.1284464, 0.00212857572848113, 0.00345963485698992, 0, 0, 0.000694512611069658, 1793613.37702732, 0.000601285764203307, 12555293.6275913, 12555293.6311062, 0.00138321907698258, 1793613.38036864, 0.0033374576680975, 25110587.2579181, 0.00650420420102476, 5380840.13312346, 0.00607507786779761, 5380840.13579954, 0, 0, 0, 0, 0, 0, -3.45535436799022e-05, 0, -3.12784942317328e-05, 0, 0, 0, -0.000150093082504418, 0, -8.39814398101008e-05, -1.44609390276587e-05, -2690420.062735, -0.000360690276462626, 0, 0, 0, -0.000326208865548262, 0, -8.12177754941333e-05, -0.000200750272023993, -7.26875460786957e-05, -8071260.18813035, -8071260.18863181, -0.000829229277674428, -0.00125402809888099, 0, 0, 0, -0.000363747430650717, 0, -0.000181604458614603, -0.000659170725521044, -0.000115625583640785, -8071260.18850284, -8071260.19007165, -0.000694732177454264, -0.00206679758258813, -0.00101300420298589, -16142520.3781587, -0.00279207613117549, -2690420.06355468, -0.000253809122851228, -2690420.06537084, -0.00328450643142778, -0.00216475850993461, -8071260.19200922, -8071260.19364996, -0.00295494354898438, -0.00528380963062152, -0.00603418717971231}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -5.22773566326533e-07, 0, 0, 0, 0, 0, 0, 0, 6.18921194141227e-06, 0, 7.53617656072494e-06, 30618.000014612, 0, 0, 0, 0, 0, 0, -4.28738887446453e-05, 0, -4.93805828295593e-05, 0, 0, -5.33694706215875e-05, -150356.250085979, -563699.250168368, -150356.250167028, 0, 0, 0, 0, 0, 0, 0.000162422746633621, 0, 0.000179605711577502, 0, 0, 0, 0.000266436217949038, 0, 0.000210376891480674, 274741.875252954, 3544580.2507949, 274741.87581961, 2579293.12590165, 549483.750702542, 2579293.12643879, 0, 0, 0, 0, 0, 0, -0.00034284170625812, 0, -0.000378262310396652, 0, 0, 0, -0.000784468705263128, 0, -0.000498706981375461, -221433.750428246, -10019877.1894396, -221433.752191844, 0, 0, -0.00065208699130277, -4373316.56470195, -664301.251410958, -14171760.0033112, -14171760.0051182, -664301.252381875, -4373316.56735256, -8746633.13092301, 0, 0, 0, 0, 0, 0, 0.000406668037561839, 0, 0.000453144790013409, 0, 0, 0, 0.00121207855706842, 0, 0.000740716297411077, 66430.1254175247, 14182831.6900785, 66430.1282495401, 0, 0, 0, 0.00140465134383675, 0, 0.000783038545061736, 3255076.12781526, 265720.501575849, 31487879.2556784, 31487879.2589731, 265720.504422146, 3255076.13341931, 20493693.5684243, 398580.752932517, 20493693.5743017, 9765228.38528428, 40987387.141402, 9765228.38875718, 0, 0, 0, 0, 0, 0, -0.000254395977489487, 0, -0.000285006745530873, 0, 0, 0, -0.000922497369462875, 0, -0.000604514140872863, -0.000215959390829454, -9864873.56428114, -0.00246248402759646, 0, 0, 0, -0.0014859748631483, 0, -0.000533984652846813, -896806.68933385, -0.000932988515477005, -30491427.3796919, -30491427.3825815, -0.00429999601423623, -896806.695068851, 0, 0, -0.000993750139800655, -12555293.6299224, -0.00175674613017562, -32285040.7564962, -32285040.7633082, -0.00313758607263435, -12555293.636903, -3587226.75831327, -64570081.518553, -3587226.76452748, -37665880.8944585, -5380840.13955199, -37665880.8996867, 0, 0, 0, 0, 0, 0, 6.53436730870638e-05, 0, 7.28791974274029e-05, 0, 0, 0, 0.000275032004263783, 0, 0.00019971520409887, 4.55269061382116e-05, 2690420.06300176, 0.000737838135567609, 0, 0, 0, 0.000574219617992054, 0, 0.000182596138703769, 0.000481559749489992, 0.000225087355384566, 10761680.2515208, 10761680.2524663, 0.00164719123862725, 0.00273991406411199, 0, 0, 0, 0.000547888163117453, 0, 0.000239136844415031, 2690420.0640621, 0.00043013579890979, 16142520.3775618, 16142520.3805565, 0.0014636351771757, 2690420.06716129, 0.00256724238468659, 32285040.7574167, 0.0060986282420815, 10761680.2526134, 0.000734195147181441, 10761680.2562183, 10761680.2579367, 0.00544481215393745, 32285040.7608222, 32285040.7639479, 0.00690853501484137, 10761680.2619241, 16142520.3891302}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 4.77065200364737e-07, 0, 0, 0, 0, 0, 0, 0, -5.3612369068764e-06, 0, -7.11958087649626e-06, -45927.0000192233, 0, 0, 0, 0, 0, 0, 3.76316325953242e-05, 0, 4.94582100469915e-05, 0, 0, 4.49340144374148e-05, 308202.975118068, 721545.975228033, 308202.975214471, 0, 0, 0, 0, 0, 0, -0.000146133465846527, 0, -0.000187855812119131, 0, 0, 0, -0.000215655837354482, 0, -0.000190573872611183, -818074.687851041, -3591901.46356727, -818074.688516305, -4409976.1512578, -1636149.375951, -4409976.15196047, 0, 0, 0, 0, 0, 0, 0.000315604096957243, 0, 0.000406984878699381, 0, 0, 0, 0.000645310332664023, 0, 0.000499236437823061, 1073953.68809196, 8436625.87755334, 1073953.69013882, 0, 0, 0.000528006140776667, 10584533.2530403, 3221861.06451002, 17947205.4421467, 17947205.4444972, 3221861.06556577, 10584533.2565778, 21169066.5082788, 0, 0, 0, 0, 0, 0, -0.000381399331469147, 0, -0.000495738316699015, 0, 0, 0, -0.00102180549035606, 0, -0.000806766661878318, -697516.313072902, -10363099.5033124, -697516.316335044, 0, 0, 0, -0.00112705584036303, 0, -0.000665728955972296, -12455648.4412888, -2790065.2523452, -31786814.8204978, -31786814.8245971, -2790065.2553727, -12455648.4487625, -33181847.4457481, -4185097.87905385, -33181847.4538507, -37366945.3269958, -66363694.8980514, -37366945.3314643, 0, 0, 0, 0, 0, 0, 0.000242383183137803, 0, 0.000314806320534647, 0, 0, 0, 0.000790126024747484, 0, 0.000688838110746488, 179361.337794786, 6457008.15222638, 179361.340369273, 0, 0, 0, 0.00124436613623544, 0, 0.000538550084102033, 7174453.50239739, 896806.688934838, 26007393.9440918, 26007393.9475356, 896806.692508782, 7174453.50994728, 0, 0, 0.000742500375266782, 26904200.6316416, 1793613.37769702, 39459494.259197, 39459494.2682713, 1793613.37900426, 26904200.6416709, 28697814.0117252, 78918988.5261472, 28697814.0195446, 80712601.902351, 43046721.0204599, 80712601.9095341, 0, 0, 0, 0, 0, 0, -6.31836220308746e-05, 0, -8.10005748649692e-05, 0, 0, 0, -0.00023692397266762, 0, -0.000231047561250732, -6.20790498248388e-05, -1614252.03810998, -0.000852546589476989, 0, 0, 0, -0.000500434187518708, 0, -0.00022183865488062, -1614252.03811317, -0.000355075590710778, -8071260.18961255, -8071260.1906956, -0.00187295647075292, -1614252.04103407, 0, 0, 0, -0.000426936694593993, 0, -0.000172058880199717, -8071260.18952467, -0.000741945114113467, -16142520.3786849, -16142520.382517, -0.00172009119725187, -8071260.1940651, -8071260.19110778, -32285040.7604708, -8071260.19546072, -16142520.3785978, -0.00106773581309478, -16142520.3836968, -32285040.7610395, -16142520.3828185, -48427561.1403282, -48427561.1445506, -16142520.3845104, -32285040.7665986, -48427561.1448118}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -4.49793914550294e-07, 0, 0, 0, 0, 0, 0, 0, 6.30913945322012e-06, 0, 6.90052485509353e-06, -45927.0000098608, 0, 0, 0, 0, 0, 0, -3.38322778156788e-05, 0, -4.51663372298577e-05, 0, 0, -7.97351074622321e-05, 308202.975076066, 308202.975110953, 721545.975113906, 0, 0, 0, 0, 0, 0, 9.8156043606667e-05, 0, 0.000149251295510104, 0, 0, 0, 0.00032604206192002, 0, 0.000447336195572797, -818074.68774607, -818074.687932004, -3591901.46297786, -1636149.37575717, -4409976.15067282, -4409976.15110685, 0, 0, 0, 0, 0, 0, -0.000169989054534346, 0, -0.000269430090279586, 0, 0, 0, -0.000667485878756809, 0, -0.00123050008361115, 1073953.68792276, 1073953.68830553, 8436625.87605741, 0, 0, -0.00145442807844041, 3221861.06447211, 10584533.2517018, 3221861.06495117, 10584533.2536769, 17947205.4395591, 17947205.441087, 21169066.5056337, 0, 0, 0, 0, 0, 0, 0.000175044878627768, 0, 0.00027248043071844, 0, 0, 0, 0.000764060728752776, 0, 0.0017156786877516, -697516.312897699, -697516.313247356, -10363099.5013293, 0, 0, 0, 0.00212289794736428, 0, 0.00308981425346047, -2790065.25252277, -12455648.4397377, -2790065.25363621, -12455648.4432257, -31786814.8156572, -31786814.8179988, -4185097.87973095, -33181847.4411205, -33181847.4462919, -37366945.3233935, -37366945.3268066, -66363694.8880796, 0, 0, 0, 0, 0, 0, -9.74709503758161e-05, 0, -0.000145831686107796, 0, 0, 0, -0.000483925582088262, 0, -0.00117739175180701, 179361.33769143, 179361.337808324, 6457008.15088011, 0, 0, 0, -0.00144333190753891, 0, -0.00293832493497255, 896806.689085644, 7174453.50146919, 896806.690023585, 7174453.5041719, 26007393.9400204, 26007393.9417621, 0, 0, -0.00307805401653836, 1793613.37891631, 26904200.6280673, 1793613.37967234, 26904200.6343656, 39459494.253349, 39459494.2582897, 28697814.0094625, 28697814.0148433, 78918988.5128569, 43046721.0182663, 80712601.8912675, 80712601.8972889, 0, 0, 0, 0, 0, 0, 2.22487184489846e-05, 0, 3.22581638175229e-05, 0, 0, 0, 0.000133397013471783, 0, 0.000318072525984335, -3.61881640807019e-05, -3.63715397700524e-05, -1614252.03773456, 0, 0, 0, 0.000383337664868895, 0, 0.00101779413801133, -0.000392774678931581, -1614252.03787847, -0.000655648054187324, -1614252.03861779, -8071260.18832092, -8071260.18884341, 0, 0, 0, 0.00104826297539614, 0, 0.00150802208962937, -0.00118752022360198, -8071260.18850262, -0.0016563113043203, -8071260.19115351, -16142520.3762203, -16142520.3778806, -8071260.19055314, -8071260.1929373, -32285040.7547936, -0.0018900921345417, -16142520.3763201, -16142520.3792247, -16142520.3827741, -32285040.7569768, -16142520.3843156, -32285040.7618081, -48427561.1325896, -48427561.1355762, -48427561.1389616}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 4.31165755139222e-07, 0, 0, 0, 0, 0, 0, 0, -7.04680662648282e-06, 0, -5.85939577070607e-06, 30618.0000031618, 0, 0, 0, 0, 0, 0, 4.30261339827704e-05, 0, 3.74931312682508e-05, 0, 0, 7.77983992354784e-05, -150356.250026976, -150356.2500168, -563699.250057781, 0, 0, 0, 0, 0, 0, -0.000137857871466118, 0, -0.00012370277130925, 0, 0, 0, -0.000357379052786401, 0, -0.000419303750304183, 274741.875091009, 274741.874951606, 3544580.25028782, 549483.750206654, 2579293.12532986, 2579293.12543414, 0, 0, 0, 0, 0, 0, 0.000258879350910454, 0, 0.000225052270673334, 0, 0, 0, 0.000794152937595154, 0, 0.00112862603166476, -221433.750160678, -221433.74961564, -10019877.1881922, 0, 0, 0.0015188706778681, -664301.250656491, -4373316.56329719, -664301.250262613, -4373316.56340599, -14171760.0011985, -14171760.0017519, -8746633.12748376, 0, 0, 0, 0, 0, 0, -0.000286469527049899, 0, -0.000230957876613653, 0, 0, 0, -0.000959722118185979, 0, -0.00155326068146822, 66430.1251530753, 66430.1241903621, 14182831.6883942, 0, 0, 0, -0.00240846095279858, 0, -0.00312014336521049, 265720.500950501, 3255076.12600324, 265720.499654458, 3255076.12563997, 31487879.2520656, 31487879.2529863, 398580.750912338, 20493693.5644313, 20493693.5655262, 9765228.38013208, 9765228.37975783, 40987387.1319829, 0, 0, 0, 0, 0, 0, 0.000171583663682368, 0, 0.000126161426398879, 0, 0, 0, 0.000624650978390544, 0, 0.00105748532778872, -7.4586961735258e-05, 0.000733400629209415, -9864873.56308582, 0, 0, 0, 0.00175683662630703, 0, 0.00290238089100388, -0.000649262619218405, -896806.688125097, 0.0008758203999649, -896806.687369066, -30491427.3767625, -30491427.3774557, 0, 0, 0.00334064425900927, -0.000981129625950223, -12555293.6264752, -0.000100082291693816, -12555293.6272227, -32285040.7519965, -32285040.7530934, -3587226.75469126, -3587226.75330853, -64570081.5077601, -5380840.13195188, -37665880.8839976, -37665880.8845747, 0, 0, 0, 0, 0, 0, -4.25520665142491e-05, 0, -2.86258237695443e-05, 0, 0, 0, -0.000173404003136453, 0, -0.000284243295819701, 1.45580603673085e-05, -0.000245864703209168, 2690420.06265066, 0, 0, 0, -0.000494161400561663, 0, -0.000991285262606598, 0.000171258572070395, 0.000151378620259828, -0.00045293975379273, -0.000254704920748632, 10761680.2505878, 10761680.2507955, 0, 0, 0, -0.00121277916904525, 0, -0.00158511396838272, 0.000349248696641937, 2690420.06293049, -0.000281182113054958, 2690420.06311149, 16142520.3757932, 16142520.3761016, 0.00155890817745728, 0.000609249281678107, 32285040.7531372, 0.000161328809335225, 10761680.2506758, 10761680.2510092, 0.00328533894454919, 10761680.2539499, 0.00257941009170558, 10761680.2539861, 32285040.7547211, 32285040.7551863, 16142520.381545}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -1.4002000883778e-07, 0, 0, 0, 0, 0, 0, 0, 2.94588194603556e-06, 0, 1.65670344895955e-06, -15309.0000002755, 0, 0, 0, 0, 0, 0, -1.85125345724257e-05, 0, -9.92128644652572e-06, 0, 0, -3.29949505638601e-05, 52214.6249990047, 52214.6249857876, 304813.125023684, 0, 0, 0, 0, 0, 0, 5.90431313751001e-05, 0, 2.8138964045146e-05, 0, 0, 0, 0.000157430909718418, 0, 0.000187997733768039, -59048.9999914275, -59048.9998602622, -2160619.31266789, -118097.999941386, -979639.312577876, -979639.312565579, 0, 0, 0, 0, 0, 0, -0.000112241900356006, 0, -3.87174891112878e-05, 0, 0, 0, -0.000327985906889583, 0, -0.000507685598192193, 22143.3749775308, 22143.3745393055, 7085880.00050469, 0, 0, -0.000745154018796321, 66430.1249129696, 1040738.62508731, 66430.1244947208, 1040738.6247518, 6244431.75049254, 6244431.75057075, 2081477.25033472, 0, 0, 0, 0, 0, 0, 0.000128775997488692, 0, 2.42889311803622e-05, 0, 0, 0, 0.000358158798947447, 0, 0.000675036376910881, 3.01954059195696e-05, 0.000728144002500051, -11492411.6257618, 0, 0, 0, 0.00116000602358319, 0, 0.0016030783052256, 5.06145693909792e-05, -365365.687500189, 0.0012992478112581, -365365.686578878, -17006112.0011763, -17006112.0013565, 0.000711939522844998, -5879066.06297205, -5879066.06257416, -1096097.06326028, -1096097.06201412, -11758132.1269662, 0, 0, 0, 0, 0, 0, -8.11086577906861e-05, 0, -4.21836660548105e-06, 0, 0, 0, -0.000213534406986898, 0, -0.000432062026942081, -1.93693916394709e-05, -0.000566241222833884, 8968066.87556268, 0, 0, 0, -0.00078446134910275, 0, -0.00150263436491505, -1.17176351994286e-05, -7.3064107410647e-05, -0.00136288915692563, -0.00102341095874868, 19729747.1262362, 19729747.1264056, 0, 0, -0.00175270098132063, -0.000434460217486266, 1793613.37509986, -0.00132295226125441, 1793613.37426612, 12555293.6258036, 12555293.6253795, 0.000760004257006742, -0.00178736178200963, 25110587.253067, -3.18216630290999e-05, 5380840.12734672, 5380840.12602444, 0, 0, 0, 0, 0, 0, 2.12407773736583e-05, 0, -1.0861426273561e-06, 0, 0, 0, 5.7267510745271e-05, 0, 0.000106653656517157, 4.50466204156291e-06, 0.000173573688865872, -2690420.0626612, 0, 0, 0, 0.000195638602340417, 0, 0.000506530332301553, 9.76518397369203e-07, 3.85912061856478e-05, 0.000514984827434507, 0.000399231761806578, -8071260.18797207, -8071260.18803815, 0, 0, 0, 0.000620262706062103, 0, 0.000859432135657203, 0.000100724645625341, 5.39068935080559e-05, 0.000710386176530309, 0.000425365036156481, -8071260.18794898, -8071260.18774823, -0.000247803944892594, 0.00109454660671449, -16142520.3765546, 0.00045167190713118, -2690420.0626187, -2690420.06214388, -0.00024207622198907, -0.000978600558281489, 0.000795979008027204, 0.000396666593002452, -8071260.18917923, -8071260.18863017, -0.000975747952332197}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.29246517236483e-07, 0, 7.32398082098263e-08, 5248.79999981047, 0, 0, 0, 0, 0, 0, 2.66014576673756e-06, 0, -8.31998667546934e-07, 0, 0, 5.88579608656961e-06, -11153.6999953777, -11153.6999912307, -109568.700006225, 0, 0, 0, 0, 0, 0, -7.70292714741413e-06, 0, 6.02585725389883e-06, 0, 0, 0, -2.78681850638259e-05, 0, -4.09733112583869e-05, 5904.89997941762, 5904.89993790302, 835051.275061153, 11809.7999395066, 220941.674994809, 220941.674984184, 0, 0, 0, 0, 0, 0, 1.46840105343274e-05, 0, -2.02222315333008e-05, 0, 0, 0, 4.18705808341139e-05, 0, 0.00011395497185614, 4.36667583815158e-05, 0.000175188736461207, -3022570.68771184, 0, 0, 0.000180213292216244, 0.000149391686745364, -110716.874916428, 0.000305314735898727, -110716.874763158, -1538964.56260777, -1538964.56260829, -221433.749836693, 0, 0, 0, 0, 0, 0, -1.88976448485929e-05, 0, 3.2428436510343e-05, 0, 0, 0, -2.00038507421562e-05, 0, -0.000141904693276044, -4.96969644028631e-05, -0.000247783015559502, 5520343.38784661, 0, 0, 0, -0.000256776009295242, 0, -0.000429795631320957, -0.000188824910229526, -0.000183694939226594, -0.000636642757320525, -0.000541299458254862, 4782969.00039043, 4782969.00042639, -0.000525272209710254, 697516.312452394, 697516.312172532, -0.000336022144286576, -0.000944628026894679, 1395032.62513758, 0, 0, 0, 0, 0, 0, 1.38187148858016e-05, 0, -2.46883703426097e-05, 0, 0, 0, -1.9824038800762e-06, 0, 7.80084829084072e-05, 2.8370096575661e-05, 0.000177493551478208, -4842756.11277051, 0, 0, 0, 0.000135776499206479, 0, 0.000411672494556348, 0.000127163532834673, 0.000167180420293694, 0.000588873518415396, 0.00048127831997693, -6636369.48798184, -6636369.48804217, 0, 0, 0.000469238576994146, 0.000401087133967079, 0.000215186828570247, 0.000730151786003208, 0.000666937572018655, -1793613.3751447, -1793613.37485743, 0.00029865736098857, 0.00146353045859888, -3587226.75069577, 0.00103913409565893, -9.26138968006385e-06, 0.00078587915662953, 0, 0, 0, 0, 0, 0, -4.13674958950598e-06, 0, 7.21321359305539e-06, 0, 0, 0, 2.00995743020778e-06, 0, -1.44437801184511e-05, -6.21254259812925e-06, -5.1421343786165e-05, 1614252.03758108, 0, 0, 0, -1.85970822936845e-05, 0, -0.000136667009218188, -3.54915171531727e-05, -5.54781345247864e-05, -0.000202828084028459, -0.00015654581813057, 3228504.07519967, 3228504.07523129, 0, 0, 0, -0.000154592374413667, 0, -0.000244857053835325, -0.000120331689861048, -0.000120261577875853, -0.000339054474225093, -0.000293183651046954, 1614252.03763906, 1614252.03753182, -0.000113167184876148, -0.000687940549223943, 3228504.07548073, -0.000274398175297681, -5.57405491482766e-05, -0.000315765199958275, -0.000356692123178884, -4.68104718189346e-05, -0.000828741767102209, -0.000807895052014141, 0.000278456101312256, -7.9490469282647e-05, -0.000406184664191157}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.20283210020793e-08, 0, 7.45170033570369e-08, -656.100000065508, 0, 0, 0, 0, 0, 0, 3.30214521175158e-07, 0, -5.87814422598646e-07, 0, 0, -2.17968777165384e-07, 12957.975000599, 656.100000953717, 6561.00000027619, 0, 0, 0, 0, 0, 0, -1.49632129521167e-06, 0, 2.21313726975886e-06, 0, 0, 0, -7.4666053460088e-07, 0, 2.16340599616186e-06, -89803.6875022745, -6.29072644486855e-06, -5904.90000040085, -12301.8750052029, -128923.650003587, -5904.90000255438, 0, 0, 0, 0, 0, 0, 3.32385386493781e-06, 0, -4.55937820174864e-06, 0, 0, 0, 4.79940529230153e-06, 0, -7.34899523258704e-06, 276792.187504405, 1.86258633882514e-05, -2.09411026544229e-07, 0, 0, -1.17551431960059e-06, 77501.8125136929, 885735.000013341, 2.52943308795801e-05, 2.09921910667399e-05, 110716.875011784, -7.81546518320547e-07, 110716.875007806, 0, 0, 0, 0, 0, 0, -4.01812699200879e-06, 0, 5.28086693632129e-06, 0, 0, 0, -8.56044700182288e-06, 0, 1.16921246092451e-05, -378651.712504588, -2.75762911861687e-05, 1.3437520698758e-06, 0, 0, 0, -7.23352447611549e-06, 0, 8.17066505931799e-06, -199290.375018089, -2690420.06252265, -5.44576614385582e-05, -5.52165272721732e-05, -2.12690277177895e-05, 1.1732115918184e-05, -4.8090813748272e-05, -697516.312536429, -2.7546092691262e-05, -697516.312515274, -4.15781264767655e-05, 7.12208071278514e-06, 0, 0, 0, 0, 0, 0, 2.55341134216583e-06, 0, -3.25033224809756e-06, 0, 0, 0, 6.55760058779359e-06, 0, -8.70857198874356e-06, 179361.337502439, 2.02336392515666e-05, -1.47867390938735e-06, 0, 0, 0, 1.10716514792502e-05, 0, -1.18768135387719e-05, 179361.337511535, 3587226.75001809, 5.21776014204931e-05, 5.72395393063945e-05, 1.91310757134153e-05, -1.57377543399332e-05, 0, 0, 2.58568188755944e-06, 4.18297519976417e-05, 1793613.37504299, 6.81173748201405e-05, 6.6434229440835e-05, 4.48700060043786e-05, 9.36541295824187e-06, 1793613.37501579, 7.91308203864087e-05, -2.08212383856623e-05, 3.91175742372208e-05, -5.242658384482e-06, 1.13824203459456e-05, 0, 0, 0, 0, 0, 0, -6.6846272242264e-07, 0, 8.30957670141054e-07, 0, 0, 0, -1.91258808582686e-06, 0, 2.46350849173211e-06, -5.1153918663989e-07, -5.89932693043894e-06, 5.72458667674307e-07, 0, 0, 0, -4.34386829224517e-06, 0, 5.16299619359401e-06, -2.85780540004693e-06, -1614252.03750555, -1.85243627915065e-05, -2.08103030299082e-05, -6.66756807896017e-06, 6.08641525243185e-06, 0, 0, 0, -4.83145060078713e-06, 0, 1.79935716615771e-06, -1.33246901919172e-05, -1614252.03751752, -3.21401970305423e-05, -3.75775703034667e-05, -1.99457586560278e-05, 3.72460337914611e-06, -1614252.03750599, -4.27404813554378e-05, 1.04502703696611e-05, -2.98258480463137e-05, -2.63371610791685e-05, -2.5962577721797e-05, -1.47537600733169e-05, -2.51065000138579e-06, -3.86949962955316e-05, -3.27386861423603e-05, 7.86407308389915e-06, 1.05695074562096e-05, 7.40223261994639e-06}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.00142380923416e-07, 0, 7.6473177430323e-08, -546.750000227376, 0, 0, 0, 0, 0, 0, 8.35803463150719e-07, 0, -8.45016887891446e-07, 0, 0, 3.97821586955593e-07, 9568.12500159817, 546.750003466303, 7927.87500159085, 0, 0, 0, 0, 0, 0, -3.2410519609872e-06, 0, 3.82373635458874e-06, 0, 0, 0, -3.56706612374325e-06, 0, 3.81232985423118e-07, -53308.1250049564, -1.97235792925571e-05, -29524.5000031753, -9021.37501649313, -138191.062513, -7381.12502028256, 0, 0, 0, 0, 0, 0, 7.16143830601043e-06, 0, -9.00632523738816e-06, 0, 0, 0, 9.69641320511353e-06, 0, -5.43229266919748e-06, 110716.87500811, 5.43046004251291e-05, 22143.3749988682, 0, 0, 6.12771923525673e-06, 44286.750034137, 763946.437540894, 7.16964165524083e-05, 9.34348750858132e-05, 509297.625036726, 22143.3750405677, 121788.562586618, 0, 0, 0, 0, 0, 0, -9.17245611948459e-06, 0, 1.15649727931773e-05, 0, 0, 0, -1.34612189350155e-05, 0, 1.28729821070449e-05, -66430.1250071676, -7.86633929020841e-05, 1.12171862226767e-05, 0, 0, 0, -1.23947333199194e-05, 0, -3.48498001621482e-06, -66430.1250340666, -1561107.93756256, -0.000144982765827156, -0.000188782221099895, -365365.687546258, -2.54892217806362e-05, -0.000107912861901589, -2756850.18760147, -0.00015764353329151, -597871.125158791, -0.000264037210889342, -365365.687653715, 0, 0, 0, 0, 0, 0, 6.26907351611227e-06, 0, -7.67502447839359e-06, 0, 0, 0, 1.01682102496055e-05, 0, -1.25111092953109e-05, 3.15028256035705e-06, 5.79672661072196e-05, -1.3334487609464e-05, 0, 0, 0, 8.35584982290783e-06, 0, -2.77323634896777e-06, 1.49287792530626e-05, 896806.687546924, 0.000137236332966124, 0.000175110665620299, 2.53326671608029e-05, -8.27444147716829e-06, 0, 0, 8.48110090773201e-06, 7.09590630915461e-05, 5380840.12511067, 0.000145611735210931, 0.000219487534199862, 1793613.37510267, 0.000102723265246944, 896806.68763481, 0.000341452967535436, 0.000100202360988279, 0.000281358965275619, 1793613.37521975, 0.000312992745837626, 0, 0, 0, 0, 0, 0, -1.75277323732482e-06, 0, 2.06076203717423e-06, 0, 0, 0, -3.33457384771388e-06, 0, 4.42390483587931e-06, -5.00485082438833e-07, -1.7125336595198e-05, 5.06775302461711e-06, 0, 0, 0, -1.86753362862831e-06, 0, 2.81879475934877e-06, -1.81412728475568e-06, -1.3870134599031e-05, -4.93607289959536e-05, -6.17862648871113e-05, -4.28786589351187e-06, 1.04653921393642e-05, 0, 0, 0, -2.55043774930465e-06, 0, -1.80972210603801e-06, -1.5741252165633e-05, -2690420.06254283, -6.87869111725724e-05, -0.000100523304455686, -3.68324328710131e-05, -1.7486294092618e-05, -4.33826775380793e-05, -0.000155641724422785, -1.56140086873241e-05, -4.92872972885857e-05, -2690420.06255885, -8.10877602194973e-05, -0.000104569610116569, -2690420.06260503, -0.000172942786584438, -0.000205400027238855, -9.1128258138956e-05, -0.000109340496674013, -0.00018165716565323}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -2.11422214534176e-08, 0, 0, 0, 0, 0, 0, 0, 2.09977270073529e-07, 0, 4.32074354964832e-07, -546.750000153353, 0, 0, 0, 0, 0, 0, -8.53422249063989e-07, 0, -3.26445110092445e-06, 0, 0, -3.81384960283318e-06, 7927.87500180306, 546.750001690216, 9568.12500065177, 0, 0, 0, 0, 0, 0, 2.15229587175321e-06, 0, 1.18933117591386e-05, 0, 0, 0, 1.25012741798169e-05, 0, 2.55417363881333e-05, -29524.5000070259, -9.4471587993145e-06, -53308.124998513, -7381.12501362863, -138191.062514938, -9021.37500088934, 0, 0, 0, 0, 0, 0, -3.61404723673023e-06, 0, -2.33631248795546e-05, 0, 0, 0, -2.26616856540619e-05, 0, -7.79754531516552e-05, 22143.3750136929, 2.55317557308749e-05, 110716.874989525, 0, 0, -6.87335027650844e-05, 22143.3750374228, 509297.625057291, 4.99370546601792e-05, 1.43140387786283e-05, 763946.437543184, 44286.7499534902, 121788.562559392, 0, 0, 0, 0, 0, 0, 3.56426951627486e-06, 0, 2.54641308126805e-05, 0, 0, 0, 2.67959102472726e-05, 0, 0.000118105024035964, -1.40194372701093e-05, -3.4321929881058e-05, -66430.1249812585, 0, 0, 0, 8.74036866022693e-05, 0, 0.000164182998351099, -5.08247151115652e-05, -365365.687596443, -9.54050967185368e-05, -5.06176105896411e-05, -1561107.93756136, -66430.1248553821, -8.67348309098578e-05, -2756850.18764128, 0.00010450005658321, -365365.6876766, -0.000136200399161174, -597871.125053158, 0, 0, 0, 0, 0, 0, -1.7234803365744e-06, 0, -1.45687304885604e-05, 0, 0, 0, -1.87730880915837e-05, 0, -8.6626663825395e-05, 6.97429238844752e-06, 2.24743157410608e-05, -1.44164520135545e-05, 0, 0, 0, -5.9103782767111e-05, 0, -0.000172543908246726, 3.40559045500938e-05, 7.5436329565203e-05, 8.67060829599241e-05, 5.56484464940728e-05, 896806.68754463, -0.000157298490741532, 0, 0, -0.00013893952859616, 6.62129389112032e-05, 1793613.37516488, 0.000101821266333178, -6.32056344080933e-05, 5380840.12514545, -0.000250565226431423, 0.000200621306649596, 0.000177247656986756, 896806.687471455, 0.000218587223332344, 1793613.37522464, 2.1929224811133e-05, 0, 0, 0, 0, 0, 0, 2.85621791642717e-07, 0, 3.42825692303983e-06, 0, 0, 0, 5.50450908917167e-06, 0, 2.45939478807345e-05, -1.25955815585819e-06, -5.81169318052721e-06, 4.23109813031847e-06, 0, 0, 0, 1.79913828690222e-05, 0, 6.53089871840361e-05, -8.80853831458426e-06, -2.22395763215556e-05, -2.94504653327296e-05, -1.91251624759691e-05, -1.32562042830274e-05, 5.78654157155441e-05, 0, 0, 0, 4.34170134039425e-05, 0, 7.47859846451809e-05, -1.92316349503143e-05, -6.59705859685974e-05, -4.60532319845985e-05, 6.97402584710666e-06, -2690420.06255515, 0.000144209939409385, -7.96794318874206e-05, -8.49143309834149e-05, 3.60344098813739e-05, -3.43471162835007e-05, -2690420.06258538, 0.000107756902016974, -0.00011047434920546, -0.000152258667501478, -0.0001130864667536, -2.63015457866427e-05, -2690420.06258436, 8.19436277247794e-05, -9.88489733364487e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -3.24717878847573e-08, 0, 0, 0, 0, 0, 0, 0, 5.21448619680323e-07, 0, 4.67187191882982e-07, -656.099999873432, 0, 0, 0, 0, 0, 0, -2.93707084089933e-06, 0, -3.11387792241129e-06, 0, 0, -6.17482295532906e-06, 6560.99999993309, 656.099996367681, 12957.9749996368, 0, 0, 0, 0, 0, 0, 9.46337603851785e-06, 0, 1.01261112156399e-05, 0, 0, 0, 2.30098454720305e-05, 0, 3.73970305428722e-05, -5904.90000104982, 2.40971493784206e-05, -89803.6875015656, -5904.8999906463, -128923.650003938, -12301.8749749979, 0, 0, 0, 0, 0, 0, -1.90168812466273e-05, 0, -1.70166048240008e-05, 0, 0, 0, -4.27595881514417e-05, 0, -0.000110181237275229, 2.53738156951879e-06, -7.17198662594721e-05, 276792.187512261, 0, 0, -0.000110626830233786, -4.66192576611137e-06, 110716.875017276, -6.50331340483949e-05, -0.000127111572987136, 885735.000028294, 77501.8124131894, 110716.874955595, 0, 0, 0, 0, 0, 0, 2.27115861996318e-05, 0, 1.4875710889057e-05, 0, 0, 0, 4.94674613489559e-05, 0, 0.000160828844075089, -1.79083577279106e-06, 0.000111460762138816, -378651.712528172, 0, 0, 0, 0.000135222504814655, 0, 0.000261261811827544, -1.07848593896695e-05, -2.74572959344285e-05, 0.000157660625931294, 0.00024790953122401, -2690420.06257679, -199290.374852596, 6.47969385151926e-05, -697516.312563877, 0.00033808094637808, -4.35565690239449e-06, 0.000269873558613955, -697516.312419338, 0, 0, 0, 0, 0, 0, -1.44571188431051e-05, 0, -6.21630348854835e-06, 0, 0, 0, -3.49906067285167e-05, 0, -0.000112329267456419, -3.1224630856931e-08, -8.74448329501826e-05, 179361.337527136, 0, 0, 0, -7.82974000708251e-05, 0, -0.000272824706402183, 1.32783821257776e-05, 1.76780136395302e-05, -0.000166445943418725, -0.000226273330495835, 3587226.75008832, 179361.337387035, 0, 0, -0.000220661599502972, -1.95083200077635e-05, 5.92155679186093e-05, -0.000125171876534047, -0.000399897887559475, 1793613.37509587, -0.000419906364731638, 6.45819286987997e-05, -0.000408578785348175, 1793613.37496552, -0.000176414024748078, 3.5423834567129e-05, -0.00043297639986585, 0, 0, 0, 0, 0, 0, 3.74755545631176e-06, 0, 9.09969639806016e-07, 0, 0, 0, 1.09988973799745e-05, 0, 2.99416711719991e-05, 3.08562288490272e-07, 2.71442661262648e-05, -9.43773538096636e-06, 0, 0, 0, 1.95806675199253e-05, 0, 0.000101724306356978, -4.06535870465348e-06, -3.61618503990184e-06, 6.50595109700886e-05, 8.14109625710313e-05, -1614252.03753588, 3.0721708831635e-05, 0, 0, 0, 6.19115846537892e-05, 0, 0.000122273778412107, -1.51156579524206e-06, -1.90639767355706e-05, 7.01958578872891e-05, 0.00015778608899344, -1614252.03755, 0.0001860767161977, -3.45192284703417e-05, 0.000195976411622234, -1614252.03751465, 3.13520515496981e-05, -3.87548818195618e-05, 0.000245552705170251, 1.4949878330594e-05, -6.02085306379085e-05, 0.000170334566798322, 0.000319498158112608, -4.61105541795354e-05, 0.000220975524698438, 0.000140633872402665}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 2.73986165764078e-08, 0, 0, 0, 0, 0, 0, 0, -2.75397118597121e-07, 0, -4.42685880230667e-07, -656.100000573336, 0, 0, 0, 0, 0, 0, 1.77330411124806e-06, 0, 3.15672690049697e-06, 0, 0, 2.63423823188308e-06, 656.100002935159, 12957.9750066564, 6561.00000704144, 0, 0, 0, 0, 0, 0, -6.43910914388053e-06, 0, -1.20862048240051e-05, 0, 0, 0, -1.15529216505034e-05, 0, -1.19067301460356e-05, -7.21047856404603e-06, -89803.6875317414, -5904.90003668783, -12301.8750329208, -5904.90002637193, -128923.650059643, 0, 0, 0, 0, 0, 0, 1.30416794044407e-05, 0, 2.61267814135005e-05, 0, 0, 0, 3.23779195926305e-05, 0, 3.24539192811226e-05, 9.45545779346379e-06, 276792.187577217, 0.000101039765410291, 0, 0, 2.99289832648756e-05, 7.5102398730467e-05, 4.38903111574883e-05, 77501.8126244367, 885735.00021424, 0.000100874779338699, 110716.875212159, 110716.875224483, 0, 0, 0, 0, 0, 0, -1.47697266255202e-05, 0, -3.17321904596282e-05, 0, 0, 0, -4.91081661073349e-05, 0, -5.19286333256254e-05, -6.27077533373551e-06, -378651.712601438, -0.000151328859098444, 0, 0, 0, -5.65354048181699e-05, 0, -4.31260582416857e-05, -8.92873263735749e-05, -3.47958862792506e-05, -199290.375214649, -2690420.06287739, -0.000199953485567923, -0.000381839439951809, -0.000215807568877637, -0.000113272227132487, -697516.313019263, -0.00035328873899346, -697516.313133448, -0.000565436315055945, 0, 0, 0, 0, 0, 0, 8.76958763961139e-06, 0, 2.01353300785719e-05, 0, 0, 0, 3.70417792086924e-05, 0, 4.33711014561575e-05, 1.55239829322015e-06, 179361.337568694, 0.000115277781813085, 0, 0, 0, 5.67783206849111e-05, 0, 3.68004036254369e-05, 5.32420990275778e-05, 1.0728397035275e-05, 179361.337676601, 3587226.75031989, 0.000196912061712757, 0.000346840756733966, 0, 0, 4.28019772649237e-05, 0.000176559716689675, 5.145485621643e-05, 0.000241017877015439, 1793613.37558191, 0.000144659449199066, 0.000552096759854778, 0.000252242498577488, 1793613.37572395, 0.000638865929935006, 0.000702662425088982, 0.000554060064031816, 0.00101496124166452, 0, 0, 0, 0, 0, 0, -2.12788976304366e-06, 0, -5.18481480304484e-06, 0, 0, 0, -1.09543848156759e-05, 0, -1.43719487645256e-05, 8.6693115093368e-08, -1.88219113301403e-05, -3.47908343030909e-05, 0, 0, 0, -2.15807082574092e-05, 0, -1.42195207747878e-05, -1.25185190789503e-05, -1.98264103018009e-08, -5.66590051610267e-05, -1614252.03760446, -7.48116403744546e-05, -0.000124980563299161, 0, 0, 0, -2.0570496416108e-05, 0, -1.28488179282314e-05, -5.51936677462321e-05, -5.56330503270584e-06, -9.48234897206757e-05, -1614252.03774038, -7.17984850979556e-05, -0.000222982423282336, -6.65435508558098e-05, -1614252.0377921, -0.000275499398136422, -9.78601408942861e-05, -3.21550659143365e-05, -0.000287962082339376, -0.000268863512147235, -0.000183077841420635, -0.000399509392530067, -0.000570321233828949, -0.000295304563559477, -0.000526062627281288, -0.000545114851730034}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.93224072391173e-08, 0, 1.19293680738216e-07, -546.750000390897, 0, 0, 0, 0, 0, 0, 8.35779666157187e-07, 0, -7.77636343870161e-07, 0, 0, -1.72791092299936e-06, 546.750003315109, 9568.12500498699, 7927.87500343874, 0, 0, 0, 0, 0, 0, -4.72899620725328e-06, 0, 1.76107381020585e-06, 0, 0, 0, 5.22221415541921e-06, 0, 1.23239619913107e-05, -1.15034219704645e-05, -53308.125023355, -29524.5000124245, -9021.37503440602, -7381.12502494593, -138191.062535646, 0, 0, 0, 0, 0, 0, 1.21237049703536e-05, 0, -1.32953661640118e-06, 0, 0, 0, -1.19776270299354e-06, 0, -3.10206018914516e-05, 2.0546730500205e-05, 110716.875055534, 22143.3750257594, 0, 0, -4.39442890764242e-05, 9.50396206334141e-05, 6.93176043072922e-05, 44286.7501213231, 763946.437630409, 22143.3750718534, 509297.62509531, 121788.562708346, 0, 0, 0, 0, 0, 0, -1.60010656367779e-05, 0, -4.76644667779881e-07, 0, 0, 0, -1.38133827666633e-05, 0, 3.49191391460716e-05, -2.01557718226976e-05, -66430.1250722672, -3.33229552796785e-05, 0, 0, 0, 5.59746676670893e-05, 0, 9.80099793764843e-05, -0.000127537340793313, -9.18633757098737e-05, -66430.1251976547, -1561107.93772919, -0.000101573606898636, -365365.687628543, -0.000250326988591142, -0.000154361439090676, -2756850.18774646, -0.000424853642282557, -597871.125547765, -365365.687946342, 0, 0, 0, 0, 0, 0, 1.06730738640004e-05, 0, 1.09218007175273e-06, 0, 0, 0, 1.81617162278134e-05, 0, -1.76761736523547e-05, 1.05217200514807e-05, 4.93539720076804e-05, 2.4078161572515e-05, 0, 0, 0, -2.2548148868817e-05, 0, -8.46830282155865e-05, 8.33468056628356e-05, 5.79086759957377e-05, 0.000154284048325087, 896806.687693872, 7.34529538892339e-05, 9.69741732015323e-05, 0, 0, -0.000103206750713392, 0.000226089063186892, 0.000142461301201373, 0.000268041775351113, 5380840.12528832, 0.000146797558968447, 1793613.37518827, 0.00037081161606664, 896806.688090377, 0.000410860974612199, 0.000735453430968651, 0.000613912008049795, 1793613.37577299, 0, 0, 0, 0, 0, 0, -2.86241146012708e-06, 0, -3.87758609895955e-07, 0, 0, 0, -6.82712905573635e-06, 0, 3.20492954981077e-06, -2.33042850884269e-06, -1.39228772297007e-05, -7.14941124998136e-06, 0, 0, 0, -6.82272141813972e-07, 0, 2.49404489935289e-05, -2.14417828309676e-05, -1.40295185373874e-05, -4.71801294657673e-05, -6.29042512084328e-05, -2.20676412586357e-05, -3.27130153708344e-05, 0, 0, 0, 3.22905582982786e-05, 0, 5.03559715598857e-05, -7.46577211867595e-05, -4.67348790118621e-05, -0.000102037134924499, -2690420.06262629, -5.1120184333438e-05, -5.60635477456243e-05, -0.000117037480085342, -0.000226608590039555, -0.000143579346025102, -0.000121628375696286, -7.08536220360102e-05, -2690420.06260097, -0.000321350721418628, -0.000270751667760768, -0.000387757186313755, -2690420.06291799, -0.000284414458117622, -0.000342048504331643, -0.000522697173530079}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -4.1179861416125e-08, 0, 0, 0, 0, 0, 0, 0, 5.63069123389065e-07, 0, 6.29220213544192e-07, -546.750000159345, 0, 0, 0, 0, 0, 0, -3.03047545514661e-06, 0, -3.70575964590107e-06, 0, 0, -7.40476583949366e-06, 546.750002232802, 7927.87500285582, 9568.12500125894, 0, 0, 0, 0, 0, 0, 7.96741881826609e-06, 0, 1.15528617780908e-05, 0, 0, 0, 3.5218492000882e-05, 0, 3.45400671066956e-05, -8.57573074371756e-06, -29524.5000112045, -53308.125000576, -7381.12502818397, -9021.37501591195, -138191.062529653, 0, 0, 0, 0, 0, 0, -1.11077374278877e-05, 0, -2.11873420436598e-05, 0, 0, 0, -8.04087292340083e-05, 0, -7.85734807633628e-05, 1.64427956879608e-05, 22143.3750197659, 110716.87498697, 0, 0, -0.000133757149357089, 8.4603857450675e-05, 4.45894387124124e-05, 22143.3750919765, 509297.62510557, 44286.750033047, 763946.437592288, 121788.562709239, 0, 0, 0, 0, 0, 0, 8.19034431363887e-06, 0, 2.3099344295462e-05, 0, 0, 0, 9.37742724725741e-05, 0, 9.56496980127008e-05, -1.73639806850517e-05, -1.86181507877044e-05, -66430.1249652818, 0, 0, 0, 0.000236960590835284, 0, 0.000226558977252127, -0.000120790223407993, -5.70736214403418e-05, -0.00013043385984609, -365365.687653158, -66430.1250127157, -1561107.93761158, -0.0002088286298152, -7.09518461036284e-05, -2756850.18777836, -0.000430052216389858, -365365.6880587, -597871.125495076, 0, 0, 0, 0, 0, 0, -2.82133397955806e-06, 0, -1.38059968811294e-05, 0, 0, 0, -5.43656060596315e-05, 0, -6.11243742965608e-05, 9.78601866117847e-06, 9.68506990323458e-06, -3.41686768982731e-05, 0, 0, 0, -0.000189852324432452, 0, -0.00017658876653123, 8.38490361429467e-05, 3.48640842229668e-05, 8.65404812433563e-05, 0.000101002442202639, -2.35200048623938e-05, 896806.687549027, 0, 0, -0.000274624135110615, 0.000200445605628811, 5.66251366537369e-05, 0.000200766614253142, 1793613.37528878, 2.93932119110134e-05, 5380840.12526394, 0.000378767592454828, 0.000562298146568517, 896806.687948744, 0.000756634112674746, 0.00065986532592391, 1793613.3759584, 0, 0, 0, 0, 0, 0, 2.80034450148082e-07, 0, 3.45998152269498e-06, 0, 0, 0, 1.25134708036805e-05, 0, 1.63521676409517e-05, -2.32251814477838e-06, -2.37161066306533e-06, 1.19650768449834e-05, 0, 0, 0, 5.62844527767105e-05, 0, 5.22630481954697e-05, -2.29242780307314e-05, -8.36622321004973e-06, -2.23511122758498e-05, -2.47651993875854e-05, 1.72747837163037e-05, -1.91614648097759e-06, 0, 0, 0, 0.000113130532251951, 0, 0.000108449883941364, -6.99871639758797e-05, -1.49358165496089e-05, -6.79023944384767e-05, -0.000102529856172278, 6.35113934626975e-06, -2690420.06258196, -0.000122064146010355, -0.000195424356469997, -0.000134482817615629, -9.76104687904427e-05, -1.13579108950011e-05, -2690420.06264182, -0.000333035958332926, -0.000282465371817963, -0.000377806790161935, -0.000509451632708797, -0.000300913877167881, -2690420.06297533, -0.000622504756308015}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.38457296696806e-08, 0, -1.20910410617557e-07, -656.100000231211, 0, 0, 0, 0, 0, 0, -7.30314278977797e-07, 0, 9.27935920236218e-07, 0, 0, 4.39826064403435e-07, 656.100000665333, 6561.00000156567, 12957.9750051868, 0, 0, 0, 0, 0, 0, 2.67188906942197e-06, 0, -3.06673864938594e-06, 0, 0, 0, 4.65398172858849e-06, 0, -5.72401668511199e-06, -1.31252119713125e-07, -5904.90000060585, -89803.6875293003, -5904.90000398528, -12301.875018223, -128923.650042649, 0, 0, 0, 0, 0, 0, -4.68115641396719e-06, 0, 4.80697592388634e-06, 0, 0, 0, -1.6771279499235e-05, 0, 2.35666906247989e-05, -2.4235019871872e-06, -1.04135142212832e-05, 276792.18757498, 0, 0, -1.28401787216305e-05, 4.15315024352763e-06, 2.43967750656065e-05, -9.71470006553371e-06, 110716.875110287, 77501.8125858086, 885735.00020177, 110716.875142675, 0, 0, 0, 0, 0, 0, 4.75153003061131e-06, 0, -3.12864104960752e-06, 0, 0, 0, 1.97657284526246e-05, 0, -4.34914632170565e-05, 4.84753740854862e-06, 2.41641782747835e-05, -378651.712598208, 0, 0, 0, 5.35475343791532e-05, 0, 6.61095536961391e-06, -1.94517181962487e-06, -9.91654181162568e-06, 5.59478728473363e-05, -0.000131932357023757, -199290.375169729, -2690420.06289244, 2.71586151150947e-05, -9.11036530871776e-05, -697516.312918805, -0.00020391596794836, -0.000295190732356937, -697516.313008143, 0, 0, 0, 0, 0, 0, -2.76982602145014e-06, 0, 1.44156823977505e-07, 0, 0, 0, -9.87771973658314e-06, 0, 3.5947066886944e-05, -3.86365404254191e-06, -2.13546434397986e-05, 179361.337563286, 0, 0, 0, -4.7979713446136e-05, 0, 1.67297437106517e-05, -2.73173692286817e-07, -5.89237800309381e-06, -7.16951574207298e-05, 7.48454239756431e-05, 179361.337655122, 3587226.7503521, 0, 0, -6.44198737612983e-05, -2.20622014279774e-05, 3.28891738097697e-05, -8.0355547409291e-05, 0.000373593308523475, 0.000122894555432551, 1793613.37553459, 0.000133984187980682, 0.000225571946942651, 1793613.37564113, 0.000298625642208259, 0.000479774976074869, 0.000778881102960545, 0, 0, 0, 0, 0, 0, 7.10802504866368e-07, 0, 4.34426821850067e-07, 0, 0, 0, 1.89907521322639e-06, 0, -1.07573356709986e-05, 1.13438471967875e-06, 6.86211902298061e-06, -1.56858169233594e-05, 0, 0, 0, 1.24799151526017e-05, 0, -1.40738778950228e-05, 4.75953051523472e-07, 4.57487742174981e-06, 2.88791398518448e-05, -1.5236665817836e-05, -5.32730943322234e-05, -1614252.03761958, 0, 0, 0, 3.53230410023043e-05, 0, 2.1149002222145e-05, 5.53575417634285e-06, 1.01177563258726e-06, 4.95933300402876e-05, -0.000127358295368547, -5.82388729807373e-05, -1614252.03773544, -3.09722673334511e-05, -5.248760230437e-05, -1614252.03777962, 3.30622603763664e-05, -2.3915839702781e-05, -0.000231915561737456, -0.000108652047374769, -0.000156688469762476, -0.000106971910272934, -0.000364998523757498, -0.000294059244613305, -0.000487377139348617, -0.000398312630717532}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -5.55288982532452e-07, 0, 0, 0, 0, 0, 0, 0, 5.93253462486912e-06, 0, 9.12174854608373e-06, 22963.5000121304, 0, 0, 0, 0, 0, 0, -3.40340832591365e-05, 0, -6.45193631487552e-05, 0, 0, -6.7851287658734e-05, -319438.687566976, -319438.68763969, -112767.187636634, 0, 0, 0, 0, 0, 0, 0.000113870499452186, 0, 0.000238190606037658, 0, 0, 0, 0.000267327454599707, 0, 0.000361699460397513, 1220961.09393824, 1220961.09442915, 206056.406943807, 4301965.68820381, 1427017.50052926, 1427017.50112953, 0, 0, 0, 0, 0, 0, -0.000224802946561487, 0, -0.00049421122180534, 0, 0, 0, -0.000622760378288333, 0, -0.0010102527576042, -2020582.96906218, -2020582.97045915, -166075.314428511, 0, 0, -0.000998909725901933, -15195891.0953273, -4207241.25103987, -15195891.0965271, -4207241.2539856, -2352733.59555409, -2352733.59765965, -17548624.6917676, 0, 0, 0, 0, 0, 0, 0.000258137738980792, 0, 0.00057791258150954, 0, 0, 0, 0.000823283315680318, 0, 0.00153640771939125, 1544500.40655813, 1544500.40856369, 49822.5966848346, 0, 0, 0, 0.00159045148329024, 0, 0.00187037172095174, 22868570.533082, 4683323.81368686, 22868570.5364149, 4683323.819598, 1693968.19096111, 1693968.19474138, 42648140.2545859, 4782969.00212922, 4782969.00908154, 47431109.257037, 47431109.2622514, 26256506.915812, 0, 0, 0, 0, 0, 0, -0.000159550567002847, 0, -0.000354659178348724, 0, 0, 0, -0.00056635630528273, 0, -0.00119318525462635, -448403.343916134, -448403.345352516, -0.00225791545379316, 0, 0, 0, -0.00132864013438689, 0, -0.00175385114246088, -15694117.0323483, -1793613.37574135, -15694117.0357988, -1793613.38112261, -448403.347227842, -448403.350676011, 0, 0, -0.00178616254333173, -44840334.3783747, -2690420.06381845, -44840334.3805993, -2690420.07257349, -1793613.37729262, -1793613.38467103, -47530754.4429625, -47530754.4522627, -17039327.0728158, -91474282.1389151, -48427561.1344047, -48427561.1420799, 0, 0, 0, 0, 0, 0, 4.0990142776711e-05, 0, 8.87157592009828e-05, 0, 0, 0, 0.000157874931003448, 0, 0.000367704337549844, 3.73553205255746e-05, 0.000447080685826957, 0.000682668003330434, 0, 0, 0, 0.000439258667450325, 0, 0.00065267899508933, 4035630.09402285, 0.000191901940628965, 4035630.09528119, 0.00202341522552216, 0.00136377551702598, 0.00261323675244375, 0, 0, 0, 0.000701282667568852, 0, 0.000752612187535043, 16142520.3759334, 0.000362612562268712, 16142520.3774374, 0.00422951244928386, 0.00109681019954784, 0.00412939897903822, 16142520.3766414, 16142520.3812935, 4035630.09832521, 24213780.5645065, 0.000533069792341703, 0.00479607675926075, 48427561.130364, 24213780.5658906, 48427561.1333826, 24213780.5725184, 16142520.3793584, 16142520.3830767, 48427561.1343096}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -3.8398855915911e-08, 0, 0, 0, 0, 0, 0, 0, 3.16418146534121e-08, 0, 9.03565113283439e-07, 3280.50000062664, 0, 0, 0, 0, 0, 0, 1.23096943723977e-06, 0, -7.87283916910739e-06, 0, 0, -2.53472512181104e-06, -61099.3125027774, -36495.5625051491, -6971.06250865757, 0, 0, 0, 0, 0, 0, -5.49289300078027e-06, 0, 3.12383872692411e-05, 0, 0, 0, -1.03507955721196e-05, 0, 3.23770143053516e-05, 384433.593755784, 66430.1250251883, 3690.5625571033, 672297.468775012, 122403.656276027, 70120.6875375963, 0, 0, 0, 0, 0, 0, 9.85464405773234e-06, 0, -6.57721158943339e-05, 0, 0, 0, 5.79859585030913e-05, 0, -0.000120994853931932, -1024131.09375811, -33215.0625729003, -0.000180243633215613, 0, 0, -2.49790126247514e-05, -4146346.96878908, -686444.625032014, -1162527.18762211, -66430.1250744327, -60894.2813835382, -33215.0626968445, -1223421.46887208, 0, 0, 0, 0, 0, 0, -8.09558351550777e-06, 0, 7.64044876804027e-05, 0, 0, 0, -0.000103346496403342, 0, 0.000203854431867714, 1145919.65626016, 0.000114570956154059, 0.000289325503722384, 0, 0, 0, -6.05389244688303e-05, 0, 0.000144408121580818, 10662035.0625153, 1444855.21876611, 548048.53153166, 0.00010086989549774, 0.000313638810855058, 0.000488423567955457, 6476937.18768104, 298935.562628978, 0.000234804352592742, 6775872.75011706, 1096097.06281343, 548048.531661591, 0, 0, 0, 0, 0, 0, 2.49025241736603e-06, 0, -4.62468277917305e-05, 0, 0, 0, 8.07434543303667e-05, 0, -0.000163267024191031, -448403.343758744, -9.06238053960991e-05, -0.000228493374132651, 0, 0, 0, 0.000114193419342324, 0, -0.000189630087492842, -11210083.5937414, -896806.687504473, -0.00029153989053808, -8.53021371280773e-05, -0.000336155441376692, -0.000542442908178407, 0, 0, -3.90062434867253e-05, -13452100.3125904, -448403.343801725, -2690420.06281203, -0.000123944308932459, -0.000201412845015289, -0.000399908050746148, -13900503.65624, -0.00042938703945097, -0.000633341756055636, -5380840.12529143, -2690420.06280531, -0.00048002908779387, 0, 0, 0, 0, 0, 0, 2.19436339015963e-08, 0, 1.13834022251119e-05, 0, 0, 0, -2.3359381961542e-05, 0, 5.05437260679502e-05, 3.09868682352063e-06, 2.83379970464392e-05, 7.03468120988709e-05, 0, 0, 0, -5.3004303048393e-05, 0, 8.03019718101371e-05, 4035630.09374542, 2.14489506069885e-06, 0.000110550094659176, 3.07934519015796e-05, 0.000133988600058376, 0.000220634170657521, 0, 0, 0, -2.01435451672417e-05, 0, 5.06205914640601e-05, 8071260.18750722, 8.08185105549124e-06, 0.000168299557716561, 2.41261086290419e-05, 0.000106930755612324, 0.000230872670514056, 8071260.18746271, 0.000216053707267892, 0.000338396734952413, 4035630.09385087, 4.02588911388168e-05, 0.000102860759799939, 8071260.1875438, 4035630.09379766, 0.000248228364670558, 0.000236916332370368, 0.000244597621888747, 0.000317368730261534, 0.000229261808959775}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -2.02959335918653e-07, 0, 0, 0, 0, 0, 0, 0, 2.16196529320797e-06, 0, 3.39673714768003e-06, 3280.50000276082, 0, 0, 0, 0, 0, 0, -1.15269494608568e-05, 0, -2.2441953450999e-05, 0, 0, -2.78750864473343e-05, -36495.562514106, -61099.312527719, -6971.06253559452, 0, 0, 0, 0, 0, 0, 3.63259942705101e-05, 0, 7.72778396982718e-05, 0, 0, 0, 0.000110183978248813, 0, 0.000141682284921, 66430.1250416797, 384433.593870735, 3690.56269010757, 672297.468882105, 70120.6876234595, 122403.656518579, 0, 0, 0, 0, 0, 0, -6.89941751414402e-05, 0, -0.000150531196382768, 0, 0, 0, -0.000251716541139145, 0, -0.000371433416927099, -33215.0625772652, -1024131.09403373, -0.000534495628804915, 0, 0, -0.000402898897225376, -1162527.18782169, -66430.1252247439, -4146346.96920312, -686444.625863405, -33215.062952398, -60894.2822599607, -1223421.46969288, 0, 0, 0, 0, 0, 0, 7.7287369389074e-05, 0, 0.000166689394298616, 0, 0, 0, 0.000328890025895285, 0, 0.000528713892301556, 8.53726125989152e-05, 1145919.65662249, 0.000809193319223465, 0, 0, 0, 0.000631731407704532, 0, 0.000726146811170257, 548048.531683566, 0.00025406644263693, 10662035.0632411, 1444855.22017939, 0.000917629987030408, 0.00193355059436192, 6476937.18829388, 0.000445977316999327, 298935.564674667, 1096097.06407589, 6775872.75247979, 548048.533557822, 0, 0, 0, 0, 0, 0, -4.69317947574223e-05, 0, -9.76827625472564e-05, 0, 0, 0, -0.000226634993179478, 0, -0.000386074269900515, -5.03926004860325e-05, -448403.344008183, -0.000615651379230474, 0, 0, 0, -0.000515035180206021, 0, -0.000646998485128402, -0.000305517419151834, -0.000167570404485029, -11210083.5943428, -896806.688672016, -0.000951987965212559, -0.00186161452015844, 0, 0, -0.000696183137354292, -2690420.06317456, -0.000208089561272952, -13452100.3133062, -448403.346000935, -0.000509839324580179, -0.00243976527918609, -0.00130456031700406, -13900503.6589645, -0.00263326631477625, -5380840.12790175, -0.00218568211213555, -2690420.0663683, 0, 0, 0, 0, 0, 0, 1.18774042775744e-05, 0, 2.34949991245217e-05, 0, 0, 0, 6.3864495405413e-05, 0, 0.000112864413689029, 1.21486511822836e-05, 7.37640005386874e-05, 0.000183703412991716, 0, 0, 0, 0.000168068271009196, 0, 0.000226694463136205, 8.72159639301298e-05, 4.71581296653305e-05, 4035630.0939391, 0.000376413142724578, 0.000378324429176914, 0.000699594701517104, 0, 0, 0, 0.00026573415457834, 0, 0.000286983310915764, 0.000225362126417782, 4.43919302039273e-05, 8071260.18779719, 0.000893706223285616, 0.000270458025437886, 0.00106351086631669, 0.000429462716770153, 8071260.18855532, 0.00120748157255807, 4035630.09407036, 6.27299780051236e-05, 0.00114057137555663, 0.00120642257196561, 0.000780126397696751, 8071260.18906972, 4035630.09586087, 0.00104789916461371, 0.00195257705881308, 0.00211608365076069}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 2.40745875020251e-07, 0, 0, 0, 0, 0, 0, 0, -2.19837527847904e-06, 0, -4.12942038689958e-06, -10206.0000038214, 0, 0, 0, 0, 0, 0, 1.1501175422891e-05, 0, 3.11810241411067e-05, 0, 0, 2.48494403745251e-05, 172590.750018232, 126663.750042391, 34809.7500441524, 0, 0, 0, 0, 0, 0, -3.88982354823708e-05, 0, -0.000118959007904079, 0, 0, 0, -7.19374008543854e-05, 0, -0.000154131919575209, -922640.625044325, -352653.750222898, -39366.0002498536, -2101980.37519519, -548663.625143927, -392019.750300818, 0, 0, 0, 0, 0, 0, 8.09194234926579e-05, 0, 0.000249527988808884, 0, 0, 0, 0.000134533365141219, 0, 0.00047788005741855, 1955998.12506582, 369056.250615142, 14762.2507510394, 0, 0, 0.000322203445941473, 10820729.2503844, 2516963.62523079, 5513700.37587537, 752874.751038213, 575727.750576232, 398580.751102728, 6089428.12599956, 0, 0, 0, 0, 0, 0, -9.86210102903165e-05, 0, -0.000291640572487645, 0, 0, 0, -0.000161692536846028, 0, -0.000768466017632429, -1793613.37506365, -132860.250899804, -0.0011963350685911, 0, 0, 0, -0.000382777731271073, 0, -0.000725384081530394, -21390500.2503701, -3786517.12521256, -5380840.1268306, -398580.751979553, -199290.376259348, -132860.252274836, -24844866.7513442, -2192194.12558239, -398580.752216398, -27037060.8764092, -10960970.6280542, -5779420.8772942, 0, 0, 0, 0, 0, 0, 6.42612196588278e-05, 0, 0.000177777067645762, 0, 0, 0, 0.000111265495588638, 0, 0.000614022620252542, 597871.125037167, 0.000662352044853441, 0.000946579746630288, 0, 0, 0, 0.000253539592123913, 0, 0.000766383536212256, 17936133.7501769, 1793613.37511395, 1793613.37676362, 0.00185501481123758, 0.00135615879400529, 0.00237888447404526, 0, 0, 0.00053697171581071, 35872267.5008745, 1793613.37528909, 20327618.251917, 0.00243845260442573, 597871.125778858, 0.00247959595284042, 37665880.8758706, 5380840.12912163, 1793613.37778029, 41253107.6281829, 21523360.5019076, 5380840.12883712, 0, 0, 0, 0, 0, 0, -1.72080713653902e-05, 0, -4.39969903601145e-05, 0, 0, 0, -3.30916524910381e-05, 0, -0.000192012225185598, -9.65628696143852e-06, -0.000193585142283094, -0.000291794918476834, 0, 0, 0, -7.10978554005262e-05, 0, -0.000306144968840371, -5380840.12503625, -2.81366953100485e-05, -0.000634479549826781, -0.000661051861108274, -0.000551118141465661, -0.000955601305136879, 0, 0, 0, -0.000166165183732587, 0, -0.000269081227603251, -16142520.3752003, -6.47348691851987e-05, -5380840.12593185, -0.00108106492140167, -0.00041820563011754, -0.00116419286385251, -16142520.3751903, -0.00195302499381952, -0.00138723383798131, -16142520.3756456, -0.000158982547986094, -0.00111160600481006, -32285040.7510597, -16142520.3755423, -16142520.3772325, -0.00239997287968274, -5380840.1260065, -0.00181233768255946, -16142520.3769178}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -1.53003997152966e-07, 0, 0, 0, 0, 0, 0, 0, 1.54850003385603e-06, 0, 2.52274171534655e-06, 2916.00000183009, 0, 0, 0, 0, 0, 0, -8.55876847500409e-06, 0, -1.77860146784273e-05, 0, 0, -1.74778854113871e-05, -45562.500008328, -45562.5000194898, -6196.50002238491, 0, 0, 0, 0, 0, 0, 2.86043249177754e-05, 0, 6.58131030263545e-05, 0, 0, 0, 6.37332291436389e-05, 0, 9.20715554656932e-05, 205031.250021421, 205031.250098299, 3280.50012672769, 705307.500085968, 90213.7500679057, 90213.750153642, 0, 0, 0, 0, 0, 0, -5.66729684617466e-05, 0, -0.000136270427044362, 0, 0, 0, -0.000148735941334968, 0, -0.000264570766454197, -295245.000036175, -295245.000264819, -0.000379650513639533, 0, 0, -0.000218023933007396, -3107453.62517972, -339531.750104408, -3107453.62536214, -339531.750518285, -44286.7502731127, -44286.7505678286, -1306459.12548799, 0, 0, 0, 0, 0, 0, 6.47143278802841e-05, 0, 0.00015818246845222, 0, 0, 0, 0.000206800331786413, 0, 0.000415068618978257, 132860.250039772, 132860.250381653, 0.000603748619503607, 0, 0, 0, 0.000317693246139811, 0, 0.000412529386626202, 4251528.00020412, 265720.50009675, 4251528.00072834, 265720.500971947, 0.000601660498937754, 0.00116222911589808, 13020304.5005624, 132860.250247292, 132860.25112319, 4849399.12571264, 4849399.12642593, 597871.126153583, 0, 0, 0, 0, 0, 0, -3.93052390050065e-05, 0, -9.59783373223157e-05, 0, 0, 0, -0.000151522879008237, 0, -0.000329346363786308, -2.50352609179701e-05, -0.000277497182288523, -0.000477146890268203, 0, 0, 0, -0.000266474843438436, 0, -0.000409387021168818, -1793613.37512764, -5.44220963606584e-05, -1793613.37568641, -0.00090782137830088, -0.000652936601725739, -0.00121262455663214, 0, 0, -0.000331152396795726, -15544649.2503964, -9.89930421039909e-05, -15544649.2507472, -0.00121342267718676, -0.000334912459983545, -0.00123919310232167, -3587226.75050397, -3587226.75184511, -0.00139845392027703, -17338262.6265046, -1793613.37596588, -1793613.37689113, 0, 0, 0, 0, 0, 0, 9.82375822461987e-06, 0, 2.36691058836704e-05, 0, 0, 0, 4.47385904793778e-05, 0, 0.000102743129606933, 6.6725897860171e-06, 8.00793077845538e-05, 0.000146893683426069, 0, 0, 0, 9.3240920975863e-05, 0, 0.000161427516377603, 3.52707311115766e-05, 1.41022931498137e-05, 0.000243602075781403, 0.000324725271663612, 0.000266551238158114, 0.000487592564009146, 0, 0, 0, 0.000119384593516082, 0, 0.000143412386900344, 5380840.1251087, 1.47210079236191e-05, 5380840.12534999, 0.000538326113785596, 0.000186584256301394, 0.000580444203856205, 0.000143774679237488, 0.000851429399883262, 0.000696141286347377, 10761680.2502557, 5.02698405610205e-05, 0.000534916176526165, 10761680.2505427, 0.000297140629031671, 10761680.2509871, 0.00113673102894746, 0.000498578121764492, 0.000907710102423305, 5380840.12595425}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 4.53724215704644e-07, 0, 0, 0, 0, 0, 0, 0, -4.94722926478747e-06, 0, -7.3605853798548e-06, -10206.000007674, 0, 0, 0, 0, 0, 0, 2.82010592610578e-05, 0, 4.98168692787187e-05, 0, 0, 5.76418757366361e-05, 126663.750040982, 172590.750083566, 34809.7500914177, 0, 0, 0, 0, 0, 0, -9.34104956653516e-05, 0, -0.000178334069128578, 0, 0, 0, -0.000234963372388413, 0, -0.000288607803807755, -352653.750118702, -922640.625390909, -39366.0004764053, -2101980.37541087, -392019.750335618, -548663.625721613, 0, 0, 0, 0, 0, 0, 0.000182163280970726, 0, 0.000361906907588771, 0, 0, 0, 0.000564557985589126, 0, 0.000773176637401862, 369056.250209921, 1955998.12596503, 14762.2513375626, 0, 0, 0.000812138924146755, 5513700.37596089, 752874.750633803, 10820729.2515332, 2516963.62744718, 398580.751188862, 575727.752579156, 6089428.12764846, 0, 0, 0, 0, 0, 0, -0.000206002558204005, 0, -0.000416091071900763, 0, 0, 0, -0.000768900113190795, 0, -0.00114533087273111, -132860.250222319, -1793613.37629958, -0.00203968249803898, 0, 0, 0, -0.00132200988859, 0, -0.00143745625922029, -5380840.12621052, -398580.750715072, -21390500.2527272, -3786517.1292563, -132860.252358074, -199290.379858043, -24844866.7525978, -398580.751285607, -2192194.1307401, -10960970.629415, -27037060.8822967, -5779420.88117881, 0, 0, 0, 0, 0, 0, 0.000125042739702044, 0, 0.000251962292034191, 0, 0, 0, 0.000543553086513423, 0, 0.000873225225187553, 0.000127715869098765, 597871.125904019, 0.00156694131726653, 0, 0, 0, 0.00113567765758237, 0, 0.00130485611280162, 1793613.37579969, 0.000453950489182787, 17936133.752334, 1793613.37863374, 0.0024156489010766, 0.00468639583687181, 0, 0, 0.00138046489336145, 20327618.252049, 0.000695274171489917, 35872267.5029568, 1793613.38118217, 0.00144391921623445, 597871.131211423, 5380840.12855888, 37665880.8834461, 1793613.38187013, 41253107.6334207, 5380840.13100702, 21523360.5106444, 0, 0, 0, 0, 0, 0, -3.14909075830747e-05, 0, -6.23524044185195e-05, 0, 0, 0, -0.000154887497926426, 0, -0.000265421220018223, -3.03760814566447e-05, -0.000254826308201717, -0.000472369288495216, 0, 0, 0, -0.000387982511697739, 0, -0.000475884666102704, -0.000217837540157806, -0.000121249950086697, -5380840.12577524, -0.0012004712770545, -0.00095492940105862, -0.00177559503149759, 0, 0, 0, -0.000549057977436449, 0, -0.000557298006403105, -5380840.12562716, -0.000167259076868178, -16142520.37622, -0.00255666467679398, -0.000730545731784072, -0.00268312896119541, -0.00112908959990586, -16142520.3784822, -0.00311000412574179, -16142520.3761063, -0.000263132270594922, -0.00296955934575307, -16142520.3783821, -0.00216934625247399, -32285040.7548295, -16142520.3810231, -0.00284711554408263, -5380840.13018496, -16142520.3808436}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 2.35674176449612e-07, 0, 0, 0, 0, 0, 0, 0, -2.07812763817878e-06, 0, -4.53517837951441e-06, 22963.5000070183, 0, 0, 0, 0, 0, 0, 7.07981082682501e-06, 0, 3.34527652847382e-05, 0, 0, 3.47308681509129e-05, -319438.687555768, -112767.187588481, -319438.68756057, 0, 0, 0, 0, 0, 0, -1.60505177900594e-05, 0, -0.000120566506559489, 0, 0, 0, -8.80896735077836e-05, 0, -0.000233938429898775, 1220961.093932, 206056.40668387, 1220961.09395213, 1427017.50055177, 4301965.6879334, 1427017.50060248, 0, 0, 0, 0, 0, 0, 2.70364842438943e-05, 0, 0.000232733121450204, 0, 0, 0, 0.000121636176928063, 0, 0.000740684819972248, -2020582.96905954, -166075.313540092, -2020582.96913428, 0, 0, 0.000497191073144364, -4207241.251338, -15195891.0950137, -2352733.59585685, -2352733.59616736, -15195891.0949513, -4207241.25140068, -17548624.6906354, 0, 0, 0, 0, 0, 0, -2.70025020529171e-05, 0, -0.00024740523270393, 0, 0, 0, -0.000143474015356505, 0, -0.00114640678074377, 1544500.40653179, 49822.5950440467, 1544500.40670493, 0, 0, 0, -0.000417976089152597, 0, -0.00133431377183286, 4683323.81408334, 22868570.5331011, 1693968.19128672, 1693968.19199835, 22868570.5329805, 4683323.81403859, 4782969.00366887, 42648140.2527058, 4782969.00423393, 47431109.2562343, 26256506.9151225, 47431109.2557057, 0, 0, 0, 0, 0, 0, 1.19989552331171e-05, 0, 0.00013739490918666, 0, 0, 0, 0.000126977711615616, 0, 0.000846105843257215, -448403.343875444, -0.000806706554499458, -448403.344051622, 0, 0, 0, 0.000160016604957251, 0, 0.0015169377556064, -1793613.37590998, -15694117.0325759, -448403.346937821, -448403.347558698, -15694117.0325974, -1793613.37590455, 0, 0, 0.000855240584514432, -2690420.06522063, -44840334.3777589, -1793613.37934405, -1793613.38058598, -44840334.3775446, -2690420.06532052, -47530754.443071, -17039327.0731915, -47530754.4419947, -48427561.1359955, -91474282.1328671, -48427561.1355192, 0, 0, 0, 0, 0, 0, -1.23380366111965e-06, 0, -3.13133257609403e-05, 0, 0, 0, -4.67804520696112e-05, 0, -0.000239322185204833, 1.97394153899493e-05, 0.000200848349345802, 8.21376735626962e-05, 0, 0, 0, -4.02491003243324e-05, 0, -0.000598503351242593, 0.000200979240203637, 4035630.09411717, 0.00101305251564008, 0.00118496366208546, 4035630.0941886, 0.000251298759861481, 0, 0, 0, -0.000160834480590542, 0, -0.000556603494035451, 0.000725020840818513, 16142520.3760357, 0.00184037506299907, 0.0025186458314611, 16142520.3759409, 0.000611332490223305, 16142520.3768561, 4035630.0982683, 16142520.376438, 0.00157176667071523, 24213780.5637942, 0.00196937166227686, 24213780.5670483, 48427561.1287232, 16142520.3813792, 16142520.3814956, 48427561.1281238, 24213780.5662799, 48427561.1317337}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.46070151046932e-07, 0, -2.47523626581981e-07, 3280.50000075652, 0, 0, 0, 0, 0, 0, -3.23408653763322e-06, 0, 2.34982110717682e-06, 0, 0, -6.69803897402294e-07, -61099.3125053883, -6971.06251105429, -36495.5625050944, 0, 0, 0, 0, 0, 0, 1.33367391337283e-05, 0, -9.74884100383494e-06, 0, 0, 0, 1.10569194239627e-05, 0, -3.08445628128372e-06, 384433.593766912, 3690.5625659748, 66430.1250151936, 122403.656305562, 672297.46878806, 70120.6875535387, 0, 0, 0, 0, 0, 0, -2.96899204364016e-05, 0, 2.16634924463112e-05, 0, 0, 0, -3.83463120607269e-05, 0, 1.86793559577412e-05, -1024131.09377755, -0.000187693197927995, -33215.0625247153, 0, 0, -1.86135892732571e-05, -686444.625123822, -4146346.96886555, -60894.2815065667, -33215.0627706039, -1162527.18761697, -66430.125100191, -1223421.46896387, 0, 0, 0, 0, 0, 0, 3.71784680280098e-05, 0, -2.66950530044479e-05, 0, 0, 0, 5.9979576000979e-05, 0, -3.74172065099112e-05, 1145919.65627399, 0.000276101046493022, 2.23881323452677e-05, 0, 0, 0, 6.29740088729664e-05, 0, 7.08281688166263e-06, 1444855.21888297, 10662035.0626727, 0.00053430327996046, 0.000593525904923474, 548048.531439239, 7.79844977153899e-05, 298935.562929201, 6476937.18779581, 0.000430801137407715, 6775872.75037933, 548048.531991252, 1096097.062802, 0, 0, 0, 0, 0, 0, -2.46052148340169e-05, 0, 1.72565886279055e-05, 0, 0, 0, -4.60394434414277e-05, 0, 3.23649424099898e-05, -448403.343760369, -0.00020434735909575, -9.90506095091418e-06, 0, 0, 0, -6.47434830568331e-05, 0, 1.59130674173991e-05, -896806.687564369, -11210083.5938764, -0.000508568872915346, -0.000577306710887616, -0.000156655196505672, -1.63013196829125e-05, 0, 0, -5.06055011759967e-05, -448403.344068646, -13452100.3128155, -0.000596615926440922, -0.000683133547897544, -2690420.06283421, -0.000283466063846547, -13900503.6565552, -0.00107004594621823, -0.000174868038934698, -2690420.06329833, -5380840.12540062, -0.000705152403975413, 0, 0, 0, 0, 0, 0, 6.66844467239767e-06, 0, -4.57514136718849e-06, 0, 0, 0, 1.43433827710396e-05, 0, -1.03268276459242e-05, 1.64316804669904e-06, 6.03368687388907e-05, 1.26819869116894e-06, 0, 0, 0, 2.19070341807965e-05, 0, -1.17046623162149e-05, 1.02368273303958e-05, 4035630.0937864, 0.000181781798542287, 0.000207678175200526, 5.16637516839395e-05, -3.67617854355714e-06, 0, 0, 0, 3.07824922867838e-05, 0, 1.05968738424326e-05, 8.39263832308858e-05, 8071260.18762129, 0.000280926205925524, 0.000339978595544337, 0.000139878320052238, 5.684341212542e-05, 8071260.18758933, 0.000517466092917153, 3.50806116713127e-05, 0.000228153215286115, 4035630.09393082, 0.000261651703851176, 4035630.09404496, 8071260.18768879, 0.000567578905697223, 0.000560087638101709, 0.000144790287017857, 0.000196702155360012, 0.000366023814570611}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 1.27710203817125e-07, 0, 0, 0, 0, 0, 0, 0, -1.91628895859597e-06, 0, -1.8994115184603e-06, 3280.49999998825, 0, 0, 0, 0, 0, 0, 1.10089527021126e-05, 0, 1.3121821385698e-05, 0, 0, 2.2272121062811e-05, -36495.56250396, -6971.0624934639, -61099.3125015786, 0, 0, 0, 0, 0, 0, -3.64404003854018e-05, 0, -4.5162111437224e-05, 0, 0, 0, -8.41634281695467e-05, 0, -0.000134200181291556, 66430.1250197884, 3690.56244634485, 384433.593756021, 70120.687506593, 672297.468794909, 122403.656206676, 0, 0, 0, 0, 0, 0, 7.37180681770312e-05, 0, 8.27763657963009e-05, 0, 0, 0, 0.00016897515979582, 0, 0.000399416998338923, -33215.0625418858, 0.000179611612078622, -1024131.09376663, 0, 0, 0.000389360963161765, -66430.1250784525, -1162527.18767369, -33215.0624232849, -60894.2809947759, -4146346.96890883, -686444.624783964, -1223421.4688371, 0, 0, 0, 0, 0, 0, -8.716240731859e-05, 0, -8.24371432424149e-05, 0, 0, 0, -0.000212370945744857, 0, -0.000594877683307665, 4.37851683329749e-05, -0.000304351134834342, 1145919.6562833, 0, 0, 0, -0.000499661204637469, 0, -0.000907115571814014, 0.000156010509398111, 548048.531539427, -0.000276520010479412, -0.000534077035049926, 10662035.0627866, 1444855.21827288, 1.32230965461164e-05, 6476937.18794001, 298935.56167063, 1096097.06301431, 548048.53103242, 6775872.75011922, 0, 0, 0, 0, 0, 0, 5.45166400815842e-05, 0, 4.21305556381658e-05, 0, 0, 0, 0.000155757661993361, 0, 0.000427262918771426, -2.1730116991972e-05, 0.000254382954282376, -448403.343783068, 0, 0, 0, 0.000323445714510836, 0, 0.000949169329796821, -0.000125522020183868, -0.000219987173735415, 0.000347329741185971, 0.000531708580267696, -11210083.594013, -896806.68705365, 0, 0, 0.000774809339868854, -9.28812884055621e-05, -2690420.06298128, 0.000153347295233917, 0.000961330893221085, -13452100.3129819, -448403.342530891, -0.000693519262138386, 0.000541268788734941, -13900503.6563183, -0.000249800979543576, -5380840.12575305, -2690420.06188643, 0, 0, 0, 0, 0, 0, -1.3853509592353e-05, 0, -8.65841766904646e-06, 0, 0, 0, -4.86087482490701e-05, 0, -0.000118004429307766, 3.90030081256498e-06, -8.25503813552415e-05, 1.19191716303726e-05, 0, 0, 0, -9.27211860357292e-05, 0, -0.000357434408528893, 3.64191210117061e-05, 6.22533693309127e-05, -0.000152061099445745, -0.000210538675594616, 4035630.09384415, -0.000147600752144662, 0, 0, 0, -0.000230996366550025, 0, -0.000421111965047708, 4.70405069078559e-05, 0.00018410560349848, -0.000120916470486223, -0.00038093736168865, 8071260.18769962, -0.000599618494943645, 0.000290852632064187, -0.000312372919194272, 8071260.18753603, -2.29980942721798e-05, 4035630.09399818, -0.000682308777349068, 0.000288236128929274, 0.00053666214230568, -8.63301557507305e-05, -0.000541227245630705, 8071260.18786553, 4035630.09327907, 8.1723770871995e-05}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -2.61988875014407e-08, 0, 0, 0, 0, 0, 0, 0, -6.93391743848375e-07, 0, 1.40288209188228e-06, -10206.0000031543, 0, 0, 0, 0, 0, 0, 8.38470011960743e-06, 0, -1.1850381308386e-05, 0, 0, -4.36400009289567e-06, 172590.750022574, 34809.7500450146, 126663.750022301, 0, 0, 0, 0, 0, 0, -3.65618037320539e-05, 0, 4.55020687636335e-05, 0, 0, 0, -1.39873414525953e-05, 0, 5.28096251683893e-05, -922640.625069753, -39366.0002587029, -352653.750065168, -548663.625236206, -2101980.37516432, -392019.750232766, 0, 0, 0, 0, 0, 0, 8.23737813787762e-05, 0, -9.35942642657915e-05, 0, 0, 0, 8.33932428245179e-05, 0, -0.000188512282713684, 1955998.12511203, 14762.250715266, 369056.250111256, 0, 0, -5.6257198667007e-05, 2516963.62552306, 10820729.2504933, 575727.751040651, 398580.751116944, 5513700.37546604, 752874.750408659, 6089428.12606467, 0, 0, 0, 0, 0, 0, -0.000103748606041086, 0, 0.000107203685612612, 0, 0, 0, -0.000141129765072049, 0, 0.00030563634530574, -1793613.3750944, -0.00102345503135758, -132860.250122755, 0, 0, 0, -6.43305235252571e-05, 0, 0.000261968710416532, -3786517.12556269, -21390500.2507364, -199290.377109206, -132860.252389336, -5380840.12570627, -398580.750263881, -2192194.12667955, -24844866.7511499, -398580.751627737, -27037060.8770119, -5779420.87845809, -10960970.6265694, 0, 0, 0, 0, 0, 0, 6.93303904252738e-05, 0, -6.49574107715423e-05, 0, 0, 0, 0.000106779583963077, 0, -0.000230817413847132, 597871.125036984, 0.000736123656372433, 7.7968212337734e-05, 0, 0, 0, 0.000122790417560523, 0, -0.000347653250832095, 1793613.37528055, 17936133.7505359, 0.00197110449525149, 0.00227810595931106, 1793613.37557764, 2.96320000151083e-05, 0, 0, -7.30250194502896e-05, 1793613.37618244, 35872267.501226, 597871.127253361, 0.00257658929940394, 20327618.2511913, 0.000826326931551683, 37665880.8767301, 1793613.37975707, 5380840.12591165, 21523360.5039274, 41253107.6272456, 5380840.12837801, 0, 0, 0, 0, 0, 0, -1.90541024659602e-05, 0, 1.63183272229947e-05, 0, 0, 0, -3.23681934205677e-05, 0, 6.62765426677747e-05, -4.24969139956321e-06, -0.000211378072630898, -2.00971661247011e-05, 0, 0, 0, -4.53889102684634e-05, 0, 0.000145953907115939, -4.81033975283141e-05, -5380840.12515105, -0.000692135138657113, -0.000800408423492945, -0.000195310167711905, 1.22046963465982e-05, 0, 0, 0, -3.09692904579341e-05, 0, 9.70958080465197e-05, -0.000294744444079622, -16142520.3754729, -0.00104827296724491, -0.00129411777484465, -5380840.12547701, -6.54106175319214e-05, -16142520.3755537, -0.00223723382584124, -0.000194467179250108, -0.000798856187227409, -16142520.3756414, -0.000823633920142508, -16142520.3765195, -32285040.7511034, -5380840.12759171, -0.00250266879866926, -16142520.3757795, -0.000892420995039231, -16142520.3770167}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 4.82375635214802e-08, 0, 0, 0, 0, 0, 0, 0, -2.71827897165949e-07, 0, -1.11103936424337e-06, 2916.00000104059, 0, 0, 0, 0, 0, 0, 4.1306320595568e-07, 0, 8.86945281673087e-06, 0, 0, 6.96365957737823e-06, -45562.5000089461, -6196.50001420291, -45562.5000068012, 0, 0, 0, 0, 0, 0, 4.87915325958023e-07, 0, -3.39716380197271e-05, 0, 0, 0, -1.80583058090022e-05, 0, -5.15252215781417e-05, 205031.250030469, 3280.50007710775, 205031.250009649, 90213.7500837142, 705307.500071932, 90213.7500718369, 0, 0, 0, 0, 0, 0, -3.07941579124478e-06, 0, 7.02783423529865e-05, 0, 0, 0, 2.91847604938229e-05, 0, 0.000166593031046524, -295245.000053447, -0.000202478384819809, -295244.99998488, 0, 0, 0.000117487284249141, -339531.750196281, -3107453.62524262, -44286.7503343614, -44286.7503359997, -3107453.62519459, -339531.75006277, -1306459.12543482, 0, 0, 0, 0, 0, 0, 6.26841440067243e-06, 0, -8.08640774361065e-05, 0, 0, 0, -3.75540066364867e-05, 0, -0.00026312285099835, 132860.250049631, 0.000275896690213486, 132860.249945458, 0, 0, 0, -0.000134744802065396, 0, -0.000305147340992399, 265720.500226553, 4251528.00038456, 0.000641652469328213, 0.00069478717487407, 4251528.00024762, 265720.499853472, 132860.25053945, 13020304.5005667, 132860.250348016, 4849399.12594147, 597871.126231936, 4849399.12566009, 0, 0, 0, 0, 0, 0, -6.11155637670873e-06, 0, 4.89334165983084e-05, 0, 0, 0, 2.97493102157394e-05, 0, 0.000199647355563672, -2.21490926552759e-05, -0.000189711090055503, 5.42124607378232e-05, 0, 0, 0, 9.38236244720685e-05, 0, 0.000339108457132486, -0.00012579747470992, -1793613.37529121, -0.000579514925521049, -0.000634747434479563, -1793613.37515381, 0.000269785004107951, 0, 0, 0.000232783723216926, -0.000375891716644479, -15544649.2506371, -0.000671389811195603, -0.00061922055733378, -15544649.2505503, 9.62214488222572e-05, -3587226.75090126, -0.00155873799742767, -3587226.75030977, -1793613.37651776, -17338262.6262033, -1793613.37620074, 0, 0, 0, 0, 0, 0, 2.24461433217029e-06, 0, -1.21832095121454e-05, 0, 0, 0, -9.09068265206405e-06, 0, -5.84550478998004e-05, 3.35706547226394e-06, 5.24383186422193e-05, -1.88659024149445e-05, 0, 0, 0, -3.30591665972616e-05, 0, -0.000133262638697938, 2.58032520571917e-05, 8.4597889090404e-05, 0.000198163150040118, 0.000212505016357834, 3.72631759761693e-05, -0.000117967845796517, 0, 0, 0, -6.91955184030378e-05, 0, -0.000134457166750714, 9.32587466881446e-05, 5380840.12525075, 0.000303070297444489, 0.000317264096869375, 5380840.12519645, -0.000191318509873282, 0.000319439690210791, 0.000697257028442025, 7.88654114266063e-07, 0.000225896925301103, 10761680.25032, 7.81965151084359e-05, 0.000635431332064115, 10761680.250656, 0.000878233515538682, 0.000802538511867331, 10761680.2504514, 0.000246315416585512, 5380840.12588164}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -2.75804805545797e-07, 0, 0, 0, 0, 0, 0, 0, 3.65648356813616e-06, 0, 4.28335799806117e-06, -10206.0000018963, 0, 0, 0, 0, 0, 0, -2.09054680398365e-05, 0, -3.04276507289787e-05, 0, 0, -4.17821307759103e-05, 126663.750022452, 34809.750016082, 172590.750017884, 0, 0, 0, 0, 0, 0, 7.04222049097148e-05, 0, 0.000109208404486466, 0, 0, 0, 0.000153863157378879, 0, 0.000250145428132038, -352653.750085267, -39366.0000412563, -922640.625043608, -392019.750175176, -2101980.37519727, -548663.62513123, 0, 0, 0, 0, 0, 0, -0.00014299538251014, 0, -0.000212073688689703, 0, 0, 0, -0.000326688237004189, 0, -0.000755075793056151, 369056.250159957, 14762.2500030867, 1955998.12503275, 0, 0, -0.000677683032508009, 752874.750512003, 5513700.37566429, 398580.750499389, 575727.750394915, 10820729.2505507, 2516963.62515, 6089428.12616994, 0, 0, 0, 0, 0, 0, 0.000166864846442914, 0, 0.000227442632397596, 0, 0, 0, 0.000443685465005023, 0, 0.00114953774471025, -132860.250157375, 0.000141646749659537, -1793613.37498309, 0, 0, 0, 0.000872828752805867, 0, 0.00157375604925295, -398580.750691392, -5380840.12604713, -132860.250652598, -199290.375542202, -21390500.250739, -3786517.12475888, -398580.751063046, -24844866.7514787, -2192194.12529544, -10960970.6278257, -5779420.87759976, -27037060.8771405, 0, 0, 0, 0, 0, 0, -0.000101822140318937, 0, -0.000127313944444432, 0, 0, 0, -0.0003414022952274, 0, -0.000847513647183138, 7.56248722664807e-05, -0.000199693519297286, 597871.124962033, 0, 0, 0, -0.00060714461239674, 0, -0.00166800121883164, 0.000447766981095488, 1793613.37578297, 0.00034963572001251, 0.000247991486214385, 17936133.7505023, 1793613.37450611, 0, 0, -0.00129921847108831, 0.000879095645631142, 20327618.2516143, 0.000972265034153915, 597871.125426526, 35872267.5013751, 1793613.37436321, 5380840.12786088, 1793613.37757447, 37665880.8765428, 5380840.12894343, 41253107.6288548, 21523360.5029082, 0, 0, 0, 0, 0, 0, 2.50589371225566e-05, 0, 2.91599994148212e-05, 0, 0, 0, 0.000107977958085973, 0, 0.000241223947696873, -1.32461670784011e-05, 8.19558820764051e-05, 1.61396880887916e-05, 0, 0, 0, 0.000193784117943735, 0, 0.000638431183045019, -0.000111599803434156, -0.000223159948700857, -4.09493236084693e-05, 1.69595352207983e-05, -5380840.12513743, 0.000218623978200452, 0, 0, 0, 0.00039117400788244, 0, 0.000706658261148259, -0.000256896074899826, -5380840.12562342, -0.000318931948312863, -0.000185855908139862, -16142520.3754769, 0.000589496685459351, -0.00104459050588409, -0.000899014567371974, -16142520.3754057, -0.000379473446884365, -16142520.3757581, 0.00025163452026725, -0.00187297074016428, -16142520.3770852, -0.00187557715896287, -5380840.12646493, -32285040.7516078, -16142520.3757821, -16142520.3775305}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.4733561112345e-07, 0, -8.74827364465661e-08, -410.062500236498, 0, 0, 0, 0, 0, 0, 1.13833415402542e-06, 0, 6.31957915251585e-07, 0, 0, 7.88450562118563e-07, 7176.09375132233, 4100.62500321066, 4100.62500201502, 0, 0, 0, 0, 0, 0, -4.56812145129551e-06, 0, -2.39861147313429e-06, 0, 0, 0, -4.07129587206104e-06, 0, -4.00532207708769e-06, -39981.093753124, -3690.56251883729, -3690.56250981451, -71350.8750161962, -71350.8750087553, -40596.1875164103, 0, 0, 0, 0, 0, 0, 1.01855688667069e-05, 0, 5.0086881537253e-06, 0, 0, 0, 1.15031614640197e-05, 0, 1.11013575295399e-05, 83037.6562530007, 5.42238610035523e-05, 2.83832407310833e-05, 0, 0, 1.41412554271671e-05, 393044.906284299, 393044.90626749, 60894.2813275894, 33215.062572356, 60894.2812802711, 33215.0625444425, 703052.156312178, 0, 0, 0, 0, 0, 0, -1.27641866565007e-05, 0, -5.86507116675045e-06, 0, 0, 0, -1.73272130974028e-05, 0, -1.60200772246756e-05, -49822.5937501597, -8.11861082325288e-05, -4.4860647591663e-05, 0, 0, 0, -2.68200633082732e-05, 0, -2.67937154318808e-05, -797161.50003144, -797161.500016263, -0.000162901379548694, -0.000156064601657379, -6.31424976385373e-05, -7.94100191752008e-05, -298935.562632267, -298935.56254436, -0.000122331719973822, -3836339.71884853, -548048.531475455, -548048.531345565, 0, 0, 0, 0, 0, 0, 8.40087242559752e-06, 0, 3.60154629600919e-06, 0, 0, 0, 1.33843749418268e-05, 0, 1.17924428564787e-05, -1.45742646085509e-06, 6.10582705590888e-05, 3.5331875269551e-05, 0, 0, 0, 2.37399533565409e-05, 0, 2.29659090148306e-05, 448403.343759738, 448403.343755691, 0.000155298136615035, 0.000154393881381138, 6.74731540231095e-05, 7.9464722269336e-05, 0, 0, 3.31080445788246e-05, 448403.343845617, 448403.343781794, 0.00018652052772131, 0.000171872217234532, 5.82372480217315e-05, 0.000112128521691711, 7622856.8438094, 0.0003300509772332, 9.12644303824741e-05, 2690420.06276258, 2690420.06258989, 0.000204767603371639, 0, 0, 0, 0, 0, 0, -2.2535571553015e-06, 0, -9.00188571538006e-07, 0, 0, 0, -4.21603880284959e-06, 0, -3.51026373894145e-06, 6.44038086917408e-07, -1.82572115922979e-05, -1.08220973719119e-05, 0, 0, 0, -7.80248425109863e-06, 0, -7.43735189701129e-06, 5.53155462201122e-07, -5.89241628612895e-08, -5.54705193953669e-05, -5.6702242910101e-05, -2.72969385935965e-05, -3.14663161564373e-05, 0, 0, 0, -1.46664904203514e-05, 0, -1.3653888369406e-05, -2.3356103115814e-05, -8.76865720710301e-06, -8.7218666706367e-05, -8.52558930183711e-05, -3.02133334598596e-05, -4.57032277332254e-05, -4035630.09375768, -0.000161245997017957, -4.53400663044245e-05, -7.09344465277022e-05, -2.06420429738726e-05, -7.10661733469093e-05, -4035630.09384577, -4035630.09377647, -0.000195931371883797, -0.000158425485619184, -3.43654305964627e-05, -6.65191369262373e-05, -0.000113320560823031}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 1.69176258810492e-08, 0, 0, 0, 0, 0, 0, 0, -2.35539411820647e-07, 0, -2.29437568595245e-07, -410.062500707086, 0, 0, 0, 0, 0, 0, 1.5244161576116e-06, 0, 1.46401052628364e-06, 0, 0, 2.28736897854538e-06, 4100.62500464359, 7176.0937582277, 4100.62500729134, 0, 0, 0, 0, 0, 0, -4.99943761119759e-06, 0, -5.34670137673857e-06, 0, 0, 0, -1.20238333831073e-05, 0, -9.78490745030953e-06, -3690.56251469132, -39981.0937878446, -3690.56253380718, -71350.8750468646, -40596.1875345956, -71350.8750649933, 0, 0, 0, 0, 0, 0, 9.21989344008059e-06, 0, 1.16837223382029e-05, 0, 0, 0, 3.0846535447306e-05, 0, 2.26866504052313e-05, 2.66389249367804e-05, 83037.6563386698, 8.83411827513892e-05, 0, 0, 3.9552786694165e-05, 60894.2813687137, 33215.062575085, 393044.906418931, 393044.90647753, 33215.0626110915, 60894.2814606657, 703052.156534452, 0, 0, 0, 0, 0, 0, -9.82946279104054e-06, 0, -1.47681684248903e-05, 0, 0, 0, -4.03788010746507e-05, 0, -3.14117713407623e-05, -2.82888777793916e-05, -49822.5938625608, -0.00012943231699161, 0, 0, 0, -7.7631320347559e-05, 0, -6.10718766474441e-05, -0.000156540203133675, -8.75212907744838e-05, -797161.500284878, -797161.500384412, -0.000202350912941711, -0.000364571764822979, -298935.562811996, -0.00015203095590282, -298935.563009611, -548048.531768652, -3836339.71952738, -548048.531866893, 0, 0, 0, 0, 0, 0, 5.72814704030602e-06, 0, 9.8700255272995e-06, 0, 0, 0, 2.59248313185221e-05, 0, 2.38839963554689e-05, 1.63489501824537e-05, 7.41203200117047e-05, 9.71255098888268e-05, 0, 0, 0, 7.05936896879499e-05, 0, 4.84210303154747e-05, 0.000105866828572919, 5.33651005529373e-05, 448403.3439803, 448403.344060964, 0.000194235233753168, 0.000330862238471346, 0, 0, 7.69327107829701e-05, 0.000263043561710955, 9.7369047723523e-05, 448403.344088742, 448403.344304591, 0.000166380598543379, 0.000522954543170681, 0.000435082907397391, 7622856.84461692, 0.00063616556924526, 2690420.06347186, 0.000679328636175008, 2690420.06358578, 0, 0, 0, 0, 0, 0, -1.42251048582788e-06, 0, -2.69003678596711e-06, 0, 0, 0, -6.47903657944723e-06, 0, -7.48476846283118e-06, -3.96014246623688e-06, -1.99633035390478e-05, -2.88685000907013e-05, 0, 0, 0, -2.37566925206001e-05, 0, -1.61588085361737e-05, -2.91778613163653e-05, -1.34139287077774e-05, -7.25440014011706e-05, -9.63326006131363e-05, -7.30747640763741e-05, -0.000119979123958106, 0, 0, 0, -3.51549102218858e-05, 0, -2.78195157768265e-05, -8.44974771986969e-05, -2.40112285773021e-05, -0.000132256895147073, -0.00022334138858207, -7.83290241529842e-05, -0.000214796102409328, -0.000139019258573241, -4035630.09409049, -0.000269399539893735, -0.000137510768759989, -4.60979808862333e-05, -0.000262281097466011, -0.000411014597311048, -0.000263138293461342, -4035630.09428914, -4035630.09435257, -0.000310026148899966, -0.000506580721560536, -0.000649375170442523}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.75984507680034e-07, 0, -4.22219044053301e-08, -410.062500104938, 0, 0, 0, 0, 0, 0, -1.70257149822977e-06, 0, 3.90984256575111e-07, 0, 0, 3.64110338548558e-07, 4100.62500134038, 4100.62500066347, 7176.09375184377, 0, 0, 0, 0, 0, 0, 7.40784859503009e-06, 0, -7.48560230612302e-07, 0, 0, 0, 1.95044163172079e-06, 0, -6.07450399914428e-06, -3690.56250583692, -3690.56249649325, -39981.0937559751, -40596.1875100085, -71350.8750121523, -71350.8750239165, 0, 0, 0, 0, 0, 0, -1.66551873882061e-05, 0, -2.37333223610516e-07, 0, 0, 0, -1.58905590736295e-05, 0, 1.67184800975164e-05, 1.30287347221618e-05, -2.6672648346204e-05, 83037.656255251, 0, 0, 1.70541432766037e-05, 33215.062530406, 60894.2812897657, 33215.0625170421, 60894.2813176494, 393044.906272152, 393044.906346515, 703052.156365438, 0, 0, 0, 0, 0, 0, 2.04005378792731e-05, 0, 2.12365571874245e-06, 0, 0, 0, 3.3816993694183e-05, 0, -1.80641123155706e-05, -1.59739686084657e-05, 5.87254682312874e-05, -49822.5937446109, 0, 0, 0, -7.38816447226428e-06, 0, -5.00889686331232e-05, -4.39091755250358e-05, -6.64492029069491e-05, 8.33788342510994e-06, -5.96795702233257e-05, -797161.499999633, -797161.500158033, -4.35984394572401e-05, -298935.562558395, -298935.562754267, -548048.531476941, -548048.531538532, -3836339.71901813, 0, 0, 0, 0, 0, 0, -1.2975746270576e-05, 0, -2.17471743665425e-06, 0, 0, 0, -3.01749361833125e-05, 0, 6.92360091113977e-06, 1.0061216313118e-05, -5.5066938550402e-05, -1.15250642222909e-05, 0, 0, 0, -1.26125038803463e-05, 0, 4.84377444840394e-05, 3.33708271453237e-05, 5.46285109695495e-05, -4.15089208627907e-05, -4.38382880691732e-06, 448403.34372103, 448403.343865263, 0, 0, 3.95214252013198e-05, 3.38606874161446e-05, 7.23670713984048e-05, 1.275237866346e-05, 0.000226957949276175, 448403.343763358, 448403.344063977, 0.000205973981948401, 0.000267796277364277, 7622856.84398311, 0.000367141125320203, 2690420.06282196, 2690420.06301832, 0, 0, 0, 0, 0, 0, 3.35553484871106e-06, 0, 6.95517821381899e-07, 0, 0, 0, 9.98732480756649e-06, 0, -6.32107969792681e-08, -2.51007523972669e-06, 1.89627615246668e-05, 5.10051433156048e-06, 0, 0, 0, 9.11567282557253e-06, 0, -1.52338731849123e-05, -1.05783496154128e-05, -1.74625828568917e-05, 2.53657736958463e-05, 1.95896364439262e-05, 1.74955725803503e-05, -2.9817793024764e-05, 0, 0, 0, -6.92682896539702e-06, 0, -2.38868921806904e-05, -1.00263385733984e-05, -3.17916412718634e-05, 9.26824460379174e-06, -6.16495728849703e-05, 1.29803897248131e-05, -0.000130723846921817, -7.3061739508408e-05, -7.64455244474734e-05, -4035630.09381797, -2.43103914296126e-06, -1.6775923957276e-05, -0.000153570684728117, -0.000149214256549443, -0.000143086217927499, -0.000181468832347666, -0.000272674990964001, -4035630.09387484, -4035630.09400902, -0.000302005692352999}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 1.30439993325187e-08, 0, 0, 0, 0, 0, 0, 0, -3.0252333451607e-07, 0, -3.52750609894197e-08, -364.500000375447, 0, 0, 0, 0, 0, 0, 2.79836081576621e-06, 0, 1.70810025879325e-07, 0, 0, 5.30834166882586e-07, 5285.25000173922, 5285.25000591571, 3645.00000266736, 0, 0, 0, 0, 0, 0, -1.17341526515185e-05, 0, -1.59385038219019e-06, 0, 0, 0, -8.70514532811265e-06, 0, 2.73681564472248e-06, -19683.0000041135, -19683.0000377435, -3280.50001170031, -76271.6250237335, -52488.0000085925, -52488.000027828, 0, 0, 0, 0, 0, 0, 2.57492240779331e-05, 0, 5.83329374351952e-06, 0, 0, 0, 3.76199597333287e-05, 0, -9.30349795063849e-06, 14762.2500056728, 14762.2501136277, 3.64762218404981e-05, 0, 0, 2.11951967981746e-06, 280482.750043394, 191909.250015768, 280482.750122619, 191909.250148652, 44286.7500222647, 44286.7500569489, 752874.75008316, 0, 0, 0, 0, 0, 0, -3.0892554401168e-05, 0, -9.52173180952231e-06, 0, 0, 0, -6.84980718906164e-05, 0, 7.16463296992184e-06, -4.08319034605804e-06, -0.000173092810231586, -6.51733337745786e-05, 0, 0, 0, -3.86800510905943e-05, 0, 2.3436100260149e-05, -199290.375041304, -132860.250021166, -199290.375277649, -132860.250355649, -5.30553065637332e-05, -8.38087942006325e-05, -996451.875166711, -132860.250017575, -132860.250211459, -2723635.12511384, -2723635.12536079, -597871.125085721, 0, 0, 0, 0, 0, 0, 1.9249571889718e-05, 0, 7.20496615179933e-06, 0, 0, 0, 5.6930845599978e-05, 0, 1.99594800492094e-06, 9.84757187141547e-07, 0.000129733956493694, 5.69420861437903e-05, 0, 0, 0, 6.28902691226374e-05, 0, -2.75642698024226e-05, 1.95863907714609e-05, 1.57656571699474e-05, 0.000281118132557412, 0.000373357564149677, 7.27195793380087e-05, 9.13988023286844e-05, 0, 0, -4.04439359497531e-07, 597871.125100448, 1.43770304494447e-05, 597871.125254063, 0.00036325814564879, 2.17341762355246e-05, 0.000134273194116821, 1793613.37508493, 1793613.3755866, 4.72410974235191e-05, 9565938.00033526, 1793613.37503946, 1793613.37529961, 0, 0, 0, 0, 0, 0, -4.87984351854454e-06, 0, -2.07096965524251e-06, 0, 0, 0, -1.78849248307123e-05, 0, -2.84186465222629e-06, 1.60699352145072e-07, -3.81074517844775e-05, -1.88545877088417e-05, 0, 0, 0, -2.93216523359812e-05, 0, 7.84828351392062e-06, -3.45603396832761e-06, -4.30667013228221e-06, -0.000104517175699139, -0.000141680106927846, -3.52727215005981e-05, -4.28645300649603e-05, 0, 0, 0, -1.39340039120723e-05, 0, 1.10012112720684e-05, -2.17284271439404e-05, -7.3904628710243e-06, -0.000130163171677173, -0.000203191877570046, -1.95779187292412e-05, -4.22479471887358e-05, -2.7064510560706e-05, -0.000308065658326946, -3.06033361490456e-05, -7.67561467496553e-05, -3.87703853878738e-06, -0.000114287092904104, -5380840.12511209, -1.24023686366946e-05, -5380840.12527512, -0.000285472944142256, 2.02789870568347e-05, -6.6465721344885e-05, -5380840.12511938}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -2.99923898991296e-08, 0, 0, 0, 0, 0, 0, 0, 3.72598506923832e-07, 0, 4.38348780576255e-07, -364.500000413986, 0, 0, 0, 0, 0, 0, -2.12676480836129e-06, 0, -2.80300959696452e-06, 0, 0, -3.86514606050789e-06, 3645.00000366156, 5285.25000551963, 5285.25000386671, 0, 0, 0, 0, 0, 0, 7.04384643630431e-06, 0, 9.4923219263387e-06, 0, 0, 0, 1.48559701488966e-05, 0, 1.87134639543138e-05, -3280.50001157922, -19683.0000217372, -19683.0000113547, -52488.0000396549, -52488.0000288823, -76271.6250525179, 0, 0, 0, 0, 0, 0, -1.36655785814877e-05, 0, -1.79564882968865e-05, 0, 0, 0, -3.47476722915272e-05, 0, -5.01717958260822e-05, 1.84971236462157e-05, 14762.2500384461, 14762.2500110728, 0, 0, -4.65067270616575e-05, 44286.7500972418, 44286.7500758822, 191909.25013501, 280482.750177441, 191909.250072189, 280482.750172391, 752874.750285065, 0, 0, 0, 0, 0, 0, 1.51588113701407e-05, 0, 1.90194021169057e-05, 0, 0, 0, 4.8477809312239e-05, 0, 7.31685480156722e-05, -1.57702495113466e-05, -3.21613423273133e-05, 4.4755254044076e-06, 0, 0, 0, 6.48841933136445e-05, 0, 8.40640924453253e-05, -0.00011233770637401, -9.70499250277509e-05, -132860.250208265, -199290.375251693, -132860.250069557, -199290.375257168, -132860.250251466, -132860.25014481, -996451.875470116, -597871.125531146, -2723635.12574099, -2723635.12566893, 0, 0, 0, 0, 0, 0, -8.84388796531918e-06, 0, -1.04922032028496e-05, 0, 0, 0, -3.62594900838988e-05, 0, -5.42757262893874e-05, 6.70311130153403e-06, 1.07210028874708e-05, -1.41078459950165e-05, 0, 0, 0, -5.03597637029287e-05, 0, -7.8022567888893e-05, 6.18953351415499e-05, 6.0084538484542e-05, 0.000150093913248466, 0.000156145254956964, 1.5724076849419e-05, 0.000179570112819572, 0, 0, -6.61239979559366e-05, 0.000190868105767329, 0.000127805289489578, 0.000269510072841549, 597871.125473021, 9.76102981618281e-05, 597871.125499819, 0.000432499958925002, 1793613.37575222, 1793613.37564815, 1793613.37592185, 1793613.37580831, 9565938.00120934, 0, 0, 0, 0, 0, 0, 2.09226447784017e-06, 0, 2.3324502055167e-06, 0, 0, 0, 1.10605653767089e-05, 0, 1.60082483548641e-05, -1.06845312293102e-06, -4.14085444922859e-07, 6.41584352971026e-06, 0, 0, 0, 1.6783688393788e-05, 0, 2.84127398467239e-05, -1.29796117512519e-05, -1.42878955963607e-05, -4.08320664431131e-05, -3.24849753879023e-05, 7.33952176298022e-06, -4.66380820617271e-05, 0, 0, 0, 2.10853296395613e-05, 0, 2.80059967051722e-05, -5.07349626534931e-05, -4.11239651442269e-05, -0.000103342401212027, -0.000162312562263406, -1.48007368025783e-05, -0.000189556357823069, -0.000130289611158883, -0.000263735506206702, -0.000225224531642518, -0.000103777091252294, -4.59571181496056e-05, -0.00025781002638552, -0.000369902786319867, -0.000327403107899991, -0.000477472358735696, -5380840.12562525, -0.000376889893870123, -5380840.12561322, -5380840.12572882}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.30295154021083e-08, 0, 0, -364.500000246885, 0, 0, 0, 0, 0, 0, -2.60701529293629e-07, 0, 3.96241273328011e-08, 0, 0, 5.66876833789308e-07, 5285.25000175305, 3645.00000339886, 5285.25000222942, 0, 0, 0, 0, 0, 0, 1.51427919721792e-06, 0, 8.56707364294774e-08, 0, 0, 0, -1.52739393212406e-06, 0, -5.2930934767643e-06, -19683.0000051496, -3280.50001578084, -19683.0000071821, -52488.0000197287, -76271.6250129046, -52488.0000282038, 0, 0, 0, 0, 0, 0, -3.82506179640437e-06, 0, -8.60875462536015e-07, 0, 0, 0, -1.88735819977517e-06, 0, 1.53095571653514e-05, 14762.2500075825, 3.47095363476507e-05, 14762.2500124087, 0, 0, 1.99773404136364e-05, 191909.250043941, 280482.75003492, 44286.7500756737, 44286.7501092258, 280482.750026679, 191909.250083145, 752874.750124178, 0, 0, 0, 0, 0, 0, 4.77780703785722e-06, 0, 1.82126034804288e-06, 0, 0, 0, 1.10980004908745e-05, 0, -2.08575049942147e-05, -5.345709605071e-06, -3.91030449721664e-05, -1.26052453329068e-05, 0, 0, 0, -2.72281959235825e-05, 0, -4.77785112787792e-05, -132860.250046334, -199290.375048945, -0.000131432840853885, -0.000186626164888755, -199290.375022832, -132860.250124378, -132860.250123673, -996451.875054168, -132860.250247579, -2723635.12522569, -597871.125373851, -2723635.12522587, 0, 0, 0, 0, 0, 0, -2.89332534250648e-06, 0, -1.62499521122536e-06, 0, 0, 0, -1.32298048575664e-05, 0, 1.40098336782586e-05, 1.18095340308159e-06, 2.17302803042034e-05, 6.74843982825674e-06, 0, 0, 0, 1.21083751889804e-05, 0, 4.46411029967466e-05, 2.23662014643789e-05, 3.35529909374589e-05, 0.00010576364659121, 0.000143870788769537, 8.35290312090053e-06, 9.86406309141353e-05, 0, 0, 5.26561388394514e-05, 8.28495646332127e-05, 597871.125055868, 0.000143450496256661, 0.000291760950884482, 597871.125027178, 0.000242091682233676, 1793613.37518748, 0.000445264744174446, 1793613.37517096, 1793613.37544009, 9565938.00025953, 1793613.37548476, 0, 0, 0, 0, 0, 0, 6.75802590637601e-07, 0, 5.39293014999567e-07, 0, 0, 0, 4.97189420657065e-06, 0, -3.72383256940602e-06, 2.26783394004113e-07, -4.72041165095803e-06, -1.3229843856637e-06, 0, 0, 0, 4.18566640727568e-07, 0, -1.51550764741712e-05, -3.54905360212633e-06, -8.73584093719296e-06, -3.20197346008775e-05, -4.05371763206268e-05, -8.80002114234061e-07, -3.25943400987708e-05, 0, 0, 0, -1.86715302952212e-05, 0, -2.56859670171384e-05, -1.87710060282726e-05, -2.15332748631683e-05, -5.87976668352966e-05, -0.000117524185144314, -2.61526692838159e-06, -9.11141981012202e-05, -5.8606488113175e-05, -0.00018201501757488, -5.17482154467143e-05, -4.70882435402699e-05, -1.59449031155829e-05, -0.000142716580393893, -0.000172401137993256, -5380840.12511023, -0.000259754083474838, -0.000302485337257452, -5380840.1250823, -0.00019305015122936, -5380840.12527639}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 1.84690741796547e-07, 0, 0, 0, 0, 0, 0, 0, -2.42979222924597e-06, 0, -3.04706597088913e-06, 22963.5000119883, 0, 0, 0, 0, 0, 0, 9.38166137241702e-06, 0, 1.68756625099152e-05, 0, 0, 3.99734924990408e-05, -112767.187587378, -319438.687652624, -319438.687626739, 0, 0, 0, 0, 0, 0, -1.14252836574061e-05, 0, -4.42024204138396e-05, 0, 0, 0, -0.000165226281777338, 0, -0.000204825230180672, 206056.406525415, 1220961.09442435, 1220961.09426637, 1427017.5009552, 1427017.50070059, 4301965.68889547, 0, 0, 0, 0, 0, 0, -1.18909130433117e-05, 0, 5.93089324560305e-05, 0, 0, 0, 0.000309761338201332, 0, 0.000486311048194034, -166075.312967229, -2020582.97022026, -2020582.96987198, 0, 0, 0.000728504365810874, -2352733.59615687, -2352733.59542349, -4207241.25337063, -15195891.0987364, -4207241.25200672, -15195891.0983237, -17548624.6941539, 0, 0, 0, 0, 0, 0, 4.3843528567485e-05, 0, -4.08889078909133e-05, 0, 0, 0, -0.000275955792730421, 0, -0.000572843063153388, 49822.5941917722, 1544500.40797183, 1544500.40764552, 0, 0, 0, -0.00111083281501175, 0, -0.00138849584440818, 1693968.19053177, 1693968.18957482, 4683323.81793993, 22868570.539414, 4683323.81536217, 22868570.5384776, 4782969.00628868, 4782969.00327396, 42648140.2621782, 26256506.9184761, 47431109.2681763, 47431109.2652584, 0, 0, 0, 0, 0, 0, -4.1013429614955e-05, 0, 1.25588961469507e-05, 0, 0, 0, 0.000112060621321554, 0, 0.000332250667514904, -0.000218755053506039, -448403.344793703, -448403.344673951, 0, 0, 0, 0.000709925339119537, 0, 0.00113557773666374, -448403.345652134, -448403.345039826, -1793613.37917182, -15694117.0375583, -1793613.3771357, -15694117.036984, 0, 0, 0.00154860347797186, -1793613.38017325, -1793613.37757235, -2690420.06918666, -44840334.3883339, -2690420.06524364, -44840334.3873488, -17039327.0726181, -47530754.4572664, -47530754.452381, -48427561.1472635, -48427561.1428046, -91474282.1533687, 0, 0, 0, 0, 0, 0, 1.33323673030189e-05, 0, -8.0616166065489e-07, 0, 0, 0, -1.65457010529297e-05, 0, -7.77739046795604e-05, 4.40027482294604e-05, 0.000259493178026472, 0.000249356297811434, 0, 0, 0, -0.000150752546151569, 0, -0.000330103899785369, 0.000475920496800338, 0.000316370934659835, 0.00123809350426249, 4035630.09560371, 0.000663970055155238, 4035630.09557397, 0, 0, 0, -0.000551331635635809, 0, -0.000683577464769929, 0.0015843573754767, 0.000780399783320359, 0.00254383719464606, 16142520.380242, 0.000896485780004354, 16142520.3796606, 4035630.09688311, 16142520.3825384, 16142520.3804686, 0.0027050536861406, 0.000990324133222926, 24213780.5690564, 16142520.3841265, 16142520.382168, 24213780.5745012, 48427561.14067, 24213780.5706473, 48427561.1387529, 48427561.1418364}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -8.80717380503965e-08, 0, 0, 0, 0, 0, 0, 0, 1.02780539811014e-06, 0, 1.19969165927776e-06, 3280.50000283796, 0, 0, 0, 0, 0, 0, -7.6030967905541e-06, 0, -8.43836904910897e-06, 0, 0, -6.49259425544964e-06, -6971.06251766388, -61099.3125336529, -36495.5625312585, 0, 0, 0, 0, 0, 0, 2.97315616878409e-05, 0, 3.39402296637651e-05, 0, 0, 0, 3.53876251301495e-05, 0, 2.04735589754146e-05, 3690.56255213554, 384433.5939086, 66430.1251498625, 122403.656438376, 70120.6876435317, 672297.469027799, 0, 0, 0, 0, 0, 0, -6.3099639269407e-05, 0, -7.76269200254604e-05, 0, 0, 0, -0.000118985070667631, 0, -5.34582235062079e-05, -8.46274265797377e-05, -1024131.0941306, -33215.0628951504, 0, 0, -4.96013756766218e-05, -60894.2817187518, -33215.0628012227, -686444.625690052, -4146346.96974849, -66430.1254889191, -1162527.18841355, -1223421.46995523, 0, 0, 0, 0, 0, 0, 7.40927222066741e-05, 0, 9.85890508422611e-05, 0, 0, 0, 0.00019895682316719, 0, 0.000102721317948006, 7.73037624355298e-05, 1145919.65674487, 0.000582142688716332, 0, 0, 0, 0.000146348807496376, 0, 1.31895584370925e-05, 0.000596609621749992, 0.000327021230807697, 1444855.21991541, 10662035.064246, 0.000887277708005909, 548048.532806613, 298935.563780711, 0.000691704621375515, 6476937.18975723, 548048.533379042, 6775872.75331553, 1096097.06529884, 0, 0, 0, 0, 0, 0, -4.55161818088713e-05, 0, -6.4613397914246e-05, 0, 0, 0, -0.000157414847045744, 0, -0.00010199017631191, -3.73618821573707e-05, -448403.344083359, -0.000441225871537772, 0, 0, 0, -0.000195322232149739, 0, 1.1135451640133e-07, -0.000380091128477817, -0.000175857527162619, -896806.688441535, -11210083.5952173, -0.000829309708929581, -0.00138078276336688, 0, 0, -2.83813848411253e-05, -0.00109002432764569, -0.000456702695217594, -448403.345151569, -13452100.3150415, -0.000774195046369058, -2690420.06475274, -0.00169889402897228, -13900503.6599639, -0.00294404001309935, -2690420.06650176, -0.00310461443744177, -5380840.12996644, 0, 0, 0, 0, 0, 0, 1.1452603032495e-05, 0, 1.70359366540402e-05, 0, 0, 0, 4.75066151792117e-05, 0, 3.78732121342866e-05, 7.45328133640613e-06, 9.14807676404899e-05, 0.000132933116417466, 0, 0, 0, 8.59690170392677e-05, 0, 8.70447073699593e-06, 9.65651965548627e-05, 3.65946036560938e-05, 0.000296996251635019, 4035630.09422394, 0.000306274200191297, 0.000494964426234334, 0, 0, 0, 4.173824780624e-05, 0, -2.03086588403795e-05, 0.000349288338311408, 0.000112616066733593, 0.000544819447834815, 8071260.18855867, 0.00034664581137246, 0.000878015343350736, 0.000505206721424427, 8071260.18897183, 0.00121029818954773, 0.000589364605949917, 0.000240080736530704, 4035630.09492384, 0.00163885128723626, 0.001176039804939, 4035630.09596317, 8071260.19026087, 0.00152719298847912, 0.00241584812782621, 0.00291246564874163}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 6.413015650546e-08, 0, 0, 0, 0, 0, 0, 0, -1.17610322663978e-06, 0, -6.46349075304621e-07, 3280.50000095666, 0, 0, 0, 0, 0, 0, 8.4095098975783e-06, 0, 3.20299978323135e-06, 0, 0, 9.97976639714544e-06, -6971.06250654954, -36495.5625092702, -61099.312517385, 0, 0, 0, 0, 0, 0, -2.72414220845487e-05, 0, -1.04054439581292e-05, 0, 0, 0, -6.84949610472392e-05, 0, -3.28890337837211e-05, 3690.56251972779, 66430.1250153251, 384433.59383281, 70120.6875645334, 122403.656334165, 672297.468934608, 0, 0, 0, 0, 0, 0, 4.73317634617466e-05, 0, 2.34474771767081e-05, 0, 0, 0, 0.000184738312777041, 0, 4.65004816629011e-05, -3.33643289459195e-05, -33215.0624746124, -1024131.09392691, 0, 0, 0.000221854778745709, -33215.0626740223, -60894.2814221559, -66430.1251335275, -1162527.18803242, -686444.625292728, -4146346.96952814, -1223421.46957338, 0, 0, 0, 0, 0, 0, -4.70557993050697e-05, 0, -3.33366975165568e-05, 0, 0, 0, -0.000233041678607904, 0, -2.3820413667655e-05, 3.22910978721933e-05, -9.79873301371968e-05, 1145919.65643909, 0, 0, 0, -0.000496189216518343, 0, -0.000316828777007633, 0.000240447493872035, 0.000182689305967074, 4.16812745523503e-05, 548048.53190772, 1444855.21919682, 10662035.0638656, 0.000295121977313114, 298935.562895349, 6476937.18932687, 548048.53269309, 1096097.06445795, 6775872.75239061, 0, 0, 0, 0, 0, 0, 2.55607854137788e-05, 0, 2.54054148329164e-05, 0, 0, 0, 0.000142836839494197, 0, 9.74108619924914e-07, -1.67064853644232e-05, 0.000100360916782036, -448403.3438438, 0, 0, 0, 0.000431165590550445, 0, 0.000183861862361797, -0.000167474702845976, -9.71129577151591e-05, 0.000102908405751685, -0.000353420976943159, -896806.68782115, -11210083.5948571, 0, 0, 0.000549219505991354, -0.000287823340110314, -0.000238628328384634, -0.000106390687286109, -2690420.06420267, -448403.344106606, -13452100.3146674, -0.00115878514308765, -0.00175462839487075, -13900503.6588682, -0.00235053995319893, -2690420.0651226, -5380840.12911881, 0, 0, 0, 0, 0, 0, -5.89245375752088e-06, 0, -7.73472769280353e-06, 0, 0, 0, -3.52875776041817e-05, 0, 3.82977153697432e-07, 3.58319576289448e-06, -3.46803387206497e-05, 1.51756488993967e-05, 0, 0, 0, -0.000129445621296398, 0, -3.11965725567004e-05, 4.67836986320015e-05, 2.0299252862278e-05, -6.95686509679147e-05, 5.80294269031736e-05, 8.7643677769795e-05, 4035630.09409001, 0, 0, 0, -0.000258791619180811, 0, -0.000198052784414937, 0.00010557044631398, 5.15467700790099e-05, -3.85189227087308e-05, 0.000570331221696149, 0.000116167992792474, 8071260.18838567, 0.000349804864401767, 0.00052138111539433, 8071260.18850359, 6.46408047914464e-05, 8.71483832605417e-05, 4035630.09476113, 0.000960887107030872, 0.000999363454212205, 0.00105727833346357, 0.00205999215784006, 4035630.09511368, 8071260.18981083, 0.00234587915537893}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 1.09982168050412e-08, 0, 0, 0, 0, 0, 0, 0, -2.29321133042019e-07, 0, 4.46625947977357e-07, -10206.0000075139, 0, 0, 0, 0, 0, 0, 7.03608993208871e-06, 0, -1.27504318392668e-06, 0, 0, -1.44342371302889e-05, 34809.7500530742, 172590.750094858, 126663.750076241, 0, 0, 0, 0, 0, 0, -4.09879758274238e-05, 0, -8.80552260061853e-06, 0, 0, 0, 3.73074099762993e-05, 0, 0.000100928426224923, -39366.0001662484, -922640.625443073, -352653.75032653, -548663.625578969, -392019.750409954, -2101980.3757747, 0, 0, 0, 0, 0, 0, 0.000104971001657757, 0, 4.69997950355106e-05, 0, 0, 0, 1.8006294648476e-05, 0, -0.00025192991174328, 14762.2502811974, 1955998.12603925, 369056.250779833, 0, 0, -0.000340590909866409, 575727.751472614, 398580.750954219, 2516963.62710336, 10820729.2528235, 752874.751242578, 5513700.37743904, 6089428.12871978, 0, 0, 0, 0, 0, 0, -0.000137951186370314, 0, -8.3636971531482e-05, 0, 0, 0, -0.000158417059804445, 0, 0.000272357971537879, -0.000266182090529593, -1793613.37631966, -132860.251076335, 0, 0, 0, 0.00037497852746179, 0, 0.000790941434007525, -199290.376881854, -132860.251140148, -3786517.12851953, -21390500.2548421, -398580.751970846, -5380840.12887548, -2192194.1289646, -398580.751977841, -24844866.7563765, -5779420.88187163, -27037060.8852217, -10960970.6333853, 0, 0, 0, 0, 0, 0, 9.17345931627485e-05, 0, 6.56348844575903e-05, 0, 0, 0, 0.000180488561511012, 0, -0.000121370176261798, 0.000133296035264233, 597871.125870713, 0.000784749646536263, 0, 0, 0, -8.65763219196481e-05, 0, -0.00069938370487464, 0.00119628766777012, 0.000676956231437343, 1793613.37781491, 17936133.7539587, 0.00165399110142231, 1793613.37819803, 0, 0, -0.000776360257074646, 597871.12834547, 0.00148157362127173, 1793613.3793315, 35872267.5071743, 0.001889629522337, 20327618.2561926, 1793613.38068314, 37665880.886305, 5380840.13330788, 21523360.512656, 5380840.13486911, 41253107.640308, 0, 0, 0, 0, 0, 0, -2.45731376196819e-05, 0, -1.93667188673832e-05, 0, 0, 0, -6.40330395065811e-05, 0, 1.4187008249869e-05, -2.76174747524497e-05, -0.000235247268780321, -0.000230882217062064, 0, 0, 0, -4.29834897445099e-05, 0, 0.000208327735265173, -0.000302688743487891, -0.000157627363254164, -0.000876951704874512, -5380840.12624082, -0.000569394003358144, -0.00107882530327279, 0, 0, 0, 0.00021833257544242, 0, 0.000396624533905152, -0.00105303012839941, -0.000420164068419637, -0.00169071644396705, -16142520.3779379, -0.000727182145484046, -5380840.12730745, -0.00174931903444273, -16142520.3794125, -0.00317122334527421, -0.00180424398883986, -0.000659912216264369, -16142520.3783029, -5380840.13025245, -0.00395262452855877, -16142520.3819044, -32285040.7584867, -0.00461076064170348, -16142520.3823098, -16142520.3842448}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 1.19106960323255e-07, 0, 0, 0, 0, 0, 0, 0, -1.48001168587519e-06, 0, -2.03378265025246e-06, 2916.00000156326, 0, 0, 0, 0, 0, 0, 6.76581106861969e-06, 0, 1.2442278505401e-05, 0, 0, 2.30376694213971e-05, -6196.50001521646, -45562.5000220614, -45562.5000136378, 0, 0, 0, 0, 0, 0, -1.40270164365115e-05, 0, -3.84962824080061e-05, 0, 0, 0, -9.90864729270162e-05, 0, -0.000117487369799943, 3280.50005339808, 205031.250094762, 205031.250037623, 90213.7501709211, 90213.7501134837, 705307.500191242, 0, 0, 0, 0, 0, 0, 1.20348790046314e-05, 0, 6.76853301438905e-05, 0, 0, 0, 0.000204677243916043, 0, 0.000283517514030674, -9.56791476400588e-05, -295245.000197112, -295245.000033634, 0, 0, 0.000416106324806684, -44286.7504728866, -44286.7503119473, -339531.75058332, -3107453.62568261, -339531.750285121, -3107453.6255705, -1306459.1261631, 0, 0, 0, 0, 0, 0, 6.04308861657275e-07, 0, -6.94744311042968e-05, 0, 0, 0, -0.00021561560448632, 0, -0.000355517959444244, 9.42137456248426e-05, 132860.250222777, 132860.249985365, 0, 0, 0, -0.000671904309692586, 0, -0.000761427297053305, 0.000630765909574576, 0.000407666517061843, 265720.500898408, 4251528.0010702, 265720.500291664, 4251528.00075972, 132860.251212992, 132860.250602002, 13020304.5016038, 597871.127323058, 4849399.12807757, 4849399.12764474, 0, 0, 0, 0, 0, 0, -7.25580259353574e-06, 0, 3.89420598307701e-05, 0, 0, 0, 0.000112404181612801, 0, 0.000227958889847006, -4.91872619801965e-05, -0.000133959778947659, 4.08445895538289e-05, 0, 0, 0, 0.000491031330566109, 0, 0.000627583909841918, -0.000410562278923455, -0.000255610388026116, -0.000656923432999061, -1793613.37578503, -0.000102575112150432, -1793613.37547464, 0, 0, 0.000848386806454479, -0.00108173410018816, -0.000534180788412469, -0.00125270618239762, -15544649.2517256, -0.000441067447179767, -15544649.2514805, -0.00200323658592669, -3587226.75318812, -3587226.75243906, -1793613.37905739, -1793613.37847176, -17338262.6298771, 0, 0, 0, 0, 0, 0, 3.23677996441357e-06, 0, -9.18823826127092e-06, 0, 0, 0, -2.3124368267179e-05, 0, -5.99471648223543e-05, 1.0785789131747e-05, 3.42939867391724e-05, -1.80666294306279e-05, 0, 0, 0, -0.000132560189330357, 0, -0.000192820032806643, 0.000105389828625929, 6.21450314069852e-05, 0.000187203382668212, 0.000218600778028075, -3.56108535398809e-06, 0.000112245215156022, 0, 0, 0, -0.000321731857630793, 0, -0.000357592130079778, 0.000352233688596438, 0.000168462073730291, 0.000460497632060102, 5380840.12565743, 9.54794764320623e-05, 5380840.1254865, 0.000632053658253373, 0.00115507092306462, 0.000796484095954959, 0.000562307495132323, 0.000207348863893778, 10761680.2507931, 0.0017419474307416, 0.00148373645598267, 0.00208692708407707, 10761680.2525951, 0.00159215061808055, 10761680.2523451, 5380840.12814064}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -2.56951177190658e-07, 0, 0, 0, 0, 0, 0, 0, 3.79702485148691e-06, 0, 3.44832115646655e-06, -10206.0000038206, 0, 0, 0, 0, 0, 0, -2.34623411109316e-05, 0, -2.02314146764081e-05, 0, 0, -4.01204722363703e-05, 34809.7500324494, 126663.750046964, 172590.750047516, 0, 0, 0, 0, 0, 0, 7.17898189311539e-05, 0, 6.6169626796058e-05, 0, 0, 0, 0.000210297037808184, 0, 0.000171577716306666, -39366.0001085445, -352653.750158465, -922640.625180816, -392019.75034656, -548663.625293971, -2101980.37558478, 0, 0, 0, 0, 0, 0, -0.00012086632365108, 0, -0.000129755996773966, 0, 0, 0, -0.000520732401757542, 0, -0.000374706342046223, 14762.2501920607, 369056.25021671, 1955998.12530453, 0, 0, -0.000725657001820522, 398580.750929256, 575727.750729151, 752874.751048968, 5513700.37689914, 2516963.62582062, 10820729.2521102, 6089428.12797259, 0, 0, 0, 0, 0, 0, 0.000115986742332747, 0, 0.000151920614086689, 0, 0, 0, 0.000650016772435032, 0, 0.000445717058506061, -0.000189038977278911, -132860.250094062, -1793613.37523242, 0, 0, 0, 0.00138073940236026, 0, 0.00117291202541761, -132860.251225222, -199290.375920314, -398580.751354784, -5380840.12766108, -3786517.12597465, -21390500.2533346, -398580.752139493, -2192194.12642198, -24844866.7554675, -5779420.8805666, -10960970.6326102, -27037060.882435, 0, 0, 0, 0, 0, 0, -5.97818549008983e-05, 0, -9.68773593660281e-05, 0, 0, 0, -0.000407271757570866, 0, -0.000287064484302867, 9.80722145078899e-05, -4.15664736190882e-05, 597871.125051202, 0, 0, 0, -0.0011418575036608, 0, -0.000863888722973927, 0.000801503099617561, 0.000578928430835623, 0.000762610682467635, 1793613.37668079, 1793613.37549116, 17936133.7524757, 0, 0, -0.00156028804887519, 0.00186816581849693, 597871.126146085, 0.00191315837758007, 20327618.2554711, 1793613.37602854, 35872267.5059154, 1793613.37965967, 5380840.13251861, 37665880.8823643, 5380840.13452706, 21523360.508848, 41253107.6385094, 0, 0, 0, 0, 0, 0, 1.27962584775294e-05, 0, 2.55934645696036e-05, 0, 0, 0, 0.000103948702024937, 0, 8.0721018813338e-05, -2.09264745508333e-05, 3.38061084834119e-05, 1.37828764518587e-05, 0, 0, 0, 0.000342561522084523, 0, 0.000237435284210159, -0.000209186873403116, -0.000144197292378209, -0.000142149839009943, -0.000373863801232443, -7.09381549794378e-05, -5380840.12570137, 0, 0, 0, 0.000669425664897301, 0, 0.000600316014715287, -0.000603334037953219, -0.000349679325681404, -0.000586283809168706, -5380840.1269292, -0.00022350483243392, -16142520.3772547, -0.00145540002823936, -0.00255000143536948, -16142520.377587, -0.000838370376848294, -0.000391455761183219, -16142520.3779704, -0.0039575500808328, -5380840.12859597, -0.00472286737998958, -16142520.3820541, -16142520.3791893, -32285040.756955, -16142520.3830921}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.6758713911647e-07, 0, 1.25321861248338e-07, -7654.50000586554, 0, 0, 0, 0, 0, 0, 5.03329674894493e-06, 0, -3.07156967605122e-07, 0, 0, 2.44879796266139e-06, 94997.8125395342, 94997.8125773885, 94997.8125557594, 0, 0, 0, 0, 0, 0, -1.95361461797013e-05, 0, -2.09864181190291e-06, 0, 0, 0, -3.47404037373971e-05, 0, -9.429592048727e-06, -264490.312618149, -264490.312875982, -264490.312723338, -1148995.12543773, -1148995.12529037, -1148995.12560812, 0, 0, 0, 0, 0, 0, 3.9564731867768e-05, 0, 1.32028729261376e-05, 0, 0, 0, 0.000106949745878365, 0, 1.15104575789555e-05, 276792.187690671, 276792.188399848, 276792.188012393, 0, 0, 0.000142531463464583, 2945068.87600896, 2945068.87569999, 2945068.87670231, 2945068.87732916, 2945068.875779, 2945068.87686976, 13584960.565163, 0, 0, 0, 0, 0, 0, -4.56267766740509e-05, 0, -2.665002321803e-05, 0, 0, 0, -0.000139795660161395, 0, -9.83704076814342e-06, -99645.1876696434, -99645.1886389556, -99645.1881909489, 0, 0, 0, -0.000368999174502976, 0, -0.000218503947683508, -2790065.25115492, -2790065.25090227, -2790065.2530563, -2790065.2541138, -2790065.25112779, -2790065.25298192, -5380840.12787778, -5380840.12627028, -5380840.13021648, -32185395.5672251, -32185395.5703341, -32185395.5679432, 0, 0, 0, 0, 0, 0, 2.87672042352519e-05, 0, 2.32092863406287e-05, 0, 0, 0, 8.14873935919417e-05, 0, 1.12983345490005e-05, 7.63058640932291e-05, 0.000733154661252194, 0.000494073666524055, 0, 0, 0, 0.000362611720180639, 0, 0.000153770054919579, 896806.688154461, 896806.688088602, 896806.690077521, 896806.690895914, 896806.688417561, 896806.689993203, 0, 0, 0.000409621427919848, 2690420.06459622, 2690420.06357053, 2690420.06592047, 2690420.06859082, 2690420.06353029, 2690420.06769073, 27801007.3163393, 27801007.3218114, 27801007.3175432, 53808401.2591989, 53808401.2560416, 53808401.2607242, 0, 0, 0, 0, 0, 0, -7.72554203955735e-06, 0, -7.48093162165855e-06, 0, 0, 0, -1.68950846206638e-05, 0, -5.88469353761294e-06, -1.28739226195845e-05, -0.000190037751199574, -0.000142406867439588, 0, 0, 0, -0.000122382129868979, 0, -4.51889511007982e-05, -0.000146470534250366, -0.000150558580792285, -0.000826917022184023, -0.00105620655968213, -0.000316999873398142, -0.000846419973161941, 0, 0, 0, -0.000214927305657369, 0, -0.00014850883763619, -0.000555622245970247, -0.000359483478510669, -0.00144536137627952, -0.00253328591907133, -0.000357511400065706, -0.00201143439632304, -8071260.18868093, -8071260.19135974, -8071260.18939997, -0.00122114801540435, -0.000395258871990132, -0.00290505556652751, -24213780.5661151, -24213780.5649156, -24213780.5679336, -24213780.5689163, -24213780.564979, -24213780.5672879, -48427561.131096}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -3.73741218690392e-08, 0, 0, 0, 0, 0, 0, 0, 7.81778705396896e-07, 0, 2.57479699635365e-07, 2187.00000148929, 0, 0, 0, 0, 0, 0, -6.47931604768339e-06, 0, -1.77144088925485e-06, 0, 0, -3.64812880900239e-06, -34171.8750082763, -24330.3750215605, -24330.3750121455, 0, 0, 0, 0, 0, 0, 2.59034355694066e-05, 0, 7.73960075158793e-06, 0, 0, 0, 2.62276746241058e-05, 0, 1.33968078567399e-05, 153773.437520835, 44286.7501247122, 44286.7500515951, 375207.187602832, 375207.187552695, 265720.500124071, 0, 0, 0, 0, 0, 0, -5.608153582722e-05, 0, -1.89666495813939e-05, 0, 0, 0, -8.46454224014215e-05, 0, -3.41844903454307e-05, -221433.750026068, -22143.375349683, -22143.3751402729, 0, 0, -6.95296428509942e-05, -1638609.75020938, -1638609.75011558, -642157.875477971, -442867.500564451, -642157.875144697, -442867.500318765, -4052237.62545897, 0, 0, 0, 0, 0, 0, 6.78082197904466e-05, 0, 2.51856657816229e-05, 0, 0, 0, 0.000132327107970224, 0, 5.61010965572832e-05, 99645.1875135598, 0.000507293818379082, 0.000222302860504377, 0, 0, 0, 0.000171368077337138, 0, 0.00010075413421978, 2192194.1252025, 2192194.12514018, 298935.563492572, 199290.376195836, 298935.56276282, 199290.375503182, 2391484.50073863, 2391484.50020962, 398580.750994002, 17238617.4382377, 6377292.00164004, 6377292.00070016, 0, 0, 0, 0, 0, 0, -4.31496082680657e-05, 0, -1.68514247631459e-05, 0, 0, 0, -0.000100369039880054, 0, -5.01328280462274e-05, 7.54182129119417e-07, -0.000368124332993521, -0.000178779283566742, 0, 0, 0, -0.000178854359279378, 0, -7.81418144432485e-05, -896806.687584871, -896806.687584572, -0.000940944790733613, -0.00115239844858222, -0.00028075982708526, -0.000473558211927944, 0, 0, -0.000161892128958797, -1793613.37549153, -1793613.37517611, -896806.688524496, -0.00141323769581379, -896806.68770909, -0.000841744672076551, -21523360.5005437, -2690420.0648357, -2690420.06305145, -23316973.8767745, -23316973.8756606, -5380840.12667929, 0, 0, 0, 0, 0, 0, 1.12518966449774e-05, 0, 4.44417762909447e-06, 0, 0, 0, 2.99788088550657e-05, 0, 1.76838307588268e-05, -2.24915975668641e-06, 0.000106009532783615, 5.58069801621329e-05, 0, 0, 0, 6.5689755470168e-05, 0, 2.75286127004571e-05, 9.33589219909681e-06, 1.87627632186393e-05, 0.000333172642378743, 0.00041070495772115, 0.000118579375049206, 0.000187330528743986, 0, 0, 0, 8.735159494042e-05, 0, 5.46788361248709e-05, 0.000113358504674675, 6.20216933952112e-05, 0.000483989639020373, 0.000694633318195703, 0.000104170457199376, 0.000310804595129382, 8071260.18764569, 0.00111982270473973, 0.000233771185637326, 0.000346659574125855, 7.68510666817584e-05, 0.000563249383277215, 16142520.3756462, 16142520.3752442, 8071260.18876005, 0.00126947441417794, 8071260.1876762, 0.000557564396957439, 16142520.3758626}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, -1.53112571216862e-08, 0, 0, 0, 0, 0, 0, 0, 4.44135541956182e-07, 0, -2.92062061541648e-08, 2187.00000276521, 0, 0, 0, 0, 0, 0, -4.55162512314179e-06, 0, 1.67281702028961e-07, 0, 0, -4.59306034261348e-07, -24330.375018276, -34171.875035633, -24330.3750263288, 0, 0, 0, 0, 0, 0, 1.86694479030957e-05, 0, 2.19151129236675e-06, 0, 0, 0, 1.91597575756827e-05, 0, -8.29029275947495e-06, 44286.7500552493, 153773.437674784, 44286.7501119222, 375207.187698812, 265720.500133103, 375207.187767878, 0, 0, 0, 0, 0, 0, -3.95471522080454e-05, 0, -1.26314074766967e-05, 0, 0, 0, -7.73859170771771e-05, 0, 3.22972084809159e-05, -22143.3750928402, -221433.750429283, -22143.3752750391, 0, 0, -2.72016425458826e-05, -642157.87547055, -442867.500303628, -1638609.75076347, -1638609.75100981, -442867.500389678, -642157.875823684, -4052237.62618713, 0, 0, 0, 0, 0, 0, 4.63056547545478e-05, 0, 2.46371668005632e-05, 0, 0, 0, 0.000126506233846215, 0, -4.00079986976447e-05, 8.87772819217946e-05, 99645.1880632931, 0.000391810651490066, 0, 0, 0, 0.000143545113349001, 0, -1.44828937176282e-05, 298935.563068325, 199290.375371138, 2192194.12636883, 2192194.12679747, 199290.375637725, 298935.563842932, 2391484.50131515, 398580.750581098, 2391484.50219551, 6377292.00211925, 17238617.4409003, 6377292.00251059, 0, 0, 0, 0, 0, 0, -2.87018213315686e-05, 0, -2.09476937068908e-05, 0, 0, 0, -9.36733788991205e-05, 0, 1.6471081673755e-05, -4.49436659445149e-05, -0.000378375596655111, -0.000291688627849298, 0, 0, 0, -0.000185253086672054, 0, 4.00359568500892e-05, -0.000347443048853948, -0.000231844313563343, -896806.688666366, -896806.689015493, -0.000571402709369596, -0.00116164694600197, 0, 0, -8.20012386055518e-05, -896806.688508929, -0.000431227511814811, -1793613.37653807, -1793613.37754343, -0.000547650822475773, -896806.689657759, -2690420.06424068, -21523360.5039706, -2690420.06496268, -23316973.8790521, -5380840.12778529, -23316973.8797123, 0, 0, 0, 0, 0, 0, 7.38780130047995e-06, 0, 6.62537080835191e-06, 0, 0, 0, 2.61133896017061e-05, 0, -2.117779345713e-07, 9.28431412614302e-06, 0.000102692928682261, 8.67874313248048e-05, 0, 0, 0, 7.51395216513061e-05, 0, -1.43400874444503e-05, 8.5795776811434e-05, 5.73682072854188e-05, 0.000381915142849957, 0.00048569782324925, 0.000208376195655395, 0.000408549992051405, 0, 0, 0, 7.17984034007463e-05, 0, 7.88516289901805e-06, 0.000289600698220699, 0.000127525671121976, 0.000646218003085732, 0.00107065594715966, 0.000225721350297763, 0.000842529032541343, 0.000543557205812739, 8071260.18913331, 0.000984254466823291, 0.000572157543479873, 0.000176955537191782, 0.00117116445027635, 8071260.18913432, 0.00109867534346217, 16142520.3773364, 16142520.377736, 0.00122324017642618, 8071260.18966256, 16142520.3777262}, 
                                                    {0, 0, 0, 0, 0, 0, 0, 0, 6.05889415955192e-08, 0, 0, 0, 0, 0, 0, 0, -9.24714544812681e-07, 0, -5.58495280845526e-07, 2187.00000126323, 0, 0, 0, 0, 0, 0, 7.01645203000007e-06, 0, 3.27719962052674e-06, 0, 0, 3.81790881567654e-06, -24330.3750111959, -24330.3750156166, -34171.8750134655, 0, 0, 0, 0, 0, 0, -2.80802035543881e-05, 0, -1.27508578961339e-05, 0, 0, 0, -2.18861209380557e-05, 0, -5.83360548035259e-06, 44286.7500378818, 44286.7500534031, 153773.437541914, 265720.500114159, 375207.187589825, 375207.187677853, 0, 0, 0, 0, 0, 0, 6.09650128858158e-05, 0, 2.87669266047561e-05, 0, 0, 0, 8.31271223680992e-05, 0, 1.43782267755565e-05, -22143.3750669012, -22143.3750646348, -221433.750049635, 0, 0, -4.54758786267032e-06, -442867.500283808, -642157.875252407, -442867.500369727, -642157.875601509, -1638609.75019782, -1638609.75060402, -4052237.62587964, 0, 0, 0, 0, 0, 0, -7.28294100530411e-05, 0, -3.58186678026527e-05, 0, 0, 0, -0.000154674772582243, 0, -3.80961154906176e-05, 6.50790503761604e-05, -2.86543029628016e-06, 99645.1875056165, 0, 0, 0, -2.43802367682683e-05, 0, 6.44153242994035e-05, 199290.375338211, 298935.562860959, 199290.375525651, 298935.563338129, 2192194.12514154, 2192194.12592934, 398580.750681835, 2391484.5004223, 2391484.50167427, 6377292.0016486, 6377292.00236102, 17238617.4394457, 0, 0, 0, 0, 0, 0, 4.50718603623011e-05, 0, 2.27891123011817e-05, 0, 0, 0, 0.000134793970213765, 0, 4.49768062509977e-05, -3.25227376599078e-05, 5.85486197725666e-05, 3.156655087468e-05, 0, 0, 0, 7.02340038125895e-05, 0, -6.1611612328274e-05, -0.000200043736498977, -0.00025624593008188, -0.000325381466619382, -0.0004873747083014, -896806.687479369, -896806.6881778, 0, 0, -8.1551041400188e-05, -0.000499991168518751, -896806.687926577, -0.00068098070527249, -896806.689205846, -1793613.37521371, -1793613.37683206, -2690420.06388994, -2690420.06496488, -21523360.5017486, -5380840.12790296, -23316973.8773152, -23316973.8787361, 0, 0, 0, 0, 0, 0, -1.12812647241229e-05, 0, -5.76888334434178e-06, 0, 0, 0, -4.45717735537673e-05, 0, -1.82343541559392e-05, 6.34110165333034e-06, -3.00450019805825e-05, -1.71874289744888e-05, 0, 0, 0, -4.09399708026428e-05, 0, 1.4821361228366e-05, 4.75042622755452e-05, 7.13406434586832e-05, 6.52677585359482e-05, 8.29074985109767e-05, -4.43955247054578e-05, 0.000187076010060773, 0, 0, 0, 1.89696779070812e-05, 0, 5.39096722971015e-05, 0.000129624658253412, 0.000160179180884259, 0.0002387707696498, 0.000584448885186556, -1.46850788517192e-06, 0.000716867239555103, 0.000444656262899951, 0.000874804401808786, 8071260.18806781, 0.000242936906206092, 0.000125943618259295, 0.000962064849566436, 0.00115987572361428, 8071260.18846996, 0.00154416553536044, 8071260.18953566, 16142520.3759613, 16142520.3767971, 16142520.3771967}};

// -----------------------------------------------------------------------------
double MAd_TetInterpolant::coef10[286][286] = {{1, -29.289682538097, -29.2896825402488, -29.289682538097, 351.454365036639, 351.454365088136, 351.454365036639, 702.908730257844, 702.908730368554, 702.908730146769, -2317.43275968095, -2317.4327601827, -2317.43275968095, -6952.29827977248, -6952.29828012818, -6952.2982836545, -6952.29828295729, -6952.2982803251, -6952.29828345409, -13904.5965692384, 9416.1430749441, 9416.14307745655, 9416.1430749441, 37664.5722991388, 37664.5723081133, 37664.5723451367, 37664.5723373046, 37664.5723098027, 37664.572338703, 56496.85846749, 56496.858463027, 56496.8584702079, 112993.717011072, 112993.717057636, 112993.717023483, -24858.217583382, -24858.2175905156, -24858.217583382, -124291.087911031, -124291.087953312, -124291.088142316, -124291.088109132, -124291.087962088, -124291.088100995, -248582.175815874, -248582.175914216, -248582.17602309, -248582.175983961, -248582.1759053, -248582.175947965, -497164.352192026, -497164.35264817, -497164.352261237, -745746.528797368, -745746.528542232, -745746.528894924, 43478.0092391827, 43478.0092510112, 43478.0092391827, 260868.055431668, 260868.055527843, 260868.056064097, 260868.05600575, 260868.055559915, 260868.055941804, 652170.138396185, 652170.138881397, 652170.139509695, 652170.139177819, 652170.138815913, 652170.138925428, 1304340.27852894, 1304340.28026553, 1304340.27883303, 869560.184855538, 869560.185085637, 869560.185464773, 2608680.5590975, 2608680.55859618, 2608680.56050856, 2608680.56184004, 2608680.55805852, 2608680.55916303, 3913020.84120991, -50016.5343638324, -50016.5343750429, -50016.5343638324, -350115.740565791, -350115.740688321, -350115.741576556, -350115.741561024, -350115.740757082, -350115.741403832, -1050347.22125665, -1050347.22224321, -1050347.22393679, -1050347.22293874, -1050347.22211679, -1050347.22206806, -2100694.4452961, -2100694.44883302, -2100694.44633197, -1750578.70201598, -1750578.70360864, -1750578.70360079, -1750578.70475406, -1750578.70330917, -1750578.70453507, -5251736.11732673, -5251736.11770188, -5251736.12264021, -5251736.1270691, -5251736.11518385, -5251736.11586654, -7002314.8283035, -7002314.82277998, -7002314.83405171, -10503472.2453329, -10503472.253905, -10503472.2430103, 36375.6613521836, 36375.6613575501, 36375.6613521836, 291005.290857789, 291005.290946562, 291005.291800533, 291005.291880456, 291005.291026759, 291005.291696758, 1018518.51750495, 1018518.51853742, 1018518.52089467, 1018518.51951068, 1018518.51847047, 1018518.51804092, 2037037.03743479, 2037037.0415424, 2037037.03935568, 2037037.03439269, 2037037.0371485, 2037037.03805864, 2037037.038569, 2037037.0364175, 2037037.0377375, 6111111.11672064, 6111111.1188898, 6111111.12525189, 6111111.13158511, 6111111.11544211, 6111111.11276588, 2546296.29384744, 2546296.29576414, 2546296.29903556, 10185185.2032511, 10185185.199217, 10185185.2084418, 10185185.2251447, 10185185.1936738, 10185185.2109391, 15277777.8111023, 15277777.8338164, 15277777.7999755, 20370370.4347406, 20370370.4183324, 20370370.4411484, -15156.525562062, -15156.5255627909, -15156.525562062, -136408.730090191, -136408.730123901, -136408.730565656, -136408.730661905, -136408.73017103, -136408.730560525, -545634.920071296, -545634.920623021, -545634.922260462, -545634.921368264, -545634.920653175, -545634.920196039, -1091269.84123106, -1091269.84380114, -1091269.84292708, -1273148.14630767, -1273148.14833987, -1273148.14969172, -1273148.14919478, -1273148.14774135, -1273148.14799398, -3819444.44680061, -3819444.44907698, -3819444.45331593, -3819444.45763291, -3819444.44770214, -3819444.44311705, -1909722.2192346, -1909722.22206947, -1909722.22093273, -1909722.22502075, -1909722.22148407, -1909722.22476036, -7638888.90033083, -7638888.90116099, -7638888.90764255, -7638888.92429644, -7638888.89313828, -7638888.9030499, -11458333.3564189, -11458333.3789446, -11458333.3419558, -9548611.13225824, -9548611.121332, -9548611.15198149, -19097222.2804825, -19097222.271812, -19097222.2941035, -19097222.31187, -19097222.2581554, -19097222.2838986, -25462963.0592983, 2755.73192013782, 2755.73191995206, 2755.73192013782, 27557.3192105484, 27557.3192156409, 27557.3193108312, 27557.3193443814, 27557.3192266145, 27557.3193230287, 124007.936376602, 124007.936496308, 124007.936945335, 124007.93673091, 124007.93652278, 124007.936376515, 248015.872947391, 248015.873621203, 248015.87351437, 330687.830196384, 330687.830754694, 330687.831339943, 330687.830978139, 330687.830599212, 330687.830418793, 992063.49237738, 992063.4931449, 992063.494323759, 992063.495483048, 992063.493256999, 992063.491222762, 578703.70261235, 578703.703754239, 578703.703621497, 578703.704650362, 578703.703403907, 578703.704371941, 2314814.81752159, 2314814.81890735, 2314814.82058697, 2314814.82624887, 2314814.81592898, 2314814.81692339, 3472222.22828807, 3472222.23635023, 3472222.22267513, 694444.443348859, 694444.444235121, 694444.445919793, 3472222.22920636, 3472222.22707953, 3472222.23087774, 3472222.24153612, 3472222.22446908, 3472222.23622579, 6944444.4636018, 6944444.46311905, 6944444.47259959, 6944444.48224552, 6944444.45223479, 6944444.46183671, 8680555.58711502, 8680555.57546063, 8680555.59912038, 11574074.1190883, 11574074.1303538, 11574074.1162392}, 
                                                     {0, -0.999999999274498, 0, 0, 28.2896825202228, 0, 0, 0, 0, 0, -323.164682331022, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1994.26807639995, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.17953113942804e-08, 0, 0, -7421.87499584452, 0, 0, -1.16083569423608e-08, -1.22166241567356e-08, 0, 0, 0, 0, -3.11335313085692e-08, -3.12615924258353e-08, -2.42044890796933e-08, 0, -2.42219542201383e-08, 0, -6.26570511030877e-08, -2.5886691901178e-08, -2.52085712304506e-08, -7.22205726333414e-08, -7.30704101670199e-08, -3.81300964136244e-08, 17436.3425835515, 0, 0, 2.07035952826118e-08, 1.94113880525038e-08, 1.19245378016331e-08, 0, 1.19259944142414e-08, 0, 8.73330350259494e-08, 8.71351346631855e-08, 5.81602250804281e-08, 0, 5.94233426909341e-08, 0, 1.7247748473892e-07, 6.04943579674e-08, 5.98249698668951e-08, 1.07950199890183e-07, 1.12921128447852e-07, 0, 3.2910088521021e-07, 3.18390675602132e-07, 2.28682779379596e-07, 1.17216870876291e-07, 2.28589669859502e-07, 1.18555647077301e-07, 3.44747149938485e-07, -26041.6666542068, 0, 0, -1.37641507080843e-08, -1.33450441808236e-08, -1.47247547488405e-08, 0, -1.42503644440239e-08, 0, -1.33543835545424e-07, -1.35453035454702e-07, -8.54643360526097e-08, 0, -8.51965751280659e-08, 0, -2.64060872723348e-07, -8.60464126617444e-08, -8.44515284370573e-08, -2.41683324020414e-07, -2.4089166572594e-07, -2.07514119665575e-07, 0, -2.07234734261874e-07, 0, -7.27517999621341e-07, -7.22861386748264e-07, -4.22897983298753e-07, -2.10936718758603e-07, -4.31792045674229e-07, -2.10866915040242e-07, -8.23071559352684e-07, -7.96435870142886e-07, -2.86782608327485e-07, -1.19484002425452e-06, -8.35830633150181e-07, -8.14130999060581e-07, 23974.8677143214, 0, 0, 0, 0, 1.0675766048962e-08, 0, 1.07281543648696e-08, 0, 1.17061347282288e-07, 1.18761022349645e-07, 7.56311351324257e-08, 0, 7.28604447886028e-08, 0, 2.44227521761786e-07, 7.58523128752131e-08, 7.73773649598297e-08, 3.02468265545031e-07, 2.98789473163197e-07, 2.181331524298e-07, 0, 2.2081070483182e-07, 0, 8.83655957295559e-07, 8.78161245054798e-07, 4.43204612565751e-07, 2.23232120788452e-07, 4.54007931693923e-07, 2.26398640279513e-07, 3.44610612046381e-07, 3.39208895638876e-07, 0, 1.37937377076014e-06, 1.3866379049432e-06, 1.04645118881308e-06, 3.77858782485418e-07, 1.09338975562423e-06, 3.77625951841765e-07, 2.08507935894886e-06, 1.11127110358211e-06, 1.1189081305929e-06, 2.21080790652195e-06, 2.19534786083386e-06, 1.48200524563435e-06, -12400.7936457965, 0, 0, 0, 0, 0, 0, 0, 0, -5.32947694864561e-08, -5.28174837199913e-08, -3.7809570585523e-08, 0, -3.7716439749147e-08, 0, -1.07776997992914e-07, -3.63194558872237e-08, -3.59876750621879e-08, -1.7813840713643e-07, -1.85542432973307e-07, -1.27544467432017e-07, 0, -1.2897638157483e-07, 0, -5.49991682419204e-07, -5.39747134098434e-07, -2.51992275934754e-07, -1.26450174775528e-07, -2.51596475209226e-07, -1.23819177133555e-07, -2.78500010608695e-07, -2.77103026746772e-07, -2.42120279381197e-07, 0, -2.4344740268134e-07, 0, -1.13448913907632e-06, -1.13514101940382e-06, -7.51902462070575e-07, -2.53622090440331e-07, -7.69783810028457e-07, -2.4626467620692e-07, -1.70503972185543e-06, -7.95581399870571e-07, -7.66244738770183e-07, -1.2269692888367e-06, -1.26180088955152e-06, -3.19594619213603e-07, -2.4667911020515e-06, -2.39619657804724e-06, -1.94904623640468e-06, -1.25425708574767e-06, -1.9862991393893e-06, -1.30818079924211e-06, -2.55657050729496e-06, 2755.73192138473, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.79273342837405e-08, 0, 0, 4.4221565076441e-08, 4.66022669343147e-08, 2.97903639534525e-08, 0, 2.9540075274781e-08, 0, 1.36174605813721e-07, 1.38293387408339e-07, 6.27996143975906e-08, 3.06780307823828e-08, 6.11989037224703e-08, 3.08904901658025e-08, 8.62015383518155e-08, 8.51072456953261e-08, 6.93090669301455e-08, 0, 6.8796836671936e-08, 0, 3.51139192389383e-07, 3.48624553225818e-07, 2.13264854664885e-07, 7.06362044411435e-08, 2.0360238295325e-07, 7.06653082716002e-08, 5.27896190760657e-07, 2.21856282678345e-07, 2.27514078687818e-07, 1.00532275837395e-07, 9.8879155530085e-08, 0, 4.73860154670547e-07, 4.82148948321992e-07, 4.12280428463419e-07, 1.05770965319607e-07, 4.20010451307462e-07, 1.09961916905377e-07, 9.71469034993788e-07, 9.96894186755526e-07, 6.27873646408261e-07, 4.16378270529094e-07, 6.271751544773e-07, 4.29556507697271e-07, 1.02613762464898e-06, 1.04755804386514e-06, 5.58005126549688e-07, 1.43058241519611e-06, 1.05919957604783e-06, 1.05817525763996e-06}, 
                                                     {0, 0, -0.999999999154637, 0, 0, 28.2896825169775, 0, 0, 0, 0, 0, -323.164682295482, 0, 0, 0, 2.22925264606613e-08, 0, 0, 0, 0, 0, 1994.26807619611, 0, -6.8818298920581e-08, 0, -2.1209422065018e-07, 0, 0, 0, -1.26164032169191e-07, 0, 0, 0, 0, 0, 0, -7421.87499513454, 0, 2.86508805569241e-07, 0, 1.14872485114422e-06, 1.52165085330997e-08, 0, 0, 6.2897001439743e-07, 0, 8.25117462996547e-07, 0, 0, 1.83591399153273e-08, -5.96989507258148e-08, -4.23087436243225e-08, -2.98597305853037e-08, -5.80939360460142e-08, 3.2035082009187e-08, 6.93965341428126e-08, 0, 17436.3425820267, 0, -7.92801599924786e-07, 0, -3.49533140081965e-06, 0, 0, 1.9980934136428e-08, -1.93830577216035e-06, 0, -3.92775501722927e-06, 3.21429012696672e-08, 0, 1.9973730256785e-07, -2.05203589182118e-07, 2.55222547748356e-08, 6.59561680853443e-08, -3.43135440465333e-06, 0, 1.80653347112813e-08, 4.89444275905982e-07, -1.91153129410808e-07, 2.81555572919436e-07, 1.43514275335708e-07, -3.82714398925101e-07, -4.35382443936083e-07, -1.43175227488984e-07, 0, -26041.6666521133, 0, 1.25944021700519e-06, 0, 6.48930520510659e-06, 6.85140175178529e-08, 0, 0, 3.4051633007106e-06, 0, 8.45527123668623e-06, 1.03311925892125e-07, 0, -4.20906150948839e-08, 4.93935566387887e-08, 1.27948514300368e-07, 1.34186990109085e-07, 6.95792084151468e-06, 0, 1.0232736495915e-05, -3.27258351311726e-07, 0, 1.18566353618307e-07, -7.50715796190037e-07, -4.34764775780841e-07, 5.46154533640997e-07, -3.98681138766562e-07, 4.43121668320059e-07, -4.03302738219034e-07, 7.18787725219875e-07, 1.33260540747352e-07, 3.03286556405543e-07, 3.01620680364832e-06, -1.48467183000847e-06, 1.67135160331588e-06, 0, 23974.8677125774, 0, -1.22414286884975e-06, 0, -6.90606572354921e-06, 0, 0, 0, -4.27173716015156e-06, 0, -1.06913477471398e-05, -4.14537096050518e-08, 0, -6.96701688705132e-08, 8.03744801193361e-08, 1.08979203418796e-07, -3.69993167789954e-07, -8.08882235250429e-06, 0, -1.39101821391584e-05, -4.5295545142636e-07, 0, -6.90543114919297e-07, -6.81604611011433e-07, -7.79325386191422e-07, -1.88248622752516e-06, 1.35269181363503e-06, -4.39173279655363e-07, 3.36389062310237e-07, -1.50033713411092e-05, 0, 6.91418289186782e-07, 9.62787530557679e-07, 4.18965304295984e-07, -1.92873171605181e-06, 6.11070912769069e-07, -7.6042418753728e-07, -3.10167343907309e-08, 2.40535803450581e-06, -3.03189590945166e-06, -3.45150913783635e-06, -9.15186778060798e-06, -3.68193074907602e-06, -1.97217239908347e-06, 0, -12400.7936449819, 0, 6.6420453137539e-07, 0, 4.00373562620081e-06, 3.63389692720853e-08, 0, 0, 2.41928360391165e-06, 0, 7.68744422555111e-06, 1.41281885928097e-07, 0, -3.63823088305607e-08, -5.1650616852381e-08, -1.01700055442921e-07, -7.92001450391505e-08, 5.39818271618613e-06, 0, 1.15545880704108e-05, 1.61004300204419e-07, 0, 2.38499944566337e-07, 6.02038327455934e-07, 1.5435565070521e-07, 3.40162920359399e-07, -6.15582465340278e-07, -1.78233062628038e-07, 2.55824361347762e-08, 1.13104926999718e-05, 0, 1.58812058556309e-05, 7.0944906585e-07, 0, 9.59280190723314e-08, 2.49931253047058e-06, 2.81597363926092e-07, -1.07606758499137e-08, -5.64638347479946e-07, -3.43693572623146e-07, -1.36572188864047e-06, 5.59001929281174e-07, 5.91849302715974e-06, 2.42620172096065e-08, 7.73473778993565e-07, -1.45702874174029e-06, 1.68721923341373e-06, -5.59538595634799e-07, 6.62383413626732e-06, 6.6043863193101e-06, -3.42617858832594e-06, -2.21374056974124e-06, -7.40412991082104e-06, -1.22509468949908e-05, 0, 2755.73192122234, 0, -1.51613283483343e-07, 0, -9.69850270932911e-07, 0, 0, 0, -6.44268823679742e-07, 0, -2.18731213998994e-06, 3.1151877717831e-08, 0, -1.68107079491367e-08, 0, 3.75850756836382e-08, -4.46295612721767e-08, -1.45122529593099e-06, 0, -3.52109165245137e-06, -1.07832984717804e-07, 0, -1.17086986845041e-07, 9.31165337790722e-08, -1.02271813148389e-07, 0, -1.98749365297464e-07, -1.68144568272926e-07, 9.08114202582531e-08, -3.53925359526153e-06, 0, -6.11551942783334e-06, 0, 0, -5.48959608134866e-08, 9.37827576954696e-08, -4.04293526549822e-07, -6.53949753002054e-07, 7.09635964527698e-07, 8.17468292322244e-08, -5.7529527713621e-07, 5.3723538045549e-07, -2.12325974973572e-07, -6.34439133525972e-07, -6.24320812775214e-06, 0, -4.1942889549074e-07, -1.95972357164432e-07, 2.52837374382219e-07, 1.6740257200764e-06, -5.0646923694391e-07, 8.54149160781539e-07, 9.22176073825792e-08, 2.86122210280836e-06, -7.68219511954779e-07, 2.24384860202408e-06, 6.15397562425634e-06, -1.9455084394265e-07, 3.02964057556531e-06, -3.19267343988507e-07, -4.33943118066225e-07, 4.44764887351932e-06, -8.34111514426665e-07, -4.60639335405333e-06, -5.92058701652869e-06}, 
                                                     {0, 0, 0, -0.999999999274498, 0, 0, 28.2896825202228, 0, 0, 0, 0, 0, -323.164682331022, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1994.26807639995, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.19131797760019e-08, 0, 0, -7421.87499584452, 0, 0, 0, 0, -1.197215482307e-08, -1.1812083755558e-08, 0, -2.4108452123528e-08, 0, -2.43354563167486e-08, -3.09880121562855e-08, -3.07813792232992e-08, -2.52871572570257e-08, -2.55054359854512e-08, -6.18421438502992e-08, -3.84356866334201e-08, -7.25989224292789e-08, -7.36000913548196e-08, 0, 0, 17436.3425835515, 0, 1.19230847417384e-08, 0, 1.20264012082316e-08, 1.8608119489727e-08, 1.86255846301719e-08, 0, 5.80205323785776e-08, 0, 5.81369477004046e-08, 8.69837890604686e-08, 8.66694733758777e-08, 5.906827027502e-08, 6.03139085342264e-08, 1.72035129253345e-07, 0, 1.04644016118982e-07, 1.10651029672226e-07, 1.19999185699271e-07, 2.28496560339408e-07, 1.18765171919222e-07, 2.3049886976878e-07, 3.2099842428579e-07, 3.25841256199055e-07, 3.47727336702519e-07, 0, 0, -26041.6666542068, 0, -1.43318530376746e-08, 0, -1.45297605058659e-08, -1.55918655764253e-08, -1.69771965374821e-08, 0, -8.42652525534504e-08, 0, -8.42419751734269e-08, -1.35639311338309e-07, -1.33683556668984e-07, -8.52082280289324e-08, -8.54061283916963e-08, -2.73886371360277e-07, 0, -1.99411601897737e-07, 0, -2.03183503799664e-07, -2.44244461100607e-07, -2.5374390588695e-07, -2.15290697269666e-07, -4.08369260185282e-07, -2.16361684124422e-07, -4.18939862356638e-07, -7.44747467251727e-07, -7.32081389287487e-07, -2.8599095003301e-07, -7.97180746303638e-07, -8.17297404864803e-07, -8.37693278299412e-07, -8.27541953185573e-07, -1.17593413051509e-06, 0, 0, 23974.8677143214, 0, 1.09115063651188e-08, 0, 1.12636655558163e-08, 0, 0, 0, 7.6475146215671e-08, 0, 7.44320516332664e-08, 1.20670222258923e-07, 1.18923992431519e-07, 7.58814167056698e-08, 7.73249695384948e-08, 2.35752509070153e-07, 0, 2.21159950797301e-07, 0, 2.1964655161355e-07, 2.89755689664162e-07, 2.93434368359158e-07, 2.21555751522828e-07, 4.48652826889884e-07, 2.28750195674365e-07, 4.39572431787383e-07, 8.97066911420552e-07, 8.68475353854592e-07, 0, 3.39069174515316e-07, 3.42235694006376e-07, 3.89267484024458e-07, 1.05688195617404e-06, 3.60722424375126e-07, 1.09562506622751e-06, 1.38589302878245e-06, 1.34212086777552e-06, 1.13269152279827e-06, 1.14610270429694e-06, 2.05639435080229e-06, 1.4868483049213e-06, 2.08507935894886e-06, 2.19534786083386e-06, 0, 0, -12400.7936457965, 0, 0, 0, 0, 0, 0, 0, -3.49253852505171e-08, 0, -3.5754844418534e-08, -5.22819618709036e-08, -5.1583469939942e-08, -3.67880303997481e-08, -3.76960684889127e-08, -1.02235617305269e-07, 0, -1.25600337241849e-07, 0, -1.26345383932858e-07, -1.81817142674845e-07, -1.77742606410902e-07, -1.28266236743002e-07, -2.58581394518842e-07, -1.26205691231007e-07, -2.44727971221437e-07, -5.37605046702083e-07, -5.67221150049591e-07, 0, -2.41654618093889e-07, 0, -2.36509038131771e-07, -2.73587261290231e-07, -2.97592123388313e-07, -2.48243736677978e-07, -7.40260929887882e-07, -2.55251904945908e-07, -7.69224925534218e-07, -1.14845897769555e-06, -1.07711957753054e-06, -7.94929519543075e-07, -7.87385715739219e-07, -1.66964946402004e-06, -3.10630639432929e-07, -1.26180088955152e-06, -1.22827327686537e-06, -1.25025235320209e-06, -1.89353932000813e-06, -1.35372238219134e-06, -1.87137402463122e-06, -2.42488158619381e-06, -2.41221550822956e-06, -2.54204178418149e-06, 0, 0, 2755.73192138473, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.71997385223221e-08, 0, 2.97903639534525e-08, 0, 3.05412442003217e-08, 4.00888211515849e-08, 4.23007122662966e-08, 3.02705771559886e-08, 6.07623462656193e-08, 3.03375173871245e-08, 6.29858760703428e-08, 1.35126867917279e-07, 1.33217668008001e-07, 0, 6.55721379416718e-08, 0, 6.79120830682223e-08, 8.65391314164299e-08, 8.82271535829204e-08, 6.99144351301584e-08, 2.0979570081181e-07, 6.81565524018879e-08, 2.16105377148779e-07, 3.44619934367074e-07, 3.42990119861497e-07, 2.05721164547867e-07, 2.13264854664885e-07, 5.15416445523442e-07, 0, 1.02779068811287e-07, 9.85881172255176e-08, 1.06015420442418e-07, 4.13025531997846e-07, 1.05957212781504e-07, 4.12792701354192e-07, 4.93231709697284e-07, 5.04454078509298e-07, 3.95749452763994e-07, 6.35231117485091e-07, 4.29183955930057e-07, 6.28758357379411e-07, 9.95124764813227e-07, 9.92796458376688e-07, 5.75188096263446e-07, 1.03414708974015e-06, 1.10492760541092e-06, 1.09822212834843e-06, 1.06748848338611e-06, 1.37097777042072e-06}, 
                                                     {0, 99.9999999854561, 0, 0, -1928.96825357668, -1.44251520543246e-07, 0, -1928.96825415258, -2.61116546505881e-07, -1928.96825391678, 15855.7539640424, 1.12207504439158e-06, 0, 31711.507928292, 31711.5079352511, 15855.7539784334, 4.82097575123407e-08, 15855.753967977, 7.32905939409572e-06, 31711.5079564643, -73185.7363072692, -3.51217118595101e-06, 0, -219557.208910788, -219557.208984975, -73185.7364627192, 7.14316022634351e-06, -73185.7363292571, -8.18756956928935e-05, -219557.208913776, -219557.208987615, -1.68011092239897e-05, -439114.418271015, -219557.2092498, -219557.209194056, 209756.944360279, 1.20466342645306e-06, 0, 839027.777450519, 839027.777734761, 209756.945180681, -2.47185505308907e-05, 209756.9444394, 0.000378643012745958, 1258541.66563042, 1258541.66665976, 839027.777703382, -0.000115642322102354, 839027.777697755, 0.000476490910650731, 2517083.33472997, 839027.779113006, 839027.778304673, 2517083.33630789, 2517083.3362401, 1258541.66894714, -388252.314631502, 2.11661700720819e-05, 0, -1941261.57338641, -1941261.57393856, -388252.316987465, -4.24872379911794e-05, -388252.314830236, -0.000938722918863297, -3882523.14447478, -3882523.14831798, -1941261.57546564, 0.00108844143498546, -1941261.57373432, -0.00200050649891309, -7765046.29960159, -1941261.57739137, -1941261.57431407, -3882523.14383232, -3882523.14786547, -0.00183090603922202, -11647569.4565018, -11647569.4584425, -7765046.30791294, -3882523.16025782, -7765046.30491131, -3882523.15284191, -11647569.4709757, 465277.777524946, -5.94466896082422e-05, 0, 2791666.66580556, 2791666.66638984, 465277.781409295, 0.00031602518357163, 465277.777847652, 0.0014124109049563, 6979166.66027553, 6979166.6671796, 2791666.67235706, -0.00299288729754548, 2791666.66611808, 0.00345078251897503, 13958333.3372316, 2791666.67118195, 2791666.66602994, 9305555.54148301, 9305555.55587992, 6979166.65926841, 0.0025788451594776, 6979166.66500234, 0.0080638147134099, 27916666.6894439, 27916666.6985771, 13958333.3541231, 6979166.6947494, 13958333.3436153, 6979166.66303146, 27916666.7033424, 27916666.7042322, 9305555.58807866, 41875000.0966007, 27916666.7595425, 27916666.7315261, -348875.661161535, 7.336858750599e-05, 0, -2442129.62898684, -2442129.62930399, -348875.664839676, -0.000590108955178897, -348875.661472827, -0.00129022946632377, -7326388.88277139, -7326388.88939183, -2442129.63877479, 0.00391728536010227, -2442129.62932717, -0.00284471605285481, -14652777.7797257, -2442129.63314186, -2442129.62909971, -12210648.129105, -12210648.1500593, -7326388.88492257, -0.000371408746892799, -7326388.88592297, -0.0119076106033599, -36631944.4655158, -36631944.4819099, -14652777.7960615, -7326388.92028336, -14652777.7834388, -7326388.86794153, -12210648.1244154, -12210648.1465839, -0.0132119383606266, -48842592.6452491, -48842592.6631229, -36631944.4925925, -12210648.2204807, -36631944.4843633, -12210648.1745525, -73263889.0443021, -36631944.5866351, -36631944.4954159, -73263889.1233474, -73263889.0922626, -48842592.7929273, 148809.523708031, -4.42879276759558e-05, 0, 1190476.19021383, 1190476.19027754, 148809.525568423, 0.000468671385959373, 148809.523866692, 0.000674862276098367, 4166666.66352847, 4166666.66680334, 1190476.19722005, -0.00245345658633764, 1190476.19052746, 0.00102107354713921, 8333333.33329892, 1190476.19202395, 1190476.19071233, 8333333.32126651, 8333333.33532364, 4166666.66849013, -0.00169748844454334, 4166666.66458418, 0.0074569312158322, 25000000.0085148, 25000000.0216574, 8333333.34067196, 4166666.68346935, 8333333.3356037, 4166666.6448445, 10416666.643397, 10416666.6673075, 8333333.31632108, 0.00945748304033837, 8333333.33020342, 0.0148476630672068, 41666666.7005063, 41666666.727175, 25000000.0273126, 8333333.39728834, 25000000.0180553, 8333333.32806372, 62500000.1150507, 25000000.0981946, 24999999.9984759, 41666666.7138062, 41666666.7217811, 10416666.7300889, 83333333.5745713, 83333333.572187, 62500000.2251951, 41666666.901217, 62500000.1460362, 41666666.8117459, 83333333.6944057, -27557.3192033819, 1.06598342431199e-05, 0, -248015.872971551, -248015.872967383, -27557.3195939089, -0.000140349266283892, -27557.3192364772, -0.000154336465009321, -992063.491382358, -992063.492039703, -248015.874906524, 0.00058381980180624, -248015.873081005, -0.000103488631987453, -1984126.9838736, -248015.873343201, -248015.873265156, -2314814.81183601, -2314814.81543978, -992063.493825932, 0.000966430171414379, -992063.491586847, -0.00167678360002755, -6944444.44533934, -6944444.44928299, -1984126.98510853, -992063.495483996, -1984126.98507933, -992063.484771398, -3472222.21458879, -3472222.22306307, -2314814.81082662, -0.00243080667575574, -2314814.8133882, -0.00513645978932243, -13888888.8964288, -13888888.9082018, -6944444.44960822, -2314814.83464876, -6944444.44774519, -2314814.80497136, -20833333.3648479, -6944444.46958494, -6944444.43376053, -3472222.21274902, -3472222.22161929, -0.00606989407770212, -17361111.1269877, -17361111.1365906, -13888888.9026808, -3472222.25636536, -13888888.9026526, -3472222.23670617, -34722222.3086244, -34722222.3173646, -20833333.4069857, -13888888.9795638, -20833333.3611287, -13888888.9137049, -34722222.335157, -34722222.318902, -17361111.2139387, -46296296.4927309, -34722222.4125328, -34722222.3713682}, 
                                                     {0, -224.999999939985, 0, 0, 5465.17856981397, -3.79716838314053e-07, 0, 3215.17857100573, 2.04272262125389e-07, 3215.17857136039, -51751.3392683651, 5.93948440367418e-06, 0, -71350.8928243438, -71350.8928549432, -19599.5535779814, 4.09523563608125e-06, -19599.5535721317, -1.51132276039788e-05, -39199.1071626978, 262665.674503142, -4.64488192250507e-05, 0, 592001.487821898, 592001.488076704, 66670.13906746, -9.14492768428423e-05, 66670.1388922675, 0.000149483034026667, 396005.951935064, 396005.952385622, 0.000162354645532851, 792011.905177529, 200010.416856049, 200010.416857584, -805303.819098177, 0.000202714030711796, 0, -2554513.88793168, -2554513.88879666, -138602.43175851, 0.000566622901223593, -138602.430572141, -0.000598358882965047, -2831718.74614915, -2831718.75013814, -1221111.10935092, -0.000377722694811522, -1221111.11103381, -0.00184424123236073, -5663437.50226733, -554409.722627125, -554409.722156622, -3663333.33707737, -3663333.33802563, -831614.585674408, 1566579.86035737, -0.000526034353231798, 0, 6446874.99819926, 6446874.99967118, 180555.559318391, -0.00166716668821989, 180555.555651643, 0.00123889857069346, 10121701.3765814, 10121701.3898524, 2288802.0817533, -0.000941439204552144, 2288802.08277151, 0.00680001655101469, 20243402.7832248, 902777.776395923, 902777.77456562, 7349652.75993015, 7349652.77785859, 0.00747118756672876, 22048958.3509021, 22048958.3584229, 9155208.34348398, 1805555.56650216, 9155208.34649367, 1805555.55767891, 13732812.5408316, -1949652.77673869, 0.000825676561755736, 0, -9892361.10914971, -9892361.11038841, -144097.228547101, 0.00275893184521763, -144097.222474596, -0.00149987849638861, -20217013.8688557, -20217013.8911304, -2670138.89395533, 0.00487853181568913, -2670138.8878501, -0.0121449046483405, -40434027.7838485, -864583.326633772, -864583.324429815, -20937499.9499838, -20937500.0032305, -11189236.074178, -0.0132073970703397, -11189236.1084854, -0.0251668895131792, -62812500.0353325, -62812500.0583062, -13350694.4496979, -2161458.35295422, -13350694.4554458, -2161458.31424703, -44756944.4849608, -44756944.5136024, -2881944.47222915, -67135416.8348622, -26701389.01578, -26701388.987616, 1505456.34832675, -0.000770748542408877, 0, 9097222.22101673, 9097222.22134133, 64484.1328612707, -0.00256577988187948, 64484.127300239, 0.00099293135216095, 22968749.9819747, 22968750.0021363, 1892361.12436865, -0.00819129735193298, 1892361.11055226, 0.0115050963093693, 45937500.0022696, 451388.878124974, 451388.878248511, 31076388.8257342, 31076388.8973603, 9999999.9653131, 0.0104762568211162, 9999999.99395733, 0.0361826833173698, 93229166.6999885, 93229166.7339181, 11354166.6520361, 1354166.67798248, 11354166.6631147, 1354166.61579091, 23871527.6881933, 23871527.7777945, 0.0383101772471791, 95486111.1760043, 95486111.2451385, 50000000.0250602, 2256944.51048289, 50000000.0734451, 2256944.44497711, 143229166.951897, 28385416.8268513, 28385416.7305411, 100000000.375799, 100000000.363346, 37847222.5059533, -657242.063075261, 0.000392993923881219, 0, -4613095.23773381, -4613095.23754955, -12400.7964830495, 0.00127508620450771, -12400.7938341554, -0.000352836770811562, -13888888.8802148, -13888888.8896102, -744047.630405103, 0.00616205536487935, -744047.619244285, -0.00539366491801739, -27777777.7768747, -99206.3414412891, -99206.3430513824, -23263888.8508927, -23263888.8964893, -4861111.099052, -0.00305839473751866, -4861111.10643236, -0.024564359347395, -69791666.6794643, -69791666.70371, -5208333.31205239, -347222.219460902, -5208333.32461385, -347222.176342223, -23437499.9196543, -23437500.0059364, -14236111.0408136, -0.0280463617568415, -14236111.1058548, -0.0386523979711143, -93750000.0435456, -93750000.1139102, -29166666.6529208, -694444.500945912, -29166666.6970969, -694444.401312305, -140625000.21748, -15625000.0779303, -15624999.9739296, -71180555.5878633, -71180555.6649839, -868055.601199011, -142361111.530788, -142361111.561108, -72916666.9783279, -26041666.9712983, -72916666.9265128, -26041666.8501076, -97222222.8314476, 124007.936423354, -8.43290635181614e-05, 0, 992063.492032105, 992063.491929334, 0.000552763438887444, -0.0002578231661438, 4.14473543274171e-05, 5.10299537620604e-05, 3472222.22046058, 3472222.22221851, 124007.939885412, -0.00175859081458575, 124007.936680972, 0.000973083670213964, 6944444.4437682, -0.00211852813180207, -0.00139395358353231, 6944444.43551962, 6944444.44670661, 992063.492024128, -0.000229298710128806, 992063.490939743, 0.00653745334271954, 20833333.334219, 20833333.3409697, 992063.484041369, -0.0035305511208087, 992063.489168254, -0.0144325649528211, 8680555.530551, 8680555.55932613, 3472222.2028243, 0.0081439234398274, 3472222.21921251, 0.013484505723373, 34722222.2315191, 34722222.2574983, 6944444.43092865, 0.0161443433303183, 6944444.44802777, -0.0258331516222239, 52083333.3932582, 3472222.23047209, 3472222.19544851, 6944444.41023426, 6944444.44450214, 0.0147062305086293, 34722222.2343078, 34722222.2726877, 20833333.3299546, 0.0304894263612687, 20833333.3620241, -0.00112227632846194, 69444444.6026866, 69444444.6269872, 20833333.4111895, 6944444.54806637, 20833333.3807234, 6944444.45913096, 52083333.5085893, 52083333.516867, 8680555.68229198, 69444444.8026444, 34722222.5221024, 34722222.4724214}, 
                                                     {0, 399.999999853126, -8.33595308893579e-08, 0, -10382.5396785909, 2.60197896402603e-06, 0, -4382.53968046231, -4.07917419743991e-07, -4382.53968255235, 105973.280380878, -3.24125094415886e-05, 0, 106208.465539736, 106208.465606544, 20235.1851636645, -2.20013125796874e-06, 20235.1851892127, 2.30129115102066e-05, 40470.3703809936, -573728.835734489, 0.000215218081157772, 0, -979404.761388312, -979404.7618839, -51074.0740911628, 0.000109746173575012, -51074.0741007283, -0.000210808709804716, -456749.998943343, -456750.000046743, -0.000290760046611938, -913500.000344947, -153222.222255271, -153222.222344006, 1854027.77693235, -0.000838281796186258, 0, 4626481.47979241, 4626481.4813543, 76157.4081157135, -0.000755791386986556, 76157.4075117576, 0.000857925037643905, 3767037.02915811, 3767037.03750876, 1070740.73565416, 0.000895184729797401, 1070740.74085004, 0.00274907901430949, 7534074.07610256, 304629.62880642, 304629.629264516, 3212222.22487286, 3212222.22656307, 456944.44614525, -3763194.44260494, 0.00200533312328379, 0, -12566203.7007912, -12566203.7031829, -67129.6322979963, 0.00235323210446296, -67129.6299610209, -0.00186277027368157, -15186574.0501259, -15186574.0764386, -1478009.24928495, -0.00010525013632137, -1478009.25895571, -0.0102032886903811, -30373148.1529764, -335648.142119134, -335648.1440109, -7794444.40614841, -7794444.44594065, -0.010177813571114, -23383333.3481053, -23383333.3581944, -5912037.03946523, -671296.301502728, -5912037.04902916, -671296.298119655, -8868055.5931431, 4847222.21968726, -0.00296847953387905, 0, 20460648.1453934, 20460648.1469323, 32407.4117692153, -0.00417936976940769, 32407.4080801325, 0.00244329968490339, 33402777.7401841, 33402777.7827461, 1201388.88310077, -0.00315345853677783, 1201388.88801745, 0.0195138972326708, 66805555.560218, 194444.428582471, 194444.433623686, 25949073.9734527, 25949074.0832707, 9328703.61922676, 0.0174703555820704, 9328703.70229185, 0.032958845630289, 77847222.2539133, 77847222.2810722, 6006944.42591821, 486111.112133873, 6006944.45142881, 486111.096947296, 37314814.8367912, 37314814.8844972, 648148.164714867, 55972222.3890475, 12013888.9904736, 12013888.9859311, -3849206.34706116, 0.00265188044505293, 0, -19745370.3691046, -19745370.3688551, -6613.76003568885, 0.00419921066538249, -6613.75738422475, -0.00173190484597524, -40995370.3377984, -40995370.3749584, -532407.413389263, 0.00659502808358204, -532407.407112251, -0.0205159513048934, -81990740.7409793, -46296.275160012, -46296.2830957523, -43518518.3956128, -43518518.5380479, -6412036.94766047, -0.01433477569015, -6412037.02997688, -0.0494054611232656, -130555555.586031, -130555555.622592, -3194444.39491247, -138888.871809503, -3194444.43287572, -138888.850601931, -24305555.3702825, -24305555.5623898, -0.0436081079639166, -97222222.2667426, -97222222.3606593, -32060185.1632429, -231481.51261488, -32060185.2571126, -231481.477075523, -145833333.625641, -7986111.20223282, -7986111.18459134, -64120370.7173435, -64120370.7443976, -10648148.4021399, 1719576.71856046, -0.00130743787995222, 0, 10416666.6665514, 10416666.6657228, 0.00111102115442963, -0.0022986497297806, 0.000440339771832533, 0.000669945829032387, 26388888.873871, 26388888.8904819, 99206.3581517284, -0.00539008698034616, 99206.3498085561, 0.0109408810930526, 52777777.7753951, -0.0140958345684807, -0.00819246563773496, 35879629.5577249, 35879629.6459113, 2314814.77286889, 0.00605922241190355, 2314814.80840757, 0.0363085006943488, 107638888.899747, 107638888.923947, 694444.398178214, -0.0228598229066818, 694444.427722311, -0.0335827810260099, 27777777.6178519, 27777777.7946212, 11805555.4051582, 0.0292527100013663, 11805555.5516893, 0.0449021009241704, 111111111.144176, 111111111.229681, 13888888.8265248, 0.0189619709453864, 13888888.9137432, -0.0353076606533836, 166666666.895147, 2083333.33657951, 2083333.32021402, 59027777.7712177, 59027777.891642, 0.0160244751338017, 118055555.970544, 118055556.029116, 34722222.4610143, 3472222.45965241, 34722222.4996575, 3472222.39902299, 46296296.8956202, -330687.830481657, 0.000273221360281458, 0, -2314814.81488708, -2314814.81458301, -6.35043624006887e-05, 0.000520624176311775, -0.000101974651461342, -0.000103370551028729, -6944444.44153995, -6944444.44448261, -0.00300222582591265, 0.00162399054236297, -0.000340836630631772, -0.00235274499247755, -13888888.8877956, 0.00374139333236253, 0.00202308229521932, -11574074.0576794, -11574074.0787553, -330687.824507097, -0.000932887102123322, -330687.829052797, -0.0106822567307215, -34722222.2224031, -34722222.22873, 0.0149496011812963, 0.00906411262005183, 0.00578707965704122, 0.0102824339059108, -11574074.0255308, -11574074.0828833, -2314814.77117202, -0.00855119174988456, -2314814.81098917, -0.0168563419781932, -46296296.3031539, -46296296.332607, -2314814.78389448, -0.00338704018737688, -2314814.81494538, 0.0187245782535497, -69444444.5071238, 0.019822537253428, 0.0212389507739323, -6944444.37497064, -6944444.44724671, -0.0148544846249246, -34722222.2228016, -34722222.2757037, -13888888.8637707, -0.0117182598627334, -13888888.9178377, 0.00914393275136651, -69444444.6077433, -69444444.6375355, -6944444.47688193, -0.0625048697192321, -6944444.49877018, -0.0206351599316264, -34722222.3763687, -34722222.4167521, -0.105019167357511, -46296296.6660206, -11574074.3496604, -11574074.3325115}, 
                                                     {0, -524.999999763694, 1.9147571659915e-07, 0, 14064.5833269837, -5.65664759362555e-06, 0, 4439.58332960949, 4.68723037304227e-07, 4439.58333347225, -149352.083265185, 6.72170245425333e-05, 0, -112311.805465864, -112311.805554963, -15459.7221645873, -5.45254859904429e-07, -15459.7222297978, -1.93034571570904e-05, -30919.4444463625, 843271.99034819, -0.00042766348293233, 0, 1095971.06417992, 1095971.06479979, 28385.4163915599, -6.61221165229823e-05, 28385.4167196131, 0.000177937954521925, 368584.489271194, 368584.490832988, 0.000218208313586152, 737168.981651655, 85156.2499223249, 85156.2500637929, -2835295.13753116, 0.00160869591040499, 0, -5509045.13689946, -5509045.13876458, -28993.0549662211, 0.000483826330534174, -28993.0557604706, -0.00078700684537995, -3281128.46203389, -3281128.47300438, -636371.520146629, -0.000583226248827789, -636371.528140095, -0.00201849516828106, -6562256.94553233, -115972.220944028, -115972.222097802, -1909114.58420754, -1909114.58593034, -173958.334104961, 5962326.38593554, -0.00374271579525712, 0, 15873148.1449286, 15873148.1475643, 15624.9990691269, -0.0014729800512039, 15625.0005829463, 0.00188550000806916, 14437731.451297, 14437731.4849566, 609664.334255082, -0.000186323636888895, 609664.352126702, 0.00801149403319722, 28875462.9654765, 78124.9934036187, 78124.9980014748, 5120949.02333538, 5120949.07716128, 0.00609349879939518, 15362847.2292503, 15362847.2377181, 2438657.40335655, 156249.99949397, 2438657.41529615, 156250.002287018, 3657986.13278358, -7920138.88482003, 0.00542158124616412, 0, -27243055.5528727, -27243055.5541272, -3472.22024000885, 0.00265594845214253, -3472.22331289136, -0.00283381564388664, -34505208.2868902, -34505208.3402538, -307291.647964487, 0.0018398689114624, -307291.66651953, -0.0165823128146366, -69010416.6684141, -20833.3183003314, -20833.3279283585, -19262152.6483292, -19262152.7922648, -4383680.44068747, -0.00855301951686459, -4383680.55683461, -0.0213992532331187, -57786458.3495891, -57786458.3707726, -1536458.30486325, -52083.3213914416, -1536458.33864893, -52083.3329868874, -17534722.2219265, -17534722.2674365, -69444.4494464915, -26302083.4354968, -3072916.71144385, -3072916.73141252, 6458333.32989079, -0.00476107989163002, 0, 27523148.1473416, 27523148.1463434, -0.00339355828857542, -0.0026781447632151, 0.00120669708940376, 0.00229392448794076, 45572916.6273343, 45572916.6729052, 63657.4002787619, -0.00340588930009019, 63657.4077398472, 0.0190189928938769, 91145833.3318157, -0.0186044528835076, -0.00668269998964761, 36226851.6962665, 36226851.8799176, 1944444.31795986, 0.0039571457372562, 1944444.43915264, 0.0352037338442679, 108680555.57115, 108680555.598066, 381944.389400338, -0.0301623215426407, 381944.437546159, -0.00748143501344176, 13599536.7953925, 13599537.051802, 0.0203979660776795, 54398148.1592357, 54398148.2377688, 9722222.17437891, -0.0034200127368826, 9722222.26958664, 0.00411214733537964, 81597222.404964, 954861.115671514, 954861.181142925, 19444444.6340353, 19444444.6902859, 1273148.29334516, -2951388.8872583, 0.00231542776815585, 0, -15104166.6669464, -15104166.6655418, 0.00291860881920558, 0.00151883005101522, -0.000687823169518654, -0.00105710767923406, -31249999.9823377, -31250000.0021591, -0.00126909202934337, 0.00251384837985726, -0.00102682899525753, -0.0109301521378049, -62499999.9973085, 0.0118569820311099, 0.00445587424931927, -32986111.0215276, -32986111.1335703, -347222.158113747, -0.000984584580167421, -347222.216082718, -0.0284519239167385, -98958333.3376484, -98958333.3545375, 0.0459435321064934, 0.027213704807444, 0.0114965222975565, 0.0067559505654108, -18229166.4613023, -18229166.6932573, -4687499.80146367, -0.00920713201000537, -4687500.00035284, -0.0249413176359741, -72916666.6776478, -72916666.7445962, -2083333.26015616, 0.0159954111628308, -2083333.34871423, 0.00524439443335143, -109375000.14767, 0.0539163034468772, -0.0177746485812496, -23437499.9672888, -23437500.0747021, 0.00368079323566842, -46875000.2440628, -46875000.3122739, -5208333.4196767, -0.104112595827759, -5208333.53322693, -0.128818627536474, -6944444.80758768, 578703.703372954, -0.000478551075303047, 0, 3472222.22240933, 3472222.22194714, -0.000932005429202818, -0.000358017736044298, 0.000161557063745549, 0.000201170229540151, 8680555.55224277, 8680555.55562302, 0.000973352393823983, -0.000685929342693402, 0.000470260817940735, 0.00256517966933106, 17361111.1100888, -0.0030698541328765, -0.00114363543984855, 11574074.0540276, 11574074.0803872, -0.0114755909260905, 0.00034312626131272, -0.00168505689095637, 0.00923807630642106, 34722222.2215469, 34722222.2257684, -0.0139928033852527, -0.0088033133705723, -0.00411375583175212, -0.00190466962888656, 8680555.49394372, 8680555.56833919, 578703.64526298, 0.00236639010293122, 578703.700399444, 0.0110476246552235, 34722222.2237657, 34722222.2461415, -0.0312144056467296, -0.0074091870333912, -0.00072242919592812, -0.0023610666835434, 52083333.3733507, -0.0329650947301271, -0.00354022516864379, 3472222.13219794, 3472222.22820792, 0.00571098458190818, 17361111.1025703, 17361111.1462775, 3472222.19122293, -0.00530087518466123, 3472222.24074426, -0.00660028125432385, 34722222.3222014, 34722222.3484368, -0.0168827238018964, 0.00847546885595454, 0.0458415696977052, 0.0309530670159255, 8680555.62965789, 8680555.68705874, 0.0597952776918328, 11574074.307465, 0.151273596189352, 0.176121163243079}, 
                                                     {0, 503.999999738638, -2.29223770752978e-07, 0, -13753.9999929805, 6.63357889391016e-06, 0, -3253.99999612784, -1.65117235557972e-07, -3254.00000017674, 149624.999924689, -7.72692080847526e-05, 0, 84458.3332555367, 84458.3333337768, 8333.33326496593, -6.72691261158603e-07, 8333.33334103979, 7.05322008519544e-06, 16666.6666650824, -868736.110677429, 0.000482266173245233, 0, -853513.888357478, -853513.888882154, -10583.3329113012, 3.46232887709884e-05, -10583.3333875514, -7.03842420353443e-05, -205361.109788679, -205361.111209047, -5.93053812170284e-05, -410722.222270279, -31749.9999185841, -31750.0000179624, 3008263.88738923, -0.00178433934290891, 0, 4474513.88726525, 4474513.88880124, 6666.66522203154, -0.000218458380384422, 6666.66687407848, 0.0003489965809783, 1916736.10229983, 1916736.11190278, 247152.77066073, -8.97962368443226e-06, 247152.778206104, 0.000670670045533219, 3833472.22256623, 26666.6657948189, 26666.6666407662, 741458.333312748, 741458.334355977, 40000.0001583968, -6512013.88562738, 0.00409472552088479, 0, -13494374.9974807, -13494374.9995324, -1666.66311351382, 0.00056300268370071, -1666.66723802198, -0.000889818652261969, -8952430.52977126, -8952430.55889148, -147222.204771622, 0.00102044990197491, -147222.22276554, -0.00315192122659951, -17904861.1118369, -8333.32936195254, -8333.33261904447, -2115624.95556493, -2115625.00337461, -0.000922188540877652, -6346875.0016732, -6346875.00622855, -588888.884617356, -16666.6646481696, -588888.892533415, -16666.6679159833, -883333.341670164, 8888888.88439612, -0.00586616290183104, 0, 24236111.1092086, 24236111.1099243, -0.00656651615891837, -0.000967360716502883, 0.00104742040060766, 0.00151012378011965, 22881944.4051822, 22881944.4508956, 34722.2007808573, -0.00181997322918517, 34722.2224783277, 0.00723623073438704, 45763888.8889162, -0.00824593511614973, -0.00200632377313578, 8645833.22133277, 8645833.34772083, 1145833.23177862, -0.000625619142025131, 1145833.33611136, 0.00577773953121739, 25937500.0040855, 25937500.0154545, 173611.091675164, -0.0105508039322804, 173611.114082973, 0.00322082913630741, 4583333.32610412, 4583333.35265106, -0.00100350572711066, 6875000.04068796, 347222.232609285, 347222.251749152, -7430555.55175474, 0.00510544270203938, 0, -25555555.5552454, -25555555.5540411, 0.00799630615944581, 0.000897796944320176, -0.00115498928090117, -0.00124720849747011, -32326388.8560665, -32326388.8946195, 0.0132311512415621, 0.00206965670689538, -0.000514432823314648, -0.00913504220889157, -64652777.7762296, 0.00975899840254382, 0.0025159777824367, -17951388.7552071, -17951388.9155225, -243055.442492169, 0.00526060180705145, -243055.552797647, -0.0120780867059718, -53854166.6700689, -53854166.684294, 0.0332758281358225, 0.0215701163948079, 0.00250775598335631, -0.00282097986783381, -3993055.34512307, -3993055.57205899, 0.00140550630317127, -15972222.2166468, -15972222.2589488, -1215277.74179239, 0.0123085116079303, -1215277.79865005, -0.00505999716835726, -23958333.4049945, 0.0208403679274731, -0.0439446724778126, -2430555.62085514, -2430555.66378619, -0.0558011155306603, 3472222.22042209, -0.00246454897696238, 0, 14583333.3337727, 14583333.3323897, -0.0053908588871171, -0.000516695486242553, 0.000662968119824917, 0.000611763628687494, 23611111.0965822, 23611111.1130875, -0.00395027635481357, -0.0010234315641197, 0.000948926555511817, 0.00547545583780991, 47222222.2204941, -0.00601273460689377, -0.00194961872999456, 18055555.4792674, 18055555.5763733, -0.0583721289959585, -0.0038936515726776, -0.00443834391044801, 0.0114713427717064, 54166666.6662498, 54166666.6749891, -0.0265231285440957, -0.0170433875984141, -0.00544515115525651, 0.00213966962198366, 6249999.821904, 6250000.02647805, 694444.271244625, -0.00607947345133073, 694444.447731092, 0.0037478530443808, 24999999.9975482, 25000000.032771, -0.0456863986531039, -0.0205855233317357, 0.00646724787507511, 0.00490657402225705, 37500000.0609234, -0.0461035194463288, 0.0213434130676199, 3472222.19390901, 3472222.25353789, -0.00759210181214135, 6944444.53246653, 6944444.57938785, 0.00612381623065523, 0.0264334443585743, 0.0991366277314591, 0.0673001802694239, 0.142369332198848, -694444.444079339, 0.000506289918534862, 0, -3472222.22242692, -3472222.22199236, 0.00148389854979595, 0.000125219618475333, -0.000157633449210224, -0.000120586502488643, -6944444.44176151, -6944444.44451416, 0.000644142171664093, 0.000152200821951803, -0.000412395718920415, -0.00135662541138614, -13888888.8883025, 0.00152188463853436, 0.000524728385391222, -6944444.4275696, -6944444.45021441, 0.0109342108798917, 0.000623355853532389, 0.00131847052439704, -0.0043358443166809, -20833333.3324391, -20833333.3345936, 0.00788926203780161, 0.00490121393512036, 0.00196104800114525, -0.000702787949810649, -3472222.16903803, -3472222.23442491, 0.0509027180815185, 0.0021140290159245, 0.00205477381236481, -0.00333638529264765, -13888888.8871612, -13888888.8990456, 0.0180339456284112, 0.0076834611505307, 0.000585594689883231, -0.00292449315804528, -20833333.349237, 0.0223971825721301, -0.00313924111542888, -694444.366105911, -694444.451136209, 0.00147813504410858, -3472222.21340158, -3472222.23656893, 0.0202151886717947, 0.00860241440157995, -0.00746604343421908, 0.00316752178138339, -6944444.48113125, -6944444.49691266, 0.0239683178450834, 0.0100200855540104, -0.0255228415516991, -0.024160624463335, -0.0181374953241147, -0.0587756507045008, -0.0255171420005485, -0.0941290066256393, -0.0542841054563202, -0.0830425369414529}, 
                                                     {0, -349.999999798641, 1.6435841226989e-07, 0, 9668.0555501497, -4.70914138546273e-06, 0, 1676.38888634758, -4.37724518470267e-08, 1676.38888899906, -106895.601793868, 5.42810965721122e-05, 0, -44263.6573622123, -44263.6574079548, -2993.05550764393, 1.51100119172572e-06, -2993.05556014595, 7.21072997710886e-08, -5986.11110962445, 632942.129295791, -0.000335256179814673, 0, 458089.120068076, 458089.120368826, 2361.11078454679, -1.8470986386303e-05, 2361.11114251362, 3.44796338384706e-06, 75424.7677330797, 75424.7685810894, -1.2505900154699e-05, 150849.537042143, 7083.33329291795, 7083.33332862901, -2240746.52662351, 0.00122884886145678, 0, -2474351.85094023, -2474351.85180865, -694.443189923844, 9.00347837487478e-05, -694.444560742946, -3.79219854283854e-05, -727100.689325811, -727100.694954927, -56689.8105117074, 0.000170019379887042, -56689.8150899939, -4.8796132321395e-06, -1454201.38893503, -2777.7774386248, -2777.77771474733, -170069.444316811, -170069.444691598, -4166.6666225288, 4965798.60860095, -0.00279753276205937, 0, 7727199.07269394, 7727199.07381478, -0.00329788885506987, -0.000179177564005961, 0.000317051737893263, 9.93515764245158e-05, 3543344.89249685, 3543344.90951883, 15856.4706552052, -0.000897077975318557, 15856.4818254299, 0.000364258876512753, 7086689.81486207, -0.00144392800459334, -0.000462176223055072, 506134.233297416, 506134.261488089, -0.000751266745805864, 1518402.77775312, 1518402.77930237, 63425.9240393273, -0.00113808166154286, 63425.9270033157, 1.27372311525825e-05, 95138.8911050599, -6940972.21876457, 0.00398137697159116, 0, -14417824.0730962, -14417824.0733948, 0.00602494773428164, 0.000275960227608705, -0.000587882905255924, -0.000276281428013511, -9539930.53294758, -9539930.55958205, 0.0140584604778526, 0.0012596940567799, -8.82464456740723e-05, -0.00113905457902743, -19079861.1107823, 0.00274244423146561, 0.00113828110648322, -2193286.9717721, -2193287.04626523, -130208.273929422, 0.0021751532679144, -130208.33545514, 0.000639605151723782, -6579861.11081261, -6579861.11510458, 0.00729472204475122, 0.00424788664843879, -0.000653856199314344, -0.000580707639667071, -520833.329002157, -520833.33858644, 0.00209720585453464, -781250.0106298, -0.00105964400716297, -0.00852804510629007, 5937499.99707494, -0.00344623753910906, 0, 15810185.1851202, 15810185.1843102, -0.00698681402392753, -0.000204335341987242, 0.000661608500073688, 0.000181990338463333, 14299768.4997378, 14299768.5220743, -0.00988947051268767, -0.00128376500558791, 0.000143022845230834, 0.0018522296201977, 28599537.0361975, -0.00321816488834107, -0.00136577376422501, 4918981.40371234, 4918981.49820424, -0.0659858669591627, -0.00490375091536434, -0.00122731343180842, 0.000551231016545071, 14756944.4434942, 14756944.4490286, -0.0115193736616499, -0.00844716710710566, -0.00144821849347277, -4.85124132406335e-05, 491898.025237802, 491898.159270889, -0.00559410987005924, 1967592.58656738, 1967592.60141455, -0.0147203103091132, -0.00683192601386711, 0.00572284454070685, 0.000615292210375624, 2951388.90540266, -0.0103673587518297, 0.0161741208200416, 0.0154287176226762, 0.0319967127450835, 0.0136224116598542, -2835648.14676281, 0.00165576131351214, 0, -9375000.00031311, -9374999.99945459, 0.00447710819781551, 0.0001148096137481, -0.000386856977683856, -9.56404239950283e-05, -11111111.1028443, -11111111.1123423, 0.00394978594413096, 0.000555461848975253, -0.000429240454271319, -0.00111712873990427, -22222222.2215001, 0.00197345673677915, 0.00110258791488522, -5439814.7706331, -5439814.82772257, 0.033842011856301, 0.00333920111513413, 0.00254109310525639, -0.00150563432484269, -16319444.4429055, -16319444.4463409, 0.00907326218276218, 0.00624476711268122, 0.00260197674587572, -3.64722834922115e-06, -868055.451348215, -868055.572589469, 0.10082159979072, 0.00632846130443908, -0.00294663393368977, 0.00273023547596623, -3472222.21792131, -3472222.23056839, 0.0166084701858706, 0.00990241353253103, -0.00127519112884911, -0.000751713543531539, -5208333.34884101, 0.0169677971721343, -0.0084232045963388, 0.0131444898916204, -0.00800004310960876, 0.00483497141032001, -0.0184400638706683, -0.0381699438261456, 0.00504165031907944, -0.00431865028846288, -0.0331628687881722, -0.0217235428317836, -0.0356019931254315, 578703.703422733, -0.000338824828101716, 0, 2314814.81494628, 2314814.8146822, -0.00118442506881266, -2.78076271491561e-05, 9.35839292236701e-05, 1.68320146280235e-05, 3472222.22070243, 3472222.22227242, -0.000893209601401318, -4.896463347284e-05, 0.000207743766082643, 0.000296394953030929, 6944444.44422468, -0.000495797011350794, -0.000293977210381472, 2314814.80510495, 2314814.81836467, -0.00630581339783846, -0.000597639242523605, -0.00080982052222526, 0.000969349149061211, 6944444.44382151, 6944444.4446787, -0.00271997580688393, -0.00164835179395922, -0.000862961743644331, -1.84632567168893e-05, 578703.67260843, 578703.711390682, -0.0293553920644898, -0.00193010879668709, -0.00098033374476832, -1.13842225336064e-05, 2314814.81292703, 2314814.81741061, -0.00613194556197005, -0.00335209675167569, -0.000494690399194195, 0.000861327181064877, 3472222.22575405, -0.00779182760786538, 0.00129755348212557, -0.0458558602607715, 0.00453218406895095, -0.00221449838139284, -0.00477474701678157, 0.00331039384469659, -0.00801494482664244, -0.00468269729641594, 0.00170369495831868, -0.00190017598092272, 0.00696988135379063, 0.0133018366418249, -0.0103716315130731, -0.0061932893523933, 0.00851836189697117, 0.0094485325746293, 0.00110174700132291, 0.0173007530218266, 0.00762983008901508, 0.0237923135778953, 0.0142867218385024, 0.0269711111330821}, 
                                                     {0, 171.428571321834, -7.15940539516332e-08, 0, -4776.19047332585, 2.04084824419036e-06, 0, -576.190475150884, 3.85529157942128e-08, -576.19047622658, 53426.1904454698, -2.33887695007883e-05, 0, 15402.3809354653, 15402.3809526179, 642.857122479541, -6.79029629038902e-07, 642.857144426209, -6.77801703486581e-07, 1285.71428518967, -320951.0580242, 0.000143596988834237, 0, -162177.248565358, -162177.248677258, -238.095091612453, 6.29094132788333e-06, -238.095248052764, 5.34913695707751e-06, -16464.2854163991, -16464.2857381253, 9.88244424041311e-06, -32928.5714277384, -714.285703270893, -714.28570972194, 1155694.44383296, -0.000523393763773212, 0, 895694.444108659, 895694.444430477, -0.000594302414823436, -2.81173836000145e-05, 3.41773973597681e-05, -1.83360118229545e-05, 162499.998079269, 162500.000204358, 5833.3316890132, -6.80421163132226e-05, 5833.33343179712, -5.0101331634274e-05, 324999.999994145, -7.32106402200954e-05, -3.89235556791879e-05, 17499.9999551, 17500.0000284075, -3.36327921785297e-05, -2610416.66533686, 0.00118571535449295, 0, -2874074.07357236, -2874074.07398035, 0.00161111953835134, 4.94800635318227e-05, -9.33985358817425e-05, 5.3085563775982e-05, -817361.105511972, -817361.111959964, 0.00416570739044575, 0.000325484571906177, -8.4806051316503e-05, 0.000104071646778322, -1634722.22218842, 0.000301414116669633, 0.000214636679780691, -53703.6939519064, -53703.7045891041, 0.000370477576406236, -161111.110994787, -161111.111284934, 0.000413695375332552, 0.000297080215603703, -0.000185096440113717, 0.000189867169084211, -0.000400804471901934, 3724206.34737451, -0.0016805262936514, 0, 5534722.22187942, 5534722.22196919, -0.00294434831169368, -7.53717558575059e-05, 0.000182274963679129, -2.96646743029994e-05, 2291666.65817839, 2291666.6682806, -0.00551408036025392, -0.000404806597367801, -9.85726421189618e-05, -0.00015713276567618, 4583333.33317452, -0.000505229971909417, -0.00047600639900508, 243055.530946765, 243055.559263088, -0.0222209523000431, -0.000660517675205761, 0.000822160262530644, -0.000658632708904474, 729166.666167805, 729166.667245133, -0.00144967793940928, -0.00081004698618989, -7.72425848949613e-05, -0.000333383644867304, -0.00117365556308033, 0.00078160847290526, -0.000941034142560376, 0.00172437023707264, 2.76653749178876e-05, 0.00122042671187411, -3253968.25241849, 0.00144967458652011, 0, -6286375.66137313, -6286375.66104624, 0.00336079184020132, 4.79629029851216e-05, -0.000216610089090604, 5.9837031418147e-05, -3611111.10407563, -3611111.11253849, 0.00407335422629357, 0.000457414759733882, 0.000120427151513802, 1.17011527086403e-05, -7222222.22194999, 0.000632385684771713, 0.000526711731173699, -578703.674326025, -578703.710385315, 0.024430264437435, 0.00158917539102155, 0.000562224967523697, 0.00073909292966891, -1736111.1103147, -1736111.11180466, 0.0020546715193422, 0.00184766825811942, 0.000697004241966076, 0.000917244568242906, 0.0463234603412678, -0.00452511963983225, 0.00217771570049137, 0.00226673935084846, -0.000844250939537544, 0.00331219876680656, 0.00165292111046165, -0.000776090465570507, 0.000428671582371748, -0.00154537895933377, 0.00187509069326593, -0.00309823223656213, -0.00297836884012234, -0.00575430450932019, -0.00151418016283001, 1587301.58656757, -0.000694233931819387, 0, 3869047.6191764, 3869047.6188419, -0.00211184754628605, -2.65951303993552e-05, 0.000131599758066231, -2.72109686852461e-05, 2976190.47309344, 2976190.47669033, -0.00179470103023382, -0.000215086208853175, 4.68758765208222e-05, -3.2527771684179e-05, 5952380.95218452, -0.00040820044201366, -0.000420337573465553, 694444.427781958, 694444.449577832, -0.0122431080562277, -0.00111096178579477, -0.00111176029685127, -0.000345852980348464, 2083333.33255202, 2083333.33350961, -0.00166471263323143, -0.00133383463783436, -0.000947091467553289, -0.000645116417397465, -0.0394909263297163, 0.0068707124245756, -0.0377433606841944, -0.00206336562635284, 0.00125564542346346, -0.00131677673959073, -0.00180679937854451, 0.00107278878910955, -0.00329557977737527, -0.00237242217940031, -0.000128558329504755, -0.000747900250222407, 0.00217427450513159, -0.00263464569019736, 0.00111756202763405, -0.0034178091856359, 0.000997708761746089, -0.00131412146943157, 0.00166854671568909, 0.0064159344020622, -0.0011140571306934, 0.000711209594234247, 0.00666913783548329, 0.00344296915775149, 0.00459866279104468, -330687.830538954, 0.000141692899038715, 0, -992063.492115384, -992063.492013451, 0.000549558198169065, 6.34783163964204e-06, -3.2770981000234e-05, 6.9386106930799e-06, -992063.491492948, -992063.492087578, 0.000448009303988137, 2.16121986435596e-05, -4.98159165885372e-05, 4.95039740336016e-07, -1984126.98407348, 0.000104763940432261, 0.000106794037370892, -330687.827036822, -330687.832095562, 0.00219708499444048, 0.000187219590611535, 0.000369439884739619, -6.87749480110762e-05, -992063.491822068, -992063.492064579, 0.000530407900086811, 0.000325184964345524, 0.000270886897808793, 0.000203876005071632, 0.0118038502887178, -0.00307760067021402, 0.0108313765427527, 0.000565805867199025, 0.000374881959867768, 0.000211694126886241, 0.000804021608026877, -0.00032673574877556, 0.00109740290928882, 0.00071615042370928, 0.000250623486788376, 0.000134589078166073, -0.000260453814804813, 0.00130025229815278, 7.31928496369166e-05, 0.0173504841609775, -0.0018599945771213, 0.000741247343116327, 0.0014558692224125, -0.000249451854722541, 0.00188368391229409, 0.00125119226292032, -0.000105717309185573, 0.000680421891092676, -0.00010502349493374, -0.00163969695368814, 0.0019850681557977, 0.00166263195835813, -0.00128023166343878, -0.00174500895752124, 0.000386578351427073, -0.00293928842113484, -0.00138246089361123, -0.0031200564124459, -0.00305642102354518, -0.0055855682213001}, 
                                                     {0, -56.2499999627369, 1.76229441411005e-08, 0, 1577.23214185732, -5.0115897616049e-07, 0, 118.749999757653, 0, 118.750000005538, -17797.7678464217, 5.72595324236745e-06, 0, -3204.01785345747, -3204.01785719102, -62.4999951273643, 9.52175465495319e-08, -62.5000002757999, 1.1848722597221e-07, -124.999999936685, 108108.382874787, -3.50430660619251e-05, 0, 34182.043626501, 34182.0436508943, -3.62136214085234e-05, -9.00139397089326e-07, 1.44664361219714e-06, -1.13835098440152e-06, 1620.5356489366, 1620.53571926716, -1.44873149717213e-06, 3241.07142853202, -1.42422837480149e-06, -8.02537172303739e-07, -394522.569231034, 0.000127285882922392, 0, -192057.291593673, -192057.291664063, 0.000151914305158879, 4.22285464217069e-06, -3.93308250381136e-06, 4.4933515345581e-06, -16284.721801389, -16284.7222687391, 0.000360763141723593, 8.16759935261151e-06, -1.77671537287877e-05, 8.59448947258584e-06, -32569.4444424027, 6.86051577861107e-06, 6.87625023565993e-06, 5.26029408903114e-06, -1.29533050755022e-06, 5.44333742265916e-06, 905121.527313625, -0.000287479352863486, 0, 629687.499891629, 629687.499980276, -0.000420478751526043, -6.94680708009934e-06, 1.1031927993322e-05, -1.42831454972397e-05, 83940.9709894189, 83940.972419349, -0.000906460766879774, -4.90599387966847e-05, -1.8355365817311e-06, -2.64076776450875e-05, 167881.944438507, -2.778188786671e-05, -3.65952115736964e-05, -0.00212992320058229, 0.000195033230287596, -5.13624239186397e-05, -1.81327454191973e-05, 1.03918321959786e-06, -3.43253083677381e-05, -3.10859211084835e-05, 2.76682223406377e-05, -4.93014344214023e-05, 4.81432552028078e-05, -1314236.11047166, 0.00040641822492246, 0, -1244791.66659388, -1244791.66661162, 0.000772695592137581, 1.17834146739742e-05, -2.56401188308896e-05, 1.03937084413104e-05, -243055.553681226, -243055.55593167, 0.00119523275521664, 4.61139105851345e-05, 7.30664339136431e-05, 5.62982445388101e-05, -486111.111082574, 3.30932144240761e-05, 7.16766228844371e-05, 0.00542719134817481, -0.000852246685209112, 0.00481283402028653, 2.99385479267478e-05, -0.000154114531807589, 9.75695047891141e-05, 0.000104933858657426, -4.00480972630617e-05, 0.000114038753650012, 4.24322525116918e-05, 1.85626401314722e-05, 0.000102010240924659, 0.000116576786847419, -5.94398614784754e-05, 0.000165207102098234, -0.000172439436132775, 8.27191189165735e-06, -2.61031886366731e-05, 1170634.92009388, -0.000349869773994319, 0, 1458333.33333513, 1458333.33326069, -0.000878611706817007, -7.08874487418968e-06, 3.50734238098099e-05, -1.70476255703331e-05, 399305.554002538, 399305.555890252, -0.000890548862355447, -6.83012136364948e-05, -8.82512108861708e-05, -3.93630240294556e-05, 798611.11106522, -5.7529272091691e-05, -6.67764864456548e-05, -0.00650886240844427, 0.00154220250471322, -0.00520388264705605, -0.000195338862071353, -0.000192641069523019, -0.000151760492196139, -0.000151760500453198, 5.91313816812168e-05, -0.000106441575341415, -0.00019870097699163, -0.000103060035267841, -0.000242507009722525, -0.0101899115257006, 0.00102876222656246, -0.000310595113415971, -0.000345201397143247, -4.91140328583619e-05, -0.000315702676147241, -0.000127064877785179, 7.14699947838477e-05, -7.72491370414625e-05, -0.000111413868381914, -6.9214106125148e-05, 0.000292721178539088, 0.000530451356926017, 0.000501497996019989, -0.000102953141174368, -582837.301331019, 0.000167122137143004, 0, -930059.52383919, -930059.523764045, 0.000547571837043474, 4.26041511644409e-06, -2.31837743357474e-05, 7.75986044473403e-06, -347222.221535875, -347222.222341064, 0.000408658496169764, 3.36246924735091e-05, 2.98908547376057e-05, 3.49457495618917e-05, -694444.444413954, 4.3962816867389e-05, 5.9508471785098e-05, 0.00369155443211092, -0.00118572093945683, 0.00250363859046883, 0.000148369850050945, 0.000312776320264166, 0.000100257699699313, 0.000151681779365884, 1.09289355902454e-05, 0.000109967050041154, 0.000138981269961048, 0.000139798068234524, 0.000152578318654342, 0.00876068248658385, -0.00158684589910675, 0.00822139781564639, 0.000249985537199295, -0.000263508230297604, 0.000176087821579305, 0.000270969250037433, -3.35119993184011e-05, 0.000258949184134347, 0.000270685265344497, 3.99750565163833e-05, 0.000221252482648856, -0.000157138710882057, 1.41961882982397e-06, 2.71787680404467e-05, 0.000370901573918558, -4.48291311677641e-05, 7.92314795079051e-05, 2.65072478563038e-05, -0.000513472727408114, -0.000122675035294235, -5.03292446713642e-05, -0.000681028044279055, -4.45228381152248e-05, 9.09908963222925e-05, 124007.936455952, -3.40484008168635e-05, 0, 248015.873027622, 248015.873004779, -0.000141517172766161, -1.101074388031e-06, 6.08656326203451e-06, -1.85371923406671e-06, 124007.93638086, 124007.936514589, -0.000109263312570258, -3.04357516844598e-06, 1.12784843900026e-06, -6.76120514201307e-06, 248015.873008167, -1.18585782194827e-05, -1.37405825172342e-05, -0.000808280457454011, 0.000325319287687393, -0.000416934029956274, -1.70094630090206e-05, -0.00010595891722519, 8.53519589300738e-07, -4.26091847991924e-05, -7.94723686479956e-06, -4.60673238491558e-05, -2.67267792826619e-05, -2.9378647152977e-05, -4.46080950269258e-05, -0.0026279413491958, 0.0007115148112601, -0.00231330367603469, -5.44980172268909e-05, -0.000100305364457587, -2.51112255143065e-05, -0.00014149691106638, 5.6459643708175e-06, -6.04714260103873e-05, -5.84815256105713e-05, -3.55607863622318e-05, -5.37792712937946e-05, -3.84747260896958e-05, -5.5971238097594e-05, -7.22724910934261e-05, -0.00383690382544187, 0.000427697206838737, -8.24887581548401e-05, -0.000193415316756943, -4.27820879400157e-05, -0.0002182990672439, -0.000142724014299688, -1.47919261967474e-05, -8.12025648347891e-05, -0.000180192674091171, 5.0215371857053e-06, -0.000121355742153299, -0.000294456648399935, -6.36861843268151e-07, 0.000102023894345491, -2.51490291398668e-05, 0.000199353321416706, 9.35714741342131e-05, 4.68622481365185e-05, 0.000480194639559335, 0.000636094757952682}, 
                                                     {0, 11.111111103374, 0, 0, -313.095237887681, 5.37820019279557e-08, 0, -11.1111110865968, 0, -11.1111111114576, 3557.16490077269, -6.13723478999258e-07, 0, 301.984126630903, 301.984126989467, -5.0227740877111e-07, 0, 1.9049790273597e-08, 0, 0, -21796.8474298678, 3.75283632230627e-06, 0, -3255.18077367975, -3255.18077604213, 3.82305893841552e-06, 0, -4.65052060866251e-08, 0, 6.28567008188343e-06, -4.51331169468621e-07, -2.17676191310469e-08, -1.24338365379298e-07, -3.86012924313337e-08, -3.15045674383338e-08, 80405.092548284, -1.35980335122163e-05, 0, 18541.6666596003, 18541.6666664784, -1.64127322352638e-05, 5.32263800823879e-08, -5.94763016695708e-08, 1.24181710092585e-08, -4.08047760717796e-05, 4.69254745365847e-06, -3.45337939472887e-05, 1.54488634127765e-08, 1.20237664456783e-06, 0, 2.71678663459084e-08, 7.18714964840446e-08, -5.73268047994567e-08, 3.16908462998637e-07, 7.96350949276664e-07, 2.97770856607687e-07, -186863.425829543, 3.06488292935585e-05, 0, -61863.4259154857, -61863.4259241363, 4.61796695876678e-05, 8.79348954336012e-08, 1.15072631417407e-07, 3.91029375647887e-08, 0.000120652177912196, -2.04293638905176e-05, 8.42599624449853e-05, 3.28267064993566e-07, 2.89055082324682e-06, 8.26354561119516e-07, -1.07899347065063e-06, -8.77983600083903e-07, -9.10805236970833e-08, 0.000204859443730113, -1.85939705472701e-05, -7.01201917863138e-07, -4.37961500907834e-07, -1.90751553529704e-06, -1.82582023832301e-06, -6.52493988721144e-07, -4.08243319065533e-06, -5.95022972168425e-08, -2.73231327279367e-06, 275462.962830157, -4.32770538984493e-05, 0, 124999.999993, 124999.999994722, -8.57597879105274e-05, 3.48420680647356e-08, 6.53601370837911e-07, -9.0980792508642e-08, -0.000184535689181002, 3.90723696292831e-05, -0.000108595061597464, 6.71377654212601e-07, -1.40124841076751e-05, -7.64317133390933e-07, -3.73190583617388e-07, 1.42124984614336e-06, 1.27873974461274e-06, -0.000531469116313745, 8.6532382965524e-05, -0.000457401964112308, 5.74596801217302e-07, 1.01908268131958e-05, 1.53138142175254e-07, -2.50870439166019e-06, 1.38383148014204e-06, 5.98399734512056e-06, -9.73530421938122e-07, 7.18814738969357e-06, -2.50131229814296e-06, 7.74163617598603e-06, 5.12491467128942e-06, 5.80856169343808e-07, 2.09217302956998e-05, -6.45046628257493e-06, 1.25464970756172e-05, -249669.312056927, 3.72276867813415e-05, 0, -150462.962963246, -150462.962955831, 9.79463393262421e-05, 1.36037081708498e-07, -1.77405324755485e-06, -5.26021138229593e-08, 0.000152589524791584, -3.50568194797001e-05, 7.89050676737676e-05, -2.15652621620502e-07, 1.6231142412929e-05, 3.7156053376506e-07, 1.47569646773286e-06, 4.80584516310312e-07, -2.4117597683469e-06, 0.000642123330884023, -0.000157410170640146, 0.000484967886083938, -1.47742077809482e-06, 2.84948872924479e-05, -1.08673758723845e-06, 2.27541288239466e-07, -3.33183181148366e-06, -1.33153983091992e-05, 1.44285894937118e-06, -5.97466707105634e-06, 4.98679035697142e-06, 0.000993656737002247, -0.00010183892091265, 2.63471395992272e-06, 2.8921813834738e-06, -1.03081145271807e-05, -1.55230751640986e-05, -2.28765077919059e-06, -1.18766905607118e-05, 1.09275122346121e-05, 6.23069457085114e-06, -1.23673638715078e-05, -8.44250599494007e-06, -3.89198517017556e-05, -1.14739939243214e-05, -2.16702555752731e-05, 126763.668377093, -1.7727261883753e-05, 0, 99206.3492093427, 99206.3492018752, -6.08526208764394e-05, 0, 1.50439103785566e-06, -2.5422859035848e-08, -6.80543556196011e-05, 1.2564590848277e-05, -3.62542347332142e-05, 8.06286016198298e-07, -8.18530645801228e-06, -3.3751480999062e-07, -1.299588014473e-06, -2.34309849273192e-06, 7.20017452489232e-07, -0.000363923326895798, 0.00012150786977394, -0.000217655248418496, 6.88800622735779e-07, -3.94887576273687e-05, 2.27059049395212e-06, -3.81151790604294e-06, -1.18667210788551e-06, 6.76737227972712e-06, -1.00254981859941e-07, 3.20563210426033e-06, 1.38094943879779e-06, -0.000864103366826835, 0.000161550851699494, -0.000790743242843221, 5.31748679213881e-06, 2.17636401958045e-05, -2.69981418978868e-07, 1.28135199183369e-06, 1.00991305537363e-06, 1.15204777920954e-05, -4.30623180385035e-06, 4.87480156758864e-06, -1.04732102312564e-05, -2.69649886917969e-06, 3.88164253811695e-05, 5.50669705579041e-06, 1.60155945108029e-05, 5.34194151578368e-06, 3.23851304938767e-06, 1.11740941596418e-05, 1.83292760899107e-05, 4.49974659257277e-05, -1.89106147012641e-05, -1.81798693103819e-06, -5.11798777642519e-05, -4.24810041021477e-05, -27557.319213185, 3.60691764819529e-06, 0, -27557.3192251612, -27557.3192228847, 1.57252350438355e-05, 1.65283229210779e-08, -4.44185519996028e-07, -1.67386671604305e-08, 1.26602486580632e-05, -7.88173091061844e-07, 1.0489545949689e-05, 1.90729958781872e-07, 1.18150876219905e-06, -6.59727904474537e-08, 4.50219519576707e-07, 4.45446484804635e-07, -3.09196971860056e-07, 7.97793965811731e-05, -3.34233720286553e-05, 3.15194522750646e-05, -5.60370485653495e-07, 1.36825840446395e-05, -3.04906111255803e-07, 1.52347129041344e-06, 7.75003363576774e-07, -1.53274136134402e-07, -1.27207643548328e-06, -1.59488100455832e-06, 6.07593821539277e-07, 0.000260613638590828, -7.27539138877643e-05, 0.000216560102204329, 7.52666698626804e-07, 1.26465935638546e-05, 5.43914204046275e-07, 2.80371606640488e-06, 5.66251699525924e-07, -4.68153176428728e-06, 9.23074590515463e-07, -2.07427742805759e-06, 1.84793088583583e-06, 5.59643641736403e-06, -8.11614416598389e-06, -1.84046260648157e-08, 0.000376702136457564, -4.26724746876714e-05, -1.36020063176563e-06, -2.41787642221657e-06, 1.40482070669631e-06, 2.80193740421431e-06, 1.43256668881515e-06, -8.35695771376816e-07, 2.28603599105698e-08, 1.00569714638371e-05, -6.72672699849158e-06, -1.65059451641794e-06, 2.59982010457311e-05, -4.13773422074069e-07, 1.55463579953985e-05, -1.12185144747428e-05, -6.51252435278249e-06, 1.67953208817166e-05, -6.83725129520289e-06, -2.59812858267517e-05, -2.19212581143983e-05}, 
                                                     {0, 0, 0, 0, 0, -5.37820019281841e-08, 0, 11.111111086059, 0, 0, 0, 6.1372347900175e-07, 0, -301.984126623138, 0, 5.07288637012895e-07, 0, 0, 0, 0, 0, -3.75283632232077e-06, 0, 3255.18077363475, 0, -3.84728454714601e-06, 0, 0, 0, -6.35108221482888e-06, 0, 2.17676191426107e-08, -1.00219017042052e-08, -3.24382169229822e-08, -4.92399108411272e-08, 0, 1.35980335122664e-05, 0, -18541.6666594656, 0, 1.64834726971589e-05, -5.32263801084076e-08, 0, -1.24181709172161e-08, 4.11504638975935e-05, 0, 3.48033082046592e-05, -1.54488634144855e-08, 0, 0, 6.65712878054391e-07, 2.22172313989596e-07, 3.42638291496198e-07, 4.87616729539701e-07, 1.40354652280532e-07, 1.70724625897644e-07, 0, -3.06488292936661e-05, 0, 61863.4259152655, 0, -4.63116625971628e-05, -8.7934895464679e-08, 0, -3.91029378827137e-08, -0.000121610625584731, 0, -8.49040636900419e-05, -3.28267064660723e-07, 0, -8.26354561884711e-07, -9.4715801096589e-07, 2.64482647140249e-07, -6.02961448211395e-07, -0.000206053789676218, 0, 7.01201917514536e-07, -3.28759424933736e-06, -1.48742335011963e-06, -6.45432455507004e-07, -6.74031780170656e-07, 1.22497857659646e-06, -7.85694088815906e-07, -1.28284272110061e-06, 0, 4.32770538985946e-05, 0, -124999.999992819, 0, 8.59185476917005e-05, -3.48420678419441e-08, 0, 9.09807929912854e-08, 0.000186038020943489, 0, 0.000109535311158661, -6.71377655394597e-07, 0, 7.64317134098773e-07, 3.75268902923886e-06, -5.20579429068588e-07, -3.68952680640736e-07, 0.0005341705060227, 0, 0.000459660757608846, -5.74596800204034e-07, 0, -1.53138139305029e-07, 1.06436166939813e-05, 7.94515125184573e-06, -1.80550799341852e-06, 3.05899524825815e-06, -1.3237717720557e-06, 4.71506259947069e-06, -2.48089153021426e-08, 1.98058369029917e-06, 2.63257490633217e-06, -6.66533942521142e-06, 1.48109555352694e-05, -4.71589781762568e-06, 0, -3.72276867814613e-05, 0, 150462.962963207, 0, -9.8065590603128e-05, -1.36037082234817e-07, 0, 5.26021135631573e-08, -0.000153914059393838, 0, -7.97337202622507e-05, 2.15652622951443e-07, 0, -3.71560533789874e-07, -4.32391890887717e-06, -1.29607493051441e-06, 1.85392097488568e-06, -0.000645413794009018, 0, -0.000487398234811146, 1.47742077713098e-06, 0, 1.08673758452612e-06, -1.02844103120878e-05, -7.35294803306589e-06, 8.26352459560285e-06, -3.51438183332718e-06, 7.26718779378647e-07, -7.70582377967292e-06, -0.00099747543611691, 0, -2.63471396209069e-06, -1.74438699818108e-05, -5.38247388991183e-06, 4.32031792207106e-06, -3.00524623200712e-06, 2.232516991366e-06, -1.55240581612579e-05, -2.66693899549171e-05, 1.28384918551351e-06, -4.66302267192321e-06, 1.39625148509479e-05, -1.87183708119565e-05, 1.57185898560812e-06, 0, 1.77272618838081e-05, 0, -99206.3492093806, 0, 6.09035302216388e-05, 0, 0, 2.54228591870741e-08, 6.86560759509013e-05, 0, 3.66597796017481e-05, -8.06286016974843e-07, 0, 3.3751480991391e-07, 2.61702702978748e-06, 2.81259420928729e-06, -1.77537609207671e-07, 0.000365969831141369, 0, 0.000219055178643266, -6.88800622006184e-07, 0, -2.27059049152701e-06, 9.74774342530546e-06, 7.17230060035603e-06, -3.53005233788763e-06, 1.25663808862096e-06, 2.18482599181483e-07, -1.72402908201138e-07, 0.000867217875204008, 0, 0.000793448796453927, -5.31748678968932e-06, 0, 2.69981424437322e-07, 1.11531246317048e-05, 1.08777237454754e-05, -3.96213668998784e-06, 7.59815747426554e-06, 4.09745829486416e-06, 1.36820646376681e-05, 2.31380963197492e-05, -3.11891940211325e-05, 1.3896951153436e-06, -1.30702187436973e-06, 9.14656924760177e-06, 1.11855136957411e-06, 1.97167648241617e-05, 2.08603859015963e-06, -2.39348956345017e-05, 3.37934914659321e-05, 1.98419847410186e-05, 6.44432634187172e-05, 6.5364373482417e-05, 0, -3.60691764820612e-06, 0, 27557.3192251806, 0, -1.57346664796787e-05, -1.6528322960713e-08, 0, 1.67386671422366e-08, -1.27664211596589e-05, 0, -1.05742247894282e-05, -1.9072995845578e-07, 0, 6.59727904540999e-08, -6.86379065737199e-07, -5.1531343370124e-07, 2.18546934322531e-07, -8.02867921257476e-05, 0, -3.18551224274386e-05, 5.60370485883351e-07, 0, 3.04906111015586e-07, -3.01452907540767e-06, -2.17027158050577e-06, -5.70740259389232e-07, 8.99844996959988e-07, 1.04328628195171e-06, -9.51425942227242e-07, -0.000261594612192571, 0, -0.000217326858535076, -7.52666699831586e-07, 0, -5.43914205874669e-07, -6.4737821023081e-06, -2.95478455352309e-06, 2.44171816282583e-06, -1.81439023841673e-06, 2.08476954444764e-07, -2.47744573415933e-06, -1.29222714091731e-05, 5.96874023024978e-06, -2.47411072284953e-06, -0.000377797501773807, 0, 1.36020062865542e-06, -3.00979505886027e-06, -5.18356622222704e-06, -7.34965179458028e-06, -2.81363892972409e-06, -3.14778771794839e-06, -7.01837853526064e-07, -2.26162278967829e-05, -4.15992964874354e-06, -3.25754897551182e-06, -3.13630486497491e-05, -9.50317507076755e-06, -2.1107766956436e-05, 2.04309281667204e-06, -3.90716048102116e-06, -2.04678827476749e-05, -1.14399369470021e-05, 1.62532591703121e-05, 8.72962019412895e-06}, 
                                                     {0, 0, 0, 0, 0, 1.84224150169869e-08, 0, 6.2499999945476, 0, 0, 0, -2.03479095022684e-07, 0, -162.053571180128, 0, -62.4999999412481, 0, 0, 0, 4.5669373691421e-08, 0, 1.22586822560872e-06, 0, 1628.47222072539, 0, 3.68577783263871e-07, 1.61388137811277e-08, 0, 0, 1620.53570947501, 0, -1.05896659988075e-07, -8.24237423233476e-07, -1.0264872631771e-06, -2.04115634432833e-07, 0, -4.19148827778665e-06, 0, -8394.09721869928, 0, -5.18126182617044e-06, 2.94758782887907e-07, 0, 8.04565896772426e-08, -16284.7221949085, 0, 2.90445572242287e-05, 2.11336117054865e-08, 0, 9.20869088012843e-08, 2.15275186953192e-06, 6.03356236584137e-06, -8.61234207534488e-08, 6.97967392101234e-06, 3.14805079209699e-06, 7.99935026926507e-06, 0, 9.15338577862691e-06, 0, 24305.5555511671, 0, 2.3011591604652e-05, 1.77126821438818e-07, 0, 2.31866779115329e-07, 83940.972156146, 0, -9.58064970313847e-05, 1.19185045483929e-06, 0, 4.12247589407525e-06, -1.61396371708128e-05, -2.63025210445567e-05, 6.05094529534922e-07, -0.000164879023524709, 0, -2.68423307454135e-06, -2.37052535512294e-05, -1.13824093428467e-05, -4.1942225436835e-05, -4.07295793239992e-05, -1.28134410048209e-05, -3.12282935475448e-05, -4.01478356439724e-05, 0, -1.2382536213304e-05, 0, -39930.5555548444, 0, -5.04359227205848e-05, 5.52538748827964e-07, 0, -3.6064988334856e-07, -243055.555475544, 0, 0.000144353269760797, 3.69722746022524e-06, 0, -3.10615748509412e-06, 2.62266160193348e-05, 5.35991825053664e-05, 1.48869714203675e-06, 0.000325924860166435, 0, 0.000436051961849944, 1.15627758999086e-06, 0, 1.34080205864335e-06, 5.6954761717502e-05, 1.55460977998077e-05, 0.000127785084861853, 8.70229321944886e-05, 1.09275886085467e-05, 4.52420947537878e-05, 0.000127879112710901, 9.39040571683873e-06, 0.000133341225662483, 0.000153820260437952, 0.000152049354752173, 0.000136499619054109, 0, 1.04909730569247e-05, 0, 34722.2222248783, 0, 6.12201050785201e-05, 2.52910427320393e-07, 0, -1.9282292319761e-07, 399305.555496725, 0, -0.000111072182678508, -1.14697414387176e-06, 0, 1.20873091922235e-06, -3.10964547360268e-05, -5.59274563691782e-05, -3.79122902697308e-06, -0.000326077548268873, 0, -0.000551021649262174, -6.94495946471697e-06, 0, -5.88446359549425e-06, -0.000129041706155444, -5.55279891609799e-05, -0.000213950505570027, -0.000119863405433577, -7.87361490512162e-07, -3.98064486813094e-05, -0.000689709589731023, 0, 1.2976605143038e-05, -0.000133293163416919, -1.78996295822464e-05, -0.000231183480569941, -0.000220307678441916, -7.74913349360356e-06, -0.000172183225976468, -8.47959146743116e-05, -0.00044938150034751, -0.000132195379496904, -0.000480199370887586, -7.60378800774953e-05, -0.0004483261095401, 0, -4.65686380170384e-06, 0, -12400.7936529907, 0, -3.68010699502305e-05, 2.86420250082282e-07, 0, -1.21606431001598e-07, -347222.222207628, 0, 4.27756383306582e-05, 3.90732896643218e-06, 0, -1.34698330059837e-06, 1.92933515734415e-05, 3.11766254810847e-05, -5.1758191563941e-06, 0.000167899739550664, 0, 0.000392421341790114, 3.80695228272992e-06, 0, 1.04212864946757e-05, 9.29023438682639e-05, 3.93671962161566e-05, 0.000148289888600665, 9.05582248171761e-05, -9.07163289996235e-06, 3.96965254476832e-05, 0.000489179739454145, 0, 0.000666931616441193, 2.25849311150244e-05, 0, -7.77361845048651e-07, 0.000177791316975784, 1.19603031215929e-05, 0.000207229274572964, 0.000153434202027836, -3.25936757189805e-05, 0.000113547211214949, 0.000117731271933525, 0.000550659308665208, 7.73250054751284e-05, 0.000211156358021492, -3.83863742818102e-05, 0.00026107553718254, 0.000239564886304726, 0.000144316053482175, 0.000647573221181773, 0.000492070947984482, -2.10135634840269e-06, 0.000136586702921424, 0.000198303297740285, 0, 9.18349441625926e-07, 0, 6.04353105609169e-07, 0, 9.27433639321404e-06, 2.92767564434682e-08, 0, -7.64960583771921e-08, 124007.936506909, 0, -3.46739636510898e-06, 8.15430877128767e-07, 0, -2.65155258921217e-07, -4.67141635460057e-06, -9.48552467713062e-06, 8.18811855479106e-07, -3.11336385871836e-05, 0, -0.00010242137101136, -2.63633723397058e-06, 0, -1.47330243382699e-06, -2.86737359086169e-05, -1.33336663083075e-05, -3.97287132004333e-05, -3.18822984269775e-05, -1.46810192834257e-06, -3.40265697872129e-06, -0.000128599195609967, 0, -0.000241494567724989, 2.55282768983471e-06, 0, 1.77536891110245e-06, -5.4686018610797e-05, -1.5408907490676e-05, -8.06384537639724e-05, -4.26856895322357e-05, 6.14373321393545e-06, -4.94258131197223e-05, -3.36184365408236e-05, -0.000168622550621421, -1.12021075242466e-05, -0.000243690883859071, 0, -7.0753280811372e-06, -8.15294302568325e-05, 1.16376062547655e-05, -4.18697717130578e-05, -8.25378341045547e-05, 2.18419550949021e-05, -0.000104062950180141, -2.78163053070017e-05, -6.15967535972515e-05, -0.000223172716601709, -0.000135004561174453, 2.14992818386835e-05, -6.95885443301929e-05, -0.000218669053468899, 6.19342622603809e-07, -0.000216639844082094, -0.000127410316177077, -0.000436531341679403, -0.000277846075881597}, 
                                                     {0, 0, 0, 0, 0, 4.17332359754716e-08, 0, 4.76190477829913, 0, 0, 0, -4.52789564648429e-07, 0, -116.666666662793, 0, -71.4285719476834, 0, 0, -1.1156814829367e-08, 7.86600970138471e-08, 0, 2.47249377441743e-06, 0, 1074.07407361513, 0, 238.09524319743, -7.9198353935024e-08, 0, 1.59248984578499e-08, 1750.00000162029, 0, 3.12001544229107e-07, 1.24427748209054e-07, -9.47362107506063e-07, -6.93568244332288e-07, 0, -8.51802032937648e-06, 0, -4861.11110849166, 0, -2.69489977161985e-05, -7.4873496324285e-07, 0, -3.0008301286302e-07, -16111.1111096821, 0, -5833.3333469243, 1.25993569701905e-07, 0, -4.65163471667019e-07, 6.82387707205315e-06, 9.98125390341609e-06, 4.54491801047032e-06, 1.10258582727025e-05, -1.84610170802311e-06, 1.41545100862659e-06, 0, 1.74562411378399e-05, 0, 11574.0740703265, 0, 7.6214838835379e-05, -8.31491174438477e-07, 0, -8.11113406911909e-07, 72916.6666534559, 0, 9.67669546697292e-05, -2.92290631278181e-06, 0, -1.21283240797824e-05, -9.87398349277443e-07, -2.35353386069296e-05, -1.17442830901899e-05, 53703.7037294699, 0, 5.70115597208114e-06, -4.43628123119672e-05, 9.72343770424098e-06, -6.31822525420701e-05, -3.35940845422087e-05, 3.12468125028651e-05, 7.98870978578404e-06, -2.16343195658114e-05, 0, -2.3555504571149e-05, 0, -13888.8888830021, 0, -0.000128598083774327, -1.09996781571687e-06, 0, 8.22389885771215e-07, -173611.111077761, 0, -0.00020748388065744, -1.02871937052292e-05, 0, 7.38775464040654e-06, 3.02890544002199e-05, 3.81325798531736e-05, 9.36311603569868e-06, -243055.55551353, 0, -9.82852799374591e-05, 1.01984208944509e-06, 0, -5.65692961460546e-06, 9.76941407129062e-05, 3.32580630998579e-05, 0.000145503231409399, 0.000144611185522076, -4.1339623187349e-05, 4.53876199715539e-05, 7.05175895258419e-05, -4.43980424618925e-05, 1.51297873542648e-05, -0.000177785370851732, 0.000348845275329342, -0.000129951258012002, 0, 1.86857486854158e-05, 0, 6613.75660988682, 0, 0.000121260163297036, -1.25918008843378e-06, 0, 3.43941090488762e-07, 208333.333324173, 0, 0.000249877990493579, 2.42656409730745e-06, 0, -1.0289341497385e-06, -4.26372696583035e-05, -5.30956302249741e-05, 2.61286774644543e-07, 578703.703620764, 0, 6.51095065168667e-05, 1.99043500306947e-05, 0, 2.08687046591507e-05, -3.62261942690897e-05, 8.20299475740639e-06, -0.000157640021391452, -0.000212672849933817, 2.79827581502355e-05, -0.000103818441198936, -4.5683424057602e-05, 0, -3.86442731190165e-05, -0.000218726595347068, 5.72507437474664e-05, -0.000226830680921031, -0.000120603105036342, 0.000107167240731146, -9.8796699232669e-05, -0.000120521127497421, -0.000502443362938209, 0.000156790580414457, 0.000153637253200623, 2.27209668646933e-05, -9.64527535922128e-05, 0, -9.01848550200363e-06, 0, 4.54405665601235e-07, 0, -6.50057967448638e-05, -6.87298346616027e-07, 0, 3.43260642789206e-07, -99206.3491915142, 0, -0.000181299650755578, -9.99005392384129e-06, 0, 3.13661867009058e-06, 2.88087770488267e-05, 3.88657798736429e-05, 1.89740229427822e-05, -694444.444414777, 0, -0.00014671193499002, -1.14604198967804e-05, 0, -2.75140530394019e-05, 6.3414152944385e-05, 3.37565543883129e-05, 0.000145039109227044, 0.000117140101387237, 1.30545315105496e-05, 3.24250873460507e-05, 7.60243264376246e-05, 0, -1.64351087364304e-05, -5.7589717265577e-05, 0, 1.73803132112144e-06, 5.8183900798536e-06, -6.9704056942721e-06, 0.000323949258169836, 0.000181617850586607, 7.24807548103653e-06, 0.000186681452970593, 4.38365312897168e-06, 0.000272615039729214, -0.00010483384012641, 0.000116221522701713, 1.64244029162138e-05, -4.77289857370987e-05, 0.000128825331357147, -0.000406380233645189, 0.000246502099108399, 0.000759154870010363, 2.23936083663621e-05, 0.000688167078500152, 0.000767114664722323, 0, 1.78361810556807e-06, 0, 1.50572393543053e-07, 0, 1.41677591293691e-05, -1.95981455041099e-07, 0, 2.19193678728243e-07, -2.44170336273889e-06, 0, 4.85021864987788e-05, -2.46992035247148e-06, 0, 7.16614053592524e-07, -8.0367406745079e-06, -5.78579336014402e-06, -2.09099349311942e-06, 330687.830680529, 0, 5.33329225003088e-05, 7.61303881389044e-06, 0, 4.49091654872363e-06, -2.76914352575677e-05, -1.29196862023685e-05, -5.41702856353723e-05, -1.14970397137823e-05, 6.28983783018493e-06, -2.55045556607776e-05, -2.00872918462625e-05, 0, 3.27227507673927e-05, -5.12834683615272e-06, 0, -3.18041785852088e-06, -5.10613709097775e-05, 1.49360489915025e-05, -0.000106505746981725, -7.39142500939528e-05, 6.14108520203732e-06, -2.64038102801501e-05, -6.31723856614431e-05, -0.000186968417163474, 2.23968597765964e-05, -1.33643608136849e-05, 0, 2.14172753742742e-05, -2.17617966957776e-05, -2.78379247233338e-06, -0.000218736672050351, -3.50190469751517e-05, -1.8039590898827e-05, 3.46083492161829e-06, -0.000150631566031477, 0.000134996129257883, -0.000502230254331237, -0.000719212796400225, 2.00596358616711e-07, -0.000139488923710593, -7.43477382051452e-05, 7.07958940372479e-05, -0.000271773643651538, 2.55778530503779e-05, 2.39937326103937e-05, 0.000318723425240505}, 
                                                     {0, 0, 0, 0, 0, 3.65879035842114e-08, 0, 4.16666667589692, 0, 0, 0, -3.74355323734284e-07, 0, -95.1388888346508, 0, -76.3888890958501, 1.02504088987155e-08, 0, 2.12944799383587e-08, -1.34704498821556e-08, 0, 2.37292629084756e-06, 0, 781.249999032597, 0, 416.666669136717, 1.71666439539358e-07, 0, -5.1055356760138e-08, 1744.21296246207, 0, -6.24419882719378e-07, -1.56042802552587e-06, -6.26866382483594e-07, 4.13520059816467e-07, 0, -7.8540753186984e-06, 0, -2951.38888562261, 0, -694.444454377808, 1.48628479210315e-06, 0, 6.88621813666395e-07, -14322.9166575338, 0, -9513.88889801656, -3.82402862330256e-07, 0, 1.27013750394081e-06, -5.49266784050776e-06, -5.3821754735625e-07, -3.62762034951643e-06, -6.21351146778102e-07, 5.24844773690394e-06, 1.28486093476293e-05, 0, 1.82724186039718e-05, 0, 5208.33332161799, 0, 3.1042663601266e-05, 1.59934828687583e-06, 0, 1.80406277091464e-06, 54108.7962607696, 0, 15856.4814694138, 4.86846451036537e-06, 0, 2.37112269818839e-05, -2.95667440313906e-05, -1.7640619420681e-05, 1.57125839817404e-06, 78124.9999351346, 0, -7.33873643597509e-06, 2.82139619013957e-05, -2.31454799384345e-05, -1.61561212589316e-05, -2.45685276416822e-05, -4.5759434690292e-05, -6.40974707436448e-05, -4.44595481247789e-05, 0, -2.40285656592191e-05, 0, -3472.2222086346, 0, -5.22565774701644e-05, 2.5567352804186e-06, 0, -1.18139574870382e-06, -95486.1110595837, 0, -4.24304239051045e-05, 1.88028555323213e-05, 0, -1.15693694108279e-05, 1.70528036537666e-05, 5.37607494769778e-05, 3.47876261720555e-05, -295138.888791751, 0, -130208.333198589, -9.70367739022119e-06, 0, 1.35660947972351e-05, -7.37686162002351e-05, -3.35514797925725e-05, 0.000176536947487447, -7.98331318286823e-07, 4.35212018320211e-05, 2.8607145169125e-05, 0.000181508667259795, 1.06486776628862e-05, 0.00014082819961522, 0.000477502733532654, -7.77566071984649e-06, 0.000302328827851702, 0, 2.09729888940194e-05, 0, -1.19284763606984e-05, 0, 6.5981697272467e-05, 1.98163000348104e-06, 0, -2.85381172863865e-07, 63657.4073066279, 0, 0.000128510168497076, -4.36457016037657e-06, 0, -2.31521907116265e-06, -1.63171125229117e-05, -3.50105214365122e-05, -6.95582224624906e-05, 520833.333215994, 0, 0.000103779025074619, -3.87651218828881e-05, 0, -4.9106270859829e-05, -0.000106247525676369, -0.000121995183524972, -0.000374253476809139, 3.5473614332276e-05, -4.33858280291607e-05, -7.35633315438289e-06, 491898.14798164, 0, 7.66611623095768e-05, 9.06828119007265e-05, -7.79082549098254e-06, -0.00037846053892818, -7.18750348907272e-05, -3.83219726634313e-05, -8.43923929968733e-05, 0.000220787283553764, -0.000811170689965719, -0.000251772143308481, -0.00131228001655313, -4.54644521262616e-05, -0.000586294001172842, 0, -9.04925709972504e-06, 0, 7.38078502961171e-06, 0, -3.7976064138056e-05, 1.77483736747453e-06, 0, -5.95802004582214e-07, 3.33767599655291e-05, 0, -7.90443834274512e-05, 1.7490498826945e-05, 0, -4.99174774074293e-06, 1.08697386221849e-05, 2.33295170733713e-06, 8.5085606985362e-06, -347222.222091631, 0, -1.2780942381289e-05, 2.22731300069501e-05, 0, 4.74091011021638e-05, 4.60405622097141e-05, 5.77432804579287e-05, 0.000202486922998186, 4.46576786315909e-05, -6.10457067827666e-06, 0.000106159225773931, -868055.555411497, 0, 0.000187005589919332, 9.90187808453411e-05, 0, -3.2404391875251e-06, 0.000227263172542398, 2.08410942189547e-05, 0.000205526672013511, -6.11959990784793e-05, -0.000128401620644824, -7.43641569658846e-05, 3.48109249130757e-05, 0.00126572588124703, 6.80249688872793e-05, 0.000248968408640851, -0.000193714533495397, 0.000295153207503138, 9.86053916181261e-05, 0.000672046275982962, 0.00154904651507595, 0.000152121879572223, -0.000388043839383058, -0.0010146805009528, -0.00084159938379753, 0, 1.78094459204303e-06, 0, -1.8047018602114e-06, 0, 9.527208703325e-06, 3.4404208675364e-07, 0, -4.33090740644343e-07, -8.53135122656942e-06, 0, 2.72816344918453e-05, 4.73255561649235e-06, 0, -1.35934545751733e-06, -2.58309024899617e-06, -4.75008495149451e-06, -8.76208794331971e-06, -3.62282738836007e-05, 0, 1.94908092138458e-05, -1.47592574479096e-05, 0, -9.32546549511299e-06, 3.61635869644079e-07, -2.42971043557414e-05, -2.57799376881289e-05, -5.27635451812949e-05, -2.67380510666462e-05, -1.15178687315805e-05, 578703.703638439, 0, -8.64761601203663e-05, 6.77911606616595e-06, 0, 2.89763388770661e-06, 6.28884561402525e-06, -5.80048880678797e-05, -0.000120389965513951, 4.48885007538859e-05, 1.77817748469415e-05, -8.77869444096323e-05, 7.49892853256719e-05, -0.000299307557157143, -2.84122512638955e-05, -9.6264767179183e-05, 0, -4.29295703894586e-05, -6.5203690115433e-05, 4.1492884482167e-05, 0.000151943376529134, 9.74654406677069e-06, 0.000110497092236955, -8.16158984809357e-05, 0.000287710222779071, -0.000158357531471165, -0.000223339220531814, 0.000469765728778591, 0.000167940175985878, 0.000276514192953055, -0.000377501977894657, 3.5826264804639e-05, 0.000181764860405814, -7.65462232865132e-05, -0.00102582869029273, -0.000628685436154252}, 
                                                     {0, 0, 0, 0, 0, 2.6871119104784e-08, 0, 4.00000000521878, 0, 0, 0, -3.49071771844161e-07, 0, -83.3333332360808, 0, -83.3333337032451, -2.30384578128347e-08, 0, -2.879586182308e-08, 1.29963154583399e-07, 0, 1.85342498236817e-06, 0, 583.333331928585, 0, 583.333337043961, -1.76090600558813e-07, 0, 9.68208171514734e-08, 1736.11111199284, 0, 8.74891190046428e-07, 1.12326897603049e-06, -1.65452049878253e-06, -1.39850340257026e-06, 0, -7.37223778489446e-06, 0, -1666.66665845126, 0, -1666.66669149072, -2.38199511449924e-06, 0, -1.01995525217405e-06, -12152.777767292, 0, -12152.7777827496, 4.70162313316639e-07, 0, -2.13188900277043e-06, 1.16542402630475e-05, 1.85551868212378e-05, 7.29427718742485e-06, 1.84297645737316e-05, -3.52497848740168e-06, -4.55833135616564e-07, 0, 1.49107773315749e-05, 0, 1666.66665086467, 0, 1666.66673923098, -9.75659743745539e-07, 0, -2.64944768213165e-06, 34722.2221720144, 0, 34722.2223581242, -5.4939387303461e-06, 0, -3.19333090549262e-05, 2.3596940831353e-05, -4.48237468326057e-05, -1.14342759612267e-05, 85069.4444538742, 0, 5.65680840547005e-06, -8.19942263853995e-05, 1.63840411512823e-05, -0.000107548434055641, -8.0607001501256e-05, 7.12801619980769e-05, 4.12516890980951e-05, -3.81732661857277e-05, 0, -2.298900429966e-05, 0, 2.99834025748918e-05, 0, -0.00013299863729242, -6.29107048965884e-06, 0, 1.08689942055678e-06, -34722.2220996029, 0, -34722.2225178129, -2.45432621147606e-05, 0, 1.27970914348115e-05, 1.57271053543429e-05, 5.63867987206746e-05, -2.29896319194644e-05, -243055.555344452, 0, -243055.555672934, 2.19834267827825e-05, 0, -2.06093354882037e-05, 0.000154211661656246, 0.000113816699583555, 0.000122966348664826, 0.000296149339568618, -0.000113036611065366, 6.04079490921843e-05, -2.92292882759432e-06, -3.50535390937285e-05, 4.9249189133176e-05, -0.000455983053345953, 0.000656774243256965, -0.000290942108364356, 0, 1.90723082978978e-05, 0, -2.85788843335772e-05, 0, 0.000118026654570555, 9.96689407433956e-08, 0, -2.90297036316312e-08, -7.11323477901492e-05, 0, 0.00034919956489234, 7.58515674907165e-06, 0, 7.47487809485456e-06, -3.12676498716023e-05, -9.47068209486016e-05, 5.86713465462636e-05, 243055.555213258, 0, 243055.555552111, 5.3654740608293e-05, 0, 7.67413598129853e-05, 8.54388947529722e-05, 2.60649733069971e-05, -3.45984907793951e-05, -0.000503909421004667, 0.000138762663572782, -0.000126150037699696, 694444.44411778, 0, -0.000104679851064389, -0.000310760495045452, -6.70551216273749e-05, -2.29031247579171e-05, -0.000366426994427983, 0.000173492716764547, -0.000170792978335801, -0.00054047086666751, -0.000475863415874673, 0.000479403498949585, 0.000730799366726051, 0.000135100088717308, -0.000109967970019068, 0, -1.00756245407822e-05, 0, 1.28311412775843e-05, 0, -6.07453395185579e-05, -4.24829391929254e-06, 0, 6.69910043051848e-07, 7.67876351890526e-05, 0, -0.000276684564603242, -2.29199262715286e-05, 0, 5.81936155893732e-06, 2.13872799233495e-05, 7.88438728303677e-05, 1.40039356833053e-05, 0.000165151197896837, 0, -0.000248279214858022, -2.99448967132667e-05, 0, -5.64827614112918e-05, -3.16653372943988e-05, 4.78903445755673e-05, 0.000120162640070159, 0.000267001547008412, -3.1133675172114e-05, -1.20504145613049e-05, -694444.44404008, 0, -694444.444324876, -0.000121303651619021, 0, 3.51808894532233e-06, -0.000221824614929687, 9.76792125844442e-05, 0.000305894601381758, 0.000480312721100615, 1.64055614528929e-05, 0.000365870680512812, 0.000106096914244563, 3.2123038680775e-05, -0.000230261549690789, 1.05072250329925e-05, 0.000226040820666119, -8.43945230725314e-05, 0.000587758703588403, -0.000946423175658935, -0.000111374397922239, 0.00148509372045107, 0.000217138775423818, 0.00164481825187267, 0.00202337127476358, 0, 2.11480865916367e-06, 0, -2.45009809272319e-06, 0, 1.27035334859173e-05, -5.49144700459981e-08, 0, 6.09856098901985e-07, -1.89121408230991e-05, 0, 7.29841463253957e-05, -5.84522490999206e-06, 0, 1.87956037638821e-06, -6.00978077055989e-06, -1.59618492054358e-05, 5.48741603800623e-06, -4.3202690232048e-05, 0, 0.000106314854396319, 1.9923520417124e-05, 0, 1.35989745236125e-05, -1.45317499390505e-05, -9.357149843967e-06, -7.37253712913621e-05, -2.20355775661878e-05, 4.12139910087993e-05, -2.02585062177604e-05, -0.000107328741879259, 0, 4.53587894891564e-05, -5.99940545506835e-06, 0, -1.70644139078622e-07, -3.24351328030014e-05, 8.04586539747122e-06, -4.00680009568058e-05, -0.000203709535807496, 2.14294718393304e-05, 1.4446397675262e-05, -0.000161676036257984, -0.00023779736522128, 0.000112262596656796, 694444.444333297, 0, 5.99123564803265e-05, -3.33073910525056e-06, -8.38527291435547e-05, -0.000369712279405454, -0.00011257972311514, -0.0001122332709563, -3.27261132801922e-05, -0.000559955745536424, 0.00010912433801066, -0.000791539090497322, -0.0015938235652054, 1.26050601989919e-05, -0.000426416355764989, -0.000200407995087955, 6.94965585796923e-05, -0.000783550902385802, -0.000155712944663051, 0.00041418685881614, 0.000801299416482649}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 4.16666668847255, 0, 0, 0, -6.31946643009032e-08, 0, -76.3888888740495, 0, -95.1388894648657, 1.98214484957747e-08, 0, 2.63863469580804e-08, 3.02610672442946e-08, 0, 7.82459083691239e-07, 0, 416.666665254109, 0, 781.25000657937, 2.14262561507365e-07, 0, -1.15023469946092e-07, 1744.21296539703, 0, -8.29915595227308e-07, -2.12194616904834e-06, -1.86038391002105e-06, 4.81995585849e-07, 0, -2.53370535613951e-06, 0, -694.444437713105, 0, -2951.3889185699, 1.95510318796946e-06, 0, 1.0009406687242e-06, -9513.88888560221, 0, -14322.9167073467, -5.16238226242028e-07, 0, 2.28693596603103e-06, -5.57960722158939e-06, 1.11646827888153e-05, -6.4754575047733e-06, 2.38617914562457e-06, 8.45155467169985e-06, 2.39006897503838e-05, 0, 7.97815197983984e-06, 0, -2.43045808382757e-05, 0, 5208.3334212274, 1.66385248464729e-06, 0, 2.61196357252069e-06, 15856.4814435451, 0, 54108.7964068408, 5.31518992979259e-06, 0, 2.97345677688533e-05, -4.16660903311743e-05, -6.37279561548533e-05, 1.95594524333919e-05, 78124.9999936126, 0, -2.00605877770377e-06, 4.87054247624724e-05, -5.19144955249585e-05, -4.65514329234541e-05, -8.60383560667718e-05, -5.1067231244439e-05, -0.000101081357580658, -7.92984209762766e-05, 0, -9.86788070611153e-06, 0, 3.46626128153428e-05, 0, -3472.22236272513, 5.15595205529277e-06, 0, -6.31435710091822e-07, 6.75865169672624e-05, 0, -95486.1114604451, 1.9983107904205e-05, 0, -1.02865369347819e-05, 3.0193722255165e-05, 0.000169476846262097, -4.22681529750002e-06, -130208.333280869, 0, -295138.888987016, -2.8065824122224e-05, 0, 2.04552816499256e-05, -0.000117671235361021, 2.63031348433193e-05, 0.000404198935919284, 0.000224522132985765, 6.19330761892358e-06, 2.59561766371183e-05, 0.000199996473479946, 6.46272841884444e-05, 0.000345817444925051, 0.000628692206783211, 0.00019085096182849, 0.000371689904891666, 0, 9.37871407216583e-06, 0, -3.35910482962493e-05, 0, 0.000152388371246805, 1.29223473843403e-06, 0, 3.0695363319176e-07, -0.000137412581642556, 0, 63657.4079415411, -5.39342274236567e-06, 0, -9.76420667604589e-06, -2.59282860234983e-05, -0.000166023768008824, -2.13406442342764e-05, -0.000141325594726794, 0, 520833.333979809, -5.08466960634787e-05, 0, -7.99270083157949e-05, -0.000113288415715818, -0.000250474668813117, -0.00082401238543781, -0.000242992914403052, 5.60580513655668e-05, 3.85515091231649e-05, 491898.148169288, 0, 9.88961654893347e-05, 0.000196132998298951, -0.000152743500023632, -0.00057821417186447, -0.000549906365552275, -8.0924195598919e-05, -0.000355691696019896, 0.000217607672821689, -0.00164808305927924, -0.000355478092398653, -0.0017025136884119, -0.000289165441239764, -0.00105666543419109, 0, -3.88931179548707e-06, 0, 2.01417632595807e-05, 0, -8.44834895481059e-05, 3.36071194500879e-06, 0, -4.9185155993661e-07, 4.93501122644505e-05, 0, -0.000332628276714927, 1.88944721731336e-05, 0, -5.04711191797357e-06, 1.65077796776378e-05, 8.10054426785795e-05, -3.4305606021822e-05, 0.000178453303156336, 0, -347222.222731813, 2.6181071825016e-05, 0, 4.72932207625181e-05, 6.91203118855164e-05, 0.000137273042896222, 0.000562195609522188, 0.000256819080990112, -0.000128050261110133, 3.26428049873699e-05, 0.00013153891551825, 0, -868055.555674278, 0.000106626098154087, 0, -1.28351662989815e-06, 0.00021627244404116, 0.000225913333709487, 0.000404230406012404, 0.00043456814403237, -0.000157719283479452, 2.31320496124074e-05, 0.000220861713245529, 0.00216101982673865, -8.00057591359409e-05, 0.000157804643534039, -8.57980387986477e-05, 0.000840845277015388, 4.10467154590606e-05, 0.00115486666722277, 0.00191982525102264, 0.000963461936414379, -0.000543272834674593, -0.000856626924741864, -0.0012107792337044, 0, 7.60836850247595e-07, 0, -4.9717591735291e-06, 0, 1.95291935079249e-05, 2.34597556522462e-07, 0, -6.04867675425872e-07, -1.26145005639039e-05, 0, 9.3325584350103e-05, 5.78143735596696e-06, 0, -1.87955110348968e-06, -4.52543408530146e-06, -2.14513371997704e-05, -5.23510331640498e-07, -5.02134218924937e-05, 0, 0.000187401419687237, -1.79415405323936e-05, 0, -1.38136876351832e-05, 2.51384460119186e-06, -4.80278867626369e-05, -0.000142900517458976, -0.000119899509683564, 1.13205926974663e-05, 1.30578882628139e-05, -0.0001064593386969, 0, 578703.703783937, 3.75776363475869e-06, 0, -2.69642611068746e-06, 3.20204016915794e-05, -0.000137590707743635, -0.000265282500042602, -0.00011241883838472, 5.7726005219062e-05, -0.000126129705477499, 3.03798787536239e-05, -0.000611431522423076, 4.06556657680507e-06, -6.93034866393085e-05, 0, -5.82243375978255e-05, 1.98650685296877e-05, -4.67005078817412e-05, 0.000221666489138079, -0.000272015391677665, 0.000103757025938654, -0.000217436008197865, 0.000328143553141131, -0.000248280620046914, -0.000326641605865956, 0.000135811568943567, 0.000195051544402404, 0.000355017659819572, -0.000286165298014684, -1.22285928247011e-05, -1.25913769362477e-05, -5.55343928744767e-05, -0.00134629203767232, -0.000740706423902075}, 
                                                     {0, 0, 0, 0, 0, 1.01166350947908e-08, 0, 4.76190478613878, 0, 0, 0, -1.31152742703347e-07, 0, -71.4285717380428, 0, -116.666667281776, -1.44342285350871e-08, 0, -1.67129980646551e-08, 1.39154351150974e-08, 0, 5.1331282807076e-07, 0, 238.095239437818, 0, 1074.07407892086, -1.34617212462099e-07, 0, 8.66338819513484e-08, 1750.00000653403, 0, 5.39825382133232e-07, 9.88461442256486e-07, 9.6057623806273e-08, -4.04479617763065e-07, 0, -1.93658598076878e-06, 0, -1.80584030509415e-06, 0, -4861.11113525544, -1.20273424269232e-06, 0, -6.49832575213205e-07, -5833.3333594676, 0, -16111.1111517891, 2.89239543357959e-07, 0, -1.58887267086938e-06, 5.48930928814457e-06, 2.82920267595929e-06, 9.85658006431064e-07, -8.08143629870648e-08, -1.00355172031791e-06, -5.69930916796477e-06, 0, 2.55443578635105e-06, 0, 2.87185833156794e-06, 0, 11574.074134114, -1.09472756823507e-06, 0, -1.70876869629449e-06, 5.46899605184789e-05, 0, 72916.6668707726, -3.24204114483195e-06, 0, -1.87393684186579e-05, 1.5731676105325e-05, -3.01952921379137e-06, 5.61784374493038e-06, 53703.703880286, 0, -2.62156110865379e-07, -4.82884030588189e-06, -4.83165735134828e-06, -2.17054969445423e-06, -2.51482124037326e-05, 3.56900731573963e-05, 4.41223207012381e-05, 1.00169052938083e-05, 0, -4.01237313030666e-06, 0, 4.23108058242159e-06, 0, -13888.8889847946, -3.59878111076343e-06, 0, 2.23980639711055e-07, -5.30342545093128e-05, 0, -173611.111499357, -1.18741888495492e-05, 0, 5.91651441615344e-06, 8.52905488471135e-06, -1.4548506047465e-05, -4.07686373925474e-05, -0.000225237548690679, 0, -243055.55610223, 2.13072173523296e-05, 0, -1.32893868718678e-05, -4.40577214885351e-06, 0.00011739050262314, -0.000105000936539991, 9.34228033656278e-05, -6.72687522070003e-05, -1.85506701992409e-05, -0.000247067936840734, 4.23823333163665e-05, -6.54248147261651e-06, -0.000325952872046184, 0.000178733290222478, -0.000205898012976579, 0, 2.46984889867209e-06, 0, -8.16537486238712e-06, 0, 6613.75668972824, -9.44807298892594e-07, 0, -3.0973289655828e-07, 7.05575216251008e-05, 0, 208333.333737916, 3.62374855874858e-06, 0, 7.3141762661802e-06, -2.11136812674407e-05, -6.87340121942877e-06, 6.70556993156939e-05, 0.000172501217188801, 0, 578703.704380428, 3.35881011799636e-05, 0, 5.47386779212138e-05, 0.000155858964766224, -6.05239541779583e-05, 0.000204459583631037, -0.000192186838910752, 9.73176530836705e-05, 3.78499359320044e-05, 0.00053158946530652, 0, -6.30459894581843e-05, 0.000250834013489978, -0.000202440925851241, 0.000523442144284706, -0.000185479900517761, 2.65256347477473e-05, -1.70744945085052e-05, -0.000291094207739201, 0.000161947485522662, 0.000437228329199837, 0.000902707552806382, 0.000168018309208489, 0.000202549911705767, 0, -1.25669592666307e-06, 0, 4.53412669957884e-06, 0, -3.3236123374128e-05, -2.13324091051734e-06, 0, 2.28655065417417e-07, -1.12320247003833e-05, 0, -99206.349462381, -1.1721584956648e-05, 0, 3.15110231997244e-06, 1.42857454860745e-05, 1.83358041220783e-05, -1.17156700760996e-05, -0.000108759530543055, 0, -694444.445043819, -1.60239300143049e-05, 0, -2.73366249081764e-05, -8.57861178151278e-05, 8.00142035430272e-05, -5.06221722263755e-05, 7.07396291806831e-05, -3.07561634931925e-05, -8.06115690799629e-05, -0.000257831506296792, 0, -0.000642708557202889, -6.56965866734046e-05, 0, -1.07979821877174e-06, -0.000458225154345209, 0.000195390402916725, -0.000260145538532363, 0.000227475494022618, 2.86724807650545e-05, 0.000107984116193053, 4.08604162735967e-05, -0.000305197627099425, -0.000229565681777677, -0.000577545784333706, 0.000270835269297109, -4.90569631600533e-05, -6.74862915172336e-05, -0.000479845076517359, -0.000688801645241462, 0.000538368439291422, 0.000170026589349616, 0.000582682260016763, 0.000975946081997093, 0, 2.29875861149224e-07, 0, -1.01635806289436e-06, 0, 6.05777359480112e-06, -1.64968735292411e-07, 0, 4.06817210164631e-07, 8.23655312371317e-07, 0, 5.8263118912129e-05, -3.57749886387378e-06, 0, 1.29835260525658e-06, -3.53827838725838e-06, -5.20460198217643e-06, 8.96862293583902e-06, 2.29079165761846e-05, 0, 330687.830869184, 1.11800136765631e-05, 0, 9.45431828167454e-06, 3.80730952800879e-06, -1.84050357671893e-05, -1.37126283587374e-05, 1.69354876828961e-05, 2.77377088670074e-05, 1.57051431674546e-05, 7.3266395310409e-05, 0, 0.00028118129536096, -1.56728719454331e-06, 0, 3.14226866826638e-06, 7.37031160956813e-05, -4.60090871028647e-05, 0.000115777968604884, -9.5699471995088e-05, 4.51283829739573e-06, 5.95290021284812e-05, -9.82517905593418e-05, -4.48574031930403e-05, 0.000123053218743156, 0.000187017100503801, 0, 3.82360509275686e-05, 0.000211269618648615, -0.0001182989145213, 5.4700053263599e-05, -5.70173645400582e-05, -0.000110207454734921, -2.96109170823809e-05, -0.000186869584067228, -6.57610329802176e-05, -0.000176151774812772, -0.000807293723232686, 4.50987623502722e-05, -0.000148038658418789, 0.000192582917130687, -4.85376142807122e-05, -0.00039728255401805, 2.11335165555295e-06, 0.00049411790175161, 0.000608099629850915}, 
                                                     {0, 0, 0, 0, 0, -1.77708303792559e-08, 0, 6.24999999872186, 0, 0, 0, 2.10347187485645e-07, 0, -62.5000000297733, 0, -162.05357134762, 0, 0, 0, -9.02391808664112e-08, 0, -1.16968852815605e-06, 0, 3.48496110154146e-08, 0, 1628.4722212987, 2.60128912009432e-08, 0, -3.96465083075259e-08, 1620.53571521647, 0, -2.2776488547389e-07, 6.03907477669229e-07, 1.27636722435928e-06, 9.27343760822498e-07, 0, 4.33623987864575e-06, 0, 8.49430173188028e-07, 0, -8394.09721497106, 6.26556455620303e-07, 0, 2.68567940920283e-07, -6.42094403773484e-06, 0, -16284.7222300416, -7.61832653352585e-08, 0, 6.86349975878288e-07, -6.91818644814842e-06, -8.40161452979628e-06, -4.20821875183538e-06, -1.43036006950008e-05, -6.18918644200907e-06, -6.27112163293274e-06, 0, -9.26369531705887e-06, 0, -6.56948253145158e-06, 0, 24305.5555319977, -2.64999256846486e-08, 0, 7.08400486867963e-07, 1.25858084827246e-05, 0, 83940.9722200877, 1.14772257946021e-06, 0, 7.60370888399018e-06, 4.92966784172218e-06, 2.17528934807825e-05, 9.35499333406074e-06, 3.29139713669802e-05, 0, 5.63725312944098e-07, 6.51239694419416e-05, 2.44929158572183e-05, 7.23514654182894e-05, 4.8811225828109e-05, 1.12125966249048e-05, 1.24152091907825e-05, 6.62406290957109e-05, 0, 1.36558105744438e-05, 0, 1.25039878792499e-05, 0, -39930.5555078731, 2.5422296342439e-06, 0, -4.65725948351892e-08, -9.59891289883018e-06, 0, -243055.555543704, 4.97459363267835e-06, 0, -2.30196625616079e-06, -1.80689047040151e-05, -2.32994452008269e-05, -4.80556456750562e-06, -9.79428217124067e-05, 0, -7.41142573211246e-05, -9.73028560694691e-06, 0, 5.43769312876157e-06, -0.00013559433465274, -8.52635944264548e-05, -0.000120557611219649, -0.000123058116962928, -2.75090466455182e-05, -5.70402267117286e-05, -0.000159154308929909, -8.61689734649126e-05, -9.03445761515297e-05, -7.96401850486699e-05, -0.000336158511095113, -0.000116930927680688, 0, -1.17541896074501e-05, 0, -1.40177550496663e-05, 0, 34722.2221736192, -6.3517690272461e-07, 0, 1.53809272802652e-07, -2.44706985059734e-05, 0, 399305.555532539, -2.09657408092241e-06, 0, -3.25602857280857e-06, 1.87339263903508e-05, 2.43539776723876e-05, -3.76572627168666e-06, 9.09120486910081e-05, 0, 0.000156161319737625, -1.43851257479344e-05, 0, -2.36253224100342e-05, 9.6579247476853e-05, 5.52638063569312e-05, 0.000104468582615677, 0.000184221497901432, 2.37539041901018e-05, 6.09048750168111e-05, 0.000211482441456285, 0, 2.57179267592533e-05, 0.000334070340142565, 0.000169107054013922, 0.000255758174296197, 0.000219693477205745, 0.000102178629553823, 0.000131391725031741, 0.000429177360927853, 0.000393786816460873, 0.000128974105690063, 0.000249067248615488, 0.000318642144167053, 0.000427817040053971, 0, 5.8276256227209e-06, 0, 8.6982978955001e-06, 0, -12400.7936234931, 1.5239059553328e-06, 0, -6.20204599252356e-08, 1.57665141234925e-05, 0, -347222.222185443, 5.22469827678523e-06, 0, -1.31532354737806e-06, -8.78128035771684e-06, -1.36486117881888e-05, -9.52455145240837e-06, -1.43524200147994e-05, 0, -6.57892582094246e-05, 6.36025704976253e-06, 0, 1.03846185672064e-05, -5.9108646939505e-05, -4.12997389280726e-05, -8.45603842675598e-05, -9.44372982600377e-05, -2.46024734421755e-05, -2.1466655675228e-05, -0.000158380376272872, 0, -0.000174735232726317, 2.65679062296718e-05, 0, 1.34194941352654e-06, -0.000159861886634441, -0.000124726988842143, -0.000277899228920638, -0.000199834992703059, -9.63635273580683e-05, -0.000128073500381411, -0.000245601424020932, -0.000254165822940937, -0.000151378757779025, -0.000320446039857532, -0.00020960279926501, -9.37552281165949e-05, -0.000595837620831463, -0.000155810767268105, -0.000490498718096934, -0.00076853539891899, -0.000494581700724617, -0.000739188323353667, -0.00112078371115264, 0, -1.21749041863153e-06, 0, -2.16321884760885e-06, 0, -6.47508285118594e-06, -1.02627668325839e-07, 0, -1.74545284333913e-07, -5.63919186467991e-06, 0, 124007.936496465, 1.26954601574036e-06, 0, -5.78382227817136e-07, 1.52728006947705e-06, 2.31690611767817e-06, -2.8801377243687e-08, -4.6691541669015e-06, 0, 1.09891000500884e-05, -4.53417120527718e-06, 0, -4.1206353916881e-06, 2.13242120601582e-05, 6.72105241331295e-06, 2.62457615516673e-05, 8.71880283666204e-06, -2.40388606332766e-06, 8.9049389887153e-06, 2.49197247907911e-05, 0, 4.94192227043758e-05, 4.4790673103175e-07, 0, -1.74567840359182e-06, 7.32451529798526e-05, 2.59499856979616e-05, 6.51185328425723e-05, 7.21498418064e-05, 2.32884731950403e-05, -4.90629017246581e-07, 0.000107879305485761, 0.000135909142065282, 1.88614121844549e-05, 6.51962316630634e-05, 0, -1.60565912642075e-05, 0.000140269804090994, 7.73353428592671e-05, 0.000213434520165661, 5.84193743731136e-05, 8.55640674921249e-05, 4.62328233104268e-05, 0.000316774657600904, 0.000156660690400763, 0.000396128625709007, 0.000535841149202261, 0.000142103682981583, 0.000268773377662699, 0.000325033494740178, 0.000203532431176119, 0.000436528608331088, 0.000299654924701176, 0.000208589591003374, 9.78671447244979e-05}, 
                                                     {0, 0, 0, 0, 0, 1.96533879024447e-08, 0, 11.1111111190617, 0, 0, 0, -2.12220810145164e-07, 0, 0, 0, -301.984127184224, 0, 0, 0, 0, 0, 1.18690209794132e-06, 0, -2.14187646385911e-07, 0, 3255.18077792665, 0, 0, 1.00280139736825e-08, 1.4465026120303e-07, 0, 5.76458513980657e-08, 1.62518617293612e-07, 3.88353818070054e-08, -1.28487030689501e-07, 0, -4.0743397694844e-06, 0, 1.02417680874237e-06, 0, -18541.6666762911, -2.30381659853307e-07, 0, -6.40874480287561e-08, 4.63622088041329e-06, 0, -3.38129632121347e-06, -1.02633792008439e-08, 0, -1.69945454292292e-07, -7.89539157018602e-08, -9.7490399563063e-08, 9.65842070821397e-07, -8.37239936346339e-07, -4.09235574362626e-07, 0, 0, 8.59804123422937e-06, 0, -1.37150884450188e-06, 0, 61863.4259522707, 2.67687135213893e-07, 0, -1.68518393857276e-07, -2.21841213140918e-05, 0, 2.56941373998335e-05, -1.33109375534272e-07, 0, -1.79379759856618e-06, 3.57965599688355e-06, 1.37748189445799e-06, -2.95491121096293e-06, -1.15727371030035e-05, 0, -2.02684499494926e-07, 2.92068880630633e-06, 4.40131880854567e-06, 2.55153366779747e-06, -5.50833487556679e-06, 1.15250624434598e-06, -4.86088454954033e-07, 6.95221141128698e-06, 0, -1.17499117722899e-05, 0, 6.55816006199654e-07, 0, -125000.000043855, -1.12245785103429e-06, 0, 0, 4.60092494875155e-05, 0, -5.89251959578696e-05, -1.46015153292518e-06, 0, 5.48161068531507e-07, -2.86873708483137e-06, -6.68154303925558e-06, 3.35713639857241e-06, 7.75219150011126e-05, 0, -6.40884017121398e-06, 2.41035242553026e-06, 0, -1.26960006506707e-06, -1.03758114670022e-05, -4.72945924390574e-06, -1.81159482204734e-05, 1.62298548345249e-05, 5.96496690077422e-06, 1.10403411893448e-05, -2.84412897557913e-05, -2.95808508401511e-06, 1.80143873944887e-05, -6.39857299430096e-05, -4.94759337938374e-06, -2.31046147981702e-05, 0, 9.74992887124933e-06, 0, 1.00432817137292e-06, 0, 150462.963004941, 6.67154705685764e-07, 0, -3.87682770529194e-08, -4.36826895725265e-05, 0, 7.04699231837306e-05, 9.81446984754061e-07, 0, 8.08089235367206e-07, 1.24804894464933e-06, 7.27188577043868e-06, -1.93671840340151e-06, -0.000133231769373079, 0, 3.09255116151933e-05, 3.70285930591909e-06, 0, 5.85733265627202e-06, 2.87153618768948e-05, 1.63856272818376e-05, 3.46829113585289e-05, -2.87141484516292e-05, -1.4452797389279e-05, -1.19599858139434e-05, -9.33720176062278e-05, 0, -5.98283225066165e-06, 3.42177195399202e-05, 9.36853533688406e-06, 4.98486728166296e-05, -5.92742900372766e-05, -3.61084773367228e-06, -3.38110607247704e-05, 4.43735992902119e-05, 5.21978151981795e-05, 2.99362714933216e-05, 0.000157350765655848, 7.91644255496266e-05, 3.76820285739499e-05, 0, -4.57012180203726e-06, 0, -1.4044178785596e-06, 0, -99206.3492285322, -6.71735118129955e-07, 0, 0, 2.41573936890017e-05, 0, -4.69846106044095e-05, -1.5805417144715e-06, 0, 3.27392234219799e-07, -1.74470233809648e-07, -4.26039611068424e-06, 3.93789533353809e-06, 9.58269126151376e-05, 0, -4.6245899720635e-05, -1.5544950879547e-06, 0, -2.37338450646409e-06, -1.77699714486663e-05, -6.69945508917135e-06, -1.87280326448652e-05, 1.25573156984219e-05, 1.70200481721555e-05, 4.24078768358165e-06, 0.000131131702853012, 0, 3.13382966006176e-05, -6.28146652871728e-06, 0, -5.58093309486036e-07, -5.635549960817e-05, -1.37660343933591e-05, -2.07991889788195e-05, 6.16429548384652e-05, 2.08340495016761e-05, 3.2408672980863e-05, -6.30717352124178e-05, -6.11937590514075e-05, 3.65805822680212e-06, -4.4077596042295e-05, 1.08781037846326e-05, 4.47910383153032e-05, -0.00010065346800969, -0.000146552639070673, -0.000121214285313687, 3.38195245564987e-05, -6.50430397943275e-06, 5.57010394995476e-05, 2.04555798936362e-05, 0, 9.15881348192925e-07, 0, 4.74370322857332e-07, 0, 27557.319228981, 1.04778111846686e-07, 0, 4.27351624935057e-08, -4.9438147972876e-06, 0, 1.1793932466352e-05, -1.77469362666297e-07, 0, 1.48850483518653e-07, -2.17908229702815e-08, 1.19931880142049e-06, -3.82772500054183e-07, -2.66767897241796e-05, 0, 1.65861532523729e-05, 1.13385264995249e-06, 0, 1.03025107796988e-06, 2.5254615243966e-06, 1.80954013644995e-06, 4.22839056768275e-06, 8.23620396681902e-07, -3.65020510840914e-06, -1.79205532515896e-06, -5.03530395793265e-05, 0, 4.09685093455578e-06, -6.94702749378104e-08, 0, 4.99506654903117e-07, 1.2788199420213e-05, 6.74455862217016e-06, 1.06655473497006e-05, -2.36883900130171e-05, -9.72446055889006e-06, 3.43190691061394e-08, 1.19818004728114e-05, 7.79932069744011e-06, 1.43539869568069e-06, -3.58887216414151e-05, 0, 3.88833732966343e-06, 1.64702902401513e-05, 6.90148758796344e-07, -5.45244168820088e-06, -2.91090733003593e-05, -1.18323414698294e-05, -2.1251143684342e-05, 2.49763053453221e-05, 3.91304303166814e-05, 3.91693479419443e-06, -7.00708192197394e-05, 3.47287941300808e-06, -2.54043134794172e-05, 4.8090547321602e-05, 3.05982136598766e-05, -4.94242764454152e-05, 6.62561626557355e-05, 8.32675571211159e-05, 8.14759742917166e-05}, 
                                                     {0, 0, 11.1111111028192, 0, 0, -313.095237872483, 0, -11.111111108565, -11.1111111118181, 0, 0, 3557.16490060374, 0, -9.32041666912672e-08, 0, 301.984126947944, 301.984126998933, 0, 3.58241655278878e-08, 0, 0, -21796.8474288645, 0, 9.25166336318855e-07, 0, -3255.18077565786, -3255.18077603911, 0, -2.22756516021661e-07, 1.16993090769115e-06, 0, -4.99361135772882e-07, -5.13714488650241e-08, -9.07951800641564e-08, -7.31673164987385e-08, 0, 80405.0925447376, 0, -3.99906237943391e-06, 0, 18541.6666640288, 18541.6666669195, 0, 1.02361579575061e-06, -1.1140805611337e-05, 0, -5.36150562796964e-06, 6.07405304862491e-06, 0, 3.00963114814707e-06, 3.733852567554e-07, 3.55674192877527e-07, 1.70931496030135e-08, 1.1872216731209e-06, 2.20568479750426e-07, 1.03427631741862e-06, 0, -186863.425821759, 0, 9.57678001676993e-06, 0, -61863.4259148859, -61863.4259251485, 0, -2.11139488333792e-06, 4.22142367215588e-05, 0, 1.52298091563007e-05, -2.20792734186084e-05, 0, -1.65274063671828e-06, -1.28777869330607e-06, -2.72330520495093e-06, -1.91994597330786e-06, 4.73296619868983e-05, 0, -2.68250485520957e-05, -4.60385295976586e-06, -3.43499255472541e-06, -2.96523845542021e-06, 1.20143734800506e-08, -1.77650574668108e-06, -2.00161643586517e-06, -7.92006926837146e-06, 0, 275462.962819441, 0, -1.37416428934629e-05, 0, 124999.999974505, 124999.999997145, 0, 4.31749391236003e-06, -8.13356558677839e-05, 0, -3.02259955780922e-05, 4.44485603915231e-05, 0, -9.32202746171975e-06, 6.12878860072203e-07, 1.39169823597557e-05, 3.50004944431249e-06, -0.000150299328855071, 0, -0.000100180171526309, 0.000111197394665741, 0, 3.38790009536186e-05, 1.19946529470666e-05, 7.15951523161815e-06, 1.98637618887602e-05, 2.29188634225392e-05, -4.94798986767706e-07, 1.80892217770277e-05, 1.6823497043649e-06, 6.27278843614807e-06, 1.64131252004153e-05, 3.91744954342192e-05, 1.10699180195557e-05, 3.31019179003156e-05, 0, -249669.312047954, 0, 1.17501923085842e-05, 0, -150462.962930906, -150462.962957617, 0, -6.17278337431155e-06, 8.75542280889467e-05, 0, 4.27702812227955e-05, -4.13439690691413e-05, 0, 1.28364638150557e-05, 5.45938631380098e-07, -2.21028314324294e-05, -6.09864235558986e-06, 0.000217830437482252, 0, 0.000116091548588313, -0.000200576851802322, 0, 8.04389276485684e-06, -1.37968345120141e-05, -1.25402594077938e-05, -2.13278682951156e-05, -2.72046399120366e-05, -4.75973244903555e-06, 5.89662373878193e-06, 0.000251420760451348, 0, -0.000138264957932966, -3.58051840518767e-05, 1.48924560419719e-06, -1.28003643006433e-05, -2.04316127445076e-05, 9.56282313722245e-06, -2.66637038184789e-05, -7.53868994895536e-05, -1.12512816547742e-05, -0.000102526016862482, -5.0308414707231e-05, -9.84437325592242e-05, 2.75014559712737e-05, 0, 126763.668372917, 0, -5.51754036481522e-06, 0, 99206.3491855152, 99206.3492027389, 0, 3.8215813410021e-06, -4.93251961699149e-05, 0, -3.43632796533356e-05, 1.60528453465495e-05, 0, -5.16007705715324e-06, -7.97511144073478e-07, 1.15967061721861e-05, 4.88937642537553e-06, -0.00015177697199554, 0, -7.50621109220889e-05, 0.000146319589165623, 0, -3.385712628804e-05, 7.64938406594623e-06, 6.84404278473622e-06, 2.07977340873173e-05, 2.11960143515149e-05, 4.53640817087407e-06, 1.48675289041009e-05, -0.000250534641386197, 0, -0.000197954160413894, 0.000208174674436608, 0, 5.2505193546767e-05, 2.89388872565277e-05, 1.62538051351141e-07, 2.16721698598979e-05, 5.37192349892884e-05, 3.54527128701781e-06, 3.98473467673272e-05, 6.77873659963824e-05, 8.7335020570235e-06, 3.85190303458635e-05, 1.63942130663013e-05, -6.04766059152953e-06, 1.73797229398891e-05, 0.000114929472211985, 4.70240819198271e-05, 3.60627268152095e-05, 7.55026288803482e-05, 0.000147949847177228, 5.75315872094879e-05, 5.05552023535635e-05, 0, -27557.3192123576, 0, 1.10307584503006e-06, 0, -27557.3192185445, -27557.3192234547, 0, -9.8592763207329e-07, 1.16572149179197e-05, 0, 1.14872666236921e-05, -1.601977378631e-06, 0, 8.59198294367217e-08, 2.89495521426291e-07, -4.14613984669126e-06, -9.68548178251786e-07, 4.1662914196423e-05, 0, 2.04591044912537e-05, -3.75194251480108e-05, 0, 8.26516582898414e-06, -1.66999777202232e-06, -1.77618862103282e-06, -6.05574722020102e-06, -1.2171997852236e-05, -3.89585753258852e-07, -7.66423092642894e-07, 8.72139800935114e-05, 0, 5.98147882277918e-05, -9.23280231490528e-05, 0, -2.34236610932506e-06, -8.72313050476918e-06, -5.66375439519386e-06, -3.48333321706883e-06, -4.70502789829163e-06, 2.51949237662356e-06, -1.90594854921869e-05, -2.08502999998455e-05, -2.39915428316335e-05, -2.70519376609928e-05, 0.000101341322180844, 0, -5.94413628963201e-05, -9.85980763820164e-06, 1.5271029934893e-06, -2.24325082573788e-06, -1.87997411285483e-05, 2.7078321714107e-06, -1.28875796502538e-05, -3.8813913105349e-05, -4.26155057267795e-05, -3.10395469646272e-05, -1.13619028725064e-05, -4.22814969359714e-05, -4.08516347218799e-05, -2.4039147408918e-05, -5.56971651885852e-05, -4.14392902958412e-05, -6.29517096942051e-05, -3.88973291603712e-05, -3.58510548904288e-05}, 
                                                     {0, 0, -56.2499999630764, 0, 0, 1577.23214186384, 0, 118.750000027956, 118.74999999275, 0, 0, -17797.7678464466, 0, -62.4999996376533, 0, -3204.0178579123, -3204.01785696781, 0, -62.5000000727726, -124.999999925228, 0, 108108.382874655, 0, -5.23868902963558e-06, 0, 34182.0436579134, 34182.0436499597, 0, -1.34198337222003e-08, 1620.53571074557, 0, 1620.53571495784, 1.85062335430732e-07, 3241.07142821159, -2.1413440019799e-06, 0, -394522.569229698, 0, 2.41731376783624e-05, 0, -192057.291695487, -192057.29166423, 0, -6.65033992688173e-07, 6.73804784678892e-05, 0, -16284.7222239497, -16284.7222565259, 0, 1.06353891358734e-05, -5.58828635085397e-06, -32569.4444470715, 1.89767373051357e-05, -9.66054894122454e-06, 9.61407781237449e-06, 1.65461883027795e-05, 0, 905121.527309188, 0, -5.7362061245754e-05, 0, 629687.500057031, 629687.499985235, 0, 7.77254499252451e-06, -0.000291331643378381, 0, 83940.9722808443, 83940.9724077159, 0, -0.000115332780855731, 2.25217355055035e-05, 167881.944463928, -6.02535939440155e-05, -0.000241819664647546, 0, 9.3658603418939e-05, 6.30109567077215e-05, 1.25070756951857e-05, 5.11345090729868e-05, -1.53394270954684e-05, -9.21562618463434e-05, -0.000135580468720216, 1.91941379245401e-05, 0, -1314236.11046399, 0, 7.87397286052393e-05, 0, -1244791.66671816, -1244791.66662589, 0, -3.19773474846105e-05, 0.000591515689623055, 0, -243055.5557056, -243055.55597846, 0, 0.000301809778444598, -3.67834705780904e-05, -486111.111183355, 0.000107589382286947, 0.000979126950229094, 0, 0.000354273610188938, -0.000759630018382113, 0, 0.000188194479205987, -0.000167701856803783, -0.000103938144116123, -0.00015391098960404, -0.000124511259507415, 0.000306592904231344, 0.000431797767939278, -0.000150356079304874, 0.000164945647125449, 0.000220016313417172, -0.000186553972386147, -0.000120555926573893, 0.000282851944121805, 0, 1170634.92008654, 0, -6.29530859367602e-05, 0, 1458333.33334336, 1458333.33327397, 0, 5.08545947645624e-05, -0.000643990441178398, 0, 399305.55569868, 399305.555980472, 0, -0.000319184376661297, 2.71586933568477e-05, 798611.111224842, -7.88168041197498e-05, -0.00157608436777098, 0, -0.000212075822073262, 0.00165227304390016, 0, -0.000736615985146727, 0.000239005238984087, 0.000196739999520405, 0.000195454443220668, 0.0002668829700726, -0.000461705715211449, -0.000743568042677158, -0.00141433925175343, 0, 0.000656085078785127, 0.000493535932517914, 4.45150768909142e-06, 0.000310632054566278, -4.36566380613827e-05, -0.000534898396784928, -0.000807324465097959, 0.000498327676529502, 0.000517497300919249, -0.000548704076572432, 0.000562085449913182, -0.00016132249730943, -0.000140605417676795, 0, -582837.30132734, 0, 2.72382017631013e-05, 0, -930059.523794482, -930059.523770783, 0, -3.675123784102e-05, 0.000361434983215514, 0, -347222.222266722, -347222.222408889, 0, 0.000145563161819968, -9.98467399823115e-06, -694444.444509999, 2.71098514932251e-05, 0.00116250228586213, 0, 3.92050733658273e-05, -0.00133099406470912, 0, 0.000727549382324707, -0.00016480410126578, -0.00013111238439034, -0.000151722256228079, -0.000250163663253734, 0.000311872713198603, 0.000504488045577912, 0.00165386543937428, 0, 0.000901037899482888, -0.00152647314692097, 0, 0.000196827589175007, -0.000442348315088075, -0.00016430836345638, -0.00027284968487675, -0.00035150606074486, 0.0005786167278376, 0.000989249723456606, -0.000618703669367365, -0.000504383036505882, 0.000883567970346997, -0.000415466026980089, 0.000275219804442145, 0.000496575379277087, -0.000797174529745873, -4.22866679129766e-07, -0.000576479965502746, -0.000894437529769835, 0.000385714477437956, 0.000344190590732415, 0.000283704927196252, 0, 124007.936455197, 0, -4.94727951759024e-06, 0, 248015.873008555, 248015.873007296, 0, 9.8511859811921e-06, -8.34901380930607e-05, 0, 124007.936503522, 124007.93653343, 0, -2.15030843640767e-05, 7.94637143887751e-07, 248015.873035367, -1.96882468931801e-06, -0.000329226638026367, 0, 3.64838090244729e-06, 0.000381129907974777, 0, -0.000218246054102321, 3.8051076815524e-05, 3.90136994269372e-05, 4.13309197671336e-05, 9.91522690223286e-05, -7.66886181708478e-05, -0.000138266806519934, -0.00063144700566058, 0, -0.0002074068473952, 0.000761703718953305, 0, -0.000274360131916314, 0.000152736706801823, 8.38173215496901e-05, 6.64725698587256e-05, 0.00014773787020989, -0.000232113332141611, -0.000381877057958639, 0.000245923401528676, 0.000267337780767475, -0.000313405810572139, -0.000587889907526641, 0, 0.000325578479524632, 0.000223811174104649, -1.73298648178798e-05, 0.000160128447654479, 5.67600300955777e-06, -0.000219908747437451, -0.000375356290126089, 0.000356313246103483, 0.000291005656246117, 0.000353596072411534, 0.000452805889391262, -0.00039057429157656, -0.000323230718732483, 0.000310344628626644, -5.59232841492231e-05, 0.000200654513667752, 0.000213106203343401, 0.000388289619622078, -8.22269969030942e-05}, 
                                                     {0, 0, 171.428571330471, 0, 0, -4776.19047354314, 0, -576.190476415534, -576.190476043779, 0, 0, 53426.1904475854, 0, 642.85714250863, 0, 15402.3809578924, 15402.3809495877, 0, 642.857140792216, 1285.71428448265, 0, -320951.058034991, 0, -238.0952224975, 0, -162177.248727145, -162177.248663581, 0, -238.095220136052, -16464.2857200801, 0, -16464.2856792291, -714.285710616167, -32928.5714162836, -714.28568764788, 0, 1155694.44386545, 0, -8.32080006515229e-05, 0, 895694.444659762, 895694.444406566, 0, -6.33825492006049e-05, 5833.33314672198, 0, 162500.000150567, 162499.999962037, 0, 5833.33300506837, 1.34460889135716e-05, 324999.999970431, -0.000194187076343139, 17500.0000054826, -0.000125088123443281, 17499.99973432, 0, -2610416.66539784, 0, 0.000209978777897055, 0, -2874074.07457542, -2874074.07396886, 0, 9.03538529409018e-05, 0.000991498792482735, 0, -817361.111881807, -817361.111523123, 0, 0.00146374494650357, -8.22076133350749e-05, -1634722.22219926, 0.000619883131672508, -53703.7032439177, 0, -53703.7028998625, -0.000227896528483165, 0.000164216139365062, -161111.111304874, -161111.110405475, 0.000903758982850273, 0.00171717009387969, 0.000657059072709127, 0, 3724206.34744694, 0, -0.000298227672825747, 0, 5534722.22288138, 5534722.22200076, 0, 2.27280220057703e-06, -0.00217733515223511, 0, 2291666.66836366, 2291666.66804721, 0, -0.00304496584855908, 0.000140850347474762, 4583333.33339403, -0.00104601766012908, -0.00304272808340496, 0, 243055.55567055, 243055.556446418, 0, -0.0042634848131034, 0.000729818455556168, 0.000255484501631898, 729166.667279933, 729166.665871372, -0.00264620893139758, -0.00521428991343067, 0.0009825645473182, -0.00178401135691885, -0.00410538563450511, -0.00120392204507881, -0.00119482294482577, -0.00456367018257079, 0, -3253968.25247163, 0, 0.000243984787596107, 0, -6286375.66186384, -6286375.66108988, 0, -0.000140658158126801, 0.00246377728316038, 0, -3611111.1129493, -3611111.11271071, 0, 0.00316455910900574, -8.24169427806132e-05, -7222222.22238498, 0.000852107880638973, 0.00560270335901238, 0, -578703.705216551, -578703.708449794, 0, 0.00806869780345354, -0.00112195509157363, -0.000821184016760038, -1736111.11193315, -1736111.11080805, 0.00380642284105996, 0.00790491485238841, 0.00354967112779311, 0, 0.00237703249047126, -0.00287372941520901, 0.000672757923307177, -0.0023169156937502, 0.0043138188968193, 0.00452618589610089, 0.0105555012763209, 0.000332712300255258, -0.000259203933250875, 0.00955760805726476, 0.00110095021768615, 0.00744527213062753, 0.00547390650859717, 0, 1587301.58658966, 0, -0.000107737146369303, 0, 3869047.61923141, 3869047.6188659, 0, 0.000141698980558939, -0.00141595763462229, 0, 2976190.47713889, 2976190.47697525, 0, -0.0015766985069229, 1.01672562380109e-05, 5952380.95247987, -0.000303384468238766, -0.00441300679211961, 0, 694444.446075715, 694444.449091835, 0, -0.00652074368093479, 0.000769534475219319, 0.000616744882794216, 2083333.333847, 2083333.33353525, -0.00242525359289967, -0.00571534858216005, -0.00535707986263946, 0, -0.00132738699466277, 0.00284532605561388, 0, -0.00595733595177309, 0.0026252382439016, 0.000795858648612291, 0.00205262881713581, -0.00160384688874019, -0.00479068131456762, -0.0116893284394552, 0.00122827405535069, 0.000699310123004282, -0.0102801096365104, 0.00350023918404965, -0.0024894530511086, -0.0082045958935288, -8.27002576692449e-05, -0.00543802870232917, 7.3603142095729e-05, 0.00025631838515813, -0.0107700862863472, -0.00848210279423937, -0.00882758777572099, 0, -330687.830542948, 0, 1.9944264199712e-05, 0, -992063.492090198, -992063.492020004, 0, -4.35527914875694e-05, 0.000330750484846743, 0, -992063.492241812, -992063.492186005, 0, 0.000305655888809755, 8.73686001373375e-06, -1984126.98415418, 2.24929304814901e-05, 0.00129883527199231, 0, -330687.831219801, -330687.832162795, 0, 0.00190678813035132, -0.000166021612965411, -0.000181788854386671, -992063.49217562, -992063.492225699, 0.000561997466593396, 0.00154767401693781, 0.00227780085798886, 0, -5.22448813809765e-05, -0.00227708648519351, 0, 0.00306636295741119, -0.00090125744045558, -0.000489577792441573, -0.000578533977012955, 0.000169190343533558, 0.00174914179540224, 0.00451843163998622, -0.000785656042628999, -0.000506963234015973, 0.00378547312529979, 0.00161708697863338, 0, 0.00045874969877142, -0.00174488806235765, 0.000170429910631179, -0.00157434169064234, 0.0020889328493139, 0.00172297017728813, 0.00484574503760493, -0.000528359947382516, 0.000323102042864382, -0.000759383936963039, -0.000958349012073, 0.00567709893686784, 0.00503434867129249, -0.000378651713429876, 0.00396564762160967, 0.00124942084341178, 0.00328273019274159, 0.00116300777639026, 0.00497261259124866}, 
                                                     {0, 0, -349.999999827934, 0, 0, 9668.05555088803, 0, 1676.38888971258, 1676.38888823982, 0, 0, -106895.601801152, 0, -2993.05555757789, 0, -44263.6574269363, -44263.6573955748, 0, -2993.05554390519, -5986.11110650622, 0, 632942.129333555, 0, 2361.11108466114, 0, 458089.120544168, 458089.120316892, 0, 2361.11101689857, 75424.7685920835, 0, 75424.7683190726, 7083.33330992488, 150849.536994887, 7083.3332314622, 0, -2240746.52673894, 0, -694.444258184105, 0, -2474351.85261136, -2474351.85172871, 0, -694.444110380295, -56689.8146009607, 0, -727100.695243256, -727100.693861478, 0, -56689.8131363446, -2777.77774013509, -1454201.38885219, -2777.77710275632, -170069.44441102, -4166.66609525018, -170069.443326083, 0, 4965798.60881985, 0, -0.000525887495784115, 0, 7727199.07590697, 7727199.07381802, 0, -0.000539152621761352, 15856.4795398565, 0, 3543344.91089217, 3543344.9075178, 0, 15856.4747708398, -1.76619926000626e-05, 7086689.81506883, -0.00198713871813475, 506134.259630485, 0, 506134.253816251, 63425.926403575, -0.00134220817122909, 1518402.77875019, 1518402.77466073, -0.00343240343450631, 63425.9190652637, 95138.8848273074, 0, -6940972.21902525, 0, 0.000807973691641138, 0, -14417824.0766357, -14417824.0735823, 0, 0.000273678362416706, 0.00484565421921982, 0, -9539930.56298543, -9539930.55815492, 0, 0.0133899689129854, 0.000130905269143632, -19079861.1119837, 0.00296066334638631, -130208.328361336, 0, -2193287.0418371, -2193287.03305256, 0, -130208.312367157, -0.00169662374220521, 0.000848822551318116, -6579861.11477838, -6579861.10689055, 0.00900344139445139, 0.0201667370487802, -520833.338029397, 0.00787196773433922, -520833.316677707, 0.010818415460494, -781249.991546195, 0.0205040731232493, 0, 5937499.99726473, 0, -0.000709939003923751, 0, 15810185.1872589, 15810185.1845463, 0, 0.000286222731821056, -0.00582465689868958, 0, 14299768.5267638, 14299768.522291, 0, -0.0140013041864916, -0.000445089949511422, 28599537.0383002, -0.00189914808664941, -0.0117411374466913, 0, 4918981.49226117, 4918981.48765411, 0, -0.0356088144360283, 0.00252853030142199, 0.000713580605117334, 14756944.4500887, 14756944.441557, -0.0117604426203772, -0.0301407102093972, 491898.145289178, 0, 491898.127547996, 0.0112467484463116, -0.00526627447967722, 1967592.60543852, 1967592.5741574, -0.0172138810519572, -0.0400456634682612, -0.0119888033343929, 2951388.88329236, -0.0398167320695931, -0.0166108474761662, -0.0417118906879264, -0.0277459679160947, 0, -2835648.14684007, 0, 0.000337439282609247, 0, -9375000.00089668, -9374999.99957876, 0, -0.000414340867725092, 0.00348021038766308, 0, -11111111.1156687, -11111111.1132274, 0, 0.0072544370296861, 0.000439209501106976, -22222222.2230337, 0.000165829063802717, 0.0101248989336535, 0, -5439814.82426311, -5439814.82399779, 0, 0.0275855174514355, -0.00150515999530633, -0.000778529410140017, -16319444.4483671, -16319444.4439145, 0.00655232630112067, 0.0219828926916608, 0.00980680941099709, 0, -868055.558631666, -868055.552941507, 0, 0.0299942251489364, -0.00974314736595444, -0.000801823260923581, -3472222.23492245, -3472222.21251985, 0.0167446262653517, 0.0439249792985978, 0.00452496200906272, -5208333.33085267, 0.0387443814672283, -0.0177205962539365, 0.0106006441243356, 0.0306810231146816, 0.0139502674634204, 0.0362408874033979, 0.00874430352051463, 0.00948121522162826, 0.0532508676576768, 0.034193104944178, 0.0503313982555021, 0, 578703.703436218, 0, -6.75394504672411e-05, 0, 2314814.81497595, 2314814.8147076, 0, 0.00014174622043612, -0.000835901482014617, 0, 3472222.22320686, 3472222.22261062, 0, -0.00150358386265703, -0.000156548748850073, 6944444.4446482, 0.000203830177234974, -0.00312483724474569, 0, 2314814.81773667, 2314814.81814958, 0, -0.00802046921290228, 0.000217659109059979, 0.000209172245974701, 6944444.44549287, 6944444.44464198, -0.00118764813615341, -0.00600142680444873, -0.00494215697924179, 0, 578703.706339852, 578703.706976477, 0, -0.0133843384256129, 0.00314960943964824, 0.00122328707170008, 2314814.81906006, 2314814.81268614, -0.00559561101456586, -0.0173257102776562, -7.58831937288923e-05, 3472222.22208346, -0.0139258000266107, -0.00205648828648015, 0, -0.00628097797777207, 0.00799955548805205, -0.00107654486719843, 0.00883275080821089, -0.00855937849838573, -0.00629380850965473, -0.0176440193357989, -0.00317207005393493, -0.00836643500256667, -0.000201389758967598, -0.00129034677370744, -0.0246101639710142, -0.0177555008829284, -0.00257766322160532, -0.0228957662625869, -0.00544322585882325, -0.0250047262115171, -0.0143327956726317, -0.0266600657769031}, 
                                                     {0, 0, 503.999999791818, 0, 0, -13753.9999943119, 0, -3254.00000187044, -3253.99999895517, 0, 0, 149624.999937818, 0, 8333.33334190292, 0, 84458.3333775293, 84458.333313645, 0, 8333.33331086392, 16666.6666634819, 0, -868736.110745392, 0, -10583.3333071672, 0, -853513.889280275, -853513.888832483, 0, -10583.3331510011, -205361.111352443, 0, -205361.11067364, -31749.9999555038, -410722.222254301, -31749.9998626166, 0, 3008263.88759549, 0, 6666.6663660411, 0, 4474513.89061996, 4474513.88890683, 0, 6666.66605714269, 247152.77788578, 0, 1916736.11332701, 1916736.10988427, 0, 247152.773937508, 26666.6664515363, 3833472.22309979, 26666.6658841203, 741458.333692549, 39999.9987328487, 741458.331700825, 0, -6512013.88601253, 0, -1666.66569941629, 0, -13494375.0042794, -13494375.0002314, 0, -1666.66586839347, -147222.219883927, 0, -8952430.56474283, -8952430.55609882, 0, -147222.206926402, -8333.3326273556, -17904861.1152032, -8333.33179476805, -2115625.00352095, 0, -2115624.98740871, -588888.890541105, -16666.6623286869, -6346875.00599566, -6346874.99792052, -16666.6601773122, -588888.878141286, -883333.324112601, 0, 8888888.88484283, 0, -0.00161639594155776, 0, 24236111.1172993, 24236111.1113212, 0, 0.000115645921483133, 34722.2150245682, 0, 22881944.4639843, 22881944.4508312, 0, 34722.1916370069, -0.00183832747546125, 45763888.8978702, -0.000184868985900023, 1145833.3296211, 0, 8645833.35026667, 8645833.32265195, 0, 1145833.28534294, 173611.11476824, -0.00585683945066541, 25937500.0209393, 25937500.0034319, -0.0141900113586088, 173611.078835985, 4583333.35268855, -0.0197507177944288, 4583333.31511191, 347222.189367513, 6874999.98700454, 347222.180125142, 0, -7430555.55206716, 0, 0.00152647975478899, 0, -25555555.5607774, -25555555.5554092, 0, -0.00141312723596884, 0.00936741607033993, 0, -32326388.9110536, -32326388.8982923, 0, 0.0322393117256747, 0.00295195302460825, -64652777.788493, -0.003744079606481, -243055.53961285, 0, -17951388.9213248, -17951388.8983804, 0, -243055.474210497, -0.00398834979441302, 0.00374986315244123, -53854166.699579, -53854166.6775573, 0.0142662094479893, 0.0495363351795924, -3993055.56146699, 0, -3993055.50624724, -1215277.81256202, 0.0188175291557748, -15972222.2754008, -15972222.2161821, 0.0375099286460786, -1215277.72685612, 0.0459904064916984, -23958333.3370832, 0.0732492076210649, -2430555.51089582, 0.113271367759378, -2430555.5127163, 0, 3472222.22054212, 0, -0.0007763127290816, 0, 14583333.3357116, 14583333.3330594, 0, 0.00137272083317982, -0.00587449839940805, 0, 23611111.1238574, 23611111.1167233, 0, -0.0169206178702171, -0.00231027528305015, 47222222.2288124, 0.00492005951224966, -0.0155553554633701, 0, 18055555.5828833, 18055555.5725114, 0, -0.0633739386734191, 0.00142552616514326, -0.00186398587879565, 54166666.6911524, 54166666.6773864, -0.00372441252881277, -0.0368512369568604, 694444.433821697, 0, 6250000.0169655, 6249999.98878963, 0, 694444.375637042, 0.0270537040921934, -0.00410590267623924, 25000000.055467, 25000000.006197, -0.0321061590284551, -0.0644432400572365, -0.0274865987858289, 37500000.012657, -0.0649835440123143, 3472222.28362975, -0.0277892477640715, 3472222.20131411, -0.0470643996527769, -0.113493032529186, 6944444.43278831, 6944444.44724009, -0.131687139232467, -0.0363994225362596, -0.128317184509763, 0, -694444.444098656, 0, 0.000165674194727868, 0, -3472222.22267627, -3472222.22212271, 0, -0.000427074940598435, 0.00146402683070534, 0, -6944444.44736648, -6944444.44562843, 0, 0.00354123078191558, 0.000716340482612122, -13888888.8905248, -0.00189364792822879, 0.0050658771433409, 0, -6944444.4528935, -6944444.45085912, 0, 0.0186031500416092, 0.000182313558277494, 0.000626037331460483, -20833333.3404678, -20833333.3370997, -0.00114320512784114, 0.0100306388152043, 0.00716520185556507, 0, -3472222.23149294, -3472222.2268985, 0, 0.0303264527650332, -0.00802778229884834, -0.00105958196198726, -13888888.9090296, -13888888.892559, 0.00914728359042636, 0.0283130590430291, 0.00611274745960763, -20833333.3398599, 0.0222185190739589, -694444.444302209, 0, -694444.428840845, -0.0247409921368173, 0.0043225195970254, -3472222.25451207, -3472222.22225911, 0.0140716692438708, 0.0182558039318976, 0.0146740288209357, 0.0339996889705912, -6944444.45182975, -6944444.4514248, 0.0552198362684651, 0.0165322499421224, 0.00459202275260595, 0.066246126942618, -0.0132778330974755, 0.076610208557601, 0.0364125302608636, 0.061660136153254}, 
                                                     {0, 0, -524.999999825291, 0, 0, 14064.5833285191, 0, 4439.58333615269, 4439.58333342822, 0, 0, -149352.083280422, 0, -15459.7222381224, 0, -112311.805623034, -112311.805553084, 0, -15459.722213269, -30919.4444627188, 0, 843271.990427604, 0, 28385.4166503468, 0, 1095971.06541609, 1095971.06494232, 0, 28385.4165753415, 368584.491173388, 0, 368584.490358083, 85156.2500100654, 737168.981918236, 85156.2500930817, 0, -2835295.13777277, 0, -28993.0551796265, 0, -5509045.14159302, -5509045.13979466, 0, -28993.0553938881, -636371.528431753, 0, -3281128.47598669, -3281128.47206402, 0, -636371.523410587, -115972.221903094, -6562256.94847465, -115972.223092291, -1909114.58555351, -173958.332007165, -1909114.58429311, 0, 5962326.38638461, 0, 15624.9986631077, 0, 15873148.1549975, 15873148.1506431, 0, 15625.0007036218, 609664.350021297, 0, 14437731.496894, 14437731.4876618, 0, 609664.332938427, 78124.9983484669, 28875462.9785511, 78125.0051558639, 5120949.08164232, 0, 5120949.06206575, 2438657.41349898, 156249.992547885, 15362847.2426549, 15362847.2375291, 156249.994354656, 2438657.40750113, 3657986.10452315, 0, -7920138.88533352, 0, -3472.21984957529, 0, -27243055.5657598, -27243055.5590668, 0, -3472.22540329017, -307291.658968431, 0, -34505208.3664699, -34505208.3523638, 0, -307291.627441763, -20833.3288243582, -69010416.6983185, -20833.3486828342, -4383680.55607505, 0, -19262152.8086507, -19262152.7786536, 0, -4383680.49867051, -1536458.34286149, -52083.3199330362, -57786458.3993752, -57786458.3865143, -52083.3267922576, -1536458.32566175, -17534722.2763859, -69444.413696588, -17534722.2617059, -3072916.6184597, -26302083.3555149, -3072916.63169524, 0, 6458333.33024012, 0, -0.00235416192815694, 0, 27523148.1570477, 27523148.1508067, 0, 0.0050172911239914, 63657.3964987408, 0, 45572916.7051301, 45572916.6904106, 0, 63657.3655012442, -0.00687919018541504, 91145833.3691994, 0.0244085466998463, 1944444.42924898, 0, 36226851.90858, 36226851.8865568, 0, 1944444.33963181, 381944.452119881, -0.0122123798187939, 108680555.657343, 108680555.636382, 0.00537257148572941, 381944.424466954, 13599537.0556241, 0, 13599536.991138, 9722222.30218105, -0.0385106756341971, 54398148.2894393, 54398148.26588, -0.0515247081906251, 9722222.26676729, 954861.028213508, 81597222.3190039, 954861.062202023, 19444444.4145564, 1273147.97659851, 19444444.482803, 0, -2951388.88738632, 0, 0.00124981235018194, 0, -15104166.6708729, -15104166.6676795, 0, -0.00360472636680844, 0.00718469648405803, 0, -31250000.0228202, -31250000.0137501, 0, 0.0219173869569227, 0.00524900329992042, -62500000.0214066, -0.0194577337809691, -347222.205103964, 0, -32986111.1588116, -32986111.1491671, 0, -347222.136836106, -0.00175195739438372, 0.00716780086640604, -98958333.4092658, -98958333.3929004, -0.0177114855047739, 0.017999956229959, -4687499.9930133, 0, -18229166.7001268, -18229166.6728982, 0, -4687499.91919799, -2083333.38949454, 0.0156736238447449, -72916666.814705, -72916666.7848199, 0.0366532984004381, -2083333.33751318, 0.0583570436130615, -109375000.108527, 0.0347153128542101, -23437500.1424565, 0.0491774424591387, -23437500.1209243, -5208333.27276105, 0.201629081186837, -46875000.0598301, -46875000.1393883, 0.187590668398982, -5208333.41467241, -6944444.29266503, 0, 578703.703391996, 0, -0.00027683802938405, 0, 3472222.22305743, 3472222.22236898, 0, 0.000991202614938923, -0.00186060068383273, 0, 8680555.56099246, 8680555.55858767, 0, -0.00446837330010695, -0.00159604667035881, 17361111.1163371, 0.00604529942031119, -0.00586345376840694, 0, 11574074.0890309, 11574074.0867801, 0, -0.025742259923914, -0.000626728757133038, -0.00221067404244553, 34722222.2444648, 34722222.2396213, 0.0096393527461458, -0.00463249314061952, 578703.696169208, 0, 8680555.57233813, 8680555.56944779, 0, 578703.664877782, 0.0154737329868603, -0.00110433217657276, 34722222.2771892, 34722222.2627982, -0.00710240437048901, -0.0111357554386992, -0.0155533742638423, 52083333.3754341, -0.00955541654400982, 3472222.22533501, 0, 3472222.20838388, 3472222.27437399, -0.010772032987138, 17361111.1870967, 17361111.1766643, -0.0217794460847918, 3472222.25804778, -0.0242615574780231, -0.0691825896603764, 34722222.2767031, 34722222.2932958, -0.0721537722623272, 0.0390565646889285, 8680555.57303345, -0.113935604120278, 8680555.6628666, -0.123271779822017, 11574074.0573813, -0.0632535887053794}, 
                                                     {0, 0, 399.999999905796, 0, 0, -10382.5396799181, 0, -4382.53968540654, -4382.53968505487, 0, 0, 105973.280394432, 0, 20235.1852021394, 0, 106208.465678716, 106208.465645203, 0, 20235.1852152788, 40470.37042194, 0, -573728.835807614, 0, -51074.0740622158, 0, -979404.762539975, -979404.762329278, 0, -51074.074267098, -456750.0004747, 0, -456750.000087002, -153222.222389147, -913500.000992513, -153222.222737937, 0, 1854027.77716326, 0, 76157.4070372942, 0, 4626481.48439773, 4626481.48358526, 0, 76157.4083901472, 1070740.74161545, 0, 3767037.0411238, 3767037.03975793, 0, 1070740.7389112, 304629.6296417, 7534074.08182071, 304629.633084435, 3212222.22727812, 456944.444362544, 3212222.22846512, 0, -3763194.44305156, 0, -67129.6282683919, 0, -12566203.7112826, -12566203.7090446, 0, -67129.633141626, -1478009.2581565, 0, -15186574.0908901, -15186574.0893791, 0, -1478009.2476654, -335648.146704222, -30373148.1761049, -335648.1620806, -7794444.45311236, 0, -7794444.44516963, -5912037.05038706, -671296.290025837, -23383333.3709598, -23383333.3774083, -671296.296509179, -5912037.06036069, -8868055.56572038, 0, 4847222.22022101, 0, 32407.4049293524, 0, 20460648.1597747, 20460648.1558007, 0, 32407.4149201669, 1201388.88256597, 0, 33402777.8145483, 33402777.8128175, 0, 1201388.86173518, 194444.439420966, 66805555.60997, 194444.477465743, 9328703.70686241, 0, 25949074.1078419, 25949074.1071582, 0, 9328703.67520406, 6006944.46417773, 486111.097196796, 77847222.3365759, 77847222.3498729, 486111.121346229, 6006944.49006035, 37314814.9079635, 648148.119908148, 37314814.9544627, 12013888.8655747, 55972222.3270265, 12013888.9068066, 0, -3849206.34744349, 0, -6613.75409777646, 0, -19745370.3808315, -19745370.3766838, 0, -6613.76571867482, -532407.397892926, 0, -40995370.4140598, -40995370.4100621, 0, -532407.377226524, -46296.2881338514, -81990740.800321, -46296.3416109175, -6412037.0258376, 0, -43518518.5802726, -43518518.5946953, 0, -6412036.96518963, -3194444.45988624, -138888.875005232, -130555555.726518, -130555555.734972, -138888.922125669, -3194444.49080353, -24305555.5780042, 0, -24305555.5641677, -32060185.3096064, -231481.437432513, -97222222.4499478, -97222222.5347389, -231481.435928427, -32060185.3981857, -7986111.04540998, -145833333.59955, -7986111.15131351, -64120370.4508728, -10648148.0219198, -64120370.5993535, 0, 1719576.71870992, 0, -0.00136242120127016, 0, 10416666.6717778, 10416666.6694727, 0, 0.0057518599972674, 99206.3427109628, 0, 26388888.9155294, 26388888.9111264, 0, 99206.3336003016, -0.00642734554184165, 52777777.8123787, 0.032817265267422, 2314814.80073674, 0, 35879629.6821868, 35879629.6943972, 0, 2314814.74868282, 694444.449186792, -0.00823698469318854, 107638889.014099, 107638889.013404, 0.0414653604160566, 694444.469878395, 11805555.5521581, 0, 27777777.8144339, 27777777.831761, 0, 11805555.5178283, 13888888.9704288, -0.0216585543237595, 111111111.3449, 111111111.400331, -0.0268228788857415, 13888889.0257241, 2083333.2787486, 166666666.932919, 2083333.37470929, 59027777.9899462, -0.0568631131004822, 59027778.1193453, 34722222.2309497, 3472222.03120685, 118055555.782161, 118055555.948673, 3472222.07983761, 34722222.5140387, 46296296.2893328, 0, -330687.830505941, 0, 0.000306690455743901, 0, -2314814.81586474, -2314814.81533716, 0, -0.00147132785112307, 0.00173338704487463, 0, -6944444.45098726, -6944444.44936925, 0, 0.00296978593097144, 0.0019912723433048, -13888888.8971699, -0.00959117383980211, -330687.825707848, 0, -11574074.0908751, -11574074.0935477, 0, -330687.809480697, 0.00018218056828837, 0.00251496663908101, -34722222.2583142, -34722222.2565826, -0.017748577439816, -0.00716590580578352, -2314814.80871872, 0, -11574074.092199, -11574074.1034048, 0, -2314814.78875681, -2314814.83606318, 0.00259740500431972, -46296296.3825595, -46296296.3926624, 0.00146073474931356, -2314814.84319431, 0.0160401647065766, -69444444.5411296, -0.0163964875644161, -6944444.44866769, 0, -6944444.44989453, -13888888.9615978, 0.0153840869173949, -34722222.3340006, -34722222.3858354, 0.0241151518297515, -13888889.0119451, -6944444.43588597, 0.0727123202599904, -69444444.5811986, -69444444.6298297, 0.053604274528806, -6944444.56875517, -34722222.2986229, 0.112144010368687, -34722222.4699466, -11574073.9870979, -46296296.3838311, -11574074.0849926}, 
                                                     {0, 0, -224.999999979284, 0, 0, 5465.1785708353, 0, 3215.17857335977, 3215.17857482271, 0, 0, -51751.3392792459, 0, -19599.5535817403, 0, -71350.8929061355, -71350.8929096551, 0, -19599.5536205628, -39199.1072015537, 0, 262665.674564747, 0, 66670.1388724126, 0, 592001.488548858, 592001.488584479, 0, 66670.13923282, 396005.952696364, 0, 396005.952808693, 200010.416935737, 792011.905821597, 200010.417313373, 0, -805303.819303602, 0, -138602.430279229, 0, -2554513.89102505, -2554513.89111504, 0, -138602.432061365, -1221111.11164806, 0, -2831718.75277271, -2831718.75398122, 0, -1221111.11209014, -554409.722738656, -5663437.50767387, -554409.72610948, -3663333.33911472, -831614.584738904, -3663333.34161621, 0, 1566579.86078012, 0, 180555.554584485, 0, 6446875.00571099, 6446875.00555119, 0, 180555.559895439, 2288802.08252937, 0, 10121701.400525, 10121701.4055751, 0, 2288802.08232776, 902777.777695599, 20243402.8042953, 902777.791683843, 7349652.78331902, 0, 7349652.78904983, 9155208.34921357, 1805555.55525376, 22048958.3711672, 22048958.384933, 1805555.56099351, 9155208.36567724, 13732812.5254004, 0, -1949652.77728126, 0, -144097.22047608, 0, -9892361.12014455, -9892361.11922779, 0, -144097.230155253, -2670138.88471936, 0, -20217013.9149339, -20217013.9232142, 0, -2670138.88213464, -864583.330742633, -40434027.8278352, -864583.363696541, -11189236.1127864, 0, -20937500.0219306, -20937500.0476899, 0, -11189236.1188094, -13350694.4682599, -2161458.33033626, -62812500.1084639, -62812500.1412457, -2161458.35063645, -13350694.5090067, -44756944.5366745, -2881944.43510017, -44756944.61325, -26701388.9161139, -67135416.8168529, -26701388.9558185, 0, 1505456.34874935, 0, 64484.1252163394, 0, 9097222.23061228, 9097222.22923084, 0, 64484.135682178, 1892361.10491563, 0, 22968750.0317568, 22968750.037242, 0, 1892361.10178913, 451388.883575603, 45937500.053461, 451388.9283912, 9999999.99266117, 0, 31076388.9298198, 31076388.9710204, 0, 9999999.98607361, 11354166.6855631, 1354166.66373469, 93229166.8231051, 93229166.8588977, 1354166.70204353, 11354166.7348457, 23871527.7911547, 0, 23871527.8318999, 50000000.1180312, 2256944.42521838, 95486111.3208794, 95486111.4676154, 2256944.42401624, 50000000.2619208, 28385416.6738285, 143229166.995011, 28385416.7682704, 100000000.188467, 37847222.2309325, 100000000.338208, 0, -657242.06325788, 0, -12400.7926938946, 0, -4613095.24233029, -4613095.24142018, 0, -12400.798816584, -744047.614806721, 0, -13888888.9087959, -13888888.909512, 0, -744047.614108052, -99206.3446400185, -27777777.8082148, -99206.3768620676, -4861111.10223496, 0, -23263888.9245496, -23263888.9515027, 0, -4861111.08836928, -5208333.33989024, -347222.220393245, -69791666.778357, -69791666.7964067, -347222.259764073, -5208333.37159575, -14236111.1081718, 0, -23437500.0226012, -23437500.0758933, 0, -14236111.1281177, -29166666.7408782, -694444.436636088, -93750000.2077901, -93750000.3231287, -694444.432293773, -29166666.8468045, -15624999.9934765, -140625000.308992, -15625000.0763117, -71180555.7434039, -868055.521605323, -71180555.945456, -72916666.7772554, -26041666.6159164, -142361111.428687, -142361111.609606, -26041666.651278, -72916667.0415466, -97222222.422491, 0, 124007.936456927, 0, -0.000215329410407963, 0, 992063.492962066, 992063.492734624, 0, 0.00127197117346546, 124007.935366441, 0, 3472222.22725408, 3472222.22683667, 0, 124007.935701213, -0.00147788249468937, 6944444.45163172, 0.00792042969621221, 992063.488959751, 0, 6944444.456131, 6944444.46225132, 0, 992063.483188101, 992063.492420404, -0.000788161691849311, 20833333.3648518, 20833333.3681335, 0.0151051925932056, 992063.50189744, 3472222.218246, 0, 8680555.56705405, 8680555.58742019, 0, 3472222.21849373, 6944444.46289457, 0.00036468020386495, 34722222.2972867, 34722222.3289192, 0.000983388536511812, 6944444.48828616, 3472222.21839165, 52083333.4409968, 3472222.24590264, 6944444.4465832, 0, 6944444.46575206, 20833333.3953549, -0.00959780138185359, 34722222.3180667, 34722222.4043905, -0.016546401067061, 20833333.4775493, 20833333.3575866, 6944444.41955095, 69444444.6077899, 69444444.6715363, 6944444.43168494, 20833333.4785202, 52083333.4481679, 8680555.51325789, 52083333.618249, 34722222.2460332, 69444444.6276071, 34722222.3261369}, 
                                                     {0, 0, 100.000000004086, 0, 0, -1928.96825406582, 0, -1928.96825470863, -1928.96825568626, 0, 0, 15855.753969305, 0, 15855.7539703814, 0, 31711.5079565628, 31711.5079628025, 0, 15855.7539942907, 31711.5079700082, 0, -73185.7363373512, 0, -73185.736310286, 0, -219557.209188897, -219557.209229571, 0, -73185.7365252801, -219557.209093873, 0, -219557.20923062, -219557.209211404, -439114.418557228, -219557.209370713, 0, 209756.94446172, 0, 209756.944287466, 0, 839027.77872459, 839027.77884616, 0, 209756.945282325, 839027.777791732, 0, 1258541.66765278, 1258541.66853682, 0, 839027.778717507, 839027.778471269, 2517083.33720418, 839027.779771055, 2517083.33690759, 1258541.6684008, 2517083.33821096, 0, -388252.314843117, 0, -388252.314346918, 0, -1941261.57668075, -1941261.57683327, 0, -388252.317087694, -1941261.57318645, 0, -3882523.15259811, -3882523.15531696, 0, -1941261.57571739, -1941261.57515939, -7765046.30913335, -1941261.58036409, -3882523.1493719, 0, -3882523.15498492, -7765046.3071109, -3882523.15286423, -11647569.464865, -11647569.473209, -3882523.15408743, -7765046.31447794, -11647569.466785, 0, 465277.777800965, 0, 465277.777002482, 0, 2791666.67089822, 2791666.67094381, 0, 465277.781678261, 2791666.66417169, 0, 6979166.67715687, 6979166.68090937, 0, 2791666.66763439, 2791666.66704002, 13958333.356904, 2791666.67939799, 6979166.66496144, 0, 9305555.56197524, 9305555.57854165, 0, 6979166.67787573, 13958333.3506144, 6979166.67560355, 27916666.7205891, 27916666.7426385, 6979166.67876696, 13958333.3662242, 27916666.7178939, 9305555.56512962, 27916666.7599656, 27916666.7154592, 41875000.1050674, 27916666.7273001, 0, -348875.661380627, 0, -348875.660630292, 0, -2442129.63365224, -2442129.63359778, 0, -348875.665454746, -2442129.62647194, 0, -7326388.9022131, -7326388.90432797, 0, -2442129.62920066, -2442129.62844505, -14652777.8025056, -2442129.645661, -7326388.88334679, 0, -12210648.1617951, -12210648.1837253, 0, -7326388.89577246, -14652777.7921473, -7326388.89920981, -36631944.5179237, -36631944.5453176, -7326388.90655151, -14652777.8078226, -12210648.1488149, 0, -12210648.1821011, -36631944.5101067, -12210648.1626758, -48842592.6968135, -48842592.7861106, -12210648.1544835, -36631944.5842421, -36631944.5024425, -73263889.0947044, -36631944.5214145, -73263889.0609296, -48842592.6826051, -73263889.1276419, 0, 148809.523804776, 0, 148809.523420921, 0, 1190476.19254964, 1190476.19251038, 0, 148809.52616287, 1190476.18847795, 0, 4166666.67531289, 4166666.67542529, 0, 1190476.18988878, 1190476.18891893, 8333333.34728323, 1190476.2016114, 4166666.66173669, 0, 8333333.34619419, 8333333.35904612, 0, 4166666.66613138, 8333333.33871529, 4166666.67267704, 25000000.0503345, 25000000.0667542, 4166666.68297634, 8333333.34688979, 8333333.32857465, 0, 10416666.6715007, 10416666.7046746, 0, 8333333.35432348, 25000000.0407347, 8333333.34802501, 41666666.7628807, 41666666.8401263, 8333333.3351343, 25000000.0928039, 25000000.035033, 62500000.1815116, 25000000.0443906, 41666666.7591452, 10416666.6699828, 41666666.8746933, 62500000.1333713, 41666666.7374458, 83333333.5593452, 83333333.655987, 41666666.7305992, 62500000.2363196, 83333333.5715265, 0, -27557.3192216092, 0, -27557.3191391589, 0, -248015.873464108, -248015.873459544, 0, -27557.3197948638, -248015.872502198, 0, -992063.494315983, -992063.494114007, 0, -248015.872900993, -248015.872433883, -1984126.98742045, -248015.876217786, -992063.490535591, 0, -2314814.81927065, -2314814.82193297, 0, -992063.490663216, -1984126.98462182, -992063.493348431, -6944444.45820998, -6944444.46205778, -992063.498402265, -1984126.98675728, -2314814.81226165, 0, -3472222.22545545, -3472222.23658361, 0, -2314814.81871809, -6944444.45418701, -2314814.82050665, -13888888.922015, -13888888.9457409, -2314814.81632989, -6944444.4659784, -6944444.45263172, -20833333.3936051, -6944444.45435031, -3472222.22136463, 0, -3472222.23576352, -13888888.9190589, -3472222.22547572, -17361111.1549412, -17361111.2080068, -3472222.2195232, -13888888.9649405, -20833333.3723133, -13888888.9123732, -34722222.3234133, -34722222.3640064, -13888888.9011906, -20833333.4140862, -34722222.3115925, -17361111.1340933, -34722222.3929573, -34722222.3097486, -46296296.4550642, -34722222.3339304}, 
                                                     {0, 0, 0, 99.9999999854561, 0, 0, -1928.96825357668, 0, -1928.96825411294, -1928.96825390833, 0, 0, 15855.7539640424, 0, 15855.753966664, 0, 15855.7539678789, 31711.507936047, 31711.5079346235, 31711.5079655675, 0, 0, -73185.7363072692, 0, -73185.736318238, 0, -73185.736370168, -219557.208993274, -219557.209007323, 0, -219557.208979728, -219557.208822497, -219557.209271663, -219557.209415051, -439114.418329526, 0, 0, 209756.944360279, 0, 209756.94438733, 0, 209756.94481908, 839027.777782601, 839027.777941777, 0, 839027.777676694, 0, 839027.776775258, 1258541.66657931, 1258541.66569357, 839027.778974083, 839027.780406102, 2517083.33490769, 1258541.66997704, 2517083.33616945, 2517083.33729387, 0, 0, -388252.314631502, 0, -388252.314661756, 0, -388252.316331937, -1941261.5741168, -1941261.57467317, 0, -1941261.57380371, 0, -1941261.57127236, -3882523.14790831, -3882523.14554525, -1941261.57696377, -1941261.58244258, -7765046.30035505, 0, -3882523.14758552, -3882523.14327149, -3882523.15917993, -7765046.30705348, -3882523.16400863, -7765046.31783873, -11647569.4546575, -11647569.4579705, -11647569.4724665, 0, 0, 465277.777524946, 0, 465277.777515506, 0, 465277.781002705, 2791666.66677119, 2791666.6678292, 0, 2791666.66623911, 0, 2791666.66298334, 6979166.66643442, 6979166.66188872, 2791666.67075535, 2791666.68192069, 13958333.3400853, 0, 6979166.66554007, 0, 6979166.65369793, 9305555.55386378, 9305555.54699994, 6979166.68595638, 13958333.3561833, 6979166.70374747, 13958333.3879992, 27916666.6852009, 27916666.68688, 9305555.5962691, 27916666.6974389, 27916666.7328248, 27916666.7439669, 27916666.77833, 41875000.0796663, 0, 0, -348875.661161535, 0, -348875.661104778, 0, -348875.665129008, -2442129.62974785, -2442129.63096139, 0, -2442129.62913972, 0, -2442129.62810263, -7326388.88896281, -7326388.88290028, -2442129.63293436, -2442129.64592791, -14652777.7852011, 0, -7326388.88798498, 0, -7326388.87058274, -12210648.1458622, -12210648.1395373, -7326388.90727918, -14652777.805245, -7326388.93471114, -14652777.8496593, -36631944.4649929, -36631944.4555791, 0, -12210648.1451484, -12210648.1320487, -12210648.2008075, -36631944.4956367, -12210648.2176063, -36631944.5763821, -48842592.6308543, -48842592.6851284, -36631944.5501906, -36631944.6463958, -73263888.9841037, -48842592.8054071, -73263889.0617698, -73263889.1463416, 0, 0, 148809.523708031, 0, 148809.523658374, 0, 148809.526076367, 1190476.19053907, 1190476.1913209, 0, 1190476.19011521, 0, 1190476.19133232, 4166666.66690774, 4166666.66249202, 1190476.19185708, 1190476.20000529, 8333333.33820176, 0, 4166666.66635668, 0, 4166666.65441155, 8333333.33202525, 8333333.32711241, 4166666.67571867, 8333333.35061707, 4166666.69559688, 8333333.38158963, 25000000.0141554, 24999999.9980736, 0, 8333333.33145832, 0, 8333333.31656593, 10416666.6633963, 10416666.6592699, 8333333.36693996, 25000000.043603, 8333333.38903062, 25000000.1153239, 41666666.6903906, 41666666.7236228, 25000000.0709749, 25000000.1702394, 62500000.0480272, 10416666.7266213, 41666666.7086319, 41666666.8021055, 41666666.8463527, 62500000.1685184, 41666666.9093696, 62500000.3268774, 83333333.4772519, 83333333.5635048, 83333333.6683463, 0, 0, -27557.3192033819, 0, -27557.3191892029, 0, -27557.319780281, -248015.873028547, -248015.873241906, 0, -248015.872900815, 0, -248015.873684297, -992063.492163803, -992063.490901503, -248015.873233783, -248015.875376073, -1984126.98552976, 0, -992063.492011882, 0, -992063.489085324, -2314814.81458347, -2314814.81249987, -992063.493856115, -1984126.98853974, -992063.499445255, -1984126.99732112, -6944444.44898156, -6944444.44163999, 0, -2314814.81446506, 0, -2314814.80828244, -3472222.22104659, -3472222.22091135, -2314814.82326526, -6944444.45902721, -2314814.8318245, -6944444.4818931, -13888888.8957035, -13888888.8998945, -6944444.46308635, -6944444.49907415, -20833333.3409269, 0, -3472222.2210173, -3472222.21841824, -3472222.24180646, -13888888.9074482, -3472222.24629029, -13888888.949721, -17361111.1227378, -17361111.1603098, -13888888.9447162, -20833333.3943254, -13888888.9873421, -20833333.4750636, -34722222.2610177, -34722222.291139, -17361111.2089886, -34722222.2962536, -34722222.3796274, -34722222.3699233, -34722222.4206719, -46296296.4475653}, 
                                                     {0, 0, 0, -224.999999939985, 0, 0, 5465.17856981397, 0, 3215.1785703944, 3215.17857134919, 0, 0, -51751.3392683651, 0, -19599.5535682421, 0, -19599.5535459712, -71350.8928575756, -71350.8928319735, -39199.107177719, 0, 0, 262665.674503142, 0, 66670.1388576814, 0, 66670.1387801019, 592001.488105909, 592001.487979706, 0, 396005.952360722, 396005.951690316, 200010.416999547, 200010.417135125, 792011.905210974, 0, 0, -805303.819098177, 0, -138602.430411141, 0, -138602.430586547, -2554513.88895989, -2554513.88874792, 0, -1221111.11089922, 0, -1221111.10686726, -2831718.74994587, -2831718.74608126, -554409.723653825, -554409.725115758, -5663437.5021389, -831614.58703629, -3663333.33717414, -3663333.33818776, 0, 0, 1566579.86035737, 0, 180555.555164424, 0, 180555.556954297, 6446875.00023678, 6446875.0002918, 0, 2288802.08256911, 0, 2288802.07080126, 10121701.3889199, 10121701.3783475, 902777.781348725, 902777.786913959, 20243402.7833699, 0, 7349652.77694211, 7349652.75673133, 1805555.56714386, 9155208.34770898, 1805555.57253361, 9155208.35988573, 22048958.3476516, 22048958.3495161, 13732812.5373299, 0, 0, -1949652.77673869, 0, -144097.221571175, 0, -144097.226355594, -9892361.1115396, -9892361.11234665, 0, -2670138.88736328, 0, -2670138.86988735, -20217013.8894792, -20217013.8718142, -864583.338747568, -864583.350029309, -40434027.7872892, 0, -11189236.1088063, 0, -11189236.0560719, -20937499.9979661, -20937499.9566302, -2161458.35222676, -13350694.4754132, -2161458.37054952, -13350694.5120032, -62812500.0260718, -62812500.0208432, -2881944.48490926, -44756944.4867861, -44756944.5259998, -26701388.9883393, -26701389.0375583, -67135416.7782194, 0, 0, 1505456.34832675, 0, 64484.1263391038, 0, 64484.132461351, 9097222.22264657, 9097222.22404753, 0, 1892361.10924193, 0, 1892361.09713547, 22968750.0013102, 22968749.9825039, 451388.89376702, 451388.907049612, 45937500.0107839, 0, 9999999.99724959, 0, 9999999.92559376, 31076388.8866804, 31076388.8413719, 1354166.68314091, 11354166.7051802, 1354166.7087382, 11354166.7559653, 93229166.695645, 93229166.6707164, 0, 23871527.772947, 23871527.6925035, 2256944.49310341, 50000000.0709163, 2256944.5082687, 50000000.1644988, 95486111.1634326, 95486111.2199464, 28385416.797391, 28385416.9342276, 143229166.797787, 37847222.4910014, 100000000.238365, 100000000.367359, 0, 0, -657242.063075261, 0, -12400.7933053857, 0, -12400.7971716363, -4613095.23830965, -4613095.23939453, 0, -744047.617763283, 0, -744047.615345638, -13888888.8900221, -13888888.8786631, -99206.3516204157, -99206.360135375, -27777777.7850869, 0, -4861111.10938368, 0, -4861111.06244513, -23263888.8881161, -23263888.8595435, -347222.22953162, -5208333.35868687, -347222.246067163, -5208333.39405617, -69791666.6869038, -69791666.6546857, 0, -14236111.1073508, 0, -14236111.0300054, -23437499.9952907, -23437499.9425267, -694444.472890571, -29166666.7286476, -694444.49000994, -29166666.8103969, -93750000.0321852, -93750000.0606569, -15625000.0848563, -15625000.2227301, -140625000.06042, -868055.607849919, -71180555.6129532, -71180555.720776, -26041666.8784147, -72916666.8943935, -26041666.9668476, -72916667.1410872, -142361111.311435, -142361111.438707, -97222222.6887434, 0, 0, 124007.936423354, 0, -7.67179889350914e-05, 0, 0.00089238018400236, 992063.492106663, 992063.492430129, 0, 124007.93613999, 0, 124007.936731812, 3472222.22257004, 3472222.21971575, 0.000497826639501543, 0.00280458473160165, 6944444.44661627, 0, 992063.491560034, 0, 992063.480084811, 6944444.4445201, 6944444.43622688, 0.00130185475484441, 992063.498899413, 0.00539205480952681, 992063.509073251, 20833333.3399532, 20833333.3263929, 0, 3472222.22122649, 0, 3472222.19342044, 8680555.55400158, 8680555.54039286, 0.00629764829703525, 6944444.46585482, 0.0117601111115287, 6944444.49109441, 34722222.2315344, 34722222.2304725, 3472222.243895, 3472222.29228741, 52083333.3400653, 0, 6944444.44240817, 6944444.41716649, 0.016501929516623, 20833333.3593555, 0.0191086546860895, 20833333.4071581, 34722222.2377145, 34722222.2811044, 6944444.50537532, 20833333.4151799, 6944444.56287662, 20833333.5411261, 69444444.4967167, 69444444.5385249, 8680555.66612289, 52083333.4353944, 52083333.5645735, 34722222.4195774, 34722222.5030148, 69444444.658922}, 
                                                     {0, 0, 0, 399.999999853126, 0, 0, -10382.5396785909, 0, -4382.53967999087, -4382.53968255908, 0, 0, 105973.280380878, 0, 20235.1851821595, 0, 20235.1851288828, 106208.46561205, 106208.465558089, 40470.3703850015, 0, 0, -573728.835734489, 0, -51074.0740346656, 0, -51074.0737481811, -979404.761945038, -979404.761631851, 0, -456750.000004651, -456749.998824364, -153222.222389856, -153222.222384796, -913500.000208206, 0, 0, 1854027.77693235, 0, 76157.4072041863, 0, 76157.4066536332, 4626481.48168358, 4626481.48087571, 0, 1070740.7405508, 0, 1070740.73354442, 3767037.03724049, 3767037.03020139, 304629.63038857, 304629.63070019, 7534074.07487363, 456944.446110248, 3212222.22458417, 3212222.22408258, 0, 0, -3763194.44260494, 0, -67129.6290605906, 0, -67129.6293286762, -12566203.7042633, -12566203.7033086, 0, -1478009.25824164, 0, -1478009.23809349, -15186574.0751369, -15186574.0550013, -335648.150107901, -335648.151425779, -30373148.1497794, 0, -7794444.44429595, -7794444.40789229, -671296.300995155, -5912037.04692893, -671296.303181448, -5912037.05011524, -23383333.3408604, -23383333.3352804, -8868055.57954769, 0, 0, 4847222.21968726, 0, 32407.406463896, 0, 32407.4092194549, 20460648.1490457, 20460648.1489674, 0, 1201388.88644711, 0, 1201388.85588697, 33402777.7804843, 33402777.7473285, 194444.447457703, 194444.450215571, 66805555.5587001, 0, 9328703.70188276, 0, 9328703.61168555, 25949074.0746465, 25949073.9923251, 486111.116663875, 6006944.46804284, 486111.121789084, 6006944.48051903, 77847222.2315917, 77847222.20666, 648148.162345536, 37314814.8408977, 37314814.8474773, 12013888.956985, 12013888.9913221, 55972222.2875602, 0, 0, -3849206.34706116, 0, -6613.75569819392, 0, -6613.76010463331, -19745370.3711969, -19745370.3722848, 0, -532407.404195823, 0, -532407.380443662, -40995370.3740512, -40995370.3415344, -46296.2990047935, -46296.3027288136, -81990740.7460457, 0, -6412037.03370811, 0, -6412036.91727923, -43518518.520635, -43518518.4225646, -138888.889739059, -3194444.47595081, -138888.892762552, -3194444.49297099, -130555555.563078, -130555555.510375, 0, -24305555.5533314, -24305555.4001166, -231481.494324063, -32060185.2369034, -231481.494965473, -32060185.2711485, -97222222.2440308, -97222222.2372999, -7986111.20253943, -7986111.29630139, -145833333.375856, -10648148.3351942, -64120370.530742, -64120370.6345535, 0, 0, 1719576.71856046, 0, -0.000476922231099436, 0, 0.00254663622415232, 10416666.6670691, 10416666.6681936, 0, 99206.346988913, 0, 99206.3390051911, 26388888.8913856, 26388888.8738969, 0.00137937433570752, 0.0041888787177563, 52777777.7828036, 0, 2314814.81202952, 0, 2314814.73810694, 35879629.6322126, 35879629.5699632, -0.00278147465829375, 694444.466083435, -0.00411028926332494, 694444.477514225, 107638888.895725, 107638888.843853, 0, 11805555.5526856, 0, 11805555.4163031, 27777777.7763379, 27777777.6599485, 0.00218454765562005, 13888888.9396624, -0.00211173232631498, 13888888.9703283, 111111111.116796, 111111111.085462, 2083333.39235608, 2083333.48057367, 166666666.644619, 0.00930656829118443, 59027777.8107875, 59027777.8458709, 3472222.36539367, 34722222.3898989, 3472222.42359868, 34722222.5954622, 118055555.658546, 118055555.754682, 46296296.6490933, 0, 0, -330687.830481657, 0, 0.000102978062603271, 0, -0.000683948305826751, -2314814.81489538, -2314814.81525698, 0, 0.000620296465402323, 0, 0.00114905816862926, -6944444.44510804, -6944444.44121028, -0.000302763577956953, -0.00120032012780593, -13888888.8907287, 0, -330687.829737522, 0, -330687.81180215, -11574074.0751531, -11574074.0584185, 0.00141377947603806, -0.00603263985252366, 0.00269000129990641, -0.00944820880328629, -34722222.2256475, -34722222.2066239, 0, -2314814.81375627, 0, -2314814.76780614, -11574074.0738455, -11574074.039523, 0.00198778363485403, -2314814.83359143, 0.00571315786578526, -2314814.84176236, -46296296.2970976, -46296296.2744068, -0.0143572210438068, -0.0425750436186349, -69444444.4224971, 0, -6944444.44322583, -6944444.3919672, -0.00209666586167387, -13888888.9077483, 0.000450694799738607, -13888888.9257228, -34722222.226935, -34722222.2368761, -0.0385962546434677, -6944444.50956133, -0.0727059658858769, -6944444.61262437, -69444444.4516674, -69444444.4776123, -0.0696480337270636, -34722222.2869691, -34722222.3968634, -11574074.226223, -11574074.3007297, -46296296.4430907}, 
                                                     {0, 0, 0, -524.999999763694, 0, 0, 14064.5833269837, 0, 4439.58333061213, 4439.58333348388, 0, 0, -149352.083265185, 0, -15459.7222210565, 0, -15459.7221645802, -112311.805561641, -112311.805504024, -30919.4444408205, 0, 0, 843271.99034819, 0, 28385.4166366389, 0, 28385.4163120565, 1095971.06487463, 1095971.06451178, 0, 368584.490777437, 368584.489628128, 85156.2500223387, 85156.249927543, 737168.98142511, 0, 0, -2835295.13753116, 0, -28993.0553774529, 0, -28993.0545556591, -5509045.13916415, -5509045.13808886, 0, -636371.527711058, 0, -636371.521096568, -3281128.47267398, -3281128.46556144, -115972.222391292, -115972.221949132, -6562256.94372825, -173958.333523899, -1909114.58395105, -1909114.58236328, 0, 0, 5962326.38593554, 0, 15624.9994831989, 0, 15624.9986761452, 15873148.1488509, 15873148.1471969, 0, 609664.35099638, 0, 609664.332474216, 14437731.4834221, 14437731.4621527, 78125.0003854305, 78124.9990843802, 28875462.9599705, 0, 5120949.07489722, 5120949.03930013, 156250.000655882, 2438657.41172259, 156250.000375365, 2438657.40740504, 15362847.2219371, 15362847.2112549, 3657986.11936839, 0, 0, -7920138.88482003, 0, -3472.2213662669, 0, -3472.22178554435, -27243055.5566051, -27243055.5554896, 0, -307291.664222706, 0, -307291.636260943, -34505208.3375956, -34505208.3018446, -20833.3335835253, -20833.3313165242, -69010416.6616147, 0, -4383680.55515326, 0, -4383680.47094633, -19262152.7815116, -19262152.695868, -52083.3323212845, -1536458.34575693, -52083.3288984914, -1536458.33853978, -57786458.3249849, -57786458.2882668, -69444.4460404996, -17534722.2294885, -17534722.2070969, -3072916.700073, -3072916.71218452, -26302083.3414374, 0, 0, 6458333.32989079, 0, -0.00081659373938579, 0, 0.000801519545250598, 27523148.1490614, 27523148.1492162, 0, 63657.4040302888, 0, 63657.381500428, 45572916.6717748, 45572916.6371129, -0.000255393868810911, -0.00216564759962047, 91145833.3301208, 0, 1944444.44184139, 0, 1944444.33753923, 36226851.8587753, 36226851.7510176, -0.00614679853659627, 381944.461773437, -0.0154162475579871, 381944.453033613, 108680555.540465, 108680555.474783, 0, 13599537.0390786, 13599536.8868208, 8.90075400968811e-05, 9722222.24748927, -0.00684323030012463, 9722222.22691547, 54398148.1403009, 54398148.0796543, 954861.161947313, 954861.196757382, 81597222.1711925, 1273148.25107503, 19444444.505163, 19444444.5514059, 0, 0, -2951388.8872583, 0, 0.000414750811668857, 0, -0.000893798518936747, -15104166.667093, -15104166.6677241, 0, 0.00233920055879782, 0, 0.010995944936341, -31250000.0031169, -31249999.9851826, 0.000329248183190283, 0.00102299396811232, -62500000.0001877, 0, -347222.219366461, 0, -347222.154846586, -32986111.117038, -32986111.0467794, 0.00725860463865719, -0.0115905887797988, 0.0168844636590774, -0.00578207085399512, -98958333.3248624, -98958333.2670316, 0, -4687499.99939162, 0, -4687499.8709498, -18229166.67011, -18229166.5455127, 0.00536234469511222, -2083333.36270382, 0.0189029118569003, -2083333.3490428, -72916666.6489656, -72916666.5713505, -0.0330469181403922, -0.0619550995168194, -109374999.904835, 0.00303251160361935, -23437500.0076277, -23437499.9740035, -0.0885564514721829, -5208333.42055437, -0.110326357000002, -5208333.52288362, -46874999.9945442, -46875000.0264981, -6944444.62777822, 0, 0, 578703.703372954, 0, -8.71881261399033e-05, 0, 0.000274450663988532, 3472222.22230563, 3472222.22254765, 0, -0.000644849893264606, 0, -0.00173130212000554, 8680555.55631385, 8680555.55253115, -9.96768944197381e-05, -0.000133556549043385, 17361111.1117736, 0, -0.00106636793574857, 0, -0.0165629450294799, 11574074.076017, 11574074.0572328, -0.00267779514564441, 0.00311745760152621, -0.00609451929257868, 0.00175953403748006, 34722222.2213313, 34722222.2017314, 0, 578703.702980771, 0, 578703.661576267, 8680555.55719174, 8680555.51799031, -0.00419420203109458, 0.0113821895346981, -0.0112549715999259, 0.00613452309837379, 34722222.215432, 34722222.1783681, 0.00699002423181955, 0.0140966392606549, 52083333.2892057, 0, 3472222.22253847, 3472222.17062712, -0.00109936829241155, 3472222.23153253, -0.00541439032382851, 3472222.22231514, 17361111.1068538, 17361111.0857084, 0.0251400599138036, 0.0394995724490431, 0.0354991596127873, 0.0899397713453139, 34722222.1876867, 34722222.1884439, 0.0444852152744367, 8680555.57609548, 8680555.62974413, 0.0904467266145788, 0.133775891801006, 11574074.1257955}, 
                                                     {0, 0, 0, 503.999999738638, 0, 0, -13753.9999929805, 0, -3253.99999836002, -3254.00000018288, 0, 0, 149624.999924689, 0, 8333.33333371378, 0, 8333.33329917945, 84458.333338934, 84458.3333026208, 16666.6666615164, 0, 0, -868736.110677429, 0, -10583.333319067, 0, -10583.3331194634, -853513.888940542, -853513.888700248, 0, -205361.111155255, -205361.110469643, -31749.9999912785, -31749.9999361431, -410722.222130481, 0, 0, 3008263.88738923, 0, 6666.66656312898, 0, 6666.66602103476, 4474513.88911666, 4474513.8883553, 0, 247152.777805905, 0, 247152.774006384, 1916736.11156293, 1916736.10717187, 26666.6666519658, 26666.6664542489, 3833472.22139393, 40000.000066246, 741458.333383441, 741458.33227805, 0, 0, -6512013.88562738, 0, -1666.66635379664, 0, -1666.66562120114, -13494375.0005543, -13494374.9992708, 0, -147222.221746827, 0, -147222.211448431, -8952430.55739382, -8952430.54376607, -8333.33310548434, -8333.33268602917, -17904861.1079316, 0, -2115625.00111168, -2115624.97994766, -16666.6670918934, -588888.890505409, -16666.6671900154, -588888.887454527, -6346874.99809015, -6346874.99051423, -883333.336218379, 0, 0, 8888888.88439612, 0, -0.000519658435768067, 0, -0.000885045709784088, 24236111.1118971, 24236111.1108294, 0, 34722.2206047007, 0, 34722.2053049475, 22881944.4482859, 22881944.4248009, -0.000933747018405683, -0.00146029555955599, 45763888.8832141, 0, 1145833.3338164, 0, 1145833.28586848, 8645833.33768266, 8645833.28442819, -0.000792127901295931, 173611.116316644, -0.00147064418074097, 173611.111528114, 25937499.9903969, 25937499.964552, 0.00319620549636744, 4583333.33422871, 4583333.31668998, 347222.23942078, 347222.244968138, 6874999.99375175, 0, 0, -7430555.55175474, 0, 0.000489809059026403, 0, 0.000306330271667493, -25555555.5562065, -25555555.5559856, 0, 0.00231482119840268, 0, 0.0147154489993787, -32326388.8932546, -32326388.8702948, 0.00156535809615923, 0.00173567378498224, -64652777.7730424, 0, -243055.554144133, 0, -243055.496943561, -17951388.8964544, -17951388.8267537, 0.00477962595162873, -0.00672009742500573, 0.00755378812612477, -0.00169108679240096, -53854166.6510643, -53854166.6056472, 0, -3993055.55924936, -3993055.46857178, -0.0037463912356442, -1215277.78863439, -0.00152505759067643, -1215277.77120804, -15972222.2110901, -15972222.1642546, -0.0275590287907057, -0.0460002546722604, -23958333.279511, -0.0629046963522767, -2430555.57624093, -2430555.59718938, 0, 0, 3472222.22042209, 0, -0.000243575380375376, 0, 3.66801030290275e-05, 14583333.3336236, 14583333.3338645, 0, -0.00160526428264304, 0, -0.00662502099295324, 23611111.1136287, 23611111.1018381, -0.00111708077629857, -0.000986786322551623, 47222222.2207942, 0, -0.00197523704372076, 0, -0.0364963779037709, 18055555.5616393, 18055555.5150147, -0.00571448678978506, 0.00372379901013872, -0.00906580644851826, 0.000758709986032651, 54166666.657106, 54166666.617824, 0, 694444.445158852, 0, 694444.371939335, 6250000.00510471, 6249999.92717249, -0.00118213843436381, 0.0137220039002977, -0.00621468754955562, 0.00292979267091908, 24999999.9835137, 24999999.9255099, 0.0166583239542409, 0.0325506894912632, 37499999.9205242, 0.00476729381504392, 3472222.22262791, 3472222.19330243, 0.0592880565146717, 0.0424205921398298, 0.0730946306223212, 0.0921500191783372, 6944444.41875053, 6944444.42666946, 0.0934442267689297, 0, 0, -694444.444079339, 0, 5.01504188629151e-05, 0, -3.85452796491162e-05, -3472222.22227718, -3472222.22239435, 0, 0.00043770004779447, 0, 0.0011812318172131, -6944444.44502222, -6944444.44257836, 0.00029712796269538, 0.000206027268190067, -13888888.8889376, 0, 0.000774187651615685, 0, 0.0089495336322558, -6944444.44631448, -6944444.43371969, 0.0021395713463407, -0.000802961508285656, 0.00337726013297856, -0.000173077774828488, -20833333.3315854, -20833333.3185025, 0, 0.000352890769568615, 0, 0.0231063959422209, -3472222.22442929, -3472222.19886195, 0.00231241953438687, -0.00520131877896426, 0.0051581427452298, -0.00156912945633086, -13888888.8827073, -13888888.8558987, -0.00253845989557894, -0.00652058017918158, -20833333.2988273, 0, -694444.445475978, -694444.414410553, -0.00160311238477577, -0.00424668853409233, 0.000237794527418482, 0.00399933653896795, -3472222.21795476, -3472222.20008494, -0.0175024496602646, -0.0208121191927574, -0.0251377774727635, -0.0446543401508997, -6944444.41263918, -6944444.40804279, -0.0343625524490132, -0.00513298043545974, -0.0286438398589157, -0.0513053409497527, -0.0755099941562107, -0.0158702751183778}, 
                                                     {0, 0, 0, -349.999999798641, 0, 0, 9668.0555501497, 0, 1676.38888825753, 1676.38888901257, 0, 0, -106895.601793868, 0, -2993.05555625192, 0, -2993.05554250369, -44263.6574107351, -44263.6573956876, -5986.11110936038, 0, 0, 632942.129295791, 0, 2361.11110714223, 0, 2361.11102844289, 458089.120399862, 458089.120296563, 0, 75424.7685477428, 75424.7682767924, 7083.33333146966, 7083.3333213588, 150849.536997968, 0, 0, -2240746.52662351, 0, -694.444403756611, 0, -694.444181695342, -2474351.85197947, -2474351.85163579, 0, -56689.8148604142, 0, -56689.8134157704, -727100.694718614, -727100.692936694, -2777.77775348098, -2777.77780520083, -1454201.38851715, -4166.66680128552, -170069.444467422, -170069.44407431, 0, 0, 4965798.60860095, 0, -0.000129418117744115, 0, -0.000468588133075427, 7727199.0743776, 7727199.07376148, 0, 15856.4813147968, 0, 15856.4775150829, 3543344.90849971, 3543344.90281293, -0.000228928379430615, 0.000120629991774058, 7086689.81333885, 0, 506134.260007038, 506134.251745721, 0.000463741319737832, 63425.9264938453, 0.000954960357894276, 63425.9258965578, 1518402.77695634, 1518402.77395675, 95138.8904542529, 0, 0, -6940972.21876457, 0, 0.000217648766664154, 0, 0.000503139950820946, -14417824.0744923, -14417824.0739174, 0, 0.000703314504613341, 0, 0.0062932838661549, -9539930.55780268, -9539930.5477541, 0.000756535381556018, -9.96953450695651e-05, -19079861.1083559, 0, -130208.33386051, 0, -130208.315753701, -2193287.03981326, -2193287.01837344, 7.92780668880058e-05, -0.001619999462114, -0.00147877870820414, -0.00158046261075929, -6579861.10700581, -6579861.09628017, -0.00325463378545942, -520833.333567301, -520833.327649397, -0.00808868464870526, -0.0117861050096266, -781249.997770917, 0, 0, 5937499.99707494, 0, -0.000204902992162577, 0, -0.000312482068664821, 15810185.1855215, 15810185.1853138, 0, -0.00103971866669486, 0, -0.00555808021278045, 14299768.5210279, 14299768.5110446, -0.00115435336649181, -1.6406461954406e-06, 28599537.0345563, 0, -0.000497943425750414, 0, -0.0215233449470913, 4918981.48624794, 4918981.45742553, -0.00212079274134373, 0.00160767155423581, -8.1226319282979e-05, 0.00242753808382917, 14756944.4376015, 14756944.4183447, 0, 491898.150766669, 491898.115522593, 0.00367413158507164, 0.00388121773365541, 0.00491684262514498, -7.92883920516013e-05, 1967592.58797686, 1967592.57007794, 0.0118633825041435, 0.0243365629937551, 2951388.86661201, 0.0318726112890062, 0.00889586800672626, 0.0202541178256495, 0, 0, -2835648.14676281, 0, 0.000100842959358475, 0, 0.00010859491118109, -9375000.00014534, -9375000.00019133, 0, 0.000724369824866663, 0, 0.00261266970701027, -11111111.1125264, -11111111.1073515, 0.000799998536060048, 6.63490829388684e-06, -22222222.22131, 0, 0.000898385379618109, 0, 0.0133858925006107, -5439814.81858415, -5439814.7989014, 0.00270002838673317, -0.000415059264158202, 0.00157635401248048, -0.00135169302719501, -16319444.4402241, -16319444.4233192, 0, -0.000757488720886625, 0, 0.0268826721641133, -868055.559055205, -868055.527796016, -0.000552543133142379, -0.0045795053015645, -0.00157819991530587, -0.00307927943852752, -3472222.21551382, -3472222.19203673, -0.00633977646761063, -0.0182544727895903, -5208333.29979857, -0.00550017525972711, -4.00521702084164e-05, 0.00918097642336489, -0.0295377001767336, -0.0176748796592543, -0.0393567981022094, -0.0418267647337303, 0.0102715656105832, 0.00633250369441893, -0.0427788183793548, 0, 0, 578703.703422733, 0, -2.06725345298362e-05, 0, -1.64290366168862e-05, 2314814.81484179, 2314814.81487962, 0, -0.00019624031597147, 0, -0.000495784014522173, 3472222.22253856, 3472222.22145534, -0.000212196854310569, -1.17267569647755e-06, 6944444.44438305, 0, -0.000364072935541303, 0, -0.00328119268009235, 2314814.81595147, 2314814.81055737, -0.00103008550187865, -4.4286012883798e-05, -0.000762898761345872, 0.00032656346616925, 6944444.44363593, 6944444.43794939, 0, -0.000113161703220951, 0, -0.00847118203141999, 578703.705191359, 578703.694673141, -0.000791720581526943, 0.00160442913926877, -0.00065524382549042, 0.00110467388190823, 2314814.81225637, 2314814.80108919, 0.000543161158372668, 0.00426020873664118, 3472222.20745203, 0, 0.000796283346892537, -0.0112892637289922, 0.0016504078303015, 0.00153488004157679, 0.00161475984735749, -0.00063827072996206, -0.00166841758248361, -0.00810741409920236, 0.00848146242591128, 0.00823149046421602, 0.0149135880187516, 0.0194487641421443, -0.0133560658915342, -0.0150278216592036, 0.018665064033292, 0.00219721500411875, 0.0125012335850241, 0.0230982018758317, 0.0348180088060611, 0.00777820114921947}, 
                                                     {0, 0, 0, 171.428571321834, 0, 0, -4776.19047332585, 0, -576.190476038863, -576.190476241982, 0, 0, 53426.1904454698, 0, 642.857143222499, 0, 642.857139628968, 15402.3809536692, 15402.3809496749, 1285.71428546433, 0, 0, -320951.0580242, 0, -238.095237610758, 0, -238.095216959994, -162177.248688379, -162177.248660583, 0, -16464.2857258744, -16464.285655384, -714.285715101819, -714.285714938528, -32928.5714205528, 0, 0, 1155694.44383296, 0, -1.0533603973349e-05, 0, -7.30983433777196e-05, 895694.444492691, 895694.444397799, 0, 5833.33335617666, 0, 5833.33299008861, 162500.000103031, 162499.999637172, -9.03080661610745e-06, 3.39999607834888e-05, 324999.999913557, 5.77758267005858e-05, 17500.0000204381, 17499.9999369792, 0, 0, -2610416.66533686, 0, 3.54953164394352e-05, 0, 0.000139543649544257, -2874074.07418904, -2874074.07401514, 0, 3.47213224121587e-05, 0, 0.000985822551127922, -817361.111520044, -817361.1100227, 8.19163668531196e-05, -0.000113574157061361, -1634722.22185113, 0, -53703.7039933593, -53703.7018968003, -0.000161248753716435, -0.000136897290806596, -0.000376532340363592, -0.000171885976774354, -161111.110976162, -161111.110267935, -0.000554499214811701, 0, 0, 3724206.34737451, 0, -6.11265025303457e-05, 0, -0.000174489648877517, 5534722.22238051, 5534722.22221175, 0, -0.000194745616507915, 0, -0.00162910704844171, 2291666.66751255, 2291666.66480514, -0.000258743690069045, 0.000169012863113963, 4583333.33260137, 0, 0.000229887807089412, 0, -0.00433022445355421, 243055.556601962, 243055.551081527, -3.32031055002815e-06, 0.000317475186910127, 0.00072046575158242, 0.000754087225703758, 729166.665810946, 729166.663141818, 0.00109860596841339, 0.000106023733042301, -0.00089976674328709, 0.00225633276596584, 0.00365790985616132, -2.08838115863819e-06, 0, 0, -3253968.25241849, 0, 5.82429937132793e-05, 0, 0.000124659463218389, -6286375.66150326, -6286375.66143497, 0, 0.000295959970648357, 0, 0.00146906262048238, -3611111.11206289, -3611111.10934063, 0.000383938762731455, -0.00014519946819035, -7222222.22153324, 0, 0.000116754004834191, 0, 0.00540646054745689, -578703.705515828, -578703.698004514, 0.000571042194744713, -0.000191343310739419, -0.000481039056776818, -0.000982970677949939, -1736111.10955659, -1736111.10461499, 0, -0.00100621159785134, 0.00789975029181369, -0.00106074939204976, -0.000838177942764876, -0.0017496033884512, -0.000850482419988912, 0.000954280540090249, 0.00474035699772604, -0.00291026775906621, -0.00748004868825581, 0.00452810258459775, -0.00911802014851754, -0.00244675136438263, -0.0071584789225236, 0, 0, 1587301.58656757, 0, -2.87274652123505e-05, 0, -5.48556265913049e-05, 3869047.61910287, 3869047.61911045, 0, -0.000207954384011377, 0, -0.000727094115584351, 2976190.4767304, 2976190.47530106, -0.0002641522432723, 9.76398118402957e-05, 5952380.95211171, 0, -0.000263703880585244, 0, -0.00343073149579061, 694444.445880867, 694444.440679713, -0.000722913927776507, -0.000111539402981451, -4.98290486246995e-05, 0.000550348577090673, 2083333.33241949, 2083333.32798649, 0, 0.000319479531003404, 0, -0.00663871131549286, 0.00135130214305799, -0.00665349376253547, 0.00018313537587058, 0.000861137718426661, 0.000897263290988064, 0.00115066991731524, -0.00136750515520775, -0.00693904761946586, 0.00145732983167416, 0.0057093309040831, -0.00760978394962143, 0.00169911058816023, -1.55531821582572e-05, -0.00113705540608551, 0.00810167540043976, 0.00436837966788027, 0.0114994805005617, 0.0124843227817684, -0.00176875478355297, 0.000185431759782746, 0.0119535579821046, 0, 0, -330687.830538954, 0, 5.99642258427636e-06, 0, 1.11320984581528e-05, -992063.4920739, -992063.492083566, 0, 5.60945019564556e-05, 0, 0.000142223794858569, -992063.492184741, -992063.491878286, 7.02975913223967e-05, -1.89964809454514e-05, -1984126.98410083, 0, 0.000108551814005352, 0, 0.00084382867499385, -330687.831122406, -330687.829679865, 0.000275365721368044, 8.16489694445414e-05, 8.15011980321209e-05, -0.000137735858678885, -992063.491898012, -992063.490365196, 0, 2.6469301103703e-05, 0, 0.00209824426869698, -0.000575461062529319, 0.0021312029326844, 0.000178866927927517, -0.000278270672236141, -2.48382293383686e-05, -0.000348642639433004, 0.000540320548383288, 0.00331602049613847, -4.16190800563149e-05, -0.0014260222838786, 0.00356760184871249, 0, -0.000315136298100203, 0.00274027269815106, -0.000469554313218448, -0.000289449073832867, -0.000558617902064689, -0.000104336491901847, 0.000349298055287632, 0.00156018184922908, -0.00220538102940739, -0.00183457274129362, -0.00450465807219308, -0.00539996005773926, 0.00309867886994274, 0.00304774573583584, -0.00525482563477547, -0.000542111786483259, -0.00397111004474954, -0.0060591072084358, -0.00958156917866185, -0.00287200651509617}, 
                                                     {0, 0, 0, -56.2499999627369, 0, 0, 1577.23214185732, 0, 118.749999981201, 118.750000012556, 0, 0, -17797.7678464217, 0, -62.5000000932079, 0, -62.4999995436563, -3204.0178574415, -3204.01785685958, -125.000000000189, 0, 0, 108108.382874787, 0, 2.79131276360277e-08, 0, -3.26016198013041e-06, 34182.0436533242, 34182.0436494111, 0, 1620.53571684957, 1620.53570598834, 6.93710340806237e-07, 1.77129314724838e-07, 3241.07142798662, 0, 0, -394522.569231034, 0, 1.63631827654589e-06, 0, 1.3335767505366e-05, -192057.291677845, -192057.291664637, 0, -5.21238062261795e-06, 0, 4.980960248767e-05, -16284.7222440964, -16284.7221747766, 0, -5.39096703376464e-06, -32569.444435289, -1.01194377323087e-05, -8.35631339827449e-06, 4.58385959428326e-06, 0, 0, 905121.527313625, 0, -5.75395367321673e-06, 0, -2.65781875631573e-05, 629687.500027199, 629687.500004512, 0, -4.36188808887297e-06, 0, -0.000148025035745824, 83940.972309901, 83940.9720957127, -7.04770534364326e-06, 1.25147422429972e-05, 167881.944395327, 0, 5.99346171720444e-05, -0.000244140959934298, 2.81186079168547e-05, 2.87919307749881e-05, 4.83105409640794e-05, 4.17228267706913e-05, 5.70316321340775e-06, -9.3050411033492e-05, 0.000106730282713955, 0, 0, -1314236.11047166, 0, 1.02537670041997e-05, 0, 3.92678229568872e-05, -1244791.66670485, -1244791.66668617, 0, 3.17308216664319e-05, 0, 0.000263046232784363, -243055.555741283, -243055.555341751, 2.97480810036834e-05, -1.12247009317226e-05, -486111.111002404, 0, -4.33464158934833e-05, 0, 0.000634330030313912, -0.000212879119714546, 0.000602635822781323, -1.78450631664489e-05, -6.63305825685037e-05, -6.93684559938809e-05, -0.000115819069409245, 6.37346410505503e-05, 0.000427217646303468, -0.0001504003493117, -4.57580237152073e-05, 5.52708239350971e-05, -0.000344437424592788, -0.000580234329425061, -0.000152864294884385, 0, 0, 1170634.92009388, 0, -1.00059953661536e-05, 0, -2.91303986488061e-05, 1458333.33336496, 1458333.33336131, 0, -4.80416702947498e-05, 0, -0.000239657815631878, 399305.555770544, 399305.555371784, -4.77695362437753e-05, 2.87064277275054e-07, 798611.111005624, 0, -2.41890721827673e-05, 0, -0.000824816735380279, 0.000381231578633172, -0.000720056384159998, -3.93950513769943e-05, 5.07933071039767e-05, 4.13797912494195e-06, 0.000116772208810137, -0.000176273010435688, -0.00085986441852382, 0, 0.000192639150316367, -0.00110256288642233, 9.69115609355125e-05, 0.000133935278874618, 0.00013547408686539, 0.000190671128909142, -6.80317936234255e-05, -0.0004894857029753, 0.00041219622793535, 0.00109400511475376, -0.000251323861400986, 0.00130656510784322, 0.000340566350015258, 0.00151465299960959, 0, 0, -582837.301331019, 0, 4.97419989211393e-06, 0, 1.46340933666556e-05, -930059.523823605, -930059.52382862, 0, 3.40254957459987e-05, 0, 0.000127927185868804, -347222.222347423, -347222.222137706, 3.44623678001401e-05, -7.84335534552417e-06, -694444.444401241, 0, 4.82266515079168e-05, 0, 0.000541264437284028, -0.000307044900171208, 0.000469225772169031, 6.03227590077838e-05, 9.11196667118937e-06, 6.50873220665364e-05, -5.04494771669728e-05, 9.06394284054984e-05, 0.000694889026893079, 0, -5.54190096207593e-05, 0, 0.000979465290635733, -0.000272464724671874, 0.000908885831672987, -3.05852602087421e-05, -0.000112951282168694, -4.72994374606172e-06, -0.000132695153462137, 8.37454721386668e-05, 0.00086711945801788, -0.000285355892937275, -0.000770886795657742, 0.000749298045058294, -0.000139402360650315, -1.76284008617719e-05, -1.66894970856067e-05, -0.00119701652430287, -0.000614424992724488, -0.0015913062899534, -0.00207591709428805, -2.51866733114691e-05, -0.000466597263208527, -0.0017178793251207, 0, 0, 124007.936455952, 0, -1.08263687599975e-06, 0, -3.34200404605068e-06, 248015.873018645, 248015.873021218, 0, -9.07678551858137e-06, 0, -2.47661232287448e-05, 124007.936536777, 124007.936490366, -9.60595996475463e-06, -8.76946325300501e-07, 248015.873010585, 0, -1.90123925110418e-05, 0, -0.000130515362032451, 9.47582747016404e-05, -0.000120752817048324, -2.57004909774885e-05, -1.0210633035608e-05, -2.34450538039149e-05, 1.17802331711228e-05, -1.5987007605883e-05, -0.000238997952723155, 0, -7.73100855742838e-06, 0, -0.000302528967183945, 0.000117977514837511, -0.000275262516029245, -9.23084435869408e-07, 3.83173872974341e-05, -3.83198129301067e-05, 4.73099374627173e-05, -5.09412661347801e-05, -0.000441931860658614, 2.94454964563588e-05, 0.000177723205449846, -0.000438915129708029, 0, 6.02217042911676e-05, -0.000388063575776536, 3.96294867677581e-05, 3.39592111462566e-05, 2.19243663040812e-05, 2.01378070255937e-05, -1.87989295255914e-05, -0.000178263554243177, 0.000305139719434209, 0.000206596092317939, 0.000578515002405071, 0.000788952267924226, -0.000377084329521124, -0.000221936147584026, 0.000699313054109568, 5.92602978493931e-05, 0.000691013913299968, 0.000889204564058105, 0.00133927478606207, 0.000646706470595104}, 
                                                     {0, 0, 0, 11.111111103374, 0, 0, -313.095237887681, 0, -11.1111111109182, -11.1111111126084, 0, 0, 3557.16490077269, 0, 1.07460732154457e-08, 0, -2.0032172336894e-08, 301.984127017574, 301.984126990834, 1.07097401241758e-08, 0, 0, -21796.8474298678, 0, -1.50408611831759e-08, 0, 1.88274078120706e-07, -3255.18077628607, -3255.18077616547, 0, -2.56937711361893e-07, 4.69255922033417e-07, -2.4781397975968e-07, 3.85283331497686e-08, -1.16923349896995e-07, 0, 0, 80405.092548284, 0, -1.06351029728146e-07, 0, -1.3012941363322e-06, 18541.6666678996, 18541.6666676157, 0, 4.71761273749961e-07, 0, -2.50999273380616e-06, 2.0617226020893e-06, -1.87447165880618e-06, 7.56911424438605e-07, 5.78577148568807e-07, 2.57436979309048e-07, 1.62804899186853e-06, 1.96123657278202e-06, 1.13903217020653e-06, 0, 0, -186863.425829543, 0, 3.81843991614572e-07, 0, 2.26999833621783e-06, -61863.4259289983, -61863.4259287987, 0, 2.84389939001056e-07, 0, 8.67078054320566e-06, -8.33109246743667e-06, 7.33329673149611e-07, -1.95894718693801e-06, 6.94865598574588e-07, 1.62832690594705e-06, 0, -5.18235367152245e-06, 1.04020414505308e-05, -6.03602097934235e-06, -6.76358090273593e-06, -2.32107902889173e-06, -7.53933813546517e-06, -6.24679762072708e-06, -4.68490850730467e-07, -1.85835929469274e-05, 0, 0, 275462.962830157, 0, -7.35835849935596e-07, 0, -4.80507401289501e-06, 125000.000004395, 125000.000006026, 0, -2.26093847373053e-06, 0, -1.71650959283316e-05, 1.83295748079406e-05, -4.32990773674075e-07, 1.96182938619478e-06, -2.24797891755345e-06, -6.00128673697085e-06, 0, 2.38408781211647e-06, 0, -3.01642924506829e-05, 1.77740268822466e-05, -3.0502467144573e-05, 1.16552296989079e-05, 1.92101504782473e-05, -1.47549373783215e-06, 1.68538211735835e-05, 8.19180599645417e-06, -4.1978399092557e-06, 1.7983590063828e-05, 1.82055019800106e-05, 1.25229449180721e-05, 4.35542947196659e-05, 7.30408075808224e-05, 4.47253364444306e-05, 0, 0, -249669.312056927, 0, 7.53792360596214e-07, 0, 2.91994859310865e-06, -150462.962966704, -150462.962968709, 0, 3.13542385793064e-06, 0, 1.53730024732545e-05, -2.22869159209582e-05, -4.14309076103391e-06, -9.49678854699575e-07, 5.56239174647056e-06, 5.66247427293801e-06, 0, 3.61869600432247e-06, 0, 4.65077784369208e-05, -3.44401870567687e-05, 2.80037873324835e-05, -1.49082854327093e-05, -2.02499236352802e-05, 1.28825397312971e-05, -4.95258360531707e-06, 4.41742276937346e-06, 3.89854385844901e-05, 0, -1.31619147032349e-05, 5.66881953911558e-05, -1.07205683334562e-05, -2.23537652151618e-05, 6.96059614617596e-06, -3.91733371079057e-05, -1.99810077417961e-05, -3.56175078263083e-05, -6.69187007331844e-05, -9.22198964160509e-05, -0.000111569284530114, -0.000132753196125308, -6.09339162021442e-05, -0.000185894359029955, 0, 0, 126763.668377093, 0, -3.68865898559701e-07, 0, -1.87768904367755e-06, 99206.3492080675, 99206.349209577, 0, -2.30421725849556e-06, 0, -9.95876785664925e-06, 1.33687615291003e-05, 3.0627963095294e-06, 3.62424687581562e-08, -1.89011359250692e-06, -3.06224640419593e-06, 0, -4.35589242123893e-06, 0, -3.31820346495947e-05, 2.84057389687808e-05, -2.11088609482488e-05, 1.01509896197191e-05, 9.60449688107169e-06, -1.85495114163043e-05, 8.91116619447478e-07, -1.48073238545449e-06, -2.08693375198609e-05, 0, 1.92476357833047e-06, 0, -5.64420091135503e-05, 2.17286006528458e-05, -5.0431634907335e-05, 1.60271954439492e-05, 2.12823959273061e-05, -1.89456466118179e-05, 3.24309018606559e-06, 2.23216983836915e-05, -3.13023430835753e-05, 9.11662319000883e-05, 4.83162903171799e-05, 1.59281355087463e-05, 1.46096751575565e-06, 2.08819456895565e-05, 5.60162894952548e-05, 0.000152105761704828, 9.90435403279125e-05, 0.000149200791735059, 0.000208293443950682, 0.000125853489429147, 0.000130678682784408, 9.86856631715245e-05, 0, 0, -27557.319213185, 0, 8.73614472869874e-08, 0, 4.7204468457369e-07, -27557.3192243411, -27557.3192248382, 0, 5.87373959814631e-07, 0, 1.43324402076333e-06, -3.18275392245411e-06, -6.80284882227381e-07, 1.33823771801678e-07, 1.15112211248639e-06, 4.44554873017545e-07, 0, 1.47961983232995e-06, 0, 6.37941104352361e-06, -9.01947385570108e-06, 2.88840963942691e-06, -2.31064962671134e-06, -2.17760560067928e-06, 4.35135060000309e-06, -9.31057812694938e-07, 1.80408286745984e-07, 1.10385111150712e-05, 0, 1.55110762121835e-06, 0, 1.06501278586594e-05, -9.66140763449991e-06, 1.21444699688622e-05, -8.4712219816725e-06, -8.70734461944662e-06, 7.06857452627189e-06, -1.37407426221988e-05, 6.5614878329454e-07, 1.96454012249954e-05, -1.37318231064611e-05, -7.71156728881161e-06, 1.548613208865e-05, 0, -3.82167494372371e-06, 1.99590302347528e-05, -2.58292243236805e-06, -5.38242102287779e-06, 4.49262670189587e-06, -1.33810804496279e-05, -8.69031136038842e-06, 9.39328539695948e-06, -3.09004569072506e-05, -1.65870201554024e-05, -3.71719421632515e-05, -5.49414486103122e-05, 2.04121097866459e-05, -1.82100538738307e-05, -6.40856531803237e-05, -1.89522225117782e-05, -7.79715898979297e-05, -0.000116246590607532, -0.000125168942767102, -0.000116326184053379}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 0, 11.1111111109198, 0, 0, 0, 0, -301.984126981703, 0, 0, -1.40385620373921e-08, 0, 0, 0, 0, 0, 0, 3255.18077599713, 0, 0, 2.22795974126787e-08, 0, 0, 3.85919036542207e-07, 0, 0, 0, 0, 0, 0, 0, 0, -18541.6666663437, 0, 0, 1.30216763280344e-07, 0, 0, -4.34685962780179e-06, 0, 0, -9.32862387427166e-07, 0, 0, -1.10819635149998e-08, 0, 0, -1.28162829097265e-07, -4.40527120378418e-08, 0, 0, 0, 0, 61863.4259239161, 0, 0, -2.47065640106686e-07, 0, 0, 1.9470916217617e-05, 0, 0, -3.53465206719769e-06, 0, 1.09256134667471e-07, -4.64640918492797e-08, 3.4076928263603e-08, 0, 1.73996246013055e-05, 0, 1.42517907859156e-07, -1.88098957418645e-07, -1.05152286516848e-07, 0, 2.81049635024142e-07, -4.74733702528499e-07, 1.50548520966294e-07, 0, 0, 0, 0, -124999.999994541, 0, 0, -4.94841591015285e-07, 0, 0, -3.75700378659379e-05, 0, 0, 1.49527336661931e-05, 0, -3.74834917598869e-07, 5.18882638078749e-08, 4.27716149495987e-08, 0, -8.38309932557146e-05, 0, 0, -7.93203331797154e-06, 0, -3.07431645946114e-08, -1.22481359169466e-06, 5.2275844955948e-07, 2.95931880958254e-07, -1.16312782076147e-06, 1.67646394953593e-07, 1.31834673514521e-06, 1.92967563341682e-06, -3.8235468118034e-08, -7.03629865158759e-07, 1.04200636473754e-06, 1.5718963494697e-06, 0, 0, 0, 0, 150462.962955791, 0, 0, 1.65480197199908e-06, 0, 0, 3.37322848764643e-05, 0, 0, -1.70597949979389e-05, 0, 1.99153233957034e-07, -1.92685163483392e-08, -2.76920135454555e-07, 0, 0.000154119707513181, 0, 0, -3.09252360165959e-05, 0, 1.85479635663507e-07, 8.13390451482986e-07, 7.99581953175642e-08, -4.32753914154996e-07, 2.76032778835436e-07, 2.14756642818239e-07, 0, 9.80202217976559e-05, 0, -7.23107883221691e-07, 4.15791931414311e-07, -9.48986400606226e-07, 1.11910234202509e-06, -2.50757006814898e-06, 4.22751273189282e-07, -2.47349936012198e-06, -1.34627413893645e-06, 6.75739874420288e-07, 6.53849525682788e-07, 5.88887742306474e-06, 3.52534482276875e-06, 0, 0, 0, 0, -99206.3492019131, 0, 0, -1.45348169336827e-06, 0, 0, -1.19628705164132e-05, 0, 0, 8.59085132423817e-06, 0, -1.13998351502015e-07, -6.22209598347828e-08, -1.35205087168416e-07, 0, -0.000119461365526729, 0, 0, 4.08886878491915e-05, 0, 2.03287225081024e-07, 1.53884255903104e-07, -3.98505874568332e-07, 2.69078545275852e-07, -5.85300638710184e-07, 2.16915107907818e-07, 0, -0.000158436343320721, 0, 0, -1.90580865859565e-05, 0, 2.35569050861835e-08, 5.70396735157124e-07, 8.41240243135836e-07, -4.41002356975203e-07, -5.726785225195e-07, -3.57931116635998e-07, -1.75454713907144e-06, 8.89210865520691e-07, 1.6200500189542e-06, -1.18080456484536e-06, -9.60742691748321e-07, -7.93410283141849e-07, -3.83532279833839e-06, 6.64022150286408e-06, -6.36168814988668e-08, -6.8880636937783e-07, 2.97495557012077e-06, 9.30684709951188e-07, 5.11523518069823e-06, 0, 0, 0, 0, 27557.3192229041, 0, 0, 4.34754084312758e-07, 0, 0, 6.82000589335985e-07, 0, 0, -1.26618760131927e-06, 0, 2.38145427497837e-08, -1.50159724437432e-08, 0, 0, 3.29159764835927e-05, 0, 0, -1.40182541959803e-05, 0, -3.11288504232036e-08, -1.26918419718606e-07, 4.65836819578178e-08, 3.26997525309749e-08, -1.25835990712943e-07, 0, 0, 7.17729402852348e-05, 0, 0, -1.34133498939343e-05, 0, -2.53828377099971e-07, -1.53536156243877e-06, -1.09877461243425e-07, 9.90750435518778e-08, -4.83890587717081e-07, -1.62725745875178e-07, 1.43999336638816e-06, -2.23603568697155e-07, 1.21507857304461e-07, 0, 4.15771093711153e-05, 0, -4.91551092699495e-08, -1.69808107699311e-06, -5.28235979580213e-08, 1.92711333022234e-07, -6.17054497766777e-07, -5.09383407651428e-07, 1.60560323657077e-06, -6.69965514413809e-08, -2.14092971402557e-06, 6.22832583246491e-07, 2.86787530394217e-06, 8.19393963115273e-07, -2.32592333563944e-06, -1.08166015297365e-06, -2.25495690524774e-06, 2.94206157163353e-06, -2.02042870535742e-06, 1.44318257302461e-06}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 0, 6.25000000434661, 0, 0, 0, 0, -162.053571459615, 0, 0, -62.5000001092209, 0, 0, 0, 0, 0, 0, 1628.47222225732, 0, 0, 1.0312833088834e-06, 0, 0, 1620.53571521361, 0, 0, 0, -2.06860092142244e-08, 0, 0, 0, 0, -8394.09722200663, 0, 0, -4.42730514058015e-06, 0, 0, -16284.7222255095, 0, 0, -7.1818077803703e-06, 0, -6.83232082848884e-08, 2.05475474618031e-08, 0, -1.09262790426178e-07, 1.93151438016806e-07, 1.6366721919805e-07, 0, 0, 0, 0, 24305.5555539399, 0, 0, 1.15448938919119e-05, 0, 0, 83940.9722290304, 0, 0, 2.73436302165153e-05, 0, -4.6350726913583e-07, 2.52414525048101e-07, 1.23520619145669e-07, 0, 2.27720415602395e-05, 0, -2.46190169900567e-08, 1.19711711532823e-06, 4.49143447733996e-07, 3.6065501434042e-07, -2.26449055342355e-06, 1.22186526573572e-06, -1.55192484136156e-06, 0, 0, 0, 0, -39930.5555517596, 0, 0, -1.80808812537126e-05, 0, 0, -243055.555565447, 0, 0, -6.06918802631352e-05, 0, 1.90522698250815e-06, -2.88652214467536e-07, 6.20322423289638e-08, 0, -4.03185395584258e-05, 0, 0, -6.46593296303888e-05, 0, 3.88918987911556e-07, 4.26777027174402e-06, -2.55518095780767e-06, -4.12699543503344e-07, -3.5772634973277e-07, 2.50138501782299e-06, -6.87536905695345e-06, -6.85440179557227e-06, 2.00180695022889e-06, 9.23654235617287e-07, -8.57711809338567e-06, -1.03586391310725e-05, 0, 0, 0, 0, 34722.222217847, 0, 0, 1.67213512496501e-05, 0, 0, 399305.55556103, 0, 0, 6.53432305406047e-05, 0, -9.32551150865618e-07, 7.53337337675257e-08, 2.3289858296031e-06, 0, 5.9123234675533e-05, 0, 0, 8.64507028466884e-05, 0, -1.2385241145634e-06, -2.50237641952204e-06, -3.91914089590281e-07, 2.45915735540667e-06, 1.00102896374437e-06, -8.69678471889671e-07, 0, 7.58242177007844e-05, 0, 4.0583700626855e-06, -7.46602481190369e-06, 4.88175626041207e-06, -3.41756020058685e-06, 1.90281261987432e-06, -9.0943283164965e-06, 7.92766634057789e-06, 9.41545369472548e-06, -4.44193620401386e-06, -1.86438776459426e-06, -3.29119557497256e-05, -1.00903217415369e-05, 0, 0, 0, 0, -12400.7936479099, 0, 0, -8.03289706109034e-06, 0, 0, -347222.222222869, 0, 0, -4.59906908424597e-05, 0, 5.370715454369e-07, 2.85268608826953e-07, 3.82402420551578e-07, 0, -3.93090652217301e-05, 0, 0, -6.68411034696218e-05, 0, -7.05626421057235e-07, 6.93359028490958e-07, 1.67923061899329e-06, -1.28906915549139e-06, 4.10738791552646e-06, 2.1164900908505e-06, 0, -6.29100151655947e-05, 0, 0, -6.61671233375845e-05, 0, -4.75715363428707e-07, 4.20664296022697e-06, -4.13900904421598e-06, -4.48482896510471e-06, 1.43022215860909e-05, -8.14672494857041e-06, 7.35426523171175e-06, -6.11181888450488e-06, -1.82024084072599e-05, 4.33132236272476e-06, 6.08083176643214e-06, 5.20316158653362e-06, 1.17967984191414e-05, -4.04904666576393e-05, 1.25548443443921e-05, -3.33223324980499e-06, -1.32878515836304e-05, 4.41686735951199e-07, -1.95909262966089e-05, 0, 0, 0, 0, -6.57238409578648e-07, 0, 0, 1.67305788119428e-06, 0, 0, 124007.936507117, 0, 0, 1.16021957559768e-05, 0, -1.13701024326403e-07, 6.94895312918591e-08, -3.34180050099307e-07, 0, 9.90210495218637e-06, 0, 0, 2.43870973094726e-05, 0, 8.70312044267163e-08, 1.00447707817105e-06, -2.07968768707732e-07, -4.84926119296686e-08, 5.38835390552619e-07, 8.5893486126623e-08, 0, 2.44999636490295e-05, 0, 0, 2.2585286893563e-05, 0, 1.0299951413198e-06, 6.69368576353917e-06, 5.45311365681192e-07, -4.86764334312995e-07, 1.21313809719198e-06, 2.1293064920673e-06, -6.66828020762423e-06, 2.22830509156381e-06, -5.33843270630535e-06, 0, 2.50649218884037e-05, 0, 5.48644615640139e-08, 1.17853782361381e-05, 2.05585905237479e-07, -2.39050717496593e-06, 2.08392437879935e-06, 1.37347520593132e-06, -4.82482820703597e-06, 7.14597711100465e-07, 9.88623778393255e-06, 6.08065255781344e-07, -1.36858934187634e-05, -6.67648565637428e-06, 8.94465539671127e-06, -9.36193227995559e-06, 1.33249804191157e-05, -1.53918345397937e-05, 9.25653029028783e-06, 6.73508781257902e-06}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 0, 4.7619047578094, 0, 0, 0, 0, -116.666666607577, 0, 0, -71.4285713507571, 0, 0, 0, 0, 0, 0, 1074.07407376367, 0, 0, 238.095237307327, 0, 0, 1749.99999886641, 0, 0, 0, 4.20348015673094e-08, 0, 0, 0, 0, -4861.1111100134, 0, 0, 2.78168055530376e-06, 0, 0, -16111.1111048983, 0, 0, -5833.33332551168, 0, 2.26695538686314e-07, -4.47636309584524e-08, 1.22131508737527e-08, 2.51548860213903e-07, -5.41868801035962e-07, -5.28575287921168e-07, 0, 0, 0, 0, 11574.0740727345, 0, 0, -7.00303206248557e-06, 0, 0, 72916.6666517595, 0, 0, -2.7703855785839e-05, 0, 1.00529669475427e-06, -6.64349208641799e-07, -7.86830165495169e-07, 0, 53703.7036660443, 0, -1.5583765900311e-07, -2.99393184154205e-06, -1.23308038762233e-06, -4.34421141499746e-07, 5.31116327425207e-06, -2.4726953194397e-06, 3.78140135602601e-06, 0, 0, 0, 0, -13888.888887873, 0, 0, 1.13707759531427e-05, 0, 0, -173611.111087963, 0, 0, 5.82231495619626e-05, 0, -4.72080746315453e-06, 7.01129196288713e-07, 1.09560113499813e-07, 0, -243055.555480803, 0, 0, 0.000111542489280232, 0, -2.4288629802003e-07, -1.05085860069596e-05, 6.46116453121941e-06, 9.957660578107e-07, -4.08765208015337e-06, -5.61849993646021e-06, 1.77027673261297e-05, 2.19894631235639e-05, -8.54657680210684e-06, -3.71071389827995e-06, 1.885694929449e-05, 3.51879147185421e-05, 0, 0, 0, 0, 6613.75661293845, 0, 0, -1.11215645185459e-05, 0, 0, 208333.333315308, 0, 0, -4.40590177501742e-05, 0, 2.28110642714336e-06, -1.81802917577266e-07, -4.14961291769357e-06, 0, 578703.703604006, 0, 0, -0.000126909566050442, 0, 2.25586639721933e-06, 8.08732265484682e-06, 2.6035828397486e-07, -6.00543133035643e-06, -7.59045348589386e-06, 3.47417252904443e-06, 0, -0.000165472466917992, 0, -1.01585350621052e-05, 1.65470706811589e-05, -1.24310048745988e-05, 6.61885081225705e-06, -1.55637144223895e-06, 2.00250635842822e-05, -1.93244178016292e-05, -2.6754273003494e-05, 1.34238646837485e-06, 1.13468928856671e-05, 9.26654324267574e-05, 3.10896299111615e-05, 0, 0, 0, 0, -3.43644689357806e-07, 0, 0, 5.18116246166528e-06, 0, 0, -99206.3491953178, 0, 0, 3.60559552702662e-05, 0, -1.2193902067816e-06, -7.78226131046383e-07, -1.70695867286141e-06, 0, -694444.444391436, 0, 0, 8.21041977186929e-05, 0, 2.0474361325793e-06, 9.34634477764585e-07, -4.36193294430997e-06, 3.03912317423635e-06, -1.04971139534661e-05, -3.87251243710813e-06, 0, 0.000126606792820966, 0, 0, 0.000157382869444899, 0, 1.89823790527258e-06, -1.76450066467412e-07, 1.11834162045674e-05, 6.76434236966728e-06, -3.70603839169861e-05, 1.62875199248047e-05, -2.23535166928259e-05, 1.27317394837806e-05, 2.47325991735475e-05, -1.01676114120246e-05, -3.12589921941458e-06, -1.20771042636968e-05, -3.28167419750943e-05, 0.000107055141296502, -2.537985282852e-05, 5.98575249596107e-06, 4.91825875453915e-05, 2.05993991565329e-05, 5.1245431606896e-05, 0, 0, 0, 0, 1.99502714685081e-08, 0, 0, -1.11669878427107e-06, 0, 0, -2.41970729258126e-06, 0, 0, -8.47434259880745e-06, 0, 2.92793947749825e-07, -1.83510417789087e-07, 9.14010632484674e-07, 0, 330687.830676706, 0, 0, -2.82759424244319e-05, 0, -3.31307510783277e-07, -2.64641736753111e-06, 5.70121517222765e-07, 1.77732013377558e-07, -7.48807950466193e-07, -1.97642820446798e-07, 0, -3.97803722859884e-05, 0, 0, -5.25055592739187e-05, 0, -2.71450207045616e-06, -1.86593765359115e-05, -1.38586626466705e-06, 1.77939683761403e-06, -5.48357989201882e-06, -5.43851943399535e-06, 1.880186899305e-05, -4.26124416532711e-06, 1.39245095586968e-05, 0, -6.65610377752015e-05, 0, -3.43365299796571e-07, -3.03288706678537e-05, -4.89216969785673e-07, 6.69115612287421e-06, -6.60276930420619e-06, -3.52724031246634e-06, 1.15273967201825e-05, -1.08384908646147e-06, -2.12216682784257e-05, 1.11075155569708e-06, 2.62815140044491e-05, 1.35287411203831e-05, -2.4223411375811e-05, 1.66868861027729e-05, -3.03838580018427e-05, 3.16227834399722e-05, -2.48102935046917e-05, -5.979928539945e-06}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 0, 4.16666666139724, 0, 0, 0, 0, -95.138888826988, 0, 0, -76.3888887519912, 0, 0, 0, 0, 0, 0, 781.249999642892, 0, 0, 416.66666578338, 0, 0, 1744.2129616349, 0, 1.24266127343428e-08, 0, -1.15809820928203e-08, 0, 0, 0, 0, -2951.38888830459, 0, 0, -694.444439835621, 0, 0, -14322.9166618193, 0, 0, -9513.88887886923, 0, -4.74554094546147e-07, 1.0123142613729e-08, -2.5629563808949e-07, -4.54912998829628e-07, 8.31034348509405e-07, 8.46882743240178e-07, 0, 0, 0, 0, 5208.33333126682, 0, 0, -1.07881088138983e-05, 0, 0, 54108.7962865526, 0, 0, 15856.4814491535, 0, -1.39720918872622e-06, 1.28089205142714e-06, 2.74713531450219e-06, 0, 78124.9999602617, 0, 9.28941953427155e-07, 5.19765271634008e-06, 2.53181586726412e-06, 3.14417573178426e-07, -6.74798426268948e-06, 3.86730726796159e-06, -4.84784927799551e-06, 0, 0, 0, 0, -3472.22221846949, 0, 0, 1.41217558499142e-05, 0, 0, -95486.1111079962, 0, 0, 4.77078886450819e-05, 0, 7.79071080245806e-06, -1.25555473278544e-06, -2.00030995866126e-06, 0, -295138.888813051, 0, 0, -130208.333231184, 0, -1.04898942408854e-06, 1.70387429543359e-05, -1.10705428952345e-05, -2.39201691984095e-06, 1.4119501044924e-05, 5.69206670024548e-06, -3.01157826108916e-05, -4.57053591679274e-05, 1.84952742358648e-05, 7.28465218001088e-06, -2.81263523665232e-05, -7.65394950257658e-05, 0, 0, 0, 0, -3.86191771584354e-06, 0, 0, -1.14216529859907e-05, 0, 0, 63657.4074097345, 0, 0, -7.24794018447888e-05, 0, -3.9309007548957e-06, 4.69621618834543e-07, 5.33084577397983e-06, 0, 520833.333299298, 0, 0, -0.000190824733647483, 0, -2.75599105430568e-06, -1.84985568135111e-05, 1.85306895138411e-06, 1.02400850539934e-05, 2.63336691156961e-05, -4.37560929942977e-06, 0, 491898.147957646, 0, 1.7035013516687e-05, -2.12246890761112e-05, 2.16820407597349e-05, -4.85473546551688e-06, -2.79161508458237e-06, -2.41093760285093e-05, 3.30023982574995e-05, 5.28136622110995e-05, 2.21738829730533e-05, -2.5704544325965e-05, -0.000160991602633425, -5.50116887496408e-05, 0, 0, 0, 0, 3.64784814260754e-06, 0, 0, 6.47174356060504e-06, 0, 0, -8.10230225644041e-06, 0, 0, 2.02204609062032e-05, 0, 1.81820163928302e-06, 1.31430259314525e-06, 3.17201630042734e-06, 0, -347222.222215863, 0, 0, 0.000129145352618843, 0, -3.30411429800268e-06, -4.76854074803528e-06, 6.89712072275196e-06, -5.08372388057303e-06, 1.44465640816229e-05, 2.06438071526386e-06, 0, -868055.555425721, 0, 0, 0.000197148886691996, 0, -4.43636061222179e-06, -1.77107908916219e-05, -2.02397230727742e-05, -4.57371688432813e-06, 5.60631634482237e-05, -2.19738518767595e-05, 4.42919061430995e-05, -1.76121205699821e-05, -6.73757038863318e-06, 1.46873321576607e-05, -2.05135283912715e-05, 1.68951289694323e-05, 5.60048224013982e-05, -0.000184943509438899, 2.48141373413652e-05, -9.40381609215874e-06, -0.000118896195877614, -7.91773105828022e-05, -9.7746631580903e-05, 0, 0, 0, 0, -8.01830561038703e-07, 0, 0, -1.52190498195916e-06, 0, 0, 1.88087135549436e-06, 0, 0, -3.8028318062073e-06, 0, -4.83181510042975e-07, 3.28094805620127e-07, -1.34930256574181e-06, 0, 1.91977496179679e-06, 0, 0, -2.28245909244849e-05, 0, 5.47299941450901e-07, 3.92517461153511e-06, -9.22921713007251e-07, -3.06160354510826e-07, 8.34104424828523e-07, 8.60124593077021e-07, 0, 578703.703680817, 0, 0, -6.86394085035434e-05, 0, 4.40368237828201e-06, 3.29002281119436e-05, 2.43039900729127e-06, -3.73382702856142e-06, 1.4888087723894e-05, 9.86083360388968e-06, -3.45703296256534e-05, 5.24885846882191e-06, -2.24309445383436e-05, 0, -6.66273714413952e-05, 0, 1.45649345979194e-06, 5.0960420484599e-05, 1.00916611248845e-06, -1.14097545847796e-05, 1.37916478494152e-05, 6.77779627801993e-06, -1.73527098518603e-05, 1.08887331051171e-06, 2.84695276955744e-05, -4.47199682365478e-06, -2.40950473677565e-05, -1.28758337851566e-05, 4.42028481390085e-05, -1.03813421002925e-05, 4.79742028287843e-05, -3.22927446029372e-05, 5.00272149018399e-05, -8.21150475607152e-06}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 0, 3.99999999808213, 0, 0, 0, 0, -83.3333332853005, 0, 0, -83.3333333670697, 0, 0, 0, 0, 0, 0, 583.333332895859, 0, 0, 583.333333361934, 0, 0, 1736.11111142615, 0, 7.88251490338191e-08, 7.14848070728186e-08, 4.7686225626719e-08, 0, 0, 0, 0, -1666.66666371831, 0, 0, -1666.66666817354, 0, 0, -12152.7777778297, 0, 0, -12152.7777803111, 0, 1.93964263870019e-07, -1.36174582346515e-07, 0, 0, -1.9154869636547e-06, -1.44826130368201e-06, 0, 0, 0, 0, 1666.66665899675, 0, 0, 1666.66667103157, 0, 0, 34722.2222232691, 0, 0, 34722.2222300994, 0, 2.3267494292739e-06, -1.24714751349416e-06, -3.07338585913863e-06, 0, 85069.4444560512, 0, 1.98305314416055e-07, -2.23789586640157e-06, -1.8796285494982e-06, 1.74830180766482e-06, 9.31453715460172e-06, -1.80120568944648e-06, 8.05786467096946e-06, 0, 0, 0, 0, 1.40870901808686e-05, 0, 0, -5.51258677582975e-06, 0, 0, -34722.2222177516, 0, 0, -34722.2222354637, 0, -1.02782410002637e-05, 9.65731863963234e-07, 1.57801619815078e-06, 0, -243055.555581565, 0, 0, -243055.555559532, 0, -1.13006072256434e-06, -2.63413518738841e-05, 1.0205987759073e-05, 8.89737781108871e-07, -3.54879038037306e-05, -6.74806650415002e-06, 2.96204859122241e-05, 4.86149059877662e-05, -3.12372996811898e-05, -2.4442666513857e-05, 1.72346247808695e-05, 9.19473188731246e-05, 0, 0, 0, 0, -1.56268417203905e-05, 0, 0, 4.25375437682649e-06, 0, 0, -7.84550019104365e-06, 0, 0, 5.42227044834516e-05, 0, 6.06125140939117e-06, -3.09372047696979e-07, -1.16334565278186e-06, 0, 243055.55556544, 0, 0, 243055.555620826, 0, 6.38006960291035e-06, 3.53487223761222e-05, -2.4950799180658e-06, -9.02972712269454e-06, -3.48667641764402e-05, 8.64093697227868e-06, 0, 694444.444483468, 0, -1.17486328509993e-05, 3.44280574637178e-05, -1.95960527363391e-05, 4.60325856094971e-06, 3.35248037914101e-05, 2.41085518205395e-05, -1.88457224669466e-05, -5.68102554362514e-05, -3.06901469471554e-05, 6.07290112716512e-05, 0.000222518162145775, 9.26687543196223e-05, 0, 0, 0, 0, 7.68542566414702e-06, 0, 0, -2.77855503118855e-06, 0, 0, 1.50423510685763e-05, 0, 0, -1.97836072121382e-05, 0, -2.49203860054793e-06, -1.69416759517025e-06, -5.76595515712026e-06, 0, -1.45616762458739e-05, 0, 0, -5.46802433395358e-05, 0, 1.20782507743539e-06, 5.02546682412516e-06, -8.52324275778643e-06, 4.36107482117762e-06, -2.00941132697896e-05, -9.87010398653947e-07, 0, -694444.4444959, 0, 0, -694444.444481833, 0, 1.26093808981482e-06, 2.85102334605073e-05, 2.07791122762102e-05, -7.20534552540058e-06, -7.10081220990401e-05, 1.17300975658592e-05, -7.30299994362177e-05, 6.47019629396459e-06, -5.43619406685367e-05, -2.00079888190275e-05, 3.36636442478778e-05, -2.26377437176828e-05, -8.76591049873383e-05, 0.000179370239372728, -2.36198133709031e-05, -1.12035686748092e-05, 0.000145240758740282, 0.000118177953304776, 9.18816907045665e-05, 0, 0, 0, 0, -1.98668675472219e-06, 0, 0, 7.58143913425967e-07, 0, 0, -4.77399957184225e-06, 0, 0, 5.47895708034617e-06, 0, 7.12136958429385e-07, -3.57059122540233e-07, 1.74816081957363e-06, 0, 3.89793841608897e-06, 0, 0, 9.48051786292118e-06, 0, -6.74576897354852e-08, -2.86650695910253e-06, 1.32351551506727e-06, 6.84569403331741e-07, 1.55486875845526e-06, -6.08805980301271e-07, 0, 1.30681718443912e-05, 0, 0, 9.78618781634883e-06, 0, -3.24598927577844e-06, -3.54873818508205e-05, -1.80054331897158e-06, 6.32611021522102e-06, -1.84067863824042e-05, -9.90450496923801e-06, 4.6787723584834e-05, -1.55230376732091e-06, 3.04294657400769e-05, 0, 694444.444455305, 0, -6.61235714002814e-07, -5.01240091164988e-05, 6.70733668409158e-07, 1.51845893721749e-05, -1.04057419718243e-05, -6.20001635209284e-06, 2.61319158304914e-05, 1.13621506971916e-05, -1.71813846523441e-05, 1.45661228230318e-05, 1.74594136646597e-05, 1.23562271172046e-05, -4.49235290388478e-05, 5.7290977137449e-06, -4.14378083398951e-05, 2.74083297481993e-05, -5.38276635385534e-05, 5.2456893679088e-05}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 0, 4.16666666687516, 0, 0, 0, 0, -76.3888888714926, 0, 0, -95.1388889073838, 0, 0, 0, 0, 0, 0, 416.666666358634, 0, 0, 781.250000473947, 0, 0, 1744.21296317302, 0, -1.18354063045297e-08, -4.20860698843954e-08, 6.98118936142634e-08, 0, 0, 0, 0, -694.444443179209, 0, 0, -2951.38889008662, 0, 0, -9513.88889081863, 0, 0, -14322.9166697969, 0, -4.99566251843559e-07, -1.20573750546404e-08, -4.50890210153537e-07, -2.40700719259698e-07, 1.00255584085047e-06, 8.44941978323097e-07, 0, 0, 0, 0, -5.35610297650303e-06, 0, 0, 5208.33333667272, 0, 0, 15856.4814840468, 0, 0, 54108.7963096984, 0, -6.74338925379564e-07, 1.26070530555261e-06, 4.24296680178612e-06, 0, 78125.0000092509, 0, 1.49763257557941e-06, 4.08074088285435e-06, 2.50004029959472e-06, -1.20397108591859e-06, -2.99869951791303e-06, 2.11976288007e-06, -2.79235060197026e-06, 0, 0, 0, 0, 9.79324244683455e-06, 0, 0, -3472.22222965506, 0, 0, 8.05461637257451e-07, 0, 0, -95486.1111431162, 0, 6.37811314000066e-06, -9.83033592105546e-07, -3.83207049453645e-06, 0, -130208.333370169, 0, 0, -295138.888935859, 0, -2.95318389524224e-06, 1.51075106358394e-05, -9.07299203087377e-06, -2.31663140519197e-06, 3.2674008021095e-05, -1.03855063242483e-06, -2.50447141165797e-05, -5.27207981429706e-05, 2.68664031188788e-05, 1.55492575991498e-05, -1.26314365496297e-05, -9.34191531382451e-05, 0, 0, 0, 0, -9.93284725232271e-06, 0, 0, 7.92425254431401e-06, 0, 0, -1.48404271131914e-05, 0, 0, 63657.4074252003, 0, -3.45823732649389e-06, 5.52962620113804e-07, -1.46708651734665e-07, 0, 6.63216474267235e-05, 0, 0, 520833.333348427, 0, -9.10130935559894e-07, -2.64738238226367e-05, 5.27803914621807e-06, 7.69584557933368e-06, 4.62803093931087e-05, -1.86441229143203e-06, 0, 491898.14822168, 0, 1.16770543722429e-05, -6.32143456288727e-06, 1.82743788945437e-05, 4.54791454324885e-06, -2.32255552607765e-05, -5.74235082215279e-06, 2.83704946430948e-05, 5.59705197876927e-05, 5.8559538069338e-05, -4.58856005688534e-05, -0.000164095964903684, -7.03003712432467e-05, 0, 0, 0, 0, 6.47857900599429e-06, 0, 0, -3.55482629082049e-06, 0, 0, 1.08909810866511e-05, 0, 0, -2.600124488133e-05, 0, 9.11696662030832e-07, 1.2564251037458e-06, 4.71020253270133e-06, 0, -2.95683105991907e-05, 0, 0, -347222.222238012, 0, -2.99750627004052e-06, -1.1602260064658e-05, 5.55416565155163e-06, -3.69829377536326e-06, 9.46118434968814e-06, -5.37024259079333e-06, 0, -6.54695420952232e-05, 0, 0, -868055.555593881, 0, -4.96643085274362e-06, -5.16386799332587e-05, -1.88603880749133e-05, 1.17246656474312e-05, 3.71071805752439e-05, -5.41705363878313e-06, 5.37657218414644e-05, -2.30598123408331e-06, 7.26416938685226e-05, 1.09623003255238e-05, -6.53200239676926e-05, 1.25354959926673e-05, 5.86394056547114e-05, -0.000155392316928121, -1.03165602910224e-05, 5.27733575066219e-06, -0.000168599653325693, -0.00014828529267941, -0.000104268643236154, 0, 0, 0, 0, -1.33678026661883e-06, 0, 0, 4.37551286990659e-07, 0, 0, -4.4576946734082e-06, 0, 0, 8.02161782251637e-06, 0, -3.86359749737151e-07, 3.05333343611458e-07, -1.18047487235882e-06, 0, 3.15420089527732e-06, 0, 0, 1.31920464928882e-05, 0, 6.15555691608259e-07, 2.79337423592237e-06, -9.45321918992272e-07, -4.88242928635228e-07, -9.53276873535909e-07, 1.42433328962178e-06, 0, 2.84191554358181e-05, 0, 0, 578703.703717606, 0, 3.21558291061689e-06, 3.24167855338805e-05, 1.74362372269296e-06, -5.57742971986394e-06, 2.273132979221e-05, 8.95547604789087e-06, -3.59950957601932e-05, -2.8400718181701e-07, -2.09599874424167e-05, 0, 2.82584697300717e-05, 0, 2.56134013233756e-06, 4.35063533832577e-05, 1.74700840543006e-07, -1.12624824071198e-05, 1.56231674564925e-05, 6.71512700109073e-06, -1.49842864262968e-05, -2.25807086241952e-06, 7.33933307678171e-06, -1.23668377633536e-05, 1.17055280055982e-05, 4.18818734630641e-06, 4.18259481859597e-05, 2.5784478132395e-05, 3.30667589348113e-05, 1.11971958076855e-05, 5.77630945682707e-05, -5.41168198060454e-05}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 0, 4.76190476100187, 0, 0, 0, 0, -71.4285713818981, 0, 0, -116.666666683235, 0, 0, 0, 0, 0, 0, 238.095237729562, 0, 0, 1074.07407421198, 0, 0, 1749.99999955177, 0, 6.04775178386825e-08, 5.63641033353927e-08, -1.58135993150057e-08, 0, 0, 0, 0, 1.71864387810051e-06, 0, 0, -4861.11111287189, 0, 0, -5833.33332865725, 0, 0, -16111.1111085211, 0, 4.69478858253924e-08, -7.03436157676457e-08, 7.13329050422535e-08, -1.63329486349534e-07, -1.10526930255021e-06, -7.16747067793622e-07, 0, 0, 0, 0, -3.47843958609173e-06, 0, 0, 11574.0740801258, 0, 0, -1.35886992318976e-05, 0, 0, 72916.6666605018, 0, 8.36200443573417e-07, -5.24201572393826e-07, -2.12056218684437e-06, 0, 53703.7036855569, 0, 1.04437776774678e-08, 0, -7.38918921282123e-07, 1.74665848322412e-06, 2.65547604371874e-06, 2.89858399477382e-07, 3.08804759746927e-06, 0, 0, 0, 0, 5.29239759413991e-06, 0, 0, -13888.8888991239, 0, 0, 2.28737306133084e-05, 0, 0, -173611.111108214, 0, -4.25252519458071e-06, 2.65282562701068e-07, 1.58588813110283e-06, 0, 5.57710622490375e-05, 0, 0, -243055.555493651, 0, 1.823684385281e-08, -1.24947550249724e-05, 3.5544416722343e-06, 2.74610928629847e-07, -2.8537005020303e-05, 3.89040243773893e-07, 1.05026729325616e-05, 2.49854598410522e-05, -1.98331299429613e-05, -1.80805401694616e-05, -2.06261877608274e-06, 5.0176146839695e-05, 0, 0, 0, 0, -4.93121962576034e-06, 0, 0, 6613.75662423786, 0, 0, -2.0462410633002e-05, 0, 0, 208333.333353054, 0, 2.85746981012275e-06, -1.63444085593396e-07, 2.61811407683757e-06, 0, -7.09932660686418e-05, 0, 0, 578703.703651689, 0, 2.55584262326904e-06, 2.23645059645038e-05, -2.87344849922598e-06, -2.53917777690478e-06, -2.63998614192083e-05, 2.76030890199429e-06, 0, -0.000118391096394335, 0, -1.24607549072365e-06, 9.19101516970541e-06, -6.81266572787734e-06, -2.47373547176749e-06, 2.91886888135482e-05, 1.57042040882998e-06, -5.23585437238658e-06, -2.700928403884e-05, -3.30241099190882e-05, 4.27864070096924e-05, 0.00011006947125193, 5.55054676980438e-05, 0, 0, 0, 0, 1.92429553315603e-06, 0, 0, -5.94836401617327e-06, 0, 0, 1.406042493213e-05, 0, 0, -99206.3492199898, 0, -7.96695184753275e-07, -7.61473145219764e-07, -3.61896290004144e-06, 0, 3.6190742540829e-05, 0, 0, -694444.444414795, 0, 6.91218468349139e-08, 5.11355827635753e-06, -3.23546527672095e-06, 1.10159817353726e-06, -7.3183159818515e-06, 3.65409299785452e-06, 0, 8.51434662055783e-05, 0, 0, 0.00010086991818927, 0, 1.67581710253825e-07, 3.07495403797534e-05, 8.60016266437479e-06, -1.29364935981776e-05, -2.3152771130667e-05, -4.47959650879384e-06, -4.05973036641078e-05, -6.24629518628699e-06, -7.15084052518744e-05, -7.92562941335732e-06, 3.83689752336054e-05, -8.3097930264747e-06, -4.48161588049271e-05, 6.09728441217188e-05, 8.75714561048741e-06, -1.6175666608748e-05, 9.28584597743275e-05, 9.30565704166462e-05, 4.64132686208411e-05, 0, 0, 0, 0, -4.56537126675794e-07, 0, 0, 1.54320691543592e-06, 0, 0, -3.27661954159845e-06, 0, 0, 4.31657499741109e-06, 0, 3.11334825909294e-07, -1.4898057155657e-07, 7.43798365173197e-07, 0, -7.99034880318929e-06, 0, 0, 330687.83068058, 0, 4.46372207428912e-08, -3.87097529019791e-07, 6.76263527240512e-07, 4.52807960589538e-07, 1.90672580397052e-06, -5.62930248019443e-07, 0, -2.54615926767126e-05, 0, 0, -3.78020713004813e-05, 0, -6.65690062178094e-07, -1.55474934845417e-05, -3.76146853996079e-07, 4.14887965376746e-06, -1.24783355190661e-05, -3.92581239455925e-06, 2.30188344177852e-05, 2.64199668745219e-06, 1.41598555024473e-05, 0, -4.52354388812642e-05, 0, -7.33770092785947e-07, -1.80764130867924e-05, 1.01937412803777e-06, 7.18332495370339e-06, -5.48170970367875e-06, -2.83124072251905e-06, 1.24463038802936e-05, 8.40245678837598e-06, 4.26345236875121e-06, 1.16141302746353e-05, -9.29374048190516e-06, -2.60159299459968e-06, -1.85514619823352e-05, -1.73451314506702e-05, -1.01742855225153e-05, -1.08051442090099e-05, -2.90856249104754e-05, 5.1088814973839e-05}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 0, 6.25000000091808, 0, 0, 0, 0, -62.500000017529, 0, 0, -162.05357144687, 0, 0, 0, 0, 0, 0, 2.46907594144642e-08, 0, 0, 1628.47222238748, 0, 0, 1620.53571475556, 0, -2.1090612007861e-08, -2.12787582455634e-08, 2.0966253796141e-08, 0, 0, 0, 0, 1.67570185888426e-07, 0, 0, -8394.09722233714, 0, 0, -2.26857970681455e-06, 0, 0, -16284.7222263379, 0, -3.04597088624995e-08, 1.55459982549326e-08, -9.00705371377476e-08, 5.97295870712699e-08, 3.86276711932926e-07, 2.35187077104155e-07, 0, 0, 0, 0, -1.19682977223375e-06, 0, 0, 24305.5555546313, 0, 0, 2.1786354460143e-06, 0, 0, 83940.972235612, 0, -2.11107666219623e-07, 2.0372809509522e-07, 1.07624346477638e-06, 0, 1.74311772061356e-05, 0, 1.03778180181166e-07, 2.20601092273931e-08, 3.35490315827707e-07, -7.07676255741957e-07, -3.73023755110832e-07, -1.61170589213051e-07, -8.15209617730968e-07, 0, 0, 0, 0, 2.08322843309026e-06, 0, 0, -39930.5555528745, 0, 0, 4.07039862065886e-06, 0, 0, -243055.55557519, 0, 1.38615720188163e-06, -1.03724420267973e-07, -1.07930739644461e-06, 0, -3.43242737682109e-05, 0, 0, -5.99955187490401e-05, 0, -2.75177854878708e-07, 4.32871005203334e-06, -1.26014757315392e-06, -2.90245375684233e-07, 1.21787940564958e-05, -1.20235866590051e-06, -3.50323262490771e-06, -1.04459020090052e-05, 8.00822496570004e-06, 7.2341469065229e-06, 2.18882584234348e-06, -2.06277222617129e-05, 0, 0, 0, 0, -1.86118060775971e-06, 0, 0, 34722.2222185368, 0, 0, -1.2073506569523e-05, 0, 0, 399305.555565051, 0, -9.86589149431232e-07, 9.41673159313801e-08, -1.11541333062734e-06, 0, 2.59285528475932e-05, 0, 0, 7.33976490791653e-05, 0, -6.95170490138909e-07, -8.95802582371901e-06, 1.49941684740746e-06, 8.02011730432768e-07, 1.2480402248684e-05, -2.70912181742088e-07, 0, 9.12864290702188e-05, 0, -1.26994399462437e-07, -8.20592599174793e-07, 2.42514762661404e-06, 2.2402220097777e-06, -1.17255679474262e-05, 1.78756489334576e-06, 2.53395606577439e-06, 1.07793314439695e-05, 1.72838769365297e-05, -1.75372200637382e-05, -4.01336891774071e-05, -2.13751427325021e-05, 0, 0, 0, 0, 1.01210093220239e-06, 0, 0, -12400.7936484044, 0, 0, 8.96207984161866e-06, 0, 0, -347222.222223341, 0, 1.844174834151e-07, 2.58211542654139e-07, 1.19433073312999e-06, 0, -2.71877547904356e-06, 0, 0, -4.83921127373153e-05, 0, -3.68056792398517e-08, -2.50378483096796e-06, 9.6504911680635e-07, -3.08480821662247e-07, 1.69004953343106e-06, -2.52344476442352e-06, 0, -5.46636726302422e-05, 0, 0, -8.66135392072727e-05, 0, -1.633504000432e-07, -1.53630002841824e-05, -3.1593542334555e-06, 6.26326970985429e-06, 5.49852719537815e-06, 3.01293360490784e-06, 1.53966382690019e-05, 3.37943868658094e-06, 3.3537282725839e-05, 2.41248290404258e-06, -1.96163998436211e-05, 2.10818487241642e-06, 1.62118183837444e-05, -1.96875197321383e-05, -8.23752412398105e-06, 7.18732156165924e-06, -4.10824375935325e-05, -4.34269769198573e-05, -2.0267946085508e-05, 0, 0, 0, 0, -1.56509803801003e-07, 0, 0, -6.79402940821841e-07, 0, 0, -2.72999436890895e-06, 0, 0, 124007.936507069, 0, -9.74662796917299e-08, 5.3722901957743e-08, -1.75679441659848e-07, 0, -2.50736700118176e-06, 0, 0, 1.18966086453328e-05, 0, -1.65669610996266e-08, -2.78780001931129e-08, -2.45678454551965e-07, -1.70658011684581e-07, -7.38550966919569e-07, 3.70969342346563e-07, 0, 1.20582239234024e-05, 0, 0, 3.22344589742463e-05, 0, 1.49391262136044e-07, 5.71436789011502e-06, 1.12400193591982e-07, -1.69730312862762e-06, 5.68931009388087e-06, 1.46373798338361e-06, -8.51032163718747e-06, -1.50772173128047e-06, -4.8762537620554e-06, 0, 3.11006298538795e-05, 0, 4.40836876457934e-07, 6.10888460392374e-06, -4.22167581874434e-07, -2.53800019011053e-06, 2.82789793693697e-06, 1.26234936201436e-06, -4.383969472384e-06, -3.10796712468352e-06, -3.44634971243241e-06, -4.68005488720409e-06, 7.21139715225701e-06, 3.0237164368247e-06, 6.7997688494318e-06, 1.1278650421496e-05, 2.67180962700287e-06, 9.5324268415941e-06, 1.25710986441168e-05, -2.21813402946811e-05}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 0, 11.1111111120705, 0, 0, 0, 0, 0, 0, 0, -301.984127009809, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3255.18077624107, 0, 0, 1.91525578436504e-07, 0, 0, 0, 0, 0, 0, 0, 0, 1.77091491331503e-07, 0, 0, -18541.6666677648, 0, 0, -2.02247016604199e-07, 0, 0, -1.71603477624952e-06, 0, -1.14461365946687e-08, 0, 0, -3.54476800901486e-08, -1.27776867855738e-07, -7.13424748910264e-08, 0, 0, 0, 0, -5.13837000276069e-07, 0, 0, 61863.4259287781, 0, 0, -9.28491183006372e-07, 0, 0, 7.37264479462814e-06, 0, 8.08979549946114e-08, -2.75412016220864e-08, -2.11102016994308e-07, 0, 3.98800772558445e-06, 0, 4.23438922237297e-08, 1.68557987709585e-07, -2.28582073066151e-08, 2.39487413772605e-07, 1.53023945188249e-07, 1.5410553696468e-07, 3.31137413001419e-07, 0, 0, 0, 0, 8.94595629697755e-07, 0, 0, -125000.000004214, 0, 0, 3.20118803234915e-06, 0, 0, -1.68272430447666e-05, 0, -3.53953037954632e-07, 0, 1.78156050166573e-07, 0, -1.25294316652355e-07, 0, 0, -1.50726371728048e-05, 0, -7.75169161344303e-08, -1.26943105376111e-06, 1.57350312293972e-07, -1.45120660396692e-08, -3.52350238172551e-06, 2.68796847619742e-07, 4.63432268553193e-07, 2.01417938183447e-06, -2.12557481833615e-06, -2.32141957810624e-06, -1.27239478555795e-06, 4.25983643676647e-06, 0, 0, 0, 0, -8.7304363608177e-07, 0, 0, 150462.962966665, 0, 0, -3.96407644297209e-06, 0, 0, 2.09623813178284e-05, 0, 2.88308081948898e-07, 0, 4.76649831049417e-07, 0, -6.04904472910334e-06, 0, 0, 3.11497239306159e-05, 0, 3.04003937413165e-07, 2.54823022527084e-06, -3.10186895127249e-07, -2.18788131783511e-08, -2.74935519686287e-06, 2.0700133636162e-07, 0, 9.34321558834281e-06, 0, 4.98588259522532e-07, 6.73209974356674e-07, -2.69733659394603e-07, -4.94006398296392e-07, 3.67264281628342e-06, -5.46007968271874e-07, 1.60450307621695e-07, -2.04000132093791e-06, -3.71331748025376e-06, 5.22073279071236e-06, 1.04366963578436e-05, 5.85156480696258e-06, 0, 0, 0, 0, 4.19775243006405e-07, 0, 0, -99206.3492081054, 0, 0, 2.70976212466896e-06, 0, 0, -1.27670411972231e-05, 0, -6.60196807156783e-08, -5.9747300059448e-08, -3.51679140282923e-07, 0, 5.75582264357176e-06, 0, 0, -2.63592347222146e-05, 0, -1.13064664755075e-07, 4.19810905727379e-07, -2.18805164724872e-07, -1.62579425291834e-08, -5.16251419714007e-07, 6.03672991249926e-07, 0, 7.80790031993471e-07, 0, 0, -1.8614092274725e-05, 0, -1.77278853686476e-07, 3.52615952057784e-06, 5.25593988050578e-07, -1.89240236757318e-06, -1.41938906950436e-06, -1.26914149375797e-06, -4.18465303359851e-06, -1.32638920967268e-06, -9.78113251073705e-06, -6.70825362566118e-07, 4.49158663692579e-06, -5.47268405932596e-07, -4.46671854253856e-06, 1.94393553166123e-06, 2.05907821526236e-06, -2.69536077274814e-06, 8.9571469394518e-06, 1.03749689783509e-05, 3.66480766325731e-06, 0, 0, 0, 0, -9.67928829610361e-08, 0, 0, 27557.3192243605, 0, 0, -6.72052798902953e-07, 0, 0, 3.07658142070425e-06, 0, 2.98704563532188e-08, 0, 5.08220387862242e-08, 0, -1.81528998379608e-06, 0, 0, 8.51207831080356e-06, 0, 3.44257036084817e-08, 9.36087874853428e-08, 6.54522708565862e-08, 5.0528220547745e-08, 2.82908639537425e-07, -6.02220932998891e-08, 0, -2.31786395159711e-06, 0, 0, 8.68043403233448e-06, 0, 5.40923635275581e-08, -1.08109782113127e-06, 2.23280570067505e-08, 4.5483489140117e-07, -1.23981076873413e-06, -2.1999747062015e-07, 2.07275566717534e-06, 5.26143300409474e-07, 1.30722715064142e-06, 0, 2.72630962720128e-06, 0, -4.00274121986181e-08, -8.61798701106929e-07, 1.79420746062636e-07, 6.28911552111756e-07, -4.52428988433913e-07, -1.88614276162297e-07, 1.31371670130001e-06, 1.27092995260089e-06, 1.35013312993007e-06, 1.35735509706718e-06, -1.73000178816564e-06, -7.2223077959396e-07, -1.03764678771907e-06, -2.7259519354327e-06, 1.16096528317354e-07, -2.525644545026e-06, -2.56240163248328e-06, 6.39607387639468e-06}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 11.1111111117703, 0, 0, 0, 0, 0, 0, 0, -301.984126997634, 0, -3.58670396334582e-08, 0, 0, 0, 0, 0, 0, 0, 3255.1807760231, 0, 2.24482826341836e-07, 0, 0, 4.98627661999546e-07, -2.54037350509892e-08, 6.18252573225351e-08, 1.56691480663379e-07, 0, 0, 0, 0, 0, 0, -18541.6666668398, 0, -1.03305439975755e-06, 0, 0, 0, -6.03607727504378e-06, 0, -3.00795581373393e-06, 2.59445271871975e-07, 1.41330487046039e-07, -6.91137225972492e-07, 2.09940425868195e-07, -1.25426579089317e-07, -1.68565542468044e-06, 0, 0, 0, 0, 0, 0, 61863.4259248612, 0, 2.09784585440142e-06, 0, 0, 0, 2.19125052720099e-05, 0, 1.58790781041871e-06, -3.76198042216148e-07, 1.05871465888369e-06, 4.2500741717349e-06, 0, 0, 2.68880766554126e-05, -2.76468587179699e-06, 8.06983624880964e-07, -2.26163211206669e-06, 4.10836998637175e-06, 4.24888114571253e-06, 6.61792545569796e-06, 2.2038649851785e-06, 0, 0, 0, 0, 0, 0, -124999.999996661, 0, -4.32893119379723e-06, 0, 0, 0, -4.40560205008625e-05, 0, 9.17471166111228e-06, 1.81749875820649e-06, -8.47477754658353e-06, -8.15854689134541e-06, 0, 0, 0, -0.000110730694651505, 0, -3.37374925405591e-05, 5.54049432992379e-06, 1.64765206765521e-06, -6.47901739152055e-06, -3.48311382607493e-05, -9.61654649382201e-06, -2.56931222176914e-05, 1.98069050204871e-05, -4.71887796687325e-06, -3.75927262397516e-05, -4.4919450327148e-06, 9.07940981918187e-06, -2.60196550856033e-05, 0, 0, 0, 0, 0, 0, 150462.962957088, 0, 6.13706918021714e-06, 0, 0, 0, 4.0834372999004e-05, 0, -1.29895351527466e-05, -2.51851204859302e-06, 1.39524304193377e-05, 1.14768054686066e-05, 0, 0, 0, 0.000200860729412945, 0, -7.72680414808342e-06, -8.56009540849708e-06, 3.36201896955312e-06, 3.8502693119845e-06, 4.30572284245295e-05, 2.33614759515223e-05, 2.60002090151732e-06, 0, 0, 0.000137772672599564, -7.85161392001359e-06, -1.41202234877916e-05, -1.92197495786092e-05, 7.41857779040402e-05, 9.51356439194673e-07, 6.14214118018101e-05, 7.83053796735591e-06, -1.29915021491467e-05, 0.000102998464660093, -2.23427674101431e-05, 5.06129863085096e-05, -4.74852274741852e-05, 0, 0, 0, 0, 0, 0, -99206.3492024329, 0, -3.82722425739367e-06, 0, 0, 0, -1.57840805709636e-05, 0, 5.1636085462231e-06, 1.41843442358196e-06, -6.37013650262631e-06, -8.07833476625073e-06, 0, 0, 0, -0.000146295660372661, 0, 3.38920327973356e-05, 4.73018690959794e-06, -1.56887779693332e-06, -5.86328067049543e-06, -2.83972095704086e-05, -1.9977975226482e-05, -1.94497857289055e-05, 0, 0, 0, -0.000208385033659837, 0, -5.25877794579982e-05, 5.13412170804664e-06, 1.11439586432637e-05, -7.2448825884089e-07, -0.00010992095131182, -2.08063750684673e-05, -5.91588316239312e-05, -9.46533518688525e-06, -1.63099704284621e-06, -4.32278788292744e-05, 2.05919585980446e-05, 8.32096572468931e-06, -7.82075525361911e-05, -7.64011713504578e-06, 3.77860231035577e-05, 2.28739441460461e-05, -7.82818219891918e-05, -0.000120098132923643, -4.02239263939323e-05, 4.01299212394524e-05, 0, 0, 0, 0, 0, 0, 27557.3192233716, 0, 9.86828578103065e-07, 0, 0, 0, 1.4862865506239e-06, 0, -9.16031948997537e-08, -3.44067260299592e-07, 2.64074713828328e-06, 1.76800610269492e-06, 0, 0, 0, 3.74297403504836e-05, 0, -8.25054026118223e-06, -1.71448248914377e-06, 9.25254786792798e-07, 1.83426465460429e-06, 1.32739575554493e-05, 5.61166258480004e-06, 1.79413854714025e-06, 0, 0, 0, 9.23881281759414e-05, 0, 2.32993425310384e-06, -5.05962699429256e-06, 2.7003435377699e-06, -1.24054249818597e-06, 2.21146256302161e-05, 6.52156961490364e-06, 2.4189891031026e-05, 3.76397416082209e-06, 1.85541307593798e-05, 3.13969895488998e-05, 0, 0, 5.94745797047073e-05, -4.50273208528261e-06, -4.54147135696315e-06, -8.4462184652788e-06, 5.17716823890345e-05, 1.35105936560779e-06, 3.38393361467949e-05, -1.25215059435325e-05, 1.08081867082367e-05, 6.18266035544941e-06, 2.31026475245041e-05, 3.95714414505393e-05, 3.82560933485308e-05, -2.04934234893309e-05, 2.84071305540244e-05, 5.13038137290962e-05, 2.79431299368066e-06, -2.50824543763061e-07, 8.11974332663083e-06}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 6.2500000043312, 0, 0, 0, 0, 0, 0, 0, -162.053571401645, 0, -62.5000001108925, 0, 0, 0, 0, 0, 0, 0, 1628.47222222254, 0, 9.97016362636708e-07, 0, 0, 1620.53571490205, -9.13623639666695e-08, -4.11822117828851e-07, 5.35883365200233e-07, 0, 0, 0, 0, 0, 0, -8394.09722177738, 0, -4.34176017113169e-06, 0, 0, 0, -16284.7222219393, 0, -4.4558222754688e-07, 1.20722937473277e-06, 1.00231644788209e-06, -5.25409954123623e-06, 4.96715536518172e-06, -3.24102994160206e-06, -5.71311898457991e-07, 0, 0, 0, 0, 0, 0, 24305.5555548491, 0, 1.61271512715806e-05, 0, 0, 0, 83940.9722349058, 0, 2.81279429300496e-06, -6.02581253727222e-06, -6.10317334592962e-06, 1.89791213112588e-05, 0, 0, -1.48769596815844e-05, -1.86741500754907e-05, -3.82183274433984e-06, -1.44492735648603e-05, 3.85874603419633e-06, 1.86884314730584e-05, 8.49980148016222e-06, -3.94356727276119e-06, 0, 0, 0, 0, 0, 0, -39930.5555520803, 0, -2.76194188348334e-05, 0, 0, 0, -243055.555579238, 0, -2.87755959605615e-05, 9.71759109262525e-06, 2.89964889062793e-05, -4.59524154776659e-05, 0, 0, 0, 2.86149493841501e-05, 0, 5.50907091088244e-05, 6.18569062021908e-05, 1.71785235291438e-05, 7.20650034134209e-05, 4.49970894913357e-05, -7.25902905324014e-05, 1.18360383577957e-06, 3.53968914515192e-05, -2.20837666484664e-05, 9.61056540877049e-06, 9.25540656681875e-05, -9.63777139309627e-06, -6.2075866252733e-05, 0, 0, 0, 0, 0, 0, 34722.2222206256, 0, 2.6999441589323e-05, 0, 0, 0, 399305.555582175, 0, 4.67078749052071e-05, -7.66828339019224e-06, -4.58183564088688e-05, 4.62182457537985e-05, 0, 0, 0, -1.53299470900474e-05, 0, -7.63743425893662e-05, -6.66472862621459e-05, -4.54268528209933e-05, -7.13991940601202e-05, -3.92162939025855e-05, 9.8195227608492e-05, 8.5349412361331e-05, 0, 0, -8.34755435863139e-05, -0.000142827193422644, 3.57139879343275e-05, -0.000101453844436326, -8.12151066585853e-06, 0.000139418880789164, -7.32982001650322e-05, -0.000232333333258679, -2.60670617690459e-05, -1.91204024896995e-06, -0.000167512234520869, -0.000112780949455931, 0.00028736575394874, 0, 0, 0, 0, 0, 0, -12400.7936500988, 0, -1.41206036376077e-05, 0, 0, 0, -347222.222240305, 0, -4.08755616271361e-05, 2.87675014385362e-06, 1.90327921516082e-05, -3.37501652933694e-05, 0, 0, 0, -3.89295983080058e-05, 0, 2.73505841502714e-05, 4.19645718998749e-05, 3.13974411698763e-05, 6.34283916640652e-05, 5.81048255162758e-06, -9.6320915998149e-05, -2.32000044370148e-05, 0, 0, 0, 7.46535786071613e-05, 0, 0.000122703375301986, 0.000139970651733681, -1.95140824776915e-06, 0.000113365287383561, 9.89345593884709e-05, -0.000108259810528343, 0.000102685175919043, 0.00022100128640106, 8.33067333986008e-05, -0.000207961953610548, 9.46847119189877e-05, -7.3872353645808e-05, 6.33186189555752e-05, 0.000354168342750907, 8.73753351680952e-05, 0.000200996191286894, -3.61726628785412e-05, 0.000120666622754653, -0.000208530278746598, -0.000158017860888928, 0, 0, 0, 0, 0, 0, -1.16165331274593e-06, 0, 3.29625316349299e-06, 0, 0, 0, 124007.936511536, 0, 9.62535863661625e-06, -7.6242890185361e-07, -6.91744848837979e-06, 1.05438359307447e-05, 0, 0, 0, 2.30239348828611e-05, 0, -9.39402263365057e-06, -1.04656681453435e-05, -9.20833420835837e-06, -1.81396662982816e-05, -1.73918290495313e-05, 2.94825759969576e-05, 2.24043412694845e-05, 0, 0, 0, -1.35040169240683e-05, 0, -7.15443005473551e-05, -3.16468068819792e-05, -1.65542082257186e-05, -2.82934468561086e-05, 3.21580474653491e-05, 5.33977759856132e-05, -5.02284610326581e-05, -6.77291589984163e-05, -6.30271421185179e-05, 3.86628339785807e-05, 0, 0, -4.96738428480345e-05, -5.80454410669019e-05, 1.79963497700717e-05, -3.16500554233329e-05, -5.07478771488931e-05, 4.18438633930827e-05, -6.940574466596e-05, -0.000117710360294426, -0.000132887721227347, -0.000129822996240457, 4.81918327429533e-05, 7.50590685886986e-05, 3.19712289846274e-05, -0.000110925909832889, -8.09086637753131e-05, 2.87254741267538e-05, -8.05887659356649e-05, -1.24109872148961e-05, 0.000162389456782075}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 4.76190476368255, 0, 0, 0, 0, 0, 0, 0, -116.666666800246, 0, -71.428571415006, -7.12702734495753e-08, 0, 0, 0, 0, 0, 0, 1074.07407474349, 0, 238.095237678542, 0, 0, 1749.99999977972, -2.59247301136632e-07, 1.63323241652981e-06, 7.45029723375188e-07, 0, 0, 0, 0, 0, 0, -4861.11111474719, 0, 1.48766172598636e-06, 0, 0, 0, -16111.1111161737, 0, -5833.33333264262, 2.91119496120245e-06, -8.0308735120359e-06, -3.71187910864795e-06, -5.80560129797011e-06, 8.11360016011252e-07, -1.34603327469974e-05, 0, 0, 0, 0, 0, 0, 11574.07408238, 0, -1.16499441314408e-05, 0, 0, 0, 72916.6666677719, 0, -1.29677960294527e-05, -1.02149010264693e-05, 3.02662393849761e-05, 9.36886654999435e-06, 0, 0, 53703.7037167695, -1.2455768457371e-05, -4.05611124101244e-06, 1.81241063126273e-05, 8.1310757642681e-05, 2.39296255248281e-05, 7.836562592267e-05, 1.8736022298088e-05, 0, 0, 0, 0, 0, 0, -13888.8889054985, 0, 2.13242451692801e-05, 0, 0, 0, -173611.111131469, 0, 5.99240673004605e-05, 2.1629047986867e-05, -8.41427716201471e-05, 5.75118615118534e-06, 0, 0, 0, -243055.555600286, 0, 7.11680283891391e-06, 2.52298324322264e-05, 5.45533516677862e-05, -0.000126552753664157, -0.000255868372404248, -8.71331963066744e-05, -0.00020208113838995, 6.90847824653848e-05, -6.33842871128999e-05, -0.00037122549909359, 5.28178169652064e-05, 3.02304979062425e-05, -4.02244290136629e-05, 0, 0, 0, 0, 0, 0, 6613.75662361884, 0, -2.16568719387464e-05, 0, 0, 0, 208333.333361858, 0, -7.59761999264695e-05, -3.03577915414145e-05, 0.000119659404584064, -7.58232660766956e-06, 0, 0, 0, 578703.703820291, 0, -8.04345501878926e-06, -7.69360055842315e-05, -7.05144355769243e-05, 0.000155726896588371, 0.000277353700168173, 0.000147798222029259, 9.81283610885862e-05, 0, 0, 1.77534967686564e-06, -8.66248379512714e-06, 1.14065503873077e-06, 8.66308296467382e-05, 0.000619704890422579, 0.000135655681201748, 0.000754863489145219, -0.00017115416216541, 0.000194474756742785, 0.000612119421654081, -0.000190997269077938, 5.3477053225911e-05, 7.73514699960464e-05, 0, 0, 0, 0, 0, 0, -4.03822204853123e-06, 0, 1.0373286459599e-05, 0, 0, 0, -99206.3492146867, 0, 5.88037930456926e-05, 2.02683326522443e-05, -5.61962305372281e-05, 2.16252153336944e-05, 0, 0, 0, -694444.444470874, 0, 2.03874777452923e-05, 5.62329077892731e-05, 6.61198069607939e-05, -0.000146151350144943, -0.000131965201648152, -5.37595605259337e-05, -6.24662179204402e-05, 0, 0, 0, -9.23899493258806e-05, 0, 7.59596084639669e-06, 2.28323390519546e-05, 4.41959647990065e-05, -0.000164934404492516, -0.000654772181436377, -0.000236125968811085, -0.000663138261965359, 0.000201838112979723, -0.00025016243903978, -0.000203546381476416, 2.36583328519567e-05, -0.000186476214470778, -0.000877364862106364, 0.000325538704169591, 0.000332317696145263, 8.94067029297376e-05, -0.000640087607572221, -0.00056153598216643, -0.00069310005943461, 0.000629362171144609, 0, 0, 0, 0, 0, 0, 2.49796560367848e-06, 0, -2.87989288894378e-06, 0, 0, 0, 1.44571699548454e-06, 0, -1.10379646985473e-05, -4.55047123751341e-06, 1.67996879757374e-05, -1.03538637484091e-05, 0, 0, 0, 330687.830681285, 0, -5.39706110838059e-06, -1.50503838796213e-05, -1.82053107739168e-05, 4.30112905847466e-05, 6.05650974858893e-05, 1.18705656934978e-05, -1.83475550157267e-05, 0, 0, 0, 3.13536927633101e-05, 0, 4.24664202603147e-05, -3.10879869499989e-05, -1.23359811272923e-05, 5.1865295062355e-05, 6.57706880812523e-05, 7.8152644355909e-05, 0.000213317507109674, -7.15238072114446e-05, 0.000121100358441292, 0.000125369891597956, 0, 0, 2.80392027624868e-06, 8.71475547408862e-06, 3.68049558205461e-05, 2.11956908652561e-05, 0.000371799236471178, 0.000110225200622019, 0.000456719574959888, -0.000264477935000503, -3.35062035729959e-05, 0.000100839305033945, 0.000139226513743267, 0.00013815337290034, 0.000489099105734537, -0.000260689683901445, 0.000159962894422157, 0.000522152604913721, -0.000393596272744949, -8.69600419208909e-05, -0.000135900517281598}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 4.16666666622287, 0, 0, 0, 0, 0, 0, 0, -95.138888830232, 0, -76.3888888404902, 2.41528904066617e-08, 0, 0, 0, 0, 0, 0, 781.250000327126, 0, 416.666665751711, 0, 0, 1744.21296398947, -6.56391216863598e-08, -2.65347925539067e-07, 7.82033582917562e-07, 0, 0, 0, 0, 0, 0, -2951.38888865432, 0, -694.444436907289, 0, 0, 0, -14322.9166771467, 0, -9513.88888428641, 3.13508136100669e-06, -2.64792261479198e-06, -7.77273592211941e-06, 1.03510228543747e-05, -7.83241333588335e-07, -1.0022676282113e-05, 0, 0, 0, 0, 0, 0, 5208.33332931808, 0, -1.40061239139562e-05, 0, 0, 0, 54108.796364097, 0, 15856.4814489798, -1.37529485153554e-05, 2.39394760385367e-06, 1.94903887500654e-05, 0, 0, 78125.0000375281, -2.86347971670931e-05, -2.7116596201823e-05, -1.80652039523377e-05, 5.51807526864712e-05, 3.57554502075203e-05, 3.00963915844031e-05, -2.69540030522817e-05, 0, 0, 0, 0, 0, 0, -3472.22220243007, 0, 1.8667236670731e-05, 0, 0, 0, -95486.1112376758, 0, 8.06381624409267e-05, 2.24579709169346e-05, 4.01183018701479e-05, -4.67686447894095e-05, 0, 0, 0, -295138.889088826, 0, -130208.333333801, 9.28862436800241e-05, 6.21409541696797e-05, 7.81463680972308e-05, -0.000122348134519201, -0.000117589238541245, -4.56229801501243e-06, 1.15212429415439e-05, 6.22106301000299e-07, -0.000278846873087293, 0.000241071377579416, -0.00015389825707152, -0.000255492040238513, 0, 0, 0, 0, 0, 0, -1.94715895215219e-05, 0, -1.78634762641636e-05, 0, 0, 0, 63657.4075212615, 0, -0.000123419492884467, -1.4277954821916e-05, -9.996383613999e-05, 4.94303450106226e-05, 0, 0, 0, 520833.333573602, 0, -0.000174700927889903, -6.76908200882828e-05, -0.000103416803108326, -8.51583557904296e-05, 0.000186832109087041, 0.000155212004952956, 5.58296649853181e-05, 0, 0, 491898.148314258, -0.000217894261761231, -4.08404438630378e-05, -0.000198869345971456, 0.000672414257533535, 0.000252867913510312, 0.000463750746128736, -0.000600623893961922, -0.000117023587184627, 8.21156693338259e-05, -0.00062974553104228, 2.32562280351273e-05, 0.000766438539394448, 0, 0, 0, 0, 0, 0, 1.33010810652926e-05, 0, 1.06745222433012e-05, 0, 0, 0, -6.64009626478222e-05, 0, 8.57938818489586e-05, 4.84220434982161e-06, 5.82675885015757e-05, -3.72090878702448e-05, 0, 0, 0, -347222.22249828, 0, 0.000212560321751915, 3.66389361769704e-05, 5.53365322209384e-05, 9.52520011185492e-05, -0.00013328106254981, -0.00014119073575578, 0.000105195981493855, 0, 0, 0, -868055.555712627, 0, -4.52840157258367e-06, 0.000153305669798888, 3.34434094280252e-05, 0.000119404460588776, -0.000456397893562895, -0.000208160311655, -0.000181753128159931, 0.000390937105232097, 0.000131782048707687, -0.000532726766048618, 0.000153642253467816, -3.85884428720067e-05, -0.000741124343928066, 0.00106747030610183, 0.000130445150543433, 0.000772044043400719, -0.00020721555937559, -0.000199751872630378, -0.00104724365550895, -0.00040693122191475, 0, 0, 0, 0, 0, 0, -5.5942597660766e-06, 0, -1.33736308285744e-06, 0, 0, 0, 1.24442096740437e-05, 0, -2.85189441131192e-05, -1.17671531745621e-06, -2.0014884882765e-05, 1.45681388002923e-05, 0, 0, 0, 9.9170190715831e-05, 0, -6.7596669818975e-05, -7.35391341732567e-06, -1.94036670553201e-05, -2.67908666438194e-05, -1.21316638653489e-05, 4.07641870931686e-05, -3.86923996961992e-05, 0, 0, 0, 578703.703796883, 0, -8.91356726940858e-05, -2.14908286577196e-05, -4.08952660767128e-05, -8.57565045501847e-06, 0.000327023153997114, 8.42444952845664e-05, -1.11064168107216e-06, -0.000136551643462647, -8.28020427453312e-05, 5.49729348805721e-05, 0, 0, 6.36203956807633e-05, -8.77357785675928e-05, 1.84168339247401e-05, -2.36327719244472e-05, 0.000384113103427528, 8.65646494809766e-05, 0.000160828668742696, -0.000302025678816351, -0.000309773563127132, -0.000410962406013696, 0.000375791872961151, 0.000396714453854914, 0.000328021073305298, -0.000381721135359433, 5.125582791745e-06, 0.000569039039633232, -0.000193750159181637, -0.000112515017263832, 0.000663308847049832}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 4.00000000704716, 0, 0, 0, 0, 0, 0, 0, -83.3333335819247, 0, -83.3333335756166, -7.84462213185109e-08, 0, 0, 0, 0, 0, 0, 583.333334173475, 0, 583.333334568606, 0, 0, 1736.11111607336, -3.79012924621104e-07, 1.69966549623724e-07, 1.81685678623031e-06, 0, 0, 0, 0, 0, 0, -1666.66666640642, 0, -1666.66667098238, 0, 0, 0, -12152.7778142964, 0, -12152.7778061166, 1.84218625956127e-06, 2.85448996759597e-06, -1.18138910748851e-05, -1.20004540602901e-06, -5.37753560740304e-06, -1.35320350109782e-05, 0, 0, 0, 0, 0, 0, 1666.66665772271, 0, 1666.66666618059, 0, 0, 0, 34722.2223229114, 0, 34722.2222849267, 3.07530020657791e-07, -2.6004581551098e-05, 4.99765915412341e-05, 0, 0, 85069.4446079156, -1.55824508158221e-05, -8.19037054459026e-06, -1.25344155791506e-05, 0.000110802306462298, 4.06408690051353e-05, 0.000119959478481226, 1.65191970811461e-05, 0, 0, 0, 0, 0, 0, 2.83711628473065e-05, 0, 1.16396626913012e-05, 0, 0, 0, -34722.2223611631, 0, -34722.2222775718, -8.28868790885148e-07, 6.75449798565348e-05, -8.30714620620913e-05, 0, 0, 0, -243055.555967915, 0, -243055.555913148, 1.25421416479305e-05, 4.1718081647402e-05, 4.17567578737463e-05, -0.000244097235243147, -0.00016047142961875, -0.000316551251702603, 0.000100690250063092, -8.33349621169281e-05, -0.000509586710449716, 0.000122194495809927, 0.000102912835470131, -8.43367082944967e-06, 0, 0, 0, 0, 0, 0, -4.42055644940201e-05, 0, -1.39663959063444e-05, 0, 0, 0, 0.000130797036430873, 0, 2.91257156060532e-05, -9.33452157959485e-08, -0.000100536356699767, 8.60511006471978e-05, 0, 0, 0, 243055.556203347, 0, 243055.556017221, -7.75402241538246e-06, -7.52683236243186e-05, -7.64084767791072e-05, 0.000187003052154004, 0.000301702686093539, 0.000179256803588542, 0, 0, 694444.445262303, -7.52479250342216e-05, -1.60229215937843e-05, 2.04816392784125e-05, 0.00116208029783973, 0.000168784638639368, 0.000834971574110368, -0.000126268933390094, 0.000381494926091263, 0.000576934722927579, -0.00018383740727372, -0.000488901924615067, 0.000433469584802279, 0, 0, 0, 0, 0, 0, 3.04201370322459e-05, 0, 6.32133503354492e-06, 0, 0, 0, -2.80368729132046e-05, 0, -8.34951226550674e-06, -2.0186394079987e-06, 9.41718424984677e-05, -6.11104662661945e-05, 0, 0, 0, -0.000330285267766665, 0, -0.00026382505790199, -1.92810989565656e-05, 6.30971052848526e-05, 8.12543690612501e-05, -5.83414432510567e-07, -0.000269742673694946, -0.000163652809780428, 0, 0, 0, -694444.445266883, 0, -694444.445078192, 1.81861081556744e-05, 0.000111461902898646, 4.27489633039026e-05, -0.00106125611114697, -0.000377384913247535, -0.00061249070736546, 0.000213567873200749, -0.000248408142967741, -8.26351000623598e-05, 2.31109418406114e-05, -0.000224834978804931, -0.0014890525362572, -2.072740609131e-05, 0.000608280986239097, -3.60555229205894e-05, -0.00158135741856564, -1.37780231302543e-05, -0.000272990105831396, 0.000781933062237573, 0, 0, 0, 0, 0, 0, -6.91997382566523e-06, 0, -1.99871957774036e-06, 0, 0, 0, 1.6812185557094e-06, 0, 2.91275377859582e-07, 5.60283116472784e-07, -3.06236215494822e-05, 1.46034293447348e-05, 0, 0, 0, 6.66576164079395e-05, 0, 3.78891592984421e-05, 9.74939489720162e-06, -1.13339359322303e-05, -3.97781431200187e-05, -7.57565881460277e-06, 9.30514664501354e-05, 3.27661366372886e-05, 0, 0, 0, 0.000276486051060278, 0, 0.000207295024288462, 8.15837624369059e-06, -1.34774126508352e-05, -2.44629978870722e-05, 0.000156404838507142, 0.000158040304639371, 0.000118371285644047, -3.12094661868807e-05, 3.06504991033727e-05, 8.27881195499731e-05, 0, 0, 694444.444729094, 2.4266783803121e-05, 2.74090664682984e-05, -2.19889954074242e-05, 0.000615741027424315, 0.000130833106178403, 0.000651754553914897, -0.00015210560769441, -0.000143025034479254, 9.68233089711945e-05, 0.000342378985093829, -0.000142021953941516, 9.76748168185017e-05, -0.000166571414265161, -0.000177197191534441, 0.000458813790007921, -0.000476424956643025, -3.93977512683437e-06, -0.000388898160513121}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 4.16666667896517, 0, 0, 0, 0, 0, 0, 0, -76.3888890499101, 0, -95.1388890404756, -1.58864379712151e-08, 0, 0, 0, 0, 0, 0, 416.666668125931, 0, 781.250000631069, 0, 0, 1744.21296695845, -2.96817027338899e-07, 4.19210781601784e-07, 9.55504518131392e-07, 0, 0, 0, 0, 0, 0, -694.44445177545, 0, -2951.38888805823, 0, 0, 0, -9513.88891597594, 0, -14322.9166846701, 1.49094356278927e-06, -1.42604556430955e-06, -7.57042999799718e-06, 9.71288505436573e-06, 3.1304339424926e-06, -1.52431342044867e-05, 0, 0, 0, 0, 0, 0, 1.02882968924697e-05, 0, 5208.3333283499, 0, 0, 0, 15856.4815697069, 0, 54108.7963248384, -9.39024936201407e-06, 8.55914765632019e-06, 2.11968228934962e-05, 0, 0, 78125.0001270616, -3.72319340289148e-05, -1.21074817888154e-05, -3.84478532439249e-05, 8.08411078321435e-05, 1.80638915490869e-05, 2.802621798602e-05, -4.97457461876061e-05, 0, 0, 0, 0, 0, 0, -1.48716323411643e-05, 0, -3472.22221322779, 0, 0, 0, -0.000198248994194028, 0, -95486.1111393247, 9.22855348391084e-06, -6.24966837153887e-06, -6.05715530542606e-05, 0, 0, 0, -130208.333761172, 0, -295138.889152123, 6.70598534723945e-05, 7.85919909100373e-05, 6.44092699546153e-05, -0.000317841517836307, -0.000116814765971629, 5.29900919488951e-05, 0.000147211842994982, 1.49934400982067e-05, -0.000384646780621165, 0.000188416222409984, -6.53766064958562e-05, 9.82876934605857e-05, 0, 0, 0, 0, 0, 0, 5.20889965820795e-06, 0, -1.08808433627084e-05, 0, 0, 0, 0.000165159891143727, 0, 63657.4073763186, -6.37071981406215e-06, -7.94081551061472e-06, 8.26433889222643e-05, 0, 0, 0, 0.00054761016172615, 0, 520833.333501655, -6.85215521738939e-05, -8.97267509267936e-05, -0.000125563036267888, 0.000484285948178106, 0.000216546979014073, -2.34130122880488e-05, 0, 0, 491898.148758045, -0.000124407465792844, -7.61551272913469e-05, -0.000304788292164818, 0.00089064199352556, 0.000237588047660867, 0.000482514861785857, -0.000461071251254519, -7.7322844579067e-05, -0.000528300853903045, -0.000524375237574266, -4.55255785424164e-05, -0.000317936536605335, 0, 0, 0, 0, 0, 0, -1.40037793944758e-06, 0, 7.33586323480726e-06, 0, 0, 0, -0.000105383212483187, 0, 5.03048837163342e-05, 3.73259878847397e-06, 2.42068833621074e-05, -4.0737333492061e-05, 0, 0, 0, -0.000429274081754439, 0, -347222.222258369, 3.590334156372e-05, 5.5085472270487e-05, 8.91144154909841e-05, -0.000321359661655067, -0.000147023582256267, 0.000147787933002039, 0, 0, 0, -0.00059997166466073, 0, -868055.556002653, 0.000112418006051498, 0.00010360056661175, 0.000177883669866882, -0.00106162037412406, -0.000214016289006778, -0.00025893999334917, 0.000358245076915072, -7.97383787547051e-05, 0.000171607956404151, 0.000288276520642478, -7.79983975612922e-06, -0.000873081169335342, 0.00107885181590587, 0.000576910195586297, 0.000608678433391368, -8.64701689769071e-06, 0.000415748225964053, -5.56051962516659e-05, 0.000146708985602774, 0, 0, 0, 0, 0, 0, -8.01921653360447e-08, 0, -1.651108996316e-06, 0, 0, 0, 1.76295280294052e-05, 0, -1.32093755718141e-05, -5.85316106268045e-08, -3.20621599567987e-06, 1.14969604386791e-05, 0, 0, 0, 0.000110297600465539, 0, 5.13511637111756e-06, -5.82733212152885e-06, -1.5576080366861e-05, -3.33160982161248e-05, 6.75435329864338e-05, 4.43127536354554e-05, -4.06821195249815e-05, 0, 0, 0, 0.000181211410241755, 0, 578703.70380143, -5.29060770480562e-05, -7.56597356865917e-05, -6.81839138488332e-05, 0.00043196668404443, 0.000118468500423731, 4.35175447489731e-05, -0.000115855878351932, 4.4586696009612e-05, -1.9275950599009e-06, 0, 0, 0.000252293339440382, -0.000111679899805129, -2.47423987191457e-05, -0.000103526892833537, 0.000656168925926982, 8.46838655334704e-05, 0.000271318058963366, -0.000263408330484793, -0.000233769908242395, -0.000182163340438417, 0.000522016879050326, 0.000129115953793361, -6.63698131591208e-05, -0.000376482390858993, -0.000136807199492936, 0.000522631080701339, -0.000438860115203849, 0.000129031993436882, -0.000204599260156619}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 4.76190476228868, 0, 0, 0, 0, 0, 0, 0, -71.4285715188668, 0, -116.666666706594, -3.17950804389488e-08, 0, 0, 0, 0, 0, 0, 238.095238942196, 0, 1074.07407411394, 0, 0, 1750.00000027145, 4.58519593958821e-07, 2.0909832178017e-07, 5.05788627207728e-07, 0, 0, 0, 0, 0, 0, -6.25919645173948e-07, 0, -4861.11111101607, 0, 0, 0, -5833.33334364316, 0, -16111.1111100665, -5.50778201981385e-07, -5.43988540129617e-06, -2.05061533875695e-06, -2.97720785632367e-06, -5.300855026409e-06, -6.11642742020634e-06, 0, 0, 0, 0, 0, 0, 3.66056171702302e-06, 0, 11574.0740746127, 0, 0, 0, 5.44599108468939e-05, 0, 72916.6666614105, 4.92513991099461e-06, 1.38343951174425e-05, 1.11230568529894e-07, 0, 0, 53703.7037467599, 1.17768503650308e-05, 7.38088143380321e-06, 1.48175734838906e-05, 7.174464335173e-05, 1.95513070647847e-05, 3.9722647186921e-05, 3.22291918265236e-05, 0, 0, 0, 0, 0, 0, 5.48501711148531e-06, 0, -13888.8888897213, 0, 0, 0, -7.08392970277504e-05, 0, -173611.111053427, -2.99718617400913e-06, -1.48410360542682e-05, 2.14262042134822e-05, 0, 0, 0, -0.000129335044772033, 0, -243055.555565775, -1.65381570260132e-05, -3.99388144796764e-05, -1.77142702316044e-05, -0.000178343711524708, -7.64025664844679e-06, -0.000105903158522324, -7.80417394493344e-05, -5.80519063218058e-05, -0.00021313294968441, -2.35084413736237e-05, -0.000106728660321666, -0.000121253634157237, 0, 0, 0, 0, 0, 0, -3.0144211880881e-06, 0, 6613.75661688828, 0, 0, 0, 9.38916320917485e-05, 0, 208333.333272834, 4.76760546490117e-06, 1.1262025086485e-06, -3.09218063690086e-05, 0, 0, 0, 0.000234697459168261, 0, 578703.703661635, 2.71703397692921e-05, 2.96527811104076e-05, 3.33286377036524e-05, 0.000231458319560035, 1.25841091507879e-06, 2.41910830175796e-05, 0, 0, 0.000204218841848225, -2.80899996998322e-05, 7.5909748929366e-05, 7.57306438383885e-05, 0.000636013156607794, 3.84911061023268e-05, 0.000369953567204488, 5.18374750174341e-05, 0.000429881063582658, 0.0003186043679064, 0.000286857141600391, -0.00016776056958096, 0.000854699029293106, 0, 0, 0, 0, 0, 0, 4.75293423132492e-06, 0, -3.62437477154236e-06, 0, 0, 0, -3.04568111554744e-05, 0, -99206.3491731005, -3.74907006935665e-06, -1.32375085775395e-05, 8.7583570474167e-06, 0, 0, 0, -0.000149977434051303, 0, -694444.444344746, -2.06134032785803e-05, -1.14001946755831e-05, -1.38282590269376e-06, -0.000159432365198815, -3.14340723873402e-05, -6.29826968723328e-05, 0, 0, 0, -0.000258703422066417, 0, -6.06739808665004e-05, -2.38564823530966e-05, -6.58356800171341e-05, -7.81555283188074e-06, -0.000468727165925826, -0.000125095959917096, -0.000164080884918534, -9.56768302725322e-05, -0.000318705453389796, -7.96983087988848e-05, -9.07949218917284e-05, -0.000139149823502306, -0.000794041064580346, -0.000526474051351885, -0.000243319688228739, -0.000541303081256921, -0.000968553353556906, -0.000210469607862377, -0.000303218863160171, -0.000217949178647759, 0, 0, 0, 0, 0, 0, -1.90262680852616e-06, 0, 1.46764425329811e-06, 0, 0, 0, 9.57470724408212e-06, 0, -1.15821152116109e-05, 4.20199636490412e-07, -3.66302964451029e-06, -1.80534816736071e-06, 0, 0, 0, 5.31326711473951e-05, 0, 330687.83064028, 4.82772020258235e-06, 6.58070017541246e-07, 7.28149113523147e-06, 4.48053690463343e-05, 1.17574449926148e-05, -8.26872972066761e-06, 0, 0, 0, 0.00014075441452213, 0, -1.05959385643614e-06, 4.00548676213409e-05, 3.83174493964634e-05, 1.18191075922513e-05, 0.000194776496584964, 2.89447599141644e-05, 1.32765633937055e-05, 1.59608705961367e-05, 0.000101408047002485, -7.23759445488422e-05, 0, 0, 8.22852237592608e-05, 3.77201054705812e-05, 4.51435585306563e-05, 2.83112478229229e-05, 0.000280428967773816, 6.58763958907749e-05, 0.000245569625596112, 6.63599154587926e-05, -1.34594047115307e-05, 0.000136035596973777, 0.000257811272567856, -0.000153624870260318, 6.9697353468329e-05, 0.000204158350448508, 6.36096950629035e-06, 0.000107909388324508, 0.00032681523848714, 0.000134506519524214, 0.000253163614767319}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 6.24999999865444, 0, 0, 0, 0, 0, 0, 0, -62.499999948381, 0, -162.053571385379, -3.68443808051343e-08, 0, 0, 0, 0, 0, 0, -3.75690280353619e-07, 0, 1628.47222187893, 0, 0, 1620.53571418298, 2.00787709817712e-08, 6.99005229290268e-07, 5.7323204681475e-07, 0, 0, 0, 0, 0, 0, -1.12794800117327e-07, 0, -8394.09722089692, 0, 0, 0, 3.01738393821252e-06, 0, -16284.722220089, -1.2389851806757e-06, -2.47776950998487e-06, -3.08281884956913e-06, -6.26465610728131e-07, -3.32264580401001e-06, -7.92923771668645e-06, 0, 0, 0, 0, 0, 0, -1.58201855059781e-06, 0, 24305.5555517733, 0, 0, 0, -1.39671538147691e-05, 0, 83940.9722062482, 1.63788610615254e-06, 1.06502895233101e-05, 1.08672152755153e-05, 0, 0, -2.34140720616979e-05, -4.9756493279687e-06, 1.83462310884115e-05, 7.90227887405216e-06, 3.43770137235948e-05, 2.08151082436558e-05, 3.40397463926903e-05, 1.72439064409361e-05, 0, 0, 0, 0, 0, 0, -2.21960411234318e-06, 0, -39930.5555497696, 0, 0, 0, 1.05393772361153e-05, 0, -243055.555542234, -5.3597661219835e-06, -2.32778059607098e-05, -2.4708000233577e-05, 0, 0, 0, 3.94935342689795e-05, 0, 2.21131840850685e-05, 1.70367454655749e-05, -2.68567615880984e-05, -3.2707765840487e-05, -9.0015036190462e-05, -6.59248459891279e-05, -4.90585718517004e-05, 3.63676443914461e-05, -8.33975976903411e-05, -0.000169966218482795, -5.09355038145497e-05, -6.47312570678682e-05, -7.1334092909893e-05, 0, 0, 0, 0, 0, 0, -1.12120175291224e-06, 0, 34722.2222174039, 0, 0, 0, -1.86093921280916e-05, 0, 399305.555532595, 4.80001256952841e-06, 3.09700713598062e-05, 2.27788656292206e-05, 0, 0, 0, -5.65214481489204e-05, 0, -3.48433023898137e-05, -3.14757011025864e-05, 3.18363971861757e-05, 6.28418395763879e-05, 0.000124571567418742, 8.10047096744565e-05, 6.8277952032885e-05, 0, 0, -8.22016847917799e-05, -4.53260173781054e-06, 0.000142228995332322, 2.77406653061222e-05, 0.000278461173252379, 0.000192182448135481, 0.000259907062466997, 3.73434342571384e-05, 8.61307523114253e-05, 2.46480967186362e-05, -8.15931488627766e-06, 0.000344688103215653, 4.3896208861204e-05, 0, 0, 0, 0, 0, 0, -3.81680944108484e-07, 0, -12400.7936483981, 0, 0, 0, 2.96460826551968e-07, 0, -347222.222207086, -2.97012132376362e-06, -9.2277764934785e-06, -8.94242624766984e-06, 0, 0, 0, 3.04547518363716e-05, 0, 2.27632968812768e-06, 2.54679847802565e-05, -1.92803504022077e-05, -6.3470732208015e-05, -8.23345059200077e-05, -3.78138930180306e-05, 2.15437359649398e-05, 0, 0, 0, 7.12333744557625e-05, 0, 3.18405728270389e-05, 4.79407534606407e-05, -9.11508851667989e-05, -6.91226615335148e-05, -0.000275660715782975, -0.0001070836187964, -0.000268053647961548, 5.03188085040311e-05, -0.000116501122715119, -4.35262996276192e-05, 1.70476719386402e-05, -0.000175913837377858, -0.000298337721756167, 0.000238356644859522, -0.000135222471073912, 0.00011439002098402, -0.000228517062842315, -0.000153867191055394, -0.000360733273671279, -0.000317202030809447, 0, 0, 0, 0, 0, 0, 3.62153116123119e-07, 0, -7.82723304035646e-07, 0, 0, 0, -1.6452520073447e-06, 0, 124007.936506135, 1.17506119615095e-06, 4.85153105414847e-06, 1.91883720009461e-06, 0, 0, 0, -1.44884178590384e-05, 0, 5.80225321415481e-07, -5.9572527682681e-06, 6.19592446320092e-06, 1.41938658339702e-05, 2.17324466908115e-05, 9.75754949973166e-06, 2.54789738588914e-06, 0, 0, 0, -4.92310494593552e-05, 0, -1.04522365174685e-05, -3.74030512617394e-05, 1.60822801085319e-05, 2.77360681635794e-05, 5.31777976193725e-05, 4.36079030438946e-05, 8.19072275746547e-05, -7.14689525734517e-06, 4.12307579792703e-05, 4.40872271301282e-05, 0, 0, -2.61077862065135e-05, -1.99064208829987e-05, 7.01225741485447e-05, 8.56686342791397e-06, 0.00016421826634857, 7.01021447184015e-05, 0.000186249918178019, -4.827409863079e-05, 0.000109920983632289, 1.65594428717834e-05, 0.00015437166201574, 0.000163194296710969, 0.000139083540228375, -9.70566945650097e-05, 0.000168646699066068, 0.000350045387431589, -7.37924743182064e-05, 0.000184951945661933, 9.54816110439337e-05}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 11.1111111103804, 0, 0, 0, 0, 0, 0, 0, 2.50434005763476e-08, 0, -301.984126983069, 0, 0, 0, 0, 0, 0, 0, -2.12499686811944e-07, 0, 3255.18077612047, 0, 0, -5.34668054911309e-07, 1.66495597294162e-07, -1.18339566494946e-07, -1.12003964574712e-08, 0, 0, 0, 0, 0, 0, 1.37203459808067e-06, 0, -18541.6666674809, 0, 0, 0, 2.77950699080578e-06, 0, 2.22015948452949e-06, -4.62503440969594e-07, -2.85913976539167e-07, 4.29216290011846e-07, -1.16815854095592e-06, -1.02491693244643e-06, -2.59146922419724e-07, 0, 0, 0, 0, 0, 0, -2.40199134524617e-06, 0, 61863.4259285785, 0, 0, 0, -9.31488178737749e-06, 0, -1.69177734569157e-06, 1.2180841874692e-06, -1.32728944329714e-06, -3.33412023628481e-06, 0, 0, -1.15963873957831e-05, 4.6737469964021e-06, 4.01861793518884e-06, 1.02400714385495e-06, 4.72344573812741e-06, 2.51073583362157e-06, -3.26865252952147e-06, 1.43878480602126e-05, 0, 0, 0, 0, 0, 0, 4.96383379435791e-06, 0, -125000.000005845, 0, 0, 0, 1.81053454863735e-05, 0, 1.93532253537805e-06, -6.5531766787595e-07, 3.19957736965156e-06, 8.82779421595189e-06, 0, 0, 0, 3.24230859480025e-05, 0, 3.32038568531833e-05, -9.19631607877544e-06, -1.32394716235413e-05, 3.69954013558596e-06, -1.21380612957339e-05, 3.43586552210489e-06, 1.20332121999133e-05, -1.52718267220905e-05, -1.1184507368263e-05, -1.36219609370788e-06, -3.18303795292428e-05, -6.23659171716861e-05, -3.54324118794684e-05, 0, 0, 0, 0, 0, 0, -3.03919986999689e-06, 0, 150462.962968669, 0, 0, 0, -1.62016550580221e-05, 0, 2.818556160082e-06, -1.73388157284005e-07, -6.38810512449338e-06, -8.78819331306747e-06, 0, 0, 0, -4.89381271576773e-05, 0, -3.12942504613724e-05, 1.21000047022129e-05, 1.27297778903858e-05, -1.50766296386737e-05, 0, -1.15394569612034e-05, -4.90638293079467e-05, 0, 0, -6.05068945083862e-05, 6.04818541165642e-06, 9.52881159899969e-06, -1.08646571548531e-05, 2.75155998367278e-05, 1.03356844849903e-06, 2.08887192978345e-05, 5.4328461614208e-05, 8.18301088941475e-05, 9.23704072811566e-05, 0.000110959411562852, 2.61937325371694e-05, 0.000155739306888522, 0, 0, 0, 0, 0, 0, 1.92859838896111e-06, 0, -99206.3492096149, 0, 0, 0, 1.03643127232668e-05, 0, -2.46107597856394e-06, 4.3705196834075e-07, 2.35713565170404e-06, 4.61736620988208e-06, 0, 0, 0, 3.45819648705051e-05, 0, 2.31553651936866e-05, -8.61246330605656e-06, -7.18549371755697e-06, 2.01937782368321e-05, 1.96395540286757e-06, 8.13649655169514e-06, 2.64051772711065e-05, 0, 0, 0, 5.91475627234796e-05, 0, 5.35461432863859e-05, -1.29989932874627e-05, -1.64089740996559e-05, 2.12709759342917e-05, 7.04889355420101e-06, -8.44427578840062e-06, 4.50295181161744e-05, -7.84651366522877e-05, -3.84734588606545e-05, 1.25400472965208e-05, 2.77351198123803e-06, -1.1845764257747e-05, -4.1941252988297e-05, -0.000133444972785488, -7.99885224692804e-05, -0.0001370657995529, -0.000184599129736812, -0.000107755100210402, -0.000113998115576262, -7.4351866278887e-05, 0, 0, 0, 0, 0, 0, -4.81476120448377e-07, 0, 27557.3192248575, 0, 0, 0, -1.51792286009187e-06, 0, 5.74112380856386e-07, -2.48577149385847e-07, -1.22628401543601e-06, -7.07721915734783e-07, 0, 0, 0, -6.71508119412966e-06, 0, -3.3958051835579e-06, 1.93669223781406e-06, 1.40656609888163e-06, -4.75633455810386e-06, 2.03098874852605e-07, -1.98550356293099e-06, -1.2500475656416e-05, 0, 0, 0, -1.14168841883019e-05, 0, -1.31254435717379e-05, 7.62488901775342e-06, 7.43875137572286e-06, -7.88314318670606e-06, 1.09362166581212e-05, -3.34023242748365e-06, -2.33492981695987e-05, 9.28805993802109e-06, 4.81441647521922e-06, -2.26792008586098e-05, 0, 0, -2.10543955539551e-05, 1.43458894178652e-06, 1.64368173506598e-06, -5.8604083523435e-06, 8.15163089413333e-06, 3.66591376007657e-06, -1.46814977227066e-05, 2.48447251974693e-05, 8.26701700255531e-06, 3.10797940107955e-05, 4.65350202870328e-05, -2.96357611818459e-05, 7.97863145992348e-06, 5.91957811983446e-05, 1.01768294655293e-05, 6.63541483611271e-05, 0.000107023779803915, 0.000115982889034935, 9.45949835176194e-05}, 
                                                     {0, 0, -9.76467416909855e-08, 0, 0, 2.60940234422014e-06, 0, 4500.00000238487, 3.4032996371114e-06, 0, 0, -2.84591372760208e-05, 0, -64303.5714252338, 0, -64303.5715007953, -4.38419033497462e-05, 0, -6.44710371377582e-05, -64303.5714950127, 0, 0.000165177509801186, 0, 391991.071273562, 0, 391991.07217325, 0.000326677073366001, 0, 0.000533272566106784, 783982.143066698, 0, 0.000589282949500381, 783982.143474261, 783982.143904106, 391991.072116687, 0, -0.000567755666299479, 0, -1333402.77696644, 0, -1333402.78157499, -0.00131704587151392, 0, -0.0022990958347316, -4000208.33234709, 0, -4000208.33618799, -0.0035339128003144, 0, -0.00374632075119061, -4000208.33583458, -4000208.3398967, -1333402.78048816, -8000416.6751192, -4000208.33897073, -4000208.34238471, 0, 0.00121209252582655, 0, 2772048.60900161, 0, 2772048.62193414, 0.00316697474511309, 0, 0.00592935886499416, 11088194.4384366, 0, 11088194.4592401, 0.0107834882685745, 0, 0.0109582073851818, 11088194.4491065, 11088194.4641512, 2772048.61753111, 16632291.665853, 0, 0.0197658656071853, 33264583.3628268, 16632291.6900603, 33264583.3740479, 16632291.7202626, 11088194.4594057, 11088194.4705369, 33264583.3972308, 0, -0.00162393145861661, 0, -3611111.10788728, 0, -3611111.12918213, -0.0047927132217918, 0, -0.00955175506440848, -18055555.5427098, 0, -18055555.5934854, -0.01711801877478, 0, -0.0174301951246354, -18055555.5583722, -18055555.5881286, -3611111.12363206, -36111111.0927262, 0, -36111111.1219103, -0.0534342413756285, 0, -0.0459154431835233, -72222222.2717299, -36111111.1624733, -72222222.3135478, -36111111.2481756, -18055555.5714641, -18055555.5778089, -108333333.446305, -36111111.1625203, -36111111.2731327, -108333333.531518, -108333333.594396, -72222222.3748152, 0, 0.00132817624604713, 0, 2881944.44150524, 0, 2881944.46202943, 0.00457564759833974, 0, 0.0093528428813509, 17291666.6524199, 0, 17291666.7177288, 0.0141370321972107, 0, 0.0164119959961737, 17291666.663862, 17291666.6976147, 2881944.4625956, 43229166.6337194, 0, 43229166.7018955, 0.0711987712170358, 0, 0.0501265342967267, 86458333.3718972, 43229166.7260844, 86458333.439385, 43229166.8371997, 17291666.6783699, 17291666.6529873, 57638888.866884, 0, 0.100153672055177, 172916666.815783, 57638888.9855618, 172916666.85737, 57638889.2504852, 43229166.698756, 43229166.8590793, 172916666.963955, 172916667.116398, 86458333.4533153, 259375000.552006, 172916667.053566, 172916667.237839, 0, -0.000606036598528685, 0, -1289682.5382057, 0, -1289682.54892687, -0.00251230094010922, 0, -0.00507528679201685, -9027777.76947113, 0, -9027777.81240579, -0.00561914972680333, 0, -0.00892172951308308, -9027777.7730321, -9027777.79393635, -1289682.55427087, -27083333.3091379, 0, -27083333.3728749, -0.0447338920420078, 0, -0.0259261445952211, -54166666.6766061, -27083333.3669403, -54166666.7291519, -27083333.4360704, -9027777.7904445, -9027777.75084778, -45138888.8532796, 0, -45138888.8846048, -0.096862966495495, 0, -0.076572943347243, -135416666.754415, -45138888.9758532, -135416666.814736, -45138889.2195063, -27083333.3306022, -27083333.4080121, -135416666.876501, -135416667.015248, -54166666.6638548, -180555555.731482, -45138888.9367506, -45138889.2458516, -270833333.852019, -180555555.982554, -270833333.930052, -180555556.288174, -135416666.908547, -135416667.128535, -270833334.151548, 0, 0.000118222908017281, 0, 248015.872702232, 0, 248015.875047877, 0.000602497423779291, 0, 0.00116549650435577, 1984126.98210166, 0, 1984126.99346787, 0.000823445008743593, 0, 0.00219262655524437, 1984126.98228932, 1984126.98778501, 248015.877653785, 6944444.43775119, 0, 6944444.45960008, 0.0104326184628922, 0, 0.00522409078593577, 13888888.887651, 6944444.45154117, 13888888.9038473, 6944444.46849608, 1984126.99112102, 1984126.97493345, 13888888.8748634, 0, 13888888.893652, 0.0334666967378881, 0, 0.0203946840015399, 41666666.6838625, 13888888.917961, 41666666.7099783, 13888888.9958774, 6944444.4410946, 6944444.4426429, 41666666.7217675, 41666666.7680519, 13888888.8653288, 17361111.0991537, 0, 0.0394438723980835, 69444444.5007564, 17361111.1374543, 69444444.5159065, 17361111.2946074, 13888888.8884751, 13888888.9968025, 104166666.841215, 69444444.6115403, 104166666.88876, 69444444.7473557, 41666666.6904484, 41666666.7723182, 138888889.162058, 69444444.5971489, 69444444.7980046, 138888889.289637, 138888889.360217, 104166667.010388}, 
                                                     {0, 0, -1.69790284734904e-08, 0, 0, 4.8273656114463e-07, 0, -124.999999749497, 0, 0, 0, -5.52247414735705e-06, 0, 3366.07142459847, 0, 124.999995043473, -9.67042192286876e-08, 0, -1.2248028558018e-07, 124.999999840507, 0, 3.38171978363878e-05, 0, -35810.5158469988, 0, 3.59665109934467e-05, 8.84000583318338e-07, 0, 1.14013997088029e-06, -3241.07135808327, 0, 1.55462815708524e-06, -3241.07142701284, 2.81498917560403e-06, 1.42526545444682e-06, 0, -0.000123094394644852, 0, 200451.388811686, 0, -0.000147087691478627, -4.51761342486587e-06, 0, -4.57380812454095e-06, 32569.4439945607, 0, -0.000391159306698807, -8.18873296452526e-06, 0, -8.68657638232472e-06, 32569.4444367167, -1.43840989147279e-05, -8.24932455180131e-06, -1.64042970453385e-05, -6.85595862482529e-06, -1.58384760659732e-05, 0, 0.000278325967085388, 0, -653993.05544166, 0, 0.000398128864649785, 6.76968025861449e-06, 0, 1.40512787192533e-05, -167881.943140742, 0, 0.00100549663964784, 4.78680883416739e-05, 0, 2.22852017546022e-05, -167881.944412239, 5.71966580468609e-05, 3.96928707465094e-05, 0.00230079368819054, 0, 5.4046656995927e-05, 6.10881763619053e-05, 2.79518331952503e-05, 8.87023642791703e-05, 7.88187300077199e-05, -1.38098933357941e-06, 8.43946437924021e-05, 1.10729200498111e-05, 0, -0.000394035688709868, 0, 1284722.22214776, 0, -0.000723055546632417, -1.23359534247101e-05, 0, -1.00330585604415e-05, 486111.109149192, 0, -0.00134429989027321, -4.98111380381131e-05, 0, -5.31920870561586e-05, 486111.111039656, -9.13021082944294e-05, -7.75018255731699e-05, -0.00576667422674066, 0, -0.00526021242475521, -3.109482551967e-05, 0, -9.89103068624008e-05, -0.000202484679516147, -2.25143133813572e-05, -0.000263389359669331, -0.000139516020161664, -6.28849015696764e-05, -0.000155237069386566, -0.000285207286086957, 1.37913029985516e-05, -0.00031310150171024, -5.54676879484477e-05, -0.000205464450747861, -0.000151032464724787, 0, 0.000339378800937981, 0, -1493055.55555975, 0, 0.00081798941273399, 6.83583444957186e-06, 0, 1.72404484946829e-05, -798611.109492555, 0, 0.00100577528985451, 6.94481877741891e-05, 0, 3.81542931104228e-05, -798611.111019915, 0.000117598205761578, 7.48522064519255e-05, 0.00685146713457321, 0, 0.00576708960299983, 0.000202283821540227, 0, 0.000157644955807843, 0.000330593496264442, 5.13325566192135e-05, 0.000345455365301567, 0.000330010914046031, 0.000132180967678187, 0.000296440213345075, 0.0108987724460351, 0, 0.000297618508283385, 0.000552782641676338, 0.000141965075171767, 0.000604368103926753, 0.000374927815271567, -2.12728721153314e-05, 0.000265472652147436, 0.000301443082529612, 0.00057909416974588, -9.8913820463086e-05, 7.45749756325885e-05, -0.000281070652871183, 0.000659456322553438, 0, -0.00016246527334157, 0, 942460.317492334, 0, -0.000511025976260178, -4.5468353677794e-06, 0, -7.63825401449309e-06, 694444.443740435, 0, -0.000453467175584011, -3.75320214358263e-05, 0, -3.35987662602209e-05, 694444.444388048, -7.7467330663708e-05, -5.72728755150863e-05, -0.00386974581616052, 0, -0.00290307867882526, -0.000152176802336838, 0, -0.000110678986204537, -0.000274640841809207, -7.88043592235965e-05, -0.00027429210826755, -0.000235420776139615, -0.00014662321258608, -0.000195755286906371, -0.0092654858226334, 0, -0.00890189536720729, -0.000272570468330205, 0, -0.000175310459756733, -0.000511302098420864, -2.9713904480026e-05, -0.000504791325506313, -0.000447480295388163, -4.21556689947431e-05, -0.000361300657523377, -6.35875351404505e-05, -0.000595698584969939, -0.000154282237719927, -0.000655404180871941, 1.41647208682309e-05, -0.000359921213980707, -0.000421464282266473, 0.000251691785924644, -0.000621215075326689, -0.000521231593872389, 0.000591655533813452, -0.000151930310292878, -0.000409719498325811, 0, 3.31300513752905e-05, 0, -248015.873028314, 0, 0.000132290116349905, 1.07179763168079e-06, 0, 1.93021529257133e-06, -248015.872887221, 0, 0.000113155201645042, 2.22814428974586e-06, 0, 7.02636040077434e-06, -248015.873002297, 2.17055161153646e-05, 1.30617917315691e-05, 0.000841970557158231, 0, 0.000521038292067482, 1.96458002410541e-05, 0, 6.19782845144285e-07, 7.88021842268258e-05, 2.88307565429689e-05, 8.94373730044104e-05, 6.05886438232591e-05, 3.35202930007212e-05, 4.99836384170993e-05, 0.00276146362095778, 0, 0.00255864266290696, 5.19451895446674e-05, 0, 2.3335856613904e-05, 0.000214798324937356, 2.28619331696825e-05, 0.000152444180664649, 0.000105304485059712, 3.78529949349683e-05, 0.000108021466517637, 0.000109161910357601, 0.000237077403606238, 9.08418140912404e-05, 0.00408608736421185, 0, 8.95640862549564e-05, 0.000302407237772938, 5.56363021583584e-05, 0.000283155949747034, 0.000230305002800855, 1.26189106268828e-05, 0.000187628254622758, 0.000271069438036745, 0.000111929010960667, 0.00037230080078903, 0.000460290809247425, 2.48164501757902e-05, -9.27975225459705e-06, 0.00028950916957885, -0.000160317402201647, 0.000146910665406221, 0.000169709739293077, 1.04091871369472e-05, -0.00027640413850536}, 
                                                     {0, 0, 0, 0, 0, -1.58087756579079e-07, 0, -125.000000076891, 1.8113319534645e-08, 0, 0, 1.66417228435782e-06, 0, 125.000000123779, 0, 3366.07143039201, -2.74490730519289e-07, 0, -3.87939883976352e-07, 124.999999944343, 0, -9.34260338366166e-06, 0, 1.90033654180099e-06, 0, -35810.5158901728, 9.08621809882704e-07, 0, 3.22362964560624e-06, -3241.07143211797, 0, 4.67972878404326e-06, -1.45582013537437e-06, -3241.07142933339, 3.45747026637966e-06, 0, 3.13348763462165e-05, 0, -1.11286217656616e-05, 0, 200451.388969623, -3.59995040496913e-06, 0, -1.39264072351853e-05, -3.06795483423878e-05, 0, 32569.444495737, -2.27215338217567e-05, 0, -3.95401943746511e-05, 1.37931847936164e-05, 32569.4444561134, -2.79575011494599e-05, 2.85499156229427e-05, -6.40963035274965e-06, -2.41634759028186e-05, 0, -6.59156506896843e-05, 0, 2.57207483110002e-05, 0, -653993.05576982, 3.98523520372429e-06, 0, 2.98633258185921e-05, 0.000185669938994053, 0, -167881.944693193, 3.4994862939798e-05, 0, 0.000134573351507739, -4.48701774550454e-05, -167881.94448011, 0.000112651371495745, 4.14447387552283e-05, 0, 0.000147067173021581, -0.000154628051565864, -4.05353138675406e-05, -0.000147405386942387, 1.51092817980994e-05, 0.00012497061523178, 0.000177492327103206, -7.18029853407868e-05, 0, 8.77695463486622e-05, 0, -3.00538326445644e-05, 0, 1284722.22256117, -5.52038789822181e-06, 0, -3.98528824882389e-05, -0.000417009173625504, 0, 486111.111673073, -2.15578721500344e-05, 0, -0.000245125810740833, 8.31391389207666e-05, 486111.111158961, -0.000221324087499116, -0.000538709636662613, 0, 0.000219742908563656, -0.000269072073415252, 0, -0.000454513914524161, 0.000386054376752408, 0.000205669823666402, 0.00030159146818121, -6.05055199107619e-05, -0.000428276457705788, -0.00062907321457354, 0.000561430477707632, -0.000140532008388768, -0.000459700219718886, 0.000442346649435552, 0.000462141063016678, -0.00040783901225435, 0, -7.18361953879421e-05, 0, 1.71453058404753e-05, 0, -1493055.55587582, 2.15173508311056e-06, 0, 3.73912689987199e-05, 0.000465692621899513, 0, -798611.111772397, -8.83327891892837e-06, 0, 0.000246829519651856, -7.41297598537203e-05, -798611.111157223, 0.000225340323423121, 0.00108690503394338, 0, -0.0006435364624625, 0.000242907529696952, 0, 0.000584125567379093, -0.000509086246917735, -0.000293602136729245, -0.000401138482847398, -2.7865543597194e-05, 0.000743653529122416, 0.000809389363602077, 0.000452944479958457, 0, 0.000572106582443783, -0.00100320174073306, -0.000265593310101163, -0.000927352007785639, 0.000544703346403282, 0.000554856430435255, 0.00119542430182725, -0.000981845518147886, -0.00117562685848383, 0.0012981863063382, -0.00155001509817962, 0.000104806966469153, -0.000516597147116595, 0, 3.29724950500889e-05, 0, -3.44616906418811e-06, 0, 942460.317626976, -8.85183094743556e-07, 0, -1.53513008358642e-05, -0.000259443797060775, 0, 694444.444840832, 1.72275975756913e-05, 0, -0.000141806365713631, 3.318777813134e-05, 694444.444475999, -0.000132180592678101, -0.000887202403286161, 0, 0.000599830953369177, -9.98519711932587e-05, 0, -0.000379045576752821, 0.000330847296730369, 0.00019433757109458, 0.000261744185680105, 5.58749790892337e-05, -0.00058182863265843, -0.000684679393458202, -0.000927696340667873, 0, 6.07958166223138e-05, -0.000382012838571412, 0, -0.000574688943953195, 0.000873265117059086, 0.000401311682050174, 0.00065311329265791, -0.000432623628648793, -0.000805199555173379, -0.00138123866378, 0.00100877161587059, 0.00105514348143425, -0.0013606063596799, 0.00106028609138637, -0.000108132964045321, -0.00101245121779572, 0.00164540003275238, 0.000861982209362332, 0.00176366131802342, 0.000757615626931905, -0.00102824655646075, -0.000530863923057437, 0.000617873523791689, 0, -6.49446677124429e-06, 0, -2.83551091885819e-07, 0, -248015.873052865, 1.62038570768156e-06, 0, 3.19558896430875e-06, 5.76400901248275e-05, 0, -248015.873111909, -7.21925586452471e-06, 0, 3.12599542036398e-05, -5.92598172199678e-06, -248015.873017734, 2.79171502038051e-05, 0.000264319292715453, 0, -0.000190251092558793, -1.64770068577793e-06, 0, 0.00013074454851962, -8.22039531780764e-05, -4.97552559820614e-05, -6.68350242380949e-05, -6.73143380643403e-06, 0.000155093742018964, 0.000170472028851053, 0.000433666217685171, 0, -0.00014256042428627, 0.000103802044842822, 0, 0.000225308172461369, -0.000300534900564515, -0.000116754066448367, -0.000194121726866994, 5.03765185397128e-05, 0.000325564791169132, 0.00054091294100193, -0.000363924420842484, -0.000302347873051058, 0.000611091136934549, 0.000213232373479457, 0, 0.000191827670938854, -0.000475665391454953, -7.78572094412104e-05, -0.000416240699042335, 0.000376183233929484, 0.000219675783463728, 0.000583417148970127, -0.000803498048392336, -0.000506977316985145, -0.000737256922582244, -0.000473061232658787, 0.000709693624276258, 0.000576574389008066, -0.000964755907442591, 3.57497455954217e-05, 4.59566011852714e-05, -0.000585905406278714, -0.000787770220295532, 3.34137150765606e-05}, 
                                                     {0, 0, 2.32401261492494e-07, 0, 0, -6.18729666506607e-06, 0, -6000.0000030411, -2.80131591835893e-06, 0, 0, 6.66391588398683e-05, 0, 125738.095199634, 0, 65738.0953504445, 2.71940298183528e-05, 0, 7.96461047263906e-05, 65738.0952782416, 0, -0.000381533437239269, 0, -960908.729678141, 0, -303527.779054095, -5.77479841610308e-05, 0, -0.000651021376218485, -1264436.50776145, 0, -0.000982783589026017, -1264436.50857917, -607055.556106304, -303527.778180674, 0, 0.00129655346992845, 0, 3801388.88688653, 0, 766111.117987213, -0.000368182693544564, 0, 0.00253521720988736, 8368888.88377752, 0, 5333611.11304114, 0.00516802011828921, 0, 0.00724494922157267, 8368888.89192643, 2298333.33552597, 766111.110835421, 10667222.2294314, 5333611.11851025, 2298333.34046734, 0, -0.00274552178331126, 0, -8803472.21774012, 0, -1142361.13137917, 0.0022503555741787, 0, -0.00575226775178201, -27552777.7582469, 0, -12230555.5731838, -0.0127820261620813, 0, -0.0222065532627604, -27552777.7831445, -4569444.44523433, -1142361.10384111, -29837499.9807022, 0, -0.0362191252329464, -59675000.0299745, -29837500.0385797, -36691666.688901, -6854166.71021981, -12230555.5708324, -4569444.45537375, -36691666.7480818, 0, 0.00366045709869589, 0, 12430555.5495051, 0, 1006944.4790253, -0.00510011136079816, 0, 0.00814238069873159, 50729166.6316999, 0, 16458333.3883926, 0.0143922171579502, 0, 0.0367424353938101, 50729166.6675641, 5034722.21162388, 1006944.4264382, 78611111.0386713, 0, 55763888.8631643, 0.0902277271507166, 0, 0.0899478340842998, 157222222.273846, 78611111.1983963, 65833333.3478751, 10069444.5445649, 16458333.3263418, 5034722.17940583, 167291666.743386, 55763888.9797493, 10069444.5788868, 167291666.979737, 98750000.2888877, 65833333.5183808, 0, -0.0029878655755406, 0, -10555555.5506028, 0, -486111.145250419, 0.00573359366553065, 0, -0.00699232988335184, -53263888.8545915, 0, -12986111.1931323, -0.00354003109144491, 0, -0.035104862373323, -53263888.8803805, -2916666.6438741, -486111.093068625, -107986111.009865, 0, -57638888.8939523, -0.114563173329092, 0, -0.106932877788076, -215972222.255558, -107986111.207321, -64930555.5245588, -7291666.76323779, -12986111.0686414, -2916666.53696739, -110416666.547943, 0, -0.176775813319415, -331250000.10798, -110416666.850024, -230555555.608249, -9722222.56256201, -57638888.9439301, -7291666.75273642, -331250000.51473, -129861111.507459, -64930555.6133648, -345833334.087529, -230555556.212606, -129861111.794472, 0, 0.0013639343243102, 0, 4960317.45805455, 0, 99206.3673201935, -0.00320385615903215, 0, 0.00334616164052464, 29861111.0930453, 0, 5555555.61471595, -0.00519044466915019, 0, 0.0184374974108167, 29861111.1011931, 694444.426030757, 99206.34146492, 74999999.9348627, 0, 31250000.0343545, 0.0693152720814431, 0, 0.0635557633248318, 149999999.998073, 75000000.0489933, 33333333.2838695, 2083333.36562976, 5555555.52020209, 694444.326806558, 100694444.321837, 0, 76388888.8086514, 0.16352878779526, 0, 0.139711732154675, 302083333.388118, 100694444.601291, 156249999.974139, 3472222.54063746, 31249999.982612, 2083333.260284, 302083333.711353, 83333333.5409825, 33333333.1814094, 305555555.617257, 76388888.9993162, 3472222.52364022, 458333334.132319, 305555556.362686, 312500000.645333, 111111111.934724, 156250000.392745, 83333333.785018, 312500001.270746, 0, -0.000266705362055121, 0, -992063.491622848, 0, -0.004005593178139, 0.000701538492940464, 0, -0.000680806441836823, -6944444.44040148, 0, -992063.508685616, 0.00291590768064828, 0, -0.00413114505213682, -6944444.44099364, 0.00531040551485886, 0.00104197488737652, -20833333.3170322, 0, -6944444.46283078, -0.0152161425467542, 0, -0.0155390457504437, -41666666.6596601, -20833333.3418808, -6944444.42494022, 0.00127480057968131, -992063.483791422, 0.03547020416049, -34722222.1813517, 0, -20833333.3173361, -0.0560720893145992, 0, -0.0399833019843582, -104166666.670277, -34722222.2701893, -41666666.6380325, -0.100051833021573, -6944444.42819997, 0.0643355955193936, -104166666.766315, -20833333.3587583, -6944444.34928858, -34722222.1728263, 0, -0.066530932290715, -138888888.907132, -34722222.2783484, -104166666.664234, -0.182509088189488, -20833333.3439623, -0.0552938195535744, -208333333.622552, -138888889.214769, -104166666.827749, -34722222.5280931, -41666666.6844215, -20833333.3599834, -208333333.673358, -104166666.966576, -34722222.6005749, -208333334.042226, -138888889.549737, -104166667.184609}, 
                                                     {0, 0, -2.69044704980226e-07, 0, 0, 6.99300492563554e-06, 0, 7000.00000123205, 1.46243425639334e-06, 0, 0, -7.26847238453313e-05, 0, -164194.444353792, 0, -59194.4445317263, -3.74105942209644e-06, 0, -6.75228867065224e-05, -59194.4444412727, 0, 0.000398890543020713, 0, 1444046.29547442, 0, 206129.630828323, -0.000187715905565043, 0, 0.000558261147311397, 1300175.92485656, 0, 0.000945787596704041, 1300175.92624627, 412259.259102504, 206129.629640064, 0, -0.00129599776573925, 0, -6430138.88592341, 0, -378472.229224552, 0.00162344273458652, 0, -0.00212117272857064, -10279027.766934, 0, -4227361.10800807, -0.00447463746332373, 0, -0.00739728314064689, -10279027.7796227, -1135416.66368734, -378472.219864279, -8454722.2247114, -4227361.11665279, -1135416.66965819, 0, 0.00262418147317824, 0, 16370138.8830943, 0, 386574.095483219, -0.00584103491830625, 0, 0.00458036481071082, 39190509.2239472, 0, 7223379.6337888, 0.00948463892143129, 0, 0.0238121421257402, 39190509.2616585, 1546296.27833385, 386574.05968727, 29657175.8766192, 0, 0.0341028142286578, 59314351.8680993, 29657175.9606865, 21670138.8813874, 2319444.46099318, 7223379.63825629, 1546296.29561549, 21670138.9512423, 0, -0.00335139104151878, 0, -24930555.5488507, 0, -208333.37039985, 0.0113452327133387, 0, -0.00605066230333033, -81006944.3867803, 0, -6840277.81411551, -0.00788275109766587, 0, -0.0421122790405176, -81006944.4400795, -1041666.61907502, -208333.300002888, -93645833.1908706, 0, -44201388.7897342, -0.0790419197773281, 0, -0.0884637178556196, -187291666.697124, -93645833.4140918, -27361111.0384831, -2083333.35353773, -6840277.75183267, -1041666.60749235, -132604166.670257, -44201388.9747968, -2083333.40096342, -132604166.949006, -41041666.849227, -27361111.260485, 0, 0.00262736911645012, 0, 22499999.9954276, 0, 46296.3329890359, -0.0122202416489464, 0, 0.00467400599517174, 93217592.5404564, 0, 3402777.84648672, -0.00329962791491449, 0, 0.0425135702601296, 93217592.5781212, 277777.713571149, 46296.2584934313, 147916666.486548, 0, 35648148.0620639, 0.0972497669954857, 0, 0.113255202471375, 295833333.347008, 147916666.751426, 17013888.7308828, 694444.423988238, 3402777.70151256, 277777.637524843, 109490740.485366, 0, 0.155981197255075, 328472222.244913, 109490740.92651, 142592592.475712, 925926.108101006, 35648148.193249, 694444.447901274, 328472222.718514, 34027777.9195237, 17013888.909759, 213888889.49242, 142592593.224789, 34027778.2993565, 0, -0.00115542044021681, 0, -11111111.1094401, 0, -0.0193654151086168, 0.00692511997491755, 0, -0.00195671688578655, -56249999.9747461, 0, -694444.49944681, 0.00912632550001769, 0, -0.0230354952368004, -56249999.9864471, 0.0433229588936656, 0.0214377529293919, -114583333.225087, 0, -14583333.3123291, -0.0590942795781458, 0, -0.0745435460435443, -229166666.653359, -114583333.370365, -4166666.52673745, 0.0513113674406337, -694444.377610975, 0.119881117710869, -118055555.326003, 0, -62499999.7990982, -0.13725783412317, 0, -0.127633704813444, -354166666.672465, -118055555.713988, -72916666.4700067, -0.162869385836751, -14583333.298312, 0.115899794099616, -354166667.041595, -10416666.61121, -4166666.49145393, -249999999.920137, -62500000.1197034, -0.149993676943049, -375000000.722461, -250000000.819295, -145833333.701044, -13888889.442303, -72916667.0472381, -10416666.9969466, -145833334.491166, 0, 0.000218326310712108, 0, 2314814.81457266, 0, 0.00424089124332858, -0.00160018006893077, 0, 0.00033510495581946, 13888888.8837054, 0, 0.016308422998641, -0.00394729634086822, 0, 0.00517473220232336, 13888888.8845524, -0.0115654681564797, -0.00491521578074398, 34722222.1967699, 0, 2314814.82023918, 0.0132503654173684, 0, 0.0202963218317974, 69444444.4343284, 34722222.2261332, -0.0447201252033405, -0.0249723413915407, -0.0196467598071236, -0.0355244878566362, 46296296.2251729, 0, 13888888.8335411, 0.0470112512844381, 0, 0.0394208986288851, 138888888.880816, 46296296.3426235, 13888888.7990533, 0.0464277687898814, 2314814.78820319, -0.0690483899289962, 138888888.986511, -0.0723205993808066, -0.10157988982594, 34722222.123732, 0, 0.0553910199278748, 138888888.864745, 34722222.2850754, 69444444.3671623, 0.104134688808859, 13888888.9015629, 0.00404795937600187, 208333333.614945, 138888889.225152, 34722222.2261114, 0.162310346274272, 13888888.9023258, -0.00776889530271713, 138888889.136645, 69444444.7525683, 0.268120508533955, 138888889.590041, 46296296.8419822, 34722222.70507}, 
                                                     {0, 0, 1.06174582003253e-07, 0, 0, -2.42903335716446e-06, 0, -6299.99999831391, -9.76384865630157e-07, 0, 0, 2.07034237681632e-05, 0, 156174.999884323, 0, 40675.0000145659, -1.67471274530919e-06, 0, 4.66048949021646e-05, 40674.999982674, 0, -8.19662273850466e-05, 0, -1479874.99910113, 0, -104166.667212793, 0.000169930673630932, 0, -0.00041288802944237, -954041.664968127, 0, -0.000557938853231893, -954041.666715079, -208333.33291978, -104166.66661363, 0, 0.000147845874955865, 0, 7159513.88594074, 0, 132291.670578836, -0.00125582051526232, 0, 0.00175383105298034, 8283819.43107612, 0, 2306597.21448691, 0.00213016353592001, 0, 0.00465769143958828, 8283819.44496647, 396874.9957923, 132291.665435867, 4613194.44361173, 2306597.22514724, 396875.000205852, 0, -3.99855484158136e-05, 0, -19704513.8837638, 0, -83333.3460813896, 0.00419421908760997, 0, -0.00424552685044114, -35221874.9595572, 0, -2758680.54276274, -0.0033306341923289, 0, -0.0165965772163309, -35221874.9997126, -333333.313816031, -83333.3262670823, -18192708.2680806, 0, -0.0175786375964689, -36385416.6698938, -18192708.353769, -8276041.64689293, -499999.996223755, -2758680.56177661, -333333.332219598, -8276041.69493671, 0, -0.000292447378684708, 0, 32138888.8839281, 0, 20833.3555948329, -0.00790961186171706, 0, 0.0062774692852335, 80624999.9372618, 0, 1631944.44578224, 0.000867067995115601, 0, 0.0330521071603753, 80624999.9936775, 104166.621556488, 20833.3170759854, 66423610.9401421, 0, 19548610.9662586, 0.0346549905023009, 0, 0.050175595345476, 132847222.230713, 66423611.159284, 6527777.68916766, 208333.302110408, 1631944.43364211, 104166.642786136, 58645833.2998189, 19548611.1689782, 208333.34774352, 58645833.4869975, 9791666.71315381, 6527777.86319508, 0, 0.00050062246613078, 0, -30763888.8864328, 0, -0.0216013854309602, 0.0084103532482084, 0, -0.00542650713277005, -101388888.835114, 0, -381944.471566241, 0.00476891680632181, 0, -0.0368414942450191, -101388888.875355, 0.0561448206002043, 0.0184803734712257, -119965277.571481, 0, -10243055.401755, -0.0383960844395265, 0, -0.0726698235243137, -239930555.553919, -119965277.826216, -1909722.06267508, 0.07669012370442, -381944.40503715, 0.0561233492046107, -59201388.5706173, 0, -0.0707995613379702, -177604166.642834, -59201389.0133933, -40972222.0690929, -0.0282826052498511, -10243055.5972826, 0.00637785892633179, -177604166.956121, -3819444.37098, -1909722.26431438, -61458333.5997023, -40972222.6045444, -3819444.68021484, 0, -0.000338044173935958, 0, 15972222.2218334, 0, 0.0109452743776671, -0.0048034106220759, 0, 0.00257311676210412, 65972222.1977616, 0, 0.0284564591486828, -0.00613036925519094, 0, 0.0216257549777558, 65972222.2107702, -0.036076916699846, -0.0105681135112986, 104166666.54732, 0, 2083333.26151613, 0.0228619564580823, 0, 0.0547207316662944, 208333333.318461, 104166666.68437, -0.129202691002068, -0.0751966374734921, -0.03728839636496, -0.0465896472790455, 76388888.618072, 0, 24305555.2928947, 0.0544443353119883, 0, 0.0639106342452605, 229166666.650016, 76388888.9948698, 10416666.4649565, 0.00759868585796394, 2083333.32306672, -0.0495387813722267, 229166666.892901, -0.19751774009949, -0.0631292877059199, 97222222.1114597, 24305555.6471989, 0.031933633311487, 145833333.717043, 97222222.7320753, 20833333.3741439, 0.170765642896369, 10416666.928436, 0.188113360808516, 20833333.9551074, 0, 8.55957035614698e-05, 0, -3472222.22229338, 0, -0.00227066282356013, 0.00112900439408945, 0, -0.000509154971878086, -17361111.106444, 0, -0.00925680452058246, 0.00231779683090846, 0, -0.00520390862955729, -17361111.1076107, 0.0093730804314251, 0.00244740714317856, -34722222.1952889, 0, 0.00990574675549805, -0.00524861576416656, 0, -0.0167534732624169, -69444444.4361291, -34722222.2223553, 0.0391445655536608, 0.026766890437555, 0.0115143023686341, 0.0131444436074826, -34722222.1410981, 0, -3472222.14429353, -0.0181308723397659, 0, -0.0229379148707723, -104166666.656867, -34722222.2527675, 0.0835101067011236, 0.00420061409739911, 0.0135402315805977, 0.0266956694043265, -104166666.724478, 0.102976566806178, 0.0430926889290317, -17361110.991842, 0, -0.0225604311583678, -69444444.4128375, -17361111.1577742, -17361111.0310866, -0.0229201847003671, -3472222.23732513, 0.00813753877006015, -104166666.831022, -69444444.6552742, 0.0981353456308135, -0.0042287785427431, -0.0289754065839772, -0.0144004888642972, -34722222.3121723, -17361111.3120022, -0.117076392644144, -34722222.632085, -0.243760529229527, -0.286461793996234}, 
                                                     {0, 0, 7.38542428298585e-08, 0, 0, -2.40933180131873e-06, 0, 4199.99999692143, 4.98206987876822e-07, 0, 0, 3.15090679661504e-05, 0, -107616.666571014, 0, -20116.6666296762, 1.19641547304696e-06, 0, -2.02805634710034e-05, -20116.6666558406, 0, -0.000218219069684507, 0, 1067513.88820782, 0, 35916.666624954, -7.27214746176436e-05, 0, 0.000195353711482903, 490930.554004659, 0, 0.0001797721517411, 490930.555524983, 71833.3330668392, 35916.6666539597, 0, 0.000879714865511764, 0, -5460277.77567579, 0, -28333.3340515748, 0.000488862020822369, 0, -0.000930084854705393, -4515208.32243025, 0, -833263.880903359, -0.000368923936258252, 0, -0.00170997289850738, -4515208.33330787, -84999.9975694557, -28333.333185603, -1666527.77665319, -833263.889984498, -84999.9996407236, 0, -0.00215701839964364, 0, 15968402.7744392, 0, 8333.33651466148, -0.00149169260660529, 0, 0.00246085859195621, 20661805.5237276, 0, 623611.093143791, -0.000294304982187573, 0, 0.00704574197972231, 20661805.5546672, 33333.3229257888, 8333.33223826122, 7058680.5005324, 0, 0.00413853349595124, 14117361.1100545, 7058680.56340195, 1870833.32028049, 49999.9935365721, 623611.114995379, 33333.3338744369, 1870833.34139201, 0, 0.00325418417670678, 0, -27652777.7750866, 0, -0.00557117315406736, 0.002701664403181, 0, -0.00390992427470006, -51402777.7297472, 0, -173611.094042617, 0.00198746972497178, 0, -0.0160391195383127, -51402777.7735757, 0.0225297563004129, 0.00264814631343091, -28402777.6394437, 0, -4826388.76245411, -0.00418494472871662, 0, -0.0153720884215211, -56805555.5549566, -28402777.7960592, -694444.39421502, 0.0281634270753389, -173611.112236491, 0.000195523333007633, -14479166.6413148, -4826388.91497536, 0.00267515966087895, -14479166.719932, -1041666.66055615, -694444.480775997, 0, -0.00294666786123117, 0, 27986111.1103929, 0, 0.00469384673581943, -0.00277312091426864, 0, 0.00359153031225547, 70208333.2934708, 0, -0.00220897568455121, -0.00319492707357166, 0, 0.0194539875382997, 70208333.3257958, -0.026723558095059, -0.00299557387263865, 57673610.9476282, 0, 1215277.63608399, 0.000997745801844882, 0, 0.0272213353124642, 115347222.217027, 57673611.1286082, -0.0843648569558468, -0.0535881889785427, -0.00721848858824168, -0.00368548762344009, 16666666.4048132, 0, 0.0108397009982619, 49999999.9779522, 16666666.7192636, 4861111.02411065, -0.0197392518666335, 1215277.80375227, 0.00190751687811563, 50000000.1057123, -0.0901673235752318, 0.0433199100129473, 7291666.73125475, 4861111.26481206, 0.0667680683388097, 0, 0.00146798429686004, 0, -15277777.7781276, 0, -0.00172211003549624, 0.00158801844474633, 0, -0.00180040180214804, -49305555.5381496, 0, -0.00617694105533553, 0.00229505661135058, 0, -0.0121038050522977, -49305555.5495418, 0.0165779356405951, 0.00164801961267609, -56249999.9073502, 0, 0.0737659701963124, -0.000145407975140907, 0, -0.0243822401793304, -112499999.990771, -56250000.0049845, 0.0657511135766948, 0.0445403355499384, 0.00923804090698229, 0.00291597483330445, -25694444.2259061, 0, -3472222.00315039, -0.00266168910941152, 0, -0.0146038510796636, -77083333.3190475, -25694444.4886356, 0.106098892794975, 0.0345166636499997, -0.004701575538184, 0.00202572732461173, -77083333.4189491, 0.140400828182158, -0.00745400179861709, -13888888.8261854, -3472222.26585301, 0.00557781039198516, -20833333.4568259, -13888889.092341, 0.0551676769778356, -0.00497767129415798, -0.127052549456881, -0.0871056008475838, -0.209807169412722, 0, -0.000309151844927612, 0, 3472222.22242678, 0, 0.000174445981471459, -0.000378344618153663, 0, 0.000372513401964506, 13888888.8857291, 0, 0.00270511405867409, -0.000674058703945493, 0, 0.00306219228187113, 13888888.8871056, -0.00422397906586541, -0.000362485890884641, 20833333.3129597, 0, -0.0136148727108073, 0.000164916651943614, 0, 0.0083916795887937, 41666666.6623158, 20833333.3325498, -0.0194420780705914, -0.014041140901281, -0.00310392690414573, -0.000340405906077401, 13888888.8243088, 0, -0.0660832893450223, 0.000473325024789293, 0, 0.00756758156994971, 41666666.6605659, 13888888.9016358, -0.0420697339570566, -0.0145033714616845, -0.00242749344917221, -0.000808651073899438, 41666666.687695, -0.0624548410852359, -0.00410109599775522, 3472222.12504776, 0, 0.00186234418682043, 13888888.8710653, 3472222.2429018, -0.0414972523411917, -0.00680010170809732, 0.00931842531906088, -0.00386581941896881, 20833333.3925087, 13888888.9714058, -0.0770285033285383, -0.0359300887785699, 0.0268126245309492, 0.0245629256970679, 0.0101569457986285, 0.0854574421011136, 0.034703579664375, 0.149632903135645, 0.0597031401912711, 0.115541016342534}, 
                                                     {0, 0, -1.15131707090137e-07, 0, 0, 3.36299753178514e-06, 0, -1999.99999759478, -5.49336093831218e-08, 0, 0, -3.96815763418696e-05, 0, 52388.8888356433, 0, 6722.22218200695, -1.43245222029601e-06, 0, 3.02040035391032e-06, 6722.22221797084, 0, 0.000250907415745651, 0, -535990.740380268, 0, -7499.99978049465, 2.30524872746506e-05, 0, -3.49746467200341e-05, -168490.7398314, 0, -9.10385924924712e-06, -168490.740715513, -14999.999901824, -7499.99998857975, 0, -0.000940421306428773, 0, 2851111.11003336, 0, 2777.77717508339, -0.00012142641566805, 0, 0.000196826732969216, 1611249.99397594, 0, 179583.328403161, -0.000133798409044573, 0, 0.00021256244318464, 1611249.999902, 8333.33252876781, 2777.77773302102, 359166.666148622, 179583.333563202, 8333.33308537878, 0, 0.00218028909758642, 0, -8731249.99837548, 0, 0.00118209376724161, 0.000300490018032872, 0, -0.000558269930007818, -7764351.83452467, 0, -63425.9137237816, 0.00074727741263677, 0, -0.00120591535208078, -7764351.85129298, 0.00325756550777456, 0.000239866594270098, -1596527.74691424, 0, 0.000403463044756384, -3193055.55441567, -1596527.77955498, -190277.773097376, 0.00298687872286616, -63425.9271842087, 0.00016638694779719, -190277.779037785, 0, -0.00315175117201763, 0, 15902777.7766309, 0, -0.002109377418603, -0.000488265937089034, 0, 0.00094342234199753, 20590277.7519142, 0, -0.0148785082754119, -0.0014540137781216, 0, 0.00340315432311054, 20590277.775996, -0.00665930812940863, -0.000594584343273666, 6874999.92359197, 0, 520833.261662267, -0.00257612657159254, 0, 0.000878828315703245, 13749999.9977727, 6875000.00394016, -0.0164197951537149, -0.0105900739280158, 0.00117066387678296, 0.000441462598168517, 1562499.98973628, 520833.340592532, -0.0040655587983796, 1562500.01107109, -0.00769394515856281, 0.00971574636100021, 0, 0.00276175381748323, 0, -16944444.444359, 0, 0.00280552397423078, 0.000439648471505806, 0, -0.000890871596959593, -30254629.6084496, 0, 0.00862676921932838, 0.00144983653590682, 0, -0.00449471033090383, -30254629.6268701, 0.00764068853944941, 0.000801914303518044, -15277777.6880602, 0, 0.0812994461520166, 0.00494847199338966, 0, -0.00335068907453736, -30555555.5513256, -15277777.7811194, 0.0262546697712884, 0.0185247665760231, 0.000648191828664414, -0.000220923503230467, -1967592.44709828, 0, 0.00446341314656979, -5902777.76673822, -1967592.60551497, 0.0291388342959607, 0.0131539213646999, -0.0088992398142669, 0.00200913168789494, -5902777.79976567, 0.0369660923740205, -0.0177322146019338, -0.00530838501205361, -0.0406661163219421, -0.00863767876728602, 0, -0.00134110687241693, 0, 9722222.222573, 0, -0.00210029634880617, -0.000248817110045076, 0, 0.000457575170365034, 22916666.6575687, 0, -0.00193575750383286, -0.00062894855662754, 0, 0.0029040333379704, 22916666.6646046, -0.00464376406827815, -0.000513411421927185, 16666666.6162501, 0, -0.0433250994770164, -0.00316561654504685, 0, 0.00453249107302946, 33333333.3288104, 16666666.6669361, -0.02003189094597, -0.0140333343383733, -0.00195835659375267, 0.000297810868107845, 3472222.10111416, 0, -0.121658294970375, -0.00609061987763586, 0, -0.00160268599952168, 10416666.6583165, 3472222.23268335, -0.0334509364793006, -0.0183186092444058, 0.00302550954037302, -0.00025705628064821, 10416666.685045, -0.0491067495152809, 0.00888717176504498, -0.0222265291033415, 0.0121407732107288, -0.00799313555278315, 0.0199179422724199, 0.0516007999336971, -0.0314723916392934, -0.0138271582615258, 0.0404626478080053, 0.023731558462683, 0.0423685458860044, 0, 0.000276762583926921, 0, -2314814.814962, 0, 0.000623328301575862, 5.92467470885251e-05, 0, -9.36052087490984e-05, -6944444.44282137, 0, 8.75424223299926e-05, 0.000101276838965377, 0, -0.00076112860600001, -6944444.44385868, 0.00117532651483654, 0.000134840489609858, -6944444.43349256, 0, 0.00847410563579263, 0.000629200135713078, 0, -0.00192864419888151, -13888888.8871435, -6944444.44392315, 0.00589706998154697, 0.00407480582286303, 0.00077693182086376, -0.000222325670858241, -2314814.77912871, 0, 0.036517403077949, 0.002133919661782, 0, -0.000701874808068191, -6944444.44122817, -2314814.81784386, 0.0127236039446837, 0.00644814428437557, 0.000116202555058594, -0.000546725564976178, -6944444.44823958, 0.0201165190173077, -0.00150585419204252, 0.0539392894510222, 0, 0.00215795917302149, 0.007107938360694, -0.00504989734153224, 0.0134450375545638, 0.0063626645785001, -0.00286462030486949, 0.00282042032177683, -0.0112940583060219, -0.0193228177710021, 0.0281484059654356, 0.0180113523880297, -0.0115839304946129, -0.0103221164603913, 0.00534178296538106, -0.0237349557642667, -0.0055424235791034, -0.0326055087244829, -0.00605628775154169, -0.0302204680948237}, 
                                                     {0, 0, 6.2896451806971e-08, 0, 0, -1.7997810388684e-06, 0, 642.857141801612, -2.89268977664609e-08, 0, 0, 2.07569435101779e-05, 0, -17107.1428380274, 0, -1357.14283736877, 6.50645520198194e-07, 0, 4.65947199844061e-07, -1357.14285599861, 0, -0.000128243329569168, 0, 178964.2855882, 0, 714.285582109148, -5.83744349210833e-06, 0, -2.83080119277273e-06, 34678.5710929528, 0, -8.25276365883003e-06, 34678.5714193865, 1428.57140649512, 714.285706673302, 0, 0.0004707882153161, 0, -979861.110738299, 0, 0.000491568276203772, 2.53308024704648e-05, 0, 4.44433339930777e-06, -341111.108953122, 0, -17499.9981407526, 6.47254921106938e-05, 0, 3.436064266086e-05, -341111.111067713, 0.000152756858016559, 4.6202330892357e-05, -34999.9998677687, -17500.0000009805, 8.59708750636852e-05, 0, -0.00107254058522757, 0, 3109027.7772276, 0, -0.00124790760811933, -4.72294526389036e-05, 0, -7.97303348723062e-06, 1707638.8826897, 0, -0.00476829454007068, -0.000259935564784954, 0, -5.08470571722639e-05, 1707638.88869948, -0.000594589678006424, -0.000193705612179294, 161111.100013288, 0, -0.000379243091510217, 322222.221794024, 161111.111242553, -0.000945112185566306, -0.00069169899279579, 0.000122753706209145, -0.000317112861211077, 2.38886420174856e-05, 0, 0.00152828262558377, 0, -5902777.7774136, 0, 0.00223795365465737, 6.96741738209455e-05, 0, -1.54709243286721e-05, -4756944.43518845, 0, 0.00629048716146592, 0.000419884136428825, 0, -6.14483217973143e-05, -4756944.44395852, 0.00112925877969838, 0.000425043497036835, -729166.639191523, 0, 0.0257046306340773, 0.000898194849331839, 0, 0.000580245238634023, -1458333.33222826, -729166.666976397, 0.00300358469000209, 0.00193988097029989, 0.000113065281556058, 0.000581191020395802, 0.00248238992472351, -0.00093992928047665, 0.00163688928743933, -0.000975288214821239, 0.00206730709457265, -0.000903592629023684, 0, -0.00132353360374182, 0, 6587301.58731271, 0, -0.00257641817741301, -4.57843934526195e-05, 0, 6.79285314608188e-06, 7430555.5479767, 0, -0.00446177443546476, -0.00042757438192852, 0, 0.000184423800915581, 7430555.55489561, -0.00129766182783149, -0.000521810921275615, 1736111.07876829, 0, -0.0289216147756819, -0.0016478822766659, 0, -0.000657331855060366, 3472222.22041218, 1736111.11124087, -0.00453341677991196, -0.00339912236767379, -0.000611753907457415, -0.00114050115283775, -0.0522643826688899, 0, -0.00206736646640343, -0.00351599562464761, 0.00118991972325277, -0.00578164575899457, -0.00319334641520022, 0.00116906175950851, -0.00188174170926983, 0.0014328730920512, -0.00717103372219601, 0.00250790371810659, -0.000921418878321586, 0.00601545365735184, -0.00116597136082508, 0, 0.000636548727493863, 0, -3968253.96840083, 0, 0.00165626480509502, 2.57244988881691e-05, 0, -5.57054576173442e-06, -6051587.29832726, 0, 0.00175706082274418, 0.000173669156131023, 0, -0.00011830107292344, -6051587.30113439, 0.000807242819378005, 0.000348181159638152, -2083333.31516275, 0, 0.0152242304644467, 0.00107698963750332, 0, 0.000136454084111824, -4166666.66509164, -2083333.33306649, 0.00342356679915852, 0.0024369595041184, 0.000843734611414137, 0.000818488796749292, 0.0438056058527624, 0, 0.0433410407979063, 0.00205503246887796, 0, 0.00131740820316951, 0.00313716524987525, -0.00100320141216407, 0.00605614378871809, 0.00413016251095959, -0.000146430819592537, 0.00156904039523837, -0.00148621838151107, 0.00857835769558372, -0.000823614438148202, 0.00515098993217154, -0.00144349754060098, 0.00264775607392672, 0.000233593203261616, -0.00703637983063012, 0.0070880293364746, 0.00414526316098303, -0.00672813978901562, -0.0022223562637608, -0.00344877762205526, 0, -0.000130321958359577, 0, 992063.492121013, 0, -0.000439736848588079, -5.78001549075115e-06, 0, -5.72276581760097e-07, 1984126.98354157, 0, -0.000370412726379858, -1.59581026136978e-05, 0, 3.57363153242099e-05, 1984126.98400694, -0.000209506029451485, -9.04170805218326e-05, 992063.488124575, 0, -0.00294232517317294, -0.000205363910758592, 0, 0.000119200411746036, 1984126.9836123, 992063.491898708, -0.00103497529844591, -0.000670048264465071, -0.000283345115363786, -0.000224851388663833, -0.0129515232147526, 0, -0.0127994891109609, -0.000646161136881396, 0, -0.000175919058755192, -0.00120971478879349, 0.000296597664163455, -0.00214412459189889, -0.00125648088651667, -0.000192980626102626, -0.000428488384154606, -3.83139180354474e-05, -0.00331716573768114, -6.00472541036968e-05, -0.0194505117286569, 0, -0.00076938369054397, -0.00203035957731031, 0.000415845834094035, -0.00270489765183983, -0.00180593675841327, 0.000237293163814013, -0.00124384656668156, 0.000315457170337208, 0.00198660872944729, -0.00511213600852253, -0.00392170824268978, 0.00190375554022377, 0.00132871893271081, -0.00237398931571383, 0.00354857710863908, 1.08120604545881e-05, 0.00306385229057853, -0.000199676427192275, 0.00421460579087331}, 
                                                     {0, 0, 1.02877823959983e-08, 0, 0, -2.8280044130075e-07, 0, -71.4285714370955, 0, 0, 0, 3.08461555529497e-06, 0, 1821.42856934181, 0, 785.714286911844, 3.25570803843584e-08, 0, 2.23011318511218e-07, 71.4285708788967, 0, -1.78261530396955e-05, 0, -17861.1110971387, 0, -714.285734056094, -3.7429948174199e-07, 0, -2.53426066332416e-06, -19964.2856791529, 0, -1.94168212547921e-06, -1749.99999389318, -714.285703391945, 2.47681961715094e-06, 0, 6.11235687872125e-05, 0, 89027.777740195, 0, 0.000130739973167658, 3.53531609205271e-06, 0, 1.41917614369779e-05, 194722.221988715, 0, 17499.9998026914, 3.19063063375093e-06, 0, 1.6205852447735e-05, 16111.1110774222, -8.50405638616495e-05, -7.43897732174092e-06, 17499.9999144133, -9.81489075162228e-06, -4.64443952080978e-05, 0, -0.00013063101040477, 0, -246527.777729011, 0, -0.000441398656141387, -1.4191197175833e-06, 0, -4.43014168842678e-05, -963194.443845576, 0, 0.000496197142036427, -6.26261008127949e-05, 0, -4.10962655360314e-05, -72916.6665399342, 0.000307239113773462, -2.02676796878586e-05, -161111.109808706, 0, 3.06435912091438e-06, -161111.110813392, -2.15885371394668e-05, 0.000557476743440895, 0.000405714034450828, -5.02169150649521e-06, 0.000107999733787053, 0.000343588229126261, 0, 0.000175799172640873, 0, 381944.444420182, 0, 0.000837364484120266, 6.79754986132278e-06, 0, 4.43132087535382e-05, 2638888.88811047, 0, -0.000554875863785489, -4.79034537903429e-06, 0, 0.000211193332835353, 173611.110802222, -0.000648075403330369, 5.36281508730085e-05, 729166.663798692, 0, -0.00335164375255481, -0.000238697016203761, 0, 8.40443999252401e-05, 729166.666082518, -0.000161171114002469, -0.00163327240927539, -0.00124543430470116, 0.0001242934732824, -0.000277810252453446, -0.00125241154877991, 0.000306329277546485, -0.000668155431376351, -0.00034158369744549, -0.00229323063856428, -6.77668141145837e-05, 0, -0.000144826731465422, 0, -307539.682550366, 0, -0.000907415329519039, -9.19329452448218e-07, 0, -6.69738256576645e-05, -4027777.77724534, 0, 0.00012616256780572, -3.22669418860965e-05, 0, -0.000195096019473936, -208333.332944904, 0.000705927895230249, -2.36907886925156e-05, -1736111.10811229, 0, 0.00438993038252285, 3.88025356035324e-05, 0, -0.000102629779326448, -1736111.11020712, 0.000387208726378236, 0.00256294108797078, 0.00172701424288514, -0.000194871870164221, 0.000283695959330029, 0.00592954581059311, 0, -7.17049610083679e-05, 0.00124189303361349, -0.000608736002881133, 0.00252121845194687, 0.00157919363425331, -0.000613791994660511, 0.00152722486989848, -8.37857756155135e-05, 0.00560928602279877, 0.000267581676576469, 0.00336338555675982, -0.000700623028194217, 0.0024372488551989, 0, 6.67036898283176e-05, 0, 99206.349223598, 0, 0.000521349074559739, 1.5579298614849e-06, 0, 3.24382538066716e-05, 3174603.17443456, 0, 0.000224998317972431, 5.14071066333485e-05, 0, 0.000147692225933317, 99206.3489404521, -0.00043103531293678, 6.32271709286245e-05, 2083333.33182626, 0, -0.00281349445682029, 4.5432568196517e-05, 0, 0.000236912949300767, 2083333.33256578, -0.000398205512202612, -0.00185650838424123, -0.0012016721660164, 0.000134059253630484, -0.000204107424222813, -0.0043441579455969, 0, -0.00554082432337083, 6.59228747988609e-05, 0, -2.36949484905887e-06, -0.00114946353587853, 5.89773688067233e-05, -0.00296739915557215, -0.0018476618170462, 0.000339156725060553, -0.000899659588765558, -0.000381560429807137, -0.00607325909760483, 2.26017570674554e-05, -0.0016321391817508, 0.000612445298332441, -0.001232991160815, -0.00156388036419161, 0.00141933268046466, -0.00597013752757169, -0.00535607987018858, 0.000295237888123976, -0.00176795632136808, -0.00155160381852906, 0, -1.31545587848619e-05, 0, -5.53592560696529e-06, 0, -0.000124130006338519, -3.71834693805589e-07, 0, -6.58552779054945e-06, -992063.492047844, 0, -0.000127363737088256, -3.18417567306032e-06, 0, -3.69479691171157e-05, 7.10192632164547e-05, 0.000109419022521232, -1.40132760793475e-05, -992063.491775912, 0, 0.000686892066237305, 1.0531281340985e-05, 0, -5.49163802830092e-05, -992063.491784821, 0.000154931475042395, 0.000547877483068244, 0.000350152113959526, -2.46628361164974e-06, 4.02640309628397e-05, 0.00115309310140005, 0, 0.00192393401006084, 8.54836164915925e-05, 0, -3.25946503078147e-05, 0.000400943824802562, -2.53423532781475e-05, 0.00111903707381691, 0.000601316326530542, -8.40878205779544e-05, 0.000301898712838399, 0.000251181339763146, 0.0021605307031115, -4.54822357739089e-05, 0.00209702844238493, 0, 6.71907198343395e-06, 0.000514460569669229, -0.000251103646374806, 0.00097108376510243, 0.000578070693334723, -0.000170463326175902, 0.000554988057165108, -0.000249579071043999, -0.000651728349258644, 0.00353458479071849, 0.00287567607278146, -0.000744137046697204, 0.000501983896004362, 0.00192551771323574, -0.000757235225654583, 0.00155635871961669, -0.000216774750468991, 0.00306539834122992, 0.000760417333515591}, 
                                                     {0, 0, 1.75410668542254e-08, 0, 0, -4.69653264214175e-07, 0, -55.5555557192653, -1.01258037060005e-08, 0, 0, 5.02991099982064e-06, 0, 1324.07407383062, 0, 888.888893727106, -2.14423735171566e-07, 0, 6.56064455266442e-07, 55.5555548609814, 0, -2.87075478850517e-05, 0, -11685.1851776443, 0, -3611.11116022921, 4.23905069624881e-06, 0, -6.14142930064381e-06, -21129.629640256, 0, -2.40299308623563e-06, -1268.51851263499, -833.333318891008, 1.33302811904147e-06, 0, 9.76733504769046e-05, 0, 49768.5184845281, 0, 2777.77802473956, -2.97028186770311e-05, 0, 2.66206941839574e-05, 185694.444395883, 0, 85231.4816210967, -5.10385184116832e-06, 0, 3.80368184619492e-05, 10416.6666327538, 2777.77766658213, 7.46764475322835e-06, 19027.7776855557, -2.38032391882689e-06, -5.96794928419835e-05, 0, -0.000207853363711475, 0, -108796.296219051, 0, -0.000707690884518709, 7.34711213348458e-05, 0, -6.44058184997009e-05, -785879.629385829, 0, -63425.9266291694, 0.000142020557143691, 0, -0.000222711687617164, -39351.8517077442, 0.000465288197568584, 3.03057955488419e-05, -740509.259195701, 0, -0.000164554298666096, -156249.999755769, 3.46920620675505e-05, -63425.9252383755, 0.000315062090868482, -8.41302293523796e-05, 3.23114732299763e-05, 0.000127408332862633, 0, 0.000279228768208483, 0, 115740.740638826, 0, 0.00120477045767237, -0.000140769627135866, 0, 0.000145944031295453, 1701388.88841782, 0, 0.00179856389856222, -0.000128303667700269, 0, 0.000385465577994776, 69444.4441298921, -0.000985156863063829, -7.29048399579913e-05, 3078703.70298826, 0, 520833.333570027, 0.00044114552041378, 0, 0.000337482181235508, 590277.777187358, 4.43920012119736e-05, -0.00240897425481562, -0.00128368979625739, 0.000117592657977692, 0.00010130003495941, 520833.331875015, 0.000137104816212649, -0.000283127656790829, -0.000141547858980115, -0.00165642388008485, -0.000225800225138126, 0, -0.000230041186995566, 0, -46296.2962162775, 0, -0.00119826111051785, 0.000131366891839622, 0, -9.84054416541252e-05, -1782407.40692884, 0, -0.00245788259349331, 0.00023081986049454, 0, -0.000564941032461731, -46296.2958517225, 0.00119961843240282, 1.70838030398764e-05, -6481481.48035105, 0, -0.000980925388955306, -5.42138276445481e-05, 0, -0.000963327657875165, -1041666.66572897, 0.000260076867614975, 0.00377670992914936, 0.00252000575575646, 6.1037611645608e-05, -0.000114001344185371, -1967592.59159905, 0, -0.000209892733346471, -1967592.59058747, -0.000126058349642923, 0.00443722329569152, 0.00108145102789073, -0.00022642933077772, -0.00162549795970102, -0.000256102441582587, 0.00640383347054799, -0.00156938368855765, 0.00184431822657725, -0.000587550448548524, 0.00117441557656227, 0, 0.000106091293737347, 0, -3.43666299762618e-05, 0, 0.000650182071945377, -7.48072586612434e-05, 0, 5.35687077955078e-05, 694444.444186235, 0, 0.00171354877516468, -0.00022047068399788, 0, 0.000331102280211839, -0.000287093377725964, -0.0007063339203818, -8.29273135429283e-05, 6481481.48070239, 0, 0.00121580089421996, -8.80595714695122e-05, 0, 0.000680785759506211, 694444.443591575, -0.000403355006882758, -0.00297140760393914, -0.00213987357996083, -0.000140547279804679, -4.2696338960587e-05, 3472222.22105013, 0, -0.000575529926340833, -8.47561111946343e-05, 0, 0.000410028816410492, 3472222.2210249, 0.000404141516218571, -0.00447862740967654, -0.00098382426163812, 0.000250934701339217, 0.000432345688002792, 0.00068726302329804, -0.0077786285255157, 0.000550904279180564, -0.00331635981199236, 0.00044010119451695, 0.000115367670133588, 0.000726449453900282, 0.00135572874926312, -0.00727016821578744, -0.00444826894206014, 0.0029129226885471, 0.0019912434592575, -0.00077230400079032, 0, -2.09691224182825e-05, 0, 6.22490799863041e-06, 0, -0.000146761812391828, 1.77756482514663e-05, 0, -1.21702398018838e-05, 5.63513532273105e-05, 0, -0.000483205718957777, 4.81476139749533e-05, 0, -9.39864231512506e-05, 7.74480634412115e-05, 0.000175771007038558, 1.51298558290723e-05, -2314814.8146088, 0, -0.000499580515968722, 9.63071332216795e-06, 0, -0.000242422226691802, 0.000270267690062769, 0.000169458564140699, 0.000826501339264774, 0.00066777962223567, 1.35075017432034e-05, 2.18241207019548e-06, -2314814.81440667, 0, 7.74583504728762e-05, -0.000142658453791713, 0, -0.00015265836380486, -2314814.81425929, 0.0001156980055757, 0.00177487843442692, 0.000527867223196966, -3.61131876848759e-05, -9.29637221153254e-05, 4.50204619312135e-05, 0.00344605178072632, -8.46414107991538e-05, 0.000461465904593842, 0, -5.30407698341319e-05, 0.000868168994885595, -0.000131449498979856, 0.00164354172384245, -0.000217750284640546, -0.000252670262138496, -0.000362215281064366, 0.000245279046937519, -0.000483131432757698, 0.00502823346001737, 0.0034280363265439, 0.000271447832284592, -0.000889577055703129, 0.00132201831071735, -0.000743161551237174, 0.000122148467592596, -0.00156284613073015, 0.00154974257744779, -0.00212145033842216}, 
                                                     {0, 0, 1.34557982225855e-08, 0, 0, -3.68553096360473e-07, 0, -50.0000000856722, 2.26948027303842e-08, 0, 0, 4.11097114105712e-06, 0, 1091.66666584269, 0, 966.666669818058, -1.59254475058859e-07, 0, -3.49094106368946e-07, 49.9999991551514, 0, -2.4550746736486e-05, 0, -8333.33332032013, 0, -5916.66670124516, 9.75107391520525e-07, 0, 4.07384053185721e-06, -21055.5555565891, 0, -4.27852035718091e-06, -1041.66665769021, -916.666656523135, 9.31651033456106e-06, 0, 8.73389804826872e-05, 0, 28124.9999390195, 0, 13333.3335218868, -2.49360366107701e-06, 0, -2.30238340512352e-05, 160069.444335533, 0, 128263.888977892, 5.22340257791838e-05, 0, -2.6396766309391e-06, 7291.66662100958, 4999.99992844205, -7.34209224205509e-05, 19097.2221048113, -6.33995442268312e-05, -4.89276845713103e-05, 0, -0.000193720230544876, 0, -41666.6665129441, 0, -8333.33390121353, 2.79962944532477e-05, 0, 6.40201955114996e-05, -536458.332850885, 0, -286111.111699336, -0.000252064230914574, 0, 0.000160900906205113, -20833.3331964386, -8333.33306644022, 0.000256726306688028, -967013.888528617, 0, -9.66047791381621e-05, -133680.555105182, 0.000207148982236539, -104166.6659984, 0.000181792005864494, 0.000294006933778344, 0.000302384527634736, 0.000573380949609086, 0, 0.000270149569462877, 0, 20833.3330972486, 0, 0.000995548990823912, -6.02412112787083e-05, 0, -0.000125896773517976, 784722.221279267, 0, 173611.11282183, 0.00028119626672615, 0, -0.000378494078642057, 20833.3330192112, -0.000556773144646845, -0.000599869014156387, 3194444.44279073, 0, 2118055.55528822, 0.000581411344901146, 0, 8.0184748964166e-05, 381944.443564382, -0.000652289924501116, 173611.109339489, -0.000273401612594432, -0.000837601955311653, -0.00122241514894976, 729166.66468896, -0.000552221266990873, 4.51475217307541e-05, -0.00127888229843101, -0.0026100185380548, -0.000794087569882506, 0, -0.000230071764074485, 0, 0.000217711486107445, 0, -0.00102109602703388, 9.14919091674577e-05, 0, 9.34957021689872e-05, -381944.443419987, 0, -0.00254880099690622, -0.000207145590277489, 0, 0.000540102677701625, 0.000354611576395805, 0.000690177637914834, 0.00073520746368189, -4548611.10867955, 0, -1215277.77843948, -0.00143652498854824, 0, 0.000734566428792566, -381944.44331762, 0.000902802910861029, 0.00271295687733612, 0.000320284994893061, 0.00126806057198909, 0.00264573631897469, -6770833.33069364, 0, -0.00109428847109173, -2083333.33083775, 0.00100365902796397, -1215277.77427641, -0.00207363904906341, 0.00137584326257986, 0.00243286616311383, 0.00203871596914093, 0.00562482022099944, 0.00330609092330101, 0.00595559248935472, 0.000885003307741284, 0.00202264128045772, 0, 0.000109253975537935, 0, -0.000111007432947145, 0, 0.000556601545109695, -5.38999866795085e-05, 0, -5.03037505703855e-05, -0.000596989816408634, 0, 0.00188512342846663, 6.64505692165015e-05, 0, -0.000304131332620998, -0.000233195805320882, -0.00053365332637215, -0.000396385149744839, 2083333.3316784, 0, 0.00131217209868228, 0.00104115958275905, 0, -0.000838404624541845, -0.000771970308212879, -0.000695228199307316, -0.00184611204564794, -0.000536512484406167, -0.00102509098333492, -0.00203972889656339, 9027777.7748966, 0, 3472222.22043353, 0.00167793706525333, 0, 0.000279425892283606, 2083333.3310372, -0.00134299594735785, -0.00359529599509886, 0.00245760669236535, -0.00146246589915664, -0.003014709881887, -0.00211909373066836, -0.00590129324505008, -0.0029832984996701, 3472222.2190687, -0.00136368430831639, 0.00188067740210282, -0.00601909019802069, -0.00129776143258966, -0.00923304253535591, -0.00223450341707418, -0.00360210304450945, -0.00253104600569235, -0.00231877930699672, 0, -2.21555456131338e-05, 0, 2.39472519159073e-05, 0, -0.000127353415661131, 1.23599569864324e-05, 0, 1.13213983159211e-05, 0.000148303406700398, 0, -0.000539185701763104, 1.67532391616048e-05, 0, 8.35286557677236e-05, 6.08254600470687e-05, 0.000135916752551682, 9.8045955204745e-05, 0.000453768571718715, 0, -0.000652735934511511, -0.000250441183690443, 0, 0.00027559204257951, 0.000237863080711089, 0.000228997306854076, 0.000578840401038189, 0.000283776915172467, 0.000316882160937499, 0.000647200521552508, -3472222.22125612, 0, 0.000328380272160904, -0.000550816301171214, 0, 0.000146690401616136, 0.000702836812246156, 0.000365470329203141, 0.00125075938031766, -0.000707085480480672, 0.000561348729947011, 0.00161457468913961, 0.000645013157128847, 0.00215495705223986, 0.00133334804978951, -3472222.22103657, 0, -0.000363112305844483, -3472222.22100719, 0.000510884016133377, 0.00141985212573138, -0.00149170602535053, 0.000722484501790245, 0.000385629760301591, 0.00117593781573831, 0.000657902936940839, 0.00481851894937511, 0.0014893968931075, 0.000485257939690725, 0.00163348015871584, 0.00436436454509709, 0.000622675431120754, -0.000164525713891158, 0.00188347499125115, 0.00352603929362855, 0.00157001629094628}, 
                                                     {0, 0, 0, 0, 0, -1.76544272492127e-07, 0, -50.0000001650757, -5.66492843779876e-08, 0, 0, 2.13122893770603e-06, 0, 966.666666146395, 0, 1091.66667227648, 7.58383124023925e-07, 0, 1.16746683437303e-06, 49.9999998690897, 0, -1.37224677501604e-05, 0, -5916.66665106355, 0, -8333.33339365533, -4.73682854323906e-06, 0, -9.37657435604638e-06, -21055.5555767329, 0, -1.20387878611597e-05, -916.666660668842, -1041.66665739744, -4.18394610473204e-06, 0, 5.19300150221214e-05, 0, 13333.3332495715, 0, 28125.0003156902, 1.98017997793794e-05, 0, 4.17249742721822e-05, 128263.888819996, 0, 160069.444720114, 7.07555629588149e-05, 0, 8.71325327688979e-05, 4999.99996015842, 7291.66657535289, 5.75051537110637e-05, 19097.2221400111, -1.93494567859293e-05, 3.13911223172575e-06, 0, -0.000120860049388985, 0, -8333.33310494738, 0, -41666.6675919339, -3.45699615729842e-05, 0, -0.000108072797870265, -286111.110640941, 0, -536458.33477521, -0.000199490383474482, 0, -0.00027696825767508, -8333.33317463559, -20833.3329738352, -0.000258617065083549, -967013.888938214, 0, -0.000381871179499209, -104166.666454754, 0.000296803255028879, -133680.554914062, 0.000262643149775101, -0.000126167000378703, -0.000250298020353083, 0.000359936561294268, 0, 0.000174854555623376, 0, -0.000367903471819254, 0, 20833.3349176839, 4.61216406559863e-05, 0, 0.000161970670250593, 173611.110077988, 0, 784722.226001705, 0.00041870590940677, 0, 0.000508944986057776, -0.000323144735254972, 20833.3325139175, 0.00066480757885069, 2118055.5541813, 0, 3194444.44576675, 0.000971358968256757, 0, 0.000807398224113528, 173611.110701608, -0.00104161350287566, 381944.442175837, -0.00091777916451238, 0.000724930427442963, 0.000801536940593054, 729166.665659341, -0.000560718364330847, 0.000389153385243864, -0.00131193859808566, -0.00299042330441012, 0.000729999874551067, 0, -0.000153134612284038, 0, 0.000356375348705504, 0, -0.00156484286693901, -3.54121629236958e-05, 0, -0.00015843256146266, 0.00116326182692719, 0, -381944.449613018, -0.000387683798110903, 0, -0.000544192391797947, 0.0003943071895226, 0.000991542856297839, -0.000761038084648667, -1215277.77516693, 0, -4548611.11512056, -0.00128234910866227, 0, -0.000937238185293133, 0.000585748780396343, 0.00159522046362304, -381944.440719908, 0.00179403875490857, -0.00144049540933839, -0.00141543833487694, -6770833.3318212, 0, -0.00154277785334153, -1215277.77688466, 0.00219555523817136, -2083333.33037563, 0.000372543722747867, -4.75384032942941e-05, -0.00169695414486952, 0.00222766572473543, 0.00848822049697697, -0.00363929096151433, 0.00481318946341701, 0.000340624305498512, 0.00266383516074474, 0, 7.42822719970844e-05, 0, -0.000190384644969528, 0, 0.000838047332650218, 1.5614013509003e-05, 0, 7.92293784845509e-05, -0.00070081936887833, 0, 0.00353030957059267, 0.000196577247949686, 0, 0.000339542239431813, -0.000253273466879587, -0.000552812841311854, 0.000520554693559865, -0.00190983207812749, 0, 2083333.33785226, 0.000960568086355082, 0, 0.000723800186630647, -0.000462538398221823, -0.00119567995377079, -0.00314775100071555, -0.00132613916629653, 0.00139485048349852, 0.00115745952279645, 3472222.21932472, 0, 9027777.77802463, 0.00149800201572946, 0, 0.00103119019504442, -0.000412302792423756, -0.00243589179763233, 2083333.330072, -0.000739821633219882, 0.000516478777238765, 0.00162105069637582, -0.0023482963535959, -0.00975721356622497, 0.00350549314551, 3472222.22115924, -0.00169047513234196, 0.000933620568338007, -0.00336337923103478, -0.00201704308236651, -0.00833974507012864, -0.00709131941468765, 0.00367729736017806, 0.00204810310326382, -0.000498120962552294, 0, -1.5310321797356e-05, 0, 4.29568541141206e-05, 0, -0.000186023693059336, -4.26405568381896e-06, 0, -1.74562852161666e-05, 0.000177109184689739, 0, -0.000962386119408347, -3.52193820519369e-05, 0, -9.20381223010058e-05, 6.78998213089828e-05, 0.000139455916005179, -0.000135436875450845, 0.000516853288704364, 0, -0.00172893133580601, -0.000225655246138765, 0, -0.000194852371704598, 0.000135692521037923, 0.000343233356753195, 0.000958886934512307, 0.000358233469308437, -0.000409878422648732, -0.000402047692994862, 0.00117168650931673, 0, -3472222.22301423, -0.000569580208736418, 0, -0.000327318728285862, 0.000168062576075394, 0.000927257058041197, 0.00140894701963195, 0.000192247307633126, -0.000376333696676126, -0.000593374465442751, 0.000906821179843209, 0.00373933775707356, -0.00160818718725377, -3472222.22128269, 0, -0.000614050209877434, 0.000203538963140697, 0.00120558256992579, -3472222.22147536, -0.000550526050790082, 0.000396385025972842, -0.000789121614797128, 0.00190806489147553, 0.00129264519968387, 0.00505207132555659, 0.00481710513035868, -0.00138710973119376, -0.00195758349242362, 0.00296734083407972, -2.7582186881991e-06, 0.000940977079339351, 0.000712316004533265, 0.00294016694725229, -0.00267900059684723}, 
                                                     {0, 0, 0, 0, 0, -9.65192221654724e-08, 0, -55.555555826331, -8.71964638061376e-08, 0, 0, 1.08768799325661e-06, 0, 888.888891162687, 0, 1324.07408092249, 1.00272430728292e-06, 0, 1.83549162369091e-06, 55.5555562313764, 0, -6.46380593892246e-06, 0, -3611.1111156458, 0, -11685.185247877, -6.66761344492293e-06, 0, -1.52993853862949e-05, -21129.6296880798, 0, -1.66273877395858e-05, -833.333335750104, -1268.51852752726, -1.44361390532902e-05, 0, 2.24444976206076e-05, 0, 2777.77776590217, 0, 49768.5188153343, 1.28446093390325e-05, 0, 6.55499727801853e-05, 85231.4816642376, 0, 185694.444921541, 7.09027986530312e-05, 0, 0.000110357667963559, 2777.77777634602, 10416.6667075463, 0.000135641474884792, 19027.7778497216, -2.61472111881552e-06, 8.36916816234573e-05, 0, -4.77943315259353e-05, 0, 7.5791984811294e-05, 0, -108796.297092067, -3.30436878331973e-05, 0, -0.000170724331234275, -63425.9261757029, 0, -785879.631566764, -0.000199607245937691, 0, -0.000440271587262493, 4.82284883444421e-05, -39351.8518889477, -0.000522217253829185, -740509.260546005, 0, -0.000494330503817268, -63425.9263558467, 0.000228112797195111, -156250.000245562, -0.000336831113717209, -0.000217277326395745, -0.000770838342146585, -0.000232896793025069, 0, 6.31882204321295e-05, 0, -0.000166200438148662, 0, 115740.741997948, 2.58887521045743e-05, 0, 0.000288669549280473, 0.000127470618419167, 0, 1701388.89315324, 0.000273519167751834, 0, 0.000863026005220092, -0.000150177502862607, 69444.4443978378, 0.00107949202564681, 520833.334839543, 0, 3078703.70815843, 0.00107429174439725, 0, 0.00120820433045534, 0.000957596508364774, -0.000801786281608925, 590277.777876458, 0.0001642278503812, 0.00111866877364749, 0.00256825083073768, 520833.335180778, -0.000623619266982138, 0.00164977144025668, 0.000222714401855382, 0.000296029273387126, 0.00106034559611161, 0, -5.05490501777352e-05, 0, 0.000189680602751853, 0, -46296.2974650906, -2.60779162120138e-05, 0, -0.000250056168329901, 6.64768589076044e-05, 0, -1782407.41256311, -0.000199751435851894, 0, -0.00107451192903872, 0.000227851199277684, -46296.2961928117, -0.00143470822852886, -0.000756319348119029, 0, -6481481.48925824, -0.00112179687870744, 0, -0.00193706555902528, -0.00102229411623894, 0.00136710382676011, -1041666.66602542, -5.143769067377e-05, -0.00204915774293046, -0.00442593625693468, -1967592.59675811, 0, -0.00204856313447566, -0.00377350618743514, 0.00206462695102233, -1967592.59563542, -0.00196180698216347, 5.62822712422705e-05, -0.00495526477387412, 0.000781641960819253, 0.00124585905517778, -0.00464955611985831, -0.00191920161982926, -0.000435352210903814, -0.00390985404900156, 0, 2.2396734824313e-05, 0, -0.000112647408025542, 0, 0.000586112552350574, 1.02998290180563e-05, 0, 0.000133337979681165, -0.000113666934241448, 0, 694444.447663459, 3.80760100470285e-05, 0, 0.000642246021342774, -0.000161976238485959, -9.48670458211886e-05, 0.000838657218431105, 7.21409183609356e-05, 0, 6481481.48801854, 0.000513016946201821, 0, 0.0013337357268162, 0.000533220684561925, -0.00102448031230364, 694444.443595463, -9.61315101254418e-05, 0.00159454250944023, 0.0033213890854404, 0.00148942378317143, 0, 3472222.22752103, 0.0015938080174046, 0, 0.00190414709623959, 0.00308643588407826, -0.00225809279903328, 3472222.22402289, 0.000963515420070958, 0.0009242940965662, 0.00467781560746138, -0.000800783464565068, -0.0028321176848769, 0.00585547436047579, 0.00473489960224802, -0.00167004477382109, 0.00403923773206235, 0.00168166017320275, -0.00170501042526971, 0.000408980161380387, -0.000601792186394167, 0.00266295312070226, 0.00777643571431471, 0.000206413689585549, 0, -4.2167979742755e-06, 0, 2.73774875493631e-05, 0, -0.000124146069122056, 6.7959163677273e-07, 0, -2.75946431805302e-05, 4.375084497966e-05, 0, -0.000804966054723849, -8.63421545803585e-06, 0, -0.00017262658248102, 4.50018677414558e-05, 3.75051428267574e-05, -0.000214931470131587, 4.25410294825438e-05, 0, -2314814.81692593, -0.000152565218391113, 0, -0.000368282845322492, -9.81328716455221e-05, 0.000288596226467556, 0.000353278435060431, 4.75909720722607e-05, -0.000598138243834815, -0.00105095524005674, -0.00010282332557392, 0, -2314814.81722678, -0.000488109913915364, 0, -0.000633339753086596, -0.000916054786851728, 0.000958293105926803, -2314814.8150973, -0.000293995729930933, -0.000533945472082262, -0.00180758416729859, 0.000336578021468957, 0.0015731713458954, -0.00202608077023763, -0.00110950920627942, 0, -0.000743757667306731, -0.0021602415586913, 0.00113272147853352, -0.0021907416981224, -0.0012055917649155, 0.000308224763219125, -0.00194184073941709, -0.000662287001266057, 0.00166696683847344, 0.000773778070411101, 0.00167681877575528, -0.00194384436440355, -0.0043762543377071, -0.0017072522122383, 0.0003703741524583, -0.00206530876519887, -0.000399655894223268, -0.000169243159992879, -0.0031883939623572}, 
                                                     {0, 0, 0, 0, 0, 1.24138445822814e-07, 0, -71.4285714906997, 3.87271665322158e-08, 0, 0, -1.42146562411736e-06, 0, 785.714286888922, 0, 1821.42857263506, -7.76257146203275e-07, 0, -7.04647030773323e-07, 71.4285719537874, 0, 8.70639881149717e-06, 0, -714.285718515412, 0, -17861.1111179665, 6.13565419429772e-06, 0, 3.82666673165085e-06, -19964.2857417562, 0, 1.47636297813217e-05, -714.28572026899, -1750.00000476325, -4.61428170527164e-06, 0, -3.19808492204354e-05, 0, -2.58950151987095e-06, 0, 89027.7777906831, -1.57777517818483e-05, 0, -6.7352891248003e-06, 17500.0001313856, 0, 194722.222406067, -8.54094841409148e-05, 0, -8.0584730756408e-05, 2.53642625794839e-05, 16111.1111312375, 2.88539706410682e-05, 17500.0000683945, 5.35734821030772e-05, 5.4346886890966e-06, 0, 7.36767272270349e-05, 0, 4.75767747226014e-05, 0, -246527.777762504, 3.49145308365282e-05, 0, 1.10118484223368e-06, -0.000268261950585856, 0, -963194.445050122, 0.000280502050641411, 0, 0.000253190500433906, -5.66118417794847e-05, -72916.6667383564, -0.000152378160716178, -161111.111914338, 0, 0.000473573467950054, -0.000313418997340658, -0.000175792777669389, -161111.111486579, 0.000191487726110204, -0.0002909641800676, 6.34355779658043e-05, -0.000156860346367305, 0, -0.000107040534272622, 0, -0.000121120429194212, 0, 381944.444339957, -3.45670388482514e-05, 0, 1.53428777355877e-05, 0.000228511185052778, 0, 2638888.88996572, -0.000510885403362639, 0, -0.000409348442926208, 3.00767005048675e-05, 173611.11124301, 0.000380337489278263, 0.00146901836288198, 0, 729166.66892279, -0.00117745929180048, 0, -0.00104364551122978, 0.000721267809181836, 0.000303416712987106, 729166.667685326, -0.000379573932218981, 0.000753302531748878, 4.72977506410878e-05, 0.00153832210208175, 0.000904030179544479, -0.000709681293655636, 0.000560341092633793, 0.000354383464379557, 0.00122241432294528, 0, 9.50199711547769e-05, 0, 0.00014626324426362, 0, -307539.682362149, 2.53104864349205e-05, 0, -5.07557449544191e-05, -6.72271993361413e-06, 0, -4027777.77879962, 0.000498118958338868, 0, 0.000400942569618534, 2.3855211742436e-05, -208333.333417367, -0.000424725765559758, -0.00124205834806155, 0, -1736111.11438706, 0.00150659745847371, 0, 0.00152742789742658, -0.000809563845337417, -0.00033875125624283, -1736111.11235215, 0.000721857374700585, -0.0011712334816125, 7.19725218273317e-05, -0.00329294661495358, 0, 0.0021017342102702, -0.00276273871401563, -0.00106369408280835, -0.00306773134296849, 0.00185563117423782, -0.00162776301694173, 0.00145993565242066, -0.0011946985401645, -0.000156991647139492, -0.00232830739531163, -0.000792630295124417, -0.00176859524983556, 0.000123438140909836, 0, -4.69771589344947e-05, 0, -8.8267455830029e-05, 0, 99206.3490760313, -8.34705580903819e-06, 0, 3.07383436853609e-05, -0.000108937225140691, 0, 3174603.17507289, -0.000247326968694172, 0, -0.000155789841117904, -4.42762049192757e-05, 99206.3492196992, 0.000340092453584343, 0.000431834755532536, 0, 2083333.33566293, -0.000975712221461522, 0, -0.00103478290053078, 0.000416377102189478, 0.000147896971126844, 2083333.33400338, -0.000525346002606137, 0.00096930604430445, 0.000292356914828147, 0.00203330273852224, 0, 0.00335520563456338, -0.00180753627056926, 0, -0.0017998464313691, 0.00222976333265045, 0.000765675461367532, 0.00277151406662714, -0.0018621901252571, 0.00153891065892013, -0.000811227344519523, 0.00085345336214296, 2.36202317605677e-05, 0.00208529972183237, 0.00404644764480123, 0.00154500742550139, -0.00240292424136249, 0.00185348355858105, 0.0011123137476948, 0.00174731182602778, -6.9456982949648e-05, 0.00276310261266557, 0.00112674526228463, 0.0018981535090211, 0, 9.89721970760332e-06, 0, 2.14358717773351e-05, 0, 3.68252596233004e-05, -1.38905469606628e-06, 0, -9.40317508308393e-06, 4.79608268725192e-05, 0, -992063.492143359, 5.50596446375021e-05, 0, 3.25053243089599e-05, 1.79386505712177e-05, -5.24832490011049e-06, -9.28395279524188e-05, -2.58570537004628e-05, 0, -992063.492693293, 0.000296141099373689, 0, 0.000224657978241659, -8.07282326952197e-05, -1.256116172158e-06, -992063.492201044, 0.000131440284654208, -0.000245994674260214, -9.68161714445004e-05, -0.000422682981009486, 0, -0.0012504906151642, 0.000586092917624747, 0, 0.000578327209672649, -0.000640663344458038, -0.000320859690184502, -0.000844732512422485, 0.000836131661590776, -0.000527471620552461, 0.000216934554054385, -0.000246262419116059, -8.53985387531787e-05, -0.000927313787972822, -0.0010965475967929, 0, 0.000774130665063892, -0.00163300614109366, -0.000623957109968665, -0.00172623892996192, 0.0012002637295826, -0.000729677754739746, 0.000941562957312735, -0.000852766508178589, -0.000573324850977576, -0.000898136158759284, 0.000553066829055876, -0.0014983693355066, -0.000938353352555324, -0.00122286018960828, -0.00105544904508729, -0.000313242800780817, -3.2428837680585e-05, 5.42176120430742e-05, -0.000843032442003237}, 
                                                     {0, 0, -2.2109900729821e-08, 0, 0, 5.73647623255591e-07, 0, 642.857143262882, -2.52721689838899e-07, 0, 0, -5.91423608219316e-06, 0, -1357.14285891717, 0, -17107.1428664751, 4.00377212959918e-06, 0, 5.22450859725437e-06, -1357.14285607344, 0, 3.24143376713028e-05, 0, 714.285710659816, 0, 178964.285797713, -2.06878796766903e-05, 0, -3.96687978518604e-05, 34678.5714723778, 0, -7.4832694565893e-05, 1428.5714326275, 34678.571420416, 714.28568038558, 0, -0.000105710614427029, 0, 4.91832177077172e-05, 0, -979861.111485198, 5.98740514494934e-05, 0, 0.00015134235949759, -17500.0000021069, 0, -341111.111523007, 0.000368434389670979, 0, 0.00053227952481368, -6.13339498478525e-05, -341111.111106923, 0.000245100155311973, -35000.0001119243, 0.000100800464751178, -17499.9996556296, 0, 0.000215464759889215, 0, -0.000155039814786322, 0, 3109027.77872461, -9.86879821099703e-05, 0, -0.000315013407455246, -0.000521902854208611, 0, 1707638.89059757, -0.000855983575387007, 0, -0.00180183471226501, 0.000217039013144344, 1707638.88891686, -0.000824032889419526, 161111.111742362, 0, -0.00224567386466789, 0.000780067975526953, 3.68809271522768e-05, 322222.222956615, 161111.109953173, -0.000973046693016729, -0.00220349352392186, -0.000587838710652263, 0, -0.000277276061314815, 0, 0.000245572192267596, 0, -5902777.77920141, 0.000106815782687288, 0, 0.000396654889980693, 0.00154618134311775, 0, -4756944.4480256, 0.00114045541042329, 0, 0.00328687849564413, -0.000354831911338726, -4756944.44444938, 0.00141826533790734, 0.000844005730443932, 0, -729166.669880778, 0.00456864212669334, 0, 0.00618208066028198, -0.00212562476691999, -0.000901383959010076, -1458333.3351196, -729166.664263617, 0.00304300426535429, 0.00644598826629816, -0.00349925375057379, 0.00127224414255976, 0.00656829828373414, 4.37690232075041e-06, 0.000533935329656684, 0.00501350389612398, 0, 0.000218969152843435, 0, -0.000215743423211696, 0, 6587301.58857254, -7.02055834362328e-05, 0, -0.000288056513764316, -0.00197285488706939, 0, 7430555.55954797, -0.000813619188144725, 0, -0.00340114250521469, 0.000263426098410327, 7430555.5554961, -0.00133024194734712, -0.0034220172306038, 0, 1736111.11719465, -0.00485071659718371, 0, -0.00861088886415669, 0.00280599224463392, 0.00160193260574638, 3472222.22452546, 1736111.10850195, -0.00465975067377009, -0.00899396542967085, 0.00132036742945653, 0, -0.0093148905837872, 0.00679307302217119, 0.000861581941046588, 0.0065957466443569, -0.00955505004528673, -0.00424641154887404, -0.0146817572901364, 0.00269927784315124, 0.00111335853790825, -0.0123719129347219, 0.00217383773523466, -0.00681639327570285, -0.00628695855519861, 0, -9.68925514319196e-05, 0, 0.000101043508058537, 0, -3968253.96887921, 2.94825107704612e-05, 0, 0.000105658738094869, 0.00120990788169302, 0, -6051587.3038541, 0.000273602259613622, 0, 0.00184745487265966, -7.48342501923078e-05, -6051587.30154216, 0.000585111754999113, 0.00336901203885584, 0, -2083333.33830028, 0.00258249384449544, 0, 0.0059509100296411, -0.00173232069039337, -0.00112352246819228, -4166666.66799944, -2083333.33181152, 0.0032107828497624, 0.00652242023365948, 0.00198090018097839, 0, -0.00358942910214154, 0.00682479081792961, 0, 0.00881766482501485, -0.00595581917553146, -0.00225023644572385, -0.00530058046769771, 0.007587349347241, 0.0052881859477393, 0.0147375194832578, -0.0038113417725619, -0.0014319502974272, 0.0122955779489761, -0.0087102439970152, 0.00125254481138791, 0.0139668837829988, -0.00463261373312255, 0.00160537490717712, -0.00462106891943488, 0.00368385862916515, 0.0127415482881392, 0.0118792698450037, 0.00608687333835712, 0, 1.83932875564193e-05, 0, -1.98035523016758e-05, 0, 992063.492194553, -9.8333151009927e-06, 0, -1.52066632078426e-05, -0.00029203301898406, 0, 1984126.98464699, -2.97835611758916e-05, 0, -0.0004146850279309, -2.37355349270365e-07, 1984126.98409311, -8.84084578206952e-05, -0.00110469728740955, 0, 992063.49353751, -0.000521135434939918, 0, -0.00171635016003778, 0.000401414970841998, 0.000266524090455671, 1984126.98441736, 992063.491650683, -0.000822772047842499, -0.00171387419449582, -0.00144742646481499, 0, 0.00185959782789968, -0.00195389427158142, 0, -0.00317729583313689, 0.00194752469533291, 0.000945408750608123, 0.00159016857131573, -0.00219698196742823, -0.00202943131547305, -0.00550678725698673, 0.00156623703969764, 0.00036061096041588, -0.00469638127473469, 0.000160267766627102, 0, -0.0032222622835566, 0.0038524778413346, 0.000560252969090856, 0.00365091092901509, -0.00530710608620432, -0.00147867901595444, -0.00699897947105316, 0.00309494721409274, 0.00110472729179375, 0.00241117789525159, -0.000857339803876565, -0.0066966633039082, -0.00666561176219141, 0.00353614460964187, -0.00353667596638982, -0.00383562587065459, -0.00224379843311209, -0.00111489015482051, -0.00603937054830641}, 
                                                     {0, 0, 4.88804905125714e-08, 0, 0, -1.23669254755953e-06, 0, -2000.0000014855, 9.80510625063581e-07, 0, 0, 1.24429782770735e-05, 0, 6722.22223203415, 0, 52388.8889231067, -1.68557382410238e-05, 0, -2.06390533502205e-05, 6722.22221902444, 0, -6.61205690614866e-05, 0, -7500.00000645522, 0, -535990.741038759, 8.68704374950466e-05, 0, 0.000147660168855861, -168490.740979368, 0, 0.00034526905886511, -15000.0000011218, -168490.740715807, -7499.99989228241, 0, 0.000206971340352274, 0, 2777.77763832944, 0, 2851111.11242361, -0.000228393053223236, 0, -0.000508425106857567, 179583.333844054, 0, 1611250.00198305, -0.00171112406248742, 0, -0.0024032490733229, 8333.33342539706, 1611250.00007056, 2777.77713304518, 359166.66699245, 8333.33283400849, 179583.331878214, 0, -0.000400318980282981, 0, 0.00057604965074561, 0, -8731250.00326345, 0.000335615910970536, 0, 0.000910775908089159, -63425.9255142334, 0, -7764351.85966167, 0.00414721998930944, 0, 0.00805583753756624, -0.000290994898285104, -7764351.85244314, 0.0016199408955015, -1596527.78259874, 0, 0.0103553766806578, -190277.780203967, 0.000867002637261429, -3193055.5585104, -1596527.77228987, 0.00316697592083261, -63425.9175155256, -190277.773005001, 0, 0.000483036798390244, 0, -0.00107075723675199, 0, 15902777.7825696, -0.000293366186085026, 0, -0.000846924916181175, -0.00321258492187664, 0, 20590277.7939342, -0.005838787448699, 0, -0.0146503027560318, 0.000141453460129426, 20590277.7791219, -0.00163675756213888, 520833.336196855, 0, 6875000.01800657, -0.021145797857926, 0, -0.0283874584814815, 0.0065745992557509, 0.000773123009114574, 13750000.0086238, 6874999.98873115, -0.00796618904840272, -0.0220951196437682, 1562500.01440168, -0.00617125286343499, 520833.306633441, -0.0098356240075729, 1562499.98895975, -0.0229847627415134, 0, -0.000353214922114418, 0, 0.00108113044602388, 0, -16944444.4485832, 0.000133155026460235, 0, 0.000267820472736332, 0.00502447246665868, 0, -30254629.6476922, 0.00457552721447834, 0, 0.0148950259161734, 0.000485454646616375, -30254629.6311542, 5.32646694179923e-06, 0.0049651068298366, 0, -15277777.8088614, 0.0231064745170628, 0, 0.0402302373819079, -0.00809049244328729, -0.00277745671156752, -30555555.5675215, -15277777.7647645, 0.00955775021707199, 0.0297978277052797, -1967592.60782321, 0, 0.0433506986128112, -0.028058080183355, 0.00131973075979006, -5902777.80900019, -1967592.55159887, 0.0149630434445733, 0.0542682916443918, 0.00605984235210684, -5902777.7653498, 0.0439556203410804, 0.0148977165026227, 0.0453385592405605, 0.0414227875902887, 0, 0.00014277726020315, 0, -0.000577797304516236, 0, 9722222.22417296, -3.25532319200801e-05, 0, 0.000106694082479185, -0.00343183617783202, 0, 22916666.6770073, -0.00184606622633432, 0, -0.00782748776381568, -0.000686720696769288, 22916666.6676164, 0.00128893115058616, -0.00744583930401657, 0, 16666666.691467, -0.0130579049235604, 0, -0.0283703338795568, 0.00433953136405018, 0.00203724080022808, 33333333.341065, 16666666.658903, -0.00449348719827121, -0.0198575475959866, 0.00118003302936938, 0, 3472222.24445619, -0.03180697823162, 0, -0.0412038190061168, 0.0238345244046746, 0.00485417521511627, 10416666.6947125, 3472222.18829061, -0.015088510163475, -0.0521858676705184, 0.00149300457143944, 10416666.6579291, -0.0379116529396497, 0.0414578969047591, -0.00680810861268864, -0.0534532752524649, -0.00369869472366482, -0.0326135093303151, -0.00379583491970772, -0.0358282473747086, -0.0591993662572726, -0.0473563701020061, -0.0589574380822301, 0, -2.43846477246468e-05, 0, 0.00012843226550766, 0, -2314814.81520318, 1.04505296100536e-05, 0, -7.2114074118467e-05, 0.000898158096414359, 0, -6944444.44683136, 0.000283828501992411, 0, 0.00167732590043319, 0.000261362458575506, -6944444.44464961, -0.00062932974615694, 0.00278467452378686, 0, -6944444.45185398, 0.00296397019168777, 0, 0.00798343692629642, -0.00077520692618224, -0.000353140035800583, -13888888.8908405, -6944444.44256633, 0.000460023731910438, 0.00489949164721833, 0.00259903644407221, 0, -2314814.82487444, 0.00913705636501513, 0, 0.0148613869418438, -0.00731807733837168, -0.00275546375653839, -6944444.45350869, -2314814.80386766, 0.00496139709259029, 0.0192245097384928, -0.00175742541545785, -6944444.44152497, 0.0127376012768286, -0.00403768796054965, 0, 0.014949610852653, -0.0178224659632312, -0.00111888954844629, -0.0190503919632879, 0.0210192459882645, 0.00499488084885392, 0.0255417133273122, -0.00368711051255974, 0.00640505150879692, -0.00164031151856376, 0.0147691906075519, 0.0258044797872193, 0.0203585870314376, -0.00537717770337409, 0.0237497146648068, 0.0171996986299317, 0.0291015494718582, 0.0244993384122055, 0.034589487426735}, 
                                                     {0, 0, -7.32921269734123e-08, 0, 0, 1.81672128066903e-06, 0, 4200.00000376754, -9.81133990061628e-07, 0, 0, -1.81039421934386e-05, 0, -20116.6666945917, 0, -107616.666755999, 2.32822541521251e-05, 0, 2.51011779116772e-05, -20116.6666748292, 0, 9.47688993634456e-05, 0, 35916.6666963904, 0, 1067513.88967366, -8.25178303645533e-05, 0, -0.000147004447252249, 490930.556264331, 0, -0.000651617764122924, 71833.3333564748, 490930.555741849, 35916.6666385009, 0, -0.000288230517694597, 0, -28333.3329906201, 0, -5460277.78128658, 3.1886659674956e-05, 0, 0.000252133026750854, -833263.890756136, 0, -4515208.33910437, 0.00312534612557308, 0, 0.00471746384319142, -84999.9999917392, -4515208.33527016, -28333.3336408432, -1666527.77936383, -84999.9989305452, -833263.887305185, 0, 0.000531221882052044, 0, 8333.33178086827, 0, 15968402.7866593, 0.00043374723203073, 0, 0.000491182642565696, 623611.111824301, 0, 20661805.5783536, -0.00691115322331037, 0, -0.0158744067055227, 33333.3328011693, 20661805.5634079, 8333.3370378225, 7058680.57065892, 0, -0.0204994424727562, 1870833.34032185, 49999.9955982387, 14117361.1247228, 7058680.55195752, 33333.3287759098, 623611.101727716, 1870833.32249349, 0, -0.000595002989204499, 0, 0.00304741318113923, 0, -27652777.7910404, -0.00109340699706948, 0, -0.00254747421331401, -173611.106137164, 0, -51402777.8259111, 0.00882618319855034, 0, 0.0284841329440024, 0.00260802229041818, -51402777.7941462, -0.0130888627744967, -4826388.9025467, 0, -28402777.8312152, 0.0392605995152394, 0, 0.0592009453444596, -694444.459824051, 0.00537402530564265, -56805555.5976192, -28402777.7757201, 0.00469551017799319, -173611.090496732, -14479166.7158697, 0.0195600763244985, -4826388.8631876, -1041666.6294357, -14479166.6495226, -694444.399092865, 0, 0.000389036930041631, 0, -0.00322539110867548, 0, 27986111.1227183, 0.00128020571014759, 0, 0.00396813057535805, -0.0095772324198692, 0, 70208333.3887007, -0.00634699712242201, 0, -0.027459406550965, -0.00499230614940582, 70208333.3522387, 0.0215590782059155, 1215277.77408924, 0, 57673611.2029608, -0.0405453759115101, 0, -0.0870797832316112, 0.0158316804527113, -0.00289076534501279, 115347222.284327, 57673611.1136227, 0.00623627535294708, -0.0223452117530896, 16666666.7154699, 0, -0.0863862708610182, 4861111.19553199, -0.0196579891934642, 50000000.1155658, 16666666.6362647, -0.0342839962923211, 1215277.72763546, -0.0465013464154603, 49999999.9943896, -0.0622964905115596, 7291666.61868179, -0.14024580653657, 4861111.04409799, 0, -0.000132442704709109, 0, 0.00179646876222338, 0, -15277777.7832951, -0.000716013257453394, 0, -0.00278260931883642, 0.00712664220872538, 0, -49305555.5882936, 0.00237002592714811, 0, 0.0130869957706059, 0.00425393379023898, -49305555.5671994, -0.0172318037739524, 0.0123267934586455, 0, -56250000.0746085, 0.0224104097541933, 0, 0.0628999253753214, -0.00634114966409822, 0.00145181939537805, -112500000.044433, -56250000.0049436, -0.0158770435729449, 0.00919475793969178, -3472222.23337739, 0, -25694444.5101458, 0.0590055376828085, 0, 0.087151675122314, -0.0659588329462959, 0.00257481577764695, -77083333.4456644, -25694444.4196255, 0.0238506168292342, 0.0519006015291696, 0.0246789717671241, -77083333.3387408, 0.0353298593997638, -13888889.0251932, 0.0290618888202464, -3472222.18559483, 0.0337483564322075, 0.137336667984239, -20833333.3324447, -13888888.8416377, 0.152014590303598, 0.0367009226705592, 0.175651531335893, 0, 1.70055178567838e-05, 0, -0.00041350577341607, 0, 3472222.22332423, 0.000152037636533015, 0, 0.00075072291730633, -0.00198316759102775, 0, 13888888.8966984, -0.00029331069777979, 0, -0.00236216770398058, -0.0013543486422087, 13888888.8917845, 0.0052865491319532, -0.00515860493807582, 0, 20833333.3562697, -0.00516884569612522, 0, -0.0175848157830388, 0.000351709041628475, -0.000901665052843121, 41666666.6792691, 20833333.3357552, 0.00819235397690128, 0.000135869800008422, -0.00360789059567211, 0, 13888888.9183503, -0.0158944821709965, 0, -0.032895510041985, 0.0187344333596663, 0.00247269384881046, 41666666.7057555, 13888888.8799764, -0.00416869118328481, -0.0208177679964411, -0.00524199349694596, 41666666.6706111, -0.00615525062075897, 3472222.23534333, 0, -0.0295329893995091, 0.0544217002522965, -0.00415188990911261, 13888888.9561092, 3472222.20966643, -0.013629462562101, -0.0203276458138672, -0.00429250018756896, -0.0419465823295423, 20833333.3479417, 13888888.8676343, -0.0559206904178277, -0.00230767890473147, 0.0141894337011759, -0.081888878766901, 0.00225483220199777, -0.103278712885642, -0.068724508673688, -0.0820553386329826}, 
                                                     {0, 0, 9.05753733595047e-08, 0, 0, -2.27267715556847e-06, 0, -6300.00000646123, -2.28990089578759e-06, 0, 0, 2.36026280232339e-05, 0, 40675.0000462484, 0, 156175.000159388, 1.74362300048004e-05, 0, 2.85278658762542e-05, 40675.000056942, 0, -0.000129655855542372, 0, -104166.666691542, 0, -1479875.00143497, -0.000266229348376246, 0, -0.000303561098310106, -954041.66793157, 0, 0.000347017984249687, -208333.333504911, -954041.667686056, -104166.667196801, 0, 0.000414323295928488, 0, 132291.66595429, 0, 7159513.89548354, 0.00152545788599039, 0, 0.0019268234896325, 2306597.22548184, 0, 8283819.45491792, -0.000915627923114712, 0, -0.00359270619360928, 396874.999955721, 8283819.45256181, 132291.671083852, 4613194.44999273, 396874.999162951, 2306597.22605321, 0, -0.000803733815624861, 0, -83333.3303230958, 0, -19704513.9060871, -0.00430363117572026, 0, -0.00689849331483268, -2758680.55695836, 0, -35221875.0423154, -0.0020662259910895, 0, 0.0135142920955702, -333333.331269998, -35221875.0299918, -83333.3540646988, -18192708.3596662, 0, 0.0144501236862559, -8276041.68391138, -499999.990542992, -36385416.7087048, -18192708.3626884, -333333.331795726, -2758680.56983622, -8276041.66273432, 0, 0.000952764617612101, 0, 20833.3274807593, 0, 32138888.9153732, 0.00677372307957949, 0, 0.0140181582106447, 1631944.43665908, 0, 80625000.0919508, 0.00877793988163051, 0, -0.0241076401309886, 104166.659035494, 80625000.0596305, 20833.3865678903, 19548611.1343288, 0, 66423611.2052164, -0.0138383068154143, 0, -0.0563778038298817, 6527777.80736174, 208333.315339176, 132847222.350079, 66423611.2017014, 104166.682047877, 1631944.48373596, 58645833.4538428, 208333.292028947, 19548611.193058, 9791666.61017895, 58645833.3827971, 6527777.74943387, 0, -0.000667761741210946, 0, 0.00622133519606143, 0, -30763888.9127869, -0.00629525058453502, 0, -0.0160429597596414, -381944.429073092, 0, -101388888.998038, -0.0113980029723527, 0, 0.019699948821219, 0.0128162105880189, -101388888.955618, -0.0750367769947742, -10243055.550479, 0, -119965277.943482, -0.00268646632032866, 0, 0.097400828372752, -1909722.24648197, 0.016407132613803, -239930555.745238, -119965277.909441, -0.0560201374197736, -381944.507730346, -59201388.9698667, 0, 0.0604552995501387, -40972222.3999248, 0.0628922271830891, -177604166.955278, -59201389.0850581, 0.0633672263409251, -10243055.6857642, -3819444.34736967, -177604166.828329, -1909722.21555424, -61458333.3280657, -3819444.20725248, -40972222.2951906, 0, 0.000250475587885495, 0, -0.00349210958392197, 0, 15972222.2339285, 0.00317655915744026, 0, 0.00961588643984154, -0.0116681647021318, 0, 65972222.2888843, 0.00655385743943177, 0, -0.00547323306285892, -0.010305679850512, 65972222.2620939, 0.05431713182867, 2083333.31457048, 0, 104166666.805367, 0.00877449352204686, 0, -0.0760245168936643, 0.00620212509043186, -0.00925542864423346, 208333333.47118, 104166666.760875, 0.0690463259304212, 0.0580935416373257, 24305555.5721496, 0, 76388888.9999052, -0.0182863907791599, 0, -0.0854542171335387, 10416666.7907819, -0.0305385782865026, 229166666.955688, 76388889.0700976, -0.0310146337076538, 2083333.41811977, -0.0665634183737615, 229166666.841203, 0.0387704221544232, 97222222.5209136, -0.0822253735658274, 24305555.7762529, 20833333.2986574, -0.284667827623983, 145833333.472298, 97222222.4008634, -0.247275554992094, 10416666.8319951, 20833333.1194635, 0, -3.78268476684189e-05, 0, 0.00080974402698683, 0, -3472222.22463011, -0.000672905658441745, 0, -0.00235036216917787, 0.003325544642777, 0, -17361111.1275313, -0.00151723566973592, 0, -0.000446488297588688, 0.00320097498092629, -17361111.1209014, -0.0156491813959627, 0.0078437637885726, 0, -34722222.2662458, -0.00288344285843089, 0, 0.0215920982819809, 0.00117275483548148, 0.00310712229067817, -69444444.4840621, -34722222.2493773, -0.0288591209572629, -0.0225619359461068, -3472222.21602592, 0, -34722222.2735458, -0.00144531125066553, 0, 0.0386993806224001, -0.0321764725960165, 0.0032588744984253, -104166666.770431, -34722222.2813203, -0.00191349802550762, -0.0185743213681005, 0.0177763639986795, -104166666.733139, -0.0286833469906485, -17361111.1314212, 0, 0.0196056026406001, -17361111.2198894, 0.0238848233319718, -69444444.5975979, -17361111.2193118, 0.033181359258774, -3472222.29875903, 0.0124331559793913, 0.0995965635928457, -104166666.768008, -69444444.5273484, 0.0837840843653708, -0.0961064113659948, -34722222.2866574, 0.167403703503524, -17361111.2694649, 0.177069365929674, -34722222.1844048, 0.0846372052314631}, 
                                                     {0, 0, -1.26490301333953e-07, 0, 0, 3.32926287172725e-06, 0, 7000.00000754058, 7.11098283621093e-06, 0, 0, -3.66376772658158e-05, 0, -59194.4444908597, 0, -164194.444638711, -8.76027162678239e-05, 0, -0.000115612070375555, -59194.444560663, 0, 0.000214713040321452, 0, 206129.629592053, 0, 1444046.29809777, 0.000776715366542599, 0, 0.000965402031988861, 1300175.92734805, 0, 0.000613657369140297, 412259.259724015, 1300175.92795576, 206129.630862358, 0, -0.000738962899778215, 0, -378472.221150624, 0, -6430138.89741014, -0.0035074744987471, 0, -0.00468029936192011, -4227361.11427138, 0, -10279027.7899887, -0.0049771947306825, 0, -0.00176522081864513, -1135416.66727905, -10279027.7927037, -378472.23056408, -8454722.23320626, -1135416.66782836, -4227361.12424707, 0, 0.00156364492324054, 0, 386574.07005713, 0, 16370138.9117871, 0.00870665462470264, 0, 0.0139239279299578, 7223379.62941982, 0, 39190509.310089, 0.0211258302265151, 0, 0.00157475877411732, 1546296.29455334, 39190509.3114716, 386574.107963022, 29657175.9535845, 0, 0.0126118755701409, 21670138.9201401, 2319444.43751532, 59314351.9262104, 29657176.0098922, 1546296.30050078, 7223379.67976455, 21670138.9180577, 0, -0.0020566880391193, 0, -208333.325818992, 0, -24930555.5919253, -0.0126978513863897, 0, -0.0252537165380553, -6840277.7661802, 0, -81006944.5582274, -0.0417792810606923, 0, -0.0011709128686945, -1041666.65741266, -81006944.5435794, -208333.414565089, -44201388.90792, 0, -93645833.4364438, -0.0609138544789493, 0, 0.000777140630733293, -27361111.1581176, -2083333.31747878, -187291666.881949, -93645833.5697662, -1041666.69525565, -6840277.88632983, -132604166.859426, -2083333.28546587, -44201389.1495594, -41041666.6711926, -132604166.884466, -27361111.1647295, 0, 0.00163758106502931, 0, 46296.2884429778, 0, 22500000.0338626, 0.0111422194750994, 0, 0.026907041746754, 3402777.75786736, 0, 93217592.7317122, 0.0418789129391606, 0, 0.00696168938284772, 277777.761189243, 93217592.6989125, 46296.4073132738, 35648148.134354, 0, 147916666.855046, 0.10741032176797, 0, -0.0395941577457031, 17013888.9215377, 694444.430863316, 295833333.642726, 147916666.990955, 277777.860058845, 3402777.91623447, 109490740.817494, 0, 0.0591676322072725, 142592592.847163, 925925.842419919, 328472222.666761, 109490741.317979, 694444.35596255, 35648148.5406875, 34027777.7421031, 328472222.7265, 17013888.977744, 213888889.133546, 34027777.6258925, 142592593.029835, 0, -0.000722109406010923, 0, 0.0043678398898807, 0, -11111111.1282267, -0.00545185897226765, 0, -0.0153832074340757, -694444.429878355, 0, -56250000.0874878, -0.0206818814066547, 0, -0.0108299514168263, 0.0136715188273261, -56250000.0610678, -0.0790529133845742, -14583333.3082479, 0, -114583333.496466, -0.0787765036925732, 0, 0.0447886879103404, -4166666.67176335, 0.00667439072486717, -229166666.885999, -114583333.551217, -0.0980570211043322, -694444.550310522, -62500000.0059966, 0, -118055555.668841, -0.101207450722006, 0, 0.00314691292271062, -72916666.8253124, 0.0449711885101146, -354166667.105313, -118055556.089392, 0.0507529737660735, -14583333.5907621, -10416666.6324442, -354166667.149352, -4166666.75831554, -250000000.419866, 0.12849888181816, -62500000.6203698, -145833333.471664, -13888888.6482775, -375000000.490305, -250000000.699588, -10416666.4492413, -72916667.1746197, -145833333.439414, 0, 0.000135249880987962, 0, -0.00100643719135486, 0, 2314814.81844634, 0.0011518180643922, 0, 0.00363438672693773, -0.00410883100556888, 0, 13888888.9110622, 0.00406249984624517, 0, 0.00479207612570331, -0.00431527739933537, 13888888.903462, 0.022570770047369, 2314814.80524282, 0, 34722222.2756729, 0.0205454321110188, 0, -0.0141654004082915, -0.00288593581802171, -0.00230055894799609, 69444444.5060062, 34722222.2802928, 0.0409189964323782, 0.0368032201517908, 13888888.8786074, 0, 46296296.3514671, 0.0442363156923413, 0, -0.0181543627624372, 13888888.9251659, -0.00585550664309729, 138888889.045109, 46296296.4727674, -0.000287761477784876, 2314814.87569049, -0.0105510904207782, 138888889.05817, 0.0427648503902216, 34722222.2386312, 0, 0.0237255730413131, 69444444.5850553, -0.0429158362111955, 138888889.104296, 34722222.5260724, -0.056032867514029, 13888889.1047539, 34722222.2528122, -0.088978217008251, 208333333.598521, 138888889.209277, -0.0930027056040878, 13888889.0958101, 138888889.08117, -0.174414984327228, 69444444.8701281, 46296296.223864, 138888889.084716, 34722222.2757394}, 
                                                     {0, 0, 1.67882747920495e-07, 0, 0, -4.53697852752468e-06, 0, -6000.00000590002, -7.72815006329996e-06, 0, 0, 5.04953599985633e-05, 0, 65738.0952629275, 0, 125738.095397857, 0.000103808190177277, 0, 0.000134929184979596, 65738.0953543013, 0, -0.000299629490081899, 0, -303527.777665664, 0, -960908.731693768, -0.000828889429684757, 0, -0.00107745838550399, -1264436.50889493, 0, -0.00115671316847719, -607055.556210236, -1264436.50994184, -303527.77901928, 0, 0.00105089986191551, 0, 766111.110025709, 0, 3801388.89637591, 0.0034642437089647, 0, 0.00479286890956531, 5333611.11244728, 0, 8368888.89770804, 0.00778245312065281, 0, 0.00582097269793763, 2298333.33504778, 8368888.90277846, 766111.117959106, 10667222.2344873, 2298333.33741035, 5333611.12674429, 0, -0.00228221985571744, 0, -1142361.10757592, 0, -8803472.24294994, -0.00820243903933519, 0, -0.0131327105118802, -12230555.5523772, 0, -27552777.8161716, -0.0273464788831461, 0, -0.0145591274296366, -4569444.44548912, -27552777.8236478, -1142361.13507478, -29837500.0157744, 0, -0.0327540836560995, -36691666.7033274, -6854166.67510098, -59675000.0726837, -29837500.0968572, -4569444.454336, -12230555.6098319, -36691666.7304627, 0, 0.00310052418728517, 0, 1006944.43822737, 0, 12430555.5894803, 0.0116874165912551, 0, 0.0224124932556193, 16458333.3201332, 0, 50729166.7557955, 0.0490074141147597, 0, 0.0223818345713841, 5034722.21706527, 50729166.7492241, 1006944.49915091, 55763888.8908081, 0, 78611111.1767509, 0.103190775447447, 0, 0.056912507660565, 65833333.388096, 10069444.460683, 157222222.41516, 78611111.3728156, 5034722.24182293, 16458333.4241811, 167291666.853044, 10069444.4373386, 55763889.1920349, 98750000.1349836, 167291666.996215, 65833333.4762231, 0, -0.00256406905162577, 0, -486111.104840822, 0, -10555555.5881025, -0.0101784815766947, 0, -0.0229119424371916, -12986111.0923419, 0, -53263889.0014848, -0.0459422865960638, 0, -0.0245110326162477, -2916666.65384603, -53263888.9732035, -486111.186261454, -57638888.865083, 0, -107986111.239225, -0.153508210315068, 0, -0.0411551425065842, -64930555.5915663, -7291666.6872064, -215972222.481808, -107986111.455763, -2916666.71094081, -12986111.1928686, -110416666.699432, 0, -0.154780326370112, -230555555.791173, -9722222.21021743, -331250000.388625, -110416667.338585, -7291666.61091355, -57638889.3078976, -129861111.275942, -331250000.657098, -64930555.6909306, -345833333.868769, -129861111.278753, -230555556.258574, 0, 0.00118011396702003, 0, 99206.3458011445, 0, 4960317.47725973, 0.0050395702657989, 0, 0.0127533459031594, 5555555.54282096, 0, 29861111.1841368, 0.0213874212640593, 0, 0.0172383565396446, 694444.432851918, 29861111.1575426, 99206.4037973177, 31249999.9740948, 0, 75000000.1163751, 0.104599798448163, 0, 0.00960595979855279, 33333333.3371102, 2083333.34536363, 150000000.174442, 75000000.2207421, 694444.501648577, 5555555.60307281, 76388888.8747503, 0, 100694444.506413, 0.173264268992246, 0, 0.0898423105868888, 156250000.13578, 3472222.2288115, 302083333.692721, 100694445.065689, 2083333.27968962, 31250000.2395723, 83333333.4373392, 302083333.906545, 33333333.3767974, 305555555.908141, 3472222.14549789, 76388889.5859943, 312500000.423762, 111111111.236652, 458333334.062261, 305555556.558896, 83333333.3548452, 156250000.658875, 312500000.654053, 0, -0.00023174134784548, 0, 0.000770756301021268, 0, -992063.495761944, -0.00109544057830872, 0, -0.00296182510121335, -992063.488611878, 0, -6944444.46350638, -0.00390586650486583, 0, -0.00536660772457608, 0.0037956709124288, -6944444.45521439, -0.015909972527276, -6944444.43576186, 0, -20833333.3729817, -0.0263345130994617, 0, 0.000211949157591739, -6944444.44050564, -0.00204273952834308, -41666666.7134704, -20833333.3887649, -0.0261017345827632, -992063.507903446, -20833333.3207854, 0, -34722222.2559456, -0.0652193070007902, 0, -0.0148176308180236, -41666666.6935218, -0.00776053621142169, -104166666.788716, -34722222.4266263, 0.00972307599063441, -6944444.48825037, -20833333.3584157, -104166666.852679, -6944444.44819486, -34722222.2252607, 0, -0.0590952291801512, -104166666.778959, 0.0228683373744091, -138888889.06024, -34722222.5707477, 0.0468497149060921, -20833333.5648905, -104166666.792276, -34722222.2700301, -208333333.660339, -138888889.330793, -20833333.2841429, -41666666.8703394, -208333333.632949, -34722222.2097755, -104166667.200444, -138888889.121079, -208333333.802102, -104166666.946233}, 
                                                     {0, 0, -4.12290837702853e-07, 0, 0, 1.1289675496354e-05, 0, 5250.00000810945, 7.09273948058849e-06, 0, 0, -0.000126119065654272, 0, -96895.8333418838, 0, -96895.8335717448, -8.62836055245695e-05, 0, -0.000158928072566576, -44395.833400606, 0, 0.000751839951677688, 0, 598555.555202725, 0, 598555.557960719, 0.000498045327533827, 0, 0.00124957479546831, 1722111.11212805, 0, 0.00178599086688328, 753152.778170083, 753152.778956646, 154597.223010701, 0, -0.00265920346980142, 0, -1829826.38674608, 0, -1829826.4010765, -0.00131727976812543, 0, -0.00510805563902544, -9929062.49939542, 0, -9929062.51091866, -0.0105662647603825, 0, -0.0110918181444627, -3943506.94515305, -3943506.95159694, -283854.169656129, -12326597.2297925, -2397534.72575485, -2397534.73501837, 0, 0.00584545466533677, 0, 3128472.21617359, 0, 3128472.25710437, 0.00115484391578893, 0, 0.0127704153645715, 27973611.0999681, 0, 27973611.162908, 0.0320255058083829, 0, 0.0313877335656553, 9675347.21986917, 9675347.24026306, 289930.562698695, 49690277.7873738, 0, 0.0574405493060586, 59945486.1330744, 10255208.3447623, 59945486.1471756, 10255208.416195, 3998263.89054213, 3998263.92639043, 35970138.9556288, 0, -0.00806182422935556, 0, -3055555.54568328, 0, -3055555.61439153, 0.00128919706395092, 0, -0.0201131247521415, -43663194.4168546, 0, -43663194.5722589, -0.0488129205081808, 0, -0.0518648043287188, -12378472.2095669, -12378472.2422994, -156250.017345382, -115711805.529295, 0, -115711805.617188, -0.160483962386511, 0, -0.1203118825032, -134670138.91161, -18958333.3546004, -134670138.956169, -18958333.5478858, -3680555.53469433, -3680555.57624114, -244583333.423816, -13159722.2173667, -13159722.497899, -142031250.189698, -142031250.305457, -54704861.2550678, 0, 0.00678203253945399, 0, 1597222.2126968, 0, 1597222.28023396, -0.00339500654871211, 0, 0.0191117264638135, 38576388.8558238, 0, 38576389.0583438, 0.0346131088394993, 0, 0.0545205083273931, 8020833.31117591, 8020833.34024633, 34722.2524665004, 139930555.49238, 0, 139930555.699508, 0.224084293404373, 0, 0.121920572313715, 156076388.879058, 16145833.3524105, 156076388.939076, 16145833.5858975, 1770833.30145916, 1770833.29140347, 205902777.764475, 0, 0.268364764891308, 428125000.09538, 16319444.4407536, 428125000.139231, 16319445.109914, 8333333.2488162, 8333333.63841075, 238541666.936507, 238541667.174765, 45659722.2495071, 444791667.248353, 164930555.894526, 164930556.302904, 0, -0.00317961391270359, 0, -347222.217191899, 0, -347222.253175443, 0.00245261545566047, 0, -0.0099307261630539, -18055555.5350714, 0, -18055555.670002, -0.00794263855785254, 0, -0.0336559842943038, -2083333.31578644, -2083333.33005123, -0.0261617798749103, -85416666.6159868, 0, -85416666.8119077, -0.146009478711247, 0, -0.0634906049566573, -90624999.9680312, -5208333.33868152, -90625000.0075949, -5208333.46483028, -347222.215530063, -347222.175034421, -168402777.714994, 0, -168402777.809635, -0.27526019851302, 0, -0.18056384002497, -343750000.022611, -6944444.44740124, -343750000.084697, -6944445.07759338, -2083333.23067894, -2083333.40930582, -182291666.844701, -182291667.009303, -19791666.5222486, -510416666.788729, -5208333.24335227, -5208333.98716875, -520833333.846052, -184027778.146224, -520833333.927347, -184027778.806882, -93750000.0947343, -93750000.5826948, -354166667.633496, 0, 0.000636549884736514, 0, -0.0011176627708416, 0, 0.00690827986988617, -0.000597132937918229, 0, 0.0021554850542339, 3472222.21696173, 0, 3472222.25315464, -0.0010060450339543, 0, 0.0090286430782834, -0.00531583115473177, -0.00208708773938057, 0.0084431173048476, 20833333.3185493, 0, 20833333.3865762, 0.0347951552035133, 0, 0.0149346190750754, 20833333.3181624, -0.00111133041037767, 20833333.3290178, 0.0225404613729781, 0.00577530939213411, -0.0108271756934217, 52083333.3050564, 0, 52083333.3623153, 0.100248126294643, 0, 0.042426598366615, 104166666.654916, 0.0020948934929996, 104166666.680797, 0.206120804732744, -0.0322155441986944, -0.0253185794947979, 52083333.3751832, 52083333.4092623, 3472222.14105527, 69444444.4271939, 0, 0.099355842128032, 208333333.36172, -0.0313601535855833, 208333333.382057, 0.363885866877709, -0.0603672635665285, 0.176354690809308, 208333333.492622, 69444444.5884345, 208333333.519573, 69444444.8664636, 20833333.2526962, 20833333.4259618, 312500000.257509, 52083333.414038, 52083333.8654159, 208333333.780292, 208333333.916347, 104166667.072152}, 
                                                     {0, 0, -6.54577098641057e-08, 0, 0, 1.77920921431402e-06, 0, 375.00000030639, 1.08734065104596e-07, 0, 0, -1.92550759063994e-05, 0, -9354.16665865623, 0, -4541.66667914113, 1.25624354821563e-07, 0, -3.76986658897579e-06, -791.666663485168, 0, 0.000110683385663635, 0, 88805.5554921409, 0, 8333.33348774995, -8.9922180245281e-06, 0, 3.77191679948398e-05, 112451.388778032, 0, 2.4637172372005e-05, 18909.7221918103, 8749.99993068415, 416.666660738729, 0, -0.00037824683018762, 0, -423576.388696092, 0, -4166.6675575934, 5.96081658059214e-05, 0, -0.000186214063538701, -1055520.83239863, 0, -198611.110622427, -3.07347161408003e-05, 0, -0.000246989785923135, -167465.277615416, -8333.33276513126, -3.09864807170175e-05, -208124.999542518, -9513.88890694051, -4166.6664307478, 0, 0.000806824609583472, 0, 1107638.88858576, 0, 0.00279630986793845, -0.000196382923651565, 0, 0.000521520109315506, 4952777.77518106, 0, 95138.8882471958, 2.91154104398344e-06, 0, 0.00104065693622003, 717013.888225855, -0.00224216640223722, 0.000212817494095774, 1752777.77291237, 0, 0.00051969673617714, 1830902.77650654, 78125.0003507577, 190277.774385892, -0.00209174905500109, 0.000374776253322527, -0.000124295214821507, 95138.8879570983, 0, -0.00108482600212686, 0, -1597222.22197078, 0, -0.00507273501961412, 0.000350518601311983, 0, -0.000811903549546206, -12656249.9963692, 0, -0.000963626478375537, 0.000303820533557554, 0, -0.00263799595266176, -1579861.10945271, 0.00481975902321248, -0.000529038303118639, -7465277.76609549, 0, -781249.988170423, -3.04685393756624e-05, 0, -0.00186948174352692, -7760416.66430633, -295138.889122893, 0.0112235555057368, 0.00756946924211952, -0.00094836591991927, -1.03633187240931e-05, -1562499.99304928, -0.00235195838503111, 0.00202462888921173, -781250.001246695, 0.0100906807621921, -0.00149831255983245, 0, 0.000893551919730722, 0, 1180555.55546881, 0, 0.00531706294237939, -0.000368661651344365, 0, 0.000807572081328463, 17673611.1083734, 0, 0.00361634754475156, -0.00039252682068334, 0, 0.00320973696223564, 1701388.88662119, -0.00556239083270232, 0.00066371405852049, 16319444.4315793, 0, -0.0143652547305693, 4.82578715158501e-05, 0, 0.0038118919868557, 16840277.7739474, 520833.332288187, -0.0179952462365201, -0.0125541090171342, 0.000935526895328471, 0.000478913689581835, 2951388.86558967, 0, 0.00126392829461696, 5902777.77112012, 0.00460625720162488, -0.0181265203566463, -0.00717045004303418, 0.00365343141205786, -0.000900556714150193, 2951388.8950252, -0.0318046444231696, 0.0036851809115414, -0.00987580236211676, 0.0101032575472813, -0.00487782840551856, 0, -0.000411696477734366, 0, -347222.222232234, 0, -0.00299050926766839, 0.000207039917583177, 0, -0.000414907652166346, -12499999.9989622, 0, -0.00366041261335655, 0.000276466892848929, 0, -0.00211301513052723, -694444.442865091, 0.00336087021171354, -0.000537748876088395, -17361111.1042878, 0, 0.00823905953228757, -0.000107798128635809, 0, -0.00375505160882934, -17708333.3297168, -347222.220391308, 0.013634927519697, 0.00984460230054248, -0.000580136925106731, -0.000349627645500569, -5208333.31549571, 0, 0.0211460081529704, -0.00025210409658057, 0, -0.00153433785373565, -10416666.6620138, -0.00272528557232574, 0.0208807625923768, 0.00924498409005821, -0.00195960235153752, 0.000393071692022279, -5208333.33754156, 0.0383479676450813, -0.0016128971288843, 0.0117247877122147, -0.00470760936055409, 0.00264640047198686, -0.00322143361259109, -0.0163078780314453, 0.03171776404428, 0.0218816497426484, -0.0103090644063406, -0.00320660301935392, -0.00588783528684268, 0, 8.12504220013353e-05, 0, 1.08373569762285e-05, 0, 0.000698607674258529, -4.95588102781237e-05, 0, 8.93765246648011e-05, 3472222.22207438, 0, 0.00126407172207897, -0.000105192375091962, 0, 0.000560079421574835, -0.000434042285508499, -0.000848298363929322, 0.000151593634887942, 6944444.44304756, 0, -0.00167836876260588, -2.64323490819186e-05, 0, 0.00121104274200107, 6944444.4430944, -0.000852730347575726, -0.00395640191265677, -0.00302864516426582, 0.000118092123272771, 0.000263004624149903, 3472222.21731721, 0, -0.00713053722958314, -6.79315273112446e-05, 0, 0.000863019760593264, 6944444.44266395, 0.000457107242519236, -0.0081779483274206, -0.00364093366747455, 0.000429694561697479, -9.08246131233561e-05, 3472222.22258156, -0.0153749682615044, 0.000325518018864932, -0.00841657651043609, 0, 0.000152509548751667, -0.00297608591295453, 0.00202322507223883, -0.00708944411011141, -0.00145218537661991, 0.00141437161448381, -0.000468565971332658, 0.00416702831984253, 0.00700661615696135, -0.0223725437249725, -0.0154978328100906, 0.00284419321047019, 0.00157062632357678, -0.00711650864923188, 0.00723043800141427, -0.00219587056467499, 0.0109067507939675, -0.00842055551011744, 0.00665356634901458}, 
                                                     {0, 0, 1.1874601528866e-08, 0, 0, -3.52136900735794e-07, 0, 375.000000643457, -1.80220528304448e-08, 0, 0, 4.04641869482804e-06, 0, -4541.66667447485, 0, -9354.16668155184, 2.57060221041114e-06, 0, -1.11728026014072e-07, -791.666670068907, 0, -2.50063730508895e-05, 0, 8333.33335582236, 0, 88805.5556761017, -2.04427205188936e-05, 0, 1.4863130442019e-05, 112451.389078919, 0, -6.03758925347688e-05, 8750.00002099379, 18909.7222545918, 416.666715391604, 0, 9.34995713949174e-05, 0, -4166.66663871878, 0, -423576.389386689, 6.20428441219984e-05, 0, -0.000128801562638745, -198611.111870972, 0, -1055520.83473461, 0.000413297292866931, 0, 0.00030933591883215, -8333.33338001991, -167465.277954513, -0.00044444563533071, -208125.000299502, -4166.66684989931, -9513.88893951021, 0, -0.00022054933951647, 0, -0.000292213798658317, 0, 1107638.89003723, -0.000112969946047475, 0, 0.000466861530881791, 95138.8902450912, 0, 4952777.78294521, -0.00138668012538452, 0, -0.00074572945268813, -3.73690763853842e-05, 717013.889451865, 0.00202520200710193, 1752777.78267672, 0, -0.00205629110277907, 190277.779204846, 0.000180055152969004, 1830902.77952746, 78124.9992081563, 0.00175988400483517, 0.000839587431012852, 95138.888721791, 0, 0.000328784493133095, 0, 0.000702458015631086, 0, -1597222.22373548, 0.000109155859845483, 0, -0.000936675321158807, -0.00100323662292528, 0, -12656250.0104249, 0.00264914232749713, 0, 0.000859355425468592, 0.000496877945092204, -1579861.11218999, -0.0047295187085832, -781250.00764782, 0, -7465277.79280668, 0.0052959431315301, 0, 0.00464953148498631, -0.0031314273415087, 0.000257998000725593, -7760416.67125729, -295138.886266763, -0.00631377160468805, -0.00487722313508745, -1562500.00804597, -0.00280145521032718, 0.00163565719472258, 0.00179244142338703, -781249.997432259, -0.00393836903382232, 0, -0.000299347607523816, 0, -0.000830761465369489, 0, 1180555.55666141, -4.19220222464861e-05, 0, 0.00106815175185707, -0.000123460638980345, 0, 17673611.1226776, -0.00278809327926656, 0, -0.000189526632460907, -0.000933551214500826, 1701388.88997741, 0.00602474141024658, 0.00564013218629528, 0, 16319444.4682368, -0.00736088155249892, 0, -0.00611761135454447, 0.00310300092887043, -0.000862653500495209, 16840277.7831924, 520833.328738675, 0.0111070462205944, 0.00869140009026897, 2951388.9072336, 0, -0.00935994218165596, 0.0150833866169907, 0.00167571850229999, 5902777.79476139, -0.00766855226072021, 0.00815093611226877, 0.001458614378874, -0.00307415022442373, 2951388.8807745, 0.0164664370258078, -0.00581338387516969, -0.000981229195389631, -0.00312664773918886, 0, 0.000151518182657989, 0, 0.000498080791530916, 0, -347222.222621766, -5.31560231664337e-06, 0, -0.000616463921650687, 0.000608133724181387, 0, -12500000.0066006, 0.00156519880566748, 0, -0.000431419266881642, 0.000757549469921232, -694444.445013945, -0.00408572371262316, -0.00162350818674432, 0, -17361111.1295418, 0.00535290124627081, 0, 0.00398996894580241, -0.0012060921766907, 0.00101159773568555, -17708333.3363201, -347222.218818223, -0.00932541679562136, -0.00854059690589087, -0.00964607197426428, 0, -5208333.35362015, 0.00820844903121567, 0, 0.00796477603224646, -0.0120825894252529, -0.000567404951461556, -10416666.6818691, 0.00960337064502939, -0.0101903421650637, -0.0047453308680491, 0.00240187458250144, -5208333.32416758, -0.0204951780908607, -0.0231011142687423, -0.00512774362950624, 0.00495509576871377, 0.00191752058745885, 0.0069557812909616, 0.0045971738895169, 0.0117481323438938, -0.0105455516246657, -0.0130884699410274, 0.0115014578849755, 0, -3.26059471685804e-05, 0, -0.000120850751347993, 0, 5.00290171302681e-05, 1.16202164887927e-05, 0, 0.000149431665844073, -0.000252829145902336, 0, 3472222.22373298, -0.000348329400225962, 0, 0.000216980583474113, -0.000231723078907329, 0.000151800815098491, 0.00109944849696123, 3.64132174481613e-06, 0, 6944444.44994562, -0.00160237093504346, 0, -0.000812876177453065, 7.477377544652e-05, -0.000449225260065628, 6944444.44505998, -0.000859208664380744, 0.00296573898439046, 0.00288045513620943, 0.00167304895008134, 0, 3472222.23044377, -0.00274245834634061, 0, -0.00260923998163521, 0.00337481470833108, 0.000227855706604135, 6944444.44911068, -0.00435042940111127, 0.00418720796084028, 0.00246041796010009, -0.000865506431636299, 3472222.21887278, 0.00879274979334302, 0.00567468457096525, 0, -0.00338635007114374, 0.00965408588336091, 0.00148718400906888, 0.0107977891306485, -0.00403004716356969, 0.00340818921620512, -0.0008600412102248, 0.000239416305713404, -0.00324607138961661, -0.00173582889348143, -0.00930723443213912, 0.00881434638309607, 0.0114216600623534, 0.00166539748162797, 0.000619858376665345, 0.00185003741356639, -0.00925359202966923, -0.0105588951272339, 0.00183760269169663}, 
                                                     {0, 0, 5.94646612537744e-07, 0, 0, -1.62275507651796e-05, 0, -4200.00000681083, -2.50255571018013e-06, 0, 0, 0.000179077667239879, 0, 90116.6666431482, 0, 69116.6668824102, 2.23957130125984e-05, 0, 8.68582600732854e-05, 27116.6666551058, 0, -0.00105481011095044, 0, -686194.443937535, 0, -340611.113390085, 2.67850784016458e-05, 0, -0.000655127339317701, -1446805.55593435, 0, -0.001255106195186, -545638.888788847, -410055.555498122, -69444.4443998906, 0, 0.00369649938100175, 0, 2485694.44205201, 0, 782638.900812994, -0.0010227903317424, 0, 0.00236977860021024, 10565694.440692, 0, 6762638.89548205, 0.00735210085484122, 0, 0.00828652259722129, 3703749.99904839, 1653472.22060962, 88194.4421476177, 8068888.88881688, 1306250.00063884, 959027.782536995, 0, -0.00807588118500094, 0, -4850694.43854041, 0, -937500.035025463, 0.00492607102499949, 0, -0.00518686916518751, -35992361.0931743, 0, -14607638.9258451, -0.0209434998959494, 0, -0.0229599884275703, -11135416.6608624, -2868055.54253422, -55555.5425288586, -44811805.5464826, 0, -0.0445778835851623, -52586111.1099398, -7774305.56185183, -30760416.6554359, -2979166.70285259, -1545138.87796988, -1104166.67426892, -17697916.704621, 0, 0.0110992425192797, 0, 5256944.43568347, 0, 569444.504892998, -0.0109634683403231, 0, 0.00726741233282483, 65138888.8538932, 0, 16354166.7665515, 0.0289435034379206, 0, 0.0363450225434629, 16631944.4273678, 2291666.62619151, 13888.8607554973, 129722222.162577, 0, 85625000.0074447, 0.123024953412095, 0, 0.0955302207134188, 148090277.762685, 18368055.5680052, 49965277.7087349, 3472222.30507397, 902777.716066606, 624999.953108327, 179131944.414142, 7881944.433361, 2361111.25497678, 101388889.03425, 51770833.4853278, 19062500.0788752, 0, -0.00932419565034802, 0, -2986111.1032453, 0, -138888.949740056, 0.0127812172002179, 0, -0.00613067938151976, -64722222.1854892, 0, -9236111.25191549, -0.015219389860143, 0, -0.0355922339214807, -12152777.7524938, -694444.384353755, 0.0286858064563462, -185069444.351913, 0, -82638888.9504058, -0.171815655088948, 0, -0.0999821752926793, -203819444.399591, -18750000.0038954, -37152777.6303469, -1388888.95412793, -208333.2243964, -138888.756334255, -196874999.912347, 0, -0.208762163220661, -406944444.387313, -13194444.446644, -251736110.977066, -1388889.27418978, -3819444.34327536, -694444.548617666, -223263889.124054, -56250000.1542519, -10069444.3620908, -259375000.365988, -94097222.5331977, -38194444.9338056, 0, 0.00437209453053758, 0, 694444.440514436, 0, 0.0330800467128816, -0.00752550215883398, 0, 0.00281328480469401, 33333333.312686, 0, 2083333.43299071, -0.00203733882481927, 0, 0.0202762541776487, 3472222.20388316, -0.0430217863172724, -0.0140135252706606, 128472222.159165, 0, 38888888.9723011, 0.112884553799897, 0, 0.0540460914266475, 135416666.61703, 6944444.43551424, 10416666.5335112, -0.000745971912324641, -0.0789042510971612, -0.115525907317587, 194444444.337748, 0, 135416666.619983, 0.212399191661958, 0, 0.142391675654016, 395833333.261952, 6944444.44873363, 156249999.813517, 0.372627382086705, 694444.308782126, -0.0708105099028145, 208333333.492436, 20833333.3187449, 2083333.08271288, 409722222.10243, 3472222.14984793, 0.361802522597829, 416666667.049051, 145833333.727144, 236111111.32219, 20833333.9704561, 41666666.7365369, 10416666.9884797, 159722223.043414, 0, -0.000876389520599588, 0, 0.000836328151846625, 0, -0.0075065777967317, 0.00177040296772818, 0, -0.000531303762967048, -6944444.4395382, 0, -0.0279829500840792, 0.00321611507576951, 0, -0.00501233544771896, 0.00522351149463585, 0.0118752599335753, 0.00264772272398347, -34722222.2055793, 0, -6944444.4798362, -0.027094207748464, 0, -0.0129320778804708, -34722222.2033369, 0.00606619927965103, 0.0437280087363, 0.0147130296852739, 0.0203152673216138, 0.0330284452360403, -69444444.4067989, 0, -34722222.2201673, -0.0781832791160122, 0, -0.03466020994844, -138888888.854099, 0.00120726392547509, -34722222.139335, -0.117440090082336, 0.0525173631158662, 0.0665001249407419, -69444444.4789686, 0.0563702970804934, 0.123453050649253, -69444444.4034109, 0, -0.075880891174474, -208333333.285046, 0.0206734197700366, -138888888.813343, -0.224833102568015, 0.0557724706576898, -0.0687467283334871, -208333333.465004, -69444444.6032461, -69444444.4190177, -0.227884045800099, -6944444.34649719, 0.013983176150396, -208333333.451282, -34722222.3241867, -0.338539200751228, -138888889.329895, -69444444.8811536, -34722222.5626897}, 
                                                     {0, 0, -4.71025205605064e-07, 0, 0, 1.27899226491644e-05, 0, 2625.00000374719, 4.58812514466202e-07, 0, 0, -0.000139579144344082, 0, -60697.9166355484, 0, -38822.9167927847, 5.42883599836556e-06, 0, -3.53423118855661e-05, -12572.9166367848, 0, 0.000812259106850489, 0, 515451.38849952, 0, 148177.084716656, -0.000166904254536873, 0, 0.000285174763687458, 882378.472111943, 0, 0.000501847173247026, 275399.305331486, 170624.999529729, 22447.9164673682, 0, -0.00281455994664198, 0, -2124045.1373185, 0, -242187.507415821, 0.00122246069793229, 0, -0.00106858606602465, -7288020.82932529, 0, -3218663.19613045, -0.00263457536671997, 0, -0.00365734347547041, -2133506.94326749, -502083.329274639, -17708.3313723857, -3683333.33044763, -464670.138701072, -277604.166380148, 0, 0.00608882491077838, 0, 4670138.88544693, 0, 182291.688852569, -0.0043328837285377, 0, 0.00239710175881386, 28820312.485449, 0, 4992187.51447149, 0.00647884464309508, 0, 0.0108230480061716, 7579861.10609991, 552083.314943457, 5208.32412498689, 24585069.4284748, 0, 0.0183105480066957, 27835069.435684, 3250000.00332168, 10329861.0872665, 562500.000766619, 345486.105072898, 197916.663013632, 5683159.73026503, 0, -0.00829885997122019, 0, -5607638.88429519, 0, -52083.3722218855, 0.00850798034290443, 0, -0.00345359841360831, -59878472.1971781, 0, -3550347.26776706, -0.00688649275042176, 0, -0.0185878445038667, -13177083.3202962, -208333.28954526, 0.0199574665818918, -85425347.1674093, 0, -34652777.7515802, -0.0477782115926996, 0, -0.0411825319049477, -95052083.3122896, -9626736.11847433, -10746527.6891937, -312499.988638578, -95486.0779337743, -52083.2974079973, -71458333.2759909, -2152777.78056881, -208333.365882714, -38958333.3920196, -10937500.0083041, -3836805.57797004, 0, 0.00692339515379978, 0, 3472222.21847867, 0, 0.0396873379523527, -0.00935305710118755, 0, 0.00292863328989485, 67013888.8649822, 0, 954861.180917749, -0.000213499437040814, 0, 0.0192363078378489, 10937499.9816093, -0.0563295853483902, -0.0218517384196485, 143749999.927708, 0, 22048611.1010252, 0.065365211784757, 0, 0.0474576494756891, 155729166.628732, 11979166.6665238, 3819444.29073394, -0.0472059460982849, -0.0618658857357238, -0.0802758680903649, 101302083.238167, 0, 0.0828271836537511, 207638888.81326, 5034722.24399112, 66666666.5009714, 0.0987562516243131, 520833.296394895, 0.00668534794538917, 111371527.893898, 5729166.57281104, 954861.054062611, 67708333.4227647, 23611111.2693768, 3819444.59455314, 0, -0.00322772967542269, 0, -868055.553841511, 0, -0.0218367642763769, 0.00537854064910334, 0, -0.00135797911118701, -38194444.4321219, 0, -0.0522857266453207, 0.00533717175634639, 0, -0.0112361798879829, -3472222.2092906, 0.0371698750874531, 0.0120933182402985, -114583333.288396, 0, -5208333.34765671, -0.0423616178311259, 0, -0.0287973905836588, -119791666.630433, -5208333.32371947, 0.123911845168405, 0.0592971320050288, 0.0480638523230742, 0.0673717098330588, -123263888.799637, 0, -51215277.70455, -0.0816297022516451, 0, -0.0581158686072185, -249999999.936394, -3472222.24097897, -20833333.1395282, -0.0905639823329288, 0.0611292797068592, 0.062508656940317, -130208333.415968, 0.188297947242946, 0.138458684451658, -154513888.752947, -868055.546724683, -0.102622471658141, -156250000.153007, -53819444.6728663, -31249999.9204791, -0.147643355697675, -5208333.37982417, -0.0988022032585774, -20833333.6814103, 0, 0.000643928351773913, 0, -0.000337529123261194, 0, 0.00500740656744054, -0.00126024245569729, 0, 0.000264020789017884, 8680555.55286031, 0, 0.0153221000711851, -0.0026118173293149, 0, 0.00276290273797298, -0.00359196782163019, -0.00977488786797044, -0.00269230280177233, 34722222.2111701, 0, 0.010739615972649, 0.00987255891366636, 0, 0.00777425478569367, 34722222.2091249, -0.00563659826732055, -0.0378434487740236, -0.0244308378906004, -0.0135592004902873, -0.0195403054924358, 52083333.3052863, 0, 8680555.53622299, 0.030222835608306, 0, 0.0150417607474596, 104166666.641355, 0.00309504576422545, -0.0780457822453958, 0.0229116615715482, -0.0263835272989753, -0.0393632921500843, 52083333.3493474, -0.096846315737142, -0.0668283926493712, 34722222.1839648, 0, 0.0287422624115714, 104166666.621349, 0.00312204974465042, 34722222.1442243, 0.0711655543499939, -0.018023392532441, 0.0103702548818844, 104166666.731486, 34722222.3187769, -0.120006941762796, 0.0186576846052967, -0.0462188679020786, -0.0251698476727634, 52083333.3369027, 8680555.62569034, 0.110638008180475, 34722222.4471797, 0.140121188119201, 0.154700938623762}, 
                                                     {0, 0, 2.27514284856111e-07, 0, 0, -6.16480421985658e-06, 0, -1200.00000144596, -3.17200151496406e-07, 0, 0, 6.6777994049239e-05, 0, 29033.3333141077, 0, 16033.333384381, -2.28973746304205e-06, 0, 1.6004776149946e-05, 4033.33332023459, 0, -0.000385014039954549, 0, -263527.777590907, 0, -44833.3339148652, 6.67551028742511e-05, 0, -0.00014757774773107, -383361.110922792, 0, -0.000154693131998225, -93027.7776700603, -49333.3330839867, -4499.9999491321, 0, 0.00132128312803831, 0, 1183611.11045706, 0, 46666.6698605341, -0.000454390427679924, 0, 0.000659597616663397, 3415972.21986896, 0, 1029027.77747909, 0.000529221385908967, 0, 0.00131629974909676, 780694.443894841, 94999.9979108067, 1666.66638881476, 1127777.77621306, 98749.9999966062, 49999.9993706287, 0, -0.00283188490239909, 0, -2871527.77654166, 0, -16666.6763823428, 0.00149182130950772, 0, -0.00172009747291395, -14933333.3259933, 0, -1022222.22478772, -0.000796465543326438, 0, -0.00482233882672121, -3097222.21996872, -49999.9912220832, 0.00125220314499785, -8567361.10005005, 0, -0.00472034238534403, -9327777.77348274, -760416.668098336, -2079166.6545751, -49999.9952227535, -34722.2218572747, -16666.6658186429, -1091666.66625113, 0, 0.00382660134897301, 0, 3798611.10972383, 0, 0.0172507058897511, -0.00280868542876079, 0, 0.00272058075449021, 34874999.9886539, 0, 347222.234578861, -0.000123618491763208, 0, 0.010243028812368, 6159722.21646998, -0.0196051985401177, -0.00261982336744685, 33576388.8585105, 0, 7847222.19820276, 0.00912954474242747, 0, 0.0126871964845754, 36180555.5469693, 2604166.66943966, 1041666.62464933, -0.0206279414832805, -0.0046836590295624, -0.00539412999581316, 15937499.9737949, 243055.561224244, 0.00142612319727821, 8333333.34771287, 1041666.64683669, 347222.228679343, 0, -0.00316746306887955, 0, -2569444.44352412, 0, -0.0177957053075881, 0.00300313306433335, 0, -0.00252176928349772, -43819444.4347427, 0, -0.0220522972768735, 0.00195434491744348, 0, -0.0120968175762869, -5833333.3252612, 0.0238290279769934, 0.00248957148166271, -65555555.5194056, 0, -2430555.53110555, -0.011129630286617, 0, -0.0194676771367061, -69513888.8744773, -3958333.33231743, 0.0695426534428835, 0.0403813168065299, 0.0107052964077961, 0.0117870707203714, -26736111.0538465, 0, -0.0181025205235076, -54166666.6377241, -694444.459340028, -7291666.59392227, -0.000469762413272794, 1.20701078683983e-05, -0.00295809210110197, -28125000.0366583, 0.0882986432054427, 0.00412107581078065, -7291666.66490057, -2430555.60688811, -0.0196774798508952, 0, 0.00146633812936964, 0, 694444.44411478, 0, 0.00988352877324329, -0.00171001793190369, 0, 0.00127301156912791, 27777777.7733696, 0, 0.0179427529143063, -0.00237053233477995, 0, 0.00763119285253271, 2083333.32767583, -0.0150477246509784, -0.00112083186947159, 60416666.6459681, 0, -0.00965347966521541, 0.00680411404890311, 0, 0.0152316885561, 62499999.9861791, 2083333.32837882, -0.054033879607638, -0.0364787562237285, -0.00879051582610679, -0.00999114302345881, 40277777.7304431, 0, 6944444.39514892, 0.0145229516268516, 0, 0.0153044547027824, 81249999.9787672, 694444.455991488, -0.0836017692916355, -0.00670109148024099, -0.00816952080370171, -0.00878642410986893, 41666666.6942988, -0.121651817088165, -0.0241094341467179, 20833333.2802748, 0.00928436553642408, 0.0140103058068946, 20833333.3727322, 6944444.52321807, -0.0847401433822461, -0.0181560742714408, 0.0285377752185276, 0.0292479515330339, 0.0807431300610273, 0, -0.000290700463969246, 0, 4.80504244716993e-05, 0, -0.00228500439604619, 0.000403664946300604, 0, -0.000269252459977601, -6944444.44361071, 0, -0.00555356319212458, 0.000894613242403973, 0, -0.00197234968454776, 0.0015667499335016, 0.00386135427211947, 0.00016702461688381, -20833333.3286585, 0, 0.000237139404861867, -0.00145285906815052, 0, -0.00478791570046649, -20833333.32824, 0.00271131576112972, 0.0159922611898172, 0.0124583457408857, 0.0024610095783376, 0.0025464776862709, -20833333.3195741, 0, 0.0155887284910294, -0.00509876257894889, 0, -0.00523576978119737, -41666666.6583899, -0.00216270456015969, 0.0330782456942146, 0.00588257752394556, 0.00448732545376682, 0.00590618552232302, -20833333.3381926, 0.0537229813322115, 0.0128301827481345, -6944444.4229278, 0, -0.00540425458554265, -20833333.3181004, -0.00580497599517069, 0.0308933780353805, -0.00890870129623032, 0.000176725375165754, -0.00218859708372791, -20833333.3558392, -6944444.47843488, 0.0726848842788474, 0.03134362333738, 0.00533452389009724, 5.22432682340625e-05, 0.0132089525373706, -0.0280329200643382, -0.0182032039730791, -0.0681400466799292, -0.0110914636350321, -0.0452635645022804}, 
                                                     {0, 0, -8.45868384542151e-08, 0, 0, 2.28223910451213e-06, 0, 300.000000711555, -3.90411215203162e-08, 0, 0, -2.47797532997548e-05, 0, -6883.33333092976, 0, -5133.33335558783, 1.94830618404189e-06, 0, -2.39954279581752e-06, -633.333328354328, 0, 0.000143664258147389, 0, 57277.7777271647, 0, 24833.3335644579, -2.94559338178028e-05, 0, 1.84376169326879e-05, 117111.11114883, 0, 3.76299906429613e-05, 13861.1110659893, 10166.66659068, 333.333301562318, 0, -0.000496625393337054, 0, -224305.555321679, 0, -36666.6678585952, 0.00018886238601503, 0, -5.44655536332631e-05, -965416.666293155, 0, -559027.778420978, -0.000204022014057747, 0, -0.000272225330483957, -106249.999756545, -41666.6660706243, 0.000235442471187963, -221805.554959498, -6944.44425427676, -4999.99981769752, 0, 0.00107298723436717, 0, 437499.99943621, 0, 16666.6701379673, -0.000590152021307676, 0, 8.76867853870012e-05, 3725694.44267806, 0, 797222.225731354, 0.000327878793157056, 0, 0.00079955047643081, 361111.110233973, 33333.3307423042, -0.00115050055167007, 4506249.99911028, 0, 0.00159494540096829, 1690972.22012765, 48611.1105445371, 901388.884916081, 16666.6661623611, -0.00105429874381171, -0.000520771722197392, 104166.66584076, 0, -0.00146178318900648, 0, -402777.776965146, 0, -0.00596929914101117, 0.00114091402389137, 0, -0.000195970385791131, -7104166.66320923, 0, -347222.231584015, -0.000300531008706256, 0, -0.00107444302125745, -541666.664724242, 0.00586220806717786, 0.00263807157894565, -16770833.3277396, 0, -6041666.66705353, -0.00492237564332769, 0, -0.00315242300768175, -5694444.43950112, -138888.887915135, -694444.431382071, 0.00307584140239234, 0.00421343637671319, 0.00316689640834145, -6770833.32290362, 0.00196752603616528, -0.00306308923840905, -729166.666057742, -347222.216440664, 0.00177120005454661, 0, 0.00121968768384342, 0, 138888.888185144, 0, 0.00600391348967471, -0.0012194006725191, 0, 8.39807962403517e-05, 6319444.44077697, 0, 0.0131970153307065, -0.000629514117194425, 0, 0.00123029469109335, 277777.775264761, -0.00749446612632426, -0.00288620807329135, 30138888.8803215, 0, 2430555.56030525, 0.00625415292566562, 0, 0.0035131207414289, 8402777.77037617, 138888.886611459, -0.0213264570215907, -0.00829783604721105, -0.00761282287298485, -0.00793350758480401, 20138888.8805819, 0, 0.00705628154410477, 22222222.2069364, -0.00239949807042158, 4861111.08541321, 0.010113657170388, -0.00697744545309474, 0.00385190315056571, 2083333.33356486, -0.0246761823841654, -0.00771656381319729, 2430555.54669798, 0.00373085054723785, 0.00579602835647278, 0, -0.000568951800262389, 0, 0.000337403934471387, 0, -0.00326440596833655, 0.000696843254010606, 0, -3.47395551318516e-05, -2083333.33125796, 0, -0.00940796547198188, 0.00105537664456646, 0, -0.000704531667000584, 0.00163179841645519, 0.00495571781148316, 0.00184335730423934, -24305555.5496485, 0, -0.00674901030765858, -0.00377626918713473, 0, -0.00142590376249879, -4166666.66062002, 0.00244270643945636, 0.0166584488977324, 0.00931019081723312, 0.00617819452485204, 0.00697325464028023, -29166666.6571751, 0, -6944444.4392452, -0.0073384224795789, 0, -0.00473140064868202, -31249999.9875833, 0.00132219790028012, 0.0268077288070495, -0.00983112610288055, 0.00793215442990583, 0.00490736101842762, -2083333.33340858, 0.03367056510749, 0.0143120327309532, -13888888.8664991, 0.00458499729156785, -0.0136432008982739, -6944444.44167748, -0.00673369574834311, 0.0333431055957623, -0.00169669739264322, 0.00343945345999889, -0.00828035807860695, -0.0119386621081724, 0, 0.000113603127315558, 0, -6.87757718767684e-05, 0, 0.000740937343947309, -0.000162844989137782, 0, 5.39430608501884e-06, -0.000490566665845983, 0, 0.00266716577365312, -0.000383663374651315, 0, 0.000181374597924595, -0.000433703724582624, -0.00128241786735718, -0.000431780938560924, 6944444.44283785, 0, 0.00297669344219389, 0.000895104225977612, 0, 0.000442889411266069, -0.00191707269557608, -0.000952645725873618, -0.00497407141251472, -0.00359870616247713, -0.00170729111081608, -0.00214997736633782, 13888888.8856282, 0, -0.000812136176153559, 0.00306822424478402, 0, 0.00085805374631398, 13888888.8841651, -0.000916755784732511, -0.0103481519898721, 0.00180458782943203, -0.00327366585067263, -0.00387074902041778, -0.00104128187775268, -0.0157299931615315, -0.00703639181235617, 6944444.44064611, 0, 0.00236697530375091, 13888888.8812292, -0.00124672365185501, -0.0108491521890849, 0.00868178147112056, -0.00280230886096363, 0.00252035544280423, 6944444.44332504, 0.00267956458223588, -0.0239704966136799, -0.00564840518437467, -0.00742622418114787, -0.00176965684722012, -0.00978096032267878, 0.000577595898051249, 0.00966358215560501, 0.0103076382617628, 0.00125474693666563, 0.00939755529915412}, 
                                                     {0, 0, -4.97178639040142e-08, 0, 0, 1.38334662446766e-06, 0, 281.250000660297, 1.27706593329388e-07, 0, 0, -1.58831168122127e-05, 0, -5749.99999676921, 0, -5750.00002281544, -2.34397614167867e-06, 0, -3.35994145854931e-06, -593.749998080291, 0, 9.76201034898169e-05, 0, 39322.9165944692, 0, 39322.9169129091, 1.59264711763019e-05, 0, 1.91925026271875e-05, 116927.083393648, 0, 6.73699753929136e-05, 11510.4166207144, 11510.416641998, 312.499989754662, 0, -0.000355936840411377, 0, -111979.166298723, 0, -111979.167973441, -5.78467816246469e-05, 0, -5.51361874933418e-05, -791753.471783098, 0, -791753.473178364, -0.000555747502890578, 0, -0.000315050155532901, -70833.3330472354, -70833.3331283316, 8.9077513513762e-05, -222482.638397763, -5729.16635576754, -5729.16686140888, 0, 0.000805301364176059, 0, 130208.332357932, 0, 130208.337202739, -7.39395717064995e-06, 0, 0.000112829937359827, 2214409.7198193, 0, 2214409.72749403, 0.00208500058882394, 0, 0.000301695482579024, 161458.332363941, 161458.332435386, -0.000277402336367087, 5262152.77706576, 0, 0.0022729065688615, 1361111.10890657, 31249.9982111038, 1361111.10820989, 31250.001444293, -0.00124362615783206, 0.00061803808844186, 105034.720541763, 0, -0.00114067409145433, 0, -52083.3317870556, 0, -52083.3400224929, 0.000176844284902155, 0, -9.25035398197746e-05, -2491319.43942386, 0, -2491319.45864685, -0.00349606639618819, 0, 0.000100502563939368, -104166.664690782, -104166.664278341, 0.000359427546464321, -14210069.4368525, 0, -14210069.4473546, -0.00810300274438449, 0, -0.00339418352769749, -3064236.10597159, -52083.3280227404, -3064236.10228776, -52083.3382603624, 0.00297987485544706, 0.00149403227793298, -8229166.6575731, 0.0056577300902224, -0.00892570896833894, -572916.659501316, -572916.65896934, -0.000487110813948115, 0, 0.000983312510835538, 0, -0.00147250378916908, 0, 0.00671111824337683, -0.000358369772501314, 0, 0.000180437105468287, 954861.105524808, 0, 954861.13104078, 0.00299653467521212, 0, -0.000640728056709845, -0.00224520800076633, -0.00339642756806659, -0.000346115427263574, 14930555.5428113, 0, 14930555.5677895, 0.013784170001418, 0, 5.522636547516e-05, 1909722.21514793, -0.00721211359563866, 1909722.20808876, 0.0054680594517714, -0.00337330433660974, -0.00594478465404692, 35763888.8785525, 0, 0.0120098513952224, 18055555.541734, -0.0120459627593647, 18055555.5373094, 0.025837753147869, -0.00988160305248999, 0.00519437887255224, 954861.097593585, 954861.091579377, -0.00285552349929982, 3124999.97596511, -0.00550881970582845, 0.00340707322860584, 0, -0.000471364025780629, 0, 0.000773898988564448, 0, -0.00362412227968735, 0.000242973291926244, 0, -8.75647102204354e-05, 0.00332824235109562, 0, -0.0139811300157481, -0.00119706673749878, 0, 0.000293003640158155, 0.00142427246211615, 0.00243687012938391, -0.000214996169350845, -5208333.32428035, 0, -5208333.34844327, -0.0101258605665549, 0, 0.0015611746115989, 0.00502477966977821, 0.00513180175208371, 0.0112015598215663, -0.00206446563795692, 0.00147533689404016, 0.0046925760265281, -32118055.5409666, 0, -32118055.5518566, -0.015274752235479, 0, -0.00520137926454965, -10416666.654422, 0.0121053461303505, -10416666.6483308, -0.0270054436075338, 0.0087532441868069, 0.00351966002935112, 0.0137699849732439, 0.0241148048771812, 0.00599414225469451, -37326388.8726124, 0.0139448770858082, -0.0255608286995986, -5208333.30609817, 0.0112137029715046, -5208333.2966007, -0.00669143018191922, 0.00646676762457735, -0.0082497855835299, -0.00187575825654666, 0, 9.62895332593859e-05, 0, -0.000171864206987182, 0, 0.000815505178190906, -5.69269856176258e-05, 0, 1.92969941169995e-05, -0.00083749244261666, 0, 0.00387605987013759, 8.87281400063762e-05, 0, -4.69461858332566e-05, -0.000372341496785095, -0.000665457976180656, 0.000103747106059453, -0.00247352074738853, 0, 0.00615081147481023, 0.00253955679613815, 0, -0.000734436521494638, -0.0014671358291713, -0.00148349441121127, -0.00352349251062229, -8.71704563746013e-05, -0.000326572716695267, -0.0012567166703316, 8680555.55013636, 0, 8680555.55682516, 0.00590518731712302, 0, 0.000308072424942182, -0.00411138298847443, -0.00387112045629938, -0.00689824163256193, 0.00943487459527922, -0.00260025719312721, -0.00410627068131387, -0.00466168170771745, -0.00964969900360858, -0.00233162645033873, 17361111.1057591, 0, 0.0045825011833756, 17361111.1044889, -0.00650110207459226, 17361111.1035225, 0.0167978450560853, -0.00589847282827932, 0.00513006679256413, -0.0107249759380909, -0.0059833457102901, -0.0206461340951714, -2.32646986970849e-05, -0.00309263157051889, 0.000691336103352933, 8680555.53520936, -0.00481774715012606, 0.0114833498653787, -0.00580881460248192, -0.00627533464479271, 0.00601119200529394}, 
                                                     {0, 0, -2.15153836497885e-08, 0, 0, 6.06903889688166e-07, 0, 300.00000120404, 8.03393783880428e-07, 0, 0, -7.1625584401973e-06, 0, -5133.33333979507, 0, -6883.33336650164, -9.85343011952039e-06, 0, -1.65326169291493e-05, -633.333340088406, 0, 4.48590583662755e-05, 0, 24833.3333183644, 0, 57277.7780988797, 6.19618943516206e-05, 0, 0.000137302758585128, 117111.111343035, 0, 0.000160993833243658, 10166.6666791563, 13861.1112019473, 333.333451940194, 0, -0.000164661886702486, 0, -36666.666467556, 0, -224305.557133578, -0.000191427567338555, 0, -0.000601842929703281, -559027.778238946, 0, -965416.668757583, -0.000770774385438246, 0, -0.00108731823917705, -41666.6666046852, -106250.00048302, -0.000986649409347621, -221805.555965809, -5000.00006121742, -6944.44548801458, 0, 0.000371352040714472, 0, 16666.6659819752, 0, 437500.004383608, 0.000401884949187124, 0, 0.00157232701203812, 797222.222115168, 0, 3725694.45355658, 0.00217077010638257, 0, 0.00389776523748803, 33333.3328703766, 361111.112335567, 0.00385062790610774, 4506250.00429127, 0, 0.00445040446305107, 901388.89051013, 16666.6651963681, 1690972.22393109, 48611.1156119716, 0.00143708279491449, 0.00665923368667402, 104166.668668384, 0, -0.000520698779480281, 0, 0.00124544221789372, 0, -402777.784943166, -0.000471003341932724, 0, -0.00252777694140385, -347222.220599977, 0, -7104166.68789717, -0.00348795772722192, 0, -0.00772356715756984, 0.00115577273352789, -541666.668355352, -0.00841144885199454, -6041666.66889736, 0, -16770833.3504032, -0.010062637989298, 0, -0.0108659265919305, -694444.447424928, 0.00509866359389583, -5694444.44670849, -138888.897553055, -0.00688396256847975, -0.0195230863890522, -6770833.34126604, 0.00426101382047383, -0.0203371280713185, -347222.221978256, -729166.672309007, -0.0114654787367835, 0, 0.000442340889738716, 0, -0.00129825523110335, 0, 138888.895729756, 0.000386756626099572, 0, 0.00232339779105838, -0.00252547493552077, 0, 6319444.47126933, 0.00296454912393104, 0, 0.00917287953048757, -0.00158604374211764, 277777.779275962, 0.0105939490479738, 2430555.55288887, 0, 30138888.9219917, 0.012040581372643, 0, 0.0153446306125725, 0.00258693659994098, -0.00773986086126982, 8402777.77790518, 138888.898734484, 0.0131690134050564, 0.0313206627562525, 20138888.8995123, 0, 0.0183413147986423, 4861111.12518377, -0.0139736526300538, 22222222.2355662, 0.0319336939908216, -0.00192046284085588, 0.0447722361620562, -0.0069985502309313, 2083333.33504499, 0.0341351320754442, 2430555.56313442, 0.00211958839666677, 0.0334767209846682, 0, -0.000208384155178569, 0, 0.0007292409931628, 0, -0.00353599809520381, -0.000172266681506268, 0, -0.00119734962970598, 0.00176919941370295, 0, -2083333.35066001, -0.00122533000549758, 0, -0.00572460983425988, 0.001099881829458, -0.000847590474923224, -0.00666873884207813, 0.00353944024797569, 0, -24305555.5856224, -0.00724645439804069, 0, -0.0109697182855325, -0.00094509226885863, 0.00542794531004175, -4166666.66484958, -0.00591741034134495, -0.0115331043237617, -0.0239227069152314, -6944444.44362462, 0, -29166666.6844039, -0.0154262160686002, 0, -0.0154959397936467, -0.0107845146078908, 0.0141538547661491, -31250000.009137, -0.0242139776057476, -0.00338143330805452, -0.0405885992652929, 0.00770564488126534, -2083333.32649528, -0.0377889112970684, -13888888.9084247, 0.0128869980048768, -0.0455660600675841, -0.00287687147866465, 0.0140523895368342, -6944444.44781526, -0.0199244722512703, -0.0164122573769497, -0.05960665764434, -0.0151159374846414, 0, 4.17690390920182e-05, 0, -0.000170614148722228, 0, 0.0007657745062218, 2.87769428022122e-05, 0, 0.000252877988154137, -0.00049703281821246, 0, 0.00448792321106193, 0.000219851392289182, 0, 0.00152034524007434, -0.000311209677950755, 0.000161349490845518, 0.00173903372781784, -0.0011671557814098, 0, 6944444.4546715, 0.00181661485760549, 0, 0.00317848069847689, 5.2930064294208e-05, -0.00146484500774973, -0.00090842069222295, 0.00152042330369056, 0.0040506127995216, 0.00761920073986149, -0.00197118623282309, 0, 13888888.8982838, 0.00502502435998951, 0, 0.00502908579005422, 0.002919488383797, -0.00527561758302888, 13888888.890749, 0.00778201685811744, 0.00275975372496437, 0.0146546438570107, -0.00281787225948585, -0.00478964290786304, 0.014405975293403, 6944444.44621496, 0, 0.00657707958784017, 0.00814968276437206, -0.00777726669686924, 13888888.8979472, 0.0179210422679853, -0.00338603485742533, 0.0205548537241694, -0.00106486044342247, -0.00999286831870192, 6944444.44171795, 0.00140302093722583, 0.0100857831729751, 0.031103787166116, 0.00471616535085173, -0.0044524863018238, 0.0255015844201342, 0.000917664964038287, 0.00749976791348977, 0.0235636143280189}, 
                                                     {0, 0, 0, 0, 0, 3.00036639133675e-07, 0, -1200.00000259701, -1.28658135882425e-06, 0, 0, -3.16076949494566e-06, 0, 16033.3333567728, 0, 29033.3333985181, 9.26738267910891e-06, 0, 2.62793734020618e-05, 4033.33335427926, 0, 1.82988954926767e-05, 0, -44833.3333683029, 0, -263527.778357516, -5.46315696364915e-05, 0, -0.000257748643562119, -383361.111751549, 0, -6.58588795708999e-05, -49333.3334026946, -93027.7780457868, -4500.00031102919, 0, -6.88288721472905e-05, 0, 46666.6664147333, 0, 1183611.11375457, 0.00021774150913846, 0, 0.00132379690120133, 1029027.77987039, 0, 3415972.22731879, 3.56179962409338e-06, 0, 0.000630979815174969, 94999.9999749734, 780694.446149888, 1666.66932779174, 1127777.77917788, 50000.0005048108, 98750.0017636331, 0, 0.000173088232181624, 0, -16666.6653734345, 0, -2871527.78461247, -0.000558155557881245, 0, -0.00388948150318205, -1022222.22501763, 0, -14933333.35351, 0.000439032234829722, 0, -0.00313684676787403, -49999.9990841118, -3097222.22787687, -0.011302191688763, -8567361.12589564, 0, -0.000307475110601511, -2079166.67161383, -49999.9983350198, -9327777.78651732, -760416.672955673, -16666.6728544351, -34722.2339457002, -1091666.66768727, 0, -0.000282204390031077, 0, -0.00272209655332812, 0, 3798611.12149502, 0.000844235947568854, 0, 0.00686042451658989, 347222.222712842, 0, 34875000.0435188, -0.0016721823573902, 0, 0.00804988225053314, -0.00315266122323848, 6159722.23295, 0.0259044941208474, 7847222.24079892, 0, 33576388.9380933, -0.000442533352483954, 0, 0.000912443716194438, 1041666.67560423, -0.00679807812611866, 36180555.5792451, 2604166.67958751, 0.0256628911720371, 0.0401955391795261, 15937500.0310999, 0.000434055119763181, 243055.584533384, 1041666.65441951, 8333333.33595912, 347222.241015065, 0, 0.000281703987112551, 0, 0.0030487612151614, 0, -2569444.45362476, -0.000862813876631276, 0, -0.00700461920507553, 0.00286350157565388, 0, -43819444.4959632, 0.00280937366559747, 0, -0.0121474810429999, 0.00489163336154955, -5833333.34507284, -0.0331552006314735, -2430555.56486206, 0, -65555555.639459, 0.00256518200388592, 0, -0.00165962644887425, -0.00707133000392616, 0.0102881550145714, -69513888.9199662, -3958333.34738935, -0.0482843937994129, -0.0668681909567422, -26736111.160899, 0, 0.00135258335165101, -7291666.71852044, 0.0129546248865879, -54166666.7353459, -694444.483335664, -0.0165275608695502, -0.0722061581458728, 0.0283249832761505, -28124999.997692, -0.076614676212808, -7291666.64942738, 0.0228800805095292, -2430555.58592525, 0, -0.000155085103444502, 0, -0.00177879870726806, 0, 694444.44881762, 0.000482080315925568, 0, 0.00384544183816404, -0.00313282691075923, 0, 27777777.8091756, -0.00222821232729519, 0, 0.00936296455438734, -0.00368003485334132, 2083333.34032442, 0.0219719316349041, -0.00018728446154959, 0, 60416666.7360366, -0.0038938358053447, 0, 0.00212856333233232, 0.00136314168697307, -0.00768766795871403, 62500000.0201569, 2083333.34195331, 0.0424894834635724, 0.0564717964490087, 6944444.4641387, 0, 40277777.8387824, -0.00089890891744241, 0, -0.00282605940395151, 0.0386180632995268, -0.013454729356979, 81250000.0642078, 694444.467993845, 0.0298808524816102, 0.0730840361536639, -0.0236358195809324, 41666666.6583764, 0.0962973141502899, 20833333.4170726, -0.00390526464945451, 0.0651476474505569, -0.0195949383813114, -0.0575499538339491, 20833333.3269363, 6944444.45185197, 0.0218391356916246, 0.104408991193591, -0.0427871679764678, 0, 3.59002609675745e-05, 0, 0.000424994665485114, 0, -0.000868194096263798, -0.00012175252960371, 0, -0.00087719039951738, 0.00105686218735876, 0, -6944444.45214794, 0.000591782085091817, 0, -0.00284567366398543, 0.00109018656753269, -0.00175335671752682, -0.00589079550565717, 0.00121021333019833, 0, -20833333.355276, 0.00175430277243908, 0, -0.00155192880039447, 0.000576603962999261, 0.00242051403037545, -20833333.3386135, -0.00258046536822771, -0.0143857405350586, -0.0186350580969249, -0.00154634004055063, 0, -20833333.3602491, 0.000359492267028564, 0, 0.00147368435466989, -0.0100917869143892, 0.0043434800012921, -41666666.6877239, -0.00451120831663515, -0.0148241328729234, -0.0277149609318796, 0.00770829525813536, -20833333.3294166, -0.0409103687192893, -6944444.45828675, 0, 0.000660591806099107, -0.0330381350664256, 0.00508216809678552, -20833333.3748164, -0.0221555762516575, -0.00397356273861494, -0.0317667457689989, 0.00784800368907267, 0.0270459543969692, -20833333.3328827, -6944444.43356364, -0.0264709846433595, -0.06786828778519, -0.00505824749241906, 0.0165096249600986, -0.042994513113113, 0.0424769356777029, 0.027239577146742, -0.0154989434883657}, 
                                                     {0, 0, -4.84184992612178e-08, 0, 0, 1.32990051307801e-06, 0, 2625.00000579313, 5.07859771233135e-06, 0, 0, -1.62942296819535e-05, 0, -38822.9167050206, 0, -60697.9168209869, -5.53605842814268e-05, 0, -9.99368964146562e-05, -12572.9167291329, 0, 0.000105023737262648, 0, 148177.083300801, 0, 515451.390334394, 0.000372451953626952, 0, 0.000867780145006288, 882378.473461771, 0, 0.000706305206190432, 170625.000158716, 275399.306483539, 22447.9175363246, 0, -0.000388424312390539, 0, -242187.499078823, 0, -2124045.14579819, -0.00141138386851073, 0, -0.00407608590451524, -3218663.19751661, 0, -7288020.84393399, -0.00355730394370332, 0, -0.00444268145035358, -502083.333022564, -2133506.95066953, -17708.3400957477, -3683333.33736357, -277604.167443416, -464670.146331172, 0, 0.000866304388558309, 0, 182291.663159702, 0, 4670138.90761552, 0.00308566784167003, 0, 0.0113906729236354, 4992187.5014578, 0, 28820312.5443577, 0.0110620936050189, 0, 0.0150533856488014, 552083.330283992, 7579861.13191596, 5208.36112097279, 24585069.4696276, 0, 0.0166948033142173, 10329861.121522, 562499.99282531, 27835069.4708336, 3250000.03793011, 197916.676455581, 345486.149590514, 5683159.7316005, 0, -0.00118413625380688, 0, -52083.3266469558, 0, -5607638.91878906, -0.00401539320227796, 0, -0.0193339050426378, -3550347.21601097, 0, -59878472.3224892, -0.0176494589755897, 0, -0.0312912596762199, -208333.32460013, -13177083.3716339, -0.0639390305524134, -34652777.7990655, 0, -85425347.3142832, -0.0449699554038518, 0, -0.0333285197910821, -10746527.7910199, -312499.976777659, -95052083.406907, -9626736.20836977, -52083.3796449765, -95486.2182064424, -71458333.4049676, -208333.312323134, -2152777.91992698, -10937499.9746941, -38958333.3915948, -3836805.6066495, 0, 0.000969887568234847, 0, -0.00713018363441998, 0, 3472222.25010799, 0.00327958177610687, 0, 0.0193425395122428, 954861.098465267, 0, 67013889.0126508, 0.013853974650373, 0, 0.040890300811589, -0.0125823567672984, 10937500.0403471, 0.0828941874267973, 22048611.1084103, 0, 143750000.168608, 0.0625130634969939, 0, 0.0347091029188092, 3819444.44861681, -0.0304186001482741, 155729166.767138, 11979166.7928669, 0.0962813114900472, 0.163667658977732, 101302083.405415, 0, 0.0683980738593092, 66666666.765402, -0.0654116042079007, 207638889.055787, 5034722.49493293, -0.00902222137607077, 520833.592271172, 5729166.59256253, 111371527.897009, 954861.266107025, 67708333.3415122, 3819444.37274322, 23611111.2983158, 0, -0.000436710925499981, 0, 0.00404232743104112, 0, -868055.569630902, -0.00154507508490621, 0, -0.0104378850670079, 0.00990187937387682, 0, -38194444.5228623, -0.00451958823008066, 0, -0.0292690337664558, 0.00923655114983219, -3472222.24526922, -0.0558610858779974, -5208333.31981452, 0, -114583333.480732, -0.0393324571152467, 0, -0.0218438200164812, 0.00517296191368348, 0.0198251814789497, -119791666.734309, -5208333.41483174, -0.0918295246551901, -0.133459650768785, -51215277.7898931, 0, -123263888.991363, -0.0729784734075194, 0, -0.0397763157463869, -20833333.3943348, 0.057385553750284, -250000000.161889, -3472222.4590195, -0.0261672644461703, -0.213946815706761, 0.0657382457327888, -130208333.433819, -0.183937790954294, -154513889.062674, 0.0641336186366349, -868055.884316261, -31249999.9558688, 0.162272529773382, -156250000.068708, -53819444.6709702, 0.0134181629985174, -5208333.65289517, -20833333.3097609, 0, 8.30627996075286e-05, 0, -0.000947648795298043, 0, 0.00297520338101685, 0.000334346781480341, 0, 0.00233683892415541, -0.00293951332460401, 0, 8680555.57552871, 0.000328537578378238, 0, 0.00858518733693278, -0.00272132434250628, 0.00549895327477937, 0.0151739505906258, -0.00559779424576143, 0, 34722222.2710246, 0.00876631559653388, 0, 0.00735446093104414, -0.00351286430541087, -0.00548953321277705, 34722222.2403874, 0.0212167164647039, 0.0330349107853958, 0.0446592091476337, 8680555.54998017, 0, 52083333.3832094, 0.0268399804015108, 0, 0.00752662221644807, 0.0128151676688945, -0.016927525960721, 104166666.722064, 0.07547088093425, 0.0217446097540397, 0.0698476689429072, -0.0204817013126371, 52083333.365695, 0.0787683990978026, 34722222.2382735, 0, 0.0243038243130077, 34722222.2827477, -0.0334159814187893, 104166666.757531, 0.148975266383595, -0.0166620528569786, 0.127943785656371, -0.0273647724686545, -0.0741326145943899, 104166666.708612, 34722222.3055626, 0.0214788822171244, 0.16472495956974, 52083333.352884, -0.0728091621306856, 8680555.75168951, -0.0864286102823439, 34722222.2347126, 0.0539639806968795}, 
                                                     {0, 0, 2.21459884988543e-07, 0, 0, -6.10682985989101e-06, 0, -4200.0000084164, -8.30815760093554e-06, 0, 0, 7.01601213294071e-05, 0, 69116.6667023819, 0, 90116.6669019647, 9.92839755136779e-05, 0, 0.000168862208444886, 27116.6667578874, 0, -0.000429710665200988, 0, -340611.110915903, 0, -686194.446730152, -0.000652387397087567, 0, -0.0013767181053857, -1446805.55706362, 0, -0.00156016822237269, -410055.555822822, -545638.890393567, -69444.4456220164, 0, 0.00155133419164535, 0, 782638.887094595, 0, 2485694.45570168, 0.00226559027102404, 0, 0.00605875459924032, 6762638.8911559, 0, 10565694.4583995, 0.00900189785553318, 0, 0.00927337425238345, 1653472.2218887, 3703750.01004334, 88194.452152945, 8068888.89585204, 959027.778983616, 1306250.01371929, 0, -0.00345715484192602, 0, -937499.994165834, 0, -4850694.47585203, -0.00418746317039972, 0, -0.0162269087201471, -14607638.884723, 0, -35992361.172434, -0.0286981359294566, 0, -0.0274128004914785, -2868055.55102423, -11135416.6981856, -55555.5850908319, -44811805.5811172, 0, -0.0434977859339642, -30760416.6827957, -2979166.65902301, -52586111.1541393, -7774305.63685667, -1104166.67120307, -1545138.94520842, -17697916.7008138, 0, 0.00480663144330542, 0, 569444.434101664, 0, 5256944.49610916, 0.00414600865568465, 0, 0.0268654820918597, 16354166.6480789, 0, 65138889.0327853, 0.0467205813498272, 0, 0.0500609459926869, 2291666.65304122, 16631944.4969447, 13888.9576730568, 85625000.0051245, 0, 129722222.328327, 0.124767367508425, 0, 0.0843356903401255, 49965277.7909185, 3472222.19739315, 148090277.889258, 18368055.7717646, 625000.023089277, 902777.893894198, 179131944.545758, 2361111.06482421, 7881944.72568538, 51770833.362434, 101388889.085104, 19062500.0924694, 0, -0.00405892815087417, 0, -138888.878330698, 0, -2986111.16079971, -0.00217232842167987, 0, -0.026467798055964, -9236111.08436943, 0, -64722222.4056698, -0.0372530578171505, 0, -0.05987172194065, -694444.424434787, -12152777.8263405, -0.0929104978324036, -82638888.856255, 0, -185069444.653999, -0.179206953177662, 0, -0.0796060455334016, -37152777.767266, -1388888.85748511, -203819444.584585, -18750000.2756929, -138888.953674759, -208333.475718288, -196875000.053356, 0, -0.19241729598547, -251736111.222667, -1388888.78378533, -406944444.668217, -13194445.0636941, -694444.358994639, -3819444.85366035, -56249999.9901838, -223263889.274166, -10069444.5699806, -259375000.221703, -38194444.437688, -94097222.7235888, 0, 0.00190411304410989, 0, -0.00581036555243995, 0, 694444.470320293, 0.000551288020002656, 0, 0.0141156321206072, 2083333.31495168, 0, 33333333.4530868, 0.0123957318842824, 0, 0.0411067454451478, -0.0148156281727591, 3472222.24620701, 0.0653556414936508, 38888888.8532054, 0, 128472222.415231, 0.118595345006177, 0, 0.0412488869995571, 10416666.6436398, -0.0199682489965158, 135416666.75228, 6944444.61094065, 0.0775144581528836, 0.109774349943803, 135416666.646306, 0, 194444444.54479, 0.207227674132769, 0, 0.115940609441508, 156250000.043563, -0.0842232919632867, 395833333.540392, 6944445.02122638, -0.0500596387365156, 694444.696663161, 20833333.3052669, 208333333.645857, 2083333.42040188, 409722222.42408, -0.147831472318547, 3472222.88129787, 236111111.21616, 20833333.2311274, 416666666.995881, 145833334.051124, 10416666.6008359, 41666667.208579, 159722222.533634, 0, -0.00038055230185861, 0, 0.00133241172644569, 0, -0.00564274302251077, -6.93051748773727e-05, 0, -0.00312521103627308, 0.0050732078986584, 0, -6944444.47582044, -0.000779591286011863, 0, -0.011859002894872, 0.00438925778074413, -0.0049978286159054, -0.018374827613727, -6944444.43261759, 0, -34722222.2887781, -0.0285595683087671, 0, -0.0111432199634141, 0.0102221318268198, 0.00556782057531721, -34722222.2427661, -0.0381925369614084, -0.0324797095867241, -0.0390065897451783, -34722222.2039553, 0, -69444444.4999988, -0.0773635155664552, 0, -0.0228954502263574, -34722222.2216288, 0.0233407350046242, -138888888.956325, -0.189701309487973, 0.00187690142906958, -0.057943453363103, 0.0117884486576775, -69444444.5361012, -0.0328766503041162, -69444444.4492209, 0, -0.0698672804676723, -138888888.94668, 0.0646092827162059, -208333333.439701, -0.335874890752762, 0.0603501849150302, -0.216202051362016, -69444444.4526887, 0.050418878950772, -208333333.476326, -69444444.7500454, 0.0488312691354147, -6944444.63255194, -208333333.44109, 0.092465323851462, -34722222.6575786, -69444444.4873782, -138888889.14003, -34722222.3989619}, 
                                                     {0, 0, -3.23476479478808e-07, 0, 0, 8.90731982928032e-06, 0, 2333.3333389084, 3.08075912042048e-06, 0, 0, -0.000100356067627193, 0, -46175.9259275828, 0, -46175.9260952253, -3.81216878310251e-05, 0, -7.67369831114357e-05, -11175.925940456, 0, 0.000604340556229314, 0, 304259.258958174, 0, 304259.26099, 0.000192348538826572, 0, 0.000575173982627708, 900185.185894463, 0, 0.000996491349702528, 207546.296194578, 207546.296647618, 19953.7039633032, 0, -0.00216099218783938, 0, -873842.590837953, 0, -873842.601447309, -0.000314574094123623, 0, -0.00222430874366197, -5768518.51770436, 0, -5768518.52648026, -0.0061464139594684, 0, -0.0059851392532257, -1204629.62842208, -1204629.63155854, -15740.7417801303, -3806250.00055169, -346527.776859434, -346527.783963515, 0, 0.00480345923554459, 0, 1238425.9210402, 0, 1238425.95145909, -0.000665290898816677, 0, 0.00534653310331446, 15829861.1015721, 0, 15829861.1497513, 0.0193942720700008, 0, 0.016113078871289, 2722222.21698019, 2722222.22517155, 4629.63204502813, 35016203.7093365, 0, 0.0321005611853619, 21537037.0340987, 1733796.28805971, 21537037.0385986, 1733796.33680203, 254629.621012166, 254629.652264843, 5961805.57254034, 0, -0.00669765175412943, 0, -856481.473550685, 0, -856481.524787013, 0.00319128338981525, 0, -0.00816665737106375, -21215277.7551365, 0, -21215277.87464, -0.030993887541479, 0, -0.0258584168187524, -2638888.87686535, -2638888.88597084, -0.00631966380329447, -87303240.7163762, 0, -87303240.784255, -0.0938347332006453, 0, -0.0630756166953293, -46388888.8711754, -2777777.7548337, -46388888.8793164, -2777777.88298016, -69444.4160759838, -69444.4694820699, -115439814.805987, -1064814.76945403, -1064814.96773422, -29201388.9119791, -29201388.9831429, -4097222.26745005, 0, 0.0056940670725836, 0, 231481.473837316, 0, 231481.524344017, -0.00467138327988769, 0, 0.00752863167962026, 13773148.1215294, 0, 13773148.278012, 0.0233204108367977, 0, 0.0272880788966909, 925925.910585487, 925925.913373747, 0.0116441680439381, 102893518.465341, 0, 102893518.626005, 0.136510648546298, 0, 0.0590282843421965, 42361111.0746649, 1388888.85744008, 42361111.0752139, 1388889.01103214, -0.0368997069735857, 0.00361747638597539, 178703703.6852, 0, 0.147951014538845, 222222222.185637, 925925.840920802, 222222222.191467, 925926.291636461, 231481.378101554, 231481.666092745, 44444444.4506019, 44444444.5889246, 1041666.66820683, 135995370.484136, 16898148.190474, 16898148.4637591, 0, -0.00269636132322021, 0, 0.00404264279462743, 0, -0.0229272919871962, 0.00302352448762153, 0, -0.00375773200766668, -3472222.20589102, 0, -3472222.31059486, -0.00637761346999754, 0, -0.0171628090428947, 0.0103169057803302, 0.0121383669996394, -0.0102679000917791, -56712962.921696, 0, -56712963.0747032, -0.0925798037481059, 0, -0.0288756743486018, -13888888.8562074, 0.0221365499808938, -13888888.8484512, -0.0602637136493524, 0.0180277422980647, 0.0031362275606877, -152777777.722678, 0, -152777777.795728, -0.158824221847398, 0, -0.0923965832996542, -173611111.059054, 0.0716237343684678, -173611111.064532, -0.352466820531595, 0.0953533452267958, -0.066898006338778, -20833333.319418, -20833333.4021418, 0.0626988193029498, -319444444.404385, 0.125759547725984, -0.369228536129269, -180555555.606526, -13888888.8955451, -180555555.645274, -13888889.3357593, -3472222.17752007, -3472222.50648089, -67129629.9435886, 0, 0.000544906076607411, 0, -0.000900711416002798, 0, 0.00512511869876118, -0.000740588451443691, 0, 0.00077228095853772, -0.00417757493731726, 0, 0.0239950154125659, -0.000282760771724919, 0, 0.00467818261943106, -0.00284555280021859, -0.00385854832819278, 0.0032875279349463, 11574074.0622672, 0, 11574074.1158119, 0.0230941281725062, 0, 0.00670418235632949, -0.0109422233618271, -0.00640479706836569, -0.015091809575626, 0.00858056994334087, -0.00174589293394849, 0.00123649751291948, 46296296.2727414, 0, 46296296.3171814, 0.0597941534745942, 0, 0.0196308420342398, 46296296.2723593, -0.0226569085049111, 46296296.27293, 0.116784605841063, -0.0294121871584551, -0.00219592143868633, -0.00890185221389858, 0.000935447414129568, -0.0331819305621837, 69444444.4279185, 0, 0.0540475763046537, 138888888.86494, -0.0523889423211727, 138888888.870522, 0.205307144390641, -0.058124340311282, 0.101980479659601, 69444444.4428457, -0.00655839059237799, 69444444.4462299, 0.179248781287983, -0.0624147248186101, 0.0575727455043867, 138888888.903029, -0.0349966355195401, 0.2613521927517, 46296296.4020315, 46296296.4972962, 11574074.2301875}, 
                                                     {0, 0, 2.171357855604e-07, 0, 0, -5.91882180339676e-06, 0, -1000.00000220662, -2.17503766030668e-08, 0, 0, 6.54737785721037e-05, 0, 21694.4444363005, 0, 18361.111182625, -1.64309682398269e-06, 0, 1.02069624175999e-05, 3361.11109822251, 0, -0.000387309491355934, 0, -165370.370185316, 0, -104166.667420361, 5.16917922334188e-05, 0, -5.99255935664802e-05, -394537.037194916, 0, -0.000216617905540175, -69120.3702198658, -57916.6665115971, -3749.99989948672, 0, 0.00136455869202471, 0, 572916.665759138, 0, 225694.44837691, -0.000414376634544226, 0, 0.000114172130977591, 2958333.33211908, 0, 2194444.44694583, 0.00157314625778905, 0, 0.00124377789095984, 477777.776896085, 284722.220955211, 1388.88787812978, 1182638.88726118, 72916.6658002578, 60416.6670901469, 0, -0.00299854303386848, 0, -960648.145832229, 0, -208333.344867453, 0.00165194475838585, 0, -4.44440195015998e-05, -9980324.06777568, 0, -4575231.4950657, -0.00485271220913738, 0, -0.00229242661570316, -1386574.070935, -437499.993900325, 0.00469883480971999, -15886574.0718069, 0, -0.00871443854018433, -8069444.43737905, -451388.885555859, -5719907.3968749, -250000.005578985, -25462.9570755491, -20833.3332960515, -1170138.88876224, 0, 0.00414541352027262, 0, 763888.885353228, 0, 69444.4643287178, -0.00348243881099046, 0, -5.88909224792525e-05, 16076388.8760299, 0, 4027777.81402012, 0.00723693322280815, 0, 0.00151888914510691, 1666666.65989942, 208333.317842119, -0.0101776968695687, 50486111.0914494, 0, 30798611.1165007, 0.0288359502074987, 0, 0.0149725198145004, 22881944.4262395, 1041666.65860446, 8437499.96440052, 208333.346777676, -0.0212618325206456, -0.0147659669764833, 37847222.1894347, 138888.873536175, 69444.4756200243, 7187499.99997494, 4791666.66780931, 381944.445175164, 0, -0.0035024943409199, 0, -231481.478235947, 0, -0.0199866422725595, 0.0040133383169115, 0, 0.000181637857301703, -12152777.7637051, 0, -1273148.19896891, -0.00375298552361094, 0, -0.000219121831548522, -694444.436089528, 0.0213953588995082, 0.0112878137191173, -73726851.8198763, 0, -24652777.803789, -0.0433772472792475, 0, -0.00902962096735044, -26388888.8612171, -694444.432033517, -3819444.38380883, -0.00620947272960212, 0.0343626788625582, 0.0391216619099549, -85185185.157702, 0, -0.0429915581007408, -101273148.094068, -231481.457832913, -51388888.8121149, -0.092468365546185, 0.0417271252526931, -0.0202983678672354, -16319444.441845, -3819444.41532628, 0.0364938515664173, -28819444.4349675, -2083333.34844939, -1273148.21161105, 0, 0.00165131714677987, 0, -0.00164803068803935, 0, 0.0108415542441565, -0.00235590809543823, 0, -0.000149202788395801, 3472222.21399931, 0, 0.0358667974393359, -0.000739055863592772, 0, -8.69217196135354e-05, -0.00540475860026802, -0.0149332839687426, -0.00630236395493861, 48611111.0885411, 0, 6944444.47752538, 0.0296233573616396, 0, 0.000196314815113986, 10416666.6447625, -0.0103618226373631, -0.0493944650114158, -0.00824682775354577, -0.025176570370534, -0.0334251466549416, 93749999.9642637, 0, 55555555.5425255, 0.0500626230711493, 0, 0.0233180245840738, 104166666.618149, -0.0194023975779266, 20833333.2522472, 0.0939148577705293, -0.0447225015941592, -0.0238263284487406, 10416666.6582183, -0.0616936292472992, -0.0680089718346396, 114583333.2595, -0.0384128623559126, 0.0906128925393768, 62499999.9881516, 3472222.24124886, 20833333.2684632, 0.0928426937920219, -0.0257904025680763, 0.0462232167063256, 6944444.53624568, 0, -0.000332713028344224, 0, 0.000354820112067513, 0, -0.00245339970208606, 0.000550212241396931, 0, 4.27281120966967e-05, 0.00201774572566187, 0, -0.0100425126655471, 0.000920659645778532, 0, 1.66900814820013e-05, 0.0014315874671658, 0.00402242489648529, 0.00142251475844526, -11574074.0678835, 0, -0.0137666326998395, -0.00726302942031972, 0, 0.0006842447511132, 0.00684067690563339, 0.00347667847096474, 0.0154286686732497, 0.0063357506743983, 0.00692633438564244, 0.00997110259897434, -34722222.2092976, 0, -11574074.074759, -0.0197977182434372, 0, -0.00312952780235783, -34722222.2041948, 0.00674145331231507, 0.0316005270297889, -0.0291666806675616, 0.0165530026986313, 0.0199353924357862, 0.00579529743452818, 0.0335186767563219, 0.0312366832238838, -34722222.2086876, 0, -0.0158030424707986, -69444444.4155791, 0.0140419050441364, -34722222.1864421, -0.0588797058691875, 0.0202035854573385, -0.016003788725599, -34722222.213644, -0.00645771759867121, 0.0550277489328036, -0.0274923946565493, 0.0340013837774732, 0.00967313110311183, -34722222.1918538, 0.00421590209126554, -0.0654197249393661, -11574074.1167791, -0.0424214235472441, -0.0458576752788849}, 
                                                     {0, 0, 9.96422508538847e-08, 0, 0, -2.79525425638519e-06, 0, -1000.00000304382, -2.34928238565979e-06, 0, 0, 3.25816568639375e-05, 0, 18361.1111190431, 0, 21694.4445338566, 3.05323055392187e-05, 0, 4.91025365380299e-05, 3361.11112983547, 0, -0.000202547634713132, 0, -104166.666549936, 0, -165370.371267574, -0.000196323548931588, 0, -0.000389589087898496, -394537.037536358, 0, -0.000541318975492278, -57916.6666563282, -69120.3706740155, -3750.00030030131, 0, 0.000742926343826987, 0, 225694.443623449, 0, 572916.671198337, 0.000649498770506284, 0, 0.00165472062387581, 2194444.44473058, 0, 2958333.33834634, 0.00307334818538651, 0, 0.00332537022622231, 284722.221740355, 477777.779612725, 1388.89109842148, 1182638.8897191, 60416.6664180325, 72916.6703702891, 0, -0.00168320479198314, 0, -208333.330868588, 0, -960648.161047654, -0.00114072384612302, 0, -0.00427715238726487, -4575231.47823544, 0, -9980324.09728593, -0.00979808177845519, 0, -0.0102006981608785, -437499.997793521, -1386574.07931836, -0.00856966752139819, -15886574.0809785, 0, -0.0154034900608594, -5719907.40842597, -249999.994269488, -8069444.44885034, -451388.908757691, -20833.3336252276, -25462.9814829633, -1170138.89462119, 0, 0.00237980274242948, 0, 69444.4402449902, 0, 763888.910458451, 0.00105503537755481, 0, 0.00674549106816196, 4027777.7683994, 0, 16076388.945369, 0.0163129064862304, 0, 0.0189664007210215, 208333.328434075, 1666666.6742766, 0.0193783837432277, 30798611.1055632, 0, 50486111.1457183, 0.04134838301131, 0, 0.0327212798555613, 8437499.99841345, 208333.315765161, 22881944.4522367, 1041666.71669425, 0.00667737179687611, 0.0439789264050253, 37847222.2319359, 69444.4230514396, 138888.967854878, 4791666.66038209, 7187500.02887341, 381944.474135393, 0, -0.00204338267220114, 0, 0.00420097090098209, 0, -231481.502514548, -0.000448536839384041, 0, -0.00641308533713265, -1273148.13633936, 0, -12152777.8516579, -0.0141648425818842, 0, -0.0217991457081223, 0.00606644032798157, -694444.450534438, -0.0250556925786158, -24652777.7567441, 0, -73726851.9271742, -0.058562357357654, 0, -0.0363581541291767, -3819444.43804644, 0.0236713329983848, -26388888.8930705, -694444.50768621, -0.0188198263332599, -0.0613110072676222, -85185185.1913128, 0, -0.0677916462840473, -51388888.8986408, 0.0514641416418295, -101273148.167705, -231481.642934085, 0.0303900338750261, -0.128181290629823, -3819444.41371713, -16319444.488916, -0.0617809945577553, -28819444.4563596, -1273148.14061612, -2083333.45787144, 0, 0.000974369229133878, 0, -0.00229047818807584, 0, 0.0110859762195885, 3.9607519490302e-05, 0, 0.0032950012546923, -0.00755078892133175, 0, 3472222.27136783, 0.00587768935550895, 0, 0.0140603817195585, -0.00406447893518053, 0.00259399229058421, 0.0169283755605741, 6944444.42643678, 0, 48611111.1845433, 0.0394797684961903, 0, 0.0227279474547488, -0.00654507095453112, -0.0155647670752391, 10416666.6648973, 0.0385587165869134, 0.0209160230425049, 0.0468134020454981, 55555555.5355463, 0, 93750000.0272751, 0.0683282589451004, 0, 0.0463293840304889, 20833333.3339155, -0.0478902432592358, 104166666.680127, 0.147055664653605, -0.0169048679128131, 0.0932855089861636, -0.0328045920562754, 10416666.6880476, 0.058892570692854, 114583333.355564, -0.0585335878331574, 0.183155438600478, 20833333.3116993, -0.0519763035224646, 62500000.0117271, 3472222.36878189, 0.00299275715408153, 0.167597596556892, 6944444.51081337, 0, -0.00019784634398054, 0, 0.00052352896695679, 0, -0.00244383917468908, 2.04498106293805e-05, 0, -0.000698401762546695, 0.00200048586764344, 0, -0.0130615028784029, -0.000817688183497305, 0, -0.00383685139174586, 0.00112803505110036, -0.00038882881289701, -0.00458158074066406, 0.00531182543162959, 0, -11574074.1003305, -0.00999262839213204, 0, -0.00620582174834691, 0.00235552876804858, 0.00407885670447598, 0.00168132750408003, -0.00898607743915355, -0.00823780809309, -0.0156330374688612, -11574074.0636066, 0, -34722222.2407395, -0.0246062810520186, 0, -0.0119340770421893, 0.00189239180117608, 0.0156732748188115, -34722222.2245931, -0.0498231809044465, 0.00131031161103866, -0.0269606906246026, 0.0108354735782173, -0.00167742979691811, -0.022501753202165, -34722222.2185873, 0, -0.0245747784047151, -34722222.2281961, 0.028819674805623, -69444444.4565211, -0.0895964620711859, 0.0215995010701495, -0.0650516076369474, 0.016448909743672, 0.0301018787899464, -34722222.2190502, -0.0537223552488585, -0.00320862516616597, -0.0699154619424375, -34722222.2155936, 0.0272590176901371, -0.109273919095513, 0.00276561755003877, -11574074.1167726, -0.0619466346428274}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -1.22306186598487e-06, 4499.99999968253, 0, 0, 0, 0, -64303.5714216891, 0, 2.87821449745822e-05, -64303.5714250529, 2.51978705250565e-05, -64303.5715001146, 0, 0, 0, 0, 391991.071375452, 0, -0.00016434387290553, 391991.071406166, -9.16074585791964e-05, 0, 783982.142788497, -0.000646044317137376, 783982.143850328, 391991.072304694, 783982.143584076, 0, 0, 0, 0, -1333402.7775571, 0, 0.000276416363328377, -1333402.77772028, 0.000157808078797899, 0, -4000208.33296663, 0, 0.0045122613357635, -4000208.33286026, 0.00262787873150147, -4000208.33832119, -1333402.78260135, -4000208.33569868, -4000208.34329216, -8000416.67687906, -4000208.34056555, 0, 0, 0, 0, 2772048.61050776, 0, 0.000706769615076794, 2772048.61109156, -0.000298149149018955, 0, 11088194.4437384, 0, -0.015991783255711, 11088194.4429301, -0.00554779491554449, 11088194.4573393, 2772048.62502641, 11088194.4475244, 0, 16632291.6642214, -0.0149362182756542, 16632291.7057876, 33264583.3799802, 11088194.4851448, 11088194.4808637, 33264583.3673514, 16632291.6824404, 33264583.4198753, 0, 0, 0, 0, -3611111.11003195, 0, -0.00354181423837088, -3611111.11125144, 0.000561759186027672, 0, -18055555.5548395, 0, 0.029562155908271, -18055555.5534944, 0.00971572146708123, -18055555.5743233, -3611111.13494673, -18055555.5566922, 0, -36111111.1072642, 0, 0.0474012649618536, -36111111.102956, 0.0177958335309798, -36111111.1864406, -72222222.3273243, -18055555.6375367, -18055555.6415662, -72222222.2756078, -36111111.1071042, -36111111.2324224, -108333333.466786, -36111111.2003244, -108333333.631026, -72222222.5260409, -108333333.556398, 0, 0, 0, 0, 2881944.44328052, 0, 0.00560020402540348, 2881944.44467349, -0.000547085448926437, 0, 17291666.6658942, 0, -0.0285017109220815, 17291666.6657824, -0.0136002128278815, 17291666.6820598, 2881944.46943629, 17291666.6661633, 0, 43229166.6652654, 0, -0.0758532627631639, 43229166.6548858, -0.00993163029899565, 43229166.7430391, 86458333.4598036, 17291666.7541963, 17291666.7718258, 86458333.3809622, 43229166.6131132, 0, 57638888.8762281, -0.0447981963061954, 57638889.0595719, 172916666.905244, 43229166.8402899, 43229166.8499846, 172916666.832423, 57638888.9642168, 172916667.130097, 86458333.829671, 172916666.882218, 172916667.389382, 259375000.645945, 172916667.308988, 0, 0, 0, 0, -1289682.53900858, 0, -0.00400616669798993, -1289682.53981363, 0.000195837862813768, 0, -9027777.77699562, 0, 0.0131958433844457, -9027777.77813704, 0.0107483811195246, -9027777.78431132, -1289682.55460577, -9027777.77782744, 0, -27083333.3342655, 0, 0.0572284561750928, -27083333.3260704, 0.00552572922286427, -27083333.3721002, -54166666.744249, -9027777.82581393, -9027777.84426211, -54166666.6932107, -27083333.2695691, 0, -45138888.8824795, 0, 0.0553417411697876, -45138888.8739104, 0.0117284409721835, -45138889.0030865, -135416666.868378, -27083333.4485669, -27083333.4898962, -135416666.765826, -45138888.8844298, -135416667.006065, -54166667.049017, -135416666.726061, -45138889.0489563, -180555555.760479, -45138889.0434336, -180555556.261576, -270833334.056827, -135416667.403992, -135416667.425233, -270833333.843046, -180555556.037757, -270833334.434837, 0, 0, 0, 0, 248015.872856055, 0, 0.0010997166824766, 248015.873040629, 1.64621168915224e-06, 0, 1984126.98380412, 0, -0.00215522071219821, 1984126.98440442, -0.0033299699619838, 1984126.98521219, 248015.876901861, 1984126.98444522, 0, 6944444.44494542, 0, -0.016283407711604, 6944444.44296326, -0.0028302610607471, 6944444.45209848, 13888888.9079392, 1984126.99492552, 1984127.0016672, 13888888.8964669, 6944444.42066335, 0, 13888888.8883423, 0, -0.0235135861676835, 13888888.8832725, 0.000992522381813619, 13888888.9180102, 41666666.7317707, 6944444.4730506, 6944444.49286778, 41666666.6921205, 13888888.8652957, 41666666.7616342, 13888889.0021976, 41666666.6564545, 0, 17361111.1061171, -0.0100211829630077, 17361111.1651122, 69444444.5343999, 13888888.9407856, 13888888.964395, 69444444.5054704, 17361111.1508258, 69444444.6895167, 104166666.948884, 41666666.9348292, 41666666.9760404, 104166666.793348, 69444444.5457408, 69444444.7880342, 138888889.201178, 69444444.7668848, 138888889.460096, 104166667.260005, 138888889.349783}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 0, -125.000000007181, 0, 0, 0, 0, 3366.07142861151, 0, -9.47590889432071e-08, 125.000000359882, -1.22315471852834e-07, 124.999999974672, 0, 0, 0, 0, -35810.515872924, 0, 8.64097737100629e-07, -2.35645955317549e-06, 1.13999219948666e-06, 0, -3241.07143415243, 1.55049207093823e-06, -3241.07142877387, 4.93554865934902e-07, 5.21519841807442e-07, 0, 0, 0, 0, 200451.388885383, 0, -4.41173717604074e-06, 8.00573949989321e-06, -4.57376110153501e-06, 0, 32569.4444925118, 0, -8.14910785327141e-06, 2.3597353759307e-05, -8.68303439160706e-06, 32569.4444477368, -2.43074592913803e-07, -5.21964302466138e-06, 2.79057465661472e-06, 1.46956868744305e-06, 9.25433922454549e-07, 0, 0, 0, 0, -653993.05553308, 0, 6.44547268866162e-06, -2.19151171608668e-05, 1.40512786821836e-05, 0, -167881.944643556, 0, 4.76766581173391e-05, -2.22787179036036e-05, 2.22841283226754e-05, -167881.944453339, 6.60982619138108e-06, 3.50334343781322e-05, 0, -0.000211813807764875, 5.40151159311375e-05, -2.16842235944435e-05, -8.95512503257405e-06, -2.9745492132457e-06, -8.17017638093486e-06, -1.40995690784389e-05, 2.02473583670506e-05, -8.29189765411516e-05, 0, 0, 0, 0, 1284722.22216241, 0, -1.17456871953915e-05, 4.29251228759282e-05, -1.00330584972834e-05, 0, 486111.111489539, 0, -4.93017795456328e-05, -1.70884189329025e-05, -5.31920869559981e-05, 486111.111127694, 9.35805128828494e-06, -7.96641633055132e-05, 0, 0.000879007206363532, 0, -3.09586689740221e-05, 0.000207447418825572, -9.89014924365324e-05, 1.41393182423693e-05, 8.82987186153408e-05, -2.97987083273058e-05, 5.55227820607813e-05, -4.33326688136124e-05, -2.86629336885107e-05, -2.9825030215099e-05, 4.94325084811958e-05, -3.56573867485477e-05, 0.000191377060502493, 0.000242989255415535, 8.65519483785836e-05, 0, 0, 0, 0, -1493055.55547828, 0, 6.20750173041471e-06, -5.11969640705984e-05, 1.72404484314117e-05, 0, -798611.111444574, 0, 6.86965349557898e-05, 2.70622251487459e-05, 3.81542930030343e-05, -798611.111124183, -3.98124497579878e-06, 8.34371185538072e-05, 0, -0.00158479856151842, 0, 0.000201980159452948, 0.000118375673342871, 0.000157644955573537, -2.52016761365315e-05, -0.000132384934849508, 2.64346089838485e-06, 6.2372989979392e-05, 0.000137166906009364, 0.000144497552862242, 0, -0.00108543511365755, 0.000297585209686004, 0.00011254286382562, 5.50287622746727e-05, 0.000190697279336886, -9.13797761923477e-05, 0, -0.000183979116726657, -0.000142248615087778, -0.000294202309395683, -0.000391891671520234, -0.000729129944088853, -0.000583126986259873, -0.000270093725024558, 0, 0, 0, 0, 942460.317412108, 0, -4.18697413359944e-06, 3.09614622334554e-05, -7.63825397928985e-06, 0, 694444.444560866, 0, -3.69596960078174e-05, 1.40667950328215e-05, -3.35987661956368e-05, 694444.444452105, 9.61523353944317e-06, -6.83389634794632e-05, 0, 0.00121473836017118, 0, -0.000151843747486553, -0.000252953963346218, -0.000110678986059626, 2.62510299394603e-06, 6.74648461145448e-05, -1.57372683663234e-05, -5.14742382832657e-05, -0.000165882730063717, -0.000122367840520493, 0, 0.00163413231766841, 0, -0.000272511706400162, 0.000316109418519353, -0.000175310459469418, -5.85371064454888e-05, 8.20685766750225e-05, -0.000124079737878218, -7.85979602155973e-05, -9.39978354534891e-05, 2.81359488698712e-05, 0.000403251152910992, 0.000223677117637886, 5.14952628390668e-05, -0.000232828960724855, 2.5069842832022e-05, 0.000147709874015217, 0.000277781705941683, 0.000550865585787899, 0.000424247723628693, 0.000769586279720964, 0.000777455549293137, 0.000740979018449653, 0.000615380055354394, 0, 0, 0, 0, -248015.87300421, 0, 9.86370689565851e-07, -7.7123411670684e-06, 1.93021528405128e-06, 0, -248015.873021158, 0, 2.0537301419031e-06, -1.23055514884016e-05, 7.02636038370728e-06, -248015.873018241, -2.48648581459754e-06, 1.76494600574278e-05, 0, -0.000332664931520042, 0, 1.95058572045976e-05, 8.32547110104083e-05, 6.19782813390477e-07, -8.90116833935502e-06, -1.92390029872607e-05, 5.45767713191428e-06, 7.78439206355882e-06, 4.05284166612489e-05, 3.50592448539228e-05, 0, -0.000731091698753012, 0, 5.19060150453263e-05, 8.15644967080808e-05, 2.33358565213782e-05, 4.18711994513984e-05, -3.60621612471148e-05, 1.79475441287759e-05, 8.73641600658465e-06, 5.43608865184518e-05, -1.37378400835063e-05, -8.3238946469796e-05, -4.66317689184846e-05, 5.79613291774396e-05, 0, -0.000447269473814452, 8.95640860945749e-05, 0.000104240941679015, 2.34173466581347e-05, 0.000122511533105939, 5.19990148884928e-05, 2.55344828919471e-05, -2.9896581892515e-05, -1.32334837942361e-05, -5.73906738828752e-05, -0.00011284344486941, -0.000123742303438162, -1.05018993309234e-05, -0.000344257658528404, -0.000141018966901413, -0.000171974044687918, -0.000493521991603892, -0.00019266821015043, -0.000711010578762053, -0.000769742341360566}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 1.27433184435629e-08, -125.000000010771, 0, 0, 0, 0, 125.000000085598, 0, -1.7879495919821e-07, 3366.07142884925, -2.64703972437898e-07, 124.999999906848, 0, 0, 0, 0, 6.8863480846884e-08, 0, 8.93904209429251e-07, -35810.5158754841, 2.21436022098579e-06, 0, -3241.07143127679, 3.29739952802501e-06, 1.31754291426099e-06, -1.47970158457881e-07, -3241.0714268478, 0, 0, 0, 0, -2.15853660689474e-06, 0, -3.73905514164096e-07, 200451.388899496, -9.0383373309853e-06, 0, 6.12935257955702e-06, 0, -2.05924365298021e-05, 32569.4444686974, -2.51304389991985e-05, -7.18785919455585e-06, -1.28137043007669e-06, 32569.4444329516, -4.23417451823389e-06, -9.4413748146821e-06, -1.95128899685736e-05, 0, 0, 0, 0, 7.61248816967638e-06, 0, 1.90403606465922e-06, -653993.055580694, 1.7657054648993e-05, 0, 5.41262837430991e-06, 0, 4.54967115013671e-05, -167881.94454069, 9.60798513323795e-05, 2.44164718869123e-05, -1.13367208518222e-05, -167881.944407613, 0, -7.13743302953581e-05, 0.00011240016727702, 1.88632717126566e-05, 4.24481757327495e-05, -2.28977263391245e-05, 5.45771543162758e-05, 5.81062884310693e-05, 0.000115323839242701, 7.04365608551416e-05, 0, 0, 0, 0, -1.31328726456573e-05, 0, 6.3205669510333e-06, 1284722.22225676, -3.11861055245456e-05, 0, -4.05150855699861e-05, 0, -7.92324358140343e-05, 486111.111308895, -0.000184447219181304, -4.85024120146131e-05, 1.43527073274925e-05, 486111.111048679, 0, 6.63442470480926e-05, 0, -0.000288292430461511, 0.000259316620061449, -0.000269003832850462, -6.66460569516926e-05, -0.000116380290724268, 5.65551055704465e-05, -0.000117127468890809, -0.000149815366567764, -0.000400993580096814, 2.35518943017154e-05, -0.000184524731821984, -0.00029967378085106, -7.58951654833045e-05, -0.000140465762133433, -0.00030844171312839, 0, 0, 0, 0, 1.24649869624192e-05, 0, -2.19767714050238e-06, -1493055.55558323, 2.20189044092194e-05, 0, 6.42694216676046e-05, 0, 6.34347953670027e-05, -798611.111328887, 0.000188117500073076, 5.67890768167688e-05, -2.69692547051746e-05, -798611.111053983, 0, 1.04458260048907e-05, 0, 0.000301492064285357, -0.000438102049846128, 0.000392758392999992, 0.000139375227626857, 0.000140315663521042, -7.370890446261e-05, 3.52587515690857e-05, 0.0001850032658147, 0.000512190042900143, 0, -0.000264774248781442, 0.000450838984979046, -1.5208874742347e-05, 0.000166738690237607, -0.00018078153849062, 0.000424200766012114, 0.000355101408249389, 0.00101912708137116, 0.00021590682151328, -0.000140404568525956, 0.00125581824254186, -0.000121096173156256, 0.000443193634580626, 0.000208466505469567, 0, 0, 0, 0, -6.24150998749259e-06, 0, 3.96850790495462e-06, 942460.317472162, -9.71804515554949e-06, 0, -4.50206166593814e-05, 0, -2.67855466128365e-05, 694444.444567696, -9.39140488710011e-05, -3.57963778057952e-05, 1.9873768685847e-05, 694444.444429158, 0, -5.25266225831099e-05, 0, -0.000177185766530372, 0.000345145368404404, -0.000249583511892689, -0.00012531566360796, -0.000111173205471199, 5.49817932004585e-05, -2.89291979674425e-05, -8.14822256241856e-05, -0.000443399373624716, 0, 9.65167471327885e-05, 0, -0.000349334978631452, 0.000343454667278542, -0.000373256304201321, -6.46254008741428e-05, -0.000164083681420619, 0.000124288071294229, -0.000150593370080407, -0.00033809624649331, -0.000769811695141316, -0.000469701293864769, 0.000223713114661964, -0.000789297273377218, 0.000155568957288593, -0.000320552189000541, -0.0007607224493485, 7.13586657892073e-05, -0.000381018841290992, 0.000313391538156829, -7.55349431366512e-05, -0.00125388842264434, -0.0011018686601134, 0.000521886956579787, 0, 0, 0, 0, 1.28642665591443e-06, 0, -1.02788914421673e-06, -248015.873018053, 2.10996259327651e-06, 0, 1.2231272593904e-05, 0, 8.05401652392105e-06, -248015.873043298, 2.14806148688263e-05, 9.00256990594722e-06, -5.96381504641398e-06, -248015.873014385, 0, 2.32026506075441e-05, 0, 5.05827563803385e-05, -0.000104098422228272, 8.84987555017136e-05, 4.02449092292132e-05, 3.40180789894997e-05, -6.37314766571609e-06, 1.45825575677687e-05, 1.47152472156045e-05, 0.000127941485763103, 0, -4.82796220413853e-07, 0, 0.000142566159285733, -0.000145288897657491, 0.000138899560619779, 4.12591987182877e-05, 5.68894267512659e-05, -6.13788913115405e-06, 0.000124513389867598, 7.71985289821318e-05, 0.00031600521803969, 7.69880719549954e-05, -7.26401135337055e-05, 0.00030219024644168, 0, -8.58296792325419e-05, 0.000160432388373094, -3.81260276725561e-05, 7.1760065823127e-05, -6.27672983528446e-05, 0.000252001026956313, 0.000147908486393871, 0.000257848267166566, -8.47463807695072e-05, 3.09130565800256e-05, -0.000223207832750857, -0.000138983127277941, 0.000318163364815029, 0.000515085173752128, -0.000194439582105451, 0.000254393561151189, 0.000378390469714006, 0.000105799495088061, 2.17311999659662e-05, 0.000458562573207348}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 2.04853845629097e-06, -5999.99999938642, 0, 0, 0, 0, 125738.095225701, 0, -4.36457539546247e-05, 65738.0952309756, -2.46603849170113e-05, 65738.0953262321, 0, 0, 0, 0, -960908.73006605, 0, 0.000379312539834358, -303527.777724748, 9.66804050969568e-05, 0, -1264436.50782045, 0.000431153202550154, -1264436.50946499, -303527.778738479, -607055.556360769, 0, 0, 0, 0, 3801388.88850267, 0, -0.00158675330124281, 766111.110935757, -0.000466745231154346, 0, 8368888.88839338, 0, -0.00380673785404583, 5333611.11020157, -0.000323367099654529, 8368888.89711764, 766111.115501863, 2298333.33471483, 5333611.12519211, 10667222.2379594, 2298333.34116749, 0, 0, 0, 0, -8803472.22111913, 0, 0.00338783520449935, -1142361.11087751, 0.00198108682624604, 0, -27552777.7774094, 0, 0.0169619691037655, -12230555.5523635, -0.00183767605846748, -27552777.7995512, -1142361.12086387, -4569444.44058692, 0, -29837499.9959038, 0.00336990456194942, -29837500.0609025, -59675000.0780888, -12230555.606851, -4569444.48131474, -36691666.7171222, -6854166.67569109, -36691666.7919195, 0, 0, 0, 0, 12430555.5534732, 0, -0.00368270706914609, 1006944.4443829, -0.00453051643433469, 0, 50729166.6674458, 0, -0.0372361217039179, 16458333.3285065, 0.000884024630855755, 50729166.6982981, 1006944.45622209, 5034722.20541977, 0, 78611111.1068096, 0, -0.0257565787605206, 55763888.8733451, 0.0180246239019611, 78611111.2356128, 157222222.401083, 16458333.4219367, 5034722.30079254, 65833333.405198, 10069444.4014741, 55763889.0588319, 167291666.893304, 10069444.517241, 167291667.146361, 65833333.7454558, 98750000.3044797, 0, 0, 0, 0, -10555555.55321, 0, 0.00171657878140731, -486111.111230146, 0.00534633458211569, 0, -53263888.8897514, 0, 0.0425103638924921, -12986111.1082821, 0.00710349797751817, -53263888.9143144, -486111.119226365, -2916666.64281642, 0, -107986111.114477, 0, 0.0575716984801489, -57638888.8650273, -0.0377911473207465, -107986111.243214, -215972222.434515, -12986111.1875541, -2916666.74889726, -64930555.6066629, -7291666.53864369, 0, -110416666.645428, -0.014527881474984, -110416666.922032, -331250000.416836, -57638889.0974222, -7291666.82604575, -230555555.842351, -9722222.22291131, -331250000.78879, -64930556.1818556, -129861111.359806, -230555556.667332, -345833334.355296, -129861111.937327, 0, 0, 0, 0, 4960317.45891654, 0, 4.47029007789297e-05, 99206.3492877799, -0.00314456601007086, 0, 29861111.1108321, 0, -0.0243396580413625, 5555555.55555994, -0.0101608562721139, 29861111.1215047, 99206.3524119108, 694444.429141399, 0, 75000000.0068416, 0, -0.05189673465215, 31249999.9845983, 0.0240917852556558, 75000000.0694066, 150000000.126009, 5555555.58607953, 694444.486156111, 33333333.3522588, 2083333.20771229, 0, 100694444.437392, 0, -0.0112993578003945, 76388888.8606753, 0.0445571034336086, 100694444.624642, 302083333.681909, 31250000.1096937, 2083333.46797542, 156250000.172106, 3472222.12022216, 302083333.929751, 33333333.7752523, 83333333.349252, 76388889.0918319, 305555555.930286, 3472222.31643067, 305555556.725633, 458333334.547405, 156250001.068967, 83333334.2668592, 312500000.785737, 111111111.641334, 312500001.666335, 0, 0, 0, 0, -992063.491723884, 0, -0.000211743752571622, -1.01822605299151e-05, 0.000723787698757697, 0, -6944444.44406052, 0, 0.0055053191565554, -992063.492426975, 0.00399502405674978, -6944444.44607298, -0.000650726227388382, 0.00378097782343375, 0, -20833333.3358182, 0, 0.0168713182792821, -6944444.44112345, -0.00408591087630009, -20833333.3473718, -41666666.6961872, -992063.496096396, -0.00820337717698713, -6944444.44784141, 0.043179258374634, 0, -34722222.2236843, 0, 0.0112071886573068, -20833333.3221934, -0.0194515317429545, -34722222.2703487, -104166666.775991, -6944444.46276597, -0.0391518782379775, -41666666.708883, 0.0643922504426365, -104166666.836457, -6944444.56250866, -20833333.2927723, 0, -34722222.2139143, -0.0120097276217212, -34722222.2937676, -138888889.052714, -20833333.3835134, -0.0579898210535901, -104166666.784756, 0.00139635719831071, -138888889.317407, -208333333.819791, -41666667.0260294, -20833333.695345, -104166666.855054, -34722222.2848537, -104166667.209024, -208333333.853771, -34722222.5750419, -208333334.267155, -104166667.5288, -138888889.553676}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -2.78751299868353e-07, 6999.99999944711, 0, 0, 0, 0, -164194.444432295, 0, 1.44974651641208e-05, -59194.444439859, -1.67881626416766e-05, -59194.444517587, 0, 0, 0, 0, 1444046.29620238, 0, -0.000250059149012137, 206129.629591298, 0.00018257575033805, 0, 1300175.92584551, 0.000283074055938276, 1300175.92725732, 206129.630442265, 412259.259905481, 0, 0, 0, 0, -6430138.88847769, 0, 0.00145420066842609, -378472.22208916, -0.000557659954660809, 0, -10279027.7775931, 0, -0.000153496047835048, -4227361.11032007, -0.00387797867642495, -10279027.7851891, -378472.225512361, -1135416.66750578, -4227361.12340851, -8454722.23565271, -1135416.67389455, 0, 0, 0, 0, 16370138.8876226, 0, -0.00411197966384273, 386574.073936123, 0.000331414448070162, 0, 39190509.2603249, 0, -0.00533468140848719, 7223379.62644302, 0.0148688035572791, 39190509.2791169, 386574.07988018, 1546296.29177175, 0, 29657175.9229209, 0.0150509931914069, 29657175.9814527, 59314351.9213373, 7223379.67197764, 1546296.32706852, 21670138.9300275, 2319444.45664909, 21670138.9989969, 0, 0, 0, 0, -24930555.5530104, 0, 0.00642501940178243, -208333.333430855, 0.00133222571427071, 0, -81006944.4482335, 0, 0.0177360965586234, -6840277.77263463, -0.0255119496605308, -81006944.4730854, -208333.337197614, -1041666.64959894, 0, -93645833.334429, 0, -0.0170754697919069, -44201388.8734764, -0.0594708585897764, -93645833.4514909, -187291666.827893, -6840277.84435905, -1041666.72616906, -27361111.1636872, -2083333.31570163, -44201389.0358881, -132604166.865101, -2083333.39679157, -132604167.103923, -27361111.4631303, -41041666.9348551, 0, 0, 0, 0, 22499999.9970213, 0, -0.00576379028541912, 46296.2965823562, -0.00273821536495917, 0, 93217592.5963804, 0, -0.02439982612988, 3402777.77464602, 0.020623466492099, 93217592.6148994, 46296.2948707835, 277777.75341653, 0, 147916666.681004, 0, -0.00094835963435443, 35648148.1218103, 0.091879286980069, 147916666.796875, 295833333.523519, 3402777.82764709, 277777.834332734, 17013888.9151359, 694444.374041071, 0, 109490740.724312, 0.079848470287575, 109490740.973012, 328472222.59743, 35648148.3101694, 694444.56020003, 142592592.835928, 925925.942186184, 328472222.960097, 17013889.3995477, 34027778.0086233, 142592593.596076, 213888889.80983, 34027778.4775206, 0, 0, 0, 0, -11111111.1092946, 0, 0.00276706097220539, -0.000168319146763403, 0.00202447327906412, 0, -56250000.0009031, 0, 0.0159375960889746, -694444.444481479, -0.0066771826970308, -56250000.0085112, 0.00332844521042001, 0.0163342072000295, 0, -114583333.349314, 0, 0.0128015372315647, -14583333.3153174, -0.0634676399846195, -114583333.404018, -229166666.776694, -694444.459401484, -0.0250419939812024, -4166666.66644528, 0.0713370464681414, 0, -118055555.558106, 0, -0.0432809710945427, -62499999.9711269, -0.0943478413402848, -118055555.728544, -354166666.981932, -14583333.4025219, -0.0910404383140642, -72916666.8023416, 0.0518790206337962, -354166667.23425, -4166667.00785593, -10416666.7070919, -62500000.163438, -250000000.335163, -0.0578421398786158, -250000001.099552, -375000001.115958, -72916667.6013582, -10416667.4167338, -145833334.041786, -13888889.3397052, -145833334.829084, 0, 0, 0, 0, 2314814.81436991, 0, -0.000549905374796713, 2.96749181117388e-05, -0.000532130582244265, 0, 13888888.8886141, 0, -0.00409500004812251, 0.000419977993414606, 0.000196583228732411, 13888888.8900421, -0.00127374186696699, -0.0043299859012585, 0, 34722222.2273775, 0, -0.0061399981518022, 2314814.81077512, 0.0160843307650973, 34722222.2369211, 69444444.4691097, 0.000483570182739754, 0.00368277138582719, -0.0031320872481713, -0.024766398880822, 0, 46296296.3026205, 0, 0.00754047665274276, 13888888.8762796, 0.0357332884665672, 46296296.345194, 138888888.98682, 2314814.82100599, 0.0259486239394248, 13888888.9177079, -0.0341049726907048, 138888889.051751, 0.085905489947039, -0.0212917670015693, 0, 34722222.2157955, 0.0305239465812383, 34722222.2832091, 138888889.03851, 13888888.9189228, 0.0321632356499866, 69444444.547659, -0.00553667915128694, 138888889.304991, 208333333.785572, 13888889.1887615, 0.276628546842895, 34722222.3969832, 0.0630198855245126, 69444444.9422681, 138888889.362137, 0.271113259778194, 138888889.755005, 34722222.9779302, 46296296.892253}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -1.07164302231508e-06, -6299.99999984651, 0, 0, 0, 0, 156174.999993261, 0, 1.11971865439655e-05, 40675.0000036321, 3.33807914634037e-05, 40675.0000386973, 0, 0, 0, 0, -1479874.99993883, 0, 2.16616760511305e-05, -104166.666692136, -0.000284114014693117, 0, -954041.666684524, -0.000480401492711451, -954041.667360606, -104166.66708692, -208333.333704923, 0, 0, 0, 0, 7159513.88858492, 0, -0.000415868091627744, 132291.666780213, 0.00109760124565447, 0, 8283819.4447995, 0, 0.00175122450911666, 2306597.22207173, 0.00410420923148145, 8283819.44845533, 132291.668119952, 396875.000834531, 2306597.22839259, 4613194.45145145, 396875.004340015, 0, 0, 0, 0, -19704513.8878358, 0, 0.00151643965960106, -83333.333759888, -0.00222598651450208, 0, -35221875.0026002, 0, -0.00202048611413539, -2758680.55425782, -0.0150501298736383, -35221875.0110483, -83333.3352393323, -333333.33283462, 0, -18192708.3335251, -0.0155444416194459, -18192708.3625256, -36385416.7039238, -2758680.57415279, -333333.348983686, -8276041.68793364, -500000.012261975, -8276041.72509774, 0, 0, 0, 0, 32138888.8866199, 0, -0.0027417844327878, 20833.334251038, 0.00246346994104659, 0, 80625000.0061471, 0, -0.000451269296436581, 1631944.44217941, 0.028249000958903, 80625000.0161889, 20833.3334550612, 104166.662021897, 0, 66423611.1201246, 0, 0.0240925529864619, 19548611.1034455, 0.0506839405384821, 66423611.176249, 132847222.310484, 1631944.46685923, 104166.692258233, 6527777.80439714, 208333.34789059, 19548611.1792343, 58645833.4376574, 208333.372417666, 58645833.5696721, 6527777.95277747, 9791666.81982564, 0, 0, 0, 0, -30763888.8861496, 0, 0.00279691071309292, -0.00104727417917088, -0.00144099615497444, 0, -101388888.894645, 0, 0.00310474750809795, -381944.443668718, -0.0282908768626878, -101388888.901476, 0.00229399022088477, 0.00772685425039032, 0, -119965277.802372, 0, -0.0180911794308869, -10243055.5381541, -0.0776799126595051, -119965277.85299, -239930555.660848, -381944.452913241, -0.0215671941248154, -1909722.2337551, -0.0078242739564401, 0, -59201388.8879083, -0.0662866492271669, -59201389.0041308, -177604166.86893, -10243055.6112888, -0.0484360349421782, -40972222.3455332, -0.0392566409616494, -177604167.074716, -1909722.45235578, -3819444.60658861, -40972222.7427321, -61458333.8406672, -3819444.81739071, 0, 0, 0, 0, 15972222.2205343, 0, -0.00153811949810618, 0.00057346089839799, 0.000368315716194949, 0, 65972222.2239532, 0, -0.00266627407162217, 0.00108824566031733, 0.0142721846334024, 65972222.2268935, -0.00240121989205826, -0.00573180386081162, 0, 104166666.689091, 0, 0.00663854630606865, 2083333.3201275, 0.0568075718957983, 104166666.709254, 208333333.394026, -0.00354852422817101, 0.00945844856675879, -0.00268273658981001, 0.00097418322580049, 0, 76388888.9055852, 0, 0.0391417381390393, 24305555.5391018, 0.0717396636992373, 76388888.9808318, 229166666.839474, 2083333.33830158, 0.029342345020871, 10416666.7297442, 0.0210567066651262, 229166666.987449, 0.13616626333172, 0.071655114640283, 24305555.6161105, 97222222.3979655, 0.0273094886723485, 97222222.813998, 145833333.951733, 10416667.1156651, 0.360342779714449, 20833333.7417398, 0.278733596314397, 20833334.1405887, 0, 0, 0, 0, -3472222.22180899, 0, 0.000353931598657377, -0.000125319096686919, -1.98850858933e-05, 0, -17361111.1109698, 0, 0.000785812636672737, -0.000613267352958262, -0.00280921565484906, -17361111.1116884, 0.000825531528566332, 0.00166483977654227, 0, -34722222.2289454, 0, -0.000845814794615048, 0.00305941173106554, -0.015872040450006, -34722222.2314519, -69444444.4578481, 0.00239600952849124, -0.00171118076719644, 0.00276744057594894, 0.000589167696465251, 0, -34722222.2333093, 0, -0.00914705806073311, -3472222.21308512, -0.0274582601392431, -34722222.2499163, -104166666.720816, 0.00762120611805912, -0.00783613248578699, -0.0111534230455365, -0.00585816441767289, -104166666.760115, -0.0299902525791013, -0.00992408285278028, 0, -17361111.1106966, -0.0222766644048084, -17361111.1364792, -69444444.5250548, -3472222.2223211, -0.00631851850961484, -17361111.1625607, -0.00384914912092072, -69444444.675194, -104166666.9193, -0.128834691005643, -0.119528315127386, -0.110450413156297, -0.0667370301568925, -17361111.3618272, -34722222.4852787, -0.139362467975547, -34722222.6986237, -0.387706462349443, -0.340978858083863}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 6.05874245911218e-07, 4200.00000015045, 0, 0, 0, 0, -107616.666665055, 0, -7.68956873907921e-06, -20116.6666752392, -1.53155033829322e-05, -20116.6666789607, 0, 0, 0, 0, 1067513.88886283, 0, 2.02647728694873e-05, 35916.6667327375, 0.000130368246474415, 0, 490930.555638046, 0.000201269252347961, 490930.555781999, 35916.6667882147, 71833.3334667121, 0, 0, 0, 0, -5460277.77760947, 0, 2.13546806562363e-05, -28333.3336078184, -0.00054758402135256, 0, -4515208.33397669, 0, -0.000723694119779487, -833263.889239986, -0.0016280629962644, -4515208.33472267, -28333.3336146687, -85000.0004280501, -833263.890696684, -1666527.78010722, -85000.0015863167, 0, 0, 0, 0, 15968402.777094, 0, -0.000164286992343121, 8333.33411044665, 0.00126922066832905, 0, 20661805.558548, 0, 0.00107460766165471, 623611.111545442, 0.00629633062310397, 20661805.5595899, 8333.33324457868, 33333.3338054673, 0, 7058680.55801169, 0.00543784390575581, 7058680.56480735, 14117361.12381, 623611.114933266, 33333.3379230298, 1870833.34042116, 50000.0064377943, 1870833.35292678, 0, 0, 0, 0, -27652777.7762012, 0, 0.000357894801378245, -0.00138924685107695, -0.00173394653118366, 0, -51402777.7838998, 0, -0.00122535776512715, -173611.111672547, -0.0126632953930151, -51402777.7840269, 0.00101367557340622, 0.000235389532367307, 0, -28402777.7902222, 0, -0.00723846231695911, -4826388.88884458, -0.017936674406986, -28402777.7998381, -56805555.5867745, -173611.11086312, -0.00467799025512541, -694444.453281549, -0.014512934116035, -4826388.90488851, -14479166.7010388, -0.0152999103555436, -14479166.7462957, -694444.497550719, -1041666.72449225, 0, 0, 0, 0, 27986111.1091648, 0, -0.000400649704120179, 0.00146173587786417, 0.00133601858052067, 0, 70208333.3388007, 0, 0.00123338265326189, 0.00145023226461596, 0.0139095445369932, 70208333.338458, -0.00136427846771402, -0.00114130584881631, 0, 57673611.136744, 0, 0.00452436561688451, 1215277.77144029, 0.0297984790322274, 57673611.1384474, 115347222.260722, -0.00801619434473698, 0.00142627875826108, 0.00307713964629715, 0.0200750904533065, 0, 16666666.6772824, 0.0189882602824086, 16666666.6974773, 50000000.0686537, 1215277.77814222, 0.0092251180433403, 4861111.14933386, 0.0281806558624267, 50000000.1406176, 0.0569847326904738, 0.0761368198807765, 4861111.27385966, 7291666.8446152, 0.133665927975483, 0, 0, 0, 0, -15277777.776575, 0, 0.000260194932177382, -0.000796865748277665, -0.000569098658354134, 0, -49305555.5572742, 0, -0.000868202580250606, -0.00186157523027402, -0.00775158578242085, -49305555.5576056, 0.000786056327413697, 0.00129653490203518, 0, -56250000.021157, 0, -0.001970286170418, 0.00614993103639809, -0.0234720707003588, -56250000.0165817, -112500000.022765, 0.00840270134157948, -0.000112919770386551, 0.00202918211361083, -0.0143356343743741, 0, -25694444.4672416, 0, -0.00821574494373073, -3472222.21914483, -0.0225237448056096, -25694444.4721506, -77083333.3941386, 0.0166320824471765, 0.000917372879817422, -0.0168809964707235, -0.0286777747436368, -77083333.4478983, -0.022708516251346, -0.0477889658088989, -3472222.22568959, -13888888.9452878, -0.00850467909799336, -13888889.0806526, -20833333.5498655, -0.12022173185216, -0.11259480821004, -0.153759765462162, -0.124888677662229, -0.275796591444424, 0, 0, 0, 0, 3472222.22192954, 0, -7.16506633994798e-05, 0.000180250209497361, 0.000103080600849558, 0, 13888888.8888366, 0, 0.000274016308413035, 0.000692883948955804, 0.00174053965628026, 13888888.8891792, -0.000208946655931255, -0.000466669901524401, 0, 20833333.3393772, 0, 0.000607709214691757, -0.00146981852015032, 0.00705176963890469, 20833333.3372257, 41666666.6718523, -0.00252675090985705, 0.000219597059561972, -0.00139682372056438, 0.00412452357871363, 0, 13888888.900591, 0, 0.00151393960992865, -0.00422164482863808, 0.00951860094179649, 13888888.8980299, 41666666.6862587, -0.00887371420784632, -0.00119869447592226, 0.00198220057275822, 0.0119147917471861, 41666666.7010381, 0.0016485415512072, 0.0120870718899252, 0, 3472222.22636521, 0.00534627398905407, 3472222.22564995, 13888888.915865, -0.00890005868902348, -0.00305054466857077, 0.0150887142756368, 0.00514401709984564, 13888888.9655023, 20833333.4224726, 0.0254156627828653, 0.0305777056031036, 0.0447739693740264, 0.0426899241287752, 0.0712567661402982, 0.0923568199176445, 0.0537266522477109, 0.163500355223072, 0.124487807988465, 0.130084059683339}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -4.83605349640882e-08, -2000.00000015833, 0, 0, 0, 0, 52388.8888892126, 0, 3.14341266527248e-07, 6722.22222859199, 1.60538321571198e-06, 6722.22222426013, 0, 0, 0, 0, -535990.740734032, 0, 4.60429263968482e-06, -7500.00004892761, -1.68344493286223e-05, 0, -168490.74081094, -1.36556540413178e-05, -168490.740788651, -7500.00001654639, -15000.0000244772, 0, 0, 0, 0, 2851111.11104161, 0, -3.33085480278906e-05, 2777.77797632862, 8.28099645429369e-05, 0, 1611250.00050612, 0, -4.05279216757106e-05, 179583.333686621, 0.000174933579864907, 1611250.00031211, 2777.77776632956, 8333.33340235217, 179583.333660706, 359166.667181668, 8333.33366010085, 0, 0, 0, 0, -8731249.9996634, 0, 5.70686379382089e-05, -0.000538279841639321, -0.00020958234033156, 0, -7764351.85393056, 0, 0.0004331417187918, -63425.9266215522, -0.00097049282389251, -7764351.85280561, 0.000279763100090903, 5.96461849235265e-05, 0, -1596527.78005029, -0.000123568425399447, -1596527.77960769, -3193055.55836474, -63425.9261409734, -0.000737252518654739, -190277.779325998, -0.0016448313500008, -190277.782358536, 0, 0, 0, 0, 15902777.7769556, 0, -6.84209668030481e-05, 0.000935393687645565, 0.000345346656070722, 0, 20590277.7817689, 0, -0.00057683110515251, 0.000976771491229264, 0.00224065660770477, 20590277.7793776, -0.00059113318599032, -0.000526608308857102, 0, 6875000.00919064, 0, -0.00120490908751439, 520833.335556, 0.00138103970539667, 6875000.00453251, 13750000.007224, -0.00207002469430313, -0.000202851841581629, 0.00172280149447594, 0.00440038219787724, 520833.334746378, 1562500.00732183, 0.00308099816915733, 1562500.01787831, 0.011728875183746, 0.0150696137835278, 0, 0, 0, 0, -16944444.4434129, 0, 3.43821219266875e-05, -0.000977574286841388, -0.000245496199806778, 0, -30254629.6331076, 0, 0.000333793630369059, -0.00132041394522617, -0.00285258098459884, -30254629.6310467, 0.000573379444631212, 0.000882822027075863, 0, -15277777.794814, 0, 0.00299103087938667, 0.000907119743696587, -0.0041712662238441, -15277777.7838529, -30555555.5647515, 0.00454240367153154, 0.00165103060591533, 0.00034918445661776, -0.00637055081906801, 0, -1967592.6031687, 0.00175041695511728, -1967592.59597144, -5902777.79250691, 0.00532962894914041, 0.00111392443154094, -0.00731306664750286, -0.00864811114109167, -5902777.80887515, -0.0098985734682485, -0.023586211585494, -0.0341622319925742, -0.0427830203725055, -0.0352284060963816, 0, 0, 0, 0, 9722222.22158396, 0, -2.03267545892737e-05, 0.00053766516240022, 0.000107617343817155, 0, 22916666.6677875, 0, -1.89501203659795e-06, 0.00126343246018894, 0.00170294288470196, 22916666.6673147, -0.000235478500624514, -0.000859581609870145, 0, 16666666.6800267, 0, -0.00204195980821755, -0.00200998151146958, 0.00402153993104977, 16666666.6706753, 33333333.3389408, -0.00378597507799153, -0.0011988508503962, -0.00167857526709949, 0.00438416179558485, 0, 3472222.23904833, 0, -0.00393702377745591, 0.00220691407177067, 0.000452768710090891, 3472222.22624482, 10416666.6803667, -0.00857870382041081, -0.00372597301443934, 0.00229733839581878, 0.00953598890700881, 10416666.6929834, 0.00143970225344312, 0.0161425374708836, -0.0052102916712122, 0.0112964153149729, -0.000479263110226969, 0.0398028434406659, 0.0505484122727564, 0.0201223328821004, 0.0275298026699312, 0.0412264388102367, 0.0395260835428365, 0.0653688037231734, 0, 0, 0, 0, -2314814.81466031, 0, 7.08688595931181e-06, -0.000123828473960338, -1.97571146953644e-05, 0, -6944444.44444121, 0, -6.97638621960715e-05, -0.000430158144500944, -0.000412255788230531, -6944444.44455688, 3.70543119829897e-05, 0.000275591260584801, 0, -6944444.44815308, 0, 0.000390643953196735, 0.000530734215773903, -0.00141678504720627, -6944444.44549534, -13888888.8902441, 0.00102710906806743, 0.000176554796115289, 0.000738906313073538, -0.00120068489851972, 0, -2314814.82263574, 0, 0.00141927925928916, 0.00121419352982014, -0.000842024039329568, -2314814.81633285, -6944444.44893256, 0.00363907380193165, 0.00161347188007058, 0.000187456388794371, -0.004053710503815, -6944444.45250627, 0.00120366476355942, -0.0040595836030744, 0, -0.00417930143691968, 0.00119504392918071, 0.00133243914174701, -0.00563548002080016, 0.00467766404007545, 0.00290193568152692, -0.00266478420197719, -0.00105921859805767, -0.0152152703701983, -0.0203391081735408, -0.0019734951579704, -0.00627611202084716, -0.0121074232361343, -0.0155612928524506, -0.0121784626520407, -0.0220701752505599, -0.0161120799726412, -0.0377122143222868, -0.0289146373150786, -0.0360379809388482}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -3.629104237796e-08, 642.857142914456, 0, 0, 0, 0, -17107.1428574167, 0, 6.10592095191428e-07, -1357.14285942308, 5.73558061691966e-07, -1357.14285720964, 0, 0, 0, 0, 178964.285713352, 0, -5.54505271357766e-06, 714.285731122244, -3.78847337365144e-06, 0, 34678.5714571886, -9.25892360557159e-06, 34678.5714348348, 714.28571463959, 1428.57142888226, 0, 0, 0, 0, -979861.11109113, 0, 2.4184159378181e-05, -6.55040850372296e-05, 1.34312780395197e-05, 0, -341111.111326777, 0, 7.0481373641688e-05, -17500.0001407636, 4.36803722050597e-05, -341111.111152941, 1.12109674208168e-05, 1.50334038158661e-05, -17500.0000394632, -35000.0000691115, -2.73964337357303e-05, 0, 0, 0, 0, 3109027.77766461, 0, -4.46551581047969e-05, 0.000176116265069466, -3.24245156094224e-05, 0, 1707638.88975504, 0, -0.000307414134027258, 0.000256123942808626, -4.55378950449357e-05, 1707638.88902073, -7.72236493528253e-05, -0.000148072015792215, 0, 161111.112102119, -0.000303315579635356, 161111.111338647, 322222.222574465, -4.52083672229767e-06, 6.08247219046064e-05, 0.000183674018054499, 0.000141484685447177, 0.000762297328453393, 0, 0, 0, 0, -5902777.77748815, 0, 5.99361665556995e-05, -0.000311295321544777, 1.95933396409213e-05, 0, -4756944.44607776, 0, 0.000409482298229336, -0.000289825481994765, 6.39282477428338e-05, -4756944.44468922, 0.00012051719501373, 0.000411520315771315, 0, -729166.670539223, 0, 0.000831340310572356, -0.0011263144262339, 0.000592092533987354, -729166.667141661, -1458333.33437301, 0.000425317667039536, 0.000150670146614542, -5.24936594086692e-06, -0.000407631168980182, 9.85104276812358e-05, -0.000840333278734016, -0.000187996663150976, -0.00262612480372013, -0.00214461096479215, -0.0024655588020787, 0, 0, 0, 0, 6587301.58693285, 0, -4.15083672956337e-05, 0.000334969625554546, -4.44264842148102e-05, 0, 7430555.55697678, 0, -0.000381402996268705, 0.000342704624585888, 9.48804721796306e-05, 7430555.55578782, -0.000128944817347484, -0.00050990977717396, 0, 1736111.11805866, 0, -0.00150783467482682, -9.54817882594991e-05, -0.000315039657194044, 1736111.11178222, 3472222.22360632, -0.000756057895320376, -0.000548733329550464, -0.000658742738747159, 0.000395346215003668, 0, 0.00480806976730964, -0.00185552460270982, -0.000397106686639744, 0.00164235502696481, -0.00154317662793086, -0.000696553028629317, 0.000472224855541308, 0.0012421500845445, 0.0040212282984635, 0.00203208379609768, 0.00416738188127901, 0.00534060443852453, 0.00673348176636325, 0.00652795305255891, 0, 0, 0, 0, -3968253.96802509, 0, 2.02977570947827e-05, -0.000188929902687161, 2.29998061745859e-05, 0, -6051587.30206301, 0, 0.000190434325656191, -0.000360896051982459, -6.86256416244502e-05, -6051587.30171141, 5.37346947695684e-05, 0.000415774271567534, 0, -2083333.3386711, 0, 0.00103225378876374, 0.00071369912735986, 4.96522870887367e-05, -2083333.33381526, -4166666.66751731, 0.000636907838283138, 0.000467326642105431, 0.000842563006407761, -0.000142034842742981, 0, -0.00712594534014315, 0, 0.00203290687916209, -0.00148748023327493, 0.00115991498347436, 9.01727839596113e-05, -0.0018100176208077, 0.00165818384389483, 0.000875033947162375, 0.000272880501217181, -0.00121248645958031, -0.003620906895761, -0.000692550139457361, -0.0026463076475794, 0.00195613909339669, -0.00109927335857366, 0.000653095660440563, -0.0053148840590598, -0.00748237400260206, -0.00274418839824111, -0.00585382052045535, -0.00734175495274251, -0.00789867972621695, -0.0109577428713149, 0, 0, 0, 0, 992063.492008165, 0, -4.45500457215157e-06, 4.46750150700444e-05, -5.16197292450715e-06, 0, 1984126.98413786, 0, -2.2965603337746e-05, 0.000130323340994984, 2.04445726962379e-05, 1984126.9841526, -5.77333756792467e-06, -0.000116884932142266, 0, 992063.493527419, 0, -0.000218697842385158, -0.000222091518512042, 7.95111349781701e-05, 992063.492217933, 1984126.98435189, -0.000166686787927227, -0.000105953355768893, -0.000287565318727558, -1.2023040526898e-05, 0, 0.00319797264947852, 0, -0.000605932390943845, -0.000289964701665979, -0.000248442034904469, -1.855679936535e-05, 0.000586108460340587, -0.000646678066726775, -0.000429512042907314, -0.000272407558637788, 0.000464827817293451, 0.00112071328288869, -0.000113402091129249, 0.000404060830799058, 0, 0.00193268353818828, -0.000765628917238318, -0.000703803184835985, 0.000610348014819758, -0.00109106190389465, -0.000803093833750139, 0.000157532844102457, -2.92107745018323e-05, 0.00150615544921118, 0.00275502279695662, 0.00026175360146706, 0.00144113032791383, 0.00179266284492943, 0.00315910126137824, 0.00139793161771361, 0.00339494975684955, 0.00343653329841834, 0.00566664374267203, 0.00529295469247287, 0.00707399526829895}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 0, -71.4285714537544, 0, 0, 0, 0, 1821.42857152331, 0, 6.38496769457609e-08, 785.714286422555, 1.1467143031887e-07, 71.4285715198491, 0, 0, 0, 0, -17861.1111105398, 0, -5.76957101714584e-07, -714.285720738777, -1.57594080761948e-06, 0, -19964.2857189085, -9.17221076835689e-07, -1750.00000178831, -1.71063378136577e-06, -714.285715940306, 0, 0, 0, 0, 89027.7777727311, 0, 4.20395113829108e-06, 2.96018439497323e-05, 5.20461185625636e-06, 0, 194722.222232498, 0, -2.7406653282932e-06, 17500.0000385059, 6.87059659487842e-06, 16111.1111153283, -3.17710731035325e-06, 1.10844930844163e-05, 3.84754851093364e-06, 17500.0000227669, -3.6570681021355e-06, 0, 0, 0, 0, -246527.777760441, 0, -2.52705009633015e-06, -7.7686584187271e-05, -1.98499346024289e-05, 0, -963194.44446123, 0, -1.42995297548674e-05, -0.000153237088608771, -4.6400749305228e-05, -72916.6666844486, 3.3537754417486e-06, -4.69361111294212e-05, 0, -161111.111196913, -7.27248656780184e-05, -4.2452027145996e-06, -161111.111195212, 1.02452115113856e-05, 1.79458848569733e-05, -1.24670254293668e-05, -8.0318272651892e-05, 1.00502867775189e-05, 0, 0, 0, 0, 381944.444409828, 0, 1.38597203214604e-05, 0.000120021325202085, 9.24894451263793e-06, 0, 2638888.88890777, 0, 3.35305953359429e-06, 0.000344222031281101, 8.58167628649858e-05, 173611.111153951, -6.33548479507333e-05, 7.74349135980299e-05, 0, 729166.666797343, 0, -0.000172439607083886, 0.000226360975946571, 7.21586884877035e-05, 6.80471845308922e-05, 729166.666759818, -6.16049508437204e-05, -0.000268875066799456, 0.000159801994169771, 0.000174733144804803, 3.86680900347711e-05, 5.46770355368742e-05, -8.21541502205839e-06, 0.000249393253407951, -1.09592392396902e-05, 0.000482359251168763, 0, 0, 0, 0, -307539.682500457, 0, -2.33889024269304e-06, -0.000109019475359367, -1.57544880251959e-05, 0, -4027777.77777369, 0, -7.51015079072589e-05, -0.000431452409169723, -0.000105552690273688, -208333.333387914, 5.01487009261973e-05, -7.36094669776243e-05, 0, -1736111.11132661, 0, -9.99127624654703e-05, -0.000376144062376626, -0.000444921976193814, -5.14500546614652e-05, -1736111.11118791, 0.00026197460646234, -2.00422900357116e-05, -0.000175516360427176, -0.000610384157349567, 0, -0.000174537601653682, -0.000283401698740677, -2.97050235013555e-06, -0.000687898086231108, 6.07479579249347e-05, 0.00101876529823513, -5.88353175648572e-06, 0.00029545653017399, -0.000771541798412093, 0.000172721060134391, -0.000144632285212219, 0.000559593376846483, -0.000126524056785987, -0.00127244878054344, 0, 0, 0, 0, 99206.3491831564, 0, 5.34314904538726e-06, 5.29095187725688e-05, 3.86790171901857e-06, 0, 3174603.17457722, 0, 3.20971297810419e-05, 0.00028402267988419, 9.80167943543908e-05, 99206.3492414632, -4.53939368898919e-05, 4.36151978185049e-05, 0, 2083333.3335154, 0, 8.87062406884086e-05, 0.000336872988138045, 0.000323714745706219, 9.70282579227063e-05, 2083333.33342268, -0.000183447116717429, 1.57784084067855e-05, 0.000218078850245696, 0.000444801691720078, 0, 0.000175172026659809, 0, 8.77923611672646e-05, 0.000114872621960471, 0.000155123723620896, -2.97966185173093e-05, 0.000382499113147063, -0.000202960046500965, -0.0001012404262173, 0.000143487888051141, 0.000240222886311764, -0.000253413219261722, -0.000333577246515342, 0.000973730455166945, -0.000232904762159654, 0.000223743269612213, -0.00135620256827145, 0.00100551634639837, 0.000584562591506242, -0.000648037965856991, -0.00090967553496733, 0.000130083034346807, 8.34999383890638e-05, 0.00156101206629675, 0, 0, 0, 0, 5.50949550362196e-06, 0, -1.30562377831602e-06, -1.09282329113389e-05, -1.99583141128504e-06, 0, -992063.492049522, 0, 4.60040509762878e-06, -7.32980547796118e-05, -2.16562264151097e-05, -5.92731371195979e-06, 8.00820601152527e-06, -7.16492188354994e-06, 0, -992063.492116194, 0, 2.44801850168828e-05, -0.00012408761378267, -1.52271033788249e-05, -1.26417023123957e-05, -992063.492091741, 4.05174783962611e-05, 8.43892131572708e-07, -2.86828393107824e-05, -0.00010964285772248, 0, -9.52587234720622e-05, 0, 4.54256056067313e-05, -4.38675062984411e-05, 3.99283262351928e-05, -5.16954569770964e-05, -6.08292479190059e-05, 0.000202851924052673, 0.000161326353815381, -7.65609855855226e-05, -7.63535084098655e-05, -7.92872551059535e-05, 0.000178016262066547, -0.000369382479160307, 0, -2.24914094056005e-05, 2.96429935862119e-06, 2.45660905970388e-05, -0.000308580041145673, 4.4599352966034e-05, 0.000305412573458445, -0.00014636047028912, 0.00025147441687408, -0.000230913641873077, -0.000692750546294593, 0.00026589889252185, 2.14987789598586e-05, -0.000488067565941825, 0.000105973652161163, -8.00155565720265e-06, -0.000552182394896028, 0.000874783026207495, -0.00128049381262222, 0.000682879017131824, 5.08509269912761e-07}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -3.41183576994242e-08, -55.5555555033835, 0, 0, 0, 0, 1324.07407339531, 0, 5.51601761198993e-07, 888.888887727254, 3.54065227526734e-07, 55.5555559580289, 0, 0, 0, 0, -11685.1851815573, 0, -4.22884679537808e-06, -3611.11110148044, -8.15671978367657e-07, 0, -21129.6296158101, -8.81497472481462e-06, -1268.51852093513, -2.85053878772181e-06, -833.333335884542, 0, 0, 0, 0, 49768.5185089085, 0, 1.17225953805947e-05, 2777.77773705825, 3.68827943062735e-06, 0, 185694.444379226, 0, 6.55073493687322e-05, 85231.4813818593, 5.65473315508912e-05, 10416.6666958595, 4.30907394680631e-05, 2777.777780129, 4.64278700056759e-05, 19027.7777815221, 9.61051152814757e-05, 0, 0, 0, 0, -108796.29627962, 0, -3.93863475891288e-05, 9.95402733958355e-05, 1.6731647043744e-05, 0, -785879.629480377, 0, -0.000191465011301114, -63425.9255819689, -0.000111046273748018, -39351.851926609, -0.000112855162884479, 3.17145015656559e-06, 0, -740509.258816895, -0.000174234793861167, -0.000198740698034532, -156250.000113154, -0.000185233620236498, -0.000368846879343923, -63425.9260294188, -0.000235288764343204, -0.000405518452057581, 0, 0, 0, 0, 115740.740719909, 0, 3.50902006667075e-05, -0.000147391483002584, -2.30948085140953e-05, 0, 1701388.88871036, 0, 0.000301997270966367, -0.000634963948762737, 0.000259163479405205, 69444.4445454962, 0.000282772156013356, 1.32752797456322e-05, 0, 3078703.70284618, 0, 0.000936174365463971, 520833.332091616, 0.000682370694304194, 0.000381565721079228, 590277.778215083, 0.00042880306127422, 0.00126744213424783, -3.92140187120651e-05, 0.000537891504216948, 0.000514259716727962, 520833.33389744, 0.00122742416216175, 0.000833106077665878, 0.00118368098212905, 0.000201917736559915, 0, 0, 0, 0, -46296.29627659, 0, -5.80922334627289e-05, 0.000134694215866912, 3.51177155508519e-05, 0, -1782407.40729172, 0, -0.00019196999127584, 0.000647316219226276, -9.3906645775118e-05, -46296.2963753194, -0.000280625219680196, 5.18354808659715e-05, 0, -6481481.48061465, 0, -0.000602439808401879, 0.00173775598151145, -0.000172098798522505, -0.000537858123143365, -1041666.66728506, -0.000844294831460316, -0.000896343118442987, 3.97465779649582e-05, -0.000111676178354841, 0, -1967592.59059254, -0.000812224709940735, -0.00112502698085921, -1967592.59256056, -0.000986911019555056, -0.00385736728680039, -0.000996989058870448, -0.00278274757870656, -0.00106401792620009, -0.00207618018156095, -0.0019694861103775, -0.00398095403685722, -0.00232723219341011, 0.00034069744743708, 0, 0, 0, 0, -1.14733333944863e-05, 0, 2.05066098644043e-05, -6.79430228367616e-05, -1.63583640360658e-05, 0, 694444.444400656, 0, 0.000161097316071974, -0.00036103862390469, -7.61866274558289e-05, 2.37218522057062e-05, 0.000210423347442323, -3.97745332747713e-05, 0, 6481481.48106867, 0, 0.000379077999827261, -0.00115588645055747, 0.000162517913281117, 0.000212822701473799, 694444.444803101, 0.000556467073848684, 0.000775623696824689, -0.000265719758021196, -2.04103041526075e-05, 0, 3472222.22076015, 0, 0.000955875156211652, -0.00196771879716176, 0.000815791973755847, 0.000862507191593456, 3472222.22249444, 0.000740209448182817, 0.000937153639323427, 0.000173153453253113, 0.00157047388009005, 0.00272471004026353, 0.00181401620000523, -0.000377527844288739, 0.00147153430964815, 0.000636761450795032, 0.00464995028355733, 0.00152492232099891, 0.0019305935410667, 0.00470281003788114, 0.00359905433165318, 0.00236783270878859, 0.00237487568112944, 6.44597793692287e-05, 0, 0, 0, 0, 2.87351155429241e-06, 0, -3.69152984507198e-06, 1.52678713259357e-05, 7.47114403735022e-06, 0, 8.69430093642283e-06, 0, -6.16693995973929e-05, 8.13153264068708e-05, 2.18649255129245e-05, -8.42092732310716e-06, -3.59398259703144e-05, 1.5500458903443e-05, 0, -2314814.81474142, 0, -0.000174126674715773, 0.000305728012102621, -0.000141749101878053, -6.46210817236174e-05, -8.76727735714781e-05, -0.000134514220691882, -0.000263793903623669, 5.15726309077315e-05, -6.34147072896942e-05, 0, -2314814.81442542, 0, -0.000391363103611909, 0.000671976714400466, -0.000488261926462289, -0.000168254496383579, -2314814.81508247, -0.000497031201936469, -0.000775010216733965, 7.7102898317056e-05, -0.00066987355561815, -0.000454758092608438, -0.0008225842324638, 5.67749629399014e-05, 0, 0.00076134082345277, -0.000556279773277162, -0.00037051164368053, 0.000172484592208238, -0.000270181884388015, -0.00134867127724903, 1.93777267425197e-05, -0.00135774118247587, -0.00108063970579556, 0.000112892037922026, -0.00161602983122266, -0.000723764822512377, 6.15013131811771e-05, -0.00125268898111528, -0.00157577173333094, 4.60825613677757e-05, -0.00285620600192682, 0.000298205455618992, -0.00335229374962284, -0.00146820287286255}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 3.81610988498554e-08, -49.9999999556691, 0, 0, 0, 0, 1091.66666601475, 0, -5.34739543299036e-07, 966.666665952108, -1.18834022938013e-07, 50.0000003628611, 0, 0, 0, 0, -8333.33332880794, 0, 6.27013479747814e-06, -5916.66666087183, -1.35111630551744e-06, 0, -21055.5555484302, 4.12372041089348e-06, -1041.66667475044, -1.0027084167096e-05, -916.666671460151, 0, 0, 0, 0, 28124.9999808081, 0, -1.9702443946362e-05, 13333.3333108395, 2.36098063308482e-06, 0, 160069.444412812, 0, -5.28953472911261e-05, 128263.888835403, -1.50165272175679e-05, 7291.66668355505, 3.88506325516533e-06, 5000.00002528256, 6.13701213857676e-05, 19097.2223127164, -1.61344553145078e-05, 0, 0, 0, 0, -41666.6666145017, 0, 7.74165841516661e-05, -8333.33328474531, -3.48115413974992e-05, 0, -536458.333278732, 0, 0.000212133986297151, -286111.110938578, -5.74956604725022e-05, -20833.3333857678, -2.41055462621174e-05, -8333.33337017857, 0, -967013.88868938, -1.3232005781931e-06, -0.000231901057035004, -133680.555836364, -0.000373468160433314, 2.08830667681909e-05, -104166.666835778, -8.16843064713164e-05, -0.00031278674893222, 0, 0, 0, 0, 20833.3332387724, 0, -8.37597172816301e-05, -6.63159237587266e-05, 2.54252090758989e-05, 0, 784722.222183092, 0, -0.000389061178030254, 173611.110850196, -8.25317836570137e-05, 20833.3334152925, -0.000115841380091313, -2.59741166242427e-05, 0, 3194444.4440778, 0, -0.000816819672161991, 2118055.55488271, -7.38227047005362e-05, 0.000458118736106962, 381944.444843098, 0.000873308947952688, -0.000570676462794935, 173611.111471647, 9.10480547338967e-05, 0.00107881996317563, 729166.667086047, 1.4105499600782e-05, 0.00185275322107899, 0.00143951670457817, 0.00218281958002003, 0, 0, 0, 0, 0.000103744707955656, 0, 0.000130046457224409, 5.50615624984866e-05, -3.18312748689252e-05, 0, -381944.444423193, 0, 0.000304436001931757, 0.000190842487216516, -0.000158381544152345, -7.68321920645503e-05, 0.000183013790054854, -2.52539708890286e-06, 0, -4548611.11093015, 0, 0.000419032700103231, -1215277.77686062, -0.000731141836832606, -0.0002711686603799, -381944.444902858, -0.000640376676263453, -0.000132712345856481, -0.000177984151575187, -0.000694764345012203, 0, -6770833.33227013, -0.000149582765438946, -0.000836141698728486, -2083333.33556737, -0.00166154201735406, 0.00236475491303424, -1215277.7776849, 0.00169500525934292, -0.00297267803520165, -0.00273888507339218, 0.000766112662338254, -0.00259344972425268, -0.00274794502300506, -0.00710451004848471, 0, 0, 0, 0, -6.18413730077946e-05, 0, -5.59013450756748e-05, -2.89713960526697e-05, 6.23054342662949e-06, 0, -2.18498781675909e-05, 0, -0.00027024650536494, -4.00640545968298e-05, 0.000236430444620898, 4.23431466086844e-05, -0.000223046996231126, 1.85106588283106e-05, 0, 2083333.33337002, 0, -0.000331583158650046, -0.000563202298256493, 0.000429648432238477, 0.000243898710693783, 0.000238236185837657, 0.000356715002544644, -0.000371527942690738, 0.000318212235709098, 0.000643723570227401, 0, 9027777.77718358, 0, -0.000398870233238671, 3472222.22110022, 0.000194338563773584, 0.000433023974489735, 2083333.33483336, 0.00151292353985923, 0.00129157672768195, 0.00039394765189011, 0.000199436612312358, -0.000840950658888696, 0.00204916848354341, 0.00231093006641103, 0.000715964563255999, 3472222.22324121, -0.00289276621210292, 0.00569535526120072, 0.0035011073729654, 0.00262789442284285, 0.00343553352234791, 0.000639583608257456, 0.00156957320770455, 0.0122663735673075, 0, 0, 0, 0, 1.52428683250414e-05, 0, 1.04993021707911e-05, 6.09449816715286e-06, -5.85499788700707e-06, 0, 1.20967233496713e-05, 0, 0.000103555861043984, -2.91989211584388e-07, -3.92618910017194e-05, -1.3871373728868e-06, 4.24913660525722e-05, -9.96389778313139e-06, 0, -3.55529617888144e-05, 0, 0.000195042898692813, 0.000108844609698059, 5.01054853986011e-06, -3.72433873557784e-05, -4.51554806163242e-05, -7.14059909353818e-05, 0.000227503519578426, -7.88391151097799e-06, 8.3651702052326e-06, 0, -3472222.22214191, 0, 0.000320018465266512, 0.000404202115557667, 0.000156995997172246, -0.000183271293370939, -0.000284124502883665, -0.000100077445153534, 0.000422277529005865, -0.000204890401495408, 0.000212417896875154, -0.000317770995043334, -0.000356076264759747, -0.000468717232847463, 0, -3472222.2218322, 0.000322107268100706, -0.00031230920484965, -3472222.22336056, -0.000627309605887792, 0.000640238442809702, -0.000594443482780684, 0.00039822787572835, -0.00143443578788977, -0.00260186981112269, -0.0015217797705007, -0.00229599657392441, -0.0016044355591863, -0.000278478237495483, -0.00221944805656437, -0.00275612078145637, 0.000723028273470047, -0.00555296637370218, -0.001437567451963, -0.00246375981919685}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 0, -49.9999999904149, 0, 0, 0, 0, 966.666666291784, 0, -8.64797875506071e-07, 1091.66666693649, -3.11666289243561e-08, 50.0000003347369, 0, 0, 0, 0, -5916.66666235592, 0, 7.23867642900218e-06, -8333.33333640569, -1.33396105534924e-06, 0, -21055.5555581791, 7.54489429237118e-06, -916.666670503315, 4.57027388858506e-06, -1041.6666706176, 0, 0, 0, 0, 13333.3333102269, 0, -4.11279167539562e-05, 28125.00001494, 1.51429777354528e-05, 0, 128263.888899491, 0, -5.19228601113991e-05, 160069.444475072, -1.83748549581259e-05, 5000.00004287051, 2.12051598051842e-05, 7291.66667841193, 5.27059953581175e-05, 19097.2222488204, 5.82045717182988e-05, 0, 0, 0, 0, -8333.33325986082, 0, 7.29195114122319e-05, -41666.6667097964, -8.73367063288747e-06, 0, -286111.111126524, 0, 0.000251490792731659, -536458.333448472, 7.8626051564953e-05, -8333.33343267806, 1.2445401230656e-05, -20833.3334142396, 0, -967013.888990512, 0.000305563981496644, -0.000237323236522863, -104166.66693408, -8.52540561854414e-05, -0.000204272138182305, -133680.555757342, 2.83882124119193e-05, -0.000555975703917924, 0, 0, 0, 0, -0.000135361314780059, 0, -0.000147447719218794, 20833.3334093206, 6.48243670193406e-05, 0, 173611.111078395, 0, -0.000527610965553043, 784722.222485893, 0.000176038002698687, 0.000140840590232258, 7.02900956886349e-05, 20833.3335788877, 0, 2118055.55589694, 0, -0.000209766939842899, 3194444.44473125, -0.000394368998195268, 0.000462330655738046, 173611.111956466, 0.000132117164762857, 0.000824539984513536, 381944.444861325, -0.000125969683065682, 0.000496891915143176, 729166.667913573, 0.000114022943818503, 0.00146250388640073, 0.00197365289629317, -0.000512602951895175, 0, 0, 0, 0, 0.000143791435825481, 0, 7.62773318923821e-05, -7.26756306138535e-05, -3.96688217752597e-05, 0, 0.00013228889915231, 0, 0.000767547324316385, -381944.444820038, 1.82482571136988e-05, -9.7215748344171e-05, -1.07351061076578e-05, -0.000258767892627924, 0, -1215277.77822427, 0, 0.00122583126581832, -4548611.11149178, 0.000700861880203461, -0.00068303919851039, -0.000876857935501871, -0.000414352705464896, 0.000121441284794878, -381944.445081357, 0.0012563532123528, 0, -6770833.33393731, 0.00145500344488893, -0.00137143200119288, -1215277.77853004, -0.000772420711006836, -0.00229099503042472, -2083333.33566087, -0.00179970090024445, -0.00281792493490548, -0.0021980163236115, -0.0024882516072374, -0.00542547642095755, -0.003412332051111, 0.0028814896402843, 0, 0, 0, 0, -8.0918989395829e-05, 0, -4.37370316439796e-05, 3.72802961962508e-05, 2.85086975837025e-05, 0, -0.000145735290562102, 0, -0.000404785945761813, 0.000263820915419529, -0.00013400860698719, 4.08815844197913e-05, 2.91577239561708e-06, 0.000126021791314216, 0, 0.000249894119987174, 0, -0.00102073739383821, 2083333.33367837, -0.000368537117401828, 0.000287896715802529, 0.00057349131372227, 0.000189883634793783, 0.000131787984341856, 0.000235623316466973, -0.00113521210819789, 0, 3472222.22285494, 0, -0.00120278219317395, 9027777.77816064, -0.000924647660060034, 0.000946333521358248, 0.00101275321691752, 0.000443438924230555, -0.000998688240746426, 2083333.33470161, -0.00105774833673392, 0.00526634468541321, 0.00170997383370666, -0.00279810593111981, 0.00119846105946747, 3472222.22418845, 0.00200709393007503, 0.00370434014979712, 0.00381610358686766, 0.00580180913199447, 0.00184529513655006, 0.00390820307001356, 0.000685398200114893, -0.00265369739864491, 0, 0, 0, 0, 1.8703496596794e-05, 0, 8.71882026251365e-06, -6.69870660168925e-06, -2.79750897559496e-06, 0, 5.18481832802674e-05, 0, 6.55333808090834e-05, -7.94682558982706e-05, 1.0946386117448e-05, -1.20932554171664e-05, 1.12661094267753e-05, -3.62151842797288e-05, 0, -3.81312469395689e-05, 0, 0.000257893174147552, -0.000127173415888421, -5.97468401024343e-05, -7.28015205175727e-05, -0.000136876722218452, -4.02147298420251e-05, -9.03375147614053e-05, -0.000169145224346523, 0.000128683288809913, 0, -0.000229756239072667, 0, 0.000259103103019029, -3472222.222339, 9.43728249748788e-05, -0.000142320621488231, -0.000443905769477996, -0.000451200242031913, -0.000299785970416394, -0.000310280489203952, 0.000223794518732952, -0.000813911506475349, -0.000499261412968615, 0.000781504380663226, 0, -3472222.22243287, 0.000238311996876147, -0.000315116769143967, -0.000246497132409412, -0.000215811171766429, -0.000559631737797743, -3472222.22276629, 0.000129163831854512, -0.0016055862065851, -0.000490035458730751, -0.00172843987393638, 0.000223536101251508, 0.000956002905656977, 0.00105408734296686, -0.00248984814985227, -0.000248820538159129, -0.00236611226620313, -0.000980690612794822, -0.00451333909199926, -0.00061236169902947}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 6.73469605980223e-08, -55.5555555495223, 0, 0, 0, 0, 888.888888459425, 0, -9.78602769922957e-07, 1324.07407426787, -1.09776902871232e-06, 55.5555554540746, 0, 0, 0, 0, -3611.11110708104, 0, 5.9671893019987e-06, -11685.1851881102, 8.01666789224819e-06, 0, -21129.6296268432, 2.06121934800913e-05, -833.333335153607, 2.90198731198947e-07, -1268.51851739098, 0, 0, 0, 0, 2777.77775960473, 0, -1.31789334014788e-05, 49768.5185367406, -3.07712861837055e-05, 0, 85231.4814527049, 0, -0.000124606251081535, 185694.444435766, -0.000135652650193088, 2777.77778524768, -4.17906456917434e-05, 10416.6666685129, -3.31554598677227e-05, 19027.7778012745, -9.36541505858714e-05, 0, 0, 0, 0, 4.75512966885251e-05, 0, 4.20119958794335e-05, -108796.296356214, 3.14856359981735e-05, 0, -63425.9258320772, 0, 0.00028139014436838, -785879.629634688, 0.000396254488908358, -6.71181388879633e-05, 7.75969312467229e-05, -39351.8518324275, 0, -740509.259151349, 0.000550900513348859, 3.73220905712826e-05, -63425.9260228539, 0.000105405936850699, 0.000389029977288728, -156250.000047998, 0.000365482099306817, 0.000443628603027872, 0, 0, 0, 0, -7.7914765271199e-05, 0, -2.58338360238587e-05, 115740.740849864, -6.43711441964883e-05, 0, -0.000177632838828899, 0, -0.00048742714885366, 1701388.88898882, -0.000912589112187953, 0.000155954766437751, -0.000257227146878544, 69444.4444011259, 0, 520833.333038915, 0, -0.00170499950271741, 3078703.7033973, -0.00159028057052975, 0.000143093459663258, 0.000273175808992667, -0.000361378514271989, -0.00108379216041413, 590277.777841011, -0.00122842875492178, -0.000280606337603387, 520833.333276405, -0.00125597129404606, -0.000465804672589427, -0.00193804348251828, -0.00040009536966612, 0, 0, 0, 0, 7.57522380345641e-05, 0, 4.55476311504618e-05, -46296.2964099013, 1.34422613508576e-05, 0, 0.000215964808898592, 0, 0.000324202880398884, -1782407.40762981, 0.000755013420702298, -0.000203206104199694, 0.000257586318408765, -46296.2963625779, 0, 0.000291416144404952, 0, 0.00157842210818417, -6481481.48111339, 0.00220483190376815, -0.000273779056266542, -0.000683494210774583, 0.000757956575941343, 0.000640699542775567, -1041666.66686631, 0.00108926437797755, 0, -1967592.5919443, 0.00245263064955622, 0.000512057629754816, -0.00120371974751138, 0.000814913871907793, 0.00272776968304408, -1967592.59224116, 0.00443883338823661, 0.000152661395583112, 0.00229246474896352, 0.00434451961881463, 0.00323704424767017, 0.00244564193086144, -0.000692396902888554, 0, 0, 0, 0, -3.98682321478292e-05, 0, -1.39569727542057e-05, 6.07082886444641e-05, -3.34646883476193e-07, 0, -0.000147601014308258, 0, -0.000184705111792332, 694444.444654047, -0.000302039726435932, 0.000116973997245149, -0.00012007194629066, 0.000145393181822676, 0, -0.000164680453344749, 0, -0.000869099607443231, 6481481.48130571, -0.00151626240836358, 0.000321199581521574, 0.000477959900895286, -0.000538389595523028, -0.000650250181887039, 694444.444878772, -0.000704948707566783, 0, -0.000419567641913317, 0, -0.0019399772878121, 3472222.22156002, -0.00239115288295865, -8.44732648086415e-06, 0.00103706845779231, -0.000858657047669504, -0.000526165589454421, 3472222.22230329, -0.00270472165688103, -0.00127818239838489, -0.00201196792330234, -0.000670920128912268, -0.000626132959355037, 0.000265356618310851, -0.00347152977469642, -0.00123143849222218, -0.000826849046357422, -0.00409551940545482, -0.00209423474123063, -0.00446131865541797, -0.00389291557333104, 0.00103311969454213, 0, 0, 0, 0, 8.69357776108457e-06, 0, 2.75307653879871e-06, -1.37520933061684e-05, -3.22894524280746e-06, 0, 4.22709359116929e-05, 0, 5.83680475706737e-05, -6.55880296794698e-05, 7.43322371886629e-05, -2.57422775090905e-05, 3.35768493531131e-05, -4.61619789605729e-05, 0, 4.92952756444875e-05, 0, 0.000218145608975578, -2314814.81480916, 0.000515814664322436, -0.000101102273121455, -0.000135319480732265, 0.000158818071592429, 0.000240411378099326, -0.000113315089969594, 0.000334132115271972, 0, 8.59506634073855e-05, 0, 0.000689181986452702, -2314814.81457194, 0.000883212543019125, -0.000156410720890741, -0.000348672820359815, 0.00052760653942497, 0.000536100959235966, -2314814.81501643, 0.00120021275451424, -0.000426441719679217, 0.000589985215041932, 0.000532497578795124, 0, 0.000246889829480996, 0.000981070968829597, 7.60338380361571e-05, -0.000573811345442252, 0.000290199666340441, 0.00108399002967194, -0.00024028512750313, 0.00127966683156752, -0.000115650757251997, -0.000888466921017528, 0.00158382737469893, -2.8556390128669e-05, 0.000460365459120673, 0.00109387213329448, 0.00145826576689036, 0.000222452420844241, 0.00256337326289413, 0.000547841137891384, 0.00359247053411686, 0.00189338402893462}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 0, -71.4285714333064, 0, 0, 0, 0, 785.714285717119, 0, -2.88832985396804e-07, 1821.42857162673, -3.04901270043402e-07, 71.4285714207435, 0, 0, 0, 0, -714.285713359339, 0, 1.81064954677412e-06, -17861.1111128852, 1.57084346332839e-06, 0, -19964.2857168052, 5.9510431071809e-06, -714.285715589299, 2.96963305135225e-06, -1749.99999950381, 0, 0, 0, 0, -6.63887958015387e-06, 0, -7.29845451079346e-06, 89027.7777838676, -3.21935306803095e-06, 0, 17500.0000043869, 0, -2.56000516283816e-05, 194722.222248582, -3.48107315956482e-05, 1.05568948329617e-05, -2.55495045656135e-06, 16111.1111006546, 1.69010288206991e-05, 17500.00000621, 4.05672843628333e-06, 0, 0, 0, 0, 2.07337837267735e-05, 0, 1.46671910944175e-06, -246527.777787761, 1.84996149788194e-05, 0, 2.53772897852659e-05, 0, 0.00010869280191621, -963194.444536038, 0.000109922462906422, -1.98147882122817e-05, 4.73123237886917e-05, -72916.6666383868, 0, -161111.111199467, 0.000197817668300416, -6.37838608237309e-05, -8.27174513568303e-05, 4.06667140104003e-06, -1.66735885708349e-05, -161111.111118088, 9.97660968835681e-05, -0.00024147519556822, 0, 0, 0, 0, -3.39058111674819e-05, 0, -4.49920945838706e-06, 381944.444452012, 1.55721365456756e-06, 0, -0.000107977327529931, 0, -0.000110356632323017, 2638888.88904268, -0.000114714921981867, 2.39093292690774e-05, -4.63949564791595e-05, 173611.111080959, 0, 0.000112536666041087, 0, -0.000211342783521529, 729166.666979401, -0.00050853115374839, 9.04626108195344e-05, 0.000228497337042676, -5.06415095750257e-05, 0.000165995792184535, 729166.666649995, -0.000301358538456459, 0.000153516032612469, 0.000282675575717196, -0.000186691081300197, 0.00064253719776417, 0.00105245944597128, 3.3025135680546e-05, 0, 0, 0, 0, 3.07385293379787e-05, 0, -1.53577612177137e-05, -307539.682539535, 6.38168753154392e-06, 0, 0.000167707155664856, 0, 0.000136311305256719, -4027777.7779105, 0.000163212200236914, -8.93058114546032e-06, 6.92214712753811e-05, -208333.333311323, 0, 7.2391730319568e-06, 0, 0.000374781710353371, -1736111.11156208, 0.000632486560091946, -6.20390113825961e-05, -0.000209595784766857, 0.000172965095064352, 0.000167719796749441, -1736111.11103466, 0.000918586052350363, 0, -0.000398931083043925, 0.000851936188149215, -0.000399979346022911, -0.000135856929306223, -9.33153975944967e-05, -0.000177616518024664, -0.000325830172094053, -0.000136682024057985, -0.00111496627845421, -0.000969122417340914, -0.00107420436661867, -0.00216664515824045, -0.00158756971886318, 0.000208686619862305, 0, 0, 0, 0, -1.43790391246901e-05, 0, 6.80331924311847e-06, 99206.3492037876, -2.86746178284576e-07, 0, -0.000120967285884206, 0, -5.10090260857379e-05, 3174603.17465079, -0.00011930894103058, 4.90313923250343e-06, -6.93599970718216e-05, 99206.3491725479, 0, -9.54663174434049e-05, 0, -0.000246333684085962, 2083333.33364093, -0.000465726647754956, -1.0510361083548e-05, 0.00015691033098062, -0.000227304648062462, -8.47981977245436e-05, 2083333.33319179, -0.000660496328111878, 0, 0.000191077521883697, 0, -0.00061184847819314, 0.000416890316731574, -0.000782461501813294, 0.000175814274409315, 0.000265253203079074, -3.60920738859418e-05, -0.000234686600949914, 0.000112168086486192, -0.00107531741792208, 0.00148088493834206, 0.000396556329695522, -0.00117133086500268, 0.000184705685796432, 0.000361290235347591, -4.01061535421302e-05, 0.00183283798923564, 0.00143044662299972, 0.00230365195833244, 0.001595440333803, 0.00180003828880061, 0.000855744728959698, 0.000577376182663636, 0, 0, 0, 0, 2.67333872811157e-06, 0, -2.71507394437122e-06, 1.15212891211919e-06, -1.10993815134872e-07, 0, 3.34166469607306e-05, 0, 3.62614215345361e-07, -992063.492068662, 1.83994420748223e-05, -1.99133220621406e-06, 1.1773756045103e-05, 3.60417858233335e-06, 0, 4.80371156959328e-05, 0, 5.58039611797605e-05, -992063.492139327, 6.89299936677874e-05, 1.29837446204166e-05, -4.73026043199602e-05, 7.66528462591785e-05, 2.93569090186399e-06, -992063.492049088, 0.000145771311438423, 0, -1.92672056892395e-05, 0, 0.000110020509584522, -0.000152307529016553, 0.000248721843684106, 2.63838330271255e-05, -0.000111430098381007, -4.74878325256688e-06, -0.000104683652618077, 4.56765534939199e-05, 0.000411803941346491, -8.0073501677765e-05, 4.46966639377102e-05, 0.000344119207637066, 0, -0.000120457630222781, 0.000264915735704901, -5.51487890984542e-05, -3.87819129071553e-05, -3.41231565919884e-05, -0.000118260144318402, -4.2551205508826e-05, 0.000458920187756019, -0.000339527540520229, -0.000248313623499469, -0.000632250501114339, -0.000286156542739366, 0.000377385095414615, 0.000486530527133658, -0.00105968806217471, -0.000321325156318747, -0.00124065669589986, -0.00141428982270134, -0.00260127914578477, -0.00095413472834805}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -1.2407550426025e-07, 642.857142906218, 0, 0, 0, 0, -1357.14285748279, 0, 2.1180331272576e-06, -17107.1428584958, 2.71194725605511e-06, -1357.14285635496, 0, 0, 0, 0, 714.285712878565, 0, -1.15378239818881e-05, 178964.285726371, -1.97990701918233e-05, 0, 34678.571442149, -4.23649659849616e-05, 1428.57142200243, 714.285708896849, 34678.5714127992, 0, 0, 0, 0, 1.65106765028628e-05, 0, 2.39960278692263e-05, -979861.111161623, 7.21426262944023e-05, 0, -17500.0000278786, 0, 0.000235987159497587, -341111.111237912, 0.000286853886390342, 3.36484882332826e-05, 1.12345398955033e-05, -341111.110992499, -9.46074956970306e-06, -34999.9999357745, -17499.9998413773, 0, 0, 0, 0, -5.47225476417251e-05, 0, -2.71119968673176e-05, 3109027.77789323, -0.000153070578827191, 0, -5.61329658688153e-05, 0, -0.000637258080115584, 1707638.88938119, -0.000973585562580687, -0.000145365472839581, 5.38605375337218e-06, 1707638.88850436, 0, 161111.111489415, -0.00136733583863638, 4.02558583722249e-05, -0.000159977234866947, 0.000221057744020881, -0.000437314691440597, 322222.221730159, 161111.109909507, -4.6204261898936e-05, 0, 0, 0, 0, 9.21116593923468e-05, 0, -2.02327446824992e-05, -5902777.77793041, 0.000198091749735027, 0, 0.000293896270155361, 0, 0.000952274148870906, -4756944.44542437, 0.00169945888496183, 0.000335247707185201, -3.16353698995676e-05, -4756944.44381406, 0, -0.000364446233852442, 0, 0.00297698912885732, -729166.668047306, 0.00342128883651194, 0.000209317031872858, 0.00035662040722917, -0.000461720289345481, 0.000632546689117421, -1458333.33187227, -729166.66269733, -0.00079404846726698, 0.000989939855473936, 0.00302912458831413, -0.000900108350603049, -0.00128271875758891, 0.0023331369420793, 0, 0, 0, 0, -8.58318068384007e-05, 0, 4.25571259026087e-05, 6587301.58741765, -0.00014247556846915, 0, -0.000455584366500862, 0, -0.000829902766760353, 7430555.55660025, -0.00174978390909758, -0.000433231764046166, 6.95203878890294e-05, 7430555.55500595, 0, -8.93103609698453e-05, 0, -0.00348021737545183, 1736111.11337696, -0.00467254383792008, -0.000820500781071602, -0.000547835537216518, 0.00032785294167461, -0.000594114626609184, 3472222.22022837, 1736111.10492213, 0, 0.00146647383619701, -0.0057699646678196, 0.00118356156667409, -0.00095085522225181, 0.00185331530211718, -0.00341399578626293, -0.00259354067223576, -0.00823049392160419, 0.00118290111983187, 0.00353955701925181, -0.00798576917734374, 0.00600787772940357, -0.000475605502641465, -0.00147980024117095, 0, 0, 0, 0, 4.19427454165703e-05, 0, -3.93218819682124e-05, -3968253.96830109, 5.43861252952249e-05, 0, 0.000320919704655464, 0, 0.000354043546612281, -6051587.30215198, 0.000932670947237512, 0.000279339180182565, -3.59090214114602e-05, -6051587.30138151, 0, 0.000343895471860797, 0, 0.00210079300511801, -2083333.33507631, 0.00312625781947037, 0.000919007032042703, 0.000477630932941062, 1.86389910023192e-05, 0.000377810512670832, -4166666.66549105, -2083333.32847629, 0, -0.000555279837491721, 0, 0.00447634612671469, -0.00182251647407564, 0.00503313388638687, -0.000144485555855739, 0.000510634685715406, -0.0010958032113446, 0.00200470279459855, 0.00285677631612164, 0.0087857569197002, 7.20956616302515e-05, -0.00293248160153109, 0.0090094625172008, -0.00197604555857159, 0.00195809322185301, 0.00651000427335779, -0.00572613218179133, -0.000813849655395019, -0.00828899469573894, -0.00290470376860992, 0.00613960802056662, 0.00743467466485203, -0.00579397896686849, 0, 0, 0, 0, -8.35412181137151e-06, 0, 1.17479402334839e-05, 992063.492071448, -9.96087672576881e-06, 0, -8.7499502347075e-05, 0, -6.81264066388862e-05, 1984126.98424743, -0.000197442171632139, -7.04489836070669e-05, 2.19726939395126e-05, 1984126.98410812, 0, -0.0001536137708795, 0, -0.000536598859359252, 992063.492573522, -0.000872381005914896, -0.000328748394819793, -0.000159677930290436, -9.1725170854124e-05, -0.00010606004937365, 1984126.98390899, 992063.490705987, 0, 6.80368988719284e-06, 0, -0.00143569709131201, 0.000750903546358357, -0.00172182248448503, -0.000257240986843643, -0.000125507891163428, 0.000106801475050127, -0.000705280937490806, -0.000909215231347035, -0.00350551037090358, -0.000303047762378856, 0.00065909273183617, -0.00339110892923443, 0, 0.000464465881091003, -0.00197992103235776, 0.000572656031805095, -0.000461356021591967, 0.000724294383490135, -0.00189646365050242, -0.00120122409323746, -0.00338729622234603, 0.00175849990628615, 0.000633972685041428, 0.00341393160448678, 0.00195005260649397, -0.00374375822550685, -0.00544384537472905, 0.00436288693872943, -0.00115763073565634, -0.000965518739815643, 0.00326120970593474, 0.00545505254073233, -0.00141589790646125}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 7.49774007736482e-07, -2000.00000012072, 0, 0, 0, 0, 6722.22222250367, 0, -1.39629296043697e-05, 52388.8888927979, -1.54764768938738e-05, 6722.22221765105, 0, 0, 0, 0, -7499.99998927532, 0, 8.18565553439343e-05, -535990.740777175, 0.000103321716528012, 0, -168490.740775194, 0.000280246538065106, -14999.9999793162, -7499.99993938938, -168490.740661115, 0, 0, 0, 0, 2777.77769924815, 0, -0.000207188547550582, 2851111.1112671, -0.000339742205295991, 0, 179583.333368492, 0, -0.00161735056131724, 1611250.00035371, -0.0017716479655442, 8333.33326611796, 2777.7774733082, 1611249.99942427, 8333.333262674, 359166.666447555, 179583.332317763, 0, 0, 0, 0, 0.000240337339660931, 0, 0.00025408927836317, -8731250.00036728, 0.000631470826740589, 0, -63425.9255842932, 0, 0.00455178205383419, -7764351.85325058, 0.00551511040820457, 0.000279854840046235, 0.000899460227483617, -7764351.84999225, 0, -1596527.77873628, 0.00888183029241384, -0.000119613525109284, -190277.777722396, 0.000185114620498303, -63425.9222198365, -3193055.55348972, -1596527.77091279, -190277.777299867, 0, 0, 0, 0, -0.000397263991673943, 0, -4.95167927773045e-06, 15902777.7782824, -0.000655839784687451, 0, -0.00124770454215858, 0, -0.00688828749817417, 20590277.7805996, -0.00926861176134354, -0.000720496558504845, -0.00159909825768299, 20590277.7748119, 0, 520833.334012366, 0, -0.0203527232070558, 6875000.00364719, -0.0221112759810875, -0.000362217161453498, 0.000615443105864453, -0.00169542436892804, -0.00724708387474215, 13749999.9933052, 6874999.97857486, 0.00160248006037435, 1562499.99692666, 520833.313541741, 0.00502860213727839, 0.00310241559122414, 1562499.9890073, 0, 0, 0, 0, 0.000369796318596201, 0, -0.00023014534495616, -16944444.4448468, 0.000352504777811295, 0, 0.00183214512030156, 0, 0.00593568458882136, -30254629.6326938, 0.0089320237323868, 0.0010182213702894, 0.00158196056387043, -30254629.6273289, 0, 0.000778960922810372, 0, 0.0243498003789189, -15277777.7838099, 0.0287597240213748, 0.0023237768269685, -0.00105729494332313, 0.00492297805224194, 0.00872216410970025, -30555555.5461845, -15277777.7449892, 0, -1967592.59614096, 0.0385498518771783, -0.00343880762323693, 0.00060191216509874, -0.00149001588421265, 0.0266646214920036, -5902777.7667245, -1967592.54433063, -0.00979243110523506, -0.00688985626015443, -5902777.74031032, -0.0233667184194209, 0.000288582221589659, 0.0089323588442977, 0, 0, 0, 0, -0.00018151622992087, 0, 0.000231286006732241, 9722222.22239511, -7.06181052982669e-05, 0, -0.00128027036348873, 0, -0.00258664516143248, 22916666.6683621, -0.00448640120279595, -0.000689821199680325, -0.000882296855118846, 22916666.6659802, 0, -0.00140845795277815, 0, -0.0148206181854546, 16666666.6713496, -0.0188736742348833, -0.00286884008197255, 0.000560909470301524, -0.00511095075968495, -0.00585820318330257, 33333333.3277554, 16666666.6422443, 0, 0.00109609651456409, 0, -0.0310483513889391, 3472222.22670745, -0.0333496999431357, 0.00053008499226619, 0.00201884693095997, -0.00257520282276408, -0.017747537851401, 10416666.6541414, 3472222.17108853, 0.00534595059222579, 0.00250336744414144, 10416666.6222534, 0.00402696779831112, -0.00651456562657562, -0.0427232082660485, 0.0264213590097043, 0.00958563466440632, 0.0275009511605438, 0.00771114388205735, -0.0256900880831491, -0.0378529702089653, 0.0248574374835336, 0, 0, 0, 0, 3.63644405646937e-05, 0, -7.20741556758481e-05, -2314814.81484595, -1.2586330784777e-06, 0, 0.000350240567581721, 0, 0.000462986460842002, -6944444.44481668, 0.000894530093860822, 0.000178868509144526, 0.000144109715823429, -6944444.44446023, 0, 0.000588871628451399, 0, 0.00366317253426466, -6944444.44583387, 0.00497886018218736, 0.00109262275386791, -8.89045406013273e-05, 0.00185636287863357, 0.0015792723575153, -13888888.8878531, -6944444.43761206, 0, 0.000122537344853945, 0, 0.009713180018939, -2314814.81666818, 0.0110430769877591, 0.000874066159684718, -0.00114447262646548, 0.00246383627843675, 0.00553349778680374, -6944444.44019169, -2314814.79495214, -0.000573567805440137, 0.00123921613865925, -6944444.42736927, 0, -0.00104591014144307, 0.013502624187342, -0.00151218151683915, 0.000603305276346359, -0.000538319274514527, 0.013707911819161, 0.00483742460382078, 0.0224266724623546, -0.00858497093981308, -0.0060528569123226, -0.00877564765912882, -0.00578787011612945, 0.0185176752968145, 0.0277778434210828, -0.0163847267615464, 0.00359640069944376, 0.00681079919756959, -0.0183127274037363, -0.023025178874158, 0.00531954885690152}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -2.72343845183042e-06, 4200.00000015906, 0, 0, 0, 0, -20116.6666648584, 0, 5.35412467460085e-05, -107616.666674118, 5.39781851560542e-05, -20116.6666530805, 0, 0, 0, 0, 35916.6666255208, 0, -0.000327116577556087, 1067513.88896363, -0.000337394215267039, 0, 490930.555603648, -0.00105568724254765, 71833.3333006094, 35916.666415209, 490930.555347235, 0, 0, 0, 0, -28333.3330917178, 0, 0.000894792353672693, -5460277.77811732, 0.00101971048375639, 0, -833263.888819253, 0, 0.00634631794527736, -4515208.33393345, 0.00634313263536917, -85000.0000531792, -28333.331845008, -4515208.33190127, -84999.9993894823, -1666527.77753961, -833263.885463267, 0, 0, 0, 0, 8333.33262859133, 0, -0.00113782967010731, 15968402.7786364, -0.0016718526854386, 0, 623611.110021266, 0, -0.0184748749636371, 20661805.55806, -0.0186704572898922, 33333.333649773, 8333.32863158028, 20661805.5510725, 0, 7058680.55680793, -0.0331359341865653, 49999.9994748976, 1870833.33529615, 33333.3290812329, 623611.096395815, 14117361.1071984, 7058680.53432106, 1870833.33126123, 0, 0, 0, 0, 0.00116521654466096, 0, 0.000185623142985574, -27652777.7790568, 0.00139933698655158, 0, -173611.107869106, 0, 0.029049420391922, -51402777.7831256, 0.0304354763466532, -0.000432965134136861, 0.00837823274637689, -51402777.7711726, 0, -4826388.88897396, 0, 0.0796384752794071, -28402777.7830809, 0.0795274914050549, 0.000845336798192851, -694444.452870718, 0.0162842680069349, -173611.078505065, -56805555.5409897, -28402777.7138417, 0.00498996358811986, -14479166.6642306, -4826388.82080762, -1041666.68097691, -694444.441532406, -14479166.6386165, 0, 0, 0, 0, -0.00110448802422222, 0, 0.0009847238508405, 27986111.1122223, -0.00039012828337589, 0, -0.00458958460927762, 0, -0.0254413616024122, 70208333.3395632, -0.0282591059679189, 0.000136128009134361, -0.00816088157190804, 70208333.3292862, 0, 1215277.77509242, 0, -0.0991594996199149, 57673611.1202787, -0.0998005313211837, -0.00379921858698897, 0.0138037649625881, -0.0309220992048109, -0.0413164277191276, 115347222.201911, 57673611.0148709, 0, 16666666.6705888, -0.144883517252291, -0.00201320081159466, 4861111.12177583, -0.0206042090825738, 1215277.67568217, 49999999.9779544, 16666666.5120134, 0.0344597703143839, -0.0118188752049352, 49999999.9016007, 0.040548118610812, 7291666.67263377, 4861111.08107891, 0, 0, 0, 0, 0.000556776492468042, 0, -0.00101391640868943, -15277777.7782982, -0.000181536131367672, 0, 0.00321774131779782, 0, 0.0113133986759026, -49305555.559293, 0.0138337422532483, 8.48584788475492e-05, 0.00408626029305421, -49305555.5553568, 0, 0.00341874318172718, 0, 0.062033523541371, -56250000.0075308, 0.0636946881462249, 0.00491173572166442, -0.0100561113395071, 0.0266325953390238, 0.0273863391501196, -112499999.988896, -56249999.9296453, 0, -3472222.2226181, 0, 0.121536027790321, -25694444.4497323, 0.120986669622139, 0.00357557840504645, -0.0186895615295291, 0.0319083047186193, 0.076430344203384, -77083333.3059455, -25694444.2848383, -0.0278144373351345, 0.0253368829969506, -77083333.2140038, 0.00937965118898037, -13888888.8806765, -3472222.07462273, -0.0558843354561212, -0.0410090747073356, -0.0287253010081088, -0.00240563236192879, -20833333.2729701, -13888888.7836641, -0.0569110949808211, 0, 0, 0, 0, -0.000115263448522051, 0, 0.000315814736747784, 3472222.22232383, 0.000102288877927373, 0, -0.000892617038499986, 0, -0.0019796168037052, 13888888.889787, -0.00271660632439673, -4.06462254419824e-05, -0.000716490768496194, 13888888.8894227, 0, -0.00136562697794642, 0, -0.01542366225573, 20833333.3357287, -0.0163413950137578, -0.00194649186414225, 0.00280956306811097, -0.00854510392570128, -0.00711115808563288, 41666666.6651631, 20833333.3136223, 0, -0.000561674333374127, 0, -0.0388381899357245, 13888888.8911648, -0.0385163282454316, -0.00289482979437922, 0.00838822673839953, -0.0156686727913132, -0.0239695997783286, 41666666.6573498, 13888888.8279976, 0.00750046386115708, -0.0147303789735214, 41666666.6198056, 0, 3472222.22309591, -0.0507088928639994, -0.0017758713626134, 0.00301327456985341, -0.00907668431110842, -0.0521820270853644, 13888888.8789197, 3472222.14733059, 0.0208075278820737, 0.0251345293103786, -0.00176385427280772, 0.00510389068504185, 20833333.2845816, 13888888.8121103, 0.0231754223664376, -0.00452150807775988, -0.0225658460912546, 0.0479582097077766, 0.0470518465705697, -0.0124054295108651}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 5.98949709820233e-06, -6300.00000001343, 0, 0, 0, 0, 40674.9999926495, 0, -0.000120056781138409, 156175.000008451, -0.000116393215224062, 40674.9999883703, 0, 0, 0, 0, -104166.666572401, 0, 0.000745041792461226, -1479875.00009611, 0.000696054218578187, 0, -954041.666685857, 0.0023411945315055, -208333.333440638, -104166.666285094, -954041.666495989, 0, 0, 0, 0, 132291.666177913, 0, -0.00206998268085581, 7159513.88936685, -0.00201587643735996, 0, 2306597.22186017, 0, -0.0144754143611748, 8283819.44498337, -0.0134825509449999, 396875.00096539, 132291.664171405, 8283819.44305597, 396874.999683908, 4613194.44620146, 2306597.21718648, 0, 0, 0, 0, -83333.3319471169, 0, 0.00247517617109509, -19704513.8902139, 0.00317567326623401, 0, -2758680.55343055, 0, 0.0434204646263691, -35221875.0026062, 0.0381110850197597, -333333.336632481, -83333.3252284433, -35221874.9954691, 0, -18192708.3335245, 0.0721143033246739, -500000.001719402, -8276041.67919586, -333333.326623177, -2758680.53183446, -36385416.6694144, -18192708.3005992, -8276041.67560023, 0, 0, 0, 0, 20833.3310182976, 0, 0.000148041523155538, 32138888.8910363, -0.00255086422596719, 0, 1631944.43906397, 0, -0.0702985877875209, 80625000.0063789, -0.0609632028769846, 104166.672543502, 20833.3192953259, 80624999.9942098, 0, 19548611.1086107, 0, -0.180290030214309, 66423611.1136505, -0.16517418406594, 208333.337549996, 6527777.81419389, 104166.635918743, 1631944.38812351, 132847222.217579, 66423611.0079584, 208333.323567522, 58645833.3590146, 19548611.0014111, 9791666.72924475, 6527777.803351, 58645833.3322896, 0, 0, 0, 0, 0.00225203946012255, 0, -0.00333739008712189, -30763888.8908923, 0.000597012158396713, 0, -381944.437072737, 0, 0.0623547890087405, -101388888.897389, 0.0565450263343376, -0.00581212736505528, 0.0124749946800165, -101388888.887886, 0, -10243055.550119, 0, 0.232845988736988, -119965277.783625, 0.19915984459774, -0.00116215698737073, -1909722.2755421, 0.0597544270307583, -381944.370898196, -239930555.545961, -119965277.616758, 0, -59201388.8869754, 0.312210481007442, 0.00823370200215227, -40972222.2908661, 0.0497212885055545, -10243055.3821233, -177604166.673859, -59201388.6367446, -3819444.56401356, -1909722.23738639, -177604166.567662, -3819444.56787327, -61458333.4520269, -40972222.2776773, 0, 0, 0, 0, -0.00117406569790162, 0, 0.00312373731576182, 15972222.2232145, 0.000451119660257277, 0, -0.00522243460968334, 0, -0.0278107214060499, 65972222.2279839, -0.028300712708201, 0.00307667587525574, -0.00503779727323167, 65972222.2254627, 0, 2083333.32818135, 0, -0.149746784411617, 104166666.672897, -0.124004199388239, -0.00287166015043638, 0.0381337858027163, -0.0511863672084331, -0.0479988322339626, 208333333.33152, 104166666.545228, 0, 24305555.5521655, 0, -0.272777920254048, 76388888.8877837, -0.248638892437146, -0.00941842636850582, 10416666.7428824, -0.0728773565199873, 2083333.19402727, 229166666.654723, 76388888.6232826, 0.0985091405107185, -0.0233683883513069, 229166666.508327, -0.0336249808409124, 97222222.2550576, 24305555.3088249, 0.140602870208014, 20833333.5186076, 0.0736583233024968, 10416666.7736055, 145833333.340414, 97222222.1546969, 20833333.5725005, 0, 0, 0, 0, 0.000252700921810709, 0, -0.000942295832304815, -3472222.22242381, -0.000233504052167222, 0, 0.00148247610801524, 0, 0.0047251421422535, -17361111.1126536, 0.00583051832250689, -0.000693560277372663, 0.00050932935383946, -17361111.1129536, 0, 0.00194728020413378, 0, 0.0378187818683633, -34722222.2247021, 0.0315245418006462, 0.00169658496242562, -0.0107040528154705, 0.0161301885411026, 0.0117625836715917, -69444444.4468249, -34722222.1869988, 0, -3472222.22088289, 0, 0.089918117791686, -34722222.2222051, 0.0759411327416992, 0.00558768198507467, -0.0297764227568651, 0.0338572131653386, 0.0452324167022793, -104166666.661982, -34722222.1187876, -0.0297816364318755, 0.0201630683863051, -104166666.597331, 0, -17361111.1095563, 0.108313232068033, 0.00975749755286287, -17361111.1357864, 0.0258192108619388, -3472222.12970176, -69444444.4440219, -17361110.9863744, -0.0513956831960386, -0.0888281328267362, 0.0035412219416638, -0.0454913879875722, -104166666.620046, -69444444.351295, -0.0413761316445226, -34722222.2706811, -17361111.1329629, -0.14696103562501, -0.140409485997213, -34722222.2817012}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -7.46056311426858e-06, 6999.99999963308, 0, 0, 0, 0, -59194.4444306905, 0, 0.000151573073151654, -164194.444446728, 0.000144537735390616, -59194.4444762129, 0, 0, 0, 0, 206129.629494392, 0, -0.00093758697506842, 1444046.29634649, -0.000820807058148362, 0, 1300175.92585746, -0.00298101505957049, 412259.259882754, 206129.62972736, 1300175.92631324, 0, 0, 0, 0, -378472.221585929, 0, 0.00251114925688333, -6430138.88920897, 0.00228501603453171, 0, -4227361.11035255, 0, 0.0189020399165453, -10279027.7776769, 0.0163079910837573, -1135416.67009152, -378472.222274299, -10279027.7790808, -1135416.67094289, -8454722.22998386, -4227361.11106122, 0, 0, 0, 0, 386574.07231793, 0, -0.00232617252495762, 16370138.8899454, -0.00360041483428718, 0, 7223379.62680872, 0, -0.0586025679918169, 39190509.2597473, -0.0440974092599233, 1546296.30614912, 386574.073470955, 39190509.2615351, 0, 29657175.9233015, -0.0888023480262942, 2319444.4621096, 21670138.9260366, 1546296.3076389, 7223379.62785582, 59314351.8785341, 29657175.9115474, 21670138.9444643, 0, 0, 0, 0, -208333.330376625, 0, -0.00248636831931217, -24930555.5574789, 0.00310099117973156, 0, -6840277.77187725, 0, 0.0974353128293812, -81006944.4474723, 0.0698517899128201, -1041666.68298581, -208333.333035398, -81006944.4488498, 0, -44201388.8817595, 0, 0.231438430857954, -93645833.3269958, 0.189874855892285, -2083333.36629079, -27361111.2017258, -1041666.67253937, -6840277.76933235, -187291666.709231, -93645833.260237, -2083333.36869227, -132604166.770962, -44201388.8503191, -41041666.8684661, -27361111.2925729, -132604166.812548, 0, 0, 0, 0, 46296.2933504643, 0, 0.00749130786550733, 22500000.0019263, -0.00105502085373391, 0, 3402777.77018496, 0, -0.0874359674832865, 93217592.5986108, -0.0667902288114832, 277777.793465745, 46296.2994997353, 93217592.6010193, 0, 35648148.1398286, 0, -0.311338955517124, 147916666.659813, -0.21656247137634, 694444.474673843, 17013889.0089279, 277777.758819768, 3402777.76137856, 295833333.376882, 147916666.52457, 0, 109490740.725584, -0.37308881645593, 925925.972068416, 142592592.791595, 694444.456715616, 35648148.0936535, 328472222.342383, 109490740.595751, 34027778.0986935, 17013889.1551355, 328472222.319554, 34027778.2157024, 213888889.355116, 142592593.038171, 0, 0, 0, 0, 0.0015831404271214, 0, -0.00619225303842281, -11111111.1121007, -0.000296480240182239, 0, -694444.43904964, 0, 0.0388050299049896, -56250000.0050906, 0.0357279917440183, -0.00816906598717152, -0.00514319472501361, -56250000.008388, 0, -14583333.3279029, 0, 0.206651709607509, -114583333.331422, 0.131475273538425, -0.0128491923817438, -4166666.74798282, 0.0272028324077501, -694444.43378986, -229166666.69844, -114583333.210588, 0, -62499999.9890145, 0, 0.342048804118521, -118055555.539899, 0.276093289167049, -0.0233814064824524, -72916666.8514391, 0.0316064194657581, -14583333.2819923, -354166666.735468, -118055555.360918, -10416666.9094323, -4166666.83915538, -354166666.6256, -0.0126010852826683, -250000000.155568, -62499999.8853764, -13888889.301415, -145833333.871601, -10416667.0515119, -72916667.2136692, -375000000.331954, -250000000.299624, -145833334.128665, 0, 0, 0, 0, -0.000352269407495811, 0, 0.00179164622773175, 2314814.81501747, 0.000235487148939991, 0, -0.00157310986139084, 0, -0.00632525473402927, 13888888.8904505, -0.00809022444265739, 0.00178900551431031, 0.00229828296185452, 13888888.8920258, 0, 2314814.81306393, 0, -0.0533233107575653, 34722222.2228451, -0.0337118826977499, 0.0018898702742396, 0.0220653862737205, -0.0101997729701606, -0.00112506931435303, 69444444.456105, 34722222.1827008, 0, 13888888.8862648, 0, -0.117374305965017, 46296296.2909658, -0.0793388244786481, 0.00234544411343367, 13888888.954307, -0.023206612748295, 2314814.79433052, 138888888.90822, 46296296.2078469, 0.0707892640372919, 0.039695974418199, 138888888.844656, 0, 34722222.2157355, -0.12624584876749, 0.00321555174000156, 69444444.519996, -0.0115917954393492, 13888888.8463094, 138888888.930108, 34722222.1499928, 0.137736634990443, 34722222.4426124, 0.109569624091869, 13888889.1104435, 208333333.389993, 138888888.911769, 0.163748980184959, 138888889.107891, 69444444.6860476, 46296296.699712, 34722222.6541883, 138888889.218088}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 5.03928379806622e-06, -5999.99999943258, 0, 0, 0, 0, 65738.0952242549, 0, -0.000101838345086163, 125738.095233342, -0.000101616938531382, 65738.0953206155, 0, 0, 0, 0, -303527.777661749, 0, 0.000605413884200861, -960908.730140926, 0.000549881440289219, 0, -1264436.50781216, 0.00211707239456966, -607055.556742278, -303527.778626342, -1264436.50885872, 0, 0, 0, 0, 766111.110608575, 0, -0.00140265271850174, 3801388.88890329, -0.00154859852022464, 0, 5333611.1102728, 0, -0.0137671753713205, 8368888.88820736, -0.010762806543358, 2298333.33935151, 766111.115537753, 8368888.89239282, 2298333.34403692, 10667222.235347, 5333611.11807282, 0, 0, 0, 0, -1142361.10976174, 0, 9.16772793664871e-05, -8803472.22250372, 0.00283757753817921, 0, -12230555.553186, 0, 0.0445984132658434, -27552777.7759958, 0.0275692572312448, -4569444.46057049, -1142361.12387949, -27552777.7843355, 0, -29837499.9956296, 0.058964044772368, -6854166.70828291, -36691666.7253376, -4569444.48537092, -12230555.5876871, -59675000.0470586, -29837500.0126381, -36691666.7696331, 0, 0, 0, 0, 1006944.44215278, 0, 0.00540530556586574, 12430555.5562991, -0.00343470624575631, 0, 16458333.3293366, 0, -0.0767333851366011, 50729166.6651304, -0.0447375154351406, 5034722.24723679, 1006944.46769742, 50729166.6742676, 0, 55763888.8797995, 0, -0.163103802190618, 78611111.0985413, -0.111127429247798, 10069444.5236411, 65833333.466538, 5034722.29851274, 16458333.405518, 157222222.303991, 78611111.0922171, 10069444.5646388, 167291666.840767, 55763888.9586599, 98750000.3468519, 65833333.6820666, 167291666.949507, 0, 0, 0, 0, -486111.10877008, 0, -0.00996698002668149, -10555555.5564116, 0.00245344255023296, 0, -12986111.1064648, 0, 0.0699290008482791, -53263888.8898408, 0.047578071914314, -2916666.68926343, -486111.138293398, -53263888.8959607, 0, -57638888.8807097, 0, 0.232698910578807, -107986111.094596, 0.114556603094891, -7291666.7464644, -64930555.7203337, -2916666.73950363, -12986111.197818, -215972222.305715, -107986111.025353, 0, -110416666.643921, 0.230546784433752, -9722222.38761059, -230555555.86249, -7291666.82455457, -57638889.0258403, -331250000.226747, -110416666.699339, -129861111.641998, -64930556.1119793, -331250000.292684, -129861111.918646, -345833334.127068, -230555556.343749, 0, 0, 0, 0, 99206.3479103012, 0, 0.00742896370223166, 4960317.46077336, -0.000817860809543067, 0, 5555555.55224955, 0, -0.0307832844195851, 29861111.1132453, -0.0293831590443067, 694444.455477061, 99206.3675021266, 29861111.1162931, 0, 31249999.9963609, 0, -0.161567144281928, 74999999.9909537, -0.0681490494233478, 2083333.37426249, 33333333.4386517, 694444.478822824, 5555555.61296229, 150000000.052009, 74999999.9044831, 0, 76388888.8740085, 0, -0.228677703143146, 100694444.420352, -0.146326789941626, 3472222.32963263, 156250000.26291, 2083333.42392605, 31250000.1066717, 302083333.479418, 100694444.392004, 83333333.7204127, 33333333.7531232, 302083333.430112, 3472222.36627265, 305555555.824379, 76388888.9887652, 111111111.873292, 312500000.867911, 83333334.1321289, 156250000.921872, 458333333.985243, 305555556.173316, 312500001.316442, 0, 0, 0, 0, 0.000297035704092121, 0, -0.00206432993417539, -992063.492153476, 5.74850909042538e-05, 0, -992063.491055625, 0, 0.00474663039729757, -6944444.44529802, 0.00769105365118639, -0.00224936036606168, -0.00529378541444109, -6944444.44629308, 0, -6944444.443609, 0, 0.0430653707513801, -20833333.331893, 0.0188397723885812, -0.00843698929960491, -6944444.47163898, -0.006381445572666, -992063.509064983, -41666666.6820889, -20833333.2979302, 0, -20833333.3302284, 0, 0.0838182506569989, -34722222.2137577, 0.0365973038199717, -0.0263588851998272, -41666666.7537521, -0.016854305202522, -6944444.47424438, -104166666.707535, -34722222.1766383, -20833333.4425129, -6944444.56654325, -104166666.660246, 0, -34722222.2131111, 0.073045562726343, -0.0479109559349157, -104166666.784338, -0.0413553144199347, -20833333.3790472, -138888888.971911, -34722222.2396214, -34722222.4788892, -104166667.003207, -20833333.6143266, -41666667.0404723, -208333333.504486, -138888889.024972, -34722222.577014, -208333333.718561, -104166667.075756, -138888889.549021, -104166667.371372, -208333333.903907}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -6.33542037096334e-06, 5249.99999876658, 0, 0, 0, 0, -96895.8333091344, 0, 0.000119962602067629, -96895.8333175284, 0.000110559205344168, -44395.8334414321, 0, 0, 0, 0, 598555.555371567, 0, -0.000845185775433326, 598555.555449694, -0.000665138513908899, 0, 1722111.11085422, -0.00189694001093278, 753152.779592826, 154597.223239683, 753152.778875049, 0, 0, 0, 0, -1829826.38812869, 0, 0.00291907387836491, -1829826.38856402, 0.00250423960712425, 0, -9929062.49859199, 0, 0.0125999283382786, -9929062.49832808, 0.00865875902465562, -3943506.95414128, -283854.171348541, -3943506.9476706, -2397534.73769476, -12326597.2415381, -2397534.72987921, 0, 0, 0, 0, 3128472.22018972, 0, -0.00504520012950853, 3128472.22178934, -0.00653023654952334, 0, 27973611.1078585, 0, -0.0440961706300768, 27973611.1059956, -0.0225881380806921, 9675347.24850653, 289930.566785667, 9675347.22426867, 0, 49690277.7691845, -0.0445276486145443, 10255208.3992608, 59945486.2030637, 3998263.94574679, 3998263.92268189, 59945486.1800366, 10255208.3399575, 35970139.0259516, 0, 0, 0, 0, -3055555.55202498, 0, 0.00364894156874719, -3055555.55539119, 0.0109127768867464, 0, -43663194.4400768, 0, 0.0835186688229683, -43663194.4375564, 0.0423289227210024, -12378472.2625751, -156250.016116368, -12378472.2159898, 0, -115711805.539847, 0, 0.129850701461962, -115711805.529383, 0.0721370642186545, -18958333.4674663, -134670139.098479, -3680555.65842905, -3680555.62559977, -134670139.0085, -18958333.2778236, -13159722.4076326, -244583333.61491, -13159722.2810984, -142031250.511139, -54704861.5525886, -142031250.34807, 0, 0, 0, 0, 1597222.21849925, 0, 0.00060089437933678, 1597222.22234102, -0.0107402904936946, 0, 38576388.8843357, 0, -0.0858037920529904, 38576388.8857119, -0.0534487510619231, 8020833.36898397, 34722.2372117589, 8020833.3207266, 0, 139930555.544186, 0, -0.198628201221606, 139930555.519703, -0.0715932329574116, 16145833.4772604, 156076389.141704, 1770833.42855222, 1770833.40445993, 156076389.007224, 16145833.1723807, 0, 205902777.734675, -0.151247007015888, 16319444.7174181, 428125000.501576, 8333333.57357338, 8333333.46352929, 428125000.391121, 16319444.4174424, 238541667.495713, 45659722.8993522, 238541666.982937, 164930556.711532, 444791667.782624, 164930556.459851, 0, 0, 0, 0, -347222.220093799, 0, -0.00229618275143598, -347222.222308627, 0.00562338705092258, 0, -18055555.5521285, 0, 0.044807007552093, -18055555.5563627, 0.0375860307362214, -2083333.35023033, -0.00873078131050041, -2083333.32494153, 0, -85416666.6638788, 0, 0.147938346086824, -85416666.6454891, 0.0484669520024698, -5208333.41198767, -90625000.1552427, -347222.265048693, -347222.258874011, -90625000.0681495, -5208333.17150516, 0, -168402777.751595, 0, 0.161347832934814, -168402777.731081, 0.0798116329033313, -6944444.63411927, -343750000.418251, -2083333.47318803, -2083333.43315141, -343750000.269315, -6944444.30406532, -182291667.292666, -19791667.1534094, -182291666.726066, -5208333.56367067, -510416667.14406, -5208333.40524417, -184027778.96483, -520833334.628634, -93750001.121754, -93750001.0284844, -520833334.230421, -184027778.384218, -354166668.415844, 0, 0, 0, 0, -0.000502629985130453, 0, 0.000908667359160362, 2.54100338860404e-06, -0.00119167685557707, 0, 3472222.22106431, 0, -0.00918888134873544, 3472222.22299091, -0.0107906427089539, 0.00333306062128652, 0.00241694492477336, -0.00188429961486992, 0, 20833333.3332325, 0, -0.0424999752723639, 20833333.3292148, -0.0163142723242796, 0.0172728039179567, 20833333.3714966, 0.0071801318208154, 0.00829334658756132, 20833333.3514622, -0.0573697467477058, 0, 52083333.3286337, 0, -0.0633906942311876, 52083333.3167339, -0.0179767684657755, 0.0504952077757996, 104166666.799571, 0.0283114158272666, 0.0241801304785981, 104166666.74387, -0.0865259590763609, 52083333.5136905, 3472222.35556412, 52083333.3014066, 0, 69444444.4275731, -0.042405069593771, 0.0790992232679867, 208333333.536873, 0.0650070795511168, 0.0443785939878396, 208333333.497345, -0.00856908942998967, 69444444.8710899, 208333333.849351, 20833333.7178725, 20833333.736096, 208333333.568322, 69444444.5262844, 52083333.8853975, 312500000.572632, 52083333.7403758, 208333334.285837, 104166667.577388, 208333334.039594}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 1.24989590321749e-07, 375.000000017082, 0, 0, 0, 0, -9354.16666605545, 0, -2.35652272213119e-06, -4541.66666756817, -2.05064033244506e-06, -791.666668382227, 0, 0, 0, 0, 88805.5555484562, 0, 1.76073602730161e-05, 8333.33334282064, 1.42514146035639e-05, 0, 112451.388885944, 3.55453120421239e-05, 18909.7222431324, 416.666685383638, 8750.00001773471, 0, 0, 0, 0, -423576.388857625, 0, -6.8232064852848e-05, -4166.66671488041, -4.92643462953223e-05, 0, -1055520.83327875, 0, -0.000194080817082476, -198611.111107511, -0.000227828258955776, -167465.277906023, -9.17304788305592e-05, -8333.33341477038, -9513.88905156305, -208125.00018881, -4166.66686177959, 0, 0, 0, 0, 1107638.88880376, 0, 0.00015460377040483, 0.000136342498333457, 9.16950533156303e-05, 0, 4952777.77763373, 0, 0.00064792792206312, 95138.8889478015, 0.000693556805245727, 717013.8892718, 0.000288997686690812, 0.000292022925812709, 0, 1752777.77745381, 0.00105602332589888, 78125.0007265531, 1830902.77881378, 0.000422194027717337, 0.000666966371748777, 190277.778332371, 0.000868232294917277, 95138.8900958414, 0, 0, 0, 0, -1597222.22206819, 0, -0.000235372721770983, -0.000218891878931037, -4.47890224760959e-05, 0, -12656249.9998336, 0, -0.000996668822256057, -0.000328809786637551, -0.00134919613695284, -1579861.11174775, -0.000495211377562719, -0.000547818300368424, 0, -7465277.77703765, 0, -0.00189504558539553, -781249.999027431, -0.00271648155946956, -295138.890736332, -7760416.66903196, -0.000569590784467922, -0.00133638506557119, -0.00123487464419485, -0.00214750383068752, -0.00169267389521629, -1562500.00266562, -0.00188023956028211, -781250.00508801, -0.00298031352373563, -0.00362221280197179, 0, 0, 0, 0, 1180555.55538494, 0, 0.000215382912631249, 0.000196186703285475, 2.86735256217533e-05, 0, 17673611.1109547, 0, 0.0011358602152599, 0.0006268281089087, 0.00109657322730416, 1701388.88949519, 0.000639050325555404, 0.000461896146074669, 0, 16319444.4440215, 0, 0.00255018348640591, -0.00115555946143483, 0.00384124027352576, 520833.335595503, 16840277.7806909, 0.000341399604136251, 0.0022386611516309, 0.00143742134992102, 0.00376890982018265, 0, 2951388.88664448, 0.00457887836187433, 0.0033659256446734, 5902777.78331503, 0.00165959219553235, 0.00161032476454798, 0.00368822926419208, 0.00336628291020874, 2951388.89883036, 0.0035944781826726, 0.00602910086909924, 0.00940780019251855, 0.00997519747130718, 0.0056836910033729, 0, 0, 0, 0, -347222.222120864, 0, -0.000110486068007404, -9.08283162713252e-05, 4.97724696598455e-06, 0, -12499999.9998559, 0, -0.000724077514553993, -0.000505253624173002, -0.000504635768331221, -694444.44473946, -0.000420012626356108, -0.000248814329707236, 0, -17361111.111217, 0, -0.00169445851730219, 0.000454621367647516, -0.00272583261884771, -347222.22363263, -17708333.335218, -0.000199980309511644, -0.00215768673775202, -0.000941252571837412, -0.00276495414594097, 0, -5208333.33188442, 0, -0.00344566423247937, 0.00243107840930305, -0.00399555571618469, -0.0026137235338639, -10416666.6711522, -6.26706594291275e-06, -0.000715987544690391, -0.00225890366245106, -0.0039400231381025, -5208333.34114421, -0.00221520712824042, -0.00554992182838172, -0.00202311415189946, -0.00462207294068275, -0.00109029121425013, -0.0127951377290599, -0.0128069453863146, -0.00836724603952383, -0.00188819419252571, -0.00919794966742571, -0.00503214350791957, -0.0122767054239697, 0, 0, 0, 0, -2.44216186147755e-05, 0, 2.25332949215329e-05, 1.67748817212996e-05, -4.11295340647185e-06, 0, 3472222.22216147, 0, 0.000173157086582079, 0.000147382342341282, 9.53552547850576e-05, 5.9402880150933e-05, 9.35507323727807e-05, 3.43745229671691e-05, 0, 6944444.44454283, 0, 0.000394124506734941, 6.01665383662596e-06, 0.000598510465134624, 0.000311886950692596, 6944444.44491478, 7.2370752965054e-05, 0.000759804945375317, 0.000170697735019105, 0.000889838606099037, 0, 3472222.22201836, 0, 0.000894968706023414, -0.000814741116738504, 0.0013387725533337, 0.000885512689888158, 6944444.44593021, -0.000217382311816396, 0.000657062193415183, 0.000570158109774717, 0.00178562875503509, 3472222.22457196, 0.000843831861458091, 0.00236303571196009, 0, -0.001015542266802, 0.00161866379338588, 0.000834935949582684, 0.00214432017662692, 8.55146957575557e-05, 0.000234164159428947, 0.00139873315760668, 0.000885000490797508, 0.00558024294457566, 0.00564718933072096, 0.00229545983425443, 0.000557762681995516, 0.00379185835110833, 0.0022534865551654, 0.00504526785052028, 0.00525278348973777, 0.000294748752556139, 0.00902745930777636, 0.00478109730369538, 0.00472704736169628}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -2.88978209839525e-07, 375.000000006614, 0, 0, 0, 0, -4541.66666598655, 0, 5.49398755093744e-06, -9354.16666765248, 5.83548454242315e-06, -791.666665556592, 0, 0, 0, 0, 8333.33332356508, 0, -3.3182487283363e-05, 88805.5555662864, -3.74471766730825e-05, 0, 112451.388893039, -0.000113293269451029, 8750.00000608884, 416.66663402125, 18909.7222079476, 0, 0, 0, 0, -4166.66661398917, 0, 9.18878776318778e-05, -423576.388938333, 0.000117942956314426, 0, -198611.111077857, 0, 0.000638930854061545, -1055520.83341121, 0.0006998144056891, -8333.33339621969, 0.000215043356373985, -167465.277678691, -4166.66659674985, -208125.000061942, -9513.88852443126, 0, 0, 0, 0, -0.000149247856460334, 0, -0.000114127376493702, 1107638.889016, -0.00020907883324759, 0, 95138.8886363906, 0, -0.00182302469008255, 4952777.77810404, -0.00210457470811614, 0.000245960398988142, -0.000758170865639729, 717013.888583868, 0, 1752777.77790632, -0.00345789763926485, 1.65287231525024e-05, 190277.778403476, -0.000775922132793556, -0.0016817183148635, 1830902.77781832, 78124.9977873147, 95138.8887564848, 0, 0, 0, 0, 0.000243085924141893, 0, 3.73588911015654e-05, -1597222.22241619, 0.000161710493558293, 0, 0.00069367661353395, 0, 0.00266668909123313, -12656250.000687, 0.00358471380545818, -0.000485073702206004, 0.00141563766450381, -1579861.11073983, 0, -781249.999886742, 0, 0.00782914166112485, -7465277.77837449, 0.00901842505729539, -0.00037523995004425, -0.00185686172444383, 0.00268974170844696, 0.00380513914932279, -7760416.66642374, -295138.882280252, 0.00102533363516071, -1562500.00123269, 0.00727177305829108, -0.00183641569539748, 0.000829729488373972, -781249.997920573, 0, 0, 0, 0, -0.000227219237819264, 0, 6.83852809620682e-05, 1180555.55572912, -4.97020636272807e-05, 0, -0.00096927700909637, 0, -0.00228147384544795, 17673611.1119098, -0.00338412450799471, 0.000531402109605062, -0.00151894631739816, 1701388.88890486, 0, -0.000567577190555775, 0, -0.00912583596319335, 16319444.4454234, -0.0117115627297126, 0.000515420898834048, 0.00282010118305273, -0.00505450830504092, -0.00518904609712257, 16840277.777502, 520833.323164757, 0, 2951388.88920869, -0.0151761423951218, -0.00056087892497623, 0.00330743766697808, -0.00409189488034234, -0.011069240718181, 5902777.77826113, -0.0163342428605813, 0.00485120938705288, -0.00439959883666987, 2951388.88150581, 0.00208335549786294, 0.00211661045067825, -0.00284530624294777, 0, 0, 0, 0, 0.000112571512874325, 0, -7.5853456648817e-05, -347222.222306237, -1.90817367713464e-05, 0, 0.00068073079059139, 0, 0.0010137201564042, -12500000.0004819, 0.00165963797454558, -0.000313072474644496, 0.000891831093223522, -694444.444679464, 0, 0.000663313200584675, 0, 0.00545174272662516, -17361111.1118932, 0.00794113207746774, -0.000375217236583811, -0.00209586420061998, 0.00420092599597529, 0.00370402740820026, -17708333.3334582, -347222.215171408, 0, 6.72180272314438e-05, 0, 0.0119585549164272, -5208333.33370952, 0.0137171540979953, 0.000130515865007572, -0.00374642306539062, 0.00521042164222822, 0.00735700507473112, -10416666.6669297, 0.0181551233567847, -0.00452780788752038, 0.00680706699365183, -5208333.32265859, 0.00263357913568585, -0.0020287389346837, 0.0157203006710534, -0.00479514082851158, -0.00532693411923615, 0.00144641744218757, 0.000802182568812574, 0.0026838172000406, 0.00923250148469486, -0.00433398774023849, 0, 0, 0, 0, -2.27724002009903e-05, 0, 2.64097268266369e-05, 1.70116830489384e-05, 1.36875855727726e-05, 0, -0.000189333034325842, 0, -0.000177940261493302, 3472222.22233894, -0.00032014698094267, 7.57652897490809e-05, -0.000188735774740358, 0.000125200421991038, 0, -0.000267414400882767, 0, -0.00130865120447759, 6944444.4446933, -0.00205338272540572, 8.47410031291469e-05, 0.000620950523110904, -0.00128841717090119, -0.00102272588424159, 6944444.44464672, -0.00197721359733669, 0, -0.000101289486257164, 0, -0.00352603564180802, 3472222.22235986, -0.00465837225199489, -7.40241638089367e-05, 0.00145702435280537, -0.00233315245993806, -0.00222043066504345, 6944444.44453587, -0.00707509768366189, 0.00134005303492953, -0.00327604003520682, 3472222.21806819, 0, 6.53231257894554e-06, -0.00542218353064986, -0.000558022826626547, 0.00136532015464079, -0.00169835148728449, -0.00503301959114322, 0.000475677681094615, -0.00877272832740801, 0.0016242948127025, 0.00310118474669152, -0.00279207023090859, -0.00061969815100389, -0.00394180922097254, -0.00746693087342639, 0.000634164251367702, 0.000583247461366074, -0.00168612724540873, 0.0054368960599152, 0.00410581639047301, -0.000179027493928518}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 2.50606027665208e-06, -4199.99999866013, 0, 0, 0, 0, 90116.6666435877, 0, -5.18174245928825e-05, 69116.6666460018, -2.97530541252929e-05, 27116.6667590999, 0, 0, 0, 0, -686194.444279934, 0, 0.000470685461978121, -340611.110954003, 0.000150514340016027, 0, -1446805.55527682, 0.000443539627016864, -545638.890381535, -69444.4454723055, -410055.556422093, 0, 0, 0, 0, 2485694.44379132, 0, -0.00208925511015749, 782638.888306465, -0.000759316239475277, 0, 10565694.4430514, 0, -0.00393808744173838, 6762638.88694062, -0.000462045895096164, 3703750.0079981, 88194.4494412128, 1653472.22454707, 1306250.01394619, 8068888.90367599, 959027.785708136, 0, 0, 0, 0, -4850694.44271537, 0, 0.00502398579526514, -937499.998826045, 0.00288485978356073, 0, -35992361.1082394, 0, 0.0178294997232862, -14607638.8826016, -0.000999919513334153, -11135416.6881193, -55555.5680180135, -2868055.55626282, 0, -44811805.5463263, 0.00435530224013684, -7774305.61520736, -52586111.1830565, -1545138.944022, -1104166.70206363, -30760416.7170075, -2979166.68077136, -17697916.7777283, 0, 0, 0, 0, 5256944.44139523, 0, -0.00673928285115187, 569444.443006673, -0.00620485713101054, 0, 65138888.8858054, 0, -0.0399341286346099, 16354166.6571176, -0.00111493148192319, 16631944.4765525, 13888.9063521225, 2291666.65932948, 0, 129722222.20583, 0, -0.0292962966956382, 85624999.9715423, 0.0152251443415432, 18368055.6789, 148090277.941266, 902777.887579344, 625000.076563234, 49965277.8606773, 3472222.20753347, 7881944.62446212, 179131944.66116, 2361111.18343826, 101388889.309391, 19062500.367792, 51770833.6022256, 0, 0, 0, 0, -2986111.10781418, 0, 0.0050777801006654, -138888.887725726, 0.00708611873259791, 0, -64722222.2196192, 0, 0.0466014526640434, -9236111.10447178, 0.00993312659262924, -12152777.8050427, -0.0136968609282409, -694444.430120696, 0, -185069444.433294, 0, 0.0637772650492598, -82638888.8495561, -0.0330798740771518, -18750000.1348619, -203819444.637525, -208333.450366371, -138888.97287376, -37152777.8523941, -1388888.8137827, 0, -196874999.954069, -0.00723928048425407, -13194444.7142994, -406944444.836009, -3819444.69941351, -694444.584877569, -251736111.409034, -1388888.92004316, -223263889.580678, -10069445.0261673, -56250000.2482323, -94097223.2151919, -259375000.872593, -38194445.129996, 0, 0, 0, 0, 694444.442510749, 0, -0.00192028299027586, -0.000607875161070301, -0.00414894752222032, 0, 33333333.3312964, 0, -0.0274038034307702, 2083333.33174326, -0.0121336676609459, 3472222.23443692, 0.00544973316767467, -0.0104050527409604, 0, 128472222.220509, 0, -0.0573980276055002, 38888888.864989, 0.0206092191974206, 6944444.51942854, 135416666.780965, 0.0641704139712791, 0.0461463045376572, 10416666.7034976, -0.0826099444675747, 0, 194444444.416428, 0, -0.0198156797025086, 135416666.617312, 0.037850512399388, 6944444.63732237, 395833333.656685, 694444.615570903, 0.113066918810398, 156250000.19933, -0.0451532321657813, 208333333.857653, 2083333.76907686, 20833333.4017694, 3472222.46596414, 409722222.599222, 0.0924626435497973, 145833334.384234, 416666667.692774, 41666667.6717385, 10416667.4421518, 236111111.801943, 20833333.7704911, 159722223.611382, 0, 0, 0, 0, 0.000467215948641469, 0, 0.000261366236015211, 0.000153201978257407, 0.000962056051186025, 0, -6944444.44366244, 0, 0.00644645320497732, -9.22176799789094e-05, 0.00458637715574331, -0.00224096084517543, -0.000871077151384343, 0.0027989442653745, 0, -34722222.2226957, 0, 0.0188267490326934, -6944444.43942114, -0.00308841352888197, -0.0166423860384944, -34722222.2489314, -0.0140813722419766, -0.0100944176424434, -0.0078724560254633, 0.0306360606782014, 0, -69444444.439421, 0, 0.0149326400068423, -34722222.2047992, -0.017207489888514, -0.0535076598792679, -138888888.989385, -0.0443814516783019, -0.0319023830186096, -34722222.2760589, 0.0377089026187119, -69444444.5949865, -0.125532975175415, 0.00995559436486481, 0, -69444444.4265492, -0.00814430746214871, -0.0846094317505535, -208333333.495095, -0.0766325132697376, -0.0476433792880303, -138888889.018712, -0.0116989104622336, -69444444.8321506, -208333333.74396, -6944444.80610845, -0.305420211672254, -69444444.6317835, -0.062514880921758, -34722222.7362144, -208333333.779933, -0.276787237985451, -138888889.667518, -34722222.9530975, -69444444.9787036}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 4.90425529030299e-07, 2624.99999919256, 0, 0, 0, 0, -60697.9166532445, 0, -5.33483297085091e-06, -38822.9166529261, -1.67301273742816e-05, -12572.9167121682, 0, 0, 0, 0, 515451.388796536, 0, -2.57186769334847e-05, 148177.083224795, 0.000140155346481138, 0, 882378.472038975, 0.000285492606961789, 275399.306226864, 22447.9172203596, 170625.000421946, 0, 0, 0, 0, -2124045.13853884, 0, 0.000302103878130579, -242187.499575894, -0.000442000794815539, 0, -7288020.83241135, 0, -0.00102948584469726, -3218663.19313364, -0.00263944319286584, -2133506.94805338, -17708.336109289, -502083.334664942, -464670.145438176, -3683333.33964557, -277604.171137621, 0, 0, 0, 0, 4670138.88798987, 0, -0.00106137630497769, 182291.665732079, 0.000481505010825599, 0, 28820312.4981094, 0, 0.000551685959713058, 4992187.49570375, 0.00921345302380969, 7579861.12080699, 5208.34082777879, 552083.335190803, 0, 24585069.4381786, 0.0110471015201027, 3250000.02785066, 27835069.4754759, 345486.13739353, 197916.685497058, 10329861.1319546, 562500.012010183, 5683159.77189926, 0, 0, 0, 0, -5607638.8873052, 0, 0.00172840224177353, -52083.3320449804, 0.00037774848115449, 0, -59878472.2203676, 0, 0.0031184544253056, -3550347.21538355, -0.015475831770449, -13177083.347716, -0.011703263833656, -208333.333778422, 0, -85425347.210669, 0, -0.0172485362584799, -34652777.7586489, -0.0378349459979093, -9626736.16943599, -95052083.4038133, -95486.1652078339, -52083.3752347429, -10746527.8116429, -312500.014728404, -2152777.86262522, -71458333.4244854, -208333.376006415, -38958333.521434, -3836805.71774772, -10937500.1230671, 0, 0, 0, 0, 3472222.22048393, 0, -0.00160188794692891, -0.00115366643654857, -0.00111739624672691, 0, 67013888.8876496, 0, -0.0060578726858073, 954861.105702136, 0.0123243972777894, 10937500.012002, 0.0104945428077635, -0.00219328872653003, 0, 143749999.9916, 0, 0.0111065437864923, 22048611.0847536, 0.054795260565093, 11979166.731174, 155729166.749398, 0.0612276954186003, 0.0511555177597898, 3819444.47420326, 0.00657204478481211, 0, 101302083.301705, 0.0557102674077315, 5034722.35104998, 207638889.055321, 520833.454816743, 0.0707488724631381, 66666666.7897194, 0.0421093940576569, 111371528.089455, 954861.368200711, 5729166.79932355, 23611111.5578065, 67708333.7189313, 3819444.7349328, 0, 0, 0, 0, -868055.554517571, 0, 0.000761380293485758, 0.000622676186731355, 0.000918052091142133, 0, -38194444.4435505, 0, 0.00429609736962074, 0.00196750187700059, -0.00414835548145707, -3472222.22740046, -0.0048398557434583, 0.00217208565671446, 0, -114583333.331737, 0, -0.00104209346590542, -5208333.31709438, -0.0368883401529754, -5208333.36952827, -119791666.714778, -0.0365517479573194, -0.0330865948568783, -0.0136106153282503, 0.00155821433758324, 0, -123263888.868695, 0, -0.034891557769152, -51215277.7444801, -0.0601274128740787, -3472222.31587977, -250000000.136771, -0.0836174391869188, -0.0575399725368082, -20833333.4123653, -0.0223335324610651, -130208333.569127, -0.197392858236583, -0.0694677206778783, -868055.671046, -154513889.047089, -0.0511633371137824, -53819444.9269846, -156250000.450897, -5208333.78799492, -0.310825149130106, -31250000.3160945, -0.195286667323587, -20833333.9421667, 0, 0, 0, 0, -0.000254629102061363, 0, -0.000145098587791289, -0.000151323865711041, -0.000248484556377231, 0, 8680555.55518395, 0, -0.00116428166261455, -0.000242727479796091, 0.000207191561753139, 0.000899776032318356, 0.000873339673133632, -0.00072844836416021, 0, 34722222.2225271, 0, -0.00124286166064371, -0.00357572019960832, 0.00908439122818795, 0.0080180627488024, 34722222.2331572, 0.00884160767474561, 0.00872875463580873, 0.00222735713295138, -0.00212131609355384, 0, 52083333.3294066, 0, 0.00738455569279067, 8680555.5439328, 0.0219473583264867, 0.0266708813576848, 104166666.70896, 0.0229900044130974, 0.0188902962244027, 0.019837027473238, 0.00393759356122126, 52083333.4008406, 0.0602565209978827, 0.0149385874773759, 0, 34722222.2097871, 0.0210194928135172, 0.0402652932103675, 104166666.735006, 0.0356875030016792, 0.0217062609444784, 34722222.2753501, 0.0143162749865779, 34722222.4037116, 104166666.846411, 0.165544749990725, 0.120639739414372, 0.0963288411088606, 0.0435265170939807, 8680555.79263072, 52083333.5309834, 0.10285143666677, 34722222.5685248, 0.311829299955874, 0.236133567933729}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -5.56685227428186e-07, -1199.99999977799, 0, 0, 0, 0, 29033.3333287202, 0, 1.01801629246088e-05, 16033.333329866, 1.06794466164733e-05, 4033.33334569831, 0, 0, 0, 0, -263527.777744031, 0, -6.23786936420032e-05, -44833.3333081633, -7.53510893378784e-05, 0, -383361.111046538, -0.000188481030775056, -93027.7779421657, -4500.00015371163, -49333.3334485224, 0, 0, 0, 0, 1183611.11098496, 0, 0.000205012747756822, 46666.6665785384, 0.000253242442504048, 0, 3415972.22184742, 0, 0.000987246978860928, 1029027.77733244, 0.00129495851695252, 780694.445368875, 1666.66744133088, 95000.0004340225, 98750.0015358962, 1127777.77924216, 50000.0013331585, 0, 0, 0, 0, -2871527.77746282, 0, -0.00038309321000977, -16666.666496801, -0.000431717486650071, 0, -14933333.3325066, 0, -0.00267485310719816, -1022222.22083452, -0.00412880086188543, -3097222.2248286, -0.00227701472039059, -50000.0010244149, 0, -8567361.1086171, -0.00587258445379892, -760416.67333232, -9327777.78526115, -34722.2280278983, -16666.6718562993, -2079166.67129344, -50000.0047085512, -1091666.67796236, 0, 0, 0, 0, 3798611.11056187, 0, 0.000517965180790223, -0.000214278113114881, 0.000238395803878119, 0, 34874999.9991474, 0, 0.00380984752665149, 347222.220202107, 0.00710280094525254, 6159722.22625226, 0.00388806092555377, 0.00146125095819071, 0, 33576388.8839505, 0, 0.0115807818801817, 7847222.21475403, 0.0165892594999697, 2604166.68141617, 36180555.5730331, 0.011493264451507, 0.0116321702716986, 1041666.67427363, 0.00988995155172275, 243055.574521642, 15937500.0206554, 0.0137591539398531, 8333333.37696809, 347222.256223573, 1041666.69651599, 0, 0, 0, 0, -2569444.44383915, 0, -0.000415654152902122, 0.000199302255370114, -3.02114329165906e-05, 0, -43819444.4438769, 0, -0.00362565257969457, 0.00133958963741187, -0.00615004996933818, -5833333.33686046, -0.00410887880673662, -0.000914062791839675, 0, -65555555.551761, 0, -0.0129985347865203, -2430555.54550677, -0.0222819094901284, -3958333.35032123, -69513888.9101979, -0.0130494198931402, -0.0157997659137927, -0.0070365119652313, -0.0137097684746529, 0, -26736111.0978453, -0.0266738418841461, -694444.474984934, -54166666.7063207, -0.025353216428476, -0.0197729853147728, -7291666.69366263, -0.019516877789736, -28125000.0754692, -0.0520019530069209, -0.0376877148950274, -2430555.65374192, -7291666.75456171, -0.0624174588222976, 0, 0, 0, 0, 694444.444080807, 0, 0.000192109389406554, -0.000123266784292196, -9.40340984514527e-05, 0, 27777777.7773938, 0, 0.0020446395857726, -0.000309227248923093, 0.00267629380152684, 2083333.33490829, 0.00233531785787518, 0.000371299906164545, 0, 60416666.665856, 0, 0.00761121499799784, -0.00598739761097006, 0.0151165074560974, 2083333.34321774, 62500000.0127937, 0.00818910742930859, 0.0119424065279229, 0.00347701550153652, 0.00972877444827134, 0, 40277777.7689498, 0, 0.0198087840967715, 6944444.43067707, 0.0250259376782594, 694444.467536331, 81250000.0332729, 0.0156120010658873, 0.0147126311872573, 0.0160741438370018, 0.0192217860482767, 41666666.7251108, 0.0390615657104854, 0.0286629987883184, 0.0253243001009615, 20833333.368848, 0.0175001026864932, 6944444.55605786, 20833333.4383908, 0.0963803287396221, 0.0582887035702768, 0.0742065688438344, 0.0470865740685065, 0.133289090375199, 0, 0, 0, 0, 8.93903716916456e-05, 0, -3.69022492393862e-05, 3.46276320530575e-05, 3.7436415550746e-05, 0, -6944444.44428839, 0, -0.000461453984746867, -1.28542955496058e-05, -0.000418355481412888, -0.000288698182351551, -0.000498882330592494, -3.01097296791708e-06, 0, -20833333.3334481, 0, -0.00176051896348425, 0.00125325827649337, -0.00371540051477417, -0.002232547046172, -20833333.3363318, -0.00213152313987369, -0.00367448533315458, -0.000480547918048776, -0.00273823750429498, 0, -20833333.3315574, 0, -0.00534588477474023, 0.00473409469058919, -0.00877276514569197, -0.00692803752975839, -41666666.677379, -0.0040574320886556, -0.00579769389794834, -0.0037220849207448, -0.0073504935798517, -20833333.3506695, -0.012763857330821, -0.00971060479100511, 0, -6944444.43908434, -0.00970086771853467, -0.00908572747684307, -20833333.3492229, -0.00649372412987864, -0.0061201374573867, -0.0111082933032485, -0.00746879926856232, -6944444.48886593, -20833333.3764996, -0.0333540386199027, -0.0215373637881491, -0.0257752218238812, -0.0160008428189494, -0.0519952637109161, -0.0450940974862953, -0.018802169981064, -0.078128946235289, -0.0633190490160621, -0.052292467060026}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 7.97518539375541e-08, 299.999999740593, 0, 0, 0, 0, -6883.33332986382, 0, -1.32464625093256e-06, -5133.33332805397, -2.2737830635517e-06, -633.3333377084, 0, 0, 0, 0, 57277.7777574742, 0, 2.06515095665048e-06, 24833.3332895314, 1.66246171045045e-05, 0, 117111.111051965, 4.16257548220666e-05, 13861.1111653457, 333.333395974948, 10166.6667066364, 0, 0, 0, 0, -224305.555488407, 0, 1.05584374855997e-05, -36666.6664846726, -5.59969367702916e-05, 0, -965416.666394869, 0, -0.000203183191482892, -559027.777343027, -0.000320494932343673, -106250.000278329, -0.000325575868287661, -41666.6668076419, -6944.4450444284, -221805.55600418, -5000.00056282829, 0, 0, 0, 0, 437499.999847752, 0, -8.07462646890974e-05, 16666.6662328797, 8.97764615607314e-05, 0, 3725694.44386779, 0, 0.000378211041774332, 797222.220737689, 0.00107384118274036, 361111.111864852, 0.000904968664050312, 33333.3335357261, 0, 4506249.99816632, 0.00135329093399257, 48611.1134999678, 1690972.22431352, 0.0028499623924804, 0.00249271288678468, 901388.890311695, 16666.668664214, 104166.670871833, 0, 0, 0, 0, -402777.777529997, 0, 0.000160388991764579, 0.000634200863351724, -4.10851645457789e-05, 0, -7104166.6660493, 0, -0.000364562199875226, -347222.219585517, -0.00160600159827186, -541666.667782727, -0.0014072411876872, -0.000102327356190519, 0, -16770833.3298391, 0, -0.00292488425227814, -6041666.66125498, -0.00433606294785855, -138888.893595826, -5694444.44938227, -0.00657286953947043, -0.00595815835735305, -694444.446076989, -0.0039807536990223, -0.00823899378009891, -6770833.33903738, -0.00826498764590456, -729166.680613902, -0.0151003829638982, -347222.233250718, 0, 0, 0, 0, 138888.888624888, 0, -0.000188378121134675, -0.000570072464268063, -2.67383422629842e-05, 0, 6319444.44407426, 0, 3.11257018902168e-05, -0.00255316209144704, 0.00152645431175016, 277777.778732148, 0.00113102879102476, -0.000151770969267701, 0, 30138888.8857748, 0, 0.00274767397116519, 2430555.54800806, 0.00521591764738843, 138888.893969047, 8402777.78398825, 0.00827708479786779, 0.00683202041112217, 0.00087098811033129, 0.00394410507968048, 0, 20138888.8802583, 0.0065349815016534, 0.0113825173914056, 22222222.2326343, 0.0136401360724076, 0.0150903556766156, 4861111.11787141, 0.0119815944603197, 2083333.35458867, 0.026783172287231, 0.0117069049938238, 0.0375987343971545, 2430555.58309444, 0.0298582237975155, 0, 0, 0, 0, 0.000157606644746659, 0, 0.000110715016930989, 0.000290924451685613, 4.44686728108636e-05, 0, -2083333.33317617, 0, 0.00012413869198465, 0.00129773248144233, -0.000642413666085729, -0.000411323432428829, -0.000464911779208606, 0.000133945535613546, 0, -24305555.5543475, 0, -0.00139329633689075, 0.00494893713780537, -0.00348894520155342, -0.00282525532267233, -4166666.67031497, -0.0051921379474819, -0.00381415004462798, 6.37816217669363e-05, -0.00249711796849796, 0, -29166666.6607516, 0, -0.00499452754685658, -6944444.4357751, -0.00644790257945511, -0.0079694437576211, -31250000.0090875, -0.0113018364257291, -0.0118988478457927, -0.00339967196890544, -0.010595114195799, -2083333.3490572, -0.0215712713444712, -0.00785114404500621, -0.012442721183121, -13888888.8982194, -0.016994821668354, -0.0367886191259855, -6944444.47385657, -0.0425576288394683, -0.037659317453149, -0.0211316098775017, -0.0239956542204143, -0.0522214673396075, 0, 0, 0, 0, -3.91973031136333e-05, 0, -2.36897133250903e-05, -6.44694374212436e-05, -1.46853716911004e-05, 0, -4.56536964198148e-05, 0, -5.74443883638638e-05, -0.000276163645901227, 7.18235680100602e-05, 7.43688728490604e-05, 6.49704675659327e-05, -4.6992079413465e-05, 0, 6944444.44429696, 0, 0.00031772080469839, -0.00123348547223562, 0.00098086567018504, 0.000667039442639477, 0.000856383566123844, 0.00126783354570265, 0.000852947681291413, -0.000183931981790365, 0.000429813616537672, 0, 13888888.8874816, 0, 0.00140465423753138, -0.00301145186464292, 0.00243372414566247, 0.00215592202981266, 13888888.8919526, 0.00360886350403623, 0.00353683322577361, 0.00056061024802775, 0.00321461664012913, 0.00469735265308604, 0.00696943857946463, 0.00183883127281416, 0, 6944444.44118912, 0.00249443906642948, 0.00386862604557804, 13888888.8930692, 0.00454089483748461, 0.0066201704401191, 0.00280794892196097, 0.00663682521435181, 0.0134272142265892, 6944444.45577381, 0.0169295293156865, 0.0149266937240347, 0.0066250163626647, 0.00839500361628483, 0.0202509568597017, 0.0124314860219206, 0.0154645138202297, 0.024263626141321, 0.0286371816544539, 0.0205657626817594}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -2.50994066462392e-07, 281.249999852767, 0, 0, 0, 0, -5749.99999723253, 0, 7.1986071286423e-06, -5749.99999868021, 1.86499291944124e-06, -593.750003860486, 0, 0, 0, 0, 39322.9166432767, 0, -6.59848759700569e-05, 39322.9166582241, 1.16727158284163e-06, 0, 116927.083319327, -6.69014502857534e-05, 11510.4167305475, 312.500035221823, 11510.4167132024, 0, 0, 0, 0, -111979.166556267, 0, 0.000292048999963964, -111979.166641667, -3.14958064492325e-05, 0, -791753.472157189, 0, 0.000554407899955947, -791753.472144615, 0.00022334810579821, -70833.3336577294, -0.000162392055289646, -70833.3335206141, -5729.1672945508, -222482.639547364, -5729.16694059822, 0, 0, 0, 0, 130208.333007658, 0, -0.000718854954086201, 130208.333307349, 8.49475297176021e-05, 0, 2214409.72210311, 0, -0.00226181225588643, 2214409.7220112, -0.0003718428359085, 161458.334147086, 0.000230485215707409, 161458.33383029, 0, 5262152.77744128, -0.00170335213959426, 31250.0025480604, 1361111.11421752, 0.00263351637426799, 0.00128703253807294, 1361111.11338597, 31250.0003075421, 105034.72709309, 0, 0, 0, 0, -52083.332738363, 0, 0.00108089060342555, -52083.3333428797, -0.000185236414153473, 0, -2491319.44425248, 0, 0.0045658664672457, -2491319.44434447, 0.00012763046097775, -104166.667827428, -3.61641730437003e-06, -104166.667470171, 0, -14210069.4442236, 0, 0.00557850425997679, -14210069.443193, 0.00273727405340805, -52083.3382578704, -3064236.11793987, -0.00578126634907843, -0.00234782330524508, -3064236.11562907, -52083.332347439, -0.00861637015355433, -8229166.67645, -0.00165114938092062, -572916.685215617, -0.0184013004202907, -572916.676718616, 0, 0, 0, 0, -0.000640882893413894, 0, -0.000958402509704256, 3.30110657178177e-05, 0.000132134528478996, 0, 954861.110728582, 0, -0.00528201130339993, 954861.111491564, -8.69465469185134e-05, 0.00090562189200962, -0.000601936996653851, 0.000899922416062643, 0, 14930555.5561188, 0, -0.00873367734718366, 14930555.553815, -0.00118554301813752, 0.00501420341195207, 1909722.22976509, 0.0061978211687919, 0.00155144900417195, 1909722.22705587, -0.00388746942900582, 0, 35763888.8872857, -0.00683196168470227, 0.0118800875494532, 18055555.5727577, 0.0137556385888668, 0.00263840324548742, 18055555.5693654, -0.000365489082674949, 954861.14203134, 0.0289573877501002, 954861.119468739, 0.0463343171583582, 3125000.03619379, 0.0279565038566955, 0, 0, 0, 0, 0.000370063014153475, 0, 0.000452845991173128, -1.54937007536653e-05, -5.71903854507073e-05, 0, 0.000418009892689154, 0, 0.00322397895727119, -0.00053414037999818, 1.99480923262514e-05, -0.000399618331021272, 0.000826882940709283, -0.000526707620613461, 0, -5208333.33404364, 0, 0.00691183215710543, -5208333.332542, 0.000465512886734987, -0.00265178633049983, -0.00439766213879243, -0.00334372484404358, 0.000171013148639838, -0.0031010318979754, 0.00374317881414885, 0, -32118055.5553728, 0, 0.00793866495427192, -32118055.5532253, 0.00390799883592134, -0.00736085508396485, -10416666.6806594, -0.0106226418181984, -0.00330997981477427, -10416666.6772335, 0.00505810183297483, -0.0231118143570191, -0.0225715648323373, 0.0018907868185359, -0.0111919373348088, -37326388.9064341, 0.00127991600977717, -0.0496387926433752, -5208333.37594817, -0.0493908329391097, -0.0344322708332871, -5208333.35941477, -0.0147177501555725, -0.0584694807282989, 0, 0, 0, 0, -8.81024410771584e-05, 0, -8.76977566508255e-05, 2.82910964723856e-07, 1.39875406929083e-05, 0, -0.000161140285314735, 0, -0.000797660035392238, 0.000197897840712373, -4.39126508064494e-06, 6.49448887258557e-05, -0.000237534647226778, 0.000166757196729693, 0, 0.000194327474779123, 0, -0.00220335884738079, -8.21882026070278e-06, -2.85681248997014e-05, 0.000541658420530075, 0.000977313316045674, 0.00067348372339619, -0.0004410870889036, 0.000928530085664671, -0.00117979330675369, 0, 8680555.55587594, 0, -0.00298635053485901, 8680555.55466805, -0.00117685361416722, 0.00174653276655152, 0.00413271029540767, 0.00306610186211597, 0.000225244206226268, 0.003146048342674, -0.00269435514996907, 0.00616760843438975, 0.00586598671577755, -0.00196336439099558, 0, 17361111.1104878, -0.00268704265009497, 0.00349426836965494, 17361111.1188018, 0.00452599497078723, 0.000372798037450921, 17361111.1177177, -0.00179058025751048, 0.0170344138154074, 0.0173509039259678, 0.0192447523509678, 0.0140654732841642, 0.00518347851164004, -0.0016702826258086, 0.0258103948349284, 8680555.57338282, 0.0112150494400036, 0.0354958623835721, 0.0344769173290684, 0.0196602064232327}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -4.52187350769153e-07, 299.999999958275, 0, 0, 0, 0, -5133.33333111754, 0, 9.59113586509158e-06, -6883.33333444311, 7.74118844933346e-06, -633.333334074027, 0, 0, 0, 0, 24833.3333112486, 0, -6.54803376700309e-05, 57277.7777931028, -4.93638444562439e-05, 0, 117111.111105719, -0.000159494358576885, 10166.6666932356, 333.3333123391, 13861.1111198234, 0, 0, 0, 0, -36666.6665608905, 0, 0.000237655086761779, -224305.555644373, 0.000156754027624961, 0, -559027.777696741, 0, 0.000977111145514798, -965416.666688657, 0.000934146348419183, -41666.6668789921, 0.0001624803230648, -106250.000049201, -5000.00000025725, -221805.555820282, -6944.44412781175, 0, 0, 0, 0, 16666.6663656277, 0, -0.000507566981029085, 437500.000283459, -0.000262331885536589, 0, 797222.221929401, 0, -0.00287126314334222, 3725694.44465837, -0.00284273141992841, 33333.3341193056, -0.000542466389952986, 361111.11127428, 0, 4506249.99978616, -0.00467514489969216, 16666.6672920821, 901388.890575604, -0.000439223067675921, -0.00145334173960573, 1690972.22327754, 48611.1089693646, 104166.666405163, 0, 0, 0, 0, 0.000521602653634666, 0, 0.000649094093566109, -402777.778292871, 0.000281979427067135, 0, -347222.221565006, 0, 0.00516447408680259, -7104166.6674745, 0.00513509061298769, -0.00150113654082758, 0.00121294033684263, -541666.667249308, 0, -6041666.66625735, 0, 0.0113015740402768, -16770833.332733, 0.0113442076970583, -0.00244651209293346, -694444.449604834, 0.00206462329307538, 0.00328995259576223, -5694444.44653009, -138888.881459859, 0.000172714978502822, -6770833.33793563, 0.00627785189685198, -347222.225334237, 0.00298854153531887, -729166.663066993, 0, 0, 0, 0, -0.000531036497477256, 0, -0.000466111320690667, 138888.889414188, -0.000103004702147438, 0, -0.000963327647724718, 0, -0.00512445329468746, 6319444.44589566, -0.00509350323106454, 0.00161008033898694, -0.00136983319452946, 277777.778934371, 0, 2430555.55536322, 0, -0.0146433070584948, 30138888.8882255, -0.0158186140420031, 0.00396149115288638, 0.00772719281917507, -0.00346487996122045, -0.00460858473283103, 8402777.78134985, 138888.877059028, 0, 20138888.8876709, -0.0205309370511694, 0.00171743235496251, 4861111.1216644, -0.00290022615131208, -0.00677824289022377, 22222222.229094, -0.0167440605750552, 0.0109083067391293, -0.00531923126299985, 2083333.31947305, 0.000688139416307901, 2430555.55533391, -0.00626164222333655, 0, 0, 0, 0, 0.000290009712417475, 0, 0.000187527255037307, -0.00027860011973818, -1.23140792180607e-05, 0, 0.000761479123297431, 0, 0.00276624021244616, -2083333.33454144, 0.00272512702864334, -0.00088381506765893, 0.000675319797396736, -0.0011054401971955, 0, 9.64561266073519e-05, 0, 0.00958496681311201, -24305555.5555435, 0.0106347611034886, -0.00286459586185916, -0.00538794534126072, 0.00286256156792034, 0.00358426920856437, -4166666.67012681, 0.00926346488518316, 0, -6944444.44410038, 0, 0.0167111807798448, -29166666.6649744, 0.0178539233910133, -0.0028990702788296, -0.0107857695933328, 0.00417305382033239, 0.00641063268677388, -31250000.005668, 0.018892819629366, -0.013484343022191, 0.00641313522841798, -2083333.31857898, 0.000269448722548075, -13888888.8983511, 0.010895185769748, -0.00704928441425859, -0.010795231050311, 0.00423603593087776, 0.00301307602705515, -6944444.43463078, 0.0167390646942415, 0.00347834115867267, 0, 0, 0, 0, -6.53963780210118e-05, 0, -3.16584789665946e-05, 5.95746977943484e-05, 1.17525755637736e-05, 0, -0.000237351128725479, 0, -0.000614485840654392, 0.000375638092010453, -0.000602085362083103, 0.000205434767855552, -0.000191495636488183, 0.000355960126312145, 0, -9.18641473525843e-05, 0, -0.0024869669703811, 6944444.44465264, -0.00291349536419173, 0.000828020827653598, 0.00145391487660852, -0.000876836826159064, -0.00110147176552535, 0.00132844734303695, -0.00274223278793486, 0, 0.0001136093769027, 0, -0.00532541903111691, 13888888.8882348, -0.00576600632301845, 0.00118566509100869, 0.00401004837378546, -0.00172015962563484, -0.00203023372954821, 13888888.8912469, -0.00792400746414537, 0.00559972386541631, -0.00239989646536822, -0.0060423057608746, 0, 6944444.44393762, -0.00690559853295804, 0.000382435789155656, 0.004658570432936, -0.00120105772456653, -0.00321403756604075, 13888888.8924018, -0.00704848008142781, 0.00607276198004982, 0.00787915120190251, -0.004195677033477, -0.000431518613049434, 6944444.43775347, -0.0105366411427661, -0.000138256887213989, 8.74016621172418e-05, -0.00342167469866255, 0.00169332806706255, -0.00249216649950571, -0.00783274495557137}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 1.66259910157738e-06, -1199.99999996339, 0, 0, 0, 0, 16033.3333294533, 0, -3.28983857661199e-05, 29033.3333363649, -3.19963386770953e-05, 4033.3333292787, 0, 0, 0, 0, -44833.333289198, 0, 0.000206249384874151, -263527.777814933, 0.000201869607451612, 0, -383361.111109937, 0.000631373799687537, -49333.3333712793, -4499.99983777942, -93027.77773286, 0, 0, 0, 0, 46666.6664457625, 0, -0.000621401353305368, 1183611.11130367, -0.000628151980302216, 0, 1029027.77759905, 0, -0.00374868925325714, 3415972.22237804, -0.00376778899253396, 95000.0004020258, 1666.66556754204, 780694.444174294, 49999.9995165574, 1127777.77828288, 98749.9981427346, 0, 0, 0, 0, -16666.6660447687, 0, 0.000944216962517682, -2871527.77833589, 0.00107474020900534, 0, -1022222.22132599, 0, 0.0109270943515681, -14933333.334144, 0.0111881070399701, -50000.0016066867, 0.00367808832805458, -3097222.22152633, 0, -8567361.11099499, 0.0191167539579106, -49999.999808941, -2079166.6705837, -16666.6621152385, -34722.2131029001, -9327777.77935231, -760416.655947848, -1091666.66499152, 0, 0, 0, 0, -0.00103961874869779, 0, -0.000631427507974376, 3798611.11205662, -0.000960560856083291, 0, 347222.220005871, 0, -0.0175690250629211, 34875000.0021231, -0.0190574041850027, 0.00321864100648249, -0.00679831647562252, 6159722.22205278, 0, 7847222.22118855, 0, -0.0454027660100903, 33576388.8890657, -0.047007808691387, 0.00206378455676719, 1041666.67869979, -0.0154680811539776, -0.0216460960300236, 36180555.5579805, 2604166.63462322, -0.00836265841569002, 15937500.0095816, 243055.51849011, 1041666.67372909, 347222.209624354, 8333333.32163184, 0, 0, 0, 0, 0.00101284257351116, 0, -0.00010193113354919, -2569444.44536392, 0.000340173304237711, 0, 0.00308093983860329, 0, 0.0159214959040186, -43819444.4474909, 0.0181974342964054, -0.00356588408474346, 0.0068952640061796, -5833333.33517239, 0, -2430555.55374854, 0, 0.0563751039113339, -65555555.5563081, 0.0610433780825882, -0.00410987306843326, -0.0184124559213491, 0.0259605749080222, 0.0290153444957233, -69513888.8935319, -3958333.28470403, 0, -26736111.1097068, 0.0833477630720729, 0.00700341864687042, -7291666.68977413, 0.0264988432045939, 0.0570178964040731, -54166666.6771433, -694444.362695866, -0.0244233856421045, 0.0341828261907534, -28124999.9596375, 0.00539061167748328, -7291666.67264327, -2430555.53124428, 0, 0, 0, 0, -0.000527967172923274, 0, 0.000338530863682053, 694444.444917707, 6.30972069877023e-05, 0, -0.00222725432301272, 0, -0.00755987896171661, 27777777.7799857, -0.009222965946192, 0.00205918801652046, -0.00351692879000623, 2083333.33573353, 0, -0.00172005059714553, 0, -0.0354335610655069, 60416666.6678917, -0.0404261185719089, 0.00321317332013987, 0.0134950310725616, -0.020468433938922, -0.0200181179523457, 62500000.0057016, 2083333.29826431, 0, 6944444.4433072, 0, -0.0683544777860193, 40277777.7760456, -0.0718199891353816, -0.00228605721261129, 0.02484919803691, -0.0310551280790451, -0.0446975772447138, 81250000.0081778, 694444.357669936, 0.0276864466515246, -0.0418992278177593, 41666666.6162533, -0.018176380232702, 20833333.3506532, -0.0771318208413219, 0.0102369195754597, 0.0292096352174226, -0.0305534090800836, -0.0205447019707377, 20833333.3144595, 6944444.39026666, 0.00242927411234548, 0, 0, 0, 0, 0.000113420194837974, 0, -0.000130563099783031, -9.96596328327641e-05, -5.71575876248773e-05, 0, 0.000641851192177157, 0, 0.00143198730955655, -6944444.44507071, 0.00189731467326901, -0.000493699772210175, 0.000720811091381955, -0.000921367677322882, 0, 0.00070428671826728, 0, 0.00884869455235706, -20833333.3339691, 0.0106155593479963, -0.000915811922186154, -0.00387204815334466, 0.00607704727835673, 0.0053790138611647, -20833333.3359736, 0.00991450704636543, 0, 0.00028162236716629, 0, 0.0213419512833349, -20833333.3326796, 0.0234605122343547, 0.000339870251248116, -0.00958666723804014, 0.0127078787818553, 0.0140863162521709, -41666666.6701804, 0.0337815095238247, -0.0103801630909864, 0.0174398671253229, -20833333.3132616, 0, -6944444.4436231, 0.0289929080228641, 0.00464776310520549, -0.00959280386054032, 0.0113006173293005, 0.0271137348595365, -20833333.3387543, 0.0408868304794015, -0.00800959214964754, -0.0188069356259831, 0.024145072704124, 0.00986366421019993, -20833333.314488, -6944444.40696133, 0.00867640726501472, -0.00247544827070001, 0.0147093888853103, -0.0140438437436165, -0.000730756143202443, 0.010766359127086}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -4.88072678466117e-06, 2624.99999969802, 0, 0, 0, 0, -38822.916654826, 0, 9.6945437174396e-05, -60697.916669798, 9.0142611665592e-05, -12572.9166720024, 0, 0, 0, 0, 148177.0832168, 0, -0.000632274109330987, 515451.388942845, -0.000557706540763613, 0, 882378.472178733, -0.00174148931573606, 170625.000284685, 22447.9164353145, 275399.305639802, 0, 0, 0, 0, -242187.499444817, 0, 0.0020175788098686, -2124045.13921707, 0.00178513795966277, 0, -3218663.19392416, 0, 0.0106639636558827, -7288020.8333407, 0.00991610476605933, -502083.335245912, -17708.3315235048, -2133506.94467789, -277604.167167274, -3683333.33691258, -464670.135954711, 0, 0, 0, 0, 182291.665106609, 0, -0.00333668191589331, 4670138.88996021, -0.00335082496160502, 0, 4992187.49806554, 0, -0.0324817978806414, 28820312.5006726, -0.0288413328935302, 552083.339610754, 5208.32685470593, 7579861.11151602, 0, 24585069.4428918, -0.0508219039306285, 562500.005534262, 10329861.1307078, 197916.662480388, 345486.095945612, 27835069.4578291, 3249999.97953619, 5683159.73430737, 0, 0, 0, 0, -52083.3306664304, 0, 0.00263368831358734, -5607638.89085078, 0.00371885225320949, 0, -3550347.21799528, 0, 0.0549272968844977, -59878472.225276, 0.0491063890787032, -208333.344676079, 0.012335999352757, -13177083.3354717, 0, -34652777.7740373, 0, 0.125738954424075, -85425347.2181649, 0.117816396237385, -312500.015692408, -10746527.8291947, -52083.3103953187, -95486.0724167897, -95052083.3589441, -9626736.04089173, -208333.330116852, -71458333.390885, -2152777.71039828, -10937500.0704696, -3836805.57753004, -38958333.3518493, 0, 0, 0, 0, -0.00268543324380361, 0, -0.000243072274261858, 3472222.22421744, -0.00230841197718371, 0, 954861.105305398, 0, -0.0518925822812689, 67013888.8946792, -0.0487122282815532, 0.0116584412410224, -0.0123222817908456, 10937500.0059251, 0, 22048611.1074892, 0, -0.163539170905375, 143749999.995346, -0.147636367691314, 0.0204518281760862, 3819444.51512804, -0.0448962684089373, -0.0555381788328247, 155729166.700503, 11979166.5495808, 0, 101302083.324379, -0.215306378306029, 0.00283002880128955, 66666666.7808036, -0.0383385334191893, 520833.232150279, 207638888.967294, 5034722.05835166, 5729166.80909881, 954861.112859361, 111371527.736032, 3819444.54778695, 67708333.4748253, 23611111.1712815, 0, 0, 0, 0, 0.00145389279664784, 0, -0.000861646980674198, -868055.556602862, 0.000666933294638845, 0, 0.00434675924045669, 0, 0.0255928625216471, -38194444.4493353, 0.0262097884943865, -0.0063860875971502, 0.00588540626224296, -3472222.22842111, 0, -5208333.3309855, 0, 0.107351746403254, -114583333.332241, 0.0968339513201225, -0.012967591281613, -0.0486178254007234, 0.0379268757684298, 0.0399796877903483, -119791666.694988, -5208333.24126772, 0, -51215277.7728281, 0, 0.184885241161951, -123263888.878214, 0.175472548898018, -0.00473723339128546, -20833333.4406274, 0.0558761377124631, 0.0842289132590565, -250000000.05994, -3472222.0337193, -0.129102536250783, 0.0345735158009735, -130208333.239358, 0.0267828921846277, -154513888.99041, -868055.411970537, -0.134465100040995, -31250000.2121599, -0.0395161281303367, -5208333.41663963, -156250000.065143, -53819444.4156417, -20833333.5293067, 0, 0, 0, 0, -0.00032542883938736, 0, 0.000380211003645528, 0.000220343555373381, -4.40719382140178e-05, 0, -0.00130925357912952, 0, -0.00503294398793397, 8680555.55707423, -0.00585618172845343, 0.00145809870061335, -0.00104572991722772, 0.00226444770898359, 0, -0.000947990946163715, 0, -0.0279438784860795, 34722222.2228867, -0.0258837657982325, 0.00327427812237541, 0.0132221011336079, -0.0116638175709953, -0.0109491972366263, 34722222.2326161, -0.0276826979100511, 0, 8680555.55493658, 0, -0.0604477603039466, 52083333.3294059, -0.0556195993663664, 0.0011448612377743, 0.0378674810829892, -0.0247692754786634, -0.0290608266019348, 104166666.687869, -0.0781766166846364, 0.043216142387128, -0.0222066458674673, 52083333.2884794, 0, 34722222.2183559, -0.0736468158514788, -0.00756279915656079, 34722222.269986, -0.0212936476219539, -0.0499375280871345, 104166666.701862, -0.0798915025859283, 0.057426822934903, 0.101473158572178, -0.0164985154553088, 0.0286483596909219, 104166666.655417, 34722222.1581446, 0.0282768037134354, 52083333.4058005, 8680555.58267743, 0.1291325150131, 0.0979255692070669, 34722222.2783271}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 7.58156772922191e-06, -4199.99999904408, 0, 0, 0, 0, 69116.6666439616, 0, -0.000149702874519725, 90116.6666595398, -0.000134471642487518, 27116.6667217338, 0, 0, 0, 0, -340611.110919634, 0, 0.00101499436291443, -686194.444420672, 0.000810730392402098, 0, -1446805.55537853, 0.0025211752432952, -410055.556575058, -69444.4447267173, -545638.889537376, 0, 0, 0, 0, 782638.888040961, 0, -0.00338194714127017, 2485694.44448989, -0.00271111460583829, 0, 6762638.88773422, 0, -0.0159445433643166, 10565694.4435093, -0.0133052138012025, 1653472.2278823, 88194.44519742, 3703750.0023683, 959027.784766403, 8068888.90056061, 1306250.00169533, 0, 0, 0, 0, -937499.99767198, 0, 0.00583465135197011, -4850694.44496932, 0.00577620672055514, 0, -14607638.885696, 0, 0.051346135994178, -35992361.1087294, 0.0372660939679435, -2868055.57176955, -55555.5556729838, -11135416.670599, 0, -44811805.5497486, 0.0686517284576481, -2979166.69768042, -30760416.7229459, -1104166.68807776, -1545138.89504633, -52586111.1546605, -7774305.54658875, -17697916.7378632, 0, 0, 0, 0, 569444.440433701, 0, -0.00477198125904863, 5256944.44579251, -0.0078038944836124, 0, 16354166.6611472, 0, -0.0910874837511621, 65138888.8872629, -0.0645847245110775, 2291666.69316892, 13888.8865892543, 16631944.448765, 0, 85624999.9888985, 0, -0.178833136516092, 129722222.205056, -0.144318898334349, 3472222.28673928, 49965277.9116184, 625000.027812551, 902777.784535812, 148090277.858431, 18368055.485814, 2361111.18311636, 179131944.617651, 7881944.43634075, 51770833.6071127, 19062500.2208546, 101388889.067706, 0, 0, 0, 0, -138888.884759577, 0, 0.000550841026580066, -2986111.11267154, 0.00641242979491269, 0, -9236111.10428119, 0, 0.08961294266075, -64722222.2247264, 0.0685791190928186, -694444.469517648, 0.002961412633373, -12152777.783202, 0, -82638888.8802445, 0, 0.246514460164338, -185069444.42158, 0.170897388056237, -1388888.95955103, -37152777.9476058, -138888.897183392, -208333.325523614, -203819444.535305, -18749999.8465053, 0, -196874999.970578, 0.273722496973858, -1388888.99302005, -251736111.42715, -694444.512494835, -3819444.45632695, -406944444.687389, -13194444.3409835, -56250000.4591074, -10069444.7632231, -223263889.014642, -38194445.019319, -259375000.600566, -94097222.7168681, 0, 0, 0, 0, -0.00229765316187446, 0, 0.00156119560937547, 694444.445290496, -0.00283658426584816, 0, 2083333.32821296, 0, -0.0456376716155142, 33333333.3374469, -0.0405573458016771, 0.0128136963938066, -0.000738893205215364, 3472222.22743676, 0, 38888888.8856476, 0, -0.170118178686892, 128472222.209959, -0.111888000706181, 0.0395847607928899, 10416666.7765169, -0.00960860792764442, -0.0150640694821697, 135416666.728934, 6944444.30442029, 0, 135416666.649427, 0, -0.249907355021703, 194444444.411781, -0.20230493266305, 0.0696441900110928, 156250000.275213, 0.0152527285451918, 694444.446560566, 395833333.509369, 6944444.262494, 20833333.6961607, 2083333.53587109, 208333333.296787, 0.064355925668795, 409722222.519779, 3472222.19023645, 20833333.9236753, 236111111.833805, 10416667.1803991, 41666667.258052, 416666667.113181, 145833333.641604, 159722223.156532, 0, 0, 0, 0, 0.000529196895388579, 0, -0.000710900956969027, -0.000171878841969276, 0.000504686681853593, 0, 0.0016045976742803, 0, 0.0092216127863446, -6944444.44602492, 0.0100837279574721, -0.00274092199021106, -0.000309610235412735, -0.00205860351576134, 0, -6944444.44365057, 0, 0.0462125233033336, -34722222.2205111, 0.0311195096753618, -0.0090362811650173, -0.0284075107166807, 0.00561328892740246, 0.00555813092473947, -34722222.2417929, 0.0467344071255691, 0, -34722222.2193757, 0, 0.0865617610022995, -69444444.4327114, 0.0605142777023921, -0.0171702307455253, -34722222.3136185, 0.00681536042577096, 0.00716404381041522, -138888888.944085, 0.0926518125789411, -0.1097209892882, -0.0448106609265543, -69444444.3968197, 0, -69444444.4328001, 0.0895357989146598, -0.0204630384023834, -138888889.019809, -0.0079625916424601, 0.000207389582753739, -208333333.437134, 0.0452616917965015, -0.213817510165125, -69444444.7455789, -0.153832844558653, -6944444.67524982, -208333333.426631, -69444444.4529084, -0.246030987395993, -208333333.637138, -34722222.4743248, -69444444.951099, -34722222.7161131, -138888889.256233}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -3.24803056355095e-06, 2333.33333229006, 0, 0, 0, 0, -46175.9259067085, 0, 7.01340999600796e-05, -46175.9259125607, 4.49451055498359e-05, -11175.9259760585, 0, 0, 0, 0, 304259.259112773, 0, -0.000562487921962612, 304259.259166493, -0.000244891429224289, 0, 900185.18499852, -0.000863546687546117, 207546.297092362, 19953.7041998072, 207546.296828687, 0, 0, 0, 0, -873842.591970771, 0, 0.00229490630061455, -873842.59228695, 0.000866959187484139, 0, -5768518.51755762, 0, 0.00601109398844812, -5768518.51730951, 0.00379111121525942, -1204629.63388119, -15740.7431214287, -1204629.63148178, -346527.784892281, -3806250.00811528, -346527.781728677, 0, 0, 0, 0, 1238425.92422328, 0, -0.00529747074540311, 1238425.92544135, -0.00219241396801221, 0, 15829861.1089658, 0, -0.0218013969364255, 15829861.10739, -0.00985953740063898, 2722222.23384934, 4629.63497428502, 2722222.22496697, 0, 35016203.6979302, -0.0206772957534995, 1733796.32618916, 21537037.0759124, 254629.658671558, 254629.648672838, 21537037.0657925, 1733796.30241235, 5961805.61263827, 0, 0, 0, 0, -856481.478488843, 0, 0.00718927384531607, -856481.481117847, 0.00355371890239922, 0, -21215277.7748605, 0, 0.0430555257225519, -21215277.7727685, 0.0179429063229902, -2638888.90688985, -0.00578016356413949, -2638888.89035886, 0, -87303240.7309014, 0, 0.0596152190309078, -87303240.7229435, 0.0360035926344606, -2777777.83876601, -46388888.9792469, -69444.5038337678, -69444.4852664919, -46388888.9395025, -2777777.7621335, -1064814.90653679, -115439814.933371, -1064814.85906847, -29201389.1005755, -4097222.42109152, -29201389.0220813, 0, 0, 0, 0, 231481.478312125, 0, -0.00562535936642631, 231481.481357734, -0.00352388425178154, 0, 13773148.1448825, 0, -0.0474467927817271, 13773148.145926, -0.021442582517472, 925925.941907076, 0.00200614558184804, 925925.925824331, 0, 102893518.512288, 0, -0.0908694530239141, 102893518.494041, -0.0386712084155833, 1388888.95562021, 42361111.2210953, 0.0632390090502563, 0.0385021297166457, 42361111.1631842, 1388888.82936816, 0, 178703703.675763, -0.0755839620801681, 925926.057376201, 222222222.438316, 231481.617164772, 231481.57854366, 222222222.388284, 925925.945825475, 44444444.7950209, 1041666.99049408, 44444444.5464539, 16898148.6544516, 135995370.796718, 16898148.5305936, 0, 0, 0, 0, 0.00181625936832257, 0, 0.00231414860877983, 4.03227867051769e-05, 0.00185881902764155, 0, -3472222.219623, 0, 0.0274191906479582, -3472222.2229537, 0.0144020517943749, -0.00762238075226218, 0.000952582049436716, 3.31018941047134e-05, 0, -56712962.9620357, 0, 0.0685674601940943, -56712962.9486777, 0.0260792923687935, -0.0376076112745891, -13888888.9564418, -0.0334648142260645, -0.0140837805999367, -13888888.9207958, 0.0623065156708925, 0, -152777777.762006, 0, 0.0767609000336546, -152777777.746851, 0.047492449311444, -0.0902239747751037, -173611111.294904, -0.0963318253627445, -0.078140804198681, -173611111.228874, 0.0383051336808297, -20833333.6055577, -0.242636423787537, -20833333.3287434, -0.123412553624256, -319444444.654628, -0.0776041726994753, -13888889.412614, -180555556.064838, -3472222.74674852, -3472222.69674324, -180555555.864222, -13888889.1286921, -67129630.3349321, 0, 0, 0, 0, -0.000430391500443955, 0, -0.000372701583338407, -2.08257767663454e-05, -0.000396275977789086, 0, -0.000894894025263562, 0, -0.00642468447996932, 0.000621075135590966, -0.00403489168096651, 0.00150684446577174, -0.000555579582272203, 0.000192036910497841, 0, 11574074.0741938, 0, -0.0203483857985143, 11574074.0714277, -0.00784836646039285, 0.00864419151467161, 0.0164118752976419, 0.00682979101167771, 0.000673982748070253, 0.00909442154918658, -0.0223273639688354, 0, 46296296.2939079, 0, -0.028815659713255, 46296296.2852258, -0.0131401510520533, 0.0239953780836176, 46296296.3550938, 0.0261645175903621, 0.0188594629241792, 46296296.3312042, -0.0312150144153076, 0.08048108597136, 0.0669624948048425, -0.023158537519127, 0, 69444444.4337221, -0.0237022243791563, 0.040184323781103, 138888888.981006, 0.0422641412913885, 0.0412557847684565, 138888888.963642, 0.0127022112001414, 0.190876542238971, 69444444.6545651, 0.191874306636423, 0.193114828653073, 69444444.51208, 0.0220720908369537, 0.259231163846177, 138888889.099412, 0.187097718930529, 46296296.681113, 11574074.4689619, 46296296.5537454}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 3.58691763314302e-07, -999.99999932767, 0, 0, 0, 0, 21694.4444344728, 0, -1.16162369217396e-05, 18361.1110995579, 7.95865958117432e-07, 3361.11113233194, 0, 0, 0, 0, -165370.370303122, 0, 0.00012814415012639, -104166.666573669, -3.36997063093168e-05, 0, -394537.036909353, 3.68037860017451e-05, -69120.3706725668, -3750.00026175148, -57916.6668827757, 0, 0, 0, 0, 572916.66640275, 0, -0.000619855794442491, 225694.444070348, 0.000146758737785676, 0, 2958333.33275207, 0, -0.000639672022639823, 2194444.44353141, 0.000346520438223006, 477777.779315248, 1388.89023347647, 284722.223000517, 72916.669821171, 1182638.89174281, 60416.668899831, 0, 0, 0, 0, -960648.147452656, 0, 0.00165248552421502, -208333.332484478, -0.00021984433021528, 0, -9980324.07289139, 0, 0.00365614286108648, -4575231.47842126, -0.00191134533767262, -1386574.07806522, -0.00335676842563403, -437500.001381606, 0, -15886574.0702309, -0.000196828249877228, -451388.901490241, -8069444.45776216, -25462.9770257773, -20833.343666965, -5719907.41704852, -250000.00622527, -1170138.91241618, 0, 0, 0, 0, 763888.88766962, 0, -0.00256290055856926, 69444.4432641702, 0.00011364070994903, 0, 16076388.8875927, 0, -0.00855440388669425, 4027777.77269059, 0.00351576170519468, 1666666.67247365, 0.00454101118166802, 208333.334566704, 0, 50486111.1044069, 0, -0.00362963346075584, 30798611.0993523, 0.00618698350158401, 1041666.6912892, 22881944.4743433, 0.0315246565823998, 0.0232073290561363, 8437500.01583576, 208333.340748198, 138888.931711858, 37847222.2621413, 69444.4719765677, 7187500.08293722, 381944.530314208, 4791666.72315298, 0, 0, 0, 0, -231481.480168901, 0, 0.00234374176086479, 0.00103039954656112, 0.000171695247885269, 0, -12152777.776708, 0, 0.0108622612411483, -1273148.14383933, -0.00297405081914084, -694444.44923048, -0.00269595908570648, -0.000425916048755812, 0, -73726851.8469859, 0, 0.0105815350757392, -24652777.761334, -0.0117544736126525, -694444.470395189, -26388888.923791, -0.0372180406057381, -0.0249986263864319, -3819444.45875282, 0.00029226433871981, 0, -85185185.167028, -0.00454384383169318, -231481.540916189, -101273148.219133, -0.0691829095780787, -0.0508413152777998, -51388888.9428937, -0.031902476075163, -16319444.575555, -0.144742549421516, -3819444.50023889, -2083333.54733471, -28819444.6107791, -1273148.30374949, 0, 0, 0, 0, -0.000769398744291189, 0, -0.00114246970192969, -0.000528430956669497, -0.000200086585581396, 0, 3472222.22145128, 0, -0.00696858587219178, -0.00174760864103131, 0.000925430616992331, 0.00208710013416207, 0.000209951147878116, 4.31408022595335e-06, 0, 48611111.1099651, 0, -0.0105238696813946, 6944444.43409947, 0.00838842318760468, 0.0141560130658313, 10416666.6870222, 0.0221959059389834, 0.0124123580327295, 0.00709240293492885, -0.00444638878094983, 0, 93749999.9888366, 0, -0.00188498962278558, 55555555.5362899, 0.0103469991665546, 0.0399287188707015, 104166666.725406, 0.054659798694699, 0.0426779016525568, 20833333.3694687, 0.0155247844418309, 10416666.7635827, 0.115071441870829, 0.0201614913016199, 0.0606593644594159, 114583333.402633, 0.0462120486049241, 3472222.43895842, 62500000.1865584, 0.234929474228876, 0.190456228785424, 20833333.4617657, 0.107821564759962, 6944444.73188768, 0, 0, 0, 0, 0.000186564610202598, 0, 0.000224920933334183, 0.000121413434227586, 5.66142479614168e-05, 0, 0.000280468747825379, 0, 0.00178653843393347, 0.000262417426113144, 3.21688524803235e-05, -0.000364445247911095, 0.000166133924534447, -1.01217140969917e-05, 0, -11574074.0741098, 0, 0.00370662596681939, 0.00233598621002577, -0.00234829843263579, -0.00314251229202347, -0.00466097314262212, -0.00517643636549628, -0.00209373739273591, -0.0015701488057313, 0.00247492577991806, 0, -34722222.2200753, 0, 0.00211922354367788, -11574074.0672655, -0.00445154563002696, -0.01049158531813, -34722222.2404555, -0.0165820733706103, -0.0116509930670768, -0.00949690726582416, -0.00146076207792327, -0.0271845803417527, -0.0342560318159302, -0.000844016127373119, 0, -34722222.2153414, -0.00194685800890541, -0.0192680718789774, -69444444.4745155, -0.0227718280961167, -0.0206338631925069, -34722222.2463995, -0.0144421288183236, -0.0765114302314818, -34722222.2953013, -0.092014226242248, -0.0774006185886906, -0.0342034820619117, -0.0218779388292773, -0.115477352790145, -34722222.3023805, -0.0721377886010528, -11574074.2264977, -0.162410731697489, -0.108567093278297}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 1.81327438604438e-06, -999.99999970713, 0, 0, 0, 0, 18361.1111030928, 0, -3.9216508731746e-05, 21694.4444441825, -2.80434862586843e-05, 3361.1111223387, 0, 0, 0, 0, -104166.666594254, 0, 0.000291811307986969, -165370.37038284, 0.000163869981958043, 0, -394537.037000437, 0.000566035480867, -57916.6668825543, -3750.00004186842, -69120.3705027262, 0, 0, 0, 0, 225694.444105503, 0, -0.00112761079232808, 572916.666780076, -0.00053324050602017, 0, 2194444.44417131, 0, -0.00366749881644009, 2958333.33320671, -0.00299705627304397, 284722.223496456, 1388.88895008449, 477777.778301242, 60416.6681012406, 1182638.89118903, 72916.6667100646, 0, 0, 0, 0, -208333.332362229, 0, 0.00249980702134347, -960648.148600928, 0.00107146503172696, 0, -4575231.4807185, 0, 0.0120050466916972, -9980324.07397526, 0.00865914841560544, -437500.003809266, 0.000445131797073353, -1386574.07523282, 0, -15886574.0729327, 0.0155731994840769, -250000.007120645, -5719907.41918179, -20833.3381284971, -25462.9634089729, -8069444.45290403, -451388.885210816, -1170138.90056011, 0, 0, 0, 0, 69444.442726582, 0, -0.00331714161417982, 763888.889807815, -0.00134221943996212, 0, 4027777.77629662, 0, -0.0223696827942196, 16076388.8898844, -0.0150583404340239, 208333.339713276, -0.00211362074027671, 1666666.66869322, 0, 30798611.1094555, 0, -0.0403312034873989, 50486111.1074684, -0.0344572801903385, 208333.349741454, 8437500.02937156, 0.00742937894492968, -0.000247022279263133, 22881944.4608787, 1041666.64690841, 69444.4622314377, 37847222.2578979, 138888.882802338, 4791666.71866397, 381944.482087862, 7187500.02024616, 0, 0, 0, 0, 0.00179540333413205, 0, 0.002535792691955, -231481.482461511, 0.000997757909843898, 0, -1273148.14594741, 0, 0.023447843647235, -12152777.7806266, 0.015733116378125, -0.00609196906409251, 0.00341112582745674, -694444.447415211, 0, -24652777.7776097, 0, 0.0554056519655954, -73726851.8471395, 0.0428363001048651, -0.0198277267385981, -3819444.48229033, -0.00459778011257735, 0.0047786007969222, -26388888.9092462, -694444.405253131, 0, -85185185.1793152, 0.0647239244801505, -0.0284380618428876, -51388888.9575282, -0.0211848606718519, 0.000231677120429717, -101273148.199607, -231481.448704489, -3819444.54288643, -0.0572460476920292, -16319444.4378525, -1273148.25758277, -28819444.5374229, -2083333.39981139, 0, 0, 0, 0, -0.00100833267160891, 0, -0.00103501365423956, 0.000520587269248536, -0.00039187054586292, 0, -0.0018665937102653, 0, -0.0131338784782187, 3472222.22500042, -0.00895832093841072, 0.00312223170846158, -0.00245353429478776, 0.00252695529511494, 0, 6944444.44500834, 0, -0.0387291637245781, 48611111.1092844, -0.0286205343489605, 0.0121012320260967, 0.0244596484133348, -4.5964701697159e-05, -0.00703088728112848, 10416666.682152, -0.0342026203544424, 0, 55555555.5535694, 0, -0.0576219223276533, 93749999.9923545, -0.0513589120203458, 0.0211781305166896, 20833333.3946923, 0.01075658067776, -0.000831799601611483, 104166666.706361, -0.0514147200044747, 0.0853976726982232, 0.0371769649612028, 10416666.6313178, 0.0216930096305283, 114583333.399648, -0.0140995971296036, 0.127225403792154, 20833333.4664193, 0.101484817673456, 0.0921342151033101, 62500000.0472714, 3472222.23670236, 6944444.58753988, 0, 0, 0, 0, 0.000233926738254571, 0, 0.000170008848146192, -0.00010766746057848, 6.07906131400589e-05, 0, 0.000621924668923434, 0, 0.00300940540528445, -0.000924385557646379, 0.0021311310625982, -0.000666094446439471, 0.000680174556923589, -0.000856251821961009, 0, -8.11949492985691e-05, 0, 0.0108519208776385, -11574074.074261, 0.00792293233678422, -0.0029997579565395, -0.0061869098239053, 0.000736680953326267, 0.00296104916850585, -0.00521750479705103, 0.0108361643990536, 0, -11574074.074421, 0, 0.0195311765828049, -34722222.2193603, 0.0163503459662092, -0.00594314457413756, -0.0203930612428401, -0.00159606677201964, 0.00226541043691279, -34722222.2356426, 0.0245975441776211, -0.0276449441895209, -0.00765676665069218, 0.0186471106982904, 0, -34722222.2198868, 0.0220633614080947, -0.00753994152420252, -34722222.2523568, -0.00579657415737338, 0.000931656422436123, -69444444.4688894, 0.0147092223399698, -0.0516651834947142, -0.0620944705804228, -0.0318604826606364, -0.0386383143423475, -34722222.22195, 0.0212699154859055, -0.0551256868556892, -34722222.2688899, -0.0317924852902329, -0.0896207756392741, -0.0810958630353037, -11574074.1117945}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9.47590889413295e-08, 0, 1.22315471857579e-07, -124.999999975288, 0, 0, 0, 0, 0, 0, -8.64097737037586e-07, 0, -1.13999219955015e-06, 0, 0, -1.55049207087676e-06, 3241.07142873329, -4.98476427452862e-07, -5.55044376221151e-07, 0, 0, 0, 0, 0, 0, 4.41173717587108e-06, 0, 4.57376110121513e-06, 0, 0, 0, 8.1491078536487e-06, 0, 8.68303439168343e-06, -32569.4444476091, 2.93955235153298e-07, 5.25878381399058e-06, -2.57180608597482e-06, -7.14305039936189e-07, -8.21201638615583e-07, 0, 0, 0, 0, 0, 0, -6.44547268839808e-06, 0, -1.4051278681282e-05, 0, 0, 0, -4.76766581197436e-05, 0, -2.22841283225961e-05, 167881.944453321, -6.66596622345139e-06, -3.55511850768042e-05, 0, 0, -5.40151159329088e-05, 2.04330574913361e-05, 8.12379311888683e-06, 3.00807823840075e-06, 7.42333905628038e-06, 1.58077705705481e-05, -1.87170922042181e-05, 8.47788162071951e-05, 0, 0, 0, 0, 0, 0, 1.1745687196859e-05, 0, 1.00330584974717e-05, 0, 0, 0, 4.9301779543325e-05, 0, 5.31920869536321e-05, -486111.111128066, -9.23495097060467e-06, 7.91633733579283e-05, 0, 0, 0, 3.09586689700588e-05, 0, 9.89014924378578e-05, -1.46062333425784e-05, -8.62242313088798e-05, 3.03500839871717e-05, -5.69874052416754e-05, 5.35157415595374e-05, 2.22081242540364e-05, 3.21460152762548e-05, -5.16257244428952e-05, 3.22912095187402e-05, -0.000186172506939677, -0.000236407606203774, -8.26959278758618e-05, 0, 0, 0, 0, 0, 0, -6.20750173093631e-06, 0, -1.72404484309447e-05, 0, 0, 0, -6.86965349571608e-05, 0, -3.81542930025818e-05, 798611.111124323, 3.94911087164829e-06, -8.58659184405764e-05, 0, 0, 0, -0.000201980159453853, 0, -0.000157644955577493, 2.62304442426318e-05, 0.000129532895801421, -2.38936233870512e-06, -6.3724648132276e-05, -0.000141576037059927, -0.000145200598599592, 0, 0, -0.000297585209683723, -0.000114283992473714, -4.59088276424292e-05, -0.000192134449364446, 8.81654620824308e-05, 1.657477651361e-05, 0.000197956540568854, 0.000143995633179655, 0.000286601964724956, 0.000397035301279614, 0.000728020437267363, 0.000593502545963586, 0.000255091871325182, 0, 0, 0, 0, 0, 0, 4.18697413357428e-06, 0, 7.63825397922029e-06, 0, 0, 0, 3.69596960075579e-05, 0, 3.3598766194202e-05, -694444.444452165, -9.61428714418844e-06, 6.88551106371811e-05, 0, 0, 0, 0.000151843747486894, 0, 0.00011067898605789, -2.73769153342726e-06, -7.05249112816613e-05, 1.58619198503982e-05, 5.14987321934851e-05, 0.000163440831889741, 0.000116585936695139, 0, 0, 0, 0.000272511706406157, 0, 0.000175310459462252, 5.90599667332859e-05, -9.75040054787708e-05, 0.000124719134339695, 9.21519039362254e-05, 7.2357422709521e-05, -7.7799734034384e-06, -0.000401352023871821, -0.000216639503965571, -2.62082527368448e-05, 0.000234014211602919, -2.9929565708577e-05, -0.000151163479856924, -0.000269845707849877, -0.000528569582710868, -0.000445766877222237, -0.000757920593820967, -0.000777975029435455, -0.000752710697453293, -0.000615738756452719, 0, 0, 0, 0, 0, 0, -9.86370689295166e-07, 0, -1.93021528408564e-06, 0, 0, 0, -2.0537301419399e-06, 0, -7.02636038339041e-06, 248015.873018251, 2.4811028509486e-06, -1.70297812910913e-05, 0, 0, 0, -1.95058572032609e-05, 0, -6.1978281232933e-07, 8.96425697563196e-06, 1.83540559051199e-05, -5.49510266241871e-06, -8.01338641650642e-06, -4.0344854316818e-05, -3.54159456216969e-05, 0, 0, 0, -5.19060150485423e-05, 0, -2.33358565249898e-05, -4.18242852140924e-05, 3.59617900897901e-05, -1.80469187052821e-05, -8.41540355688461e-06, -5.22297287297125e-05, 1.07636694687381e-05, 8.23769359382428e-05, 4.38056447691904e-05, -4.8104316179445e-05, 0, 0, -8.95640861030189e-05, -0.000104295022298474, -3.22312707551531e-05, -0.000122637714897589, -4.86943850548662e-05, -2.42208318982438e-05, 3.21176441008503e-05, 9.92440264395661e-06, 5.62488309154809e-05, 0.00011052557817935, 0.000116076230228095, 1.38496356209581e-05, 0.00035155013802834, 0.000144049511908716, 0.000199347002469631, 0.000486069205351093, 0.000195786756664921, 0.000706582752094495, 0.000740063898548969}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -1.86767990581513e-08, 0, 0, 0, 0, 0, 0, 0, 2.89358136891645e-07, 0, 3.88966519342742e-07, -124.999999767989, 0, 0, 0, 0, 0, 0, -1.06232936304747e-06, 0, -3.22446296334463e-06, 0, 0, -4.70557650124382e-06, -1.26517196226541e-06, 3241.07142631706, -4.18801344159165e-06, 0, 0, 0, 0, 0, 0, 4.42911041104971e-06, 0, 1.39266601135272e-05, 0, 0, 0, 2.29703742076098e-05, 0, 3.95601891825987e-05, 1.45117971613883e-06, -32569.4444375005, 2.75744458766192e-05, 8.6646109932877e-06, 2.2859911024683e-05, 3.6648371008395e-05, 0, 0, 0, 0, 0, 0, -6.57054753687939e-06, 0, -2.98633258307496e-05, 0, 0, 0, -3.62052956217839e-05, 0, -0.000134579125474718, 2.36083329596879e-06, 167881.944421316, -0.000104143490944458, 0, 0, -0.000147245509009575, -9.74593509813532e-06, -5.01988173650176e-05, -2.69376732246596e-05, -6.64932958137755e-05, -0.000157319481477376, -0.000197266900652383, -0.000126370434868892, 0, 0, 0, 0, 0, 0, 1.03355369488371e-05, 0, 3.98528825032883e-05, 0, 0, 0, 2.48088983028847e-05, 0, 0.000245125810789415, -1.96762548962392e-05, -486111.111046671, 0.000198312889108952, 0, 0, 0, 0.000269843727919453, 0, 0.00045456132839017, -5.09338125247766e-06, 1.65962221845361e-05, 0.000111248883694483, 0.000167537476527374, 0.000425350332990649, 0.000648412440739791, 2.70431686402157e-05, 0.000347468976525412, 0.00048115464869628, 0.000306877409531667, 0.000272614010020803, 0.000856160852438319, 0, 0, 0, 0, 0, 0, -7.4218159442678e-06, 0, -3.73912690117598e-05, 0, 0, 0, 3.9751748951399e-06, 0, -0.000246829519659024, 2.32738280433623e-05, 798611.111033163, -0.000196437536893955, 0, 0, 0, -0.000244634421197555, 0, -0.000584125567702441, 4.94934109965658e-05, 2.88900099782826e-05, -0.000125395626888167, -9.80374185665209e-05, -0.000653995059349196, -0.000821066649464166, 0, 0, -0.000572285701024874, -1.6378682852173e-06, -0.000177593636724572, -4.56326463063464e-05, -0.000463674192243796, -0.000729116990604826, -0.00116095398471747, -0.000322429226530942, -9.36304075368779e-05, -0.00181149331530606, -0.000367607411605996, -0.0014607933208921, -0.000702422427811412, 0, 0, 0, 0, 0, 0, 4.004399601848e-06, 0, 1.53513008422233e-05, 0, 0, 0, -1.34668744182016e-05, 0, 0.000141806365696908, -1.31247069165167e-05, -694444.444404655, 0.000114589956555075, 0, 0, 0, 0.00010175565921234, 0, 0.000379045576882692, -4.7943208292267e-05, -4.65356142852549e-05, 9.15248478290953e-05, 3.34505026283252e-05, 0.000473855968556097, 0.000690570206983051, 0, 0, 0, 0.000382328929572022, 0, 0.000574688944544373, -3.71969411941997e-05, -1.54624477034721e-05, 0.000107507622136411, 0.000284962286596496, 0.000806802223031705, 0.00135481140542226, 7.24581973517862e-05, 4.01029287950374e-06, 0.0016315635088567, -7.41628702411718e-07, 0.000449363254140391, 0.000796690498795119, 0.000511616644484222, 0.000777723245072758, 0.000353030556439721, 0.000521637706991495, 0.00212533777959755, 0.00153348831364929, 0.00157679404102658, 0, 0, 0, 0, 0, 0, -2.38930437880559e-06, 0, -3.19558896568985e-06, 0, 0, 0, 6.04889156770764e-06, 0, -3.12599541959129e-05, 2.99541152269866e-06, 248015.873000419, -2.37522840586513e-05, 0, 0, 0, 8.41528772409667e-07, 0, -0.000130744548531867, 1.49294015245565e-05, 1.94381377701498e-05, -2.87965997755332e-05, -2.06706774240667e-05, -0.000115139167067614, -0.000171989157433911, 0, 0, 0, -0.000104012771927994, 0, -0.000225308172614198, 3.18027536968705e-05, 9.01052017429305e-07, -3.58930439630692e-05, 9.92389974707219e-06, -0.000283447230287684, -0.000537239949471163, 2.90048856440879e-05, -5.25996854991732e-05, -0.000652727169624018, 0, 0, -0.000191827671249325, 3.2130615844623e-05, -9.11268947426369e-05, 2.43427099796199e-05, -0.000223870406827296, -0.000291490397098585, -0.000488932827647289, -0.000100517893460764, -0.000181475271999698, -0.000122311279769034, -1.188037535708e-06, -0.00102024684172738, -0.000907951420300008, -0.000159343357317011, -0.00074114590078141, -0.000523876541614252, -0.000629027080557205, -0.000392270356988117, -0.000947091167098868}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -1.27433184434617e-08, 0, 0, 0, 0, 0, 0, 0, 1.78794959198315e-07, 0, 2.64703972438398e-07, -124.999999904963, 0, 0, 0, 0, 0, 0, -8.93904209413369e-07, 0, -2.21436022099303e-06, 0, 0, -3.29739952795858e-06, -1.35289057672525e-06, 1.2522460190489e-07, 3241.07142682385, 0, 0, 0, 0, 0, 0, 3.73905514145696e-07, 0, 9.0383373309205e-06, 0, 0, 0, 2.05924365295443e-05, 0, 2.51304389989847e-05, 7.37723072578808e-06, 1.34251959370084e-06, -32569.4444328056, 4.43877877421496e-06, 9.92480950837354e-06, 1.9726193885505e-05, 0, 0, 0, 0, 0, 0, -1.9040360644742e-06, 0, -1.76570546490689e-05, 0, 0, 0, -4.54967115014797e-05, 0, -9.60798513320395e-05, -2.48654862151002e-05, 1.11664802373126e-05, 167881.944407456, 0, 0, -0.000112400167276359, -1.95448832235727e-05, -4.40709663301217e-05, 2.24926997884023e-05, -5.54244832543951e-05, -5.86428477587087e-05, -0.000116191783572087, -7.24790951621389e-05, 0, 0, 0, 0, 0, 0, -6.32056694993427e-06, 0, 3.11861055245756e-05, 0, 0, 0, 7.92324358133616e-05, 0, 0.000184447219179598, 4.92141042310237e-05, -1.41820489250947e-05, -486111.111048187, 0, 0, 0, 0.000288292430463923, 0, 0.000269003832848504, 6.80921500183811e-05, 0.000120145133701589, -5.57709039105934e-05, 0.000117775036360361, 0.000158655037693486, 0.000401218309088423, -2.15738142136668e-05, 0.000185405986569559, 0.000303195438581067, 8.17402395043658e-05, 0.000148217749967657, 0.00030981642009698, 0, 0, 0, 0, 0, 0, 2.19767714019574e-06, 0, -2.20189044090049e-05, 0, 0, 0, -6.34347953659535e-05, 0, -0.00018811750007177, -5.75977932387208e-05, 2.68495812849197e-05, 798611.111051853, 0, 0, 0, -0.000301492064281601, 0, -0.000392758393003102, -0.000140869108653143, -0.000146050971781018, 7.32367115215495e-05, -3.62355785059924e-05, -0.000185047136087482, -0.000513620583970399, 0, 0, -0.000450838984983502, 1.14684966912277e-05, -0.000170875462910699, 0.000178687672614051, -0.000424585674376427, -0.000362197516509525, -0.00101808598457749, -0.000222044078589749, 0.000138332477663183, -0.00125746672061103, 0.000106672871493414, -0.000462536428380727, -0.000216904178223569, 0, 0, 0, 0, 0, 0, -3.96850790480024e-06, 0, 9.71804515563402e-06, 0, 0, 0, 2.67855466124881e-05, 0, 9.39140488708833e-05, 3.62166761790227e-05, -1.98311314981076e-05, -694444.444428312, 0, 0, 0, 0.000177185766528735, 0, 0.000249583511891516, 0.000126378444724695, 0.000112495144069741, -5.45329115315009e-05, 2.97429993332898e-05, 8.35810232444145e-05, 0.000442795495003648, 0, 0, 0, 0.000349334978631345, 0, 0.000373256304198658, 6.67465384781215e-05, 0.00016295811498901, -0.000123187596022197, 0.000155391345869278, 0.000340976654064679, 0.000774332540638556, 0.000477675356728714, -0.00022085571288282, 0.000810627203356776, -0.000153893160226856, 0.00031982961087854, 0.000761818150923888, -6.09198453881468e-05, 0.00039779337082575, -0.000314993538309418, 8.71230041271802e-05, 0.00125303045172416, 0.0010952297066191, -0.000516928323370275, 0, 0, 0, 0, 0, 0, 1.02788914422863e-06, 0, -2.10996259329266e-06, 0, 0, 0, -8.05401652392263e-06, 0, -2.14806148684899e-05, -9.1044870761633e-06, 5.95252448907475e-06, 248015.873014232, 0, 0, 0, -5.05827563794671e-05, 0, -8.84987555008234e-05, -4.0528204907179e-05, -3.47873257287851e-05, 6.28599450720371e-06, -1.46702019976352e-05, -1.58362719926431e-05, -0.000128105464254157, 0, 0, 0, -0.000142566159284976, 0, -0.00013889956061959, -4.18213016447145e-05, -5.77390456326776e-05, 5.84285031456444e-06, -0.000125543460089334, -7.74870962004473e-05, -0.000316833628506411, -7.93320563438348e-05, 7.10205004280361e-05, -0.000304660697058038, 0, 0, -0.000160432388376325, 3.74185214984773e-05, -7.54615067391973e-05, 6.2192309874033e-05, -0.000253179168808613, -0.000148831634835157, -0.000258670274215355, 7.96625491758772e-05, -3.78706623095403e-05, 0.000219504179959554, 0.000135946463859603, -0.0003177879038188, -0.000514662416106622, 0.000190887795303916, -0.000253820725928119, -0.000384123551126992, -0.000105587791947098, -2.26738754258133e-05, -0.000471619171207175}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.98300904653452e-08, 0, -2.2337696817399e-07, -71.4285715095079, 0, 0, 0, 0, 0, 0, 4.50418039206033e-07, 0, 2.53449577786333e-06, 0, 0, 1.95094731200822e-06, 1750.00000063003, 714.285715155085, 1.32247933763439e-06, 0, 0, 0, 0, 0, 0, -3.95258879893678e-06, 0, -1.4191823571993e-05, 0, 0, 0, -3.28068162555626e-06, 0, -1.6211508989396e-05, -16111.1111158918, 3.35238637717662e-06, -6.05073115249059e-06, -17500.0000066806, -1.16165969345836e-05, -2.8800468407482e-05, 0, 0, 0, 0, 0, 0, 2.74665671691333e-06, 0, 4.43014170999094e-05, 0, 0, 0, 6.30699455158993e-05, 0, 4.10976827283626e-05, 72916.6666828579, -1.54370781665401e-05, 5.1449593087566e-05, 0, 0, -3.013692803986e-06, 161111.111139757, 9.38010780067967e-05, -1.58243896094518e-05, 4.27854294103084e-05, 2.05921395923489e-05, 0.000149700864724752, 6.45888529775754e-05, 0, 0, 0, 0, 0, 0, -9.33092176157042e-06, 0, -4.43132091234168e-05, 0, 0, 0, 3.5771089618987e-06, 0, -0.000211193333245176, -173611.111119986, 7.66948714731883e-05, -9.77914119513699e-05, 0, 0, 0, 0.000238476412525925, 0, -8.40560363047174e-05, -729166.666852654, -6.66401043540736e-05, 7.74399392108082e-05, 5.60774642578296e-05, -9.37242203266126e-05, -0.000250386384467688, -5.13440597310717e-05, -0.000250332748231224, -0.000532934716290046, -0.00045685697403535, -8.47635670535702e-05, -0.00070038310750343, 0, 0, 0, 0, 0, 0, 3.77148677718898e-06, 0, 6.69738260355296e-05, 0, 0, 0, 3.41218567786485e-05, 0, 0.000195096019948571, 208333.333338594, -6.7956902495659e-05, 5.76672895878347e-05, 0, 0, 0, -3.83043782387077e-05, 0, 0.000102629779957416, 1736111.11142017, 6.75742668251615e-05, -0.00022840937991369, -7.37966139794405e-05, 0.000131344578952906, 0.000377963255891708, 0, 0, 7.17489196927847e-05, 0.000527027250221186, 0.000551203538061953, 0.000175422308174037, 0.000663091533607139, 0.000350577131773462, 0.000301690160652187, 0.0010945589611969, -0.000113266365401976, 0.000445578382359508, 0.000389168138945334, 0.001374478599858, 0.00163481845515908, 0, 0, 0, 0, 0, 0, -3.3001111448543e-06, 0, -3.24382540173905e-05, 0, 0, 0, -5.28852147275896e-05, 0, -0.000147692226260978, -99206.3491918308, 4.26962812843571e-05, -7.38498560752663e-05, 0, 0, 0, -4.59889848686246e-05, 0, -0.000236912949588133, -2083333.33356409, -3.88031647845072e-05, 0.000134593388206641, 6.6663193339577e-05, -9.54001653926613e-05, -0.000224778959089596, 0, 0, 0, -6.60004477848154e-05, 0, 2.36949384674183e-06, -0.000422808066842169, -0.000148634117187809, 0.000230868833218525, -5.34091786920062e-05, -9.3991938772662e-05, -0.000606448440118391, -0.000630721927960723, 0.000518782233566597, -0.000336411766136692, -0.000466401205146824, -0.000318971455358051, -0.000908941318648243, -0.00075033531157637, -0.0015643267040956, -0.000363862271983019, -0.00123632207612729, -0.000317616448053669, -0.0014540988458796, -0.0024279526817078, 0, 0, 0, 0, 0, 0, 8.16074310903241e-07, 0, 6.58552784010506e-06, 0, 0, 0, 3.66057123145718e-06, 0, 3.69479692141645e-05, -6.52607462981862e-06, -1.22256425858324e-05, 1.41337791324319e-05, 0, 0, 0, -1.02909470544042e-05, 0, 5.49163803390081e-05, 992063.492111035, -2.77242294413322e-07, -4.17448787139495e-05, -3.97167281036034e-05, -4.52843400506796e-06, 7.0444035404467e-05, 0, 0, 0, -8.54319017023849e-05, 0, 3.25946505050667e-05, 0.000212536184475292, 0.000116590698956265, -8.81556570667488e-05, -1.33087564445086e-05, 1.23208658883969e-05, 0.000213576942225465, 0.000227967253326442, -0.000149755053979431, 9.70885260987955e-05, 0, 0, -6.71907136773153e-06, 0.000217849558822861, 0.000168430431317944, 1.87282151755943e-05, 0.000253861960923098, 2.69539934143324e-05, 0.000304317497401325, 0.000889546465798124, 0.000749358781935648, -0.000123351161899468, 0.000441351112051616, 0.000584997484705927, 0.000496189230574839, 9.1789054580073e-05, 0.000521722959940222, 0.0010844872483577, 0.00118219370409448, 0.000647050062729106, 0.000674499466326058}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.18296736808428e-07, 0, -6.55746007753473e-07, -55.555555401202, 0, 0, 0, 0, 0, 0, -4.27868509337074e-06, 0, 6.14108157525161e-06, 0, 0, 2.39506634966054e-06, 1268.51851786823, 833.333328450486, 2.46511776704196e-06, 0, 0, 0, 0, 0, 0, 2.99139010924892e-05, 0, -2.66205522264431e-05, 0, 0, 0, 5.17887397012906e-06, 0, -3.80285355812999e-05, -10416.6666629936, -2777.77774058923, -2.65758462330018e-05, -19027.7777674786, -5.39188582317568e-06, 3.19195799688014e-07, 0, 0, 0, 0, 0, 0, -7.4119596022458e-05, 0, 6.44058187043059e-05, 0, 0, 0, -0.000142376928981777, 0, 0.000222708444761263, 39351.85182856, -0.00020546337866385, 5.22952060304712e-05, 0, 0, 0.000164481269046373, 156249.999996551, -2.47940622044054e-05, 63425.9258013303, 5.61715761429561e-05, 3.33574499245349e-05, 0.000146295851613216, 0.00021591363499807, 0, 0, 0, 0, 0, 0, 0.000141958736693512, 0, -0.00014594403165447, 0, 0, 0, 0.000129230904029039, 0, -0.000385465578213237, -69444.4444215939, 0.000470794915308996, -0.000133715133928981, 0, 0, 0, -0.00044083491992888, 0, -0.000337455551095883, -590277.777653544, -0.000169611660077926, 0.00064012899697762, 0.00018891195015212, 1.59758964661157e-05, -0.000592134042248718, -520833.333136982, 0.00014108408957035, -0.000434115400817375, -0.000386890334706539, -0.000937350060906791, -5.75234102246537e-05, 0, 0, 0, 0, 0, 0, -0.000132650026249844, 0, 9.840544202348e-05, 0, 0, 0, -0.000232149663911329, 0, 0.000564941032776084, 46296.2962419987, -0.000644195145565576, 0.000256648460614218, 0, 0, 0, 5.35362988511998e-05, 0, 0.000963327657851718, 1041666.66640321, 0.000158922685203807, -0.00113860135426835, -0.000997557948990237, -0.000116405561459192, 0.000942529718973971, 0, 0, 0.000209792130446781, 1967592.59189019, -0.000183454760611037, -0.00129038397452254, 0.000220693887737602, -0.000401368022187738, 0.00271656559418645, 0.000420795567169851, 0.000434014357895193, 0.00128987838692468, 0.00201151470276567, 0.000344319544232917, 0.00211038370078523, 0, 0, 0, 0, 0, 0, 7.55593711274479e-05, 0, -5.35687080016154e-05, 0, 0, 0, 0.000221449399953987, 0, -0.000331102280473326, 2.34616658254934e-05, 0.000412046743023221, -0.000100060319177584, 0, 0, 0, 8.87779207054955e-05, 0, -0.00068078575924341, -694444.444215735, -8.43170996707993e-05, 0.00106753374697088, 0.00107318063454774, 5.55417291672283e-05, -0.000680671977817761, 0, 0, 0, 8.49336462975676e-05, 0, -0.00041002881677216, -3472222.22184727, -0.000412644322229908, 0.00118039373815073, -0.000121771242008939, 0.000235795826192464, -0.00166451917906962, -0.000801286053858953, 0.000604228395207674, -0.000153151956255377, 0.00136514728090817, 0.000173239524704329, -0.00095053018047161, -0.0023169548041573, -0.000620186695655355, -0.00080827983365312, -0.00196722460050362, -0.00165208088725595, -0.00321904077958238, -0.00189683705248972, 0, 0, 0, 0, 0, 0, -1.79605419054661e-05, 0, 1.21702398489118e-05, 0, 0, 0, -4.84359317009246e-05, 0, 9.39864232396685e-05, -6.70552359801277e-06, -0.000109922619837195, 3.35838216495992e-05, 0, 0, 0, -9.91642379659395e-06, 0, 0.000242422226615666, -5.98929192462334e-05, 2.21689809547161e-05, -0.000317232428777876, -0.0003760428749442, 4.89511652017984e-05, 0.000233559734186923, 0, 0, 0, 0.000142540096655811, 0, 0.000152658363698795, 2314814.81462053, -5.26729087308042e-06, -0.000530348507051682, -0.000130065351022548, -7.39418624951468e-05, 0.00063472528164847, 0.00012145517513481, -0.000794421075780122, -1.91837778157351e-05, 0, 0, 5.30407702035357e-05, -0.000402095225397569, -2.51371373366418e-05, -0.000555767304287057, 0.000430675993360495, -2.37526679980192e-05, 0.000750057239396699, -0.000168057261085264, 0.000142084224211652, -0.000219789333753118, 0.000235099642452422, -0.00101207072421304, 0.000954141956941273, 0.000931293970578764, 0.000103797581377078, 0.0014439968071796, 0.00174938768652563, 0.00237403342478865, 0.00218520186456077}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -2.37844665915286e-08, 0, 0, 0, 0, 0, 0, 0, 1.8814012620809e-07, 0, 3.50781980323148e-07, -49.9999997503883, 0, 0, 0, 0, 0, 0, -1.27576401812135e-06, 0, -4.07513635651958e-06, 0, 0, 4.23595888375388e-06, 1041.66666242776, 916.666669503321, -4.90579383677494e-06, 0, 0, 0, 0, 0, 0, 4.13019507385885e-06, 0, 2.30242407328749e-05, 0, 0, 0, -5.18233116915632e-05, 0, 2.67074673325381e-06, -7291.66664225936, -5000.00001986397, 5.73160984879895e-05, -19097.2221934126, 4.53045637836269e-05, -4.2394245642922e-05, 0, 0, 0, 0, 0, 0, -3.31588813767858e-05, 0, -6.40201953271275e-05, 0, 0, 0, 0.000250059571470529, 0, -0.000160910193549149, 20833.3332765613, 8333.3334018968, -0.0002061230991072, 0, 0, 9.63279746770757e-05, 133680.555303996, -0.000120486309757721, 104166.666620247, 0.000388372495168483, -0.00034235808863031, -4.77065390950446e-05, -6.19247235987736e-05, 0, 0, 0, 0, 0, 0, 6.99971626476759e-05, 0, 0.000125896773197186, 0, 0, 0, -0.000275785937934996, 0, 0.00037849407852269, -20833.3332109567, -7.86950635198865e-05, 0.000479149892518146, 0, 0, 0, -0.000580215618717006, 0, -8.01084839290311e-05, -381944.443805709, 0.000307641108422641, -173611.111173943, -0.00129199897340504, 0.00108967558068626, 0.000756617545277575, -729166.665863052, 0.000656239771826018, -0.00125149693893634, 0.000348205561471971, -0.00075499096623029, 0.000282307417246564, 0, 0, 0, 0, 0, 0, -0.000102359105427998, 0, -9.34957018340374e-05, 0, 0, 0, 0.000199005112805843, 0, -0.000540102677625091, -0.000104858872281913, 8.58297796430476e-07, -0.000567564592853294, 0, 0, 0, 0.00143385556949633, 0, -0.000734566428633102, 381944.443618223, -0.000230355644579964, 9.22786475495844e-06, 0.0020349363359599, -0.0016194271264757, -0.00193275791963271, 0, 0, 0.00109400035852227, 2083333.3318064, -0.000714023828706216, 1215277.77699989, 0.0044419672303536, -0.0019701840666006, -0.000941400447926931, -0.000905770575077891, 0.00256634609807539, -0.00372151406361004, -0.000525164950001791, -0.000704675070720721, 0.0023382869795763, 0, 0, 0, 0, 0, 0, 6.04658738374953e-05, 0, 5.03037503777443e-05, 0, 0, 0, -6.00871091150399e-05, 0, 0.000304131332549385, 6.48597929752158e-05, 0.000102264051104161, 0.00027934378485784, 0, 0, 0, -0.00103822744539027, 0, 0.000838404624759221, 0.000585600094822119, 8.52957201017773e-05, -5.61176979786638e-05, -0.00117123743463473, 0.00121510150618833, 0.00146542961194689, 0, 0, 0, -0.00167742863022074, 0, -0.000279425893162255, -2083333.33161741, 0.000666296042465374, 0.000744907598406467, -0.00470926032261048, 0.00210597785121963, 0.00172038263569693, 0.00120133226980182, -0.00271987600808855, 0.0037575351481808, -3472222.22068, 0.00150721547033079, -0.00352240583442326, 0.00263631003169657, 0.00102270029151136, -0.000418121957999945, -0.00639848642101341, 0.00510695124668482, 0.000686497612687563, -0.00204537819807167, 0, 0, 0, 0, 0, 0, -1.40164766845906e-05, 0, -1.13213982686064e-05, 0, 0, 0, -1.87612932729202e-05, 0, -8.35286557354528e-05, -1.47892444829886e-05, -2.79592338250149e-05, -6.56675090572318e-05, 0, 0, 0, 0.000249206336118713, 0, -0.000275592042714152, -0.000175055109616264, -2.10711934040122e-05, -3.95596190486476e-05, 0.000205891872255902, -0.0003471651837941, -0.000464230283582857, 0, 0, 0, 0.000550477343825181, 0, -0.000146690401565701, -0.000579302240888276, -1.23574456335512e-05, -0.000209482253309034, 0.00147446197422483, -0.000777810767265192, -0.00106296647937131, -0.000310924192862698, 0.000988566670882553, -0.00159239751982636, 0, 0, 0.000363112306596599, 3472222.22139161, -0.000340327720217174, -0.000608569472948251, 0.00212663181383744, -0.000904994772498632, 0.000251638569767927, -0.000384622501362022, -0.000212770941002446, 0.000481962091092446, 0.00344909817078706, -0.00143002675553801, -0.00141975863842514, -0.00112499834564684, -0.00119921851890846, 0.00261739052798141, -0.00056540732484126, 0.0012104430036685, -0.00119686405986007}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 5.55806555841496e-08, 0, 0, 0, 0, 0, 0, 0, -7.30319030107201e-07, 0, -1.1652602207815e-06, -50.0000007597285, 0, 0, 0, 0, 0, 0, 4.44861112143195e-06, 0, 9.37455056871803e-06, 0, 0, 1.19834641269178e-05, 916.666667355656, 1041.66667779381, 9.25234397146837e-06, 0, 0, 0, 0, 0, 0, -1.82605346136528e-05, 0, -4.17242848427429e-05, 0, 0, 0, -7.0226267796226e-05, 0, -8.7084098663725e-05, -4999.99999175383, -7291.66674701438, -5.67036374077464e-05, -19097.222256526, -1.38139212488518e-05, -0.000143682751413593, 0, 0, 0, 0, 0, 0, 2.9816052427105e-05, 0, 0.000108072798226198, 0, 0, 0, 0.000196937918338884, 0, 0.000276952513053779, 8333.33329126244, 20833.3336210189, 0.00020207556535579, 0, 0, 0.000381440938767004, 104166.66664079, -7.81821155733373e-05, 133680.555813015, 0.000737092620732815, 4.86738435111954e-05, 0.000628633490852626, 0.000318271588386066, 0, 0, 0, 0, 0, 0, -3.73803250941607e-05, 0, -0.000161970670868718, 0, 0, 0, -0.000411928980376267, 0, -0.000508944986352925, 6.07675981448148e-05, -20833.3338729666, -0.000504798932290957, 0, 0, 0, -0.000969508478294912, 0, -0.000807268933252029, -173611.110892591, 0.00024173870137849, -381944.445255805, -0.00211365102086316, -0.000136153233604901, -0.00137095699991354, -729166.666852211, 0.000317535447379288, -0.00285338768149336, 3.30057744861572e-05, -0.00173176513524508, -0.00147806811292053, 0, 0, 0, 0, 0, 0, 2.59821453037324e-05, 0, 0.000158432562107326, 0, 0, 0, 0.000377706478096112, 0, 0.000544192392073732, -5.98923991612101e-05, 0.000579459922774346, 0.000554457107061557, 0, 0, 0, 0.00127824417550083, 0, 0.000937238185838905, -0.000409451429337066, -0.000224901852780685, 381944.445800086, 0.00256200428930033, 0.000316328679737106, 0.00186567107171502, 0, 0, 0.00154228942006656, 1215277.77738165, -0.000767638239766106, 2083333.33385458, 0.00523870342881546, -0.000369647823419241, 0.00493353489968614, -0.000425086837455419, 0.00384005415780339, 0.00290185222578873, 0.00104420014041452, 0.000236608576180785, 0.00456942249609702, 0, 0, 0, 0, 0, 0, -1.01174476639554e-05, 0, -7.92293788496981e-05, 0, 0, 0, -0.000188994716158459, 0, -0.00033954223963182, 3.19338323471661e-05, -0.00036718385705414, -0.000391722205718382, 0, 0, 0, -0.000956101869552911, 0, -0.000723800186431677, 0.000261919013511677, 0.00010130352267425, -0.000971122683085433, -0.00176747606542353, -0.000459217333536335, -0.00135404519018241, 0, 0, 0, -0.00149714007250824, 0, -0.00103119019673605, 0.000524610926011627, 0.000552873996688603, -2083333.33417656, -0.00462013617226743, 0.000401568290258922, -0.00392020516999432, 0.000570628541167577, -0.00346497924467638, -0.00186689498786876, -3472222.22217748, 0.00088326436260168, -0.00578014966764669, -0.000313964717027047, 0.000661723732816987, -0.00273161495962119, -0.00709626164082593, -0.00179140111836708, -0.0057538918804578, -0.00376723424973352, 0, 0, 0, 0, 0, 0, 2.92981070037639e-06, 0, 1.74562853022012e-05, 0, 0, 0, 3.29084356917528e-05, 0, 9.20381223708366e-05, -8.32223487603693e-06, 8.5355907264062e-05, 0.000104257646628067, 0, 0, 0, 0.000223802109583952, 0, 0.000194852371563657, -6.73954364427297e-05, -7.40559374987262e-06, 0.000312772149795884, 0.0004969049703263, 0.000115893851664496, 0.000431061446230956, 0, 0, 0, 0.000569005578899389, 0, 0.000327318728472549, -0.000188483199718549, -0.000146292479450926, 0.000358801436281073, 0.00170147124199306, -7.53844629744614e-05, 0.00125004881814931, -0.000203475169881417, 0.00135590874576453, 0.000858267203694259, 0, 0, 0.00061405021093849, -0.00025406666946036, -0.000452531240511839, 3472222.22217235, 0.00282450655754595, -0.000351462000556809, 0.00244202045120097, -0.000640393429704932, -0.000301216093595133, 0.00143764734990568, 0.00328135394591493, -5.42181228517303e-05, 0.00232679295226197, 0.000112331855739261, -0.000239829156936184, 0.00396224836393788, 0.000407001994069012, 0.00295216914579082, 0.00243812877823191}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 8.71737522708466e-08, 0, 0, 0, 0, 0, 0, 0, -1.00236189571652e-06, 0, -1.83496500922195e-06, -55.5555564795088, 0, 0, 0, 0, 0, 0, 6.66735683920721e-06, 0, 1.52986354106655e-05, 0, 0, 1.66144472880714e-05, 833.333338084767, 1268.51853488091, 1.5326549245705e-05, 0, 0, 0, 0, 0, 0, -1.28680182501122e-05, 0, -6.55496585335494e-05, 0, 0, 0, -7.07827718832301e-05, 0, -0.000110339817159456, -2777.77779056367, -10416.6667739317, -0.000116399402088391, -19027.7778571821, -2.96989928947087e-05, -0.000152080759987413, 0, 0, 0, 0, 0, 0, 3.32136039654063e-05, 0, 0.00017072433135954, 0, 0, 0, 0.000199054026815597, 0, 0.00044026441095117, 2.77268108705042e-05, 39351.8521753884, 0.000398590459060503, 0, 0, 0.000494173301102539, 63425.9261503793, 4.99548808485891e-05, 156250.000425692, 0.000890686490088636, 0.000199947176115659, 0.000911364247318154, 0.000532682175013913, 0, 0, 0, 0, 0, 0, -2.64226556585621e-05, 0, -0.000288669549501301, 0, 0, 0, -0.000272142733347562, 0, -0.000863026005183128, -4.89789190403197e-05, -69444.4450059078, -0.000775980447863449, 0, 0, 0, -0.00107362435072883, 0, -0.00120814540098417, -0.000363180227254921, -0.000169942449362341, -590277.778836284, -0.00183432996775493, -0.000709730766401591, -0.00250062093529574, -520833.334235046, 1.44909840380428e-05, -0.00312468545922974, -0.000709857958224512, -0.00243368655588292, -0.00121815312413926, 0, 0, 0, 0, 0, 0, 2.69374298358636e-05, 0, 0.000250056168566107, 0, 0, 0, 0.00019790576498988, 0, 0.00107451192889912, 3.23210141991569e-05, 46296.2969131429, 0.00105910988824007, 0, 0, 0, 0.00112034503630121, 0, 0.00193706555915266, 0.00042761003875342, 0.00019757147929771, 1041666.66810015, 0.00248216384232209, 0.00116968265402515, 0.00398587101251932, 0, 0, 0.00204834051178911, 0.00106591605166921, 5.72345170587869e-05, 1967592.59434804, 0.00568129860011126, 0.000179879492543867, 0.00659531050435009, 0.000363750853913145, 0.00491405900123945, 0.00333199380582825, 0.00324516669183886, 0.000954682627191944, 0.00703756392139992, 0, 0, 0, 0, 0, 0, -1.09885969984688e-05, 0, -0.000133337979823703, 0, 0, 0, -3.6850724626878e-05, 0, -0.000642246021229514, -8.81480204107912e-06, -0.000374178023562001, -0.000603007853062708, 0, 0, 0, -0.000511484262926006, 0, -0.00133373572668483, -0.000242091007930728, -0.000164381017094512, -694444.44552583, -0.00152468481108985, -0.000841693822457495, -0.00283179958610299, 0, 0, 0, -0.0015934151531265, 0, -0.00190414709678485, -0.000845962166869104, -0.000201811323159021, -3472222.22375684, -0.00459048817858575, -0.000550559734156037, -0.00547525436892479, -0.00063002593942396, -0.00444240610890006, -0.00385235192019351, -0.00117186171208026, 8.00116799470945e-05, -0.00736911784617402, -0.00133881188928017, -0.000407412358046631, -0.00374714873301855, -0.00738539661959763, -0.0015391240346836, -0.00832384500411058, -0.00161293789694154, 0, 0, 0, 0, 0, 0, -4.62606264184553e-07, 0, 2.75946432170444e-05, 0, 0, 0, 8.32628427528956e-06, 0, 0.000172626582451754, 2.03995740663077e-07, 8.41683197594634e-05, 0.000155442191080134, 0, 0, 0, 0.00015196008815254, 0, 0.000368282845212595, 5.51484866410105e-05, 6.03864984367201e-05, 0.000327065352933967, 0.00036136859113357, 0.000361268630025271, 0.000874036311997533, 0, 0, 0, 0.000487848004101175, 0, 0.00063333975318023, 0.000240355595585426, -4.25429872656859e-06, 2314814.81538057, 0.00145871573949266, 0.000265916774349564, 0.00194863061916984, 0.000247541998651909, 0.00132400998149252, 0.00121061479300501, 0, 0, 0.000743757667637266, 0.000485199601667416, -6.03690303028112e-05, 0.00074058271786708, 0.00289717939333836, 2.22663800059086e-05, 0.00305163645200594, 0.000443565721263278, -7.15452806191809e-05, 0.00172135060244682, 0.00339897092221406, 0.000971595185700589, 0.00374452747010075, 0.00140188117585712, 0.000115858818867119, 0.00462223360080894, 0.000821741616652352, 0.00255453550966854, 0.00242102418652525}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -3.78637238312345e-08, 0, 0, 0, 0, 0, 0, 0, 7.53272521989975e-07, 0, 7.03499784357427e-07, -71.4285712322469, 0, 0, 0, 0, 0, 0, -5.89502335217189e-06, 0, -3.8259544720388e-06, 0, 0, -1.47345310315647e-05, 714.285711832847, 1749.99999362575, -1.79130625975361e-06, 0, 0, 0, 0, 0, 0, 1.44581783166131e-05, 0, 6.73512463864559e-06, 0, 0, 0, 8.51262719755244e-05, 0, 8.05675544078862e-05, 1.455980453991e-05, -16111.1110692993, 2.80889775387626e-06, -17499.9999482773, 2.78116079381498e-06, 6.81893302474984e-05, 0, 0, 0, 0, 0, 0, -3.07149546787876e-05, 0, -1.10118498299025e-06, 0, 0, 0, -0.000279103681119366, 0, -0.000253186743674414, -4.09257534918419e-05, 72916.6665621649, 5.94361432784865e-05, 0, 0, -0.000473419073125344, -0.000202681793808409, -2.51351557302824e-05, 161111.110891644, -0.00052473914893732, 7.66244557627614e-05, -0.000315689643171374, -0.000442434368418156, 0, 0, 0, 0, 0, 0, 2.65509137733915e-05, 0, -1.53428775125977e-05, 0, 0, 0, 0.000507055120771141, 0, 0.000409348443346332, 9.96593947660142e-05, -173611.110991325, -0.000213718695307683, 0, 0, 0, 0.00117678360074153, 0, 0.00104361466088367, 0.000401061543320532, 5.27240346559481e-05, -729166.666303082, 0.00113824296632866, -0.000299050743820388, 0.000461860049558138, 0.000587579255442697, -0.000301812598909406, 0.00150613743800556, 0.00120051648141595, 0.00191595270800576, 0.000413548479116773, 0, 0, 0, 0, 0, 0, -1.62849271694919e-05, 0, 5.07557447361388e-05, 0, 0, 0, -0.000492250366835644, 0, -0.000400942570007981, -0.000114034176698065, 208333.333236291, 0.000243422615628519, 0, 0, 0, -0.00150506049227674, 0, -0.00152742789860533, -0.000516677710423226, 2.39283279769927e-05, 1736111.11073239, -0.00166944225334235, 0.00065949290029195, -0.000692312650944703, 0, 0, -0.00210161766283179, -0.000809111983688618, 0.000256291699248287, -0.000802246296710119, -0.00312447520335885, 0.000652720757841822, -0.00249404416636983, -0.00146319631267234, -0.00385467515108564, -0.000216150417407052, -0.0044073579152131, -0.00158689214035395, -0.00435441735666287, 0, 0, 0, 0, 0, 0, 2.83460951158911e-06, 0, -3.0738343566493e-05, 0, 0, 0, 0.00024264263840093, 0, 0.000155789841329387, 7.22276483288614e-05, -99206.3491603737, -0.000230507044912276, 0, 0, 0, 0.000973977555069714, 0, 0.0010347829011427, 0.000365977879891705, -3.28696569797199e-05, -2083333.3331151, 0.00110319383978445, -0.0006776513834132, 0.000119259828423068, 0, 0, 0, 0.00180733059668687, 0, 0.00179984643323384, 0.000625043181391208, -0.000273099815758445, 0.000409045279961199, 0.00286972428895138, -0.000798040565913051, 0.00150135797689587, 0.00116346331436081, 0.00332312584232507, -0.000135902897556278, 0.000469828767405462, -0.000630514956217998, 0.00338677453493047, 0.00333129604821284, 0.00196794342901084, 0.00465667751259148, 0.00479413102349912, 0.000735446603665466, 0.00310837312479552, 0.00402030981535807, 0, 0, 0, 0, 0, 0, 2.79445979180537e-06, 0, 9.40317505567667e-06, 0, 0, 0, -5.35496848069678e-05, 0, -3.25053243617982e-05, -1.98333167778686e-05, 3.64945268536975e-06, 6.47769682061937e-05, 0, 0, 0, -0.000295380361551085, 0, -0.000224657978355654, -0.000108439718603991, -4.50867553223993e-06, 992063.492017991, -0.000267743930920835, 0.000178336392145851, -2.18493249208748e-05, 0, 0, 0, -0.0005859558009042, 0, -0.000578327210169566, -0.000204460366942189, 0.000222239845290006, -9.49033637142879e-05, -0.00113160783665183, 0.00032024401502699, -0.000415716762965439, -0.000350996625881315, -0.000947961104172696, 0.000357297623520534, 0, 0, -0.000774130666072286, -0.000187390435936492, 0.000329000511754425, -0.000202504182890593, -0.00156542824660781, 0.000368801607793115, -0.00122374205140311, -0.00112140027702217, -0.000563769769559021, -0.00184510605419519, -0.00242549040075717, 0.000123197767968691, -0.000718490891115612, -0.00205884118635889, -0.000514577914320744, -0.00189108132013064, -0.00252786715470682, -0.00344321328239442, -0.0020384098262185}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -2.04512126325581e-08, 0, 0, 0, 0, 0, 0, 0, 6.52910644438726e-08, 0, 5.33477563534635e-07, -71.4285712263542, 0, 0, 0, 0, 0, 0, -1.72459407132113e-07, 0, -4.76931664791883e-06, 0, 0, -5.71652403480715e-06, -3.22791672100169e-07, 1749.99999942689, 714.285709604776, 0, 0, 0, 0, 0, 0, -4.83122672298715e-06, 0, 2.03259247710695e-05, 0, 0, 0, 4.64869131406267e-06, 0, 3.73949613583541e-05, 1.43629818302866e-06, -16111.1111065514, 3.86372173660694e-05, -1.13889281710434e-05, 2.94632312099343e-05, -17499.9999720254, 0, 0, 0, 0, 0, 0, 1.3909818661122e-05, 0, -6.13298995731088e-05, 0, 0, 0, -1.15801255862557e-06, 0, -0.000151136133203615, 1.11406565810151e-06, 72916.6666651592, -0.000129299211169459, 0, 0, -5.38813799700662e-05, 6.12689167092495e-05, -7.37595856912182e-05, 6.74258844763122e-05, 161111.110998934, -0.000154912611253907, -0.000246294190175479, -5.37879875034974e-05, 0, 0, 0, 0, 0, 0, -2.51722779121867e-05, 0, 9.02708608007319e-05, 0, 0, 0, -7.41431918133446e-05, 0, 0.000391933806790278, 8.21647896321875e-06, -173611.11116214, 0.000253203669632868, 0, 0, 0, -7.85654267410334e-05, 0, 0.000252202745112138, -0.000169314325802289, 0.000144226789789284, -0.000296705254940127, -729166.66670447, 0.000509876749850346, 0.000640611506510046, -0.000287953763139003, 0.000434438481930258, 0.000418297542461093, -0.000114526744127123, -0.000251678500717439, 0.00049384164934772, 0, 0, 0, 0, 0, 0, 1.68775796248981e-05, 0, -7.64397031038195e-05, 0, 0, 0, 0.00012385677055957, 0, -0.000497930670669556, -8.72528947538931e-06, 208333.333411594, -0.000229882734558953, 0, 0, 0, 0.000225157207105415, 0, -0.000593559486066452, 0.000166550961527041, -0.00012068732580569, 0.000225662331305676, 1736111.11111081, -0.000626951761458943, -0.00137032292096274, 0, 0, -0.000285631475384647, 0.000607181231523939, -0.000853356409868293, 0.000548399323943647, -0.000432044243954304, -0.00115059190018464, -0.000754427299279715, 0.00028599659190684, 0.000224574221075095, -0.00124976476093985, 0.000905554783865627, -0.000308534740619539, -0.000999662579607509, 0, 0, 0, 0, 0, 0, -4.80204052912556e-06, 0, 3.79103952300136e-05, 0, 0, 0, -8.529365516165e-05, 0, 0.000331431046923063, 4.10061625674405e-06, -99206.3492226624, 0.000125874765665623, 0, 0, 0, -0.000208901040194262, 0, 0.000637222972797004, -9.44535156345912e-05, 7.76082713452485e-05, -0.000125089790640308, -2083333.33328958, 0.00043465380085838, 0.000901814428153485, 0, 0, 0, -1.29986930265129e-05, 0, 0.00035755393189565, -0.000527457647991522, 0.000552236068245536, -0.000584530822269486, -6.40601789359775e-05, 0.000869121540081732, 0.000829663574529494, -0.000270171103990715, -0.000200222241521644, 0.00186642678586386, -0.000592555224984375, 0.00111689705223575, 0.000631375665702792, -0.00154565568804517, 0.000956751420228227, -0.00158460285194743, -0.000196594740000144, 0.00121440517695755, 0.00108189750116833, -0.000191421061058957, 0, 0, 0, 0, 0, 0, 3.63886424711511e-06, 0, -7.16430773590482e-06, 0, 0, 0, 2.35858541765349e-05, 0, -8.39791382138819e-05, 1.76636623689133e-07, 1.38019794444325e-05, -3.08269788189416e-05, 0, 0, 0, 5.75214825571985e-05, 0, -0.00015470812768808, 2.37875170283913e-05, -2.24607058638269e-05, 2.84380024848766e-05, 992063.492097506, -0.000115947146430856, -0.000347528286711878, 0, 0, 0, 0.000100669083120671, 0, -9.87476040227042e-05, 0.000110272813670151, -0.00010514428100261, 0.000146537018784713, -0.00011182159635836, -0.000327594960480411, -0.000312514494552662, 3.23952336550024e-05, 0.000129893333041989, -0.000646858468829159, 0, 0, -4.83925193265057e-05, 0.000247508909717231, -0.00035695364482167, 0.000236901900875535, -3.04223542608205e-05, -0.000432910149827363, -0.000320811658126751, 0.000427078551556139, -0.000214529066954845, 0.000763639923168712, 8.05501372908293e-05, -0.00103385363261181, -0.00055401785814969, 0.00101535366794347, -0.00069482213349745, 0.000201260853164392, 0.000572812535349115, 0.000864140821410915, -0.000487414852402608}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 4.73767993694609e-08, 0, 0, 0, 0, 0, 0, 0, -7.901863751815e-07, 0, -7.67284436266996e-07, -55.5555564220245, 0, 0, 0, 0, 0, 0, 7.36611665661433e-06, 0, 5.05272595533149e-06, 0, 0, 1.3860312888166e-05, 2.89380845735747e-06, 1268.51853238922, 833.333344621802, 0, 0, 0, 0, 0, 0, -2.82608796968412e-05, 0, -1.76037463553594e-05, 0, 0, 0, -7.56150451576968e-05, 0, -5.70967630291885e-05, -6.37736927367767e-06, -10416.666774249, -2777.77784696281, -4.8214585536618e-05, -4.93803713624734e-05, -19027.7779284814, 0, 0, 0, 0, 0, 0, 7.18911837292553e-05, 0, 5.52536014175558e-05, 0, 0, 0, 0.000301887265509829, 0, 0.000129830012826763, 8.09642543019935e-06, 39351.8522276245, 0.000237379059974236, 0, 0, 0.000235084114272101, 0.00010339780129037, 4.37940333437813e-05, 0.000365165641778425, 156250.00079706, 0.000165636797999881, 63425.9266894737, 0.000457023808579068, 0, 0, 0, 0, 0, 0, -9.79766582565934e-05, 0, -7.14576335695871e-05, 0, 0, 0, -0.000460076305488748, 0, -0.000266257177882463, -2.43707801848269e-06, -69444.4451407066, -0.000455292667690431, 0, 0, 0, -0.000570334754828688, 0, -0.000567594893245244, -0.00010368386343709, -8.61578043328194e-05, -0.00091105466086415, -590277.779541204, -0.000424516962533347, -0.00182944694675491, -0.000733540281961264, -0.000259590858950518, -520833.33610587, -0.000533969108005675, -0.00196739414096602, -0.00212944154461492, 0, 0, 0, 0, 0, 0, 9.64277999630432e-05, 0, 7.104306123551e-05, 0, 0, 0, 0.000503686893769313, 0, 0.000245553556124803, -3.54873570409709e-06, 46296.2970425339, 0.000374247049355451, 0, 0, 0, 0.000890979245612883, 0, 0.000640712741365298, 0.000181964422863191, 2.98021935928152e-05, 0.00176003628780637, 1041666.66926586, 0.000462135680082933, 0.00259328185854395, 0, 0, 0.00133480540844952, 0.000443227492873604, 0.000509373929492855, 0.00156661469655877, 1967592.59776021, 0.000456980308163416, 0.0043943262301675, 0.000517263977835583, 0.00430912163116568, 0.00315888977912236, 0.00229016988560808, 0.00117021811855779, 0.00700935984300741, 0, 0, 0, 0, 0, 0, -5.04020439337071e-05, 0, -3.66987555199368e-05, 0, 0, 0, -0.000284479960571435, 0, -0.000165756196926085, 8.56775120900373e-06, -0.000505325013053006, -0.000267471938613829, 0, 0, 0, -0.000691096096919336, 0, -0.000460888909277348, -0.000126464037466951, -4.03395460941287e-05, -0.00139381059012162, -694444.446483406, -0.000475082496126, -0.00161229164917205, 0, 0, 0, -0.00106763487354718, 0, -0.000944783190219403, -0.000147086214896937, -0.000315552599317115, -0.00144164350258408, -3472222.22599794, -7.69437778899805e-05, -0.00385963390723107, -0.000103792498750913, -0.00395265729859447, -0.00343049979263751, -0.00122406392999535, -0.000386311048796756, -0.0057938484561522, -0.00130876013919452, -0.00143928867622439, -0.0032014424025646, -0.00748412283547287, -0.00102702596177349, -0.00707174103673648, -0.0061868913561762, 0, 0, 0, 0, 0, 0, 5.78741817361127e-06, 0, 9.3076783955129e-06, 0, 0, 0, 6.71581725491071e-05, 0, 2.38800639593303e-05, -7.36448306116698e-06, 0.000113860206509193, 7.52550886368589e-05, 0, 0, 0, 0.000216271105091976, 0, 3.39723109714297e-05, 3.71107896111752e-05, -4.42543139047008e-06, 0.000435354788545238, 0.000533452802635972, 0.000135810449410428, 0.000510717401089533, 0, 0, 0, 0.00032908070092278, 0, 0.000120226534627406, 0.000152075680199012, 0.000119971444731358, 0.00066654843528436, 2314814.81630749, 7.25081868079222e-05, 0.00111311381919598, 7.11026612886431e-05, 0.00132842249257471, 0.0010306338807901, 0, 0, 0.000400349424196177, 0.00027747866415744, 0.000206853589202438, 0.000562981244384021, 0.0020821243895897, -2.35902754195157e-06, 0.00236270758814526, 0.000614143844835404, 0.000119628496846044, 0.0013391860591247, 0.00294452277799467, 0.000923700528710361, 0.00278617667016021, 0.000953256231852192, 8.85821049749433e-05, 0.0037519171342135, 0.00105675151470919, 0.00246911353542592, 0.002995589174849}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 2.62655813609009e-08, 0, 0, 0, 0, 0, 0, 0, -1.91092251121181e-07, 0, -5.62663358682841e-07, -50.0000007016713, 0, 0, 0, 0, 0, 0, -4.81657035252163e-07, 0, 5.81754889663273e-06, 0, 0, 1.28859752393675e-06, 2.98908304875725e-06, 1041.66668024272, 916.666675187365, 0, 0, 0, 0, 0, 0, -5.16902475110361e-06, 0, -3.12929838520545e-05, 0, 0, 0, 1.98008650993972e-06, 0, -3.17349963863989e-05, -8.68122221184929e-06, -7291.66673128655, -5000.00004876944, -8.62019811109659e-05, -1.70063292489163e-05, -19097.2223586649, 0, 0, 0, 0, 0, 0, 6.61099119734828e-07, 0, 7.1921494043421e-05, 0, 0, 0, -8.45933742490375e-05, 0, 0.000132924618982839, 3.91330266347807e-06, 20833.3335508983, 8333.33347866939, 0, 0, 4.54195668206248e-05, 0.000186876398828355, 0.000106595780263521, 0.000415347801937546, 133680.556177997, 9.24156544330009e-05, 104166.667236735, 0.000575701133244194, 0, 0, 0, 0, 0, 0, -3.3424309907773e-05, 0, -0.000120207013157753, 0, 0, 0, -5.58089732064899e-06, 0, -0.000323543027517551, 6.1219749649332e-07, -20833.3337819336, -0.000282974129593916, 0, 0, 0, -1.4486065522159e-05, 0, -0.000192577372033058, -0.000346088907696021, -2.84234563037892e-05, -0.00132454101744224, -381944.446218231, -0.000245729554474748, -173611.11233805, -0.000657706498359562, -7.68966308512226e-05, -729166.668983627, -0.00118686135801086, -0.00251726944886125, -0.00130713689426475, 0, 0, 0, 0, 0, 0, 7.24022099682159e-06, 0, 0.000118033475760765, 0, 0, 0, -9.71113327638156e-06, 0, 0.00039986579140072, -3.05911703163525e-05, 0.000562714566942159, 0.00041608674152418, 0, 0, 0, 1.82523055323857e-05, 0, 0.000463379211946346, 0.000221027416796587, 0.000120666065321521, 0.00157137808741203, 381944.446607828, 0.000344811885736715, 0.00166238971756635, 0, 0, -0.000130215473587653, 0.00145112105673019, 9.29686379778015e-05, 0.0018677717757069, 2083333.33670277, 8.15716773170195e-05, 1215277.78186091, 0.000910615182120941, 0.00504360132340719, 0.00267809122081514, 0.00423832978413389, 0.00313537288409203, 0.004242264572957, 0, 0, 0, 0, 0, 0, -3.52406533239962e-06, 0, -6.44672990225902e-05, 0, 0, 0, -5.49561933403057e-05, 0, -0.000226285019625607, 3.99057696684501e-05, -0.000245244812576943, -0.000182498192962312, 0, 0, 0, 2.0182180994842e-05, 0, -0.000517432002530002, -0.00013196465628154, -1.37959771782265e-05, -0.00116505181466005, -0.00122671367064274, 3.79705596257544e-06, -0.00117282267415664, 0, 0, 0, 2.86254132182167e-05, 0, -0.000141306610936899, -0.000918833069609568, -0.000117373519596775, -0.00199206830317685, -2083333.33683361, 9.73444475943002e-05, -0.0036347746584568, -0.00026658914674712, -0.00475963878998498, -0.00157581740101147, -0.00120096247393098, -4.03483479086938e-05, -3472222.22602492, -0.00209726190205059, -0.000518092096380111, -0.0060156037653168, -0.00741102794156332, -0.00281406731426388, -0.00630146872543959, -0.00446168476960631, 0, 0, 0, 0, 0, 0, 6.64311605925855e-06, 0, 1.15119992515772e-05, 0, 0, 0, 6.67025540331284e-06, 0, 8.7576101975118e-05, -8.82601285301859e-06, 7.33581443429901e-05, 3.32305663354634e-05, 0, 0, 0, -5.24314830517234e-05, 0, 0.000233188788923961, 2.71331576888443e-05, 2.0290708914257e-05, 0.000307085930653433, 0.000386135755933056, -2.11104583473852e-05, 0.000303921649217737, 0, 0, 0, -0.000126518503134428, 0, 0.000292710708375563, 0.000135148651790238, 1.61249712151055e-05, 0.000605799084266696, 0.000801980535111893, -0.00010231867673113, 0.00135623033239798, 3.79404722636104e-05, 0.00179590460449902, 0.000832091158673354, 0, 0, -4.35518460030989e-06, 0.000430152573061803, -3.61163447177159e-06, 0.000613255840826321, 3472222.22404185, -6.58474733833553e-05, 0.00179239098955498, 0.000382190107005249, 0.000566966645717173, 0.00320860747037521, 0.00377062028948038, 0.000889042879939969, 0.00286111846721772, 0.00172126692428843, 0.00105381137270615, 0.00460615443468046, 0.00168848431058522, 0.00499983972406167, 0.00239405975656148}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 1.05603146818453e-08, 0, 0, 0, 0, 0, 0, 0, -1.0097657347131e-07, 0, -1.9303838576844e-07, -50.0000005660128, 0, 0, 0, 0, 0, 0, 2.84288738592626e-06, 0, 1.82184830111989e-06, 0, 0, -3.92593852224693e-06, 4.8098190345198e-06, 916.666675543264, 1041.66667570028, 0, 0, 0, 0, 0, 0, -2.90106061155583e-06, 0, -1.00496066376837e-05, 0, 0, 0, 2.38623929017635e-05, 0, 2.56030585975796e-05, -4.2431206126716e-06, -5000.00007435137, -7291.66672634699, -6.16344416115094e-05, -3.81896836346466e-05, -19097.2223310944, 0, 0, 0, 0, 0, 0, 3.72498966393087e-05, 0, 5.19623548209989e-05, 0, 0, 0, 2.08319751945438e-05, 0, 1.05894342250719e-05, 4.93786804438848e-06, 8333.33354938231, 20833.333515016, 0, 0, -9.75785216554431e-05, 0.000203302689528772, -6.15759077404419e-05, 0.000399040225430793, 104166.667207137, 0.000174832774665293, 133680.556129014, 0.000536220552224786, 0, 0, 0, 0, 0, 0, -3.38737907954953e-05, 0, -9.68718324287395e-05, 0, 0, 0, 8.13367881395818e-05, 0, -7.1710016698491e-06, 1.40896438669692e-05, -0.00030118904894674, -20833.3336336167, 0, 0, 0, 0.00044594243918862, 0, 0.000321432377428123, -0.000244300431683971, 8.80369831787327e-05, -0.000872584735915463, -173611.112069201, -0.00027324576354093, -381944.446111698, -0.00126681320212012, 9.67126120398013e-05, -729166.668336841, -0.00068556050814842, -0.00257272219381821, -0.00250977016289519, 0, 0, 0, 0, 0, 0, 8.40488515902775e-05, 0, 0.000100667140056037, 0, 0, 0, 6.33543687826913e-05, 0, 0.000108717416585617, 1.01885501644894e-05, 0.000212854352846507, 0.000279323793598287, 0, 0, 0, -0.000752866332627167, 0, -0.000352658225129694, 0.000346785112861442, -0.000243645036909916, 0.00131746442964117, 0.00118498505756588, 0.000352405645867555, 381944.446891272, 0, 0, -0.000772876273650302, 0.000641277112401463, -0.000594108278140288, 0.00183245605952235, 1215277.78075843, -0.000232579439232057, 2083333.33677396, 0.000386692894967806, 0.00469948065144355, 0.00434912534944142, 0.00367217560720537, 0.00111317991838984, 0.00855680765947338, 0, 0, 0, 0, 0, 0, -4.49472782188842e-05, 0, -5.5153922816292e-05, 0, 0, 0, -3.10315337168783e-05, 0, -0.000102656306623931, -2.06452214774223e-05, -0.000221167787716657, -0.000211724350212788, 0, 0, 0, 0.000431614499568212, 0, 0.000269360918054127, -0.000211999661746842, 9.86222197336712e-05, -0.000804099388060749, -0.000868260389437868, -0.000387264621695899, -0.00188445153230679, 0, 0, 0, 0.00102344490097509, 0, 0.000712555485122869, -0.000581447282857955, 0.000507257799942849, -0.00151772532103476, -0.0011848239979367, -5.06561230258341e-05, -2083333.33621582, -0.000328146587795335, -0.00379271072367857, -0.00437122205096292, -0.00169872619189283, 0.000743972117699332, -3472222.22560102, -0.00314376693663953, -0.00145143821179457, -0.0053250207844439, -0.0071140889599797, -0.00306727983443537, -0.00744538593041071, -0.00630593187983178, 0, 0, 0, 0, 0, 0, 4.74674832520412e-06, 0, 1.61256068707847e-05, 0, 0, 0, 3.63650283077179e-05, 0, 3.05522339768943e-06, 2.93031504386847e-06, 2.49604120483473e-05, 5.61135312959918e-05, 0, 0, 0, -6.97143232348956e-06, 0, -8.16797627845908e-05, 6.3738897815012e-05, -3.77104459751028e-05, 0.000260458978831218, 0.000172011230952864, 0.000113422804303325, 0.000495625900476952, 0, 0, 0, -0.000181907823311711, 0, -0.00025376031393595, 0.000315245997977381, -0.000123998732476155, 0.000641137433337301, 0.000756939104533147, -9.46207344793159e-06, 0.00105391443212269, 4.40027663334073e-06, 0.00117062315249903, 0.00104096459109249, 0, 0, -0.000323351304148844, 0.000486628746693858, -0.00034999040467188, 0.000769891668355997, 0.000739253044228109, -0.000193263790506949, 3472222.22329387, 0.000655624113988145, -0.000595163321577218, 0.0019388031314181, 0.00258231823163655, 0.00102938279900924, 0.0030732911209894, 0.0022350230713783, 2.50428522745502e-05, 0.00278330859983243, 0.00246418084953154, 0.00292381160351559, 0.00420957440191759}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9.56541148726922e-08, 0, -1.90030872561306e-07, -55.5555556309744, 0, 0, 0, 0, 0, 0, -3.00862628281963e-06, 0, 1.90293859392338e-06, 0, 0, 6.71125423021904e-06, -1.45101748436853e-06, 833.333336823679, 1268.51851863074, 0, 0, 0, 0, 0, 0, 6.28756297370464e-06, 0, -8.07837987160584e-06, 0, 0, 0, -1.39723092090654e-05, 0, -4.47439388916673e-05, -4.81556399641405e-06, -2777.77777958359, -10416.6666543573, -1.05796873630419e-05, 6.15434624529901e-06, -19027.7778278108, 0, 0, 0, 0, 0, 0, -3.74197664954677e-05, 0, 4.05954112962173e-06, 0, 0, 0, -5.9929815153216e-05, 0, 9.03258827312313e-05, 1.76928280331674e-05, -1.30252043955307e-05, 39351.8518083748, 0, 0, 6.40671728876756e-05, 1.27291729059872e-06, 3.43815582024451e-05, 4.59155535531112e-05, 63425.926154074, -9.49582991267689e-05, 156250.000259039, 0.000134393314883923, 0, 0, 0, 0, 0, 0, 3.56773263769103e-05, 0, 8.77230767015672e-06, 0, 0, 0, 8.6575902893191e-05, 0, -0.000301640081501804, -6.41926866527709e-05, 2.9402337805487e-05, -69444.4443333527, 0, 0, 0, -7.33046305792274e-05, 0, -0.00051895325348137, -4.73204966413081e-05, -1.7967797982371e-05, -0.000138960445480583, -0.000503277450867652, 0.000178898625821381, -590277.778253207, 0.000154153232274098, 4.90845129839615e-05, -520833.334554715, -0.000294815291820079, -0.000499965726772698, -5.7016644679431e-05, 0, 0, 0, 0, 0, 0, -7.01881258522278e-05, 0, -1.97416169150048e-05, 0, 0, 0, -0.000195105700504785, 0, 0.000264423800890224, 6.23007653006232e-05, -3.61826289451776e-05, 46296.2960865774, 0, 0, 0, -2.98664083851298e-06, 0, 0.0010004324094189, 2.58845242308714e-05, 0.000105444036701175, 0.00013698160208622, 0.000469823848184995, -0.000436624613090723, 1041666.66709854, 0, 0, 0.000490791406522522, 0.000488837519082884, 0.000110550271954929, 0.000369988609323875, 0.00168071412753953, -0.000228139486795358, 1967592.59461923, 0.00038300791840064, 0.000983600755865875, -1.42799479102767e-05, 0.000541585547696072, 0.000957263454687735, 9.50407451272298e-05, 0, 0, 0, 0, 0, 0, 3.36757509833444e-05, 0, 1.47430272123744e-05, 0, 0, 0, 8.99763487606709e-05, 0, -0.000140970776787791, -3.73906497998139e-05, 0.000143483777299601, 0.000199887998005184, 0, 0, 0, 0.000134736360729026, 0, -0.0009012581000571, -2.43543181442484e-05, -0.000101966821332381, -0.000124091751834233, -0.000139285071170132, 0.00048487308702564, -694444.444366878, 0, 0, 0, -0.00020061364522668, 0, -0.000893520361341414, -0.000139970214914133, -9.43990741172263e-05, -0.000383804253381108, -0.00179822408031602, 0.000613521562639551, -3472222.22429049, 3.24798934609838e-05, -0.0010018901749498, 0.000340325016783275, -0.000191884134419151, 3.76861737110306e-05, -0.00195928881318699, 0.000188328801500093, 0.000590207505807699, -0.000270153280460622, -0.00245399042433389, 0.000920182818954897, -0.00210007637812557, -0.00113869604096065, 0, 0, 0, 0, 0, 0, -4.7805069485611e-06, 0, -6.58041901243369e-06, 0, 0, 0, -3.77017222795169e-05, 0, 4.53226135052245e-05, 1.14788276566172e-05, -2.12573018096254e-05, -6.12034250133209e-05, 0, 0, 0, -0.000130520161903277, 0, 0.000256448493006685, 1.02405407846457e-05, 4.9747871799141e-05, -2.28849254935967e-05, 1.69282652998255e-05, -0.000173175942647875, 1.95995355478484e-05, 0, 0, 0, -0.000165818665565241, 0, 0.00035703217714938, -6.17237458776158e-05, 3.01618131845341e-05, 5.61642895571378e-05, 0.000104213197304764, -0.000202508100141221, 2314814.81547201, 8.29683684035696e-05, 0.000200177528166112, 0.000171429032471056, 0, 0, 0.000126717453639513, 0.000103761153522971, 8.55471297790758e-05, 0.000115096725511937, 0.000929965231479869, -0.000187935398241332, 0.00137450798631994, 0.000189346716074279, 0.000369630524894538, 0.00051622304340996, 0.00115518064018772, 0.000110300929929525, 0.00078838933396351, -1.83498835673331e-05, -4.66797939151752e-05, 0.00234964798187441, -0.000744635605828685, 0.00122655837627146, -0.000416383642092165}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6.2208538815649e-08, 0, -1.08448941727603e-07, -71.4285714272813, 0, 0, 0, 0, 0, 0, -5.38016617090531e-07, 0, 1.63246125209815e-06, 0, 0, -3.14617059189637e-06, 2.03663217530303e-06, 714.28571360757, 1749.99999980831, 0, 0, 0, 0, 0, 0, 6.7556968171117e-06, 0, -8.91127796436177e-06, 0, 0, 0, 1.46529439223282e-05, 0, 8.05915058629884e-06, -6.72877241770727e-06, -2.57890070299103e-06, -16111.1111100928, -1.12988141184357e-05, -5.78847976805041e-06, -17499.9999938587, 0, 0, 0, 0, 0, 0, -8.30250521222977e-06, 0, 2.88910590813705e-05, 0, 0, 0, -4.18406051148345e-05, 0, 4.00490174888375e-05, 2.43161254889476e-05, 1.81978433235786e-06, 72916.66664702, 0, 0, -3.18269834508652e-05, 6.67804023640488e-05, 4.79139625159965e-05, 2.81893340675937e-05, -2.82452041107495e-05, 4.93009799215006e-07, 161111.111061047, 6.01122141196004e-05, 0, 0, 0, 0, 0, 0, 2.40865974858977e-05, 0, -5.66689901217922e-05, 0, 0, 0, 7.0143374526103e-05, 0, -6.07654520130009e-05, -3.23700924666629e-05, -1.71118963133047e-06, -173611.111055544, 0, 0, 0, 0.000158027955957543, 0, 0.000177265558114789, -0.000162100724214581, -0.00018564968236229, -8.44572615999845e-05, 4.1904558065036e-05, 0.00012494808710382, -729166.66663195, -0.000271722692732334, -7.33713940596497e-05, 0.0002298005350581, -0.000301155781371542, -0.000310329945596735, -0.000221242566819946, 0, 0, 0, 0, 0, 0, -9.89061998575954e-06, 0, 5.27081826589989e-05, 0, 0, 0, -3.54234871461382e-05, 0, 0.000125574673155435, 3.43215137977902e-05, -1.68240153443538e-05, 208333.33329491, 0, 0, 0, -0.00022753998547327, 0, -0.000139474813240382, 0.000218326707079485, 0.000249007706367374, 3.45259909769214e-05, -4.32843055961151e-05, -0.000234331052127172, 1736111.11108263, 0, 0, -0.000292298817652922, 0.000188617047307467, 0.000285839344565887, 0.000112037079830955, -0.000437878780719427, -0.000169603557360505, -0.000143813677855924, 0.000786163273774906, 0.000315046748835374, 0.000423511066096309, 0.000660136694204625, 0.000320489699838113, 0.000785271523203204, 0, 0, 0, 0, 0, 0, 8.39052906836413e-06, 0, -2.6630033297486e-05, 0, 0, 0, 2.79181509711975e-05, 0, -8.63313343852067e-05, -1.7444060608457e-05, -1.24156172002233e-05, -99206.3491895107, 0, 0, 0, 0.000130732952658354, 0, 0.00013056811853315, -0.000154675380715283, -0.000159951238692047, 2.20357578615263e-05, 6.05419525541646e-06, 0.000195367490778627, -2083333.33352322, 0, 0, 0, 0.000329336559791652, 0, 0.000335505583551577, -0.000285422668140854, -0.000394916795453971, -7.49863716973071e-05, 0.000348186033596919, 0.000126297066530975, 0.00031523346745022, -0.000879085103533908, -0.000168521448539326, -6.63473452486619e-05, -0.000191622080149348, -0.000117459880073849, 0.000323142061633021, -0.0013148033485035, -0.00131989033968427, -0.000883585277725188, -0.00016585686546154, -0.000938622610720733, -9.41402944366197e-06, -0.000330309491880351, 0, 0, 0, 0, 0, 0, -2.65527746975003e-06, 0, 7.18335374047399e-06, 0, 0, 0, 5.59899080443233e-07, 0, 1.84101552475757e-05, 2.62773673435682e-06, -3.59577095225921e-06, -1.93911345489222e-06, 0, 0, 0, -7.34979013370853e-07, 0, -7.44274647831354e-06, 3.64491272847923e-05, 3.4669263854459e-05, 1.26398378200255e-05, 4.32678821774136e-06, -6.55099947862188e-05, 992063.492070625, 0, 0, 0, -2.92652329750438e-05, 0, -8.45516537496911e-05, 0.000131336080761147, 0.000146293496335959, 3.17621107659374e-05, 7.90773196768603e-05, -0.000137154654633291, -7.62739877919733e-05, 0.000196707755827485, 0.000101250214860727, -0.000126276733328976, 0, 0, -0.000127172677118287, 0.00010495277284117, 0.000116686884508463, 8.19317116591735e-05, -0.000181009437750272, 9.93451839911432e-06, -0.000386778072870685, 0.000342895752797751, 0.000465635528560698, 0.000231628205444623, 0.000101855455404211, -0.000165554303630847, -4.6037975633163e-05, 0.000561201979828624, 0.000444029575230483, -0.000360088492751214, 0.00128956114160171, 0.000299297169177609, 0.000563950740932044}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.88832985389293e-07, 0, 3.04901270034088e-07, -71.4285714128604, 0, 0, 0, 0, 0, 0, -1.81064954679186e-06, 0, -1.57084346327103e-06, 0, 0, -5.95104310725752e-06, 714.285715560596, -2.97246499701803e-06, 1749.99999940071, 0, 0, 0, 0, 0, 0, 7.29845451081926e-06, 0, 3.21935306824027e-06, 0, 0, 0, 2.56000516282657e-05, 0, 3.48107315956937e-05, -1.05151786837021e-05, 2.63963421238325e-06, -16111.1111001197, -1.68334315225047e-05, -17500.0000062965, -3.90725656832131e-06, 0, 0, 0, 0, 0, 0, -1.46671911030108e-06, 0, -1.84996149784269e-05, 0, 0, 0, -0.000108692801911636, 0, -0.000109922462906161, 1.94489002780852e-05, -4.73406490537654e-05, 72916.6666355669, 0, 0, -0.000197817668298146, 6.3129903693287e-05, 8.23246707494269e-05, -4.9708679113052e-06, 1.73949614709506e-05, 161111.111112181, -0.00010043694793958, 0.000241504591991332, 0, 0, 0, 0, 0, 0, 4.49920945518203e-06, 0, -1.55721365480666e-06, 0, 0, 0, 0.000110356632318129, 0, 0.00011471492198535, -2.32075059872211e-05, 4.66328608832076e-05, -173611.111076528, 0, 0, 0, 0.000211342783498854, 0, 0.000508531153750105, -8.86813351630743e-05, -0.000228811951860411, 5.17417331307323e-05, -0.000163872207491782, -729166.666666978, 0.000309294924138977, -0.000152944475228369, -0.000264504310129188, 0.000181185275432683, -0.000642042910698877, -0.00106615285785524, -1.15213128356462e-05, 0, 0, 0, 0, 0, 0, 1.53577612174268e-05, 0, -6.38168753229388e-06, 0, 0, 0, -0.000136311305253814, 0, -0.000163212200238418, 8.70082529282793e-06, -6.94743459942638e-05, 208333.333312386, 0, 0, 0, -0.000374781710347684, 0, -0.000632486560074279, 6.00126417111421e-05, 0.000215590579632216, -0.000175214077968917, -0.000168621952307728, 1736111.11101259, -0.000924687117766364, 0, 0, -0.0008519361881189, 0.000402193833739159, 0.000111368282499077, 9.22415631145251e-05, 0.000168002808464148, 0.000316915063507119, 0.000118132786338854, 0.00110492202494302, 0.000958356853683409, 0.00103570633091408, 0.00218761928126384, 0.00159965514941273, -0.000213829229186588, 0, 0, 0, 0, 0, 0, -6.8033192426585e-06, 0, 2.86746177894186e-07, 0, 0, 0, 5.10090260848253e-05, 0, 0.000119308941028307, -4.67856694211088e-06, 6.952964706646e-05, -99206.3491713048, 0, 0, 0, 0.000246333684078402, 0, 0.000465726647756905, 1.12111327482957e-05, -0.000152041753215684, 0.000227686323225723, 8.38318285446362e-05, -2083333.33318617, 0.000670977578636229, 0, 0, 0, 0.000611848478178993, 0, 0.000782461501817889, -0.000175279146595166, -0.000233027953701157, 3.78462011196156e-05, 0.000222748206681304, -9.02693840310936e-05, 0.00106930193597233, -0.00148432145993724, -0.000393429479960354, 0.00112001682925089, -0.000181747513732128, -0.000318642800393217, 4.73839832520109e-05, -0.00183758900143564, -0.00143090788425737, -0.00226429148770236, -0.00161055090875161, -0.00173735848445325, -0.000785106827965405, -0.0005401033891287, 0, 0, 0, 0, 0, 0, 2.71507394407354e-06, 0, 1.1099381518825e-07, 0, 0, 0, -3.62614214955334e-07, 0, -1.83994420751456e-05, 1.92060121389808e-06, -1.1794051541572e-05, -4.26898515386185e-06, 0, 0, 0, -5.58039611794965e-05, 0, -6.89299936696806e-05, -1.30643904781745e-05, 4.84318333012055e-05, -7.67968068142797e-05, -3.09157748041931e-06, 992063.492049302, -0.000147295023477258, 0, 0, 0, -0.000110020509580904, 0, -0.00024872184368355, -2.67749981954817e-05, 0.000108158886762933, 4.30938793346097e-06, 0.000107050559168038, -5.36687847692284e-05, -0.000412330177151602, 8.11615777846151e-05, -4.12127333374809e-05, -0.000345874527597331, 0, 0, -0.00026491573568805, 5.44230460081803e-05, 3.97783998662201e-05, 3.33833925119529e-05, 0.000119325113993312, 3.24000671494711e-05, -0.000460576681383664, 0.000345448536684002, 0.000254033415267293, 0.000638854049407188, 0.000288609167629291, -0.000404172904568538, -0.000503461563078032, 0.00105764001770203, 0.000288128248351798, 0.00124603970617508, 0.00138366560166331, 0.00258080771812921, 0.000972813477800882}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -6.73469605970999e-08, 0, 0, 0, 0, 0, 0, 0, 9.78602769923011e-07, 0, 1.0977690287173e-06, -55.5555554364793, 0, 0, 0, 0, 0, 0, -5.9671893018641e-06, 0, -8.01666789231138e-06, 0, 0, -2.06121934794993e-05, 833.333334818115, -5.04199464289763e-07, 1268.51851717041, 0, 0, 0, 0, 0, 0, 1.31789334013059e-05, 0, 3.07712861830709e-05, 0, 0, 0, 0.000124606251079039, 0, 0.000135652650191087, -2777.77778340683, 4.23557013764251e-05, -10416.66666723, 3.51132260505758e-05, -19027.777796813, 9.56288478149979e-05, 0, 0, 0, 0, 0, 0, -4.20119958779983e-05, 0, -3.14856359988791e-05, 0, 0, 0, -0.00028139014436963, 0, -0.000396254488904537, 6.28501253158085e-05, -7.91427170179905e-05, 39351.8518314138, 0, 0, -0.000550900513343906, -4.40750655911035e-05, 63425.9260070302, -0.000109286644563051, -0.000397184447731988, 156250.000043328, -0.000373739101906131, -0.000461763178033516, 0, 0, 0, 0, 0, 0, 2.58338360337015e-05, 0, 6.43711441967541e-05, 0, 0, 0, 0.000487427148848257, 0, 0.000912589112172179, -0.00014952044708322, 0.000258797983757621, -69444.444397576, 0, 0, 0, 0.0017049995027405, 0, 0.00159028057051043, -0.00013022310310893, -0.000238229428866344, 0.000368775784715423, 0.00108939462029564, -590277.777759384, 0.00123189070453021, 0.000297895783979366, -520833.333262274, 0.00129023844376021, 0.000520707666592507, 0.00201047187580559, 0.000410384372277098, 0, 0, 0, 0, 0, 0, -4.55476311532709e-05, 0, -1.34422613488616e-05, 0, 0, 0, -0.00032420288038978, 0, -0.000755013420689122, 0.000196074954023482, -0.000258704279548686, 46296.2963447384, 0, 0, 0, -0.00157842210815085, 0, -0.00220483190379677, 0.000259204230094506, 0.000630961334903809, -0.000762364100177587, -0.000648776658662642, 1041666.66686629, -0.00110218279315251, 0, 0, -0.00245263064959338, -0.0005455342235988, 0.00117385122966189, -0.000835322643594907, -0.00273142386995496, 1967592.59216815, -0.0044296775644186, -0.000215233264610752, -0.00231613833498426, -0.00435256880524786, -0.00337874386521444, -0.00261355465989621, 0.000606849735075747, 0, 0, 0, 0, 0, 0, 1.39569727555831e-05, 0, 3.3464688437949e-07, 0, 0, 0, 0.00018470511178741, 0, 0.00030203972643435, -0.000113371341007248, 0.000120446407233717, -0.000138746475929803, 0, 0, 0, 0.000869099607426844, 0, 0.00151626240835126, -0.00031138627811213, -0.000467059932534278, 0.000542642306927105, 0.000657967235576654, -694444.444857882, 0.000698172020675526, 0, 0, 0, 0.00193997728780783, 0, 0.00239115288293354, 2.8328603361858e-05, -0.00104756642985754, 0.000869126526245128, 0.000570386397019422, -3472222.22227931, 0.00274622537162373, 0.00135052342225583, 0.00203534811642273, 0.000870558739932157, 0.000642023882867551, -0.000271750188765742, 0.00348125959392617, 0.00133492752664068, 0.000967047537520224, 0.00407641151540372, 0.00220571177715754, 0.00446333352873123, 0.00383147394965156, -0.000973436950374015, 0, 0, 0, 0, 0, 0, -2.7530765387697e-06, 0, 3.22894524265644e-06, 0, 0, 0, -5.83680475705094e-05, 0, -7.43322371858821e-05, 2.49189994062701e-05, -3.36789613163032e-05, 4.51636048023556e-05, 0, 0, 0, -0.000218145608966607, 0, -0.000515814664315192, 9.85071709629237e-05, 0.000127408792455109, -0.000159622081190915, -0.000241225653263338, 0.000103339658733084, -0.000335475572868991, 0, 0, 0, -0.000689181986446182, 0, -0.000883212543018207, 0.000151117845238534, 0.000337474556278823, -0.000530379226586958, -0.000545852514343704, 2314814.81501154, -0.0012073514418498, 0.000405105523935208, -0.000606039157737754, -0.000552846662102488, 0, 0, -0.000981070968860612, -8.27623620994992e-05, 0.000536649793083383, -0.000295664952408294, -0.00109512093518723, 0.000227025913412447, -0.00128651892625181, 6.77173693843729e-05, 0.000823763493514348, -0.00162214257254971, -1.41436295331597e-06, -0.000454962211999947, -0.00109485077879584, -0.00148950160122875, -0.000218385441202138, -0.00261650269696927, -0.00054993793615067, -0.00360231959063996, -0.0020141047260207}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 8.64797875471213e-07, 0, 3.11666288856995e-08, -50.0000002838041, 0, 0, 0, 0, 0, 0, -7.23867642900266e-06, 0, 1.3339610555612e-06, 0, 0, -7.5448942924607e-06, 916.666670074117, -4.72188200880953e-06, 1041.6666699996, 0, 0, 0, 0, 0, 0, 4.11279167538839e-05, 0, -1.5142977734727e-05, 0, 0, 0, 5.19228601096304e-05, 0, 1.83748549579231e-05, -5000.00004093586, -2.04417557364195e-05, -7291.66667546698, -5.08734961216954e-05, -19097.2222458839, -5.63513601987115e-05, 0, 0, 0, 0, 0, 0, -7.29195114152782e-05, 0, 8.73367063443758e-06, 0, 0, 0, -0.000251490792710455, 0, -7.86260515635163e-05, 8333.33342619683, -1.34775272325769e-05, 20833.3334021592, 0, 0, -0.00030556398148569, 0.00022779305337542, 104166.666918457, 7.78431312056072e-05, 0.00020309141381688, 133680.555724365, -3.60827315062685e-05, 0.000543098519478223, 0, 0, 0, 0, 0, 0, 0.000147447719208373, 0, -6.48243670204522e-05, 0, 0, 0, 0.000527610965524906, 0, -0.00017603800268748, -0.00012966748817182, -6.8266948147541e-05, -20833.333562085, 0, 0, 0, 0.000209766939749868, 0, 0.000394368998201431, -0.00043957770810496, -173611.111924306, -0.000121571357714077, -0.000810980415833122, -381944.444896284, 0.000164234119484402, -0.000479998673937997, -729166.66780422, -0.000121688656983344, -0.00141530766418701, -0.00199552051223148, 0.000628116988535034, 0, 0, 0, 0, 0, 0, -7.62773318950453e-05, 0, 3.96688217729083e-05, 0, 0, 0, -0.000767547324295615, 0, -1.82482571166649e-05, 8.80319074621859e-05, 8.90576262468098e-06, 0.000264279220700616, 0, 0, 0, -0.00122583126577536, 0, -0.000700861880146189, 0.000655088602713162, 0.000855765884968377, 0.000400175121157725, -0.000130503585604484, 381944.444976325, -0.00128899509339463, 0, 0, -0.00145500344477635, 0.00135397125489812, 1215277.77837048, 0.000751724072103596, 0.00224375966905676, 2083333.33556198, 0.00171567331438155, 0.00270753703619128, 0.00212892402898216, 0.00229137523208343, 0.00543172432981121, 0.0033528878307053, -0.00296304407663713, 0, 0, 0, 0, 0, 0, 4.3737031646803e-05, 0, -2.85086975852786e-05, 0, 0, 0, 0.000404785945753068, 0, 0.000134008606976195, -3.58401779538673e-05, -1.90585902034131e-06, -0.000124833772544594, 0, 0, 0, 0.00102073739379296, 0, 0.000368537117412042, -0.000271759322198856, -0.000528358964417606, -0.000185401783279266, -0.000132448573159327, -0.000201057037897626, 0.00117836649214023, 0, 0, 0, 0.00120278219311113, 0, 0.000924647660080764, -0.000923635134910411, -0.000837499762873787, -0.000426834020337956, 0.000965337319632757, -2083333.33457537, 0.00105015665816409, -0.00521664787508862, -0.0016790441380104, 0.00266179668610626, -0.00116892356735166, -3472222.22396161, -0.00196622280438089, -0.00364642004327043, -0.0037027249710934, -0.00561075753680178, -0.00184607504578591, -0.00358563595511347, -0.000373720856001643, 0.00288812147776367, 0, 0, 0, 0, 0, 0, -8.71882026388624e-06, 0, 2.79750897581647e-06, 0, 0, 0, -6.55333808067311e-05, 0, -1.09463861175371e-05, 1.09843421871679e-05, -1.14157534198465e-05, 3.5229906863891e-05, 0, 0, 0, -0.000257893174141268, 0, 5.97468400952613e-05, 6.88797726145312e-05, 0.000133367787879958, 3.90252099018762e-05, 8.9098545291383e-05, 0.000167436981390441, -0.000135122893250502, 0, 0, 0, -0.000259103102999738, 0, -9.43728249771131e-05, 0.000134265090176483, 0.00041170134389095, 0.000447053586092855, 0.000305922832897434, 0.00026714773236182, -0.000230145798696615, 0.000797450259625888, 0.000506530746470371, -0.000801876092185931, 0, 0, -0.000238311996816259, 0.000304887135456102, 0.000218958441373418, 0.000207556880986182, 0.000559605656105269, 3472222.22270275, -0.000139549761912402, 0.00159522378360088, 0.000463355088464625, 0.0017355045541569, -0.000228818722272686, -0.00109103212346779, -0.0011362320163945, 0.00245399453809768, 7.68144632165862e-05, 0.00236172269223483, 0.000818584188163632, 0.0044042113011414, 0.000625377390065716}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -3.81610988486255e-08, 0, 0, 0, 0, 0, 0, 0, 5.34739543301906e-07, 0, 1.18834022948455e-07, -50.0000003440456, 0, 0, 0, 0, 0, 0, -6.27013479730773e-06, 0, 1.35111630541229e-06, 0, 0, -4.1237204100857e-06, 1041.66667434885, 9.74961775334137e-06, 916.666671215673, 0, 0, 0, 0, 0, 0, 1.97024439461655e-05, 0, -2.36098063397635e-06, 0, 0, 0, 5.28953472878742e-05, 0, 1.50165272146246e-05, -7291.66668133002, -3.17476512398726e-06, -5000.0000238002, -5.88869168536545e-05, -19097.2223071435, 1.87491225487995e-05, 0, 0, 0, 0, 0, 0, -7.74165841494796e-05, 0, 3.48115413961443e-05, 0, 0, 0, -0.000212133986300448, 0, 5.74956604774257e-05, 20833.3333809762, 2.20421406471912e-05, 8333.33336993133, 0, 0, 1.32320058452761e-06, 0.00022382140692574, 133680.555816895, 0.000368673037107108, -3.20725904951488e-05, 104166.666832362, 7.05718061926306e-05, 0.000289616033275247, 0, 0, 0, 0, 0, 0, 8.37597172961314e-05, 0, -2.54252090753356e-05, 0, 0, 0, 0.000389061178026684, 0, 8.25317836343198e-05, -20833.3334085463, 0.000117847749710929, 2.96485631484369e-05, 0, 0, 0, 0.000816819672202198, 0, 7.3822704672845e-05, -0.000443117872762604, -381944.444800932, -0.000863924593884263, 0.000577171384271479, -173611.111358028, -8.81242017503026e-05, -0.00105673496669942, -729166.667080815, 3.42616224290946e-05, -0.00178412476449325, -0.00133831043887974, -0.00218017431647249, 0, 0, 0, 0, 0, 0, -0.00013004645722809, 0, 3.18312748717658e-05, 0, 0, 0, -0.000304436001921436, 0, 0.000158381544171302, 6.91450584278287e-05, -0.000184403150606939, -2.29791437651801e-05, 0, 0, 0, -0.000419032700062069, 0, 0.000731141836788501, 0.000255039843527126, 381944.444837912, 0.000635375456075368, 0.000122653700651746, 0.000188063301579493, 0.000677458212439452, 0, 0, 0.000149582765377901, 0.000792053880476493, 2083333.3355429, 0.0016348199749082, -0.00236524035052362, 1215277.77759725, -0.00167525316752656, 0.0029001096233067, 0.00271349300509821, -0.000760109191039832, 0.00239883691619594, 0.00252761373225389, 0.00699756567271452, 0, 0, 0, 0, 0, 0, 5.59013450773216e-05, 0, -6.23054342520241e-06, 0, 0, 0, 0.000270246505359775, 0, -0.000236430444621822, -3.86427631764297e-05, 0.000223481131537098, -9.13751346045625e-06, 0, 0, 0, 0.000331583158632289, 0, -0.000429648432256418, -0.000233151310826565, -0.000231227026853753, -0.000351151029738288, 0.00038238310511104, -0.000292878619510984, -0.000656122877410898, 0, 0, 0, 0.000398870233239496, 0, -0.000194338563808709, -0.000408374543959887, -2083333.33486896, -0.00149969192907147, -0.00122790359336174, -0.000369892385483938, -0.000140509194013205, 0.000932672905162318, -0.00202107371479514, -0.00202445247397389, -0.000696796000614006, -3472222.22327623, 0.00290142039325955, -0.00555699476480323, -0.00332558758456697, -0.00267090095582325, -0.00328907263114142, -0.000665363857566017, -0.00168463668803063, -0.0122015431387222, 0, 0, 0, 0, 0, 0, -1.04993021706668e-05, 0, 5.85499788676002e-06, 0, 0, 0, -0.000103555861044101, 0, 3.92618910054684e-05, 5.49707776356933e-07, -4.26195168772653e-05, 8.59883530370939e-06, 0, 0, 0, -0.000195042898681137, 0, -5.0105485290797e-06, 3.44592431600002e-05, 3.58592853726455e-05, 7.03890396291887e-05, -0.000228441993915157, -6.20506367713907e-06, -9.91754935683094e-06, 0, 0, 0, -0.000320018465259627, 0, -0.000156995997169487, 0.000177030013247927, 0.000274984165273424, 9.6550191166989e-05, -0.000436129474250056, 0.000202546436339443, -0.000221497632482742, 0.00029187312962054, 0.000333262223581548, 0.000443536789899466, 0, 0, -0.000322107268148874, 0.000303966450314686, 3472222.22331569, 0.000620383817195728, -0.000655986449752555, 0.000582699112267808, -0.000407924168086754, 0.00137207106196989, 0.00251901832060502, 0.0014694557152145, 0.00225487661281395, 0.00162621970260095, 0.000282390365992382, 0.00217982208930759, 0.00277743872179292, -0.000795650749586372, 0.00556434405964648, 0.00143014310897598, 0.0023005045049396}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 3.4118357699185e-08, 0, 0, 0, 0, 0, 0, 0, -5.51601761213751e-07, 0, -3.54065227546969e-07, -55.5555559392582, 0, 0, 0, 0, 0, 0, 4.22884679533743e-06, 0, 8.15671978489256e-07, 0, 0, 8.81497472461458e-06, 1268.51852085604, 2.83585656151932e-06, 833.333335657434, 0, 0, 0, 0, 0, 0, -1.17225953805114e-05, 0, -3.68827943014042e-06, 0, 0, 0, -6.55073493687972e-05, 0, -5.65473315503202e-05, -10416.6666957474, -4.28699902139347e-05, -2777.77777902337, -4.61818283560037e-05, -19027.7777815246, -9.58524296521552e-05, 0, 0, 0, 0, 0, 0, 3.93863475870465e-05, 0, -1.67316470426302e-05, 0, 0, 0, 0.00019146501131072, 0, 0.000111046273748436, 39351.8519258463, 0.000112780069811718, -9.12447338472607e-06, 0, 0, 0.000174234793864934, 0.000196776986404758, 156250.000112376, 0.000183004863516668, 0.00037056201504716, 63425.9260163798, 0.000234449086520359, 0.000404692238848843, 0, 0, 0, 0, 0, 0, -3.50902006742058e-05, 0, 2.30948085136175e-05, 0, 0, 0, -0.000301997270978047, 0, -0.000259163479396147, -69444.4445441707, -0.000282184511630126, -5.31692927087496e-06, 0, 0, 0, -0.000936174365514899, 0, -0.000682370694297586, -0.000377176287060075, -590277.778216426, -0.000425863463209729, -0.001262077320825, -1.31585895482083e-06, -0.000521699992398758, -0.000511899725991592, -520833.333852808, -0.00124114909866833, -0.000829185967181768, -0.00121494437231795, -0.000150555319710164, 0, 0, 0, 0, 0, 0, 5.80922334620781e-05, 0, -3.5117715552672e-05, 0, 0, 0, 0.000191969991281673, 0, 9.39066457698509e-05, 46296.2963753139, 0.000280001318603534, -4.55108887357479e-05, 0, 0, 0, 0.000602439808411667, 0, 0.000172098798561764, 0.000533110442645522, 1041666.66729994, 0.000838681047156152, 0.000893636595350112, -8.88839216022371e-05, 0.000101818045350214, 0, 0, 0.000812224710007992, 0.00112896440599993, 1967592.59251006, 0.000983596922231562, 0.00383417098562964, 0.000977370845140454, 0.00273888227629784, 0.00103917677917986, 0.00204809483785942, 0.00188082595381392, 0.00402815160024489, 0.00235248442577505, -0.000358000013726382, 0, 0, 0, 0, 0, 0, -2.05066098632883e-05, 0, 1.63583640351813e-05, 0, 0, 0, -0.000161097316074479, 0, 7.61866274507677e-05, -2.35520927728566e-05, -0.000210018257421168, 3.98037898973259e-05, 0, 0, 0, -0.00037907799984331, 0, -0.0001625179132753, -0.000211459653416608, -694444.444792841, -0.000555446114254625, -0.000778038963996872, 0.000276300024532537, 4.32063072697302e-05, 0, 0, 0, -0.000955875156242547, 0, -0.000815791973744305, -0.00086043654190451, -3472222.22242228, -0.000735371639620596, -0.000965172706768334, -0.000130332039520884, -0.0015869321891186, -0.00273366664123846, -0.00180381882819136, 0.000253683680621, -0.00146296467588017, -0.000537129948407489, -0.00463024235303764, -0.00153739998187393, -0.00192929623191063, -0.00460668815961539, -0.00362347783253764, -0.00222624172793849, -0.00220406947396125, 1.8843561002719e-05, 0, 0, 0, 0, 0, 0, 3.69152984442084e-06, 0, -7.47114403721943e-06, 0, 0, 0, 6.16693995983723e-05, 0, -2.1864925513766e-05, 8.3522431030726e-06, 3.58910717052407e-05, -1.61885969265564e-05, 0, 0, 0, 0.000174126674715578, 0, 0.000141749101873061, 6.46027814409644e-05, 9.06936241558749e-05, 0.000134140216198092, 0.000263317177020267, -4.92059247935452e-05, 6.06872870441164e-05, 0, 0, 0, 0.000391363103619406, 0, 0.000488261926461207, 0.000167226727192321, 2314814.81507536, 0.000495804529690474, 0.000780369482126286, -9.48128652767843e-05, 0.000668383771202352, 0.000457322584218278, 0.000830686610016402, -6.1559791251703e-05, 0, 0, 0.000556279773315945, 0.000368361844991915, -0.000170281332999769, 0.000268007395012465, 0.00135173598580522, -4.10766570567266e-05, 0.00135520497950358, 0.00109154094675589, -0.000100547233910324, 0.00162933417042617, 0.000729757916785709, -0.000125574237303113, 0.00121615924478352, 0.00156906754527356, -0.000121379246768753, 0.00286927212907179, -0.000369397947204423, 0.00330778868740823, 0.0015077955925614}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.38496769408676e-08, 0, -1.14671430311996e-07, -71.4285715243811, 0, 0, 0, 0, 0, 0, 5.76957101728699e-07, 0, 1.57594080757552e-06, 0, 0, 9.17221076973035e-07, 1750.00000180301, 1.69318534533261e-06, 714.285715980693, 0, 0, 0, 0, 0, 0, -4.20395113828872e-06, 0, -5.20461185647162e-06, 0, 0, 0, 2.74066532801117e-06, 0, -6.87059659546412e-06, -16111.1111153682, 3.18145488274154e-06, -1.12169577360731e-05, -3.79952394368559e-06, -17500.0000225026, 3.86917868143828e-06, 0, 0, 0, 0, 0, 0, 2.52705009700369e-06, 0, 1.98499346019096e-05, 0, 0, 0, 1.42995297518754e-05, 0, 4.64007493058461e-05, 72916.666684955, -3.55398726727845e-06, 4.83693671239467e-05, 0, 0, 7.27248656779621e-05, 4.79482956910534e-06, 161111.111195099, -9.92379415811658e-06, -1.95981059868564e-05, 1.63023052581383e-05, 7.91686316201576e-05, -1.11564969325298e-05, 0, 0, 0, 0, 0, 0, -1.38597203177486e-05, 0, -9.24894451234631e-06, 0, 0, 0, -3.35305952988267e-06, 0, -8.58167628703542e-05, -173611.111155073, 6.34209267029138e-05, -7.85553935721349e-05, 0, 0, 0, 0.000172439607101307, 0, -7.21586884934735e-05, -7.00012972208643e-05, -729166.666760215, 6.15017268320605e-05, 0.000268038619589104, -0.000138555481675345, -0.000178231957484549, -3.84811294072033e-05, -7.16067518188556e-05, 1.73306997330814e-05, -0.000248067956591185, 2.84742298801524e-05, -0.000497822788523209, 0, 0, 0, 0, 0, 0, 2.33889024230298e-06, 0, 1.57544880258135e-05, 0, 0, 0, 7.5101507905966e-05, 0, 0.000105552690277605, 208333.333388517, -5.01461067705677e-05, 6.81254372321285e-05, 0, 0, 0, 9.99127624675917e-05, 0, 0.000444921976180309, 5.44791510019859e-05, 1736111.11118495, -0.000260817135065573, 1.96806008622667e-05, 0.000190320347715894, 0.000609560776624953, 0, 0, 0.000283401698718215, 4.25626692819591e-07, 0.0007080940256382, -6.13582270229255e-05, -0.00101264869372148, 8.58667787647316e-06, -0.000281745459327568, 0.000780565704879772, -0.000166304991807943, 0.000170960910927951, -0.000586616232434305, 0.000107415404718312, 0.00126945199166176, 0, 0, 0, 0, 0, 0, -5.34314904550252e-06, 0, -3.86790171854108e-06, 0, 0, 0, -3.20971297800075e-05, 0, -9.80167943531078e-05, -99206.3492417833, 4.53447782828908e-05, -4.18454634811254e-05, 0, 0, 0, -8.87062406850754e-05, 0, -0.000323714745710816, -9.87282550491668e-05, -2083333.33343027, 0.0001836997604319, -1.38962824871921e-05, -0.000218525073143425, -0.000451729859761441, 0, 0, 0, -8.77923611592956e-05, 0, -0.000155123723629399, 2.87851137098284e-05, -0.000415163040022408, 0.000202391354594196, 0.000115980212006046, -0.000154103957890126, -0.000227883163909896, 0.00025617352725516, 0.000332955481449935, -0.000911337294795295, 0.000230497094841165, -0.000262797436055017, 0.00135127228749037, -0.000994394720413739, -0.000579745405231897, 0.000613410871893595, 0.000927673260990118, -0.000176836861424277, -0.000144723164246973, -0.0015810934957265, 0, 0, 0, 0, 0, 0, 1.30562377848978e-06, 0, 1.99583141119406e-06, 0, 0, 0, -4.60040509812687e-06, 0, 2.16562264156737e-05, 5.9605482460022e-06, -8.01172988125485e-06, 6.97924849820418e-06, 0, 0, 0, -2.44801850158145e-05, 0, 1.52271033816224e-05, 1.30563151671225e-05, 992063.492091583, -4.05168505522462e-05, -8.26312990919444e-07, 2.63148666928071e-05, 0.000109943300808714, 0, 0, 0, -4.54256056078945e-05, 0, -3.99283262334886e-05, 5.23944678806068e-05, 6.54818160481178e-05, -0.000202753061046153, -0.000164182944067338, 8.19155204811257e-05, 7.56525277806086e-05, 7.92683026840947e-05, -0.000182709139266889, 0.000369169207539098, 0, 0, -2.96429937352784e-06, -2.38809370237784e-05, 0.0003077071807293, -4.42193938410422e-05, -0.000307997061494672, 0.000153307736095503, -0.000252223976856364, 0.000226238991945905, 0.00068733812543931, -0.000273732738066266, -2.74824251533388e-05, 0.000509285275399797, -9.63882220272497e-05, 8.55298928668167e-06, 0.000579901593209365, -0.000885960451463232, 0.00130567862734802, -0.000671299844835209, -2.43277306418749e-05}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.90999503050494e-07, 0, -1.20871039532108e-07, -41.6666665130467, 0, 0, 0, 0, 0, 0, 1.0905788726236e-06, 0, 2.11447431513856e-07, 0, 0, 2.74937890111439e-06, 951.388889348244, 416.666664957994, 416.666664287257, 0, 0, 0, 0, 0, 0, -3.86607656828233e-06, 0, 6.93966409743317e-07, 0, 0, 0, -2.30268853417002e-05, 0, -7.95375407971254e-06, -7812.50000283978, 9.64075885595263e-06, 9.86869693405367e-06, -9513.88888454497, -9513.88889484452, -4166.66660056701, 0, 0, 0, 0, 0, 0, 1.84209189509755e-05, 0, -4.87919157223354e-06, 0, 0, 0, -1.94296777405181e-05, 0, 5.10152378918857e-05, 29513.888895651, -1.51923799997001e-05, -6.32895841699527e-05, 0, 0, 5.03266168889565e-05, 78125.0000972452, 78125.0000198898, -2.93070533359949e-05, -0.000289848209049367, 0.000113065911166506, -0.000400842699910162, 95138.8889253757, 0, 0, 0, 0, 0, 0, -2.58430389301902e-05, 0, -3.82811079573625e-05, 0, 0, 0, -0.000110582950551256, 0, 9.75005397648897e-05, -52083.3333162643, -0.000116061566062142, 9.86549871551398e-05, 0, 0, 0, -0.000474599567881357, 0, -0.00022227676658823, -295138.889065766, -295138.889135428, 0.000173499854715044, 0.000418215945516347, -5.91273699857885e-05, 0.000702567557167712, 4.02156199800024e-05, -0.000302602937212883, 0.00154620116011307, -781250.000563456, 0.000348700867400173, 0.000387864098661183, 0, 0, 0, 0, 0, 0, 2.21911403742482e-05, 0, 1.03721946989516e-05, 0, 0, 0, 7.32380807984104e-05, 0, 3.1092287212892e-05, 34722.2221574717, 0.000153448190411771, 2.45524753443905e-05, 0, 0, 0, 0.000200799611554857, 0, 0.000136494855340771, 520833.333543828, 520833.333626667, -0.000155293916425692, -0.000379857570449844, 9.67693293651034e-05, -0.00112996230331112, 0, 0, 0.000160564458589284, -5.45939998481686e-05, 0.000281774197440513, -0.000791269474177761, -0.00213823859411329, 0.000278431977739546, -0.00255472503302449, 2951388.88962931, -0.00102190372603549, -0.000852703239316482, 0.00102274159818791, 0.00142997290384174, -0.00218337115846768, 0, 0, 0, 0, 0, 0, -7.14226502711079e-06, 0, -1.11007279235222e-05, 0, 0, 0, -1.8533594399478e-05, 0, -3.62644032918083e-06, 4.8085331766259e-05, -0.000105166095687719, -1.38255732643837e-05, 0, 0, 0, -4.53515367440687e-05, 0, 0.000121718506291298, -347222.222268749, -347222.2222357, 0.000106611696228251, 0.000190323658580397, -4.71293193378497e-05, 0.000491220287192206, 0, 0, 0, -7.63694765374356e-05, 0, -0.000422189888149308, -0.00028603936834462, -0.000592109190085039, 0.000518851808766042, 0.000911704595189399, -0.00038994437186365, 0.00212244412966289, -5208333.33436069, 0.000633487109175785, 0.000363466299264013, 0.000582470903178137, -0.000651901959922232, 0.00255568413276921, -0.00128250358261121, -0.00206171294324802, 0.00100090394937976, 0.00169765996547254, -0.00111335679163204, 0.00259366765101449, -0.00112783902575894, 0, 0, 0, 0, 0, 0, 1.62378479252773e-06, 0, 3.59098268285506e-06, 0, 0, 0, 1.98767763415293e-05, 0, -7.30629907454904e-06, -1.29548467627431e-05, 4.30115727180435e-05, 2.02568232666663e-05, 0, 0, 0, 8.27243912658097e-05, 0, 5.59943699548784e-05, 1.77533651826276e-05, -4.00480119923794e-05, -2.63960206359103e-05, 6.51220965931462e-06, 8.25587193080194e-05, -0.000168151167451542, 0, 0, 0, 0.000129088097463728, 0, 0.00016539552044264, 2.22175738131765e-05, 0.000118862994962468, -0.000202971772418436, -0.000354932702414085, 0.000141075556287511, -0.000722547672678927, 3472222.22251548, -0.000218233666947377, -0.000214422821598165, 0, 0, 3.56523774520822e-05, -7.08691642145154e-05, 0.000269986117871367, -0.000303740237652937, -0.00117914441456979, 0.000298346722246694, -0.00119042728292484, 0.000467065767440151, 0.000750463766698446, -0.000656470523921203, -0.00111112992500505, 0.000197718720889056, -0.000804483993948411, 0.00032263961347283, 0.00121251471735906, -0.00162184115878739, 0.000791795909197865, -0.000186242669635814, 0.000600997896191698}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.82848992885146e-08, 0, 2.6922673478417e-07, -41.6666665571738, 0, 0, 0, 0, 0, 0, 3.55909256837379e-06, 0, -3.49701555897744e-06, 0, 0, -1.68428257819481e-06, 416.666665043276, 951.38889116245, 416.666661393912, 0, 0, 0, 0, 0, 0, -1.08464427090182e-05, 0, 2.05239032735782e-05, 0, 0, 0, 5.42675098794204e-07, 0, 4.84263187510717e-05, 4.63180742310462e-06, -7812.50004466361, 2.47285497240171e-05, -9513.88891011412, -4166.66663522062, -9513.88884295282, 0, 0, 0, 0, 0, 0, 2.71451582549331e-05, 0, -4.15433269471946e-05, 0, 0, 0, 8.33033758218045e-05, 0, -0.000222901527884259, -3.73283482501615e-05, 29513.8890688989, -0.000130146108094827, 0, 0, -6.63893135162823e-05, 6.72631797148969e-05, -7.91165333260687e-05, 78125.0002005401, 78125.0001358615, -0.000193647189783216, -0.000231358990141869, 95138.8887278389, 0, 0, 0, 0, 0, 0, -3.78051204109628e-05, 0, 5.44059932129331e-05, 0, 0, 0, -0.000255652184581629, 0, 0.000446850466594132, 4.28451552289133e-05, -52083.333646522, 0.000256049652014607, 0, 0, 0, -5.81975178740083e-05, 0, 0.000571733034212932, -3.47082478231314e-05, 0.00019869464113581, -295138.889436528, -295138.889103724, 0.000275950590013102, 0.00114384869758988, -0.000480476474658704, 0.000552837328425279, 0.00027731050214093, 0.000485447848830198, -781249.999636175, 0.00188468054815803, 0, 0, 0, 0, 0, 0, 3.10451710740804e-05, 0, -5.91608344344454e-05, 0, 0, 0, 0.000234098312917426, 0, -0.000513965849087201, -7.15478199256115e-05, 34722.2226317157, -0.000230076996490171, 0, 0, 0, 0.0002799371728705, 0, -0.000737758381986017, -2.55375968687057e-05, -0.000360245934075169, 520833.33449143, 520833.334096233, -0.000453216191357611, -0.000983625751161937, 0, 0, -2.89497581281806e-05, 0.000705152812145543, -0.000111546114223017, 0.00148224006760903, 1.407747725535e-05, -0.000336676002286177, -0.000999579102861035, -0.00101168155000287, 2951388.88983479, -0.00308748550932694, 0.000669591674542432, -0.00288676733001, -0.00125010698897526, 0, 0, 0, 0, 0, 0, -1.61332038334415e-05, 0, 1.57386719879163e-05, 0, 0, 0, -0.000127039002142453, 0, 0.000353338894241262, 5.57943230383048e-05, -0.000301733478483881, 0.000191071207803917, 0, 0, 0, -0.000257169302028839, 0, 0.000454914913585417, 4.27665876858487e-05, 0.000307070204794045, -347222.22324566, -347222.222921682, 0.00040175334859504, 0.0013566916743999, 0, 0, 0, -0.000283081338090454, 0, 0.000372022179009688, -0.000307908347316609, 0.000167212558117134, -0.0014524482176371, -0.000216286949011129, 0.000575191235604178, 0.0018015078835953, 0.000822972112434503, -5208333.33471351, 0.00216656451426571, -0.00111539912759085, 0.000193621118501318, 1.6226101744557e-06, 0.000577690871318588, 0.00133418309713033, -0.00101099014236015, 0.00203250956787662, 0.00330564269990233, 0.00246030341990597, 0.00394763023580794, 0, 0, 0, 0, 0, 0, -1.70608615426961e-06, 0, -3.9765485549381e-06, 0, 0, 0, 2.19626626120567e-05, 0, -7.0744618257341e-05, -1.49315211625251e-05, 5.73454550318331e-05, -4.07672300633865e-05, 0, 0, 0, 0.000107243796682625, 0, -0.000270528069296022, -2.18286840312371e-05, -7.13932139516619e-05, 0.000330305155522399, 0.000184457745537447, -5.6621139754382e-05, -0.000317366599144743, 0, 0, 0, 3.99072679886769e-05, 0, -0.000271393568441857, 7.5342136530186e-05, -0.000285266438727631, 0.000621026366806783, 0.000467594012949894, -0.000103887424894655, -0.000835583161006456, -0.000247213671816306, 3472222.22268626, -0.000852093275898011, 0, 0, 4.39662401205988e-05, 0.000365389763304353, -0.000202676553977537, 0.000746033527805243, 0.000198847096808104, -0.000204433042396419, -0.000115963669454779, -8.92623796448341e-05, -0.000529177904663722, 0.000667639353039385, -4.60969725715108e-05, -0.00127150440387404, -0.00127267613105565, 0.000504608511769833, -0.000988786511649722, -0.00151195368917466, -0.0011245718593594, -0.000712198909819894, -0.00178655459050614}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.83219039763883e-07, 0, 2.25664831422698e-07, -41.6666660052428, 0, 0, 0, 0, 0, 0, -7.02026979022081e-07, 0, -1.0553534086196e-06, 0, 0, -4.93883584800987e-06, 416.666658065385, 416.666656417296, 951.388881842282, 0, 0, 0, 0, 0, 0, -6.0979478531959e-06, 0, -1.19223331942186e-06, 0, 0, 0, 5.98994303793506e-05, 0, 2.71119229960992e-05, 3.58751525278714e-05, 6.53857575187318e-05, -7812.49997427936, -4166.66656778657, -9513.8888368543, -9513.88877941815, 0, 0, 0, 0, 0, 0, 3.10982325720205e-05, 0, 2.46842995188431e-05, 0, 0, 0, -0.000128672951334844, 0, -6.21813815425416e-05, -9.80635913208113e-05, -0.000162008157354054, 29513.888857676, 0, 0, -0.000306842795954902, -0.00034625421713462, -0.000270003152704609, -0.000437537178450821, -0.000633523271639394, 78124.9998596009, 78124.9996305081, 95138.8883143154, 0, 0, 0, 0, 0, 0, -8.3894347357039e-05, 0, -2.409698804148e-05, 0, 0, 0, 0.000286159478124373, 0, -4.03600734576779e-05, 0.000152001522706753, 0.000263339083447656, -52083.3333501281, 0, 0, 0, 0.000936266737635521, 0, 0.00034338194694392, 0.000728020599196191, 0.000585042385270525, 0.00127547016132022, 0.0016683021081143, -295138.888625457, -295138.888204261, 0.00141560083449933, 0.000646647915485748, 0.00215528695272044, 0.00163755180852425, 0.00281158013698196, -781249.998828813, 0, 0, 0, 0, 0, 0, 0.000109565608067354, 0, 3.55059862253779e-05, 0, 0, 0, -0.000179784708315592, 0, 0.000155892330095286, -0.000135432098704439, -0.000180794142496786, 34722.2222377409, 0, 0, 0, -0.00134955746964759, 0, -0.000357467908087415, -0.000768186176026343, -0.000748545351375117, -0.00140219129450026, -0.00183709977373802, 520833.333080836, 520833.333037862, 0, 0, -0.00144542384874883, -0.00201837559088734, -0.00125486536771204, -0.00250120512891854, -0.00389586885794549, -0.000588659427260828, -0.00385812284991783, -0.00293166514094692, -0.00464052636660614, 2951388.88694102, -0.00461744406134123, -0.00247514534568424, -0.00439266611056306, 0, 0, 0, 0, 0, 0, -7.22767878846919e-05, 0, -2.03696334211817e-05, 0, 0, 0, 2.58113633492434e-05, 0, -0.0001413303774207, 6.84063687199188e-05, -9.26775756617188e-06, -6.25990038541285e-05, 0, 0, 0, 0.000833177366364615, 0, -0.000122395778263558, 0.000507282536889691, 0.000470298246370432, 0.000849903380022857, 0.000929490084873517, -347222.222151214, -347222.222045389, 0, 0, 0, 0.00143158602316668, 0, 0.000763317305943204, 0.00141660115582375, 0.00123922432846197, 0.00237791464791045, 0.0041338516116455, 0.000690329928164072, 0.00211698444041416, 0.00235192924275512, 0.00365110644617675, -5208333.33322539, 0.00212185053661079, 0.00119185019856769, 0.00486375912039886, 0.00500374554660192, 0.00276482308040563, 0.00623263549308365, 0.00660491595276356, 0.00209048426792868, 0.00396499706252058, 0.00374974942665735, 0, 0, 0, 0, 0, 0, 1.59477316094202e-05, 0, 2.72895686153301e-06, 0, 0, 0, 1.21000455125265e-05, 0, 3.77961680762456e-05, -1.38229411283473e-05, 8.10183947327959e-06, 1.71022885516731e-05, 0, 0, 0, -0.000188726268738498, 0, 4.69001531111952e-05, -0.000118313331109014, -0.000124850786372917, -0.000223399077563114, -0.000193120308916279, -7.73831808730223e-06, 8.32921753818198e-05, 0, 0, 0, -0.000545683879308108, 0, -0.000112404282102, -0.000367070090726845, -0.000399190513850016, -0.000757780071573952, -0.00126440597857946, -0.00013693410368429, -0.000420542024516074, -0.000535112554694652, -0.00108029040773513, 3472222.2222221, 0, 0, -0.000423414363743392, -0.000770236269290683, -0.000624273848569062, -0.00108574676475968, -0.00239742818475722, -0.000474194004477617, -0.00158902051559827, -0.00132462005650343, -0.00145112962263907, -0.00282857988756424, -0.00292144434359498, 0.000190076013854437, -0.00102291209482244, -0.00298680494464644, -0.00108367664030102, -0.00330563517860192, -0.00339849401898266, -0.00417459451341053, -0.00155218895178741}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.38563993574865e-07, 0, 2.63108562504628e-07, -33.3333336292098, 0, 0, 0, 0, 0, 0, 1.8048824700953e-06, 0, -3.20163928713809e-06, 0, 0, 5.52113313705671e-06, 694.444448111037, 500.000002362486, 333.333331780262, 0, 0, 0, 0, 0, 0, -1.13442778580297e-05, 0, 1.67264661181424e-05, 0, 0, 0, -2.47381544739595e-05, 0, -5.15731220894455e-05, -4861.11113548233, -1666.66665828442, 4.70156368007964e-05, -10416.6667051154, -6944.44450095945, -5000.00006793489, 0, 0, 0, 0, 0, 0, -6.83363532903993e-06, 0, -5.14951413419183e-05, 0, 0, 0, 0.000179581977368115, 0, 7.8552639313693e-05, 13888.8889463953, -2.54501071739624e-05, -0.000181233086383151, 0, 0, 0.000170874256343448, 72916.6668074964, 48611.1114635049, 34722.2223390554, 16666.6667339008, 5.77939372911601e-05, 0.000129030121463265, 104166.667044257, 0, 0, 0, 0, 0, 0, -7.89947497489164e-06, 0, 0.000211747869939965, 0, 0, 0, 2.39958317990761e-05, 0, -0.000215203257867026, -13888.8889327936, 0.000236723794633624, 0.000432040842974318, 0, 0, 0, 0.000233025808775927, 0, -0.000729729076240339, -208333.333801507, -138888.889282114, -0.000108320852807164, 1.54191683649688e-05, 0.000144057755032637, 0.00069653803272434, -243055.556141536, -0.000869152087739541, -0.000759542109482981, -729166.668753186, -347222.222850294, -0.00202785034527993, 0, 0, 0, 0, 0, 0, 1.9482854323496e-06, 0, -0.000156212069182417, 0, 0, 0, -1.43305315034581e-05, 0, -8.91106070590906e-05, 4.48039541350305e-05, -0.000235351042141359, -0.000810330696557657, 0, 0, 0, 4.37920328720208e-06, 0, 0.000494644323263898, 208333.333770287, 138888.889292792, -8.1092392571902e-05, -0.000210866223562996, -0.000521498022637665, -0.00124543138687334, 0, 0, 0.00121953843399809, 694444.445958987, 0.00172684037785602, 0.00120525139685974, 0.000700712973671023, 0.00139743601747119, -0.00124693155548043, 2083333.33748633, 0.00139145044327162, 9.69228627090815e-05, 2430555.55991806, 0.00508495217109847, 0.00361871088644694, 0, 0, 0, 0, 0, 0, -6.82443541040325e-06, 0, 9.98388099772645e-05, 0, 0, 0, -6.01545416597146e-05, 0, 0.000131490958107146, 1.4602189101438e-05, 8.76118767884983e-05, 0.000428182120498835, 0, 0, 0, -6.74408342898773e-05, 0, -0.000710449452345968, -0.000225735659137969, -0.000183114075616776, 0.000122777477607599, 8.84807334404161e-05, 0.000475788718231187, 0.00130310553186376, 0, 0, 0, -0.000966890316475842, 0, -0.000881192378188036, -694444.445301751, -0.00043495943924344, -0.000100135614879649, 0.00145018906537813, -0.000104259521995872, 0.000356635021259798, -2083333.33547359, 1.06716028969815e-05, 0.000771776500132057, -0.00208350811500656, -0.00122975499889195, -0.00042811990621315, -6944444.45036703, -0.00625459212063163, -0.00614125972354911, -0.00275869095499907, -0.00314307978999311, -0.00118314545833145, -0.0109740584125699, 0, 0, 0, 0, 0, 0, 2.62088538625122e-06, 0, -2.55458636298389e-05, 0, 0, 0, -1.11176424634254e-05, 0, -5.92842590766802e-05, -6.63731967870708e-06, -4.487433409593e-05, -0.00013731711928269, 0, 0, 0, -8.2376973406173e-05, 0, 5.58986828760295e-05, 8.78722271469453e-06, 3.31969785095404e-05, -7.38070175728446e-05, -2.59479746126915e-05, -0.000340567255138636, -0.000440310844634012, 0, 0, 0, -5.23538047789265e-05, 0, 0.000497865417263434, 0.000344683705301476, 0.000234134061966998, -6.50654119337188e-05, -0.000232156742304804, -0.000125421108111747, -0.000220362983401387, 0.000714612289063066, 0.000343249845981492, -0.000459999694817752, 0, 0, 0.000602318595415825, 0.000559768314012782, 0.000296105872281804, 0.000213496137345318, -6.90192698931494e-05, 0.000170711341502054, 0.000149249214804838, 6944444.44769484, 0.00263619108555408, 0.00185084662485976, 0.00100925645818542, 0.00139056098800537, -1.87868304837773e-06, 0.00251993640702588, 0.00244744694480042, 0.000733632475864503, 0.00515985111013555, 0.00401817711290561, 0.00343140766761404}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -4.72390758309202e-08, 0, 0, 0, 0, 0, 0, 0, 7.92510692681857e-07, 0, 9.56911925236467e-07, -31.2499994331466, 0, 0, 0, 0, 0, 0, -3.26646153071452e-06, 0, -8.9238230003851e-06, 0, 0, -1.70920760180954e-05, 572.916670347864, 572.916660051653, 312.499995225106, 0, 0, 0, 0, 0, 0, 2.79871687110831e-05, 0, 4.58699383762314e-05, 0, 0, 0, 8.57544374063418e-05, 0, 0.000128669201351271, -3125.00002361574, -3124.99997375034, 2.07239788358031e-05, -10503.472250385, -5729.16666816794, -5729.16654545312, 0, 0, 0, 0, 0, 0, -2.90931008518952e-05, 0, -0.000122641325717483, 0, 0, 0, -0.000246581448938427, 0, -0.000327303982618397, 5208.33340059871, 5208.33317595415, -0.000196417311345847, 0, 0, -0.00024433953410601, 57291.6669465962, 31250.0000633338, 57291.6666432036, 31249.9999195842, 3.89108922502538e-05, -0.000583268006055219, 105034.722455812, 0, 0, 0, 0, 0, 0, 7.48787040214866e-05, 0, 0.000101107139281993, 0, 0, 0, 0.000166328248721245, 0, 0.000781723387817979, -0.000160720984721276, 0.000371371158774814, 0.000499030036077375, 0, 0, 0, 8.3094433096471e-05, 0, 0.00122878704148339, -95486.1116194077, -52083.333845555, -95486.1105969635, -52083.3329471562, -8.69703431538503e-05, 0.00149140063723633, -312500.000299415, 3.68487298878801e-05, 0.00210430152135905, -572916.66762851, -572916.667818251, 0.0017525340641877, 0, 0, 0, 0, 0, 0, -3.79583112430296e-05, 0, -0.000163223884724214, 0, 0, 0, -0.000123124278294595, 0, -0.00069022876237944, 0.000125617443277095, -0.000513361560841744, -0.000288450519172767, 0, 0, 0, -0.000432967305920311, 0, -0.00106884173776506, 0.000741771785913116, 0.000455486428066535, -0.00107976234066568, -0.000206207393811234, -1.16021951961986e-05, -0.00162363112488844, 0, 0, -0.00170828924551013, 520833.333542678, 0.000190482286599111, 520833.332577194, -0.00223882042079468, -4.28699558436206e-05, -0.00237281092234147, 954861.112317942, 954861.112792717, -0.00103922666226152, 3125000.00332366, -2.47276158314854e-05, 0.000165356702822473, 0, 0, 0, 0, 0, 0, 2.36223777596138e-05, 0, 6.79272172229855e-05, 0, 0, 0, 0.000156771666554396, 0, 0.000464317831312681, -8.70553839792365e-05, 0.000328491588650253, 0.000348368243471619, 0, 0, 0, 0.000339163238219135, 0, 0.00110220190964056, -0.000470869696819213, -0.000257773868924546, 0.00107125101793532, 0.000258826219253649, 0.000249984882884494, 0.00159686157757353, 0, 0, 0, 0.0012787951311837, 0, 0.00143701607491449, -0.000753563456669061, -0.000836784632233995, 0.000951673273696134, 0.0001170584691724, -0.000619073714306342, 0.00473510606191911, -0.00237991028003825, -0.00180139202742724, 0.00134765301118152, -868055.554441938, -0.000289403637647982, 0.00415874273610594, -5208333.33585903, -0.00190595829161456, -5208333.33442623, 0.00163482630169348, -0.00156348743681044, 0.00168726529318513, 0.00416167454679379, 0, 0, 0, 0, 0, 0, -9.90098984428441e-06, 0, -1.43530533089136e-05, 0, 0, 0, -8.05239133306357e-06, 0, -9.89255171891476e-05, 2.02078110666328e-05, -0.000102353342983334, -8.32057351428139e-05, 0, 0, 0, 5.68709009846995e-05, 0, -0.000317033983244281, 0.000146674967347516, 6.92564896143769e-05, -0.000317357390930493, -0.000146622172337989, 9.84270999089407e-05, -0.000427526072465877, 0, 0, 0, -4.01962038984065e-05, 0, -0.000637169482855879, 0.000201224833081314, -8.00437296660409e-05, -0.000461600248016033, 0.000225519461325495, 0.000266469073153808, -0.00165408343186152, 0.000686489049278873, -5.25682992171627e-05, -0.000899027771994081, 0, 0, -0.000592590143235739, -2.47213392653891e-05, 6.94780439121154e-05, -0.000190509247082976, -0.000966379089551937, 0.000341584046442912, -0.0024544039765463, 8.61733102156597e-05, 0.000444231217288442, 0.000248735994298471, 0.000140207373511823, -0.000664442213153893, -0.00172082074674903, 8680555.55714855, 0.000733534240472892, -0.00213983886743167, 0.00148306064871139, 0.000487725565491037, -0.00022477535186425}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.39338001551575e-07, 0, 4.572276311248e-08, -33.3333337222176, 0, 0, 0, 0, 0, 0, -1.61811782857746e-07, 0, 5.65325230266627e-07, 0, 0, 3.20496214032315e-06, 500.000000616526, 694.444451223691, 333.333328771609, 0, 0, 0, 0, 0, 0, -2.37721022868821e-05, 0, -9.60227780255813e-06, 0, 0, 0, -1.96473905366678e-05, 0, -1.79035359502128e-05, -1666.66666935835, -4861.11115736237, 6.5325945615425e-05, -10416.6667008888, -4999.99997186652, -6944.44443007966, 0, 0, 0, 0, 0, 0, 4.5475640629129e-05, 0, 2.58655004493612e-05, 0, 0, 0, 2.89174021673653e-05, 0, -4.54467325254588e-05, 1.45449599742068e-05, 13888.8891453543, -8.609247819171e-05, 0, 0, -0.000256149569385079, 34722.2222992957, 16666.666513337, 72916.6670885307, 48611.1108081141, -0.000317649053897599, -0.00051344436708599, 104166.666328582, 0, 0, 0, 0, 0, 0, -0.000102658386748146, 0, -3.99072393826972e-06, 0, 0, 0, 2.38396491642149e-05, 0, -6.40355672426968e-05, 3.11745878701831e-05, -13888.8895785997, -1.7242504455136e-06, 0, 0, 0, 0.000631173228239145, 0, 0.000233300730129814, -0.000210087385608097, 0.000430731202728471, -208333.334969337, -138888.889175011, 0.000938554575569102, 0.001234631639254, -243055.556646704, 0.00066425239855114, 0.00142627018538936, -347222.221675234, -729166.666410624, 0.000997870413657323, 0, 0, 0, 0, 0, 0, 6.88497901490427e-05, 0, 5.20266019881066e-05, 0, 0, 0, 2.93555217030332e-05, 0, 5.32544472208101e-05, -2.08899104940493e-05, 0.000783313966538241, -0.000119706507094701, 0, 0, 0, -0.000308303724624839, 0, -0.000889522693794384, 0.00015501031898475, -0.000298140664296114, 208333.335883291, 138888.889632507, -0.00100514487606991, -0.00274867764541907, 0, 0, -0.0010502781278149, 0.00160520154001611, -0.0016688852046395, 694444.447753651, -0.00172637503458935, -0.00201739317740142, -0.00529394064978698, -5.80542406729207e-05, 2083333.33384343, -0.00520936337825888, 2430555.55470613, -0.00305852316178689, -0.0063111569444262, 0, 0, 0, 0, 0, 0, -3.7748325600276e-05, 0, -1.14422048910745e-05, 0, 0, 0, -8.62702071072738e-05, 0, -0.000173623502177631, 2.33815887636209e-05, -0.000383095113604121, -0.000131067814295637, 0, 0, 0, -1.09436720511993e-05, 0, 0.000535249530386364, -6.4868533524215e-05, 5.88676637839623e-05, -0.00191855846099561, -0.000862473062452157, 0.000231276506799455, 0.00112142402036436, 0, 0, 0, 0.000773968091989786, 0, 0.00109936482776297, -0.000597546204613417, 0.00149916562270638, -694444.448051647, 0.000655654488209668, 0.001811566153148, 0.00334919428311063, 0.000714321603619466, -2083333.33503458, 0.00513532326506905, -0.00324835960232533, 0.00222784391964782, 0.00419884208887494, -0.000884489687704131, 0.00457013119799985, -6944444.44707305, 0.00135095394530847, 0.0080478698693915, 0.00903660456005929, -0.000322340471817567, 0, 0, 0, 0, 0, 0, 1.47159556626603e-05, 0, 3.31117214129994e-06, 0, 0, 0, 5.39916921773238e-06, 0, 3.76218889568581e-05, -1.03970096424988e-05, 0.000128700287762451, 4.77562265539972e-05, 0, 0, 0, -8.96840194763142e-05, 0, -1.97757236676409e-05, 1.10369484462314e-05, -2.32309810484239e-05, 0.00049375927531155, 0.000361904446271479, -0.000119066275802385, -0.000274498335495028, 0, 0, 0, -0.00026369221807724, 0, -0.000396620656251521, 9.37286301495883e-05, -6.65292266539367e-05, 0.00129172274859463, -0.000450760097414992, -0.000529842646294585, -0.00120301337713133, -0.000142183831337245, 0.00115880777082488, -0.00126612622242972, 0, 0, -0.000585233683795706, 0.000797440307860737, -0.00064596101386983, 0.00140417461820286, -0.00141935767825231, -0.000979463945489, -0.00196314367033811, 0.000784634785163806, -0.00117608791206894, 6944444.44665773, -0.000514694678544602, -0.00283553082517336, -0.0042949351896274, 0.00111870891436386, -0.00334421823200243, -0.00226135875958744, -0.00248744835817424, -0.00114489104697948, -0.0049926857808962}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 9.48888012680515e-08, 0, 0, 0, 0, 0, 0, 0, -1.70743362313368e-06, 0, -1.5319101236231e-06, -33.333334273514, 0, 0, 0, 0, 0, 0, 1.19035374996909e-05, 0, 1.08959559972064e-05, 0, 0, 1.06314946864592e-05, 333.333341473531, 694.444458133936, 500.000016837438, 0, 0, 0, 0, 0, 0, -4.25370870101555e-05, 0, -4.52201805383116e-05, 0, 0, 0, -9.72607927729059e-05, 0, -7.60832913879134e-05, -1.89780473708972e-05, -4861.11120117055, -1666.66677263896, -6944.44450791446, -5000.00004781237, -10416.6668086964, 0, 0, 0, 0, 0, 0, 0.000122270967378463, 0, 9.52757719103382e-05, 0, 0, 0, 0.000286334188389331, 0, 0.000320904379282545, 5.97417212199142e-05, 13888.8891771119, 0.000334426513845767, 0, 0, 0.000620549770246992, 0.000219681532215176, 0.000104120222799558, 48611.1114309199, 72916.6674541742, 16666.6670702199, 34722.2228628637, 104166.667019537, 0, 0, 0, 0, 0, 0, -0.000152874978339495, 0, -0.000187413916080578, 0, 0, 0, -0.000503290905772133, 0, -0.000288777168901229, -5.73543615222262e-05, -13888.8894817422, -0.000595996643579753, 0, 0, 0, -0.00174843214157676, 0, -0.00119588341736108, -0.000426535416132661, -0.000394459047909952, -138888.889974186, -208333.335723651, -0.000897170073874997, -0.00201790589290215, -0.000869755992121005, -0.000654621121976001, -243055.558433929, -0.000411277635407315, -729166.668784109, -347222.223619661, 0, 0, 0, 0, 0, 0, 0.000161219332345521, 0, 0.00016754598400452, 0, 0, 0, 0.00053409793709713, 0, 0.00028570227450922, 3.9959805649639e-05, 0.000581249971073941, 0.000836961269335341, 0, 0, 0, 0.00194487249750631, 0, 0.0012227483389848, 0.000499180303048969, 0.000490949833127284, 138888.890090535, 208333.336229324, 0.00158360549813742, 0.00193009313381485, 0, 0, 0.00203447141864914, 0.000749933332739652, 0.000243107262194328, 0.001130485499258, 694444.451116453, 0.000732631555130008, 0.00666785086686292, 0.000520337603360649, 2083333.33667316, 0.00401218214331816, 0.000728313975927679, 0.00112072059741391, 2430555.55951488, 0, 0, 0, 0, 0, 0, -7.25307825690607e-05, 0, -8.81564049450925e-05, 0, 0, 0, -0.000191552585978078, 0, -0.00011725824308769, -2.76618972190429e-05, -0.000303242376368866, -0.000511675006229033, 0, 0, 0, -0.00112209483047554, 0, -0.000191687840494052, -0.000281529459179899, -0.000293800155939436, -0.000698121060671707, -0.00167413106306261, -0.00123573151171497, -0.00163920376448555, 0, 0, 0, -0.00195378312178499, 0, -0.00157190707693443, -0.00112373897152096, -0.000722715898371787, -0.0015148228928223, -694444.451066645, -0.00169763084236734, -0.00473433262022334, -0.00116589382430974, -2083333.33664994, -0.0027133288190245, -0.000919655591899822, -0.000600364122669212, -0.00782257227247074, -0.00160196533005084, -0.000255436410343242, -0.00184714792098565, -6944444.45008019, -0.0037531925637463, -0.00507121019508443, 0.00151462832731055, 0, 0, 0, 0, 0, 0, 1.9473868772345e-05, 0, 2.29097813327013e-05, 0, 0, 0, 5.87051320973922e-05, 0, 8.5713845977786e-06, 5.99761460501603e-06, 7.58146605641856e-05, 0.000126908287961397, 0, 0, 0, 0.000312693812567288, 0, 0.000135367164096593, 7.68244901412326e-05, 4.6824187625812e-05, 0.000231228322949496, 0.000429224576770936, 0.000425137445154064, 0.000241732200319684, 0, 0, 0, 0.000982051288197828, 0, 0.000537225115211947, 0.000309076813915878, 0.000330721702313124, 0.000543428239877152, 0.00198539218594882, 0.000524989405198158, 0.00157109255195045, 0.000201600679301564, 0.00125735483582016, 0.000742472438385079, 0, 0, 0.000815873923998431, 0.000467447139031254, 0.000387603826800254, 0.000590546155395313, 0.00394832659710497, 0.000640005970145102, 0.00277662729331584, -0.000252040113161196, -2.70253821143529e-05, 0.00122458284582596, 6944444.44712239, 0.00053990624647132, 0.00242752259960655, 0.00104934239726103, 0.000583313273418391, 0.00350097151651768, 0.000715925521522988, 0.000544909544614549, 0.000832487187160215}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 7.0667551519229e-08, 0, 0, 0, 0, 0, 0, 0, -1.00186896930679e-06, 0, -1.07071952812944e-06, -31.2500011792383, 0, 0, 0, 0, 0, 0, 7.38761903787556e-06, 0, 8.32679425722094e-06, 0, 0, 2.19070230113104e-05, 312.500004470259, 572.916682453103, 572.916681020487, 0, 0, 0, 0, 0, 0, -4.62094319355973e-05, 0, -3.32737458489096e-05, 0, 0, 0, -0.000108521798591642, 0, -8.49792935363788e-05, -3.01771091798598e-05, -3125.00009627144, -3125.00008201647, -5729.16674098082, -5729.16676046059, -10503.4723712311, 0, 0, 0, 0, 0, 0, 0.000107056253560546, 0, 9.34066844169857e-05, 0, 0, 0, 0.000289808077147763, 0, 8.71144648642314e-05, 6.57172780541482e-05, 5208.33366256268, 5208.33368947829, 0, 0, 0.000168803543206314, 0.000226366980027713, 0.000234574386145914, 31250.0004402174, 57291.6673474789, 31250.0003485103, 57291.6673444741, 105034.722924297, 0, 0, 0, 0, 0, 0, -0.000199613631483952, 0, -0.000144765669695366, 0, 0, 0, -0.000581350616302671, 0, -0.00034287100809338, -0.000128070271721886, -0.000698630582665683, -0.000805805999689191, 0, 0, 0, -0.000619503605429215, 0, -0.000727936497713183, -0.000381077174414428, -0.000367741599840704, -52083.3343107255, -95486.1127444556, -52083.3344756806, -95486.1126486371, -0.000895427922009682, -0.000992838508523726, -312500.002390821, -0.00160063851323093, -572916.66889633, -572916.669162686, 0, 0, 0, 0, 0, 0, 0.000156956549988124, 0, 0.000145435219239463, 0, 0, 0, 0.000557790988773128, 0, 0.000299665426944188, 0.000109855330083852, 0.000765457973065742, 0.000777174577056684, 0, 0, 0, 0.00104771849312705, 0, 0.00105415060703395, 0.000438650801538632, 0.000361901526555503, 0.00176392584864681, 0.00255204797067989, 0.00146923406231877, 0.0018337058738985, 0, 0, 0.00156388973009003, 0.00163544821181026, 0.00149869037107255, 0.00235050558523701, 520833.337254298, 0.0019652636554739, 520833.336049306, 0.00201966491627493, 954861.115284226, 954861.11445941, 0.00361233516705485, 0.00473551288801796, 3125000.00466533, 0, 0, 0, 0, 0, 0, -9.42112103755472e-05, 0, -8.23018987893094e-05, 0, 0, 0, -0.000381257454031991, 0, -0.000219057025363374, -5.75286061862493e-05, -0.000376392577888552, -0.00051797297537359, 0, 0, 0, -0.000760592110728281, 0, -0.000858619854684197, -0.000233894693117333, -0.000287715199774893, -0.00137838290250879, -0.00204115040940511, -0.00126460723654574, -0.000933358926785741, 0, 0, 0, -0.00114469508008746, 0, -0.00153038775586004, -0.000648581134034921, -0.000816245425541332, -0.0020330357291039, -0.00379776231573566, -0.00118114779746632, -0.00285623526294154, -0.000636319668945959, -0.00430408836725535, -0.0036179205467469, -0.00204497302536585, -0.00167502505061637, -868055.559290007, -0.00211329768584537, -0.00327005091805433, -0.00296158639111603, -5208333.34071778, -0.00250469692271934, -5208333.33974599, -0.00973484776900638, 0, 0, 0, 0, 0, 0, 1.96820453774944e-05, 0, 1.90157218960528e-05, 0, 0, 0, 8.08815449757966e-05, 0, 6.19683318380936e-05, 9.73286249812968e-06, 0.000117837356204883, 0.000157214045469254, 0, 0, 0, 0.000141931764791972, 0, 0.000147199902251838, 5.60809934149093e-05, 8.75935197003868e-05, 0.000323285819323052, 0.000583490283288955, 0.000396283921662688, 0.000463620385951281, 0, 0, 0, 0.00014728752411114, 0, 0.000220493878360515, 0.000152043660593511, 0.000228501598108234, 0.000691219608120382, 0.00101302577985084, 0.000590100081867763, 0.000641886945710027, 0.00024577736861864, 0.00123356137446052, 0.00155295450218071, 0, 0, 0.000364657522947008, 0.000666059434884906, 0.000657485223938452, 0.000898951670391294, 0.00136643245562547, 0.000454106373566053, 0.00203551247137925, 0.00117449365781545, 0.00130100177076761, 0.0016365367558479, 0.00277411590827922, 0.00251677672183131, 0.0021712658801995, 0.00110530484755101, 0.0019216147429784, 8680555.55986494, 0.000976010166123854, 0.00333714594460038, 0.00327199852313304}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.06807778888858e-07, 0, -1.36600230675131e-07, -33.3333336286363, 0, 0, 0, 0, 0, 0, 1.00613440126185e-06, 0, 2.49461503607035e-06, 0, 0, -9.14741834323465e-06, 333.333337859286, 500.000002193924, 694.44445027465, 0, 0, 0, 0, 0, 0, 5.99477714260503e-07, 0, -1.93198642044162e-05, 0, 0, 0, 3.70504230441872e-05, 0, 1.70473916511184e-05, -7.96214985520409e-07, -1666.66667805497, -4861.11116125924, -5000.00002736745, -6944.44448142547, -10416.6667029877, 0, 0, 0, 0, 0, 0, 1.56032539734175e-05, 0, 5.60353903888417e-05, 0, 0, 0, -0.000126825864362905, 0, 0.000125235044256621, -7.0884020493099e-06, 5.20061049897487e-05, 13888.8890205064, 0, 0, -2.11596021990161e-05, 9.35591210272577e-06, 0.000145034017793659, 16666.6667164566, 34722.2222520822, 48611.1114265282, 72916.6668178613, 104166.666969484, 0, 0, 0, 0, 0, 0, -3.01764300148672e-05, 0, -0.000108313399291962, 0, 0, 0, 0.000143882604773502, 0, -0.000113193129113676, 4.56244401073149e-05, -0.000149626115517196, -13888.8891336373, 0, 0, 0, 0.000276137013475874, 0, 0.000141657017299805, 8.20740704007521e-05, -0.000187096181290664, -0.000411554202445384, -0.000277809113660564, -138888.889579119, -208333.334001716, 4.35798630346356e-05, -0.000606743266659242, -243055.555573934, -0.000616193109309203, -347222.222912926, -729166.668316409, 0, 0, 0, 0, 0, 0, 5.89492000503219e-05, 0, 0.000111640185410984, 0, 0, 0, -6.59143366780333e-05, 0, 0.000242332356410596, -3.85668656514587e-05, 0.000236099846818802, 0.000409521827475963, 0, 0, 0, -0.000616582497204352, 0, -0.000124458158069031, -0.00026876890014203, 0.000276849956857791, 0.000397602392489311, 0.000347142820231282, 138888.890004721, 208333.334619664, 0, 0, -0.000500083055523323, -0.000746274124368523, 0.000494731060973893, -0.000457183652225232, -0.000800457798247167, 0.00119580804237852, 694444.445649524, 0.000339238342053914, 0.00125862566190463, 2083333.33767574, 0.000693165145221582, 0.00311520532833775, 2430555.55811509, 0, 0, 0, 0, 0, 0, -2.77247408146524e-05, 0, -6.09932699267302e-05, 0, 0, 0, -5.06419908160067e-06, 0, -0.000120449350804501, 2.73400761614718e-05, -0.000212014613761764, -0.000240023443297014, 0, 0, 0, 0.000444853329407151, 0, 6.59218825507237e-05, 0.000154699202222335, -7.40933532807436e-05, -0.000271716180587017, -0.000280666485405866, -0.000619256027848237, -0.00148292799386363, 0, 0, 0, 0.000528234997907262, 0, 0.00029501751446826, 0.00043858348445773, -0.000327352298697067, -1.274891917027e-05, 0.000888826110253203, -0.00147094408011123, -694444.446084777, -0.000327911347334672, -0.00117400940422546, -2083333.33705239, 0.000871414379599853, -0.000837487421537134, -0.000257756442145155, 0.000127133939147123, -0.00292872029286484, -0.00175996378445375, -0.000922341701473436, -0.00710444519299823, -6944444.44760374, -0.000902462250287712, 0, 0, 0, 0, 0, 0, 9.41192615783017e-06, 0, 1.49091518572034e-05, 0, 0, 0, 2.26781081177431e-05, 0, 2.56505843380531e-05, -3.95221010446754e-06, 4.39067360503557e-05, 5.19921501217266e-05, 0, 0, 0, -4.86734910428345e-05, 0, 8.18688692218685e-05, -4.47594353687516e-05, 2.11227532922479e-05, 0.000127870283096085, 0.000169357519512764, 0.000141006317066297, 0.000240171347854544, 0, 0, 0, -0.000143246045650737, 0, 0.000240728988828245, -0.000150593998971704, 8.62005038895459e-05, 5.87861185982402e-05, 3.03037107844355e-06, 0.000354364269449552, 0.000884610116796258, -4.75606885639301e-05, 0.00101345097849765, 0.00119864387378908, 0, 0, -4.93914491900433e-05, -0.000439003389134938, 3.11962355053292e-05, -0.00029880516080596, -0.000303199474150308, 0.000515325142893978, 3.82955624334519e-05, -0.000766411939440128, 0.000794136584799535, 0.000903372553799373, 0.00056862841693273, 0.00229343316240657, 6944444.44707316, -0.000318977124768812, 0.00261593058833493, 0.000163288858471045, 0.00214343311831394, 0.000981023574804723, 0.00324805106952825}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -1.52004146693596e-08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.95339401858343e-07, -33.3333333765217, 0, 0, 0, 0, 0, 0, -1.00211811817038e-07, 0, -2.83426088185493e-06, 0, 0, 1.75419168054646e-07, 500.000000789449, 333.333333456084, 694.444446146651, 0, 0, 0, 0, 0, 0, -6.31517227570476e-06, 0, 1.81378242235504e-05, 0, 0, 0, -1.27282508059973e-05, 0, -3.1943061463674e-06, -1666.66666727811, -3.45841790272577e-06, -4861.11110989302, -4999.99999955379, -10416.6666880397, -6944.44447748203, 0, 0, 0, 0, 0, 0, 3.85138831028269e-05, 0, -7.85368336415407e-05, 0, 0, 0, -0.000101309847986657, 0, -2.46460931311717e-05, -6.72826334590337e-06, -2.91375337486947e-05, 13888.8888868351, 0, 0, 6.42923475055462e-05, 16666.6665701362, 34722.2223047713, -0.000158209262947475, 0.000137436342919813, 72916.6666535493, 48611.1113714005, 104166.666916086, 0, 0, 0, 0, 0, 0, -9.04949865422913e-05, 0, 0.000105664610665482, 0, 0, 0, 3.69150830701909e-05, 0, -1.15366634876336e-05, -6.91951606050363e-06, -8.62336539292657e-05, -13888.8888020792, 0, 0, 0, -0.000173623368683186, 0, -0.000226362874185536, 0.0002556432006342, -0.000107879238560973, 0.000349958320625431, -3.22404078744252e-05, -208333.333491014, -138888.8897474, 0.000448681667053999, -243055.555762458, -0.0010949883485955, -347222.22255952, -0.000983609952701487, -729166.667656519, 0, 0, 0, 0, 0, 0, 9.0338780480681e-05, 0, -0.000138119584929816, 0, 0, 0, -8.70120461679028e-05, 0, -0.000198676897516284, -5.12091300976474e-06, 0.000168075643790583, -0.000224894695566905, 0, 0, 0, 0.000227612316624368, 0, 0.00110093390679628, -0.000390302055653879, 8.61949996527508e-05, -0.000672917003795119, -0.000305821712821041, 208333.333285665, 138888.889405452, 0, 0, 0.0010815995772429, -0.000610480850808911, 0.000386923873297167, -0.000983658742100357, 0.000377268839482029, 694444.444171981, 0.00345148935130213, 0.000753103833902979, -0.000625370590137335, 2083333.33633934, -0.00113220699355189, 2430555.55748453, 0.001560226307231, 0, 0, 0, 0, 0, 0, -6.23424955604041e-05, 0, 7.98418192388562e-05, 0, 0, 0, 1.35189103298815e-05, 0, 0.000188621682549308, -1.07815103022669e-05, -3.14894562913234e-05, 0.000256497225344946, 0, 0, 0, 1.96448231774975e-05, 0, -0.000501835758107069, 0.000217455648680971, -0.000192244558578289, 0.000385242845830383, 0.000240406330657482, 0.000129450260581461, -0.000136842118186259, 0, 0, 0, -0.000435030097346117, 0, -0.00121677353884719, 0.000817293435832659, -0.000458158782301116, 0.000852414826849294, -0.00120739295871305, -694444.444346024, -0.00226873745546461, -0.000153497819903221, 0.00123345935487212, -2083333.33432212, 0.00153998758362836, -6.47359933546302e-05, -0.00208549776216179, 0.001034897810163, -0.000778025372065639, 0.00322118418952934, -0.00191605122841206, -6944444.4462925, -0.00434535342428938, -0.000436078922946951, 0, 0, 0, 0, 0, 0, 1.78258918036144e-05, 0, -1.53913451501509e-05, 0, 0, 0, 4.29729006173966e-06, 0, -5.67829090798783e-05, -2.13782381047883e-07, 5.70997298829733e-05, -6.54401259399293e-05, 0, 0, 0, -9.56969516604664e-05, 0, 9.35906974514728e-05, -7.70170962745779e-05, 8.08348853726328e-05, -7.87363374649496e-05, -4.60084411141357e-05, -5.34514589231782e-05, -1.31671051868417e-05, 0, 0, 0, 4.03560615639054e-05, 0, 0.00016895080583767, -0.000315438271202617, 0.00014728902890094, -0.000349602531628721, -2.88088952453263e-05, -8.79971317293225e-05, 0.000515031534513916, -0.000283131166096277, -0.000968772010842355, 0.000542762903089739, 0, 0, 0.000174463960764523, -0.000421969359857973, 0.000286157880780321, -0.000396879663498133, 0.000835335793907578, -0.000100579517030419, 0.00132844377413064, -0.00114700854744657, 0.000605215519697475, -0.00145149779624644, 4.99234743410827e-05, 6944444.44479497, 0.00102041422335255, -0.0013886007038679, 0.000721299392107429, 0.00277976065998083, 0.000797813909960266, 0.000674943954614009, 0.000850264767625229}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -2.99717328176966e-08, 0, 0, 0, 0, 0, 0, 0, -1.51655742402685e-07, 0, 9.410165080095e-07, -31.250000450493, 0, 0, 0, 0, 0, 0, 4.62397786868407e-06, 0, -1.07431789728315e-05, 0, 0, -3.77528224250743e-06, 572.916670487608, 312.500006480382, 572.916669975301, 0, 0, 0, 0, 0, 0, -2.33992427487808e-05, 0, 4.9297155665056e-05, 0, 0, 0, 3.11538849870897e-07, 0, 3.32202237983163e-05, -3125.00001570995, -3.27476306153521e-05, -3125.00001130199, -5729.16669420018, -10503.4722942663, -5729.16673292832, 0, 0, 0, 0, 0, 0, 5.22020447758055e-05, 0, -0.000116684958430666, 0, 0, 0, 0.000154293449949968, 0, -0.000131194589481664, 5208.33336554562, 0.000152562779745899, 5208.33324644168, 0, 0, -4.75814049341484e-05, 31250.0001157072, 57291.6669554745, 0.000225756920400316, 0.000328413482789066, 57291.6671082937, 31250.0001318355, 105034.722633988, 0, 0, 0, 0, 0, 0, -9.84364404319727e-05, 0, 0.000247436462883635, 0, 0, 0, -0.000249074088680788, 0, 0.000542427886223176, -8.14653508671991e-06, -0.00018019655956583, 0.000261376601764829, 0, 0, 0, -2.87307888349575e-05, 0, -1.14129926909764e-05, -52083.3334661815, -95486.1116854002, -0.000381448480859965, -0.00108424630879392, -95486.1116922003, -52083.3327846184, -0.000310441199069789, -312500.001214109, -0.00123756063921842, -572916.667909767, -0.000674182521750948, -572916.668555442, 0, 0, 0, 0, 0, 0, 0.000107829515576874, 0, -0.000213174252908584, 0, 0, 0, 0.000358437325472691, 0, -0.000600177895004373, 3.25152219516354e-06, 0.000217281365148783, -0.000376651395242624, 0, 0, 0, 4.27337933989887e-05, 0, -0.00111757459694785, 1.0292234066911e-05, 0.000786361843472717, 0.000631181418909748, 0.00161179308398396, 0.000706161915594531, -0.000944084527588969, 0, 0, -1.96137847603353e-05, 0.000145692096531561, 520833.335353261, 0.000513236060780112, 0.00377832966588223, 520833.336007838, -0.000177010147086371, 954861.111949553, 0.00349269859798451, 954861.111519069, 0.00407508602170031, 3125000.00410638, 0.00601773521740506, 0, 0, 0, 0, 0, 0, -5.47951968520725e-05, 0, 0.000110435335260938, 0, 0, 0, -0.000270698961873396, 0, 0.000452628368184681, 3.01476331993083e-05, -0.000281774301626252, 0.000181123024949288, 0, 0, 0, -0.000266658002060512, 0, 0.000606595618599378, 0.000117326006148464, -0.000264593966844048, -0.000318783815031571, -0.00109876066484199, -0.000361485476603442, 0.00108967895140982, 0, 0, 0, -0.000558673727493506, 0, 0.000148783656401651, -0.000137121421351349, -0.00132369027235875, -0.000195189281159146, -0.00185062543085426, -0.00162417032387867, -0.000291546301289582, -0.00028221488098871, -0.00274897510283339, -0.00119919682285811, -0.00109826221503318, -868055.558085687, -0.00430138127519562, -0.00181586915905278, -5208333.33840422, -0.00558886006946327, -0.00755298851004222, -5208333.3374207, -0.00381039756913525, -0.00971155555067, 0, 0, 0, 0, 0, 0, 1.03649917159773e-05, 0, -2.74036154468578e-05, 0, 0, 0, 7.20153411528162e-05, 0, -0.000140766808199257, -4.31269324539494e-06, 3.58566396416984e-05, -5.91128329911132e-05, 0, 0, 0, 0.000204613866352576, 0, -0.000213714589599922, -3.01170457354035e-05, 4.22367795824362e-05, 4.65357090379642e-05, 0.000312374832250406, 4.9243064350658e-05, -0.000492895318355827, 0, 0, 0, 0.000149373699121623, 0, 0.000253050608435239, 3.4541166909837e-05, 0.00042252458793573, 5.82400795885786e-05, 0.00105538026100392, 0.000532707329064601, -0.000148785613152021, 0.000338803916865118, 0.00146073844255449, -0.00021370956288345, 0, 0, 0.000387235347118559, 0.000114331118788165, 0.00062240480649351, 3.20278408159815e-05, 0.00152146336208866, 0.000938179920589544, 0.00130718102284265, 0.00118689369629715, 0.00150586704506773, 0.00187944370952841, 0.00346403396506215, 0.00186637110267672, 0.00220671816209547, 0.00125895379966235, 8680555.55824889, 0.00285730159381995, 0.00243153666277225, 0.00422130184896847, 0.00492164261680461}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 2.65802278094129e-08, 0, 0, 0, 0, 0, 0, 0, -9.53573798407926e-08, 0, -6.31529823602556e-07, -33.3333336665354, 0, 0, 0, 0, 0, 0, 1.4047040393666e-06, 0, 5.90497865126419e-06, 0, 0, 4.33448210141851e-06, 694.444449822603, 333.333337325681, 500.000004866962, 0, 0, 0, 0, 0, 0, 3.14285224768194e-06, 0, -2.15024480805389e-05, 0, 0, 0, -1.88823713182169e-05, 0, -3.13938426076989e-05, -4861.11114069403, -3.91594345213348e-05, -1666.6666904474, -6944.4445000672, -10416.6667143266, -5000.00003326601, 0, 0, 0, 0, 0, 0, -6.32965882116037e-06, 0, 4.46144347237012e-05, 0, 0, 0, 9.14163266548872e-05, 0, 8.74696836683888e-05, 13888.8889664325, 0.000108126392158221, 9.12146121238805e-05, 0, 0, 0.000279623866044366, 48611.1112719008, 72916.666932044, 0.000119935087282922, 0.000335755282987023, 34722.2222952424, 16666.6668573012, 104166.667019496, 0, 0, 0, 0, 0, 0, 3.11893256469651e-05, 0, -7.3056685206859e-05, 0, 0, 0, -8.75610172101957e-05, 0, -0.000348176448177585, -13888.8890475131, -0.000151591506002091, -0.000139011574367314, 0, 0, 0, -0.000452424523138837, 0, -0.000346264291935601, -138888.8892894, -208333.333873045, -0.000263080922850634, -0.000466293200026534, -0.000456488273885801, -0.000225597781804078, -0.000444668534554379, -243055.556173876, -0.000614768788786949, -729166.66785825, -0.00161537914039379, -347222.222685533, 0, 0, 0, 0, 0, 0, -2.56446334563224e-05, 0, 5.76079669157446e-05, 0, 0, 0, 6.56816706398158e-05, 0, 0.000209607326190911, 0.000153024049135452, 0.000188544201250741, 0.000179283785107861, 0, 0, 0, 0.000812288948576143, 0, 0.00118191693947794, 138888.889466427, 208333.333797093, 0.000246520319480683, 0.000713351645447398, 0.000451741920597224, 0.000525568055283709, 0, 0, 0.00122905377784448, 0.000817939752320857, 694444.445969987, 0.000654820977468917, 0.000805664194410863, 0.000881177500572758, 0.00168212161166106, 2083333.3359227, 0.00189292616606737, 0.00200416971614282, 0.00194703197934501, 2430555.55765417, 0.00266579133104188, 0, 0, 0, 0, 0, 0, 1.0865462103576e-05, 0, -2.99139839469584e-05, 0, 0, 0, 1.09163172334848e-05, 0, -0.000111584942389913, -9.11765490445589e-05, -9.10472005213007e-05, -5.22683487002811e-05, 0, 0, 0, -0.000431788402948805, 0, -0.00070710003841249, -0.000451280183390422, -0.000436921702532348, -0.000287923926566432, -0.000560969549532625, -0.000277518865569911, 3.04186877465701e-05, 0, 0, 0, -0.000922667681417468, 0, -0.000883965218489788, -0.000619907912779082, -694444.445660424, -0.000261721594153028, -0.00118843904799134, -0.000759178629684902, 0.000333459005835733, -2083333.33559028, -0.00157770278672683, -0.000296589619363403, -4.83876254247427e-06, -0.00119164344636541, 2.50346569872839e-05, -0.00255907021455075, -6944444.44758073, -0.00136490330757711, -0.000585097394993173, -0.0028303831116811, -0.00257867278849916, 1.43214712760134e-05, 0, 0, 0, 0, 0, 0, -4.22942566610083e-06, 0, 6.78081479938652e-06, 0, 0, 0, -1.31172289665007e-05, 0, 4.6711791034584e-05, 1.6999844046462e-05, 1.97476878011184e-05, 1.04834596602482e-05, 0, 0, 0, 2.47656341327201e-05, 0, 7.84590543433861e-05, 0.000100317992578837, 0.000130383095955868, 0.000148007798015176, 0.000122290902148194, 0.000104254981557824, 0.000108438058140624, 0, 0, 0, 0.00030483250561751, 0, 4.79484328782958e-05, 0.000290219998765216, 0.000258771152298006, 0.000103584713180656, 0.000469362248563262, 0.00020774690154182, -0.000179941243043688, 0.000542963930617287, 0.000151557930277795, 0.000157311075775725, 0, 0, 0.000378704251164195, 0.000203889622212822, 0.000637296429883602, 0.000168242303611758, 0.000695741737364474, 0.000406607225823054, -0.000148999328198573, 0.000697242669474018, 6944444.44585055, 0.000449189238624876, 0.000757940719185061, 0.000476362764638293, -0.000380658966230675, 0.000810925983418802, 0.0012966092372548, 0.00063380355992834, 0.00265342330357266, 0.00108077144094609, 0.000441867600687087}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -2.42605350464152e-08, 0, 0, 0, 0, 0, 0, 0, 2.32084657257725e-07, 0, 3.35548389111156e-07, -27.7777779287087, 0, 0, 0, 0, 0, 0, -7.01988417540637e-06, 0, -1.13747531412864e-06, 0, 0, 1.75615510056025e-08, 509.259264806931, 416.666669698038, 416.666665721062, 0, 0, 0, 0, 0, 0, 1.94841436529376e-05, 0, 5.01942120223568e-06, 0, 0, 0, 1.88129045354295e-06, 0, -1.6543302809106e-05, -2777.77780229187, -1388.8888890988, -1388.88884630893, -7638.88895210674, -7638.88893444847, -6250.00000958695, 0, 0, 0, 0, 0, 0, -3.71663205897356e-05, 0, -3.75726942477365e-05, 0, 0, 0, -0.000302813225136711, 0, 6.67268124565931e-05, 4629.62968746811, -4.13184731551797e-05, -0.000174274330717179, 0, 0, -0.0001018688142867, 41666.6669844959, 41666.666804534, 25462.9632375627, 20833.3331723749, 25462.963094672, 20833.3331391102, 114583.333919022, 0, 0, 0, 0, 0, 0, 3.57370683626758e-05, 0, 1.91067917908213e-05, 0, 0, 0, 0.000443741027979286, 0, -5.87247126726819e-06, -4.07972062615665e-05, -0.000158572258341039, 0.000412718105646721, 0, 0, 0, 3.30874700211784e-05, 0, -0.000323020575405319, -69444.4450464242, -69444.4447498607, -0.000678833130717054, 2.72445579615696e-05, 0.000138939629399565, -1.0005607231823e-05, -138888.890031837, -138888.889610741, -69444.4439491536, -625000.002056753, -381944.446687355, -381944.446137462, 0, 0, 0, 0, 0, 0, -4.52778581900676e-05, 0, -5.39803111248045e-05, 0, 0, 0, -0.000615320864364619, 0, 3.04679802813566e-05, -6.404813666983e-06, 0.000108121476674698, -0.000605304578617966, 0, 0, 0, -0.00062656264194821, 0, 0.000594219099137834, 0.000566457131747938, 0.000270957935795381, 0.000596368507561302, -0.000389985413271048, -0.000582406271747975, -0.00160325383876118, 0, 0, -1.18751694991238e-05, 231481.483302784, 231481.481728271, 0.00140397367135378, -0.00144277130794474, 9.139053821393e-06, -0.000725604321045187, 1041666.66947294, 0.00173009351036751, 0.0016968332190956, 2083333.33781493, 2083333.33865112, 1273148.14856556, 0, 0, 0, 0, 0, 0, 1.77008221065196e-05, 0, 3.7100264256245e-05, 0, 0, 0, 0.000373894394839839, 0, -9.2893547259528e-06, 1.4661328327585e-06, 8.56151318616329e-05, 0.000413268277814955, 0, 0, 0, 0.000701224727503143, 0, -7.76694540605088e-05, -0.000249982976433347, -3.29552985418397e-05, -0.000215233580974672, 0.000380322672097341, 0.000596433006124732, 0.000737396164536436, 0, 0, 0, 0.000609092162811643, 0, -0.000740240335134388, -0.00135044713468269, -0.000321766478536532, -0.00142275984259774, -0.000204869399973902, -3.4354658440334e-05, 3.36883732669968e-05, -0.0015429381094597, -0.00137702282049204, 3.88810647866722e-05, -0.00164762577194777, -0.000475318326063972, 0.00134117320437274, -3472222.22791675, -3472222.22544762, -0.0035954495574391, -0.00394755267422498, -0.00320669018109358, -0.000655380444055147, -6944444.45331671, 0, 0, 0, 0, 0, 0, 3.52711383923794e-06, 0, -5.08463311277544e-06, 0, 0, 0, -8.14352896244745e-05, 0, -2.69472156988768e-05, -2.35732639323493e-06, 3.50997582913332e-05, -9.26636332404724e-05, 0, 0, 0, -0.000251593030175713, 0, 0.000177987187057928, 5.904540152836e-05, -3.92404213864638e-05, -1.76849479005586e-05, -5.20468984059837e-05, -0.000222820397670472, -0.000395847731107898, 0, 0, 0, -0.00020876041882007, 0, 0.000204974241142441, 0.000246471897629979, 0.00016757693246541, 0.00026796371814662, -0.000636730077487504, -0.000137860775336652, -0.000225915910375886, 8.38377764000072e-05, 0.000227573235802698, -5.77835013762127e-05, 0, 0, -0.000243616300702632, 0.000749557343990717, 0.000388181128636053, 0.000643685906933376, -0.000905869532072198, 0.000205631798876567, -0.00034444065735215, 0.00139001882473883, 0.00114644272626875, 0.00160807784414245, 0.000311251147516198, 0.000815457644641033, 6.19712021919479e-06, 0.00293286643038554, 0.00206656939517698, 0.00165204886685145, 11574074.0773201, 0.00369096967661485, 0.0026931498901531}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 2.20204594696996e-08, 0, 0, 0, 0, 0, 0, 0, -1.4983226380392e-07, 0, -3.18376821163802e-07, -27.7777779998769, 0, 0, 0, 0, 0, 0, 5.05378596176197e-06, 0, 2.18360680422473e-06, 0, 0, 6.1806458846433e-07, 416.666668933801, 509.259257895669, 416.666669076776, 0, 0, 0, 0, 0, 0, -1.22198792445376e-05, 0, -1.02439236377034e-05, 0, 0, 0, 5.08951496789657e-06, 0, -7.16296500992066e-06, -1388.88890618592, -2777.77776765673, -1388.88892379926, -7638.8889133152, -6250.00002417503, -7638.88890301896, 0, 0, 0, 0, 0, 0, 1.70698499264274e-05, 0, 4.52239079812698e-05, 0, 0, 0, 0.000128961121660062, 0, 2.3425422780149e-05, 4.7985014231788e-05, 4629.62956631269, 8.95460808898949e-05, 0, 0, 7.68320568021223e-05, 25462.9630444448, 20833.3335154097, 41666.666733946, 41666.6668185251, 20833.3333967831, 25462.9630746672, 114583.333344032, 0, 0, 0, 0, 0, 0, -3.39487026229345e-05, 0, -2.13460932438377e-05, 0, 0, 0, -0.000147786995605746, 0, -9.52940614028965e-05, -0.000122433240804752, 0.000418197438252411, -0.000217052360913709, 0, 0, 0, 0.000171046702601186, 0, 4.78468978110273e-05, -0.000225685244623041, -0.000223273592578852, -69444.4444557555, -69444.4443505762, -0.000358108156068265, 0.000671435940417683, -138888.889351303, -69444.4445616251, -138888.889475584, -381944.444534476, -624999.999563268, -381944.444184002, 0, 0, 0, 0, 0, 0, 4.28244509856906e-05, 0, 4.54718759930559e-05, 0, 0, 0, 0.000118605603586681, 0, 4.26087985297841e-05, 0.000164339188520969, -0.000490256331575165, 0.000305454000298767, 0, 0, 0, 2.44041589210437e-05, 0, -0.000181640568346671, 0.000233189919249393, 0.000407633834743479, -0.000111949391465467, -0.000226340611016696, 0.000631604213267577, 0.00013948962071829, 0, 0, 0.000722471489546261, 0.000925474048467284, 0.00116809117840825, 231481.482795551, 231481.482285742, 0.00112677176469847, -0.000217476139621145, 0.000103902222977306, 1041666.6680552, -0.00107728561427163, 2083333.33477006, 1273148.14729064, 2083333.33562562, 0, 0, 0, 0, 0, 0, -2.79008421585092e-05, 0, -2.76555745594948e-05, 0, 0, 0, -0.000106362480107943, 0, 2.30224281874029e-05, -7.91588100490197e-05, 0.000199958533024361, -0.000223650006564792, 0, 0, 0, -9.74859236993001e-05, 0, -0.0002600694136851, -0.000224837046283704, -0.000286096549289602, 0.000215977140007234, 0.000312256217522397, -0.000455209951044609, 0.000398317900734432, 0, 0, 0, -0.000871711572968185, 0, -0.000189114824230241, -0.00046264700309953, -0.000578573800518818, -0.000622361595686654, 0.000831648744455344, -0.000460114037706119, 0.000785714558646717, 6.57103854648177e-05, -0.000498617651407985, 0.000349294726991057, -0.00176775831593742, -0.000852907903903196, -0.00188965597969721, 0.000398573532929476, -0.00131123012187222, -3472222.22582675, -3472222.22274207, 0.000254691193370872, -0.000398404034072755, -6944444.44388297, 0, 0, 0, 0, 0, 0, 2.04237179959998e-06, 0, 2.95844187451661e-06, 0, 0, 0, 8.33537526309263e-06, 0, 1.08669327146032e-05, 2.17270038412121e-05, -0.000104338733511864, 5.3550288798991e-05, 0, 0, 0, 4.48915837212568e-05, 0, -9.1920687742851e-05, 5.0894708104831e-05, 0.000128897905204669, -0.000109034071304969, -0.000200392810980861, 0.000133734695337364, -5.66137560228374e-05, 0, 0, 0, 1.93680799284464e-05, 0, 0.000170464007098147, 0.000232390480669073, 9.35607646000806e-05, 0.000182984904096709, 0.00038348566766765, 0.000227463034292625, -0.000436949068187811, 0.000213978913091018, 0.00033883614391225, -0.000245441331742482, 0, 0, 0.000503606422648144, 0.000505368808727567, -3.24072601852527e-06, 0.000685174510767273, 0.000674578620186355, 9.7609884945179e-05, 0.000521148734021979, 0.000371215918704495, 0.00058039287022696, 0.00131332176650014, 0.00109898845824019, 2.71673284974385e-05, 0.00023644876733048, 0.000383255625835113, 0.000118260445696528, -0.000323097009483084, 0.000319644234990994, 11574074.0750206, 0.000553294227591597}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7.50906504624557e-08, 0, 0, -27.7777779304667, 0, 0, 0, 0, 0, 0, 1.82177046554202e-06, 0, -4.71344064608866e-07, 0, 0, -8.69198815717671e-07, 416.666670288366, 416.666671005042, 509.259261723889, 0, 0, 0, 0, 0, 0, -7.39962313513817e-06, 0, 4.6509712858782e-06, 0, 0, 0, 6.86432890409556e-06, 0, -7.86273667213386e-07, -1388.88890716052, -1388.88891890435, -2777.77777729399, -6250.00003533512, -7638.88892542256, -7638.88892580192, 0, 0, 0, 0, 0, 0, 3.04035001283584e-05, 0, -1.83630991641349e-05, 0, 0, 0, 6.19508714081779e-05, 0, 2.01576985573472e-05, 4.99223168937407e-05, 0.000106980604540166, 4629.62958832197, 0, 0, 0.00014434926614522, 20833.3334624107, 25462.9631279616, 20833.3334933955, 25462.9632429438, 41666.6667756255, 41666.6668549306, 114583.333759342, 0, 0, 0, 0, 0, 0, -4.5104393315525e-05, 0, 3.96134260901466e-05, 0, 0, 0, 1.68539165611234e-05, 0, -5.85258699257684e-05, -7.51726113650471e-05, -0.000161966865469313, 0.000196510011050912, 0, 0, 0, -0.000244118413252667, 0, -0.000398598322457022, -0.000184639317367348, -0.000482367599383291, -0.000219349966238949, -0.000502937444193669, -69444.4444721008, -69444.444794095, -69444.4447245458, -138888.889383532, -138888.88964619, -381944.445782775, -381944.446224785, -625000.001273899, 0, 0, 0, 0, 0, 0, 7.38402990123338e-05, 0, -4.27486028010395e-05, 0, 0, 0, 0.000103213984505779, 0, -2.64699278016335e-05, 6.67734784665417e-05, 0.000202784772454013, -0.00025540725340254, 0, 0, 0, 0.000234469342152288, 0, 0.000681932341486081, 0.000221767743990689, 0.000490056220793449, 0.000227892634029052, 0.00064822891634515, -0.000262856741826313, 0.000226817086587986, 0, 0, 0.000317656765544866, 7.79375176454093e-05, 0.000804787188739924, 0.000299636937925702, 0.0019615340536246, 231481.481779457, 231481.483386299, 0.00192040758794779, 0.00248781175888565, 1041666.66853916, 1273148.15139866, 2083333.33672086, 2083333.3382666, 0, 0, 0, 0, 0, 0, -4.36793756057628e-05, 0, 2.22479081975533e-05, 0, 0, 0, -4.39712666185213e-05, 0, 5.06096864801799e-05, -4.02016160969793e-05, -0.000162344298869482, 0.000186203599736808, 0, 0, 0, -9.37114787841513e-05, 0, -0.000372472217198024, -7.8603673821766e-05, -0.000410675640264765, -0.000149615701822322, -0.000433445618749414, 0.000289372953431763, 0.00028889909544031, 0, 0, 0, -0.000200735061142147, 0, -0.000507812671528991, -0.000101092643497529, -0.000844483264479743, -5.46060725406555e-05, -0.00117891460292205, -1.68208567012546e-05, -0.000963337868689459, -0.00159558785276468, -0.00202891445601857, -0.000617787397377943, 0.000123066430357004, -0.000427887154310137, -0.00122363798135258, -0.00306544244647627, -0.00342328236969636, -0.00290963793888651, -0.00384074464819453, -3472222.2258842, -3472222.22717077, -6944444.44993673, 0, 0, 0, 0, 0, 0, 9.93894752792667e-06, 0, -5.27087809577257e-06, 0, 0, 0, 2.5817684316305e-05, 0, -2.12917895002638e-05, 7.21641641169847e-06, 3.9637945486216e-05, -5.98306323952577e-05, 0, 0, 0, 1.34592199502603e-05, 0, 6.52094715959144e-05, 2.11260769526714e-05, 0.000102697837443577, 7.94522102250656e-05, 0.000119146632971891, -8.63562109610038e-05, -0.000100240801108243, 0, 0, 0, 0.000177763968509963, 0, 0.000195403607740635, 2.59426230367259e-05, 0.000286967848626307, 3.81058964801285e-05, 0.000228602319020902, -0.00010216268096195, 0.000398644606260838, 0.000469904032212746, 0.000405269802178549, -0.000215757461466537, 0, 0, 0.00024740235933297, -6.91303709763562e-05, 0.000327348814463074, -4.90663420219706e-05, 0.000852397794932338, 9.98650207010673e-05, 0.000581928469373565, 0.000790329010023961, 0.000982750618404429, 0.0009408395320534, 0.00160358833885293, 0.000823225774584093, 0.00108215941303931, 0.00145036690831425, 0.00157899180311053, 0.00226296944792717, 0.00307508710959534, 0.00281768744863804, 11574074.0769567}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 4500.00000483629, 0, 0, 0, 0, 0, 0, 0, -64303.5715084984, 0, -64303.5714953526, -64303.571541869, 0, 0, 0, 0, 0, 0, 391991.072154349, 0, 391991.07191823, 0, 0, 783982.14349664, 391991.072188222, 783982.144709432, 783982.144235299, 0, 0, 0, 0, 0, 0, -1333402.78121248, 0, -1333402.77985524, 0, 0, 0, -4000208.33816235, 0, -4000208.33608776, -1333402.78036709, -4000208.3455936, -4000208.34068217, -4000208.34462188, -4000208.34044121, -8000416.68455318, 0, 0, 0, 0, 0, 0, 2772048.62054899, 0, 2772048.61648529, 0, 0, 0, 11088194.4624503, 0, 11088194.4508362, 2772048.61636534, 11088194.4850678, 11088194.4663759, 0, 0, 16632291.6846466, 11088194.4770509, 11088194.4637949, 16632291.7288656, 33264583.4324266, 16632291.6949311, 33264583.4052561, 33264583.4169689, 0, 0, 0, 0, 0, 0, -3611111.12675689, 0, -3611111.11988815, 0, 0, 0, -18055555.5927038, 0, -18055555.5621066, -3611111.11749011, -18055555.6310918, -18055555.5955602, 0, 0, 0, -36111111.1631183, 0, -36111111.1499147, -18055555.6081025, -18055555.5898298, -36111111.2748792, -72222222.4790998, -36111111.1720899, -72222222.3666408, -36111111.2565641, -36111111.1668912, -108333333.663155, -72222222.3977008, -108333333.703462, -108333333.589976, 0, 0, 0, 0, 0, 0, 2881944.45982401, 0, 2881944.45335853, 0, 0, 0, 17291666.7106943, 0, 17291666.6674522, 2881944.44857762, 17291666.7472205, 17291666.7114704, 0, 0, 0, 43229166.7405433, 0, 43229166.7083391, 17291666.7158207, 17291666.7054464, 43229166.8915409, 86458333.6777656, 43229166.7462488, 86458333.4837741, 0, 0, 57638888.9784659, 43229166.8452395, 43229166.740848, 57638889.1800425, 172916667.311682, 57638888.9638595, 172916667.191638, 86458333.526378, 172916667.391183, 172916667.039771, 172916667.242463, 172916667.0626, 259375000.877219, 0, 0, 0, 0, 0, 0, -1289682.54791999, 0, -1289682.54480368, 0, 0, 0, -9027777.80580987, 0, -9027777.77495301, -1289682.54066053, -9027777.8239252, -9027777.80608041, 0, 0, 0, -27083333.3862029, 0, -27083333.3513528, -9027777.80301047, -9027777.802156, -27083333.4886721, -54166666.9001938, -27083333.3931664, -54166666.7429218, 0, 0, 0, -45138888.9717838, 0, -45138888.9638023, -27083333.4466552, -27083333.3938954, -45138889.1583341, -135416667.233204, -45138888.9424717, -135416667.056603, -54166666.7751593, -135416667.320665, -135416666.924972, -45138889.1303977, -45138888.9466441, -180555556.260532, -135416667.076718, -135416666.964622, -180555556.308847, -270833334.481783, -180555555.925223, -270833334.224957, -270833334.214411, 0, 0, 0, 0, 0, 0, 248015.87486509, 0, 248015.874276672, 0, 0, 0, 1984126.99152357, 0, 1984126.98286321, 248015.872907785, 1984126.99516524, 1984126.99179231, 0, 0, 0, 6944444.45988594, 0, 6944444.44574253, 1984126.98961786, 1984126.99043201, 6944444.48711724, 13888888.9524434, 6944444.46390769, 13888888.9035169, 0, 0, 0, 13888888.9166037, 0, 13888888.9110912, 6944444.47444381, 6944444.46544082, 13888888.9824955, 41666666.8529518, 13888888.9073448, 41666666.7732587, 13888888.9137553, 41666666.8893697, 41666666.736216, 0, 0, 17361111.1474345, 13888888.9663548, 13888888.9127384, 17361111.2241772, 69444444.7546592, 17361111.1257121, 69444444.7198243, 41666666.7763891, 41666666.7556659, 69444444.7889551, 104166667.179636, 69444444.5648392, 104166666.989819, 69444444.7141856, 69444444.5840219, 138888889.484439, 104166666.979298, 138888889.456713, 138888889.326415}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -124.999999996547, 0, 0, 0, 0, 0, 0, 0, 3366.07142835546, 0, 125.000000182841, 124.999999832244, 0, 0, 0, 0, 0, 0, -35810.5158720315, 0, -9.87216871729788e-07, 0, 0, -3241.07142984023, 1.66635207499007e-06, -3241.07142608073, 1.60295491580849e-06, 0, 0, 0, 0, 0, 0, 200451.388885221, 0, 5.03145999953973e-06, 0, 0, 0, 32569.4444781834, 0, -1.02180509208795e-05, -6.23333760726958e-06, 32569.4444341814, -1.04073991444585e-05, -2.05966687832683e-05, -1.52955105806389e-05, -1.9569471886484e-05, 0, 0, 0, 0, 0, 0, -653993.055537528, 0, -2.3874888174907e-05, 0, 0, 0, -167881.944641365, 0, 0.000112636356140397, 1.3129221319794e-05, -167881.944419501, 2.60480177488464e-05, 0, 0, -7.88496395284941e-05, 7.28119285276509e-05, 5.02370709727023e-05, 7.64980077626127e-05, 1.97432680971079e-05, 7.92554966167894e-05, 0.000117261742956477, 0.00011188523585577, 0, 0, 0, 0, 0, 0, 1284722.22217331, 0, 5.96204500358325e-05, 0, 0, 0, 486111.111554349, 0, -0.000272544136751413, -1.6605626662639e-05, 486111.111070798, -2.82484573601333e-05, 0, 0, 0, 0.00072917921481704, 0, -0.000243967781634059, -0.000169292304932434, -6.70875606031603e-05, -0.000191824772343317, 7.4746600620502e-05, -0.000186213772883036, -0.000407660720519408, -0.000285168613727604, -0.000256925954344892, -0.00014553747942257, -0.000449422387742655, -0.000336507718731844, -0.000483218147791759, 0, 0, 0, 0, 0, 0, -1493055.55548947, 0, -7.77356031032574e-05, 0, 0, 0, -798611.111557904, 0, 0.000273044642400226, 1.54735533525528e-05, -798611.111076742, -6.70354211510371e-06, 0, 0, 0, -0.00163643603919349, 0, 0.000812260025718082, 0.000164778684778721, 2.85548217247574e-05, 0.000228581691916106, -0.000232131547258949, 0.000232678797711779, 0.000624486059994988, 0, 0, -0.000571467691015603, 0.00036091980412856, 0.000244838709340609, 0.000435220801003442, -0.000227336083802992, 0.000439688841860062, 0.000709118374607752, 0.000707091213857289, 0.00027115140903711, 0.000816734279088128, 0.000927616277662176, 0.00119876462539784, 0.000573695259256605, 0, 0, 0, 0, 0, 0, 942460.317417856, 0, 5.08897733566967e-05, 0, 0, 0, 694444.444645737, 0, -0.000104710775079381, -5.85481060028291e-06, 694444.444429083, 3.06175691877597e-05, 0, 0, 0, 0.00136824392624723, 0, -0.000754770455851894, -7.97577301147551e-05, -9.39070815493519e-07, -0.000160956413854208, 0.000228292693938939, -8.92060568076704e-05, -0.000463187805537486, 0, 0, 0, 0.00145167835704057, 0, -0.000319355249076298, -0.000286023512796534, -8.55944366343885e-05, -0.000323763637291206, 0.000576129796452474, -0.000392752062436009, -0.000992555368358038, -0.000420678184139782, -0.000163400545465404, -0.000786672967617766, -0.000388989902757445, -0.000385526897132586, -0.000181114432808771, -0.00115255183362683, -0.00131624672468509, -0.000987773197616264, 0.000292231500126656, -0.00119775385887336, -0.000719967831285608, -0.00164492225313481, 0, 0, 0, 0, 0, 0, -248015.873005377, 0, -1.31481531177348e-05, 0, 0, 0, -248015.873043857, 0, 1.1912808244846e-05, 1.68744683601735e-06, -248015.873012397, -1.44084961218933e-05, 0, 0, 0, -0.000403334751506524, 0, 0.000227674294060543, 2.17801607275505e-05, -1.01682622475272e-05, 4.62312227459505e-05, -7.12024340781481e-05, 3.09800290624358e-06, 0.000111379013669247, 0, 0, 0, -0.000748195867279105, 0, 0.00034589447587619, 7.8977062171862e-05, 7.02408138511322e-06, 0.000102875121322236, -0.00028738675594369, 0.000115687612179153, 0.000408397278433236, 7.04246052353655e-05, 2.83857725830871e-06, 0.000274563150954413, 0, 0, -0.000275950476048518, 0.000127347873054141, 0.000100476107451081, 0.000118742858197408, -0.000177889790436519, 0.000195541812430671, 0.000302603028156101, 0.000456017547682746, 0.000343114980658307, 0.000325485910337689, -0.000318319334596434, 0.000486323365658756, 0.000468669707739605, 0.000580976515187459, 0.000627155888184477, -3.40102950135487e-05, 0.000765518553945726, 0.000407445201242062, 0.000521192678149138}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -124.999999977151, 0, 0, 0, 0, 0, 0, 0, 124.999999466898, 0, 3366.07142820584, 124.999999986656, 0, 0, 0, 0, 0, 0, 3.7573196284797e-06, 0, -35810.5158710624, 0, 0, -3241.07141984299, -2.92948733099794e-07, -4.67708124747789e-07, -3241.07142790019, 0, 0, 0, 0, 0, 0, -1.35776208504425e-05, 0, 200451.388884847, 0, 0, 0, -5.41785941748646e-05, 0, 32569.4443931288, -3.39626131382155e-07, 6.37344879925391e-06, 32569.4444348421, 8.35368062719243e-06, 6.75442454180681e-06, -1.15793612856912e-06, 0, 0, 0, 0, 0, 0, 2.88219108398554e-05, 0, -653993.055555149, 0, 0, 0, 0.000165221565292548, 0, -167881.944297138, 9.37846485887954e-06, -1.98899956105333e-05, -167881.944397467, 0, 0, 0.000273546496074882, -1.59480779953209e-05, -3.26199284303098e-05, -4.95262363322486e-05, -6.16273323367944e-05, -7.63429610528723e-06, 7.80141718352813e-05, -0.000101740190792903, 0, 0, 0, 0, 0, 0, -3.78440960618e-05, 0, 1284722.22223498, 0, 0, 0, -0.00027829947530046, 0, 486111.110876407, -3.12614275591804e-05, 2.96603096320447e-05, 486111.111012542, 0, 0, 0, -0.000685150007205351, 0, -0.000638307025263514, -1.23007556743916e-05, 6.15244646733796e-05, 9.06140095538775e-05, 0.000181907456997499, -5.95023154926572e-05, -0.000417855500033683, 9.96403132142916e-05, 9.34144981537891e-05, 5.78597412764291e-05, 0.000335155054519719, 0.000587886119799848, 0.000175492747260693, 0, 0, 0, 0, 0, 0, 3.0849411397492e-05, 0, -1493055.55557846, 0, 0, 0, 0.000262421452561669, 0, -798611.11089767, 4.89495061733752e-05, -2.70469525829009e-05, -798611.111011741, 0, 0, 0, 0.00089352349017719, 0, 0.000771426864429877, 8.5613834794179e-05, -4.30109013394255e-05, -5.54530117622684e-05, -0.000216243492079505, 0.000132413302435182, 0.000842869453384685, 0, 0, 0.00120391590183112, -6.75882020476791e-05, -0.000210280375496306, -0.000142622649193064, -0.000410061082229917, -2.87236557965171e-05, 0.000303355301428121, -0.000383622688952538, -0.0011265294352592, 0.000325948349645526, -0.00118336037480795, -0.00050392490325874, -0.0014068833258525, 0, 0, 0, 0, 0, 0, -1.45076215899908e-05, 0, 942460.317477171, 0, 0, 0, -0.000130256687769315, 0, 694444.444341724, -3.41386356424892e-05, 1.47286097853268e-05, 694444.444402007, 0, 0, 0, -0.000578737935666412, 0, -0.000481793746356081, -9.39625758165301e-05, -2.88105756915641e-06, -1.01763772698522e-05, 0.000118047374957331, -8.74893160876562e-05, -0.000744180372998767, 0, 0, 0, -0.00106426460020597, 0, -0.000956350001104582, -3.36049936628419e-05, 0.000178478254654465, 6.16197709952669e-05, 0.000355182144226917, -4.75351745780472e-05, -0.000669093975692856, 0.000168972520670015, 0.000838457215113102, -0.000854380537868847, 0.000100401767207399, 0.000146051540528385, 0.000243993656555256, 0.000860852802842066, 0.000647279255404821, 0.00141559935532605, 0.00218037757573226, 8.53349214229759e-05, 0.000741737115694451, 0.00191001573717344, 0, 0, 0, 0, 0, 0, 3.02713090702949e-06, 0, -248015.873020523, 0, 0, 0, 2.68358679436924e-05, 0, -248015.872995954, 9.05580200435946e-06, -3.59149728400247e-06, -248015.87301108, 0, 0, 0, 0.000146686670982246, 0, 0.000122729052841504, 3.36596855300423e-05, 8.2077749712536e-06, 1.12860987380326e-05, -2.98584357460624e-05, 1.57584172159963e-05, 0.000243912447841791, 0, 0, 0, 0.000355604435784568, 0, 0.000290637828704419, 4.27103653743602e-05, -4.74841787243812e-05, 1.47379014172044e-05, -8.59801240022275e-05, 2.16249248537189e-05, 0.000379081610217177, 4.41500640675333e-07, -0.000203940832718735, 0.00043171306391353, 0, 0, 0.000419664016909909, -1.34991195574415e-05, -8.34227140759598e-05, -2.35367966335677e-05, -0.00015757341635255, -2.57446900952064e-05, 1.90365684111553e-05, -0.00022362622808021, -0.000271110789996669, -0.000564152851045483, -0.000900266498404822, 0.000261229723877587, 0.000134332682540047, -0.000575789851975322, -0.000182092195246246, -0.000980331741497354, -0.000766255031148782, -0.00147695396625429, -0.000630061760173617}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -6000.00001095717, 0, 0, 0, 0, 0, 0, 0, 125738.095426544, 0, 65738.0953949987, 65738.0954301977, 0, 0, 0, 0, 0, 0, -960908.731796466, 0, -303527.778813062, 0, 0, -1264436.50980094, -303527.778755794, -1264436.51127979, -607055.557897016, 0, 0, 0, 0, 0, 0, 3801388.89611971, 0, 766111.115187282, 0, 0, 0, 8368888.90330415, 0, 5333611.11839078, 766111.113537444, 8368888.91197802, 2298333.34686074, 5333611.12886547, 2298333.34107776, 10667222.253464, 0, 0, 0, 0, 0, 0, -8803472.24038457, 0, -1142361.12148854, 0, 0, 0, -27552777.8316076, 0, -12230555.568654, -1142361.11434554, -27552777.8561469, -4569444.48824348, 0, 0, -29837500.0510301, -12230555.6020829, -4569444.45440893, -29837500.1093881, -59675000.1788782, -6854166.70131579, -36691666.7981066, -36691666.7753792, 0, 0, 0, 0, 0, 0, 12430555.5828538, 0, 1006944.46164563, 0, 0, 0, 50729166.7736136, 0, 16458333.3422144, 1006944.44591094, 50729166.8137414, 5034722.30712302, 0, 0, 0, 78611111.2723754, 0, 55763888.9755044, 16458333.404121, 5034722.22825132, 78611111.4206714, 157222222.69564, 10069444.5355316, 65833333.6149463, 55763889.1353623, 10069444.4712761, 167291667.274696, 65833333.4886275, 167291667.215877, 98750000.3408158, 0, 0, 0, 0, 0, 0, -10555555.5799134, 0, -486111.12890453, 0, 0, 0, -53263889.0065017, 0, -12986111.1081067, -486111.108879567, -53263889.0453238, -2916666.765466, 0, 0, 0, -107986111.355457, 0, -57638888.9555249, -12986111.1770679, -2916666.66998616, -107986111.558184, -215972222.862999, -7291666.80682908, -64930555.8733238, 0, 0, -110416666.897383, -57638889.1863601, -7291666.67178152, -110416667.214177, -331250001.203658, -9722222.27040341, -230555556.551184, -64930555.6521205, -331250001.157093, -129861111.666203, -230555556.264962, -129861111.415818, -345833334.631702, 0, 0, 0, 0, 0, 0, 4960317.47223535, 0, 99206.3595230584, 0, 0, 0, 29861111.1790151, 0, 5555555.54912718, 99206.3457193328, 29861111.1996478, 694444.507997846, 0, 0, 0, 75000000.1783099, 0, 31250000.014192, 5555555.59098366, 694444.446627346, 75000000.3204316, 150000000.434597, 2083333.44994242, 33333333.513768, 0, 0, 0, 100694444.687643, 0, 76388889.039088, 31250000.1893403, 2083333.33876101, 100694444.988185, 302083334.400042, 3472222.27587724, 156250000.757968, 33333333.3500943, 302083334.425239, 83333333.7721399, 76388889.3403992, 3472222.21913468, 305555556.892531, 156250000.410319, 83333333.407292, 305555556.652238, 458333335.124447, 111111111.422772, 312500001.420558, 312500000.970107, 0, 0, 0, 0, 0, 0, -992063.494527886, 0, -0.00253706099648136, 0, 0, 0, -6944444.46053086, 0, -992063.489982546, 0.00142823305833658, -6944444.46520873, -0.0172531661524382, 0, 0, 0, -20833333.3840823, 0, -6944444.44039157, -992063.500351523, -0.000488770930018626, -20833333.4238279, -41666666.7840304, -0.0398324986698768, -6944444.48586597, 0, 0, 0, -34722222.3121619, 0, -20833333.366901, -6944444.49623417, -0.00530134546366605, -34722222.4210463, -104166667.018754, -0.0268107694497737, -41666666.8802595, -6944444.44115506, -104166667.049764, -20833333.4684567, 0, 0, -34722222.3083468, -20833333.4772614, 0.000994403852102703, -34722222.4506769, -138888889.481583, 0.00280296615919772, -104166667.192006, -41666666.7525331, -20833333.3181763, -138888889.442635, -208333334.154584, -34722222.3423948, -104166667.226136, -104166667.030364, -34722222.2542164, -208333334.309033, -104166666.890487, -208333334.06673, -138888889.402278}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 7000.00000957447, 0, 0, 0, 0, 0, 0, 0, -164194.444612964, 0, -59194.4445706885, -59194.4446121001, 0, 0, 0, 0, 0, 0, 1444046.29785285, 0, 206129.63037172, 0, 0, 1300175.92740017, 206129.630295582, 1300175.92899937, 412259.261241913, 0, 0, 0, 0, 0, 0, -6430138.89585995, 0, -378472.225041349, 0, 0, 0, -10279027.7913965, 0, -4227361.11441821, -378472.223110001, -10279027.8001107, -1135416.67868926, -4227361.12598215, -1135416.67069656, -8454722.24884664, 0, 0, 0, 0, 0, 0, 16370138.9060307, 0, 386574.081693368, 0, 0, 0, 39190509.3149901, 0, 7223379.6257531, 386574.073108799, 39190509.3372794, 1546296.33741266, 0, 0, 29657175.9647721, 7223379.66519045, 1546296.29176646, 29657176.0274064, 59314352.0135892, 2319444.46569928, 21670139.0003843, 21670138.9603818, 0, 0, 0, 0, 0, 0, -24930555.5801845, 0, -208333.347225418, 0, 0, 0, -81006944.5592604, 0, -6840277.75157278, -208333.32761421, -81006944.5934779, -1041666.75030582, 0, 0, 0, -93645833.4870594, 0, -44201388.9221687, -6840277.82767055, -1041666.64438698, -93645833.6361294, -187291667.107019, -2083333.40310834, -27361111.3477671, -44201389.1116368, -2083333.31630907, -132604167.215868, -27361111.1511453, -132604167.09994, -41041666.8866304, 0, 0, 0, 0, 0, 0, 22500000.0206632, 0, 46296.3119556622, 0, 0, 0, 93217592.7190505, 0, 3402777.73714172, 46296.2863420918, 93217592.7530116, 277777.879352807, 0, 0, 0, 147916666.926203, 0, 35648148.1199742, 3402777.8216862, 277777.749541481, 147916667.119152, 295833333.937, 694444.569889707, 17013889.1491171, 0, 0, 109490740.914854, 35648148.4141075, 694444.388721334, 109490741.274396, 328472223.34441, 925925.91680704, 142592593.48676, 17013888.8226346, 328472223.20377, 34027778.1635694, 142592593.046548, 34027777.7785707, 213888889.911639, 0, 0, 0, 0, 0, 0, -11111111.1205118, 0, -0.00966232855058285, 0, 0, 0, -56250000.0714379, 0, -694444.417860114, 0.00811690157081935, -56250000.0914732, -0.0676750088893668, 0, 0, 0, -114583333.532937, 0, -14583333.2729231, -694444.467122865, 0.0179336173054224, -114583333.666207, -229166667.078812, -0.115600183696011, -4166666.80885808, 0, 0, 0, -118055555.784915, 0, -62500000.0652596, -14583333.500942, 0.0345660210593918, -118055556.10881, -354166667.674183, -0.0206693112443867, -72916667.3283779, -4166666.57407319, -354166667.622605, -10416666.9908423, -62500000.4403352, 0.0546847451576666, -250000001.250202, -72916666.8471204, -10416666.4367382, -250000000.85841, -375000001.514353, -13888888.8976834, -145833334.514235, -145833333.84055, 0, 0, 0, 0, 0, 0, 2314814.81660799, 0, 0.00247462301126509, 0, 0, 0, 13888888.9051569, 0, -0.00638494352633606, -0.00260397633198843, 13888888.9104128, 0.0188416763799573, 0, 0, 0, 34722222.2802428, 0, 2314814.78907744, 0.00510371872548765, -0.00516748387968783, 34722222.3183803, 69444444.5561443, 0.0420583245208866, 0.0317168443500801, 0, 0, 0, 46296296.3912471, 0, 13888888.884725, 2314814.86084529, -0.00574012863095127, 46296296.5046326, 138888889.222713, 0.020978312319007, 13888889.0659431, -0.0323932381098911, 138888889.230623, 0.104308021183756, 0, 0, 34722222.2869253, 13888889.0287432, -0.0159948571877727, 34722222.4558371, 138888889.453801, -0.0176526379325699, 69444444.9281715, 13888888.901008, -0.119755750426343, 138888889.359347, 208333334.042481, 0.0157988384634758, 34722222.7043145, 69444444.7106725, -0.125218991694889, 138888889.754546, 34722222.215537, 138888889.400963, 46296296.5848777}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -6300.00000352693, 0, 0, 0, 0, 0, 0, 0, 156175.000066536, 0, 40675.0000322821, 40675.0000693126, 0, 0, 0, 0, 0, 0, -1479875.00075202, 0, -104166.666755984, 0, 0, -954041.666957168, -104166.666796822, -954041.668105618, -208333.334072589, 0, 0, 0, 0, 0, 0, 7159513.89251802, 0, 132291.666922575, 0, 0, 0, 8283819.45031163, 0, 2306597.21959537, 132291.666006684, 8283819.45594702, 396875.004708729, 2306597.22813127, 396874.999628807, 4613194.45508919, 0, 0, 0, 0, 0, 0, -19704513.897918, 0, -83333.3347294566, 0, 0, 0, -35221875.0299101, 0, -2758680.53583877, -83333.3295688998, -35221875.0421897, -333333.350610208, 0, 0, -18192708.3399798, -2758680.566246, -333333.321616157, -18192708.3820707, -36385416.7415802, -500000.001346533, -8276041.70876373, -8276041.68153803, 0, 0, 0, 0, 0, 0, 32138888.9015377, 0, 20833.3376014462, 0, 0, 0, 80625000.0671679, 0, 1631944.39436005, 20833.3244530949, 80625000.0833513, 104166.704320907, 0, 0, 0, 66423611.1794441, 0, 19548611.0756202, 1631944.45349895, 104166.63709494, 66423611.2678325, 132847222.437735, 208333.350600796, 6527777.85772386, 19548611.2112169, 208333.29519338, 58645833.5848913, 6527777.72496685, 58645833.4958147, 9791666.70808246, 0, 0, 0, 0, 0, 0, -30763888.8989369, 0, -0.00642237020486755, 0, 0, 0, -101388888.965328, 0, -381944.383150656, 0.0118213364469563, -101388888.981251, -0.0491325524226976, 0, 0, 0, -119965277.92214, 0, -10243055.4505931, -381944.447105782, 0.0344825765027981, -119965278.02273, -239930555.859261, -0.0459853836539048, -1909722.29173311, 0, 0, -59201388.9233374, -10243055.6703704, 0.0656550195708819, -59201389.1594079, -177604167.215614, 0.0493834508834121, -40972222.6196278, -1909722.07654797, -177604167.08887, -3819444.53290491, -40972222.3019606, -3819444.26038413, -61458333.7344275, 0, 0, 0, 0, 0, 0, 15972222.226436, 0, 0.00466708210252514, 0, 0, 0, 65972222.2656117, 0, -0.036373141242125, -0.00830574002449188, 65972222.2762457, 0.0349894499701276, 0, 0, 0, 104166666.788556, 0, 2083333.23087347, -0.00133603081590169, -0.0221126939410769, 104166666.853107, 208333333.545502, 0.0524833946652793, 0.0223235110423219, 0, 0, 0, 76388888.9920371, 0, 24305555.5122078, 2083333.40171725, -0.0386226115584328, 76388889.1858792, 229166667.171237, -0.0222050665662678, 10416666.9352453, -0.128012814519627, 229166667.097622, 0.0772180479688901, 24305555.777646, -0.0643055624426503, 97222222.8339422, 10416666.6032809, -0.308879701849547, 97222222.5247655, 145833334.012612, -0.186318157099678, 20833333.8450378, 20833333.3431607, 0, 0, 0, 0, 0, 0, -3472222.22294107, 0, -0.00131714842891543, 0, 0, 0, -17361111.1208319, 0, 0.00845935974543295, 0.00241034656885311, -17361111.1240985, -0.0102712320692349, 0, 0, 0, -34722222.2593686, 0, 0.034527886062051, 0.000962373752058857, 0.00635712284868492, -34722222.2777565, -69444444.503278, -0.0216105092317392, -0.000949523007232268, 0, 0, 0, -34722222.2751472, 0, -3472222.1864614, -0.0183284729656148, 0.00789372890344713, -34722222.3381462, -104166666.835557, -0.00248339415649728, -0.0592989629737327, 0.0397595648148453, -104166666.825317, -0.023377471740178, 0, 0, -17361111.1256444, -3472222.29227856, 0.0166028490123648, -17361111.2372921, -69444444.7299968, 0.0229512977976519, -17361111.340824, 0.0497731559518174, 0.128041891482996, -69444444.6393134, -104166666.993227, 0.0646117208548638, -0.215214390677933, -17361111.1822439, 0.162700660699955, -34722222.6462771, 0.141488568155378, -34722222.3756518, -0.0351840566167791}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 4199.99999979809, 0, 0, 0, 0, 0, 0, 0, -107616.666665725, 0, -20116.6666516061, -20116.6666717441, 0, 0, 0, 0, 0, 0, 1067513.88902153, 0, 35916.6664874707, 0, 0, 490930.55526618, 35916.6665725599, 490930.555802624, 71833.3333137131, 0, 0, 0, 0, 0, 0, -5460277.77865355, 0, -28333.3326071461, 0, 0, 0, -4515208.33387064, 0, -833263.884965658, -28333.332537399, -4515208.33605902, -84999.9999586848, -833263.889362892, -84999.9983768827, -1666527.77835475, 0, 0, 0, 0, 0, 0, 15968402.7800803, 0, 8333.33213217781, 0, 0, 0, 20661805.5638539, 0, 623611.092906927, 8333.33058464676, 20661805.5666616, 33333.3340590111, 0, 0, 7058680.54699255, 623611.109568116, 33333.3248005376, 7058680.56586811, 14117361.1247227, 49999.9935607769, 1870833.33227386, 1870833.3247526, 0, 0, 0, 0, 0, 0, -27652777.7808302, 0, 0.000461238291654596, 0, 0, 0, -51402777.8007542, 0, -173611.071422631, 0.0054481944009997, -51402777.8009464, -0.00373366386150077, 0, 0, 0, -28402777.7867704, 0, -4826388.84012367, -173611.103861265, 0.0177830559604914, -28402777.8170373, -56805555.6034025, 0.0106716001019338, -694444.430008003, -4826388.9065977, 0.0291692959498465, -14479166.7075618, -694444.389449566, -14479166.6771002, -1041666.63250656, 0, 0, 0, 0, 0, 0, 27986111.1131757, 0, 0.00105186307126136, 0, 0, 0, 70208333.3613667, 0, -0.0450113331168655, -0.00657481187538742, 70208333.3600983, 0.00794557291565597, 0, 0, 0, 57673611.1565169, 0, 1215277.68352687, -0.0109109012177906, -0.0194573530493388, 57673611.1771627, 115347222.295094, -0.00577055836617526, -0.0371667524371887, 0, 0, 16666666.6339332, 1215277.79576122, -0.0399953350182017, 16666666.7338771, 50000000.1196992, -0.0454559581859491, 4861111.15883605, -0.101456327242237, 50000000.069014, -0.0546603852011243, 4861111.05088639, -0.155981001256849, 7291666.70428809, 0, 0, 0, 0, 0, 0, -15277777.7784008, 0, -0.0013264392323125, 0, 0, 0, -49305555.5717452, 0, 0.0255755138730318, 0.00427748842469831, -49305555.5717401, -0.00765799467344888, 0, 0, 0, -56250000.0463731, 0, 0.0789208203785598, 0.00794719083442164, 0.0121858754886339, -56250000.0524427, -112500000.053774, -0.00478925219377606, 0.037832999493556, 0, 0, 0, -25694444.460235, 0, -3472222.15367461, -0.00759780180597804, 0.0219837045696445, -25694444.5263612, -77083333.4534733, 0.0324201343581036, -0.00553010983071486, 0.0791740144497046, -77083333.4144391, 0.0504620751003796, -3472222.27840891, 0.0424802050856158, -13888889.0151868, 0.103101747363087, 0.191675548223305, -13888888.8972935, -20833333.4690715, 0.168612296851853, -0.0735260814400375, 0.131389829914192, 0, 0, 0, 0, 0, 0, 3472222.2222667, 0, 0.000455679108064236, 0, 0, 0, 13888888.8924443, 0, -0.00583658545687693, -0.00117258116200128, 13888888.8928783, 0.00270084582175326, 0, 0, 0, 20833333.3486907, 0, -0.0244518193882446, -0.00236677196343292, -0.0033927752951587, 20833333.3495047, 41666666.6822931, 0.00435778657197259, -0.0128894031710454, 0, 0, 0, 13888888.9059928, 0, -0.0341432732588329, 0.00162290833553203, -0.00458550942977855, 13888888.9224188, 41666666.7084404, -0.00693805117970374, -0.0121143838313675, -0.0233252156037601, 41666666.6985158, -0.0195556951474449, 0, 0, 3472222.21217473, 0.0177269353471072, -0.0102987956511551, 3472222.25872932, 13888888.9559807, -0.0170773245868154, 0.0398289788855364, -0.0455512704038854, -0.0713820590272699, 13888888.9162469, 20833333.4044144, -0.066540609639543, 0.0297778703757842, -0.0165141946887051, -0.102047996392763, 0.102717313876194, -0.112201823752439, -0.0118297069233474, -0.0531447595807983}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -1999.99999945706, 0, 0, 0, 0, 0, 0, 0, 52388.8888791655, 0, 6722.22221107994, 6722.22221601873, 0, 0, 0, 0, 0, 0, -535990.74071541, 0, -7499.99989969629, 0, 0, -168490.740545705, -7499.99993203998, -168490.740687696, -14999.9998948387, 0, 0, 0, 0, 0, 0, 2851111.11112964, 0, 2777.77739914829, 0, 0, 0, 1611249.99964256, 0, 179583.331443554, 2777.77743848788, 1611250.0000006, 8333.33268947058, 179583.332840754, 8333.33246982281, 359166.665617429, 0, 0, 0, 0, 0, 0, -8731250.00010159, 0, 0.000638803196606159, 0, 0, 0, -7764351.85329245, 0, -63425.9179581622, 0.000958141834897599, -7764351.85242387, 0.00185195319452245, 0, 0, -1596527.77223136, -63425.9237612192, 0.00326441354097206, -1596527.77726954, -3193055.55384936, 0.0039805427481549, -190277.771399071, -190277.771333979, 0, 0, 0, 0, 0, 0, 15902777.7777704, 0, -0.0003745246337801, 0, 0, 0, 20590277.7835734, 0, -0.0164035084861251, -0.00167524010583545, 20590277.7794186, -0.00270037740367328, 0, 0, 0, 6874999.99828293, 0, 520833.309361141, -0.00510094552189503, -0.0059516056650092, 6875000.00137134, 13750000.0003309, -0.00886791151500708, -0.0202167618038366, 520833.33039948, -0.0122094573731684, 1562499.9904521, -0.0238156755575144, 1562499.98682572, -0.0251551094362735, 0, 0, 0, 0, 0, 0, -16944444.4441723, 0, -0.000256632365432107, 0, 0, 0, -30254629.6372131, 0, 0.017724083513707, 0.00186498106457838, -30254629.6318633, 0.00159470436062146, 0, 0, 0, -15277777.7898296, 0, 0.042322223900395, 0.00611421284210584, 0.00591227137932984, -15277777.7820007, -30555555.5593088, 0.0100181479557493, 0.0326926005688689, 0, 0, -1967592.57067865, 0.00438654773663065, 0.0146674595334021, -1967592.59524454, -5902777.77676581, 0.020913691313266, 0.0281832847240974, 0.0381959020673791, -5902777.76722019, 0.0452263741041547, 0.0414310518323172, 0.0654167813326556, 0.0276214306938731, 0, 0, 0, 0, 0, 0, 9722222.22193807, 0, 0.000437691464505566, 0, 0, 0, 22916666.6710195, 0, -0.00958764582047232, -0.00112092741448897, 22916666.6681594, 9.7395159564683e-05, 0, 0, 0, 16666666.6815744, 0, -0.0335191271817328, -0.0039076999944281, -0.00343115681081705, 16666666.6705513, 33333333.3374746, -0.00434234223484969, -0.0260574177837784, 0, 0, 0, 3472222.2218488, 0, -0.0348326632718002, -0.00436585924380687, -0.0074487772030716, 3472222.22921149, 10416666.6733167, -0.0170295938759144, -0.0372190170018304, -0.0276684426826629, 10416666.6611146, -0.0422534586986114, 0.00310389935621751, -0.0172972896579769, -0.00903674835868702, -0.0492529212922558, -0.070162431399944, -0.029978186238765, -0.00196229070571554, -0.0742240168536056, -0.0261680776600021, -0.0739107068511419, 0, 0, 0, 0, 0, 0, -2314814.81472467, 0, -0.000157360518881511, 0, 0, 0, -6944444.4453384, 0, 0.00209943545460226, 0.00029351490906549, -6944444.4448316, -0.000302641958249726, 0, 0, 0, -6944444.44969131, 0, 0.00992593013483157, 0.00108603107722187, 0.000851107408096302, -6944444.445751, -13888888.890353, 0.000307860495535979, 0.00774612097272177, 0, 0, 0, -2314814.81992285, 0, 0.0159350659605082, 0.00136867666740145, 0.00146550522072031, -2314814.81816159, -6944444.44797415, 0.00472498978986349, 0.0167421747959709, 0.00758840407726233, -6944444.44324182, 0.0154953949405896, 0, 0, 0.00701307422532969, -0.00106788205520589, 0.00416373752242098, -0.00427852169069155, -0.00184437100044407, 0.00756918110974734, 0.00939906401910287, 0.0195621088585826, 0.0239500401131418, 0.00746490559436714, -0.00322927477716853, 0.0311967979794343, 0.0137804065675109, 0.0175852039843263, 0.0380973143632917, -0.00513516872746089, 0.0446095291340687, 0.0222923658211317, 0.0344195108968801}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 642.857142728709, 0, 0, 0, 0, 0, 0, 0, -17107.1428542813, 0, -1357.14285546507, -1357.14285509771, 0, 0, 0, 0, 0, 0, 178964.285701286, 0, 714.285699306097, 0, 0, 34678.5713925054, 714.285697116468, 34678.5714020269, 1428.57139986629, 0, 0, 0, 0, 0, 0, -979861.111078677, 0, 4.92439198728377e-05, 0, 0, 0, -341111.1110756, 0, -17499.9996375666, 6.94513374023347e-05, -341111.111000128, 0.00018197292904745, -17499.9998208263, 0.000197018867454764, -34999.999680927, 0, 0, 0, 0, 0, 0, 3109027.77768837, 0, -3.7091373959137e-05, 0, 0, 0, 1707638.8894061, 0, -0.00161642985080147, -0.000166442647715614, 1707638.88860709, -0.000541923300749198, 0, 0, 161111.110150459, -0.00063409019277921, -0.000647350314403627, 161111.1105045, 322222.221361553, -0.000988246491282719, -0.00183042060272262, -0.00155508510284074, 0, 0, 0, 0, 0, 0, -5902777.77756926, 0, -0.0001045502645613, 0, 0, 0, -4756944.44612764, 0, 0.00330930831834695, 0.00025031527565492, -4756944.44397066, 0.000842480624390552, 0, 0, 0, -729166.66746517, 0, 0.00500566003606298, 0.00137301980264295, 0.00102244066168009, -729166.665348168, -1458333.33218705, 0.00238339050214186, 0.00558662576391911, 0.00187066651241445, 0.00275021403194777, 0.00397960778031145, 0.00517125831177346, 0.00410877019937426, 0.00638907621198233, 0, 0, 0, 0, 0, 0, 6587301.5870207, 0, 0.000250761075383858, 0, 0, 0, 7430555.55754877, 0, -0.00341188057242825, -0.000252399645459126, 7430555.55509232, -0.000604712486098089, 0, 0, 0, 1736111.11617546, 0, -0.00931971305352977, -0.00147066409711975, -0.00084038014184988, 1736111.10950718, 3472222.22152051, -0.00297481726523971, -0.0084942375674965, 0, 0, -0.00338102879374296, -0.002394430434865, -0.00295315630713121, -0.00236293618671422, -0.00330957016534318, -0.00495426832880303, -0.0103634745848576, -0.00785509847931842, -0.00479366237277699, -0.0119039292390327, -0.0101842890609628, -0.0142975564056401, -0.00912495130388569, 0, 0, 0, 0, 0, 0, -3968253.96806854, 0, -0.000203622474360628, 0, 0, 0, -6051587.30261465, 0, 0.00168005373088289, 0.000132430066961911, -6051587.30133945, 0.000124018881309407, 0, 0, 0, -2083333.33855748, 0, 0.00742163189781422, 0.000836407564019593, 0.000403367662299546, -2083333.33218072, -4166666.6665625, 0.00151797050463772, 0.00637417498562388, 0, 0, 0, -0.00239807740148735, 0, 0.00738294566625626, 0.0019070778100385, 0.00130365638180789, 0.00156718568018858, 0.000539119167903543, 0.00431057016660896, 0.0120446028879554, 0.00529239747927661, 0.00375395928372002, 0.0112987046182337, 0.00180811462637807, 0.00395124693511121, 0.00673599439446157, 0.0114727814386858, 0.0145959193638995, 0.00933019475400325, 0.00326595628091218, 0.0165474073211483, 0.0108447623275039, 0.0179000637351565, 0, 0, 0, 0, 0, 0, 992063.492016108, 0, 5.87389540986545e-05, 0, 0, 0, 1984126.98430283, 0, -0.000328429309564011, -3.35538955006656e-05, 1984126.9840678, 3.00912387091374e-05, 0, 0, 0, 992063.493773213, 0, -0.00215104179269444, -0.000223968998350842, -5.81186691186614e-05, 992063.491726541, 1984126.984155, -0.000240243347775135, -0.00176192666985153, 0, 0, 0, 0.0023246202111477, 0, -0.00368482697509791, -0.00054494221356553, -0.000207218918350585, -0.000481363813454606, 0.000321741127215579, -0.00129855464221081, -0.00483856865743563, -0.00127459455591874, -0.00115982065932158, -0.00408066062663053, 0, 0, -0.000924485030727623, -0.000558405423892875, -0.00094996065359717, -0.000337531998642722, -0.00120186927086626, -0.00186672277068681, -0.00445872773844281, -0.00447720635444162, -0.00452262811340237, -0.00312452860582955, -0.000209760380343437, -0.0071878134913543, -0.00583099639570816, -0.00499808614049834, -0.00794156165795424, -0.00177280925551485, -0.00929801089119216, -0.00566028457258476, -0.00830522518960797}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -71.4285714506127, 0, 0, 0, 0, 0, 0, 0, 1821.42857154591, 0, 785.714286091898, 71.428571476805, 0, 0, 0, 0, 0, 0, -17861.1111129952, 0, -714.285716844187, 0, 0, -19964.2857131448, 1.89467376393818e-06, -1750.00000076371, -714.285716385186, 0, 0, 0, 0, 0, 0, 89027.7777897516, 0, 1.2618318631562e-05, 0, 0, 0, 194722.222230746, 0, 17499.9999652596, -2.02260001445632e-05, 16111.1111215352, 1.74705743540223e-05, -1.26733826381807e-05, -3.55418566313478e-06, 17500.0000164048, 0, 0, 0, 0, 0, 0, -246527.777810981, 0, -4.16234453025929e-05, 0, 0, 0, -963194.44455538, 0, 0.000165623469742002, 8.01087789092546e-05, -72916.6667008562, -6.4518088242616e-05, 0, 0, -161111.110967667, 0.000133386616123745, 0.000117092249545788, 2.52870354926396e-06, -161111.111269517, -8.13807486802545e-05, -6.34682570762134e-05, 3.04730821217062e-05, 0, 0, 0, 0, 0, 0, 381944.444490741, 0, 8.09258094964967e-05, 0, 0, 0, 2638888.88922438, 0, -0.000325225439690985, -0.000153660470850795, 173611.111162233, 0.000126945023389705, 0, 0, 0, 729166.66662373, 0, -0.000747665103007525, -0.00042552125703538, -0.000449396906511331, 5.59329926447659e-05, 729166.667068054, 0.000378006723564528, -2.72867733248601e-05, -0.000215196742325732, -5.75920513206808e-05, 0.000624241245272001, -0.000823196243963995, 0.000217951978708954, 0.000159411149368671, 0, 0, 0, 0, 0, 0, -307539.682572683, 0, -8.86679864419685e-05, 0, 0, 0, -4027777.77821801, 0, 0.000322168681214724, 0.000162799669194122, -208333.333375977, -0.000147942940458185, 0, 0, 0, -1736111.11155111, 0, 0.0012596511531745, 0.000614928140179908, 0.00072065475322648, -7.53991382630283e-05, -1736111.11158263, -0.000672793927956824, 0.000350506627762049, 0, 0, 0.00100023772677809, 0.000612739097178326, 0.000528884928137489, 3.07311648001185e-05, -0.00131772934678091, -0.000457506354896517, -0.000853772363995263, 0.00220035733796181, -0.000563418121365981, -0.000622879836855606, 0.00102346733611252, 0.00110061422791534, -0.00167329658321598, 0, 0, 0, 0, 0, 0, 99206.3492173915, 0, 5.15205694075571e-05, 0, 0, 0, 3174603.17486779, 0, -0.000162104483418569, -8.82091838654855e-05, 99206.3492265663, 0.000100089482879521, 0, 0, 0, 2083333.33394395, 0, -0.000921713408278595, -0.000421829645064821, -0.00054195953213091, 2.36162870492514e-05, 2083333.33360197, 0.000559927722685184, -0.000497520859062668, 0, 0, 0, -0.000353082899376253, 0, -0.00143319418234991, -0.000717929153156209, -0.000674808767791391, 0.000117303088291524, 0.0011124924036271, 0.000785372234783919, 0.000273026521702517, -0.00220242529498655, 0.000227382154733852, 0.000534614530067781, -0.000255355771597479, 8.93965517741413e-05, 0.00146217011851244, -0.00239366997474742, -0.00265069557457126, -0.000377397899259678, 0.00244201389386399, -0.00050095830110922, 0.00283201749235311, -0.000341295207458119, 0, 0, 0, 0, 0, 0, -1.22959974725889e-06, 0, -1.23050813431201e-05, 0, 0, 0, -992063.49212251, 0, 3.37095758160719e-05, 1.97421528163526e-05, -4.29433744506753e-06, -2.86992629192404e-05, 0, 0, 0, -992063.492294893, 0, 0.000249461228662188, 0.000107981394826095, 0.000154880354186066, -6.73492072682855e-07, -992063.492126954, -0.000178745556506755, 0.000204032597795679, 0, 0, 0, -7.94581169116209e-05, 0, 0.000576176662961805, 0.000244463981056965, 0.000286198783650203, -6.02048717681652e-05, -0.00033141149368573, -0.000373076142008811, 9.82521398443566e-05, 0.000754044395802416, 5.67145270093137e-05, -8.04655868837441e-05, 0, 0, 0.000463011557981912, 0.000204557830367909, 9.2716165173603e-05, -3.86843275350739e-05, -0.000622598786934873, -0.000260415888484223, -0.000452392954125518, 0.0013679477846927, 0.00144200028454566, 4.30611464182051e-05, -0.000866702341730028, -2.15773904341987e-05, -0.00121908701602104, 0.000708596092819329, 0.000879939971740436, -0.00201205827198253, 0.00130313492895447, -0.000458880160146588, -0.000492042203276403}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -55.5555556155191, 0, 0, 0, 0, 0, 0, 0, 1324.07407565087, 0, 888.888889349836, 55.5555568474694, 0, 0, 0, 0, 0, 0, -11685.1851991795, 0, -3611.11110916494, 0, 0, -21129.6296465691, -7.28250685636578e-07, -1268.5185433325, -833.333353553697, 0, 0, 0, 0, 0, 0, 49768.5185746721, 0, 2777.77775226789, 0, 0, 0, 185694.444603185, 0, 85231.4815165605, -2.82192550433889e-05, 10416.666842987, 2777.77790792657, 4.59062624441733e-05, 4.55071939462685e-05, 19027.778071291, 0, 0, 0, 0, 0, 0, -108796.296417613, 0, 9.30190541969352e-05, 0, 0, 0, -785879.630225848, 0, -63425.9258340674, 0.000123469296671505, -39351.8524583127, -0.000387663797204596, 0, 0, -740509.259768789, 0.000112411030424752, 9.50997222425614e-05, -0.000412449746600272, -156250.001619564, -0.000490207098241109, -63425.9274078279, -0.00047056793223936, 0, 0, 0, 0, 0, 0, 115740.740887974, 0, -0.000171154342779932, 0, 0, 0, 1701388.89003905, 0, -0.000499222199364434, -0.000249825908618779, 69444.4455952407, 0.000604029627837557, 0, 0, 0, 3078703.70548593, 0, 520833.333911229, -0.000474794320858623, -0.000574038314302912, 0.00150678045844437, 590277.781937678, 0.00156985553670986, 0.00338705295653463, 0.000130301735754593, 0.000748676148594406, 520833.340000257, -0.000882378611165147, 0.00232561571619066, 0.00348216894554582, 0, 0, 0, 0, 0, 0, -46296.2963935834, 0, 0.000171524040647129, 0, 0, 0, -1782407.40860238, 0, 0.000831316298743857, 0.000287085324760596, -46296.297519709, -0.000523301418634971, 0, 0, 0, -6481481.48427398, 0, 7.6893738491076e-05, 0.000689640103256295, 0.000900263913191042, -0.00252640813059266, -1041666.67205672, -0.00220306775820674, -0.00363173855381738, 0, 0, -1967592.59472044, 0.000559443216874185, -0.000339830258973678, -0.00130248693977629, -1967592.60515193, -0.00253387798694225, -0.0121040008638419, 0.00323888708451488, -0.00508766882252509, -0.00805630274950044, 0.00198627977174295, -0.00170039574335067, -0.0105825109314852, 0, 0, 0, 0, 0, 0, 2.99935734793482e-05, 0, -8.97264637887144e-05, 0, 0, 0, 694444.445075464, 0, -0.00062422530320019, -0.000178404986978074, 0.00068224759651224, 0.000257193003252194, 0, 0, 0, 6481481.48354333, 0, -0.000539116994755993, -0.000452247382441176, -0.000667606761907724, 0.00205072610928472, 694444.448003072, 0.0014514357381547, 0.00168077789568663, 0, 0, 0, 3472222.22469587, 0, 0.00100599362568899, -0.000668562906545962, -0.00012905751685714, 0.00211869301384072, 3472222.23277739, 0.00270322751151112, 0.00972039750266166, -0.00301223389625551, 0.00508234580689674, 0.00721535385194144, 0.000335783288817486, 0.00194300996482208, 0.0154196815852206, -0.00565095367924619, -0.0016749080793859, -0.000853648745539949, 0.0128253462420521, 0.00605157727923545, 0.0164843395205528, 0.0029420854992346, 0, 0, 0, 0, 0, 0, -2.52941535683414e-06, 0, 1.88937873641692e-05, 0, 0, 0, -0.000131980476062583, 0, 0.000178677576906415, 4.56761290288911e-05, -0.000157848556546149, -5.64169357499366e-05, 0, 0, 0, -2314814.8153987, 0, 0.000274391117557439, 0.000107057434296444, 0.000216436350451522, -0.000638440723517811, -0.000925459176321777, -0.000382717030121693, -0.000219148741419941, 0, 0, 0, -2314814.81577832, 0, -9.40577751585172e-05, 0.000179882367194314, 0.000127890692289242, -0.00099885613380014, -2314814.81814869, -0.000995347345071789, -0.00276979091570879, 0.0010203384183553, -0.00183813278027213, -0.00220805146754706, 0, 0, -0.000715805372656383, 4.81122955120836e-05, -0.000566113468477376, -0.000592637458569868, -0.00648153305604175, -0.00118455091939823, -0.00649885990161129, 0.00259233484272656, 0.00139316265562489, -0.000310700858095318, -0.00546818978438912, -0.00364026252211079, -0.00780125623961658, 0.00263369937964784, -0.00129646263066946, -0.00996794826938298, 0.00272578497645849, -0.00126450894171584, -0.00468235973660909}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -50.0000000589641, 0, 0, 0, 0, 0, 0, 0, 1091.66666782088, 0, 966.666668133552, 50.0000009077104, 0, 0, 0, 0, 0, 0, -8333.33333980853, 0, -5916.66667250637, 0, 0, -21055.5555886526, -1.70077264580299e-07, -1041.66667869781, -916.666688751642, 0, 0, 0, 0, 0, 0, 28125.0000047699, 0, 13333.3333370923, 0, 0, 0, 160069.444699557, 0, 128263.889034385, -2.14088072053456e-05, 7291.66673002248, 5000.00015402225, 2.19792509491836e-05, 4.80996313684386e-05, 19097.222483598, 0, 0, 0, 0, 0, 0, -41666.6666036368, 0, -8333.33330779898, 0, 0, 0, -536458.334211955, 0, -286111.111346007, 8.06105511349682e-05, -20833.3334221988, -8333.33384893278, 0, 0, -967013.889996432, 9.53171844008629e-05, 0.000114529399865761, -0.000204381739999169, -133680.557077469, -0.000511805911972863, -104166.668022488, -0.000452868237019349, 0, 0, 0, 0, 0, 0, 20833.3331079887, 0, -6.95879558882827e-05, 0, 0, 0, 784722.22377039, 0, 173611.111172467, -0.000134656420552908, 20833.3331683456, 0.000982921285798496, 0, 0, 0, 3194444.4478229, 0, 2118055.55753098, -0.000335139247541569, -0.000570748317124554, 0.000521381864663978, 381944.448168346, 0.0017501223984413, 173611.11398012, -1.39517636610631e-05, 0.000479984278279012, 729166.673321261, -0.00105312329195949, 0.00271777759432443, 0.00285548097998714, 0, 0, 0, 0, 0, 0, 0.000331821396034578, 0, 7.77748378573858e-05, 0, 0, 0, -381944.445883346, 0, 0.000259544747803917, 0.000114783338057442, 0.000597817039013491, -0.00112558549421066, 0, 0, 0, -4548611.11602931, 0, -1215277.77943736, 0.000256039339471445, 0.000911965586311586, -0.000536815515298896, -381944.44869351, -0.00279936408358229, -0.00289963031167102, 0, 0, -6770833.33841061, 0.000439317647811982, 0.000211756001448962, -0.000575435001998554, -2083333.34679874, -0.002454012917195, -1215277.78888037, 0.00359820374732092, -0.00592795296419938, -0.00613735804856271, 0.000912378264959926, -0.000470119518423532, -0.0110450460037299, 0, 0, 0, 0, 0, 0, -0.00023171847032139, 0, -4.06498582019187e-05, 0, 0, 0, 0.000647289242472529, 0, -0.000292480189770378, -5.53888690118541e-05, -0.000606798416676191, 0.000710454445899645, 0, 0, 0, 2083333.33678258, 0, 0.000490101044535558, -1.25560048736076e-06, -0.000645021546211804, 9.34043882913078e-05, 0.00204003588900984, 0.00222459009430987, 0.00135392569239928, 0, 0, 0, 9027777.78297429, 0, 3472222.22556603, -0.000235338015241263, -0.00065647671372042, 0.00087031315489419, 2083333.34504683, 0.00308642367050335, 0.00786669481916533, -0.00335086866099953, 0.00491273805360928, 0.00520410332021945, 9.53696208782519e-06, 0.00144395664844128, 3472222.23830178, -0.00473711186790286, -0.0034245262360752, 0.000861118905567585, 0.0164763489844246, 0.00399088236875151, 0.0140637650425018, 0.00264609486819547, 0, 0, 0, 0, 0, 0, 6.31807214329174e-05, 0, 8.45913876434031e-06, 0, 0, 0, -0.000101417551954461, 0, 9.51282109606544e-05, 1.23860541929551e-05, 0.000212064462729945, -0.00019511571338376, 0, 0, 0, -0.00093024164912136, 0, 1.81762502766802e-05, -3.24808961634926e-05, 0.000154616283835896, 8.35207813345092e-05, -0.000276296166739283, -0.00070353742923189, -0.000231470289828853, 0, 0, 0, -3472222.22411824, 0, -0.000778664639251047, -3.09316963396924e-05, 0.000299848573748083, -0.000378254646669371, -0.0036216209836268, -0.00122405055433549, -0.00192824635953247, 0.000986013770919141, -0.00134282534265757, -0.0015744766020581, 0, 0, -3472222.22406567, 3.44032183599451e-05, -0.000257564845308481, -0.000293520366059891, -3472222.22955459, -0.00113117178568899, -0.00613288372429175, 0.00243150646787111, 0.00214726450684989, -0.000893956103271163, -0.00757195000537724, -0.00253841809915335, -0.00537887620875651, 0.00174461761329553, -0.000119949442780341, -0.0102361358305408, 0.00210338248234129, -0.00387877544822727, -0.00435158999547422}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -50.0000001242394, 0, 0, 0, 0, 0, 0, 0, 966.666669066738, 0, 1091.66666901428, 50.0000010379392, 0, 0, 0, 0, 0, 0, -5916.66668330731, 0, -8333.33334520012, 0, 0, -21055.5556022026, -6.00933532478187e-07, -916.66667865011, -1041.66669041101, 0, 0, 0, 0, 0, 0, 13333.3333833119, 0, 28125.0000247008, 0, 0, 0, 128263.889222782, 0, 160069.444674664, -1.56710719789683e-05, 5000.00006691849, 7291.66683165399, 1.68125431779756e-06, 4.09066982018649e-05, 19097.2224917105, 0, 0, 0, 0, 0, 0, -8333.33339549328, 0, -41666.6666810133, 0, 0, 0, -286111.11221154, 0, -536458.333826682, 5.59916690949493e-05, -8333.33346996508, -20833.3338904449, 0, 0, -967013.890425746, 0.000125120227179822, 0.000189357961113588, -6.90716425305731e-05, -104166.668258118, -0.000478763471152568, -133680.556883092, -0.0002634131598318, 0, 0, 0, 0, 0, 0, -7.03545967810797e-06, 0, 20833.3333077484, 0, 0, 0, 173611.11302473, 0, 784722.222662523, -7.00386837462633e-05, -1.56992257639549e-06, 20833.3344163008, 0, 0, 0, 2118055.55992037, 0, 3194444.44757916, -0.00026567646429513, -0.000819636946541033, 0.000203580843944227, 173611.115320608, 0.00180718190643763, 381944.4472688, -0.00024887810786398, 0.000186079845247862, 729166.673261443, -0.00121932466204627, 0.00232608633950434, 0.0017798920697475, 0, 0, 0, 0, 0, 0, 9.91152075562967e-05, 0, 4.58208661694995e-05, 0, 0, 0, -0.00180535109103946, 0, -381944.444467241, 3.13760298325857e-05, 0.00034025537530842, -0.00122992029422058, 0, 0, 0, -1215277.7838096, 0, -4548611.11411626, 0.000154078024632885, 0.00120905381495563, -8.08517754714266e-05, -0.00521939350036746, -0.00322144236895361, -381944.447509792, 0, 0, -6770833.34055348, 0.000601414215613059, 0.00121556591359669, 0.000115074492123337, -1215277.79221192, -0.00211290574959778, -2083333.34394025, 0.00331007175013884, -0.00588982844807781, -0.00356302517424736, 0.000869314912059125, 0.00181286763488671, -0.00836756675201881, 0, 0, 0, 0, 0, 0, -9.58608919993061e-05, 0, -2.69504648445786e-05, 0, 0, 0, 0.000866325581222395, 0, -0.000195028455413312, 7.74640670936956e-06, -0.000406396888626917, 0.000768652924210406, 0, 0, 0, 0.00402381181282242, 0, 2083333.33458593, 7.29487180355468e-05, -0.000779883683625838, -0.000209997755378036, 0.00293638810521077, 0.00266249749553529, 0.0016785953502734, 0, 0, 0, 3472222.22917936, 0, 9027777.78309839, -0.000269841203518697, -0.00175823011230321, 8.50813522529855e-05, 0.013316801905388, 0.00329162542656311, 2083333.34075133, -0.00309386272859309, 0.0054439903524169, 0.00357080511000078, -0.000384763580576995, 0.000400214569690677, 3472222.2385246, -0.00376753938053035, -0.00551351522515041, 0.00137668346697882, 0.0148459442232757, 0.000225554310966175, 0.00869772976136008, 0.00101663036295163, 0, 0, 0, 0, 0, 0, 2.99669526834785e-05, 0, 5.18855771497641e-06, 0, 0, 0, -0.000162665695319681, 0, 8.32583080138751e-05, -5.61007651478108e-06, 0.000148821715466765, -0.000200600141197541, 0, 0, 0, -0.0010466609134357, 0, -0.000129696555343465, -7.39595364922706e-05, 0.000195996203120777, 0.000158652209335018, -0.000569920757610159, -0.000858218668497117, -0.000375637154889906, 0, 0, 0, -0.00238158060581392, 0, -3472222.22361478, -3.62161299410387e-06, 0.000686991848800023, -5.7331651230978e-05, -0.0043297348361737, -0.00152845292387683, -0.0018636475170847, 0.000911427428645289, -0.00155882839530058, -0.00137831304859514, 0, 0, -3472222.22493382, 0.00013289474071693, 0.000387464222962057, 4.39627996419634e-05, -0.00798800365371591, -0.000998857166253871, -3472222.22826225, 0.00179645434731312, 0.00284824785333729, -0.00115920614814346, -0.00762534100724479, -0.00104376126166855, -0.00293411264304875, 0.00114674228552133, 0.00179683963346463, -0.00825532457181045, 0.00291213748035813, -0.00352632386821664, -0.00078873209017556}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -55.5555556127061, 0, 0, 0, 0, 0, 0, 0, 888.88889011001, 0, 1324.07407518688, 55.555555913468, 0, 0, 0, 0, 0, 0, -3611.11112031813, 0, -11685.1851898903, 0, 0, -21129.6296542427, -5.27092866552997e-07, -833.333339951022, -1268.51852625409, 0, 0, 0, 0, 0, 0, 2777.77780618504, 0, 49768.5185235594, 0, 0, 0, 85231.4816806396, 0, 185694.444551399, 1.15283722568458e-06, 2777.77782534252, 10416.666699606, -9.77342996617198e-06, 1.97793262551556e-05, 19027.7779318815, 0, 0, 0, 0, 0, 0, -3.26741619666036e-05, 0, -108796.296284903, 0, 0, 0, -63425.9266259796, 0, -785879.62978921, -1.19875415154508e-05, -0.000123893893622419, -39351.8518996703, 0, 0, -740509.260140859, 7.79654430432679e-05, -3.33338885156889e-05, 9.50811241890234e-06, -63425.9270142455, -0.00012182191093545, -156250.000660749, -0.000161026145639101, 0, 0, 0, 0, 0, 0, -1.3661857212066e-05, 0, 115740.7407059, 0, 0, 0, 0.00124289662050565, 0, 1701388.88886047, 4.12130167274042e-05, 0.000111074602471762, 69444.4444321243, 0, 0, 0, 520833.336091477, 0, 3078703.7053234, -0.000140842933148619, -2.62808982964686e-05, -5.5897176176667e-06, 0.00311031680706492, 0.000357233990951315, 590277.778812967, -0.000220779832599363, 0.000329468473971154, 520833.337666143, -0.000349381158982539, 0.00152105287600148, 0.000502955162393884, 0, 0, 0, 0, 0, 0, 7.01723441038701e-05, 0, -46296.2962615715, 0, 0, 0, -0.00117901052183887, 0, -1782407.4070846, -6.40136284355771e-05, 4.85435340898606e-05, -46296.2962026091, 0, 0, 0, -0.0039664922687208, 0, -6481481.48278582, 9.82021839203541e-05, 7.63581176123891e-05, 0.000142894915565013, -0.00416248554643179, -0.000478496227062332, -1041666.66714012, 0, 0, -1967592.59701226, 0.000299561034523858, -0.000342913559697954, 0.000355078744967469, -0.0100141499821701, -0.00101390664349525, -1967592.59919804, 0.00117394885687351, -0.00389423228904998, -0.000240394360258407, -0.000421012408887304, 0.000572972026163845, -0.00498968821627274, 0, 0, 0, 0, 0, 0, -6.29217652795056e-05, 0, -1.37971453104432e-05, 0, 0, 0, 0.000565934070495698, 0, 694444.444132765, 4.5485216656762e-05, -0.000138335920797529, -8.70707951465099e-05, 0, 0, 0, 0.00270504010699866, 0, 6481481.4817461, -2.94360673367808e-06, -5.77675394645163e-05, -0.000227511426804236, 0.00265065380271973, 0.000373016166261378, 694444.444186811, 0, 0, 0, 0.00453316963222029, 0, 3472222.22527794, -0.000205204291632759, 0.000124174609687526, -0.000337126882119068, 0.00960554095427954, 0.00121008485470757, 3472222.22647433, -0.000901309366193597, 0.00363971797202762, -0.000607449194600321, -0.00035717949651706, 0.0010348051959586, 0.0113335112681158, -0.000845202509687637, -0.00124542576281305, 0.00167101446180333, 0.00960505689820187, -0.00105984582484652, 0.00448340258766213, 0.00196310285668517, 0, 0, 0, 0, 0, 0, 1.89524551251359e-05, 0, 1.64686794793211e-06, 0, 0, 0, -0.000105775199866583, 0, 9.4084006250085e-05, -1.36900088776276e-05, 5.97872052992149e-05, 2.53379186659584e-05, 0, 0, 0, -0.00070957354372613, 0, -2314814.8147265, -1.46039174075947e-05, 1.33286126774948e-05, 0.000104960008714351, -0.00063566820410134, -0.000106066493330544, 0.000213893016579593, 0, 0, 0, -0.00160593000184862, 0, -2314814.8155294, 2.27036863408564e-06, -1.56322479928525e-06, 0.000149989165453331, -0.0032913028175044, -0.000510943290722656, -2314814.81570834, 0.000302193966656215, -0.00114205652939409, 0.000385191764180282, 0, 0, -0.00173370665788106, 0.000104088955791712, -0.00032185404660041, 0.000128327188167417, -0.00569399273805555, -0.000649082125554012, -0.0042509088491341, 0.000500537286451363, 0.000726719646238945, -0.000760819872392389, -0.00506960602244369, 0.000508473977984006, -0.000982035351449771, 8.38613861923365e-05, 0.000609389931907203, -0.00521875901691321, 0.000196134676794978, -0.00272066029606871, -0.00050355274755916}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -71.4285714185844, 0, 0, 0, 0, 0, 0, 0, 785.714285485717, 0, 1821.42857128546, 71.4285718410157, 0, 0, 0, 0, 0, 0, -714.285713065159, 0, -17861.1111099992, 0, 0, -19964.2857110475, -3.58266253302605e-06, -714.285720092485, -1750.00000625619, 0, 0, 0, 0, 0, 0, -4.20955799043109e-06, 0, 89027.7777747847, 0, 0, 0, 17499.9999912423, 0, 194722.22219305, 1.87319323102174e-05, 3.99014665628499e-05, 16111.1111432975, 2.49977485760915e-05, 5.01781184932879e-05, 17500.0000802119, 0, 0, 0, 0, 0, 0, 1.17498516849615e-05, 0, -246527.777776612, 0, 0, 0, -4.62164031694746e-06, 0, -963194.444332195, -5.2019173889903e-05, -0.000134083456178889, -72916.6667403128, 0, 0, -161111.111014456, -5.76055862946467e-05, -0.000225926709785486, -0.000135467406646506, -0.000479473793788168, -0.000239231109275193, -161111.111466257, -0.00029970181086753, 0, 0, 0, 0, 0, 0, -2.39400110155143e-05, 0, 381944.444454025, 0, 0, 0, 5.28021478650543e-05, 0, 2638888.88866024, 8.49859108235536e-05, 0.000243311922349244, 173611.111202721, 0, 0, 0, -0.000104346144158936, 0, 729166.666330589, 6.33317760386798e-05, 0.000533231903609771, 0.000403704950981573, 0.00127024861886907, 0.000467575834217441, 729166.667354351, 0.000183716014275361, 0.000956175270014737, 0.0018557952390852, 0.000792945907350585, 0.00116943577716489, 0.00119290139428587, 0, 0, 0, 0, 0, 0, 2.99659051600248e-05, 0, -307539.682560425, 0, 0, 0, -8.68398065115311e-05, 0, -4027777.77752104, -8.7574438482122e-05, -0.000246249274161703, -208333.333401492, 0, 0, 0, -5.83211950962657e-07, 0, -1736111.11054488, -2.85583237852728e-05, -0.000608541822134625, -0.000653773463833334, -0.00172703868582267, -0.000471463293614632, -1736111.11172488, 0, 0, 0.000361442690388827, -3.63899892106084e-05, -0.00173345605277762, -0.000530625673577804, -0.00383907920539792, -0.00160067876046408, -0.00317925785516039, -0.00123152556473103, -0.0022852525392257, -0.00152664307474917, -0.0014130510174216, -0.0027116774102601, -0.00373104777573471, 0, 0, 0, 0, 0, 0, -2.02228140631193e-05, 0, 99206.3492231772, 0, 0, 0, 5.93715263246513e-05, 0, 3174603.17445603, 5.31540526190236e-05, 0.000129804533106853, 99206.3492412741, 0, 0, 0, 7.16066747020778e-05, 0, 2083333.33287098, 4.78179649659864e-07, 0.000354676872006477, 0.000508216478657773, 0.00117684655090411, 0.000235566200046937, 2083333.33354068, 0, 0, 0, -0.00015087756738734, 0, -0.000410916756340127, -6.74757881667733e-05, 0.00137949553103459, 0.000656102042628301, 0.00341632389818951, 0.00118578649925229, 0.00243127063748829, 0.00073885382980717, 0.00212444061034402, 0.00070511440914011, 0.000268883055274762, 0.00204645447487893, 0.00460150683867562, 0.000863086265131463, 0.00318772372089099, 0.00151327712318344, 0.00500364273368506, 0.00282373587156265, 0.003898024942974, 0.00365775206155884, 0, 0, 0, 0, 0, 0, 5.55770206922537e-06, 0, -5.0319924715549e-06, 0, 0, 0, -1.47222919353061e-05, 0, -992063.492030565, -1.34952476150948e-05, -2.62220106144904e-05, -1.02916551590178e-05, 0, 0, 0, -3.7499219629958e-05, 0, -992063.491913848, -7.31970280263565e-08, -8.76464594630561e-05, -0.000153116819393791, -0.00032405207190455, -5.6362124000752e-05, -992063.492067633, 0, 0, 0, 1.71368270564545e-05, 0, 0.000175540868979599, 6.76587664157069e-05, -0.000398068422794055, -0.000293589016237906, -0.00113531625958029, -0.000328939481707661, -0.000650674272311403, -0.000176630995071534, -0.000752072292737822, -1.79988242745869e-05, 0, 0, 9.43108030997054e-05, -3.73890729147749e-05, -0.000852419145862955, -0.000224194193020258, -0.00204995246400967, -0.000816524894121406, -0.00192318744320393, -0.000135121647966331, -0.00134658470532225, -0.000757160781960718, -0.00223876547922946, -0.00072528255193097, -0.00129022067660206, -0.000319379869346969, -0.0018971843717053, -0.00289162518220437, -0.00165667922743818, -0.00223131680507557, -0.00211540462848718}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 642.85714268709, 0, 0, 0, 0, 0, 0, 0, -1357.14285352089, 0, -17107.1428541369, -1357.14285711733, 0, 0, 0, 0, 0, 0, 714.285690720987, 0, 178964.285695786, 0, 0, 34678.5713651812, 714.285716926107, 1428.57143358139, 34678.5714242538, 0, 0, 0, 0, 0, 0, 7.89906578429981e-05, 0, -979861.111059504, 0, 0, 0, -17499.9996336601, 0, -341111.110714975, -4.15420019048334e-06, -6.38957331471022e-05, -341111.111043864, -7.24055252358046e-05, -17500.000050066, -34999.9999973499, 0, 0, 0, 0, 0, 0, -0.000156925950013877, 0, 3109027.77771369, 0, 0, 0, -0.00104528387455999, 0, 1707638.88767909, -4.74888069849872e-05, 0.00022376333667504, 1707638.88856387, 0, 0, 161111.109146642, 0.000184337507795693, 0.000309960192833743, 0.000476393766080833, 0.000535797788192282, 161111.111136951, 322222.221636848, 0.000752514562615538, 0, 0, 0, 0, 0, 0, 0.00019531638609637, 0, -5902777.77777306, 0, 0, 0, 0.00166119125432459, 0, -4756944.4423893, 0.000198377449507438, -0.000382773461164683, -4756944.44377946, 0, 0, 0, 0.0045976549440289, 0, -729166.661805936, -1.59187816920488e-06, -0.000711788805754482, -0.00107101376211471, -0.00177319508165156, -729166.666084527, -1458333.33026782, -0.00115892881465257, -0.000888932548644922, -0.000573147786354245, -0.00284237928687879, -0.00454587134455783, -0.000910135308020389, 0, 0, 0, 0, 0, 0, -0.000153392450625652, 0, 6587301.5873776, 0, 0, 0, -0.00148849409687661, 0, 7430555.55356875, -0.000319993323983608, 0.000377485790630337, 7430555.55491407, 0, 0, 0, -0.00567688524377565, 0, 1736111.10483849, -0.000616270606318163, 0.00064949569586182, 0.0010647915303281, 0.00240270161693529, 1736111.10990372, 3472222.21615864, 0, 0, -0.00852247176516795, 0.00104833655497716, 0.0022859477268331, 0.00213989491748254, 0.00387188391543087, 0.000318806237915458, -0.00215782736145224, 0.0038816053365669, 0.00916668690512037, -0.00363833233591553, 0.00989274288029287, 0.00477187944958942, 0.00957917243931309, 0, 0, 0, 0, 0, 0, 7.10860430488527e-05, 0, -3968253.96833038, 0, 0, 0, 0.000704237860113241, 0, -6051587.30057476, 0.000223470183767177, -0.00020717012204111, -6051587.3013361, 0, 0, 0, 0.00353001827148269, 0, -2083333.32917525, 0.000768961858217486, -0.00018825815068448, -0.000430526948107469, -0.00152302293197397, -2083333.33251498, -4166666.6613778, 0, 0, 0, 0.00708871298652602, 0, 0.00717163040365698, -4.01982890313147e-05, -0.00208268215737904, -0.00150674079470654, -0.00394044113340027, 0.000529587356790802, 0.0048811982590875, -0.00189830142002669, -0.00736944075312112, 0.00742302730635683, -0.00180860397627787, -0.00172771632768608, -0.0024548965929341, -0.00813759005786458, -0.00698945199489896, -0.0122724537331516, -0.0161412338726141, -0.000531770797979437, -0.00349278297518109, -0.01499682116026, 0, 0, 0, 0, 0, 0, -1.49280713475236e-05, 0, 992063.492087374, 0, 0, 0, -0.000138341183142194, 0, 1984126.98391877, -5.92466812366596e-05, 4.77130587474961e-05, 1984126.98410819, 0, 0, 0, -0.00086447731647828, 0, 992063.490945802, -0.000286481279526978, -7.69582453242148e-06, 5.8204177955185e-05, 0.000406019148984641, 992063.491913013, 1984126.98241798, 0, 0, 0, -0.00226759131763488, 0, -0.00232035132430536, -0.000300217543875728, 0.000614456290933691, 0.000293855353490468, 0.00124564494679782, -0.000288225504176778, -0.00273679168850322, 0.000133767696074826, 0.00203189933466421, -0.00358978905026498, 0, 0, -0.0029332322111621, 0.000449103410749805, 0.00103007344452033, 0.000733206191256678, 0.00179287651352489, 0.000320341426431414, 3.49911870561281e-05, 0.00216978438643531, 0.00305175011418632, 0.00503490107086099, 0.0072513010326317, -0.0023741071933471, -0.00198817129740031, 0.00528505155434601, 0.0022783335675381, 0.0065686407089748, 0.0072221890043168, 0.0115263642332909, 0.00440023767362414}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -1999.99999935723, 0, 0, 0, 0, 0, 0, 0, 6722.22220829236, 0, 52388.888877577, 6722.2222213075, 0, 0, 0, 0, 0, 0, -7499.99991122232, 0, -535990.740672017, 0, 0, -168490.740494808, -7500.00000493747, -15000.0000031728, -168490.740718117, 0, 0, 0, 0, 0, 0, 2777.77749186685, 0, 2851111.11092701, 0, 0, 0, 179583.331927322, 0, 1611249.99845026, 2777.77780530877, 8333.33347203418, 1611249.99970327, 8333.33348298077, 179583.333475761, 359166.666599773, 0, 0, 0, 0, 0, 0, 0.00054102655830326, 0, -8731249.99980015, 0, 0, 0, -63425.922032214, 0, -7764351.84711262, 5.3293215239356e-05, -0.000500527985535672, -7764351.85047791, 0, 0, -1596527.77010092, -0.000422771645313209, -63425.9271899313, -0.00107817178877838, -190277.779688433, -1596527.77752262, -3193055.55310646, -190277.78014311, 0, 0, 0, 0, 0, 0, -0.0006408127520652, 0, 15902777.7778441, 0, 0, 0, -0.00595409365215444, 0, 20590277.7697428, -0.00041789168504156, 0.000751675681514786, 20590277.7749942, 0, 0, 0, 520833.31572273, 0, 6874999.98081354, -0.000191983310285041, 0.00340475417192674, 0.00203656213098401, 0.00703796330827581, 6874999.99659065, 13749999.9877451, 0.00282751307202295, 520833.336574629, 1562500.00221088, 0.0100787152648282, 0.015586017286569, 1562500.00151957, 0, 0, 0, 0, 0, 0, 0.000484444398255992, 0, -16944444.4448228, 0, 0, 0, 0.00508712010318907, 0, -30254629.6218973, 0.000765901390464156, -0.000584214432190215, -30254629.6269328, 0, 0, 0, 0.0212213110307076, 0, -15277777.7528279, 0.00223136966767276, -0.00379796814701617, -0.000978100815758008, -0.0101149173187148, -15277777.7711935, -30555555.5317464, 0, 0, -1967592.55950087, -0.00259368907376244, -0.00991158525224014, -0.00455505578842061, -0.0169062261646935, -1967592.59257474, -5902777.76888167, -0.0152749391550293, -0.032379974121783, -5902777.75927788, -0.0343895546806446, -0.0173970552752933, -0.0336282804263742, 0, 0, 0, 0, 0, 0, -0.000221725051938974, 0, 9722222.22256195, 0, 0, 0, -0.00226465225991522, 0, 22916666.6627598, -0.000551692751205749, 0.00022588908831482, 22916666.6655626, 0, 0, 0, -0.0128672654255542, 0, 16666666.6500022, -0.00273137866102895, 0.00180598287742997, -0.00070120070436702, 0.00667992355020209, 16666666.6619956, 33333333.3128081, 0, 0, 0, -0.026964708768149, 0, 3472222.19408465, -0.00031632403810293, 0.00978567732187308, 0.00191672585866686, 0.0189010444806255, 3472222.21829616, 10416666.6470257, 0.00853874254930213, 0.0263134362119944, 10416666.6336993, 0.00454648888601503, 0.00700344933453243, 0.0109102450634587, 0.029828120978625, 0.0285660256225528, 0.0419562562654456, 0.0620410751719742, -0.00183408010263787, 0.0083266933128798, 0.0549358590146345, 0, 0, 0, 0, 0, 0, 4.73836523119239e-05, 0, -2314814.81491645, 0, 0, 0, 0.000411492095020217, 0, -6944444.44365364, 0.00014646088843881, -2.83029308999165e-05, -6944444.44433662, 0, 0, 0, 0.00306372119976393, 0, -6944444.439933, 0.00101851302514326, -0.000304565991776822, 0.000503492027350973, -0.00177858702027863, -6944444.4434074, -13888888.8823261, 0, 0, 0, 0.00846393398160809, 0, -2314814.8056101, 0.0011452601719586, -0.00313517222834123, 0.000334896660474135, -0.00661434696714655, -2314814.81285735, -6944444.43366951, -0.00107846157220583, -0.00722619106721393, -6944444.42902868, 0, 0, 0.0112658071656348, -0.0012624829124945, -0.00464665489125378, -0.00152660881167678, -0.0087352702387572, -0.000924019964522617, -0.000271271085381196, -0.00833980185489904, -0.0130794850681665, -0.0164689541895187, -0.0297798187587843, 0.0115133992470937, 0.0104841811829451, -0.0181411326297543, -0.00874220551871197, -0.0247101853289727, -0.028295841437054, -0.0428879932433852, -0.0145273252762938}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 4199.99999858059, 0, 0, 0, 0, 0, 0, 0, -20116.6666335836, 0, -107616.666642335, -20116.6666645484, 0, 0, 0, 0, 0, 0, 35916.6664579703, 0, 1067513.88875627, 0, 0, 490930.554961193, 35916.6666624588, 71833.3333243991, 490930.555537222, 0, 0, 0, 0, 0, 0, -28333.3327021555, 0, -5460277.77750946, 0, 0, 0, -833263.885389478, 0, -4515208.32969012, -28333.3333452148, -85000.0001504673, -4515208.3329057, -84999.9999834989, -833263.889206423, -1666527.77804635, 0, 0, 0, 0, 0, 0, 8333.33229809911, 0, 15968402.777782, 0, 0, 0, 623611.101324594, 0, 20661805.5448886, 8333.33330135595, 33333.3337977773, 20661805.553275, 0, 0, 7058680.53673365, 33333.3326901468, 623611.113636496, 50000.0002672119, 1870833.33960937, 7058680.55517346, 14117361.1083499, 1870833.3391092, 0, 0, 0, 0, 0, 0, 0.000935345132457178, 0, -27652777.7787059, 0, 0, 0, -173611.096384995, 0, -51402777.7603754, 0.000315795522185473, -0.000250474280694028, -51402777.7730075, 0, 0, 0, -4826388.84424914, 0, -28402777.7326976, 0.00328900354565175, -173611.118438509, 0.00150370070119524, -694444.465462856, -28402777.7706575, -56805555.5352447, 0.00136961510195597, -4826388.89508992, -14479166.6824843, -694444.463377062, -1041666.70364632, -14479166.6763572, 0, 0, 0, 0, 0, 0, -0.000446707038524929, 0, 27986111.1127356, 0, 0, 0, -0.0118815105273308, 0, 70208333.3169989, -0.00060699419562536, -0.000467100112001258, 70208333.3288025, 0, 0, 0, 1215277.72266501, 0, 57673611.0549997, -0.00728363997238502, 0.00946166537057815, -0.00692095912945665, 0.030068493713784, 57673611.0968755, 115347222.178546, 0, 0, 16666666.5859305, -0.00518365999519947, 1215277.79786369, -0.00567216731465443, 4861111.16968823, 16666666.6645564, 50000000.0006403, 0.0262389053202087, 0.0774430242190565, 49999999.9713378, 0.0638747222488804, 4861111.15259985, 7291666.76528595, 0, 0, 0, 0, 0, 0, 0.000105698102809653, 0, -15277777.7789624, 0, 0, 0, 0.00465294818156057, 0, -49305555.5474144, 0.000389961873639885, 0.000666252027501722, -49305555.553899, 0, 0, 0, 0.0337454139554879, 0, -56249999.9636033, 0.006775636904809, -0.00576155374030742, 0.00850896960072457, -0.0202668040221879, -56249999.9897128, -112499999.960414, 0, 0, 0, -3472222.1538607, 0, -25694444.3792868, 0.00765891654563036, -0.0212411665685495, 0.0107870710119829, -0.0617743544022951, -25694444.4337471, -77083333.3079484, -0.0138372229795425, -0.0644117626508023, -77083333.2707365, 0.00480431945063709, -3472222.23446858, -13888888.9380662, -0.0465680449934821, -0.0598797962118194, -0.0761746956333513, -0.174685880054601, -13888888.8904488, -20833333.3820923, -0.129703443374485, 0, 0, 0, 0, 0, 0, -1.30574892970927e-05, 0, 3472222.22254785, 0, 0, 0, -0.000645061501668589, 0, 13888888.8872594, -8.50709559684067e-05, -0.000251773951446789, 13888888.888833, 0, 0, 0, -0.00796246557759654, 0, 20833333.323578, -0.00224021027122733, 0.00142640498969488, -0.00327349568695879, 0.00542757155312971, 20833333.330984, 41666666.6535827, 0, 0, 0, -0.0221294442640131, 0, 13888888.8685615, -0.00408550428511029, 0.00740401836434313, -0.00612322112594473, 0.0218278762017711, 13888888.883867, 41666666.6490482, 0.00136763156394482, 0.0181801300165338, 41666666.6358518, 0, 0, 3472222.19505936, -0.00185943202596664, 0.00917486785858745, -0.0034592458580061, 0.0304689813561199, 3472222.22259831, 13888888.9003092, 0.0107616267433548, 0.0268196410247778, 0.0285525155199428, 0.0861640236269351, 13888888.8663665, 20833333.3221564, 0.0255968879576153, 0.0188953729885996, 0.0775803294509182, 0.0590405592280095, 0.10095955055244, 0.0410012122007173}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -6299.99999848264, 0, 0, 0, 0, 0, 0, 0, 40674.9999571823, 0, 156174.999973712, 40675.0000145857, 0, 0, 0, 0, 0, 0, -104166.666422119, 0, -1479874.99990576, 0, 0, -954041.665795891, -104166.666731167, -208333.333568159, -954041.666954579, 0, 0, 0, 0, 0, 0, 132291.666144901, 0, 7159513.88903366, 0, 0, 0, 2306597.21690567, 0, 8283819.43929636, 132291.666788599, 396875.001632959, 8283819.44579491, 396875.000836714, 2306597.2239175, 4613194.4487146, 0, 0, 0, 0, 0, 0, -83333.3332966891, 0, -19704513.8904183, 0, 0, 0, -2758680.54061564, 0, -35221874.9861906, -83333.3335908789, -333333.338326008, -35221875.0030541, 0, 0, -18192708.3049222, -333333.332538263, -2758680.56078119, -500000.00449042, -8276041.69428169, -18192708.3394731, -36385416.6822549, -8276041.69026507, 0, 0, 0, 0, 0, 0, 20833.334982582, 0, 32138888.8927421, 0, 0, 0, 1631944.42324963, 0, 80624999.9792128, 20833.3340190542, 104166.674749553, 80625000.0047269, 0, 0, 0, 19548611.0392981, 0, 66423611.0488588, 104166.659494268, 1631944.45600344, 208333.341257319, 6527777.85301503, 66423611.1180135, 132847222.241148, 208333.331164955, 19548611.1276445, 58645833.4291427, 6527777.82201477, 9791666.78122806, 58645833.4144557, 0, 0, 0, 0, 0, 0, -0.0027442705487828, 0, -30763888.8937027, 0, 0, 0, -381944.430454613, 0, -101388888.870093, -0.00110974178209983, -0.0073651510500621, -101388888.894546, 0, 0, 0, -10243055.4628142, 0, -119965277.70827, 0.0133200916800915, -381944.460424407, -0.00319631213663245, -1909722.32497734, -119965277.780567, -239930555.555365, 0, 0, -59201388.7679573, 0.0170125456388692, -10243055.5863804, 0.00795893192507031, -40972222.4322634, -59201388.9012193, -177604166.800947, -1909722.25997509, -3819444.67225055, -177604166.749397, -3819444.58563159, -40972222.3622236, -61458333.6744598, 0, 0, 0, 0, 0, 0, 0.00182038451620515, 0, 15972222.2252902, 0, 0, 0, -0.00260891804185684, 0, 65972222.2127753, 0.000956971204127678, 0.00354889694715094, 65972222.2267456, 0, 0, 0, 2083333.27588642, 0, 104166666.624935, -0.0103572416435291, 0.0120150996462816, -0.00378769401900984, 0.0701633174823988, 104166666.66796, 208333333.316922, 0, 0, 0, 24305555.4442025, 0, 76388888.8028106, -0.0198611264050167, 2083333.36390914, -0.0140137752897915, 10416666.8601001, 76388888.8881015, 229166666.747941, 0.0112120756490275, 0.198859040523255, 229166666.683222, -0.0181079542329759, 24305555.5749741, 97222222.4401932, 0.0615560686379862, 10416666.7860644, 0.171476406992609, 20833333.8318223, 97222222.3334273, 145833333.640071, 20833333.6781699, 0, 0, 0, 0, 0, 0, -0.000443705749797085, 0, -3472222.22301468, 0, 0, 0, -0.000646623514258633, 0, -17361111.1091573, -0.000306260028014928, -0.000694316529775837, -17361111.1126954, 0, 0, 0, 0.0131298896453262, 0, -34722222.2112158, 0.00293776753577327, -0.0037674975442823, 0.00267738273980459, -0.0192650161502235, -34722222.2232945, -69444444.4357972, 0, 0, 0, -3472222.18392486, 0, -34722222.1988508, 0.0079477839564319, -0.0115966538967469, 0.00783035613107496, -0.0655412659053434, -34722222.2203035, -104166666.680975, 0.00104947437245826, -0.062901470656295, -104166666.657255, 0, 0, -17361111.0712508, 0.00802780574267864, -3472222.23385696, 0.00879916858953202, -17361111.2123224, -17361111.1125695, -69444444.5255247, -0.0026200125400976, -0.0404057037681773, -0.0706408034684802, -0.235683564801138, -69444444.4612876, -104166666.74919, -0.0315339646393537, -17361111.167317, -34722222.4748161, -0.117370278820929, -0.240883961824559, -34722222.3869473}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 7000.00000130654, 0, 0, 0, 0, 0, 0, 0, -59194.4444468559, 0, -164194.444461777, -59194.4445249155, 0, 0, 0, 0, 0, 0, 206129.629767433, 0, 1444046.29654915, 0, 0, 1300175.92549519, 206129.630065095, 412259.260490378, 1300175.92718106, 0, 0, 0, 0, 0, 0, -378472.223512449, 0, -6430138.8905518, 0, 0, 0, -4227361.10837677, 0, -10279027.7752711, -378472.223383207, -1135416.67442369, -10279027.7849054, -1135416.67260357, -4227361.11729327, -8454722.23834686, 0, 0, 0, 0, 0, 0, 386574.079243237, 0, 16370138.8942572, 0, 0, 0, 7223379.62300335, 0, 39190509.2546004, 386574.076411044, 1546296.32094675, 39190509.2805649, 0, 0, 29657175.9093617, 1546296.3070064, 7223379.6443965, 2319444.47557255, 21670138.9748885, 29657175.9541094, 59314351.923288, 21670138.962532, 0, 0, 0, 0, 0, 0, -208333.344085471, 0, -24930555.565321, 0, 0, 0, -6840277.77368317, 0, -81006944.4401934, -208333.337336296, -1041666.71100973, -81006944.4823677, 0, 0, 0, -44201388.8410674, 0, -93645833.3067066, -1041666.67460387, -6840277.80231704, -2083333.41044143, -27361111.3291354, -93645833.3946609, -187291666.818287, -2083333.38000999, -44201388.942917, -132604166.977177, -27361111.23873, -41041666.9833374, -132604166.939532, 0, 0, 0, 0, 0, 0, 46296.3084053028, 0, 22500000.0104012, 0, 0, 0, 3402777.78594364, 0, 93217592.5890951, 46296.3009931806, 277777.823879458, 93217592.6334077, 0, 0, 0, 35648148.0815245, 0, 147916666.651424, 277777.778488407, 3402777.80791327, 694444.543008218, 17013889.1793271, 147916666.742357, 295833333.50045, 0, 0, 109490740.674866, 694444.474693366, 35648148.2153575, 925926.016108297, 142592593.180125, 109490740.820765, 328472222.737308, 17013888.9899111, 34027778.3875468, 328472222.644754, 34027778.1706159, 142592592.995829, 213888889.791296, 0, 0, 0, 0, 0, 0, -0.00700441037063829, 0, -11111111.1171884, 0, 0, 0, -694444.457634608, 0, -56249999.9973395, -0.00305030947666276, -0.0258570906459301, -56250000.0248396, 0, 0, 0, -14583333.292936, 0, -114583333.329339, 0.00137541741648041, -694444.467146676, -0.0625296766849995, -4166666.86420512, -114583333.385453, -229166666.756139, 0, 0, 0, -62499999.9238763, 0, -118055555.531556, -0.00735469473859193, -14583333.3873415, -0.0796473018366865, -72916667.1743382, -118055555.613291, -354166667.071722, -4166666.6989607, -10416667.2097511, -354166666.968253, -0.0464011874475713, -62500000.059915, -250000000.655351, -10416666.8492466, -72916666.9430923, -13888889.3819802, -145833334.512197, -250000000.435226, -375000000.953509, -145833334.185345, 0, 0, 0, 0, 0, 0, 0.00163204744709929, 0, 2314814.81631436, 0, 0, 0, 0.00526257189066034, 0, 13888888.8880628, 0.000786640773927216, 0.00606410164828078, 13888888.8954709, 0, 0, 0, 2314814.80713062, 0, 34722222.2206763, -0.000244189945504721, 0.00728683075761974, 0.0155687434954661, 0.0543181266316743, 34722222.2377948, 69444444.4621802, 0, 0, 0, 13888888.8588748, 0, 46296296.2971019, 0.000145960046334938, 2314814.8345078, 0.0259295337549383, 13888889.0555645, 46296296.3143037, 138888889.007685, 0.00179649554205176, 0.181261216246529, 138888888.971093, 0, 0, 34722222.2037449, 0.0101178253977161, 13888888.9118479, 0.0231631127824726, 69444444.7210048, 34722222.2395618, 138888889.156211, 0.0180304545636692, 13888888.9627115, 0.220173469005176, 34722222.7545499, 138888889.043561, 208333333.68247, 0.10937519439344, 69444444.6060501, 138888889.520728, 34722222.4853879, 46296296.8406911, 138888889.360096}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -6000.00000501908, 0, 0, 0, 0, 0, 0, 0, 65738.0953147278, 0, 125738.095306802, 65738.0953814205, 0, 0, 0, 0, 0, 0, -303527.778511494, 0, -960908.730721444, 0, 0, -1264436.5083766, -303527.778680748, -607055.557799629, -1264436.50989711, 0, 0, 0, 0, 0, 0, 766111.114879422, 0, 3801388.89151565, 0, 0, 0, 5333611.11424471, 0, 8368888.89090458, 766111.11401451, 2298333.3476612, 8368888.89982712, 2298333.34598377, 5333611.1211592, 10667222.2471401, 0, 0, 0, 0, 0, 0, -1142361.12237709, 0, -8803472.22930537, 0, 0, 0, -12230555.5679282, 0, -27552777.7846849, -1142361.11717557, -4569444.49103436, -27552777.8105554, 0, 0, -29837500.0101433, -4569444.47726332, -12230555.5810216, -6854166.73384437, -36691666.7981592, -29837500.0444924, -59675000.1079616, -36691666.7811315, 0, 0, 0, 0, 0, 0, 1006944.46447014, 0, 12430555.5671944, 0, 0, 0, 16458333.3631769, 0, 50729166.6775963, 1006944.45314872, 5034722.30812838, 50729166.7247807, 0, 0, 0, 55763888.9109306, 0, 78611111.1441822, 5034722.27087865, 16458333.3749603, 10069444.6173082, 65833333.6663649, 78611111.211321, 157222222.454603, 10069444.585651, 55763888.9764597, 167291667.126509, 65833333.5541431, 98750000.4844185, 167291667.062887, 0, 0, 0, 0, 0, 0, -486111.131811097, 0, -10555555.5672829, 0, 0, 0, -12986111.1538345, 0, -53263888.8950289, -486111.119066874, -2916666.75796416, -53263888.9506866, 0, 0, 0, -57638888.9145894, 0, -107986111.163471, -2916666.71126041, -12986111.1573556, -7291666.89987613, -64930555.9997541, -107986111.239003, -215972222.483862, 0, 0, -110416666.722325, -7291666.82549012, -57638888.9953984, -9722222.4994378, -230555556.416954, -110416666.802801, -331250000.766469, -64930555.7681502, -129861112.042076, -331250000.626133, -129861111.797824, -230555556.154886, -345833334.593745, 0, 0, 0, 0, 0, 0, 99206.3606577256, 0, 4960317.46700235, 0, 0, 0, 5555555.58766731, 0, 29861111.1105185, 99206.3532037351, 694444.496656845, 29861111.1479815, 0, 0, 0, 31250000.0203499, 0, 75000000.0328611, 694444.469014827, 5555555.58633754, 2083333.49160888, 33333333.6359322, 75000000.089241, 150000000.144537, 0, 0, 0, 76388888.9215264, 0, 100694444.517158, 2083333.43080545, 31250000.0795191, 3472222.47600347, 156250000.741379, 100694444.549516, 302083333.939089, 33333333.4341027, 83333334.1691842, 302083333.801164, 3472222.43472251, 76388888.9902963, 305555556.519136, 83333333.7552019, 156250000.422487, 111111111.987907, 312500001.604109, 305555556.196307, 458333334.674371, 312500001.232187, 0, 0, 0, 0, 0, 0, -0.00261736011172295, 0, -992063.493704598, 0, 0, 0, -992063.501682475, 0, -6944444.44339682, -0.000795814005439645, -0.012469467680248, -6944444.45397969, 0, 0, 0, -6944444.45259336, 0, -20833333.3392499, -0.00621317703016835, -992063.500904214, -0.0426217769953026, -6944444.52764622, -20833333.3598316, -41666666.6969845, 0, 0, 0, -20833333.3393821, 0, -34722222.2515327, -0.0265303783290128, -6944444.4711622, -0.0873318135365567, -41666666.9093047, -34722222.2564959, -104166666.845396, -6944444.46372662, -20833333.6170781, -104166666.801302, 0, 0, -34722222.2479556, -0.0662973200797261, -20833333.3697113, -0.0993866962395988, -104166667.073366, -34722222.2554784, -138888889.281624, -20833333.425007, -41666666.7812152, -34722222.6203834, -104166667.379884, -138888889.125851, -208333333.840693, -34722222.4906012, -104166666.903059, -208333334.186531, -104166667.074643, -138888889.658646, -208333333.998263}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 5250.00001416414, 0, 0, 0, 0, 0, 0, 0, -96895.8335763304, 0, -96895.83355282, -44395.8335792867, 0, 0, 0, 0, 0, 0, 598555.557535018, 0, 598555.557045154, 0, 0, 1722111.1139417, 154597.223452444, 753152.781846042, 753152.78117304, 0, 0, 0, 0, 0, 0, -1829826.39734293, 0, -1829826.39467122, 0, 0, 0, -9929062.51937732, 0, -9929062.51366372, -283854.169654926, -3943506.97094857, -3943506.96456829, -2397534.74266386, -2397534.73471063, -12326597.2669853, 0, 0, 0, 0, 0, 0, 3128472.2433174, 0, 3128472.23611788, 0, 0, 0, 27973611.1779173, 0, 27973611.1464799, 289930.560097856, 9675347.30889212, 9675347.28583031, 0, 0, 49690277.8558402, 3998263.9388851, 3998263.90691914, 10255208.4522347, 59945486.3494207, 10255208.3964231, 59945486.3150101, 35970139.0478309, 0, 0, 0, 0, 0, 0, -3055555.58744238, 0, -3055555.57688822, 0, 0, 0, -43663194.5734353, 0, -43663194.4945487, -156250.005063589, -12378472.3802847, -12378472.3388398, 0, 0, 0, -115711805.765321, 0, -115711805.724164, -3680555.6297749, -3680555.564344, -18958333.6605108, -134670139.495917, -18958333.4979923, -134670139.356442, -13159722.4717526, -13159722.3023881, -244583334.190303, -54704861.3387092, -142031250.726048, -142031250.576917, 0, 0, 0, 0, 0, 0, 1597222.25088036, 0, 1597222.24279072, 0, 0, 0, 38576389.0318699, 0, 38576388.9260627, 34722.2259848624, 8020833.4968101, 8020833.45919661, 0, 0, 0, 139930555.844651, 0, 139930555.749441, 1770833.40774413, 1770833.33140081, 16145833.7975834, 156076389.695381, 16145833.5657516, 156076389.454782, 0, 0, 205902778.121903, 8333333.62071168, 8333333.37846307, 16319444.9858316, 428125001.605007, 16319444.6140258, 428125001.473887, 45659722.3493377, 238541668.133674, 238541667.684608, 164930556.442226, 164930556.190025, 444791668.556166, 0, 0, 0, 0, 0, 0, -347222.236272397, 0, -347222.233566905, 0, 0, 0, -18055555.6404506, 0, -18055555.5683591, -0.00127441353184902, -2083333.4230762, -2083333.40823328, 0, 0, 0, -85416666.8695922, 0, -85416666.7737023, -347222.269214743, -347222.220318693, -5208333.66057277, -90625000.5410169, -5208333.50357225, -90625000.343523, 0, 0, 0, -168402778.084908, 0, -168402778.057484, -2083333.52043392, -2083333.34532987, -6944444.97871043, -343750001.385574, -6944444.61001425, -343750001.199172, -19791666.671382, -182291668.028234, -182291667.519238, -5208333.75708245, -5208333.41164656, -510416668.495727, -93750000.4472183, -93750000.2585667, -184027779.073724, -520833335.784209, -184027778.571914, -520833335.486715, -354166668.16485, 0, 0, 0, 0, 0, 0, 0.00288709219019688, 0, 0.00271140693663369, 0, 0, 0, 3472222.24308013, 0, 3472222.2236831, 6.93535200648656e-06, 0.0203507257169325, 0.0190512629477104, 0, 0, 0, 20833333.3911541, 0, 20833333.3545142, 0.0133262976966042, 7.1660583587525e-05, 0.0905860662831584, 20833333.4781006, 0.0510866259556146, 20833333.4158319, 0, 0, 0, 52083333.4357451, 0, 52083333.4190364, 0.0557131124348222, 0.00280024835044831, 0.195096939578596, 104166667.119389, 0.0620414940371389, 104166667.033977, 3472222.21083797, 52083333.805762, 52083333.6058535, 0, 0, 69444444.5703734, 0.134327558694134, 0.0159738301935536, 0.213577521295332, 208333334.109142, 0.0401074653372094, 208333334.083991, 20833333.4034375, 20833333.3453238, 69444445.0863193, 208333334.436625, 69444444.7916018, 208333334.215649, 52083333.7086638, 52083333.5298737, 312500001.356407, 104166667.032038, 208333334.353345, 208333334.208896}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 375.000000170514, 0, 0, 0, 0, 0, 0, 0, -9354.16667051447, 0, -4541.66666769329, -791.66666922083, 0, 0, 0, 0, 0, 0, 88805.5555984382, 0, 8333.33332620734, 0, 0, 112451.388909417, 416.666662529946, 18909.7222826768, 8750.00003268485, 0, 0, 0, 0, 0, 0, -423576.389092595, 0, -4166.6666041043, 0, 0, 0, -1055520.83370926, 0, -198611.110939748, 0.000116882974461152, -167465.278264164, -8333.3335343101, -9513.88904636813, -4166.66665649664, -208125.00057309, 0, 0, 0, 0, 0, 0, 1107638.88938923, 0, -0.000178678199440936, 0, 0, 0, 4952777.77964647, 0, 95138.8875722632, -0.000500591203323115, 717013.890641135, 0.000585119893935014, 0, 0, 1752777.77814765, -0.000205082071392244, -0.000867849676209267, 78125.0016013429, 1830902.78165624, 0.000541465510815929, 190277.780251851, 95138.8897739593, 0, 0, 0, 0, 0, 0, -1597222.22290534, 0, 0.00025336010515063, 0, 0, 0, -12656250.0042468, 0, 0.00343324731796792, 0.00101131954742545, -1579861.11453453, -0.000901731671253129, 0, 0, 0, -7465277.78163618, 0, -781249.997573748, 0.00143620811423598, 0.00315785215690472, -295138.894356637, -7760416.67718604, -0.0023332194557655, -0.0043373835468363, -0.00209429704943128, 0.000806602994374367, -1562500.01457043, 0.00370337565716037, -781250.007591608, -0.005326308724651, 0, 0, 0, 0, 0, 0, 1180555.55607293, 0, -0.000183007990484799, 0, 0, 0, 17673611.1160423, 0, -0.00442943128837438, -0.00114077057208191, 1701388.89265746, 0.000836858076954453, 0, 0, 0, 16319444.4528903, 0, -0.00661554654011598, -0.00241793560862387, -0.00480797672999298, 520833.34200599, 16840277.7919316, 0.00388373640760978, 0.00222840443846845, 0, 0, 2951388.88954005, 0.000619140056036704, -0.00399473678101523, 0.00778973992423, 5902777.80738191, 0.00172638226466232, 0.0243287976576176, -0.0122646225751639, 2951388.90625676, 0.0117241664224949, 0.000276935964461999, -0.00395578290563865, 0.0272620628535034, 0, 0, 0, 0, 0, 0, -347222.22242218, 0, 5.57373846983647e-05, 0, 0, 0, -12500000.0028309, 0, 0.00287153507957889, 0.000677000936113779, -694444.446637061, -0.000512303014378492, 0, 0, 0, -17361111.1186392, 0, 0.00626080155553024, 0.00172663757648979, 0.00356192938783342, -347222.228932215, -17708333.3430223, -0.00306799744065817, 0.00144441435673975, 0, 0, 0, -5208333.33789479, 0, 0.00383653836897154, 0.00106025100364436, 0.00448295565137689, -0.00951399434609902, -10416666.69242, -0.00347336861455361, -0.0168962633611874, 0.0124389373163069, -5208333.34923039, -0.00927689353497502, -0.00488966926625428, 0.000438722489908022, -0.0348986405304393, 0.0104767384177522, 0.0152962413269073, -0.0084870020094552, -0.0373336392975837, -0.00107946157596319, -0.039357086179448, -0.00866872903739547, 0, 0, 0, 0, 0, 0, 2.98769817529644e-05, 0, -1.9480135391144e-06, 0, 0, 0, 3472222.22285635, 0, -0.00074584448392598, -0.000166566082570664, 0.000527091661372406, 0.00015012372781926, 0, 0, 0, 6944444.44684849, 0, -0.00211197162318119, -0.000432119299245805, -0.00106485438898994, 0.0020424303229376, 6944444.44710569, 0.00097898154134184, -0.0012380634175537, 0, 0, 0, 3472222.22492967, 0, -0.00236791889183711, -0.000432261080278431, -0.00176299031200226, 0.00392756162528299, 6944444.45278803, 0.0017458040162441, 0.00360704499993993, -0.00437731454843126, 3472222.22750283, 0.00206845288498475, 0, 0, -8.01548513013226e-05, 0.000950198663257363, -0.000706418991291684, 0.00359859626946576, 0.0151763803327642, 0.00100165624831802, 0.0138055282618688, -0.00659304554892487, -0.00867942853570634, 0.00560524094229782, 0.0155808287402916, 0.00200521586043444, 0.017289967235401, -0.000965091145638257, -0.00478149338367682, 0.0270357886108932, -0.00662989929539721, 0.00932723117815162, 0.00954671865265296}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 375.000000049258, 0, 0, 0, 0, 0, 0, 0, -4541.66666699563, 0, -9354.16666826496, -791.666668171364, 0, 0, 0, 0, 0, 0, 8333.33333568122, 0, 88805.5555660499, 0, 0, 112451.388907218, 416.666677025513, 8750.00002387557, 18909.7222497347, 0, 0, 0, 0, 0, 0, -4166.66666665319, 0, -423576.388930773, 0, 0, 0, -198611.111284109, 0, -1055520.83339044, -6.53042570725389e-05, -8333.3335001005, -167465.277900986, -4166.6666971372, -9513.88906323769, -208125.00047014, 0, 0, 0, 0, 0, 0, -4.61860009724214e-05, 0, 1107638.88900202, 0, 0, 0, 95138.8895922676, 0, 4952777.77779224, 0.000224673781367733, 0.00051593245657213, 717013.889075877, 0, 0, 1752777.77840398, -3.25810058337779e-05, 0.000839089410576615, 0.000196107627721578, 190277.780814078, 78125.0007950866, 1830902.77989981, 95138.8900462442, 0, 0, 0, 0, 0, 0, 0.000161555467054031, 0, -1597222.22242535, 0, 0, 0, -0.0014113786414861, 0, -12656249.9997542, -0.000435520431893376, -0.000786708219006706, -1579861.11110826, 0, 0, 0, -781250.002365473, 0, -7465277.7786905, 9.48999590862916e-05, -0.00204821483382246, -0.000691089501569676, -0.00839812482564541, -295138.890205257, -7760416.6705841, 0.000402277033098448, -0.00360881682531314, -1562500.01218196, -0.00221035032747084, -0.00561718228341269, -781250.004704638, 0, 0, 0, 0, 0, 0, -0.000243850093910432, 0, 1180555.55578337, 0, 0, 0, 0.00150908456656189, 0, 17673611.1106003, 0.000499626869252157, 0.000571394429369012, 1701388.88860447, 0, 0, 0, 0.00379209412155263, 0, 16319444.444783, -3.89630631227627e-05, 0.00245460947916228, 0.00112077459358108, 0.011518962371789, 520833.33413086, 16840277.7808874, 0, 0, 2951388.89234427, -0.00109161907359155, 0.0069071092377946, -0.000273483169511285, 0.0264816989280851, 0.0059738677740647, 5902777.79798547, 0.0031514729107101, 0.0123758111472066, 2951388.89412423, 0.00419086553413688, 0.00911876346189682, 0.0196082782391169, 0, 0, 0, 0, 0, 0, 0.000175960873120066, 0, -347222.222363479, 0, 0, 0, -0.000820448086368933, 0, -12499999.9996094, -0.000317163831972113, -0.000132476635562784, -694444.444179375, 0, 0, 0, -0.00283540123674192, 0, -17361111.1106485, -5.81613413619565e-05, -0.00148778249267117, -0.00079162462994602, -0.00774823510633551, -347222.222228497, -17708333.3340036, 0, 0, 0, -0.00393037272655314, 0, -5208333.33565506, 0.000849222186380983, -0.00566007093864271, -0.000263739239605742, -0.0246781094568827, -0.00425470725020201, -10416666.6816229, -0.0020786092507507, -0.0119119645436144, -5208333.33433632, 0.000876477018723112, -0.00837779015943684, -0.0309878809117191, -0.00115381698449631, -0.0109274654212082, -0.00531711499533566, -0.0305871954351224, -0.00850862324601566, -0.0197356262322881, -0.0151294985847247, 0, 0, 0, 0, 0, 0, -4.955057532029e-05, 0, 3.63826395209889e-05, 0, 0, 0, 0.000174918049867829, 0, 3472222.22212069, 8.39924828854386e-05, -2.47216032796253e-05, -7.25598171734183e-05, 0, 0, 0, 0.000826581192123969, 0, 6944444.44411192, 4.59190505228753e-05, 0.00037905555241614, 0.000200210853264578, 0.00204208499684457, -0.000104249275162766, 6944444.44425148, 0, 0, 0, 0.00145442261559786, 0, 3472222.22269363, -0.000273459946106575, 0.00166496063996756, 0.000265248544145934, 0.00845590853322258, 0.00110301861375846, 6944444.4483638, 0.00048915566724209, 0.00415370171150277, 3472222.22142448, 0, 0, 0.00153692369910418, -0.000335878264073486, 0.00360747769100022, -6.82060506055296e-05, 0.0145742495420203, 0.00333451141856498, 0.0125260785184305, -0.000156911879874623, 0.00466126288375896, 0.00268060329578157, 0.0151411328336173, 0.00154742838785973, 0.00593246577630687, -4.39479588665428e-05, 0.00642699223215235, 0.0172906995109427, 0.00582254419998912, 0.0109956759548462, 0.008130309619902}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -4200.00001283561, 0, 0, 0, 0, 0, 0, 0, 90116.6668958019, 0, 69116.6668597635, 27116.6668823791, 0, 0, 0, 0, 0, 0, -686194.446319767, 0, -340611.11229919, 0, 0, -1446805.55828029, -69444.4454065044, -545638.892549603, -410055.558503361, 0, 0, 0, 0, 0, 0, 2485694.45223039, 0, 782638.893102439, 0, 0, 0, 10565694.4643609, 0, 6762638.9005721, 88194.4463014679, 3703750.02444465, 1653472.24008504, 1306250.01764247, 959027.786917876, 8068888.92781441, 0, 0, 0, 0, 0, 0, -4850694.4628051, 0, -937500.009648908, 0, 0, 0, -35992361.1820138, 0, -14607638.9131648, -55555.5570082166, -11135416.7473882, -2868055.61317835, 0, 0, -44811805.6309836, -1545138.93127092, -1104166.67368246, -7774305.66389023, -52586111.3241242, -2979166.71687771, -30760416.8436923, -17697916.7885683, 0, 0, 0, 0, 0, 0, 5256944.47013581, 0, 569444.459167475, 0, 0, 0, 65138889.0254875, 0, 16354166.690937, 13888.8882091601, 16631944.5915223, 2291666.77436102, 0, 0, 0, 129722222.444207, 0, 85625000.1391169, 902777.840998629, 624999.987596983, 18368055.8626908, 148090278.326721, 3472222.36418971, 49965278.1798031, 7881944.67429959, 2361111.15831659, 179131945.212464, 19062500.1386316, 101388889.48665, 51770833.7767526, 0, 0, 0, 0, 0, 0, -2986111.13216989, 0, -138888.903371209, 0, 0, 0, -64722222.3689168, 0, -9236111.11889628, 0.00297540595668726, -12152777.9286523, -694444.563177308, 0, 0, 0, -185069444.768566, 0, -82638889.017315, -208333.398013708, -138888.860804552, -18750000.4439449, -203819445.172303, -1388889.10153777, -37152778.2555831, 0, 0, -196875000.326232, -3819444.71423355, -694444.446655636, -13194444.9633552, -406944445.913448, -1388889.01847969, -251736112.419695, -10069444.4685578, -223263890.133804, -56250000.8053343, -94097222.9007059, -38194444.8120437, -259375001.548399, 0, 0, 0, 0, 0, 0, 694444.453744505, 0, 0.00821989922540558, 0, 0, 0, 33333333.4159462, 0, 2083333.32975081, -0.00304588732796751, 3472222.30390057, 0.0722379319299814, 0, 0, 0, 128472222.455082, 0, 38888888.9401618, 0.0415616510599276, -0.0211993774292877, 6944444.76189854, 135416667.149614, 0.162906228035513, 10416666.9507625, 0, 0, 0, 194444444.767846, 0, 135416666.892448, 694444.622786157, -0.0188234678107872, 6944444.96886787, 395833334.611993, 0.146077306300012, 156250001.039161, 2083333.27934789, 208333334.498213, 20833334.0218957, 3472222.63525268, 0.0493564276667931, 409722223.894714, 41666666.9761, 10416666.6944618, 145833334.413111, 416666668.737327, 20833333.8361904, 236111112.911351, 159722223.290343, 0, 0, 0, 0, 0, 0, -0.00169893413051696, 0, -0.00201762334772452, 0, 0, 0, -6944444.46343551, 0, 0.0021685608208197, 0.00112839336368367, -0.0181736785861905, -0.0187281642616474, 0, 0, 0, -34722222.2880424, 0, -6944444.4490601, -0.0118619068987436, 0.00610982389426342, -0.0890242320314148, -34722222.3492011, -0.0507006783324621, -0.0671259408211846, 0, 0, 0, -69444444.5607312, 0, -34722222.2797005, -0.053803375117954, 0.00755185086858629, -0.194901070135015, -138888889.305994, -0.0597948343751603, -34722222.5300336, 0.0253331640156858, -69444444.8483383, -0.222789162080403, 0, 0, -69444444.5609468, -0.131443031808206, -0.00799520310762666, -0.212818705436729, -208333334.056596, -0.0358072418999423, -138888889.562119, -6944444.48337204, 0.0607929898067573, -69444444.996773, -208333334.27088, -0.238811939136003, -69444445.1895934, -34722222.5378105, -0.0654398577525503, -208333334.50564, -34722222.4121498, -138888889.686618, -69444445.0824077}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 2625.00000600349, 0, 0, 0, 0, 0, 0, 0, -60697.916781767, 0, -38822.9167493412, -12572.9167668155, 0, 0, 0, 0, 0, 0, 515451.389868604, 0, 148177.083761838, 0, 0, 882378.473529913, 22447.9170257691, 275399.307345434, 170625.001351583, 0, 0, 0, 0, 0, 0, -2124045.14298235, 0, -242187.501239289, 0, 0, 0, -7288020.84383503, 0, -3218663.19909249, -17708.3335407108, -2133506.95687186, -502083.34161725, -464670.146660097, -277604.170095494, -3683333.35177139, 0, 0, 0, 0, 0, 0, 4670138.89837659, 0, 182291.669084423, 0, 0, 0, 28820312.5391452, 0, 4992187.50566015, 5208.33174162142, 7579861.15321562, 552083.359971558, 0, 0, 24585069.4810921, 345486.127293565, 197916.665340579, 3250000.05131633, 27835069.5490222, 562500.021747451, 10329861.1951499, 5683159.77221497, 0, 0, 0, 0, 0, 0, -5607638.9016748, 0, -52083.3368381803, 0, 0, 0, -59878472.2990909, 0, -3550347.21873863, 0.00478651920348145, -13177083.4117521, -208333.382618347, 0, 0, 0, -85425347.3409821, 0, -34652777.8336283, -95486.1315385273, -52083.315710987, -9626736.2629437, -95052083.6067888, -312500.065655072, -10746527.9648641, -2152777.87993777, -208333.346970817, -71458333.715129, -3836805.58772919, -38958333.5979165, -10937500.1926403, 0, 0, 0, 0, 0, 0, 3472222.23211512, 0, 0.00363301371025942, 0, 0, 0, 67013888.9713959, 0, 954861.095647015, -0.00657572789950937, 10937500.0821698, 0.0540374023159404, 0, 0, 0, 143750000.183205, 0, 22048611.1445557, 0.0185182411449154, -0.0296809970149638, 11979166.8932706, 155729167.031566, 0.100951735009582, 3819444.65633836, 0, 0, 101302083.490961, 520833.446821285, -0.0138721182924905, 5034722.46886578, 207638889.631816, 0.055450803734825, 66666667.3169912, 954861.064573902, 111371528.344193, 5729167.03054196, 23611111.3625477, 3819444.55048241, 67708334.0515274, 0, 0, 0, 0, 0, 0, -868055.559583578, 0, -0.00227521054205063, 0, 0, 0, -38194444.4901627, 0, 0.0139646863608947, 0.00466357005658649, -3472222.26764837, -0.0330287801693182, 0, 0, 0, -114583333.469035, 0, -5208333.33154809, -0.0116875967109444, 0.0224762644038056, -5208333.50029558, -119791666.910166, -0.078498482745573, -0.116268918564877, 0, 0, 0, -123263889.061083, 0, -51215277.8695026, -0.0717791011833886, 0.0214122568125845, -3472222.4824087, -250000000.650029, -0.0690352617738508, -20833333.8388619, 0.0718646107699851, -130208333.868853, -0.311360600259471, -868055.745596207, -0.0188772190890162, -154513889.743192, -5208333.40080627, 0.0690641266094206, -53819444.8923472, -156250000.977127, -0.181995007499798, -31250000.8743397, -20833333.7503516, 0, 0, 0, 0, 0, 0, 0.000656200362985532, 0, 0.000616936763332801, 0, 0, 0, 8680555.56576239, 0, -0.00417938583078703, -0.00136227259327506, 0.0103226904376327, 0.0086650267178389, 0, 0, 0, 34722222.2610485, 0, -0.00651599754387194, 0.00350260846380382, -0.00684758338481284, 0.0482437459225993, 34722222.2866215, 0.0248129392050185, 0.0244474774686544, 0, 0, 0, 52083333.3991972, 0, 8680555.5728182, 0.0217512557502649, -0.00823792823769239, 0.0999506841463019, 104166666.878954, 0.0295613521875258, 0.143485534618227, -0.0267828816980775, 52083333.5203801, 0.0981535803580621, 0, 0, 34722222.2778406, 0.0586080861158796, 0.00293477865477918, 0.103223435329929, 104166667.039941, 0.0179932080749058, 34722222.5639473, -0.0116996710724091, -0.0620501447160626, 34722222.4616583, 104166667.107621, 0.0953649788357958, 0.367753999136911, 8680555.67072952, -0.00380281762962477, 52083333.9155155, 0.0294302822808263, 34722222.5585478, 0.274052748226658}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -1200.00000142628, 0, 0, 0, 0, 0, 0, 0, 29033.3333638589, 0, 16033.333349205, 4033.33335747636, 0, 0, 0, 0, 0, 0, -263527.778059885, 0, -44833.3333760016, 0, 0, -383361.111415562, -4500.00005183616, -93027.7782569246, -49333.3336482392, 0, 0, 0, 0, 0, 0, 1183611.11232592, 0, 46666.6666349403, 0, 0, 0, 3415972.22517179, 0, 1029027.77833232, 1666.66642156668, 780694.447964938, 95000.0019148496, 98750.0017719658, 50000.0005340963, 1127777.78252632, 0, 0, 0, 0, 0, 0, -2871527.78059787, 0, -16666.6663337758, 0, 0, 0, -14933333.3452412, 0, -1022222.22036947, 0.00145935300161602, -3097222.23451806, -50000.0059235669, 0, 0, -8567361.11969005, -34722.2246369866, -16666.6643896009, -760416.679893371, -9327777.80637406, -50000.004795702, -2079166.68813145, -1091666.67733435, 0, 0, 0, 0, 0, 0, 3798611.11483189, 0, -0.000596700987290383, 0, 0, 0, 34875000.0243412, 0, 347222.213474607, -0.00326000192453009, 6159722.24573792, 0.0103352884266291, 0, 0, 0, 33576388.922199, 0, 7847222.22936392, 0.000466139338950825, -0.0100683150371158, 2604166.70815713, 36180555.6315935, 0.0159127235508518, 1041666.71191353, 243055.578790834, 0.000257908103649945, 15937500.106179, 347222.21861408, 8333333.39751279, 1041666.71292529, 0, 0, 0, 0, 0, 0, -2569444.4471955, 0, 0.000427197477030626, 0, 0, 0, -43819444.4710481, 0, 0.013784165777671, 0.00395642949425794, -5833333.35863124, -0.010796443510528, 0, 0, 0, -65555555.6115941, 0, -2430555.54911877, 0.00188190189424788, 0.0156814757915252, -3958333.39755242, -69513888.9910615, -0.0247659779070196, -0.0448204126454153, 0, 0, -26736111.1472302, -0.0217945076268542, 0.0105880515500814, -694444.507724867, -54166666.8767201, -0.0123729395517218, -7291666.84773247, 0.0363099136005383, -28125000.1415059, -0.0918291598351045, -2430555.59766601, -0.0079027588990669, -7291666.85799631, 0, 0, 0, 0, 0, 0, 694444.445479752, 0, -7.80405177241836e-05, 0, 0, 0, 27777777.7925448, 0, -0.00988573498563477, -0.00254733223662862, 2083333.34768554, 0.00652551804413355, 0, 0, 0, 60416666.7103612, 0, -0.0136148464388564, -0.00157306038459984, -0.0118230052017907, 2083333.38222893, 62500000.06905, 0.0191974182409272, 0.0181115659544044, 0, 0, 0, 40277777.8247711, 0, 6944444.45735122, 0.0108754039772355, -0.0123708980566859, 694444.515435021, 81250000.1836552, 0.0179024816655026, 0.136690858138341, -0.040013975121952, 41666666.8010918, 0.0762619108644782, 0.0454334022280215, 0.00273581079597261, 20833333.57894, -0.0139907325692409, -0.0467589342572886, 6944444.54003374, 20833333.5952734, 0.0332755779758104, 0.250005434139244, 0.0895215093253335, 0, 0, 0, 0, 0, 0, -0.000146768636854931, 0, -2.55405500650656e-05, 0, 0, 0, -6944444.44770087, 0, 0.00271217910157599, 0.000679162544027016, -0.00334915057734255, -0.00174523454347784, 0, 0, 0, -20833333.3462249, 0, 0.00596154519761808, 0.000299707753563899, 0.0036299682986877, -0.014589646458415, -20833333.3518962, -0.00608799714869154, -0.00151292020001285, 0, 0, 0, -20833333.3530784, 0, 0.00085865195380926, -0.00316444896142298, 0.00481079803508915, -0.0283853708793183, -41666666.7265783, -0.00811539084137682, -0.0360799092445741, 0.0142795510299436, -20833333.38044, -0.0218153056175301, 0, 0, -6944444.4568005, -0.0127327916589022, 0.000272794991438838, -0.027179777059188, -20833333.4406862, -0.00506178476865344, -0.0984744923911998, 0.0168633256904515, 0.0297119632081964, -6944444.49904345, -20833333.4491657, -0.0206602847628449, -0.107423476819531, -0.0164807230478903, 0.0115259462432002, -0.169045526830614, 0.0116748748159942, -0.0780852670020405, -0.0698231402538138}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 300.000000630364, 0, 0, 0, 0, 0, 0, 0, -6883.33334589094, 0, -5133.33334300736, -633.33334450143, 0, 0, 0, 0, 0, 0, 57277.7778751871, 0, 24833.3333774705, 0, 0, 117111.111295291, 333.333358985703, 13861.1113014298, 10166.6668475959, 0, 0, 0, 0, 0, 0, -224305.55591084, 0, -36666.6667510587, 0, 0, 0, -965416.668067576, 0, -559027.778532047, 7.75113993352743e-05, -106250.001264729, -41666.6678325521, -6944.44502172948, -5000.00048232344, -221805.557941969, 0, 0, 0, 0, 0, 0, 437500.000674625, 0, 16666.6667116274, 0, 0, 0, 3725694.44936041, 0, 797222.223450257, -0.000470179235997589, 361111.115189047, 33333.3370452577, 0, 0, 4506250.00549229, 0.000354076010550536, -8.30696345259765e-05, 48611.1151360963, 1690972.23515612, 16666.6704279356, 901388.900773311, 104166.672040778, 0, 0, 0, 0, 0, 0, -402777.778423369, 0, 7.77363578876456e-05, 0, 0, 0, -7104166.67578852, 0, -347222.22258468, 0.00100703050342608, -541666.673810724, -0.00660278444879477, 0, 0, 0, -16770833.3499232, 0, -6041666.6761987, 0.000788332241576861, 0.0022787485075114, -138888.901330631, -5694444.47707937, -0.011443654109841, -694444.471915003, -0.00481409799429972, -0.00499545355268876, -6770833.38515235, 0.000667771222206199, -729166.692830598, -347222.250078168, 0, 0, 0, 0, 0, 0, 138888.889104348, 0, -0.000129886360784174, 0, 0, 0, 6319444.45367007, 0, -0.00130158854632877, -0.0011787938021015, 277777.784693553, 0.00687868917924493, 0, 0, 0, 30138888.9132593, 0, 2430555.56322196, -0.00120502990465144, -0.00395248211571759, 138888.90795083, 8402777.81948292, 0.016642788105989, 0.0317870527254719, 0, 0, 20138888.9126882, 0.00234919063883861, 0.00279802011293113, 0.0150816742958183, 22222222.3203536, 0.015681357294275, 4861111.20265034, -0.0141217803756825, 2083333.38826753, 0.0583910350857492, 0.00781915030173098, 0.0175517229952352, 2430555.64662491, 0, 0, 0, 0, 0, 0, 8.33999479845728e-05, 0, 6.60490360364403e-05, 0, 0, 0, -2083333.33810001, 0, 0.00153181372347498, 0.000759886602548064, -0.00344851166110182, -0.003989313998426, 0, 0, 0, -24305555.5729242, 0, -0.00215874791821878, 0.00048650824946556, 0.00303229064688713, -0.0142405615816326, -4166666.69306709, -0.0119667845133085, -0.0177709278569688, 0, 0, 0, -29166666.6910268, 0, -6944444.45980079, -0.000789485957043558, -0.000223859137191636, -0.0185415044170969, -31250000.0837243, -0.0168535357758776, -0.0728101812037799, 0.0147102936271908, -2083333.38459389, -0.0514684676772884, -0.0082120193294578, -0.0112226646751621, -13888889.0061695, 0.0184001565243375, 0.00668052764977519, -0.0220958361887521, -6944444.56288574, -0.0407194920214141, -0.122785525709008, -0.0411598103467258, 0, 0, 0, 0, 0, 0, -5.80956586528341e-05, 0, -9.51419105675673e-06, 0, 0, 0, 0.000972120130320615, 0, -0.000512407655968039, -0.000205251675994109, 0.000686815560377658, 0.0010040098254933, 0, 0, 0, 6944444.44924568, 0, -0.000169133368829275, -2.13705835652575e-05, -0.000938142614044861, 0.00411798604512519, 0.00652242986925017, 0.00348263397813172, 0.00380085190587106, 0, 0, 0, 13888888.8978225, 0, 0.00352985383560443, 0.000550757012641526, -0.000310173637391517, 0.00782142518254001, 13888888.9154778, 0.00626655508999988, 0.0211601525265941, -0.00485810173511526, 0.0175456487444411, 0.0163267037321322, 0, 0, 6944444.45280344, 0.0019863840194751, 0.00340085212730242, 0.0060923678651105, 13888888.9392045, 0.00666988763736689, 0.0478735470408796, -0.0110263295104511, -0.00748614588145249, 0.0142790894415409, 6944444.4967338, 0.0224428375015173, 0.0537500676725006, -0.00486638786160728, 0.00755448779900486, 0.0777861259173706, -0.00587095526642592, 0.0298105043759482, 0.0394507654200954}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 281.250000636402, 0, 0, 0, 0, 0, 0, 0, -5750.00001119698, 0, -5750.00001253619, -593.75000924642, 0, 0, 0, 0, 0, 0, 39322.9167434161, 0, 39322.9167362228, 0, 0, 116927.083552511, 312.500024414342, 11510.4168011602, 11510.4168408128, 0, 0, 0, 0, 0, 0, -111979.166895156, 0, -111979.166861788, 0, 0, 0, -791753.473797667, 0, -791753.473257483, 4.25824819634736e-05, -70833.3341241242, -70833.3344909074, -5729.1671625724, -5729.16709532667, -222482.64101283, 0, 0, 0, 0, 0, 0, 130208.333593796, 0, 130208.33366294, 0, 0, 0, 2214409.72756193, 0, 2214409.7244208, -0.000279713324021419, 161458.335386165, 161458.337207992, 0, 0, 5262152.78458515, 0.000591809901616877, -0.000559918611400107, 31250.0031643107, 1361111.12279877, 31250.0034178414, 1361111.12148211, 105034.727376475, 0, 0, 0, 0, 0, 0, -52083.3332000171, 0, -52083.3337121907, 0, 0, 0, -2491319.45406166, 0, -2491319.44661226, 0.000460740441312456, -104166.668953806, -104166.674131068, 0, 0, 0, -14210069.4640749, 0, -14210069.4576918, -0.000264702898431411, 0.00361984961228939, -52083.3417584021, -3064236.14041886, -52083.3445887042, -3064236.13438099, -0.00496357776937445, -0.00223029058715976, -8229166.71310649, 0.000141249495960984, -572916.69438981, -572916.689949562, 0, 0, 0, 0, 0, 0, -0.000634731152560811, 0, 0.000318701232101233, 0, 0, 0, 954861.120581846, 0, 954861.111731053, -0.000341398373360696, 0.000470029772567629, 0.00845476457503542, 0, 0, 0, 14930555.5834069, 0, 14930555.5682604, 0.000748980345726772, -0.00566581330085406, 0.0107114024571525, 1909722.25828322, 0.0183178602841199, 1909722.24939433, 0, 0, 35763888.9195433, 0.00410163543267066, -0.00427221111473428, 0.0127146398101525, 18055555.6478707, 0.0124467821637135, 18055555.6330048, -0.0133177776472567, 954861.170099523, 954861.1561084, 0.0175514172276697, 0.0101960500732926, 3125000.08559978, 0, 0, 0, 0, 0, 0, 0.000585147907859524, 0, -0.000178976830364773, 0, 0, 0, -0.00476865226670214, 0, 0.000398306646180995, 0.000119237823230113, 0.000967070952728205, -0.00524210503495032, 0, 0, 0, -5208333.35265001, 0, -5208333.3387565, -0.00127652086608347, 0.00374309162303152, -0.00608152768110643, -0.0207634609625614, -0.0145375403552951, -0.016167070388749, 0, 0, 0, -32118055.5854432, 0, -32118055.5774153, -0.00318575514992232, 0.00695238613519658, -0.0133124010350918, -10416666.7477565, -0.0168997514283564, -10416666.7251023, 0.0151236475125471, -0.0525144081296583, -0.0401633422706679, -0.00830600953461462, -0.00377839865617804, -37326388.9942629, 0.00752751697046948, 0.0168772047263433, -0.0364465234430069, -5208333.45710992, -0.0264021748140255, -5208333.43133485, -0.0435951348081818, 0, 0, 0, 0, 0, 0, -0.000181898977241188, 0, 4.69389068980293e-05, 0, 0, 0, 0.000941846071685122, 0, -0.000221589877708379, -2.31287043343371e-05, -0.000540763841043424, 0.00137820817612893, 0, 0, 0, 0.00524754770024022, 0, 0.000661813340604287, 0.000563572486691846, -0.000890133538030525, 0.00115110347245005, 0.00437361424992607, 0.00457035660988476, 0.00392882960101989, 0, 0, 0, 8680555.56607011, 0, 8680555.56123812, 0.00138904430082326, -0.00271379159253428, 0.00495331594103639, 0.0258272034216927, 0.00720492665414064, 0.0163949863028905, -0.00471402677870068, 0.0163252510309345, 0.013580090983748, 0, 0, 17361111.1222597, 0.00254055925806903, -0.000851117589123645, 0.00498891619558926, 17361111.1587322, 0.00499819577222395, 17361111.1516946, -0.00750845067613937, -0.0123501431028961, 0.0197054904066772, 0.0576978005964352, 0.0158111066000835, 0.0380160762557632, 0.00330174985260487, -0.000835544681180114, 8680555.62551837, -0.00258012209634758, 0.0415164674444008, 0.0316519876158445}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 300.000000471191, 0, 0, 0, 0, 0, 0, 0, -5133.33334211236, 0, -6883.33334266469, -633.333338089319, 0, 0, 0, 0, 0, 0, 24833.3333942892, 0, 57277.7778282362, 0, 0, 117111.1112864, 333.333343518953, 10166.6667410016, 13861.1112054291, 0, 0, 0, 0, 0, 0, -36666.6668483421, 0, -224305.55568276, 0, 0, 0, -559027.779046363, 0, -965416.667529737, 1.54405955380064e-05, -41666.6671178448, -106250.000557342, -5000.00013347217, -6944.44465844341, -221805.556973398, 0, 0, 0, 0, 0, 0, 16666.6668843458, 0, 437500.000164936, 0, 0, 0, 797222.226426895, 0, 3725694.44631244, -5.12638846550953e-05, 33333.3344625455, 361111.112719636, 0, 0, 4506250.00583737, -0.00017249808320071, -0.000270500892067453, 16666.6677058622, 901388.897311287, 48611.1127321482, 1690972.22894467, 104166.66926531, 0, 0, 0, 0, 0, 0, 4.76653227157127e-05, 0, -402777.777890592, 0, 0, 0, -347222.229544592, 0, -7104166.66840721, -6.61438352589213e-05, -0.00106259797665647, -541666.669288531, 0, 0, 0, -6041666.68323522, 0, -16770833.3453913, 0.000639215018725677, 0.00192854916292866, -0.00287837123813572, -694444.466604729, -138888.894155287, -5694444.45846576, -0.000562687695421093, -0.00127869263972563, -6770833.36723804, 0.00145660538580809, -347222.238473871, -729166.677863887, 0, 0, 0, 0, 0, 0, -0.000390347802147052, 0, 138888.88894183, 0, 0, 0, 0.00697219994450776, 0, 6319444.44462914, 0.000282138895880741, -0.000237578339317914, 277777.780281137, 0, 0, 0, 2430555.57825576, 0, 30138888.9009773, -0.000324899486773232, -0.00287368560405024, 0.00311630423337405, 0.0281617300562952, 138888.897206759, 8402777.79238991, 0, 0, 20138888.9167203, -0.000537209942453554, -0.00230309491003385, 0.00267252630802861, 4861111.18217279, 0.00701932729746248, 22222222.2779758, -0.00941839970806358, 0.0371236353091137, 2083333.35262264, 0.00790774635917396, 0.00071597813162043, 2430555.60644283, 0, 0, 0, 0, 0, 0, 0.000368608277058698, 0, -3.21838408158104e-05, 0, 0, 0, -0.00341728936868637, 0, -2083333.33264261, -0.000273979090615932, 0.00100891867646372, -0.00133617610595435, 0, 0, 0, -0.015110527025556, 0, -24305555.5608159, -0.000367463876509386, 0.00174057586054158, -0.00114319279220036, -0.0170101600712054, -0.00642729729913286, -4166666.6739072, 0, 0, 0, -6944444.47079498, 0, -29166666.6878151, 0.000383354827061324, 0.00402169175035834, -0.00316169745743409, -0.0645059998122306, -0.00982055526900943, -31250000.0413219, 0.00915219001788774, -0.0341553255713641, -2083333.34876818, -0.000834409141701609, -0.002768079597536, -13888888.9702975, 0.00637139187273872, 0.0129943404192357, -0.0199854337438212, -0.0817595412569976, -0.00522614890716627, -6944444.49840057, -0.0229277882446314, 0, 0, 0, 0, 0, 0, -0.000114565143265205, 0, 1.27591952815918e-05, 0, 0, 0, 0.000663963027490086, 0, -0.000306199809644444, 8.93369188191276e-05, -0.000454112892183134, 0.000310673493797344, 0, 0, 0, 0.00394743456390503, 0, 6944444.44505295, 0.000286502656104986, -0.000382769527052524, -6.28844855970725e-05, 0.00381165175520658, 0.00191754040363687, 0.00128427860272904, 0, 0, 0, 0.008892410761365, 0, 13888888.8946662, 0.000214876883136104, -0.00160126711266311, 0.00112881217763026, 0.021179112830174, 0.00435612889430371, 13888888.9001668, -0.00288169987069208, 0.010725286034217, 0.00496282346433344, 0, 0, 6944444.45502347, 0.000126288813673148, -0.000417610492916163, 0.000999106111085544, 0.0381688797319724, 0.00342398065066203, 13888888.9206186, -0.00467453750217681, -0.00812010766503016, 0.0106739485640776, 0.0405529410986938, 0.00429298463045402, 6944444.46357092, 0.00072619978573773, -0.00399750573023925, 0.04515013361301, -0.0029932374446017, 0.025093059138779, 0.0121365137898116}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -1200.00000059584, 0, 0, 0, 0, 0, 0, 0, 16033.3333412294, 0, 29033.3333472475, 4033.33334065453, 0, 0, 0, 0, 0, 0, -44833.3333846179, 0, -263527.777875417, 0, 0, -383361.111294399, -4500.00001795278, -49333.333454473, -93027.7779300228, 0, 0, 0, 0, 0, 0, 46666.6668153491, 0, 1183611.11148861, 0, 0, 0, 1029027.7790316, 0, 3415972.22319306, 1666.66670910738, 95000.0007579646, 780694.445283488, 50000.0000740439, 98750.0004948545, 1127777.78030442, 0, 0, 0, 0, 0, 0, -16666.6668286836, 0, -2871527.77867034, 0, 0, 0, -1022222.22643983, 0, -14933333.335931, -0.000245939529188481, -50000.0021086636, -3097222.22431186, 0, 0, -8567361.11685257, -16666.6656138196, -34722.2231971609, -50000.0007694282, -2079166.68144685, -760416.669571989, -9327777.79018833, -1091666.67213349, 0, 0, 0, 0, 0, 0, -0.000113032659257032, 0, 3798611.11245555, 0, 0, 0, 347222.229998175, 0, 34875000.003656, 0.0008292247595591, 0.00275061451060033, 6159722.22489975, 0, 0, 0, 7847222.23775243, 0, 33576388.902101, -0.00273366519735381, 0.00130510361266278, 0.00271437573817823, 1041666.70567719, 2604166.6733298, 36180555.5825795, -0.00369527736385355, 243055.562468273, 15937500.0607608, 347222.224182961, 1041666.6952652, 8333333.35927586, 0, 0, 0, 0, 0, 0, 0.000471106630817902, 0, -2569444.445726, 0, 0, 0, -0.00816429218293337, 0, -43819444.4469476, -0.00133797298408773, -0.00130842914360918, -5833333.33505999, 0, 0, 0, -2430555.57646681, 0, -65555555.5712536, 0.00251736158189419, -0.00138970328411336, -0.00371393815555598, -0.0517270867657728, -3958333.34055311, -69513888.9184287, 0, 0, -26736111.1389382, 0.00867990987658286, -0.00887005946134251, 0.00416922874725424, -7291666.79014441, -694444.459464072, -54166666.7719362, 0.00681988548487825, -0.0636196215208979, -28125000.0425482, -0.010665877304697, -2430555.58011892, -7291666.7679705, 0, 0, 0, 0, 0, 0, -0.000443535640035614, 0, 694444.445149729, 0, 0, 0, 0.00456708236771901, 0, 27777777.7784463, 0.00102838283272393, -0.000356280229392234, 2083333.33381725, 0, 0, 0, 0.0142232085398387, 0, 60416666.6751628, -0.000654045485540887, 0.00118580729519739, 0.00191002453997988, 0.0337125397563646, 2083333.33699817, 62500000.0152505, 0, 0, 0, 6944444.46891048, 0, 40277777.8020316, -0.0064252853750381, 0.00555895975007331, -0.00140251163600288, 0.111531779559302, 694444.457887314, 81250000.0842156, -0.00843844997459754, 0.0612487756272808, 41666666.6967387, -0.00814608893556697, 0.0151123703708952, 20833333.4786494, -0.014331306723385, 0.0105232091343737, 0.0224293160651727, 0.15217807127471, 6944444.47792173, 20833333.4464551, 0.058774272413896, 0, 0, 0, 0, 0, 0, 0.00014274853034424, 0, -0.000168825133820559, 0, 0, 0, -0.00104249160982256, 0, -6944444.44446266, -0.000303878960826484, 0.000378215186160214, -3.73879878749313e-05, 0, 0, 0, -0.00396774938542077, 0, -20833333.3348394, -0.000144701327799239, -0.000470933006677463, -0.000177948094921254, -0.00851964143439766, -0.000647035752088539, -20833333.3360927, 0, 0, 0, -0.00779450777170818, 0, -20833333.3409968, 0.00150522588831765, -0.00138195445937038, -0.0001230500106521, -0.0373865270583679, -0.00433009811766896, -41666666.6917426, 0.00296089557519074, -0.0209877002433022, -20833333.3410333, 0, 0, -6944444.45551066, 0.00310612779482017, -0.00517817538845071, 0.00213047232531546, -0.0654953937182072, -0.00749223290778933, -20833333.3934737, 0.00942400285570213, -0.00107831759179871, -0.0132955642425855, -0.0751618305346566, -6944444.45739018, -20833333.3753689, 0.00675167456692292, -0.0119833694902275, -0.0868664736102857, -0.00781697546287, -0.0475529041428925, -0.0370771469539534}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 2625.00000273504, 0, 0, 0, 0, 0, 0, 0, -38822.9167068032, 0, -60697.9167217218, -12572.9167084756, 0, 0, 0, 0, 0, 0, 148177.083619774, 0, 515451.389286405, 0, 0, 882378.472920678, 22447.9167979621, 170625.000681893, 275399.306296357, 0, 0, 0, 0, 0, 0, -242187.501063744, 0, -2124045.14048571, 0, 0, 0, -3218663.19893941, 0, -7288020.8371581, -17708.3333877618, -502083.33749901, -2133506.94890972, -277604.168832005, -464670.141148008, -3683333.34407925, 0, 0, 0, 0, 0, 0, 182291.668937318, 0, 4670138.89276091, 0, 0, 0, 4992187.51473852, 0, 28820312.5112453, 5208.33324184084, 552083.345626439, 7579861.12460716, 0, 0, 24585069.4647393, 197916.667731497, 345486.112232394, 562500.013246316, 10329861.1686, 3250000.01357455, 27835069.4972254, 5683159.75341564, 0, 0, 0, 0, 0, 0, -52083.3361447639, 0, -5607638.89476623, 0, 0, 0, -3550347.2499126, 0, -59878472.2403959, -0.00072650033213848, -208333.352555493, -13177083.3563707, 0, 0, 0, -34652777.8282225, 0, -85425347.2718555, -52083.3297011735, -95486.1059248249, -312500.036950554, -10746527.923103, -9626736.14737446, -95052083.4568792, -208333.344892997, -2152777.79324574, -71458333.5594127, -3836805.57847383, -10937500.1477453, -38958333.4689444, 0, 0, 0, 0, 0, 0, 0.00185407422398667, 0, 3472222.22777545, 0, 0, 0, 954861.141706595, 0, 67013888.9045033, 0.00213731251972955, 0.0157038418527756, 10937500.0226963, 0, 0, 0, 22048611.1762358, 0, 143750000.064787, -0.00406886301888019, -0.0083847895970632, 0.0509084527099725, 3819444.63414287, 11979166.7156879, 155729166.818224, 0, 0, 101302083.425857, -0.00309479291760186, 520833.333744726, 0.0352009523306433, 66666667.0942016, 5034722.26530913, 207638889.287541, 954861.086112507, 5729166.97336655, 111371528.025642, 3819444.55626246, 23611111.2375265, 67708333.7917875, 0, 0, 0, 0, 0, 0, -0.000490357992514785, 0, -868055.55855799, 0, 0, 0, -0.0183406377551593, 0, -38194444.4510557, -0.00197603299354338, -0.00580146688776304, -3472222.23446143, 0, 0, 0, -5208333.37760382, 0, -114583333.373647, -0.000229933546486384, 0.0037253271426429, -0.0332026001907616, -0.123540348908867, -5208333.36627595, -119791666.759291, 0, 0, 0, -51215277.8524881, 0, -123263888.974511, 0.00584076577960546, 0.00938898822092689, -0.0416336520368325, -20833333.7049376, -3472222.26779098, -250000000.332964, 0.0415063769519385, -0.287021193449635, -130208333.542708, -0.00893966387525475, -868055.574235074, -154513889.386902, 0.0126992880791338, -5208333.35572399, -0.183467145006818, -31250000.6232369, -53819444.6333254, -156250000.530971, -20833333.6415796, 0, 0, 0, 0, 0, 0, -8.18561646651197e-06, 0, 0.000706186117569482, 0, 0, 0, 0.00454010469181571, 0, 8680555.55664876, 0.0006152888390355, 0.000551368806838954, 0.00285482137797476, 0, 0, 0, 0.0127305980700475, 0, 34722222.2313686, 0.0011038295473479, -0.000102801197031534, 0.00801558399294347, 0.031613235789321, 0.00875225629525282, 34722222.2442361, 0, 0, 0, 8680555.57818313, 0, 52083333.3626292, -0.000519016136735884, -0.00380370568270163, 0.0167563163505211, 0.121864429873116, 0.0166783707670643, 104166666.772137, -0.0148004069800733, 0.0981334778210814, 52083333.4004596, 0, 0, 34722222.2572861, -2.71903729129121e-05, 0.000618636953840829, 0.0100252637462752, 34722222.434485, 0.0137492818647126, 104166666.875569, -0.0288731605246725, -0.0156409874964615, 0.0971692463760181, 0.292101034548006, 34722222.3106489, 104166666.88242, 0.011011198041197, 8680555.59797676, 52083333.6840684, 0.0455292144726411, 0.22457866523238, 34722222.4227388}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -4200.00000879494, 0, 0, 0, 0, 0, 0, 0, 69116.6668063889, 0, 90116.6668212143, 27116.6668150105, 0, 0, 0, 0, 0, 0, -340611.112179575, 0, -686194.445553822, 0, 0, -1446805.55748586, -69444.4450698834, -410055.557943968, -545638.891202271, 0, 0, 0, 0, 0, 0, 782638.893253182, 0, 2485694.44891928, 0, 0, 0, 6762638.90124991, 0, 10565694.4547454, 88194.4454993528, 1653472.23715849, 3703750.01413313, 959027.788103394, 1306250.00797939, 8068888.91966895, 0, 0, 0, 0, 0, 0, -937500.010543409, 0, -4850694.45539253, 0, 0, 0, -14607638.9297468, 0, -35992361.1411342, -55555.5566118151, -2868055.60238931, -11135416.7115019, 0, 0, -44811805.6091109, -1104166.68559997, -1545138.89752955, -2979166.72642039, -30760416.8262256, -7774305.59980371, -52586111.2590032, -17697916.7691362, 0, 0, 0, 0, 0, 0, 569444.459915495, 0, 5256944.46122962, 0, 0, 0, 16354166.7442713, 0, 65138888.9378451, 13888.8907058704, 2291666.74807552, 16631944.5258871, 0, 0, 0, 85625000.1335278, 0, 129722222.349561, 625000.01982856, 902777.774457274, 3472222.38564574, 49965278.176564, 18368055.673757, 148090278.128689, 2361111.2119036, 7881944.49861734, 179131945.047094, 19062500.1184119, 51770833.7933713, 101388889.301993, 0, 0, 0, 0, 0, 0, -138888.902284322, 0, -2986111.12705818, 0, 0, 0, -9236111.197517, 0, -64722222.2659127, -0.00305594027437975, -694444.523860227, -12152777.8639041, 0, 0, 0, -82638889.0637002, 0, -185069444.607727, -138888.907364151, -208333.321902621, -1388889.11820424, -37152778.2992779, -18750000.165102, -203819444.884842, 0, 0, -196875000.23596, -694444.534853337, -3819444.46478203, -1388889.11819625, -251736112.218995, -13194444.5734636, -406944445.506857, -10069444.4594379, -56250000.9241417, -223263889.64902, -38194444.9121209, -94097222.6576937, -259375001.315741, 0, 0, 0, 0, 0, 0, 0.006222939034398, 0, 694444.453057396, 0, 0, 0, 2083333.38545266, 0, 33333333.3533098, 0.0023593898725628, 0.0405227024566158, 3472222.27204198, 0, 0, 0, 38888889.0088792, 0, 128472222.324139, 0.0152037581913838, -0.00514091669759923, 0.15837174806686, 10416667.0105485, 6944444.56147549, 135416666.944913, 0, 0, 0, 135416666.860742, 0, 194444444.655027, 0.0528219686922806, 694444.438609234, 0.234569978052413, 156250000.946736, 6944444.57440487, 395833334.223884, 2083333.2794085, 20833334.1873694, 208333333.988, 0.149128184758084, 3472222.27960768, 409722223.504543, 10416666.8007433, 41666666.8124713, 20833334.0282223, 236111112.810126, 145833333.93987, 416666668.205872, 159722223.206528, 0, 0, 0, 0, 0, 0, -0.00118192197750467, 0, -0.00201424687880566, 0, 0, 0, -0.0130334332767157, 0, -6944444.44818412, -0.000626327711252895, -0.00837839917296068, -0.0122978147295268, 0, 0, 0, -6944444.47885239, 0, -34722222.2460386, -0.00563212947618631, -0.000105391020580218, -0.0424911566267515, -0.0901623882476521, -0.0334689933595512, -34722222.2917979, 0, 0, 0, -34722222.2828116, 0, -69444444.5151201, -0.0174716294350287, 0.00298995743518271, -0.0877592958617151, -34722222.5295263, -0.0475645451834402, -138888889.172357, 0.0238149031776311, -0.293589081980751, -69444444.6587788, 0, 0, -69444444.5311592, -0.0425524856712845, -0.00753477282625209, -0.0820704634063145, -138888889.424125, -0.0342129033472216, -208333333.869855, 0.0190626639582018, -6944444.43299292, -0.350125271631632, -69444445.2127836, -69444444.7238864, -208333333.973733, -0.137705332010505, -34722222.3721769, -208333334.29066, -34722222.4230161, -69444445.1113298, -138888889.512316}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 2333.33334077936, 0, 0, 0, 0, 0, 0, 0, -46175.9260506336, 0, -46175.9260551469, -11175.9260492461, 0, 0, 0, 0, 0, 0, 304259.260194956, 0, 304259.260102342, 0, 0, 900185.187012727, 19953.7042060504, 207546.298272067, 207546.298209194, 0, 0, 0, 0, 0, 0, -873842.596158315, 0, -873842.595678563, 0, 0, 0, -5768518.53090543, 0, -5768518.52749419, -15740.7414955742, -1204629.64206623, -1204629.64157532, -346527.786534896, -346527.783809282, -3806250.02468472, 0, 0, 0, 0, 0, 0, 1238425.93357885, 0, 1238425.93301494, 0, 0, 0, 15829861.1528198, 0, 15829861.1335886, 4629.62989022109, 2722222.26077728, 2722222.26102491, 0, 0, 35016203.7556345, 254629.647579548, 254629.633974718, 1733796.34850341, 21537037.1670643, 1733796.3320939, 21537037.1548856, 5961805.63127951, 0, 0, 0, 0, 0, 0, -856481.491000419, 0, -856481.492057807, 0, 0, 0, -21215277.8550507, 0, -21215277.8088313, 0.000305159249374292, -2638888.95387806, -2638888.9612928, 0, 0, 0, -87303240.8821481, 0, -87303240.8487777, -69444.4678368556, -69444.4348333871, -2777777.92172605, -46388889.2147797, -2777777.88038772, -46388889.1648057, -1064814.91327523, -1064814.85376353, -115439815.305495, -4097222.2975978, -29201389.2452995, -29201389.1913241, 0, 0, 0, 0, 0, 0, 231481.488071516, 0, 231481.491503011, 0, 0, 0, 13773148.228926, 0, 13773148.1715057, -0.000277573082913775, 925925.986281347, 925926.004941479, 0, 0, 0, 102893518.714329, 0, 102893518.63674, 0.0251263211645846, -0.019770569985222, 1388889.09046259, 42361111.5306892, 1388889.04108227, 42361111.4521499, 0, 0, 178703703.929656, 231481.58522869, 231481.487628285, 925926.149735079, 222222223.142925, 925926.036991169, 222222223.071266, 1041666.65524297, 44444445.1720218, 44444445.0114475, 16898148.505575, 16898148.4138178, 135995371.371303, 0, 0, 0, 0, 0, 0, -0.00219750474449203, 0, -0.00546186000946233, 0, 0, 0, -3472222.26683724, 0, -3472222.23105742, 0.000196836234290197, -0.0286870100947851, -0.0471934123390445, 0, 0, 0, -56712963.0988341, 0, -56712963.0259778, -0.0190923534723127, 0.0130326754523845, -0.138422282083125, -13888889.156945, -0.113597722911783, -13888889.1025953, 0, 0, 0, -152777777.985482, 0, -152777777.951371, -0.06834295585664, 0.0146859499080648, -0.227115712829822, -173611111.902467, -0.124022785554531, -173611111.800078, 0.0550867398220181, -20833334.0006641, -20833333.8304625, -0.163219647772067, -0.0505224158570628, -319444445.502312, -3472222.3294329, -3472222.25170179, -13888889.4540775, -180555556.878891, -13888889.2883818, -180555556.730415, -67129630.3132943, 0, 0, 0, 0, 0, 0, 0.000220769649549904, 0, 0.00129231576320093, 0, 0, 0, 0.0100796596488326, 0, 0.00133403674276548, -0.000122704603653539, 0.00534926571319698, 0.0119701700318846, 0, 0, 0, 11574074.1117455, 0, 11574074.0864762, 0.00625379060418107, -0.00298067906101224, 0.0368905546236788, 0.0671806200004971, 0.034063300524422, 0.0537368172411966, 0, 0, 0, 46296296.3670125, 0, 46296296.3466996, 0.0224104426417286, -0.00650192794217822, 0.0843888530608492, 46296296.5508892, 0.0486590548580316, 46296296.5078744, -0.0209061328364881, 0.22512253411952, 0.16553017078494, 0, 0, 69444444.5254235, 0.0503612084069921, 0.00728310657859987, 0.0865167046191327, 138888889.341191, 0.036011489982295, 138888889.318273, -0.00915389931285609, -0.0400270941704457, 0.287573246123652, 69444445.0425188, 0.197039718864446, 69444444.9317786, 0.128595974665704, 0.0715728782531238, 138888889.640717, 11574074.1827824, 46296296.7957506, 46296296.7324335}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -1000.00000309129, 0, 0, 0, 0, 0, 0, 0, 21694.4444993769, 0, 18361.1111628754, 3361.11116235359, 0, 0, 0, 0, 0, 0, -165370.37078596, 0, -104166.666972844, 0, 0, -394537.037847738, -3750.00017403053, -69120.3711967135, -57916.6674807044, 0, 0, 0, 0, 0, 0, 572916.668216548, 0, 225694.445435753, 0, 0, 0, 2958333.33907481, 0, 2194444.44817477, 1388.88895962918, 477777.783073065, 284722.227434571, 72916.6699316097, 60416.6690259672, 1182638.89920009, 0, 0, 0, 0, 0, 0, -960648.15129254, 0, -208333.335368869, 0, 0, 0, -9980324.09380046, 0, -4575231.48967891, 0.000720051123103061, -1386574.09069644, -437500.017063441, 0, 0, -15886574.0974614, -25462.9685365366, -20833.3341021497, -451388.909287966, -8069444.49933805, -250000.015710891, -5719907.45748437, -1170138.91694391, 0, 0, 0, 0, 0, 0, 763888.892395621, 0, 69444.4472577152, 0, 0, 0, 16076388.9254789, 0, 4027777.78690975, -0.00182240155980831, 1666666.69493287, 208333.365214665, 0, 0, 0, 50486111.1780932, 0, 30798611.1559417, 0.00579325409306245, -0.00672425658713315, 1041666.72480293, 22881944.5824106, 208333.380190928, 8437500.11680519, 138888.923846029, 69444.4620349599, 37847222.4340113, 381944.461732016, 7187500.13494582, 4791666.78861043, 0, 0, 0, 0, 0, 0, -231481.483425093, 0, -0.00258154865406287, 0, 0, 0, -12152777.8153582, 0, -1273148.15263749, 0.00218499074708872, -694444.470896516, -0.0349004493132534, 0, 0, 0, -73726851.9477371, 0, -24652777.8212395, -0.00612678722027562, 0.0123944915053329, -694444.528208792, -26388889.065654, -0.0703790825493148, -3819444.58604474, 0, 0, -85185185.2865047, -0.0339539310977549, -0.00466829841122602, -231481.566799901, -101273148.549183, -0.0545603573915651, -51388889.2562984, 0.0269785841206005, -16319444.7230907, -3819444.68343233, -2083333.43961598, -1273148.2388318, -28819444.849714, 0, 0, 0, 0, 0, 0, 0.000324199882041781, 0, 0.00141147878667485, 0, 0, 0, 3472222.2422629, 0, 0.000237360368156618, -0.00144427749598459, 0.0126705343335934, 0.020979343401242, 0, 0, 0, 48611111.1786691, 0, 6944444.46393972, 0.00532584898995921, -0.00902328083116795, 0.0591007999141, 10416666.7783886, 0.0527899371810232, 0.0863633168662862, 0, 0, 0, 93750000.098801, 0, 55555555.6268118, 0.0223534961944446, -0.00469306355987439, 0.0910518054094008, 104166667.011679, 0.0613246086408699, 20833333.6266898, -0.0390597396623116, 10416666.922637, 0.211646885416674, 0.0592971501321851, 0.0304277637883634, 114583333.79938, 0.000528423072869544, -0.0114798708488145, 3472222.40790293, 62500000.5372657, 0.161945215888571, 20833333.8315526, 6944444.68221216, 0, 0, 0, 0, 0, 0, 7.09551140462725e-05, 0, -0.000341474470285713, 0, 0, 0, -0.00430146190458075, 0, 0.000340036669602188, 0.000419538693564135, -0.00238326359252729, -0.00537004536612991, 0, 0, 0, -11574074.0927579, 0, -0.00280681830045014, -0.00189216026360725, 0.00251304764928374, -0.0161971895683346, -0.0275699818402971, -0.0159520341794385, -0.0211221601519249, 0, 0, 0, -34722222.2573633, 0, -11574074.0926594, -0.00783590853077886, 0.00261819165666725, -0.0351907849298848, -34722222.3326525, -0.0238822185649918, -0.0875427709513962, 0.0133364553511349, -0.0861655640839991, -0.0699101848174997, 0, 0, -34722222.2581583, -0.0179632896903876, -0.00705850851111717, -0.0336603948118344, -69444444.645621, -0.0203893222784689, -34722222.4097812, 0.0189672802386633, 0.0254338524576366, -0.0992561355905328, -34722222.4634103, -0.0846269000537392, -0.209861295684342, -0.0285033546159212, -0.0256493574057135, -34722222.5428094, -0.0161719290757382, -11574074.2534694, -0.172709264942561}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -1000.00000208835, 0, 0, 0, 0, 0, 0, 0, 18361.1111462225, 0, 21694.444484991, 3361.11114246617, 0, 0, 0, 0, 0, 0, -104166.666912751, 0, -165370.370620243, 0, 0, -394537.037678355, -3750.00010825816, -57916.6671594488, -69120.3709124795, 0, 0, 0, 0, 0, 0, 225694.445250926, 0, 572916.667497543, 0, 0, 0, 2194444.44889563, 0, 2958333.33651787, 1388.88894424866, 284722.225197099, 477777.781149397, 60416.6686214552, 72916.6681723834, 1182638.89622292, 0, 0, 0, 0, 0, 0, -208333.334631367, 0, -960648.149872419, 0, 0, 0, -4575231.49628736, 0, -9980324.08169612, 0.000233593092941661, -437500.008318326, -1386574.08480004, 0, 0, -15886574.0936995, -20833.3364105456, -25462.962509265, -250000.011992929, -5719907.44736359, -451388.898748678, -8069444.47952991, -1170138.90915633, 0, 0, 0, 0, 0, 0, 69444.4452363069, 0, 763888.891269989, 0, 0, 0, 4027777.80441809, 0, 16076388.898332, -0.000150337099640616, 208333.344821752, 1666666.68628062, 0, 0, 0, 30798611.1651283, 0, 50486111.1522603, 0.00284563844033107, -0.00784254823066102, 208333.36561654, 8437500.10077234, 1041666.69667695, 22881944.5239961, 69444.4645326755, 138888.896275741, 37847222.3769309, 381944.456504909, 4791666.7699137, 7187500.08255874, 0, 0, 0, 0, 0, 0, 0.000415473833556693, 0, -231481.48366592, 0, 0, 0, -1273148.17494151, 0, -12152777.7833421, -0.000446959526980342, -0.0071010334704145, -694444.465467329, 0, 0, 0, -24652777.8518223, 0, -73726851.8960631, -0.00383015949941738, 0.0123626961615139, -0.0422432499419563, -3819444.57174275, -694444.490693634, -26388888.983685, 0, 0, -85185185.2740266, -0.0180053472252288, 0.0101893727898332, -0.0479875579698359, -51388889.1900216, -231481.513533903, -101273148.410108, 0.0231647383721617, -3819444.66126603, -16319444.5961998, -1273148.23826579, -2083333.38824898, -28819444.7425337, 0, 0, 0, 0, 0, 0, -0.000804843795305801, 0, 0.00120201683942501, 0, 0, 0, 0.0141263640391388, 0, 3472222.22324942, 0.000577937275246439, 0.000790566180756794, 0.0123768603684401, 0, 0, 0, 6944444.49487793, 0, 48611111.1334709, 0.00444794299346289, -0.00753537434777161, 0.0259091433376971, 0.0778795075824805, 0.0352152194359414, 10416666.7238148, 0, 0, 0, 55555555.6373172, 0, 93750000.0689524, 0.0116926013220738, -0.0175097358604491, 0.0488985401975985, 20833333.5966007, 0.0417770987354676, 104166666.87317, -0.0326231151686025, 0.197112357817838, 10416666.7984739, 0.0318798334298642, 0.0083480683359421, 114583333.676763, 0.00928528608789212, -0.024841618362315, 0.156265708810674, 20833333.7538679, 3472222.31548833, 62500000.3344235, 6944444.63083352, 0, 0, 0, 0, 0, 0, 0.000302102509084626, 0, -0.000293858449253455, 0, 0, 0, -0.0030119171730181, 0, 0.00013622850696829, -0.000187596581460994, 0.000628866079911056, -0.00311005341561187, 0, 0, 0, -0.0137015982669288, 0, -11574074.0779796, -0.00185732175206667, 0.00156691456789498, -0.00578635780628046, -0.018160988838212, -0.0105848712027632, -0.0137512525872319, 0, 0, 0, -11574074.101555, 0, -34722222.2421031, -0.0045077815208652, 0.00663091457186275, -0.0178495522518482, -0.0848660693069341, -0.0176333840225536, -34722222.2838136, 0.0108731929782648, -0.0640856658533715, -0.0440916470296241, 0, 0, -34722222.2548797, -0.00929556627605807, 0.00313227504039487, -0.0179968910987991, -34722222.3738028, -0.0116108184342655, -69444444.5812212, 0.0121511264673166, 0.0276568676999331, -0.0801208253924115, -0.196441747115004, -0.0500314529723661, -34722222.3540401, -0.0282356430092494, -0.00503033552648959, -34722222.4542013, -0.017121686892231, -0.154326307045194, -11574074.1884446}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -5.6114262220361e-06, 0, 0, 0, 0, 0, 0, 0, 8.46986307955893e-05, 0, 7.5459029913829e-05, 120000.000214705, 0, 0, 0, 0, 0, 0, -0.00079909790996952, 0, -0.000687403483688884, 0, 0, -0.000427349540716606, -1314761.90696255, -1314761.90797991, -1314761.90699794, 0, 0, 0, 0, 0, 0, 0.00411806652324516, 0, 0.00308034173549192, 0, 0, 0, 0.00234561380016765, 0, 0.00364008178072096, 6070555.56528794, 6070555.57573802, 6070555.56454967, 12141111.1378961, 12141111.1320981, 12141111.1385981, 0, 0, 0, 0, 0, 0, -0.0127257485719169, 0, -0.00767372155798529, 0, 0, 0, -0.00588352125303251, 0, -0.0130510640181099, -15322222.2453156, -15322222.2868, -15322222.2419779, 0, 0, -0.0193776762624941, -45966666.7597625, -45966666.7513544, -45966666.7945181, -45966666.8175127, -45966666.735166, -45966666.7514743, -91933333.5430961, 0, 0, 0, 0, 0, 0, 0.0238385646099456, 0, 0.0117201491990559, 0, 0, 0, 0.00990722109719392, 0, 0.018544500239029, 22847222.2531859, 22847222.3405177, 22847222.2513441, 0, 0, 0, 0.0382052563396215, 0, 0.0702551608509949, 91388889.0549026, 91388889.0711892, 91388889.1902542, 91388889.2736588, 91388888.9976786, 91388888.9950674, 137083333.674279, 137083333.560474, 137083333.767976, 274166667.29018, 274166667.486932, 274166667.219459, 0, 0, 0, 0, 0, 0, -0.0262436712054121, 0, -0.0112163091303929, 0, 0, 0, -0.0138123127580231, 0, -0.00841872068719335, -20138888.9115235, -20138889.0158395, -20138888.9200768, 0, 0, 0, -0.0303189059509502, 0, -0.100014773986645, -100694444.603258, -100694444.661005, -100694444.823205, -100694444.950532, -100694444.548659, -100694444.471334, 0, 0, -0.130139672591355, -201388889.337421, -201388889.276499, -201388889.473581, -201388889.777382, -201388889.145133, -201388889.454292, -402777778.675084, -402777779.250428, -402777778.378957, -604166668.342825, -604166667.997932, -604166668.418998, 0, 0, 0, 0, 0, 0, 0.0155466666342393, 0, 0.00622420424284287, 0, 0, 0, 0.0122459413777448, 0, -0.00265359985088114, 9722222.22992647, 9722222.29691037, 9722222.2435115, 0, 0, 0, 0.00845543813337982, 0, 0.0578044660836413, 58333333.4098123, 58333333.4665178, 58333333.57765, 58333333.6709135, 58333333.4001377, 58333333.2885748, 0, 0, 0, 0.0932503072396916, 0, 0.145867501088539, 145833333.620945, 145833333.660545, 145833333.808534, 145833334.122546, 145833333.468648, 145833333.643516, 291666667.290887, 291666667.91176, 291666666.917706, 194444444.954875, 194444444.738856, 194444445.321612, 583333334.827491, 583333334.712725, 583333335.227635, 583333335.521475, 583333334.39242, 583333334.805144, 875000002.485344, 0, 0, 0, 0, 0, 0, -0.0038137413264629, 0, -0.00151714305224554, 0, 0, 0, -0.00446593363341842, 0, 0.00226384723846432, -1984126.98480812, -1984127.0027781, -1984126.99058299, 0, 0, 0, -0.000149373443853054, 0, -0.0102018279231353, -13888888.9030337, -13888888.9218508, -13888888.9527854, -13888888.9800051, -13888888.9107268, -13888888.8629837, 0, 0, 0, -0.02241462945373, 0, -0.0512135310635659, -41666666.7384152, -41666666.7739218, -41666666.8144435, -41666666.9249109, -41666666.6994533, -41666666.7133086, -83333333.5018516, -83333333.7364901, -83333333.3533949, 0, 0, -0.0601481756129765, -69444444.610611, -69444444.5742708, -69444444.6508528, -69444444.8545289, -69444444.5152533, -69444444.7455599, -208333333.818948, -208333333.848992, -208333334.105272, -208333334.275143, -208333333.59751, -208333333.738471, -277777778.584683, -277777778.368117, -277777778.776969, -416666667.837766, -416666668.090131, -416666667.749343}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 2.66315380059241e-08, 0, 0, 0, 0, 0, 0, 0, -5.7076200472367e-07, 0, -3.50181093544905e-07, 1428.57142872568, 0, 0, 0, 0, 0, 0, 5.09463467410051e-06, 0, 1.25397759609355e-06, 0, 0, 7.30797629315081e-06, -36428.5714353127, -1428.57142974816, -1428.57143010769, 0, 0, 0, 0, 0, 0, -2.02315705785516e-05, 0, 7.6054553398366e-07, 0, 0, 0, -6.7200692017263e-05, 0, -2.74688632153767e-05, 357222.222268328, -1.47824982596707e-05, -9.01978975099719e-06, 35000.0000451569, 35000.0000773228, 5.55452491435723e-05, 0, 0, 0, 0, 0, 0, 4.19085013860348e-05, 0, -1.18769014935755e-05, 0, 0, 0, 0.00024434418852466, 0, 4.44021231552832e-06, -1780555.55570402, 9.31376393935794e-05, 9.71974567391845e-05, 0, 0, 0.00030632927244688, -322222.222473844, -322222.222664476, 1.98869970232799e-05, -9.87435524436816e-05, -0.00021578719949988, -0.000296024835911912, -0.00083233776986069, 0, 0, 0, 0, 0, 0, -5.06052448042312e-05, 0, 2.47198694814397e-05, 0, 0, 0, -0.000413059407185993, 0, 0.000147265085517861, 4930555.55581183, -0.000197834798980721, -0.000310515889162635, 0, 0, 0, -0.00106981672310209, 0, -0.000508036497680135, 1458333.33399797, 1458333.33443094, -0.000505017835720649, -0.000200341645158845, 3.32734628077828e-05, 0.000687326272345181, -5.68992919759712e-05, 0.00111699305891535, 0.000724968036632451, 0.00305815005053177, 0.00218288605343037, 0.0031675720891206, 0, 0, 0, 0, 0, 0, 3.77368805210555e-05, 0, -2.25473418232576e-05, 0, 0, 0, 0.000347281139496862, 0, -0.000289976492133908, -7638888.88912755, 0.000197054282737135, 0.000467934298371058, 0, 0, 0, 0.00154613905306742, 0, 0.000212409877265773, -3472222.2232096, -3472222.22365844, 0.000982215734499992, 0.000628598208772942, 0.000527545498533996, -0.000769992767827732, 0, 0, 0.00178377568303156, -0.000119367082027852, -0.00225272486445699, 0.00137430256875034, 4.02081953602213e-05, -0.000891847548066073, -0.00161528704301813, -0.00513103362530576, -0.00189507361999568, -0.00466036977691976, -0.00569443619916315, -0.00812786872598377, -0.00809910194409986, 0, 0, 0, 0, 0, 0, -1.69976459498304e-05, 0, 9.43844784253781e-06, 0, 0, 0, -0.00013754911092843, 0, 0.000216317867889828, 6150793.65090384, -9.63921426984914e-05, -0.000345947678291253, 0, 0, 0, -0.000986264803899404, 0, 0.000187260662511887, 4166666.66738151, 4166666.66757617, -0.000772367610268242, -0.000536208860926102, -0.000737567215298983, 0.000397219464901552, 0, 0, 0, -0.00196690643140811, 0, -0.0011622844772814, 0.000331211440106249, 0.00205608139084348, -0.00189128541694602, -0.000892235211053264, -7.71316618061497e-05, 0.00172473569847293, 0.00423961810953685, 0.00014581763803041, 0.0028111058012834, -0.00149232228870312, 0.00147944736550411, 0.000273176148564218, 0.00602403641774866, 0.00894652817807614, 0.00323146108640948, 0.00701865745342938, 0.00771205334363981, 0.00946462483761883, 0.0134074558648313, 0, 0, 0, 0, 0, 0, 3.63893025996163e-06, 0, -1.42355491537304e-06, 0, 0, 0, 1.9305032106811e-05, 0, -5.73925419122599e-05, -1984126.98414616, 1.80267956643927e-05, 0.000100482049852105, 0, 0, 0, 0.000228988789433559, 0, -0.000134427515324647, -1984126.98432966, -1984126.98434784, 0.000208601334421452, 0.000146610684042757, 0.000293725412856454, -5.73802013463059e-05, 0, 0, 0, 0.000691364292660696, 0, 0.000215847384412764, -0.000194821633085855, -0.000706816428059366, 0.000735204752920326, 0.000444990978466146, 0.000245413282550156, -0.000665696256544646, -0.00134470737316132, 0.000280043176704313, -0.00054034005753026, 0, 0, 0.000772347988655756, 0.000485585988982419, -0.000741900686292128, 0.00107257474871691, 0.000539064952847164, -0.000196639115895069, -0.000283607240030352, -0.00237641256172814, -0.00349169972634126, -0.000119952599392594, -0.0018458340242561, -0.00241006596927519, -0.00369544809722854, -0.00150164481648523, -0.0040557652681836, -0.00448094361485478, -0.00688736017408915, -0.0059355999073501, -0.00760240556307861}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 2.74647782496484e-07, 0, 0, 0, 0, 0, 0, 0, -4.10773578929794e-06, 0, -5.76118518326627e-06, 1428.57142580354, 0, 0, 0, 0, 0, 0, 2.12565660191059e-05, 0, 4.44409567860922e-05, 0, 0, 8.06295217082234e-05, -1428.57141212018, -36428.5713954113, -1428.57138024934, 0, 0, 0, 0, 0, 0, -5.71544411910393e-05, 0, -0.000171669198216127, 0, 0, 0, -0.000373852740127185, 0, -0.000569742598748323, -4.5549648356328e-05, 357222.222069023, -0.000314618670371104, 34999.9998430084, -0.000258123962064782, 34999.9994971824, 0, 0, 0, 0, 0, 0, 9.12597669568492e-05, 0, 0.000376343307227519, 0, 0, 0, 0.00086086220812929, 0, 0.00195299171251789, 6.17294347891906e-05, -1780555.55514314, 0.0010113380614517, 0, 0, 0.00230016155685199, 0.000378787299165294, 0.000630255374370538, -322222.221674186, -322222.220375055, 0.00148667572843393, 0.00279792785499833, 0.0020901085750534, 0, 0, 0, 0, 0, 0, -9.3480797015443e-05, 0, -0.000486925751080076, 0, 0, 0, -0.00107622144587633, 0, -0.00367881230306714, -5.47074039902555e-05, 4930555.55485862, -0.00174031899686138, 0, 0, 0, -0.00449269259955606, 0, -0.00643445475742694, -0.000453800567216392, -0.000558867791447841, 1458333.33223544, 1458333.32967668, -0.00401045656858571, -0.00769556817178269, -0.000870344309898191, -0.00321522196515764, -0.0079452965433077, -0.0048387294998944, -0.00571560748155332, -0.00905648795208099, 0, 0, 0, 0, 0, 0, 6.5977747799702e-05, 0, 0.000364496217153898, 0, 0, 0, 0.000704406139730003, 0, 0.00389907317631178, 5.25923981210317e-05, -7638888.88817387, 0.00161662959052422, 0, 0, 0, 0.0046313888103259, 0, 0.00920444835277685, 0.000313699105028548, -2.31172803685412e-05, -3472222.22082666, -3472222.21801216, 0.00543702533131822, 0.0110469636481549, 0, 0, 0.00960116938539676, 0.000480971806477087, 0.00273321532598578, 0.000937140831631067, 0.0110876873267475, 0.00716226279200199, 0.0164125936163578, 0.00521738904482493, 0.00819409871127879, 0.018351277906617, 0.010293206897789, 0.0162367184658386, 0.0167321556398648, 0, 0, 0, 0, 0, 0, -3.19593364256516e-05, 0, -0.000143569133477931, 0, 0, 0, -0.000199482975890452, 0, -0.00217888591972862, -4.25570669840365e-05, 6150793.65038704, -0.000743592120687344, 0, 0, 0, -0.00237997886349136, 0, -0.00658813300359712, -0.000158188883909569, 0.000292368501175198, 4166666.66568656, 4166666.66405882, -0.00349084970059976, -0.00786433009011273, 0, 0, 0, -0.00681293446074331, 0, -0.00917521876130233, 4.8998356056085e-05, -0.000478407870628383, -0.000605555051441443, -0.00808842993913112, -0.00694747995913399, -0.0161962857491832, -0.00253028977356457, -0.00636675993642926, -0.017263939866678, 0.000214361920631841, -0.00479643256886419, -0.014825462773256, -0.00816467882364851, -0.0126110277046418, -0.00927352855778623, -0.0137647215560484, -0.0221237394813541, -0.0214546667169176, -0.0211060458259176, 0, 0, 0, 0, 0, 0, 7.93161496119321e-06, 0, 2.23709709787244e-05, 0, 0, 0, 9.61462497770388e-06, 0, 0.000498664166166738, 1.43501728194517e-05, -1984126.98402814, 0.000128797764650274, 0, 0, 0, 0.000466292287827035, 0, 0.00187105828788923, 4.93870274806147e-05, -0.000107184487155766, -1984126.98384249, -1984126.98347175, 0.000839838477075402, 0.00218641002222695, 0, 0, 0, 0.00185398674382035, 0, 0.00327604343830519, -6.56446458119642e-05, -0.000235053908452478, 0.000210690566386935, 0.00241667619514178, 0.00239938480219033, 0.00601419958060734, 0.000420884215839859, 0.00211122739421323, 0.00609213754764963, 0, 0, 0.00327065480522394, -0.00037182950615013, 0.000837178858533371, -0.000255417766288216, 0.00514390431614416, 0.0026035484399342, 0.0072516063533454, 0.002286251085016, 0.00317985028713775, 0.00328096777813248, 0.0047515729695797, 0.0104326205538507, 0.0102767627433371, 0.00352521808726019, 0.00881003987081884, 0.00790307089064635, 0.00935837734248471, 0.00865664931182057, 0.013258633722367}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 1.23414915830921e-07, 0, 0, 0, 0, 0, 0, 0, -2.18024166606551e-06, 0, -2.6034983143197e-06, 1428.57142776569, 0, 0, 0, 0, 0, 0, 1.20758405989329e-05, 0, 1.81666089396956e-05, 0, 0, 4.55111365766625e-05, -1428.57142385689, -1428.57142240334, -36428.5714123969, 0, 0, 0, 0, 0, 0, -3.07517246862908e-05, 0, -6.32313483299952e-05, 0, 0, 0, -0.000250640103418705, 0, -0.000294913036975804, -2.77771798355749e-05, -9.00844080643446e-06, 357222.222101406, 1.98054746744782e-05, 34999.9999395885, 34999.9998341616, 0, 0, 0, 0, 0, 0, 3.54145020796687e-05, 0, 0.00012417951974573, 0, 0, 0, 0.000679098685226051, 0, 0.000933536545090062, 0.000123386073626552, -6.45286515392981e-06, -1780555.55514776, 0, 0, 0.00139916282208225, -0.000103371636210546, 0.000119450074852037, -0.000246534425913267, 0.000468416991460639, -322222.221722072, -322222.220966114, -5.30644256850573e-06, 0, 0, 0, 0, 0, 0, -3.85385280460536e-06, 0, -0.000141422759613115, 0, 0, 0, -0.00102241752338879, 0, -0.00163869343294536, -0.000306718937135024, 3.23635441354606e-05, 4930555.55486282, 0, 0, 0, -0.00313501708480134, 0, -0.00359855439462029, -5.53211010090374e-05, -0.000186905541538384, 0.000541656365033254, -0.000679411210266541, 1458333.3317277, 1458333.32931972, 0.00105678036709397, -0.000939942528650126, -0.00326833674085183, 0.0011759300077137, 0.00157452967326625, -0.00214050260629866, 0, 0, 0, 0, 0, 0, -3.2666505915208e-05, 0, 8.97673858100279e-05, 0, 0, 0, 0.000865326253898891, 0, 0.00162420923593821, 0.000402673406645494, -5.19138285078053e-05, -7638888.88829313, 0, 0, 0, 0.00370775736090284, 0, 0.00481201865115467, 0.000610849424840896, 0.000316927280856964, -0.000357975720598976, 0.000642411787527531, -3472222.21997029, -3472222.21599218, 0, 0, 0.00606226348545906, -0.00135860681997891, 0.00070853778297411, -0.00195521500087368, 0.00386407431882021, 0.002803731761095, 0.00838982130352738, -0.0019311285363545, -0.00383379042492726, 0.0076096890678814, -0.006627896827069, 0.000224166303026157, 0.000736716596227387, 0, 0, 0, 0, 0, 0, 3.09313528987052e-05, 0, -2.77560919977026e-05, 0, 0, 0, -0.000381961697580919, 0, -0.000846339612849402, -0.000263961626810711, 4.79640403885164e-05, 6150793.65056621, 0, 0, 0, -0.00223152595776171, 0, -0.00325682593800062, -0.000769719226430129, -0.000328731736021019, -4.30035894493166e-05, -0.000386330549650066, 4166666.66528081, 4166666.66199106, 0, 0, 0, -0.00480568268649032, 0, -0.00536863946993194, 0.000420190748669265, -0.000145311993633497, 0.00116421635769781, -0.00236034335769114, -0.0030184084390842, -0.00911365208487759, 0.000776045509880445, 0.00308572484199302, -0.00897967676422152, 0.00215727791403884, -0.00188358148523899, -0.00684552473800719, 0.00700185123507387, 0.0020636369529932, 0.00913750763832409, 0.00303740883352778, -0.00526505753930511, -0.00747220315398157, 0.00606238302093029, 0, 0, 0, 0, 0, 0, -9.09266276345991e-06, 0, 2.77752298530692e-06, 0, 0, 0, 6.75665075580305e-05, 0, 0.000179032016383502, 6.83367022985926e-05, -1.83072631064416e-05, -1984126.98410485, 0, 0, 0, 0.000537333838368532, 0, 0.000879823752391486, 0.000293623996617378, 0.000127181995721227, 7.96106677649766e-05, 0.000102279968485849, -1984126.98383889, -1984126.9827743, 0, 0, 0, 0.00146496232428001, 0, 0.00180637413823334, 0.000128763342612441, -1.35078753084012e-05, -0.000136830659806521, 0.000628118957185346, 0.00105607166434621, 0.00358573684212806, 0.000115335361180385, -0.000757100922516239, 0.00352954039837515, 0, 0, 0.00210709370947181, -0.00067378252250082, 0.000359665435506386, -0.000802964150166395, 0.00208147342367114, 0.00120611400875522, 0.00377918938003519, -0.00208624307034391, -0.00107618271491778, -0.00363638311117951, -0.0020416805755823, 0.00393609496296611, 0.00550673000244494, -0.00490704192128065, 0.000746907709973266, 0.0013447465260537, -0.0045180909392764, -0.00572830907671807, 0.000887073308019336}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 7.50084259170324e-07, 0, 0, 0, 0, 0, 0, 0, -1.01672652695134e-05, 0, -2.45208845068818e-05, -105000.000166992, 0, 0, 0, 0, 0, 0, 1.89370389330945e-05, 0, 0.000235579268234849, 0, 0, 0.000447996364641656, 1937916.6693151, 887916.668877954, 887916.668075891, 0, 0, 0, 0, 0, 0, 0.000105647217033959, 0, -0.000389273372006637, 0, 0, 0, -0.00179824633035368, 0, -0.00573235034706443, -11971111.1246053, -3091944.45631355, -3091944.445304, -15063055.5828098, -15063055.5823056, -6183888.90737148, 0, 0, 0, 0, 0, 0, 8.62416787238347e-05, 0, -0.00130714370518326, 0, 0, 0, -0.00184609392021925, 0, 0.0182917455708992, 36596527.8114176, 5677083.36417345, 5677083.31766532, 0, 0, 0.0347801714218104, 78870138.9991273, 78870139.0188243, 47950694.5519721, 17031250.1057559, 47950694.521919, 17031250.0224231, 95901389.1309326, 0, 0, 0, 0, 0, 0, -0.00184427125236858, 0, 0.00537203899147011, 0, 0, 0, 0.0173428435288587, 0, -0.0202149577480534, -62569444.4897003, -5798611.15541261, -5798611.06168701, 0, 0, 0, -0.070891604432699, 0, -0.116806684147783, -193506944.656228, -193506944.736428, -79965277.9786255, -23194444.7030455, -79965277.8618598, -23194444.3393742, -205104166.996361, -205104167.023596, -34791666.9126232, -410208334.216177, -239895834.196434, -239895833.8984, 0, 0, 0, 0, 0, 0, 0.00408835275523663, 0, -0.00738177798092365, 0, 0, 0, -0.0301115999390368, 0, -5.06100827101885e-05, 61111111.1433343, 3125000.03732609, 3124999.93652255, 0, 0, 0, 0.0595150564303165, 0, 0.153856060816394, 247569444.652716, 247569444.785564, 73611111.3009173, 15625000.3076567, 73611111.1349902, 15624999.6854586, 0, 0, 0.222604043086181, 379166667.133474, 379166667.321693, 263194444.881907, 31250000.6020866, 263194444.846758, 31250000.1076116, 758333334.739339, 294444445.84274, 294444444.816274, 789583335.480615, 789583335.201578, 441666668.415076, 0, 0, 0, 0, 0, 0, -0.00358022931231796, 0, 0.00456592523843104, 0, 0, 0, 0.0216769740165981, 0, 0.0137105659643543, -31944444.4548518, -694444.462618305, -694444.406061804, 0, 0, 0, -0.015129869419112, 0, -0.0850643227561602, -160416666.764648, -160416666.866737, -35416666.7527154, -4166666.84333024, -35416666.6546741, -4166666.36039114, 0, 0, 0, -0.173071377106401, 0, -0.224106015892703, -322916666.975571, -322916667.210766, -166666666.922095, -10416667.2282379, -166666666.86654, -10416666.4812375, -645833334.367914, -177083334.380484, -177083333.208049, -326388889.295055, -326388889.42896, -13888889.3526007, -979166668.819442, -979166668.858215, -666666668.818664, -354166668.793959, -666666668.024035, -354166667.777782, -1000000003.24956, 0, 0, 0, 0, 0, 0, 0.0011345556380129, 0, -0.00107159531875362, 0, 0, 0, -0.00567399299603854, 0, -0.0064338952109207, 6944444.44525769, 0.00415556921452931, -0.00902817813240918, 0, 0, 0, -0.00262421501351641, 0, 0.0148872576202285, 41666666.6829944, 41666666.713065, 6944444.45885266, 0.0391969495750477, 6944444.43940691, -0.103164195987396, 0, 0, 0, 0.0475841539195112, 0, 0.0732748289868817, 104166666.7425, 104166666.836152, 41666666.7173433, 0.181310302698494, 41666666.705805, -0.14348409689879, 208333333.620339, 41666666.9622776, 41666666.5196154, 0, 0, 0.0964249454760938, 138888889.025729, 138888889.125752, 104166666.782056, 0.275461337693764, 104166666.815665, 0.0773466052014508, 416666667.426232, 416666667.533047, 208333334.093104, 104166667.541734, 208333333.599329, 104166666.771894, 416666667.69559, 416666667.586397, 138888889.724609, 625000001.757781, 416666668.412138, 416666667.963755}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 1.45073621522904e-07, 0, 0, 0, 0, 0, 0, 0, -9.71817462595061e-06, 0, 3.73821180358686e-05, 84000.0000909501, 0, 0, 0, 0, 0, 0, 0.000319586771097207, 0, -0.000293598524466412, 0, 0, -0.000908419143368964, -1802333.33525587, -542333.334484711, -542333.333956103, 0, 0, 0, 0, 0, 0, -0.00229672499561851, 0, 0.000478575323999408, 0, 0, 0, 0.00382580696578269, 0, 0.00862076068336532, 13723888.8994417, 1388888.89338982, 1388888.88618423, 10912777.7967485, 10912777.7976121, 2777777.78893721, 0, 0, 0, 0, 0, 0, 0.00736490868033402, 0, 0.00130827890509717, 0, 0, 0, -0.00293760537026532, 0, -0.0278288703291705, -49713888.9157285, -1763888.89382089, -1763888.86383665, 0, 0, -0.0441991830593603, -74075000.084696, -74075000.105744, -26125000.063471, -5291666.72709192, -26125000.0510296, -5291666.67118792, -52250000.1861381, 0, 0, 0, 0, 0, 0, -0.0127633988006679, 0, -0.00577153266581795, 0, 0, 0, -0.0110813545401728, 0, 0.0412514939824459, 97013888.9241057, 1111111.10434249, 1111111.04519971, 0, 0, 0, 0.0889881194984942, 0, 0.134374481195723, 222708333.507376, 222708333.576364, 30902777.8623514, 4444444.57491278, 30902777.8055542, 4444444.33407863, 155486111.330124, 155486111.397082, 6666666.79336324, 310972222.961841, 92708333.9666357, 92708333.7550438, 0, 0, 0, 0, 0, 0, 0.0124251222499025, 0, 0.00832244090383697, 0, 0, 0, 0.0272413407063443, 0, -0.0276740027156229, -105138888.911759, -277777.756776441, -277777.694699664, 0, 0, 0, -0.0857392251578861, 0, -0.180344774569742, -332638889.069187, -332638889.168548, -18055555.5879057, -1388889.0162039, -18055555.5049901, -1388888.61624464, 0, 0, -0.231922854236556, -367361111.450986, -367361111.664731, -157638889.097766, -2777778.09748278, -157638889.188098, -2777777.74913088, -734722223.456216, -72222223.1609849, -72222222.4416885, -472916668.417448, -472916668.219677, -108333334.624636, 0, 0, 0, 0, 0, 0, -0.0063858590101369, 0, -0.00545180651773251, 0, 0, 0, -0.0233202380922716, 0, 0.00511981139451168, 59722222.2277767, -0.0181872809910752, -0.0522111492605647, 0, 0, 0, 0.0368037575313106, 0, 0.111985604286602, 243055555.643362, 243055555.711716, 4166666.64680171, 0.0497728871815562, 4166666.59678496, -0.247827447587672, 0, 0, 0, 0.178040648559319, 0, 0.222174903959689, 375000000.244858, 375000000.462736, 76388888.9324341, 0.295216829924608, 76388889.0059604, -0.226621524080126, 750000000.933239, 20833333.9525417, 20833333.1557478, 263888889.102582, 263888889.344337, 0.180382768798176, 791666668.54352, 791666668.570614, 305555557.185344, 41666668.1671091, 305555556.638482, 41666667.4059456, 458333336.0371, 0, 0, 0, 0, 0, 0, 0.00134609086790275, 0, 0.00137016822453847, 0, 0, 0, 0.00714321631358052, 0, 0.00142347239538407, -13888888.8886739, 0.0054314266869071, 0.0132091340842517, 0, 0, 0, -0.00423735841240313, 0, -0.0257846183267468, -69444444.4592864, -69444444.4776844, 0.0138055081828339, -0.00311256098421951, 0.0244766708183708, 0.0804760155007651, 0, 0, 0, -0.0515511499589682, 0, -0.0738483470860169, -138888888.953823, -138888889.029767, -13888888.8686411, -0.0921704289212221, -13888888.8976263, 0.128857534595322, -277777778.039258, -0.146525304924449, 0.143233291156745, 0, 0, -0.0895268897419812, -138888888.96975, -138888889.096447, -69444444.4642691, -0.130367518241628, -69444444.5656507, 0.0179599051751397, -416666667.364663, -416666667.435883, -69444444.9553547, -0.580873796663025, -69444444.6336102, -0.0149043181711318, -277777778.62152, -277777778.556557, -0.549305435469056, -416666668.214793, -138888890.297729, -138888889.944286}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 1.50522200277521e-06, 0, 0, 0, 0, 0, 0, 0, -1.23235257942005e-05, 0, -5.7620070728592e-05, -52500.0000374377, 0, 0, 0, 0, 0, 0, -0.000105402794531229, 0, 0.000472061890915849, 0, 0, 0.000886287812198205, 1213958.33419634, 251458.333730792, 251458.333659007, 0, 0, 0, 0, 0, 0, 0.00115685474843529, 0, -0.00172930052497452, 0, 0, 0, -0.00377179310247312, 0, -0.00719772657578047, -10309027.7828182, -448958.333950647, -448958.33271941, -5507986.11897921, -5507986.12050691, -897916.671956583, 0, 0, 0, 0, 0, 0, -0.00420464209931537, 0, 0.00337677685138167, 0, 0, 0, 0.00625690944403176, 0, 0.0246756789015527, 42480902.7909901, 354166.662956891, 354166.658291657, 0, 0, 0.0309324969311535, 42670138.9278051, 42670138.941358, 9293402.79747452, 1062500.02227893, 9293402.80250993, 1062500.0084189, 18586805.6430527, 0, 0, 0, 0, 0, 0, 0.00795964683762676, 0, -0.00365048103025016, 0, 0, 0, -0.00264363978575981, 0, -0.0441829424624011, -93402777.7949554, -104166.651404674, -104166.644022297, 0, 0, 0, -0.0591715810139355, 0, -0.091045971601932, -151597222.308861, -151597222.345746, -6909722.22738317, -416666.701779396, -6909722.23614334, -416666.645036957, -65000000.0807049, -65000000.1429263, -625000.056282686, -130000000.366247, -20729166.9382698, -20729166.8817941, 0, 0, 0, 0, 0, 0, -0.00843924935039865, 0, 0.00207021328572862, 0, 0, 0, -0.00505230036839544, 0, 0.0430500504290136, 112152777.788019, -0.0232047823001879, -0.0284419903325128, 0, 0, 0, 0.0596263982918254, 0, 0.129423140426966, 263541666.762764, 263541666.809202, 1909722.18570982, 0.0179988867851592, 1909722.19867807, -0.0679411118863321, 0, 0, 0.13838393529357, 192534722.367032, 192534722.508065, 43055555.5830793, 0.104995320041454, 43055555.7023707, 0.0133148808388018, 385069445.080564, 7638889.23376575, 7638889.04152996, 129166667.485413, 129166667.456552, 11458333.9392955, 0, 0, 0, 0, 0, 0, 0.00474520427628177, 0, -0.00049711010717375, 0, 0, 0, 0.00693474240579124, 0, -0.0213885961484921, -69444444.4459602, 0.0164962378512867, 0.0178167838885399, 0, 0, 0, -0.0306556616022156, 0, -0.0899493023811628, -218750000.049558, -218750000.077527, 0.0451479098308031, 0.00655837245770089, 0.0334702709042911, 0.062112542066344, 0, 0, 0, -0.0999978243024825, 0, -0.133303246806574, -239583333.451266, -239583333.576249, -10416666.6094905, -0.081857709498715, -10416666.7189781, 0.0510019216172035, -479166667.162317, -0.176160805654962, 0.0249636751051969, -100694444.489373, -100694444.678953, -0.0561531079890855, -302083334.2675, -302083334.322279, -41666667.3442312, -0.630779799372759, -41666667.2388555, -0.394369107927975, -62500001.3378315, 0, 0, 0, 0, 0, 0, -0.00110166593998246, 0, 1.39093595963854e-05, 0, 0, 0, -0.00255771121420898, 0, 0.00417590181020885, 17361111.1105681, -0.00457482792407133, -0.00452380315473122, 0, 0, 0, 0.00602807735987539, 0, 0.024408576877476, 69444444.4533501, 69444444.4598813, -0.015861639237874, -0.00672013546178274, -0.0118926245681255, -0.0196484120916853, 0, 0, 0, 0.0288770176092503, 0, 0.0469921338116251, 104166666.70141, 104166666.740662, -0.0374111872256749, 0.0223838075971176, 0.000992739426672197, -0.0302484119302567, 208333333.474567, 0.0245534415312781, -0.0420857298126937, 0, 0, 0.048138316874946, 69444444.4693478, 69444444.5548333, 17361111.0850657, 0.0336774891641324, 17361111.1717094, -0.00837990297437401, 208333333.698916, 208333333.739754, 0.172251167597881, 0.215409352792533, 0.113870797420158, 0.0448036734839286, 69444444.8263284, 69444444.8500038, 0.245259553965441, 104166667.461627, 0.654487407325952, 0.547524649167634}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -1.02867901357444e-06, 0, 0, 0, 0, 0, 0, 0, 9.48482506363145e-06, 0, 2.99107248387418e-05, 24000.0000110799, 0, 0, 0, 0, 0, 0, 5.07976375302788e-06, 0, -0.000260887083452956, 0, 0, -0.000357415900860243, -580666.666918903, -80666.6667468628, -80666.6668106558, 0, 0, 0, 0, 0, 0, -0.000250984378511513, 0, 0.0011289273642786, 0, 0, 0, 0.00132740406179245, 0, 0.00287148026584634, 5270555.55714716, 89999.9997206452, 90000.0003848898, 1860555.55754518, 1860555.55845848, 180000.001778734, 0, 0, 0, 0, 0, 0, 0.00100239412735861, 0, -0.00277594535629599, 0, 0, 0, -0.00218904680325874, 0, -0.0107886702926722, -23672222.2266813, -33333.3304295914, -33333.3334210436, 0, 0, -0.00988193427493203, -15613888.9002908, -15613888.9054557, -1975000.00220264, -100000.004580083, -1975000.00843038, -100000.006890608, -3950000.02614475, 0, 0, 0, 0, 0, 0, -0.00197393927537879, 0, 0.00406346910105685, 0, 0, 0, 0.00185037568434706, 0, 0.021899312433081, 57430555.5617014, -0.00774158688791891, -0.000905648965763189, 0, 0, 0, 0.0154667700704954, 0, 0.0310120738295315, 61944444.4727443, 61944444.4847889, 694444.433525757, 0.000135381021144073, 694444.451823654, 0.0122930557548869, 15208333.3492158, 15208333.3783927, 0.0201469159323318, 30416666.7792241, 2083333.40370563, 2083333.40897928, 0, 0, 0, 0, 0, 0, 0.00218298690502246, 0, -0.0035172585335676, 0, 0, 0, -0.000645573721616852, 0, -0.0245708847714813, -75972222.2261453, 0.00977788307700368, 0.0013968445667061, 0, 0, 0, -0.0137347162799555, 0, -0.0491329513838744, -123194444.47921, -123194444.492592, 0.029137305503192, 0.0113469701963128, 0.00177923263834623, -0.0138111348988261, 0, 0, -0.036968717863263, -52083333.3714234, -52083333.4257583, -4861111.09450026, -0.016283282332425, -4861111.15678897, -0.0301050503808505, -104166666.869804, -0.0638564833266271, -0.0806219451249953, -14583333.5643468, -14583333.5866913, -0.19291984957166, 0, 0, 0, 0, 0, 0, -0.00129036382726783, 0, 0.00167098353118928, 0, 0, 0, -0.000195329274358024, 0, 0.0142868567096005, 51388888.8896122, -0.00618532191560286, -0.000920693932215673, 0, 0, 0, 0.00732092779020206, 0, 0.0383923306624257, 116666666.686618, 116666666.693419, -0.0258204236946824, -0.0132701293036598, -0.00597598154883322, 0.0100689866782715, 0, 0, 0, 0.0213321861072476, 0, 0.0394801760886193, 79166666.7036811, 79166666.7479492, -0.0436770126340081, 0.000250964051414591, 0.0157032807006838, 0.0244736291943811, 158333333.498114, 0.00596915216196602, 0.0318357956412732, 13888888.8835307, 13888888.962512, 0.0172371686425996, 41666666.947046, 41666666.9845391, 0.156722286290161, 0.170384523296405, 0.196874871640716, 0.166787451810693, 0.416326428305681, 0, 0, 0, 0, 0, 0, 0.000316391714926508, 0, -0.000338150424830107, 0, 0, 0, 0.000182692516510221, 0, -0.00335367875515868, -13888888.8887283, 0.00158474124727029, 0.000249565041314195, 0, 0, 0, -0.00180403022871999, 0, -0.0116822758552323, -41666666.6708553, -41666666.6720352, 0.00781826246171131, 0.00469856809499617, 0.00232017101778987, -0.00318078425927503, 0, 0, 0, -0.00544560623298627, 0, -0.0158085242129466, -41666666.6791436, -41666666.6922507, 0.0205510200998009, 0.00303066295559941, -0.000176030206372822, -0.00944642972458427, -83333333.3820144, 0.0104757016869223, -0.00303125550971721, 0, 0, -0.0109170121108579, -13888888.8921862, -13888888.9249588, 0.0192459034393629, -0.000768805341338893, -0.0179427953218442, -0.00566598233001472, -41666666.7823627, -41666666.7994588, -0.0214944813564632, -0.0443138287687767, -0.0470021675368981, -0.0455516421022708, -0.0992442959998582, -0.130639375094085, -0.084384483903703, -0.250203289203104, -0.187957615350565, -0.187722525202461}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 1.57203872140684e-07, 0, 0, 0, 0, 0, 0, 0, -7.20089577607404e-07, 0, -4.84129992095289e-06, -7500.00000206604, 0, 0, 0, 0, 0, 0, -7.99485211276693e-06, 0, 4.86243636511454e-05, 0, 0, 3.86707932012179e-05, 187083.33338481, 15833.3333424738, 15833.3333669969, 0, 0, 0, 0, 0, 0, 5.78535312272654e-05, 0, -0.00024170953479735, 0, 0, 0, -1.86454853355474e-05, 0, -0.000391857862513189, -1776111.11145622, -8333.33320113975, -8333.33347719788, -378194.44478857, -378194.445062668, -16666.6670666576, 0, 0, 0, 0, 0, 0, -0.000156794116983844, 0, 0.000661817440509331, 0, 0, 0, -0.000306114759128582, 0, 0.00183945712385003, 8471527.77881967, -0.000838525472487692, 0.000280436782890609, 0, 0, 0.000529195359735231, 3349305.55778732, 3349305.55903678, 190277.777588094, 0.000555972860565545, 190277.779794518, 0.00224181000640713, 380555.56110221, 0, 0, 0, 0, 0, 0, 0.000243209129061364, 0, -0.00104958728386989, 0, 0, 0, 0.000637574320367687, 0, -0.00439568644432632, -22152777.7793691, 0.00189346614125242, -0.000257709717722796, 0, 0, 0, 0.00114029538075156, 0, -0.00313534734435345, -14340277.7838375, -14340277.7863372, 0.0042873140108043, 0.00180805127309477, -0.00227053116284179, -0.00596008300556739, -1562500.00138298, -1562500.00955154, -0.00476777866765258, -3125000.02303267, -0.0131864328504597, -0.0195000592103166, 0, 0, 0, 0, 0, 0, -0.000239193334443186, 0, 0.00096503503581433, 0, 0, 0, -0.000415811511949495, 0, 0.00552846920358085, 31944444.4456611, -0.00209448205743657, 5.12800372488215e-05, 0, 0, 0, -0.00279461458950702, 0, 0.007156325410511, 31597222.2305124, 31597222.2328223, -0.00875217237989651, -0.00503676088986276, 0.000490943395838678, 0.00856865669360956, 0, 0, -0.000535698656653218, 5902777.78366577, 5902777.79704775, -0.00808703353915933, -0.000105214181435921, 0.0098707348401044, 0.0113200385979275, 11805555.5971375, 0.00781068552881224, 0.0275875677513306, 0.0431717203976955, 0.0561885489248217, 0.0464236287967718, 0, 0, 0, 0, 0, 0, 0.000139114776117464, 0, -0.000480057017019291, 0, 0, 0, 3.41767127562852e-05, 0, -0.0034884821767441, -23611111.1115042, 0.00117243163607699, 0.000104140293979976, 0, 0, 0, 0.00179897472826984, 0, -0.00697408727509562, -34027777.7831394, -34027777.7839241, 0.0070875728323119, 0.00427083757627444, 0.000990817787762337, -0.00637618820903398, 0, 0, 0, 0.00352358537092243, 0, -0.00199926763967536, -10416666.6737694, -10416666.6838718, 0.0135315934605828, 0.00524119282092203, -0.00347742635903216, -0.0127893860400594, -20833333.3679874, 0.00506593036736249, -0.017461812204521, 0.00540204867994656, -0.0149527672341334, -0.00289377523053663, -0.0528957559095436, -0.0674513395212201, -0.0230502475340121, -0.0366579324132398, -0.0493487018851751, -0.0513990229022144, -0.0896596172910087, 0, 0, 0, 0, 0, 0, -3.56358805638139e-05, 0, 0.000100554382453346, 0, 0, 0, 3.42791938510398e-05, 0, 0.000871042487999845, 6944444.44446732, -0.000267404084694539, -5.92999251545745e-05, 0, 0, 0, -0.000323941817849957, 0, 0.00244139993267231, 13888888.8902134, 13888888.8901964, -0.0020162494652681, -0.00121728724319232, -0.000523191790007417, 0.0018419429829323, 0, 0, 0, -0.00150064199699825, 0, 0.00171778095723448, 6944444.44723125, 6944444.45004223, -0.00572675993001879, -0.00264315260715237, 5.82929554360337e-05, 0.00537220617817922, 13888888.8995173, -0.0044079761098215, 0.00433856014011434, 0, 0, -0.00105992277041444, -0.000568940332888201, 0.00732500589975608, -0.00616887790434206, -0.00242455733525117, 0.00347278022745037, 0.002555692628057, 0.0224211418916916, 0.0277468329572752, -0.000685930304274085, 0.00754125552371423, 0.0145213365489625, 0.0194103893549438, 0.0154320747977129, 0.0288829093026499, 0.0232480943046549, 0.0516179628653603, 0.0379815948255867, 0.0463018620484241}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -1.18821268463855e-07, 0, 0, 0, 0, 0, 0, 0, 1.87451574265567e-07, 0, 3.89166271305511e-06, 833.33333320151, 0, 0, 0, 0, 0, 0, 7.66924456709383e-06, 0, -3.79309958984845e-05, 0, 0, -2.74102055083893e-05, -19861.1111144315, -9166.66665457616, -833.333345337515, 0, 0, 0, 0, 0, 0, -5.44588842934023e-05, 0, 0.000185520122477294, 0, 0, 0, 5.39945330427697e-05, 0, 0.000254952818228298, 175277.777808333, 8333.33317617276, 0.000102331620393385, 217638.888898179, 19027.7778953255, 8333.33340846007, 0, 0, 0, 0, 0, 0, 0.000173845075072199, 0, -0.00051664091887503, 0, 0, 0, 1.53499692875777e-05, 0, -0.00109167274471365, -746527.777840874, 0.000762434729495709, -0.000396103240490316, 0, 0, -0.000570108203397258, -1909027.77818752, -156250.000657528, -190277.77724864, 0.000111060960376235, -0.000475795187337957, -0.000735356341035072, -190277.778953695, 0, 0, 0, 0, 0, 0, -0.000316746898925619, 0, 0.000850184659456459, 0, 0, 0, -0.000189974119258136, 0, 0.00254049541479338, 1631944.44440515, -0.00177074937322103, 0.000915396708543603, 0, 0, 0, 0.000505448626699863, 0, 0.00209176319003298, 8055555.55694965, 590277.77930163, -0.00284908741233313, -0.0018041401867379, 0.000705452342238521, 0.00209073782913806, 1562499.99979938, 0.00203123172954613, 0.00211999456071286, 1562500.00561313, 0.00252098743277747, 0.00449326459773876, 0, 0, 0, 0, 0, 0, 0.000337461238758539, 0, -0.000817944278025382, 0, 0, 0, 0.000314167545478139, 0, -0.00324082925174185, -1736111.11085187, 0.00216423123590049, -0.00123463452449627, 0, 0, 0, -0.000249952588723233, 0, -0.00394838684457741, -17361111.1130425, -1041666.66825934, 0.00535104570308856, 0.00436457484153873, -0.000728189184487284, -0.00315546093404897, 0, 0, -0.00142451042893463, -5902777.77961339, -0.00420858172941034, 0.00402792271575834, -0.00125414642086521, -0.00200985406177357, -0.00523063503178421, -5902777.78865133, 0.00158629335218812, -0.00520430939413811, -0.011131338393243, -0.0137346194448644, -0.0134808579723163, 0, 0, 0, 0, 0, 0, -0.000194347392654596, 0, 0.000426008381204924, 0, 0, 0, -0.000253731100674724, 0, 0.00211664157250356, 694444.444163666, -0.00134877834232259, 0.000869535166410664, 0, 0, 0, 0.000154207159243813, 0, 0.00363333310325169, 18055555.5566714, 694444.445011277, -0.00436519591157134, -0.00413721458357855, 0.000621538323450572, 0.00260783444089045, 0, 0, 0, 0.000328504760290825, 0, 0.00195652774624578, 10416666.6693752, 0.00367512078528958, -0.00612407792535773, -0.00121904611016098, 0.000706340593256501, 0.0051453934694277, 10416666.6758328, -0.00703400942006234, 0.0019162125559244, -0.00154735399063462, 0.00333724760191607, 0.0042989231481592, 0.0155067387255121, 0.0177781371386217, 0.0034580730766158, 0.0113071099293137, 0.00965863122469479, 0.0148512705543752, 0.0261611990265682, 0, 0, 0, 0, 0, 0, 4.65095365128087e-05, 0, -9.29675076074865e-05, 0, 0, 0, 8.39206000449349e-05, 0, -0.000552773123004146, 9.65748561947775e-05, 0.000340141745949116, -0.000245932220999034, 0, 0, 0, -5.67857115358681e-05, 0, -0.00126703711206142, -6944444.44465528, 1.9704091234636e-05, 0.00130554135819977, 0.00141399239358585, -0.00027123159787077, -0.000878669713390408, 0, 0, 0, -6.1177358976405e-05, 0, -0.00102841528163393, -6944444.44551614, -0.00109073995677094, 0.00261599458264574, 0.00115105320743383, -0.000139951729572191, -0.00198615215086628, -6944444.44713983, 0.00395452221449615, -0.000177478890916176, 0, 0, -0.000188161929084145, -0.000361141460021644, -0.00174987412907816, 0.00204490560314808, -0.000889488924050236, -0.000758762501620627, -0.00222983424634729, -0.0070879433411443, -0.00758674758394249, 0.00282636015847388, -0.00159204391294285, -0.00134517576119856, -0.00471714557999827, -0.00416537862439832, -0.00659582930451374, -0.00868127565998083, -0.015572565204458, -0.0114564751579665, -0.0128160779096942}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 5.00031060191053e-08, 0, 0, 0, 0, 0, 0, 0, -1.88145807209275e-06, 0, 2.1319701903986e-06, 666.666664984864, 0, 0, 0, 0, 0, 0, 2.83954837470204e-05, 0, -1.52322316226345e-05, 0, 0, -4.30388776570701e-05, -14555.5555408103, -10666.6666510999, -666.666655396435, 0, 0, 0, 0, 0, 0, -0.000181556460297317, 0, 3.77378085277963e-05, 0, 0, 0, 0.000227681549330058, 0, 0.000323708805578632, 111111.111032405, 43333.3331584008, -0.000183740641637748, 232222.222112087, 13888.8888223206, 10000.0002742507, 0, 0, 0, 0, 0, 0, 0.000609672996322881, 0, -3.61916450898849e-05, 0, 0, 0, -0.000502226418245591, 0, -0.000878073904354489, -374999.999728593, -33333.3322515415, 0.000951920931219016, 0, 0, -0.00176502339095085, -1763888.88812146, -97222.2222268073, -936111.110348817, -33333.3353140517, 0.00122881633387223, -0.000752980722539864, -208333.335077907, 0, 0, 0, 0, 0, 0, -0.00115688323541637, 0, -1.57774823212576e-05, 0, 0, 0, 0.000262515006309043, 0, 0.00128964628008729, 555555.55497629, -0.00312318622587427, -0.00214537104881814, 0, 0, 0, 0.00468592549142296, 0, 0.00388191220590142, 5902777.77529447, 277777.778307316, 694444.440939656, 0.00378221721775246, -0.00522514760011283, -0.0012349934017564, 7013888.88551955, -0.00214461385468175, 0.00879509408159681, 1458333.33848109, 694444.454671825, 0.00239404706450599, 0, 0, 0, 0, 0, 0, 0.00124391900088841, 0, 7.22312710506905e-05, 0, 0, 0, 0.000664738288787929, 0, -0.00114118408520717, -277777.777060597, 0.00451298776656305, 0.00246588289294397, 0, 0, 0, -0.00625093737717195, 0, -0.00400776506500733, -8611111.10691988, -277777.778609916, 0.007359073070886, -0.00143296954669215, 0.00892778290154581, 0.00494197239329568, 0, 0, -0.00827491377125912, -22916666.6590718, 0.00102612664456489, -4861111.10231281, -0.0202466756793254, 0.00867278049785102, -0.00946229558700847, -4166666.67394559, -0.0157580623866843, 0.00904886409324015, -4861111.1302978, -0.00866438890832864, -0.0293427465856641, 0, 0, 0, 0, 0, 0, -0.000705943446372655, 0, -6.50992537794627e-05, 0, 0, 0, -0.00101136789002257, 0, 0.000573040709881648, -0.000464125818629246, -0.0032076797312263, -0.00144362009039792, 0, 0, 0, 0.00383544922213071, 0, 0.00213635321359689, 4166666.66321848, 0.000536772214267068, -0.00679176460370262, -0.00231152501945791, -0.00670148015080862, -0.00471012997145709, 0, 0, 0, 0.0083037136033528, 0, 0.00561259302991628, 31944444.436734, 0.000532766865160804, -0.0102624652878438, 0.0169440560713375, -0.010565030552404, 0.00155947167773523, 4166666.67113444, 0.00733397214057692, -0.0176670716195153, 13888888.8781649, -0.00552731029921633, 0.0205155399882933, 13888888.9041461, 0.0113307861030398, 0.0187575959788239, 0.0433445955579501, -0.00707614199088146, 0.0176690491720988, 0.0407722275570933, 0, 0, 0, 0, 0, 0, 0.000164228261238728, 0, 2.01515572942305e-05, 0, 0, 0, 0.000399813219825647, 0, -0.000122090339197512, 0.000121289363737752, 0.000893331524979123, 0.000347240518407005, 0, 0, 0, -0.000809301220396689, 0, -0.000498788093627295, 0.00110954004205686, -0.000114821041582128, 0.00231865751072837, 0.00164021159662661, 0.00189208293554544, 0.00144322853861252, 0, 0, 0, -0.00301480430887486, 0, -0.0013559191633726, -13888888.8859299, -0.000391155566846568, 0.00429349700562417, -0.00456202827886973, 0.00414359828176746, 0.00115565745225919, -0.000889433429692172, 0.000320543388396593, 0.00827022149104641, 0, 0, -0.00296929390193998, -13888888.8840857, 0.00108646126031465, 0.00549059027518855, -0.0107939517237482, 0.00384201679820294, -0.0053647601244183, -13888888.892892, -0.00516091935552321, -0.00430738696402529, -0.0175402824144111, 0.0101701312292342, 0.000946359599885076, -0.00665738896661047, 9.68827034890082e-05, -0.0207403609910918, -0.01886660941424, -0.0277946735491702, -0.0140281914825125}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -7.47547185149644e-08, 0, 0, 0, 0, 0, 0, 0, 1.40066660413564e-06, 0, 2.39284095352372e-06, 625.000001484919, 0, 0, 0, 0, 0, 0, -1.10998549463284e-05, 0, -1.02600668646561e-05, 0, 0, -5.00217465101338e-05, -12083.3333212398, -12083.3333908066, -625.000011130322, 0, 0, 0, 0, 0, 0, 2.14341380192472e-05, 0, 9.26341901921421e-06, 0, 0, 0, 0.000467532176093698, 0, 0.00018617465685014, 73958.3332085798, 73958.3337981607, -6.29714803007591e-05, 232986.111188839, 11458.3331651391, 11458.3340333479, 0, 0, 0, 0, 0, 0, 6.28026432644219e-05, 0, 9.81138693691518e-06, 0, 0, 0, -0.00182648202701786, 0, 2.56731313733148e-05, -166666.666272416, -166666.66820605, 0.00047152224414575, 0, 0, -0.0020267320757075, -1418402.77687235, -62499.9990848966, -1418402.77883864, -62500.0055824818, 0.00149497390753325, -0.00176239671172455, -210069.446181024, 0, 0, 0, 0, 0, 0, -0.000300863205445932, 0, -8.60359699494845e-06, 0, 0, 0, 0.00329778970706998, 0, -0.000882225950685077, 104166.666018195, 104166.66915869, -0.000933902296280313, 0, 0, 0, 0.00801200057393951, 0, 0.00216486575073759, 3159722.21874163, 104166.664849841, 3159722.22571895, 104166.683547523, -0.00498695647686391, -0.000143634575168863, 8541666.66409969, -0.00525280426287637, 0.0167573341786652, 1145833.33305537, 1145833.34820328, 0.00225033150073629, 0, 0, 0, 0, 0, 0, 0.000450318007107418, 0, -1.72132233169902e-05, 0, 0, 0, -0.00282586695720263, 0, 0.00133095681820026, 0.00062182832466752, -0.00203676834065576, 0.000712614870958327, 0, 0, 0, -0.0133336103255353, 0, 0.00101361537037023, -1909722.2171476, 0.00146799707234754, -1909722.22731119, -0.0232732152263544, 0.00709440339838437, 0.00439216632055256, 0, 0, -0.0102995571453993, -18576388.8792054, 0.00732413931929629, -18576388.8860875, -0.0452040982103101, 0.0123839751109008, -0.0156321820109415, -1909722.21775003, -1909722.25854684, 0.00658425984768616, -6250000.00920112, 0.00318569949270441, -0.034488015395686, 0, 0, 0, 0, 0, 0, -0.000298725714765834, 0, 1.96374944652036e-05, 0, 0, 0, 0.00100365140174436, 0, -0.00075732147079476, -0.000339110833436157, 0.000786489971141899, -0.000153491100264921, 0, 0, 0, 0.00976747355211516, 0, -0.00266337652231517, -0.00355360845746411, -0.000476170167875172, 0.00334435106796455, 0.0146957441679787, -0.00456154466276184, -0.00420027729760498, 0, 0, 0, 0.0139924188527631, 0, 0.00376436319653971, 10416666.6564818, -0.00460198042444019, 10416666.6654262, 0.047526592561243, -0.0120249053069721, 0.00144141900286437, -0.00427442450570974, 0.0352386333501266, -0.0134226487946819, 38194444.4339831, -0.011810244761211, 0.039020041175806, 10416666.6607376, -0.00474404729187849, 10416666.6856586, 0.0655978868850183, -0.0133348414699162, 0.0215622027170653, 0.0208658082147708, 0, 0, 0, 0, 0, 0, 7.48058119170887e-05, 0, -4.94394115901983e-06, 0, 0, 0, -6.93096359448329e-05, 0, 0.000145871702766065, 8.15861663868173e-05, -0.000105886752117176, -2.5515540029305e-05, 0, 0, 0, -0.00258839677495314, 0, 0.00105147050546714, 0.000986355391892961, 1.35385050075765e-05, -0.00084337714358568, -0.0033797691816367, 0.00104512319647571, 0.00111349637451952, 0, 0, 0, -0.0058626322744092, 0, 0.000329097057244221, 0.00361998477189955, 0.00101055571380884, 9.60549028859011e-05, -0.0168975963014716, 0.00400654745705077, 0.00251809912821846, 0.00126256728323225, -0.0119694511806593, 0.00579893551258791, 0, 0, -0.00398991104496527, -17361111.1053738, 0.00400524473017561, -17361111.1072741, -0.0237636445940171, 0.00585036112645667, -0.00887537174366937, 0.00516017629278797, 0.0016985593650199, -0.0105985291163473, -0.0346604242766025, 0.00981607963185233, -0.000645656482985996, -17361111.1090705, 0.00602424952462961, -0.0289713817543528, -0.00225689172086309, -0.0223596463419539, -0.00634636562771355}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -8.09619923529631e-07, 0, 0, 0, 0, 0, 0, 0, 1.01304320440901e-05, 0, 1.64806213110787e-05, 666.666676555115, 0, 0, 0, 0, 0, 0, -6.11694397169441e-05, 0, -0.000137861590058764, 0, 0, -0.000164042253397517, -10666.6667018899, -14555.5557202464, -666.666795601365, 0, 0, 0, 0, 0, 0, 0.000211895508422039, 0, 0.000611442828303194, 0, 0, 0, 0.000788934590608221, 0, 0.00110506664220155, 43333.3333887757, 111111.11221544, 0.000865215601305325, 232222.222938714, 10000.0002154419, 13888.8904466618, 0, 0, 0, 0, 0, 0, -0.000437438175992215, 0, -0.00159819251362732, 0, 0, 0, -0.00219258675849901, 0, -0.00385226417107576, -33333.3333779021, -375000.003752146, -0.00327013283357309, 0, 0, -0.00419288072962929, -936111.112572536, -33333.3330852871, -1763888.89346723, -97222.2303670817, -0.000950020466906249, -0.00744510030686848, -208333.337444998, 0, 0, 0, 0, 0, 0, 0.000556026139781625, 0, 0.00253176766733072, 0, 0, 0, 0.00344552139274302, 0, 0.00778760272530936, 2.82128909170961e-06, 555555.562710234, 0.00713983833073259, 0, 0, 0, 0.0094255792595155, 0, 0.0106321796874195, 694444.446211562, -0.000707391744820109, 5902777.791079, 277777.798641937, 0.00352856261501838, 0.0193543883627055, 7013888.89743168, -0.00235260815970442, 0.0290400302926766, 694444.44811754, 1458333.3556205, 0.0127845822947973, 0, 0, 0, 0, 0, 0, -0.000437361064817171, 0, -0.00237542439513919, 0, 0, 0, -0.00296702634922213, 0, -0.00922613397780134, 7.81697995789546e-05, -277777.785618616, -0.00887354988540929, 0, 0, 0, -0.0117193032631199, 0, -0.014455107920475, -0.00128580832474763, 0.000147034131495051, -8611111.13077723, -277777.805582707, -0.00717700209845208, -0.0281795596348899, 0, 0, -0.0172893511198339, -4861111.11933196, 0.00574583069738009, -22916666.686414, -0.0544179822704595, 0.00382556481567898, -0.0518095926719771, -0.000716296028989242, -4166666.7145826, -0.0235232899324002, -4861111.13476931, -0.00251634281999161, -0.0528331366144854, 0, 0, 0, 0, 0, 0, 0.000199896157527386, 0, 0.00120879183580908, 0, 0, 0, 0.00129166356790526, 0, 0.0058982333364533, -0.000119105635185884, 0.00462339377860105, 0.0057949954023051, 0, 0, 0, 0.00724341521277326, 0, 0.0104344687543798, 0.000535660880522596, 0.000601200737666346, 4166666.68128213, 0.0186721122888196, 0.00707150113125045, 0.0216748140649695, 0, 0, 0, 0.0146492734678391, 0, 0.0143965749714032, 0.00412905878330498, -0.0037954043909011, 31944444.4647137, 0.0469666696516758, -0.00131891452155473, 0.0447618942660802, 9.41139722923014e-05, 4166666.71251148, 0.0230760007290046, 13888888.9025065, -0.00817967841743751, 0.0626683101248674, 0.0108822681911059, -0.00845787755405001, 13888888.9290249, 0.0743959219514898, 0.00141601332318877, 0.0605039643331357, 0.0295348250046463, 0, 0, 0, 0, 0, 0, -4.11734199160629e-05, 0, -0.000256189160590698, 0, 0, 0, -0.000219354805541495, 0, -0.00155796712906924, 5.43383773264349e-05, -0.00113981084045049, -0.00153561409241292, 0, 0, 0, -0.00172121662480446, 0, -0.00315870497417006, -7.35688001078235e-05, -0.000325306186823182, -0.00431944031678688, -0.00504703165026563, -0.00259990443844341, -0.00684358539288331, 0, 0, 0, -0.00475934913290573, 0, -0.00463246513456982, -0.000859417129718039, 0.000653259379067905, -13888888.8965034, -0.0152416414880876, -0.000514432830114894, -0.0153035557839015, -0.000429991830261639, -0.0162691047283562, -0.00903535807432362, 0, 0, -0.00599184590742674, -0.00355455055569402, 0.00338166439381433, -13888888.8961464, -0.026949197770343, 0.00313538011620285, -0.0257480133409937, -0.00187073747641583, 0.00381169670230589, -13888888.9099519, -0.0343038421979144, -0.00136245275488863, -0.0256479434284424, -0.011525659974652, 0.00644833217819495, -0.0419335271112203, -0.00251906614738873, -0.0266227678102645, -0.0159039196245986}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 1.49194496339535e-08, 0, 0, 0, 0, 0, 0, 0, -2.50348711603643e-06, 0, -1.54013031465729e-07, 833.333334366839, 0, 0, 0, 0, 0, 0, 1.59841588294633e-05, 0, -1.13676393077852e-05, 0, 0, 6.19711348122012e-05, -9166.66665863237, -19861.1111161647, -833.333355000902, 0, 0, 0, 0, 0, 0, -4.62259931044471e-05, 0, 0.000108277799432906, 0, 0, 0, -0.000412964408863161, 0, -0.000357725122358852, 8333.3332504388, 175277.777872636, 0.000276225064069031, 217638.888778043, 8333.33332986023, 19027.7776218401, 0, 0, 0, 0, 0, 0, 6.98605317297365e-05, 0, -0.000425318203604769, 0, 0, 0, 0.00129899405577943, 0, 0.000968627780230377, 0.000315925409463262, -746527.778309189, -0.00137135894864477, 0, 0, 0.00212234190355652, -190277.777053432, 0.000207175475244133, -1909027.77779831, -156249.998993637, -0.000763413524890832, 8.2647819429184e-05, -190277.775739237, 0, 0, 0, 0, 0, 0, -4.05576689756623e-05, 0, 0.000882269327455859, 0, 0, 0, -0.00238127543265117, 0, -0.00130620589367472, -0.000707404776598694, 1631944.44583451, 0.0034104643313614, 0, 0, 0, -0.0052362332144639, 0, -0.00522123823499354, -0.00188578833396343, -0.000274479341410271, 8055555.55700725, 590277.775200938, 0.00376922608407847, 0.00179396329887716, 1562499.99920427, 0.00097834504042794, -0.00248482999193242, -0.00792193527455059, 1562499.99210009, -0.00358136168081791, 0, 0, 0, 0, 0, 0, -2.41897476031902e-05, 0, -0.00100899091697284, 0, 0, 0, 0.00253488181729093, 0, 0.000703492483218881, 0.00093070415857232, -1736111.11303936, -0.00457057202728327, 0, 0, 0, 0.00707740462428286, 0, 0.00685536974061233, 0.00272326013429427, -2.20789904368427e-05, -17361111.1147725, -1041666.66315618, -0.0074627379954867, -0.00478115134366874, 0, 0, 0.00938879263813617, 0.00157370444982273, -0.00182957481279943, -5902777.77986187, 0.0064403008438966, -0.00460355726211773, 0.000701163187796597, 0.0120056887040066, -5902777.7653916, -0.00221200845814764, 0.0220054039322073, 0.0151447543390155, 0.0147247581855321, 0, 0, 0, 0, 0, 0, 4.31176261899439e-05, 0, 0.000600725249440238, 0, 0, 0, -0.00142252816112397, 0, 7.80803716636246e-05, -0.000648709745299349, 694444.445816052, 0.00314307287882081, 0, 0, 0, -0.00509157770248631, 0, -0.00444488386180648, -0.00216762892508471, 8.87740161528263e-05, 18055555.559148, 694444.442159537, 0.00672623765449098, 0.00493071083324431, 0, 0, 0, -0.00792519244898594, 0, -0.00833679821718547, -0.00159891045691135, 0.00233564440698819, 10416666.6709963, -0.00738260075607349, 0.0063307067421121, 0.00170187400376026, -0.00840029789249716, 10416666.6577166, 0.00822139628338337, 0.00199089297491975, 0.00397583071199896, -0.00330087129866433, -0.0218924883283308, -0.016185084782295, -0.0218954779469592, -0.0184846066125376, -0.00707685837753462, -0.00364335874644391, -0.0344038013414931, 0, 0, 0, 0, 0, 0, -1.55023512771116e-05, 0, -0.000145455117243545, 0, 0, 0, 0.000321195988801812, 0, -0.000146235964969007, 0.000182665153902676, -0.000390250088406451, -0.000867931016164031, 0, 0, 0, 0.00149320456269371, 0, 0.00108340424733288, 0.000709492135766744, 2.40768484483118e-06, -6944444.44572788, 0.000529010416619178, -0.00230659846933139, -0.00185733992395321, 0, 0, 0, 0.00270243424645669, 0, 0.00288063355172175, 0.000714191767443597, -0.00103640601210627, -6944444.44662246, 0.00313216214583936, -0.00291742740104945, -0.00112409600374229, 0.00240491490870385, -6944444.44217295, -0.00462689073495412, 0, 0, 0.00334238383419605, -0.000703870313378497, -0.00182128372193796, -0.00213914908182702, 0.00251189938421133, -0.00236960005268881, 0.000257318615200603, 0.00739746219331942, 0.00546225516085833, 0.00735923201150681, 0.00831811680116361, -0.00108181153242048, -0.00325558649776076, 0.0102977771248739, 0.005306879432593, 0.003571292637741, 0.0188665391427172, 0.0197503909293171, 0.0113685829187908}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -1.03967975061635e-06, 0, 0, 0, 0, 0, 0, 0, 1.57998751218176e-05, 0, 2.36197194495161e-05, -7499.99999155031, 0, 0, 0, 0, 0, 0, -6.64867388517423e-05, 0, -0.000180233973006395, 0, 0, -0.000363091011133658, 15833.3332663744, 187083.333255093, 15833.3331638466, 0, 0, 0, 0, 0, 0, 0.00011516794341287, 0, 0.000666604061456241, 0, 0, 0, 0.00159351706084704, 0, 0.00266375336667687, -8333.33307346, -1776111.11100078, -8333.33231816639, -378194.444009833, -16666.6655629362, -378194.442551587, 0, 0, 0, 0, 0, 0, -3.83124381030662e-05, 0, -0.00134364520150383, 0, 0, 0, -0.00338840147237078, 0, -0.00934939520590776, -0.00061170047081234, 8471527.77820698, -0.00286142697525611, 0, 0, -0.0103986400024763, 190277.776499731, -0.0031717292704533, 3349305.55549264, 3349305.54937182, -0.00580718275158349, 190277.767246314, 380555.546255374, 0, 0, 0, 0, 0, 0, -0.000158525767162513, 0, 0.0014895780945725, 0, 0, 0, 0.00386965067206194, 0, 0.0179127076635227, 0.00110294211284982, -22152777.7795948, 0.00398977375601692, 0, 0, 0, 0.0192406850426763, 0, 0.0301955886036423, 0.00205381385344369, 0.00369753369833156, -14340277.7811099, -14340277.7671783, 0.0140961552362463, 0.0282925060591535, -1562500.00119422, 0.0145717610197287, -1562499.97218497, 0.0248179210519566, -3124999.97722001, 0.0386414750581858, 0, 0, 0, 0, 0, 0, 0.000241952504472375, 0, -0.000821186900947662, 0, 0, 0, -0.00220864999407103, 0, -0.0191263867582405, -0.00140582703525079, 31944444.4472641, -0.00238546910774979, 0, 0, 0, -0.0190269585867912, 0, -0.0447651069567756, -0.00236867128761588, -0.00157942865304889, 31597222.2298976, 31597222.2117815, -0.0166816283489498, -0.04030305368787, 0, 0, -0.042732821320931, 0.00373393073377206, -0.01362919392595, 5902777.78806935, 5902777.74149059, -0.0298800496649726, -0.0564356517308148, -0.030322691980621, 11805555.5278742, -0.0720589516060097, -0.046587808551666, -0.081083840321402, -0.0676490362034117, 0, 0, 0, 0, 0, 0, -0.000133339934595102, 0, 0.000137963658954248, 0, 0, 0, 0.000410435793898878, 0, 0.0106470828473829, 0.00101103843562003, -23611111.1131546, 7.04148906187633e-05, 0, 0, 0, 0.00960054628058417, 0, 0.0330346728453444, 0.00192282667626871, 1.25722043756987e-05, -34027777.7848287, -34027777.7726014, 0.00864579195254391, 0.0285965335613124, 0, 0, 0, 0.0288823228686153, 0, 0.0425574033116838, -0.00340331428610386, 0.00315602856891203, -10416666.6804204, -10416666.6406651, 0.0267524896974872, 0.056562447431322, 0.0176872303881014, -20833333.3156533, 0.0635178569521875, -0.0131166439195443, 0.0215251320288757, 0.046474884018788, 0.0389551486312228, 0.0696131164233941, 0.0342660850389219, 0.0524936293075936, 0.102467262792193, 0.0767795987131312, 0.103001066479332, 0, 0, 0, 0, 0, 0, 2.47960490111635e-05, 0, 2.83381297989578e-05, 0, 0, 0, 6.19975800212417e-05, 0, -0.00239882497136569, -0.000295508646500216, 6944444.44501299, 0.000319784563234101, 0, 0, 0, -0.00189552125550151, 0, -0.00959421675496864, -0.000696250348821317, 4.04543792155787e-05, 13888888.8912835, 13888888.8880321, -0.00132640199892093, -0.0078790401882455, 0, 0, 0, -0.00763330179548531, 0, -0.0158403227291645, 0.000874263469876732, 0.000849075306272287, 6944444.45021848, 6944444.43656312, -0.00834875514445116, -0.0218002525887062, -0.00420194843471144, 13888888.8843135, -0.0212041214120373, 0, 0, -0.0143738424254346, 0.00547169074907175, -0.00381318612447354, 0.00810392045410262, -0.0155662334975194, -0.0109404132439005, -0.0228574043732753, -0.0114165183118168, -0.0199917303930079, -0.00894505501274545, -0.0178809262199724, -0.0451071649607093, -0.0335260521983194, -0.012329881024247, -0.0453470488626491, -0.0246024591589449, -0.0494260485020942, -0.0408803270180982, -0.0600449322833966}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 3.61334219362174e-07, 0, 0, 0, 0, 0, 0, 0, -5.96575466244823e-06, 0, -2.35855845896291e-05, 24000.0000114508, 0, 0, 0, 0, 0, 0, -0.000103224674645111, 0, 0.000177943629728817, 0, 0, 0.000569053405260155, -80666.6666226824, -580666.667045872, -80666.6666145582, 0, 0, 0, 0, 0, 0, 0.000858322843491372, 0, -0.0004726833554464, 0, 0, 0, -0.00179405867743365, 0, -0.00500053999554003, 89999.999500805, 5270555.55952227, 89999.9999255854, 1860555.55712082, 179999.997939902, 1860555.55508716, 0, 0, 0, 0, 0, 0, -0.00270397985194977, 0, 0.000141647285920729, 0, 0, 0, 0.000934218433947837, 0, 0.0189623375963974, -33333.3314845046, -23672222.2388149, -33333.3354027787, 0, 0, 0.0175210098405804, -1975000.0030271, -99999.9915362586, -15613888.9073367, -15613888.8979488, -99999.9893972377, -1974999.99258779, -3949999.98632589, 0, 0, 0, 0, 0, 0, 0.00452402945512626, 0, 0.00168276893532267, 0, 0, 0, 0.00414547453253129, 0, -0.0375833902553046, -0.00436002321651307, 57430555.591489, 0.00991287132842046, 0, 0, 0, -0.0236829477107413, 0, -0.0597755612025173, 694444.447524765, -0.011995949779312, 61944444.5106651, 61944444.4851934, -0.0206052688640494, 694444.42081849, 15208333.3766512, -0.0394892025146395, 15208333.3450379, 2083333.27570655, 30416666.6699703, 2083333.26357442, 0, 0, 0, 0, 0, 0, -0.00426717658578134, 0, -0.00341812195807634, 0, 0, 0, -0.00828744915404818, 0, 0.0401121133014465, 0.00625798284165794, -75972222.2651811, -0.0186629358255659, 0, 0, 0, 0.0128257225529807, 0, 0.0978468002615359, 0.000168311779117371, 0.0073672946317219, -123194444.554295, -123194444.513343, 0.0132309125031575, 0.036953239938369, 0, 0, 0.0705522294153825, -4861111.16731302, 0.0436117034708726, -52083333.4580418, -52083333.402158, 0.0758372498475861, -4861111.11252729, 0.0837124254375763, -104166666.740975, 0.123327045677733, -14583333.2830654, 0.218982016671663, -14583333.2848916, 0, 0, 0, 0, 0, 0, 0.00214298085156537, 0, 0.00268573846385007, 0, 0, 0, 0.00614316685408622, 0, -0.02162201139314, -0.00473478138345267, 51388888.9157811, 0.0160315495514866, 0, 0, 0, -0.000696103362160513, 0, -0.0766674106189557, -0.00317134063347536, -0.00236692971104432, 116666666.754395, 116666666.721396, 0.00508719980726206, -0.0264903340257689, 0, 0, 0, -0.0350304837214118, 0, -0.0835678227494933, 0.0363943045644809, -0.0134812801709133, 79166666.8064367, 79166666.743464, -0.0605327570016488, -0.0234802735775443, -0.0551127694562756, 158333333.437662, -0.0980948343871401, 13888889.0011133, -0.06501804283009, 13888888.9569992, -0.0605164700083563, -0.218009696421774, 41666666.7140355, 41666666.7163887, -0.272621704177879, -0.0440576385227213, -0.219434862559262, 0, 0, 0, 0, 0, 0, -0.000445393896637347, 0, -0.000774062274538286, 0, 0, 0, -0.0016851074002645, 0, 0.00456193462299444, 0.00142847565396009, -13888888.8957568, -0.00517738815520713, 0, 0, 0, -0.00116658817569047, 0, 0.0229637017934091, 0.00178766556343739, 0.000688140820880352, -41666666.6939439, -41666666.6838541, -0.00626304816289758, 0.00657775134560589, 0, 0, 0, 0.00581554592872012, 0, 0.0348942749357106, -0.00895359763570986, -0.0013986384480317, -41666666.722005, -41666666.6940321, 0.0153638126343752, 0.0177159593928629, 0.0147125438316982, -83333333.3787401, 0.0283948577022182, 0, 0, 0.0230089967763798, -0.0378484705188399, 0.0133609725326089, -13888888.951868, -13888888.9309299, 0.0308011836922133, -0.0183123439270812, 0.0205279792841084, 0.0693319313471643, -41666666.7204757, -41666666.7091244, 0.113977708359002, 0.0150608698442942, -0.0178391529569464, 0.141174139605504, -0.0433059193712134, 0.132615455216703, 0.049125775128108, 0.131462338478477}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 6.12146182971792e-06, 0, 0, 0, 0, 0, 0, 0, -9.3085599687313e-05, 0, -6.98222944751959e-05, -52500.000117199, 0, 0, 0, 0, 0, 0, 0.000973226200374, 0, 0.000522716216093815, 0, 0, 0.000208053225129888, 251458.333761449, 1213958.33561492, 251458.334386126, 0, 0, 0, 0, 0, 0, -0.004783946445347, 0, -0.00267921297334085, 0, 0, 0, -0.00420340321291752, 0, 0.0023417973716942, -448958.33387671, -10309027.7959126, -448958.340072559, -5507986.12245786, -897916.666573031, -5507986.12406159, 0, 0, 0, 0, 0, 0, 0.0126594952616539, 0, 0.00878326528312295, 0, 0, 0, 0.0225977040724788, 0, -0.0148465327652635, 354166.665836002, 42480902.8455564, 354166.694154854, 0, 0, 0.000232653434027601, 9293402.80635704, 1062499.99032745, 42670138.9743835, 42670138.9820414, 1062499.9961589, 9293402.82173469, 18586805.5762816, 0, 0, 0, 0, 0, 0, -0.0195715463927312, 0, -0.0174574680654737, 0, 0, 0, -0.0510994388015468, 0, 0.0342091680866937, -104166.66127076, -93402777.9147763, -104166.733436535, 0, 0, 0, -0.0416303287896534, 0, 0.0409356150422707, -6909722.26387804, -416666.651837982, -151597222.491629, -151597222.502343, -416666.672888679, -6909722.30957044, -65000000.1966543, -624999.937945948, -65000000.2678593, -20729166.635295, -130000000.235951, -20729166.6634332, 0, 0, 0, 0, 0, 0, 0.0178390527104682, 0, 0.0200784462958275, 0, 0, 0, 0.0583158645300452, 0, -0.0365353099600381, -0.0104716675443127, 112152777.93314, 0.0925768753676797, 0, 0, 0, 0.0932064989182493, 0, -0.102329915650788, 1909722.25560657, -0.00495157900902215, 263541667.084603, 263541667.075733, 0.0496195605016918, 1909722.32096061, 0, 0, 0.00641139358550633, 43055555.7970257, -0.0813244351888596, 192534722.71126, 192534722.841549, -0.120067890284868, 43055555.9542596, 7638888.80378332, 385069445.080073, 7638888.86283941, 129166666.89422, 11458333.0337327, 129166667.040909, 0, 0, 0, 0, 0, 0, -0.00891968619793465, 0, -0.012232849728028, 0, 0, 0, -0.0335818118788257, 0, 0.0173067223484398, 0.00915355735308863, -69444444.5376914, -0.0670050632473681, 0, 0, 0, -0.0769155530626627, 0, 0.0949911421568766, -0.0120372051461999, -0.00222355003920872, -218750000.317745, -218750000.293685, -0.0737008056219831, -0.065272246592666, 0, 0, 0, -0.0643808414426381, 0, 0.0550390268203753, -10416666.8172491, 0.0255049978377632, -239583333.849642, -239583333.923425, 0.0830391898801134, -10416666.9297062, 0.0640722548120519, -479166667.345142, 0.00761454941782542, -100694444.858743, 0.130308081723285, -100694445.096956, -41666666.7710645, 0.369799681571567, -302083333.921811, -302083334.087345, 0.417228873637774, -41666667.0542341, -62499999.9325492, 0, 0, 0, 0, 0, 0, 0.00189047192868166, 0, 0.0030487898401423, 0, 0, 0, 0.00783433370676892, 0, -0.00257606393082406, -0.0030166172118788, 17361111.1341832, 0.0194890448450237, 0, 0, 0, 0.0223010196218026, 0, -0.0305418578816196, 0.000779611231781434, 0.000687785088339615, 69444444.5392002, 69444444.5281608, 0.0336702643700599, 0.0207489731624033, 0, 0, 0, 0.0341778546656518, 0, -0.0373583860574855, 0.0376527256380177, 0.004200249394635, 104166666.863547, 104166666.868284, -0.0123738481246819, 0.0592613217744068, -0.0172024426673237, 208333333.592329, 0.00946778256640015, 0, 0, 0.00537929117755898, 17361111.2431719, -0.0310141238346808, 69444444.6658791, 69444444.7635444, -0.0602120473802769, 17361111.3457307, 0.0193483291550985, -0.125868045621492, 208333333.691224, 208333333.729532, -0.175797425832616, 0.154954582393734, 69444444.6582097, -0.261550149337109, 69444444.8542909, -0.157205279863215, 104166666.845752, -0.0957163527679423}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -1.57128315624277e-05, 0, 0, 0, 0, 0, 0, 0, 0.000240856878813517, 0, 0.000224962181690376, 84000.0002771803, 0, 0, 0, 0, 0, 0, -0.0021336811778039, 0, -0.00168841759295055, 0, 0, -0.00185709597193805, -542333.334848387, -1802333.33825499, -542333.336174025, 0, 0, 0, 0, 0, 0, 0.00969514557053039, 0, 0.00756507829652934, 0, 0, 0, 0.0144925895674521, 0, 0.00638889552790434, 1388888.89333811, 13723888.9246549, 1388888.90505951, 10912777.8059788, 2777777.78571171, 10912777.8137036, 0, 0, 0, 0, 0, 0, -0.0248797291215553, 0, -0.0213091053979984, 0, 0, 0, -0.0565241382556974, 0, -0.00959655754407288, -1763888.89645784, -49713889.0151099, -1763888.94614919, 0, 0, -0.0444779369531628, -26125000.0800292, -5291666.6782997, -74075000.181266, -74075000.2235785, -5291666.68938063, -26125000.1297756, -52250000.122701, 0, 0, 0, 0, 0, 0, 0.0382273750924158, 0, 0.0375627438061739, 0, 0, 0, 0.112462037171511, 0, 0.00573081867280421, 1111111.1165871, 97013889.1328645, 1111111.23704146, 0, 0, 0, 0.160505329915231, 0, 0.0529141337018323, 30902777.9057203, 4444444.46977491, 222708333.860727, 222708333.960561, 4444444.50716609, 30902778.0330493, 155486111.543776, 6666666.6427017, 155486111.78001, 92708333.5252947, 310972222.940231, 92708333.586165, 0, 0, 0, 0, 0, 0, -0.0351412009746119, 0, -0.0397125086973506, 0, 0, 0, -0.120303497734335, 0, -0.00303912426626927, -277777.775044015, -105138889.155163, -277777.941913805, 0, 0, 0, -0.256113435559496, 0, -0.000251577292245083, -18055555.6719783, -1388888.93394746, -332638889.663787, -332638889.766878, -1388889.02185531, -18055555.8369059, 0, 0, -0.20125387149313, -157638889.423852, -2777777.741587, -367361112.086941, -367361112.552711, -2777777.6876538, -157638889.882849, -72222222.4077173, -734722223.849614, -72222222.5218617, -472916667.687045, -108333333.40137, -472916668.002755, 0, 0, 0, 0, 0, 0, 0.0178963363615087, 0, 0.0228517009507589, 0, 0, 0, 0.0666912242544443, 0, 0.00519047711260472, -0.00685564446264931, 59722222.3769279, 0.114375850523915, 0, 0, 0, 0.185765403086292, 0, -0.0387018283198189, 4166666.72221647, 0.0366146718296339, 243055556.118592, 243055556.162619, 0.146338619079262, 4166666.84020538, 0, 0, 0, 0.249004663746339, 0, 0.0895474906286961, 76388889.22218, 0.0169185621233669, 375000000.974139, 375000001.339822, -0.0641141714580528, 76388889.5684092, 20833333.4453683, 750000001.603327, 20833333.535706, 263888889.712118, -0.130580483392347, 263888890.38536, 305555556.269953, 41666666.5804992, 791666668.365189, 791666668.76404, 41666666.5047206, 305555556.81775, 458333334.211376, 0, 0, 0, 0, 0, 0, -0.00388949993074809, 0, -0.00547875438759275, 0, 0, 0, -0.0151973152886103, 0, -0.00308040141514349, 0.00301557313169273, -13888888.9262034, -0.0325073509057531, 0, 0, 0, -0.0504761752329809, 0, 0.0180227571030267, -0.0103370074410406, -0.0102224844622913, -69444444.6056323, -69444444.6106316, -0.0599689440981272, -0.0490244936433939, 0, 0, 0, -0.0978587328766848, 0, -0.00212251623646282, -13888888.9727405, -0.0218259288814485, -138888889.245026, -138888889.341525, 0.0011980475800074, -13888889.0583705, -0.0316392315961042, -277777778.357088, -0.0672745446320552, 0, 0, -0.0764373418130924, -69444444.7037704, 0.0301742145762119, -138888889.305987, -138888889.602097, 0.0718985933224541, -69444444.9855422, -69444444.6432693, 0.031797222981536, -416666667.557334, -416666667.690756, 0.116388869187391, -69444444.8926242, -277777778.4217, 0.175987495719209, -277777778.820184, -138888889.076919, -416666667.511409, -138888889.185714}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 1.68482323685396e-05, 0, 0, 0, 0, 0, 0, 0, -0.000267267797703176, 0, -0.000250458703849885, -105000.000320871, 0, 0, 0, 0, 0, 0, 0.00227341393714481, 0, 0.00190578616857914, 0, 0, 0.00244055026532466, 887916.668973081, 1937916.67207744, 887916.670041498, 0, 0, 0, 0, 0, 0, -0.0102264474678798, 0, -0.00816019330538412, 0, 0, 0, -0.0165956375079331, 0, -0.012131684373375, -3091944.45283195, -11971111.1480398, -3091944.46118228, -15063055.5911099, -6183888.90654407, -15063055.5997675, 0, 0, 0, 0, 0, 0, 0.0268968158873792, 0, 0.0213532287560637, 0, 0, 0, 0.0574312648177897, 0, 0.0316105942654193, 5677083.35068414, 36596527.9026736, 5677083.38387361, 0, 0, 0.0692838921712458, 47950694.5538942, 17031250.0536491, 78870139.0898634, 78870139.1472714, 17031250.0550927, 47950694.599913, 95901389.1079344, 0, 0, 0, 0, 0, 0, -0.0432349374991827, 0, -0.0351124582871002, 0, 0, 0, -0.107795688032283, 0, -0.0450958393814772, -5798611.13093141, -62569444.679387, -5798611.21073426, 0, 0, 0, -0.195816410887075, 0, -0.146340877098035, -79965277.9596062, -23194444.5533227, -193506944.983822, -193506945.138276, -23194444.5466881, -79965278.0537156, -205104167.163791, -34791666.7712304, -205104167.445405, -239895833.845744, -410208334.352906, -239895833.880419, 0, 0, 0, 0, 0, 0, 0.0419122835936416, 0, 0.0353096608917218, 0, 0, 0, 0.114089374717731, 0, 0.0369553975414353, 3125000.01013059, 61111111.3635015, 3125000.12300347, 0, 0, 0, 0.2708935869996, 0, 0.150825881138753, 73611111.2786341, 15625000.1347142, 247569445.193641, 247569445.389628, 15625000.1390925, 73611111.3675531, 0, 0, 0.332932055026062, 263194445.068084, 31250000.1624996, 379166667.678296, 379166668.288859, 31250000.06897, 263194445.572699, 294444445.091536, 758333335.395799, 294444445.088935, 789583335.066969, 441666667.568371, 789583335.409374, 0, 0, 0, 0, 0, 0, -0.0224988360799943, 0, -0.0197343776185562, 0, 0, 0, -0.0648918499354149, 0, -0.0175799726298337, -694444.444175262, -31944444.5901577, -694444.528597326, 0, 0, 0, -0.181159964045917, 0, -0.0697776839431766, -35416666.7465604, -4166666.75478304, -160416667.188081, -160416667.308528, -4166666.79060054, -35416666.7890306, 0, 0, 0, -0.320637104328874, 0, -0.252571510255105, -166666667.056022, -10416666.8244086, -322916667.610137, -322916668.144224, -10416666.6825479, -166666667.416324, -177083333.755066, -645833335.236894, -177083333.685593, -326388889.74775, -13888888.9312037, -326388890.556298, -666666668.036566, -354166667.447445, -979166669.007127, -979166669.529393, -354166667.282656, -666666668.544704, -1000000002.19489, 0, 0, 0, 0, 0, 0, 0.0051281081831028, 0, 0.00467201015073345, 0, 0, 0, 0.0155915822311448, 0, 0.00409207420586774, -0.00154369109254893, 6944444.47966944, 0.0239248006456098, 0, 0, 0, 0.0472029123811354, 0, 0.0104413909846522, 6944444.45936697, 0.0228624388285563, 41666666.8108355, 41666666.8401353, 0.047508261397557, 6944444.46992658, 0, 0, 0, 0.109676875727182, 0, 0.0675972235034594, 41666666.7636737, 0.060754530795598, 104166666.994524, 104166667.16037, 0.00960728951501957, 41666666.845816, 41666666.7783011, 208333333.991508, 41666666.7423492, 0, 0, 0.129513427726539, 104166666.938905, 0.0307781732867, 138888889.294247, 138888889.670345, -0.0237051758062491, 104166667.269055, 208333333.742422, 104166666.942387, 416666667.763058, 416666667.982834, 104166666.785876, 208333333.93989, 416666667.587324, 138888889.149323, 416666668.036915, 416666667.520619, 625000001.479865, 416666667.598084}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 1.82456637906614e-06, 0, 0, 0, 0, 0, 0, 0, -1.79740632922928e-05, 0, 1.16981018044388e-05, -105000.000244025, 0, 0, 0, 0, 0, 0, 0.000465261653278588, 0, 0.000216886037377863, 0, 0, -0.00126609533478285, 887916.669146629, 887916.670001779, 1937916.6695631, 0, 0, 0, 0, 0, 0, -0.00395295837755491, 0, -0.00188373631187169, 0, 0, 0, 0.0092996351596204, 0, 0.00467662510846141, -3091944.45530905, -3091944.46442512, -11971111.1247108, -6183888.91628075, -15063055.5824248, -15063055.5871428, 0, 0, 0, 0, 0, 0, 0.0161771814356985, 0, 0.005688729702226, 0, 0, 0, -0.031977139679238, 0, -0.00389732548519782, 5677083.36008955, 5677083.39795349, 36596527.8121393, 0, 0, -0.0333144733665112, 17031250.091523, 47950694.5465299, 17031250.1224591, 47950694.5977331, 78870138.9913426, 78870139.0044623, 95901389.1373954, 0, 0, 0, 0, 0, 0, -0.0353464292044395, 0, -0.0092422306716634, 0, 0, 0, 0.0506742761721268, 0, 0.000707211915344986, -5798611.15073898, -5798611.23567668, -62569444.496783, 0, 0, 0, 0.122910711669288, 0, 0.0158613085062739, -23194444.6042792, -79965277.9902953, -23194444.7229049, -79965278.1479578, -193506944.637451, -193506944.647894, -34791666.9779912, -205104166.977233, -205104167.138147, -239895834.028172, -239895834.228255, -410208334.103946, 0, 0, 0, 0, 0, 0, 0.0422431251257125, 0, 0.00915247822914973, 0, 0, 0, -0.0316854455124172, 0, -0.00780362010526422, 3125000.03480998, 3125000.14331415, 61111111.1607971, 0, 0, 0, -0.207170040727888, 0, 0.0423286230431522, 15625000.1547339, 73611111.3641894, 15625000.3446812, 73611111.6013889, 247569444.651157, 247569444.611393, 0, 0, -0.0939279674663743, 31250000.397443, 263194444.936727, 31250000.5070412, 263194445.280255, 379166667.079537, 379166667.38325, 294444445.385082, 294444445.988273, 758333334.388702, 441666668.398199, 789583335.051325, 789583335.467773, 0, 0, 0, 0, 0, 0, -0.0260917561653127, 0, -0.00533743884276258, 0, 0, 0, -0.000535827312676898, 0, 0.0131496153803084, -694444.460974202, -694444.534928049, -31944444.4724557, 0, 0, 0, 0.153338372175159, 0, -0.0442477187825141, -4166666.74602457, -35416666.8275929, -4166666.88865875, -35416667.0120213, -160416666.79282, -160416666.709741, 0, 0, 0, 0.157041201087718, 0, -0.0305741700196528, -10416666.9206044, -166666667.067476, -10416667.0625288, -166666667.351436, -322916666.930054, -322916667.179895, -177083333.958263, -177083334.629694, -645833333.995312, -13888889.3178043, -326388889.311545, -326388889.777835, -354166668.10377, -666666668.318762, -354166668.538418, -666666669.134892, -979166668.294466, -979166668.708364, -1000000002.91401, 0, 0, 0, 0, 0, 0, 0.00652151261625856, 0, 0.00139191560312692, 0, 0, 0, 0.00549831583887986, 0, -0.00552299542431979, 0.00321716473716185, 0.0240234076723519, 6944444.45171975, 0, 0, 0, -0.0402408556054927, 0, 0.00960133906272915, 0.0169276182204062, 6944444.48660312, 0.0585075924144648, 6944444.54580931, 41666666.7013366, 41666666.6592013, 0, 0, 0, -0.0736884135512695, 0, 0.0356395949066956, 0.0652974074481428, 41666666.7975362, 0.118703230817863, 41666666.8850184, 104166666.737213, 104166666.79623, 41666666.8313451, 41666667.0920068, 208333333.486697, 0, 0, -0.0153473140803824, 0.139982793528779, 104166666.838847, 0.166017360892999, 104166667.022801, 138888889.010871, 138888889.245441, 104166667.101605, 208333333.916849, 104166667.416603, 208333334.362553, 416666667.193727, 416666667.350488, 138888889.620849, 416666667.478842, 416666667.875119, 416666667.960659, 416666668.24059, 625000001.431627}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 3.7972331406404e-06, 0, 0, 0, 0, 0, 0, 0, -9.76209592174992e-05, 0, -9.1393522101222e-05, 84000.0001163883, 0, 0, 0, 0, 0, 0, 0.000442960601092194, 0, 0.000254007781220282, 0, 0, 0.00274026922601651, -542333.334461132, -542333.334732314, -1802333.3349545, 0, 0, 0, 0, 0, 0, 0.00031033441171957, 0, 0.000641605882173893, 0, 0, 0, -0.0191200988435976, 0, -0.013028561339644, 1388888.89359809, 1388888.89678284, 13723888.8974017, 2777777.78876743, 10912777.7926632, 10912777.7932066, 0, 0, 0, 0, 0, 0, -0.00740586773645351, 0, -0.00434166140069989, 0, 0, 0, 0.0616743825448625, 0, 0.0256045550982621, -1763888.90082213, -1763888.91492804, -49713888.9130705, 0, 0, 0.0846040758998204, -5291666.69665688, -26125000.0548831, -5291666.70957862, -26125000.0656107, -74075000.0595032, -74075000.0627018, -52250000.1379364, 0, 0, 0, 0, 0, 0, 0.0219990781735354, 0, 0.00927728876124886, 0, 0, 0, -0.0993050864085667, 0, -0.0266059432417806, 1111111.13030628, 1111111.16597407, 97013888.9305451, 0, 0, 0, -0.252005099430941, 0, -0.140803951891038, 4444444.4851025, 30902777.895974, 4444444.52917356, 30902777.9230728, 222708333.446082, 222708333.457469, 6666666.74570387, 155486111.280191, 155486111.325905, 92708333.6921197, 92708333.8028508, 310972222.706649, 0, 0, 0, 0, 0, 0, -0.029651223017383, 0, -0.0104097070357875, 0, 0, 0, 0.0771566847636331, 0, 0.018368469884367, -277777.796596867, -277777.84857932, -105138888.932918, 0, 0, 0, 0.366094109205349, 0, 0.104163055591128, -1388888.91661623, -18055555.7070228, -1388888.98194176, -18055555.7473174, -332638889.009754, -332638889.01167, 0, 0, 0.347458574883491, -2777777.84264435, -157638889.160023, -2777777.87006431, -157638889.20841, -367361111.321861, -367361111.465701, -72222222.6765875, -72222222.988511, -734722222.932707, -108333334.14518, -472916667.69196, -472916667.97436, 0, 0, 0, 0, 0, 0, 0.0193860222469793, 0, 0.0062478860440025, 0, 0, 0, -0.0226747485597314, 0, -0.00967745777129556, 0.0101171699961982, 0.049798371885618, 59722222.2488584, 0, 0, 0, -0.249296153315557, 0, -0.0341794069907285, 0.00796903847535523, 4166666.77133124, 0.0556677559328243, 4166666.81219008, 243055555.629969, 243055555.606891, 0, 0, 0, -0.374292147442418, 0, -0.186702921453351, 0.0192410104477114, 76388889.124089, 0.0438285633804169, 76388889.1138559, 375000000.120446, 375000000.284104, 20833333.6229314, 20833333.9536732, 750000000.474334, 0.0389121643173073, 263888889.097985, 263888889.245206, 41666667.2431154, 305555556.501076, 41666667.4559276, 305555557.016678, 791666667.694298, 791666668.015723, 458333335.033905, 0, 0, 0, 0, 0, 0, -0.00498730741179505, 0, -0.00158188953330208, 0, 0, 0, -0.000393140477380115, 0, 0.00267978700039413, -0.00226068484021235, -0.0144442956329828, -13888888.8959953, 0, 0, 0, 0.0630414064570024, 0, 0.00483250375238124, -0.000373267075925601, -0.0296492450332215, -0.014531948978612, -0.0478401754964071, -69444444.4652924, -69444444.4488131, 0, 0, 0, 0.141312155882083, 0, 0.0306894905188436, 0.000878500812213398, -13888888.971938, -0.00467047673922021, -13888888.9553923, -138888888.917533, -138888888.971098, -0.0750622669045777, -0.200215383455095, -277777777.896134, 0, 0, 0.112983458721394, -0.00419289619464534, -69444444.5348844, -0.00274910490416589, -69444444.5546009, -138888888.937096, -138888889.049439, -0.145213388997136, -69444444.7741595, -0.289691047328365, -69444445.0390089, -416666667.010614, -416666667.145473, -0.235018132903378, -277777778.265149, -277777778.54857, -138888889.591719, -138888889.772491, -416666667.581181}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -5.51634253705067e-06, 0, 0, 0, 0, 0, 0, 0, 0.000120180431276114, 0, 0.000108634898494342, -52499.999998506, 0, 0, 0, 0, 0, 0, -0.000723167258004506, 0, -0.000525222745669525, 0, 0, -0.00253558107528805, 251458.333438355, 251458.333131898, 1213958.33337041, 0, 0, 0, 0, 0, 0, 0.00161238427382434, 0, 0.000899273047234391, 0, 0, 0, 0.0168871180084603, 0, 0.0134043127849484, -448958.333941464, -448958.331809733, -10309027.7777971, -897916.666596142, -5507986.11378183, -5507986.10908828, 0, 0, 0, 0, 0, 0, 6.05557196334868e-05, 0, 0.000166485775065648, 0, 0, 0, -0.052594883390289, 0, -0.0322867173910069, 354166.668780927, 354166.662436518, 42480902.7777871, 0, 0, -0.0806572482167388, 1062499.99795964, 9293402.79052952, 1062499.99335564, 9293402.76273319, 42670138.8965148, 42670138.8755609, 18586805.5817686, 0, 0, 0, 0, 0, 0, -0.0063263389304276, 0, -0.00284559704985884, 0, 0, 0, 0.0848807886092223, 0, 0.04239678675451, -104166.670992789, -104166.663134729, -93402777.7793046, 0, 0, 0, 0.220461996178052, 0, 0.163526223610997, -416666.657461807, -6909722.25653511, -416666.635966407, -6909722.17955846, -151597222.226099, -151597222.180135, -624999.972405502, -65000000.0283689, -64999999.9319308, -20729166.7500954, -20729166.7406015, -130000000.09394, 0, 0, 0, 0, 0, 0, 0.0110888032960501, 0, 0.00438621189995368, 0, 0, 0, -0.0707791584948882, 0, -0.0322486139258304, 0.00503319905779439, 0.00342132909854394, 112152777.782336, 0, 0, 0, -0.300497483187954, 0, -0.166121410914144, -0.0171054325767435, 1909722.27405327, -0.0554333029925719, 1909722.17094071, 263541666.660632, 263541666.600445, 0, 0, -0.352243761773948, -0.0506945087534983, 43055555.6178673, -0.0844802398233551, 43055555.4041687, 192534722.225529, 192534722.085956, 7638889.00477855, 7638888.97622876, 385069444.542705, 11458333.4772656, 129166666.927561, 129166666.928812, 0, 0, 0, 0, 0, 0, -0.0079672018958769, 0, -0.00296610751711956, 0, 0, 0, 0.0270596848218534, 0, 0.0136344030032582, -0.00309048826268813, -0.00715817952281668, -69444444.4493892, 0, 0, 0, 0.197227363468749, 0, 0.085353479183271, 0.0145305056739965, -0.0397872006713514, 0.0443670303092857, 0.0231214800205963, -218749999.994206, -218749999.94752, 0, 0, 0, 0.338783586451464, 0, 0.244932924749554, 0.0453197516405371, -10416666.7365774, 0.10081784768681, -10416666.5222505, -239583333.306572, -239583333.211212, -0.0794047383275685, -0.0473417982286783, -479166666.68993, 0.0850161267854072, -100694444.458361, -100694444.249893, -0.094158760032455, -41666666.9382835, -0.0708172412072458, -41666666.950165, -302083333.553285, -302083333.565456, -62500000.4358699, 0, 0, 0, 0, 0, 0, 0.00214020921603677, 0, 0.000781862593365188, 0, 0, 0, -0.00302053188671909, 0, -0.00248155700824591, 0.000775831518686985, 0.00311633716089481, 17361111.1129994, 0, 0, 0, -0.0492795010562814, 0, -0.0181845685734739, -0.00457674999978921, 0.0120367940034364, -0.0130584645812114, -0.0012630099477429, 69444444.4436886, 69444444.4278671, 0, 0, 0, -0.118424872004133, 0, -0.0628823207566433, -0.0162454622457262, 0.0290994378060059, -0.0416444011509484, -0.0474283755697315, 104166666.652651, 104166666.623899, 0.0214937169801903, 0.0100711790399136, 208333333.324127, 0, 0, -0.122267202883455, -0.0303916793594982, 17361111.1262764, -0.047115768654163, 17361111.0129968, 69444444.4297671, 69444444.3654006, 0.0188976796357616, 0.105139517761843, -0.00265604497674648, 0.111944104836403, 208333333.383257, 208333333.388905, -0.00212146333686848, 69444444.5661444, 69444444.5858196, 0.194918089803939, 0.211095156132066, 104166666.893073}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 2.75187410586334e-06, 0, 0, 0, 0, 0, 0, 0, -5.87914565635207e-05, 0, -5.23551944428092e-05, 23999.9999818083, 0, 0, 0, 0, 0, 0, 0.000370859522350189, 0, 0.000282460002346603, 0, 0, 0.00115743809422026, -80666.6665847215, -80666.6663526291, -580666.666417535, 0, 0, 0, 0, 0, 0, -0.00100571470582321, 0, -0.000640081447579113, 0, 0, 0, -0.00747281971674656, 0, -0.00647533411121837, 89999.9998130975, 89999.9980804935, 5270555.55396731, 179999.99881526, 1860555.55496851, 1860555.55179939, 0, 0, 0, 0, 0, 0, 0.0010436463225695, 0, 0.000489072956324545, 0, 0, 0, 0.0226218631906993, 0, 0.0170114076528106, -33333.3331288393, -33333.3273958079, -23672222.2173416, 0, 0, 0.0373090921494745, -99999.9969187211, -1975000.00044631, -99999.9924194325, -1974999.98177935, -15613888.8833796, -15613888.8680694, -3949999.99706174, 0, 0, 0, 0, 0, 0, 0.000511231848151839, 0, 0.000543457019963176, 0, 0, 0, -0.0362832904478601, 0, -0.0243223332807561, -0.000146762360804931, -0.0098797382977193, 57430555.5480341, 0, 0, 0, -0.0965538261819917, 0, -0.0826594935069497, -0.00626028531769882, 694444.450260806, -0.0250940789093067, 694444.400916365, 61944444.4245999, 61944444.3869563, -0.0163931711408081, 15208333.3252983, 15208333.2561869, 2083333.3421363, 2083333.32174662, 30416666.6453899, 0, 0, 0, 0, 0, 0, -0.0021726412135516, 0, -0.00139980482479979, 0, 0, 0, 0.0311911607055895, 0, 0.019456436649588, 0.000111730481888879, 0.00868053124528075, -75972222.2166411, 0, 0, 0, 0.126104872444522, 0, 0.0936303423156659, 0.00957377187524537, -0.0116335171014229, 0.0418531193771755, 0.054179538170076, -123194444.414414, -123194444.362991, 0, 0, 0.165784238141865, 0.0201525396463447, -4861111.11327501, 0.0440216330881288, -4861110.97805741, -52083333.297789, -52083333.1768445, -0.0237565889604502, 0.0323334076983679, -104166666.590358, -0.0131880424185514, -14583333.3459322, -14583333.3057932, 0, 0, 0, 0, 0, 0, 0.00183217634440895, 0, 0.00106162913288371, 0, 0, 0, -0.0131825893759521, 0, -0.00807534471913274, -5.31608416936195e-05, -0.00357506549766602, 51388888.8874252, 0, 0, 0, -0.0807219554109035, 0, -0.0536033574105338, -0.00816886207499856, 0.00953324777833765, -0.0333867880726771, -0.0331708852646897, 116666666.6467, 116666666.609352, 0, 0, 0, -0.149435231211334, 0, -0.12753241417016, -0.0158049160798944, 0.0132478105106641, -0.0528650677852587, -0.109189090855372, 79166666.6235536, 79166666.5202037, 0.0197022504927858, -0.0408189341552075, 158333333.238713, -0.0324226016263352, 13888888.8665538, 13888888.7135491, 0.0247962821629065, 0.0439615012392089, -0.0131841682878408, -0.0155262709036165, 41666666.6289407, 41666666.5832542, 0.0517436187423386, 0, 0, 0, 0, 0, 0, -0.000523483815507861, 0, -0.000287138718895719, 0, 0, 0, 0.00201843768159741, 0, 0.00130792178385757, 4.15680208972412e-06, 0.000459154066687419, -13888888.8890075, 0, 0, 0, 0.0200322762745696, 0, 0.0123740102973774, 0.00269695928402069, -0.00287958408815068, 0.0101199063167398, 0.00774157811643882, -41666666.6619813, -41666666.6508355, 0, 0, 0, 0.0498097579888895, 0, 0.0369082325382156, 0.0059869010348401, -0.00725601102456854, 0.0224729826672264, 0.03401540251034, -41666666.6503792, -41666666.6149978, -0.00507528178156317, 0.0179658432825432, -83333333.2948062, 0, 0, 0.0584486596281275, 0.00942124733248974, 0.00203118274646791, 0.0194634953509398, 0.0714085313658215, -13888888.8708264, -13888888.8091027, -0.00914134306114917, -0.0243962271032299, 0.0201101075041748, 0.0040267337669946, -41666666.6294787, -41666666.6073625, 0.00104617937414821, -0.00188379501275091, 0.0264530642348098, -0.042079450250466, -0.0308540119659257, -0.00146875377232281}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -7.53906242154057e-07, 0, 0, 0, 0, 0, 0, 0, 1.51941199363172e-05, 0, 1.47539567734868e-05, -7499.99999415206, 0, 0, 0, 0, 0, 0, -9.30687508779675e-05, 0, -8.9940551173762e-05, 0, 0, -0.000299863482143476, 15833.3333066416, 15833.3332539629, 187083.333237234, 0, 0, 0, 0, 0, 0, 0.000256731573126272, 0, 0.000255881720800972, 0, 0, 0, 0.00182150745581841, 0, 0.00179869570147904, -8333.3332508562, -8333.3329196182, -1776111.11044441, -16666.666499838, -378194.444140023, -378194.443257377, 0, 0, 0, 0, 0, 0, -0.000323834977555814, 0, -0.000364786808402094, 0, 0, 0, -0.00524333514661426, 0, -0.0051764235652512, -0.000264751963175357, -0.00123634783846084, 8471527.77560144, 0, 0, -0.00961388501668544, -8.6050784263687e-05, 190277.777570232, -0.000767176680362238, 190277.773044912, 3349305.55296975, 3349305.54806649, 380555.554587558, 0, 0, 0, 0, 0, 0, 7.83005260739239e-05, 0, 0.000208924465661173, 0, 0, 0, 0.00814902497954828, 0, 0.00816421895440692, 0.000659490038437296, 0.00217025409164263, -22152777.7741731, 0, 0, 0, 0.0232511439441408, 0, 0.0230649548034479, 0.000535032418084858, -0.000799633759477667, 0.00333619761625902, 0.00991224961511152, -14340277.769147, -14340277.7555901, -0.000470375263869782, -1562499.99595402, -1562499.97652554, -0.00447164561606717, -0.000693839107673901, -3124999.98815581, 0, 0, 0, 0, 0, 0, 0.000242626122439417, 0, 6.65394929571348e-05, 0, 0, 0, -0.00695371384354971, 0, -0.00720203954970875, -0.000913813115920492, -0.00212902307843327, 31944444.441426, 0, 0, 0, -0.0287937556174607, 0, -0.0285316306080285, -0.00220775972876214, 0.00157829528932255, -0.00715289921060631, -0.0118074155806347, 31597222.2095901, 31597222.1894119, 0, 0, -0.0426935181156886, 0.00177630881068187, -0.00101066220276665, -0.00189092183298714, -0.034342372047017, 5902777.76357259, 5902777.72465949, 0.0103873255091246, 0.00199657805666231, 11805555.5168994, 0.019961401688929, -0.0012275699618158, -0.0120845245741935, 0, 0, 0, 0, 0, 0, -0.000250940887237961, 0, -0.000136848217452911, 0, 0, 0, 0.0030193751373645, 0, 0.00329653903628744, 0.000621832740217784, 0.00104954175133816, -23611111.1100026, 0, 0, 0, 0.0177593817494049, 0, 0.0178059249359399, 0.00269598770885426, -0.000991638487979997, 0.00669878310650624, 0.00744921983173904, -34027777.7695261, -34027777.7541218, 0, 0, 0, 0.0359663222365858, 0, 0.0355774063318854, -0.000101462469408202, -0.00251076507563038, 0.0059627253119063, 0.0249507794768862, -10416666.6503523, -10416666.6129308, -0.00702211199915818, 0.0015229036227166, -20833333.2881512, -0.00163988880724778, 0.00837719942735671, 0.0510474887689906, -0.0256652927994593, -0.0123252597237494, -0.0220985011672644, -0.000500421562066227, 0.025238454719747, 0.0419885308932496, -0.0215641268279666, 0, 0, 0, 0, 0, 0, 7.57394657297044e-05, 0, 4.62290423533921e-05, 0, 0, 0, -0.000507202362970798, 0, -0.000597181547464358, -0.000163587153773739, -0.000194528061090359, 6944444.44434925, 0, 0, 0, -0.00431271826836799, 0, -0.00443687239716188, -0.00103819857995904, 0.000195486329646986, -0.00224765763959079, -0.00190013891890647, 13888888.8869787, 13888888.8822859, 0, 0, 0, -0.0113795745282029, 0, -0.0112036247240714, -0.000764460857986725, 0.00146468885906351, -0.00361553484563758, -0.00726661646782248, 6944444.43842684, 6944444.42442813, 0.00115057134909741, -0.00222820128347297, 13888888.8711407, 0, 0, -0.0150599595012698, 0.000963477517084338, -0.00089187102382205, -0.000784655992012528, -0.0180150359168066, -0.00619594292119528, -0.0257140057458532, 0.00885102596253072, 0.00728750577917306, 0.00588723633536809, 0.00290477804759801, -0.0192835778139456, -0.0287117597674595, 0.0145523398297229, -0.00286115030895126, -0.0114603455047527, 0.0192778203316407, 0.0192569941499101, -0.00536778430451706}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -1.34250766563676e-07, 0, 0, 0, 0, 0, 0, 0, 2.64752222518493e-06, 0, 2.17151137198366e-06, 833.333334883042, 0, 0, 0, 0, 0, 0, -1.86979391456352e-05, 0, -1.44628620351071e-05, 0, 0, -3.82946909434247e-05, -19861.1111324187, -833.333350252836, -9166.66668183634, 0, 0, 0, 0, 0, 0, 7.20981414209747e-05, 0, 4.85703798856106e-05, 0, 0, 0, 0.000217107702425233, 0, 0.000235782013036483, 175277.777908883, 8.17414403436854e-05, 8333.33340372666, 19027.7779355733, 217638.889082681, 8333.33346132059, 0, 0, 0, 0, 0, 0, -0.000173024689355004, 0, -8.68158617432856e-05, 0, 0, 0, -0.0006284982443258, 0, -0.000744572043140785, -746527.778168299, -0.000273005967310734, -0.00022485451048879, 0, 0, -0.00110634994279305, -156250.000822914, -1909027.77883256, -0.000390857519287634, -0.000373976572137399, -190277.77844027, -0.000462712766309221, -190277.779015575, 0, 0, 0, 0, 0, 0, 0.000261215760699323, 0, 8.30701304330533e-05, 0, 0, 0, 0.00110725177281157, 0, 0.00125169559719301, 1631944.44506631, 0.000610158402720529, 0.000441395536193679, 0, 0, 0, 0.00236964515329019, 0, 0.00293875832606879, 590277.779802949, 8055555.55817202, 0.000392141877739271, 0.000912668937956076, 0.00127945419697916, 0.00143408854051662, 0.00164849855270802, 1562500.00296025, 0.00032317643864242, 1562500.005641, 0.00261662262512764, 0.00321290471051729, 0, 0, 0, 0, 0, 0, -0.000237574053003302, 0, -3.90457203201413e-05, 0, 0, 0, -0.001209098296062, 0, -0.0011276655145221, -1736111.1116552, -0.000791564087571937, -0.00047714588236543, 0, 0, 0, -0.00275098309798147, 0, -0.00397773512887894, -1041666.66914358, -17361111.1143305, -0.000181883980262141, -0.00185323050590844, -0.00151442916756055, -0.00262464546263176, 0, 0, -0.00473944282048513, -0.0033088209011582, -5902777.78359877, -0.000861700710744821, 0.000541857628102529, -0.0039470918647365, -0.000793146270190502, -5902777.78845591, -0.00255236712951706, -0.00513866041766885, -0.0104064869973191, -0.0113737177410384, -0.00346230355331425, 0, 0, 0, 0, 0, 0, 0.000117628333033161, 0, 6.12348095710188e-06, 0, 0, 0, 0.000742611108953749, 0, 0.000508262208663303, 694444.444692486, 0.000524744595063518, 0.000256464948025436, 0, 0, 0, 0.00173981005405825, 0, 0.00260411411256444, 694444.445905908, 18055555.5574662, 9.09783988136817e-05, 0.00196433061639799, 0.000974042641957684, 0.00226397904115978, 0, 0, 0, 0.00352203370903625, 0, 0.0044177456043491, 0.00290074307386591, 10416666.6717532, -0.000517632662006393, -0.000204813845139159, 0.00262861704092801, 0.00180273160666077, 10416666.6755108, 0.00156769510685373, 0.00515377670294066, 0.00143642207056333, 0.00526515750147716, -0.00148050657090508, 0.0140636294295849, 0.0148544146006603, 0.00734250441015325, 0.000157119482205517, 0.0102736088998165, 0.002393788629807, 0.0133604625118177, 0, 0, 0, 0, 0, 0, -2.41570797137194e-05, 0, 5.21970723733381e-07, 0, 0, 0, -0.000193033862923385, 0, -8.80489557111804e-05, -4.65166579335325e-05, -0.000136476330712204, -5.28747937668452e-05, 0, 0, 0, -0.000476848898003521, 0, -0.000654504835091943, -0.000331255075147722, -6944444.44487908, -4.54163337197572e-05, -0.000765629361463152, -0.000248879338232496, -0.000718901568357564, 0, 0, 0, -0.00102405680349436, 0, -0.00150416807378004, -0.000908888290310744, -6944444.44605527, 0.000421828265801738, -0.000299722168064254, -0.000704306463712411, -0.00105833779362186, -6944444.44709189, -0.000621753755566113, -0.00213842179415284, 0, 0, -0.00165431617084397, -0.000763437252883487, -0.00241548046115052, 0.000220068557873687, 0.000949795940442273, -0.00169119070581923, 0.00031145520535233, -0.0060460306264094, -0.00639775608892646, -0.00163052436593819, 0.000564520310518043, -0.00396957940887613, -0.00143076376806642, -0.00535963165182314, -0.00645238286689413, 0.00134549942142493, -0.00980338481306443, -0.00457450312645527, -0.00530232324877518}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -1.06332081748791e-07, 0, 0, 0, 0, 0, 0, 0, 1.42000363081562e-06, 0, 2.90531288719592e-06, 666.666671277122, 0, 0, 0, 0, 0, 0, -3.46985499622951e-06, 0, -2.25295957559356e-05, 0, 0, -4.59602369245464e-05, -14555.5556140873, -666.666732681069, -10666.6667102654, 0, 0, 0, 0, 0, 0, -1.37012897331249e-05, 0, 7.7499384851133e-05, 0, 0, 0, 0.000222065562808241, 0, 0.000351888774956148, 111111.111414029, 0.000362563924367446, 43333.333491496, 13888.8895385687, 232222.222707076, 10000.0005896646, 0, 0, 0, 0, 0, 0, 8.70759235121603e-05, 0, -0.000134390896284849, 0, 0, 0, -0.000469627368453573, 0, -0.00116131086641994, -375000.000818857, -0.00100849942292095, -33333.3336077472, 0, 0, -0.00163291480006291, -97222.2247490032, -1763888.89120225, -0.00295264501516998, -0.00281025755835426, -936111.112557724, -33333.3354943992, -208333.337840874, 0, 0, 0, 0, 0, 0, -0.000191578317418394, 0, 0.000114141849753063, 0, 0, 0, 0.000452123217128947, 0, 0.00195417804646905, 555555.556812032, 0.00155278083256055, 0.000208088563517556, 0, 0, 0, 0.00337730877549432, 0, 0.00468232723983176, 277777.782839941, 5902777.78317102, 0.00680700877692767, 0.00639434950923641, 694444.446423806, 0.00414667639302732, 0.00863032821984925, 7013888.89506297, 0.00881991837380321, 1458333.34831773, 0.0166022223573138, 694444.455769397, 0, 0, 0, 0, 0, 0, 0.000214022754600987, 0, -3.08696246529566e-05, 0, 0, 0, -9.68073725708761e-05, 0, -0.00173606163796578, -277777.778869064, -0.00131471007177138, 7.12298873340642e-06, 0, 0, 0, -0.00355996291986079, 0, -0.00639783458689669, -277777.783386906, -8611111.11769724, -0.00849536684976, -0.00751628438684191, -0.00118817181012354, -0.00439719863157169, 0, 0, -0.00776403527958157, -0.0121205946181635, -22916666.6783864, -0.0142282138200043, -0.0158235180918405, -4861111.11850705, -0.0135675554185107, -4166666.69028198, -0.0285369752559898, -0.0134417396472093, -0.0392860054130936, -4861111.14036967, -0.0322499558063949, 0, 0, 0, 0, 0, 0, -0.000121580479041441, 0, -1.45546888640619e-05, 0, 0, 0, -0.000135055009199998, 0, 0.000753998608493559, 0.00049423054532799, 0.000553772940922062, -0.000101852695674845, 0, 0, 0, 0.00182508473992766, 0, 0.00419604523998485, 0.00324982838300452, 4166666.6707052, 0.00546514180056019, 0.0043598372323784, 0.000127252617258741, 0.00242045494267317, 0, 0, 0, 0.00591719522836873, 0, 0.00733186779797863, 0.00853144210554112, 31944444.4545947, 0.0115205413594176, 0.0130414993828742, 0.00396232910949287, 0.0101858988346407, 4166666.6844824, 0.0230604376110966, 0.00793468829245714, 0.0123810132648976, 13888888.8991707, 0.016893558268235, 0.0390903805675422, 13888888.9210646, 0.0437120499057412, 0.0380255934951606, 0.0235592136414057, 0.0262853202534113, 0.0517954206096638, 0, 0, 0, 0, 0, 0, 2.79191389930407e-05, 0, 7.90455689183598e-06, 0, 0, 0, 7.05616173276746e-05, 0, -0.000118535359049886, -8.93956510356466e-05, -8.42964495056363e-05, 4.15840576199969e-05, 0, 0, 0, -0.00034248643885576, 0, -0.00105932472453543, -0.000761479201633201, -0.000976952845449098, -0.00141253601337971, -0.000971864936677444, 0.000102642513195007, -0.000526032678810836, 0, 0, 0, -0.00170948674318721, 0, -0.00248167257854493, -0.00242987048034726, -13888888.8921704, -0.00370121042568124, -0.00399438767508065, -0.000707116643078314, -0.00301132897981357, -0.00519431754773142, -0.00709805839889608, -0.00193037655852424, 0, 0, -0.00287314331761457, -0.00404832236417061, -13888888.8936203, -0.00468781852988799, -0.00729229375171145, -0.00311870938304399, -0.00645828538787715, -0.0140313675070078, -13888888.9014904, -0.0173124572391078, -0.015616411641122, -0.00694405769525936, -0.00790248742860519, -0.0209581282396716, -0.013531711470995, -0.0159857568245282, -0.026715019562556, -0.0295552098335198, -0.0207948275078152}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 2.80965799276554e-07, 0, 0, 0, 0, 0, 0, 0, -7.0469513861568e-06, 0, -2.80600942737665e-06, 625.000004125921, 0, 0, 0, 0, 0, 0, 6.13608981009106e-05, 0, 9.57590738971391e-06, 0, 0, 7.06767325263589e-05, -12083.3333988277, -625.000040562238, -12083.3333809024, 0, 0, 0, 0, 0, 0, -0.00026864975721447, 0, -1.78013492153414e-05, 0, 0, 0, -0.000554719438792667, 0, -0.000256568329588594, 73958.3336624172, 0.000191234395742612, 73958.3335202082, 11458.3339772978, 232986.111818909, 11458.3336618397, 0, 0, 0, 0, 0, 0, 0.000666652909312727, 0, 3.17374287122327e-05, 0, 0, 0, 0.00210751880589013, 0, 0.000503037425373011, -166666.667482766, -0.000374893996426589, -166666.667044002, 0, 0, 0.00175093354448375, -62500.0026169267, -1418402.7810804, -0.0028271628930271, -0.00158289172432029, -1418402.78039894, -62500.0003905901, -210069.449631358, 0, 0, 0, 0, 0, 0, -0.000982454163004325, 0, -6.22000487287637e-05, 0, 0, 0, -0.00431679237848005, 0, -0.000670058347170856, 104166.667787993, 0.000173161298324479, 104166.667154948, 0, 0, 0, -0.00554977347100275, 0, -0.00272586106066079, 104166.67162427, 3159722.22942891, 0.00611007404879664, 0.00337889260030719, 3159722.22711278, 104166.665025681, 0.00882342328082926, 8541666.6773585, 0.00278249096690882, 1145833.35281767, 0.0188659335172526, 1145833.34496107, 0, 0, 0, 0, 0, 0, 0.000850572994144046, 0, 8.10397244283135e-05, 0, 0, 0, 0.00492357397784596, 0, 0.000687124441882041, -0.000864065806477112, 0.00039317288118491, -0.000471120806060463, 0, 0, 0, 0.00869094355356384, 0, 0.00230311761504988, -0.00490770787465011, -1909722.23032295, -0.00677837675587634, -0.00311217942956057, -1909722.22751111, 0.00495957242806419, 0, 0, 0.00685157546932604, -0.0118629679721306, -18576388.9076318, -0.0141438594414642, -0.00628916955291209, -18576388.9048818, 0.000712561776488766, -1909722.25349692, -0.0321979202519698, -1909722.23048637, -0.0499125925668912, -6250000.03955968, -0.0334717378010283, 0, 0, 0, 0, 0, 0, -0.000398050794332907, 0, -5.32449498098709e-05, 0, 0, 0, -0.00295327999536047, 0, -0.000472576460479331, 0.000346204204879951, -0.000548925522865398, 0.000317996269258193, 0, 0, 0, -0.00664517415487908, 0, -0.00107210850531366, 0.00246311240320716, 0.00452389688152807, 0.00363600459553584, 0.000901038815636788, 0.00330416168000425, -0.00491360645990377, 0, 0, 0, -0.00737999122661397, 0, -0.00405678249228216, 0.00737254510259897, 10416666.6816153, 0.0107390090810901, 0.00508050489798449, 10416666.6784708, -0.00490010472023723, 0.0230214900432581, 0.025162018654267, -0.00108527303930834, 0.0121609085781461, 38194444.4640126, 0.00288798517091525, 0.0509413025591516, 10416666.7135881, 0.0545877109755199, 0.0415901401763847, 10416666.6960596, 0.0180031642508522, 0.067401286202606, 0, 0, 0, 0, 0, 0, 7.73327649380873e-05, 0, 1.34160747540084e-05, 0, 0, 0, 0.00072564469423342, 0, 0.000145158073270243, -5.54486711019559e-05, 0.000202412846952862, -0.000101492067939892, 0, 0, 0, 0.00199874498098126, 0, 0.00024228271449029, -0.000493736317040477, -0.000985439022049307, -0.000714183528168321, 0.000134590973026964, -0.000936757716586454, 0.00169356476088354, 0, 0, 0, 0.00283697683566838, 0, 0.000923803005732268, -0.00174320538156114, -0.0044451655485822, -0.00310404904805963, -0.00126004902977835, -0.00355909066246791, 0.0028843805155266, -0.00639405066393078, -0.00728596852325212, 0.00229873472245178, 0, 0, 0.00229980730294487, -0.00355946501929015, -17361111.1192316, -0.0045178806981707, -0.00185310864571285, -17361111.11846, 0.000534852841648408, -0.0180263664543346, -0.0185665956490052, -0.0210029793248685, -0.0174075215005852, -0.00675061009704559, -0.000333921499547709, -0.0268699153104548, -17361111.1312382, -0.0138693632654915, -0.0375313544619585, -0.0383965252638859, -0.024181045109086}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 4.67387765435377e-07, 0, 0, 0, 0, 0, 0, 0, -9.58249813303771e-06, 0, -8.23652785112647e-06, 666.666667290149, 0, 0, 0, 0, 0, 0, 6.55805494814322e-05, 0, 5.21981053378534e-05, 0, 0, 0.000159318939407173, -10666.6666920935, -666.666644814624, -14555.5555639989, 0, 0, 0, 0, 0, 0, -0.000231339914485392, 0, -0.000174891851848068, 0, 0, 0, -0.000964382894697297, 0, -0.000930952042265958, 43333.3335365208, -0.000162368428165739, 111111.111148964, 9999.99998991639, 232222.222488425, 13888.888595228, 0, 0, 0, 0, 0, 0, 0.000469053097928117, 0, 0.000340868719177291, 0, 0, 0, 0.00297257299128832, 0, 0.00286737751304477, -33333.3340858167, 0.000578580821128978, -375000.000132867, 0, 0, 0.00461085255214785, -33333.3338215106, -936111.112798474, 0.000624971955849332, 0.00134385377502901, -1763888.88984766, -97222.2202987517, -208333.333238223, 0, 0, 0, 0, 0, 0, -0.000558599107032874, 0, -0.000387644037731167, 0, 0, 0, -0.00520138916979814, 0, -0.00512355394947413, 0.00146479788199347, -0.00113581736193114, 555555.556005102, 0, 0, 0, -0.0111279506714722, 0, -0.0111178448228248, 0.00210347192121398, 694444.449536999, -0.00245957074788426, -0.00330331562396753, 5902777.77983924, 277777.771115777, -0.000710253700348595, 7013888.89342825, -0.00527417308479292, 694444.447625579, -0.00218553470465364, 1458333.33045166, 0, 0, 0, 0, 0, 0, 0.000375772540224271, 0, 0.00024112428707588, 0, 0, 0, 0.00521146534078349, 0, 0.00529218012854526, -0.00156362079279162, 0.00120904404716291, -277777.778664746, 0, 0, 0, 0.0144156947416769, 0, 0.0147176801351771, -0.00346775345140557, -0.0076061146950964, 0.00418099549690839, 0.00495814990458195, -8611111.11441713, -277777.766354711, 0, 0, 0.0194493374738081, -0.000966442466443383, -4861111.12162335, 0.00399068427901529, 0.00651039397686559, -22916666.6728322, 0.0134385324942414, -0.0112369386802107, 0.00616117392598394, -4166666.65537712, 0.00087336719929463, -4861111.11217136, 0.00513579990372235, 0, 0, 0, 0, 0, 0, -0.000125184759487072, 0, -6.75277400203359e-05, 0, 0, 0, -0.00277975912274288, 0, -0.00291374871116496, 0.000872861785852377, -0.00064709738053506, 0.000825289932842159, 0, 0, 0, -0.00960461163614403, 0, -0.0101329253453644, 0.0025830979388915, 0.00545275658163068, -0.00327040804000325, -0.00384755039776251, 4166666.66986037, -0.00919555713036871, 0, 0, 0, -0.0162761506823551, 0, -0.0166371498521314, 0.00197278528643079, 0.0109117563387654, -0.00509258468650594, -0.00527188010860385, 31944444.4496729, -0.0167385559665965, 0.0133083850965175, -0.00778252262325095, 4166666.65255917, -0.00192004331124718, 13888888.8979651, -0.00892410940841378, 0.00556941734822274, 0.0109019364529761, -0.0077941410427993, -0.00143757000449587, 13888888.8802439, -0.0128489929287175, -0.00371875124606162, 0, 0, 0, 0, 0, 0, 1.38325871657673e-05, 0, 3.638769586391e-06, 0, 0, 0, 0.00061018855058729, 0, 0.000658868271154489, -0.00020033553037993, 0.000135025285291877, -0.000285228377360002, 0, 0, 0, 0.00258266392200034, 0, 0.0028199046667324, -0.000734776584771091, -0.00150287602334727, 0.000960550897170103, 0.0011525152347181, -0.00123987446061168, 0.00277314242742616, 0, 0, 0, 0.00528506296949266, 0, 0.00559705551718168, -0.000837019685630665, -0.00405610279943529, 0.00208701527228651, 0.00207666509852663, -13888888.8910538, 0.0074442192506626, -0.0052161828401655, 0.00340356885443068, 0.0056048075993306, 0, 0, 0.00673113457216635, 8.17269458678864e-05, -0.00477224044202481, 0.00163210267997289, 0.00241394695746529, -13888888.8921267, 0.00576409285877819, -0.00475587863102887, -0.00822800786504241, 0.00575102805610877, 0.000486452107287422, -13888888.882286, 0.00969162587060182, 0.00169851451166424, -0.000461916123175081, 0.000816912984699689, -0.00214595988702101, 0.00207766904561902, 0.00733069378640301}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 2.97721190413026e-07, 0, 0, 0, 0, 0, 0, 0, -5.777206590675e-06, 0, -6.0611493738203e-06, 833.33333150695, 0, 0, 0, 0, 0, 0, 3.38845142622456e-05, 0, 3.85025300815923e-05, 0, 0, 0.00011823210529841, -9166.66666354294, -833.333290103927, -19861.1110890821, 0, 0, 0, 0, 0, 0, -8.57899297785091e-05, 0, -0.000116750722994773, 0, 0, 0, -0.000698830284436675, 0, -0.000726926328682349, 8333.333357414, -0.000281588471173115, 175277.777649104, 8333.33316133417, 217638.888892309, 19027.7773003369, 0, 0, 0, 0, 0, 0, 8.30291439225562e-05, 0, 0.000184394533728369, 0, 0, 0, 0.00195169764140297, 0, 0.00216675608965191, -0.000140174797021569, 0.000922627958781782, -746527.777429853, 0, 0, 0.00376474043520406, 0.000342270018132952, -190277.778108203, 0.0012225786398301, 0.00232506448134585, -1909027.77764986, -156249.997402856, -190277.777043258, 0, 0, 0, 0, 0, 0, 4.65354562519024e-05, 0, -0.000137613505516342, 0, 0, 0, -0.00295284856933173, 0, -0.00354435373198879, 0.000320932076174621, -0.00168220850572185, 1631944.44406862, 0, 0, 0, -0.00876540839871427, 0, -0.00936180700421604, -0.000378481493448825, 0.00122024086197006, -0.00398037182419325, -0.00548962043870374, 8055555.55498593, 590277.770451162, -0.00246912276795681, 1562500.00049991, -0.00945948762690176, 0.000113520034451324, -0.00370226463657018, 1562499.99665741, 0, 0, 0, 0, 0, 0, -0.000177950889024071, 0, 1.41960774015832e-05, 0, 0, 0, 0.00246125855373932, 0, 0.00322823217788475, -0.000384446721265425, 0.00170234715631351, -1736111.11111901, 0, 0, 0, 0.01047539343277, 0, 0.0120690306377803, 0.000281803881096091, -0.00201381790445065, 0.00647145917781306, 0.00704179875659632, -17361111.1105041, -1041666.65616134, 0, 0, 0.016621566243819, 0.0026220321882975, -0.00192174886371751, 0.00662772957682109, 0.0150042692672818, -5902777.77753078, 0.0202447250566972, -0.00178985398008022, 0.00911422522504476, -5902777.76829702, 0.00267360074123765, 0.000569451021932796, 0.00738673721544787, 0, 0, 0, 0, 0, 0, 0.000148130244529762, 0, 3.9451370192552e-05, 0, 0, 0, -0.00103953151974264, 0, -0.00151830759711433, 0.000238454013376342, -0.000883740190183714, 694444.44472741, 0, 0, 0, -0.00628492009293767, 0, -0.00781873629919208, -0.000149244935570745, 0.00159204452621509, -0.00505863964536061, -0.00464175825850099, 18055555.5555574, 694444.437189637, 0, 0, 0, -0.0133901409395365, 0, -0.0144804714039163, -0.00157899198108513, 0.00240951228116081, -0.00761047959769147, -0.0115155864320022, 10416666.6661258, -0.0203135316728982, 0.00207911541744589, -0.0105051652333909, 10416666.6557648, -0.00479000318336043, 0.000695019929984297, -0.0206252725903038, -0.000348117019354615, 0.00235111546337081, -0.00779041261414242, -0.00752493673806875, -0.00499956290337342, -0.013354973763798, 0.000360274517759126, 0, 0, 0, 0, 0, 0, -4.23574584350424e-05, 0, -1.64165424342648e-05, 0, 0, 0, 0.000165840215979132, 0, 0.000282350812863363, -6.04489723940379e-05, 0.000180861949503041, -0.00013838080736029, 0, 0, 0, 0.00149737747320161, 0, 0.00200648257229027, 3.77332707787113e-05, -0.000488276056454999, 0.00151355217758374, 0.00121767029682484, -6944444.44462467, 0.00190139020227135, 0, 0, 0, 0.00407171952109314, 0, 0.00477077653409366, 0.0004504316418229, -0.00102925650461097, 0.00309670475592162, 0.00349122089248962, -6944444.44436316, 0.00750854726992649, -0.000776547204853189, 0.00436851044801325, -6944444.44025302, 0, 0, 0.00584559789437991, 0.00134098723821073, -0.000688641435573116, 0.00279500134083472, 0.00744321306179135, 5.44817755654892e-05, 0.0103776918286095, -0.000248267279084931, -0.00157218461581427, 0.00565544755693086, 0.00357778657508399, 0.00383983599384865, 0.00855121194492262, 0.00240754765992172, 0.000611687045391329, 0.00505295215837689, -0.00192794006241421, 0.000156221927275877, 0.00184837776189296}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 1.04098647852705e-08, 0, 0, 0, 0, 0, 0, 0, 1.8810585911491e-06, 0, -2.20777455171295e-06, 833.333334209412, 0, 0, 0, 0, 0, 0, -2.80900126201104e-05, 0, 2.75149667917931e-05, 0, 0, 3.82199016820964e-06, -833.333336773654, -19861.1111529646, -9166.66665507858, 0, 0, 0, 0, 0, 0, 0.00014316524286803, 0, -0.000140573011780309, 0, 0, 0, 0.000194544635853723, 0, -0.000203261462383193, 2.64314700843474e-07, 175277.778155065, 8333.33320458901, 19027.7780531394, 8333.33319993464, 217638.889036917, 0, 0, 0, 0, 0, 0, -0.000373812856169269, 0, 0.000377615691266345, 0, 0, 0, -0.00106699161811838, 0, 0.00116367750800314, 3.90495487101043e-05, -746527.779251702, 0.000513158951401024, 0, 0, -0.000193115664190962, -0.000919153383557391, 0.000577492528892112, -156250.002020942, -1909027.77958026, 0.00104344478419268, -190277.777755937, -190277.778421477, 0, 0, 0, 0, 0, 0, 0.000557058265241867, 0, -0.000571195543723747, 0, 0, 0, 0.00244556431113371, 0, -0.00299614772694699, -0.000139198851665248, 1631944.44754388, -0.00102785567389902, 0, 0, 0, 0.0024809999279595, 0, -0.0012401234400659, 0.00168309675724403, -0.00122771720558724, 590277.783880184, 8055555.56143438, -0.00303884052836397, -0.0015460199469171, 0.00539883791299718, -0.00351098054689933, 1562500.00450504, 0.00150321897466312, 1562500.00698309, -0.00106759174813371, 0, 0, 0, 0, 0, 0, -0.000477571298973007, 0, 0.000482323365935283, 0, 0, 0, -0.00289396035547245, 0, 0.00398580728373412, 0.000221143184279925, -1736111.11475229, 0.00106678533264085, 0, 0, 0, -0.00498275642314377, 0, 0.00389415682450722, -0.00175571908278119, 0.00153076573181546, -1041666.67603045, -17361111.1202499, 0.00413396145171661, 0.00425086958997009, 0, 0, -0.00195423833421856, -0.0073926023195533, 0.00508480882285981, -0.0118947102973304, -5902777.7898378, 0.00784271452795055, -0.00553425285453561, -0.00174520543914384, -5902777.7940639, 0.00590992132834992, -0.0148047556577934, 0.0047232604534783, -0.0108015147535578, 0, 0, 0, 0, 0, 0, 0.000218626793562373, 0, -0.000207958985367685, 0, 0, 0, 0.00173720272816585, 0, -0.00265383888559661, -0.000169780659020679, 694444.446684324, -0.000524328268828738, 0, 0, 0, 0.00416170412111922, 0, -0.00420345005206747, 0.00100577096785303, -0.00108245244156782, 694444.451649083, 18055555.5626443, -0.00257843401122138, -0.00452604809008237, 0, 0, 0, 0.00399503545761246, 0, -0.000408801100568001, 0.005106075404305, -0.00350806763540171, 0.011786413908308, 10416666.6787525, -0.00734443791432305, 0.00194314793362077, 0.000672242863694174, 10416666.68253, -0.00781589308413604, 0.010528042637869, -0.00753073940570976, 0.0124729991061372, 0.0143152021762269, -0.00560644942984263, 0.0232947946003613, 0.0192489475755606, -0.0126407497513106, 0.00784255813741669, 0.0118330765198337, 0, 0, 0, 0, 0, 0, -4.12700606984071e-05, 0, 3.44682658027448e-05, 0, 0, 0, -0.000417923431401706, 0, 0.000697619006742516, 5.11320219839444e-05, -0.000560438015375107, 8.80751316291596e-05, 0, 0, 0, -0.00129015777569015, 0, 0.00157680751691617, -0.000247473321358609, 0.000327374079355096, -0.0021968418017207, -6944444.44663136, 0.00057062362142868, 0.00168794746217601, 0, 0, 0, -0.00183466541154694, 0, 0.000858709248794974, -0.00145310826523775, 0.000992361722067836, -0.00440377581394154, -6944444.44889523, 0.00240106179365085, 0.00062552288796045, 5.01944103420364e-05, -6944444.4502765, 0.00337682947163986, 0, 0, -0.000976117859327893, -0.00348360161132204, 0.00243910718362843, -0.00505450971871925, -0.00632925550200748, 0.00361098727952894, -0.00447632456275142, -0.0048640973999463, 0.0023183917263929, -0.0110520823265792, -0.00901667220674578, 0.0074116155671049, -0.00165426159675839, -0.0115385506575464, 0.00693310017836334, -0.00936314098474306, -0.00510858492946852, -0.0122694283543798, -0.00148793292971457}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -6.14950334571901e-07, 0, 0, 0, 0, 0, 0, 0, 9.52679021134892e-06, 0, 1.04228724229058e-05, 666.666677112097, 0, 0, 0, 0, 0, 0, -7.15320133501263e-05, 0, -7.89368708385405e-05, 0, 0, -0.000118247348547845, -666.666714058457, -14555.5557298214, -10666.6668090692, 0, 0, 0, 0, 0, 0, 0.000297693353017009, 0, 0.000331846995320362, 0, 0, 0, 0.000689748306271022, 0, 0.000665412471189853, 0.000116029770634058, 111111.112263197, 43333.3341999529, 13888.8896902368, 10000.0004861872, 232222.224049245, 0, 0, 0, 0, 0, 0, -0.000738241886323548, 0, -0.000842310911787013, 0, 0, 0, -0.00222289581382319, 0, -0.0020466256431155, -0.000178707313501407, -375000.003932217, -33333.3361608163, 0, 0, -0.00300550996998591, -0.00190836573417988, -0.000926368308251729, -97222.2269681341, -1763888.89821021, -33333.3356375673, -936111.119736209, -208333.339404237, 0, 0, 0, 0, 0, 0, 0.00111902607368318, 0, 0.0013260164908085, 0, 0, 0, 0.00413329611780614, 0, 0.00366586430119952, 0.000121732825894647, 555555.563151691, 0.00533970114276386, 0, 0, 0, 0.00739268864920181, 0, 0.00692201572074049, 0.00298693543343828, 0.00143762686352887, 277777.791342296, 5902777.80135898, 0.00569268285366561, 694444.465493945, 0.0093194552002002, 0.00323925488880737, 7013888.92237439, 0.00917075792739761, 1458333.36074144, 694444.466406303, 0, 0, 0, 0, 0, 0, -0.00101752344998033, 0, -0.00126698452324448, 0, 0, 0, -0.00442379806181867, 0, -0.00385043071072483, 7.60078096976499e-05, -277777.786166638, -0.00587779691702033, 0, 0, 0, -0.00990683093419094, 0, -0.00865086860907435, -0.00316270070736516, -0.00154539642153899, -277777.798092623, -8611111.14278957, -0.00778919703350589, -0.0276575250568337, 0, 0, -0.01235851801282, -0.0106106502439687, -0.00419102890311594, -0.0203544330906536, -22916666.7273449, -0.00502717657022454, -4861111.17004287, -0.00786407140909381, -4166666.72202803, -0.0397254432779305, -0.0342202999230763, -0.0229536142605076, -4861111.18193915, 0, 0, 0, 0, 0, 0, 0.000509898669572751, 0, 0.000672274599578751, 0, 0, 0, 0.00253522679330745, 0, 0.00221216865484592, -0.000170435231137712, 0.00494285238779682, 0.00350720150862261, 0, 0, 0, 0.0067945308869523, 0, 0.00558657080495493, 0.00196130397327246, 0.000805734567744802, 0.0152965109368035, 4166666.68833125, 0.00564755692158493, 0.0186840496294707, 0, 0, 0, 0.0108165130438731, 0, 0.00996087582157409, 0.00714995876553443, 0.0038739112471377, 0.0207252475453525, 31944444.496424, 0.00375093119265997, 0.0496109819449812, 0.00398796675691522, 4166666.71865093, 0.0347486405081044, 0.0149231595877207, 0.00393990659715294, 13888888.9579646, 0.0204079154978681, 0.0119851522715386, 0.0492392972025201, 13888888.9790513, 0.0262829679970299, 0.0825753916943187, 0.055356961266373, 0, 0, 0, 0, 0, 0, -0.00010823021659253, 0, -0.000151717546899874, 0, 0, 0, -0.000603755768289481, 0, -0.000538452546188493, 7.29113726955916e-05, -0.00120621679879876, -0.000877025156214697, 0, 0, 0, -0.00186644110103796, 0, -0.00147042292343366, -0.000510348546796333, -0.000122055640890315, -0.00456354707680871, -0.00594635382609105, -0.00169111988276981, -0.00510905995696862, 0, 0, 0, -0.00364838160367005, 0, -0.00309309459704663, -0.00220517421579301, -0.00139604619910754, -0.00794004462915273, -13888888.905888, -0.00118099109423726, -0.016054833690203, -0.000932809227097394, -0.0184755959452966, -0.0117857227237368, 0, 0, -0.00428687184293375, -0.00463146035689391, -0.00206611643104142, -0.00748301805214169, -13888888.9177596, -0.00107560426638609, -0.0282966616277923, -0.00501660222812666, -0.00291209832330965, -0.0243186522705197, -13888888.9289136, -0.0109856762136368, -0.0350045255837757, -0.0154276302058482, -0.00634030241444913, -0.0506047658066546, -0.0151455498050381, -0.0376612095434835, -0.0315439675744684}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -3.30688530479258e-07, 0, 0, 0, 0, 0, 0, 0, 3.61035785208574e-06, 0, 5.97050495425178e-06, 625.000008910033, 0, 0, 0, 0, 0, 0, -2.7857754551282e-05, 0, -5.47649233540876e-05, 0, 0, -2.8769144303256e-05, -625.000049666136, -12083.3334829428, -12083.3334525658, 0, 0, 0, 0, 0, 0, 0.000139606249673745, 0, 0.000278197032877629, 0, 0, 0, 8.24894125128583e-05, 0, 0.000184880511952593, 0.000123875316976452, 73958.3342560726, 73958.3340497182, 11458.3342587706, 11458.3337898122, 232986.112680768, 0, 0, 0, 0, 0, 0, -0.000427234431860465, 0, -0.000809561964648614, 0, 0, 0, -0.000259084364994946, 0, -0.00089237557119574, -0.000178232930982999, -166666.669609304, -166666.669040772, 0, 0, -5.43347923465918e-05, -0.00251712858616, -0.000639707799284965, -62500.0051724479, -1418402.78522245, -62500.0019896804, -1418402.78502301, -210069.451582368, 0, 0, 0, 0, 0, 0, 0.000775517746575071, 0, 0.00139359180119836, 0, 0, 0, 0.000820977418801519, 0, 0.00230790746042781, 0.00016607990831603, 104166.672043072, 104166.671238351, 0, 0, 0, -0.000907346171193654, 0, 0.000738647634853112, 0.00392796280783064, 0.000273824399936802, 104166.680311024, 3159722.23985858, 104166.67143128, 3159722.24029391, 0.0119863581626646, 0.00168997920973629, 8541666.69192041, 0.0128114117500957, 1145833.36444051, 1145833.35777825, 0, 0, 0, 0, 0, 0, -0.000809932390689177, 0, -0.0014091777951169, 0, 0, 0, -0.00148484678650221, 0, -0.00311622766489342, -8.65712577063793e-05, -0.00567269077885225, -0.00509214376673035, 0, 0, 0, 0.00158582220681986, 0, -0.0025207216615372, -0.00387367663035385, 5.12265555546297e-05, -0.0188358714088318, -1909722.24494905, -0.00633091728635661, -1909722.24710454, 0, 0, 0.00169998971379048, -0.0144952264591454, 0.000264669253818759, -0.0243220399421502, -18576388.929886, -0.00243747520826421, -18576388.9333808, -0.0108060947403649, -1909722.28185144, -1909722.26515205, -0.047985836741183, -0.0306470905183534, -6250000.07512293, 0, 0, 0, 0, 0, 0, 0.00044871305565576, 0, 0.000775120556615297, 0, 0, 0, 0.00126812847095525, 0, 0.00211459562530745, 2.28601527819819e-06, 0.00321742475006095, 0.00305132564237711, 0, 0, 0, -0.000758964506530176, 0, 0.00282385072629709, 0.0022784925205895, 3.44704454784729e-05, 0.0131198776948543, 0.015217477936832, 0.00436864386555766, 0.0176368362614597, 0, 0, 0, -0.00311941471533365, 0, -0.000103873776832475, 0.00950284061596641, -0.000805131349834939, 0.0226985613353589, 10416666.6991106, 0.00173124780487075, 10416666.7062089, 0.00425776721432562, 0.0533510775425768, 0.037374164183447, 0.0196497128964049, -0.000954101554740214, 38194444.4890139, 0.0319156746265998, 0.0153592497137086, 0.0654845354385035, 10416666.7568809, 0.0345352908338631, 10416666.7497988, 0.0737393902519118, 0, 0, 0, 0, 0, 0, -0.0001020929314367, 0, -0.000179041614711577, 0, 0, 0, -0.000404856811384145, 0, -0.000574428646197566, 1.37143395657547e-05, -0.000760447310117064, -0.000763783824585204, 0, 0, 0, 1.25631113999025e-08, 0, -0.00104626920793417, -0.000590321579774983, -5.89174869861944e-05, -0.00363883428302025, -0.00414251777524616, -0.00122488348075167, -0.00501075323553333, 0, 0, 0, 0.00127009837667071, 0, -0.000574326420287934, -0.00273038626039049, 0.000156139592984608, -0.00797485988686023, -0.0101574872105852, -0.000469477217101834, -0.0137971325987096, -0.000702299975468129, -0.0181127953914864, -0.0126816299882311, 0, 0, 0.00104387470064515, -0.00628192510607463, 0.000703926231465942, -0.00914925207646861, -17361111.1272696, 0.000649931117983908, -17361111.1299931, -0.00781582330258765, -0.00213463819703391, -0.0304661070653902, -0.0383983725338929, -0.014501644015572, -0.0348369939201228, -0.0233250286840503, -0.00853280575168678, -17361111.1578683, -0.0231184028668657, -0.0478448300396281, -0.0395600750291635}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -7.38177622023336e-08, 0, 0, 0, 0, 0, 0, 0, 1.05572592846521e-08, 0, 2.18187108078567e-06, 666.666669987358, 0, 0, 0, 0, 0, 0, 5.62565750884926e-06, 0, -2.41790991694371e-05, 0, 0, -7.44839154262293e-06, -666.666685209153, -10666.6667283217, -14555.5556028236, 0, 0, 0, 0, 0, 0, -3.01173254855177e-05, 0, 0.000126540527422843, 0, 0, 0, -8.0500531752084e-05, 0, 0.000144068290122397, 4.5681067616537e-05, 43333.3336712482, 111111.111368384, 10000.0003672239, 13888.8890652339, 232222.223040129, 0, 0, 0, 0, 0, 0, 5.58289491469383e-05, 0, -0.000348860778037276, 0, 0, 0, 0.000545391345485758, 0, -0.000851485012198669, -0.000103181546811344, -33333.3342462618, -375000.000700087, 0, 0, 4.57523567817068e-05, -0.000834015063714704, -0.000182638249526277, -33333.3353524484, -936111.114744556, -97222.2228945951, -1763888.89305568, -208333.336838337, 0, 0, 0, 0, 0, 0, -3.25546659519301e-05, 0, 0.000546910976525731, 0, 0, 0, -0.00117138216320773, 0, 0.00212988781710386, 0.000237575173200035, 0.0013650182427924, 555555.556525521, 0, 0, 0, -0.00199534953256703, 0, 0.00158976534237399, 0.001086728174621, 4.72840634548271e-06, 0.00526164470750804, 694444.452031521, 277777.778948537, 5902777.78863394, 0.00411750124367508, 0.000172313017160917, 7013888.90326445, 0.0054543636707328, 694444.459051529, 1458333.34674381, 0, 0, 0, 0, 0, 0, -1.7208199790972e-05, 0, -0.000498372092711897, 0, 0, 0, 0.00109512014262366, 0, -0.00258344726147797, -0.000350574663502657, -0.00111134288889476, -277777.778376102, 0, 0, 0, 0.00438256164272508, 0, -0.00451486733998621, -0.00104125228281898, -8.94645854790411e-05, -0.0070828564428873, -0.00853827480861766, -277777.778461565, -8611111.12603168, 0, 0, 0.00112153549200847, -0.00416832593577921, 0.00106912159250023, -0.00892061861074697, -4861111.13184185, 0.000416102582107302, -22916666.6951662, -0.00392618624661516, -0.0271911405551979, -4166666.69179925, -0.0192670163684028, -0.0153320308391631, -4861111.1507076, 0, 0, 0, 0, 0, 0, 2.58919631816716e-05, 0, 0.000249443399071982, 0, 0, 0, -0.000426585858385385, 0, 0.00151360743481521, 0.000260339900073529, 0.000435072528017901, 6.53129663745504e-05, 0, 0, 0, -0.00349924046246674, 0, 0.00438258464505085, 0.000721984651921606, 0.000317674035020403, 0.00468226303350076, 0.00496868228290876, -0.000335819808636977, 4166666.67674038, 0, 0, 0, -0.00409339096038004, 0, 0.00209119425592631, 0.00232707102076648, -0.00105977914130198, 0.00875296589365494, 0.0140441505767327, -0.00109936753988316, 31944444.4719997, 0.00143897954966579, 0.0240445410658277, 4166666.6891037, 0.00642242205801583, -0.00204029345686973, 13888888.9150867, 0.0106183707929842, 0.00703485575620769, 0.0263653689799802, 0.0452787246023197, 0.0192661560031577, 13888888.9372927, 0.0334773367195865, 0, 0, 0, 0, 0, 0, -7.40383240841493e-06, 0, -5.35938836378203e-05, 0, 0, 0, 4.34308809626127e-05, 0, -0.00034550031491234, -7.45583985302217e-05, -5.52468615298008e-05, 4.95371683958217e-05, 0, 0, 0, 0.000945698942541002, 0, -0.00143021710335481, -0.000238694808214612, -0.000180797498863445, -0.00119083749401711, -0.00115747607044576, 0.000363806423905481, -0.00259940408915925, 0, 0, 0, 0.00208806617730717, 0, -0.00170113179330958, -0.000620497708013165, 0.000233589708909107, -0.00314006715131699, -0.00373747094395162, 0.000707004238448273, -13888888.8989299, -0.000314532062608169, -0.00805350526113787, -0.0074561321161207, 0, 0, 0.000630373071209, -0.00179705514333682, 0.00102631901536021, -0.0033373027368815, -0.00803611182172776, 0.00113710897742494, -13888888.9016911, -0.00213004626346658, -0.0010086235215324, -0.012742692914591, -0.0188491544512264, -0.00871544532702956, -13888888.9109691, -0.00794948587292151, -0.00445221946780612, -0.0256784817757655, -0.00898028573939731, -0.0206790078838399, -0.0202255932662949}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.32684444685779e-06, 0, 9.12550992911388e-07, 833.33333221072, 0, 0, 0, 0, 0, 0, 1.42208218170536e-05, 0, -1.52841039480253e-05, 0, 0, 1.29056898490518e-05, -833.333326743371, -9166.66665187677, -19861.1110957576, 0, 0, 0, 0, 0, 0, -5.58305885496318e-05, 0, 9.19388643663652e-05, 0, 0, 0, -0.000190184585297975, 0, 1.76962029528267e-05, -6.31266677281498e-06, 8333.33322756222, 175277.777680074, 8333.3332522988, 19027.7776956454, 217638.888772476, 0, 0, 0, 0, 0, 0, 0.000107165465687224, 0, -0.000270743559467645, 0, 0, 0, 0.000751482907954188, 0, -0.000429012725675483, -4.39033934745828e-05, 0.000346382039137571, -746527.777434459, 0, 0, 0.000667987551406709, 0.000186645624575833, 8.21488076153681e-05, 0.000523116980978344, -190277.776992957, -156249.999425593, -1909027.77743403, -190277.777462132, 0, 0, 0, 0, 0, 0, -0.000109026173167828, 0, 0.000438143011355532, 0, 0, 0, -0.00134731338430572, 0, 0.00140603288842204, 0.00014305710051199, -0.000595916211137952, 1631944.44372588, 0, 0, 0, -0.00282511610657061, 0, -0.000434725568910587, -8.78850799709331e-05, 0.000277128063188049, -0.00148025295516702, -0.00213855010437422, 590277.775754586, 8055555.55531524, -0.00124491926708523, -0.000904537289037723, 1562499.99758423, -0.000141372993446181, -0.00182126697846159, 1562499.99933854, 0, 0, 0, 0, 0, 0, 5.77073483611808e-05, 0, -0.000399302653852893, 0, 0, 0, 0.00121313495526864, 0, -0.00199440798354843, -0.000183979110952775, 0.000579515732000025, -1736111.11021859, 0, 0, 0, 0.00444694187948424, 0, -0.00122852582273704, -0.000183386696626431, -0.000710476905252509, 0.00207189319309496, 0.00259241542716426, -1041666.66308167, -17361111.1115807, 0, 0, 0.00365287483205728, 0.00111097793353702, 0.000101783742560036, 0.00296213159273811, 0.00594019711900488, 0.00318204527211133, -5902777.77502296, -0.0011620776838912, 0.00380654300294726, -5902777.77679416, 0.00266874625510651, -0.00033042480330465, 0.00284949776227111, 0, 0, 0, 0, 0, 0, -1.40208704722238e-05, 0, 0.000192723295538703, 0, 0, 0, -0.000522706324706672, 0, 0.00133083294328191, 0.000114785793920307, -0.00030903231539352, 694444.443844361, 0, 0, 0, -0.00307349992475098, 0, 0.00196900739898501, 0.00022236290684952, 0.000583302036399675, -0.00145262837310273, -0.0014399575987413, 694444.441358342, 18055555.5562838, 0, 0, 0, -0.00471735720249996, 0, -0.00133418602743965, -0.000242779930151413, 0.000726215249055856, -0.00297224651731723, -0.00536985825837647, -0.00423834485615549, 10416666.666192, 0.00178757004639792, -0.00295491719026994, 10416666.665729, -0.00214576099732719, -0.00169540675102671, -0.00630587169695568, -0.000753056619778518, 0.00246531993048644, -0.00496934099291637, -0.0045942949861679, -0.000153581535788512, -0.00181176397307254, -0.00184358154614727, 0, 0, 0, 0, 0, 0, 1.11519689330746e-06, 0, -3.8389990262539e-05, 0, 0, 0, 8.19176244104753e-05, 0, -0.000342671159897067, -2.90651679671596e-05, 7.13473620136021e-05, 0.000166261416176146, 0, 0, 0, 0.000780065896026682, 0, -0.000798436278025106, -7.08145154338631e-05, -0.000167148438924089, 0.000411691978988217, 0.000301323607212758, 0.00102640737403029, -6944444.44470425, 0, 0, 0, 0.00183221317486219, 0, -0.000196484440831583, -6.13761065865304e-05, -0.000388332111884764, 0.00108729396571872, 0.00161978728548401, 0.00191899584519965, -6944444.44496657, -0.000701774398816973, 0.000772168853092273, -6944444.44399774, 0, 0, 0.00143061786030821, 0.000426740397749358, 0.000129898395678875, 0.00119233497668996, 0.00335863805295749, 0.00147071778490608, 0.00188948582234456, -0.000526863510872466, -0.0017148972789431, 0.00232501847273865, 0.00167762500636037, 0.00052857652307321, 6.10665316769632e-05, 0.00177150939246751, -0.000847424756734467, 0.00220999105013311, -0.000378016997264407, 0.00241638719240788, 0.000298381291208712}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -8.48331105426566e-06, 0, 0, 0, 0, 0, 0, 0, 0.000133608504399302, 0, 0.000150788628376828, 63000.0001730936, 0, 0, 0, 0, 0, 0, -0.000988798581008831, 0, -0.00110032891343159, 0, 0, -0.00193311926510682, -1036750.00185517, -1036750.00286946, -406750.001467745, 0, 0, 0, 0, 0, 0, 0.00377181122021623, 0, 0.0039919343219006, 0, 0, 0, 0.012334671331928, 0, 0.0117664888773947, 5109166.67474275, 5109166.68530587, 1041666.66938742, 16518333.3578629, 6150833.34894331, 6150833.35653786, 0, 0, 0, 0, 0, 0, -0.00845736527258136, 0, -0.00861219995907687, 0, 0, 0, -0.0370842316270471, 0, -0.0307297228633917, -11739583.3511674, -11739583.3905794, -1322916.66339256, 0, 0, -0.0698818751114753, -75893750.0838608, -24802083.4012772, -75893750.1212991, -24802083.4866075, -14385416.6996397, -14385416.7107304, -90279166.8493391, 0, 0, 0, 0, 0, 0, 0.0120777990919588, 0, 0.0118936500685889, 0, 0, 0, 0.0560842329370272, 0, 0.0399849934071047, 14062500.0205663, 14062500.0948298, 833333.320054323, 0, 0, 0, 0.191137337716586, 0, 0.158525179043561, 160416666.806881, 43020833.4841428, 160416666.950536, 43020833.7521739, 16562500.0095143, 16562499.9529763, 292708333.650749, 44687500.1388689, 44687500.3999744, 337395833.905552, 337395834.129776, 193541667.029147, 0, 0, 0, 0, 0, 0, -0.0111623907787645, 0, -0.0104021742869594, 0, 0, 0, -0.0421183543026703, 0, -0.0269463564718229, -8541666.67708937, -8541666.75533883, -208333.323378111, 0, 0, 0, -0.252737359418471, 0, -0.165274272142769, -175000000.117393, -34375000.1812064, -175000000.340493, -34375000.5460626, -9374999.97179938, -9374999.77893122, 0, 0, -0.357323535967135, -482291667.057792, -52083333.5946694, -482291667.217451, -52083334.3404505, -35416666.7418528, -35416667.0066452, -534375000.851198, -534375001.502824, -203125000.107746, -1000000001.63995, -553125001.047144, -553125001.520183, 0, 0, 0, 0, 0, 0, 0.00614996542214753, 0, 0.00522689407505316, 0, 0, 0, 0.0134098339926664, 0, 0.00881535134068853, 2083333.33324277, 2083333.37838622, 0.000835765637021524, 0, 0, 0, 0.156504851445805, 0, 0.077876243538895, 95833333.3758184, 10416666.7784207, 95833333.5361121, 10416667.0054148, 2083333.31706545, 2083333.11183205, 0, 0, 0, 0.344979492727732, 0, 0.271676708177176, 364583333.554553, 20833333.5631077, 364583333.773648, 20833334.3168989, 10416666.6455496, 10416666.6443371, 385416667.273113, 385416667.945055, 104166666.395254, 541666667.12622, 20833333.4553743, 20833334.2087252, 1104166668.13382, 583333334.528629, 1104166668.59139, 583333335.500763, 395833333.892364, 395833334.275043, 1125000002.24836, 0, 0, 0, 0, 0, 0, -0.0015160696771364, 0, -0.00114005266092268, 0, 0, 0, -0.000865656739312189, 0, -0.00109059317460344, 0.00139549810161165, -0.00992808418539348, -0.00221273208443427, 0, 0, 0, -0.0356868677203398, 0, -0.0126650263146667, -20833333.336395, -0.02752196943285, -20833333.3810456, -0.080177413539972, -0.00218456925531267, 0.0711547861165881, 0, 0, 0, -0.118121089379086, 0, -0.0715529161461469, -104166666.710389, -0.076881281149645, -104166666.797297, -0.332222518123814, 0.0148240190621255, 0.0964645926930474, -104166666.833497, -104166667.069483, -20833333.1622201, 0, 0, -0.140622221269393, -208333333.469408, -0.0616651284856484, -208333333.521341, -0.499734740606281, 0.00208942300506932, -0.216253030231661, -416666667.143541, -208333333.807115, -416666667.462559, -208333334.325193, -104166666.66156, -104166666.739873, -625000000.814758, -208333333.751461, -208333334.215128, -625000001.098015, -625000001.391436, -416666667.492047}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 3.96586128233692e-07, 0, 0, 0, 0, 0, 0, 0, -1.01938378634084e-07, 0, -1.70779736265963e-05, -4499.99999741318, 0, 0, 0, 0, 0, 0, -5.24642563524885e-05, 0, 0.000149826204958662, 0, 0, 0.000194949567282936, 103250.00001654, 54499.9999410155, 9500.00000565327, 0, 0, 0, 0, 0, 0, 0.000407055963081014, 0, -0.000642105392181163, 0, 0, 0, -0.000779568179648409, 0, -0.00156979682188813, -859166.666819174, -99999.9992472523, -4999.99983667414, -1240416.66677543, -207916.667239875, -105000.000432237, 0, 0, 0, 0, 0, 0, -0.00141462124996642, 0, 0.0016069365283735, 0, 0, 0, 0.00136660598841179, 0, 0.00553132376745378, 3364583.333546, 49999.9960409245, -0.00111046357306555, 0, 0, 0.00611278578545424, 10185416.668331, 1593750.00403375, 2183333.33100955, 100000.001617962, 104166.667111695, 50000.001262196, 2287500.00838694, 0, 0, 0, 0, 0, 0, 0.00270293054672203, 0, -0.00243964186075362, 0, 0, 0, -0.000611746987547315, 0, -0.0109041573985943, -6562499.99943409, 0.00992147995249302, 0.00231430007881607, 0, 0, 0, -0.0123340176262116, 0, -0.0168772031444451, -39062500.0044048, -5416666.67669947, -1041666.65246135, 0.00204339755201316, 0.00524159683113342, 0.00274458026922356, -16666666.6640556, -729166.675863874, -0.0124358676704822, -17395833.3719698, -2083333.36034959, -1041666.68727506, 0, 0, 0, 0, 0, 0, -0.00292389709634261, 0, 0.00220250438382221, 0, 0, 0, -0.00145155233324428, 0, 0.012342626997277, 6041666.66485082, -0.0129220901488667, -0.00205715263831861, 0, 0, 0, 0.0140274324706514, 0, 0.0240768038453441, 73750000.0040531, 8125000.01078847, -0.0284740622386239, -0.0133243006058269, -0.0120795728920057, -0.00919801215364491, 0, 0, 0.0251613709935647, 56250000.0013038, 2083333.35666992, 7291666.64174374, 0.022952110993191, 0.00068848877072923, 0.0121572781876516, 58333333.4041174, 0.0202567491326411, -0.000786906993609597, 14583333.4216592, 7291666.75167133, 0.0867074618248053, 0, 0, 0, 0, 0, 0, 0.00167564646761832, 0, -0.00108708936943615, 0, 0, 0, 0.00213498685375817, 0, -0.00741244296958043, -2083333.33160614, 0.00850266702401517, 0.000748719932633807, 0, 0, 0, -0.00814786339646637, 0, -0.0178950178003424, -64583333.3338358, -4166666.67119967, 0.0242485970734053, 0.0168412285663194, 0.00938305644095966, 0.00757324224034094, 0, 0, 0, -0.0197427261366525, 0, -0.0222895984195504, -83333333.3366822, -2083333.3547007, 0.0365191496585998, -0.0137042088450074, 0.00988526601565092, 0.000528576696972209, -85416666.722678, 0.0119481531865529, 0.0290139589529267, -20833333.3153812, -0.0131107367494494, -0.0257538126473094, -41666666.77346, -20833333.4476416, -0.0545504116451266, -0.0952403825694476, -0.0412024812969232, -0.060074581523056, -0.179483241967346, 0, 0, 0, 0, 0, 0, -0.000394952836084122, 0, 0.000226239664955998, 0, 0, 0, -0.00083776075147615, 0, 0.00181875809380049, -0.000557136165902575, -0.0022406383648688, -6.53546906199848e-05, 0, 0, 0, 0.00175641589827098, 0, 0.00540488635263006, 20833333.3326918, 0.000327212313010191, -0.00756688421958465, -0.00675734065594094, -0.00244151868931681, -0.00190802463566222, 0, 0, 0, 0.00639599358812436, 0, 0.00779253283612907, 41666666.6676258, 0.00638933143573161, -0.0157503416096954, 0.00128075270207467, -0.00518232749172172, -0.00249406673300017, 41666666.6822085, -0.0133463766795494, -0.0158083560073425, 0, 0, 0.0081769197173037, 20833333.3292437, 0.00851817413416979, -0.0152109392889488, 0.0124799475224047, -6.46329770258033e-05, 0.00563114513030644, 41666666.7102676, 20833333.3824983, 3.38719264446929e-05, 0.0280111496092613, -0.00647914135260271, 0.00328104054031045, 0.0358094730411324, 0.0390174777114955, 0.0490039303080495, 0.106130454866455, 0.0901760756731907, 0.0735665452554837}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 1.497240787197e-06, 0, 0, 0, 0, 0, 0, 0, -1.35605343004661e-05, 0, -2.73600201571197e-05, -4500.00002161117, 0, 0, 0, 0, 0, 0, 0.000100986958868932, 0, 0.000257646581619362, 0, 0, 8.43842941717798e-05, 54500.0000446248, 103250.000340028, 9500.00027018373, 0, 0, 0, 0, 0, 0, -0.000445358694098791, 0, -0.00130925052190966, 0, 0, 0, -0.000242789419134397, 0, -0.000522835571025153, -99999.9998600829, -859166.66931606, -5000.00217091869, -1240416.66782159, -105000.000208353, -207916.668631474, 0, 0, 0, 0, 0, 0, 0.0011236018026839, 0, 0.00388182233993037, 0, 0, 0, 0.00073651983309427, 0, 0.00218354842256817, 49999.9990611238, 3364583.34341067, 0.00930412906752761, 0, 0, 0.000322393011554073, 2183333.3345566, 99999.9984456475, 10185416.6770868, 1593750.01136541, 50000.0030493159, 104166.676706438, 2287499.99883463, 0, 0, 0, 0, 0, 0, -0.00172406853048605, 0, -0.00687380522714393, 0, 0, 0, -0.000797802606031603, 0, -0.00591102300480121, 0.00245306672915283, -6562500.02090994, -0.0219407147883436, 0, 0, 0, -0.00136932342459979, 0, 0.00104763547234358, -1041666.66543964, 0.0030107759190624, -39062500.0361653, -5416666.69901045, -0.0157462907550079, -0.0305230574435098, -16666666.6859796, 0.003468090038824, -729166.705240571, -2083333.31065002, -17395833.3449093, -1041666.66482513, 0, 0, 0, 0, 0, 0, 0.00162508507530753, 0, 0.00710891539891246, 0, 0, 0, -0.000251544565729586, 0, 0.0094861209337658, -0.00350805726236617, 6041666.69114733, 0.0287602561577096, 0, 0, 0, 0.00192699710999756, 0, -0.00335163867360446, -0.00502887699410651, -0.00147759048590563, 73750000.0597281, 8125000.04592866, 0.0338489413402921, 0.049749985717666, 0, 0, -0.000707877964315208, 7291666.68550072, -0.00710847733658802, 56250000.0591791, 2083333.40945713, 0.00348317069024191, 0.0744985425736215, -0.0412824072471777, 58333333.3827988, 0.0214721999854889, 14583333.3031743, -0.0664668664343024, 7291666.68536827, 0, 0, 0, 0, 0, 0, -0.000868965119108723, 0, -0.00394513769533126, 0, 0, 0, 0.000825530431124066, 0, -0.00781079945625653, 0.00261186757018447, -2083333.34855728, -0.0195692271566483, 0, 0, 0, 0.000165254460020251, 0, 0.00214387654549557, 0.00544956042156781, -0.000214042895493257, -64583333.3810053, -4166666.69926471, -0.0328707972328595, -0.0418483959342335, 0, 0, 0, -0.00233888816281052, 0, 0.00546264635403498, -0.00805735671616241, 0.00106956157736941, -83333333.402207, -2083333.39887933, -0.0145650972831055, -0.0666323888140031, 0.0299251469848973, -85416666.7297686, -0.0419487093404494, -20833333.3799554, 0.00519035484468843, -0.0949834650689309, 0.0486630687866528, 0.0835245433304993, -41666666.6773717, -20833333.3855367, 0.0516354124733501, -0.0543876125947246, 0.0916707245907603, 0, 0, 0, 0, 0, 0, 0.000200801457532823, 0, 0.000905657453360135, 0, 0, 0, -0.000326476905381786, 0, 0.00250542212539765, -0.000781919572725878, 0.0039074631712776, 0.00536873770188167, 0, 0, 0, -0.000768905763148631, 0, 1.57419924341953e-06, -0.00206924929336607, 0.000164975878251149, 20833333.3481594, 0.00937733451613814, 0.0119097327940816, 0.0142498910303978, 0, 0, 0, 0.00100910477392725, 0, -0.00338906488612583, 0.000651354931168093, 0.00157191006412322, 41666666.6945374, 0.0202315772830264, 0.00948657771659862, 0.0228784511248151, -0.00820721485654513, 41666666.6936903, 0.0220783312002128, 0, 0, -0.000540185909389782, 0.0136861274006985, -0.00110420982809586, 20833333.3618241, 0.0410943609548972, 0.000590628184027874, 0.0401147516865987, -0.0192773476945102, -0.0294595713603929, 41666666.6890995, 20833333.363514, -0.00820863536430408, 0.0367144612585267, -0.00554913574162764, -0.0403017897719847, 0.0458363893198771, -0.0675705106075544, -0.0295958376174822, -0.0345592229482674}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 8.17109095305055e-07, 0, 0, 0, 0, 0, 0, 0, -1.28614737191541e-05, 0, -4.58461634526643e-05, -35000.0000367173, 0, 0, 0, 0, 0, 0, -7.13865658062692e-05, 0, 0.000269992401359909, 0, 0, 0.00108046338273543, 692638.889695677, 517638.88969236, 167638.888942125, 0, 0, 0, 0, 0, 0, 0.00118354886922066, 0, -0.000301380055365311, 0, 0, 0, -0.00662348979238824, 0, -0.0075582992293139, -4563888.89287307, -1975694.44855688, -299305.550580267, -10039583.3439562, -3113194.45206783, -2275000.00905768, 0, 0, 0, 0, 0, 0, -0.00506501409185391, 0, -0.00127904421718634, 0, 0, 0, 0.017079372244675, 0, 0.0193029763789974, 13107638.8973936, 3229166.67270381, 236111.081839566, 0, 0, 0.0467707222070292, 63708333.3734289, 18069444.487198, 36329861.1524869, 6694444.51236647, 5197916.66946575, 3701388.88781456, 41527777.8906895, 0, 0, 0, 0, 0, 0, 0.0104652324168404, 0, 0.00457999559705572, 0, 0, 0, -0.0187033591277956, 0, -0.0227578082237521, -18576388.8965223, -2430555.550594, -69444.3760342385, 0, 0, 0, -0.124552476735408, 0, -0.10914054770403, -171909722.290214, -40833333.4317132, -55798611.1720935, -7361111.28715422, -3819444.37944542, -2638888.7665749, -206701389.003211, -26006944.5213809, -7500000.1657103, -232708333.74399, -115416667.141024, -63437500.2046545, 0, 0, 0, 0, 0, 0, -0.0114873791421024, 0, -0.0062168718786678, 0, 0, 0, 0.00281925309744153, 0, 0.0112091653534315, 12847222.2222809, 694444.422510129, -0.0802146055205368, 0, 0, 0, 0.162646620852052, 0, 0.11637171600192, 225347222.273884, 39583333.4452671, 39236111.134013, 2777777.97769391, 1041666.52040362, 694444.158717067, 0, 0, 0.238324826302671, 450000000.138869, 41666666.8417492, 276041666.780296, 4166667.14838197, 15972222.2081318, 2777777.79543572, 491666667.317522, 118750000.805159, 42361111.0281797, 568055556.670649, 307986111.898846, 120833334.296578, 0, 0, 0, 0, 0, 0, 0.00642556673156126, 0, 0.00401719607408051, 0, 0, 0, 0.0091583040922175, 0, 0.000113565770846671, -3472222.21806718, 0.0207567000983069, 0.0475642264399843, 0, 0, 0, -0.0999943844521332, 0, -0.0578398105294253, -142361111.121697, -13888888.9503996, -10416666.6459031, -0.0946314932530615, 0.119712661519766, 0.25019946649254, 0, 0, 0, -0.228623028125333, 0, -0.180882043513205, -416666666.737814, -20833333.4864413, -163194444.46175, -0.483324221619495, -3472222.11743438, 0.217824708861083, -437500000.474163, -41666667.2574959, -10416666.2783945, -430555555.638106, -13888888.9541694, -0.341829071181682, -875000001.080558, -458333334.305302, -493055556.70331, -62500001.3449665, -173611111.469083, -41666667.1341967, -500000001.726865, 0, 0, 0, 0, 0, 0, -0.00143860843247297, 0, -0.00102492011504137, 0, 0, 0, -0.00475778944314216, 0, -0.00138367982925289, -0.00187243387175909, -0.00646959774754375, -0.0114844886922901, 0, 0, 0, 0.0223317331727351, 0, 0.0105139979631034, 34722222.2188187, 0.0127656643899252, -0.0142000903859635, 0.0118101190975441, -0.0346081131105954, -0.0776457426571325, 0, 0, 0, 0.0794782045423327, 0, 0.048425126834469, 138888888.897832, 0.0470785526850033, 34722222.2037587, 0.160659804685917, -0.0519975192637974, -0.133801127772425, 138888889.01766, 0.152561663007627, -0.20225368367938, 0, 0, 0.0913796095844904, 208333333.352661, 0.041721634887036, 138888888.888325, 0.235013829648743, -0.0188339663588454, 0.0259142409914793, 416666667.039758, 208333333.726493, 138888889.285207, 0.584449257533968, 34722222.158121, -0.0741364352425657, 416666667.186542, 138888889.21915, 0.508685601601316, 416666667.58545, 208333334.351454, 138888889.508078}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -1.29436634060276e-07, 0, 0, 0, 0, 0, 0, 0, -4.60895995512719e-06, 0, 2.70359152889367e-05, 14999.9999960006, 0, 0, 0, 0, 0, 0, 0.000163505741085154, 0, -0.00018964534344255, 0, 0, -0.000543571564588259, -325416.66683709, -200416.666664904, -50416.6665698925, 0, 0, 0, 0, 0, 0, -0.00126848496666065, 0, 0.000483739971997951, 0, 0, 0, 0.00296845460807549, 0, 0.00401387327023572, 2480555.55655693, 560416.665777878, 56249.9975105046, 4290972.22427458, 1036805.55805774, 616666.66898453, 0, 0, 0, 0, 0, 0, 0.00453482467899401, 0, -0.000446002428099488, 0, 0, 0, -0.00678439597111315, 0, -0.0113484207004988, -8593750.00178698, -583333.325902829, -20833.3200502517, 0, 0, -0.0220809331488236, -31972222.2319517, -7166666.68351868, -11461805.5574963, -1187500.01746268, -1093749.99927609, -624999.9974968, -12555555.5958332, 0, 0, 0, 0, 0, 0, -0.00879368415801519, 0, -0.000301429627354879, 0, 0, 0, 0.00570432803770877, 0, 0.0166722907158123, 14409722.2219723, 208333.310717178, -0.0303461759305934, 0, 0, 0, 0.0550326379878119, 0, 0.0533733060164876, 106736111.12882, 20798611.1511375, 11319444.4233156, 625000.036348138, 381944.411379214, 208333.282788707, 78437500.0096586, 6770833.36698078, 625000.052285064, 85208333.4984913, 23020833.4882095, 12083333.4123007, 0, 0, 0, 0, 0, 0, 0.00952429055126112, 0, 0.00103654826446316, 0, 0, 0, 0.0025710068788844, 0, -0.0134523061981399, -11458333.3291051, 0.0328721642023396, 0.0353299849564044, 0, 0, 0, -0.069668469786402, 0, -0.0619631322920623, -169097222.233523, -25000000.0430802, -3819444.38448993, -0.0226032709731159, 0.0719000323385449, 0.113120272487397, 0, 0, -0.105640080445946, -223611111.125625, -15625000.0835075, -69791666.621852, -0.149074649324901, -2083333.32694308, 0.00380612401830649, -239236111.390355, -11458333.5498406, -3819444.40263101, -141666667.0793, -73958333.6778162, -11458333.7015441, 0, 0, 0, 0, 0, 0, -0.00541274461217573, 0, -0.000853652812091672, 0, 0, 0, -0.00694780355840355, 0, 0.00546374709811713, 3472222.21770004, -0.0231800771204617, -0.0209435978083484, 0, 0, 0, 0.0422732141424458, 0, 0.0359940540280455, 124999999.997691, 10416666.6866672, -0.0586391235433958, -0.00847066810846263, -0.0593096933737335, -0.0968722194791742, 0, 0, 0, 0.0969036446284398, 0, 0.0824412133240379, 260416666.673223, 10416666.7398647, 20833333.2460381, 0.141205628973432, -0.0501743968660367, -0.0886419369916804, 270833333.542397, 0.11028533189537, -0.170092870628122, 159722222.180269, 3472222.26221289, 0.106137404716631, 322916667.10553, 166666667.114746, 62500000.3505589, 0.476416656856724, 20833333.4944409, 0.184433046507664, 62500000.7330775, 0, 0, 0, 0, 0, 0, 0.00125706788316704, 0, 0.000243575182506633, 0, 0, 0, 0.00300581504485385, 0, -0.000800947811990494, 0.00150448924640156, 0.00638599025182638, 0.00507219782122345, 0, 0, 0, -0.00933927748311675, 0, -0.00842549754024616, -34722222.2186369, -0.00271618070791887, 0.0198736966320847, 0.00981022841208334, 0.0174957615578852, 0.0295766470812544, 0, 0, 0, -0.033251853447177, 0, -0.0237271055361831, -104166666.665358, -0.0212851998527755, 0.0417358076766115, -0.0428249337546059, 0.025885474466719, 0.0510207416796736, -104166666.722728, -0.00959627910142097, 0.0875213763876378, 0, 0, -0.0385099112138367, -104166666.652841, -0.0255579263527204, -34722222.180539, -0.0746050548666554, 0.00488968406502332, -0.00148019201137082, -208333333.496404, -104166666.852352, -0.0836680326296099, -0.186854702367036, 0.0290548240927042, 0.0324732599655193, -104166666.843716, -34722222.3791014, -0.191390022424926, -104166667.082399, -0.400662082284047, -0.276271137372127}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -2.84968283463298e-07, 0, 0, 0, 0, 0, 0, 0, 5.06495163121026e-06, 0, -4.0242167565134e-06, -3749.99999770901, 0, 0, 0, 0, 0, 0, -7.31133787786397e-05, 0, 2.58298351512444e-06, 0, 0, 0.000175723527489865, 76666.6666516383, 64166.6667318714, 7916.66661352782, 0, 0, 0, 0, 0, 0, 0.000524324706444815, 0, 0.000180420173139487, 0, 0, 0, -0.00134519192303592, 0, -0.00101944648413966, -524305.555395432, -310416.6669908, -4166.66548503066, -1303472.22259893, -153472.222107288, -127083.335235358, 0, 0, 0, 0, 0, 0, -0.00193924093497002, 0, -0.000822672093932451, 0, 0, 0, 0.00417751674991464, 0, 0.0014202460282012, 1493055.55479715, 458333.33275363, -0.00614208900834592, 0, 0, 0.00833816897410906, 8809027.77682473, 944444.445111114, 6211805.55715119, 520833.348936361, 76388.8830631212, 62500.0024032634, 2454861.12359646, 0, 0, 0, 0, 0, 0, 0.00391406528378113, 0, 0.00165901498442594, 0, 0, 0, -0.00549527968234485, 0, 0.000652821724595014, -1736111.10925135, -208333.32803132, 0.0138811807774862, 0, 0, 0, -0.0269961020561699, 0, -0.0143688717706919, -24548611.1039495, -2152777.78071171, -8819444.44315812, -416666.708183392, 0.0261015756064168, 0.015961885476506, -40798611.1052489, -416666.656459683, -208333.382142563, -15000000.034447, -9965277.84865116, -1145833.35063826, 0, 0, 0, 0, 0, 0, -0.00435825206640967, 0, -0.00180698473399948, 0, 0, 0, 0.00182470646061576, 0, -0.00318201657836131, 694444.441967284, -0.0105203229116907, -0.0160577821632765, 0, 0, 0, 0.0401886170758572, 0, 0.00826905096840403, 27430555.5412128, 1388888.89321384, 3819444.43312792, 0.0467663261490679, -0.0445987942554429, -0.0438505354129258, 0, 0, 0.0423000621050832, 107638888.86428, 694444.437477012, 53472222.2020779, 0.132234888179371, -0.0423864787549656, 0.0327976353254997, 33680555.6021062, 7638889.02264882, -0.042446738881472, 59722222.3438447, 6250000.05745125, 3819444.60851662, 0, 0, 0, 0, 0, 0, 0.00251450299645296, 0, 0.00104203908745866, 0, 0, 0, 0.00186467853859802, 0, 0.00274952812733001, 0.00169080295475125, 0.00867265530178324, 0.00948915857031963, 0, 0, 0, -0.0270674705291138, 0, 0.000534853165937622, -10416666.6540431, -0.00276859619050158, 0.0147166612031781, -0.0204146896075649, 0.0337904238812739, 0.0393818914180725, 0, 0, 0, -0.0485428372454662, 0, -0.023610803216807, -107638888.859635, 0.00096896979101351, -20833333.3059616, -0.131189779960254, 0.0512015636236569, 0.0196250479040848, -20833333.3633386, -0.101695034755332, 0.0901171688444819, -118055555.518614, 0.0301218465798175, -0.110329157207074, -128472222.311627, -10416666.7386788, -41666666.8047657, -0.263465349620167, 0.0187074175763002, -0.0841840784481598, -20833333.5518068, 0, 0, 0, 0, 0, 0, -0.000587069164460545, 0, -0.000250111800354284, 0, 0, 0, -0.00119751526705823, 0, -0.000803648168483801, -0.000462171319164409, -0.0026176417259912, -0.00230263725000177, 0, 0, 0, 0.00652160958331202, 0, -0.00135736176419767, -0.00418948547861948, 0.000647599670473915, -0.00588853011806226, 0.00148934664246226, -0.00960957062558402, -0.0121034022799297, 0, 0, 0, 0.0188155205942376, 0, 0.00353681607805233, 34722222.2103954, 0.000491084187704314, -0.0124362666063876, 0.0434186655319977, -0.0197787090022026, -0.0188009446869634, 0.00708409313790138, 0.0255041849772581, -0.0422412761810719, 0, 0, 0.0160241195776143, 69444444.4255347, -0.00812581583933203, 34722222.2028331, 0.0681356461973277, -0.019764990766784, 0.020082015632413, 69444444.46783, 0.0300598112417571, 0.0491343210128948, 0.122022561258296, -0.0426325987607218, -0.0124117975312408, 69444444.4879047, 0.00853465347940207, 0.110042530051285, 34722222.3208204, 0.152819820290736, 0.0724592028015285}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 2.2249246674535e-06, 0, 0, 0, 0, 0, 0, 0, -2.89136456137801e-05, 0, -4.54506704732091e-05, -3750.00002941783, 0, 0, 0, 0, 0, 0, 0.000187685457133193, 0, 0.000359532177737745, 0, 0, 0.000485035138903915, 64166.6667639276, 76666.6672211241, 7916.66699558531, 0, 0, 0, 0, 0, 0, -0.000655521119742594, 0, -0.00153403405664353, 0, 0, 0, -0.00279164235126427, 0, -0.00288741850014778, -310416.666735254, -524305.559502154, -4166.66850237116, -1303472.22465804, -127083.333537748, -153472.227233087, 0, 0, 0, 0, 0, 0, 0.00122433963709223, 0, 0.00398799205771873, 0, 0, 0, 0.00907597545155029, 0, 0.00869146851255177, 458333.333161352, 1493055.56897258, 0.00663626974349114, 0, 0, 0.013230809189015, 6211805.55945681, 520833.331127106, 8809027.79432584, 944444.475874136, 62499.9984001955, 76388.9074708282, 2454861.12520731, 0, 0, 0, 0, 0, 0, -0.00121850597429712, 0, -0.00639760582105291, 0, 0, 0, -0.0156097837651879, 0, -0.0162553697224294, -208333.332897344, -1736111.13564769, -0.0154013647116755, 0, 0, 0, -0.0374821250308273, 0, -0.0263343560732245, -8819444.44590588, -416666.662352971, -24548611.1596605, -2152777.86647935, 0.00389081146187193, -0.0363982946960252, -40798611.1368649, -208333.314219279, -416666.763395223, -9965277.7900791, -15000000.0924521, -1145833.35980243, 0, 0, 0, 0, 0, 0, 0.000605174337444639, 0, 0.00609356529138028, 0, 0, 0, 0.0138038620047944, 0, 0.0193421876917643, -0.000597429773517976, 694444.469532652, 0.0209225931210504, 0, 0, 0, 0.0550681169231649, 0, 0.0270986236354478, 3819444.44183195, -0.00107628242573778, 27430555.6258945, 1388889.01109374, 0.00136309657760224, 0.0447010379952807, 0, 0, 0.0567228707786093, 53472222.2360398, -0.0326018229939025, 107638888.953221, 694444.663976818, -0.0411908910765824, 0.132082060742983, 7638888.88952642, 33680555.7701166, 0.0168272352814779, 59722222.3196855, 3819444.43110743, 6250000.1875007, 0, 0, 0, 0, 0, 0, -0.000118337164359718, 0, -0.00312993348138355, 0, 0, 0, -0.00570347919922371, 0, -0.0129871107510325, 0.000549865792637831, -0.0136687367765295, -0.0147632738668582, 0, 0, 0, -0.0381042327268324, 0, -0.01596578431034, 0.00317048141377847, -0.00261735499612204, -10416666.7165942, -0.0808937039909766, -0.00843074774787911, -0.0345780360942275, 0, 0, 0, -0.0607332930003722, 0, -0.035067714718316, -20833333.3308425, 0.0204037986821007, -107638888.956403, -0.213515679807229, 0.0323997836233471, -0.0829241980980676, 0.00053003948281623, -20833333.5428466, -0.00170192315769873, -118055555.587458, 0.0516819651938296, -0.212323371172192, -41666666.7011244, 0.0384411024104901, -128472222.397031, -10416666.9860242, 0.0534090275089552, -0.154587731266006, -20833333.4414991, 0, 0, 0, 0, 0, 0, 1.86872329541478e-06, 0, 0.000663678977327245, 0, 0, 0, 0.00078775904290887, 0, 0.0036368970379018, -0.000217863306644934, 0.0031237589888535, 0.00413886096106626, 0, 0, 0, 0.00979183015811018, 0, 0.00461716486933195, -0.00111475960237826, 0.00147812871797597, 0.0139630206456931, 0.0206559361994245, 0.00488217382830507, 0.0121878535865875, 0, 0, 0, 0.0231331785356842, 0, 0.00762858575648052, -0.00335114285985182, -0.00370602783920503, 34722222.2474231, 0.0738219780836721, -0.00762049777560131, 0.0212003449767536, 0.00167464905806918, 0.0735217589617003, 0.00102605491203783, 0, 0, 0.0201118767113458, 34722222.2252623, -0.0192980069859186, 69444444.4642902, 0.110138713474236, -0.0228501911032663, 0.0676899397404104, 0.000821459014550608, -0.0137678267965822, 69444444.5381842, 0.161921957268648, -0.032030587668869, 0.0418842923854401, 69444444.4858356, -0.0398852289905744, 0.150090675785294, 0.0093403964221741, 34722222.3396687, 0.0339337784393969}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -6.21687237384733e-06, 0, 0, 0, 0, 0, 0, 0, 7.66932263976956e-05, 0, 0.000109374694378199, 15000.0000852424, 0, 0, 0, 0, 0, 0, -0.000581060953446313, 0, -0.000909988259749405, 0, 0, -0.000796331195949343, -200416.666982714, -325416.668103369, -50416.6676124436, 0, 0, 0, 0, 0, 0, 0.0024172762704122, 0, 0.00421589937648488, 0, 0, 0, 0.00460536845443573, 0, 0.00419162736317477, 560416.667089846, 2480555.56634871, 56250.0063382856, 4290972.22871396, 616666.667840537, 1036805.56528382, 0, 0, 0, 0, 0, 0, -0.00569553632195104, 0, -0.0118252357705379, 0, 0, 0, -0.0158573829452161, 0, -0.0123054755923945, -583333.332976871, -8593750.03909537, -20833.3588058033, 0, 0, -0.017591544501423, -11461805.5689018, -1187499.99699423, -31972222.2699036, -7166666.72883128, -625000.005238654, -1093750.03716372, -12555555.5717787, 0, 0, 0, 0, 0, 0, 0.00805091800300255, 0, 0.0202726419959546, 0, 0, 0, 0.027892058479762, 0, 0.0243730909651142, 208333.330677826, 14409722.2988367, 0.0600049870572387, 0, 0, 0, 0.0546878055062591, 0, 0.0273453553618978, 11319444.4575058, 624999.995086698, 106736111.260917, 20798611.2924306, 208333.361938826, 381944.528589023, 78437500.0961751, 624999.976130406, 6770833.51873858, 23020833.3167858, 85208333.478022, 12083333.3481657, 0, 0, 0, 0, 0, 0, -0.00692929543386211, 0, -0.0205364149515558, 0, 0, 0, -0.0247783901266687, 0, -0.0325455045686857, 0.00509301301784746, -11458333.4176746, -0.0802080316363672, 0, 0, 0, -0.0827162095511243, 0, -0.0171827537217127, -3819444.44736688, -0.00306415248136401, -169097222.452446, -25000000.2572407, -0.0723945552774913, -0.11640230219401, 0, 0, -0.0719505951507098, -69791666.7635881, 0.0422065567380643, -223611111.36129, -15625000.421595, 0.0342663293221972, -2083333.60859722, -11458333.2799, -239236111.500384, -3819444.46617784, -141666666.768718, -11458333.1819585, -73958333.5633216, 0, 0, 0, 0, 0, 0, 0.00340574006590002, 0, 0.0112212264922776, 0, 0, 0, 0.0103539813953354, 0, 0.0246028959016996, -0.00450677318592867, 3472222.27165045, 0.0557737956427704, 0, 0, 0, 0.0555290146204079, 0, 0.00512512460677581, -0.00465141026426809, 0.00787006095875962, 125000000.172335, 10416666.8435325, 0.0792769090192818, 0.0939943044648468, 0, 0, 0, 0.0870463191353197, 0, 0.0304199548127674, 20833333.3778081, -0.0176197058711647, 260416666.936384, 10416667.0714181, -0.00255440884321804, 0.190041677856264, -0.0375024347026468, 270833333.750195, 0.0405540147933455, 159722222.411641, -0.0626246710807295, 3472222.65762252, 62500000.0019554, -0.198231294851155, 322916666.969253, 166666667.136381, -0.193064060966507, 20833333.5548745, 62499999.9342252, 0, 0, 0, 0, 0, 0, -0.000738579576467951, 0, -0.00254123844250193, 0, 0, 0, -0.00153371721186217, 0, -0.0075974302372534, 0.00152560970626125, -0.0120528956671177, -0.015575790115406, 0, 0, 0, -0.0131926903078481, 0, -0.00180775797646651, 0.00282014355477417, -0.00312398694866025, -34722222.272724, -0.0476501968483284, -0.0311757924367414, -0.0330068050383244, 0, 0, 0, -0.0339587151784063, 0, -0.000419083862514451, -0.00630644180475099, -0.00135951758966649, -104166666.769862, -0.13756365836336, -0.0128842016648456, -0.0516110384717966, 0.00751625374934356, -104166666.824501, -0.027784223751492, 0, 0, -0.0252286034649502, -34722222.2739699, 0.0213608506637559, -104166666.774792, -0.218979321337228, 0.0227278404718796, -0.15144516218779, 0.0150078881870921, 0.0665728489285749, -208333333.52895, -104166666.923115, 0.0708546300977047, -0.0840634175779635, -104166666.747791, 0.130025124494518, -34722222.4634011, 0.114827022593739, -104166666.748474, 0.0608448362340626}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 1.04272641996639e-05, 0, 0, 0, 0, 0, 0, 0, -0.000141177962678626, 0, -0.000184171519927139, -35000.0001584701, 0, 0, 0, 0, 0, 0, 0.0010605338613432, 0, 0.00143560451201033, 0, 0, 0.00171867999591928, 517638.88988258, 692638.891611815, 167638.890443601, 0, 0, 0, 0, 0, 0, -0.00423085489753551, 0, -0.00615757749738969, 0, 0, 0, -0.0107644280459879, 0, -0.00910240565490458, -1975694.44771451, -4563888.90831625, -299305.563202665, -10039583.3499276, -2275000.00548247, -3113194.46463456, 0, 0, 0, 0, 0, 0, 0.00951607733156669, 0, 0.0163080896572145, 0, 0, 0, 0.0360693291811056, 0, 0.0236821394843001, 3229166.67285182, 13107638.9546719, 236111.137007594, 0, 0, 0.0473520277873081, 36329861.1570816, 6694444.45860591, 63708333.4357925, 18069444.5790296, 3701388.89741365, 5197916.72451935, 41527777.8591245, 0, 0, 0, 0, 0, 0, -0.0128636634877041, 0, -0.0270009358444807, 0, 0, 0, -0.0614109903241048, 0, -0.0382192708000293, -2430555.56155021, -18576389.0083703, -69444.5048460239, 0, 0, 0, -0.143748186991773, 0, -0.0839584860581133, -55798611.1765378, -7361111.14605685, -171909722.509118, -40833333.7194947, -2638888.90171097, -3819444.51504386, -206701389.122854, -7499999.98788723, -26006944.8167345, -115416666.834095, -232708333.805143, -63437500.1143432, 0, 0, 0, 0, 0, 0, 0.0107021442753824, 0, 0.0268233524106524, 0, 0, 0, 0.0527068935914249, 0, 0.0426503022988828, 694444.445538874, 12847222.3435387, 0.085218562295071, 0, 0, 0, 0.210867472817895, 0, 0.064415809200571, 39236111.1577382, 2777777.83401072, 225347222.623713, 39583333.8632509, 694444.487974913, 1041666.70377804, 0, 0, 0.214618672496154, 276041666.923131, 4166666.65484905, 450000000.515313, 41666667.5727768, 2777777.70207925, 15972222.6573105, 118750000.209753, 491666667.727883, 42361111.0967247, 568055556.272882, 120833333.433267, 307986111.908222, 0, 0, 0, 0, 0, 0, -0.00519716619305792, 0, -0.0144943290062324, 0, 0, 0, -0.0207467762717592, 0, -0.0293910276914814, 0.00264299410735002, -3472222.28791497, -0.0629083950074387, 0, 0, 0, -0.141054916308124, 0, -0.0207626062301196, -10416666.6793701, -0.045041012963907, -142361111.386744, -13888889.2356715, -0.0665496146270261, -0.0160862692032732, 0, 0, 0, -0.239198326941725, 0, -0.12147344510118, -163194444.572005, -0.0187609531413496, -416666667.166555, -20833334.2242991, 0.0616505965406849, -3472222.40936465, -41666666.8189607, -437500001.020709, -10416666.5415637, -430555555.956713, 0.0989525245511473, -13888889.7348082, -493055556.070364, -62500000.0862839, -875000001.115137, -458333334.70932, -41666666.5037691, -173611111.636875, -500000000.65802, 0, 0, 0, 0, 0, 0, 0.00114374380210524, 0, 0.00325949321224748, 0, 0, 0, 0.00262687839206315, 0, 0.00880021372999631, -0.00150245089936541, 0.0149525887516933, 0.0184430280211407, 0, 0, 0, 0.0342035363245904, 0, 0.00281111707903503, -0.000627833307165537, 0.0135124518238088, 34722222.2965429, 0.087173145672421, 0.0318646999169543, 0.00938920305073442, 0, 0, 0, 0.0901628787614259, 0, 0.0192354613093085, 34722222.2432681, 0.017004830385652, 138888889.064691, 0.305315057349752, -0.00613946655833315, 0.0127400334740051, 0.047908259630237, 138888889.243478, -0.0550332720244753, 0, 0, 0.0794075396213105, 138888888.998464, -0.0307693087471434, 208333333.537974, 0.459058799888424, -0.0549474011428341, 0.249631977332554, 138888889.032065, 0.0508976752636323, 416666667.229379, 208333334.03205, -0.145418337566638, 34722222.3034887, 416666667.072632, -0.0768393440864092, 138888889.480711, 208333333.542529, 416666667.268831, 138888889.044828}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -2.18590991316522e-06, 0, 0, 0, 0, 0, 0, 0, 2.78677586466409e-05, 0, 5.60572222299975e-05, 13333.3333721258, 0, 0, 0, 0, 0, 0, -0.000131324430536147, 0, -0.000395938850321374, 0, 0, -0.000792725628106807, -244814.815090444, -244814.815681796, -44814.8150991975, 0, 0, 0, 0, 0, 0, 0.000115469087545368, 0, 0.00136860557290574, 0, 0, 0, 0.00502466940702118, 0, 0.00470820889750841, 1388888.88972593, 1388888.89472672, 49999.9996160445, 4444444.4501549, 772222.223383399, 772222.230913712, 0, 0, 0, 0, 0, 0, 0.000991970010302208, 0, -0.00292117301144031, 0, 0, 0, -0.0158895281858975, 0, -0.0116223783596034, -3009259.26051849, -3009259.27613008, -18518.5134260827, 0, 0, -0.0271546340996977, -24629629.6430753, -3796296.30061023, -24629629.6621784, -3796296.35861933, -805555.545858209, -805555.574059893, -13194444.491545, 0, 0, 0, 0, 0, 0, -0.00343795052785047, 0, 0.00401381905771466, 0, 0, 0, 0.0244850222154008, 0, 0.0163490569042993, 2777777.77833815, 2777777.80135462, -0.0107164978342021, 0, 0, 0, 0.0817699971512737, 0, 0.0511472468851274, 50972222.2341238, 5833333.3459308, 50972222.301343, 5833333.50791434, 277777.730015363, 277777.767449335, 129722222.28231, 3333333.30514173, 3333333.50626586, 63611111.2277448, 63611111.3759784, 12916666.7317464, 0, 0, 0, 0, 0, 0, 0.00463129647522253, 0, -0.0033355507250241, 0, 0, 0, -0.0166955090946987, 0, -0.0153491947729284, -925925.9248326, -925925.941402236, 0.00863377034413245, 0, 0, 0, -0.120428194756864, 0, -0.0424868096864804, -44444444.4422709, -2777777.79891678, -44444444.5368133, -2777778.00362681, 0.074852916007492, 0.0634183654052757, 0, 0, -0.1277323025247, -240740740.778165, -2777777.74650716, -240740740.853453, -2777778.2259037, -925925.817461491, -925926.07497077, -93055555.7092453, -93055556.1033194, -4166666.59026134, -292592593.035785, -52777777.9163303, -52777778.289422, 0, 0, 0, 0, 0, 0, -0.00287733908984059, 0, 0.00146461678998323, 0, 0, 0, 0.00267300426713171, 0, 0.00917492896759507, -0.00167257787555194, 0.00350872130917901, -0.00224405914256863, 0, 0, 0, 0.0817613914667591, 0, 0.0161868975102377, 13888888.8787285, 0.0173088553476966, 13888888.9392502, 0.133548086398845, -0.0483580634192822, -0.0581796566303933, 0, 0, 0, 0.141540785346726, 0, 0.0756302677378207, 180555555.548712, -0.0114473602145588, 180555555.647026, 0.449839718533618, -0.117612189023793, -0.000109495918900573, 41666666.7741444, 41666667.1470059, -0.189538098438007, 333333333.384691, -0.0998595616273965, 0.375881275072064, 375000000.326601, 41666666.8185825, 375000000.590169, 41666667.5120671, 13888888.8169204, 13888889.168998, 208333333.930055, 0, 0, 0, 0, 0, 0, 0.000682190242259599, 0, -0.000248201987205169, 0, 0, 0, 0.00113284265602566, 0, -0.00248793228618317, 0.000680168741185091, 0.000246393021245072, -0.000128948712568187, 0, 0, 0, -0.0201769502301408, 0, -0.00264045314189203, 0.00471592298272931, -0.00533037517450227, -0.0100396445829165, -0.0283279500854919, 0.0107608495430415, 0.0157913953647199, 0, 0, 0, -0.0543292086431808, 0, -0.0139395285900973, -46296296.2851369, -0.0008227236501528, -46296296.3223572, -0.154393506837209, 0.0401122371336639, 0.031371151507066, -0.0313497204988743, -0.150331984547475, 0.0874164536541426, 0, 0, -0.0472404548953047, -138888888.888172, 0.0327884985096362, -138888888.911738, -0.222770614668244, 0.057062796211437, -0.0831789775954227, -138888888.978878, -0.0652901079351376, -138888889.148125, -0.41074269043201, 0.1118277977698, 0.00659913417327052, -277777777.982412, 0.00942319030354547, -0.342543215452338, -138888889.132227, -138888889.346314, -46296296.474879}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 5.47459862862766e-06, 0, 0, 0, 0, 0, 0, 0, -7.48643888130603e-05, 0, -0.000126369928679928, 63000.0001837996, 0, 0, 0, 0, 0, 0, 0.000351530050139211, 0, 0.000831862094303785, 0, 0, 0.00174314525058824, -1036750.00291737, -406750.002004888, -1036750.00181921, 0, 0, 0, 0, 0, 0, -0.000296574672182562, 0, -0.00366878923620332, 0, 0, 0, -0.0110971891042873, 0, -0.00686805499952514, 5109166.68155962, 1041666.67701438, 5109166.67129615, 6150833.35951326, 16518333.3652042, 6150833.34940777, 0, 0, 0, 0, 0, 0, -0.00339665614665491, 0, 0.0110757867492848, 0, 0, 0, 0.0411150480257231, 0, 0.0170016460045214, -11739583.3719335, -1322916.69587869, -11739583.3310499, 0, 0, 0.0275308630318, -24802083.4399943, -75893750.1454637, -14385416.7642905, -14385416.7437712, -75893750.1151851, -24802083.3586784, -90279166.9077976, 0, 0, 0, 0, 0, 0, 0.0128266882325377, 0, -0.0203191657696444, 0, 0, 0, -0.0797980559163436, 0, -0.0413400777649154, 14062500.0570611, 833333.384930572, 14062499.9722434, 0, 0, 0, -0.0973501539376804, 0, -0.00633651969207459, 43020833.5413198, 160416666.984729, 16562500.1809037, 16562500.1756449, 160416666.862913, 43020833.2661466, 44687500.3057981, 292708333.78796, 44687500.1733069, 337395834.186876, 193541667.441984, 337395833.990014, 0, 0, 0, 0, 0, 0, -0.0193310655045888, 0, 0.0211184019677802, 0, 0, 0, 0.0791677428801587, 0, 0.0691102655844952, -8541666.71537649, -208333.391732957, -8541666.61973544, 0, 0, 0, 0.178481719176012, 0, -0.0245854203100117, -34375000.213525, -175000000.372052, -9375000.17629701, -9375000.21265074, -175000000.180793, -34374999.7465886, 0, 0, 0.0264116798814494, -52083333.774178, -482291667.442545, -35416667.071358, -35416667.0286563, -482291667.298205, -52083333.4281941, -534375001.31754, -203125001.203616, -534375000.678655, -553125001.919161, -1000000001.99655, -553125001.677981, 0, 0, 0, 0, 0, 0, 0.0136421482393211, 0, -0.0113743755638735, 0, 0, 0, -0.036811871922633, 0, -0.0578783638626298, 2083333.35564862, 0.0383277265503027, 2083333.30031823, 0, 0, 0, -0.147058169139091, 0, 0.00682115268562823, 10416666.7779762, 95833333.5578366, 2083333.42040069, 2083333.47238611, 95833333.425536, 10416666.3879195, 0, 0, 0, -0.0781460682920735, 0, 0.0679750302364077, 20833333.634138, 364583333.957338, 10416666.9076151, 10416666.9663764, 364583333.756666, 20833333.2176469, 385416667.615151, 104166667.565602, 385416666.879934, 20833333.7222769, 541666667.401746, 20833333.637388, 583333335.207215, 1104166668.88279, 395833335.194367, 395833335.194948, 1104166668.40071, 583333334.593848, 1125000003.07658, 0, 0, 0, 0, 0, 0, -0.00372646445014477, 0, 0.00245707148712888, 0, 0, 0, 0.00578942546218683, 0, 0.0182913552799506, -0.0042305714082918, -0.0108584635235154, 0.00854824844544858, 0, 0, 0, 0.0440997292427147, 0, 0.00748716622256658, -0.0231822862190923, -20833333.3882834, -0.0174341663747103, -0.0396466983243126, -20833333.3548926, 0.103826913962748, 0, 0, 0, 0.0461360354241551, 0, -0.036052458298721, -0.0787955499829551, -104166666.860021, -0.0497632192880457, -0.0900026187966211, -104166666.783045, 0.109496902742639, -104166666.928146, -20833333.5951959, -104166666.639837, 0, 0, -0.0161530294013227, -0.136140979113129, -208333333.633399, -0.115656060570423, -0.152451408108557, -208333333.579398, -0.0734244691873729, -208333333.974669, -416666667.511436, -104166667.318858, -104166667.401119, -416666667.162228, -208333333.583164, -208333334.202484, -625000001.04704, -208333334.115487, -625000001.61298, -416666668.22525, -625000001.350949}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 6.71813456178504e-07, 0, 0, 0, 0, 0, 0, 0, -1.25522534424934e-05, 0, -1.26497070105805e-05, -4500.00001275586, 0, 0, 0, 0, 0, 0, 7.65076751217272e-05, 0, 8.67090508412384e-05, 0, 0, 0.000227610931494234, 103250.000190031, 9500.00015760779, 54500.000108773, 0, 0, 0, 0, 0, 0, -0.000239104973397033, 0, -0.000273980344793653, 0, 0, 0, -0.0012323035564056, 0, -0.00152834500806097, -859166.667742191, -5000.00080546663, -100000.000326613, -207916.668413464, -1240416.66840648, -105000.001121938, 0, 0, 0, 0, 0, 0, 0.000464526773453688, 0, 0.00041466611464118, 0, 0, 0, 0.00315949116135463, 0, 0.00484673055991926, 3364583.33631425, 0.0024219543020477, 50000.0004732898, 0, 0, 0.00737438235632853, 1593750.00801401, 10185416.6757623, 104166.672833995, 50000.0043679221, 2183333.33893691, 100000.002859595, 2287500.0132335, 0, 0, 0, 0, 0, 0, -0.000591258425984235, 0, -0.00024226145916626, 0, 0, 0, -0.00466361391736561, 0, -0.00789690271667491, -6562500.00446385, -0.00457868597599561, -0.000277922364669027, 0, 0, 0, -0.0154732725050266, 0, -0.0203960289612379, -5416666.68461499, -39062500.0214227, -0.0113764304532914, -0.0100755630995746, -1041666.67558528, -0.00448048687407448, -729166.687864393, -16666666.6927386, -0.00679889938364353, -17395833.388141, -1041666.70444089, -2083333.36573511, 0, 0, 0, 0, 0, 0, 0.000474185115122484, 0, -7.52334369456586e-05, 0, 0, 0, 0.00421810244513642, 0, 0.00669457731739712, 6041666.67031351, 0.00515101686317494, -0.000158877449524641, 0, 0, 0, 0.0167098297611546, 0, 0.0275783088093875, 8125000.02084337, 73750000.0253126, 0.0125396681266385, 0.014838291836935, 0.00776690413625679, 0.00518889634291995, 0, 0, 0.0337430674759964, 2083333.3688155, 56250000.0494011, 0.0245953092062939, 0.00910539435019734, 7291666.70062938, 0.00512617504309289, 58333333.4295921, 0.0531403023581701, 0.0382847519023035, 7291666.78483967, 14583333.4464693, 0.0625445988628357, 0, 0, 0, 0, 0, 0, -0.000212662257810764, 0, 0.000150203507737682, 0, 0, 0, -0.00221604183082534, 0, -0.00270061981449789, -2083333.33484206, -0.00307405778081925, 0.00033454519904815, 0, 0, 0, -0.00932520235681711, 0, -0.0177764872860295, -4166666.6786166, -64583333.3478748, -0.00802673198199161, -0.0121043846035739, -0.00348791543018116, -0.00364235913262582, 0, 0, 0, -0.0249858532424589, 0, -0.0320138542226822, -2083333.36171498, -83333333.3747985, -0.013245853493603, -0.00903311378123937, -0.0208958219716537, -0.00341360009744588, -85416666.7427439, -0.0390283922850449, -0.0252018470930581, -0.0255615443856181, -20833333.3787854, -0.000785862160799781, -20833333.4728303, -41666666.8065286, -0.105513942584284, -0.0530360222634603, -0.096198804646642, -0.0387309769135657, -0.160402967427736, 0, 0, 0, 0, 0, 0, 3.96747181265733e-05, 0, -4.81225329655333e-05, 0, 0, 0, 0.000518283191163864, 0, 0.000370759939332596, 0.00024228190093999, 0.000739194301345702, -0.000138985561489637, 0, 0, 0, 0.00217354169560292, 0, 0.00435493153867739, 0.00264710628165076, 20833333.3365378, 0.00224571999376486, 0.00399647029314288, 0.000581165799357758, 0.00101458137479922, 0, 0, 0, 0.00700131361665006, 0, 0.0108796112150813, 0.00860037087188127, 41666666.6793846, 0.00292783109133092, 0.00425772895380616, 0.00489995359882265, 0.0015899609221746, 41666666.688848, 0.0123077892393507, 0.00773422581777206, 0, 0, 0.0122499184969819, 0.0102733865065557, 20833333.353455, 0.00543619919308206, 0.000234598092255452, 0.0145592174139332, -0.000875824333072895, 20833333.3891374, 41666666.724141, 0.0339251673007511, 0.0180835801238864, 0.0311090127552065, 0.00900284084362224, 0.0614763748437463, 0.0616927442544877, 0.0111213992059462, 0.101516684506936, 0.0725625078319996, 0.0636933999397354}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -1.74688654232137e-06, 0, 0, 0, 0, 0, 0, 0, 3.4285750084425e-05, 0, 3.45972290506461e-05, -4499.999992324, 0, 0, 0, 0, 0, 0, -0.00020786493179117, 0, -0.000211069675877872, 0, 0, -0.000680805378758423, 54499.9999918218, 9499.99977782238, 103249.99992535, 0, 0, 0, 0, 0, 0, 0.000554852715826649, 0, 0.00063082449864215, 0, 0, 0, 0.00416208949759863, 0, 0.00396676136995338, -100000.000158026, -4999.99852584544, -859166.666318902, -104999.998924406, -1240416.66694588, -207916.664292258, 0, 0, 0, 0, 0, 0, -0.000576740310411051, 0, -0.00103180168264954, 0, 0, 0, -0.0123031103224639, 0, -0.0113900189305931, 50000.0008988581, -0.00464786522008508, 3364583.33267761, 0, 0, -0.0209352461029644, 99999.9972322996, 2183333.33593568, 49999.9922512628, 104166.654056464, 10185416.6677874, 1593749.98804005, 2287499.99597897, 0, 0, 0, 0, 0, 0, -0.000283999403932539, 0, 0.000889546825987155, 0, 0, 0, 0.0197523288314404, 0, 0.0184812863006876, -0.00210754627052012, 0.00778579092279207, -6562500.00015302, 0, 0, 0, 0.0514121958655336, 0, 0.0488368565983231, 0.00373377964880378, -1041666.6753156, 0.0237906684744039, 0.0312462504275207, -39062500.0023532, -5416666.63426224, 0.0191648989772054, -16666666.673483, -729166.619410612, -2083333.33179451, -1041666.64059217, -17395833.3202925, 0, 0, 0, 0, 0, 0, 0.0012352283786603, 0, -0.000286535701089418, 0, 0, 0, -0.0173742025711985, 0, -0.0172536036551795, 0.00254701244601786, -0.00687495229282405, 6041666.66876615, 0, 0, 0, -0.0653971753337381, 0, -0.0602191138354333, -0.00281841584821022, 0.0140572141741608, -0.0360907564182077, -0.0403123597969033, 73750000.0048333, 8124999.95366909, 0, 0, -0.0903297645227678, -0.0235973312490946, 7291666.68309759, -0.0466973373563427, -0.0819360678008657, 56250000.0073982, 2083333.24016123, 0.00810326202643079, -0.0628768833727514, 58333333.2982377, -0.0377373710080075, 14583333.3348151, 7291666.62382077, 0, 0, 0, 0, 0, 0, -0.00106632016240497, 0, -8.41180696643828e-05, 0, 0, 0, 0.00769294012397022, 0, 0.00860489767173203, -0.00157009397244964, 0.00285884495027964, -2083333.33584115, 0, 0, 0, 0.0416228584586576, 0, 0.0382196773493164, 0.00117662302771238, -0.0110106958229826, 0.026444387319635, 0.0256259770990907, -64583333.3392728, -4166666.63396996, 0, 0, 0, 0.0772508793173693, 0, 0.0737389519907593, 0.0150245051621558, -0.0183497766053087, 0.048944800395104, 0.0683888167641726, -83333333.3391457, -2083333.24350664, -0.0130289713240137, 0.066125373712346, -85416666.6256604, 0.0386267349558073, -20833333.3440246, 0.101166697722055, 0.0260231099521478, -0.0182942700492467, 0.0770312822237965, 0.0546135388355538, -41666666.6534838, -20833333.2735501, 0.0190914086257989, 0, 0, 0, 0, 0, 0, 0.000312294060918352, 0, 6.03073391183824e-05, 0, 0, 0, -0.0012821607625293, 0, -0.00176294177855786, 0.000391962688488913, -0.000379681304752109, 0.000935806063367309, 0, 0, 0, -0.0104397211747208, 0, -0.00983797689494423, -0.000208457435229542, 0.00332370209391337, -0.00745391371993691, -0.00622330480602957, 20833333.3359945, -0.00890912466262084, 0, 0, 0, -0.0249923635298808, 0, -0.0231754813340801, -0.00409390722273973, 0.00748432701574675, -0.0185262671895524, -0.0219783598069523, 41666666.6695557, -0.0329954035583322, 0.00572322828486686, -0.0249681528308861, 41666666.650641, 0, 0, -0.0313078374510204, -0.0119177617867545, 0.00620277469236895, -0.0201886376805674, -0.0401267035616835, 20833333.3360884, -0.0469382364838515, -0.00598261625507145, 0.0125425991724912, -0.0446319222762308, -0.0266931510040272, 41666666.6543912, 20833333.297573, -0.0329711833495989, 0.00256152756910396, -0.0268458232765043, -0.00119485682476021, -0.0216304675834661, -0.0122150709024694}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -3.7449489641492e-06, 0, 0, 0, 0, 0, 0, 0, 5.13160264363043e-05, 0, 6.33492697657056e-05, -35000.0001161468, 0, 0, 0, 0, 0, 0, -0.000424565456193757, 0, -0.000475783272653514, 0, 0, -0.000477594545291638, 692638.890998494, 167638.890230143, 517638.889816673, 0, 0, 0, 0, 0, 0, 0.00185579919996692, 0, 0.00260984700045516, 0, 0, 0, 0.00368728761133387, 0, 0.000195637285747107, -4563888.90013858, -299305.562078027, -1975694.44482989, -3113194.46424459, -10039583.3548181, -2275000.00960528, 0, 0, 0, 0, 0, 0, -0.00404396925935803, 0, -0.00892996580817863, 0, 0, 0, -0.0186595705673409, 0, 0.000646525722049366, 13107638.9182803, 236111.12738482, 3229166.65406616, 0, 0, 0.00413304864674518, 18069444.5305892, 63708333.4384281, 5197916.74149029, 3701388.93513639, 36329861.1828142, 6694444.45026547, 41527777.9441633, 0, 0, 0, 0, 0, 0, 0.0043333935497833, 0, 0.0174686744494455, 0, 0, 0, 0.0444889847397195, 0, 0.00949781234896317, -18576388.9314609, -69444.4682725785, -2430555.5126405, 0, 0, 0, 0.0119283010395796, 0, -0.046171117893261, -40833333.5107372, -171909722.455473, -3819444.5836917, -2638888.99079735, -55798611.2182929, -7361111.03324315, -26006944.6973246, -206701389.209452, -7500000.07574724, -232708333.970979, -63437500.5388457, -115416667.073244, 0, 0, 0, 0, 0, 0, -0.00178403728998975, 0, -0.0190285375052933, 0, 0, 0, -0.0529723725226081, 0, -0.0303062129569101, 12847222.2569055, 0.0214912762800107, 694444.382680574, 0, 0, 0, -0.053752499708446, 0, 0.0739369608467354, 39583333.5244382, 225347222.489653, 1041666.80312078, 694444.559957439, 39236111.1867562, 2777777.56607517, 0, 0, 0.072464448278543, 41666667.055067, 450000000.569216, 15972222.5594136, 2777777.95139032, 276041667.10193, 4166666.63266783, 491666667.700282, 42361111.9417901, 118750000.352506, 307986112.589908, 568055556.937899, 120833334.366319, 0, 0, 0, 0, 0, 0, -0.000321350236490072, 0, 0.010756774848069, 0, 0, 0, 0.0308427646155807, 0, 0.0312685890579154, -3472222.2370162, -0.0113052505281525, 0.042377101271695, 0, 0, 0, 0.0588611224820237, 0, -0.0393587923777671, -13888888.9925255, -142361111.264908, -0.0679348077561427, -0.0676995928393282, -10416666.6874823, 0.211017188159189, 0, 0, 0, -0.0262207816105144, 0, -0.110262939542108, -20833333.6159566, -416666667.124318, -3472222.4267423, -0.148889322941077, -163194444.724059, 0.167078371381441, -437500000.767031, -10416667.2746905, -41666666.7131024, -13888889.2288337, -430555556.102235, -0.0845887625736493, -458333334.889104, -875000001.614724, -173611112.56349, -41666667.8060067, -493055556.681674, -62500000.6402445, -500000002.161762, 0, 0, 0, 0, 0, 0, 0.000337026058645463, 0, -0.00246056277502809, 0, 0, 0, -0.00696940366448638, 0, -0.0108696129964813, 0.00255091753080908, 0.00266953535818964, -0.0113473526564405, 0, 0, 0, -0.0207473884623909, 0, 0.0040050462005237, 0.0222105445625628, 34722222.2572513, 0.0135654700675737, 0.0166353741094773, 1.84005362031512e-05, -0.0749289817523438, 0, 0, 0, -0.0045146918239774, 0, 0.0419748207843438, 0.0789469009297354, 138888889.02684, 0.0452334660736626, 0.0449787799089921, 34722222.2929989, -0.0999221348112467, 138888889.103203, 0.171725724859461, -0.0485332364070718, 0, 0, 0.0384867995372038, 0.124855506362467, 208333333.561766, 0.0994685529042314, 0.0579745276976191, 138888889.073617, -0.0127506883091668, 208333333.899829, 416666667.299013, 34722222.7287224, 0.443706719809858, 138888889.19075, 0.0662000442184299, 138888889.633186, 416666667.394387, 0.400837555342997, 416666667.883151, 138888889.997173, 208333334.200386}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -4.07969197702774e-07, 0, 0, 0, 0, 0, 0, 0, 8.99725088570253e-06, 0, 1.3264289139373e-05, 15000.000051759, 0, 0, 0, 0, 0, 0, -1.19415164458445e-06, 0, -8.81712492977552e-05, 0, 0, -0.000359928327672345, -325416.667531046, -50416.6672995724, -200416.667086536, 0, 0, 0, 0, 0, 0, -0.000257501041888737, 0, 5.68170210981667e-05, 0, 0, 0, 0.00157660221088061, 0, 0.00317828558229285, 2480555.56022348, 56250.0031651016, 560416.667302627, 1036805.56388053, 4290972.23063993, 616666.671126412, 0, 0, 0, 0, 0, 0, 0.000978393442820697, 0, 0.000926703477846794, 0, 0, 0, -0.0015575852735817, 0, -0.0104593776690846, -8593750.01228754, -20833.3419231689, -583333.330940457, 0, 0, -0.0159731489500417, -7166666.70343731, -31972222.2642892, -1093750.03150714, -625000.019642617, -11461805.5833622, -1187500.00632404, -12555555.6216288, 0, 0, 0, 0, 0, 0, -0.00162059820311743, 0, -0.00304934506818571, 0, 0, 0, -0.00254347132390492, 0, 0.0151138829829503, 14409722.2398619, 0.0140331523000736, 208333.323061889, 0, 0, 0, 0.0284116529767832, 0, 0.0518341059310738, 20798611.1891547, 106736111.205256, 381944.504907669, 208333.377968963, 11319444.4871146, 624999.988223197, 6770833.43976385, 78437500.1269096, 625000.028191309, 85208333.5945021, 12083333.5407274, 23020833.493073, 0, 0, 0, 0, 0, 0, 0.00139141616096463, 0, 0.00408138066034985, 0, 0, 0, 0.00656205009576648, 0, -0.00859773029413384, -11458333.3472623, -0.0136549004028667, 0.0156663469137557, 0, 0, 0, -0.0209498808884479, 0, -0.0728589658102182, -25000000.0866095, -169097222.329029, -0.0645617720991932, -0.0573293569915412, -3819444.4767282, 0.0421724143830455, 0, 0, -0.0836947877997782, -15625000.1683033, -223611111.339604, -2083333.47266822, -0.0547198854754891, -69791666.8391229, 0.0079361144104858, -239236111.546312, -3819444.76023218, -11458333.4904728, -73958333.9332658, -141666667.220975, -11458333.6999317, 0, 0, 0, 0, 0, 0, -0.000600888429339671, 0, -0.00256292510507673, 0, 0, 0, -0.00501469130785554, 0, -0.000190992647276251, 3472222.22781253, 0.00716680399307265, -0.011130457956579, 0, 0, 0, 0.00456429637961054, 0, 0.0457184930998179, 10416666.7146713, 125000000.059556, 0.0371541485869287, 0.0394221610207436, 0.0103438285022705, -0.0434071248807629, 0, 0, 0, 0.055555232384056, 0, 0.0881942271784187, 10416666.7942765, 260416666.850001, 0.0856835065245203, 0.0514871158389431, 20833333.4435615, -0.0422026355295592, 270833333.662446, 0.236867845160575, 0.0589616047941021, 3472222.36223071, 159722222.442766, 0.00855629591741927, 166666667.327359, 322916667.3261, 20833333.9129742, 0.37530467332106, 62500000.4630695, 0.217529506733815, 62500000.8404929, 0, 0, 0, 0, 0, 0, 0.000101831912215008, 0, 0.000622671086227155, 0, 0, 0, 0.0013252557256872, 0, 0.00130345494557931, -0.000868178467245564, -0.00153781261844255, 0.00307145298827999, 0, 0, 0, 0.000808081434765421, 0, -0.0102149197206045, -0.0103825197799178, -34722222.2350744, -0.00897878463897419, -0.0111425338209886, -0.000464752255119703, 0.0155961181110268, 0, 0, 0, -0.0128785721984155, 0, -0.0310462457661574, -0.0371640639567483, -104166666.720976, -0.0206538711614213, -0.0190956146587399, -0.0270183948825401, 0.0230794287014719, -104166666.759384, -0.0709913310720784, -0.0030571429958241, 0, 0, -0.032613667028697, -0.052905615161863, -104166666.759452, -0.0385076392888407, -0.00962642444860798, -34722222.2967162, 0.0120641197818351, -104166666.916932, -208333333.594182, -0.201324396464493, -0.143435753544527, -0.136391948358931, -0.027968554362625, -34722222.5340208, -104166666.964131, -0.108476306667151, -104166667.163163, -0.415131733037774, -0.333909192917845}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -3.99343338928542e-07, 0, 0, 0, 0, 0, 0, 0, 1.12011546567075e-05, 0, 5.28235105452065e-07, -3750.00002453102, 0, 0, 0, 0, 0, 0, -0.000117632607203899, 0, 1.80838107594838e-05, 0, 0, -3.16138574549442e-05, 76666.6670173087, 7916.66697692264, 64166.6669126343, 0, 0, 0, 0, 0, 0, 0.000584876184196668, 0, -4.8401450856394e-05, 0, 0, 0, 0.000568390802703626, 0, -0.000399969054429049, -524305.557327806, -4166.66832481636, -310416.667492694, -153472.225858264, -1303472.22564885, -127083.33598002, 0, 0, 0, 0, 0, 0, -0.00156423010450134, 0, -9.84245699415624e-05, 0, 0, 0, -0.00322731501509562, 0, 0.00195582410573451, 1493055.56011081, 0.00439923609681608, 458333.334378955, 0, 0, 0.00117315094094878, 944444.458825239, 8809027.79370131, 76388.9049044436, 62500.0129313421, 6211805.56738696, 520833.340519005, 2454861.13906549, 0, 0, 0, 0, 0, 0, 0.0024097806342399, 0, 0.000549381214842665, 0, 0, 0, 0.00811563703684752, 0, -0.00313695134397174, -1736111.11771462, -0.0062713980266542, -208333.333210927, 0, 0, 0, 0.00144817611374023, 0, -0.00863190496524944, -2152777.80586373, -24548611.1462632, -0.0354576720320685, -0.0296011695741568, -8819444.46418118, -416666.672557229, -416666.714830806, -40798611.159901, -208333.367540655, -15000000.0974903, -1145833.43402815, -9965277.84689471, 0, 0, 0, 0, 0, 0, -0.00214088035291095, 0, -0.000845795345048185, 0, 0, 0, -0.0104432205877885, 0, 0.00172029794037979, 694444.449880206, 0.00472420596868842, -0.00156637639914165, 0, 0, 0, -0.00775246617281831, 0, 0.0145033931906276, 1388888.91851453, 27430555.5960228, 0.0418760126289044, 0.0339761459301894, 3819444.46207575, -0.00510962035358348, 0, 0, 0.0107125851239241, 694444.511738968, 107638888.975457, 0.0768307412861841, 0.0665968376689381, 53472222.290481, 0.0375937688085534, 33680555.7080148, 0.171341021760752, 7638888.95849, 6250000.24825671, 59722222.4250697, 3819444.63767096, 0, 0, 0, 0, 0, 0, 0.00101552909446877, 0, 0.000568672089439159, 0, 0, 0, 0.00676444527629159, 0, 0.00029240166085335, -0.00235400916358081, -0.00162872054726589, 0.00139221964501428, 0, 0, 0, 0.00886096783414632, 0, -0.00978635100177493, -0.0161702493921922, -10416666.6900732, -0.0252122537979966, -0.0187303098980418, -0.00836861353775912, 0.0109215738277856, 0, 0, 0, -0.00382231662017717, 0, -0.0156694798196111, -0.0458706241957204, -107638888.959175, -0.0590233587749092, -0.0559418152885653, -20833333.379102, -0.0141550073371496, -20833333.4448589, -0.136445385470274, -0.0253244488377479, -0.0679443057788978, -118055555.641182, -0.0601020400649076, -10416666.9138009, -128472222.450235, -0.270591049089529, -0.230950772296946, -41666666.8308623, -0.136957315394928, -20833333.6764108, 0, 0, 0, 0, 0, 0, -0.000198253695055818, 0, -0.000143639055773752, 0, 0, 0, -0.00174688856880271, 0, -0.000409659810109166, 0.000413094890969687, 0.000149323778424563, -0.000388228150971191, 0, 0, 0, -0.00333108235652207, 0, 0.00225778655121819, 0.0035679976856308, 0.00535134706427721, 0.00604787550226991, 0.00382553593338174, 0.00173357453426978, -0.00478575988907353, 0, 0, 0, -0.000446925317498368, 0, 0.0061625966496437, 0.0124188291625979, 34722222.243632, 0.017708045369613, 0.0168803046925005, 0.0121371515610293, -0.000735031948883416, 0.031123374221321, 0.0411730206244873, 0.000294490993724828, 0, 0, 0.00497428602878643, 0.0220777456637237, 69444444.4807342, 0.0248078440774262, 0.0283215374449916, 34722222.2522238, 0.0180344628512458, 0.0870820177880771, 69444444.5325748, 0.104848170629801, 0.0956293454349249, 0.0452983223082443, 0.0282881303401052, 0.13001758989436, 69444444.5455897, 0.0896350025683754, 34722222.4038665, 0.192308309442189, 0.13550437608812}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -2.00757602307053e-06, 0, 0, 0, 0, 0, 0, 0, 3.81954201478092e-05, 0, 3.43113273676243e-05, -3750.00001317556, 0, 0, 0, 0, 0, 0, -0.000269169248248144, 0, -0.000222135695496174, 0, 0, -0.00058082264787167, 64166.6669043944, 7916.66674296871, 76666.6668222477, 0, 0, 0, 0, 0, 0, 0.000983768398161604, 0, 0.00083292215177537, 0, 0, 0, 0.00360184662905496, 0, 0.00308890473361219, -310416.66801412, -4166.66693828929, -524305.556095251, -127083.33492988, -1303472.22495147, -153472.222787674, 0, 0, 0, 0, 0, 0, -0.00201737025664132, 0, -0.00198224848957, 0, 0, 0, -0.0117683183481976, 0, -0.00916232687458223, 458333.337251801, 0.000241551220993463, 1493055.5561772, 0, 0, -0.0149974691096906, 520833.340560947, 6211805.56903763, 62500.0051618778, 76388.8919931957, 8809027.78801441, 944444.443044099, 2454861.12652365, 0, 0, 0, 0, 0, 0, 0.00234874131509989, 0, 0.00296879056180102, 0, 0, 0, 0.0215769187603871, 0, 0.0169420195481484, -208333.339775806, 0.000556041616604413, -1736111.11104378, 0, 0, 0, 0.0385155032063653, 0, 0.0320405376120653, -416666.682341047, -8819444.47721519, -0.00757381479157432, -0.00564675636647772, -24548611.1302586, -2152777.76113249, -208333.350401235, -40798611.1531351, -416666.673286036, -9965277.838812, -1145833.38261852, -15000000.0356823, 0, 0, 0, 0, 0, 0, -0.00146274443966374, 0, -0.00266780776860265, 0, 0, 0, -0.0222490528828219, 0, -0.0190965756325762, 0.00606410471813951, -0.00121901765118849, 694444.443985517, 0, 0, 0, -0.0533772988934896, 0, -0.0398048365866078, 0.0178693927082916, 3819444.48641029, 0.00401747631797861, 0.00181661343381857, 27430555.5773573, 1388888.84897922, 0, 0, -0.0580775323545216, 0.0253181776773361, 53472222.3014021, 0.018164984062899, 0.0221963391725567, 107638888.949218, 694444.433678591, 7638888.99715284, 0.0735092136350381, 33680555.5702293, 3819444.57078329, 59722222.3481922, 6250000.1099658, 0, 0, 0, 0, 0, 0, 0.000392428063422353, 0, 0.00130107439975725, 0, 0, 0, 0.0120269055524522, 0, 0.011803415785465, -0.00305980415319565, 0.000775532136169786, 4.40012636593725e-05, 0, 0, 0, 0.0374732926997554, 0, 0.0274747542047733, -0.0103834221892694, -0.0270175841524554, 0.00058119172801729, 0.00281314758851831, -10416666.6818529, 0.0383372291703466, 0, 0, 0, 0.0523202642484567, 0, 0.0439799321390098, -0.017219934779735, -20833333.4032225, -0.00691226872018255, -0.0163694256488835, -107638888.933759, 0.0353716135779319, -0.0897092047105972, -0.0463863606301522, -20833333.3091106, -0.0172291452592303, -118055555.632023, -0.0183215186982853, -0.13480854753817, -41666666.8303444, -0.113652671158207, -0.142347860458309, -128472222.303153, -10416666.7298092, -20833333.5297267, 0, 0, 0, 0, 0, 0, -1.18451294756254e-05, 0, -0.000262893213171421, 0, 0, 0, -0.00264063089691341, 0, -0.00303152532029096, 0.000644574964883352, -0.000144480948920689, 0.000124971632017133, 0, 0, 0, -0.0104569094447551, 0, -0.00806619987126682, 0.00244857076848498, 0.00688312591308055, -0.000830136158343844, -0.00165627149997463, 0.00482580060193869, -0.0131675919705344, 0, 0, 0, -0.0181973921950622, 0, -0.0137735505502266, 0.00436009358322416, 0.0232007793011188, -4.8607242731293e-05, 0.00258612087794931, 34722222.2368141, -0.0213596602167655, 0.0281634076726985, 0.00961484800007092, -0.017268132009002, 0, 0, -0.0188452572548111, 0.00540790366532925, 34722222.2564923, 0.00349707338947405, 0.01199376539686, 69444444.4719305, -0.000681149719029465, 0.051602703622268, 0.0727174436879289, 0.033857573439077, 0.0569778222806703, 69444444.4548795, -0.00420521086561331, 0.0557155385886094, 69444444.50869, 0.0635216981848556, 0.109915917292648, 0.107912110513489, 34722222.2909227}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 5.08412110775138e-06, 0, 0, 0, 0, 0, 0, 0, -9.41494611820112e-05, 0, -0.0001014331136251, 15000.0000058089, 0, 0, 0, 0, 0, 0, 0.00055963468537405, 0, 0.000624653949649517, 0, 0, 0.00186588163821075, -200416.666921504, -50416.6664186008, -325416.666809747, 0, 0, 0, 0, 0, 0, -0.00142333945185193, 0, -0.00201947857186572, 0, 0, 0, -0.0114818045293637, 0, -0.0103358680185372, 560416.66824389, 56249.9980655546, 2480555.55623191, 616666.66639403, 4290972.22657111, 1036805.55299941, 0, 0, 0, 0, 0, 0, 0.00101640862117076, 0, 0.00399328058205801, 0, 0, 0, 0.03513540185243, 0, 0.0289024943543625, -583333.338347766, -20833.3272839002, -8593750.00169469, 0, 0, 0.0536602211209081, -1187500.00016587, -11461805.576658, -624999.990254999, -1093749.98373393, -31972222.2406575, -7166666.65066422, -12555555.5730253, 0, 0, 0, 0, 0, 0, 0.00235431140293059, 0, -0.00494720159403834, 0, 0, 0, -0.0580804729148036, 0, -0.048066789346457, 208333.342444554, -0.00849847572135469, 14409722.2258277, 0, 0, 0, -0.137168073904446, 0, -0.115613159747757, 625000.00073023, 11319444.4964925, 208333.294927568, 381944.39986273, 106736111.148911, 20798611.0557757, 624999.970682499, 78437500.0688445, 6770833.27101393, 23020833.4089316, 12083333.3563232, 85208333.394867, 0, 0, 0, 0, 0, 0, -0.00559132089678964, 0, 0.00365139778329112, 0, 0, 0, 0.0517054110007071, 0, 0.0485648676699385, -0.00950816275327619, 0.00396820812200156, -11458333.3392028, 0, 0, 0, 0.18298309556201, 0, 0.135121307005808, 3.59642038531578e-05, -3819444.51489768, 0.063620192432443, 0.0599574712566712, -169097222.269612, -24999999.9035317, 0, 0, 0.219620750128998, 0.044678746684699, -69791666.7907072, 0.0906684006000537, -2083333.21993011, -223611111.209826, -15624999.8643094, -11458333.4636575, -3819444.42794148, -239236111.158829, -11458333.4058431, -141666666.8929, -73958333.4618082, 0, 0, 0, 0, 0, 0, 0.00443723641418424, 0, -0.0014124745897235, 0, 0, 0, -0.0225817606047445, 0, -0.0273463444628241, 0.0053041213827929, 0.00165661847748679, 3472222.22750853, 0, 0, 0, -0.121012142332448, 0, -0.0849668929361315, -0.00098329440405027, 0.0491009634599553, -0.047834446402439, -0.0363835917915171, 125000000.035638, 10416666.5861029, 0, 0, 0, -0.198359085696806, 0, -0.163821905283863, -0.0352709450197436, 20833333.4446667, -0.10287662979478, -0.104679469964137, 260416666.74078, 10416666.5180689, 0.106826950470827, -0.053698531325122, 270833333.317275, -0.087016118991685, 159722222.334052, 3472222.0861934, 0.0679883657533393, 62500000.2704951, -0.0334397639073132, 20833333.4520242, 322916666.859352, 166666666.756518, 62500000.2804244, 0, 0, 0, 0, 0, 0, -0.00126382614681137, 0, 0.000206164984109227, 0, 0, 0, 0.00353307593404589, 0, 0.00650687684213464, -0.00122578880509802, -0.00150110264485774, -0.00186941120742344, 0, 0, 0, 0.0311466343040774, 0, 0.0227141495883262, 0.000496601199032816, -0.0137025178070926, 0.0133518144667769, 0.00729196283935225, -34722222.2340381, 0.0256233949512054, 0, 0, 0, 0.0676245653790346, 0, 0.0481410798926441, 0.0115060691296586, -0.0390611007128076, 0.0401965132003779, 0.0365397555725331, -104166666.691695, 0.0622390777255732, -0.0340026012980563, 0.0268881297204863, -104166666.64612, 0, 0, 0.073002454241677, 0.0308535969898465, -34722222.2711131, 0.0476822749176388, 0.0571399511121668, -104166666.704283, 0.0644972014869822, -0.0229950988488013, -0.113707250523368, 0.0534006719102247, -0.0329131500364316, -208333333.379858, -104166666.659724, 0.0279521788354418, -104166666.792698, -34722222.2912129, -0.146879664240693, -0.108448336947233, -104166666.80697}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -7.05991203880938e-06, 0, 0, 0, 0, 0, 0, 0, 0.000123482394606498, 0, 0.000143858065340987, -35000.0000913214, 0, 0, 0, 0, 0, 0, -0.000699580172923134, 0, -0.000866955388225042, 0, 0, -0.00250624743404813, 517638.890255726, 167638.889624513, 692638.89006357, 0, 0, 0, 0, 0, 0, 0.00152231008545167, 0, 0.00302503775233838, 0, 0, 0, 0.0157254508046171, 0, 0.0124171431777976, -1975694.45133222, -299305.558923496, -4563888.89378848, -2275000.00969185, -10039583.350812, -3113194.45192119, 0, 0, 0, 0, 0, 0, 0.000488439375563628, 0, -0.00714374910254973, 0, 0, 0, -0.0511867716243457, 0, -0.032656305220434, 3229166.68513756, 236111.120565947, 13107638.8980363, 0, 0, -0.0623879291701581, 6694444.48028858, 36329861.1879996, 3701388.91801061, 5197916.69746369, 63708333.4033386, 18069444.4622545, 41527777.8960187, 0, 0, 0, 0, 0, 0, -0.00791014548257957, 0, 0.0110015715529244, 0, 0, 0, 0.0892988221368042, 0, 0.0578277419663503, -2430555.58465275, -69444.4638479312, -18576388.8972278, 0, 0, 0, 0.172341661483857, 0, 0.108826245055954, -7361111.17499452, -55798611.2830024, -2638888.92761882, -3819444.50496351, -171909722.356646, -40833333.3179806, -7500000.07381663, -206701389.141491, -26006944.5366196, -115416667.058017, -63437500.3559253, -232708333.705989, 0, 0, 0, 0, 0, 0, 0.013925750092111, 0, -0.0101169705613584, 0, 0, 0, -0.0822683228477491, 0, -0.0683310771120662, 694444.471443758, 0.0276136837832673, 12847222.2261871, 0, 0, 0, -0.251159227351482, 0, -0.108676499952685, 2777777.83750797, 39236111.3227884, 694444.463575458, 1041666.73093223, 225347222.368907, 39583333.2286524, 0, 0, -0.217305422469232, 4166666.75088955, 276041667.09649, 2777777.82968674, 15972222.3894069, 450000000.365279, 41666666.752417, 118750000.585952, 42361111.624598, 491666667.105935, 120833334.118145, 568055556.583723, 307986112.058715, 0, 0, 0, 0, 0, 0, -0.0103099642507291, 0, 0.00491914321987208, 0, 0, 0, 0.036275833802723, 0, 0.0458707459283591, -0.0136505021254076, -0.0221410033230172, -3472222.22392114, 0, 0, 0, 0.178342939392398, 0, 0.0687876319849233, -0.0280768996134859, -10416666.8028481, 0.00114169869608849, -0.0420376006022693, -142361111.203134, -13888888.7602616, 0, 0, 0, 0.221583798532717, 0, 0.120200216661505, -0.0416009245502758, -163194444.802359, 0.0107700792381032, -3472222.33618889, -416666666.926737, -20833333.3118199, -41666667.0875757, -10416667.0170899, -437500000.186416, -0.0277408296424548, -430555555.966045, -13888889.0725662, -62500000.694818, -493055556.659552, -41666667.3372163, -173611112.151036, -875000000.942842, -458333334.166107, -500000001.53498, 0, 0, 0, 0, 0, 0, 0.00286664951455096, 0, -0.000954836499230168, 0, 0, 0, -0.00547012860033549, 0, -0.0127342014491978, 0.00289645427446035, 0.00721257995035218, 0.000752648035250742, 0, 0, 0, -0.0484252589023678, 0, -0.0216491033663928, 0.00521609904678385, 0.0356627171478744, -0.00221941428677316, 0.0141748136363563, 34722222.2482849, -0.0502472997650585, 0, 0, 0, -0.084247762303664, 0, -0.0267267219483719, 0.00616112022664497, 34722222.338923, -0.0180240032303518, 0.0263325389255181, 138888888.966857, -0.046165270782783, 0.118283946012657, 0.0929229822984418, 138888888.89823, 0, 0, -0.0610162827098934, 0.00324481001675101, 138888889.059259, -0.0120870233494296, 0.0768776879811543, 208333333.473432, 0.0599444246546876, 0.219251518183724, 138888889.310313, 0.199748856764411, 34722222.6200787, 416666666.947278, 208333333.542794, 0.261272716494597, 416666667.209023, 138888889.410483, 208333334.078996, 138888889.646669, 416666667.40855}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 3.75389436895403e-06, 0, 0, 0, 0, 0, 0, 0, -6.7683158666995e-05, 0, -6.03979782617083e-05, 13333.3333969172, 0, 0, 0, 0, 0, 0, 0.000504672041275651, 0, 0.000389090749592521, 0, 0, 0.000882765523517025, -244814.815809628, -44814.8155083316, -244814.815482062, 0, 0, 0, 0, 0, 0, -0.00196114710469207, 0, -0.00165605696687344, 0, 0, 0, -0.0058045736121916, 0, -0.00371309866107834, 1388888.89404834, 50000.0035309928, 1388888.89086598, 772222.231167521, 4444444.45516045, 772222.22839308, 0, 0, 0, 0, 0, 0, 0.00415090126280842, 0, 0.00471418748868142, 0, 0, 0, 0.0212091521414809, 0, 0.010061538494158, -3009259.27296951, -18518.5273793968, -3009259.25976392, 0, 0, 0.0169041903234887, -3796296.33226682, -24629629.6797142, -805555.591296508, -805555.586481339, -24629629.6682402, -3796296.30976317, -13194444.5231977, 0, 0, 0, 0, 0, 0, -0.00480639402071856, 0, -0.00829199713115551, 0, 0, 0, -0.042439448346687, 0, -0.0213680819881343, 2777777.79854355, 0.0123320636629499, 2777777.77057381, 0, 0, 0, -0.0509814164206938, 0, -0.0222537041229769, 5833333.40398832, 50972222.3354127, 277777.848769714, 277777.845424115, 50972222.2880006, 5833333.32128265, 3333333.44170179, 129722222.379187, 3333333.42147681, 63611111.3876952, 12916666.9368278, 63611111.3158999, 0, 0, 0, 0, 0, 0, 0.00279387825445192, 0, 0.00847221633124506, 0, 0, 0, 0.0464682767465994, 0, 0.0300260543219318, -925925.944050236, -0.0102124673603066, -925925.912319744, 0, 0, 0, 0.0828136565536231, 0, 0.020444599587251, -2777777.85196205, -44444444.5799646, -0.0736620168597837, -0.0702703874588309, -44444444.5056509, -2777777.7027042, 0, 0, 0.0445621343505084, -2777777.92939648, -240740741.021319, -925926.081101645, -925926.107627397, -240740740.961433, -2777777.86141123, -93055555.9928585, -4166667.10170076, -93055555.7445073, -52777778.4293761, -292592593.201459, -52777778.3620256, 0, 0, 0, 0, 0, 0, -0.00054783979902059, 0, -0.00458213274108388, 0, 0, 0, -0.0261382503942501, 0, -0.0225446120055057, 0.00849689300261255, 0.00500559628996124, -0.00983830132655395, 0, 0, 0, -0.0653425290491556, 0, -0.0169413571308498, 0.0401089805221107, 13888888.9714077, 0.0385884453244934, 0.034287292183752, 13888888.9210161, -0.0889104836216912, 0, 0, 0, -0.0525533592334901, 0, -0.014012194730298, 0.101414631680253, 180555555.78948, 0.10385621774995, 0.144598776507475, 180555555.706864, -0.00447042001665621, 41666666.9921096, 0.321192092496135, 41666666.6988049, 0.139996154829391, 333333333.606955, 0.181084157435795, 41666667.3066435, 375000000.695107, 13888889.5519231, 13888889.5893641, 375000000.484963, 41666667.0969892, 208333334.329557, 0, 0, 0, 0, 0, 0, -7.00988103811063e-05, 0, 0.00101131509063379, 0, 0, 0, 0.0058798542579387, 0, 0.00671839768167586, -0.00166008694464382, -0.00116833309383946, 0.00256264243920414, 0, 0, 0, 0.0198407364979066, 0, 0.00724536016438128, -0.00880229742233844, -0.02009870254594, -0.00812141517353734, -0.00632319417310754, -0.00787416852645806, 0.0339106078613219, 0, 0, 0, 0.0226609254395471, 0, 0.00151233524308781, -0.0263903882226113, -46296296.3702649, -0.0262588197906968, -0.0390070713771703, -46296296.3391135, 0.0289767455329776, -0.0928562712227648, -0.0885718077478041, 0.0229284565431133, 0, 0, 0.0087442638993027, -0.0456425509063325, -138888889.006334, -0.0456032496562327, -0.0865724221779492, -138888888.984253, -0.0529295281956211, -0.22338368914737, -138888889.164488, -0.2370185826217, -0.281428634566474, -138888889.009946, -0.0775435500820459, -0.310615252154404, -277777778.086192, -0.307069482938925, -138888889.407107, -46296296.8419266, -138888889.295093}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -1.06876899016273e-07, 0, 0, 0, 0, 0, 0, 0, 2.51065451140919e-06, 0, 1.70715183154593e-06, 500.000000812697, 0, 0, 0, 0, 0, 0, -1.59338639496109e-05, 0, -8.15632221598693e-06, 0, 0, -4.46510338547501e-05, -10916.6666972108, -5500.00000708072, -5499.99999369176, 0, 0, 0, 0, 0, 0, 4.5436503498406e-05, 0, 4.01143617107737e-06, 0, 0, 0, 0.000269794732009072, 0, 0.000284959613190872, 83333.333513838, 5000.00002538035, 4999.9998548174, 119583.333589878, 119583.333679038, 59999.999865471, 0, 0, 0, 0, 0, 0, -7.4599928118533e-05, 0, 6.85465133510171e-05, 0, 0, 0, -0.000664220031492821, 0, -0.000796482337329253, -281250.000442829, -0.000125886678300511, 0.000704464197796453, 0, 0, -0.00167267215883529, -906250.00151441, -906250.002011187, -104166.667167113, -49999.9992708218, -104166.667762604, -49999.9983177505, -1300000.00237898, 0, 0, 0, 0, 0, 0, 8.11927201798399e-05, 0, -0.000207882214651423, 0, 0, 0, 0.000829770558861383, 0, 0.00100930502925844, 416666.667157813, 0.000462456036944778, -0.00156479639161904, 0, 0, 0, 0.00365946481596142, 0, 0.00453649853744643, 3020833.33704465, 3020833.33774775, 2.94916246073256e-05, -0.00152956946424062, 0.00131080129968888, -0.00602192430418248, 729166.669547234, 729166.673120909, -0.00611646807555756, 9791666.68010622, 1041666.67121566, 1041666.67146327, 0, 0, 0, 0, 0, 0, -6.04792476676025e-05, 0, 0.000261656939043699, 0, 0, 0, -0.000578119333888744, 0, -0.000455416740963635, -208333.333508543, -0.000813742767586924, 0.00182799444637758, 0, 0, 0, -0.00371567417775502, 0, -0.00579104364246525, -4375000.00433268, -4375000.0044693, 0.000552979421765959, 0.00113076791883032, -0.000388631014529282, 0.00966178021476327, 0, 0, -0.00828876402571648, -2083333.33987645, -2083333.3450255, -0.000529709468016419, 0.00986314135253089, -0.00865701800075267, 0.0155002310992826, -32291666.6918515, -0.00271508217209291, -0.00104978732005451, -7291666.69133273, -7291666.69744707, -0.00410704629770982, 0, 0, 0, 0, 0, 0, 2.7377303824536e-05, 0, -0.000156008219262417, 0, 0, 0, 0.000231556786693298, 0, -0.000107164945160721, -7.50031669695823e-05, 0.000654284323771318, -0.00110020790459885, 0, 0, 0, 0.00178142819299329, 0, 0.00337042928224146, 2083333.33564193, 2083333.33528461, -0.00026480864543955, 9.55572811823244e-05, -0.000345803547369798, -0.00732147867785337, 0, 0, 0, 0.00614395946202481, 0, 0.007869108922544, 2083333.33953732, 2083333.34209797, -0.00221125539820111, -0.00706358781198737, 0.00517013349911105, -0.0160558417279141, 45833333.3532625, 7.65415645573906e-05, -0.00394551377820862, 0.00239776835488586, 0.0114319768917898, -0.0161761095816177, 20833333.3674323, 20833333.3748039, 0.0155409885245285, -0.0021991483176981, 0.0256275192796143, -0.00678193258671283, 0.0386031638734986, 0, 0, 0, 0, 0, 0, -5.39335427605682e-06, 0, 3.62319810442569e-05, 0, 0, 0, -4.57115639512906e-05, 0, 0.000106879801163786, 5.12528431827247e-05, -0.000196051157985069, 0.000270370487932484, 0, 0, 0, -0.000330645758681386, 0, -0.000695429706766824, -0.000425805348359542, -0.000239702349713261, -4.49576774715099e-05, -0.000300945152630245, 0.000207093942781914, 0.00214554123728562, 0, 0, 0, -0.00160307503707627, 0, -0.00260471148664048, -0.00203108531372753, -0.00226816839390085, 0.00118003017772264, 0.00175494296686371, -0.00112985720729524, 0.00594695199320527, -20833333.3389304, 8.11836974711862e-05, 0.00234741822290798, 0, 0, -0.00315136937382877, -0.00177347069919571, -0.00461374624965819, 0.000818297634383299, 0.00592019612580759, -0.00373212905876955, 0.00815238470500247, -20833333.3480641, -20833333.3504144, -0.00250779304520003, 0.00235103414615648, -0.00692787889781436, 0.00684668645644227, -0.0121248060615096, -0.0192743487845822, 0.00679277855949697, -0.0287879573073593, -0.0134714613790835, -0.0151049072404427}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -3.01342821448392e-07, 0, 0, 0, 0, 0, 0, 0, 5.88636716551011e-06, 0, 2.61163460817219e-06, 500.000003375855, 0, 0, 0, 0, 0, 0, -5.70299631947576e-05, 0, -1.07972360545276e-05, 0, 0, -2.91296590740872e-05, -5500.0000172675, -10916.6667435705, -5500.00000388893, 0, 0, 0, 0, 0, 0, 0.000263181888646054, 0, 3.06758903553735e-05, 0, 0, 0, 0.000336168381269795, 0, -3.19825548966308e-05, 5000.00004415265, 83333.3340312634, 4999.99996298986, 119583.333773783, 59999.9998731202, 119583.33343546, 0, 0, 0, 0, 0, 0, -0.000664630725975382, 0, -8.76166081799527e-05, 0, 0, 0, -0.00145994126454129, 0, 0.000632346260152333, -2.24755623542514e-05, -281250.002792358, 8.56691716994134e-05, 0, 0, -0.000636141977512966, -104166.668054879, -49999.9993941059, -906250.0035549, -906250.002341966, -49999.9990571336, -104166.665615432, -1299999.99906065, 0, 0, 0, 0, 0, 0, 0.000986667732280145, 0, 0.000200794625623296, 0, 0, 0, 0.00296666344195404, 0, -0.00185008207392566, -0.000142904992669404, 416666.672494274, 0.000122752455786332, 0, 0, 0, 0.00356304768310226, 0, -0.000763804032459721, 0.00231557423039581, -0.00126146021349763, 3020833.34512535, 3020833.34271238, -0.00191285876417678, -0.00486206102929414, 729166.675918682, -0.0033790656616993, 729166.669312502, 1041666.66342092, 9791666.67011394, 1041666.65629166, 0, 0, 0, 0, 0, 0, -0.000869215370562936, 0, -0.000271842176638914, 0, 0, 0, -0.00307936239492886, 0, 0.00237565783120528, 0.000339255852538992, -208333.339999473, -0.000601351861351248, 0, 0, 0, -0.00644270812154769, 0, 0.00378851677695002, -0.00205992900677286, 0.00155589539965017, -4375000.01884452, -4375000.01562871, 0.00120678656691242, 0.00876718755806576, 0, 0, -0.00268065669951799, -0.0120022244979734, 0.00482604872714104, -2083333.35663858, -2083333.34847689, 0.00590429907563089, 0.00205567621029428, 0.00445752396665719, -32291666.6846545, 0.0228471777203337, -7291666.67124679, 0.0235635127045013, -7291666.65239723, 0, 0, 0, 0, 0, 0, 0.000424714965524331, 0, 0.000187852261334028, 0, 0, 0, 0.00158191640577046, 0, -0.00143490685290118, -0.000304581411516596, 0.00397712710949421, 0.000683260520564261, 0, 0, 0, 0.00485607115197058, 0, -0.00408075203158277, 0.000943847472529034, -0.00111608630416554, 2083333.34781972, 2083333.34532216, 0.000444677441193041, -0.00712976274787241, 0, 0, 0, 0.00519419177411445, 0, -0.00106467986688578, 0.00750498382357959, -0.00328158337159253, 2083333.35829677, 2083333.35241275, -0.00393348187178519, -0.0072748352688508, -0.00344184545360949, 45833333.3572757, -0.0209433636088004, 0.0203284599098718, -0.00593633207661887, 0.0110953170321247, 0.00484185495597872, -0.0207213749863813, 20833333.3545144, 20833333.3337941, -0.0335658017249204, -0.0268524866974292, -0.0275300984639484, 0, 0, 0, 0, 0, 0, -8.92814157659743e-05, 0, -5.13768349669418e-05, 0, 0, 0, -0.000319238148900201, 0, 0.000331680153473026, 0.00010070204382881, -0.000970952703220364, -0.000253305472416211, 0, 0, 0, -0.00129983078925819, 0, 0.00141498743517238, -0.00017444763805814, 0.000350806525934731, -0.0043066393323611, -0.00350204084576103, -0.000540306690366304, 0.00214751172104378, 0, 0, 0, -0.00235238937542118, 0, 0.0013781554137497, -0.00186118390091621, 0.000920546117113555, -0.00958577427127484, -0.00754163529794646, 0.00066979262945541, 0.00420216436313388, 0.00118133069566122, -20833333.3435715, 0.00698154450331679, 0, 0, -0.000936279823678384, -0.00636688187719414, 0.00198145868029586, -0.010940391179923, -0.00879933568544118, 0.00230040793563896, -0.00188784033655292, -0.00161045166954195, 0.00683787123965415, -20833333.3481879, -20833333.3394699, 0.0148304269763678, 0.013505157141082, -0.0102876386982073, 0.0147194101827227, 0.00586082247308364, 0.0118900183127955, 0.00626577877524472, 0.0235795876693441}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 7.94863220055618e-08, 0, 0, 0, 0, 0, 0, 0, -1.28055653954571e-06, 0, -2.46429014299549e-06, 499.999996951237, 0, 0, 0, 0, 0, 0, 6.09412516469488e-07, 0, 6.70545339066806e-06, 0, 0, 5.85789974171178e-05, -5499.99996163411, -5499.99994393026, -10916.6666462437, 0, 0, 0, 0, 0, 0, 6.59491597634108e-05, 0, 1.66473458635483e-05, 0, 0, 0, -0.00045045066740618, 0, -0.000216019769083479, 4999.99977277826, 4999.99964217517, 83333.3333242378, 59999.9994519259, 119583.333178133, 119583.332870061, 0, 0, 0, 0, 0, 0, -0.000383079906083777, 0, -9.8973916743002e-05, 0, 0, 0, 0.00150284183532999, 0, 7.66768463956625e-05, 0.000671743413437426, 0.000905574590479771, -281250.000224959, 0, 0, 0.00183965174732506, -49999.9975001013, -104166.665632655, -49999.9968994384, -104166.663254531, -906250.000004782, -906249.998983974, -1299999.9980986, 0, 0, 0, 0, 0, 0, 0.000945603341938461, 0, 0.000179327429385655, 0, 0, 0, -0.00232718637342229, 0, 0.000689311013377864, -0.00108765538333703, -0.000821848911731952, 416666.667325653, 0, 0, 0, -0.00628556686913889, 0, -0.00197070492433355, -0.00573762294036014, -0.00291181844051874, -0.00783375502512646, -0.00924225474541565, 3020833.33426839, 3020833.3338033, -0.0106969869274747, 729166.664961234, 729166.656654861, 1041666.65913139, 1041666.65419118, 9791666.6674448, 0, 0, 0, 0, 0, 0, -0.00119224644518713, 0, -0.000165277788557232, 0, 0, 0, 0.00151862100398132, 0, -0.0011861629975697, 0.000991586156679009, -0.000269255143129895, -208333.334098339, 0, 0, 0, 0.00963865391995319, 0, -0.000699806089014003, 0.00703196031466843, 0.00374860300785391, 0.00965793741477083, 0.0108727592018873, -4375000.00169203, -4375000.00378248, 0, 0, 0.0074820845064469, 0.0171087230568258, 0.00547572405998948, 0.0204751700175368, 0.0255257284928277, -2083333.33218622, -2083333.32337247, 0.015280643549483, 0.0270610052903256, -32291666.6770302, 0.030927310633473, -7291666.66638293, -7291666.65144293, 0, 0, 0, 0, 0, 0, 0.00075551403955573, 0, 8.20141326029063e-05, 0, 0, 0, -0.000127996963228165, 0, 0.00073851762521428, -0.000481460074568664, 0.000875874621783073, 0.00040255972580952, 0, 0, 0, -0.00663415072281445, 0, 0.00214051933999878, -0.00444399968157913, -0.00221175703525992, -0.0056649917853112, -0.00528674790994981, 2083333.3344381, 2083333.33731542, 0, 0, 0, -0.00942463652952322, 0, -0.00221398036992806, -0.0130003129943022, -0.00563245640143562, -0.0177626620143564, -0.0244587000395633, 2083333.33302201, 2083333.33212427, -0.0137944814529869, -0.0228160750176312, 45833333.3470553, -0.0211376201768036, -0.00503851271162148, -0.0235748296595477, -0.0356453902647828, -0.00686637995534983, -0.0441438783759442, -0.0325548715710921, 20833333.3472878, 20833333.3316779, -0.0194648355141454, 0, 0, 0, 0, 0, 0, -0.0001911428872981, 0, -1.80589033508109e-05, 0, 0, 0, -0.000172504655136125, 0, -0.000160023479034193, 9.76125240757616e-05, -0.000386152081212654, -8.00929012785363e-05, 0, 0, 0, 0.00163970011347868, 0, -0.000859451322256935, 0.00114397871934639, 0.000478782622787855, 0.00124031902841057, 0.000666008979561263, -0.000227484510002475, -0.00127922423916612, 0, 0, 0, 0.00379365829230598, 0, -0.000525759889096775, 0.00385159700267696, 0.00185433482277449, 0.00573005659687437, 0.00785777798283273, 8.90485704090353e-05, -0.00173347303631172, 0.00454379215846532, 0.00645397729755439, -20833333.3387627, 0, 0, 0.00236432087740201, 0.00763975120485959, 0.00307492224674125, 0.00912946004651425, 0.0132537044760349, 0.00185684372724873, 0.0059350595519528, 0.0144794541218534, 0.00504827815259972, 0.0194055155229124, 0.0160779242603105, -20833333.3426468, -20833333.337991, 0.02432525459168, -0.00328561866531574, 0.011667709959185, 0.0125149445604042, 0.0209335379364271, -0.00239712740954661}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 3.18264217756633e-07, 0, 0, 0, 0, 0, 0, 0, -3.31793525323245e-06, 0, -6.49644186071344e-06, 416.666666396523, 0, 0, 0, 0, 0, 0, 2.49596265452176e-05, 0, 5.84617573982829e-05, 0, 0, 4.03273076587264e-05, -8055.55560673944, -6666.66666092328, -4583.33330103307, 0, 0, 0, 0, 0, 0, -0.000110588186424809, 0, -0.000299605634545508, 0, 0, 0, -0.000224554639185262, 0, -0.000207349216094586, 49305.5558635249, 27083.3331807553, 4166.66619577204, 128472.222738369, 88194.4449075561, 72916.6665506989, 0, 0, 0, 0, 0, 0, 0.000265441213000658, 0, 0.000904688811533742, 0, 0, 0, 0.000952378973518052, 0, 0.000805314673309509, -111111.11189254, -20833.3324440182, 0.00233808912760929, 0, 0, 0.000474236183482939, -781250.00289022, -534722.22481757, -517361.112490663, -291666.666140343, -76388.8897809733, -62499.9976298173, -1399305.56028716, 0, 0, 0, 0, 0, 0, -0.000356761155639485, 0, -0.00161923086044774, 0, 0, 0, -0.00211674707981915, 0, -0.00216583840165658, 69444.4454914215, -0.00228461480703165, -0.00570993197612795, 0, 0, 0, -0.00185612932924739, 0, -0.000279485198266544, 1736111.11759654, 1180555.56078635, 381944.444947975, 208333.33137947, -0.000847990534036194, -0.0101621115093151, 3090277.78568558, 416666.673543055, 208333.327042391, 8437500.02150955, 5555555.57000569, 1145833.34228264, 0, 0, 0, 0, 0, 0, 0.000268148154258945, 0, 0.00167932719479315, 0, 0, 0, 0.00244132625337883, 0, 0.0033706704329292, -0.000767309294055962, 0.00302481335333469, 0.00748787126660136, 0, 0, 0, 0.00377783711954816, 0, 0.000166350903279695, -1041666.67389475, -694444.449463712, 0.00291735754149234, 0.00344553686199365, 0.00469548146873364, 0.0186432945195715, 0, 0, 0.000699055915310861, -6597222.23697674, -694444.455582857, -2083333.34115699, 0.0101390104046616, -0.00767287835684236, 0.0193519771320132, -18402777.8134922, -3819444.46366995, -0.000879371350780413, -32986111.1625943, -6250000.04363439, -3819444.46566751, 0, 0, 0, 0, 0, 0, -0.000103294580707697, 0, -0.000929936567273344, 0, 0, 0, -0.00142055196186574, 0, -0.0026533170556473, 0.00028564041977831, -0.00202714931998521, -0.00504847716076389, 0, 0, 0, -0.0032163461117375, 0, -0.000653498523409486, 0.00396500752230213, 0.00225756396216464, -0.00420320545605234, -0.00287931218370757, -0.00548373516678351, -0.0161268819311931, 0, 0, 0, -0.00212126282197869, 0, 0.00103301895247357, 3472222.23405173, 0.00705528429347404, 0.00120829227752802, -0.00808517337546164, 0.00321713605187943, -0.0222093709644392, 10416666.6927526, 0.0113270121611796, -0.0100505683717064, 3472222.23410727, 0.00971529724832356, -0.0191856785243316, 69444444.5035234, 10416666.7188055, 20833333.3819729, 0.0195254766402471, 0.0326001520456815, 0.00163965399105459, 20833333.4038157, 0, 0, 0, 0, 0, 0, 1.50983911457283e-05, 0, 0.000212468322308645, 0, 0, 0, 0.0003337307866219, 0, 0.000813905303662795, -4.06853328837306e-05, 0.000544166377811314, 0.00137169422321696, 0, 0, 0, 0.000976018268464225, 0, 0.000495129824116316, -0.000850256740485587, -0.000362816853844071, 0.00170297831180858, 0.000957851100463447, 0.00212634476312907, 0.00539846638039095, 0, 0, 0, 0.00118588127930779, 0, -0.000612262516133623, -0.00347230929484102, -0.00148126784782448, 0.00133032867228031, 0.00237026848431162, -5.87469774837887e-05, 0.00902276069705404, -0.00706611548724867, -0.00220026538272168, 0.00641421451292138, 0, 0, 2.25392056756191e-05, -0.00530661332669595, -0.00313871753258614, -0.00248775010575808, 0.00736665808530964, -0.00299351321557942, 0.0102211591991915, -34722222.2445713, -0.0195640162253349, -0.0156579434510918, -0.0065541414107644, -0.0072622703207214, 0.00709180436543141, -34722222.2509511, -0.0224931758722475, 0.000357039173976563, -34722222.2658265, -0.0343751262252351, -0.0256584039952499}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 8.98812849685555e-08, 0, 0, 0, 0, 0, 0, 0, -1.14148949470234e-06, 0, -3.30826890635207e-06, 416.666666909889, 0, 0, 0, 0, 0, 0, 2.19191591983503e-07, 0, 2.78501947174929e-05, 0, 0, 5.50334538355917e-05, -6666.66669381821, -8055.55555885756, -4583.33329689878, 0, 0, 0, 0, 0, 0, 3.6262285718656e-05, 0, -0.000110434677251523, 0, 0, 0, -0.000280745737978138, 0, -0.000430235872394981, 27083.3335022014, 49305.5556514087, 4166.66633854394, 128472.22260476, 72916.6665960728, 88194.4438977952, 0, 0, 0, 0, 0, 0, -0.00015636959721153, 0, 0.00024393642556015, 0, 0, 0, 0.000563970202649365, 0, 0.00148562229685526, -20833.3338386469, -111111.111546987, 0.00103378794531543, 0, 0, 0.00209093968626978, -517361.113007933, -291666.666740839, -781250.002338444, -534722.220743789, -62499.9987277329, -76388.8841400044, -1399305.55492415, 0, 0, 0, 0, 0, 0, 0.000300055428731541, 0, -0.000326539160815089, 0, 0, 0, -0.000477864152812937, 0, -0.00261573693952336, 0.00088694965229833, 69444.4451523059, -0.00150019692789894, 0, 0, 0, -0.00401620424060827, 0, -0.00643327673168526, 381944.44848502, 208333.334075829, 1736111.11761878, 1180555.55494401, -0.00340403765892457, -0.0144598899491567, 3090277.78635912, 208333.330234231, 416666.6522608, 5555555.55744853, 8437500.00265627, 1145833.32005708, 0, 0, 0, 0, 0, 0, -0.000310523897067524, 0, 0.000274048177038949, 0, 0, 0, 0.000128281823226729, 0, 0.0024143492190914, -0.000901587375536333, -0.000395844635337973, 0.000980657939973139, 0, 0, 0, 0.00342307510650945, 0, 0.00944117106811455, -0.00471007119889267, -0.0017300564063293, -1041666.67504614, -694444.446292409, 0.00339203617427278, 0.020524713946421, 0, 0, 0.0103406601991748, -2083333.34568846, 0.00406892931683992, -6597222.24052086, -694444.428191128, 0.00714702042031981, 0.0372310782188948, -3819444.45040292, -18402777.7917587, 0.0338688978992729, -32986111.1251361, -3819444.42675695, -6249999.97486948, 0, 0, 0, 0, 0, 0, 0.000171610571651146, 0, -0.000137412202936544, 0, 0, 0, 1.89196751594726e-05, 0, -0.00109629339754287, 0.000473886143726377, -4.82817172725443e-05, -0.000184184758609421, 0, 0, 0, -0.00122724828390489, 0, -0.00650209372800014, 0.0030139822616335, 0.00167622821701932, 0.00475723544631312, 0.00205215476331139, -0.00101060866339342, -0.0137013192443445, 0, 0, 0, -0.00671329467008925, 0, -0.0110725545080216, 0.00805990622985578, -0.00224415234967021, 3472222.2391212, -0.00677879370347664, -0.00521374022372839, -0.0387012492782919, 0.00675691058167364, 10416666.6841029, -0.0329278729595172, 3472222.24008918, -0.00758787542369464, -0.0331080881832047, 20833333.3489239, -0.0110108783956664, 69444444.4734752, 10416666.6543927, -0.0343822554686067, -0.0526226939414013, 20833333.3152412, 0, 0, 0, 0, 0, 0, -4.02040926353517e-05, 0, 3.1764344356718e-05, 0, 0, 0, -4.88827360974119e-06, 0, 0.000189087421507334, -9.59746892276372e-05, 7.87907725464535e-05, -3.79781685226038e-05, 0, 0, 0, 0.000134877247079554, 0, 0.0016805775646069, -0.000819972477272832, -0.000571374052140589, -0.000899293492207013, -0.00055462591376462, -0.000147607462060722, 0.00343858565481744, 0, 0, 0, 0.00144709462325738, 0, 0.00413502728111247, -0.00207341582528444, 0.000297603102781942, -0.00540386413864499, 0.00054008315266831, 0.000991816026932024, 0.0141615203349756, -0.00286697621015904, -0.00660730530915456, 0.0111135509283624, 0, 0, 0.00395929528374653, -0.00570453891699573, 0.0028598078383973, -0.00873912924404577, 0.00932939864880307, 0.00316179852880649, 0.0187332089363707, -0.00574572984532753, 0.00192217676007211, -34722222.2388259, -0.000863592499470934, 0.0170703585521883, 0.0284620676056053, -34722222.2366466, 0.0130645106306824, 0.0210246571384401, 0.00501772978770472, -34722222.2217439, 0.0243983438064673}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -8.66830013321811e-07, 0, 0, 0, 0, 0, 0, 0, 1.38117716572876e-05, 0, 1.36200029246709e-05, 416.6666773917, 0, 0, 0, 0, 0, 0, -0.000109165220068711, 0, -0.000100240361749816, 0, 0, -0.000152685634130944, -4583.33340251907, -8055.55569146763, -6666.66682335581, 0, 0, 0, 0, 0, 0, 0.000459384663525951, 0, 0.000411798644937779, 0, 0, 0, 0.000973554405015042, 0, 0.000803329700957363, 4166.6669500405, 49305.5564028307, 27083.334344296, 88194.4451800317, 72916.667388205, 128472.223676303, 0, 0, 0, 0, 0, 0, -0.00113910692226874, 0, -0.00103809533989034, 0, 0, 0, -0.00316534574887454, 0, -0.00219860906287214, -0.00074805195670008, -111111.113820103, -20833.3368528078, 0, 0, -0.00429727596159427, -76388.8912805494, -62500.0020913034, -534722.226025915, -781250.007533397, -291666.670418837, -517361.117775104, -1399305.56050401, 0, 0, 0, 0, 0, 0, 0.00174414112153885, 0, 0.00165407855101052, 0, 0, 0, 0.00559979964822321, 0, 0.00337425387143196, 0.00122995720596161, 69444.4492166777, 0.00716096871722355, 0, 0, 0, 0.011409783642397, 0, 0.00946216898065322, 0.00466442870682783, 0.00422917113630337, 1180555.56550601, 1736111.13057526, 208333.343832826, 381944.459677253, 416666.676529405, 208333.341218746, 3090277.80535411, 1145833.34281001, 8437500.0194437, 5555555.57315103, 0, 0, 0, 0, 0, 0, -0.00163048557493229, 0, -0.0016222183046211, 0, 0, 0, -0.00557382086662791, 0, -0.00301462357363447, -0.00120514453927245, -0.00473793659003253, -0.00854339936685683, 0, 0, 0, -0.0150435257482884, 0, -0.0105579636060815, -0.0053126628741953, -0.0053374947177549, -694444.457796591, -1041666.69224592, -0.0162068832157672, -0.0184515872845675, 0, 0, -0.0198570944007963, -0.0140635058928577, -0.0109512786884654, -694444.464435582, -6597222.27695484, -0.015309772629148, -2083333.37983856, -0.0118731081492105, -18402777.8171566, -3819444.47664122, -6250000.02458663, -3819444.4682126, -32986111.1593489, 0, 0, 0, 0, 0, 0, 0.000854353819720024, 0, 0.000889345071322542, 0, 0, 0, 0.00297584090143788, 0, 0.0015257501775244, 0.000641923820514246, 0.00251596848158767, 0.00552184605429954, 0, 0, 0, 0.0096241493724658, 0, 0.00566826176479968, 0.00325502816709706, 0.0036236985007262, 0.00871919250401883, 0.0164707816928191, 0.0130078605544619, 0.0115954166619342, 0, 0, 0, 0.0181102646470729, 0, 0.0157417030837181, 0.0102909361858272, 0.00880836259330909, 0.0186172964297888, 3472222.27103013, 0.0149063070036421, 0.0356447238863462, 0.00876581748541244, 10416666.7039397, 0.0284374665726345, 0.0191479888630432, 0.0125477956756449, 3472222.28575129, 0.0168378769475153, 0.0193119066534199, 10416666.7011542, 69444444.5076899, 0.0297253005494618, 20833333.3892871, 20833333.3672244, 0, 0, 0, 0, 0, 0, -0.000192064054058404, 0, -0.000207429371073367, 0, 0, 0, -0.0006720095660896, 0, -0.000351966850900033, -0.000142718350220874, -0.000564425311056334, -0.00148565085867031, 0, 0, 0, -0.0023988507640565, 0, -0.0011492449599071, -0.000827857345761263, -0.00100192139193369, -0.00218678687896507, -0.00412942305694393, -0.00421828781075132, -0.00305790237457162, 0, 0, 0, -0.00591893248215091, 0, -0.00454354690455953, -0.00303367286817929, -0.00292859979886726, -0.00631381976167565, -0.0159062581351106, -0.00580288392042961, -0.0100150434548933, -0.00275900390909749, -0.0129644597044234, -0.00977813542232286, 0, 0, -0.00753887447077284, -0.00693055516702693, -0.00460665093919121, -0.00908025926264867, -0.0290414090341245, -0.00540100874997401, -0.025248694450307, -0.00512739448611688, -0.0071928343283078, -0.0178958139835255, -34722222.2518988, -0.0135053577201552, -0.0231682472010808, -0.0116709259357381, -0.011870586522439, -34722222.2582458, -0.00973206473252794, -34722222.2430787, -0.0203022374546502}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -3.30955701702478e-07, 0, 0, 0, 0, 0, 0, 0, 4.91044079645986e-06, 0, 5.36175880984065e-06, 416.666673482207, 0, 0, 0, 0, 0, 0, -4.24936730424681e-05, 0, -4.72773276240992e-05, 0, 0, -4.28771070484574e-05, -4583.33338538174, -6666.66675603836, -8055.55565253385, 0, 0, 0, 0, 0, 0, 0.000208920487462964, 0, 0.000229491673392958, 0, 0, 0, 0.00023589982814316, 0, 0.000275689103340051, 4166.66685155032, 27083.3338859508, 49305.5561565705, 72916.6672145614, 88194.445128669, 128472.223133578, 0, 0, 0, 0, 0, 0, -0.000620029416155658, 0, -0.000640070082674995, 0, 0, 0, -0.00077119694249282, 0, -0.00108691994589642, -0.000391311266620349, -20833.3352960724, -111111.113162688, 0, 0, -0.00105438053093292, -62500.0013606756, -76388.8910887893, -291666.669177511, -517361.11502537, -534722.225582161, -781250.004204091, -1399305.56126301, 0, 0, 0, 0, 0, 0, 0.00111635352620976, 0, 0.00107376757072278, 0, 0, 0, 0.00167289599537627, 0, 0.00225985776536453, 0.000547580085146179, 0.00411806111170764, 69444.4484941576, 0, 0, 0, 0.00189410746687972, 0, 0.00355001516512331, 0.00168100578049291, 0.00397879196272761, 208333.339967423, 381944.454162691, 1180555.56381283, 1736111.12199547, 208333.337611789, 416666.675237783, 3090277.79013811, 1145833.34709748, 5555555.57367471, 8437500.02237034, 0, 0, 0, 0, 0, 0, -0.0011798062539974, 0, -0.00108114051082205, 0, 0, 0, -0.00232472518297825, 0, -0.00249432251192228, -0.000489230909558205, -0.00496614816534668, -0.00463083916353391, 0, 0, 0, -0.00176833547169747, 0, -0.00595181876220746, -0.00115150196826196, -0.00441301687513842, -0.00990385116598916, -0.0143916919975084, -694444.455358458, -1041666.68194165, 0, 0, -0.00498622944483444, -0.0028816135783637, -0.0111994210166522, -0.0081100227779689, -2083333.35166404, -694444.459400844, -6597222.24570938, -0.0159192669372702, -3819444.4762325, -18402777.819121, -3819444.47283276, -6250000.04540944, -32986111.159212, 0, 0, 0, 0, 0, 0, 0.000669629463475537, 0, 0.000603301951584721, 0, 0, 0, 0.00176698204516602, 0, 0.00142084177257149, 0.000240317612804015, 0.00316831565001344, 0.0028696353067678, 0, 0, 0, 0.00110486403732971, 0, 0.00448360575858152, 0.000453149421391929, 0.00284373035133339, 0.00752046882537994, 0.0114465677565063, 0.00748941555086529, 0.010717886965678, 0, 0, 0, 0.0026427830450555, 0, 0.00638536862129009, 0.000542022338141552, 0.00763699342638186, 0.00824363047470771, 0.0146838044126659, 0.0122015915164387, 3472222.24512951, 0.00927615660086027, 0.0302487667272806, 10416666.7025768, 0.00333117687959402, 0.0122819092158929, 3472222.2430347, 0.0171841336123621, 0.0393718392675209, 0.0299081023074484, 20833333.3804194, 10416666.7239675, 69444444.5023976, 20833333.3942845, 0, 0, 0, 0, 0, 0, -0.000157152756429046, 0, -0.000143104607316415, 0, 0, 0, -0.000542657612675828, 0, -0.000337896526313914, -4.66540523523528e-05, -0.0008278573252555, -0.000748463522876927, 0, 0, 0, -0.000410584682608453, 0, -0.00122928750011818, -8.18527119278415e-05, -0.000798879051392436, -0.00224394627273046, -0.00368695826185535, -0.00211491321165263, -0.00296440807218227, 0, 0, 0, -0.00040101019785808, 0, -0.00268054690236292, 6.69442447260475e-05, -0.00223981225250225, -0.00320230350515024, -0.00519155492698288, -0.003840208170718, -0.00834930441438493, -0.00231472580291019, -0.0115459462108336, -0.0118006917876397, 0, 0, -0.00202198772346014, -0.000157425499752478, -0.00382779495031546, -0.00188809084943272, -0.00660651521089791, -0.00466625827996302, -0.0104908724148109, -0.00331283421507361, -0.0124837619886447, -0.0139879580426903, -0.0192170690184402, -0.0244135082100983, -34722222.2479625, -0.00716118199238394, -0.0257414390474808, -34722222.24772, -0.0242446484112451, -0.0272291918923992, -34722222.2615279}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 1.93388240029305e-07, 0, 0, 0, 0, 0, 0, 0, 1.09617923415519e-06, 0, -6.27416897296104e-06, 416.666669306903, 0, 0, 0, 0, 0, 0, -2.44638302029945e-05, 0, 5.87370576035554e-05, 0, 0, 1.56563658259487e-05, -6666.66669859427, -4583.33337541516, -8055.55558787943, 0, 0, 0, 0, 0, 0, 0.000151242017299436, 0, -0.00030433261704253, 0, 0, 0, 5.8787858442509e-05, 0, -9.1062186924118e-05, 27083.3334572977, 4166.66691842616, 49305.5556054377, 72916.6668972954, 128472.222754423, 88194.4450373678, 0, 0, 0, 0, 0, 0, -0.000512840264837035, 0, 0.000929146557009797, 0, 0, 0, -0.000298679214772017, 0, 0.00048302076046943, -20833.3335810816, -0.000817276164220893, -111111.110628104, 0, 0, -0.000720079640401206, -291666.667039404, -517361.113256442, -62500.0006198455, -76388.8919215965, -781250.002165187, -534722.224829982, -1399305.56000192, 0, 0, 0, 0, 0, 0, 0.00101350469242636, 0, -0.00166618454793352, 0, 0, 0, 0.000775910117022155, 0, -0.00182515324428626, 0.0002800793579882, 0.00175038764470234, 69444.4423109076, 0, 0, 0, 0.00205981869387918, 0, 0.00281534090056725, 208333.333076636, 381944.448907771, 0.000516094782111198, 0.00655089635485486, 1736111.11446174, 1180555.55932716, 208333.333194292, 3090277.78551951, 416666.680438242, 5555555.5668299, 1145833.34736433, 8437500.01762311, 0, 0, 0, 0, 0, 0, -0.00114688855135191, 0, 0.00171279846156395, 0, 0, 0, -0.00130200474867946, 0, 0.00338992918237227, -0.000174158912171652, -0.0024195583069299, 0.00353302453794799, 0, 0, 0, -0.00226282241361049, 0, -0.00371339585964127, 0.00139222860831491, -0.00529202221648023, 0.000206039508977361, -0.00739141876175898, -1041666.66858155, -694444.444324346, 0, 0, -0.00696404889077515, 0.00256858830146628, -2083333.34609636, 0.0023586686111913, -0.0247594571140099, -6597222.23283602, -694444.468853727, -3819444.45760555, -0.0194824740024293, -18402777.8023784, -3819444.46604439, -32986111.1496567, -6250000.04988339, 0, 0, 0, 0, 0, 0, 0.000686264966457345, 0, -0.000931451762093171, 0, 0, 0, 0.00115094419227482, 0, -0.00289570453362706, 4.85532100396838e-05, 0.00185140334786984, -0.00267649449445833, 0, 0, 0, 0.0013495825031204, 0, 0.001518252361326, -0.00142726656338205, 0.00338506397281005, -0.000308996907275549, 0.00472857962219622, -0.000126813965666535, -0.00418959289577977, 0, 0, 0, 0.00550239313985707, 0, 0.00788679255274656, -0.00349107722175678, 0.0104694413712704, -0.00356210872552977, 0.0186123121640755, 3472222.2285525, 0.0173936358838019, 0.00700406639173968, 0.0116978105468037, 10416666.6795592, -0.00421275380320703, 3472222.23429966, 0.0340317127599473, 0.0121441613251951, 20833333.3695899, 0.0162147401927195, 0.0552026844667883, 69444444.4840378, 10416666.7218084, 20833333.3943966, 0, 0, 0, 0, 0, 0, -0.000168102666207901, 0, 0.000207373478127023, 0, 0, 0, -0.000394592192669596, 0, 0.000921686047221378, -1.56292875297432e-06, -0.000577461028597879, 0.0007730247335118, 0, 0, 0, -0.000408470652695927, 0, 7.80580629135456e-05, 0.000476714944426956, -0.000902419089611237, -2.86900517853416e-06, -0.00144096010160336, 0.000359014571154034, 0.0023712682665962, 0, 0, 0, -0.0015115483560505, 0, -0.00277219902372499, 0.00144604713326065, -0.00342556598893614, 0.00154801396495931, -0.00513975863154818, -0.00142046027053541, -0.00375592947969046, -0.00132112660411326, -0.00248141686484109, -0.00151756368206493, 0, 0, -0.00346550651252459, 0.00205868657357745, -0.00503338971800685, 0.00223269930956496, -0.0138970725559699, -0.00371983838091482, -0.0147590842711752, -0.00129737277329074, -0.0124593694338237, -0.00328905788414123, -0.0202971767841394, -34722222.2346298, -0.0187369344234163, -0.00261915505818684, -34722222.2425353, -0.0345821459174536, -0.024527279031257, -0.0305112534563333, -34722222.257254}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 6.49121106688257e-08, 0, 0, 0, 0, 0, 0, 0, 1.82997607583683e-07, 0, -1.65964945285801e-06, 416.666670560154, 0, 0, 0, 0, 0, 0, -3.4916587460622e-06, 0, 1.67533708646423e-05, 0, 0, -5.20749009326688e-06, -8055.55561455548, -4583.33338756196, -6666.66670095778, 0, 0, 0, 0, 0, 0, 1.5495466591609e-05, 0, -0.000103376708132723, 0, 0, 0, 6.93999294516272e-05, 0, 6.99919189956056e-05, 49305.5558391475, 4166.66698974926, 27083.3333665552, 88194.4450157276, 128472.222892528, 72916.6671175194, 0, 0, 0, 0, 0, 0, -5.10890991305857e-05, 0, 0.000355841594406468, 0, 0, 0, -0.000126014620746186, 0, -0.000111205580474711, -111111.111797001, -0.00102069883930457, -20833.3329029131, 0, 0, -0.000874453876676363, -534722.224425581, -781250.002950054, -76388.8911917134, -62500.0025145629, -517361.113653695, -291666.667549158, -1399305.56079162, 0, 0, 0, 0, 0, 0, 0.000117382855993861, 0, -0.000682128716585205, 0, 0, 0, -4.97417832838091e-06, 0, -0.000372937890032361, 69444.4453809364, 0.00191458367176617, -0.0016341395015024, 0, 0, 0, 0.00214836987666291, 0, 0.00276794203892546, 1180555.55984101, 1736111.11709429, 0.00473675412528275, 0.00649432574207885, 381944.448692538, 208333.332073659, 416666.673389917, 3090277.78806084, 208333.339767868, 8437500.01731394, 1145833.35089557, 5555555.57061266, 0, 0, 0, 0, 0, 0, -0.000157583549804786, 0, 0.000728080408289835, 0, 0, 0, 0.000196280211190803, 0, 0.00122328489858051, -0.000734991726787333, -0.00215690092904053, 0.0024631179788144, 0, 0, 0, -0.00220250626045219, 0, -0.0033431386770097, -694444.448935716, -1041666.67297945, -0.00537518283808654, -0.00871051409510383, -0.00333008266765668, 0.00611295713787473, 0, 0, -0.00615686612239241, -694444.454494932, -6597222.2391093, -0.00934489797458135, -0.0146196870437736, -2083333.348719, -0.00537402094411944, -18402777.8030166, -0.028927764207125, -3819444.46112658, -6250000.039892, -32986111.1546487, -3819444.48367558, 0, 0, 0, 0, 0, 0, 0.000109239785383167, 0, -0.000405685768114225, 0, 0, 0, -0.000169753799023963, 0, -0.00120854292319563, 0.000312936825489159, 0.00134236792092779, -0.00173534963612608, 0, 0, 0, 0.000961677119357032, 0, 0.00147559726816839, 0.0024118172226863, 0.00336107254057558, 0.00329509041484581, 0.00600207629834464, 0.00105474873984244, -0.00701753186759554, 0, 0, 0, 0.0050982140797533, 0, 0.00606272098807423, 0.00757065040010228, 3472222.23487796, 0.00597276034351009, 0.0136619833587647, 0.010556931601169, -0.00108233179015429, 10416666.6836335, 0.0231295760704224, 0.00605235548507039, 0.00923131850103223, 3472222.24015997, 0.012870640879583, 10416666.7039332, 69444444.4908542, 0.0401781037543076, 0.0453825314961556, 20833333.3741021, 0.0310373914253337, 20833333.3996598, 0, 0, 0, 0, 0, 0, -3.01943521254108e-05, 0, 9.21094409248102e-05, 0, 0, 0, 4.17854245068713e-05, 0, 0.0004044381733505, -5.71171570514773e-05, -0.00035233283156883, 0.000473383622924627, 0, 0, 0, -0.000123950580012287, 0, -8.74753056149154e-05, -0.000520020644778758, -0.000716507709602075, -0.000867774427688178, -0.00169395460611214, -3.87697177585295e-05, 0.00265577424983721, 0, 0, 0, -0.001463537807195, 0, -0.00191602526075226, -0.00225564899904299, -0.00357110560904286, -0.00144222439498337, -0.00464229189294971, -0.00277831945398537, 0.00232260789622638, -0.00426172447099126, -0.00724221907277419, 0.000320486842026472, 0, 0, -0.00278381171909726, -0.003670854072004, -0.00702902713162825, -0.00277360621445456, -0.00688123289197822, -0.0064736614114784, -0.00337222600438387, -0.0123983574990409, -34722222.2390636, -0.0147289382471624, -0.0188314949483185, -0.0126074916536529, -0.00689786860781975, -0.0179358233052614, -34722222.246174, -0.0196327402143789, -34722222.2556008, -0.0342987922578155, -0.0305681091687989}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.02939756382e-07, 0, 0, 370.370372387995, 0, 0, 0, 0, 0, 0, 1.65778875087628e-06, 0, -3.69572757401322e-06, 0, 0, 1.08052770618889e-06, -5925.92597208279, -5925.92594839157, -5925.92594082534, 0, 0, 0, 0, 0, 0, -4.25074671691637e-06, 0, 9.67944064075612e-06, 0, 0, 0, -6.32124971406137e-05, 0, 0.000104116296662562, 24074.0743189024, 24074.0741427766, 24074.074034151, 94444.4449387834, 94444.4448707095, 94444.4446716257, 0, 0, 0, 0, 0, 0, -3.15713436036114e-05, 0, 2.34200815829763e-05, 0, 0, 0, 0.000499540938212745, 0, -0.000483732466853796, -18518.519154025, -18518.5184760171, -18518.5179883677, 0, 0, -0.000493104870943114, -379629.631792625, -379629.631602691, -379629.631621475, -379629.630636655, -379629.630819933, -379629.62997797, -1500000.00412712, 0, 0, 0, 0, 0, 0, 0.000159721662879496, 0, -0.00011949486793513, 0, 0, 0, -0.00136391696287767, 0, 0.000726790530503282, 0.00097439660776957, -0.00049708103893881, -0.00161336602626373, 0, 0, 0, 0.000143685559379706, 0, 0.002808526837298, 277777.781933735, 277777.781932182, 277777.781322381, 277777.779052796, 277777.778638508, 277777.776352518, 1481481.48911971, 1481481.48682331, 1481481.48466209, 5972222.23645234, 5972222.23670299, 5972222.23289946, 0, 0, 0, 0, 0, 0, -0.00027719230350578, 0, 0.000174399607417094, 0, 0, 0, 0.00169444433126999, 0, -0.000285522680700438, -0.000911328471635306, 0.000834652830354356, 0.00225761938279184, 0, 0, 0, 0.00160024038137636, 0, -0.00463180053747625, -0.0042034742479386, -0.00479316028927126, -0.00264360250958609, 0.000233302123177913, 0.00109853707774556, 0.00523568772317349, 0, 0, -0.00423073595778979, -925925.937443156, -925925.934945494, -925925.938024448, -925925.930830154, -925925.931492553, -925925.929490484, -4166666.68659386, -4166666.68903404, -4166666.67612026, -23148148.1855735, -23148148.1800575, -23148148.1788431, 0, 0, 0, 0, 0, 0, 0.000214168515374044, 0, -0.000110269107270518, 0, 0, 0, -0.000975298823579475, 0, -0.000198980128714996, 0.000478933063088195, -0.000562220488110408, -0.00151818801462046, 0, 0, 0, -0.00218140124123808, 0, 0.00302910716570686, 0.00227310031359003, 0.00299010452783309, 0.000389364720086716, -0.00135113834980994, -0.00188400488326845, -0.00595038513343319, 0, 0, 0, 0.00185227700589026, 0, 0.00598896188265729, 0.00785699645952787, 0.007151926206627, 0.00828237592847735, 0.00186240669324576, 0.00180376033632511, 0.000298631737259987, 0.0127554645822192, 0.0153535484164819, 7.66547021856972e-05, 0.0132968769286892, 0.00701446471773406, 0.00675126915069518, 13888888.9233742, 13888888.9214622, 13888888.9297835, 13888888.9221619, 13888888.9152763, 13888888.9125534, 87962963.0191176, 0, 0, 0, 0, 0, 0, -6.24284291987676e-05, 0, 2.59374948645509e-05, 0, 0, 0, 0.000206730626487681, 0, 0.000136256975305053, -0.000107188944318818, 0.000133292708025782, 0.000396451753930463, 0, 0, 0, 0.000826925906983755, 0, -0.000651120970760806, -0.000533279294902338, -0.000788541254640569, 0.000251917058627445, 0.000633069746408948, 0.000733834020184789, 0.00228279523628523, 0, 0, 0, 6.30883812210462e-05, 0, -0.00235507724166521, -0.00207231996729204, -0.00225351453054507, -0.00187848868759181, 0.000292990322217986, 0.000144419980171124, 0.00115216111337151, -0.00318762878365104, -0.00371532967171755, 0.00241530230490722, 0, 0, -0.00208832511568328, -0.0048237741009195, -0.00290394664537901, -0.00511188001312524, -0.00233297927953389, -0.0015481540256616, -0.00288872727694235, -0.0105527593162828, -0.0111946776534964, -0.0154732671595604, -0.012060961796413, -0.00632184034217954, -0.00489320200748173, -0.0196044825866312, -0.0152158062305283, -0.014343597892421, -46296296.3237007, -46296296.3266126, -46296296.3209432}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -1.67584480833857e-05, 0, 0, 0, 0, 0, 0, 0, 0.000278184412060581, 0, 0.000237445630294264, 63000.0003483766, 0, 0, 0, 0, 0, 0, -0.00238255869087151, 0, -0.00188378898664357, 0, 0, -0.00233697393043236, -406750.002578982, -1036750.00549535, -1036750.00419228, 0, 0, 0, 0, 0, 0, 0.0112516113023595, 0, 0.00825735412030238, 0, 0, 0, 0.013843730653936, 0, 0.0136589456486069, 1041666.67620483, 5109166.70188209, 5109166.68911036, 6150833.36931507, 6150833.35766226, 16518333.3854786, 0, 0, 0, 0, 0, 0, -0.0319730895684423, 0, -0.0211254455889962, 0, 0, 0, -0.0441144191561806, 0, -0.045842592153525, -1322916.68781687, -11739583.4506915, -11739583.4002854, 0, 0, -0.062914979877859, -14385416.7768642, -14385416.7375659, -24802083.5217539, -75893750.2665156, -24802083.4308305, -75893750.2163379, -90279166.936831, 0, 0, 0, 0, 0, 0, 0.0555070407309814, 0, 0.0331521968150562, 0, 0, 0, 0.0868891319185738, 0, 0.0809865211359297, 833333.36268637, 14062500.2240345, 14062500.119125, 0, 0, 0, 0.133377780013154, 0, 0.181284644458952, 16562500.1875487, 16562500.1269387, 43020833.8224293, 160416667.33912, 43020833.5415036, 160416667.139722, 44687500.4604165, 44687500.2089562, 292708334.202996, 193541667.279931, 337395834.427037, 337395834.191427, 0, 0, 0, 0, 0, 0, -0.0571801593540051, 0, -0.0319099352667896, 0, 0, 0, -0.106983261336983, 0, -0.0747650332677461, -208333.357736154, -8541666.91398721, -8541666.79352679, 0, 0, 0, -0.147195032309645, 0, -0.263437398120582, -9375000.18736776, -9375000.1424003, -34375000.6724519, -175000000.910152, -34375000.2559541, -175000000.554295, 0, 0, -0.297532442165068, -35416667.2512326, -35416666.9372956, -52083334.230285, -482291668.242236, -52083333.619741, -482291668.11854, -203125000.714949, -534375002.080935, -534375001.344135, -553125001.817911, -553125001.460646, -1000000002.52857, 0, 0, 0, 0, 0, 0, 0.0319759014102778, 0, 0.0173860066614974, 0, 0, 0, 0.0735779945318277, 0, 0.0341695458601955, 0.0106451260191231, 2083333.48045584, 2083333.40862217, 0, 0, 0, 0.0902222593968238, 0, 0.176657513651753, 2083333.43782199, 2083333.423697, 10416667.1338076, 95833333.9685567, 10416666.8407267, 95833333.6585651, 0, 0, 0, 0.206329219790675, 0, 0.315778432647214, 10416667.0521059, 10416666.8711382, 20833334.1534439, 364583334.674057, 20833333.5273924, 364583334.510709, 104166667.091245, 385416668.561582, 385416667.690496, 20833334.0942257, 20833333.5477074, 541666668.404688, 395833334.687337, 395833334.50338, 583333335.632216, 1104166669.72878, 583333334.826647, 1104166669.30846, 1125000002.76763, 0, 0, 0, 0, 0, 0, -0.0074600393898216, 0, -0.00409725998725332, 0, 0, 0, -0.0211331901947324, 0, -0.00618198016740583, -0.00173489955581546, -0.036565821642187, -0.0192353704022299, 0, 0, 0, -0.0255973502787098, 0, -0.0431281794130475, -0.025064309780935, -0.0242115039635847, -0.12911483554346, -20833333.5129748, -0.0507959711583309, -20833333.4079768, 0, 0, 0, -0.0524377044376139, 0, -0.117097301449946, -0.106483991919587, -0.0670485714918695, -0.282139788523155, -104166667.10692, -0.0566856791160927, -104166667.027146, -20833333.4377986, -104166667.327084, -104166666.969932, 0, 0, -0.116567560607786, -0.249732107371989, -0.0767345493068095, -0.348643651271325, -208333334.047112, -0.053046858618108, -208333334.053117, -104166667.042494, -104166667.02695, -208333334.389917, -416666668.005081, -208333333.878472, -416666667.690411, -208333334.184397, -208333333.935117, -625000001.557097, -416666667.717062, -625000001.633955, -625000001.426064}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 1.20847980239072e-06, 0, 0, 0, 0, 0, 0, 0, -2.66479170005625e-05, 0, -1.13845014620188e-05, -4500.00002226666, 0, 0, 0, 0, 0, 0, 0.000257650384330654, 0, 4.21918332033516e-05, 0, 0, 0.000199737596674589, 9500.00010330172, 103250.000478701, 54500.0001846537, 0, 0, 0, 0, 0, 0, -0.00118172266930552, 0, -8.00083011185211e-05, 0, 0, 0, -0.00202501576337425, 0, -0.000350880121069103, -5000.00023543137, -859166.67033662, -100000.000878413, -207916.669247491, -105000.000202172, -1240416.67002925, 0, 0, 0, 0, 0, 0, 0.00300291941357478, 0, 0.000137559490199538, 0, 0, 0, 0.0082937134499167, 0, -0.00117412442041006, 0.000201794658246223, 3364583.34685466, 50000.0026001717, 0, 0, 0.00594098486748715, 104166.674090633, 49999.9989995777, 1593750.01743779, 10185416.6888924, 99999.9993183683, 2183333.34542732, 2287500.0104632, 0, 0, 0, 0, 0, 0, -0.00450314204137939, 0, -0.000341104200937147, 0, 0, 0, -0.0171227462339174, 0, 0.00575693603188721, 0.00035936399355983, -6562500.02735737, -0.00497196813283611, 0, 0, 0, -0.0229663485513318, 0, -0.00615568836839698, -0.0126594692866867, 0.00297928868796512, -5416666.71909102, -39062500.0637117, 0.00268355098899693, -1041666.68929776, -729166.708213787, 0.00948599029264202, -16666666.7345992, -1041666.68259429, -17395833.4009518, -2083333.35335768, 0, 0, 0, 0, 0, 0, 0.003977806742863, 0, 0.000598942431294107, 0, 0, 0, 0.0190998124666734, 0, -0.0092021418257661, -0.00112607083437007, 6041666.69781722, 0.00608496190236553, 0, 0, 0, 0.0376312023084945, 0, -0.00257952761546553, 0.0129552810517205, -0.00415165558140671, 8125000.08031622, 73750000.0924503, -0.00218470100255292, 0.020397520615189, 0, 0, 0.0283260460014772, 0.0532606514051442, -0.0146332194157594, 2083333.42726109, 56250000.1451745, -0.0228033580517493, 7291666.77251749, 0.0136545999205986, 58333333.4874931, 0.0199329635902421, 7291666.77191951, -0.00305869929517783, 14583333.4612114, 0, 0, 0, 0, 0, 0, -0.00191837793173006, 0, -0.000510936446068744, 0, 0, 0, -0.0110151229466951, 0, 0.0065491319844684, 0.00108939351780834, -2083333.35204854, -0.00427227403374464, 0, 0, 0, -0.0285535052109387, 0, 0.00869834643138253, -0.00741640621109811, 0.0034186571582415, -4166666.72787947, -64583333.400517, -0.00101282771069799, -0.00728867111889188, 0, 0, 0, -0.0348259648677831, 0, -0.0134034216096818, -0.0358585868614423, 0.00871052688635587, -2083333.42892001, -83333333.4689243, 0.0211420988081207, -0.0753769678855422, -0.0050784427151714, -85416666.8200948, -0.00842729280778538, -0.0786193224123608, 0.0237267954115858, -20833333.4891896, -0.0834101627063266, 0.020107626389934, -20833333.5053741, -41666666.8644478, 0.0327239411625642, -0.124422204184993, -0.0950276818079439, 0, 0, 0, 0, 0, 0, 0.000390317478597238, 0, 0.000163544904972845, 0, 0, 0, 0.00258113330660106, 0, -0.0017488904741372, -0.000371391023134934, 0.00461430749265891, 0.00127817438031479, 0, 0, 0, 0.00826413382085985, 0, -0.00414165187367883, 0.00180777306009897, -0.00123433615331218, 0.0183876018674656, 20833333.3527648, 0.00132473615087741, 0.000336449612637071, 0, 0, 0, 0.0138576846378334, 0, 0.000801619002236734, 0.0101996322596862, -0.00204119894562935, 0.0362042978785422, 41666666.7137914, -0.00635367986713593, 0.0186428021025692, 0.000126410912075258, 41666666.7235223, 0.000735390472404166, 0, 0, 0.0108152196645845, 0.0253116043716864, -0.00662041494371458, 0.0393265961452114, 20833333.4058859, -0.0120059360533057, 0.0594965141477911, 0.0254891599567188, -0.00987936708305456, 20833333.4211016, 41666666.7615636, -0.0209952317888034, 0.0457124008611948, 0.0718757426781074, -0.0269297748793503, 0.10053514328459, 0.0278990897427634, 0.0901659120068784, 0.0295300866764426}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 3.48217934821461e-08, 0, 0, 0, 0, 0, 0, 0, 5.34062913182482e-06, 0, -3.61182340814258e-06, -4499.99999841499, 0, 0, 0, 0, 0, 0, -5.22114896895484e-05, 0, 7.72914637882757e-05, 0, 0, -9.03765216106783e-05, 9499.99996752915, 54499.9999918199, 103250.000000792, 0, 0, 0, 0, 0, 0, 0.000143940738248502, 0, -0.000496119956961562, 0, 0, 0, 0.00118313991033925, 0, -3.74698728578573e-05, -4999.99981602, -99999.9998379875, -859166.666726229, -104999.999747074, -207916.666487858, -1240416.66706208, 0, 0, 0, 0, 0, 0, 1.10450175484001e-06, 0, 0.00147967815232877, 0, 0, 0, -0.00471321154781212, 0, 0.00249994521502569, -0.000383141772644077, 49999.9994728936, 3364583.33350797, 0, 0, -0.00412133179810794, 49999.9981369699, 104166.665519761, 99999.9983361882, 2183333.33296773, 1593749.99903791, 10185416.6707464, 2287500.00223792, 0, 0, 0, 0, 0, 0, -0.000630426534406698, 0, -0.00239283315060963, 0, 0, 0, 0.0083239154311335, 0, -0.00823585988127946, 0.000264707058887812, 0.000422183721819383, -6562499.99984976, 0, 0, 0, 0.018464757996177, 0, 0.00122080438219215, 0.00444100522080889, 0.00225341636297706, 0.00434732839398198, -1041666.66245127, -5416666.66259169, -39062500.015794, 0.00841171744569021, -729166.661753036, -16666666.6704614, -1041666.66634768, -2083333.33698934, -17395833.3514547, 0, 0, 0, 0, 0, 0, 0.00112107671093728, 0, 0.00218763806082963, 0, 0, 0, -0.00690827361470272, 0, 0.0112773991631577, 0.000153393363894154, 0.000391291390438776, 6041666.66555531, 0, 0, 0, -0.0305750681350456, 0, 0.0110377145705521, -0.0049305000089011, -0.00153709072719414, -0.00509432134297313, -0.00543426554270261, 8124999.99098596, 73750000.0274555, 0, 0, -0.0212493801081591, -0.0143957313218241, -0.00830754133920838, -0.0161601369712291, 7291666.65364676, 2083333.32033941, 56250000.0234736, -0.00651446698979318, 0.00233181698610342, 58333333.3769369, -0.0110867573007771, 7291666.68856454, 14583333.3675776, 0, 0, 0, 0, 0, 0, -0.000799204620700111, 0, -0.00107438247777134, 0, 0, 0, 0.00232680809220269, 0, -0.00716934866225856, -0.000304112841047551, -0.000730722813245598, -2083333.33198439, 0, 0, 0, 0.0217560578326747, 0, -0.0147432762987492, 0.00265631663833554, -7.44216279132371e-05, 0.00283845963761648, 0.00164482577474033, -4166666.65732208, -64583333.3546896, 0, 0, 0, 0.0309691494806815, 0, 0.00374199480706456, 0.0103214001817758, 0.005499812023163, 0.0136135533708352, 0.0197819519890733, -2083333.31701319, -83333333.371346, 0.00650070094121546, -0.00499908479735487, -85416666.7100168, 0.0181474267482059, 0.0147106791688192, -20833333.3286089, 0.0229348659658164, -0.0096019874543775, 0.0203204453646098, -0.0207904215591097, -20833333.37326, -41666666.730193, -0.0230880790948675, 0, 0, 0, 0, 0, 0, 0.00021032616284805, 0, 0.000222318117735531, 0, 0, 0, -0.000118616787170485, 0, 0.00175112963203899, 0.000115888799436879, 0.000288710702578112, -0.000504546456694033, 0, 0, 0, -0.00554734152912561, 0, 0.00547928158250007, -0.000553058146923988, 0.000330326187598771, -0.000636070914737871, 0.000363738995236104, -0.00359211291455709, 20833333.3393512, 0, 0, 0, -0.0128777264072682, 0, 0.00356298781444992, -0.00273852781236401, -0.00109928748770016, -0.00427731137089645, -0.00703487410579837, -0.00749982718183577, 41666666.6849375, -0.00196726544815332, 0.00370510326909383, 41666666.6815956, 0, 0, -0.0080467885312482, -0.00627095771376654, -0.00546040646723339, -0.00776563992897796, -0.0118685661008901, -0.00876685766915154, 20833333.3402458, -0.0099384255591098, 0.00125379497751768, -0.00737607164412099, 0.0073089220513128, 20833333.353013, 41666666.7034383, -0.0182358928831801, 0.0113808224506976, 0.0193069819227271, 0.00118271419284366, 0.00198333285563065, 0.030456832670947}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 1.54344262457491e-05, 0, 0, 0, 0, 0, 0, 0, -0.000253626370659195, 0, -0.000223899854637649, -35000.0002810483, 0, 0, 0, 0, 0, 0, 0.00213251575953816, 0, 0.00164088490291397, 0, 0, 0.00234768577071234, 167638.890699919, 692638.893540495, 517638.892167192, 0, 0, 0, 0, 0, 0, -0.0095648139020845, 0, -0.00696817551315374, 0, 0, 0, -0.0156299888280948, 0, -0.0110849151703945, -299305.561880216, -4563888.92004187, -1975694.46286363, -3113194.47305989, -2275000.01406282, -10039583.3746647, 0, 0, 0, 0, 0, 0, 0.0251458114233917, 0, 0.0181854215175446, 0, 0, 0, 0.0545205440825442, 0, 0.029652464495156, 236111.124799038, 13107638.994459, 3229166.7258591, 0, 0, 0.0587699760173762, 5197916.75454785, 3701388.92195402, 18069444.6083361, 63708333.557503, 6694444.50002523, 36329861.2805412, 41527777.9515953, 0, 0, 0, 0, 0, 0, -0.040181568204506, 0, -0.0297809502206373, 0, 0, 0, -0.106532426023588, 0, -0.0456108952021516, -69444.4627832674, -18576389.0890372, -2430555.66970471, 0, 0, 0, -0.158586456861892, 0, -0.123080008158685, -3819444.59953722, -2638888.94377007, -40833333.7828328, -171909722.809498, -7361111.24285054, -55798611.4838497, -26006944.8522845, -7500000.08006297, -206701389.607532, -63437500.3495916, -232708334.13276, -115416667.170052, 0, 0, 0, 0, 0, 0, 0.0384043708819795, 0, 0.0298469455001107, 0, 0, 0, 0.119214389028689, 0, 0.0404994014467106, 0.0140280259721194, 12847222.4381267, 694444.575545087, 0, 0, 0, 0.219596010987813, 0, 0.137167860020659, 1041666.82999845, 694444.506321913, 39583333.9721407, 225347223.025313, 2777777.96605606, 39236111.5575165, 0, 0, 0.252884107272299, 15972222.7543188, 2777777.86472091, 41666667.5152109, 450000001.376404, 4166666.77293019, 276041667.837268, 42361111.4989439, 491666668.303662, 118750000.787524, 307986112.380504, 120833333.982502, 568055557.254778, 0, 0, 0, 0, 0, 0, -0.0201776815580257, 0, -0.0166773081848876, 0, 0, 0, -0.0715993009749942, 0, -0.0200274354665385, -0.00503719317655935, -3472222.3464376, -0.0826594496147904, 0, 0, 0, -0.153043019682817, 0, -0.0738410835126793, -0.0962234362951298, -0.0393319118151233, -13888889.3418062, -142361111.665279, -0.148081515957773, -10416666.9419573, 0, 0, 0, -0.230157978310095, 0, -0.19790371993328, -3472222.58316881, -0.0721607024704647, -20833334.1440353, -416666667.881084, -0.0843115806065141, -163194445.36375, -10416666.8984238, -437500001.545223, -41666667.2925149, -13888889.622637, -0.048718283635139, -430555557.040456, -173611112.064607, -41666667.0895312, -458333335.139047, -875000002.244382, -62500000.5956632, -493055557.302185, -500000001.575215, 0, 0, 0, 0, 0, 0, 0.00447955953609603, 0, 0.00396178250265935, 0, 0, 0, 0.0179142574268259, 0, 0.00450160564903252, 0.000438790177353231, 0.0296823953324134, 0.0219621350169437, 0, 0, 0, 0.0428448057047036, 0, 0.0145684006646497, 0.0241738302465872, 0.0101343443380732, 0.126735554924349, 34722222.374808, 0.0484645865758884, 0.0692080668365406, 0, 0, 0, 0.0774354983746134, 0, 0.0579257839431725, 0.102476361702045, 0.0270433143437616, 0.288269308571016, 138888889.29347, 0.0332258604700768, 34722222.495819, 0.0598699042824445, 138888889.435766, 0.198337232234644, 0, 0, 0.0919227624720134, 0.24233910339362, 0.0204998509873388, 0.350288827205933, 208333333.977427, 0.00586419022717139, 138888889.476471, 34722222.4991586, 0.115666132096093, 208333334.214815, 416666667.693311, 0.21348213571568, 138888889.571434, 138888889.587167, 0.167928992704755, 416666667.792337, 138888889.421092, 416666667.752987, 208333334.072856}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -6.7382667940245e-06, 0, 0, 0, 0, 0, 0, 0, 0.000119955934321696, 0, 8.94318595202094e-05, 15000.000116121, 0, 0, 0, 0, 0, 0, -0.00105002051844764, 0, -0.000588800877283491, 0, 0, -0.00105953196286676, -50416.6673003101, -325416.668750401, -200416.667927902, 0, 0, 0, 0, 0, 0, 0.00471220234893956, 0, 0.00234341161307322, 0, 0, 0, 0.00807962082201048, 0, 0.00402889757874372, 56250.0019100277, 2480555.57035129, 560416.67375597, 1036805.56724036, 616666.670542905, 4290972.23926851, 0, 0, 0, 0, 0, 0, -0.0120845969460915, 0, -0.00604012845084256, 0, 0, 0, -0.0302529623320169, 0, -0.0073202210419903, -20833.3368597035, -8593750.05214315, -583333.35650343, 0, 0, -0.0271985553685079, -1093750.0343417, -625000.005450868, -7166666.73928449, -31972222.3214788, -1187500.01407418, -11461805.6236547, -12555555.6149028, 0, 0, 0, 0, 0, 0, 0.0185869071661942, 0, 0.0101277192068555, 0, 0, 0, 0.060453728183736, 0, 0.00550195547415259, 0.0035556413929353, 14409722.3240016, 208333.37908335, 0, 0, 0, 0.0859656814195948, 0, 0.0442050935043515, 381944.504161633, 208333.339461207, 20798611.3205171, 106736111.38202, 625000.035509142, 11319444.5904515, 6770833.50873558, 625000.002386375, 78437500.3141386, 12083333.4334161, 85208333.6469864, 23020833.4882731, 0, 0, 0, 0, 0, 0, -0.0170174696505909, 0, -0.0105477571068246, 0, 0, 0, -0.0668043919089968, 0, 0.000205445224864456, -0.00109981799013496, -11458333.4456722, -0.0540220340684304, 0, 0, 0, -0.130803451620616, 0, -0.0313367577609114, -0.0627159726231529, -0.00591631944521844, -25000000.3095845, -169097222.602256, -0.0579078927619093, -3819444.61307757, 0, 0, -0.118056658265879, -2083333.56003442, 0.00683802697119586, -15625000.3838283, -223611111.744018, 0.00814538545643029, -69791667.1695739, -3819444.54118624, -239236111.793705, -11458333.5660582, -73958333.8021626, -11458333.4504554, -141666667.301699, 0, 0, 0, 0, 0, 0, 0.00856912578957557, 0, 0.0061275862119013, 0, 0, 0, 0.0387036320489695, 0, -0.00213529087558707, -0.000942151116438765, 3472222.28801045, 0.0350091251916727, 0, 0, 0, 0.0956502345220891, 0, 0.00444821871456687, 0.0369404348282518, 0.00268861245867549, 10416666.8942987, 125000000.267055, 0.0519870194221785, 0.0999064316054485, 0, 0, 0, 0.126309188677338, 0, 0.0724809160254887, 0.153802121517259, 0.00127397950610572, 10416667.0472299, 260416667.24236, -0.00553775411118056, 20833333.714092, 0.0511630995031013, 270833333.998628, 0.183218836090829, 3472222.54913871, -0.0271071760180277, 159722222.901479, 20833333.6804234, 0.024593331705093, 166666667.398368, 322916667.57564, 0.0600922062488182, 62500000.6464311, 62500000.4916948, 0, 0, 0, 0, 0, 0, -0.00182938911051706, 0, -0.00150477207247729, 0, 0, 0, -0.00921280919762327, 0, 0.000647566897155479, 0.000627139186698009, -0.0159169325694063, -0.00952241153233649, 0, 0, 0, -0.0271034680994146, 0, 0.00283821978158303, -0.00925139580326084, -8.44611472788886e-05, -0.0659900709776191, -34722222.296833, -0.0188712559230197, -0.0242695746555922, 0, 0, 0, -0.0472513193518436, 0, -0.0146269618668899, -0.0439333152585618, -0.00305043639947916, -0.140007694767049, -104166666.862194, -0.0024256376622027, -0.107392175947059, -0.0115469849693901, -104166666.907689, -0.0588225091086375, 0, 0, -0.042844159795973, -0.106946731901249, 0.00550278406711343, -0.160976124144419, -104166666.973767, 0.0162615890731154, -34722222.4840884, -0.101197462911922, 0.00427530633413823, -104166667.038479, -208333333.76344, -0.0108578171838955, -0.250877442158383, -34722222.5087542, 0.026014743198579, -104166667.126816, -0.140620302116141, -104166667.068237, -0.203699656873391}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 3.22472748653579e-06, 0, 0, 0, 0, 0, 0, 0, -4.7766879166376e-05, 0, -5.38974907456278e-05, -3750.0000573281, 0, 0, 0, 0, 0, 0, 0.000367627603462359, 0, 0.000422567008108664, 0, 0, 0.000536725143771859, 7916.66696594517, 76666.6675853465, 64166.6674508038, 0, 0, 0, 0, 0, 0, -0.00158889045493556, 0, -0.00185916101775957, 0, 0, 0, -0.00306290343586133, 0, -0.00298795180825944, -4166.66751507462, -524305.561444075, -310416.671472325, -153472.227059074, -127083.336275135, -1303472.23184887, 0, 0, 0, 0, 0, 0, 0.00411666360429693, 0, 0.0049129560929018, 0, 0, 0, 0.00994082425225555, 0, 0.0095340528952587, 0.00158106407578256, 1493055.57504351, 458333.349304621, 0, 0, 0.0126630933964086, 76388.9017679639, 62500.0057140209, 944444.471904999, 8809027.82588836, 520833.34725119, 6211805.6004001, 2454861.14712019, 0, 0, 0, 0, 0, 0, -0.00653305613808304, 0, -0.00798512932217207, 0, 0, 0, -0.0189863637529549, 0, -0.0179042198816158, -0.00186638677903554, -1736111.14766444, -208333.364168686, 0, 0, 0, -0.0300160290096483, 0, -0.0298213838501008, -0.0211562013838861, -0.00811014113913514, -2152777.85285863, -24548611.2308116, -416666.702061083, -8819444.55369383, -416666.727204408, -208333.35271989, -40798611.2810287, -1145833.3937399, -15000000.1577255, -9965277.90460462, 0, 0, 0, 0, 0, 0, 0.00621001139016925, 0, 0.00782078122747761, 0, 0, 0, 0.0212410989509439, 0, 0.019637426154708, 0.00114539280261726, 694444.483748025, 0.0346832492910126, 0, 0, 0, 0.0394074922482262, 0, 0.0391392249065494, 0.0222116023378927, 0.00858747318707831, 1388888.99627662, 27430555.7139303, 0.0499708968576194, 3819444.5888424, 0, 0, 0.0502568039506213, 0.074106721146621, 0.0201458702629722, 694444.570935565, 107638889.191632, 0.0329061374784987, 53472222.5193541, 0.0547514072771775, 33680555.8675723, 7638889.11631646, 6250000.21684555, 3819444.59164645, 59722222.6157081, 0, 0, 0, 0, 0, 0, -0.0032412463208933, 0, -0.00423058955872629, 0, 0, 0, -0.012865893228059, 0, -0.0117101377244296, -0.000173130619590573, -0.0226361550749337, -0.0211125298605605, 0, 0, 0, -0.0271359518669091, 0, -0.0265080556080901, -0.0134630765079579, -0.00529518873741981, -0.077213639278092, -10416666.7734439, -0.0370453202548259, -0.0983982639933183, 0, 0, 0, -0.0415353141803766, 0, -0.0426744073975587, -0.0504205291405192, -0.0160773925956823, -0.122340182943666, -107638889.144624, -0.0279814907983164, -20833333.5844465, -0.0275839660890316, -20833333.6209793, -0.199257168074239, -0.101865009809088, -0.019739791595388, -118055555.896953, -0.139070262400295, -0.0807804800156636, -10416666.9706471, -128472222.713587, -0.170117539843428, -41666667.116069, -20833333.6676425, 0, 0, 0, 0, 0, 0, 0.000713424112050947, 0, 0.000969260409689669, 0, 0, 0, 0.00325141452784101, 0, 0.00294635875308245, -8.37961373250394e-05, 0.00542607092812595, 0.00537683321231705, 0, 0, 0, 0.00767439394510306, 0, 0.00723948878475676, 0.00347463506488608, 0.00129621739710112, 0.0220291256629359, 0.0289077344959123, 0.011235967660082, 0.0271724437033134, 0, 0, 0, 0.0134234704176213, 0, 0.0141140540284688, 0.0149843570326554, 0.00583071371965492, 0.0444146782776828, 34722222.304915, 0.00980087754913857, 0.0818627941191573, 0.00643469207775556, 0.100169539683484, 0.0678381061215959, 0, 0, 0.0171387425698751, 0.0332112252600252, 0.00800217452534442, 0.0488780200817217, 69444444.5844329, 0.00650292148175078, 34722222.3626721, 0.0353085203559458, 0.0181900703659323, 0.146907319480651, 69444444.6605778, 0.0726194344082346, 0.189402671736465, 0.102729765662731, 0.0451432867295674, 69444444.7126929, 0.0977172955789189, 34722222.4431051, 0.186144869952591}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 1.20516642269312e-06, 0, 0, 0, 0, 0, 0, 0, -1.25081717331233e-05, 0, -2.32752041380374e-05, -3750.00003355374, 0, 0, 0, 0, 0, 0, 8.66840061308697e-05, 0, 0.000221751826800502, 0, 0, 0.000122696909354701, 7916.66686514274, 64166.6672256073, 76666.6671246546, 0, 0, 0, 0, 0, 0, -0.000425262177394574, 0, -0.00112454147135501, 0, 0, 0, -0.000210673937109466, 0, -0.00102866204250105, -4166.66720723648, -310416.67002369, -524305.558249977, -127083.336735449, -153472.224212753, -1303472.22845549, 0, 0, 0, 0, 0, 0, 0.00137458051771086, 0, 0.00319350360747033, 0, 0, 0, -4.88774700533081e-05, 0, 0.00511293149495926, 0.000987954973344389, 458333.34384396, 1493055.5641636, 0, 0, 0.00106024882391055, 62500.0086949256, 76388.8925161731, 520833.351849224, 6211805.58427438, 944444.453142904, 8809027.80743563, 2454861.14011587, 0, 0, 0, 0, 0, 0, -0.00266069650581124, 0, -0.00532887934093794, 0, 0, 0, -0.000477184885610733, 0, -0.0124817448254734, -0.00145833020686, -208333.352308819, -1736111.12685207, 0, 0, 0, 0.0050930861352252, 0, -0.00915996363094483, -0.0126295956069435, -0.00386930968575272, -416666.715525939, -8819444.51064693, -2152777.79767206, -24548611.1866353, -208333.373292254, -416666.677523426, -40798611.2123669, -1145833.38601523, -9965277.8971639, -15000000.1049111, 0, 0, 0, 0, 0, 0, 0.0029213894100235, 0, 0.00524082466565259, 0, 0, 0, 0.00275045824012652, 0, 0.0155356262654244, 0.00153797634049599, 0.019790401124619, 694444.460893688, 0, 0, 0, -0.012724557262456, 0, 0.0227756417718693, 0.0119613483830317, 0.00383814255471878, 0.067847436140468, 3819444.52879033, 1388888.91316215, 27430555.6598806, 0, 0, -0.00149862741776321, 0.0441454835829135, 0.00552974517119843, 0.0822598779222442, 53472222.3789246, 694444.460272394, 107638889.076743, 0.0461836180981527, 7638889.11422178, 33680555.7455155, 3819444.61959536, 6250000.14189969, 59722222.5236676, 0, 0, 0, 0, 0, 0, -0.00167445760182475, 0, -0.00282684602641112, 0, 0, 0, -0.00341812741680672, 0, -0.00964434635589337, -0.000900353976992881, -0.0110585145954401, -0.00921025002791762, 0, 0, 0, 0.00937710342629577, 0, -0.0216111464032977, -0.00713805302566637, -0.00334944825011392, -0.0472673289636659, -0.0565423610629391, -0.0147643265840693, -10416666.7393245, 0, 0, 0, 0.014438785820359, 0, -0.0114594835802298, -0.0267349963615723, -0.00165113382040059, -0.0786698990556497, -20833333.4520301, -0.0103662237639736, -107638889.0636, -0.020431391984434, -0.203549432603376, -20833333.5003035, -0.0614255963655537, -0.00202844859474484, -118055555.735511, -0.108700064198907, -0.0831982995347461, -0.232576362057648, -41666667.0148288, -10416666.8366389, -128472222.571058, -20833333.6259201, 0, 0, 0, 0, 0, 0, 0.000389067285656515, 0, 0.000646259343786268, 0, 0, 0, 0.00130461047822665, 0, 0.00239926490915646, 0.000207071706495242, 0.00256418357314014, 0.00216736578258914, 0, 0, 0, -0.00197977939756705, 0, 0.00707938531596479, 0.00195284108536997, 0.00129546834795063, 0.0129896152738147, 0.0154974647664801, 0.00349816763067712, 0.0199119741673773, 0, 0, 0, -0.00762253870004454, 0, 0.00822444221579502, 0.00737874825720335, 0.000896166157117348, 0.0281343239315389, 0.036228273349987, 0.00219765123656999, 34722222.2846992, 0.00425015218251247, 0.070305837594587, 0.056041992197004, 0, 0, -0.00182912134998278, 0.0181695074695311, -0.00129113887363687, 0.0298530717393814, 34722222.2829973, -0.00090392187684816, 69444444.5266054, 0.0244863303117443, 0.0175026754413717, 0.110051219462507, 0.147079710844918, 0.0732883576299726, 69444444.5966223, 0.0755418540968259, 0.0507526452789235, 69444444.6308406, 0.0918160948150516, 0.177109893646327, 34722222.3904541}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -1.37389803471443e-06, 0, 0, 0, 0, 0, 0, 0, 8.96525599417421e-06, 0, 2.56256792719294e-05, 15000.0000360799, 0, 0, 0, 0, 0, 0, -8.02191289995426e-05, 0, -0.000343635925250692, 0, 0, 9.91701812136265e-05, -50416.6668231238, -200416.667273371, -325416.667229207, 0, 0, 0, 0, 0, 0, 0.000748423359501202, 0, 0.00199424081245312, 0, 0, 0, -0.00228549920924276, 0, 0.000953754686958507, 56250.0001023043, 560416.670149829, 2480555.55902998, 616666.669511085, 1036805.55809292, 4290972.23007312, 0, 0, 0, 0, 0, 0, -0.00348373907570623, 0, -0.00579802217816144, 0, 0, 0, 0.00956697397337153, 0, -0.0101778481714364, -20833.3330546765, -583333.345020026, -8593750.01103494, 0, 0, 0.00749161129383643, -625000.002749741, -1093750.00360694, -1187500.01392603, -11461805.586234, -7166666.67865487, -31972222.2647706, -12555555.5955836, 0, 0, 0, 0, 0, 0, 0.00809690047645226, 0, 0.0094519558287185, 0, 0, 0, -0.014920412457854, 0, 0.0290637965149283, -0.000110704717197641, 208333.357921723, 14409722.2411924, 0, 0, 0, -0.0443432198627244, 0, 0.0102618538228579, 208333.32926556, 381944.446577788, 625000.036753832, 11319444.5086149, 20798611.1356906, 106736111.231207, 625000.011123607, 6770833.34622872, 78437500.1202963, 12083333.3962066, 23020833.4731616, 85208333.5110985, 0, 0, 0, 0, 0, 0, -0.00987011044931525, 0, -0.00893290440464244, 0, 0, 0, 0.0071684691660666, 0, -0.0371156768002959, -0.000783840023608837, -0.030604552373676, -11458333.3508776, 0, 0, 0, 0.0787719772018181, 0, -0.053756733005717, 0.00960506086705743, -0.00322661813082605, -0.055065313076815, -3819444.53065923, -25000000.0226799, -169097222.398555, 0, 0, 0.0399621339445563, 0.0108591510132541, -2083333.33441372, -0.0254352899678891, -69791666.8082052, -15625000.0111609, -223611111.374993, -3819444.47855835, -11458333.5770616, -239236111.459717, -11458333.4867983, -73958333.5922896, -141666667.105536, 0, 0, 0, 0, 0, 0, 0.00607273319165585, 0, 0.00463118062567966, 0, 0, 0, 0.00295599172326961, 0, 0.0224223454550474, 0.000947557553277644, 0.0202948243680692, 3472222.23038947, 0, 0, 0, -0.0561189001584395, 0, 0.0579554478844062, -0.00645158884524518, 0.0054090893871371, 0.042216622304938, 0.067516525883175, 10416666.6736104, 125000000.125428, 0, 0, 0, -0.0773536627162329, 0, 0.0154106316554109, -0.0173701098332071, -0.00258035451277593, 0.0302695194383772, 20833333.4182386, 10416666.6616397, 260416666.949394, -0.00150639625178283, 0.22618038127918, 270833333.647396, -0.00801724994666705, 3472222.20794639, 159722222.414029, 0.0340178001806274, 20833333.4904557, 0.171507962083183, 62500000.4431844, 166666667.020276, 322916667.240022, 62500000.4224908, 0, 0, 0, 0, 0, 0, -0.00149153085402169, 0, -0.0010270882695245, 0, 0, 0, -0.00261103417776154, 0, -0.00526137367916254, -0.000311795531124109, -0.00550517783913972, -0.00151314468620301, 0, 0, 0, 0.0134929175055797, 0, -0.0196062781240884, 0.00143677382439305, -0.00269188636965296, -0.0127483640015135, -0.0222599905499363, 0.000740522732295705, -34722222.2561468, 0, 0, 0, 0.034985713039191, 0, -0.0210016384889232, 0.00576024995097573, -0.000731679544893612, -0.0130419073961407, -0.0247444655929076, 0.00732091490493493, -104166666.777505, 0.00435639332511782, -0.0843540786573547, -104166666.771388, 0, 0, 0.0147435726494826, 0.00816729126791868, 0.00949523985288844, -0.000144271224005819, -34722222.2632001, 0.0139497804234466, -104166666.7776, 0.0148517164212426, -0.0331474628230411, -0.0840610406061197, -0.179296429666226, -104166666.823857, -208333333.602912, -0.01106186568583, -34722222.3399906, -104166666.931101, -0.119592155117364, -0.205914398655862, -104166666.965945}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 8.96481449271475e-06, 0, 0, 0, 0, 0, 0, 0, -0.000127457861080779, 0, -0.000144706534837346, -35000.0001909781, 0, 0, 0, 0, 0, 0, 0.00105363975472303, 0, 0.00133543039324366, 0, 0, 0.00103389497588489, 167638.89005147, 517638.891848803, 692638.891571166, 0, 0, 0, 0, 0, 0, -0.0053160571896826, 0, -0.00647116851913367, 0, 0, 0, -0.00374609035104365, 0, -0.00828963080947999, -299305.55896097, -1975694.46249337, -4563888.9047344, -2275000.01651225, -3113194.45821286, -10039583.3666261, 0, 0, 0, 0, 0, 0, 0.0166118080062831, 0, 0.0174651238757982, 0, 0, 0, 0.00843472968710525, 0, 0.0368577716664857, 236111.117679136, 3229166.72612863, 13107638.9388408, 0, 0, 0.0223730277061361, 3701388.92816962, 5197916.70004036, 6694444.52848434, 36329861.2639658, 18069444.5072887, 63708333.4931401, 41527777.9391492, 0, 0, 0, 0, 0, 0, -0.031249343493348, 0, -0.0278928284913931, 0, 0, 0, -0.0201441317196414, 0, -0.080445720396091, -69444.4538285299, -2430555.6712082, -18576388.9790232, 0, 0, 0, -0.0121821342887445, 0, -0.103079664048421, -2638888.93989072, -3819444.49651597, -7361111.32850869, -55798611.470973, -40833333.4752253, -171909722.622688, -7500000.16166701, -26006944.5602695, -206701389.442619, -63437500.3086743, -115416667.276982, -232708333.951193, 0, 0, 0, 0, 0, 0, 0.0340505324229112, 0, 0.0266040776362441, 0, 0, 0, 0.0372770835193403, 0, 0.0900277792890119, 0.00909396097033064, 694444.576330018, 12847222.3159504, 0, 0, 0, -0.022660682253261, 0, 0.19785367855927, 694444.485541556, 1041666.72544994, 2777778.08206962, 39236111.5881416, 39583333.5053184, 225347222.7573, 0, 0, 0.0973005490232351, 2777777.93732098, 15972222.349114, 4166666.98537859, 276041667.546805, 41666666.8383702, 450000001.038411, 42361111.3958583, 118750001.113519, 491666667.775401, 120833334.183242, 307986112.066892, 568055557.220361, 0, 0, 0, 0, 0, 0, -0.0197292388564436, 0, -0.0141278663614764, 0, 0, 0, -0.034770534410408, 0, -0.0503254281712876, -0.00490824654074312, -0.0812759387409441, -3472222.2750111, 0, 0, 0, 0.0237019151932339, 0, -0.161797536601963, -0.0209794330616732, -0.0419857976596595, -0.217164299881302, -10416667.0042162, -13888888.9967974, -142361111.468314, 0, 0, 0, -0.00817178943464889, 0, -0.171203179011765, -0.0871468671839303, -3472222.31165069, -0.301294677825543, -163194445.13165, -20833333.4481165, -416666667.619199, -10416666.7944376, -41666667.6726499, -437500000.941144, -0.215685460708547, -13888888.9920715, -430555556.605696, -41666667.1427266, -173611111.782105, -62500001.0305281, -493055557.420511, -458333334.490738, -875000001.966273, -500000001.677865, 0, 0, 0, 0, 0, 0, 0.00469703823684668, 0, 0.00322305172732087, 0, 0, 0, 0.0120672711855042, 0, 0.0112910187380255, 0.00105433969188608, 0.0208683789928005, 0.0126198424363942, 0, 0, 0, -0.0040199363356278, 0, 0.0474077148330128, 0.00520701894550262, 0.0130601116505177, 0.0618250597015855, 0.0986408744143829, 0.0277282413888189, 34722222.3155492, 0, 0, 0, -0.0159388396340812, 0, 0.0820373955334466, 0.0226070391743627, 0.0309163632382174, 0.107818994743903, 34722222.4392155, 0.0288280479689074, 138888889.219586, 0.0238081606342088, 0.354271281728232, 138888889.192713, 0, 0, 0.0381491327706947, 0.0607753934177791, 0.0306659455127515, 0.104460347270527, 138888889.263332, 0.0221601656743547, 208333333.817114, 0.0845215337140746, 34722222.4013951, 0.47643510658729, 138888889.670891, 208333333.813352, 416666667.512335, 0.281140114053209, 138888889.31194, 416666667.708368, 138888889.441977, 208333334.251859, 416666667.690153}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -7.95863847043699e-06, 0, 0, 0, 0, 0, 0, 0, 0.00011253602247723, 0, 0.000135804224307306, 13333.3334865759, 0, 0, 0, 0, 0, 0, -0.000862475087517823, 0, -0.00111731397954245, 0, 0, -0.00122548725045229, -44814.8157177633, -244814.817226462, -244814.816905481, 0, 0, 0, 0, 0, 0, 0.00381403150269826, 0, 0.00508946437122523, 0, 0, 0, 0.00641291945290694, 0, 0.0073898371655048, 50000.0028098425, 1388888.90392173, 1388888.90147614, 772222.236052329, 772222.231266687, 4444444.47015412, 0, 0, 0, 0, 0, 0, -0.0103300806465635, 0, -0.013648946423348, 0, 0, 0, -0.0195263599443137, 0, -0.0259821747698823, -18518.5242813612, -3009259.3079723, -3009259.30052396, 0, 0, -0.0280674231205773, -805555.593373355, -805555.575602606, -3796296.37169903, -24629629.7545337, -3796296.33765335, -24629629.749109, -13194444.5534067, 0, 0, 0, 0, 0, 0, 0.0172486403617293, 0, 0.0222451190153527, 0, 0, 0, 0.0373270492398674, 0, 0.0517044864835035, 0.00803904123927101, 2777777.86787295, 2777777.85633377, 0, 0, 0, 0.0559424779776793, 0, 0.0774757230700714, 277777.838903955, 277777.808134806, 5833333.53405925, 50972222.5259762, 5833333.43418429, 50972222.5149807, 3333333.50280399, 3333333.39937641, 129722222.661292, 12916666.8687166, 63611111.564683, 63611111.4959605, 0, 0, 0, 0, 0, 0, -0.0171795330111074, 0, -0.0217641895121974, 0, 0, 0, -0.0443187885829643, 0, -0.0576385693622061, -0.00697488576152311, -925926.02183424, -925926.012946669, 0, 0, 0, -0.0624861119956767, 0, -0.116583331813108, -0.0623098937307559, -0.0333979926119527, -2777778.05865736, -44444444.8423682, -2777777.91389783, -44444444.8333652, 0, 0, -0.110903398797736, -925926.129951892, -925925.994946239, -2777778.12360893, -240740741.493493, -2777777.88494782, -240740741.521363, -4166666.86317164, -93055556.4288273, -93055556.2396466, -52777778.4364062, -52777778.2964868, -292592593.72261, 0, 0, 0, 0, 0, 0, 0.00931721934308484, 0, 0.0117714742723251, 0, 0, 0, 0.0292553994606455, 0, 0.0339386194652275, 0.0031862174992848, 0.054801362491456, 0.0522519068018666, 0, 0, 0, 0.0399750455924765, 0, 0.0858840437668024, 0.0371710298380524, 0.022545936405393, 0.197481568449714, 13888889.1572858, 0.0963806655150901, 13888889.152946, 0, 0, 0, 0.0722015687694079, 0, 0.115471126064511, 0.134145806979074, 0.049846063163578, 0.328556115921741, 180555556.173606, 0.0856388861415742, 180555556.234078, 0.102952949755727, 41666667.4636733, 41666667.2584953, 0.276957922933181, 0.0634975242060762, 333333334.190018, 13888889.3186775, 13888889.2193558, 41666667.5489183, 375000001.357683, 41666667.2696508, 375000001.288278, 208333334.387493, 0, 0, 0, 0, 0, 0, -0.00211235836640005, 0, -0.00270349988414373, 0, 0, 0, -0.00806369290622281, 0, -0.00831236756666176, -0.000539353622899191, -0.0130526019634785, -0.0131858808767228, 0, 0, 0, -0.0117223833131884, 0, -0.0243313740599903, -0.00964613229143058, -0.00649541686705574, -0.0549670686764051, -0.0731426347025802, -0.0280194722107184, -0.0719076604814902, 0, 0, 0, -0.0181031797981947, 0, -0.0445292475397066, -0.0385943698630264, -0.0173423266966947, -0.116990003513877, -46296296.4933907, -0.027649713255294, -46296296.5230603, -0.0246033269184085, -0.276584202257258, -0.197744440619783, 0, 0, -0.0381389628759046, -0.0884358699747786, -0.0210393188601261, -0.131896008836272, -138888889.223601, -0.0196892564795842, -138888889.254163, -0.107077340952186, -0.0839303980136469, -0.419683514862277, -138888889.476365, -0.25388929991851, -138888889.434865, -0.297219086325843, -0.191754841269915, -277777778.510729, -46296296.6317815, -138888889.540072, -138888889.482794}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 3.56362088079723e-06, 0, 0, 0, 0, 0, 0, 0, -9.35384280778678e-05, 0, -1.57479675578067e-05, -26250.0001339493, 0, 0, 0, 0, 0, 0, 0.000921671152176732, 0, 4.92673232767855e-05, 0, 0, 0.000486800604563668, 388229.168465311, 388229.168443651, 388229.167920505, 0, 0, 0, 0, 0, 0, -0.00461699479520847, 0, 0.000354167320597072, 0, 0, 0, -0.00388611237512846, 0, -0.00336205780090399, -1481770.84189247, -1481770.84395769, -1481770.83666353, -5588541.68540447, -5588541.68478536, -5588541.68036743, 0, 0, 0, 0, 0, 0, 0.0139283941624691, 0, -0.00274263677336668, 0, 0, 0, 0.0107077836361318, 0, 0.00898884604132368, 2421875.02096064, 2421875.03372578, 2421874.99975418, 0, 0, 0.0276498923301452, 19838541.7394853, 19838541.7428089, 19838541.7464499, 19838541.7416798, 19838541.7311509, 19838541.694481, 78500000.151532, 0, 0, 0, 0, 0, 0, -0.0257753642251926, 0, 0.00678457271599821, 0, 0, 0, -0.0172690177698393, 0, -0.00288937129655818, -1822916.69641094, -1822916.7297073, -1822916.65134214, 0, 0, 0, -0.0607950262767486, 0, -0.0910487607140884, -29739583.4704945, -29739583.4892772, -29739583.5060766, -29739583.5279961, -29739583.4434424, -29739583.3135083, -55833333.5770942, -55833333.5660984, -55833333.5154611, -259843750.498048, -259843750.521585, -259843750.407163, 0, 0, 0, 0, 0, 0, 0.0282383233067969, 0, -0.00783604759513064, 0, 0, 0, 0.0206141388135104, 0, -0.0188438426524164, 520833.35803738, 520833.403848938, 520833.306304584, 0, 0, 0, 0.0545422341047921, 0, 0.122639236407847, 20312500.1376422, 20312500.1751538, 20312500.1991472, 20312500.2650463, 20312500.1022003, 20312499.8619319, 0, 0, 0.18184674204422, 58333333.6804801, 58333333.6996484, 58333333.7250607, 58333333.7350156, 58333333.6522165, 58333333.5061468, 358854167.388784, 358854167.554671, 358854167.105665, 677083334.521682, 677083334.496869, 677083334.400364, 0, 0, 0, 0, 0, 0, -0.0167046334648214, 0, 0.00429533304878483, 0, 0, 0, -0.0160791149543074, 0, 0.0248467226843954, -0.0110203543113308, -0.043847172202007, 0.0189664686960772, 0, 0, 0, -0.0206763622771664, 0, -0.0652624130299998, -5208333.40491284, -5208333.43784797, -5208333.44929676, -5208333.51943485, -5208333.386584, -5208333.16634915, 0, 0, 0, -0.132780868708878, 0, -0.197544324200798, -20833333.5687691, -20833333.6101969, -20833333.628551, -20833333.7028581, -20833333.5438492, -20833333.3479213, -223958333.823947, -223958334.073916, -223958333.489334, -31250000.3821301, -31250000.3332841, -31250000.3703074, -645833334.421623, -645833334.561116, -645833334.600086, -645833334.561751, -645833334.382104, -645833334.160129, -1250000001.8096, 0, 0, 0, 0, 0, 0, 0.0040984342082607, 0, -0.00089233011051798, 0, 0, 0, 0.00546188349663506, 0, -0.00912429875267179, 0.00199042416205457, 0.0116480827013761, -0.00479670221676812, 0, 0, 0, 0.00276305583453234, 0, 0.00897298925875648, 0.0152502273949122, 0.0259645157045731, 0.0268714731950054, 0.0534789431663292, 0.0130484694943606, -0.0645878245304229, 0, 0, 0, 0.0322372052569427, 0, 0.0687041191358356, 0.062246469112825, 0.0828229243665275, 0.0821818962055441, 0.125243136330105, 0.0577188978344325, -0.0442025711220276, 52083333.4625721, 52083333.5733899, 52083333.3235744, 0, 0, 0.0823367929464354, 0.134552365724997, 0.123938990543541, 0.148384125949263, 0.186332033238124, 0.10573159846156, 0.118811515767412, 208333333.682302, 208333333.77803, 208333333.838784, 208333333.852003, 208333333.645039, 208333333.520714, 312500000.560107, 312500000.603745, 312500000.510045, 625000000.904024, 625000000.931381, 625000000.81818}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -3.53481473262566e-07, 0, 0, 0, 0, 0, 0, 0, -1.13699335408543e-06, 0, 9.00536817059944e-06, -2812.50000705829, 0, 0, 0, 0, 0, 0, 5.21966356249088e-06, 0, -0.000101522031582125, 0, 0, 5.12004890644113e-05, 57500.000246113, 34062.5000832738, 34062.4999767954, 0, 0, 0, 0, 0, 0, 3.79981814726677e-05, 0, 0.000638919469888105, 0, 0, 0, -0.000408316164367525, 0, -0.000460162374659124, -393229.168035636, -62500.0002740008, -62499.9988883396, -690104.168960559, -690104.169292566, -406250.000027282, 0, 0, 0, 0, 0, 0, -0.000153365249981745, 0, -0.00219025597449042, 0, 0, 0, 0.000300101789162629, 0, 0.000767913954525311, 1119791.67003662, 31250.0003871767, 31249.9937073069, 0, 0, 0.0046961507804272, 4640625.01164644, 4640625.01379024, 1208333.34004905, 687500.000460531, 1208333.34140309, 687499.992667614, 8166666.68781929, 0, 0, 0, 0, 0, 0, 0.000174078412926212, 0, 0.00418972030819856, 0, 0, 0, 0.00137543572992122, 0, 0.00174606380143825, -1302083.33757166, -0.000447963081031602, 0.0157553804167732, 0, 0, 0, -0.00939008182175109, 0, -0.0149484617761853, -12812500.0259261, -12812500.0288154, -572916.67419863, -312500.001322396, -572916.675105894, -312499.965134164, -7395833.36279468, -7395833.37676633, -624999.981641779, -54114583.4277371, -13229166.7257489, -13229166.7148999, 0, 0, 0, 0, 0, 0, -1.97180569998963e-05, 0, -0.00448008772369228, 0, 0, 0, -0.00282237938525429, 0, -0.00640710865433217, 520833.33598755, 0.000674133876993993, -0.0204939980798023, 0, 0, 0, 0.0064984672877007, 0, 0.018966196079453, 14062500.0288366, 14062500.0290931, 0.00164090910256861, 0.00308549786560187, -0.00146057502479341, -0.0653854773143337, 0, 0, 0.0289937537218056, 16666666.7210905, 16666666.740788, 3125000.026769, -0.023552354914086, 3125000.058383, -0.0665241564607011, 145312500.159197, 5729166.74362156, 5729166.69445888, 80208333.5360508, 80208333.5476367, 11458333.4319068, 0, 0, 0, 0, 0, 0, -8.08196612336084e-05, 0, 0.0025068823639837, 0, 0, 0, 0.00198237423363729, 0, 0.00652834735315994, -0.00065534300388339, -0.000639657237027475, 0.013555341354783, 0, 0, 0, -0.000645020039730127, 0, -0.00927885633297613, -5208333.34877516, -5208333.34701999, 0.00245305988178704, -0.00379456460817585, 0.00804977071204965, 0.0561364389308328, 0, 0, 0, -0.0198212069236147, 0, -0.0305368493315483, -10416666.7119873, -10416666.7202452, -0.00444400216521638, 0.0137827063708854, -0.0348118130871106, 0.0816155445437278, -151041666.784534, -0.0494554475231016, 0.0181039256981154, -5208333.37118505, -5208333.40633988, 0.0572372545240123, -177083333.569342, -177083333.593997, -31250000.173597, -0.0867640293260138, -31250000.1800667, -0.0270590105124736, -62500000.308406, 0, 0, 0, 0, 0, 0, 3.80692742844007e-05, 0, -0.000572301798848934, 0, 0, 0, -0.000486652458374759, 0, -0.00222562629383292, -3.14471923120758e-07, 0.000224857157949814, -0.00360581583920043, 0, 0, 0, -0.000598108943697126, 0, 0.00095243265150814, 0.00309896163516663, 0.00226236693910252, -0.00123145565364416, 0.00162012823282381, -0.00386706322993121, -0.0183815439525019, 0, 0, 0, 0.00434833143001455, 0, 0.0103483194386384, 0.0138467711319806, 0.0137925648944376, -0.00314780326090812, -0.00232894643509133, 0.00729160455222326, -0.0342633977180197, 52083333.3651589, 0.0131057352943053, -0.0170548658302738, 0, 0, 0.012164225152631, 0.0180931875820034, 0.0279574955614229, 0.00560571976254253, -0.0183570569730928, 0.0245104073885489, -0.0339932425464713, 104166666.758177, 104166666.767573, 0.0515263752438838, 0.0295612087253691, 0.048803471261177, -0.0203503596265495, 52083333.4357175, 52083333.4559107, 0.00798954101506624, 104166666.860011, 0.138021894909534, 0.124751763890233}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 1.93898555211988e-06, 0, 0, 0, 0, 0, 0, 0, -3.42575470000205e-05, 0, -2.20055826994449e-05, -2812.50002584891, 0, 0, 0, 0, 0, 0, 0.00031150183356761, 0, 0.000134143739926486, 0, 0, 0.000221263426015346, 34062.5001993721, 57500.0004185546, 34062.500194898, 0, 0, 0, 0, 0, 0, -0.00142456083211287, 0, -0.000518347132553644, 0, 0, 0, -0.00191741449322941, 0, -0.000467822135453793, -62500.0008509122, -393229.169942591, -62500.0009334195, -690104.169612529, -406250.000711215, -690104.168380272, 0, 0, 0, 0, 0, 0, 0.00365700789993339, 0, 0.00137925149862036, 0, 0, 0, 0.00769115830083267, 0, -0.00063621537432229, 31250.0021090239, 1119791.67872993, 31250.0034591499, 0, 0, 0.00502985542736231, 1208333.34395521, 687500.00198781, 4640625.01899725, 4640625.01557877, 687500.001434677, 1208333.33379025, 8166666.67142502, 0, 0, 0, 0, 0, 0, -0.00560359571772787, 0, -0.00244804982787191, 0, 0, 0, -0.0153132716218583, 0, 0.00388678584568312, -0.00309603730866889, -1302083.35674714, -0.00861278308367827, 0, 0, 0, -0.0205279092666641, 0, -0.0027527110775712, -572916.687054417, -312500.004907138, -12812500.0564546, -12812500.0533455, -312500.005814418, -572916.655663884, -7395833.38717415, -624999.999129779, -7395833.36155204, -13229166.6778729, -54114583.3779759, -13229166.6479866, 0, 0, 0, 0, 0, 0, 0.00514326338025461, 0, 0.0026706585174261, 0, 0, 0, 0.0160139115171547, 0, -0.00562983809229012, 0.00257358830758268, 520833.358303171, 0.0125995333639531, 0, 0, 0, 0.0342422214465533, 0, -0.00802253619483515, 0.0219765607248245, 0.00746152580144637, 14062500.0827211, 14062500.0821807, 0.01596418063141, -0.025840169774902, 0, 0, 0.0218519329398745, 3125000.07441644, -0.000909811643725419, 16666666.7882494, 16666666.7661311, -0.00229409191391803, 3125000.01151964, 5729166.68612053, 145312500.130502, 5729166.60698398, 80208333.4101272, 11458333.2858481, 80208333.3359026, 0, 0, 0, 0, 0, 0, -0.00262891283180563, 0, -0.00159038459350294, 0, 0, 0, -0.00848148531451854, 0, 0.00335944471015184, -0.00104475597470007, -0.0139550443216213, -0.00956988933360999, 0, 0, 0, -0.0251061946652223, 0, 0.0119090534901315, -0.0128675252485603, -0.00563916210772953, -5208333.39173905, -5208333.39226925, -0.0188837794917942, 0.0221338242327989, 0, 0, 0, -0.031044368799644, 0, -0.00485495438546564, -0.0498375722863496, -0.00238577168807428, -10416666.7875943, -10416666.7805646, -0.00464365260480132, 0.0211960000816025, -0.0176999874022059, -151041666.813352, 0.0601408561347819, -5208333.44581272, 0.00691112664378153, -5208333.41145215, -31250000.0690974, 0.0363422319469303, -177083333.491176, -177083333.415697, 0.0903516807137007, -31249999.9402089, -62499999.9601195, 0, 0, 0, 0, 0, 0, 0.000577633241722479, 0, 0.000392813166455963, 0, 0, 0, 0.00181110386925212, 0, -0.000697758580142752, 0.000133714580582462, 0.0032221808391369, 0.00288990920317045, 0, 0, 0, 0.00670175921288913, 0, -0.00454465789274927, 0.00319195502401813, 0.00160764789740465, 0.0159048079984229, 0.0160627029456536, 0.00784017705986217, -0.0064174372245685, 0, 0, 0, 0.0128976820790007, 0, -0.0027844853268187, 0.0133446734653944, 0.00189112419275494, 0.043284088594418, 0.0427986075320396, 0.00547517989616077, -0.016591967987652, 0.00639796425986457, 52083333.3898491, -0.0201356273531855, 0, 0, 0.00809899609290094, 0.03709605227905, -0.00207704083703093, 0.0568256522971575, 0.0536664049361351, -0.00231863895538948, 0.0164564192391479, 0.0235683680557212, -0.00792748701516157, 104166666.759835, 104166666.729791, -0.0427527027402708, -0.04119250722317, 52083333.4036545, -0.03789864025235, 52083333.3431884, -0.0194313781579624, 104166666.684919, -0.0640397545522798}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 5.68938270776297e-08, 0, 0, 0, 0, 0, 0, 0, -6.70454781917976e-06, 0, 6.99946267812248e-06, -2812.5000068894, 0, 0, 0, 0, 0, 0, 0.000107745477959537, 0, -2.79968755202862e-05, 0, 0, -0.000103422238445771, 34062.5000269507, 34062.5000616453, 57500.0001511435, 0, 0, 0, 0, 0, 0, -0.000756950413540995, 0, 3.81226846612473e-05, 0, 0, 0, 0.00069046284398611, 0, 0.000229053442709035, -62499.9998653094, -62500.0003480898, -393229.167679059, -406249.999747927, -690104.168122229, -690104.167856265, 0, 0, 0, 0, 0, 0, 0.00283324645732763, 0, -9.57922221991551e-05, 0, 0, 0, -0.00217221510664183, 0, 0.000938644020122725, 31249.9991874453, 31250.0020969232, 1119791.66979553, 0, 0, -0.00195274020274854, 687499.995959544, 1208333.33719208, 687499.996645232, 1208333.3356032, 4640625.00846497, 4640625.00819401, 8166666.67745755, 0, 0, 0, 0, 0, 0, -0.00590473961128953, 0, 0.000299347439808054, 0, 0, 0, 0.00206163065159446, 0, -0.00295658648938267, 0.00161181057805333, -0.00730174619644703, -1302083.3382866, 0, 0, 0, 0.0101545156195228, 0, -0.00502531515678361, -312499.986749011, -572916.671730383, -312499.990407746, -572916.669088234, -12812500.0201899, -12812500.0246827, -624999.977708243, -7395833.35349073, -7395833.34881136, -13229166.6852412, -13229166.6844481, -54114583.3975428, 0, 0, 0, 0, 0, 0, 0.00685724287513181, 0, -0.000407280514530668, 0, 0, 0, 0.00195410547442057, 0, 0.00234201769641285, -0.00158030267996862, 0.0125906526383344, 520833.337374709, 0, 0, 0, -0.0187233076790173, 0, 0.0174127288405053, -0.0192935465953973, 0.00477742552488909, -0.0103732003086849, 0.00763573694955194, 14062500.0240711, 14062500.0346054, 0, 0, -0.00089203210883348, -0.0455946551624382, 3125000.02162089, -0.0460550042756858, 3125000.00272816, 16666666.7020456, 16666666.7163365, 5729166.67046853, 5729166.67708566, 145312500.129576, 11458333.3293425, 80208333.4618547, 80208333.4471866, 0, 0, 0, 0, 0, 0, -0.0041510076866604, 0, 0.000224541242290868, 0, 0, 0, -0.00439682650787811, 0, -6.52287786331562e-05, 0.000798665989739172, -0.0103500616786053, -0.00157438384175962, 0, 0, 0, 0.0133161240033044, 0, -0.0154920442877856, 0.0134368419511553, -0.00361933386465061, 0.00365490183252723, -0.0131143220933337, -5208333.34755611, -5208333.35496709, 0, 0, 0, 0.0119757565706088, 0, -0.0165056244793844, 0.0396013299207829, -0.0124093151284729, 0.0404646091902537, 0.00922605773242761, -10416666.6935009, -10416666.7271879, 0.0120901731225951, -0.00748376979981803, -151041666.778544, 0.0584173653360908, -5208333.35638807, -5208333.35853167, 0.0492816132323285, -31250000.0976316, 0.0443800563865077, -31250000.076748, -177083333.518943, -177083333.505745, -62500000.1389316, 0, 0, 0, 0, 0, 0, 0.00102108585423114, 0, -3.8004160474617e-05, 0, 0, 0, 0.00196164412156283, 0, -0.000374766919246948, -0.000172513717085064, 0.00325996021625745, 0.000228420513856105, 0, 0, 0, -0.00293410290038226, 0, 0.00425170380772951, -0.00367123447842192, 0.0013346953317123, 0.000257064109700217, 0.00678556382714102, 0.00331889205030259, 0.00465109785162095, 0, 0, 0, -0.00692308240164071, 0, 0.00993857249771765, -0.0127149339117536, 0.00367125329946141, -0.0128322572678273, -0.00321338542982563, 0.00758473297784412, 0.0238195053931135, -0.00667794437186858, 0.0058018246410732, 52083333.3684042, 0, 0, 0.00230169181017049, -0.02359893952256, 0.00510859406940355, -0.0252349955707513, -0.00178831829370209, 0.00746467858562353, 0.0240671633286169, -0.0314889985400414, 0.0246268963239662, -0.0240466366724584, 0.021220881827672, 104166666.747605, 104166666.748227, -0.0493488398828373, 52083333.4123728, 52083333.3972464, 0.0426678704264395, 0.0356769186120778, 104166666.788584}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 2.01332776024534e-06, 0, 0, 0, 0, 0, 0, 0, -7.74559694618067e-06, 0, -5.09524213917754e-05, 10000.000027665, 0, 0, 0, 0, 0, 0, 1.10528708580243e-05, 0, 0.000471878586755448, 0, 0, 0.000178558142775801, -183611.111949413, -133611.11145986, -133611.111091672, 0, 0, 0, 0, 0, 0, 4.94088019584338e-05, 0, -0.00257185208531312, 0, 0, 0, -0.000615049641487248, 0, -0.00045802571693446, 1041666.67119798, 373611.112565118, 373611.10773615, 2415277.78599251, 2415277.78613255, 1747222.22341408, 0, 0, 0, 0, 0, 0, -0.000665730791284957, 0, 0.00820572118121564, 0, 0, 0, 0.00424606189727286, 0, 0.00294001329330778, -2256944.45574991, -388888.891437642, -388888.868708643, 0, 0, -0.00573509183254843, -13263888.9272865, -13263888.9311107, -6395833.36208773, -4527777.78702206, -6395833.35722282, -4527777.75953151, -31152777.8488877, 0, 0, 0, 0, 0, 0, 0.00230380420284907, 0, -0.0151660072033394, 0, 0, 0, -0.0118036928489786, 0, -0.0145442036201933, 2083333.34868829, 138888.891036568, 138888.837225667, 0, 0, 0, 0.00862711271021583, 0, 0.0276302236928842, 26944444.5239675, 26944444.5334395, 6250000.04477081, 4305555.57748477, 6250000.02013659, 4305555.45225535, 30972222.334716, 30972222.346863, 8333333.31150399, 166111111.399435, 76666666.8934324, 76666666.8259082, 0, 0, 0, 0, 0, 0, -0.00354448514213288, 0, 0.015941075967964, 0, 0, 0, 0.0144089051299808, 0, 0.0295323742181296, -694444.456039402, -0.00130056179846825, 0.0678909365637647, 0, 0, 0, 0.00469048429079765, 0, -0.036817128673189, -22916666.7507537, -22916666.7621942, -2083333.36240501, -1388888.91382926, -2083333.31893132, -1388888.68027677, 0, 0, -0.0582201753746111, -45833333.5173044, -45833333.5459575, -25694444.5858475, -4166666.6656851, -25694444.5998834, -4166666.52183407, -317361111.569938, -68750000.3381928, -68750000.0811638, -361805556.225216, -361805556.190016, -133333333.740866, 0, 0, 0, 0, 0, 0, 0.00257210659111543, 0, -0.008840109343396, 0, 0, 0, -0.00792138016296053, 0, -0.0260012550184582, 0.00449323552684404, 0.0010721955307468, -0.045047042745893, 0, 0, 0, -0.0132235118492126, 0, 0.0146842754943061, 6944444.4887194, 6944444.49598486, 0.00272706044900092, 0.0151165596991444, -0.0314798307377168, -0.187522256617897, 0, 0, 0, 0.0363619963210281, 0, 0.0711004504333091, 20833333.4707161, 20833333.4925223, 6944444.51565937, 0.0115917788369262, 6944444.528947, -0.21208901883864, 250000000.327731, 20833333.5713001, 20833333.2508283, 20833333.49903, 20833333.5276491, -0.0773430001282952, 500000000.705597, 500000000.739696, 277777778.425682, 62500000.4405354, 277777778.282705, 62500000.1795373, 541666667.660588, 0, 0, 0, 0, 0, 0, -0.000720331459019772, 0, 0.00200817679751035, 0, 0, 0, 0.00155134150902279, 0, 0.00836174205073007, -0.000670229255469323, -0.000509551239125012, 0.0119680388231387, 0, 0, 0, 0.00559355813518077, 0, 0.000900383198482579, -0.00915437522823558, -0.0110960176313843, 0.00293791242293845, -0.00412079256028995, 0.0133923060904071, 0.0632532294587289, 0, 0, 0, -0.0059076220891051, 0, -0.0249740977066782, -0.0386359472054204, -0.0445236153052373, -0.00830171372071286, -0.00556842813127975, -0.0157646896171268, 0.0990604072096972, -69444444.5320887, -0.0638184355174139, 0.0689410621207907, 0, 0, -0.0287503549950736, -0.0664191398397318, -0.0755777342021577, -0.0472740784714215, 0.0134358263285708, -0.0605570978147116, 0.0629901683959318, -208333333.584907, -208333333.613439, -69444444.665852, -0.169516552950247, -69444444.5644925, 0.037702997582532, -208333333.672834, -208333333.671481, -0.120739839540557, -416666667.234922, -208333333.829259, -208333333.729637}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -8.72944695281167e-07, 0, 0, 0, 0, 0, 0, 0, 9.72782259854258e-06, 0, 2.06187518941593e-05, -2500.00000273582, 0, 0, 0, 0, 0, 0, -5.72302564490202e-05, 0, -0.000175025174064013, 0, 0, -0.000202741640465584, 42777.7780133807, 42777.7778029869, 30277.7776518481, 0, 0, 0, 0, 0, 0, 0.000171556924797419, 0, 0.000841358279384322, 0, 0, 0, 0.00116624260032928, 0, 0.00117257373457963, -206944.445831119, -206944.444399241, -55555.5537870278, -726388.89153284, -512500.001486427, -512499.99891569, 0, 0, 0, 0, 0, 0, -0.000203391788057, 0, -0.00243041708260277, 0, 0, 0, -0.0040088786577666, 0, -0.00402679506624101, 305555.559257857, 305555.554423836, 27777.7697570242, 0, 0, -0.00458804917094566, 3451388.9023545, 2416666.67546155, 3451388.89968347, 2416666.66369128, 888888.889916518, 888888.874400961, 8618055.57049759, 0, 0, 0, 0, 0, 0, -9.53119433951797e-05, 0, 0.00423271976965177, 0, 0, 0, 0.00770286124703383, 0, 0.0088410952597249, -138888.894457552, -138888.884625949, 0.0179271985979968, 0, 0, 0, 0.0121259139573829, 0, 0.00951572586001791, -4861111.13987164, -3333333.35270827, -4861111.13099953, -3333333.32870121, -416666.658848601, -416666.61268465, -15694444.4909813, -3611111.12777404, -3611111.07224704, -40000000.0718935, -40000000.0574027, -14166666.6741292, 0, 0, 0, 0, 0, 0, 0.000492416438272739, 0, -0.00432473197429127, 0, 0, 0, -0.00819562261588624, 0, -0.0116268915199076, 0.00487599297283203, -0.00693258598031251, -0.0216548960306012, 0, 0, 0, -0.0178300210121917, 0, -0.0125916064704437, 2083333.36512939, 1388888.9109449, 2083333.34851449, 1388888.88309838, -0.0192849056725626, -0.0905625416156648, 0, 0, -0.0162953744901743, 19444444.5197222, 2777777.80777353, 19444444.5178962, 2777777.728323, 1388888.89909202, 1388888.77625186, 52777777.8981552, 52777777.879002, 6249999.94530635, 178472222.418094, 56944444.5503296, 56944444.4827808, 0, 0, 0, 0, 0, 0, -0.000464335610328413, 0, 0.00238113642537505, 0, 0, 0, 0.00454976950989226, 0, 0.00813625052101971, -0.00229471476745049, 0.00523773494805578, 0.0135821226602503, 0, 0, 0, 0.0129986089276799, 0, 0.0100321418834902, -0.0182167594891682, -0.013112777682203, -0.00164391447004528, 0.00540819588770861, 0.0171806267238272, 0.0721850895702175, 0, 0, 0, 0.0156137915690925, 0, 0.0112851663893598, -6944444.49887105, -0.0222273482316491, -6944444.49509941, 0.0308369534032143, 0.00138463390116244, 0.126319139841873, -20833333.4238491, -20833333.4138154, 0.0906066557800965, -13888888.9722732, -0.0164913184486169, 0.0991832309850309, -208333333.5459, -41666666.8070798, -208333333.558067, -41666666.7300912, -20833333.3722111, -20833333.2631011, -222222222.420643, 0, 0, 0, 0, 0, 0, 0.000147427301312298, 0, -0.000544745590167192, 0, 0, 0, -0.0010289019450901, 0, -0.00230521552039979, 0.000439691395512848, -0.00150084226590017, -0.00347539018212776, 0, 0, 0, -0.00371811193281603, 0, -0.00347781771099796, 0.00431009224685026, 0.00321555613288534, -0.00205129244290263, -0.00226299633801879, -0.00551793066564301, -0.022252841803193, 0, 0, 0, -0.00569372849852685, 0, -0.00353163981399006, 0.0149789834563167, 0.00628665964184045, 0.011442918259719, -0.00772173040910792, -0.00232102617081976, -0.050053959673122, 0.0263040189867636, 0.0228315184595953, -0.0409245749965748, 0, 0, -0.00496619867972492, 0.0308094749801197, 0.00566292567214515, 0.0315435840295516, -0.0319259098135303, 0.00237824123567098, -0.055621862870024, 69444444.5211261, 0.0560721339898642, 69444444.5346988, 0.0318839703245598, -0.0104375090044441, -0.0667537892488977, 138888889.010258, 0.0451012335786525, -0.0232765400093523, 138888889.014521, 138888889.006319, 69444444.4862308}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -4.34035020644551e-06, 0, 0, 0, 0, 0, 0, 0, 7.67931097590748e-05, 0, 5.0683412250568e-05, 10000.0000793722, 0, 0, 0, 0, 0, 0, -0.000686046389826849, 0, -0.000335840391986626, 0, 0, -0.000523455328595265, -133611.11197348, -183611.112236924, -133611.111813313, 0, 0, 0, 0, 0, 0, 0.003166720975206, 0, 0.00127752964741023, 0, 0, 0, 0.00396407657808744, 0, 0.00202273214742882, 373611.115205515, 1041666.6742214, 373611.11390301, 2415277.78825917, 1747222.2284426, 2415277.7845475, 0, 0, 0, 0, 0, 0, -0.0085060602185005, 0, -0.0030530049898937, 0, 0, 0, -0.0140411814442033, 0, -0.00317558332322428, -388888.89939425, -2256944.46958375, -388888.896157676, 0, 0, -0.0149804308840463, -6395833.3744154, -4527777.80232416, -13263888.943335, -13263888.9337418, -4527777.79755128, -6395833.34337149, -31152777.8301404, 0, 0, 0, 0, 0, 0, 0.01395446752969, 0, 0.0048146653759439, 0, 0, 0, 0.0261506652282669, 0, -0.00058302789766144, 138888.904823923, 2083333.37871976, 138888.902934316, 0, 0, 0, 0.0453175378548595, 0, 0.0281929982637422, 6250000.08077843, 4305555.60775127, 26944444.5840143, 26944444.5773248, 4305555.59508048, 6249999.98383976, 30972222.3899856, 8333333.39300562, 30972222.31449, 76666666.8342583, 166111111.338281, 76666666.7529317, 0, 0, 0, 0, 0, 0, -0.0138132453621944, 0, -0.00488171938531459, 0, 0, 0, -0.0271467370763613, 0, 0.00695047423972756, -0.0142070835290601, -694444.490204843, -0.0186734851784159, 0, 0, 0, -0.0624752025585815, 0, -0.0214888053408462, -2083333.42053742, -1388888.95204501, -22916666.8494287, -22916666.8571541, -1388888.94364963, -2083333.26878643, 0, 0, -0.0761537413929899, -25694444.6857862, -4166666.75940121, -45833333.6595407, -45833333.5866837, -4166666.74516176, -25694444.5068075, -68750000.2533984, -317361111.584388, -68750000.0333159, -361805556.050314, -133333333.584575, -361805555.861103, 0, 0, 0, 0, 0, 0, 0.00757899904696929, 0, 0.0028424491243482, 0, 0, 0, 0.015110587235982, 0, -0.00683226630926069, 0.00674920984042466, 0.02450369081611, 0.0141351640950875, 0, 0, 0, 0.0403034781458355, 0, 0.00169109196994322, 0.0503664073476014, 0.0408627966666551, 6944444.56093019, 6944444.5756083, 0.0461524439245259, -0.0669859454234974, 0, 0, 0, 0.075712362430907, 0, 0.0553052731347363, 6944444.61059159, 0.0742803631371368, 20833333.6243481, 20833333.5974518, 0.0619283350846876, 6944444.41770189, 20833333.5194116, 250000000.45144, 20833333.2914401, 20833333.6492781, 0.0711291605173936, 20833333.5376032, 277777778.239564, 62500000.2850277, 500000000.662532, 500000000.472483, 62500000.1465694, 277777777.911542, 541666667.127365, 0, 0, 0, 0, 0, 0, -0.00176728285705629, 0, -0.000710524303967185, 0, 0, 0, -0.00357625599308864, 0, 0.00206040922653662, -0.00127883780315101, -0.00551008759412929, -0.00441134436852711, 0, 0, 0, -0.00987198202551202, 0, 0.00325431536975742, -0.0122288955173926, -0.0108632170482309, -0.0285062484549626, -0.0348303104956052, -0.0165774183697143, 0.0228645929885043, 0, 0, 0, -0.0255677464219543, 0, -0.0129971688898465, -0.0452672699557993, -0.0242144617203462, -0.0949133758008135, -0.0950038119227534, -0.0237869470604016, 0.0347398385457015, -0.0547456578881877, -69444444.602445, 0.0296851547460215, 0, 0, -0.0309853764609185, -0.10828804963041, -0.0258797749676853, -0.144101788779148, -0.123042321858319, -0.0221679714357909, -0.0509169896975517, -69444444.5993361, -0.11246082477406, -208333333.649301, -208333333.576426, -0.0199339959916397, -69444444.4334993, -208333333.635578, -0.102580577162142, -208333333.481093, -208333333.565606, -416666666.971971, -208333333.461897}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -2.27599641301046e-06, 0, 0, 0, 0, 0, 0, 0, 5.32657557762299e-05, 0, 1.90783815300454e-05, 10000.0000673728, 0, 0, 0, 0, 0, 0, -0.000531879870232662, 0, -0.000176122519196988, 0, 0, -0.000288681669280051, -133611.111735101, -133611.111953796, -183611.112065223, 0, 0, 0, 0, 0, 0, 0.00282364323067614, 0, 0.000708360838872003, 0, 0, 0, 0.00180622960523551, 0, 0.00264619125148082, 373611.113504999, 373611.116006369, 1041666.67167652, 1747222.22776341, 2415277.78661758, 2415277.78763936, 0, 0, 0, 0, 0, 0, -0.00899577590020672, 0, -0.00124549680356748, 0, 0, 0, -0.00479521812796501, 0, -0.0105996454157333, -388888.894040588, -388888.905453808, -2256944.45752691, 0, 0, -0.0147793155208322, -4527777.79196545, -6395833.36445862, -4527777.79836531, -6395833.37702229, -13263888.9290468, -13263888.9333543, -31152777.8494333, 0, 0, 0, 0, 0, 0, 0.0172452663289772, 0, 0.000980324105358766, 0, 0, 0, 0.00982885394174524, 0, 0.0170768177750582, 138888.895907208, 138888.924042751, 2083333.35215234, 0, 0, 0, 0.0252964106120953, 0, 0.0594852215188896, 4305555.57124667, 6250000.0589112, 4305555.59756301, 6250000.10013968, 26944444.5296888, 26944444.5391139, 8333333.3680773, 30972222.3382089, 30972222.3875555, 76666666.8478426, 76666666.8866214, 166111111.390191, 0, 0, 0, 0, 0, 0, -0.0192312504083043, 0, -0.000336452284643374, 0, 0, 0, -0.016183283921515, 0, -0.0100420562381317, -0.0060843548780388, -0.0451404681152369, -694444.45981421, 0, 0, 0, -0.0153663662154315, 0, -0.0956409914352141, -1388888.8943073, -2083333.39947717, -1388888.93965162, -2083333.46630068, -22916666.7633601, -22916666.7629179, 0, 0, -0.098021716511374, -4166666.68790446, -25694444.6098804, -4166666.70915215, -25694444.7139397, -45833333.5141575, -45833333.6370391, -68750000.2082019, -68750000.3430786, -317361111.566135, -133333333.691932, -361805556.158058, -361805556.216396, 0, 0, 0, 0, 0, 0, 0.0114377873537844, 0, 0.000122313309067776, 0, 0, 0, 0.014944435832887, 0, -0.000538048195412165, 0.00301134208889386, 0.0314871785669942, 0.00697411609058962, 0, 0, 0, 0.00256977298108377, 0, 0.0643506951656731, -0.0024414320520478, 0.0419609924803939, 0.0343941261158119, 0.0987095957327361, 6944444.50266443, 6944444.48450102, 0, 0, 0, 0.0590596630658255, 0, 0.124387743639066, -0.00127636762402711, 6944444.56627095, 0.0252780996387838, 6944444.65115312, 20833333.4641146, 20833333.5880575, 20833333.4444334, 20833333.6121571, 250000000.3324, 0.0131418370246515, 20833333.4986638, 20833333.6824237, 62500000.2185904, 277777778.31756, 62500000.3244537, 277777778.45368, 500000000.694845, 500000000.75118, 541666667.522918, 0, 0, 0, 0, 0, 0, -0.00279869293644852, 0, -6.97681785483236e-05, 0, 0, 0, -0.00534045460550327, 0, 0.0016832430774572, -0.00062837345402358, -0.00906464254198427, -0.00147273002897299, 0, 0, 0, -0.000348565546121741, 0, -0.0146817346523698, 0.00163441311136298, -0.0114803439989121, -0.0101110655662379, -0.031400748549693, -0.0148979865266386, -0.00309009229376015, 0, 0, 0, -0.0106763305776848, 0, -0.0488782096873212, 0.00393605305804513, -0.0375441601580166, -0.00550156048748874, -0.0636760711514736, -0.0374952593960373, -0.0755319249332694, -0.0225433363956755, -0.0931650759995492, -69444444.5338365, 0, 0, -0.0453130121612676, 0.00114030512007374, -0.0580995036343483, -0.0021788219179004, -0.130996745733608, -0.0553641738512266, -0.162973162929384, -0.0360865914306288, -69444444.6178834, -0.120036024779565, -69444444.7061391, -208333333.596657, -208333333.624242, -0.0760007656562759, -208333333.664024, -208333333.722598, -208333333.680326, -208333333.726905, -416666667.186618}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, -5.31771487385417e-07, 0, 0, 0, 0, 0, 0, 0, -2.19816927244086e-06, 0, 1.57657591166528e-05, -2500.00001627076, 0, 0, 0, 0, 0, 0, 5.11043059718017e-05, 0, -0.00014268719960002, 0, 0, -2.09174282762024e-05, 42777.7780324976, 30277.7780053426, 42777.7779377821, 0, 0, 0, 0, 0, 0, -0.000317288569957393, 0, 0.000789662262389045, 0, 0, 0, -0.000148397394010538, 0, -0.000174965885788762, -206944.445639663, -55555.5568064111, -206944.444560138, -512500.002363728, -726388.89201659, -512500.002498002, 0, 0, 0, 0, 0, 0, 0.00116107097628578, 0, -0.00259041751023819, 0, 0, 0, -3.62572651848802e-05, 0, 0.000258305950455774, 305555.558386974, 27777.7813602824, 305555.5528807, 0, 0, 0.00418937824390005, 2416666.67508147, 3451388.90239334, 888888.897689966, 888888.901808064, 3451388.90039711, 2416666.67452603, 8618055.58187141, 0, 0, 0, 0, 0, 0, -0.0025086527849153, 0, 0.00487911918994275, 0, 0, 0, 0.000895626583087171, 0, 0.00279367919621679, -138888.892760383, -0.00653250154397402, -138888.878692245, 0, 0, 0, -0.00894853487141662, 0, -0.0166062622462674, -3333333.34747001, -4861111.13893403, -416666.682094502, -416666.695156819, -4861111.12793257, -3333333.33676704, -3611111.13550096, -15694444.4897499, -3611111.15818792, -40000000.0807252, -14166666.7540377, -40000000.0842677, 0, 0, 0, 0, 0, 0, 0.00306584896458864, 0, -0.00516820356288022, 0, 0, 0, -0.000834533900176806, 0, -0.00834055792252116, 0.00312967629433828, 0.00790954267445868, -0.0157643292459872, 0, 0, 0, 0.00643115192876303, 0, 0.0230400499339379, 1388888.90093814, 2083333.36447498, 0.01345998415341, 0.0320467532762068, 2083333.34213045, 1388888.8686768, 0, 0, 0.0345300921971249, 2777777.80942519, 19444444.5193853, 1388888.91985503, 1388888.97834121, 19444444.5062177, 2777777.84624366, 52777777.8869619, 6250000.13358692, 52777777.8780322, 56944444.6284654, 178472222.441585, 56944444.6774411, 0, 0, 0, 0, 0, 0, -0.0019525764830413, 0, 0.00286123839527389, 0, 0, 0, -0.000199278949816053, 0, 0.00831851791183588, -0.00139927807581944, -0.00561727309860305, 0.0113824646968244, 0, 0, 0, -0.000946043979080154, 0, -0.0119069729898369, -0.0049285889001275, -0.0184005230081201, -0.00587656169423691, -0.0193140706157693, 0.00123677382199221, 0.0317035080988278, 0, 0, 0, -0.0251881062321595, 0, -0.0392801016393243, -0.0192634584740008, -6944444.5031302, -0.0156203514681736, -0.0696001895092789, -6944444.48146699, -0.0356654268508523, -20833333.4013853, -0.0942933278291242, -20833333.3726749, -0.0288211257700717, -13888888.9612576, -0.109088346360003, -41666666.8206123, -208333333.5546, -20833333.5112984, -20833333.5942406, -208333333.540558, -41666666.8833768, -222222222.574168, 0, 0, 0, 0, 0, 0, 0.000503186451136818, 0, -0.000643935049583327, 0, 0, 0, 0.000329764220247337, 0, -0.00283062990878284, 0.000266950183241147, 0.00169907293729786, -0.0031666844393348, 0, 0, 0, -0.00036416819035759, 0, 0.00134604795446007, 0.000734188254161431, 0.00452441165408614, 0.00118128194717347, 0.00524944837499555, -0.00185859206601474, -0.0136918976366552, 0, 0, 0, 0.00607227781217213, 0, 0.0138124290435268, 0.00443821856741701, 0.0179185498948427, 0.00190231187960366, 0.0195863307353071, 0.00819768360109593, 0.00175334607467736, 0.0159510723711875, 0.025213034507132, -0.00126432703381246, 0, 0, 0.0165426791727061, 0.0100441399408661, 0.0292788515149922, 0.00798664755600757, 0.0472054453779562, 0.0231371511298531, 0.0428829705167767, 0.0437489520700455, 69444444.522113, 0.059976042916539, 0.0999557893118675, 69444444.505818, 0.0612567930493114, 0.0717017942254119, 138888889.00404, 0.135769084021578, 138888889.051552, 69444444.6262282, 138888889.063956}, 
                                                     {0, 0, 0, 0, 0, 0, 0, 0, 2.52628193467687e-06, 0, 0, 0, 0, 0, 0, 0, -3.91496138977433e-05, 0, -4.05528971850963e-05, -2500.00003960825, 0, 0, 0, 0, 0, 0, 0.00031372736281363, 0, 0.00032169192077459, 0, 0, 0.000415807999491029, 30277.7781001942, 42777.7782813733, 42777.7783294931, 0, 0, 0, 0, 0, 0, -0.00138085375137825, 0, -0.00138886244977529, 0, 0, 0, -0.0024634306578715, 0, -0.0024853670377029, -55555.5569102478, -206944.44745463, -206944.447744488, -512500.00343077, -512500.003548469, -726388.894199133, 0, 0, 0, 0, 0, 0, 0.00368952112754613, 0, 0.00355585875354852, 0, 0, 0, 0.00751486531894491, 0, 0.00786756135194029, 27777.7811557597, 305555.565146664, 305555.566371207, 0, 0, 0.0121922424250116, 888888.900096088, 888888.900752061, 2416666.68280464, 3451388.91445565, 2416666.68305024, 3451388.9122574, 8618055.58409537, 0, 0, 0, 0, 0, 0, -0.00615440573047148, 0, -0.00564466580961093, 0, 0, 0, -0.0131367010704352, 0, -0.0134053119330837, -0.00526886069983779, -138888.906444183, -138888.909857056, 0, 0, 0, -0.0284662870816737, 0, -0.0322588538042053, -416666.686547483, -416666.690012057, -3333333.37292242, -4861111.17418966, -3333333.37294477, -4861111.16385536, -3611111.15232456, -3611111.15280885, -15694444.5334493, -14166666.7372074, -40000000.1007484, -40000000.0992492, 0, 0, 0, 0, 0, 0, 0.00622392553208137, 0, 0.00549566537171452, 0, 0, 0, 0.0138783870386153, 0, 0.0125229355568012, 0.00501366146069861, 0.0186621254943767, 0.0241390800027596, 0, 0, 0, 0.0331097866549274, 0, 0.042755318599043, 0.0203693465551381, 0.0275780747165463, 1388888.94061268, 2083333.41654918, 1388888.94260809, 2083333.39638454, 0, 0, 0.0594272649898959, 1388888.94223426, 1388888.94793194, 2777777.85440227, 19444444.6057998, 2777777.84741404, 19444444.5975783, 6250000.08937996, 52777777.9604252, 52777777.943955, 56944444.6135201, 56944444.6368402, 178472222.470864, 0, 0, 0, 0, 0, 0, -0.00347084174949774, 0, -0.0030060609294714, 0, 0, 0, -0.00836688633581641, 0, -0.0061838438659329, -0.00262609925757378, -0.0107931109924047, -0.0152813916467756, 0, 0, 0, -0.0192965882297534, 0, -0.0267269003688895, -0.0115561401751939, -0.0180178184958949, -0.0337631495286648, -0.0562163223930741, -0.0391107870152795, -0.038040140457418, 0, 0, 0, -0.0468925881033151, 0, -0.0564528236446317, -0.035079974603442, -0.0446765475860031, -0.0678579557069396, -6944444.58092359, -0.057246027619746, -6944444.56870039, -0.0580663335443503, -20833333.4971184, -20833333.465728, -0.0670985117436098, -0.0608227410291173, -13888889.0770817, -20833333.4591472, -20833333.5055806, -41666666.8642223, -208333333.615535, -41666666.885735, -208333333.607202, -222222222.507778, 0, 0, 0, 0, 0, 0, 0.000815534564129747, 0, 0.00070443758566508, 0, 0, 0, 0.00220227485352783, 0, 0.00131062588063927, 0.000572602125175157, 0.00264509868763576, 0.00408598761133948, 0, 0, 0, 0.0046474580969763, 0, 0.00622520613928238, 0.00281634478756619, 0.00494948961254201, 0.00863190869593778, 0.0153265241747461, 0.0119054254315109, 0.00926878290313146, 0, 0, 0, 0.0132992130734961, 0, 0.0190969757226591, 0.00954079086235602, 0.0140802310161289, 0.0223398550758523, 0.0439551810839409, 0.0191616455971295, 0.0371985414169978, 0.0157706807892406, 0.056474232719696, 0.0406062182803679, 0, 0, 0.023516496503512, 0.0222058475559071, 0.021536139317417, 0.0304421914938748, 0.0787144033564238, 0.0224589259992954, 0.0798761610597883, 0.0346955907007597, 0.0583706796600678, 0.0899482768271552, 69444444.5652305, 0.086876535632621, 69444444.5536723, 0.070004289533568, 0.100687297257254, 138888889.039759, 69444444.5587, 138888889.036778, 138888889.05097}};


