// -*- C++ -*- 
// -----------------------------------------------------------------------------
// MAdLiB - Copyright (C) Cenaero
//
// See the Copyright.txt and License.txt files for license information. 
// ------------------------------------------------------------------

#include "MAd_TriInterpolant.h"

#include <cmath>
using std::abs;

// -----------------------------------------------------------------------------

double MAd_TriInterpolant::getCoef(int order,int i,int j){
  switch(order) {
  case 0: return coef0[i][j];
  case 1: return coef1[i][j];
  case 2: return coef2[i][j];
  case 3: return coef3[i][j];
  case 4: return coef4[i][j];
  case 5: return coef5[i][j];
  case 6: return coef6[i][j];
  case 7: return coef7[i][j];
  case 8: return coef8[i][j];
  case 9: return coef9[i][j];
  case 10: return coef10[i][j];
  case 11: return coef11[i][j];
  default : throw; break;
  }         
}

// -----------------------------------------------------------------------------

MAd_TriInterpolant::MAd_TriInterpolant(const int &o)
  :MAd_Interpolant(MSH_TRI_3, o), ndof(size(o)) 
{
  
  coef.resize(size(),size());
  for (int i=0; i< ndof; i++) for (int j=0; j< ndof; j++) coef(i,j) = getCoef(o,i,j);
  
  pp = P;
  
  NodeNumberu = new double[ndof];
  NodeNumberv = new double[ndof];
  nodePositionTriangle(order(), NodeNumberu, NodeNumberv);
}

// -----------------------------------------------------------------------------

MAd_TriInterpolant::~MAd_TriInterpolant(){    
  delete []NodeNumberu;
  delete []NodeNumberv;
  return;
}
// -----------------------------------------------------------------------------
void MAd_TriInterpolant::fcts (const double &u, const double &v, const double &w, 
                                     doubleVector &f) const
{
  doubleVector m(ndof);
  f = 0.;
  for (int i=0; i< ndof ; i++) m(i) = monomeof(u,v,i);
  coef.mult(m,f,false);
  return;
}
// -----------------------------------------------------------------------------
void MAd_TriInterpolant::grads  (const double &u, const double &v, const double &w, 
                                       doubleVector gr[3])const
{
  doubleVector m(ndof);
  for (int k=0; k< 2; k++){ 
    gr[k] = 0.;
    for (int i=0; i< ndof ; i++) m(i) = dmonomeof(u,v,i,k);
    coef.mult(m,gr[k],false);
  }
  return;
}

void MAd_TriInterpolant::nodeNumberCoordinate(const int & i, double &u, double &v, double &w) const{

  u= NodeNumberu[i];
  v= NodeNumberv[i];
  w=0.;
  return;
}

int MAd_TriInterpolant::size(int q){
  return (q+1)*(q+2)/2;
}

bool MAd_TriInterpolant::reorientData(doubleMatrix& sol,int orientation) const {
  
  if (sol.size1() != size())  return false;

  doubleMatrix tmp(sol);
  
  if (orientation < 0) {
    int n=0;
    int locOrder = order();
    while (locOrder>=0){
      for (int j=0; j<sol.size2(); j++) {
        if (locOrder==0) tmp(n,j) = sol(n,j);
        else if (locOrder>2){
          // vertices
          tmp(n,j) = sol(n,j); tmp(n+1,j) = sol(n+2,j); tmp(n+2,j) = sol(n+1,j);
          // edges
          int i=0;
          for (int l=3; l>0; l--){
            for (int k=0; k<(locOrder-1); k++){
              tmp(n+3+i,j) = sol(n+3+l*(locOrder-1)-(k+1),j);
              i++;
            }
          }
        }
      }
      // inside, recursively
      n = n+3*locOrder;
      locOrder = locOrder-3;
    }
    sol = tmp;
  }
  
  for (int i=1;i<abs(orientation);i++) {

    int n=0;
    int locOrder = order();
    while (locOrder>=0){
      for (int j=0; j<sol.size2(); j++) {
        if (locOrder==0) tmp(n,j) = sol(n,j);
        else if (locOrder>2){
          // vertices
          for (int vert=0; vert<3; vert++) tmp(n+vert,j) = sol(n+(vert+2)%3,j);
          // edges
          for (int edge=0; edge<3; edge++){
            for (int k=0; k<(locOrder-1); k++) tmp(n+3+edge*(locOrder-1)+k,j) = sol(n+3+((edge+2)%3)*(locOrder-1)+k,j);
          }
        }
      }
      // inside, recursively
      n = n+3*locOrder;
      locOrder = locOrder-3;
    }
    sol = tmp;
  }
  return true;
}

// -----------------------------------------------------------------------------

bool MAd_TriInterpolant::findOriginalCoordinates(int iRot,int iSwp,
                                                       const doubleMatrix& uvwTfo,
                                                       doubleMatrix& uvwOri) const {
  

  if (abs(iSwp) != 1) return false;
  if (iRot > 2 || iRot < 0) return false;

  const doubleVector uTfo(uvwTfo.seeColumn(0));
  const doubleVector vTfo(uvwTfo.seeColumn(1));
  
  doubleMatrix uvwTmp(uvwOri.size1(),2);
  
  doubleVector uTmp(uvwTmp.touchColumn(0));
  doubleVector vTmp(uvwTmp.touchColumn(1));

  switch(iSwp) {
  case 1:
    uTmp = uTfo;
    vTmp = vTfo;
    break;
 
  case -1:
    uTmp = vTfo;
    vTmp = uTfo;
    break;
  default:
    return false;
  }
  
  doubleVector uOri(uvwOri.touchColumn(0));
  doubleVector vOri(uvwOri.touchColumn(1));
  
  switch (iRot) {
  case 0:
    uOri = uTmp;
    vOri = vTmp;
    break;
  case 1:
    uOri = 1;
    uTmp *= -1;
    uOri += uTmp;
    vTmp *= -1;
    uOri += vTmp;
    vOri = uTmp;
    break;
  case 2:
    uOri = vTmp;
    vOri = 1;
    uTmp *=-1;
    vOri += uTmp;
    vTmp *=-1;
    vOri += vTmp;
    break;
  default:
    return false;
  }
  return true;
}



//-------------------------------
//  Private Members
//-------------------------------

double MAd_TriInterpolant::monomeof(const double &u, const double &v, int i) const{
  return pow(u,pp[i][0])*pow(v,pp[i][1]);
};

double MAd_TriInterpolant::dmonomeof(const double &u, const double &v, int i, int k) const{
  double Zero=0.;
  switch (k){
  case 0:
    if (pp[i][0] > 0)
      return pp[i][0] * pow(u,pp[i][0]-1)*pow(v,pp[i][1]);
    else return Zero;
    break;
  case 1:    
    if (pp[i][1] > 0)
      return pp[i][1] * pow(u,pp[i][0])*pow(v,pp[i][1]-1);
    else return Zero;  
    break;
  default:
    throw;
    break;
  }
};

int MAd_TriInterpolant::getOrientationCode(int facenumber, int dir) const 
{
  if (dir >0)
    return facenumber*2;
  else
    return facenumber*2+1;
}


double MAd_TriInterpolant::coef0[1][1] = {{1}};




double MAd_TriInterpolant::coef1[3][3] = {{1, -1, -1}, 
                                                {0, 1, 0}, 
                                                {0, 0, 1}};




double MAd_TriInterpolant::coef2[6][6] = {{1, -3, -3, 2, 2, 4}, 
                                                {0, -1, 0, 2, 0, 0}, 
                                                {0, 0, -1, 0, 2, 0}, 
                                                {0, 4, 0, -4, 0, -4}, 
                                                {0, 0, 0, 0, 0, 4}, 
                                                {0, 0, 4, 0, -4, -4}};




double MAd_TriInterpolant::coef3[10][10] = {{1, -5.5, -5.5, 9, 9, 18, -4.5, -4.5, -13.5, -13.5}, 
                                                  {0, 1, 0, -4.5, 0, 0, 4.5, 0, 0, 0}, 
                                                  {0, 0, 1, 0, -4.5, 0, 0, 4.5, 0, 0}, 
                                                  {0, 9, 0, -22.5, 0, -22.5, 13.5, 0, 27, 13.5}, 
                                                  {0, -4.5, 0, 18, 0, 4.5, -13.5, 0, -13.5, 0}, 
                                                  {0, 0, 0, 0, 0, -4.5, 0, 0, 13.5, 0}, 
                                                  {0, 0, 0, 0, 0, -4.5, 0, 0, 0, 13.5}, 
                                                  {0, 0, -4.5, 0, 18, 4.5, 0, -13.5, 0, -13.5}, 
                                                  {0, 0, 9, 0, -22.5, -22.5, 0, 13.5, 13.5, 27}, 
                                                  {0, 0, 0, 0, 0, 27, 0, 0, -27, -27}};




double MAd_TriInterpolant::coef4[15][15] = {
  {1, -8.33333333333333, -8.33333333333333, 23.3333333333333, 23.3333333333333, 46.6666666666667, -26.6666666666667, -26.6666666666667, -80, -80, 10.6666666666667, 10.6666666666667, 42.6666666666667, 42.6666666666667, 64}, 
  {0, -1, 0, 7.33333333333333, 0, 0, -16, 0, 0, 0, 10.6666666666667, 0, 0, 0, 0}, 
  {0, 0, -1, 0, 7.33333333333333, 0, 0, -16, 0, 0, 0, 10.6666666666667, 0, 0, 0}, 
  {0, 16, 0, -69.3333333333333, 0, -69.3333333333333, 96, 0, 192, 96, -42.6666666666667, 0, -128, -42.6666666666667, -128}, 
  {0, -12, 0, 76, 0, 28, -128, 0, -144, -16, 64, 0, 128, 0, 64}, 
  {0, 5.33333333333333, 0, -37.3333333333333, 0, -5.33333333333333, 74.6666666666667, 0, 32, 0, -42.6666666666667, 0, -42.6666666666667, 0, 0}, 
  {0, 0, 0, 0, 0, 5.33333333333333, 0, 0, -32, 0, 0, 0, 42.6666666666667, 0, 0}, 
  {0, 0, 0, 0, 0, 4, 0, 0, -16, -16, 0, 0, 0, 0, 64}, 
  {0, 0, 0, 0, 0, 5.33333333333333, 0, 0, 0, -32, 0, 0, 0, 42.6666666666667, 0}, 
  {0, 0, 5.33333333333333, 0, -37.3333333333333, -5.33333333333333, 0, 74.6666666666667, 0, 32, 0, -42.6666666666667, 0, -42.6666666666667, 0}, 
  {0, 0, -12, 0, 76, 28, 0, -128, -16, -144, 0, 64, 0, 128, 64}, 
  {0, 0, 16, 0, -69.3333333333333, -69.3333333333333, 0, 96, 96, 192, 0, -42.6666666666667, -42.6666666666667, -128, -128}, 
  {0, 0, 0, 0, 0, 96, 0, 0, -224, -224, 0, 0, 128, 128, 256}, 
  {0, 0, 0, 0, 0, -32, 0, 0, 160, 32, 0, 0, -128, 0, -128}, 
  {0, 0, 0, 0, 0, -32, 0, 0, 32, 160, 0, 0, 0, -128, -128}};




double MAd_TriInterpolant::coef5[21][21] = {{1, -11.4166666666667, -11.4166666666667, 46.875, 46.875, 93.75, -88.5416666666667, -88.5416666666667, -265.625, -265.625, 78.125, 78.125, 312.5, 312.5, 468.75, -26.0416666666667, -26.0416666666667, -130.208333333333, -130.208333333333, -260.416666666667, -260.416666666667}, 
                                                  {0, 1, 0, -10.4166666666667, 0, 0, 36.4583333333333, 0, 0, 0, -52.0833333333333, 0, 0, 0, 0, 26.0416666666667, 0, 0, 0, 0, 0}, 
                                                  {0, 0, 1, 0, -10.4166666666667, 0, 0, 36.4583333333333, 0, 0, 0, -52.0833333333333, 0, 0, 0, 0, 26.0416666666667, 0, 0, 0, 0}, 
                                                  {0, 25, 0, -160.416666666667, 0, -160.416666666667, 369.791666666667, 0, 739.583333333333, 369.791666666667, -364.583333333333, 0, -1093.75, -364.583333333333, -1093.75, 130.208333333333, 0, 520.833333333333, 130.208333333333, 781.25, 520.833333333333}, 
                                                  {0, -25, 0, 222.916666666667, 0, 97.9166666666667, -614.583333333333, 0, -739.583333333333, -125, 677.083333333333, 0, 1406.25, 52.0833333333333, 781.25, -260.416666666667, 0, -781.25, 0, -781.25, -260.416666666667}, 
                                                  {0, 16.6666666666667, 0, -162.5, 0, -37.5, 510.416666666667, 0, 322.916666666667, 20.8333333333333, -625, 0, -781.25, 0, -156.25, 260.416666666667, 0, 520.833333333333, 0, 260.416666666667, 0}, 
                                                  {0, -6.25, 0, 63.5416666666667, 0, 6.25, -213.541666666667, 0, -57.2916666666667, 0, 286.458333333333, 0, 156.25, 0, 0, -130.208333333333, 0, -130.208333333333, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, -6.25, 0, 0, 57.2916666666667, 0, 0, 0, -156.25, 0, 0, 0, 0, 130.208333333333, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, -4.16666666666667, 0, 0, 31.25, 20.8333333333333, 0, 0, -52.0833333333333, 0, -156.25, 0, 0, 0, 0, 260.416666666667, 0}, 
                                                  {0, 0, 0, 0, 0, -4.16666666666667, 0, 0, 20.8333333333333, 31.25, 0, 0, 0, -52.0833333333333, -156.25, 0, 0, 0, 0, 0, 260.416666666667}, 
                                                  {0, 0, 0, 0, 0, -6.25, 0, 0, 0, 57.2916666666667, 0, 0, 0, -156.25, 0, 0, 0, 0, 130.208333333333, 0, 0}, 
                                                  {0, 0, -6.25, 0, 63.5416666666667, 6.25, 0, -213.541666666667, 0, -57.2916666666667, 0, 286.458333333333, 0, 156.25, 0, 0, -130.208333333333, 0, -130.208333333333, 0, 0}, 
                                                  {0, 0, 16.6666666666667, 0, -162.5, -37.5, 0, 510.416666666667, 20.8333333333333, 322.916666666667, 0, -625, 0, -781.25, -156.25, 0, 260.416666666667, 0, 520.833333333333, 0, 260.416666666667}, 
                                                  {0, 0, -25, 0, 222.916666666667, 97.9166666666667, 0, -614.583333333333, -125, -739.583333333333, 0, 677.083333333333, 52.0833333333333, 1406.25, 781.25, 0, -260.416666666667, 0, -781.25, -260.416666666667, -781.25}, 
                                                  {0, 0, 25, 0, -160.416666666667, -160.416666666667, 0, 369.791666666667, 369.791666666667, 739.583333333333, 0, -364.583333333333, -364.583333333333, -1093.75, -1093.75, 0, 130.208333333333, 130.208333333333, 520.833333333333, 520.833333333333, 781.25}, 
                                                  {0, 0, 0, 0, 0, 250, 0, 0, -979.166666666667, -979.166666666667, 0, 0, 1250, 1250, 2500, 0, 0, -520.833333333333, -520.833333333333, -1562.5, -1562.5}, 
                                                  {0, 0, 0, 0, 0, 41.6666666666667, 0, 0, -354.166666666667, -41.6666666666667, 0, 0, 833.333333333333, 0, 312.5, 0, 0, -520.833333333333, 0, -520.833333333333, 0}, 
                                                  {0, 0, 0, 0, 0, 41.6666666666667, 0, 0, -41.6666666666667, -354.166666666667, 0, 0, 0, 833.333333333333, 312.5, 0, 0, 0, -520.833333333333, 0, -520.833333333333}, 
                                                  {0, 0, 0, 0, 0, -125, 0, 0, 906.25, 281.25, 0, 0, -1562.5, -156.25, -1718.75, 0, 0, 781.25, 0, 1562.5, 781.25}, 
                                                  {0, 0, 0, 0, 0, 31.25, 0, 0, -187.5, -187.5, 0, 0, 156.25, 156.25, 1093.75, 0, 0, 0, 0, -781.25, -781.25}, 
                                                  {0, 0, 0, 0, 0, -125, 0, 0, 281.25, 906.25, 0, 0, -156.25, -1562.5, -1718.75, 0, 0, 0, 781.25, 781.25, 1562.5}};




double MAd_TriInterpolant::coef6[28][28] = {{1, -14.7, -14.7, 81.2, 81.2, 162.4, -220.5, -220.5, -661.5, -661.5, 315, 315, 1260, 1260, 1890, -226.8, -226.8, -1134, -1134, -2268, -2268, 64.8, 64.8, 388.8, 388.8, 972, 972, 1296}, 
                                                  {0, -1, 0, 13.7, 0, 0, -67.5, 0, 0, 0, 153, 0, 0, 0, 0, -162, 0, 0, 0, 0, 0, 64.8, 0, 0, 0, 0, 0, 0}, 
                                                  {0, 0, -1, 0, 13.7, 0, 0, -67.5, 0, 0, 0, 153, 0, 0, 0, 0, -162, 0, 0, 0, 0, 0, 64.8, 0, 0, 0, 0, 0}, 
                                                  {0, 36, 0, -313.2, 0, -313.2, 1044, 0, 2088, 1044, -1674, 0, -5022, -1674, -5022, 1296, 0, 5184, 1296, 7776, 5184, -388.8, 0, -1944, -388.8, -3888, -1944, -3888}, 
                                                  {0, -45, 0, 526.5, 0, 256.5, -2074.5, 0, -2610, -535.5, 3699, 0, 7884, 486, 4671, -3078, 0, -9396, -162, -9720, -3564, 972, 0, 3888, 0, 5832, 972, 3888}, 
                                                  {0, 40, 0, -508, 0, -148, 2232, 0, 1692, 180, -4356, 0, -6120, -72, -1836, 3888, 0, 8424, 0, 5184, 648, -1296, 0, -3888, 0, -3888, 0, -1296}, 
                                                  {0, -22.5, 0, 297, 0, 49.5, -1381.5, 0, -598.5, -27, 2889, 0, 2376, 0, 297, -2754, 0, -3726, 0, -972, 0, 972, 0, 1944, 0, 972, 0, 0}, 
                                                  {0, 7.2, 0, -97.2, 0, -7.2, 468, 0, 90, 0, -1026, 0, -378, 0, 0, 1036.8, 0, 648, 0, 0, 0, -388.8, 0, -388.8, 0, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, 7.2, 0, 0, -90, 0, 0, 0, 378, 0, 0, 0, 0, -648, 0, 0, 0, 0, 0, 388.8, 0, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, 4.5, 0, 0, -49.5, -27, 0, 0, 162, 0, 297, 0, 0, -162, 0, -972, 0, 0, 0, 0, 0, 972, 0, 0}, 
                                                  {0, 0, 0, 0, 0, 4, 0, 0, -36, -36, 0, 0, 72, 72, 324, 0, 0, 0, 0, -648, -648, 0, 0, 0, 0, 0, 0, 1296}, 
                                                  {0, 0, 0, 0, 0, 4.5, 0, 0, -27, -49.5, 0, 0, 0, 162, 297, 0, 0, 0, -162, 0, -972, 0, 0, 0, 0, 0, 972, 0}, 
                                                  {0, 0, 0, 0, 0, 7.2, 0, 0, 0, -90, 0, 0, 0, 378, 0, 0, 0, 0, -648, 0, 0, 0, 0, 0, 388.8, 0, 0, 0}, 
                                                  {0, 0, 7.2, 0, -97.2, -7.2, 0, 468, 0, 90, 0, -1026, 0, -378, 0, 0, 1036.8, 0, 648, 0, 0, 0, -388.8, 0, -388.8, 0, 0, 0}, 
                                                  {0, 0, -22.5, 0, 297, 49.5, 0, -1381.5, -27.0000000000001, -598.5, 0, 2889, 0, 2376, 297, 0, -2754, 0, -3726, 0, -972, 0, 972, 0, 1944, 0, 972, 0}, 
                                                  {0, 0, 40, 0, -508, -148, 0, 2232, 180, 1692, 0, -4356, -72.0000000000002, -6120, -1836, 0, 3888, 0, 8424, 648.000000000001, 5184, 0, -1296, 0, -3888, 0, -3888, -1296}, 
                                                  {0, 0, -45, 0, 526.5, 256.5, 0, -2074.5, -535.5, -2610, 0, 3699, 486, 7884, 4671, 0, -3078, -162, -9396, -3564, -9720, 0, 972, 0, 3888, 972, 5832, 3888}, 
                                                  {0, 0, 36, 0, -313.2, -313.2, 0, 1044, 1044, 2088, 0, -1674, -1674, -5022, -5022, 0, 1296, 1296, 5184, 5184, 7776, 0, -388.8, -388.8, -1944, -1944, -3888, -3888}, 
                                                  {0, 0, 0, 0, 0, 540, 0, 0, -3078, -3078, 0, 0, 6426, 6426, 12852, 0, 0, -5832, -5832, -17496, -17496, 0, 0, 1944, 1944, 7776, 7776, 11664}, 
                                                  {0, 0, 0, 0, 0, -54, 0, 0, 648, 54, 0, 0, -2538, 0, -594, 0, 0, 3888, 0, 1944, 0, 0, 0, -1944, 0, -1944, 0, 0}, 
                                                  {0, 0, 0, 0, 0, -54, 0, 0, 54, 648, 0, 0, 0, -2538, -594, 0, 0, 0, 3888, 0, 1944, 0, 0, 0, -1944, 0, -1944, 0}, 
                                                  {0, 0, 0, 0, 0, -360, 0, 0, 3492, 1332, 0, 0, -9612, -1620, -11232, 0, 0, 10368, 648, 21384, 11664, 0, 0, -3888, 0, -11664, -3888, -11664}, 
                                                  {0, 0, 0, 0, 0, 180, 0, 0, -2016, -396, 0, 0, 7020, 216, 3996, 0, 0, -9072, 0, -11016, -1944, 0, 0, 3888, 0, 7776, 0, 3888}, 
                                                  {0, 0, 0, 0, 0, -36, 0, 0, 360, 252, 0, 0, -972, -216, -2484, 0, 0, 648, 0, 6480, 1944, 0, 0, 0, 0, -3888, 0, -3888}, 
                                                  {0, 0, 0, 0, 0, -36, 0, 0, 252, 360, 0, 0, -216, -972, -2484, 0, 0, 0, 648, 1944, 6480, 0, 0, 0, 0, 0, -3888, -3888}, 
                                                  {0, 0, 0, 0, 0, 180, 0, 0, -396, -2016, 0, 0, 216, 7020, 3996, 0, 0, 0, -9072, -1944, -11016, 0, 0, 0, 3888, 0, 7776, 3888}, 
                                                  {0, 0, 0, 0, 0, -360, 0, 0, 1332, 3492, 0, 0, -1620, -9612, -11232, 0, 0, 648, 10368, 11664, 21384, 0, 0, 0, -3888, -3888, -11664, -11664}, 
                                                  {0, 0, 0, 0, 0, 135, 0, 0, -1107, -1107, 0, 0, 1944, 1944, 8748, 0, 0, -972, -972, -13608, -13608, 0, 0, 0, 0, 5832, 5832, 11664}};




double MAd_TriInterpolant::coef7[36][36] = {{1, -18.15, -18.15, 127.672222222222, 127.672222222222, 255.344444444444, -460.668055555556, -460.668055555555, -1382.00416666667, -1382.00416666667, 933.722222222222, 933.722222222222, 3734.88888888889, 3734.88888888889, 5602.33333333333, -1073.78055555556, -1073.78055555556, -5368.90277777778, -5368.90277777778, -10737.8055555556, -10737.8055555556, 653.605555555555, 653.605555555555, 3921.63333333333, 3921.63333333333, 9804.08333333333, 9804.08333333333, 13072.1111111111, -163.401388888889, -163.401388888889, -1143.80972222222, -1143.80972222222, -3431.42916666667, -3431.42916666667, -5719.04861111111, -5719.04861111111}, 
                                                  {0, 1, 0, -17.15, 0, 0, 110.522222222222, 0, 0, 0, -350.145833333333, 0, 0, 0, 0, 583.576388888889, 0, 0, 0, 0, 0, -490.204166666667, 0, 0, 0, 0, 0, 0, 163.401388888889, 0, 0, 0, 0, 0, 0, 0}, 
                                                  {0, 0, 0.999999999999999, 0, -17.15, 0, 0, 110.522222222222, 0, 0, 0, -350.145833333333, 0, 0, 0, 0, 583.576388888889, 0, 0, 0, 0, 0, -490.204166666667, 0, 0, 0, 0, 0, 0, 163.401388888889, 0, 0, 0, 0, 0, 0}, 
                                                  {0, 49, 0, -546.35, 0, -546.35, 2431.48888888889, 0, 4862.97777777778, 2431.48888888889, -5552.3125, 0, -16656.9375, -5552.3125, -16656.9375, 6886.20138888889, 0, 27544.8055555556, 6886.20138888889, 41317.2083333333, 27544.8055555556, -4411.8375, 0, -22059.1875, -4411.8375, -44118.375, -22059.1875, -44118.375, 1143.80972222222, 0, 6862.85833333333, 1143.80972222222, 17157.1458333333, 6862.85833333333, 22876.1944444444, 17157.1458333333}, 
                                                  {0, -73.5, 0, 1076.775, 0, 562.275, -5615.19583333333, 0, -7294.46666666667, -1679.27083333333, 14205.9166666667, 0, 30862.8541666667, 2451.02083333333, 19107.9583333333, -18907.875, 0, -58474.3541666667, -1750.72916666667, -61975.8125, -24160.0625, 12745.3083333333, 0, 51471.4375, 490.204166666667, 78432.6666666667, 14706.125, 53922.4583333333, -3431.42916666667, 0, -17157.1458333333, 0, -34314.2916666667, -3431.42916666666, -34314.2916666667, -17157.1458333333}, 
                                                  {0, 81.6666666666667, 0, -1291.69444444444, 0, -434.194444444444, 7412.61111111111, 0, 6264.51388888889, 852.736111111111, -20325.1319444444, 0, -30746.1388888889, -733.638888888888, -11154.6458333333, 28828.6736111111, 0, 65827.4166666667, 233.430555555555, 45402.2430555555, 8636.93055555555, -20425.1736111111, 0, -63726.5416666667, 0, -68628.5833333333, -2451.02083333333, -27778.2361111111, 5719.04861111111, 0, 22876.1944444444, 0, 34314.2916666667, 0, 22876.1944444444, 5719.04861111111}, 
                                                  {0, -61.25, 0, 1004.5, 0, 218.458333333333, -6062.04861111111, 0, -3318.04861111111, -257.25, 17607.3333333333, 0, 17607.3333333333, 100.041666666666, 3601.5, -26377.6527777778, 0, -41550.6388888889, 0, -16456.854166667, -1283.86805555555, 19608.1666666667, 0, 44118.375, 0, 29412.25, 0, 4902.04166666666, -5719.04861111111, 0, -17157.1458333333, 0, -17157.1458333333, 0, -5719.04861111111, 0}, 
                                                  {0, 29.4, 0, -492.45, 0, -63.7, 3064.13333333333, 0, 997.558333333333, 34.3, -9253.85416666667, 0, -5552.3125, 0, -500.208333333333, 14496.0375, 0, 14005.8333333333, 0, 2451.02083333333, 0, -11274.6958333333, 0, -16176.7375, 0, -4902.04166666667, 0, 0, 3431.42916666667, 0, 6862.85833333333, 0, 3431.42916666667, 0, 0, 0}, 
                                                  {0, -8.16666666666667, 0, 138.697222222222, 0, 8.16666666666667, -880.843055555556, 0, -130.530555555556, 0, 2734.47222222222, 0, 750.3125, 0, 0, -4435.18055555556, 0, -1984.15972222222, 0, 0, 0, 3594.83055555556, 0, 2451.02083333333, 0, 0, 0, 0, -1143.80972222222, 0, -1143.80972222222, 0, 0, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, -8.16666666666667, 0, 0, 130.530555555556, 0, 0, 0, -750.3125, 0, 0, 0, 0, 1984.15972222222, 0, 0, 0, 0, 0, -2451.02083333333, 0, 0, 0, 0, 0, 0, 1143.80972222222, 0, 0, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, -4.9, 0, 0, 71.4583333333333, 34.3, 0, 0, -350.145833333333, 0, -500.208333333333, 0, 0, 700.291666666667, 0, 2451.02083333333, 0, 0, 0, -490.204166666667, 0, -4902.04166666667, 0, 0, 0, 0, 0, 0, 3431.42916666667, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, -4.08333333333333, 0, 0, 52.4027777777778, 42.875, 0, 0, -200.083333333333, -100.041666666667, -550.229166666667, 0, 0, 233.430555555556, 0, 2100.875, 1283.86805555556, 0, 0, 0, 0, -2451.02083333333, 0, -4902.04166666667, 0, 0, 0, 0, 0, 0, 5719.04861111111, 0}, 
                                                  {0, 0, 0, 0, 0, -4.08333333333333, 0, 0, 42.875, 52.4027777777778, 0, 0, -100.041666666667, -200.083333333333, -550.229166666667, 0, 0, 0, 233.430555555556, 1283.86805555556, 2100.875, 0, 0, 0, 0, 0, -2451.02083333333, -4902.04166666667, 0, 0, 0, 0, 0, 0, 0, 5719.04861111111}, 
                                                  {0, 0, 0, 0, 0, -4.9, 0, 0, 34.3, 71.4583333333333, 0, 0, 0, -350.145833333333, -500.208333333333, 0, 0, 0, 700.291666666667, 0, 2451.02083333333, 0, 0, 0, -490.204166666667, 0, -4902.04166666667, 0, 0, 0, 0, 0, 0, 3431.42916666667, 0, 0}, 
                                                  {0, 0, 0, 0, 0, -8.16666666666667, 0, 0, 0, 130.530555555556, 0, 0, 0, -750.3125, 0, 0, 0, 0, 1984.15972222222, 0, 0, 0, 0, 0, -2451.02083333333, 0, 0, 0, 0, 0, 0, 1143.80972222222, 0, 0, 0, 0}, 
                                                  {0, 0, -8.16666666666666, 0, 138.697222222222, 8.16666666666664, 0, -880.843055555555, 0, -130.530555555555, 0, 2734.47222222222, 0, 750.312499999999, 0, 0, -4435.18055555555, 0, -1984.15972222222, 0, 0, 0, 3594.83055555555, 0, 2451.02083333333, 0, 0, 0, 0, -1143.80972222222, 0, -1143.80972222222, 0, 0, 0, 0}, 
                                                  {0, 0, 29.4, 0, -492.45, -63.6999999999999, 0, 3064.13333333333, 34.2999999999998, 997.558333333332, 0, -9253.85416666666, 0, -5552.3125, -500.20833333333, 0, 14496.0375, 0, 14005.8333333333, 0, 2451.02083333332, 0, -11274.6958333333, 0, -16176.7375, 0, -4902.04166666666, 0, 0, 3431.42916666667, 0, 6862.85833333333, 0, 3431.42916666666, 0, 0}, 
                                                  {0, 0, -61.25, 0, 1004.5, 218.458333333333, 0, -6062.04861111111, -257.25, -3318.04861111111, 0, 17607.3333333333, 100.041666666667, 17607.3333333333, 3601.49999999999, 0, -26377.6527777778, 0, -41550.6388888889, -1283.86805555554, -16456.8541666666, 0, 19608.1666666667, 0, 44118.375, 0, 29412.25, 4902.04166666664, 0, -5719.04861111111, 0, -17157.1458333333, 0, -17157.1458333333, 0, -5719.0486111111}, 
                                                  {0, 0, 81.6666666666666, 0, -1291.69444444444, -434.194444444444, 0, 7412.61111111111, 852.736111111111, 6264.51388888889, 0, -20325.1319444444, -733.638888888889, -30746.1388888889, -11154.6458333333, 0, 28828.6736111111, 233.430555555558, 65827.4166666666, 8636.93055555554, 45402.2430555555, 0, -20425.1736111111, 0, -63726.5416666666, -2451.02083333332, -68628.5833333333, -27778.2361111111, 0, 5719.04861111111, 0, 22876.1944444444, 0, 34314.2916666667, 5719.04861111109, 22876.1944444444}, 
                                                  {0, 0, -73.5, 0, 1076.775, 562.275, 0, -5615.19583333333, -1679.27083333333, -7294.46666666667, 0, 14205.9166666667, 2451.02083333333, 30862.8541666667, 19107.9583333333, 0, -18907.875, -1750.72916666667, -58474.3541666667, -24160.0625, -61975.8125, 0, 12745.3083333333, 490.20416666667, 51471.4375, 14706.125, 78432.6666666667, 53922.4583333333, 0, -3431.42916666667, 0, -17157.1458333333, -3431.42916666667, -34314.2916666667, -17157.1458333333, -34314.2916666667}, 
                                                  {0, 0, 49, 0, -546.35, -546.35, 0, 2431.48888888889, 2431.48888888889, 4862.97777777778, 0, -5552.3125, -5552.3125, -16656.9375, -16656.9375, 0, 6886.20138888889, 6886.20138888889, 27544.8055555556, 27544.8055555555, 41317.2083333333, 0, -4411.8375, -4411.8375, -22059.1875, -22059.1875, -44118.375, -44118.375, 0, 1143.80972222222, 1143.80972222222, 6862.85833333333, 6862.85833333333, 17157.1458333333, 17157.1458333333, 22876.1944444444}, 
                                                  {0, 0, 0, 0, 0, 1029, 0, 0, -7871.85, -7871.85, 0, 0, 23509.7916666667, 23509.7916666667, 47019.5833333333, 0, 0, -34314.2916666667, -34314.2916666667, -102942.875, -102942.875, 0, 0, 24510.2083333333, 24510.2083333333, 98040.8333333333, 98040.8333333333, 147061.25, 0, 0, -6862.85833333333, -6862.85833333333, -34314.2916666667, -34314.2916666667, -68628.5833333333, -68628.5833333333}, 
                                                  {0, 0, 0, 0, 0, 68.6, 0, 0, -1069.01666666667, -68.6, 0, 0, 5902.45833333333, 0, 1000.41666666667, 0, 0, -14706.125, 0, -4902.04166666667, 0, 0, 0, 16666.9416666667, 0, 9804.08333333333, 0, 0, 0, 0, -6862.85833333333, 0, -6862.85833333333, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, 68.6, 0, 0, -68.6, -1069.01666666667, 0, 0, 0, 5902.45833333333, 1000.41666666667, 0, 0, 0, -14706.125, 0, -4902.04166666667, 0, 0, 0, 16666.9416666667, 0, 9804.08333333334, 0, 0, 0, 0, -6862.85833333333, 0, -6862.85833333333, 0, 0}, 
                                                  {0, 0, 0, 0, 0, -857.5, 0, 0, 10561.5416666667, 4559.04166666667, 0, 0, -40867.0208333333, -8953.72916666667, -49820.75, 0, 0, 70379.3125, 7703.20833333333, 148461.833333333, 85785.7291666667, 0, 0, -56373.4791666667, -2451.02083333333, -171571.458333333, -63726.5416666667, -176473.5, 0, 0, 17157.1458333333, 0, 68628.5833333333, 17157.1458333333, 102942.875, 68628.5833333333}, 
                                                  {0, 0, 0, 0, 0, 571.666666666667, 0, 0, -8041.44444444444, -2038.94444444444, 0, 0, 37815.75, 2401, 26210.9166666667, 0, 0, -76098.3611111111, -933.722222222222, -103176.305555556, -28011.6666666667, 0, 0, 68628.5833333333, 0, 147061.25, 9804.08333333332, 88236.75, 0, 0, -22876.1944444444, 0, -68628.5833333333, 0, -68628.5833333333, -22876.1944444444}, 
                                                  {0, 0, 0, 0, 0, -257.25, 0, 0, 3858.75, 557.375, 0, 0, -20058.3541666667, -300.125, -7753.22916666667, 0, 0, 45869.1041666667, 0, 35014.5833333333, 3851.60416666666, 0, 0, -46569.3958333333, 0, -61275.5208333333, 0, -14706.125, 0, 0, 17157.1458333333, 0, 34314.2916666667, 0, 17157.1458333333, 0}, 
                                                  {0, 0, 0, 0, 0, 42.875, 0, 0, -593.104166666667, -343, 0, 0, 2651.10416666667, 300.125, 4701.95833333333, 0, 0, -4551.89583333333, 0, -20658.6041666667, -3851.60416666667, 0, 0, 2451.02083333333, 0, 34314.2916666667, 0, 14706.125, 0, 0, 0, 0, -17157.1458333333, 0, -17157.1458333333, 0}, 
                                                  {0, 0, 0, 0, 0, 38.1111111111111, 0, 0, -438.277777777778, -438.277777777778, 0, 0, 1333.88888888889, 1333.88888888889, 5002.08333333333, 0, 0, -933.722222222222, -933.722222222222, -14939.5555555556, -14939.5555555556, 0, 0, 0, 0, 9804.08333333333, 9804.08333333333, 42484.3611111111, 0, 0, 0, 0, 0, 0, -22876.1944444444, -22876.1944444444}, 
                                                  {0, 0, 0, 0, 0, 42.875, 0, 0, -343, -593.104166666667, 0, 0, 300.125, 2651.10416666667, 4701.95833333333, 0, 0, 0, -4551.89583333333, -3851.60416666667, -20658.6041666667, 0, 0, 0, 2451.02083333333, 0, 34314.2916666667, 14706.125, 0, 0, 0, 0, 0, -17157.1458333333, 0, -17157.1458333333}, 
                                                  {0, 0, 0, 0, 0, -257.25, 0, 0, 557.375, 3858.75, 0, 0, -300.125, -20058.3541666667, -7753.22916666667, 0, 0, 0, 45869.1041666667, 3851.60416666666, 35014.5833333333, 0, 0, 0, -46569.3958333333, 0, -61275.5208333333, -14706.125, 0, 0, 0, 17157.1458333333, 0, 34314.2916666667, 0, 17157.1458333333}, 
                                                  {0, 0, 0, 0, 0, 571.666666666667, 0, 0, -2038.94444444444, -8041.44444444444, 0, 0, 2401, 37815.75, 26210.9166666667, 0, 0, -933.722222222222, -76098.3611111111, -28011.6666666667, -103176.305555556, 0, 0, 0, 68628.5833333333, 9804.08333333333, 147061.25, 88236.75, 0, 0, 0, -22876.1944444444, 0, -68628.5833333333, -22876.1944444444, -68628.5833333333}, 
                                                  {0, 0, 0, 0, 0, -857.5, 0, 0, 4559.04166666667, 10561.5416666667, 0, 0, -8953.72916666667, -40867.0208333333, -49820.75, 0, 0, 7703.20833333333, 70379.3125, 85785.7291666667, 148461.833333333, 0, 0, -2451.02083333333, -56373.4791666667, -63726.5416666667, -171571.458333333, -176473.5, 0, 0, 0, 17157.1458333333, 17157.1458333333, 68628.5833333333, 68628.5833333333, 102942.875}, 
                                                  {0, 0, 0, 0, 0, 428.75, 0, 0, -4530.45833333333, -4530.45833333333, 0, 0, 12505.2083333333, 12505.2083333333, 46019.1666666667, 0, 0, -13305.5416666667, -13305.5416666667, -114847.833333333, -114847.833333333, 0, 0, 4902.04166666667, 4902.04166666667, 107844.916666667, 107844.916666667, 205885.75, 0, 0, 0, 0, -34314.2916666667, -34314.2916666667, -102942.875, -102942.875}, 
                                                  {0, 0, 0, 0, 0, -171.5, 0, 0, 2172.33333333333, 1572.08333333333, 0, 0, -8303.45833333333, -2801.16666666667, -19508.125, 0, 0, 11204.6666666667, 1400.58333333333, 71429.75, 32213.4166666667, 0, 0, -4902.04166666667, 0, -88236.75, -14706.125, -98040.8333333333, 0, 0, 0, 0, 34314.2916666667, 0, 68628.5833333333, 34314.2916666667}, 
                                                  {0, 0, 0, 0, 0, -171.5, 0, 0, 1572.08333333333, 2172.33333333333, 0, 0, -2801.16666666667, -8303.45833333333, -19508.125, 0, 0, 1400.58333333333, 11204.6666666667, 32213.4166666667, 71429.75, 0, 0, 0, -4902.04166666667, -14706.125, -88236.75, -98040.8333333333, 0, 0, 0, 0, 0, 34314.2916666667, 34314.2916666667, 68628.5833333333}};




double MAd_TriInterpolant::coef8[45][45] = {{1, -21.7428571428571, -21.7428571428571, 187.498412698413, 187.498412698413, 374.996825396825, -854.4, -854.4, -2563.2, -2563.2, 2280.53333333333, 2280.53333333333, 9122.13333333333, 9122.13333333333, 13683.2, -3686.4, -3686.4, -18432, -18432, -36864, -36864, 3549.86666666667, 3549.86666666666, 21299.2, 21299.2, 53247.9999999999, 53248, 70997.3333333333, -1872.45714285714, -1872.45714285714, -13107.2, -13107.2, -39321.5999999999, -39321.6, -65535.9999999999, -65535.9999999999, 416.101587301588, 416.101587301587, 3328.81269841269, 3328.8126984127, 11650.8444444444, 11650.8444444444, 23301.6888888889, 23301.6888888889, 29127.1111111111}, 
                                                  {0, -1, 0, 20.7428571428571, 0, 0, -166.755555555556, 0, 0, 0, 687.644444444444, 0, 0, 0, 0, -1592.88888888889, 0, 0, 0, 0, 0, 2093.51111111111, 0, 0, 0, 0, 0, 0, -1456.35555555556, 0, 0, 0, 0, 0, 0, 0, 416.101587301587, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                  {0, 0, -0.999999999999999, 0, 20.7428571428571, 0, 0, -166.755555555555, 0, 0, 0, 687.644444444444, 0, 0, 0, 0, -1592.88888888889, 0, 0, 0, 0, 0, 2093.51111111111, 0, 0, 0, 0, 0, 0, -1456.35555555555, 0, 0, 0, 0, 0, 0, 0, 416.101587301587, 0, 0, 0, 0, 0, 0, 0}, 
                                                  {0, 64, 0, -879.542857142857, 0, -879.542857142857, 4963.55555555556, 0, 9927.11111111111, 4963.55555555555, -14973.1555555556, 0, -44919.4666666667, -14973.1555555556, -44919.4666666667, 26168.8888888889, 0, 104675.555555556, 26168.8888888889, 157013.333333333, 104675.555555556, -26578.4888888889, 0, -132892.444444444, -26578.4888888889, -265784.888888889, -132892.444444444, -265784.888888889, 14563.5555555556, 0, 87381.3333333333, 14563.5555555556, 218453.333333333, 87381.3333333333, 291271.111111111, 218453.333333333, -3328.8126984127, 0, -23301.6888888889, -3328.8126984127, -69905.0666666666, -23301.6888888889, -116508.444444444, -69905.0666666667, -116508.444444444}, 
                                                  {0, -112, 0, 1987.2, 0, 1091.2, -13051.0222222222, 0, -17372.4444444444, -4321.42222222222, 43488.7111111111, 0, 95894.7555555555, 8917.33333333333, 61323.3777777778, -81464.8888888889, 0, -254520.888888889, -10126.2222222222, -274773.333333333, -111843.555555556, 87017.2444444445, 0, 354076.444444444, 6007.46666666666, 546133.333333333, 111047.111111111, 384113.777777778, -49516.0888888889, 0, -249036.8, -1456.35555555555, -502442.666666666, -56797.8666666666, -509724.444444444, -262144, 11650.8444444444, 0, 69905.0666666666, 0, 174762.666666667, 11650.8444444444, 233016.888888889, 69905.0666666667, 174762.666666667}, 
                                                  {0, 149.333333333333, 0, -2848.71111111111, 0, -1056.71111111111, 20403.2, 0, 18568.5333333333, 2944, -73181.8666666667, 0, -116588.088888889, -4039.11111111111, -47445.3333333333, 145408, 0, 344746.666666667, 2730.66666666666, 256000, 59392, -162747.733333333, 0, -522467.555555555, -728.177777777772, -591644.444444444, -36408.8888888889, -267605.333333333, 96119.4666666667, 0, 393216, 0, 611669.333333333, 8738.13333333331, 436906.666666667, 131072, -23301.6888888889, 0, -116508.444444444, 0, -233016.888888889, 0, -233016.888888889, -23301.688888889, -116508.444444445}, 
                                                  {0, -140, 0, 2764, 0, 710.666666666666, -20721.7777777778, 0, -13100.4444444444, -1338.66666666667, 78172.4444444445, 0, 88078.2222222222, 1109.33333333333, 22961.7777777778, -162929.777777778, 0, -282168.888888889, -341.333333333336, -136533.333333333, -17635.5555555556, 190236.444444444, 0, 461482.666666666, 0, 357262.222222222, 5006.22222222228, 91022.2222222223, -116508.444444444, 0, -371370.666666667, 0, -415061.333333333, 0, -182044.444444444, -21845.3333333335, 29127.1111111111, 0, 116508.444444444, 0, 174762.666666667, 0, 116508.444444444, 0, 29127.1111111112}, 
                                                  {0, 89.6, 0, -1804.8, 0, -311.466666666667, 13912.1777777778, 0, 5907.91111111111, 358.4, -54294.7555555556, 0, -41426.4888888889, -136.533333333334, -6382.93333333333, 117464.177777778, 0, 140174.222222222, 0, 40277.3333333333, 2275.55555555557, -142358.755555556, 0, -244303.644444444, 0, -114688, 0, -12743.1111111112, 90294.0444444445, 0, 209715.2, 0, 148548.266666667, 0, 29127.1111111111, 0, -23301.6888888889, 0, -69905.0666666666, 0, -69905.0666666667, 0, -23301.6888888889, 0, 0}, 
                                                  {0, -37.3333333333333, 0, 761.955555555556, 0, 80, -5984, 0, -1546.66666666667, -42.6666666666666, 23918.9333333333, 0, 11158.7555555556, 0, 779.377777777778, -53248, 0, -39253.3333333333, 0, -5120, 0, 66628.2666666667, 0, 71907.5555555555, 0, 15473.7777777778, 0, 0, -43690.6666666667, 0, -65536, 0, -21845.3333333333, 0, 0, 0, 11650.8444444444, 0, 23301.6888888889, 0, 11650.8444444444, 0, 0, 0, 0}, 
                                                  {0, 9.14285714285714, 0, -188.342857142857, 0, -9.14285714285714, 1499.02222222222, 0, 179.2, 0, -6098.48888888889, 0, -1319.82222222222, 0, 0, 13880.8888888889, 0, 4778.66666666666, 0, 0, 0, -17840.3555555556, 0, -9102.22222222222, 0, 0, 0, 0, 12066.946031746, 0, 8738.13333333333, 0, 0, 0, 0, 0, -3328.8126984127, 0, -3328.8126984127, 0, 0, 0, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, 9.14285714285714, 0, 0, -179.2, 0, 0, 0, 1319.82222222222, 0, 0, 0, 0, -4778.66666666666, 0, 0, 0, 0, 0, 9102.22222222222, 0, 0, 0, 0, 0, 0, -8738.13333333333, 0, 0, 0, 0, 0, 0, 0, 3328.8126984127, 0, 0, 0, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, 5.33333333333333, 0, 0, -97.4222222222222, -42.6666666666667, 0, 0, 640, 0, 779.377777777778, 0, 0, -1934.22222222222, 0, -5120, 0, 0, 0, 2730.66666666667, 0, 15473.7777777778, 0, 0, 0, 0, -1456.35555555555, 0, -21845.3333333333, 0, 0, 0, 0, 0, 0, 0, 11650.8444444444, 0, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, 4.26666666666666, 0, 0, -71.1111111111111, -51.2, 0, 0, 398.222222222222, 136.533333333333, 853.333333333333, 0, 0, -910.222222222222, 0, -4778.66666666667, -2275.55555555555, 0, 0, 728.177777777777, 0, 10922.6666666667, 0, 12743.1111111111, 0, 0, 0, 0, -8738.13333333333, 0, -29127.1111111111, 0, 0, 0, 0, 0, 0, 0, 23301.6888888889, 0, 0}, 
                                                  {0, 0, 0, 0, 0, 4, 0, 0, -58.6666666666666, -58.6666666666666, 0, 0, 256, 256, 860.444444444444, 0, 0, -341.333333333333, -341.333333333333, -3754.66666666667, -3754.66666666666, 0, 0, 0, 0, 5006.22222222222, 5006.22222222222, 16384, 0, 0, 0, 0, 0, 0, -21845.3333333333, -21845.3333333333, 0, 0, 0, 0, 0, 0, 0, 0, 29127.1111111111}, 
                                                  {0, 0, 0, 0, 0, 4.26666666666666, 0, 0, -51.1999999999999, -71.1111111111111, 0, 0, 136.533333333333, 398.222222222222, 853.333333333333, 0, 0, 0, -910.222222222222, -2275.55555555555, -4778.66666666666, 0, 0, 0, 728.177777777778, 0, 10922.6666666667, 12743.1111111111, 0, 0, 0, 0, 0, -8738.13333333333, 0, -29127.1111111111, 0, 0, 0, 0, 0, 0, 0, 23301.6888888889, 0}, 
                                                  {0, 0, 0, 0, 0, 5.33333333333333, 0, 0, -42.6666666666666, -97.4222222222222, 0, 0, 0, 640, 779.377777777777, 0, 0, 0, -1934.22222222222, 0, -5120, 0, 0, 0, 2730.66666666667, 0, 15473.7777777778, 0, 0, 0, 0, -1456.35555555556, 0, -21845.3333333333, 0, 0, 0, 0, 0, 0, 0, 11650.8444444444, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, 9.14285714285715, 0, 0, 0, -179.2, 0, 0, 0, 1319.82222222222, 0, 0, 0, 0, -4778.66666666667, 0, 0, 0, 0, 0, 9102.22222222223, 0, 0, 0, 0, 0, 0, -8738.13333333334, 0, 0, 0, 0, 0, 0, 0, 3328.8126984127, 0, 0, 0, 0, 0}, 
                                                  {0, 0, 9.14285714285713, 0, -188.342857142857, -9.14285714285693, 0, 1499.02222222222, 0, 179.199999999999, 0, -6098.48888888888, 0, -1319.82222222222, 0, 0, 13880.8888888889, 0, 4778.66666666665, 0, 0, 0, -17840.3555555555, 0, -9102.2222222222, 0, 0, 0, 0, 12066.946031746, 0, 8738.13333333332, 0, 0, 0, 0, 0, -3328.8126984127, 0, -3328.81269841269, 0, 0, 0, 0, 0}, 
                                                  {0, 0, -37.3333333333333, 0, 761.955555555555, 79.9999999999992, 0, -5983.99999999999, -42.6666666666553, -1546.66666666666, 0, 23918.9333333333, 0, 11158.7555555555, 779.377777777742, 0, -53248, 0, -39253.3333333333, 0, -5119.99999999995, 0, 66628.2666666666, 0, 71907.5555555555, 0, 15473.7777777777, 0, 0, -43690.6666666666, 0, -65536, 0, -21845.3333333333, 0, 0, 0, 11650.8444444444, 0, 23301.6888888889, 0, 11650.8444444444, 0, 0, 0}, 
                                                  {0, 0, 89.5999999999999, 0, -1804.8, -311.466666666665, 0, 13912.1777777778, 358.399999999975, 5907.9111111111, 0, -54294.7555555555, -136.533333333181, -41426.4888888889, -6382.93333333325, 0, 117464.177777778, 0, 140174.222222222, 2275.55555555515, 40277.3333333332, 0, -142358.755555555, 0, -244303.644444444, 0, -114688, -12743.1111111107, 0, 90294.0444444444, 0, 209715.2, 0, 148548.266666667, 0, 29127.1111111109, 0, -23301.6888888889, 0, -69905.0666666667, 0, -69905.0666666666, 0, -23301.6888888888, 0}, 
                                                  {0, 0, -140, 0, 2764, 710.666666666664, 0, -20721.7777777778, -1338.66666666663, -13100.4444444444, 0, 78172.4444444444, 1109.33333333312, 88078.2222222222, 22961.7777777776, 0, -162929.777777778, -341.333333332707, -282168.888888889, -17635.5555555549, -136533.333333333, 0, 190236.444444444, 0, 461482.666666667, 5006.22222222082, 357262.222222222, 91022.2222222214, 0, -116508.444444444, 0, -371370.666666667, 0, -415061.333333333, -21845.3333333322, -182044.444444444, 0, 29127.1111111111, 0, 116508.444444444, 0, 174762.666666667, 0, 116508.444444444, 29127.1111111108}, 
                                                  {0, 0, 149.333333333333, 0, -2848.71111111111, -1056.71111111111, 0, 20403.2, 2943.99999999997, 18568.5333333333, 0, -73181.8666666666, -4039.11111111091, -116588.088888889, -47445.3333333332, 0, 145408, 2730.6666666661, 344746.666666667, 59391.9999999994, 256000, 0, -162747.733333333, -728.177777776936, -522467.555555556, -36408.8888888875, -591644.444444444, -267605.333333332, 0, 96119.4666666666, 0, 393216, 8738.13333333196, 611669.333333333, 131071.999999999, 436906.666666666, 0, -23301.6888888889, 0, -116508.444444444, 0, -233016.888888889, -23301.6888888882, -233016.888888889, -116508.444444444}, 
                                                  {0, 0, -112, 0, 1987.2, 1091.2, 0, -13051.0222222222, -4321.4222222222, -17372.4444444444, 0, 43488.7111111111, 8917.33333333322, 95894.7555555555, 61323.3777777777, 0, -81464.8888888888, -10126.2222222219, -254520.888888889, -111843.555555555, -274773.333333333, 0, 87017.2444444444, 6007.46666666618, 354076.444444445, 111047.11111111, 546133.333333333, 384113.777777777, 0, -49516.0888888888, -1456.3555555552, -249036.8, -56797.8666666658, -502442.666666666, -262143.999999999, -509724.444444444, 0, 11650.8444444444, 0, 69905.0666666667, 11650.8444444441, 174762.666666667, 69905.0666666662, 233016.888888889, 174762.666666666}, 
                                                  {0, 0, 64, 0, -879.542857142857, -879.542857142857, 0, 4963.55555555555, 4963.55555555555, 9927.11111111111, 0, -14973.1555555555, -14973.1555555555, -44919.4666666667, -44919.4666666666, 0, 26168.8888888889, 26168.8888888888, 104675.555555556, 104675.555555555, 157013.333333333, 0, -26578.4888888889, -26578.4888888887, -132892.444444444, -132892.444444444, -265784.888888889, -265784.888888889, 0, 14563.5555555555, 14563.5555555554, 87381.3333333334, 87381.333333333, 218453.333333333, 218453.333333333, 291271.111111111, 0, -3328.8126984127, -3328.81269841266, -23301.6888888889, -23301.6888888888, -69905.0666666667, -69905.0666666665, -116508.444444444, -116508.444444444}, 
                                                  {0, 0, 0, 0, 0, 1792, 0, 0, -17459.2, -17459.2, 0, 0, 69142.7555555555, 69142.7555555556, 138285.511111111, 0, 0, -142677.333333333, -142677.333333333, -428032, -428032, 0, 0, 162019.555555555, 162019.555555556, 648078.222222222, 648078.222222222, 972117.333333333, 0, 0, -96119.4666666666, -96119.4666666668, -480597.333333333, -480597.333333333, -961194.666666666, -961194.666666666, 0, 0, 23301.6888888889, 23301.6888888889, 139810.133333333, 139810.133333333, 349525.333333333, 349525.333333333, 466033.777777777}, 
                                                  {0, 0, 0, 0, 0, -85.3333333333333, 0, 0, 1644.08888888889, 85.3333333333333, 0, 0, -11798.7555555556, 0, -1558.75555555556, 0, 0, 41187.5555555555, 0, 10240, 0, 0, 0, -74638.2222222222, 0, -30947.5555555555, 0, 0, 0, 0, 66992.3555555555, 0, 43690.6666666666, 0, 0, 0, 0, 0, -23301.6888888889, 0, -23301.6888888889, 0, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, -85.3333333333334, 0, 0, 85.3333333333333, 1644.08888888889, 0, 0, 0, -11798.7555555556, -1558.75555555556, 0, 0, 0, 41187.5555555556, 0, 10240, 0, 0, 0, -74638.2222222223, 0, -30947.5555555556, 0, 0, 0, 0, 66992.3555555556, 0, 43690.6666666667, 0, 0, 0, 0, 0, -23301.6888888889, 0, -23301.6888888889, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, -1792, 0, 0, 27016.5333333333, 12680.5333333333, 0, 0, -136772.266666667, -35328, -172100.266666667, 0, 0, 331093.333333333, 48469.3333333333, 710656, 428032, 0, 0, -420522.666666666, -32768, -1294336, -518826.666666666, -1359872, 0, 0, 270882.133333333, 8738.13333333333, 1092266.66666667, 305834.666666666, 1660245.33333333, 1135957.33333333, 0, 0, -69905.0666666666, 0, -349525.333333333, -69905.0666666666, -699050.666666666, -349525.333333333, -699050.666666666}, 
                                                  {0, 0, 0, 0, 0, 1493.33333333333, 0, 0, -25500.4444444444, -7580.44444444444, 0, 0, 153031.111111111, 14279.1111111111, 119523.555555556, 0, 0, -425756.444444444, -11832.8888888889, -620771.555555555, -206848, 0, 0, 602567.111111111, 3640.88888888886, 1354410.66666667, 156558.222222222, 904760.888888889, 0, 0, -422343.111111111, 0, -1310720, -43690.6666666666, -1398101.33333333, -553415.111111111, 0, 0, 116508.444444444, 0, 466033.777777778, 0, 699050.666666666, 116508.444444445, 466033.777777778}, 
                                                  {0, 0, 0, 0, 0, -896, 0, 0, 16256, 3114.66666666667, 0, 0, -106609.777777778, -3584, -52849.7777777778, 0, 0, 329728, 1365.33333333333, 308565.333333333, 56661.3333333333, 0, 0, -515185.777777778, 0, -784611.555555555, -20024.8888888889, -289450.666666667, 0, 0, 393216, 0, 873813.333333333, 0, 567978.666666667, 87381.3333333335, 0, 0, -116508.444444444, 0, -349525.333333333, 0, -349525.333333333, 0, -116508.444444445}, 
                                                  {0, 0, 0, 0, 0, 358.4, 0, 0, -6741.33333333333, -768, 0, 0, 46660.2666666666, 409.6, 13619.2, 0, 0, -154965.333333333, 0, -85333.3333333333, -6826.66666666668, 0, 0, 263236.266666667, 0, 240298.666666667, 0, 38229.3333333334, 0, 0, -218453.333333333, 0, -305834.666666667, 0, -87381.3333333333, 0, 0, 0, 69905.0666666666, 0, 139810.133333333, 0, 69905.0666666666, 0, 0}, 
                                                  {0, 0, 0, 0, 0, -51.2, 0, 0, 904.533333333333, 460.8, 0, 0, -5632, -409.599999999999, -8089.6, 0, 0, 15701.3333333333, 0, 49834.6666666667, 6826.66666666666, 0, 0, -19660.8, 0, -136533.333333333, 0, -38229.3333333333, 0, 0, 8738.13333333333, 0, 166024.533333333, 0, 87381.3333333333, 0, 0, 0, 0, 0, -69905.0666666667, 0, -69905.0666666667, 0, 0}, 
                                                  {0, 0, 0, 0, 0, -42.6666666666666, 0, 0, 668.444444444444, 554.666666666666, 0, 0, -3356.44444444444, -1877.33333333333, -8647.11111111111, 0, 0, 6371.55555555555, 1365.33333333333, 43008, 28899.5555555555, 0, 0, -3640.88888888888, 0, -80099.5555555555, -20024.8888888889, -140174.222222222, 0, 0, 0, 0, 43690.6666666667, 0, 247580.444444444, 87381.3333333333, 0, 0, 0, 0, 0, 0, -116508.444444444, 0, -116508.444444444}, 
                                                  {0, 0, 0, 0, 0, -42.6666666666666, 0, 0, 554.666666666666, 668.444444444444, 0, 0, -1877.33333333333, -3356.44444444444, -8647.11111111111, 0, 0, 1365.33333333333, 6371.55555555555, 28899.5555555555, 43008, 0, 0, 0, -3640.88888888889, -20024.8888888889, -80099.5555555555, -140174.222222222, 0, 0, 0, 0, 0, 43690.6666666666, 87381.3333333333, 247580.444444444, 0, 0, 0, 0, 0, 0, 0, -116508.444444444, -116508.444444444}, 
                                                  {0, 0, 0, 0, 0, -51.2, 0, 0, 460.8, 904.533333333333, 0, 0, -409.599999999998, -5632, -8089.6, 0, 0, 0, 15701.3333333333, 6826.66666666665, 49834.6666666666, 0, 0, 0, -19660.8, 0, -136533.333333333, -38229.3333333333, 0, 0, 0, 8738.13333333335, 0, 166024.533333333, 0, 87381.3333333333, 0, 0, 0, 0, 0, -69905.0666666667, 0, -69905.0666666666, 0}, 
                                                  {0, 0, 0, 0, 0, 358.4, 0, 0, -768, -6741.33333333334, 0, 0, 409.599999999998, 46660.2666666667, 13619.2, 0, 0, 0, -154965.333333333, -6826.66666666664, -85333.3333333333, 0, 0, 0, 263236.266666667, 0, 240298.666666667, 38229.3333333332, 0, 0, 0, -218453.333333333, 0, -305834.666666667, 0, -87381.3333333332, 0, 0, 0, 69905.0666666667, 0, 139810.133333333, 0, 69905.0666666666, 0}, 
                                                  {0, 0, 0, 0, 0, -896, 0, 0, 3114.66666666666, 16256, 0, 0, -3583.99999999999, -106609.777777778, -52849.7777777778, 0, 0, 1365.33333333332, 329728, 56661.3333333332, 308565.333333333, 0, 0, 0, -515185.777777778, -20024.8888888888, -784611.555555556, -289450.666666666, 0, 0, 0, 393216, 0, 873813.333333333, 87381.3333333329, 567978.666666666, 0, 0, 0, -116508.444444445, 0, -349525.333333333, 0, -349525.333333333, -116508.444444444}, 
                                                  {0, 0, 0, 0, 0, 1493.33333333333, 0, 0, -7580.44444444444, -25500.4444444444, 0, 0, 14279.1111111111, 153031.111111111, 119523.555555556, 0, 0, -11832.8888888888, -425756.444444445, -206848, -620771.555555555, 0, 0, 3640.88888888884, 602567.111111112, 156558.222222222, 1354410.66666667, 904760.888888888, 0, 0, 0, -422343.111111112, -43690.6666666665, -1310720, -553415.11111111, -1398101.33333333, 0, 0, 0, 116508.444444445, 0, 466033.777777778, 116508.444444444, 699050.666666666, 466033.777777777}, 
                                                  {0, 0, 0, 0, 0, -1792, 0, 0, 12680.5333333333, 27016.5333333333, 0, 0, -35328, -136772.266666667, -172100.266666667, 0, 0, 48469.3333333333, 331093.333333333, 428032, 710656, 0, 0, -32767.9999999999, -420522.666666667, -518826.666666666, -1294336, -1359872, 0, 0, 8738.13333333328, 270882.133333334, 305834.666666666, 1092266.66666667, 1135957.33333333, 1660245.33333333, 0, 0, 0, -69905.0666666668, -69905.0666666666, -349525.333333333, -349525.333333333, -699050.666666666, -699050.666666666}, 
                                                  {0, 0, 0, 0, 0, 1120, 0, 0, -14645.3333333333, -14645.3333333333, 0, 0, 56192, 56192, 184064, 0, 0, -94549.3333333333, -94549.3333333333, -647509.333333333, -647509.333333333, 0, 0, 73727.9999999999, 73728, 980309.333333333, 980309.333333333, 1813162.66666667, 0, 0, -21845.3333333333, -21845.3333333334, -677205.333333333, -677205.333333333, -1922389.33333333, -1922389.33333333, 0, 0, 0, 0, 174762.666666667, 174762.666666667, 699050.666666666, 699050.666666666, 1048576}, 
                                                  {0, 0, 0, 0, 0, 224, 0, 0, -3765.33333333333, -2272, 0, 0, 21632, 4095.99999999999, 37674.6666666667, 0, 0, -53589.3333333333, -2047.99999999999, -211285.333333333, -64170.6666666666, 0, 0, 57344, 0, 502442.666666667, 30037.3333333333, 322218.666666667, 0, 0, -21845.3333333333, 0, -502442.666666667, 0, -611669.333333333, -131072, 0, 0, 0, 0, 174762.666666667, 0, 349525.333333333, 0, 174762.666666667}, 
                                                  {0, 0, 0, 0, 0, 224, 0, 0, -2272, -3765.33333333333, 0, 0, 4095.99999999999, 21632, 37674.6666666666, 0, 0, -2047.99999999999, -53589.3333333333, -64170.6666666666, -211285.333333333, 0, 0, 0, 57344, 30037.3333333332, 502442.666666667, 322218.666666666, 0, 0, 0, -21845.3333333334, 0, -502442.666666667, -131072, -611669.333333333, 0, 0, 0, 0, 0, 174762.666666667, 0, 349525.333333333, 174762.666666666}, 
                                                  {0, 0, 0, 0, 0, -597.333333333333, 0, 0, 9244.44444444444, 6855.11111111111, 0, 0, -46421.3333333333, -19000.8888888889, -103651.555555556, 0, 0, 96028.4444444444, 20024.8888888889, 497891.555555555, 268970.666666667, 0, 0, -87381.3333333333, -7281.77777777774, -953912.888888889, -262144, -1121393.77777778, 0, 0, 29127.1111111111, 0, 786432, 87381.3333333332, 1572864, 902940.444444444, 0, 0, 0, 0, -233016.888888889, 0, -699050.666666667, -233016.888888889, -699050.666666666}, 
                                                  {0, 0, 0, 0, 0, 199.111111111111, 0, 0, -2816, -2816, 0, 0, 11719.1111111111, 11719.1111111111, 39367.1111111111, 0, 0, -16384, -16384, -159744, -159744, 0, 0, 7281.77777777776, 7281.77777777777, 211171.555555555, 211171.555555555, 611669.333333333, 0, 0, 0, 0, -87381.3333333333, -87381.3333333332, -699050.666666666, -699050.666666666, 0, 0, 0, 0, 0, 0, 233016.888888889, 233016.888888889, 466033.777777778}, 
                                                  {0, 0, 0, 0, 0, -597.333333333333, 0, 0, 6855.11111111111, 9244.44444444444, 0, 0, -19000.8888888889, -46421.3333333333, -103651.555555556, 0, 0, 20024.8888888888, 96028.4444444444, 268970.666666666, 497891.555555555, 0, 0, -7281.77777777773, -87381.3333333334, -262144, -953912.888888889, -1121393.77777778, 0, 0, 0, 29127.1111111112, 87381.3333333331, 786432, 902940.444444444, 1572864, 0, 0, 0, 0, 0, -233016.888888889, -233016.888888889, -699050.666666666, -699050.666666666}};




double MAd_TriInterpolant::coef9[55][55] = {{1, -25.4607142857143, -25.4607142857143, 261.763392857143, 261.763392857142, 523.526785714285, -1453.82142857143, -1453.82142857142, -4361.46428571429, -4361.46428571428, 4869.4921875, 4869.49218749997, 19477.96875, 19477.9687499999, 29216.953125, -10295.9859375, -10295.9859374999, -51479.9296875001, -51479.9296874995, -102959.859375, -102959.859375, 13839.609375, 13839.6093749999, 83037.6562500003, 83037.6562499989, 207594.140625001, 207594.140625, 276792.187500001, -11467.1049107143, -11467.1049107142, -80269.7343750004, -80269.7343749986, -240809.203125001, -240809.203124999, -401348.671875002, -401348.671875001, 5338.13504464286, 5338.1350446428, 42705.0803571431, 42705.080357142, 149467.78125, 149467.781249999, 298935.562500002, 298935.5625, 373669.453125002, -1067.62700892857, -1067.62700892856, -9608.64308035722, -9608.6430803569, -38434.5723214287, -38434.5723214282, -89680.6687500004, -89680.6687499999, -134521.003125001, -134521.003125001}, 
                                                  {0, 0.999999999999999, 0, -24.4607142857143, 0, 0, 237.302678571428, 0, 0, 0, -1216.51875, 0, 0, 0, 0, 3652.9734375, 0, 0, 0, 0, 0, -6643.01249999999, 0, 0, 0, 0, 0, 0, 7196.59687499999, 0, 0, 0, 0, 0, 0, 0, -4270.50803571428, 0, 0, 0, 0, 0, 0, 0, 0, 1067.62700892857, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                  {0, 0, 0.999999999999976, 0, -24.4607142857137, 0, 0, 237.302678571424, 0, 0, 0, -1216.51874999998, 0, 0, 0, 0, 3652.97343749994, 0, 0, 0, 0, 0, -6643.0124999999, 0, 0, 0, 0, 0, 0, 7196.59687499991, 0, 0, 0, 0, 0, 0, 0, -4270.50803571424, 0, 0, 0, 0, 0, 0, 0, 0, 1067.62700892856, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                  {0, 81, 0, -1333.31785714286, 0, -1333.31785714286, 9202.97410714286, 0, 18405.9482142857, 9202.97410714285, -34932.76875, 0, -104798.30625, -34932.7687499999, -104798.30625, 80033.9484375, 0, 320135.79375, 80033.9484374998, 480203.690624999, 320135.79375, -113669.325, 0, -568346.625000001, -113669.325, -1136693.25, -568346.625, -1136693.25, 97984.434375, 0, 587906.606250001, 97984.4343749994, 1469766.515625, 587906.606249999, 1959688.6875, 1469766.515625, -46975.5883928571, 0, -328829.118750001, -46975.5883928567, -986487.356249998, -328829.118749999, -1644145.59374999, -986487.35625, -1644145.59375, 9608.64308035714, 0, 76869.1446428575, 9608.64308035702, 269042.00625, 76869.1446428569, 538084.012499998, 269042.00625, 672605.015624998, 538084.0125}, 
                                                  {0, -162, 0, 3395.63571428571, 0, 1937.63571428571, -27125.3089285714, 0, -36811.8964285714, -9686.58749999999, 113455.18125, 0, 253186.25625, 26275.8937499999, 166006.96875, -278309.41875, 0, -876754.631250001, -41826.3749999999, -960407.381249998, -403788.54375, 415557.3375, 0, 1701349.3125, 39119.9624999997, 2649823.875, 572037.1875, 1896949.125, -372008.7, 0, -1879972.5375, -19929.0374999995, -3819732.18749999, -471653.887500001, -3919377.37499999, -2059333.875, 183631.845535714, 0, 1106061.58125, 4270.50803571391, 2780100.73125, 209254.89375, 3736694.53124999, 1165848.69375, 2839887.84375, -38434.5723214286, 0, -269042.006250001, 0, -807126.018749999, -38434.5723214285, -1345210.03124999, -269042.006250001, -1345210.03125, -807126.018750001}, 
                                                  {0, 252, 0, -5660.1, 0, -2258.1, 48984.075, 0, 47071.8, 8293.72499999999, -219410.775, 0, -363360.9375, -15992.4375, -159942.6, 568879.70625, 0, 1387829.19375, 17085.9375000001, 1086105.20625, 284241.65625, -887949.3375, 0, -2913699.09375001, -9595.46250000016, -3422996.71874999, -278637.468750002, -1666288.96875, 823733.55, 0, 3428901.61875001, 2214.33750000003, 5480485.31249999, 142824.768750003, 4107596.06249999, 1372889.25, -418509.7875, 0, -2122442.49375001, 0, -4334565.65624999, -29893.5562500015, -4484033.43749998, -567977.568750006, -2391484.5, 89680.66875, 0, 538084.012500002, 0, 1345210.03125, 0, 1793613.37499999, 89680.668750002, 1345210.03125, 538084.012500004}, 
                                                  {0, -283.5, 0, 6580.2375, 0, 1902.4875, -59404.3875, 0, -41490.73125, -5049.84375, 278498.503125, 0, 340634.3625, 6629.34375000005, 103210.453125, -753879.403125, 0, -1394602.059375, -4305.65625000028, -763023.796874998, -126606.796875, 1222683.35625, 0, 3127751.71875001, 1107.16875000051, 2661818.20312499, 76579.1718750024, 832221.843749999, -1172491.70625, 0, -3902769.84375001, 0, -4691627.57812499, -18268.2843750038, -2383180.73437499, -440099.578125006, 612817.903125, 0, 2540952.28125001, 0, 4035630.09374999, 0, 2989355.62499998, 89680.6687500077, 971540.578125002, -134521.003125, 0, -672605.015625002, 0, -1345210.03125, 0, -1345210.03124999, 0, -672605.015624997, -134521.003125005}, 
                                                  {0, 226.8, 0, -5366.25, 0, -1113.75, 49708.6875, 0, 24953.0625, 2035.125, -240251.0625, 0, -212799.65625, -1640.25000000005, -43079.3437499999, 671948.915625, 0, 913332.206250001, 492.075000000244, 335615.653124999, 32722.9875000003, -1125621.5625, 0, -2158979.0625, 0, -1254791.25, -9226.40625000191, -230660.15625, 1112704.59375, 0, 2839887.84375001, 0, 2399788.26562499, 0, 730731.374999995, 58126.3593750046, -597871.125, 0, -1943081.15625, 0, -2242016.71875, 0, -1046274.46874999, 0, -149467.781250002, 134521.003125, 0, 538084.012500001, 0, 807126.018749999, 0, 538084.012499995, 0, 134521.003124998, 0}, 
                                                  {0, -126, 0, 3019.05, 0, 429.750000000001, -28453.6125, 0, -9803.36250000001, -486.000000000001, 140501.08125, 0, 85789.6312500001, 182.250000000027, 10534.05, -402804.39375, 0, -380770.368750001, 0, -85046.9624999996, -3745.23750000016, 693087.6375, 0, 937402.875000002, 0, 333995.906249998, 0, 27679.21875, -704159.325, 0, -1290958.7625, 0, -680908.781249997, 0, -94109.3437499981, 0, 388616.23125, 0, 926700.243750002, 0, 687551.793749998, 0, 149467.781249996, 0, 0, -89680.6687499999, 0, -269042.006250001, 0, -269042.00625, 0, -89680.6687499981, 0, 0, 0}, 
                                                  {0, 46.2857142857142, 0, -1118.95714285714, 0, -98.357142857143, 10677.2464285714, 0, 2272.91785714286, 52.0714285714289, -53563.275, 0, -20266.2, 0, -1148.17499999999, 156520.85625, 0, 92223.0562500001, 0, 9513.44999999991, 0, -275315.9625, 0, -234350.71875, 0, -38750.9062499996, 0, 0, 286598.539285714, 0, 335472.131250001, 0, 83037.6562499994, 0, 0, 0, -162279.305357143, 0, -251959.974107143, 0, -89680.6687499995, 0, 0, 0, 0, 38434.5723214285, 0, 76869.1446428573, 0, 38434.5723214285, 0, 0, 0, 0, 0}, 
                                                  {0, -10.125, 0, 246.399107142857, 0, 10.125, -2373.15535714286, 0, -236.274107142857, 0, 12050.1421875, 0, 2136.88125, 0, 0, -35747.1984375, 0, -9913.26093750001, 0, 0, 0, 64031.259375, 0, 25833.9375, 0, 0, 0, 0, -68090.878125, 0, -38197.3218750001, 0, 0, 0, 0, 0, 39502.1993303571, 0, 29893.55625, 0, 0, 0, 0, 0, 0, -9608.64308035714, 0, -9608.64308035716, 0, 0, 0, 0, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, -10.125, 0, 0, 236.274107142858, 0, 0, 0, -2136.88125, 0, 0, 0, 0, 9913.26093750002, 0, 0, 0, 0, 0, -25833.9375000001, 0, 0, 0, 0, 0, 0, 38197.3218750001, 0, 0, 0, 0, 0, 0, 0, -29893.5562500001, 0, 0, 0, 0, 0, 0, 0, 0, 9608.64308035716, 0, 0, 0, 0, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, -5.7857142857143, 0, 0, 127.575, 52.0714285714286, 0, 0, -1057.05, 0, -1148.175, 0, 0, 4305.65625000001, 0, 9513.44999999999, 0, 0, 0, -9226.40625000003, 0, -38750.90625, 0, 0, 0, 0, 9964.51875000004, 0, 83037.6562499999, 0, 0, 0, 0, 0, -4270.50803571432, 0, -89680.6687499999, 0, 0, 0, 0, 0, 0, 0, 0, 38434.5723214286, 0, 0, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, -4.49999999999999, 0, 0, 92.475, 60.7499999999998, 0, 0, -683.4375, -182.249999999999, -1248.4125, 0, 0, 2323.6875, 0, 9226.40624999999, 3745.23749999999, 0, 0, -3690.56250000001, 0, -31369.78125, 0, -27679.2187499999, 0, 0, 2214.33750000002, 0, 49822.59375, 0, 94109.3437499998, 0, 0, 0, 0, 0, -29893.55625, 0, -149467.78125, 0, 0, 0, 0, 0, 0, 0, 0, 89680.66875, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, -4.04999999999999, 0, 0, 75.9375, 66.8249999999998, 0, 0, -478.40625, -328.049999999998, -1252.96875, 0, 0, 1230.1875, 492.074999999993, 7893.70312499999, 6150.93749999999, 0, 0, -1107.16875, 0, -20298.09375, -9226.40624999998, -38750.90625, 0, 0, 0, 0, 18268.284375, 0, 99645.1874999999, 58126.359375, 0, 0, 0, 0, 0, 0, -89680.6687499999, 0, -149467.78125, 0, 0, 0, 0, 0, 0, 0, 0, 134521.003125, 0}, 
                                                  {0, 0, 0, 0, 0, -4.04999999999999, 0, 0, 66.825, 75.9374999999998, 0, 0, -328.05, -478.406249999999, -1252.96875, 0, 0, 492.075, 1230.18749999999, 6150.9375, 7893.70312499999, 0, 0, 0, -1107.16874999999, -9226.40624999999, -20298.09375, -38750.90625, 0, 0, 0, 0, 0, 18268.284375, 58126.359375, 99645.1875, 0, 0, 0, 0, 0, 0, 0, -89680.66875, -149467.78125, 0, 0, 0, 0, 0, 0, 0, 0, 0, 134521.003125}, 
                                                  {0, 0, 0, 0, 0, -4.49999999999999, 0, 0, 60.7499999999999, 92.4749999999998, 0, 0, -182.249999999999, -683.437499999999, -1248.4125, 0, 0, 0, 2323.68749999999, 3745.23749999999, 9226.40624999999, 0, 0, 0, -3690.56249999999, 0, -31369.78125, -27679.21875, 0, 0, 0, 2214.33749999998, 0, 49822.59375, 0, 94109.3437499999, 0, 0, 0, 0, 0, -29893.55625, 0, -149467.78125, 0, 0, 0, 0, 0, 0, 0, 0, 89680.66875, 0, 0}, 
                                                  {0, 0, 0, 0, 0, -5.78571428571428, 0, 0, 52.0714285714285, 127.575, 0, 0, 0, -1057.05, -1148.175, 0, 0, 0, 4305.65624999999, 0, 9513.45, 0, 0, 0, -9226.40624999998, 0, -38750.90625, 0, 0, 0, 0, 9964.51874999997, 0, 83037.65625, 0, 0, 0, 0, 0, -4270.50803571427, 0, -89680.66875, 0, 0, 0, 0, 0, 0, 0, 0, 38434.5723214286, 0, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, -10.125, 0, 0, 0, 236.274107142858, 0, 0, 0, -2136.88125, 0, 0, 0, 0, 9913.26093750001, 0, 0, 0, 0, 0, -25833.9375, 0, 0, 0, 0, 0, 0, 38197.321875, 0, 0, 0, 0, 0, 0, 0, -29893.55625, 0, 0, 0, 0, 0, 0, 0, 0, 9608.64308035714, 0, 0, 0, 0, 0, 0}, 
                                                  {0, 0, -10.1249999999998, 0, 246.399107142852, 10.124999999999, 0, -2373.15535714281, 0, -236.274107142814, 0, 12050.1421874998, 0, 2136.88124999954, 0, 0, -35747.1984374994, 0, -9913.2609374979, 0, 0, 0, 64031.2593749991, 0, 25833.9374999951, 0, 0, 0, 0, -68090.8781249991, 0, -38197.3218749939, 0, 0, -1.263980062129e-08, 0, 0, 39502.1993303567, 0, 29893.5562499961, 0, 0, 1.18739934339215e-08, 0, 1.05259715868624e-08, 0, -9608.64308035705, 0, -9608.64308035611, 0, 0, 0, 0, 0, 0}, 
                                                  {0, 0, 46.2857142857133, 0, -1118.95714285712, -98.3571428571389, 0, 10677.2464285712, 52.0714285715498, 2272.91785714268, 0, -53563.2749999991, 0, -20266.1999999981, -1148.17500000104, 0, 156520.856249998, 0, 92223.0562499911, 0, 9513.45000000109, 0, -275315.962499996, 0, -234350.718749979, -2.24942243522263e-08, -38750.9062499918, -2.33477007656512e-08, 0, 286598.539285711, 0, 335472.131249974, 2.92827757446078e-08, 83037.6562499757, 5.49652465868241e-08, 2.09286072194397e-08, 0, -162279.305357141, 0, -251959.974107126, -1.87094831968936e-08, -89680.6687499757, -5.13642589033018e-08, 0, -4.65206095618329e-08, 0, 38434.5723214282, 0, 76869.1446428528, 0, 38434.5723214201, 1.70705524401938e-08, 0, 2.45193531860009e-08, 1.09791205822977e-08}, 
                                                  {0, 0, -125.999999999998, 0, 3019.04999999995, 429.74999999999, 0, -28453.6124999995, -486.000000000304, -9803.36249999956, 0, 140501.081249998, 182.250000002224, 85789.6312499952, 10534.0500000027, 0, -402804.393749994, 0, -380770.368749978, -3745.2375000212, -85046.9625000034, 0, 693087.637499991, 1.22872607005248e-08, 937402.874999949, 5.63421767849967e-08, 333995.906249981, 27679.2187500602, 0, -704159.324999991, -1.24967588143699e-08, -1290958.76249994, -7.27375218962187e-08, -680908.781249942, -1.39882793503198e-07, -94109.3437500562, 0, 388616.231249996, 0, 926700.243749959, 4.59995752007917e-08, 687551.793749941, 1.29912668443732e-07, 149467.781250003, 1.20228278800505e-07, 0, -89680.668749999, 0, -269042.006249989, -1.14619250670293e-08, -269042.006249979, -4.30101959688498e-08, -89680.668749988, -6.26653484738108e-08, -2.926747049186e-08}, 
                                                  {0, 0, 226.799999999996, 0, -5366.24999999991, -1113.74999999999, 0, 49708.6874999992, 2035.12500000049, 24953.0624999993, 0, -240251.062499997, -1640.25000000355, -212799.656249992, -43079.3437500045, 0, 671948.915624991, 492.075000011098, 913332.206249965, 32722.9875000344, 335615.653125006, 0, -1125621.56249999, -1.92194569038282e-08, -2158979.06249992, -9226.40625009077, -1254791.24999997, -230660.1562501, 0, 1112704.59374999, 1.93975939348018e-08, 2839887.8437499, 1.16126534812891e-07, 2399788.26562491, 58126.3593752293, 730731.375000097, 0, -597871.124999993, -1.08649442281946e-08, -1943081.15624993, -7.26261457885223e-08, -2242016.71874991, -2.11581330444429e-07, -1046274.46875001, -149467.7812502, 0, 134521.003124999, 0, 538084.012499983, 1.78729989584236e-08, 807126.018749968, 6.97686427639398e-08, 538084.012499983, 1.03100390548607e-07, 134521.00312505}, 
                                                  {0, 0, -283.499999999996, 0, 6580.23749999991, 1902.48749999998, 0, -59404.3874999992, -5049.84375000053, -41490.7312499993, 0, 278498.503124996, 6629.34375000376, 340634.362499992, 103210.453125005, 0, -753879.403124991, -4305.65625001166, -1394602.05937496, -126606.796875037, -763023.796875008, 0, 1222683.35624999, 1107.16875002001, 3127751.71874991, 76579.1718750974, 2661818.20312497, 832221.84375011, 0, -1172491.70624999, -2.00399073697498e-08, -3902769.84374989, -18268.2843751234, -4691627.57812491, -440099.578125251, -2383180.73437511, 0, 612817.903124993, 1.11780825601743e-08, 2540952.28124993, 7.62658325806891e-08, 4035630.0937499, 89680.6687502297, 2989355.62500002, 971540.578125221, 0, -134521.003124998, 0, -672605.015624981, -1.84983795405654e-08, -1345210.03124997, -7.54539905347877e-08, -1345210.03124998, -134521.003125113, -672605.015625057}, 
                                                  {0, 0, 251.999999999997, 0, -5660.09999999994, -2258.09999999999, 0, 48984.0749999995, 8293.72500000037, 47071.7999999995, 0, -219410.774999998, -15992.4375000027, -363360.937499994, -159942.600000004, 0, 568879.706249993, 17085.9375000082, 1387829.19374997, 284241.656250027, 1086105.20625001, 0, -887949.33749999, -9595.4625000139, -2913699.09374994, -278637.468750069, -3422996.71874998, -1666288.96875008, 0, 823733.549999991, 2214.33750001384, 3428901.61874992, 142824.768750087, 5480485.31249994, 1372889.25000018, 4107596.06250008, 0, -418509.787499995, 0, -2122442.49374995, -29893.5562500532, -4334565.65624993, -567977.568750165, -4484033.43750002, -2391484.50000016, 0, 89680.668749999, 0, 538084.012499987, 1.27003464836953e-08, 1345210.03124998, 89680.6687500542, 1793613.37499999, 538084.012500082, 1345210.03125004}, 
                                                  {0, 0, -161.999999999999, 0, 3395.63571428569, 1937.63571428571, 0, -27125.3089285712, -9686.58750000017, -36811.8964285712, 0, 113455.181249999, 26275.8937500012, 253186.256249997, 166006.968750002, 0, -278309.418749997, -41826.3750000037, -876754.631249988, -403788.543750012, -960407.381250003, 0, 415557.337499995, 39119.9625000063, 1701349.31249997, 572037.187500031, 2649823.87499999, 1896949.12500004, 0, -372008.699999996, -19929.0375000063, -1879972.53749996, -471653.887500039, -3819732.18749997, -2059333.87500008, -3919377.37500004, 0, 183631.845535712, 4270.50803571785, 1106061.58124998, 209254.893750024, 2780100.73124997, 1165848.69375008, 3736694.53125001, 2839887.84375007, 0, -38434.5723214281, 0, -269042.006249994, -38434.5723214341, -807126.018749989, -269042.006250025, -1345210.03125, -807126.018750037, -1345210.03125002}, 
                                                  {0, 0, 80.9999999999997, 0, -1333.31785714285, -1333.31785714286, 0, 9202.97410714279, 9202.9741071429, 18405.9482142856, 0, -34932.7687499997, -34932.7687500003, -104798.306249999, -104798.30625, 0, 80033.9484374993, 80033.948437501, 320135.793749996, 320135.793750003, 480203.690625001, 0, -113669.324999999, -113669.325000002, -568346.624999992, -568346.625000008, -1136693.25, -1136693.25000001, 0, 97984.4343749989, 97984.434375002, 587906.60624999, 587906.606250009, 1469766.51562499, 1469766.51562502, 1959688.68750001, 0, -46975.5883928566, -46975.5883928583, -328829.118749993, -328829.118750006, -986487.356249991, -986487.356250019, -1644145.59375, -1644145.59375002, 0, 9608.64308035703, 9608.64308035746, 76869.1446428553, 76869.1446428585, 269042.006249997, 269042.006250006, 538084.012499999, 538084.012500009, 672605.015625005}, 
                                                  {0, 0, 0, 0, 0, 2916, 0, 0, -34877.4428571428, -34877.4428571428, 0, 0, 174358.575, 174358.575, 348717.15, 0, 0, -472966.0875, -472966.087499999, -1418898.2625, -1418898.2625, 0, 0, 752874.750000001, 752874.749999998, 3011498.99999999, 3011499, 4517248.49999999, 0, 0, -704159.325000002, -704159.324999997, -3520796.62499999, -3520796.625, -7041593.24999999, -7041593.24999999, 0, 0, 358722.675000002, 358722.674999998, 2152336.05, 2152336.05, 5380840.12499999, 5380840.125, 7174453.49999999, 0, 0, -76869.1446428578, -76869.1446428566, -538084.012499999, -538084.012499999, -1614252.03749999, -1614252.0375, -2690420.06249999, -2690420.0625}, 
                                                  {0, 0, 0, 0, 0, 104.142857142857, 0, 0, -2400.49285714286, -104.142857142858, 0, 0, 21323.25, 0, 2296.35, 0, 0, -96528.7125000002, 0, -19026.8999999999, 0, 0, 0, 243577.125, 0, 77501.8124999997, 0, 0, 0, 0, -345436.650000001, 0, -166075.312499999, 0, 0, 0, 0, 0, 256230.482142858, 0, 179361.3375, 0, 0, 0, 0, 0, 0, -76869.1446428573, 0, -76869.144642857, 0, 0, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, 104.142857142857, 0, 0, -104.142857142859, -2400.49285714286, 0, 0, 0, 21323.25, 2296.35000000003, 0, 0, 0, -96528.7125, 0, -19026.9000000002, 0, 0, 0, 243577.125, 0, 77501.8125000005, 0, 0, 0, 0, -345436.65, 0, -166075.312500001, 0, 0, 0, 0, 0, 256230.482142857, 0, 179361.3375, 0, 0, 0, 0, 0, 0, -76869.1446428571, 0, -76869.1446428572, 0, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, -3402, 0, 0, 61102.35, 30484.3499999999, 0, 0, -386324.4375, -111965.2875, -498289.724999999, 0, 0, 1223585.49375, 215897.906249999, 2663068.89375, 1655381.30625, 0, 0, -2173741.3125, -230660.156249997, -6751884.09374999, -2865721.78125, -7213204.40624999, 0, 0, 2205480.15000001, 129538.743749997, 8951459.34374999, 2723635.125, 13751035.875, 9599153.06249999, 0, 0, -1195742.25000001, -29893.5562499978, -6008604.80624999, -1345210.03125, -12106890.28125, -6277646.8125, -12256358.0625, 0, 0, 269042.006250002, 0, 1614252.0375, 269042.006249999, 4035630.09374999, 1614252.0375, 5380840.12499999, 4035630.09375}, 
                                                  {0, 0, 0, 0, 0, 3402, 0, 0, -68756.85, -22829.85, 0, 0, 506582.1, 60598.1249999998, 429399.224999999, 0, 0, -1822235.7375, -79552.1249999995, -2799414.675, -1056731.0625, 0, 0, 3579845.62500001, 51667.8749999988, 8336980.68749999, 1280625.1875, 5986092.37499999, 0, 0, -3932663.40000001, -13286.0249999983, -12522078.5625, -763946.4375, -13983541.3125, -6144786.5625, 0, 0, 2271910.27500001, 0, 9267002.43749999, 179361.337499999, 14348907, 2989355.625, 10163809.125, 0, 0, -538084.012500003, 0, -2690420.0625, 0, -5380840.12499998, -538084.012500001, -5380840.12499999, -2690420.0625}, 
                                                  {0, 0, 0, 0, 0, -2551.5, 0, 0, 54629.4375000001, 12529.6875, 0, 0, -436306.500000001, -22895.15625, -252530.15625, 0, 0, 1721134.828125, 18452.8125, 1825803.28125, 433128.515625, 0, 0, -3686871.93750001, -5535.84375000002, -6166929.93749999, -326614.781250002, -2801136.9375, 0, 0, 4367780.71875001, 0, 10412922.09375, 91341.4218750029, 7813843.45312498, 1860043.5, 0, 0, -2690420.06250001, 0, -8519663.53124999, 0, -9416470.21874997, -448403.343750007, -4035630.09375, 0, 0, 672605.015625003, 0, 2690420.0625, 0, 4035630.09374999, 0, 2690420.06249999, 672605.015625004}, 
                                                  {0, 0, 0, 0, 0, 1360.8, 0, 0, -30156.3, -4641.3, 0, 0, 253017.675, 5248.80000000004, 97521.9749999999, 0, 0, -1061979.8625, -1968.30000000021, -750988.462499999, -104319.9, 0, 0, 2438723.7, 0, 2760540.75, 36905.6250000019, 730731.374999999, 0, 0, -3095643.82500001, 0, -5154977.69999999, 0, -2291839.31249999, -232505.437500004, 0, 0, 2032761.82500001, 0, 4663394.77499999, 0, 3228504.07499999, 0, 597871.125000001, 0, 0, -538084.012500002, 0, -1614252.0375, 0, -1614252.03749999, 0, -538084.012499997, 0}, 
                                                  {0, 0, 0, 0, 0, -486.000000000001, 0, 0, 11020.05, 1032.75, 0, 0, -95581.0125000002, -546.750000000027, -22316.5125, 0, 0, 419042.868750001, 0, 179320.33125, 11235.7125000002, 0, 0, -1014904.6875, 0, -699361.593749998, 0, -83037.6562499999, 0, 0, 1368460.575, 0, 1411640.15625, 0, 282328.031249998, 0, 0, 0, -956593.800000002, 0, -1404997.14375, 0, -448403.343749995, 0, 0, 0, 0, 269042.006250001, 0, 538084.0125, 0, 269042.006249998, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, 60.7500000000001, 0, 0, -1309.1625, -607.499999999999, 0, 0, 10474.81875, 546.749999999998, 13030.875, 0, 0, -40596.1875000001, 0, -103499.775, -11235.7125, 0, 0, 81192.3750000002, 0, 396735.46875, 0, 83037.6562499998, 0, 0, -79716.1500000003, 0, -780553.968749999, 0, -282328.031249999, 0, 0, 0, 29893.5562500003, 0, 747338.90625, 0, 448403.343749999, 0, 0, 0, 0, 0, 0, -269042.00625, 0, -269042.006249999, 0, 0, 0}, 
                                                  {0, 0, 0, 0, 0, 48.5999999999999, 0, 0, -959.85, -704.699999999998, 0, 0, 6652.125, 2624.39999999998, 13869.225, 0, 0, -20503.125, -1968.29999999994, -95544.5624999999, -51175.7999999999, 0, 0, 28048.2750000001, 0, 291554.4375, 36905.6249999997, 346912.874999999, 0, 0, -13286.0250000001, 0, -391937.7375, 0, -1029666.9375, -232505.437499999, 0, 0, 0, 0, 179361.3375, 0, 1315316.475, 0, 597871.124999999, 0, 0, 0, 0, 0, 0, -538084.0125, 0, -538084.012499999, 0}, 
                                                  {0, 0, 0, 0, 0, 45.5624999999999, 0, 0, -797.34375, -797.343749999998, 0, 0, 4442.34375, 4442.34374999998, 13907.953125, 0, 0, -9226.40625, -9226.40624999993, -76989.2343749999, -76989.2343749999, 0, 0, 5535.84375000003, 5535.84374999986, 157771.546875, 157771.546875, 420724.125, 0, 0, 0, 0, -91341.4218749999, -91341.4218749997, -838680.328125, -838680.328125, 0, 0, 0, 0, 0, 0, 448403.343749999, 448403.34375, 1569411.703125, 0, 0, 0, 0, 0, 0, 0, 0, -672605.015625, -672605.015625}, 
                                                  {0, 0, 0, 0, 0, 48.5999999999999, 0, 0, -704.699999999999, -959.849999999998, 0, 0, 2624.39999999999, 6652.12499999998, 13869.225, 0, 0, -1968.29999999999, -20503.1249999999, -51175.7999999999, -95544.5624999999, 0, 0, 0, 28048.2749999999, 36905.6249999998, 291554.4375, 346912.875, 0, 0, 0, -13286.0249999998, 0, -391937.7375, -232505.437499999, -1029666.9375, 0, 0, 0, 0, 0, 179361.3375, 0, 1315316.475, 597871.124999999, 0, 0, 0, 0, 0, 0, 0, -538084.0125, 0, -538084.0125}, 
                                                  {0, 0, 0, 0, 0, 60.7499999999999, 0, 0, -607.499999999999, -1309.1625, 0, 0, 546.749999999995, 10474.81875, 13030.875, 0, 0, 0, -40596.1874999999, -11235.7124999999, -103499.775, 0, 0, 0, 81192.3749999998, 0, 396735.46875, 83037.6562499998, 0, 0, 0, -79716.1499999998, 0, -780553.96875, 0, -282328.03125, 0, 0, 0, 29893.5562499998, 0, 747338.90625, 0, 448403.34375, 0, 0, 0, 0, 0, 0, -269042.00625, 0, -269042.00625, 0, 0}, 
                                                  {0, 0, 0, 0, 0, -486, 0, 0, 1032.75, 11020.05, 0, 0, -546.750000000009, -95581.0125, -22316.5125000001, 0, 0, 0, 419042.86875, 11235.7125000003, 179320.331250001, 0, 0, 0, -1014904.6875, 0, -699361.593750002, -83037.6562500022, 0, 0, 0, 1368460.575, 0, 1411640.15625, 0, 282328.031250004, 0, 0, 0, -956593.8, 0, -1404997.14375, 0, -448403.343750003, 0, 0, 0, 0, 269042.00625, 0, 538084.0125, 0, 269042.006250001, 0, 0}, 
                                                  {0, 0, 0, 0, 0, 1360.8, 0, 0, -4641.30000000001, -30156.3, 0, 0, 5248.8, 253017.675, 97521.9750000001, 0, 0, -1968.2999999999, -1061979.8625, -104319.9, -750988.462500001, 0, 0, 0, 2438723.7, 36905.6249999997, 2760540.75, 730731.375000004, 0, 0, 0, -3095643.825, 0, -5154977.7, -232505.437500006, -2291839.31250001, 0, 0, 0, 2032761.825, 0, 4663394.775, 0, 3228504.07500001, 597871.125000008, 0, 0, 0, -538084.0125, 0, -1614252.0375, 0, -1614252.0375, 0, -538084.012500004}, 
                                                  {0, 0, 0, 0, 0, -2551.5, 0, 0, 12529.6875, 54629.4375, 0, 0, -22895.1562499999, -436306.5, -252530.15625, 0, 0, 18452.8124999998, 1721134.828125, 433128.515625, 1825803.28125, 0, 0, -5535.8437499995, -3686871.9375, -326614.781249998, -6166929.9375, -2801136.9375, 0, 0, 0, 4367780.71875, 91341.4218749951, 10412922.09375, 1860043.5, 7813843.453125, 0, 0, 0, -2690420.0625, 0, -8519663.53125, -448403.34375, -9416470.21875, -4035630.09375001, 0, 0, 0, 672605.015624999, 0, 2690420.0625, 0, 4035630.09375, 672605.015625001, 2690420.0625}, 
                                                  {0, 0, 0, 0, 0, 3402, 0, 0, -22829.85, -68756.8499999999, 0, 0, 60598.1249999999, 506582.1, 429399.225, 0, 0, -79552.1249999997, -1822235.7375, -1056731.0625, -2799414.675, 0, 0, 51667.8749999996, 3579845.625, 1280625.1875, 8336980.6875, 5986092.375, 0, 0, -13286.0249999998, -3932663.39999999, -763946.437499993, -12522078.5625, -6144786.56249999, -13983541.3125, 0, 0, 0, 2271910.275, 179361.337499994, 9267002.4375, 2989355.62499999, 14348907, 10163809.125, 0, 0, 0, -538084.012499999, 0, -2690420.0625, -538084.012499998, -5380840.125, -2690420.0625, -5380840.125}, 
                                                  {0, 0, 0, 0, 0, -3402, 0, 0, 30484.35, 61102.3499999999, 0, 0, -111965.2875, -386324.4375, -498289.725, 0, 0, 215897.90625, 1223585.49375, 1655381.30625, 2663068.89375, 0, 0, -230660.15625, -2173741.3125, -2865721.78124999, -6751884.09374999, -7213204.40624999, 0, 0, 129538.743750001, 2205480.14999999, 2723635.12499999, 8951459.34374999, 9599153.06249998, 13751035.875, 0, 0, -29893.5562500009, -1195742.25, -1345210.03124999, -6008604.80624999, -6277646.81249999, -12106890.28125, -12256358.0625, 0, 0, 0, 269042.006249999, 269042.006249999, 1614252.0375, 1614252.0375, 4035630.09375, 4035630.09374999, 5380840.125}, 
                                                  {0, 0, 0, 0, 0, 2551.5, 0, 0, -40085.8875, -40085.8874999999, 0, 0, 199550.08125, 199550.081249999, 605771.662499999, 0, 0, -468701.4375, -468701.437499998, -2793017.7, -2793017.7, 0, 0, 575727.750000002, 575727.749999995, 5984247.09374999, 5984247.09374999, 10817038.6875, 0, 0, -358722.675000003, -358722.674999994, -6626404.96874999, -6626404.96874999, -18085601.53125, -18085601.53125, 0, 0, 89680.6687500023, 89680.6687499959, 3676907.41875, 3676907.41874999, 13900503.65625, 13900503.65625, 20626553.8125, 0, 0, 0, 0, -807126.018749999, -807126.018749997, -4035630.09374999, -4035630.09375, -8071260.18749999, -8071260.18749999}, 
                                                  {0, 0, 0, 0, 0, -291.6, 0, 0, 6087.15, 3244.04999999999, 0, 0, -46391.7375000001, -5904.89999999997, -67058.8874999999, 0, 0, 167920.59375, 2952.44999999992, 503023.668749999, 116621.775, 0, 0, -306685.743750001, 0, -1777005.84375, -55358.4374999997, -808233.187499998, 0, 0, 269042.006250001, 0, 3128858.8875, 0, 2491129.6875, 348758.156249999, 0, 0, -89680.6687500011, 0, -2600739.39375, 0, -3407865.41249999, 0, -896806.687499998, 0, 0, 0, 0, 807126.01875, 0, 1614252.0375, 0, 807126.018749998, 0}, 
                                                  {0, 0, 0, 0, 0, -291.599999999999, 0, 0, 3244.04999999999, 6087.14999999999, 0, 0, -5904.89999999997, -46391.7374999999, -67058.8874999999, 0, 0, 2952.44999999992, 167920.59375, 116621.775, 503023.66875, 0, 0, 0, -306685.743749999, -55358.4374999991, -1777005.84375, -808233.187499999, 0, 0, 0, 269042.006249999, 0, 3128858.8875, 348758.156249998, 2491129.6875, 0, 0, 0, -89680.6687499993, 0, -2600739.39375, 0, -3407865.4125, -896806.687499998, 0, 0, 0, 0, 0, 807126.01875, 0, 1614252.0375, 0, 807126.018749999}, 
                                                  {0, 0, 0, 0, 0, -1701, 0, 0, 31316.625, 23662.1249999999, 0, 0, -196921.125, -90441.5624999996, -425143.687499999, 0, 0, 556659.843750001, 149672.812499999, 2559610.125, 1532608.59375, 0, 0, -787935.093750002, -114407.437499997, -6759265.21874999, -2367495.84375, -8224418.53124999, 0, 0, 548048.531250004, 33215.0624999965, 8785384.03124999, 1677360.65624999, 17537553, 10944363.09375, 0, 0, -149467.781250003, 0, -5530307.90625, -448403.343749995, -16142520.375, -6427114.59374999, -16740391.5, 0, 0, 0, 0, 1345210.03125, 0, 5380840.12499999, 1345210.03125, 8071260.18749999, 5380840.12499999}, 
                                                  {0, 0, 0, 0, 0, 850.5, 0, 0, -16934.0625, -10555.3125, 0, 0, 120034.40625, 29387.8124999999, 206830.96875, 0, 0, -393660.000000001, -30754.6874999995, -1427222.53125, -547638.468749999, 0, 0, 638467.312500002, 11071.687499999, 4487724, 540667.406249998, 3448830.65624999, 0, 0, -498225.937500003, 0, -6842302.87499999, -182682.843749998, -9200572.31249999, -3039178.21875, 0, 0, 149467.781250002, 0, 4932436.78125, 0, 10313276.90625, 896806.687499998, 6427114.59374999, 0, 0, 0, 0, -1345210.03125, 0, -4035630.09374999, 0, -4035630.09374999, -1345210.03125}, 
                                                  {0, 0, 0, 0, 0, -243, 0, 0, 4525.875, 3796.87499999999, 0, 0, -28476.5625, -17085.9374999999, -70166.2499999999, 0, 0, 75861.5625000001, 24603.7499999997, 435281.34375, 310212.281249999, 0, 0, -84882.9375000003, -11071.6874999994, -1131157.40625, -428105.249999999, -1861888.78125, 0, 0, 33215.0625000005, 0, 1212349.78125, 182682.843749999, 4550463.5625, 2358269.4375, 0, 0, 0, 0, -448403.34375, 0, -4334565.65625, -896806.687499998, -4782969, 0, 0, 0, 0, 0, 0, 1345210.03125, 0, 2690420.0625, 1345210.03125}, 
                                                  {0, 0, 0, 0, 0, -242.999999999999, 0, 0, 3796.875, 4525.87499999999, 0, 0, -17085.9375, -28476.5624999999, -70166.2499999999, 0, 0, 24603.75, 75861.5624999997, 310212.28125, 435281.34375, 0, 0, -11071.6875, -84882.9374999993, -428105.249999999, -1131157.40625, -1861888.78125, 0, 0, 0, 33215.0624999991, 182682.843749999, 1212349.78125, 2358269.4375, 4550463.5625, 0, 0, 0, 0, 0, -448403.343749999, -896806.687499998, -4334565.65625, -4782969, 0, 0, 0, 0, 0, 0, 0, 1345210.03125, 1345210.03125, 2690420.0625}, 
                                                  {0, 0, 0, 0, 0, 850.499999999998, 0, 0, -10555.3125, -16934.0625, 0, 0, 29387.8124999999, 120034.40625, 206830.96875, 0, 0, -30754.6874999998, -393659.999999999, -547638.468749999, -1427222.53125, 0, 0, 11071.6874999998, 638467.312499998, 540667.406249997, 4487724, 3448830.65625, 0, 0, 0, -498225.937499997, -182682.843749997, -6842302.875, -3039178.21874999, -9200572.31249999, 0, 0, 0, 149467.781249998, 0, 4932436.78125, 896806.687499993, 10313276.90625, 6427114.59374999, 0, 0, 0, 0, 0, -1345210.03125, 0, -4035630.09375, -1345210.03125, -4035630.09375}, 
                                                  {0, 0, 0, 0, 0, -1701, 0, 0, 23662.125, 31316.6249999999, 0, 0, -90441.5624999999, -196921.124999999, -425143.687499999, 0, 0, 149672.8125, 556659.843749998, 1532608.59375, 2559610.125, 0, 0, -114407.4375, -787935.093749996, -2367495.84374999, -6759265.21874999, -8224418.53124999, 0, 0, 33215.0625000006, 548048.531249995, 1677360.65624999, 8785384.03124999, 10944363.09375, 17537553, 0, 0, 0, -149467.781249996, -448403.343749997, -5530307.90624999, -6427114.59374998, -16142520.375, -16740391.5, 0, 0, 0, 0, 0, 1345210.03125, 1345210.03125, 5380840.125, 5380840.12499999, 8071260.18749999}, 
                                                  {0, 0, 0, 0, 0, 755.999999999998, 0, 0, -12784.5, -12784.5, 0, 0, 68343.75, 68343.7499999998, 213232.5, 0, 0, -144888.75, -144888.749999999, -1109082.375, -1109082.375, 0, 0, 132860.25, 132860.249999998, 2236480.875, 2236480.875, 5447270.24999999, 0, 0, -44286.7500000008, -44286.7499999975, -1926473.625, -1926473.625, -9787371.74999999, -9787371.74999999, 0, 0, 0, 0, 597871.124999999, 597871.124999996, 7174453.49999999, 7174453.49999999, 13153164.75, 0, 0, 0, 0, 0, 0, -1793613.375, -1793613.375, -5380840.12499999, -5380840.125}};




double MAd_TriInterpolant::coef10[66][66] = {{1, -29.2896825396823, -29.2896825396831, 351.45436507936, 351.45436507938, 702.908730158736, -2317.43276014103, -2317.43276014125, -6952.29828042333, -6952.29828042342, 9416.14307760106, 9416.14307760232, 37664.5723104059, 37664.5723104068, 56496.8584656094, -24858.2175925913, -24858.2175925956, -124291.087962964, -124291.087962968, -248582.175925931, -248582.175925932, 43478.0092592564, 43478.0092592657, 260868.055555557, 260868.05555557, 652170.138888905, 652170.138888908, 869560.185185211, -50016.5343915303, -50016.5343915428, -350115.740740742, -350115.740740763, -1050347.22222225, -1050347.22222225, -1750578.70370377, -1750578.70370377, 36375.6613756577, 36375.6613756682, 291005.291005291, 291005.291005313, 1018518.51851855, 1018518.51851855, 2037037.03703713, 2037037.03703711, 2546296.2962964, -15156.5255731905, -15156.5255731953, -136408.73015873, -136408.730158742, -545634.920634936, -545634.920634932, -1273148.14814821, -1273148.1481482, -1909722.22222232, -1909722.2222223, 2755.73192239822, 2755.73192239918, 27557.3192239857, 27557.3192239886, 124007.93650794, 124007.936507938, 330687.830687849, 330687.830687842, 578703.70370374, 578703.703703729, 694444.444444475}, 
                                                   {0, -0.999999999999925, 0, 28.2896825396804, 0, 0, -323.164682539659, 0, 0, 0, 1994.26807760127, 0, 0, 0, 0, -7421.8749999995, 0, 0, 0, 0, 0, 17436.3425925915, 0, 0, 0, 0, 0, 0, -26041.6666666651, 0, 0, 0, 0, 0, 0, 0, 23974.8677248663, 0, 0, 0, 0, 0, 0, 0, 0, -12400.793650793, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2755.73192239845, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                   {0, 0, -1.00000000000021, 0, 28.2896825396883, 0, 0, -323.164682539743, 0, 0, 0, 1994.26807760175, 0, 0, 0, 0, -7421.87500000114, 0, 0, 0, 0, 0, 17436.342592595, 0, 0, 0, 1.50946202204238e-08, 1.1645162611984e-08, 0, -26041.6666666698, 0, -1.31843545733701e-08, -1.11561089810586e-08, -3.24960945929931e-08, -2.15227413669083e-08, -3.40096555362379e-08, 0, 23974.8677248702, 0, 1.22743567029099e-08, 1.26464469729637e-08, 3.86618165260089e-08, 2.81839873661482e-08, 5.56401786866445e-08, 3.92080341526935e-08, 0, -12400.7936507947, 0, 0, 0, -2.36153814558993e-08, -2.20441492837509e-08, -4.58821665105802e-08, -2.75295248819418e-08, -3.96503159194594e-08, 0, 2755.7319223988, 0, 0, 0, 0, 0, 1.46434050440902e-08, 0, 1.64801710056395e-08, 1.12944789076863e-08}, 
                                                   {0, 99.9999999999983, 0, -1928.96825396821, 0, -1928.96825396825, 15855.7539682534, 0, 31711.5079365079, 15855.7539682539, -73185.7363315665, 0, -219557.208994709, -73185.7363315692, -219557.208994708, 209756.944444433, 0, 839027.777777779, 209756.944444442, 1258541.66666666, 839027.777777771, -388252.314814789, 0, -1941261.57407408, -388252.314814811, -3882523.14814814, -1941261.57407404, -3882523.14814813, 465277.777777741, 0, 2791666.66666668, 465277.777777774, 6979166.66666663, 2791666.66666659, 9305555.55555553, 6979166.66666662, -348875.661375629, 0, -2442129.62962964, -348875.661375661, -7326388.88888885, -2442129.62962953, -12210648.1481481, -7326388.88888878, -12210648.1481481, 148809.523809508, 0, 1190476.1904762, 148809.523809526, 4166666.66666664, 1190476.19047612, 8333333.33333327, 4166666.66666656, 10416666.6666667, 8333333.33333328, -27557.3192239826, 0, -248015.873015874, -27557.3192239869, -992063.492063486, -248015.873015853, -2314814.81481479, -992063.492063453, -3472222.22222222, -2314814.81481479, -3472222.22222221}, 
                                                   {0, -224.999999999993, 0, 5465.17857142838, 0, 3215.17857142857, -51751.3392857121, 0, -71350.892857143, -19599.5535714286, 262665.674603162, 0, 592001.488095239, 66670.1388888888, 396005.952380951, -805303.819444398, 0, -2554513.88888889, -138602.430555555, -2831718.75, -1221111.11111111, 1566579.86111101, 0, 6446875.00000001, 180555.555555555, 10121701.3888889, 2288802.08333333, 7349652.77777778, -1949652.77777763, 0, -9892361.11111113, -144097.222222222, -20217013.8888888, -2670138.88888885, -20937500, -11189236.1111112, 1505456.34920622, 0, 9097222.22222225, 64484.1269841292, 22968749.9999999, 1892361.11111104, 31076388.8888888, 10000000.0000001, 23871527.7777779, -657242.063492, 0, -4613095.23809525, -12400.7936507969, -13888888.8888888, -744047.619047557, -23263888.8888887, -4861111.1111111, -23437500.0000002, -14236111.1111113, 124007.936507923, 0, 992063.492063495, 0, 3472222.2222222, 124007.936507918, 6944444.44444436, 992063.492063473, 8680555.55555554, 3472222.22222228, 6944444.44444453}, 
                                                   {0, 399.999999999984, 0, -10382.5396825392, 0, -4382.53968253969, 105973.280423275, 0, 106208.465608466, 20235.1851851853, -573728.835978805, 0, -979404.761904765, -51074.0740740749, -456749.999999999, 1854027.77777767, 0, 4626481.48148149, 76157.4074074105, 3767037.03703705, 1070740.74074076, -3763194.4444442, 0, -12566203.7037037, -67129.6296296371, -15186574.0740741, -1478009.25925933, -7794444.44444453, 4847222.22222187, 0, 20460648.1481482, 32407.4074074148, 33402777.7777776, 1201388.88888897, 25949074.0740743, 9328703.70370413, -3849206.34920604, 0, -19745370.3703704, -6613.75661376209, -40995370.3703702, -532407.407407456, -43518518.5185184, -6412037.0370375, -24305555.5555561, 1719576.71957657, 0, 10416666.6666667, 0, 26388888.8888887, 99206.349206341, 35879629.6296293, 2314814.81481506, 27777777.7777784, 11805555.5555563, -330687.830687799, 0, -2314814.81481482, 0, -6944444.44444442, 0, -11574074.0740739, -330687.830687869, -11574074.0740741, -2314814.81481508, -6944444.44444476}, 
                                                   {0, -524.999999999974, 0, 14064.5833333326, 0, 4439.58333333335, -149352.083333325, 0, -112311.805555556, -15459.7222222225, 843271.990740692, 0, 1095971.06481482, 28385.4166666682, 368584.490740741, -2835295.13888872, 0, -5509045.1388889, -28993.0555555612, -3281128.47222225, -636371.527777815, 5962326.3888885, 0, 15873148.1481482, 15625.0000000124, 14437731.4814815, 609664.351851978, 5120949.07407423, -7920138.88888833, 0, -27243055.5555556, -3472.22222223275, -34505208.3333332, -307291.666666854, -19262152.7777781, -4383680.55555625, 6458333.33333284, 0, 27523148.1481482, 0, 45572916.6666665, 63657.4074075595, 36226851.8518518, 1944444.44444525, 13599537.0370379, -2951388.88888865, 0, -15104166.6666667, 0, -31249999.9999998, -4.14800307650047e-08, -32986111.1111107, -347222.22222269, -18229166.6666676, -4687500.00000128, 578703.703703654, 0, 3472222.22222222, 0, 8680555.55555553, 0, 11574074.0740739, 9.02695400780826e-08, 8680555.5555556, 578703.703704142, 3472222.22222271}, 
                                                   {0, 503.999999999972, 0, -13753.9999999992, 0, -3254.00000000002, 149624.999999991, 0, 84458.3333333338, 8333.33333333357, -868736.111111058, 0, -853513.888888893, -10583.3333333348, -205361.111111112, 3008263.8888887, 0, 4474513.8888889, 6666.66666667189, 1916736.11111114, 247152.777777814, -6512013.88888846, 0, -13494375, -1666.66666667769, -8952430.5555556, -147222.222222347, -2115625.00000015, 8888888.88888828, 0, 24236111.1111111, 0, 22881944.4444444, 34722.2222224139, 8645833.33333366, 1145833.333334, -7430555.55555502, 0, -25555555.5555556, 0, -32326388.8888888, -1.63662924800733e-07, -17951388.8888888, -243055.555556335, -3993055.55555639, 3472222.22222196, 0, 14583333.3333333, 0, 23611111.111111, 5.27563337819806e-08, 18055555.5555552, 4.67027895991275e-07, 6250000.00000085, 694444.444445667, -694444.44444439, 0, -3472222.22222222, 0, -6944444.44444443, 0, -6944444.4444443, -9.44446831300962e-08, -3472222.22222226, -4.21124533318454e-07, -694444.444444897}, 
                                                   {0, -349.999999999978, 0, 9668.05555555496, 0, 1676.3888888889, -106895.601851845, 0, -44263.6574074077, -2993.05555555569, 632942.12962959, 0, 458089.120370372, 2361.11111111192, 75424.7685185189, -2240746.52777763, 0, -2474351.85185186, -694.444444447443, -727100.694444464, -56689.8148148367, 4965798.61111079, 0, 7727199.07407409, 0, 3543344.90740743, 15856.4814815574, 506134.259259349, -6940972.22222176, 0, -14417824.0740741, 0, -9539930.5555555, -1.19044982939938e-07, -2193287.03703723, -130208.333333735, 5937499.9999996, 0, 15810185.1851852, 0, 14299768.5185185, 1.03508522268254e-07, 4918981.48148144, 4.73256714866396e-07, 491898.148148626, -2835648.14814795, 0, -9375000.00000001, 0, -11111111.111111, -3.46093550559986e-08, -5439814.81481459, -2.88927113323431e-07, -868055.555556036, -7.37975082055511e-07, 578703.703703662, 0, 2314814.81481481, 0, 3472222.22222221, 0, 2314814.81481473, 5.94727472081209e-08, 578703.703703712, 2.56500982865728e-07, 2.62487617710933e-07}, 
                                                   {0, 171.42857142856, 0, -4776.19047619016, 0, -576.190476190479, 53426.190476187, 0, 15402.380952381, 642.857142857193, -320951.058201037, 0, -162177.248677249, -238.095238095524, -16464.2857142858, 1155694.44444437, 0, 895694.444444446, 0, 162500.000000007, 5833.33333334141, -2610416.6666665, 0, -2874074.07407408, 0, -817361.111111119, -2.87051223521632e-08, -53703.7037037355, 3724206.34920611, 0, 5534722.22222223, 0, 2291666.66666664, 4.58141992807439e-08, 243055.555555621, 1.50870753814333e-07, -3253968.25396804, 0, -6286375.66137566, 0, -3611111.11111109, -4.04098051613309e-08, -578703.703703672, -1.78941931279499e-07, -1.65687684466184e-07, 1587301.58730148, 0, 3869047.61904762, 0, 2976190.47619045, 1.36894224664739e-08, 694444.444444349, 1.11020905475885e-07, 1.65340307698433e-07, 2.78945575714599e-07, -330687.830687809, 0, -992063.492063492, 0, -992063.492063489, 0, -330687.830687798, -2.30408716630545e-08, 0, -9.79388272227778e-08, -9.44535525019484e-08}, 
                                                   {0, -56.2499999999961, 0, 1577.23214285703, 0, 118.750000000001, -17797.7678571416, 0, -3204.01785714288, -62.5000000000101, 108108.382936501, 0, 34182.0436507938, 0, 1620.5357142857, -394522.569444418, 0, -192057.291666667, 0, -16284.7222222237, 0, 905121.527777719, 0, 629687.500000001, 0, 83940.9722222234, 0, 0, -1314236.11111103, 0, -1244791.66666667, 0, -243055.555555549, -1.00751830063481e-08, -1.19206252296888e-08, -3.24189289814957e-08, 1170634.92063485, 0, 1458333.33333333, 0, 399305.55555555, 0, -1.16050002188013e-08, 3.85808116512662e-08, 3.16892332233494e-08, -582837.301587266, 0, -930059.523809524, 0, -347222.222222216, 0, 2.45766754268109e-08, -2.42955234356391e-08, -3.1586951596996e-08, -6.04652423217441e-08, 124007.936507929, 0, 248015.873015873, 0, 124007.936507936, 0, 0, 0, 0, 2.14415241628369e-08, 1.92763828352856e-08}, 
                                                   {0, 11.1111111111103, 0, -313.095238095216, 0, -11.1111111111112, 3557.16490299799, 0, 301.984126984129, 0, -21796.8474426793, 0, -3255.18077601413, 0, 0, 80405.0925925872, 0, 18541.6666666667, 0, 0, 0, -186863.425925914, 0, -61863.425925926, 0, 0, 0, 0, 275462.962962946, 0, 125000, 0, 0, 0, 0, 0, -249669.312169297, 0, -150462.962962963, 0, 0, 0, 0, 0, 0, 126763.668430328, 0, 99206.3492063493, 0, 0, 0, 0, 0, 0, 0, -27557.3192239843, 0, -27557.3192239859, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                   {0, 0, 0, 0, 0, 11.1111111111112, 0, 0, -301.984126984129, 0, 0, 0, 3255.18077601412, 0, 0, 0, 0, -18541.6666666667, 0, 0, 0, 0, 0, 61863.425925926, 0, 0, 0, 0, 0, 0, -125000, 0, 0, 0, 0, 0, 0, 0, 150462.962962963, 0, 0, 0, 0, 0, 0, 0, 0, -99206.3492063492, 0, 0, 0, 0, 0, 0, 0, 0, 0, 27557.3192239859, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                   {0, 0, 0, 0, 0, 6.25000000000008, 0, 0, -162.053571428573, -62.5000000000013, 0, 0, 1628.47222222224, 0, 1620.53571428574, 0, 0, -8394.09722222227, 0, -16284.7222222224, 0, 0, 0, 24305.5555555556, 0, 83940.9722222228, 0, 0, 0, 0, -39930.5555555556, 0, -243055.555555556, 0, 0, 0, 0, 0, 34722.2222222222, 0, 399305.555555556, 0, 0, 0, 0, 0, 0, -12400.7936507936, 0, -347222.222222222, 0, 0, 0, 0, 0, 0, 0, 0, 0, 124007.936507937, 0, 0, 0, 0, 0, 0}, 
                                                   {0, 0, 0, 0, 0, 4.76190476190473, 0, 0, -116.666666666666, -71.4285714285707, 0, 0, 1074.07407407407, 238.095238095235, 1749.99999999998, 0, 0, -4861.1111111111, 0, -16111.111111111, -5833.33333333319, 0, 0, 11574.074074074, 0, 72916.6666666662, 0, 53703.703703703, 0, 0, -13888.8888888889, 0, -173611.11111111, 0, -243055.555555554, 0, 0, 0, 6613.75661375664, 0, 208333.333333332, 0, 578703.703703702, 0, 0, 0, 0, 0, 0, -99206.3492063485, 0, -694444.444444443, 0, 0, 0, 0, 0, 0, 0, 0, 0, 330687.83068783, 0, 0, 0, 0}, 
                                                   {0, 0, 0, 0, 0, 4.16666666666657, 0, 0, -95.1388888888867, -76.3888888888871, 0, 0, 781.249999999985, 416.666666666659, 1744.21296296292, 0, 0, -2951.38888888884, -694.444444444433, -14322.9166666664, -9513.88888888866, 0, 0, 5208.33333333326, 0, 54108.7962962953, 15856.4814814808, 78124.9999999986, 0, 0, -3472.22222222218, 0, -95486.1111111091, 0, -295138.888888885, -130208.33333333, 0, 0, 0, 0, 63657.4074074051, 0, 520833.333333328, 0, 491898.148148142, 0, 0, 0, 0, 0, 0, -347222.222222219, 0, -868055.55555555, 0, 0, 0, 0, 0, 0, 0, 0, 0, 578703.703703702, 0, 0}, 
                                                   {0, 0, 0, 0, 0, 3.99999999999988, 0, 0, -83.3333333333309, -83.3333333333308, 0, 0, 583.333333333319, 583.333333333316, 1736.11111111106, 0, 0, -1666.66666666663, -1666.66666666661, -12152.7777777775, -12152.7777777775, 0, 0, 1666.66666666661, 1666.66666666655, 34722.2222222213, 34722.2222222212, 85069.444444443, 0, 0, 0, 0, -34722.2222222206, -34722.2222222202, -243055.555555552, -243055.555555552, 0, 0, 0, 0, 0, 0, 243055.555555551, 243055.55555555, 694444.444444437, 0, 0, 0, 0, 0, 0, 0, 0, -694444.444444439, -694444.444444438, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 694444.444444442}, 
                                                   {0, 0, 0, 0, 0, 4.1666666666665, 0, 0, -76.3888888888864, -95.1388888888849, 0, 0, 416.666666666655, 781.249999999969, 1744.2129629629, 0, 0, -694.444444444425, -2951.38888888876, -9513.88888888863, -14322.9166666662, 0, 0, 0, 5208.33333333305, 15856.4814814809, 54108.7962962945, 78124.9999999983, 0, 0, 0, -3472.22222222186, 0, -95486.1111111074, -130208.33333333, -295138.888888884, 0, 0, 0, 0, 0, 63657.4074074031, 0, 520833.333333325, 491898.14814814, 0, 0, 0, 0, 0, 0, 0, -347222.222222216, 0, -868055.555555547, 0, 0, 0, 0, 0, 0, 0, 0, 0, 578703.7037037, 0}, 
                                                   {0, 0, 0, 0, 0, 4.76190476190461, 0, 0, -71.4285714285691, -116.666666666663, 0, 0, 238.09523809523, 1074.07407407404, 1749.99999999995, 0, 0, 0, -4861.11111111098, -5833.33333333314, -16111.1111111107, 0, 0, 0, 11574.0740740738, 0, 72916.6666666649, 53703.7037037023, 0, 0, 0, -13888.8888888885, 0, -173611.111111107, 0, -243055.555555551, 0, 0, 0, 6613.7566137564, 0, 208333.333333329, 0, 578703.703703696, 0, 0, 0, 0, 0, 0, -99206.3492063462, 0, -694444.444444438, 0, 0, 0, 0, 0, 0, 0, 0, 0, 330687.830687829, 0, 0, 0}, 
                                                   {0, 0, 0, 0, 0, 6.24999999999993, 0, 0, -62.4999999999991, -162.05357142857, 0, 0, 0, 1628.4722222222, 1620.53571428569, 0, 0, 0, -8394.09722222214, 0, -16284.7222222219, 0, 0, 0, 24305.5555555554, 0, 83940.9722222209, 0, 0, 0, 0, -39930.5555555554, 0, -243055.555555552, 0, 0, 0, 0, 0, 34722.2222222222, 0, 399305.555555551, 0, 0, 0, 0, 0, 0, -12400.7936507937, 0, -347222.222222219, 0, 0, 0, 0, 0, 0, 0, 0, 0, 124007.936507936, 0, 0, 0, 0, 0}, 
                                                   {0, 0, 0, 0, 0, 11.111111111111, 0, 0, 0, -301.984126984125, 0, 0, 0, 3255.18077601409, 0, 0, 0, 0, -18541.6666666665, 0, 0, 0, 0, 0, 61863.4259259256, 0, 0, 0, 0, 0, 0, -124999.999999999, 0, 0, 0, 0, 0, 0, 0, 150462.962962963, 0, 0, 0, 0, 0, 0, 0, 0, -99206.3492063491, 0, 0, 0, 0, 0, 0, 0, 0, 0, 27557.3192239859, 0, 0, 0, 0, 0, 0, 0}, 
                                                   {0, 0, 11.1111111111134, 0, -313.0952380953, -11.1111111111429, 0, 3557.16490299889, 0, 301.984126984915, 0, -21796.8474426845, 0, -3255.18077602117, 0, 0, 80405.0925926049, 0, 18541.6666666996, 2.3319427993631e-08, 4.17830733562433e-08, 0, -186863.425925951, 0, -61863.4259260139, -6.41949805685292e-08, -1.59804490996916e-07, -1.25783493816858e-07, 0, 275462.962962996, 0, 125000.00000014, 1.207843967248e-07, 3.42138427684956e-07, 2.35377196316978e-07, 3.64347838781069e-07, 0, -249669.312169339, 0, -150462.962963093, -1.36248955006321e-07, -4.05087912793517e-07, -3.08500152288624e-07, -5.90958726655297e-07, -4.28031423862723e-07, 0, 126763.668430347, 0, 99206.3492064151, 7.95805246334325e-08, 2.46517163408686e-07, 2.39786332728412e-07, 4.83750436652019e-07, 3.04082355147171e-07, 4.28945337087654e-07, 0, -27557.3192239882, 0, -27557.3192239999, -1.81061865335341e-08, -6.01051224103467e-08, -7.60857483810002e-08, -1.53348535390838e-07, -1.07709787512988e-07, -1.76347310165884e-07, -1.24523151391141e-07}, 
                                                   {0, 0, -56.2500000000111, 0, 1577.23214285744, 118.750000000153, 0, -17797.767857146, -62.5000000011618, -3204.01785714665, 0, 108108.382936526, 0, 34182.0436508275, 1620.53571431343, 0, -394522.569444504, -1.05975430655881e-08, -192057.291666824, -1.14134623440975e-07, -16284.7222224229, 0, 905121.527777902, 1.14543750564578e-08, 629687.500000421, 3.1566001872797e-07, 83940.9722229846, 6.13854471613616e-07, 0, -1314236.11111127, 0, -1244791.66666733, -5.91846884056659e-07, -243055.555557178, -1.16458763296146e-06, -1.7618523211942e-06, 0, 1170634.92063505, -3.1570340050258e-08, 1458333.33333396, 6.64143436891326e-07, 399305.555557466, 1.52883733503896e-06, 2.82969509467472e-06, 2.11225193649353e-06, 0, -582837.30158736, 3.04622422895775e-08, -930059.52380984, -3.86092799181829e-07, -347222.222223379, -1.18108943201469e-06, -2.29699195247124e-06, -1.5200118301725e-06, -2.09440600763472e-06, 0, 124007.936507948, 0, 248015.87301594, 8.76568295234069e-08, 124007.936508218, 3.72599880891861e-07, 7.22657244798611e-07, 5.37521976148971e-07, 8.5082739890835e-07, 6.2056293819616e-07}, 
                                                   {0, 0, 171.428571428604, 0, -4776.19047619134, -576.190476190914, 0, 53426.1904761997, 642.857142860482, 15402.3809523918, 0, -320951.05820111, -238.095238108633, -162177.248677345, -16464.2857143653, 0, 1155694.44444462, 3.09737066359786e-08, 895694.444444893, 5833.33333366587, 162500.000000572, 0, -2610416.66666703, -3.32539656784335e-08, -2874074.07407527, -9.25030702053371e-07, -817361.111113268, -53703.7037054855, 0, 3724206.34920682, -1.9034838861695e-08, 5534722.22222412, 1.72908626449737e-06, 2291666.66667122, 3.43300409517613e-06, 243055.555560616, 0, -3253968.25396863, 9.23647744590168e-08, -6286375.66137743, -1.92974481418805e-06, -3611111.11111644, -4.51860749427184e-06, -578703.70371174, -6.20206257089234e-06, 0, 1587301.58730176, -8.85870078926422e-08, 3869047.61904852, 1.11629140699692e-06, 2976190.47619369, 3.47109112197545e-06, 694444.444450905, 4.52798276240706e-06, 6.07320358657733e-06, 0, -330687.830687863, 2.82827934183956e-08, -992063.492063685, -2.52742336863049e-07, -992063.492064268, -1.08832119770058e-06, -330687.830689846, -1.60117945585234e-06, -2.43494565912929e-06, -1.83953970372673e-06}, 
                                                   {0, 0, -350.000000000061, 0, 9668.05555555721, 1676.38888888971, 0, -106895.60185187, -2993.05555556187, -44263.6574074278, 0, 632942.129629729, 2361.11111113678, 458089.120370551, 75424.7685186688, 0, -2240746.52777811, -694.444444504176, -2474351.85185269, -56689.8148154532, -727100.694445516, 0, 4965798.6111118, 6.37932880318169e-08, 7727199.07407631, 15856.4814832706, 3543344.90741141, 506134.259262663, 0, -6940972.22222313, 3.77554572817205e-08, -14417824.0740776, -3.33624133776513e-06, -9539930.55556393, -130208.333340008, -2193287.03704658, 0, 5937500.00000072, -1.78202344837329e-07, 15810185.1851885, 3.70283285703969e-06, 14299768.5185282, 8.8225310316521e-06, 4918981.48149645, 491898.148160135, 0, -2835648.14814847, 1.6989266107014e-07, -9375000.00000168, -2.13059740562712e-06, -11111111.1111169, -6.74444647024365e-06, -5439814.81482671, -8.89804657490608e-06, -868055.55556712, 0, 578703.703703766, -5.40292876959863e-08, 2314814.81481518, 4.8069099039729e-07, 3472222.22222362, 2.10124701391161e-06, 2314814.81481849, 3.1537305268985e-06, 578703.703708273, 3.58674144106863e-06}, 
                                                   {0, 0, 504.000000000081, 0, -13754.0000000022, -3254.00000000106, 0, 149625.000000023, 8333.33333334151, 84458.3333333596, 0, -868736.111111243, -10583.3333333672, -853513.888889121, -205361.111111306, 0, 3008263.88888933, 6666.66666674606, 4474513.88888996, 247152.777778621, 1916736.11111248, 0, -6512013.8888898, -1666.66666675124, -13494375.0000029, -147222.222224607, -8952430.55556062, -2115625.00000446, 0, 8888888.88889009, -5.11852762035315e-08, 24236111.1111156, 34722.2222266637, 22881944.4444549, 1145833.33334227, 8645833.33334567, 0, -7430555.55555652, 2.36653876528352e-07, -25555555.5555598, -4.90270931781808e-06, -32326388.8889009, -243055.555567444, -17951388.8889079, -3993055.55557146, 0, 3472222.22222265, -2.24347045657832e-07, 14583333.3333355, 2.80465927781076e-06, 23611111.1111183, 9.05633429891077e-06, 18055555.5555705, 694444.444456484, 6250000.00001507, 0, -694444.444444527, 7.10730762332398e-08, -3472222.22222269, -6.29890191404007e-07, -6944444.44444615, -2.80371081032013e-06, -6944444.444449, -4.29117965838036e-06, -3472222.22222808, -694444.444449238}, 
                                                   {0, 0, -525.000000000075, 0, 14064.5833333353, 4439.58333333429, 0, -149352.083333355, -15459.7222222296, -112311.805555579, 0, 843271.990740862, 28385.4166666977, 1095971.06481502, 368584.490740915, 0, -2835295.13888929, -28993.0555556291, -5509045.13888984, -636371.527778552, -3281128.47222344, 0, 5962326.38888973, 15625.0000000785, 15873148.1481507, 609664.351854067, 14437731.4814859, 5120949.07407813, 0, -7920138.88889, -3472.22222217456, -27243055.5555596, -307291.666670793, -34505208.3333423, -4383680.55556388, -19262152.7777888, 0, 6458333.33333422, -2.18432051469429e-07, 27523148.1481519, 63657.407411938, 45572916.6666769, 1944444.44445563, 36226851.8518685, 13599537.0370517, 0, -2951388.88888929, 2.06093471192055e-07, -15104166.6666686, -2.57492005684526e-06, -31250000.000006, -347222.222230729, -32986111.111124, -4687500.00001133, -18229166.6666802, 0, 578703.703703781, -6.50438107009914e-08, 3472222.22222264, 5.74812578260162e-07, 8680555.55555696, 2.61793626900186e-06, 11574074.0740779, 578703.703707784, 8680555.55556071, 3472222.22222665}, 
                                                   {0, 0, 400.000000000047, 0, -10382.539682541, -4382.53968254027, 0, 105973.280423294, 20235.1851851897, 106208.46560848, 0, -573728.835978913, -51074.0740740935, -979404.761904888, -456750.000000106, 0, 1854027.77777804, 76157.4074074541, 4626481.48148206, 1070740.74074123, 3767037.03703776, 0, -3763194.44444498, -67129.6296296799, -12566203.7037052, -1478009.25926067, -15186574.0740767, -7794444.44444696, 0, 4847222.22222293, 32407.4074073778, 20460648.1481506, 1201388.88889152, 33402777.777783, 9328703.70370901, 25949074.0740807, 0, -3849206.34920692, -6613.75661361926, -19745370.3703727, -532407.407410281, -40995370.3703762, -6412037.03704426, -43518518.5185283, -24305555.5555647, 0, 1719576.71957698, -1.29228165573691e-07, 10416666.6666679, 99206.3492079695, 26388888.8888922, 2314814.81482032, 35879629.6296371, 11805555.5555628, 27777777.777786, 0, -330687.83068788, 4.06320020664298e-08, -2314814.81481507, -3.58614453786817e-07, -6944444.44444521, -330687.830689516, -11574074.0740763, -2314814.81481749, -11574074.0740771, -6944444.44444721}, 
                                                   {0, 0, -225.00000000002, 0, 5465.17857142911, 3215.1785714288, 0, -51751.33928572, -19599.5535714303, -71350.8928571486, 0, 262665.674603207, 66670.1388888968, 592001.488095288, 396005.952380994, 0, -805303.819444553, -138602.430555575, -2554513.88888912, -1221111.11111131, -2831718.75000028, 0, 1566579.86111134, 180555.555555577, 6446875.0000006, 2288802.08333392, 10121701.3888899, 7349652.77777878, 0, -1949652.77777808, -144097.222222211, -9892361.11111207, -2670138.88888998, -20217013.8888908, -11189236.1111133, -20937500.0000026, 0, 1505456.34920659, 64484.1269840718, 9097222.22222313, 1892361.1111123, 22968750.000002, 10000000.0000031, 31076388.8888926, 23871527.7777815, 0, -657242.063492171, -12400.7936507416, -4613095.23809571, -744047.61904828, -13888888.88889, -4861111.11111345, -23263888.8888916, -14236111.1111142, -23437500.0000032, 0, 124007.936507957, -1.63246504225024e-08, 992063.492063597, 124007.936508081, 3472222.22222247, 992063.492064205, 6944444.44444522, 3472222.22222337, 8680555.55555669, 6944444.44444555}, 
                                                   {0, 0, 100.000000000005, 0, -1928.96825396839, -1928.96825396831, 0, 15855.7539682554, 15855.7539682544, 31711.5079365093, 0, -73185.7363315777, -73185.7363315716, -219557.20899472, -219557.208994718, 0, 209756.944444472, 209756.94444445, 839027.77777783, 839027.777777825, 1258541.66666673, 0, -388252.314814871, -388252.314814821, -1941261.57407421, -1941261.57407422, -3882523.14814835, -3882523.14814838, 0, 465277.777777852, 465277.777777777, 2791666.66666689, 2791666.66666693, 6979166.66666704, 6979166.66666721, 9305555.55555613, 0, -348875.661375721, -348875.66137565, -2442129.62962984, -2442129.62962991, -7326388.88888928, -7326388.88888965, -12210648.1481489, -12210648.148149, 0, 148809.523809551, 148809.523809513, 1190476.1904763, 1190476.19047634, 4166666.66666687, 4166666.66666725, 8333333.33333388, 8333333.3333341, 10416666.6666674, 0, -27557.3192239911, -27557.3192239823, -248015.873015898, -248015.873015906, -992063.492063532, -992063.492063668, -2314814.81481496, -2314814.81481511, -3472222.22222246, -3472222.22222248}, 
                                                   {0, 0, 0, 0, 0, 4499.99999999998, 0, 0, -64303.5714285712, -64303.5714285709, 0, 0, 391991.071428571, 391991.071428568, 783982.142857135, 0, 0, -1333402.77777778, -1333402.77777776, -4000208.3333333, -4000208.33333327, 0, 0, 2772048.61111112, 2772048.61111108, 11088194.4444444, 11088194.4444442, 16632291.6666665, 0, 0, -3611111.11111113, -3611111.11111107, -18055555.5555554, -18055555.555555, -36111111.1111108, -36111111.1111105, 0, 0, 2881944.44444446, 2881944.44444443, 17291666.6666665, 17291666.6666659, 43229166.6666664, 43229166.6666656, 57638888.8888881, 0, 0, -1289682.53968255, -1289682.53968255, -9027777.77777767, -9027777.7777773, -27083333.3333332, -27083333.3333324, -45138888.8888885, -45138888.888888, 0, 0, 248015.873015877, 248015.873015878, 1984126.98412696, 1984126.98412686, 6944444.44444439, 6944444.44444415, 13888888.8888888, 13888888.8888885, 17361111.1111108}, 
                                                   {0, 0, 0, 0, 0, -125.000000000001, 0, 0, 3366.07142857145, 125.000000000012, 0, 0, -35810.515873016, 0, -3241.07142857144, 0, 0, 200451.388888889, 0, 32569.4444444461, 0, 0, 0, -653993.055555556, 0, -167881.944444446, 0, 0, 0, 0, 1284722.22222222, 0, 486111.111111105, 1.07899810150016e-08, 1.50917515508309e-08, 3.5431787424142e-08, 0, 0, -1493055.55555556, 0, -798611.111111105, 0, 0, -4.19338055442575e-08, -3.83217266735371e-08, 0, 0, 942460.317460318, 0, 694444.444444438, 0, -2.2866143783855e-08, 2.61682411154023e-08, 3.79226858502809e-08, 6.55448715569938e-08, 0, 0, -248015.873015873, 0, -248015.873015872, 0, 0, 0, 0, -2.29165416732974e-08, -2.1901680063978e-08}, 
                                                   {0, 0, 0, 0, 0, -124.999999999999, 0, 0, 124.999999999995, 3366.07142857141, 0, 0, 0, -35810.5158730156, -3241.07142857125, 0, 0, 0, 200451.388888888, 0, 32569.444444442, 0, 0, 0, -653993.055555552, 0, -167881.94444443, 0, 0, 0, 0, 1284722.22222222, 0, 486111.11111107, 2.05336392593048e-08, 0, 0, 0, 0, -1493055.55555555, 0, -798611.111111051, -2.99172149423586e-08, 4.11198153345807e-08, -3.46893090597269e-08, 0, 0, 0, 942460.317460316, 0, 694444.444444401, 1.76075887850327e-08, -5.60878043800405e-08, 4.28175050163873e-08, 1.60672593758251e-08, 0, 0, 0, -248015.873015873, 0, -248015.873015861, 0, 2.48596707915477e-08, -1.43234123485582e-08, 0, -1.79898566650505e-08}, 
                                                   {0, 0, 0, 0, 0, -5999.99999999998, 0, 0, 125738.095238095, 65738.0952380947, 0, 0, -960908.730158729, -303527.777777774, -1264436.5079365, 0, 0, 3801388.88888889, 766111.111111097, 8368888.88888884, 5333611.11111104, 0, 0, -8803472.22222223, -1142361.11111108, -27552777.7777776, -12230555.5555553, -29837499.9999997, 0, 0, 12430555.5555556, 1006944.44444442, 50729166.6666663, 16458333.3333328, 78611111.1111106, 55763888.8888882, 0, 0, -10555555.5555556, -486111.11111111, -53263888.8888885, -12986111.1111104, -107986111.11111, -57638888.8888877, -110416666.666666, 0, 0, 4960317.46031749, 99206.3492063626, 29861111.1111108, 5555555.55555511, 74999999.9999994, 31249999.9999991, 100694444.444444, 76388888.8888878, 0, 0, -992063.4920635, 0, -6944444.44444438, -992063.492063375, -20833333.3333331, -6944444.44444416, -34722222.222222, -20833333.3333329, -34722222.2222219}, 
                                                   {0, 0, 0, 0, 0, 7000, 0, 0, -164194.444444444, -59194.4444444442, 0, 0, 1444046.2962963, 206129.629629628, 1300175.92592592, 0, 0, -6430138.88888889, -378472.222222218, -10279027.7777777, -4227361.11111108, 0, 0, 16370138.8888889, 386574.07407407, 39190509.2592591, 7223379.62962953, 29657175.9259258, 0, 0, -24930555.5555556, -208333.333333337, -81006944.444444, -6840277.77777752, -93645833.333333, -44201388.8888888, 0, 0, 22500000.0000001, 46296.2962963128, 93217592.5925921, 3402777.77777744, 147916666.666666, 35648148.1481477, 109490740.74074, 0, 0, -11111111.1111111, -1.74542044174751e-08, -56249999.9999996, -694444.444444204, -114583333.333333, -14583333.3333329, -118055555.555556, -62499999.9999999, 0, 0, 2314814.81481482, 0, 13888888.8888888, -6.52712574496448e-08, 34722222.2222219, 2314814.81481467, 46296296.2962962, 13888888.8888888, 34722222.2222222}, 
                                                   {0, 0, 0, 0, 0, -6300.00000000001, 0, 0, 156175, 40675.0000000001, 0, 0, -1479875, -104166.666666668, -954041.666666664, 0, 0, 7159513.8888889, 132291.666666671, 8283819.44444445, 2306597.22222224, 0, 0, -19704513.8888889, -83333.3333333459, -35221875, -2758680.55555562, -18192708.3333334, 0, 0, 32138888.8888889, 20833.3333333505, 80624999.9999997, 1631944.4444445, 66423611.1111113, 19548611.1111116, 0, 0, -30763888.8888889, -1.79269710633807e-08, -101388888.888888, -381944.444444444, -119965277.777777, -10243055.555556, -59201388.8888896, 0, 0, 15972222.2222223, 1.20934123810052e-08, 65972222.2222219, -4.69226503436484e-08, 104166666.666666, 2083333.33333352, 76388888.8888898, 24305555.5555565, 0, 0, -3472222.22222223, 0, -17361111.111111, 1.75637033603711e-08, -34722222.2222219, 0, -34722222.2222223, -3472222.22222251, -17361111.1111115}, 
                                                   {0, 0, 0, 0, 0, 4200.00000000002, 0, 0, -107616.666666667, -20116.6666666669, 0, 0, 1067513.88888889, 35916.6666666683, 490930.555555556, 0, 0, -5460277.77777779, -28333.3333333393, -4515208.33333336, -833263.888888926, 0, 0, 15968402.7777778, 8333.33333334626, 20661805.5555556, 623611.111111232, 7058680.55555572, 0, 0, -27652777.7777778, -1.21539415950805e-08, -51402777.7777776, -173611.111111282, -28402777.7777782, -4826388.88888958, 0, 0, 27986111.1111111, 0, 70208333.3333331, 1.3248226755139e-07, 57673611.1111111, 1215277.77777853, 16666666.6666676, 0, 0, -15277777.7777778, 0, -49305555.5555554, -3.34307954877551e-08, -56249999.9999996, -4.15137092061178e-07, -25694444.4444455, -3472222.2222235, 0, 0, 3472222.22222222, 0, 13888888.8888889, 0, 20833333.3333332, 7.4361726914735e-08, 13888888.888889, 4.09810747781615e-07, 3472222.22222275}, 
                                                   {0, 0, 0, 0, 0, -2000.00000000001, 0, 0, 52388.8888888892, 6722.22222222239, 0, 0, -535990.740740743, -7500.000000001, -168490.740740742, 0, 0, 2851111.11111112, 2777.77777778127, 1611250.00000002, 179583.333333359, 0, 0, -8731250.00000001, 0, -7764351.85185188, -63425.9259260096, -1596527.77777789, 0, 0, 15902777.7777778, 0, 20590277.7777777, 1.25827016639131e-07, 6875000.00000025, 520833.333333788, 0, 0, -16944444.4444445, 0, -30254629.6296296, -1.04719670774342e-07, -15277777.7777778, -5.1801512044941e-07, -1967592.59259321, 0, 0, 9722222.22222223, 0, 22916666.6666666, 3.34313123213276e-08, 16666666.6666665, 3.03401074753229e-07, 3472222.22222285, 8.32310309197369e-07, 0, 0, -2314814.81481481, 0, -6944444.44444443, 0, -6944444.44444436, -6.03261562458172e-08, -2314814.81481488, -2.75893436916536e-07, -3.23282832793612e-07}, 
                                                   {0, 0, 0, 0, 0, 642.857142857147, 0, 0, -17107.142857143, -1357.14285714292, 0, 0, 178964.285714287, 714.285714286061, 34678.5714285717, 0, 0, -979861.111111113, 0, -341111.11111112, -17500.0000000094, 0, 0, 3109027.77777778, 0, 1707638.8888889, 3.21886571401361e-08, 161111.111111151, 0, 0, -5902777.77777778, 0, -4756944.44444442, -5.00712444402117e-08, -729166.666666755, -1.71672838080726e-07, 0, 0, 6587301.58730159, 0, 7430555.55555553, 4.29182095201815e-08, 1736111.11111111, 2.00284977760847e-07, 2.14591047600907e-07, 0, 0, -3968253.96825397, 0, -6051587.30158728, -1.43060698400605e-08, -2083333.33333325, -1.21601593640514e-07, -2.14591047600907e-07, -3.14733536481331e-07, 0, 0, 992063.492063491, 0, 1984126.98412698, 0, 992063.49206346, 2.50356222201059e-08, 1.25178111100529e-08, 1.07295523800454e-07, 1.14448558720484e-07}, 
                                                   {0, 0, 0, 0, 0, -71.428571428572, 0, 0, 1821.42857142858, 785.714285714295, 0, 0, -17861.1111111112, -714.285714285769, -19964.2857142859, 0, 0, 89027.7777777781, 0, 194722.222222223, 17500.0000000012, 0, 0, -246527.777777778, 0, -963194.444444448, 0, -161111.111111119, 0, 0, 381944.444444445, 0, 2638888.88888889, 0, 729166.666666688, 1.80576735932243e-08, 0, 0, -307539.682539683, 0, -4027777.77777778, 0, -1736111.11111114, -1.73346118519481e-08, -4.46917343169598e-08, 0, 0, 99206.3492063491, 0, 3174603.17460317, 0, 2083333.33333335, 0, 4.66186123736348e-08, 3.15899928009525e-08, 0, 0, 0, 0, -992063.492063492, 0, -992063.492063493, 0, -1.586575867053e-08, 0, -1.86565797152245e-08}, 
                                                   {0, 0, 0, 0, 0, -55.5555555555549, 0, 0, 1324.07407407406, 888.888888888876, 0, 0, -11685.1851851851, -3611.11111111106, -21129.6296296293, 0, 0, 49768.5185185182, 2777.77777777773, 185694.444444442, 85231.4814814795, 0, 0, -108796.296296296, 0, -785879.629629622, -63425.9259259198, -740509.259259248, 0, 0, 115740.740740741, 0, 1701388.88888887, -1.07229937006973e-08, 3078703.70370368, 520833.333333305, 0, 0, -46296.2962962967, 0, -1782407.40740739, 0, -6481481.48148145, 4.01729906631512e-08, -1967592.59259255, 0, 0, 0, 0, 694444.444444434, 0, 6481481.48148145, -2.88892064836536e-08, 3472222.22222219, -4.54070351319543e-08, 0, 0, 0, 0, 0, 0, -2314814.81481481, 0, -2314814.8148148, 1.77753588128833e-08, 1.59400445186277e-08}, 
                                                   {0, 0, 0, 0, 0, -49.9999999999988, 0, 0, 1091.66666666664, 966.666666666642, 0, 0, -8333.33333333317, -5916.66666666653, -21055.555555555, 0, 0, 28124.9999999995, 13333.333333333, 160069.444444441, 128263.888888886, 0, 0, -41666.6666666659, -8333.3333333329, -536458.333333322, -286111.111111101, -967013.888888873, 0, 0, 20833.3333333331, 0, 784722.222222201, 173611.111111094, 3194444.4444444, 2118055.55555552, 0, 0, 0, 0, -381944.444444421, 1.76575909664238e-08, -4548611.11111106, -1215277.77777773, -6770833.33333326, 0, 0, 0, 0, -1.32349094787248e-08, -1.01995965128019e-08, 2083333.3333333, -3.21836550508292e-08, 9027777.77777772, 3472222.22222216, 0, 0, 0, 0, 0, 0, 1.02329012563733e-08, 0, -3472222.2222222, 2.10401736441734e-08, -3472222.22222219}, 
                                                   {0, 0, 0, 0, 0, -49.9999999999983, 0, 0, 966.666666666638, 1091.66666666663, 0, 0, -5916.66666666652, -8333.33333333304, -21055.5555555549, 0, 0, 13333.333333333, 28124.9999999989, 128263.888888886, 160069.44444444, 0, 0, -8333.33333333303, -41666.6666666642, -286111.111111102, -536458.333333317, -967013.88888887, 0, 0, 0, 20833.3333333305, 173611.111111098, 784722.222222188, 2118055.55555552, 3194444.44444439, 0, 0, 0, 0, 1.24078157567606e-08, -381944.444444405, -1215277.77777774, -4548611.11111103, -6770833.33333324, 0, 0, 0, 0, 0, -2.47563461715043e-08, -2.04830856668141e-08, 2083333.33333328, 3472222.22222216, 9027777.77777769, 0, 0, 0, 0, 0, 0, 0, 1.67912257321298e-08, 1.53845462533669e-08, -3472222.22222219, -3472222.22222219}, 
                                                   {0, 0, 0, 0, 0, -55.5555555555537, 0, 0, 888.888888888861, 1324.07407407403, 0, 0, -3611.111111111, -11685.1851851848, -21129.629629629, 0, 0, 2777.77777777755, 49768.518518517, 85231.481481479, 185694.444444439, 0, 0, 0, -108796.296296293, -63425.9259259207, -785879.629629609, -740509.259259242, 0, 0, 0, 115740.740740737, 0, 1701388.88888884, 520833.333333304, 3078703.70370365, 0, 0, 0, -46296.2962962938, 0, -1782407.40740735, 2.35557288742986e-08, -6481481.48148139, -1967592.59259251, 0, 0, 0, 0, 0, 694444.444444411, 0, 6481481.48148141, -4.63472127598852e-08, 3472222.22222213, 0, 0, 0, 0, 0, 0, 0, -2314814.81481479, 0, -2314814.81481478, 3.14982156829121e-08}, 
                                                   {0, 0, 0, 0, 0, -71.4285714285704, 0, 0, 785.714285714271, 1821.42857142854, 0, 0, -714.285714285693, -17861.1111111109, -19964.2857142853, 0, 0, 0, 89027.7777777767, 17499.9999999992, 194722.222222218, 0, 0, 0, -246527.777777775, 0, -963194.444444427, -161111.111111103, 0, 0, 0, 381944.444444442, 0, 2638888.88888885, 0, 729166.666666631, 0, 0, 0, -307539.682539681, 0, -4027777.77777773, 0, -1736111.11111104, 2.95115937747992e-08, 0, 0, 0, 99206.3492063495, 0, 3174603.17460314, 0, 2083333.33333327, 0, -4.69824572894803e-08, 0, 0, 0, 0, 0, -992063.492063483, 0, -992063.492063471, 0, 2.2487834456397e-08, 1.06831969464773e-08}, 
                                                   {0, 0, 0, 0, 0, 642.857142857138, 0, 0, -1357.14285714282, -17107.1428571427, 0, 0, 714.285714285908, 178964.285714284, 34678.5714285702, 0, 0, 0, -979861.111111104, -17500.0000000027, -341111.111111096, 0, 0, 0, 3109027.77777776, 1.91935771104078e-08, 1707638.88888881, 161111.111111111, 0, 0, 0, -5902777.77777775, -3.41356447905733e-08, -4756944.44444422, -8.29021072701888e-08, -729166.666666593, 0, 0, 0, 6587301.58730157, 2.84827348604317e-08, 7430555.55555524, 1.34780552634473e-07, 1736111.11111086, 1.18916109565331e-07, 0, 0, 0, -3968253.96825396, -1.13726669052722e-08, -6051587.30158708, -8.50327737376616e-08, -2083333.33333303, -1.79958333738032e-07, -1.97214612315652e-08, 0, 0, 0, 992063.492063493, 0, 1984126.98412692, 1.88465245157898e-08, 992063.492063362, 6.6297282162362e-08, -3.72962281721944e-08, 6.63283037561617e-08}, 
                                                   {0, 0, 0, 0, 0, -1999.99999999998, 0, 0, 6722.22222222207, 52388.8888888885, 0, 0, -7500.00000000034, -535990.740740736, -168490.740740736, 0, 0, 2777.77777778054, 2851111.11111109, 179583.333333336, 1611249.99999995, 0, 0, 0, -8731249.99999995, -63425.9259259706, -7764351.85185158, -1596527.77777774, 0, 0, 0, 15902777.7777777, 9.00518574394465e-08, 20590277.7777771, 520833.333333501, 6874999.99999966, 0, 0, 0, -16944444.4444444, -8.07721357665898e-08, -30254629.6296286, -3.36255335518765e-07, -15277777.7777769, -1967592.59259274, 0, 0, 0, 9722222.22222222, 3.36790763921418e-08, 22916666.666666, 2.34618629921561e-07, 16666666.6666656, 3.99575978423783e-07, 3472222.22222205, 0, 0, 0, -2314814.81481482, 0, -6944444.44444426, -5.52671284338404e-08, -6944444.44444404, -1.73051415787576e-07, -2314814.81481462, -1.06911996801986e-07}, 
                                                   {0, 0, 0, 0, 0, 4199.99999999996, 0, 0, -20116.6666666662, -107616.666666666, 0, 0, 35916.6666666667, 1067513.88888888, 490930.555555543, 0, 0, -28333.3333333374, -5460277.77777773, -833263.888888878, -4515208.33333321, 0, 0, 8333.33333334342, 15968402.7777777, 623611.111111162, 20661805.5555549, 7058680.55555538, 0, 0, -1.1131641424996e-08, -27652777.7777776, -173611.111111245, -51402777.7777762, -4826388.88888901, -28402777.7777768, 0, 0, 0, 27986111.111111, 1.30600590885566e-07, 70208333.3333312, 1215277.77777825, 57673611.1111089, 16666666.6666664, 0, 0, 0, -15277777.7777778, -5.42870216238081e-08, -49305555.5555541, -3.93653913216326e-07, -56249999.9999977, -3472222.22222265, -25694444.4444436, 0, 0, 0, 3472222.22222223, 0, 13888888.8888885, 1.00151431769774e-07, 20833333.3333325, 2.68813164926445e-07, 13888888.8888883, 3472222.22222221}, 
                                                   {0, 0, 0, 0, 0, -6299.99999999994, 0, 0, 40674.9999999992, 156174.999999998, 0, 0, -104166.666666665, -1479874.99999998, -954041.666666645, 0, 0, 132291.666666669, 7159513.88888882, 2306597.22222218, 8283819.44444424, 0, 0, -83333.3333333449, -19704513.8888887, -2758680.55555555, -35221874.9999991, -18192708.3333329, 0, 0, 20833.3333333512, 32138888.8888887, 1631944.44444452, 80624999.9999977, 19548611.1111109, 66423611.1111093, 0, 0, -1.52135871777169e-08, -30763888.8888888, -381944.444444532, -101388888.888886, -10243055.5555558, -119965277.777774, -59201388.8888877, 0, 0, 0, 15972222.2222222, 2.75512677485277e-08, 65972222.2222202, 2083333.3333337, 104166666.666663, 24305555.5555555, 76388888.8888868, 0, 0, 0, -3472222.22222224, 0, -17361111.1111106, -1.04140935029891e-07, -34722222.222221, -3472222.22222244, -34722222.2222212, -17361111.1111107}, 
                                                   {0, 0, 0, 0, 0, 6999.99999999993, 0, 0, -59194.4444444434, -164194.444444443, 0, 0, 206129.629629627, 1444046.29629628, 1300175.9259259, 0, 0, -378472.222222221, -6430138.88888882, -4227361.11111104, -10279027.7777775, 0, 0, 386574.074074081, 16370138.8888887, 7223379.62962952, 39190509.2592583, 29657175.9259254, 0, 0, -208333.333333353, -24930555.5555554, -6840277.77777768, -81006944.4444421, -44201388.8888883, -93645833.3333312, 0, 0, 46296.29629632, 22499999.9999999, 3402777.7777777, 93217592.5925896, 35648148.1481479, 147916666.666663, 109490740.740739, 0, 0, -1.59176839281162e-08, -11111111.1111111, -694444.44444438, -56249999.999998, -14583333.3333334, -114583333.33333, -62499999.9999992, -118055555.555553, 0, 0, 0, 2314814.81481483, -2.26262202347562e-08, 13888888.8888884, 2314814.81481485, 34722222.222221, 13888888.8888889, 46296296.296295, 34722222.2222215}, 
                                                   {0, 0, 0, 0, 0, -5999.99999999995, 0, 0, 65738.0952380944, 125738.095238094, 0, 0, -303527.777777775, -960908.730158719, -1264436.50793649, 0, 0, 766111.111111108, 3801388.88888884, 5333611.11111104, 8368888.88888873, 0, 0, -1142361.11111111, -8803472.22222212, -12230555.5555554, -27552777.7777771, -29837499.9999995, 0, 0, 1006944.44444446, 12430555.5555554, 16458333.3333331, 50729166.6666651, 55763888.8888882, 78611111.1111095, 0, 0, -486111.111111135, -10555555.5555555, -12986111.1111109, -53263888.8888869, -57638888.8888884, -107986111.111108, -110416666.666665, 0, 0, 99206.3492063672, 4960317.46031747, 5555555.55555542, 29861111.1111098, 31249999.9999998, 74999999.9999976, 76388888.888888, 100694444.444442, 0, 0, 0, -992063.492063504, -992063.492063453, -6944444.4444441, -6944444.4444444, -20833333.3333325, -20833333.3333332, -34722222.2222212, -34722222.2222215}, 
                                                   {0, 0, 0, 0, 0, 5249.99999999994, 0, 0, -96895.8333333322, -96895.8333333318, 0, 0, 598555.55555555, 598555.555555544, 1722111.11111108, 0, 0, -1829826.38888888, -1829826.38888885, -9929062.49999987, -9929062.49999981, 0, 0, 3128472.22222221, 3128472.22222214, 27973611.1111107, 27973611.1111104, 49690277.7777771, 0, 0, -3055555.55555557, -3055555.55555547, -43663194.4444438, -43663194.444443, -115711805.555554, -115711805.555553, 0, 0, 1597222.22222226, 1597222.22222219, 38576388.8888882, 38576388.8888872, 139930555.555554, 139930555.555552, 205902777.777775, 0, 0, -347222.222222255, -347222.222222239, -18055555.5555552, -18055555.5555545, -85416666.6666658, -85416666.6666642, -168402777.777776, -168402777.777774, 0, 0, 1.04565964595258e-08, 1.18595994447588e-08, 3472222.22222212, 3472222.22222194, 20833333.3333331, 20833333.3333326, 52083333.3333329, 52083333.333332, 69444444.4444432}, 
                                                   {0, 0, 0, 0, 0, 375.000000000001, 0, 0, -9354.16666666669, -4541.66666666668, 0, 0, 88805.5555555557, 8333.33333333346, 112451.388888889, 0, 0, -423576.38888889, -4166.6666666671, -1055520.83333333, -198611.111111113, 0, 0, 1107638.88888889, 0, 4952777.77777778, 95138.888888891, 1752777.77777779, 0, 0, -1597222.22222222, 0, -12656250, 0, -7465277.77777782, -781250.000000027, 0, 0, 1180555.55555556, 0, 17673611.1111111, 0, 16319444.4444445, 0, 2951388.88888899, 0, 0, -347222.222222223, 0, -12500000, 0, -17361111.1111111, 1.28810926970024e-08, -5208333.33333345, -5.20778861279668e-08, 0, 0, 0, 0, 3472222.22222222, 0, 6944444.44444443, 0, 3472222.22222227, 0, 4.61644972460914e-08}, 
                                                   {0, 0, 0, 0, 0, 374.999999999993, 0, 0, -4541.66666666657, -9354.1666666665, 0, 0, 8333.33333333309, 88805.555555554, 112451.388888886, 0, 0, -4166.66666666666, -423576.388888882, -198611.111111104, -1055520.83333331, 0, 0, 0, 1107638.88888887, 95138.8888888822, 4952777.77777768, 1752777.77777772, 0, 0, 0, -1597222.2222222, 0, -12656249.9999998, -781249.999999935, -7465277.77777756, 0, 0, 0, 1180555.55555555, 1.0901402952888e-08, 17673611.1111108, -1.46264117200186e-08, 16319444.444444, 2951388.88888866, 0, 0, 0, -347222.222222223, 0, -12499999.9999998, -2.22830509910531e-08, -17361111.1111108, 8.75273749407302e-08, -5208333.33333302, 0, 0, 0, 0, 0, 3472222.22222217, 0, 6944444.44444433, 0, 3472222.22222208, -8.83169508202835e-08}, 
                                                   {0, 0, 0, 0, 0, -4199.99999999996, 0, 0, 90116.6666666658, 69116.6666666657, 0, 0, -686194.44444444, -340611.111111105, -1446805.55555553, 0, 0, 2485694.44444443, 782638.888888867, 10565694.4444443, 6762638.88888876, 0, 0, -4850694.44444443, -937499.999999962, -35992361.1111107, -14607638.8888884, -44811805.555555, 0, 0, 5256944.44444445, 569444.444444418, 65138888.8888882, 16354166.6666658, 129722222.222221, 85624999.9999985, 0, 0, -2986111.11111114, -138888.888888899, -64722222.2222215, -9236111.11111011, -185069444.444443, -82638888.8888866, -196874999.999997, 0, 0, 694444.444444476, 2.53358764225327e-08, 33333333.3333329, 2083333.33333273, 128472222.222221, 38888888.8888872, 194444444.444443, 135416666.666664, 0, 0, -1.0426257247576e-08, -1.03694715375883e-08, -6944444.44444433, 1.50005346400433e-07, -34722222.2222219, -6944444.44444397, -69444444.444444, -34722222.2222213, -69444444.4444435}, 
                                                   {0, 0, 0, 0, 0, 2624.99999999998, 0, 0, -60697.9166666663, -38822.9166666663, 0, 0, 515451.388888887, 148177.083333331, 882378.472222213, 0, 0, -2124045.13888888, -242187.499999995, -7288020.83333327, -3218663.19444439, 0, 0, 4670138.88888888, 182291.666666662, 28820312.4999998, 4992187.49999982, 24585069.4444442, 0, 0, -5607638.88888889, -52083.3333333415, -59878472.2222218, -3550347.22222188, -85425347.2222216, -34652777.7777771, 0, 0, 3472222.22222224, 2.3899348824114e-08, 67013888.8888884, 954861.11111074, 143749999.999999, 22048611.1111101, 101302083.333332, 0, 0, -868055.555555573, -2.10767237150509e-08, -38194444.4444441, 2.10723332206976e-07, -114583333.333333, -5208333.33333262, -123263888.888888, -51215277.7777767, 0, 0, 0, 0, 8680555.55555548, -4.92589352912638e-08, 34722222.222222, -2.01579015101978e-07, 52083333.3333332, 8680555.55555514, 34722222.2222219}, 
                                                   {0, 0, 0, 0, 0, -1200, 0, 0, 29033.3333333333, 16033.3333333333, 0, 0, -263527.777777777, -44833.3333333332, -383361.111111109, 0, 0, 1183611.11111111, 46666.6666666667, 3415972.22222221, 1029027.77777777, 0, 0, -2871527.77777778, -16666.666666669, -14933333.3333333, -1022222.22222219, -8567361.11111107, 0, 0, 3798611.11111111, 0, 34874999.9999999, 347222.222222147, 33576388.8888888, 7847222.22222211, 0, 0, -2569444.44444445, -1.21485717362752e-08, -43819444.4444443, 8.44722385704814e-08, -65555555.5555554, -2430555.55555534, -26736111.111111, 0, 0, 694444.44444445, 0, 27777777.7777777, -4.56172053572947e-08, 60416666.6666665, -1.83629844024687e-07, 40277777.7777778, 6944444.44444427, 0, 0, 0, 0, -6944444.44444442, 0, -20833333.3333333, 5.52583997483721e-08, -20833333.3333333, 9.38859352908173e-08, -6944444.44444443}, 
                                                   {0, 0, 0, 0, 0, 299.999999999995, 0, 0, -6883.33333333322, -5133.33333333324, 0, 0, 57277.7777777771, 24833.3333333329, 117111.111111109, 0, 0, -224305.555555553, -36666.6666666657, -965416.666666651, -559027.777777765, 0, 0, 437499.999999997, 16666.6666666668, 3725694.44444439, 797222.222222183, 4506249.99999993, 0, 0, -402777.777777777, 0, -7104166.66666657, -347222.222222153, -16770833.3333332, -6041666.6666665, 0, 0, 138888.888888891, 0, 6319444.44444434, -6.83820827018152e-08, 30138888.8888887, 2430555.55555532, 20138888.8888886, 0, 0, 0, 0, -2083333.33333327, 3.49908622160292e-08, -24305555.5555554, 1.59493206955158e-07, -29166666.6666664, -6944444.44444416, 0, 0, 0, 0, -1.52132131586189e-08, 0, 6944444.44444439, -4.26049049399031e-08, 13888888.8888888, -1.00762642861287e-07, 6944444.44444433}, 
                                                   {0, 0, 0, 0, 0, 281.249999999992, 0, 0, -5749.99999999986, -5749.99999999983, 0, 0, 39322.9166666659, 39322.9166666655, 116927.08333333, 0, 0, -111979.166666664, -111979.166666663, -791753.472222204, -791753.472222202, 0, 0, 130208.333333331, 130208.333333326, 2214409.72222217, 2214409.72222215, 5262152.77777769, 0, 0, -52083.3333333335, -52083.333333326, -2491319.44444435, -2491319.44444431, -14210069.4444442, -14210069.4444442, 0, 0, 0, 0, 954861.111111018, 954861.111110959, 14930555.5555553, 14930555.5555552, 35763888.8888885, 0, 0, 0, 0, 5.33340748413525e-08, 9.26692987415922e-08, -5208333.33333318, -5208333.3333331, -32118055.5555552, -32118055.5555551, 0, 0, 0, 0, -1.3020441402357e-08, -2.38038745289768e-08, -3.81541315187923e-08, -6.67994436324921e-08, 8680555.55555546, 8680555.55555542, 17361111.1111109}, 
                                                   {0, 0, 0, 0, 0, 299.999999999991, 0, 0, -5133.33333333319, -6883.33333333312, 0, 0, 24833.3333333327, 57277.777777776, 117111.111111108, 0, 0, -36666.6666666653, -224305.555555548, -559027.777777763, -965416.666666641, 0, 0, 16666.6666666656, 437499.999999984, 797222.222222188, 3725694.44444434, 4506249.99999991, 0, 0, 0, -402777.777777759, -347222.222222176, -7104166.66666645, -6041666.66666649, -16770833.333333, 0, 0, 0, 138888.888888877, -3.87208051789173e-08, 6319444.44444419, 2430555.5555554, 30138888.8888884, 20138888.8888885, 0, 0, 0, 0, 2.07347105970242e-08, -2083333.33333317, 6.47573719197446e-08, -24305555.5555552, -6944444.44444416, -29166666.6666662, 0, 0, 0, 0, 0, -4.23779824699903e-08, -1.12500919539377e-08, 6944444.44444434, -5.7188492413804e-08, 13888888.8888887, 6944444.44444427}, 
                                                   {0, 0, 0, 0, 0, -1199.99999999998, 0, 0, 16033.333333333, 29033.3333333327, 0, 0, -44833.3333333322, -263527.777777773, -383361.111111102, 0, 0, 46666.6666666654, 1183611.11111109, 1029027.77777775, 3415972.22222215, 0, 0, -16666.6666666682, -2871527.77777773, -1022222.22222217, -14933333.333333, -8567361.11111089, 0, 0, 0, 3798611.11111105, 347222.222222185, 34874999.9999993, 7847222.22222191, 33576388.8888881, 0, 0, 0, -2569444.44444441, 1.23253448798874e-08, -43819444.4444436, -2430555.55555537, -65555555.5555542, -26736111.1111102, 0, 0, 0, 694444.444444444, 0, 27777777.7777772, 0, 60416666.6666655, 6944444.44444398, 40277777.7777766, 0, 0, 0, 0, 0, -6944444.44444429, -1.1687802701313e-08, -20833333.333333, 4.40837774648527e-08, -20833333.3333328, -6944444.44444408}, 
                                                   {0, 0, 0, 0, 0, 2624.99999999995, 0, 0, -38822.9166666658, -60697.9166666654, 0, 0, 148177.08333333, 515451.388888878, 882378.472222202, 0, 0, -242187.499999994, -2124045.13888884, -3218663.19444437, -7288020.83333317, 0, 0, 182291.666666665, 4670138.88888879, 4992187.49999983, 28820312.4999994, 24585069.4444439, 0, 0, -52083.3333333439, -5607638.88888877, -3550347.22222202, -59878472.2222208, -34652777.7777769, -85425347.2222205, 0, 0, 1.85604628948092e-08, 3472222.22222216, 954861.11111095, 67013888.8888872, 22048611.1111104, 143749999.999997, 101302083.333331, 0, 0, -1.39619957402736e-08, -868055.555555554, 8.89300703633933e-08, -38194444.4444433, -5208333.3333331, -114583333.333331, -51215277.7777764, -123263888.888886, 0, 0, 0, 0, -2.37600021958429e-08, 8680555.55555526, -3.38044544247123e-08, 34722222.2222215, 8680555.55555533, 52083333.3333323, 34722222.2222213}, 
                                                   {0, 0, 0, 0, 0, -4199.99999999992, 0, 0, 69116.6666666654, 90116.6666666649, 0, 0, -340611.111111105, -686194.44444443, -1446805.55555553, 0, 0, 782638.888888877, 2485694.44444439, 6762638.88888876, 10565694.4444442, 0, 0, -937499.999999991, -4850694.44444432, -14607638.8888886, -35992361.1111102, -44811805.5555548, 0, 0, 569444.444444456, 5256944.44444431, 16354166.6666662, 65138888.888887, 85624999.9999985, 129722222.22222, 0, 0, -138888.888888921, -2986111.11111105, -9236111.11111064, -64722222.22222, -82638888.8888874, -185069444.44444, -196874999.999996, 0, 0, 2.75483303412631e-08, 694444.444444451, 2083333.33333306, 33333333.3333319, 38888888.8888882, 128472222.222219, 135416666.666664, 194444444.44444, 0, 0, 0, -1.14572301704543e-08, 6.81115218943395e-08, -6944444.44444408, -6944444.44444429, -34722222.2222213, -34722222.2222217, -69444444.4444429, -69444444.444443}, 
                                                   {0, 0, 0, 0, 0, 2333.33333333328, 0, 0, -46175.925925925, -46175.9259259248, 0, 0, 304259.259259254, 304259.259259252, 900185.185185165, 0, 0, -873842.59259258, -873842.592592565, -5768518.51851841, -5768518.51851838, 0, 0, 1238425.92592591, 1238425.92592587, 15829861.1111108, 15829861.1111106, 35016203.7037031, 0, 0, -856481.481481483, -856481.48148143, -21215277.7777772, -21215277.7777768, -87303240.7407395, -87303240.7407391, 0, 0, 231481.481481505, 231481.481481466, 13773148.1481476, 13773148.1481471, 102893518.518517, 102893518.518516, 178703703.703701, 0, 0, -2.37111594288355e-08, -1.1832534701883e-08, -3472222.22222191, -3472222.22222157, -56712962.9629621, -56712962.9629613, -152777777.777776, -152777777.777775, 0, 0, 0, 0, -7.73818552761628e-08, -1.67770690332335e-07, 11574074.0740739, 11574074.0740736, 46296296.2962958, 46296296.2962954, 69444444.4444434}, 
                                                   {0, 0, 0, 0, 0, -999.999999999981, 0, 0, 21694.4444444441, 18361.1111111107, 0, 0, -165370.370370368, -104166.666666664, -394537.037037029, 0, 0, 572916.66666666, 225694.444444437, 2958333.33333328, 2194444.44444439, 0, 0, -960648.148148139, -208333.333333323, -9980324.07407391, -4575231.48148131, -15886574.0740738, 0, 0, 763888.888888887, 69444.4444444411, 16076388.8888886, 4027777.77777746, 50486111.1111105, 30798611.1111105, 0, 0, -231481.481481492, 0, -12152777.7777775, -1273148.14814781, -73726851.8518511, -24652777.7777769, -85185185.185184, 0, 0, 1.20224309081597e-08, 1.20485653670004e-08, 3472222.22222203, -1.92856244844689e-07, 48611111.1111106, 6944444.44444383, 93749999.9999991, 55555555.5555545, 0, 0, 0, 0, 4.5468708657259e-08, 4.65796648232122e-08, -11574074.0740739, 1.69037483864037e-07, -34722222.222222, -11574074.0740737, -34722222.2222218}, 
                                                   {0, 0, 0, 0, 0, -999.999999999973, 0, 0, 18361.1111111106, 21694.4444444438, 0, 0, -104166.666666664, -165370.370370366, -394537.037037027, 0, 0, 225694.444444439, 572916.666666648, 2194444.44444439, 2958333.33333326, 0, 0, -208333.333333328, -960648.148148109, -4575231.48148135, -9980324.0740738, -15886574.0740738, 0, 0, 69444.4444444466, 763888.888888844, 4027777.77777757, 16076388.8888883, 30798611.1111105, 50486111.1111102, 0, 0, -1.05927909073127e-08, -231481.481481458, -1273148.14814795, -12152777.7777771, -24652777.7777772, -73726851.8518506, -85185185.1851838, 0, 0, 0, 0, -1.10757689369792e-07, 3472222.2222218, 6944444.44444413, 48611111.1111102, 55555555.5555546, 93749999.9999986, 0, 0, 0, 0, 2.74799060900844e-08, 1.09717321893481e-07, 7.18585357514495e-08, -11574074.0740738, -11574074.0740738, -34722222.2222217, -34722222.2222217}};




double MAd_TriInterpolant::coef11[78][78] = {{1, -33.2186507936508, -33.2186507936566, 457.478432539684, 457.478432539862, 914.956865079441, -3509.76278549384, -3509.76278549606, -10529.2883564829, -10529.2883564831, 16870.6780836641, 16870.6780836792, 67482.7123346673, 67482.7123346707, 101224.068502008, -53820.583094963, -53820.5830950252, -269102.915474865, -269102.915474882, -538205.830949775, -538205.830949826, 117058.35357639, 117058.353576556, 702350.121458496, 702350.121458521, 1755875.3036463, 1755875.30364668, 2341167.07152887, -174492.021412866, -174492.021413159, -1221444.14989039, -1221444.14989036, -3664332.44967101, -3664332.4496721, -6107220.749453, -6107220.7494543, 175442.536532741, 175442.536533079, 1403540.2922624, 1403540.29226223, 4912391.02291776, 4912391.02291931, 9824782.0458371, 9824782.04584144, 12280977.5573004, -113712.75516011, -113712.755160356, -1023414.79644142, -1023414.79644117, -4093659.18576484, -4093659.18576592, -9551871.43345188, -9551871.4334579, -14327807.150183, -14327807.1501871, 42885.9533746704, 42885.9533747726, 428859.533746917, 428859.533746764, 1929867.90186064, 1929867.90186092, 5146314.40496141, 5146314.40496533, 9006050.20868542, 9006050.20869061, 10807260.2504269, -7147.65889577844, -7147.65889579696, -78624.2478536086, -78624.2478535714, -393121.239267933, -393121.23926792, -1179363.7178036, -1179363.71780457, -2358727.43560774, -2358727.43560971, -3302218.40985243, -3302218.40985346}, 
                                                   {0, 0.999999999999991, 0, -32.2186507936505, 0, 0, 425.259781746027, 0, 0, 0, -3084.50300374776, 0, 0, 0, 0, 13786.1750799161, 0, 0, 0, 0, 0, -40034.4080150458, 0, 0, 0, 0, 0, 0, 77023.9455613417, 0, 0, 0, 0, 0, 0, 0, -97468.07585152, 0, 0, 0, 0, 0, 0, 0, 0, 77974.4606812161, 0, 0, 0, 0, 0, 0, 0, 0, 0, -35738.2944788907, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7147.65889577816, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
                                                   {0, 0, 1.00000000000207, 0, -32.2186507937143, 0, 0, 425.259781746819, 0, 0, 0, -3084.50300375309, 0, 0, -1.10734992599115e-08, 0, 13786.175079938, 1.96463179405471e-08, 2.96086038811816e-08, 7.33813172032471e-08, 8.74149909124711e-08, 0, -40034.4080151041, -5.45321438127216e-08, -8.21365999867374e-08, -2.51936567025046e-07, -3.51824691158242e-07, -5.07546220163135e-07, 0, 77023.9455614438, 1.0223017111663e-07, 1.42797088459323e-07, 5.09222769895512e-07, 8.00580233217829e-07, 1.4592023491744e-06, 1.75607746788486e-06, 0, -97468.0758516371, -1.29036531420004e-07, -1.58004636832687e-07, -6.50057209866661e-07, -1.08401824361337e-06, -2.29303073773224e-06, -3.2684642787828e-06, -4.1840246458591e-06, 0, 77974.4606813008, 1.03983273706499e-07, 1.08042261078717e-07, 5.30573198098683e-07, 8.7274667381898e-07, 2.02691068862393e-06, 3.35239152438049e-06, 5.07459043586067e-06, 6.02779092773176e-06, 0, -35738.2944789258, -4.78924847372038e-08, -4.14809805429613e-08, -2.56706667258241e-07, -3.87439388155331e-07, -9.65736126845584e-07, -1.79947938638847e-06, -3.02394468798539e-06, -4.2047837620629e-06, -5.05260984517975e-06, 0, 7147.65889578447, 0, 0, 5.58346042910827e-08, 7.31053768531357e-08, 1.98365122388306e-07, 3.98294339176708e-07, 7.05954998347913e-07, 1.13674685131571e-06, 1.55676961099971e-06, 1.81407970023528e-06}, 
                                                   {0, 121, 0, -2688.45674603176, 0, -2688.45674603175, 25781.8661309525, 0, 51563.7322619053, 25781.8661309515, -141080.769604278, 0, -423242.308812839, -141080.76960427, -423242.308812816, 489463.582476304, 0, 1957854.32990527, 489463.58247627, 2936781.49485766, 1957854.32990514, -1128191.14725117, 0, -5640955.73625613, -1128191.14725105, -11281911.4725108, -5640955.73625573, -11281911.472511, 1753958.16298034, 0, 10523748.9778829, 1753958.16298004, 26309372.4447028, 10523748.977882, 35079163.2596032, 26309372.4447047, -1819994.79817297, 0, -12739963.5872124, -1819994.79817251, -38219890.7616291, -12739963.5872105, -63699817.9360445, -38219890.7616338, -63699817.9360496, 1208604.14055888, 0, 9668833.12447266, 1208604.14055846, 33840915.9356461, 9668833.12447028, 67681831.8712846, 33840915.9356515, 84602289.8391103, 67681831.8712967, -464597.828225592, 0, -4181380.45403124, -464597.828225384, -16725521.8161204, -4181380.45402961, -39026217.5709413, -16725521.8161233, -58539326.3564129, -39026217.5709518, -58539326.3564204, 78624.247853562, 0, 786242.478535831, 78624.2478535204, 3538091.15341022, 786242.47853537, 9434909.7424253, 3538091.15341079, 16511092.0492436, 9434909.74242847, 19813310.4590948, 16511092.0492479}, 
                                                   {0, -302.500000000001, 0, 8384.89186507939, 0, 5057.3918650794, -92270.3205853178, 0, -128909.330654764, -36639.0100694421, 554216.479392639, 0, 1259620.32741405, 151187.368628735, 856591.216650076, -2055189.48364887, 0, -6557696.8796795, -392128.428732607, -7341953.73714397, -3231574.76984652, 4977184.22615743, 0, 20572508.4147293, 663771.510098304, 32518191.3973353, 7632270.26655055, 23891365.9652175, -8035638.71299194, 0, -40912345.6668729, -734152.101909547, -84027147.6394605, -11706399.2225403, -87697908.1490042, -47519714.5840537, 8587823.5559359, 0, 52039091.7705504, 512150.434928627, 131890255.948601, 11660726.1655089, 179438727.64262, 59209197.8595493, 139060362.0376, -5838337.74350616, 0, -41073047.1638376, -204682.959287947, -124037873.328639, -7271118.45852259, -208640163.167726, -45166706.3495989, -211505724.597765, -129768996.18871, 2287250.84664905, 0, 18333745.0676748, 35738.2944787792, 64328930.0620019, 2573157.20247924, 129086719.657728, 19298679.0186042, 162108903.756213, 66044368.1969933, 130587728.025854, -393121.239267807, 0, -3538091.15341106, 1.82120591532467e-08, -14152364.6136409, -393121.239267477, -33022184.0984887, -3538091.15341095, -49533276.1477295, -14152364.6136425, -49533276.1477343, -33022184.0984946}, 
                                                   {0, 605.000000000002, 0, -17878.9503968254, 0, -7896.45039682546, 211218.300231482, 0, 218940.168915347, 44324.3686838571, -1348939.384375, 0, -2359995.57182546, -139852.052083305, -1150908.2395336, 5260649.16457508, 0, 13371789.1881785, 271465.999826324, 11233096.8824557, 3393422.85867991, -13272405.8356771, 0, -44960741.8791695, -332167.687499888, -55579958.3109313, -6140027.43463426, -29699482.0145773, 22154815.849566, 0, 94600742.2746608, 250356.016319239, 157105166.619255, 6982902.94166564, 125509560.721847, 47582867.4530287, -24333348.1730407, 0, -126010272.754528, -106328.810019543, -265157469.986957, -4875175.93940882, -287300444.673544, -47114295.7197864, -166165347.858466, 16920457.9678241, 0, 103316160.402627, 19493.615170041, 264586838.706523, 1910374.28668859, 365427309.982435, 27973337.7693876, 289967525.658192, 128813809.045346, -6754537.65651045, 0, -47603408.2458909, 1.1809080210438e-07, -144096803.338884, -321644.650308926, -243163355.634318, -9006050.20868278, -247666380.738635, -54036301.2520793, -153102853.547531, 1179363.71780341, 0, 9434909.74242911, -1.87167540515459e-08, 33022184.0984952, -3.74082316874319e-07, 66044368.1969762, 1179363.71780406, 82555460.2462094, 9434909.74242791, 66044368.1969699, 33022184.0984901}, 
                                                   {0, -907.500000000002, 0, 27650.3005952381, 0, 9349.05059523814, -339123.350892858, 0, -270092.51478175, -40776.6638888804, 2252520.51288029, 0, 3068820.65616741, 97642.2524305102, 1115256.14571738, -9115708.9505043, 0, -18403971.2124264, -138682.805555449, -11673429.4359355, -2523849.97957068, 23773979.5447917, 0, 65306832.9056753, 116873.815972057, 62439118.5369117, 3381138.99189598, 24170530.3519577, -40852012.1223958, 0, -144212959.508634, -54131.030555317, -190046154.171908, -2681741.47210416, -113434957.718533, -29377361.3744036, 46008476.0955853, 0, 200390819.657081, 10632.8810016863, 342602058.764896, 1166072.61654911, 286680193.281726, 20914876.9308877, 118219915.273658, -32690792.6406002, 0, -170432677.433991, 2.51302517001705e-07, -362015927.327694, -214429.766871894, -397767217.549927, -8050863.06532971, -235385403.181292, -69728661.4641282, 13294645.5461475, 0, 81054451.8781367, -1.14925193340227e-07, 207139154.799645, -9.9487675799606e-07, 285191589.941466, 1286578.60123964, 225151255.216892, 21014117.1535711, 99066552.2954181, -2358727.43560682, 0, -16511092.0492504, 1.89898832198424e-08, -49533276.1477427, 3.22426696736684e-07, -82555460.246217, -2.69462128580184e-07, -82555460.2461957, -2358727.43560469, -49533276.1477094, -16511092.0492354}, 
                                                   {0, 1016.4, 0, -31527.3566666667, 0, -8234.85666666668, 395620.894166667, 0, 243821.823055559, 27553.1788888795, -2696956.92800926, 0, -2862190.10254637, -48742.3291666161, -777654.003703466, 11214051.9626157, 0, 17834291.9116906, 48091.6180554345, 8501572.86330794, 1309378.18229045, -30032326.3399652, 0, -65921527.6651544, -25097.1141664886, -47947072.919029, -1227585.15127053, -13260359.6309497, 52906992.6271179, 0, 151565081.187882, 5413.10305534431, 154653256.481039, 607620.817983066, 66836034.5188184, 11443074.3134625, -60958306.7843748, 0, -218775070.909511, 2.19052340749887e-07, -295376117.794581, -124050.278354225, -183185046.048907, -5172896.60744403, -50674538.7085439, 44211519.2062499, 0, 192674892.343308, -1.73707629188015e-07, 329539564.453965, -1.25904868740557e-06, 275639718.507965, 955187.143333391, 113257904.139323, 19649564.0916028, -18312302.0909838, 0, -94563527.1911576, 8.20536004527344e-08, -198133104.590963, 6.12658709548267e-07, -213143188.272019, 2.89174871871291e-06, -121581677.817058, -3002016.7362007, -33322385.7720376, 3302218.40984953, 0, 19813310.4591, -1.42749308690308e-08, 49533276.1477425, -1.85334679841778e-07, 66044368.1969686, -1.09589307097007e-07, 49533276.1476987, -4.49116742729281e-06, 19813310.4590596, 3302218.4098329}, 
                                                   {0, -846.999999999999, 0, 26583.3638888889, 0, 5309.5472222222, -338748.065231481, 0, -159811.752731483, -13226.8124999928, 2351730.95972222, 0, 1918138.76145838, 16369.4513888499, 381321.794791485, -9977957.57737265, 0, -12281163.9589126, -10065.6874999055, -4294665.37308858, -451408.804049978, 27293111.6562499, 0, 46822418.7425375, 2460.50138875564, 25153090.5732561, 265119.024650657, 4749690.36856793, -49111054.1094095, 0, -111281895.978237, 1.33059819312536e-07, -84825580.3401425, -61799.5932150278, -25185589.6957396, -2592650.81763326, 57757809.6027775, 0, 166128132.775014, -1.0952352596459e-07, 170147361.793722, -1.82483432639023e-06, 73499789.9261342, 558226.252585572, 12280977.5572298, -42710510.8381364, 0, -151192479.260895, 7.89975445267287e-08, -199634112.959065, 7.58345436104142e-07, -118852571.693244, 1.12157801381974e-05, -30020167.3621585, -2319740.20521268, 18012100.417361, 0, 76551426.7737931, -3.96670482179501e-08, 126084702.921521, -2.14235683656845e-07, 99066552.2954223, -3.29644113692672e-06, 36024200.834626, -2.39038813806215e-05, 4503025.10427559, -3302218.40984953, 0, -16511092.0492496, 0, -33022184.0984948, 5.64410552576898e-08, -33022184.0984793, 2.45035656113503e-07, -16511092.0492152, 4.5034575932274e-06, -3302218.40982036, 1.44183043976654e-05}, 
                                                   {0, 518.571428571427, 0, -16411.3452380952, 0, -2435.84523809522, 211445.987499999, 0, 74186.6125000004, 4270.29166666293, -1487790.43554894, 0, -904899.980489438, -3311.65476188427, -125019.324702286, 6410695.23609454, 0, 5912150.7880128, 958.636904711964, 1438956.60013139, 93083.6434518792, -17835867.0053818, 0, -23087294.6155106, 6.74401699609421e-08, -8668653.95572512, -25835.2645821803, -1007165.23518073, 32675326.6318698, 0, 56379723.7832215, -5.6188633672609e-08, 30264884.7295617, -1.42249672499104e-06, 5546175.17231934, 261633.314342488, -39139634.9683033, 0, -86675701.6345143, 3.19916717903084e-08, -63228426.8782844, 9.72293055824039e-07, -16994888.1347964, 1.06113057454075e-05, -1302527.92271, 29474346.1374998, 0, 81327362.4905165, -1.84468609598219e-08, 77643069.223311, -2.9877857326373e-07, 29201435.5250646, -6.73353072185347e-06, 3411382.65474561, -2.98752217754477e-05, -12651356.2455274, 0, -42457093.8409267, 1.09466083790421e-08, -51463144.0495994, 0, -26160431.5585161, 2.08746675675582e-06, -4503025.10429076, 1.40026507879746e-05, 3.51964023783197e-05, 2358727.4356068, 0, 9434909.7424282, 0, 14152364.6136405, 0, 9434909.7424192, -1.74650168801744e-07, 2358727.43559033, -2.74808560551593e-06, -1.54752051131953e-05, -8.23219914846915e-06}, 
                                                   {0, -226.874999999999, 0, 7224.52827380949, 0, 753.729166666655, -93856.6930059519, 0, -23161.1825892857, -831.874999998718, 667224.479075723, 0, 285997.458614423, 305.020833326297, 24641.3258928245, -2910101.43150213, 0, -1898004.54201809, 1.71263361174133e-08, -288365.969592949, -8699.62991054059, 8209155.32135411, 0, 7554149.34745411, -2.20124108213463e-08, 1775559.31475556, -4.06346139363909e-07, 96164.5959475354, -15270025.4007811, 0, -18864100.2837107, 1.45247966559415e-08, -6372237.25320837, 5.06025039533812e-07, -545257.359860529, 3.31239617762871e-06, 18591592.4319691, 0, 29747256.7498865, 0, 13769580.8975646, -3.30579390485234e-07, 1736703.89697667, -3.87270402066018e-06, -1.15470569625513e-05, -14240085.8819071, 0, -28704348.3382753, 0, -17602734.498781, 6.82393496615302e-08, -3138472.04240183, 2.52825274337709e-06, 1.98288378327374e-05, 1.07656047121223e-05, 6218463.23932701, 0, 15438943.2148823, 0, 12222496.7117794, 2.98537762876802e-08, 3002016.73621612, -8.00631109170364e-07, -1.67955889939947e-05, -5.20703645066794e-06, -1.24643140177113e-05, -1179363.7178034, 0, -3538091.15341053, 0, -3538091.15341009, -1.11969370472497e-08, -1179363.71780073, 6.87020120096717e-08, 5.52598016092819e-06, 1.04456262198631e-06, 5.37474950817772e-06, 3.01812906238097e-06}, 
                                                   {0, 67.2222222222218, 0, -2150.87103174602, 0, -141.166666666663, 28123.8744819222, 0, 4368.26805555552, 73.9444444441824, -201560.429546956, 0, -54463.406431879, 0, -2210.6748015806, 887732.835066681, 0, 365971.177540797, 0, 26212.5059578298, -3.56356069692337e-08, -2532932.39852428, 0, -1479222.67873271, 0, -164238.46770805, 8.4261501805273e-08, 3.15344430527424e-07, 4773116.39221639, 0, 3763234.35341458, 0, 602771.579831471, -1.05239327018756e-07, -1.29546404295753e-06, -6.9083946233469e-07, -5895932.51560014, 0, -6066058.61163236, 0, -1339743.006249, 6.49861451939594e-08, 2.84055133073835e-06, 8.25632374690839e-07, 2.38657398817958e-06, 4587497.43674489, 0, 6004033.47245414, 0, 1773918.98049691, 0, -3.43917480624696e-06, -5.48134354721042e-07, -4.05000938687081e-06, -2.27743722970475e-06, -2037082.78529678, 0, -3323661.38653712, 0, -1286578.60123979, -1.24037540585592e-08, 2.14352464717508e-06, 1.74617905522439e-07, 3.3912032898325e-06, 1.1269378582322e-06, 2.60395839117344e-06, 393121.239267799, 0, 786242.478535662, 0, 393121.23926777, 0, -5.33858288828669e-07, -1.48025412299237e-08, -1.10565439168475e-06, -2.29138422224982e-07, -1.10518520456315e-06, -6.47660670906425e-07}, 
                                                   {0, -12.0999999999999, 0, 388.635674603171, 0, 12.0999999999994, -5107.98979166662, 0, -376.535674603164, 0, 36849.3409336416, 0, 4731.45411706347, 0, 0, -163600.929785327, 0, -32117.886816579, 0, 0, 0, 471268.03268518, 0, 131483.042968755, 0, -2.70165458891848e-08, 0, -2.93283272172441e-08, -898011.242320592, 0, -339784.989716451, 0, 6.79331602604354e-08, 0, 1.19903833593274e-07, 6.48221199823757e-08, 1123541.09254298, 0, 558226.252604198, 0, -9.81473197548924e-08, 0, -2.61882533250904e-07, -7.85949000414811e-08, -2.2256540672909e-07, -886959.490248834, 0, -565314.839938856, 0, 7.90106260339607e-08, 0, 3.16228157350514e-07, 5.26446354265841e-08, 3.74404409470564e-07, 2.15842661388228e-07, 400268.898163577, 0, 321644.650310039, 0, -3.15125199077341e-08, 0, -1.97198802542015e-07, -1.66922586707657e-08, -3.10907371450277e-07, -1.08628322400842e-07, -2.44610932399232e-07, -78624.2478535598, 0, -78624.247853565, 0, 0, 0, 4.92549337414543e-08, 0, 1.00779013019754e-07, 2.22870433585637e-08, 1.02514907604145e-07, 6.21388617053169e-08}, 
                                                   {0, 0, 0, 0, 0, -12.0999999999999, 0, 0, 376.535674603177, 0, 0, 0, -4731.45411706362, 0, 0, 0, 0, 32117.8868165799, 0, 0, 0, 0, 0, -131483.042968758, 0, 2.45318328246999e-08, 0, 2.86207587334361e-08, 0, 0, 339784.989716459, 0, -5.99470046445434e-08, 0, -1.16390517605631e-07, -6.41674654167704e-08, 0, 0, -558226.252604209, 0, 8.29244150318084e-08, 0, 2.52356243880387e-07, 7.8237356055487e-08, 2.1978023826714e-07, 0, 0, 565314.839938866, 0, -6.20389377843797e-08, 0, -3.0199566194883e-07, -5.25252630069862e-08, -3.68351744860615e-07, -2.14630080587237e-07, 0, 0, -321644.650310044, 0, 2.13056873491032e-08, 0, 1.86241221447656e-07, 1.66636535428428e-08, 3.04527893907444e-07, 1.08364442628556e-07, 2.42705890682499e-07, 0, 0, 78624.2478535661, 0, 0, 0, -4.58676721954951e-08, 0, -9.82100061446456e-08, -2.22638244393527e-08, -1.01433233697885e-07, -6.19136069295941e-08}, 
                                                   {0, 0, 0, 0, 0, -6.72222222222213, 0, 0, 200.970436507937, 73.9444444444341, 0, 0, -2382.95508708118, 0, -2210.67480158706, 0, 0, 14930.7697916672, 0, 26212.5059578901, 0, 0, 0, -54797.4163483827, 0, -164238.467708322, 0, 1.00181716598533e-08, 0, 0, 121794.81875001, 0, 602771.579832145, 0, -4.91786826075484e-08, -1.40379242011948e-08, 0, 0, -161265.361863444, 0, -1339743.00624995, 0, 1.24246834095716e-07, 0, 6.41830630367253e-08, 0, 0, 116961.691021844, 0, 1773918.98049765, 0, -1.68428176225393e-07, 0, -1.43621002241975e-07, -2.8639797551864e-08, 0, 0, -35738.2944789019, 0, -1286578.60124007, 0, 1.12719433142268e-07, 0, 1.45955259083484e-07, 0, 5.72299868706649e-08, 0, 0, 0, 0, 393121.239267801, 0, -2.9875435989675e-08, 0, -5.35629824553099e-08, 0, -3.66126846340707e-08, 0}, 
                                                   {0, 0, 0, 0, 0, -5.04166666666695, 0, 0, 143.795535714294, 83.187500000001, 0, 0, -1589.4974537038, -305.020833333351, -2372.62633928574, 0, 0, 9012.51834490803, 0, 26226.7079861111, 8699.62991071468, 0, 0, -28705.8495370395, 0, -148706.552690971, 0, -96164.5959490771, 0, 0, 51875.5709490806, 0, 473646.517361105, 0, 545257.359866905, 1.60217721001178e-08, 0, 0, -49620.1113426035, 0, -855946.920659707, 0, -1736703.89699074, -3.14507082306866e-08, -3.86255435544585e-08, 0, 0, 19493.6151703155, 0, 818731.837152762, 0, 3138472.04241896, 3.14014079421586e-08, 3.98539424102807e-08, 6.81773690053425e-08, 0, 0, 0, 0, -321644.650310011, 0, -3002016.73622683, -1.51261501918595e-08, 0, -4.52108295945833e-08, -5.57512312818653e-08, 0, 0, 0, 0, 0, 0, 1179363.7178034, 0, 0, 1.16324030781767e-08, 1.66609297990961e-08, 1.94448554669105e-08}, 
                                                   {0, 0, 0, 0, 0, -4.32142857142876, 0, 0, 116.462500000005, 87.148809523811, 0, 0, -1179.41388888894, -522.892857142874, -2348.6604166667, 0, 0, 5871.65104166698, 958.63690476196, 23784.8467592593, 14091.9625000005, 0, 0, -15378.1336805568, 0, -118411.629340277, -25835.2645833352, -142709.080555558, 0, 0, 20299.1364583365, 0, 310125.695891198, 0, 710469.776041669, 261633.314351864, 0, 0, -10632.8810019896, 0, -409365.918576377, 0, -1860754.17534721, -2.72681829246466e-08, -1302527.92274308, 0, 0, 0, 0, 214429.766873332, 0, 2456195.51145831, 2.7359113831512e-08, 3411382.65480325, 4.6840506170682e-08, 0, 0, 0, 0, 0, 0, -1286578.60124006, -1.40610231591915e-08, -4503025.10434027, -3.7272800506841e-08, -3.21673627603964e-08, 0, 0, 0, 0, 0, 0, 0, 0, 2358727.4356068, 1.19507607718144e-08, 0, 1.35731012199135e-08}, 
                                                   {0, 0, 0, 0, 0, -4.03333333333345, 0, 0, 101.303888888892, 92.4305555555567, 0, 0, -915.062500000033, -711.715277777786, -2321.54745370374, 0, 0, 3802.5930555557, 2236.81944444448, 20970.1822916669, 17875.915393519, 0, 0, -7381.5041666672, -2460.50138888885, -87142.7575231479, -56181.4483796317, -161470.403645836, 0, 0, 5413.10305555689, 0, 169159.470486108, 61799.593217597, 670999.232928243, 507478.411458351, 0, 0, 0, 0, -124050.278356475, 0, -1302527.92274305, -558226.252604208, -2108854.73206021, 0, 0, 0, 0, 0, 0, 955187.143344883, 5.0038648971751e-08, 4093659.1857639, 2319740.20526626, 0, 0, 0, 0, 0, 0, 2.13152343733917e-08, -2.96033793375501e-08, -3002016.73622683, -5.86934425518769e-08, -4503025.10434031, 0, 0, 0, 0, 0, 0, 0, 0, -1.1500602690408e-08, 1.96227340622645e-08, 3302218.40984954, 1.58673365200647e-08}, 
                                                   {0, 0, 0, 0, 0, -4.03333333333333, 0, 0, 92.4305555555558, 101.303888888889, 0, 0, -711.715277777775, -915.062500000006, -2321.54745370371, 0, 0, 2236.81944444443, 3802.59305555559, 17875.9153935184, 20970.1822916668, 0, 0, -2460.50138888882, -7381.50416666673, -56181.448379628, -87142.7575231498, -161470.403645834, 0, 0, 0, 5413.10305555558, 61799.5932175878, 169159.470486115, 507478.411458327, 670999.232928251, 0, 0, 0, 0, 0, -124050.278356487, -558226.25260414, -1302527.92274309, -2108854.73206019, 0, 0, 0, 0, 0, 0, -3.67204205163704e-08, 955187.143344964, 2319740.20526617, 4093659.18576393, 0, 0, 0, 0, 0, 0, 2.26713199154034e-08, -3.97217180446434e-08, 4.2997514410213e-08, -3002016.73622689, -4503025.10434027, 0, 0, 0, 0, 0, 0, 0, 1.04606174009694e-08, -1.71860456156677e-08, 1.73756937205017e-08, 0, 3302218.40984954}, 
                                                   {0, 0, 0, 0, 0, -4.32142857142837, 0, 0, 87.148809523803, 116.462499999997, 0, 0, -522.892857142795, -1179.41388888888, -2348.66041666654, 0, 0, 958.636904761567, 5871.65104166674, 14091.9624999989, 23784.8467592587, 0, 0, 0, -15378.1336805561, -25835.2645833276, -118411.629340277, -142709.08055555, 0, 0, 0, 20299.1364583344, -1.42891276591166e-08, 310125.695891205, 261633.31435182, 710469.776041659, 0, 0, 0, -10632.8810019859, 1.93547597407392e-08, -409365.91857639, 6.970559232615e-08, -1860754.17534723, -1302527.92274299, 0, 0, 0, 0, -1.5039747851286e-08, 214429.766873345, -7.96323480492325e-08, 2456195.51145837, -1.19020313785649e-07, 3411382.65480319, 0, 0, 0, 0, 0, 0, 4.36350392923297e-08, -1286578.60124012, 1.04001837660865e-07, -4503025.10434028, 7.1759072968721e-08, 0, 0, 0, 0, 0, 0, 0, 1.32320259308422e-08, -3.28817786012148e-08, 2358727.43560682, -3.50027290370945e-08, -1.49105244554904e-08}, 
                                                   {0, 0, 0, 0, 0, -5.04166666666632, 0, 0, 83.1874999999903, 143.79553571428, 0, 0, -305.02083333324, -1589.49745370367, -2372.62633928554, 0, 0, 0, 9012.51834490739, 8699.62991071271, 26226.7079861101, 0, 0, 0, -28705.8495370373, 0, -148706.55269097, -96164.5959490657, 0, 0, 0, 51875.5709490751, -1.79679513769495e-08, 473646.517361109, -3.89819224025903e-08, 545257.359866883, 0, 0, 0, -49620.1113425941, 2.42103144452559e-08, -855946.920659721, 8.1669640441409e-08, -1736703.89699074, 8.74918878357333e-08, 0, 0, 0, 19493.6151703058, -1.84438397319648e-08, 818731.837152773, -8.74809714003495e-08, 3138472.04241902, -1.50993460816855e-07, -8.00794054256943e-08, 0, 0, 0, 0, 0, -321644.650310012, 4.85088515111651e-08, -3002016.7362269, 1.14750004204845e-07, 3.15337944892571e-08, 1.15136757915588e-07, 0, 0, 0, 0, 0, 0, -1.02256890955067e-08, 1179363.71780342, -3.51439042275799e-08, 0, -4.26164270322302e-08, -3.45783660393771e-08}, 
                                                   {0, 0, 0, 0, 0, -6.72222222222149, 0, 0, 73.944444444436, 200.970436507918, 0, 0, 0, -2382.95508708093, -2210.67480158715, 0, 0, 0, 14930.7697916655, 0, 26212.5059578915, 0, 0, 0, -54797.4163483753, 0, -164238.46770833, 0, 0, 0, 0, 121794.81874999, -1.40555078788096e-08, 602771.579832168, -3.30011642412917e-08, -1.13339413928106e-08, 0, 0, 0, -161265.36186341, 1.81982702669105e-08, -1339743.00624998, 6.40328253819428e-08, 0, 8.46209666681162e-08, 0, 0, 0, 116961.691021811, -1.48527127121552e-08, 1773918.98049766, -5.76921008944538e-08, 4.51090150436146e-08, -1.47052646525921e-07, -9.50326697489545e-08, 0, 0, 0, -35738.2944788842, 0, -1286578.60124005, 2.55203013547572e-08, -4.81347454043712e-08, 9.47655880998119e-08, 2.86843898684883e-08, 1.44402443718824e-07, 0, 0, 0, 0, 0, 393121.239267793, 0, 1.55382367489449e-08, -2.16734965784946e-08, 0, -5.12433786260195e-08, -4.81248866857787e-08}, 
                                                   {0, 0, 0, 0, 0, -12.0999999999981, 0, 0, 0, 376.53567460312, 0, 0, 0, -4731.45411706284, 0, 0, 0, 0, 32117.8868165741, 0, 0, 0, 0, 0, -131483.042968732, 0, -1.22276554238315e-08, 0, 0, 0, 0, 339784.989716389, -2.83021384620774e-08, 3.41286983480548e-08, -2.47067793396698e-08, 5.73557718852215e-08, 0, 0, 0, -558226.252604092, 4.4713368080476e-08, -4.59547027082643e-08, 9.15578152101116e-08, -1.8192814031908e-07, 0, 0, 0, 0, 565314.839938751, -3.99876013314561e-08, 2.30233374904019e-08, -1.16205220044724e-07, 2.80475848242196e-07, -1.2748986862353e-07, 1.06659895202299e-07, 0, 0, 0, -321644.650309982, 1.9364852093697e-08, 0, 7.10928203640846e-08, -2.04998046048742e-07, 1.21615285240815e-07, -1.73196595457732e-07, 7.37268363972579e-08, 0, 0, 0, 78624.2478535519, 0, 0, -1.80237336298563e-08, 5.59790983120823e-08, -3.56980480111344e-08, 8.28565397042831e-08, -5.08315615648224e-08, 0}, 
                                                   {0, 0, -12.1000000000244, 0, 388.635674603924, 12.1000000003661, 0, -5107.98979167596, 0, -376.535674611375, 0, 36849.3409337046, 5.2340644011622e-08, 4731.45411713722, 1.29248440989153e-07, 0, -163600.929785586, -2.27799592411275e-07, -32117.8868169252, -8.53561927890643e-07, -1.02189668686113e-06, 0, 471268.032685869, 6.31923407440036e-07, 131483.042969711, 2.92241619919893e-06, 4.12030007607617e-06, 5.90966965565299e-06, 0, -898011.242321799, -1.18436344569822e-06, -339784.989718105, -5.90060559530596e-06, -9.39486028753902e-06, -1.69349536454206e-05, -2.04834803555354e-05, 0, 1123541.09254436, 1.49472009984099e-06, 558226.252606009, 7.52999052072454e-06, 1.27457393573365e-05, 2.65603963339559e-05, 3.82122939412762e-05, 4.85947997740398e-05, 0, -886959.490249835, -1.20445326864612e-06, -565314.839940077, -6.14695526060984e-06, -1.02728578931175e-05, -2.34616892897642e-05, -3.93101800825088e-05, -5.87873950627264e-05, -7.01047609482681e-05, 0, 400268.898163991, 5.54564905412038e-07, 321644.650310499, 2.97510594170106e-06, 4.5591018694704e-06, 1.11769877366284e-05, 2.11766929176605e-05, 3.49933492954127e-05, 4.90071843985585e-05, 5.8554135939041e-05, 0, -78624.2478536345, -1.10045281504398e-07, -78624.2478536385, -6.47359506518289e-07, -8.58270106195314e-07, -2.29513872402188e-06, -4.70293989412237e-06, -8.16854681523885e-06, -1.329010399474e-05, -1.80122205554273e-05, -2.10425050970101e-05}, 
                                                   {0, 0, 67.2222222223537, 0, -2150.87103175007, -141.166666668618, 0, 28123.8744819725, 73.9444444798701, 4368.26805559932, 0, -201560.429547295, -2.77198164861356e-07, -54463.406432272, -2210.6748022751, 0, 887732.835068079, 1.20405664934748e-06, 365971.177542637, 4.52600075363921e-06, 26212.5059633386, 0, -2532932.398528, -3.33796718726795e-06, -1479222.67873777, -1.54528642189048e-05, -164238.467730332, -3.13662973515017e-05, 0, 4773116.3922229, 6.25457537041373e-06, 3763234.35342326, 3.11695264664908e-05, 602771.579882438, 8.9569199373616e-05, 0.000108905981611587, 0, -5895932.51560761, -7.891741935997e-06, -6066058.61164175, -3.97719517437859e-05, -1339743.00631832, -0.000140209430463766, -0.000203651086241401, -0.000257192267304666, 0, 4587497.43675029, 6.35797061117191e-06, 6004033.47246035, 3.24777521160462e-05, 1773918.9805528, 0.000123787252657964, 0.000210163862229, 0.000310299882360513, 0.000371534665211955, 0, -2037082.78529901, -2.92610231889292e-06, -3323661.3865394, -1.57251880291998e-05, -1286578.60126452, -5.89792857001175e-05, -0.000113640976576363, -0.00018451947966173, -0.000260312490584688, -0.000309143031004271, 0, 393121.239268203, 5.80445408022036e-07, 786242.478536014, 3.4229343751166e-06, 393121.239272389, 1.21119505153995e-05, 2.5326510137182e-05, 4.30752492412357e-05, 7.08272197229246e-05, 9.49384196915327e-05, 0.000111194823500631}, 
                                                   {0, 0, -226.875000000427, 0, 7224.52827382265, 753.729166672928, 0, -93856.6930061152, -831.875000113288, -23161.1825894263, 0, 667224.479076825, 305.020834216933, 285997.458615684, 24641.3258950603, 0, -2910101.43150667, -3.83032291094276e-06, -1898004.54202398, -8699.62992515818, -288365.969610724, 0, 8209155.32136618, 1.06117539586039e-05, 7554149.34747023, 4.91748112166509e-05, 1775559.31482764, 96164.5960492647, 0, -15270025.4008023, -1.98794597549581e-05, -18864100.2837381, -9.9098011576688e-05, -6372237.25337365, -0.000285022584266597, -545257.360215358, 0, 18591592.4319934, 2.507508989294e-05, 29747256.7499157, 0.000126464976303871, 13769580.8977898, 0.000445314434075515, 1736703.89764399, 0.000818904452597732, 0, -14240085.8819247, -2.01947043769961e-05, -28704348.3382942, -0.000103326370764819, -17602734.4989626, -0.000393017190201292, -3138472.04309539, -0.000985173784385439, -0.00118457622978451, 0, 6218463.23933429, 9.28881065293885e-06, 15438943.214889, 5.00519073961883e-05, 12222496.7118596, 0.000187334137589956, 3002016.73659405, 0.000585263993328762, 0.000831966211469896, 0.000981624462034527, 0, -1179363.71780471, -1.8416147485066e-06, -3538091.15341152, -1.08986418789937e-05, -3538091.15342495, -3.84885525735649e-05, -1179363.71788555, -0.000136654239230273, -0.000227173050810872, -0.000300931537531731, -0.000353381774882735}, 
                                                   {0, 0, 518.571428572357, 0, -16411.3452381238, -2435.84523810869, 0, 211445.987500354, 4270.29166690903, 74186.6125003024, 0, -1487790.43555133, -3311.6547637888, -904899.980492143, -125019.324707102, 0, 6410695.23610443, 958.636912912143, 5912150.7880254, 93083.6434832114, 1438956.60016976, 0, -17835867.0054081, -2.25656289212678e-05, -23087294.6155449, -25835.2646879889, -8668653.95588111, -1007165.23539922, 0, 32675326.6319159, 4.22649246184134e-05, 56379723.7832792, 0.00021072864926651, 30264884.7299203, 261633.314958244, 5546175.17308364, 0, -39139634.9683563, -5.32909213995608e-05, -86675701.6345749, -0.000269036278713492, -63228426.878774, -0.000945561192513353, -16994888.1362396, -1302527.92448616, 0, 29474346.1375382, 4.28976588248092e-05, 81327362.4905548, 0.000219984665776692, 77643069.2237057, 0.000834341165773459, 29201435.5265724, 0.00209055129295029, 3411382.65732828, 0, -12651356.2455433, -1.97171155389165e-05, -42457093.8409398, -0.000106625039855478, -51463144.0497729, -0.000397977512073754, -26160431.5593424, -0.00124071931676531, -4503025.1061183, -0.00208311031713751, 0, 2358727.43560967, 3.90626132401303e-06, 9434909.74243002, 2.32252501437387e-05, 14152364.6136724, 8.18389275412105e-05, 9434909.74260559, 0.000289778960722936, 2358727.43609419, 0.000637424866118068, 0.000750553670617437}, 
                                                   {0, 0, -847.000000001421, 0, 26583.3638889326, 5309.54722224252, 0, -338748.065232024, -13226.8125003643, -159811.752731939, 0, 2351730.95972589, 16369.451391711, 1918138.76146246, 381321.794798773, 0, -9977957.57738778, -10065.6875121824, -12281163.9589315, -451408.804097149, -4294665.37314677, 0, 27293111.6562902, 2460.50142259924, 46822418.7425886, 265119.024809199, 25153090.5734933, 4749690.36889752, 0, -49111054.1094801, -6.31304878876059e-05, -111281895.978322, -61799.5935323178, -84825580.3406894, -2592650.81855619, -25185589.6968963, 0, 57757809.6028587, 7.95677818643901e-05, 166128132.775102, 0.000402100725890302, 170147361.79447, 558226.254013534, 73499789.9283283, 12280977.5598964, 0, -42710510.8381952, -6.40101043949598e-05, -151192479.260948, -0.000329150236504244, -199634112.959667, -0.0012434537310618, -118852571.695549, -2319740.20837944, -30020167.3660475, 0, 18012100.4173854, 2.93952362957556e-05, 76551426.7738106, 0.000159666424965299, 126084702.921784, 0.000593714206328522, 99066552.2966934, 0.00184566591366738, 36024200.8373908, 4503025.10744253, 0, -3302218.40985393, -5.81816674380644e-06, -16511092.0492518, -3.47923228553796e-05, -33022184.0985427, -0.000122256875220398, -33022184.0987677, -0.000431179600008151, -16511092.0499823, -0.000947361769922914, -3302218.41096829}, 
                                                   {0, 0, 1016.40000000156, 0, -31527.3566667147, -8234.85666668864, 0, 395620.894167264, 27553.1788892815, 243821.823056051, 0, -2696956.92801329, -48742.3291696976, -2862190.10255077, -777654.003711372, 0, 11214051.9626324, 48091.618068611, 17834291.9117109, 1309378.18234134, 8501572.86337122, 0, -30032326.3400095, -25097.1142027767, -65921527.665209, -1227585.15144069, -47947072.9192876, -13260359.631306, 0, 52906992.6271957, 5413.10312318013, 151565081.187972, 607620.818322974, 154653256.481637, 11443074.3144529, 66836034.5200735, 0, -60958306.7844643, -8.52020342572795e-05, -218775070.909601, -124050.278787315, -295376117.7954, -5172896.60896974, -183185046.051301, -50674538.7114111, 0, 44211519.2063147, 6.84953774896339e-05, 192674892.343361, 0.000353198060197765, 329539564.454624, 955187.144672089, 275639718.510496, 19649564.0949864, 113257904.14352, 0, -18312302.0910106, -3.14231149903805e-05, -94563527.1911737, -0.000171528610903214, -198133104.591249, -0.000634460283687496, -213143188.273424, -3002016.73819228, -121581677.820059, -33322385.7754267, 0, 3302218.40985439, 6.21254690475518e-06, 19813310.4591017, 3.73966272705723e-05, 49533276.1477937, 0.00013089322687274, 66044368.1972893, 0.00045916976838085, 49533276.1485372, 3302218.41085758, 19813310.4602916}, 
                                                   {0, 0, -907.500000001231, 0, 27650.300595276, 9349.05059525514, 0, -339123.350893328, -40776.6638891923, -270092.514782131, 0, 2252520.51288347, 97642.2524328901, 3068820.6561708, 1115256.14572353, 0, -9115708.95051744, -138682.80556559, -18403971.212442, -2523849.97961003, -11673429.4359848, 0, 23773979.5448267, 116873.815999969, 65306832.9057168, 3381138.99202674, 62439118.5371139, 24170530.3522339, 0, -40852012.1224573, -54131.030607532, -144212959.508701, -2681741.47236482, -190046154.172377, -29377361.3751642, -113434957.71951, 0, 46008476.095656, 10632.8810674621, 200390819.657146, 1166072.61688176, 342602058.765539, 20914876.9320526, 286680193.283601, 118219915.275866, 0, -32690792.6406515, -5.26068008489597e-05, -170432677.434027, -214429.7671451, -362015927.32821, -8050863.06634952, -397767217.551924, -69728661.4667128, -235385403.18454, 0, 13294645.5461688, 2.4109244673244e-05, 81054451.8781465, 0.000132192917753031, 207139154.799866, 1286578.60172534, 285191589.942584, 21014117.1550855, 225151255.219229, 99066552.2980136, 0, -2358727.43561066, -4.76083284404825e-06, -16511092.0492511, -2.88445852433624e-05, -49533276.1477816, -0.000100345457715923, -82555460.2464737, -2358727.43595645, -82555460.2468546, -16511092.0500155, -49533276.1486567}, 
                                                   {0, 0, 605.000000000683, 0, -17878.9503968464, -7896.45039683469, 0, 211218.300231743, 44324.3686840273, 218940.168915554, 0, -1348939.38437677, -139852.052084598, -2359995.57182729, -1150908.23953696, 0, 5260649.16458237, 271465.999831813, 13371789.1881868, 3393422.85870128, 11233096.8824827, 0, -13272405.8356966, -332167.687515002, -44960741.8791914, -6140027.43470471, -55579958.3110425, -29699482.0147276, 0, 22154815.8496002, 250356.016347574, 94600742.2746954, 6982902.94180569, 157105166.619513, 47582867.4534379, 125509560.722382, 0, -24333348.1730801, -106328.8100553, -126010272.75456, -4875175.93958797, -265157469.987311, -47114295.7204078, -287300444.674579, -166165347.85966, 0, 16920457.9678527, 19493.6151987939, 103316160.402643, 1910374.28683659, 264586838.706806, 27973337.7699291, 365427309.983548, 128813809.046728, 289967525.65996, 0, -6754537.6565223, -1.30499712342124e-05, -47603408.2458944, -321644.650381589, -144096803.339004, -9006050.20894031, -243163355.634946, -54036301.2528828, -247666380.739919, -153102853.548926, 0, 1179363.71780556, 2.57504265751658e-06, 9434909.74242911, 1.56406114134968e-05, 33022184.0985157, 1179363.71785724, 66044368.1971215, 9434909.74261346, 82555460.2465752, 33022184.0989057, 66044368.197482}, 
                                                   {0, 0, -302.500000000254, 0, 8384.89186508719, 5057.39186508275, 0, -92270.3205854148, -36639.0100695045, -128909.330654838, 0, 554216.479393295, 151187.368629207, 1259620.32741471, 856591.216651305, 0, -2055189.48365159, -392128.428734607, -6557696.87968248, -3231574.76985428, -7341953.73715388, 0, 4977184.22616467, 663771.510103837, 20572508.414737, 7632270.26657584, 32518191.3973763, 23891365.9652723, 0, -8035638.71300467, -734152.101919996, -40912345.6668846, -11706399.2225903, -84027147.6395563, -47519714.5842004, -87697908.1492013, 0, 8587823.55595057, 512150.434941906, 52039091.7705606, 11660726.1655731, 131890255.948732, 59209197.8597684, 179438727.643006, 139060362.038032, 0, -5838337.74351682, -204682.959298678, -41073047.1638419, -7271118.45857624, -124037873.328742, -45166706.3497874, -208640163.168145, -129768996.189201, -211505724.598412, 0, 2287250.84665348, 35738.2944837062, 18333745.0676752, 2573157.202506, 64328930.062045, 19298679.018693, 129086719.657967, 66044368.1972746, 162108903.756689, 130587728.026355, 0, -393121.239268607, -9.5240900921415e-07, -3538091.15341087, -393121.239273474, -14152364.6136479, -3538091.15342925, -33022184.0985442, -14152364.6137065, -49533276.1478669, -33022184.0986419, -49533276.1479203}, 
                                                   {0, 0, 121.000000000057, 0, -2688.4567460335, -2688.45674603249, 0, 25781.8661309742, 25781.8661309657, 51563.7322619214, 0, -141080.769604425, -141080.769604379, -423242.308812981, -423242.308813086, 0, 489463.582476912, 489463.582476743, 1957854.32990591, 1957854.32990682, 2936781.49485986, 0, -1128191.14725279, -1128191.14725241, -5640955.73625773, -5640955.73626108, -11281911.47252, -11281911.472523, 0, 1753958.1629832, 1753958.16298274, 10523748.9778852, 10523748.9778924, 26309372.4447244, 26309372.4447355, 35079163.2596473, 0, -1819994.79817627, -1819994.79817609, -12739963.5872141, -12739963.587224, -38219890.7616585, -38219890.7616779, -63699817.9361323, -63699817.9361435, 0, 1208604.14056127, 1208604.14056146, 9668833.12447316, 9668833.12448202, 33840915.935669, 33840915.9356874, 67681831.8713813, 67681831.8713996, 84602289.8392548, 0, -464597.828226587, -464597.828226808, -4181380.45403108, -4181380.4540358, -16725521.8161297, -16725521.8161393, -39026217.570997, -39026217.5710081, -58539326.3565212, -58539326.3565291, 0, 78624.2478537422, 78624.2478538082, 786242.478535735, 786242.478536849, 3538091.15341163, 3538091.15341395, 9434909.74243834, 9434909.74244048, 16511092.0492755, 16511092.0492789, 19813310.459136}, 
                                                   {0, 0, 0, 0, 0, 6654.99999999992, 0, 0, -111262.621031746, -111262.621031742, 0, 0, 806058.221527785, 806058.221527744, 1612116.44305553, 0, 0, -3326122.10983252, -3326122.10983225, -9978366.32949707, -9978366.32949728, 0, 0, 8626825.43211844, 8626825.43211728, 34507301.7284704, 34507301.7284724, 51760952.5927076, 0, 0, -14602973.2221656, -14602973.2221624, -73014866.1108166, -73014866.1108223, -146029732.221636, -146029732.221646, 0, 0, 16151346.2420163, 16151346.2420108, 96908077.4520756, 96908077.4520828, 242270193.630187, 242270193.630222, 323026924.840271, 0, 0, -11267309.5684384, -11267309.568433, -78871166.9790448, -78871166.9790478, -236613500.937124, -236613500.937174, -394355834.895229, -394355834.895263, 0, 0, 4503025.10434176, 4503025.10433883, 36024200.83472, 36024200.8347186, 126084702.921509, 126084702.921543, 252169405.843027, 252169405.843071, 315211757.303812, 0, 0, -786242.478535953, -786242.478535307, -7076182.30682019, -7076182.30681908, -28304729.2272771, -28304729.2272857, -66044368.1969801, -66044368.1969977, -99066552.2954794, -99066552.2954875}, 
                                                   {0, 0, 0, 0, 0, 147.888888888888, 0, 0, -4569.23849206352, -147.888888888624, 0, 0, 56846.3615189609, 0, 4421.34960316788, 0, 0, -380901.947332469, 0, -52425.0119157223, 3.63888584696828e-08, 0, 0, 1534020.09508111, 0, 328476.935416385, -8.62505552566415e-08, -3.21558634005773e-07, 0, 0, -3885029.17216463, 0, -1205543.15966366, 1.08281111418893e-07, 1.32577288916435e-06, 7.01340501998955e-07, 0, 0, 6227323.97349586, 0, 2679486.01249904, -7.14309633255962e-08, -2.91368023467178e-06, -8.29239677870394e-07, -2.43573646499152e-06, 0, 0, -6120995.16347604, 0, -3547837.96099465, 1.47055950700372e-08, 3.53129401889502e-06, 5.45763948710679e-07, 4.16104449218092e-06, 2.29949165845837e-06, 0, 0, 3359399.68101605, 0, 2573157.20247991, 0, -2.19753472173563e-06, -1.73710007457926e-07, -3.50286373197757e-06, -1.12153823385417e-06, -2.65087770370139e-06, 0, 0, -786242.47853567, 0, -786242.478535584, 0, 5.45595814952284e-07, 1.49428676063214e-08, 1.14542322732515e-06, 2.25506999792231e-07, 1.13596016231659e-06, 6.49251889787653e-07}, 
                                                   {0, 0, 0, 0, 0, 147.888888888867, 0, 0, -147.888888888858, -4569.23849206287, 0, 0, 0, 56846.3615189521, 4421.34960317523, 0, 0, 0, -380901.947332402, 0, -52425.0119158041, 0, 0, -2.13530267602642e-08, 1534020.09508082, -1.03991806360133e-07, 328476.935416789, 5.70287282748533e-08, 0, 0, 3.53225699318379e-08, -3885029.17216383, 3.22205345460011e-07, -1205543.15966469, 3.14215548944548e-07, -5.95675421208882e-07, 0, 0, -3.35888027557314e-08, 6227323.97349454, -5.02382723983239e-07, 2679486.01250044, -1.07256422282152e-06, 1.96048200804084e-06, -1.36956561827668e-07, 0, 0, 1.72033378611678e-08, -6120995.16347473, 4.47177212514485e-07, -3547837.96099554, 1.31948627064705e-06, -3.0765246624923e-06, 1.59413614812437e-06, -9.80241004020419e-07, 0, 0, 0, 3359399.68101534, -2.16521438164446e-07, 2573157.20248004, -7.9036635221861e-07, 2.27465773245447e-06, -1.43909924925985e-06, 1.80357335622315e-06, -1.02679588066379e-06, 0, 0, 0, -786242.47853551, 4.35747090198418e-08, -786242.478535517, 1.97446240351524e-07, -6.25382032240672e-07, 4.10543622136111e-07, -9.01378476669666e-07, 6.24279390540874e-07, 1.071479004354e-07}, 
                                                   {0, 0, 0, 0, 0, -9982.50000000003, 0, 0, 240098.931547622, 130291.431547615, 0, 0, -2164557.83030758, -731352.083283708, -2895909.91359119, 0, 0, 10352431.7754964, 2307558.85937492, 23012422.4103663, 14967549.4942457, 0, 0, -29862336.4502622, -4479188.99713515, -94066198.3479122, -43299903.4416667, -103024576.342185, 0, 0, 54751845.8122448, 5480766.84374927, 224488150.092696, 76674913.1872405, 350434142.248419, 251891984.311459, 0, 0, -64419309.5505301, -4130874.26926963, -326227422.021856, -85073680.8968749, -664847466.851509, -363405290.445329, -685501838.197898, 0, 0, 47194042.3273161, 1754425.36532628, 284918679.329152, 57720594.5192676, 718437187.101489, 309480634.443776, 970197227.025982, 742999142.216157, 0, 0, -19620323.6689166, -321644.650309527, -137663910.332684, -21871836.2210774, -414278309.599257, -144096803.338906, -693465866.068336, -423284359.80801, -697968891.172728, 0, 0, 3538091.15341147, -7.84939165849297e-08, 28304729.2272815, 3538091.15340875, 99066552.2954737, 28304729.227286, 198133104.590946, 99066552.2954958, 247666380.738701, 198133104.590976}, 
                                                   {0, 0, 0, 0, 0, 13310.0000000001, 0, 0, -356734.408730165, -137119.408730152, 0, 0, 3665782.98108475, 598057.737037004, 3458585.71812153, 0, 0, -19595763.258268, -1432086.3689814, -32327888.6573396, -14164211.7680549, 0, 0, 61845932.6604206, 2034014.48148127, 151389319.371982, 31765483.0141199, 119274855.244206, 0, 0, -121916613.568761, -1714149.30092535, -402739378.252523, -42131985.4490747, -478432602.646028, -238027674.110413, 0, 0, 152135261.376409, 793921.781480498, 638412352.533761, 33046994.1541665, 1033090718.15268, 276384020.178254, 790944574.800885, 0, 0, -116961691.021846, -155948.921361536, -598220062.34627, -14191351.8439782, -1236830895.32533, -185579216.421322, -1304512727.19665, -720484016.694446, 0, 0, 50605424.9821211, -3.94022489275535e-07, 306205707.09513, 2573157.20247622, 774520317.946438, 66044368.1970072, 1050705857.67928, 342229907.929886, 810544518.781217, 0, 0, -9434909.7424299, 5.85879043038356e-08, -66044368.1969903, 1.42355374451414e-06, -198133104.590949, -9434909.74243098, -330221840.984908, -66044368.1970012, -330221840.984928, -198133104.590975}, 
                                                   {0, 0, 0, 0, 0, -13975.5000000001, 0, 0, 395068.529166674, 113229.279166657, 0, 0, -4353348.83958346, -378856.209722171, -3041169.54930531, 0, 0, 25111448.4512743, 670207.026041547, 30991897.6494188, 9650887.97418868, 0, 0, -85136731.2449707, -661259.748263637, -159993795.249814, -16160880.6848941, -90356684.9414823, 0, 0, 178818476.250884, 345085.3197911, 466438068.458764, 15060831.522279, 410791369.047632, 137887523.042232, 0, 0, -235720338.933014, -74430.1670129998, -801315178.07148, -7405801.6178804, -996806011.733373, -115924985.124132, -539804786.268159, 0, 0, 189945786.219473, -7.92042463406845e-07, 804540485.30876, 1501008.36810992, 1320205087.40868, 50761373.9034849, 1034331220.93619, 377981198.152171, 0, 0, -85557476.9824811, 3.50818811584478e-07, -436793435.120995, 3.41629219602404e-06, -900605020.867937, -9006050.20869023, -945635271.9113, -130587728.025875, -517847886.999069, 0, 0, 16511092.0492514, -5.48035794794686e-08, 99066552.2954854, -1.16011722266909e-06, 247666380.738685, 2.20687335233327e-06, 330221840.984895, 16511092.0492535, 247666380.738677, 99066552.2954813}, 
                                                   {0, 0, 0, 0, 0, 11180.4000000001, 0, 0, -326303.523333339, -70086.0233333225, 0, 0, 3753606.70972234, 174593.924999943, 1955325.88472195, 0, 0, -22784913.9069455, -216076.758333195, -21017714.7048585, -4649341.05624868, 0, 0, 81582229.4260469, 132867.074999748, 115872391.906932, 5483227.3451364, 46461032.6010296, 0, 0, -180788169.125223, -32478.6183328841, -363136665.206148, -3207263.56041409, -229806523.844736, -50632812.7058829, 0, 0, 250531942.168776, -6.22896581969302e-07, 670020363.458981, 744301.670136395, 614123308.031464, 27167010.960056, 221057596.031168, 0, 0, -211232813.985444, 5.33905201724132e-07, -718846553.020103, 2.8517392437335e-06, -894873898.007813, -5731122.86006775, -483870515.757126, -102341479.644045, 0, 0, 99066552.2955013, -2.39067223499988e-07, 414278309.599294, -2.21700681833653e-06, 666447715.442247, 2.20366100985295e-06, 504338811.685954, 18012100.4173484, 171114953.964851, 0, 0, -19813310.4591008, 3.88414279029881e-08, -99066552.2954853, 7.08141392476516e-07, -198133104.590943, -7.71266165029782e-07, -198133104.590916, 8.8484713146386e-08, -99066552.2954444, -19813310.4590848}, 
                                                   {0, 0, 0, 0, 0, -6655.00000000002, 0, 0, 198411.430555559, 31260.0138888808, 0, 0, -2349802.55023155, -54802.0763888451, -894751.501620168, 0, 0, 14785339.2474544, 42499.5694443361, 9972505.32997494, 1503944.19363323, 0, 0, -55164133.5762187, -12302.5069442641, -57589675.3411947, -1116657.54698857, -15709378.6800255, 0, 0, 127759381.67935, -2.63806240698284e-07, 190705876.106806, 308997.966085424, 82447198.1850867, 10878081.6820429, 0, 0, -185107825.36356, 3.17494653216915e-07, -373912349.022074, 2.16528973207294e-06, -237246157.356676, -2791131.26300479, -51232764.9611612, 0, 0, 162927635.593421, -2.59935018883057e-07, 426832197.768957, -1.6523275199296e-06, 376480189.783965, -7.71170678259799e-06, 124174328.634715, 11598701.0262805, 0, 0, -79553443.5100219, 1.1881640442912e-07, -261175456.051728, 1.00289740602265e-06, -306205707.095061, 1.52115148813371e-06, -147098820.075004, 1.89355584736453e-05, -22515125.521635, 0, 0, 16511092.0492501, -2.00973300119056e-08, 66044368.1969901, -3.00704040293387e-07, 99066552.2954663, 3.3090225224849e-08, 66044368.1969517, -2.83803016506882e-06, 16511092.0492157, -1.29603462499687e-05}, 
                                                   {0, 0, 0, 0, 0, 2852.14285714285, 0, 0, -86340.7023809534, -9475.45238094829, 0, 0, 1044234.46547622, 10457.8571428345, 276279.155952278, 0, 0, -6747025.00548973, -3834.5476189911, -3161243.66322655, -293342.892856604, 0, 0, 25981664.4159738, -8.74824457882294e-08, 18892959.9145789, 103341.058332128, 3164204.78610631, 0, 0, -62372479.9576437, 1.06235474984192e-07, -65277511.9307758, 1.48000731145875e-06, -17348995.2930351, -1046533.25739779, 0, 0, 93952136.5335401, -1.07340380167936e-07, 134569741.961095, -1.16027986783313e-06, 52845418.5798145, -1.01566271906917e-05, 5210111.69093725, 0, 0, -86083804.5920722, 8.26873330529945e-08, -162498776.059644, 6.39653019200321e-07, -90060502.0867474, 5.94582474733371e-06, -13645530.61915, 2.96854835770468e-05, 0, 0, 43743672.4421675, -3.93656704175505e-08, 105499445.301682, -2.77110699434415e-07, 79767873.2768479, -1.69207241049919e-06, 18012100.4173056, -1.28575619204596e-05, -3.65344065309633e-05, 0, 0, -9434909.74242838, 0, -28304729.2272814, 7.5951114262127e-08, -28304729.2272724, 1.20259701734328e-07, -9434909.74240843, 2.33431392774386e-06, 1.67147947050196e-05, 7.95756707927976e-06}, 
                                                   {0, 0, 0, 0, 0, -831.874999999995, 0, 0, 25473.2008928574, 1746.93749999864, 0, 0, -313007.29548612, -915.062499992449, -51655.2781249657, 0, 0, 2063925.28435029, -1.88558745576966e-08, 602958.6471723, 26098.8897319601, 0, 0, -8147796.56796925, 2.69788655973138e-08, -3699825.18220341, 4.23568788477019e-07, -288493.787845604, 0, 0, 20141818.1507827, -2.62201411210636e-08, 13218121.0237811, -5.29664509321659e-07, 1635772.07959392, -3.40569418868385e-06, 0, 0, -31372315.3963568, 1.95188231631775e-08, -28395108.7157936, 3.84252793526241e-07, -5210111.69095709, 3.85296750517626e-06, 1.20505200601874e-05, 0, 0, 29825231.2105682, -1.31272802873107e-08, 36024200.8347184, -1.47626532806957e-07, 9415416.12723836, -2.43826211804363e-06, -2.1055558981654e-05, -1.08845446094425e-05, 0, 0, -15760587.8651925, 0, -24766638.0738702, 2.75325698501986e-08, -9006050.20866887, 7.56054723852479e-07, 1.81002029373272e-05, 5.06788442588423e-06, 1.28816630667229e-05, 0, 0, 3538091.15341057, 0, 7076182.30682034, 0, 3538091.15340729, -6.41937773830915e-08, -6.01727471686829e-06, -9.84767503249454e-07, -5.68314861970591e-06, -3.00305242796991e-06}, 
                                                   {0, 0, 0, 0, 0, 83.1875, 0, 0, -2455.81383928575, -998.249999999903, 0, 0, 28599.3343253976, 915.062499999456, 29386.5785714263, 0, 0, -174933.260677091, 0, -340819.385565454, -26098.889732133, 0, 0, 622353.070052121, 0, 2072972.42013878, -1.64116328727153e-08, 288493.787847133, 0, 0, -1329593.43802095, 1.17429991712628e-08, -7319530.28793371, 2.13610782346195e-08, -1635772.07960023, 8.88095777776749e-08, 0, 0, 1674678.75781271, -1.71641932057716e-08, 15481474.7388884, -5.10756507509337e-08, 5210111.690971, 4.47747293498839e-08, -5.13716031197709e-07, 0, 0, -1140376.48746302, 1.44662755994805e-08, -19240198.1730899, 7.991310636767e-08, -9415416.12725524, -1.19192429836525e-07, 1.292691306529e-06, 7.23651276059958e-08, 0, 0, 321644.650310144, 0, 12865786.0124006, -5.96693079932272e-08, 9006050.20867939, 5.9166606640918e-08, -1.40637043013961e-06, 1.7952852567688e-07, -3.95287349409796e-07, 0, 0, -2.93063008842808e-08, 0, -3538091.1534102, 1.71271503331204e-08, -3538091.1534099, 0, 5.40578482094987e-07, -7.09808342678863e-08, 3.10749987279045e-07, -3.04544534384072e-08}, 
                                                   {0, 0, 0, 0, 0, 63.3809523809556, 0, 0, -1771.49761904771, -1109.16666666668, 0, 0, 19006.1870370381, 4880.33333333355, 30937.8273809527, 0, 0, -103415.618981488, -3834.54761904848, -330900.156481482, -135359.530952386, 0, 0, 311663.509259285, 0, 1792475.26180554, 103341.058333359, 1435292.47685189, 0, 0, -523266.628703772, 0, -5367993.86342585, -4.41609069258399e-08, -7677584.50046302, -1046533.2574076, 0, 0, 453669.589418105, 0, 8931620.04166649, 3.29350022703052e-08, 22577150.6608796, 3.82654806414474e-07, 5210111.69097265, 0, 0, -155948.921362553, 0, -7641497.14675907, 1.07327927915985e-08, -36570022.0594904, -3.82800913633418e-07, -13645530.6192133, -7.78361857721283e-07, 0, 0, 6.76342187273124e-08, 0, 2573157.20248005, -2.85624393533649e-08, 30020167.3622682, 1.86213650544195e-07, 18012100.4173611, 5.46928022913903e-07, 6.0605274417993e-07, 0, 0, -1.61917683703183e-08, 0, 2.20071306051917e-08, 1.11286256706572e-08, -9434909.74242715, -3.10477840256112e-08, -9434909.74242717, -1.5144013243551e-07, -1.73137054448702e-07, -2.24932063564369e-07}, 
                                                   {0, 0, 0, 0, 0, 55.4583333333354, 0, 0, -1448.38680555561, -1173.86805555557, 0, 0, 13975.0378472228, 7828.86805555573, 30602.0623842597, 0, 0, -64867.7638888919, -19012.9652777784, -294412.039293983, -203345.527662043, 0, 0, 153781.336805567, 12302.5069444447, 1360452.22627314, 489844.818171321, 1944718.78524309, 0, 0, -175925.849305586, 0, -3202752.64120365, -308997.966088018, -8903426.79658567, -4622564.46348398, 0, 0, 74430.1670139416, 0, 3622268.12800914, 3.52406063186543e-08, 20654371.346354, 2791131.26302126, 20716396.4855327, 0, 0, -5.51100236825257e-08, 0, -1501008.36811329, 1.72392152139348e-08, -22788036.1340853, -4.70302639134779e-07, -46394804.1053242, -11598701.0263317, 0, 0, 3.15754438444101e-08, 0, -6.6139190031059e-08, -3.58924672778782e-08, 9006050.20868027, 2.61787361224289e-07, 48032267.7796295, 5.95620401388024e-07, 22515125.5217018, 0, 0, 0, 0, 1.24697836500048e-08, 1.39770258104845e-08, 8.72701553261142e-08, -5.66216457811959e-08, -16511092.0492476, -1.94836226396931e-07, -16511092.0492478, -1.87535383653624e-07}, 
                                                   {0, 0, 0, 0, 0, 53.2400000000007, 0, 0, -1273.32333333336, -1273.32333333334, 0, 0, 10614.7250000002, 10614.7250000001, 30400.4097222225, 0, 0, -38920.6583333341, -38920.6583333337, -252648.756250001, -252648.756250004, 0, 0, 62004.6350000028, 62004.6349999999, 921457.770138877, 921457.770138913, 2088350.55381947, 0, 0, -32478.6183333427, -32478.6183333302, -1453418.17041662, -1453418.17041672, -7544512.38368053, -7544512.38368073, 0, 0, 1.79680310358542e-08, 0, 744301.670138804, 744301.670138947, 11685536.2211803, 11685536.2211811, 26794860.1250002, 0, 0, -1.95533301157593e-08, 0, 6.31192582528807e-08, 0, -5731122.86006906, -5731122.86007013, -40117860.0204859, -40117860.0204867, 0, 0, 1.10442295114309e-08, 0, -2.21514585261913e-08, -4.84387540195027e-08, -2.80016327217398e-07, 4.44045991976476e-07, 18012100.4173607, 18012100.4173618, 55837511.2938196, 0, 0, 0, 0, 0, 2.26322149547182e-08, 6.8275677810525e-08, -1.06551438941617e-07, 1.84194157095041e-07, -2.38355606603036e-07, -19813310.4590972, -19813310.4590974}, 
                                                   {0, 0, 0, 0, 0, 55.4583333333317, 0, 0, -1173.86805555551, -1448.38680555553, 0, 0, 7828.86805555508, 13975.0378472222, 30602.0623842584, 0, 0, -19012.9652777751, -64867.7638888895, -203345.527662028, -294412.039293978, 0, 0, 12302.5069444378, 153781.336805559, 489844.818171245, 1360452.22627315, 1944718.78524302, 0, 0, 0, -175925.849305562, -308997.966087831, -3202752.64120374, -4622564.46348352, -8903426.79658565, 0, 0, 0, 74430.1670138981, -1.79618198848484e-07, 3622268.1280093, 2791131.26302018, 20654371.3463545, 20716396.485532, 0, 0, 0, 0, 1.33079479532716e-07, -1501008.36811342, 7.88204568100973e-07, -22788036.1340862, -11598701.0263299, -46394804.1053239, 0, 0, 0, 0, -4.54390927749192e-08, -4.88522172753118e-08, -4.4508837568428e-07, 9006050.20868104, -1.00247796697391e-06, 48032267.7796299, 22515125.5217008, 0, 0, 0, 0, 0, 2.57516848215391e-08, 9.40168905254022e-08, -1.53077123599827e-07, 3.35866421811422e-07, -16511092.0492478, 3.12031855559218e-07, -16511092.0492476}, 
                                                   {0, 0, 0, 0, 0, 63.3809523809485, 0, 0, -1109.16666666656, -1771.49761904755, 0, 0, 4880.33333333229, 19006.1870370367, 30937.8273809504, 0, 0, -3834.54761904239, -103415.618981482, -135359.530952363, -330900.156481471, 0, 0, -1.40245221826106e-08, 311663.509259263, 103341.058333246, 1792475.26180553, 1435292.47685176, 0, 0, 2.22274056760994e-08, -523266.628703715, 2.10458286384554e-07, -5367993.86342591, -1046533.25740695, -7677584.50046281, 0, 0, -2.2067271170152e-08, 453669.589418007, -2.84975261605313e-07, 8931620.04166666, -9.72288337930114e-07, 22577150.6608797, 5210111.69097125, 0, 0, 1.35850678051565e-08, -155948.921362452, 2.18931479027655e-07, -7641497.14675921, 1.06397047004454e-06, -36570022.0594912, 1.74909583029466e-06, -13645530.6192121, 0, 0, 0, 1.22010536972168e-08, -9.10718386955967e-08, 2573157.20248007, -5.89707282223537e-07, 30020167.3622691, -1.38882947854986e-06, 18012100.4173609, -1.24827713504403e-06, 0, 0, 0, 0, 1.61252507891488e-08, 4.05764497335453e-08, 1.24193003546032e-07, -9434909.74242744, 4.3135029558683e-07, -9434909.7424273, 4.97587606894171e-07, 3.44649424688603e-07}, 
                                                   {0, 0, 0, 0, 0, 83.1874999999919, 0, 0, -998.249999999891, -2455.81383928552, 0, 0, 915.06249999899, 28599.3343253948, 29386.5785714267, 0, 0, 0, -174933.260677072, -26098.8897321272, -340819.385565464, 0, 0, -1.47432273213033e-08, 622353.070052042, -7.94956739671902e-08, 2072972.42013885, 288493.787847142, 0, 0, 2.52929857468161e-08, -1329593.43802074, 1.85874407369476e-07, -7319530.28793394, 4.25002265014434e-07, -1635772.07960055, 0, 0, -2.87507451700225e-08, 1674678.75781235, -2.43709833210177e-07, 15481474.7388887, -8.43339838297641e-07, 5210111.69097231, -1.05148328774982e-06, 0, 0, 2.11849134363716e-08, -1140376.48746266, 1.95390818158777e-07, -19240198.17309, 8.00916512085404e-07, -9415416.12725751, 1.82920154592461e-06, 1.11659219927797e-06, 0, 0, 0, 321644.650309956, -9.13205320691368e-08, 12865786.0124005, -3.82803231312383e-07, 9006050.20868116, -1.23117363019561e-06, -3.45538632834413e-07, -1.69539504270535e-06, 0, 0, 0, 1.20489326290581e-08, 1.95970526715366e-08, -3538091.15341012, 7.41309073915517e-08, -3538091.15341042, 3.07729323974565e-07, -9.09986755609917e-08, 6.07768619028482e-07, 5.53169413691887e-07}, 
                                                   {0, 0, 0, 0, 0, -831.874999999885, 0, 0, 1746.93749999976, 25473.2008928539, 0, 0, -915.062499994911, -313007.295486073, -51655.2781250014, 0, 0, -3.78491736886062e-08, 2063925.28434995, 26098.889732091, 602958.6471727, 0, 0, 1.14929740399038e-07, -8147796.56796772, 5.61437656167721e-07, -3699825.18220541, -288493.787847438, 0, 0, -1.91401207000072e-07, 20141818.1507786, -1.6924224738999e-06, 13218121.0237862, -1.8430998504299e-06, 1635772.07960347, 0, 0, 1.87092036996438e-07, -31372315.39635, 2.60073075610168e-06, -28395108.7158005, 5.81406403253714e-06, -5210111.6909818, 1.4864417647855e-06, 0, 0, -1.02525808646559e-07, 29825231.2105615, -2.29844557561035e-06, 36024200.8347227, -6.94705074792112e-06, 9415416.12727232, -9.13063335096888e-06, 3.84201285482013e-06, 0, 0, 2.54876827630608e-08, -15760587.8651889, 1.11100505161772e-06, -24766638.0738706, 4.07880458870267e-06, -9006050.20869208, 7.87481019567056e-06, -8.45520036287994e-06, 6.32518126136648e-06, 0, 0, 0, 3538091.15340975, -2.2482861577507e-07, 7076182.30681991, -1.00117191260278e-06, 3538091.15341341, -2.20501382736374e-06, 4.44914837465754e-06, -3.48667966143983e-06, -9.83361245074657e-07}, 
                                                   {0, 0, 0, 0, 0, 2852.1428571425, 0, 0, -9475.45238095134, -86340.7023809424, 0, 0, 10457.8571428395, 1044234.46547607, 276279.155952379, 0, 0, -3834.54761892266, -6747025.00548865, -293342.892856948, -3161243.6632277, 0, 0, -3.75608447295304e-07, 25981664.4159691, 103341.058331483, 18892959.9145847, 3164204.7861115, 0, 0, 6.284696662462e-07, -62372479.957631, 5.4207319592964e-06, -65277511.930791, -1046533.25740085, -17348995.2930632, 0, 0, -6.28163549914885e-07, 93952136.5335192, -8.20507253376717e-06, 134569741.961116, -1.92624988138462e-05, 52845418.5798892, 5210111.69096496, 0, 0, 3.62454016819519e-07, -86083804.5920515, 7.1848247450847e-06, -162498776.059656, 2.24529282352746e-05, -90060502.0868518, 3.15706736301552e-05, -13645530.6192211, 0, 0, -1.03002930930854e-07, 43743672.4421564, -3.457274096021e-06, 105499445.301683, -1.29526840653877e-05, 79767873.2769202, -2.63282731203546e-05, 18012100.4173847, -2.28386952942432e-05, 0, 0, 0, -9434909.74242587, 7.01323059557002e-07, -28304729.2272798, 3.11645389181703e-06, -28304729.2272916, 7.32081000698518e-06, -9434909.74244043, 1.16924958486632e-05, 4.33560690810539e-06}, 
                                                   {0, 0, 0, 0, 0, -6654.99999999926, 0, 0, 31260.0138888861, 198411.430555535, 0, 0, -54802.0763888479, -2349802.55023125, -894751.50162035, 0, 0, 42499.5694441673, 14785339.2474522, 1503944.19363377, 9972505.32997711, 0, 0, -12302.5069436199, -55164133.5762091, -1116657.54698663, -57589675.341206, -15709378.6800349, 0, 0, -1.38279410253197e-06, 127759381.679325, 308997.966076222, 190705876.106835, 10878081.6820446, 82447198.1851411, 0, 0, 1.40373999329893e-06, -185107825.363517, 1.75180548263981e-05, -373912349.022113, -2791131.26297764, -237246157.356826, -51232764.9612061, 0, 0, -8.39726708426645e-07, 162927635.593379, -1.51735569965084e-05, 426832197.768979, -4.94693226296028e-05, 376480189.784179, 11598701.0262582, 124174328.634847, 0, 0, 2.59712056642271e-07, -79553443.5099992, 7.24067451732393e-06, -261175456.051727, 2.81575084142538e-05, -306205707.095211, 5.95840917521263e-05, -147098820.07516, -22515125.5216478, 0, 0, -2.90428518567923e-08, 16511092.049245, -1.46616574366797e-06, 66044368.1969862, -6.63425274209306e-06, 99066552.2955066, -1.66617765766743e-05, 66044368.1970169, -2.60815325427838e-05, 16511092.0492364}, 
                                                   {0, 0, 0, 0, 0, 11180.3999999989, 0, 0, -70086.0233333282, -326303.523333304, 0, 0, 174593.924999934, 3753606.70972189, 1955325.88472217, 0, 0, -216076.758332904, -22784913.9069423, -4649341.05624917, -21017714.7048613, 0, 0, 132867.074998737, 81582229.4260331, 5483227.34513246, 115872391.906947, 46461032.6010411, 0, 0, -32478.6183312191, -180788169.125187, -3207263.5603987, -363136665.206188, -50632812.7058771, -229806523.844809, 0, 0, -2.1564956007094e-06, 250531942.168716, 744301.670112389, 670020363.459034, 27167010.9600011, 614123308.031674, 221057596.031213, 0, 0, 1.3105184541423e-06, -211232813.985384, 2.2681686596136e-05, -718846553.020131, -5731122.85999173, -894873898.00812, -102341479.643981, -483870515.757298, 0, 0, -4.22312931343252e-07, 99066552.2954693, -1.06868433591308e-05, 414278309.599289, -4.39486702743151e-05, 666447715.442466, 18012100.4172659, 504338811.68617, 171114953.964846, 0, 0, 5.22785896405475e-08, -19813310.4590936, 2.1475620312411e-06, -99066552.2954788, 1.01578273119564e-05, -198133104.591002, 2.71446601741844e-05, -198133104.591009, -19813310.4590569, -99066552.2954676}, 
                                                   {0, 0, 0, 0, 0, -13975.4999999989, 0, 0, 113229.279166661, 395068.529166637, 0, 0, -378856.20972215, -4353348.839583, -3041169.54930548, 0, 0, 670207.026041205, 25111448.451271, 9650887.97418886, 30991897.6494213, 0, 0, -661259.748262544, -85136731.2449569, -16160880.6848887, -159993795.249828, -90356684.9414918, 0, 0, 345085.319789448, 178818476.250847, 15060831.5222604, 466438068.458803, 137887523.042216, 410791369.047702, 0, 0, -74430.1670116745, -235720338.932953, -7405801.61785302, -801315178.071531, -115924985.124055, -996806011.733586, -539804786.268186, 0, 0, -1.30342628701591e-06, 189945786.219413, 1501008.36808894, 804540485.308784, 50761373.9033838, 1320205087.40899, 377981198.152069, 1034331220.93635, 0, 0, 4.01239374307117e-07, -85557476.9824491, 1.13442327665486e-05, -436793435.120987, -9006050.20863029, -900605020.868166, -130587728.025759, -945635271.91152, -517847886.999041, 0, 0, -4.55011022768585e-08, 16511092.0492442, -2.24370031560491e-06, 99066552.2954777, -1.14767445357691e-05, 247666380.738747, 16511092.0492156, 330221840.984992, 99066552.2954426, 247666380.738696}, 
                                                   {0, 0, 0, 0, 0, 13309.9999999992, 0, 0, -137119.408730154, -356734.408730138, 0, 0, 598057.737036986, 3665782.98108443, 3458585.71812163, 0, 0, -1432086.36898116, -19595763.2582657, -14164211.7680548, -32327888.6573412, 0, 0, 2034014.48148056, 61845932.6604109, 31765483.0141148, 151389319.371992, 119274855.244212, 0, 0, -1714149.30092451, -121916613.568736, -42131985.4490586, -402739378.25255, -238027674.110394, -478432602.646078, 0, 0, 793921.781480265, 152135261.376366, 33046994.1541437, 638412352.533797, 276384020.178179, 1033090718.15284, 790944574.800895, 0, 0, -155948.921361922, -116961691.021804, -14191351.8439622, -598220062.346284, -185579216.421223, -1236830895.32557, -720484016.694345, -1304512727.19677, 0, 0, -4.06453073216884e-08, 50605424.9820988, 2573157.20247138, 306205707.095122, 66044368.1969482, 774520317.946612, 342229907.929773, 1050705857.67945, 810544518.781188, 0, 0, -2.61823323484918e-08, -9434909.74242493, 1.68571468881723e-06, -66044368.1969839, -9434909.74241749, -198133104.590996, -66044368.1969632, -330221840.984984, -198133104.59094, -330221840.984943}, 
                                                   {0, 0, 0, 0, 0, -9982.49999999965, 0, 0, 130291.431547617, 240098.931547609, 0, 0, -731352.083283712, -2164557.83030743, -2895909.91359123, 0, 0, 2307558.85937488, 10352431.7754954, 14967549.4942456, 23012422.410367, 0, 0, -4479188.99713516, -29862336.4502577, -43299903.4416635, -94066198.3479174, -103024576.342187, 0, 0, 5480766.84374989, 54751845.8122328, 76674913.1872304, 224488150.09271, 251891984.311446, 350434142.248446, 0, 0, -4130874.26927129, -64419309.5505102, -85073680.8968611, -326227422.021874, -363405290.445276, -664847466.851599, -685501838.197904, 0, 0, 1754425.36532824, 47194042.3272966, 57720594.5192592, 284918679.329159, 309480634.443703, 718437187.101622, 742999142.216093, 970197227.026058, 0, 0, -321644.650310627, -19620323.6689062, -21871836.2210763, -137663910.332679, -144096803.33886, -414278309.59935, -423284359.807934, -693465866.068439, -697968891.172719, 0, 0, 1.60104744648143e-07, 3538091.15340917, 3538091.15340933, 28304729.2272779, 28304729.2272749, 99066552.2954985, 99066552.2954682, 198133104.59099, 198133104.590957, 247666380.738714}, 
                                                   {0, 0, 0, 0, 0, 9982.49999999992, 0, 0, -212647.056547619, -212647.056547615, 0, 0, 1579778.42480159, 1579778.42480156, 4367439.34960312, 0, 0, -6008041.10729171, -6008041.10729153, -30467709.6641363, -30467709.6641368, 0, 0, 13340223.4052088, 13340223.4052078, 107634633.256941, 107634633.256945, 188588819.703471, 0, 0, -18066231.4479184, -18066231.4479152, -220292995.224641, -220292995.224655, -570547828.434359, -570547828.434383, 0, 0, 14737173.0687536, 14737173.0687475, 273009852.606926, 273009852.606945, 959404852.808979, 959404852.809064, 1402264346.54166, 0, 0, -6666816.38824805, -6666816.38824181, -202226763.776721, -202226763.77673, -917798389.448197, -917798389.448319, -1788929064.17876, -1788929064.17885, 0, 0, 1286578.60124242, 1286578.60123908, 82341030.4793592, 82341030.4793573, 468314610.851345, 468314610.851428, 1152774426.71104, 1152774426.71115, 1531028535.47568, 0, 0, -5.61935021584638e-07, 1.55055871778042e-07, -14152364.61364, -14152364.6136378, -99066552.2954753, -99066552.2954964, -297199656.886432, -297199656.886477, -495332761.477416, -495332761.477436}, 
                                                   {0, 0, 0, 0, 0, -475.357142857148, 0, 0, 13809.1250000003, 6227.17857142817, 0, 0, -157161.258134926, -11503.6428571408, -179848.998214276, 0, 0, 932418.185416708, 5751.82142856554, 2029402.37281737, 321526.817857108, 0, 0, -3190655.17604187, 1.93608270316881e-08, -11898369.5913189, -155011.587499969, -3449622.94722189, 0, 0, 6495723.66666727, -5.11662995090373e-08, 40070495.3687485, -1.95419823013897e-08, 18769934.845137, 1569799.88611101, 0, 0, -7719471.60744155, 7.99092225212433e-08, -79863569.2059004, 1.5829248723607e-07, -56566926.9305501, -7.95873965767044e-07, -7815167.53645671, 0, 0, 4912391.0229178, -6.89856926813939e-08, 92282774.2162183, -3.50911049727286e-07, 94972893.1097143, 1.13819893931527e-06, 20468295.9288143, 8.73652999802425e-07, 0, 0, -1286578.60124072, 2.99718102660244e-08, -56609458.4545627, 2.97196423468359e-07, -82341030.4793595, -5.66328591570123e-07, -27018150.6260355, -1.64390849623407e-06, 8.38020479834966e-07, 0, 0, 1.51355868270027e-07, 0, 14152364.6136408, -9.04806153323311e-08, 28304729.2272802, 8.21054928035999e-08, 14152364.6136384, 5.5223114165121e-07, -1.11640865681129e-06, 4.76955620273842e-07}, 
                                                   {0, 0, 0, 0, 0, -475.357142857102, 0, 0, 6227.17857142795, 13809.124999999, 0, 0, -11503.6428571371, -157161.258134912, -179848.998214275, 0, 0, 5751.82142854107, 932418.185416617, 321526.817857051, 2029402.37281739, 0, 0, 8.3337126176523e-08, -3190655.17604149, -155011.587499536, -11898369.5913192, -3449622.94722176, 0, 0, -1.40589095132596e-07, 6495723.66666625, -1.09604497808714e-06, 40070495.3687496, 1569799.88610866, 18769934.8451381, 0, 0, 1.55432387684655e-07, -7719471.60743985, 1.45336284457157e-06, -79863569.205902, 4.97962216718363e-06, -56566926.9305561, -7815167.53645252, 0, 0, -1.10746810637202e-07, 4912391.02291611, -1.15394120376283e-06, 92282774.2162189, -4.94272674130819e-06, 94972893.1097255, -1.02510468720345e-05, 20468295.9288136, 0, 0, 4.60281309773603e-08, -1286578.60123982, 5.24218278191935e-07, -56609458.4545622, 2.48413961658685e-06, -82341030.4793685, 7.22596792209169e-06, -27018150.62604, 8.91816563290669e-06, 0, 0, 0, -4.72809139126359e-08, -1.07768056758137e-07, 14152364.6136404, -4.95794693966303e-07, 28304729.2272829, -1.93384668732071e-06, 14152364.6136414, -3.27818206569387e-06, -2.8087476986946e-06}, 
                                                   {0, 0, 0, 0, 0, -9317.00000000008, 0, 0, 229216.686111115, 177973.186111109, 0, 0, -2061771.37731486, -1082918.8537037, -4272047.23101849, 0, 0, 9181137.25069473, 3225083.55532406, 36921650.0761569, 24765132.8807871, 0, 0, -23093650.9107652, -5355691.35648134, -155433768.529973, -69721792.5644691, -196706218.827199, 0, 0, 34535597.4944485, 5079295.03379567, 365370923.492349, 108916144.396995, 725892608.830775, 498894132.196077, 0, 0, -30516368.4757017, -2580245.7898136, -505852225.08204, -96759217.118058, -1423476944.14058, -689099296.270295, -1543061412.47628, 0, 0, 14737173.0687577, 545821.224767371, 410184650.413525, 45848982.8805522, 1537032568.94808, 529173677.413132, 2341573054.2569, 1683858478.41093, 0, 0, -3002016.73623125, 4.89905274356039e-07, -180121004.173604, -9006050.20867504, -864580820.033282, -210141171.535913, -1726159623.33038, -918617121.285473, -1756179790.69272, 0, 0, 1.03956110802683e-06, -6.28267375559621e-08, 33022184.0984945, -2.18364138896602e-06, 198133104.590959, 33022184.0985032, 495332761.477403, 198133104.590992, 660443681.969899, 495332761.477445}, 
                                                   {0, 0, 0, 0, 0, 6987.75000000012, 0, 0, -184723.389583338, -120669.014583332, 0, 0, 1838014.87222228, 590962.613541664, 3133575.61076386, 0, 0, -9186030.29322955, -1335381.20833333, -30324784.8902774, -14723556.430382, 0, 0, 25727310.0848975, 1568569.63541657, 145673369.603816, 31718938.5295151, 133045153.788021, 0, 0, -42242502.9697965, -933760.277082866, -387236476.647385, -35388161.2256973, -574377598.846171, -263838026.117197, 0, 0, 40415580.6885502, 223290.50104076, 596781079.117343, 19947284.7597237, 1300853243.9852, 267762525.832495, 992526277.130216, 0, 0, -20877661.8474048, 8.52591616262238e-07, -530538230.474984, -4503025.10433758, -1588339764.07633, -135500119.048821, -1795069552.95744, -847387451.45317, 0, 0, 4503025.10434532, -3.71623663727986e-07, 252169405.843049, -3.8744822831152e-06, 990665522.954815, 27018150.626062, 1531028535.47565, 342229907.929906, 1103241150.56338, 0, 0, -1.19155257925164e-06, 5.40264899735513e-08, -49533276.1477422, 1.42845096183324e-06, -247666380.738703, -4.20044920321468e-06, -495332761.477409, -49533276.1477578, -495332761.477427, -247666380.738729}, 
                                                   {0, 0, 0, 0, 0, -3993.00000000008, 0, 0, 110262.258333336, 62679.008333332, 0, 0, -1167304.5618056, -239197.337499996, -1708472.52430553, 0, 0, 6303133.51250029, 387193.445833326, 17733560.4760413, 6281492.28437498, 0, 0, -19231586.4182305, -287878.662499929, -93123211.1906225, -9744200.62534774, -61424879.3602424, 0, 0, 34346138.8875037, 81196.5458330432, 273635742.647735, 6922005.5322929, 294425441.570478, 88771506.9217057, 0, 0, -35503189.6656315, 5.30423183195188e-07, -465560694.671862, -1860754.17534752, -753270505.264035, -58241605.6883825, -379593851.770835, 0, 0, 19649564.0916734, -4.85924756533887e-07, 453577437.782627, -1.86742032733041e-06, 1036514505.83537, 14327807.1501907, 805222761.839738, 216963936.845509, 0, 0, -4503025.10434411, 2.11539816915815e-07, -234157305.42569, 2.19170176751493e-06, -720484016.694413, -8.9967523547727e-06, -801538468.572539, -45030251.0434257, -355738983.24289, 0, 0, 9.05249469308172e-07, -3.22915569633658e-08, 49533276.1477425, -7.50757829900207e-07, 198133104.590963, 1.65004357437087e-06, 297199656.886444, 7.33220680854356e-06, 198133104.59097, 49533276.1477506}, 
                                                   {0, 0, 0, 0, 0, 1663.75000000003, 0, 0, -47315.2013888901, -23828.5972222213, 0, 0, 522392.235648166, 66901.2361111071, 671496.586342571, 0, 0, -2978045.48784736, -69341.4027777675, -7313486.21539328, -1821255.67199068, 0, 0, 9696220.84826453, 24605.0138888445, 40906245.6738412, 1815439.94143526, 18791464.232291, 0, 0, -18607541.7534741, 1.41860553251141e-07, -129799444.893399, -617995.932176176, -96899055.6136527, -17585818.551737, 0, 0, 20691586.4298644, -2.38968749370317e-07, 240756780.234252, -1.87237645365201e-07, 272042260.435749, 5582262.52604615, 82121284.2719887, 0, 0, -12280977.5572951, 2.12025148587682e-07, -256263065.028814, 9.70807687297176e-07, -417280326.33551, -5.50595700902045e-06, -196495640.916655, -23197402.0526687, 0, 0, 3002016.7362288, -9.18996883765841e-08, 144096803.338887, -9.48987675239642e-07, 324217807.512484, 2.78418477873776e-06, 228153271.953224, 7.75865992982052e-06, 45030251.0434037, 0, 0, -4.60052615834439e-07, 1.44743365005187e-08, -33022184.0984951, 3.04840190043922e-07, -99066552.2954817, -4.58585951734301e-07, -99066552.2954789, -2.4799236050193e-06, -33022184.0984932, -2.52380742815177e-06}, 
                                                   {0, 0, 0, 0, 0, -369.722222222238, 0, 0, 10062.606481482, 6876.83333333342, 0, 0, -103788.422222227, -35585.76388889, -186347.394444446, 0, 0, 534935.37013892, 53683.6666666709, 1909088.11550926, 954189.894675957, 0, 0, -1500905.84722234, -24605.0138888921, -9742970.3746527, -1397564.78888903, -9615024.30243076, 0, 0, 2300568.79861143, -1.45833044033435e-08, 26952742.2974533, 617995.932176204, 47869874.6879632, 13415473.7393529, 0, 0, -1786324.00833388, 3.85412742465091e-08, -40490010.8555546, -1.86521914217769e-07, -127647736.428819, -5582262.526044, -61777038.62153, 0, 0, 545821.224769078, -3.85277826551386e-08, 30565988.5870361, -9.17675927657128e-08, 181485557.235531, 2.43003643147855e-06, 144642624.563659, 23197402.0526664, 0, 0, -3.18049302374005e-07, 1.66670775295842e-08, -9006050.20868004, 1.91870777262092e-07, -126084702.921526, -1.24304533800923e-06, -162108903.75625, -3.34445668331644e-06, -45030251.0434059, 0, 0, 7.60419645931971e-08, 0, -1.02290871880797e-07, -7.29512459542256e-08, 33022184.0984948, 2.31096362716802e-07, 66044368.1969901, 9.91395752294462e-07, 33022184.0984961, 1.25170826671085e-06}, 
                                                   {0, 0, 0, 0, 0, -332.750000000008, 0, 0, 8324.29583333358, 7409.2333333334, 0, 0, -75096.129166669, -54720.7375000007, -184618.943055558, 0, 0, 316230.348958345, 165748.320833336, 1654519.42256945, 1353331.68437503, 0, 0, -655108.494791711, -199300.6125, -6894017.328993, -4041988.65659737, -11976182.9477432, 0, 0, 629273.230208455, 81196.5458333096, 14043619.2397566, 4729698.79479199, 48887086.9704861, 34911131.5189247, 0, 0, -223290.50104188, 5.63149581714258e-08, -13136924.4779508, -1860754.17534749, -96275421.0324642, -38889762.2647596, -136765431.888022, 0, 0, 2.25429411262982e-07, -5.7247168293953e-08, 4503025.1043397, -6.96548249949763e-08, 85148111.0638868, 14327807.150177, 250531942.16875, 139184412.315976, 0, 0, -1.28543956371216e-07, 2.56425638442235e-08, 2.57540683461356e-07, 2.45684873639297e-07, -27018150.62604, -2.02470031031467e-06, -193630079.48663, -45030251.0434065, -211642179.903995, 0, 0, 3.04722084269866e-08, 0, -4.03519305706837e-08, -1.05922034727373e-07, -4.59784427616333e-07, 4.63303209912168e-07, 49533276.1477421, 1.27221034504138e-06, 99066552.2954862, 49533276.1477441}, 
                                                   {0, 0, 0, 0, 0, -332.749999999996, 0, 0, 7409.23333333324, 8324.29583333325, 0, 0, -54720.737499999, -75096.1291666666, -184618.943055554, 0, 0, 165748.320833327, 316230.348958335, 1353331.68437497, 1654519.42256944, 0, 0, -199300.612499986, -655108.494791674, -4041988.65659703, -6894017.32899312, -11976182.947743, 0, 0, 81196.5458333277, 629273.230208339, 4729698.79479111, 14043619.2397572, 34911131.5189227, 48887086.9704866, 0, 0, -2.61121786260816e-08, -223290.501041659, -1860754.17534642, -13136924.4779517, -38889762.2647542, -96275421.0324676, -136765431.88802, 0, 0, 4.56757100829123e-08, -1.06550294027365e-08, -5.93792770493975e-07, 4503025.10434022, 14327807.15017, 85148111.0638926, 139184412.315968, 250531942.168751, 0, 0, -2.9826133838824e-08, 0, 2.0344682848474e-07, 2.93624117702869e-07, 2.17108797482312e-06, -27018150.6260444, -45030251.0433985, -193630079.486634, -211642179.903992, 0, 0, 0, 0, -2.28251456938821e-08, -1.4416244411877e-07, -4.82211871776374e-07, 7.68865450383591e-07, -1.56942082423811e-06, 49533276.1477442, 49533276.1477419, 99066552.2954863}, 
                                                   {0, 0, 0, 0, 0, -369.722222222204, 0, 0, 6876.83333333284, 10062.6064814812, 0, 0, -35585.763888884, -103788.422222221, -186347.394444436, 0, 0, 53683.6666666417, 534935.370138889, 954189.894675842, 1909088.11550921, 0, 0, -24605.0138888227, -1500905.84722224, -1397564.78888846, -9742970.3746527, -9615024.30243014, 0, 0, -1.02630190798363e-07, 2300568.79861116, 617995.932174872, 26952742.2974537, 13415473.7393496, 47869874.6879624, 0, 0, 9.64038594238951e-08, -1786324.0083334, 1.4347057650966e-06, -40490010.8555556, -5582262.52603672, -127647736.42882, -61777038.6215233, 0, 0, -5.47607513330314e-08, 545821.224768587, -1.09577966001226e-06, 30565988.5870368, -5.57217385307446e-06, 181485557.235536, 23197402.0526534, 144642624.563654, 0, 0, 1.79335324261995e-08, -4.79846519691557e-08, 4.37176839080575e-07, -9006050.20868013, 3.11397878290866e-06, -126084702.921531, 7.19310772209589e-06, -162108903.75625, -45030251.0433971, 0, 0, 0, 1.63903144782824e-08, -7.34725661892107e-08, -2.0899950235311e-07, -6.57463092962213e-07, 33022184.0984964, -2.29357332219852e-06, 66044368.1969913, -2.46125613623522e-06, 33022184.098494}, 
                                                   {0, 0, 0, 0, 0, 1663.74999999988, 0, 0, -23828.5972222202, -47315.2013888861, 0, 0, 66901.236111092, 522392.235648123, 671496.586342557, 0, 0, -69341.4027776779, -2978045.48784709, -1821255.67199043, -7313486.21539331, 0, 0, 24605.0138886159, 9696220.84826344, 1815439.94143362, 40906245.673842, 18791464.2322901, 0, 0, 4.53019091562064e-07, -18607541.7534712, -617995.932172171, -129799444.893402, -17585818.5517279, -96899055.6136551, 0, 0, -4.85256789646793e-07, 20691586.4298595, -5.03084326210423e-06, 240756780.234257, 5582262.52602452, 272042260.435767, 82121284.2719722, 0, 0, 3.3258255408382e-07, -12280977.5572903, 3.9636092770501e-06, -256263065.028816, 1.77115477479297e-05, -417280326.335544, -23197402.0526284, -196495640.91665, 0, 0, -1.33560310702657e-07, 3002016.73622621, -1.7461613972578e-06, 144096803.338885, -9.22887025626699e-06, 324217807.512512, -2.48900312057666e-05, 228153271.953237, 45030251.0433755, 0, 0, 2.361310689935e-08, 1.0871552174487e-07, 3.43937343978564e-07, -33022184.0984939, 1.87897833131432e-06, -99066552.29549, 7.04831607689707e-06, -99066552.295488, 1.04383863852968e-05, -33022184.0984872}, 
                                                   {0, 0, 0, 0, 0, -3992.99999999976, 0, 0, 62679.0083333292, 110262.258333328, 0, 0, -239197.337499961, -1167304.56180551, -1708472.52430548, 0, 0, 387193.445833127, 6303133.51249978, 6281492.28437436, 17733560.4760413, 0, 0, -287878.66249944, -19231586.4182284, -9744200.62534386, -93123211.190624, -61424879.36024, 0, 0, 81196.5458324288, 34346138.8874982, 6922005.53228341, 273635742.647742, 88771506.921684, 294425441.570483, 0, 0, 9.18055207546174e-07, -35503189.6656223, -1860754.17533601, -465560694.671872, -58241605.6883302, -753270505.264072, -379593851.770797, 0, 0, -5.84197324114342e-07, 19649564.0916643, -8.77598003650936e-06, 453577437.782632, 14327807.1501328, 1036514505.83545, 216963936.845415, 805222761.839728, 0, 0, 2.16531059940627e-07, -4503025.1043392, 3.74900447236322e-06, -234157305.425687, 2.20147176880991e-05, -720484016.694472, -45030251.0433475, -801538468.572567, -355738983.242829, 0, 0, -3.54000851776992e-08, -1.70254545467552e-07, -7.03467060157103e-07, 49533276.1477401, -4.57147151034718e-06, 198133104.590981, -1.65429564282607e-05, 297199656.886464, 49533276.1477217, 198133104.590958}, 
                                                   {0, 0, 0, 0, 0, 6987.7499999997, 0, 0, -120669.014583328, -184723.389583327, 0, 0, 590962.613541617, 1838014.87222217, 3133575.6107638, 0, 0, -1335381.20833307, -9186030.2932289, -14723556.4303811, -30324784.8902773, 0, 0, 1568569.63541596, 25727310.0848949, 31718938.5295091, 145673369.603819, 133045153.788017, 0, 0, -933760.277082288, -42242502.9697894, -35388161.2256822, -387236476.647395, -263838026.117163, -574377598.846179, 0, 0, 223290.501040784, 40415580.6885382, 19947284.7597051, 596781079.117358, 267762525.832409, 1300853243.98527, 992526277.130163, 0, 0, 3.88536708684658e-07, -20877661.8473928, -4503025.10432686, -530538230.474991, -135500119.048721, -1588339764.07644, -847387451.453028, -1795069552.95744, 0, 0, -6.18600937488141e-08, 4503025.10433892, -5.58872925667491e-06, 252169405.843045, 27018150.6260057, 990665522.954906, 342229907.929778, 1531028535.4757, 1103241150.5633, 0, 0, 0, 2.05220348885773e-07, 9.9128854539447e-07, -49533276.1477389, 7.7172699443975e-06, -247666380.738729, -49533276.1477167, -495332761.477443, -247666380.738686, -495332761.477416}, 
                                                   {0, 0, 0, 0, 0, -9316.99999999977, 0, 0, 177973.186111108, 229216.686111105, 0, 0, -1082918.85370367, -2061771.37731477, -4272047.23101844, 0, 0, 3225083.5553239, 9181137.25069422, 24765132.8807863, 36921650.0761571, 0, 0, -5355691.3564811, -23093650.9107631, -69721792.5644629, -155433768.529977, -196706218.827197, 0, 0, 5079295.03379606, 34535597.4944424, 108916144.396978, 365370923.492362, 498894132.196043, 725892608.830792, 0, 0, -2580245.78981529, -30516368.4756911, -96759217.1180367, -505852225.082059, -689099296.270198, -1423476944.14066, -1543061412.47624, 0, 0, 545821.224769506, 14737173.068747, 45848982.8805403, 410184650.413534, 529173677.413008, 1537032568.94821, 1683858478.41079, 2341573054.25695, 0, 0, -7.04412472609494e-07, -3002016.73622551, -9006050.20867427, -180121004.173601, -210141171.535837, -864580820.033384, -918617121.285333, -1726159623.33047, -1756179790.69267, 0, 0, 1.84788456069827e-07, -2.01880003082524e-07, -1.05371033046799e-06, 33022184.0984912, 33022184.0984857, 198133104.590987, 198133104.590943, 495332761.477449, 495332761.477406, 660443681.969903}, 
                                                   {0, 0, 0, 0, 0, 6211.33333333333, 0, 0, -141423.67962963, -141423.679629629, 0, 0, 1115237.50555556, 1115237.50555555, 3169939.17777776, 0, 0, -4076156.07361114, -4076156.07361111, -24318104.5148145, -24318104.5148149, 0, 0, 7922814.47222247, 7922814.47222217, 85230537.8604142, 85230537.8604179, 177350479.609722, 0, 0, -8498571.79722324, -8498571.79722187, -157476189.724529, -157476189.72454, -572453691.801843, -572453691.801861, 0, 0, 4763530.68889103, 4763530.68888814, 159578278.077765, 159578278.077781, 950473232.767328, 950473232.767395, 1591316970.75694, 0, 0, -1091642.44953948, -1091642.44953631, -84056468.6143413, -84056468.6143498, -844385434.716851, -844385434.716948, -2120515458.22566, -2120515458.22573, 0, 0, 1.44276507171176e-06, -2.82824466880065e-07, 18012100.4173568, 18012100.4173565, 378254108.764552, 378254108.764617, 1350907531.30203, 1350907531.30212, 1981331045.90972, 0, 0, -3.47173229264693e-07, 1.70994624763153e-08, 6.5300190841821e-07, 2.03901045093899e-06, -66044368.196983, -66044368.1969997, -330221840.984934, -330221840.98497, -660443681.969898, -660443681.969915}, 
                                                   {0, 0, 0, 0, 0, 1331.00000000004, 0, 0, -34923.9611111125, -26383.377777778, 0, 0, 341081.074074088, 158366.81666667, 687340.724074081, 0, 0, -1632058.0273149, -349838.561111123, -6637165.21620372, -4063182.52083344, 0, 0, 4166449.01851884, 324786.183333336, 31226633.2099535, 8696232.07546348, 38273509.1877322, 0, 0, -5719845.56203789, -108262.061111038, -77804334.5851839, -7767802.88472326, -173305005.242825, -77619386.8974575, 0, 0, 3969608.90740886, -1.73882173209387e-07, 103309071.815275, 2481005.56713038, 407480354.345367, 64754245.3020922, 321290220.943294, 0, 0, -1091642.44953855, 1.73468124559079e-07, -68773474.3208305, 3.37713252788113e-07, -496697314.539345, -19103742.8669081, -647343972.575466, -237432232.77432, 0, 0, 8.56621039979656e-07, -7.65359522315467e-08, 18012100.4173597, -7.87759152331476e-07, 294197640.150225, 5.52402278644903e-06, 594399313.772913, 60040334.7245496, 378254108.764592, 0, 0, -2.0409029399325e-07, 1.07969553110406e-08, 2.5646474046907e-07, 3.1316002411526e-07, -66044368.1969889, -1.13677053220696e-06, -198133104.590969, -3.98275850408331e-06, -198133104.590974, -66044368.1969949}, 
                                                   {0, 0, 0, 0, 0, 1330.99999999995, 0, 0, -26383.3777777766, -34923.9611111103, 0, 0, 158366.816666655, 341081.07407407, 687340.724074052, 0, 0, -349838.561111049, -1632058.02731481, -4063182.52083312, -6637165.21620361, 0, 0, 324786.183333171, 4166449.01851855, 8696232.07546178, 31226633.2099536, 38273509.1877305, 0, 0, -108262.06111088, -5719845.5620371, -7767802.8847192, -77804334.5851856, -77619386.8974477, -173305005.242824, 0, 0, -1.75572432196201e-07, 3969608.90740748, 2481005.56712544, 103309071.815278, 64754245.302069, 407480354.345376, 321290220.943276, 0, 0, 5.8581655043615e-08, -1091642.44953711, 3.1820541006981e-06, -68773474.3208327, -19103742.8668813, -496697314.539364, -237432232.774282, -647343972.575458, 0, 0, 0, 7.28301845965801e-08, -1.22731216082981e-06, 18012100.4173597, -9.67502330106166e-06, 294197640.150242, 60040334.7245158, 594399313.77292, 378254108.76457, 0, 0, 0, -3.32412029670446e-08, 1.91407769477407e-07, 6.6232893381499e-07, 2.07902575236806e-06, -66044368.196994, 7.07022596968696e-06, -198133104.590975, -66044368.196984, -198133104.59097}, 
                                                   {0, 0, 0, 0, 0, -3327.50000000006, 0, 0, 82734.5902777798, 70533.7569444447, 0, 0, -745233.678240762, -486609.902777782, -1735127.95601853, 0, 0, 3203218.64525475, 1418982.33506946, 15356511.0925926, 11726898.7407409, 0, 0, -7217470.74074122, -2014535.51215277, -64261632.4198487, -33008138.7364016, -100217246.778358, 0, 0, 8773737.8692143, 1387107.65798592, 139455067.468747, 44889284.8180001, 396571823.958621, 264734571.310772, 0, 0, -5458212.24768756, -372150.835068997, -161885613.255202, -29461941.1096663, -791316725.635984, -331958544.881966, -937758079.235832, 0, 0, 1364553.0619238, -4.45975847809293e-07, 95518714.3344848, 7505041.84056628, 821460943.276599, 197860193.978615, 1544674066.0949, 1007722436.22891, 0, 0, -1.42704358081211e-06, 1.92388397707267e-07, -22515125.5216988, 2.18730843381186e-06, -420282343.071743, -45030251.0434192, -1170786527.12846, -487827719.636893, -1215816778.17189, 0, 0, 3.393073651233e-07, -2.36998857432362e-08, -4.23700507670644e-07, -9.21171268580758e-07, 82555460.2462341, 3.81718782587758e-06, 330221840.984945, 82555460.2462486, 495332761.47743, 330221840.984962}, 
                                                   {0, 0, 0, 0, 0, 1247.81250000001, 0, 0, -29309.7291666669, -29309.7291666666, 0, 0, 239161.751736113, 239161.751736112, 683831.289930556, 0, 0, -872359.583333342, -872359.58333334, -5514272.53501154, -5514272.53501162, 0, 0, 1507057.1006945, 1507057.10069445, 19720918.631944, 19720918.6319448, 43532420.8229169, 0, 0, -1217948.1875002, -1217948.18749995, -33031205.9369198, -33031205.9369223, -150100836.811341, -150100836.811345, 0, 0, 372150.835069861, 372150.835069297, 25678407.6197892, 25678407.6197926, 236687931.10416, 236687931.104176, 484168236.425348, 0, 0, -4.72634430712414e-07, 1.51566809989655e-07, -7505041.84056506, -7505041.84056687, -169886856.209191, -169886856.209214, -676136042.181995, -676136042.182012, 0, 0, 2.7614186923788e-07, -6.02838589062482e-08, -7.99254817230831e-07, -9.83104936122797e-07, 45030251.0433959, 45030251.0434113, 405272259.390614, 405272259.390636, 720484016.694445, 0, 0, -6.56465965130453e-08, 0, 1.06075232774731e-07, 4.50880651840594e-07, 1.67975907036552e-06, -2.18285825067542e-06, -82555460.2462339, -82555460.2462428, -247666380.738713, -247666380.738718}, 
                                                   {0, 0, 0, 0, 0, -3327.49999999994, 0, 0, 70533.7569444432, 82734.5902777765, 0, 0, -486609.902777764, -745233.678240735, -1735127.95601849, 0, 0, 1418982.33506937, 3203218.64525463, 11726898.7407404, 15356511.0925925, 0, 0, -2014535.51215261, -7217470.74074075, -33008138.7363984, -64261632.4198499, -100217246.778355, 0, 0, 1387107.657986, 8773737.8692129, 44889284.817992, 139455067.468752, 264734571.310754, 396571823.958625, 0, 0, -372150.835069632, -5458212.24768498, -29461941.1096561, -161885613.25521, -331958544.881918, -791316725.636013, -937758079.235809, 0, 0, 4.045000560437e-07, 1364553.0619211, 7505041.84056067, 95518714.3344895, 197860193.978555, 821460943.276651, 1007722436.22884, 1544674066.09491, 0, 0, -2.84355280255815e-07, 3.29408427353139e-08, 2.49116591799305e-06, -22515125.5216984, -45030251.0433825, -420282343.071783, -487827719.636824, -1170786527.12849, -1215816778.17186, 0, 0, 7.26990408775038e-08, 2.61823916826221e-08, -3.67901346507803e-07, -1.39505846940058e-06, -4.57653778784126e-06, 82555460.2462452, 82555460.2462242, 330221840.984962, 330221840.984943, 495332761.47743}};



//---------------------------------------------------------------------
// Tri

/* 2
   |\
   0-1 */
    
/* 2
   |\
   5 4
   |  \
   0-3-1*/
       
/* 
   2
   |\
   7 6
   |  \
   8 9 5
   |    \
   0-3-4-1 */       
       
/* 2
   | \
   9  8 
   |    \
   10 14 7
   |      \
   11 12 13 6
   |         \   
   0__3__4__5__1 */
       
/*  2
    |\
    | \
    11  10
    |   \
    |    \
    12 17  9 
    |      \
    |       \
    13 20 19  8
    |         \
    |          \
    14 15 18 16  7
    |            \
    |             \   
    0__3__4__5__6__1 */
   
/* 2					   
   | \
   13 12
   |    \
   14 20 11
   |       \
   15 25 24 10 
   |          \
   16 26 27 23 9
   |             \
   17 18 21 22 19 8
   |                \   
   0__3__4__5__6__7__1 */     
      
   
/* 2			    		     
   | \
   15 14
   |    \
   16 23 13
   |       \
   17 30 29 12 
   |          \
   18 31 35 28 11
   |             \
   19 32 33 34 27 10
   |                \
   20 21 24 25 26 22 9
   |                   \ 
   0__3__4__5__6__7__8__1    
*/

void MAd_TriInterpolant::nodePositionTriangle(int o, double *u, double *v){ 
  int nb = size(o);
  if (o ==0) {u[0]=0.;v[0]=0.; return;}     
   
  u[0]= 0.;    v[0]= 0.;
  u[1]= o; v[1]= 0;
  u[2]= 0.;    v[2]= o;
  // edges
  for (int k=0; k< (o-1); k++){
    u[3+k] = k+1; v[3+k]=0.;
    u[3+o-1+k] = o -1 -k ; v [3+o-1+k] = k+1;
    u[3+2*(o-1)+k] = 0. ; v [3+2*(o-1)+k] = o -1 -k;
  }
  if (o >2){
    double *utemp;
    double *vtemp;
    int nbdoftemp = size(o-3);
    utemp = new double[nbdoftemp];
    vtemp = new double[nbdoftemp];
    nodePositionTriangle(o-3,utemp, vtemp);
    for (int k=0; k< nbdoftemp; k++){
      u[3+k+3*(o-1)] = utemp[k]*(o-3)+1.;
      v[3+k+3*(o-1)] = vtemp[k]*(o-3)+1.;
    }
    delete []utemp;
    delete []vtemp;
  }
  for (int k=0; k< nb; k++){
    u[k] = u[k]/o;
    v[k] = v[k]/o;
  }
  return;     
}

int MAd_TriInterpolant::P[78][2] = {
  //order 0 :    
  {0,0},
  //order 1:   
  {1,0},	
  {0,1},
  // order 2:
  {2,0},	
  {0,2},
  {1,1},
  // order 3:
  {3,0},
  {0,3},
  {2,1},
  {1,2},
  //order 4:
  {4,0},
  {0,4},
  {3,1},
  {1,3},
  {2,2},
  // order 5:
  {5,0},
  {0,5},
  {4,1},
  {1,4},
  {3,2},
  {2,3},
  //order 6
  {6,0},
  {0,6},
  {5,1},
  {1,5},
  {4,2},
  {2,4},
  {3,3},
  //order 7
  {7,0},
  {0,7},
  {6,1},
  {1,6},
  {5,2},
  {2,5},
  {4,3},
  {3,4},  
  //order 8
  {8,0},
  {0,8},
  {7,1},
  {1,7},
  {6,2},
  {2,6},
  {5,3},
  {3,5},
  {4,4},
  //order 9
  {9,0},
  {0,9},
  {8,1},
  {1,8},
  {7,2},
  {2,7},
  {6,3},
  {3,6},
  {5,4},
  {4,5},
  // order 10
  {10,0},
  {0,10},
  {9,1},
  {1,9},
  {8,2},
  {2,8},
  {7,3},
  {3,7},
  {6,4},
  {4,6},
  {5,5},
  // order 11
  {11,0},
  {0,11},
  {10,1},
  {1,10},
  {9,2},
  {2,9},
  {8,3},
  {3,8},
  {7,4},
  {4,7},
  {6,5},
  {5,6} 
};
  
